#! /usr/bin/perl -w
#
# Generate documentation from XML Docbook sources using XSL transformations.
# Note: configuration files that are expected:
#       - docs_config ... contains information needed for launching external
#                         tools such as Saxon, FOP, etc.
#       - docs_makefile ... describes all the docs stuff that should be
#                         processed
#
# Note: launch perl docgen.pl -h for usage information.
#

use warnings;
use strict;

use Cwd;
use Getopt::Std;
use File::Copy;
use File::Path;
use File::stat;
use File::Temp qw/ tempfile tempdir /;


###
# Declare all global variables.
my $tmpfile_extension;
    # Extension of temporary files.
my %config_values = ();
    # Values obtained from the configuration file.
my %options = ();
    # Array that will receive the command-line parameters.
my @formats_processed = ();
    # Contains all format string that should be processed.
my @formats_ignored = ();
    # Contains all format string that should be ignored. Has greater priority
    # that the @formats_processed list.
my @books = ();
    # Array of references to the book_description structures.
my $dtd_reference = "http://www.oasis-open.org/docbook/xml/4.2/docbookx.dtd";
    # Reference to the docbook DTD.
my $tag_sequence = '\$\$';
    # Sequence that encloses tags.
my $cmd_separator = ';';
    # Separator that is used to split commands in the commands list.


###
# Reads the configuration file. If some error occures, prints the error
# message.
#
# Parameters:
#   - filename Name of the configuration file, including path (if needed).
#   - config_values Reference to a hash that will be filled with
#     configuration values.
#
sub read_config_file
    {
    my( $filename, $r_config_values ) = @_;

    my( $line, $line_part ) = ( "", "" );
    my( @config_item );

    open( FCONF, $filename ) or error( "Can't open ".$filename."." );

    # todo - more strict syntax checking

    LINE : while( $line = <FCONF> )
        {
        # Concatenate lines ending with '\'.

        chomp( $line );
        if( $line =~ s/\s*\\\s*$/ / )
            {
            my $next_line = <FCONF>;
            if( defined( $next_line ) )
                {
                chomp( $next_line );
                $next_line =~ s/^\s*//;
                $line .= $next_line;
                redo LINE;
                }
            }

        @config_item = parse_assignment( $line );

        if( $config_item[ 0 ] && $config_item[ 1 ] )
            {
            $$r_config_values{ $config_item[ 0 ] } = $config_item[ 1 ];
            }
        }

    close( FCONF );
    }  # read_config_file()


###
# Verifies config values - in this version just tests whether all mandatory
# values are specified
#
# Parameters:
#   - r_config_values Reference to the hash of configuration values.
#
# Returns:
#   1 (true) on success, otherwise 0 (false).
#
sub verify_config_values
    {
    my( $r_config_values ) = @_;
    my( @mandatory_settings ) =
        (
        "DOCBOOK2HTML_COMMAND",
        "DOCBOOK2CHUNK_COMMAND",
        "DOCBOOK2PDF_COMMAND",
        "VALIDATE_COMMAND",
        "DOCBOOK_HTML_XSL",
        "DOCBOOK_CHUNK_XSL",
        "DOCBOOK_FO_XSL",
        );
    my( $setting );

    foreach $setting ( @mandatory_settings )
        {
        if( !exists( $$r_config_values{ $setting } ) ||
            !defined( $$r_config_values{ $setting } ) )
            {
            error( $setting." unspecified in the configuration." );
            }

        }

    if( exists( $$r_config_values{ "DTD_REFERENCE" } ) )
        {
        $dtd_reference = $$r_config_values{ "DTD_REFERENCE" };
        }

    if( exists( $$r_config_values{ "TAG_SEQUENCE" } ) &&
        defined( $$r_config_values{ "TAG_SEQUENCE" } ) )
        {
        $tag_sequence = $$r_config_values{ "TAG_SEQUENCE" };
        }

    if( exists( $$r_config_values{ "COMMAND_SEPARATOR" } ) &&
        defined( $$r_config_values{ "COMMAND_SEPARATOR" } ) )
        {
        $cmd_separator = $$r_config_values{ "COMMAND_SEPARATOR" };
        }
    }  # verify_config_values()



###
# Checks whether the SGML_CATALOG_FILES environment variable is set; if
# not, it sets it appropriately according to the configuration file settings.
#
# Parameters:
#   - r_config_values Reference to the hash of configuration values.
#
sub set_catalog
    {
    my( $r_config_values ) = @_;

    if( defined( $ENV{ "SGML_CATALOG_FILES" } ) )
        {
        return 0;
            # OK - user is responsible for setting this variable correctly
        }

    # SGML_CATALOG_FILES is not defined

    if( !defined( $$r_config_values{ "SGML_CATALOG_FILES" } ) )
        {
        error( "SGML_CATALOG_FILES not defined; either set appropriate environment variable\n"
               ." or define this variable in the configuration file.\n" );
        }

    # SGML_CATALOG_FILES was defined in the configuration

    $ENV{ "SGML_CATALOG_FILES" } = $$r_config_values{ "SGML_CATALOG_FILES" };
    print(
        "SGML_CATALOG_FILES set to ".$$r_config_values{ "SGML_CATALOG_FILES" }."\n"
        );
    }  # set_catalog()



###
# Read makefile.
#
# Parameters:
#   - filename Name of the makefile.
#   - rbooks Reference to the books array that contains information about
#     what should be processed.
#
sub read_makefile
    {
    my( $filename, $rbooks ) = @_;

    my( $line );
    my( $line_number ) = ( 0 );
    my( $book_open ) = ( 0 );
    my( $rbook_desc );
    my( @assignment );

    open( FMAKEFILE, $filename ) or error( "Can't open $filename." );

    while( $line = <FMAKEFILE> )
        {
        $line_number++;

        chomp( $line );
        while( $line =~ /\\\s*$/ )
            {
            # $line ends by '\' - "to be continued"

            $line =~ s/^\s*(\S.*\S)\s*\\\s*$/$1 /;
            $line .= <FMAKEFILE>;
            chomp( $line );
            $line_number++;
            }

        if( ( $line =~ /^\s*#.*$/ ) || ( $line =~ /^\s*$/ ) )
            {
            # The line is comment or empty
            next;
            }
        elsif( $line =~ /^\s*book\s*(.*\S)\s*$/ )
            {
            # The line starts description of a new book.

            if( $book_open )
                {
                error_file
                    (
                    $filename,
                    $line_number,
                    "Nested book definitions."
                    );
                }

            $book_open = 1;
            $rbook_desc = { "name" => $1 };
            next;
            }
        elsif( $line =~ /^\s*endbook\s*$/ )
            {
            # The line ends description of a book.

            $book_open = 0;
            validate_book_desc( $rbook_desc );
            push( @$rbooks, $rbook_desc );
            next;
            }
        elsif( $line =~ /^\s*\S.*\S\s*=\s*\S.*\S\s*$/ )
            {
            @assignment = parse_assignment( $line );
            $rbook_desc->{ $assignment[ 0 ] } = $assignment[ 1 ];
            next;
            }
        else
            {
            error_file( $filename, $line_number, "Invalid line: ".$line );
            }
        }  # while( $line = <FMAKEFILE> )

    if( $book_open )
        {
        error_file
            (
            $filename,
            $line_number,
            "Book ".$rbook_desc->{ "name" }." not closed."
            );
        }

    close( FMAKEFILE );
    }  # read_makefile()


###
# Validates book description - i.e. checks whether all the mandatory fields
# are given and eventually changes the description in some ways. In case the
# description is invalid the routine prints the error message and quits.
#
# Parameters:
#   - rbook_desc Reference to the book description structure.
#
sub validate_book_desc
    {
    my( $rbook_desc ) = @_;
    my( @mandatory_settings ) =
        (
        "format", "path", "output_path", "name", "output_name", "copy_images"
        );
    my( $key );
    my( $error_msg );

    # Check whether all mandatory information are given
    foreach $key ( @mandatory_settings )
        {
        if( !exists( $$rbook_desc{ $key } ) ||
            !defined( $$rbook_desc{ $key } ) )
            {
            $error_msg = $key." unspecified in the book ";
            $error_msg .= $$rbook_desc{ "name" }." description.";
            error( $error_msg );
            }
        }

    # Niether path nor output_path may end with slash
    if( $$rbook_desc{ "path" } =~ /^.*[\/\\]\s*$/ )
        {
        chop( $$rbook_desc{ "path" } );
        }

    if( $$rbook_desc{ "output_path" } =~ /^.*[\/\\]\s*$/ )
        {
        chop( $$rbook_desc{ "output_path" } );
        }
    }  # validate_book_desc()


###
# Preprocesses a xslt file so that it replaces the $$DTD_REFERENCE$$ tag
# with the value assigned to DTD_REFERENCE in the configuration file.
#
# Parameters:
#   - rbook_desc Description of the book that should be generated.
#
# Returns:
#   Name of the temporary file containing the copy of the preprocessed xslt
#   file with $$DTD_REFERENCE$$ replaced.
#
sub preprocess_main_xslt
    {
    my( $rbook_desc ) = @_;
    my( $fh, $dest );

    open( FSOURCE, $rbook_desc->{ "path" }.'/'.$rbook_desc->{ "name" } ) or
        error( "Cannot open ".$rbook_desc->{ "path" }.'/'.$rbook_desc->{ "name" }."\n" );

    ( $fh, $dest ) = File::Temp::tempfile(
                                    $rbook_desc->{ "name" }.".XXXX",
                                    DIR => $rbook_desc->{ "path" }
                                    );

    while( my $line = <FSOURCE> )
        {
        chomp( $line );

        if( $line =~ /^(.*)($tag_sequence)DTD_REFERENCE($tag_sequence)(.*)$/ )
            {
            $line = $1.'"'.$dtd_reference.'"'.$4."\n";
            print( $fh $line );
            }
        else
            {
            print( $fh $line."\n" );
            }
        }

    close( FSOURCE );
    close( $fh );

    return $dest;
    } # preprocess_main_xslt()


###
# Preprocesses a user-defined XSLT stylesheet.
# The stylesheet is copied into a temporary file (its name is the same as
# the name of the original stylesheet, but with some suitable extension
# appended) and
#   - $$IMPORTED_XSLT$$ is replaced with correct docbook template
#     (according to the configuration file)
#   - if there is a CSS stylesheet attribute defined in the makefile,
#     appends a line that will cause inserting link to this stylesheet
#     to each html page generated.
# Called from generate_book.
#
# Parameters:
#   - rbook_desc Description of the book to generate.
#
# Returns:
#   Name of the temporary file containing the template to be used for
#   the transformation or undef.
#
sub preprocess_user_xslt
    {
    my( $rbook_desc ) = @_;
    my( $imported_style );
    my( $fh );

    if( !defined( $rbook_desc ) or !defined( $rbook_desc->{ "user_xslt" } ) )
        {
        if( defined( $rbook_desc->{ "stylesheet" } ) )
            {
            error( "Using stylesheet requires a user-predefined style." );
                # hack
            }

        return undef;
        }

    my( $source ) = ( $rbook_desc->{ "user_xslt" } );
    my( $dest );

    if( $rbook_desc->{ "format" } eq "html" )
        {
        $imported_style = $config_values{ "DOCBOOK_HTML_XSL" };
        }
    elsif( $rbook_desc->{ "format" } eq "chunk" )
        {
        $imported_style = $config_values{ "DOCBOOK_CHUNK_XSL" };
        }
    elsif( $rbook_desc->{ "format" } eq "pdf" )
        {
        $imported_style = $config_values{ "DOCBOOK_FO_XSL" };
        }
    else
        {
        error( "Unknown format (".$rbook_desc->{ "format" }.")\n" );
        }

    open( FSOURCE, $source ) or error( "Cannot open ".$source."\n" );
    ( $fh, $dest ) = create_temporary_file( $source );

    while( my $line = <FSOURCE> )
        {
        chomp( $line );

        if( $line =~ /^(.*)($tag_sequence)IMPORTED_XSLT($tag_sequence)(.*)$/ )
            {
            $line = $1.'"file:///'.$imported_style.'"'.$4."\n";

            print( $fh $line );
            }
        elsif( $line =~ /^(.*)<\/xsl:stylesheet>.*$/ )
            {
            print( $fh $1."\n" );

            if( defined( $rbook_desc->{ "stylesheet" } ) )
                {
                # define global XSLT variable that makes the processor
                # put reference to user-defined CSS stylesheet into each
                # generated html page -->

                if( $rbook_desc->{ "format" } eq "pdf" )
                    {
                    print( "Warning: CSS stylesheets don't make sense in PDF format.\n\n" );
                    }
                else
                    {
                    print( $fh '<xsl:param name="html.stylesheet">' );
                    print( $fh $rbook_desc->{ "stylesheet" } );
                    print( $fh "</xsl:param>\n" );
                    }
                }

            print( $fh "<\/xsl:stylesheet>\n" );
            }
        else
            {
            print( $fh $line."\n" );
            }
        }

    close( FSOURCE );
    close( $fh );

    return $dest;
    }  # preprocess_user_xslt()


###
# Preprocesses the command string from the configuration. Replaces all the
# $$tagname$$ tags.
#
# Parameters:
#   - command    The command string.
#   - source     The source xslt file name with path relatively to the docs root.
#   - xslt_style Path to the xslt style that should be used for the transormation.
#   - rbook_desc Description of the book.
#
# Returns:
#   The result command line.
#
sub preprocess_command_pattern
    {
    my( $command, $source, $xslt_style, $rbook_desc ) = @_;

    my( $source_plain, $source_path, $source_name, $source_ext );
    my( $target, $target_name, $target_ext );

    $source = get_absolute_path_file( $source );

    if( $source =~ /^(.*)[\/\\]([^\/\\]*)$/ )
        {
        $source_path = $1;
        $source_plain = $2;
        }
    else
        {
        $source_path = getcwd();
        $source_plain = $source;
        }

    if( $source_plain =~ /^(.*)[\.]([^\.]*)$/ )
        {
        $source_name = $1;
        $source_ext = $2;
        }
    else
        {
        $source_name = $source_plain;
        $source_ext = "";
        }

    $target = get_absolute_path( $rbook_desc->{ "output_path" } );

    if( $rbook_desc->{ "output_name" } =~ /^(.*)[\.]([^\.]*)$/ )
        {
        $target_name = $1;
        $target_ext = $2;
        }
    else
        {
        $target_name = $rbook_desc->{ "output_name" };
        $target_ext = "";
        }

    $xslt_style = get_absolute_path_file( $xslt_style );

    $command =~ s/($tag_sequence)SOURCE($tag_sequence)/$source/g;
    $command =~ s/($tag_sequence)SOURCE_PATH($tag_sequence)/$source_path/g;
    $command =~ s/($tag_sequence)SOURCE_PLAIN($tag_sequence)/$source_plain/g;
    $command =~ s/($tag_sequence)SOURCE_NAME($tag_sequence)/$source_name/g;
    $command =~ s/($tag_sequence)SOURCE_EXT($tag_sequence)/$source_ext/g;
    $command =~ s/($tag_sequence)TARGET($tag_sequence)/$target\/$rbook_desc->{ "output_name" }/g;
    $command =~ s/($tag_sequence)TARGET_PATH($tag_sequence)/$target/g;
    $command =~ s/($tag_sequence)TARGET_PLAIN($tag_sequence)/$rbook_desc->{ "output_name" }/g;
    $command =~ s/($tag_sequence)TARGET_NAME($tag_sequence)/$target_name/g;
    $command =~ s/($tag_sequence)TARGET_EXT($tag_sequence)/$target_ext/g;
    $command =~ s/($tag_sequence)XSLT_STYLE($tag_sequence)/$xslt_style/g;

    return $command;
    }  # preprocess_command_pattern()


###
# If the stylesheet has been specified in the makefile, copies the relevant
# file from the source to destination tree.
#
# Parameters:
#   - rbook_desc Description of the book.
#
sub copy_css_stylesheet_to_output
    {
    my( $rbook_desc ) = @_;
    my( $stylesheet_path );
    my( $stylesheet_name );
    my( $output_stylesheet_path );

    if( !defined( $rbook_desc->{ "stylesheet" } ) )
        {
        return;
        }

    $stylesheet_path = $rbook_desc->{ "stylesheet" };

    # Extract stylesheet path and plain name
    if( $stylesheet_path =~ /\// )
        {
        $stylesheet_path =~ s/\/([^\/]*)$//;
        $stylesheet_path .= "/";
            # we must append '/' in case the script is run on Win32 platform
            # and the stylesheet is in the root directory and is specified
            # using the absolute path (merge_paths must get for example e:/,
            # not just e:

        $stylesheet_name = $1;

        $output_stylesheet_path = merge_paths
                                (
                                $rbook_desc->{ "output_path" },
                                $stylesheet_path
                                );

        $stylesheet_path = merge_paths
                                (
                                $rbook_desc->{ "path" },
                                $stylesheet_path
                                );

        if( !defined( $output_stylesheet_path ) or !defined( $stylesheet_path ) )
            {
            error( "The CSS stylesheet cannot be copied into the destination." );
            }
        }
    else
        {
        $stylesheet_name = $stylesheet_path;
        $stylesheet_path = $rbook_desc->{ "path" };

        $output_stylesheet_path = $rbook_desc->{ "output_path" };
        }

    my( $stylesheet_in, $stylesheet_out ) =
                            ( $stylesheet_path, $output_stylesheet_path );

    if( ( length( $stylesheet_in ) > 0 ) &&
        !( substr( $stylesheet_in, length( $stylesheet_in ) - 1, 1 ) eq "/" ) )
        {
        $stylesheet_in .= '/'
        }

    if( ( length( $stylesheet_out ) > 0 ) &&
        !( substr( $stylesheet_out, length( $stylesheet_out ) - 1, 1 ) eq "/" ) )
        {
        $stylesheet_out .= '/';
        }

    $stylesheet_in .= $stylesheet_name;
    $stylesheet_out .= $stylesheet_name;

    if( $stylesheet_in eq $stylesheet_out )
        {
        return;
        }

    print( "Copying CSS stylesheet...\n" );
    print( "[$stylesheet_in ==> $stylesheet_out]\n\n" );

    if( !( -e $stylesheet_in ) )
        {
        error( "The $stylesheet_in doesn't exist." );
        }

    if( compare_files_timestamps( $stylesheet_in, $stylesheet_out ) == 1 )
        {
        copy_file( $stylesheet_in, $stylesheet_out );
        }
    }  # copy_css_stylesheet_to_output()


###
# Copies images from the source to the destination tree according to the
# settings and makefile.
#
# Parameters:
#   - rbook_desc Description of the book.
#
sub copy_images_to_output
    {
    my( $rbook_desc ) = @_;
    my( @extension_list, @imagedirs_list );

    if( $rbook_desc->{ "copy_images" } eq "true" )
        {
        if( !defined( $rbook_desc->{ "image_extensions" } ) ||
            !defined( $rbook_desc->{ "image_directories" } ) )
            {
            print( "Warning: IMAGE_EXTENSIONS or IMAGE_DIRECTORIES not ".
                   "defined in the book description (book ".
                   $rbook_desc->{ "path" }."/".$rbook_desc->{ "name" }.")".
                   "; images won't be copied.\n\n" );
            }
        else
            {
            @extension_list =
                split( /\s+/, $rbook_desc->{ "image_extensions" } );

            @imagedirs_list =
                split( /\s+/, $rbook_desc->{ "image_directories" } );

            foreach my $imagedir ( @imagedirs_list )
                {
                print( "Copying image files...\n" );
                print( "[".$rbook_desc->{ "path" }."/".$imagedir." ==> " );
                print( $rbook_desc->{ "output_path" }."/".$imagedir."]\n\n" );
                actualize_directory(
                    $rbook_desc->{ "path" }."/".$imagedir,
                    $rbook_desc->{ "output_path" }."/".$imagedir,
                    \@extension_list
                    );
                }
            }
        }
    }  # copy_images_to_output()


###
# Generates one book of the documentation
#
# Parameters:
#   - rbook_desc Description of the book to generate.
#   - rformats_processed Array containing the format string ("html", "pdf" or
#     "chunk") for all formats that should be processed.
#   - rformats_ignored Array containing the format string for all formats that
#     should be ignored.
#
sub generate_book
    {
    my( $rbook_desc, $rformats_processed, $rformats_ignored ) = @_;
    my( $cmd_string );
    my( $current_dir );
    my( $full_source_dir );
    my( $actual_xslt ) = ( undef );
        # The XLST stylesheet that will be actualy used for the transformation.

    if( ( find_in_array( $rformats_processed, $rbook_desc->{ "format" } ) == -1 ) ||
        ( find_in_array( $rformats_ignored, $rbook_desc->{ "format" } ) >= 0 ) )
        {
        print "Book ".$rbook_desc->{ "path" }."/".$rbook_desc->{ "name" };
        print " [".$rbook_desc->{ "format" }."] ignored.\n\n";
        return 0;
        }
    else
        {
        print( "\n\nTransforming ".$rbook_desc->{ "path" }."/" );
        print( $rbook_desc->{ "name" }." ==> " );
        print( $rbook_desc->{ "output_path" }."/" );
        print( $rbook_desc->{ "output_name" }."\n\n" );
        }

    # Create the target directory
    create_path( $rbook_desc->{ "output_path" } );

    # Find absolute path for user-defined style
    if( defined( $rbook_desc->{ "user_xslt" } ) )
        {
        $current_dir = getcwd();
        $rbook_desc->{ "user_xslt" } =~ /^(.*\/)([^\/]*)$/;

        my( $user_xslt_name ) = ( $2 );
        chdir( $1 );
        $rbook_desc->{ "user_xslt" } = getcwd();
        chdir( $current_dir );

        if( $rbook_desc->{ "user_xslt" } =~ /[^\/]$/ )
            {
            $rbook_desc->{ "user_xslt" } .= '/';
            }

        $rbook_desc->{ "user_xslt" } .= $user_xslt_name;

        # Preprocess user-defined XSL styles$rbook_desc );
        }

    # Copy CSS stylesheet if needed
    copy_css_stylesheet_to_output( $rbook_desc );

    # Preprocess the user xslt stylesheet; replaces the $$IMPORTED_XSLT$$
    $actual_xslt = preprocess_user_xslt( $rbook_desc );

    # Perform the transformations...
    my $src_name = preprocess_main_xslt( $rbook_desc );

    # Copy images from the source to the destination tree
    copy_images_to_output( $rbook_desc );

    if( $rbook_desc->{ "format" } eq "pdf" )
        {
        $cmd_string = preprocess_command_pattern(
            $config_values{ "DOCBOOK2PDF_COMMAND" },
            $src_name,
            ( $actual_xslt ? $actual_xslt : $config_values{ "DOCBOOK_FO_XSL" } ),
            $rbook_desc
            );

        process_commands( $cmd_string );
        }
    else
        {
        if( $rbook_desc->{ "format" } eq "html" )
            {
            $cmd_string = preprocess_command_pattern(
                $config_values{ "DOCBOOK2HTML_COMMAND" },
                $src_name,
                ( $actual_xslt ? $actual_xslt : $config_values{ "DOCBOOK_HTML_XSL" } ),
                $rbook_desc
                );

            process_commands( $cmd_string );
            }
        elsif( $rbook_desc->{ "format" } eq "chunk" )
            {
            # Prepare the command string
            $cmd_string = preprocess_command_pattern(
                $config_values{ "DOCBOOK2CHUNK_COMMAND" },
                $src_name,
                ( $actual_xslt ? $actual_xslt : $config_values{ "DOCBOOK_CHUNK_XSL" } ),
                $rbook_desc
                );

            # Switch into the target directory
            chdir( $rbook_desc->{ "output_path" } );

            # Perform the transformation
            process_commands( $cmd_string );

            # Rename the index.html if needed
            rename( "index.html", $rbook_desc->{ "output_name" } );

            # Switch directory back
            chdir( $current_dir );
            }
        else
            {
            # Unsupported format string

            error( "Unsupported format ".$rbook_desc->{ "format" }."." );
            }
        }


    # Delete temporary copies of user-defined XSLT stylesheets
    if( $actual_xslt )
        {
        unlink( $actual_xslt );
        }

    unlink( $src_name );
        # Delete the copy of the source file
    }  # generate_book()


###
# Validates docbook file
#
# Parameters:
#   - rbook_desc Description of the book to validate.
#
sub validate_book
    {
    my( $rbook_desc ) = @_;

    # Preprocess the xml; replaces $$DTD_REFERENCE$$
    my $src_name = preprocess_main_xslt( $rbook_desc );

    # Validate source.
    my $cmd_string = preprocess_command_pattern(
        $config_values{ "VALIDATE_COMMAND" },
        $src_name,
        'none',
        $rbook_desc
        );

    process_commands( $cmd_string );

    unlink( $src_name );
    }  # validate_book()


###
# Takes a list of commands (';'-separated) and launches each of the commands.
#
# Parameters:
#   - loc The list of commands.
#
sub process_commands
    {
    my( $loc ) = @_;
    my( @aoc );
        # array of commands
    my( $command );

    @aoc = split( /$cmd_separator/, $loc );

    foreach $command ( @aoc )
        {
        print $command."\n\n";
        system( $command );
        }
    }  # process_commands()


###
# Copies all files from source to destination directory. The files that are
# already present in the destination directory and are newer or equally old
# as the matching files in the source directory are not copied.
#
# Note that files in subdirectories will not be copied.
#
# Parameters:
#   - source The source directory path.
#   - dest The destination directory path.
#   - extensions List of extensions of files that can be touched.
#
sub actualize_directory
    {
    my( $source, $dest, $rextensions ) = @_;
    my( @files, $filename );
    my( $extension );

    if( not opendir( SDIR, $source ) )
        {
        error( "Cannot open the directory ".$source."\n" );
        }
    @files = readdir( SDIR );
    closedir( SDIR );

    create_path( $dest );

    foreach $filename ( @files )
        {
        foreach $extension ( @$rextensions )
            {
            if( $filename =~ /^.*$extension$/ )
                {
                if( compare_files_timestamps(
                            $source."/".$filename,
                            $dest."/".$filename
                            ) ==1 )
                    {
                    copy_file( $source."/".$filename, $dest."/".$filename );
                    }

                last;
                }
            }
        }
    }  # actualize_directory()



###
# Compares timestamps of two files. In this context, 'timestamp' means the
# time when the file has been modified last.
#
# Parameters:
#   - first Path to the first file.
#   - second Path to the second file.
#
# Returns:
#   -1 if the first file is older, 0 if both the files are equally old,
#   1 if the first file is younger.
#
# Note: non-existent file is assumed the older one. If neither of the files
#   exists, the routine returns 0. The files might be actually directories,
#   we don't care about that.
#
sub compare_files_timestamps
    {
    my( $first, $second ) = @_;
    my( @stat_first, @stat_second );
    my( $exists_first, $exists_second );
    my( $result );

    $exists_first = -f $first;
    $exists_second = -f $second;

    if( !$exists_first )
        {
        if( !$exists_second )
            {
            $result = 0;
                # Neither of the files exists.
            }
        else
            {
            $result = -1;
                # The first file doesn't exist.
            }
        }
    else
        {
        if( !$exists_second )
            {
            $result = 1;
                # The second file doesn't exist.
            }
        else
            {
            # Both files exist.
            my $sb_first = stat( $first );
            my $sb_second = stat( $second );

            if( $sb_first->mtime < $sb_second->mtime )
                {
                $result = -1;
                }
            elsif( $sb_first->mtime > $sb_second->mtime )
                {
                $result = 1;
                }
            else
                {
                $result = 0;
                }
            }
        }

    return $result;
    }  # compare_files_timestamps()



###
# Copies files using.
#
# Parameters:
#   - source The source file (can contain path).
#   - dest The destination file (can contain path).
#
# Returns:
#   1 if the file has been successfuly copied, prints error message and quits otherwise.
#
sub copy_file
    {
    my( $source, $dest ) = @_;
    copy( $source, $dest ) or error( "Cannot copy $source to ${dest}: $!\n" );
    return 1;
    }  # copy_file()



###
# Creates path, i.e. all directories contained withing.
#
# Parameters:
#   - path Path to be created.
#
# On error, prints error message and exits.
#
sub create_path
    {
    my( $path ) = @_;
    mkpath( $path );
    }  # create_path()



###
# Parses lines like "A = B".
#
# Parameters:
#   - line The source line.
#
# Returns:
#   Array that contains the 'A' part at index 0 and 'B' part at index 1.
#   If the line contains multiple '=' characters, the first one will be used
#   as a separator between 'A' part and 'B' part.
#
sub parse_assignment
    {
    my( $line ) = @_;
    my( @result );
    @result = $line =~ m/^\s*([^#\s]\S*?)\s*=\s*(.*\S)\s*$/;
    return @result;
    }  # parse_assignment()



###
# Formats and prints the error message and quits.
#
# Parameters:
#   - message The error message.
#
sub error
    {
    my( $message ) = @_;
    die "Error: ".$message."\n";
    }  # error()


###
# Formats and prints the error message about parsing some file.
#
# Parameters:
#   - filename Name of the file that caused the error.
#   - line_number Number of line that caused the error.
#   - message The error message.
#
sub error_file
    {
    my( $filename, $line_number, $message ) = @_;
    die "Error in ".$filename.", line ".$line_number.": ".$message."\n";
    }  # error_file()


###
# Finds the index in an array for a specified string value.
#
# Parameters:
#   - rarray Reference to the array.
#   - value The value to look for.
#
# Returns:
#   The lowest index of the value if found; -1 otherwise.
#
sub find_in_array
    {
    my( $rarray, $value ) = @_;
    my( $i );

    for( $i = 0; $i < @$rarray; $i++ )
        {
        if( $$rarray[ $i ] eq $value ) { return $i; }
        }

    return -1;
    }  # find_in_array()


###
# Gets a fully-qualified path for a given relative one.
#
# Parameters:
#   - relative_path The relative path.
#
# Return:
#   The corresponding absolute path.
#
sub get_absolute_path
    {
    my( $relative_path ) = @_;
    my( $absolute_path );
    my( $current_dir ) = ( getcwd() );

    chdir( $relative_path );
    $absolute_path = getcwd();
    chdir( $current_dir );

    return $absolute_path;
    }  # get_absolute_path()


###
# Gets a fully-qualified path for a given relative one with filename.
#
# Parameters:
#   - relative_path The relative path ending with a filename.
#
# Return:
#   The corresponding absolute path ending with the same filename.
#
sub get_absolute_path_file
    {
    my( $relative_path ) = @_;

    if( $relative_path =~ /^(.*)[\/\\]([^\/\\]*)$/ )
        {
        return get_absolute_path( $1 ).'/'.$2;
        }
    else
        {
        return getcwd().'/'.$2;
        }
    }  # get_absolute_path()


###
# Creates a temporary file that will be automaticaly deleted when
# closed.
#
# Parameters:
#   - filename Name of the file; some extra characters will be added
#     to the name to make it unique.
#
# Returns:
#   ( Tmp file name; Handle to the file )
#
sub create_temporary_file
    {
    my( $filename ) = @_;
    my( $fh, $fn );

    # Split the filename into the path and the filename.

    if( $filename =~ /^(.*)[\/\\]([^\/\\]*)$/ )
        {
        ( $fh, $fn ) = File::Temp::tempfile( $2.".XXXX", DIR => $1 );
        }
    else
        {
        ( $fh, $fn ) = tempfile( $filename."XXX" );
        }

    return ( $fh, $fn );
    }  # create_temporary_file()


###
# Merges two paths into a single absolute path. The result will be
# correct even if the second path is absolute.
#
# If some of the directories given with both arguments don't exists
# they may be created.
#
# Parameters:
#   - path1 The first path (it doesn't matter whether absolute or relative).
#   - path2 The second path (it doesn't matter whether absolute or relative).
#
# Returns:
#   The merged path or undef on error.
#
sub merge_paths
    {
    my( $path1, $path2 ) = @_;
    my( $actual_directory ) = ( getcwd() );
    my( $result );

    if( !chdir( $path1 ) )
        {
        create_path( $path1 ) or return undef;
        }

    if( !chdir( $path2 ) )
        {
        if( !create_path( $path2 ) )
            {
            chdir( $actual_directory );
            return undef;
            }

        chdir( $path2 );
        }

    $result = getcwd();

    chdir( $actual_directory ) or
        error( "Cannot change directory to ".$actual_directory."\n" );

    return $result;
    }  # merge_paths()


###
# Prints the information about the script
#
sub print_info
    {
    print "\ndocgen.pl - documentation transformator\n";
    print "Usage: perl docgen.pl [options]\n\n";
    print "Options:\n";
    print "    -c filename ... specifies the configuration file; if ommited, the default\n";
    print "                    name is docs_config.\n";
    print "    -f filename ... specifies the makefile name; if ommited, the default name\n";
    print "                    is docs_makefile.\n";
    print "    -p formats  ... specifies result formats that will be processed; if\n";
    print "                    ommited, all formats will be processed. formats contains a\n";
    print "                    comma-separated list of format strings (pdf,html,chunk).\n";
    print "    -i formats  ... specifies result formats that should be ignored; if\n";
    print "                    ommited, no formats will be ignored.\n";
    print "    -v          ... validates the book, does not process it.\n";
    print "    -h          ... prints this message.\n\n";
    }  # print_info()



###
# MAIN
#

# Parse the command line

getopts( "hvf:c:p:i:", \%options );

if( $options{ "h" } )
    {
    print_info();
    exit( 0 );
    }


# Process the formats lists

if( !defined( $options{ "p" } ) ) { $options{ "p" } = "html,pdf,chunk"; }
if( !defined( $options{ "i" } ) ) { $options{ "i" } = ""; }
@formats_processed = split( /,/, $options{ "p" } );
@formats_ignored = split( /,/, $options{ "i" } );


# Process the configuration file

if( defined( $options{ "c" } ) )
    {
    read_config_file( $options{ "c" }, \%config_values );
    }
else
    {
    read_config_file( "docs_config", \%config_values );
    }

verify_config_values( \%config_values );
set_catalog( \%config_values );

# Process the makefile

if( defined( $options{ "f" } ) )
    {
    read_makefile( $options{ "f" }, \@books );
    }
else
    {
    read_makefile( "docs_makefile", \@books );
    }

# Validate or generate all books specified in the configuration file

if( defined( $options{ "v" } ) )
    {
    # Validate all unique docbook inputs specified in the configuration file.

    my %processed;
    foreach my $book ( @books )
        {
        unless( defined( $processed{ $book->{ "name" } } ) )
            {
            validate_book( $book );
            $processed{ $book->{ "name" } }++;
            }
        }
    }
else
    {
    # Gnerate all books specified in the configuration file

    foreach my $book ( @books )
        {
        generate_book( $book, \@formats_processed, \@formats_ignored );
        }
    }
