=head1 NAME

dependency_cache.pm - cache that keeps lists of existing files and their dependencies

=head1 SYNOPSIS

    use dependency_cache ;

    load_cache( %state ) ;

    save_cache( %state ) ;

    find_prerequisities( %state ) ;

=head1 DESCRIPTION

=head2 Functions

This package offers functions to manage and query file dependency cache.

=over

=item load_cache

Load dependency cache from current directory which should be projects'
root directory and set $state{ CACHE } to reference to the cache.

=item save_cache

Save cache structures to current directory. Required elements of the
I<state> structure are CACHE and PLATFORM.

=item find_prerequisities

Find complete dependencies for all files - keys of DEPENDS structure of I<state>.
I<State> structure should contain CACHE, PLATFORM, VERBOSE, MAKEFILEDIR, INCLUDES,
REPLACEMENTS, DEPENDS, GENERATED and EXTGENERATED fields. CACHE is reference to the
cache structure, see L<mkgen_parser.pm> for more information about other elements
of I<state>structure.

=back

=head2 Internal cache structure

The cache is hash, where keys are directory names and values are hashes
containing information about each file in the directory. Keys of this hash
are file names, values are structures:

=over

=item MTIME

Modification time of the file.
It will be undefined for nonexisting generated files.

=item DEPENDS

Set of automatically determined dependencies (see Supported file types).
It does not include dependencies supplied by user in F<makefile.gen>.

This will be defined as empty set if there are no dependencies and undefined
if set of dependencies is unknown yet. The dependencies will not be resolved
yet, just plain string as parsed from the sources.

=item GENERATED

Defined if this file is marked as generated. Such files are never parsed and
information about their dependencies is not saved. It must be always supplied
in the I<state> structure. This is to prevent inconsistent behavior of B<mkgen>,
bacause generated files may or may not exist. We don't want B<mkgen> users to
hope that dependencies mentioned in F<makefile.gen> are "good enough" and
that the right dependencies will be found out after the files are generated.

=back

=head2 External cache structure

Cache is saved to F<.mkgen_cache/PLATFORM> subdirectory of projects' root
directory. This subdirectory will contain single file for each cached source
directory, with '/' replaced by '__'. These files are simple line-oriented
(lines beginning with # or lines containing only whitespace are ignored).

Each line contains file name and mtime (in seconds). If word "depends" is
next, set of file dependencies will follow (this set may be empty).
Otherwise dependencies for this file are not known yet.

=head2 Supported file types.

When determining dependencies, files are parsed and searched for include-style
commands. The following file types are supported:

=over

=item C/C++ sources (.c, .C, .cpp, .cc, .h, .hpp)

S<C<#include "foo">> preprocessor directives are recognized, F<foo> will be added
to dependencies. S<C<#include E<lt>barE<gt>>> preprocessor directives are ignored.

C<#include> anywhere in the file will be recognized as preprocessor directive as
long as no non-whitespace characters precede it on the line. Comments, ifdefs, etc.
are ignored. Preprocessor directives with spaces between the C<#> character and the
command will not be recognized. This can be used for special hacks. Example:

    #include "foo.h"        // foo.h included and added as dependency
    #include <cstdio>       // cstdio included, not added
    /*
     #include "bar.h"       // bar not included, but added
     */
    # include "baz.h"       // baz.h included, but not added
    // #include "a.h"       // a.h not included, not added

=item Massiv IDL (.idl)

Parsed the same way as C/C++ sources - S<C<#import "foo">> preprocessor directives
are recognized.

=item Perl scripts (.pl, .pm)

These files are ignored, dependencies must be supplied manually.

=item Special files (factlist.cache, idl.list)

Not parsed, but no "unknown file type" warning.

=back

=head1 KNOWN BUGS

Most functions can't handle empty string as directory name. Use '.' instead.

=cut


package dependency_cache ;

use strict ;
use warnings ;

use FileHandle ;
use DirHandle ;
use File::Path ;
use File::stat ;
use Fcntl ':mode' ;

use utils ;

our( @ISA, @EXPORT ) ;
@ISA = qw( Exporter ) ;
@EXPORT = qw(
    load_cache
    save_cache
    find_prerequisities
) ;


my $CACHE_DIRNAME = '.mkgen_cache' ;


###
# get_line
#
# handle - stream to read from
# return line read, undef on eof.
#
# Read single line from given stream. Skip empty lines and comments.
sub get_line( $ ) {
    my $handle = shift ;

    my $line ;
    LINE: while ( $line = $handle->getline() ) {
        chomp $line ;
        if ( $line =~ /^\s*(#|$)/ ) {
            next LINE ;
        }
        last LINE ;
    }

    return $line ;
}


###
# dir2fname
#
# name - directory name
#
# Convert cache directory name to file name.
sub dir2fname( $ ) {
    my $name = shift ;
    $name =~ s|[/\\]|__|g ;
    if ( $name eq '.' ) {
        $name = '__' ;
    }
    return $name ;
}


###
# load_cache_directory
#
# platform - platform directory name
# cache - cache to modify
# dirname - directory name, relative to projects' root
# verbose
#
# Try to load previous version of cache info about given directory.
# $$cache{ $dirname } must not be defined yet.
# If cache file does not exist, empty info will be created.
# Directory will not be scanned for file changes.
sub load_cache_directory( $\%$;$ ) {
    my $platform = shift ;
    my $cache = shift ;
    my $dirname = shift ;
    my $verbose = shift ;

    die "Empty dirname.\n" unless $dirname ;

    print "Loading info about directory '$dirname'.\n"
        if $verbose >= 4 ;

    die "Cache for '$dirname' already exists, can't load, stopped"
        if defined( $$cache{ $dirname } ) ;
    $$cache{ $dirname } = { } ;
    my $dirinfo = $$cache{ $dirname } ;

    my $dirfname = $CACHE_DIRNAME . '/' . $platform . '/' . dir2fname( $dirname ) ;
    my $handle = new FileHandle( $dirfname, 'r' ) ;

    if ( defined( $handle ) ) {
        while ( my $line = get_line( $handle ) ) {
            my( $fname, $mtime, $depends ) = $line =~ /^\s*(\S+)\s+(\d+)\s*(.*)$/ ;
            die "$dirfname: Unable to parse cache line '$line', stopped"
                unless defined( $depends ) ;
            die "$dirfname: Duplicate cache entry for '$fname', stopped"
                if defined( $$dirinfo{ $fname } ) ;

            $$cache{ $dirname }{ $fname } = {
                MTIME => $mtime
            } ;

            my( $dnames ) = $depends =~ /^depends\s*(.*)$/ ;
            if ( defined( $dnames ) ) {
                $$dirinfo{ $fname }{ DEPENDS } = { } ;
                foreach my $dname ( split( /\s+/, $dnames ) ) {
                    $$dirinfo{ $fname }{ DEPENDS }{ $dname }++ ;
                }
            }
        }

        $handle->close() ;
    }
}


###
# save_cache_directory
#
# platform - platform directory name
# dirinfo - info about all files in directory
# dirname - directory name, relative to projects' root
# verbose
#
# Save directory info.
sub save_cache_directory( $\%$;$ ) {
    my $platform = shift ;
    my $dirinfo = shift ;
    my $dirname = shift ;
    my $verbose = shift ;

    die "Empty dirname.\n" unless $dirname ;

    print "Saving info about directory '$dirname'.\n"
        if $verbose >= 4 ;

    my $dirfname = $CACHE_DIRNAME . '/' . $platform . '/' . dir2fname( $dirname ) ;
    my $handle = new FileHandle( $dirfname, 'w' ) or
        die "Can't create '$dirfname', stopped" ;

    print $handle <<EOT
# mkgen dependency cache for '$platform', directory '$dirname'.
# Automatically generated and modified by mkgen.
# It's safe to remove this file, but never modify its contents.

EOT
    ;

    foreach my $fname ( sort( keys( %$dirinfo ) ) ) {
        my $fileinfo = $$dirinfo{ $fname } ;

        # Skip nonexisting generated files.

        next if defined( $$fileinfo{ GENERATED } ) && ! defined( $$fileinfo{ MTIME } ) ;

        # Save info.

        print $handle "$fname $$fileinfo{ MTIME }" ;
        if ( defined( $$fileinfo{ DEPENDS } ) ) {
            print $handle " depends" ;
            foreach my $depend ( sort( keys( %{$$fileinfo{ DEPENDS }} ) ) ) {
                print $handle " $depend" ;
            }
        }
        print $handle "\n" ;
    }

    print $handle <<EOT

## EOF ##
EOT
    ;

    $handle->close() ;
}


###
# update_cache_directory
#
# cache - cache to modify
# dirname - directory name, relative to projects' root
# verbose
# return array of names of directory's subdirectories, relative to projects' root
#
# Scan given directory and update cache info.
# Info about non-existing files will be removed. Empty info about new files
# will be added. Dependency info about modified files will be removed.
sub update_cache_directory( \%$;$ ) {
    my $cache = shift ;
    my $dirname = shift ;
    my $verbose = shift ;

    die "Empty dirname.\n" unless $dirname ;

    print "Updating info about directory '$dirname'.\n"
        if $verbose >= 4 ;

    die "Cache for '$dirname' does not exist, can't update, stopped"
        unless defined( $$cache{ $dirname } ) ;
    my $dirinfo = $$cache{ $dirname } ;

    # Get info about all files and subdirectories in given directory.

    my %directory_files ;
    my @subdirs ;
    my $dirhandle = new DirHandle( $dirname ) ;
    if ( defined( $dirhandle ) ) {
        while ( defined( my $name = $dirhandle->read() ) ) {
            next if ( $name eq '..' ) or ( $name eq '.' ) ;

            my $complete_name = simplify_path( $dirname . '/' . $name ) ;
            my $st = stat( $complete_name ) ;
            if ( defined( $st ) ) {
                my $mode = $st->mode ;
                my $mtime = $st->mtime ;

                if ( S_ISREG( $mode ) ) {
                    # Regular file.

                    $directory_files{ $name } = { MTIME => $mtime } ;
                }
                elsif ( S_ISDIR( $mode ) ) {
                    # Subdirectory.

                    if ( ( $name ne 'CVS' ) && ( $name ne '.svn' ) ) {
                        push( @subdirs, $complete_name ) ;
                    }
                }
            }
            else {
                warn( "Warning: Can't stat '$complete_name'.\n" ) ;
            }
        }
    }
    else {
        # No such directory - empty directory info.

        warn( "Warning: Directory '$dirhandle' does not exist.\n" ) ;
    }

    # Merge new info into cache.

    foreach my $fname ( keys( %directory_files ) ) {
        if ( defined( $$dirinfo{ $fname } ) ) {
            # Known file. Remove DEPENDS if modification time does not match.

            if ( $$dirinfo{ $fname }{ MTIME } != $directory_files{ $fname }{ MTIME } ) {
                print "Modified file '$fname' in '$dirname' found.\n"
                    if $verbose >= 4 ;
                $$dirinfo{ $fname }{ MTIME } = $directory_files{ $fname }{ MTIME } ;
                delete $$dirinfo{ $fname }{ DEPENDS } if defined( $$dirinfo{ $fname }{ DEPENDS } ) ;
            }
            else {
                print "Unmodified file '$fname' in '$dirname' found.\n"
                    if $verbose >= 5 ;
            }
        }
        else {
            # New file.

            print "New file '$fname' in '$dirname' found.\n"
                if $verbose >= 3 ;
            $$dirinfo{ $fname } = $directory_files{ $fname } ;
        }
    }

    # Cache cleanup.

    foreach my $fname ( keys( %$dirinfo ) ) {
        unless ( defined( $directory_files{ $fname } ) ) {
            print "File '$fname' removed from '$dirname'.\n"
                if $verbose >= 3 ;
            delete $$dirinfo{ $fname } ;
        }
    }

    return \@subdirs ;
}


###
# cache_directories
#
# platform - platform directory name
# cache - cache to modify
# dirnames - directory names, relative to projects' root
# verbose
#
# Add info about given directories and all their subdirectories to mkgen cache.
sub cache_directories( $\%\@;$ ) {
    my $platform = shift ;
    my $cache = shift ;
    my $dirnames = shift ;
    my $verbose = shift ;

    my @queue = @$dirnames ;    # Directories to process
    my %dirs_processed ;        # Set of directories in queue or already done

    while ( defined( my $dirname = shift( @queue ) ) ) {
        die "Empty dirname.\n" unless $dirname ;

        # Skip known directories.

        next if defined( $$cache{ $dirname } ) ;

        # Process directory.

        load_cache_directory( $platform, %$cache, $dirname, $verbose ) ;
        my $subdirs = update_cache_directory( %$cache, $dirname, $verbose ) ;

        # Add subdirectories to queue.

        foreach my $subdir ( @$subdirs ) {
            unless ( defined( $dirs_processed{ $subdir } ) ) {
                $dirs_processed{ $subdir }++ ;
                push( @queue, $subdir ) ;
            }
        }
    }
}


###
# cache_generated
#
# state - see file documentation
#
# Add information about generated files to the cache.
sub cache_generated( \% ) {
    my $state = shift ;

    # Get required data from state.

    my $project_dir = $$state{ MAKEFILEDIR } ;
    my $generated = $$state{ GENERATED } ;
    my $ext_generated = $$state{ EXTGENERATED } ;
    my $replacements = $$state{ REPLACEMENTS }{ FILE } ;
    my $cache = $$state{ CACHE } ;

    foreach my $fname ( keys( %$generated ), keys( %$ext_generated ) ) {
        # Complete file name.

        my $replaced_fname = apply_replacements( $project_dir . '/' . $fname, %$replacements ) ;
        my $simplified_fname = simplify_path( $replaced_fname ) ;

        # Split into path and name.

        my $path = simplify_path( select_path( $simplified_fname ) ) ;
        my $basename = select_basename( $simplified_fname ) ;

        # Update cache.

        my $dirinfo = $$cache{ $path } ;
        unless ( defined( $dirinfo ) ) {
            warn( "Warning: Generated file '$simplified_fname' in uncached directory.\n" ) ;
            $dirinfo = { } ;
            $$cache{ $path } = $dirinfo ;
        }
        $$dirinfo{ $basename }{ GENERATED }++ ;
    }
}


###
# prepare_cache
#
# state - see file documentation
#
# Cache directories specified in INCLUDES and add info about generated files.
sub prepare_cache( \% ) {
    my $state = shift ;

    # Get required data from state.

    my $platform = $$state{ PLATFORM } ;
    my $project_dir = $$state{ MAKEFILEDIR } ;
    my $includes = $$state{ INCLUDES } ;
    my $replacements = $$state{ REPLACEMENTS }{ FILE } ;
    my $cache = $$state{ CACHE } ;
    my $verbose = $$state{ VERBOSE } ;

    # Cache include directories.

    my @directories ;
    foreach my $include ( '', keys( %$includes ) ) {
        my $dir = apply_replacements( $include, %$replacements ) ;
        $dir = simplify_path( $project_dir . '/' . $dir ) ;
        unless ( $dir ) {
            $dir = '.' ;
        }
        push( @directories, $dir ) ;
    }
    cache_directories( $platform, %$cache, @directories, $verbose ) ;

    # Add info about generated files to the cache.

    cache_generated( %$state ) ;
}


###
# parse_file
#
# fname - file name relative to projects' root
# verbose
#
# Parse given file and return reference to set of files it includes.
sub parse_file( $;$ ) {
    my $fname = shift ;
    my $verbose = shift ;

    my %result ;

    # Determine file type.

    my( $ext ) = $fname =~ m|\.([^./]*)$| ;
    $ext = "" unless defined( $ext ) ;

    my $type ;

    # C/C++ sources.

    if (
        $ext eq 'c' or
        $ext eq 'cpp' or
        $ext eq 'cc' or
        $ext eq 'C' or
        $ext eq 'h' or
        $ext eq 'hpp'
    ) {
        $type = 'c' ;
    }

    # Perl scripts. Currently skipped.

    elsif (
        $ext eq 'pl' or
        $ext eq 'pm'
    ) {
        $type = 'skip' ;
    }

    # Massiv IDL. Not defined yet, skipped.

    elsif (
        $ext eq 'idl'
    ) {
        $type = 'idl' ;
    }

    # Special files to skip. Quite hacky, should be defined externally.
    # factlist.chache is obsolete file - no longer used anywhere.
    # idl.list is dependency of class_list.* sources.

    elsif (
        $fname =~ /factlist\.cache$/ or
        $fname =~ /idl\.list$/
    ) {
        $type = 'skip' ;
    }
    else {
        warn( "Warning: Can't parse '$fname' - unknown type.\n" ) ;
        return \%result ;
    }

    # Parse the file.

    my $handle = new FileHandle( $fname, "r" ) ;
    local $_ ;

    if ( defined( $handle ) ) {
        print "Determining '$fname' prerequisities.\n"
            if $verbose >= 3 ;
        if ( $type ne 'skip' ) {
            while ( $_ = $handle->getline() ) {
                if ( $type eq 'c' ) {
                    # C/C++ source.

                    if ( my( $include ) = /^\s*#include\s+"(.*?)"/ ) {
                        $result{ $include }++ ;
                    }
                }
                elsif ( $type eq 'idl' ) {
                    # Massiv IDL.

                    if ( my( $include ) = /^\s*#import\s+"(.*?)"/ ) {
                        $result{ $include }++ ;
                    }
                }
            }
        }

        $handle->close() ;
    }
    else {
        warn( "Warning: Can't open '$fname', check permissions.\n" ) ;
    }

    return \%result ;
}


###
# get_cache_depends
#
# fname - file name relative to projects' root
# cache
# verbose
# return reference to set of file dependencies
#
# Make sure that list of dependencies for given file is up to date and
# return the dependencies.
sub get_cache_depends( $\%;$ ) {
    my $fname = shift ;
    my $cache = shift ;
    my $verbose = shift ;

    my( $path, $basename ) = split_filename( $fname ) ;
    my $dirinfo = $$cache{ $path } ;
    unless ( defined( $dirinfo ) ) {
        warn( "Warning: Nothing is known about directory '$path', " .
            "determining dependencies for '$fname'.\n" ) ;
        return { } ;
    }

    # Parse the file unless dependency set already exists.

    my $fileinfo = $$dirinfo{ $basename } ;
    unless ( defined( $fileinfo ) ) {
        warn( "Warning: Nothing is known about file '$fname', weird.\n" ) ;
        return { } ;
    }

    unless ( defined( $$fileinfo{ GENERATED } ) ) {
        unless ( defined( $$fileinfo{ DEPENDS } ) ) {
            $$fileinfo{ DEPENDS } = parse_file( $fname, $verbose ) ;
        }
    }

    return $$fileinfo{ DEPENDS } ;
}


###
# find_file
#
# source_fname - name of file to include from, relative to project directory
# include_fname - name of file to find
# state - state to used to find the file, see file documentation.
# return reference to set of files found, without FILE replacements applied,
#       relative to projects directory
#
# Note that all include directories must be already cached.
#
# Because this used to be really slow, special nasty optimization is done
# in this function - for each $include_fname's path ($include_path), array of
# structures with information about all directories to search is remembered in
# $$state{ INCLUDE_DIRINFOS }{ $include_path }.
# The same structure is remembered for each $source_path, $include_path
# pair in $$state{ SOURCE_DIRINFOS }{ $source_path }{ $include_path }.
#
# This structure contains two elements:
#   PATH - path to directory, relative to project directory, with trailing slash
#   DIRINFO - valid reference to cache directory info
sub find_file( $$\% ) {
    my $source_fname = shift ;
    my $include_fname = shift ;
    my $state = shift ;

    # Get required data from state.

    my $project_dir = $$state{ MAKEFILEDIR } ;
    my $includes = $$state{ INCLUDES } ;
    my $replacements = $$state{ REPLACEMENTS }{ FILE } ;
    my $cache = $$state{ CACHE } ;
    my $include_dirinfos = $$state{ INCLUDE_DIRINFOS } ;
    my $source_dirinfos = $$state{ SOURCE_DIRINFOS } ;

    # Split both names.

    my( $source_path, $source_basename ) = split_filename( $source_fname ) ;
    my( $include_path, $include_basename ) = split_filename( $include_fname ) ;

    # Cache references to include directory infos.

    unless ( defined( $$include_dirinfos{ $include_path } ) ) {
        my @dirinfos ;

        foreach my $include ( keys( %$includes ) ) {
            my $path = simplify_path( apply_replacements(
                $project_dir . '/' . $include . '/' . $include_path, %$replacements
            ) ) ;

            my $dirinfo = $$cache{ $path } ;
            if ( defined( $dirinfo ) ) {
                my $relative = relative_path( $path, $project_dir ) ;
                if ( $relative ) {
                    $relative .= '/' ;
                }
                push( @dirinfos, {
                    PATH => $relative,
                    DIRINFO => $dirinfo
                } ) ;
            }
        }

        $$include_dirinfos{ $include_path } = \@dirinfos ;
    }

    # Cache reference to source directory info.

    unless (
        defined( $$source_dirinfos{ $source_path } ) and
        defined( $$source_dirinfos{ $source_path }{ $include_path } )
    ) {
        my $path = simplify_path( apply_replacements(
            $project_dir . '/' . $source_path . '/' . $include_path, %$replacements
        ) ) ;

        my $dirinfo = $$cache{ $path } ;
        if ( defined( $dirinfo ) ) {
            my $relative = relative_path( $path, $project_dir ) ;
            if ( $relative ) {
                $relative .= '/' ;
            }
            $$source_dirinfos{ $source_path }{ $include_path } = {
                PATH => $relative,
                DIRINFO => $dirinfo
            } ;
        }
    }

    # Check directories.

    my %result ;
    foreach my $dirinfo (
        $$source_dirinfos{ $source_path }{ $include_path },
        @{$$include_dirinfos{ $include_path }}
    ) {
        if ( defined( $$dirinfo{ DIRINFO }{ $include_basename } ) ) {
            my $result_fname = $$dirinfo{ PATH } . $include_basename ;
            $result{ $result_fname }++ ;
        }
    }

    # Done.

    return \%result ;
}


###
# load_cache
sub load_cache( \% ) {
    my $state = shift ;

    # Cache is lazy, so we don't do anything here.

    die "There is already a cache in state structure, can't create a new one, stopped"
        if defined( $$state{ CACHE } ) ;
    $$state{ CACHE } = { } ;
}


###
# save_cache
sub save_cache( \% ) {
    my $state = shift ;

    my $platform = $$state{ PLATFORM } or
        die "PLATFORM not defined in state, stopped" ;
    my $cache = $$state{ CACHE } or
        die "CACHE not defined in state, stopped" ;
    my $verbose = $$state{ VERBOSE } ;

    mkpath( simplify_path( $CACHE_DIRNAME . '/' . $platform ) ) ;

    # Save each cached directory info.

    foreach my $dir ( keys( %$cache ) ) {
        save_cache_directory( $platform, %{$$cache{ $dir }}, $dir, $verbose ) ;
    }
}


###
# find_prerequisities
sub find_prerequisities( \% ) {
    my $state = shift ;

    # Check state.

    my $platform = $$state{ PLATFORM } or
        die "PLATFORM not defined in state, stopped" ;
    my $cache = $$state{ CACHE } or
        die "CACHE not defined in state, stopped" ;
    my $verbose = $$state{ VERBOSE } ;
    my $project_dir = $$state{ MAKEFILEDIR } or
        die "MAKEFILEDIR not defined in state, stopped" ;
    my $incldues = $$state{ INCLUDES } or
        die "INCLUDES not defined in state, stopped" ;
    my $replacements = $$state{ REPLACEMENTS }{ FILE } or
        die "REPLACEMENTS{ FILE } not defined in state, stopped" ;
    my $depends = $$state{ DEPENDS } or
        die "DEPENDS not defined in state, stopped" ;
    my $generated = $$state{ GENERATED } or
        die "GENERATED not defined in state, stopped" ;
    my $ext_generated = $$state{ EXTGENERATED } or
        die "EXTGENERATED not defined in state, stopped" ;

    # Prepare cache.

    print "Determining source prerequisities...\n"
        if $verbose >= 2 ;
    prepare_cache( %$state ) ;
    $$state{ INCLUDE_DIRINFOS } = { } ;
    $$state{ SOURCE_DIRINFOS } = { } ;

    # Create fifo queue of files to parse.

    my @sources_to_parse = keys( %{$depends} ) ;

    # Process sources.

    while ( my $source = shift( @sources_to_parse ) ) {
        my $source_depends = $$depends{ $source } or
            die "Depends{ $source } not defined, stopped" ;

        # Get complete file name.

        my $source_name = apply_replacements( $source, %$replacements ) ;
        $source_name = simplify_path( $project_dir . '/' . $source_name ) ;

        # Check all files this file includes.

        my $includes = get_cache_depends( $source_name, %$cache, $verbose ) ;
        foreach my $include ( keys( %$includes ) ) {
            # Find candidates for prerequisity.

            my $prerequisities_set = find_file( $source, $include, %$state ) ;
            my @prerequisities = keys( %$prerequisities_set ) ;
            warn( "Warning: Can't find '$include', included from '$source_name'.\n" )
                unless @prerequisities ;

            # Update list of file prerequisities.

            for my $prerequisity ( @prerequisities ) {
                $$source_depends{ $prerequisity }++ ;
            }
        }

        # Add prerequisities of this file to list of sources to process.

        foreach my $prerequisity ( keys( %$source_depends ) ) {
            next if defined( $$depends{ $prerequisity } ) ;
            $$depends{ $prerequisity } = { } ;
            push( @sources_to_parse, $prerequisity ) ;
        }
    }

    # Remove optimization structures.

    delete( $$state{ INCLUDE_DIRINFOS } ) ;
    delete( $$state{ SOURCE_DIRINFOS } ) ;
}


1 ;


__END__
