=head1 NAME

makefile_writer.pm - makefile generator

=head1 SYNOPSIS

    use makefile_writer ;

    write_makefile( %state ) ;

=head1 DESCRIPTION

=head2 Functions

Use this package to generate makefile for platform and projects specified by
mkgen I<state>.

=over

=item  write_makefile

Use I<state> to generate makefile called F<makefile.I<platform_suffix>>.

See L<mkgen_parser.pm> for more information about I<state> structure.

=back

=head1 KNOWN BUGS

None.

=cut


package makefile_writer ;

use strict ;
use warnings ;

use FileHandle ;

use utils ;

require Exporter ;
our( @ISA, @EXPORT ) ;
@ISA = qw( Exporter ) ;
@EXPORT = qw( write_makefile ) ;


###
# prerequisities_closure
#
# state - mkgen state
# primary - array of primary prerequisities.
#
# Starting with given list of direct prerequisities, determine the closure
# and return reference to array containing all the prerequisities.
sub prerequisities_closure( \%\@ ) {
    my $state = shift ;
    my $primary = shift ;

    my $depends = $$state{ DEPENDS } ;
    
    my @sources_to_process = @$primary ;
    my @prerequisities = @sources_to_process ;  # the closure
    my %sources_done ;                          # set of already processed sources
    foreach my $source ( @sources_to_process ) {
        $sources_done{ $source }++ ;
    }

    # Gather closure of prerequisities.

    while ( defined( my $source = shift @sources_to_process ) ) {
        # Try each direct prerequisity of $source.

        foreach my $prerequisity_name ( keys( %{$$depends{ $source }} ) ) {
            # Skip already processed ones.

            next if defined( $sources_done{ $prerequisity_name } ) ;

            # Add this source to the list.

            $sources_done{ $prerequisity_name }++ ;
            push( @prerequisities, $prerequisity_name ) ;
            push( @sources_to_process, $prerequisity_name ) ;
        }
    }

    # Done.

    return \@prerequisities ;
}



###
# replace_vname
#
# commands - commands hash of state.
# cmd_name - command name (key to commands ).
# vname - variable name.
#
# Apply @VNAME -> vname replacements in commands and return new list by value.
sub replace_vname( \%$$ ) {
    my $commands = shift ;
    my $cmd_name = shift ;
    my $vname = shift ;

    my $cmds = $$commands{ $cmd_name } ;
    die "Command $cmd_name not defined, stopped " unless defined( $cmds ) ;
    my @result = map { my $el = $_ ; $el =~ s/\@VNAME/$vname/g ; $el } @$cmds ;

    return @result ;
}


###
# write_makefile
#
# state - mkgen state
# Write makefile.
sub write_makefile( \% ) {
    my $state = shift ;

    # Get necessary data from state.

    my $version = $$state{ VERSION } ;
    my $verbose = $$state{ VERBOSE } ;
    
    my $mkgen_dir = $$state{ MKGENDIR } ;
    my $makefile_dir = $$state{ MAKEFILEDIR } ;
    my $root_dir = $$state{ ROOTDIR } ;
    my $config_subdir = $$state{ MKCONFSUBDIR } ;
    
    my $platform_subdir = $$state{ PLATFORMSUBDIR } ;
    my $platform = $$state{ PLATFORM } ;
    my $platform_suffix = $$state{ PLATFORM_SUFFIX } ;

    my $replacements = $$state{ REPLACEMENTS } ;

    my $separate_library_paths = $$state{ SEPARATE_LIB_PATHS } ;
    my $makefile_prefix = $$state{ MAKEFILE_PREFIX } ;
    my $makefile_postfix = $$state{ MAKEFILE_POSTFIX } ;

    my $projects = $$state{ PROJECTS } ;

    my $includes = $$state{ INCLUDES } ;
    my $defines = $$state{ DEFINES } ;
    my $precompile = $$state{ PRECOMPILE } ;
    my $generated = $$state{ GENERATED } ;
    my $ext_generated = $$state{ EXTGENERATED } ;
    my $depends = $$state{ DEPENDS } ;
    my $regenerate_depends = $$state{ REGENERATE_DEPENDS } ;

    my $commands = $$state{ COMMANDS } ;

    my $pure_out_begin = $$state{ VERBATIM }{ PREFIX } ;
    my $pure_out_middle = $$state{ VERBATIM }{ INFIX } ;
    my $pure_out_end = $$state{ VERBATIM }{ SUFFIX } ;
    #...

    # Create makefile.

    my $makefile_name = $makefile_dir . "makefile." . $platform_suffix ;
    my $handle = new FileHandle( $makefile_name, "w" ) ;   
    die "Can't write '$makefile_name', stopped" unless defined( $handle ) ;

    print "Writing '$makefile_name'...\n" if $verbose >= 2 ;

    # Insert handle into state temporarily. Hack.

    $$state{ HANDLE } = $handle ;
    
    # Define function which writes single line to makefile.
    
    sub wl( \%$ ) {
        my $state = shift ;
        my $line = shift ;

        # Apply write-time replacements.
        
        $line = apply_replacements( $line, %{$$state{ REPLACEMENTS }{ WRITE }} ) ;

        # Remove escape backslashes.
        
        $line =~ s/\\(.)/$1/g ;

        # Write the line.

        my $handle = $$state{ HANDLE } ;
        print $handle $line . "\n" ;
    }

    ###
    # Header.
    
    wl( %{$state}, "# makefile.$platform_suffix" ) ;
    wl( %{$state}, "# Automatically generated from makefile.gen by mkgen $version." ) ;
    wl( %{$state}, "" ) ;
    wl( %{$state}, "PLATFORM            = $platform" ) ;
    wl( %{$state}, "PLATFORMSUFFIX      = $platform_suffix" ) ;
    wl( %{$state}, "" ) ;
    wl( %{$state}, "ROOT                = $root_dir#" ) ;
    wl( %{$state}, "PROJECTDIR          = $makefile_dir#" ) ;
    wl( %{$state}, "OBJROOT             = \$(ROOT)\$(BDIR)\$(PROJECTDIR)" ) ;
    wl( %{$state}, "MKGENDIR            = $mkgen_dir#" ) ;
    wl( %{$state}, "MKGENROOT           = \$(ROOT)\$(MKGENDIR)" ) ;
    wl( %{$state}, "MKCONFSUBDIR        = $config_subdir#" ) ;
    wl( %{$state}, "MKCONFROOT          = \$(MKGENROOT)\$(MKCONFSUBDIR)" ) ;
    wl( %{$state}, "ALLPLATFORMSUBDIR   = $platform_subdir#" ) ;
    wl( %{$state}, "ALLPLATFORMROOT     = \$(MKGENROOT)\$(ALLPLATFORMSUBDIR)" ) ;
    wl( %{$state}, "PLATFORMSUBDIR      = \$(ALLPLATFORMSUBDIR)\$(PLATFORM)/#" ) ;
    wl( %{$state}, "PLATFORMROOT        = \$(MKGENROOT)\$(PLATFORMSUBDIR)" ) ;
    wl( %{$state}, "" ) ;

    ###
    # User prefix.

    foreach my $line ( @{$pure_out_begin} ) {
        chomp( $line ) ;
        wl( %{$state}, $line ) ;
    }
    wl( %{$state}, "" ) if scalar( @{$pure_out_begin} ) ;

    ###
    # Copy contents of platform prefix file.

    if ( defined( $makefile_prefix ) ) {
        my $prefix_name = $mkgen_dir . $platform_subdir . $makefile_prefix ;

        if ( -f $prefix_name ) {
            # Open the file.

            my $prefix_handle = new FileHandle( $prefix_name, "r" ) ;
            die "Can't open '$prefix_name', stopped" unless $prefix_handle ;

            # Copy its contents without any replacements.

            while ( my $line = $prefix_handle->getline() ) {
                print $handle $line ;
            }

            # Close it.

            $prefix_handle->close() ;
            wl( %{$state}, "" ) ;
        }
        else {
            # File not found warning.

            warn( "Prefix file '$prefix_name' not found.\n" ) ;
        }
    }

    ###
    # Includes and defines.

    # Ugly hack - delete default include directory.

    delete $$includes{ "" } ;

    # Include commands for compiler.

    wl( %{$state}, "INCLUDES = " . join( ' ', map( { "\$(INCPREFIX)$_" } sort( keys( %{$includes} ) ) ) ) ) ;
    wl( %{$state}, "" ) ;

    # Macro definitions for compiler.

    wl( %{$state}, "DEFINES = " . join( ' ', map( { "\$(DEFPREFIX)$_" } @{$defines} ) ) ) ;
    wl( %{$state}, "" ) ;

    ###
    # Program and library prerequisities (both as pure list and as compiler/linker options).

    # Write all necessary info for each project.

    foreach my $project_name ( keys( %{$projects} ) ) {
        my $project = $$projects{ $project_name } ;
        my $uc_name = uc( $project_name ) ;

        # Get source names without extensions.

        my @source_names = map( { /^(.*)\.[^.]*$/ } sort( keys( %{$$project{ SOURCES }} ) ) ) ;

        # List of objects as prerequisities.

        my @objects = map( { "\$(OBJROOT)$_\$(OBJ)" } @source_names ) ;
        wl( %{$state}, "${uc_name}_OBJS = " . join( ' ', @objects ) ) ;

        # Project type-spefic stuff.

        if ( defined( $$project{ PROGRAM } ) ) {
            # Program.

            # List of objects for linker.

            wl( %{$state}, "${uc_name}_LINK_OBJS = " . join( '$(LOBJSEP)', @objects ) ) ;
        }
        elsif ( defined( $$project{ LIBRARY } ) ) {
            # List of objects as linker paramters.

            wl( %{$state}, "${uc_name}_BUILD_OBJS = " . join( '$(BOBJSEP)', @objects ) ) ;
        }
        elsif ( defined( $$project{ SHARED_LIBRARY } ) ) {
            # List of objects as dll linker paramters.

            wl( %{$state}, "${uc_name}_BUILD_OBJS = " . join( '$(SHBOBJSEP)', @objects ) ) ;
        }
        else {
            # Unknown project type.

            die "Internal error, stopped" ;
        }

        if ( defined( $$project{ PROGRAM } ) or defined( $$project{ SHARED_LIBRARY } ) ) {
            # List of libraries as prerequisities.

            my @libraries = map( {
                    mangle( $_, '$(OBJROOT)', '$(LIBPFX)', '$(LIBSFX)' ) ;
                } @{$$project{ LIBS }}
            ) ;
            my @shared_libraries = map( {
                    mangle( $_, '$(OBJROOT)', '$(SHLIBPFX)', '$(SHLIBSFX)' ) ;
                } @{$$project{ SHLIBS }}
            ) ;
            wl( %{$state}, "${uc_name}_LIBS = " . join( ' ', @libraries, @shared_libraries ) ) ;
        }

        # List of libraries as prerequisities of symbolic program/library name - symbolic names.

        my @symbolic_libraries = map( {
                my $basename = select_basename( $_ ) ;
                ( "library_" . $basename )
            } @{$$project{ LIBS }}, @{$$project{ SHLIBS }}
        ) ;
        wl( %{$state}, "${uc_name}_SYMBOLIC_LIBS = " . join( ' ', @symbolic_libraries ) ) ;

        if ( defined( $$project{ PROGRAM } ) or defined( $$project{ SHARED_LIBRARY } ) ) {
            # List of libraries for linker.

            if ( $separate_library_paths ) {
                # Separate command to specify path to library and library name.

                my @link_libraries = (
                    # Static libraries.

                    map( {
                        my( $path, $name ) = m|^(.*/)([^/]*)$| ;
                        $path = "" unless $path ;
                        $name = $_ unless $name ;
                        ( "\$(LLIBPATH)\$(OBJROOT)${path}", "\$(LLIB)\$(LLIBPFX)${name}\$(LLIBSFX)" )
                    } @{$$project{ LIBS }} ),

                    # Dynamic libraries.

                    map( {
                        my( $path, $name ) = m|^(.*/)([^/]*)$| ;
                        $path = "" unless $path ;
                        $name = $_ unless $name ;
                        ( "\$(LSHLIBPATH)\$(OBJROOT)${path}", "\$(LSHLIB)\$(LSHLIBPFX)${name}\$(LSHLIBSFX)" )
                    } @{$$project{ SHLIBS }} ),

                    # And system libraries.

                    map( { "\$(LSYSLIB)\$(LSYSLIBPFX)$_\$(LSYSLIBSFX)" } @{$$project{ SYSLIBS }} )
                ) ;

                # Write it.

                wl( %{$state}, "${uc_name}_LINK_LIBS = " . join( ' ', @link_libraries ) ) ;
            }
            else {
                # Libraries specified including full path.

                my @link_libraries = (
                    # Static libraries.

                    map( {
                        mangle( $_, '$(LLIB)$(OBJROOT)', '$(LLIBPFX)', '$(LLIBSFX)' )
                    } @{$$project{ LIBS }} ),

                    # Dynamic libraries.

                    map( {
                        mangle( $_, '$(LSHLIB)$(OBJROOT)', '$(LSHLIBPFX)', '$(LSHLIBSFX)' )
                    } @{$$project{ SHLIBS }} ),

                    # System libraries.

                    map( { "\$(LSYSLIB)\$(LSYSLIBPFX)$_\$(LSYSLIBSFX)" } @{$$project{ SYSLIBS }} )
                ) ;

                # Write it.

                wl( %{$state}, "${uc_name}_LINK_LIBS = " . join( '$(LLIBSEP)', @link_libraries ) ) ;
            }
        }

        # Compiler defines used when building objects belonging to this project.

        my @project_defines ; # basic defines
        my @project_sh_defines ; # additional defines if shared libraries are enabled

        if ( defined( $$project{ SHARED_LIBRARY } ) ) {
            push( @project_sh_defines, "OPTION_BUILD_SHARED_${uc_name}" ) ;
        }

        foreach my $shlib_name ( @{$$project{ SHLIBS }} ) {
            my( $local_name ) = ( $shlib_name =~ m|([^/]*)$| ) ;
            push( @project_sh_defines, "OPTION_USE_SHARED_" . uc( $local_name ) ) ;
        }

        foreach my $line ( replace_vname( %$commands, 'IFDEF', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
            wl( %{$state}, $line ) ;
        }
        wl( %{$state}, "${uc_name}_DEFINES = " . join( ' ',
                    map( { "\$(DEFPREFIX)$_" } @project_defines, @project_sh_defines ) ) ) ;
        foreach my $line ( replace_vname( %$commands, 'ELSE', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
            wl( %{$state}, $line ) ;
        }
        wl( %{$state}, "${uc_name}_DEFINES = " . join( ' ',
                    map( { "\$(DEFPREFIX)$_" } @project_defines ) ) ) ;
        foreach my $line ( replace_vname( %$commands, 'ENDIF', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
            wl( %{$state}, $line ) ;
        }

        # Done.

        wl( %{$state}, "" ) ;
    }

    ###
    # User middle.

    foreach my $line ( @{$pure_out_middle} ) {
        chomp( $line ) ;
        wl( %{$state}, $line ) ;
    }
    wl( %{$state}, "" ) if scalar( @{$pure_out_middle} ) ;

    ###
    # "All" rule.

    {
        my @all ;

        # Add programs to @all.

        foreach my $project_name ( sort( keys( %{$projects} ) ) ) {
            my $project = $$projects{ $project_name } ;
            if ( defined( $$project{ PROGRAM } ) ) {
                push( @all, "program_" . select_basename( $project_name ) ) ;
            }
        }

        # Add libraries to @all.

        foreach my $project_name ( sort( keys( %{$projects} ) ) ) {
            my $project = $$projects{ $project_name } ;
            if ( defined( $$project{ LIBRARY } ) || defined( $$project{ SHARED_LIBRARY } ) ) {
                push( @all, "library_" . select_basename( $project_name ) ) ;
            }
        }

        # Write it.

        wl( %{$state}, "all : " . join( ' ', @all ) ) ;
        wl( %{$state}, "" ) ;
    } # eob

    ###
    # "generate" and "degenerate" rules.

    wl( %{$state}, "generate : " . join( ' ', sort( keys( %$generated ) ) ) ) ;
    wl( %{$state}, "" ) ;

    wl( %{$state}, "degenerate :" ) ;
    foreach my $file_name ( sort( keys( %$generated ) ) ) {
        foreach my $command ( @{$$commands{ REMOVE }} ) {
            # Replace target name.

            my $c = $command ;
            $c =~ s/\@TNAME/${file_name}/g ;

            # Write it.

            wl( %{$state}, "\t$c" ) ;
        }
    }
    wl( %{$state}, "" ) ;
    
    ###
    # Program/library rules.

    # Set of bdirs to create.

    my %bdirs ;

    # Programs.

    foreach my $project_name ( sort( keys( %{$projects} ) ) ) {
        my $project = $$projects{ $project_name } ;
        my $uc_name = uc( $project_name ) ;

        if ( defined( $$project{ PROGRAM } ) ) {
            die "cmd_link not defined, stopped" unless defined( $$commands{ LINK } ) ;

            # Add OBJROOT to list of directories to make.

            $bdirs{ '$(OBJROOT)' }++ ;

            wl( %{$state}, "program_${project_name} : \$(OBJROOT)\$(BDTAG) \$(${uc_name}_SYMBOLIC_LIBS) \$(OBJROOT)${project_name}\$(EXE) \$(ALWAYS)" ) ;
            wl( %{$state}, "\$(OBJROOT)${project_name}\$(EXE) : \$(${uc_name}_LIBS) \$(${uc_name}_OBJS)" ) ;

            # Write commands to link the program.

            foreach my $command ( @{$$commands{ LINK }} ) {
                # Replace target names.

                my $c = $command ;
                $c =~ s/\@TNAME/\$(OBJROOT)${project_name}\$(EXE)/g ;
                $c =~ s/\@UNAME/${uc_name}/g ;

                # Write it.

                wl( %{$state}, "\t$c" ) ;
            }

            # Next program.

            wl( %{$state}, "" ) ;
        }
    }

    # Static libraries.

    foreach my $project_name ( sort( keys( %{$projects} ) ) ) {
        my $project = $$projects{ $project_name } ;
        my $uc_name = uc( $project_name ) ;

        if ( defined( $$project{ LIBRARY } ) ) {
            die "cmd_build not defined, stopped" unless defined( $$commands{ BUILD } ) ;

            # Add OBJROOT to list of directories to make.

            $bdirs{ '$(OBJROOT)' }++ ;

            # Write library rules.

            wl( %{$state}, "library_${project_name} : \$(OBJROOT)\$(BDTAG) \$(${uc_name}_SYMBOLIC_LIBS) \$(OBJROOT)\$(LIBPFX)${project_name}\$(LIBSFX) \$(ALWAYS)" ) ;
            wl( %{$state}, "\$(OBJROOT)\$(LIBPFX)${project_name}\$(LIBSFX) : \$(${uc_name}_OBJS)" ) ;

            # Write commands to build the library.

            foreach my $command ( @{$$commands{ BUILD }} ) {
                # Replace target names.

                my $c = $command ;
                $c =~ s/\@TNAME/\$(OBJROOT)\$(LIBPFX)${project_name}\$(LIBSFX)/g ;
                $c =~ s/\@UNAME/${uc_name}/g ;

                # Write it.
                wl( %{$state}, "\t$c" ) ;
            }

            # Next library.

            wl( %{$state}, "" ) ;
        }
    }

    # Shared libraries.

    foreach my $project_name ( sort( keys( %{$projects} ) ) ) {
        my $project = $$projects{ $project_name } ;
        my $uc_name = uc( $project_name ) ;

        if ( defined( $$project{ SHARED_LIBRARY } ) ) {
            die "cmd_shbuild not defined, stopped" unless defined( $$commands{ SHBUILD } ) ;

            # Add OBJROOT to list of directories to make.

            $bdirs{ '$(OBJROOT)' }++ ;

            # Write library rules.

            wl( %{$state}, "library_${project_name} : \$(OBJROOT)\$(BDTAG) \$(${uc_name}_SYMBOLIC_LIBS) \$(OBJROOT)\$(SHLIBPFX)${project_name}\$(SHLIBSFX) \$(ALWAYS)" ) ;

            # Shared version.

            foreach my $line ( replace_vname( %$commands, 'IFDEF', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
                wl( %{$state}, $line ) ;
            }

            wl( %{$state}, "\$(OBJROOT)\$(SHLIBPFX)${project_name}\$(SHLIBSFX) : \$(${uc_name}_OBJS) \$(${uc_name}_LIBS)" ) ;
            foreach my $command ( @{$$commands{ SHBUILD }} ) {
                # Replace target names.

                my $c = $command ;
                $c =~ s/\@TNAME/\$(OBJROOT)\$(SHLIBPFX)${project_name}\$(SHLIBSFX)/g ;
                $c =~ s/\@UNAME/${uc_name}/g ;

                # Write it.

                wl( %{$state}, "\t$c" ) ;
            }

            # Static version.
            
            foreach my $line ( replace_vname( %$commands, 'ELSE', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
                wl( %{$state}, $line ) ;
            }
            
            wl( %{$state}, "\$(OBJROOT)\$(SHLIBPFX)${project_name}\$(SHLIBSFX) : \$(${uc_name}_OBJS)" ) ;
            foreach my $command ( @{$$commands{ BUILD }} ) {
                # Replace target names.

                my $c = $command ;
                $c =~ s/\@TNAME/\$(OBJROOT)\$(SHLIBPFX)${project_name}\$(SHLIBSFX)/g ;
                $c =~ s/\@UNAME/${uc_name}/g ;

                # Write it.

                wl( %{$state}, "\t$c" ) ;
            }

            # Next library.

            foreach my $line ( replace_vname( %$commands, 'ENDIF', 'CFG_ENABLE_SHARED_LIBRARIES' ) ) {
                wl( %{$state}, $line ) ;
            }
            wl( %{$state}, "" ) ;
        }
    }

    ###
    # External libraries.

    {
        # Create set of referenced libraries which can't be built in current directory.

        my %external_libs ;

        # Check each local project.

        foreach my $project_name ( keys( %{$projects} ) ) {
            my $project = $$projects{ $project_name } ;

            # All static libraries it requires.

            foreach my $library_name ( @{$$project{ LIBS }} ) {
                # Skip local libraries.

                if ( defined( $$projects{ $library_name } ) ) {
                    next if defined( $$projects{ $library_name }{ LIBRARY } ) ;
                    if ( defined( $$projects{ $library_name }{ PROGRAM } ) ) {
                        die( "Tried to link '$project_name' with static library " .
                             "'$library_name', but it's a program.\n" ) ;
                    }
                    die( "Tried to link '$project_name' with static library " .
                         "'$library_name', but it's a shared library.\n" ) ;
                }

                # Remember the rest.

                $external_libs{ $library_name } = 's' ;
            }

            # All shared libraries it requires.

            foreach my $library_name ( @{$$project{ SHLIBS }} ) {
                # Skip local libraries.

                if ( defined( $$projects{ $library_name } ) ) {
                    next if defined( $$projects{ $library_name }{ SHARED_LIBRARY } ) ;
                    if ( defined( $$projects{ $library_name }{ PROGRAM } ) ) {
                        die( "Tried to link '$project_name' with shared library " .
                             "'$library_name', but it's a program.\n" ) ;
                    }
                    die( "Tried to link '$project_name' with shared library " .
                         "'$library_name', but it's a static library.\n" ) ;
                }
                if ( defined( $external_libs{ $library_name } ) and $external_libs{ $library_name } eq 's' ) {
                    die( "Tried to link '$library_name' as both static and shared library." ) ;
                }

                # Remember it if it's ok.

                $external_libs{ $library_name } = 'd' ;
            }
        }

        # Write rules for external libraries.

        foreach my $library_name ( sort( keys( %external_libs ) ) ) {
            die "cmd_build_in not defined, stopped" unless defined( $$commands{ BUILDIN } ) ;

            # Write rule.

            my $path = select_path( $library_name ) ;
            my $path_noslash = $path ;
            $path_noslash =~ s|/$|| ;
            my $name = select_basename( $library_name ) ;
            wl( %{$state}, "library_${name} : \$(ALWAYS)" ) ;

            # Write commands to build library in its directory.

            foreach my $command ( @{$$commands{ BUILDIN }} ) {
                # Apply target and path replacements.

                my $c = $command ;
                $c =~ s/\@TNAME/library_${name}/g ;
                $c =~ s/\@PATH/${path_noslash}/g ;

                # Write it.

                wl( %{$state}, "\t$c" ) ;
            }

            # No command for library binary.

            if ( $external_libs{ $library_name } eq 's' ) {
                wl( %{$state}, mangle( $library_name, '$(OBJROOT)', '$(LIBPFX)', '$(LIBSFX)' ) . " :" ) ;
            }
            else {
                wl( %{$state}, mangle( $library_name, '$(OBJROOT)', '$(SHLIBPFX)', '$(SHLIBSFX)' ) . " :" ) ;
            }

            # Next library.

            wl( %{$state}, "" ) ;
        }
    } # eob

    ###
    # Object rules.

    {
        # Create set of all referenced sources.
        # Remember which source belongs to whic project.

        my %sources ;

        foreach my $project_name ( keys( %{$projects} ) ) {
            my $project = $$projects{ $project_name } ;

            foreach my $source_name ( keys( %{$$project{ SOURCES }} ) ) {
                if ( defined( $sources{ $source_name } ) ) {
                    my $pname = $sources{ $source_name } ;
                    die( "Source '$source_name' belongs to multiple projects: '$pname' and '$project_name'.\n" ) ;
                }
                $sources{ $source_name } = $project_name ;
            }
        }

        # Write object rules.

        my %obj_shortcuts ;

        foreach my $source_name ( sort( keys( %sources ) ) ) {
            my $project_name = $sources{ $source_name } ;

            # Determine source project info.

            my $uc_project_name = uc( $project_name ) ;
            my $project = $$projects{ $project_name } ;
            my $linkage = ( defined( $$project{ SHARED_LIBRARY } ) ? 'DYNAMIC' : 'STATIC' ) ;

            # Get source names elements.

            my $path = select_path( $source_name ) ;
            my $name = select_basename( $source_name ) ;
            my $extension = "" ;
            if( $name =~ /\./ ) {
                ( $name, $extension ) = ( $name =~ /^(.*)\.([^\.]*)$/ ) ;
            }

            # Create closure of source prerequisities.

            my @primary_prerequisities = ( $source_name ) ;
            my $prerequisities = prerequisities_closure( %$state, @primary_prerequisities ) ;

            # Add sources' object location to set of bdirs to make.

            $bdirs{ "\$(OBJROOT)${path}" }++ ;

            # Remember shortcut to this object.

            $obj_shortcuts{ $name } = [ ] unless exists( $obj_shortcuts{ $name } ) ;
            push( @{$obj_shortcuts{ $name }}, "\$(OBJROOT)${path}${name}\$(OBJ)" ) ;

            # Write object rule.

            wl( %{$state}, "\$(OBJROOT)${path}${name}\$(OBJ) : \$(OBJROOT)${path}\$(BDTAG) ${source_name} " .
                join( ' ', @$prerequisities ) ) ;

            # Write rule commands if they are defined for its source's extension.

            if ( exists( $$commands{ COMPILE }{ $extension } ) ) {
                foreach my $command ( @{$$commands{ COMPILE }{ $extension }} ) {
                    # Apply target and source replacements.

                    my $c = $command ;
                    $c =~ s/\@TNAME/\$(OBJROOT)${path}${name}\$(OBJ)/g ;
                    $c =~ s/\@SNAME/${source_name}/g ;
                    $c =~ s/\@PNAME/${uc_project_name}/g ;
                    $c =~ s/\@LINKAGE/${linkage}/g ;
                    if ( $precompile ) {
                        $c =~ s/\@PRECOMPILE/\$(PRECOMPILEPREFIX)${precompile}/g ;
                    }
                    else {
                        $c =~ s/\@PRECOMPILE//g ;
                    }

                    # Write it.

                    wl( %{$state}, "\t$c" ) ;
                }
            }
        }
        wl( %{$state}, "" ) ;

        # Write object shortcuts

        foreach my $shortcut_name ( sort( keys( %obj_shortcuts ) ) ) {
            wl( %{$state}, "${shortcut_name}_obj : " . join( ' ', @{$obj_shortcuts{ $shortcut_name }} ) ) ;
        }
        wl( %{$state}, "" ) ;

    } # eob

    ###
    # Bdir rules.

    # Write rules to make build directories.

    foreach my $bdir_name ( sort( keys( %bdirs ) ) ) {
        die "cmd_mkbdir not defined, stopped" unless defined( $$commands{ MKBDIR } ) ;

        # Write rule.

        wl( %{$state}, "${bdir_name}\$(BDTAG) : " ) ;

        # Write commands.
        
        my $bdir_name_noslash = $bdir_name ;
        $bdir_name_noslash =~ s|/$|| ;

        foreach my $command ( @{$$commands{ MKBDIR }} ) {
            my $c = $command ;
            $c =~ s/\@TNAME/${bdir_name}\$(BDTAG)/g ;
            $c =~ s/\@PATH/${bdir_name_noslash}/g ;
            wl( %{$state}, "\t$c" ) ;
        }
    }
    wl( %{$state}, "" ) if scalar( %bdirs ) ;

    ###
    # Generated files.

    foreach my $target_name ( sort( keys( %$generated ) ) ) {

        unless ( defined( @{$$commands{ EXTRA }{ $target_name }} ) ) {
            die "No 'cmd_extra' defined for generated target '$target_name', stopped" ;
        }

        unless ( defined( %{$$regenerate_depends{ $target_name }} ) ) {
            die "No 'regenerate_depends' defined for target '$target_name', but 'cmd_extra'\n" .
                "exists for the target - unable to create rule for the target. Check your\n" .
                "'makefile.gen'. Each target with extra commands should be marked as 'generated'\n" .
                "and both 'depends' and 'regenerate_depends' list should be defined for it.\n" .
                "Stopped";
        }

        # Prerequisities.

        my @primary_prerequisities = sort( keys( %{$$regenerate_depends{ $target_name }} ) ) ;
        my $prerequisities = prerequisities_closure( %$state, @primary_prerequisities ) ;

        # Write rule.

        wl( %{$state}, "${target_name} : " . join( ' ', @$prerequisities ) ) ;

        # Write commands.

        foreach my $command ( @{$$commands{ EXTRA }{ $target_name }} ) {
            wl( %{$state}, "\t$command" ) ;
        }

        # Done.

        wl( %{$state}, "" ) ;
    }

    ###
    # Externally generated files.

    foreach my $target_name ( sort( keys( %$ext_generated ) ) ) {
        if ( defined( $$regenerate_depends{ $target_name } ) ) {
            # Prerequisities.

            my @primary_prerequisities = sort( keys( %{$$regenerate_depends{ $target_name }} ) ) ;
            my $prerequisities = prerequisities_closure( %$state, @primary_prerequisities ) ;

            # Write rule.

            wl( %$state, "${target_name} : " . join( ' ', @$prerequisities ) ) ;
        }
        else {
            # No regenerate_depends supplied - always try to regenerate the file.

            wl( %$state, "${target_name} : \$(ALWAYS)" ) ;
        }

        # Write commands to generate the file in its directory.

        my $tname = $$ext_generated{ $target_name }{ TARGET } ;
        my $path = $$ext_generated{ $target_name }{ PROJECTDIR } ;
        foreach my $command ( @{$$commands{ BUILDIN }} ) {
            # Apply target and path replacements.

            my $c = $command ;
            $c =~ s/\@TNAME/${tname}/g ;
            $c =~ s/\@PATH/${path}/g ;

            # Write it.

            wl( %{$state}, "\t$c" ) ;
        }

        # Done.

        wl( %{$state}, "" ) ;
    }

    ###
    # User postfix.

    foreach my $line ( @{$pure_out_end} ) {
        chomp( $line ) ;
        wl( %{$state}, $line ) ;
    }
    wl( %{$state}, "" ) if scalar( @{$pure_out_end} ) ;

    ###
    # Platform postfix.

    if ( defined( $makefile_postfix ) ) {
        my $postfix_name = $mkgen_dir . $platform_subdir . $makefile_postfix ;

        if ( -f $postfix_name ) {
            # Open the file.

            my $postfix_handle = new FileHandle( $postfix_name, "r" ) ;
            die "Can't open '$postfix_name', stopped" unless $postfix_handle ;

            # Copy its contents without any replacements.

            while ( my $line = $postfix_handle->getline() ) {
                print $handle $line ;
            }

            # Close it.

            $postfix_handle->close() ;
            wl( %{$state}, "" ) ;
        }
        else {
            # File not found error.

            warn( "Post file '$postfix_name' not found.\n" ) ;
        }
    }

    ###
    # Finish.

    wl( %{$state}, "### EOF" ) ;

    $handle->close() ;
    delete $$state{ HANDLE } ;
}


1 ;


__END__
