=head1 NAME

mkgen_parser.pm - makefile description reader

=head1 SYNOPSIS

    use mkgen_parser ;

    read_makefile_gen( $file-name, %mkgen-state ) ;

=head1 DESCRIPTION

=head2 Functions

Use the following function to read and parse F<makefile.gen> file and create
corresponding entries in mkgen state structure.

=over

=item read_makefile_gen

Parse F<makefile.gen> file from I<file-name> using options stored in I<mkgen-state>
structure and add all necessary information into this structure.

C<read_makefile_gen> expects the following information in I<mkgen-state> structure:

=over

=item VERBOSE

Scalar value. Verbosity level (0 being quiet).

=item MKGENDIR

Path from root directory to directory containing F<mkgen.pl> script and
F<platform> subdirectory, including trailing slash.

=item PLATFORMSUBDIR

Name of subdirectory of I<MKGENDIR> where platform definitions are stored,
including trailing slash.

=item PLATFORM

Platform name.

=item REPLACEMENTS

Associative array containing several search-replace hashes:

=over

=item READ

These replacements are copied to LOCAL when F<makefile.gen> processing
starts.

=item LOCAL

Replacements applied when reading lines from F<makefile.gen>. Defined using
'->' syntax in platform description file. These replacements are applied just
before line processing begins. For macros this means that -> replacements
are applied when macros are expanded, not when they are defined.

LOCAL replacements may be modified during F<makefile.gen> processing using
I<replace> and I<noreplace> commands.

=item WRITE

Replacements applied when writing line to generated makefile. Defined using
'=>' syntax in platform description file.

Not used by C<read_makefile_gen>.

=item FILE

Replacements applied to filenames. Defined using '#i->' syntax in platform
description file.

FIXME: this is a bit weird, as these replacements are applied to (nearly?) all
filenames except those generated by find_file when generating list of
dependencies.

=back

=back

The following entries will be added to I<mkgen-state> on success:

=over

=item MAKEFILEDIR

Directory containing parsed F<makefile.gen>, relative to root directory,
including trailing slash.

=item ROOTDIR

Path back from I<MAKEFILEDIR> to current (root) directory.

=item PROJECTS

Associative array of projects (programs and libraries) defined by makefile.gen.
Each element contains:

=over

=item NAME

Name of the program/library. Equal to this element's key. It should be unique
withing set of all projects that may be compiled/linked together.

=item PROGRAM

Defined if this project is a program.

=item LIBRARY

Defined if this project is a static library.

=item SHARED_LIBRARY

Defined if this project is a library which can be built as shared (dll) if
enabled in global config. When building this library as shared,
OPTION_BUILD_SHARED_libname will be defined, when building program/library
that uses this one, OPTION_USE_SHARED_libname will be defined.

=item SOURCES

Reference to set of sources to build this project from. Paths are relative
to project directory.

=item LIBS

List of static libraries to build before this project. If project is PROGRAM,
it will be linked with the libraries. Order is not important. Names are
relative to project directory.

=item SHLIBS

List of shared libraries to build before this project. If project is PROGRAM,
it will be linked with the libraries. Order is not important. Names are
relative to project directory.

=item SYSLIBS

List of system libraries to link this project with.

=back

=item INCLUDES

Set (hash) of include directories, without FILE replacements applied, without
trailing slash.

Empty string used to be part of this set as a "current directory" hack. Source
parser supports including from current directory directly now.

Directory names are relative to current directory, but unlike other
file/directory names, they will not be simplified (ie. F<foo/../bar/baz> will
not be converted to F<bar/baz>.)

=item DEFINES

List (array) of macro definitions to pass to compiler.

=item PRECOMPILE

Name of header to precompile or undefined if no header should be precompiled.

=item GENERATED

Set (hash) containing names of files that must/can be generated, relative to
project directory.

=item EXTGENERATED

Hash. Keys are names of externally generated files, relative to current
project's directory (ie. "compatible" with GENERATED keys). Values are
structures:

=over

=item PROJECTDIR

Directory (relative to current project's directory) containing makefiles that
can be used to generate the file.

=item TARGET

Name of target used to generate the file.

=back

Set (hash) containing names of files that must/can be generated, relative to
project directory.

=item DEPENDS

Associative array of direct prerequisities for given files. Keys are filenames,
values are sets (hashes) of filenames that the key filename includes. The meaning
of this hash is that if the key filename is required to rebuild something, value
filenames are required as well.

Value will be defined for each existing source (even if the value is empty set).

Prerequisities for existing files should be gathered automatically by mkgen
(if the files contain S<C<#include "">> directives).

=item REGENERATE_DEPENDS

The structure is the same as DEPENDS, the meaning is different. REGENERATE_DEPENDS
contains primary prerequisities for makefile targets. DEPENDS of the primary
prerequisities will be used to gather set of all prerequisities.

See makefile.gen.txt for more info.

=item COMMANDS

Structure containing makefile commands.
Each structure attribute is list (array) of commands to emit to generated
makefile for corresponding rule. This structure contains the following attributes:
C<IFDEF>, C<ELSE>, C<ENDIF>, C<LINK>, C<BUILD>, C<SHBUILD>, C<BUILDIN>, C<COMPILE>,
C<EXTRA>, C<MKBDIR> and C<RM>.

See mkgen documentation for more info.

=item VERBATIM

Structure containing strings to write to generated makefile. Each structure
attribute is list (array) of lines to write near the beginning (attribute
C<PREFIX>), middle (attribute C<INFIX>) and end (attribute C<SUFFIX>) of
the makefile.

=back

I<mkgen-state> may contain more information, but it will not be used by
C<read_makfile_gen>:

=over

=item VERSION

mkgen.pl version.

=item MKCONFSUBDIR

Subdirectory of I<MKGENDIR> containing mkgen configs, including trailing
slash.

=item PLATFORM_SUFFIX

Platform suffix.

=item SEPARATE_LIB_PATHS

Hackish flag specifying how options for linker will be generated.

=item MAKEFILE_PREFIX

Text to write near the beginning of generated makefile.

=item MAKEFILE_POSTFIX

Text to write near the end of generated makefile.

=back

=head1 TODO

Macros with arguments.

=head1 KNOWN BUGS

None.

=cut


package mkgen_parser ;

use strict ;
use warnings ;

use FileHandle ;

use utils ;

require Exporter ;
our( @ISA, @EXPORT ) ;
@ISA = qw( Exporter ) ;
@EXPORT = qw( read_makefile_gen ) ;


###
# Structures used in this file:


# SourceStack
#
# Array of structures containing state of "opened" input file or macro.
#
# array of structures:
#   NAME - File or macro name.
#   HANDLE - I/O handle. Undefined for macros.
#   MACRO - Reference to macro structure. Undefined for files.
#   LINE - Next line number, counting from 0.


# IfStack
#
# Array of strucutres which store information about 'if' blocks.
#
# Each array element is structure describing single block:
#   TOPLEVEL - Top level block if defined (no elses can be here).
#   SKIPREST - Ignore all following blocks on this level if defined.
#     (previous condition already satisfied or this block is subblock of a SKIP block)
#   SKIP - Ignore this block if defined - condition not satisfied.


# Macro
#
# Macro definitions.
#
# Hash. Keys are macro names. Values are structures:
#   ARGUMENTS - number of required arguments. Any if undefined.
#   LINES - array of macro lines.


###
# get_line
#
# source_stack - SourceStack.
# return line read from current input or undef at toplevel eof.
#
# Read line from input file or macro.
sub get_line ( \@ ) {
    my $source_stack = shift ;

    # Check recursion.
    
    if ( @{$source_stack} > 100 ) {
        die "Source stack too deep - infinite recursion? Stopped" ;
    }

    # Try to read next line.
    
    while ( @{$source_stack} ) {
        my $current_source = $$source_stack[ $#{$source_stack} ] ;
        my $line ;

        if( $$current_source{ HANDLE } ) {
            # File.
            
            $line = $$current_source{ HANDLE }->getline() ;
            if ( $line ) {
                # Another line read.
                
                $$current_source{ LINE }++ ;
            }
            else {
                # EOF.
                
                $$current_source{ HANDLE }->close() ;
            }
        }
        elsif ( $$current_source{ MACRO } ) {
            # Macro.
            
            die "Internal error. Unknown macro line. Stopped"
                unless defined( $$current_source{ LINE } ) ;
                
            my $macro = $$current_source{ MACRO } ;
            my $macro_lines = $$macro{ LINES } ;
            
            if ( $$current_source{ LINE } < scalar( @{$macro_lines} ) ) {
                # Read next line.
                
                $line = $$macro_lines[ $$current_source{ LINE } ] ;

                # FIXME: replace arguments.
                
                $$current_source{ LINE }++ ;
            }
        }
        else {
            die "Internal error. Unknown source stack element. Stopped" ;
        }

        # Line found?
        
        return $line if $line ;

        # No line - eof or end-of-macro.
        
        pop @{$source_stack} ;
    }

    # Game over.

    return undef ;
}


###
# parse_warning
#
# message - message to write
# line - line on which the error occured.
# source_stack - current source stack
#
# Print warning message.
sub parse_warning( $$\@ ) {
    my $message = shift ;
    my $line = shift ;
    my $source_stack = shift ;

    # Print warning location.
        
    my $current_source = $$source_stack[ $#{$source_stack} ]
        or die "Internal error, stopped" ;

    if ( $$current_source{ HANDLE } ) {
        # Warning reading file.
        
        print STDERR "File '$$current_source{ NAME }'" ;
    }
    elsif ( $current_source->{MACRO} ) {
        # Warning expanding macro.
        
        print STDERR "Macro '$$current_source{ NAME }'" ;
    }
    else {
        die "Internal error, stopped" ;
    }

    # Print rest of warning message.
    # Note: Line number is 0-based, but it points to one line _after_ error.
    
    print STDERR " ($$current_source{ LINE }): $line\n$message\n" ;
}


###
# parse_error
#
# message - message to write
# line - line on which the error occured.
# source_stack - current source stack
#
# Print error message and exit.
sub parse_error( $$\@ ) {
    my $message = shift ;
    my $line = shift ;
    my $source_stack = shift ;

    # Print error.

    parse_warning( $message, $line, @$source_stack ) ;

    # Exit.
    
    die "Stopped" ;
}


###
# normalize
#
# name - file name relative to makefile.gen directory
# state - state structure
# return Normalized file name
#
# First FILE replacements are applied.
# Then name is converted to relative to root directory, simplified and
# converted back to relative to makefile.gen directory.
sub normalize( $\% ) {
    my $name = shift ;
    my $state = shift ;
    my $replaced = apply_replacements( $name, %{$$state{ REPLACEMENTS }{ FILE }} ) ;
    my $simplified = relative_path( $$state{ MAKEFILEDIR } . $replaced, $$state{ MAKEFILEDIR } ) ;
    return $simplified ;
}


###
# read_makefile_gen
#
# makefile_gen - Name of file to parse.
# state - mkgen state structure.
#
# Parse specified file using information stored in state and update the state.
sub read_makefile_gen( $\% ) {
    my $makefile_gen = shift ;
    my $state = shift ;

    # Use our own local $_.

    local $_ ;

    
    ###
    # Read input from state.
    
    # Extract all necessary information from the state.

    my $verbose = $$state{ VERBOSE } ;
    my $mkgen_dir = $$state{ MKGENDIR } ;
    my $platform_subdir = $$state{ PLATFORMSUBDIR } ;
    my $platform = $$state{ PLATFORM } ;
    my $replacements = $$state{ REPLACEMENTS } ;

    
    ###
    # Prepare state structures generated by this function.

    # Program/library definitions.

    my %projects ;
    $$state{ PROJECTS } = \%projects ;
    my $current_project = undef ;

    # Include directories.
    
    my %includes ;
    $$state{ INCLUDES } = \%includes ;

    # Macro defines.
    
    my @defines ;
    $$state{ DEFINES } = \@defines ;

    # Precompiled header.

    $$state{ PRECOMPILE } = undef ;

    # Generated files.
    
    my %generated ;
    my %ext_generated ;
    $$state{ GENERATED } = \%generated ;
    $$state{ EXTGENERATED } = \%ext_generated ;

    # Direct prerequisities.
    
    my %depends ;
    $$state{ DEPENDS } = \%depends ;
    my %regenerate_depends ;
    $$state{ REGENERATE_DEPENDS } = \%regenerate_depends ;

    # Make commands.

    my @cmd_ifdef ;    
    my @cmd_else ;
    my @cmd_endif ;
    my @cmd_link ;
    my @cmd_build ;
    my @cmd_shbuild ;
    my @cmd_build_in ;
    my %cmd_compile ;
    my %cmd_extra ;
    my @cmd_mkbdir ;
    my @cmd_remove ;
    $$state{ COMMANDS } = { } ;
    $$state{ COMMANDS }{ IFDEF } = \@cmd_ifdef ;
    $$state{ COMMANDS }{ ELSE } = \@cmd_else ;
    $$state{ COMMANDS }{ ENDIF } = \@cmd_endif ;
    $$state{ COMMANDS }{ LINK } = \@cmd_link ;
    $$state{ COMMANDS }{ BUILD } = \@cmd_build ;
    $$state{ COMMANDS }{ SHBUILD } = \@cmd_shbuild ;
    $$state{ COMMANDS }{ BUILDIN } = \@cmd_build_in ;
    $$state{ COMMANDS }{ COMPILE } = \%cmd_compile ;
    $$state{ COMMANDS }{ EXTRA } = \%cmd_extra ;
    $$state{ COMMANDS }{ MKBDIR } = \@cmd_mkbdir ;
    $$state{ COMMANDS }{ REMOVE } = \@cmd_remove ;

    # Verbatim output.
    
    my @pure_out_begin ;
    my @pure_out_middle ;
    my @pure_out_end ;
    $$state{ VERBATIM } = { } ;
    $$state{ VERBATIM }{ PREFIX } = \@pure_out_begin ;
    $$state{ VERBATIM }{ INFIX } = \@pure_out_middle ;
    $$state{ VERBATIM }{ SUFFIX } = \@pure_out_end ;   

    # Directory containing makefile.gen.

    my( $makefile_dir ) = ( $makefile_gen =~ m|^(.*?)[^/\\]*$| ) ;
    $makefile_dir = simplify_path( $makefile_dir ) ;
    if ( $makefile_dir ) {
        $makefile_dir .= '/' ;
    }
    else {
        $makefile_dir = './' ;
    }
    $$state{ MAKEFILEDIR } = $makefile_dir ;

    # Path from makefile.gen directory to current (root) directory.

    my $root_dir = relative_path( '', $makefile_dir ) ;
    if ( $root_dir ) {
        $root_dir .= '/' ;
    }
    else {
        $root_dir = './' ;
    }
    $$state{ ROOTDIR } = $root_dir ;

    
    ###
    # Init.

    print "Processing '$makefile_gen'...\n" if $verbose >= 2 ;

    
    ###
    # Prepare local temporary structures.

    # Initialize stack containing info about parsed files/macros. See SourceStack.
    
    my @source_stack ;

    # Main makefile.gen.
    
    push( @source_stack, {
            NAME => $makefile_gen,
            HANDLE => new FileHandle( $makefile_gen, "r" ),
            LINE => 0
        } ) ;
    die "Can't open 'makefile.gen' in '$makefile_dir', stopped"
        unless $source_stack[ 0 ]->{ HANDLE } ;

    # Always parse makegen.common first.

    {
        my $name = simplify_path( $mkgen_dir . $platform_subdir . 'common/makegen.common' ) ;
        my $handle = new FileHandle( $name, "r" ) ;
        die "Can't open '$name', stopped" unless defined( $handle ) ;
        push( @source_stack, { NAME => $name, HANDLE => $handle, LINE => 0 } ) ;
    } # eob

    # Initialize if stack which stores information about ifdef-like constructs. See IfStack.

    my @if_stack ;
    push( @if_stack, { TOPLEVEL => 1 } ) ;

    # Copy READ replacements to LOCAL.

    $$replacements{ LOCAL } = { } ;
    foreach my $replace_key ( keys( %{$$replacements{ READ }} ) ) {
        $$replacements{ LOCAL }{ $replace_key } = $$replacements{ READ }{ $replace_key } ;
    }

    # Initialize macro structures.

    my %macros ;
    my $current_macro = undef ; # Reference to macro we are defining.
    my $macro_depth ;           # Macro definition nesting level.

    # Initialize lists of macros to expand at beginning/end of program/library blocks.
    # Endmacros will be expanded in reverse order.
    # Default lists are empty.

    my %block_macros ;
    $block_macros{ PROGRAM } = [ ] ;
    $block_macros{ ENDPROGRAM } = [ ] ;
    $block_macros{ LIBRARY } = [ ] ;
    $block_macros{ ENDLIBRARY } = [ ] ;


    ###
    # Parse makefile.gen
    
    LINE: while ( $_ = get_line( @source_stack ) ) {
        # Concatenate lines ending with '\'.
        
        chomp ;
        if ( s/\s*\\\s*$/ / ) {
            my $next_line = get_line( @source_stack ) ;
            if ( $next_line ) {
                $next_line =~ /^\s*/ ;
                $_ .= $' ;
                redo LINE ;
            }
        }

        # Skip comments and empty lines.
        
        s/^\s*//;
        next if /^(#|$)/ ;


        ###
        # Inside macro definition.
        
        # If we are in a middle of macro definition just read and remember the lines.

        if ( $current_macro ) {
            if ( /^macro\s+/ ) {
                # Nested macro beginning.
                
                $macro_depth++ ;
            }
            elsif ( /^endmacro\b/ ) {
                # End of macro.
                
                $macro_depth-- ;
            }

            # Was last endmacro toplevel?
            
            if ( $macro_depth <= 0 ) {
                # Yes, macro definition finished.
                
                $current_macro = undef ;
            }
            else {
                # Otherwise read next macro line.
                
                push( @{$$current_macro{ LINES }}, $_ ) ;
            }

            # Stop line processing.
            
            next LINE ;
        }

        # Apply '->' and 'replace/noreplace' replacements to line.
        # Hack: do not replace key name in 'replace' commands.

        unless ( /^noreplace\s+\S+/ or /^require\s+\S+/ ) {
            if ( my( $key, $value ) = /^replace\s+(\S+)\s+(.+?)\s*$/ ) {
                $value = apply_replacements( $value, %{$$replacements{ LOCAL }} ) ;
                $_ = "replace $key $value" ;
            }
            elsif ( /^replace\s+(\S+)\s*$/ ) {
                # replacement to nothing...
                }
            else {
                $_ = apply_replacements( $_, %{$$replacements{ LOCAL }} ) ;
            }
        }

        # Parse the line.
        
        study ;

        
        ###
        # "Preprocessing" conditions.

        if ( /^(ifplatform|ifnplatform)\b/ ) {
            # If-block.

            my %if_stack_element ;
            if ( defined( $if_stack[ $#if_stack ]{ SKIP } ) ) {
                # Skipping current if-block so ignore all nested if-blocks.
                
                $if_stack_element{ SKIPREST }++ ;
                $if_stack_element{ SKIP }++ ;
            }
            elsif ( /^ifplatform\s+/ ) {
                # Check platform for equality.

                unless ( $' =~ /\b$platform\b/ ) {
                    $if_stack_element{ SKIP }++ ;
                }
            }
            elsif ( /^ifnplatform\s+/ ) {
                # Check platform for inequality.

                if ( $' =~ /\b$platform\b/ ) {
                    $if_stack_element{ SKIP }++ ;
                }
            }

            # New if-stack element.
            
            push( @if_stack, \%if_stack_element ) ;

            # Done.
            
            next LINE ;
        }
        
        if ( /^(elifplatform|elifnplatform|else)\b/ ) {
            # Else- or ElseIf-block.
            # FIXME. Merge this with if-block somehow.

            my $if_stack_element = $if_stack[ $#if_stack ] ;
            parse_error( "else without if", $_, @source_stack )
                if defined( $$if_stack_element{ TOPLEVEL } ) ;

            unless ( defined( $$if_stack_element{ SKIP } ) ) {
                # Current block has been parsed so skip everything else.
                
                $$if_stack_element{ SKIP }++ ;
                $$if_stack_element{ SKIPREST }++ ;
            }
            
            unless ( defined( $$if_stack_element{ SKIPREST } ) ) {
                # Check if condition.
                # FIXME: merge with standard ifs.

                delete $$if_stack_element{ SKIP } ;
                
                if ( /^elifplatform\s+/ ) {
                    # Check platform for equality.
                    
                    unless ( $' =~ /\b$platform\b/ ) {
                        $$if_stack_element{ SKIP }++ ;
                    }
                }
                elsif ( /^elifnplatform\s+/ ) {
                    # Check platform for inequality.

                    if ( $' =~ /\b$platform\b/ ) {
                        $$if_stack_element{ SKIP }++ ;
                    }
                }
            }

            # Done.
            
            next LINE ;
        }
        
        if ( /^endif\b/ ) {
            # End of series of if-blocks.
            
            parse_error( "endif without if", $_, @source_stack )
                if $if_stack[ $#if_stack ]{ TOPLEVEL } ;

            # Remove current if stack element.
            
            pop( @if_stack ) ;
            
            # Done.

            next LINE ;
        }

        # Skip line if current condition is not satisfied.
        
        next if defined( $if_stack[ $#if_stack ]{ SKIP } ) ;


        ###
        # Macro expansion

        # FIXME: macro arguments.
        
        foreach my $macro_name ( keys( %macros ) ) {
            if ( /^${macro_name}\s*$/ ) {
                # Expand macro here.
                
                push( @source_stack, {
                    NAME => $macro_name,
                    MACRO => $macros{ $macro_name },
                    LINE => 0
                } ) ;

                # Done.

                next LINE ;
            }
        }


        ###
        # Includes.
        
        if ( my( $type, $name ) = /^(sysinc|inc)lude\s+(\S+)/ ) {           
            if ( $type eq "sysinc" ) {
                # Apply sysinclude replacements.

                $name =~ s/\@PLATFORM/$platform/g ;

                # Prepend system path.
                
                $name = simplify_path( $mkgen_dir . $platform_subdir . $name ) ;
            }
            else {
                # Prepend makefile_dir.

                $name = simplify_path( $makefile_dir . $name ) ;
            }

            # Try to open the file.
            
            my $handle = new FileHandle( $name, "r" ) ;
            die "Can't open '$name', stopped" unless defined( $handle ) ;

            # Parse it.
            
            push( @source_stack, {
                NAME => $name,
                HANDLE => $handle,
                LINE => 0
            } ) ;

            # Done.
            
            next LINE ;
        }


        ###
        # Replacements.

        if ( my( $key, $value ) = /^replace\s+(\S+)\s+(.+?)\s*$/ ) {
            # no warning - this is actually used by factgen macros
            #parse_warning( "Replacement of '$key' redefined.", $_, @source_stack )
            #    if defined( $$replacements{ LOCAL }{ $key } ) ;
            $$replacements{ LOCAL }{ $key } = $value ;
            next LINE ;
        }

        if ( my( $key ) = /^replace\s+(\S+)\s*$/ ) {
            # no warning - this is actually used by factgen macros
            #parse_warning( "Replacement of '$key' redefined.", $_, @source_stack )
            #    if defined( $$replacements{ LOCAL }{ $key } ) ;
            $$replacements{ LOCAL }{ $key } = "" ;
            next LINE ;
        }

        if ( my( $key ) = /^noreplace\s+(\S+)/ ) {
            # no warning - this is actually used by factgen macros
            #parse_warning( "Replacement of '$key' not defined - noreplace ignored.", $_, @source_stack )
            #    unless defined( $$replacements{ LOCAL }{ $key } ) ;
            delete( $$replacements{ LOCAL }{ $key } ) ;
            next LINE ;
        }

        if ( my( $key ) = /^require\s+(\S+)/ ) {
            parse_error( "Required replacement for '$key' not defined,", $_, @source_stack )
                unless defined( $$replacements{ LOCAL }{ $key } ) ;
            next LINE ;
        }

        
        ###
        # Macros.

        # Macro beginning.
        
        if ( my( $name ) = /^macro\s+(\w+)/ ) {
            # FIXME: Macro arguments.

            # Create new macro.
            
            $current_macro = { LINES => [ ] } ;
            $macro_depth = 1 ;

            # Remember it.
            
            $macros{ $name } = $current_macro ;

            # Done.
            
            next LINE ;
        }

        # Add macro to program/library macro list.
        
        if ( my( $type, $name ) = /^add_(program|endprogram|library|endlibrary)_macro\s+(\w+)/ ) {
            $type = uc( $type ) ;
            push( @{$block_macros{ $type }}, $name ) ;
            next LINE ;
        }

        # Remove macro from program/library macro list.
        
        if ( my( $type, $name ) = /^remove_(program|endprogram|library|endlibrary)_macro\s+(\w+)/ ) {
            $type = uc( $type ) ;

            # Try to find the macro.
            
            my $index ;
            for ( $index = $#{$block_macros{ $type }} ; $index >= 0 ; $index-- ) {
                if ( $block_macros{ $type }[ $index ] eq $name ) {
                    delete $block_macros{ $type }[ $index ] ;
                    next LINE ;
                }
            }

            # Not found.
            
            parse_error( "No macro '$name' in '$type' list - can't remove.",
                $_, @source_stack ) ;
        }


        ###
        # Program/libarary block.
        
        # Program/library beginning.
        
        if ( my( $type, $name ) = /^(program|library|shared_library)\s+(\w+)/ ) {
            # Check if program/library name is ok.
            
            parse_error( "Programs/libraries can't be nested.\n" .
                "Forgot endprogram/endlibrary/endshared_library?", $_, @source_stack )
                if defined( $current_project ) ;
            parse_error( "Duplicate defintion of program/library $name.",
                $_,  @source_stack ) if exists( $projects{ $name } ) ;

            # Create new project structure.
            
            my %project ;
            $project{ NAME } = $name ;
            $project{ PROGRAM }++ if $type eq "program" ;
            $project{ LIBRARY }++ if $type eq "library" ;
            $project{ SHARED_LIBRARY }++ if $type eq "shared_library" ;
            $project{ SOURCES } = { } ;
            $project{ LIBS } = [ ] ;
            $project{ SHLIBS } = [ ] ;
            $project{ SYSLIBS } = [ ] ;
            $current_project = \%project ;
            $projects{ $name } = $current_project ;

            # Parse starting program/library macros first.
            # Bottom source_stack element will be parsed first.

            my $macros_type = ( $type eq 'shared_library' ? 'LIBRARY' : uc( $type ) ) ;
            foreach my $name ( reverse @{$block_macros{ $macros_type }} ) {
                push( @source_stack, {
                    NAME => $name,
                    MACRO => $macros{ $name },
                    LINE => 0
                } ) ;
            }

            # Done.

            next LINE ;
        }

        # Program/library end.
        
        if ( my( $type) = /^end(program|library|shared_library)\b/ ) {
            # Check if endprogram/library is valid.
            
            parse_error( "Not in program/library block.",
                $_, @source_stack ) unless $current_project ;
            if ( $type eq "program" ) {
                parse_error( "Not in program block.", $_, @source_stack )
                    unless defined( $$current_project{ PROGRAM } ) ;
            }
            elsif ( $type eq "library" ) {
                parse_error( "Not in static library block.", $_, @source_stack )
                    unless defined( $$current_project{ LIBRARY } ) ;
            }
            elsif ( $type eq "shared_library" ) {
                parse_error( "Not in shared library block.", $_, @source_stack )
                    unless defined( $$current_project{ SHARED_LIBRARY } ) ;
            }

            # Nasty hack - this creates dummy macro containing
            # "__internal_endproject" line. When all endprogram/endlibrary
            # macros are processed, this dummy macro will be parsed next
            # and will cause $current_project to become undefined.
            
            push( @source_stack, {
                NAME => "__internal",
                MACRO => { LINES => [ "__internal_endproject" ] },
                LINE => 0
            } ) ;

            # Add endmacros to source stack. The first macro will be parsed as last.
            
            my $macros_type = ( $type eq 'shared_library' ? 'LIBRARY' : uc( $type ) ) ;
            foreach my $name ( @{$block_macros{ "END" . $macros_type }} ) {
                push( @source_stack, {
                    NAME => $name,
                    MACRO => $macros{ $name },
                    LINE => 0
                } ) ;
            }

            # Done.
            
            next LINE ;
        }

        # Program/library end, internal hack.
        
        if ( /^__internal_endproject\b/ ) {
            parse_error( "Not in program/library block.", $_, @source_stack )
                unless $current_project ;

            # End of project definition.
            
            $current_project = undef ;

            # Done.

            next LINE ;
        }

        ###
        # Program/library options.

        if ( $current_project ) {
            # Modify list of project sources.
            
            if ( my( $operation ) = /^sources\s*([+-]?=)\s*/ ) {
                # Clear set of sources if operation is '='.
                
                $$current_project{ SOURCES } = { } if $operation eq "=" ;

                # Create list of file names.

                my %sources ;
                foreach my $fname ( split( /\s+/, $' ) ) {
                    # Apply replacements.
                    
                    my $replaced = apply_replacements( $fname, %{$$replacements{ FILE }} ) ;

                    # Convert names to relative to root directory.

                    my $root_relative = simplify_path( $makefile_dir . $replaced ) ;

                    # Glob if name contains '*' or '?'.

                    my @globbed ;
                    if ( $root_relative =~ /[*?]/ ) {
                        if ( $operation eq '-=' ) {
                            parse_warning( "Used glob pattern in 'sources -='", $_, @source_stack ) ;
                        }
                        @globbed = glob( $root_relative ) ;
                    }
                    else {
                        @globbed = ( $root_relative ) ;
                    }

                    # Convert file names back to makefile.gen-dir-relative.

                    foreach my $fn ( @globbed ) {
                        my $makefile_relative = relative_path( $fn, $makefile_dir ) ;
                        $sources{ $makefile_relative }++ ;
                    }
                }

                # Perform requested operation.                
                
                if ( $operation eq "-=" ) {
                    # Remove specified sources from list.
                    
                    foreach my $source ( keys( %sources ) ) {
                        delete $$current_project{ SOURCES }{ $source } ;
                    }
                }
                else { # '=' or '+='
                    # Add specified source from list.
                    
                    foreach my $source ( keys( %sources ) ) {
                        $$current_project{ SOURCES }{ $source }++ ;
                    }
                }

                # Done.
                
                next LINE ;
            }

            # Modify list of libraries.
            
            if ( my( $group, $operation ) = /^(shlibs|libs)\s*(\+?=)\s*/ ) {
                # Clear list of libraries if operation is '='.

                my $libs_group = uc( $group ) ;
                $$current_project{ $libs_group } = [ ] if $operation eq "=" ;

                # Add libraries to list applying file replacements.
                
                foreach my $lib ( split( /\s+/, $' ) ) {
                    push( @{$$current_project{ $libs_group }}, normalize( $lib, %$state ) ) ;
                }

                # Done.
                
                next LINE ;
            }

            # Modify lis of system libs.

            if( my( $operation ) = /^syslibs\s*(\+?=)\s*/ ) {
                # Clear list of system libraries if operation is '='.

                $$current_project{ SYSLIBS } = [ ] if $operation eq "=" ;

                # Add libraries to list applying file replacements.

                foreach my $syslib ( split( /\s+/, $' ) ) {
                    push( @{$$current_project{ SYSLIBS }}, $syslib ) ;
                }

                # Done.

                next LINE ;
            }
        }


        ###
        # Global options.
        
        # Includes. Will not be simplified.
        
        if ( my( $operation ) = /^includes\s*(\+?=)\s*/ ) {
            # Clear set of include directories if operation is '='.
            
            %includes = () if $operation eq "=" ;

            # Add includes to set of include directories.
            
            foreach my $include ( split( /\s+/, $' ) ) {
                # Remove trailing slash.
                
                $include =~ s|/$|| ;

                # Remember it.
                
                $includes{ $include }++ ;
            }

            # Done.
            
            next LINE ;
        }

        # Defines.
        
        if ( my( $operation ) = /^defines\s*(\+?=)\s*/ ) {
            # Clear list of defines if operation is '='.

            @defines = () if $operation eq "=" ;

            # Remember defines.
            
            push( @defines, split( /\s+/, $' ) ) ;

            # Done.
            
            next LINE ;
        }

        # Header to precompile.

        if( my( $header ) = /^precompile\s*=\s*(\S+)/ ) {
            $$state{ PRECOMPILE } = $header ;
            next LINE ;
        }

        # Set of generated files.
        
        if ( my( $operation ) = /^generated\s*(\+?=)\s*/ ) {
            # Clear set of generated files if operation is '='.

            %generated = () if $operation eq "=" ;

            # Add the names to set of generated files.
            
            foreach my $name ( split( /\s+/, $' ) ) {
                my $normalized = normalize( $name, %$state ) ;
                $generated{ $normalized }++ ;
            }

            # Done.
            
            next LINE ;
        }

        if ( my( $directory, $operation ) = /^extern_generated\s+(\S+?)\s*(\+?=)\s*/ ) {
            # FIXME: Syntax does not map to structures nicely, so '=' is
            # currently not supported; we would have to grep all elements in
            # the set and remove those with matching PROJECTDIR.

            parse_warning( "Command currently implemented as extern_generated +=",
                $_, @source_stack ) if $operation eq '=' ;

            # Add the names to set of generated files.

            foreach my $name ( split( /\s+/, $' ) ) {
                my $complete_name = normalize( $directory . '/' . $name, %$state ) ;
                $ext_generated{ $complete_name } = {
                    PROJECTDIR => normalize( $directory, %$state ),
                    TARGET => normalize( $name, %$state )
                } ;
            }

            # Done.

            next LINE ;
        }

        # Additional prerequisities.
        
        if ( my( $name, $operation ) = /^depends\s+(\S+?)\s*(\+?=)\s*/ ) {
            # Clear set of file prerequisities if operation is '='.

            my $name_normalized = normalize( $name, %$state ) ;
            $depends{ $name_normalized } = { } if $operation eq "=" ;

            # Add the names to set of file prerequisities.
            
            foreach my $prerequisity ( split( /\s+/, $' ) ) {
                my $prereq_normalized = normalize( $prerequisity, %$state ) ;
                $depends{ $name_normalized }{ $prereq_normalized }++ ;
            }

            # Done.
            
            next LINE ;
        }

        if ( my( $name, $operation ) = /^regenerate_depends\s+(\S+?)\s*(\+?=)\s*/ ) {
            # Clear set of file primary prerequisities if operation is '='.

            my $name_normalized = normalize( $name, %$state ) ;
            $regenerate_depends{ $name_normalized } = { } if $operation eq "=" ;

            # Add the names to set of file prerequisities.

            foreach my $prerequisity ( split( /\s+/, $' ) ) {
                my $prereq_normalized = normalize( $prerequisity, %$state ) ;
                $regenerate_depends{ $name_normalized }{ $prereq_normalized }++ ;
            }

            # Done.

            next LINE ;
        }

        ###
        # Rule commands.

        # Ifdef/else/endif commands.

        if ( my( $operation ) = /^cmd_ifdef\s*(\+?=)\s*/ ) {
            @cmd_ifdef = () if $operation eq "=" ;
            push( @cmd_ifdef, $' ) ;
            next LINE ;
        }
        if ( my( $operation ) = /^cmd_else\s*(\+?=)\s*/ ) {
            @cmd_else = () if $operation eq "=" ;
            push( @cmd_else, $' ) ;
            next LINE ;
        }
        if ( my( $operation ) = /^cmd_endif\s*(\+?=)\s*/ ) {
            @cmd_endif = () if $operation eq "=" ;
            push( @cmd_endif, $' ) ;
            next LINE ;
        }

        # Link commands.
        
        if ( my( $operation ) = /^cmd_link\s*(\+?=)\s*/ ) {
            @cmd_link = () if $operation eq "=" ;
            push( @cmd_link, $' ) ;
            next LINE ;
        }

        # Build "local" static library commands.
        
        if ( my( $operation ) = /^cmd_build\s*(\+?=)\s*/ ) {
            @cmd_build = () if $operation eq "=" ;
            push( @cmd_build, $' ) ;
            next LINE ;
        }

        # Build "local" dynamic library commands.
        # It should be the same as cmd_build if dynamic libraries are globally disabled.

        if ( my( $operation ) = /^cmd_shbuild\s*(\+?=)\s*/ ) {
            @cmd_shbuild = () if $operation eq "=" ;
            push( @cmd_shbuild, $' ) ;
            next LINE ;
        }

        # Build library in another directory.
        
        if ( my( $operation ) = /^cmd_build_in\s*(\+?=)\s*/ ) {
            @cmd_build_in = () if $operation eq "=" ;
            push( @cmd_build_in, $' ) ;
            next LINE ;
        }

        # Compile files with given extension.
        
        if ( my( $extension, $operation ) = /^cmd_compile\s+(\S+?)\s*(\+?=)\s*/ ) {
            $cmd_compile{ $extension } = [ ] if $operation eq "=" ;
            push( @{$cmd_compile{ $extension }}, $' ) ;
            next LINE ;
        }

        # Compile given file.
        
        if ( my( $filename, $operation ) = /^cmd_extra\s+(\S+?)\s*(\+?=)\s*/ ) {
            my $normalized = normalize( $filename, %$state ) ;
            $cmd_extra{ $normalized } = [ ] if $operation eq "=" ;
            push( @{$cmd_extra{ $normalized }}, $' ) ;
            next LINE ;
        }

        # Make directory containing built objects, libraries and executables.
        
        if ( my( $operation ) = /^cmd_mkbdir\s*(\+?=)\s*/ ) {
            @cmd_mkbdir = () if $operation eq "=" ;
            push( @cmd_mkbdir, $' ) ;
            next LINE ;
        }

        # Remove given file.

        if ( my( $operation ) = /^cmd_remove\s*(\+?=)\s*/ ) {
            @cmd_remove = () if $operation eq "=" ;
            push( @cmd_remove, $' ) ;
            next LINE ;
        }

        
        ###
        # Verbatim makefile output.
        
        if ( my( $position ) = /^\.(<|-|>)/ ) {
            # Get reference to list to modify.
            
            my $list ;
            $list = \@pure_out_begin if $position eq "<" ;
            $list = \@pure_out_middle if $position eq "-" ;
            $list = \@pure_out_end if $position eq ">" ;

            # Apply replacements to line.
            
            my $line = $' ;
            if ( $current_project ) {
                my $name = $$current_project{ NAME } ;
                my $ucname = uc( $name ) ;
                $line =~ s/\@NAME/${name}/g ;
                $line =~ s/\@UNAME/${ucname}/g ;
            }

            # Update the list.
            
            push( @{$list}, $line ) ;

            # Done.
            
            next LINE ;
        }

        # Echo.
        
        if ( /^echo\s*/ ) {
            print "$'\n" ;
            next LINE ;
        }

        # Unknown command.
        
        parse_error( "Unknown command", $_, @source_stack ) ;
    }

    # Determine for which source dependencies should be gathered:
    # - Sources of all projects.
    # - All files required to generate other files.

    my @sources_to_process ;
    foreach my $project ( keys( %projects ) ) {
        push( @sources_to_process, keys( %{$projects{ $project }{ SOURCES }} ) ) ;
    }
    foreach my $generated ( keys( %regenerate_depends ) ) {
        push( @sources_to_process, keys( %{$regenerate_depends{ $generated }} ) ) ;
    }

    # Make empty set of prerequisities for those sources unless the set exists.

    foreach my $source ( @sources_to_process ) {
        $depends{ $source } = { } unless defined $depends{ $source } ;
    }
    
    # Done.
}
    

1 ;


__END__
