=head1 NAME

source_parser.pm - C++ source reader

=head1 SYNOPSIS

    use source_parser ;

    find_prerequisities( %state ) ;

=head1 DESCRIPTION

=head2 Functions

This package offers functions that parse C++ sources to update mkgen state.

=over

=item find_prerequisities

Parse all sources mentioned in I<state> and update their direct prerequisities
according to S<#include ""> directives.

See L<mkgen_parser.pm> for more information about I<state> structure.

=back

=head1 KNOWN BUGS

Will be removed soon, replaced by L<dependency_cache.pm>.

=cut


package source_parser ;

use strict ;
use warnings ;

use FileHandle ;
use Cwd ;

use utils ;

our( @ISA, @EXPORT ) ;
@ISA = qw( Exporter ) ;
@EXPORT = qw(
    load_cache
    save_cache
    find_prerequisities
) ;


###
# find_file
#
# source_filename - name of file to include from
# include_filename - name of file to find
# state - mkgen state to use to find the file
# return list of files found, without FILE replacements applied
sub find_file( $$\% ) {
    my $source_filename = shift ;
    my $include_filename = shift ;
    my $state = shift ;

    # Get required data from state.
    
    my $includes = $$state{ INCLUDES } ;
    my $generated = $$state{ GENERATED } ;
    my $ext_generated = $$state{ EXTGENERATED } ;
    my $file_replacements = $$state{ REPLACEMENTS }{ FILE } ;

    # Try to find the file. "" means current directory.

    my %result ;
    foreach my $include ( "", keys( %{$includes} ) ) {
        # Create file name including path.
        
        my $complete_filename ;
        if ( $include eq "" ) {
            # Current directory. Get it from source_filename.

            my $source_path = select_path( $source_filename ) ;
            $complete_filename = $source_path . $include_filename ;
        }
        else {
            # Another directory.
            
            $complete_filename = $include . '/' . $include_filename ;
        }

        # Simplify file name.

        $complete_filename = simplify_path( $complete_filename ) ;

        # Apply replacemetns to this name.

        my $replaced_filename =
            apply_replacements( $complete_filename, %{$file_replacements} ) ;

        # Does it exist? Or is it generated?
        
        if (
            -f $replaced_filename or
            defined( $$generated{ $complete_filename } ) or
            defined( $$ext_generated{ $complete_filename } )
        ) {
            # Yes - add it to result.
            
            $result{ $complete_filename }++ ;
        }
    }

    # Done.
    
    return keys( %result ) ;
}


###
# load_cache and save_cache
# No-ops to make source_parser API compatible with dependency_cache.
sub load_cache( \% ) {
    my $state = shift ;
    $$state{ CACHE } = { } ;
}

sub save_cache( \% ) {
}


###
# find_prerequisities
#
# state - mkgen state to read and modify
#
# Find all direct prerequisities for each source.
sub find_prerequisities( \% ) {
    my $state = shift ;

    # Get required input from state.

    my $verbose = $$state{ VERBOSE } ;
    my $project_dir = $$state{ MAKEFILEDIR } ;
    my $file_replacements = $$state{ REPLACEMENTS }{ FILE } ;
    my $depends = $$state{ DEPENDS } ;
    my $generated = $$state{ GENERATED } ;
    my $ext_generated = $$state{ EXTGENERATED } ;

    # Chdir to directory containing makefile.gen.

    my $original_dir = cwd() ;
    chdir( $project_dir ) or die "Can't chdir to '$project_dir', stopped" ;

    print "Determining source prerequisities...\n" if $verbose >= 2 ;
        
    # Create fifo queue of files to parse.

    my @sources_to_parse = keys( %{$depends} ) ;

    # Process sources.
        
    while ( my $source = shift( @sources_to_parse ) ) {
        # Apply "file" replacements.
        
        my $source_name = apply_replacements( $source, %{$file_replacements} ) ;

        # Try to open the file and search it for #include if it exists.

        my $handle = new FileHandle( $source_name, "r" ) ;
        if ( $handle ) {
            while ( $_ = $handle->getline() ) {
                # Check for #include"" directive.

                if ( my( $include ) = /^\s*#include\s+"(.*?)"/ ) {
                    # Found new direct prerequisity. Try to find it. May find multiple files.

                    my @prerequisities = find_file( $source, $include, %{$state} ) ;
                    print STDERR "Can't find '$include', #included from '$source'.\n"
                        unless @prerequisities ;

                    for my $prerequisity ( @prerequisities ) {
                        # Update list of file prerequisities.

                        $$depends{ $source }{ $prerequisity }++ ;
                    }
                }
            }

            $handle->close() ;
        }
        elsif (
            not defined( $$generated{ $source } ) and
            not defined( $$ext_generated{ $source } )
        ) {
            # Warning if file can't be opened and it's not marked as generated.
            
            print STDERR "Can't open '$source_name', check permissions.\n" ;
            next ;
        }

        # Add prerequisities of this file to list of sourcess to process.

        foreach my $prerequisity ( keys( %{$$depends{ $source }} ) ) {
            next if defined( $$depends{ $prerequisity } ) ;
            $$depends{ $prerequisity } = { } ;
            push( @sources_to_parse, $prerequisity ) ;
        }
    }

    # Done.
    
    chdir( $original_dir ) or die "Can't chdir back to '$original_dir', stopped" ;
}


1 ;


__END__
