=head1 NAME

utils.pm - various mkgen utility functions

=head1 SYNOPSIS

    use utils ;

    $replaced = apply_replacements( $string, %replacements ) ;

    $path = select_path( $filename ) ;

    $basename = select_basename( $filename ) ;

    ( $path, $basename ) = split_path( $filename ) ;

    $mangled = mangle( $filename, $prefix, $infix, $suffix ) ;

    $simplified = simplify_path( $filename ) ;

    $relative = relative_path( $path, $base ) ;

=head1 DESCRIPTION

=head2 Functions

Utility package offers the following functions.

=over

=item apply_replacements

Applies replacements specified by I<replacements> hash to I<string> and returns
the result. If replacement source is single character, no replacement will be
made if character is preceded by '\' in I<string>. If replacementssource is
word (it matches ^\w+$), only separate words (\b\w+\b) will be replaced.

The replacements will be applied repeatedly (in "random" order) until no more
replacements are possible.

=item select_path

Get path component of filename, including trailing slash.

=item select_basename

Get base name component of filename.

=item split_filename

Return path (without trailing slash!) and basename.

=item mangle.

Get filename with prefix prepended, suffix appended and infix inserted inbetween
file path and name.

=item simplify_path

Get rid of superfluous ".." in path, remove "." elements and all trailing slashes.
For example F<./foo/.//bar/../boo/> will be converted to F<foo/boo>.

=item relative_path

Convert I<path> to path relative to I<base>. Both I<path> and I<base> should be
either absolute paths or paths relative to the same directory.

=back

=head1 TODO

Common utils for factgen and mkgen.

=head1 KNOWN BUGS

None.

=cut


package utils ;

use strict ;
use warnings ;

our( @ISA, @EXPORT ) ;
@ISA = qw( Exporter ) ;
@EXPORT = qw(
    apply_replacements
    select_path
    select_basename
    split_filename
    mangle
    simplify_path
    relative_path
) ;


###
# apply_replacements
#
# string - string to replace in.
# replacements - hash containing search-replace pairs.
# return string with applied replacements.
sub apply_replacements( $\% ) {
    my $string = shift ;
    my $replacements = shift ;

    my $repeats = 0 ;
    my $count ;
    do {
        $count = 0 ;

        while ( my( $search, $replace ) = each %{$replacements} ) {
            if ( length( $search ) <= 1 ) {
                # Single character - do not replace '\c'.

                $count += $string =~ s/([^\\]|^)\Q$search\E/$1$replace/g ;
            }
            elsif ( $search =~ /^\w+$/ ) {
                # Replace only whole words.

                $count += $string =~ s/\b\Q$search\E\b/$replace/g ;
            }
            else {
                # Replace any occurance.

                $count += $string =~ s/\Q$search\E/$replace/g ;
            }
        }

        $repeats++ ;
        if ( $repeats > 100 ) {
            warn "Cyclic replacements detected, current string is '$string'." ;
            $count = 0 ;
        }
    } while ( $count > 0 ) ;

    return $string ;
}


###
# select_path
#
# string - file name
# return path component of file name.
sub select_path( $ ) {
    my $string = shift ;
    
    my( $path ) = $string =~ m|^(.*/)[^/]*$| ;
    $path = "" unless $path ;
    
    return $path ;
}


###
# select_basename
#
# string - file name
# return file base name component of file name.
sub select_basename( $ ) {
    my $string = shift ;
    
    my( $basename ) = $string =~ m|^.*/([^/]*)$| ;
    $basename = $string unless $basename ;
    
    return $basename ;
}


###
# split_path
sub split_filename( $ ) {
    my $string = shift ;

    my( $path, $basename ) = $string =~ m|^(.*)/([^/]*)$| ;
    unless ( defined( $basename ) ) {
        $path = "" ;
        $basename = $string ;
    }

    return ( $path, $basename ) ;
}


###
# mangle
#
# string - file name
# prefix
# infix
# suffix
# return file name with prefix prepended, suffix appended and infix
#     inserted between file path and name
sub mangle( $$$$ ) {
    my $string = shift ;
    my $prefix = shift ;
    my $infix = shift ;
    my $suffix = shift ;
    
    my $path = select_path( $string ) ;
    my $name = select_basename( $string ) ;
    
    return $prefix . $path . $infix . $name . $suffix ;
}


###
# simplify_path
sub simplify_path( $ ) {
    my $string = shift ;
    my @elements = split( /\//, $string ) ;
    
    my @simplified ;
    foreach my $element ( @elements ) {
        if (
            $element eq '..' and
            scalar( @simplified ) and
            $simplified[ $#simplified ] ne '..'
        ) {
            pop( @simplified ) ;
        }
        elsif (
            $element ne '.' and
            $element ne ''
        ) {
            push( @simplified, $element ) ;
        }
    }
    
    return join( '/', @simplified ) ;
}


###
# relative_path
sub relative_path( $$ ) {
    my $path = shift ;
    my $base = shift ;

    my @pelements = split( /\//, simplify_path( $path ) ) ;
    my @belements = split( /\//, simplify_path( $base ) ) ;

    # Skip common part.

    ELEMENT: while ( @pelements and @belements ) {
        if ( $pelements[ 0 ] eq $belements[ 0 ] ) {
            shift( @pelements ) ;
            shift( @belements ) ;
        }
        else {
         last ELEMENT ;
        }
    }

    # Chdir using '..' to common base.

    my @result ;
    while ( @belements ) {
        push( @result, '..' ) ;
        shift( @belements ) ;
    }

    return join( '/', ( @result, @pelements ) ) ;
}


1 ;


__END__
