#! /usr/bin/perl -w
#
# Generate makefiles

=head1 NAME

mkgen.pl - makefile generator

=head1 SYNOPSIS

B<mkgen.pl>
    S<[ B<-g>|B<--generate> ]>
    S<[ B<-c>|B<--clean> ]>
    S<[ B<-r>|B<--root> I<root-directory> ]>
    S<[ B<-M>|B<--mkgendir> I<mkgen-directory> ]>
    S<[ B<-p>|B<--platform> I<platform> ]>
    S<[ B<-v>|B<--verbose> ]>
    S<[ I<project-directory>... ]>

=head1 DESCRIPTION

Generete makefiles from F<makefile.gen> files found in directories specified
by I<project-directory> and their subdirectories. See documentation in
F<doc/mkgen> directory for more information about F<makefile.gen> format
and/or mkgen internals.

Commands:

=over

=item B<-g>|B<--generate>

Generate makefiles. This is default.

=item B<-c>|B<--clean>

Remove generated makefiles.

=back

Options:

=over

=item B<-r>|B<--root> I<root-directory>

Chdir to I<root-directory> before doing anything. This means that all other
directories (including, but not limited to I<mkgen-directory> and
I<project-directory>) should be relative to I<root-directory>. At least
I hope that's what it does.

=item B<-M>|B<--mkgendir> I<mkgen-directory>

Specify path to directory containing F<mkgen.pl> and F<platform> and
F<config> subdirectories. Should be relative to I<root-directory> if
it's specified.

=item B<-p>|B<--platform> I<platform>

Build makefiles for specified platform. To specify more platforms, use this
option multiple times. If this option is not specified, mkgen will try to
guess local platform.

=item B<-v>|B<--verbose>

Be more verbose. This option may be specified multiple times.

=back

=head1 TODO

Macro arguments.

Cleaning of generated files.

=head1 KNOWN BUGS

None known, many unknown.

=cut


use strict ;
use warnings ;

use Getopt::Long ;
use File::Find ;
use Cwd ;
use Config ;

use FindBin ;
use lib "$FindBin::Bin/lib" ;
use mkgen_parser ;
use dependency_cache ;
#use source_parser ;
use makefile_writer ;


my $version = "0.95" ;


###
# guess_platform
sub guess_platform() {
    if ( $Config{ 'osname' } eq 'linux' ) {
        return 'unix' ;
    }
    if ( $Config{ 'osname' } eq 'cygwin' ) {
        # FIXME: msvc or cygwin? Or both?

        return 'cygwin' ;
    }
    if ( $Config{ 'osname' } eq 'MSWin32' ) {
        # FIXME: copied from another script. Is this returned by Active Perl?

        return 'msvc' ;
    }

    # Unknown platform. Please update guess_platform() if this happens.
    # Check file Config.pm somewhere in /usr/lib/perl*.
    # Or use environment variables.

    return undef ;
}


###
# parse_platform_description
#
# platform_file - name of file containing platform description.
# state - mkgen state to read and modify
#
# Parse platform_file and update mkgen state.
sub parse_platform_description( $\% ) {
    my $platform_file = shift ;
    my $state = shift ;
    
    # Open file.
    
    my $handle = new FileHandle( $platform_file ) ;
    die "Can't open '$platform_file', stopped" unless defined( $handle ) ;

    # Parse it.

    my %replacements ;
    $$state{ REPLACEMENTS } = \%replacements  ;

    while ( $_ = $handle->getline() ) {
        study ;

        # Skip comments and empty lines.
        
        next if /^\s*(#|$)/ ;

        # Name of prefix file.

        $$state{ MAKEFILE_PREFIX } = $1, next
            if /^prefix\s*=\s*(\S+)/ ;

        # Name of postfix file.
        
        $$state{ MAKEFILE_POSTFIX } = $1, next
            if /^postfix\s*=\s*(\S+)/ ;

        # Platform suffix.
        
        $$state{ PLATFORM_SUFFIX} = $1, next
            if /^suffix\s*=\s*(\S+)/ ;

        # Replacement strings.

        $replacements{ READ }{ $1 } = $2, next
            if /^(\S+)\s*->\s*(\S+)/ ;
        $replacements{ WRITE }{ $1 } = $2, next
            if /^(\S+)\s*=>\s*(\S+)/ ;
        $replacements{ FILE }{ $1 } = $2, next
            if /^(\S+)\s*#i->\s*(\S+)/ ;

        # separate_lib_paths flag.
        
        $$state{ SEPARATE_LIB_PATHS }++, next
            if /^separate_lib_paths/ ;

        # Unknown decription line.
        
        print STDERR "Unknown platform description line: '$_'.\n" ;
    }

    # Done.

    $handle->close() ;
}


###
# process_platform
#
# platform_file - name of file with platform description.
# state_base - basic mkgen state conataining global information such as
#              mkgen version, path to mkgen and platform descriptions, ...
# mode - "generate" or "clean"
# makefile_gens - list of filenames of all makefile.gen files found
#
# Generate makefiles for given platform.
sub process_platform( $\%$\@ ) {
    my $platform_file = shift ;
    my $state_base = shift ;
    my $mode = shift ;
    my $makefile_gens = shift ;

    # Copy state.

    my %state = %$state_base ;
    
    # Get platform name.

    my( $platform ) = ( $platform_file =~ m|^.*/([^/]*)\.platform$| ) ;
    $state{ PLATFORM } = $platform ;

    # Write info message.
    
    print "Generating makefiles for '$platform'...\n"
        if $state{ VERBOSE } >= 1 and $mode eq 'generate';
    print "Cleaning '$platform' makefiles...\n"
        if $state{ VERBOSE } >= 1 and $mode eq 'clean';

    # Parse platform description.

    parse_platform_description( $platform_file, %state ) ;

    # Prepare cache.

    load_cache( %state ) ;

    # For each makefile.gen found.
    
    foreach my $makefile_gen ( @{$makefile_gens} ) {
        # Check mode to see what should be done.
        
        if ( $mode eq 'generate' ) {
            ###
            # Parse makefile.gen and generated makefile.platform_suffix.

            # Copy state.

            my %local_state = %state ;

            # Parse makefile.gen.

            read_makefile_gen( $makefile_gen, %local_state ) ;

            # Find source prerequisities.

            find_prerequisities( %local_state ) ;

            # Write makefile.

            write_makefile( %local_state ) ;
        }
        elsif( $mode eq 'clean' ) {
            ###
            # Remove makefile.platform_suffix.

            # Create name of makefile.
            
            my $filename = $makefile_gen ;
            $filename =~ s/\.gen$/\.$state{ PLATFORM_SUFFIX }/ ;

            # Unlink it if it exists.
            
            if ( -f $filename ) {
                print "Removing '$filename'\n" if $state{ VERBOSE } >= 2 ;
                unlink( $filename ) ;
            }
        }
    }

    # Save cache.

    save_cache( %state ) ;
}


###
# Main.

# Global variables.

my $opt_mode = "generate" ;
my $opt_root = "." ;
my $opt_mkgendir = "mkgen" ;
my @opt_platforms = () ;
my $opt_shared = 1 ;
my $opt_verbose =  0 ;

# Options.

Getopt::Long::Configure( "bundling" ) ;
GetOptions(
    # Mode
    'g|generate'    => sub { $opt_mode = 'generate' },
    'c|clean'       => sub { $opt_mode = 'clean' },
    'r|root=s'      => \$opt_root,
    'M|mkgendir=s'  => \$opt_mkgendir,
    'p|platform=s'  => \@opt_platforms,
    'shared!'       => \$opt_shared,
    'v|verbose+'    => \$opt_verbose,
) or die "Invalid options, stopped" ;

# Add trailing slash to mkgendir and rootdir.

$opt_mkgendir =~ s|([^/])$|$1/| ;
$opt_root =~ s|([^/])$|$1/| ;

# Chdir to root (FIXME: probably wrong).

my $original_dir = cwd() ;
chdir( $opt_root ) or die "Can't chdir to root '$opt_root', stopped" ;

# Create basic mkgen state.

my %state ;
$state{ VERBOSE } = $opt_verbose ;
$state{ MKGENDIR } = $opt_mkgendir ;
$state{ MKCONFSUBDIR } = "config/" ;
$state{ PLATFORMSUBDIR } = "platform/" ;
$state{ VERSION } = $version ;

# Find all makefile templates.

my @project_dirs = @ARGV ;
@project_dirs = ( "." ) unless scalar( @project_dirs ) ;
my @makefile_gens ;
find( sub { push( @makefile_gens, $File::Find::name ) if /^makefile\.gen$/ ; }, @project_dirs ) ;
die "No 'makefile.gen' found, stopped" unless scalar( @makefile_gens ) ;

# Try to guess local platform unless it has been specified.

unless ( scalar( @opt_platforms ) ) {
    my $platform = guess_platform() ;
    die "Unable to guess platform, modify function guess_platform in mkgen.pl."
        unless defined( $platform ) ;
    push( @opt_platforms, $platform ) ;
}

# Find all platform descriptions, filtered by opt_platforms.

my %platforms_found ;
my @platform_files ;
find(
    sub {
        my $pass ; # use this platform?

        # Filter platforms.
            
        foreach my $platform ( @opt_platforms ) {
            if ( /^\Q$platform\E\.platform$/ ) {
                $pass = 1 ;
                $platforms_found{ $platform }++ ;
            }
        }

        # New platform if the file passed the tests.
        
        push( @platform_files, $File::Find::name ) if $pass ;
    }, $state{ MKGENDIR } . $state{ PLATFORMSUBDIR } ) ;

# Print warning for each requested platform that has not been found.

foreach my $platform ( @opt_platforms ) {
    warn( "Platform description '$platform.platform' not found.\n" )
        unless $platforms_found{ $platform } ;
}
die "No platform descriptions found, stopped" unless @platform_files ;

# Generate makefiles for each platform.

foreach my $platform_file ( @platform_files ) {
    process_platform( $platform_file, %state, $opt_mode, @makefile_gens ) ;
}

# Done.

chdir( $original_dir ) or die "Can't chdir back to '$original_dir', stopped" ;

# EOF #
