/** @file balancer.cpp
 * Balancer implementation.
 */

/*
 * begin        : Tue Oct 14 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "balance/balancer.h"
#include "object/object_status.h"
#include "object/object_manager.h"
#include "object/object_property_data.h"
#include "pointer/object_pointer.h"
#include "pointer/pointer_to_object.h"
#include "system/system.h"
#include "global/global.h"
#include "utility/debug.h"
#include <algorithm>
#include <limits>


/** @def LOG_BALANCER
 * Define this to enable Balancer logging.
 */
//#define LOG_BALANCER

/** @def NO_DATA_FLOW_HISTORY
 * Define this to disable utilization of data flow history.
 * If defined each data flow run is isolated and ignores samples from
 * previous runs.
 */
//#define NO_DATA_FLOW_HISTORY


namespace Massiv {
namespace Core {

namespace {


const int MAX_DUMP_OBJECTS = 4;


/** Converts @c num_bits to bytes. */
int to_bytes
    (
    int num_bits
    )
    {
    return ( num_bits + 7 ) / 8;
    }


/** Clears a map container that associates keys with pointers to owned objects. */
template< typename T >
void clear_map
    (
    T & container
    )
    {
    }


/** ObjectOperation that dumps ids of enumerated objects. */
struct DumpMigrationGroup : public GarbageCollector::ObjectOperation
    {

    /** Constructor. */
    DumpMigrationGroup
        (
        std::ostream &  stream
        ) : stream( stream ),
            num_objects( 0 )
        {
        stream << "{ ";
        }

    /** Destructor. */
    ~DumpMigrationGroup()
        {
        stream << " } (" << num_objects << " object(s))";
        }

    /** Dumps @c object's object id. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        if( num_objects < MAX_DUMP_OBJECTS )
            {
            if( num_objects > 0 )
                {
                stream << ", ";
                }

            stream << object->get_object_id();
            }
        else
        if( num_objects == MAX_DUMP_OBJECTS )
            {
            if( num_objects > 1 )
                {
                stream << ", ";
                }

            stream << "...";
            }

        num_objects++;
        }

    std::ostream &  stream;
        /**< The stream. */
    int             num_objects;
        /**< Number of objects within the migration group. */

    };

} // anonymous namespace


/**
 * Gathers information about a migration group.
 * Defined in the main namespace so that it could be ObjectProperty's friend.
 */
struct GatherMigrationGroupInfo : public GarbageCollector::ObjectOperation
    {

    /** Constructor. */
    GatherMigrationGroupInfo() :
        has_events( false ),
        no_balance( false ),
        no_migrate( false ),
        has_pinned_objects( false ),
        has_stack_referenced_objects( false ),
        has_replicas( false ),
        has_replica_requests( false ),
        size( 0 ),
        triggered( false )
        {
        }

    /** Tests @c object. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        const MetaObject * metaobject =
            object->get_object_id().get_class_type_id().get_metaobject_no_throw();

        if( metaobject && metaobject->get_attribute_no_balance() )
            {
            no_balance = true;
            }

        if( metaobject && metaobject->get_attribute_no_migrate() )
            {
            no_migrate = true;
            }

        SystemTime my_touch_time = object->get_touch_time();

        if( my_touch_time > touch_time )
            {
            touch_time = my_touch_time;
            }

        if( object->has_scheduled_event() )
            {
            STime my_event_time = object->get_time_of_first_event();

            if( !triggered || my_event_time < first_event_time )
                {
                first_event_time = my_event_time;
                }

            has_events = true;
            }

        if( object->is_pinned() )
            {
            has_pinned_objects = true;
            }

        if( object->is_stack_referenced() )
            {
            has_stack_referenced_objects = true;
            }

        if( object->is_replica() )
            {
            has_replicas = true;
            }

        if( object->has_replica_requests() )
            {
            has_replica_requests = true;
            }

        size += to_bytes( object->bitstream_size() );

        triggered = true;
        }

    SystemTime  touch_time;
        /**< Maximum of enumerated objects' touch times. */
    STime       first_event_time;
        /**< Time of the first event scheduled to some object from the group. */
    bool        has_events : 1;
        /**< Are there any scheduled events for objects in the group? */
    bool        no_balance : 1;
        /**< Does migration group contain a "no-balance" object? */
    bool        no_migrate : 1;
        /**< Does migration group contain a "no-migrate" object? */
    bool        has_pinned_objects : 1;
        /**< Are there any pinned objects in the group? */
    bool        has_stack_referenced_objects : 1;
        /**< Are there any stack-strong referenced objects in the group? */
    bool        has_replicas : 1;
        /**< Is this a migration group of object replicas? */
    bool        has_replica_requests : 1;
        /**< Are there any objects replicated to other nodes in the group? */
    int         size;
        /**< Migration group size. */
    bool        triggered;
        /**< Was operation triggered? */

    };


namespace {

/**
 * Gathers statistics information about objects registered on the system.
 * For each enumerated object determines if its migration group could be
 * swapped-out.
 */
struct OptimizeMemory : public GarbageCollector::ObjectOperation
    {

    /** Constructor. */
    OptimizeMemory
        (
        Float32     object_timeout
        ) : object_timeout( object_timeout ),
            memory_load( 0 ),
            num_objects( 0 )
        {
        }

    /** Tests if migration group of @c object can be swapped-out. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        GarbageCollector &          gc = Global::garbage_collector();
        SystemTime                  current_time = SystemTime::current_time();
        GatherMigrationGroupInfo    operation;

        // Update state.

        memory_load += object->memory_size();
        num_objects++;

        // Get object's migration group id.
        // If @c object belongs to a not yet explored migration
        // group @c operation (which gathers information about the migration
        // group) will be triggered.

        if( !object->is_consistent() )
            {
            // Migration groups of inconsistent objects can not be enumerated.

            return;
            }

        gc.get_migration_group_id( object, operation );

        if( !operation.triggered )
            {
            // Migration group of @c object was already processed.

            return;
            }

        // Can the group be swapped-out?

        if( operation.no_balance || operation.has_pinned_objects ||
            operation.has_stack_referenced_objects ||
            operation.has_replicas || operation.has_replica_requests )
            {
            // Migration group contains a "no-balance", pinned object, object
            // replica, or object replicated to different node.

            return;
            }

        if( operation.touch_time + object_timeout > current_time )
            {
            // All the objects in the migration group were not touched
            // for at least @c object_timeout seconds.

            return;
            }

        if( operation.has_events && operation.first_event_time <
            current_time + object_timeout )
            {
            // There is at least one object in the migration group with
            // an event to be fired during the next @c object_timeout seconds.

            return;
            }

        groups_to_swap_out.push( object->get_object_id() );
        }

    Float32                         object_timeout;
        /**< Object timeout. */
    Int32                           memory_load;
        /**< Memory load. */
    Int32                           num_objects;
        /**< Enumerated objects count. */
    GarbageCollector::ObjectIdQueue groups_to_swap_out;
        /**< Object id's of migration groups representatives that can be
             swapped-out. */

    };


/**
 * Chooses random subset of all registered objects. Each object is
 * chosen with the same probability, "no-migrate" or "no-balance"
 * migration groups are rejected.
 */
struct RandomChooseObjects : public GarbageCollector::ObjectOperation
    {

    /** Constructor. */
    RandomChooseObjects
        (
        Float32 probability
        ) : threshold( static_cast<Int32>( probability * RAND_MAX ) )
        {
        }

    /** Puts @c object to @c object with given probability. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        GarbageCollector &          gc = Global::garbage_collector();
        GatherMigrationGroupInfo    operation;

        // Reject object with probability @c 1.0f - probability.

        if( std::rand() > threshold )
            {
            return;
            }

        // Get object's migration group id.
        // If @c object belongs to a not yet explored migration
        // group @c operation (which gathers information about the migration
        // group) will be triggered.

        if( !object->is_consistent() )
            {
            // Migration groups of inconsistent object replicas can not be
            // enumerated.

            return;
            }

        gc.get_migration_group_id( object, operation );

        if( !operation.triggered )
            {
            // Migration group of @c object was already processed,
            // either rejected or chosen.

            return;
            }

        // Check migration group of object if it can migrate.

        if( operation.no_balance || operation.no_migrate ||
            operation.has_pinned_objects || operation.has_stack_referenced_objects ||
            operation.has_replicas )
            {
            return;
            }

        objects.push( object );
        }

    GarbageCollector::ObjectQueue   objects;
        /**< Chosen objects. */
    Int32                           threshold;
        /**< Random value with uniform distribution over [ 0, RAND_MAX ]
             threshold. */

    };


} // anonymous namespace


// Implementation of Balancer::BalancerSettings
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::BalancerSettings::BalancerSettings() :
    Settings( "Balancer" ),
    enabled( access_boolean( "enabled", true ) ),
    data_flow_optimization_enabled( access_boolean( "data_flow_optimization_enabled", true ) ),
    resource_load_optimization_enabled( access_boolean( "resource_load_optimization_enabled", true ) ),
    memory_load_optimization_enabled( access_boolean( "memory_load_optimization_enabled", true ) ),
    min_data_run_period( access_float( "min_data_run_period", 5.0f ) ),
    max_data_run_period( access_float( "max_data_run_period", 120.0f ) ),
    resource_run_period( access_float( "resource_run_period", 10.0f ) ),
    memory_run_period( access_float( "memory_run_period", 60.0f ) ),
    samples_per_run_fraction( access_float( "samples_per_run_fraction", 0.02f ) ),
    size_of_object_samples_fraction( access_float( "size_of_object_samples_fraction", 0.2f ) ),
    data_flow_fraction( access_float( "data_flow_fraction", 0.75f ) ),
    data_flow_aging_coef( access_float( "data_flow_aging_coef", 0.9f ) ),
    object_timeout( access_float( "object_timeout", 60.0f ) ),
    cpu_load_weight( access_float( "cpu_load_weight", 0.5f ) ),
    rank_compare_factor( access_float( "rank_compare_factor", 0.6f ) ),
    balance_object_count_fraction( access_float( "balance_object_count_fraction", 0.05f ) )
    {
    }


// Implementation of Balancer::LocalStatistics
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::LocalStatistics::LocalStatistics() :
    Statistics( "Balancer/Local", Status::FACILITY_BALANCER,
        Statistics::OPEN_NORMAL ),
    last_data_run( create_float( "last_data_run", 0.0f ) ),
    last_memory_run( create_float( "last_memory_run", 0.0f ) ),
    last_resource_run( create_float( "last_resource_run", 0.0f ) ),
    current_data_run_period( create_float( "current_data_run_period", 5.0f ) ),
    min_recent_data_size( create_integer( "min_recent_data_size", 0 ) ),
    max_recent_data_size( create_integer( "max_recent_data_size", 0 ) ),
    data_size_threshold( create_integer( "data_size_threshold", 0 ) ),
    size_of_object_samples( create_integer( "size_of_object_samples", 0 ) ),
    num_samples( create_integer( "num_samples", 0 ) ),
    consumed_time( create_float( "consumed_time", 0.0f ) )
    {
    }


// Implementation of Balancer::GlobalStatistics
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::GlobalStatistics::GlobalStatistics() :
    Statistics( "Balancer/Global", Status::FACILITY_BALANCER,
        Statistics::OPEN_NORMAL ),
    num_total_objects( create_integer( "num_total_objects", 0 ) ),
    total_memory_load( create_integer( "total_memory_load", 0 ) )
    {
    }


// Implementation of Balancer::NodeStatistics
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::NodeStatistics::NodeStatistics
    (
    const NodeId & node
    ) : Statistics( ( "Balancer/Nodes/" + node.to_identifier() ).c_str(),
            Status::FACILITY_BALANCER, Statistics::OPEN_OVERWRITE ),
        memory_load( create_integer( "memory_load", 0 ) ),
        num_objects( create_integer( "num_objects", 0 ) ),
        object_count_fraction( create_float( "object_count_fraction", 0.0f ) ),
        cpu_load( create_float( "cpu_load", 0.0f ) ),
        resource_load_rank( create_float( "resource_load_rank", 0.0f ) )
    {
    }


// Implementation of Balancer::ObjectToObjectSample
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::ObjectToObjectSample::ObjectToObjectSample
    (
    const ObjectId &    producer,
    const ObjectId &    consumer,
    int                 size
    ) : producer( producer ),
        consumer( consumer ),
        size( size )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                          stream,
    const Balancer::ObjectToObjectSample &  sample
    )
    {
    stream << sample.producer << " to " << sample.consumer
        << " (" << sample.size << " byte(s))";

    return stream;
    }


// Definition of Balancer::GroupSample::ObjectSample::Less
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

struct Balancer::GroupSample::ObjectSample::Less
    {

    /** Less comparator. */
    bool operator()
        (
        const ObjectSample &    sample,
        const ObjectId &        consumer
        ) const
        {
        return less( sample.consumer, consumer );
        }
    
    /** Less comparator. */
    bool operator()
        (
        const ObjectId &        consumer,
        const ObjectSample &    sample
        ) const
        {
        return less( consumer, sample.consumer );
        }

    /** Less comparator. */
    bool operator()
        (
        const ObjectSample &    sample1,
        const ObjectSample &    sample2
        )
        {
        return less( sample1.consumer, sample2.consumer );
        }

    ObjectId::Less less;
        /**< ObjectId less predicate. */

    };


// Implementation of Balancer::GroupSample::ObjectSample
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::GroupSample::ObjectSample::ObjectSample
    (
    const ObjectId &    consumer,
    int                 size
    ) : consumer( consumer ),
        size( size )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                              stream,
    const Balancer::GroupSample::ObjectSample & sample
    )
    {
    stream << sample.consumer << " (" << sample.size << " byte(s))";
    return stream;
    }


// Definition of Balancer::GroupSample::NodeSample::Less
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

struct Balancer::GroupSample::NodeSample::Less
    {

    /** Less comparator. */
    bool operator()
        (
        const NodeSample &    sample,
        const NodeId &        node
        ) const
        {
        return less( sample.node, node );
        }
    
    /** Less comparator. */
    bool operator()
        (
        const NodeId &        node,
        const NodeSample &    sample
        ) const
        {
        return less( node, sample.node );
        }

    /** Less comparator. */
    bool operator()
        (
        const NodeSample &    sample1,
        const NodeSample &    sample2
        ) const
        {
        return less( sample1.node, sample2.node );
        }

    NodeId::Less less;
        /**< NodeId less predicate. */

    };


// Implementation of Balancer::GroupSample::NodeSample
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::GroupSample::NodeSample::NodeSample
    (
    const NodeId &  node,
    int             size
    ) : node( node ),
        size( size )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                              stream,
    const Balancer::GroupSample::NodeSample &   sample
    )
    {
    stream << sample.node << " (" << sample.size << " byte(s))";
    return stream;
    }


// Implementation of Balancer::GroupSample
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::GroupSample::GroupSample() :
    total_external_size( 0 ),
    total_internal_size( 0 )
    {
    }


void Balancer::GroupSample::account
    (
    const ObjectToObjectSample &    sample,
    MigrationGroupId                group_id
    )
    {
    GarbageCollector &  gc = Global::garbage_collector();
    ObjectManager &     object_manager = Global::object_manager();
    const ObjectId &    consumer = sample.consumer;
    ObjectProperty *    consumer_object = gc.find_local_object( consumer );

    if( group_representative.is_null() )
        {
        // Set group representative object.

        group_representative = sample.producer;
        }

    // Guess node containing @c sample.consumer object.
    // Since we have been talking to that object recently the
    // guess can not be "too bad".

    NodeId node;

    if( consumer_object == NULL )
        {
        ObjectStatus object_status;

        // Object is remote.

        object_manager.get_remote_object_status
            (
            consumer,
            object_status,
            node
            );

        if( object_status.get_value() != ObjectStatus::ALIVE )
            {
            // @c sample can not be accounted.

            return;
            }
        }
    else
        {
        // Object is local.

        if( !consumer_object->is_consistent() )
            {
            // Do not account this sample as consumer is inconsistent.

            return;
            }

        if( gc.get_migration_group_id( consumer_object ) == group_id )
            {
            // Do not account this sample as producer and consumer belong
            // to the same migration group.

            total_internal_size += sample.size;
            return;
            }

        node = System::get_local_node_id();
        }

    // Account @c sample to the consumer.

    #ifdef LOG_BALANCER

    ObjectSample::Less      o_predicate;
    ObjectSamples::iterator o_it =
        std::lower_bound
            (
            object_samples.begin(),
            object_samples.end(),
            consumer,
            o_predicate
            );

    if( o_it == object_samples.end() || o_it->consumer != consumer )
        {
        // Add new object sample.

        object_samples.insert( o_it, ObjectSample( consumer, sample.size ) );
        }
    else
        {
        // Update existing object sample.

        o_it->size += sample.size;
        }

    #endif

    // Account @c sample to the consumer's node.

    NodeSample::Less        n_predicate;
    NodeSamples::iterator   n_it =
        std::lower_bound
            (
            node_samples.begin(),
            node_samples.end(),
            node,
            n_predicate
            );

    if( n_it == node_samples.end() || n_it->node != node )
        {
        // Add new node sample.

        node_samples.insert( n_it, NodeSample( node, sample.size ) );
        }
    else
        {
        // Update existing node sample.

        n_it->size += sample.size;
        }

    // Account @c sample to this.

    total_external_size += sample.size;
    }


void Balancer::GroupSample::optimize_data_flow()
    {
    Balancer &          balancer = Global::balancer();
    GarbageCollector &  gc = Global::garbage_collector();
    Float32             flow_fraction;

    if( !node_samples.size() )
        {
        return;
        }

    // Determine if there is a remote node that consumes the majority
    // of migration group's data flow.
    // Utilize that @c node_samples are sorted by data flow size.

    if( System::get_local_node_id() == node_samples[ 0 ].node )
        {
        return;
        }

    flow_fraction = static_cast<Float32>( node_samples[ 0 ].size ) / total_external_size;

    if( flow_fraction < balancer.settings.data_flow_fraction.read() )
        {
        return;
        }

    // Test if the second most major consumer node consumes substantially
    // less data than the most major node.
    // This condition is a heuristics only and might not be actually needed.

    if( node_samples.size() >= 2 )
        {
        flow_fraction =
            static_cast<Float32>( node_samples[ 1 ].size ) / node_samples[ 0 ].size;

        if( flow_fraction > balancer.settings.data_flow_fraction.read() )
            {
            return;
            }
        }

    // Test if migration group is still on the current node and is consistent.

    ObjectProperty * object = gc.find_local_object( group_representative );

    if( object == NULL || !object->is_consistent() )
        {
        return;
        }

    // Try migrating the group to the most major node.
    // This globally saves at least @c data_flow_save bytes of data flow
    // generated by this migration group (data flow consumed by this group
    // will be redirected to the new node without any important penalty)
    // under a condition that data flow in the future will be statistically
    // similar to the gathered results.
    // How often the migration takes place does not matter as long as
    // the migration is cheaper than leaving the group on the current node
    // and the target node is ready to accept the group.

    GatherMigrationGroupInfo group_info;
    gc.enumerate_migration_group( object, group_info );

    const NodeId &  node = node_samples[ 0 ].node;
    int             migration_benefit = node_samples[ 0 ].size;
    int             migration_cost = group_info.size;
    int             data_flow_save = migration_benefit - migration_cost;

    if( !node.is_type( NodeId::TYPE_SERVER ) )
        {
        return;
        }

    if( data_flow_save < 0 )
        {
        #ifdef LOG_BALANCER

        std::ostringstream ss;

        ss  << "Migration of " << group_representative << " to " << node
            << " rejected because migration_cost = " << migration_cost
            << ", migration_benefit = " << migration_benefit << ".";

        Global::log_debug
            (
            Status::FACILITY_BALANCER,
            Status::PRIORITY_NORMAL,
            ss.str()
            );

        #endif

        return;
        }

    if( balancer.get_resource_info( node ).is_more_loaded_than(
        balancer.get_local_resource_info() ) )
        {
        return;
        }

    if( group_info.no_balance || group_info.no_migrate ||
        group_info.has_pinned_objects || group_info.has_stack_referenced_objects ||
        group_info.has_replicas )
        {
        return;
        }

    // Log.

    std::ostringstream ss;

    ss << "Data flow optimization: Migrating group of " << object->get_object_id()
        << " to " << node << " and saving " << data_flow_save << " byte(s).";

    Global::log_debug
        (
        Status::FACILITY_BALANCER,
        Status::PRIORITY_LOW,
        ss.str()
        );

    // Migrate.

    UpcallToObject upcall = object;

    upcall.tag_upcall( "::Massiv::Core::Balancer::GroupSample::optimize_data_flow()" );
    upcall->migrate_to( node, STime::current_time(), false );
    }


void Balancer::GroupSample::apply_aging()
    {
    Balancer &  balancer = Global::balancer();
    Float32     aging = balancer.settings.data_flow_aging_coef;
    std::size_t i;

    // Apply aging to gathered data flow.
    // The aging is applied to @c object_samples which due to sum distributivity
    // easily propagates to @c node_samples, @c total_extrernal_size and
    // @c total_internal_size.

    #ifdef LOG_BALANCER

    for( i = 0; i < object_samples.size(); i++ )
        {
        ObjectSample & sample = object_samples[ i ];

        sample.size = static_cast<Int32>( sample.size * aging );

        if( sample.size < 1 )
            {
            object_samples.erase( object_samples.begin() + i-- );
            }
        }

    #endif

    for( i = 0; i < node_samples.size(); i++ )
        {
        NodeSample & sample = node_samples[ i ];

        sample.size = static_cast<Int32>( sample.size * aging );

        if( sample.size < 1 )
            {
            node_samples.erase( node_samples.begin() + i-- );
            }
        }

    total_external_size = static_cast<Int32>( total_external_size * aging );
    total_internal_size = static_cast<Int32>( total_internal_size * aging );
    }


std::ostream & operator<<
    (
    std::ostream &                  stream,
    const Balancer::GroupSample &   sample
    )
    {
    GarbageCollector &  gc = Global::garbage_collector();
    ObjectProperty *    object = gc.find_local_object( sample.group_representative );
    std::size_t         i;

    if( object != NULL )
        {
        DumpMigrationGroup operation( stream );

        if( object->is_replica() )
            {
            stream << "replica of " << object->get_object_id();
            }
        else
            {
            hope( object->is_consistent() );
            gc.enumerate_migration_group( object, operation );
            }
        }
    else
        {
        stream << sample.group_representative;
        }

    stream << ":";

    #ifdef LOG_BALANCER

    if( sample.object_samples.size() )
        {
        stream << "\nto objects:";

        for( i = 0; i < sample.object_samples.size(); i++ )
            {
            stream << '\n' << sample.object_samples[ i ];
            }
        }

    #endif

    if( sample.node_samples.size() )
        {
        stream << "\nto nodes:";

        for( i = 0; i < sample.node_samples.size(); i++ )
            {
            stream << '\n' << sample.node_samples[ i ];
            }
        }

    stream << "\n(external: " << sample.object_samples.size() << " object(s), " <<
        sample.node_samples.size() << " node(s), " <<
        sample.total_external_size << " byte(s), internal: " <<
        sample.total_internal_size << " byte(s))";

    return stream;
    }


// Implementation of Balancer::ResourceLoadInfo
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::ResourceLoadInfo::ResourceLoadInfo
    (
    const NodeId & node
    ) : NodeStatistics( node ),
        node( node )
    {
    }


void Balancer::ResourceLoadInfo::evaluate_local_load()
    {
    GarbageCollector &  gc = Global::garbage_collector();
    Balancer &          balancer = Global::balancer();

    hope( node == System::get_local_node_id() );

    // Evaluate CPU load (compare elapsed time to total sleep time).
    // The formula is correct as long as we are the only process
    // running on that CPU.
    // To prevent jittering, actual value written to the statistics
    // is a blend of the previous value and the newly computed one.

    SystemTime  current_time = System::system_time();
    SystemTime  total_sleep_time = System::total_sleep_time();
    Float32     new_cpu_load;

    if( current_time > balancer.previous_system_time )
        {
        new_cpu_load = static_cast<Float32>( 1.0f - ( total_sleep_time -
            balancer.previous_total_sleep_time ).get_value() / ( current_time -
            balancer.previous_system_time ).get_value() );
        }
    else
        {
        new_cpu_load = 0.0f;
        }

    balancer.previous_system_time = current_time;
    balancer.previous_total_sleep_time = total_sleep_time;

    cpu_load = ( 3.0f * cpu_load + new_cpu_load ) / 4.0f;

    // Rank local node's resource load.
    // Balance between CPU load and object count fraction.

    if( balancer.settings.cpu_load_weight < 0.0f ||
        balancer.settings.cpu_load_weight > 1.0f )
        {
        panic( "Invalid Balancer/cpu_load_weight." );
        }

    resource_load_rank = balancer.settings.cpu_load_weight * cpu_load + ( 1.0f -
        balancer.settings.cpu_load_weight ) * object_count_fraction;

    // Evaluate number of local objects.

    num_objects = gc.get_statistics().num_objects;
    }


void Balancer::ResourceLoadInfo::distribute()
    {
    hope( node == System::get_local_node_id() );

    // Multicast resource load info to all server nodes.

    ResourceLoadMessage message( num_objects, memory_load, cpu_load,
        resource_load_rank );

    message.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
    }


bool Balancer::ResourceLoadInfo::is_more_loaded_than
    (
    const ResourceLoadInfo & info
    ) const
    {
    return resource_load_rank > info.resource_load_rank;
    }


std::ostream & operator<<
    (
    std::ostream &                      stream,
    const Balancer::ResourceLoadInfo &  info
    )
    {
    stream << info.node << ":\n";
    stream << static_cast<const Balancer::NodeStatistics &>( info );

    return stream;
    }


// Implementation of Balancer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

Balancer::Balancer() :
    initialized( false ),
    enabled( false )
    {
    }


Balancer::~Balancer()
    {
    if( initialized )
        {
        shutdown();
        }
    }


void Balancer::initialize()
    {
    hope( !initialized );

    local_statistics.reset();
    global_statistics.reset();

    initialized = true;
    enabled = true;
    }


void Balancer::shutdown()
    {
    hope( initialized );

    enabled = false;
    object_samples.clear();
    group_samples.clear();
    clear_map( resource_infos );
    least_loaded_node = NodeId();
    most_loaded_node = NodeId();

    initialized = false;
    }


std::ostream & operator<<
    (
    std::ostream &      stream,
    const Balancer &    balancer
    )
    {
    std::size_t i;

    if( !balancer.group_samples.size() )
        {
        if( balancer.object_samples.size() )
            {
            stream << "Object-to-object data flow samples:";

            for( i = 0; i < balancer.object_samples.size(); i++ )
                {
                stream << '\n' << i << ": " << balancer.object_samples[ i ];
                }
            }
        }
    else
        {
        stream << "\nMigration group data flow samples:";

        for( i = 0; i < balancer.group_samples.size(); i++ )
            {
            stream << '\n' << i << ": " << balancer.group_samples[ i ];
            }
        }

    stream << "\nResource loads:";

    Balancer::ResourceLoadInfos::const_iterator it =
        balancer.resource_infos.begin();

    while( it != balancer.resource_infos.end() )
        {
        stream << '\n' << *it->second;
        ++it;
        }

    return stream;
    }


void Balancer::account_for_local_call
    (
    ObjectProperty *    callee,
    int                 packet_size
    )
    {
    GarbageCollector &  gc = Global::garbage_collector();
    ObjectProperty *    caller = gc.who( true );

    packet_size *= 2; // A magic constant to strenghten the sample.

    if( caller != NULL )
        {
        // @c caller might not be identified reliably.
        // That's because @c GC::who() keeps track of objects directly upcalled
        // from the core only. If upcalled object then directly calls another
        // object the call will not be "registered" by GC.
        // However the good news is that @c caller is always from the "correct
        // migration group" (unless weak pointers are used to call local objects).
        // Also note that calls issued by object replicas will be accounted to
        // non-replica objects that invoked code on that replicas.

        add_sample
            (
            caller->get_object_id(),
            callee->get_object_id(),
            packet_size
            );

        // Touch objects involved in the call.

        caller->touch_object();
        callee->touch_object();
        }
    }


void Balancer::account_for_object_migration
    (
    ObjectProperty *    object,
    const ObjectId &    addressee
    )
    {
    GarbageCollector &  gc = Global::garbage_collector();
    ObjectProperty *    initiator = gc.who( true );
    ObjectProperty *    target = gc.find_local_object( addressee );

    if( initiator != NULL )
        {
        int data_size = to_bytes( object->bitstream_size() );

        // Note: @c initiator might not be identified reliably.
        // See @c account_for_local_call() for more info.

        add_sample
            (
            initiator->get_object_id(),
            addressee,
            data_size
            );

        // Touch objects involved in the migration.
        // Both @c initiator and addressee will be touched at @c object
        // delivery time too.

        initiator->touch_object();

        if( target != NULL )
            {
            target->touch_object();
            }
        }
    }


void Balancer::account_for_object_replication
    (
    ObjectProperty *    object,
    const NodeId &      node,
    int                 data_size
    )
    {
    // Add sample.

    data_size *= 2; // A magic constant to strenghten the sample.

    if( node.is_type( NodeId::TYPE_SERVER ) )
        {
        add_sample( object->get_object_id(), node, data_size );
        }

    // Touch object involved in the replication.

    object->touch_object();
    }


void Balancer::add_sample
    (
    const ObjectId &    producer,
    const ObjectId &    consumer,
    int                 data_size
    )
    {
    const MetaObject * metaobject =
        producer.get_class_type_id().get_metaobject_no_throw();

    // Drop sample?

    if( !settings.data_flow_optimization_enabled )
        {
        return;
        }

    if( metaobject && metaobject->get_attribute_no_balance() )
        {
        // If @c producer is a "no-balance" object drop the sample.

        return;
        }

    if( data_size < local_statistics.data_size_threshold.read() )
        {
        // Sample size is too small.

        return;
        }

    // Add new sample.

    object_samples.push_back
        ( ObjectToObjectSample( producer, consumer, data_size ) );

    local_statistics.size_of_object_samples += data_size;
    local_statistics.num_samples++;
    }


void Balancer::aggregate_samples()
    {
    GarbageCollector &  gc = Global::garbage_collector();
    std::size_t         i;

    if( object_samples.size() )
        {
        local_statistics.min_recent_data_size = std::numeric_limits<int>::max();
        }
    else
        {
        local_statistics.min_recent_data_size = 0;
        }

    local_statistics.max_recent_data_size = 0;
    gc.freeze_migration_groups();

    // Update previous @c group_samples.

    #ifndef NO_DATA_FLOW_HISTORY

    for( i = 0; i < group_samples.size(); i++ )
        {
        GroupSample & sample = group_samples[ i ];
        ObjectProperty * object = gc.find_local_object( sample.group_representative );

        if( object != NULL && object->is_consistent() )
            {
            // Enumerate object's migration group and check if its group id
            // matches @c sample position in the array.
            // This is needed by the aggregation algorithm that expects
            // @c sample to be present at @c group_id position.

            MigrationGroupId group_id = gc.get_migration_group_id( object );

            if( group_id != i )
                {
                // Migration groups linked together, split up or became
                // inconsistent since the last run (@c object belongs to an
                // already enumerated group).
                // Previous data flow samples can not be easily reused.

                #ifdef LOG_BALANCER

                Global::log_debug
                    (
                    Status::FACILITY_BALANCER,
                    Status::PRIORITY_NORMAL,
                    "Discarding data flow history because of migration group change(s)."
                    );

                #endif

                group_samples.clear();
                break;
                }

            // Apply aging.

            sample.apply_aging();
            }
        else
            {
            // Object is no longer local or consistent. Kill off the sample.

            group_samples.erase( group_samples.begin() + i-- );
            }
        }

    #else // !NO_DATA_FLOW_HISTORY

    group_samples.clear();

    #endif

    // Aggregate @c object_samples to @c group_samples by migration groups.

    for( i = 0; i < object_samples.size(); i++ )
        {
        const ObjectToObjectSample & sample = object_samples[ i ];
        ObjectProperty * object = gc.find_local_object( sample.producer );

        if( object != NULL && object->is_consistent() )
            {
            // Determine migration group of @c object.

            MigrationGroupId group_id = gc.get_migration_group_id( object );

            if( group_samples.size() <= group_id )
                {
                group_samples.resize( group_id + 1 );
                }

            // Account @c sample to @c group_id.

            group_samples[ group_id ].account( sample, group_id );
            }

        // Update statistics.

        if( local_statistics.min_recent_data_size > sample.size )
            {
            local_statistics.min_recent_data_size = sample.size;
            }

        if( local_statistics.max_recent_data_size < sample.size )
            {
            local_statistics.max_recent_data_size = sample.size;
            }
        }

    object_samples.clear();
    gc.unfreeze_migration_groups();
    }


Balancer::ResourceLoadInfo & Balancer::get_resource_info
    (
    const NodeId & node
    )
    {
    ResourceLoadInfos::value_type::second_type & entry = resource_infos[ node ];

    if( entry == NULL )
        {
        // Resource load info has not been allocated for @c node yet.
        // Allocate now.

        entry = new ResourceLoadInfo( node );
        }

    return *entry;
    }


Balancer::ResourceLoadInfo & Balancer::get_local_resource_info()
    {
    // FIXME: Optimize access to local node resource info.
    return get_resource_info( System::get_local_node_id() );
    }


void Balancer::handle_resource_load_message
    (
    const ResourceLoadMessage & message
    )
    {
    std::ostringstream  ss;

    // Read per-node information from the message.

    const NodeId &      node = message.get_initiator();
    ResourceLoadInfo &  node_info = get_resource_info( node );

    node_info.memory_load = message.get_memory_load();
    node_info.num_objects = message.get_num_objects();
    node_info.cpu_load = message.get_cpu_load();
    node_info.resource_load_rank = message.get_rank();

    // Update global info.

    ResourceLoadInfos::iterator it;
    Float32                     least_rank = 1.0f;
    Float32                     greatest_rank = 0.0f;

    global_statistics.num_total_objects = 0;
    global_statistics.total_memory_load = 0;

    ss << "Resource load infos:";

    for( it = resource_infos.begin(); it != resource_infos.end(); ++it )
        {
        const ResourceLoadInfo & info = *it->second;

        if( info.resource_load_rank < least_rank )
            {
            least_loaded_node = info.node;
            least_rank = info.resource_load_rank;
            }

        if( info.resource_load_rank > greatest_rank )
            {
            most_loaded_node = info.node;
            greatest_rank = info.resource_load_rank;
            }

        global_statistics.num_total_objects += info.num_objects;
        global_statistics.total_memory_load += info.memory_load;
        }

    // Update node infos according to global information.

    for( it = resource_infos.begin(); it != resource_infos.end(); ++it )
        {
        ResourceLoadInfo & info = *it->second;

        if( global_statistics.num_total_objects != 0 )
            {
            info.object_count_fraction = info.num_objects /
                static_cast<Float32>( global_statistics.num_total_objects );
            }
        else
            {
            info.object_count_fraction = 0.0f;
            }

        ss << '\n' << info;
        }

    // Done.

    Global::log_debug
        (
        Status::FACILITY_BALANCER,
        Status::PRIORITY_LOW,
        ss.str()
        );
    }


void Balancer::optimize_data_flow()
    {
    SystemTime  time = SystemTime::current_time();
    std::size_t i;

    // Should optimize?

    if( !System::is_server() )
        {
        return;
        }

    if( !settings.data_flow_optimization_enabled )
        {
        return;
        }

    if( local_statistics.last_data_run + local_statistics.current_data_run_period >
        time.get_value() )
        {
        return;
        }

    local_statistics.last_data_run = static_cast<Float32>( time.get_value() );

    // Aggregate object-to-object data flow samples to migration-group-to-node
    // samples. Migrate objects to the "major" consumer nodes to reduce
    // data flow among different nodes.
    // If we were optimizing data flow only we would end up in a state
    // where all objects have been centered to a single node.
    // @c optimize_resource_load() monitors overall resource load globally
    // and ensures that all nodes will be loaded uniformly.

    aggregate_samples();

    #ifdef LOG_BALANCER

    std::ostringstream ss;

    ss << *this;

    Global::log_debug
        (
        Status::FACILITY_BALANCER,
        Status::PRIORITY_NORMAL,
        ss.str()
        );

    ss.str( std::string() );

    #endif

    for( i = 0; i < group_samples.size(); i++ )
        {
        group_samples[ i ].optimize_data_flow();
        }

    // Schedule next data flow optimization run.

    ResourceLoadInfo & local_info = get_local_resource_info();

    if( local_statistics.num_samples > static_cast<Int32>(
        settings.samples_per_run_fraction * local_info.num_objects ) )
        {
        local_statistics.data_size_threshold =
            ( local_statistics.data_size_threshold + 1 ) * 0x200 / 0x100;

        local_statistics.current_data_run_period =
            local_statistics.current_data_run_period * 0x100 / 0x200;
        }
    else
        {
        local_statistics.data_size_threshold =
            local_statistics.data_size_threshold * 0x150 / 0x200;

        local_statistics.current_data_run_period =
            ( local_statistics.current_data_run_period + 1 ) * 0x200 / 0x100;
        }

    if( local_statistics.size_of_object_samples < static_cast<Int32>(
        settings.size_of_object_samples_fraction *
        local_info.memory_load ) )
        {
        local_statistics.current_data_run_period =
            ( local_statistics.current_data_run_period + 1 ) * 0x420 / 0x200;
        }
    else
        {
        local_statistics.current_data_run_period =
            local_statistics.current_data_run_period * 0x150 / 0x200;
        }

    if( local_statistics.current_data_run_period < settings.min_data_run_period )
        {
        local_statistics.current_data_run_period = settings.min_data_run_period;
        }

    if( local_statistics.current_data_run_period > settings.max_data_run_period )
        {
        local_statistics.current_data_run_period = settings.max_data_run_period;
        }

    // Done.

    local_statistics.size_of_object_samples = 0;
    local_statistics.num_samples = 0;

    #ifdef NO_DATA_FLOW_HISTORY

    group_samples.clear();

    #endif
    }


void Balancer::optimize_resource_load()
    {
    GarbageCollector &  gc = Global::garbage_collector();
    const NodeId &      my_node = System::get_local_node_id();
    SystemTime          time = SystemTime::current_time();

    // Should optimize?

    if( my_node == NodeId() )
        {
        return;
        }

    if( !System::is_server() )
        {
        return;
        }

    if( !settings.resource_load_optimization_enabled )
        {
        return;
        }

    if( local_statistics.last_resource_run + settings.resource_run_period >
        time.get_value() )
        {
        return;
        }

    local_statistics.last_resource_run = static_cast<Float32>( time.get_value() );

    // Distribute local node's resource load info.

    ResourceLoadInfo & local_info = get_local_resource_info();

    local_info.evaluate_local_load();
    local_info.distribute();

    // Test if the least loaded node is substantially less loaded than our
    // node and we are the most loaded node (node resource load ranks are
    // compared). If yes, some of our objects need be migrated to that node
    // in order to keep uniform node load.

    if( least_loaded_node == my_node || most_loaded_node != my_node ||
        least_loaded_node == NodeId() )
        {
        return;
        }

    Float32 our_rank = local_info.resource_load_rank;
    Float32 least_rank = get_resource_info( least_loaded_node ).resource_load_rank;

    if( our_rank * settings.rank_compare_factor < least_rank )
        {
        return;
        }

    // Choose objects that will be migrated to the least loaded node.
    // Currently approximately @c object_count objects will be chosen
    // by random (regardless of data flow statistics). It is okay because
    // @c optimize_data_flow() will tune/correct "bad migrations".
    // @c settings.balance_object_count_fraction must be sufficiently
    // small so that the adaptation would converge.
    // This reduces local node's load.

    std::size_t object_count = static_cast<Int32>(
        global_statistics.num_total_objects *
        settings.balance_object_count_fraction );

    RandomChooseObjects operation
        ( object_count / static_cast<Float32>( local_info.num_objects ) );

    gc.freeze_migration_groups();
    gc.enumerate_all_objects( operation );
    gc.unfreeze_migration_groups();

    if( operation.objects.empty() )
        {
        return;
        }

    // Migrate the objects.

    std::ostringstream  ss;

    ss  << "Resource load optimization: " << least_loaded_node
        << " is substantially less loaded than the local node. "
        << "The following objects will be migrated to that node:";

    while( !operation.objects.empty() )
        {
        ObjectProperty * object = operation.objects.front();
        operation.objects.pop();

        ss << '\n' << object->get_object_id();

        UpcallToObject upcall = object;

        upcall.tag_upcall( "::Massiv::Core::Balancer::optimize_resource_load()" );
        upcall->migrate_to( least_loaded_node, STime(), false );
        }

    // Log.

    Global::log_debug
        (
        Status::FACILITY_BALANCER,
        Status::PRIORITY_NORMAL,
        ss.str()
        );
    }


void Balancer::optimize_memory_load()
    {
    ObjectManager &     object_manager = Global::object_manager();
    GarbageCollector &  gc = Global::garbage_collector();
    SystemTime          time = SystemTime::current_time();

    // Should optimize?

    if( !settings.memory_load_optimization_enabled )
        {
        return;
        }

    if( local_statistics.last_memory_run + settings.memory_run_period >
        time.get_value() )
        {
        return;
        }

    local_statistics.last_memory_run = static_cast<Float32>( time.get_value() );

    // Scan registered objects.
    // Find migration groups that were not used recently.

    gc.freeze_migration_groups();

    OptimizeMemory operation( settings.object_timeout );
    gc.enumerate_all_objects( operation );

    gc.unfreeze_migration_groups();

    // Update local resource load info.

    ResourceLoadInfo & local_info = get_local_resource_info();

    local_info.memory_load = operation.memory_load;
    local_info.num_objects = operation.num_objects;

    // Swap-out the groups and reduce memory load.

    while( !operation.groups_to_swap_out.empty() )
        {
        const ObjectId & object_id = operation.groups_to_swap_out.front();

        object_manager.try_to_swap_out( object_id );

        operation.groups_to_swap_out.pop();
        }
    }


void Balancer::next_tick()
    {
    Statistics::MeasureSystemTime clock( local_statistics.consumed_time );

    // Optimize.

    if( System::inside_simulation() )
        {
        return;
        }

    if( !System::is_enabled( System::CAPABILITY_SIMULATION ) )
        {
        return;
        }

    if( is_load_balancing_enabled() )
        {
        optimize_data_flow();
        optimize_resource_load();
        optimize_memory_load();
        }
    }


void Balancer::enable_load_balancing()
    {
    enabled = true;
    }


void Balancer::disable_load_balancing()
    {
    enabled = false;
    }


bool Balancer::is_load_balancing_enabled() const
    {
    return settings.enabled.read() && enabled;
    }


} // namespace Massiv::Core
} // namespace Massiv
