/** @file balancer.h
 * Balancer.
 */

/*
 * begin        : Tue Oct 14 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BALANCER_H
#define MASSIV_BALANCER_H


#include "global/common.h"
#include "pointer/objectid.h"
#include "pointer/garbage_collector.h"
#include "node/nodeid.h"
#include "registry/statistics.h"
#include "sysevent/balancer_msg.h"
#include "utility/hash.h"
#include <vector>
#include <iosfwd>

namespace Massiv {
namespace Core {


/** @addtogroup balance
 * @{
 */

/**
 * Balancer monitors data flow among migration groups and nodes and performs
 * object migrations in order to decrease network or CPU load, swaps out
 * not-recently-used objects and migrates object among nodes in order to
 * keep uniform node loads.
 *
 * There are three different optimization techniques:
 *
 *  1) local data flow optimization:
 *  The optimizer keeps track of data flow among migration groups and nodes.
 *  An example of data flow between object A and B is a RPC call on object B
 *  issued by object A or migration of object C, initiated by A, to object B.
 *  Simple data flow samples are described by ObjectToObjectSample, that
 *  are aggregated by @c aggregate_samples() to migration group to
 *  node samples (GroupSample).
 *  Migration groups that produce data flow consumed by a "single major"
 *  node will be migrated to that node. This reduces data flow among
 *  different nodes. 
 *  If we were running only this kind of optimization all objects would
 *  end up on a single node.
 *
 *  2) global resource optimization:
 *  The optimizer keeps track of resource loads on all nodes and migrates
 *  the objects among the nodes to ensure uniform node loads.
 *  Each node "ranks" its resource load which is in turn distributed to all
 *  other nodes. The most loaded node periodically chooses a portion of
 *  its objects (which should be sufficiently small in order to converge)
 *  and migrates them to the least loaded node.
 *  The objects are chosen by a random process which implies that the
 *  elected sets need not be "compact" (with respect to some data flow
 *  criterion). However (1) will tune the bad migrations ("companion" objects
 *  will be pushed to that node later).
 *
 *  3) local memory optimization:
 *  Migration groups that were not touched for some time are transparently
 *  swapped out and read back into memory when needed. Swap-out operation is
 *  technically equal to a "migration to swap" which means that swapped-out
 *  objects are not local. The swapping always operates on a per migration
 *  group basis. Objects are swapped-in when a scheduled event should be
 *  dispatched, or an object is delivered to the swapped-out object.
 *  This implies that RPC calls wake up the objects.
 */
class MASSIV_EXPORTED Balancer
    {
public:
        
    NO_COPY_CONSTRUCTOR( Balancer );
    NO_ASSIGNMENT_OPERATOR( Balancer );

    friend class ResourceLoadMessage;

public:

    /**
     * Balancer settings.
     */ 
    class BalancerSettings : public Settings
        {
    public:
            
        /** Constructor. */
        BalancerSettings();

    public:
        
        Boolean &   enabled;
            /**< Is Balancer enabled? */
        Boolean &   data_flow_optimization_enabled;
            /**< Is data flow optimization enabled?
                 If @c enabled is false then this has no effect. */
        Boolean &   resource_load_optimization_enabled;
            /**< Is resource load optimization enabled?
                 If @c enabled is false then this has no effect. */
        Boolean &   memory_load_optimization_enabled;
            /**< Is memory load optimization enabled?
                 If @c enabled is false then this has no effect. */
        Float &     min_data_run_period;
            /**< Minimum time difference between two data flow optimization
                 runs. */
        Float &     max_data_run_period;
            /**< Maximum time difference between two data flow optimization
                 runs. */
        Float &     resource_run_period;
            /**< Time difference between two resource optimization runs. */
        Float &     memory_run_period;
            /**< Time difference between two memory optimization runs. */
        Float &     samples_per_run_fraction;
            /**< Suggested number of object-to-object samples per run
                 relative to the total number of objects.
                 The actual value is obtained by multiplying the constant
                 by @c statistics.num_objects. */
        Float &     size_of_object_samples_fraction;
            /**< Suggested minimum total size of object-to-object samples
                 in a run, as a fraction relative to @c memory_load.
                 The actual value is obtained by multiplying by
                 @c statistics.memory_load. */
        Float &     data_flow_fraction;
            /**< Minimum data flow fraction that must be consumed by the
                 "major" consumer node in order to trigger group migration
                 to that node. A float value from [ 0, 1 ] range. */
        Float &     data_flow_aging_coef;
            /**< A coefficient from [ 0, 1 ] that determines how much old
                 samples are obsoleted. Their weight will be multiplied by
                 this coefficient. */
        Float &     object_timeout;
            /**< Swap-out migration group if it has not been touched
                 for @c object_timeout seconds. */
        Float &     cpu_load_weight;
            /**< CPU load versus object count fraction weight.
                 A positive float value <= 1.0f. This controls actually
                 what is balanced. 0.0f optimizes for uniform number of
                 objects on all nodes whereas 1.0f optimizes for uniform
                 CPU load. */
        Float &     rank_compare_factor;
            /**< Factor used when comparing resource load ranks of two
                 nodes to determine whether one of the nodes is
                 "substantially less loaded than the other node".
                 A float value from [ 0, 1 ] range. The lesser values imply
                 the bigger required differences in node ranks. */
        Float &     balance_object_count_fraction;
            /**< How many objects to migrate in a single resource optimization
                 run. Defined as a fraction relative to the total number of
                 objects registered to the simulation. Must be sufficiently
                 small or the resource optimization will not converge. */
        };

    /**
     * Balancer statistics, local information only.
     */
    class LocalStatistics : public Statistics
        {
    public:

        /** Constructor. */
        LocalStatistics();

    public:

        Float &     last_data_run;
            /**< Time of the last data flow optimization run. */
        Float &     last_memory_run;
            /**< Time of the last memory optimization run. */
        Float &     last_resource_run;
            /**< Time of the last resource load optimization run. */
        Float &     current_data_run_period;
            /**< Time difference between two data flow optimization runs. */
        Integer &   min_recent_data_size;
            /**< Minimum data size of a "recent" sample. Updated at the end
                 of the run. */
        Integer &   max_recent_data_size;
            /**< Maximum data size of a "recent" sample. Updated at the end
                 of the run. */
        Integer &   data_size_threshold;
            /**< Current data size threshold. Smaller samples are dropped. */
        Integer &   size_of_object_samples;
            /**< Total size of object samples. Sum of @c sizes of
                 @c object_samples. */
        Integer &   num_samples;
            /**< Size of @c object_samples. */
        Float &     consumed_time;
            /**< Total time consumed by the balancer. */
        };
    
    /**
     * Balancer statistics, global information.
     * Need not be always up to date.
     */
    class GlobalStatistics : public Statistics
        {
    public:

        /** Constructor. */
        GlobalStatistics();

    public:
        
        Integer &   num_total_objects;
            /**< Total number of objects on all nodes. Swapped-out objects
                 are not counted. */
        Integer &   total_memory_load;
            /**< Total memory load. */
        };

    /**
     * Balancer statistics, node-specific.
     * The information need not be always up to date.
     */
    class NodeStatistics : public Statistics
        {
    public:

        /** Constructor. */
        NodeStatistics
            (
            const NodeId & node
            );

    public:

        Integer &   memory_load;
            /**< Total memory occupied by all objects on the node. In bytes. */
        Integer &   num_objects;
            /**< Total number of objects on the node. */
        Float &     object_count_fraction;
            /**< @c num_objects relative to the total number of objects
                 in the simulation. A float value from [ 0, 1 ] range.
                 Always computed locally by each node from @c num_objects
                 and global statistics. */
        Float &     cpu_load;
            /**< Recent CPU load. A float value from [ 0, 1 ] range. */
        Float &     resource_load_rank;
            /**< Resource load rank used to compare resource loads of nodes.
                 Node A is said to be less loaded than node B if A.rank
                 is lesser than B.rank. Rank is a node-specific weighted sum
                 of the above attributes. A float value from [ 0, 1 ] range. */
        };
    
    typedef GarbageCollector::MigrationGroupId  MigrationGroupId;
    
    /**
     * Object to object data flow sample.
     */
    class ObjectToObjectSample
        {
    public:

        /** Constructor. */
        ObjectToObjectSample
            (
            const ObjectId &    producer,
            const ObjectId &    consumer,
            int                 size
            );

        /** Dumps the object. */
        friend std::ostream & operator<<
            (
            std::ostream &                  stream,
            const ObjectToObjectSample &    sample
            );

    public:
            
        ObjectId    producer;
            /**< Data producer. */
        ObjectId    consumer;
            /**< Data consumer. */
        int         size;
            /**< Data size. */
        };
    
    /**
     * Migration group data flow sample.
     */
    class GroupSample
        {
    public:
        
        /**
         * Migration group to object data flow sample.
         */
        struct ObjectSample
            {
        public:

            struct Less;

            /** Constructor. */
            ObjectSample
                (
                const ObjectId &    object_id,
                int                 size
                );
            
            /** Dumps the object. */
            friend std::ostream & operator<<
                (
                std::ostream &          stream,
                const ObjectSample &    sample
                );

        public:
                
            ObjectId    consumer;
                /**< Consumer object. */
            int         size;
                /**< Aggregated size. */
            };

        /**
         * Migration group to node data flow sample.
         */
        struct NodeSample
            {
        public:

            struct Less;

            /** Constructor. */
            NodeSample
                (
                const NodeId &  node,
                int             size
                );
            
            /** Dumps the object. */
            friend std::ostream & operator<<
                (
                std::ostream &      stream,
                const NodeSample &  sample
                );

        public:
                
            NodeId  node;
                /**< Consumer node. */
            int     size;
                /**< Aggregated size. */
            };

        typedef std::vector<NodeSample>     NodeSamples;
        typedef std::vector<ObjectSample>   ObjectSamples;

    public:
        
        /** Constructor. */
        GroupSample();

        /**
         * Accounts object-to-object sample to this group sample.
         * Assumes @c sample.producer belongs to the migration group
         * identified by @c group_id and sampled by @c this.
         */
        void account
            (
            const ObjectToObjectSample &    sample,
            MigrationGroupId                group_id
            );
        
        /**
         * Reduces data flow generated by this migration group
         * by migrating it to the most major consumer node.
         */
        void optimize_data_flow();

        /**
         * Applies aging to @c object_samples and @c node_samples.
         */
        void apply_aging();

        /** Dumps the object. */
        friend std::ostream & operator<<
            (
            std::ostream &      stream,
            const GroupSample & sample
            );

    public:

        ObjectId        group_representative;
            /**< Migration group representative. Any object that belongs
                 to the migration group. */
        ObjectSamples   object_samples;
            /**< Object samples. Sorted by @c ObjectSample::consumer. */
        NodeSamples     node_samples;
            /**< Node samples. Sorted by @c NodeSample::node. */
        int             total_external_size;
            /**< Total size of external data flow accounted to this migration group.
                 This covers data flow produced by this migration group and consumed
                 by objects in different migration groups. */
        int             total_internal_size;
            /**< Total size of internal data flow account to this migration group.
                 This covers data flow produced and also consumed by this migration
                 group. */
        };

    /**
     * Resource load information, specific to a particular node.
     * The information is stored into NodeStatistics.
     */
    class ResourceLoadInfo : public NodeStatistics
        {
    public:
            
        /** Constructor. */
        ResourceLoadInfo
            (
            const NodeId & node
            );

        /**
         * Fills/updates info according to local load.
         * Can be called only if @c node refers to the local node.
         */
        void evaluate_local_load();

        /**
         * Distributes load info to all server nodes.
         * Can be called only if @c node refers to the local node.
         */
        void distribute();

        /**
         * Tests if @c node is more loaded than another node described
         * by @c info.
         * Comparing resource ranks.
         */
        bool is_more_loaded_than
            (
            const ResourceLoadInfo & info
            ) const;

        /** Dumps the object. */
        friend std::ostream & operator<<
            (
            std::ostream &              stream,
            const ResourceLoadInfo &    info
            );
    
    public:

        NodeId  node;
            /**< Node. */
        };

    friend class GroupSample;
    friend class ResourceLoadInfo;

    typedef std::vector<ObjectToObjectSample>       ObjectToObjectSamples;
    typedef std::vector<GroupSample>                GroupSamples;
    typedef ext::hash_map
        <
        NodeId,
        ResourceLoadInfo *,
        NodeId::Hash,
        NodeId::EqualKey
        > ResourceLoadInfos;

public:
        
    /**
     * Constructor.
     * @c initialize() should be called then.
     */
    Balancer();

    /** Destructor. */
    ~Balancer();

    /** Initializes Balancer. */
    void initialize();

    /** Shutdowns Balancer. */
    void shutdown();
    
    /** Dumps the object. */
    friend std::ostream & operator<<
        (
        std::ostream &      stream,
        const Balancer &    balancer
        );

public:

    //
    // Balancer input.
    //

    /**
     * Accounts for a local call.
     * Caller object, the object that issues the call, is determined by
     * @c GarbageCollector::who(), @c callee refers to the object being
     * called.
     * @c packet_size is an approximation of potential packet that must have
     * been sent if @c callee was not local.
     * Called by RPC stubs when a local RPC call is processed.
     * Remote RPC calls are translated to object migrations.
     * @note Caller object can not be always identified reliably, see
     * @c GarbageCollector.
     */
    void account_for_local_call
        (
        ObjectProperty *    callee,
        int                 packet_size
        );

    /**
     * Accounts for an object migration.
     * Initiator object, the object responsible for initiating migration
     * of @c object to @c addressee, is determined by
     * @c GarbageCollector::who().
     * Called by Object when a migration is requested.
     * @note Initiator object can not be always identified reliably, see
     * @c GarbageCollector.
     */
    void account_for_object_migration
        (
        ObjectProperty *    object,
        const ObjectId &    addressee
        );
    
    /**
     * Accounts for object replication of @c object to @c node with
     * @c data_size size of replication update.
     */
    void account_for_object_replication
        (
        ObjectProperty *    object,
        const NodeId &      node,
        int                 data_size
        );

    /**
     * Adds generic data flow sample.
     */
    void add_sample
        (
        const ObjectId &    producer,
        const ObjectId &    consumer,
        int                 data_size
        );

private:

    //
    // Private.
    //

    /**
     * Aggregates object to object samples to migration group samples.
     */
    void aggregate_samples();
    
    /**
     * Returns resource load info for a specified node.
     */
    ResourceLoadInfo & get_resource_info
        (
        const NodeId & node
        );
    
    /**
     * Returns resource load info for local node.
     */
    ResourceLoadInfo & get_local_resource_info();
    
    /**
     * Handles resource load message.
     */
    void handle_resource_load_message
        (
        const ResourceLoadMessage & message
        );

private:

    //
    // Optimizers.
    //

    /**
     * Optimizes data flow.
     */
    void optimize_data_flow();

    /**
     * Optimizes resource load.
     */
    void optimize_resource_load();

    /**
     * Optimizes memory load.
     */
    void optimize_memory_load();

public:

    //
    // Control.
    //

    /**
     * Processes next tick.
     */
    void next_tick();
    
    /**
     * Enables load balancing.
     * Load balancing may still be globally disabled via settings in the
     * registry.
     */
    void enable_load_balancing();

    /**
     * Disables load balancing.
     */
    void disable_load_balancing();

    /**
     * Checks if load balancing is enabled.
     */
    bool is_load_balancing_enabled() const;

private:

    //
    // Main.
    //

    bool                    initialized;
        /**< Is Balancer initialized? */
    bool                    enabled;
        /**< Is load balancing locally enabled? If @c settings.enabled is
             false then @c enabled has no effect. */

    //
    // Settings and statistics accessible through registry.
    //

    BalancerSettings        settings;
        /**< Settings. */
    LocalStatistics         local_statistics;
        /**< Local statistics. */
    GlobalStatistics        global_statistics;
        /**< Global statistics. */
    ResourceLoadInfos       resource_infos;
        /**< Per-node resource load infos.
             Use @c get_resource_info() / @c get_local_resource_info() to
             access as entries are allocated on demand. */
    
    //
    // Data flow samples.
    //

    ObjectToObjectSamples   object_samples;
        /**< Object to object data flow samples. */
    GroupSamples            group_samples;
        /**< @c object_samples grouped by migration groups. Indexed
             by MigrationGroupId. Computed from @c object_samples at
             the end of the run. */

    //
    // Resource load helper statistics.
    //

    SystemTime              previous_system_time;
        /**< System time when CPU load was evaluated for the last time. */
    SystemTime              previous_total_sleep_time;
        /**< Total sleep time when CPU load was evaluated for the last time. */
    NodeId                  least_loaded_node;
        /**< Least loaded node. Updated by @c handle_resource_load_message(). */
    NodeId                  most_loaded_node;
        /**< Most loaded node. Updated by @c handle_resource_load_message(). */

    }; // Balancer

/** @} */ // addtogroup balance


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_BALANCER_H
