/** @file cmdline_parser.cpp
 * Simple command line parser class.
 * @see cmdline_parser.h
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "cmdline/cmdline_parser.h"
#include "status/exception.h"
#include <sstream>
#include <string>


namespace Massiv {

using namespace Massiv::Core;


CmdLineParser::CmdLineParser
    (
    int                     argc,
    const char * const *    argv,
    const std::string &     non_parameter_switches,
    const std::string &     parameter_switches
    ) : argc( argc - 1 ),   // don't process the first field - program name
        argv( argv ),
        non_parameter_switches( non_parameter_switches ),
        parameter_switches( parameter_switches ),
        expect_switches( true ),
        processing_mode( PROCESS_STANDARD )
    { 
    if( this->argc > 0 )
        {
        actual_arg = 1;
        }
    else
        {
        actual_arg = 0;
        }
    }


CmdLineParser::~CmdLineParser()
    {
    }


char CmdLineParser::get_next
    (
    std::string &       parameter
    )
    {
    parameter.erase();

    if( actual_arg > argc )
        {
        return 0;
        }

    if( processing_mode == PROCESS_STANDARD )
        {
        return get_next_in_standard_mode( parameter );
        }
    else
    if( processing_mode == PROCESS_SWITCHES_GROUP )
        {
        return get_next_in_switches_group_mode( parameter );
        }

    return 0;
    } // CmdLineParser::get_next()


int CmdLineParser::get_argc()
    {
    return argc;
    } // CmdLineParser::get_argc()


char CmdLineParser::get_next_in_standard_mode
    (
    std::string &       parameter
    )
    {
    char                result = 0;
    hope( actual_arg <= argc );

    do
        {
        if( argc == 0 )
            {
            break;
            }

        result = '-';

        std::string     aux_par( argv[ actual_arg ] );
        
        hope( aux_par.size() );

        if( expect_switches && ( aux_par[ 0 ] == '-' ) )
            {
            if( aux_par.size() < 2 )
                {
                actual_arg++;

                throw ExMissingSwitch
                    ( 
                    "Missing switch letter in the command line." 
                    );
                }
            else
            if( ( aux_par.size() == 2 ) && ( aux_par[ 1 ] == '-' ) )
                {
                expect_switches = false;
                    // End the switches section
                                    
                actual_arg++;
                parameter = get_parameter();
                }
            else
                {
                processing_mode = PROCESS_SWITCHES_GROUP;
                actual_switch_in_group = 1;

                result = get_next_in_switches_group_mode( parameter );
                }
            }
        else
            {
            expect_switches = false;
                // End the switches section

            parameter = get_parameter();
                // actual_arg is incremented in this method!
            }

        break;
        } while( true );

    return result;
    } // CmdLineParser::get_next_in_standard_mode()


char CmdLineParser::get_next_in_switches_group_mode
    (
    std::string &       parameter
    )
    {
    std::string         aux_par( argv[ actual_arg ] );
    char                result = 0;

    hope( actual_arg <= argc );
    hope( actual_switch_in_group < aux_par.size() );

    result = aux_par[ actual_switch_in_group ];

    if( non_parameter_switches.find( result ) != std::string::npos )
        {
        // The switch found among 'non-parameter' switches
        
        if( ++actual_switch_in_group == aux_par.size() )
            {
            processing_mode = PROCESS_STANDARD;
            actual_arg++;
            }
        }
    else
    if( parameter_switches.find( result ) != std::string::npos )
        {
        // The switch found among 'parameter' switches

        if( ++actual_switch_in_group != aux_par.size() )
            {
            parameter = aux_par.substr( actual_switch_in_group );                        
            actual_arg++;
            }
        else
            {
            actual_arg++;
            parameter = get_parameter();
            }

        processing_mode = PROCESS_STANDARD;
        }
    else
        {
        if( ++actual_switch_in_group == aux_par.size() )
            {
            processing_mode = PROCESS_STANDARD;
            actual_arg++;
            }

        std::ostringstream  ss;
        ss << "Invalid commandline switch ('" << result << "')";
        throw ExUnexpectedCommandLineSwitch( ss.str() );
        }       

    return result;
    } // CmdLineParser::get_next_in_switches_group_mode()


std::string CmdLineParser::get_parameter()
    {
    if( actual_arg > argc )
        {
        throw ExMissingCommandLineArgument( "Missing command line argument." );
        }

    hope( argv[ actual_arg ][ 0 ] );

    return argv[ actual_arg++ ];
    } // CmdLineParser::get_parameter()

 
} // namespace Massiv
