/** @file cmdline_parser.h
 * Simple command line parser class.
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CMDLINE_PARSER_H
#define MASSIV_CMDLINE_PARSER_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <string>


namespace Massiv {


/** @addtogroup utility
 * @{
 */

/**
 * Interface for uniform simple command line parsing. Command line can consist
 * of one-letter switches without parameters (i.e. <tt>-e</tt>) or with (i.e.
 * <tt>-f name</tt>). Switches may be grouped (<tt>-ef name</tt>). 
 * <tt>--</tt> sequence ends switches section - all subsequent parameters 
 * are not considered as switches, even though they start with <tt>-</tt>. 
 * Arbitrary word with initial letter other than <tt>-</tt> which is not 
 * predecesed with some 'with-parameter-switch' also ends up switches.
 * @note <tt>-ef name</tt> has the same meaning as <tt>-efname</tt>.
 */
class MASSIV_EXPORTED CmdLineParser
    {

public:

    /**
     * Constructor.
     * @param argc argc parameter passed to main().
     * @param argv argv parameter passed to main().
     * @param non_parameter_switches Each character in this string declares
     *      one switch expecting no parameter.
     * @param parameter_switches Each character in this string declares
     *      one switch expecting parameter.
     * @note There should not be duplicate characters in 
     *      @c non_parameter_switches or @c parameter_switches. There should
     *      not also be a character contained in both of the strings. However,
     *      No runtime checks are done.
     */
    CmdLineParser
        (
        int                     argc,
        const char * const *    argv,
        const std::string &     non_parameter_switches = std::string(),
        const std::string &     parameter_switches = std::string()
        );

    /**
     * Destructor.
     */
    ~CmdLineParser();

private:

    NO_DEFAULT_CONSTRUCTOR( CmdLineParser );
    NO_COPY_CONSTRUCTOR( CmdLineParser );
    NO_ASSIGNMENT_OPERATOR( CmdLineParser );

public:

    /**
     * Call this method to get the next switch/parameter.
     * @param parameter Receives a parameter that was contained on the command
     *      line. If there is a parameter enclosed in quotes (""), the quotes
     *      are removed. @c parameter may receive an empty string if there
     *      is a switch expecting no parameter.
     * @return Character of the switch or <tt>-</tt> if switches section has 
     *      already ended. Returns 0 the parsing is finished.
     *
     * @exception ExMissingSwitch There was only <tt>-</tt> on the command line,
     *      but a switch characted didn't follow.
     * @exception ExUnexpectedCommandLineSwitch There was a switch on the
     *      command line that was defined neither in @c non_parameter_switches
     *      nor in @c parameter_switches.
     * @exception ExMissingCommandLineArgument There was a parameter switch
     *      on the command line, but it was not followed by an argument.
     */
    char get_next
        (
        std::string &       parameter
        );

    /**
     * Number of parameters.
     * @return Number of text 'fields' that were passed on the commandline.
     */
    int get_argc();

protected:

    // Helper implementation routines

    /**
     * get_next() implementation in the @c PROCESS_STANDARD mode.
     * @see get_next()
     *
     * @exception ExMissingSwitch There was only '-' on the command line,
     *      but a switch characted didn't follow.
     * @exception ExMissingCommandLineArgument There was a parameter switch
     *      on the command line, but it was not followed by an argument.
     */
    char get_next_in_standard_mode
        (
        std::string &       parameter
        );

    /**
     * get_next() implementation in the @c PROCESS_SWITCHES_GROUP mode.
     * @see get_next().
     *
     * @exception ExUnexpectedCommandLineSwitch There was a switch on the
     *      command line that was defined neither in @c non_parameter_switches
     *      nor in @c parameter_switches.
     * @exception ExMissingCommandLineArgument There was a parameter switch
     *      on the command line, but it was not followed by an argument.
     */
    char get_next_in_switches_group_mode
        (
        std::string &       parameter
        );

    /**
     * Finds and returns parameter. This is being called either after closing
     * the switches section or after processing a switch expecting a parameter.
     *
     * @exception ExMissingCommandLineArgument There was a parameter switch
     *      on the command line, but it was not followed by an argument.
     */
    std::string get_parameter();

protected:

    int                     argc;
        /**< argc parameter passed to main(). */

    const char * const *    argv;
        /**< argv parameter passed to main(). */

    std::string             non_parameter_switches;
        /**< Each character in this string declares one switch expecting no
             parameter. */

    std::string             parameter_switches;
        /**< Each character in this string declares one switch expecting 
             parameter. */

protected:

    /**
     * Mode of processing - determines what will be done after @c get_next()
     * is called next time.
     */
    enum EProcessingMode
        {
        PROCESS_STANDARD = 0,
            /**< Next command line arguments will be processed. */
        PROCESS_SWITCHES_GROUP = 1,
            /**< The same argument will be processed as during the last call
                 of get_next() - this happens when the argument is a group
                 of switches ('-vf', for instance). */
        PROCESS_COUNT
        }; // enum EProcessingMode

protected:

    // Implementation stuff

    int                     actual_arg;
        /**< Determines actually parsed field in @c argv. */

    bool                    expect_switches;
        /**< true if the switches section hasn't been ended yet. */

    EProcessingMode         processing_mode;
        /**< @see EProcessingMode. */

    unsigned int            actual_switch_in_group;
        /**< This parameter makes sense only during processing in the 
             PROCESS_SWITCHES_GROUP mode. It determines which switch
             of the group is to be processed next.
             @note @c actual_switch_in_group is one-based. */
        
    }; // class CmdLineParser

/** @} */ // addtogroup utility


} // namespace Massiv


#endif // MASSIV_CMDLINE_PARSER_H
