/** @file account_manager.cpp
 * AccountManager declaration.
 */

/* 
 * begin        : Sat Sep 6 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "data/account_manager.h"
#include "data/path_manager.h"
#include "sysevent/account_msg.h"
#include "utility/debug.h"
#include "node/nodeid.h"
#include "global/global.h"
#include "utility/quote_string.h"
#include "status/exception.h"
#include "database/node_database.h"
#include "net/massiv_rsa.h"
#include "data/data_manager.h"
#include "filesys/volume_manager.h"
#include "filesys/file_system.h"
#include "object/lib/account_object_interface.h"
#include "object/lib/account_object_interface_rpc.h"
#include "pointer/object_pointer.h"
#include "node/node_manager.h"
#include "net/net.h"
#include "system/system.h"
#include <sstream>
#include <ctime>
#include <fstream>

namespace Massiv {
namespace Core {


#define AM_PENDING_ACCOUNTS_DIRECTORY       "pending_accounts"
#define AM_CONFIG_NODES_PRIVATE             "config_nodes_private"

namespace {

void log_message
    (
    const std::string &     message
    )
    {
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
        std::string( "AccountManager: " ) + message
        );
    }

}


AccountManager::AccountManager()
    {
    // Test if config_nodes_private is writeble.
    #if 0
        {
        std::string name = Global::path_manager().get_source_data_directory() +
            "/" + AM_CONFIG_NODES_PRIVATE; // FIXME: dat to do #define
        VolumeManager::IFStream     ifs( name, std::ios_base::in );
        if( ifs.is_open() )
            {
            ifs.close();
            // File is there. Check read-only flag.
            try
                {
                FileSystem & fs = VolumeManager::file_system();
                fs.rename_file( name, name + ".read_only_test" );
                fs.rename_file( name + ".read_only_test", name );

                VolumeManager::OFStream ofs( name.c_str(), std::ios_base::out | std::ios_base::ate );
                ofs << "x";
                if
                ofs.close();
                }
            catch( Exception )
                {
                // Cannot open.
                std::ostringstream  ss;
                ss << "File \"" << name << "\" is read-only. Remove read-only flag "
                    << "and run service again.";
                throw ExCommonError( ss.str() ); // FIXME: account exception
                }
            }        
        }
    #endif
    
        {
        std::string name = Global::path_manager().get_source_data_directory() + "/" + AM_CONFIG_NODES_PRIVATE + ".manual";
        VolumeManager::IFStream     ifs( name, std::ios_base::in );
        if( ifs.is_open() )
            {
            ifs.close();
            std::ostringstream  ss;
            ss << "File \"" << name << "\" with new accounts still exist. Do manual account update "
            << "and run service again.";
            throw ExCommonError( ss.str() ); // FIXME: account exception
            }
        }
    }


AccountManager::~AccountManager()
    {
    // TODO: cleanup?
    }


bool AccountManager::create_account()
    {
    hope( is_anonymous() );
    
    if( account_request.account_request_status != ACCOUNT_REQUEST_STATUS_IDLE )
        {
        return false;
        }
        
    generate_public_private_key( public_key, private_key );

    std::ostringstream ss1;
    ss1 << quote_string( public_key, true );
    public_key = ss1.str();

    std::ostringstream ss2;
    ss2 << quote_string( private_key, true );
    private_key = ss2.str();

    account_request.anonymous_node_id = Global::get_local_node_id();
    account_request.account_request_status = ACCOUNT_REQUEST_STATUS_PENDING;
    account_request.public_key = public_key;
    //account_request.magic_number = Global::system_time().get_value();
    
    AccountRegisterRequestMessage     request_to_service( account_request ); 
    request_to_service.send( get_service(), NetworkCommon::STREAM_RELIABLE );    
    return true;
    }
    
    
bool AccountManager::save_account
    (
    const std::string &         file_name,
    const std::string &         pass
    )
    {    
    if( account_request.account_request_status != ACCOUNT_REQUEST_STATUS_SUCCESS )
        {
        return false;
        }

    std::ostringstream  oss;
    
    oss << "\n[ ]\n";
    oss << "node_id : string = \"" << account_request.generated_node_id << "\"\n";   
    oss << "rsa_public_key : string = " << public_key << "\n";
    oss << "rsa_private_key : string = " << private_key << "\n";

    encrypt_file( file_name, pass, oss.str() );
    return true;
    }
        

void AccountManager::encrypt_file
    (
    const std::string &         file_name,
    const std::string &         pass,
    const std::string &         content
    )
    {
    std::ostringstream  ss;
    ss << "AccountManager: writing login file to " << file_name;
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );
        
    VolumeManager::OFStream     ofs; //( file_name.c_str(), std::ios_base::binary );
    ofs.open( file_name, std::ios_base::out | std::ios_base::binary | std::ios_base::trunc );
    if( !ofs.is_open() )
        {
        throw ExCommonError( std::string( "Cannot create file " ) + file_name );
        }
    
    std::string crypted;        
    encrypt_string( content, pass, crypted );
    ofs << quote_const_string( crypted );
    
    ofs.close();
    }

 
 void AccountManager::encrypt_string
    (
    const std::string &         input,
    const std::string &         password,
    std::string &               crypted
    )
    {
    unsigned int padding = 0;
    std::string  crypted_string;
    std::string  pom = input;
    std::string  pass = "Mas" + password + "siv";
    
    BlowfishEncryptor benc( pass );
    
    // How long padding do I need?
    padding = ( 8 - pom.size() % 8 ) % 8;
    
    // Add the padding.
    if( padding != 0 )
        {
        for( unsigned int i = 0; i < padding ; i++ )
            {
            pom += " ";
            }
        }
    
    // Compute checksum.
    Massiv::Core::BitIOStreamBuffer buffer;
    for( unsigned int i = 0; i < pom.size(); i++ )
        {
        buffer.push_back( pom.at( i ));
        }
        
    Massiv::Core::CRC32< Massiv::Core::BitIOStreamBuffer > crc32( &buffer );
    
    UInt32 crc = crc32.compute_checksum();
    
    benc.process( pom, crypted_string );
    
    std::ostringstream oss2;
    
    // Write info about crc.
    oss2 << ( char ) ( crc >> 24 );
    oss2 << ( char ) ( crc >> 16 );
    oss2 << ( char ) ( crc >> 8 );
    oss2 << ( char ) crc;
    
    // Write info about padding.
    oss2 << ( char ) ( padding ^ pass[ 0 ] );
    oss2 << crypted_string;
    
    crypted = oss2.str();
    }


void AccountManager::decrypt_string
    (
    const std::string &         input,
    const std::string &         password,
    std::string       &         decrypted
    )
    {    
    std::string pass = "Mas" + password + "siv";   
        
    BlowfishDecryptor bdec( pass );
   
    // Read CRC32.
    UInt32 crc3 = ( UInt32 ) ( ( UInt8 ) input[ 3 ] );
    UInt32 crc2 = ( UInt32 ) ( ( UInt8 ) input[ 2 ] << 8 );
    UInt32 crc1 = ( UInt32 ) ( ( UInt8 ) input[ 1 ] << 16 );
    UInt32 crc0 = ( UInt32 ) ( ( UInt8 ) input[ 0 ] << 24 );
    UInt32 crc = crc3 + crc2 + crc1 + crc0;

    // How long is the padding?
    unsigned int padding = ( unsigned int ) input[ 4 ] ^ pass[ 0 ];
    
    // Remove info about padding.
    std::string pom = input.substr( 5, input.size() - 5 );
    
    if( pom.size() % 8 != 0 )
        {
        std::ostringstream ss;
        ss  << "Error decrypting file.";
        throw ExCommonError( ss.str() );
        }
    bdec.process( pom, decrypted );
    
    // Verify checksum.
    Massiv::Core::BitIOStreamBuffer buffer;
    for( unsigned int i = 0; i < decrypted.size(); i++ )
        {
        buffer.push_back( decrypted.at( i ));
        }
        
    Massiv::Core::CRC32< Massiv::Core::BitIOStreamBuffer > crc32( &buffer );
    
    UInt32 current_crc = crc32.compute_checksum();
    
    if( current_crc != crc )
        {
        std::ostringstream ss;
        ss  << "Bad password entered - CRC Checksum failed!";
        throw ExCommonError( ss.str() );
        }    

    // Remove padding from the decrypted string.
    decrypted = decrypted.substr( 0, decrypted.size() - padding );
    }    
    

AccountManager::AccountRequestStatus AccountManager::get_account_request_status() const
    {
    hope( is_anonymous() );
    return account_request.account_request_status;
    }


void AccountManager::set_account_request_idle_status()
    {
    hope( is_anonymous() );
    account_request.account_request_status = ACCOUNT_REQUEST_STATUS_IDLE;
    }


const AccountManager::AccountRequest & AccountManager::get_account_request() const
    {
    hope( is_anonymous() );
    return account_request;
    }
    

void AccountManager::generate_public_private_key
    (
    std::string &         public_key,
    std::string &         private_key    
    ) const
    {
    // FIXME: prevzato z Octova zdrojaku pro generovani klicu.
    
    unsigned int bitlen = RSACipher::DEFAULT_PRIVATE_KEY_LENGTH;
    unsigned int seedlen = RSACipher::DEFAULT_PRIVATE_KEY_SEED_LENGTH;

    RSACipher rsa( bitlen, RSACipher::generate_seed( seedlen ) );
    private_key = rsa.get_private_key();
    public_key = rsa.get_public_key();
    }


// Messages handling.

void AccountManager::handle_account_register_request_message
    (
    const AccountRegisterRequestMessage &       msg
    )
    {
    hope( is_service() );
    
    const NodeId &      initiator = msg.get_initiator();
    hope( initiator.is_type( NodeId::TYPE_ANONYMOUS ) );
    
    std::ostringstream ss;
    ss << "received handle_account_register_request_message from node " << msg.get_initiator();
    log_message( ss.str() );

    if( find_account_request2( initiator ) )
        {
        // Client cannot send simultaneous create account requests.
        // Discard request.
        AccountRequest  ar = msg.account_request;
        ar.account_request_status = ACCOUNT_REQUEST_STATUS_ERROR;
        AccountRegisterRespondMessage     respond_to_client( ar );
        respond_to_client.send( initiator, NetworkCommon::STREAM_RELIABLE );

        std::ostringstream ss;
        ss << "request from " << msg.get_initiator() << " was discarted." ;
        log_message( ss.str() );

        return;
        }
        
    AccountRequest ar_temp = msg.account_request;
    ar_temp.anonymous_node_id = initiator; // FIXME - doc proc je to tady
    AccountRequest & ar = add_account_request( ar_temp );
    
    // Generate unique node id.
    ar.generated_node_id = generate_unique_node_id();
    
    // Save account do disk to be able to create account even 
    // if service crashes for some reason.
    ar.account_request_status = ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER;
    save_pending_account( ar );

    // Send request to server.
    //AccountCreateRequestMessage     request_to_server( ar );
    //request_to_server.send( get_server(), NetworkCommon::STREAM_RELIABLE );    

    // Send response to client.
    ar.account_request_status = ACCOUNT_REQUEST_STATUS_SUCCESS;
    AccountRegisterRespondMessage     respond_to_client( ar );
    respond_to_client.send( ar.anonymous_node_id, NetworkCommon::STREAM_RELIABLE );
    
    ar.account_request_status = ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER;

        {    
        std::ostringstream ss;
        ss << "request from " << msg.get_initiator() << " was succesfully added to list."
        << "Generated node id is " << ar.generated_node_id;
        log_message( ss.str() );
        }
    }


void AccountManager::handle_account_register_respond_message
    (
    const AccountRegisterRespondMessage &       msg
    )
    {
    hope( is_anonymous() );    
    hope( account_request.public_key == msg.account_request.public_key );    

    std::ostringstream ss;
    ss << "handle_account_register_respond_message received from node " << msg.get_initiator();
    log_message( ss.str() );
    
    account_request = msg.account_request;
    }


void AccountManager::handle_account_create_request_message
    (
    const AccountCreateRequestMessage &       msg
    )
    {
    hope( is_server() );
    
    std::ostringstream ss;
    ss << "handle_account_create_request_message received from node " << msg.get_initiator()
       << "Generated node id is " << msg.account_request.generated_node_id;
    log_message( ss.str() );

    AccountRequest  ar( msg.account_request );
    if( !create_account_object( ar.account_id ) )
        {
        ar.account_request_status = ACCOUNT_REQUEST_STATUS_ERROR;

        std::ostringstream ss;
        ss << "error when creating account for client."
           << "Generated node id is " << msg.account_request.generated_node_id;     
        log_message( ss.str() );
        }
    else
        {
        ar.account_request_status = ACCOUNT_REQUEST_STATUS_SUCCESS;

        std::ostringstream ss;
        ss << "account created for node id " << msg.account_request.generated_node_id
           << "Account id is " << ar.account_id;
        log_message( ss.str() );
        }

    AccountCreateRespondMessage     respond_to_service( ar );
    respond_to_service.send( get_service(), NetworkCommon::STREAM_RELIABLE );
    }


void AccountManager::handle_account_create_respond_message
    (
    const AccountCreateRespondMessage &       msg
    )
    {
    hope( is_service() );
    hope( msg.account_request.anonymous_node_id.is_type( NodeId::TYPE_ANONYMOUS ) );

    AccountRequest & ar = find_account_request( msg.account_request.anonymous_node_id );
    hope( ar.account_request_status == ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER );
    ar = msg.account_request;
    
    std::ostringstream ss;
    ss << "handle_account_create_respond_message: received response from node " << msg.get_initiator()
       << "Generated node id is " << msg.account_request.generated_node_id;     
    log_message( ss.str() );

    if( ar.account_request_status == ACCOUNT_REQUEST_STATUS_SUCCESS )
        {
        // Automatic management.
        std::string     name = Global::path_manager().get_source_data_directory() + "/" + AM_CONFIG_NODES_PRIVATE;
        if( account_manager_settings.manual_management )
            {
            // Manual management.
            name = Global::path_manager().get_source_data_directory() + "/" + AM_CONFIG_NODES_PRIVATE + ".manual";
            VolumeManager::IFStream     ifs( name, std::ios_base::in );
            if( !ifs.is_open() )
                {
                // File doesn't exist. Create empty file.
                VolumeManager::OFStream     ofs( name, std::ios_base::out );
                if( !ofs.is_open() )
                    {
                    std::string     temp = "Cannot create file \"" + name + "\"";
                    panic( temp.c_str() );
                    }
                // File created.
                ofs.close();
                }
            else
                {
                // File exist.
                ifs.close();
                }
            }
        FileSystem & fs = VolumeManager::file_system();
        std::string     work_name = name + ".work";
        try
            {
            fs.delete_file( work_name );
            }
        catch( ... )
            {
            // do nothing.
            }
        if( !cp.copy_file( work_name, name ) )
            {
            std::ostringstream  ss;
            ss << "Cannot copy file \"" << name << "\" to \"" << work_name << "\"";
            panic( ss.str().c_str() );
            }
        while( !cp.is_copied() )
            {
            cp.next_tick();
            }
        // File is duplicated. Lest's update working copy.
        VolumeManager::OFStream   ofs( work_name, std::ios_base::app ); // std::ios_base::out | 
        if( !ofs.is_open() )
            {
            std::string temp = "Cannot update file \"" + work_name + "\". Check read-only flag.";
            panic( temp.c_str() );
            }

        ofs << "\n[ client_" << ar.generated_node_id.get_relative_id() << " ]\n";
        ofs << "node_id : string = \"" << ar.generated_node_id << "\"\n";
        ofs << "rsa_public_key : string = " << ar.public_key << "\n";
        ofs << "address : string = \"localhost\"\n";
        ofs << "port : integer = 0\n";
        ofs << "account_object : string = \"" << ar.account_id << "\"\n";
        ofs << "name : string = \"Client" << ar.generated_node_id.get_relative_id() << "\"\n";
        ofs.close();
        
        // Let's rename working copy to original name.
        try
            {
            fs.delete_file( name );
            }
        catch( ... )
            {
            std::string temp = "Cannot update file \"" + name + "\". Check read-only flag.";
            panic( temp.c_str() );
            }   
         
        fs.rename_file( work_name, name );
        
        // Delete saved pending account.
        delete_pending_account( ar );
        
        Global::data_manager().switch_to_update_phase( DataManager::AUTOMATIC );
                
        remove_account_request( ar.anonymous_node_id );

        std::ostringstream ss;
        ss << "account saved to " AM_CONFIG_NODES_PRIVATE "."
           << "Generated node id is " << msg.account_request.generated_node_id;
        log_message( ss.str() );
        } // success
    else
        {
        // Request was not processed correctly.
        hope( ar.account_request_status == ACCOUNT_REQUEST_STATUS_ERROR );
        
        ar.try_time = SystemTime::current_time();
        std::ostringstream ss;
        ss << "handle_account_create_respond_message(): error when creating account for client."
           << "Generated node id is " << msg.account_request.generated_node_id;     
        log_message( ss.str() );

        // FIXME - ted to request odstrani, ale melo by to asi nechat spadnout server nebo service
        //delete_pending_account( ar );
        //remove_account_request( ar.anonymous_node_id );
        }
    }


AccountManager::AccountRequest & AccountManager::add_account_request
    (
    const AccountRequest &      account_request
    )
    {
    hope( is_service() );
    hope( account_request.anonymous_node_id.is_type( NodeId::TYPE_ANONYMOUS ) );
    
    std::pair< AccountRequestHash::iterator, bool> result = account_request_hash.insert
        (
        AccountRequestHash::value_type( account_request.anonymous_node_id, account_request )
        );
    if( !result.second )
        {
        std::ostringstream  ss;
        ss << "Cannot add account request for NodeId = " << account_request.anonymous_node_id 
           << " to account request hash table.";
        throw ExCommonError( ss.str() );
        }    
    
    return result.first->second;
    }
    

AccountManager::AccountRequest & AccountManager::find_account_request
    (
    const NodeId &              node_id
    )
    {
    hope( is_service() );
    hope( node_id.is_type( NodeId::TYPE_ANONYMOUS ) );
    
    AccountRequestHash::iterator it = account_request_hash.find( node_id );
    if( it == account_request_hash.end() )
        {
        std::ostringstream  ss;
        ss << "Cannot find account request for NodeId = " << node_id
           << " in account request hash table.";
        throw ExCommonError( ss.str() );
        }    
    return it->second;
    }


bool AccountManager::find_account_request2
    (
    const NodeId &              node_id
    ) const
    {
    hope( is_service() );
    hope( node_id.is_type( NodeId::TYPE_ANONYMOUS ) );
    
    AccountRequestHash::const_iterator it = account_request_hash.find( node_id );
    if( it == account_request_hash.end() )
        {
        return false;
        }    
    return true;
    }


void AccountManager::remove_account_request
    (
    const NodeId &      node_id
    )
    {
    hope( is_service() );
    
    AccountRequestHash::iterator it = account_request_hash.find( node_id );
    if( it == account_request_hash.end() )
        {
        hope( 0 );
        }
    account_request_hash.erase( it );
    }
    

NodeId AccountManager::generate_unique_node_id()
    {
    static bool once = true;
    static UInt32 last_relative_id = 0;
    if( once )
        {
        once = false;
        // Find the greatest node id in the first call of this method.
        EnumeratorGenerateNodeId    en;
        Global::node_database().enumerate_nodes( NodeId::TYPE_CLIENT, en );
        hope( last_relative_id < en.max_relative_id || last_relative_id == 0 );
        last_relative_id = en.max_relative_id;
        }
    last_relative_id++; 
    return NodeId::create_from_type_and_relative_id( NodeId::TYPE_CLIENT, last_relative_id );
    }


bool AccountManager::is_service() const
    {
    return Global::get_local_node_id().is_type( NodeId::TYPE_SERVICE );
    }


bool AccountManager::is_client() const
    {
    return Global::get_local_node_id().is_type( NodeId::TYPE_CLIENT );
    }


bool AccountManager::is_server() const
    {
    return Global::get_local_node_id().is_type( NodeId::TYPE_SERVER );
    }


bool AccountManager::is_anonymous() const
    {
    return Global::get_local_node_id().is_type( NodeId::TYPE_ANONYMOUS );
    }


bool AccountManager::create_account_object
    (
    ObjectId &      account_id
    )
    {
    hope( is_server() );
    
    if( System::get_phase() != System::PHASE_RUNNING )
        {
        // Simulation is not ready to create new objects.
        return false;
        }
    
    NodeDatabase & node_database = Global::node_database();
    NodeId local_node_id = Global::get_local_node_id();
    const NodeDatabase::NodeData & node_data = node_database.get_node_data( local_node_id );
    ObjectId local_node_object_id = node_data.get_proxy_node_object_id();
    
    if( local_node_object_id.is_null() )
        {
        // Should never happen.
        return false;
        }
    
    // Access local node object.
    WeakPointer< Lib::NodeObjectInterface > local_node_object = local_node_object_id;
    
    // Check if local node object was created.
    if( !local_node_object.is_local() )
        {
        // Node object was not yet created.
        return false;
        }
    
    // Create new account object.
    Pointer< Lib::AccountObjectInterface > account = local_node_object->create_account_object();
    if( account == null )
        {
        // Should never happen.
        return false;
        }
    
    account_id = account.get_object_id();
    return true;
    }


NodeId AccountManager::get_service()
    {
    EnumeratorNodeId    en;
    Global::node_database().enumerate_nodes( NodeId::TYPE_SERVICE, en );
    return en.node_id;
    }
    

NodeId AccountManager::get_server()
    {
    EnumeratorNodeId    en;
    Global::node_database().enumerate_nodes( NodeId::TYPE_SERVER, en );
    return en.node_id;
    }


AccountManager::AccountRequest::AccountRequest() :
    account_request_status( ACCOUNT_REQUEST_STATUS_IDLE )
    {
    }
    

AccountManager::AccountRequest::AccountRequest
    (
    const AccountRequest &      ar
    ) :
    anonymous_node_id( ar.anonymous_node_id ),
    generated_node_id( ar.generated_node_id ),
    account_request_status( ar.account_request_status ),
    public_key( ar.public_key ),
    account_id( ar.account_id )
    {
    }
    

AccountManager::AccountRequest & AccountManager::AccountRequest::operator=
    (
    const AccountRequest &      ar
    )
    {
    anonymous_node_id = ar.anonymous_node_id;
    generated_node_id = ar.generated_node_id;
    account_request_status = ar.account_request_status;
    public_key = ar.public_key;
    account_id = ar.account_id;
    return *this;
    }


void AccountManager::AccountRequest::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    anonymous_node_id.bitstr_write( bos, desc );
    generated_node_id.bitstr_write( bos, desc );
        
    UInt32  temp = static_cast< UInt32 >( account_request_status );
    Serializer::BitStrInt::write( bos, temp, ACCOUNT_REQUEST_STATUS_BITS );
        
    public_key.bitstr_write( bos, desc );
    account_id.bitstr_write( bos, desc );
    }


void AccountManager::AccountRequest::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    anonymous_node_id.bitstr_read( bis, desc );
    generated_node_id.bitstr_read( bis, desc );
        
    UInt32  temp;
    Serializer::BitStrInt::read( bis, temp, ACCOUNT_REQUEST_STATUS_BITS );
    account_request_status = static_cast< AccountRequestStatus >( temp );
        
    public_key.bitstr_read( bis, desc );
    account_id.bitstr_read( bis, desc );
    }


void AccountManager::AccountRequest::text_read
    (
    TextReader &                        tr,
    const Serializer::Description &     desc
    )
    {
    tr.next_line();
    anonymous_node_id.text_read( tr, desc );
    tr.next_line();
    generated_node_id.text_read( tr, desc );
    tr.next_line();
        
    UInt32  temp;
    Serializer::TextInt::read( tr, temp, ACCOUNT_REQUEST_STATUS_BITS );
    account_request_status = static_cast< AccountRequestStatus >( temp );
        
    tr.next_line();
    public_key.text_read( tr, desc );
    tr.next_line();
    account_id.text_read( tr, desc );
    }


void AccountManager::AccountRequest::text_write
    (
    TextWriter &                        tw,
    const Serializer::Description &     desc
    ) const
    {
    anonymous_node_id.text_write( tw, desc );
    tw.end_line();
    generated_node_id.text_write( tw, desc );
    tw.end_line();
        
    UInt32  temp = static_cast< UInt32 >( account_request_status );
    Serializer::TextInt::write( tw, temp, ACCOUNT_REQUEST_STATUS_BITS );
    tw.end_line();
        
    public_key.text_write( tw, desc );
    tw.end_line();
    account_id.text_write( tw, desc );
    tw.end_line();
    }


// Enumerator.

void AccountManager::EnumeratorNodeId::operator()
    (
    const NodeDatabase::NodeData & data
    )
    {
    node_id = data.get_node_id();
    }


AccountManager::EnumeratorGenerateNodeId::EnumeratorGenerateNodeId() :
    max_relative_id( 0 )
    {
    }
    
    
void AccountManager::EnumeratorGenerateNodeId::operator()
    (
    const NodeDatabase::NodeData & data
    )
    {    
    UInt32 temp = data.get_node_id().get_relative_id();
    if( max_relative_id < temp )
        {
        max_relative_id = temp;
        }
    }


AccountManager::EnumeratorFindNodeId::EnumeratorFindNodeId
    (
    const NodeId &      _node_id
    ) :
    node_id( _node_id ),
    result( false )
    {
    }


void AccountManager::EnumeratorFindNodeId::operator()
    (
    const NodeDatabase::NodeData & data
    )
    {
    if( node_id == data.get_node_id() )
        {
        result = true;
        }
    }


AccountManager::EnumeratorFindRunningNode::EnumeratorFindRunningNode() :
    found( false )
    {
    }


void AccountManager::EnumeratorFindRunningNode::operator()
    (
    const NodeDatabase::NodeData & data
    )
    {
    //if( Global::network().is_node_connected( data.get_node_id() ) )
    if( !found )
        {
        node_id = data.get_node_id();
        found = true;
        }
    }


AccountManager::AccountManagerSettings::AccountManagerSettings() :
    Settings( "AccountManager" ),
    manual_management( access_boolean( "manual_management", false ) )
    {
    }


/*void AccountManager::fix_local_node_config_file
    (
    std::string &           decrypted_content,
    const std::string &     config_file_name
    )
    {
    if( decrypted_content.find( "[ UNKNOWN ]" ) == std::string::npos ) // FIXME: jmeno pres #define
        {
        // Nothing to fix.
        return;
        }
    // Let's replace [ UNKNOWN ].
    EnumeratorGenerateNodeId    en;
    Global::node_database().enumerate_nodes( NodeId::TYPE_CLIENT, en );
    }*/
    

std::string AccountManager::get_pending_account_directory() const
    {
    return Global::path_manager().get_source_data_directory() + "/" + AM_PENDING_ACCOUNTS_DIRECTORY;
    }


void AccountManager::save_pending_account
    (
    const AccountRequest &      ar
    ) const
    {
    hope( is_service() );
    hope( ar.generated_node_id.is_type( NodeId::TYPE_CLIENT ) );

    std::ostringstream  ss;
    TextWriter  tw( ss );
    tw.line() << get_pending_account_directory() << "/";
    ar.generated_node_id.text_write( tw, Serializer::Description() );
    
    std::string name = ss.str();
    
    VolumeManager::OFStream     ofs( name, std::ios_base::out );
    if( !ofs.is_open() )
        {
        std::ostringstream  ss;
        ss << "Cannot create file \"" << name << "\"";
        throw ExCommonError( ss.str() ); // FIXME: account exception
        }
    TextWriter  tw2( ofs );
    ar.text_write( tw2, Serializer::Description() );
    ofs.close();
    }


void AccountManager::delete_pending_account
    (
    const AccountRequest &      ar
    ) const
    {
    hope( is_service() );
    hope( ar.generated_node_id.is_type( NodeId::TYPE_CLIENT ) );

    std::ostringstream  ss;
    TextWriter  tw( ss );
    tw.line() << get_pending_account_directory() << "/";
    ar.generated_node_id.text_write( tw, Serializer::Description() );
    
    std::string name = ss.str();

    FileSystem & fs = VolumeManager::file_system();
    fs.delete_file( name );
    }


void AccountManager::load_pending_account
    (
    const std::string &     file_name,
    AccountRequest &        ar
    ) const
    {
    hope( is_service() );
        
    VolumeManager::IFStream     ifs( file_name, std::ios_base::in );
    if( !ifs.is_open() )
        {
        std::ostringstream  ss;
        ss << "Cannot open file \"" << file_name << "\"";
        throw ExCommonError( ss.str() ); // FIXME: account exception
        }
    TextReader  tr( ifs );
    ar.text_read( tr, Serializer::Description() );
    ifs.close();
    }


void AccountManager::load_all_pending_accounts()
    {
    FileSystem & fs = VolumeManager::file_system();
    FileIterator it;
    try
        {
        it = fs.open_directory( get_pending_account_directory(), "*" );
        }
    catch( ... )
        {
        // Directory doesn't exist. That's ok.
        return;
        }
    
    while( it.is_valid() )
        {
        if( it.is_file() )
            {
            AccountRequest  ar;
            std::string     file_name = get_pending_account_directory();
            file_name += "/" + it.name();
            load_pending_account( file_name, ar );
            if( find_client_node_id( ar.generated_node_id ) )
                {
                // Request was sucessfully processed, but service crashed before file deletion.
                delete_pending_account( ar );
                }
            else
                {
                add_account_request( ar );
                ar.account_request_status = ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER;
                }
            }
        it++;
        }
    }


bool AccountManager::find_client_node_id
    (
    const NodeId &      node_id
    )
    {
    EnumeratorFindNodeId    en( node_id );
    Global::node_database().enumerate_nodes( NodeId::TYPE_CLIENT, en );
    return en.result;
    }
    

void AccountManager::next_tick()
    {
    // Only service can process next_tick.
    if( is_service() )
        {        
        // Load pending account requests from disk to account request hash table.
        static bool pending_accounts_loaded = false;
        if( !pending_accounts_loaded )
            {
            load_all_pending_accounts();
            pending_accounts_loaded = true;
            }
        NodeId  server_node_id;
        if( any_server_connected( server_node_id ) && account_request_hash.size() )
            {
            // Some server is running. Let's process pending requests.
            AccountRequestHash::iterator it = account_request_hash.begin();
            while( it != account_request_hash.end() )
                {
                AccountRequest & ar = it->second;
/*                if( ar.account_request_status == ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER || 
                    ar.account_request_status == ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER )
                    {
                    const int time_interval = 30; // Interval between tries.
                    if( ar.try_time.is_invalid() || ( SystemTime::current_time() - ar.try_time ) > time_interval )
                        {
                        AccountCreateRequestMessage     request_to_server( ar );
                        request_to_server.send( server_node_id, NetworkCommon::STREAM_RELIABLE );                
                        ar.account_request_status = ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER;
        
                        ar.try_time = SystemTime::current_time();
                        }
                    }*/
                if( ar.account_request_status == ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER )
                    {
                    AccountCreateRequestMessage     request_to_server( ar );
                    request_to_server.send( server_node_id, NetworkCommon::STREAM_RELIABLE );                
                    ar.account_request_status = ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER;
                    }
                if( ar.account_request_status == ACCOUNT_REQUEST_STATUS_ERROR )
                    {
                    // Server refused to create account. Let's try it in some time.
                    const int time_interval = 30; // Interval between tries.
                    if( SystemTime::current_time() - ar.try_time > time_interval )
                        {
                        AccountCreateRequestMessage     request_to_server( ar );
                        request_to_server.send( server_node_id, NetworkCommon::STREAM_RELIABLE );                
                        ar.account_request_status = ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER;
                        }
                    }
                ++it;
                }
            }
        }
    }
    

bool AccountManager::any_server_connected
    (
    NodeId &    server_node_id
    ) const
    {
    EnumeratorFindRunningNode   en;
    Global::node_database().enumerate_nodes( NodeId::TYPE_SERVER, en );
    if( en.found )
        {
        server_node_id = en.node_id;
        return true;
        }
    return false;
    }


} // namespace Massiv::Core
} // namespace Massiv
