/** @file account_manager.h
 * AccountManager declaration.
 */

/* 
 * begin        : Sat Sep 6 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_ACCOUNT_MANAGER_H
#define MASSIV_ACCOUNT_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_STYPE_FLOAT_H
#include "serializer/stype_float.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_NODE_DATABASE_H
#include "database/node_database.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_COPY_FILE_H
#include "data/copy_file.h"
#endif

#include <string>
#include <map>
#include <vector>
#include <stack>
#include <list>
#include <sstream>
#include <queue>
#include <fstream>

namespace Massiv {
namespace Core {


/**
 * Class AccountManager.
 * Manages client's account creation, which is processed in several steps.
 * 1) anonynous node sends system message to service node with a 
 * request to create account.
 * 2) service sends system message to some server with a request to create new instance 
 * of account object.
 * 3) server send respond to service with account object id.
 * 4) service adds new account to client database and sends respond to anonymous node.
 * 5) anonymous node adds information about new account to user named file into login directory.
 *
 * @ingroup unclassified
 */
class MASSIV_EXPORTED AccountManager
    {
public:
    
    enum AccountRequestStatus
        {
        ACCOUNT_REQUEST_STATUS_IDLE                             = 0,
            /**< Account status idle. */
        ACCOUNT_REQUEST_STATUS_SUCCESS                          = 1,
            /**< Account status success. Account object successfuly created in simulation. */
        ACCOUNT_REQUEST_STATUS_PENDING                          = 2,
            /**< Account status pending. Operation is not finished yet. */
        ACCOUNT_REQUEST_STATUS_SEND_REQUEST_TO_SERVER           = 3,
            /**< Account request is prepared to be sent to server. */
        ACCOUNT_REQUEST_STATUS_WAITING_FOR_RESPOND_FROM_SERVER  = 4,
            /**< Account request is waiting for server response. */
        ACCOUNT_REQUEST_STATUS_ERROR                            = 5,
            /**< Account status error. An error occured during account creation. */
        ACCOUNT_REQUEST_STATUS_BITS         = 3
            /**< Number of bits needed for serialization. */
        }; // AccountRequestStatus
        
    /**
     * Information about one account creation request are stored here.
     */
    struct MASSIV_EXPORTED AccountRequest
        {
        
        /** Constructor. */
        AccountRequest();

        /** Copy constructor. */
        AccountRequest
            (
            const AccountRequest & ar
            );

        /** operator= */
        AccountRequest & operator=
            (
            const AccountRequest & ar
            );

        /**
         * Read from a bistream.
         */
        void bitstr_read
            (
            BitIStream &                    bis,
            const Serializer::Description & desc
            );

        /**
         * Write to a bistream.
         */
        void bitstr_write
            (
            BitOStream &                    bos,
            const Serializer::Description & desc
            ) const;
        
        /**
        * Read from text stream.
        */
        void text_read
            (
            TextReader &                        tr,
            const Serializer::Description &     desc
            );

        /**
        * Serialize to text stream.
        */
        void text_write
            (
            TextWriter &                        tw,
            const Serializer::Description &     desc
            ) const;

        NodeId                      anonymous_node_id;
            /**< NodeId of massiv node which "owns" this structure. Must be always anonymous type. */

        NodeId                      generated_node_id;
            /**< Newly generated node id. */

        AccountRequestStatus        account_request_status;
            /**< Account status. */

        SString                     public_key;
            /**< Public key. */

        ObjectId                    account_id;
            /**< Accound object id. */
        
        SystemTime                  try_time;
            /**< Last time when this request tries to be send to a server. 
                 This attribute is not serialized. */
            
        // Magic number temporary disabled.
        //SFloat64                    magic_number;
            /* Magic mumber to distinguish between multiple requests from one node. */
            
        }; // AccountRequest

    typedef ext::hash_map
                <
                NodeId,
                AccountRequest,
                NodeId::Hash,
                NodeId::EqualKey
                > AccountRequestHash;

    /**
     * Enumerates all nodes of specified node type and stores its node ids
     * to attribute node_id. Node id of the last enumerated node remains in this attribute.
     */
    class EnumeratorNodeId : public NodeDatabase::Enumerator
        {
        
        /**
         * Set node id of @param data to attribute node_id.
         */        
        void operator()
            (
            const NodeDatabase::NodeData & data
            );
            
    public:
        
        NodeId      node_id;
            /** Last node id of a requested node type. */
        
        }; // EnumeratorNodeId


    /**
     * Generates information to create uniques client's node id.
     */
    class EnumeratorGenerateNodeId : public NodeDatabase::Enumerator
        {
    public:
        
        /**
         * Constructor.
         */
        EnumeratorGenerateNodeId();
    
        /**
         * Stores the greatest relative id of all enumerated nodes 
         * to attribute max_relative_id.
         */        
        void operator()
            (
            const NodeDatabase::NodeData & data
            );
            
    public:
        
        UInt32      max_relative_id;
            /** Maximum relative id of all client's node ids. */
        
        }; // EnumeratorGenerateNodeId

    /**
     * Enumerator used to find specified node id.
     */
    class EnumeratorFindNodeId : public NodeDatabase::Enumerator
        {    
    public:        
        
        /**
         * Find node id.
         * @param _node_id node id to be found
         */
        EnumeratorFindNodeId
            (
            const NodeId &      _node_id
            );
            
    private:
        /**
         * Set node id of @param data to attribute node_id.
         */        
        void operator()
            (
            const NodeDatabase::NodeData & data
            );
            
    public:
    
        NodeId      node_id;
            /** Node id to be found. */
        bool        result;
            /** Result. True if node id was found. */        
        
        }; // EnumeratorFindNodeId
        
    /**
     * Enumerate running nodes.
     */
    class EnumeratorFindRunningNode : public NodeDatabase::Enumerator
        {    
    public:        
        
        /**
         * Constructor.
         */
        EnumeratorFindRunningNode();
            
    private:
        
        /**
         * Set node id of @param data to attribute node_id.
         */        
        void operator()
            (
            const NodeDatabase::NodeData & data
            );

    public:
        
        NodeId      node_id;
            /** Node id of first found running node. */
        bool        found;
            /** True is some running node was found. */        
        
        }; // EnumeratorFindRunningNode

    /**
     * Constructor.
     */
    AccountManager();
    
    /**
     * Destructor.
     */
    ~AccountManager();

    /**
     * Create new account.
     * This method is permitted to be called only by anonymous node.
     * Result of account cration will be stored in attribute account_request.
     * Method get_account_request_status() should be called to gain operation status.
     * Return true is request was sent to service.
     */
    bool create_account();
    
    /**
     * @return pending account directory.
     * Pending account requests are saved here untill account create request are finished.
     */
    std::string get_pending_account_directory() const;

    /**
     * Save and encrypt account to a file.
     * This method is permitted to be called only by anonymous node.
     * Return true if account successfuly saved.
     */
    bool save_account
        (
        const std::string &         file_name,
        const std::string &         password
        );

    /**
     * Encrypt file.
     * @param file_name file to writed encrypted context to.
     * @param pass password.
     * @param context to encrypt.
     */
    static void encrypt_file
        (
        const std::string &         file_name,
        const std::string &         pass,
        const std::string &         content
        );
   
    /**
     * Encrypts std::string.
     * @param input string to encrypt.
     * @param password password to use.
     * @param crypted is returned crypted string.
     */
    static void encrypt_string
        (
        const std::string &         input,
        const std::string &         password,
        std::string &               crypted
        );
   
    /**
     * Decrypts std::string.
     * @param input string to decrypt.
     * @param password password to use.
     * @param decrypted is returned decrypted string.
     */
    static void decrypt_string
        (
        const std::string &         input,
        const std::string &         password,
        std::string       &         decrypted
        );
    
    /**
     * Return account request status.
     * This method is permitted to be called only by anonymous node.
     * When returned status is equal to ACCOUNT_REQUEST_STATUS_SUCCESS, account
     * is created, and its just a question of time when information about a new account
     * will be distributed by data manager to all massiv servers.
     */
    AccountRequestStatus get_account_request_status() const;

    /**
     * Set idle status.
     * Must be called before next account create request.
     */
    void set_account_request_idle_status();

    /**
     * Return account request structure.
     * This method is permitted to be called only by anonymous node.
     */
    const AccountRequest & get_account_request() const;
    
    /**
     * Generate public/private key using cryptopp library.
     */
    void generate_public_private_key
        (
        std::string &         public_key,
        std::string &         private_key    
        ) const;
    
public:
    // Message handling.
    
    /**
     * Handle account register request message.
     * This message can handle service node only.
     */
    void handle_account_register_request_message
        (
        const AccountRegisterRequestMessage &       msg
        );

    /**
     * Handle account register respond message.
     * This message can handle client node only.
     */
    void handle_account_register_respond_message
        (
        const AccountRegisterRespondMessage &       msg
        );

    /**
     * Handle account create request message.
     * This message can handle server node only.
     */
    void handle_account_create_request_message
        (
        const AccountCreateRequestMessage &       msg
        );

    /**
     * Handle account create respond message.
     * This message can handle service node only.
     */
    void handle_account_create_respond_message
        (
        const AccountCreateRespondMessage &       msg
        );

    /**
     * FIXME
     */
/*    void fix_local_node_config_file
        (
        std::string &           decrypted_content,
        const std::string &     config_file_name
        );
*/
        
public:

    /**
     * AccountManager settings.
     */
    class AccountManagerSettings : public Settings
        {
    public:

        /** Constructor. */
        AccountManagerSettings();

    public:

        Boolean &           manual_management;
            /**< True - manual account management, false - automatic account management. */
        }; // AccountManagerSettings.
    
private:
        
    /**
     * Add account request to account request hash table.
     * This method can called by service node only.
     */
    AccountRequest & add_account_request
        (
        const AccountRequest &      account_request
        );
        
    /**
     * Find account request in account request hash table.
     * This method can called by service node only.
     */
    AccountRequest & find_account_request
        (
        const NodeId &              node_id
        );

    /**
     * Find account request in account request hash table.
     * This method can called by service node only.
     * Return true is request for @param node_id was found.
     */
    bool find_account_request2
        (
        const NodeId &              node_id
        ) const;

    /**
     * Remove account request from account request hash table.
     * This method can called by service node only.
     */
    void remove_account_request
        (
        const NodeId &      node_id
        );

    /**
     * Generate unique client's node id.
     * This method can called by service node only.
     */
    NodeId generate_unique_node_id();
    
    /** Return true if this node is service node. */
    bool is_service() const;
    
    /** Return true if this node is client node. */
    bool is_client() const;
    
    /** Return true if this node is server node. */
    bool is_server() const;

    /** Return true if this node is anonymous node. */
    bool is_anonymous() const;

    /**
     * Return true if account object was created.
     * This method can called by server node only.
     */
    bool create_account_object
        (
        ObjectId &      account_id
        );

    /** Return service node id. */
    NodeId get_service();
    
    /** Return node id of some server. */
    NodeId get_server();
            
    /**
     * Save text serialized account request to special file to disk.
     * @param ar account request
     */
    void save_pending_account
        (
        const AccountRequest &      ar
        ) const;
    
    /**
     * Delete file which contain serialized account request.
     * @param ar account request
     */
    void delete_pending_account
        (
        const AccountRequest &      ar
        ) const;

    /**
     * Load pending account from file to structure AccountRequest.
     * @param file_name file name with text serialization of account request
     * @param ar AccountRequest structure where to load request
     */
    void load_pending_account
        (
        const std::string &     file_name,
        AccountRequest &        ar
        ) const;
    
    /**
     * Load all pending accounts (located in special directory)
     * to account request hash table.
     */
    void load_all_pending_accounts();
    
    /**
     * Find if client's node id is already presented in simulation.
     * @return true is client's node id is presented in simulation
     */
    bool find_client_node_id
        (
        const NodeId &      node_id
        );

    /**
     * @return true is some server is connected to simulation (means that
     * service has this server node id in its node tables) and service
     * can send a message to this server.
     */
    bool any_server_connected
        (
        NodeId &    server_node_id
        ) const;
    

public:

    /**
     * Next tick.
     * If there is some pending account request in a account request hash table,
     * process this request.
     */
    void next_tick();
    
private:

    AccountRequestHash      account_request_hash;
        /** Hash table of account requests, each request per one anonymous node.
            This attribute is used only by service node. */
    
    AccountRequest          account_request;
        /**< Account request structure. 
             This attribute is used only by anonymous node. */
    
    std::string             public_key;
        /**< Public key in quote string form. */
    std::string             private_key;
        /**< Private key in quote string form. */
        
    AccountManagerSettings  account_manager_settings;
        /**< Account manager settings. */

    FileCopy        cp;
        /**< Used for synchronous data files copying. */

    }; // AccountManager


}; // namespace Massiv::Core
}; // namespace Massiv


#endif // MASSIV_ACCOUNT_MANAGER_H_
