/** @file compare_file.cpp
 * Compare files implementation.
 */

/* 
 * begin        : Wed Jul 31 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "data/compare_file.h"
#include "utility/debug.h"
#include <cstring>

namespace Massiv {
namespace Core {

#define COMPARE_BUFFER_SIZE        512000
    /**< Maximum size of buffer used for data comparing. */

CompareFile::CompareFile() :
    busy( false ),
    file_size1( 0 ),
    file_size2( 0 ),
    offset( 0 ),
    buffer1( NULL ),
    buffer2( NULL ),
    result( IDLE )
    {
    }
    
    
CompareFile::~CompareFile()
    {
    }
    
    
bool CompareFile::compare_file
    (
    const std::string &     file1,
    const std::string &     file2
    )
    {
    if( busy )
        {
        return false;
        }
    ifs1.clear();
    ifs1.open( file1.c_str(), std::ios_base::in | std::ios_base::binary );
    if( !ifs1.is_open() )
        {
        return false;
        }
    ifs1.seekg( 0, std::ios_base::end );
    hope( ifs1.good() );

    file_size1 = ifs1.tellg();

    ifs1.seekg( 0, std::ios_base::beg );
    hope( ifs1.good() );
    
    ifs2.clear();
    ifs2.open( file2.c_str(), std::ios_base::in | std::ios_base::binary );
    if( !ifs2.is_open() )
        {
        return false;
        }
    ifs2.seekg( 0, std::ios_base::end );
    hope( ifs2.good() );

    file_size2 = ifs2.tellg();
    if( file_size1 != file_size2 )
        {
        // File differs.
        busy = false;
        ifs1.close();
        ifs2.close();
        result = FILES_DIFFERS;
        return true;
        }

    ifs2.seekg( 0, std::ios_base::beg );
    hope( ifs2.good() );

    UInt32 length = COMPARE_BUFFER_SIZE;
    if( file_size1 < length && file_size2 < length )
        {
        length = file_size1;
        if( length < file_size2 )
            {
            length = file_size2;
            }
        }
    buffer1 = new UInt8[ length ];
    buffer2 = new UInt8[ length ];
    if( !buffer1 || !buffer2 )
        {
        throw ExCommonError( "Memory allocation error." );
        }
    offset = 0;
    busy = true;
    return true;
    }


void CompareFile::next_tick()
    {
    if( !busy )
        {
        return;
        }
    static UInt32 length;
    // Read data to buffer.
    length = COMPARE_BUFFER_SIZE;
    UInt32 rest = file_size1 - offset;
    if( rest == 0 )
        {
        // Copying done.
        ifs1.close();
        ifs2.close();
        delete buffer1;
        delete buffer2;
        buffer1 = NULL;
        buffer2 = NULL;
        busy = false;
        result = FILES_ARE_IDENTICAL;
        return;
        }
    if( rest < length )
        {
        length = rest;
        }
    ifs1.read( ( char * ) buffer1, length );
    ifs2.read( ( char * ) buffer2, length );
    offset += length;
    if( std::memcmp( buffer1, buffer2, length ) != 0 )
        {
        // File differs
        ifs1.close();
        ifs2.close();
        delete buffer1;
        delete buffer2;
        buffer1 = NULL;
        buffer2 = NULL;
        busy = false;            
        result = FILES_DIFFERS;
        }
    }


bool CompareFile::is_compared()
    {
    return !busy;
    }


CompareFile::Result CompareFile::get_result()
    {
    if( busy )
        return COMPARE_PENDING;
    return result;
    }


void CompareFile::set_idle()
    {
    result = IDLE;
    }
  

} // namespace Massiv::Core
} // namespace Massiv
