/** @file compare_file.h
 * File comparing declaration.
 */

/* 
 * begin        : Thu Jul 31 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_COMPARE_FILE_H
#define MASSIV_COMPARE_FILE_H

#ifndef MASSIV_VOLUME_MANAGER_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_VOLUME_MANAGER_H
#include "filesys/volume_manager.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

namespace Massiv {
namespace Core {

/**
 * Provides method for files comparing.
 * @code
 * CompareFile cf;
 * if( !cf.compare_file( "song_copy.mp3", "song.mp3" ) )
 *      {
 *      throw Exception;
 *      }
 * while( !cf.is_compared() )
 *      {
 *      cf.next_tick();
 *      }
 * if( cf.is_compared() )
 *      {
 *   // Files are compared.
 *      }
 * switch( cf.get_result() )
 *      {
 *      case FILES_ARE_IDENTICAL:
 *          break;
 *      case FILES_DIFFERS:
 *          break;
 *      }
 * @endcode
 *
 * @ingroup utility
 */
class CompareFile
    {
public:

    enum Result
        {
        IDLE,
            /**< Idle value. */
        FILES_ARE_IDENTICAL,
            /**< Files are identical. */
        FILES_DIFFERS,
            /**< Files differs. */
        COMPARE_PENDING
            /**< Comparing is still in progress. */
        }; // Result

    /**
     * Constructor.
     */
    CompareFile();

    /**
     * Destructor.
     */
    ~CompareFile();
    
    /**
     * Compare files.
     * Call @method is_compared to find comparing is finished.
     * Returns false when it is not possible to compare files.
     */
    bool compare_file
        (
        const std::string &     file1,
        const std::string &     file2
        );
        
    /**
     * Next tick.
     * Must be called till @method is_compared() return true.
     */
    void next_tick();

    /**
     * Return true when files coparing is finished.
     */
    bool is_compared();

    /**
     * Return result of file comparing.
     */
    Result get_result();
    
    /**
     * Set idle result.
     */
    void set_idle();

private:

    bool                        busy;
        /**< True during file coparing. */
        
    VolumeManager::FStream      ifs1;
        /**< First input file. */

    VolumeManager::FStream      ifs2;
        /**< Second input file. */
    
    UInt32                      file_size1;
        /**< File1 size. */
    
    UInt32                      file_size2;
        /**< File2 size. */

    UInt32                      offset;
        /**< Actual file offset. */

    UInt8 *                     buffer1;
        /**< Compare buffer for file1. */
        
    UInt8 *                     buffer2;
        /**< Compare buffer for file2. */

    Result                      result;
        /** Comparing result. */
                    
    }; // CompareFile

}; // namespace Massiv::Core
}; // namespace Massiv


#endif // MASSIV_COMPARE_FILE_H
