/** @file copy_file.cpp
 * Asynchronous/synchronous copy file implementation.
 */

/* 
 * begin        : Wed Feb 2 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "data/copy_file.h"
#include "utility/debug.h"
#include "global/global.h"
#include "filesys/file_system.h"

namespace Massiv {
namespace Core {

#define MAX_COPY_REQUESTS       1
    /**< Maximum copy requests in one time. */

#define COPY_BUFFER_SIZE        512000
    /**< Maximum size of buffer used for copying. */

#define SELECT_FILE_STREAM \
    std::iostream *     _ifs = NULL; \
    std::iostream *     _ofs = NULL; \
    if( use_vm ) \
        { \
        _ifs = &this->ifs1; \
        _ofs = &this->ofs1; \
        } \
    else \
        { \
        _ifs = &this->ifs2; \
        _ofs = &this->ofs2; \
        } \
    std::iostream &     ifs = *_ifs; \
    std::iostream &     ofs = *_ofs;

FileCopy::FileCopy
    (
    bool        _use_vm,
    Mode        _mode
    ) :
    async_io( NULL ),
    busy( false ),
    file_size( 0 ),
    offset( 0 ),
    buffer( NULL ),
    activity( READ ),
    mode( _mode ),
    use_vm( _use_vm )
    {
    if( mode == ASYNCHRONOUS )
        {
        async_io = new AsyncIO( MAX_COPY_REQUESTS );
        }            
    }
    
    
FileCopy::~FileCopy()
    {
    if( !async_io )
        {
        delete async_io;
        async_io = NULL;
        }
    }
    
    
bool FileCopy::copy_file
    (
    const std::string &     destination,
    const std::string &     source
    )
    {
    if( busy )
        {
        return false;
        }

    SELECT_FILE_STREAM
    
    ifs.clear();
    if( use_vm )
        {
        this->ifs1.open( source.c_str(), std::ios_base::in | std::ios_base::binary );
        if( !this->ifs1.is_open() )
            {
            return false;
            }
        }
    else
        {
        this->ifs2.open( source.c_str(), std::ios_base::in | std::ios_base::binary );
        if( !this->ifs2.is_open() )
            {
            return false;
            }
        }
    ifs.seekg( 0, std::ios_base::end );
    hope( ifs.good() );

    file_size = ifs.tellg();
    offset = 0;

    ifs.seekg( 0, std::ios_base::beg );
    hope( ifs.good() );
    
    ofs.clear();
    if( use_vm )
        {
        this->ofs1.open( destination.c_str(), std::ios_base::out | std::ios_base::binary );
        if( !this->ofs1.is_open() )
            {
            return false;
            }
        }
    else
        {
        std::string::size_type  pos = destination.rfind( "/" );
        if( pos != std::string::npos )
            {
            // Create directory first.
            // FIXME
            std::string temp( destination, 0, pos );
            Global::native_file_system().create_directory( temp );
            }
        
        this->ofs2.open( destination.c_str(), std::ios_base::out | std::ios_base::binary );
        if( !this->ofs2.is_open() )
            {
            return false;
            }
        }
    UInt32 length = COPY_BUFFER_SIZE;
    if( file_size < length )
        {
        length = file_size;
        }
    buffer = new UInt8[ length ];
    if( !buffer )
        {
        throw ExCommonError( "Memory allocation error." );
        }
    activity = READ;
    busy = true;
    return true;
    }


void FileCopy::next_tick()
    {
    if( !busy )
        {
        return;
        }
    if( mode == ASYNCHRONOUS && async_io->are_requests_pending() )
        {
        return;
        }

    SELECT_FILE_STREAM

    // Previous request finished.
    static UInt32 length;
    if( activity == READ )
        {
        // Read data to buffer.
        length = COPY_BUFFER_SIZE;
        UInt32 rest = file_size - offset;
        if( rest == 0 )
            {
            // Copying done.
            if( use_vm )
                {
                this->ifs1.close();
                this->ofs1.close();
                }
            else
                {
                this->ifs2.close();
                this->ofs2.close();
                }
            delete buffer;
            buffer = NULL;
            busy = false;            
            return;
            }
        if( rest < length )
            {
            length = rest;
            }
        switch( mode )
            {
            case SYNCHRONOUS:
                {
                ifs.read( ( char * )buffer, length );
                break;
                }
            case ASYNCHRONOUS:
                {
                async_io->async_read( ifs , buffer, offset, length );
                break;
                }
            default:
                panic( "Unknown copying mode." );
            }
        activity = WRITE;
        return;
        }
    if( activity == WRITE )
        {
        // Write data from buffer.
        switch( mode )
            {
            case SYNCHRONOUS:
                {
                ofs.write( ( char * )buffer, length );
                break;
                }
            case ASYNCHRONOUS:
                {
                async_io->async_write( ofs , buffer, offset, length );
                break;
                }
            default:
                panic( "Unknown copying mode." );
            }
        offset += length;
        activity = READ;
        return;
        }
    }

bool FileCopy::is_copied()
    {
    return !busy;
    }
    
}; // namespace Massiv::Core
}; // namespace Massiv
