/** @file copy_file.h
 * Asynchronous/synchronous file copying declaration.
 */

/* 
 * begin        : Wed Feb 2 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_COPY_FILE_H
#define MASSIV_COPY_FILE_H


#ifndef MASSIV_ASYNC_IO_H
#include "filesys/async_io.h"
#endif

#ifndef MASSIV_VOLUME_MANAGER_H
#include "filesys/volume_manager.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#include <fstream>

namespace Massiv {
namespace Core {

/**
 * Provides method for asynchronous/synchronous file copying.
 * @code
 * FileCopy cp;
 * if( !cp.copy_file( "song_copy.mp3", "song.mp3" ) )
 *      {
 *      throw Exception;
 *      }
 * while( !cp.is_copied() )
 *      {
 *      cp.next_tick();
 *      }
 * if( cp.is_copied() )
 *   {
 *   // File is already copied.
 *   }
 * @endcode
 *
 * @ingroup utility
 */
class FileCopy
    {
public:

    enum Activity
        {
        READ,
            /**< Reading from source file to buffer. */
        WRITE        
            /**< Writing to destination file from buffer. */
        }; // Activity
        
    enum Mode
        {
        ASYNCHRONOUS,
            /**< Asynchronous copying using asycn_io. */
        SYNCHRONOUS
            /**< Synchronous copying. */
        }; // Mode

    /**
     * Constructor.
     * Default copy mode is SYNCHRONOUS.
     * @param _use_vm true - use volume manager, false - use std::fstream.
     */
    FileCopy
        (
        bool        _use_vm = true,
        Mode        _mode = SYNCHRONOUS
        );

    /**
     * Destructor.
     */
    ~FileCopy();
    
    /**
     * Copy file source to destination.
     * Call @method is_copied to find copying is finished.
     * Returns false when it is not possible to copy file.
     */
    bool copy_file
        (
        const std::string &     destination,
        const std::string &     source
        );
        
    /**
     * Next tick.
     * Must be called till @method is_copied() return true.
     */
    void next_tick();

    /**
     * Return true when file copying is finished.
     */
    bool is_copied();
    
private:

    AsyncIO *                   async_io;
        /**< Request queue for asynchronous IO when copying file. */

    bool                        busy;
        /**< True during file copying. */
        
    VolumeManager::FStream      ifs1;
        /**< Input file. */

    VolumeManager::FStream      ofs1;
        /**< Output file. */
    
    std::fstream                ifs2;
        /**< Input file. */

    std::fstream                ofs2;
        /**< Output file. */

    UInt32                      file_size;
        /**< File size. */
    
    UInt32                      offset;
        /**< Actual file offset. */

    UInt8 *                     buffer;
        /**< Copy buffer. */
        
    Activity                    activity;
        /**< Copy activity. */

    Mode                        mode;
        /** Copying mode. */
                    
    bool                        use_vm;
        /**< FIXME */
        
    }; // FileCopy

}; // namespace Massiv::Core
}; // namespace Massiv


#endif // MASSIV_COPY_FILE_H
