/** @file data_manager.cpp
 * DataManager implementation.
 */

/*
 * begin        : Tue Nov 20 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "data/data_manager.h"
#include "data/path_manager.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "utility/debug.h"
#include "global/common_garbage.h"
#include "status/exception.h"
#include "serializer/serializer_bool.h"
#include "serializer/serializer_int.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "time/system_time.h"
#include "sysevent/data_msg.h"
#include "registry/registry.h"
#include "global/global.h"
#include "system/system.h"
#include "filesys/file_system.h"
#include "filesys/volume_manager.h"
#include "filesys/compact_volume.h"
#include "serializer/serializer.h"
#include <stack>
#include <sstream>
#include <time.h>

namespace Massiv {
namespace Core {

typedef DataManager::DataVersion                      DataVersion;
typedef DataManager::DataNodeTree                     DataNodeTree;
typedef DataManager::DataObjectList                   DataObjectList;
typedef DataManager::DataDescription                  DataDescription;
typedef DataManager::DownloadProvider                 DownloadProvider;
typedef DataManager::UploadProvider                   UploadProvider;
typedef DataManager::UploadProvider::UploadInfo       UploadInfo;
typedef DataManager::FileSizeType                     FileSizeType;
typedef DataManager::MagicNumberType                  MagicNumberType;
typedef DataManager::DataFileDownloadType             DataFileDownloadType;
typedef DataManager::DataFileList                     DataFileList;
typedef DataManager::DataFileInfo                     DataFileInfo;
typedef DataManager::DataVersion                      DataVersion;
typedef DataManager::UpdateProvider                   UpdateProvider;
typedef DataManager::DataNodeTree::DataNode           DataNode;

#define DM_DD_FORMAT_VERSION                "1.01"
    /**< Data description format version. */
#define DM_DD_FORMAT_VERSION_STRING         "FORMAT VERSION="
    /**< Data description format version string. */

#define DM_DD_NODE                          "node"
    /**< Data description node string. */
#define DM_DD_ENDNODE                       "endnode"
    /**< Data description endnode string. */
#define DM_DD_TYPE                          "type"
    /**< Data description type string. */
#define DM_DD_FILE                          "file"
    /**< Data description file string. */
#define DM_DD_OBSOLETE                      "obsolete"
    /**< Data description obsolete string. */
#define DM_DD_DATAOBJECT                    "dataobject"
    /**< Data description dataobject string. */
#define DM_DD_NECESSARY_CLIENT              "necessary_client"
    /**< Data description necessary client string. */
#define DM_DD_NECESSARY_SERVER              "necessary_server"
    /**< Data description necessary server string. */
#define DM_DD_CLIENT                        "client"
    /**< Data description client string. */
#define DM_DD_SERVER                        "server"
    /**< Data description server string. */


#define DM_DNT_ROOT_NAME                    "root"
    /**< Name of a root of a data node tree. */
#define DM_DOWNLOAD_FILE_NAME               ".download"
    /**< Temporary download file name. */
#define DM_COPY_FILE_NAME                   ".copy"
    /**< Temporary copy file name. */

#define DOWNLOAD_TIMEOUT                    10
    /**< Download timeout in seconds. After this time new request is sent to
         a server. */

#define UPLOAD_TIMEOUT                      10
    /**< Upload timeout in seconds. After this time data upload to a node is canceled. */

#define MAX_DOWNLOAD_SPEED                  16384
    /**< Maximum download/upload speed per data request in kilobyte/second.
         This value can be changed in configuration files. */

#define MAX_PATH                            256
    /**< Maximum length of buffer to store directory name. */
    
#define CHECK_DATA_VERSION_INTERVAL         60
    /**< Interval between data version check. Time is in seconds. */

#define DM_NUM_UPLOAD_PER_SECOND            4
    /**< Number of data uploads per one second. The greater number, the better continuousness
         on a slow modem connection. 
         Example: if an upload speed is 2048 b/s, and the value is 2, server 
                  will send 1024b every 0.5 second. */

#define THROW_DATA_MANAGER_NOT_READY \
    throw ExNotReady( "Data object manager is not completely initilized yet." );

#define THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE \
    throw ExNotMasterDataServer( "Data object manager type is not master data service." );


/**
 * Assert within realase mode.
 */
#define assert_release( condition, text ) \
    if( ( condition ) == 0 ) \
        { \
        std::ostringstream  ss; \
        ss << text << std::endl << "Source file: " << __FILE__ << " line: " << __LINE__ << std::endl; \
        throw ExDataManagementError( ss.str() ); \
        }


// **** non-class functions ****

/**
 * Return next name token from @param str where @param separator 
 * is name token separator and move @param iter to point to next name token.
 * Example: return "ahoj" from "/name1/name2/name3" when @param iter
 *          points to the beginning of "ahoj", if separator == '/'
 */
static std::string get_next_name_token
    (
    const std::string &                 str,
    std::string::const_iterator &       iter,
    char                                separator = '/'
    )
    {
    std::string::size_type       pos     = iter - str.begin();
    assert_release( ( pos >= 0 ) && ( pos < str.length() ), "Read error." );
    std::string::size_type       newpos  = str.find( separator, pos );
    if( newpos == std::string::npos )
        {
        newpos = str.length();
        }
    std::string temp( str, pos, newpos - pos);
    iter += newpos - pos;
    return temp;
    }

/**
 * Return file size of file @param file_name.
 */
static FileSizeType get_file_size
    (
    const std::string &     file_name
    )
    {
    VolumeManager::IFStream   ifs( file_name.c_str(), std::ios::in );
    if( !ifs )
        {
        std::ostringstream  ss;
        ss << "Data file \"" << file_name << "\" cannot be opened for reading.";
        throw ExDataManagementError( ss.str() );
        }
    ifs.seekg( 0, std::ios_base::end );
    if( ifs.fail() )
        {
        std::ostringstream  ss;
        ss << "Unexpected error occured when reading size of file \"" << file_name << "\".";
        throw ExDataManagementError( ss.str() );
        }
    FileSizeType  file_size = ifs.tellg();
    //assert_release( file_size != std::ifstream::pos_type( std::ifstream::off_type( -1 ) ), "File access error." );
    ifs.close();
    return file_size;
    }


/**
 * Rename file.
 * @param old_file_name old file name
 * @param new_file_name new file name
 */
int rename
    (
    const char * old_file_name,
    const char * new_file_name
    )
    {
    try
        {
        FileSystem & fs = VolumeManager::file_system();
        fs.rename_file( old_file_name, new_file_name );
        }
    catch( ... )
        {
        return -1;
        }
    return 0;
    }


/**
 * Remove file.
 * @param file_name file to be removed
 */
int remove
    (
    const char * file_name
    )
    {
    try
        {
        FileSystem & fs = VolumeManager::file_system();
        fs.delete_file( file_name );
        }
    catch( ... )
        {
        return -1;
        }
    return 0;
    }

// **** DataManager start ****

DataManager::DataManager()
    {
    initialized = false;
    data_description = NULL;
    ready_for_use = false;
    }


DataManager::~DataManager()
    {
    destroy();
    }


void DataManager::initialize
    (
    Type    the_type
    )
    {
    NodeId              data_service_node;
    
    // Extract data service node id from cvar.
    try    
        {
        std::string         str = data_manager_settings.data_service_node;
        std::istringstream  iss( str );
        TextReader          tr( iss );
    
        tr.next_line();
        data_service_node.text_read( tr, Serializer::Description() );
        }
    catch( Exception & )
        {
        Global::log_warning
            (
            Status::FACILITY_DATA_MANAGEMENT,
            Status::PRIORITY_HIGH,
            "Could not extract data service node id from registry."
            );
        }
    
    // Initialize.
    _initialize
        ( 
        Global::path_manager().get_data_directory(),//data_manager_settings.data_directory,
        Global::path_manager().get_source_data_directory(),//data_manager_settings.source_directory,
        the_type,
        static_cast< DownloadType >( data_manager_settings.download_type.read() ),
        false,
        true,
        data_service_node
        );
    }


void DataManager::_initialize
    (
    const std::string &             data_dir,
    const std::string &             source_dir,
    Type                            init_type,
    DownloadType                    init_download_type,
    bool                            test,
    bool                            exclusive_access,
    NodeId                          data_service_node_id
    )
    {
    destroy();

    initialized = true;

    data_directory = data_dir;
    source_directory = source_dir;
    std::string::size_type  size = data_directory.size() - 1;
    if( data_directory.find_last_of( "/\\" ) == size )
        {
        // Remove last char.
        std::string::iterator it = data_directory.begin() + size;
        data_directory.erase( it );
        }
    data_description = NULL;
    ready_for_use = false;
    type = init_type;
    download_type = init_download_type;
    dp.data_service_node_id = data_service_node_id;
    // @todo: inicializovat up, dp
    phase = PHASE_INIT;
    test_mode = test;
    update.initialize();
    
    if( data_directory.empty() )
        {
        data_directory = "."; // Actual directory.
        }

    // Mount volume if requested.
    if( data_manager_settings.mount_data_volume )
        {
        // Create data directory.
        Global::native_file_system().create_directory( data_directory );
        std::string image_path = data_directory + "/" + 
            data_manager_settings.data_volume_image_name.read();
        // Mount volume.
        try
            {
            VolumeManager::mount_volume( image_path );
            }
        catch( ExFileSystemError & )
            {
            // Format the volume.
            VolumeManager::format_volume( image_path, DM_DATA_VOLUME_NAME );
            }
        // Redirect access to the mounted volume.
        data_directory = std::string( DM_DATA_VOLUME_NAME ) + ":/";
        }
    }


void DataManager::destroy()
    {
    ready_for_use = false;
    data_file_list.destroy();
    up.destroy( NOT_READY_ERROR );
    dp.destroy();
    if( !used_data_object_hash.empty() )
        {
        // Some data objects are still in use.

        std::ostringstream oss;
        oss << "Registered data objects must be deleted before data object manager deletion.\n" <<
            "The following data objects still exist:";

        UsedDataObjectHash::const_iterator it;
        bool first = true;
        for( it = used_data_object_hash.begin(); it != used_data_object_hash.end(); ++it )
            {
            if( first )
                {
                oss << " ";
                first = false;
                }
            else
                {
                oss << ", ";
                }
            oss << it->first;
            }
        
        throw ExUsedDataObjectHashTableNotEmpty( oss.str() );
        }

    delete data_description;
    data_description = NULL;
    initialized = false;
    }


bool DataManager::everything_idle()
    {
    bool result = phase == PHASE_IDLE;
    if( type == DATA_CLIENT || type == DATA_SERVER )
        {
        return result;
        }
    // data service.
    UploadProvider::UploadInfoHash::iterator it = up.upload_info_hash.begin();
    while( it != up.upload_info_hash.end() )
        {
        UploadProvider::UploadInfo & ui = it->second;        
        if( !ui.deleted )
            {
            result = result && ( ui.phase == UploadProvider::PHASE_IDLE );
            }
        ++it;
        }
    return result;
    }


void DataManager::next_tick()
    {
    if( !initialized )
        return;
    
    SystemTime  old_time = Global::system_time();
    bool once = true;
    while( once || ( !everything_idle() && Global::system_time() - old_time < 0.020f ) ) // 20ms
        {
    
        switch( phase )
            {
            case PHASE_INIT:
                {
                phase_init();
                break;
                }
            case PHASE_IDLE:
                {
                phase_idle();
                break;
                }
            case PHASE_DOWNLOAD:
                {
                phase_download();
                break;
                }
            case PHASE_CLEANUP:
                {
                phase_cleanup();
                break;
                }
            case PHASE_UPDATE:
                {
                phase_update();
                break;
                }
            default:
                panic( "Unexpected phase." );
            }

        if( phase == PHASE_IDLE )
            {
            up_next_tick(); // Upload provider next tick.
            }
        once = false;
        } // while
    }


void DataManager::phase_init()
    {
    hope( phase == PHASE_INIT );

    std::string     data_description_file_name( data_directory + "/" + DM_DD_FILE_NAME );
    // Load data description file from disk.
    VolumeManager::IFStream ifs_dd( data_description_file_name.c_str(), std::ios_base::in );
    if( !ifs_dd.is_open() )
        {
        // Cannot open data description file.
        hope( dp.data_version == DataVersion( 0, 0 ) );
        // Create data description file for directory access test.
        VolumeManager::OFStream ofs( data_description_file_name.c_str(), std::ios_base::out );
        if( !ofs )
            {
            throw ExCanNotCreateDataDescriptionFile
                    (
                    "Cannot create data description file \"" + data_description_file_name + "\"."
                    );
            }
        // File successfully created - directory access test passed.

        // Simulate data description.
        TextWriter          tw( ofs );
        // Write data description format version.
        tw.line() << DM_DD_FORMAT_VERSION_STRING << DM_DD_FORMAT_VERSION;
        tw.end_line();
        // Write version.
        tw.line() << "[1.0." << generate_magic_number() << "]";
        tw.end_line();

        dp.data_version = DataVersion( 1, 0 );
        ofs.close();

        ifs_dd.clear();
        ifs_dd.open( data_description_file_name.c_str(), std::ios_base::in );
        if( !ifs_dd.is_open() )
            {
            std::ostringstream      ss;
            ss << "Cannot open data description file \"" + data_description_file_name + "\".";
            throw ExDataManagementError( ss.str() );
            }
        }

    DataDescription::get_initial_information_from_data_description
        (
        data_directory, 
        ifs_dd,
        dp.data_version,
        dp.magic_number
        );
        
    ifs_dd.close();

    std::string     data_file_list_file_name( data_directory + "/" + DM_DATA_FILE_LIST_FILE_NAME );
    // Load data file list from disk.
    VolumeManager::IFStream ifs_dfl( data_file_list_file_name.c_str(), std::ios_base::in );
    if( !ifs_dfl )
        {
        VolumeManager::OFStream ofs( data_file_list_file_name.c_str(), std::ios_base::out );
        if( !ofs )
            {
            std::ostringstream  ss;
            ss << "Cannot create data file list \"" << data_file_list_file_name << "\".";
            throw ExDataManagementError( ss.str() );
            }
        ofs.close();
        }

    load_data_file_list();
    
    switch( type )
        {
        case MASTER_DATA_SERVICE:
            {
            download_type = NO_DOWNLOAD;
            dp.data_description = load_data_description();
            hope( dp.data_description != NULL );
            dp.data_file_names.str( "" );
            dp.data_file_names.clear();
            int num;
            num = get_data_file_names_to_download
                (
                FULL_DOWNLOAD,
                dp.data_file_names,
                dp.data_description
                );
            if( num != 0 )
                {
                std::ostringstream  ss;
                ss << "Data files:" << std::endl << dp.data_file_names.str() << " are missing in data directory.";
                throw ExDataManagementError( ss.str() );
                }
            data_description = dp.data_description;
            dp.data_description = NULL;
            
            num = set_usable_data_nodes();
            hope( ( UInt32 ) num == data_file_list.data_file_list_map.size() );
                       
            phase = PHASE_IDLE;
            if( !test_mode )
                {
                // cant switch to automatic update in test mode.
                switch_to_update_phase( DataManager::AUTOMATIC );
                }
            else
                {
                // no update in test mode.
                ready_for_use = true;
                }
            break;
            }
        case SECONDARY_DATA_SERVICE:
        case DATA_SERVER:
        case DATA_CLIENT:
            {
            data_description = load_data_description();
            
            // Download actual data version first from some data service.
            DataVersionRequestMessage     msg;
            msg.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );

            phase = PHASE_IDLE;
            break;
            }
        default:
            panic( "Unexpected type." );
        }
    }


void DataManager::phase_idle()
    {
    hope( phase == PHASE_IDLE );
    switch( type )
        {
        case MASTER_DATA_SERVICE:
            {
            // Check if update is requested.
            if( update.next_update )
                {
                update.next_update = false;
                // Next update request was received during last update phase.
                switch_to_update_phase( DataManager::AUTOMATIC );
                }
            else
                {
                // Check for newer data version (locally changed).
                static DataVersion     old_data_version = DataVersion( 0, 0 );
                DataVersion            actual_version = get_data_version();
                hope( actual_version >= old_data_version );
                if( actual_version > old_data_version )
                    {
                    old_data_version = actual_version;
                    // Send data version message to massiv servers (not clients).
                    UploadProvider::UploadInfoHash::iterator it = up.upload_info_hash.begin();
                    it = up.upload_info_hash.begin();
                    while( it != up.upload_info_hash.end() )
                        {
                        UploadInfo & ui = it->second;
                        if( !ui.deleted && ui.node_id.is_type( NodeId::TYPE_SERVER ) )
                            {
                            DataVersionMessage    msg( actual_version, get_magic_number() );
                            msg.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
                            }
                        ++it;
                        }
                    }
                }
            break;
            }
        case SECONDARY_DATA_SERVICE:
        case DATA_SERVER:
        case DATA_CLIENT:
            {
            // FIXME: temporary hack
            static bool reg_done = false;
            if( !reg_done )
                {
                const NodeId & node_id = Global::get_local_node_id();
                if( !node_id.is_type( NodeId::TYPE_ANONYMOUS ) )
                    {
                    register_to_data_service();
                    reg_done = true;
                    }
                }
                
            // Check for newer data version (received by system message).
            static SystemTime   last_check_time = Massiv::System::system_time();
            SystemTime          check_time = Massiv::System::system_time();
            if( check_time - last_check_time >= ( int )data_manager_settings.check_data_version_interval )
                // && type == DATA_CLIENT )
                {
                // Only clients send periodically check data request message.
                last_check_time = check_time;

                // Download actual data version now.
                DataVersionRequestMessage     msg;
                msg.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );
                }
            if( dp.fresh_download_info.valid )
                {
                dp.fresh_download_info.valid = false;
                DataVersion         new_version         = dp.fresh_download_info.data_version;
                DataVersion         actual_version      = dp.data_version;
                if( dp.magic_number != dp.fresh_download_info.magic_number && dp.magic_number != 0 )
                    {
                    // Magic number was changed - all data must be deleted.
                    // This can happen only when data manager is not ready for use.
                    if( ready_for_use != false )
                        {
                        panic( "Data magic number was changed." );
                        }
                    
                    std::ostringstream  ss;
                    ss << "Data magic number of master data service was changed."
                       << "\nAll data must be updated.";
                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
                        ss.str()
                        );
                    // Empty downloaded file list.
                    VolumeManager::OFStream     ofs( data_directory + "/" + DM_DATA_FILE_LIST_FILE_NAME, std::ios_base::out );
                    if( !ofs.is_open() )
                        {
                        std::ostringstream  ss;
                        ss << "Cannot create file " << data_directory << "/" << DM_DATA_FILE_LIST_FILE_NAME;
                        throw ExDataManagementError( ss.str() );
                        }
                    ofs.close();
                    data_file_list.data_file_list_map.clear();
                    
                    dp.active_download_info = dp.fresh_download_info;
                    dp.phase = dp.PHASE_DATA_DESCRIPTION_DOWNLOAD;
                    dp.status = dp.STATUS_INIT;

                    phase = PHASE_DOWNLOAD;                    
                    }
                else
                    {
                    // magic number is the same.
                    assert_release( new_version >= actual_version, 
                        "New data version is lower than actual data version whilst magic number was not changed." );
                    if( new_version > actual_version )
                        {
                        dp.active_download_info = dp.fresh_download_info;
                        dp.phase = dp.PHASE_DATA_DESCRIPTION_DOWNLOAD;
                        dp.status = dp.STATUS_INIT;

                        phase = PHASE_DOWNLOAD;
                        }
                    else
                        {
                        // new_version == actual_version -> no need to download data description,
                        // but some data files may be missing.
                        if( !ready_for_use )
                            {                        
                            dp.data_description = load_data_description();
                            dp.active_download_info = dp.fresh_download_info;
                    
                            delete data_description;
                            data_description = dp.data_description;
                            dp.data_description = NULL;

                            dp.download_type = NECESSARY_DOWNLOAD;
                            dp.phase = dp.PHASE_DATA_FILES_DOWNLOAD;
                            dp.status = dp.STATUS_INIT;
        
                            phase = PHASE_DOWNLOAD;
                            }
                        }
                    }
                }
            if( !dp.custom_data_file_download.empty() && phase == PHASE_IDLE )
                {
                dp.actual_downloaded_file_name = dp.custom_data_file_download.front();
                dp.custom_data_file_download.pop_front();
                DataNode & dn = data_description->data_node_tree.get_data_node( data_description->get_data_node_name( dp.actual_downloaded_file_name ) );
                dp.actual_downloaded_data_version = dn.data_file_info.version; // Version from DD.
                
                // Data file in actual version is not downloaded yet.
                dp.status = dp.STATUS_INIT;
                dp.phase = dp.PHASE_FILE_DOWNLOAD;
                dp.phase_before = dp.PHASE_IDLE;

                phase = PHASE_DOWNLOAD;
                }
            break;
            }
        default:
            panic( "Unexpected type." );
        }
    }


DataDescription * DataManager::load_data_description
    (
    bool                    light_weigth,
    const std::string &     file_name
    ) const
    {
    std::string     data_description_file_name;
    if( file_name.empty() )
        {
        // default data description file name.
        data_description_file_name = data_directory + "/" + DM_DD_FILE_NAME;
        }
    else
        {
        // custom data description file name.
        data_description_file_name = file_name;
        }
    VolumeManager::IFStream   ifs( data_description_file_name.c_str(), std::ios_base::in );
    if( !ifs.is_open() )
        {
        std::ostringstream      ss;
        ss << "Cannot open data description file \"" + data_description_file_name + "\".";
        throw ExDataManagementError( ss.str() );
        }
    DataDescription * dd = new DataDescription( ifs, data_directory, light_weigth );
    hope( dd != NULL );
    ifs.close();
    return dd;
    }


void DataManager::phase_cleanup()
    {
    hope( phase == PHASE_CLEANUP );
    // @todo: implementace - co to ma vubec delat - ma to smysl?
    }


void DataManager::phase_update()
    {
    hope( phase == PHASE_UPDATE );
    hope( type == MASTER_DATA_SERVICE );
    
    switch( update.phase )
        {
        case UpdateProvider::PHASE_IDLE:
            {
            // Nothing to do.
            break;
            }
        case UpdateProvider::PHASE_LOAD_AND_CHECK_UPDATE_FILE:
            {
            updatep_phase_load_and_check_update_file();
            break;
            }
        case UpdateProvider::PHASE_BACKUP_DATA_FILES:
            {
            updatep_phase_backup_data_files();
            break;
            }
        case UpdateProvider::PHASE_COMPARE_DATA_FILES:
            {
            updatep_phase_compare_data_files();
            break;
            }
        case UpdateProvider::PHASE_COPY_DATA_FILES:
            {
            updatep_phase_copy_data_files();
            break;
            }
        case UpdateProvider::PHASE_UPDATE_DATA_DESCRIPTION:
            {
            updatep_phase_update_data_description();
            break;
            }
        default:
            panic( "Unexpected update phase." );
        }
/*        volatile UpdateNotify & un = *( static_cast< UpdateNotify * >( shmem->get_memory() ) );
        un = UPDATE_NOTIFY_UPDATE_ERROR;*/
    }
    

void DataManager::updatep_phase_load_and_check_update_file()
    {
    hope( update.phase == update.PHASE_LOAD_AND_CHECK_UPDATE_FILE );
    hope( type == MASTER_DATA_SERVICE );

    // Load update config file.
    if( !update.load_and_check( *this ) )
        {
        // Serious error. Cannot update.
        std::ostringstream  ss;
        ss << "Data update was not finished. Data in source directory must be repaired. "
           << "Service will run anyway...";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        phase = PHASE_IDLE;
        return;
        }

    // Update file is ok. Proceed to data file copying.
    if( !update.test )
        {
/*        // Create backup file list.
        std::ostringstream ss;
        DataVersion dv = data_description->get_data_version();
        ss << ".backup/" << dv.get_major() << "." << dv.get_minor() << "." << data_description->get_magic_number();
        update.backup_dir = data_directory + "/" + ss.str();
        if( test_file_exist( update.backup_dir + "/" + DM_DD_FILE_NAME, false ) )
            {
            // Backup of this version of data already exist - no backup will be processed.
            update.status = update.STATUS_INIT;
            update.phase = update.PHASE_COMPARE_DATA_FILES;
            return;
            }
        hope( update.backup_file_list.empty() );

        update.backup_file_list.push_back( DM_DD_FILE_NAME );
        update.backup_file_list.push_back( DM_DATA_FILE_LIST_FILE_NAME );

        DataFileList::DataFileListMap::iterator it = data_file_list.data_file_list_map.begin();
        while( it != data_file_list.data_file_list_map.end() )
            {
            const DataFileInfo & dfi = it->second;
            update.backup_file_list.push_back( dfi.name );
            ++it;
            }

        update.phase = update.PHASE_BACKUP_DATA_FILES; */
        
        // Backup is disabled.
        update.status = update.STATUS_INIT;
        update.phase = update.PHASE_COMPARE_DATA_FILES;        
        }
    else
        {
        // No updating.
        phase = PHASE_IDLE;
        }
    }


void DataManager::updatep_phase_backup_data_files()
    {
    hope( false ); // Backup is disabled.
    
    hope( update.phase == update.PHASE_BACKUP_DATA_FILES );
    hope( type == MASTER_DATA_SERVICE );

    if( !update.cp.is_copied() )
        {
        update.cp.next_tick();
        }
    else
        {
        // Next data file to backup
        if( update.backup_file_list.size() )
            {
            const std::string   file_name = update.backup_file_list.back();
            update.backup_file_list.pop_back();
            const std::string   source = data_directory + "/" + file_name;
            const std::string   destination = update.backup_dir + "/" + file_name;
            if( !update.cp.copy_file( destination, source ) )
                {
                std::ostringstream  ss;
                ss << "Update error: cannot copy data file \"" << source << "\" to \"" << destination << "\".";
                throw ExDataManagementError( ss.str() );
                }            
            }
        else
            {
            // All data files are backuped.
            update.status = update.STATUS_INIT;
            update.phase = update.PHASE_COMPARE_DATA_FILES;
            }
        }
    }
    
    
void DataManager::updatep_phase_compare_data_files()
    {
    hope( update.phase == update.PHASE_COMPARE_DATA_FILES );
    hope( type == MASTER_DATA_SERVICE );
    
    static DataDescription::DataFileNameList::iterator it;
    
    switch( update.status )
        {
        case UpdateProvider::STATUS_INIT:
            {
            update.cf.set_idle();
            it = update.ud->file_list.begin();
            update.status = update.STATUS_PENDINGX;
            break;            
            }
        case UpdateProvider::STATUS_PENDINGX:
            {
            if( !update.cf.is_compared() )
                {
                update.cf.next_tick();
                }
            else
                {
                switch( update.cf.get_result() )
                    {
                    case CompareFile::IDLE:
                        // No file was compared yet.
                        break;
                    case CompareFile::FILES_ARE_IDENTICAL:
                        // This data file will not be copied.
                        break;
                    case CompareFile::FILES_DIFFERS:
                        update.copy_file_list.push_back( update.actual_file_name );
                        break;
                    default:
                        panic( "Unexpected compare file result." );
                    }
                if( it != update.ud->file_list.end() )
                    {
                    // Next data file to compare.
                    update.actual_file_name = it->first;
                    ++it;
                    if( data_file_list.find( update.actual_file_name ) )
                        {
                        // This data file exist in actual data version. We must compare it with
                        // a version from source directory.
                        const std::string    file1 = source_directory + "/" + update.actual_file_name;
                        const std::string    file2 = data_directory + "/" + update.actual_file_name;
                        if( !update.cf.compare_file( file1, file2 ) )
                            {
                            std::ostringstream  ss;
                            ss << "Update error: cannot compare data file \"" << file1 << "\" with \"" << file2 << "\".";
                            throw ExDataManagementError( ss.str() );
                            }
                        }
                    else
                        {
                        // New data file in data system.
                        update.copy_file_list.push_back( update.actual_file_name );
                        }
                    }
                else
                    {
                    // Every data file was compared with a new version from source directory.
                    update.status = update.STATUS_INIT;
                    update.phase = update.PHASE_COPY_DATA_FILES;
                    }
                }
            break;            
            }
        default:
            panic( "Unexpected update status." );
        }
    }


void DataManager::updatep_phase_copy_data_files()
    {
    hope( update.phase == update.PHASE_COPY_DATA_FILES );
    hope( type == MASTER_DATA_SERVICE );
    
    static UpdateProvider::VecStr::iterator it;
    
    switch( update.status )
        {
        case UpdateProvider::STATUS_INIT:
            {
            update.actual_file_name = "";
            it = update.copy_file_list.begin();
            update.status = update.STATUS_PENDINGX;
            break;            
            }
        case UpdateProvider::STATUS_PENDINGX:
            {
            if( false ) // FIXME: dirty hack.//!update.cp.is_copied() )
                {
                //update.cp.next_tick();
                }
            else
                {
                std::string     file_name_tmp( data_directory + "/" + update.actual_file_name + ".tmp" );
                std::string     file_name( data_directory + "/" + update.actual_file_name );
                // See if we did copy of some data file before and this 
                // data file is already present in data system, so 
                // some data object can use this file, or it can be used
                // when uploading data.
                if( data_file_list.find( update.actual_file_name ) )
                    {
                    // File successfully copied.
                    remove( file_name_tmp.c_str() ); // Shouldn't be there.

                    // Close file if it is open.
                    up.close_open_file( update.actual_file_name );
                    // Close open files of data objects.
                    int num_open_files = close_data_object_open_file( file_name );
                        
                    if( num_open_files == 0 )
                        {
                        rename( file_name.c_str(), file_name_tmp.c_str() );
                        }
                    else
                        {
                        // Old version of data file is still in use.
                        const DataFileInfo & dfi = data_file_list.get( update.actual_file_name );
                        std::ostringstream ss;
                        ss << file_name << "." << dfi.version.get_major() << "_" << dfi.version.get_minor();
                        file_name_tmp = ss.str();
            
                        remove( file_name_tmp.c_str() );

                        if( rename( file_name.c_str(), file_name_tmp.c_str() ) != 0 )
                            {
                            std::ostringstream  ss;
                            ss << "Cannot rename file \"" << file_name << "\" to \"" << file_name_tmp << "\".";
                            throw ExDataManagementError( ss.str() );
                            }
                        }
                    if( rename( ( data_directory + "/" + DM_COPY_FILE_NAME ).c_str(), file_name.c_str() ) != 0 )
                        {
                        std::ostringstream  ss;
                        ss << "Cannot rename file \"" << data_directory << "/" << DM_COPY_FILE_NAME << "\" to \"" << file_name << "\".";
                        throw ExDataManagementError( ss.str() );
                        }
                    
                    update_data_file_after_download( update.actual_file_name );
                    update_used_data_object_after_data_file_update( update.actual_file_name );

                    if( num_open_files )
                        {
                        // Restore open files of data object.
                        restore_data_object_file( file_name, file_name_tmp );
                        }                    
                    }
                else
                    {
                    // This data file is a new one in data system.
                    if( update.actual_file_name.size() )
                        {
                        remove( file_name_tmp.c_str() ); // Shouldn't be there.
                        rename( file_name.c_str(), file_name_tmp.c_str() );
                        if( rename( ( data_directory + "/" + DM_COPY_FILE_NAME ).c_str(), file_name.c_str() ) != 0 )
                            {
                            std::ostringstream  ss;
                            ss << "Cannot rename file \"" << data_directory << "/" << DM_COPY_FILE_NAME << "\" to \"" << file_name << "\".";
                            throw ExDataManagementError( ss.str() );
                            }                        
                        }
                    }
                // Next data file to copy.
                if( it != update.copy_file_list.end() )
                    {
                    const std::string &  file_name = *it;
                    update.actual_file_name = file_name;
                    const std::string    source = source_directory + "/" + file_name;
                    const std::string    destination = data_directory + "/" + DM_COPY_FILE_NAME;
                    if( !update.cp.copy_file( destination, source ) )
                        {
                        std::ostringstream  ss;
                        ss << "Update error: cannot copy data file \"" << source << "\" to \"" << destination << "\".";
                        throw ExDataManagementError( ss.str() );
                        }            
                    while( !update.cp.is_copied() )
                        {
                        update.cp.next_tick();
                        }                    
                    // Now, file is copied.
                    ++it;
                    }
                else
                    {
                    // All data files are copied into data directory.
                    update.phase = update.PHASE_UPDATE_DATA_DESCRIPTION;
                    }
                }
            break;            
            }
        default:
            panic( "Unexpected update status." );
        }
    }


void DataManager::updatep_phase_update_data_description()
    {
    hope( update.phase == update.PHASE_UPDATE_DATA_DESCRIPTION );
    hope( type == MASTER_DATA_SERVICE );
    
    // Delete data files from data description and from data file list.
    UpdateProvider::VecStr::iterator it = update.delete_file_list.begin();
    UpdateProvider::VecStr::iterator end = update.delete_file_list.end();
    while( it != end )
        {
        data_description->file_list.erase( *it );
        data_file_list.data_file_list_map.erase( *it );
        ++it;
        }

    bool should_update = false; // Should we increase data version?
    DataNodeTree::DataNodeStack s;
    DataNodeTree::add_sons_to_stack( update.ud->data_node_tree.root_node, s );
    while( !s.empty() )
        {
        const DataNode * node = s.top();
        hope( node != NULL );
        s.pop();
        DataNodeTree::add_sons_to_stack( node, s );

        // Try if node already existed.
        try
            {
            data_description->data_node_tree.get_data_node( node->full_name );
            }
        catch( ExDataNodeNotExist & )
            {
            DataNode * dn = new DataNode();
            dn->name = node->name;
            dn->full_name = node->full_name;
            dn->download_type = node->download_type;
            dn->data_file_info.name = node->data_file_info.name;
            dn->data_file_info.version = DataVersion( 0, 99 ); // Automatically increased to 1.0
            
            add_data( dn, "", false );
            should_update = true;
            }

        // Try if data file is the same.
        DataNode & dn = data_description->data_node_tree.get_data_node( node->full_name );
        if( node->data_file_info.name != dn.data_file_info.name )
            {
            // File name differs. Version stays.
            should_update = true;
            
            std::string old_name = dn.data_file_info.name;
            if( !data_description->file_list.insert
                (
                DataDescription::DataFileNameList::value_type( node->data_file_info.name, dn.full_name )
                ).second )
                {
                // Should not happen.
                throw ExDataFileAlreadyUsed( "Data file " + node->data_file_info.name + " already used in another data node." );
                }
            dn.data_file_info.name = node->data_file_info.name;
            dn.data_file_info.load_file_info( data_directory );
            
            // Update data file list.
            data_file_list.add( dn.data_file_info );
            
            // Add this data file name to copy file list(if not already presented)
            // to enable data file update.
            bool found = false;
            UpdateProvider::VecStr::const_iterator it = update.copy_file_list.begin();
            while( it != update.copy_file_list.end() )
                {
                if( *it == dn.data_file_info.name )
                    {
                    found = true;
                    }
                ++it;
                }
            if( !found )
                {
                update.copy_file_list.push_back( dn.data_file_info.name );
                }
            }

        } // while
    
    UpdateProvider::VecStr::iterator it3 = update.copy_file_list.begin();
    while( it3 != update.copy_file_list.end() )
        {
        update_data_file( *it3, false );
        should_update = true;
        ++it3;
        }
    update.copy_file_list.clear();    
    update.delete_file_list.clear();

    DataObjectList::DataObjectInfoHash::iterator it2 = update.ud->data_object_list.data_object_info_hash.begin();
    while( it2 != update.ud->data_object_list.data_object_info_hash.end() )
        {
        DataObjectList::DataObjectInfo & doi = it2->second;
        try
            {
            if( doi.node_path != data_description->data_object_list.get_data_object_node_path( doi.data_object_name ) )
                {
                add_data_object( doi.data_object_name, doi.node_path, false );    
                should_update = true;
                }
            }
        catch( ExDataObjectNotExist & )
            {
            add_data_object( doi.data_object_name, doi.node_path, false );    
            should_update = true;
            }        
        ++it2;
        }    
    //@todo: zapsat nejake checksum at se da overit ze data jsou konzistentni.

    if( should_update )
        {
        data_description->data_version.increase();
        save_data_description();
        save_data_file_list();

        std::ostringstream  ss;
        ss << "Data update successfully finished.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        }
    else
        {
        std::ostringstream  ss;
        ss << "Data was not updated. Data is up to date.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        }    
/*    remove( ( data_directory + "/" + DM_UPDATE_FILE_NAME ).c_str() );
    
    volatile UpdateNotify & un = *( static_cast< UpdateNotify * >( shmem->get_memory() ) );
    un = UPDATE_NOTIFY_UPDATE_FINISHED;*/
    
    set_usable_data_nodes();
    if( !ready_for_use )
        {
        // Startup update.
        ready_for_use = true;
        }
    phase = PHASE_IDLE;
    }


void DataManager::phase_download()
    {
    hope( phase == PHASE_DOWNLOAD );

    switch( dp.phase )
        {
        case DownloadProvider::PHASE_IDLE:
            {
            // Custom download of one data file finished.
            phase = PHASE_IDLE;
            break;
            }
        case DownloadProvider::PHASE_DATA_DESCRIPTION_DOWNLOAD:
            {
            dp_phase_data_description_download();
            break;
            }
        case DownloadProvider::PHASE_DATA_FILES_DOWNLOAD:
            {
            dp_phase_data_files_download();
            break;
            }
        case DownloadProvider::PHASE_FILE_DOWNLOAD:
            {
            dp_phase_file_download();
            break;
            }
        default:
            panic( "Unexpected download phase." );
        }
    // If phase == PHASE_IDLE , downloading is finished. Status of downloading 
    // is available in dp.status.
    }


void DataManager::dp_phase_data_description_download()
    {
    hope( phase == PHASE_DOWNLOAD );
    hope( dp.phase == dp.PHASE_DATA_DESCRIPTION_DOWNLOAD );
    
    switch( dp.status )
        {
        case DownloadProvider::STATUS_INIT:
            {
            dp.actual_downloaded_file_name = DM_DD_FILE_NAME;
            dp.actual_downloaded_data_version = dp.active_download_info.data_version;
            
            dp.phase = dp.PHASE_FILE_DOWNLOAD;
            dp.phase_before = dp.PHASE_DATA_DESCRIPTION_DOWNLOAD;
            break;
            }
        case DownloadProvider::STATUS_SUCCESS:
            {
            // Data description file successfully downloaded.
            // Load data description file from disk.
            std::string     data_description_file_name( data_directory + "/" + DM_DD_FILE_NAME );
            VolumeManager::IFStream   ifs_dd( data_description_file_name.c_str(), std::ios_base::in );
            if( !ifs_dd )
                {
                std::ostringstream  ss;
                ss << "Cannot open downloaded data description file \"" << data_description_file_name << "\".";
                throw ExDataManagementError( ss.str() );
                }
            dp.data_description = new DataDescription( ifs_dd, data_directory );
            dp.magic_number = dp.data_description->magic_number;
            assert_release( dp.data_description != NULL, "Memory allocation error." );
            
            // Test if some data file names was changed.
            data_file_list.refresh( *data_description, *dp.data_description, data_directory, *this );            
            save_data_file_list();
            
            delete data_description;
            data_description = dp.data_description;
            dp.data_description = NULL;
            dp.data_version = data_description->get_data_version();
            
            ifs_dd.close();
            assert_release
                (
                dp.active_download_info.data_version == data_description->data_version,
                "Downloaded data description version differs."
                );            
            
            set_usable_data_nodes();
            
            dp.download_type = NECESSARY_DOWNLOAD;
            dp.phase = dp.PHASE_DATA_FILES_DOWNLOAD;
            dp.status = dp.STATUS_INIT;
            break;
            }
        default:
            panic( "Unexpected download status." );
        }
    }


void DataManager::dp_phase_file_download()
    {
    hope( phase == PHASE_DOWNLOAD );
    hope( dp.phase == dp.PHASE_FILE_DOWNLOAD );
    
    switch( dp.status )
        {
        case DownloadProvider::STATUS_INIT:
            {
            dp.data_stream.open( ( data_directory + "/" + DM_DOWNLOAD_FILE_NAME ).c_str(), std::ios_base::out | std::ios::binary );
            if( !dp.data_stream.is_open() )
                {
                std::ostringstream  ss;
                ss << "Cannot create download file \"" << data_directory << "/" << DM_DOWNLOAD_FILE_NAME << "\".";
                throw ExDataManagementError( ss.str() );
                }
            dp.offset = 0;
            DataRequestMessage     msg
                ( 
                dp.actual_downloaded_file_name,
                dp.actual_downloaded_data_version,
                data_manager_settings.max_download_speed
                );
            msg.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );
            dp.last_request_time = SystemTime::current_time();
            dp.status = dp.STATUS_PENDINGX;
            break;
            }        
        case DownloadProvider::STATUS_SUCCESS:
            {
            // File successfully downloaded.
            dp.data_stream.close();
            std::string     file_name_tmp( data_directory + "/" + dp.actual_downloaded_file_name + ".tmp" );
            std::string     file_name( data_directory + "/" + dp.actual_downloaded_file_name );

            remove( file_name_tmp.c_str() );

            // Close file if it is open.
            up.close_open_file( dp.actual_downloaded_file_name );
            // Close open files of data objects.
            int num_open_files = close_data_object_open_file( file_name );
                
            if( num_open_files == 0 )
                {
                rename( file_name.c_str(), file_name_tmp.c_str() );
                }
            else
                {
                // Old version of data file is still in use.
                const DataFileInfo & dfi = data_file_list.get( dp.actual_downloaded_file_name );
                std::ostringstream ss;
                ss << file_name << "." << dfi.version.get_major() << "_" << dfi.version.get_minor();
                file_name_tmp = ss.str();
    
                remove( file_name_tmp.c_str() );

                if( rename( file_name.c_str(), file_name_tmp.c_str() ) != 0 )
                    {
                    std::ostringstream  ss;
                    ss << "Cannot rename file \"" << file_name << "\" to \"" << file_name_tmp << "\".";
                    throw ExDataManagementError( ss.str() );
                    }
                }
            if( rename( ( data_directory + "/" + DM_DOWNLOAD_FILE_NAME ).c_str(), file_name.c_str() ) != 0 )
                {
                std::ostringstream  ss;
                ss << "Cannot rename file \"" << data_directory << "/" << DM_DOWNLOAD_FILE_NAME << "\" to \"" << file_name << "\".";
                throw ExDataManagementError( ss.str() );
                }
            
            if( file_name != ( data_directory + "/" + DM_DD_FILE_NAME ) )
                {
                update_data_file_after_download( dp.actual_downloaded_file_name );
                update_used_data_object_after_data_file_update( dp.actual_downloaded_file_name );
                }

            if( num_open_files )
                {
                // Restore open files of data object.
                restore_data_object_file( file_name, file_name_tmp );
                }
            dp.phase = dp.phase_before;
            if( dp.phase == dp.PHASE_DATA_FILES_DOWNLOAD )
                {
                dp.status = dp.STATUS_PENDINGX;
                }
            break;
            }
        case DownloadProvider::STATUS_FAILED:
            {
            dp.data_stream.close();
            switch( dp.error )
                {
                case NOT_READY_ERROR:
                    {
                    throw ExDataManagementError( "Data server is not ready. Maybe server shutdown." );
                    break;
                    }
                case DATA_FILE_ERROR:
                    {
                    std::ostringstream ss;
                    ss << "Data file \"" << dp.actual_downloaded_file_name
                       << "\" doesn't exist in actual data description.";
                    throw ExDataManagementError( ss.str() );
                    break;
                    }
                case DATA_FILE_VERSION_ERROR:
                case DATA_SERVICE_UPDATE:
                    {
                    // We have to download newer version of data description.
                    phase = PHASE_IDLE;
                    //dp.phase = dp.PHASE_DATA_DESCRIPTION_DOWNLOAD;
                    //dp.status = dp.STATUS_INIT;
                    break;
                    }
                default:
                    panic( "Unexpected data message error type." );
                }
            break;
            }
        case DownloadProvider::STATUS_PENDINGX:
            {
            dp.current_time = SystemTime::current_time();
            hope( dp.current_time >= dp.last_request_time );
            if( dp.current_time - dp.last_request_time > DOWNLOAD_TIMEOUT )
                {
                dp.data_stream.close();
                dp.status = dp.STATUS_INIT;
                }
            break;
            }
        default:
            panic( "Unexpected download status." );
        }
    }


void DataManager::dp_phase_data_files_download()
    {
    hope( phase == PHASE_DOWNLOAD );
    hope( dp.phase == dp.PHASE_DATA_FILES_DOWNLOAD );    

    switch( dp.status )
        {
        case DownloadProvider::STATUS_INIT:
            {
            dp.data_file_names.str( "" );
            dp.data_file_names.clear();
            int num;
            num = get_data_file_names_to_download
                (
                dp.download_type,
                dp.data_file_names,
                data_description
                );
            if( num == 0 )
                {
                // No files to download.
                dp.status = dp.STATUS_SUCCESS;
                }
            else
                {
                dp.status = dp.STATUS_PENDINGX;
                }
            break;
            }
        case DownloadProvider::STATUS_SUCCESS:
            {
            // Data files successfully downloaded.            
            switch( dp.download_type )
                {
                case NECESSARY_DOWNLOAD:
                    {
                    // Data manager can offer data objects now.
                    switch( download_type )
                        {
                        case NECESSARY_DOWNLOAD:
                            {
                            set_usable_data_nodes();
                            ready_for_use = true;
                            phase = PHASE_IDLE;
                            break;
                            }
                        case FULL_DOWNLOAD:
                            {
                            dp.download_type = FULL_DOWNLOAD;
                            dp.status = dp.STATUS_INIT;
                            break;
                            }
                        default:
                            panic( "Unexpected download type." );
                        }
                    break;
                    }
                case FULL_DOWNLOAD:
                    {
                    set_usable_data_nodes();
                    ready_for_use = true;
                    phase = PHASE_IDLE;
                    break;
                    }
                default:
                    panic( "Unexpected download type." );
                }
            break;
            }
        case DownloadProvider::STATUS_FAILED:
            {
            phase = PHASE_IDLE;
            break;
            }
        case DownloadProvider::STATUS_PENDINGX:
            {
            std::getline( dp.data_file_names, dp.actual_downloaded_file_name );
            if( !dp.actual_downloaded_file_name.empty() )
                {
                DataNode & dn = data_description->data_node_tree.get_data_node( data_description->get_data_node_name( dp.actual_downloaded_file_name ) );
                dp.actual_downloaded_data_version = dn.data_file_info.version;
                
                dp.status = dp.STATUS_INIT;
                dp.phase = dp.PHASE_FILE_DOWNLOAD;
                dp.phase_before = dp.PHASE_DATA_FILES_DOWNLOAD;
                }
            else
                {
                // All files downloaded
                dp.status = dp.STATUS_SUCCESS;
                }
            break;
            }
        default:
            panic( "Unexpected download status." );
        }
    
    }

void DataManager::up_next_tick()
    {
    // Client cannot upload data.
    if( type == DATA_SERVER || type == DATA_CLIENT )
        {
        return;
        }

    if( !ready_for_use )
        {
        return; 
        }

    up.clear_deleted_requests();
    
    UploadProvider::UploadInfoHash::iterator it = up.upload_info_hash.begin();
    while( it != up.upload_info_hash.end() )
        {
        UploadProvider::UploadInfo & ui = it->second;
        hope( ui.deleted == false );
        
        switch( ui.phase )
            {
            case UploadProvider::PHASE_SEND_DATA_VERSION:
                {
                up_send_data_version( ui );
                break;
                }
            case UploadProvider::PHASE_UPLOAD_DATA_FILE:
                {
                if( !ui.wait_for_confirmation )
                    {
                    up_upload_data( ui );
                    }
                else
                    {
                    // Waiting for confirmation.
                    ui.current_time = SystemTime::current_time();
                    hope( ui.current_time >= ui.last_upload_time );
                    if( ui.current_time - ui.last_upload_time >= UPLOAD_TIMEOUT )
                        {
                        // Uploading failed.
                        ui.phase = up.PHASE_IDLE;
                        }
                    }
                break;
                }
            case UploadProvider::PHASE_IDLE:
                {
                break;
                }
            default:
                panic( "Unexpected upload phase." );
            }
        ++it;
        }
    }


void DataManager::up_send_data_version
    (
    UploadProvider::UploadInfo &    ui
    )
    {
    DataVersionMessage      msg( get_data_version(), get_magic_number() );
    msg.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
    ui.phase = up.PHASE_IDLE;
    }


void DataManager::up_upload_data
    (
    UploadProvider::UploadInfo &    ui
    )
    {
    if( ui.num_uploads == -1 )
        {
        // First data packet.
        ui.first_upload_time = SystemTime::current_time();
        ui.num_uploads = 0;
        }
    
    // Check speed limit.
    SystemTime time = SystemTime::current_time();
    hope( time >= ui.first_upload_time );
    Float64 link_speed = 0;
    if( time > ui.first_upload_time )
        {
        link_speed = ( ui.offset + ui.buffer_size ) / ( time - ui.first_upload_time ).get_value();
        }
    if( link_speed > ui.average_speed )
        {
        // Cannot upload now - link speed would be faster than speed limit.
        return;
        }
    // Test actual upload speed. It can differ from an average upload speed.
    if( time < ui.last_upload_time + ( Float64 )1 / DM_NUM_UPLOAD_PER_SECOND )
        {
        // Cannot upload now - actual link speed would be faster upload speed limit.
        return;
        }
    
    ui.last_upload_time = time;
    std::ostringstream  ss;
    ss << "DataManager: uploading file \"" << ui.data_file_name << "\" version "
        << ui.data_version.get_major() << "." << ui.data_version.get_minor()
        << " to node " << ui.node_id << ". Actual offset = " << ui.offset
        << " Average speed = " << ( int )link_speed << " B/s";
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );

    assert_release( ( int )ui.data_stream.tellg() == ui.offset, "File \"" + ui.data_file_name + "\" seek error.");
    int rest = ui.file_size - ui.offset;
    assert_release( rest >= 0, "Error: message size less than zero." );
    if( rest < ui.buffer_size )
        {
        ui.buffer_size = rest;
        }
    DataMessage      msg( ui.buffer_size, ui.data_file_name );
    ui.data_stream.read( ( char *) msg.buffer, ui.buffer_size );
    assert_release( !ui.data_stream.fail() , "File \"" + ui.data_file_name + "\" read error." );
    
    msg.offset = ui.offset;
    
    ui.offset += ui.buffer_size;
    if( ui.file_size != ui.offset )
        {
        // We must wait for confirmation message before we can send next data message.
        ui.wait_for_confirmation = true;
        }
    else
        {
        // Passed to the end of stream.
        std::ostringstream  ss;
        ss << "DataManager: file \"" << ui.data_file_name << "\" version "
           << ui.data_version.get_major() << "." << ui.data_version.get_minor()
           << " successfuly uploaded to node " << ui.node_id << ". Size = " << ui.offset
           << " Overall upload speed = " << ( int )link_speed << " B/s";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );
        ui.phase = up.PHASE_IDLE;
        ui.data_stream.close();
        ui.data_stream.clear();
        
        msg.end = true;
        }
    msg.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
    ui.num_uploads++;
    }


DataObject * DataManager::get_data_object
    (
    const std::string &             data_object_name,
    bool                            open_data_file
    )
    {
    if( !ready_for_use )
        {
        THROW_DATA_MANAGER_NOT_READY
        //return NULL;
        }
    DataObject * data_object = NULL;
    try
        {
        data_object = new DataObject( this, data_object_name, open_data_file );
            
        UsedDataObjectHash::iterator it = used_data_object_hash.find( data_object_name );
        if( it == used_data_object_hash.end() )
            {
            // Data objects of this name don't have list yet.
            UsedDataObjectList     used_data_object_list;
            std::pair< UsedDataObjectHash::iterator, bool > result;
            result = used_data_object_hash.insert
                ( 
                UsedDataObjectHash::value_type( data_object_name, used_data_object_list )
                );
            if( !result.second )
                {
                std::ostringstream  ss;
                ss << "Cannot insert data object name \"" << data_object_name << "\" into used data object hash table.";
                
                throw ExDataManagementError( ss.str() );
                }
            it = result.first;
            }
        UsedDataObjectList & udolist = it->second;        
        udolist.push_front( data_object );
        } // try
    catch( ... )
        {
        delete data_object;
        throw;
        }
    return data_object;
    }


void DataManager::update_used_data_object_after_data_file_update
    (
    const std::string &     data_file_name
    )
    {
    // Update data objects which actually use this data file.
    std::string     full_data_file_name = data_directory + "/" + data_file_name;
    UsedDataObjectHash::iterator it = used_data_object_hash.begin();
    while( it != used_data_object_hash.end() )
        {
        UsedDataObjectList & udolist = it->second;
        UsedDataObjectList::iterator it_list = udolist.begin();
        while( it_list != udolist.end() )
            {
            DataObject *    data_object = *it_list;
            if( data_object->data_file_full_name == full_data_file_name )
                {
                data_object->update();
                }
            ++it_list;
            }
        ++it;
        }
    
    std::string     data_node_name = data_description->get_data_node_name( data_file_name );
    
    // Update data objects which points to data node which own this data file.
    DataObjectList::DataObjectInfoHash::iterator it2 
        = data_description->data_object_list.data_object_info_hash.begin();
    while( it2 != data_description->data_object_list.data_object_info_hash.end() )
        {
        DataObjectList::DataObjectInfo & doinfo = it2->second;
        if( doinfo.node_path == data_node_name )
            {
            // I found data object which points to data node which owns this data file.
            update_used_data_object( doinfo.data_object_name );
            }
        it2++;
        }
    }


void DataManager::update_used_data_object
    (
    const std::string &     data_object_name
    )
    {
    UsedDataObjectHash::iterator it = used_data_object_hash.find( data_object_name );
    if( it != used_data_object_hash.end() )
        {
        UsedDataObjectList & udolist = it->second;
        UsedDataObjectList::iterator it_list = udolist.begin();
        while( it_list != udolist.end() )
            {
            DataObject *    data_object = *it_list;
            data_object->update();
            ++it_list;
            }
        }
    }


bool DataManager::is_ready_for_use() const
    {
    return ready_for_use;
    }


DataVersion DataManager::get_data_version() const
    {
    if( !ready_for_use )
        {
        THROW_DATA_MANAGER_NOT_READY
        }
    hope( data_description != NULL );
    return data_description->get_data_version();
    }


MagicNumberType DataManager::get_magic_number() const
    {
    if( !ready_for_use )
        {
        THROW_DATA_MANAGER_NOT_READY
        }
    hope( data_description != NULL );
    return data_description->get_magic_number();
    }


void DataManager::destroy_request
    (
    const NodeId    & node_id
    )
    {
    // Only master data service has some registered nodes.
    if( type == MASTER_DATA_SERVICE )
        {
        up.remove_request( node_id );
        }
    }


void DataManager::register_to_data_service() const
    {
    // Sending data version request message will cause registration
    // in data manager of data service.
    DataVersionRequestMessage     msg;
    msg.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );
    }


void DataManager::add_data
    (
    DataNodeTree::DataNode *        new_data_node,
    const std::string &             data_object_name,
    bool                            should_save_to_disk
    )
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }
    if( phase != PHASE_UPDATE )
        {
        throw ExDataManagementError( "DataManager is not in update phase" );
        }
    hope( new_data_node != NULL );
    hope( data_description != NULL );
    
    // Test file name.
    if( new_data_node->data_file_info.name[0] == '.' || new_data_node->data_file_info.name[0] == '[')
        {
        std::ostringstream  ss;
        ss << "Data file name \"" << new_data_node->data_file_info.name << "\" cannot start with \".\" or \"[\"" << std::endl;
        throw ExDataManagementError( ss.str() );
        }

    if( !test_file_exist( new_data_node->data_file_info.name ) )
        {
        if( !new_data_node->deprecated )
            {
            std::ostringstream  ss;
            ss << "Data file \"" << new_data_node->data_file_info.name << "\" doesn't exist in data directory.";
            throw ExDataManagementError( ss.str() );
            }
        }
    else
        {
        // Data file exist.
        new_data_node->data_file_info.load_file_info( data_directory );
        data_file_list.add( new_data_node->data_file_info );
        }

    data_description->add_data_node( new_data_node );
    if( !data_object_name.empty() )
        {
        data_description->add_data_object( data_object_name, new_data_node->full_name );
        update_used_data_object( data_object_name );
        }
    if( should_save_to_disk )
        {
        data_description->data_version.increase();
        save_data_description();
        save_data_file_list();
        }
    }


void DataManager::add_data_object
    (
    const std::string &             data_object_name,
    const std::string &             node_path,
    bool                            should_save_to_disk
    )
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }
    if( phase != PHASE_UPDATE )
        {
        throw ExDataManagementError( "DataManager is not in update phase" );
        }
    assert_release( data_object_name != "", "Undefined data object name." );
    assert_release( node_path != "", "Undefined data node path." );
    hope( data_description != NULL );
    
    data_description->add_data_object( data_object_name, node_path );
    update_used_data_object( data_object_name );
    
    if( should_save_to_disk )
        {
        data_description->data_version.increase();
        save_data_description();
        }
    }


void DataManager::remove_data_object
    (
    const std::string &             data_object_name,
    bool                            should_save_to_disk
    )
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }
    if( phase != PHASE_UPDATE )
        {
        throw ExDataManagementError( "DataManager is not in update phase" );
        }
    hope( data_description != NULL );
    data_description->remove_data_object( data_object_name );
    if( should_save_to_disk )
        {
        data_description->data_version.increase();
        save_data_description();
        }
    }


void DataManager::save_data_description()
    {
    hope( type == MASTER_DATA_SERVICE );

    std::string     old_dd( data_directory + "/" + DM_DD_FILE_NAME + ".tmp" );
    std::string     dd( data_directory + "/" + DM_DD_FILE_NAME );
    remove( old_dd.c_str() );

    // We must check uploading of data_description file.
    up.close_open_file( DM_DD_FILE_NAME );

    if( rename( dd.c_str(), old_dd.c_str() ) != 0 )
        {
        throw ExCanNotRenameDataDescriptionFile
            (
            "Cannot rename data description file " + dd + " to " + old_dd + " ."
            );
        }        
    VolumeManager::OFStream   ofs( dd.c_str(), std::ios_base::out );
    if( !ofs )
        {
        throw ExCanNotCreateDataDescriptionFile
                (
                "Cannot create data description file " + dd + " ."
                );
        }
    data_description->save_description( ofs );
    ofs.close();
    }


int DataManager::set_usable_data_nodes()
    {
    int num = 0;
     
    DataFileList::DataFileListMap::iterator  it = data_file_list.data_file_list_map.begin();
    while( it != data_file_list.data_file_list_map.end() )
        {
        const std::string & data_file_name = it->first;
        const std::string   data_node_name = data_description->get_data_node_name( data_file_name );
        DataNode & data_node = data_description->data_node_tree.get_data_node( data_node_name );
        
        data_node.unusable = false;
        num++;
        
        ++it;
        }
    return num;
    }


int DataManager::get_data_file_names_to_download
    (
    DownloadType                    download_type,
    std::ostream &                  stream,
    const DataDescription *         data_description
    )
    {
    hope( data_description != NULL );

    int num = 0;
    DataNodeTree::DataNodeStack s;

    DataNodeTree::add_sons_to_stack( data_description->data_node_tree.root_node, s );
    while( !s.empty() )
        {
        DataNode * node = s.top();
        hope( node != NULL );
        s.pop();
        bool        should = false;
        switch( download_type )
            {
            case NECESSARY_DOWNLOAD:
                {
                // Note sons of root node are automatically NECESSARY_DATA_FILE.
                if( node->deprecated )
                    {
                    break;
                    }
                if( type == DATA_CLIENT && flags_on( node->download_type, NECESSARY_CLIENT_DATA_FILE ) )
                    {
                    should = true;
                    }
                if( type == DATA_SERVER && flags_on( node->download_type, NECESSARY_SERVER_DATA_FILE ) )
                    {
                    should = true;
                    }
                break;
                }
            case FULL_DOWNLOAD:
                {
                if( node->deprecated )
                    {
                    break;
                    }
                if( type == DATA_CLIENT && ( flags_on( node->download_type, CLIENT_DATA_FILE ) 
                    || flags_on( node->download_type, NECESSARY_CLIENT_DATA_FILE ) ) )
                    {
                    should = true;
                    }
                if( type == DATA_SERVER && ( flags_on( node->download_type, SERVER_DATA_FILE )
                    || flags_on( node->download_type, NECESSARY_SERVER_DATA_FILE ) ) )
                    {
                    should = true;
                    }
                break;
                }
            default:
                panic( "Unexpected download type." );
            } // switch
        DataNodeTree::add_sons_to_stack( node, s );
        if( should )
            {
            should = false;
            // Now we know this data file should be downloaded.
            // Test data file version.
            if( !data_file_list.find( node->data_file_info.name ) )
                {
                // Data file is not present in data directory in any version.
                should = true;
                }
            else
                {
                DataFileInfo & dfi = data_file_list.get( node->data_file_info.name );
                std::ostringstream  ss;
                ss << "Wrong data file \"" << node->data_file_info.name << "\" version.";
                assert_release( dfi.version <= node->data_file_info.version, ss.str() );
                if( dfi.version < node->data_file_info.version )
                    {
                    // New version is available to download.
                    should = true;
                    }
                }
            if( should )
                {
                stream << node->data_file_info.name << std::endl;
                num++;
                }
            } //if
        } // while
    return num;
    }


bool DataManager::test_file_exist
    (
    const std::string &         file_name,
    bool                        in_data_dir
    ) const
    {
    std::string  temp;
    if( in_data_dir )
        temp = data_directory + "/";
    VolumeManager::IFStream ifs( ( temp + file_name ).c_str(), std::ios_base::in | std::ios_base::binary );
    if( !ifs.is_open() )
        {
        return false;
        }
    ifs.close();
    return true;
    }


bool DataManager::test_files_exist
    (
    std::stringstream &         file_names,
    bool                        throw_exception
    )
    {
    TextReader      tr( file_names );

    while( tr.next_line() )
        {
        if( !test_file_exist( tr.str() ) )
            {
            if( throw_exception )
                {
                throw ExDataFileNotExist( "Data file " + tr.str() + " doesn't exist in data directory." );
                }
            return false;
            }
        }
    return true;
    }


void DataManager::release_data_object
    (
    DataObject *             data_object
    )
    {
    UsedDataObjectHash::iterator it = used_data_object_hash.find( data_object->data_object_name );
    if( it == used_data_object_hash.end() )
        {
        throw ExDataObjectNotRegistered( "Data object is not registered in data object manager." );
        }
    
    UsedDataObjectList & udolist = it->second;
    udolist.remove( data_object );
    if( udolist.empty() )
        {
        used_data_object_hash.erase( it );
        }
    }

void DataManager::update_data_file
    (
    
    const std::string &             data_file_name,
    bool                            should_save_to_disk
    )
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }
    if( phase != PHASE_UPDATE )
        {
        throw ExDataManagementError( "DataManager is not in update phase" );
        }
    hope( data_description != NULL );

    std::string     data_node_name = data_description->get_data_node_name( data_file_name );
    DataNode & data_node = data_description->data_node_tree.get_data_node( data_node_name );
    if( !data_file_list.find( data_file_name ) )
        {
        std::ostringstream  ss;
        ss << "Data file \"" << data_file_name << "\" doesn't exist in data file list.";
        throw ExDataManagementError( ss.str() );
        }
     
    DataFileInfo & dfi = data_file_list.get( data_file_name );
        
    hope( data_node.data_file_info == dfi );

    dfi.load_file_info( data_directory );
    dfi.version.increase();
    data_node.data_file_info = dfi;

    if( should_save_to_disk )
        {
        data_description->data_version.increase();
        save_data_description();
        save_data_file_list();
        }
    update_used_data_object_after_data_file_update( data_file_name );
    }


void DataManager::update_data_file_after_download
    (
    const std::string &             data_file_name
    )
    {
    hope( data_description != NULL );

    // Get data node name for this data file.
    std::string     data_node_name = data_description->get_data_node_name( data_file_name );

    // Test if data file version is the same as in data node which owns this data file.
    DataNode & data_node = data_description->data_node_tree.get_data_node( data_node_name );
    hope( data_node.data_file_info.name == data_file_name );

    data_node.data_file_info.load_file_info( data_directory );
    data_node.unusable = false;
    
    if( data_file_list.find( data_file_name ) )
        {
        data_file_list.get( data_file_name ) = data_node.data_file_info;
        }
    else
        {
        data_file_list.add( data_node.data_file_info );
        }
        
    save_data_file_list();
    }
    

void DataManager::save_data_file_list()
    {
    std::string     old_dfl( data_directory + "/" + DM_DATA_FILE_LIST_FILE_NAME + ".tmp" );
    std::string     dfl( data_directory + "/" + DM_DATA_FILE_LIST_FILE_NAME );
    remove( old_dfl.c_str() );
    if( rename( dfl.c_str(), old_dfl.c_str() ) != 0 )
        {
        std::ostringstream  ss;
        ss << "Cannot rename data file list \"" << dfl << "\" to \"" << old_dfl << "\".";
        throw ExDataManagementError( ss.str() );
        }        
    VolumeManager::OFStream   ofs( dfl.c_str(), std::ios_base::out );
    if( !ofs )
        {
        std::ostringstream  ss;
        ss << "Cannot create data file list \"" << dfl << "\".";
        throw ExDataManagementError( ss.str() );                
        }
    data_file_list.save( ofs );
    ofs.close();
    }


void DataManager::load_data_file_list()
    {
    std::string     dfl( data_directory + "/" + DM_DATA_FILE_LIST_FILE_NAME );
    VolumeManager::IFStream   ifs( dfl.c_str(), std::ios_base::in );
    if( !ifs )
        {
        std::ostringstream  ss;
        ss << "Cannot open data file list \"" << dfl << "\".";
        throw ExDataManagementError( ss.str() );                
        }
    data_file_list.load( ifs, data_directory );
    ifs.close();
    }


// **** System message handling methods ****

void DataManager::handle_data_version_message
    (
    const DataVersionMessage &    msg
    )
    {
    hope( type != MASTER_DATA_SERVICE );
    hope( dp.data_service_node_id == msg.get_initiator() );
    dp.fresh_download_info.valid = true;
    dp.fresh_download_info.data_version = msg.data_version;
    dp.fresh_download_info.magic_number = msg.magic_number;
    dp.fresh_download_info.node_id = msg.get_initiator();
    }


void DataManager::handle_data_message
    (
    const DataMessage &    msg
    )
    {
    hope( type != MASTER_DATA_SERVICE );
    hope( dp.data_service_node_id == msg.get_initiator() );

    if( msg.offset == 0 )
        {
        // First data message.
        dp.first_download_time = SystemTime::current_time();
        }

    // Check speed limit.
    SystemTime time = SystemTime::current_time();
    hope( time >= dp.first_download_time );
    Float64 link_speed = 0;
    if( time > dp.first_download_time )
        {
        link_speed = ( dp.offset + msg.buffer_size ) / ( time - dp.first_download_time ).get_value();
        }
    std::ostringstream  ss;
    ss << "DataManager: downloading file \"" << dp.actual_downloaded_file_name << "\" version "
       << dp.actual_downloaded_data_version.get_major() << "." 
       << dp.actual_downloaded_data_version.get_minor()
       << " from node " << dp.data_service_node_id << ". Actual offset = " << dp.offset
       << " Actual speed = " << ( int )link_speed << " B/s";
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );

    if( msg.offset != dp.offset )
        {
        std::ostringstream  ss;
        ss << "Warning: while downloading data, message offset differs from expected offset: msg.offset= " << msg.offset << " dp.offset= " << dp.offset
           << "\nFile download will be initialized again.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        // Simulate timeout - this will start downloading again.
        dp.data_stream.close();
        dp.status = dp.STATUS_INIT;
        return;
        }
    if( msg.data_file_name != dp.actual_downloaded_file_name )
        {
        std::ostringstream  ss;
        ss << "Warning: while downloading data, message data file name differs from expected"
           << " data_file_name: msg.data_file_name= " << msg.data_file_name
           << " dp.actual_downloaded_file_name= " << dp.actual_downloaded_file_name
           << "\nFile download will be initialized again.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        // Simulate timeout - this will start downloading again.
        dp.data_stream.close();
        dp.status = dp.STATUS_INIT;
        return;
        }
    dp.data_stream.write( reinterpret_cast< char * >( msg.buffer ), msg.buffer_size );
    assert_release( dp.data_stream.good(), "Write error." );
    dp.offset += msg.buffer_size;
    if( msg.end )
        {
        // Last data message. No confirmation.
        std::ostringstream  ss;
        ss << "DataManager: file \"" << dp.actual_downloaded_file_name << "\" version "
           << dp.actual_downloaded_data_version.get_major() << "." 
           << dp.actual_downloaded_data_version.get_minor()
           << " successfully downloaded from node " << dp.data_service_node_id << ". Size = " << dp.offset
           << " Overall download speed = " << ( int )link_speed << " B/s";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );
        dp.status = dp.STATUS_SUCCESS;
        }
    else
        {
        DataConfirmationMessage     dcm( dp.offset, dp.actual_downloaded_file_name );
        dcm.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );
        dp.last_request_time = SystemTime::current_time();
        }
    }


void DataManager::handle_data_version_request_message
    (
    const DataVersionRequestMessage &    msg
    )
    {
    // Only data server can handle data request message.
    hope( type == MASTER_DATA_SERVICE || type == SECONDARY_DATA_SERVICE );
    if( !ready_for_use )
        {        
        DataErrorMessage        dem( NOT_READY_ERROR );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    if( phase == PHASE_UPDATE )
        {
        DataErrorMessage        dem( DATA_SERVICE_UPDATE );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    if( !up.find_request( msg.get_initiator() ) )
        {
        up.add_request( msg.get_initiator() );
        }
    UploadInfo &    ui = up.get_request( msg.get_initiator() );
    ui.node_id = msg.get_initiator();
    ui.phase = up.PHASE_SEND_DATA_VERSION;
    }


void DataManager::handle_data_request_message
    (
    const DataRequestMessage &    msg
    )
    {
    // Only data server can handle data request message.
    hope( type == MASTER_DATA_SERVICE || type == SECONDARY_DATA_SERVICE );
    if( !ready_for_use )
        {
        DataErrorMessage        dem( NOT_READY_ERROR );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    if( phase == PHASE_UPDATE )
        {
        DataErrorMessage        dem( DATA_SERVICE_UPDATE );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    if( !up.find_request( msg.get_initiator() ) )
        {
        up.add_request( msg.get_initiator() );
        }
    UploadInfo &    ui = up.get_request( msg.get_initiator() );        

    ui.node_id = msg.get_initiator();
    ui.phase = up.PHASE_UPLOAD_DATA_FILE;
    hope( msg.data_file_name.size() );
    ui.data_file_name = msg.data_file_name;
    ui.data_version = msg.data_version;
    ui.offset = 0;
    ui.average_speed = msg.download_speed;    
    ui.buffer_size = msg.download_speed / DM_NUM_UPLOAD_PER_SECOND;
    if( msg.download_speed > data_manager_settings.max_download_speed )
        {
        ui.average_speed = data_manager_settings.max_download_speed;
        ui.buffer_size = data_manager_settings.max_download_speed / DM_NUM_UPLOAD_PER_SECOND;
        }
    ui.buffer_len = 0;
    ui.num_uploads = -1;  // Set to -1 to enable set ui.first_upload_time
    
    // Check if file is correct data file.
    hope( data_description != NULL );
    if( ui.data_file_name != DM_DD_FILE_NAME )
        {
        std::string     data_node_name;
        try
            {
            data_node_name = data_description->get_data_node_name( ui.data_file_name );
            }
        catch( ... )
            {
            // File doesn't exist in data file list.
            DataErrorMessage        dem( DATA_FILE_ERROR );
            dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
            ui.phase = up.PHASE_IDLE;
            return;
            }
        DataNode & dn = data_description->data_node_tree.get_data_node( data_node_name );
        if( ui.data_version != dn.data_file_info.version )
            {
            // Requested version of data file is not available.
            DataErrorMessage        dem( DATA_FILE_VERSION_ERROR );
            dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
            ui.phase = up.PHASE_IDLE;
            return;
            }
        }
    else
        {
        // Special case - data description file.
        if( ui.data_version != get_data_version() )
            {
            // Requested version of data description is not available.
            DataErrorMessage        dem( DATA_FILE_VERSION_ERROR );
            dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
            ui.phase = up.PHASE_IDLE;
            return;
            }
        }
     
    ui.data_stream.close();
    ui.data_stream.clear();
    ui.data_stream.open( ( data_directory + "/" + ui.data_file_name ).c_str(), std::ios_base::in | std::ios_base::binary );
    if( !ui.data_stream.is_open() )
        {
        std::ostringstream  ss;
        ss << "Cannot open data file \"" << data_directory << "/" << ui.data_file_name << "\".";
        throw ExDataManagementError( ss.str() );
        }                        
    // Get file size.
    ui.data_stream.seekg( 0, std::ios_base::end );
    hope( ui.data_stream.good() );

    ui.file_size = ui.data_stream.tellg();

    ui.data_stream.seekg( 0, std::ios_base::beg );
    hope( ui.data_stream.good() );

    ui.wait_for_confirmation = false;
    }


void DataManager::handle_data_confirmation_message
    (
    const DataConfirmationMessage &    msg
    )
    {
    hope( type == MASTER_DATA_SERVICE || type == SECONDARY_DATA_SERVICE );
    if( !ready_for_use )
        {
        DataErrorMessage        dem( NOT_READY_ERROR );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    if( phase == PHASE_UPDATE )
        {
        DataErrorMessage        dem( DATA_SERVICE_UPDATE );
        dem.send( msg.get_initiator(), NetworkCommon::STREAM_RELIABLE );
        return;
        }
    hope( up.find_request( msg.get_initiator() ) );
    UploadInfo &    ui = up.get_request( msg.get_initiator() );        
    // @todo: otestovat zda potvrzovaci paket byl reakci na posledne zaslana data
    if( msg.data_file_name != ui.data_file_name )
        {
        std::ostringstream  ss;
        ss << "Warning: while uploading data, message data_file_name differs from expected"
           << " data_file_name: msg.data_file_name= " << msg.data_file_name
           << " ui.data_file_name= " << ui.data_file_name
           << "\nFile download will be initialized again.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        // Simulate data file version error. Client will initialize download again.
        DataErrorMessage    dem( DATA_FILE_VERSION_ERROR );
        dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
        ui.phase = up.PHASE_IDLE;
        return;
        }
    if( msg.offset != ui.offset )
        {
        std::ostringstream  ss;
        ss << "Warning: while uploading data, message offset differs from expected"
           << " offset: msg.offset= " << msg.offset
           << " ui.offset= " << ui.offset
           << "\nFile download will be initialized again.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        // Simulate data file version error. Client will initialize download again.
        DataErrorMessage    dem( DATA_FILE_VERSION_ERROR );
        dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
        ui.phase = up.PHASE_IDLE;
        return;
        }
        
    std::ostringstream  ss;
    ss << "DataConfirmation message received from node " << msg.get_initiator() << " Offset = " << msg.offset;
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );
    ui.wait_for_confirmation = false;
    }


void DataManager::handle_data_error_message
    (
    const DataErrorMessage &    msg
    )
    {
    hope( type != MASTER_DATA_SERVICE );
    hope( dp.data_service_node_id == msg.get_initiator() );

    dp.error = msg.error_type;
    dp.status = dp.STATUS_FAILED;
    }


void DataManager::handle_data_update_request_message
    (
    const DataUpdateRequestMessage &   msg
    )
    {
    hope( type == MASTER_DATA_SERVICE );
    hope( msg.get_initiator().is_type( NodeId::TYPE_SERVER ) );
        
    switch_to_update_phase( DataManager::AUTOMATIC );
    }

bool DataManager::request_master_data_update()
    {
    hope( type != MASTER_DATA_SERVICE );
    if( type == DATA_CLIENT )
        {
        return false;
        }
    hope( type == DATA_SERVER );
    DataUpdateRequestMessage    msg;
    msg.send( dp.data_service_node_id, NetworkCommon::STREAM_RELIABLE );
    return true;
    }


void DataManager::switch_from_to_anonymous_mode()
    {
    if( type != DATA_CLIENT && type != DATA_SERVER )
        {
        return;
        }
    //hope( !( Global::get_local_node_id().is_type( NodeId::TYPE_ANONYMOUS ) ) );
    // Reactivate download request if there is some.
    if( phase == PHASE_DOWNLOAD && dp.phase == dp.PHASE_FILE_DOWNLOAD )
        {
        // Simulate timeout.
        dp.data_stream.close();
        dp.status = dp.STATUS_INIT;
        }
    }


bool DataManager::download_success()
    {
    if( phase == PHASE_IDLE && dp.status == dp.STATUS_SUCCESS )
        return true;
    return false;
    }


int DataManager::close_data_object_open_file
    (
    const std::string       file_name
    )
    {
    int num_open_files = 0;
    UsedDataObjectHash::iterator it2 = used_data_object_hash.begin();
    while( it2 != used_data_object_hash.end() )
        {
        UsedDataObjectList & udolist = it2->second;
        UsedDataObjectList::iterator it_list = udolist.begin();
        while( it_list != udolist.end() )
            {
            DataObject *    data_object = *it_list;
            if( data_object->data_file_full_name == file_name )
                {
                data_object->update();
                if( data_object->data.is_open() )
                    {
                    data_object->offset = data_object->data.tellg();
                    data_object->data.close();
                    num_open_files++;
                    }
                else
                    {
                    data_object->offset = -1; // Means that file was not opened.
                    }
                }
            ++it_list;
            }
        ++it2;
        }
    return num_open_files;
    }


void DataManager::restore_data_object_file
    (
    const std::string       old_file_name,
    const std::string       new_file_name
    )
    {
    UsedDataObjectHash::iterator it2 = used_data_object_hash.begin();
    while( it2 != used_data_object_hash.end() )
        {
        UsedDataObjectList & udolist = it2->second;
        UsedDataObjectList::iterator it_list = udolist.begin();
        while( it_list != udolist.end() )
            {
            DataObject *    data_object = *it_list;
            if( data_object->data_file_full_name == old_file_name )
                {
                data_object->data_file_full_name = new_file_name;
                if( data_object->offset != -1 )
                    {
                    data_object->open();
                    data_object->data.seekg( data_object->offset );
                    // Now @attribute data_object->data is restored to its previous state.
                    }
                }
            ++it_list;
            }
        ++it2;
        }
    }


UInt32 DataManager::generate_magic_number()
    {
    if( type == MASTER_DATA_SERVICE )
        {
        srand( static_cast<unsigned>( ::time( NULL ) ) );
        UInt32 temp = std::rand();
        temp <<= 16;
        temp |= std::rand(); 
        return temp;
        }
    return 0;
    }


void DataManager::switch_to_update_phase
    (
    UpdateMode              update_mode
    )
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }    
    switch( phase )
        {
        case PHASE_IDLE:
            {
            // Switch to update mode.
            hope( update.next_update == false );
            switch( update_mode )
                {
                case AUTOMATIC:
                    {
                    update.phase = update.PHASE_LOAD_AND_CHECK_UPDATE_FILE;
                    break;
                    }
                case MANUAL:
                    {
                    update.phase = update.PHASE_IDLE;
                    break;
                    }
                default:
                    panic( "Unexpected update mode." );
                }
            phase = PHASE_UPDATE;
            up.destroy( DATA_SERVICE_UPDATE ); // @todo - spravit tohle
            break;
            }
        case PHASE_UPDATE:
            {
            // Data manager is already in update phase. Let's go to update phase
            // again after this update phase.
            update.next_update = true;
            break;
            }
        default:
            panic( "Unsupported phase." );
        }
    }


void DataManager::finish_update_phase()
    {
    if( type != MASTER_DATA_SERVICE )
        {
        THROW_DATA_MANAGER_NOT_MASTER_DATA_SERVICE
        }
    hope( phase == PHASE_UPDATE );
    phase = PHASE_IDLE;
    }


const std::string & DataManager::get_data_directory()
    {
    return data_directory;
    }


void DataManager::updatep_set_test()
    {
    update.test = true;
    }


bool DataManager::update_finished()
    {
    return phase == PHASE_IDLE;
    }

// **** DataManager end ****
// **** DataObject start ****


DataObject::DataObject
    (
    DataManager *               _data_manager,
    const std::string &         _data_object_name,
    bool                        open_data_file
    ) : data(),
        data_manager( _data_manager ),        
        data_updated( true ),
        data_object_name( _data_object_name )
    {
    refresh( open_data_file );
    }

DataObject::~DataObject()
    {
    release();
    }

void DataObject::refresh
    (
    bool            open_data_file
    )
    {
    if( !data_updated )
        return;
    
    data_updated = false;

    DataDescription &   data_description = *( data_manager->data_description );
    node_path = data_description.data_object_list.get_data_object_node_path( data_object_name );
    
    // Log used node.
    std::ostringstream  ss;
    ss << "DataObject::refresh()\n"
          "data object name: \"" << data_object_name << "\"\n";

    DataNode *      node = &( data_description.data_node_tree.get_data_node( node_path ) );

    // Add to log.
    ss << "-----\nnode full_name: \"" << node->full_name << "\"\n"
            "node deprecated: \"" << node->deprecated << "\"\n"
            "node unusable: \"" << node->unusable << "\"\n"
            "node data_file_info.name: \"" << node->data_file_info.name << "\"\n";

    while( ( node->full_name != DM_DNT_ROOT_NAME ) 
                            && ( ( node->deprecated ) || ( node->unusable ) ) )
        {
        node_path = data_description.data_node_tree.get_data_node_father_name( node_path );
        node = &( data_description.data_node_tree.get_data_node( node_path ) );

        // Add to log.
        ss << "-----\nnode full_name: \"" << node->full_name << "\"\n"
              "node deprecated: \"" << node->deprecated << "\"\n"
              "node unusable: \"" << node->unusable << "\"\n"
              "node data_file_info.name: \"" << node->data_file_info.name << "\"\n";
        }

    if( node->full_name == DM_DNT_ROOT_NAME )
        {
        throw ExNoAvailableDataNode( "No available data node for data object " + data_object_name + "." );
        }
          
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );
    
    // Test data file.
    data_manager->data_file_list.get( node->data_file_info.name );    

    data_file_full_name = data_manager->data_directory + "/" + node->data_file_info.name;
    file_size = get_file_size( data_file_full_name );

    if( open_data_file )
        {
        open();
        }
    }


bool DataObject::is_data_updated() const
    {
    return data_updated;
    }


std::string DataObject::get_name() const
    {
    return data_object_name;
    }



void DataObject::update()
    {
    data_updated = true;
    }
    

void DataObject::release()
    {
    data.close();
    data_manager->release_data_object( this );
    }


FileSizeType DataObject::get_size() const 
    {
    return file_size;
    }


void DataObject::download_actual_version()
    {    
    DataDescription &   data_description = *( data_manager->data_description );
    std::string node_path = data_description.data_object_list.get_data_object_node_path( data_object_name );    
    DataNode &  node = data_description.data_node_tree.get_data_node( node_path );
    DataVersion dv = node.data_file_info.version; // Version from DD.
            
    if( data_manager->dp.actual_downloaded_file_name == node.data_file_info.name )
        {
        // We are downloading right now requested file.
        return;
        }

    bool    download = false; // Should we download new version?

    // Find if there is such a request to download.
    DataManager::StringQueue::iterator it = data_manager->dp.custom_data_file_download.begin();
    DataManager::StringQueue::iterator end = data_manager->dp.custom_data_file_download.end();
    while( it != end )
        {
        if( *it == node.data_file_info.name )
            {
            // Already in deque.
            return;
            }
        ++it;
        }

    if( !( data_manager->data_file_list.find( node.data_file_info.name ) ) )
        {
        download = true;
        }
    else
        {
        const DataFileInfo & dfi = data_manager->data_file_list.get( node.data_file_info.name );
        assert_release
            (
            dfi.version <= dv,
            std::string( "Data file " ) + node.data_file_info.name + " version error."
            );                    
        if( dfi.version < dv )
            {
            download = true;
            }
        }
    if( download )
        {
        std::ostringstream  ss;
        ss << "DataObject::download_actual_version():\n"
              "data object name: \"" << data_object_name << "\"\n"
              "file name for download: \"" << node.data_file_info.name << "\"";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );
        
        data_manager->dp.custom_data_file_download.push_back( node.data_file_info.name );
        }
    }


void DataObject::open()
    {
    data.close();
    data.clear();
    
    data.open( data_file_full_name.c_str(), std::ios_base::in | std::ios::binary );
    if( !data.is_open() )
        {
        throw ExCanNotOpenDataFile( "Cannot open data file \"" + data_file_full_name + "\"." );
        }    
    }


void DataObject::close()
    {
    data.close();
    }


// **** DataObject end ****
// **** DataNode start ****


DataNode::DataNode() :
    name( "" ),
    full_name( "" ),
    download_type( BLANK ),
    deprecated( false ),
    unusable( true ),
    data_file_info( "", DataVersion( 0, 0 ) )
    {
    }


DataNode::DataNode
    (
    const std::string &     _full_name, 
    const std::string &     _file_name, 
    DataFileDownloadType    _download_type,
    bool                    _deprecated,
    const DataVersion &     _file_version
    ) : 
    full_name( _full_name ),
    download_type( _download_type ),
    deprecated( _deprecated ),
    unusable( false ),
    data_file_info( _file_name, _file_version )
    {
    if( full_name != DM_DNT_ROOT_NAME )
        {
        name = DataNodeTree::get_data_node_name( _full_name );
        }
    else
        {
        name = "";
        }
    }


void DataNode::text_write
    (
    TextWriter &            tw
    ) const
    {
    tw.line() << full_name;
    tw.end_line();
    
    tw.line();
    Serializer::TextBool::write( tw, deprecated );
    tw.end_line();
    
    tw.line();
    Serializer::TextInt::write( tw, download_type, 32 );
    tw.end_line();
    
    data_file_info.text_write( tw );
    }


void DataNode::text_read
    (
    TextReader &            tr,
    bool                    light_weigth
    )
    {
    full_name = tr.str();
    
    bool res = tr.next_line();
    assert_release( res, "Read error." );
    Serializer::TextBool::read( tr, deprecated );
    res = tr.next_line();
    assert_release( res, "Read error." );
    int temp;
    Serializer::TextInt::read( tr, temp, 32 );
    download_type = static_cast<DataFileDownloadType>( temp );
    res = tr.next_line();
    assert_release( res, "Read error." );
    data_file_info.text_read( tr, light_weigth );

    name = DataNodeTree::get_data_node_name( full_name );
    }


// **** DataNode end ****
// **** DataNodeTree start ****

DataNodeTree::DataNodeTree()
    {
    root_node = new DataNode( DM_DNT_ROOT_NAME, "", BLANK, false, DataVersion( 0, 0 ) );
    }

DataNodeTree::~DataNodeTree()
    {
    DataNodeTree::DataNodeStack s;
    
    DataNodeTree::add_sons_to_stack( root_node, s );
    while( !s.empty() )
        {
        DataNode * node = s.top();
        s.pop();
        DataNodeTree::add_sons_to_stack( node, s );
        delete node;
        }
    }


std::string DataNodeTree::get_data_node_father_name
    ( 
    const std::string &         data_node_full_name
    )
    {
    std::string::size_type      pos = data_node_full_name.rfind( "/" );
    assert_release( pos != std::string::npos, "Parse error." );

    std::string                 node_path = data_node_full_name;
    node_path.erase( pos );   // Get data node path - erase last data node name.
    
    return node_path;
    }


std::string DataNodeTree::get_data_node_name
    ( 
    const std::string &         data_node_full_name
    )
    {
    std::string::size_type     pos = data_node_full_name.rfind( "/" );
    assert_release( pos != std::string::npos, "Parse error." );

    std::string::const_iterator  it = data_node_full_name.begin() + pos + 1;
    std::string                  node_name = get_next_name_token( data_node_full_name, it );
    return node_name;
    }


DataNode & DataNodeTree::get_data_node
    ( 
    const std::string &         data_node_full_name
    )
    {    
    DataNode *       akt_node = root_node;
    if( data_node_full_name.empty() )
        {
        // Special case - root node.
        return *akt_node;
        }

    std::string       token;
    std::string::const_iterator it_str = data_node_full_name.begin();
    while( ( it_str != data_node_full_name.end() ) && ( *it_str == '/' ) )
        {
        token = get_next_name_token( data_node_full_name, ++it_str );
        
        if( !token.empty() )
            {   
            // There is another token in the node_path.
            DataNode::NodeList::iterator    it_node = akt_node->node_list.find( token );
            if( it_node != akt_node->node_list.end() )
                {   
                // Found son of requested name.
                akt_node = it_node->second;
                }
            else
                {   
                // No son of requested name exist.
                std::ostringstream  ss;
                ss << "Data node \"" << data_node_full_name << "\" doesn't exist in data node tree.";
                throw ExDataNodeNotExist( ss.str() );
                }
            }
        }
    return *akt_node;
    }


std::string DataNodeTree::get_data_node_file_name
    ( 
    const std::string &         data_node_full_name
    )
    {
    DataNode &      data_node = get_data_node( data_node_full_name );
    return data_node.data_file_info.name;
    }




void DataNodeTree::add_sons_to_stack
    (
    const DataNode *                  father,
    DataNodeStack &                   s
    )
    {
    hope( father != NULL );
    DataNode::NodeList::const_iterator it = father->node_list.end();
    while( it != father->node_list.begin() )
        {
        --it;
        s.push( it->second );
        }
    }


// **** DataNodeTree end ****
// **** DataDescription start ****

DataDescription::DataDescription
    (
    std::istream &          stream,
    const std::string &     data_directory,
    bool                    light_weight
    ) :
    magic_number( 0 )
    {
    load_description( stream, data_directory, light_weight );
    validate_all_data_objects();
    }

void DataDescription::validate_all_data_objects()
    {
    DataObjectList::DataObjectInfoHash::const_iterator it = 
                                data_object_list.data_object_info_hash.begin();
    while( it != data_object_list.data_object_info_hash.end() )
        {        
        data_node_tree.get_data_node( ( it->second ).node_path ); // Throws exception if data node doesn't exist.
        ++it;
        }
    }

void DataDescription::validate_data_object
    (
    const std::string &     data_object_name
    )
    {
    DataObjectList::DataObjectInfoHash::const_iterator it = 
                        data_object_list.data_object_info_hash.find( data_object_name );
    if( it != data_object_list.data_object_info_hash.end() )
        {        
        data_node_tree.get_data_node( ( it->second ).node_path ); // Throws exception if data node doesn't exist.
        }
    else
        {
        throw ExDataObjectNotExist( "Data object " + data_object_name + " doesn't exist in data object list." );
        }
    }


void DataDescription::read_data_version
    (
    TextReader &                        tr,
    DataVersion &                       data_version,
    DataManager::MagicNumberType &      magic_number
    )
    {
    bool res = tr.next_line();
    assert_release( res, "Read error." );
    char c;
    tr.line() >> c;
    assert_release( c == '[', "Data version read error." );
    data_version.text_read( tr );
    tr.line() >> c;     assert_release( c == '.', "Data version read error." );
    tr.line() >> magic_number;
    tr.line() >> c;     assert_release( c == ']', "Data version read error." );
    }
    
void DataDescription::read_format_version
    (
    TextReader &                    tr,
    const std::string &             data_directory
    )
    {
    bool res = tr.next_line();
    assert_release( res, "Read error." );
    std::string str = tr.str();
    int pos = str.find( DM_DD_FORMAT_VERSION_STRING );
    if( pos != 0 )
        {
        // Data description does not support format version. */
        std::ostringstream  ss;
        ss << "Data description format version error. Data description file does not contain field "
           << "\""DM_DD_FORMAT_VERSION_STRING << "\"\nDelete data directory ("
           << data_directory << ") to update data.";
        throw ExDataManagementError( ss.str() );
        }
    std::string version = str.substr( sizeof( DM_DD_FORMAT_VERSION_STRING ) - 1 );
    if( version != DM_DD_FORMAT_VERSION )
        {
        // This version of data description is not supported any more.
        std::ostringstream  ss;
        ss << "Data description format version error. Data description file format version is "
           << version << ", actually supported version is " << DM_DD_FORMAT_VERSION << "."
           << "\nDelete data directory (" << data_directory << ") to update data.";
        throw ExDataManagementError( ss.str() );
        }
    }
    
    
void DataDescription::load_description
    (
    std::istream &          stream,
    const std::string &     data_directory,    
    bool                    light_weight
    )
    {
    TextReader      tr( stream );
    if( !light_weight )
        {
        // Read version and magic number.
        read_format_version( tr, data_directory );
        read_data_version( tr, data_version, magic_number );
        }

    DataManager::DataNodeTree::DataNodeStack                temp_stack;
    std::vector<DataNode *>                                 temp_vector;

    typedef std::pair< std::string, std::string >   StrPair;
    typedef std::vector< StrPair >                  VecStr;
    VecStr                                                  do_vector;
    
    int                                                     node_level = 0;
    
    while( tr.next_line() )
        {
        std::string tag;
        tr.line() >> tag;
        if( tag == DM_DD_NODE )
            {
            // New node definition.
            std::string father_name;
            if( temp_stack.size() )
                {
                // This data node has a father.
                DataNode * father = temp_stack.top();
                hope( father );
                father_name = father->full_name;
                }
            std::string my_name;
            tr.line() >> my_name;
            
            // Fill information. Other information (if available) will be filled later.
            DataNode *  data_node = new DataNode();
            data_node->name = my_name;
            data_node->full_name = father_name + "/" + my_name;            
            
            temp_stack.push( data_node );
            node_level++;
            } // node
        else if( tag == DM_DD_TYPE )
            {
            hope( temp_stack.size() );

            std::string type;
            tr.line() >> type;
            
            DataNode * dn = temp_stack.top();
            hope( dn );
            if( type == DM_DD_NECESSARY_CLIENT )
                dn->download_type = static_cast< DataManager::DataFileDownloadType>
                    ( dn->download_type + DataManager::NECESSARY_CLIENT_DATA_FILE );
            else if( type == DM_DD_NECESSARY_SERVER )
                dn->download_type = static_cast< DataManager::DataFileDownloadType>
                    ( dn->download_type + DataManager::NECESSARY_SERVER_DATA_FILE );
            else if( type == DM_DD_CLIENT )
                dn->download_type = static_cast< DataManager::DataFileDownloadType>
                    ( dn->download_type + DataManager::CLIENT_DATA_FILE );
            else if( type == DM_DD_SERVER )
                dn->download_type = static_cast< DataManager::DataFileDownloadType>
                    ( dn->download_type + DataManager::SERVER_DATA_FILE );
            else
                {
                std::ostringstream  ss;
                ss << "Data description: type tag error: unknown data node download type."
                   << " type= \"" << type << "\"" << " line= " << tr.line_number();
                throw ExDataManagementError( ss.str() );
                }
            } // type
        else if( tag == DM_DD_FILE )
            {
            hope( temp_stack.size() );

            DataNode * dn = temp_stack.top();
            hope( dn );
            dn->data_file_info.text_read( tr, light_weight );
            } // file
        else if( tag == DM_DD_OBSOLETE )
            {
            hope( temp_stack.size() );
            DataNode * dn = temp_stack.top();
            hope( dn );
            dn->deprecated = true;
            } // obsolete
        else if( tag == DM_DD_ENDNODE )
            {
            // Test if everything is properly filled for this node.
            hope( temp_stack.size() );
            if( node_level == 0 )
                {        
                std::ostringstream  ss;
                ss << "Data description structure error."
                   << " line= " << tr.line_number();
                throw ExDataManagementError( ss.str() );
                }
            DataNode * dn = temp_stack.top();
            hope( dn );
            if( dn->full_name.empty() || dn->name.empty() || dn->data_file_info.name.empty() )
                {
                std::ostringstream  ss;
                ss << "Data description: endnode tag error: data node attributes are not properly filled."
                   << " line= " << tr.line_number();
                throw ExDataManagementError( ss.str() );
                }
            if( !light_weight )
                {
                if( dn->data_file_info.version.get_major() < 1 || dn->data_file_info.size == 0 )
                    {
                    // Test mode?
                    if( dn->data_file_info.version.get_major() == 0 && dn->data_file_info.version.get_minor() != 99 )
                        {
                        std::ostringstream  ss;
                        ss << "Data description: endnode tag error: data node file information are not properly filled."
                        << " line= " << tr.line_number();
                        throw ExDataManagementError( ss.str() );
                        }
                    }
                }
            // Data node seems to be correctly filled.
            temp_stack.pop();
            temp_vector.push_back( dn );
            node_level--;
            } // endnode
        else if( tag == DM_DD_DATAOBJECT )
            {
            std::string _name;
            tr.line() >> _name;
            std::string _path;
            tr.line() >> _path;

            if( _path != "=" )
                {
                std::ostringstream  ss;
                ss << "Data description file read error. String \"=\" expected. "
                << "line= " << tr.line_number();
                throw ExDataManagementError( ss.str() );
                }            
            tr.line() >> _path;
            do_vector.push_back( StrPair( _name, _path ) );
            } // dataobject
        else
            {
            std::ostringstream  ss;
            ss << "Unsupported tag in data description. tag= \"" + tag + "\" line = "
               << tr.line_number();
            throw ExDataManagementError( ss.str() );
            }          
        } // while 
    // Now, every node is stored in temp_vector. Let's go to build the tree.
    std::vector<DataNode *>::reverse_iterator    it = temp_vector.rbegin();
    while( it != temp_vector.rend() )
        {
        add_data_node( *it ); // Dont increase data version.
        ++it;
        }
    // Data node tree is successfully built.
    
    VecStr::iterator it2 = do_vector.begin();
    while( it2 != do_vector.end() )
        {
        add_data_object( it2->first, it2->second );
        ++it2;
        }
    }


void DataDescription::save_description
    (
    std::ostream &     stream
    )
    {
    TextWriter          tw( stream );

    // Write data description format version.
    tw.line() << DM_DD_FORMAT_VERSION_STRING << DM_DD_FORMAT_VERSION;
    tw.end_line();

    // Write version.
    tw.line() << "[";
    data_version.text_write( tw );
    tw.line() << "." << magic_number;
    tw.line() << "]";
    tw.end_line();
    tw.end_line();

    DataNodeTree::DataNodeStack       s;
    data_node_tree.add_sons_to_stack( data_node_tree.root_node , s );
    while( !s.empty() )
        {
        const DataNode *      node = s.top();
        s.pop();
        write_node( tw, *node );
        }

    tw.end_line();
    DataObjectList::DataObjectInfoHash::iterator    it = data_object_list.data_object_info_hash.begin();
    while( it != data_object_list.data_object_info_hash.end() )
        { 
        DataObjectList::DataObjectInfo &    data_object_info = it->second;
        tw.line() << "dataobject " << data_object_info.data_object_name 
                  << " = " << data_object_info.node_path;
        tw.end_line();
        ++it;
        }
    }


void DataDescription::write_node
    (
    TextWriter &        tw,
    const DataNode &    node
    ) const
    {
    tw.line() << DM_DD_NODE << " " << node.name;
    tw.end_line();
    tw.change_indent( 4 );
    if( node.deprecated )
        {
        tw.line() << "obsolete";
        tw.end_line();
        }
    if( flags_on( node.download_type, NECESSARY_CLIENT_DATA_FILE ) )
        {
        tw.line() << DM_DD_TYPE << " " << DM_DD_NECESSARY_CLIENT;
        tw.end_line();        
        }
    if( flags_on( node.download_type, NECESSARY_SERVER_DATA_FILE ) )
        {
        tw.line() << DM_DD_TYPE << " " << DM_DD_NECESSARY_SERVER;
        tw.end_line();        
        }
    if( flags_on( node.download_type, CLIENT_DATA_FILE ) )
        {
        tw.line() << DM_DD_TYPE << " " << DM_DD_CLIENT;
        tw.end_line();        
        }
    if( flags_on( node.download_type, SERVER_DATA_FILE ) )
        {
        tw.line() << DM_DD_TYPE << " " << DM_DD_SERVER;
        tw.end_line();        
        }
    tw.line() << DM_DD_FILE << " ";
    node.data_file_info.text_write( tw );
    DataNode::NodeList::const_iterator it = node.node_list.begin();
    if( node.node_list.size() )
        {
        tw.end_line();
        }
    while( it != node.node_list.end() )
        {
        // Recursive call        
        write_node( tw, *(it->second) );
        ++it;
        }
    tw.change_indent( -4 );
    tw.line() << DM_DD_ENDNODE;
    tw.end_line();
    tw.end_line();
    }


void DataDescription::copy_stream_by_lines
    (
    std::ostream &          dest_stream,
    std::istream &          source_stream
    )
    {
    TextWriter      tw( dest_stream );
    TextReader      tr( source_stream );

    while( tr.next_line() )
        {
        tw.line() << tr.str();
        tw.end_line();
        }
    }


void DataDescription::get_initial_information_from_data_description
    (
    const std::string &     data_directory,
    std::istream &          stream,
    DataVersion &           data_version,
    UInt32 &                magic_number
    )
    {
    TextReader  tr( stream );
    read_format_version( tr, data_directory );
    read_data_version( tr, data_version, magic_number );
    }


void DataDescription::add_data_node
    (
    DataNode *              new_node
    )
    {
    if( !file_list.insert
            (
            DataFileNameList::value_type( new_node->data_file_info.name, new_node->full_name )
            ).second )
        {
        throw ExDataFileAlreadyUsed( "Data file " + new_node->data_file_info.name + " already used in another data node." );
        }
    std::string     node_path = data_node_tree.get_data_node_father_name( new_node->full_name );
    
    DataNode &      father = data_node_tree.get_data_node( node_path );
    hope( father.full_name == node_path || father.full_name == DM_DNT_ROOT_NAME );
    if( !father.node_list.insert
                            ( 
                            DataNode::NodeList::value_type
                                                    ( 
                                                    new_node->name,
                                                    new_node
                                                    )
                            ).second )
        {
        throw ExDataNodeAlreadyExist( "Data node " + new_node->full_name + " already exist in data node tree." );
        }
    }

void DataDescription::add_data_object
    (
    const std::string &         data_object_name,
    const std::string &         node_path
    )
    {
    DataNode & data_node = data_node_tree.get_data_node( node_path ); // Throws exception if data node doesn't exist.
    if( data_node.deprecated )
        {
        throw ExDeprecatedDataNode
                (
                "Data object " + data_object_name + 
                " cannot point to deprecated data node " + node_path + " ."
                );
        }
    DataObjectList::DataObjectInfoHash::iterator it = 
                          data_object_list.data_object_info_hash.find( data_object_name );
    if( it == data_object_list.data_object_info_hash.end() )
        {
        // Data object is not in hash table.
        DataObjectList::DataObjectInfo      data_object_info( data_object_name, node_path );
        data_object_list.data_object_info_hash.insert
            (
            DataObjectList::DataObjectInfoHash::value_type
                (
                data_object_name,
                data_object_info
                )
            );
        }
    else
        {
        // Data object is already in hash table - update path.
        ( it->second ).node_path = node_path;
        }
    }

void DataDescription::remove_data_object
    (
    const std::string &         data_object_name
    )
    {
    DataObjectList::DataObjectInfoHash::iterator    it = 
                         data_object_list.data_object_info_hash.find( data_object_name );
    hope( it != data_object_list.data_object_info_hash.end() );
    
    data_object_list.data_object_info_hash.erase( it );
    }


DataVersion DataDescription::get_data_version() const
    {
    return data_version;
    }


MagicNumberType DataDescription::get_magic_number() const
    {
    return magic_number;
    }


const std::string & DataDescription::get_data_node_name
    (
    const std::string &     data_file_name
    )
    {
    DataFileNameList::iterator it = file_list.find( data_file_name );
    if( it == file_list.end() )
        {
        std::ostringstream  ss;
        ss << "Data file \"" << data_file_name << "\" is not in data file list.";
        throw ExDataManagementError( ss.str() );        
        }
    return it->second;
    }


// **** DataDescription end ****
// **** DataObjectList start ****

DataObjectList::DataObjectInfo::DataObjectInfo
    (
    std::string     _data_object_name,
    std::string     _node_path
    ) : data_object_name( _data_object_name ),
        node_path( _node_path )
    {
    }

std::string DataObjectList::get_data_object_node_path
    (
    const std::string &         data_object_name
    )
    {
    DataObjectInfoHash::iterator it = data_object_info_hash.find( data_object_name );
    if( it == data_object_info_hash.end() )
        {
        throw ExDataObjectNotExist( "Data object " + data_object_name + " doesn't exist in data object list." );
        }

    return ( it->second ).node_path;
    }


void DataObjectList::add
    (
    const std::string &         data_object_name,
    const std::string &         node_path
    )
    {
    DataObjectInfoHash::iterator it = data_object_info_hash.find( data_object_name );
    if( it != data_object_info_hash.end() )
        {
        DataObjectInfo & dfi = it->second;
        dfi.node_path = node_path;
        return;
        }

    // Data object doesn't exist yet.
    DataObjectInfo  doi( data_object_name, node_path );
    std::pair< DataObjectInfoHash::iterator, bool > result;
    result = data_object_info_hash.insert
        (
        DataObjectInfoHash::value_type( data_object_name, doi )
        );
    if( !result.second )
        {
        std::ostringstream  ss;
        ss << "Cannot insert data object name \"" << data_object_name << "\" into used data object hash table.";
        throw ExDataManagementError( ss.str() );
        }
    }

// **** DataObjectList end ****
// **** DownloadProvider start ****


DownloadProvider::DownloadProvider():
    phase( PHASE_IDLE ),
    phase_before( PHASE_IDLE ),    
    data_description( NULL ),
    offset( 0 ),
    magic_number( 0 ),
    download_type( DataManager::NECESSARY_DOWNLOAD ),
    status( STATUS_IDLE )
    {
    }


DownloadProvider::~DownloadProvider()
    {
    if( data_description )
        {
        // During minimal data files download, otherwise is always set to NULL.
        delete data_description;
        }
    }


DownloadProvider::DownloadInfo::DownloadInfo() :
    valid( false )
    {
    }


void DownloadProvider::destroy()
    {
    if( data_description )
        {
        delete data_description;
        data_description = NULL;
        }
    data_file_names.str( "" );
    data_file_names.clear();
    if( data_stream.is_open() )
        {
        data_stream.close();
        }
    }


// **** DownloadProvider end ****
// **** UploadProvider start ****

void UploadProvider::add_request
    (
    const NodeId &              node_id
    )
    {
    UploadInfoHash::iterator it = upload_info_hash.find( node_id );
    if( it != upload_info_hash.end() )
        {
        UploadInfo & ui = it->second;
        if( ui.deleted )
            {
            ui.deleted = false; // Reuse this structure.
            }
        return;
        }
    // No entry for node id.
    UploadInfo  upload_info;
    if( !upload_info_hash.insert
            (
            UploadInfoHash::value_type( node_id, upload_info )
            ).second
      )
        {
        std::ostringstream  ss;
        ss << "Cannot add upload request for NodeId = " << node_id << " to upload info hash table.";
        throw ExDataManagementError( ss.str() );
        }
    }


bool UploadProvider::find_request
    (
    const NodeId &              node_id
    )
    {
    UploadInfoHash::iterator it = upload_info_hash.find( node_id );
    if( it != upload_info_hash.end() )
        {
        const UploadInfo & ui = it->second;
        if( ui.deleted )
            {
            return false;
            }
        return true;
        }
    return false;
    }


void UploadProvider::remove_request
    (
    const NodeId &              node_id
    )
    {
    // FIXME: zatim to odregistrovava pouze anonymous ucty
    if( node_id.is_type( NodeId::TYPE_ANONYMOUS ) )
        {
        UploadInfoHash::iterator it = upload_info_hash.find( node_id );
        if( it != upload_info_hash.end() )
            {
            UploadProvider::UploadInfo & ui = it->second;
            ui.deleted = true;
            }
        }
    }


void UploadProvider::clear_deleted_requests()
    {
    // Must not be called when iterating requests.
    UploadInfoHash::iterator it = upload_info_hash.begin();
    while( it != upload_info_hash.end() )
        {
        const UploadInfo & ui = it->second;
        if( ui.deleted )
            {
            upload_info_hash.erase( it );
            it = upload_info_hash.begin();
            }
        else
            {
            ++it;
            }
        }
    }


UploadInfo & UploadProvider::get_request
    (
    const NodeId &              node_id
    )
    {
    UploadInfoHash::iterator it = upload_info_hash.find( node_id );
    if( it == upload_info_hash.end() )
        {
        std::ostringstream  ss;
        ss << "Upload request for NodeId = " << node_id << " doesn't exist in upload info hash table.";
        throw ExDataManagementError( ss.str() );
        }    
    if( ( it->second ).deleted )
        {
        std::ostringstream  ss;
        ss << "Upload request for NodeId = " << node_id << " exist in upload info hash table, but is marked as deleted.";
        throw ExDataManagementError( ss.str() );
        }
    return it->second;
    }


void UploadProvider::destroy
    (
    DataManager::DataMessageErrorType      data_message_error_type
    )
    {
    UploadInfoHash::iterator it = upload_info_hash.begin();
    it = upload_info_hash.begin();
    while( it != upload_info_hash.end() )
        {
        UploadInfo & ui = it->second;
        if( !ui.deleted && ui.phase != PHASE_IDLE )
            {
            if( ui.data_stream.is_open() )
                {
                ui.data_stream.close();
                ui.data_stream.clear();
                }
            DataErrorMessage    dem( data_message_error_type );
            dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
            ui.phase = PHASE_IDLE;
            }
        ++it;
        }
    if( data_message_error_type == NOT_READY_ERROR )
        {
        // Method is called when destroying data manager.
        upload_info_hash.clear();
        }
    }

void UploadProvider::close_open_file
    ( 
    const std::string &         file_name
    )
    {
    UploadInfoHash::iterator it = upload_info_hash.begin();
    while( it != upload_info_hash.end() )
        {
        UploadInfo & ui = it->second;
        if( ( !ui.deleted ) && ( ui.phase != PHASE_IDLE ) && ( ui.data_file_name == file_name ) )
            {
            if( ui.data_stream.is_open() )
                {
                ui.data_stream.close();
                ui.data_stream.clear();
                }
            DataErrorMessage    dem( DATA_SERVICE_UPDATE );
            dem.send( ui.node_id, NetworkCommon::STREAM_RELIABLE );
            }
        ++it;
        }
    }


UploadProvider::UploadInfo::UploadInfo() :
    deleted( false ),
    node_id( NodeId() ),
    phase( PHASE_IDLE ),
    data_file_name( "" ),
    data_version( DataVersion( 0, 0 ) ),
    offset( 0 ),
    buffer_size( 0 ),
    file_size( 0 ),
    wait_for_confirmation( false ),
    last_upload_time( 0 ),
    current_time( 0 ),
    first_upload_time( 0 ),
    average_speed( 0 ),
    num_uploads( 0 )
    {
    }


UploadProvider::UploadInfo::UploadInfo
    (
    const UploadInfo & ui
    ) :
    deleted( ui.deleted ),
    node_id( ui.node_id ),
    phase( ui.phase ),
    data_file_name( ui.data_file_name ),
    data_version( ui.data_version ),
    offset( ui.offset ),
    buffer_size( ui.buffer_size ),
    file_size( ui.file_size ),
    wait_for_confirmation( ui.wait_for_confirmation ),
    last_upload_time( ui.last_upload_time ),
    current_time( ui.current_time ),
    first_upload_time( ui.first_upload_time ),
    average_speed( ui.average_speed ),
    num_uploads( ui.num_uploads )
    {
    }

// **** DownloadProvider end ****
// **** DataVersion start ****


DataVersion::DataVersion() : 
    data_version( 0 )
    {
    }


DataVersion::DataVersion
    (
    int         major_version,
    int         minor_version
    ) :
    data_version( ( major_version << 8 ) | ( minor_version ) )
    {
    }


void DataVersion::text_write
    (
    TextWriter &            tw
    ) const
    {
    tw.line() << get_major() << "." << get_minor();
    }


void DataVersion::text_read
    (
    TextReader &            tr
    )
    {
    int major, minor;
    tr.line() >> major;
    char c;
    tr.line() >> c;
    assert_release( c == '.', "Read error." );
    tr.line() >> minor;
    data_version = ( major << 8 ) | ( minor );
    }


void DataVersion::bitstr_read
    (
    BitIStream &            bis
    )
    {
    Serializer::BitStrInt::read( bis, data_version, 32 );
    }


void DataVersion::bitstr_write
    (
    BitOStream &            bos
    ) const
    {
    Serializer::BitStrInt::write( bos, data_version, 32 );
    }


bool DataVersion::operator==
    (
    const DataVersion & v
    ) const
    {
    return data_version == v.data_version;
    }


bool DataVersion::operator!=
    (
    const DataVersion & v
    ) const
    {
    return data_version != v.data_version;
    }


bool DataVersion::operator<
    (
    const DataVersion & v
    ) const
    {
    return data_version < v.data_version;
    }


bool DataVersion::operator>
    (
    const DataVersion & v
    ) const
    {
    return data_version > v.data_version;
    }


bool DataVersion::operator<=
    (
    const DataVersion & v
    ) const
    {
    return data_version <= v.data_version;
    }


bool DataVersion::operator>=
    (
    const DataVersion & v
    ) const
    {
    return data_version >= v.data_version;
    }


void DataVersion::increase()
    {
    if( ( data_version & 0xff ) < DO_MAX_LOW_VERSION_NUMBER )
        {
        data_version++;
        }    
    else
        {
        data_version &= 0xffffff00;
        data_version += 0x100;
        }
    }


int DataVersion::get_major() const
    {
    return data_version >> 8;
    }


int DataVersion::get_minor() const
    {
    return data_version & 0xff;
    }

// **** DataVersion end ****
// **** DataFileInfo start ***


DataFileInfo::DataFileInfo()
    {
    }

DataFileInfo::DataFileInfo
    (
    const std::string &         _name,
    const DataVersion &         _version
    ) :
    name( _name ),
    version( _version ),
    size( 0 )
    {
    }

bool DataFileInfo::operator==
    (
    const DataFileInfo & dfi
    ) const
    {
    bool    result = true;
    if( name != dfi.name ) result = false;
    if( version != dfi.version ) result = false;
    if( size  != dfi.size ) result = false;
    return result;
    }

void DataFileInfo::load_file_info
    (
    const std::string &         data_directory
    )
    {
    size = get_file_size( data_directory + "/" + name );
    // @todo: load crc and creation time of file.
    }


void DataFileInfo::text_write
    (
    TextWriter &            tw
    ) const
    {
    tw.line() << "\"" << name << "\" ";
    version.text_write( tw );
    tw.line() << " " << size;
    tw.end_line();
    }


void DataFileInfo::text_read
    (
    TextReader &            tr,
    bool                    light_weigth
    )
    {
    name = tr.read_quote_string();
    if( !light_weigth )
        {
        version.text_read( tr );
        size = tr.read_int();
        }
    
    /*std::string                     str = tr.str();
    std::string::const_iterator     it = str.begin();
    name = get_next_name_token( str, it, '|' );
    std::istringstream  ss( get_next_name_token( str, ++it, '|' ) );
    TextReader      tr2( ss );
    bool res = tr2.next_line();
    assert_release( res, "Read error." );
    version.text_read( tr2 );
    // @todo: temporary hack.
    std::istringstream  ss2( get_next_name_token( str, ++it, '|' ) );
    int temp;
    ss2 >> temp;
    size = temp;*/
    }


// **** DataFileInfo end ***
// **** DataFileList start ****


void DataFileList::add
    (
    const DataFileInfo &        data_file_info
    )
    {
    std::pair< DataFileListMap::iterator, bool > result;
    result = data_file_list_map.insert
                (
                DataFileListMap::value_type( data_file_info.name, data_file_info )
                );
    if( !result.second )
        {
        std::ostringstream  ss;
        ss << "Cannot add data file \"" << data_file_info.name << "\" to data file list.";
        throw ExDataManagementError( ss.str() );
        }
    }


bool DataFileList::find
    (
    const std::string &         data_file_name
    ) const
    {
    DataFileListMap::const_iterator it = data_file_list_map.find( data_file_name );
    if( it == data_file_list_map.end() )
        {
        return false;
        }
    return true;
    }


DataFileInfo & DataFileList::get
    (
    const std::string &         data_file_name
    )
    {
    DataFileListMap::iterator it = data_file_list_map.find( data_file_name );
    if( it == data_file_list_map.end() )
        {
        std::ostringstream  ss;
        ss << "Data file \"" << data_file_name << "\" is not in data file list.";
        throw ExDataManagementError( ss.str() );
        }
    return it->second;
    }


void DataFileList::save
    (
    std::ostream &          stream            
    ) const
    {
    TextWriter          tw( stream );

    DataFileListMap::const_iterator it = data_file_list_map.begin();
    while( it != data_file_list_map.end() )
        {
        const DataFileInfo & dfi = it->second;
        dfi.text_write( tw );
        ++it;
        }    
    }


void DataFileList::load
    (
    std::istream &          stream,
    const std::string &     data_directory
    )
    {
    TextReader      tr( stream );
    while( tr.next_line() )
        {
        DataFileInfo    dfi;
        dfi.text_read( tr, false );
        dfi.validate( data_directory );
        add( dfi );
        }
    }


void DataFileList::refresh
    (
    DataDescription &       data_description,
    DataDescription &       new_data_description,
    const std::string &     data_directory,
    DataManager &           dm
    )
    {
    DataDescription::DataFileNameList::iterator it = data_description.file_list.begin();
    while( it != data_description.file_list.end() )
        {
        std::string node_name = data_description.get_data_node_name( it->first );
        try
            {
            std::string new_node_name = new_data_description.get_data_node_name( it->first );
            if( node_name != new_node_name )
                {
                // Node name differs - delete data file name from file list.
                dm.data_file_list.data_file_list_map.erase( it->first );
                }
            }
        catch( ... )
            {
            // Error. Delete data file name from file list.
            dm.data_file_list.data_file_list_map.erase( it->first );
            }        
        ++it;
        }
    
    // Delete data file names which are not present in data description.
    /*DataFileListMap::iterator it = data_file_list_map.begin();
    
    while( it != data_file_list_map.end() )
        {
        const std::string file_name = it->first;
        if( data_description.file_list.find( file_name ) == data_description.file_list.end() )
            {
            // Data file was removed from data description.
            data_file_list_map.erase( it );
            }
        else
            {
            ++it;
            }
        }*/
    
    /*data_file_list_map.clear();
    
    DataDescription::DataFileNameList::const_iterator it = data_description.file_list.begin();
    DataDescription::DataFileNameList::const_iterator end = data_description.file_list.end();
    while( it != end )
        {
        const std::string file_name = it->first;
        const std::string node_name = data_description.get_data_node_name( file_name );
        const DataNode & dn = data_description.data_node_tree.get_data_node( node_name );
        hope( file_name == dn.data_file_info.name );
        DataFileInfo dfi( dn.data_file_info.name, dn.data_file_info.version );
        if( dm.test_file_exist( dn.data_file_info.name ) )
            {
            dfi.load_file_info( data_directory );
            add( dfi );
            }
        ++it;
        }*/
    }


void DataFileList::destroy()
    {
    data_file_list_map.clear();
    }


void DataFileInfo::validate
    (
    const std::string &     data_directory
    ) const
    {
    if( size != get_file_size( data_directory + "/" + name ) )
        {
        std::ostringstream  ss;
        ss << "Size of data file \"" << name << "\" differs.";
        throw ExDataManagementError( ss.str() );
        }
    }

// **** DataFileList end ****
// **** DataManagerSettings start ****

DataManager::DataManagerSettings::DataManagerSettings() :
    Settings( "DataManager" ),
    download_type( access_integer( DM_SETTINGS_DOWNLOAD_TYPE, static_cast< int >( NECESSARY_DOWNLOAD ) ) ),
    data_service_node( access_string( DM_SETTINGS_DATA_SERVICE_NODE, "[ Service 1 ]" ) ),
    max_download_speed( access_integer( DM_SETTINGS_MAX_DOWNLOAD_SPEED, MAX_DOWNLOAD_SPEED ) ),
    check_data_version_interval( access_integer( DM_SETTINGS_CHECK_DATA_VERSION_INTERVAL, CHECK_DATA_VERSION_INTERVAL ) ),
    mount_data_volume( access_boolean( DM_SETTINGS_MOUNT_DATA_VOLUME, false ) ),
    data_volume_image_name( access_string( DM_SETTINGS_DATA_VOLUME_IMAGE_NAME, DM_DATA_VOLUME_IMAGE_NAME ) )
    {
    }

// **** DataManagerSettings end ****
// **** UpdateProvider start ****

UpdateProvider::UpdateProvider()
    {
    initialize();
    }

void UpdateProvider::initialize()
    {
    phase = PHASE_IDLE;
    copy_file_list.clear();
    // @todo: inicializovat cp
    test = false;
    next_update = false;
    }
    
    
bool UpdateProvider::load_and_check
    (
    const DataManager &     data_manager
    )
    {
    hope( phase == PHASE_LOAD_AND_CHECK_UPDATE_FILE );
    
    try
        {      
        ud.reset( data_manager.load_data_description( true, data_manager.source_directory + "/" + DM_UPDATE_FILE_NAME ) );

        // Test data node tree.
        DataNodeTree::DataNodeStack s;

        delete_file_list.clear();
        DataNodeTree::add_sons_to_stack( data_manager.data_description->data_node_tree.root_node, s );
        while( !s.empty() )
            {
            const DataNode * node = s.top();
            hope( node != NULL );
            s.pop();
            DataNodeTree::add_sons_to_stack( node, s );

            const DataNode & data_node = ud->data_node_tree.get_data_node( node->full_name );
            // Node exist. Test contents.
            if( data_node.download_type != node->download_type )
                {
                throw ExDataManagementError( data_node.full_name + ":wrong download type." );
                }
            if( data_node.data_file_info.name != node->data_file_info.name )
                {
                //throw ExDataManagementError( data_node.full_name + ":wrong data file name." );
                
                // Add file name to delete file list.
                delete_file_list.push_back( node->data_file_info.name );
                }
            // test data file in source directory.
            if( !data_manager.test_file_exist( data_manager.source_directory + "/" + data_node.data_file_info.name, false ) )
                {
                throw ExDataManagementError( data_node.full_name + ":data file does not exist in source directory." );
                }
            } // while

        // Test data objects.
        DataObjectList::DataObjectInfoHash::iterator it = 
            data_manager.data_description->data_object_list.data_object_info_hash.begin();
        DataObjectList::DataObjectInfoHash::iterator end =
            data_manager.data_description->data_object_list.data_object_info_hash.end();
        DataObjectList::DataObjectInfoHash::iterator ud_end =
            ud->data_object_list.data_object_info_hash.end();
        while( it != end )
            {
            if( ud->data_object_list.data_object_info_hash.find( it->first ) == ud_end )
                {
                // Error, some data object are missing.
                throw ExDataManagementError( "data object \"" + it->first + "\"  is missing in new data description" );
                }
            ++it;
            }            
        } // try
    catch( std::exception & e )
        {
        std::ostringstream  ss;
        ss << "Update description is not compatible with actual data description. Error message: "
        << e.what() << "\nDelete data directory to update data.";
        Global::log
            (
            STATUS_ERROR( Status::FACILITY_DATA_MANAGEMENT ),
            ss.str()
            );
        return false;
        }        
    return true;
    }
    
    
#if 0

// FIXME: tohle se nepouziva, pokud by bylo treba pak v anonymous namespace

/**
 * Convert windows-like-slashes in @param str to unix-like-slashes.
 */
void to_unix_slash
    (
    std::string &   str
    )
    {
    std::string::iterator   it = str.begin();
    while( it != str.end() )
        {
        if( *it == '\\' )
            {
            *it = '/';
            }
        ++it;
        }
    }


/**
 * Test if exist file @params directory/data_file_name.
 * If @param new_data_file is true, it test whether it is correct to
 * add new data node with this data file.
 * If @param new_data_file is false it just tests file exitence.
 */
void test_data_file
    (
    const std::string &     directory,
    const std::string &     data_file_name,
    const DataManager &     data_manager,
    bool                    new_data_file
    )
    {
    if( new_data_file && data_manager.data_file_list.find( data_file_name ) )
        {
        // Data file already exist.
        std::ostringstream  ss;
        ss << "Update error: data file \"" << data_file_name << "\" already exists.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );                    
        throw ExDataManagementError( ss.str() );
        }
    // Test if file really exist in directory with new versions of data files.
    std::string  file_name = directory + "/" + data_file_name;
    if( !data_manager.test_file_exist( file_name, false ) )
        {
        std::ostringstream  ss;
        ss << "Cannot open data file \"" << file_name << "\" for update.";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );                    
        throw ExDataManagementError( ss.str() );
        }
    }


/**
 * Parse @param line. Format of this string is ABC=XYZ.
 * Function stores the first part in @param part1 <- ABC,
 * and the second part is stored in @param part2 <- XYZ.
 */
void parse_equality_line
    (
    const std::string &     line,
    std::string &           part1,
    std::string &           part2
    )
    {
    std::string::size_type     pos = line.find( "=" );
    assert_release( pos != std::string::npos, "Read error." );
    
    part1.assign( line, 0, pos );
    part2.assign( line, pos + 1, std::string::npos );
    }

#endif


// **** UpdateProvider end ****

} // namespace Massiv::Core
} // namespace Massiv
