/** @file data_manager.h
 * DataManager declaration.
 */

/* 
 * begin        : Tue Nov 20 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_DATA_MANAGER_H
#define MASSIV_DATA_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_COPY_FILE_H
#include "data/copy_file.h"
#endif

#ifndef MASSIV_COMPARE_FILE_H
#include "data/compare_file.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_VOLUME_MANAGER_H
#include "filesys/volume_manager.h"
#endif

#include <string>
#include <map>
#include <vector>
#include <stack>
#include <list>
#include <sstream>
#include <queue>

namespace Massiv {
namespace Core {


/** @addtogroup data_manager
 * @{
 */

class DataObject; // Forward class declaration.

#define DM_DD_FILE_NAME                         ".description"
    /**< Data description file name. */
#define DM_DATA_FILE_LIST_FILE_NAME             ".filelist"
    /**< Data file list file name. */
#define DM_UPDATE_FILE_NAME                     ".update_description"
    /**< Name of file with data update information. */

#if 0

#define DM_UPDATE_CFG_DATA_FILE_UPDATE          "[DATA FILE UPDATE]"
    /**< Name of section in data update information. */
#define DM_UPDATE_CFG_ADD_DATA_NODES            "[ADD DATA NODES]"
    /**< Name of section in data update information. */
#define DM_UPDATE_CFG_ADD_DATA_OBJECTS          "[ADD DATA OBJECTS]"
    /**< Name of section in data update information. */
#define DM_UPDATE_CFG_SETUP                     "[SETUP]"
    /**< Name of section in data update information. */
#define DM_UPDATE_CFG_SOURCE_DATA_DIRECTORY     "SOURCE_DATA_DIRECTORY"
    /**< Name of subsection in data update information. */
#define DM_UPDATE_CFG_SETTINGS_DIRECTORY        "SETTINGS_DIRECTORY"
    /**< Name of subsection in data update information. */
#define DM_UPDATE_CFG_DATA_VERSION_UPDATE       "DATA_VERSION_UPDATE"
    /**< Name of subsection in data update information. */

#endif

// DataManager settings strings. Used in configuration file massiv.setting.conf
#define DM_SETTINGS_DOWNLOAD_TYPE               "download_type"
#define DM_SETTINGS_DATA_SERVICE_NODE           "data_service_node"
#define DM_SETTINGS_MAX_DOWNLOAD_SPEED          "max_download_speed"
#define DM_SETTINGS_CHECK_DATA_VERSION_INTERVAL "check_data_version_interval"
#define DM_SETTINGS_MOUNT_DATA_VOLUME           "mount_data_volume"
#define DM_SETTINGS_DATA_VOLUME_IMAGE_NAME      "data_volume_image_name"
    

#define DM_DATA_VOLUME_NAME                     "data_volume"
    /**< Data volume name. */
#define DM_DATA_VOLUME_IMAGE_NAME               "data.image"
    /**< Data volume image name. */
    
/**
 * Manage data objects.
 * Consists of two main parts - data description and used data object hash table.
 *     Data description - consists of two parts - data node tree and data object list.
 *     ----------------
 *     Data description describes the whole data system for data manager.
 *     Data object manager has @attribute data_description which points to DataDescription 
 *     structure.
 *
 *         Data node tree
 *         --------------
 *         Is a tree which consists of data nodes. Data node is a structure which
 *         represent one "data unit" ( actually one file in data directory, but it may be changed 
 *         in future ). If one data node is an ancestor of another data node in a tree means that 
 *         first data node can be used instead second one if data file for second data node
 *         is not available ( for example is not downloaded yet ). Path in data node tree
 *         is described like unix directory paths.
 *         Example: /graphic/figure/man/warrior
 *                  if data file for data node "warrior" is not available, then data file for
 *                  "man" or "figure" or "graphic" can be used instead.
 *
 *         Data object list
 *         ----------------
 *         Is a list of data objects ( reprezented by string ), which can 
 *         data manager offer by @method get_data_object().
 *         Each data object has its own data node path, which points
 *         to some data node in a data node tree ( usually it points to some leaf node, 
 *         but it is not necessary ).
 * 
 *     Used data object hash table
 *     ---------------------------
 *     Each instance of data object returned by @method get_data_object() is registered
 *     in so called "used data object hash table", where are stored pointers to all
 *     created data objects. The purpose is to be able to actualize data object when its
 *     data node path is actualised or when data file for data node is downloaded and 
 *     can be used instead data file of some ancestor.
 *
 * Behavior of data object manager - when its type is "master data service":
 *     1) in initialization phase tries to load data description from file in data directory.
 *     2) builds data description ( data node tree and data object list ) and validate it.
 *     3) Now is capable to offer data objects.
 *     4) Every next call of @method next_tick() is in idle phase and checks if 
 *        data version of its data description is locally changed - when positive,
 *        it sends system message to other massiv nodes that new data description is 
 *        available for download.
 *     5) In @method next_tick() is called @method up_next_tick() which services
 *        upload requests.
 * Behavior of data object manager - when its type is "secondary data service","data client"
 *     or "data_server": 
 *     1) in initialization phase tries to load data description from file in data directory.
 *     2) loads data version from data description and sends system message to get
 *        actual data version in simulation. Master data service or secondary data service 
 *        sends system message with data version to local node
 *     3) checks if actual simulation data version is newer then local one. If positive,
 *        it sets download phase, new data description and data files of data nodes with flag 
 *        NECESSARY_DATA_FILE are downloaded.
 *     4) builds data description ( data node tree and data object list ) and validate it.
 *     5) Now is capable to offer data objects.
 *     4) Every next call of @method next_tick() is in idle phase and checks if 
 *        data version of data description in simulation is changed - when positive,
 *        it sets download phase.
 *     5) In @method next_tick() is called @method up_next_tick() which services
 *        upload requests when "secondary data service". 
 *        "Data client" and "data server" cannot upload data.
 *
 * @code - get data object
 * static DataManager &   data_manager = Global::data_manager();
 * DataObject *           warrior               = data_manager.get_data_object( "warrior" );
 * DataObject *           texture_warrior       = data_manager.get_data_object( "texture_warrior" );
 * DataObject *           sound_warrior_default = data_manager.get_data_object( "sound_warrior_default" );
 * @endcode
 *
 * See in test/test_dataobject.cpp for more details of use.
 */
class MASSIV_EXPORTED DataManager
    {

public:

    friend class DataObject; // Data object can access data object manager which owns it.
    friend void test_data_file
        (
        const std::string &     directory,
        const std::string &     data_file_name,
        const DataManager &     data_manager,
        bool                    new_data_file
        );

    typedef int                             FileSizeType;
    typedef UInt32                          MagicNumberType;
    typedef SInt< UInt32,32 >               SMagicNumberType; // used in data_msg.h
    
    typedef std::list<DataObject *>         UsedDataObjectList;
    typedef ext::hash_map
                        <
                        std::string,
                        UsedDataObjectList
                        > UsedDataObjectHash;
    
    typedef std::deque<std::string>         StringQueue;
    
    /**
     * Type of data manager.
     */
    enum Type
        {        
        MASTER_DATA_SERVICE     = 0,
            /**< Master data service. Cannot download data. 
                 Other massiv servers and clients upload data from master data service.
                 Master data service can be only one in whole simulation. */
        SECONDARY_DATA_SERVICE  = 1,
            /**< Secondary data service. Download data from the master data service
                 and it behaves like master data service for other massiv nodes. */
        DATA_SERVER             = 2,
            /**< Usually massiv server, which doesnt offer data for download. */                 
        DATA_CLIENT             = 3,
            /**< Data client, usually massiv client. Download data from a data service.
                 Downloads only client data files. */                 

        }; // Type
    

    /**
     * Data manager phase.
     */ 
    enum Phase
        {
        PHASE_INIT,
            /**< Initialization phase - is set in constructor. */
        PHASE_IDLE,
            /**< Idle phase - data object manager just check if new data version is available. */
        PHASE_CLEANUP,
            /**< Cleanup phase. Not used now. */
        PHASE_DOWNLOAD,
            /**< Download phase - during this phase data object manager is downloading
                 new data description or data files required for minimal data download. */
        PHASE_UPDATE
            /**< Master data service phase. When masters data server is manually switched
                 to this phase, it cannot upload data to other massiv nodes. It is supposed
                 data are locally updated by administrator during this phase and after
                 update master data service is switched back to idle phase. */
        }; // Phase

    /**
     * Update mode of data manager.
     */
    enum UpdateMode
        {
        AUTOMATIC,
            /**< This mode is set when data update utilitity signalizes about
                 data update. */
        MANUAL
            /**< This mode is set in testing mode of data manager. */
        }; // UpdateMode

    /**
     * Values of data update notification. Data update notification is implemented 
     * using shared memory between processes.
     *
     * There are two possibilities how to update data in data directory.
     * First: data update utility detects (using OS lock) whether any massiv process
     * is runing and its data manager is using data direcotry. If positive,
     * data update is processed by data manager of running massiv process.
     * If negative, data update is processed by data manager created in data update
     * utility.
     *
     * This values are used in the first case. Data update utility 
     * sets value in shared memory to UPDATE_NOTIFY_START_UPDATE. Data manager of running
     * massiv process detects this and switches itself to update phase and sets value to
     * UPDATE_NOTIFY_UPDATE_IN_PROGRESS. After successful data update sets 
     * UPDATE_NOTIFY_UPDATE_FINISHED. Data update utility detects this value and 
     * data update is finished.
     */
    enum UpdateNotify
        {
        UPDATE_NOTIFY_IDLE,
            /**< Must be set when no data update is in progess. */
        UPDATE_NOTIFY_START_UPDATE,
            /**< Set by data update utility when data update is required. */
        UPDATE_NOTIFY_UPDATE_IN_PROGRESS,
            /**< Set by data manager when data updating is in progress. */
        UPDATE_NOTIFY_UPDATE_FINISHED,
            /**< Set by data manager when data updating is sucessfully finished. */
        UPDATE_NOTIFY_UPDATE_ERROR,
            /**< Set by data manager when data update error occured. */
        }; // UpdateNotify
    
    /**
     * Download type of data object manager. It specifies which data files to download.
     *
     * You can define other download type. Maximum number of all download types is 32.
     *
     * Combination of these values is allowed - for example download type of 
     * data object manager can be FRESH_DOWNLOAD | USER_DOWNLOAD,
     * so data object manager will download data files for FRESH_DOWNLOAD and
     * also data files of data nodes which has set @attribute user_download_type
     * to the same value as data object manager's attribute user_download_type. 
     * Method DataManager::get_data_file_names_to_download() has @param of this type
     * and returns list of data files, which has set requsted type.
     */ 
    enum DownloadType
        {
        NO_DOWNLOAD                 = 0x0,
            /**< Used just for master data service. */
        NECESSARY_DOWNLOAD          = 0x1,
            /**< Download only necessary data files. Data manager always download
                 this data files. Data files of sons of root node are automatically 
                 included in NECESSARY_DOWNLOAD to guarantee every path in data node tree
                 has downloaded at least one data file which can be used when creating
                 a new data object. */
        FULL_DOWNLOAD               = 0x2
            /**< Download data file for every non-deprecated data node. */

        }; // DownloadType

    /**
     * Data file download type. Each data file has download type, so 
     * data object manager can decide whether to download it or not.
     * Is is possible to combinate these values.
     */
    enum DataFileDownloadType
        {
        BLANK                       = 0x0,
        NECESSARY_CLIENT_DATA_FILE  = 0x1,
            /**< Data file is necessary to be downloaded to client. */
        NECESSARY_SERVER_DATA_FILE  = 0x2,
            /**< Data file is necessary to be downloaded to server. */
        CLIENT_DATA_FILE            = 0x4,
            /**< Client data file. */
        SERVER_DATA_FILE            = 0x8
            /**< Server data file. */            
        
        }; // DataFileDownloadType

    
    /**
     * Data message error values.
     */
    enum DataMessageErrorType
        {
        NOT_READY_ERROR,
            /**< Massiv server if not fully initialized. */
        DATA_FILE_ERROR,
            /**< Requested data file doesnt exist in data system. */
        DATA_FILE_VERSION_ERROR,
            /**< Requested data file exists in newer version. */
        DATA_SERVICE_UPDATE,
            /**< Actually downloaded data file was updated on server. 
                 Download it again. */
        ERROR_TYPE_COUNT        = 4 
            /**< Count of data message errors. */
        }; // DataMessageErrorType

    enum
        {
        ERROR_TYPE_BIT_COUNT    = 3
            /**< Number of bits required to store data error message type. */
        };

    /** 
     * Version of data description. 
     * Low-order byte of low-order word - minor data version number.
     * High-order byte of low-order word - major data version number.
     * When minor data version reach number DO_MAX_LOW_VERSION_NUMBER,
     * major data version is increased and minor data version is set to 0.
     * Example: 1.99 --\> 2.0
     *          1.21 --\> 1.22
     * Each data description have data version.
     * When data description is changed ( data node is added or data object 
     * is added ), data version is increased.
     * Data file have data version. When @method DataManager::update_data_file()
     * is called, data file version is increased.
     */
    class DataVersion
        {

        enum Param
            {
            DO_MAX_LOW_VERSION_NUMBER       = 99
                /**< Maximum minor data version number. */
            }; // Param

    public:
        
        /** Constructor. */
        DataVersion();

        /** Constructor. */
        DataVersion
            (
            int         major_version,
            int         minor_version
            );
        
        /**
         * Read serialized data version from text stream.
         */
        void text_write
            (
            TextWriter &            tw
            ) const;

        /**
         * Serialize to text.
         */ 
        void text_read
            (
            TextReader &            tr
            );

        /**
         * Read serialized data version from bit stream.
         */ 
        void bitstr_read
            (
            BitIStream &            bis
            );

        /**
         * Serialize to bit stream.
         */
        void bitstr_write
            (
            BitOStream &            bos
            ) const;

        /**
         * Operator==.
         */
        bool operator==
            (
            const DataVersion & v
            ) const;

        /**
         * Operator!=.
         */
        bool operator!=
            (
            const DataVersion & v
            ) const;

        /**
         * Binary operator<.
         * Returns true if this data version is bigger then @param v data version.
         */
        bool operator<
            (
            const DataVersion & v
            ) const;

        /**
         * Binary operator>.
         * Returns true if @param v data version is bigger than this data version.
         */
        bool operator>
            (
            const DataVersion & v
            ) const;

        /**
         * Binary operator<=.
         */
        bool operator<=
            (
            const DataVersion & v
            ) const;

        /**
         * Binary operator>=.
         */
        bool operator>=
            (
            const DataVersion & v
            ) const;

        /** 
         * Increase version by one.
         */
        void increase();

        /**
         * Return major data version.
         */
        int get_major() const;

        /**
         * Return minor data version.
         */
        int get_minor() const;

    private:
        
        unsigned int        data_version;
            /**< Data version. */

    }; // DataVersion



    /**
     * Contain information about one data file.
     * Used for simple test if file in a data directory is really
     * requested data file.
     */
    struct DataFileInfo
        {
        
        /** Constructor. */
        DataFileInfo();

        /** Constructor. */
        DataFileInfo
            (
            const std::string &         _name,
            const DataVersion &         _version
            );            

        /**
         * Binary operator==.
         */
        bool operator==
            (
            const DataFileInfo & v
            ) const;

        /**
         * Load file information from disk
         * Currently size of file is read.
         */
        void load_file_info
            (
            const std::string &         data_directory
            );

        /**
         * Serialize to text.
         */
        void text_write
            (
            TextWriter &            tw
            ) const;

        /**
         * Read serialized information from text.
         * @param light_weight - true = read just data file name
         *                     - false = read complete data file information (version, size, etc.)
         */ 
        void text_read
            (
            TextReader &            tr,
            bool                    light_weigth
            );


        /**
         * Validate if file in a data directory is really requested data file.
         */
        void validate
            (
            const std::string &     data_directory
            ) const;

        std::string                 name;
            /**< Data file name. */

        DataVersion                 version;
            /**< Version of data file. Initial version is always set to 1.0 */

        FileSizeType                size;
            /** Size of data file. Just for simple check if data file is valid. */

        }; // DataFileInfo

    /**
     * Tree of data nodes.
     */ 
    struct DataNodeTree
        {

        /**
         * Represent node in data node tree.
         */
        struct DataNode
            {
    
            typedef std::map
                        <
                        std::string,
                        DataNode *
                        > NodeList;

            /** Constructor. */
            DataNode();
            
            /** Contructor. */
            DataNode
                (
                const std::string &     _full_name, 
                const std::string &     _file_name,
                DataFileDownloadType    _download_type = BLANK,
                bool                    _deprecated = false,
                const DataVersion &     _file_version = DataVersion( 0, 99 )
                );
                
            /**
             * Serialize to text.
             */
            void text_write
                (
                TextWriter &            tw
                ) const;

            /**
             * Read serialized information from text.
             * @param light_weight - true = read data node without data file information (version, size)
             */ 
            void text_read
                (
                TextReader &            tr,
                bool                    light_weigth
                );
                
            
            std::string             name;   
                /**< Name of a data node.
                     Format: "my_name" */

            std::string             full_name;
                /**< Full name of a data node.
                     Format: "/name1/name2/my_name" */
            
            NodeList                node_list;
                /**< List of sons of this node. */

            DataFileDownloadType    download_type;
                /**< Download type for data file. */

            bool                    deprecated;    
                /**< == true -> data node is deprecated and can't be used. */

            bool                    unusable;
                /**< == true -> data node is unusable. For example data node file 
                     is not downloaded yet. */
            
            DataFileInfo            data_file_info;
                /**< Information about data file. */

            }; // DataNode

        typedef std::stack<DataNode *>        DataNodeStack;
            // Used when walking across tree.

        /** Constructor. */
        DataNodeTree();

        /** Destructor. */
        ~DataNodeTree();

        
        /**
         * Return reference to data node.
         */
        DataNode & get_data_node
            ( 
            const std::string &         data_node_full_name
            );


        /**
         * Return full name of father of data node.
         */
        static std::string get_data_node_father_name
            ( 
            const std::string &         data_node_full_name
            );

        /**
         * Return data node name from full name.
         */
        static std::string get_data_node_name
            ( 
            const std::string &         data_node_full_name
            );

        /**
         * Return file name for @param data node.
         */
        std::string get_data_node_file_name
            ( 
            const std::string &         data_node_full_name
            );

           
        /**
         * Add sons of node to stack. Last son added first.
         */
        static void add_sons_to_stack
            (
            const DataNode *                   father,
            DataNodeStack &                    s
            );

        DataNode *                      root_node;
            /** Root of a tree. */

        }; // DataNodeTree


    /**
     * Store information about all available data objects and their actual
     * pointers to the data node tree.
     */
    struct DataObjectList
        {

        /**
         * Information about data objects and data object paths are
         * stored here.
         */
        struct DataObjectInfo
            {                       
            
            /** Constructor. */
            DataObjectInfo
                (
                const std::string     _data_object_name,
                const std::string     _node_path
                );

            std::string     data_object_name;
                /**< Data object name. */
            std::string     node_path;
                /**< Data node path in data node tree. */
            };
        
        typedef std::map
                    <
                    std::string,
                    DataObjectInfo
                    > DataObjectInfoHash;

        /**
         * Return path in data node tree to the data node which
         * represents requested data object.
         */
        std::string get_data_object_node_path
            (
            const std::string &         data_object_name
            );

        /**
         * Add data object name to list. If data object name already existed,
         * its node path is actualized.
         */
        void add
            (
            const std::string &         data_object_name,
            const std::string &         node_path
            );
            
        DataObjectInfoHash              data_object_info_hash;
            /**< Hash table for information about data object names and their pointers
                 to a data node tree. */
                
        }; // DataObjectList

    
    struct DataDescription; // Forward declaration.
    
    /**
     * List of data files downloaded in data directory. 
     * Contain all data files if master data service ( secondary data service ).
     * Contain downloaded data files if data client.
     */
    struct DataFileList
        {

        typedef std::map
                    <
                    std::string,
                    DataFileInfo
                    > DataFileListMap;


        /**
         * Add data file info structure to table.
         */ 
        void add
            (
            const DataFileInfo &        data_file_info
            );

        /**
         * Find data file name in table.
         * Return true when success.
         */
        bool find
            (
            const std::string &         data_file_name
            ) const;

        /**
         * Return reference to DataFileInfo structure for requested data file.
         */
        DataFileInfo & get
            (
            const std::string &         data_file_name
            );

        /**
         * Save data file list map to stream.
         */
        void save
            (
            std::ostream &          stream            
            ) const;

        /**
         * Load data file list map from stream and validate data file.
         */
        void load
            (
            std::istream &          stream,
            const std::string &     data_directory
            );

        /**
         * Refresh data file list.
         * Some data files may be changed in data description, so this method
         * ensures that data file list of downloaded data files is updated.
         */
        void refresh
            (
            DataDescription &       data_description,
            DataDescription &       new_data_description,
            const std::string &     data_directory,
            DataManager &           dm
            );

        /**
         * Destroy object.
         */
        void destroy();


        DataFileListMap             data_file_list_map;
            /** Data file list. Maps data file name to data node full name. */            

        }; // DataFileList
    
    friend struct DataFileList;

    /**
     * Data description structure. Encapsulates whole data system description
     * into one complex.
     */
    struct DataDescription
        {

        /** Maps data file name to data node name. */
        typedef ext::hash_map
                    <
                    std::string,
                    std::string
                    > DataFileNameList;

        /** 
         * Constructor.
         * @param light_weight - true = load data description without data file information (version, size)
         */
        DataDescription
            (
            std::istream &          stream,
            const std::string &     data_directory,
            bool                    light_weight = false
            );

        /** 
         * Validate if data object's node path of all data objects point to 
         * valid data nodes in data node tree.
         */
        void validate_all_data_objects();

        /** 
         * Validate if data object's node path point to 
         * valid data node in data node tree.
         */
        void validate_data_object
            (
            const std::string &     data_object_name
            );

        /** Save description to stream. */
        void save_description
            (
            std::ostream &          stream
            );  

        /**
         * Recursive function. Write one node to text writer and
         * call itself to sons.
         */
        void write_node
            (
            TextWriter &                        tw,
            const DataNodeTree::DataNode &      node
            ) const;

        /** 
         * Load description from stream.
         * @param light_weight true = load data description without data file information (version, size)
         * @param data_directory data directory - used in error message
         */
        void load_description
            (
            std::istream &          stream,
            const std::string &     data_directory,
            bool                    light_weight
            );  

        /** Copy one stream to other by lines. */
        static void copy_stream_by_lines
            (
            std::ostream &          dest_stream,
            std::istream &          source_stream
            );
        
        /**
         * Add a new node to existing tree.
         * This method changes tree, so data version is increased by default.
         */
        void add_data_node
            (
            DataNodeTree::DataNode *        new_node
            );

        /**
         * Add new data object with name @param data_object_name to hash table.
         * @param node_path - pointer to data node tree.
         * @param point_to_deprecated - data object can point to deprecated 
         * data node. It's used within local update of master data service.
         */
        void add_data_object
            (
            const std::string &         data_object_name,
            const std::string &         node_path
            );

        /**
         * Remove data object from data object hash table. This data object can't be 
         * used in simulation after that.
         */
        void remove_data_object
            (
            const std::string &         data_object_name
            );
            
        /**
         * Return data version.
         */
        DataVersion get_data_version() const;

        /**
         * Return magic number.
         */
        MagicNumberType get_magic_number() const;

        /**
         * Reads data version from stream, where syntax of a stream is expected like "[2.56.27658]".
         * First two numbers are data version, the second one is data magic number.
         */
        static void get_initial_information_from_data_description
            (
            const std::string &     data_directory,
            std::istream &          stream,
            DataVersion &           data_version,
            UInt32 &                magic_number
            );
        
        /**
         * Return data node name which owns @param data file.
         * Used for fast access to data_node.
         */
        const std::string & get_data_node_name
            (
            const std::string &     data_file_name
            );
        
        /**
         * Read data version and magic number from text reader.
         * @param tr text reader to read from
         * @param data_version store data version here
         * @param data_version store magic number here
         */
        static void read_data_version
            (
            TextReader &        tr,
            DataVersion &       data_version,
            MagicNumberType &   magic_number
            );
        
        /**
         * Read format version form text reader.
         * @param tr text reader to read from
         * @param data_directory data directory - used in error message
         */
        static void read_format_version
            (
            TextReader &                    tr,
            const std::string &             data_directory
            );

        DataObjectList                  data_object_list;
            /**< Data object list. */

        DataNodeTree                    data_node_tree;
            /**< Data node tree. */

        DataVersion                     data_version;  
            /**< Data version. */

        MagicNumberType                 magic_number;
            /**< Data magic number. */
            
        DataFileNameList                file_list;
            /**< List of all data files. One data file belong to one data node.
                 One data node can own just one data file.
                 Data file name is unique, so two data nodes cannot own data 
                 file with the same name. 
                 This attribute ensures faster access to data node using
                 key data file name. */

        }; // DataDescription


    /**
     * Provide service for downloading data description file and data files.
     * Note that data object manager doesn't need to be configured for
     * downloading files(master data service). See Type.
     */
    class DownloadProvider
        {
    
        friend class DataManager;
        friend class DataObject;
        
        enum DownloadPhase
            {
            PHASE_IDLE,
                /**< Idle download phase. */
            PHASE_DATA_DESCRIPTION_DOWNLOAD,
                /**< Data description download. */
            PHASE_DATA_FILES_DOWNLOAD,
                /**< Data files download. */
            PHASE_FILE_DOWNLOAD
                /**< Phase during downloading of data file. */
            };

        enum Status
            {
            STATUS_INIT,
                /**< Initialization of some download phase. */
            STATUS_PENDINGX,
                /**< Last phase in not completed. */
            STATUS_SUCCESS,
                /**< Last phase is successfully completed. */
            STATUS_FAILED,
                /**< Last phase failed to complete. */
            STATUS_IDLE
                /**< Idle status. */
            };

        /**
         * Keep information about download.
         * Handler routine of data version system message stores here
         * information about actual simulation data version.
         */
        struct DownloadInfo
            {
            
            /** Constructor. */
            DownloadInfo();

            bool                        valid;
                /**< True when stored information are valid. */

            DataVersion                 data_version;
                /**< Data version of master data service. */

            MagicNumberType             magic_number;
                /**< Data magic number of master data service. */
                
            NodeId                      node_id;
                /**< NodeId of a massiv server to download data from. 
                     Not used now.*/

            };

        /** Constructor. */
        DownloadProvider();

        /** Destructor. */
        ~DownloadProvider();

        /**
         * Destroy object.
         */
        void destroy();

    private:        
    
        DownloadPhase               phase;
            /**< Phase of download provider. */

        DownloadPhase               phase_before;
            /**< Phase before actual phase. */

        DataDescription *           data_description;
            /**< Downloaded data description. */

        FileSizeType                offset;
            /**< Actual offset of downloaded file. */

        DataVersion                 data_version;
            /**< Data version of actual downloaded data description. */

        MagicNumberType             magic_number;
            /**< Data magic number of actual downloaded data description. */
            
        std::stringstream           data_file_names;
            /**< Stream to store list of data files to be downloaded. */

        std::string                 actual_downloaded_file_name;
            /**< Actual downloaded file name. */

        DataVersion                 actual_downloaded_data_version;
            /**< Data version of downloaded file. */

        VolumeManager::OFStream     data_stream;
            /**< Stream used to store downloaded data file. */
             
        DownloadType                download_type;
            /**< Download type. */

        DownloadInfo                fresh_download_info;
            /**< The most recent download information. System message stores here information
                 about the most recent data version. Not used when message
                 came during download executive phases. */

        DownloadInfo                active_download_info;
            /**< Download information. Contains a copy of attribute fresh_download_info.
                 Information are not changed during download executive phases. */

        Status                      status;
            /**< Actual downloading status of particular phases. */

        DataMessageErrorType        error;
            /**< Data message error type of last data error message. */
        
        NodeId                      data_service_node_id;
            /**< NodeId of massiv data server to download from. */

        SystemTime                  last_request_time;
            /**< Platform time when last request message was sent. */
        
        SystemTime                  current_time;
            /**< Current system time. */
        
        SystemTime                  first_download_time;
            /**< Starting platform time of download of one data file. */

        StringQueue                 custom_data_file_download;
            /**< Queue of data file names which should be downloaded.
                 Used when particular data files are requested. */
        };

    /**
     * Provide service for uploading data description file and data files.
     * Note that data object manager doesn't need to be configured for
     * uploading files(client). See Type.
     */
    class UploadProvider
        {
    
        friend class DataManager;

    public: 

        enum UploadPhase
            {
            PHASE_IDLE,
                /**< Idle upload phase. */
            PHASE_SEND_DATA_VERSION,
                /**< Phase when data version system message is sent to massiv node. */
            PHASE_UPLOAD_DATA_FILE
                /**< Phase when data file is uploaded to massiv node. */
            }; // UploadPhase

        struct UploadInfo
            {
            
            /** Constructor. */
            UploadInfo();

            /** Copy constructor. */
            UploadInfo
                (
                const UploadInfo & ui
                );

            bool                        deleted;
                /**< True when this structure is marked as deleted. */
            
            NodeId                      node_id;
                /**< NodeId of massiv node which "owns" this structure. */

            UploadPhase                 phase;
                /**< Phase of this UploadInfo. */

            std::string                 data_file_name;
                /**< Data files name to be uploaded to node. */

            DataVersion                 data_version;
                /**< Data version when uploading data description. */
            
            FileSizeType                offset;
                /**< Offset of uploaded data file. */

            int                         buffer_size;
                /**< Size of a data message buffer for data. */

            int                         buffer_len;
                /**< Size of data in data message buffer. */

            VolumeManager::IFStream     data_stream;
                /**< Stream of uploaded data file. */

            FileSizeType                file_size;
                /**< Data file size. */
            
            bool                        wait_for_confirmation;
                /**< True when waiting for confirmation of last data message. */
            
            SystemTime                  last_upload_time;
                /**< Platform time of last data upload. */
                        
            SystemTime                  current_time;
                /**< Actual platform time. */

            SystemTime                  first_upload_time;
                /**< Platform time of a first data upload. */

            int                         average_speed;
                /**< Average upload speed in bytes per second. */

            int                         num_uploads;
                /**< Number of data messages sent to a node. */
            }; // UploadInfo

        typedef ext::hash_map
                    <
                    NodeId,
                    UploadInfo,
                    NodeId::Hash,
                    NodeId::EqualKey
                    > UploadInfoHash;

        /**
         * Clear all deleted requests - removes all upload requests 
         * marked as deleted from hash table.
         */
        void clear_deleted_requests();
        
        /**
         * Add upload request into hash table.
         */
         void add_request
            (
            const NodeId &              node_id
            );

        /**
         * Mark request for @param node_id as deleted.
         */
         void remove_request
            (
            const NodeId &              node_id
            );

        /**
         * Return true if request for @param node id exist.
         */
        bool find_request
            (
            const NodeId &              node_id
            );

        /**
         * Return reference to a structure for @param node id.
         */
        UploadInfo & get_request
            (
            const NodeId &              node_id
            );

        /**
         * Destroy object. 
         * Send data error message to nodes uploading to.
         */
        void destroy
            (
            DataMessageErrorType      data_message_error_type
            );

        /**
         * Close files of all requests which currently uploading it.
         * It also sends data error message to a massiv node.
         */
        void close_open_file
            ( 
            const std::string &         file_name
            );

    private:        
    
        UploadInfoHash              upload_info_hash;
            /**< Hash NodeId to structure UploadInfo. Every node which want to download
                 data has its own structure here. */

        }; // UploadProvider
    
    /**
     * UpdateProvider.
     * Provides data update using asynchronous data file copying.
     */
    struct UpdateProvider
        {
        
        typedef std::vector< std::string >                      VecStr;

        /**
         * Data update phase.
         */
        enum Phase
            {
            PHASE_IDLE,
                /**< No update is in progress. */
            PHASE_LOAD_AND_CHECK_UPDATE_FILE,
                /**< Update information checked for validity. If positive,
                     next update phase is set. */                     
            PHASE_BACKUP_DATA_FILES,
                /**< Actual version of data is backuped. Data update is not
                     processed unless old version data is successfuly backuped. */
            PHASE_COMPARE_DATA_FILES,
                /**< Data files are compared with new versions in source directory. */
            PHASE_COPY_DATA_FILES,
                /**< Data files are copied to data directory. */
            PHASE_UPDATE_DATA_DESCRIPTION
                /**< Data description file is actualized and saved to data directory. */
            }; // Phase
        
        /**
         * Data update status.
         */
        enum UPStatus
            {
            STATUS_INIT,
                /**< Used by some update phases to determine its beginning. */
            STATUS_PENDINGX
                /**< Used by some update phases. This status is set after init status. */
            }; // Status
            
        /**
         * Constructor.
         */
        UpdateProvider();
        
        /**
         * Initializes attributes to basic values.
         */
        void initialize();
        
        /**
         * Loads and checks data update information from data update file.
         * Returns true if it is possible to update data.
         */
        bool load_and_check
            (
            const DataManager &     data_manager
            );                
                        
        std::auto_ptr< DataDescription > ud;
            /** Safe pointer to update description. */
        
        Phase           phase;
            /**< Update provider phase. */

        UPStatus        status;
            /**< Update provider status. */
        
        VecStr          copy_file_list;
            /**< List of data file names to be copied to data directory. */

        VecStr          backup_file_list;
            /**< List of data file names to be backuped. */

        VecStr          delete_file_list;
            /**< List of data file names to be deleted from file lists. */

        FileCopy        cp;
            /**< Used for synchronous data files copying. */

        CompareFile     cf;
            /**< Used for data file comparing. */
        
        bool            test;
            /**< True when data manager just check validity of data update information.
                 No data update is really processed. */
        
        std::string     backup_dir;
            /**< Name of backup directory. Actually it's data_directory/[old data version]. */
                  
        std::string     actual_file_name;
            /**< Actual copied/compared file name. */
              
        bool            next_update;
            /**< FIXME */
            
        }; // UpdateProvider

    friend struct UpdateProvider;
    
public:

    /**
     * DataManager settings.
     */
    class DataManagerSettings : public Settings
        {
    public:

        /** Constructor. */
        DataManagerSettings();

    public:

        Integer &           download_type;
            /**< DataManager download type. */
        String &            data_service_node;
            /**< Data service node id. */
        Integer &           max_download_speed;
            /**< Maximal upload speed per connection. */
        Integer &           check_data_version_interval;
            /**< Data version check interval in seconds. */
        Boolean &           mount_data_volume;
            /**< If true data volume containing the downloaded data will
                 be stored in the data directory instead of the data itself.
                 The data volume image name can be specified by
                 @c data_volume_image_name variable. */
        String &            data_volume_image_name;
            /**< Data volume image name. */

        }; // DataManagerSettings.

public:

    /**
     * Constructor.
     * @param data_directory - directory for storing and loading data node files.
     * @param type - initial flags that restrict using of data object manager.
     */
    DataManager();
    
    /**
     * Destructor.
     */
    ~DataManager();

   
    /**
     * Must be called to initialize DataManager from registry.
     */
    void initialize
        (
        Type    type
        );

    /**
     * Manual initialization of DataManager.
     * If @param test is true, data manager is in test mode. Used only
     * when testing functionality of data manager.
     */
    void _initialize
        (
        const std::string &             data_dir,
        const std::string &             source_dir,
        Type                            init_type,
        DownloadType                    init_download_type,
        bool                            test = false,
        bool                            exclusive_access = true,
        NodeId                          data_service_node_id = NodeId()
        );

    /**
     * Destroy DataManager. Instance is not deleted and DataManager can be 
     * initialized again.
     */
    void destroy();

    /**
     * Run next data object manager tick.
     */
    void next_tick();

    
    /**
     * Return data object of @param requested name.
     *
     * If @a open_data_file is @c true, DataObject::open() will be called
     * automatically on the object.
     */
    DataObject * get_data_object
        (
        const std::string &             data_object_name,
        bool                            open_data_file = false
        );

    /**
     * Return true when data object manager is ready to offer data object.
     * To get data object call method get_data_object().
     */
    bool is_ready_for_use() const;


    /**
     * Return actual data version of data node manager.
     */
    DataVersion get_data_version() const;

    /**
     * Return actual data magic number of data node manager.
     */
    MagicNumberType get_magic_number() const;

    /**
     * Remove node id from upload info hash table.
     * This method can be called only on data service.
     * FIXME: ted se vola i na serveru a clientovi, ale nic to nedela
     */
    void destroy_request
        (
        const NodeId    & node_id
        );

    /**
     * Register to data service. This method is called when anonymous connection
     * to data service is closed and regular connection to simulation is created.
     */
    void register_to_data_service() const;
    
    /**
     * Add a @param new data node to existing data node tree.
     * If @param data_object_name if not empty, new data object is added to
     * existing data object list. If data object already existed its data node path
     * is updated.
     * This method changes data description, so data version is increased by default.
     * If @param should_save_to_disk is true, data version is increased and 
     * data description is saved to disk.
     * Only "master data service" can directly call this method. Else exception is thrown.
     */
    void add_data
        (
        DataNodeTree::DataNode *        new_data_node,
        const std::string &             data_object_name,
        bool                            should_save_to_disk = true
        );

    /**
     * Add new data object with name @param data_object_name.
     * @param node_path - pointer to data node tree.
     * If data object already exists, its data node path will be updated.
     * This method changes data description, so data version is increased by default.
     * If @param should_save_to_disk is true, data version is increased and 
     * data description is saved to disk.
     */
    void add_data_object
        (
        const std::string &             data_object_name,
        const std::string &             node_path,
        bool                            should_save_to_disk = true
        );

    /**
     * Remove data object from data object hash table. This data object can't be 
     * used in simulation after that.
     * This method changes data description, so data version is increased by default.
     * If @param should_save_to_disk is true, data version is increased and 
     * data description is saved to disk.
     */
    void remove_data_object
        (
        const std::string &             data_object_name,
        bool                            should_save_to_disk = true
        );    
    
    /**
     * Return true if download phase was succesfully finished.
     */
    bool download_success();


    /**
     * Data file in data directory was updated ( new version of file was copied
     * to data directory ). Data file version will be increased by one and data node
     * ( which belong to this data file ) in data node tree will be updated too.
     * If @param should_save_to_disk is true, data version is increased and 
     * data description is saved to disk.
     */ 
    void update_data_file
        (
        const std::string &             data_file_name,
        bool                            should_save_to_disk = true
        );        

    /**
     * Data message handling methods.
     */
    void handle_data_version_message
        (
        const DataVersionMessage &          msg
        );
    void handle_data_message
        (
        const DataMessage &                 msg
        );
    void handle_data_version_request_message
        (
        const DataVersionRequestMessage &   msg
        );
    void handle_data_request_message
        (
        const DataRequestMessage &          msg
        );
    void handle_data_confirmation_message
        (
        const DataConfirmationMessage &     msg
        );
    void handle_data_error_message
        (
        const DataErrorMessage &            msg
        );
    void handle_data_update_request_message
        (
        const DataUpdateRequestMessage &    msg
        );

    
    /**
     * Send data update request do master data service.
     */
    bool request_master_data_update();

    /**
     * This method must be called when this node is switched from/to
     * anonymous mode to/from full mode.
     */
    void switch_from_to_anonymous_mode();
    
private:

    /**
     * Update downloaded data file list after data file download.
     */
    void update_data_file_after_download
        (
        const std::string &             data_file_name
        );

    /**
     * Update all registered data object which use @param data file.
     */
    void update_used_data_object_after_data_file_update
        (
        const std::string &     data_file_name
        );
    
    /**
     * Calls @method DataObject::update() on all data objects with name @param
     * data_object_name in @attribute used_data_object_hash.
     * Usually called by data object manager when data node path for this data object is
     * updated ( that happens when new data file is downloaded or new data description is
     * downloaded ).
     */
    void update_used_data_object
        (
        const std::string &     data_object_name
        );

    
    /**
     * Load data description from data directory, 
     * build and return intance of DataDescription.
     * @param light_weight - true = read data description without data file information (version, size)
     * @param file_name - relative file name of data description file; if this value is "", 
     *                    default data description file name is used
     */
    DataDescription * load_data_description
        (
        bool                    light_weigth = false,
        const std::string &     file_name = ""
        ) const;

    /**
     * Save data description to file in data directory.
     */
    void save_data_description();

    /** 
     * Mark data nodes as usable when its data file is downloaded.
     * Return number of marked data nodes.
     */
    int set_usable_data_nodes();
    
    /**
     * Save downloaded data file list to disk.
     */ 
    void save_data_file_list();

    /**
     * Load downloaded data file list from data directory.
     */
    void load_data_file_list();

    /**
     * Release data object from hash table. Called by destructor of data object.
     */ 
    void release_data_object
        (
        DataObject *             data_object
        );

    
    /**
     * Phase methods.
     */
    void phase_init();
    void phase_idle();
    void phase_cleanup();
    void phase_download();
    void phase_update();


    /**
     * Download provider phases.
     */
    void dp_phase_check_data_version();
    void dp_phase_data_description_download();
    void dp_phase_data_files_download();
    void dp_phase_file_download();

    /**
     * Update provider phases.
     */
    void updatep_phase_load_and_check_update_file();
    void updatep_phase_backup_data_files();
    void updatep_phase_compare_data_files();
    void updatep_phase_copy_data_files();
    void updatep_phase_update_data_description();

public:

    /**
     * Set update provider to test mode. When called update provider just check validity
     * of data update infromation.
     */
    void updatep_set_test();
    
    /**
     * Returns false when data update is in progress.
     */
    bool update_finished();
    
private:
    /**
     * Next tick of upload provider.
     */
    void up_next_tick();
    
    /**
     * Send data version message to massiv node.
     */
    void up_send_data_version
        (
        UploadProvider::UploadInfo &    ui
        );

    
    /**
     * Send data message to massiv node.
     */
    void up_upload_data
        (
        UploadProvider::UploadInfo &    ui
        );


    /**
     * Get list of file names to be downloaded. Depending on @param download_type @param
     * stream is filled with file names, each file name in new line.
     * Every data file is decided whether to
     * download it or not. If DataNode::download_type of a data node is in compliance with
     * @param download_type, data file for this data node is added to @param stream.
     * @Return number of data files to be downloaded.
     */ 
    int get_data_file_names_to_download
        (
        DownloadType                    download_type,
        std::ostream &                  stream,
        const DataDescription *         data_description
        );    
public:

    /**
     * Return true if @param file exists and can be opened for reading.
     * @param in_data_dir specifies whether to file is located in data_directory.
     */
    bool test_file_exist
        (
        const std::string &         file_name,
        bool                        in_data_dir = true
        ) const;

private:

    /**
     * Return true, if phase is PHASE_IDLE and every upload info record in upload provider
     * is in phase UploadProvider::PHASE_IDLE. 
     * Otherwise return false.
     */
    bool everything_idle();

    /**
     * Return true if files exist ( each file name in new line in @param file_names )
     * in data_directory and can be opened for reading.
     * @param throw_exception - if true it throws an exception if some data file doesnt exist.
     */
    bool test_files_exist
        (
        std::stringstream &         file_names,
        bool                        throw_exception = true
        );


    /**
     * This method is called when new version of data file @param file_name 
     * is downloaded while some data object has this data file opened. 
     * This method saves internal file position and closes files of data objects.
     * After that is supposed data file is renamed (usually to name <original name>.<version> )
     * Then is supposed to call @method restore_data_object_file with parameters 
     * old_file_name (original data file name) and new_file_name (new data file name).
     */
    int close_data_object_open_file
        (
        const std::string       file_name
        );

    /**
     * Open data file with of data objects with data file name @param old_file_name.
     * Renames data file of data objects to @param new_file_name, opens it and
     * seeks saved saved file position. See @method close_data_object_open_file.
     */
    void restore_data_object_file
        (
        const std::string       old_file_name,
        const std::string       new_file_name
        );

    /**
     * Generate magic number for this series of data versions.
     * Each series of data version have its own data magic number. When data structure
     * is changed to version which it not compatible with last data version, new magic
     * number is generated. This ensures data clients will delete its actual data
     * and download new one.
     */
    UInt32 generate_magic_number();
    
public:    

    /**
     * Switch master data service to update phase. This is the only one phase
     * when master data service can be locally updated. During this phase are data not
     * uploaded to other massiv nodes.
     * @param update_mode = AUTOMATIC - data manager reads data update information
     *                                  and processes data update.
     *                    = MANUAL - used in internal tests of data manager functionality.
     */
    void switch_to_update_phase
        (
        UpdateMode              update_mode
        );

    /**
     * Switch master data phase back to idle phase.
     */
    void finish_update_phase();

    
    /**
     * Return data directory.
     */
    const std::string & get_data_directory();

    /**
     * Backup data into data_directory\version_number
     * If backup of current version of data already exist, nothing is done.
     * Return true when data are backuped.
     */
    bool backup_data();
        

private:

    
    std::string                     data_directory;
        /**< Path to data directory where to store and load data files. */
    
    std::string                     source_directory;
        /**< Path to data source directory where. Data files for update are stored here. */

    DataDescription *               data_description;
        /**< Description of actual version of data. */

    DataFileList                    data_file_list;
        /**< List of data files in data directory. */

    UsedDataObjectHash              used_data_object_hash;
        /**< Registered instances of DataObject. */

    bool                            ready_for_use;
        /**< True when data object manager is ready to offer data object. */

    Type                            type;
        /**< Data object manager type. */

    DownloadType                    download_type;         
        /**< Download type. */

    DownloadProvider                dp;
        /**< Provide data download. */

    UploadProvider                  up;
        /**< Provide data upload. */

    Phase                           phase;  
        /**< Data object manager phase. */

    DataManagerSettings             data_manager_settings;
        /**< Data manager settings. */

    bool                            initialized;
        /**< True when DataManager was initialized by method _initialize(). */
    
    bool                            test_mode;
        /**< True when data manager is in test mode. Used for testing and debugging. */
            
    UpdateProvider                  update;
        /**< Provide master data service update. */
                
    }; // DataManager


/**
 * Data object name ( std::string ) is required to get data object.
 * To get data object call @method DataManager::get_data_object().
 * To use data of data object use @attribute data. 
 * @Method is_data_updated() must be called before every access 
 * to @attribute data, because data file can be changed ( new version downloaded )
 * during particular accesses.
 */
class MASSIV_EXPORTED DataObject
    {

    friend class DataManager;

public:
        
    /** 
     * Constructor.
     * If @param open_data_file is true, data file of data object
     * can be used without necessity to call @method open().
     */
    DataObject
        (
        DataManager *               _data_manager,
        const std::string &         _data_object_name,
        bool                        open_data_file = false
        );

    /** Destructor. */
    ~DataObject();

    /**
     * User of data object must call this method to obtain new data.
     * Any other structure that use information from @attribute data
     * must be reinitialized. It is on user of data object to ensure this.
     * Call @method is_data_updated() before calling this method.
     * Do nothing when new data are not available.
     *
     * If @param open_data_file is true, data file of data object
     * can be used without necessity to call @method open() otherwise
     * @method open() must be called.
     */
    void refresh
        (
        bool        open_data_file = false
        );

    /**
     * True when new data are available. User of this data object must call
     * method refresh() to obtain new data.
     */
    bool is_data_updated() const;

    
    /** Return data object name. */
    std::string get_name() const;
        
    /**
     * Return size of data file used by data object.
     */
    DataManager::FileSizeType get_size() const;

    /**
     * Force to download the most actual version of data file of data object.
     * When new version is available, @method is_data_updated() return true.
     * After that @method refresh() must be called to use new downloaded data version
     * of data file.
     */
    void download_actual_version();

    /**
     * Open data file of data object.
     */
    void open();
    
    /**
     * Close data file of data object.
     */
    void close();
        
private:

    /**
     * Must be called to release data object from data object manager hash table.
     * Called by destructor if data object was not released before.
     */
    void release();


    /**
     * Called by data object manager when new data is available.
     */
    void update();

public:

    VolumeManager::IFStream         data;
        /**< Data of data object. Using of a stream is guaranteed 
             even if data is updated (for example a new version is downloaded). */

private:

    DataManager * const             data_manager;
        /**< Reference to data object manager which owns this data object. */


    bool                            data_updated;
        /**< True when new data appears for this data object - data could be
             downloaded or locally added (master data service). */
    
    std::string                     data_object_name;
        /**< Name of data object. */

    std::string                     node_path;
        /**< Actual pointer to data node tree of
             data node manager which owns this data object. */

    std::string                     data_file_full_name;
        /**< Data file full name for fast access. */

    DataManager::FileSizeType       file_size;
        /**< Size of data file. */

    DataManager::FileSizeType       offset;
        /**< Offset in data file. */

    }; // DataObject

/** @} */ // addtogroup data_manager


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_DATA_MANAGER_H
