/** @file path_manager.h
 * PathManager declaration.
 */

/* 
 * begin        : Sun Nov 23 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_PATH_MANAGER_H
#define MASSIV_PATH_MANAGER_H

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

namespace Massiv {

class System;

namespace Core {

/**
 * Class PathManager.
 * Class offers several methods to obtain paths to various Massiv directories.
 * Call Global::PathManager() to obtain global instance of PathManager.
 *
 * @ingroup unclassified
 */
class MASSIV_EXPORTED PathManager
    {
public:
    
    friend class Massiv::System;

    /**
     * PathManager settings.
     * If some path from settings begins with substring "$HOME", it will be
     * replaced by real home directory, i.e. "/home/homer" on a linux platform,
     * "c:\Documents and Settings\bart" on a windows platform.
     */
    class PathManagerSettings : public Settings
        {
    public:

        /** Constructor. */
        PathManagerSettings();

    public:

        String &    login_directory;
            /**< Login directory. */
        String &    source_data_directory;
            /**< Source data directory. */
        String &    data_directory;
            /**< Data directory. */
        String &    archive_directory;
            /**< Archive directory. */
        String &    log_directory;
            /**< Log directory. */
        String &    work_directory;
            /**< Work directory. */
        }; // PathManagerSettings.
    
public:

    /** Constructor. */
    PathManager();
        
    /**
     * Replace $HOME shortcut in @param directory_path by the absolute
     * path to the actual home directory.
     */
    std::string replace_home_shortcut
        (
        const std::string & directory_path
        );

    /** Get home directory for current user. */
    const std::string & get_home_directory() const;
    
    /** Get login directory. */
    const std::string & get_login_directory() const;

    /** Get source data directory. */
    const std::string & get_source_data_directory() const;

    /** Get data directory. */
    const std::string & get_data_directory() const;

    /** Get archive directory. */
    const std::string & get_archive_directory() const;

    /** Get log directory. */
    const std::string & get_log_directory() const;

    /** Get work directory. */
    const std::string & get_work_directory() const;

private:

    /**
     * Resolves path to initial system config.
     *
     * @param directory_path    Path to the config file. May contain $HOME
     *                          shortcut.
     *
     * Made static so that system can read config from user's home directory
     * (PathManager does not run yet).
     */
    static std::string resolve_path_to_config
        (
        const std::string & directory_path
        );
    
private:
    
    PathManagerSettings path_manager_settings;
        /**< Path manager settings. */

    std::string         home_directory;
        /**< Home directory for current user. */
    std::string         login_directory;
        /**< Login directory. Contains login files to connect to simulation. */
    std::string         source_data_directory;
        /**< Source data directory. Used by master data service only. */
    std::string         data_directory;
        /**< Data directory. Contains all data files used in simulation. */
    std::string         archive_directory;
        /**< Archive direcory with archives. Used by simulation servers only. */
    std::string         log_directory;
        /**< Log directory with several log files. */
    std::string         work_directory;
        /**< Work directory used for temporary files. */
    
    }; // class PathManager
    
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PATH_MANAGER_H
