/** @file archive_database.cpp
 * Archive database.
 */

/*
 * begin        : Mon Feb 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "database/archive_database.h"
#include "filesys/file_ack_manager.h"
#include "filesys/file_system.h"
#include "status/exception.h"
#include "global/global.h"
#include "system/system.h"
#include "object/archive.h"
#include "registry/statistics.h"
#include "object/archive_manager.h"
#include "data/path_manager.h"
#include <map>
#include <sstream>


namespace Massiv {
namespace Core {


/**
 * ArchiveDatabase internal data structure.
 */
struct ArchiveDatabase::My
    {
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );
    
public:

    /**
     * File ack state of an archive file.
     */
    enum ArchiveAckState
        {
        ARCHIVE_BEING_ACKED,
            /**< The archive is beeing acked. */
            
        ARCHIVE_ACKED
            /**< The archive was successfully acked. */

        }; // ArchiveAckState

    /**
     * Stores miscenalleous information about an archive file.
     */
    struct ArchiveInfo
        {
    public:
    
        /** Constructor. */
        ArchiveInfo
            (
            ArchiveAckState _ack_state
            ) :
            ack_state( _ack_state )
            {
            }
            
    public:
        ArchiveAckState     ack_state;
            /**< Ack state of corresponding archive. */
        
        }; // ArchiveInfo
    
    /**
     * Map associating #ArchivationId with #ArchiveInfo.
     */
    typedef std::map
        <
        ArchivationId,
        ArchiveInfo,
        std::less< ArchivationId >
        > ArchiveInfoMap;

    /** ArchiveDatabase settings. */
    class ArchiveDatabaseSettings : public Settings
        {
    public:

        /** Constructor. */
        ArchiveDatabaseSettings() :
            Settings( "ArchiveDatabase" ),
            keep_archives_count( access_integer( "keep_archives_count", 5 ) ),
            milestone_divider( access_integer( "milestone_divider", 10 ) )
            {
            }

    public:
    
        Integer &   keep_archives_count;
            /**< How many archives counted from last valid archive are keep on disk.
             * Older archives that do not satisfy milestone criteria are removed from disk.
             */
        
        Integer &   milestone_divider;
            /**
             * Specifier milestone criteria that is used to detect whether
             * old archive should be removed from disk or not.
             * If archivation id of an archive modulo @a milestone_divider
             * is zero, the archive will never be automatically removed.
             */

        }; // ArchiveDatabaseSettings

public:

    /** Constructor. */
    My()
        {
        }

    /**
     * Create archive database directory structure.
     * The directory structure is created on native file system, not using
     * volume manager. This will ensure that if somebody mounts some
     * volume with archive directories, new archives won't be created
     * inside the volume but inside native file system.
     */
    void create_directory_structure() const
        {
        FileSystem & native_fs = Global::native_file_system();
        native_fs.create_directory( Global::path_manager().get_archive_directory() + "/regular" );
        }
    
    //
    // ArchivationId to/from file name.
    //
    
    /** Convert #ArchivationId to file name (without path). */
    std::string archivation_id_to_file_name
        (
        const ArchivationId & archivation_id
        ) const
        {
        return archivation_id.to_string();
        }
    
    /**
     * Create #ArchivationId from a file name (without path).
     * Returns invalid #ArchivationId if the file name doesn't
     * correspond to any valid #ArchivationId.
     */
    ArchivationId file_name_to_archivation_id
        (
        const std::string & file_name
        ) const
        {
        return ArchivationId::from_string( file_name );
        }

    /**
     * Convert #ArchivationId to file name (with path relative to base
     * directory). Path of the file corresponds to regular archive directory.
     */
    std::string archivation_id_to_regular_path_file_name
        (
        const ArchivationId & archivation_id
        ) const
        {
        std::string file_name = archivation_id_to_file_name( archivation_id );
        return Global::path_manager().get_archive_directory() + "/regular/" + file_name;
        }
    
    
    /**
     * Schedule ack creation for specified regular archive.
     */
    void ack_regular_archive
        (
        const ArchivationId & archivation_id
        )
        {
        FileAckManager &    file_ack_manager = Global::file_ack_manager();
        std::string         file_name = archivation_id_to_regular_path_file_name( archivation_id );
        
        // Schedule ack creation of the archive.
        file_ack_manager.add_job( file_name );
        
        // Insert the archive into regular archive list.
        ArchiveInfoMap::iterator it = regular_archives.insert
            (
            ArchiveInfoMap::value_type
                (
                archivation_id,
                ArchiveInfo( ARCHIVE_BEING_ACKED )
                )
            ).first;
        it->second.ack_state = ARCHIVE_BEING_ACKED;
        }


    /**
     * Remove ack file for specified regular archive.
     */
    void unack_regular_archive
        (
        const ArchivationId & archivation_id
        )
        {
        /** @todo There may be more archives for same archivation_id in
         * different mounted volumes! See FileAckManager::remove_ack().
         */
        FileAckManager &    file_ack_manager = Global::file_ack_manager();
        std::string         file_name = archivation_id_to_regular_path_file_name( archivation_id );
        
        // Unack the archive.
        file_ack_manager.remove_ack( file_name );
        
        // Remove the archive from regular archive list.
        regular_archives.erase( archivation_id );
        }


    /**
     * Enumerate all valid regular archives and insert them into
     * regular archive list.
     */
    void enumerate_regular_archives()
        {
        FileAckManager &    file_ack_manager = Global::file_ack_manager();
        std::string         regular_archives_dir_name = Global::path_manager().get_archive_directory() + "/regular";
        FileSystem &        fs = Global::native_file_system(); /** @todo FIXME Use volume manager? */
        FileIterator        it = fs.open_directory( regular_archives_dir_name, "*" );
        
        // Iterate through all files and find valid archives.
        for( ; it.is_valid(); it++ )
            {
            std::string     file_name;
            std::string     full_file_name;
            ArchivationId   archive_id;
            
            // Ignore directories.
            if( it.is_directory() )
                continue;
            
            // Get file name of the file.
            file_name = it.name();
            
            // Check if the file is an archive file.
            archive_id = file_name_to_archivation_id( file_name );
            if( archive_id.is_invalid() )
                continue;
            
            full_file_name = archivation_id_to_regular_path_file_name( archive_id );
                
            // Check if the file is acked.
            if( !file_ack_manager.is_file_acked( full_file_name ) )
                {
                Global::log_warning
                    (
                    Status::FACILITY_ARCHIVE_DATABASE,
                    Status::PRIORITY_HIGHEST,
                    msg_found_invalid_archive( full_file_name )
                    );
                
                continue;
                }
            
            // Try open the archive.
            if( !verify_archive( full_file_name ) )
                {
                Global::log_warning
                    (
                    Status::FACILITY_ARCHIVE_DATABASE,
                    Status::PRIORITY_HIGHEST,
                    msg_found_invalid_archive( full_file_name )
                    );
                
                continue;
                }
            
            // Insert the archive into regular archive list.
            regular_archives.insert
                (
                ArchiveInfoMap::value_type
                    (
                    archive_id,
                    ArchiveInfo( ARCHIVE_ACKED )
                    )
                );
            }
        
        // All regular archives enumerated.
        
        if( regular_archives.empty() )
            {
            Global::log_warning
                (
                Status::FACILITY_ARCHIVE_DATABASE,
                Status::PRIORITY_HIGHEST,
                msg_no_valid_archive_found()
                );
            }
        else
            {
            Global::log_info
                (
                Status::FACILITY_ARCHIVE_DATABASE,
                Status::PRIORITY_HIGHEST,
                msg_found_latest_valid_archive( get_latest_valid_archive() )
                );
            }
        }
    
    /** Simple archive verification. */
    bool verify_archive
        (
        const std::string & path_file_name
        )
        {
        bool failed = false;
        
        try 
            {
            Archive     archive( path_file_name, Archive::MODE_READ_ONLY );
            }
        catch( std::exception & )
            {
            failed = true;
            }
        
        return !failed;
        }
        
    /**
     * Get #ArchivationId of latest valid archive file.
     * If there is no valid archive, invalid #ArchivationId is returned.
     */
    ArchivationId get_latest_valid_archive() const
        {
        if( regular_archives.empty() )
            {
            ArchivationId   archivation_id;
            hope( archivation_id.is_invalid() );
            return archivation_id;
            }
        else
            {
            ArchiveInfoMap::const_iterator last = regular_archives.end();
            last--;
            const ArchivationId & archivation_id = last->first;
            hope( !archivation_id.is_invalid() );
            return archivation_id;
            }
        }
    
    /** Remove archives that are too old. */
    void remove_obsolete_archives()
        {
        unsigned int    limit = settings.keep_archives_count;
        UInt32          milestone_divider = settings.milestone_divider;
        
        if( limit < 0 )
            {
            return;
            }
        
        if( regular_archives.size() <= limit )
            {
            return;
            }
            
        // Note: We can't use reverse iterators here, because
        // delete_archive() removes items from regular_archives
        // and removing an item from std::map can invalidate
        // reverse iterator pointing before the item.
        
        bool not_yet_logged = true ;
        ArchiveInfoMap::iterator        it = regular_archives.end();
        hope( it != regular_archives.begin() );
        it--;
        
        for( ; ; )
            {
            if( limit == 0 )
                {
                bool            quit = ( it == regular_archives.begin() );
                ArchivationId   archivation_id = it->first;
                
                if( !quit )
                    {
                    it--;
                    }
                
                // Should be delete the archive?
                UInt32 order = archivation_id.get_order();
                if( ( order % milestone_divider ) != 0 )
                    {
                    // We will remove at least one archive.
                    // Log a message before the first archive is removed.
                    if( not_yet_logged )
                        {
                        Global::log_info
                            (
                            Status::FACILITY_ARCHIVE_DATABASE,
                            Status::PRIORITY_NORMAL,
                            "Removing obsolete archives..."
                            );

                        not_yet_logged = false ;
                        }
                    
                    delete_archive( archivation_id );
                    }
                
                if( quit )
                    {
                    break;
                    }
                }
            else
                {
                // Skip first 'limit' archives.
                
                if( it == regular_archives.begin() )
                    {
                    break;
                    }
    
                it--;
                limit--;
                }
            }
        }
    
    /**
     * Delete archive.
     */
    void delete_archive
        (
        const ArchivationId & archivation_id
        )
        {
        FileSystem & fs = Global::native_file_system();
        ArchiveManager & archive_manager = Global::archive_manager();
        ArchiveInfoMap::iterator it = regular_archives.find( archivation_id );
        if( it != regular_archives.end() )
            {
            // Do not delete archives beeing used.
            if( archive_manager.get_current_archivation_id() != archivation_id )
                {
                std::string file_name = archivation_id_to_regular_path_file_name( archivation_id );
                std::ostringstream ss;
                
                ss  << "Removing '" << file_name << "' archive file.";
                
                Global::log_info
                    (
                    Status::FACILITY_ARCHIVE_DATABASE,
                    Status::PRIORITY_HIGH,
                    ss.str()
                    );
                
                try 
                    {
                    fs.delete_file( file_name );
                    }
                catch( std::exception & )
                    {
                    }
                
                // This will remove the archive from regular_archives.
                unack_regular_archive( archivation_id );
                }
            }
        }

    static std::string msg_found_invalid_archive
        (
        const std::string & path_file_name
        )
        {
        std::ostringstream ss;
        ss  << "Found invalid archive '" << path_file_name << "'.";
        return ss.str();
        }
    
    
    static std::string msg_no_valid_archive_found()
        {
        return "No valid archives found.";
        }
    
    
    static std::string msg_found_latest_valid_archive
        (
        const ArchivationId & arch_id
        )
        {
        std::ostringstream ss;
        ss  << "Latest valid archive found: " << arch_id << ".";
        return ss.str();
        }


public:

    ArchiveDatabaseSettings settings;
        /**< Archive database settings. */

    ArchiveInfoMap          regular_archives;
        /**< List of all valid regular archives. */

    }; // ArchiveDatabase::My

    
//-----------------------------------------------------------------------------


// ArchiveDatabase
// ~~~~~~~~~~~~~~~


ArchiveDatabase::ArchiveDatabase() :    
    my( new My() )
    {
    }


ArchiveDatabase::~ArchiveDatabase()
    {
    }
    
    
void ArchiveDatabase::initialize
    (
    NodeId::Type    node_type
    )
    {
    hope( my->regular_archives.empty() );
    
    if( node_type != NodeId::TYPE_SERVER )
        {
        return;
        }

    try 
        {
        // Initialize directory structure.
        my->create_directory_structure();
        
        // Enumerate valid regular archives.
        my->enumerate_regular_archives();
        }
    catch( Exception & e )
        {
        // Cleanup.
        my->regular_archives.clear();
        
        // Rethrow.
        std::ostringstream ss;
        ss  << "Unable to initialize archive database. "
            << e.what();
        throw ExArchiveDatabaseInitializationError( ss.str() );
        }
    }


void ArchiveDatabase::next_tick()
    {
    // Remove obsolete archives each period seconds.
    SystemTime          period = SystemTime( 20.0 );
    static SystemTime   last_time;
    SystemTime          current_time = Global::system_time();
    
    if( ( last_time + period ).is_older_than( current_time ) )
        {
        my->remove_obsolete_archives();
        last_time = current_time;
        }
    }
    
    
Archive * ArchiveDatabase::create_regular_archive
    (
    const ArchivationId &   archivation_id
    )
    {
    std::string file_name = my->archivation_id_to_regular_path_file_name( archivation_id );
    std::string volume_name = archivation_id.to_string(); /** @todo FIXME */
    
    // If there is already an archive for this archivation_id, invalidate it.
    my->unack_regular_archive( archivation_id );
    
    // Create new archive file.
    return new Archive( file_name, volume_name );
    }


void ArchiveDatabase::commit_regular_archive
    (
    const ArchivationId &   archivation_id,
    Archive *               archive
    )
    {
    hope( archive != NULL );
    
    // Close the archive file.
    delete archive;
    
    // Schedule ack creation and insert the archive into regular archive list.
    my->ack_regular_archive( archivation_id );
    }


ArchivationId ArchiveDatabase::get_latest_valid_archive() const
    {
    return my->get_latest_valid_archive();
    }
    
    
std::string ArchiveDatabase::archivation_id_to_file_name
    (
    const ArchivationId & archivation_id
    ) const
    {
    return my->archivation_id_to_file_name( archivation_id );
    }
    

std::string ArchiveDatabase::archivation_id_to_regular_path_file_name
    (
    const ArchivationId & archivation_id
    ) const
    {
    return my->archivation_id_to_regular_path_file_name( archivation_id );
    }


} // namespace Massiv::Core
} // namespace Massiv
