/** @file archive_database.h
 * Archive database.
 */

/*
 * begin        : Mon Feb 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ARCHIVE_DATABASE_H
#define MASSIV_ARCHIVE_DATABASE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_ARCHIVATION_ID_H
#include "object/archivation_id.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#include <string>
#include <memory> // auto_ptr


namespace Massiv {
namespace Core {


class Archive;


/**
 * ArchiveDatabase manages all archive files.
 *
 * All regular archives are stored in one directory 'archives/regular'.
 * Archives can be valid or invalid. During initialization ArchiveDatabase
 * enumerates all valid archive files. When checking whether an archive file
 * is valid, ArchiveDatabase uses #FileAckManager to detect whether the
 * file has its ack file. If ack file exists, the ArchiveDatabase tries
 * to open the archive file and do some verification. After this, the archive
 * file is inserted into valid regular archive list.
 *
 * ArchiveDatabase also manages acking of new archive files created
 * during at runtime. New archive object must be created by
 * calling @c create_regular_archive() method and commited by
 * calling @c commit_regular_archive() method.
 *
 * @todo FIXME CZ - Zminka, ze se archivuje vzdy na nativni file system.
 *
 * @ingroup archivation
 */
class ArchiveDatabase
    {
    NO_COPY_CONSTRUCTOR( ArchiveDatabase );
    NO_ASSIGNMENT_OPERATOR( ArchiveDatabase );

public:

    /** Constructor. */
    ArchiveDatabase();

    /** Destructor. */
    ~ArchiveDatabase();

    /**
     * Initialize the database.
     * @throw ExArchiveDatabaseInitializationError Initialization failed.
     */
    void initialize
        (
        NodeId::Type    node_type
        );

    /**
     * Advance archive database.
     * Archive database deletes too old archives.
     */
    void next_tick();

public:

    /**
     * Create regular archive for specified #ArchivationId.
     * @return Pointer to archive file. The caller claims ovnership
     *      of the archive object. The caller is responsible for
     *      destroying the returned archive object by simple deleting
     *      it if the archivation failed and the archive should not
     *      become valid archive file, or the caller must commit
     *      the archive by calling @c commit_regular_archive() passing
     *      returned object as an argument if the archivation was successfull
     *      and the archive should become valid archive file.
     */
    Archive * create_regular_archive
        (
        const ArchivationId &   archivation_id
        );

    /**
     * Commit regular archive file which was previously created
     * by calling @c create_regular_archive().
     * #ArchiveDatabase claims ovnership of the archive and caller
     * should never access the arhive object any more.
     */
    void commit_regular_archive
        (
        const ArchivationId &   archivation_id,
        Archive *               archive
        );

    /**
     * Get #ArchivationId of latest valid archive file found.
     * @pre The ArchiveDatabase is properly initialized.
     */
    ArchivationId get_latest_valid_archive() const;

    /**
     * Convert #ArchivationId to file name (without path).
     */
    std::string archivation_id_to_file_name
        (
        const ArchivationId & archivation_id
        ) const;

    /**
     * Convert #ArchivationId to file name (with path relative to base
     * directory). Path of the file corresponds to regular archive directory.
     */
    std::string archivation_id_to_regular_path_file_name
        (
        const ArchivationId & archivation_id
        ) const;

private:

    // Internal data structure. Forward declaration.
    struct My;

    std::auto_ptr< My >     my;
        /**< Internal data structure. Always valid pointer. */

    }; // ArchiveDatabase


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVE_DATABASE_H
