/** @file node_database.cpp
 * Node database.
 * @bug Poradne tady pouzit GlobalCallbacks.
 */

/*
 * begin        : Sun 12 2003
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "database/node_database.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "registry/registry.h"
#include "status/exception.h"
#include "utility/debug.h"
#include "global/global_callbacks.h"
#include "data/data_manager.h"
#include "serializer/serializer.h"
#include "utility/text_reader.h"
#include "utility/hash.h"
#include "utility/quote_string.h"
#include <memory>
#include <sstream>
#include <vector>


namespace Massiv {
namespace Core {


// NodeDatabase::NodeData
// ~~~~~~~~~~~~~~~~~~~~~~


NodeDatabase::NodeData::NodeData()
    {
    }


NodeDatabase::NodeData::NodeData
    (
    const NodeId &          _node_id,
    const std::string &     _rsa_public_key,
    const ServerAddress &   _server_address,
    const ObjectId &        _account_object_id,
    const ObjectId &        _proxy_node_object_id,
    const std::string &     _name
    ) :
    node_id( _node_id ),
    rsa_public_key( _rsa_public_key ),
    server_address( _server_address ),
    account_object_id( _account_object_id ),
    proxy_node_object_id( _proxy_node_object_id ),
    name( _name )
    {
    }


NodeDatabase::NodeData::NodeData
    (
    const NodeData & other
    ) :
    node_id( other.node_id ),
    rsa_public_key( other.rsa_public_key ),
    server_address( other.server_address ),
    account_object_id( other.account_object_id ),
    proxy_node_object_id( other.proxy_node_object_id ),
    name( other.name )
    {
    }


NodeDatabase::NodeData & NodeDatabase::NodeData::operator=
    (
    const NodeData & other
    )
    {
    node_id                 = other.node_id;
    rsa_public_key          = other.rsa_public_key;
    server_address          = other.server_address;
    account_object_id       = other.account_object_id;
    proxy_node_object_id    = other.proxy_node_object_id;
    name                    = other.name;
    return *this;
    }


// Implementation of NodeDatabase::Enumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

NodeDatabase::Enumerator::Enumerator()
    {
    }


NodeDatabase::Enumerator::~Enumerator()
    {
    }


// NodeDatabase::My
// ~~~~~~~~~~~~~~~~

struct NodeDatabase::My
    {
private:

    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );
    
public:

    /** Hash map used to store NodeData. */
    typedef ext::hash_map
                <
                NodeId,
                NodeData,
                NodeId::Hash,
                NodeId::EqualKey
                > NodeDataSet;

    /** Flags used when reading NodeData. */
    enum RequireFlags
        {
        REQUIRE_ADDRESS     = 0x01,
            /**< Address (host address and port) must be specified. */
        REQUIRE_ACCOUNT     = 0x02
            /**< Account object id must be specified. */
        };
public:

    /** Constructor. */
    My();
    
    /** Destructor. */
    ~My();

public:

    /** Clear everything. Reset to @c NodeDatabase::MODE_NONE mode. */
    void clear();
    
    void switch_to_mode
        (
        NodeDatabase::Mode new_mode
        );
    
    bool refresh
        (
        bool force_refresh
        );

private:

    void switch_to_simple_mode();

    void switch_to_full_mode();

    void initialize_static_data();

public:

    void read_local_node_data();
    
private:

    void read_anonymous_node_data();
    
    void read_master_data_service_node_data();
    
    void initialize_data_objects();

    NodeId extract_node_id_from_cvar
        (
        const ConfigVariableString & cvar
        );
        
    void read_node_data_set
        (
        NodeDataSet & node_data_set,
        DataObject *  data_object
        );
        
    void read_node_data_set_from_registry
        (
        NodeDataSet &               node_data_set,
        Registry &                  registry,
        Registry::RegistryNode *    base_registry_node
        );
    
    /**
     * Read NodeData from a registry node.
     * @param require_flags Combination of #RequireFlags. Specified
     *      which part of NodeData must be spefied in the registry
     *      and which are optional. Node id and rsa public key are
     *      always required, node name is always optional.
     */
    NodeData read_node_data_from_registry
        (
        Registry &                  registry,
        Registry::RegistryNode *    registry_node,
        unsigned int                require_flags
        );
        
    NodeData read_node_data_from_cvars
        (
        const std::string &     serialized_node_id,
        const std::string &     rsa_public_key,
        const std::string &     address,
        Int32                   port,
        const std::string &     serialized_account_object_id,
        const std::string &     name
        );

public:
    //
    // Mode.
    //
    
    NodeDatabase::Mode              mode;
        /**< Current database mode. */

    //
    // Data objects used to update node lists.
    //
    
    std::auto_ptr< DataObject >     public_node_list_data_object;
        /**< Node database data object containing public node list. */
        
    std::auto_ptr< DataObject >     private_node_list_data_object;
        /**< Node database data object containing private node list. */
    
    //
    // Node database.
    //
    
    NodeDataSet         nodes;
        /**< List of all nodes read from the database data objects.
         * This is the main part of node database and refering to "node database"
         * often means refering to this set.
         * Static nodes (@c static_node) and local node are automatically inserted into
         * this set.
         */
    
    //
    // Static data.
    //
        
    NodeDataSet         static_nodes;
        /**< List of all nodes read from registry during initialization.
         * Contains at least master data service node.
         * @warning Local node data are not included in this set.
         */
         
    NodeData            local_node_data;
        /**< Node data for the local node. Read when system connects into the
             simulation (the data is incomplete at that point) and updated
             at node database refresh (the data will become complete). */
    
    std::string         local_node_rsa_private_key;
        /**< Rsa private key for local node. */
    
    bool                anonymous_local_node;
        /**< True if local node data was not read yet.
         * Node database uses anonymous node data instead.
         */
    
    NodeData            anonymous_node_data;
        /**< Node data for all anonymous nodes. Anonymous nodes share
         * one common node data.
         */
    
    std::string         anonymous_rsa_private_key;
        /**< Anonymous node rsa private key. */

    }; // NodeDatabase::My


//-----------------------------------------------------------------------------    


NodeDatabase::My::My() :
    mode( NodeDatabase::MODE_NONE ),
    anonymous_local_node( true )
    {
    }

    
NodeDatabase::My::~My()
    {
    }
    
    
void NodeDatabase::My::clear()
    {
    public_node_list_data_object.reset();
    private_node_list_data_object.reset();
    
    nodes.clear();
    static_nodes.clear();
    
    local_node_data = NodeData();
    local_node_rsa_private_key.clear();
    
    anonymous_local_node = true;
    anonymous_node_data = NodeData();
    anonymous_rsa_private_key.clear();

    mode = NodeDatabase::MODE_NONE;
    }


void NodeDatabase::My::switch_to_mode
    (
    NodeDatabase::Mode new_mode
    )
    {
    if( new_mode == mode )
        {
        return;
        }

    switch( new_mode )
        {
        case NodeDatabase::MODE_SIMPLE:
            {
            if( mode != MODE_NONE )
                {
                // Switching from full mode. Clear all data.
                clear();
                }
            
            switch_to_simple_mode();
            break;
            }
            
        case NodeDatabase::MODE_FULL:
            {
            hope( mode == MODE_SIMPLE );
            
            switch_to_full_mode();
            break;
            }
            
        case NodeDatabase::MODE_NONE:
            {
            // Clear all data.
            clear();
            mode = MODE_NONE;
            break;
            }
        }
    }
    

bool NodeDatabase::My::refresh
    (
    bool force_refresh
    )
    {
    // We may not check 'mode' value here (for example to verify if the 'mode'
    // is MODE_FULL) because refresh() is called by switch_to_full_mode()
    // while the 'mode' is still set to MODE_SIMPLE.
    
    // Do nothing if database data objects are not opened.
    if( public_node_list_data_object.get() == NULL ||
        private_node_list_data_object.get() == NULL )
        {
        return false;
        }
    
    // Do nothing if there is not newer version of database data objects available.
    if( !force_refresh )
        {
        // If newer version is available set force_refresh to true.
        force_refresh = public_node_list_data_object->is_data_updated() ||
                        private_node_list_data_object->is_data_updated();
        }
        
    if( !force_refresh )
        {
        return false;
        }
        
    // Log.
        {
        Global::log_info
            (
            Status::FACILITY_NODE_DATABASE,
            Status::PRIORITY_NORMAL,
            "Refreshing node database..."
            );
        }
    
    try 
        {
        NodeDataSet                 public_database;
        NodeDataSet                 private_database;
        NodeDataSet                 new_database;
        NodeDataSet::const_iterator it;
        
        public_node_list_data_object->refresh();
        private_node_list_data_object->refresh();
        
        // Read public node list.
        read_node_data_set( public_database, public_node_list_data_object.get() );
        
        // Read private node list.
        read_node_data_set( private_database, private_node_list_data_object.get() );
        
        // Update main database.
        // Note: Doing it this way will ensure that if an exception is thrown
        // from NodeDataSet::insert() then this->nodes will stay unmodified.
        new_database.insert( public_database.begin(), public_database.end() );
        new_database.insert( private_database.begin(), private_database.end() );
        
        // Insert static node data into new database.
        new_database.insert( static_nodes.begin(), static_nodes.end() );
        
        // Update @c local_node_data from the database so that it would
        // become complete (all fields will be properly filled).
        // Insert an incomplete entry, if corresponding data can not be
        // found in the database.
        hope( !anonymous_local_node );
        it = new_database.find( local_node_data.get_node_id() );
        if( it == new_database.end() )
            {
            // Nothing known about the local node. Insert the incomplete
            // entry and issue a warning.
            // This may happen on client nodes if they did not download
            // private node data set.
            
            std::ostringstream ss;
            ss << "Nothing known about the local node (" << 
                local_node_data.get_node_id() << "). Such a node is not "
                "registered in the database. Inserting an incomplete "
                "entry.";
            
            Global::log_info
                (
                Status::FACILITY_NODE_DATABASE,
                Status::PRIORITY_HIGH,
                ss.str()
                );
            
            new_database.insert( NodeDataSet::value_type( 
                local_node_data.get_node_id(), local_node_data ) );
            }
        else
            {
            // Update missing local node data fields that were not read by
            // @c read_local_node_data().
            
            const NodeData & data = it->second;

            if( local_node_data.get_rsa_public_key() != data.get_rsa_public_key() )
                {
                std::ostringstream ss;
                ss << "RSA public key of the local node does not match the "
                    "corresponding entry in the node data object:\n"
                    "supplied key = " << quote_const_string( 
                    local_node_data.get_rsa_public_key() ) << "\n"
                    "key in the database = " << quote_const_string(
                    data.get_rsa_public_key() );

                throw ExNodeDatabaseError( ss.str() );
                }

            local_node_data = NodeData
                (
                local_node_data.get_node_id(),
                data.get_rsa_public_key(),
                data.get_server_address(),
                data.get_account_object_id(),
                data.get_proxy_node_object_id(),
                data.get_name()
                ); 
            }
        
        // Commit.
        nodes = new_database;
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error refreshing node database. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }

    return true;
    }

    
void NodeDatabase::My::switch_to_simple_mode()
    {
    Global::log_debug
        (
        Status::FACILITY_NODE_DATABASE,
        Status::PRIORITY_NORMAL,
        "Node database switching to simple mode..."
        );

    try 
        {
        anonymous_local_node = true;

        // Initialize static data.
        initialize_static_data();
        
        // Set anonymous local node data.
        local_node_data = anonymous_node_data;
        local_node_rsa_private_key = anonymous_rsa_private_key;

        // Insert static nodes into node database.
        nodes = static_nodes;
        }
    catch( std::exception & e )
        {
        // Cleanup everything.
        clear();

        std::ostringstream ss;
        ss  << "Error switching node database to simple mode. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
        
    mode = NodeDatabase::MODE_SIMPLE;
    }
    
    
void NodeDatabase::My::switch_to_full_mode()
    {
    hope( !anonymous_local_node );
    
    Global::log_debug
        (
        Status::FACILITY_NODE_DATABASE,
        Status::PRIORITY_NORMAL,
        "Node database switching to full mode..."
        );
    
    try 
        {
        // Initialize data objects used to refresh node database.
        initialize_data_objects();
            
        // Do initial database refresh.
        refresh( true );
        }
    catch( std::exception & e )
        {
        // Cleanup everything.
        clear();

        std::ostringstream ss;
        ss  << "Error switching node database to full mode. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }

    mode = NodeDatabase::MODE_FULL;
    }


void NodeDatabase::My::initialize_static_data()
    {
    // Read anonymous node data.
    read_anonymous_node_data();
    
    // Read master data service node data.
    read_master_data_service_node_data();
    }


void NodeDatabase::My::read_local_node_data()
    {
    Registry &                  registry = Global::registry();
    std::string                 registry_node_path = "Settings/NodeDatabase/local_node";
    Registry::RegistryNode *    registry_node;
    
    // When reading local node, only node id, rsa public key and
    // rsa private key are required.
    // At this point local node data will be incomplete and the rest
    // will be read at the next node database refresh from the node data
    // objects if available. Because of this the data is not inserted into
    // the database yet.

    unsigned int                require_flags = 0;
    NodeData                    node_data;
    NodeDataSet::const_iterator it;
    
    try 
        {
        // Access registry node.
        registry_node = registry.find_node_for_path( registry_node_path );
        
        // Read local node data.
        node_data = read_node_data_from_registry( registry, registry_node, require_flags );

        // Read additional data specific for the local node.
        { // sob
        // Access private rsa key.
        ConfigVariable &        rsa_private_key_cvar = registry.access_variable( "rsa_private_key", registry_node );
        ConfigVariableString &  rsa_private_key_str  = rsa_private_key_cvar.as_string();

        // Read private rsa key.
        local_node_rsa_private_key = rsa_private_key_str;
        } // eob
        }
    catch( ExRegistryError & e )
        {
        std::ostringstream ss;
        ss  << "Unable to read local node data from registry. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
    
    // Store it.
    local_node_data = node_data;
    
    // We are not anonymous any more.
    anonymous_local_node = false;
    }


void NodeDatabase::My::read_anonymous_node_data()
    {
    Registry &                  registry = Global::registry();
    std::string                 registry_node_path = "Settings/NodeDatabase/anonymous_node";
    Registry::RegistryNode *    registry_node;
    // When reading anonymous node, only node id, rsa public key and
    // rsa private key are required. The rest will be set to zeros.
    unsigned int                require_flags = 0;
    NodeData                    node_data;

    try 
        {
        // Access registry node.
        registry_node = registry.find_node_for_path( registry_node_path );

        // Read anonymous node data.
        node_data = read_node_data_from_registry( registry, registry_node, require_flags );
        
        // Read anonymous private rsa key.
        { // sob
        // Access private rsa key.
        ConfigVariable &        rsa_private_key_cvar = registry.access_variable( "rsa_private_key", registry_node );
        ConfigVariableString &  rsa_private_key_str  = rsa_private_key_cvar.as_string();

        // Read private rsa key.
        anonymous_rsa_private_key = rsa_private_key_str;
        } // eob
        }
    catch( ExRegistryError & e )
        {
        std::ostringstream ss;
        ss  << "Unable to read anonymous node data from registry. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }

    // Store it.
    anonymous_node_data = node_data;
    }
    

void NodeDatabase::My::read_master_data_service_node_data()
    {
    Registry &                  registry = Global::registry();
    std::string                 registry_node_path = "Settings/NodeDatabase/master_data_node";
    Registry::RegistryNode *    registry_node;
    // For master data service node id and rsa public key and address are required.
    unsigned int                require_flags = REQUIRE_ADDRESS;
    NodeData                    node_data;

    try 
        {
        // Access registry node.
        registry_node = registry.find_node_for_path( registry_node_path );

        // Read node data.
        node_data = read_node_data_from_registry( registry, registry_node, require_flags );
        }
    catch( ExRegistryError & e )
        {
        std::ostringstream ss;
        ss  << "Unable to read master data service node data from registry. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
    
    // Store it.
    static_nodes.insert( NodeDataSet::value_type( node_data.get_node_id(), node_data ) );
    }
    
    
void NodeDatabase::My::initialize_data_objects()
    {
    DataManager &               data_manager = Global::data_manager();
    std::auto_ptr< DataObject > public_node_list_do;
    std::auto_ptr< DataObject > private_node_list_do;
    std::string                 public_node_list_do_name = "config/nodes/public";
    std::string                 private_node_list_do_name = "config/nodes/private";

    // Get data objects for public and private node lists.
    public_node_list_do.reset( data_manager.get_data_object( public_node_list_do_name ) );
    private_node_list_do.reset( data_manager.get_data_object( private_node_list_do_name ) );
    hope( public_node_list_do.get() != NULL );
    hope( private_node_list_do.get() != NULL );
    
    public_node_list_data_object  = public_node_list_do;
    private_node_list_data_object = private_node_list_do;
    }
    

NodeId NodeDatabase::My::extract_node_id_from_cvar
    (
    const ConfigVariableString & cvar
    )
    {
    NodeId              node_id;
    std::string         str = cvar;
    std::istringstream  iss( str );
    TextReader          tr( iss );
    
    tr.next_line();
    node_id.text_read( tr, Serializer::Description() );
    
    return node_id;
    }
    
    
void NodeDatabase::My::read_node_data_set
    (
    NodeDataSet & node_data_set,
    DataObject *  data_object
    )
    {
    hope( data_object );
    
    try 
        {
        Registry                    tmp_registry;
        Registry::RegistryNode *    registry_node;
    
        // Read the file like a registry config file.
        data_object->open();
        tmp_registry.load_config_file( data_object->data );
        data_object->close();
        
        // Get root node of the registry.
        registry_node = tmp_registry.find_node_for_path( "" );
        
        // Read node data from the temporary registry.
        read_node_data_set_from_registry( node_data_set, tmp_registry, registry_node );
        }
    catch( std::exception & e )
        {
        data_object->close();
        std::ostringstream ss;
        ss  << "Error reading node data from \""
            << data_object->get_name() << "\" data object. "
            << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
    }
    
    
void NodeDatabase::My::read_node_data_set_from_registry
    (
    NodeDataSet &               node_data_set,
    Registry &                  registry,
    Registry::RegistryNode *    base_registry_node
    )
    {
    hope( base_registry_node != NULL );
    
    Registry::SubnodeList::const_iterator   subnode_it;
    const Registry::SubnodeList &           subnodes = base_registry_node->subnodes;
    
    // For each subnode of the base_node...
    for( subnode_it = subnodes.begin(); subnode_it != subnodes.end(); ++subnode_it )
        {
        // ... read NodeData from the subnode.
        Registry::RegistryNode * registry_node = subnode_it->second;
        hope( registry_node != NULL );
        
        // When reading list of node data, full node data (except node name) is required.
        unsigned int            require_flags = REQUIRE_ADDRESS | REQUIRE_ACCOUNT;
        NodeData                node_data = read_node_data_from_registry( registry, registry_node, require_flags );
        
        // Insert the NodeData into the node_data_set.
        node_data_set.insert( NodeDataSet::value_type( node_data.get_node_id(), node_data ) );
        }
    }


NodeDatabase::NodeData NodeDatabase::My::read_node_data_from_registry
    (
    Registry &                  registry,
    Registry::RegistryNode *    registry_node,
    unsigned int                require_flags
    )
    {
    hope( registry_node != NULL );
    
    // Node id and rsa public key are always required.
    // Node name is always optional.
    bool require_address = flags_on( require_flags, REQUIRE_ADDRESS );
    bool require_account = flags_on( require_flags, REQUIRE_ACCOUNT );
    
    // Optional part.
    std::string     address;
    Int32           port;
    std::string     account_object_id;
    std::string     node_name;
    bool            invalid_address;

    // Access all cvars.
    
    // Required.
    ConfigVariable & rsa_public_key_cvar    = registry.access_variable( "rsa_public_key", registry_node );
    ConfigVariable & node_id_cvar           = registry.access_variable( "node_id", registry_node );
    
    // Read address.
    try 
        {
        invalid_address = false;
        ConfigVariable & address_cvar       = registry.access_variable( "address", registry_node );
        ConfigVariable & port_cvar          = registry.access_variable( "port", registry_node );
        invalid_address = true;
        
        port    = port_cvar.as_integer();
        address = address_cvar.as_string();
        }
    catch( ExRegistryError & )
        {
        if( require_address || invalid_address )
            {
            throw;
            }
        else
            {
            // Set default values.
            address = "127.0.0.1";
            port    = 0;

            // Fall through.
            }
        }

    // Read account object id.
    try 
        {
        ConfigVariable & account_object_id_cvar = registry.access_variable( "account_object", registry_node );
        
        account_object_id = account_object_id_cvar.as_string();
        }
    catch( ExRegistryError & )
        {
        if( require_account )
            {
            throw;
            }
        else
            {
            // Set default value.
            account_object_id = "[ null ]";
            
            // Fall through.
            }
        }
    
    // Read node name.
    try 
        {
        ConfigVariable & name_cvar  = registry.access_variable( "name", registry_node );
        
        node_name = name_cvar.as_string();
        }
    catch( ExRegistryError & )
        {
        // Set default value.
        node_name = "unnamed";
        
        // Fall through.
        }
    
    ConfigVariableString &  rsa_public_key_cvar_str    = rsa_public_key_cvar.as_string();
    ConfigVariableString &  node_id_cvar_str           = node_id_cvar.as_string();
    
    // Extract node data from the cvars.
    return read_node_data_from_cvars
        (
        node_id_cvar_str,
        rsa_public_key_cvar_str,
        address,
        port,
        account_object_id,
        node_name
        );
    }
    

NodeDatabase::NodeData NodeDatabase::My::read_node_data_from_cvars
    (
    const std::string &     serialized_node_id,
    const std::string &     rsa_public_key,
    const std::string &     address,
    Int32                   port,
    const std::string &     serialized_account_object_id,
    const std::string &     name
    )
    {
    NodeId          node_id;
    ServerAddress   host_address;
    ObjectId        account_object_id;
    ObjectId        proxy_node_object_id;
    
    
    // Extract node id.
    try 
        {
        node_id = extract_node_id_from_cvar( serialized_node_id );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error extracting node id. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
    
    // Extract host address.
    try 
        {
        host_address = ServerAddress( address, port );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error extracting host address. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
    
    // Extract account object id.
    try 
        {
        std::istringstream iss( serialized_account_object_id );
        TextReader tr( iss );
        tr.next_line();
        account_object_id.text_read( tr, Serializer::Description() );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error extracting account object id. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
        
    // Create proxy node object id.
    try 
        {
        proxy_node_object_id = ObjectId( node_id );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error creating proxy node object id. " << e.what();
        throw ExNodeDatabaseError( ss.str() );
        }
        
    // Create node data.
    return NodeData
        (
        node_id,
        rsa_public_key,
        host_address,
        account_object_id,
        proxy_node_object_id,
        name
        );
    }

    
//-----------------------------------------------------------------------------    
    

// Implementation of NodeDatabase
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

NodeDatabase::NodeDatabase() :
    my( new My() )
    {
    }


NodeDatabase::~NodeDatabase()
    {
    }


void NodeDatabase::load_local_node_config()
    {
    hope( my->mode == MODE_SIMPLE );

    // Read local node data.
    my->read_local_node_data();
    
    /// @todo FIXME Temporary hack.
        {
        std::ostringstream ss;
        ss  << "**** Local node id: " << my->local_node_data.get_node_id();
        Global::log_debug
            (
            Status::FACILITY_NODE_DATABASE,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
    
    // Notify core.
    GlobalCallbacks::node_database_changed();
    }

    
void NodeDatabase::switch_to_mode
    (
    NodeDatabase::Mode new_mode
    )
    {
    // Switch mode.
    my->switch_to_mode( new_mode );
    
    // Notify core.
    GlobalCallbacks::node_database_changed();
    }
    
    
bool NodeDatabase::refresh
    (
    bool force_refresh
    )
    {
    bool refreshed = my->refresh( force_refresh );

    if( refreshed )
        {
        // Notify core.
        GlobalCallbacks::node_database_changed();
        }

    return refreshed;
    }
    
    
void NodeDatabase::next_tick()
    {
    refresh();
    }


void NodeDatabase::enumerate_all_nodes
    (
    Enumerator & enumerator
    )
    {
    My::NodeDataSet::const_iterator it;
    
    for( it = my->nodes.begin(); it != my->nodes.end(); ++it )
        {
        const NodeData & node_data = it->second;
        enumerator.operator()( node_data );
        }
    }


void NodeDatabase::enumerate_nodes
    (
    NodeId::Type    node_type,
    Enumerator &    enumerator
    )
    {
    My::NodeDataSet::const_iterator it;

    for( it = my->nodes.begin(); it != my->nodes.end(); ++it )
        {
        const NodeData & node_data = it->second;
        const NodeId &   node_id = node_data.get_node_id();
        
        if( node_id.get_type() == node_type )
            {
            enumerator.operator()( node_data );
            }
        }
    }
    

const NodeId & NodeDatabase::get_local_node_id() const
    {
    hope( my->mode == MODE_SIMPLE || my->mode == MODE_FULL );
    return my->local_node_data.get_node_id();
    }


const std::string & NodeDatabase::get_local_node_rsa_public_key() const
    {
    hope( my->mode == MODE_SIMPLE || my->mode == MODE_FULL );
    return my->local_node_data.get_rsa_public_key();
    }


const std::string & NodeDatabase::get_local_node_rsa_private_key() const
    {
    hope( !my->local_node_rsa_private_key.empty() );
    return my->local_node_rsa_private_key;
    }


const NodeDatabase::NodeData & NodeDatabase::get_node_data
    (
    const NodeId & node
    ) const
    {
    const NodeData *  node_data;
    get_node_data( node, &node_data );
    
    if( node_data )
        {
        return *node_data;
        }
    else
        {
        std::ostringstream ss;
        ss  << "Unable to get node data from node database. "
            << "The node " << node << " is not registered in the database.";
        
        throw ExNodeNotFoundInNodeDatabase( ss.str() );
        }
    }
    
    
void NodeDatabase::get_node_data
    (
    const NodeId &      node,
    const NodeData **   output
    ) const
    {
    hope( output != NULL );
    
    // Check if node is local node.
    if( my->local_node_data.get_node_id() == node )
        {
        ( *output ) = &( my->local_node_data );
        return;
        }
    
    // If requesting data for anonymous node, use anonymous node data.
    if( node.is_type( NodeId::TYPE_ANONYMOUS ) )
        {
        ( *output ) = &( my->anonymous_node_data );
        return;
        }
    
    My::NodeDataSet::const_iterator it = my->nodes.find( node );
    
    if( it != my->nodes.end() )
        {
        const NodeData & node_data = it->second;
        ( *output ) = &node_data;
        }
    else
        {
        ( *output ) = NULL;
        }
    }


ServerAddress NodeDatabase::get_node_server_address
    (
    const NodeId & node
    ) const
    {
    /** @todo FIXME Service node is not "server" node. Rename
     * server_address to host_address etc? */
    hope( node.get_type() == NodeId::TYPE_SERVER || node.get_type() == NodeId::TYPE_SERVICE );
    
    const NodeData &        node_data = get_node_data( node );
    const ServerAddress &   sa = node_data.get_server_address();
        
    return sa;
    }


} // namespace Massiv::Core
} // namespace Massiv
