/** @file node_database.h
 * Node database.
 */

/*
 * begin        : Sun Jan 12 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_DATABASE_H
#define MASSIV_NODE_DATABASE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_SERVERADDRESS_H
#include "net/serveraddress.h"
#endif

#ifndef MASSIV_CLASS_TYPE_ID_H
#include "object/class_type_id.h"
#endif

#include <string>
#include <memory> // auto_ptr


namespace Massiv {
namespace Core {

/**
 * NodeDatabase.
 *
 * NodeDatabase keeps up-to-date list of nodes in one massiv system.
 * For each node NodeDatabase stores miscenaleous public data (like
 * public rsa key or account object id) that are used by local node's massiv core
 * to communicate with other nodes.
 *
 * NodeDatabase uses data manager to keeps its node database up-to-date.
 * Two data objects stores list of all nodes in the massiv system.
 * The list of nodes is divided into two data objects "config/nodes/list.public"
 * and "config/nodes/list.private".
 * The public sublist contains nodes which are visible by each node in the system.
 * The private sublist contains nodes which are visible by privileged
 * nodes only (for example by server nodes).
 * The public sublist usually contains list of all server and service nodes.
 * The private sublist usually contains all client nodes.
 *
 * The data objects are stored in format compatible with registry format.
 * One registry node represents one node (its node data).
 * Node data contains node id of the node, its rsa public key, its
 * address (host name and port number), object id of its account object
 * and node name.
 * Node name is optional and may be omited. The others are required (except
 * situation when reading local node data and master data service node data,
 * see comment on @c MODE_SIMPLE below).
 * Node id of the node is stored as string config variable name "node_id"
 * containnig serialized #NodeId.
 * Rsa public key is stored as string config variable named "rsa_public_key".
 * Host address is stored as string config variable named "address"
 * containing host address name or ip address of the node.
 * Address port is stored as integer config variable named "port".
 * Account object id is stored as string config variable named "account_object"
 * containing serialized #ObjectId of the account object.
 * Node name is stored as string config variable named "name".
 *
 * NodeDatabase can work in three different modes (#Mode).
 *
 * Initial mode is @c MODE_NONE. This mode indicates that the NodeDatabase
 * is not initialized at all and it doesn't contain any valid data.
 * @c MODE_NONE is default mode once NodeDatabase object is created.
 *
 * Simple mode (@c MODE_SIMPLE) is used while the NodeDatabase is waiting
 * until data manager is ready to use.
 * Before NodeDatabase can create node database from the sublists, the data
 * objects for the sublists must be downloaded from master data service node.
 * But before we (the data manager) can contact the data service,
 * we have to initialize the node database with at least two nodes - local node
 * and the master data service node - to be able to contact the master data service
 * node and to download the sublists. For this reason simple mode exists.
 * When NodeDatabase is switched to simple mode, it extracts data for
 * the master data service node from registry and creates node
 * database using these data (this database is refered to as "static database").
 * Data for local node is initialized to anonymous mode i.e. local node id
 * is set to anonymous node id and rsa keys are set to anonymous
 * rsa keys.
 * When reading the master data service node data from registry at the
 * beginning of simple mode, account object id may be ommited.
 * Thus static database in the registry may contain only node id
 * and public rsa key for local node, and node id and public rsa key
 * and address for master data service node.
 *
 * Before switching node database to full mode, local node data must
 * be switched to non-anonymous mode.
 * This is done by @c load_local_node_config() function. 
 * When reading the local node data from registry, address (host name and
 * port number) and account object id may be omited.
 *
 * Once data manager is ready to use, NodeDatabase is switched to full mode.
 * In full mode NodeDatabase uses data objects for the sublists to create
 * the final node database.
 * NodeDatabase then periodically checks whether the sublists get updated.
 * If new version of the sublists is available, NodeDatabase recreate
 * the node database using the updated sublists.
 *
 * @todo Document local node stuff and client and server node object class type id stuff.
 * @todo CZ Co kdyz nekdo smazne node ze sublistu za behu? Pak to cele popada?
 *      Udelat neco jako freezovni server nodu? Nebo si to maji hlidat vyssi
 *      vrtsvy (node manager) samy?
 *
 * @ingroup node
 */
class NodeDatabase
    {
private:
    
    NO_COPY_CONSTRUCTOR( NodeDatabase );
    NO_ASSIGNMENT_OPERATOR( NodeDatabase );

public:

    /**
     * NodeData represents one data element stored in the node database.
     */
    class NodeData
        {
    public:
    
        /** Constructor. */
        NodeData();

        /** Constructor. */
        NodeData
            (
            const NodeId &        node_id,
            const std::string &   rsa_public_key,
            const ServerAddress & server_address,
            const ObjectId &      account_object_id,
            const ObjectId &      proxy_node_object_id,
            const std::string &   name
            );

        /** Copy constructor. */
        NodeData
            (
            const NodeData & other
            );

        /** Operator=. */
        NodeData & operator=
            (
            const NodeData & other
            );

    public:

        /** Get NodeId of the node. */
        const NodeId & get_node_id() const
            {
            return node_id;
            }

        /** Get RSA public key of the node. */
        const std::string & get_rsa_public_key() const
            {
            return rsa_public_key;
            }

        /** Get ServerAddress of the node. */
        const ServerAddress & get_server_address() const
            {
            return server_address;
            }
            
        /** Get ObjectId of account object for the node. */
        const ObjectId & get_account_object_id() const
            {
            return account_object_id;
            }
            
        /** Get ObjectId of node object of the node. */
        const ObjectId & get_proxy_node_object_id() const
            {
            return proxy_node_object_id;
            }
        
        /**
         * Get name of the node.
         * @note There are no rules for node names e.g. two nodes may have
         * the same name, the name can be empty etc.
         */
        const std::string & get_name() const
            {
            return name;
            }

    private:
    
        NodeId          node_id;
            /**< NodeId of the node. */
            
        std::string     rsa_public_key;
            /**< RSA public key of the node. */

        ServerAddress   server_address;
            /**< Address of the server node. */
            
        ObjectId        account_object_id;
            /**< ObjectId of account object for the node. */
            
        ObjectId        proxy_node_object_id;
            /**< ObjectId of node object of the node. */
            
        std::string     name;
            /**< Name of the nodes. Arbitrary ascii string. There are no
             * rules for node names e.g. two nodes may have same name,
             * the name can be empty etc.
             */

        }; // NodeData

    /**
     * Functor used to enumerate node data elements in the database.
     */
    class Enumerator
        {
    public:

        /** Default constructor. */
        Enumerator();

        /** Destructor. */
        virtual ~Enumerator();

        /**
         * Enumeration operation.
         */
        virtual void operator()
            (
            const NodeData & data
            ) = 0;
        
        }; // Enumerator
        
public:

    /**
     * NodeDatabase mode.
     */
    enum Mode
        {
        MODE_NONE,
            /**< The database doesn't contain any valid data.
             * @see NodeDatabase
             */
        MODE_SIMPLE,
            /**< The database contains "static database" extracted from registry.
             * Static database includes anonymous local node data and
             * master data service node data.
             * @see NodeDatabase
             */
        MODE_FULL
            /**< The database contains up-to-date database downloaded
             * from a data service node ane periodically checks for
             * new version available.
             * @see NodeDatabase
             */
        }; // Mode

public:

    /**
     * Default constructor.
     * Initializes NodeDatabase to @c MODE_NONE mode.
     */
    NodeDatabase();
    
    /** Destructor. */
    ~NodeDatabase();

public:

    /**
     * Load node data for local node.
     * The database must be in simple mode.
     * This function extract local node data from registry (thus overriding
     * anonymous local node data read when the database was switched to simple
     * mode).
     * Once local node data are loaded, you can switch the database to full mode.
     */
    void load_local_node_config();

    /**
     * Switch the NodeDatabase to @a new_mode mode.
     * If this operation fails for some reason, the database
     * is cleared and reset to @c MODE_NONE mode.
     * 
     * Node database can be switched to @c MODE_SIMPLE or @c MODE_NONE mode at any time.
     * Switching the database to @c MODE_FULL mode must be done only if the database
     * is running in simple mode and local node data has been set by calling
     * @c load_local_node_config().
     *
     * When switching to full mode, the database automatically refreshes itself
     * (see refresh()).
     * @throw ExNodeDatabaseError The operation has failed for some reason.
     */
    void switch_to_mode
        (
        Mode new_mode
        );

    /**
     * Refresh the database.
     * NodeDatabase checks if new version of node database data objects is available.
     * In that case it recreates the node database using new available data.
     * @note NodeDatabase keeps the database up-to-date using the data objects
     *      only in full mode. If the mode of NodeDatabase is not MODE_FULL,
     *      this function does nothing (even if @a force_refresh is true).
     * @note This function is automatically called when NodeDatabase is switched
     *      to full mode.
     * @param force_refresh Override detection of new version of the data objects
     *      and force the database to be recreated using available data objects.
     * @return True if the database was updated, false if newer version of databse
     *      is not available.
     */
    bool refresh
        (
        bool force_refresh = false
        );
        
    /**
     * NodeDatabase next tick function.
     * The NodeDatabase periodically calls refresh() in this function.
     */
    void next_tick();

public:

    /**
     * Enumerate all nodes stored in the database.
     */
     void enumerate_all_nodes
        (
        Enumerator & enumerator
        );

    /**
     * Enumerate all nodes stored in the database whose node type is @a node_type.
     */
    void enumerate_nodes
        (
        NodeId::Type    node_type,
        Enumerator &    enumerator
        );

    /**
     * Get local node id.
     */
    const NodeId & get_local_node_id() const;
    
    /**
     * Get local node rsa public key.
     */
    const std::string & get_local_node_rsa_public_key() const;

    /**
     * Get local node rsa private key.
     */
    const std::string & get_local_node_rsa_private_key() const;
     
    /**
     * Get NodeData for some @a node.
     * @throw ExNodeNotFoundInNodeDatabase The specified nod was not found
     *      in the node database.
     */
    const NodeData & get_node_data
        (
        const NodeId & node
        ) const;
        
    /**
     * Get NodeData for some @a node.
     * If specified node is not found in the node database,
     * the @a output_node_data_ptr is set to NULL.
     */
    void get_node_data
        (
        const NodeId &      node,
        const NodeData **   output_node_data_ptr
        ) const;

    /**
     * Returns ServerAddres bound to specified server node.
     */
    ServerAddress get_node_server_address
        (
        const NodeId & node
        ) const;

private:

    /** Node database internal sctructure. */
    struct My;
            
private:

    std::auto_ptr< My >     my;
        /**< Internal structure. Always a valid pointer. */
 
    }; // NodeDatabase


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_DATABASE_H
