#!/usr/bin/perl -w
#
# Gather list of managed classes.

=head1 NAME

classlist.pl - gather list of known managed classes

=head1 SYNOPSIS

B<classlist.pl>
    S<[ B<-v>|B<--verbose> ]>
    S<[ I<directory> ]>

=head1 Description

Use F<classlist.pl> to update list of known object classes and their
metaobjects and factories. It reads F<idl.list> in specified I<directory> (or
in current directory if no directory is specified) and parses all idls
specified in the list and its subproject lists. Then it writes the list into
F<class_list.list> file and C++ sources in F<class_list.cpp> and F<class_list.h>.

The list file contains database of all known classes. It's no longer read by
F<classlist.pl> (F<factlist.pl> used to read old database), but it can still
be used by other scripts or browsed by curious programmer, as it's format is very
simple.

The C++ sources will contain code needed to instantiate and register all
metaobjects and object factories. The functions will be generated into
namespace S<Massiv::Generated>.

Both the database and the C++ source will contain magic number generated from
idls. This number should change whenever something important in idls is
changed and remain the same otherwise.

B<Note>: You should never use this utility directly. Use F<genmkgen.pl> to
generate makefiles from F<idl.list> - these makefiles will contain correct
rules to generate class lists.

=over

=item Database format

Lines beginning with # and empty lines or lines containing only whitespace
are ignored.

The database begins with "ClassList" string followed by F<classlist.pl>
version. Next line contains "Magic" and the magic number. Information about
each known managed class follows:

    class_kind class_type_id class_name idl_file_name metaobject_class_name factory_class_name

Class kinds is written as a number. Class name includes namespace the class is defined
in. IDL file name is relative to I<directory> (directory containing F<idl.list>).
Metaobject and factory classes are templates and their names currently contain spaces,
which makes the file hard to parse.

=back

=head1 KNOWN BUGS

Magic number is currently always 0.

Metaobject/factory names contain spaces. Change format to something easier to parse?

=head1 SEE ALSO

Format of I<idl.list> is described in documentation of L<genmkgen.pl>.

=cut

use strict ;
use warnings ;

use Getopt::Long ;

use FindBin ;
use lib "$FindBin::Bin/lib" ;
use ClassList ;


###
# Classlist database version.
my $classlist_version = 1 ;


###
# ClassList object.
my $class_list ;


###
# Main

# Options

my $opt_verbose = 0 ;

Getopt::Long::Configure( "bundling" ) ;
GetOptions(
    'v|verbose+'    => \$opt_verbose,
) or die "Invalid options, stopped" ;

my $directory = $ARGV[ 0 ] || '.' ;

# Parse IDLs.

$class_list = ClassList->new() ;
$class_list->parse( $directory, $opt_verbose ) ;

# Write files.

print( "Writing list and sources...\n" ) if $opt_verbose ;
$class_list->write_list() ;
$class_list->write_header() ;
$class_list->write_source() ;


## EOF ##
