#! /usr/bin/perl -w
#
# Generate makefile.gen code for metaobjects and object factories.

=head1 NAME

genmkgen.pl - generate parts of makefile.gen used to generate metaobjects and factories.

=head1 SYNOPSIS

B<genmkgen.pl>
    S<[ B<-v>|B<--verbose> ]>
    [ I<directory> ]

=head1 Description

This utility is used to generate F<makefile_idl.gen> from F<idl.list> files.
F<makefile_idl.gen> should be included by F<makefile.gen> in the same directory. It
instructs F<mkgen.pl> how to generate makefiles that in turn generate metaobjects
and factories from IDL code. F<genmkgen.pl> must be run from massiv root directory
(the directory containing F<src>, F<mkgen> and F<doc> directories).

F<genmkgen.pl> will process all subdirectories of I<directory> (or all
subdirectories of current directory if no I<directory> is specified) which
contain F<idl.list> file.

The following commands can be used in F<idl.list>:

=over

=item B<depends> I<project-directory>

Project in I<directory> requires stuff from project in I<project-directory>.
I<project-directory> must contain another F<idl.list>. All required
"subprojects" must be listed, I<depends> of subproject F<idl.list> files will
be ignored.

=item B<directory> I<include-directory>

Add I<include-directory> to list of directories to search when "#include"
directive is encountered in IDL file. Include directories specified in
subprojects of this project will be used too.

=item B<idl> I<file-name>

All IDL files belonging to this project should be listed using I<IDL> commands.

Generated makefile will contain code that will generate metaobjects, factories
and RPC classes for all managed classes defined in I<file-name>.

For each I<file-name> defined in projects specified by I<depends>, generated
makefile will contain correct dependencies for all generated files.

=item B<class_list>

Generate list of managed classes belonging to this project and its subprojects.

=back

=head1 GENERATED MAKEFILE

Generated file, F<makefile_idl.gen> will contain:

Rules and dependencies used to generate metaclasses for each IDL specified by
I<idl> command in both main and subproject F<idl.list>. If IDL belongs to subproject,
make will be run in the subproject directory to generate it if necessary.

If I<class_list> command is used in main project F<idl.list>, rules and dependencies
used to generate class list will be written to the makefile.

If I<class_list> command is used in subproject F<idl.list>, dependencies for the
class list will be written, makefile in the subproject directory will be used to
generate the class list if necessary.

=head1 KNOWN BUGS

None.

=head1 SEE ALSO

If you really want to examine contents of generated makefile_idl.gen, check
macros defined in F<mkgen/platform/common/makegen.factgen>, which are used to
generate all the rules with correct dependencies.

=cut

use strict ;
use warnings ;

use Getopt::Long ;
use File::Find ;

use FindBin ;
use lib "$FindBin::Bin/lib" ;
use FileNames ;
use GenMkgen ;


###
# Main

# Options

my $opt_verbose = 0 ;

Getopt::Long::Configure( "bundling" ) ;
GetOptions(
    'v|verbose+'    => \$opt_verbose,
) or die "Invalid options, stopped" ;

my $directory = $ARGV[ 0 ] || '.' ;

# Find all idl.list files in $directory.

my @project_dirs ;
my $idl_list = FileNames::idl_list() ;
find( sub {
    if ( $_ eq $idl_list ) {
        my $path = $File::Find::name ;
        $path =~ s|/$idl_list$||o ;
        push( @project_dirs, $path ) ;
    }
}, $directory ) ;

# Process lists.

my $makefile_idl = FileNames::makefile_idl() ;
foreach my $dir ( @project_dirs ) {
    print( "Generating '$makefile_idl' in '$dir'.\n" ) if $opt_verbose ;
    my $genmkgen = GenMkgen->new() ;
    $genmkgen->parse( $dir, $opt_verbose > 1 ? 1 : 0 ) ;
    $genmkgen->write_makefile_gen() ;
}
