=head1 NAME

ClassInfo.pm - additional information about class symbol.

=head1 SYNOPSIS

    use ClassInfo ;

    # Create empty class info.
    $class_info = ClassInfo->new() ;

    # Create and initialize class info.
    $class_info = ClassInfo->new( $class_symbol ) ;

    # Get factory/metaobject class template name, relative to Massiv::Generated namespace.
    $factory_template_name = ClassInfo->factory_name() ;
    $metaobject_template_name = ClassInfo->metaobject_name() ;

    # Initialize empty class info.
    $class_info->initialize( $class_symbol ) ;

    # Get factory/metaobject class name, relative to Massiv::Generated namespace.
    $factory_name = $class_info->factory_name() ;
    $metaobject_name = $class_info->metaobject_name() ;

    # Get reference to array of all parents (references to SymbolClass)
    # Please do not modify the array.
    $parents = $class_info->parents() ;

    # Get reference to array of all virtual parents (references to SymbolClass)
    # Please do not modify the array.
    $virtual_parents = $class_info->virtual_parents() ;

    # Get number of class components, excluding class itself.
    $subcomponent_count = $class_info->component_count() ;
    
    # Get reference to class symbol object for given component.
    #   index - 1-based index to COMPONENTS, 0 for this component.
    $class_symbol = $class_info->class_symbol( $index ) ;

    # Get reference to array of all properties, including inherited ones.
    # Properties will be sorted by component.
    # Values are structures:
    #   SYMBOL - reference to SymbolProperty
    #   COMPONENT - 1-base component index
    # Please do not modify it.
    $properties = $class_info->properties() ;

    # Write code containing casts from 'Object * object' to all components
    # of the class. These pointers will be called 'p_COMPNENT_INDEX'.
    #   handle - write the code here
    #   is_const - should all the pointers be constant?
    # returns reference to array of flags indicating which components were
    # referenced while performing the casts
    $referenced = $class_info->write_casts( $handle, $is_const ) ;

    # Write code referencing each unreferenced p_INDEX.
    # FIXME: Should this be part of ClassInfo?
    $class_info->write_references( $handle, $referenced ) ;

    # Write arguments to MetaObject::ComponentInfo constructor for
    # given component.
    #   handle - write the code here
    #   index - 1-based index into COMPONENTS, 0 for this component
    $class_info->write_componentinfo_arguments( $handle, $index ) ;

    # Write code containing casts from p_0 (pointer to object of this class)
    # to pointer to specified parent class.
    #   handle - write the code here
    #   index - 1-based index into COMPONENTS, 0 for this component
    #   indent - indentation to use (string)
    #   is_const - should all the pointers be constant?
    # Result will be stored in p_$index pointer.
    $class_info->write_component_casts( $handle, $index, $indent, $is_const ) ;

    # Create component label used by textual serializers.
    #   index - 1-based component index
    # Retuns string in ':flags:component_path format'
    # FIXME: This string can be very long because component names include
    #        namespace names. Textual serialization should be probably changed.
    $component_string = $class_info->component_string( $index ) ;

=head1 DESCRIPTION

C<ClassInfo> class offers additional information about class symbol
(see F<SymbolClass.pm>).

=head1 TODO

Clean this up, because some functionality has been moved to MetaObject (C++ code).

=head1 KNOWN BUGS

None.

=cut

package ClassInfo ;

use strict ;
use warnings ;

use SymbolClass ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = {
        # Reference to class symbol object.
        SYMBOL => undef,

        # Factory class name.
        FACTORY_NAME => undef,

        # Metaobject class name.
        METAOBJECT_NAME => undef,

        # Array of references to SymbolClass for all parent classes.
        # Each parent class will be listed at most once (even if it's
        # inherited multiple times).
        PARENTS => [ ],

        # Subset of PARENTS, virtual parents only.
        VIRTUAL_PARENTS => [ ],

        # Array describing all components and how to cast to them.
        # Each element is structure:
        #   FROM_INDEX
        #       Cast from pointer to component described by COMPONENTS[ index - 1 ]
        #       or "complete component" pointer if index == 0.
        #   TO_INDEX
        #       Index of this element + 1.
        #   SYMBOL
        #       Reference to SymbolClass for this component.
        #   NAME
        #       Unique component name (withing this class)
        #       Currently if inheritance is C : B : A, where A, B, and C are
        #       fully qualified names without initial '::', component A in C
        #       has unique name "B/A".
        #   VIRTUAL
        #       True if this component is inherited virtually.
        COMPONENTS => [ ],

        # Array of 1-based indices to COMPONENTS.
        # This order should be is to order in which constructors are called
        # in C++. 0 is always last and represents "complete component".
        COMPONENT_ORDER => [ ],

        # All properties (including inherited ones).
        # Array of structures:
        #   SYMBOL - reference property symbol object
        #   COMPONENT - 1-based index to COMPONENTS
        PROPERTIES => [ ],
    } ;

    # Bless it.

    bless( $self, $class ) ;

    # Initialize.

    if ( @_ ) {
        $self->initialize( shift ) ;
    }

    # Done.

    return $self ;
}


###
# factory_template_name
sub factory_template_name {
    return 'ObjectFactoryImpl' ;
}


###
# metaobject_template_name
sub metaobject_template_name {
    return 'MetaObjectImpl' ;
}


###
# initialize
sub initialize {
    my $self = shift ;
    my $symbol = shift ;

    if ( defined( $$self{ SYMBOL } ) ) {
        require Carp ;
        Carp::confess( "Tried to reinitialize ClassInfo.\n" ) ;
    }
    $$self{ SYMBOL } = $symbol ;
    my $self_name = $symbol->full_name() ;

    # Class names.

    $$self{ FACTORY_NAME } = $self->factory_template_name() . "< $self_name >" ;
    $$self{ METAOBJECT_NAME } = $self->metaobject_template_name() . "< $self_name >" ;

    # Add temp sets (names, not symbols) and initialize recursively.

    $$self{ PARENT_SET } = { } ;
    $$self{ VIRTUAL_PARENT_SET } = { } ;
    $self->_add_component( $symbol, '' ) ;

    # Check inheritance of Massiv::Core::Object.

    if( $self_name ne '::Massiv::Core::Object' ) {
        unless ( defined( $$self{ PARENT_SET }{ '::Massiv::Core::Object' } ) ) {
            die( "$self_name is not descendent of Massiv::Core::Object.\n" ) ;
        }

        if ( $$self{ PARENT_SET }{ '::Massiv::Core::Object' } > 1 ) {
            die( "static_cast from '$self_name *' to 'Massiv::Core::Object *' is ambiguous.\n" ) ;
        }
    }

    # Note:
    # Former $$self{ PARENTS } included self.

    # Delete temporary sets.

    delete $$self{ PARENT_SET } ;
    delete $$self{ VIRTUAL_PARENT_SET } ;

    # Add properties of each component into the list.

    my $properties = $$self{ PROPERTIES } ;
    my $component_order = $$self{ COMPONENT_ORDER } ;
    foreach my $index ( @$component_order ) {
        my $class = $self->class_symbol( $index ) ;

        # Add all properties.

        my $component_properties = $class->properties() ;
        foreach my $property ( @$component_properties ) {
            push( @$properties, {
                SYMBOL => $property,
                COMPONENT => $index,
            } ) ;
        }
    }
}


###
# factory_name
sub factory_name {
    return $_[ 0 ]{ FACTORY_NAME } ;
}


###
# metaobject_name
sub metaobject_name {
    return $_[ 0 ]{ METAOBJECT_NAME } ;
}


###
# parents
sub parents {
    return $_[ 0 ]{ PARENTS } ;
}


###
# virtual_parents
sub virtual_parents {
    return $_[ 0 ]{ VIRTUAL_PARENTS } ;
}


###
# component_count
sub component_count {
    return scalar( @{$_[ 0 ]{ COMPONENTS }} ) ;
}


###
# class_symbol
sub class_symbol {
    my $self = shift ;
    my $index = shift ;
    unless ( defined( $index ) ) {
        require Carp ;
        Carp::confess( "Class symbol index required, is should probably be 0.\n" ) ;
    }
    if ( $index == 0 ) {
        return $$self{ SYMBOL } ;
    }
    my $components = $$self{ COMPONENTS } ;
    if ( $index > scalar( @$components ) ) {
        require Carp ;
        Carp::confess( "Component index '$index' out of range.\n" ) ;
    }
    return $$components[ $index - 1 ]{ SYMBOL } ;
}


###
# properties
sub properties {
    return $_[ 0 ]{ PROPERTIES } ;
}


###
# write_casts
sub write_casts {
    my( $self, $handle, $is_const ) = @_ ;
    my $symbol = $$self{ SYMBOL } ;
    my $components = $$self{ COMPONENTS } ;
    my $self_name = $symbol->full_name() ;

    # Find longest parent name to layout generated source nicely.

    my $max_length = length( $self_name ) ;
    foreach my $component ( @$components ) {
        my $length = length( $$component{ SYMBOL }->full_name() ) ;
        if ( $length > $max_length ) { $max_length = $length ; }
    }

    # Mark pointers as unreferenced.

    my @referenced ;
    $#referenced = scalar( @$components ) ;

    # Generate code which casts Object * to class_name *.

    $handle->print( "    hope( object );\n" ) ;
    $handle->print( '    ' ) ;
    $handle->print( 'const ' ) if $is_const ;
    $handle->print( $self_name ) ;
    $handle->print( ' ' x ( 1 + $max_length - length( $self_name ) ) ) ;
    $handle->print( "* const p_0 = get_complete_object_pointer( object );\n" ) ;
    
    # Generate static_casts for each component.
    # Mark referenced pointers.

    foreach my $component ( @$components ) {
        my $cname = $$component{ SYMBOL }->full_name() ;
        $handle->print( '    ' ) ;
        $handle->print( 'const ' ) if $is_const ;
        $handle->print( $cname ) ;
        $handle->print( ' ' x ( 1 + $max_length - length( $cname ) ) ) ;
        $handle->print( "* const p_$$component{ TO_INDEX } = static_cast< " ) ;
        $handle->print( 'const ' ) if $is_const ;
        $handle->print( "$cname * >( p_$$component{ FROM_INDEX } );\n" ) ;
        $referenced[ $$component{ FROM_INDEX } ] = 1 ;
    }

    return \@referenced ;
}


###
# write_references
sub write_references {
    my $self = shift ;
    my $handle = shift ;
    my $referenced = shift ;

    for ( my $i = 0 ; $i < scalar( @$referenced ) ; $i++ ) {
        next if $$referenced[ $i ] ;
        $handle->print( "    ::Massiv::Core::reference_value( p_$i );\n" ) ;
    }
}


###
# write_componentinfo_arguments
sub write_componentinfo_arguments {
    my( $self, $handle, $index ) = @_ ;

    my $symbol = $self->class_symbol( $index ) ;
    my $class_name = $symbol->full_name() ;
    $handle->print( "typeid( $class_name ), " ) ;

    if ( $index == 0 ) {
        $handle->print( "false, -1, \"\"" ) ;
    }
    else {
        my $components = $$self{ COMPONENTS } ;
        my $component = $$components[ $index - 1 ] ;
        my $virtual = ( $$component{ VIRTUAL } ? 'true' : 'false' ) ;
        $handle->print( "$virtual, $$component{ FROM_INDEX }, \"$$component{ NAME }\"" ) ;
    }
}


###
# write_component_casts
sub write_component_casts {
    my( $self, $handle, $index, $indent, $is_const ) = @_ ;
    my $const_string = ( $is_const ? 'const ' : '' ) ;

    if ( $index == 0 ) {
        return 0 ;
    }

    $self->_write_component_cast( $handle, $index - 1, $indent, $const_string ) ;
}


###
# component_string
sub component_string {
    my $self = shift ;
    my $index = shift ;

    my $flags = '' ;
    my $component_path ;

    if( $index == 0 ) {
        # Main component.

        $flags .= 'm' ;
        $component_path = '' ;
    }
    else {
        # Parent component.

        my $component = $$self{ COMPONENTS }[ $index - 1 ] ;
        if ( $$component{ VIRTUAL } ) {
            # Virtual component - do not print path, only name.

            $flags .= 'v' ;
            $component_path = $$component{ SYMBOL }->name() ;
        }
        else {
            # Normal component.

            $component_path = $$component{ NAME } ;
        }
    }

    return ":${flags}:${component_path}" ;
}


###
# _add_component
# Add information how to cast from given component to its parents.
#   symbol - reference to SymbolClass component to add info about
#   my_name - unique name of this component including trailing separator
sub _add_component {
    my $self = shift ;
    my $symbol = shift ;
    my $my_name = shift ;

    my $parents = $symbol->parents() ;

    # Component index.

    my $my_index = @{$$self{ COMPONENTS }} ; # yes, not $#, it's 1-based

    # Process parents.

    foreach my $parent_info ( @$parents ) {
        my $parent_symbol = $$parent_info{ SYMBOL } ;
        my $parent_virtual = $$parent_info{ VIRTUAL } ;
        my $parent_name = $parent_symbol->full_name() ;

        # Skip virtual parent if already done before.

        if( $parent_virtual ) {
            next if defined( $$self{ VIRTUAL_PARENT_SET }{ $parent_name } ) ;
            $$self{ VIRTUAL_PARENT_SET }{ $parent_name }++ ;
            push( @{$$self{ VIRTUAL_PARENTS }}, $parent_symbol ) ;
        }

        # Update parent list.

        push( @{$$self{ PARENTS }}, $parent_symbol )
            unless defined( $$self{ PARENT_SET }{ $parent_name } ) ;
        $$self{ PARENT_SET }{ $parent_name }++ ;

        # Add information about this parent to COMPONENTS.

        my $complete_parent_name = $my_name . substr( $parent_name, 2 ) ;
        push( @{$$self{ COMPONENTS }}, {
            FROM_INDEX => $my_index,
            TO_INDEX => 1 + @{$$self{ COMPONENTS }},
            SYMBOL => $parent_symbol,
            NAME => $complete_parent_name,
            VIRTUAL => $parent_virtual,
        } ) ;

        # Add information about parents of this parent to result.

        $self->_add_component( $parent_symbol, $complete_parent_name . '/' ) ;
    }

    # Insert us into component order.

    push( @{$$self{ COMPONENT_ORDER }}, $my_index ) ;
}


###
# _write_component_cast
# Recurse to write casts to FROM_INDEX component and then write cast to
# the specified component.
#   handle - output
#   i - index to components - cast to this component (0-based)
#   indent - indentation string
#   const_string - either 'const ' or ''
sub _write_component_cast ;
sub _write_component_cast {
    my( $self, $handle, $i, $indent, $const_string ) = @_ ;
    my $components = $$self{ COMPONENTS } ;

    # Recurse if necessary.

    my $component = $$components[ $i ] ;
    my $from_i = $$component{ FROM_INDEX } ;
    if ( $from_i > 0 ) {
        $self->_write_component_cast( $handle, $from_i - 1, $indent, $const_string ) ;
    }

    # Write cast to i.

    my $to_i = $i + 1 ;
    my $cname = $$component{ SYMBOL }->full_name() ;
    $handle->print( "$indent$const_string$cname * const p_$to_i = static_cast< $const_string$cname * >( p_$from_i );\n" ) ;
}



1 ;


__END__
