=head1 NAME

ClassList - generator of class list files.

=head1 SYNOPSIS

    use ClassList ;

    # Constructor
    $class_list = ClassList->new() ;

    # Parsing.
    $class_list->parse( $directory [, $verbose ] ) ;

    # Generate files.
    $class_list->write_header() ;
    $class_list->write_source() ;
    $class_list->write_list() ;


=head1 DESCRIPTION

C<ClassList> generates list of all classes defined in specified project
and its subprojects.

=head1 KNOWN BUGS

None.

=cut

package ClassList ;

use strict ;
use warnings ;

use FileHandle ;

use IdlList ;
use FileNames ;
use Utils ;


###
# classlist_version
my $classlist_version = 2 ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate.

    my $self = {
        # Directory.
        DIRECTORY => undef,

        # Magic number.
        # Used to be random. Currently not used. Should be computed by IDL
        # parser as CRC of symbol hierarchy. Or simply CRC of IDL files?
        MAGIC => 0,

        # Set of all classes.
        # Keys are fully qualified class names. Values are structures:
        #   KIND - class kind id.
        #   TYPE_ID - class type id. Always >= 1.
        #   METAOBJECT - metaobject class name.
        #   FACTORY - factory class name.
        #   IDL_NAME - name of idl file with definition of this class.
        CLASSES => { },
    } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# parse
sub parse  {
    my( $self, $directory, $verbose ) = @_ ;

    $$self{ DIRECTORY } = $directory ;

    # Parse everything.
    # Note that this is quite an overkill - we need only metaobject/factory
    # names and file name of IDL definition.

    my $idl_list = IdlList->new() ;
    print( "Parsing IDL lists...\n" ) if $verbose ;
    $idl_list->parse( $directory ) ;

    # Check if class list should be generated here.

    unless ( defined( $idl_list->class_kinds() ) ) {
        print( STDERR "Warning: class_list not specified in IDL list.\n" ) ;
    }

    $idl_list->parse_subprojects() ;

    print( "Parsing IDLs...\n" ) if $verbose ;
    $idl_list->parse_idls() ;

    # Extract necessary info.

    my %class_type_id ;
    my $class_kind_numbers = $idl_list->class_kind_numbers() ;
    if ( defined( $class_kind_numbers ) ) {
        foreach my $kind_number ( @$class_kind_numbers ) {
            $class_type_id{ $kind_number } = 1 ;
        }
    }

    my $classes = $$self{ CLASSES } ;
    my $class_infos = $idl_list->class_infos() ;

    CLASS : foreach my $class_name ( sort( keys( %$class_infos ) ) ) {
        my $class_info = $$class_infos{ $class_name } ;
        my $class_symbol = $class_info->class_symbol( 0 ) ;
        my $kind = $class_symbol->require_attribute( '::kind' ) ;

        # Skip this kind?

        next CLASS unless defined( $class_type_id{ $kind } ) ;

        # Add this class to class list.

        $$classes{ $class_name } = {
            KIND => $kind,
            TYPE_ID => $class_type_id{ $kind },
            METAOBJECT => $class_info->metaobject_name(),
            FACTORY => $class_info->factory_name(),
            IDL_NAME => $class_symbol->file_name(),
        } ;
        $class_type_id{ $kind }++ ;
    }
}


###
# write_header
sub write_header {
    my $self = shift ;
    my $directory = $$self{ DIRECTORY } ;

    my $file_name = Utils::simplify_path( $directory . '/' . FileNames::class_list_header() ) ;
    my $handle = FileHandle->new( $file_name, 'w' ) ;
    unless ( defined( $handle ) ) {
        die( "Can't create '$file_name'.\n" ) ;
    }

    my $header_guard = $file_name ;
    $header_guard =~ s|[./]|_|g ;

    my $namespace = 'Massiv::Generated' ;

    # Write intro.

    $handle->print( <<EOT
/** \@file $file_name
 * List of all managed classes.
 *
 * Automatically generated by classlist.pl version ${classlist_version}. Do not modify.
 */

#ifndef ${header_guard}
#define ${header_guard}


#ifndef MASSIV_COMMON
#include "global/common.h"
#endif


EOT
    ) ;

    Utils::write_open_namespace( $handle, $namespace ) ;

    # Body.

    $handle->print( <<EOT


extern const unsigned class_list_magic;
    /**< Magic number - changes every time something in class definitions is changed. */


/**
 * Instantiate and register metaobjects and factories for all known managed classes.
 */
void register_classes();


EOT
    ) ;

    # Ending

    Utils::write_close_namespace( $handle, $namespace ) ;

    $handle->print( <<EOT


#endif // ${header_guard}
EOT
    ) ;

    $handle->close() ;
}


###
# write_source
sub write_source( $$\%$ ) {
    my $self = shift ;
    my $directory = $$self{ DIRECTORY } ;

    my $file_name = Utils::simplify_path( $directory . '/' . FileNames::class_list_source() ) ;
    my $handle = FileHandle->new( $file_name, 'w' ) ;
    unless ( defined( $handle ) ) {
        die( "Can't create '$file_name'.\n" ) ;
    }

    my $header_file_name = Utils::simplify_path( $directory . '/' . FileNames::class_list_header() ) ;

    my $namespace_generated = 'Massiv::Generated' ;
    my $namespace_core = 'Massiv::Core' ;

    # Write intro.

    $handle->print( <<EOT
/** \@file $file_name
 * List of all managed classes.
 *
 * Automatically generated by classlist.pl version ${classlist_version}. Do not modify.
 */

#include "global/common.h"
#include "global/global.h"
#include "object/class_manager.h"
#include "pointer/default_cast_impl.h"

EOT
    ) ;

    # Write generated header includes.

    my %generated_header_names ;
    foreach my $class_name ( keys( %{$$self{ CLASSES }} ) ) {
        my $idl_name = $$self{ CLASSES }{ $class_name }{ IDL_NAME } ;
        my $generated_header_name =
            FileNames::generated_header( FileNames::base( $idl_name ) ) ;
        $generated_header_names{ $generated_header_name }++ ;
    }

    foreach my $header_name ( sort( keys( %generated_header_names ) ) ) {
        $handle->print( "#include \"$header_name\"\n" ) ;
    }

    # Write rest of intro.

    $handle->print( <<EOT

#include "${header_file_name}"


EOT
    ) ;

    Utils::write_open_namespace( $handle, $namespace_generated ) ;

    # Body.

    $handle->print( <<EOT


const unsigned class_list_magic = $$self{ MAGIC };


void register_classes()
    {
EOT
    ) ;

    # Create all metaobjects, their create() functions do the rest.

    foreach my $class_name ( sort( keys( %{$$self{ CLASSES }} ) ) ) {
        my $class_info = $$self{ CLASSES }{ $class_name } ;
        my $metaobject = $$class_info{ METAOBJECT } ;
        my $kind = $$class_info{ KIND } ;
        my $type_id = $$class_info{ TYPE_ID } ;
        $handle->print( "    ${metaobject}::create( $kind, $type_id );\n" ) ;
    }

    # End.

    $handle->print( <<EOT
    }

    
EOT
    ) ;
    
    # Instantiate classes and data members in namespace Massiv::Core.
    
    Utils::write_close_namespace( $handle, $namespace_generated ) ;        
    $handle->print( "\n\n" ) ;        
    Utils::write_open_namespace( $handle, $namespace_core ) ;
    $handle->print( "\n\n" ) ;

    # Instantiate DefaultCast classes.
    
    foreach my $class_name ( sort( keys( %{$$self{ CLASSES }} ) ) ) {
        $handle->print( "template class DefaultCast< ${class_name} >;\n" ) ;
    }
    $handle->print( "\n\n" ) ;    

    # Define storage for PointerBase< DefaultCast<> >::cast instances.    
    
    $handle->print( <<EOT
#ifdef COMPILER_MSVC

// Define storage for PointerBase< DefaultCast<> >::cast instances.
// This is to workaround a regression in MSVC7.1, MSVC8 and MSVC9 front ends and 
// prevent linking errors. The generic definition of PointerBase<>::cast member in 
// "core/pointer/pointer_base.h" should be sufficient though and works correctly
// in earlier versions.


EOT
    ) ;

    foreach my $class_name ( sort( keys( %{$$self{ CLASSES }} ) ) ) {
        $handle->print( "template <>\n" ) ;
        $handle->print( "DefaultCast< ${class_name} > PointerBase< DefaultCast< ${class_name} > >::cast;\n\n" ) ;
    }
    
    $handle->print( <<EOT
    
#endif // COMPILER_MSVC


EOT
    ) ;

    Utils::write_close_namespace( $handle, $namespace_core ) ;

    # Done.

    $handle->close() ;
}


###
# write_list
sub write_list {
    my $self = shift ;
    my $directory = $$self{ DIRECTORY } ;

    my $file_name = Utils::simplify_path( $directory . '/' . FileNames::class_list() ) ;
    my $handle = FileHandle->new( $file_name, 'w' ) ;
    unless ( defined( $handle ) ) {
        die( "Can't create '$file_name'.\n" ) ;
    }

    # Write intro.

    $handle->print( <<EOT
# Class list.
# Automatically generated by classlist.pl version ${classlist_version}. Do not modify.

ClassList ${classlist_version}

magic $$self{ MAGIC }

EOT
    ) ;

    # Write entries

    foreach my $class_name ( sort( keys( %{$$self{ CLASSES }} ) ) ) {
        my $class_info = $$self{ CLASSES }{ $class_name } ;
        my $kind = $$class_info{ KIND } ;
        my $type_id = $$class_info{ TYPE_ID } ;
        my $idl_name = $$class_info{ IDL_NAME } ;
        my $metaobject = $$class_info{ METAOBJECT } ;
        my $factory = $$class_info{ FACTORY } ;
        $handle->print( "$kind $type_id $class_name $idl_name $metaobject $factory\n" ) ;
    }
}


1 ;


__END__
