=head1 NAME

FactoryGen.pm - factory generator

=head1 SYNOPSIS

    use FactoryGen ;

    FactoryGen::write_class( $handle, $class_info ) ;

    FactoryGen::write_source( $handle, $class_info, $idl_parser ) ;

=head1 DESCRIPTION

Use this module to write interface and implementation of factory for class
specified by I<$class_info> using information gathered by parser I<$idl_parser>.

See C<ClassInfo.pm> and C<IdlParser.pm> for information about the classes.

=head1 KNOWN BUGS

None.

=cut

package FactoryGen ;

use strict ;
use warnings ;

use ClassInfo ;
use IdlParser ;


###
# write_class
sub write_class {
    my( $handle, $class_info ) = @_ ;

    # Extract info.

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $factory_template_name = $class_info->factory_template_name() ;
    my $factory_name = $class_info->factory_name() ;
    my $metaobject_name = $class_info->metaobject_name() ;

    my $symbol_attribute = $class_symbol->require_attribute( '::symbol_attribute' ) ;
    $symbol_attribute .= ' ' if $symbol_attribute ne '' ;
    
    # Write class definition.

    $handle->print( <<EOT

/**
 * Object factory for $class_name.
 */
template<>
class ${symbol_attribute}$factory_name : public ::Massiv::Core::ObjectFactory
    {
    /** This class. */
    typedef $factory_template_name Self;

    /** Superclass. */
    typedef ::Massiv::Core::ObjectFactory Super;

    /** Metaobject. */
    typedef $metaobject_name Meta;

public:

    /** Pointer to the only instance of $factory_name. */
    static const Self * self;

public:

    /**
     * Constructor.
     */
    $factory_template_name
        (
        const Meta & the_metaobject
        );

    /**
     * Destructor.
     */
    virtual ~$factory_template_name();

public:

    /**
     * Create new object of class $class_name.
     * Generates new object id, initializes the object and its properties.
     */
    virtual ::Massiv::Core::ObjectProperty * create_object() const;

    /**
     * Create new object of class $class_name using given object id.
     * Initializes the object and its properties.
     */
    virtual ::Massiv::Core::ObjectProperty * create_object
        (
        const ::Massiv::Core::ObjectId & object_id
        ) const;

    /**
     * Create new object replica of class $class_name using given object id.
     * Initializes the object and its properties.
     */
    virtual ::Massiv::Core::ObjectProperty * create_replica
        (
        const ::Massiv::Core::ObjectId & object_id
        ) const;

    /**
     * Clone object of class $class_name.
     */
    virtual ::Massiv::Core::ObjectProperty * clone_object
        (
        const ::Massiv::Core::ObjectProperty * object
        ) const;

    /**
     * Throw object of class $class_name by value.
     */
    virtual void raise_exception
        (
        const ::Massiv::Core::ObjectProperty * object
        ) const;

    }; // class $factory_name
EOT
    ) ;

}


###
# write_source
sub write_source {
    my( $handle, $class_info, $idl_parser ) = @_ ;

    # Extract info.

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $factory_template_name = $class_info->factory_template_name() ;
    my $factory_name = $class_info->factory_name() ;
    my $metaobject_name = $class_info->metaobject_name() ;
    my $properties = $class_info->properties() ;
    my $all_attributes = $idl_parser->attributes() ;

    my $abstract = $class_symbol->require_attribute( '::abstract' ) ;

    my $referenced ;

    $handle->print( <<EOT

const ${factory_name} * ${factory_name}::self = NULL;

${factory_name}::${factory_template_name}
    (
    const Meta & the_metaobject
    ) : Super( the_metaobject )
    {
    hope( !self );
    self = this;
    }
    
${factory_name}::~${factory_template_name}()
    {
    }
EOT
    ) ;
    
    if ( $abstract ) {
        $handle->print( <<EOT

::Massiv::Core::ObjectProperty * ${factory_name}::create_object() const
    {
    _abstract_instance();
    /*NOTREACHED*/
    }

::Massiv::Core::ObjectProperty * ${factory_name}::create_object
    (
    const ::Massiv::Core::ObjectId &
    ) const
    {
    _abstract_instance();
    /*NOTREACHED*/
    }

::Massiv::Core::ObjectProperty * ${factory_name}::create_replica
    (
    const ::Massiv::Core::ObjectId &
    ) const
    {
    _abstract_instance();
    /*NOTREACHED*/
    }

::Massiv::Core::ObjectProperty * ${factory_name}::clone_object
    (
    const ::Massiv::Core::ObjectProperty *
    ) const
    {
    _abstract_instance();
    /*NOTREACHED*/
    }


void ${factory_name}::raise_exception
    (
    const ::Massiv::Core::ObjectProperty * object
    ) const
    {
    _abstract_instance();
    /*NOTREACHED*/
    }
EOT
        ) ;
    } # if $abstract
    else {
        $handle->print( <<EOT

::Massiv::Core::ObjectProperty * ${factory_name}::create_object() const
    {
    $class_name * const object = new $class_name;
    initialize_object( *object );
    return object;
    }

::Massiv::Core::ObjectProperty * ${factory_name}::create_object
    (
    const ::Massiv::Core::ObjectId & object_id
    ) const
    {
    $class_name * const object = new $class_name;
    initialize_object( *object, object_id );
    return object;
    }

::Massiv::Core::ObjectProperty * ${factory_name}::create_replica
    (
    const ::Massiv::Core::ObjectId & object_id
    ) const
    {
    $class_name * const object = new $class_name;
    initialize_replica( *object, object_id );
    return object;
    }

::Massiv::Core::ObjectProperty * ${factory_name}::clone_object
    (
    const ::Massiv::Core::ObjectProperty * other_object
    ) const
    {
    const $class_name * const other = ::Massiv::Core::variant_cast< $class_name * >( get_complete_object( other_object ) );
    $class_name * const object = new $class_name( *other );
    initialize_object( *object );
    return object;
    }

void ${factory_name}::raise_exception
    (
    const ::Massiv::Core::ObjectProperty * object
    ) const
    {
    const $class_name * const my_object = ::Massiv::Core::variant_cast< $class_name * >( get_complete_object( object ) );
    hope( my_object->get_metaobject()->get_attribute_throwable() ); // FIXME
    throw *my_object;
    }
EOT
        ) ;
    } # if $abstract

}


1 ;


__END__
