=head1 NAME

GenMkgen - generator of factgen/classlist-related makefile.gen snippets.

=head1 SYNOPSIS

    use GenMkgen ;

    # Constructor
    $genmkgen = GenMkgen->new() ;

    # Parsing.
    $genmkgen->parse( $directory [, $verbose ] ) ;

    # Generate files.
    $genmkgen->write_makefile_gen() ;


=head1 DESCRIPTION

C<GenMkgen> generates parts of makefile.gen based on information read from IDLs
specified in idl.list in given directory and in subproject idl.lists.

=head1 KNOWN BUGS

None.

=cut

package GenMkgen ;

use strict ;
use warnings ;

use IdlList ;
use FileNames ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate.

    my $self = {
        # IdlList object.
        IDL_LIST => undef,

        # Additional per-IDL info.
        # Keys are IDL file names relative to current directory, values are
        # structures (all file names are relative to IDL_LIST->directory()):
        #   MAIN_IDL - copied from IdlList's all_idls.
        #   CLASSES - reference to array in IdlList's all_idls.
        #   HEADERS - copy of array in IdlList's all_idls, relative file names
        #   PATH - path to project directory relative to IDL_LIST->directory()
        #   FILE_NAME - IDL file name
        #   BASE_NAME - FILE_NAME without extension
        #   GENERATED_SOURCE - generated source file name
        #   GENERATED_HEADER - generated header file name
        #   RPC_HEADER - RPC header file name        
        #   PARENT_RPC_HEADERS - array of rpc header file names to include from RPC_HEADER
        IDL_INFOS => { },
    } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# parse
sub parse {
    my( $self, $directory, $verbose ) = @_ ;

    # Parse everything.

    my $idl_list = IdlList->new() ;
    $$self{ IDL_LIST } = $idl_list ;
    print( "Parsing IDL lists...\n" ) if $verbose ;
    $idl_list->parse( $directory ) ;
    $idl_list->parse_subprojects() ;

    print( "Parsing IDLs...\n" ) if $verbose ;
    $idl_list->parse_idls() ;

    $self->_gather_idl_info() ;
}


###
# write_makefile_gen
sub write_makefile_gen {
    my $self = shift ;
    my $idl_list = $$self{ IDL_LIST } ;
    my $idl_infos = $$self{ IDL_INFOS } ;
    my @all_idls = sort( keys( %$idl_infos ) ) ;
    my $directory = $idl_list->directory() ;

    my $file_name = Utils::simplify_path( $directory . '/' . FileNames::makefile_idl() ) ;
    my $handle = FileHandle->new( $file_name, 'w' ) ;
    unless ( defined( $handle ) ) {
        die( "Can't create '$file_name'.\n" ) ;
    }

    $handle->print( <<EOT
# makefile_idl.gen, automatically generated by genmkgen,pl, do not modify.
# To be included into makefile.gen.

# Path to src/core

EOT
    ) ;

    my $core_path = Utils::relative_path( 'src/core', $directory ) ;
    if ( length( $core_path ) ) {
        $core_path .= '/' ;
    }
    $handle->print( 'replace @CORE_PATH@ ' . $core_path . "\n" ) ;

    $handle->print( <<EOT

# Include helper macros

sysinclude common/makegen.factgen

# List of all generated sources.

EOT
    ) ;


    ## List of all generated sources.

    $handle->print( "replace \@GENERATED_SOURCES\@" ) ;
    foreach my $idl ( @all_idls ) {
        my $idl_info = $$idl_infos{ $idl } ;
        next unless $$idl_info{ MAIN_IDL } ;
        $handle->print( " \\\n    $$idl_info{ GENERATED_SOURCE }" ) ;
    }
    if ( defined( $idl_list->class_kinds() ) ) {
        $handle->print( " \\\n    " . FileNames::class_list_source() ) ;
    }
    $handle->print( "\n" ) ;


    ## Factgen includes.

    $handle->print( "\n# Include directories.\n\n.-FACTGEN_INCLUDES =" ) ;
    my $includes = $idl_list->all_includes() ;
    foreach my $include ( @$includes ) {
        my $directory = Utils::relative_path( $include, $directory ) ;
        $directory =~ s|/|\\/|g ;
        if ( length( $directory ) ) {
            $handle->print( ' -I ' . $directory ) ;
        }
    }
    $handle->print( "\n" ) ;


    ## Generated metaclasses.

    $handle->print( <<EOT

# Rules to generate local metaclasses.
EOT
    ) ;

    foreach my $idl ( @all_idls ) {
        my $idl_info = $$idl_infos{ $idl } ;
        next unless $$idl_info{ MAIN_IDL } ;
        my $headers = join( ' ', @{$$idl_info{ HEADERS }} ) ;
        my $rpc_depends = join( ' ', @{$$idl_info{ PARENT_RPC_HEADERS }} ) ;
        $handle->print( <<EOT

replace \@BASE_NAME\@ $$idl_info{ BASE_NAME }
replace \@INCLUDE_HEADERS\@ $headers
replace \@PARENT_RPC_HEADERS\@ $rpc_depends
generate_metaclasses
EOT
        ) ;
    }

    
    ## Extern metaclasses.

    $handle->print( <<EOT

# Dependencies for extern metaclasses.
EOT
    ) ;

    foreach my $idl ( @all_idls ) {
        my $idl_info = $$idl_infos{ $idl } ;
        next if $$idl_info{ MAIN_IDL } ;
        my $subproject_directory = Utils::relative_path(
            $directory . '/' . $$idl_info{ PATH }, $directory
        ) . '/' ;
        my $relative_base_name = Utils::relative_path(
            $directory . '/' . $$idl_info{ BASE_NAME }, $directory . '/' . $subproject_directory
        ) ;
        my $headers = join( ' ', @{$$idl_info{ HEADERS }} ) ;
        my $rpc_depends = join( ' ', @{$$idl_info{ PARENT_RPC_HEADERS }} ) ;
        $handle->print( <<EOT

replace \@EXTERN_PROJECT_PATH\@ $subproject_directory
replace \@BASE_NAME\@ $relative_base_name
replace \@INCLUDE_HEADERS\@ $headers
replace \@PARENT_RPC_HEADERS\@ $rpc_depends
extern_metaclasses
EOT
        ) ;
    }


    ## Class list.

    if ( defined( $idl_list->class_kinds() ) ) {
        $handle->print( "\n# Generate class list.\n\ngenerate_class_list\n" ) ;
    }


    ## Extern class list.

    my $extern_class_list_project ;
    my $subprojects = $idl_list->subprojects() ;
    foreach my $subproject ( keys( %$subprojects ) ) {
        my $subproject_list = $$subprojects{ $subproject } ;
        next unless defined( $subproject_list->class_kinds() ) ;
        if ( defined( $idl_list->class_kinds() ) ) {
            die( "Class list requested both in main project '$directory' and its\n" .
                 "subproject '$directory/$subproject'.\n" ) ;
        }
        if ( defined( $extern_class_list_project ) ) {
            die( "Class list requested in multiple subprojects of project '$directory':\n" .
                 "'$extern_class_list_project' and '$subproject'.\n" ) ;
        }

        $extern_class_list_project = $subproject ;

        my $subproject_directory = Utils::relative_path(
            $directory . '/' . $subproject, $directory
        ) . '/' ;
        $handle->print( <<EOT

# External class list.

replace \@EXTERN_PROJECT_PATH\@ $subproject_directory
extern_class_list
EOT
        ) ;
    }

    # Done.

    $handle->close() ;
}


###
# _gather_idl_info {
sub _gather_idl_info {
    my $self = shift ;
    my $idl_list = $$self{ IDL_LIST } ;
    my $directory = $idl_list->directory() ;

    my %results ;
    my $all_idls = $idl_list->all_idls() ;
    foreach my $idl ( keys( %$all_idls ) ) {
        my $idl_info = $$all_idls{ $idl } ;

        # Subproject path.

        my $main_idl = $$idl_info{ MAIN_IDL } ;
        my $subproject_path = Utils::relative_path(
            $$idl_info{ IDL_LIST }->directory(), $directory
        ) ;

        # File names.

        my $complete_name = $$idl_info{ COMPLETE_NAME } ;
        my $file_name = Utils::relative_path( $complete_name, $directory ) ;
        my $base_name = FileNames::base( $file_name ) ;

        # Headers.

        my @headers = map( { Utils::relative_path( $_, $directory ) }
            @{$$idl_info{ HEADERS }} ) ;

        # Parent RPC headers.

        my $rpc_header = FileNames::rpc_header( $base_name ) ;

        my %rpc_headers ;
        my $idl_classes = $$idl_info{ CLASSES } ;
        foreach my $class ( @$idl_classes ) {
            my $parents = $class->parents() ;
            foreach my $parent ( @$parents ) {
                my $parent_class = $$parent{ SYMBOL } ;
                my $file_name = Utils::relative_path(
                    $parent_class->file_name(), $directory
                ) ;
                $file_name = FileNames::rpc_header( FileNames::base( $file_name ) ) ;
                if ( $file_name ne $rpc_header ) {
                    $rpc_headers{ $file_name }++ ;
                }
            }
        }

        # Create info structure for this IDL.

        my @rpc_headers_array = sort( keys( %rpc_headers ) ) ;
        my $result = {
            MAIN_IDL => $main_idl,
            CLASSES => $idl_classes,
            HEADERS => \@headers,
            PATH => $subproject_path,
            FILE_NAME => $file_name,
            BASE_NAME => $base_name,
            GENERATED_SOURCE => FileNames::generated_source( $base_name ),
            GENERATED_HEADER => FileNames::generated_header( $base_name ),
            RPC_HEADER => $rpc_header,
            PARENT_RPC_HEADERS => \@rpc_headers_array,
        } ;
        $results{ $complete_name } = $result ;
    }

    $$self{ IDL_INFOS } = \%results ;
}


1 ;


## EOF ##
