=head1 NAME

IdlList - parser of idl.list files.

=head1 SYNOPSIS

    use IdlList ;

    # Constructor
    $idl_list = IdlList->new() ;

    # Parse main idl.list.
    $idl_list->parse( $directory ) ;

    # Parse subproject idl lists.
    $idl_list->parse_subprojects() ;

    # Parse IDLs specified in all_idls().
    # Subproject IDLs will be included if parse_subprojects() has been called.
    $idl_list->parse_idls() ;

    # Get directory.
    $directory = $idl_list->directory() ;

    # Get reference to array of project directories this project depends on.
    $depends = $idl_list->depends() ;

    # Get reference to array of paths to search when including idls.
    $includes = $idl_list->includes() ;

    # Get reference to array of idl files that belong to this project.
    # The array contains plain idl file names as read from idl.list.
    $idls = $idl_list->idls() ;

    # Should class list be generated here?
    # Return reference to array of class kind names to include in the class
    # list or undef if class list should not be generated.
    $class_kinds = $idl_list->class_kinds() ;

    # Get class kinds as numbers.
    # IDLs must be parsed before this call and symbol 'enum ::ClassKind'
    # must be defined in one of the IDL.
    $class_kind_numbers = $idl_list->class_kind_numbers() ;

    ###
    # Must parse_subprojects() first to make these available.
    # all_* will include only main includes/idls otherwise.

    # Get reference to map of subproject IdlList objects.
    # Keys are subproject directory names (relative to directory()),
    # values are IdlList objects.
    $subprojects = $idl_list->subprojects() ;

    # Get reference to array of main and subproject includes, relative to
    # current directory (ie. including $directory).
    $all_includes = $all_includes() ;

    # Get reference to set of information about main and subproject idls.
    # Keys are plain idl file names, values are structures:
    #   MAIN_IDL - true iff IDL has been specified in main (primary) idl.list.
    #   IDL_LIST - reference to IDL list object.
    #   COMPLETE_NAME - name relative to current directory.
    #   CLASSES - array of refences to class symbols of classes defined
    #             in the IDL.
    #   HEADERS - array of additional headers to include to generated header.
    #             This is copy of idl_infos(){ HEADERS } array from the parser
    #             plus header_name()s of all CLASSES, relative to current dir.
    #
    # Note that COMPLETE_NAME, CLASSES and HEADERS are available only if
    # parse_idls() has been called.
    $all_idls = $idl_list->all_idls() ;

    ###
    # Must parse_idls() first to make these available

    # Get reference to IdlParser
    $idl_parser = $idl_list->parser() ;

    # Get reference to map of class infos.
    # Keys are fully qualified class names, values are ClassInfo objects.
    $class_infos = $idl_list->class_infos() ;

=head1 DESCRIPTION

C<IdlList> parses F<idl.list> file in specified directory.

=head1 KNOWN BUGS

None.

=cut

package IdlList ;

use strict ;
use warnings ;

use FileHandle ;

use IdlParser ;
use ClassInfo ;
use FileNames ;
use Utils ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate.

    my $self = {
        DIRECTORY => undef,
        DEPENDS => [ ],
        INCLUDES => [ ],
        IDLS => [ ],
        CLASS_KINDS => undef,
        SUBPROJECTS => { },
        ALL_INCLUDES => undef,
        ALL_IDLS => undef,
        PARSER => undef,
        CLASS_INFOS => undef,
        IDL_CLASSES => undef,
    } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# parse
sub parse  {
    my $self = shift ;
    my $directory = shift ;

    die( "Use '.' instead of '' for current directory.\n" ) if $directory eq '' ;
    $$self{ DIRECTORY } = $directory ;
    my $file_name = Utils::simplify_path( $directory . '/' . FileNames::idl_list() ) ;
    my $handle = FileHandle->new( $file_name, 'r' ) ;
    unless ( defined( $handle ) ) {
        die( "Can't open '$file_name'\n" ) ;
    }

    local $_ ;
    while ( $_ = Utils::get_line( $handle ) ) {
        s/^\s*// ;
        s/\s*$// ;
        
        if ( my( $depends ) = /^depends\s+(.*)/ ) {
            foreach my $depend ( split( /\s+/, $depends ) ) {
                push( @{$$self{ DEPENDS }}, $depends ) ;
            }
        }
        elsif ( my( $directories ) = /^directory\s+(.*)/ ) {
            foreach my $directory ( split( /\s+/, $directories ) ) {
                push( @{$$self{ INCLUDES }}, $directory ) ;
            }
        }
        elsif ( my( $idls ) = /^idl\s+(.*)/ ) {
            foreach my $idl ( split( /\s+/, $idls ) ) {
                push( @{$$self{ IDLS }}, $idl ) ;
            }
        }
        elsif ( my( $kinds ) = /^class_list\s+(.*)$/ ) {
            my @class_kinds = split( /\s+/, $kinds ) ;
            $$self{ CLASS_KINDS } = \@class_kinds ;
        }
        else {
            die "'$file_name': Unknown idl.list command '$_'.\n" ;
        }
    }

    $handle->close() ;
}


###
# parse_subprojects
sub parse_subprojects {
    my $self = shift ;

    my $directory = $$self{ DIRECTORY } ;
    my $depends = $$self{ DEPENDS } ;
    my $subprojects = $$self{ SUBPROJECTS } ;
    foreach my $depend ( @$depends ) {
        my $idl_list = $self->new() ;
        $idl_list->parse( Utils::simplify_path( $directory . '/' . $depend ) ) ;
        $$subprojects{ $depend } = $idl_list ;
    }
}


###
# parse_idls
sub parse_idls {
    my $self = shift ;
    my $includes = $self->all_includes() ;
    my $idls = $self->all_idls() ;

    # Create parser.

    my $parser = IdlParser->new() ;
    $$self{ PARSER } = $parser ;
    foreach my $include ( @$includes ) {
        $parser->add_include_path( $include ) ;
    }

    foreach my $idl ( keys( %$idls ) ) {
        my $complete_name = $parser->import_file( $idl ) ;
        $$idls{ $idl }{ COMPLETE_NAME } = $complete_name ;
        $parser->parse() ;
    }

    $self->_gather_idl_classes() ;
}


###
# directory
sub directory {
    return $_[ 0 ]{ DIRECTORY } ;
}


###
# depends
sub depends {
    return $_[ 0 ]{ DEPENDS } ;
}


###
# includes
sub includes {
    return $_[ 0 ]{ INCLUDES } ;
}


###
# idls
sub idls {
    return $_[ 0 ]{ IDLS } ;
}


###
# class_kinds
sub class_kinds {
    return $_[ 0 ]{ CLASS_KINDS } ;
}


###
# class_kind_numbers
sub class_kind_numbers {
    my $self = shift ;

    my $class_kinds = $$self{ CLASS_KINDS } ;
    unless ( defined( $class_kinds ) ) {
        return undef ;
    }

    my $parser = $$self{ PARSER } ;
    if ( defined( $parser ) ) {
        my $enums = $parser->enums() ;
        my $class_kind_enum = $$enums{ '::ClassKind' } ;
        if ( defined( $class_kind_enum ) ) {
            # Convert names to values.

            my @result ;
            foreach my $kind_name ( @$class_kinds ) {
                my $kind_symbol = $class_kind_enum->get_symbol( $kind_name ) ;
                die( "Class kind '$kind_name' not defined in IDL.\n" )
                    unless defined( $kind_symbol ) ;
                push( @result, $kind_symbol->value() ) ;
            }
            return \@result ;
        }
        else {
            die( "enum ::ClassKind not defined in IDL.\n" ) ;
        }
    }
    else {
        die( "Must parse IDLs first.\n" ) ;
    }
}


###
# subprojects
sub subprojects {
    return $_[ 0 ]{ SUBPROJECTS } ;
}


###
# all_includes
sub all_includes {
    my $self = shift ;
    unless ( defined( $$self{ ALL_INCLUDES } ) ) {
        $self->_gather_includes() ;
    }
    return $$self{ ALL_INCLUDES } ;
}


###
# all_idls
sub all_idls {
    my $self = shift ;
    unless ( defined( $$self{ ALL_IDLS } ) ) {
        $self->_gather_idls() ;
    }
    return $$self{ ALL_IDLS } ;
}


###
# parser
sub parser {
    return $_[ 0 ]{ PARSER } ;
}


###
# class_infos
sub class_infos {
    my $self = shift ;
    unless ( defined( $$self{ CLASS_INFOS } ) ) {
        $self->_create_class_infos() ;
    }
    return $$self{ CLASS_INFOS } ;
}


###
# _gather_includes
sub _gather_includes {
    my $self = shift ;
    my $directory = $$self{ DIRECTORY } ;

    # Main includes.

    my %all_includes ;
    $all_includes{ Utils::simplify_path( $directory ) }++ ;
    foreach my $include ( @{$$self{ INCLUDES }} ) {
        $all_includes{ Utils::simplify_path( $directory . '/' . $include ) }++ ;
    }

    # Subproject includes.

    my $subprojects = $$self{ SUBPROJECTS } ;
    foreach my $subproject_directory ( keys( %$subprojects ) ) {
        my $subproject = $$subprojects{ $subproject_directory } ;
        $all_includes{ Utils::simplify_path( $directory . '/' . $subproject_directory ) }++ ;
        foreach my $include ( @{$subproject->includes()} ) {
            $all_includes{ Utils::simplify_path( $directory . '/' . $subproject_directory . '/' . $include ) }++ ;
        }
    }

    @{$$self{ ALL_INCLUDES }} = sort( keys( %all_includes ) ) ;
}


###
# _gather_idls
sub _gather_idls {
    my $self = shift ;
    my $directory = $$self{ DIRECTORY } ;

    # Main idls.

    my %all_idls ;
    foreach my $idl ( @{$$self{ IDLS }} ) {
        $all_idls{ $idl } = {
            MAIN_IDL => 1,
            IDL_LIST => $self,
            COMPLETE_NAME => undef,
            CLASSES => [ ],
            HEADERS => [ ],
        } ;
    }

    # Subproject includes.

    my $subprojects = $$self{ SUBPROJECTS } ;
    foreach my $subproject_directory ( keys( %$subprojects ) ) {
        my $subproject = $$subprojects{ $subproject_directory } ;

        foreach my $idl ( @{$subproject->idls()} ) {
            # Check if IDL is not in multiple lists.

            if ( defined( $all_idls{ $idl } ) ) {
                my $idl_info = $all_idls{ $idl } ;
                my $cur_directory = Utils::simplify_path(
                    $directory . '/' . $subproject_directory
                ) ;
                my $prev_directory = $$idl_info{ IDL_LIST }->directory() ;
                
                warn( "IDL '$idl' specified in both '$prev_directory' and '$cur_directory'.\n" ) ;
                next ;
            }

            # Add info about the IDL.

            $all_idls{ $idl } = {
                MAIN_IDL => 0,
                IDL_LIST => $subproject,
                COMPLETE_NAME => undef,
                CLASSES => [ ],
                HEADERS => [ ],
            } ;
        }
    }

    $$self{ ALL_IDLS } = \%all_idls ;
}


###
# _create_class_infos
sub _create_class_infos {
    my $self = shift ;
    my $parser = $$self{ PARSER } or die( "No parser.\n" ) ;
    my $classes = $parser->classes() ;
    my $class_infos = { } ;

    foreach my $class_name ( keys( %$classes ) ) {
        my $class_symbol = $$classes{ $class_name } ;
        my $class_info = ClassInfo->new( $class_symbol ) ;
        $$class_infos{ $class_name } = $class_info ;
    }

    $$self{ CLASS_INFOS } = $class_infos ;
}


###
# _gather_idl_classes
sub _gather_idl_classes {
    my $self = shift ;
    my $all_idls = $self->all_idls() ;
    my $parser = $$self{ PARSER } or die( "No parser.\n" ) ;
    my $classes = $parser->classes() ;

    # all_idls keys are simple names, we want complete names as keys.

    my %idl_infos ;
    foreach my $idl ( keys( %$all_idls ) ) {
        my $idl_info = $$all_idls{ $idl } ;
        $idl_infos{ $$idl_info{ COMPLETE_NAME } } = $idl_info ;
    }

    my %idl_headers ;
    
    # Copy IDL infos from parser.

    my $parser_idl_infos = $parser->idl_infos() ;
    foreach my $idl_name ( keys( %$parser_idl_infos ) ) {
        my $parser_idl_info = $$parser_idl_infos{ $idl_name } ;
        my $idl_info = $idl_infos{ $idl_name } ;
        unless ( defined( $idl_info ) ) {
            die( "'$idl_name' has been read by the parser but it's not listed in asy idl.list.\n" ) ;
        }

        my $project_directory = $$idl_info{ IDL_LIST }->directory() ;
        die( "Invalid project directory '$project_directory'.\n" ) unless $project_directory ;
        foreach my $header_name ( @{$$parser_idl_info{ HEADERS }} ) {
            my $relative_name = Utils::simplify_path(
                $project_directory . '/' . $header_name
            ) ;
            $idl_headers{ $idl_name }{ $relative_name }++ ;
        }
    }

    # Gather classes.

    foreach my $class_name ( keys( %$classes ) ) {
        my $class_symbol = $$classes{ $class_name } ;
        my $idl_name = $class_symbol->file_name() ;
        my $idl_info = $idl_infos{ $idl_name } ;
        unless ( defined( $idl_info ) ) {
            die( "Class '$class_name' defined in '$idl_name', but there is no info for such file.\n" .
                 "Have you forgotten to put the file in idl.list?\n" ) ;
        }

        push( @{$$idl_info{ CLASSES }}, $class_symbol ) ;
        $idl_headers{ $idl_name }{ $class_symbol->header_name() }++ ;
    }

    # Copy keys of %idl_headers.

    foreach my $idl_name ( keys( %idl_headers ) ) {
        my $idl_info = $idl_infos{ $idl_name } ;
        my @headers = sort( keys( %{$idl_headers{ $idl_name }} ) ) ;
        $$idl_info{ HEADERS } = \@headers ;
    }
}


1 ;


__END__
