=head1 NAME

IdlParser - simple top-down IDL parser class

=head1 SYNOPSIS

    use IdlParser ;

    # Parent class is Parser - see Parser.pm

    # Constructor
    $parser = IdlParser->new() ;

    # Parsing.
    $parser->parse() ;

    # Get reference to root namespace symbol.
    $root = $parser->root() ;

    # Get set of symbols of given type.
    # Keys are fully qualified symbol names and values are references
    # to symbol objects.
    $enums = $parser->enums() ;
    $attributes = $parser->attributes() ;
    $classes = $parser->classes() ;

    # Get reference to hash of information about IDL files.
    # Keys are file names (relative to current directory), values are structures:
    #   HEADERS - array of additional headers to include into generated files
    #             (as read from .idl file, should be relative to project directory)
    $idl_infos = $parser->idl_infos() ;

=head1 DESCRIPTION

C<IdlParser> extends C<Parser> and is able to parse files in Massiv IDL syntax.

The most simple way to parse file is:

    # Create parser.

    $parser = IdlParser->new() ;

    # Add several paths to search when including files.

    $parser->add_include_path( $path ) ;

    # Parse file.

    $parser->push_file( $file ) ;
    $parser->parse() ;    

All information about parsed files will be stored in single symbol table
(C<root_namespace()>) - it will contain symbol objects describing nested
namespaces, classes, etc. See F<SymbolBase.pm> and other "Symbol*" files for
information about all available symbol class types.

It's possible to access list of all known enumeration types, attributes and
classes using C<enums()>, C<attributes()> and C<classes()>.

There is a nasty hack in C<IdlParser> - it's method C<_finish_symbol()>,
which checks if all required attributes are defined, if their values make
sense, etc. It may be necessary to change this method when new attributes
are defined in IDL sources.

=head1 KNOWN BUGS

There is single symbol table per namespace/class - it's not possible to
declare subclass with the same name as property.

=cut

package IdlParser ;

use strict ;
use warnings ;

use FileHandle ;
use POSIX qw( strtod ) ;

use Parser ;
our( @ISA ) ;
@ISA = qw( Parser ) ;

use MassivType ;
use SymbolBase ;
use SymbolTable ;
use SymbolAttribute ;
use SymbolConstant ;
use SymbolProperty ;
use SymbolMethod ;
use SymbolClass ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Parent constructor.
    
    my $self = $class->SUPER::new() ;

    # Root namespace.
    $$self{ ROOT } = SymbolNamespace->new( '' ) ;

    # Symbol table stack.
    # Elements are structures:
    #   SYMBOL - reference to symbol table symbol.
    #   ATTRIBUTES - attribute defaults defined in current block
    #                it's a hash with fully qualified attribute names as keys
    #                and values equal to default attribute values
    $$self{ SYMBOL_STACK } = [ {
        SYMBOL => $$self{ ROOT },
        ATTRIBUTES => { },
    } ] ;

    # All definitions of symbols of given type.
    # Keys are fully qualified names, values are references to the symbols.
    $$self{ ENUMS } = { } ;
    $$self{ ATTRIBUTES } = { } ;
    $$self{ CLASSES } = { } ;

    # Information about IDL files.
    $$self{ IDL_INFOS } = { } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# dump
sub dump {
    my $self = shift ;
    $$self{ ROOT }->dump( 0 ) ;
}


###
# root
sub root {
    return $_[ 0 ]{ ROOT } ;
}


###
# enums
sub enums {
    return $_[ 0 ]{ ENUMS } ;
}


###
# attributes
sub attributes {
    return $_[ 0 ]{ ATTRIBUTES } ;
}


###
# classes
sub classes {
    return $_[ 0 ]{ CLASSES } ;
}


###
# idl_infos
sub idl_infos {
    return $_[ 0 ]{ IDL_INFOS } ;
}


###
# parse
sub parse {
    my $self = shift ;

    # Read in first token.

    $self->next_token() ;
    $self->_parse_global() ;

    # Current token should be undefined. Otherwise _parse_global returned
    # on '}', but root namespace can't be closed.

    if ( defined( $self->token() ) ) {
        if ( $self->token_string() ne '}' ) {
            $self->internal_error( "Token should be '}'." ) ;
        }
        $self->error( "Can't close root namespace." ) ;
    }
}


###
# _required_attributes
# Attributes required to generate sources.
my %_required_attributes = (
    'SymbolClass' => [
        '::abstract',
        '::assignable',
        '::client_server_migrate',
        '::comparable',
        '::has_strong_pointer',
        '::kind',
        '::no_archive',
        '::no_balance',
        '::no_migrate',
        '::node_object',
        '::root',
        '::simulation_startup_notify',
        '::symbol_attribute',
        '::throwable',
        '::tracked',
        '::value_type',
    ],
    'SymbolProperty' => [
        '::repflags',
        '::ptr_repflags',
        '::strongptr_repflags',
    ],
    'SymbolMethod' => [
        '::const',
        '::virtual',
    ],
) ;


###
# _finish_symbol
sub _finish_symbol {
    my $self = shift ;
    my $symbol = shift ;
    my $type = ref( $symbol ) ;

    my $attributes ;

    # Check required attributes.

    foreach my $reqattr_type ( keys( %_required_attributes ) ) {
        next if $type ne $reqattr_type ;
        $attributes = $symbol->attributes() ;
        my $required_attributes = $_required_attributes{ $reqattr_type } ;

        foreach my $required_attribute ( @$required_attributes ) {
            unless ( defined( $$attributes{ $required_attribute } ) ) {
                my $type_name = $symbol->type_name() ;
                $self->error( $self->_error_symbol( $symbol,
                    "Required $type_name attribute '$required_attribute' is not defined." ) ) ;
            }
        }
    }

    # Check class attribute values.

    if ( $type eq 'SymbolClass' ) {
        # Can't change 'kind' if parent is non-shared.

        my $kind = $$attributes{ '::kind' } ;
        my $parents = $symbol->parents() ;
        foreach my $parent ( @$parents ) {
            my $parent_symbol = $$parent{ SYMBOL } ;
            my $parent_attributes = $parent_symbol->attributes() ;
            my $parent_kind = $$parent_attributes{ '::kind' } ;
            if ( ( $parent_kind != 0 ) && ( $parent_kind != $kind ) ) {
                my $my_name = $symbol->full_name() ;
                my $parent_name = $parent_symbol->full_name() ;
                $self->warning( $self->_error_symbol( $symbol,
                    "Class '$my_name' kind ($kind) is different from parent class " .
                    "'$parent_name' non-shared kind ($parent_kind)."
                ) ) ;
            }
        }

        # Throwables must be value types.

        my $throwable = $$attributes{ '::throwable' } ;
        my $value_type = $$attributes{ '::value_type' } ;
        if ( $throwable && ! $value_type ) {
            $self->error( $self->_error_symbol( $symbol,
                "Throwable must be value type." ) ) ;
        }

        # Assign has_strong_pointer attribute value.

        my $strong_pointer_property = $symbol->find_strong_pointer() ;
        $$attributes{ '::has_strong_pointer' } = defined( $strong_pointer_property ) ;

        # Value types can't contain strong pointers.

        if ( $value_type && defined( $strong_pointer_property ) ) {
            $self->error( $self->_error_symbol( $strong_pointer_property,
                        "Value types can't contain strong pointers." ) ) ;
        }
    }

    # Check property attribute values.

    if ( $type eq 'SymbolProperty' ) {
        my $repflags = $symbol->require_attribute( '::repflags' ) ;

        # Check repflags - MIGRATE bit must be always set.
        # FIXME: Get MIGRATE value from the enum?

        if ( ! ( $repflags & 0b0001 ) ) {
            $self->error( $self->_error_symbol( $symbol,
                "MIGRATE bit in 'repflags' attribute is not set." ) ) ;
        }

        # Check ptr_repflags if property is pointer.

        my $property_type = $symbol->property_type() ;
        my $property_type_name = ref( $property_type ) ;
        if ( $property_type_name eq 'TypePointer' ) {
            my $ptr_repflags = $symbol->require_attribute( '::ptr_repflags' ) ;
            my $kind = $property_type->kind() ;

            # Use default from strongptr_repflags if ptr_repflags is zero and pointer is strong.

            if ( ( $kind eq 'strong' ) && ( $ptr_repflags == 0 ) ) {
                $$attributes{ '::ptr_repflags' } = $symbol->require_attribute( '::strongptr_repflags' ) ;
            }

            # MIGRATE bit must be set if ptr_repflags is non-zero.

            if ( ( $ptr_repflags != 0 ) && ! ( $ptr_repflags & 0b0001 ) ) {
                $self->error( $self->_error_symbol( $symbol,
                    "MIGRATE bit in non-zero 'ptr_repflags' attribute is not set." ) ) ;
            }
        }
    }
}


###
# _error_symbol
# Produce argument set for "previous definition here" type of messages.
#   symbol - set FILE_NAME and LINE_NUMBER from this
#   message - set MESSAGE to this
# Creates set of arguments (CURRENT_LINE will be set to '') to pass to
# Parser warning/error message methods.
sub _error_symbol {
    my( $self, $symbol, $message ) = @_ ;
    return {
        MESSAGE => $message,
        FILE_NAME => $symbol->file_name(),
        LINE_NUMBER => $symbol->line_number(),
        CURRENT_LINE => '',
    } ;
}


###
# _symbol_stack_top
# Get reference to top of symbol table stack.
sub _symbol_stack_top {
    my $self = shift ;
    my $sstack = $$self{ SYMBOL_STACK } ;
    my $num_symbols = scalar( @$sstack ) ;
    if ( $num_symbols == 0 ) {
        $self->internal_error( "Symbol table stack is empty." ) ;
    }
    return $$sstack[ $num_symbols - 1 ] ;
}


###
# _symbol_table
# Get reference to current symbol table (symbol stack top table).
sub _symbol_table {
    my $self = shift ;
    my $stop = $self->_symbol_stack_top() ;
    return $$stop{ SYMBOL } ;
}


###
# _enter_symbol
# Push given symbol table to symbol table stack.
#   symbol - reference to symbol
sub _enter_symbol {
    my $self = shift ;
    my $symbol = shift ;
    my $sstack = $$self{ SYMBOL_STACK } ;
    push( @$sstack, {
        SYMBOL => $symbol,
        ATTRIBUTES => { },
    } ) ;
}


###
# _leave_symbol
# Pop top entry from symbol table stack.
sub _leave_symbol {
    my $self = shift ;
    my $sstack = $$self{ SYMBOL_STACK } ;
    if ( scalar( @$sstack ) == 0 ) {
        $self->internal_error( "Symbol table stack is empty." ) ;
    }
    pop( @$sstack ) ;
}


###
# _find_symbol
# Find symbol in current or parent symbol table.
#   name - symbol name as a list
# Returns reference to the symbol or undef if not found.
sub _find_symbol {
    my $self = shift ;
    my $symbol_table = $self->_symbol_table() ;
    return $symbol_table->find_symbol_list( @_ ) ;
}


###
# _get_symbol
# Get symbol from current symbol table.
#   name - simple symbol name
# Unlike _find_symbol, this method only looks into current symbol table
# and does not support qualified symbol names.
sub _get_symbol {
    my $self = shift ;
    my $name = shift ;
    my $symbol_table = $self->_symbol_table() ;
    return $symbol_table->get_symbol( $name ) ;
}


###
# _create_symbol
# Create symbol in current namespace.
#   type - symbol type name
#   name - symbol name
# Dies with error if symbol with the same name but different type already exists.
sub _create_symbol {
    my( $self, $type, $name ) = @_ ;
    my $symbol_table = $self->_symbol_table() ;
    return $symbol_table->create_symbol( $self, $type, $name ) ;
}


###
# _create_new_symbol
# Create new symbol in current namespace.
#   type - symbol type name
#   name - symbol name
# Dies with error if symbol with the same name already exists.
sub _create_new_symbol {
    my( $self, $type, $name ) = @_ ;
    my $symbol_table = $self->_symbol_table() ;
    return $symbol_table->create_new_symbol( $self, $type, $name ) ;
}


###
# _parse_symbol
# Parse symbol name.
# Expects identifiers separated by "::" tokens. It may start with "::" token.
# Parsing will stop at first token following the symbol.
# Returns symbol name as a list.
sub _parse_symbol {
    my $self = shift ;
    my @symbol_name ;

    # Read first token and check if it's "::".

    $self->require_token() ;
    if ( $self->token_string eq '::' ) {
        push( @symbol_name, '' ) ;
        $self->next_token() ;
    }

    # Read the symbol.

    TOKEN : for ( ; ; ) {
        # Read the identifier and push it.

        $self->ensure_identifier() ;
        push( @symbol_name, $self->token_value() ) ;
        $self->next_token() ;

        # Check if it's followed by "::".

        my $token = $self->token() ;
        if ( ! defined( $token ) || ( $token->token() ne '::' ) ) {
            last TOKEN ;
        }
        $self->next_token() ;
    }

    return @symbol_name ;
}


###
# _parse_and_find_symbol
# Parse symbol and then try to find it.
# Returns reference to the symbol, or dies with error.
sub _parse_and_find_symbol {
    my $self = shift ;
    my @symbol_name = $self->_parse_symbol() ;
    my $symbol = $self->_find_symbol( @symbol_name ) ;
    unless ( defined( $symbol ) ) {
        my $name = join( '::', @symbol_name ) ;
        $self->error( "Can't find symbol '$name'." ) ;
    }
    return $symbol ;
}


###
# _ignore_semicolon
sub _ignore_semicolon {
    my $self = shift ;
    $self->warning( {
        MESSAGE => "Superfluous semicolon.",
        LEVEL => 3,
    } ) ;
    $self->consume_token( ';' ) ;
}


# Other private methods.

require IdlParserAttribute ;
require IdlParserClass ;
require IdlParserEnum ;
require IdlParserNamespace ;
require IdlParserType ;
require IdlParserValue ;
require IdlParserFile ;


1 ;


__END__
