=head1 NAME

IdlParserAttribute - private attribute parsing methods of C<IdlParser>.

=cut

package IdlParser ;


###
# _set_attribute_default
#   name - fully qualified attribute name
#   value - new default value
# Set block-scoped default value of given attribute.
sub _set_attribute_default {
    my( $self, $name, $value ) = @_ ;
    my $sstop = $self->_symbol_stack_top() ;
    $$sstop{ ATTRIBUTES }{ $name } = $value ;
}


###
# _get_attribute_default
#   name - fully qualified attribute name
#   symbol - optional class symbol.
# Get block-scoped default value of given attribute. If default value has not
# been defined for current block, parent blocks will be searched. If no default
# value is found at block-scope and symbol is defined and it's a class and
# attribute is inherited class attribute, value from parents will be returned.
# Otherwise default value from attribute type definition will be returned.
sub _get_attribute_default {
    my( $self, $name, $symbol ) = @_ ;

    my $attribute = $$self{ ATTRIBUTES }{ $name } ||
        $self->internal_error( "There is no attribute '$name'." ) ;

    # Try block-defaults.
    
    my $sstack = $$self{ SYMBOL_STACK } ;
    for ( my $i = $#{$sstack} ; $i >= 0 ; $i-- ) {
        my $default = $$sstack[ $i ]{ ATTRIBUTES }{ $name } ;
        if ( defined( $default ) ) {
            return $default ;
        }
    }

    # Try inherited value.

    if ( defined( $symbol ) && ( ref( $symbol ) eq 'SymbolClass' ) &&
         ( ref( $attribute ) eq 'SymbolClassAttribute' ) ) {
        my $flags = $attribute->attribute_flags() ;
        if ( $$flags{ 'inherit' } ) {
            # Inherited values must be the same in all parents.

            my $value ;
            my $parents = $symbol->parents() ;
            PARENT : foreach my $parent ( @$parents ) {
                my $parent_class = $$parent{ SYMBOL } ;
                my $parent_attributes = $parent_class->attributes() ;
                my $parent_attribute = $$parent_attributes{ $name } ;
                next PARENT unless defined( $parent_attribute ) ;

                if ( defined( $value ) && ( $value != $parent_attribute ) ) {
                    $self->error( "Ambiguous inherited value of class attribute '$name'." ) ;
                }
                $value = $parent_attribute ;
            }

            return $value if defined( $value ) ;
        }
    }

    # Try attribute-default.

    return $attribute->default_value() ;
}


###
# _parse_attribute_type
# Parse name of attribute type - either builtin or enum symbol name.
# Returns name of builtin type or reference to type symbol.
sub _parse_attribute_type {
    my $self = shift ;

    my $type_name = $self->token_string() ;

    # Bultin types.
    
    if ( ( $type_name eq 'bool' ) || ( $type_name eq 'int' ) || ( $type_name eq 'string' ) ) {
        $self->next_token() ;
        return $type_name ;
    }

    # Try symbol.

    my $symbol = $self->_parse_and_find_symbol() ;
    if ( ref( $symbol ) ne 'SymbolEnum' ) {
        my $full_name = $symbol->full_name() ;
        $type_name = $symbol->type_name() ;
        $self->error( {
            MESSAGE => "Symbol '$type_name $full_name' is not 'enum'.",
            NO_DIE => 1,
        } ) ;
        $self->error( $self->_error_symbol( $symbol, "Defined here." ) ) ;
    }

    # Done.

    return $symbol ;
}


###
# _parse_attribute_definition
# Parse attribute type definition - classattr, propattr or methodattr
sub _parse_attribute_definition {
    my $self = shift ;

    # Which attribute kind?

    $self->ensure_identifier() ;
    my $kind_name = $self->token_value() ;
    my $kind ;
    if ( $kind_name eq 'classattr' ) {
        $kind = 'SymbolClassAttribute' ;
    }
    elsif ( $kind_name eq 'propattr' ) {
        $kind = 'SymbolPropertyAttribute' ;
    }
    elsif ( $kind_name eq 'methodattr' ) {
        $kind = 'SymbolMethodAttribute' ;
    }
    else {
        $self->internal_error( "Unknown attribute kind '$kind_name'." ) ;
    }
    $self->next_token() ;

    # Optional attribute flags.

    my %attribute_flags ;
    if ( $self->token_string() eq '<' ) {
        $self->next_token() ;
        
        ATTRIBUTE_FLAG : while ( $self->token_string() ne '>' ) {
            # Parse attribute flag.

            my $flag_token ;
            if ( $kind_name eq 'classattr' ) {
                $flag_token = $self->consume_token( 'inherit', 'idl_internal' ) ;
            }
            elsif ( $kind_name eq 'propattr' ) {
                $flag_token = $self->consume_token( 'idl_internal' ) ;
            }
            elsif ( $kind_name eq 'methodattr' ) {
                $flag_token = $self->consume_token( 'idl_internal' ) ;
            }
            $attribute_flags{ $flag_token->value() }++ ;

            # Another flag?

            if ( $self->token_string() eq '>' ) {
                last ATTRIBUTE_FLAG ;
            }
            $self->consume_token( ',' ) ;
        }

        $self->consume_token( '>' ) ;
    }

    # Attribute type.

    my $type = $self->_parse_attribute_type() ;

    # Read attribute name and create it.

    $self->ensure_identifier() ;
    my $name = $self->token_value() ;
    my $attribute = $self->_create_new_symbol( $kind, $name ) ;
    $$self{ ATTRIBUTES }{ $attribute->full_name() } = $attribute ;
    $self->next_token() ;

    # Read default attribute value.

    $self->consume_token( '=' ) ;
    my $value = $self->_parse_value( $type ) ;
    $self->consume_token( ';' ) ;

    # Initialize the attribute.

    $attribute->attribute_flags( \%attribute_flags ) ;
    $attribute->attribute_type( $type ) ;
    $attribute->default_value( $value ) ;

    # Finish attribute.

    $self->_finish_symbol( $attribute ) ;
}


###
# _parse_attribute_default
# Parse attribute value that will be default value for the rest of current block.
sub _parse_attribute_default {
    my $self = shift ;

    # Skip 'attribute'

    $self->consume_token( 'attribute' ) ;

    # Parse attribute name and find the symbol.

    my $attribute = $self->_parse_and_find_symbol() ;
    my $attr_name = $attribute->full_name() ;
    unless ( $attribute->isa( 'SymbolAttribute' ) ) {
        $self->error( {
            MESSAGE => "'$attr_name' is not an attribute type.",
            NO_DIE => 1,
        } ) ;
        $self->error( $self->_error_symbol( $attribute, "Declared here." ) ) ;
    }

    # Parse attribute value.

    $self->consume_token( '=' ) ;
    my $value = $self->_parse_value( $attribute->attribute_type() ) ;
    $self->consume_token( ';' ) ;

    # Remember new attribute default.

    $self->_set_attribute_default( $attr_name, $value ) ;
}


###
# _parse_attribute_list
#   kind - kind of attributes allowed in the list (class name)
# Parse list of attribute assignments.
# Starting token should be '<', ending token '>'. The block should contain
# comma-separated list of attribute assignments 'name = value'. For boolean
# attributes 'name' is allowed shortcut for 'name = true'.
# Returns reference to hash containing ( attribute_name, value ) pairs.
sub _parse_attribute_list {
    my $self = shift ;
    my $kind = shift ;

    my %result ;

    $self->consume_token( '<' ) ;
    my $kind_name = $kind->type_name() ;

    ATTRIBUTE : while ( $self->token_string() ne '>' ) {
        # Read attribute name and find its definition symbol.

        my $attribute = $self->_parse_and_find_symbol() ;
        my $attr_name = $attribute->full_name() ;
        if ( ref( $attribute ) ne $kind ) {
            my $attr_kind = $attribute->type_name() ;
            $self->error( "'$attr_kind $attr_name' is not '$kind_name'." ) ;
        }
        if ( defined( $result{ $attr_name } ) ) {
            $self->error( "Redefininition of attribute '$attr_name' value." ) ;
        }

        # Read value.
        # Special case - boolean attributes can be easily set to true.

        my $value ;
        my $attr_type = $attribute->attribute_type() ;
        if ( ( $self->token_string() ne '=' ) && ( $attr_type eq 'bool' ) ) {
            $value = 1 ;
        }
        else {
            $self->consume_token( '=' ) ;
            $value = $self->_parse_value( $attr_type ) ;
        }

        # Create result entry.

        $result{ $attr_name } = $value ;

        # Next attribute?

        if ( $self->token_string() eq '>' ) {
            last ATTRIBUTE ;
        }
        $self->consume_token( ',' ) ;
    }

    # Done.

    $self->consume_token( '>' ) ;

    return \%result ;
}


###
# _complete_attribute_list
#   symbol - symbol with attributes() field to complete
# Add block-scoped values, inherited values or default values (in this order)
# of attributes with undefined value to attribute list specified symbol.
sub _complete_attribute_list {
    my( $self, $symbol ) = @_ ;

    my $all_attributes = $$self{ ATTRIBUTES } ;
    my $symbol_attributes = $symbol->attributes() ;
    my $kind = ref( $symbol ) . 'Attribute' ;
    
    foreach my $attr_name ( keys( %$all_attributes ) ) {
        next if defined( $$symbol_attributes{ $attr_name } ) ;
        my $attribute = $$all_attributes{ $attr_name } ;
        next if ref( $attribute ) ne $kind ;
        $$symbol_attributes{ $attr_name } = $self->_get_attribute_default( $attr_name, $symbol ) ;
    }
}


1 ;


__END__
