=head1 NAME

IdlParserClass - private class parsing methods of C<IdlParser>.

=cut

package IdlParser ;


###
# _parse_class
sub _parse_class {
    my $self = shift ;

    # Skip 'class'.

    $self->consume_token( 'class' ) ;

    # Attribute list.

    my $attributes ;
    if ( $self->token_string() eq '<' ) {
        $attributes = $self->_parse_attribute_list( 'SymbolClassAttribute' ) ;
    }
    
    # Read class name and create class symbol.

    $self->ensure_identifier() ;
    my $name = $self->token_value() ;
    my $class = $self->_create_symbol( 'SymbolClass', $name ) ;
    $self->next_token() ;

    # Definition or declaration?

    my $token = $self->consume_token( ':', '{', ';' ) ;
    my $token_string = $token->token() ;
    $token = undef ;

    if ( $token_string eq ';' ) {
        # Class declaration. Nothing to do.

        if ( defined( $attributes ) ) {
            $self->error( "Defined attribute list in declaration of 'class $name'." ) ;
        }
        return ;
    }

    # Redefinition?

    if ( $class->defined() ) {
        $self->error( {
            MESSAGE => "Redefinition of class '$name'." ,
            NO_DIE => 1,
        } ) ;
        $self->error( $self->_error_symbol( $class, "Previous definition here." ) ) ;
    }

    # Parents.
    
    if ( $token_string eq ':' ) {
        $self->_parse_class_parents( $class ) ;
    }

    # Remember that class has been defined.

    $class->defined( 1 ) ;
    $$self{ CLASSES }{ $class->full_name() } = $class ;

    # Initialize class.

    $class->set_context_from_parser( $self ) ;
    unless ( defined( $attributes ) ) {
        $attributes = { } ;
    }
    $class->attributes( $attributes ) ;
    $self->_complete_attribute_list( $class ) ;

    # Enter class scope.

    $self->_enter_symbol( $class ) ;
    
    # Class body.

    $self->dispatch( {
        DISPATCH => {
            'attribute' => \&_parse_attribute_default,
            'class' => \&_parse_class,
            'enum' => \&_parse_enum,
            'property' => \&_parse_class_property,
            'method' => \&_parse_class_method,
            ';' => \&_ignore_semicolon,
            '}' => 'end',
        },
        REQUIRE_TOKEN  => 1,
        REPEAT => 1,
    } ) ;

    # Finish class.

    $self->consume_token( '}' ) ;
    $self->_finish_symbol( $class ) ;
    $self->_leave_symbol() ;
}


###
# _parse_class_parents
#   class - class symbol
sub _parse_class_parents {
    my $self = shift ;
    my $class = shift ;

    PARENT : for ( ; ; ) {
        $self->require_token() ;

        # Virtual parent?

        my $virtual = 0 ;
        if ( $self->token_string() eq 'virtual' ) {
            $virtual = 1 ;
            $self->next_token() ;
        }

        # Read parent name.

        my $parent = $self->_parse_and_find_symbol() ;

        # Add it to class parents.

        my $parent_type_name = $parent->type_name() ;
        my $parent_name = $parent->full_name() ;
        if ( $parent_type_name ne 'class' ) {
            $self->error( "Unexpected type of '$parent_name'. " .
                "Expected 'class', found '$parent_type_name'." ) ;
        }
        unless ( $parent->defined() ) {
            $self->error( {
                MESSAGE => "Can't inherit undefined class '$parent_name'.",
                NO_DIE => 1,
            } ) ;
            $self->error( $self->_error_symbol( $parent, "Declared here." ) ) ;
        }
        $class->add_parent( $parent, $virtual ) ;

        # Another parent follows?

        my $token = $self->consume_token( ',', '{' ) ;
        last if $token->token() eq '{' ;
    }
}


###
# _parse_class_property
sub _parse_class_property {
    my $self = shift ;

    $self->consume_token( 'property' ) ;
    my $class = $self->_symbol_table() ;

    # Property attributes.

    my $attributes ;
    if ( $self->token_string() eq '<' ) {
        $attributes = $self->_parse_attribute_list( 'SymbolPropertyAttribute' ) ;
    }
    else {
        $attributes = { } ;
    }

    # Property type.

    my $type = $self->_parse_type() ;

    # Read property name and create property symbol.

    $self->ensure_identifier() ;
    my $name = $self->token_value() ;
    my $property = $self->_create_property_symbol( $name ) ;
    $property->attributes( $attributes ) ;
    $self->_complete_attribute_list( $property ) ;
    $property->property_type( $type ) ;
    $self->next_token() ;

    # Store property into array of properties.

    my $properties = $class->properties() ;
    push( @$properties, $property ) ;

    # Finish.
    
    $self->consume_token( ';' ) ;
    $self->_finish_symbol( $property ) ;
}


###
# _parse_class_method
sub _parse_class_method {
    my $self = shift ;

    $self->consume_token( 'method' ) ;
    my $class = $self->_symbol_table() ;

    # Method attributes.

    my $attributes ;
    if ( $self->token_string() eq '<' ) {
        $attributes = $self->_parse_attribute_list( 'SymbolMethodAttribute' ) ;
    }
    else {
        $attributes = { } ;
    }

    # Read and check method name.

    $self->ensure_identifier() ;
    my $name = $self->token_value() ;
    if ( my( $prefix ) = ( $name =~ /^(a?sync)/ ) ) {
        $self->error( "Reserved method name prefix '$prefix' used." ) ;
    }

    # Create the symbol.

    my $method = $self->_create_method_symbol( $name ) ;
    $method->attributes( $attributes ) ;
    $self->_complete_attribute_list( $method ) ;
    $self->next_token() ;

    # Set method index and store method into array of methods.

    my $methods = $class->methods() ;
    push( @$methods, $method ) ;
    $method->method_index( $#{$methods} ) ;

    # Method arguments.

    $self->consume_token( '(' ) ;

    my %all_arguments ;
    ARGUMENT : while ( $self->token_string() ne ')' ) {
        # Read argument kind, type. Check type.

        my $arg_kind_token = $self->consume_token( 'in', 'out', 'inout' ) ;
        my $arg_type = $self->_parse_type() ;
        unless ( $arg_type->is_argument() ) {
            my $idl_type = $arg_type->idl_type() ;
            $self->error( "Can't use '$idl_type' as method argument." ) ;
        }

        # Read and check argument name.

        $self->ensure_identifier() ;
        my $arg_name = $self->token_value() ;
        if ( $arg_name =~ /^_/ ) {
            $self->error( "Bad argument name '$arg_name', names beginning with '_' are reserved." ) ;
        }
        if ( defined( $all_arguments{ $arg_name } ) ) {
            $self->error( "Multiple arguments named '$arg_name'." ) ;
        }
        $all_arguments{ $arg_name }++ ;
        $self->next_token() ;

        # Add argument.

        $method->add_argument( $arg_kind_token->token(), $arg_type, $arg_name ) ;

        # Next argument?

        if ( $self->token_string() eq ')' ) {
            last ARGUMENT ;
        }
        $self->consume_token( ',' ) ;
    }
    $self->consume_token( ')' ) ;

    # Optional return type.

    my $token = $self->consume_token( ';', ':' ) ;
    if ( $token->token() eq ':' ) {
        my $result_type = $self->_parse_type() ;
        unless ( $result_type->is_argument() ) {
            my $idl_type = $result_type->idl_type() ;
            $self->error( "Can't use '$idl_type' as method result." ) ;
        }
        $method->result_type( $result_type ) ;
        $self->consume_token( ';' ) ;
    }

    $self->_finish_symbol( $method ) ;
}


###
# _create_property_symbol
sub _create_property_symbol {
    my( $self, $name ) = @_ ;
    return $self->_create_class_symbol( 'SymbolProperty', $name ) ;
}


###
# _create_method_symbol
sub _create_method_symbol {
    my( $self, $name ) = @_ ;
    return $self->_create_class_symbol( 'SymbolMethod', $name ) ;
}


###
# _create_class_symbol
sub _create_class_symbol {
    my( $self, $type, $name ) = @_ ;
    my $class = $self->_symbol_table() ;

    # Check for symbol redefinitions.

    my $symbol = $class->find_class_symbol( $name ) ;
    if ( defined( $symbol ) ) {
        my $type_name = $type->type_name() ;
        my $symbol_type_name = $symbol->type_name() ;
        $self->error( {
            MESSAGE => "Can't define '$type_name $name'." ,
            NO_DIE => 1,
        } ) ;
        $self->error( $self->_error_symbol(
            $symbol, "Previous definition of '$symbol_type_name $name' here."
        ) ) ;
    }

    # Create the symbol.

    return $self->_create_symbol( $type, $name ) ;
}


1 ;


__END__
