=head1 NAME

MassivType - classes for Massiv types

=head1 SYNOPSIS

    use MassivType ;

    ###
    # MassivType

    # Constructor.
    $type = MassivType->new() ;

    # Each Type* class should implement the following methods, which
    # return string which can be used to represent the type in various
    # contexts.

    # If some typess can't be used in the context, the method should return
    # undef. However these cases should be documented as bugs and fixed in
    # Massiv::Core.

    # Can type be used as "true" stype? (can it be element of value_array?)
    # Default implementation uses stype_types table.
    $is_stype = $type->is_stype() ;

    # Is this type strong pointer or does it contain strong pointer?
    # Default implementation returns 0.
    $has_strong_pointer = $type->has_strong_pointer() ;
    
    # Can type be used as method argument?
    # Default implementation returns ! $type->has_strong_pointer().
    $is_argument = $type->is_argument() ;

    # Note that even "simple_type" may return property type if it's the most
    # simple type.

    # Get type in IDL syntax.
    $idl_type = $type->idl_type() ;

    # Get the most simple C++ type which can hold values of this type
    # on the stack.
    $simple_type = $type->simple_type() ;

    # Get the most simple C++ type which is serializable and can hold values
    # of this type on the stack.
    $serializable_type = $type->serializable_type() ;

    # Get type used as in parameter.
    # By default it's equal to "simple_type" or "const simple_type &" depending
    # on const_reference_in_types.
    $in_type = $type->in_type() ;

    # Get type used as out parameter.
    # By default it's qual to "simple_type &".
    $out_type = $type->out_type() ;

    # Get property type.
    $property_type = $type->property_type() ;

    # Estimate how many bits will be occupied by in, inout or out argument
    # "name" when serialized to bitstream.
    # This method should return C++ code which estimates the size. Default
    # implementation uses memory_size(), which works for properties and stypes,
    # but it's utterly wrong. Specialized implementations are wrong too.
    $estimate_code = $type->estimate_size( $name ) ;

=cut

package MassivType ;


###
# array_ckind
my %array_ckind = (
    'property' => '::Massiv::Core::PArray',
    'value' => '::Massiv::Core::PValueArray',
) ;


###
# dictionary_ckind
my %dictionary_ckind = (
    'simple' => '::Massiv::Core::PDictionary',
    'multi' => '::Massiv::Core::PMultiDictionary',
) ;


###
# dictionary_ckind
my %set_ckind = (
    'simple' => '::Massiv::Core::PSet',
    'multi' => '::Massiv::Core::PMultiSet',
) ;


###
# dictionary_set_idlkind
my %dictionary_set_idlkind = (
    'simple' => '',
    'multi' => 'multi_',
) ;


###
# float_ctype
my %float_ctype = (
    'float32' => 'Float32',
    'float64' => 'Float64',
) ;


###
# floatq_idltype
my %floatq_idltype = (
    'float32' => 'floatq32',
    'float64' => 'floatq64',
) ;


###
# int_ctype
my %int_ctype = (
    'int8' => '::Massiv::Core::Int8',
    'uint8' => '::Massiv::Core::UInt8',
    'int16' => '::Massiv::Core::Int16',
    'uint16' => '::Massiv::Core::UInt16',
    'int32' => '::Massiv::Core::Int32',
    'uint32' => '::Massiv::Core::UInt32',
    'int64' => '::Massiv::Core::Int64',
    'uint64' => '::Massiv::Core::UInt64',
) ;


###
# pointer_ckind
my %pointer_ckind = (
    'weak' => 'WeakPointer',
    'strong' => 'Pointer',
    'remote' => 'Remote',
) ;


###
# stype_types
my %stype_types = (
    'TypeBool' => 1,
    'TypeColor' => 1,
    'TypeEnum' => 1,
    'TypeFloat' => 1,
    'TypeFloatQ' => 1,
    'TypeInt' => 1,
    'TypeOrientation' => 1,
    'TypeString' => 1,
    'TypeTime' => 1,
    'TypeVector' => 1,
    'TypeVLInt' => 1,
) ;


###
# const_reference_in_types
my %const_reference_in_types = (
    'TypeArray' => 1,
    'TypeColor' => 1,
    'TypeDictionary' => 1,
    'TypeEventHandle' => 1,
    'TypeOrientation' => 1,
    'TypePair' => 1,
    'TypeSet' => 1,
    'TypeString' => 1,
    'TypeTime' => 1,
    'TypeVector' => 1,
) ;


###
# FIXME: key_types


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = { } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# is_stype
# Generic implementation could be class method, but some types may have
# to rely on their data (type parameters) to determine if they are stypes,
# so calling this as Class->is_stype() is forbidden.
sub is_stype {
    my $self = shift ;
    my $class = ref( $self ) ; # || $proto ; - see remark above
    unless ( defined( $class ) ) {
        require Carp ;
        Carp::confess( "is_stype() is no longer class method, called with '$self' proto.\n" ) ;
    }
    return $stype_types{ $class } ? 1 : 0 ;
}


###
# has_strong_pointer
sub has_strong_pointer {
    return 0 ;
}


###
# is_argument
# See is_stype remark, the same applies here.
sub is_argument {
    my $self = shift ;
    my $class = ref( $self ) ;
    unless ( defined( $class ) ) {
        require Carp ;
        Carp::confess( "is_argument() is no longer class method, called with '$self' proto.\n" ) ;
    }
    return $self->has_strong_pointer() ? 0 : 1 ;
}


###
# idl_type
sub idl_type {
    return undef ;
}


###
# simple_type
sub simple_type {
    return undef ;
}


###
# serializable_type
sub serializable_type {
    return undef ;
}


###
# in_type
sub in_type {
    my $self = shift ;
    my $class = ref( $self ) ;
    if ( $const_reference_in_types{ $class } ) {
        return 'const ' . $self->simple_type() . ' &' ;
    }
    else {
        return $self->simple_type() ;
    }
}


###
# out_type
sub out_type {
    my $self = shift ;
    return $self->simple_type() . ' &' ;
}


###
# property_type
sub property_type {
    return undef ;
}


###
# estimate_size
sub estimate_size {
    my $self = shift ;
    my $name = shift ;
    return "8 * ${name}.memory_size()" ;
}


=pod

    ###
    # TypeArray

    # Constructor.
    $array_type = TypeArray->new( [ $kind [, $element_type ] ] ) ;

    # Get/set array kind ('property' or 'value')
    $kind = $array_type->kind() ;
    $array_type->kind( $kind ) ;

    # Get/set reference to array element type.
    $element_type = $array_type->element_type() ;
    $array_type->element_type( $element_type ) ;

=cut

{
package TypeArray ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $kind = shift ;
    my $element_type = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ KIND } = undef ;
    $self->kind( $kind ) ;
    $$self{ ELEMENT_TYPE } = undef ;
    $self->element_type( $element_type ) ;
    return $self ;
}

###
# kind
sub kind {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $array_ckind{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid kind name '$name'.\n" )
        }
        $$self{ KIND } = $name ;
    }
    return $$self{ KIND } ;
}

###
# element_type
sub element_type {
    my $self = shift ;

    if ( @_ ) {
        my $type = shift ;
        if ( defined( $$self{ KIND } ) ) {
            if ( $$self{ KIND } eq 'value' ) {
                unless ( $type->is_stype() ) {
                    my $idl_type = $type->idl_type() ;
                    require Carp ;
                    Carp::confess( "Can't create value_array of '$idl_type'.\n" ) ;
                }
            }
        }
        else {
            require Carp ;
            Carp::cluck( "Setting element type of array with undefined kind!\n" ) ;
        }
        $$self{ ELEMENT_TYPE } = $type ;
    }

    return $$self{ ELEMENT_TYPE } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    return $self->element_type()->has_strong_pointer() ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $$self{ KIND } . '_array< ' . $$self{ ELEMENT_TYPE }->idl_type() . ' >' ;
}

sub simple_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub property_type {
    my $self = shift ;
    my $kind = $$self{ KIND } ;
    my $kind_name = $array_ckind{ $kind } ;
    my $element_type = $$self{ ELEMENT_TYPE } ;
    my $element_name ;
    if ( $kind eq 'property' ) { $element_name = $element_type->property_type() ; }
    else { $element_name = $element_type->serializable_type() ; }
    return $kind_name . '< ' . $element_name . ' >' ;
}


} # package TypeArray


=pod

    ###
    # TypeBool

=cut

{
package TypeBool ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# *_type
sub idl_type {
    return 'bool' ;
}

sub simple_type {
    return 'bool' ;
}

sub serializable_type {
    return '::Massiv::Core::SBool' ;
}

sub property_type {
    return '::Massiv::Core::PBool' ;
}

sub estimate_size {
    return "1" ;
}


} # package TypeBool


=pod

    ###
    # TypeClass

    # Constructor.
    $class_type = TypeClass->new( [ $class_symbol ] ) ;

    # Get/set class symbol.
    $symbol = $class_type->class_symbol() ;
    $class_type->class_symbol( $symbol ) ;

    # Class can be used as property or argument to property_array
    # or dictionary (probably only as value), and value_type clasess can
    # be used as method arguments.

=cut

{
package TypeClass ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $class_symbol = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ CLASS_SYMBOL } = $class_symbol ;
    return $self ;
}

###
# class_symbol
sub class_symbol {
    my $self = shift ;
    if ( @_ ) { $$self{ CLASS_SYMBOL } = shift ; }
    return $$self{ CLASS_SYMBOL } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    my $symbol = $self->class_symbol() ;
    return defined( $symbol->find_strong_pointer() ) ;
}


###
# is_argument
sub is_argument {
    my $self = shift ;

    # Class can be used as argument if it's value type without strong pointers.

    my $symbol = $self->class_symbol() ;
    if( $symbol->require_attribute( '::value_type' ) ) {
        return $self->has_strong_pointer() ? 0 : 1 ;
    }

    return 0 ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub simple_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub in_type {
    my $self = shift ;
    if( $self->is_argument() ) {
        return 'const ' . $self->property_type() . ' &' ;
    }
    else {
        require Carp ;
        Carp::confess( "Tried to use non-value-type class as 'in' argument.\n" ) ;
    }
}

sub out_type {
    my $self = shift ;
    if( $self->is_argument() ) {
        return $self->property_type() . ' &' ;
    }
    else {
        require Carp ;
        Carp::confess( "Tried to use non-value-type class as 'out' argument.\n" ) ;
    }
}

sub property_type {
    my $self = shift ;
    my $class_symbol = $$self{ CLASS_SYMBOL } ;
    return $class_symbol->full_name() ;
}


} # package TypeClass


=pod

    ###
    # TypeColor

=cut

{
package TypeColor ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# *_type
sub idl_type {
    return 'color' ;
}

sub simple_type {
    return '::Massiv::Core::SColor' ;
}

sub serializable_type {
    return '::Massiv::Core::SColor' ;
}

sub property_type {
    return '::Massiv::Core::PColor' ;
}


} # package TypeBool


=pod

    ###
    # TypeDictionary

    # Constructor.
    $dictionary_type = TypeDictionary->new( [ $kind [, $key_type [, $value_type ] ] ] ) ;

    # Get/set dictionary kind ('simple' or 'multi')
    $kind = $dictionary_type->kind() ;
    $dictionary_type->kind( $kind ) ;

    # Get/set reference to key type.
    $key_type = $dictionary_type->key_type() ;
    $dictionary_type->key_type( $key_type ) ;

    # Get/set reference to value type.
    $value_type = $dictionary_type->value_type() ;
    $dictionary_type->value_type( $value_type ) ;

=cut

{
package TypeDictionary ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $kind = shift ;
    my $key_type = shift ;
    my $value_type = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ KIND } = undef ;
    $self->kind( $kind ) ;
    $$self{ KEY_TYPE } = $key_type ;
    $$self{ VALUE_TYPE } = $value_type ;
    return $self ;
}

###
# kind
sub kind {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $dictionary_ckind{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid kind name '$name'.\n" )
        }
        $$self{ KIND } = $name ;
    }
    return $$self{ KIND } ;
}

###
# key_type
sub key_type {
    my $self = shift ;
    if ( @_ ) { $$self{ KEY_TYPE } = shift ; }
    return $$self{ KEY_TYPE } ;
}

###
# value_type
sub value_type {
    my $self = shift ;
    if ( @_ ) { $$self{ VALUE_TYPE } = shift ; }
    return $$self{ VALUE_TYPE } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    return $self->key_type()->has_strong_pointer() ||
           $self->value_type()->has_strong_pointer() ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $dictionary_set_idlkind{ $$self{ KIND } } . 'dictionary< ' .
        $$self{ KEY_TYPE }->idl_type() . ', ' . $$self{ VALUE_TYPE }->idl_type() . ' >' ;
}

sub simple_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub property_type {
    my $self = shift ;
    my $kind = $$self{ KIND } ;
    my $key_type = $$self{ KEY_TYPE } ;
    my $value_type = $$self{ VALUE_TYPE } ;
    return $dictionary_ckind{ $kind } . '< ' .
        $key_type->property_type() . ', ' . $value_type->property_type() . ' >' ;
}


} # package TypeDictionary


=pod

    ###
    # TypeEnum

    # Constructor.
    $enum_type = TypeEnum->new( [ $enum_symbol ] ) ;

    # Get/set enun symbol.
    $symbol = $enum_type->enum_symbol() ;
    $enum_type->enum_symbol( $symbol ) ;

=cut

{
package TypeEnum ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $enum_symbol = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ ENUM_SYMBOL } = $enum_symbol ;
    return $self ;
}

###
# enum_symbol
sub enum_symbol {
    my $self = shift ;
    if ( @_ ) { $$self{ ENUM_SYMBOL } = shift ; }
    return $$self{ ENUM_SYMBOL } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $$self{ ENUM_SYMBOL }->full_name() ;
}

sub simple_type {
    my $self = shift ;
    return $$self{ ENUM_SYMBOL }->full_name() ;
}

sub serializable_type {
    my $self = shift ;
    return '::Massiv::Core::SEnum< ' . $self->simple_type() . ' >' ;
}

sub property_type {
    my $self = shift ;
    return '::Massiv::Core::PEnum< ' . $self->simple_type() . ' >' ;
}

sub estimate_size {
    return "8" ;
}


} # package TypeEnum


=pod

    ###
    # TypeEventHandle

=cut

{
package TypeEventHandle ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# *_type
sub idl_type {
    return 'event_handle' ;
}

sub simple_type {
    return '::Massiv::Core::EventHandle' ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub property_type {
    return '::Massiv::Core::PEventHandle' ;
}

sub estimate_size {
    return '8 * sizeof( ::Massiv::Core::EventHandle )' ;
}


} # package TypeEventHandle


=pod

    ###
    # TypeFloat

    # Constructor.
    $float_type = TypeFloat->new( [ $type_name ] ) ;

    # Get/set float type (float32 or float64)
    $type_name = $float_type->float_type() ;
    $float_type->float_type( $type_name ) ;

=cut

{
package TypeFloat ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $type_name = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ FLOAT_TYPE } = undef ;
    $self->float_type( $type_name ) ;
    return $self ;
}

###
# float_type
sub float_type {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $float_ctype{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid float type name '$name'.\n" )
        }
        $$self{ FLOAT_TYPE } = $name ;
    }
    return $$self{ FLOAT_TYPE } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $$self{ FLOAT_TYPE } ;
}

sub simple_type {
    my $self = shift ;
    return '::Massiv::Core::' . $float_ctype{ $$self{ FLOAT_TYPE } } ;
}

sub serializable_type {
    my $self = shift ;
    return '::Massiv::Core::S' . $float_ctype{ $$self{ FLOAT_TYPE } } ;
}

sub property_type {
    my $self = shift ;
    return '::Massiv::Core::P' . $float_ctype{ $$self{ FLOAT_TYPE } } ;
}

sub estimate_size {
    my $self = shift ;
    return $$self{ FLOAT_TYPE } eq 'float32' ? '32' : '64' ;
}


} # package TypeFloat


=pod

    ###
    # TypeFloatQ

    # FloatQ* types require traits structure as their template parameter. This
    # structure must be defined somewhere in sources "visible" from headers
    # containing declaration of classes containing FloatQ types. That's why
    # traits can't be generated by factgen.

    # Constructor.
    $floatq_type = TypeFloatQ->new( [ $type_name, [ $traits_name ] ] ) ;

    # Get/set float type (float32 or float64).
    $type_name = $floatq_type->float_type() ;
    $floatq_type->float_type( $type_name ) ;

    # Get/set traits class name (as string).
    $traits_name = $floatq_type->traits_name() ;
    $floatq_type->traits_name( $traits_name ) ;

=cut

{
package TypeFloatQ ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $type_name = shift ;
    my $traits_name = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ FLOAT_TYPE } = undef ;
    $self->float_type( $type_name ) ;
    $$self{ TRAITS_NAME } = $traits_name ;
    return $self ;
}

###
# float_type
sub float_type {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $floatq_idltype{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid float type name '$name'.\n" )
        }
        $$self{ FLOAT_TYPE } = $name ;
    }
    return $$self{ FLOAT_TYPE } ;
}

###
# traits_name
sub traits_name {
    my $self = shift ;
    if ( @_ ) { $$self{ TRAITS_NAME } = shift ; }
    return $$self{ TRAITS_NAME } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $floatq_idltype{ $$self{ FLOAT_TYPE } } . '< "' . $$self{ TRAITS_NAME } . '" >' ;
}

sub simple_type {
    my $self = shift ;
    return '::Massiv::Core::' . $float_ctype{ $$self{ FLOAT_TYPE } } ;
}

sub serializable_type {
    my $self = shift ;
    return '::Massiv::Core::SFloatQ< ' . $float_ctype{ $$self{ FLOAT_TYPE } } .
                ', ' . $$self{ TRAITS_NAME } . ' >' ;
}

sub property_type {
    my $self = shift ;
    return '::Massiv::Core::PFloatQ< ' . $float_ctype{ $$self{ FLOAT_TYPE } } .
                ', ' . $$self{ TRAITS_NAME } . ' >' ;
}

sub estimate_size {
    # Rough estimate. Heh.
    return '16' ;
}


} # package TypeFloatQ


=pod

    ###
    # TypeInt

    # Constructor.
    $int_type = TypeInt->new( [ $type_name [, $bits ] ] ) ;

    # Get/set integral type (using IDL type names)
    $type_name = $int_type->int_type() ;
    $int_type->int_type( $type_name ) ;

    # Get/set number of bits to use during serialization.
    $bits = $int_type->bits() ;
    $int_type->bits( $bits ) ;

=cut

{
package TypeInt ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $type_name = shift ;
    my $bits = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ INT_TYPE } = undef ;
    $self->int_type( $type_name ) ;
    $$self{ BITS } = $bits ;
    return $self ;
}

###
# int_type
sub int_type {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $int_ctype{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid int type name '$name'.\n" )
        }
        $$self{ INT_TYPE } = $name ;
    }
    return $$self{ INT_TYPE } ;
}

###
# bits
sub bits {
    my $self = shift ;
    if ( @_ ) { $$self{ BITS } = shift ; }
    return $$self{ BITS } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return "$$self{ INT_TYPE }< $$self{ BITS } >" ;
}

sub simple_type {
    my $self = shift ;
    return $int_ctype{ $$self{ INT_TYPE } } ;
}

sub serializable_type {
    my $self = shift ;
    my $simple = $self->simple_type() ;
    return "::Massiv::Core::SInt< $simple, $$self{ BITS } >" ;
}

sub property_type {
    my $self = shift ;
    my $simple = $self->simple_type() ;
    return "::Massiv::Core::PInt< $simple, $$self{ BITS } >" ;
}

sub estimate_size {
    my $self = shift ;
    # Wow, correct.
    return $$self{ BITS } ;
}


} # package TypeInt


=pod

    ###
    # TypeOrientation

=cut

{
package TypeOrientation ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# *_type
sub idl_type {
    return 'orientation' ;
}

sub simple_type {
    return '::Massiv::Core::SOrientation' ;
}

sub serializable_type {
    return '::Massiv::Core::SOrientation' ;
}

sub property_type {
    return '::Massiv::Core::POrientation' ;
}


} # package TypeOrientation


=pod

    ###
    # TypePair

    # Constructor.
    $pair_type = TypePari->new( [ $first_type [, $second_type ] ] ) ;

    # Get/set reference to first type.
    $first_type = $pair_type->first_type() ;
    $pair_type->first_type( $first_type ) ;

    # Get/set reference to second type.
    $second_type = $pair_type->second_type() ;
    $pair_type->second_type( $second_type ) ;

=cut

{
package TypePair ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $first_type = shift ;
    my $second_type = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ FIRST_TYPE } = $first_type ;
    $$self{ SECOND_TYPE } = $second_type ;
    return $self ;
}

###
# first_type
sub first_type {
    my $self = shift ;
    if ( @_ ) { $$self{ FIRST_TYPE } = shift ; }
    return $$self{ FIRST_TYPE } ;
}

###
# second_type
sub second_type {
    my $self = shift ;
    if ( @_ ) { $$self{ SECOND_TYPE } = shift ; }
    return $$self{ SECOND_TYPE } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    return $self->first_type()->has_strong_pointer() ||
           $self->second_type()->has_strong_pointer() ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return 'pair< ' . $$self{ FIRST_TYPE }->idl_type() . ', ' .
        $$self{ SECOND_TYPE }->idl_type() . ' >' ;
}

sub simple_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub property_type {
    my $self = shift ;
    my $first_type = $$self{ FIRST_TYPE } ;
    my $second_type = $$self{ SECOND_TYPE } ;
    return '::Massiv::Core::PPair< ' . $$self{ FIRST_TYPE }->property_type() . ', ' .
        $$self{ SECOND_TYPE }->property_type() . ' >' ;
}


} # package TypePair


=pod

    ###
    # TypePointer

    # Constructor.
    $pointer_type = TypePointer->new( [ $kind [, $class_symbol ] ] ) ;

    # Get/set pointer kind ('weak', 'strong' or 'remote')
    $kind = $pointer_type->kind() ;
    $pointer_type->kind( $kind ) ;

    # Get/set reference to class symbol - type of this pointer.
    $class_symbol = $pointer_type->class_symbol() ;
    $pointer_type->class_symbol( $class_symbol ) ;

=cut

{
package TypePointer ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $kind = shift ;
    my $class_symbol = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ KIND } = undef ;
    $self->kind( $kind ) ;
    $$self{ CLASS_SYMBOL } = $class_symbol ;
    return $self ;
}

###
# kind
sub kind {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $pointer_ckind{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid kind name '$name'.\n" )
        }
        $$self{ KIND } = $name ;
    }
    return $$self{ KIND } ;
}

###
# class_symbol
sub class_Symbol {
    my $self = shift ;
    if ( @_ ) { $$self{ CLASS_SYMBOL } = shift ; }
    return $$self{ CLASS_SYMBOL } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    return ( $self->kind() eq 'strong' ? 1 : 0 ) ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    my $class_name = $$self{ CLASS_SYMBOL }->full_name() ;
    return "$$self{ KIND }_pointer< $class_name >" ;
}

sub simple_type {
    my $self = shift ;
    return $self->serializable_type() ;
}

sub serializable_type {
    my $self = shift ;
    my $kind_name = $pointer_ckind{ $$self{ KIND } } ;
    my $class_name = $$self{ CLASS_SYMBOL }->full_name() ;
    return "::Massiv::Core::$kind_name< $class_name >" ;
}

sub property_type {
    my $self = shift ;
    my $kind_name = $pointer_ckind{ $$self{ KIND } } ;
    my $class_name = $$self{ CLASS_SYMBOL }->full_name() ;
    return "::Massiv::Core::P$kind_name< $class_name >" ;
}

sub estimate_size {
    return '8 * sizeof( ::Massiv::Core::ObjectId )' ;
}


} # package TypePointer


=pod

    ###
    # TypeSet

    # Constructor.
    $set_type = TypeSet->new( [ $kind [, $value_type ] ] ) ;

    # Get/set set kind ('simple' or 'multi')
    $kind = $set_type->kind() ;
    $set_type->kind( $kind ) ;

    # Get/set reference to value type.
    $value_type = $set_type->value_type() ;
    $set_type->value_type( $value_type ) ;

=cut

{
package TypeSet ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $kind = shift ;
    my $value_type = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ KIND } = undef ;
    $self->kind( $kind ) ;
    $$self{ VALUE_TYPE } = $value_type ;
    return $self ;
}

###
# kind
sub kind {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $set_ckind{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid kind name '$name'.\n" )
        }
        $$self{ KIND } = $name ;
    }
    return $$self{ KIND } ;
}

###
# value_type
sub value_type {
    my $self = shift ;
    if ( @_ ) { $$self{ VALUE_TYPE } = shift ; }
    return $$self{ VALUE_TYPE } ;
}

###
# has_strong_pointer
sub has_strong_pointer {
    my $self = shift ;
    return $self->value_type()->has_strong_pointer() ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return $dictionary_set_idlkind{ $$self{ KIND } } . 'set< ' .
        $$self{ VALUE_TYPE }->idl_type() . ' >' ;
}

sub simple_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub serializable_type {
    my $self = shift ;
    return $self->property_type() ;
}

sub property_type {
    my $self = shift ;
    my $kind = $$self{ KIND } ;
    my $value_type = $$self{ VALUE_TYPE } ;
    return $set_ckind{ $kind } . '< ' . $value_type->property_type() . ' >' ;
}


} # package TypeSet


=pod

    ###
    # TypeString

=cut

{
package TypeString ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# *_type
sub idl_type {
    return 'string' ;
}

sub simple_type {
    return 'std::string' ;
}

sub serializable_type {
    return '::Massiv::Core::SString' ;
}

sub property_type {
    return '::Massiv::Core::PString' ;
}

sub estimate_size {
    my $self = shift ;
    my $name = shift ;
    return "7 * ( ${name}.size() + 1 )" ;
}


} # package TypeString


=pod

    ###
    # TypeTime

    # FIXME: There some problems with time type conversions, so no C++ native
    #        types are used here.

=cut

{
package TypeTime ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# type_name
sub type_name {
    'type_time' ;
}

###
# *_type
sub idl_type {
    return 'time' ;
}

sub simple_type {
    return '::Massiv::Core::STime' ;
}

sub serializable_type {
    return '::Massiv::Core::STime' ;
}

sub property_type {
    return '::Massiv::Core::PTime' ;
}


} # package TypeTime


=pod

    ###
    # TypeVector

    # Constructor.
    $vector_type = TypeVector->new( [ $length ] ) ;

    # Get/set vector length. Must be 2 or 3.
    $length = $vector_type->length() ;
    $vector_type->length( $length ) ;

=cut

{
package TypeVector ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $length = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ LENGTH } = undef ;
    $self->length( $length ) ;
    return $self ;
}

###
# length
sub length {
    my $self = shift ;
    if ( @_ ) {
        my $length = shift ;
        if ( ( $length == 2 ) || ( $length == 3 ) ) {
            $$self{ LENGTH } = $length ;
        }
        else {
            require Carp ;
            Carp::confess( "Invalid vector length '$length'.\n" )
        }
    }
    return $$self{ LENGTH } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return 'vector' . $$self{ LENGTH } ;
}

sub simple_type {
    my $self = shift ;
    return $self->serializable_type() ;
}

sub serializable_type {
    my $self = shift ;
    return '::Massiv::Core::SVector' . $$self{ LENGTH } ;
}

sub property_type {
    my $self = shift ;
    return '::Massiv::Core::PVector' . $$self{ LENGTH } ;
}


} # package TypeVector


=pod

    ###
    # TypeVLInt

    # Constructor.
    $vlint_type = TypeVLInt->new( [ $type_name ] ) ;

    # Get/set integral type (using IDL int type names - without "vl" prefix)
    $type_name = $vlint_type->int_type() ;
    $vlint_type->int_type( $type_name ) ;

    # FIXME: vlint as alias for vlint32?

=cut

{
package TypeVLInt ;

our( @ISA ) ;
@ISA = qw( MassivType ) ;

###
# new
sub new {
    my $proto = shift ;
    my $type_name = shift ;

    my $class = ref( $proto ) || $proto ;
    my $self = $class->SUPER::new( @_ ) ;

    $$self{ INT_TYPE } = undef ;
    $self->int_type( $type_name ) ;
    return $self ;
}

###
# int_type
sub int_type {
    my $self = shift ;
    if ( @_ ) {
        my $name = shift ;
        unless ( defined( $int_ctype{ $name } ) ) {
            require Carp ;
            Carp::confess( "Invalid int type name '$name'.\n" )
        }
        $$self{ INT_TYPE } = $name ;
    }
    return $$self{ INT_TYPE } ;
}

###
# *_type
sub idl_type {
    my $self = shift ;
    return 'vl' . $$self{ INT_TYPE } ;
}

sub simple_type {
    my $self = shift ;
    return $int_ctype{ $$self{ INT_TYPE } } ;
}

sub serializable_type {
    my $self = shift ;
    my $simple = $self->simple_type() ;
    return "::Massiv::Core::SVli< $simple >" ;
}

sub property_type {
    my $self = shift ;
    my $simple = $self->simple_type() ;
    return "::Massiv::Core::PVli< $simple >" ;
}

sub estimate_size {
    return '16' ;
}


} # package TypeVLInt


1 ;


__END__


=pod

=head1 DESCRIPTION

Package F<MassivType.pm> defines classes which represent various Massiv types.
Each type object can return string which represents the type in specified
context - types in IDL syntax, simple C++ types, serializable types, method
arguments and property types.

=head1 KNOWN BUGS

List of types which return C<undef> in some cases should be here.

More checks should be done - only some types should be allowed as hash keys, etc.

=cut
