=head1 NAME

MetaobjectGen.pm - metaobject generator

=head1 SYNOPSIS

    use MetaobjectGen ;

    MetaobjectGen::write_class( $handle, $class_info, $idl_parser ) ;

    MetaobjectGen::write_source( $handle, $class_info, $idl_parser ) ;

=head1 DESCRIPTION

Use this module to write interface and implementation of metaobject for class
specified by I<$class_info> using information gathered by parser I<$idl_parser>.

See C<ClassInfo.pm> and C<IdlParser.pm> for information about the classes.

=head1 KNOWN BUGS

None.

=cut

package MetaobjectGen ;

use strict ;
use warnings ;

use ClassInfo ;
use IdlParser ;
use RPCGen ;


###
# write_class
sub write_class {
    my( $handle, $class_info, $idl_parser ) = @_ ;

    # Extract info.

    my $metaobject_template_name = $class_info->metaobject_template_name() ;
    my $factory_name = $class_info->factory_name() ;
    my $metaobject_name = $class_info->metaobject_name() ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;

    my $comparable = $class_symbol->require_attribute( '::comparable' ) ;
    my $assignable = $class_symbol->require_attribute( '::assignable' ) ;
    my $symbol_attribute = $class_symbol->require_attribute( '::symbol_attribute' ) ;
    $symbol_attribute .= ' ' if $symbol_attribute ne '' ;

    $handle->print( <<EOT

/**
 * Metaobject for $class_name.
 */
template<>
class ${symbol_attribute}$metaobject_name : public ::Massiv::Core::MetaObject
    {
    /** This class. */
    typedef $metaobject_template_name Self;

    /** Superclass. */
    typedef ::Massiv::Core::MetaObject Super;

    /** Object factory. */
    typedef $factory_name Factory;

public:

    /** Pointer to the only instance of $metaobject_name. */
    static const Self * self;

    /** Class type id for $class_name. */
    static ::Massiv::Core::ClassTypeId class_id;

public:

    /**
     * Constructor.
     */
    $metaobject_template_name
        (
        int     kind,
        int     type_id
        );

    /**
     * Destructor.
     */
    virtual ~$metaobject_template_name();

    /**
     * Create metaobject and factory objects and register them to the class manager.
     */
    static void create
        (
        int     kind,
        int     type_id
        );

    /**
     * Initialize introspection structures.
     */
    virtual void initialize_introspection();

public:

    /**
     * Convert ObjectProperty * to $class_name *.
     */
    virtual ::Massiv::Core::VariantPointer downcast
        (
        ::Massiv::Core::ObjectProperty * object
        ) const;

    /**
     * Convert $class_name * to pointer to given component.
     */
    virtual UpcastResult upcast
        (
        const ::Massiv::Core::VariantPointer &  object,
        std::size_t                             component_index
        ) const;

    /**
     * Create new method arguments and results packet for method \@a method_index.
     */
    virtual std::auto_ptr< ::Massiv::Core::MethodPacket > create_method_packet
        (
        std::size_t method_index
        ) const;
    
private:

    /**
     * Convert ObjectProperty * to $class_name * using Object::complete_object.
     */
    $class_name * get_complete_object_pointer
        (
        ::Massiv::Core::ObjectProperty * object
        ) const;

    /**
     * Convert const ObjectProperty * to const $class_name * using Object::complete_object.
     */
    const $class_name * get_complete_object_pointer
        (
        const ::Massiv::Core::ObjectProperty * object
        ) const;
EOT
    ) ;

    if ( $comparable || $assignable ) {
        $handle->print( <<EOT

protected:
EOT
        ) ;
    }

    if ( $comparable ) {
        $handle->print( <<EOT

    /**
     * Memberwise compare two objects.
     */
    virtual bool compare_objects
        (
        const ::Massiv::Core::ObjectProperty *  object1,
        const ::Massiv::Core::ObjectProperty *  object2
        ) const;
EOT
        ) ;
    }

    if ( $assignable ) {
        $handle->print( <<EOT

    /**
     * Memberwise copy.
     */
    virtual void copy_objects
        (
        ::Massiv::Core::ObjectProperty *        destination,
        const ::Massiv::Core::ObjectProperty *  source
        ) const;
EOT
        ) ;
    }

    $handle->print( <<EOT

    }; // class $metaobject_name
EOT
    ) ;

}


###
# write_source
sub write_source {
    my( $handle, $class_info, $idl_parser ) = @_ ;

    # Extract info.

    my $metaobject_template_name = $class_info->metaobject_template_name() ;
    my $factory_name = $class_info->factory_name() ;
    my $metaobject_name = $class_info->metaobject_name() ;
    my $all_virtual_parents = $class_info->virtual_parents() ;
    my $all_properties = $class_info->properties() ;
    my $subcomponent_count = $class_info->component_count() ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $short_class_name = substr( $class_name, 2 ) ; # without initial '::'
    my $class_parents = $class_symbol->parents() ;
    my $class_attributes = $class_symbol->attributes() ;
    my $class_properties = $class_symbol->properties() ;
    my $class_methods = $class_symbol->methods() ;

    my $comparable = $class_symbol->require_attribute( '::comparable' ) ;
    my $assignable = $class_symbol->require_attribute( '::assignable' ) ;

    my $all_attributes = $idl_parser->attributes() ;

    my $referenced ;

    $handle->print( <<EOT

const ${metaobject_name} * ${metaobject_name}::self = NULL;

::Massiv::Core::ClassTypeId ${metaobject_name}::class_id;

${metaobject_name}::${metaobject_template_name}
    (
    int     kind,
    int     type_id
    )
    {
    hope( !self );
    self = this;
    class_name = "$short_class_name";
    class_type_info = &typeid( $class_name );
    class_pointer_type_info = &typeid( $class_name * );
    class_type_id = class_id = ::Massiv::Core::ClassTypeId( kind, type_id );
    object_factory.reset( new Factory( *this ) );
EOT
    ) ;

    # Initialize attributes.

    foreach my $attribute_name ( sort( keys( %$class_attributes ) ) ) {
        my $attribute = $$all_attributes{ $attribute_name } ;
        my $attribute_flags = $attribute->attribute_flags() ;
        next if $$attribute_flags{ 'idl_internal' } ;

        my $cpp_name = $attribute->cpp_name() ;
        my $value = $$class_attributes{ $attribute_name } ;
        my $cpp_value = $attribute->cpp_value( $value ) ;
        $handle->print( "    attributes.$cpp_name = $cpp_value;\n" ) ;
    }

    $handle->print( <<EOT
    }

${metaobject_name}::~${metaobject_template_name}()
    {
    // object_factory deleted by auto_ptr
    }

void ${metaobject_name}::create
    (
    int     kind,
    int     type_id
    )
    {
    std::auto_ptr< ::Massiv::Core::MetaObject > metaobject( new ${metaobject_name}( kind, type_id ) );
    ::Massiv::Core::ClassManager & manager = ::Massiv::Core::Global::class_manager();
    manager.register_class( *metaobject );
    metaobject.release();
    }

void ${metaobject_name}::initialize_introspection()
    {
    // Component list.

EOT
    ) ;

    # Components.

    for ( my $component_index = 0 ; $component_index <= $subcomponent_count ; $component_index++ ) {
        $handle->print( "    component_infos.push_back( ComponentInfo( " ) ;
        $class_info->write_componentinfo_arguments( $handle, $component_index ) ;
        $handle->print( " ) );\n" ) ;
    }

    # Properties.

    if ( scalar( @$class_properties ) ) {
        $handle->print( <<EOT

    // Properties.

    const $class_name * const p_0 = reinterpret_cast< $class_name * >( 0x10000 );
    ::Massiv::Core::MetaObject::ComponentInfoIterator component_it = component_infos.begin();
    ::Massiv::Core::Property::Attributes attributes;
EOT
        ) ;
    }
    foreach my $property ( @$class_properties ) {
        my $property_name = $property->name() ;
        my $property_type = $property->property_type()->idl_type() ;
        my $escaped_type = Utils::escape_string( $property_type ) ;
        my $attributes = $property->attributes() ;

        # Attributes.

        $handle->print( "    attributes.reset();\n" ) ;

        foreach my $attribute ( keys( %$attributes ) ) {
            my $attribute_symbol = $$all_attributes{ $attribute } ;
            my $attribute_flags = $attribute_symbol->attribute_flags() ;
            next if $$attribute_flags{ 'idl_internal' } ;

            my $cpp_name = $attribute_symbol->cpp_name() ;
            my $value = $$attributes{ $attribute } ;
            my $cpp_value = $attribute_symbol->cpp_value( $value ) ;

            $handle->print( "    attributes.$cpp_name = $cpp_value;\n" ) ;
        }

        # Property info.

        $handle->print( "    property_infos.push_back( PropertyInfo( \"$property_name\", " .
            "\"$escaped_type\", component_it, attributes, PROPERTY_OFFSET( $property_name ) ) );\n" ) ;
    }

    # Methods.

    if ( scalar( @$class_methods ) ) {
        $handle->print( "\n    // Methods.\n" ) ;
    }
    foreach my $method ( @$class_methods ) {
        my $method_name = $method->name() ;
        my $method_index = $method->method_index() ;
        my $arguments = $method->arguments() ;
        my $result_type = $method->result_type() ;
        if ( defined( $result_type ) ) {
            $result_type = $result_type->idl_type() ;
        }
        else {
            $result_type = '' ;
        }
        my $escaped_result_type = Utils::escape_string( $result_type ) ;
        my $attributes = $method->attributes() ;

        $handle->print( "\n    method_infos.push_back( MethodInfo( \"$method_name\", " .
                    "\"$escaped_result_type\" ) );\n" ) ;

        foreach my $argument ( @$arguments ) {
            my $argument_kind = uc( $$argument{ KIND } ) ;
            my $argument_type = $$argument{ TYPE }->idl_type() ;
            my $escaped_argument_type = Utils::escape_string( $argument_type ) ;
            my $argument_name = $$argument{ NAME } ;
            $handle->print( "    method_infos[ $method_index ].arguments.push_back" .
                "( ArgumentInfo( \"$argument_name\", \"$escaped_argument_type\", " .
                "ArgumentInfo::ARGUMENT_$argument_kind ) );\n" ) ;
        }

        foreach my $attribute_name ( keys( %$attributes ) ) {
            my $attribute_value = $$attributes{ $attribute_name } ;
            my $attribute = $$all_attributes{ $attribute_name } ;
            my $attribute_name = $attribute->cpp_name() ;
            $handle->print( "    method_infos[ $method_index ].attributes.a_$attribute_name = $attribute_value;\n" ) ;
        }
    }

    $handle->print( <<EOT
    }

::Massiv::Core::VariantPointer ${metaobject_name}::downcast
    (
    ::Massiv::Core::ObjectProperty * object
    ) const
    {
    hope( object );
    hope( typeid( *object ) == typeid( $class_name ) );
EOT
    ) ;

    if ( scalar( @$all_virtual_parents ) ) {
        $handle->print( "    return ::Massiv::Core::VariantPointer( dynamic_cast< $class_name * >( object ) );\n" ) ;
    }
    else {
        $handle->print( "    return ::Massiv::Core::VariantPointer( checked_cast< $class_name * >( object ) );\n" ) ;
    }

    $handle->print( <<EOT
    }

::Massiv::Core::MetaObject::UpcastResult ${metaobject_name}::upcast
    (
    const ::Massiv::Core::VariantPointer &  object,
    std::size_t                             component_index
    ) const
    {
    $class_name * const p_0 = ::Massiv::Core::variant_cast< $class_name * >( object );
    hope( p_0 );
    switch( component_index )
        {
EOT
    ) ;

    for ( my $component_index = 0 ; $component_index <= $subcomponent_count ; $component_index++ ) {
        $handle->print( <<EOT
        case ${component_index}:
            {
EOT
        ) ;
        $class_info->write_component_casts( $handle, $component_index, ' ' x 12, 0 ) ;
        $handle->print( <<EOT
            return UpcastResult( ::Massiv::Core::VariantPointer( p_$component_index ), component_index );
            }
EOT
        ) ;
    }

    $handle->print( <<EOT
        }
    panic_upcast_index();
    /*NOTREACHED*/
    }

std::auto_ptr< ::Massiv::Core::MethodPacket > ${metaobject_name}::create_method_packet
    (
    std::size_t method_index
    ) const
    {
EOT
    ) ;

    if ( @$class_methods ) {
        $handle->print( <<EOT
    switch( method_index )
        {
EOT
        ) ;

        # For each method.
        # FIXME: Use indices from MethodTags?

        for ( my $i = 0 ; $i < @$class_methods ; $i++ ) {
            my $arguments_name = RPCGen::method_packet_name( $class_info, $i ) ;
            $handle->print( ( ' ' x 8 ) . "case ${i}: return std::auto_ptr< ::Massiv::Core::MethodPacket >( new $arguments_name );\n" ) ;
        }

        $handle->print( ( ' ' x 8 ) . "}\n" ) ;

    }

    $handle->print( <<EOT
    return std::auto_ptr< ::Massiv::Core::MethodPacket >();
    }

$class_name * ${metaobject_name}::get_complete_object_pointer
    (
    ::Massiv::Core::ObjectProperty * object
    ) const
    {
    hope( object );
    return ::Massiv::Core::variant_cast< $class_name * >( get_complete_object( object ) );
    }

const $class_name * ${metaobject_name}::get_complete_object_pointer
    (
    const ::Massiv::Core::ObjectProperty * object
    ) const
    {
    hope( object );
    return ::Massiv::Core::variant_cast< $class_name * >( get_complete_object( object ) );
    }
EOT
    ) ;

    if ( $comparable ) {
        $handle->print( <<EOT

bool ${metaobject_name}::compare_objects
    (
    const ::Massiv::Core::ObjectProperty * object1,
    const ::Massiv::Core::ObjectProperty * object2
    ) const
    {
    // FIXME: do memberwise compare ourselves?

    hope( object1 );
    hope( object2 );
    const $class_name * const o1 = get_complete_object_pointer( object1 );
    const $class_name * const o2 = get_complete_object_pointer( object2 );
    return *o1 == *o2;
    }
EOT
        ) ;
    }

    if ( $assignable ) {
        $handle->print( <<EOT

void ${metaobject_name}::copy_objects
    (
    ::Massiv::Core::ObjectProperty *       destination,
    const ::Massiv::Core::ObjectProperty * source
    ) const
    {
    // FIXME: do memberwise copy ourselves?

    hope( destination );
    hope( source );
    $class_name * const dest = get_complete_object_pointer( destination );
    const $class_name * const src = get_complete_object_pointer( source );
    *dest = * src;
    }
EOT
        ) ;
    }

    # Buah, done.
}


1 ;


__END__
