=head1 NAME

Parser - simple top-down parser base class

=head1 SYNOPSIS

    use Parser ;

    ###
    # Class methods.

    # Constructor
    $parser = Parser->new ;


    ###
    # Messages.

    # Convert current source stack context to set of strings and return
    # reference to hash containing them:
    #   FILE_NAME - name of current file.
    #   LINE_NUMBER - current line number.
    #   CURRENT_LINE - complete current line.
    # Arguments:
    #   no_here - if true, 'HERE' mark won't be added to CURRENT_LINE.
    $context = $parser->source_context( [ $no_here ] ) ;

    # Print error or warning message.
    # Arguments to this method are passed in a structure:
    #   MESSAGE - message string (required)
    #   TYPE - 'w' for warnings, 'e' for errors, 'i' for internal errors (required)
    #   FILE_NAME, LINE_NUMBER, CURRENT_LINE - if not set, source_context() will be used.
    #   NO_HERE - do not add 'HERE' mark to line. Ignored if CURRENT_LINE is defined.
    $parser->print_message( \%arguments ) ;

    # Get/set warning level.
    # Only warnings with level lower or equal to this number will be reported.
    # Initial value is 1.
    $warning_level = $parser->warning_level() ;
    $parser->warning_level( $warning_level ) ;

    # Print warning message.
    # See print_message, TYPE will be set automatically to 'w'.
    # You can set LEVEL to warning level, 1 being the highest (and default).
    # If string is passed instead of reference to hash, MESSAGE will be set to this string.
    $parser->warning( \%arguments ) ;
    $parser->warning( $message ) ;

    # Print error message and optionally die.
    # See print_message, TYPE will be set automatically to 'e'.
    # If string is passed instead of reference to hash, MESSAGE will be set to this string.
    # If argument NO_DIE is set, this method will return, otherwise it will die.
    $parser->error( \%arguments ) ;
    $parser->error( $message ) ;

    # Print internal error message and die.
    # See print_message, TYPE will be set automatically to 'i'.
    # If string is passed instead of reference to hash, MESSAGE will be set to this string.
    $parser->internal_error( \%arguments ) ;
    $parser->internal_error( $message ) ;


    ###
    # Source stack.

    # Push info about given file to source stack.
    #   file_name - name of file to parse, the file must exist
    $parser->push_file( $file_name ) ;

    # Push info about given file to source stack unless it has been already
    # pushed in the past.
    #   file_name - name of file to parse, the file must exist
    $parser->push_file_unless_parsed( $file_name ) ;

    # Include paths.
    $parser->add_include_path( $path ) ;

    # Push info about given file to source stack if file exists in search path.
    #   file_name - name of file to include
    # Returns complete file name relative to current directory.
    $complete_name = $parser->include_file( $file_name ) ;

    # Push info about given file to source stack if file exists in search path
    # and it has not been pushed in the past.
    #   file_name - name of file to import
    # Returns complete file name relative to current directory.
    $complete_name = $parser->import_file( $file_name ) ;

    # Note that when file is added to source stack next token will _not_ be
    # automatically read.

    ###
    # Token reading.

    # Get reference to current token object.
    # The token object will be reallocated for each new token.
    # See F<Token.pm> for info about C<Token> class.
    $token = $parser->token() ;

    # Get current token string.
    $string = $parser->token_string() ;

    # Get current token type character.
    $type = $parser->token_type() ;

    # Get current token value.
    $value = $parser->token_value() ;
    
    # Read in next token.
    # This methods skips comments and interprets preprocessor directives.
    $parser->next_token() ;

    # Freeze current token.
    # Increases "freeze" counter. If the counter is non-zero, next_token() will
    # just decrease the counter instead of reading next token. This is pretty
    # useless since recommended usage of next_token() has been changed (and it
    # has been renamed to next_token() from read_token())
    $parser->freeze_token() ;

    # Die if current token is undefined.
    $parser->require_token() ;

    # Make sure that type of current token is equal to one of given types.
    # Print error message and die if the type is not equal to neither of given types.
    $parser->ensure_token_type( $token_type_character, ... ) ;

    # Make sure that current token is an integer.
    $parser->ensure_integer() ;

    # Make sure that current token is a float.
    $parser->ensure_float() ;

    # Make sure that current token is an integer or a float.
    $parser->ensure_number() ;

    # Make sure that current token is a string.
    $parser->ensure_string() ;

    # Make sure that current token is a quoted character.
    $parser->ensure_character() ;

    # Make sure that current token is an identifier.
    $parser->ensure_identifier() ;

    # Make sure that current token is a "symbol" ("other" token type).
    $parser->ensure_symbol() ;

    # Make sure that current token string is equal to one of given strings.
    #   string - expected token string (not token value)
    $parser->ensure_token( $string, ... ) ;

    # Ensure the token string is equal to one of given strings.
    #   string - expected token string
    # Dies with error if no string token matches. Otherwise reads in next
    # token and returns the consumed token.
    $token = $parser->consume_token( $string, ... ) ;


    ###
    # Generic parsing.

    # Analyze current token and perform action specified by supplied
    # dispatch table.
    # See DESCRIPTION for documentation of this method.
    $parser->dispatch( \%arguments ) ;
    

=head1 DESCRIPTION

The Parser class implements basic top-down parser.

The SYNOPSIS contains nearly all "documentation" that exists.

This class offers basic token-basde parser. Use C<token()> or similar
methods (such as C<token_type()>) to analyze current token and C<next_token()>
to read in next token. Use C<require_token()>, C<ensure_*()> and
C<consume_token()> methods for easier checking of token existence, type
and value.

Note that C<next_token()> will skip C++-like comments and interpret simple
preprocessing directives such as C<#include> or C<#import>.

When calling function that analyzes piece of code, current token will be the
first token to analyze - functions should not start with a call to
C<next_token()>. When such function exits, current token should the first
token after the last analyzed token.

You can use generic parsing method, C<dispatch()>. It should make standard
top-down parser dispatch easier. It's arguments are passed in a single
structure containing:

=over

=item DISPATCH

Dispatch table. It's a simple hash, where keys are token strings (B<not>
values), and values are one of:

=over

=item reference to code

If given token is found, supplied code will be executed. Reference to parser
instance (C<$self>) will be passed to this function as the only argument.

=item reference to array

First member of the array must be reference to code. It will be executed and
the reference to the array will be passed to the function as second argument.

=item "end"

Parsing of this block will be stopped. Use this for block ending token if
REPEAT is defined.

=back

B<Warning>: C<next_token()> will not be called - current token will still be
the token dispatcher analyzed (both when calling dispatch routine and when
returning from C<dispatch()>). Skip the token using C<consume_token()>
(preferred to C<next_token()> to prevent weird errors) if you don't care
about it.

=item REPEAT

If true, continue parsing when returned from dispatch routine.

=item STARTING_TOKEN

If defined, parsed block must start with this token string.

=item REQUIRE_TOKEN

If true, C<parse()> will exit with error if there is no token to parse.
Otherwise it will just return.

=back

=head1 ISSUES

Set of symbol tokens is hardwired into the parser. It should be possible to
pass it to the constructor and to precompile get_token regexp at run-time.

Parsing is still partially line-oriented and even stuff similar to C/C++ is not
100% compatible. For example, tokens can't span multiple lines, C-like multiline
strings are not supported.

=head1 KNOWN BUGS

None.

=cut

package Parser ;

use strict ;
use warnings ;

use FileHandle ;
use Cwd ;

use Utils ;
use Token ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = {

        # Print stack backtrace when dying.
        # FIXME: configurable.
        DEBUG => 0,

        # Directory the parser works in.
        ROOT_DIRECTORY => undef,

        # Array of directories to search when including files.
        # FIXME: Get rid of "" hack and use C-like searching.
        INCLUDES => [ "" ],

        # Report warnings with level <= this leve.
        WARNING_LEVEL => 1,

        # Set of parsed files. Used by #import.
        PARSED_FILES => { },

        # Source stack. Array of structures:
        #   FILE_NAME - file name, relative to current directory.
        #   FILE_HANDLE - opened FileHandle.
        #   CURRENT_LINE - current line, the rest to parse.
        #   COMPLETE_LINE - current line, complete.
        #   LINE_NUMBER - current line number.
        #   LINE_TOKEN_COUNT - number of tokens read from current line.
        # These should probably be implemented as objects too, but...
        SOURCE_STACK => [ ],
        
        # Current token.
        TOKEN => undef,

        # Freeze counter.
        TOKEN_FROZEN => 0,
        
    } ;

    # Determine current directory.

    $$self{ ROOT_DIRECTORY } = getcwd() ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# warning_level
sub warning_level {
    my $self = shift ;
    if ( @_ ) { $$self{ WARNING_LEVEL } = shift ; }
    return $$self{ WARNING_LEVEL } ;
}


###
# token
sub token {
    $_[ 0 ]{ TOKEN } ;
}


###
# token_string
sub token_string {
    my $self = shift ;
    my $token = $$self{ TOKEN } || $self->internal_error( "No token defined." ) ;
    $token->token() ;
}


###
# token_type
sub token_type {
    my $self = shift ;
    my $token = $$self{ TOKEN } || $self->internal_error( "No token defined." ) ;
    $token->type() ;
}


###
# token_value
sub token_value {
    my $self = shift ;
    my $token = $$self{ TOKEN } || $self->internal_error( "No token defined." ) ;
    $token->value() ;
}


###
# Get source stack depth.
sub source_stack_depth {
    my $self = shift ;
    my $source_stack = $$self{ SOURCE_STACK } ;
    return scalar( @$source_stack ) ;
}


###
# Get reference to topmost source stack structure - current source info.
sub current_source {
    my $self = shift ;
    my $source_stack = $$self{ SOURCE_STACK } ;
    die 'Source stack is empty' unless scalar( @$source_stack ) ;
    return $$source_stack[ $#{$source_stack} ] ;
}


###
# Get current file name.
sub file_name {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ FILE_NAME } ;
}


###
# Get current file handle.
sub file_handle {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ FILE_HANDLE } ;
}


###
# Get rest of current line to parse.
sub current_line {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ CURRENT_LINE } ;
}


###
# Get complete current line.
sub complete_line {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ COMPLETE_LINE } ;
}


###
# Get current line number.
sub line_number {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ LINE_NUMBER } ;
}


###
# Get number of tokens parser on current line.
sub line_token_count {
    my $cs = current_source( $_[ 0 ] ) ;
    return $$cs{ LINE_TOKEN_COUNT } ;
}


###
# source_context
sub source_context {
    my( $self, $no_here ) = @_ ;

    my $file_name ;
    my $line_number ;
    my $current_line ;
    
    if ( $self->source_stack_depth() ) {
        $file_name = $self->file_name() ;
        $line_number = $self->line_number() ;
        if ( $no_here ) {
            $current_line = $self->complete_line() ;
        }
        else {
            # Note: reported position will be one token off if token is frozen.
            
            my $line_rest = $self->current_line() ;
            my $complete_line = $self->complete_line() ;
            my $position = length( $complete_line ) - length( $line_rest ) ;
            $current_line = substr( $complete_line, 0, $position ) .
                ' ***HERE*** ' . substr( $complete_line, $position ) ;
        }
    }

    # Stupid "unknown" defaults.
    
    $file_name = 'unknown file'
        unless defined( $file_name ) ;
    $line_number = 'unknown line'
        unless defined( $line_number ) ;
    $current_line = ''
        unless defined( $current_line ) ;

    # Return it.

    return {
        FILE_NAME => $file_name,
        LINE_NUMBER => $line_number,
        CURRENT_LINE => $current_line
    } ;
}


###
# print_message
sub print_message {
    my( $self, $arguments ) = @_ ;

    # Get arguments.

    my $message = $$arguments{ MESSAGE } || 'Unknown error.' ;
    my $type = $$arguments{ TYPE } || 'i' ;
    my $file_name = $$arguments{ FILE_NAME } ;
    my $line_number = $$arguments{ LINE_NUMBER } ;
    my $current_line = $$arguments{ CURRENT_LINE } ;

    # Print source stack if file name and line number is read from it.

    my $print_stack ;
    unless ( defined( $file_name ) || defined( $line_number ) ) {
        $print_stack = 1 ;
    }

    # Get source context.

    my $source_context = $self->source_context( $$arguments{ NO_HERE } ) ;

    $file_name = $$source_context{ FILE_NAME }
        unless defined( $file_name ) ;
    $line_number = $$source_context{ LINE_NUMBER }
        unless defined( $line_number ) ;
    $current_line = $$source_context{ CURRENT_LINE }
        unless defined( $current_line ) ;

    # Print the message.

    print STDERR "$file_name ($line_number): " ;
    if ( $type eq 'w' ) {
        print STDERR 'warning: ' ;
    }
    elsif ( $type eq 'e' ) {
        # nothing if error
    }
    else {
        print STDERR 'INTERNAL ERROR: ' ;
    }
    print STDERR "$message\n" ;

    # Print line context.

    if ( length( $current_line ) ) {
        print STDERR "$current_line\n" ;
    }

    # Optionally print source stack.

    if ( $print_stack ) {
        my $sstack = $$self{ SOURCE_STACK } ;
        my $my_file_name = $file_name ;
        for ( my $i = $#{$sstack} - 1 ; $i >= 0 ; $i-- ) {
            my $sinfo = $$sstack[ $i ] ;
            my $parent_file_name = $$sinfo{ FILE_NAME } ;
            my $parent_line_number = $$sinfo{ LINE_NUMBER } ;
            print STDERR "$parent_file_name ($parent_line_number): '$my_file_name' included from here.\n" ;
            $my_file_name = $parent_file_name ;
        }
    }
}


###
# warning
sub warning {
    my( $self, $arguments ) = @_ ;

    my $level = $$arguments{ LEVEL } || 1 ;
    if ( $level > $$self{ WARNING_LEVEL } ) {
        return ;
    }

    unless ( ref( $arguments ) eq 'HASH' ) {
        $arguments = {
            MESSAGE => $arguments
        }
    }
    $$arguments{ TYPE } = 'w' ;

    $self->print_message( $arguments ) ;
}


###
# error
sub error {
    my( $self, $arguments ) = @_ ;

    unless ( ref( $arguments ) eq 'HASH' ) {
        $arguments = {
            MESSAGE => $arguments
        }
    }
    $$arguments{ TYPE } = 'e' ;

    $self->print_message( $arguments ) ;

    unless ( $$arguments{ NO_DIE } ) {
        if ( $$self{ DEBUG } ) {
            require Carp ;
            Carp::confess() ;
        }
        else {
            die ;
        }
    }
}


###
# internal error
sub internal_error {
    my( $self, $arguments ) = @_ ;

    unless ( ref( $arguments ) eq 'HASH' ) {
        $arguments = {
            MESSAGE => $arguments
        }
    }
    $$arguments{ TYPE } = 'i' ;

    $self->print_message( $arguments ) ;

    require Carp ;
    Carp::confess() ;
}


###
# add_include_path
sub add_include_path {
    my $self = shift ;
    my $path = shift ;
    push( @{$$self{ INCLUDES }}, $path ) ;
}


###
# push_file
sub push_file {
    my( $self, $file_name ) = @_ ;

    # Open file.

    my $handle = FileHandle->new( $file_name, 'r' ) ;
    unless ( $handle ) {
        $self->error( "Can't open file '$file_name'." ) ;
    }

    # If current stack has single entry with null file handle, remove it.
    # This will happen iff previous file has been completely parsed - its
    # entry has not been removed by _read_line() to allow sane error messages.

    my $sstack = $$self{ SOURCE_STACK } ;
    if ( scalar( @$sstack ) == 1 ) {
        unless ( defined( $$sstack[ 0 ]{ FILE_HANDLE } ) ) {
            pop( @$sstack ) ;
        }
    }

    # Push new entry to source stack.

    push( @$sstack, {
        FILE_NAME => $file_name,
        FILE_HANDLE => $handle,
        CURRENT_LINE => '',
        COMPLETE_LINE => '',
        LINE_NUMBER => 0,
        LINE_TOKEN_COUNT => 0
    } ) ;

    $$self{ PARSED_FILES }{ $file_name }++ ;
}


###
# push_file_unless_parsed
sub push_file_unless_parsed {
    my( $self, $file_name ) = @_ ;
    unless ( defined( $$self{ PARSED_FILES }{ $file_name } ) ) {
        $self->push_file( $file_name ) ;
    }
}

###
# include_file
sub include_file {
    my( $self, $file_name ) = @_ ;

    my $found = Utils::find_file( $file_name, $$self{ ROOT_DIRECTORY }, @{$$self{ INCLUDES }} ) ;
    if ( defined( $found ) ) {
        $self->push_file( $found ) ;
    }
    else {
        $self->error( "Can't find file '$file_name'." ) ;
    }

    return $found ;
}


###
# import_file
sub import_file {
    my( $self, $file_name ) = @_ ;

    my $found = Utils::find_file( $file_name, $$self{ ROOT_DIRECTORY }, @{$$self{ INCLUDES }} ) ;
    if ( defined( $found ) ) {
        $self->push_file_unless_parsed( $found ) ;
    }
    else {
        $self->error( "Can't find file '$file_name'." ) ;
    }

    return $found ;
}


###
# Read in next line.
# Return 1 if line is read, 0 otherwise.
# Note that the last entry from source stack will be never removed.
sub _read_line {
    my $self = shift ;
    my $source_stack = $$self{ SOURCE_STACK } ;

    while ( scalar( @$source_stack ) ) {
        # Try to read line from current source.

        my $cs = $$source_stack[ $#{$source_stack} ] ;
        last unless defined( $$cs{ FILE_HANDLE } ) ;
        my $line = $$cs{ FILE_HANDLE }->getline() ;

        if ( defined( $line ) ) {
            chomp( $line ) ;
            $$cs{ CURRENT_LINE } = $line ;
            $$cs{ COMPLETE_LINE } = $line ;
            $$cs{ LINE_NUMBER }++ ;
            $$cs{ LINE_TOKEN_COUNT } = 0 ;
            return 1 ;
        }

        # End of this file.

        $$cs{ FILE_HANDLE }->close() ;
        if ( scalar( @$source_stack ) == 1 ) {
            # Leave the last stack entry there.
            
            $$cs{ FILE_HANDLE } = undef ;
            last ;
        }

        # Finished current stack entry.

        pop( @$source_stack ) ;
    }

    # End of all files.

    return 0 ;
}


###
# Read next token from current line.
# Modifies source stack, current line and current token.
# Returns 1 if token has been read, 0 otherwise.
sub _read_current_line_token {
    my $self = shift ;
    my $cs = $self->current_source() ;

    # Parse remaining part of current line.

    my( $token ) = $$cs{ CURRENT_LINE } =~ m{
        ^\s*                    # Skip initial whitespace.
        (                       # Start of token.
            0x[0-9a-fA-F]+          # Hexadecimal number.
        |            
            0b[01]+                 # Binary number.
        |
                                    # Decimal or floating point number. Octal too, but 8 and 9 are allowed. FIXME.
            (?= \d | \.\d )             # Begins with "N" or ".N", ie. at least one digit.
            \d*                         # Whole part.
            (?: \.\d* )?                # Optional fractional part.
            (?: [eE] [+-]? \d+ )?       # Optional exponent.
        |
            "                       # Doublequoted string.
            (?: \\. | [^"\\] )*         # Escaped chars or neither doublequote nor escape.
            "  
        |
            '                       # Single character in quotes.
            (?: \\. | [^'\\] )
            '
        |
            \w+                     # Identifier.
        |
            /\* | \*/ | // |        # Multicharacter symbols.
            ::
        |
            \S                      # Single nonwhitespace character.
        )                       # End of token.
    }x ;

    if ( defined( $token ) ) {
        $$self{ TOKEN } = Token->new( $token ) ;
        $$cs{ CURRENT_LINE } = $' ;
        $$cs{ LINE_TOKEN_COUNT }++ ;
        return 1 ;
    }

    $$self{ TOKEN } = undef ;
    return 0 ;
}


###
# Skip till the end of /* */ comment.
sub _skip_comment {
    my $self = shift ;
    my $source_stack = $$self{ SOURCE_STACK } ;

    my $error_arguments = $self->source_context() ;

    LINE : for ( ; ; ) {
        my $cs = $$source_stack[ $#{$source_stack} ] ;

        # Check current line for '*/'

        if ( $$cs{ CURRENT_LINE } =~ s|^.*?\*/|| ) {
            last LINE ;
        }

        # Read in next line.
        
        unless ( $self->_read_line() ) {
            $$error_arguments{ MESSAGE } = 'Unterminated comment.' ;
            $self->error( $error_arguments ) ;
        }
    }
}


###
# Read and analyze next token.
#   single_line - if defined to true, parser will not advance to next line
#                 unless end-of-line is commented out using /* */
# This method skips comments but does not interpret other tokens.
sub _read_token {
    my( $self, $single_line ) = @_ ;
    my $source_stack = $$self{ SOURCE_STACK } ;

    TOKEN : for ( ; ; ) {
        my $cs = $$source_stack[ $#{$source_stack} ] ;

        # Make sure we have non-empty line.

        if ( $$cs{ CURRENT_LINE } =~ /^\s*$/ ) {
            unless ( $single_line ) {
                if ( $self->_read_line() ) {
                    next TOKEN ;
                }
            }

            # End of all files or end of line and single_line is set - no token.

            $$self{ TOKEN } = undef ;
            last TOKEN ;
        }

        # Read token from current line.

        unless ( $self->_read_current_line_token() ) {
            # Failed. Line is not empty, so this is an error.
            # This should never happen as all nonwhitespace chars are recognized as symbols.

            $self->internal_error( "Can't read next token." ) ;
        }

        # Skip comments.

        if ( $$self{ TOKEN }->type() eq 'S' ) {
            my $value = $$self{ TOKEN }->value() ;

            if ( $value eq '//' ) {
                # Comment till the end of line.

                $$cs{ CURRENT_LINE } = '' ;
                next TOKEN ;
            }
            if ( $value eq '/*' ) {
                # Skip till */ token.

                $self->_skip_comment() ;
                next TOKEN ;
            }
        }

        # Return current token.

        last TOKEN ;
    }
}


###
# Ensure that there are no more tokens on current line.
sub _ensure_preprocessor_end {
    my $self = shift ;
    $self->_read_token( 1 ) ;
    if ( defined( $$self{ TOKEN } ) ) {
        $self->error( 'Garbage after preprocessor directive.' ) ;
    }
}


###
# Parse preprocessor directive.
# Note: this is "old-style" parsing - it begins with _read_token()
#       and leaves last analyzed token as current token.
sub _preprocessor_directive {
    my $self = shift ;

    $self->_read_token( 1 ) ;
    unless ( defined( $$self{ TOKEN } ) ) {
        $self->error( 'Unspecified # directive.' ) ;
    }
    my $directive = $$self{ TOKEN }->token() ;

    if ( $directive eq 'include' ) {
        # Read file name.
        
        $self->_read_token( 1 ) ;
        $self->ensure_string() ;
        my $file_name = $$self{ TOKEN }->value() ;
        $self->_ensure_preprocessor_end() ;

        # Include the file.

        $self->include_file( $file_name ) ;
    }
    elsif ( $directive eq 'import' ) {
        # Read file name.

        $self->_read_token( 1 ) ;
        $self->ensure_string() ;
        my $file_name = $$self{ TOKEN }->value() ;
        $self->_ensure_preprocessor_end() ;

        # Import the file.

        $self->import_file( $file_name ) ;
    }
    elsif ( $directive eq 'warn' ) {
        # Read the warning.

        $self->_read_token( 1 ) ;
        $self->ensure_string() ;

        # Print it.
        # FIXME: Interpolate the string?

        $self->warning( {
            MESSAGE => $$self{ TOKEN }->value,
            CURRENT_LINE => ''
        } ) ;
    }
    elsif ( $directive eq 'error' ) {
        # Read the error.

        $self->_read_token( 1 ) ;
        $self->ensure_string() ;

        # Print it.
        # FIXME: Interpolate the string?

        $self->error( {
            MESSAGE => $$self{ TOKEN }->value,
            CURRENT_LINE => ''
        } ) ;
    }
    else {
        $self->error( "Unknown # directive '$directive'." ) ;
    }
}


###
# next_token
sub next_token {
    my $self = shift ;
    my $source_stack = $$self{ SOURCE_STACK } ;

    # Do nothing if current token is frozen.

    if ( $$self{ TOKEN_FROZEN } ) {
        $$self{ TOKEN_FROZEN }-- ;
        return ;
    }

    TOKEN : for ( ; ; ) {

        # Read next token.

        $self->_read_token() ;
        unless ( defined( $$self{ TOKEN } ) ) {
            last TOKEN ;
        }

        # Preprocess.

        if ( $$self{ TOKEN }->token eq '#' ) {
            my $cs = $$source_stack[ $#{$source_stack} ] ;
            if ( $$cs{ LINE_TOKEN_COUNT } == 1 ) {
                $self->_preprocessor_directive() ;
                next TOKEN ;
            }
        }

        # Return current token.

        last TOKEN ;
    }
}


###
# freeze_token
sub freeze_token {
    my $self = shift ;
    $$self{ TOKEN_FROZEN }++ ;
}


###
# require_token
sub require_token {
    my $self = shift ;
    unless ( defined( $$self{ TOKEN } ) ) {
        $self->error( 'Parse error at end of input.' ) ;
    }
}


###
# ensure_token_type
sub ensure_token_type {
    my $self = shift ;
    $self->require_token() ;
    my $error = $$self{ TOKEN }->ensure_type( @_ ) ;
    if ( defined( $error ) ) {
        $self->error( 'Parse error. ' . $error ) ;
    }
}


###
# ensure_integer
sub ensure_integer {
    $_[ 0 ]->ensure_token_type( 'i' ) ;
}


###
# ensure_float
sub ensure_float {
    $_[ 0 ]->ensure_token_type( 'f' ) ;
}


###
# ensure_number
sub ensure_number {
    $_[ 0 ]->ensure_token_type( 'i', 'f' ) ;
}


###
# ensure_string
sub ensure_string {
    $_[ 0 ]->ensure_token_type( 's' ) ;
}


###
# ensure_character
sub ensure_character {
    $_[ 0 ]->ensure_token_type( 'c' ) ;
}


###
# ensure_identifier
sub ensure_identifier {
    $_[ 0 ]->ensure_token_type( 'I' ) ;
}


###
# ensure_symbol
sub ensure_symbol {
    $_[ 0 ]->ensure_token_type( 'S' ) ;
}


###
# ensure_token
sub ensure_token {
    my $self = shift ;
    $self->require_token() ;
    my $error = $$self{ TOKEN }->ensure_token( @_ ) ;
    if ( defined( $error ) ) {
        $self->error( 'Parse error. ' . $error ) ;
    }
}


###
# consume_token
sub consume_token {
    my $self = shift ;
    $self->ensure_token( @_ ) ;
    my $result = $$self{ TOKEN } ;
    $self->next_token() ;
    return $result ;
}


###
# dispatch
sub dispatch {
    my( $self, $arguments ) = @_ ;
    my $dispatch = $$arguments{ DISPATCH } or
        $self->internal_error( "No dispatch table supplied to parse()." ) ;
    my $repeat = $$arguments{ REPEAT } ;
    my $start = $$arguments{ STARTING_TOKEN } ;
    my $require = $$arguments{ REQUIRE_TOKEN } ;

    # Consume starting token.

    if ( defined( $start ) ) {
        $self->consume_token( $start ) ;
    }

    TOKEN : for ( ; ; ) {
        # Read token.

        if ( $require ) {
            $self->require_token() ;
        }
        my $token = $$self{ TOKEN } ;
        last TOKEN unless defined( $token ) ;
        my $token_string = $token->token() ;

        # Dispatch.

        my $value = $$dispatch{ $token_string } ;
        unless ( defined( $value ) ) {
            my $message = "Parse error.\n" ;
            my @dispatch_tokens = sort( keys( %$dispatch ) ) ;
            if ( scalar( @dispatch_tokens ) > 1 ) {
                $message .= "Expected one of " .
                     join( ', ', map( { "'$_'" } @dispatch_tokens ) ) ;
            }
            else {
                $message .= "Expected '$dispatch_tokens[ 0 ]'" ;
            }
            $message .= ",\nfound '$token_string'." ;
            $self->error( $message ) ;
        }

        # next_token() not called!

        my $ref_value = ref( $value ) ;
        if ( $ref_value eq 'CODE' ) {
            &$value( $self ) ;
        }
        elsif ( $ref_value eq 'ARRAY' ) {
            my $code = $$value[ 0 ] ;
            if ( ref( $code ) ne 'CODE' ) {
                $self->internal_error( "First element of the array must be reference to code." ) ;
            }
            &$code( $self, $value ) ;
        }
        elsif ( $value eq 'end' ) {
            last TOKEN ;
        }
        else {
            $self->internal_error( "Unsupported dispatch table entry '$value'." ) ;
        }

        # Repeat.

        last TOKEN unless $repeat ;
    }
}


1 ;


__END__
