=head1 NAME

RPCGen.pm - RPC-related code generator

=head1 SYNOPSIS

    use RPCGen ;

    $class_name = RPCGen::method_packet_name( $class_info, $method_index ) ;

    RPCGen::write_tag_class( $handle, $class_info ) ;

    RPCGen::write_packet_classes( $handle, $class_info ) ;

    RPCGen::write_packet_source( $handle, $class_info ) ;

    RPCGen::write_stub_class( $handle, $class_info ) ;

    RPCGen::write_stub_source( $handle, $class_info ) ;

=head1 DESCRIPTION

Use this module to write interfaces and implementations of RPC-related classes
for all methods defined in class specified by I<$class_info>.

=head1 TODO

Do not ignore method return type - treat it as out argument "_result".

RPC results.

=head1 KNOWN BUGS

None.

=cut

package RPCGen ;

use strict ;
use warnings ;

use ClassInfo ;


###
# method_packet_name
sub method_packet_name {
    my $class_info = shift ;
    my $method_index = shift ;
    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $methods = $class_symbol->methods() ;
    my $method_symbol = $$methods[ $method_index ] ;
    my $method_name = $method_symbol->name() ;

    return "MethodPacketImpl< ::Massiv::Generated::MethodTags< $class_name >::$method_name >" ;
}


###
# write_tag_class
sub write_tag_class {
    my( $handle, $class_info ) = @_ ;
    
    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $methods = $class_symbol->methods() ;

    my $symbol_attribute = $class_symbol->require_attribute( '::symbol_attribute' ) ;
    $symbol_attribute .= ' ' if $symbol_attribute ne '' ;

    # Tags.

    $handle->print( <<EOT

/**
 * Method tags for $class_name.
 */
template<>
struct ${symbol_attribute}MethodTags< $class_name >
    {
EOT
    ) ;

    my $method_index = 0 ;
    foreach my $method ( @$methods ) {
        my $method_name = $method->name() ;
        $handle->print( "    struct $method_name { enum { index = $method_index }; };\n" ) ;
        $method_index++ ;
    }

    $handle->print( <<EOT    
    };
EOT
    ) ;
}


###
# write_packet_classes
sub write_packet_classes {
    my( $handle, $class_info ) = @_ ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $methods = $class_symbol->methods() ;

    my $symbol_attribute = $class_symbol->require_attribute( '::symbol_attribute' ) ;
    $symbol_attribute .= ' ' if $symbol_attribute ne '' ;

    # Packets.
    
    my $method_index = 0 ;
    foreach my $method ( @$methods ) {
        my $method_name = $method->name() ;
        my $arguments = $method->arguments() ;
        my $result_type = $method->result_type() ;
        my $has_arguments = $method->has_in_arguments() ;
        my $has_results = $method->has_out_arguments() ;
        my $packet_class_name = method_packet_name( $class_info, $method_index ) ;

        # Write arguments/results class.

        $handle->print( <<EOT

/**
 * Method packet class for ${class_name}::${method_name}.
 */
template<>
class ${symbol_attribute}$packet_class_name : public ::Massiv::Core::MethodPacket
    {
public:
EOT
        ) ;

        if ( $has_arguments ) {
            _write_packet_serializer_declarations( $handle, 'arguments' ) ;
        }
        if ( $has_results ) {
            _write_packet_serializer_declarations( $handle, 'results' ) ;
        }
        if ( $has_arguments or $has_results ) {
            $handle->print( <<EOT

    /** Get value of $method_name \@a index-th argument. */
    virtual std::string get_argument_value
        (
        int index
        ) const;
EOT
            ) ;
        }

        $handle->print( <<EOT

    /** Call the method using pointer to $class_name component. */
    virtual void call_method
        (
        const ::Massiv::Core::VariantPointer & object
        );

    /** Call the method using object pointer to #Object. */
    virtual bool call_method
        (
        const ::Massiv::Core::WeakPointer< ::Massiv::Core::Object > & object
        );
EOT
        ) ;

        if ( $has_arguments || $has_results ) {
            $handle->print( "\npublic:\n\n" ) ;
        }

        foreach my $argument ( @$arguments ) {
            my $name = $$argument{ NAME } ;
            my $type = $$argument{ TYPE } ;
            $handle->print( '    ' . $type->serializable_type() . ' ' . $name . ";\n" ) ;
        }
        if ( defined( $result_type ) ) {
            $handle->print( '    ' . $result_type->serializable_type() . " _result;\n" ) ;
        }

        $handle->print( <<EOT

    }; // class $packet_class_name
EOT
        ) ;

        $method_index++ ;
    }
}


###
# _write_packet_serializer_declarations
sub _write_packet_serializer_declarations {
    my( $handle, $kind ) = @_ ;

    $handle->print( <<EOT

    /** Serialize $kind to bitstream. */
    virtual void bitstr_write_$kind
        (
        ::Massiv::Core::BitOStream &                    bos,
        const ::Massiv::Core::Serializer::Description & desc
        ) const;

    /** Read $kind from bitstream. */
    virtual void bitstr_read_$kind
        (
        ::Massiv::Core::BitIStream &                    bis,
        const ::Massiv::Core::Serializer::Description & desc
        );

    /** Serialize $kind to text stream. */
    virtual void text_write_$kind
        (
        ::Massiv::Core::TextWriter &                    textw,
        const ::Massiv::Core::Serializer::Description & desc
        ) const;

    /** Read $kind from text stream. */
    virtual void text_read_$kind
        (
        ::Massiv::Core::TextReader &                    textr,
        const ::Massiv::Core::Serializer::Description & desc
        );
EOT
    ) ;
}


###
# write_packet_source
sub write_packet_source {
    my( $handle, $class_info ) = @_ ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $methods = $class_symbol->methods() ;

    my $method_index = 0 ;
    foreach my $method ( @$methods ) {
        my $packet_class_name = method_packet_name( $class_info, $method_index ) ;
        my $method_name = $method->name() ;
        my $arguments = $method->arguments() ;
        my $result_type = $method->result_type() ;

        $handle->print( "\n// Method packet for ${class_name}::${method_name}.\n" ) ;

        # Write implementation of serializers.

        if ( $method->has_in_arguments() ) {
            _write_packet_serializer_definitions
                ( $handle, 'a', $packet_class_name, $arguments );
        }
        if ( $method->has_out_arguments() ) {
            _write_packet_serializer_definitions
                ( $handle, 'r', $packet_class_name, $arguments, $result_type );
        }

        # Write implementation of get_argument_value().

        if ( $method->has_in_arguments() or $method->has_out_arguments() ) {
            $handle->print( <<EOT

std::string ${packet_class_name}::get_argument_value
    (
    int index
    ) const
    {
    std::ostringstream ss;
    Massiv::Core::TextWriter tw( ss );
    Massiv::Core::Serializer::Description desc;
    switch( index )
        {
EOT
        ) ;

            # Result.

            if ( defined( $result_type ) ) {
                $handle->print( <<EOT
        case -1:
            _result.text_write( tw, desc );
            break;
EOT
                ) ;
            }

            # Arguments.

            my $argument_index = 0 ;
            foreach my $argument ( @$arguments ) {
                my $aname = $$argument{ NAME } ;
                $handle->print( <<EOT
        case ${argument_index}:
            ${aname}.text_write( tw, desc );
            break;
EOT
                ) ;
                $argument_index++ ;
            }

            # Done.

            $handle->print( <<EOT
        default:
            return MethodPacket::get_argument_value( index );
        }
    return ss.str();
    }
EOT
            ) ;
        }

        # Write implementation of calls.

        $handle->print( <<EOT

void ${packet_class_name}::call_method
    (
    const ::Massiv::Core::VariantPointer & _object
    )
    {
    $class_name * _component = ::Massiv::Core::variant_cast< $class_name * >( _object );
    ::Massiv::Core::UpcallToObject _upcall = _component;
    _upcall.tag_upcall( "${class_name}::${method_name}( ... )" );
EOT
        ) ;

        $handle->print( '    ' ) ;
        if ( defined( $result_type ) ) {
            $handle->print( '_result = ' ) ;
        }
        $handle->print( '_component->' . $method_name . "(" ) ;
        my $argument_index = 0 ;
        foreach my $argument ( @$arguments ) {
            if ( $argument_index == 0 ) {
                $handle->print( ' ' ) ;
            }
            else {
                $handle->print( ', ' ) ;
            }
            $handle->print( $$argument{ NAME } ) ;
            $argument_index++ ;
        }
        if ( $argument_index > 0 ) {
            $handle->print( " " ) ;
        }
        $handle->print( ");\n" ) ;
        
        $handle->print( <<EOT
    }

bool ${packet_class_name}::call_method
    (
    const ::Massiv::Core::WeakPointer< ::Massiv::Core::Object > & _object
    )
    {
    ::Massiv::Core::WeakPointer< $class_name > _o;
    ::Massiv::Core::Pointer< $class_name > _o_strong;
    try
        {
        _o = _object.convert();
        }
    catch( ::Massiv::Core::Lib::IllegalPointerConversionException & )
        {
        return false;
        }
    if( !_o->is_replica() )
        {
        _o_strong = _o;
        }
    call_method( ::Massiv::Core::VariantPointer( _o.dereference() ) );
    return true;
    }
EOT
        ) ;

        $method_index++ ;
    }
}


###
# _write_packet_serializer_definitions
sub _write_packet_serializer_definitions {
    my( $handle, $type, $packet_class_name, $arguments, $return_type ) = @_ ;

    my $skip_kind ;
    my $suffix ;
    if ( $type eq 'a' ) {
        $skip_kind = 'out' ;
        $suffix = 'arguments' ;
        $return_type = undef ;
    }
    else {
        $skip_kind = 'in' ;
        $suffix = 'results' ;
    }

    $handle->print( <<EOT
    
void ${packet_class_name}::bitstr_write_$suffix
    (
    ::Massiv::Core::BitOStream &                    _bos,
    const ::Massiv::Core::Serializer::Description & _desc
    ) const
    {
EOT
    ) ;

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq $skip_kind ;
        $handle->print( '    ' . $$argument{ NAME } . ".bitstr_write( _bos, _desc );\n" ) ;
    }
    if ( defined( $return_type ) ) {
        $handle->print( "    _result.bitstr_write( _bos, _desc );\n" ) ;
    }

    $handle->print( <<EOT
    }

void ${packet_class_name}::bitstr_read_$suffix
    (
    ::Massiv::Core::BitIStream &                    _bis,
    const ::Massiv::Core::Serializer::Description & _desc
    )
    {
EOT
    ) ;

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq $skip_kind ;
        $handle->print( '    ' . $$argument{ NAME } . ".bitstr_read( _bis, _desc );\n" ) ;
    }
    if ( defined( $return_type ) ) {
        $handle->print( "    _result.bitstr_read( _bis, _desc );\n" ) ;
    }

    $handle->print( <<EOT
    }

void ${packet_class_name}::text_write_$suffix
    (
    ::Massiv::Core::TextWriter &                    _textw,
    const ::Massiv::Core::Serializer::Description & _desc
    ) const
    {
EOT
    ) ;

    my $argument_index = 0 ;
    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq $skip_kind ;
        if ( $argument_index > 0 ) {
            $handle->print( "    _textw.write_space();\n" ) ;
        }
        $handle->print( '    ' . $$argument{ NAME } . ".text_write( _textw, _desc );\n" ) ;
        $argument_index++ ;
    }
    if ( defined( $return_type ) ) {
        if ( $argument_index > 0 ) {
            $handle->print( "    _textw.write_space();\n" ) ;
        }
        $handle->print( "    _result.text_write( _textw, _desc );\n" ) ;
    }

    $handle->print( <<EOT
    }

void ${packet_class_name}::text_read_$suffix
    (
    ::Massiv::Core::TextReader &                    _textr,
    const ::Massiv::Core::Serializer::Description & _desc
    )
    {
EOT
    ) ;

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq $skip_kind ;
        $handle->print( '    ' . $$argument{ NAME } . ".text_read( _textr, _desc );\n" ) ;
    }
    if ( defined( $return_type ) ) {
        $handle->print( "    _result.text_read( _textr, _desc );\n" ) ;
    }

    $handle->print( <<EOT
    }
EOT
    ) ;
}


###
# write_stub_class
sub write_stub_class {
    my( $handle, $class_info ) = @_ ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $parents = $class_symbol->parents() ;
    my $methods = $class_symbol->methods() ;

    my $symbol_attribute = $class_symbol->require_attribute( '::symbol_attribute' ) ;
    $symbol_attribute .= ' ' if $symbol_attribute ne '' ;

    my $stub_class_name = "RPCStubs< $class_name >" ;

    # RPCStubs class, parents.

    $handle->print( <<EOT

/**
 * RPC stubs for $class_name.
 */
template<>
EOT
    ) ;
    $handle->print( "class ${symbol_attribute}$stub_class_name :" ) ;
    my $parent_index = 0 ;
    if ( scalar( @$parents ) ) {
        foreach my $parent ( @$parents ) {
            my $parent_name = $$parent{ SYMBOL }->full_name() ;
            my $parent_virtual = $$parent{ VIRTUAL } ;

            $handle->print( ',' ) if $parent_index > 0 ;
            $handle->print( ' public' ) ;
            $handle->print( ' virtual' ) if $parent_virtual ;
            $handle->print( " RPCStubs< $parent_name >" ) ;

            $parent_index++ ;
        }
    }
    else {
        $handle->print( ' public ::Massiv::Core::RPCStubsBase' ) ;
    }
    $handle->print( "\n" ) ;

    # Parameter modification.

    $handle->print( <<EOT
    {
public:

    /** Set call flags and delivery time. */
    $stub_class_name & param
        (
        int                             flags = ::Massiv::Core::RPC_DEFAULT,
        const ::Massiv::Core::STime &   time = ::Massiv::Core::STime()
        );

    /** Request replication of callee for \@a timeout seconds. */
    $stub_class_name & request_replica
        (
        const ::Massiv::Core::STime & timeout
        );

    /** Request replication of callee, allow optimized replica SRPC. */
    $stub_class_name & optimize_replica
        (
        const ::Massiv::Core::STime & timeout
        );

public:    
EOT
    ) ;

    # Method stubs.

    foreach my $method ( @$methods ) {
        my $method_name = $method->name() ;
        my $arguments = $method->arguments() ;
        my $cpp_result = $method->cpp_result_type() ;

        # Asynchronous shortcut version.

        $handle->print( "\n    /** Call $method_name asynchronously. */\n" .
            "    ::Massiv::Core::Pointer< ::Massiv::Core::ResultObject > async_$method_name" ) ;
        _write_arguments( $handle, $arguments, 't', 'out' ) ;
        $handle->print( ";\n" ) ;

        # Synchronous version.

        $handle->print( "\n    /** Call $method_name synchronously. */\n    $cpp_result sync_$method_name" ) ;
        _write_arguments( $handle, $arguments, 't' ) ;
        $handle->print( ";\n" ) ;
    }

    # The end.

    $handle->print( "\n    }; // class $stub_class_name\n" ) ;
}


###
# write_stub_source
sub write_stub_source {
    my( $handle, $class_info ) = @_ ;

    my $class_symbol = $class_info->class_symbol( 0 ) ;
    my $class_name = $class_symbol->full_name() ;
    my $methods = $class_symbol->methods() ;

    my $stub_class_name = "RPCStubs< $class_name >" ;

    # Parameter modification.

    $handle->print( <<EOT

// RPC stubs for $class_name.

$stub_class_name & ${stub_class_name}::param( int flags, const ::Massiv::Core::STime & time )
    {
    _param( flags, time );
    return *this;
    }

$stub_class_name & ${stub_class_name}::request_replica( const ::Massiv::Core::STime & timeout )
    {
    _replica_timeout = timeout;
    return *this;
    }

$stub_class_name & ${stub_class_name}::optimize_replica( const ::Massiv::Core::STime & timeout )
    {
    _replica_timeout = timeout;
    return param( ::Massiv::Core::RPC_ALLOW_LOCAL_REPLICA_CALL );
    }
EOT
    ) ;

    # Write method stubs.

    my $method_index = 0 ;
    foreach my $method ( @$methods ) {
        _write_async_method_stub
            (
            $handle,
            $stub_class_name,
            $class_info,
            $method_index,
            $method
            ) ;

        _write_sync_method_stub
            (
            $handle,
            $stub_class_name,
            $class_info,
            $class_name,
            $method_index,
            $method
            ) ;

        $method_index++ ;
    }
}


###
# _write_async_method_stub
sub _write_async_method_stub {
    my( $handle, $stub_class_name, $class_info, $method_index, $method ) = @_ ;
    my $metaobject_name = $class_info->metaobject_name() ;
    my $method_name = $method->name() ;
    my $arguments = $method->arguments() ;
    my $packet_class_name = method_packet_name( $class_info, $method_index ) ;

    $handle->print( <<EOT

::Massiv::Core::Pointer< ::Massiv::Core::ResultObject > ${stub_class_name}::async_$method_name
EOT
    ) ;

    $handle->print( '    ' ) ;
    _write_arguments( $handle, $arguments, 'tn', 'out' ) ;
    $handle->print( <<EOT

    {
    $packet_class_name _packet;
EOT
    ) ;

    # Initialize arguments.

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq 'out' ;
        my $name = $$argument{ NAME } ;

        # Because arguments are always passed by value, class type must match.
        
        my $type = $$argument{ TYPE } ;
        if( ref( $type ) eq 'TypeClass' ) {
            my $arg_symbol = $type->class_symbol() ;
            my $arg_name = $arg_symbol->full_name() ;
            $handle->print( "    _check_class_argument_type( typeid( $arg_name ), typeid( $name ) );\n" ) ;
        }

        # Copy by value to packet.
        
        $handle->print( "    _packet.$name = $name;\n" ) ;
    }

    # Create and send RPC request.

    $handle->print( <<EOT
    const ::Massiv::Core::ClassTypeId & _class_type_id = ${metaobject_name}::class_id;
    if( _flags & ::Massiv::Core::RPC_REPLICAS )
        {
EOT
    ) ;

    if ( $method->require_attribute( '::const' ) ) {
        $handle->print( <<EOT
    ::Massiv::Core::RPCObject::create_and_send_to_replicas
        ( _object_id, _get_absolute_time(), _packet, _class_type_id, $method_index );
    return ::Massiv::Core::Pointer< ::Massiv::Core::ResultObject >();
EOT
        ) ;
    }
    else {
        $handle->print( <<EOT
        _tried_to_call_non_const_method();
        /*NOTREACHED*/
EOT
        ) ;
    }

    $handle->print( <<EOT
        }
    else
        {
        return ::Massiv::Core::RPCObject::create_and_send
            ( _object_id, _get_absolute_time(), _packet, _class_type_id, $method_index, _require_reply(), _replica_timeout, _no_archive() );
        }
    }
EOT
    ) ;

}


###
# _write_sync_method_stub
sub _write_sync_method_stub {
    my( $handle, $stub_class_name, $class_info, $class_name, $method_index, $method ) = @_ ;
    my $method_name = $method->name() ;
    my $arguments = $method->arguments() ;
    my $result_type = $method->result_type() ;
    my $cpp_result = $method->cpp_result_type() ;
    my $packet_class_name = method_packet_name( $class_info, $method_index ) ;

    $handle->print( "\n$cpp_result ${stub_class_name}::sync_$method_name" ) ;
    _write_arguments( $handle, $arguments, 'tn' ) ;
    $handle->print( <<EOT

    {
    _sync();
EOT
    ) ;

    # Check out value type arguments.

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind ne 'out' ;
        my $name = $$argument{ NAME } ;
        my $type = $$argument{ TYPE } ;
        if ( ref( $type ) eq 'TypeClass' ) {
            my $arg_symbol = $type->class_symbol() ;
            my $arg_name = $arg_symbol->full_name() ;
            $handle->print( "    _check_class_argument_type( typeid( $arg_name ), typeid( $name ) );\n" ) ;
        }
    }

    # Can method be called directly?

    $handle->print( "    ::Massiv::Core::ObjectProperty * const _object = " ) ;
    if( $method->require_attribute( '::const' ) ) {
        $handle->print( "_local_const_srpc_call();\n" ) ;
    }
    else {
        $handle->print( "_local_srpc_call();\n" ) ;
    }

    $handle->print( <<EOT
    if( _object )
        {
        std::size_t _arg_size = 0;
EOT
    ) ;

    # Check in and inout value type arguments.
    # (otherwise they will be checked in async_*() call)

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq 'out' ;
        my $name = $$argument{ NAME } ;
        my $type = $$argument{ TYPE } ;
        if ( ref( $type ) eq 'TypeClass' ) {
            my $arg_symbol = $type->class_symbol() ;
            my $arg_name = $arg_symbol->full_name() ;
            $handle->print( "        _check_class_argument_type( typeid( $arg_name ), typeid( $name ) );\n" ) ;
        }
    }

    # Compute memory occupied by in and inout arguments for balancer.

    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        next if $kind eq 'out' ;
        my $name = $$argument{ NAME } ;
        my $type = $$argument{ TYPE } ;
        $handle->print( "        _arg_size += " . $type->estimate_size( $name ) . ";\n" ) ;
    }

    # Direct call.

    my $divide = '/' ; # hack because of stupid kdevelop syntax highlighting, ignore it...
    $handle->print( <<EOT
        ::Massiv::Core::Global::balancer().account_for_local_call( _object, ( _arg_size + 7 ) $divide 8 );
        ::Massiv::Core::UpcallToObject _upcall = _object;
        _upcall.tag_upcall( "${class_name}::${method_name}( ... )" );
        ::Massiv::Core::WeakPointer< $class_name > _o( _object_id );
        ::Massiv::Core::Pointer< $class_name > _o_strong;
        if( !_o->is_replica() )
            {
            _o_strong = _o;
            }
EOT
        ) ;

    $handle->print( ' ' x 8 ) ;
    if ( defined( $result_type ) ) {
        $handle->print( 'return ' ) ;
    }
    $handle->print( '_o->' . $method_name . "(" ) ;
    my $argument_index = 0 ;
    foreach my $argument ( @$arguments ) {
        if ( $argument_index == 0 ) {
            $handle->print( ' ' ) ;
        }
        else {
            $handle->print( ', ' ) ;
        }
        $handle->print( $$argument{ NAME } ) ;
        $argument_index++ ;
    }
    if ( $argument_index > 0 ) {
        $handle->print( " " ) ;
    }
    $handle->print( ");\n" ) ;
    unless ( defined( $result_type ) ) {
        $handle->print( ' ' x 8 . "return;\n" ) ;
    }

    # Call async version otherwise.

    $handle->print( <<EOT
        }
    ::Massiv::Core::Pointer< ::Massiv::Core::ResultObject > _result
EOT
    ) ;
    $handle->print( ' ' x 8 . "( async_$method_name" ) ;
    _write_arguments( $handle, $arguments, 'n', 'out' ) ;
    $handle->print( " );\n" ) ;

    # Wait for results.

    $handle->print( <<EOT
    _result->wait_for_reply();
    hope( _result->state == ::Massiv::Core::ResultObject::STATE_SUCCESS );
EOT
    ) ;

    # Create method packet if method returns or has [in]out arguments.

    if ( $method->has_out_arguments() ) {
        $handle->print( <<EOT
    std::auto_ptr< $packet_class_name > _packet
        ( checked_cast< $packet_class_name * >( _result->create_results().release() ) );
EOT
        ) ;

        # Update out and inout arguments.

        foreach my $argument ( @$arguments ) {
            my $kind = $$argument{ KIND } ;
            next if $kind eq 'in' ;
            my $name = $$argument{ NAME } ;
            $handle->print( "    $name = _packet->$name;\n" ) ;
        }
    }

    # Return.

    if ( defined( $method->result_type() ) ) {
        $handle->print( "    return _packet->_result;\n" ) ;
    }

    $handle->print( "    }\n" ) ;
}


###
# _write_arguments
sub _write_arguments {
    my( $handle, $arguments, $write_mask, $skip_kind ) = @_ ;

    $handle->print( '(' ) ;

    my $argument_index = 0 ;
    foreach my $argument ( @$arguments ) {
        my $arg_kind = $$argument{ KIND } ;
        next if ( defined( $skip_kind ) ) && ( $arg_kind eq $skip_kind ) ;

        $handle->print( ',' ) if $argument_index > 0 ;

        if ( $write_mask =~ /t/ ) {
            my $arg_type = $$argument{ TYPE } ;
            $handle->print( ' ' ) ;
            if ( $arg_kind eq 'in' ) {
                $handle->print( $arg_type->in_type() ) ;
            }
            else {
                $handle->print( $arg_type->out_type() ) ;
            }
        }

        if ( $write_mask =~ /n/ ) {
            $handle->print( ' ' . $$argument{ NAME } ) ;
        }

        $argument_index++ ;
    }

    if ( $argument_index > 0 ) {
        $handle->print( ' ' ) ;
    }
    $handle->print( ')' ) ;
}


1 ;


__END__
