=head1 NAME

SourceGen - factory/metaobject/RPC generator

=head1 SYNOPSIS

    use SourceGen ;

    # Parent class is IdlParser - see IdlParser.pm

    # Constructor
    $generator = SourceGen->new() ;

    # Parse IDL and create all helper data structures.
    $generator->parse( $file_name ) ;

    # Generate sources.
    $generator->write_generated_header() ;
    $generator->write_rpc_header() ;
    $generator->write_generated_source() ;

=head1 DESCRIPTION

C<SourceGen> extends C<IdlParser>. It performs additional checks while parsing
IDL and is able to generate factories, metaobjects and RPC-related sources.

=head1 KNOWN BUGS

None.

=cut

package SourceGen ;

use strict ;
use warnings ;

use FileHandle ;

use IdlParser ;
our( @ISA ) ;
@ISA = qw( IdlParser ) ;

use ClassInfo ;
use FactoryGen ;
use MetaobjectGen ;
use FileNames ;
use Utils ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Parent constructor.
    
    my $self = $class->SUPER::new() ;

    # File name of primary IDL.
    $$self{ IDL_NAME } = undef ;

    # Array of C++ header file names with definition of the classes.
    $$self{ USER_HEADER_NAMES } = [ ] ;

    # Other file names.
    $$self{ GENERATED_SOURCE_NAME } = undef ;
    $$self{ GENERATED_HEADER_NAME } = undef ;
    $$self{ GENERATED_HEADER_GUARD } = undef ;
    $$self{ RPC_HEADER_NAME } = undef ;
    $$self{ RPC_HEADER_GUARD } = undef ;

    # Class descriptions for all primary classes.
    # Keys are fully qualified class names, values are class infos.
    $$self{ CLASS_INFOS } = { } ;

    # Class names sorted in order which should be used when writing headers.
    $$self{ CLASS_ORDER } = [ ] ;

    return $self ;
}


###
# parse
sub parse {
    my $self = shift ;
    my $file_name = shift || die "No file name specified.\n" ;

    $self->_set_file_name( $file_name ) ;

    # Parse IDL sources.

    $self->SUPER::parse() ;

    # Generate class infos for "primary" classes.

    my $all_classes = $self->classes() ;
    my $class_infos = $$self{ CLASS_INFOS } ;
    foreach my $class_name ( keys( %$all_classes ) ) {
        my $class_symbol = $$all_classes{ $class_name } ;
        next if ( $class_symbol->file_name() ne $file_name ) ;

        my $class_info = ClassInfo->new( $class_symbol ) ;
        $$class_infos{ $class_name } = $class_info ;
    }

    # Create array of C++ header file names.

    $self->_gather_header_names() ;

    # Create class info order array.

    $self->_sort_classes() ;
}


###
# write_generated_header
sub write_generated_header {
    my $self = shift ;

    my $idl_name = $$self{ IDL_NAME } ;
    my $user_header_names = $$self{ USER_HEADER_NAMES } ;
    my $header_name = $$self{ GENERATED_HEADER_NAME } ;
    my $header_guard = $$self{ GENERATED_HEADER_GUARD } ;

    my $class_infos = $$self{ CLASS_INFOS } ;
    my $class_order = $$self{ CLASS_ORDER } ;

    my $generated_namespace = 'Massiv::Generated' ;

    # Create file.

    my $handle = FileHandle->new( $header_name, 'w' ) or
        die( "Can't create '$header_name'.\n" ) ;

    # Write intro.

    $handle->print( <<EOT
/** \@file $header_name
 * Metaobjects, object factories and other internal generated stuff.
 * "Script" sources should not include this file.
 * Automatically generated from $idl_name. Do not modify.
 */

#ifndef $header_guard
#define $header_guard


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

EOT
    ) ;

    foreach my $header_name ( @$user_header_names ) {
        $handle->print( "#include \"${header_name}\"\n" ) ;
    }

    $handle->print( "\n" ) ;
    Utils::write_open_namespace( $handle, $generated_namespace ) ;

    # Object forward declarations.

    $handle->print( <<EOT


// Forward declarations.
EOT
    ) ;

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        my $metaobject_name = $class_info->metaobject_name() ;
        my $factory_name = $class_info->factory_name() ;
        
        $handle->print( <<EOT

template <>
class $factory_name;

template <>
class $metaobject_name;
EOT
        ) ;
    }

    # Write factory class declarations.

    $handle->print( "\n\n// Factories.\n" ) ;

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        FactoryGen::write_class( $handle, $class_info ) ;
    }

    # Write metaobject class declarations.

    $handle->print( "\n\n// Metaobjects.\n" ) ;

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        MetaobjectGen::write_class( $handle, $class_info, $self ) ;
    }

    $handle->print( "\n\n" ) ;
    Utils::write_close_namespace( $handle, $generated_namespace ) ;

    # Write outro.

    $handle->print( <<EOT


#endif // $header_guard
EOT
    ) ;

    # Close file.

    $handle->close() ;
}


###
# write_rpc_header
sub write_rpc_header {
    my $self = shift ;

    my $idl_name = $$self{ IDL_NAME } ;
    my $user_header_names = $$self{ USER_HEADER_NAMES } ;
    my $header_name = $$self{ RPC_HEADER_NAME } ;
    my $header_guard = $$self{ RPC_HEADER_GUARD } ;

    my $class_infos = $$self{ CLASS_INFOS } ;
    my $class_order = $$self{ CLASS_ORDER } ;

    my $generated_namespace = 'Massiv::Generated' ;

    # Set of parent RPC headers.

    my %parent_headers ;
    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        my $class_symbol = $class_info->class_symbol( 0 ) ;
        my $parents = $class_symbol->parents() ;
        foreach my $parent ( @$parents ) {
            my $parent_symbol = $$parent{ SYMBOL } ;
            my $file_name = $parent_symbol->file_name() ;
            my $rpc_file_name = FileNames::rpc_header( FileNames::base( $file_name ) ) ;
            $parent_headers{ $rpc_file_name }++ ;
        }
    }

    # Create file.

    my $handle = FileHandle->new( $header_name, 'w' ) or
        die( "Can't create '$header_name'.\n" ) ;

    # Write intro.

    $handle->print( <<EOT
/** \@file $header_name
 * Public RPC-related classes.
 * Automatically generated from $idl_name. Do not modify.
 */

#ifndef $header_guard
#define $header_guard


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

EOT
    ) ;

    foreach my $header_name ( @$user_header_names ) {
        $handle->print( "#include \"${header_name}\"\n" ) ;
    }

    # Include parent RPC headers.

    if ( scalar( %parent_headers ) ) {
        $handle->print( "\n// Definitions of parent RPCStubs.\n\n" ) ;
        foreach my $header ( sort( keys( %parent_headers ) ) ) {
            $handle->print( "#include \"${header}\"\n" ) ;
        }
    }

    # Write RPC packet classes.

    $handle->print( "\n\n" ) ;
    Utils::write_open_namespace( $handle, $generated_namespace ) ;

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        $handle->print( "\n\n// RPC classes for $class_name.\n" ) ;
        RPCGen::write_tag_class( $handle, $class_info ) ;
        RPCGen::write_packet_classes( $handle, $class_info ) ;
        RPCGen::write_stub_class( $handle, $class_info ) ;
    }

    $handle->print( "\n\n" ) ;
    Utils::write_close_namespace( $handle, $generated_namespace ) ;

    # Done.

    $handle->print( <<EOT


#endif // $header_guard
EOT
    ) ;

    # Close file.

    $handle->close() ;
}


###
# write_generated_source
sub write_generated_source {
    my $self = shift ;

    my $idl_name = $$self{ IDL_NAME } ;
    my $source_name = $$self{ GENERATED_SOURCE_NAME } ;
    my $header_name = $$self{ GENERATED_HEADER_NAME } ;
    my $rpc_header_name = $$self{ RPC_HEADER_NAME } ;

    my $class_infos = $$self{ CLASS_INFOS } ;
    my $class_order = $$self{ CLASS_ORDER } ;

    my $generated_namespace = 'Massiv::Generated' ;

    # Create file.

    my $handle = FileHandle->new( $source_name, 'w' ) or
        die( "Can't create '$source_name'.\n" ) ;

    # Write intro.

    $handle->print( <<EOT
/** \@file $source_name
 * Imlementation of metaobjects, object factories and other stuff generated
 * from '${idl_name}'. Do not modify.
 */

#include "core.h"
#include "${header_name}"
#include "${rpc_header_name}"
#include "global/global.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "serializer/serializer.h"
#include "net/bitostream.h"
#include "net/bitistream.h"
#include "object/object_manager.h"
#include "balance/balancer.h"


EOT
    ) ;

    Utils::write_open_namespace( $handle, $generated_namespace ) ;

    # Global stuff.

    $handle->print( <<EOT


// Macro used by metaobjects.
// Can't use offsetof from cstddef because managed classes are not PODs.
#define PROPERTY_OFFSET( name ) \\
    ( reinterpret_cast< const ::Massiv::Core::UInt8 * >( static_cast< const ::Massiv::Core::Property * >( &p_0->name ) ) - \\
      reinterpret_cast< const ::Massiv::Core::UInt8 * >( p_0 ) )
EOT
    ) ;

    # Write factories and metaobject.

    $handle->print( "\n\n// Implementation of object factories and metaobjects.\n" ) ;

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;

        $handle->print( "\n\n// Object factory for $class_name\n" ) ;
        FactoryGen::write_source( $handle, $class_info, $self ) ;
    }

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;

        $handle->print( "\n\n// Metaobject for $class_name\n" ) ;
        MetaobjectGen::write_source( $handle, $class_info, $self ) ;
    }

    # Write implementation of RPC-related stuff.

    foreach my $class_name ( @$class_order ) {
        my $class_info = $$class_infos{ $class_name } ;
        $handle->print( "\n\n// Implementation of RPC classes for $class_name.\n" ) ;
        RPCGen::write_packet_source( $handle, $class_info ) ;
        RPCGen::write_stub_source( $handle, $class_info ) ;
    }

    $handle->print( "\n\n" ) ;
    Utils::write_close_namespace( $handle, $generated_namespace ) ;

    # Close file.

    $handle->close() ;
}


###
# _set_file_name
# Set name of IDL file to parse and determine names of files to generate.
#   file_name
sub _set_file_name {
    my $self = shift ;
    my $file_name = shift ;

    if ( $file_name =~ /\\/ ) {
        die( <<EOT
File name '$file_name' contains backslashes!
Please use only slashes when passing file names to factgen scripts.
This is probably error in makefile.gen - your slashes are not escaped and
are changed into backslashes by mkgen.pl when generating makefiles for
Microsoft(TM) Windows(TM) platform.
EOT
        ) ;
    }

    $$self{ IDL_NAME } = $file_name ;
    $self->push_file( $file_name ) ;

    my $base_name = FileNames::base( $file_name ) ;

    $$self{ GENERATED_SOURCE_NAME } = FileNames::generated_source( $base_name ) ;

    my $generated_header_name = FileNames::generated_header( $base_name ) ;
    my $generated_header_guard = 'GENERATED_' . uc( $generated_header_name ) ;
    $generated_header_guard =~ s|[./]|_|g ;
    $$self{ GENERATED_HEADER_NAME } = $generated_header_name ;
    $$self{ GENERATED_HEADER_GUARD } = $generated_header_guard ;
    
    my $rpc_header_name = FileNames::rpc_header( $base_name ) ;
    my $rpc_header_guard = 'GENERATED_' . uc( $rpc_header_name ) ;
    $rpc_header_guard =~ s|[./]|_|g ;
    $$self{ RPC_HEADER_NAME } = $rpc_header_name ;
    $$self{ RPC_HEADER_GUARD } = $rpc_header_guard ;
}


###
# _gather_header_names
# Create list of C++ header file names with definitions of the required classes.
sub _gather_header_names {
    my $self = shift ;
    my @header_names ;
    my %header_name_set ;

    # File headers. Order from IDL file is preserved.

    my $idl_info = $$self{ IDL_INFOS }{ $$self{ IDL_NAME } } ;
    if ( defined( $idl_info ) ) {
        foreach my $header_name ( @{$$idl_info{ HEADERS }} ) {
            unless ( defined( $header_name_set{ $header_name } ) ) {
                $header_name_set{ $header_name }++ ;
                push( @header_names, $header_name ) ;
            }
        }
    }
    
    # Gather class headers. Warning: in random order.

    my $class_infos = $$self{ CLASS_INFOS } ;
    foreach my $class_name ( keys( %$class_infos ) ) {
        my $class_symbol = $$class_infos{ $class_name }->class_symbol( 0 ) ;
        my $header_name = $class_symbol->header_name() ;
        unless ( defined( $header_name_set{ $header_name } ) ) {
            $header_name_set{ $header_name }++ ;
            push( @header_names, $header_name ) ;
        }
    }

    $$self{ USER_HEADER_NAMES } = \@header_names ;
}


###
# _sort_classes
# Topological sort in O(n^2), please don't kill me.
sub _sort_classes {
    my $self = shift ;

    # Gather set of class names.

    my $class_infos = $$self{ CLASS_INFOS } ;
    my %class_names ;
    foreach my $class_name ( keys( %$class_infos ) ) {
        $class_names{ $class_name }++ ;
    }

    # Sort them.

    my @sorted_class_names ;
    while ( %class_names ) {
        my $picked_class ;

        # Try each class in class_names.

        CLASS_NAME : foreach my $class_name ( keys( %class_names ) ) {
            # Check if all class_name's parents are outside class_names set.

            my $class_info = $$class_infos{ $class_name } ;
            my $parents = $class_info->parents() ;
            foreach my $parent ( @$parents ) {
                my $parent_name = $parent->full_name() ;
                if ( defined( $class_names{ $parent_name } ) ) {
                    # class_name has parent in class_names - skip it.

                    next CLASS_NAME ;
                }
            }

            # All parents are outside class_names - add class_name to the list.

            $picked_class = $class_name ;
            last CLASS_NAME ;
        }

        if ( defined( $picked_class ) ) {
            delete( $class_names{ $picked_class } ) ;
            push( @sorted_class_names, $picked_class ) ;
        }
        else {
            die( "Cyclic inheritance in " . join( ", ", keys( %class_names ) ) . "\n" ) ;
        }
    }

    $$self{ CLASS_ORDER } = \@sorted_class_names ;
}


1 ;


__END__
