=head1 NAME

SymbolAttribute - attribute type symbol classes

=head1 SYNOPSIS

    use SymbolAttribute ;

    ###
    # SymbolAttribute

    # Parent class is SymbolBase

    # Get C++ attribute name.
    # Initial '::' will be removed and other '::' will be replaced by '_'.
    $cpp_name = $attribute->cpp_name() ;

    # Get/set reference to set of attribute flags.
    # Note: set version will replace the hash (set).
    # Attribute flags is simple set of strings - all attributes are boolean.
    # The following flags are supported:
    # - idl_internal (all kinds): This attribute is not exported to generated sources.
    # - inherit (class kind): Inherit value of this attribute from parents if it's
    #                         not explicitly specified in attribute list.
    $flags = $attribute->attribute_flags() ;
    $attribute->attribute_flags( $flags ) ;

    # Get/set attribute type.
    # Attribute type can be either builtin or user-defined types. For builtin
    # types C<attribute_type()> will be string describing the type, for
    # user-defined types it will be reference to type symbol.
    # Supported builtin types are 'bool', 'int' and 'string'.
    # Supported user-defined types are enums (SymbolEnum).
    $type = $attribute->attribute_type() ;
    $attribute->attribute_type( $type ) ;

    # Get C++ attribute type.
    $cpp_type = $attribute->cpp_type() ;

    # Get/set attribute default value.
    # The value is "perlish" scalar. Boolean (scalar that evaluates to true
    # or false), integer or string according to attribute type.
    $default = $attribute->default_value() ;
    $attribute->default_value( $default ) ;

    # Convert specified perlish value to C++ value.
    $cpp_value = $attribute->cpp_value( $value ) ;

=cut

{
package SymbolAttribute ;

our( @ISA ) ;
@ISA = qw( SymbolBase ) ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = $class->SUPER::new( @_ ) ;

    # Add data.

    $$self{ ATTRIBUTE_FLAGS } = { } ;
    $$self{ ATTRIBUTE_TYPE } = undef ;
    $$self{ DEFAULT_VALUE } = undef ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'attribute' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    
    $self->SUPER::dump( $indent ) ;

    my $type = $$self{ ATTRIBUTE_TYPE } ;
    if ( ref( $type ) ) {
        $type = $type->full_name() ;
    }

    $indent += 2 ;
    print ' ' x $indent . "attribute_type = $type\n" ;
    print ' ' x $indent . "default_value = $$self{ DEFAULT_VALUE }\n" ;
}


###
# cpp_name
sub cpp_name {
    my $self = shift ;
    my $name = substr( $self->full_name(), 2 ) ;
    $name =~ s/::/_/g ;
    return $name ;
}


###
# attribute_flags
sub attribute_flags {
    my $self = shift ;
    if ( @_ ) { $$self{ ATTRIBUTE_FLAGS } = shift ; }
    return $$self{ ATTRIBUTE_FLAGS } ;
}


###
# attribute_type
sub attribute_type {
    my $self = shift ;
    if ( @_ ) { $$self{ ATTRIBUTE_TYPE } = shift ; }
    return $$self{ ATTRIBUTE_TYPE } ;
}


###
# cpp_type
sub cpp_type {
    my $self = shift ;
    my $type = $$self{ ATTRIBUTE_TYPE } ;

    # Currently all references are to enum (->int in C++)

    if ( ref( $type ) ) {
        return 'int' ;
    }

    # String is std::string.

    if ( $type eq 'string' ) {
        return 'std::string' ;
    }

    # Other builtin types match C++ types.

    return $type ;
}


###
# default_value
sub default_value {
    my $self = shift ;
    if ( @_ ) { $$self{ DEFAULT_VALUE } = shift ; }
    return $$self{ DEFAULT_VALUE } ;
}


###
# cpp_value
sub cpp_value {
    my $self = shift ;
    my $value = shift ;
    
    my $type = $$self{ ATTRIBUTE_TYPE } ;
    if ( $type eq 'bool' ) {
        if ( $value ) { return 'true' ; }
        return 'false' ;
    }

    # Enclose string in double-quotes. It will _not_ be escaped.

    if ( $type eq 'string' ) {
        return '"' . $value . '"' ;
    }
    
    # Otherwise type is int or enum (int in C++) - no conversion needed.
    
    return $value ;
}


} # SymbolAttribute


=pod

    ###
    # SymbolClassAttribute

    # parent is SymbolAttribute

=cut

{
package SymbolClassAttribute ;

our( @ISA ) ;
@ISA = qw( SymbolAttribute ) ;


###
# type_name
sub type_name {
    'classattr' ;
}


} # package SymbolClassAttribute


=pod

    ###
    # SymbolPropertyAttribute

    # parent is SymbolAttribute

=cut

{
package SymbolPropertyAttribute ;

our( @ISA ) ;
@ISA = qw( SymbolAttribute ) ;


###
# type_name
sub type_name {
    'propattr' ;
}


} # package SymbolPropertyAttribute


=pod

    ###
    # SymbolMethodAttribute

    # parent is SymbolAttribute

=cut

{
package SymbolMethodAttribute ;

our( @ISA ) ;
@ISA = qw( SymbolAttribute ) ;


###
# type_name
sub type_name {
    'methodattr' ;
}


} # package SymbolMethodAttribute


1 ;


__END__


=pod

=head1 DESCRIPTION

Package F<SymbolAttribute.pm> defines attribute type symbol classes.
C<SymbolAttribute> is generic attribute type class, which keeps information
about attribute type (reference to enum symbol) and default attribute value.
Three classes inherit C<SymbolAttribute>: C<SymbolClassAttribute>,
C<SymbolPropertyAttribute> and C<SymbolMethodAttribute>. These classes do
not define any new methods or data - they are only different types to
distinguish different attribute kinds.

A bit of terminology: attribute B<kind> is one of C<classattr>, C<propattr>,
C<methodattr>. Attribute B<type> is user-specified type of specific attribute,
it may be builtin type (boolean, integer, string) or user-defined enumeration type.

When parsing value of string attribute, escape sequences will _not_ be interpolated,
and cpp_value() of string attribute simply encloses it in double quotes (weird
characters won't be escaped).

=head1 KNOWN BUGS

None.

=cut
