=head1 NAME

SymbolBase - base symbol class and several small symbol classes

=head1 SYNOPSIS

    use SymbolBase ;

    ###
    # SymbolBase

    # Constructor.
    $symbol = SymbolBase->new( $name [, $parent ] ) ;

    # Get "nice" type name. Each symbol class should redefine this.
    $type_name = SymbolBase->type_name() ;

    # Get/set parent symbol. Should be a SymbolTable.
    $parent_symbol = $symbol->parent_symbol() ;
    $symbol->parent_symbol( $parent_symbol ) ;
    
    # Get/set symbol name.
    $name = $symbol->name() ;
    $symbol->name( $name ) ;

    # Get name of file and number of line this symbol has been defined on.
    $file_name = $symbol->file_name() ;
    $line_number = $symbol->line_number() ;

    # Get fully qualified smbol name - name including names of all parent
    # symbol tables, separated by "::".
    # Note that because there is root namespace with empty name, the fully
    # qualified name WILL be prefixed by "::".
    $full_name = $symbol->full_name() ;

    # Set file name and line.
    $symbol->set_context( $file_name, $line_number ) ;

    # Set file name and line from current parser context.
    $symbol->set_context_from_parser( $parser ) ;

    # Dump symbol info.
    $symbol->dump( $indent ) ;

=cut

package SymbolBase ;

use warnings ;
use strict ;

use SymbolTable ;
use Parser ;


###
# new
sub new {
    my( $proto, $name, $parent ) = @_ ;
    my $class = ref( $proto ) || $proto ;

    unless ( defined( $name ) ) {
        use Carp ;
        croak( "No name specified, stopped" ) ;
    }
    
    # Allocate the object.

    my $self = {
        PARENT_SYMBOL => $parent,
        NAME => $name,
        FILE_NAME => undef,
        LINE_NUMBER => undef,
        # FULL_NAME cached if requested
    } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'base' ;
}


###
# parent_symbol
sub parent_symbol {
    my $self = shift ;
    if ( @_ ) { $$self{ PARENT_SYMBOL } = shift ; }
    return $$self{ PARENT_SYMBOL } ;
}


###
# name
sub name {
    my $self = shift ;
    if ( @_ ) { $$self{ NAME } = shift ; }
    return $$self{ NAME } ;
}


###
# file_name
sub file_name {
    return $_[ 0 ]{ FILE_NAME } ;
}


###
# line_number
sub line_number {
    return $_[ 0 ]{ LINE_NUMBER } ;
}


###
# full_name
sub full_name {
    my $self = shift ;

    unless ( defined( $$self{ FULL_NAME } ) ) {
        my $parent = $$self{ PARENT_SYMBOL } ;
        my $full_name = $$self{ NAME } ;
        while ( defined( $parent ) ) {
            $full_name = $$parent{ NAME } . '::' . $full_name ;
            $parent = $$parent{ PARENT_SYMBOL } ;
        }
        $$self{ FULL_NAME } = $full_name ;
    }

    return $$self{ FULL_NAME } ;
}


###
# set_context
sub set_context {
    my $self = shift ;
    $$self{ FILE_NAME } = shift ;
    $$self{ LINE_NUMBER } = shift ;
}


###
# set_context_from_parser
sub set_context_from_parser {
    my $self = shift ;
    my $parser = shift ;
    $self->set_context( $parser->file_name(), $parser->line_number() ) ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    print( ' ' x $indent . $self->type_name . ' ' . $self->name . "\n" ) ;
}


=pod

    ###
    # SymbolNamespace

    # parent is SymbolTable

=cut

{
package SymbolNamespace ;

our( @ISA ) ;
@ISA = qw( SymbolTable ) ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = $class->SUPER::new( @_ ) ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'namespace' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    $self->SUPER::dump( $indent ) ;
}


} # package SymbolNamespace


=pod

    ###
    # SymbolEnum

    # parent is SymbolTable, only SymbolConstInt are allowed as children symbols

=cut

{
package SymbolEnum ;

our( @ISA ) ;
@ISA = qw( SymbolTable ) ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;

    # Allocate the object.

    my $self = $class->SUPER::new( @_ ) ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'enum' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    $self->SUPER::dump( $indent ) ;
}


} # package SymbolEnum


1 ;


__END__


=pod

=head1 DESCRIPTION

Package F<SymbolBase.pm> defines multiple symbol types:

=over

=item SymbolBase

Generic symbol. It has a name and parent symbol, which will be undefined
for root symbols.

=item SymbolNamespace

Namespace symbol, symbol subtable. Name of root symbol may be undefined (and
will be left undefined by C<IdlParser>.)

=item SymbolEnum

Enumeration type. It's a symbol table, but only C<SymbolConstInt> symbols
are allowed.

=item Symbol classes defined in other modules

More complex symbol classes are defined in other modules. These include:

=over

FIXME: out of date

=item SymbolTable

Defined in F<SymbolTable.pm>

=item SymbolClass

Defined in F<SymbolClass.pm>

=item SymbolConstant and other constant value symbol types

Defined in F<SymbolConstant.pm>

=back

=back

=head1 TODO

Nicer dump.

=head1 KNOWN BUGS

None.

=cut
