=head1 NAME

SymbolClass - symbol class that represents a class

=head1 SYNOPSIS

    use SymbolClass ;

    ###
    # SymbolClass

    # Parent class is SymbolTable.

    # Constructor - see SymbolTable.

    # Is class already defined?
    $is_defined = $class->defined() ;
    $class->defined( 1 ) ;

    # Get reference to array with information about direct parent classes:
    #   SYMBOL - reference to parent symbol class; always defined
    #   VIRTUAL - true if class is inherited virtually
    $parents = $class->parents() ;
    
    # Add information about parent class.
    #   $parent - reference to parent class symbol
    #   $virtual - is parent class inherited virtually?
    $class->add_parent( $parent, $virtual ) ;

    # Get/set reference to attribute values.
    # Gets/sets reference to hash, where keys of this hash are fully qualified
    # attribute name and values are attribute values.
    # (hash will not be copied, only reference set)
    $attributes = $class->attributes() ;
    $class->attributes( $attributes ) ;

    # Get value of given attribute.
    # Dies with error if specified attribute is not defined for this class.
    # Note that attributes defined after the class (when parsing the sources)
    # will not be defined for the class, not even to default values.
    $value = $class->require_attribute( $fully_qulified_name ) ;

    # Get reference to array of all property symbols defined in the class,
    # not including inherited properties (order from IDL is preserved).
    $properties = $class->properties() ;

    # Get reference to array of all method symbols (order from IDL is preserved).
    $methods = $class->methods() ;

    # Find first property containing strong pointer.
    # Return reference to property symbol defined in this or parent class if it
    # contains strong pointers, or undef otherwise.
    $property = $class->find_strong_pointer() ;

    # Try to find symbol with given name in class and its parents.
    $symbol = $class->find_class_symbol( $name ) ;

    # Get name of file with C++ definition of this class.
    $header_name = $class->header_name() ;

=head1 DESCRIPTION

This module defines C<SymbolClass> class. Its parent is C<SymbolTable>.
It represents a class. Subsymbols can be subclasses, properties and methods.
In addition to these symbols, C<SymbolClass> keeps information about parents
and class attributes.

See F<SymbolBase.pm> for more info about various symbol types.

=head1 KNOWN BUGS

None.

=cut

package SymbolClass ;

use strict ;
use warnings ;

use SymbolTable ;
our( @ISA ) ;
@ISA = qw( SymbolTable ) ;

use FileNames ;


###
# Constructor
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Allocate the object.
    
    my $self = $class->SUPER::new( @_ ) ;

    # Add data.

    # Is class defined or only declared?
    $$self{ DEFINED } = 0 ;

    # Hash of attribute-value pairs.
    $$self{ ATTRIBUTES } = { } ;

    # Array of structures containing info about parents:
    #   SYMBOL - reference to parent class symbol
    #   VIRTUAL - is parent virtual?
    $$self{ PARENTS } = [ ] ;

    # Array of references to properties and methods in IDL order.
    $$self{ PROPERTIES } = [ ] ;
    $$self{ METHODS } = [ ] ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'class' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    
    $self->SUPER::dump( $indent ) ;

    $indent += 2 ;
    foreach my $parent_info ( @{$$self{ PARENTS }} ) {
        my $parent_symbol = $$parent_info{ SYMBOL } ;
        my $parent_name = $parent_symbol->full_name() ;
        print ' ' x $indent . "parent $parent_name" .
            ( $$parent_info{ VIRTUAL } ? " (virtual)\n" : "\n" ) ;
    }
    foreach my $attr_name ( keys( %{$$self{ ATTRIBUTES }} ) ) {
        print ' ' x $indent . "attribute " . $attr_name .
            " = " . $$self{ ATTRIBUTES }{ $attr_name } . "\n" ;
    }
}


###
# defined
sub defined {
    my $self = shift ;
    if ( @_ ) { $$self{ DEFINED } = shift ; }
    return $$self{ DEFINED } ;
}


###
# parents
sub parents {
    return $_[ 0 ]{ PARENTS } ;
}


###
# add_parent
sub add_parent {
    my( $self, $parent, $virtual ) = @_ ;
    push( @{$$self{ PARENTS }}, {
        SYMBOL => $parent,
        VIRTUAL => $virtual,
    } ) ;
}


###
# attributes
sub attributes {
    my $self = shift ;
    if ( @_ ) { $$self{ ATTRIBUTES } = shift ; }
    return $$self{ ATTRIBUTES } ;
}


###
# require_attribute
sub require_attribute {
    my $self = shift ;
    my $name = shift ;
    unless ( defined( $$self{ ATTRIBUTES }{ $name } ) ) {
        my $cname = $self->full_name() ;
        die "No attribute '$name' defined for class '$cname'.\n" ;
    }
    return $$self{ ATTRIBUTES }{ $name } ;
}


###
# properties
sub properties {
    return $_[ 0 ]{ PROPERTIES } ;
}


###
# methods
sub methods {
    return $_[ 0 ]{ METHODS } ;
}


###
# find_strong_pointer
sub find_strong_pointer {
    my $self = shift ;

    # Try self.

    my $properties = $self->properties() ;

    # Check all properties for strong pointer.

    foreach my $property ( @$properties ) {
        if( $property->property_type()->has_strong_pointer() ) {
            return $property ;
        }
    }

    # Try parents.

    my $parents = $$self{ PARENTS } ;
    foreach my $parent ( @$parents ) {
        my $parent_symbol = $$parent{ SYMBOL } ;
        my $strong_symbol = $parent_symbol->find_strong_pointer() ;
        if ( defined( $strong_symbol ) ) {
            return $strong_symbol ;
        }
    }

    # No strong pointer property found.

    return undef ;
}


###
# find_class_symbol
sub find_class_symbol {
    my $self = shift ;
    my $name = shift ;

    # Try self.
    
    my $symbol = $self->get_symbol( $name ) ;
    return $symbol if defined( $symbol ) ;

    # Try parents.

    my $parents = $$self{ PARENTS } ;
    foreach my $parent ( @$parents ) {
        my $parent_symbol = $$parent{ SYMBOL } ;
        $symbol = $parent_symbol->find_class_symbol( $name ) ;
        last if defined( $symbol ) ;
    }

    return $symbol ;
}


###
# header_name
sub header_name {
    # FIXME: Allow overriding of this value.

    my $self = shift ;
    my $idl_name = $self->file_name() ;
    return FileNames::header( FileNames::base( $idl_name ) ) ;
}


1 ;


__END__
