=head1 NAME

SymbolMethod - symbol class that represents a class method

=head1 SYNOPSIS

    use SymbolMethod ;

    ###
    # SymbolMethod

    # Parent class is SymbolBase.

    # Constructor - see SymbolBase.

    # Get/set reference to attribute values.
    # Gets/sets reference to hash, where keys of this hash are fully qualified
    # attribute names and values are attribute values.
    # (hash will not be copied, only reference set)
    $attributes = $method->attributes() ;
    $method->attributes( $attributes ) ;

    # Get value of given attribute.
    # Dies with error if specified attribute is not defined for this method.
    # Note that attributes defined after the method (when parsing the sources)
    # will not be defined for the method, not even to default values.
    $value = $method->require_attribute( $fully_qulified_name ) ;

    # Get/set method index.
    # All class methods are indexed from 0 based on order they appear in
    # class definition.
    $index = $method->method_index() ;
    $method->method_index( $index ) ;

    # Get reference to array of method arguments.
    # Each argument is a structure:
    #   KIND - 'in', 'out', or 'inout'
    #   TYPE - argument type (reference to MassivType object)
    #   NAME - argument name
    $arguments = $method->arguments() ;

    # Does method have non-zero number of in or inout arguments?
    $has_in_arguments = $method->has_in_arguments() ;

    # Does method have non-zero number of out or inout arguments or return value?
    $has_out_arguments = $method->has_out_arguments() ;

    # Add method argument.
    $method->add_argument( $kind, $type, $name ) ;

    # Get/set result type. Undef if method is "void".
    $result_type = $method->result_type() ;
    $method->result_type( $result_type ) ;

    # Get simple C++ result type.
    $cpp_result_type = $method->cpp_result_type() ;

=head1 DESCRIPTION

This module defines C<SymbolMethod> class.
It represents a class method - list of property attributes and arguments.

See F<SymbolBase.pm> for more info about various symbol types.

=head1 KNOWN BUGS

None.

=cut

package SymbolMethod ;

use strict ;
use warnings ;

use SymbolBase ;
our( @ISA ) ;
@ISA = qw( SymbolBase ) ;


###
# Constructor
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Allocate the object.
    
    my $self = $class->SUPER::new( @_ ) ;

    # Add data.

    $$self{ ATTRIBUTES } = { } ;
    $$self{ METHOD_INDEX } = undef ;
    $$self{ ARGUMENTS } = [ ] ;
    $$self{ RESULT_TYPE } = undef ;

    # DOne.

    return $self ;
}


###
# type_name
sub type_name {
    'method' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    
    $self->SUPER::dump( $indent ) ;

    $indent += 2 ;

    print ' ' x $indent . "index " . $$self{ METHOD_INDEX } . "\n" ;

    foreach my $attr_name ( keys( %{$$self{ ATTRIBUTES }} ) ) {
        print ' ' x $indent . "attribute " . $attr_name .
            " = " . $$self{ ATTRIBUTES }{ $attr_name } . "\n" ;
    }

    foreach my $argument ( @{$$self{ ARGUMENTS }} ) {
        print ' ' x $indent . "argument " . $$argument{ KIND } . ' ' .
            $$argument{ TYPE }->idl_type() . ' ' . $$argument{ NAME } . "\n" ;
    }

    if ( defined( $$self{ RESULT_TYPE } ) ) {
        print ' ' x $indent . "result " . $$self{ RESULT_TYPE }->idl_type() . "\n" ;
    }
}


###
# attributes
sub attributes {
    my $self = shift ;
    if ( @_ ) { $$self{ ATTRIBUTES } = shift ; }
    return $$self{ ATTRIBUTES } ;
}


###
# require_attribute
sub require_attribute {
    my $self = shift ;
    my $name = shift ;
    unless ( defined( $$self{ ATTRIBUTES }{ $name } ) ) {
        my $mname = $self->full_name() ;
        die "No attribute '$name' defined for method '$mname'.\n" ;
    }
    return $$self{ ATTRIBUTES }{ $name } ;
}


###
# method_index
sub method_index {
    my $self = shift ;
    if ( @_ ) { $$self{ METHOD_INDEX } = shift ; }
    return $$self{ METHOD_INDEX } ;
}


###
# arguments
sub arguments {
    return $_[ 0 ]{ ARGUMENTS } ;
}


###
# has_in_arguments
sub has_in_arguments {
    my $self = shift ;

    my $arguments = $$self{ ARGUMENTS } ;
    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        if ( ( $kind eq 'in' ) || ( $kind eq 'inout' ) ) {
            return 1 ;
        }
    }

    return 0 ;
}


###
# has_out_arguments
sub has_out_arguments {
    my $self = shift ;

    if ( defined( $$self{ RESULT_TYPE } ) ) {
        return 1 ;
    }

    my $arguments = $$self{ ARGUMENTS } ;
    foreach my $argument ( @$arguments ) {
        my $kind = $$argument{ KIND } ;
        if ( ( $kind eq 'out' ) || ( $kind eq 'inout' ) ) {
            return 1 ;
        }
    }

    return 0 ;
}


###
# add_argument
sub add_argument {
    my( $self, $kind, $type, $name ) = @_ ;
    push( @{$$self{ ARGUMENTS }}, {
        KIND => $kind,
        TYPE => $type,
        NAME => $name,
    } ) ;
}


###
# result_type
sub result_type {
    my $self = shift ;
    if ( @_ ) { $$self{ RESULT_TYPE } = shift ; }
    return $$self{ RESULT_TYPE } ;
}


###
# cpp_result_type
sub cpp_result_type {
    my $self = shift ;
    my $result_type = $$self{ RESULT_TYPE } ;
    if ( defined( $result_type ) ) {
        return $result_type->simple_type() ;
    }
    return 'void' ;
}


1 ;


__END__
