=head1 NAME

SymbolProperty - symbol class that represents a class property

=head1 SYNOPSIS

    use SymbolProperty ;

    ###
    # SymbolProperty

    # Parent class is SymbolBase.

    # Constructor - see SymbolBase.

    # Get/set reference to attribute values.
    # Gets/sets reference to hash, where keys of this hash are fully qualified
    # attribute names and values are attribute values.
    # (hash will not be copied, only reference set)
    $attributes = $property->attributes() ;
    $property->attributes( $attributes ) ;

    # Get value of given attribute.
    # Dies with error if specified attribute is not defined for this property.
    # Note that attributes defined after the property (when parsing the sources)
    # will not be defined for the property, not even to default values.
    $value = $property->require_attribute( $fully_qulified_name ) ;

    # Get/set property type.
    $type = $property->property_type() ;
    $property->property_type( $type ) ;

=head1 DESCRIPTION

This module defines C<SymbolProperty> class.
It represents a class property - list of property attributes and property type.

See F<SymbolBase.pm> for more info about various symbol types.

=head1 KNOWN BUGS

None.

=cut

package SymbolProperty ;

use strict ;
use warnings ;

use SymbolBase ;
our( @ISA ) ;
@ISA = qw( SymbolBase ) ;


###
# Constructor
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Allocate the object.
    
    my $self = $class->SUPER::new( @_ ) ;

    # Add data.

    $$self{ ATTRIBUTES } = { } ;
    $$self{ PROPERTY_TYPE } = undef ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'property' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;
    
    $self->SUPER::dump( $indent ) ;

    $indent += 2 ;
    print ' ' x $indent . "type = " . $$self{ PROPERTY_TYPE }->idl_type() . "\n" ;
    foreach my $attr_name ( keys( %{$$self{ ATTRIBUTES }} ) ) {
        print ' ' x $indent . "attribute " . $attr_name .
            " = " . $$self{ ATTRIBUTES }{ $attr_name } . "\n" ;
    }
}


###
# attributes
sub attributes {
    my $self = shift ;
    if ( @_ ) { $$self{ ATTRIBUTES } = shift ; }
    return $$self{ ATTRIBUTES } ;
}


###
# require_attribute
sub require_attribute {
    my $self = shift ;
    my $name = shift ;
    unless ( defined( $$self{ ATTRIBUTES }{ $name } ) ) {
        my $pname = $self->full_name() ;
        die "No attribute '$name' defined for property '$pname'.\n" ;
    }
    return $$self{ ATTRIBUTES }{ $name } ;
}


###
# property_type
sub property_type {
    my $self = shift ;
    if ( @_ ) { $$self{ PROPERTY_TYPE } = shift ; }
    return $$self{ PROPERTY_TYPE } ;
}


1 ;


__END__
