=head1 NAME

SymbolTable - symbol class that can have subsymbols

=head1 SYNOPSIS

    use SymbolTable ;

    # Parent class is SymbolBase.

    # Constructor - see SymbolBase.

    # Get hash of symbols.
    $symbols = $table->symbols() ;

    # Create symbol with given type.
    #   parser - parser to read source context from and to send error to
    #   type - symbol type name
    #   name - symbol name
    # Dies with error if symbol with the same name but different type
    # already exists. On success, returns reference to the symbol.
    $result = $table->create_symbol( $parser, $type, $name ) ;

    # Create new symbol with given type.
    #   parser - parser to read source context from and to send error to
    #   type - symbol type name
    #   name - symbol name
    # Dies with error if symbol with the same name is already defined.
    # On success, returns reference to the symbol.
    $result = $table->create_new_symbol( $parser, $type, $name ) ;

    # Find symbol with given name.
    #   name - symbol name as a list
    # Recurses to parent symbol tables to find the symbol. Is able to resolve
    # partially or fully qualified names (names containing "::"). Returns
    # reference to first symbol found, undef if no symbol of given name is
    # found.
    # $name should be equal to split( "::", $string_name ) - ie. list of
    # identifiers, starting with empty string if it's a fully qualified name
    # (name beginning with "::".
    $symbol = $table->find_symbol_list( $name_list ) ;
    
    # Find symbol with given name.
    #   name - symbol name as a string
    # It's assumed that $name is well-formed symbol name - it consists of
    # identifiers separated by "::", with no trailing "::" characters and
    # all identifiers at least one character long.
    # See find_symbol_list() for more info.
    $symbol = $table->find_symbol_string( $name_string ) ;

    # Get symbol with given name.
    #   name - simple symbol name
    # Unlike find_* methods this method just looks into this symbol table,
    # it's just shortcut for $table->symbols(){ $name }.
    $symbol = $table->get_symbol( $name ) ;

=head1 DESCRIPTION

This module defines C<SymbolTable> class. Its parent is C<SymbolBase>.
See F<SymbolBase.pm> for more info about various symbol types.

=head1 KNOWN BUGS

None.

=cut

package SymbolTable ;

use strict ;
use warnings ;

use SymbolBase ;
our( @ISA ) ;
@ISA = qw( SymbolBase ) ;


###
# Constructor
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Allocate the object.
    
    my $self = $class->SUPER::new( @_ ) ;

    # Add data.

    $$self{ SYMBOLS } = { } ;

    # Bless it.

    return bless( $self, $class ) ;
}


###
# type_name
sub type_name {
    'symbol_table' ;
}


###
# dump
sub dump {
    my $self = shift ;
    my $indent = shift ;

    $self->SUPER::dump( $indent ) ;

    $indent += 2 ;
    my $symbols = $$self{ SYMBOLS } ;
    foreach my $sname ( sort( keys( %$symbols ) ) ) {
        $$symbols{ $sname }->dump( $indent ) ;
    }
}


###
# symbols
sub symbols {
    $_[ 0 ]{ SYMBOLS } ;
}


###
# create_symbol
sub create_symbol {
    my( $self, $parser, $type, $name ) = @_ ;

    my $symbols = $$self{ SYMBOLS } ;
    my $symbol = $$symbols{ $name } ;
    if ( defined ( $symbol ) ) {
        if ( ref( $symbol ) ne $type ) {
            my $symbol_type_name = $symbol->type_name ;
            my $type_name = $type->type_name ;
            $parser->error( {
                MESSAGE => "'$symbol_type_name $name' redeclared as '$type_name'.",
                NO_DIE => 1,
            } ) ;
            $parser->error( {
                MESSAGE => "Previous declaration here.",
                FILE_NAME => $symbol->file_name(),
                LINE_NUMBER => $symbol->line_number(),
                CURRENT_LINE => '',
            } ) ;
        }
    }
    else {
        $symbol = $type->new( $name, $self ) ;
        $symbol->set_context_from_parser( $parser ) ;
        $$symbols{ $name } = $symbol ;
    }

    return $symbol ;
}


###
# create_new_symbol
sub create_new_symbol {
    my( $self, $parser, $type, $name ) = @_ ;

    my $symbols = $$self{ SYMBOLS } ;
    my $symbol = $$symbols{ $name } ;
    if ( defined ( $symbol ) ) {
        my $symbol_type_name = $symbol->type_name ;
        my $type_name = $type->type_name ;
        $parser->error( {
            MESSAGE => "Redefinition of '$type_name $name'.",
            NO_DIE => 1,
        } ) ;
        $parser->error( {
            MESSAGE => "Previously defined as '$symbol_type_name' here.",
            FILE_NAME => $symbol->file_name(),
            LINE_NUMBER => $symbol->line_number(),
            CURRENT_LINE => '',
        } ) ;
    }
    $symbol = $type->new( $name, $self ) ;
    $symbol->set_context_from_parser( $parser ) ;
    $$symbols{ $name } = $symbol ;

    return $symbol ;
}


###
# find_symbol_list
sub find_symbol_list {
    my $self = shift ;
    my @name = @_ ;

    # Search this and all parent namespaces.

    my $symbol ;
    my $parent ;
    NAMESPACE : while ( defined( $self ) ) {
        $parent = $self->parent_symbol() ;

        # Search only root namespace if name is fully qualified.

        if ( ( $name[ 0 ] eq '' ) && ( defined( $parent ) ) ) {
            next NAMESPACE ;
        }

        # Try to find the symbol.

        $symbol = $self ;
        foreach my $name_component ( @name ) {
            # Skip zero-width name components. (there should be at most one - the starting one)

            next if $name_component eq '' ;

            # Try to find child symbol.

            my $symbols = $$symbol{ SYMBOLS } ;
            if ( defined( $symbols ) ) {
                $symbol = $$symbols{ $name_component } ;
            }
            else {
                $symbol = undef ; # Current symbol is not a symbol table.
            }

            # Not found if $symbol became undef.

            last unless( defined( $symbol ) ) ;
        }

        # Found if $symbol is defined.

        last NAMESPACE if defined( $symbol ) ;
    }
    continue {
        # Try parent namespace.

        $self = $parent ;
    }

    # Done.

    return $symbol ;
}


###
# find_symbol_string
sub find_symbol_string {
    my $self = shift ;
    my $name = shift ;
    return $self->find_symbol_list( split( '::', $name ) ) ;
}


###
# get_symbol
sub get_symbol {
    my $self = shift ;
    my $name = shift ;
    return $$self{ SYMBOLS }{ $name } ;
}


1 ;


__END__
