=head1 NAME

Token - class representing a token

=head1 SYNOPSIS

    use Token ;

    ###
    # Class methods.

    # Constructor.
    $token = Token->new ;

    # Create initialized and analyzed token.
    $token = Token->new( $string ) ;

    
    ###
    # Object data methods.

    # Token string.
    $string = $token->token ;

    # Token type.
    # It's one of:
    #   i - integral number (value is the number)
    #   f - floating-point number (value is the number)
    #   s - doublequoted string (value is without the quotes, escape sequences will not be interpolated)
    #   c - quoted character (value is without the quotes)
    #   I - identifier (including keywords), Perl's "word" but not a number
    #   S - other weird thing, from fixed set of symbols or single non-whitespace character
    $type = $token->type ;

    # Token value.
    $value = $token->value ;

    ###
    # Other object methods.

    # Set and analyze token.
    $token->set( $string ) ;

    # Analyze token - determine its type and value.
    $token->analyze() ;

    # Get token type string.
    $type_string = $token->type_string() ;
    $type_string = $token->type_string( $type_character ) ;

    ###
    # Check methods. These return undef on success and error string on error.

    # Make sure that type of current token is equal to one of given types
    $error = $token->ensure_type( $type_character1, ... ) ;

    # Make sure that current token is an integer.
    $error = $token->ensure_integer() ;

    # Make sure that current token is a float.
    $error = $token->ensure_float() ;

    # Make sure that current token is an integer or a float.
    $error = $token->ensure_number() ;

    # Make sure that current token is a string.
    $error = $token->ensure_string() ;

    # Make sure that current token is a quoted character.
    $error = $token->ensure_character() ;

    # Make sure that current token is an identifier.
    $error = $token->ensure_identifier() ;

    # Make sure that current token is punctuation, operator or other symbol.
    $error = $token->ensure_symbol() ;

    # Make sure that current token string is equal to one of given string.
    #   string - expected token string (not token value)
    $error = $token->ensure_token( $string, ... ) ;

=head1 DESCRIPTION

Blah blah.

=head1 ISSUES

Punctuation, operators and other non-alphanumeric tokens are called 'symbols',
which is probably stupid name. They are not Symbols (see F<Symbol.pm>).

=head1 KNOWN BUGS

None.

=cut

package Token ;

use strict ;
use warnings ;

use POSIX qw( strtod ) ;


###
# Table of token type names.
my %TOKEN_TYPE_TO_STRING = (
    'i' => 'integer',
    'f' => 'float',
    's' => 'string',
    'c' => 'character',
    'I' => 'identifier',
    'S' => 'symbol'
) ;


###
# new
sub new {
    my $proto = shift ;
    my $class = ref( $proto ) || $proto ;
    
    # Allocate the object.
    
    my $self = {
        TOKEN => undef,
        VALUE => undef,
        TYPE => undef,
    } ;

    # Bless it.

    bless( $self, $class ) ;

    # Initialize.

    my $string = shift ;
    if ( defined( $string ) ) {
        $self->set( $string ) ;
    }

    return $self ;
}


###
# token
sub token {
    $_[ 0 ]{ TOKEN } ;
}


###
# value
sub value {
    $_[ 0 ]{ VALUE } ;
}


###
# type
sub type {
    $_[ 0 ]{ TYPE } ;
}


###
# set
sub set {
    my $self = shift ;
    $$self{ TOKEN } = shift ;
    $self->analyze() ;
}


###
# analyze
sub analyze {
    my $self = shift ;

    unless ( defined( $$self{ TOKEN } ) ) {
        die( 'Token not defined in analyze, stopped' ) ;
    }
    
    local $_ ;
    for ( $$self{ TOKEN } ) {
        if ( /^[0-9.]/ ) {
            # Number.

            if ( /[.eE]/ ) {
                # Float. See perlfaq4.

                $! = 0 ;
                my( $value, $unparsed ) = strtod( $_ ) ;
                if ( ( $value eq '' ) || ( $unparsed != 0 ) || $! ) {
                    die( "Unable to parse '$_' as a floating-point number, stopped" ) ;
                }
                
                $$self{ TYPE } = 'f' ;
                $$self{ VALUE } = $value ;
            }
            else {
                # Integer. Binary, octal, decimal or hexadecimal.

                $$self{ TYPE } = 'i' ;
                if ( /^0/ ) {
                    $$self{ VALUE } = oct( $_ ) ;
                }
                else {
                    $$self{ VALUE } = $_ ;
                }
            }
        }
        elsif ( /^"/ ) {
            # String.

            my $value = $_ ;
            $value =~ s/^"// ;
            $value =~ s/"$// ;
            $$self{ TYPE } = 's' ;
            $$self{ VALUE } = $value ;
        }
        elsif ( /^'/ ) {
            # Quoted character.

            my $value = $_ ;
            $value =~ s/^'// ;
            $value =~ s/'$// ;
            $$self{ TYPE } = 'c' ;
            $$self{ VALUE } = $value ;
        }
        elsif ( /^\w+$/ ) {
            # Identifier.

            $$self{ TYPE } = 'I' ;
            $$self{ VALUE } = $_ ;
        }
        else {
            # Other symbol.

            $$self{ TYPE } = 'S' ;
            $$self{ VALUE } = $_ ;
        }
    }
}


###
# type_string
sub type_string {
    my $self = shift ;
    my $type = shift || $$self{ TYPE } ;
    my $result = $TOKEN_TYPE_TO_STRING{ $type } || 'unknown' ;
    return $result ;
}


###
# ensure_type
sub ensure_type {
    my $self = shift ;
    foreach my $type ( @_ ) {
        if ( $$self{ TYPE } eq $type ) {
            return undef ;
        }
    }

    my $message = "Unexpected token type.\nExpected " ;
    for ( my $i = 0 ; $i < scalar( @_ ) ; $i++ ) {
        if ( $i > 0 ) {
            if ( $i == $#_ ) {
                $message .= ' or ' ;
            }
            else {
                $message .= ', ' ;
            }
        }
        $message .= $self->type_string( $_[ $i ] ) ;
    }
    $message .= ', found ' ;
    $message .= $self->type_string() ;
    $message .= '.' ;

    return $message ;
}


###
# ensure_integer
sub ensure_integer {
    return $_[ 0 ]->ensure_type( 'i' ) ;
}


###
# ensure_float
sub ensure_float {
    return $_[ 0 ]->ensure_type( 'f' ) ;
}


###
# ensure_number
sub ensure_number {
    return $_[ 0 ]->ensure_type( 'i', 'f' ) ;
}


###
# ensure_string
sub ensure_string {
    return $_[ 0 ]->ensure_type( 's' ) ;
}


###
# ensure_character
sub ensure_character {
    return $_[ 0 ]->ensure_type( 'c' ) ;
}


###
# ensure_identifier
sub ensure_identifier {
    return $_[ 0 ]->ensure_type( 'I' ) ;
}


###
# ensure_symbol
sub ensure_symbol {
    return $_[ 0 ]->ensure_type( 'S' ) ;
}


###
# ensure_token
sub ensure_token {
    my $self = shift ;
    foreach my $string ( @_ ) {
        if ( $$self{ TOKEN } eq $string ) {
            return undef ;
        }
    }

    my $message = "Unexpected token.\nExpected " ;
    for ( my $i = 0 ; $i < scalar( @_ ) ; $i++ ) {
        if ( $i > 0 ) {
            if ( $i == $#_ ) {
                $message .= ' or ' ;
            }
            else {
                $message .= ', ' ;
            }
        }
        $message .= "'" . $_[ $i ] . "'" ;
    }
    $message .= ', found ' ;
    $message .= "'" . $$self{ TOKEN } . "'" ;
    $message .= '.' ;

    return $message ;
}


1 ;


__END__
