=head1 NAME

Utils.pm - various helper functions

=head1 SYNOPSIS

    use Utils ;

    $line = Utils::get_line( $handle ) ;
    
    $file_name = Utils::find_file( $file_name, $root_dir, @directories ) ;

    Utils::write_open_namespace( $handle, $namespace ) ;

    Utils::write_close_namespace( $handle, $namespace ) ;

    $path = Utils::simplify_path( $path ) ;

    $path = Utils::relative_path( $path, $base ) ;

    $escaped = Utils::escape_string( $string ) ;

=head1 DESCRIPTION

=head2 Functions

This module offsers various simple helper functions.

=over

=item get_line

Reads and chomps single line from given stream.
Skips empty lines and '#' comments.

=item find_file

Tries to find file specified by I<file_name> in given I<directories>, which
are relative to I<root_dir>. On success, simplified fully qualified file
name relative to I<root_dir> will be returned. Otherwise C<undef> will be
returned. When I<root_dir> is absolute directory, returned string can be used
as unique identifier of the file (unless symbolic links are used). Otherwise
different strings may be returned for the same file when I<file_name> is
relative to different directory in I<directories>.

=item write_open_namespace

Writes C++ code opening given I<namespace> into file I<handle>.

=item write_close_namespace

Writes C++ code closing given I<namespace> into file I<handle>.

=item simplify_path

Get rid of superfluous ".." in path, remove "." elements and all trailing slashes.
For example F<./foo/.//bar/../boo/> will be converted to F<foo/boo>.

=item relative_path

Convert I<path> to path relative to I<base>. Both I<path> and I<base> should be
either absolute paths or paths relative to the same directory.

=item escape_string

Replace various characters by their C escape sequence counterparts and return
the modified string which should form valid C/C++ string constant when enclosed
in double-qutoes. Only the following characters will be replaced: backslash,
new-line, carriage return, horizontal tab, double-quote.

=back

=head1 KNOWN BUGS

None.

=cut

package Utils ;

use strict ;
use warnings ;

use FileHandle ;


###
# get_line
sub get_line( $ ) {
    my $handle = shift ;

    my $line ;
    LINE: while ( $line = $handle->getline() ) {
        chomp( $line ) ;
        if ( $line =~ /^\s*(#|$)/ ) {
            next LINE ;
        }
        last LINE ;
    }

    return $line ;
}


###
# find_file
sub find_file( $$\@ ) {
    my $fname = shift ;
    my $root = shift ;
    my $includes = shift ;

    $root .= '/' unless $root eq '' ;
    foreach my $i ( @$includes ) {
        my $inc = ( $i eq '' ? '' : $i . '/' ) ;
        my $cfname = simplify_path( $root . $inc . $fname ) ;

        if ( -f $cfname ) {
            # File found.

            return relative_path( $cfname, $root ) ;
        }
    }

    return undef ;
}


###
# write_open_namespace
sub write_open_namespace( $$ ) {
    my $handle = shift ;
    my $namespace = shift ;
    
    foreach my $name ( split( /::/, $namespace ) ) {
        $handle->print( "namespace $name {\n" ) ;
    }
}


###
# write_close_namespace
sub write_close_namespace( $$ ) {
    my $handle = shift ;
    my $namespace = shift ;

    my @names =  split( /::/, $namespace ) ;
    for ( my $i = $#names ; $i >= 0 ; $i-- ) {
        my $partial_namespace = join( '::', @names[ 0 .. $i ] ) ;
        $handle->print( "} // namespace $partial_namespace\n" ) ;
    }
}


###
# simplify_path
sub simplify_path( $ ) {
    my $string = shift ;
    my @elements = split( /\//, $string ) ;

    my @simplified ;
    my $first_element = 1 ;
    foreach my $element ( @elements ) {
        if (
            $element eq '..' and
            scalar( @simplified ) and
            $simplified[ $#simplified ] ne '..'
        ) {
            # Chdir back.

            pop( @simplified ) ;
        }
        elsif (
            $element eq '.' or
            ( ( $element eq '' ) and ( not $first_element ) )
        ) {
            # Ignore '.' elements and empty elements (multiple slashes).
        }
        else {
            # Chdir into $element.

            push( @simplified, $element ) ;
        }
        $first_element = 0 ;
    }

    return join( '/', @simplified ) ;
}


###
# relative_path
sub relative_path( $$ ) {
    my $path = shift ;
    my $base = shift ;

    my @pelements = split( /\//, simplify_path( $path ) ) ;
    my @belements = split( /\//, simplify_path( $base ) ) ;

    # Skip common part.

    ELEMENT: while ( @pelements and @belements ) {
        if ( $pelements[ 0 ] eq $belements[ 0 ] ) {
            shift( @pelements ) ;
            shift( @belements ) ;
        }
        else {
         last ELEMENT ;
        }
    }

    # Chdir using '..' to common base.

    my @result ;
    while ( @belements ) {
        push( @result, '..' ) ;
        shift( @belements ) ;
    }

    return join( '/', ( @result, @pelements ) ) ;
}


###
# escape_string
sub escape_string( $ ) {
    my $string = shift ;
    $string =~ s/\\/\\\\/g ;
    $string =~ s/\n/\\n/g ;
    $string =~ s/\r/\\r/g ;
    $string =~ s/\t/\\t/g ;
    $string =~ s/"/\\"/g ;
    return $string ;
}


1 ;


__END__
