/** @file file_system_unix.cpp
 * Class implementing traversing through the filesystem directory graph.
 */

/* 
 * begin        : Wed Jan 29 2003
 * copyright    : (C) 2002 by Marek Svantner, Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "global/global.h"
#include "utility/debug.h"
#include "utility/wildcards.h"
#include "utility/quote_string.h"
#include "filesys/native_file_system.h"
#include "filesys/arch/unix/file_system_unix.h"
#include <iosfwd>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdio>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>

namespace Massiv {
namespace Core {

#define CHECK_VALID()  \
    { \
    if( !is_valid() ) \
        { \
        throw ExInvalidFsIterator( "FS iterator is not valid." ); \
        } \
    }

namespace {


/**
 * Calls stat() for the given file.
 * @param file_path  The path of the file.
 * @param buf_stat Receives the result of stat.
 * @exception ExCannotGetFileInformation stat() failed.
 */
void get_file_stat
    (
    const char * const  file_path,
    struct stat *       buf_stat
    )
    {
    if( ::stat( file_path, buf_stat ) == -1 )
        {
        std::ostringstream  ss;
        ss << "Cannot get file information for '" << file_path <<
            "' (" << ::strerror( errno ) << ").";
        throw ExCanNotGetFileInformation( ss.str() );
        }
    }

    
} // anonymous namespace


// NativeFileSystem implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileSystem * NativeFileSystem::instantiate()
    {
    return new UnixFileSystem();
    }


std::string NativeFileSystem::get_home_directory()
    {
    char * home = ::getenv( "HOME" );

    if( home != NULL )
        {
        return home;
        }
    else
        {
        return ".";
        }
    }


// UnixFileSystem implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileIterator UnixFileSystem::open_directory
    (
    const std::string &         path,
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) const
    {
    return FileIterator( new UnixFileSystem::IteratorImplUnix( path, mask, mode ) );
    }


bool UnixFileSystem::is_file
    (
    const std::string & path
    ) const
    {
    std::string path_normalized = path;
    struct stat buf_stat;
    
    normalize_path( path_normalized );
    if( ::stat( path.c_str(), &buf_stat ) == -1 )
        {
        return false;
        }
    
    return S_ISREG( buf_stat.st_mode ) || S_ISLNK( buf_stat.st_mode );
    }


bool UnixFileSystem::is_directory
    (
    const std::string & path
    ) const
    {
    std::string path_normalized = path;
    struct stat buf_stat;    
    
    normalize_path( path_normalized );
    if( ::stat( path.c_str(), &buf_stat ) == -1 )
        {
        return false;
        }

    return S_ISDIR( buf_stat.st_mode );
    }


// UnixFileSystem::IteratorImplUnix implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UnixFileSystem::IteratorImplUnix::IteratorImplUnix
    (
    DirectoryEnumerationMode    mode
    ) : IteratorImpl( mode )
    {
    invalidate_handle();
    }  // UnixFileSystem::IteratorImplUnix::IteratorImplUnix()


UnixFileSystem::IteratorImplUnix::IteratorImplUnix
    (
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) : IteratorImpl( mask, mode )
    {
    invalidate_handle();
    bind_current( mask );
    }  // UnixFileSystem::IteratorImplUnix::IteratorImplUnix()


UnixFileSystem::IteratorImplUnix::IteratorImplUnix
    (
    const std::string &         path,
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) : IteratorImpl( path, mask, mode )
    {
    invalidate_handle();
    bind( path, mask );
    }  // UnixFileSystem::IteratorImplUnix::IteratorImplUnix()


UnixFileSystem::IteratorImplUnix::~IteratorImplUnix()
    {
    if( hdir != 0 )
        {
        close_handle();
        }
    }  // UnixFileSystem::IteratorImplUnix::~IteratorImplUnix()


void UnixFileSystem::IteratorImplUnix::bind
    (
    const std::string &     path,
    const std::string &     mask
    )
    {
    FileSystem &    unix_fs = Global::native_file_system();
    std::string     path_normalized( path );
    
    unix_fs.normalize_path( path_normalized );

    if( valid_flag && ( hdir != 0 ) )
        {
        // Iterator is already bound with some directory
        close_handle();
        invalidate_handle();
        }

    absolute_directory.clear();

    // Bind iterator with a new directory
    std::string current_dir( unix_fs.get_current_directory() );
    initial_current_directory = current_dir;

    this->path = path_normalized;
    this->mask = mask;
    valid_flag = false;


    path_normalized.erase( path_normalized.size() - 1, 1 );
        // erase trailing '/'

    hdir = ::opendir( path_normalized.c_str() );
        // Note that we have to do mask-matching ourselves, because
        // it is different under Windows and Linux.

    path_normalized += '/';
        // put the trailing '/' back
    
    if( hdir != 0 )
        {
        valid_flag = true;
        next_aux();
        }
    else
        {
        valid_flag = false;

        std::ostringstream  ss;
        ss << "Directory " << quote_const_string( path ) << " not found.";
        throw ExDirectoryNotFound( ss.str() );
        }
    }  // UnixFileSystem::IteratorImplUnix::bind()


Reference< FileIteratorImpl > UnixFileSystem::IteratorImplUnix::enter()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when enter() was called." 
            );
        }

    std::string     path( get_absolute_directory() );
    path.append( fname );

    return new IteratorImplUnix( path, mask );
    }  // UnixFileSystem::IteratorImplUnix::enter()


Reference< FileIteratorImpl > 
UnixFileSystem::IteratorImplUnix::descent()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when descent() was called." 
            );
        }

    std::string     file_name( fname );

    if( close_handle() )
        {
        // closedir successful
        std::string     abs_dir( get_absolute_directory() );

        valid_flag = false;
        invalidate_handle();

        std::string     path( abs_dir + file_name );
        bind( path, mask );
        }
    else
        {
        // closedir unsuccessful

        valid_flag = false;
        invalidate_handle();
        // fixme: throw exception or ignore?
        }

    return this;
    }  // UnixFileSystem::IteratorImplUnix::descent()


Reference< FileIteratorImpl >
UnixFileSystem::IteratorImplUnix::leave()
    {
    CHECK_VALID();

    std::string path( get_absolute_directory() );
    path.append( ".." );

    return new IteratorImplUnix( path, mask );
    }  // UnixFileSystem::IteratorImplUnix::leave()
    

bool UnixFileSystem::IteratorImplUnix::is_file() const
    {
    CHECK_VALID();
    FileSystem & unix_fs = Global::native_file_system();
    return unix_fs.is_file( get_absolute_directory() + fname );
    }  // UnixFileSystem::IteratorImplUnix::is_file()


bool UnixFileSystem::IteratorImplUnix::is_directory() const
    {
    CHECK_VALID();
    FileSystem & unix_fs = Global::native_file_system();
    return unix_fs.is_directory( get_absolute_directory() + fname );
    }  // UnixFileSystem::IteratorImplUnix::is_directory()


std::string UnixFileSystem::IteratorImplUnix::name() const
    {
    CHECK_VALID();
    return fname;
    }  // UnixFileSystem::IteratorImplUnix::name()


UInt64 UnixFileSystem::IteratorImplUnix::size() const
    {
    CHECK_VALID();
    off_t   result = 0;
       
    // FIXME: Hrmm, this is really effecient. Two get_absolute_directory()
    //        and stats.

    if( is_file() )
        {
        struct stat     buf_stat;
        std::string     path( get_absolute_directory() );
        path.append( fname );

        get_file_stat( path.c_str(), &buf_stat );
            // ExCanNotGetFileInformation might be thrown here

        result = buf_stat.st_size;
        }

    return result;
    }  // UnixFileSystem::IteratorImplUnix::size()


std::string UnixFileSystem::IteratorImplUnix::get_absolute_directory() const
    {
    CHECK_VALID();
    char    aux_abs_dir[ PATH_MAX + 1 ];
    
    if( absolute_directory.empty() )
        {
        // Absolute directory bound with this iterator has not yet been 
        // computed

        // todo: rewrite the following lines to be more effective?

        FileSystem &    unix_fs = Global::native_file_system();
        std::string     old_current_dir( unix_fs.get_current_directory() );
        bool            success = true;

        do
            {
            if( ::chdir( initial_current_directory.c_str() ) == -1 )
                {
                success = false;
                break;
                }

            if( ::chdir( path.c_str() ) == -1 )
                {
                success = false;
                break;
                }

            if( ::getcwd( aux_abs_dir, PATH_MAX + 1 ) == NULL )
                {
                success = false;
                break;
                }
            } while( 0 );

        unix_fs.set_current_directory( old_current_dir );

        if( !success )
            {
            throw ExCanNotChangeDirectory( "Cannot change directory." );
            }

        absolute_directory.assign( aux_abs_dir );
        if( absolute_directory[ absolute_directory.size() - 1 ] != '/' )
            {
            absolute_directory += '/';
            }        
        }
    return absolute_directory;
    }  // UnixFileSystem::IteratorImplUnix::get_absolute_directory()


bool UnixFileSystem::IteratorImplUnix::is_in_root() const
    {
    CHECK_VALID();

    std::string dir = get_absolute_directory();
    hope( dir.size() >= 1 );

    return ( dir.size() == 1 ) && ( dir[ 0 ] == '/' );
    }  // UnixFileSystem::IteratorImplUnix::is_in_root()


FileIteratorImpl * UnixFileSystem::IteratorImplUnix::next_aux()
    {
    CHECK_VALID();
    struct dirent * fdata;

    if( valid_flag )
        {
        while( true )
            {
            errno = 0;
            fdata = ::readdir( hdir );
            
            if( !fdata )
                {
                if( errno )
                    {
                    std::ostringstream  ss;
                    ss << "Error while enumerating directory "
                       << get_absolute_directory();
                    throw ExErrorEnumeratingDirectory( ss.str() );
                    }

                valid_flag = false;
                break;
                }
                
            if( Wildcards::match( mask.c_str(), fdata->d_name ) )
                {
                valid_flag = true;
                fname.assign( fdata->d_name );
                break;
                }
            else
            if( dir_enum_mode == FileIterator::ENUMERATE_ALL_DIRS )   
                {
                struct stat     buf_stat;
                std::string     path( get_absolute_directory() );
                path.append( fdata->d_name );

                get_file_stat( path.c_str(), &buf_stat );
                    // ExCanNotGetFileInformation might be thrown here

                if( S_ISDIR( buf_stat.st_mode ) )
                    {
                    valid_flag = true;
                    fname.assign( fdata->d_name );
                    break;
                    }
                }
            }
        }
    else
        {
        throw ExInvalidFsIterator
            ( 
            "Increment attempted on invalid iterator." 
            );
        }

    return this;
    }  // UnixFileSystem::IteratorImplUnix::next()


bool UnixFileSystem::IteratorImplUnix::is_handle_valid() const
    {
    return hdir != 0;
    }
    

void UnixFileSystem::IteratorImplUnix::invalidate_handle()
    {
    hdir = 0;
    }


bool UnixFileSystem::IteratorImplUnix::close_handle()
    {
    if( ::closedir( hdir ) == 0 )
        {
        return true;
        }
    return false;
    }

std::string UnixFileSystem::get_current_directory_API() const
    {
    char cwd[ PATH_MAX + 1 ];
    if( !::getcwd( cwd, PATH_MAX + 1 ) )
        {
        throw ExPathTooLong( "Path too long" );
        }
    return std::string( cwd );
    }
    

bool UnixFileSystem::set_current_directory_API
    (
    const std::string &     directory
    ) const
    {
    if( ::chdir( directory.c_str() ) == 0 )
        {
        return true;
        }
    return false;
    }


bool UnixFileSystem::is_absolute
    (
    const std::string & path
    ) const
    {
    return ( path.length() > 0 ) && ( path[ 0 ] == get_slash() );
    }


bool UnixFileSystem::is_relative
    (
    const std::string & path
    ) const
    {
    return ( path.length() == 0 ) || ( path[ 0 ] != get_slash() );
    }


char UnixFileSystem::get_slash() const
    {
    return '/';
    }
    
    
char UnixFileSystem::get_back_slash() const
    {
    return '\\';
    }


std::string UnixFileSystem::get_last_error_message() const
    {
    return strerror( errno );
    }


bool UnixFileSystem::create_directory_API
    (
    const std::string &     directory
    ) const
    {
    // Test if can be opened.
    DIR * dir = opendir( directory.c_str() );
    if( dir )
        {
        // Directory exist.
        closedir( dir );
        return true;
        }
    
    errno = 0;
    if( ::mkdir( directory.c_str(), 0777 ) == 0 )
        {
        return true;
        }
    if( errno == EEXIST )
        {
        // Test there is not a file with the same name.
        DIR * dir = opendir( directory.c_str() );
        if( dir )
            {
            // Directory exist.
            closedir( dir );
            return true;
            }
        }
    return false;
    }


bool UnixFileSystem::delete_file_API
    (
    const std::string &     file_name
    ) const
    {
    if( ::unlink( file_name.c_str() ) == 0 )
        {
        return true;
        }
    return false;
    }


bool UnixFileSystem::delete_directory_API
    (
    const std::string &     directory
    ) const
    {
    errno = 0;
    if( ::rmdir( directory.c_str() ) == 0 )
        {
        return true;
        }
    return false;
    }


#undef CHECK_VALID


} // namespace Core
} // namespace Massiv
