/** @file file_system_win32.cpp
 * Class implementing traversing through the filesystem directory graph.
 */

/* 
 * begin        : Wed Jan 29 2003
 * copyright    : (C) 2002 by Marek Svantner, Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "utility/debug.h"
#include "utility/wildcards.h"
#include "global/global.h"
#include "utility/quote_string.h"
#include "filesys/native_file_system.h"
#include "filesys/file_iterator.h"
#include "filesys/arch/win32/file_system_win32.h"
#include <cstdio>
#include <iosfwd>
#include <fstream>
#include <sstream>
#include <string>
#include <ctype.h>
#include <errno.h>
#include <windows.h>

namespace Massiv {
namespace Core {


#define CHECK_VALID()  \
    { \
    if( !is_valid() ) \
        { \
        throw ExInvalidFsIterator( "FS iterator is not valid." ); \
        } \
    }


// NativeFileSystem implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileSystem * NativeFileSystem::instantiate()
    {
    return new Win32FileSystem();
    }


std::string NativeFileSystem::get_home_directory()
    {
    char * home = ::getenv( "USERPROFILE" );

    if( home != NULL )
        {
        return home;
        }
    else
        {
        return ".";
        }
    }


// Win32FileSystem implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


FileIterator Win32FileSystem::open_directory
    (
    const std::string &         path,
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) const
    {
    return FileIterator( new Win32FileSystem::IteratorImplWin32( path, mask, mode ) );
    }
    

bool Win32FileSystem::is_file
    (
    const std::string & path
    ) const
    {
    std::string     path_normalized = path;
    WIN32_FIND_DATA data;
    HANDLE          handle;

    normalize_path( path_normalized, false );
    handle = ::FindFirstFile( path_normalized.c_str(), &data );
    
    if( handle == INVALID_HANDLE_VALUE )
        {
        ::FindClose( handle );
        return false;
        }

    ::FindClose( handle );
    return !flags_on( data.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY );
    }


bool Win32FileSystem::is_directory
    (
    const std::string & path
    ) const
    {
    std::string     path_normalized = path;
    WIN32_FIND_DATA data;
    HANDLE          handle;

    normalize_path( path_normalized, false );
    handle = ::FindFirstFile( path_normalized.c_str(), &data );
    
    if( handle == INVALID_HANDLE_VALUE )
        {
        ::FindClose( handle );
        return false;
        }
    
    ::FindClose( handle );
    return flags_on( data.dwFileAttributes, FILE_ATTRIBUTE_DIRECTORY );
    }


// Win32FileSystem::IteratorImplWin32 implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Win32FileSystem::IteratorImplWin32::IteratorImplWin32
    (
    DirectoryEnumerationMode    mode
    ) : IteratorImpl( mode )
    {
    handle = INVALID_HANDLE_VALUE;
    }  // Win32FileSystem::IteratorImplWin32::IteratorImplWin32()


Win32FileSystem::IteratorImplWin32::IteratorImplWin32
    (
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) : IteratorImpl( mask, mode )
    {
    handle = INVALID_HANDLE_VALUE;
    bind_current( mask );
    }  // Win32FileSystem::IteratorImplWin32::IteratorImplWin32()


Win32FileSystem::IteratorImplWin32::IteratorImplWin32
    (
    const std::string &                     path,
    const std::string &                     mask,
    DirectoryEnumerationMode                mode
    ) : IteratorImpl( path, mask, mode )
    {
    handle = INVALID_HANDLE_VALUE;
    bind( path, mask );
    }  // Win32FileSystem::IteratorImplWin32::IteratorImplWin32()


Win32FileSystem::IteratorImplWin32::~IteratorImplWin32()
    {
    if( handle != INVALID_HANDLE_VALUE )
        {
        FindClose( handle );
        }
    }  // Win32FileSystem::IteratorImplWin32::~IteratorImplWin32()


void Win32FileSystem::IteratorImplWin32::bind
    (
    const std::string &     path,
    const std::string &     mask
    )
    {
    FileSystem &    win32_fs = Global::native_file_system();
    std::string     path_normalized( path );
    
    win32_fs.normalize_path( path_normalized );

    if( valid_flag && ( handle != INVALID_HANDLE_VALUE ) )
        {
        // Iterator is already bound with some directory
        FindClose( handle );
        handle = INVALID_HANDLE_VALUE;
        }

    absolute_directory.clear();

    // Bind iterator with a new directory
    char temp[ MAX_PATH + 1 ];
    ::GetCurrentDirectory( MAX_PATH + 1, temp );
    initial_current_directory.assign( temp );

    this->path = path_normalized;
    this->mask = mask;
    valid_flag = false;

    handle = ::FindFirstFile( ( path_normalized + "*" ).c_str(), &fdata );
        // Note that we have to do mask-matching ourselves, because
        // it is different under Windows and Linux.
    
    if( handle != INVALID_HANDLE_VALUE )
        {
        valid_flag = true;

        if( !( fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) ||
            !( dir_enum_mode == FileIterator::ENUMERATE_ALL_DIRS ) ) 
            {
            if( !Wildcards::match( mask.c_str(), fdata.cFileName ) )   
                {
                next_aux();
                }
            }
        }
    else
        {
        valid_flag = false;
        
        std::ostringstream  ss;
        ss << "Directory " << quote_const_string( path ) << " not found.";
        throw ExDirectoryNotFound( ss.str() );
        }
    }  // Win32FileSystem::IteratorImplWin32::bind()


Reference< FileIteratorImpl > 
Win32FileSystem::IteratorImplWin32::enter()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when enter() was called." 
            );
        }

    return new IteratorImplWin32
        ( 
        std::string( get_absolute_directory() ) + fdata.cFileName,
        mask,
        dir_enum_mode
        );
    }  // Win32FileSystem::IteratorImplWin32::enter()


Reference< FileIteratorImpl > 
Win32FileSystem::IteratorImplWin32::descent()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when descent() was called." 
            );
        }

    if( FindClose( handle ) )
        {
        std::string     abs_dir( get_absolute_directory() );

        valid_flag = false;
        handle = INVALID_HANDLE_VALUE;

        bind( abs_dir + fdata.cFileName, mask );
        }
    else
        {
        valid_flag = false;
        handle = INVALID_HANDLE_VALUE;
        // fixme: throw exception or ignore?
        }

    return this;
    }  // Win32FileSystem::IteratorImplWin32::descent()


Reference< FileIteratorImpl > 
Win32FileSystem::IteratorImplWin32::leave()
    {
    CHECK_VALID();

    return new IteratorImplWin32
        ( 
        std::string( get_absolute_directory() ) + "..",
        mask,
        dir_enum_mode
        );
    }  // Win32FileSystem::IteratorImplWin32::leave()


bool Win32FileSystem::IteratorImplWin32::is_file() const
    {
    CHECK_VALID();
    return ( ( fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 );
    }  // Win32FileSystem::IteratorImplWin32::is_file()


bool Win32FileSystem::IteratorImplWin32::is_directory() const
    {
    CHECK_VALID();
    return ( ( fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 );
    }  // Win32FileSystem::IteratorImplWin32::is_directory()


std::string Win32FileSystem::IteratorImplWin32::name() const
    {
    CHECK_VALID();
    return std::string( fdata.cFileName );
    }  // Win32FileSystem::IteratorImplWin32::name()


UInt64 Win32FileSystem::IteratorImplWin32::size() const
    {
    CHECK_VALID();
    return ( fdata.nFileSizeHigh * MAXDWORD ) + fdata.nFileSizeLow;
    }  // Win32FileSystem::IteratorImplWin32::size()


std::string Win32FileSystem::IteratorImplWin32::get_absolute_directory() const
    {
    CHECK_VALID();
    
    if( absolute_directory.empty() )
        {
        // Absolute directory bound with this iterator has not yet been 
        // computed

        // todo: rewrite the following lines to be more effective?

        char old_current_dir[ MAX_PATH + 1 ];

        ::GetCurrentDirectory( MAX_PATH + 1, old_current_dir );

        ::SetCurrentDirectory( initial_current_directory.c_str() );
        ::SetCurrentDirectory( path.c_str() );
        char temp[ MAX_PATH + 1 ];
        ::GetCurrentDirectory( MAX_PATH + 1, temp );
        absolute_directory.assign( temp );
        
        ::SetCurrentDirectory( old_current_dir );

        int len = strlen( absolute_directory.c_str() ); // FIXME

        if( len <= MAX_PATH )
            {
            if( ( len > 0 ) && ( absolute_directory[ len - 1 ] ) != '\\' )
                {
                absolute_directory += '\\';
                }
            }
        else
            {
            panic( "Path too long." );
            }
        }

    return absolute_directory;
    }  // Win32FileSystem::IteratorImplWin32::get_absolute_directory()


bool Win32FileSystem::IteratorImplWin32::is_in_root() const
    {
    CHECK_VALID();

    std::string dir = get_absolute_directory();
    hope( dir.size() >= 3 );

    return 
        ( dir.size() == 3 ) && ( dir[ 1 ] == ':' ) && ( dir[ 2 ] == '\\' );
    }  // Win32FileSystem::IteratorImplWin32::is_in_root()


FileIteratorImpl * Win32FileSystem::IteratorImplWin32::next_aux()
    {
    CHECK_VALID();

    if( valid_flag )
        {
        int     result;

        while( true )
            {
            result = ::FindNextFile( handle, &fdata );
            
            if( !result )
                {
                valid_flag = false;
                break;
                }
                
            if( Wildcards::match( mask.c_str(), fdata.cFileName ) )
                {
                valid_flag = true;
                break;
                }

            if( ( fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) &&
                ( dir_enum_mode == FileIterator::ENUMERATE_ALL_DIRS ) ) 
                {
                valid_flag = true;
                break;
                }
            }
        }
    else
        {
        throw ExInvalidFsIterator
            ( 
            "Increment attempted on invalid iterator." 
            );
        }

    return this;
    }  // Win32FileSystem::IteratorImplWin32::next()


bool Win32FileSystem::IteratorImplWin32::is_handle_valid() const
    {
    return handle != INVALID_HANDLE_VALUE;
    }


void Win32FileSystem::IteratorImplWin32::invalidate_handle()
    {
    handle = INVALID_HANDLE_VALUE;
    }

bool Win32FileSystem::IteratorImplWin32::close_handle()
    {
    if( FindClose( handle ) != 0 )
        {
        return true;
        }
    return false;
    }


std::string Win32FileSystem::get_current_directory_API() const
    {
    char cwd[ MAX_PATH + 1 ];
    if( ::GetCurrentDirectory( MAX_PATH + 1, cwd ) == 0 )
        {
        throw ExCanNotChangeDirectory( strerror( GetLastError() ) );
        }
    return std::string( cwd );
    }

bool Win32FileSystem::set_current_directory_API
    (
    const std::string &     directory
    ) const
    {
    if( ::SetCurrentDirectory( directory.c_str() ) != 0 )
        {
        return true;
        }
    return false;
    }


bool Win32FileSystem::is_absolute
    (
    const std::string & path
    ) const
    {
    if( ( path.length() > 3 ) &&
        ( isalpha( path[ 0 ] ) ) &&
        ( path[ 1 ] == ':' ) &&
        ( path[ 2 ] == get_slash() ) )
        {
        return true;
        }
    else
        {
        return false;
        }
    }


bool Win32FileSystem::is_relative
    (
    const std::string & path
    ) const
    {
    return !is_absolute( path );
    }


char Win32FileSystem::get_slash() const
    {
    return '\\';
    }
    
    
char Win32FileSystem::get_back_slash() const
    {
    return '/';
    }


std::string Win32FileSystem::get_last_error_message() const
    {
    static char message_buffer[ 1024 ];
    FormatMessage
        (
        FORMAT_MESSAGE_IGNORE_INSERTS |
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
        message_buffer,
        sizeof( message_buffer ) / sizeof( char ) - 1,
        NULL
        );
    return std::string( message_buffer );
    }


bool Win32FileSystem::create_directory_API
    (
    const std::string &     directory
    ) const
    {
    // Test if there is a file with the same name.
    std::string current_dir = get_current_directory_API();
    if( set_current_directory_API( directory ) )
        {
        // Directory exist.
        set_current_directory_API( current_dir );
        return true;
        }

    /*HANDLE h;
    h = CreateFile( directory.c_str(), GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL );
    if( h != INVALID_HANDLE_VALUE )
        {
        // It is file.
        CloseHandle( h );
        return false;
        }*/
    if( ::CreateDirectory( directory.c_str(), NULL ) != 0 )
        {
        return true;
        }
    if( GetLastError() == ERROR_ALREADY_EXISTS )
        {
        return true;
        }
    return false;
    }


bool Win32FileSystem::delete_file_API
    (
    const std::string &     file_name
    ) const
    {
    if( ::DeleteFile( file_name.c_str() ) != 0 )
        {
        return true;
        }
    return false;
    }


bool Win32FileSystem::delete_directory_API
    (
    const std::string &     directory
    ) const
    {
    if( ::RemoveDirectory( directory.c_str() ) != 0 )
        {
        return true;
        }
    return false;        
    }


#undef CHECK_VALID


} // namespace Core
} // namespace Massiv
