/** @file file_system_win32.h
 * Class implementing traversing through the win32 filesystem directory graph.
 */

/* 
 * begin        : Wed Jan 29 2003
 * copyright    : (C) 2003 by Marek Svantner, Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FILE_SYSTEM_WIN32_H
#define MASSIV_FILE_SYSTEM_WIN32_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_FILE_SYSTEM_COMMON_H
#include "filesys/file_system_common.h"
#endif

#ifndef MASSIV_FILE_ITERATOR_H
#include "filesys/file_iterator.h"
#endif

#include <string>
#include <windows.h>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/**
 * Win32 file system.
 * On Win32 NativeFileSystem instantiates this.
 *
 * @note See FileSystem (and partially FileSystemCommon) documentation 
 *      for information about each specific method.
 */
class Win32FileSystem : public FileSystemCommon
    {
public:

    friend class NativeFileSystem;

    /**
     * Iterator that enumerates files from Win32 file system.
     */
    class IteratorImplWin32 : public FileSystemCommon::IteratorImpl
                              // MSVC6 requires FileSystemCommon:: prefix.
        {
        friend class Win32FileSystem;

    protected:
        
        /**
         */
        IteratorImplWin32
            (
            DirectoryEnumerationMode    mode
            );

        /**
         */
        IteratorImplWin32
            (
            const std::string &         mask,
            DirectoryEnumerationMode    mode
            );

        /**
         */
        IteratorImplWin32
            (
            const std::string &         path,
            const std::string &         mask,
            DirectoryEnumerationMode    mode
            );

        /**
         */
        virtual ~IteratorImplWin32();

    private: 

        NO_ASSIGNMENT_OPERATOR( IteratorImplWin32 );

    protected:

        /**
         */
        virtual void bind
            (
            const std::string &     path,
            const std::string &     mask
            );

        /**
         */
        virtual Reference< FileIteratorImpl > enter();

        /**
         */
        virtual Reference< FileIteratorImpl > descent();

        /**
         */
        virtual Reference< FileIteratorImpl > leave();

        /**
         */
        virtual bool is_file() const;

        /**
         */
        virtual bool is_directory() const;

        /**
         */
        virtual std::string name() const;

        /**
         */
        virtual UInt64 size() const;

        /**
         */
        virtual std::string get_absolute_directory() const;

        /**
         */
        virtual bool is_in_root() const;

        /**
         */
        virtual bool is_handle_valid() const;

        /**
         */
        virtual void invalidate_handle();

        /**
         */
        virtual bool close_handle();

    private:

        /**
         * Advances to the next file or subdirectory in the current directory.
         * @return Pointer to this iterator.
         * @exception ExInvalidFsIterator Iterator is invalid.
         */
        virtual FileIteratorImpl * next_aux();

    private:

        HANDLE              handle;
            /**< Contains handle needed for FindFirstFile()/FindNextFile() 
                 function or INVALID_HANDLE_VALUE. */

        WIN32_FIND_DATA     fdata;
            /**< Contains data about each found file. Must be properly
                 filled whenever @c valid_flag is true. */

        }; // class IteratorImplWin32

protected:

    /**
     */
    virtual FileIterator open_directory
        (
        const std::string &         path,
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const;

    /**
     */
    virtual bool is_file
        (
        const std::string & path
        ) const;
    
    /**
     */
    virtual bool is_directory
        (
        const std::string & path
        ) const;

    /**
     */
    virtual std::string get_current_directory_API() const;

    /**
     */
    virtual bool set_current_directory_API
        (
        const std::string &     directory
        ) const;

    /**
     */
    virtual bool create_directory_API
        (
        const std::string &     directory
        ) const;

    /**
     */
    virtual bool delete_file_API
        (
        const std::string &     file_name
        ) const;

    /**
     */
    virtual bool delete_directory_API
        (
        const std::string &     directory
        ) const;

    /**
     * Tests whether a given @c path is absolute.
     *
     * @param path The path.
     * @return @c true if the @c path is absolute, @c false otherwise.
     */
    virtual bool is_absolute
        (
        const std::string & path
        ) const;

    /**
     * Tests whether a given @c path is relative.
     *
     * @param path The path.
     * @return @c true if the @c path is relative, @c false otherwise.
     */
    virtual bool is_relative
        (
        const std::string & path
        ) const;

    /**
     */
    virtual char get_slash() const;

    /**
     */
    virtual char get_back_slash() const;

    /**
     */
    virtual std::string get_last_error_message() const;

    }; // class Win32FileSystem

/** @} */ // addtogroup filesys


} // namespace Core
} // namespace Massiv

#endif // MASSIV_FILE_SYSTEM_WIN32_H
