/** @file async_io.cpp
 * Asynchronous IO implementation.
 */

/*
 * begin        : Thu 6 Dec 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/async_io.h"
#include "filesys/basic_volume.h"

namespace Massiv {
namespace Core {


// Implementation of AsyncIO::ReadRequest
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

AsyncIO::ReadRequest::ReadRequest
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    ) : stream( stream ),
        buffer( ptr ),
        offset( offset ),
        length( length )
    {
    }


AsyncIO::ReadRequest::ProcessingResult AsyncIO::ReadRequest::process()
    {
    try
        {
        BasicVolume::read_from_stream( stream, buffer, offset, length );
        }
    catch( ... )
        {
        signal_error
            (
            STATUS_INFO
                (
                Status::FACILITY_FILE_SYSTEM,
                Status::PRIORITY_HIGH
                ),
            "IO error",
            false
            );
        }

    return RESULT_DONE;
    }


// Implementation of AsyncIO::WriteRequest
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

AsyncIO::WriteRequest::WriteRequest
    (
    std::ostream &  stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    ) : stream( stream ),
        buffer( ptr ),
        offset( offset ),
        length( length )
    {
    }


AsyncIO::WriteRequest::ProcessingResult AsyncIO::WriteRequest::process()
    {
    try
        {
        BasicVolume::write_to_stream( stream, buffer, offset, length );
        }
    catch( ... )
        {
        signal_error
            (
            STATUS_INFO
                (
                Status::FACILITY_FILE_SYSTEM,
                Status::PRIORITY_HIGH
                ),
            "IO error",
            false
            );
        }

    return RESULT_DONE;
    }


// Implementation of AsyncIO::CopyAndWriteRequest
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

AsyncIO::CopyAndWriteRequest::CopyAndWriteRequest
    (
    std::ostream &  stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    ) : WriteRequest( stream, buffer_copy, offset, length )
    {
    UInt32 i;

    hope( length <= MAX_BUFFER_LENGTH );

    for( i = 0; i < length; i++ )
        {
        buffer_copy[ i ] = ptr[ i ];
        }
    }


// Implementation of AsyncIO::FlushRequest
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

AsyncIO::FlushRequest::FlushRequest
    (
    std::ostream & stream
    ) : stream( stream )
    {
    }


AsyncIO::FlushRequest::ProcessingResult AsyncIO::FlushRequest::process()
    {
    try
        {
        stream.flush();
        }
    catch( ... )
        {
        signal_error
            (
            STATUS_INFO
                (
                Status::FACILITY_FILE_SYSTEM,
                Status::PRIORITY_HIGH
                ),
            "IO error",
            false
            );
        }

    return RESULT_DONE;
    }


// Implementation of AsyncIO
// ~~~~~~~~~~~~~~~~~~~~~~~~~

AsyncIO::AsyncIO
    (
    int max_requests
    ) : RequestQueue( max_requests )
    {
    }


AsyncIO::~AsyncIO()
    {
    stop();
    log_errors();
    }


void AsyncIO::read
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    )
    {
    RequestPtr r;

    // Submit request.

    r = async_read( stream, ptr, offset, length );
    r->wait_for_completion();

    // Log new errors.

    log_errors();
    }


AsyncIO::RequestPtr AsyncIO::async_read
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    )
    {
    RequestPtr r;

    // Log old errors.

    log_errors();

    // Submit request.

    r = new ReadRequest( stream, ptr, offset, length );
    queue_request( r.get() );

    return r;
    }


void AsyncIO::write
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    )
    {
    RequestPtr r;

    // Submit request.

    r = async_write( stream, ptr, offset, length );
    r->wait_for_completion();

    // Log new errors.

    log_errors();
    }


AsyncIO::RequestPtr AsyncIO::async_write
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    )
    {
    RequestPtr r;

    // Log old errors.

    log_errors();

    // Submit request.

    r = new WriteRequest( stream, ptr, offset, length );
    queue_request( r.get() );

    return r;
    }


AsyncIO::RequestPtr AsyncIO::async_copy_and_write
    (
    std::iostream & stream,
    UInt8 *         ptr,
    UInt32          offset,
    UInt32          length
    )
    {
    RequestPtr r;

    // Log old errors.

    log_errors();

    // Submit request.

    r = new CopyAndWriteRequest( stream, ptr, offset, length );
    queue_request( r.get() );

    return r;
    }


void AsyncIO::flush
    (
    std::ostream & stream
    )
    {
    RequestPtr r;

    // Submit request.

    r = async_flush( stream );
    r->wait_for_completion();

    // Log new errors.

    log_errors();
    }


AsyncIO::RequestPtr AsyncIO::async_flush
    (
    std::ostream & stream
    )
    {
    RequestPtr r;

    // Log old errors.

    log_errors();

    // Submit request.

    r = new FlushRequest( stream );
    queue_request( r.get() );

    return r;
    }


} // namespace Massiv::Core
} // namespace Massiv
