/** @file async_io.h
 * Asynchronous IO.
 */

/*
 * begin        : Thu 6 Dec 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ASYNC_IO_H
#define MASSIV_ASYNC_IO_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_REQUEST_QUEUE_H
#include "thread/request_queue.h"
#endif

#include <iostream>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/**
 * Asynchronous IO helper class.
 *
 * Each %AsyncIO instance launches one worker thread that processes
 * submitted IO requests in FIFO order. Particular requests processing does
 * not overlap, thus an older request blocks a newer one.
 *
 * Since IO stream library does not allow parallel access to
 * IO stream objects, client must not use those streams that
 * are currently used by %AsyncIO.
 */
class AsyncIO : public RequestQueue
    {
private:

    /**
     * Read request.
     */
    class ReadRequest : public RequestQueue::Request
        {
    public:

        /**
         * Constructs read request.
         * When request gets processed read data will be directly written
         * to buffer pointed by @c ptr.
         */
        ReadRequest
            (
            std::iostream & stream,
            UInt8 *         ptr,
            UInt32          offset,
            UInt32          length
            );

        /**
         * Processes request.
         * Runs in a different thread!
         */
        virtual ProcessingResult process();

    private:

        std::iostream & stream;
            /**< Stream to read from. */
        UInt8 *         buffer;
            /**< Buffer to write to. */
        UInt32          offset;
            /**< Offset within the stream. */
        UInt32          length;
            /**< Number of bytes. */
        }; // ReadRequest

    /**
     * Write request.
     */
    class WriteRequest : public RequestQueue::Request
        {
    public:

        /**
         * Constructs write request.
         * When request gets processed data will be directly read
         * from buffer pointed by @c ptr.
         */
        WriteRequest
            (
            std::ostream &  stream,
            UInt8 *         ptr,
            UInt32          offset,
            UInt32          length
            );

        /**
         * Processes request.
         * Runs in a different thread!
         */
        virtual ProcessingResult process();

    private:

        std::ostream &  stream;
            /**< Stream to write to. */
        UInt8 *         buffer;
            /**< Buffer to write. */
        UInt32          offset;
            /**< Offset within the stream. */
        UInt32          length;
            /**< Number of bytes. */
        }; // WriteRequest

    /**
     * Copy & write request.
     */
    class CopyAndWriteRequest : public WriteRequest
        {
    public:

        /**
         * Constants.
         */
        enum Constants
            {
            MAX_BUFFER_LENGTH = 4096
                /**< Maximum length of buffer that can be written. */
            }; // Constants

        /**
         * Constructs copy & write request.
         * Prior processing copies data to an internal store.
         */
        CopyAndWriteRequest
            (
            std::ostream &  stream,
            UInt8 *         ptr,
            UInt32          offset,
            UInt32          length
            );

    private:

        UInt8   buffer_copy[ MAX_BUFFER_LENGTH ];
            /**< Buffer to write. */
        }; // CopyAndWriteRequest

    /**
     * Flush request.
     */
    class FlushRequest : public RequestQueue::Request
        {
    public:

        /**
         * Constructs flush request.
         */
        FlushRequest
            (
            std::ostream & stream
            );

        /**
         * Processes request.
         * Runs in a different thread!
         */
        virtual ProcessingResult process();

    private:

        std::ostream & stream;
            /**< Stream to be flushed. */
        }; // FlushRequest

public:

    /**
     * Constructs IO request queue.
     * @c max_requests determines maximum number of pending requests.
     */
    AsyncIO
        (
        int max_requests
        );

    /**
     * Destructor.
     */
    ~AsyncIO();

    /**
     * Synchronous read.
     * Submits read request and waits for its completion.
     */
    void read
        (
        std::iostream & stream,
        UInt8 *         ptr,
        UInt32          offset,
        UInt32          length
        );

    /**
     * Asynchronous read.
     * Submits read request.
     * @c ptr and @c stream must remain valid until request completes.
     */
    RequestPtr async_read
        (
        std::iostream & stream,
        UInt8 *         ptr,
        UInt32          offset,
        UInt32          length
        );

    /**
     * Synchronous write.
     * Submits write request and waits for its completion.
     */
    void write
        (
        std::iostream & stream,
        UInt8 *         ptr,
        UInt32          offset,
        UInt32          length
        );

    /**
     * Asynchronous write.
     * Submits write request.
     * @c ptr and @c stream must remain valid until request completes.
     */
    RequestPtr async_write
        (
        std::iostream & stream,
        UInt8 *         ptr,
        UInt32          offset,
        UInt32          length
        );

    /**
     * Asynchronous write.
     * Submits copy & write request.
     * Copies buffer to internal store so that at the time
     * of processing of the request @c ptr might be invalid
     * or already point to different data.
     * @c stream must remain valid until request completes.
     */
    RequestPtr async_copy_and_write
        (
        std::iostream & stream,
        UInt8 *         ptr,
        UInt32          offset,
        UInt32          length
        );

    /**
     * Synchronous flush.
     * Submits flush request and waits for completion.
     */
    void flush
        (
        std::ostream & stream
        );

    /**
     * Asynchronous flush.
     * Submits flush request.
     */
    RequestPtr async_flush
        (
        std::ostream & stream
        );

    }; // AsyncIO

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_ASYNC_IO_H
