/** @file basic_volume.cpp
 * Basic volume.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/basic_volume.h"
#include "utility/quote_string.h"
#include "utility/debug.h"
#include "data/copy_file.h"
#include <iosfwd>
#include <sstream>
#include <cstring>
#include <typeinfo>

namespace Massiv {
namespace Core {


// Implementation of BasicVolume::FStreamDescriptor
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::FStreamDescriptor::FStreamDescriptor()
    {
    name = "<noname>";
    flags = DEFAULT_STREAM_FLAGS;
    }


void BasicVolume::FStreamDescriptor::init
    (
    const std::string & stream_name,
    int                 stream_flags
    )
    {
    name = stream_name;
    flags = stream_flags;
    }


void BasicVolume::FStreamDescriptor::write
    (
    std::ostream & str
    ) const
    {
    StreamDescriptor::write( str );

    str << quote_const_string( name ) << " ";
    str << flags << '\n';
    }


void BasicVolume::FStreamDescriptor::restore
    (
    StreamPool *    owner_pool,
    std::istream &  str
    )
    {
    StreamDescriptor::restore( owner_pool, str );

    str >> quote_string( name );
    str >> flags;
    }


void BasicVolume::FStreamDescriptor::set_name
    (
    const std::string & name
    )
    {
    this->name = name;
    }


std::string BasicVolume::FStreamDescriptor::get_name()
    {
    return name;
    }


// Implementation of BasicVolume::FSDHierarchyNode
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::FSDHierarchyNode::FSDHierarchyNode
    (
    FSDHierarchyNode *  parent_node,
    const std::string & path
    ) : volume( NULL ),
        path( path ),
        parent_node( parent_node ),
        lock_counter( 0 )
    {
    }


BasicVolume::FSDHierarchyNode::FSDHierarchyNode
    (
    BasicVolume &                   basic_volume,
    BasicVolume::FSDHierarchyNode * parent_node,
    const std::string &             path
    ) : volume( &basic_volume ),
        path( path ),
        parent_node( parent_node ),
        lock_counter( 0 )
    {
    }  // BasicVolume::FSDHierarchyNode::FSDHierarchyNode()


BasicVolume::FSDHierarchyNode::~FSDHierarchyNode()
    {
    SubnodeList::iterator it( subnode_list.begin() );

    while( it != subnode_list.end() )
        {
        delete it->second;
        it++;
        }
    }  // BasicVolume::FSDHierarchyNode::~FSDHierarchyNode()


void BasicVolume::FSDHierarchyNode::set_volume
    (
    BasicVolume &       volume
    )
    {
    this->volume = &volume;
    }  // BasicVolume::FSDHierarchyNode::set_volume()


BasicVolume::FStreamDescriptor * 
BasicVolume::FSDHierarchyNode::get_stream_descriptor
    (
    const std::string &     stream_name
    ) const
    {
    const FStreamDescriptor *     result = NULL;
    StreamList::const_iterator it;

    it = stream_list.find( stream_name );

    if( it == stream_list.end() )
        {
        // Stream does not exist in this directory.
        result = NULL;
        }
    else
        {
        result = &it->second;
        }

    return const_cast<FStreamDescriptor *>( result );
    }  // BasicVolume::FSDHierarchyNode::get_stream_descriptor()


BasicVolume::FSDHierarchyNode * BasicVolume::FSDHierarchyNode::get_subnode
    (
    const std::string &     subnode_name
    ) const
    {
    const FSDHierarchyNode *      result = NULL;
    SubnodeList::const_iterator it;

    it = subnode_list.find( subnode_name );

    if( it == subnode_list.end() )
        {
        // Subnode does not exist in this directory.
        result = NULL;
        }
    else
        {
        result = it->second;
        }
       
    return const_cast<FSDHierarchyNode*>( result );
    }  // BasicVolume::FSDHierarchyNode::get_subnode()


BasicVolume::FSDHierarchyNode * BasicVolume::FSDHierarchyNode::create_subnode
    (
    const std::string & subnode_name
    )
    {
    hope( volume );

    std::pair<SubnodeList::iterator, bool>  result;
    std::string                             new_path( path );
    
    if( new_path.size() > 1 )
        {
        new_path.push_back( volume->slash );
        }

    new_path += subnode_name;

    // Check whether the subnode_name syntax is correct
    if( subnode_name.find( volume->slash ) != std::string::npos )
        {
        std::ostringstream  ss;
        ss << "Cannot create FSDHierarchyNode '" << subnode_name << "' ("
           << "Node name cannot contain '" << volume->slash << "')";
        throw ExCanNotCreateDirectory( ss.str() );
        }

    // Insert the subnode
    result = subnode_list.insert
        ( 
        SubnodeList::value_type
            ( subnode_name, new FSDHierarchyNode( *volume, this, new_path ) )
        );

    return result.first->second;    
        // It doesn't matter if the subnode already existed or not.
    }  // BasicVolume::FSDHierarchyNode::create_subnode()


void BasicVolume::FSDHierarchyNode::remove_subnode
    (
    const std::string &     subnode_name
    )
    {
    SubnodeList::iterator it;

    it = subnode_list.find( subnode_name );

    if( it == subnode_list.end() )
        {
        std::ostringstream ss;
        ss << "Cannot delete FSDHierarchyNode " << subnode_name;
        throw ExCanNotDeleteDirectory( ss.str() );
        }
    else
        {
        // Delete subnode
        delete it->second;
        subnode_list.erase( it );
        }
    }  // BasicVolume::FSDHierarchyNode::remove_subnode()


BasicVolume::FStreamDescriptor & 
BasicVolume::FSDHierarchyNode::create_stream_descriptor
    (
    const std::string &     name,
    bool &                  existed_flag
    )
    {
    return insert_stream_descriptor( name, FStreamDescriptor(), existed_flag );
    }  // BasicVolume::FSDHierarchyNode::create_stream_descriptor()


BasicVolume::FStreamDescriptor & 
BasicVolume::FSDHierarchyNode::insert_stream_descriptor
    (
    const std::string &         name,
    const FStreamDescriptor &   descriptor,
    bool &                      existed_flag
    )
    {
    std::pair<StreamList::iterator, bool>   result;

    // Register the stream.

    result = stream_list.insert
        ( 
        StreamList::value_type( name, descriptor ) 
        );

    FStreamDescriptor & f = result.first->second;

    if( !result.second )
        {
        // Stream already registered.
        existed_flag = true;
        return f;
        }
    else
        {
        // Initialize new stream.
        existed_flag = false;
        return f;
        }
    }  // BasicVolume::FSDHierarchyNode::insert_stream_descriptor()


bool BasicVolume::FSDHierarchyNode::remove_stream_descriptor
    (
    const std::string &     name
    )
    {
    StreamList::iterator it;

    it = stream_list.find( name );

    return remove_stream_descriptor( it );
    }  // BasicVolume::FSDHierarchyNode::remove_stream_descriptor()


bool BasicVolume::FSDHierarchyNode::remove_stream_descriptor
    (
    FSDHierarchyNode::StreamList::iterator &    it
    )
    {
    if( it == stream_list.end() )
        {
        // Stream does not exist.
        return false;
        }
    else
        {
        // Erase the descriptor from the descriptor list.
        stream_list.erase( it );

        // Delete this node - must be empty and unlocked, otherwise
        // this is an empty operation.
        delete_empty();

        return true;
        }
    }  // BasicVolume::FSDHierarchyNode::remove_stream_descriptor()


void BasicVolume::FSDHierarchyNode::unlock()
    {
    hope( lock_counter > 0 );
    lock_counter--;
    if( stream_list.size() + subnode_list.size() == 0 )
        {
        delete_empty();
        }
    }  // BasicVolume::FSDHierarchyNode::unlock()


void BasicVolume::FSDHierarchyNode::delete_empty()
    {
    hope( volume );

    FSDHierarchyNode *  node_to_delete = this;

    while( node_to_delete->stream_list.size() + 
           node_to_delete->subnode_list.size() == 0 )
        {
        if( node_to_delete->lock_counter )  break;

        FSDHierarchyNode *  parent = node_to_delete->parent_node;
        
        if( parent == node_to_delete )
            {
            // This is a root node that cannot be deleted.
            break;
            }

        std::string             node_name;
        std::string::size_type  last_slash = 
                                node_to_delete->path.rfind( volume->slash );

        if( last_slash < node_to_delete->path.size() - 1 )
            {
            node_name = node_to_delete->path.substr( last_slash + 1 );
            }

        SubnodeList::iterator   it_sub_parent
            ( 
            parent->subnode_list.find( node_name )
            );

        if( it_sub_parent != parent->subnode_list.end() )
            {
            parent->subnode_list.erase( it_sub_parent );
            }
        
        delete node_to_delete;
        node_to_delete = parent;
        }
    }  // BasicVolume::FSDHierarchyNode::delete_empty()


void BasicVolume::FSDHierarchyNode::save_streams
    (
    StreamPool::OStream &   stream
    )
    {
    // Save all streams of this node.
    if( stream_list.size() )
        {
        StreamList::iterator    it;           

        for( it = stream_list.begin(); it != stream_list.end(); it++ )
            {
            it->second.write( stream );
            }
        }

    // Save recursively streams of all subnodes.
    if( subnode_list.size() )
        {
        SubnodeList::iterator   it;

        for( it = subnode_list.begin(); it != subnode_list.end(); it++ )
            {
            it->second->save_streams( stream );
            }
        }
    }  // BasicVolume::FSDHierarchyNode::save_streams()


BasicVolume::FSDHierarchyNode::StreamList::size_type 
BasicVolume::FSDHierarchyNode::stream_count()
    {
    return stream_list.size();
    }


BasicVolume::FSDHierarchyNode::SubnodeList::size_type 
BasicVolume::FSDHierarchyNode::subnode_count()
    {
    return subnode_list.size();
    }


bool BasicVolume::FSDHierarchyNode::empty()
    {
    return ( !stream_list.size() && !subnode_list.size() );
    }


void BasicVolume::FSDHierarchyNode::lock()
    {
    lock_counter++;
    }


BasicVolume::FSDHierarchyNode * 
BasicVolume::FSDHierarchyNode::get_parent_node()
    {
    hope( parent_node );
    return parent_node;
    }


void BasicVolume::FSDHierarchyNode::set_parent_node
    (
    FSDHierarchyNode *  parent_node
    )
    {
    hope( parent_node );
    this->parent_node = parent_node;
    }


std::string BasicVolume::FSDHierarchyNode::get_path()
    {
    return path;
    }


// Implementation of BasicVolume::XXXStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define COMMON_STREAM_IMPLEMENTATION( StreamName, CompressedStreamName ) \
    \
    BasicVolume::StreamName::~StreamName() \
        { \
        if( compressed_stream ) \
            { \
            delete compressed_stream; \
            } \
        } \
    \
    bool BasicVolume::StreamName::open \
        ( \
        FStreamDescriptor &     descriptor, \
        std::ios_base::openmode mode \
        ) \
        { \
        if( !raw_stream.open( descriptor, mode ) ) \
            { \
            return false; \
            } \
        \
        if( descriptor.flags & STREAM_COMPRESSED ) \
            { \
            compressed_stream = new CompressedStreamName( raw_stream ); \
            Super::init( compressed_stream->rdbuf() ); \
            } \
        else \
            { \
            Super::init( raw_stream.rdbuf() ); \
            } \
        \
        return true; \
        }


// Implementation of BasicVolume::IFStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::IFStream::IFStream()
    : Super( NULL ),
      compressed_stream( NULL )
    {
    }


BasicVolume::IFStream::IFStream
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode
    ) : Super( NULL ),
        compressed_stream( NULL )
    {
    open( volume, stream_name, mode );
    }


BasicVolume::IFStream::IFStream
    (
    FStreamDescriptor &     descriptor,
    std::ios_base::openmode mode
    ) : Super( NULL ),
        compressed_stream( NULL )
    {
    open( descriptor, mode );
    }


bool BasicVolume::IFStream::open
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode
    )
    {
    return open( volume.access( stream_name, mode, 0 ), mode );
    }

COMMON_STREAM_IMPLEMENTATION( IFStream, CompressedIStream );


// Implementation of BasicVolume::OFStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::OFStream::OFStream()
    : Super( NULL ),
      compressed_stream( NULL )
    {
    }


BasicVolume::OFStream::OFStream
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode,
    int                     creation_flags
    ) : Super( NULL ),
        compressed_stream( NULL )
    {
    open( volume, stream_name, mode, creation_flags );
    }


BasicVolume::OFStream::OFStream
    (
    FStreamDescriptor &     descriptor,
    std::ios_base::openmode mode
    ) : Super( NULL ),
        compressed_stream( NULL )
    {
    open( descriptor, mode );
    }


bool BasicVolume::OFStream::open
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode,
    int                     creation_flags
    )
    {
    return open( volume.access( stream_name, mode, creation_flags ), mode );
    }


COMMON_STREAM_IMPLEMENTATION( OFStream, CompressedOStream );

#undef COMMON_STREAM_IMPLEMENTATION


// Implemenation of BasicVolume::FStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::FStream::FStream()
    : Super( NULL )
    {
    }


BasicVolume::FStream::FStream
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode
    ) : Super( NULL )
    {
    open( volume, stream_name, mode );
    }


BasicVolume::FStream::FStream
    (
    FStreamDescriptor &     descriptor,
    std::ios_base::openmode mode
    ) : Super( NULL )
    {
    open( descriptor, mode );
    }


bool BasicVolume::FStream::open
    (
    BasicVolume &           volume,
    const std::string &     stream_name,
    std::ios_base::openmode mode
    )
    {
    return open( volume.access( stream_name, mode, 0 ), mode );
    }


bool BasicVolume::FStream::open
    (
    FStreamDescriptor &     descriptor,
    std::ios_base::openmode mode
    )
    {
    if( descriptor.flags & STREAM_COMPRESSED )
        {
        return false;
        }

    if( !raw_stream.open( descriptor, mode ) )
        {
        return false;
        }

    Super::init( raw_stream.rdbuf() );

    return true;
    }


// Implementation of BasicVolume
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::BasicVolume
    (
    int flags,
    int block_size,
    int cache_size
    ) : StreamPool( block_size, cache_size ),
        root_directory( &root_directory, std::string( 1, get_slash() ) ),
        current_directory_node( &root_directory ),
        stream_counter( 0 ),
        volume_flags( flags ),
        slash( get_slash() ),
        backslash( get_back_slash() )
    {
    root_directory.set_volume( *this );
    }


BasicVolume::~BasicVolume()
    {
    }


void BasicVolume::flush()
    {
    flush_pool();
    }


void BasicVolume::set_current_directory
    (
    const std::string & path
    )
    {
    hope( current_directory_node );
    
    // Save old current directory info
    PathVector          old_current_directory_name( current_directory_name );
    FSDHierarchyNode *  old_current_directory_node = current_directory_node;
   
    std::string::size_type  begpos = 0;
        // begpos - position in the path where to begin a search for '/'

    if( ( path.size() > 0 ) && ( path[ 0 ] == slash ) )
        {
        // Absolute path has been given - we set the current directory to
        // root and convert the path to relative (but this time it is 
        // relative to root).
        
        current_directory_name.clear();
        current_directory_node = &root_directory;
        begpos = 1;  // Will cause skipping of the leading '/' in the path.
        }

    // Now we can suppose that a relative path has been given.
    std::string::size_type  endpos = path.find( slash, begpos );
    std::string             component;
    
    if( endpos == std::string::npos )
        {
        component = path.substr( begpos ); // Substring from begpos to the end.
        }
    else
        {
        component = path.substr( begpos, endpos - begpos ); 
        }

    while( !component.empty() )
        {
        if( component == "." )
            {
            break;
            }
        else
        if( component == ".." )
            {
            current_directory_node = current_directory_node->parent_node;
            if( current_directory_name.size() > 0 )
                {
                current_directory_name.pop_back();
                }
            }
        else
            {
            current_directory_node = 
                    current_directory_node->get_subnode( component );

            if( !current_directory_node )
                {
                // The desired directory doesn't exist - 'rollback' the operation
                // and throw exception

                current_directory_name = old_current_directory_name;
                current_directory_node = old_current_directory_node;

                std::ostringstream  ss;
                ss << "Cannot change Volume directory to '" << path << "'";
                throw ExCanNotChangeDirectory( ss.str() );
                }
            else
                {
                current_directory_name.push_back( component );
                }
            }

        if( endpos == std::string::npos )
            {
            break;
            }

        begpos = endpos;
        begpos++;
        endpos = path.find( slash, begpos );
        
        if( endpos == std::string::npos )
            {
            component = path.substr( begpos );
            }
        else
            {
            component = path.substr( begpos, endpos - begpos );
            }
        } // while( !component.empty() )
    } // BasicVolume::set_current_directory()


FileIterator BasicVolume::open_current_directory
    (
    const std::string &                     mask,
    FileSystem::DirectoryEnumerationMode    mode
    ) const
    {
    return open_directory
        ( 
        get_current_directory(), 
        mask,
        mode
        );
    }  // BasicVolume::open_current_directory()


FileIterator BasicVolume::open_directory
    (
    const std::string &                     path,
    const std::string &                     mask,
    FileSystem::DirectoryEnumerationMode    mode
    ) const
    {
    BasicVolume * self = const_cast<BasicVolume *>( this );
    return FileIterator( new IteratorImpl( *self, path, mask, mode ) );
    }  // BasicVolume::open_directory()


std::string BasicVolume::get_current_directory() const
    {
    std::string result( 1, slash );

    PathVector::const_iterator it( current_directory_name.begin() );

    while( it != current_directory_name.end() )
        {
        if( result.size() > 1 )
            {
            result.push_back( slash );
            }

        result.append( *it );
        it++;
        }

    return result;
    }  // BasicVolume::get_current_directory()


void BasicVolume::normalize_path
    (
    std::string &   path,
    bool            slash_terminated
    ) const
    {
    std::string::iterator it( path.begin() );

    while( it != path.end() )
        {
        if( *it == backslash )
            {
            *it = slash;
            }

        it++;
        }

    if( path.size() > 1 )
        {
        if( slash_terminated )
            {
            if( path[ path.size() - 1 ] != slash )
                {
                path.append( 1, slash );
                }
            }
        else
        if( path[ path.size() - 1 ] == slash )
            {
            path.erase( path.size() - 1 );
            }
        }

        //if( ( !slash_terminated ) && ( path[ path.size() - 1 ] == '/' ) )
        //    {
        //    path.erase( path.size() - 1 );
        //    }
        //}
    }  // BasicVolume::normalize_path()


void BasicVolume::create_directory
    (
    const std::string & path
    )
    {
    // Empty operation.
    }   // BasicVolume::create_directory


void BasicVolume::delete_directory
    (
    const std::string &                 directory_name,
    FileSystem::DirectoryDeleteMode     mode
    )
    {
    // TODO: implement this
    }  // BasicVolume::delete_directory()


void BasicVolume::create_file
    (
    const std::string & path
    )
    {
    std::string full_path = get_current_directory() + path;

    if( &access( full_path, std::ios_base::out, 0 ) == &invalid_stream )
        {
        std::ostringstream ss;
        
        ss << "Can not create stream " << quote_const_string( full_path ) << ".";
        throw ExCanNotCreateFile( ss.str() );
        }
    }


void BasicVolume::delete_file
    (
    const std::string & path
    )
    {
    std::string full_path = get_current_directory() + path;

    if( !remove( full_path ) )
        {
        std::ostringstream ss;
        
        ss << "Can not delete stream " << quote_const_string( full_path ) << ".";
        throw ExCanNotDeleteFile( ss.str() );
        }
    }
        

void BasicVolume::rename_file
    (
    const std::string & from,
    const std::string & to
    )
    {
    // Setup.

    BasicVolume::FStreamDescriptor * src_descriptor = 
        get_descriptor_by_path( from );

    if( !src_descriptor ) 
        {
        // The source file doesn't exist.
        std::ostringstream  ss;
        ss << "File '" << from << "' doesn't exist.";
        throw ExNotAFile( ss.str() );
        }

    if( src_descriptor->num_readers + src_descriptor->num_writers > 0 )
        {
        // The source file is opened - it can't be moved.
        std::ostringstream  ss;
        ss << "File '" << from << "' can't be moved because it is opened.";
        throw ExCanNotRenameFile( ss.str() );
        }

    std::string dest_stream_name;
    std::string dest_stream_path;

    if( is_file( to ) )
        {
        BasicVolume::FStreamDescriptor * dest_descriptor = 
            get_descriptor_by_path( to );

        if( src_descriptor == dest_descriptor )
            {
            return;
                // Copying the stream on itself - empty operation.
            }

        delete_file( to );
        }

    wrench_plain_name( to, dest_stream_path, dest_stream_name );

    std::string src_stream_name;
    std::string src_stream_path;

    wrench_plain_name( from, src_stream_path, src_stream_name );
    
    BasicVolume::FSDHierarchyNode * src_node = 
        get_node_by_path( src_stream_path );

    hope( src_node );
        // src_node must exist because the source file has already been found.

    BasicVolume::FSDHierarchyNode * dest_node = 
        get_node_by_path( dest_stream_path );
    
    if( !dest_node )
        {
        // The destination node doesn't exist.
        std::ostringstream  ss;
        ss << "Moving file: destination node '" << dest_stream_path 
           << "'doesn't exist.";
        throw ExCanNotRenameFile( ss.str() );
        }

    // Rename.
    
    bool existed_flag;
    
    #if 0
    list_descriptors( &root_directory );
    #endif

    BasicVolume::FStreamDescriptor &    dest_descriptor =
        dest_node->insert_stream_descriptor
            (
            dest_stream_name, 
            BasicVolume::FStreamDescriptor( *src_descriptor ), 
            existed_flag 
            );
        // Note that the original stream descriptor is invalidated in the
        // new descriptor copy constructor - see FStreamDescriptor.

    dest_descriptor.set_name( dest_node->get_path() + slash + dest_stream_name );
    
    hope( !existed_flag );
        // If the file had existed, it was already deleted in this routine.

    // Delete the original descriptor.
    src_node->remove_stream_descriptor( src_stream_name );

    // Notify that the state was modified. Must be done explicitly because
    // @c rename_file() is not mapped to any StreamPool operation.
    modified();
    }


bool BasicVolume::is_file
    (
    const std::string & path
    ) const
    {
    BasicVolume * self = const_cast<BasicVolume *>( this );
    std::string full_path = get_current_directory() + path;

    return &self->access( full_path, std::ios_base::in, 0 ) != &invalid_stream;
    }


bool BasicVolume::is_directory
    (
    const std::string & path
    ) const
    {
    BasicVolume * self = const_cast<BasicVolume *>( this );
    std::string full_path = get_current_directory() + path;

    return self->get_node_by_path( full_path, false ) != NULL;
    }


bool BasicVolume::is_absolute
    (
    const std::string & path
    ) const
    {
    return ( path.length() > 0 ) && ( path[ 0 ] == slash );
    }


bool BasicVolume::is_relative
    (
    const std::string & path
    ) const
    {
    return ( path.length() == 0 ) || ( path[ 0 ] != slash  );
    }


char BasicVolume::get_slash() const
    {
    return '/';
    }


char BasicVolume::get_back_slash() const
    {
    return '\\';
    }


BasicVolume::FStreamDescriptor * BasicVolume::get_descriptor_by_path
    (
    const std::string & descriptor_path,
    bool                create_flag
    )
    {
    FStreamDescriptor * result = NULL;
    FSDHierarchyNode *  target_node;

    std::string::size_type last_slash = descriptor_path.find_last_of( "/\\" );

    if( last_slash != std::string::npos )
        {
        target_node = get_node_by_path
            ( 
            descriptor_path.substr( 0, last_slash ? last_slash : 1 ),
            create_flag
            );

        if( target_node )
            {
            result = target_node->get_stream_descriptor
                ( 
                descriptor_path.substr( last_slash + 1 )
                );
            }
        }
    else
        {
        result = current_directory_node->get_stream_descriptor( descriptor_path );
        }

    return result;
    }  // BasicVolume::get_descriptor_by_path()


BasicVolume::FSDHierarchyNode * BasicVolume::get_node_by_path
    (
    const std::string & path,
    bool                create_flag
    )
    {
    FSDHierarchyNode * target_node = current_directory_node;

    if( !path.size() || ( path == "." ) )
        {
        return target_node;
        }

    std::string::size_type  begpos = 0;

    // Set target_node to point to the node the search will be done
    // relative to.
    if( path[ 0 ] == slash )
        {
        // Absolute path has been given

        target_node = &root_directory;
        begpos = 1; // Skip the leading '/' in the next search.
        }
    else
        {
        // Relative path has been given.
        target_node = current_directory_node;
        }

    // Search for the node containing the desired descriptor.
    std::string::size_type  endpos = path.find( slash, begpos );
    std::string             component;
    
    if( endpos == std::string::npos )
        {
        component = path.substr( begpos ); 
                                // Substring from begpos to the end.
        }
    else
        {
        component = path.substr( begpos, endpos - begpos ); 
        }

    while( !component.empty() )
        {
        if( component == ".." )
            {
            target_node = target_node->parent_node;
            }
        else
            {
            FSDHierarchyNode *  aux_node = target_node->get_subnode( component );
            if( aux_node )
                {
                // The subnode named 'component' exists.
                target_node = aux_node;
                }
            else
                {
                // The subnode named 'component' doesn't exist.
                if( create_flag )
                    {
                    aux_node = target_node->create_subnode( component );

                    if( !aux_node )
                        {
                        // Creation of the new subnode failed.
                        return NULL;
                        }
                    else
                        {
                        // Creation of the new subnode succeeded.
                        target_node = aux_node;
                        }
                    }
                else
                    {
                    return NULL;
                    }
                }
            }

        if( endpos == std::string::npos )
            {
            return target_node;
            }

        begpos = endpos;
        begpos++;
        endpos = path.find( slash, begpos );
        
        if( endpos == std::string::npos )
            {
            component = path.substr( begpos );
            }
        else
            {
            component = path.substr( begpos, endpos - begpos );
            }
        } // while( !component.empty() )        

    return target_node;
    }  // BasicVolume::get_node_by_path()


std::string BasicVolume::get_volume_name() const
    {
    return volume_name;
    }


BasicVolume::FStreamDescriptor & BasicVolume::create
    (
    const std::string & stream_name,
    int                 creation_flags
    )
    {
    bool existed_flag;

    // Create new descriptor.
    FStreamDescriptor & f = 
        insert( stream_name, FStreamDescriptor(), existed_flag );

    // Initialize the descriptor if it hadn't existed before.
    if( existed_flag )
        {
        // Stream already registered.
        return f;
        }
    else
        {
        // Initialize new stream.

        StreamPool::create_stream( f );
        f.init( stream_name, creation_flags );
            // Streams are still initialized using full name...

        // Don't increase stream_counter here - it has been already increased
        // in BasicVolume::insert().

        return f;
        }
    }  // BasicVolume::create()


BasicVolume::FStreamDescriptor & BasicVolume::insert
    (
    const std::string &                     stream_name,
    const BasicVolume::FStreamDescriptor &  descriptor,
    bool &                                  existed_flag
    )
    {
    // Find the directory node that will hold the given stream.
    FSDHierarchyNode *  target_node;
    std::string         plain_name;

    std::string::size_type  last_slash = stream_name.find_last_of( "/\\" );

    if( last_slash != std::string::npos )
        {
        target_node = get_node_by_path
            ( 
            stream_name.substr( 0, last_slash ? last_slash : 1 ),
            true
            );

        if( target_node )
            {
            plain_name = stream_name.substr( last_slash + 1 );
            }
        }
    else
        {
        plain_name = stream_name;
        target_node = current_directory_node;
        }

    // Insert the new descriptor.
    FStreamDescriptor & f = target_node->insert_stream_descriptor
        ( 
        plain_name, 
        descriptor, 
        existed_flag 
        );

    if( !existed_flag )
        {
        // Stream didn't exist before.
        stream_counter++;
        }

    return f;
    }  // BasicVolume::insert()


bool BasicVolume::remove
    (
    const std::string & stream_name
    )
    {
    // Find the directory node that will hold the given stream.
    FSDHierarchyNode *  target_node;
    std::string         plain_name;

    std::string::size_type  last_slash = stream_name.find_last_of( "/\\" );

    if( last_slash != std::string::npos )
        {
        target_node = get_node_by_path
            ( 
            stream_name.substr( 0, last_slash ? last_slash : 1 ),
            true
            );

        if( target_node )
            {
            plain_name = stream_name.substr( last_slash + 1 );
            }
        }
    else
        {
        plain_name = stream_name;
        target_node = current_directory_node;
        }

    // Remove the descriptor
    return remove( plain_name, target_node );
    }  // BasicVolume::remove()


bool BasicVolume::remove
    (
    const std::string & plain_name,
    FSDHierarchyNode *  target_node
    )
    {
    FStreamDescriptor * f = target_node->get_stream_descriptor( plain_name );
    if( f )
        {
        StreamPool::delete_stream( *f );
        if( target_node->remove_stream_descriptor( plain_name ) )
            {
            hope( stream_counter > 0 );
            stream_counter--;
            return true;
            }
        }

    return false;
    }  // BasicVolume::remove()


bool BasicVolume::remove
    (
    FileIterator &  it
    )
    {
    if( !it.is_valid() )
        {
        throw ExInvalidFsIterator( "Iterator is invalid." );
        }

    if( !it.is_file() )
        {
        throw ExNotAFile( "Iterator does not reference a file." );
        }

    Reference<IteratorImpl> it_vol = 
        reference_cast<IteratorImpl>( it.get_impl() );

    StreamPool::delete_stream( *it_vol->get_refered_descriptor() );

    if( it_vol->node->remove_stream_descriptor( it_vol->name() ) )
        {
        hope( stream_counter > 0 );
        stream_counter--;
        return true;
        }

    it.unbind();
        // Invalidate the iterator.

    return false;
    }  // BasivVolume::remove()


BasicVolume::FStreamDescriptor & BasicVolume::access
    (
    const std::string &     stream_name,
    std::ios_base::openmode mode,
    int                     creation_flags
    )
    {
    FStreamDescriptor *     result = get_descriptor_by_path( stream_name,
        ( mode & std::ios_base::out ) != 0 );

    if( !( get_open_mode() & std::ios_base::out ) &&
        mode & std::ios_base::out )
        {
        // Volume is mounted read-only.
        return invalid_stream;
        }

    if( result )
        {
        // Stream found.
        return *result;
        }

    // Stream not found.

    if( ( mode & std::ios_base::in ) && !( mode & std::ios_base::out ) )
        {
        // Return invalid stream.
        return invalid_stream;
        }
    else
    if( mode & std::ios_base::out )
        {
        // Create the stream.
        return create( stream_name, creation_flags );
        }

    panic( "Invalid parameters to BasicVolume::access()." );
    return invalid_stream;
    }  // BasicVolume::access()


unsigned BasicVolume::get_stream_count() const
    {
    return stream_counter;
    // todo: remove: return stream_list.size();
    }


void BasicVolume::format
    (
    const std::string & name
    )
    {
    volume_name = name;
    StreamPool::format_pool();
    }


void BasicVolume::freeing_block
    (
    BlockIndex  block
    )
    {
    if( volume_flags & CLEAR_FREE_BLOCKS )
        {
        BlockRef b = access_block( block );

        // Clear block data.

        b.lock();
        memset( b.get_data(), 0, block_size );
        b.unlock( true );
        }
    }


void BasicVolume::adding_block
    (
    BlockIndex  block
    )
    {
    hope( get_open_mode() & std::ios_base::out );

    if( volume_flags & DIRTY_NEW_BLOCKS )
        {
        BlockRef b = access_block( block );

        // Note: Because block has not been mirrored in the image file yet,
        //       load_block() will zero-fill it.

        b.lock();
        b.unlock( true );
        }
    }


std::string BasicVolume::get_pool_name()
    {
    if( volume_name.size() )
        {
        return volume_name;
        }
    else
        {
        return StreamPool::get_pool_name();
        }
    }


void BasicVolume::write_root_stream()
    {
    StreamPool::OStream stream
        (
        get_root_stream(),
        std::ios_base::out | std::ios_base::trunc
        );

    stream.exceptions( std::ios_base::failbit | std::ios_base::badbit );

    // Save volume name.
    stream << quote_const_string( volume_name ) << '\n';

    // Save the total number of streams.
    stream << get_stream_count() << '\n';

    // Save stream list.
    root_directory.save_streams( stream );
    }


void BasicVolume::read_root_stream()
    {
    StreamPool::IStream stream
        (
        get_root_stream(),
        std::ios_base::in
        );

    unsigned num_streams;

    stream.exceptions( std::ios_base::failbit | std::ios_base::badbit );

    // Load volume name.

    stream >> quote_string( volume_name );

    // Load stream list.

    hope( !root_directory.stream_list.size() );
    hope( !get_stream_count() );

    stream >> num_streams;

    while( num_streams-- > 0 )
        {
        FStreamDescriptor   descriptor;
        bool                existed_flag;

        descriptor.restore( this, stream );
        
        insert( descriptor.name, descriptor, existed_flag );
        }
    }


void BasicVolume::read_from_stream
    (
    std::istream &  stream,
    UInt8 *         buffer,
    Int32           offset,
    Int32           length
    )
    {
    stream.clear();
    stream.seekg( 0, std::ios_base::end );
    hope( !stream.fail() );

    if( stream.tellg() >= offset + length )
        {
        // Read data.

        stream.seekg( offset, std::ios_base::beg );
        hope( !stream.fail() );

        stream.read( reinterpret_cast<char *>( buffer ), length );
        hope( !stream.fail() );
        }
    else
        {
        // Reading past end of the stream.
        // Zero-fill buffer.

        while( length-- > 0 )
            {
            *buffer = 0;
            buffer++;
            }
        }
    }


void BasicVolume::write_to_stream
    (
    std::ostream &  stream,
    UInt8 *         buffer,
    Int32           offset,
    Int32           length
    )
    {
    stream.clear();
    stream.seekp( offset, std::ios_base::beg );
    hope( !stream.fail() );

    stream.write( reinterpret_cast<char *>( buffer ), length );
    hope( !stream.fail() );
    }


void BasicVolume::zero_fill_stream
    (
    std::ostream &  stream,
    Int32           offset,
    Int32           length
    )
    {
    static char buffer[ DEFAULT_BLOCK_SIZE ];

    stream.clear();
    stream.seekp( offset, std::ios_base::beg );
    hope( !stream.fail() );

    while( length > 0 )
        {
        stream.write( buffer, length > DEFAULT_BLOCK_SIZE ? DEFAULT_BLOCK_SIZE : length );
        hope( !stream.fail() );

        length -= DEFAULT_BLOCK_SIZE;
        }
    }


void BasicVolume::wrench_plain_name
    (
    const std::string & full_name,
    std::string &       path,
    std::string &       plain_name
    )
    {
    std::string::size_type offs = full_name.find_first_of( "/\\" );

    if ( offs < std::string::npos )
        {
        /* 'slash' found */

        path.insert( 0, full_name, 0, offs );
                // Wrenches the path component. It won't be slash-terminated.

        plain_name.insert( 0, full_name, offs + 1, full_name.length() - offs );
            // Wrenches the plain name component.
        }
    else
        {
        /* path contains only the plain name. */

        plain_name = full_name;
        }
    }


void BasicVolume::list_descriptors
    (
    FSDHierarchyNode *  node
    )
    {
    if( node->stream_list.size() )
        {
        FSDHierarchyNode::StreamList::iterator    it;           

        for( it = node->stream_list.begin(); 
             it != node->stream_list.end(); it++ )
            {
            std::cout << "Stream: " << it->second.name <<
                         " [ " << it->second.current_length << 
                         " ]" << std::endl;
            }
        }
    else
        {
        std::cout << "The node is empty." << std::endl;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
