/** @file basic_volume.h
 * Basic volume.
 */

/*
 * begin        : Fri 28 Jun 2002
 * copyright    : (C) 2002 by Marek Vondrak, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BASIC_VOLUME_H
#define MASSIV_BASIC_VOLUME_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STREAM_POOL_H
#include "filesys/stream_pool.h"
#endif

#ifndef MASSIV_COMPRESSED_STREAM_H
#include "filesys/compressed_stream.h"
#endif

#ifndef MASSIV_FILE_SYSTEM_H
#include "filesys/file_system.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <iostream>
#include <fstream>
#include <string>

namespace Massiv {
namespace Core {


namespace Test  { 

class FileSystemTest; 
class TestVolumeEnumeration;
class VolumeEnumerationTest; 

}


/** @addtogroup filesys
 * @{
 */

/** @class BasicVolume
 * BasicVolume manages streams of data as it does @c StreamPool.
 *
 * Here streams are called file streams.
 * Unlike @c StreamPool each file stream is identified by its name
 * (@c StreamDescriptor is extended to @c FStreamDescriptor)
 * plus can define additional attributes. Root stream stores
 * "main directory" (list of @c FStreamDescriptors).
 * File streams can be (transparently) compressed.
 *
 * @note New operations that are not mapped to StreamPool operations
 *       and that modify BasicVolume state (such as node hierarchy)
 *       must call @c modified() explicitly.
 *       Currently only @c rename_file() does this.
 */
class BasicVolume : public StreamPool, public FileSystem
    {
    friend class Massiv::Core::Test::VolumeEnumerationTest;

public:

    class IFStream;
    class OFStream;
    class FStream;

public:

    //
    // Parameters.
    //

    enum Parameters
        {
        DEFAULT_CACHE_SIZE      =   2048 * 1024
            /**< Block cache size.
                 Note that each open not-flushed (@c flush()) stream
                 occupies at least one block of cache memory. */
        }; // Parameters

    enum Flags
        {
        CLEAR_FREE_BLOCKS       =   1,
            /**< Clear free blocks? */
        DIRTY_NEW_BLOCKS        =   2,
            /**< If BlockPool increases its size, the newly
                 allocated blocks are not mirrored in the image
                 file until they are used.
                 This flag instructs BasicVolume to dirty the new blocks. */
        DEFAULT_VOLUME_FLAGS    =   0
            /**< Default flags. */
        }; // Flags
    
    enum StreamFlags
        {
        STREAM_COMPRESSED       =   1,
            /**< Stream is compressed. */
        DEFAULT_STREAM_FLAGS    =   0
            /**< Default flags. */
        }; // StreamFlags

private:

    //
    // File stream desriptors.
    //

    class IteratorImpl;

    /**
     * This class stores information about file stream.
     */
    class FStreamDescriptor : public StreamPool::StreamDescriptor
        {
        friend class BasicVolume;
        friend class IFStream;
        friend class OFStream;
        friend class FStream;

    public:

        /** Default constructor. */
        FStreamDescriptor();

        /** Initializes FStream part of the descriptor. */
        void init
            (
            const std::string & stream_name,
            int                 stream_flags
            );

    public:

        /**
         */
        virtual void write
            (
            std::ostream & str
            ) const;

        /**
         */
        virtual void restore
            (
            StreamPool *    owner_pool,
            std::istream &  str
            );

    public:
        
        /**
         * @return The name of this descriptor, which is the absolute name,
         *      qualified relatively to the root.
         */
        std::string get_name();

        /**
         * Sets a new ABSOLUTE name for this stream descriptor.
         */
        void set_name
            (
            const std::string & name
            );

    private:

        std::string     name;
            /**< Absolute stream name, qualified relatively to the root
                 node. */
        int             flags;
            /**< Stream flags. */

        }; // FStreamDescriptor

private:

    //
    // File stream descriptor hierarchy.
    //

    /**
     * File Stream Descriptors Hierarchy Node.<br>
     * Each node represents a "directory" in FStreamDescriptors' hierarchy.
     * The node interface accepts plain (parent-relative) names.
     */
    class FSDHierarchyNode
        {
        friend class BasicVolume;
        friend class IteratorImpl;

    public:

        /**
         * Constructor.
         *
         * @param parent_node Node representing the parent directory;
         *      'self' pointer in case of the root node.
         * @param path Path of this node in the volume. Path of the root
         *      directory should be "/" (where '/' is a value of 
         *      get_slash() method).
         *
         * @note You should set the volume that owns this node by set_volume()
         *      later.
         */
        FSDHierarchyNode
            (
            FSDHierarchyNode *  parent_node,
            const std::string & path
            );

        /**
         * Constructor.
         *
         * @param volume The volume that owns this node.
         * @param parent_node Node representing the parent directory;
         *      'self' pointer in case of the root node.
         * @param path Path of this node in the volume. Path of the root
         *      directory should be "/" (where '/' is a value of 
         *      get_slash() method).
         */
        FSDHierarchyNode
            (
            BasicVolume &       volume,
            FSDHierarchyNode *  parent_node,
            const std::string & path
            );

        /** 
         * Destructor. Recursively deletes all subdirectories. 
         */
        ~FSDHierarchyNode();

        NO_DEFAULT_CONSTRUCTOR( FSDHierarchyNode );
        NO_COPY_CONSTRUCTOR( FSDHierarchyNode );
        NO_ASSIGNMENT_OPERATOR( FSDHierarchyNode );

    public:

        typedef ext::hash_map<std::string, FStreamDescriptor>    StreamList;
        typedef ext::hash_map<std::string, FSDHierarchyNode *>   SubnodeList;
        typedef StreamList::iterator    StreamListIterator;
        typedef SubnodeList::iterator   SubnodeListIterator;

    public:

        /**
         * Sets the volume that owns this %FSDHierarchyNode.
         *
         * @param volume The owner volume.
         */
        void set_volume
            (
            BasicVolume &       volume
            );

    public:

        /**
         * @param stream_name Plain name of a descriptor.
         * @return FStreamDescriptor denoted with @c name if there is one
         *      in the list; otherwise NULL.
         */
        FStreamDescriptor * get_stream_descriptor
            (
            const std::string &     stream_name
            ) const;

        /**
         * @param subnode_name Plain name of a subnode.
         * @return Pointer to a subnode denoted with @c name if there is one
         *      present in the list; otherwise NULL.
         */
        FSDHierarchyNode * get_subnode
            (
            const std::string &     subnode_name
            ) const;

        /**
         * Creates new subnode of this node.
         * @param subnode_name Name of the new subnode.
         * @return Pointer to the created subnode.
         * @exception ExCannotCreateDirectory Cannot create subnode.
         *      @c subnode_name has probably wrong syntax.
         */
        FSDHierarchyNode * create_subnode
            (
            const std::string &     subnode_name
            );

        /**
         * Deletes a given subnode of this node.
         * @exception ExCannotDeleteDirectory The specified subnode was
         *      not found.
         * @bug Is it needed to close contained descriptors, etc.?
         * @note Just a helper method!!! Doesn't check whether there
         *      are iterators referencing the deleted node or whether
         *      it has another subnodes or not!!!
         */
        void remove_subnode
            (
            const std::string &     subnode_name
            );

        /**
         * Creates a stream descriptor and stores it in this node.
         * @param name Plain name of the descriptor.
         * @param existed_flag Receives true if the descriptor 'name'
         *      already existed in the node. False otherwise.
         * @return Reference to the created descriptor.
         */
        FStreamDescriptor & create_stream_descriptor
            (
            const std::string &     name,
            bool &                  existed_flag
            );

        /**
         * Inserts a created stream descriptor into this node under the 
         * condition the node doesn't already contain an item of the same
         * name.
         *
         * @param name Plain name of the descriptor (cannot contain '/', '\\'
         *      and cannot be empty).
         * @param descriptor The descriptor to insert.
         * @param existed_flag Receives true if the descriptor already
         *      existed in the node. False otherwise.
         * @return Reference to the inserted descriptor.
         */
        FStreamDescriptor & insert_stream_descriptor
            (
            const std::string &         name,
            const FStreamDescriptor &   descriptor,
            bool &                      existed_flag
            );

        /**
         * Deletes a stream descriptor from this node.
         * @param name Plain name of the descriptor.
         * @return true if the descriptor was present in the node, false
         *      otherwise.
         * @note Doesn't delete the stream itself! This method should be
         *      called only from BasicVolume::remove() to delete both the
         *      descriptor and the stream!
         * @note IMPORTANT!!! Deleting of the last descriptor in the node
         *      can cause deleting of the node itself (if not locked) even
         *      with its map entry in its parent, which can cause deleting
         *      of the parent, and so on. You should not have iterator
         *      referencing subnode that is root node of hierarchy in
         *      which some deleting is being done!!!
         */
        bool remove_stream_descriptor
            (
            const std::string &     name
            );

    protected:

        /**
         * Deletes a stream descriptor from this node.
         * @param it Iterator; should reference the descriptor to be deleted.
         *      Must reference a descriptor within this node!!!
         * @return false if iterator was stream_list.end(), true otherwise.
         * @note Doesn't delete the stream itself!
         * @note IMPORTANT!!! Deleting of the last descriptor in the node
         *      can cause deleting of the node itself (if not locked) even
         *      with its map entry in its parent, which can cause deleting
         *      of the parent, and so on. You should not have iterator
         *      referencing subnode that is root node of hierarchy in
         *      which some deleting is being done!!!
         */
        bool remove_stream_descriptor
            (
            StreamList::iterator &  it
            );

    public:

        /**
         * @return number of streams contained in this node.
         */
        StreamList::size_type stream_count();

        /**
         * @return number of subnodes contained in this node.
         */
        SubnodeList::size_type subnode_count();

        /**
         * @return true if this node is empty, i.e. it contains neither
         *      descriptors nor subnodes.
         */
        bool empty();

    public:

        /**
         * Locks the node in memory. This means that it can't be deleted
         * until unlocked. If the node has been locked three times, it
         * must be three times unlocked until it can be deleted. This is
         * intended for use with iterators.
         */
        void lock();

        /**
         * Unlocks the node. Deletes it if it contains niether subnodes 
         * nor descriptors. Recursively deletes even its ancestors if needed.
         */
        void unlock();

    protected:

        /**
         * Deletes this node if empty. Recursively deletes even its 
         * ancestors.
         */
        void delete_empty();

        /**
         * @return The parent node of this one.
         */
        FSDHierarchyNode * get_parent_node();

        /**
         * Sets the parent node of this one.
         */
        void set_parent_node
            (
            FSDHierarchyNode *  parent_node
            );

        /**
         * @return Path that is represented by this node in the volume.
         */
        std::string get_path();

    protected:
        
        /**
         * Serializes all streams holded in this node and all its subnodes
         * to the stream.
         * @param stream Reference to the target stream.
         * @note Loading from stream is done by BasicVolume itself.
         */
        void save_streams
            (
            StreamPool::OStream &   stream
            );
            
    protected:

        BasicVolume *       volume;
            /**< The volume that owns this node. */
        std::string         path;
            /**< Absolute path of this node in the volume. */
        StreamList          stream_list;
            /**< List of FStreamDescriptors saved in this node.
                 Indexed by @c parent_node relative name. */
        SubnodeList         subnode_list;
            /**< List of nodes - subdirectories of this node.
                 Indexed by @c parent_node relative name. */
        FSDHierarchyNode *  parent_node;
            /**< Parent directory of this node. 'self' pointer for the 
                 root node. */
        int                 lock_counter;
            /**< How many times the node has been locked in memory.
                 The nodes are locked when there is a file iterator
                 pointing to it, for example. */

        }; // FSDHierarchyNode
    
    /**
     * File iterator implementation object for BasicVolume.
     * Actual iterator instances are created through BasicVolume's FileSystem
     * interface.
     * @note Inherited member @c path might not be kept in consistent state
     *      because of efficiency (we don't need it for implementation).
     */
    class IteratorImpl : public FileIteratorImpl
        {
        friend class BasicVolume;
        friend class Massiv::Core::Test::VolumeEnumerationTest;

    protected:
        
        /**
         * Default constructor.
         */
        IteratorImpl();

        /**
         * Constructor.
         * Opens @c volume's current directory.
         */
        IteratorImpl
            (
            BasicVolume &               volume,
            const std::string &         mask,
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         * Full featured constructor.
         * @param volume The volume to bind this iterator with.
         * @param path Specifies the directory of volume to bind this
         *      iterator with.
         * @param mask Mask for filtering of files and directories.
         * @exception ExDirectoryNotFound The specified directory doesn't 
         *      exist.
         */
        IteratorImpl
            (
            BasicVolume &               volume,
            const std::string &         path,
            const std::string &         mask,
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         * Full featured constructor.
         * @param volume The volume to bind this iterator with.
         * @param fsd_node Specifies the directory of volume to bind this
         *      iterator with.
         * @param mask Mask for filtering of files and directories.
         * @exception ExDirectoryNotFound The specified directory doesn't 
         *      exist.
         */
        IteratorImpl
            (
            BasicVolume &               volume,
            FSDHierarchyNode *          fsd_node,
            const std::string &         mask,
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         * Destructor.
         */
        virtual ~IteratorImpl();

    private:

        NO_ASSIGNMENT_OPERATOR( IteratorImpl );

    protected:

        /**
         */
        virtual void bind_current
            (
            const std::string & mask
            );

        /**
         */
        virtual void bind
            (
            const std::string & path,
            const std::string & mask
            );

        /**
         * Binds the iterator with a new directory. If it has already been
         * bound, this old binding is abolished.
         *
         * @param node Node specifying the directory to be open.
         * @param mask Mask describing files that should be enumerated. May
         *      contain wildcards, i.e. * and ?.
         * @exception ExDirectoryNotFound The specified directory doesn't 
         *      exist.
         */
        virtual void bind
            (
            FSDHierarchyNode *  node,
            const std::string & mask
            );

        /**
         */
        virtual void unbind();

        /**
         */
        virtual Reference<FileIteratorImpl> next();

        /**
         */
        virtual Reference<FileIteratorImpl> rewind();

        /**
         */
        virtual Reference<FileIteratorImpl> enter();

        /**
         */
        virtual Reference<FileIteratorImpl> descent();

        /**
         */
        virtual Reference<FileIteratorImpl> leave();

        /**
         */
        virtual Reference<FileIteratorImpl> ascent();

        /**
         */
        virtual bool is_valid() const;

        /**
         */
        virtual bool is_file() const;

        /**
         */
        virtual bool is_directory() const;

        /**
         */
        virtual bool is_trivial_directory() const;

        /**
         */
        virtual std::string name() const;

        /**
         */
        virtual UInt64 size() const;

        /**
         */
        virtual std::string get_absolute_directory() const;

        /**
         */
        virtual bool is_in_root() const;

        /**
         * @return Pointer to FStreamDescriptor refered to by this iterator.
         * @note Iterator must be valid.
         * @exception ExNotAFile The iterator doesn't refer to a file.
         */
        FStreamDescriptor * get_refered_descriptor();

    protected:

        BasicVolume *                           volume;
            /**< Volume this iterator is bound to. */
        FSDHierarchyNode *                      node;
            /**< FSDHierarchyNode this iterator is bound with. */
        bool                                    refers_dir;
            /**< If true, the iterator refers to a directory. Otherwise 
                 refers to a file. Makes sense only if the iterator is
                 valid. */
        FSDHierarchyNode::StreamListIterator    stream_iterator;
            /**< Iterator for stream lists. */
        FSDHierarchyNode::SubnodeListIterator   subnode_iterator;
            /**< Iterator for subnode lists. */

        };  // class IteratorImpl

    friend class FStreamDescriptor;
    friend class FSDHierarchyNode;
    friend class IteratorImpl;

    typedef std::vector<std::string>    PathVector;

public:

    //
    // Raw streams.
    // Return stored data as they were written to the volume.
    //

    typedef StreamPool::IStream         RawIStream;
    typedef StreamPool::OStream         RawOStream;
    typedef StreamPool::IOStream        RawIOStream;

    //
    // Compressed streams.
    // Treat stream data as compressed and do automatic
    // compression/decompression.
    //

    typedef CompressedStream::IStream   CompressedIStream;
    typedef CompressedStream::OStream   CompressedOStream;

    typedef RawIStream                  UncompressedIStream;
    typedef RawOStream                  UncompressedOStream;
    typedef RawIOStream                 UncompressedIOStream;

    //
    // Generic (mutable) streams.
    // Depending on stream flags work like raw streams
    // or compressed streams.
    //

    /**
     * Input file stream.
     */
    class IFStream : public std::istream
        {
    private:

        typedef std::istream    Super;

    public:

        /**
         * Constructor.
         */
        IFStream();

        /**
         * Constructor.
         */
        IFStream
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in
            );

        /**
         * Constructor.
         */
        IFStream
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::in
            );

        /**
         * Destructor.
         */
        ~IFStream();

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::in
            );

    private:

        RawIStream          raw_stream;
            /**< Raw stream. */
        CompressedIStream * compressed_stream;
            /**< Compressed stream. */

        }; // IFStream

    /**
     * Output file stream.
     */
    class OFStream : public std::ostream
        {
    private:

        typedef std::ostream    Super;

    public:

        /**
         * Constructor.
         */
        OFStream();

        /**
         * Constructor.
         */
        OFStream
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::out,
            int                     creation_flags = DEFAULT_STREAM_FLAGS
            );

        /**
         * Constructor.
         */
        OFStream
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::out
            );

        /**
         * Destructor.
         */
        ~OFStream();

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::out,
            int                     creation_flags = DEFAULT_STREAM_FLAGS
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::out
            );

    private:

        RawOStream          raw_stream;
            /**< Raw stream. */
        CompressedOStream * compressed_stream;
            /**< Compressed stream. */

        }; // OFStream

    /**
     * Input & output file stream.
     * Stream can not be compressed.
     */
    class FStream : public std::iostream
        {
    private:

        typedef std::iostream   Super;

    public:

        /**
         * Constructor.
         */
        FStream();

        /**
         * Constructor.
         */
        FStream
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

        /**
         * Constructor.
         */
        FStream
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            BasicVolume &           volume,
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            FStreamDescriptor &     descriptor,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

    private:

        RawIOStream     raw_stream;
            /**< Raw stream. */

        }; // FStream

    friend class IFStream;
    friend class OFStream;
    friend class FStream;

public:

    //
    // BasicVolume construction.
    //

    /**
     * Constructor.
     */
    BasicVolume
        (
        int flags = DEFAULT_VOLUME_FLAGS,
        int block_size = DEFAULT_BLOCK_SIZE,
        int cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Destructor.
     */
    virtual ~BasicVolume();

protected:

    /**
     * Formats volume.
     */
    void format
        (
        const std::string & name
        );

public:
    
    //
    // BasicVolume operations.
    //

    /**
     * Creates file stream.
     */
    FStreamDescriptor & create
        (
        const std::string & stream_name,
        int                 creation_flags = DEFAULT_STREAM_FLAGS
        );

    /**
     * Removes file stream.
     * @param stream_name Fully qualified name of the stream.
     * @return true if stream was removed, false otherwise.
     * @note IMPORTANT!!! Deleting the last descriptor in the node
     *      can cause deleting of the node itself (if not locked) inluding
     *      its map entry in its parent, which can cause deleting
     *      of the parent, and so on. You should not have iterator
     *      referencing subnode that is root node of hierarchy in
     *      which some deleting is being done!!!
     */
    bool remove
        (
        const std::string & stream_name
        );

    /**
     * Removes file stream referenced by the given iterator. The iterator
     * will be invalidated afterwards.
     * @return true if stream was remove, false otherwise.
     * @note IMPORTANT!!! Deleting of the last descriptor in the node
     *      can cause deleting of the node itself (if not locked) even
     *      with its map entry in its parent, which can cause deleting
     *      of the parent, and so on. You should not have iterator
     *      referencing subnode that is root node of hierarchy in
     *      which some deleting is being done!!!
     */
    bool remove
        (
        FileIterator & it
        );

    /**
     * Tries to find stream identified by @c stream_name.
     * If no such stream exists then, according to @c mode and @c creation_flags,
     * new stream will be created or invalid stream will be returned (if the 
     * stream is being opened for input but not for output, the new stream
     * won't be created).
     * Invalid stream will always fail to open (stream will not be @c good()).
     */
    FStreamDescriptor & access
        (
        const std::string &     stream_name,
        std::ios_base::openmode mode,
        int                     creation_flags
        );
        
    /**
     * Get volume name.
     */
    std::string get_volume_name() const;

    /**
     * Get number of file streams stored in the volume.
     */
    unsigned get_stream_count() const;

    /**
     * Equivalent to @c flush_pool().
     * @todo automaticky flush otevrenych streamu?
     */
    void flush();

public:

    //
    // FileSystem interface implementation.
    // Note that the file/directory paths are specified relatively
    // to "current directory".
    //
    
    /**
     */
    virtual FileIterator open_current_directory
        (
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const;

    /**
     */
    virtual FileIterator open_directory
        (
        const std::string &         path,
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const;
    
    /**
     * @return The absolute path to the current directory.
     */
    virtual std::string get_current_directory() const;
    
    /**
     * Sets the current directory. The new directory must exist.
     * @path Specifies the new directory (can be given either as a relative
     *      or as an absolute path.
     * @exception ExCannotChangeDirectory The target directory doesn't exist.
     */
    virtual void set_current_directory
        (
        const std::string & path
        );

    /**
     * Replaces all backslashes in the path with slashes.
     *
     * @param path The path to be normalized.
     * @param slash_terminated If true, ensures that the @c path will be
     *      slash-terminated.
     */
    virtual void normalize_path
        (
        std::string &   path,
        bool            slash_terminated = true
        ) const;

    /**
     * Empty operation because BasicVolume manages directories by "itself".
     * The directory is logically created if it should hold a stream.
     */
    virtual void create_directory
        (
        const std::string & path
        );
    
    /**
     * Removes directory.
     * Returns true if directory was successfuly removed. Note that some 
     * subdirectories may have been removed even if @c directory_name 
     * could not be removed (no atomicity).
     */
    void delete_directory
        (
        const std::string &             directory_name,
        FileSystem::DirectoryDeleteMode mode = FileSystem::DELETE_ALL_RECURSIVE
        );

    /**
     * Alias for @c create().
     */
    virtual void create_file
        (
        const std::string & path
        );

    /**
     * Alias for @c remove().
     */
    virtual void delete_file
        (
        const std::string & path
        );

    /**
     * Renames or moves the file in the volume.
     *
     * @param from Name of the source file, either full or plain.
     * @param to Either name of the destination file (full or plain) or the
     *      destination path.
     *
     * @exception ExNotAFile The source file doesn't exist.
     * @exception ExCanNotRenameFile
     *
     * @note    Simply moves corresponding descriptor from one node to another.
     *          Must call @c modified() because the operation is not mapped to
     *          a StreamPool operation.
     */
    virtual void rename_file
        (
        const std::string & from,
        const std::string & to
        );

    /**
     */
    virtual bool is_file
        (
        const std::string & path
        ) const;
    
    /**
     */
    virtual bool is_directory
        (
        const std::string & path
        ) const;


public:
 
    //
    // Interoperability with native file system maintained by the OS.
    //
    
    /**
     * Unpacks all streams stored in the BasicVolume into the external filesystem.
     * Exports the hierarchy maintained by @c BasicVolume (source tree)
     * into an external one.
     * @param external_path Path to the root of the external tree.
     * @param volume_path Path to the root of the source tree in the 
     *      volume.
     * @param mask Mask for filtering files to unpack.
     * @param erase If true, all unpacked streams will be erased from the
     *      volume.
     *
     * @exception ExCannotOpenFile Some of the destination files or source
     *      streams could not be opened.
     * @exception ExCannotCreateDirectory Some of the directories in the
     *      destination tree could not be created.
     * @exception ExIOError If an IO error occurs.
     */
    void unpack
        (
        const std::string &         external_path,
        const std::string &         volume_path,
        const std::string &         mask,
        bool                        erase = false,
        bool                        recursive = true,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        );

    /**
     * Packs all files matching the @c mask from external file system to
     * Volume.
     * @param volume_path Path to the root of the destination (Volume) tree.
     * @param external_path Path to the root of the source (external) tree.
     * @param mask Mask for filtering files to unpack.
     * @param erase If true, all files being packed will be erased from the
     *      external filesystem. All empty directories that are contained
     *      (even indirectly) in @c external_path will also be deleted,
     *      regardless of whether their names match mask or not (it is
     *      done this way to keep the same semantics as unpack has, i.e.
     *      all empty directories are deleted automaticaly).
     *
     * @exception ExCannotOpenFile Some of the destination files or source
     *      streams could not be opened.
     * @exception ExCannotCreateDirectory Some of the directories in the
     *      destination tree could not be created.
     * @exception ExIOError If an IO error occurs.
     */
    void pack
        (
        const std::string &         volume_path,
        const std::string &         external_path,
        const std::string &         mask,
        int                         creation_flags,
        bool                        erase = false,
        bool                        recursive = true,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        );

    /**
     * Unpacks the directory refered to by the iterator @c it together
     * with all subdirectories. All the files will be unpacked relatively
     * to the current directory (in a 'real' filesystem, not in the Volume).
     * @note @c it should be an iterator over volume.
     */
    void unpack_directory
        (
        FileIterator    it,
        bool            erase,
        bool            recursive
        );

    /**
     * Packs the directory refered to by the iterator @c it (together
     * with all subdirectories, recursively) to the volume. All the files
     * will be packed relatively to the current directory in the volume.
     * @note @c it should be an iterator over 'harrdisk' filesystem.
     */
    void pack_directory
        (
        FileIterator        it,
        const std::string & destination_path,
        int                 creation_flags,
        bool                erase,
        bool                recursive
        );

    /**
     * Copies BasicVolume stream @c stream_name to an external file into
     * @c external path directory.
     */
    void unpack_stream_to_external_file
        (
        const std::string & external_path,
        const std::string & stream_name
        );

    /**
     * Copies external file @c external_name into a directory @c volume_path
     * on BasicVolume.
     * @param creation_flags See Volume and CompactVolume, respectively.
     */
    void pack_external_file_to_stream
        (
        const std::string & external_name,
        const std::string & volume_path,
        int                 creation_flags
        );

public:

    //
    // Implemented from FileSystem.
    //

    /**
     * Tests whether a given @c path is absolute.
     *
     * @param path The path.
     * @return @c true if the @c path is absolute, @c false otherwise.
     */
    virtual bool is_absolute
        (
        const std::string & path
        ) const;

    /**
     * Tests whether a given @c path is relative.
     *
     * @param path The path.
     * @return @c true if the @c path is relative, @c false otherwise.
     */
    virtual bool is_relative
        (
        const std::string & path
        ) const;

    /**
     * Gets the character that is used as a file separator in paths.
     *
     * @return The file separator.
     */
    virtual char get_slash() const;

    /**
     * Gets the reverted character to the one that is used as a file
     * separator in filesystem paths.
     *
     * @return The reverted file separator.
     */
    virtual char get_back_slash() const;

private:
    
    //
    // Private.
    //

    /**
     * Inserts a created stream descriptor into this node.
     * @param name Path and name of the descriptor.
     * @param descriptor The descriptor to insert.
     * @param existed_flag Receives true if the descriptor already
     *      existed in the node. False otherwise.
     * @return Reference to the inserted descriptor.
     */
    FStreamDescriptor & insert
        (
        const std::string &         stream_name,
        const FStreamDescriptor &   descriptor,
        bool &                      existed_flag
        );    
    
    /**
     * Removes file stream.
     * @param plain_name Plain name of the stream.
     * @param node Node that hold the stream.
     * @return true if stream was removed, false otherwise.
     * @note IMPORTANT!!! Deleting of the last descriptor in the node
     *      can cause deleting of the node itself (if not locked) even
     *      with its map entry in its parent, which can cause deleting
     *      of the parent, and so on. You should not have iterator
     *      referencing subnode that is root node of hierarchy in
     *      which some deleting is being done!!!
     */
    bool remove
        (
        const std::string & plain_name,
        FSDHierarchyNode *  node
        );

    /**
     * Returns pointer to a stream descriptor according to the 
     * @c descriptor_path. Path can be given either absolute or
     * relative.
     *
     * @param descriptor_path Path to the descriptor.
     * @param create_flag If true and the descriptor or path doesn't
     *      exist, it will be created.
     *
     * @return Pointer to the descriptor if found; otherwise NULL.
     *
     * @note In case @c descriptor_path is just a plain name without
     *      slashes, the descriptor will be looked after in the
     *      current directory node.
     */
    FStreamDescriptor * get_descriptor_by_path
        (
        const std::string & descriptor_path,
        bool                create_flag = false
        );

    /**
     * Find the directory node given by @c node_path. If some of directories
     * on the path don't exist, they will be created (in case that 
     * @c create_flag is true) or the method will fail.
     * @return Pointer to the node found if the method was sucessful, 
     *      NULL otherwise.
     * @note In case the method failed, some of directory nodes might
     *      stay created and others not.
     * @note Returns NULL if the node_path is empty.
     */
    FSDHierarchyNode * get_node_by_path
        (
        const std::string & node_path,
        bool                create_flag = false
        );

protected:

    //
    // BlockPool interface implementation.
    //

    /**
     */
    virtual void freeing_block
        (
        BlockIndex  block
        );

    /**
     */
    virtual void adding_block
        (
        BlockIndex  block
        );

    /**
     */
    virtual std::string get_pool_name();

protected:

    //
    // StreamPool interface implementation.
    //

    /**
     */
    virtual void write_root_stream();

    /**
     */
    virtual void read_root_stream();

public:

    //
    // Helper functions.
    //

    /**
     * Reads requested portion of @c stream.
     * Zero-fills buffer if reading past end of stream.
     */
    static void read_from_stream
        (
        std::istream &  stream,
        UInt8 *         buffer,
        Int32           offset,
        Int32           length
        );

    /**
     * Writes @c buffer to @c stream.
     */
    static void write_to_stream
        (
        std::ostream &  stream,
        UInt8 *         buffer,
        Int32           offset,
        Int32           length
        );

    /**
     * Zero-fills portion of @c stream.
     */
    static void zero_fill_stream
        (
        std::ostream &  stream,
        Int32           offset,
        Int32           length
        );

    /**
     * Splits the fully-qualified name of the descriptor to the path and the
     * plain descriptor name.
     *
     * @param full_name Fully-qualified descriptor name.
     * @param path      Path to the descriptor node.
     * @param plain_name The plain name of the descriptor.
     */
    static void wrench_plain_name
        (
        const std::string & full_name,
        std::string &       path,
        std::string &       plain_name
        );
    
private:

    /**
     * Auxiliary debugging routine: lists all the descriptor names contained
     * within the FSDHierarchyNode to std::cout.
     */
    void list_descriptors
        (
        FSDHierarchyNode *  node
        );

protected:

    //
    // Volume data.
    //

    FSDHierarchyNode            root_directory;
        /**< Root directory node of the FStreamDescriptors hierarchy. */
    FSDHierarchyNode *          current_directory_node;
        /**< Current directory. Relative paths will be resolved relatively
             to this directory. */
    PathVector                  current_directory_name;
        /**< Name of the current directory. The first component of the
             name is stored in current_directory_name[ 0 ], etc. */
    unsigned                    stream_counter;
        /**< Number of stream descriptors registered with this 
             BasicVolume. */
    std::string                 volume_name;
        /**< Volume name. */
    int                         volume_flags;
        /**< Volume flags. */

private:

    FStreamDescriptor           invalid_stream;
        /**< Invalid stream. This stream will always fail to open. */

private:

    char                        slash;
        /**< The value returned from @c get_slash() method. File separator
             in paths. */
    char                        backslash;
        /**< The value returned from @c get_back_slash() method. Reverted
             file separator in paths. */

    }; // BasicVolume

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_BASIC_VOLUME_H
