/** @file basic_volume_iterator.cpp
 * Implementation of BasicVolume::IteratorImpl.
 */

/*
 * begin        : Sat 01 Mar 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/debug.h"
#include "utility/wildcards.h"
#include "status/exception.h"
#include "filesys/basic_volume.h"
#include <sstream>
#include <stack>
#include <cstring>

namespace Massiv {
namespace Core {


#define CHECK_VALID()  \
    { \
    if( !is_valid() ) \
        { \
        throw ExInvalidFsIterator( "FS iterator is not valid." ); \
        } \
    hope( volume ); \
    hope( node ); \
    }


// Implementation of BasicVolume::IteratorImpl
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BasicVolume::IteratorImpl::IteratorImpl()
    : FileIteratorImpl( FileIterator::MATCH_DIRS_NAMES ),
      volume( NULL ),
      node( NULL ),
      refers_dir( false )
    {
    }  // BasicVolume::IteratorImpl::IteratorImpl()


BasicVolume::IteratorImpl::IteratorImpl
    (
    BasicVolume &                           volume,
    const std::string &                     mask,
    FileSystem::DirectoryEnumerationMode    mode
    ) 
    : FileIteratorImpl( mode ),
      volume( &volume ),
      node( NULL ),
      refers_dir( false )
    {
    bind_current( mask );
    }  // BasicVolume::IteratorImpl::IteratorImpl()


BasicVolume::IteratorImpl::IteratorImpl
    (
    BasicVolume &                           volume,
    const std::string &                     path,
    const std::string &                     mask,
    FileSystem::DirectoryEnumerationMode    mode
    )
    : FileIteratorImpl( mode ),
      volume( &volume ),
      node( NULL ),
      refers_dir( false )
    {
    bind( path, mask );
    }  // BasicVolume::IteratorImpl::IteratorImpl()


BasicVolume::IteratorImpl::IteratorImpl
    (
    BasicVolume &                           volume,
    FSDHierarchyNode *                      fsd_node,
    const std::string &                     mask,
    FileSystem::DirectoryEnumerationMode    mode
    )
    : FileIteratorImpl( mode ),
      volume( &volume ),
      node( NULL ),
      refers_dir( false )
    {
    bind( fsd_node, mask );
    }  // BasicVolume::IteratorImpl::IteratorImpl()


BasicVolume::IteratorImpl::~IteratorImpl()
    {
    if( volume && node )
        {
        unbind();
        }
    }  // BasicVolume::IteratorImpl::~IteratorImpl()


void BasicVolume::IteratorImpl::bind_current
    (
    const std::string &     mask
    )
    {
    hope( volume );
    bind( volume->get_current_directory(), mask );
    }  // BasicVolume::IteratorImpl::bind_current()


void BasicVolume::IteratorImpl::bind
    (
    const std::string &     path,
    const std::string &     mask
    )
    {
    std::string path_normalized( path );
    
    hope( volume );
    volume->normalize_path( path_normalized );

    FSDHierarchyNode * aux_node = volume->get_node_by_path( path_normalized );

    if( !aux_node )
        {
        std::ostringstream  ss;
        ss << "Volume directory '" << path.c_str() << "' not found.";
        throw ExDirectoryNotFound( ss.str() );
        }

    bind( aux_node, mask );
    }  // BasicVolume::IteratorImpl::bind()


void BasicVolume::IteratorImpl::bind
    (
    FSDHierarchyNode *  node,
    const std::string & mask
    )
    {
    hope( volume );

    if( valid_flag && node )
        {
        // Iterator is already bound with some directory.
        unbind();
        }

    // Bind iterator with a new directory.
    this->mask = mask;
    this->node = node;
    valid_flag = false;

    stream_iterator = node->stream_list.begin();
    refers_dir = false;

    while( stream_iterator != node->stream_list.end() )
        {
        if( Wildcards::match( mask.c_str(), stream_iterator->first.c_str() ) )
            {
            valid_flag = true;
            break;
            }

        stream_iterator++;
        }

    if( !valid_flag )
        {
        // No file found to be refered to - try to find directory.

        subnode_iterator = node->subnode_list.begin();
        refers_dir = true;

        while( subnode_iterator != node->subnode_list.end() )
            {
            if( dir_enum_mode == FileIterator::ENUMERATE_ALL_DIRS )
                {
                valid_flag = true;
                break;
                }
            else
            if( Wildcards::match
                    ( 
                    mask.c_str(), 
                    subnode_iterator->first.c_str() 
                    ) )
                {
                valid_flag = true;
                break;
                }

            subnode_iterator++;
            }
        }

    node->lock();
    }  // BasicVolume::IteratorImpl::bind()


void BasicVolume::IteratorImpl::unbind()
    {
    hope( volume && node );
    node->unlock();
    node = NULL;
    valid_flag = false;
    }  // BasicVolume::IteratorImpl::unbind()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::next()
    {
    CHECK_VALID();

    bool increment_subnode_iterator = true;

    valid_flag = false;

    if( !refers_dir )
        {
        stream_iterator++;
        while( stream_iterator != node->stream_list.end() )
            {
            if( Wildcards::match
                    ( 
                    mask.c_str(), 
                    stream_iterator->first.c_str() 
                    ) )
                {
                valid_flag = true;
                break;
                }

            stream_iterator++;
            }

        if( !valid_flag )
            {
            subnode_iterator = node->subnode_list.begin();
            increment_subnode_iterator = false;
            }
        }

    if( !valid_flag )
        {
        // Find a directory to refer to.

        if( increment_subnode_iterator )
            {
            subnode_iterator++;
            }

        refers_dir = true;

        while( subnode_iterator != node->subnode_list.end() )
            {
            if( dir_enum_mode == FileIterator::ENUMERATE_ALL_DIRS )
                {
                valid_flag = true;
                break;
                }
            else
            if( Wildcards::match
                    ( 
                    mask.c_str(), 
                    subnode_iterator->first.c_str() 
                    ) )
                {
                valid_flag = true;
                break;
                }

            subnode_iterator++;
            }
        }

    return this;
    }  // BasicVolume::IteratorImpl::next()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::rewind()
    {
    FSDHierarchyNode * aux_node = node;

    hope( aux_node && volume );

    unbind();
    bind( aux_node, mask );

    return this;
    }  // BasicVolume::IteratorImpl::rewind()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::enter()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when enter() was called." 
            );
        }

    FSDHierarchyNode * fsd = subnode_iterator->second;

    return new IteratorImpl( *volume, fsd, mask );
    }  // BasicVolume::IteratorImpl::enter()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::descent()
    {
    CHECK_VALID();

    if( !is_directory() )
        {
        throw ExNotADirectory
            ( 
            "Iterator didn't point to a directory when descent() was called." 
            );
        }

    FSDHierarchyNode * fsd = subnode_iterator->second;

    unbind();
    bind( fsd, mask );

    return this;
    }  // BasicVolume::IteratorImpl::descent()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::leave()
    {
    CHECK_VALID();

    return new IteratorImpl
        ( 
        *volume, 
        node->get_parent_node(), 
        mask 
        );
    }  // BasicVolume::IteratorImpl::leave()


Reference<FileIteratorImpl> BasicVolume::IteratorImpl::ascent()
    {
    CHECK_VALID();

    FSDHierarchyNode * fsd = node->get_parent_node();

    unbind();
    bind( fsd, mask );

    return this;
    }  // BasicVolume::IteratorImpl::ascent()


bool BasicVolume::IteratorImpl::is_valid() const
    {
    return valid_flag && volume && node;
    }  // BasicVolume::IteratorImpl::is_valid()


bool BasicVolume::IteratorImpl::is_file() const
    {
    CHECK_VALID();
    return is_valid() && !refers_dir;
    }  // BasicVolume::IteratorImpl::is_file()


bool BasicVolume::IteratorImpl::is_directory() const
    {
    CHECK_VALID();
    return is_valid() && refers_dir;
    }  // BasicVolume::IteratorImpl::is_directory()


bool BasicVolume::IteratorImpl::is_trivial_directory() const
    {
    CHECK_VALID();
    return false;
    }  // BasicVolume::IteratorImpl::is_trivial_directory()


std::string BasicVolume::IteratorImpl::name() const
    {
    CHECK_VALID();
    
    hope( volume );
    hope( node );

    if( refers_dir )
        {
        return subnode_iterator->first;
        }
    else
        {
        BasicVolume::FStreamDescriptor * fsd = &stream_iterator->second;
        std::string name( fsd->get_name() );
        std::string::size_type sn = name.find_last_of( "/\\" );
        if( sn == std::string::npos )
            {
            return fsd->get_name();
            }
        else
            {
            return fsd->get_name().substr
                (
                sn < fsd->get_name().size() - 1 ? sn + 1 : sn 
                );
            }
        }
    }  // BasicVolume::IteratorImpl::name()


UInt64 BasicVolume::IteratorImpl::size() const
    {
    CHECK_VALID();

    hope( volume );
    hope( node );

    if( refers_dir )
        {
        // Iterator references a directory
        return 0;
        }
    else
        {
        // Iterator references a file
        BasicVolume::FStreamDescriptor * fsd = &stream_iterator->second;
        return fsd->get_current_length();
        }
    }  // BasicVolume::IteratorImpl::size()


std::string BasicVolume::IteratorImpl::get_absolute_directory() const
    {
    CHECK_VALID();

    return node->get_path();
    }  // BasicVolume::IteratorImpl::get_absolute_directory()


bool BasicVolume::IteratorImpl::is_in_root() const
    {
    hope( volume );
    hope( node );

    return ( node == &volume->root_directory );
    }  // BasicVolume::IteratorImpl::is_in_root()


BasicVolume::FStreamDescriptor *
BasicVolume::IteratorImpl::get_refered_descriptor()
    {
    CHECK_VALID();

    if( !is_file() )
        {
        throw ExNotAFile( "Volume descriptor doesn't refer to a file!" );
        }

    return &stream_iterator->second;
    }  // BasicVolume::IteratorImpl::get_refered_descriptor()


#undef CHECK_VALID


} // namespace Massiv::Core
} // namespace Massiv
