/** @file basic_volume_packing.cpp
 * Basic volume interoperability with the native file system.
 */

/*
 * begin        : Sat 01 Mar 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/basic_volume.h"
#include "global/global.h"
#include <iosfwd>
#include <sstream>
#include <stack>
#include <cstring>

namespace Massiv {
namespace Core {


// Helper routines
// ~~~~~~~~~~~~~~~

static std::string qualify_name
    (
    const std::string & directory_name,
    const std::string & directory_relative_name
    )
    {
    if( directory_name.size() &&
        directory_name[ directory_name.size() - 1 ] != '/' )
        {
        return directory_name + "/" + directory_relative_name;
        }
    else
        {
        return directory_name + directory_relative_name;
        }
    }


static void remove_file
    (
    const std::string &     filename,
    const std::string &     path
    )
    {
    FileSystem & native_fs = Global::native_file_system();
    
    try
        {
        native_fs.delete_file( filename );
        }
    catch( ExCanNotDeleteFile & )
        {  
        // File could not be deleted - log error.

        Global::log
            ( 
            STATUS_WARNING
                ( 
                Status::FACILITY_LOGGER, 
                Status::PRIORITY_LOW
                ),
            "Cannot delete file " + qualify_name( path, filename ) + 
            " while packing directory."
            );
        }  
    } // remove_file()


// BasicVolume implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

void BasicVolume::unpack
    (
    const std::string &         external_path,
    const std::string &         volume_path,
    const std::string &         mask,
    bool                        erase,
    bool                        recursive,
    DirectoryEnumerationMode    mode
    )
    {
    FileSystem &    native_fs = Global::native_file_system();
    bool            unpack_dir = true;
    std::string     old_dir( native_fs.get_current_directory() );
    
    if( ( external_path.size() > 0 ) && ( external_path != "." ) )
        {
        native_fs.set_current_directory( external_path );
        }

    FileIterator it;

    try
        {
        it = open_directory
            ( 
            volume_path, 
            mask, 
            mode 
            );
        }
    catch( Exception & )
        {
        // @c volume_path wasn't a path in the volume - try whether the 
        // last component of @c volume_path couldn't be a file.
        
        std::string::size_type pos = external_path.rfind( '/' );
        
        if( pos == std::string::npos )
            {
            unpack_stream_to_external_file( external_path, "" );
            }
        else
            {
            unpack_stream_to_external_file
                (
                external_path.substr( pos + 1 ),
                external_path.substr( 0, pos )
                );
            }

        if( erase )
            {
            BasicVolume::remove( volume_path );
            }

        unpack_dir = false;
        }

    if( unpack_dir )
        {
        unpack_directory( it, erase, recursive );
        }

    native_fs.set_current_directory( old_dir );
    }  // BasicVolume::unpack()


void BasicVolume::pack
    (
    const std::string &         volume_path,
    const std::string &         external_path,
    const std::string &         mask,
    int                         creation_flags,
    bool                        erase,
    bool                        recursive,
    DirectoryEnumerationMode    mode
    )
    {
    FileSystem &    native_fs = Global::native_file_system();
    std::string     old_dir( native_fs.get_current_directory() );
    bool            pack_dir = true;

    if( !native_fs.is_directory( external_path ) )
        {
        // Try whether the last component of the external_path 
        // couldn't be a file.

        std::string::size_type pos = external_path.rfind( '/' );

        if( pos != std::string::npos )
            {
            native_fs.set_current_directory
                ( 
                external_path.substr( 0, pos ) 
                );

            pos++;
            }
        else
            {
            pos = 0;
            }

        std::string filename( external_path.substr( pos ) );
        pack_external_file_to_stream( filename, volume_path, creation_flags );

        if( erase )
            {
            remove_file( filename, volume_path );
            }

        pack_dir = false;
        }
    
    if( pack_dir )
        {
        try
            {
            native_fs.set_current_directory( external_path );
            }
        catch( Exception & )
            {
            std::ostringstream  ss;
            ss << "Cannot change directory to '" << external_path
               << "'";
            throw ExCanNotChangeDirectory( ss.str() );
            }

        FileIterator it
            ( 
            native_fs.open_current_directory( mask, mode ) 
            );
    
        std::string dest_path( volume_path );

        if( ( dest_path.size() > 1 ) && 
            ( dest_path[ dest_path.size() - 1 ] != '/' ) )
            {
            dest_path.push_back( '/' );
            }

        pack_directory( it, dest_path, creation_flags, erase, recursive );
        }

    native_fs.set_current_directory( old_dir );
    }  // BasicVolume::pack()


void BasicVolume::unpack_directory
    (
    FileIterator    it,
    bool            erase,
    bool            recursive
    )
    {
    FileSystem &                native_fs = Global::native_file_system();
    std::stack< std::string >   to_delete_stack;
        // Remembers names within this directory that will be deleted after
        // unpacking it.
    FSDHierarchyNode *          node;
        // Node that represent this directory.
    
    Reference< IteratorImpl > it_vol;
    it_vol = reference_cast< IteratorImpl >( it.get_impl() );

    node = it_vol->node;
        
    while( it.is_valid() )
        {
        if( recursive && it.is_directory() )
            {
            native_fs.create_directory( it.name() );
            native_fs.set_current_directory( it.name() );
            
            FileIterator it_enter = it.enter();
            
            it++; 
                // We must move @c it to the next element, because deleting
                // descriptors (and thus even nodes, possibly) could cause
                // deleting of the current item.
            
            unpack_directory( it_enter, erase, recursive );

            native_fs.set_current_directory( ".." );
            }
        else
        if( it.is_file() )
            {
            unpack_stream_to_external_file( it.name(), it.get_absolute_directory() );

            if( erase )
                {
                to_delete_stack.push( it.name() );
                }
            
            it++;
            }
        else
            {
            // Ignore this file.
            it++;
            }
        }

    while( !to_delete_stack.empty() )
        {
        remove( to_delete_stack.top(), node );
        to_delete_stack.pop();
        }
    }  // BasicVolume::unpack_directory()


void BasicVolume::pack_directory
    (
    FileIterator        it,
    const std::string & destination_path,
    int                 creation_flags,
    bool                erase,
    bool                recursive
    )
    {
    FileSystem &                native_fs = Global::native_file_system();
    std::stack< std::string >   to_delete_stack;
        // Remembers names within this directory that will be deleted after
        // unpacking it.
    std::string                 absolute_directory;
    
    if( it.is_valid() )
        {
        absolute_directory = it.get_absolute_directory();
        }

    while( it.is_valid() )
        {
        if( recursive && it.is_directory() )
            {
            if( ( it.name() != "." ) && ( it.name() != ".." ) )
                {
                native_fs.set_current_directory( it.name() );

                pack_directory
                    ( 
                    it.enter(), 
                    std::string( destination_path + it.name() + '/' ),
                    creation_flags,
                    erase,
                    recursive
                    );

                native_fs.set_current_directory( ".." );

                if( erase )
                    {
                    // Delete the directory if empty
                    try
                        {
                        native_fs.delete_directory
                            ( 
                            it.name(), 
                            FileSystem::DELETE_DIRS_RECURSIVE
                            );
                        }
                    catch( ExDirectoryNotEmpty & )
                        {
                        }
                    }
                }
            }
        else
        if( it.is_file() )
            {
            pack_external_file_to_stream
                ( it.name(), destination_path, creation_flags );

            if( erase )
                {
                to_delete_stack.push( it.name() );
                }
            }

        it++;
        }

    if( erase )
        {
        // Delete all files matching the given mask
        while( !to_delete_stack.empty() )
            {
            remove_file( to_delete_stack.top(), absolute_directory );
            to_delete_stack.pop();
            }  // while
        }
    }  // BasicVolume::pack_directory()


void BasicVolume::unpack_stream_to_external_file
    (
    const std::string &             name,
    const std::string &             external_path
    )
    {
    std::ofstream output
        ( 
        name.c_str(), 
        std::ios_base::out | std::ios_base::binary 
        );

    if( !output.good() )
        {
        std::ostringstream  ss;
        ss << "Cannot open file " << name << " for writing.";
        throw ExCanNotOpenFile( ss.str() );
        }

    std::string path = qualify_name( external_path, name );

    BasicVolume::IFStream input
        (
        *this,
        path,
        std::ios_base::in | std::ios_base::binary
        );
    
    if( !input.good() )
        {
        std::ostringstream  ss;
        ss << "Cannot open stream " << name << " for reading.";
        throw ExCanNotOpenFile( ss.str() );
        }

    // Test whether input is not empty; if input was empty, after processing
    // of line 'output << input.rdbuf()' would output.bad() be true and we'd
    // get an error message.
    // FIXME: how to handle this situation in a better way?

    char   c;
    input.get( c );

    if( !input.eof() )
        {
        input.putback( c );
        output << input.rdbuf();

        if( output.fail() || input.fail() )
            {
            std::ostringstream  ss;

            ss << "Error copying " << path << " to " << name << ".";
            throw ExIOError( ss.str() );
            }
        }
    }  // BasicVolume::unpack_stream_to_external_file()


void BasicVolume::pack_external_file_to_stream
    (
    const std::string &             external_name,
    const std::string &             volume_path,
    int                             creation_flags
    )
    {
    std::ifstream input
        (
        external_name.c_str(), 
        std::ios_base::in | std::ios_base::binary 
        );

    if( !input.good() )
        {
        std::ostringstream  ss;
        ss << "Cannot open stream " << external_name << " for reading.";
        throw ExCanNotOpenFile( ss.str() );
        }

    std::string target = qualify_name( volume_path, external_name );

    BasicVolume::OFStream output
        (
        *this,
        target,
        std::ios_base::out | std::ios_base::binary,
        creation_flags
        );

    if( !output.good() )
        {
        std::ostringstream  ss;
        ss << "Cannot open file " << external_name << " for writing.";
        throw ExCanNotOpenFile( ss.str() );
        }

    // Test whether input is not empty; if input was empty, after processing
    // of line 'output << input.rdbuf()' would output.bad() be true and we'd
    // get an error message.
    // FIXME: how to handle this situation in a better way?

    char   c;
    input.get( c );

    if( !input.eof() )
        {
        input.putback( c );
        output << input.rdbuf();

        if( !output.good() || !input.good() )
            {
            std::ostringstream  ss;

            ss << "Error copying " << external_name << " to " << target << ".";
            throw ExIOError( ss.str() );
            }
        }
    }  // BasicVolume::pack_external_file_to_stream()


} // namespace Massiv::Core
} // namespace Massiv
