/** @file block_pool.cpp
 * Block pool.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/block_pool.h"
#include "status/exception.h"
#include "global/global.h"
#include <vector>
#include <sstream>

namespace Massiv {
namespace Core {

namespace {

const Int64     MAGIC_ID = UINT64( 0xd1e2a3d4b5e6e7f8 );
    /**< Block pool ID. */
const Int32     BLOCK_POOL_VERSION = 0x101;
    /**< Block pool version. */
const Int32     SYNCED_FLAG_OFFSET = 16;
    /**< Offset of synced_flag within the master block. */

}
    

// Implementation of BlockPool::BlockMapEntry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BlockPool::BlockMapEntry::BlockMapEntry()
    {
    next = INVALID_BLOCK;
    prev = INVALID_BLOCK;
    }


// Implementation of BlockPool
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

BlockPool::BlockPool
    (
    int     block_size,
    int     cache_size
    ) : SimpleBlockPool( block_size, cache_size )
    {
    mounted = false;
    synced = true;

    num_blocks = 0;
    first_free_block = INVALID_BLOCK;
    last_free_block = INVALID_BLOCK;
    num_free_blocks = 0;

    block_map_chain = INVALID_BLOCK;

    /* The last entry in each block is reserved for internal purposes */
    /* (link in block map chain). */

    block_map_entries_per_block = block_size / BLOCK_MAP_ENTRY_SIZE - 1;
    }


void BlockPool::restore_state()
    {
    load_state();

    #ifdef OPTION_DEBUG

    test_block_map_integrity();

    #endif

    mounted = true;
    }


void BlockPool::format_pool()
    {
    hope( !block_map.size() );

    /* Create master block chain. */

    create_chain( 1 );

    mounted = true;
    }


void BlockPool::flush_pool()
    {
    /* Flush state unless it is already synced. */

    if( can_write() && !is_synced() )
        {
        /* Two-phase flush. First externalize pool state and flush all the
           blocks. Then reopen master block and write that the pool was
           synced properly. */

        save_state();
        SimpleBlockPool::flush_pool();
        
        write_synced_flag( true );
        SimpleBlockPool::flush_pool();
        }
    }


void BlockPool::close_pool()
    {
    flush_pool();
    SimpleBlockPool::close_pool();
    }


bool BlockPool::can_write() const
    {
    if( !SimpleBlockPool::can_write() )
        {
        return false;
        }
    else
        {
        return mounted;
        }
    }


void BlockPool::block_modified
    (
    BlockIndex index
    )
    {
    SimpleBlockPool::block_modified( index );

    if( index == MASTER_BLOCK_INDEX )
        {
        /* Ignore writes to the master block because they are always initiated
           the BlockPool itself. This also avoids infinite loop while calling
           @c write_synced_flag(). */

        return;
        }

    modified();
    }


void BlockPool::modified()
    {
    if( is_synced() )
        {
        /* We are not synced any more until the pool is flushed. */

        if( can_write() )
            {
            write_synced_flag( false );
            }
        else
            {
            synced = false;
            }
        }
    }


bool BlockPool::is_synced()
    {
    return synced;
    }


void BlockPool::save_master_block_extension
    (
    BlockRef & block
    )
    {
    }


void BlockPool::load_master_block_extension
    (
    BlockRef & block
    )
    {
    }


void BlockPool::freeing_block
    (
    BlockIndex  block
    )
    {
    }


void BlockPool::adding_block
    (
    BlockIndex  block
    )
    {
    }


std::string BlockPool::get_pool_name()
    {
    std::ostringstream ss;

    ss << "Pool$" << reinterpret_cast<UInt32>( this );
    return ss.str();
    }


//
// Block chain management.
//


BlockPool::BlockIndex BlockPool::get_block
    (
    BlockIndex  start,
    int         index_in_chain
    )
    {
    try
        {
        if( index_in_chain > 0 )
            {
            while( index_in_chain > 0 )
                {
                start = block_map[ start ].next;
                index_in_chain--;
                }
            }
        else
            {
            while( index_in_chain < 0 )
                {
                start = block_map[ start ].prev;
                index_in_chain++;
                }
            }

        return start;
        }
    catch( ExOutOfRange & )
        {
        throw ExIntegrityError( "Corrupted block map (get_block)." );
        }
    }


BlockPool::BlockIndex BlockPool::get_next_block
    (
    BlockIndex  chain
    )
    {
    return get_block( chain, 1 );
    }


BlockPool::BlockIndex BlockPool::get_prev_block
    (
    BlockIndex  chain
    )
    {
    return get_block( chain, -1 );
    }


BlockPool::BlockIndex BlockPool::grow_pool
    (
    int growth
    )
    {
    BlockIndex  first;
    int         i;

    hope( growth > 0 );
    modified();

    first = block_map.size();
    block_map.resize( first + growth );

    adding_block( first );

    for( i = first + 1; i < first + growth; i++ )
        {
        block_map[ i ].prev = i - 1;
        block_map[ i - 1 ].next = i;

        adding_block( i );
        }

    num_blocks += growth;

    return first;
    }


void BlockPool::erase_chain
    (
    BlockIndex  block
    )
    {
    BlockIndex  prev;
    BlockIndex  last;
    int         size;

    modified();

    try
        {
        /* Cut off block chain tail. */

        prev = block_map[ block ].prev;

        if( prev != INVALID_BLOCK )
            {
            block_map[ prev ].next = INVALID_BLOCK;
            block_map[ block ].prev = INVALID_BLOCK;
            }

        /* Free blocks. */

        last = block;
        size = 1;

        freeing_block( last );

        while( block_map[ last ].next != INVALID_BLOCK )
            {
            last = block_map[ last ].next;
            freeing_block( last );

            size++;
            }

        /* Link block chain tail to free block chain. */

        if( last_free_block != INVALID_BLOCK )
            {
            block_map[ last_free_block ].next = block;
            block_map[ block ].prev = last_free_block;

            last_free_block = last;
            }
        else
            {
            first_free_block = block;
            last_free_block = last;
            }

        num_free_blocks += size;

        hope( block_map[ first_free_block ].prev == INVALID_BLOCK );
        hope( first_free_block != INVALID_BLOCK && last_free_block != INVALID_BLOCK );
        }
    catch( ExOutOfRange & )
        {
        throw ExIntegrityError( "Corrupted block map (erase_chain)." );
        }
    }


BlockPool::BlockIndex BlockPool::create_chain
    (
    int size
    )
    {
    BlockIndex  first;

    hope( size > 0 );
    modified();

    if( !num_free_blocks )
        {
        first = grow_pool( size );
        }
    else
    if( num_free_blocks < size )
        {
        BlockIndex  new_chain;

        /* Grow pool and use existing free block chain. */

        first = first_free_block;
        new_chain = grow_pool( size - num_free_blocks );

        try
            {
            hope( block_map[ last_free_block ].next == INVALID_BLOCK );

            block_map[ last_free_block ].next = new_chain;
            block_map[ new_chain ].prev = last_free_block;

            first_free_block = INVALID_BLOCK;
            last_free_block = INVALID_BLOCK;
            num_free_blocks = 0;
            }
        catch( ... )
            {
            erase_chain( new_chain );
            throw;
            }
        }
    else
        {
        BlockIndex  new_first_free;
        BlockIndex  prev;
        int         i;

        /* Allocate chain. */

        new_first_free = first = first_free_block;
        prev = INVALID_BLOCK;

        try
            {
            for( i = 0; i < size; i++)
                {
                prev = new_first_free;
                new_first_free = block_map[ new_first_free ].next;
                }

            if( new_first_free != INVALID_BLOCK )
                {
                block_map[ new_first_free ].prev = INVALID_BLOCK;
                }

            block_map[ prev ].next = INVALID_BLOCK;
            }
        catch( ExOutOfRange & )
            {
            throw ExIntegrityError( "Corrupted block map (create_chain)." );
            }

        first_free_block = new_first_free;
        num_free_blocks -= size;

        if( !num_free_blocks )
            {
            hope( first_free_block == INVALID_BLOCK );
            last_free_block = INVALID_BLOCK;
            }
        }

    hope( block_map[ first ].prev == INVALID_BLOCK );

    #ifdef OPTION_DEBUG

    hope( get_chain_length( first ) == size );

    #endif

    return first;
    }


void BlockPool::link_chains
    (
    BlockIndex  chain,
    BlockIndex  head
    )
    {
    modified();

    try
        {
        hope( block_map[ head ].prev == INVALID_BLOCK );

        while( block_map[ chain ].next != INVALID_BLOCK )
            {
            chain = block_map[ chain ].next;
            }

        hope( chain != last_free_block );

        block_map[ chain ].next = head;
        block_map[ head ].prev = chain;
        }
    catch( ExOutOfRange & )
        {
        throw ExIntegrityError( "Corrupted block map (link_chains)." );
        }
    }


void BlockPool::append_blocks
    (
    BlockIndex  chain,
    int         size
    )
    {
    BlockIndex  new_chain;

    modified();
    new_chain = create_chain( size );

    try
        {
        link_chains( chain, new_chain );
        }
    catch( ... )
        {
        erase_chain( new_chain );
        throw;
        }
    }


int BlockPool::get_chain_length
    (
    BlockIndex  head
    )
    {
    try
        {
        int length;

        hope( block_map[ head ].prev == INVALID_BLOCK );
        length = 0;

        while( head != INVALID_BLOCK )
            {
            head = block_map[ head ].next;
            length++;
            }

        return length;
        }
    catch( ExOutOfRange & )
        {
        throw ExIntegrityError( "Corrupted block map (get_chain_length)." );
        }
    }


void BlockPool::test_block_map_integrity()
    {
    std::vector<bool>   visited_blocks;
    BlockIndex          block;
    bool                free_block_chain_ok;

    visited_blocks.resize( num_blocks, false );
    free_block_chain_ok = false;

    try
        {
        for( block = 0; block < num_blocks; block++ )
            {
            BlockIndex  prev;
            BlockIndex  first, last;
            int         backward_length, forward_length;
            int         total_length;

            /* Examine @c block. */

            if( visited_blocks[ block ] )
                continue;

            /* Examine block chain containing @c block (go backward). */

            visited_blocks[ block ] = true;

            first = block;
            prev = INVALID_BLOCK;
            backward_length = 0;

            while( true )
                {
                prev = first;
                first = block_map[ first ].prev;

                if( first == INVALID_BLOCK )
                    break;

                if( visited_blocks[ first ] )
                    {
                    throw ExIntegrityError( "Cross-linked block chains." );
                    }

                if( block_map[ first ].next != prev )
                    {
                    throw ExIntegrityError( "Inconsistent links in block chain." );
                    }

                visited_blocks[ first ] = true;
                backward_length++;
                }

            first = prev;

            /* Examine block chain containing @c block (go forward). */

            last= block;
            prev = INVALID_BLOCK;
            forward_length = 0;

            while( true )
                {
                prev = last;
                last = block_map[ last ].next;

                if( last == INVALID_BLOCK )
                    break;

                if( visited_blocks[ last ] )
                    {
                    throw ExIntegrityError( "Cross-linked block chains." );
                    }

                if( block_map[ last ].prev != prev )
                    {
                    throw ExIntegrityError( "Inconsistent links in block chain." );
                    }

                visited_blocks[ last ] = true;
                forward_length++;
                }

            last = prev;

            /* Inspect chain length. */

            total_length = backward_length + forward_length + 1;

            if( first == first_free_block )
                {
                if( last != last_free_block || num_free_blocks != total_length )
                    {
                    throw ExIntegrityError( "Invalid free block chain." );
                    }

                free_block_chain_ok = true;
                }
            }

        if( !free_block_chain_ok && first_free_block != INVALID_BLOCK )
            {
            throw ExIntegrityError( "Free block chain corrupted." );
            }
        }
    catch( ExOutOfRange & )
        {
        throw ExIntegrityError( "Corrupted block map (test_integrity)." );
        }
    }


//
// BlockPool state serialization.
//


int BlockPool::get_num_block_map_blocks()
    {
    return ( num_blocks + block_map_entries_per_block - 1 ) /
            block_map_entries_per_block;
    }


void BlockPool::save_block_map()
    {
    BlockRef            block;
    BlockIndex          new_block_map;
    BlockIndex          block_map_block;
    int                 num_block_map_blocks;
    int                 old_num_blocks;
    int                 num_entries, i, j;
    BlockMap::Iterator  it;

    /* Erase old block map. */

    if( block_map_chain != INVALID_BLOCK )
        {
        erase_chain( block_map_chain );
        block_map_chain = INVALID_BLOCK;
        }

    /* Allocate new block map. */

    do
        {
        old_num_blocks = num_blocks;
        num_block_map_blocks = get_num_block_map_blocks();

        new_block_map = create_chain( num_block_map_blocks );

        if( num_blocks != old_num_blocks )
            {
            /* create_chain() caused block pool to expand => try again. */

            erase_chain( new_block_map );
            }
        }
    while( num_blocks != old_num_blocks );

    /* Save new block map. */

    block_map_block = new_block_map;
    it = block_map.begin();

    for( i = 0; i < num_block_map_blocks; i++ )
        {
        /* Determine number of entries in block. */

        num_entries = block_map_entries_per_block;

        if( i * block_map_entries_per_block + num_entries > num_blocks )
            {
            num_entries = num_blocks - i * block_map_entries_per_block;
            }

        /* Write block of block map. */

        block = access_block( block_map_block );
        block.lock();

        for( j = 0; j < num_entries; j++ )
            {
            /* Write block map entry. */

            block << ( *it ).prev;
            block << ( *it ).next;

            it++;
            }

        /* Write reserved block map entry. */

        block << block_map[ block_map_block ].prev;
        block << block_map[ block_map_block ].next;

        block.unlock( true );

        /* Next block of block map. */

        block_map_block = block_map[ block_map_block ].next;
        }

    /* Done. */

    block_map_chain = new_block_map;
    }


void BlockPool::load_block_map()
    {
    BlockIndex  block_map_block;
    BlockRef    block;
    int         num_block_map_blocks;
    int         i, j, num_entries;

    block_map.clear();

    num_block_map_blocks = get_num_block_map_blocks();
    block_map_block = block_map_chain;

    for( i = 0; i < num_block_map_blocks; i++ )
        {
        BlockMapEntry   entry;

        /* Determine number of entries in block. */

        num_entries = block_map_entries_per_block;

        if( ( int ) block_map.size() + num_entries > num_blocks )
            {
            num_entries = num_blocks - block_map.size();
            }

        /* Read block of block map. */

        block = access_block( block_map_block );
        block.lock();

        for( j = 0; j < num_entries; j++ )
            {
            /* Read block map entry. */

            block >> entry.prev;
            block >> entry.next;

            block_map.push_back( entry );
            }

        /* Read reserved block map entry. */

        block >> entry.prev;
        block >> entry.next;

        block.unlock( false );

        /* Next iteration. */

        block_map_block = entry.next;
        }
    }


void BlockPool::save_state()
    {
    BlockRef block;

    /* Save block map. */

    save_block_map();

    /* Write signature. */

    block = access_block( MASTER_BLOCK_INDEX );
    block.lock();

    block << MAGIC_ID;
    block << BLOCK_POOL_VERSION;
    block << block_size;
    
    /* Write that block pool was not synced properly. This will be updated
       once the pool has been successfully flushed. */

    hope( block.tellp() == SYNCED_FLAG_OFFSET );
    block << static_cast<Int32>( 0 );

    /* Write master block. */

    block << num_blocks;
    block << block_map_chain;
    block << first_free_block;
    block << last_free_block;
    block << num_free_blocks;

    /* Save master block extension. */

    save_master_block_extension( block );

    /* Done. */

    block.unlock( true );
    }


void BlockPool::write_synced_flag
    (
    bool synced_flag
    )
    {
    BlockRef block;
    
    /* Write synced flag. */

    synced = synced_flag;
    block = access_block( MASTER_BLOCK_INDEX );
    block.lock();
    
    block.seekp( SYNCED_FLAG_OFFSET, std::ios_base::beg );
    block << static_cast<Int32>( synced );

    block.unlock( true, true );

    /* Log. */

    #ifdef OPTION_DEBUG
    
    Global::log_info
        (
        Status::FACILITY_FILE_SYSTEM,
        Status::PRIORITY_NORMAL,
        get_pool_name() + ( synced_flag ? " has been synced." : " is dirty." )
        );
    
    #endif
    }


void BlockPool::load_state()
    {
    BlockRef    block;
    Int64       id;
    Int32       version;
    Int32       saved_block_size;
    Int32       synced_flag;

    /* Verify signature. */

    block = access_block( MASTER_BLOCK_INDEX );
    block.lock();

    block >> id;
    block >> version;
    block >> saved_block_size;

    if( id != MAGIC_ID )
        {
        block.unlock( false );
        throw ExUnrecognizedImageFile( "Unrecognized image file." );
        }

    if( version != BLOCK_POOL_VERSION )
        {
        block.unlock( false );
        std::ostringstream ss;
        ss << "Unrecognized image file version. Expected " << BLOCK_POOL_VERSION <<
            ", found " << version << ".";
        throw ExUnrecognizedImageFile( ss.str() );
        }

    if( block_size != saved_block_size )
        {
        block.unlock( false );
        std::ostringstream ss;
        ss << "The image has different block size settings. Expected " <<
            block_size << ", found " << saved_block_size << ".";
        throw ExUnrecognizedImageFile( ss.str() );
        }

    /* Test if pool was synced properly. */

    hope( block.tellg() == SYNCED_FLAG_OFFSET );
    block >> synced_flag;

    if( !synced_flag )
        {
        throw ExIntegrityError
            (
            get_pool_name() + ": "
            "The image was not synced properly. "
            "To prevent file system inconsistencies and/or random crashes "
            "the pool will not be opened."
            );
        }

    /* Read master block. */

    block >> num_blocks;
    block >> block_map_chain;
    block >> first_free_block;
    block >> last_free_block;
    block >> num_free_blocks;

    /* Load block map. */

    load_block_map();

    /* Read master block extension. */

    load_master_block_extension( block );

    block.unlock( false );
    }


} // namespace Massiv::Core
} // namespace Massiv
