/** @file block_pool.h
 * Block pool.
 */

/*
 * begin        : Fri 28 Jun 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BLOCK_POOL_H
#define MASSIV_BLOCK_POOL_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SIMPLE_BLOCK_POOL_H
#include "filesys/simple_block_pool.h"
#endif

#ifndef MASSIV_ARRAY_H
#include "utility/array.h"
#endif


namespace Massiv {
namespace Core {


namespace Test {
    class FileSystemTest;
}


/** @addtogroup filesys
 * @{
 */

/** @class BlockPool
 * BlockPool is file-system-like block-based structure.
 * In contrast to SimpleBlockPool, BlockPool manages allocation
 * of block chains (block chain is a list of linked blocks).
 * Besides blocks that hold client-supplied data, there are
 * system-managed helper blocks (such as master block or block map)
 * where internal BlockPool state is stored.
 *
 * Internal blocks organization ("image structure"):
 *  - master block:
 *      pointer to block map
 *      free block chain pointers
 *      [ optional client-defined master block extension ]
 *  - block map ("cluster map", allocation table):
 *      for each block of the "image":
 *          block map entry defining block chain links
 *      block map may be spanned among multiple blocks
 *  - client data
 */
class BlockPool : public SimpleBlockPool
    {
public:

    friend class Test::FileSystemTest;

    enum Parameters
        {
        DEFAULT_BLOCK_SIZE      = 4096
            /**< Default block size. */
        }; // Parameters

    enum Constants
        {
        MASTER_BLOCK_INDEX      = 0,
            /**< Block index of master block. */
        BLOCK_MAP_ENTRY_SIZE    = 2 * sizeof( Int32 ),
            /**< How many bytes is required to serialize BlockMapEntry. */
        }; // Constants

private:

    /**
     * BlockMapEntry.
     */
    struct BlockMapEntry
        {
    public:

        /**
         * Default constructor.
         */
        BlockMapEntry();

    public:

        BlockIndex  next;
            /**< Next block in chain. */
        BlockIndex  prev;
            /**< Previous block in chain. */
        }; // BlockMapEntry


    typedef NormalizedArray< DynamicStorage<BlockMapEntry> > BlockMap;

protected:

    //
    // Main.
    //

    /**
     * Creates empty block pool.
     * Most-derived class is supposed to call either @c restore_state()
     * or @c format().
     */
    BlockPool
        (
        int     block_size = DEFAULT_BLOCK_SIZE,
        int     cache_size = DEFAULT_BLOCK_SIZE * 16
        );

    /**
     * Restores internal block pool state to the state stored in the "image".
     * Can only be called from the most derived-class just *after* construction
     * of the object.
     */
    virtual void restore_state();

    /**
     * Initializes block pool to default.
     * Can only be called from the most derived-class just *after* construction
     * of the object.
     */
    void format_pool();

protected:

    //
    // SimpleBlockPool Extensions.
    //

    /**
     * Flushes block pool state.
     * Override this to flush your metadata unless @c is_synced() already.
     */
    virtual void flush_pool();

    /**
     */
    virtual void close_pool();

    /**
     * BlockPool is read-only until it is successfully mounted
     * (@c restore_state() or @c format() completes without an error).
     */
    virtual bool can_write() const;
    
    /**
     * Called by the caching mechanism when a dirty block is unlocked.
     * This means that externalized state has not been synchronized with the
     * pool state yet.
     */
    virtual void block_modified
        (
        BlockIndex index
        );
    
public:

    /**
     * To be called by BlockPool and its subclasses when object state
     * "metadata" (stored in the memory, not accessed through the cache)
     * is modified.
     *
     * When changing data stored directly in the cache @c modified() is called
     * automatically by the caching mechanism.
     *
     * This is needed in order to optimize out unneccessary pool flushes.
     */
    virtual void modified();

protected:
    
    /**
     * Tests if pool state is synchronized with an externalized state.
     */
    bool is_synced();

    /**
     * Called by BlockPool when it saves master block.
     * This may be used by "client" to store additional data
     * into master block ("master block extension").
     * @c block is already locked and writer pointer points
     * to the right place.
     * Note that size of master block extension plus size of
     * master block is limited to the size of block.
     * Method can access @c block only and must not modify
     * BlockPool state!
     */
    virtual void save_master_block_extension
        (
        BlockRef & block
        );

    /**
     * Called by BlockPool when it loads master block.
     * This may be used to retrieve "client"-stored
     * master block extension.
     * @c block is already locked and reader pointer points
     * to the right place.
     * BlockPool state is restored already.
     */
    virtual void load_master_block_extension
        (
        BlockRef & block
        );
    
    /**
     * Called by BlockPool when it frees @c block.
     */
    virtual void freeing_block
        (
        BlockIndex  block
        );

    /**
     * Called by BlockPool when new @c block was added to the pool.
     */
    virtual void adding_block
        (
        BlockIndex  block
        );

    /**
     * Returns pool name that will be used in debugging dumps.
     * To be overriden by a subclass.
     */
    virtual std::string get_pool_name();

protected:

    //
    // Block manipulation.
    //

    /**
     * Returns index of @c index_in_chain-th block in block chain
     * for block @c start.
     * If @c index_in_chain is negative then the chain is being
     * walked backward.
     */
    BlockIndex get_block
        (
        BlockIndex  start,
        int         index_in_chain
        );

    /**
     * Returns index of next block in block chain @c chain.
     * @c chain need not be the first block of the chain.
     * Returns INVALID_BLOCK if no such block exists.
     */
    BlockIndex get_next_block
        (
        BlockIndex  chain
        );

    /**
     * Returns index of previous block in block chain @c chain.
     * Returns INVALID_BLOCK if no such block exists.
     */
    BlockIndex get_prev_block
        (
        BlockIndex  chain
        );

    /**
     * Expands block pool by @c growth blocks (does not try to
     * allocate block chain from free blocks).
     * Returns pointer to first block in newly created block chain.
     */
    BlockIndex grow_pool
        (
        int growth
        );

    /**
     * Erases block chain.
     * @c block does not have to point to the first block of the
     * corresponding block chain, in that case it truncates the chain.
     */
    void erase_chain
        (
        BlockIndex  block
        );

    /**
     * Creates block chain of @c size blocks.
     * Returns pointer to first block in newly created block chain.
     */
    BlockIndex create_chain
        (
        int size
        );

    /**
     * Links block chain @c chain with head of another chain (@c head).
     */
    void link_chains
        (
        BlockIndex  chain,
        BlockIndex  head
        );

    /**
     * Appends free blocks to an existing block chain.
     */
    void append_blocks
        (
        BlockIndex  chain,
        int         size
        );

    /**
     * Returns current length of block chain beginning at @c head.
     */
    int get_chain_length
        (
        BlockIndex  head
        );

    /**
     * Tests block map integrity.
     */
    void test_block_map_integrity();

private:

    //
    // Private.
    //

    /**
     * Returns number of blocks required to store block map.
     */
    int get_num_block_map_blocks();

    /**
     * Saves block map.
     */
    void save_block_map();

    /**
     * Loads block map.
     * Master block must already be loaded.
     */
    void load_block_map();
    
    /**
     * Performs the following steps:
     *  1) saves block map
     *  2) saves master block
     *  3) saves master block extension
     */
    void save_state();
    
    /**
     * Writes "synced_flag" to master block which is then externalized
     * automatically.
     */
    void write_synced_flag
        (
        bool synced_flag
        );

    /**
     * Performs the following steps:
     *  1) loads master block
     *  2) loads block map
     *  3) loads master block extension
     */
    void load_state();

private:

    //
    // Master block.
    //

    Int32       num_blocks;
        /**< Number of blocks (current pool size). */
    BlockIndex  block_map_chain;
        /**< Pointer to block map chain. */
    BlockIndex  first_free_block;
        /**< Pointer to first free block
             (first block of free block chain). */
    BlockIndex  last_free_block;
        /**< Pointer to last free block
             (last block of free block chain). */
    Int32       num_free_blocks;
        /**< Length of free block chain. */

private:

    //
    // Block map.
    //

    BlockMap    block_map;
        /**< Current block map is cached in memory. */
    int         block_map_entries_per_block;
        /**< Number of block map entries per block. */

private:    

    //
    // Internal state.
    //

    bool        mounted : 1;
        /**< Is block pool fully initialized/mounted? */
    bool        synced : 1;
        /**< Is BlockPool state properly synchronized with an external saved
             state? */

    }; // BlockPool

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_BLOCK_POOL_H
