/** @file compact_volume.cpp
 * Compact volume.
 */

/*
 * begin        : Sat 22 Sep 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/compact_volume.h"
#include "utility/quote_string.h"
#include <sstream>

namespace Massiv {
namespace Core {

#define STORAGE_NAME_SUFFIX     "{compact_volume_storage}"


// Implementation of CompactVolume::BlockInfo
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CompactVolume::BlockInfo::BlockInfo()
    {
    stream_index = INVALID_INDEX;
    slot_index = INVALID_INDEX;
    }


bool CompactVolume::BlockInfo::is_mapping_defined() const
    {
    return stream_index != INVALID_INDEX;
    }


void CompactVolume::BlockInfo::save
    (
    std::ostream &  s
    )
    {
    s << stream_index << " ";
    s << slot_index << " ";
    s << '\n';
    }


void CompactVolume::BlockInfo::restore
    (
    std::istream &  s
    )
    {
    s >> stream_index;
    s >> slot_index;
    }


// Implementation of CompactVolume::HelperStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CompactVolume::HelperStream::HelperStream()
    :   slot_size( 0 ),
        stream_index( INVALID_INDEX ),
        volume_flags( DEFAULT_VOLUME_FLAGS )
    {
    }


void CompactVolume::HelperStream::init
    (
    BasicVolume &           storage,
    std::ios_base::openmode mode,
    HelperStreamIndex       index,
    int                     size,
    int                     flags
    )
    {
    std::ostringstream  s;
    std::string         stream_name;

    // Initialize.

    slot_size = size;
    stream_index = index;
    volume_flags = flags;

    // Generate stream name.

    s << "helper_stream{" << stream_index << "}$";
    stream_name = s.str();

    // Open stream.

    if( flags_on( mode, std::ios_base::out ) )
        {
        // Create stream if does not exist.

        storage.create( stream_name );
        }

    stream.open
        (
        storage,
        stream_name,
        mode | std::ios_base::binary
        );

    stream.exceptions( std::ios_base::failbit | std::ios_base::badbit );
    }


void CompactVolume::HelperStream::flush()
    {
    stream.flush();
    }


void CompactVolume::HelperStream::save
    (
    std::ostream &  s
    )
    {
    SlotUsage::iterator it;

    s << slot_usage.size() << " ";

    for( it = slot_usage.begin(); it != slot_usage.end(); it++ )
        {
        s << *it << " ";
        }

    s << '\n';
    }


void CompactVolume::HelperStream::restore
    (
    std::istream &  s
    )
    {
    int count;

    hope( !slot_usage.size() );

    s >> count;

    while( count-- > 0 )
        {
        bool state;

        s >> state;

        slot_usage.push_back( state );
        }
    }


void CompactVolume::HelperStream::allocate
    (
    BlockInfo & block
    )
    {
    SlotUsage::iterator it;

    hope( !block.is_mapping_defined() );

    // Find a free slot.

    /**< @todo: better allocation algorithm */

    for( it = slot_usage.begin(); it != slot_usage.end(); it++)
        {
        if( !*it )
            {
            // Free slot found.

            *it = true;

            block.stream_index = stream_index;
            block.slot_index = it - slot_usage.begin();

            return;
            }
        }

    // Create a new slot.

    slot_usage.push_back( true );

    block.stream_index = stream_index;
    block.slot_index = slot_usage.size() - 1;
    }


void CompactVolume::HelperStream::free
    (
    BlockInfo & block
    )
    {
    int offset;

    hope( block.is_mapping_defined() );
    hope( block.stream_index == stream_index );

    if( volume_flags & CLEAR_FREE_BLOCKS )
        {
        offset = block.slot_index * slot_size;
        BasicVolume::zero_fill_stream( stream, offset, slot_size );
        }

    slot_usage[ block.slot_index ] = false;

    block.slot_index = INVALID_INDEX;
    block.stream_index = INVALID_INDEX;
    }


void CompactVolume::HelperStream::save_block
    (
    BlockInfo & block,
    UInt8 *     ptr
    )
    {
    int offset;

    hope( block.is_mapping_defined() );
    hope( block.stream_index == stream_index );

    offset = block.slot_index * slot_size;

    BasicVolume::write_to_stream( stream, ptr, offset, slot_size );
    }


void CompactVolume::HelperStream::load_block
    (
    BlockInfo & block,
    UInt8 *     ptr
    )
    {
    int offset;

    hope( block.is_mapping_defined() );
    hope( block.stream_index == stream_index );

    offset = block.slot_index * slot_size;

    BasicVolume::read_from_stream( stream, ptr, offset, slot_size );
    }


int CompactVolume::HelperStream::get_slot_size()
    {
    return slot_size;
    }


// Implementation of CompactVolume
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CompactVolume::CompactVolume
    (
    int flags,
    int block_size,
    int cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        storage( DEFAULT_VOLUME_FLAGS, STORAGE_BLOCK_SIZE, STORAGE_CACHE_SIZE ),
        mapping_dirty( false )
    {
    }


CompactVolume::CompactVolume
    (
    std::iostream &         image,
    std::ios_base::openmode mode,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        storage( DEFAULT_VOLUME_FLAGS, STORAGE_BLOCK_SIZE, STORAGE_CACHE_SIZE ),
        mapping_dirty( false )
    {
    try
        {
        if( !open( image, mode ) )
            {
            throw ExCanNotOpenVolumeImage( "Can not read from image stream." );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


CompactVolume::CompactVolume
    (
    const std::string &     file_name,
    std::ios_base::openmode mode,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        storage( DEFAULT_VOLUME_FLAGS, STORAGE_BLOCK_SIZE, STORAGE_CACHE_SIZE ),
        mapping_dirty( false )
    {
    try
        {
        if( !open( file_name, mode ) )
            {
            std::ostringstream ss;

            ss << "Can not open " << quote_const_string( file_name ) << ".";
            throw ExCanNotOpenVolumeImage( ss.str() );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


CompactVolume::CompactVolume
    (
    std::iostream &         image,
    const std::string &     volume_name,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        storage( DEFAULT_VOLUME_FLAGS, STORAGE_BLOCK_SIZE, STORAGE_CACHE_SIZE ),
        mapping_dirty( false )
    {
    try
        {
        if( !format( image, volume_name ) )
            {
            throw ExCanNotOpenVolumeImage( "Can not format image stream." );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


CompactVolume::CompactVolume
    (
    const std::string &     file_name,
    const std::string &     volume_name,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        storage( DEFAULT_VOLUME_FLAGS, STORAGE_BLOCK_SIZE, STORAGE_CACHE_SIZE ),
        mapping_dirty( false )
    {
    try
        {
        if( !format( file_name, volume_name ) )
            {
            std::ostringstream ss;

            ss << "Can not format " << quote_const_string( file_name ) << ".";
            throw ExCanNotOpenVolumeImage( ss.str() );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


CompactVolume::~CompactVolume()
    {
    close_pool();
    }


bool CompactVolume::open
    (
    std::iostream &         image,
    std::ios_base::openmode mode
    )
    {
    if( !storage.open( image, mode ) )
        {
        return false;
        }

    init( mode );
    restore_state();

    return true;
    }


bool CompactVolume::open
    (
    const std::string &     file_name,
    std::ios_base::openmode mode
    )
    {
    try
        {
        volume_name = file_name;

        if( !storage.open( file_name, mode ) )
            {
            volume_name = std::string();
            return false;
            }
        }
    catch( ... )
        {
        volume_name = std::string();
        throw;
        }

    init( mode );
    restore_state();

    return true;
    }


bool CompactVolume::format
    (
    std::iostream &     image,
    const std::string & volume_name
    )
    {
    if( !storage.format( image, volume_name + STORAGE_NAME_SUFFIX ) )
        {
        return false;
        }

    init( std::ios_base::in | std::ios_base::out );
    BasicVolume::format( volume_name );

    return true;
    }


bool CompactVolume::format
    (
    const std::string & file_name,
    const std::string & volume_name
    )
    {
    if( !storage.format( file_name, volume_name + STORAGE_NAME_SUFFIX ) )
        {
        return false;
        }

    init( std::ios_base::in | std::ios_base::out );
    BasicVolume::format( volume_name );

    return true;
    }


void CompactVolume::load_block
    (
    BlockIndex  index,
    UInt8 *     ptr
    )
    {
    hope( flags_on( get_open_mode(), std::ios_base::in ) );

    BlockInfo & block = extra_block_map[ index ];
    int         i;

    // Load block.

    if( !block.is_mapping_defined() )
        {
        // Block has not been allocated on the storage yet.
        // Zero-fill the block.

        for( i = 0; i < block_size; i++ )
            {
            ptr[ i ] = 0;
            }

        return;
        }

    HelperStream & stream = helper_stream[ block.stream_index ];
    stream.load_block( block, ptr );

    // Zero-fill rest of the block.

    for( i = stream.get_slot_size(); i < block_size; i++ )
        {
        ptr[ i ] = 0;
        }
    }


void CompactVolume::save_block
    (
    BlockIndex  index,
    UInt8 *     ptr
    )
    {
    hope( flags_on( get_open_mode(), std::ios_base::out ) );
    
    BlockInfo & block = extra_block_map[ index ];

    // Allocate block.

    #ifndef ENABLE_SPARSE_FILES

    if( !block.is_mapping_defined() )
        {
        // We have not been notified about this block
        // (@c notify_block_usage()).
        // -> assume that the whole block is used.

        internal_notify_block_usage( index, block_size, true );
        }

    #else

    // Sparse files:
    //  1) determine how many bytes block exactly uses (@c used_size),
    //     the rest is assumed to be zero-filled
    //  2) reallocate the block so that it would be held by the appropriate
    //     helper stream with optimal @c slot_size

    int used_size = block_size;

    while( used_size >= 1 )
        {
        if( ptr[ used_size - 1 ] != 0 )
            {
            break;
            }
                
        used_size--;
        }

    internal_notify_block_usage( index, used_size, true );

    if( !used_size )
        {
        return;
        }

    #endif // ENABLE_SPARSE_FILES

    // Save block.

    helper_stream[ block.stream_index ].save_block( block, ptr );
    }


void CompactVolume::freeing_block
    (
    BlockIndex  block
    )
    {
    notify_block_usage( block, 0 );
    }


void CompactVolume::adding_block
    (
    BlockIndex  block
    )
    {
    notify_block_usage( block, block_size );
    BasicVolume::adding_block( block );
    }


void CompactVolume::modified()
    {
    // We must propagate the call to @c storage so that it would be made dirty
    // too (CompactVolume's state is externalized to @c storage).
    
    BasicVolume::modified();
    storage.modified();
    }


void CompactVolume::notify_block_usage
    (
    BlockIndex  index,
    int         allocated_size
    )
    {
    internal_notify_block_usage( index, allocated_size, false );
    }
    

void CompactVolume::internal_notify_block_usage
    (
    BlockIndex  index,
    int         allocated_size,
    bool        called_from_save_block
    )
    {
    hope( flags_on( get_open_mode(), std::ios_base::out ) );

    if( index >= ( BlockIndex ) extra_block_map.size() )
        {
        // This is a new block.

        extra_block_map.resize( index + 1 );
        mapping_dirty = true;
        }

    BlockInfo & block = extra_block_map[ index ];

    if( !block.is_mapping_defined() )
        {
        // Block has not been allocated on the storage yet.

        if( allocated_size )
            {
            helper_stream[ choose_stream( allocated_size ) ].allocate( block );
            mapping_dirty = true;
            }
        }
    else
        {
        // If it is needed change block mapping to point to a slot of
        // different size (move block from one helper stream to another).

        HelperStreamIndex new_stream;

        if( allocated_size )
            {
            new_stream = choose_stream( allocated_size );

            if( block.stream_index != new_stream )
                {
                if( !called_from_save_block )
                    {
                    BlockRef b = access_block( index );

                    b.lock();

                    // Note: Since block is now locked in memory we can safely
                    //       delete it from the storage.

                    helper_stream[ block.stream_index ].free( block );
                    helper_stream[ new_stream ].allocate( block );

                    b.unlock();
                    }
                else
                    {
                    // We are called from @c save_block(), thus upcalled
                    // from LRUCache and can not use @c access_block() because
                    // that could reenter the cache.
                    // But it is okay to delete the block from the storage as
                    // it is in the memory and it will be saved to the new slot
                    // as soon as this method is left.
                    
                    helper_stream[ block.stream_index ].free( block );
                    helper_stream[ new_stream ].allocate( block );
                    }
                
                mapping_dirty = true;
                }
            }
        else
            {
            // Undefine block mapping.

            helper_stream[ block.stream_index ].free( block );
            mapping_dirty = true;
            }
        }
    }


void CompactVolume::flush_pool()
    {
    int i;

    // Flush CompactVolume's BlockPool.

    BasicVolume::flush_pool();

    if( can_write() && mapping_dirty )
        {
        BasicVolume::OFStream   ebms( storage, "extra_block_map$" );
        BasicVolume::OFStream   hsis( storage, "helper_stream_info$" );
        ExtraBlockMap::iterator it;

        ebms.exceptions( std::ios_base::failbit | std::ios_base::badbit );
        hsis.exceptions( std::ios_base::failbit | std::ios_base::badbit );

        // Save extra block map.

        ebms << extra_block_map.size() << '\n';

        for( it = extra_block_map.begin(); it != extra_block_map.end(); it++ )
            {
            it->save( ebms );
            }

        // Save helper stream info.

        for( i = 0; i < NUM_HELPER_STREAMS; i++ )
            {
            helper_stream[ i ].save( hsis );
            }

        // Done.

        mapping_dirty = false;
        }
    
    // Flush @c storage.

    for( i = 0; i < NUM_HELPER_STREAMS; i++ )
        {
        helper_stream[ i ].flush();
        }

    storage.flush();
    }


void CompactVolume::restore_state()
    {
    BasicVolume::IFStream   ebms( storage, "extra_block_map$" );
    BasicVolume::IFStream   hsis( storage, "helper_stream_info$" );
    int                     i;

    ebms.exceptions( std::ios_base::failbit | std::ios_base::badbit );
    hsis.exceptions( std::ios_base::failbit | std::ios_base::badbit );

    // Load extra block map.

    hope( !extra_block_map.size() );

    ebms >> i;

    while( i-- > 0 )
        {
        BlockInfo info;

        info.restore( ebms );

        extra_block_map.push_back( info );
        }

    // Load helper stream info.

    for( i = 0; i < NUM_HELPER_STREAMS; i++ )
        {
        helper_stream[ i ].restore( hsis );
        }

    // Restore CompactVolumes's BlockPool.

    BasicVolume::restore_state();
    }


void CompactVolume::init
    (
    std::ios_base::openmode mode
    )
    {
    HelperStreamIndex   i;
    int                 prev_slot_size;
    int                 slot_size;

    set_open_mode( mode );

    prev_slot_size = 0;

    for( i = 0; i < NUM_HELPER_STREAMS; i++ )
        {
        // Choose @c slot_size for helper stream @c i.
        // The only restrictions on @c slot_size are given
        // according to the following asserts.

        slot_size = ( i + 1 ) * ( i + 1 ) * block_size /
                    ( NUM_HELPER_STREAMS * NUM_HELPER_STREAMS );

        hope( slot_size > 0 );
        hope( slot_size > prev_slot_size );
        hope( ( i == NUM_HELPER_STREAMS - 1 && slot_size == block_size ) ||
                ( i != NUM_HELPER_STREAMS - 1 ) );

        // Initialize helper stream.

        helper_stream[ i ].init( storage, mode, i, slot_size, volume_flags );

        // Prepare for next iteration.

        prev_slot_size = slot_size;
        }
    }


CompactVolume::HelperStreamIndex CompactVolume::choose_stream
    (
    int allocated_size
    )
    {
    HelperStreamIndex i;

    // Find the smallest slot that @c allocated_size fits to.

    hope( allocated_size <= block_size );

    for( i = 0; i < NUM_HELPER_STREAMS; i++ )
        {
        if( helper_stream[ i ].get_slot_size() >= allocated_size )
            break;
        }

    return i;
    }


} // namespace Massiv::Core
} // namespace Massiv
