/** @file compact_volume.h
 * Compact volume.
 */

/*
 * begin        : Mon 16 Sep 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_COMPACT_VOLUME_H
#define MASSIV_COMPACT_VOLUME_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_BASIC_VOLUME_H
#include "filesys/basic_volume.h"
#endif

#ifndef MASSIV_VOLUME_H
#include "filesys/volume.h"
#endif

#include <vector>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/** @def ENABLE_SPARSE_FILES
 * Define this to enable sparse files (works for CompactVolume).
 * This option reduces allocated size of files that consists of
 * zero-filled or partially-zeroed blocks.
 */
#define ENABLE_SPARSE_FILES

/**
 * Space-saving representation of BasicVolume.
 * @c CompactVolume creates helper Volume (@c storage) that
 * holds set of @c helper streams where blocks of @c CompactVolume
 * are stored. @see BlockInfo, HelperStream.
 */
class CompactVolume : public BasicVolume
    {
private:

    typedef int     HelperStreamIndex;
    typedef int     SlotIndex;

    class           BlockInfo;
    class           HelperStream;

    friend class    BlockInfo;
    friend class    HelperStream;

    /**
     * Parameters.
     */
    enum Parameters
        {
        NUM_HELPER_STREAMS =    16,
            /**< Number of helper streams. */
        }; // Parameters

    /**
     * Constants.
     */
    enum Constants
        {
        INVALID_INDEX       =   -1,
            /**< Invalid index. */
        STORAGE_BLOCK_SIZE  =   DEFAULT_BLOCK_SIZE,
            /**< Block size of @c storage. */
        STORAGE_CACHE_SIZE  =   NUM_HELPER_STREAMS * 2 * STORAGE_BLOCK_SIZE
            /**< Storage cache size. */
        }; // Constants

    /**
     * Definition of place where block of @c CompactVolume
     * is stored on @c storage.
     */
    class BlockInfo
        {
    private:

        friend class HelperStream;
        friend class CompactVolume;

    public:

        /**
         * Default constructor.
         */
        BlockInfo();

    public:

        /**
         * Tests whether mapping for corresponding block
         * is defined (valid).
         */
        bool is_mapping_defined() const;

        /**
         * Saves block info.
         */
        void save
            (
            std::ostream &  s
            );

        /**
         * Restores block info.
         */
        void restore
            (
            std::istream &  s
            );

    private:

        HelperStreamIndex   stream_index;
            /**< Helper stream. */
        SlotIndex           slot_index;
            /**< Slot in helper stream. */
        }; // BlockInfo

    typedef std::vector<BlockInfo>  ExtraBlockMap;

    /**
     * Description of a helper stream.
     * Each helper stream is divided into a number of slots of fixed size.
     * Each slot is either used or free.
     * If a slot is used then it holds a block of @c CompactVolume.
     * There is a mapping of blocks of @c CompactVolume to
     * slots of suitable helper streams (via @c extra_block_map),
     * each helper stream has different slot size.
     */
    class HelperStream
        {
    private:

        typedef std::vector<bool>   SlotUsage;

    public:

        /**
         * Constructor.
         */
        HelperStream();

        /**
         * Initializes class.
         */
        void init
            (
            BasicVolume &           storage,
            std::ios_base::openmode mode,
            HelperStreamIndex       index,
            int                     size,
            int                     flags
            );

        /**
         * Flushes helper stream slot data.
         * Helper stream state is not flushed.
         */
        void flush();

    public:

        /**
         * Saves helper stream state.
         */
        void save
            (
            std::ostream &  s
            );

        /**
         * Restores helper stream state.
         */
        void restore
            (
            std::istream &  s
            );

    public:

        /**
         * Allocates slot for @c block.
         * Uses first-free algorithm.
         */
        void allocate
            (
            BlockInfo & block
            );

        /**
         * Frees slot occupied by @c block.
         */
        void free
            (
            BlockInfo & block
            );

        /**
         * Saves @c block to previously allocated slot.
         */
        void save_block
            (
            BlockInfo & block,
            UInt8 *     ptr
            );

        /**
         * Loads @c block from previously allocated slot.
         */
        void load_block
            (
            BlockInfo & block,
            UInt8 *     ptr
            );

        /**
         * Returns @c slot_size.
         */
        int get_slot_size();

    private:

        int                     slot_size;
            /**< Slot size. */
        SlotUsage               slot_usage;
            /**< Determines what slots are used.
                 This info is not saved into @c stream. */
        BasicVolume::FStream    stream;
            /**< Stream data. Contains slots (saved blocks) only. */
        HelperStreamIndex       stream_index;
            /**< Index of this helper stream. */
        int                     volume_flags;
            /**< CompactVolume's volume flags. */

        }; // HelperStream

public:

    /**
     * Construct uninitialized volume.
     * Either @c format() or @c open() is supposed to be called then.
     * @c block_size & @c cache_size parametrize @c CompactVolume,
     * not @c storage.
     */
    CompactVolume
        (
        int flags = DEFAULT_VOLUME_FLAGS,
        int block_size = DEFAULT_BLOCK_SIZE,
        int cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Opens volume.
     */
    CompactVolume
        (
        std::iostream &         image,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Opens volume.
     */
    CompactVolume
        (
        const std::string &     file_name,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Formats volume.
     */
    CompactVolume
        (
        std::iostream &         image,
        const std::string &     volume_name,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Formats volume.
     */
    CompactVolume
        (
        const std::string &     file_name,
        const std::string &     volume_name,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Destructor.
     */
    ~CompactVolume();

    /**
     * Opens volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool open
        (
        std::iostream &         image,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
        );

    /**
     * Opens volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool open
        (
        const std::string &     file_name,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
        );

    /**
     * Formats volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool format
        (
        std::iostream &     image,
        const std::string & volume_name
        );

    /**
     * Formats volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool format
        (
        const std::string & file_name,
        const std::string & volume_name
        );

protected:

    //
    // BlockPool interface implementation.
    //

    /**
     */
    virtual void load_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        );

    /**
     */
    virtual void save_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        );

    /**
     */
    virtual void freeing_block
        (
        BlockIndex  block
        );

    /**
     */
    virtual void adding_block
        (
        BlockIndex  block
        );
    
    /**
     */
    virtual void modified();

protected:

    //
    // StreamPool interface implementation.
    //

    /**
     */
    virtual void notify_block_usage
        (
        BlockIndex  index,
        int         allocated_size
        );

protected:

    //
    // SimpleBlockPool interface implementation.
    //

    /**
     */
    virtual void flush_pool();

    /**
     */
    virtual void restore_state();

private:
    
    /**
     * Internal version of @c notify_block_usage().
     */
    void internal_notify_block_usage
        (
        BlockIndex  index,
        int         allocated_size,
        bool        called_from_save_block
        );

    /**
     * Initializes object.
     */
    void init
        (
        std::ios_base::openmode mode
        );

    /**
     * Chooses what helper stream should hold
     * block of size @c allocated_size.
     */
    HelperStreamIndex choose_stream
        (
        int allocated_size
        );

private:

    Volume          storage;
        /**< This is the volume that is actually created.
             We use this volume to store our blocks there. */
    HelperStream    helper_stream[ NUM_HELPER_STREAMS ];
        /**< Array of helper streams allocated on @c storage. */
    ExtraBlockMap   extra_block_map;
        /**< Mapping of blocks managed by @c BlockPool
             to slots of helper streams. */
    bool            mapping_dirty;
        /**< True if CompactVolume's metadata is dirty. */

    }; // CompactVolume

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_COMPACT_VOLUME_H
