/** @file compressed_stream.cpp
 * Compressed streams.
 */

/*
 * begin        : Tue 3 Sep 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


// Massiv:

#include "global/common.h"
#include "utility/types.h"
#include "utility/array.h"
#include "filesys/compressed_stream.h"

// CryptoPP:

#include "net/cryptopp_begin.h"
#include "cryptopp/gzip.h"
#include "cryptopp/files.h"
#include "net/cryptopp_end.h"


namespace Massiv {
namespace Core {
    
typedef CryptoPP::Gzip      Compressor;
typedef CryptoPP::Gunzip    Decompressor;


// Definition of CompressedStream::StreamBufferWrite
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * StreamBufferWrite.
 * @c std::streambuf object implementation that compresses provided data into
 * a bound @c std::ostream.
 */
class CompressedStream::StreamBufferWrite : public std::streambuf
    {
private:

    enum Parameters
        {
        BUFFER_LENGTH   =   16
            /**< Default put buffer length. */
        }; // Parameters

public:

    /**
     * Constructor.
     * Bounds buffer to @c output_stream that will hold compressed data.
     * @c output_stream must be empty.
     */
    StreamBufferWrite
        (
        std::ostream &  output_stream
        ) : compressor( new CryptoPP::FileSink( output_stream ) ),
            bytes_put( 0 )
        {
        // Tell std::streambuf that stream is opened for writing.
        // Put area is empty and next access to the stream will cause
        // @c overflow() to be called. Put area can not be NULL.

        setp( reserve_area, reserve_area );
        }

    /**
     * Destructor.
     */
    ~StreamBufferWrite()
        {
        sync();

        if( bytes_put )
            {
            compressor.MessageEnd();
            }
        }

protected:

    //
    // std::streambuf interface implementation.
    //

    /**
     * Called by IO stream library.
     * Compresses content of the buffer.
     */
    virtual int sync()
        {
        overflow( traits_type::eof() );

        return 0;
        }

    /**
     * Called by IO stream library.
     * Compresses content of the buffer and initializes a new buffer.
     */
    virtual int_type overflow
        (
        int_type    chr
        )
        {
        UInt8 *             data;
        int                 length;

        // Compress put area into output stream.

        data = reinterpret_cast<UInt8 *>( pbase() );
        length = pptr() - pbase();

        if( length )
            {
            hope( data );
            compressor.Put( data, length );

            bytes_put += length;
            }

        // Initialize new put area.

        if( chr != traits_type::eof() )
            {
            reserve_area[ 0 ] = traits_type::to_char_type( chr );

            setp( reserve_area, reserve_area + BUFFER_LENGTH );
            pbump( 1 );
            }
        else
            {
            setp( reserve_area, reserve_area + BUFFER_LENGTH );
            }

        return traits_type::not_eof( chr );
        }

private:

    Compressor  compressor;
        /**< Compressor. Writes output to @c output stream. */
    char        reserve_area[ BUFFER_LENGTH ];
        /**< Default put buffer. */
    int         bytes_put;
        /**< Number of bytes put to @c compressor. */

    }; // StreamBufferWrite


// Definition of CompressedStream::StreamBufferRead
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * StreamBufferRead.
 * @c std::streambuf object implementation that decompresses data from a bound
 * @c std::istream.
 */
class CompressedStream::StreamBufferRead : public std::streambuf
    {
private:

    enum Parameters
        {
        PUMP_BYTES  =   32
            /**< How many bytes pump into @c filter_queue from the compressed
                 stream at @c underflow(). */
        }; // Parameters

    /**
     * Sink.
     * CryptoPP decompressor puts decompressed data here.
     */
    class Sink : public CryptoPP::Sink
        {
    public:

        typedef Array< DynamicStorage<UInt8> >  Storage;

    protected:

        //
        // CryptoPP::Sink interface implementation.
        //

        /**
         * Called by @c Decompressor when outputs decompressed data.
         */
        virtual void Put
            (
            UInt8   b
            )
            {
            reserve_area.push_back( b );            
            }

        /**
         * Called by @c Decompressor when outputs decompressed data.
         */
        virtual void Put
            (
            const UInt8 *   data,
            unsigned int    length
            )
            {
            while( length-- > 0 )
                {
                Sink::Put( *data );
                data++;
                }
            }

        #if( CRYPTOPP_VERSION >= 500 )

        /**
         * Called by @c Decompressor when outputs decompressed data.
         * @note Applicable for Cryptopp 5.0 and above.
         */
        virtual unsigned int Put2
            (
            const UInt8 *   data,
            unsigned int    length,
            int             message_end,
            bool            blocking
            )
            {
            Put( data, length );
            return 0;
            }

        /**
         * Called by @c Decompressor to flush intermediate buffers.
         * @note Applicable for Cryptopp 5.0 and above.
         */
        bool IsolatedFlush
            (
            bool    hard_flush,
            bool    blocking
            )
            {
            // This is the bottom of buffered transformation chain.
            // Data stays at @c Sink.

            return false;
            }

        #endif

    public:

        Storage reserve_area;
            /**< Variable-length reserve area. */

        }; // Sink

public:

    /**
     * Bounds buffer to @c input_stream that holds compressed data.
     */
    StreamBufferRead
        (
        std::istream &  input_stream
        ) : sink( new Sink() ),
            decompressor( new Decompressor( sink ) ),
            filter_queue( input_stream, false, decompressor )
        {
        static char void_area;

        // Note: filters must be initialized in this (exact) order
        //       (sink -> filter_queue).

        // Tell std::streambuf that stream is opened for reading.
        // Get area is empty and next access to the stream will cause
        // @c underflow() to be called. Get area can not be NULL.

        setg( &void_area, &void_area, &void_area );
        }

protected:

    //
    // std::streambuf interface implementation.
    //

    /**
     * Called by IO stream library.
     * Initializes buffer by decompressed data.
     */
    virtual int_type underflow()
        {
        Sink::Storage & reserve_area = sink->reserve_area;
        const char *    ptr;
        char *          data;

        // Did get area really underflow?

        ptr = gptr();

        if( ptr != egptr() )
            {
            return traits_type::to_int_type( *ptr );
            }

        // Decompress data.

        reserve_area.clear();

        while( reserve_area.empty() )
            {
            // Pump compressed data into the queue until we are at the end of
            // the compressed stream or the decompressor put decompressed data
            // into @c reserve_area.

            if( !filter_queue.Pump( PUMP_BYTES ) )
                {
                filter_queue.PumpAll();
                break;
                }
            }

        // Initialize get area.

        if( !reserve_area.empty() )
            {
            // Bind streambuf to @c reserve_area.

            data = reinterpret_cast<char *>( reserve_area.get_buffer() );
            setg( data, data, data + reserve_area.size() );

            return traits_type::to_int_type( reserve_area[ 0 ] );
            }
        else
            {
            // End of stream.

            static char_type void_area;

            // Can not set get area to NULL.
            // That would mean that stream is not opened for reading!

            setg( &void_area, &void_area, &void_area );

            return traits_type::eof();
            }
        }

private:

    Sink *                  sink;
        /**< Filter queue tail. Stores uncompressed data. */
    Decompressor *          decompressor;
        /**< Decompresses data. Writes data to @c sink. */
    CryptoPP::FileSource    filter_queue;
        /**< Filter queue head. Pumps data from @c input stream to
             @c decompressor. */
    }; // StreamBufferRead


// Implementation of CompressedStream::IStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CompressedStream::IStream::IStream
    (
    std::istream &  input_stream
    ) : Super( NULL ),
        owned_stream( NULL ),
        buffer( new StreamBufferRead( input_stream ) )
    {
    Super::init( buffer );
    }


CompressedStream::IStream::IStream
    (
    std::istream *  input_stream
    ) : Super( NULL ),
        owned_stream( input_stream ),
        buffer( new StreamBufferRead( *input_stream ) )
    {
    Super::init( buffer );
    }


CompressedStream::IStream::~IStream()
    {
    delete buffer;
    delete owned_stream;
    }


// Implementation of CompressedStream::OStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CompressedStream::OStream::OStream
    (
    std::ostream &  output_stream
    ) : Super( NULL ),
        owned_stream( NULL ),
        buffer( new StreamBufferWrite( output_stream ) )
    {
    Super::init( buffer );
    }


CompressedStream::OStream::OStream
    (
    std::ostream *  output_stream
    ) : Super( NULL ),
        owned_stream( output_stream ),
        buffer( new StreamBufferWrite( *output_stream ) )
    {
    Super::init( buffer );
    }


CompressedStream::OStream::~OStream()
    {
    delete buffer;
    delete owned_stream;
    }


} // namespace Massiv::Core
} // namespace Massiv
