/** @file compressed_stream.h
 * Compressed streams.
 */

/*
 * begin        : Tue 3 Sep 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_COMPRESSED_STREAM_H
#define MASSIV_COMPRESSED_STREAM_H

#include <iostream>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/** @class CompressedStream
 * CompressedStream defines @c IStream and @c OStream classes
 * that provide IO stream interface to compressed data streams.
 * @c IStream decompresses data stored in a bound @c std::istream.
 * @c OStream compresses data into a bound @c std::ostream.
 * Both @c IStream and @c OStream classes do not support
 * seeking in uncompressed streams and simultaneous read & write access.
 * CompressedStream uses @c CryptoPP::Gzip and @c CryptoPP::Unzip engines.
 * Format of compressed data streams is standard gzip.
 */
class CompressedStream
    {
private:

    class StreamBufferWrite;
    class StreamBufferRead;

public:

    /**
     * IStream.
     */
    class IStream : public std::istream
        {
    private:

        typedef std::istream Super;

    public:

        /**
         * Opens stream for reading (exclusive).
         */
        IStream
            (
            std::istream & input_stream
            );
        
        /**
         * Opens stream for reading (exclusive).
         * Will own @c input_stream.
         */
        IStream
            (
            std::istream * input_stream
            );

        /**
         * Destructor.
         */
        ~IStream();

    private:
        
        std::istream *      owned_stream;
            /**< Owned stream. */
        StreamBufferRead *  buffer;
            /**< Stream buffer. */

        }; // IStream

    /**
     * OStream.
     */
    class OStream : public std::ostream
        {
    private:

        typedef std::ostream Super;

    public:

        /**
         * Opens stream for writing (exclusive).
         */
        OStream
            (
            std::ostream & output_stream
            );
        
        /**
         * Opens stream for writing (exclusive).
         * Will own @c output_stream.
         */
        OStream
            (
            std::ostream * output_stream
            );
        
        /**
         * Destructor.
         */
        ~OStream();

    private:

        std::ostream *      owned_stream;
            /**< Owned stream. */
        StreamBufferWrite * buffer;
            /**< Stream buffer. */

        }; // OStream

    }; // CompressedStream

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_COMPRESSED_STREAM_H
