/** @file file_ack_manager.cpp
 * File ack manager (implementation).
 */

/*
 * begin        : Sun Jun 29 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/file_ack_manager.h"
#include "filesys/file_system.h"
#include "global/global.h"
#include "utility/heap.h"
#include "utility/heap_impl.h"
#include "filesys/volume_manager.h"
#include <sstream>


namespace Massiv {
namespace Core {


/**
 * FileAckManager internal data structure.
 */
struct FileAckManager::My
    {
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );

public:

    /** Constructor. */
    My()
        {
        }

public:

    /**
     * Job describes which file ack to create.
     * Executing job means creating ack file to the file the job describes.
     */
    struct Job
        {
        std::string         file_name;
            /**< File name of the file whose ack must be created. */
        }; // Job
    
    /**
     * Jobs is a schedule which stores jobs sorted by their executing time.
     */
    typedef Heap
        <
        SystemTime,         // Time when the job must be executed.
        Job,                // Job itself.
        SystemTime::Less    // SystemTime comparator.
        > Jobs;

public:

    /** Get ack file name for specified file. */
    std::string get_ack_file_name
        (
        const std::string & file_name
        ) const
        {
        return file_name + ".__ack__";
        }


    /** Check whether specified file has its ack file. */
    bool is_file_acked
        (
        const std::string & file_name
        )
        {
        std::string             ack_file_name = get_ack_file_name( file_name );
        VolumeManager::IFStream ack_file( ack_file_name );
        return ack_file.good();
        }


    /** Execute specified job. */
    void execute_job
        (
        const Job & job
        )
        {
        if( job.file_name.empty() )
            return;
        
        std::string                 ack_file_name = get_ack_file_name( job.file_name );
        
        // Check if the ack file already exist.
        if( is_file_acked( job.file_name ) )
            {
            std::ostringstream ss;
            ss  << "Error creating ack file '" << ack_file_name << "'. "
                << "The ack file already exists.";
            
            Global::log_warning
                (
                Status::FACILITY_GENERIC,
                Status::PRIORITY_HIGHEST,
                ss.str()
                );
            }
        else
            {
            // Create the ack file.
            VolumeManager::OFStream     ack_file( ack_file_name );
        
            if( !ack_file.good() )
                {
                std::ostringstream ss;
                ss  << "Error creating ack file '" << ack_file_name << "'. "
                    << "Unable to open output stream.";
            
                Global::log_warning
                    (
                    Status::FACILITY_GENERIC,
                    Status::PRIORITY_HIGHEST,
                    ss.str()
                    );
                }
            else
                {
                // Ack file successfully created.
                // Nothing to do.
                }
            }
        }

public:

    Jobs        jobs;
        /**< Job schedule sorted by their execution time. */

    }; // FileAckManager::My


// FileAckManager::Setting
// ~~~~~~~~~~~~~~~~~~~~~~~

namespace {

const Float32 DEFAULT_CREATE_ACK_DELAY = 10.0f; // ten seconds

}

FileAckManager::Settings::Settings() :
    Super( "FileAckManager" ),
    create_ack_delay( access_float( "create_ack_delay", DEFAULT_CREATE_ACK_DELAY ) )
    {
    }


// FileAckManager
// ~~~~~~~~~~~~~~


FileAckManager::FileAckManager() :
    my( new My() )
    {
    }


FileAckManager::~FileAckManager()
    {
    // Wait until all jobs have been completed.
    // Report how long we have to wait.
    
    SystemTime          wait_time = get_seconds_until_work_done();
    std::ostringstream  ss;

    ss  << "Waiting " << wait_time.get_value() << " seconds for finishing I/O operations...";

    Global::log_info
        (
        Status::FACILITY_FILE_SYSTEM,
        Status::PRIORITY_NORMAL,
        ss.str()
        );
    
    // Wait until done.

    while( !work_done() )
        {
        next_tick();
        }
    }


void FileAckManager::add_job
    (
    const std::string & file_name_to_ack
    )
    {
    // Get default ack creation delay from settings.
    SystemTime  time_delta = SystemTime( settings.create_ack_delay );
    
    // Add job.
    add_job( file_name_to_ack, time_delta );
    }


void FileAckManager::add_job
    (
    const std::string & file_name_to_ack,
    const SystemTime &  time_delta
    )
    {
    // Compute execution time of new job.
    SystemTime  current_time = Global::system_time();
    SystemTime  job_time = current_time + time_delta;
    
    // Create the job.
    My::Job     job;
    job.file_name = file_name_to_ack;
    
    // Schedule the job.
    my->jobs.push( job_time, job );
    }


void FileAckManager::remove_ack
    (
    const std::string & file_name
    )
    {
    /** @todo There may be more files with same name in different mounted volumes! */
    
    FileSystem &    file_system = Global::volume_manager().file_system();
    bool            done = false;
    std::string     ack_file_name = my->get_ack_file_name( file_name );
    
    // Remove ack file.
    try 
        {
        file_system.delete_file( ack_file_name );
        }
    catch( std::exception & )
        {
        }
    
    // Remove all ack requests for the file from the job schedule.
    while( !done )
        {
        unsigned int    i;
        
        done = true;
    
        for( i = 0; i < (unsigned int) my->jobs.get_size(); ++i )
            {
            const My::Job & job = my->jobs[ i ];
            
            if( job.file_name == file_name )
                {
                // Remove the job from the schedule.
                my->jobs.pop( i );
                
                // Job schedule has changed internal structure. Loop again...
                done = false;
                break;
                }
            }
        }
    }


bool FileAckManager::is_file_acked
    (
    const std::string & file_name
    ) const
    {
    return my->is_file_acked( file_name );
    }


bool FileAckManager::work_done() const
    {
    // No jobs, no work.
    return my->jobs.is_empty();
    }


SystemTime FileAckManager::get_seconds_until_work_done() const
    {
    SystemTime      current_time = Global::system_time();
    SystemTime      last_job_time = current_time;
    unsigned int    i;
    
    // Store execution time of last job to 'last_job_time'...
    for( i = 0; i < (unsigned int) my->jobs.get_size(); ++i )
        {
        SystemTime  job_time = my->jobs.get_key( i );
        if( last_job_time < job_time )
            {
            last_job_time = job_time;
            }
        }
    
    return SystemTime( last_job_time - current_time );
    }


void FileAckManager::next_tick()
    {
    // Get current time.
    SystemTime  current_time = Global::system_time();
    
    // Execute all jobs scheduled for current tick.
    while( !my->jobs.is_empty() )
        {
        // Get time of first job.
        SystemTime  job_time = my->jobs.get_key( 0 );
        
        // Check if the job must be executed in current tick.
        if( job_time <= current_time )
            {
            // The job must be executed in current tick.
            My::Job     job;
            
            // Remove the job from the schedule.
            my->jobs.pop( 0, &job );
            
            // Execute the job.
            my->execute_job( job );
            }
        else
            {
            // There are no more jobs for current tick.
            break;
            }
        }
    }


} // namespace Core
} // namespace Massiv
