/** @file file_ack_manager.h
 * File ack manager.
 */

/*
 * begin        : Sun Jun 29 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FILE_ACK_MANAGER_H
#define MASSIV_FILE_ACK_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#include <string>
#include <memory> // auto_ptr


namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/**
 * FileAckManager is used to ack files.
 * When a file is written to disk (ie when a programator closes its output
 * stream), we don't know whether data of the file will be properly stored
 * on the disk, because the application or OS can crash soon after the stream
 * was closed.
 * Sometimes it is not possible to verify the data when the file is opened
 * just by reading its entire content. For such files, the FileAckManager
 * can create so called ack files. These ack files can be used to detect
 * whether the application or OS has crashed in few seconds after a
 * file stream has been closed.
 *
 * Usage of the FileAckManager is simple. When you close output stream of
 * a file, schedule ack file creation in the FileAckManager. The FileAckManager
 * will wait for specified number of seconds (by default the number of
 * seconds is taken from settings) and then it will create the ack file.
 * Before opening the file for reading, you can ask FileAckManager whether
 * ack of the file was created or not.
 *
 * Of course this method is simple and assumes that the number of seconds
 * the FileAckManager waits until it creates the ack file is specified
 * in a hope that all data of the file being acked is flushed from all caches
 * in the specified time. There are many reason why this hope will never
 * be true, but we don't care because such errors can't be handled
 * by software anyway. The ack files just says that "X seconds after a file
 * was written its ack was properly created. And we hope that with big
 * probability the data of the file is ok".
 */
class FileAckManager
    {
    NO_COPY_CONSTRUCTOR( FileAckManager );
    NO_ASSIGNMENT_OPERATOR( FileAckManager );

public:

    /** Constructor. */
    FileAckManager();

    /** Destructor. */
    ~FileAckManager();

public:

    /**
     * Add new ack file creation request.
     * Time delay until the ack file will be created is determined from
     * FileAckManager settings.
     */
    void add_job
        (
        const std::string & file_name_to_ack
        );

    /**
     * Add new ack file creation request.
     * Time delay until the ack file will be created is determined from
     * @a time_delta argument (in seconds).
     */
    void add_job
        (
        const std::string & file_name_to_ack,
        const SystemTime &  time_delta
        );
    
    /**
     * Remove ack file for specified file.
     * If there is scheduled ack request for the file, it is removed
     * from the schedule.
     * @bug There may be more files with same name in different mounted volumes.
     *      Only the first found ack is removed. Should we remove all acks?
     *      If someone don't want some ack to be removed, he has to mount
     *      the volume containing the ack as read only??
     */
    void remove_ack
        (
        const std::string & file_name
        );
    
    /**
     * Check whether specified file has its ack file.
     * @note This function doesn't check if the file itself exists,
     *      it only check if its ack file exists.
     */
    bool is_file_acked
        (
        const std::string & file_name
        ) const;

    /**
     * Check whether the manager has scheduled some work.
     */
    bool work_done() const;
    
    /**
     * Get number of seconds from now (current system time) until the manager
     * finishes all its work.
     */
    SystemTime get_seconds_until_work_done() const;
    
public:
    
    /**
     * Advance next tick.
     */
    void next_tick();
    
public:

    /**
     * FileAckManager settings.
     */
    class Settings : public ::Massiv::Core::Settings
        {
        typedef ::Massiv::Core::Settings    Super;
        
    public:
    
        /** Constructor. */
        Settings();
        
    public:
    
        Float &     create_ack_delay;
            /**< Default time delay between ack creation request
             * and ack file creation itself (in seconds).
             */

        }; // Settings
    
private:

    // Internal data structure. Forward declaration.
    struct My;
    
    std::auto_ptr< My >     my;
        /**< Internal data structure. Always valid pointer. */
    
    Settings                settings;
        /**< FileAckManager settings. */

    }; // FileAckManager

/** @} */ // addtogroup filesys


} // namespace Core
} // namespace Massiv


#endif // MASSIV_FILE_ACK_MANAGER_H
