/** @file file_iterator.cpp
 * File iterator implementation.
 */

/* 
 * begin        : Wed Jan 29 2003
 * copyright    : (C) 2002 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/file_iterator.h"

namespace Massiv {
namespace Core {


// FileIterator implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileIterator::FileIterator()
    : impl( NULL )
    {
    }


FileIterator::FileIterator
    (
    Reference< FileIteratorImpl >     it
    ) : impl( it )
    {
    }


FileIterator::FileIterator
    (
    const FileIterator &  src
    ) : impl( src.impl )
    {
    }


FileIterator::~FileIterator()
    {
    }


bool FileIterator::operator==
    (
    const FileIterator & it
    ) const
    {
    bool result = true;
    
    if( ( impl == NULL ) || ( it.impl == NULL ) )
        {
        if( impl == it.impl )
            {
            result = true;
            }
        else
            {
            result = false;
            }
        }
    else
    if( impl->is_valid() && it.impl->is_valid() )
        {
        if( impl->get_absolute_directory() != 
            it.impl->get_absolute_directory() )
            {
            result = false;
            }
        else
        if( impl->name() != it.impl->name() )
            {
            result = false;
            }
        }
    else
        {
        if( impl->is_valid() || it.impl->is_valid() )
            {
            // Exactly one of the iterators is valid.
            result = false;
            }
        }

    return result;
    }  // FileIterator::operator==()


void FileIterator::bind_current
    (
    const std::string &     mask
    )
    {
    impl->bind_current( mask );
    }


void FileIterator::bind
    (
    const std::string &     path,
    const std::string &     mask
    )
    {
    impl->bind( path, mask );
    }


void FileIterator::unbind()
    {
    impl->unbind();
    }


bool FileIterator::operator!=
    (
    const FileIterator & it
    ) const
    {
    return !operator==( it );
    }  


FileIterator & FileIterator::operator=
    (
    const FileIterator & it
    )
    {
    impl = it.impl;
    return *this;
    }


FileIterator & FileIterator::next()
    {
    impl->next();
    return *this;
    }


FileIterator & FileIterator::operator++( int )
    {
    impl->next();
    return *this;
    }


FileIterator & FileIterator::rewind()
    {
    impl->rewind();
    return *this;
    }


FileIterator FileIterator::enter()
    {
    return FileIterator( impl->enter() );
    }


FileIterator & FileIterator::descent()
    {
    impl->descent();
    return *this;
    }


FileIterator FileIterator::leave()
    {
    return FileIterator( impl->leave() );
    }


FileIterator & FileIterator::ascent()
    {
    impl->ascent();
    return *this;
    }


bool FileIterator::is_valid() const
    {
    if( impl == NULL )
        {
        return false;
        }
    else
        {
        return impl->is_valid();
        }
    }


bool FileIterator::is_file() const
    {
    return impl->is_file();
    }


bool FileIterator::is_directory() const
    {
    return impl->is_directory();
    }


bool FileIterator::is_trivial_directory() const
    {
    return impl->is_trivial_directory();
    }


std::string FileIterator::name() const
    {
    return impl->name();
    }


UInt64 FileIterator::size() const
    {
    return impl->size();
    }


std::string FileIterator::get_absolute_directory() const
    {
    return impl->get_absolute_directory();
    }


bool FileIterator::is_in_root() const
    {
    return impl->is_in_root();
    }


std::string FileIterator::get_mask()
    {
    return impl->get_mask();
    }


void FileIterator::set_mask
    (
    const std::string &     mask
    )
    {
    impl->set_mask( mask );
    }


FileIterator::DirectoryEnumerationMode FileIterator::get_dir_enum_mode()
    {
    return impl->get_dir_enum_mode();
    }


void FileIterator::set_dir_enum_mode
    (
    DirectoryEnumerationMode mode
    )
    {
    impl->set_dir_enum_mode( mode );
    }


Reference< FileIteratorImpl > FileIterator::get_impl()
    {
    return impl;
    }


// FileIteratorImpl implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileIteratorImpl::FileIteratorImpl
    (
    FileIterator::DirectoryEnumerationMode  mode
    )
    : valid_flag( false ),
      dir_enum_mode( mode )
    {
    }


FileIteratorImpl::FileIteratorImpl
    (
    const std::string &                     mask,
    FileIterator::DirectoryEnumerationMode  mode
    ) 
    {
    set_dir_enum_mode( mode );

    // Class will be initialized from descendant class according to whether 
    // the iterator binding will be successful.
    }


FileIteratorImpl::FileIteratorImpl
    (
    const std::string &                     path,
    const std::string &                     mask,
    FileIterator::DirectoryEnumerationMode  mode
    )
    {
    set_dir_enum_mode( mode );

    // Class will be initialized from descendant class according to whether 
    // the iterator binding will be successful.
    }


FileIteratorImpl::~FileIteratorImpl()
    {
    }


std::string FileIteratorImpl::get_mask()
    {
    return mask;
    }


void FileIteratorImpl::set_mask
    (
    const std::string &     mask
    )
    {
    this->mask = mask;
    }


FileIterator::DirectoryEnumerationMode FileIteratorImpl::get_dir_enum_mode()
    {
    return dir_enum_mode;
    }


void FileIteratorImpl::set_dir_enum_mode
    (
    FileIterator::DirectoryEnumerationMode mode
    )
    {
    dir_enum_mode = mode;
    }


} // namespace Massiv::Core
} // namespace Massiv
