/** @file file_iterator.h
 * File iterator.
 * Abstract interface that allows to enumerate items from a file-system-like
 * hierarchical structure.
 */

/* 
 * begin        : Wed Jan 29 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FILE_ITERATOR_H
#define MASSIV_FILE_ITERATOR_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#include <string>

namespace Massiv {
namespace Core {


namespace Test {

class FsEnumerationTest;
class VolumeEnumerationTest;

}

class BasicVolume;
class FileIterator;
class FileIteratorImpl;


/** @addtogroup filesys
 * @{
 */


/**
 * Represents a "pointer" to a specific point in a file system
 * (i.e. either to a file or to a directory) or another file-system-like
 * hierarchical structure.
 * Accesses information on the referenced item and allows to traverse
 * through the directory structure.
 * The interface is designed to be independent on the target filesystem
 * as much as possible. For example there might be various iterator
 * implementations that enumerate files on the filesystem provided by
 * the OS or enumerate items from client-defined structure.
 * The interface defines its own (architecture-independent) naming
 * conventions. File name path components are always separated by '/',
 * regardless of the enumerated structure.
 * There may be little deviations like how absolute paths are specified
 * (drives on Windows, ...).
 * The iterator is either bound or unbound. A bound iterator is linked
 * with a specific point in the iterator-enumerated structure.
 * By default iterator is constructed unbound. To bind it, use
 * enumerated-structure helper methods, such as
 * @c FileSystem::open_directory().
 *
 * Iterator operator= and copy constructor do not actualy create a new
 * iterator instance. Iterator contexts (including iterator position)
 * are shared instead.
 * The main reason for this is that iterator contexts might not be
 * copyable.
 * This has the consequence that unbind(), operator++(), etc. called on
 * one of the shared iterators affects the other iterator too.
 *
 * The iterator behaviour for the following cases is left undefined:
 *  - new files are created in an already opened (just being enumerated)
 *    directory
 *  - files are deleted from an opened directory
 *
 * Additional restrictions may apply for various iterator kinds,
 * see @c FileIteratorImpl.
 *
 * FileIterator is a proxy to @c FileIteratorImpl
 * which holds the (shared) iterator context.
 *
 * @code
 * {
 * FileSystem & fs = ...;
 * FileIterator it( fs.open_directory( "../dir", "*" ) );
 *
 * while( it.is_valid() )
 *     {
 *     if( it.is_directory() )
 *         {
 *         std::cout << it.name() << "\t\tDIR" << std::endl;
 *         }
 *     else
 *         {
 *         std::cout << it.name() << "\t\t" << it.size() << std::endl;
 *         }
 *
 *     it++;
 *     }
 * } // Iterator's binding with '../dir' is abolished here
 * @endcode
 */
class FileIterator
    {
private:

    friend class Massiv::Core::Test::FsEnumerationTest;
    friend class Massiv::Core::Test::VolumeEnumerationTest;

public:
    
    /**
     * Specifies how subdirectory names should be enumerated.
     */
    enum DirectoryEnumerationMode
        {
        MATCH_DIRS_NAMES            = 0,
            /**< Match directory names with mask, don't enumerate directories
                    that don't match it. */

        ENUMERATE_ALL_DIRS          = 1,
            /**< Enumerate all subdirectory names regardless whether they 
                    match the mask or not. */

        DIRECTORY_ENUMERATING_MODE_COUNT,
        }; // enum DirectoryEnumerationMode

public:

    /**
     * Default constructor.
     * Creates unbound/invalid iterator.
     */
    FileIterator();

    /**
     * Constructor.
     * Used by iterator implementation object to created a bound iterator.
     * @param it    Iterator implementation (architecture specific).
     * @exception   ExDirectoryNotFound The specified directory doesn't 
     *              exist.
     */
    explicit FileIterator
        (
        Reference<FileIteratorImpl> it
        );

    /**
     * Copy constructor.
     * Shares iterator context with @c src.
     */
    FileIterator
        (
        const FileIterator & src
        );

    /**
     * Destructor.
     */
    ~FileIterator();

public:

    /**
     * Binds the iterator with the directory that is current to the
     * iterator-enumerated structure. If the iterator has already been bound,
     * the old binding is abolished.
     *
     * @param path  Path specifying the directory to be opened.
     * @param mask  Mask describing files that should be enumerated. May
     *              contain wildcards, i.e. * and ?.
     *
     * @exception   ExDirectoryNotFound The specified directory doesn't 
     *              exist.
     */
    void bind_current
        (
        const std::string & mask
        );

    /**
     * Binds the iterator with an explictly specified directory.
     * If the iterator has already been bound, this old binding is abolished.
     *
     * @param path  Path specifying the directory to be opened.
     * @param mask  Mask describing files that should be enumerated. May
     *              contain wildcards, i.e. * and ?.
     * @exception   ExDirectoryNotFound The specified directory doesn't 
     *              exist.
     */
    void bind
        (
        const std::string & path,
        const std::string & mask
        );

    /**
     * Unbinds/invalidates the iterator.
     * Abolishes existing binding.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    void unbind();

public:

    /**
     * Compares the iterators for equality. Two iterators are said
     * to be equal iff both of them point to the same item or if both
     * of them are invalid.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    bool operator==
        (
        const FileIterator & it
        ) const;

    /**
     * Compares the iterators for non-equality.
     * "Negative" version of operator==.
     */
    bool operator!=
        (
        const FileIterator & it
        ) const;

    /**
     * Unbinds the iterator and rebinds it to the context specified by @c it.
     * The two contexts will be shared.
     */
    FileIterator & operator=
        (
        const FileIterator & it
        );

public:

    /**
     * Advances to the next file or subdirectory in the current directory.
     * @return Reference to this iterator.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    FileIterator & next();

    /**
     * Advances to the next file or subdirectory in the current directory.
     * Note that this operator returns reference to the iterator AFTER
     * performing the increment!!!
     * @return Reference to this iterator.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    FileIterator & operator++( int );

    /**
     * Resets the iterator - i.e. it will be pointing to the first file
     * or subdirectory in this directory. If the iterator is invalid,
     * the path and mask from the most recent binding are used to create
     * new binding.
     * @return Reference to this iterator.
     * @exception ExInvalidFsIterator Iterator is invalid.
     * @exception ExDirectoryNotFound The directory we want to rewind
     *            doesn't exist (might have been deleted meanwhile).
     */
    FileIterator & rewind();

public:

    /**
     * Enters the directory - returns a new iterator bound with the
     * entered directory; this iterator still remains valid and points
     * to the same item as before.<br>
     *
     * @return The newly created iterator bound with the entered directory.
     *
     * @exception ExInvalidFsIterator Iterator is invalid.
     * @exception ExNotADirectory Iterator doesn't point to a directory.
     * @exception ExDirectoryNotFound The directory we want to enter 
     *      into doesn't exist. This might happen if somebody deleted
     *      the directory the iterator had been pointing to.
     */
    FileIterator enter();

    /**
     * Enters the directory being referenced.
     * Updates iterator context instead of creating a new one.
     * @return Reference to this iterator.
     *
     * @exception ExInvalidFsIterator Iterator is invalid.
     * @exception ExNotADirectory Iterator doesn't point to a directory.
     * @exception ExDirectoryNotFound The directory we want to descent
     *      into doesn't exist. This might happen if somebody deleted
     *      the directory the iterator had been pointing to.
     */
    FileIterator & descent();

    /**
     * Returns a new iterator bound with the directory that is a direct
     * ascendant of the directory being referenced by this iterator; this
     * iterator will remain valid and will point to the same item as before.
     * <br>
     *
     * @return The newly created iterator bound with the directory that 
     *      is direct ascendant of the leaved one.
     *
     * @exception ExInvalidFsIterator Iterator is invalid.
     * @exception ExDirectoryNotFound The directory we want to leave
     *      doesn't exist. This might happen if somebody deleted
     *      the directory the iterator had been pointing to.
     *
     * @note When @c leave is called for an iterator bound with the
     *      root directory, the result will be another iterator bound
     *      with the root. Use @c is_in_root() for testing whether
     *      the iterator is bound with root.
     */
    FileIterator leave();

    /**
     * Leaves the directory.
     * Updates iterator context instead of creating a new one.
     * @return Reference to this iterator.
     *
     * @exception ExInvalidFsIterator Iterator is invalid.
     * @exception ExDirectoryNotFound The directory we want to ascent 
     *      from doesn't exist. This might happen if somebody deleted
     *      the directory the iterator had been pointing to.
     *
     * @note @c descent(), ascent() sequence does not preserve iterator
     *      position within the directory. The whole directory will be
     *      enumerated instead as if it was visited for the first time.
     * @note When @c leave is called for an iterator bound with the
     *      root directory, the result will be the same as if we
     *      called @c rewind().
     */
    FileIterator & ascent();

public:
    
    /**
     * @return true if the iterator is valid (i.e. points to a file or
     *      to a directory), false otherwise.
     */
    bool is_valid() const;

    /**
     * @return true if the iterator points to a file, false otherwise.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    bool is_file() const;

    /**
     * @return true if the iterator points to a directory, false otherwise.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    bool is_directory() const;

    /**
     * @return true if the iterator points to a directory "." or "..", 
     *      false otherwise.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    bool is_trivial_directory() const;

    /**
     * @return Name of the referenced file or directory
     *         The name will be relative to the parent directory.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    std::string name() const;

    /**
     * @return Size of the referenced file in bytes. Returns 0 in the case the
     *      referenced item was a directory.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    UInt64 size() const;

    /**
     * @return Absolute path to the directory bound with the iterator.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    std::string get_absolute_directory() const;

    /**
     * @return true if the iterator is bound with the root directory, 
     *      false otherwise.
     * @exception ExInvalidFsIterator Iterator is invalid.
     */
    bool is_in_root() const;

    /**
     * @return The current mask for filtering files.
     */
    std::string get_mask();

    /**
     * Sets the current mask for filtering files.
     */
    void set_mask
        (
        const std::string & mask
        );

    /**
     * @return The current directory enumeration mode. 
     */
    DirectoryEnumerationMode get_dir_enum_mode();

    /**
     * Sets the current directory enumeration mode.
     */
    void set_dir_enum_mode
        (
        DirectoryEnumerationMode mode
        );

public:

    /**
     * @return Pointer to the implementation object (@c impl). 
     * This is needed for better extensibility.
     */
    Reference<FileIteratorImpl> get_impl();

private:

    Reference<FileIteratorImpl> impl;
        /**< Shared pointer to implementation object.
             The object is architecture-dependent. */

    };  // class FileIterator


/**
 * FileIterator implementation object interface.
 * Holds FileSystem iterator context, specific to enumerated structure.
 * This is the "actual iterator".
 * For information on interface description, see @c FileIterator.
 */
class FileIteratorImpl : public SharedObject
    {
protected:

    friend class Massiv::Core::Test::FsEnumerationTest;
    friend class Massiv::Core::Test::VolumeEnumerationTest;

    typedef FileIterator::DirectoryEnumerationMode DirectoryEnumerationMode;

public:
    
    /**
     * Default constructor.
     * Initializes the iterator as invalid.
     */
    FileIteratorImpl
        (
        DirectoryEnumerationMode  mode
        );

    /**
     * Constructor.
     * Opens the current directory.
     */
    FileIteratorImpl
        (
        const std::string &                     mask,
        DirectoryEnumerationMode  mode
        );

    /**
     * Full-featured constructor.
     */
    FileIteratorImpl
        (
        const std::string &                     path,
        const std::string &                     mask,
        DirectoryEnumerationMode  mode
        );

protected:
    
    /**
     * Destructor.
     */
    virtual ~FileIteratorImpl();

private:

    NO_ASSIGNMENT_OPERATOR( FileIteratorImpl );

public:

    /**
     * Binds current directory to the iterator.
     */
    virtual void bind_current
        (
        const std::string & mask
        ) = 0;

    /**
     * Binds the iterator with directory specified by @c path.
     */
    virtual void bind
        (
        const std::string & path,
        const std::string & mask
        ) = 0;

    /**
     * Unbinds the iterator from the enumerated structure.
     */
    virtual void unbind() = 0;

public:

    /**
     * Advances iterator.
     */
    virtual Reference<FileIteratorImpl> next() = 0;

    /**
     * Resets the iterator.
     */
    virtual Reference<FileIteratorImpl> rewind() = 0;

public:

    /**
     * Creates new iterator referencing the subdirectory.
     */
    virtual Reference<FileIteratorImpl> enter() = 0;

    /**
     * Descents into the subdirectory.
     */
    virtual Reference<FileIteratorImpl> descent() = 0;

    /**
     * Creates new iterator referenced the parent directory.
     */
    virtual Reference<FileIteratorImpl> leave() = 0;

    /**
     * Ascents into the parent directory.
     */
    virtual Reference<FileIteratorImpl> ascent() = 0;

public:
    
    /**
     * Tests if the iterator is valid.
     */
    virtual bool is_valid() const = 0;

    /**
     * Tests if the iterators points to a file.
     */
    virtual bool is_file() const = 0;

    /**
     * Tests if the iterator points to a directory.
     */
    virtual bool is_directory() const = 0;

    /**
     * Tests if the iterator points to "." or ".." directory, 
     * false otherwise.
     */
    virtual bool is_trivial_directory() const = 0;

    /**
     * Returns name of the referenced file or directory (relative to the
     * parent directory).
     */
    virtual std::string name() const = 0;

    /**
     * Returns size of the referenced file in bytes.
     */
    virtual UInt64 size() const = 0;

    /**
     * Returns absolute path to the directory bound with the iterator.
     */
    virtual std::string get_absolute_directory() const = 0;

    /**
     * Tests if the iterator points to the root directory.
     */
    virtual bool is_in_root() const = 0;

    /**
     * Returns current mask that is used to filter directory entries.
     */
    std::string get_mask();

    /**
     * Sets current filter mask.
     */
    void set_mask
        (
        const std::string & mask
        );

    /**
     * Returns current directory enumeration mode. 
     */
    DirectoryEnumerationMode get_dir_enum_mode();

    /**
     * Sets directory enumeration mode.
     */
    void set_dir_enum_mode
        (
        DirectoryEnumerationMode mode
        );

protected:

    bool                        valid_flag;
        /**< True if the iterator is valid, i.e. it points to file or a 
             directory. */
    std::string                 mask;
        /**< Mask for filtering files. */
    DirectoryEnumerationMode    dir_enum_mode;
        /**< Enumeration mode. */

    };  // class FileIteratorImpl

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_FILE_ITERATOR_H
