/** @file file_system.h
 * File system abstraction.
 */

/* 
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FILE_SYSTEM_H
#define MASSIV_FILE_SYSTEM_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_FILE_ITERATOR_H
#include "filesys/file_iterator.h"
#endif

#include <string>

namespace Massiv {
namespace Core {


namespace Test { class FsEnumerationTest; }


/** @addtogroup filesys
 * @{
 */


/**
 * File system abstraction interface.
 * The interface allows to create/delete directories and enumerate
 * files stored in a hierarchical file-system-like structure.
 */
class FileSystem
    {
public:
        
    /**
     * Constructor.
     */
    FileSystem() { }

    /**
     * Destructor.
     */
    virtual ~FileSystem() { }

    NO_COPY_CONSTRUCTOR( FileSystem );
    NO_ASSIGNMENT_OPERATOR( FileSystem );

public:

    /**
     * Specifies how a directory should be deleted.
     */
    enum DirectoryDeleteMode
        {
        DELETE_DIRS_RECURSIVE       = 0,
            /**< Delete directories recursively; however, if some directory
                    contains files, it won't be deleted. */

        DELETE_ALL_RECURSIVE        = 1,
            /**< Delete all files and directories recursively. */

        DELETE_EMPTY                = 2,
            /**< If the directory is not empty, it won't be deleted. */
        }; // enum DirectoryDeleteMode

    typedef FileIterator::DirectoryEnumerationMode DirectoryEnumerationMode;

public:

    /**
     * Returns iterator bound with the current directory and the given
     * mask. This iterator can be used for enumerating files in the
     * current directory.
     *
     * @param mask Mask describing files that should be enumerated. May
     *      contain wildcards, i.e. * and ?. 
     */
    virtual FileIterator open_current_directory
        (
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const = 0;

    /**
     * Returns iterator bound with a given directory and mask that can be
     * used for enumerating files in the directory.
     *
     * @param path Path specifying the directory to be open.
     * @param mask Mask describing files that should be enumerated. May
     *      contain wildcards, i.e. * and ?.
     *
     * @return Iterator bound with the specified path and mask.
     */
    virtual FileIterator open_directory
        (
        const std::string &         path,
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const = 0;

    /**
     * @return Path to the current directory in a string.
     */
    virtual std::string get_current_directory() const = 0;

    /**
     * Sets current directory.
     *
     * @param path Specifies the prospective current directory. Absolute or
     *      relative.
     *
     * @exception ExCannotChangeDirectory Wrong or non-existent directory
     *      name was given.
     */
    virtual void set_current_directory
        (
        const std::string & path
        ) = 0;

    /**
     * Normalizes path, i.e. ensures that the path contains correct type of
     * slashes and that it has the correct termination slash if needed.
     *
     * @param path The path to be normalized.
     * @param slash_terminated If true, the result string will be 
     *      '/'-terminated (and vice versa).
     *
     * @bug Should remove unneccessary '.' and '..' (not implemented yet).
     */
    virtual void normalize_path
        (
        std::string &   path,
        bool            slash_terminated = true
        ) const = 0;

    /**
     * Creates directory given by a specified path. If more directories 
     * contained in the path don't exist, all will be created.
     * 
     * @param path Specifies a directory to be created. Can be either
     *      absolute or relative (taken relatively to the current directory).
     *
     * @exception ExCanNotCreateDirectory Directory can not be created.
     *
     * @note If more directories has to be created, no atomicity is
     *      ensured (some of the directories may be created and others
     *      not if some error occurs).
     */
    virtual void create_directory
        (
        const std::string & path
        ) = 0;

    /**
     * Deletes directory given by a specified path.
     * @param path Specifies a directory to be deleted. Can be either
     *      absolute or relative (taken relatively to the current directory).
     * @param mode Specifies whether the directory should be deleted even
     *      if it contains some other directories or files. Its values
     *      may be:
     *      <ul>
     *       <li><b>DELETE_DIRS_RECURSIVE:</b> If the deleted 
     *           directory contains other directories, they will be deleted 
     *           recursively. However, directories containing one or more 
     *           'regular' files will not be deleted.
     *       <li><b>DELETE_ALL_RECURSIVE:</b> The directory will 
     *           be deleted even with all subdirectories and contained files.
     *       <li><b>DELETE_EMPTY:</b> The directory will be 
     *           deleted only if it is empty.
     *      </ul>
     * @exception ExNotADirectory @c path does not specify a directory.
     * @exception ExCannotDeleteDirectory Couldn't delete the directory.
     * @exception ExDirectoryNotEmpty Directory was not empty and the 
     *      delete mode was DELETE_DIRS_RECURSIVE or DELETE_EMPTY.
     */
    virtual void delete_directory
        (
        const std::string & path,
        DirectoryDeleteMode mode = DELETE_EMPTY
        ) = 0;

    /**
     * Creates file refered by @c path.
     * @c path may refer to a directory that does not exist yet.
     * In that case the directory is created first.
     * @exception ExCanNotCreateFile File can not be created.
     * @exception ExCanNotCreateDirectory Directory can not be created.
     */
    virtual void create_file
        (
        const std::string & path
        ) = 0;

    /**
     * Deletes file refered by @c path.
     * @exception ExCanNotDeleteFile Access denied.
     * @exception ExNotAFile @c path does not specify a file.
     * @exception EXIOError Unspecified IO error.
     */
    virtual void delete_file
        (
        const std::string & path
        ) = 0;

    /**
     * Renames file refered by @c path.
     * @exception ExCanNotRenameFile Access denied.
     * @exception ExNotAFile @c from or @c to does not specify a file.
     * @exception EXIOError Unspecified IO error.
     */
    virtual void rename_file
        (
        const std::string & from,
        const std::string & to
        ) = 0;

    /**
     * Tests if @c path refers to a file.
     */
    virtual bool is_file
        (
        const std::string & path
        ) const = 0;
    
    /**
     * Tests if @c path refers to a directory.
     */
    virtual bool is_directory
        (
        const std::string & path
        ) const = 0;

    /**
     * Tests whether a given @c path is absolute.
     *
     * @param path The path (can contain even a filename).
     * @return @c true if the @c path is absolute, @c false otherwise.
     *
     * @note Doesn't check whether the path actualy exists.
     */
    virtual bool is_absolute
        (
        const std::string & path
        ) const = 0;

    /**
     * Tests whether a given @c path is relative.
     *
     * @param path The path (can contain even a filename).
     * @return @c true if the @c path is relative, @c false otherwise.
     */
    virtual bool is_relative
        (
        const std::string & path
        ) const = 0;

public:

    /**
     * Gets the character that is used as a file separator in paths.
     *
     * @return The file separator.
     */
    virtual char get_slash() const = 0;

    /**
     * Gets the reverted character to the one that is used as a file
     * separator in filesystem paths.
     *
     * @return The reverted file separator.
     */
    virtual char get_back_slash() const = 0;

    }; // class FileSystem

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_FILE_SYSTEM_H
