/** @file file_system_common.cpp
 * Class implementing traversing through the Win32/Unix filesystem directory graph.
 */

/* 
 * begin        : Mon Oct 6 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "global/global.h"
#include "utility/debug.h"
#include "utility/wildcards.h"
#include "utility/quote_string.h"
#include "filesys/native_file_system.h"
#include "filesys/file_iterator.h"
#include "filesys/file_system_common.h"
#include <iosfwd>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdio>
#include <cerrno>


namespace Massiv {
namespace Core {


#define CHECK_VALID()  \
    { \
    if( !is_valid() ) \
        { \
        throw ExInvalidFsIterator( "FS iterator is not valid." ); \
        } \
    }


// FileSystemCommon implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileIterator FileSystemCommon::open_current_directory
    (
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) const
    {
    return open_directory( get_current_directory(), mask, mode );
    }


std::string FileSystemCommon::get_current_directory() const
    {
    char slash = get_slash();
    std::string result = get_current_directory_API();

    if( result.size() > 0 )
        {
        if( result[ result.size() - 1 ] != slash )
            {
            result += slash;
            }
        }
    else
        {
        panic( "get_current_directory_API returned empty path." );
        }

    return result;
    }


void FileSystemCommon::set_current_directory
    (
    const std::string &     path
    )
    {
    std::string path_normalized( path );
    normalize_path( path_normalized );
    if( !set_current_directory_API( path_normalized ) )
        {
        std::ostringstream  ss;
        ss << "Cannot change directory to " << quote_const_string( path ) << " ("
           << get_last_error_message() << ")";
        throw ExCanNotChangeDirectory( ss.str() );
        }
    }


void FileSystemCommon::normalize_path
    (
    std::string &   path,
    bool            slash_terminated
    ) const
    {
    char            slash = get_slash();
    char            back_slash = get_back_slash();
    
    std::string::iterator it( path.begin() );
    while( it != path.end() )
        {
        if( *it == back_slash )
            {
            *it = slash;
            }
        it++;
        }

    if( path.size() > 0 )
        {
        if( ( slash_terminated ) && ( path[ path.size() - 1 ] != slash ) )
            {
            path += slash;
            }
        }
    else
        {
        panic( "Path is empty." );
        }
    }


void FileSystemCommon::create_directory
    (
    const std::string &     path
    )
    {
    std::string             component;
    std::string::size_type  pos = 0;
    std::string::size_type  pos_old = 0;
    while( pos != std::string::npos )
        {
        pos = path.find_first_of( "/\\", pos_old );
        component = path.substr( 0, pos );
        pos_old = pos + 1;
        if( component.empty() )
            {
            // If component is empty string:
            // - path is empty string: while-loop will end; or
            // - path is absolute and this is first component: skip it
            continue;
            }        
        if( !create_directory_API( component ) )
            {
            std::ostringstream  ss;
            ss << "Cannot create directory " << quote_const_string( path ) << " ("
               << get_last_error_message() << ")";
            throw ExCanNotCreateDirectory( ss.str() );
            }                
        }
    }


void FileSystemCommon::delete_directory
    (
    const std::string &     path,
    DirectoryDeleteMode     mode
    )
    {
    bool delete_permitted = true;

    if( ( mode == DELETE_DIRS_RECURSIVE ) || ( mode == DELETE_ALL_RECURSIVE ) )
        {
        // Delete recursively.

        FileIterator it( open_directory( path, "*" ) );
        while( it.is_valid() )
            {
            if( it.is_directory() && !it.is_trivial_directory() )
                {
                std::string subdir( path );
                if( ( subdir[ subdir.size() - 1 ] != '/' ) &&
                    ( subdir[ subdir.size() - 1 ] != '\\' ) )
                    {
                    subdir += get_slash();
                    }

                subdir.append( it.name() );

                try
                    {
                    delete_directory( subdir, mode );
                    }
                catch( ExDirectoryNotEmpty & )
                    {
                    delete_permitted = false;
                    }
                }
            else
            if( it.is_file() )
                {
                if( mode == DELETE_ALL_RECURSIVE )
                    {
                    std::string file_name( path );
                    if( ( file_name[ file_name.size() - 1 ] != '/' ) &&
                        ( file_name[ file_name.size() - 1 ] != '\\' ) )
                        {
                        file_name += get_slash();
                        }
                    file_name.append( it.name() );
                    if( !delete_file_API( file_name ) )
                        {
                        delete_permitted = false;
                        }
                    }
                else
                    {
                    delete_permitted = false;
                    }
                }
            it++;
            }  // while( it.is_valid() )
        }

    if( delete_permitted )
        {
        if( !delete_directory_API( path ) )
            {
            std::ostringstream  ss;
            ss << "Cannot delete directory " << quote_const_string( path ) << " ("
               << get_last_error_message() << ").";
            throw ExCanNotDeleteDirectory( ss.str() );
            }
        }
    else
        {
        std::ostringstream  ss;
        ss << "Directory " << quote_const_string( path ) << " not empty.";
        throw ExDirectoryNotEmpty( ss.str() );
        }
    }


void FileSystemCommon::create_file
    (
    const std::string & path
    )
    {
    std::size_t pos = path.find_last_of( "/\\" );
    std::string directory;

    if( pos != std::string::npos )
        {
        create_directory( path.substr( 0, pos ) );
        }

    std::ofstream stream( path.c_str(), std::ios_base::out );

    if( !stream.good() )
        {
        std::ostringstream ss;
        ss << "Can not create file " << quote_const_string( path ) << ".";
        throw ExCanNotCreateFile( ss.str() );
        }
    }


void FileSystemCommon::delete_file
    (
    const std::string & path
    )
    {
    if( ::remove( path.c_str() ) )
        {
        std::ostringstream ss;

        switch( errno )
            {
            case EACCES:
                {
                ss << "File " << quote_const_string( path ) << " can not be deleted.";
                throw ExCanNotDeleteFile( ss.str() );
                }
            case ENOENT:
                {
                ss << quote_const_string( path ) << " is not a file.";
                throw ExNotAFile( ss.str() );
                }

            default:
                {
                ss << "IO error while deleting " << quote_const_string( path ) << ".";
                throw ExIOError( ss.str() );
                }
            }
        }
    }


void FileSystemCommon::rename_file
    (
    const std::string & from,
    const std::string & to
    )
    {
    try
        {
        create_file( to );
        delete_file( to );
        }
    catch( ExCanNotCreateFile & e )
        {
        throw ExCanNotRenameFile( e.what() );
        }
    catch( ExCanNotDeleteFile & e )
        {
        throw ExCanNotRenameFile( e.what() );
        }

    if( ::rename( from.c_str(), to.c_str() ) )
        {
        std::ostringstream ss;

        switch( errno )
            {
            case EACCES:
                {
                ss << "File " << quote_const_string( from ) << " can not be renamed to " <<
                    quote_const_string( to ) << ".";
                throw ExCanNotRenameFile( ss.str() );
                }
            case ENOENT:
                {
                ss << quote_const_string( from ) << " is not a file.";
                throw ExNotAFile( ss.str() );
                }

            default:
                {
                ss << "IO error while renaming " << quote_const_string( from ) << " to " <<
                    quote_const_string( to ) << ".";
                throw ExIOError( ss.str() );
                }
            }
        }
    }


// FileSystemCommon::IteratorImpl implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

FileSystemCommon::IteratorImpl::IteratorImpl
    (
    DirectoryEnumerationMode    mode
    ) : FileIteratorImpl( mode )
    {
    valid_flag = false;
    }


FileSystemCommon::IteratorImpl::IteratorImpl
    (
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) : FileIteratorImpl( mask, mode )
    {
    valid_flag = false;
    }


FileSystemCommon::IteratorImpl::IteratorImpl
    (
    const std::string &         path,
    const std::string &         mask,
    DirectoryEnumerationMode    mode
    ) : FileIteratorImpl( path, mask, mode )
    {
    valid_flag = false;
    }


FileSystemCommon::IteratorImpl::~IteratorImpl()
    {
    // No functionality.
    }


void FileSystemCommon::IteratorImpl::bind_current
    (
    const std::string &     mask
    )
    {
    // MSVC6 can't handle this: FileSystem & fs( Global::native_file_system() );
    FileSystem & fs = Global::native_file_system();
    initial_current_directory = fs.get_current_directory();
    bind( initial_current_directory, mask );
    }


void FileSystemCommon::IteratorImpl::unbind()
    {
    CHECK_VALID();

    absolute_directory.clear();
    valid_flag = false;

    close_handle();
    invalidate_handle();
    }


Reference< FileIteratorImpl > FileSystemCommon::IteratorImpl::next()
    {
    return next_aux();
    }


Reference< FileIteratorImpl > FileSystemCommon::IteratorImpl::rewind()
    {
    if( is_handle_valid() )
        {
        close_handle();
        invalidate_handle();
        }
    bind( path, mask );
    return this;
    }


Reference< FileIteratorImpl > FileSystemCommon::IteratorImpl::ascent()
    {
    CHECK_VALID();

    if( close_handle() )
        {
        std::string     abs_dir = get_absolute_directory();
        std::string     path( abs_dir + ".." );

        valid_flag = false;
        invalidate_handle();
        bind( path, mask );
        }
    else
        {
        valid_flag = false;
        invalidate_handle();
        // fixme: throw exception or ignore?
        }

    return this;
    }


bool FileSystemCommon::IteratorImpl::is_valid() const
    {
    return valid_flag && is_handle_valid();
    }


bool FileSystemCommon::IteratorImpl::is_trivial_directory() const
    {
    CHECK_VALID();
    return is_directory() && ( ( name() == "." ) || ( name() == ".." ) );
    }


#undef CHECK_VALID


} // namespace Core
} // namespace Massiv
