/** @file file_system_common.h
 * Win32/Unix file system abstraction.
 */

/* 
 * begin        : Mon Oct 6 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FILE_SYSTEM_COMMON_H
#define MASSIV_FILE_SYSTEM_COMMON_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_FILE_ITERATOR_H
#include "filesys/file_iterator.h"
#endif

#ifndef MASSIV_FILE_SYSTEM_H
#include "filesys/file_system.h"
#endif

#ifndef MASSIV_NATIVE_FILE_SYSTEM_H
#include "filesys/native_file_system.h"
#endif

#include <string>

namespace Massiv {
namespace Core {


namespace Test { class FsEnumerationTest; }


/** @addtogroup filesys
 * @{
 */


/**
 * Win32/Unix common file system abstraction interface.
 */
class FileSystemCommon : public NativeFileSystem
    {
public:
        
    /**
     * Iterator that enumerates files from Win32/Unix file system.
     */
    class IteratorImpl : public FileIteratorImpl
        {

    protected:
        
        /**
         */
        IteratorImpl
            (
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         */
        IteratorImpl
            (
            const std::string &         mask,
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         */
        IteratorImpl
            (
            const std::string &         path,
            const std::string &         mask,
            DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
            );

        /**
         */
        virtual ~IteratorImpl();

    private: 

        NO_ASSIGNMENT_OPERATOR( IteratorImpl );

    protected:

        /**
         */
        virtual void bind_current
            (
            const std::string & mask
            );

        /**
         */
        virtual void bind
            (
            const std::string &     path,
            const std::string &     mask
            ) = 0;

        /**
         */
        virtual void unbind();

        /**
         */
        virtual Reference< FileIteratorImpl > next();

        /**
         */
        virtual Reference< FileIteratorImpl > rewind();


        /**
         */
        virtual Reference< FileIteratorImpl > enter() = 0;

        /**
         */
        virtual Reference< FileIteratorImpl > descent() = 0;

        /**
         */
        virtual Reference< FileIteratorImpl > leave() = 0;

        /**
         */
        virtual Reference< FileIteratorImpl > ascent();

        /**
         */
        virtual bool is_valid() const;

        /**
         */
        virtual bool is_file() const = 0;

        /**
         */
        virtual bool is_directory() const = 0;

        /**
         */
        virtual bool is_trivial_directory() const;

        /**
         */
        virtual std::string name() const = 0;

        /**
         */
        virtual UInt64 size() const = 0;

        /**
         */
        virtual std::string get_absolute_directory() const = 0;

        /**
         */
        virtual bool is_in_root() const = 0;

        /**
         * Return true if handle is valid.
         */
        virtual bool is_handle_valid() const = 0;        

        /**
         * Invalidate handle.
         */
        virtual void invalidate_handle() = 0;

        /**
         * Close handle.
         */
        virtual bool close_handle() = 0;
                
    private:

        /**
         * Advances to the next file or subdirectory in the current directory.
         * @return Pointer to this iterator.
         * @exception ExInvalidFsIterator Iterator is invalid.
         */
        virtual FileIteratorImpl * next_aux() = 0;

    protected:

        std::string         path;
            /**< Path that was given for binding the iterator with a 
                 directory. Before storing the path here, all bash slashes are
                 converted to slashes and the path is terminated by slash. */

        std::string         initial_current_directory;
            /**< The current directory at the time of creation of this 
                 iterator or its binding with new directory. The format
                 of the path remains the same that has been returned from

                 GetCurrentDirectory(). */

        mutable std::string  absolute_directory;
            /**< Contains either an empty string or the absolute path to the
                 directory bounded with this iterator. Before storing a path
                 here, all back-slashes are turned into slashes and the string is
                 terminated by slashes. */

        }; // class IteratorImpl

    /**
     * Constructor.
     */
    FileSystemCommon() { }

    /**
     * Destructor.
     */
    virtual ~FileSystemCommon() { }

    NO_COPY_CONSTRUCTOR( FileSystemCommon );
    NO_ASSIGNMENT_OPERATOR( FileSystemCommon );

public:

    /**
     */
    virtual FileIterator open_current_directory
        (
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const;

    /**
     */
    virtual FileIterator open_directory
        (
        const std::string &         path,
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const = 0;

    /**
     */
    virtual std::string get_current_directory() const;

    /**
     */
    virtual void set_current_directory
        (
        const std::string & path
        );

    /**
     */
    virtual void normalize_path
        (
        std::string &   path,
        bool            slash_terminated = true
        ) const;

    /**
     */
    virtual void create_directory
        (
        const std::string & path
        );

    /**
     */
    virtual void delete_directory
        (
        const std::string & path,
        DirectoryDeleteMode mode = DELETE_EMPTY
        );

    /**
     */
    virtual void create_file
        (
        const std::string & path
        );

    /**
     */
    virtual void delete_file
        (
        const std::string & path
        );

    /**
     */
    virtual void rename_file
        (
        const std::string & from,
        const std::string & to
        );

    /**
     */
    virtual bool is_file
        (
        const std::string & path
        ) const = 0;
    
    /**
     */
    virtual bool is_directory
        (
        const std::string & path
        ) const = 0;
            
    /**
     * Implementation calls API function.
     * Return current directory.
     */
    virtual std::string get_current_directory_API() const = 0;

    /**
     * Implementation calls API function.
     * Set current directory.
     * Return true if directory was set.
     */
    virtual bool set_current_directory_API
        (
        const std::string &     directory
        ) const = 0;

    /**
     * Implementation calls API function.
     * Create directory.
     * Return true if directory was created or already existed.
     */
    virtual bool create_directory_API
        (
        const std::string &     directory
        ) const = 0;

    /**
     * Implementation calls API function.
     * Delete file.
     * Return true if file was deleted.
     */
    virtual bool delete_file_API
        (
        const std::string &     file_name
        ) const = 0;

    /**
     * Implementation calls API function.
     * Delete directory.
     * Return true if directory was deleted.
     */
    virtual bool delete_directory_API
        (
        const std::string &     directory
        ) const = 0;
        
    /**
     * Return last error message.
     */
    virtual std::string get_last_error_message() const = 0;
    
    }; // class FileSystemCommon

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_FILE_SYSTEM_COMMON_H
