/** @file simple_block_pool.cpp
 * Simple block pool.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/simple_block_pool.h"
#include "global/global.h"
#include "utility/debug.h"
#include <sstream>


namespace Massiv {
namespace Core {


// Implementation of SimpleBlockPool::Block
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SimpleBlockPool::Block::Block()
    {
    index = INVALID_BLOCK;
    pool = NULL;
    data = NULL;
    dirty = false;
    reference_count = 0;
    }


SimpleBlockPool::Block::Block
    (
    BlockIndex          index,
    SimpleBlockPool *   pool
    ) : index( index ),
        pool( pool ),
        data( NULL ),
        dirty( false ),
        reference_count( 0 )
    {
    }


SimpleBlockPool::Block::Block
    (
    const Block & block
    )
    {
    hope( !block.data );

    index = block.index;
    pool = block.pool;
    data = NULL;
    dirty = false;
    reference_count = 0;
    }


SimpleBlockPool::Block::~Block()
    {
    hope( !reference_count );
    hope( !data );
    }


void SimpleBlockPool::Block::unlock
    (
    bool modified
    )
    {
    #ifdef OPTION_DEBUG
    
    if( modified && !pool->can_write() )
        {
        Global::log_warning
            (
            Status::FACILITY_FILE_SYSTEM,
            Status::PRIORITY_HIGH,
            "We have a dirty block in a read-only pool. The changes will be lost."
            );
        }

    #endif

    dirty |= modified;
    
    if( modified )
        {
        pool->block_modified( index );
        }

    LRUItem::unlock();
    }


UInt8 & SimpleBlockPool::Block::operator[]
    (
    int offset
    )
    {
    hope( data );
    return data[ offset ];
    }


SimpleBlockPool::BlockIndex SimpleBlockPool::Block::get_index()
    {
    return index;
    }


int SimpleBlockPool::Block::get_block_size()
    {
    hope( pool );
    return pool->block_size;
    }


SimpleBlockPool * SimpleBlockPool::Block::get_pool() const
    {
    hope( pool );
    return pool;
    }


UInt8 * SimpleBlockPool::Block::get_data()
    {
    hope( data );
    return data;
    }


int SimpleBlockPool::Block::memory_size()
    {
    return get_block_size();
    }


void SimpleBlockPool::Block::cache_in()
    {
    hope( !data );
    hope( pool );
    hope( !dirty );

    data = new UInt8[ get_block_size() ];
    pool->load_block( index, data );
    }


void SimpleBlockPool::Block::cache_out()
    {
    hope( pool );
    hope( data );
    hope( !dirty );

    delete [] data;
    data = NULL;
    }


void SimpleBlockPool::Block::flush()
    {
    hope( pool );
    hope( data );
    hope( dirty );
    
    if( pool->can_write() )
        {
        pool->save_block( index, data );
        }
    else
        {
        Global::log_warning
            (
            Status::FACILITY_FILE_SYSTEM,
            Status::PRIORITY_HIGH,
            "Discarding a dirty block because the pool is in read-only mode."
            );
        }
    
    dirty = false;
    }


bool SimpleBlockPool::Block::is_dirty()
    {
    return dirty;
    }


std::string SimpleBlockPool::Block::get_name()
    {
    std::ostringstream ss;

    ss << "block$" << index;

    return ss.str();
    }


// Implementation of SimpleBlockPool::BlockRef
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SimpleBlockPool::BlockRef::BlockRef() :
    block( NULL ),
    read_offset( INVALID_OFFSET ),
    write_offset( INVALID_OFFSET )
    {
    }


SimpleBlockPool::BlockRef::BlockRef
    (
    Block * block
    ) : block( block ),
        read_offset( 0 ),
        write_offset( 0 )
    {
    if( block )
        {
        block->reference_count++;
        }
    }


SimpleBlockPool::BlockRef::BlockRef
    (
    const BlockRef & b
    ) : block( b.block ),
        read_offset( b.read_offset ),
        write_offset( b.write_offset )
    {
    if( block )
        {
        block->reference_count++;
        }
    }


SimpleBlockPool::BlockRef::~BlockRef()
    {
    if( block )
        {
        block->reference_count--;
        }
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator=
    (
    const BlockRef & b
    )
    {
    if( block )
        {
        block->reference_count--;
        }

    block = b.block;
    read_offset = b.read_offset;
    write_offset = b.write_offset;

    if( block )
        {
        block->reference_count++;
        }

    return *this;
    }


void SimpleBlockPool::BlockRef::lock()
    {
    hope( block );
    block->lock();
    }


void SimpleBlockPool::BlockRef::unlock
    (
    bool modified,
    bool flush
    )
    {
    hope( block );
    block->unlock( modified );

    /* NOTE: @c unlock() calls @c pool->block_modified() which could flush
       all the blocks. Thus @c block need not be dirty any more and it
       must be tested now. Flushing is allowed only if the block is
       unlocked as a modified block. */

    if( flush && block->is_dirty() )
        {
        hope( modified );
        block->flush();
        }
    }


UInt8 & SimpleBlockPool::BlockRef::operator[]
    (
    int offset
    )
    {
    hope( block );
    hope( offset >= 0 && offset < block->get_block_size() );

    return (*block)[ offset ];
    }


SimpleBlockPool::BlockIndex SimpleBlockPool::BlockRef::get_index()
    {
    hope( block );
    return block->get_index();
    }


SimpleBlockPool * SimpleBlockPool::BlockRef::get_pool() const
    {
    hope( block );
    return block->get_pool();
    }


UInt8 * SimpleBlockPool::BlockRef::get_data()
    {
    hope( block );
    return block->get_data();
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::seekp
    (
    int                 offset,
    std::ios::seekdir   direction
    )
    {
    hope( block );

    switch( direction )
        {
        case std::ios::beg:
            write_offset = offset;
            break;
        case std::ios::cur:
            write_offset += offset;
            break;
        case std::ios::end:
            write_offset = block->get_block_size() - offset - 1;
            break;
        default:
            panic( "Unexpected direction in seekp." );
        }

    hope( write_offset >= 0 && write_offset < block->get_block_size() );

    return *this;
    }


int SimpleBlockPool::BlockRef::tellp()
    {
    return write_offset;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator<<
    (
    UInt8   byte
    )
    {
    ( *this )[ write_offset ] = byte;
    write_offset++;

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator<<
    (
    Int32   integer
    )
    {
    ( *this )[ write_offset + 0 ] = ( integer >> 0 ) & 0xff;
    ( *this )[ write_offset + 1 ] = ( integer >> 8 ) & 0xff;
    ( *this )[ write_offset + 2 ] = ( integer >> 16 ) & 0xff;
    ( *this )[ write_offset + 3 ] = ( integer >> 24 ) & 0xff;
    write_offset += 4;

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator<<
    (
    Int64   large_integer
    )
    {
    operator<<( static_cast<Int32>( large_integer & 0xffffffff ) );
    operator<<( static_cast<Int32>( ( large_integer >> 32 ) & 0xffffffff ) );

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator<<
    (
    const std::string & string
    )
    {
    int size;
    int i;

    size = string.size();

    operator<<( size );

    for( i = 0; i < size; i++, write_offset++ )
        ( *this )[ write_offset ] = string[ i ];

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::seekg
    (
    int                 offset,
    std::ios::seekdir   direction
    )
    {
    hope( block );

    switch( direction )
        {
        case std::ios::beg:
            read_offset = offset;
            break;
        case std::ios::cur:
            read_offset += offset;
            break;
        case std::ios::end:
            read_offset = block->get_block_size() - offset - 1;
            break;
        default:
            panic( "Unexpected direction in seekg." );
        }

    hope( read_offset >= 0 && read_offset < block->get_block_size() );

    return *this;
    }


int SimpleBlockPool::BlockRef::tellg()
    {
    return read_offset;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator>>
    (
    UInt8 & byte
    )
    {
    byte = ( *this )[ read_offset ];
    read_offset++;

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator>>
    (
    Int32 & integer
    )
    {
    integer = ( ( *this )[ read_offset + 0 ] << 0 ) |
              ( ( *this )[ read_offset + 1 ] << 8 ) |
              ( ( *this )[ read_offset + 2 ] << 16 ) |
              ( ( *this )[ read_offset + 3 ] << 24 );
    read_offset += 4;

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator>>
    (
    Int64 & large_integer
    )
    {
    Int32 low;
    Int32 high;

    operator>>( low );
    operator>>( high );

    large_integer = static_cast<UInt32>( low );
    large_integer |= static_cast<Int64>( high ) << 32;

    return *this;
    }


SimpleBlockPool::BlockRef & SimpleBlockPool::BlockRef::operator>>
    (
    std::string & string
    )
    {
    int size;
    int i;

    operator>>( size );
    string.resize( size );

    for( i = 0; i < size; i++, read_offset++ )
        string[ i ] = ( *this )[ read_offset ];

    return *this;
    }


// Implementation of SimpleBlockPool
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SimpleBlockPool::SimpleBlockPool
    (
    int block_size,
    int cache_size
    ) : block_cache( cache_size, true ),
        tidy_counter( CACHE_TIDYING ),
        block_size( block_size )
    {
    }


SimpleBlockPool::~SimpleBlockPool()
    {
    hope( !blocks.size() );
    }


void SimpleBlockPool::flush_pool()
    {
    block_cache.flush( LRUCache::FLUSH_UNLOCKED );
    tidy_cache();
    }


void SimpleBlockPool::close_pool()
    {
    block_cache.unregister_all();
    blocks.clear();
    }


bool SimpleBlockPool::can_write() const
    {
    return true;
    }


void SimpleBlockPool::block_modified
    (
    BlockIndex index
    )
    {
    }


SimpleBlockPool::BlockRef SimpleBlockPool::access_block
    (
    BlockIndex  index
    )
    {
    /* Tidy cache. */

    if( tidy_counter-- <= 0 )
        {
        tidy_counter = CACHE_TIDYING;
        tidy_cache();
        }

    /* Access block. */

    BlockList::iterator it;
    Block * block;

    hope( index >= 0 );

    it = blocks.find( index );

    if( it != blocks.end() )
        {
        /* Block is already registered. */

        block = &it->second;
        }
    else
        {
        /* Register new block. */

        BlockList::value_type value( index, Block( index, this ) );

        it = blocks.insert( value ).first;
        block = &it->second;

        block_cache.register_item( block );
        }

    return block;
    }


int SimpleBlockPool::get_block_size() const
    {
    return block_size;
    }


void SimpleBlockPool::tidy_cache()
    {
    BlockList::iterator it;
    Block *             block;

    for( it = blocks.begin(); it != blocks.end(); )
        {
        block = &it->second;

        if( !block->reference_count )
            {
            block_cache.unregister_item( block );
            blocks.erase( it++ );
            }
        else
            {
            it++;
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
