/** @file simple_block_pool.h
 * Simple block pool.
 */

/*
 * begin        : Fri 28 Jun 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SIMPLE_BLOCK_POOL_H
#define MASSIV_SIMPLE_BLOCK_POOL_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_LRU_CACHE_H
#include "utility/lru_cache.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <iostream>
#include <string>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/** @class SimpleBlockPool
 * Simple class managing access to cacheable blocks of fixed size.
 */
class SimpleBlockPool
    {
public:

    /**
     * Constants.
     */
    enum Constants
        {
        INVALID_BLOCK   = -1,
            /**< Invalid block index. */
        INVALID_OFFSET  = -1,
            /**< Invalid offset. */
        }; // Constants

    /**
     * Parameters.
     */
    enum Parameters
        {
        CACHE_TIDYING =     16384
            /**< How "often" to tidy the block cache. */
        }; // Parameters

    typedef int BlockIndex;

    class BlockRef;

private:

    /**
     * Block that repesents data.
     */
    class Block : public LRUCache::LRUItem
        {
    private:

        friend class BlockRef;
        friend class SimpleBlockPool;

    public:

        /**
         * Default constructor.
         */
        Block();

        /**
         * Constructor.
         */
        Block
            (
            BlockIndex          index,
            SimpleBlockPool *   pool
            );

        /**
         * Copy constructor.
         */
        Block
            (
            const Block & block
            );

        /**
         * Destructor.
         */
        ~Block();

    public:

        /**
         * Unlocks the block.
         * Client is responsible for setting @c modified properly,
         * otherwise changes made to the block may be lost.
         */
        void unlock
            (
            bool modified
            );

        /**
         * Returns reference to a byte at specified offset.
         * Block must be locked.
         */
        UInt8 & operator[]
            (
            int offset
            );

        /**
         * Returns block index.
         */
        BlockIndex get_index();

        /**
         * Returns block size.
         */
        int get_block_size();

        /**
         * Returns pointer to the pool that owns this block.
         */
        SimpleBlockPool * get_pool() const;

        /**
         * Returns pointer to block data.
         * Block must be locked.
         * The pointer remain valid until the block is unlocked.
         */
        UInt8 * get_data();

    protected:

        //
        // LRUItem interface implementation.
        //

        /**
         * Returns block size.
         */
        virtual int memory_size();

        /**
         * Allocates and loads block data.
         */
        virtual void cache_in();

        /**
         * Releases block data.
         */
        virtual void cache_out();
        
    public:

        /**
         * Saves block data.
         * Block must be dirty.
         */
        virtual void flush();

        /**
         * Tests if the block is dirty (= has been modified since it was
         * (loaded).
         * The block is made dirty at @c unlock() time.
         */
        virtual bool is_dirty();

        /**
         * Returns block identification.
         */
        virtual std::string get_name();

    private:

        BlockIndex          index;
            /**< Block index. */
        SimpleBlockPool *   pool;
            /**< Pointer to block pool that owns this block. */
        UInt8 *             data;
            /**< Pointer to block data. May be NULL if block is not
                 locked in memory. */
        bool                dirty;
            /**< Is block dirty? */
        int                 reference_count;
            /**< Number of BlockRef objects referencing this block. */
        }; // Block

    typedef ext::hash_map<BlockIndex, Block> BlockList;

    friend class Block;

public:

    /**
     * Reference to cachable block.
     * Since block data is cached, each access must be coupled by lock() and 
     * unlock(). lock() ensures that data will be present in memory.
     */
    class BlockRef
        {
    public:

        /**
         * Default constructor.
         */
        BlockRef();

        /**
         * Constructor.
         */
        BlockRef
            (
            Block * block
            );

        /**
         * Copy constructor.
         */
        BlockRef
            (
            const BlockRef & b
            );

        /**
         * Destructor.
         */
        ~BlockRef();

        /**
         * Operator=.
         */
        BlockRef & operator=
            (
            const BlockRef & b
            );

    public:

        /**
         * Locks the block.
         */
        void lock();

        /**
         * Unlocks the block.
         * Client must set @c modified to true if the block has been modified
         * since previous lock().
         * If @c flush is true then @c modified must also be true and the block
         * will be immediatelly saved.
         */
        void unlock
            (
            bool modified = true,
            bool flush = false
            );

        /**
         * Returns reference to a byte in the block at specified offset.
         * Block must be locked.
         */
        UInt8 & operator[]
            (
            int offset
            );

        /**
         * Returns block index.
         */
        BlockIndex get_index();

        /**
         * Returns pointer to the pool that owns the referenced block.
         */
        SimpleBlockPool * get_pool() const;

        /**
         * Returns pointer to block data.
         */
        UInt8 * get_data();

    public:

        //
        // Writing to block.
        //

        /**
         * Moves writer pointer.
         */
        BlockRef & seekp
            (
            int                 offset,
            std::ios::seekdir   direction
            );

        /**
         * Returns current writer pointer.
         */
        int tellp();

        /**
         * Writes UInt8.
         */
        BlockRef & operator<<
            (
            UInt8   byte
            );

        /**
         * Writes Int32.
         */
        BlockRef & operator<<
            (
            Int32   integer
            );

        /**
         * Writes Int64.
         */
        BlockRef & operator<<
            (
            Int64   large_integer
            );

        /**
         * Writes string.
         */
        BlockRef & operator<<
            (
            const std::string & string
            );

    public:

        //
        // Reading from block.
        //

        /**
         * Moves reader pointer.
         */
        BlockRef & seekg
            (
            int                 offset,
            std::ios::seekdir   direction
            );

        /**
         * Returns current reader pointer.
         */
        int tellg();

        /**
         * Reads UInt8.
         */
        BlockRef & operator>>
            (
            UInt8 & byte
            );

        /**
         * Reads Int32.
         */
        BlockRef & operator>>
            (
            Int32 & integer
            );

        /**
         * Reads Int64.
         */
        BlockRef & operator>>
            (
            Int64 & large_integer
            );

        /**
         * Reads string.
         * Note: string may be safely retrieved by operator<< even
         * if contains spaces (this is a difference from IO streams).
         */
        BlockRef & operator>>
            (
            std::string & string
            );

    private:

        Block * block;
            /**< Native pointer to block. */
        int     read_offset;
            /**< Current reader pointer. */
        int     write_offset;
            /**< Current writer pointer. */

        }; // BlockRef

protected:

    /**
     * Constructor.
     */
    SimpleBlockPool
        (
        int block_size,
        int cache_size
        );

    /**
     * Destructor.
     * Every block must be unreferenced, @c close_pool() is supposed
     * to have been called before.
     */
    virtual ~SimpleBlockPool();

    /**
     * Flushes all unlocked dirty blocks.
     */
    virtual void flush_pool();

    /**
     * This is a replacement for destructor.
     * Must be called by the most-derived class from its destructor.
     */
    virtual void close_pool();
    
    /**
     * Tests if the pool is in read-write mode.
     * By default, the pool is writeable.
     * However subclasses may override this and can make pool read only
     * until the pool is successfully opened. 
     * @note Read-only mode does not prevent client from modifying
     * the pool. However the changes will be discarded when the block is
     * removed from the cache.
     */
    virtual bool can_write() const;

    /**
     * Client-defined callback.
     * The implementation should load data for block @c index
     * and place it at @c ptr.
     * If requested block has not been yet backed up on the storage,
     * block should be zero-filled.
     */
    virtual void load_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        ) = 0;

    /**
     * Client-defined callback.
     * The implementation should save data for block @c index
     * pointed by @c ptr to a backing up store so that it
     * could be later retrieved by @c load_block().
     */
    virtual void save_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        ) = 0;

    /**
     * Client-defined callback.
     * Called by the pool when a dirty block is unlocked.
     */
    virtual void block_modified
        (
        BlockIndex  index
        );
                 
protected:

    /**
     * Accesses block.
     * Returns reference to block @c index.
     */
    BlockRef access_block
        (
        BlockIndex  index
        );

    /**
     * Returns block size.
     */
    int get_block_size() const;

    /**
     * Unregisters unreferenced blocks from block cache.
     */
    void tidy_cache();

private:

    LRUCache    block_cache;
        /**< Block cache. */
    BlockList   blocks;
        /**< List of known blocks. */
    int         tidy_counter;
        /**< When the counter drops to zero we will tidy block cache.
             @c tidy_counter is then reinitialized to @c CACHE_TIDYING. */

protected:

    Int32       block_size;
        /**< Block size. */

    }; // SimpleBlockPool

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_SIMPLE_BLOCK_POOL_H
