/** @file stream_pool.cpp
 * Stream pool.
 */

/*
 * begin        : Mon 12 Aug 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "utility/debug.h"
#include "filesys/stream_pool.h"
#include <sstream>

namespace Massiv {
namespace Core {


// Implementation of StreamPool::StreamDescriptor
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

StreamPool::StreamDescriptor::StreamDescriptor()
    {
    init();
    }


StreamPool::StreamDescriptor::StreamDescriptor
    (
    const StreamDescriptor & s
    )
    {
    hope( !s.num_writers && !s.num_readers );

    // Copy.

    pool = s.pool;
    first = s.first;
    allocated_length = s.allocated_length;
    index_map = NULL;
    current_length = s.current_length;
    block_size = s.block_size;

    num_writers = 0;
    num_readers = 0;

    // Invalidate the other descriptor.

    s.pool = NULL;
    }


StreamPool::StreamDescriptor::~StreamDescriptor()
    {
    // If this fails, stream described by this descriptor is still opened.

    hope( !num_readers && !num_writers );
    hope( !index_map );
    }


StreamPool::StreamDescriptor & StreamPool::StreamDescriptor::operator=
    (
    const StreamDescriptor &s
    )
    {
    hope( !s.num_writers && !s.num_readers );
    hope( !index_map );
        // Should be zero because there are neither readers nor writers
        // to this stream.

    // Copy.

    pool = s.pool;
    first = s.first;
    allocated_length = s.allocated_length;
    index_map = NULL;
    current_length = s.current_length;
    block_size = s.block_size;

    num_writers = 0;
    num_readers = 0;

    // Invalidate the other descriptor.

    s.pool = NULL;

    return *this;
    }


bool StreamPool::StreamDescriptor::is_valid() const
    {
    return pool != NULL;
    }


void StreamPool::StreamDescriptor::write
    (
    std::ostream & str
    ) const
    {
    str << first << " ";
    str << current_length << " ";
    }


void StreamPool::StreamDescriptor::restore
    (
    StreamPool *    owner_pool,
    std::istream &  str
    )
    {
    int stream_blocks;

    hope( !pool );
    hope( !num_writers && !num_readers );

    pool = owner_pool;
    block_size = pool->block_size;

    // Read descriptor.

    str >> first;
    str >> current_length;

    // Analyze descriptor.

    if( first == INVALID_BLOCK )
        {
        // We have restored an empty stream descriptor.

        pool = NULL;

        if( allocated_length || current_length )
            {
            throw ExIntegrityError( "Invalid empty stream descriptor." );
            }
        }
    else
        {
        // We have restored a non-empty descriptor.

        if( pool->get_prev_block( first ) != INVALID_BLOCK )
            {
            throw ExIntegrityError( "Damaged block chain." );
            }

        stream_blocks = pool->get_chain_length( first );
        allocated_length = stream_blocks * block_size;
        }
    }


void StreamPool::StreamDescriptor::notify_change()
    {
    hope( pool != NULL );
    pool->modified();
    }


void StreamPool::StreamDescriptor::init()
    {
    pool = NULL;
    first = INVALID_BLOCK;

    allocated_length = 0;
    index_map = NULL;
    current_length = 0;
    block_size = 0;

    num_writers = 0;
    num_readers = 0;
    }


StreamPool::BlockIndex StreamPool::StreamDescriptor::get_block
    (
    int index_in_chain
    ) const
    {
    BlockIndex block;

    if( index_map )
        {
        block = index_map->at( index_in_chain );
        hope( block == pool->get_block( first, index_in_chain ) );
        }
    else
        {
        block = pool->get_block( first, index_in_chain );
        hope( block != INVALID_BLOCK );
        }

    return block;
    }


StreamPool::BlockIndex StreamPool::StreamDescriptor::get_last_block() const
    {
    BlockIndex block;

    block = get_block( get_num_stream_blocks() - 1 );
    hope( pool->get_next_block( block ) == INVALID_BLOCK );

    return block;
    }


void StreamPool::StreamDescriptor::open_close_notify()
    {
    if( !num_readers && !num_writers )
        {
        // Stream was totally closed.
        // Delete index map.

        if( index_map )
            {
            delete index_map;
            index_map = NULL;
            }
        }
    else
        {
        // Stream was opened.
        // Reconstruct index map.

        if( !index_map )
            {
            BlockIndex  block;
            int         i;

            index_map = new IndexMap();
            block = first;

            for( i = 0; i < allocated_length / block_size; i++ )
                {
                index_map->push_back( block );
                block = pool->get_next_block( block );
                }

            // Now we should be at the end of the block chain.

            hope( block == INVALID_BLOCK );
            }
        }
    }


void StreamPool::StreamDescriptor::grow
    (
    int length
    )
    {
    BlockIndex  block;
    int         blocks_required;
    int         stream_blocks;
    int         delta;

    // Test.

    blocks_required = get_num_blocks_required( length );
    stream_blocks = get_num_stream_blocks();

    if( blocks_required <= stream_blocks )
        {
        return;
        }

    // Grow.

    delta = blocks_required - stream_blocks;
    block = get_last_block();
    pool->append_blocks( block, delta );

    // Update index map for appended blocks.

    while( stream_blocks < blocks_required )
        {
        pool->notify_block_usage( block, block_size );
        block = pool->get_next_block( block );

        if( index_map )
            {
            index_map->push_back( block );
            }

        allocated_length += block_size;
        stream_blocks++;
        }

    hope( pool->get_next_block( block ) == INVALID_BLOCK );

    // Note: ~StreamBuffer will truncate the stream to the
    //       correct length.

    // Done.

    notify_change();
    }


void StreamPool::StreamDescriptor::truncate
    (
    int length
    )
    {
    BlockIndex  block, unused;
    int         blocks_required;
    int         stream_blocks;

    // Test.

    hope( !num_writers );

    blocks_required = get_num_blocks_required( length );
    stream_blocks = get_num_stream_blocks();

    if( length >= current_length )
        {
        return;
        }

    // Truncate.

    if( blocks_required < stream_blocks )
        {
        block = get_block( blocks_required );

        pool->notify_block_usage( block, length % block_size );
        unused = pool->get_next_block( block );

        while( unused != INVALID_BLOCK )
            {
            pool->notify_block_usage( unused, 0 );
            unused = pool->get_next_block( unused );
            }

        pool->erase_chain( block );

        if( index_map )
            {
            index_map->resize( blocks_required );
            }
        }
    else
        {
        hope( blocks_required == stream_blocks );

        block = get_last_block();
        pool->notify_block_usage( block, length % block_size );
        }

    // Done.

    current_length = length;
    allocated_length = blocks_required * block_size;

    notify_change();
    }


int StreamPool::StreamDescriptor::block_size_align
    (
    int value
    ) const
    {
    return ( value + block_size - 1 ) / block_size * block_size;
    }


int StreamPool::StreamDescriptor::get_num_blocks_required
    (
    int length
    ) const
    {
    int blocks_required;

    blocks_required = block_size_align( length ) / block_size;

    if( !blocks_required )
        {
        // Each block chain consists of at least one block.

        blocks_required = 1;
        }

    return blocks_required;
    }


int StreamPool::StreamDescriptor::get_num_stream_blocks() const
    {
    int stream_blocks;

    if( index_map )
        {
        stream_blocks = index_map->size();

        hope( stream_blocks == allocated_length / block_size );
        }
    else
        {
        stream_blocks = allocated_length / block_size;
        }

    hope( stream_blocks > 0 );
    hope( pool->get_chain_length( first ) == stream_blocks );

    return stream_blocks;
    }


int StreamPool::StreamDescriptor::get_current_length() const
    {
    return current_length;
    }


// Implementation of StreamPool::StreamBuffer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

StreamPool::StreamBuffer::StreamBuffer()
      : open_mode( std::ios_base::openmode( 0 ) ),
        put_offset( 0 ),
        get_offset( 0 ),
        stream( NULL ),
        read_block_locked( false ),
        write_block_locked( false )
    {
    // Initialize std::streambuf.

    // Setting put area and get area to NULL tells streambuf
    // that stream is not opened.
    // @c open() is supposed to be called then.

    setp( NULL, NULL );
    setg( NULL, NULL, NULL );
    }


StreamPool::StreamBuffer::~StreamBuffer()
    {
    if( stream )
        {
        // Synchronize and unlock read and write blocks.

        sync();

        // Update open stream count.

        if( flags_on( open_mode, std::ios_base::in ) )
            {
            stream->num_readers--;
            }

        if( flags_on( open_mode, std::ios_base::out ) )
            {
            stream->num_writers--;
            stream->truncate( stream->current_length );
            }

        stream->open_close_notify();
        stream->pool->num_open_streams--;
        }
    }


bool StreamPool::StreamBuffer::open
    (
    StreamDescriptor *          s,
    std::ios_base::openmode     mode
    )
    {
    // Test.

    hope( !stream );

    if( s->first == INVALID_BLOCK )
        return false;

    // Initialize.

    stream = s;
    open_mode = mode;

    put_offset = 0;
    get_offset = 0;

    if( flags_on( open_mode, std::ios_base::out ) )
        {
        // Open for writing.

        hope( flags_on( stream->pool->get_open_mode(), std::ios_base::out ) );

        if( flags_on( open_mode, std::ios_base::trunc ) )
            {
            stream->truncate( 0 );
            }
        else
        if( flags_on( open_mode, std::ios_base::app ) )
            {
            put_offset = stream->current_length;
            }

        stream->num_writers++;

        // Initialize put area.

        init_put_area( INVALID_BLOCK );
        }

    if( flags_on( open_mode, std::ios_base::in ) )
        {
        // Open for reading.

        stream->num_readers++;

        // Initialize get area.

        init_get_area( INVALID_BLOCK );
        }

    // Update open stream count.

    stream->open_close_notify();
    stream->pool->num_open_streams++;

    return true;
    }


int StreamPool::StreamBuffer::sync()
    {
    if( write_block_locked )
        {
        overflow( traits_type::eof() );
        }

    if( read_block_locked )
        {
        underflow();
        }

    if( flags_on( open_mode, std::ios_base::out ) )
        {
        init_put_area( INVALID_BLOCK );
        }
        
    if( flags_on( open_mode, std::ios_base::in ) )
        {
        init_get_area( INVALID_BLOCK );
        }

    return 0;
    }


std::streambuf * StreamPool::StreamBuffer::setbuf
    (
    char *  reserve_area,
    int     length
    )
    {
    return NULL;
    }


StreamPool::StreamBuffer::pos_type StreamPool::StreamBuffer::seekoff
    (
    off_type                offset,
    std::ios_base::seekdir  direction,
    std::ios_base::openmode mode
    )
    {
    BlockIndex  block;
    off_type    block_offset;
    int         length;
    pos_type    return_value;

    // Init, check open mode.
    
    return_value = pos_type( traits_type::eof() );

    if( !stream )
        {
        return return_value;
        }            
    
    if( ( flags_on( mode, std::ios_base::out ) && !flags_on( open_mode, std::ios_base::out ) ) ||
        ( flags_on( mode, std::ios_base::in ) && !flags_on( open_mode, std::ios_base::in ) ) )
        {
        return return_value;
        }

    // Flush put area if needed to "synchronize" with
    // the get area.

    if( flags_on( open_mode, std::ios_base::out ) &&
        flags_on( mode, std::ios_base::in ) )
        {
        // We are trying to seek a read pointer in a stream buffer
        // that is also opened for writing.
        // Flush put area and adjust current stream length so that
        // client could read from the appended portion of the stream.
        // Note: This is called automatically whenever get area underflows.

        if( put_offset + pptr() - pbase() > stream->current_length )
            {
            seekoff( 0, std::ios_base::cur, std::ios_base::out );
            }
        }
    
    // Seek.

    if( flags_on( mode, std::ios_base::out ) )
        {
        // Adjust stream length.

        put_offset += pptr() - pbase();

        if( put_offset > stream->current_length )
            {
            stream->current_length = put_offset;
            }
        
        // Seek put pointer.

        seekoff( offset, direction, stream->allocated_length, put_offset, block, block_offset, length );

        if( length <= 0 )
            {
            // Seeking past end of allocated stream.
            // Increase allocated stream length.

            stream->grow( put_offset + stream->block_size );
            seekoff( 0, std::ios_base::cur, stream->allocated_length, put_offset, block, block_offset, length );

            hope( length > 0 );
            hope( put_offset + length <= stream->allocated_length );
            }

        init_put_area( block, block_offset, length );
        
        // Succeeded.

        return_value = pos_type( put_offset );
        }

    if( flags_on( mode, std::ios_base::in ) )
        {
        // Seek get pointer.
        // Note: Both the put and get pointers must be moved if requested.

        get_offset += gptr() - eback();
        hope( get_offset <= stream->current_length );

        seekoff( offset, direction, stream->current_length, get_offset, block, block_offset, length );

        if( length <= 0 )
            {
            // Seeking past end of the stream.

            hope( get_offset >= stream->current_length );
            get_offset = stream->current_length;

            init_get_area( INVALID_BLOCK );

            // Can not return traits_type::eof(), that would
            // mean that error occured.

            return pos_type( get_offset );
            }

        init_get_area( block, block_offset, length );

        // Succeeded.

        return_value = pos_type( get_offset );
        }

    // Done.

    return return_value;
    }
     

StreamPool::StreamBuffer::pos_type StreamPool::StreamBuffer::seekpos
    (
    pos_type                position,
    std::ios_base::openmode mode
    )
    {
    return seekoff( position, std::ios_base::beg, mode );
    }
    
    
StreamPool::StreamBuffer::int_type StreamPool::StreamBuffer::overflow
    (
    int_type    chr
    )
    {
    pos_type pos;

    // Check write permission.

    if( !flags_on( open_mode, std::ios_base::out ) )
        {
        return traits_type::eof();
        }

    // Consume put area.

    pos = seekoff( 0, std::ios_base::cur, std::ios_base::out );

    if( pos == pos_type( traits_type::eof() ) )
        {
        // Error.

        return pos;
        }

    if( chr != traits_type::eof() )
        {
        // @c chr is the character that overflowed the old area.
        // It must be written to the new area.

        char * ptr = pptr();
        hope( ptr && ptr != epptr() );

        *ptr = traits_type::to_char_type( chr );
        seekoff( 1, std::ios_base::cur, std::ios_base::out );
        }

    return traits_type::not_eof( chr );
    }


StreamPool::StreamBuffer::int_type StreamPool::StreamBuffer::underflow()
    {
    pos_type pos;

    // Check read permission.

    if( !flags_on( open_mode, std::ios_base::in ) )
        {
        return traits_type::eof();
        }

    // Load get area.

    pos = seekoff( 0, std::ios_base::cur, std::ios_base::in );

    if( pos == pos_type( traits_type::eof() ) )
        {
        // Error.

        return pos;
        }

    // Return first element.

    char * ptr = gptr();

    if( ptr == egptr() )
        {
        // End of stream.

        return traits_type::eof();
        }

    hope( ptr != NULL );

    return traits_type::to_int_type( *ptr );
    }


void StreamPool::StreamBuffer::init_put_area
    (
    BlockIndex  block,
    off_type    offset_in_block,
    int_type    length
    )
    {
    char * data;

    if( write_block_locked )
        {
        // Unlock current write block.

        write_block.unlock( true );
        write_block_locked = false;
        }

    if( block != INVALID_BLOCK )
        {
        // Lock new write block.

        hope( offset_in_block >= 0 && offset_in_block < stream->block_size );
        hope( length >= 0 && offset_in_block + length <= stream->block_size );

        write_block = stream->pool->access_block( block );
        write_block.lock();
        write_block_locked = true;

        // Claim whole block.
        // That is for simplicity and effecient block allocation.

        stream->pool->notify_block_usage( block, stream->block_size );

        // Initialize put area.

        data = reinterpret_cast<char *>( write_block.get_data() );
        data += offset_in_block;
        setp( data, data + length );
        }
    else
        {
        static char_type void_area;

        // Can not set put area to NULL.
        // That would mean that stream is not opened for writing!

        setp( &void_area, &void_area );
        }
    }


void StreamPool::StreamBuffer::init_get_area
    (
    BlockIndex  block,
    off_type    offset_in_block,
    int_type    length
    )
    {
    char * data;

    if( read_block_locked )
        {
        // Unlock current read block.

        read_block.unlock( false );
        read_block_locked = false;
        }

    if( block != INVALID_BLOCK )
        {
        // Lock new write block.

        hope( offset_in_block >= 0 && offset_in_block < stream->block_size );
        hope( length >= 0 && offset_in_block + length <= stream->block_size );

        read_block = stream->pool->access_block( block );
        read_block.lock();
        read_block_locked = true;

        // Initialize get area.

        data = reinterpret_cast<char *>( read_block.get_data() );
        data += offset_in_block;
        setg( data, data, data + length );
        }
    else
        {
        static char_type void_area;

        // Can not set get area to NULL.
        // That would mean that stream is not opened for reading!

        setg( &void_area, &void_area, &void_area );
        }
    }


void StreamPool::StreamBuffer::seekoff
    (
    off_type                offset,
    std::ios_base::seekdir  direction,
    int_type                stream_limit,
    /* In/out. */
    off_type &              position,
    /* Out. */
    BlockIndex &            block,
    off_type &              offset_in_block,
    int &                   length
    )
    {
    int index_in_chain;

    // Update global position.

    switch( direction )
        {
        case std::ios_base::beg:
            {
            position = offset;
            break;
            }

        case std::ios_base::end:
            {
            position = stream->current_length + offset;
            break;
            }

        case std::ios_base::cur:
            {
            position += offset;
            break;
            }

        default:
            panic( "Unexpected direction in seekoff." );
        }

    if( position < 0 )
        {
        position = 0;
        }

    // Select block mapping.

    index_in_chain = position / stream->block_size;
    offset_in_block = position % stream->block_size;
    length = stream->block_size - offset_in_block;

    if( length > 0 )
        {
        // Seeking into stream.

        if( index_in_chain * stream->block_size + offset_in_block + length > stream_limit )
            {
            // Map only that part of the block that contains stream data
            // until @c stream_limit.

            length = stream_limit - index_in_chain * stream->block_size - offset_in_block;
            }

        // Get absolute block index.

        if( length > 0 )
            {
            block = stream->get_block( index_in_chain );
            }
        else
            {
            block = INVALID_BLOCK;
            }
        }
    else
        {
        // Seeking past end of the stream.

        block = INVALID_BLOCK;
        position = stream->allocated_length;
        }

    // @c length might be negative.
    // This occurs when seeking past @c stream_limit.
    }


// Implementation of StreamPool::XXXStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define STREAM_IMPLEMENTATION( StreamName ) \
    \
    StreamPool::StreamName::StreamName() \
        : Super( NULL ), opened( false ) \
        { \
        } \
    \
    StreamPool::StreamName::StreamName \
        ( \
        StreamDescriptor &      stream, \
        std::ios_base::openmode mode \
        ) : Super( NULL ), opened( false ) \
        { \
        open( stream, mode ); \
        } \
    \
    bool StreamPool::StreamName::open \
        ( \
        StreamDescriptor &      stream, \
        std::ios_base::openmode mode \
        ) \
        { \
        if( opened ) \
            { \
            return false; \
            } \
        \
        if( buffer.open( &stream, mode ) ) \
            { \
            /* Bind std::ios_base to our stream. */ \
            \
            Super::init( &buffer ); \
            \
            opened = true; \
            return true; \
            } \
        \
        return false; \
        }

STREAM_IMPLEMENTATION( IStream );
STREAM_IMPLEMENTATION( OStream );
STREAM_IMPLEMENTATION( IOStream );

#undef STREAM_IMPLEMENTATION


// Implementation of StreamPool
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

StreamPool::StreamPool
    (
    int     block_size,
    int     cache_size
    ) : BlockPool( block_size, cache_size ),
        num_open_streams( 0 ),
        open_mode( std::ios_base::in | std::ios_base::out )
    {
    }


StreamPool::~StreamPool()
    {
    hope( !num_open_streams );
    }


std::ios_base::openmode StreamPool::get_open_mode() const
    {
    return open_mode;
    }


void StreamPool::set_open_mode
    (
    std::ios_base::openmode mode
    )
    {
    open_mode = mode;
    }


void StreamPool::create_stream
    (
    StreamDescriptor & s
    )
    {
    hope( can_write() );
    hope( !s.is_valid() );
    modified();

    s.pool = this;

    s.first = create_chain( 1 );
    s.allocated_length = 1 * block_size;

    s.current_length = 0;
    s.block_size = block_size;

    s.num_writers = 0;
    s.num_readers = 0;
    }


void StreamPool::delete_stream
    (
    StreamDescriptor & s
    )
    {
    hope( can_write() );
    hope( !s.num_writers && !s.num_readers );
    modified();

    s.truncate( 0 );
    erase_chain( s.first );

    s.first = INVALID_BLOCK;
    s.pool = NULL;
    }


StreamPool::StreamDescriptor & StreamPool::get_root_stream()
    {
    if( !root_stream.is_valid() )
        {
        create_stream( root_stream );
        }

    return root_stream;
    }


void StreamPool::flush_pool()
    {
    if( can_write() && !is_synced() )
        {
        write_root_stream();

        BlockPool::flush_pool();
        }
    }


bool StreamPool::can_write() const
    {
    if( !BlockPool::can_write() )
        {
        return false;
        }
    else
        {
        return flags_on( get_open_mode(), std::ios_base::out );
        }
    }


void StreamPool::restore_state()
    {
    BlockPool::restore_state();

    read_root_stream();
    }


void StreamPool::save_master_block_extension
    (
    BlockRef & block
    )
    {
    // Save root stream descriptor.

    if( root_stream.is_valid() )
        {
        std::ostringstream packed;
        root_stream.write( packed );

        block << 1;
        block << packed.str();
        }
    else
        {
        block << 0;
        }
    }


void StreamPool::load_master_block_extension
    (
    BlockRef & block
    )
    {
    int root_stream_valid;

    // Restore root stream info.

    block >> root_stream_valid;

    if( root_stream_valid )
        {
        // BlockRef::operator>>() reads whole string
        // stored at StreamPool::save_master_block_extension()
        // (ie. string may contain spaces).

        std::string packed_string;
        block >> packed_string;

        std::istringstream packed( packed_string );
        root_stream.restore( this, packed );
        }
    }


void StreamPool::write_root_stream()
    {
    }


void StreamPool::read_root_stream()
    {
    }


void StreamPool::notify_block_usage
    (
    BlockIndex  index,
    int         allocated_size
    )
    {
    }


} // namespace Massiv::Core
} // namespace Massiv
