/** @file stream_pool.h
 * Stream pool.
 */

/*
 * begin        : Mon 12 Aug 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STREAM_POOL_H
#define MASSIV_STREAM_POOL_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_BLOCK_POOL_H
#include "filesys/block_pool.h"
#endif

#include <iostream>
#include <vector>

namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/**
 * StreamPool defines IO stream interface to block chains represented in
 * @c BlockPool. Each managed chain is described by a @c StreamDescriptor.
 * Unlike @c BlockPool @c StreamPool is not fully serializable. It does not
 * store @c StreamDescriptors for created block chains
 * (except for @c root stream) to the @c BlockPool.
 * On the other side it defines special @c root stream whose state is
 * fully persistent. It is up to client what is stored in the root stream,
 * usually it is used to store "main directory" or stream descriptor list.
 * Stream descriptor can be saved by @c write() and restored back by
 * @c restore().
 * Implementation detail: @c StreamPool does not distinguish among text
 * and binary streams. Both the streams are encoded in the same way.
 */
class StreamPool : public BlockPool
    {
public:

    friend class Test::FileSystemTest;

    class StreamBuffer;
    class StreamDescriptor;

    friend class StreamBuffer;
    friend class StreamDescriptor;

    /**
     * Stream descriptor.
     * Description of managed block chain.
     */
    class StreamDescriptor
        {
    public:

        friend class StreamBuffer;
        friend class StreamPool;

    public:

        //
        // Constructors.
        //

        /**
         * Constructs empty stream decriptor.
         */
        StreamDescriptor();

        /**
         * Copies stream descriptor.
         * Invalidates @c s because there can not be
         * two descriptors describing the same stream (block chain).
         */
        StreamDescriptor
            (
            const StreamDescriptor & s
            );

        /**
         * Destructor.
         */
        virtual ~StreamDescriptor();

        /**
         * Assigns stream descriptor.
         * Invalidates @c s because there can not be two descriptors
         * describing the same stream (block chain).
         */
        StreamDescriptor & operator=
            (
            const StreamDescriptor & s
            );

    public:

        /**
         * Checks if descriptor represents valid stream.
         */
        bool is_valid() const;

        /**
         * Translates chain-relative block index to global BlockIndex.
         */
        BlockIndex get_block
            (
            int index_in_chain
            ) const;

        /**
         * Returns index of last block in the block chain.
         */
        BlockIndex get_last_block() const;

        /**
         * Called by StreamBuffer when opens or closes the stream.
         * According to open stream count @c index_map will be
         * either reconstructed or deleted.
         */
        void open_close_notify();

    protected:

        //
        // State.
        //

        /**
         * Writes descriptor state to @c str.
         * It is safe to write to root stream.
         */
        virtual void write
            (
            std::ostream & str
            ) const;

        /**
         * Restores descriptor state from @c str.
         * Desriptor will be owned by @c owner_pool.
         * It is safe to read from root stream.
         */
        virtual void restore
            (
            StreamPool *    owner_pool,
            std::istream &  str
            );

        /**
         * Client-defined callback.
         * Called when descriptor is changed.
         */
        virtual void notify_change();

    private:

        //
        // Private.
        //

        /**
         * Initializes empty stream descriptor.
         */
        void init();

        /**
         * Allocates new blocks to the stream and increases
         * @c allocated_length. Does not modify @c current_length.
         * Stream must be opened.
         */
        void grow
            (
            int length
            );

        /**
         * Truncates stream length (@c current_length) to @c length.
         * Stream must not be opened for writing.
         */
        void truncate
            (
            int length
            );

        /**
         * Aligns @c value up to multiply of block_size.
         */
        int block_size_align
            (
            int value
            ) const;

        /**
         * Returns number of blocks required to represent a stream
         * of given @c length.
         */
        int get_num_blocks_required
            (
            int length
            ) const;

        /**
         * Returns number of blocks allocated in the block chain.
         */
        int get_num_stream_blocks() const;

    public:

        /**
         * Returns current length of the stream.
         */
        int get_current_length() const;        

    protected:

        //
        // Stream info.
        //

        typedef std::vector<BlockIndex> IndexMap;

        mutable StreamPool *    pool;
            /**< Pool that owns this stream.
                 Descriptor is empty iff @c pool == NULL. */
        BlockIndex              first;
            /**< First block of block chain that represents the stream. */
        int                     allocated_length;
            /**< Allocated stream length (may be larger than @c current_length).
                 Note that when appending at the end of the stream,
                 @c allocated_length grows by multiples of @c block_size.
                 When closing the stream the length is adjusted to
                 the correct size.
                 allocated_length == index_map->size() * block_size. */
        IndexMap *              index_map;
            /**< Map [relative index of block in stream block chain] -> BlockIndex.
                 index_map->size() == length of chain = allocated_length / block_size.
                 @c index_map is always reconstructible from BlockPool::BlockMap
                 and is allocated if the stream is opened only. */
        int                     current_length;
            /**< Current stream length. Updated during seeks. */
        int                     block_size;
            /**< Block size. */

    protected:

        //
        // Stream usage.
        //

        int                     num_writers;
            /**< How many times the stream has been opened for writing. */
        int                     num_readers;
            /**< How many times the stream has been opened for reading. */

        }; // StreamDescriptor

    /**
     * std::streambuf object implemented over a block chain.
     * Maintains read & write position in the stream and
     * sets up get and put area to point to (reference) appropriate
     * blocks of block chain. These blocks are locked in memory
     * until read and/or write position moves to another block or
     * @c sync() is called.
     * Every stream may be opened multiple-times, both for
     * reading and/or writing. Moreover changes made to the stream
     * (through a @c XXXStream) are visible to other
     * @c XXXStream objects even if they _DO_NOT_SHARE_ the same
     * streambuf object!
     */
    class StreamBuffer : public std::streambuf
        {
    public:

        //
        // Constructors & destructors.
        //

        /**
         * Constructs unbound stream buffer.
         * @c open() is expected to be called then.
         */
        StreamBuffer();

        /**
         * Destructor.
         */
        ~StreamBuffer();

        /**
         * Binds StreamBuffer to @c s.
         * Returns true on success.
         */
        bool open
            (
            StreamDescriptor *      s,
            std::ios_base::openmode mode
            );

    public:

        //
        // std::streambuf interface implementation.
        //

        /**
         * Flushes reserve area (get & put areas).
         * Unlocks all locked blocks.
         */
        virtual int sync();

        /**
         * Called when there is a request to use
         * another reserve area, which is not managed by our
         * StreamBuffer.
         * The request is rejected.
         */
        virtual std::streambuf * setbuf
            (
            char *  reserve_area,
            int     length
            );

        /**
         * Seeks to a specified offset.
         */
        virtual pos_type seekoff
            (
            off_type                offset,
            std::ios_base::seekdir  direction,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

        /**
         * Seeks to a specified position.
         */            
        virtual pos_type seekpos
            (
            pos_type                position,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
            );

        /**
         * Write pointer moved out of put area.
         * The only thing we have to do is to reinitialize put area
         * to point to a new (portion of) write block.
         */
        virtual int_type overflow
            (
            int_type    chr
            );

        /**
         * Read pointer moved out of get area.
         * The only thing we have to do is to reinitialize get area
         * to point to a new (portion of) read block.
         */
        virtual int_type underflow();

    private:

        /**
         * Initializes put area to point to portion of @c block.
         * If @c block == INVALID_BLOCK then put area will be empty-initialized.
         */
        void init_put_area
            (
            BlockIndex  block,
            off_type    offset_in_block = 0,
            int_type    length = 0
            );

        /**
         * Initializes get area to point to portion of @c block.
         * If @c block == INVALID_BLOCK then get area will be empty-initialized.
         */
        void init_get_area
            (
            BlockIndex  block,
            off_type    offset_in_block = 0,
            int_type    length = 0
            );

        /**
         * Updates global stream position and selects best mapping
         * for stream data beginning at updated position.
         * @c position is either @c put_offset or @c get_offset.
         * @c stream_limit is either @c stream-\>current_length
         * or @c stream-\>allocated_length.
         * Computes @c block that contains data at updated position,
         * @c offset within block (where that data starts) and length of
         * that data until end of the block or the stream itself.
         */
        void seekoff
            (
            off_type                offset,
            std::ios_base::seekdir  direction,
            int_type                stream_limit,
            /* In/out. */
            off_type &              position,
            /* Out. */
            BlockIndex &            block,
            off_type &              offset_in_block,
            int &                   length
            );

    private:

        std::ios_base::openmode open_mode;
            /**< Stream mode. */
        off_type                put_offset;
            /**< Offset of the beginning of put area within the stream. */
        off_type                get_offset;
            /**< Offset of the beginning of get area within the stream. */
        StreamDescriptor *      stream;
            /**< Pointer to stream descriptor. */

    private:

        BlockRef    read_block;
            /**< Reference to block we are currently reading from.
                 std::streambuf (get area) is bound to a portion
                 of read_block. */
        BlockRef    write_block;
            /**< Reference to block we are currently writing to.
                 std::streambuf (put area) is bound to a portion
                 of write_block. */
        bool        read_block_locked : 1;
            /**< Was read_block locked at @c init_get_area(). */
        bool        write_block_locked : 1;
            /**< Was write_block locked at @c init_put_area()? */

        }; // StreamBuffer

    /**
     * IStream.
     */
    class IStream : public std::istream
        {
    private:

        typedef std::istream    Super;

    public:

        /**
         * Creates closed stream object.
         */
        IStream();

        /**
         * Opens stream for reading.
         */
        IStream
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

    private:

        StreamBuffer    buffer;
            /**< Stream buffer. */
        bool            opened : 1;
            /**< Is stream opened? */

        }; // IStream

    /**
     * OStream.
     */
    class OStream : public std::ostream
        {
    private:

        typedef std::ostream    Super;

    public:

        /**
         * Creates closed stream object.
         */
        OStream();

        /**
         * Opens stream for writing.
         */
        OStream
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

    private:

        StreamBuffer    buffer;
            /**< Stream buffer. */
        bool            opened : 1;
            /**< Is stream opened? */

        }; // OStream

    /**
     * IOStream.
     */
    class IOStream : public std::iostream
        {
    private:

        typedef std::iostream   Super;

    public:

        /**
         * Creates closed stream object.
         */
        IOStream();

        /**
         * Opens stream for both reading and writing.
         */
        IOStream
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            StreamDescriptor &      stream,
            std::ios_base::openmode mode
            );

    private:

        StreamBuffer    buffer;
            /**< Stream buffer. */
        bool            opened : 1;
            /**< Is stream opened? */

        }; // IOStream

protected:

    /**
     * Constructor.
     */
    StreamPool
        (
        int     block_size = DEFAULT_BLOCK_SIZE,
        int     cache_size = DEFAULT_BLOCK_SIZE * 16
        );

    /**
     * Destructor.
     */
    ~StreamPool();

protected:

    /**
     * Creates a new empty stream.
     * Descriptor @c s must be empty.
     */
    void create_stream
        (
        StreamDescriptor & s
        );

    /**
     * Deletes stream.
     * Stream must be closed.
     * After return from the function descriptor will be describing
     * empty stream, which may be reused by @c create_stream().
     */
    void delete_stream
        (
        StreamDescriptor & s
        );

    /**
     * Returns root stream descriptor.
     * Note that root stream is on optional "extension"
     * and is not allocated from BlockPool until get_root_stream().
     */
    StreamDescriptor & get_root_stream();

protected:

    /**
     * Flushes StreamPool buffers.
     * IO stream library buffers (blocks locked by StreamPool::StreamBuffer)
     * remain untouched, thus open streams are not flushed.
     * To flush a @c stream call @c stream.flush().
     */
    virtual void flush_pool();

    /**
     */
    virtual bool can_write() const;

    /**
     */
    virtual void restore_state();

    /**
     */
    virtual void save_master_block_extension
        (
        BlockRef & block
        );

    /**
     */
    virtual void load_master_block_extension
        (
        BlockRef & block
        );

protected:

    /**
     * Client-defined StreamPool callback.
     * Called by StreamPool when it is being flushed.
     * Client should save its state here.
     */
    virtual void write_root_stream();

    /**
     * Client-defined StreamPool callback.
     * Called by StreamPool when it restores its state.
     */
    virtual void read_root_stream();

    /**
     * Client-defined StreamPool callback.
     * Called by StreamPool to inform client about
     * allocated block size.
     */
    virtual void notify_block_usage
        (
        BlockIndex  index,
        int         allocated_size
        );

    /**
     * Returns StreamPool open mode.
     * Default is std::ios_base::in | std::ios_base::out.
     */
    std::ios_base::openmode get_open_mode() const;

    /**
     * Sets StreamPool open mode.
     * @c std::ios_base::in for read-only mode.
     */
    void set_open_mode
        (
        std::ios_base::openmode mode
        );

private:

    int                         num_open_streams;
        /**< Number of open streams. */
    std::ios_base::openmode     open_mode;
        /**< Stream pool open mode.
             Set up by a derived class. */

private:

    //
    // StreamPool master block extension.
    //

    StreamDescriptor    root_stream;
        /**< Root stream descriptor. */

    }; // StreamPool

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_STREAM_POOL_H
