/** @file volume.cpp
 * Volume.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "filesys/volume.h"
#include "global/global.h"
#include "utility/quote_string.h"
#include "status/exception.h"
#include <sstream>

namespace Massiv {
namespace Core {


// Implementation of Volume
// ~~~~~~~~~~~~~~~~~~~~~~~~

Volume::Volume
    (
    int flags,
    int block_size,
    int cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        file( NULL ),
        image_stream( NULL ),
        async_io( NULL )
    {
    }


Volume::Volume
    (
    std::iostream &         image,
    std::ios_base::openmode mode,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        file( NULL ),
        image_stream( NULL ),
        async_io( NULL )
    {
    try
        {
        if( !open( image, mode ) )
            {
            throw ExCanNotOpenVolumeImage( "Can not read from image stream." );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


Volume::Volume
    (
    const std::string &     file_name,
    std::ios_base::openmode mode,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        file( NULL ),
        image_stream( NULL ),
        async_io( NULL )
    {
    try
        {
        if( !open( file_name, mode ) )
            {
            std::ostringstream ss;

            ss << "Can not open " << quote_const_string( file_name ) << ".";
            throw ExCanNotOpenVolumeImage( ss.str() );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


Volume::Volume
    (
    std::iostream &         image,
    const std::string &     volume_name,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        file( NULL ),
        image_stream( NULL ),
        async_io( NULL )
    {
    try
        {
        if( !format( image, volume_name ) )
            {
            throw ExCanNotOpenVolumeImage( "Can not format image stream." );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


Volume::Volume
    (
    const std::string &     file_name,
    const std::string &     volume_name,
    int                     flags,
    int                     block_size,
    int                     cache_size
    ) : BasicVolume( flags, block_size, cache_size ),
        file( NULL ),
        image_stream( NULL ),
        async_io( NULL )
    {
    try
        {
        if( !format( file_name, volume_name ) )
            {
            std::ostringstream ss;

            ss << "Can not format " << quote_const_string( file_name ) << ".";
            throw ExCanNotOpenVolumeImage( ss.str() );
            }
        }
    catch( ... )
        {
        close_pool();
        throw;
        }
    }


Volume::~Volume()
    {
    close_pool();
    }


bool Volume::open
    (
    std::iostream &         image,
    std::ios_base::openmode mode
    )
    {
    if( !image.good() )
        {
        return false;
        }

    if( image_stream )
        {
        return false;
        }

    // Init image stream.

    image_stream = &image;
    image_stream->exceptions( std::ios_base::failbit | std::ios_base::badbit );

    // Restore state.

    try
        {
        set_open_mode( mode );
        BasicVolume::restore_state();
        }
    catch( Exception & )
        {
        image_stream = NULL;
        throw;
        }

    return true;
    }


bool Volume::open
    (
    const std::string &     file_name,
    std::ios_base::openmode mode
    )
    {
    if( image_stream )
        {
        return false;
        }

    // Open file.

    file = new std::fstream( file_name.c_str(), mode | std::ios_base::binary );
    volume_name = file_name;
    stream_name = file_name;

    try
        {
        if( !open( *file, mode ) )
            {
            delete file;
            file = NULL;
            volume_name = std::string();
            stream_name = std::string();

            return false;
            }
        }
    catch( ... )
        {
        delete file;
        file = NULL;
        volume_name = std::string();
        stream_name = std::string();

        throw;
        }

    // Start async IO.

    if( mode & std::ios_base::out )
        {
        start_async_io();
        }

    return true;
    }


bool Volume::format
    (
    std::iostream &     image,
    const std::string & volume_name
    )
    {
    if( !image.good() )
        {
        return false;
        }

    if( image_stream )
        {
        return false;
        }

    // Init image stream.

    image_stream = &image;
    image_stream->exceptions( std::ios_base::failbit | std::ios_base::badbit );

    // Format.

    try
        {
        set_open_mode( std::ios_base::in | std::ios_base::out );
        BasicVolume::format( volume_name );
        }
    catch( Exception & )
        {
        image_stream = NULL;
        throw;
        }

    return true;
    }


bool Volume::format
    (
    const std::string & file_name,
    const std::string & volume_name
    )
    {
    std::ios_base::openmode mode;

    if( image_stream )
        {
        return false;
        }

    // Create file.

    {
    std::ofstream( file_name.c_str(), std::ios_base::out );
    }

    // Format file.

    mode = std::ios_base::in | std::ios_base::out;
    file = new std::fstream( file_name.c_str(), mode | std::ios_base::binary );
    stream_name = file_name;

    try
        {
        if( !format( *file, volume_name ) )
            {
            delete file;
            file = NULL;
            stream_name = std::string();

            return false;
            }
        }
    catch( ... )
        {
        delete file;
        file = NULL;
        stream_name = std::string();

        throw;
        }

    // Start async IO.

    start_async_io();

    return true;
    }


bool Volume::start_async_io()
    {
    if( async_io )
        {
        return true;
        }

    // Starts asynchronous mode.
    // Each access to @c image_stream must be done via @c async_io object.

    async_io = new AsyncIO( MAX_WRITE_BACK_REQUESTS );

    return true;
    }


bool Volume::stop_async_io()
    {
    if( !async_io )
        {
        return true;
        }

    delete async_io;
    async_io = NULL;

    return true;
    }


bool Volume::is_io_pending()
    {
    if( !async_io )
        {
        return false;
        }

    return async_io->are_requests_pending();
    }


void Volume::load_block
    (
    BlockIndex  index,
    UInt8 *     ptr
    )
    {
    hope( get_open_mode() & std::ios_base::in );

    // Log operation.

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream ss;

    ss << volume_name << ": Loading block " << index << ".";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_FILE_SYSTEM,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );

    #endif

    // Load block.

    UInt32 offset = index * block_size;
    UInt32 length = block_size;

    if( !async_io )
        {
        // Synchronous read.
        read_from_stream( *image_stream, ptr, offset, length );
        }
    else
        {
        // Synchronous read.
        async_io->read( *image_stream, ptr, offset, length );
        }
    }


void Volume::save_block
    (
    BlockIndex  index,
    UInt8 *     ptr
    )
    {
    hope( get_open_mode() & std::ios_base::out );

    // Log operation.

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream ss;

    ss << volume_name << ": Saving block " << index << ".";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_FILE_SYSTEM,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );

    #endif

    // Save block.
    
    UInt32 offset = index * block_size;
    UInt32 length = block_size;

    if( index != MASTER_BLOCK_INDEX )
        {
        if( !async_io )
            {
            // Synchronous write.
            write_to_stream( *image_stream, ptr, offset, length );
            }
        else
            {
            // Asynchronous write (write-back).
            async_io->async_copy_and_write( *image_stream, ptr, offset, length );
            }
        }
    else
        {
        // Write master block always synchronously. When done it is ensured
        // that no IO is pending.

        if( !async_io )
            {
            write_to_stream( *image_stream, ptr, offset, length );
            }
        else
            {
            async_io->write( *image_stream, ptr, offset, length );
            }
        
        // Flush the image stream.

        hope( !is_io_pending() );
        image_stream->flush();
        }
    }


std::string Volume::get_pool_name()
    {
    return BasicVolume::get_pool_name() + " (" + stream_name + ")";
    }


void Volume::flush_pool()
    {
    BasicVolume::flush_pool();

    if( image_stream )
        {
        if( !async_io )
            {
            // Synchronous flush.
            image_stream->flush();
            }
        else
            {
            // Asynchronous flush (write-back).
            async_io->async_flush( *image_stream );
            }
        }
    }
    

void Volume::close_pool()
    {
    try
        {
        stop_async_io();
        BasicVolume::close_pool();

        if( file )
            {
            delete file;
            file = NULL;
            }

        image_stream = NULL;
        }
    catch( Exception & )
        {
        Global::log_warning
            (
            Status::FACILITY_FILE_SYSTEM,
            Status::PRIORITY_HIGHEST,
            "Failed to close Volume."
            );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
