/** @file volume.h
 * Volume.
 */

/*
 * begin        : Fri 28 Jun 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VOLUME_H
#define MASSIV_VOLUME_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_BASIC_VOLUME_H
#include "filesys/basic_volume.h"
#endif

#ifndef MASSIV_ASYNC_IO_H
#include "filesys/async_io.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup filesys
 * @{
 */

/** @class Volume
 * @c Volume implements an effecient file system with IO stream interface.
 * The image file (file system image) is stored in a @c std::fstream
 * or any @c std::iostream.
 * Unlike other simple "filesystems" our implementation is fully dynamic.
 * Streams can be dynamically created, deleted or modified without
 * any performance loss or ineffecient storage usage.
 * Its design allows to easily switch among std::fstreams
 * and Volume::FStreams.
 * Asynchronous IO (deferred writing to the image file initiated by
 * file system logic) is supported. Turned on for std::fstream read-write
 * volume image streams by-default.
 * @note The first byte of image file saved by Volume should be 1 (unless
 * the Volume was used as a helper class in some other type of volume, 
 * for example in CompactVolume. Value 1 helps to recognize 'Volume' images.
 */
class Volume : public BasicVolume
    {
public:

    enum Parameters
        {
        MAX_WRITE_BACK_REQUESTS =   16
            /**< Maximum number of write-back requests (async IO). */
        }; // Parameters

public:

    //
    // Open & close.
    //

    /**
     * Constructs uninitialized volume.
     * Either @c format() or @c open() is supposed to be called then.
     */
    Volume
        (
        int flags = DEFAULT_VOLUME_FLAGS,
        int block_size = DEFAULT_BLOCK_SIZE,
        int cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Opens volume.
     * Expects volume image to be stored in @c image.
     * @c mode determines whether the volume should be opened in
     * read-write mode or read-only mode.
     * @c block_size must correspond to the block size of the stored
     * volume image.
     */
    Volume
        (
        std::iostream &         image,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Opens volume.
     * Expects volume image to be stored on disk in @c file_name.
     */
    Volume
        (
        const std::string &     file_name,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Formats volume.
     * Volume image will be stored in @c image.
     */
    Volume
        (
        std::iostream &         image,
        const std::string &     volume_name,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Formats volume.
     * Volume image will be stored on disk in @c file_name.
     */
    Volume
        (
        const std::string &     file_name,
        const std::string &     volume_name,
        int                     flags = DEFAULT_VOLUME_FLAGS,
        int                     block_size = DEFAULT_BLOCK_SIZE,
        int                     cache_size = DEFAULT_CACHE_SIZE
        );

    /**
     * Destructor.
     */
    ~Volume();

    /**
     * Opens volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool open
        (
        std::iostream &         image,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
        );

    /**
     * Opens volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool open
        (
        const std::string &     file_name,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
        );

    /**
     * Formats volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool format
        (
        std::iostream &     image,
        const std::string & volume_name
        );

    /**
     * Formats volume.
     * Returns true on success.
     * Throws exception if image integrity error is detected.
     */
    bool format
        (
        const std::string & file_name,
        const std::string & volume_name
        );

public:

    //
    // Asynchronous IO.
    //

    /**
     * Switches to asynchronous mode.
     * Returns true on success.
     */
    bool start_async_io();

    /**
     * Switches to normal synchronous mode.
     * Returns true on success.
     */
    bool stop_async_io();

    /**
     * Tests if there are pending IO operations.
     * Valid for asynchronous mode.
     */
    bool is_io_pending();

private:

    /**
     */
    virtual void load_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        );

    /**
     */
    virtual void save_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        );
    
    /**
     */
    virtual std::string get_pool_name();

    /**
     */
    virtual void flush_pool();

    /**
     */
    virtual void close_pool();

private:

    std::fstream *  file;
        /**< Std. file stream that represents the image. May be NULL. */
    std::string     stream_name;
        /**< File name used to open the @c file. */
    std::iostream * image_stream;
        /**< Stream that represents the image.
             @c image is either external stream or @c file. */
    AsyncIO *       async_io;
        /**< Request queue for asynchronous IO when accessing
             image file. */

    }; // Volume

/** @} */ // addtogroup filesys


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_BASIC_VOLUME_H
