/** @file volume_manager.cpp
 * Virtual file system with IO stream interface.
 */

/*
 * begin        : Fri 21 Jun 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "filesys/volume_manager.h"
#include "filesys/basic_volume.h"
#include "filesys/volume.h"
#include "filesys/compact_volume.h"
#include "global/global.h"
#include "utility/hash.h"
#include "utility/quote_string.h"
#include "status/exception.h"
#include <sstream>
#include <algorithm>


namespace Massiv {
namespace Core {


// Definition of VolumeManager::VolumeManagerSettings
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * VolumeManager settings.
 * @note Besides the attributes listed here there is a list of preset
 * volumes that are mounted upon VolumeManager start up. See
 * @c VolumeManager::My::My().
 */
class VolumeManager::VolumeManagerSettings : public Settings
    {
public:

    /** Constructor. */
    VolumeManagerSettings() :
        Settings( "VolumeManager" ),
        flush_period( access_float( "flush_period", 30.0f ) ),
        volume_cache( access_integer( "volume_cache", BasicVolume::DEFAULT_CACHE_SIZE ) )
        {
        }

public:

    Float &     flush_period;
        /**< How often (in seconds) the volume manager flushes file systems. */
    Integer &   volume_cache;
        /**< Cache size per mounted volume in bytes. */
    
    }; // VolumeManagerSettings.


// Definition of VolumeManager::My
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
/**
 * VolumeManager state.
 * Defined as an inner class to hide implementation details.
 */
class VolumeManager::My : public FileSystem
    {
public:
       
    class IteratorImpl;

    friend class VolumeManager::IFStream;
    friend class VolumeManager::OFStream;
    friend class VolumeManager::FStream;
    friend class IteratorImpl;

    typedef ext::hash_map<std::string, BasicVolume *> VolumeMap;
    typedef std::list<FileSystem *> FileSystemList;

    /** Result of name resolution. */
    struct ResolvedName
        {
    public:

        /** Constructor. */
        ResolvedName() :
            file_system( NULL ),
            relative_name(),
            is_native( false ),
            is_directory( false ),
            is_file( false ),
            is_explicit( false ),
            volume_name( std::string() )
            {
            }
            
    public:
                
        FileSystem *    file_system;
            /**< File system that holds @c relative_name. */
        std::string     relative_name;
            /**< @c file_system relative name. */
        bool            is_native : 1;
            /**< Name refers to a native file system. */
        bool            is_directory : 1;
            /**< Name refers to a directory. */
        bool            is_file : 1;
            /**< Name refers to a file. */
        bool            is_explicit : 1;
            /**< Name refers to @c file_system explicitly. */
        std::string     volume_name;
            /**< May contain the name of volume, if applicable. */
        }; // ResolvedName

    /**
     * Iterator that enumerates files from the virtual file system
     * implemented by VolumeManager.
     * The iterators are invalidated by volume mount/dismount operations.
     * IteratorImpl servers as a proxy to various iterators implemented
     * by the mounted file systems. There is an "underyling iterator"
     * that performs the enumeration on the current file system.
     * When that iterator reaches the end of the enumerated sequence,
     * it is automatically rebound to the next file system (with the
     * same context).
     * This allows to mount "overlapping" directories (multiple file systems
     * hold different parts of the same virtual directory) and handle
     * the enumeration properly.
     */
    class IteratorImpl : public FileIteratorImpl
        {
        friend class My;

    protected:
        
        /**
         */
        IteratorImpl
            (
            DirectoryEnumerationMode    mode
            ) : FileIteratorImpl( mode ),
                current_fs( My::self().file_system_list.end() )
            {
            valid_flag = false;
            }

        /**
         */
        IteratorImpl
            (
            const std::string &         mask,
            DirectoryEnumerationMode    mode
            ) : FileIteratorImpl( mode ),
                current_fs( My::self().file_system_list.end() )
            {
            valid_flag = false;
            bind_current( mask );
            }

        /**
         */
        IteratorImpl
            (
            const std::string &         path,
            const std::string &         mask,
            DirectoryEnumerationMode    mode
            ) : FileIteratorImpl( mode ),
                current_fs( My::self().file_system_list.end() )
            {
            valid_flag = false;
            bind( path, mask );
            }

        /**
         */
        virtual ~IteratorImpl()
            {
            }

    private: 

        NO_ASSIGNMENT_OPERATOR( IteratorImpl );

    protected:

        /**
         */
        virtual void bind_current
            (
            const std::string & mask
            )
            {
            bind( My::self().get_current_directory(), mask );
            }

        /**
         * Binds the iterator with @c path that refers to VolumeManager
         * file system namespace.
         */
        virtual void bind
            (
            const std::string &     path,
            const std::string &     mask
            )
            {
            // Rebind.

            if( valid_flag )
                {
                it.unbind();
                current_fs = My::self().file_system_list.end();
                }
            
            if( !initial_path.size() )
                {
                initial_path = path;
                }

            My::self().resolve_name( path, resolved_path );
            
            bound_path = resolved_path.relative_name;
            bound_mask = mask;

            if( resolved_path.is_explicit )
                {
                // Bind the iterator with the explicitly qualified
                // file system. In this case the iterator will
                // work as a proxy to 'it'.

                bind_underlying_iterator
                    (
                    resolved_path.file_system,
                    resolved_path.relative_name
                    );
                }
            else
                {
                // Bind the iterator with the first available file system.
                    
                current_fs = My::self().file_system_list.begin();
                
                while( current_fs != My::self().file_system_list.end() )
                    {
                    bind_underlying_iterator( *current_fs, bound_path );
                    
                    if( it.is_valid() )
                        {
                        break;
                        }

                    ++current_fs;
                    }
                }
            
            // Validate iterator.

            if( !it.is_valid() )
                {
                std::ostringstream ss;

                ss << "Directory " << quote_const_string( path ) << " not found.";

                throw ExDirectoryNotFound( ss.str() );
                }

            valid_flag = true;
            }

        /**
         */
        virtual void unbind()
            {
            ensure_validity();
            valid_flag = false;

            bound_path.clear();
            bound_mask.clear();
            resolved_path = ResolvedName();

            it.unbind();
            current_fs = My::self().file_system_list.end();
            }

        /**
         */
        virtual Reference<FileIteratorImpl> next()
            {
            ensure_validity();
            next_aux();
            return this;
            }

        /**
         */
        virtual Reference<FileIteratorImpl> rewind()
            {
            bind( initial_path, bound_mask );
            return this;
            }

        /**
         */
        virtual Reference<FileIteratorImpl> enter()
            {
            ensure_validity();

            if( !is_directory() )
                {
                throw ExNotADirectory
                    ( 
                    "Iterator didn't point to a directory when enter() was called." 
                    );
                }

            return new IteratorImpl
                ( 
                get_absolute_directory() + it.name(),
                bound_mask,
                dir_enum_mode
                );
            }

        /**
         */
        virtual Reference<FileIteratorImpl> descent()
            {
            ensure_validity();

            if( !is_directory() )
                {
                throw ExNotADirectory
                    ( 
                    "Iterator didn't point to a directory when descent() was called."
                    );
                }

            bind( get_absolute_directory() + it.name(), bound_mask );

            return this;
            }

        /**
         */
        virtual Reference<FileIteratorImpl> leave()
            {
            ensure_validity();

            return new IteratorImpl
                ( 
                get_absolute_directory() + "..",
                bound_mask,
                dir_enum_mode
                );
            }

        /**
         */
        virtual Reference<FileIteratorImpl> ascent()
            {
            ensure_validity();

            bind( get_absolute_directory() + "..", bound_mask );
            return this;
            }

        /**
         */
        virtual bool is_valid() const
            {
            return valid_flag;
            }

        /**
         */
        virtual bool is_file() const
            {
            ensure_validity();

            if( it.is_valid() )
                {
                return it.is_file();
                }
            else
                {
                return false;
                }
            }

        /**
         */
        virtual bool is_directory() const
            {
            ensure_validity();

            if( it.is_valid() )
                {
                return it.is_directory();
                }
            else
                {
                return false;
                }
            }

        /**
         */
        virtual bool is_trivial_directory() const
            {
            ensure_validity();

            if( it.is_valid() )
                {
                return it.is_trivial_directory();
                }
            else
                {
                return false;
                }
            }

        /**
         */
        virtual std::string name() const
            {
            ensure_validity();
            
            return it.name();
            }

        /**
         */
        virtual UInt64 size() const
            {
            ensure_validity();
            
            return it.size();
            }

        /**
         */
        virtual std::string get_absolute_directory() const
            {
            ensure_validity();
            
            return bound_path + ( *current_fs )->get_slash();
            }

        /**
         */
        virtual bool is_in_root() const
            {
            ensure_validity();

            if( it.is_valid() )
                {
                return it.is_in_root();
                }
            else
                {
                return false;
                }
            }

    private:

        /**
         */
        virtual FileIteratorImpl * next_aux()
            {
            // Advance iterator to the underlying file system
            // until it is invalidated.

            it.next();
                
            if( !it.is_valid() && !resolved_path.is_explicit )
                {
                // The current file system has been fully examined.
                // Proceed to the next file system.

                FileSystemList::const_iterator end = My::self().file_system_list.end();

                while( current_fs != end )
                    {
                    ++current_fs;

                    if( current_fs == end )
                        {
                        break;
                        }

                    bind_underlying_iterator( *current_fs, bound_path );
                    
                    if( it.is_valid() )
                        {
                        break;
                        }
                    }
                }
            
            if( !it.is_valid() )
                {
                // No other file system can process the request.
                // Invalidate the iterator.

                valid_flag = false;
                }

            // TODO: Skip trivial directories if they were already
            //       enumerated from a previous file system.

            return this;
            }

        /**
         * Binds the underlying iterator with @c path on @c file_system.
         */
        void bind_underlying_iterator
            (
            const FileSystem *  file_system,
            const std::string & path
            )
            {
            try
                {
                it = file_system->open_directory( path, bound_mask );
                hope( it.is_valid() );
                }
            catch( ExDirectoryNotFound & )
                {
                hope( !it.is_valid() );
                }
            }

        /**
         * Ensures that the iterator is valid.
         */
        void ensure_validity() const
            {
            if( !is_valid() )
                {
                throw ExInvalidFsIterator( "FS iterator is not valid." );
                }
            }

    private:

        std::string                     initial_path;
            /**< Initial path bound to the iterator.
                 Must be stored in order to be able to implement rewind. */
        std::string                     bound_path;
            /**< Path bound to the iterator. */
        std::string                     bound_mask;
            /**< Mask bound to the iterator. */
        ResolvedName                    resolved_path;
            /**< Resolved @c bound_path. */
        FileIterator                    it;
            /**< Iterator to the file system being inspected.
                 The "underlying iterator". */
        FileSystemList::const_iterator  current_fs;
            /**< File system being inspected. Needed to keep track
                 which file system is "current". @c next_aux()
                 uses this to proceed to the next file system.
                 My::file_system_list.end() if unknown. */
        }; // class IteratorImpl

public:

    /**
     * Constructor.
     */
    My()
        {
        std::string registry_path = settings.get_registry_path();

        // Register native file system.

        file_system_list.push_back( &Global::native_file_system() );

        // Mount preset volumes.

        try
            {
            mount_from_registry
                ( registry_path + "/Volumes", DEFAULT_MOUNT_FLAGS );
            }
        catch( Exception & )
            {
            }
            
        try
            {
            mount_from_registry
                ( registry_path + "/ReadOnlyVolumes", DEFAULT_MOUNT_FLAGS | MOUNT_READ_ONLY );
            }
        catch( Exception & )
            {
            }
        }

    /**
     * Destructor.
     */
    ~My()
        {
        while( image_to_volume.size() )
            {
            VolumeMap::iterator it = image_to_volume.begin();                        
            dismount_volume( it->first );
            }
        }

    /**
     * Mounts volume.
     *
     * @param image_path Specifies the image of volume to be mounted.
     * @param volume_name Specifies the name of the volume.
     * @param mount_flags Specifies how the volume should be mounted.
     *      See VolumeManager::MountFlags for more information.
     * @param format Format a new volume? Set this to @c true if a new
     *      volume should be created or an old rewritten.
     */
    void mount_volume
        (
        const std::string & image_path,
        const std::string & volume_name,
        int                 mount_flags,
        bool                format
        )
        {
        BasicVolume *               volume;
        std::ios_base::openmode     mode;
        std::string                 path;
        int                         volume_flags;
        int                         block_size;
        int                         cache_size;

        // Test mount validity.

        path = image_path;
        normalize_name( path );

        if( image_to_volume.find( path ) != image_to_volume.end() )
            {
            std::ostringstream ss;
            
            ss << "Volume " << quote_const_string( path ) << " already mounted.";

            throw ExVolumeAlreadyMounted( ss.str() );
            }

        // Open volume.

        if( flags_on( mount_flags, MOUNT_READ_ONLY ) )
            {
            hope( !format );
            mode = std::ios_base::in;
            }
        else
            {
            mode = std::ios_base::in | std::ios_base::out;
            }

        volume_flags = BasicVolume::DEFAULT_VOLUME_FLAGS;
        
        block_size = BasicVolume::DEFAULT_BLOCK_SIZE;
        
        cache_size = settings.volume_cache;

        if( cache_size < BasicVolume::DEFAULT_CACHE_SIZE )
            {
            cache_size = BasicVolume::DEFAULT_CACHE_SIZE;
            }

        if( flags_on( mount_flags, MOUNT_VOLUME ) )
            {
            if( !format )
                {
                volume = new Volume
                    ( path, mode, volume_flags, block_size, cache_size );
                }
            else
                {
                volume = new Volume
                    ( path, volume_name, volume_flags, block_size, cache_size );
                }
            }
        else
            {
            if( !format )
                {
                volume = new CompactVolume
                    ( path, mode, volume_flags, block_size, cache_size );
                }
            else
                {
                volume = new CompactVolume
                    ( path, volume_name, volume_flags, block_size, cache_size );
                }
            }

        // Register volume.

        // FIXME: exception safety

        image_to_volume.insert( VolumeMap::value_type( path, volume ) );
        name_to_volume.insert( VolumeMap::value_type( volume->get_volume_name(), volume ) );
        file_system_list.push_back( volume );
        
        // Log.

        Global::log_info
            (
            Status::FACILITY_VOLUME_MANAGER,
            Status::PRIORITY_LOW,
            "\"" + image_path + "\" successfully mounted."
            );
        }
    
    /**
     * Dismounts a volume.
     *
     * @param image_path Specifies the image of the volume to be dismounted.
     */
    void dismount_volume
        (
        const std::string & image_path
        )
        {
        std::string path = image_path;
        normalize_name( path );

        // Find volume.

        VolumeMap::iterator i_it = image_to_volume.find( path );

        if( i_it == image_to_volume.end() )
            {
            std::ostringstream ss;
            
            ss << "Volume " << quote_const_string( path ) << " not mounted.";

            throw ExVolumeNotMounted( ss.str() );
            }

        // Log.

        Global::log_info
            (
            Status::FACILITY_VOLUME_MANAGER,
            Status::PRIORITY_LOW,
            "Dismounting \"" + image_path + "\"."
            );

        // Unregister volume.

        BasicVolume *               volume =
            i_it->second;
        VolumeMap::iterator         n_it = 
            name_to_volume.find( volume->get_volume_name() );
        FileSystemList::iterator    f_it = 
            std::find( file_system_list.begin(), file_system_list.end(), volume );

        image_to_volume.erase( i_it );
        name_to_volume.erase( n_it );
        file_system_list.erase( f_it );
                
        delete volume;
        }

    /**
     * Flushes mounted volumes.
     */
    void flush()
        {
        VolumeMap::iterator it = image_to_volume.begin();

        while( it != image_to_volume.end() )
            {
            const std::string & image_path = it->first;
            BasicVolume *       volume = it->second;
            
            volume->flush();
            
            Global::log_info
                (
                Status::FACILITY_VOLUME_MANAGER,
                Status::PRIORITY_LOW,
                "\"" + image_path + "\" flushed."
                );

            ++it;
            }
        }

    /**
     * Resolves file/directory name and fills @c resolved.
     *
     * 'Resolves' means that it finds out which volume does the given @a name
     * belong to and also finds out some additional information such as
     * whether the @a name refers to a file or a directory, etc.
     * 
     * @see ResolvedName
     */
    void resolve_name
        (
        const std::string & name,
        ResolvedName &      resolved
        ) const
        {
        FileSystem & native_fs = Global::native_file_system();
        std::string  normalized_name = name;

        normalize_name( normalized_name );

        // Does 'name' explicitly qualify volume name?
        
        std::string qualifier = "://";
        std::size_t pos = normalized_name.find( qualifier );

        if( pos != std::string::npos )
            {
            resolved.volume_name = normalized_name.substr( 0, pos );
            
            // Construct qualified name.

            normalized_name = normalized_name.substr( pos + qualifier.size() );
            
            if( resolved.volume_name == "native" )
                {
                // Native file system.

                resolved.file_system = &native_fs;
                
                if( resolved.file_system->is_relative( normalized_name ) )
                    {
                    normalized_name = current_directory + normalized_name;
                    }
                
                resolved.relative_name = normalized_name;
                resolved.is_native = true;
                resolved.is_directory = native_fs.is_directory( normalized_name );
                resolved.is_file = native_fs.is_file( normalized_name );
                resolved.is_explicit = true;
                }
            else
                {
                VolumeMap::const_iterator it = name_to_volume.find( resolved.volume_name );
                BasicVolume * volume;

                // Mounted volume.

                if( it == name_to_volume.end() )
                    {
                    std::ostringstream ss;
                    
                    ss << "No such volume " 
                       << quote_const_string( resolved.volume_name ) << ".";

                    throw ExVolumeNotMounted( ss.str() );
                    }

                volume = it->second;
                    
                resolved.file_system = volume;

                if( resolved.file_system->is_relative( normalized_name ) )
                    {
                    normalized_name = current_directory + normalized_name;
                    }

                resolved.relative_name = normalized_name;
                resolved.is_native = false;
                resolved.is_directory = volume->is_directory( normalized_name );
                resolved.is_file = volume->is_file( normalized_name );
                resolved.is_explicit = true;
                }

            resolved.file_system->normalize_path( resolved.relative_name, false );

            return;
            }

        resolved.volume_name = "";
       
        // Find filesystem that holds normalized_name.
            
        FileSystemList::const_iterator it = file_system_list.begin();
        
        bool            file = false;
        bool            directory = false;

        while( it != file_system_list.end() )
            {
            FileSystem *    file_system = *it;
            std::string     aux_name = normalized_name;
           
            if( file_system->is_relative( aux_name ) )
                {
                aux_name = current_directory + aux_name;
                }

            file = file_system->is_file( aux_name );
            directory = file_system->is_directory( aux_name );

            if( file || directory )
                {
                resolved.file_system = file_system;
                resolved.relative_name = aux_name;
                resolved.is_native = file_system == &native_fs;
                resolved.is_directory = directory;
                resolved.is_file = file;
                resolved.is_explicit = false;

                resolved.file_system->normalize_path( resolved.relative_name, false );

                return;
                }

            ++it;
            }

        // normalized_name not found anywhere.
        // If client later requests to create such stream we want to
        // create it in normalized_name (grand)parent's directory.

        std::string parent_directory = normalized_name;
        
        while( true )
            {
            // Go to parent directory of parent_directory.

            pos = parent_directory.find_last_of( "/\\" );
            
            if( pos == std::string::npos )
                {
                break;
                }

            parent_directory = parent_directory.substr( 0, pos );

            // Find file system that holds parent_directory.

            it = file_system_list.begin();

            while( it != file_system_list.end() )
                {
                FileSystem *    file_system = *it;
                std::string     aux_name = normalized_name;

                if( file_system->is_relative( aux_name ) )
                    {
                    aux_name = current_directory + aux_name;
                    }

                file = file_system->is_file( parent_directory );
                directory = file_system->is_directory( parent_directory );

                if( file || directory )
                    {
                    resolved.file_system = file_system;
                    resolved.relative_name = aux_name;
                    resolved.is_native = file_system == &native_fs;
                    resolved.is_directory = directory;
                    resolved.is_file = file;
                    resolved.is_explicit = false;

                    resolved.file_system->normalize_path( resolved.relative_name, false );

                    return;
                    }

                ++it;
                }
            }

        // Native file system.
        
        resolved.file_system = &native_fs;
        
        if( resolved.file_system->is_relative( normalized_name ) )
            {
            normalized_name = current_directory + normalized_name;
            }
        
        resolved.relative_name = normalized_name;
        resolved.is_native = true;
        resolved.is_directory = false;
        resolved.is_file = false;
        resolved.is_explicit = false;

        resolved.file_system->normalize_path( resolved.relative_name, false );
        }  // VolumeManager::My::resolve_name()

    /**
     * Normalizes @c name.
     *
     * Removes unneccessary ".", "..".
     * @note Does not check if ".", ".." is really applicable
     *      (ie. "my_directory/my_file.txt/subdir/.." ).
     */
    void normalize_name
        (
        std::string & name
        ) const
        {
        std::vector<std::string>    components;
        std::string                 component;
        std::string::iterator       it;

        // Split name into components.

        for( it = name.begin(); it != name.end(); ++it )
            {
            char    ch = *it;
            bool    last_char = it + 1 == name.end();

            if( ch == '/' || ch == '\\' || last_char )
                {
                if( last_char && ch != '/' && ch != '\\' )
                    {
                    component += ch;
                    }

                if( component == "." )
                    {
                    component.clear();
                    }
                else
                if( component == ".." )
                    {
                    if( components.size() )
                        {
                        components.pop_back();
                        }

                    component.clear();
                    }
                else
                    {
                    components.push_back( component );
                    component.clear();
                    }
                }
            else
                {
                component += ch;
                }
            }

        // Construct normalized name.

        std::size_t i;
        name.clear();

        for( i = 0; i < components.size(); ++i )
            {
            if( i > 0 )
                {
                name += "/";
                }
            
            name += components[ i ];
            }
        }  // VolumeManager::My::normalize_name()
    
    /**
     * Mounts volumes listed in the registry under @a registry_path.
     * Each volume is described by a @c ConfigVariableString variable
     * holding path to the image file on the native file system.
     *
     * @param registry_path Specifies a registry path of a node that
     *      contains information about what volumes should be mounted.
     * @param mount_flags Specifies a manner of mounting the volume.
     *      See VolumeManager::MountFlags from closer information.
     */
    void mount_from_registry
        (
        const std::string & registry_path,
        int                 mount_flags
        )
        {
        Registry &          registry = Global::registry();
        Registry::Iterator  it = registry.begin( registry_path );

        while( it != registry.end() )
            {
            ConfigVariable & cvar = *it;
            
            try
                {
                ConfigVariableString & image_path = cvar.as_string();

                try
                    {
                    mount_volume
                        (
                        image_path,
                        std::string(),
                        mount_flags,
                        false
                        );
                    }
                catch( ... )
                    {
                    std::ostringstream ss;

                    ss << quote_const_string( image_path ) << " could not be mounted.";

                    Global::log_warning
                        (
                        Status::FACILITY_VOLUME_MANAGER,
                        Status::PRIORITY_HIGH,
                        ss.str()
                        );
                    }
                }
            catch( ... )
                {
                std::ostringstream ss;

                ss << quote_const_string( cvar.get_name() ) << " is not a string variable.";

                Global::log_warning
                    (
                    Status::FACILITY_VOLUME_MANAGER,
                    Status::PRIORITY_NORMAL,
                    ss.str()
                    );
                }

            ++it;
            }
        }  // VolumeManager::My::mount_from_registry()

    /**
     * Returns pointer to self.
     */
    static My & self()
        {
        hope( VolumeManager::my );
        return *VolumeManager::my;
        }

public:

    //
    // FileSystem interface implementation.
    //

    /**
     */
    virtual FileIterator open_current_directory
        (
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const
        {
        return open_directory( get_current_directory(), mask, mode );
        }

    /**
     */
    virtual FileIterator open_directory
        (
        const std::string &         path,
        const std::string &         mask,
        DirectoryEnumerationMode    mode = FileIterator::MATCH_DIRS_NAMES
        ) const
        {
        return FileIterator( new IteratorImpl( path, mask, mode ) );
        }

    /**
     */
    virtual std::string get_current_directory() const
        {
        return current_directory;
        }

    /**
     */
    virtual void set_current_directory
        (
        const std::string & path
        )
        {
        ResolvedName resolved;
        
        // Resolve path.

        resolve_name( path, resolved );

        if( !resolved.is_directory )
            {
            std::ostringstream ss;

            ss << "Can not change directory to " << quote_const_string( path ) << ".";

            throw ExCanNotChangeDirectory( ss.str() );
            }

        current_directory = resolved.relative_name;

        // Append path separator so that @c current_directory can be simply
        // prepended in front of current-directory-relative name to
        // get qualified name.

        if( current_directory.size() &&
            current_directory[ current_directory.size() - 1 ] != get_slash() )
            {
            current_directory += get_slash();
            }
        }

    /**
     */
    virtual void normalize_path
        (
        std::string &   path,
        bool            slash_terminated = true
        ) const
        {
        ResolvedName    resolved;

        resolve_name( path, resolved );
        resolved.file_system->normalize_path( path, slash_terminated );

        path = resolved.volume_name + "://" + resolved.relative_name;
        }

    /**
     */
    virtual void create_directory
        (
        const std::string & path
        )
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        resolved.file_system->create_directory( resolved.relative_name );
        }

    /**
     */
    virtual void delete_directory
        (
        const std::string & path,
        DirectoryDeleteMode mode = DELETE_EMPTY
        )
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        resolved.file_system->delete_directory( resolved.relative_name, mode );
        }
    
    /**
     */
    virtual void create_file
        (
        const std::string & path
        )
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        resolved.file_system->create_file( resolved.relative_name );
        }

    /**
     */
    virtual void delete_file
        (
        const std::string & path
        )
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        resolved.file_system->delete_file( resolved.relative_name );
        }

    /**
     */
    virtual void rename_file
        (
        const std::string & from,
        const std::string & to
        )
        {
        ResolvedName from_resolved;
        ResolvedName to_resolved;

        resolve_name( from, from_resolved );
        resolve_name( to, to_resolved );
        
        if( from_resolved.file_system == to_resolved.file_system )
            {
            from_resolved.file_system->rename_file
                ( from_resolved.relative_name, to_resolved.relative_name );
            }
        else
            {
            // Move file from one file system to the other.

            VolumeManager::IFStream input( from, std::ios_base::in | std::ios_base::binary );
            VolumeManager::OFStream output( to, std::ios_base::out | std::ios_base::binary );

            if( !input.good() )
                {
                std::ostringstream ss;
                
                ss << "Can not open " << quote_const_string( from ) << ".";
                throw ExIOError( ss.str() );
                }

            if( !output.good() )
                {
                std::ostringstream ss;
                
                ss << "Can not create " << quote_const_string( to ) << ".";
                throw ExIOError( ss.str() );
                }

            output << input.rdbuf();

            if( !output.good() || !input.good() )
                {
                std::ostringstream ss;
                
                ss  << "IO error while copying " << quote_const_string( from )
                    << " to " << quote_const_string( to ) << ".";
                throw ExIOError( ss.str() );
                }

            from_resolved.file_system->delete_file( from_resolved.relative_name );
            }
        }

    /**
     */
    virtual bool is_file
        (
        const std::string & path
        ) const
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        return resolved.is_file;
        }
    
    /**
     */
    virtual bool is_directory
        (
        const std::string & path
        ) const
        {
        ResolvedName resolved;
        
        resolve_name( path, resolved );
        return resolved.is_directory;
        }

public:

    //
    // Implemented from FileSystem
    //

    /**
     */
    virtual bool is_absolute
        (
        const std::string & path
        ) const
        {
        ResolvedName        resolved_path;

        My::self().resolve_name( path, resolved_path );

        return resolved_path.file_system->is_absolute
                                                ( 
                                                resolved_path.relative_name 
                                                );
        }

    /**
     */
    virtual bool is_relative
        (
        const std::string & path
        ) const
        {
        return !is_absolute( path );
        }

    char get_slash() const
        {
        return '/';
        }

    char get_back_slash() const
        {
        return '\\';
        }

public:

    VolumeManagerSettings   settings;
        /**< Volume manager settings. */

    SystemTime              last_time_flushed;
        /**< Last time file systems were flushed. */

private:

    VolumeMap       image_to_volume;
        /**< List of mounted volumes. Indexed by paths to their image
             files. */
    VolumeMap       name_to_volume;
        /**< List of mounted volumes. Indexed by volume names. */
    FileSystemList  file_system_list;
        /**< List of available file systems. Ordered by the time of
             the corresponding mount operations. Native file system is the
             first one. */
    std::string     current_directory;
        /**< Current directory (current VolumeManager path). */
    }; // VolumeManager::My


// Implementation of VolumeManager::IFStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define RESOLVE_AND_INIT() \
    My::ResolvedName resolved; \
    if( stream.get() ) \
        return false; \
    My::self().resolve_name( stream_name, resolved ); \
    close();

#define CREATE_FILE() \
    if( mode & std::ios_base::out ) \
        { \
        if( !resolved.file_system->is_file( resolved.relative_name ) ) \
            resolved.file_system->create_file( resolved.relative_name ); \
        }

#define BIND_STREAM() \
    if( !stream->good() ) \
        { \
        setstate( std::ios_base::failbit ); \
        delete stream.release(); \
        return false; \
        } \
    Super::init( stream->rdbuf() ); \
    return true;

#define CLOSE_IMPLEMENTATION( StreamName ) \
    bool VolumeManager::StreamName::is_open() const \
        { \
        return rdbuf() != NULL; \
        } \
    void VolumeManager::StreamName::close() \
        { \
        if( is_open() ) \
            { \
            delete stream.release(); \
            Super::init( NULL ); \
            } \
        }


VolumeManager::IFStream::IFStream()
    : Super( NULL )
    { 
    }


VolumeManager::IFStream::IFStream
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) : Super( NULL ) 
    { 
    open( stream_name, mode, flags ); 
    } 


bool VolumeManager::IFStream::open 
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) 
    { 
    RESOLVE_AND_INIT();

    if( resolved.is_native )
        { 
        if( !flags_on( flags, STREAM_COMPRESSED ) ) 
            { 
            stream.reset
                (
                new std::ifstream( resolved.relative_name.c_str(), mode )
                ); 
            } 
        else 
            { 
            stream.reset
                (
                new CompressedStream::IStream
                    (
                    new std::ifstream( resolved.relative_name.c_str(), mode )
                    )
                );
            }
        }
    else
        {
        BasicVolume * volume = checked_cast<BasicVolume *>( resolved.file_system );
        
        stream.reset
            (
            new BasicVolume::IFStream
                ( *volume, resolved.relative_name, mode )
            );
        }
    
    BIND_STREAM();
    }


CLOSE_IMPLEMENTATION( IFStream );

// Implementation of VolumeManager::OFStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

VolumeManager::OFStream::OFStream()
    : Super( NULL )
    { 
    } 


VolumeManager::OFStream::OFStream
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) : Super( NULL ) 
    { 
    open( stream_name, mode, flags ); 
    } 


bool VolumeManager::OFStream::open 
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) 
    { 
    RESOLVE_AND_INIT();

    // When opening a file, that does not exist or refers
    // to a not existing directory, for writing, the directory is created
    // first. BasicVolume does this automatically.

    if( resolved.is_native )
        { 
        CREATE_FILE();

        if( !flags_on( flags, STREAM_COMPRESSED ) ) 
            { 
            stream.reset
                (
                new std::ofstream( resolved.relative_name.c_str(), mode )
                ); 
            } 
        else 
            { 
            stream.reset
                (
                new CompressedStream::OStream
                    ( new std::ofstream( resolved.relative_name.c_str(), mode ) )
                );
            } 
        } 
    else 
        { 
        BasicVolume * volume = checked_cast<BasicVolume *>( resolved.file_system );

        stream.reset
            (
            new BasicVolume::OFStream
                (
                *volume,
                resolved.relative_name,
                mode,
                flags_on( flags, STREAM_COMPRESSED ) ?
                BasicVolume::STREAM_COMPRESSED : 0
                )
            );
        } 
    
    BIND_STREAM();    
    }


CLOSE_IMPLEMENTATION( OFStream );

// Implementation of VolumeManager::FStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

VolumeManager::FStream::FStream()
    : Super( NULL )
    {
    } 


VolumeManager::FStream::FStream
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) : Super( NULL ) 
    { 
    open( stream_name, mode, flags ); 
    } 


bool VolumeManager::FStream::open 
    ( 
    const std::string &     stream_name, 
    std::ios_base::openmode mode, 
    int                     flags 
    ) 
    { 
    RESOLVE_AND_INIT();
    
    // When opening a file, that does not exist or refers
    // to a not existing directory, for writing, the directory is created
    // first. BasicVolume does this automatically.

    if( resolved.is_native ) 
        { 
        CREATE_FILE();

        hope( !flags_on( flags, STREAM_COMPRESSED ) );
        stream.reset
            ( new std::fstream( resolved.relative_name.c_str(), mode ) ); 
        } 
    else 
        { 
        BasicVolume * volume = checked_cast<BasicVolume *>( resolved.file_system );

        stream.reset
            (
            new BasicVolume::FStream
                (
                *volume,
                resolved.relative_name.c_str(),
                mode
                )
            );
        } 
    
    BIND_STREAM();
    }


CLOSE_IMPLEMENTATION( FStream );

#undef RESOLVE_AND_INIT
#undef CREATE_FILE
#undef BIND_STREAM
#undef CLOSE_IMPLEMENTATION


// Implementation of VolumeManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

VolumeManager::My * VolumeManager::my = NULL;


VolumeManager::VolumeManager()
    {
    hope( !my );
    my = new My();
    }


VolumeManager::~VolumeManager()
    {
    delete my;
    my = NULL;
    }


void VolumeManager::mount_volume
    (
    const std::string & image_path,
    int                 mount_flags
    )
    {
    My::self().mount_volume( image_path, std::string(), mount_flags, false );
    }


void VolumeManager::dismount_volume
    (
    const std::string & image_path
    )
    {
    My::self().dismount_volume( image_path );
    }


void VolumeManager::format_volume
    (
    const std::string & image_path,
    const std::string & volume_name,
    int                 mount_flags
    )
    {
    My::self().mount_volume( image_path, volume_name, mount_flags, true );
    }


FileSystem & VolumeManager::file_system()
    {
    return My::self();
    }


void VolumeManager::next_tick()
    {
    SystemTime flush_period = My::self().settings.flush_period.read_float();
    SystemTime last_time_flushed = My::self().last_time_flushed;
    SystemTime current_time = Global::system_time();
    
    if( ( last_time_flushed + flush_period ).is_older_than( current_time ) )
        {
        flush();
        My::self().last_time_flushed = current_time;
        }
    }


void VolumeManager::flush()
    {
    My::self().flush();
    }


} // namespace Core
} // namespace Massiv
