/** @file volume_manager.h
 * Volume manager.
 * Virtual file system with IO stream interface.
 */

/*
 * begin        : Fri 21 Jun 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VOLUME_MANAGER_H
#define MASSIV_VOLUME_MANAGER_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#include <iostream>
#include <memory>

namespace Massiv {
namespace Core {


class FileSystem;


/** @addtogroup filesys
 * @{
 */

/**
 * Virtual file system that defines single interface for accessing
 * streams on the native file system and a set of "mounted" volumes
 * (such as Volume or CompactVolume images).
 *
 * When VolumeManager looks up a stream it searches the native file system
 * first, then mounted volumes in the order as they were mounted.
 * Potential name clashes are ignored.
 * Volumes are always mounted to the "root" of the virtual file system
 * and each volume can be mounted at most once.
 * Each volume can have a volume name that can be used to force VolumeManager
 * to look up for streams on that volume only. This is done by prefixing
 * stream name by "volume_name://" pattern. "native://" refers to the
 * native file system.
 * Stream semantics should be the same for both the native file system
 * and mounted volumes.
 * There can not be multiple VolumeManager instances.
 */
class MASSIV_EXPORTED VolumeManager
    {
public:
    
    /**
     * Flags to XXXStream::open().
     */
    enum StreamFlags
        {
        STREAM_COMPRESSED           =   1,
            /**< Treat stream as compressed.
                 Must be specified when opening a compressed stream
                 that is stored on the native file system. */

        DEFAULT_INPUT_FLAGS         =   0,
            /**< Default flags used by IFStream. */
        DEFAULT_OUTPUT_FLAGS        =   0,
            /**< Default flags used by OFStream. */
        DEFAULT_INPUT_OUTPUT_FLAGS  =   0,
            /**< Default flags used by FStream. */

        }; // StreamFlags
    
    /**
     * Flags to @c mount_volume().
     */
    enum MountFlags
        {
        MOUNT_VOLUME            =   1,
            /**< Interpret image file as Volume image. */
        MOUNT_COMPACT_VOLUME    =   2,
            /**< Interpret image file as CompactVolume image. */
        MOUNT_READ_ONLY         =   4,
            /**< Mount the volume in read-only mode. */

        DEFAULT_MOUNT_FLAGS     =   MOUNT_COMPACT_VOLUME,
            /**< Default mount flags. */
        DEFAULT_FORMAT_FLAGS    =   MOUNT_COMPACT_VOLUME
            /**< Default format flags. */
        };

private:

    class My;
    class VolumeManagerSettings;

    friend class My;

public:

    //
    // VolumeManager streams.
    //
        
    /**
     * Input stream.
     */
    class MASSIV_EXPORTED IFStream : public std::istream
        {
    private:

        typedef std::istream Super;

    public:
            
        /** Default constructor. */
        IFStream();

        /** Constructor. */
        IFStream
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in,
            int                     flags = DEFAULT_INPUT_FLAGS
            );

        /**
         * Opens stream.
         * Returns true if stream was opened successfully.
         * Throws @c ExVolumeNotMounted if @c stream_name explictly refers
         * to unknown volume.
         */
        bool open
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in,
            int                     flags = DEFAULT_INPUT_FLAGS
            );

        /** Tests if stream is opened. */
        bool is_open() const;
        
        /** Closes the stream. */
        void close();

    private:

        std::auto_ptr<std::istream> stream;
            /**< Pointer to device-specific istream. */

        }; // IFStream
    
    /**
     * Output stream.
     * For information on stream interface see @c IFStream.
     */
    class MASSIV_EXPORTED OFStream : public std::ostream
        {
    private:

        typedef std::ostream Super;
    
    public:
            
        /** Default constructor. */
        OFStream();

        /** Constructor. */
        OFStream
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::out,
            int                     flags = DEFAULT_OUTPUT_FLAGS
            );

        /** Opens stream. */
        bool open
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::out,
            int                     flags = DEFAULT_OUTPUT_FLAGS
            );
        
        /** Tests if stream is opened. */
        bool is_open() const;
        
        /** Closes the stream. */
        void close();

    private:

        std::auto_ptr<std::ostream> stream;
            /**< Pointer to device-specific ostream. */

        }; // OFStream
    
    /**
     * Input/output stream.
     * For information on stream interface see @c IFStream.
     */
    class MASSIV_EXPORTED FStream : public std::iostream
        {
    private:

        typedef std::iostream Super;

    public:
            
        /** Default constructor. */
        FStream();

        /** Constructor. */
        FStream
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
            int                     flags = DEFAULT_INPUT_OUTPUT_FLAGS
            );

        /** Opens stream. */
        bool open
            (
            const std::string &     stream_name,
            std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out,
            int                     flags = DEFAULT_INPUT_OUTPUT_FLAGS
            );
        
        /** Tests if stream is opened. */
        bool is_open() const;

        /** Closes the stream. */
        void close();

    private:

        std::auto_ptr<std::iostream> stream;
            /**< Pointer to device-specific ostream. */

        }; // FStream
    
    friend class IFStream;
    friend class OFStream;
    friend class FStream;

public:

    //
    // Volume management.
    //

    /** Constructor. */
    VolumeManager();
    
    /** Destructor. */
    ~VolumeManager();

    /**
     * Mounts a volume whose image is stored on the native file system
     * under @c image_path name.
     * The new volume will be mounted to the root of the volume manager
     * namespace.
     * Each volume can be mounted at most once.
     */
    static void mount_volume
        (
        const std::string & image_path,
        int                 mount_flags = DEFAULT_MOUNT_FLAGS
        );

    /**
     * Dismounts volume with this @c image_path.
     * Any streams that would refer to that volume must have already
     * been closed.
     */
    static void dismount_volume
        (
        const std::string & image_path
        );

    /**
     * Creates a new volume image with location specified by @c image_path 
     * and mounts it.
     *
     * @param image_path Specifies the location directory of the image.
     */
    static void format_volume
        (
        const std::string & image_path,
        const std::string & volume_name,
        int                 mount_flags = DEFAULT_FORMAT_FLAGS
        );

    /**
     * Returns file system interface to the virtual file system.
     */
    static FileSystem & file_system();
    
    /**
     * VolumeManager's next tick function.
     * Calls @c flush() every few seconds as specified
     * in #VolumeManagerSettings.
     */
    static void next_tick();

    /**
     * Flushes file systems.
     */
    static void flush();

private:

    static My * my;
        /**< Volume manager state. */

    }; // VolumeManager

/** @} */ // addtogroup filesys


} // namespace Core
} // namespace Massiv


#endif // MASSIV_VOLUME_MANAGER_H
