/** @file common_garbage.h
 * Simple utility functions and other garbage.
 * Probably should be moved to other headers, should not be included from common.h.
 */

/*
 * begin        : 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_COMMON_GARBAGE_H
#define MASSIV_COMMON_GARBAGE_H


namespace Massiv {
namespace Core {


/** @def NO_DEFAULT_CONSTRUCTOR
 * Disable implicit default constructor.
 *
 * @ingroup utility
 */
#define NO_DEFAULT_CONSTRUCTOR( Type ) private: Type()

/** @def NO_COPY_CONSTRUCTOR
 * Disable implicit copy constructor.
 *
 * @ingroup utility
 */
#define NO_COPY_CONSTRUCTOR( Type ) private: Type( const Type & )


/** @def NO_ASSIGNMENT_OPERATOR
 * Disable implicit assignment operator.
 *
 * @ingroup utility
 */
#define NO_ASSIGNMENT_OPERATOR( Type ) private: Type & operator=( const Type & )


/** @def FRIEND_AUTO_PTR
 * Make <tt>std::auto_ptr< @a Class ></tt> friend.
 *
 * @ingroup utility
 */
#ifdef HACK_AUTO_PTR_IS_STRUCT
#define FRIEND_AUTO_PTR( Class ) friend struct std::auto_ptr< Class >
#else
#define FRIEND_AUTO_PTR( Class ) friend class std::auto_ptr< Class >
#endif // HACK_AUTO_PTR_IS_STRUCT


/**
 * Get rid of "unreferenced variable" warnings.
 *
 * @ingroup utility
 */
template< class Type >
void reference_value
    (
    Type
    )
    {
    }


/** @def checked_cast
 * checked_cast<> performs checked pointer conversion
 * in DEBUG mode, static conversion otherwise.
 * May be used on polymorphic types only
 * (so that dynamic_cast<> would work on it)!
 *
 * @ingroup utility
 */
#ifdef OPTION_DEBUG
#define checked_cast        dynamic_cast
#else
#define checked_cast        static_cast
#endif // OPTION_DEBUG


/**
 * Checks if @c yes_flags are set and @c no_flags are not set
 * in ORed flag @c combination.
 *
 * @ingroup utility
 */
inline bool flags_on
    (
    int combination,
    int yes_flags,
    int no_flags = 0
    )
    {
    return ( combination & ( yes_flags | no_flags ) ) == yes_flags;
    }

/**
 * Checks if @c value is represented in @c T in its @c bit_count lower bits.
 *
 * @ingroup utility
 */
template <class T>
inline bool only_lower_bits
    (
    T   value,
    int bit_count
    )
    {
    return ( value & ~( ( 1 << bit_count ) - 1 ) ) == 0;
    }

/**
 * Returns @c bit_count lower bits of @c value.
 *
 * @ingroup utility
 */
template <class T>
inline T get_lower_bits
    (
    T   value,
    int bit_count
    )
    {
    return value & ( ( 1 << bit_count ) - 1 );
    }


/**
 * Return @c true if given integral @a value is power of 2
 * (whether exactly single bit of @a value is set).
 *
 * @ingroup utility
 */
template< typename T >
inline bool is_power_of_two
    (
    T value
    )
    {
    if( value == 0 )
        {
        return false;
        }
    return ( value & ( value - 1 ) ) ? false : true;
    }


/** @def RELEASE_NO_THROW
 * Use this to specify that method throws no exceptions on release mode
 * (ie. OPTION_DEBUG not defined) to hint code optimizer.
 *
 * @ingroup utility
 */
#ifndef OPTION_DEBUG
#define RELEASE_NO_THROW()  throw()
#else
#define RELEASE_NO_THROW()
#endif


/** @def DEBUG_ONLY
 * Expands to its argument if @c OPTION_DEBUG is defined and to nothing otherwise.
 *
 * @ingroup utility
 */

/** @def DEBUG_COMMA
 * Expands to comma if @c OPTION_DEBUG is defined and to nothing otherwise.
 *
 * @ingroup utility
 */
#ifdef OPTION_DEBUG
#define DEBUG_ONLY( x )     x
#define DEBUG_COMMA         ,
#else
#define DEBUG_ONLY( x )
#define DEBUG_COMMA
#endif


/** @def INLINE
 * Declare method or function as inline.
 *
 * Every function or method that should be inlined when compiling with
 * optimiziations turned on (@c OPTION_OPTIMIZE defined) should be declared
 * using @c INLINE keyword, which expands to @c inline when @c OPTION_OPTIMIZE
 * is defined and to nothing otherwise.
 *
 * Functions declared with @c INLINE in @c file.h should be defined in
 * @c file_inline.h. @c file_inline.h should be included from @c file.h if
 * @c ENABLE_INLINE is defined and from @c file.cpp if ENABLE_INLINE is
 * not defined.
 *
 * file.h:
 * @code
 * #ifndef _FILE_H_
 * #define _FILE_H_
 *
 * class X
 *     {
 * public:
 *     void fni();
 *     INLINE void fi();
 *     }; // class X
 *
 * #ifdef ENABLE_INLINE
 * #include "file_inline.h"
 * #endif
 *
 * #endif // _FILE_H_
 * @endcode
 *
 * file.cpp:
 * @code
 * #include "file.h"
 *
 * void X::fni() { ... }
 *
 * #ifndef ENABLE_INLINE
 * #include "file_inline.h"
 * #endif
 * @endcode
 *
 * file_inline.h:
 * @code
 * void X::fi() { ... }
 * @endcode
 *
 * Never include @c file_inline.h from other files than @c file.cpp or @c file.h.
 * Do not use #ifdef constructs to guard contents of @c file_inline.h from multiple
 * inclusions. Do not use @c INLINE when defining function, only in declarations.
 * Do not write implementations of templates in @c file_inline.h, these should be
 * either in @a file.h or in @c file_impl.h if @a file.h is "too big". @c file_impl.h
 * should be always included from @c file.h of course.
 *
 * Use files in @c src/template/ directory as templates.
 *
 * @ingroup utility
 */

/** @def ENABLE_INLINE
 * Enable inlining.
 *
 * @ingroup utility
 */

#ifdef OPTION_OPTIMIZE
#define ENABLE_INLINE
#endif

#ifdef ENABLE_INLINE
#define INLINE inline
#else
#define INLINE
#endif


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_COMMON_GARBAGE_H
