/** @file common_win32.h
 * Common win32 stuff.
 */

/*
 * begin        : 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_COMMON_WIN32_H
#define MASSIV_COMMON_WIN32_H


#if defined( COMPILER_MSVC )
#ifndef MASSIV_COMMON_MSVC_H
#include "global/common_msvc.h"
#endif

#elif defined( COMPILER_GCC )
#ifndef MASSIV_COMMON_GCC_H
#include "global/common_gcc.h"
#endif

#else
#error "Unsupported Win32 compiler."
#endif // compiler


//
// Define additional Win32 build specific macros.
//

/** @def MASSIV_EXPORTED
 * Define MASSIV_EXPORTED macro which is used to mark symbols defined by and
 * exported from the core.
 *
 * If shared libraries are enabled, define MASSIV_EXPORTED to __declspec(dllexport)
 * when compiling core library objects and to __declspec(dllimport) when compiling
 * objects that may reference symbols from core library. 
 * This has the effect that symbols are exported while building core and imported
 * from the DLL when linking against (using) core.
 *
 * Under Win32 symbols that are exported/imported from a DLL must be explicitly 
 * marked by the __declspec(dllimport/dllexport) attribute.
 * At the time of compile the compiler must be able to determine whether the symbol
 * will be imported from a DLL or not (access to DLL-imported symbols/classes might be
 * compiled in a special way).
 *
 * General rules on how to use MASSIV_EXPORTED:
 *
 *  - each symbol that is exported from (published by) the core must be marked
 *    MASSIV_EXPORTED:
 *      class MASSIV_EXPORTED MyClass; // Declares an exported class.
 *      MASSIV_EXPORTED void f(); // Declares an exported global function.
 *      // Global variables can also be exported but in Massiv we do not
 *      // have them (all public global variables are wrapped by
 *      // Massiv::Core::Global class).
 *
 *  - each symbol that is not defined by (implemented in) the core must not be
 *    MASSIV_EXPORTED
 *  
 * Advanced rules that deal with template classes:
 *
 *  - if a template class can be instantiated outside the core and the set
 *    of all potential instantiations (ie. template arguments) is not known
 *    apriori when building the core, the template definition must be visible to
 *    the code that uses the core. In that case template class <b>must not</b> be
 *    MASSIV_EXPORTED (or you will get unresolved references because
 *    the linker will not find code for your template instances).
 *
 *  - if a template class is instantiated by the core only or there is a known
 *    preset set of all potential instantiations, the class can be exported.
 *    However the template class must also be explicitly instantiated in the
 *    core in all its instances so that its definitions are correctly exported
 *    from the core.
 *
 *    header_published_by_the_core.h:
 *
 *    // T can be only 'int' or 'float'.
 *    template< class T >
 *    class MASSIV_EXPORTED MyClass { void f(); }
 *
 *    file_linked_to_the_core_dll.cpp:
 *
 *    // Instantiate the MASSIV_EXPORTED template for all possible arguments
 *    // and ensure that its code will be exported from the core.
 *    template class MASSIV_EXPORTED MyClass< int >;
 *    template class MASSIV_EXPORTED MyClass< float >;
 *
 *  - if some instances of a template class should be exported because they
 *    are defined by (implemented in) the core and the other instances not because
 *    they are defined outside the core, template class must be specialized.
 *
 *    header_published_by_the_core.h:
 *
 *    template< class T >
 *    class MyClass { ... };
 *
 *    template<>
 *    class MASSIV_EXPORTED MyClass<int>;
 *
 *    // All MyClass instances, except MyClass<int>, are not exported
 *    // from the core.
 *
 *    file_linked_to_the_core_dll.cpp:
 *
 *    // Instantiate the MyClass instances exported from the core.
 *    template class MASSIV_EXPORTED MyClass<int>;
 *
 * Note:
 *  This model with a single MASSIV_EXPORTED macro assumes that the whole
 *  core is either built as a DLL or not.
 *
 * @ingroup utility
 */
#ifdef OPTION_SHARED_LIBRARIES
    #if defined( OPTION_BUILD_SHARED_MASSIV )
        #define MASSIV_EXPORTED __declspec(dllexport)
    #elif defined( OPTION_USE_SHARED_MASSIV )
        #define MASSIV_EXPORTED __declspec(dllimport)
    #endif
#else
    #define MASSIV_EXPORTED
#endif


#endif // MASSIV_COMMON_WIN32_H
