/** @file global.h
 * Global objects.
 */

/*
 * begin        : Tue Nov 22 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_GLOBAL_H
#define MASSIV_GLOBAL_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATUS_H
#include "status/status.h"
#endif

#include <string>


namespace Massiv {

class System;
class SystemContext;

namespace Core {


// Note: Headers must be included by user.
class Logger;
class LogMessage;
class ObjectManager;
class GlobalSearchManager;
class ObjectProxyManager;
class GarbageCollector;
class ClassManager;
class NodeManager;
class NodeActionManager;
class ArchiveManager;
class STime;
class Status;
class SystemEventQueue;
class TimeManager;
class Network;
class Registry;
class NodeDatabase;
class ArchiveDatabase;
class WellKnownObjectIdDatabase;
class DataManager;
class ThreadManager;
class CriticalSectionManager;
class SemaphoreManager;
class NodeId;
class ReplicaClient;
class ReplicaServer;
class VolumeManager;
class SystemTime;
class ReplicaTime;
class FileSystem;
class FileAckManager;
class Property;
class ReplicaManager;
class AccountManager;
class PathManager;
class Exception;
class Balancer;
class StackTracer;

          
/** @class Global
 * This class encapsulates all global objects in namespace Massiv::Core.
 * References to global objects are acquired via calls to static methods.
 * While Global is intended to be used by the core itself, another
 * static object @c Massiv::System defines public API to the system and the
 * global objects. @c Massiv::System is responsible for core startup
 * and Global initialization.
 *
 * @note The statemenet above is actually not true. Code using Massiv core
 *       may want (will have) to use some global core objects, such as logger,
 *       object manager or replica manager.
 *
 * @ingroup global
 */
class MASSIV_EXPORTED Global
    {
public:

    friend class Massiv::System;
    friend class Exception;
    friend class StackFrameEntry;

    //
    // Global objects.
    //

    /**
     * Check if logger is available.
     */
    static INLINE bool logger_available();

    /**
     * Get log manager (complete Logger).
     */
    static INLINE Logger & logger();

    /**
     * Get object manager.
     */
    static INLINE ObjectManager & object_manager();

    /**
     * Get global search manager.
     */
    static INLINE GlobalSearchManager & global_search_manager();

    /**
     * Get object proxy manager.
     * Object manager must already be running.
     */
    static INLINE ObjectProxyManager & object_proxy_manager();

    /**
     * Get garbage collector.
     * Object manager must already be running.
     */
    static INLINE GarbageCollector & garbage_collector();
    
    /**
     * Get balancer.
     * Object manager must already be running.
     */
    static INLINE Balancer & balancer();

    /**
     * Get class manager.
     */
    static INLINE ClassManager & class_manager();

    /**
     * Get node manager.
     */
    static INLINE NodeManager & node_manager();
    
    /**
     * Get node action manager.
     */
    static INLINE NodeActionManager & node_action_manager();

    /**
     * Get archive manager.
     */
    static INLINE ArchiveManager & archive_manager();

    /**
     * Get system event queue.
     */
    static INLINE SystemEventQueue & system_event_queue();

    /**
     * Get time manager.
     */
    static INLINE TimeManager & time_manager();

    /**
     * Get network object.
     */
    static INLINE Network & network();

    /**
     * Get registry object.
     */
    static INLINE Registry & registry();

    /**
     * Get node database object.
     */
    static INLINE NodeDatabase & node_database();

    /**
     * Get archive database object.
     */
    static INLINE ArchiveDatabase & archive_database();
    
    /**
     * Get well know object id database.
     */
    static INLINE WellKnownObjectIdDatabase & well_known_object_id_database();

    /**
     * Get data object manager.
     */
    static INLINE DataManager & data_manager();
    
    static INLINE DataManager & data_manager_test_server();
    static INLINE DataManager & data_manager_test_client();

    /**
     * Get account manager.
     */
    static INLINE AccountManager & account_manager();

    /**
     * Get path manager.
     */
    static INLINE PathManager & path_manager();

    /**
     * Get thread manager.
     */
    static INLINE ThreadManager & thread_manager();

    /**
     * Get critical section global object.
     */
    static INLINE CriticalSectionManager & critical_section_manager();

    /**
     * Get semaphore global object.
     */
    static INLINE SemaphoreManager & semaphore_manager();
    
    /**
     * Get stack tracer.
     */
    static INLINE StackTracer & stack_tracer();

    /**
     * Test if stack tracer is available.
     */
    static INLINE bool stack_tracer_available();

    /**
     * Get access to native file system.
     */
    static INLINE FileSystem & native_file_system();

    /**
     * Get volume manager object.
     */
    static INLINE VolumeManager & volume_manager();

    /**
     * Get file ack manager.
     */
    static INLINE FileAckManager & file_ack_manager();
    
    /**
     * Get replication manager.
     */
    static INLINE ReplicaManager & replica_manager();

    /**
     * Get replication server.
     */
    static INLINE ReplicaServer & replica_server();

    /**
     * Get local node id.
     */
    static NodeId get_local_node_id();

    /**
     * Get system context.
     */
    static INLINE SystemContext & system_context();
    
public:

    /**
     * Get system time.
     */
    static INLINE const SystemTime & system_time();

public:

    //
    // Logging.
    //

    /**
     * Send message to logger.
     */
    static bool log
        (
        const LogMessage &  message
        );

    /**
     * Send message to logger.
     */
    static bool log
        (
        const Status &      status,
        const std::string & message
        );

    /**
     * Send debug message to logger.
     */
    static bool log_debug
        (
        Status::Facility    facility,
        Status::Priority    priority,
        const std::string & message
        );

    /**
     * Send info message to logger.
     */
    static bool log_info
        (
        Status::Facility    facility,
        Status::Priority    priority,
        const std::string & message
        );

    /**
     * Send warning message to logger.
     */
    static bool log_warning
        (
        Status::Facility    facility,
        Status::Priority    priority,
        const std::string & message
        );

    /**
     * Send error message to logger.
     */
    static bool log_error
        (
        Status::Facility    facility,
        const std::string & message
        );

public:

    //
    // Properties & replication.
    //

    /**
     * Are writes to replica properties allowed?
     * Nasty hack.
     */
    static INLINE bool replica_writes_allowed_hack();
    
    /**
     * Get current replication timestamp.
     */
    static INLINE const ReplicaTime & replica_time();

    /**
     * Get stack root object.
     */
    static INLINE Property * get_stack_root_object();

protected:

    /// Only ReplicaClient can allow writes to replicas.
    friend class ReplicaClient;

    /**
     * (Dis)allow writes to replicas.
     */
    static INLINE void allow_replica_writes
        (
        bool allow
        );

private:

    // Globals.

    static Logger *                 g_logger;
        /**< Logger. */
    static ObjectManager *          g_object_manager;
        /**< Object manager. */
    static GlobalSearchManager *    g_global_search_manager;
        /**< Global search manager. */
    static ClassManager *           g_class_manager;
        /**< Class manager. */
    static NodeManager *            g_node_manager;
        /**< Node manager. */
    static NodeActionManager *      g_node_action_manager;
        /**< Node action manager. */
    static ArchiveManager *         g_archive_manager;
        /**< Archive manager. */
    static ArchiveDatabase *        g_archive_database;
        /**< Archive database. */
    static WellKnownObjectIdDatabase * g_well_known_object_id_database;
        /**< Well known object id database. */
    static SystemEventQueue *       g_system_event_queue;
        /**< System event queue. */
    static TimeManager *            g_time_manager;
        /**< Time manager. */
    static DataManager *            g_data_manager;
        /**< Data manager. */
    static AccountManager *         g_account_manager;
        /**< Account manager. */
    static PathManager *            g_path_manager;
        /**< Path manager. */
    static ThreadManager *          g_thread_manager;
        /**< Thread manager. */
    static SemaphoreManager *       g_semaphore_manager;
        /**< Semaphore manager. */
    static CriticalSectionManager * g_critical_section_manager;
        /**< Critical section manager. */
    static StackTracer *            g_stack_tracer;
        /**< Stack trace. */
    static VolumeManager *          g_volume_manager;
        /**< Volume manager. */
    static FileSystem *             g_native_file_system;
        /**< Native file system. */
    static FileAckManager *         g_file_ack_manager;
        /**< File ack manager. */
    static ReplicaManager *         g_replica_manager;
        /**< Replica manager. */
    static Network *                g_network;
        /**< Network. */
    static Registry *               g_registry;
        /**< Registry. */
    static NodeDatabase *           g_node_database;
        /**< Node database. */
    static SystemTime               g_system_time;
        /**< System time. */
    static bool                     g_replica_writes_allowed;
        /**< Are writes to replica properties allowed? Nasty hackish flag. */
    static bool                     g_log_exceptions;
        /**< Is exception logging enabled? */
    static bool                     g_print_stack_traces;
        /**< Can print stack traces? Must be false unless stack tracing
             facilities are ready to be used. */
    static SystemContext *          g_system_context;
        /**< System context. */

    // Cached pointers to components of Globals.

    static GarbageCollector *       g_gc;
        /**< Garbage collector. Valid only if ObjectManager exists. */
    static ObjectProxyManager *     g_proxy_manager;
        /**< Object proxy manager. Valid only if ObjectManager exists.  */
    static ReplicaServer *          g_replica_server;
        /**< Replica server. Valid only if ObjectManager exists. */
    static const ReplicaTime *      g_replica_time_ptr;
        /**< Replica time. Valid only if ObjectManager exists. */
    static Balancer *               g_balancer;
        /**< Balancer. Valid only if ObjectManager exists. */
    static Property *               g_stack_root_object;
        /**< Stack root object. Non-NULL iff ObjectManager exists. */
    
    // For testing purposes only.

    static DataManager *            g_data_manager_test_server;
    static DataManager *            g_data_manager_test_client;
    
    }; // Global


} // namespace Massiv::Core
} // namespace Massiv

#ifdef ENABLE_INLINE
#include "global_inline.h"
#endif

#endif // MASSIV_GLOBAL_H
