/** @file global_callbacks.cpp
 * Global callbacks for the massiv core.
 */
 
/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/global_callbacks.h"
#include "system/system.h"
#include "global/global.h"
#include "object/providerid.h"
#include "object/object_provider_state.h"
#include "object/object_manager.h"
#include "database/archive_database.h"
#include "object/archive_manager.h"
#include "status/status.h"
#include "database/node_database.h"
#include "pointer/garbage_collector.h"
#include "replication/replica_server.h"
#include "object/lib/account_object_interface.h"
#include "object/lib/account_object_interface_rpc.h"
#include "pointer/object_pointer.h"
#include "pointer/pointer_to_object.h"
#include "data/data_manager.h"
#include "data/path_manager.h"
#include "time/time_manager.h"
#include "net/net.h"
#include "sysevent/system_event_queue.h"
#include "node/node_action_manager.h"
#include "node/node_manager.h"
#include "thread/stack_tracer.h"


namespace Massiv {
namespace Core {

namespace {


/**
 * Update proxy node objects naming - update local name mapping for registered nodes
 * (name mapping for their proxy node objects).
 */
void refresh_proxy_node_objects_name_mapping()
    {
    class DefineNameEnumerator : public NodeDatabase::Enumerator
        {
        virtual void operator()
            (
            const NodeDatabase::NodeData & data
            )
            {
            GarbageCollector & gc = Global::garbage_collector();
            gc.define_local_name( data.get_proxy_node_object_id(), data.get_name() );
            }
        
        }; // DefineNameEnumerator

    NodeDatabase & node_database = Global::node_database();
    DefineNameEnumerator define_name_enumerator;
    
    // Mapping for server nodes is defined on all nodes.
    node_database.enumerate_nodes( NodeId::TYPE_SERVER, define_name_enumerator );
    
#if 0 // Mapping for client nodes is useless on server nodes?

    // Mapping for client nodes is defined only on server nodes
    // because node database on a client node should not contain
    // information about nother client nodes.
    if( System::is_server() )
        {
        node_database.enumerate_nodes( NodeId::TYPE_CLIENT, define_name_enumerator );
        }
#endif // 0
    
    // For client nodes define name "client" for local node itself.
    if( System::is_client() )
        {
        GarbageCollector & gc = Global::garbage_collector();
        NodeDatabase::NodeData node_data = node_database.get_node_data( node_database.get_local_node_id() );
        gc.define_local_name( node_data.get_proxy_node_object_id(), "client" );
        }
    }


void load_simulation_from_archive
    (
    const std::string & archive_file_name
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    std::string swap_file_name = Global::path_manager().get_archive_directory() + "/swap_file";

    // Log this.
        {
        std::ostringstream ss;
        ss  << "Starting simulation from archive file \"" << archive_file_name << "\"...";
        Global::log
            (
            STATUS_INFO
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );
        }

    // Copy archive to swap file.
        {
        // Log.
            {
            std::ostringstream ss;
            ss  << "Copying archive file \"" << archive_file_name << "\" to swap file \""
                << swap_file_name << "\".";
            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_ARCHIVATION,
                    Status::PRIORITY_NORMAL
                    ),
                ss.str()
                );
            }

        if( swap_file_name == archive_file_name )
            {
            std::ostringstream ss;
            ss  << "Archive file can't have same name as swap file \"" << swap_file_name << "\"!";
            throw ExIOError( ss.str() );
            }

        std::ifstream   in( archive_file_name.c_str(), std::ios_base::in | std::ios_base::binary );
        std::ofstream   out( swap_file_name.c_str(), std::ios_base::out | std::ios_base::trunc | std::ios_base::binary );

        if( in.fail() )
            {
            std::ostringstream ss;
            ss  << "Unable to open archive file \"" << archive_file_name << "\" for reading.";
            throw ExIOError( ss.str() );
            }

        if( out.fail() )
            {
            std::ostringstream ss;
            ss  << "Unable to open swap file \"" << swap_file_name << "\" for writting.";
            throw ExIOError( ss.str() );
            }

        out << in.rdbuf();

        if( out.fail() )
            {
            std::ostringstream ss;
            ss  << "Error copying archive file \"" << archive_file_name << "\" "
                << "to swap file \"" << swap_file_name << "\".";
            throw ExIOError( ss.str() );
            }
        }

    object_manager.open_swap( swap_file_name );
    }


} // anonymous namespace


void GlobalCallbacks::system_initialized()
    {
    // Warning: Local node id is anonymous!
    }


void GlobalCallbacks::system_shutdown()
    {
    }


void GlobalCallbacks::identity_changed()
    {
    // Reinitialize network.
    Global::network().initialize();

    // Switch data manager to full mode.
    Global::data_manager().switch_from_to_anonymous_mode();

    Global::node_manager().initialize();
    
    // Throw away messages for old identity.
    Global::system_event_queue().reset();

    // Reset time manager.
    Global::time_manager().initialize();
    }


void GlobalCallbacks::system_connected()
    {
    // If the local system is client node we must create dummy object provider
    // so the client node will be able to generate object ids.
    // Note: If the local system is server, object providers are restored
    //       from archive. Service nodes can't create objects at all.
    
    if( System::is_client() )
        {
        ObjectManager &     object_manager = Global::object_manager();
        ProviderId          dummy_provider_id = ProviderId::create_dummy_provider_id();
        ObjectProviderState dummy_provider_state( dummy_provider_id, 1 );
        
        // Register dummy object provider only if it doesn't exist yet.
        if( !object_manager.is_local_provider( dummy_provider_id ) )
            {
            // Log.
            Global::log_info
                (
                Status::FACILITY_SYSTEM,
                Status::PRIORITY_NORMAL,
                "Creating dummy object provider..."
                );
    
            // Register the dummy object provider.
            object_manager.register_local_provider( dummy_provider_state );
            }
        }
        
    if( System::is_server() )
        {
        NodeActionManager & node_action_manager = Global::node_action_manager();
        node_action_manager.initialize();
        }

    System::enable( System::CAPABILITY_SIMULATION );
    System::enable( System::CAPABILITY_AUTO_ARCHIVATION );
    System::enable( System::CAPABILITY_LOAD_BALANCING );
    }
    
    
void GlobalCallbacks::system_disconnecting()
    {
    System::disable( System::CAPABILITY_SIMULATION );
    System::disable( System::CAPABILITY_AUTO_ARCHIVATION );
    System::disable( System::CAPABILITY_LOAD_BALANCING );
    }

    
void GlobalCallbacks::next_tick_start()
    {
    }
    

void GlobalCallbacks::next_tick_finish()
    {
    }


void GlobalCallbacks::thread_starting()
    {
    // Install exception handler for the current thread.
    if( Global::stack_tracer_available() )
        {
        Global::stack_tracer().install_parachute();
        }
    }


void GlobalCallbacks::thread_stopping()
    {
    }


void GlobalCallbacks::simulation_thread_resumed
    (
    int simulation_thread_index
    )
    {
    UpcallToObject::simulation_thread_resumed( simulation_thread_index );
    }


void GlobalCallbacks::node_database_changed()
    {
    // There are some new nodes registered in node database.
    // Update system message queues for the nodes.
    Global::system_event_queue().update_server_node_queues();
    
    // Node naming - define local name mapping for registered nodes
    // (for their proxy node objects).
    refresh_proxy_node_objects_name_mapping();
    }


void GlobalCallbacks::client_connected
    (
    const NodeId &      client_node,
    bool                notify_account_object
    )
    {
    hope( client_node.is_type( NodeId::TYPE_CLIENT ) );
    
    if( notify_account_object )
        {
        NodeDatabase &                  node_database = Global::node_database();
        const NodeDatabase::NodeData &  node_data = node_database.get_node_data( client_node );
        ObjectId                        client_account_object_id = node_data.get_account_object_id();
        ObjectId                        client_node_object_id = node_data.get_proxy_node_object_id();
    
        Remote< Lib::AccountObjectInterface > account = client_account_object_id;

        account->async_client_connected( client_node_object_id );
        }
    }


void GlobalCallbacks::client_disconnected
    (
    const NodeId &      client_node,
    bool                notify_account_object
    )
    {
    hope( client_node.is_type( NodeId::TYPE_CLIENT ) );

    // Notify account object.
    
    if( notify_account_object )
        {
        NodeDatabase &                  node_database = Global::node_database();
        const NodeDatabase::NodeData &  node_data = node_database.get_node_data( client_node );
        ObjectId                        client_account_object_id = node_data.get_account_object_id();
    
        Remote< Lib::AccountObjectInterface > account = client_account_object_id;

        account->async_client_disconnected();
        }
    }


void GlobalCallbacks::node_disconnected
    (
    const NodeId &  node_id
    )
    {
    // Stop replication to client node.

    ObjectManager & object_manager = Global::object_manager();
    ReplicaServer & replica_server = *object_manager.replica_server;
    hope( &replica_server );
    replica_server.node_disconnected( node_id );
    
    // Cleanup data manager.

    Global::data_manager().destroy_request( node_id );
    
    // TODO FIXME: account manager? time manager? log routing?
    }


void GlobalCallbacks::node_not_accessible
    (
    const NodeId & node_id
    )
    {
    // FIXME ?
    
    // Warning: This function is called directly from SystemMessage::send()
    // thus it is not safe to call node_disconnected() or notify some
    // core subsystem directly from this function.
    // Is there anything we can do in this function?
    // Everything should be fine even if this function does nothing
    // because node_disconnected() should be called soon (there can be
    // some delay caused by communication protocol used to disconnect
    // client nodes from simulation for example).
    }


void GlobalCallbacks::load_simulation
    (
    const ArchivationId & archive_id_to_load
    )
    {
    hope( !archive_id_to_load.is_invalid() );
    
    ArchiveManager & archive_manager = Global::archive_manager();
    ArchiveDatabase & archive_database = Global::archive_database();
    std::string archive_file;

    // Get file name path of the archive to load simulation from.
    archive_file = archive_database.archivation_id_to_regular_path_file_name( archive_id_to_load );
    
    // Load simulation from the archive to memory.
    load_simulation_from_archive( archive_file );
    
    // Initialize archive manager.
    archive_manager.initialize( archive_id_to_load );
    }


void GlobalCallbacks::shutdown_local_node()
    {
    System::request_quit();
    }


} // namespace Massiv::Core
} // namespace Massiv
