/** @file global_callbacks.h
 * Global callbacks for the massiv core.
 */

/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_GLOBAL_CALLBACKS_H
#define MASSIV_GLOBAL_CALLBACKS_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_FWD
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_OBJECT_PROVIDER_FWD_H
#include "object/object_provider_fwd.h"    // NodeId
#endif

#ifndef MASSIV_ARCHIVATION_ID_H
#include "object/archivation_id.h"
#endif


namespace Massiv {
namespace Core {


/**
 * Abstract encapsulation class of callbacks used by various
 * massiv core subsystems.
 *
 * Core subsystems sometimes need to do something that
 * affects entire system or synchronize with other subsystems and
 * it is not good idead to leave this behaviour hidden inside
 * the subsystem. The only reason for this class is to unite global system
 * behaviour to one place and not to be scattered inside implementation
 * of many core subsystems.
 *
 * Some callbacks (namely those related to system state changes like
 * @c system_initialized(), @c identity_changed(), @c system_shutdown(), 
 * @c system_connected(), @c system_disconnecting(), @c node_database_changed())
 * should not throw exceptions or the system may be halted immediatelly.
 *
 * @ingoup global
 */
class GlobalCallbacks
    {
    NO_DEFAULT_CONSTRUCTOR( GlobalCallbacks );
    NO_COPY_CONSTRUCTOR( GlobalCallbacks );
    NO_ASSIGNMENT_OPERATOR( GlobalCallbacks );

public:

    /**
     * Local system has just been initialized.
     * This function is called just after @c System has been
     * successfully initialized and its purpose is to do
     * additional setups of the local system.
     * @warning Local node id is anonymous. If you want to do
     *      some node type specific stuff, use @c identity_changed().
     *
     * @note Called from System.
     *
     * @see System.
     */
    static void system_initialized();
    
    /**
     * Local node identity has changed.
     * Called before node connects to the simulation (the identity
     * changed to non-anonymous) or after node disconnected from the
     * simulation (the identity changed back to anonymous).
     */
    static void identity_changed();
    
    /**
     * Local system is going down.
     * This function is called before @c System's shutdown.
     */
    static void system_shutdown();
    
    /**
     * Local system has just been connected to simulation.
     */
    static void system_connected();
    
    /**
     * Local system is disconnecting.
     * Called when disconnect has been initiated.
     */
    static void system_disconnecting();
    
    /**
     * Callback which is called before any core tick begins.
     */
    static void next_tick_start();
    
    /**
     * Callback which is called at the end of core tick.
     */
    static void next_tick_finish();

    /**
     * Called to notify that new a thread is starting, invoked within
     * the context of that thread.
     */
    static void thread_starting();

    /**
     * Called to notify that current thread is stopping.
     */
    static void thread_stopping();

    /**
     * Callback which is called when simulation thread with a specified
     * @c simulation_thread_index was resumed.
     *
     * @warning The @c simulation_thread_index might be different from
     *          the global thread index obtained from
     *          @c Thread::get_current_thread_index().
     *
     * @see     System::run()
     */
    static void simulation_thread_resumed
        (
        int simulation_thread_index
        );

    /**
     * Node database has changed callback.
     */
    static void node_database_changed();
    
public:

    /**
     * Client connected notification.
     *
     * This function is called by node manager on each server node
     * during client connection to the simulation.
     *
     * Main purpose of this callback is to prepare massiv subsystems
     * for new client and to notify account object of the client's
     * connection.
     *
     * This function is called by node manager just after the client
     * node was successfully connected to the simulation (ie. all server
     * nodes has accepted the client node).
     * Main purpose of this callback is to notify account object
     * of the client node that the client was connected.
     *
     * @param client_node_id NodeId of the client.
     * @param notify_account_object Whether account object of the client
     *      should be notified of the cliennt's connection.
     *
     * @note This function is called with @a notify_account_object parameter
     *      set to true on one server node only (the forward node of the
     *      client). On all othe server nodes this function is called
     *      with the parameter set to false.
     */     
    static void client_connected
        (
        const NodeId & client_node_id,
        bool           notify_account_object
        );
        
    /**
     * Client disconnected notification.     
     *
     * This function is called by node manager on each server node just after
     * the client node was disconnected from the simulation.
     *
     * Main purpose of this callback is to clean-up all the client related
     * data stored in various massiv subsystems and to notify account
     * object of the client node that the client was disconnected.
     *
     * @param client_node_id NodeId of the client.
     * @param notify_account_object Whether account object of the client
     *      should be notified of the client's disconnection.
     *
     * @note This function is called with @a notify_account_object parameter
     *      set to true on one server node only (the forward node of the client).
     *      On all other server nodes this function is called with the parameter
     *      set to false.
     *
     * @warning Do not assume that this callback is properly paired
     *      with client_connected() callback. This callback can be called
     *      twice or more times on one server node withou corresponding
     *      client_connected() call.
     *      On the other hand if client_connected() is called there
     *      will be at least one corresponding client_disconnected() call.
     */
    static void client_disconnected
        (
        const NodeId & client_node_id,
        bool           notify_account_object
        );
    
    /**
     * Node disconnected notification.
     *
     * This function is called by node manager when a connection
     * to @a node_id is lost or closed.
     *
     * Main purpose of this callback is to clean-up all the node related
     * data stored in various massiv subsystems like replica or data management.
     *
     * @warning The function must be able to handle situation that the
     *      node is already dicconnected because of node disconnecting
     *      protocol that can cause calling this callback two or more
     *      times.
     */
    static void node_disconnected
        (
        const NodeId &  node_id
        );
    
    /**
     * Function used by node manager notifying the core that
     * somebody is trying to send a message to node that is not
     * accessible by local node. It includes case the node is not
     * registered in node database.
     * @warning The function is called directly from message sending
     *     functions (like @c SystemMessage::send() etc).
     */
    static void node_not_accessible
        (
        const NodeId & node_id
        );
        
public:

    /**
     * Load simulation from archive file identified by @a archive_id_to_load.
     * Assumes that there is no simulation running right now.
     */
    static void load_simulation
        (
        const ArchivationId & archive_id_to_load
        );
        
public:

    /**
     * Request to shutdown local node.
     */
    static void shutdown_local_node();

    }; // GlobalCallbacks

    
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_GLOBAL_CALLBACKS_H
