/** @file logdest.h
 * Abstract log destination class; base for all destinations. 
 * Destination describes an output where log messages may be written.
 */

/*
 * begin        : Tue Nov 25 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LOGDEST_H
#define MASSIV_LOGDEST_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_LOGMESSAGE_H
#include "logger/logmessage.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup logger
 * @{
 */

/**
 * Destination of LogMessages.
 */
class MASSIV_EXPORTED LogDestination
    {
    friend class Logger;

public:

    /**
     * Destructor.
     */
    virtual ~LogDestination()
        {
        }

public:

    /**
     * Gets short name describing destination.
     * For example "console" or "#filename".
     * @return The destination name.
     */
    virtual std::string get_name() const = 0;

    /**
     * Gets short description of destination.
     * @return The destination description.
     */
    virtual std::string get_description() const = 0;

protected:

    /**
     * Matches the destination name.
     * For example, file destination may match both "#a" and "#/var/log/bzz".
     *
     * @param name  Name of message rule destination.
     * @return      true if this destination name matches the @a name parameter.
     */
    virtual bool match_name
        (
        const std::string & name
        ) = 0;

    /**
     * Logs message (writes it into this destination).
     * @param name Name of this destination.
     * @param message The message to be logged. See LogMessage.
     */
    virtual void log
        (
        const std::string & name,
        const LogMessage &  message
        ) = 0;

    /**
     * Flushes output stream.
     */
    virtual void flush() = 0;

    /**
     * Flushes output stream hard.
     * While @c flush() is called when core is in a consistent state,
     * @c flush_hard() may be called any time.
     * It is used by logger when "crash-proof mode" was set on so that
     * any message would be externalized properly.
     *
     * @see flush
     */
    virtual void flush_hard() = 0;

    }; // class LogDestination

/** @} */ // addtogroup logger


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_LOGDEST_H
