/** @file logdest_clients.cpp
 * Logging to subscribed client accounts.
 * @see logdest_clients.h
 */

/*
 * begin        : Sat Aug 9 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "logger/logdest_clients.h"
#include "object/lib/account_object_interface_rpc.h"


namespace Massiv {
namespace Core {
    
namespace {

/**
 * Converts message to 7-bit ASCII.
 * This is to prevent serialization errors when message is sent via
 * RPC to subscribed clients and message contains localized strings.
 */
std::string to_ascii_7_bit
    (
    const std::string & message
    )
    {
    std::string s;
    std::size_t i;

    for( i = 0; i < message.size(); i++ )
        {
        unsigned char c = message[ i ];

        if( c > 0 && c < 128 )
            {
            s += c;
            }
        else
            {
            s = '?';
            }
        }

    return s;
    }

}
    

LogDestinationClients::LogDestinationClients() :
    distributing_now( false )
    {
    }


LogDestinationClients::~LogDestinationClients()
    {
    }
    

void LogDestinationClients::subscribe
    (
    const WeakPointer<Lib::AccountObjectInterface> & account
    )
    {
    subscribers.insert( account.get_object_id() );
    }


void LogDestinationClients::unsubscribe
    (
    const WeakPointer<Lib::AccountObjectInterface> & account
    )
    {
    Subscribers::iterator it = subscribers.find( account.get_object_id() );

    if( it != subscribers.end() )
        {
        subscribers.erase( it );
        }
    }


std::string LogDestinationClients::get_name() const
    {
    return "clients";
    }


std::string LogDestinationClients::get_description() const
    {
    return "subscribed clients";
    }


bool LogDestinationClients::match_name
    (
    const std::string & name
    )
    {
    return name == "clients";
    }


void LogDestinationClients::log
    (
    const std::string & ,
    const LogMessage &  message
    )
    {
    // Can log?

    if( System::get_phase() != System::PHASE_RUNNING )
        {
        // We are using RPC to deliver the message, thus system
        // must be properly running.

        return;
        }

    if( distributing_now )
        {
        // The message is probably related to queue distribution.
        // Discard the message to prevent an infinite loop.

        return;
        }

    // Queue message.
    // Can not be distributed immediatelly because we may be logging
    // a message from the core and core is not "interruptible"
    // (its state may be inconsistent now, RPC can not be used).

    messages.push_back( message.get_string() );
    }


void LogDestinationClients::flush()
    {
    Subscribers::iterator   it = subscribers.begin();
    PValueArray<SString>    packet;
    std::size_t             i;

    // Empty @c messages and construct a new message packet.

    for( i = 0; i < messages.size(); i++ )
        {
        packet.push_back( to_ascii_7_bit( messages[ i ] ) );
        }

    messages.clear();

    // Distribute the @c packet.

    distributing_now = true;

    try
        {
        while( it != subscribers.end() )
            {
            Remote<Lib::AccountObjectInterface> account_interface = *it;
            
            account_interface->async_log_system_messages
                (
                ObjectId( System::get_local_node_id() ),
                packet
                );
            
            ++it;
            }
        }
    catch( Exception & )
        {
        }

    distributing_now = false;
    }


void LogDestinationClients::flush_hard()
    {
    // Messages can not be distributed here too because this may be called
    // by @c Logger when core is not in a consistent state.
    }


} // namespace Massiv::Core
} // namespace Massiv
