/** @file logdest_clients.h
 * Logging to subscribed client accounts using RPC.
 */

/*
 * begin        : Sat Aug 9 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LOGDEST_CLIENTS_H
#define MASSIV_LOGDEST_CLIENTS_H


#ifndef MASSIV_LOGGER_H
#include "logger/logger.h"
#endif

#ifndef MASSIV_LOGDEST_H
#include "logger/logdest.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#include <vector>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup logger
 * @{
 */
    
namespace Lib {

class AccountObjectInterface;

}


/**
 * Logging to subscribed client accounts.
 * Registers list of subscribed AccountObjects the messages should be
 * delivered to.
 * Destination name is "clients".
 */
class MASSIV_EXPORTED LogDestinationClients : public LogDestination
    {
public:

    /**
     * Constructor.
     */
    LogDestinationClients();

    /**
     * Destructor.
     */
    ~LogDestinationClients();

public:

    /**
     * Subscribes account. Each account is generally bound with some client;
     * subscription of client's account is the same as subscription of 
     * client for receiving messages.
     *
     * @param account Weak pointer to the account object the messages
     *      should be sent to.
     * @note See the Lib documentation for information about account object.
     */
    void subscribe
        (
        const WeakPointer<Lib::AccountObjectInterface> & account
        );
    
    /**
     * Unsubscribes account.
     * @param account The account object to unsubscribe.
     */
    void unsubscribe
        (
        const WeakPointer<Lib::AccountObjectInterface> & account
        );
    
public:

    /**
     * Gets short name describing destination. 
     * @returns <tt>"clients"</tt>.
     */
    std::string get_name() const;

    /**
     * Gets short textual description of destination.
     */
    std::string get_description() const;

protected:

    /**
     * Matches destination name. 
     * @return true if the @a name is <tt>"clients"</tt>.
     */
    virtual bool match_name
        (
        const std::string & name
        );

    /**
     * Logs message (distribute it via RPC to the subscribed accounts).
     * @param name This parameter is ignored.
     * @param message The message to distribute.
     * @note The message will be actualy distributes as soon as
     *      flush() is called.
     */
    virtual void log
        (
        const std::string & name,
        const LogMessage &  message
        );
    
    /**
     * Flushes output.
     * Distributes all queued messages.
     * @note RPC facilities must be ready to be used now.
     */
    virtual void flush();
    
    /**
     * Flushes output hard.
     * Does nothing because this may be called when core is in an inconsistent
     * state (RPC can not be utilized). 
     */
    virtual void flush_hard();
    
protected:
    
    typedef ext::hash_set
        <
        ObjectId,
        ObjectId::Hash,
        ObjectId::EqualKey
        >   Subscribers;

    typedef std::vector<std::string> Messages;

    Subscribers subscribers;
        /**< List of subscribed accounts. */
    Messages    messages;
        /**< Queued messages. */
    bool        distributing_now;
        /**< Messages are being distributed now. */
        
    }; // class LogDestinationClients

/** @} */ // addtogroup logger


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_LOGDEST_CLIENTS_H
