/** @file logdest_file.cpp
 * File output log destination.
 * @see logdest_file.h
 */

/*
 * begin        : Thu Mar 06 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "data/path_manager.h"
#include "logger/logdest_file.h"
#include "global/global.h"
#include "filesys/file_system.h"
#include "utility/debug.h"
#include <sstream>
#include <ctime>

namespace Massiv {
namespace Core {

namespace {

std::string create_log_file
    (
    const std::string & relative_path
    )
    {
    const std::string & log_directory = Global::path_manager().get_log_directory();
    std::string         path = log_directory + "/" + relative_path;

    Global::native_file_system().create_file( path );

    return path;
    }

}
    

LogDestinationFile::LogDestinationFile
    (
    const std::string & path
    ) : ofs( create_log_file( path ).c_str(), std::ios::out | std::ios::trunc ),
        name( "#" + path )
    {
    if( !ofs.good() )
        {
        std::ostringstream ss;
        ss << "Cannot open logger destination '" << path << "'";
        throw ExCanNotOpenDestination( ss.str() );
        }
    }


LogDestinationFile::~LogDestinationFile()
    {
    LogDestinationFile::flush_hard();
    }


std::string LogDestinationFile::get_name() const
    {
    return name;
    }


std::string LogDestinationFile::get_description() const
    {
    hope( name.size() > 0 );  // name should contain at least '#' sign

    std::ostringstream ss;
    ss << "File output (" << name.substr( 1 ) << ")";
    return ss.str();
    }


bool LogDestinationFile::match_name
    (
    const std::string & name
    )
    {
    return name == this->name;
    // TODO: match full description
    }


void LogDestinationFile::log
    (
    const std::string & ,
    const LogMessage &  message
    )
    {
    static std::string  format = "D.M.Y h:m:s";
    ofs << convert_time_to_string( format ) << message.get_string() << '\n';
    }


void LogDestinationFile::flush()
    {
    SystemTime      current_time = SystemTime::current_time();
    const Float32   flush_period = 5.0f;
    
    if( flush_time + flush_period > current_time )
        {
        // Do nothing.
        // This is to prevent permanent disk writes.
        // If persistent log is requested enable "crash-proof logging".
        
        return;
        }

    flush_time = current_time;
    flush_hard();
    }


void LogDestinationFile::flush_hard()
    {
    ofs.flush();
    }


std::string LogDestinationFile::convert_time_to_string
    (
    const std::string &     format
    ) const
    {
    time_t      t;
    std::tm *   tt;
    
    t = std::time( NULL );
    tt = std::localtime( &t );

    std::string::const_iterator     it = format.begin();
    std::ostringstream              ss;

    while( it != format.end() )
        {
        switch( *it )
            {
            case 'Y':
                ss << tt->tm_year + 1900;
                break;
            case 'M':
                ss << ( tt->tm_mon + 1 );
                break;
            case 'D':
                ss << tt->tm_mday;
                break;
            case 'h':
                ss << tt->tm_hour;
                break;
            case 'm':
                if( tt->tm_min < 10 )
                    {
                    ss << "0";
                    }
                ss << tt->tm_min;
                break;
            case 's':
                if( tt->tm_sec < 10 )
                    {
                    ss << "0";
                    }
                ss << tt->tm_sec;
                break;
            case ':':
            case ' ':
            case '.':
                ss << *it;
                break;
            default:
                panic( "Unknown time format." );
            }            
        ++it;
        }
    
    return ss.str();
    }


} // namespace Massiv::Core
} // namespace Massiv
