/** @file logdest_file.h
 * File output log destination.
 */

/*
 * begin        : Thu Mar 06 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LOGDEST_FILE_H
#define MASSIV_LOGDEST_FILE_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_LOGDEST_H
#include "logger/logdest.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#include <fstream>

namespace Massiv {
namespace Core {


/** @addtogroup logger
 * @{
 */

/**
 * File output destination. 
 *
 * Destination name is <tt>#</tt> concatenated with the
 * path to the file given in the constructor. For example: 
 * <tt>#blabla/logs/log</tt>. Note that the path must be
 * relative, because it is merged with a path specifying
 * the <em>logs root directory</em> (this path can be set
 * in the configuration files and is can be obtained from
 * the PathManager in the runtime.
 * 
 * @exception ExCannotOpenDestination The destination file could not be opened.
 */
class MASSIV_EXPORTED LogDestinationFile : public LogDestination
    {
public:

    /**
     * Constructor.
     * @param path Path to the file.
     */
    LogDestinationFile
        (
        const std::string &     path
        );

    /**
     * Destructor.
     */
    ~LogDestinationFile();

    NO_DEFAULT_CONSTRUCTOR( LogDestinationFile );

public:

    /**
     * Gets name of the destination.
     *
     * @return Name of the destination.
     */
    std::string get_name() const;

    /**
     * Gets short description of the destination.
     */
    std::string get_description() const;

protected:

    /**
     * Matches destination name. 
     *
     * @return true if the @a name parameter matches this destination name.
     */
    virtual bool match_name
        (
        const std::string & name
        );

    /**
     * Logs message (writes it into the file bound with the destination).
     *
     * @param name Ignored.
     * @param message The message to be logged into the destination.
     */
    virtual void log
        (
        const std::string & name,
        const LogMessage &  message
        );
   
    /**
     * Flushes output.
     */
    virtual void flush();
    
    /**
     * Flushes output hard.
     */
    virtual void flush_hard();
    
    /**
     * Converts actual system time to string using defined format of output 
     * string.
     *
     * Following formating can be used:
     *
     * @c Y - year, @c M - month, @c D - day of month, @c h - hour, 
     * @c m - minute, @c s - second. ':', ' ', '.' can be used as a separators.
     *
     * Example of the format string: <tt>"h:m:s D.M.Y"</tt>
     */
    std::string convert_time_to_string
        (
        const std::string & format
        ) const;
        
protected:

    std::ofstream       ofs;
        /**< Output stream used for logging. */

    std::string         name;
        /**< Name of this destination. */

    SystemTime          flush_time;
        /**< When the stream was flushed for the last time. */

    }; // class LogDestinationFile

/** @} */ // addtogroup logger


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_LOGDEST_FILE_H
