/** @file logger.cpp
 * Logger enables logging messages to global logs.
 * @see logger.h
 */

/*
 * begin        : Tue Nov 22 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/global.h"
#include "logger/logger.h"
#include "logger/logdest_clients.h"
#include "logger/logdest_cerr.h"
#include "logger/logdest_cout.h"
#include "logger/logdest_file.h"
#include "registry/registry.h"
#include <algorithm>
#include <iosfwd>
#include <sstream>

namespace Massiv {
namespace Core {


// Logger::LoggerSettings implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Logger::LoggerSettings::LoggerSettings() : 
    Settings( "Logger" ),
    enabled( access_boolean( "enabled", true ) ),
    crash_proof_logging( access_boolean( "crash_proof_logging", true ) ),
    collapse_messages( access_boolean( "collapse_messages", true ) )
    {
    }


// Logger::SelectorElement implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

bool Logger::SelectorElement::valid() const
    {
    if( facility < Status::FACILITY_ANY || facility >= Status::FACILITY_COUNT ||
        type < Status::TYPE_ANY || type >= Status::TYPE_COUNT ||
        priority < Status::PRIORITY_LOWEST ||  priority > Status::PRIORITY_HIGHEST ||
        ( type != Status::TYPE_ANY && !( flags & ( TYPE_LEQ | TYPE_GEQ ) ) ) ||
        !( flags & ( PRIORITY_LEQ | PRIORITY_GEQ ) ) )
        {
        return false;
        }
    return true;
    }


bool Logger::SelectorElement::match_status
    (
    Status status
    ) const
    {
    bool result;

    if( ( facility == Status::FACILITY_ANY || facility == status.value.c.facility ) && // facility ok, and
        ( type == Status::TYPE_ANY ||                                                  // ( any status, or
          ( ( !( flags & TYPE_LEQ ) || status.value.c.type <= type ) &&                //   ( <=status ok, and
            ( !( flags & TYPE_GEQ ) || status.value.c.type >= type )                   //     >=status ok )
          )                                                                            // )
        ) &&                                                                           // and
        ( !( flags & PRIORITY_LEQ ) || status.value.c.priority <= priority ) &&        // <=priority ok, and
        ( !( flags & PRIORITY_GEQ ) || status.value.c.priority >= priority )           // >=priority ok
      )
        {
        result = true;
        }
    else
        {
        result = false;
        }

    return( flags & NEGATE ) ? !result : result;
    }


bool Logger::SelectorElement::operator<
    (
    const SelectorElement &     se2
    ) const
    {
    return
        facility < se2.facility ||
          (
          facility == se2.facility &&
            (
            type > se2.type ||
              (
              type == se2.type &&
                (
                priority > se2.priority ||
                  (
                  priority == se2.priority && flags < se2.flags
                  )
                )
              )
            )
          );
    }


bool Logger::SelectorElement::Less::operator()
    (
    const SelectorElement &     se1,
    const SelectorElement &     se2
    ) const
    {
    return se1 < se2;
    }


// Logger::Selector implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

bool Logger::Selector::valid() const
    {
    if( empty() )
        return false;
    for( const_iterator sit = begin(); sit != end(); ++sit )
        {
        if( !sit->valid() )
            return false;
        }
    return true;
    }


bool Logger::Selector::match_status
    (
    Status status
    ) const
    {
    for( Selector::const_iterator iter = begin(); iter != end(); ++iter )
        {
        if( !iter->match_status( status ) )
            {
            return false;
            }
        }

    return true;
    }


// Logger::RuleIterator implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

std::string Logger::RuleIterator::to_string() const
    {
    std::ostringstream  s;
    bool                first = true;

    for( Logger::Selector::const_iterator s_iter = iterator->first.begin();
         s_iter != iterator->first.end(); ++s_iter )
        {
        const Logger::SelectorElement & se = *s_iter;

        if( first )
            first = false;
        else
            s << ", ";

        if( se.flags & Logger::SelectorElement::NEGATE )
            s << "!";

        if( se.facility == Status::FACILITY_ANY )
            s << "*";
        else
            s << Status::facility_string( se.facility );

        s << ".";
        if( se.type == Status::TYPE_ANY )
            {
            s << "*";
            }
        else
            {
            if( se.flags & Logger::SelectorElement::TYPE_LEQ )
                {
                if( se.flags & Logger::SelectorElement::TYPE_GEQ )
                    s << "=";
                else
                    s << "<=";
                }
            else
                {
                hope( se.flags & Logger::SelectorElement::TYPE_GEQ );
                s << ">=";
                }
            s << Status::type_string( se.type );
            }

        s << ".";
        if( se.flags & Logger::SelectorElement::PRIORITY_LEQ )
            {
            if( se.flags & Logger::SelectorElement::PRIORITY_GEQ )
                s << "=";
            else
                s << "<=";
            }
        else
            {
            hope( se.flags & Logger::SelectorElement::PRIORITY_GEQ );
            s << ">=";
            }
        s << Status::priority_string( se.priority );
        }

    s << " -> ";
    first = true;
    for( Logger::RuleDestinations::const_iterator rd_iter = iterator->second.begin();
         rd_iter != iterator->second.end(); ++rd_iter )
        {
        if( first )
            first = false;
        else
            s << ", ";
        s << (*rd_iter)->first;
        }

    return s.str();
    }


// Logger implementation
// ~~~~~~~~~~~~~~~~~~~~~

Logger::Logger() :
    message_match_count( 0 ),
    loading_configuration( false )
    {
    LogDestinationCout *    dest = new LogDestinationCout();
    Logger::Selector        selector;
    Logger::RuleIterator    rule_it;

    // Register "cout" destination.
    register_destination( dest );

    // Create logger rule.
    selector.insert(
        Logger::SelectorElement
            (
            Logger::SelectorElement::PRIORITY_LEQ,
            Status::FACILITY_ANY,
            Status::TYPE_ANY,
            Status::PRIORITY_HIGHEST
            )
        );

    // Register the rule.
    rule_it = insert_rule( selector );
    hope( rule_it != end_rule() );

    // Send messages that pass the rule filter to "cout".
    insert_destination( rule_it, "cout" );
    }


Logger::~Logger()
    {
    unregister_all_destinations();
    }


void Logger::initialize_from_registry()
    {
    Registry & registry = Global::registry();

    // Reset to default state.

    unregister_all_destinations();
    clear_rules();

    // Does registry contain logger settings?

    try
        {
        registry.find_node_for_path( "Settings/Logger" );
        }
    catch( ExRegistryNodeNotFound & )
        {
        // Logger settings not present in the registry - do not touch logger.
        return;
        }

    // Load logger destinations & rules.
    // Attention! log() could be called while loading logger state.
    // Turn off logging.

    loading_configuration = true;

    try
        {
        register_destinations_from_registry( registry );
        
        SelectorElementsMap  selector_elements_map =
            load_selector_elements_from_registry( registry );

        load_selectors_from_registry( registry, selector_elements_map );
        }
    catch( Exception & e )
        {
        unregister_all_destinations();
        clear_rules();
        message_match_count = 0;
        loading_configuration = false;

        // Raise caught exception as a new exception (do not rethrow it).
        // This ensures that it will be eventually logged now.

        throw ExLoggerError
            (
            std::string( "An error occured while initializing logger: " ) + 
            e.what()
            );
        }

    loading_configuration = false;

    // Update state.

    message_match_count = 0;
    }


void Logger::flush()
    {
    RegisteredDestinations::const_iterator it = registered_destinations.begin();

    while( it != registered_destinations.end() )
        {
        ( *it )->flush();
        ++it;
        }
    }


bool Logger::log
    (
    const LogMessage & message
    ) const
    {
    static bool inside_logger = false;
    bool        logged = false;
    
    // Can log?
    if( !settings.enabled )
        {
        return false;
        }

    if( inside_logger )
        {
        return false;
        }

    if( loading_configuration )
        {
        return false;
        }

    inside_logger = true;

    // Log.
    try
        {
        // Collapse messages.
        if( settings.collapse_messages )
            {
            // Are we logging the same message once again? */
            if( message == previous_message )
                {
                // Remember the message but do not log now.
                message_match_count++;
                inside_logger = false;

                return false;
                }
            else
                {
                // Write how many times was the last message repeated.
                if( message_match_count )
                    {
                    std::ostringstream ss;
                    
                    ss  << "The previous message was repeated " << message_match_count + 1 
                        << " times.";
                    
                    message_match_count = 0;
                    inside_logger = false;

                    return log( LogMessage( previous_message.status, ss.str() ) );
                    }
                }
            }

        // Foreach rule if match call.
        for( Rules::const_iterator r_iter = rules.begin(); r_iter != rules.end(); ++r_iter )
            {
            if( r_iter->first.match_status( message.status ) )
                {
                logged = true;
                for( RuleDestinations::const_iterator rd_iter = r_iter->second.begin();
                     rd_iter != r_iter->second.end(); ++rd_iter )
                    {
                    if( ( *rd_iter )->second )
                        {
                        // Log.
                        ( *rd_iter )->second->log( ( *rd_iter )->first, message );

                        // Flush.
                        if( settings.crash_proof_logging )
                            {
                            ( *rd_iter )->second->flush_hard();
                            }
                        }
                    }
                }
            }
        }
    catch( Exception & )
        {
        inside_logger = false;

        throw;
        }

    // Done.
    previous_message = message;
    inside_logger = false;

    return logged;
    }


bool Logger::register_destination
    (
    LogDestination * destination
    )
    {
    hope( destination );

    // Check whether a destination of the same name isn't already registered.
    if( find_destination( destination->get_name() ) )
        {
        return false;
        }

    // Add to registered_destinations.
    std::pair< RegisteredDestinations::iterator, bool > result;
    result = registered_destinations.insert( destination );

    // Update destination_mapping (if there is a NULL pointer in 
    // destination_mapping for name of the destination being registered).
    for( DestinationMapping::iterator dm_iter = destination_mapping.begin();
         dm_iter != destination_mapping.end(); ++dm_iter )
        {
        if( !dm_iter->second && destination->match_name( dm_iter->first ) )
            {
            dm_iter->second = destination;
            }
        }

    return result.second;
    }


void Logger::unregister_destination
    (
    const std::string & name
    )
    {
    // Find destination in DestinationMapping table.
    DestinationMapping::iterator    dm_iter = destination_mapping.find( name );
    LogDestination *                log_destination = NULL;

    if( dm_iter != destination_mapping.end() )
        {
        log_destination = dm_iter->second;
        }

    // Delete the destination if found.
    if( log_destination )
        {
        hope( log_destination->match_name( dm_iter->first ) );
        dm_iter->second = NULL;
        registered_destinations.erase( log_destination );
        delete log_destination;
        }
    }


void Logger::enable
    (
    bool enabled
    )
    {
    settings.enabled = enabled;
    }


bool Logger::is_enabled()
    {
    return settings.enabled;
    }


LogDestination * Logger::find_destination
    (
    const std::string & name
    )
    {
    RegisteredDestinations::const_iterator it = registered_destinations.begin();

    while( it != registered_destinations.end() )
        {
        LogDestination * dest = *it;

        if( dest->match_name( name ) )
            {
            return dest;
            }

        ++it;
        }
    
    return NULL;
    }


void Logger::unregister_all_destinations()
    {
    RegisteredDestinations::iterator    it_rdest
        ( 
        registered_destinations.begin() 
        );

    while( it_rdest != registered_destinations.end() )
        {
        if( *it_rdest )
            {
            delete *it_rdest;
            }

        it_rdest++;
        }

    registered_destinations.erase
        ( 
        registered_destinations.begin(), 
        registered_destinations.end() 
        );

    DestinationMapping::iterator        it_dest
        (
        destination_mapping.begin()
        );

    while( it_dest != destination_mapping.end() )
        {
        it_dest->second = NULL;
        it_dest++;
        }
    }


Logger::RuleIterator Logger::insert_rule
    (
    const Selector &    selector
    )
    {
    // Check selector.
    if( !selector.valid() )
        {
        return end_rule();
        }

    // Add to rules.
    std::pair<Rules::iterator, bool> rules_pair =
        rules.insert( Rules::value_type( selector, RuleDestinations() ) );
    return rules_pair.first;
    }


Logger::RuleIterator Logger::find_rule
    (
    const Selector &    selector
    )
    {
    return rules.find( selector );
    }


void Logger::erase_rule
    (
    RuleIterator        rule
    )
    {
    rules.erase( rule.iterator );
    }


void Logger::erase_rule
    (
    const Selector &    selector
    )
    {
    rules.erase( selector );
    }


void Logger::clear_rules()
    {
    rules.clear();
    destination_mapping.clear();
    }


Logger::RuleIterator Logger::begin_rule()
    {
    return rules.begin();
    }


Logger::RuleIterator Logger::end_rule()
    {
    return rules.end();
    }


void Logger::insert_destination
    (
    RuleIterator            rule,
    const std::string &     destination
    )
    {
    // Add destination to destination_mapping.
    std::pair<DestinationMapping::iterator, bool> dm_pair =
        destination_mapping.insert( DestinationMapping::value_type( destination, NULL ) );

    if( dm_pair.second )
        {
        // New entry inserted - try to match to registered destination.
        hope( !dm_pair.first->second );
        for( RegisteredDestinations::const_iterator rd_iter = registered_destinations.begin();
             rd_iter != registered_destinations.end(); ++rd_iter )
            {
            hope( *rd_iter );
            if( (*rd_iter)->match_name( dm_pair.first->first ) )
                {
                dm_pair.first->second = *rd_iter;
                }
            }
        }

    // Add destination to rule.
    rule.iterator->second.insert( dm_pair.first );
    }


void Logger::erase_destination
    (
    RuleIterator            rule,
    const std::string &     destination
    )
    {
    hope( rule.iterator != rules.end() );

    for( RuleDestinations::iterator rud_iter = rule.iterator->second.begin();
         rud_iter != rule.iterator->second.end(); ++rud_iter )
        {
        if( (*rud_iter)->first == destination )
            {
            rule.iterator->second.erase( rud_iter );
            break; // Only one destination with this name.
            }
        }

    // FIXME: remove from destination_mapping if no longer referenced
    }


void Logger::unregister_destination
    (
    LogDestination * destination
    )
    {
    hope( destination );

    // Update destination_mapping (don't delete the whole record for the
    // destination - keep its name and set NULL pointer instead).
    for( DestinationMapping::iterator dm_iter = destination_mapping.begin();
         dm_iter != destination_mapping.end(); ++dm_iter )
        {
        if( dm_iter->second == destination )
            {
            hope( destination->match_name( dm_iter->first ) );
            dm_iter->second = NULL;
            }
        }

    // Remove from registered destinations.
    registered_destinations.erase( destination );
    
    delete destination;
    }


void Logger::register_destinations_from_registry
    (
    const Registry &    registry
    )
    {
    Registry::const_iterator        it;

    try
        {
        it = registry.begin( "Settings/Logger/Destinations", "*" );
        }
    catch( ExRegistryNodeEmpty & )
        {
        // No destinations defined in registry.
        return;
        }

    while( it != registry.end() )
        {
        LogDestination *    logdest = NULL;
        std::string         name( it->read_string() );

        if( name == "cerr" )
            {
            logdest = new LogDestinationCerr;
            }
        else
        if( name == "cout" )
            {
            logdest = new LogDestinationCout;
            }
        else
        if( name == "clients" )
            {
            logdest = new LogDestinationClients;
            }
        else        
        if( name[ 0 ] == '#' )
            {
            logdest = new LogDestinationFile( name.substr( 1 ) );
            }
        else
            {
            std::ostringstream  ss;
            ss << "LogDestination name '" << name << "' is not valid.";
            throw ExInvalidDestinationName( ss.str() );
            }

        if( !register_destination( logdest ) )
            {
            // Destination already existed.
            delete logdest;
            }

        it++;
        }
    }


Logger::SelectorElementsMap Logger::load_selector_elements_from_registry
    (
    const Registry &    registry
    )
    {
    SelectorElementsMap                 selector_elements_map;
    Registry::const_subnode_iterator    sub_it;

    try
        {
        sub_it = registry.begin_subnodes( "Settings/Logger/SelectorElements" );
        }
    catch( ExRegistryNodeEmpty & )
        {
        // No SelectorElements defined in registry.
        return selector_elements_map;
        }

    while( sub_it != registry.end_subnodes() )
        {
        selector_elements_map[ *sub_it ] =
            load_selector_element_from_registry( registry, *sub_it );

        sub_it++;
        }

    return selector_elements_map;
    }


Logger::SelectorElement Logger::load_selector_element_from_registry
    (
    const Registry &    registry,
    const std::string & name
    )
    {
    Registry::const_iterator        it;

    try
        {
        it = registry.begin( "Settings/Logger/SelectorElements/" + name );
        }
    catch( ExRegistryNodeEmpty & )
        {
        std::ostringstream  ss;
        ss << "Selector element '" << name << "' defined without any data.";
        ExRegistryNodeEmpty( ss.str() );
        }

    SelectorElement     selector_element
        (   // default settings
        SelectorElement::PRIORITY_LEQ,
        Status::FACILITY_ANY,
        Status::TYPE_ANY,
        Status::PRIORITY_HIGHEST
        );


    while( it != registry.end() )
        {
        if( it->get_name() == "priority" )
            {
            selector_element.flags &= ~SelectorElement::PRIORITY_LEQ;
                // Erase default settings which could colide with the new one.

            process_priority_string( it->read_string(), selector_element );
            }
        else
        if( it->get_name() == "type" )
            {
            process_type_string( it->read_string(), selector_element );
            }
        else
        if( it->get_name() == "facility" )
            {
            process_facility_string( it->read_string(), selector_element );
            }
        else
        if( it->get_name() == "negate" )
            {
            if( it->read_boolean() )
                {
                selector_element.flags |= SelectorElement::NEGATE;
                }
            else
                {
                selector_element.flags &= ~SelectorElement::NEGATE;
                }
            }
        else
            {
            std::ostringstream   ss;
            ss << "Unexpected logger selector element setting ("
               << it->get_name() << ")";
            throw ExUnexpectedLoggerSetting( ss.str() );
            }

        it++;
        }

    return selector_element;
    }


void Logger::load_selectors_from_registry
    (
    const Registry &                registry,
    const SelectorElementsMap &     selector_elements_map
    )
    {
    Registry::const_subnode_iterator    sub_it;

    try
        {
        sub_it = registry.begin_subnodes( "Settings/Logger/Selectors" );
        }
    catch( ExRegistryNodeEmpty & )
        {
        // No selectors defined in registry.
        return;
        }

    while( sub_it != registry.end_subnodes() )
        {
        load_selector_from_registry
            ( 
            registry, 
            *sub_it, 
            selector_elements_map 
            );

        sub_it++;
        }
    }


void Logger::load_selector_from_registry
    (
    const Registry &                registry,
    const std::string &             name,
    const SelectorElementsMap &     selector_elements_map
    )
    {
    Selector                        selector;

    // Insert a set of selector elements into the selector.
    Registry::const_iterator        it;

    try
        {
        it = registry.begin( "Settings/Logger/Selectors/" + name + "/Elements" );
        }
    catch( ExRegistryNodeEmpty & )
        {
        // Attempt to define empty selector.
        std::ostringstream  ss;
        ss << "Selector '" << name << "' defined empty.";
        throw ExEmptySelector( ss.str() );
        }

    while( it != registry.end() )
        {
        SelectorElementsMap::const_iterator     sem_it
            (
            selector_elements_map.find( it->read_string() )
            );
     
        if( sem_it != selector_elements_map.end() )
            {
            selector.insert( sem_it->second );
            }

        it++;
        }

    // Register the selector with the logger.
    RuleIterator    rule_it( insert_rule( selector ) );

    if( rule_it == end_rule() )
        {
        std::ostringstream  ss;
        ss << "Logger rule registration failed (selector " << name << ")";
        throw ExRuleRegistrationFailed( ss.str() );
        }

    // Insert all destinations for this rule.
    try
        {
        it = registry.begin( "Settings/Logger/Selectors/" + name + "/Destinations" );
        }
    catch( ExRegistryNodeEmpty & )
        {
        // No destinations defined for selector.
        return;
        }

    while( it != registry.end() )
        {
        insert_destination( rule_it, it->read_string() );
        it++;
        }
    }


namespace {

enum EComparisons
    {
    LEQ = 1,
    GEQ = 2,
    EQ  = 4,
    LT  = 8,
    GT  = 16,
    ERR = 128
    };


/**
 * Parses a string of type <tt>"<= warning"</tt>. 
 *
 * In this case, it returns 
 * <tt>EComparisons::LEQ</tt> and sets <tt>"warning"</tt> into @c keyword.
 * If the string is wrong, <tt>EComparisons::ERR</tt> is returned.
 */
EComparisons parse_settings_string
    (
    const std::string & sstring,
    std::string &       keyword
    )
    {
    EComparisons        result = ERR;

    std::string::size_type  x;
    std::string::size_type  y;

    x = sstring.find_first_not_of( " \t" );

    if( ( x == std::string::npos ) || ( x + 1 == std::string::npos ) )
        {
        return ERR;
        }

    y = sstring.find_first_not_of( "<>=", x );

    if( ( y == std::string::npos ) || ( y + 1 == std::string::npos ) )
        {
        return ERR;
        }

    std::string         comp( sstring.substr( x, y - x ) );

    if( comp == "<=" )  result = LEQ;
    else
    if( comp == ">=" )  result = GEQ;
    else
    if( comp == "<" )   result = LT;
    else
    if( comp == ">" )   result = GT;
    else
    if( comp == "=" )   result = EQ;

    x = sstring.find_first_not_of( " \t", y );

    if( ( x == std::string::npos ) || ( x + 1 == std::string::npos ) )
        {
        return ERR;
        }

    y = sstring.find_first_of( " \t", x );

    if( y == std::string::npos )
        {
        y = sstring.size();
        }

    keyword = sstring.substr( x, y - x );

    return result;
    }

} // anonymous namespace


void Logger::process_priority_string
    (
    const std::string & priority_string,
    SelectorElement &   selector_element
    )
    {
    std::string         keyword;
    bool                decrement_keyword = false;
    bool                increment_keyword = false;

    switch( parse_settings_string( priority_string, keyword ) )
        {
    case LT:
        decrement_keyword = true;
    case LEQ:
        selector_element.flags |= SelectorElement::PRIORITY_LEQ;
        break;

    case GT:
        increment_keyword = true;
    case GEQ:
        selector_element.flags |= SelectorElement::PRIORITY_GEQ;
        break;

    case EQ:
        selector_element.flags |= SelectorElement::PRIORITY_EQ;
        break;

    case ERR:
    default:
        throw ExInvalidPriorityString
            ( 
            "Invalid priority string in logger settings!" 
            );
        break;
        }

    selector_element.priority = Status::priority_from_string( keyword );
    if( selector_element.priority == Status::PRIORITY_COUNT )
        {
        std::ostringstream ss;
        ss << "Invalid priority value in registry logger settings (" 
           << keyword << ")";
        throw ExInvalidPriorityString( ss.str() );
        }


    hope( !increment_keyword || !decrement_keyword );

    if( decrement_keyword )
        {
        if( selector_element.priority == 0 )
            {
            std::ostringstream  ss;
            ss << "Invalid comparison (too low value) in registry logger settings ("
               << priority_string << ")";
            throw ExInvalidPriorityString( ss.str() );
            }

        int new_priority = ( int ) selector_element.priority;
        selector_element.priority = ( Status::Priority ) --new_priority;
        }

    if( increment_keyword )
        {
        if( selector_element.priority == Status::PRIORITY_COUNT )
            {
            std::ostringstream  ss;
            ss << "Invalid comparison (too high value) in registry logger settings ("
               << priority_string << ")";
            throw ExInvalidPriorityString( ss.str() );
            }

        int new_priority = ( int ) selector_element.priority;
        selector_element.priority = ( Status::Priority ) ++new_priority;
        }
    }  // Logger::process_priority_string()


void Logger::process_type_string
    (
    const std::string & type_string,
    SelectorElement &   selector_element
    )
    {
    std::string         keyword;
    bool                decrement_keyword = false;
    bool                increment_keyword = false;

    switch( parse_settings_string( type_string, keyword ) )
        {
    case LT:
        decrement_keyword = true;
    case LEQ:
        selector_element.flags |= SelectorElement::TYPE_LEQ;
        break;

    case GT:
        increment_keyword = true;
    case GEQ:
        selector_element.flags |= SelectorElement::TYPE_GEQ;
        break;

    case EQ:
        selector_element.flags |= SelectorElement::TYPE_EQ;
        break;

    case ERR:
    default:
        std::ostringstream  ss;
        ss << "Invalid type string in logger settings (" << type_string << ")";
        throw ExInvalidPriorityString( ss.str() );
        break;
        }

    
    selector_element.type = Status::type_from_string( keyword );
    if( selector_element.type == Status::TYPE_COUNT )
        {
        std::ostringstream ss;
        ss << "Invalid type value in registry logger settings (" 
           << keyword << ")";
        throw ExInvalidTypeString( ss.str() );
        }

    hope( !increment_keyword || !decrement_keyword );

    if( decrement_keyword )
        {
        if( selector_element.type == 0 )
            {
            std::ostringstream  ss;
            ss << "Invalid comparison (too low value) in registry logger settings ("
               << type_string << ")";
            throw ExInvalidTypeString( ss.str() );
            }

        int new_type = ( int ) selector_element.type;
        selector_element.type = ( Status::Type ) --new_type;
        }

    if( increment_keyword )
        {
        if( selector_element.type == Status::TYPE_COUNT )
            {
            std::ostringstream  ss;
            ss << "Invalid comparison (too high value) in registry logger settings ("
               << type_string << ")";
            throw ExInvalidTypeString( ss.str() );
            }

        int new_type = ( int ) selector_element.type;
        selector_element.type = ( Status::Type ) ++new_type;
        }
    }


void Logger::process_facility_string
    (
    const std::string & facility_string,
    SelectorElement &   selector_element
    )
    {
    selector_element.facility = 
        Status::facility_from_string( facility_string );

    if( selector_element.facility == Status::FACILITY_COUNT )
        {
        std::ostringstream  ss;
        ss << "Invalid facility (" << facility_string 
           << ") in registry logger settings.";
        
        throw ExInvalidFacilityString( ss.str() );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
