/** @file logger.h
 * Logger enables logging messages to %global logs.
 *
 * Each log is specified by some of the LogDestination descendants.
 * The Logger has a set of rules that help it decide to which %logger
 * destinations it should write each specific %log %message.
 */

/*
 * begin        : Tue Nov 22 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LOGGER_H
#define MASSIV_LOGGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATUS_H
#include "status/status.h"
#endif

#ifndef MASSIV_LOGMESSAGE_H
#include "logger/logmessage.h"
#endif

#ifndef MASSIV_LOGDEST_H
#include "logger/logdest.h"
#endif

#ifndef MASSIV_REGISTRY_H
#include "registry/registry.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#include <set>
#include <map>


namespace Massiv {
namespace Core {


/** @addtogroup logger
 * @{
 */

namespace Test { class LoggerTest; }


/**
 * Global logger.
 * 
 * It filters messages according its table of rules and sends them to 
 * registered %log destinations.
 *
 * Use @a Global to get pointer to logger object.
 */
class MASSIV_EXPORTED Logger
    {
    friend class Massiv::Core::Test::LoggerTest;

public:
    
    /**
     * Logger settings.
     * @see Settings
     */
    class LoggerSettings : public Settings
        {
    public:

        /**
         * Constructor.
         */
        LoggerSettings();

    public:
            
        Boolean &   enabled;
            /**< Is logging enabled? */

        Boolean &   crash_proof_logging;
            /**< Crash-proof logging (persistent log)? */

        Boolean &   collapse_messages;
            /**< Collapse repeated messages to a single message? */
        
        }; // class LoggerSettings

    /**
     * Specifies a part of conditions a message must meet to be sent to %log 
     * %destinations bound by the specific rule that contains this %selector 
     * %element.
     *
     * Full set of conditions is specified by Selector that can contain even
     * more %selector %elements.
     */
    class SelectorElement
        {

    public:

        /**
         * %Flags specify the way %selector %element is applied in rule 
         * matching.
         */
        enum Flags
            {
            TYPE_LEQ        = 0x0001,
                /**< %Message %type is lower than or equal to a @a type 
                     specified by this %element. */
            TYPE_GEQ        = 0x0002,
                /**< %Message %type is greater than or equal to a @a type
                     specified by this %element. */
            TYPE_EQ         = ( TYPE_LEQ | TYPE_GEQ ),
                /**< %Message %type is equal to a @a type
                     specified by this %element. */
            PRIORITY_LEQ    = 0x0004,
                /**< %Message %priority is lower than or equal to a @a priority
                     specified by this %element. */
            PRIORITY_GEQ    = 0x0008,
                /**< %Message %priority is greater than or equal to a 
                     @a priority specified by this %element. */
            PRIORITY_EQ     = ( PRIORITY_LEQ | PRIORITY_GEQ ),
                /**< %Message %priority is equal to a @a priority
                     specified by this %element. */
            NEGATE          = 0x8000
                /**< Negates this %selector %element. */
            };

        unsigned            flags;
            /**< Flags (see SelectorElement::Flags). */
        Status::Facility    facility;
            /**< Facility to match or Status::FACILITY_ANY, which matches any facility. */
        Status::Type        type;
            /**< Type to match or Status::TYPE_ANY. */
        Status::Priority    priority;
            /**< Priority to match. */

    public:

        /**
         * Default constructor.
         */
        SelectorElement()
            : flags( 0 ),
              facility( Status::FACILITY_ANY ),
              type( Status::TYPE_WARNING ),
              priority( Status::PRIORITY_NORMAL )
            {
            }

        /**
         * Constructor.
         * No checks are done in constructor.
         */
        SelectorElement
            (
            unsigned            flags,
            Status::Facility    facility,
            Status::Type        type,
            Status::Priority    priority
            ) : flags( flags ),
                facility( facility ),
                type( type ),
                priority( priority )
            {
            }

    public:

        /**
         * Checks if this %selector %element is valid.
         * %Selector %element may be invalid if facility/type/priority contains
         * out-of-range number, weird combination of flags, etc.
         *
         * @return true if valid.
         */
        bool valid() const;

        /**
         * Checks if message status satisfies conditions specified by 
         * this SelectorElement.
         * 
         * @return true if the conditions are met.
         */
        bool match_status
            (
            Status status
            ) const;

        /**
         * Compares SelectorElements.
         *
         * @return true if this %selector %element &lt; @a se2.
         */
        bool operator<
            (
            const SelectorElement &     se2
            ) const;

        /** @struct Less
         * %Comparator for SelectorElement instances.
         *
         * Ordering rules are as follows:
         * - lower facility first
         * - higher type first 
         * - higher priority first
         * - lower flags 'dword' value 
         * .
         * 
         * @note Sorting by flags is a bit weird, but we implement it to
         *      make two %selector %elements that differ only in %flags
         *      to yield the 'non-equal' %result.
         */
        struct Less
            {
            /** @return true if @a se1 &lt; @a se2 */
            bool operator()
                (
                const SelectorElement &     se1,
                const SelectorElement &     se2
                ) const;

            }; // struct Less

        }; // class SelectorElement

    /**
     * Set of %selector %elements. 
     * 
     * Only %messages that pass all tests specified 
     * by %selector elements will be written to the %log destinations
     * corresponding to the %rule that contains this %selector.
     *
     * @see SelectorElement
     */
    class Selector : public std::set<SelectorElement, SelectorElement::Less>
        {
    public:

        /**
         * Tests if the selector is valid.
         * 
         * Selector is invalid if it is empty or contains an invalid selector 
         * element.
         * 
         * @return true if the selector is valid, false otherwise.
         */
        bool valid() const;

        /**
         * Checks if message status satisfies conditions specified by Selector.
         *
         * @return true if the status matches the conditions, false otherwise.
         */
        bool match_status
            (
            Status status
            ) const;
        }; // class Selector
    
private:

    /** @typedef DestinationMapping
     * %Map converting %destination names to %destination pointers. 
     * LogDestination pointer may be @c NULL if no matching destination 
     * is registered.
     *
     * @todo Entries are never removed from this list. Remove them when 
     * they are no longer referenced? (use reference counting?).
     */
    typedef std::map<std::string, LogDestination *> DestinationMapping;
    
    /** @typedef SelectorElementsMap
     * %SelectorElementsMap is being used during initialization of the logger
     * from registry.
     */
    typedef std::map< std::string, SelectorElement >    SelectorElementsMap;

    /** @typedef RegisteredDestinations
     * Set of all registered destinations. Used when new destination is added 
     * to DestinationMapping.
     */
    typedef std::set<LogDestination *> RegisteredDestinations;
    
    /** @struct RuleDestinationsLess
     * Comparator for %rule %destinations.
     */
    struct RuleDestinationsLess
        {
        /** @return true if @a dmi1 &lt; @a dmi2, false otherwise. */
        bool operator()
            (
            const DestinationMapping::const_iterator & dmi1,
            const DestinationMapping::const_iterator & dmi2
            ) const
            {
            //return std::less<std::string>()( dmi1->first, dmi2->first );
            return dmi1->first < dmi2->first;
            }

        }; // struct RuleDestinationsLess
    
    /** @typedef RuleDestinations
     * Set of destinations (iterators to DestinationMapping) of a specific rule.
     */
    typedef std::set<DestinationMapping::const_iterator, RuleDestinationsLess> RuleDestinations;
    
    /** @typedef Rules
     * The rules %table.
     */
    typedef std::map<Selector, RuleDestinations> Rules;

public:

    /**
     * %Iterator for table of filtering rules that is maintaned by the Logger. 
     *
     * It is similar to standard STL iterators, but it's not compatible with 
     * them. You can use it to add more destinations, remove rules, convert 
     * rules to strings, etc. 
     * Iterator pointing to a rule remains valid until whole rule is removed. 
     * 
     * @note This iterator can't be dereferenced.
     */
    class RuleIterator
        {
        friend class Logger;

    public:

        /**
         * Constructor.
         */
        RuleIterator()
            {
            }

        /**
         * Destructor.
         */
        ~RuleIterator()
            {
            }

        /**
         * Copy constructor
         */
        RuleIterator
            (
            const RuleIterator & i
            )
            {
            iterator = i.iterator;
            }

        /**
         * Converts Rules::iterator (the real STL %iterator) to RuleIterator.
         * @param i Rules %iterator.
         */
        RuleIterator
            (
            const Rules::iterator & i
            )
            {
            iterator = i;
            }

    public:

        /**
         * Assignment operator.
         */
        RuleIterator & operator=
            (
            const RuleIterator & i
            )
            {
            iterator = i.iterator;
            return *this;
            }

        /**
         * Preincrement.
         */
        RuleIterator & operator++()
            {
            ++iterator;
            return *this;
            }

        /**
         * Postincrement.
         */
        RuleIterator operator++( int )
            {
            RuleIterator old = *this;
            ++iterator;
            return old;
            }

    public:

        /**
         * Compares iterators for equality.
         */
        bool operator==
            (
            const RuleIterator & i
            ) const
            {
            return iterator == i.iterator;
            }

        /**
         * Compares iterators for inequality.
         */
        bool operator!=
            (
            const RuleIterator & i
            ) const
            {
            return !(*this == i);
            }

    public:

        /**
         * Converts rule to string.
         * String format:
         * <tt><br>
         * <br>rule: selector(, selector)* -\> destination_name(, destination_name)*
         * <br>selector: [!]facility.type.priority
         * <br>facility: *|facility_name
         * <br>type: *|(&lt;=|=|&gt;=)type_name
         * <br>priority: (&lt;=|=|&gt;=)priority_name
         * <br></tt>
         */
        std::string to_string() const;

    private:

        Rules::iterator iterator;
            /**< Real STL iterator. */

        }; // class RuleIterator

    friend class RuleIterator;
    
public:

    // Initialization.

    /**
     * Constructor.
     * The %logger will be default-initialized.
     */
    Logger();

    /**
     * Destructor.
     */
    ~Logger();

    /**
     * Initializes %logger according to the variables stored in 
     * the @link Registry registry @endlink.
     
     * See @ref logger_configuration for format of the %logger configuration 
     * file.
     *
     * @exception ExLoggerInitializationError Wrong initialization data.
     *      ExLoggerInitializatonError or any of its descendants may be 
     *      thrown.
     * @note Discards current %logger %settings.
     */
    void initialize_from_registry();

    /**
     * Flushes all %logger destinations.
     * It cannot be called when core is in an inconsistent state.
     */
    void flush();

public:

    // Input

    /**
     * Processes %log %message.
     *
     * For each %rule in the @link Logger::Rules table of rules @endlink, 
     * if message matches rule @link Logger::Selector selector @endlink
     * (i.e. every @link Logger::SelectorElement selector element@endlink
     * contained within) it will be sent to all destinations bound with
     * this specific %rule. 
     * 
     * @note This means that message may be sent to single destination 
     * several times if rules are weird.
     *
     * @return true if at the message has been sent to at least one
     *         destination, false otherwise.
     */
    bool log
        (
        const LogMessage & message
        ) const;

public:

    /** @name Output configuration methods */
    //@{

    /**
     * Registers new @link LogDestination log destination @endlink.
     *
     * @param destination %Pointer to the %destination to %register. Should 
     *      be pointing to %object created by the @c new %operator (the 
     *      %logger will %delete it in its %destructor using @c %delete).
     *
     * @return false if the @c %destination has already been registered;
     *      true if the @c %destination was not registered so far. If true,
     *      Logger takes responsibility for deleting the %destination 
     *      %object when it is needed not any more.
     *
     * @note No %pointer should be held to @c %destination after it is 
     *      successfuly registered!!!
     */
    bool register_destination
        (
        LogDestination * destination
        );

    /**
     * Unregisters a @link LogDestination log destination @endlink.
     * 
     * @note This won't change existing rules, but nothing will be sent 
     *      to unregistered destinations. 
     */
    void unregister_destination
        (
        const std::string & name
        );

    /**
     * Searches for a %destination specified by its %name. 
     *
     * @param name Name of the %destination.
     * @return %Pointer to a %destination registered under the given 
     *      @c %name, NULL if not found.
     */
    LogDestination * find_destination
        (
        const std::string & name
        );

    /**
     * Unregisters all destinations and deletes all %destination objects.
     */
    void unregister_all_destinations();

    //@}

public:
    
    /** @name Filter configuration methods */
    //@{

    /**
     * Adds a %rule into the %logger @link Logger::Rules rules table @endlink.
     *
     * @param selector      %Rule %selector.
     * @return %Iterator that can be used to add destinations, convert %rule 
     *      to %string, etc. Returns end_rule() if the %rule is invalid.
     */
    RuleIterator insert_rule
        (
        const Selector &    selector
        );

    /**
     * Searches for a rule in the @link Logger::Rules rules table @endlink.
     *
     * @param selector A %selector that specifies the %rule to be found.
     * @return An %iterator pointing to the %rule if some has been found,
     *      otherwise the same %iterator that is returned from end_rule().
     *      
     */
    RuleIterator find_rule
        (
        const Selector &    selector
        );

    /**
     * Removes a rule from the @link Logger::Rules rules table @endlink.
     * 
     * @param rule An %iterator that specifies the %rule.
     */
    void erase_rule
        (
        RuleIterator        rule
        );

    /**
     * Removes a rule from the @link Logger::Rules rules table @endlink.
     *
     * @param selector A %selector that specifies the %rule.
     */
    void erase_rule
        (
        const Selector &    selector
        );

    /**
     * Removes all rules from the @link Logger::Rules rules table @endlink.
     */
    void clear_rules();

    /**
     * Gets iterator pointing to the first rule in the 
     * @link Logger::Rules rules table @endlink.
     */
    RuleIterator begin_rule();

    /**
     * Gets iterator pointing beyond the last rule in the
     * the @link Logger::Rules rules table @endlink
     */
    RuleIterator end_rule();

    //@}

    /** @name Destination manipulating methods */
    //@{

    /**
     * Inserts %destination into a specified %rule.
     *
     * @param rule          %Iterator pointing to the %rule to be modified.
     * @param destination   Name of %destination to be added.
     *
     * @note The %destination %object is not inserted.
     */
    void insert_destination
        (
        RuleIterator            rule,
        const std::string &     destination
        );

    /**
     * Removes %destination from a specified %rule.
     *
     * @param rule          %Iterator pointing to the %rule to be modified.
     * @param destination   Name of the destination to be deleted.
     */
    void erase_destination
        (
        RuleIterator            rule,
        const std::string &     destination
        );

    /**
     * Unregisters %destination. 
     *
     * @param destination The %destination to be unregistered.
     * @note The existing rules will not be changed, but nothing will be 
     *      sent to unregistered destinations. Deletes the %destination 
     *      %object.
     */
    void unregister_destination
        (
        LogDestination * destination
        );

    //@}

public:

    /**
     * Enables/disables logger.
     * @param enabled true if the %logger has to be enabled, false to disable.
     */
    void enable
        (
        bool        enabled
        );

    /**
     * Finds out whether the %logger is enabled or disabled.
     * @return true if the %logger is enabled, false otherwise.
     */
    bool is_enabled();

private:

    /** @name Init from registry helper routines */
    //@{
    
    /**
     * Registers all destinations that were configured in the 
     * @link Registry registry @endlink.
     *
     * @param registry %Reference to the %global @link Registry registry
     *      @endlink object.
     */
    void register_destinations_from_registry
        (
        const Registry &    registry
        );

    /**
     * Loads all SelectorElements from the @link Registry registry @endlink.
     *
     * @param registry %Reference to the %global @link Registry registry
     *      @endlink object.
     * @return %Map name-\>[ loaded selector ].
     */
    SelectorElementsMap load_selector_elements_from_registry
        (
        const Registry &    registry
        );

    /**
     * Loads one SelectorElement from the @link Registry registry @endlink.
     *
     * @param registry %Reference to the %global @link Registry registry
     *      @endlink object.
     * @param name Name of the @link SelectorElement selector element 
     *      @endlink (last component of path specifying the %node that 
     *      holds this SelectorElement's data).
     */
    SelectorElement load_selector_element_from_registry
        (
        const Registry &    registry,
        const std::string & name
        );

    /**
     * Loads all Selectors from the @link Registry registry @endlink.
     *
     * @param registry %Reference to the %global @link Registry registry
     *      @endlink object .
     * @param selector_elements_map Mapping 
     *      <tt>name -\> [ selector_element ]</tt> that has been obtained 
     *      from method load_selector_elements_from_registry().
     */
    void load_selectors_from_registry
        (
        const Registry &            registry,
        const SelectorElementsMap & selector_elements_map
        );

    /**
     * Loads one Selector from the @link Registry registry @endlink.
     *
     * @param registry %Reference to the %global @link Registry registry
     *      @endlink object.
     * @param name Name of the %selector (last component of path specifying
     *      the %node that holds this selector's data).
     * @param selector_elements_map Mapping 
     *      <tt>name -\> [ selector_element ]</tt> that has been obtained 
     *      from method load_selector_elements_from_registry().
     */
    void load_selector_from_registry
        (
        const Registry &                registry,
        const std::string &             name,
        const SelectorElementsMap &     selector_elements_map
        );

    /**
     * Processes priority %settings string from %selector %element definition
     * in the @link Registry registry @endlink.
     *
     * @param priority_string The %priority %string that was in the 
     *      configuration file.
     * @param selector_element %Reference to variable that will obtain the
     *      %result of the processing.
     */
    void process_priority_string
        (
        const std::string & priority_string,
        SelectorElement &   selector_element
        );

    /**
     * Processes type %settings string from %selector %element definition
     * in the @link Registry registry @endlink. 
     *
     * @param type_string The %type %string that was in the configuration
     *      file.
     * @param selector_element %Reference to variable that will obtain the
     *      %result of the processing.     
     */
    void process_type_string
        (
        const std::string & type_string,
        SelectorElement &   selector_element
        );

    /**
     * Processes facility %settings string from %selector %element definition
     * in the @link Registry registry @endlink. 
     *
     * @param facility_string The %facility %string that was in the 
     *      configuration file.
     * @param selector_element %Reference to variable that will obtain the
     *      %result of the processing.
     */
    void process_facility_string
        (
        const std::string & facility_string,
        SelectorElement &   selector_element
        );

    //@}

private:
    
    RegisteredDestinations  registered_destinations;
        /**< %Set of registered destinations. */

    DestinationMapping      destination_mapping;
        /**< Mapping <tt>%Destination name -> destination pointer</tt>. */

    Rules                   rules;
        /**< %Table of %logger %rules. */

    LoggerSettings          settings;
        /**< %Logger %settings. */

    mutable LogMessage      previous_message;
        /**< The last logged message. */

    mutable Int32           message_match_count;
        /**< How many times was the @c %previous_message repeated. */

    bool                    loading_configuration;
        /**< Are we loading logger configuration from registry now? */

    }; // class Logger

/** @} */ // addtogroup logger


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_LOGGER_H
