/** @file byte_order.h
 * Conversions between host and network byte order.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BYTE_ORDER_H
#define MASSIV_BYTE_ORDER_H


#ifdef COMPILER_ICC

// ICC is always little endian.

namespace Massiv {
namespace Core {


/** @addtogroup net
 * @{
 */

/**
 * Functions used to convert integer byte order between host-order and network-order.
 * Implemented because of ICC - it does not like gcc-trickery in declaration of
 * ntohs-like methods.
 */
namespace ByteOrder {


/** Convert 16bit integer in host byteorder to network byteorder. */
inline UInt16 network_to_host_16
    (
    UInt16 value
    )
    {
    return ( ( value & 0xff00 ) >> 8 ) |
           ( ( value & 0x00ff ) << 8 );
    }


/** Convert 16bit integer in network byteorder to host byteorder. */
inline UInt16 host_to_network_16
    (
    UInt16 value
    )
    {
    return network_to_host_16( value );
    }


/** Convert 32bit integer in host byteorder to network byteorder. */
inline UInt32 network_to_host_32
    (
    UInt32 value
    )
    {
    return ( ( value & 0xff000000 ) >> 24 ) |
           ( ( value & 0x00ff0000 ) >> 8 ) |
           ( ( value & 0x0000ff00 ) << 8 ) |
           ( ( value & 0x000000ff ) << 24 );
    }


/** Convert 32bit integer in network byteorder to host byteorder. */
inline UInt32 host_to_network_32
    (
    UInt32 value
    )
    {
    return network_to_host_32( value );
    }

    
} // namespace Massiv::Core::ByteOrder

/** @} */ // addtogroup net

} // namespace Massiv::Core
} // namespace Massiv


#else // compiler


#include <netinet/in.h>


namespace Massiv {
namespace Core {


/** @addtogroup net
 * @{
 */

namespace ByteOrder {


/** Convert 16bit integer in host byteorder to network byteorder. */
inline UInt16 network_to_host_16
    (
    UInt16 value
    )
    {
    return ntohs( value );
    }


/** Convert 16bit integer in network byteorder to host byteorder. */
inline UInt16 host_to_network_16
    (
    UInt16 value
    )
    {
    return htons( value );
    }


/** Convert 32bit integer in host byteorder to network byteorder. */
inline UInt32 network_to_host_32
    (
    UInt32 value
    )
    {
    return ntohl( value );
    }


/** Convert 32bit integer in network byteorder to host byteorder. */
inline UInt32 host_to_network_32
    (
    UInt32 value
    )
    {
    return htonl( value );
    }


} // namespace Massiv::Core::ByteOrder

/** @} */ // addtogroup net

} // namespace Massiv::Core
} // namespace Massiv


#endif // compiler


#endif // MASSIV_BYTE_ORDER_H

