/** @file socket_unix.cpp
 * Unix socket objects implementation.
 */

/*
 * begin        : Mon May 20 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/arch/unix/socket_unix.h"
#include "net/arch/unix/byte_order.h"
#include "net/net.h"
#include "logger/logger.h"
#include "global/global.h"
#include <sstream>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <cerrno>
#include <cstring>
#include <sstream>

namespace Massiv {
namespace Core {

using namespace NetworkCommon;


namespace {

const int SOCKET_ERROR  =   -1;
const int INVALID_SOCKET =  -1;


/** Format Unix errno error to a string. */
std::string format_message
    (
    int err_no
    )
    {
    return std::strerror( err_no );
    }


/** Resolves @c address to socket library address. */
bool resolve
    (
    const ServerAddress &   address,
    sockaddr_in &           sock_address
    )
    {
    memset( &sock_address, 0, sizeof( sock_address ) );

    sock_address.sin_family = AF_INET;

    if( inet_aton( address.get_dns_name().c_str(), &sock_address.sin_addr ) == 0 )
        {
        hostent * host = ::gethostbyname( address.get_dns_name().c_str() );
        
        if( host == NULL )
            {
            std::ostringstream s;
            
            s   << "Adress resolving failed for " << address.get_dns_name()
                << ". " << format_message( errno );

            Global::log_debug
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_HIGHEST,
                s.str()
                );

            return false;
            }

        sock_address.sin_addr.s_addr = ( ( in_addr * ) host->h_addr )->s_addr;
        }

    sock_address.sin_port = ByteOrder::host_to_network_16( address.get_port() );

    return true;
    }
    

/** Binds @c socket_handle to a specified port. */
bool bind
    (
    int     socket_handle,
    UInt16  port    
    )
    {
    sockaddr_in sock_address;
    int         result;

    memset( &sock_address, 0, sizeof( sock_address ) );

    sock_address.sin_family = AF_INET;
    sock_address.sin_addr.s_addr = ByteOrder::host_to_network_32( INADDR_ANY );
    sock_address.sin_port = ByteOrder::host_to_network_16( port );
    
    result = ::bind
        (
        socket_handle,
        reinterpret_cast<sockaddr *>( &sock_address ),
        sizeof( sock_address )
        );
        
    if( result == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot bind socket! " << format_message( errno );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );
        
        return false;
        }

    return true;
    }


/** Enables/disable non-blocking IO. */
bool set_io_mode
    (
    int     socket_handle,
    bool    non_blocking
    )
    {
    unsigned long l = non_blocking;

    if( ::ioctl( socket_handle, FIONBIO, &l ) == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot switch socket to ";
        
        if( non_blocking )
            {
            oss << "non";
            }

        oss << "blocking mode. " << format_message( errno );

        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }

    return true;
    }


/** Returns number of bytes that can be received from the socket. */
int pending_data_size
    (
    int socket_handle
    )
    {
    unsigned long byte_count;
    
    if( ::ioctl( socket_handle, FIONREAD, &byte_count ) == SOCKET_ERROR )
        {
        return 0;
        }
    
    return byte_count;
    }


/** Copies socket library address. */
void copy_address
    (
    sockaddr_in &       target,
    const sockaddr_in & source
    )
    {
    memcpy( &target, &source, sizeof( source ) );
    }


/** Reuses address that will be bound to @c socket_handle. */
bool reuse_address
    (
    int socket_handle
    )
    {
    int flag = 1;
    
    return ::setsockopt
        (
        socket_handle,
        SOL_SOCKET,
        SO_REUSEADDR,
        &flag,
        sizeof( flag )
        ) == 0;
    }

}


// Implementation of TCPSocketUnix
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TCPSocketUnix::TCPSocketUnix()
    {
    sock = INVALID_SOCKET;
    }


TCPSocketUnix::TCPSocketUnix
    (
    int socket
    ) : sock( socket )
    {
    }


TCPSocketUnix::~TCPSocketUnix()
    {
    if( sock != INVALID_SOCKET )
        {
        close();
        }
    }


bool TCPSocketUnix::connect
    ( 
    const ServerAddress & address
    )
    {
    sockaddr_in sock_address;
    int         result;

    hope( sock == INVALID_SOCKET );

    // Create socket.

    if( ( sock = ::socket( AF_INET, SOCK_STREAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( errno );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }
    
    // Resolve @c address.

    if( !resolve( address, sock_address ) )
        {
        return false;
        }

    // Connect.

    result = ::connect
        (
        sock,
        reinterpret_cast<sockaddr *>( &sock_address ),
        sizeof( sock_address )
        );
         
    if( result == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot connect socket! " << format_message( errno );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        close();

        return false;
        }
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


bool TCPSocketUnix::start_server
    (
    UInt16  port,
    int     backlog
    )
    {
    hope( sock == INVALID_SOCKET );

    // Create socket.
    
    if( ( sock = ::socket( AF_INET, SOCK_STREAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( errno );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }
    
    // Reuse socket address.

    // reuse_address( sock );
       
    // Bind socket to a specified port.

    if( !bind( sock, port ) )
        {
        close();
        return false;
        }

    // Start listening.

    if( ::listen( sock, backlog ) == SOCKET_ERROR )
        {
        std::ostringstream oss;

        oss << "Cannot listen on socket! " << format_message( errno );

        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        close();
        return false;
        }

    touch_server_start_timestamp();
    server_port = port;
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


Socket * TCPSocketUnix::accept()
    {
    int                 new_socket_handle;
    TCPSocketUnix *     new_socket = NULL;
    
    hope( sock != INVALID_SOCKET );
    
    // Accept.

    if( ( new_socket_handle = ::accept( sock, NULL, NULL) ) == INVALID_SOCKET )
        {
        return NULL;
        }
    
    // Create socket object and bind it to @c new_socket_handle.

    new_socket = new TCPSocketUnix( new_socket_handle );

    // Enable non-blocking mode.

    if( !new_socket->set_non_block( true ) )
        {
        delete new_socket;
        return NULL;
        }

    new_socket->server_port = server_port;

    return new_socket;
    }

    
void TCPSocketUnix::close()
    {
    TCPSocket::close();

    ::close( sock );
    sock = INVALID_SOCKET;
    }

    
int TCPSocketUnix::internal_send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    int bytes_sent = ::send
        (
        sock,
        reinterpret_cast<const char *>( buf ),
        len,
        MSG_NOSIGNAL
        );

    if( bytes_sent == SOCKET_ERROR )
        {
        if( errno != EWOULDBLOCK && errno != EAGAIN )
            {
            std::ostringstream  oss;
            oss << "Can't send data to socket. " << format_message( errno );
            throw ExSendError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    return bytes_sent;
    }


int TCPSocketUnix::internal_receive
    (
    UInt8 * buf,
    int     len
    )
    {
    int bytes_fetched = ::recv( sock, reinterpret_cast<char *>( buf ), len, 0 );
    
    if( bytes_fetched == SOCKET_ERROR )
        {
        if( errno != EAGAIN && errno != EWOULDBLOCK )
            {
            std::ostringstream oss;
            oss << "Can not fetch data from the socket." << format_message( errno );
            throw ExReceiveError( oss.str() );
            }
        else
            {
            return 0;
            }
        }

    return bytes_fetched;
    }


int TCPSocketUnix::get_pending_data_amount()
    {
    return pending_data_size( sock );
    }


bool TCPSocketUnix::set_non_block
    (
    bool non_blocking
    )
    {
    if( !set_io_mode( sock, non_blocking ) )
        {
        close();
        return false;
        }
    else
        {
        return true;
        }
    }


// Implementation of UDPSocketUnix
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UDPSocketUnix::UDPSocketUnix()
      : udp_connected( false ),
        sock( INVALID_SOCKET )
    {
    }


UDPSocketUnix::UDPSocketUnix
    (
    int socket
    ) : udp_connected( false ),
        sock( socket )
    {
    }


UDPSocketUnix::~UDPSocketUnix()
    {
    if( sock != INVALID_SOCKET )
        {
        close();
        }
    }


bool UDPSocketUnix::connect
    ( 
    const ServerAddress & address 
    )
    {
    sockaddr_in     sock_address;
    int             result;

    hope( sock == INVALID_SOCKET );

    // Create socket.

    sock = socket( AF_INET, SOCK_DGRAM, 0 );
    
    // Resolve @c address.
    
    resolve( address, sock_address );
    
    // Bind.
    
    if( !bind( sock, 0 ) )
        {
        close();
        return false;
        }
    
    // Connect.

    result = ::connect
        (
        sock,
        reinterpret_cast<sockaddr *>( &sock_address ),
        sizeof( sock_address )
        );
    
    if( result == SOCKET_ERROR )
        {
        close();
        return false;
        }
    
    copy_address( destination_address, sock_address );
    udp_connected = true;
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


UInt16 UDPSocketUnix::get_bound_port()
    {
    hope( sock != INVALID_SOCKET );
    
    // getsockname
    struct sockaddr_in s_in;
    socklen_t          l = sizeof( struct sockaddr_in );
    
    if( getsockname( sock, ( sockaddr * ) &s_in, &l ) == SOCKET_ERROR )
        {
        return 0;
        }
    
    return ( UInt16 ) htons( s_in.sin_port );
    }


bool UDPSocketUnix::start_server
    (
    UInt16  port,
    int     backlog
    )
    {
    hope( sock == INVALID_SOCKET );

    // Create socket.

    if( ( sock = ::socket( AF_INET, SOCK_DGRAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( errno );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );
        
        return false;
        }

    // Reuse address.

    // reuse_address( sock );

    // Bind socket to a specified port.

    if( !bind( sock, 0 ) )
        {
        close();
        return false;
        }

    // Enable non-blocking mode.

    return set_non_block( true );
    }


void UDPSocketUnix::close()
    {
    ::close( sock );
    sock = INVALID_SOCKET;
    udp_connected = false;
    }


int UDPSocketUnix::internal_send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    hope( sock != INVALID_SOCKET );
    hope( udp_connected );

    int bytes_sent = ::sendto
        (
        sock,
        reinterpret_cast<const char *>( buf ),
        len,
        MSG_NOSIGNAL,
        reinterpret_cast<sockaddr *>( &destination_address ),
        sizeof( destination_address )
        );

    if( bytes_sent == SOCKET_ERROR )
        {
        if( errno != EAGAIN && errno != EWOULDBLOCK )
            {
            std::ostringstream  oss;
            oss << "Can't send data to socket. " << format_message( errno );
            throw ExSendError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    return bytes_sent;
    }


int UDPSocketUnix::internal_receive
    (
    UInt8 * buf,
    int     len
    )
    {
    sockaddr_in received_from;
    socklen_t   address_length;

    address_length = sizeof( received_from );

    int bytes_fetched = ::recvfrom
        (
        sock,
        reinterpret_cast<char *>( buf ),
        len,
        0,
        reinterpret_cast<sockaddr *>( &received_from ),
        &address_length
        );

    if( bytes_fetched == SOCKET_ERROR )
        {
        if( errno != EAGAIN && errno != EWOULDBLOCK )
            {
            std::ostringstream  oss;
            oss << "Can't receive data from socket. " << format_message( errno );
            throw ExReceiveError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    hope( address_length == sizeof( received_from ) );
    
    // Update socket state.

    if( !udp_connected )
        {
        copy_address( destination_address, received_from );
        udp_connected = true;
        }

    return bytes_fetched;
    }


int UDPSocketUnix::get_pending_data_amount()
    {
    return pending_data_size( sock );
    }


bool UDPSocketUnix::set_non_block
    (
    bool    non_blocking
    )
    {
    if( !set_io_mode( sock, non_blocking ) )
        {
        close();
        return false;
        }
    else
        {
        return true;
        }
    }


void UDPSocketUnix::set_destination_address
    (
    const ServerAddress & sa
    )
    {
    sockaddr_in     winsock_address;
    resolve( sa, winsock_address );
    copy_address( destination_address, winsock_address );    
    udp_connected = true;
    }

    
} // namespace Massiv::Core
} // namespace Massiv
