/** @file socket_win32.cpp
 * Win32 socket objects implementation.
 */

/*
 * begin        : Mon May 20 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/arch/win32/socket_win32.h"
#include "net/net.h"
#include "logger/logger.h"
#include "global/global.h"
#include <sstream>

#include <windows.h>
#include <winsock2.h>

namespace Massiv {
namespace Core {

using namespace NetworkCommon;


namespace {


/** Format Win32 error to a string. */
std::string format_message
    (
    int win32_error
    )
    {
    static char message_buffer[ 1024 ];

    FormatMessage
        (
        FORMAT_MESSAGE_IGNORE_INSERTS |
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        win32_error,
        MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
        message_buffer,
        sizeof( message_buffer ) / sizeof( char ) - 1,
        NULL
        );

    return std::string( message_buffer );
    }


/** Resolves @c address to WinSock address. */
bool resolve
    (
    const ServerAddress &   address,
    sockaddr_in &           winsock_address
    )
    {
    memset( &winsock_address, 0, sizeof( winsock_address ) );

    winsock_address.sin_family = AF_INET;
    winsock_address.sin_addr.s_addr = ::inet_addr( address.get_dns_name().c_str() );

    if( winsock_address.sin_addr.s_addr == INADDR_NONE )
        {
        hostent * host = ::gethostbyname( address.get_dns_name().c_str() );
        
        if( host == NULL )
            {
            std::ostringstream s;
            
            s   << "Adress resolving failed for " << address.get_dns_name()
                << ". " << format_message( ::WSAGetLastError() );

            Global::log_debug
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_HIGHEST,
                s.str()
                );

            return false;
            }

        winsock_address.sin_addr.s_addr = ( ( in_addr * ) host->h_addr )->s_addr;
        }

    winsock_address.sin_port = ::htons( address.get_port() );

    return true;
    }
    

/** Binds @c socket_handle to a specified port. */
bool bind
    (
    int     socket_handle,
    UInt16  port    
    )
    {
    sockaddr_in winsock_address;
    int         result;

    memset( &winsock_address, 0, sizeof( winsock_address ) );

    winsock_address.sin_family = AF_INET;
    winsock_address.sin_addr.s_addr = ::htonl( INADDR_ANY );
    winsock_address.sin_port = ::htons( port );
    
    result = ::bind
        (
        socket_handle,
        reinterpret_cast<sockaddr *>( &winsock_address ),
        sizeof( winsock_address )
        );
        
    if( result == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot bind socket! " << format_message( ::WSAGetLastError() );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );
        
        return false;
        }

    return true;
    }


/** Enables/disable non-blocking IO. */
bool set_io_mode
    (
    int     socket_handle,
    bool    non_blocking
    )
    {
    unsigned long l = non_blocking;

    if( ::ioctlsocket( socket_handle, FIONBIO, &l ) == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot switch socket to ";
        
        if( non_blocking )
            {
            oss << "non";
            }

        oss << "blocking mode. " << format_message( ::WSAGetLastError() );

        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }

    return true;
    }


/** Returns number of bytes that can be received from the socket. */
int pending_data_size
    (
    int socket_handle
    )
    {
    unsigned long byte_count;
    
    if( ::ioctlsocket( socket_handle, FIONREAD, &byte_count ) == SOCKET_ERROR )
        {
        return 0;
        }
    
    return byte_count;
    }


/** Copies winsock address. */
void copy_address
    (
    sockaddr_in &       target,
    const sockaddr_in & source
    )
    {
    memcpy( &target, &source, sizeof( source ) );
    }

}


// Implementation of TCPSocketWin32
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

TCPSocketWin32::TCPSocketWin32()
    {
    sock = INVALID_SOCKET;
    }


TCPSocketWin32::TCPSocketWin32
    (
    int socket
    ) : sock( socket )
    {
    }


TCPSocketWin32::~TCPSocketWin32()
    {
    if( sock != INVALID_SOCKET )
        {
        close();
        }
    }


bool TCPSocketWin32::connect
    ( 
    const ServerAddress & address
    )
    {
    sockaddr_in winsock_address;
    int         result;

    hope( sock == INVALID_SOCKET );

    // Create socket.

    if( ( sock = ::socket( AF_INET, SOCK_STREAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( ::WSAGetLastError() );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }
    
    // Resolve @c address.

    if( !resolve( address, winsock_address ) )
        {
        return false;
        }

    // Connect.

    result = ::connect
        (
        sock,
        reinterpret_cast<sockaddr *>( &winsock_address ),
        sizeof( winsock_address )
        );
         
    if( result == SOCKET_ERROR )
        {
        std::ostringstream oss;
        
        oss << "Cannot connect socket! " << format_message( ::WSAGetLastError() );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        close();

        return false;
        }
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


bool TCPSocketWin32::start_server
    (
    UInt16  port,
    int     backlog
    )
    {
    hope( sock == INVALID_SOCKET );

    // Create socket.
    
    if( ( sock = ::socket( AF_INET, SOCK_STREAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( ::WSAGetLastError() );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        return false;
        }
       
    // Socket reusing was removed!    
    // Win32 socket reusing have other behaviour than on unix. It allows to bind
    // multiple addresses to one port, but that's something, what we do not
    // want!

    // Bind socket to a specified port.

    if( !bind( sock, port ) )
        {
        close();
        return false;
        }

    // Start listening.

    if( ::listen( sock, backlog ) == SOCKET_ERROR )
        {
        std::ostringstream oss;

        oss << "Cannot listen on socket! " << format_message( ::WSAGetLastError() );

        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );

        close();
        return false;
        }

    touch_server_start_timestamp();
    server_port = port;
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


Socket * TCPSocketWin32::accept()
    {
    int                 new_socket_handle;
    TCPSocketWin32 *    new_socket = NULL;
    
    hope( sock != INVALID_SOCKET );
    
    // Accept.

    if( ( new_socket_handle = ::accept( sock, NULL, NULL) ) == INVALID_SOCKET )
        {
        return NULL;
        }
    
    // Create socket object and bind it to @c new_socket_handle.

    new_socket = new TCPSocketWin32( new_socket_handle );

    // Enable non-blocking mode.

    if( !new_socket->set_non_block( true ) )
        {
        delete new_socket;
        return NULL;
        }

    new_socket->server_port = server_port;

    return new_socket;
    }

    
void TCPSocketWin32::close()
    {
    TCPSocket::close();

    closesocket( sock );
    sock = INVALID_SOCKET;
    }

    
int TCPSocketWin32::internal_send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    int bytes_sent = ::send
        (
        sock,
        reinterpret_cast<const char *>( buf ),
        len,
        0
        );

    if( bytes_sent == SOCKET_ERROR )
        {
        int win32_error = ::WSAGetLastError();
        
        if( win32_error != WSAEWOULDBLOCK )
            {
            std::ostringstream  oss;
            oss << "Can't send data to socket. " << format_message( win32_error );
            throw ExSendError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    return bytes_sent;
    }


int TCPSocketWin32::internal_receive
    (
    UInt8 * buf,
    int     len
    )
    {
    int bytes_fetched = ::recv( sock, reinterpret_cast<char *>( buf ), len, 0 );
    
    if( bytes_fetched == SOCKET_ERROR )
        {
        UInt32 win32_error = ::WSAGetLastError();
        
        if( win32_error != WSAEWOULDBLOCK )
            {
            std::ostringstream oss;
            oss << "Can not fetch data from the socket. " << format_message( win32_error );
            throw ExReceiveError( oss.str() );
            }
        else
            {
            return 0;
            }
        }

    return bytes_fetched;
    }


int TCPSocketWin32::get_pending_data_amount()
    {
    return pending_data_size( sock );
    }


bool TCPSocketWin32::set_non_block
    (
    bool non_blocking
    )
    {
    if( !set_io_mode( sock, non_blocking ) )
        {
        close();
        return false;
        }
    else
        {
        return true;
        }
    }
    

// Implementation of UDPSocketWin32
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UDPSocketWin32::UDPSocketWin32() : 
    sock( INVALID_SOCKET ),
    udp_connected( false )
    {
    }


UDPSocketWin32::UDPSocketWin32
    (
    int socket
    ) : sock( socket ),
        udp_connected( false )
    {
    }


UDPSocketWin32::~UDPSocketWin32()
    {
    if( sock != INVALID_SOCKET )
        {
        close();
        }
    }


bool UDPSocketWin32::connect
    ( 
    const ServerAddress & address 
    )
    {
    sockaddr_in     winsock_address;
    int             result;

    hope( sock == INVALID_SOCKET );

    // Create socket.

    sock = socket( AF_INET, SOCK_DGRAM, 0 );
    
    // Resolve @c address.
    
    resolve( address, winsock_address );
    
    // bind
    
    if( !bind( sock, 0 ) )
        {
        close();
        return false;
        }
    
    // Connect.

    result = ::connect
        (
        sock,
        reinterpret_cast<sockaddr *>( &winsock_address ),
        sizeof( winsock_address )
        );
    
    if( result == SOCKET_ERROR )
        {
        close();
        return false;
        }
    
    copy_address( destination_address, winsock_address );
    udp_connected = true;
    
    // Enable non-blocking mode.

    return set_non_block( true );
    }


UInt16 UDPSocketWin32::get_bound_port()
    {
    hope( sock != INVALID_SOCKET );
    
    // getsockname
    struct sockaddr_in s_in;
    int                l = sizeof( struct sockaddr_in );
    
    if( getsockname( sock, ( sockaddr * ) &s_in, &l ) == SOCKET_ERROR )
        {
        return 0;
        }
    
    return ( UInt16 ) ::htons( s_in.sin_port );
    }


bool UDPSocketWin32::start_server
    (
    UInt16   port,
    int      backlog
    )
    {
    hope( sock == INVALID_SOCKET );

    // Create socket.

    if( ( sock = ::socket( AF_INET, SOCK_DGRAM, 0 ) ) == INVALID_SOCKET )
        {
        std::ostringstream oss;
        
        oss << "Cannot create socket! " << format_message( ::WSAGetLastError() );
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );
        
        return false;
        }

    // Socket reusing removed!
    // Win32 socket reusing have other behaviour than unix. It allows to bind
    // multiple addresses to one port, but that's something, what we do not
    // want!

    // Bind socket to a specified port.

    if( !bind( sock, 0 ) )
        {
        close();
        return false;
        }

    // Enable non-blocking mode.

    return set_non_block( true );
    }


void UDPSocketWin32::close()
    {
    closesocket( sock );
    sock = INVALID_SOCKET;
    udp_connected = false;
    }


int UDPSocketWin32::internal_send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    hope( sock != INVALID_SOCKET );
    hope( udp_connected );

    int bytes_sent = ::sendto
        (
        sock,
        reinterpret_cast<const char *>( buf ),
        len,
        0,
        reinterpret_cast<sockaddr *>( &destination_address ),
        sizeof( destination_address )
        );

    if( bytes_sent == SOCKET_ERROR )
        {
        UInt32 win32_error = ::WSAGetLastError();
        
        if( win32_error != WSAEWOULDBLOCK )
            {
            std::ostringstream oss;
            oss << "Can't send data to socket. " << format_message( ::WSAGetLastError() );
            throw ExSendError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    return bytes_sent;
    }


int UDPSocketWin32::internal_receive
    (
    UInt8 * buf,
    int     len
    )
    {
    sockaddr_in received_from;
    int         address_length;
    
    address_length = sizeof( received_from );

    int bytes_fetched = ::recvfrom
        (
        sock,
        reinterpret_cast<char *>( buf ),
        len,
        0,
        reinterpret_cast<sockaddr *>( &received_from ),
        &address_length
        );

    if( bytes_fetched == SOCKET_ERROR )
        {
        UInt32 win32_error = ::WSAGetLastError();
        
        if( win32_error != WSAEWOULDBLOCK )
            {
            std::ostringstream oss;
            oss << "Can't receive data from socket. " << format_message( win32_error );
            throw ExReceiveError( oss.str() );
            }
        else
            {
            return 0;
            }
        }
    
    hope( address_length == sizeof( received_from ) );

    // Update socket state.

    if( !udp_connected )
        {
        copy_address( destination_address, received_from );
        udp_connected = true;
        }

    return bytes_fetched;
    }


int UDPSocketWin32::get_pending_data_amount()
    {
    return pending_data_size( sock );
    }


bool UDPSocketWin32::set_non_block
    (
    bool    non_blocking
    )
    {
    if( !set_io_mode( sock, non_blocking ) )
        {
        close();
        return false;
        }
    else
        {
        return true;
        }
    }


void UDPSocketWin32::set_destination_address
    (
    const ServerAddress & sa
    )
    {
    sockaddr_in     winsock_address;
    resolve( sa, winsock_address );
    copy_address( destination_address, winsock_address );    
    udp_connected = true;
    }
    
    
} // namespace Massiv::Core
} // namespace Massiv
