/** @file bitistream.cpp
 * Bit-oriented input stream buffer.
 */

/*
 * begin        : Tue Dec 18 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/bitistream.h"
#include "utility/debug.h"
#include "status/exception.h"
#include <cstring>
#include <sstream>

namespace Massiv {
namespace Core {


BitIStream & BitIStream::operator=
    (
    const BitIStream & stream
    )
    {
    bit = stream.bit;
    offset = stream.offset;
    buffer = stream.buffer;
    BitIOStreamBuffer::ConstIterator beg = stream.buffer.begin();
    data = buffer.begin() + ( stream.data - beg );
    end = buffer.begin() + ( stream.end - beg );
    return *this;
    }


BitIOStreamBuffer BitIStream::bind
    (
    BitIOStreamBuffer   buf,
    bool                dup,
    SizeType            offset
    )
    {
    hope( offset % 8 == 0 );

    if( offset / 8 > buf.size() )
        throw_out_of_range();

    this->offset = offset;

    BitIOStreamBuffer   prev = buffer;

    if( dup )
        {
        BitIOStreamBuffer           b;
        BitIOStreamBuffer::Iterator si( buf.begin() );
        BitIOStreamBuffer::Iterator se( buf.end() );
        BitIOStreamBuffer::Iterator di;

        b.resize( buf.size() );
        for( di = b.begin(); si != se; si++, di++ )
            {
            (*di) = (*si);
            }

        buf = b;
        }

    bit    = 0;
    buffer = buf;
    data   = buffer.begin() + offset / 8;
    end    = buffer.end();

    return prev;
    }


BitIOStreamBuffer BitIStream::bind_part
    (
    BitIOStreamBuffer   buf,
    bool                dup,
    SizeType            begin,
    SizeType            len
    )
    {
    hope( begin + len <= buf.size() );

    this->offset = 0;

    BitIOStreamBuffer   prev = buffer;

    if( dup )
        {
        BitIOStreamBuffer           b;
        BitIOStreamBuffer::Iterator si( buf.begin() + begin );
        BitIOStreamBuffer::Iterator se( buf.begin() + begin + len );
        BitIOStreamBuffer::Iterator di;

        b.resize( len );
        for( di = b.begin(); si != se; si++, di++ )
            {
            (*di) = (*si);
            }

        buffer = b;
        data   = buffer.begin();
        end    = buffer.end();
        }
    else
        {
        buffer = buf;
        data   = buffer.begin() + begin;
        end    = buffer.begin() + begin + len;
        }

    bit    = 0;
    return prev;
    }


BitIStream BitIStream::head
    (
    bool duplicate
    )
    {
    BitIStream  istr;
    SizeType    end_offset;

    if( duplicate )
        {
        BitIOStreamBuffer           newbuf;
        BitIOStreamBuffer::Iterator si( buffer.begin() + offset / 8 );
        BitIOStreamBuffer::Iterator di;

        newbuf.resize( data - buffer.begin() - offset / 8 + 1 );
        for( di = newbuf.begin(); si != data + 1; si++, di++ )
            {
            (*di) = (*si);
            }

        end_offset = newbuf.size();
        istr.bind( newbuf, false, 0 );
        }
    else
        {
        end_offset = data - buffer.begin() - offset / 8 + 1;
        istr.bind( buffer, false, offset );
        }

    istr.change_stream_end( end_offset * 8 );

    return istr;
    }  // BitIStream::head()


BitIStream BitIStream::tail
    (
    bool duplicate
    )
    {
    BitIStream  istr;

    // Align the source stream to byte boundary
    // ???: Is this really needed? Boovie considered missing alignment
    //      a bug, but to let user align the source explicitly (if he
    //      wants) leads to greater generality.
    // Boovie: Documentation to tail() says that the stream
    //         is first aligned, so it was a bug.
    // Boovie: Also offset passed to istr.bind() below is specified
    //         byte aligned so the alignment must be done!
    // Boovie: If anybody change this, do not forget to tell me to update
    //         migration message serialization, which works with
    //         bytes (not bits) and expects tail() to align the stream.
    seek( ( 8 - bit ) & 0x07, WHENCE_CURRENT );

    if( duplicate )
        {
        BitIOStreamBuffer           newbuf;
        BitIOStreamBuffer::Iterator si( data );
        BitIOStreamBuffer::Iterator di;

        newbuf.resize( end - data );
        for( di = newbuf.begin(); si != end; si++, di++ )
            {
            (*di) = (*si);
            }

        istr.bind( newbuf, false, 0 );
        }
    else
        {
        istr.bind( buffer, false, ( data - buffer.begin() ) * 8 );
        }

    istr.change_stream_end( ( end - data ) * 8 );

    return istr;
    }  // BitIStream::tail()


BitIStream::SizeType BitIStream::change_offset
    (
    SizeType new_offset
    )
    {
    hope( new_offset % 8 == 0 );

    if( new_offset > buffer.size() * 8 - 1 )
        {
        throw_out_of_range();
        }

    int old_offset = offset;
    offset  = new_offset;

    return old_offset;
    }


BitIStream::SizeType BitIStream::change_stream_end
    (
    SizeType new_end_offset
    )
    {
    hope( new_end_offset % 8 == 0 );

    SizeType byte_offset = new_end_offset / 8;

    if( byte_offset + offset / 8 > buffer.size() )
        {
        throw_out_of_range();
        }

    int result = 8 * ( end - buffer.begin() ) - offset;

    end = buffer.begin() + byte_offset + offset / 8;

    if( data > end )  data = end;

    return result;
    }


void BitIStream::seek
    (
    int         ofs,
    SeekWhence  whence
    )
    {
    hope( ( offset % 8 ) == 0 );
    
    switch( whence )
        {
        case WHENCE_START:
            {
            int       bofs; // How many bytes the data pointer will move.
            const int stream_size = end - buffer.begin() - offset / 8;
            const int stream_bit_size = stream_size * 8;
            hope( stream_bit_size >= 0 );
        
            if( ofs == 0 )
                {
                bit = 0;
                data = buffer.begin();
                bofs = offset / 8;
                }
            else if( ofs < 0 )
                {
                // Trying to seek before beginning of the stream.
                throw_out_of_range();
                }
            else
                {
                if( ofs > stream_bit_size )
                    {
                    // Trying to seek past end of the stream.
                    throw_out_of_range();
                    }
            
                data = buffer.begin();
                bofs = offset / 8 + ofs / 8;
                bit = ofs % 8;
                }
                
            data += bofs;
            break;
            }
            
        case WHENCE_CURRENT:
            {
            const int distance = data - buffer.begin() - offset / 8;
            const int bit_distance = distance * 8 + bit;
            const int new_offset = bit_distance + ofs;
            seek( new_offset, WHENCE_START );
            break;
            }

        case WHENCE_END:
            {
            int       bofs; // How many bytes the data pointer will move.
            const int stream_size = end - buffer.begin() - offset / 8;
            const int stream_bit_size = stream_size * 8;
            hope( stream_bit_size >= 0 );
        
            if( ofs == 0 )
                {
                bit = 0;
                data = end;
                bofs = 0;
                }
            else if( ofs > 0 )
                {
                // Trying to seek past end of the stream.
                throw_out_of_range();
                }
            else
                {
                if( ( - ofs ) > stream_bit_size )
                    {
                    // Trying to seek before beginning of the stream.
                    throw_out_of_range();
                    }
        
                data = end;
                bofs = ( ( ofs + 1 ) / 8 ) - 1;
                bit = 8 - ( ( - ofs ) % 8 );
                if( bit == 8 )
                    {
                    bit = 0;
                    }
                }
                
            data += bofs;
            break;
            }
            
        default:
            panic( "Unexpected 'whence' argument." );
        }
    }  // BitIStream::seek()


Float32 BitIStream::read_float()
    {
    union
        {
        UInt32  i;
        Float32 f;
        } value;
    value.i = read_int( 32 );
    return value.f;
    }


Float64 BitIStream::read_float64()
    {
    union
        {
        UInt64  i;
        Float64 f;
        } value;
    value.i = read_int64( 64 );
    return value.f;
    }


std::string BitIStream::read_string()
    {
    std::string s;
    UInt32      c;

    while( ( c = read_int( 7 ) ) != 0 )
        {
        s.push_back( static_cast<char>( c ) );
        }

    return s;
    }


std::string BitIStream::read_full_string()
    {
    std::string s;
    UInt32      c;
    UInt32      length;
    UInt32      len;
    
    length = read_int( 32 );
    len = 0;

    while( len < length )
        {
        c = read_int( 8 );
        s.push_back( static_cast<char>( c ) );
        len++;
        }

    return s;
    }


void BitIStream::read_raw
    (
    UInt8 *     buf,
    UInt32      length
    )
    {
    // Must be byte aligned.
    align_to_byte();
    hope( !bit );

    hope( end - data >= static_cast< Int32 >( length ) );
    std::memcpy( buf, &( *data ), length );
    data += length;
    }


UInt32 BitIStream::read_int
    (
    int     bits
    )
    {
    int     shift;
    UInt32  result;
    UInt32  mask;

    hope( 0 < bits );
    hope( bits <= 32 );
    hope( 0 <= bit );
    hope( bit < 8 );

    if( bits < 32 )
        mask = ( 1 << bits ) - 1;
    else
        mask = ~0;

    if( data == end )
        {
        throw_end_of_stream();
        }

    result = (UInt32) *data >> bit;
    shift = 8 - bit;
    bit = ( bit + bits ) & 7;
    bits -= shift;

    while( bits > 0 )
        {
        data++;

        if( data == end )
            {
            throw_end_of_stream();
            }

        result |= (UInt32) *data << shift;
        shift += 8;
        bits -= 8;
        }

    if( !bit )
        {
        // Move to next byte
        hope( !bits );
        data++;
        }

    return result & mask;
    }  // BitIStream::read_int()


UInt64 BitIStream::read_int64
    (
    int     bits
    )
    {
    int     shift;
    UInt64  result;
    UInt64  mask;

    hope( 0 < bits );
    hope( bits <= 64 );
    hope( 0 <= bit );
    hope( bit < 8 );

    if( bits < 64 )
        mask = ( UINT64( 1 ) << bits ) - 1;
    else
        mask = ~UINT64( 0 );

    if( data == end )
        {
        throw_end_of_stream();
        }

    result = (UInt64) *data >> bit;
    shift = 8 - bit;
    bit = ( bit + bits ) & 7;
    bits -= shift;

    while( bits > 0 )
        {
        data++;

        if( data == end )
            {
            throw_end_of_stream();
            }

        result |= (UInt64) *data << shift;
        shift += 8;
        bits -= 8;
        }

    if( !bit )
        {
        // Move to next byte
        hope( !bits );
        data++;
        }

    return result & mask;
    }  // BitIStream::read_int64()


UInt32 BitIStream::read_unary()
    {
    UInt32 value = 0;
    while( read_int( 1 ) == 0 )
        {
        value++;
        }
    return value;
    }


void BitIStream::throw_out_of_range() const
    {
    throw ExOutOfRange( "Offset out of range of the BitIOStreamBuffer." );
    }


void BitIStream::throw_end_of_stream() const
    {
    throw ExUnexpectedEof( "Reading past end of stream." );
    }


} //namespace Massiv::Core
} //namespace Massiv


#ifndef ENABLE_INLINE
#include "net/bitistream_inline.h"
#endif
