/** @file bitistream.h
 * Bit-oriented input stream buffer.
 */

/*
 * begin        : Tue Dec 18 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BITISTREAM_H
#define MASSIV_BITISTREAM_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#ifndef MASSIV_BITIOSTREAM_COMMON_H
#include "net/bitiostream_common.h"
#endif

#include <string>
#include <iosfwd>


namespace Massiv {
namespace Core {

class NetIBuffer;
class BitOStream;

namespace Test { class BitIOStreamTest; }


/** @class BitIStream
 * Bit-oriented input stream buffer.
 *
 * @ingroup utility
 * @ingroup serialization
 * @ingroup net
 */
class BitIStream : public BitIOStream
    {
public:
        
    friend class Massiv::Core::Test::BitIOStreamTest;
    friend class Massiv::Core::BitOStream;
    friend class Massiv::Core::NetIBuffer;

    typedef unsigned int SizeType;

public:

    /**
     * Default constructor.
     * Binds empty input buffer.
     */
    INLINE BitIStream();

    /**
     * Constructor.
     * Binds specified input buffer.
     * @param buffer Buffer used as data source.
     * @param duplicate Specifies whether BitIStream should duplicate @a buffer.
     *  If the @a buffer is not duplicated, it must remain valid (i.e. do not call operations
     *  which can invalidate iterators) until the buffer is unbound.
     * @param offset Skip @a offset bits in the buffer. Must be byte-aligned.
     */
    INLINE BitIStream
        (
        BitIOStreamBuffer   buffer,
        bool                duplicate = false,
        SizeType            offset = 0
        );

    /**
     * Copy constructor.
     * Creates a copy of input bitstream bound to the same input buffer.
     */
    INLINE BitIStream
        (
        const BitIStream &  stream
        );

    /**
     * Destructor.
     * Unbinds input buffer.
     */
    INLINE ~BitIStream();

    /**
     * Operator=.
     * Resulting stream will share buffer with source stream, current position
     * and begin/end offsets will remain valid.
     */
    BitIStream & operator=
        (
        const BitIStream & stream
        );

public:

    /**
     * Binds specified buffer with BitIStream.
     * @param buffer Buffer used as data source.
     * @param duplicate Specifies whether BitIStream should duplicate @a buffer.
     *  If the @a buffer is not duplicated, it must remain valid (i.e. do not call operations
     *  which can invalidate iterators) until the buffer is unbound.
     * @param offset Skip @a offset bits in the buffer. Must be byte-aligned.
     * @return Previously bound buffer.
     */
    BitIOStreamBuffer bind
        (
        BitIOStreamBuffer   buffer,
        bool                duplicate = false,
        SizeType            offset = 0
        );

    /**
     * Binds part of specified buffer with BitIStream.
     * @param buffer Buffer used as data source.
     * @param duplicate Specifies whether BitIStream should duplicate @a buffer.
     *  If the @a buffer is not duplicated, it must remain valid (i.e. do not call operations
     *  which can invalidate iterators) until the buffer is unbound.
     * @param begin Skip @a bytes in the source buffer.
     * @param len use only @a bytes from the source buffer.
     * @return Previously bound buffer.
     */
    BitIOStreamBuffer bind_part
        (
        BitIOStreamBuffer   buf,
        bool                dup,
        SizeType            begin,
        SizeType            len
        );

    /**
     * Unbinds input buffer.
     * Binds empty buffer with BitIStream.
     * @return Previously bound buffer.
     */
    INLINE BitIOStreamBuffer unbind();

    /**
     * Moves the BitIOBuffer "iterator" position to the start of range valid
     * for this BitIStream.
     */
    INLINE void rewind();

    /**
     * Moves the BitIOBuffer "iterator" positon to the given value.
     * @param ofs     Offset (in bites) in the BitIOBuffer.
     * @param whence  Whence to move. WHENCE_START, WHENCE_CURRENT or
     *                WHENCE_END values are possible.
     * @exception ExOutOfRange
     * @note Seeking at the first bit after the end of the stream is allowed
     *      ie it will not throw exception (but of course trying to read
     *      a bit from the position will throw).
     *      For example if bit size of a stream is 24, following
     *      commands will seek to same position and will not throw exception: 
     *      @code
     *      stream.seek( 24, WHENCE_START );
     *      stream.seek( 0, WHENCE_END );
     *      @endcode
     */
    void seek
        (
        int         ofs,
        SeekWhence  whence
        );

    /**
     * Skips @a offset bits in the buffer.
     * @exception ExOutOfRange
     */
    INLINE void skip
        (
        SizeType    offset
        );

    /**
     * Returns shared input buffer.
     * The buffer is not unbound.
     * @note The whole buffer is returned regardless if BitIStream is bound
     * to its portion only.
     */
    INLINE BitIOStreamBuffer get_buffer();

public:

    /**
     * Read up to 32 bits of integer value.
     * @note If @bits bits cannot be all read because of the end of
     *       stream, the position at the stream stays at @end.
     * @todo Optimize out do { } while.
     */
    UInt32 read_int
        (
        int     bits
        );

    /**
     * Read up to 64 bits of integer value.
     * @note If @bits bits cannot be all read because of the end of
     *       stream, the position at the stream stays at @end.
     */
    UInt64 read_int64
        (
        int     bits
        );

    /**
     * Read unary-coded integer value (as written by BitOStream::write_unary()).
     */
    UInt32 read_unary();

    /**
     * Read 32bit floating-point value.
     * @note If @bits bits cannot be all read because of the end of
     *       stream, the position at the stream stays at @end.
     */
    Float32 read_float();

    /**
     * Read 64bit floating-point value.
     * @note If @bits bits cannot be all read because of the end of
     *       stream, the position at the stream stays at @end.
     */
    Float64 read_float64();

    /**
     * Read std string.
     * @todo Is returning std::string a good idea?
     */
    std::string read_string();

    /**
     * Read std string. Assumes each char is 8 bits.
     */
    std::string read_full_string();

    /**
     * Read raw binary data from bitstream.
     * @param length - number of bytes to copy from bitstream
     * @param buf - buffer to store data from bitstream
     */
    void read_raw
        (
        UInt8 *     buf,
        UInt32      length
        );
        
public:

    /**
     * Creates BitIStream operating just on the part of bitbuffer - from
     * the begining to the actual position aligned to bytes.
     *
     * @param duplicate If false, the new stream operates on the portion of
     *        the same instance of buffer. Otherwise, the necessary part
     *        of the buffer is duplicated.
     * @return BitIStream operating on a selected buffer / portion of buffer.
     */
    BitIStream head
        (
        bool duplicate = false
        );

    /**
     * Creates BitIStream operating just on the part of bitbuffer - from
     * the actual position aligned to bytes to the end.
     *
     * @param duplicate If false, the new stream operates on the portion of
     *        the same instance of buffer. Otherwise, the necessary part
     *        of the buffer is duplicated.
     * @return BitIStream operating on a selected buffer / portion of buffer.
     */
    BitIStream tail
        (
        bool duplicate = false
        );

public:

    /**
     * Changes starting offset of the stream in the buffer. Offset is given in
     * bits, but must be byte-aligned.
     * @param new_offset New starting offset (absolute).
     * @return Old offset in bits.
     * @note Doesn't change stream's position in the buffer and thus the position
     *       might be invalid after return from @a change_offset. Therefore
     *       user should remember to call @a seek if necessary.
     * @exception ExOutOfRange
     */
    SizeType change_offset
        (
        SizeType new_offset
        );

    /**
     * Changes the stream's end position and thus constraints the length
     * of the buffer. The position is counted from the actual begining
     * of the stream, considering @a offset member.
     * @param new_end_offset Offset of the end of stream. It is given
     *        in bits relative to current starting offset, and must be
     *        byte-aligned.
     * @return Old position of end of stream in bits (as the starting position
     *        is considered the @a offset member).
     */
    SizeType change_stream_end
        (
        SizeType new_end_offset
        );

    /**
     * Skips so many bits in the buffer, that the actual position
     * is byte-aligned.
     */
    INLINE void align_to_byte();

    /**
     * Returns length of the stream in bits.
     */
    INLINE SizeType bit_size() const;

    /**
     * Returns length of the stream in bytes.
     */
    INLINE SizeType byte_size() const;     

    /**
     * Returns the actual position in the buffer in bits. The position
     * is counted from the actual @a offset.
     */
    INLINE SizeType get_position() const;

    /**
     * Returns number of bits remaining in the buffer (from the actual
     * position of the stream).
     */
    INLINE SizeType count_remaining_bits() const;
    
private:

    /** Throw exception. */
    void throw_out_of_range() const;

    /** Throw exception. */
    void throw_end_of_stream() const;

private:

    int                         bit;
        /**< Bit to read from. Next byte is not yet allocated if @a bit is 0. */

    SizeType                    offset;
        /**< Starting position (in bits) of the BitIStream in the buffer.
             It has to be aligned to bytes (ie. multiples of 8). */

    BitIOStreamBuffer           buffer;
        /**< Input data buffer. Valid portion is specified by <data, end)
             range. */

    BitIOStreamBuffer::Iterator data;
        /**< Iterator to buffer. Points to byte to read from. */

    BitIOStreamBuffer::Iterator end;
        /**< Iterator to buffer. Points to input buffer end. */

    };  // class BitIStream

    
} // namespace Massiv::Core
} // namespace Massiv

#ifdef ENABLE_INLINE
#include "bitistream_inline.h"
#endif

#endif // MASSIV_BITISTREAM_H
