/** @file bitostream.cpp
 * Bit-oriented output stream buffer.
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/bitostream.h"
#include "net/bitistream.h"
#include <cstring>
#include <algorithm>


namespace Massiv {
namespace Core {


BitIOStreamBuffer BitOStream::bind
    (
    BitIOStreamBuffer   buf,
    bool                dup
    )
    {
    BitIOStreamBuffer   prev = buffer;

    if( dup )
        {
        BitIOStreamBuffer           b;
        BitIOStreamBuffer::Iterator si( buf.begin() );
        BitIOStreamBuffer::Iterator se( buf.end() );
        BitIOStreamBuffer::Iterator di;

        b.resize( buf.size() );
        for( di = b.begin(); si != se; si++, di++ )
            {
            (*di) = (*si);
            }

        buf = b;
        }

    bit         = 0;
    num_bits    = buf.size() * 8;
    buffer      = buf;

    return prev;
    }


void BitOStream::align_fill
    (
    UInt8   value,
    int     count
    )
    {
    align_to_byte();
    for( int i = 1; i <= count; i++ )
        {
        buffer.push_back( (UInt8) value );
        }

    num_bits += count * 8;
    }


void BitOStream::write_int
    (
    UInt32  value,
    int     bits
    )
    {
    hope( 0 < bits );
    hope( bits <= 32 );

    // Get rid of high bits.

    if( bits < 32 )
        {
        value &= ( 1 << bits ) - 1;
        }

    // Write unaligned bits.

    int shift;
    if( bit )
        {
        BitIOStreamBuffer::Iterator last( buffer.end() - 1 );
        ( *last ) |= static_cast< UInt8 >( value << bit );
        shift = 8 - bit;
        }
    else
        {
        buffer.push_back( static_cast< UInt8 >( value ) );
        shift = 8;
        }

    // Update bit pointer and counter.

    bit = ( bit + bits ) & 7;
    num_bits += bits;
    bits -= shift;

    // Write aligned bytes.

    if( bits > 0 )
        {
        value >>= shift;
        do
            {
            buffer.push_back( static_cast< UInt8 >( value ) );
            value >>= 8;
            bits -= 8;
            }
        while( bits > 0 );
        }
    }


void BitOStream::write_int64
    (
    UInt64  value,
    int     bits
    )
    {
    hope( 0 < bits );
    hope( bits <= 64 );

    // Get rid of high bits.

    if( bits < 64 )
        {
        value &= ( UINT64( 1 ) << bits ) - 1;
        }

    // Write unaligned bits.

    int shift;
    if( bit )
        {
        BitIOStreamBuffer::Iterator last( buffer.end() - 1 );
        (*last) |= static_cast< UInt8 >( value << bit );
        shift = 8 - bit;
        }
    else
        {
        buffer.push_back( static_cast< UInt8 >( value ) );
        shift = 8;
        }

    // Update bit pointer and counter.

    bit = ( bit + bits ) & 7;
    num_bits += bits;
    bits -= shift;

    // Write aligned bytes.

    if( bits > 0 )
        {
        value >>= shift;
        do
            {
            buffer.push_back( static_cast< UInt8 >( value ) );
            value >>= 8;
            bits -= 8;
            }
        while( bits > 0 );
        }
    }


void BitOStream::write_unary
    (
    UInt32 value
    )
    {
    hope( value >= 0 && value <= 64 );
    if( value > 0 )
        {
        // Write @a value zeroes.

        write_int64( 0, value );
        }
    write_int( 1, 1 );
    }


void BitOStream::write_float
    (
    Float32 value
    )
    {
    union
        {
        Float32 f;
        UInt32  i;
        } v;
    v.f = value;
    write_int( v.i, 32 );
    }


void BitOStream::write_float64
    (
    Float64 value
    )
    {
    union
        {
        Float64 f;
        UInt64  i;
        } v;
    v.f = value;
    write_int64( v.i, 64 );
    }


void BitOStream::write_string
    (
    const std::string & str
    )
    {
    for( std::string::const_iterator si = str.begin(); si != str.end(); ++si )
        {
        UInt8 c = static_cast<UInt8>( *si );
        hope( c > 0 && c < 128 );
        write_int( c, 7 );
        }

    write_int( 0, 7 );
    }


void BitOStream::write_full_string
    (
    const std::string & str
    )
    {
    write_int( str.size(), 32 );

    for( std::string::const_iterator si = str.begin(); si != str.end(); ++si )
        {
        write_int( static_cast< Int32 >( *si ), 8 );
        }
    }


void BitOStream::write_raw
    (
    UInt8 *     buf,
    UInt32      length
    )
    {
    // Must be byte aligned.
    align_to_byte();
    hope( !bit );    

    int current_size = buffer.size();
    buffer.resize( current_size + length );
    std::memcpy( ( char * ) buffer.get_buffer() + current_size, buf, length );

    num_bits = buffer.size() * 8;
    }


void BitOStream::write_bitistream
    (
    BitIStream &    istream,
    bool            byte_align
    )
    {
    int bits_to_copy = istream.count_remaining_bits();
        /* number of bits to be copied */

    if( byte_align )  bits_to_copy = ( bits_to_copy + 8 ) & ~0x07;

    int bits_in      = byte_align ? ( ( num_bits + 8 ) & ~0x07 ) : num_bits;
        /* number of bits that will precede the copied istream */

    buffer.reserve( ( bits_in + bits_to_copy ) / 8 + 1 );
    
    if( byte_align )
        {
        /* copying will be done byte-per-byte */
        BitIOStreamBuffer::Iterator   it_in( istream.data );

        bit = 0;
        num_bits = bits_in + bits_to_copy;

        while( it_in != istream.end )
            {
            buffer.push_back( *it_in++ );
            }
        }
    else
        {
        /* copying will be done on a level of bits */
        const int aux_bits = 8 - istream.bit;
        hope( aux_bits > 0 );
        hope( aux_bits <= 8 );
        write_int( istream.read_int( aux_bits ), aux_bits );
        bits_to_copy -= aux_bits;

        while( bits_to_copy > 32 )
            {
            write_int( istream.read_int( 32 ), 32 );
            bits_to_copy -= 32;
            }

        if( bits_to_copy > 0 )
            {
            hope( bits_to_copy <= 32 );
            write_int( istream.read_int( bits_to_copy ), bits_to_copy );
            }
        }

    // Move source stream position to its end.

    istream.seek( 0, WHENCE_END );
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "bitostream_inline.h"
#endif
