/** @file bitostream.h
 * Bit-oriented output stream buffer.
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BITOSTREAM_H
#define MASSIV_BITOSTREAM_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_BITIOSTREAM_COMMON_H
#include "net/bitiostream_common.h"
#endif

#include <vector>
#include <string>


namespace Massiv {
namespace Core {

namespace Test
    {
    class BitIOStreamTest;
    class SocketsTest;
    }

class BitIStream;


/** @class BitOStream
 * Bit-oriented output stream buffer.
 *
 * @ingroup utility
 * @ingroup serialization
 * @ingroup net
 */
class BitOStream : public BitIOStream
    {
    friend class Massiv::Core::Test::BitIOStreamTest;
    friend class Massiv::Core::Test::SocketsTest;

public:

    /**
     * Default constructor.
     * Binds empty output buffer with default reserved size.
     */
    INLINE BitOStream();

    /**
     * Constructor.
     * Binds empty output buffer with specified reserved size.
     */
    INLINE BitOStream
        (
        int size
        );

    /**
     * Destructor.
     */
    INLINE ~BitOStream();

public:

    /**
     * Binds specified output buffer.
     * The @a buffer is not cleared, data is being added at the @a buffer back.
     * @param buffer Buffer used as data output.
     * @param duplicate Specifies whether BitOStream should duplicate @a buffer.
     * @return Previously bound buffer.
     */
    BitIOStreamBuffer bind
        (
        BitIOStreamBuffer buffer,
        bool              duplicate = false
        );

    /**
     * Unbinds current output buffer and binds new empty one with
     * default reserved size.
     * @return Previously bound buffer.
     */
    INLINE BitIOStreamBuffer unbind();

    /**
     * Unbinds current output buffer and binds new empty one with
     * default reserved size.
     * @return Previously bound buffer.
     */
    INLINE BitIOStreamBuffer reset();

    /**
     * Unbinds current output buffer and binds new empty one with
     * specified reserved size.
     * @return Previously bound buffer.
     */
    INLINE BitIOStreamBuffer reset
        (
        int size
        );

    /**
     * Returns shared output buffer.
     * The buffer is not unbound.
     */
    INLINE BitIOStreamBuffer get_buffer();

public:

    /**
     * First aligns buffer to next byte
     * then fills buffer with integer value count-times.
     * @todo Another values ???
     */
    void align_fill
        (
        UInt8   value,
        int     count
        );

    /**
     * Aligns buffer to next byte - adds padding of at most 7 zero bits.
     */
    INLINE void align_to_byte();

    /**
     * Write up to 32 bits of integer value.
     * @todo Optimize out do { } while.
     */
    void write_int
        (
        UInt32  value,
        int     bits
        );

    /**
     * Write up to 64 bits of integer value.
     */
    void write_int64
        (
        UInt64  value,
        int     bits
        );

    /**
     * Write @a value using simple unary coding: @a value zero bits followed
     * by single one bit, e.g. number 4 is written "00001".
     * @a Value must be less than or equal to 64.
     */
    void write_unary
        (
        UInt32 value
        );

    /**
     * Write 32bit floating-point value.
     */
    void write_float
        (
        Float32 value
        );

    /**
     * Write 64bit floating-point value.
     */
    void write_float64
        (
        Float64 value
        );

    /**
     * Write std string.
     * Each char consumes 7 bits! Terminated by '\0' (7bits), which
     * implies that @c str must not contain '\0'.
     * @bug Assumes US-ASCII encoding. Highest bit should never be set.
     * @todo Transfer eight bit? Support non-US-ASCII encodings?
     */
    void write_string
        (
        const std::string & str
        );

    /**
     * Write std string.
     * Each char consumes 8 bits!
     * @todo String length is currently encoded using 32-bits, this might
     * be ineffecient. Rewrite.
     */
    void write_full_string
        (
        const std::string & str
        );

    /**
     * Write raw binary data to bitstream.
     * @param length - number of bytes to copy to bitstream
     * @param buf - buffer from store data to bitstream
     */
    void write_raw
        (
        UInt8 *     buf,
        UInt32      length
        );

    /**
     * Writes istream from the actual position to the end. Position of
     * the source istream will be moved to its end.
     * @param byte_align If true, the istream is copied "byte-aligned".
     *        It means that the byte pointed to by @a data iterator is
     *        copied first, and, moreover, the ostream is aligned to
     *        bytes before the copy is done.
     */
    void write_bitistream
        (
        BitIStream &    istream,
        bool            byte_align = true
        );

public:

    /**
     * Get length of buffer (in bits).
     */
    INLINE int bit_size() const;

    /**
     * Get length of buffer (in bytes).
     */
    INLINE int byte_size() const;

private:

    enum { DEFAULT_STRBUF_SIZE = 1024 };
        /**< Default initial stream buffer size. */

    int                 bit;
        /**< Bit to write to. Next byte is not yet allocated if @a bit is 0.
             It is bit in a byte, NOT bit in a buffer. */

    int                 num_bits;
        /**< Number of bits written. */

protected:

    BitIOStreamBuffer   buffer;
        /**< Output data buffer. */

    };  // class BitOStream


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "net/bitostream_inline.h"
#endif


#endif // MASSIV_BITOSTREAM_H
