/** @file block_cipher.h
 * Block cipher.
 */

/*
 * begin        : Sun Apr 06 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_BLOCK_CIPHER_H
#define MASSIV_BLOCK_CIPHER_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup net
 * @{
 */

/**
 * Block cipher.
 * Base class for ciphers operating on equi-size blocks of input and output.
 * Size of input block is equal to the size of output block.
 * @c BlockCipher is treated as an abstract class, whose implementation
 * is defined at @c Implementation. @c Implementation is supposed
 * to be derived from @c BlockCipher.
 * The inteface for @c Implementation looks like this:
 * @code
 * // Processes @c input block and writes output to @c output block.
 * void process_block
 *      (
 *      const InputElement *    input,
 *      OutputElement *         output
 *      );
 * @endcode
 * @c Implementation parameter allows us to call implementation
 * methods directly without the need for defining them as pure
 * virtual methods at @c BlockCipher.
 */
template
    <
    int     block_size,
    class   Implementation,
    class   InputElement = UInt8,
    class   OutputElement = InputElement
    >
class BlockCipher
    {
public:

    /**
     * Processes stream of data.
     * Note! There must be enough space in the output container.
     * Note! first - last must be block_size-aligned!
     */
    template <class InputIterator, class OutputIterator>
    void process_range
        (
        InputIterator   first,
        InputIterator   last,
        OutputIterator  output
        )
        {
        do
            {
            int todo = last - first;
            int i;

            if( todo <= 0 )
                break;

            hope( todo % block_size == 0 );

            /* Read block from input container. */

            for( i = 0; i < block_size; i++ )
                {
                in_block[i] = *first;
                first++;
                }

            /* Process block. */

            downcast().process_block( in_block, out_block );

            /* Write processed block to output container. */

            for( i = 0; i < block_size; i++ )
                {
                *output = out_block[i];
                output++;
                }
            }
        while( true );
        }

    /**
     * Processes stream of data.
     * Optimized version.
     * Note! There must be enough space in the output container.
     * Note! first - last must be block_size-aligned!
     */
    void process_range
        (
        const InputElement *    first,
        const InputElement *    last,
        OutputElement *         output
        )
        {
        do
            {
            int todo = last - first;

            if( todo <= 0 )
                break;

            hope( todo % block_size == 0 );

            /* Process block. */

            downcast().process_block( first, output );

            first += block_size;
            output += block_size;
            }
        while( true );
        }

    /**
     * Processes stream of data and places the result to
     * the end of the output container.
     * Note! first - last must be block_size-aligned!
     */
    template <class InputIterator, class OutputContainer>
    void process_and_append_range
        (
        InputIterator       first,
        InputIterator       last,
        OutputContainer &   output
        )
        {
        int todo = last - first;

        output.resize( output.size() + todo );
        process_range( first, last, output.end() - todo );
        }

    /**
     * Processes all content of intput container
     * and places it to the end of the output container.
     * Note! input.size() must be block_size-aligned!
     */
    template <class InputContainer, class OutputContainer>
    void process_and_append
        (
        InputContainer      input,
        OutputContainer &   output
        )
        {
        process_and_append_range( input.begin(), input.end(), output );
        }

    /**
     * Processes all content of intput container
     * and places it to the output container.
     * Note! input.size() must be block_size-aligned!
     */
    template <class InputContainer, class OutputContainer>
    void process
        (
        InputContainer &    input,
        OutputContainer &   output
        )
        {
        output.clear();
        process_and_append( input, output );
        }

    /**
     * Aligns size of the container so that it could
     * be used by process(), process_and_append().
     */
    template <class InputContainer>
    void align
        (
        InputContainer &    input
        )
        {
        int size = input.size();
        int delta = size % block_size;

        if( delta )
            input.resize( size + block_size - delta );
        }

private:

    /***
     * Returns reference to ancestor of @c BlockCipher that
     * implements @c BlockCipher interface.
     */
    Implementation & downcast()
        {
        return static_cast<Implementation &>( *this );
        }

protected:

    InputElement    in_block[ block_size ];
        /**< Input block. */

    OutputElement   out_block[ block_size ];
        /**< Output block. */

    }; // BlockCipher

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_BLOCK_CIPHER_H
