/** @file connection_info.cpp
 * Keeps info about created connections.
 */

/*
 * begin        : Tue Jul 30 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/connection_info_impl.h"
#include "utility/debug.h"
#include "utility/quote_string.h"
#include <sstream>

namespace Massiv {
namespace Core {


ConnectionInfo::ConnectionInfo
    (
    bool server_side
    ) : connection_status( NOT_CONNECTED ),
        inform_node_manager( false ),
        disconnection_reason( NOT_SPECIFIED_REASON ),
        tcpsocket( NULL ),
        udpsocket( NULL ),
        benc_odd( NULL ),
        bdec_odd( NULL ),
        benc_even( NULL ),
        bdec_even( NULL ),
        blowfish_key_timestamp( SystemTime::INVALID ),
        blowfish_key_period( BLOWFISH_PERIOD_ODD ),
        server_side( server_side ),
        idle_ticks( 0 ),
        last_message_time( SystemTime::INVALID ),
        ping_messages_sent( 0 ),
        node_id(), 
        permit_auto_close( true ),
        udp_connected( false ),
        tcp_confirmed( false )
    {
    }  // ConnectionInfo::ConnectionInfo()


ConnectionInfo::~ConnectionInfo()
    {
    delete udpsocket;
    delete tcpsocket;
    delete benc_odd;
    delete bdec_odd;
    delete benc_even;
    delete bdec_even;
    }  // ConnectionInfo::~ConnectionInfo()


std::string ConnectionInfo::to_string() const
    {
    std::ostringstream ss;

    ss << "node_id = " << node_id << '\n';
    ss << "connection_status = ";

    switch( connection_status )
        {
        case NOT_CONNECTED:
            ss << "NOT_CONNECTED\n";
            break;

        case CONNECTED:
            ss << "CONNECTED\n";
            break;

        case NODE_ID_EXCHANGE_SERVER:
            ss << "NODE_ID_EXCHANGE_SERVER\n";
            break;

        case BLOWFISH_EXCHANGE_SERVER:
            ss << "BLOWFISH_EXCHANGE_SERVER\n";
            break;

        case BLOWFISH_EXCHANGE_CLIENT:
            ss << "BLOWFISH_EXCHANGE_CLIENT\n";
            break;

        case UDP_PORT_RECEIVE_CLIENT:
            ss << "UDP_PORT_RECEIVE_CLIENT\n";
            break;

        case UDP_CONNECTION_SETUP_SERVER:
            ss << "UDP_CONNECTION_SETUP_SERVER\n";
            break;

        case TRYING_TO_DISCONNECT:
            ss << "TRYING_TO_DISCONNECT\n";
            break;

        default:
            panic( "Invalid connection_status." );
        }

    ss << "disconnection_reason = ";

    switch( disconnection_reason )
        {
        case NOT_SPECIFIED_REASON:
            ss << "NOT_SPECIFIED_REASON\n";
            break;

        case REQUEST_REASON:
            ss << "REQUEST_REASON\n";
            break;
        
        case ERROR_REASON:
            ss << "ERROR_REASON\n";
            break;

        default:
            panic( "Invalid disconnection_reason." );
        }

    ss << "inform_node_manager = " << inform_node_manager << '\n';
    ss << "permit_auto_close = " << permit_auto_close << '\n';
    ss << "tcp_confirmed = " << tcp_confirmed;

    return ss.str();
    }


void ConnectionInfo::set_connection_status
    ( 
    ConnectionStatus status 
    )
    {
    if( ( connection_status != CONNECTED ) && ( status == CONNECTED ) )
        {
        // The connection has just been established.
   
        blowfish_key_period = BLOWFISH_PERIOD_ODD;
        blowfish_key_even = blowfish_key_odd;
        blowfish_key_timestamp = Global::system_time();
        touch_blowfish_key_timestamp();
        inform_node_manager = true;
        }
    
    connection_status = status;
    }  // ConnectionInfo::set_connection_status()

    
ConnectionInfo::ConnectionStatus ConnectionInfo::get_connection_status()
    {
    return connection_status;
    }  // ConnectionInfo::get_connection_status()


bool ConnectionInfo::should_inform_node_manager()
    {
    return inform_node_manager;
    }  // ConnectionInfo::should_inform_node_manager()


void ConnectionInfo::set_inform_node_manager
    (
    bool inform
    )
    {
    inform_node_manager = false;
    } // ConnectionInfo::set_inform_node_manager()


void ConnectionInfo::set_disconnection_reason
    ( 
    DisconnectionReason reason
    )
    {
    disconnection_reason = reason;
    }  // ConnectionInfo::set_disconnection_reason()

    
ConnectionInfo::DisconnectionReason ConnectionInfo::get_disconnection_reason()
    {
    return disconnection_reason;
    }  // ConnectionInfo::get_disconnection_reason()



BlowfishEncryptor * ConnectionInfo::get_blowfish_encryptor_for_udp()
    {
    BlowfishEncryptor * result;

    switch( blowfish_key_period )
        {
        case BLOWFISH_PERIOD_ODD:
        case BLOWFISH_PERIOD_EVEN_TRANSITION:
            result = get_blowfish_encryptor_odd();
            break;

        case BLOWFISH_PERIOD_EVEN:
        case BLOWFISH_PERIOD_ODD_TRANSITION:
            result = get_blowfish_encryptor_even();
            break;

        default:
            panic( "Unexpected blowfish resend algorithm period." );
            break;
        }

    return result;
    }  // ConnectionInfo::get_blowfish_encryptor_for_udp()


BlowfishDecryptor * ConnectionInfo::get_blowfish_decryptor_for_udp
    (
    BlowfishKeyPeriod   period
    )
    {
    BlowfishDecryptor * result;

    if( period == BLOWFISH_PERIOD_ODD )
        {
        result = get_blowfish_decryptor_odd();
        }
    else
    if( period == BLOWFISH_PERIOD_EVEN )
        {
        result = get_blowfish_decryptor_even();
        }
    else
        {
        panic( "Unexpected blowfish resend algorithm phase in UDP packet." );
        }

    return result;
    }  // ConnectionInfo::get_blowfish_decryptor_for_udp()


BlowfishEncryptor * ConnectionInfo::get_blowfish_encryptor_for_tcp()
    {
    BlowfishEncryptor * result = NULL;

    switch( blowfish_key_period )
        {
        case BLOWFISH_PERIOD_ODD:
        case BLOWFISH_PERIOD_EVEN_TRANSITION:
            result = get_blowfish_encryptor_odd();
            break;

        case BLOWFISH_PERIOD_EVEN:
        case BLOWFISH_PERIOD_ODD_TRANSITION:
            result = get_blowfish_encryptor_even();
            break;

        default:
            /* Invalid blowfish period. */
            hope( 0 );
        }

    return result;
    }  // ConnectionInfo::get_blowfish_encryptor_for_tcp()


BlowfishDecryptor * ConnectionInfo::get_blowfish_decryptor_for_tcp
    (
    BlowfishKeyPeriod   period
    )
    {
    BlowfishDecryptor * result = NULL;

    switch( period )
        {
        case BLOWFISH_PERIOD_ODD:
        case BLOWFISH_PERIOD_EVEN_TRANSITION:
            result = get_blowfish_decryptor_odd();
            break;

        case BLOWFISH_PERIOD_EVEN:
        case BLOWFISH_PERIOD_ODD_TRANSITION:
            result = get_blowfish_decryptor_even();
            break;

        default:
            /* Invalid blowfish period. */
            hope( 0 );
            break;
        }

    return result;
    }  // ConnectionInfo::get_blowfish_decryptor_for_tcp()


void ConnectionInfo::set_blowfish_key
    (
    const std::string & blowfish_key
    )
    {
    // Log setting of new blowfish key
    // FIXME: Is it sensible to write the key into log?

    std::ostringstream  ss;
    ss << "Setting blowfish new key: " << quote_const_string( blowfish_key )
       << " (period " << blowfish_key_period << ")";

    Global::log
        (
        STATUS_DEBUG( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        ss.str()
        );

    // Set the key

    switch( blowfish_key_period )
        {
        case BLOWFISH_PERIOD_ODD:
            set_blowfish_key_even( blowfish_key );
            break;

        case BLOWFISH_PERIOD_EVEN:
            set_blowfish_key_odd( blowfish_key );
            break;

        case BLOWFISH_PERIOD_ODD_TRANSITION:
        case BLOWFISH_PERIOD_EVEN_TRANSITION:
            panic( "Changing blowfish key during transitional period" );
            break;

        default:
            panic( "Invalid blowfish key phase." );
            break;
        }
    }  // ConnectionInfo::set_blowfish_key()


void ConnectionInfo::set_blowfish_key_odd
    (
    const std::string & key_odd
    )
    {
    blowfish_key_odd = key_odd;

    set_blowfish_odd(
        new BlowfishEncryptor( blowfish_key_odd ),
        new BlowfishDecryptor( blowfish_key_odd )
        );
    }  // ConnectionInfo::set_blowfish_key_odd()


void ConnectionInfo::set_blowfish_key_even
    (
    const std::string & key_even
    )
    {
    blowfish_key_even = key_even;

    set_blowfish_even(
        new BlowfishEncryptor( blowfish_key_even ),
        new BlowfishDecryptor( blowfish_key_even )
        );
    }  // ConnectionInfo::set_blowfish_key_even()


void ConnectionInfo::set_blowfish_key_period
    (
    BlowfishKeyPeriod   blowfish_key_period
    )
    {
    std::ostringstream  ss;
    ss << "Setting blowfish key period to " << blowfish_key_period;

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
        ss.str()
        );

    this->blowfish_key_period = blowfish_key_period;
    }  // ConnectionInfo::set_blowfish_key_period()


void ConnectionInfo::set_blowfish
    ( 
    BlowfishEncryptor * blenc, 
    BlowfishDecryptor * bldec 
    )
    {
    switch( blowfish_key_period )
        {
        case BLOWFISH_PERIOD_ODD:
            set_blowfish_even( blenc, bldec );
            break;

        case BLOWFISH_PERIOD_EVEN:
            set_blowfish_odd( blenc, bldec );
            break;

        case BLOWFISH_PERIOD_ODD_TRANSITION:
        case BLOWFISH_PERIOD_EVEN_TRANSITION:
            panic( "Changing blowfish key during transitional period" );
            break;

        default:
            panic( "Invalid blowfish key phase." );
            break;
        }
    }  // ConnectionInfo::set_blowfish()


void ConnectionInfo::set_blowfish_odd
    (
    BlowfishEncryptor * blenc,
    BlowfishDecryptor * bldec
    )
    {
    if( benc_odd )  delete benc_odd;
    if( bdec_odd )  delete bdec_odd;

    benc_odd = blenc;
    bdec_odd = bldec;
    }  // ConnectionInfo::set_blowfish_odd()


void ConnectionInfo::set_blowfish_even
    (
    BlowfishEncryptor * blenc,
    BlowfishDecryptor * bldec
    )
    {
    if( benc_even ) delete benc_even;
    if( bdec_even ) delete bdec_even;

    benc_even = blenc;
    bdec_even = bldec;
    }  // ConnectionInfo::set_blowfish_even()


} // namespace Massiv::Core
} // namespace Massiv
