/** @file connection_info.h
 * Connection information class.
 */

/*
 * begin        : Tue Jul 30 2002
 * copyright    : (C) 2002 by Ondrej Pecta, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CONNECTION_INFO_H
#define MASSIV_CONNECTION_INFO_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_MASSIV_BLOWFISH_H
#include "net/massiv_blowfish.h"
#endif

#ifndef MASSIV_NETIBUFFER_H
#include "net/netibuffer.h"
#endif

#ifndef MASSIV_SOCKET_H
#include "net/socket.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#include <list>

namespace Massiv {
namespace Core {

namespace Test { class SocketsTest; }

class Network;

/** @addtogroup net
 * @{
 */

/**
 * Contains information about a network connection.
 */
class ConnectionInfo
    {
    friend class SocketsTest;

public:    

    /**
     * Constructor.
     * @param server_side true if this is a server_side of the connection,
     *      otherwise false.
     */
    explicit ConnectionInfo
        (
        bool    server_side
        );

    /**
     * Destructor.
     */
    ~ConnectionInfo();

    /**
     * Dumps connection info state.
     */
    std::string to_string() const;

public:    
    
    /**
     * Status of connection;
     */
    enum ConnectionStatus
        {
        NOT_CONNECTED               = 0,
            /**< Disconnected and new connection is NOT being established. */

        CONNECTED                   = 1,
            /**< Connection is established, both parties can communicate
                 in a "normal" way. */

        NODE_ID_EXCHANGE_SERVER     = 2,
            /**< Server is waiting for the client NodeId. This state
                 can emerge only on server. */

        BLOWFISH_EXCHANGE_SERVER    = 3,
            /**< Server is waiting for the client blowfish key. This state 
                 can emerge only on server and only after completing the
                 RSA keys exchange (because blowfish key is being sent
                 RSA-encrypted). */

        BLOWFISH_EXCHANGE_CLIENT    = 4,
            /**< Client is waiting for the server blowfish key. This state
                 can emerge only on client and only after completing the 
                 RSA keys echange (because blowfish key is being sent
                 RSA-encrypted). */

        UDP_PORT_RECEIVE_CLIENT     = 5,
            /**< Client is waiting for udp port assigned by the server.
                 udp port is crypted with new blowfish.  */

        UDP_CONNECTION_SETUP_SERVER = 6, 
            /**< Server is waiting for udp message from the client.
                 This initializes the UDP connection.  */

        TRYING_TO_DISCONNECT        = 7 
            /**< I am trying to disconnect from some node.  */

        };  // enum ConnectionInfo::ConnectionStatus

    /**
     * Reason of connection;
     */
    enum DisconnectionReason
        {
        NOT_SPECIFIED_REASON    = 0,
            /**< Reason is not specified. */

        REQUEST_REASON          = 1,
            /**< There was a request to close the connection. */
        
        ERROR_REASON            = 2
            /**< There was an error and connnection must be closed. */

        };  // enum ConnectionInfo::DisconnectionReason

public:

    /**
     * Blowfish key recomputation and resend algorithm phase.
     * @note It is important not to change values assigned to specific
     *      names!!!
     */
    enum BlowfishKeyPeriod
        {
        BLOWFISH_PERIOD_ODD                 = 0,
            /**< Odd period; this is even the initial phase, set up
                 immediately after establishing a connection. */

        BLOWFISH_PERIOD_EVEN                = 1,
            /**< Even period. */

        BLOWFISH_PERIOD_EVEN_TRANSITION     = 2,
            /**< This period can arise only on a server. It is a
                 transitional phase between BLOWFISH_PERIOD_ODD and
                 BLOWFISH_PERIOD_EVEN. Server in 
                 BLOWFISH_PERIOD_EVEN_TRANSITION phase has just sent
                 the new key to client and is waiting for the
                 acknowledgement. */

        BLOWFISH_PERIOD_ODD_TRANSITION      = 3,
            /**< This period can arise only on a server. It is a 
                 transitional phase between BLOWFISH_PERIOD_EVEN and
                 BLOWFISH_PERIOD_ODD. Server in
                 BLOWFISH_PERIOD_ODD_TRANSITION phase has just sent
                 the new key to client and is waiting for the
                 acknowledgement. */
        };  // enum BlowfishKeyPeriod

public:
    // Set & get methods for accessing private members.

    /**
     * Sets the connection status.
     */    
    void set_connection_status
        ( 
        ConnectionStatus status 
        );
    
    /**
     * @return the connection status.
     */
    ConnectionStatus get_connection_status();
   
    /**
    * @return true when node manager should be informed.
    */
    bool should_inform_node_manager();
   
    /**
    * Setter fot boolean inform_node_manager.
    */
    void set_inform_node_manager
        (
        bool inform
        );

    /**
     * Sets the disconnection reason.
     */    
    void set_disconnection_reason
        ( 
        DisconnectionReason reason
        );
    
    /**
     * @return the disconnectionreason.
     */
    DisconnectionReason get_disconnection_reason();

    /**
     * Sets a TCP socket to be used by this connection.
     */ 
    inline void set_tcpsocket
        (
        Socket * socket
        );
       
    /**
     * Sets an UDP socket to be used by this connection.
     */  
    inline void set_udpsocket
        (
        Socket * socket 
        );

    /**
     * @return TCP socket bound with this connection.
     */
    inline Socket * get_tcpsocket();
    
    /**
     * @return UDP socket bound with this connection.
     */
    inline Socket * get_udpsocket();

    /**
     * @return Blowfish encryptor bound with this connection that should
     *      be used for encrypting UDP messages according to the actual 
     *      phase of the blowfish key resend algorithm.
     */
    BlowfishEncryptor * get_blowfish_encryptor_for_udp();

    /**
     * @param period The blowfish key period flag from the message header.
     * @return Blowfish decryptor bound with this connection that should
     *      be used for decrypting UDP messages according to the actual 
     *      phase of the blowfish key resend algorithm and according to 
     *      the blowfish period flag in the message header.
     */
    BlowfishDecryptor * get_blowfish_decryptor_for_udp
        (
        BlowfishKeyPeriod   period
        );

    /**
     * @return Blowfish encryptor that should be used for encrypting TCP
     *      messages. 
     * @note Using the returned encryptor for UDP messages can cause that
     *      the other party won't be able to decrypt the message.
     */
    BlowfishEncryptor * get_blowfish_encryptor_for_tcp();

    /**
     * @return Blowfish decryptor that should be used for decrypting TCP
     *      messages.
     * @param period Blowfish period of the TCP message.
     * @note Using the returned decryptor for TCP messages can yield
     *      wrong result.
     */
    BlowfishDecryptor * get_blowfish_decryptor_for_tcp
        (
        BlowfishKeyPeriod   period
        );

    /**
     * Stores a RSA public key of the other party.
     */
    inline void set_other_side_public_key
        (
        const std::string & public_key 
        );
    
    /**
     * @return The other party RSA public key.
     */
    inline std::string & get_other_side_public_key();

    /**
     * Sets blowfish key according to the actual blowfish resend phase
     * (odd/even). For the odd phase sets the key for the prospective
     * even phase and vice versa.
     * @param blowfish_key The blowfish key to be set.
     */
    void set_blowfish_key
        (
        const std::string & blowfish_key
        );

    /**
     * Sets blowfish key to be used by this connection in the
     * BLOWFISH_PERIOD_ODD phase.
     * @param key_odd The odd-phase blowfish key.
     */
    void set_blowfish_key_odd
        (
        const std::string & key_odd
        );


    /**
     * Sets blowfish key to be used by this connection in the 
     * BLOWFISH_PERIOD_EVEN phase.
     * @param key_even The even-phase blowfish key.
     */
    void set_blowfish_key_even
        (
        const std::string & key_even
        );

    /**
     * @return Actual blowfish key for the connection in odd period.
     */
    inline std::string & get_blowfish_key_odd();

    /**
     * @return Actual blowfish key for the connection in even period.
     */
    inline std::string & get_blowfish_key_even();

    /**
     * Increments the PING_MESSAGEs counter.
     */
    inline void inc_ping_messages_sent();

    /**
     * Sets the @a server_side attribute.
     */
    inline void set_server_side
        (
        bool        server_side
        );

    /**
     * @return true if the connection_info is being used by server, otherwise 
     *         false.
     */
    inline bool is_server_side();

    /**
     * Increments the idle ticks counter.
     */
    inline void inc_idle_ticks();

    /**
     * Sets the idle ticks counter to zero. 
     */
    inline void reset_idle_ticks();

    /**
     * @return Actual value of the PING_MESSAGEs counter.
     */
    inline unsigned int get_ping_messages_sent();

    /**
     * @return Actual value of the idle ticks counter.
     */
    inline int get_idle_ticks();

    /**
     * Sets the timestamp of the last message received from this
     * connection to the current time.
     */
    inline void touch_last_message_time();

    /**
     * @return Time of reception of the last message from this connection.
     */
    inline SystemTime get_last_message_time();

    /**
     * @return Actual value of @a max_idle_seconds
     */
    inline unsigned int get_max_idle_seconds();

    /**
     * @return Number of idle ticks (since receiving of the last 
     *      message on the connection).
     */
    inline SystemTime get_idle_seconds();

    /**
     * @return Actual value of @a next_ping_delay.
     */
    inline unsigned int get_next_ping_delay();

    /**
     * @return Actual value of @a max_pings.
     */
    inline unsigned int get_max_pings();

    // Blowfish recomputation and resend algorithm routines
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    /**
     * Sets the timestamp of the blowfish key to the current time.
     */
    inline void touch_blowfish_key_timestamp();

    /**
     * @return Timestamp of the current blowfish key.
     */
    inline SystemTime get_blowfish_key_timestamp();

    /**
     * Sets the blowfish key timestamp.
     * @note Being used for debugging.
     */
    inline void set_blowfish_key_timestamp
        (
        SystemTime  timestamp_new
        );

    /**
     * Sets the @a blowfish_key_period attribute.
     */
    void set_blowfish_key_period
        (
        BlowfishKeyPeriod   blowfish_key_period
        );

    /**
     * @return The actual blowfish resend algorithm period.
     */
    inline BlowfishKeyPeriod get_blowfish_key_period();

    /**
     * @return Blowfish encryptor for the BLOWFISH_PERIOD_ODD phase
     *      bound with this connection.
     */
    inline BlowfishEncryptor * get_blowfish_encryptor_odd();

    /**
     * @return Blowfish decryptor for the BLOWFISH_PERIOD_ODD phase
     *      bound with this connection.
     */
    inline BlowfishDecryptor * get_blowfish_decryptor_odd();

    /**
     * @return Blowfish encryptor for the BLOWFISH_PERIOD_EVEN phase
     *      bound with this connection.
     */
    inline BlowfishEncryptor * get_blowfish_encryptor_even();

    /**
     * @return Blowfish decryptor for the BLOWFISH_PERIOD_EVEN phase
     *      bound with this connection.
     */
    inline BlowfishDecryptor * get_blowfish_decryptor_even();

    /**
     * @return node id I am connected to.
     */
    inline const NodeId & get_node_id() const;

    /**
     * Set node id I am connected to.
     */
    inline void set_node_id
        (
        NodeId node
        );

    /**
     * @return permit of auto close feature.
     */
    inline bool get_permit_auto_close();

    /**
     * Control auto close feature on this connection.
     */
    inline void set_permit_auto_close
        (
        bool auto_close
        );

    /**
     * Controls UDP connected status.
     */
    inline void set_udp_connected
        (
        bool connected
        );
    
    /**
     * Return @true when UDP connected.
     */    
    inline bool is_udp_connected();

    /**
     * Controls TCP confirmation status.
     */
    inline void set_tcp_confirmed
        (
        bool confirmed
        );
    
    /**
     * Return @true when TCP confirmed.
     */    
    inline bool is_tcp_confirmed();

private:
    
    /**
     * Sets a blowfish encryptor and decryptor according to the actual phase
     * (odd/even) for this connection. In the odd phase sets 
     * decryptor/encryptor for the prospective even phase and vice versa.
     * @param blenc Blowfish encryptor.
     * @param bldec Blowfish decryptor.
     */
    void set_blowfish
        (
        BlowfishEncryptor * blenc,
        BlowfishDecryptor * bldec
        );

    /**
     * Sets a blowfish encryptor and decryptor for the odd phase for this
     * connection.
     * @param blenc Blowfish encryptor.
     * @param bldec Blowfish decryptor.
     */
    void set_blowfish_odd
        (
        BlowfishEncryptor * blenc,
        BlowfishDecryptor * bldec
        );

    /**
     * Sets a blowfish encryptor and decryptor for the even phase for this
     * connection.
     * @param blenc Blowfish encryptor.
     * @param bldec Blowfish decryptor.
     */
    void set_blowfish_even
        (
        BlowfishEncryptor * blenc,
        BlowfishDecryptor * bldec
        );

private:
    
    ConnectionStatus        connection_status;
        /**< Connection status. */

    bool                    inform_node_manager;
        /**< Whether to inform node manager or not. */
    
    DisconnectionReason     disconnection_reason;
        /**< Disconnection reason. */

    Socket *                tcpsocket;
        /**< TCP socket. */
    
    Socket *                udpsocket;
        /**< UDP socket. */
    
    BlowfishEncryptor *     benc_odd;
        /**< Blowfish encryptor class that will be used for scrambling
             messages in BLOWFISH_PERIOD_ODD period of the connection. */
    
    BlowfishDecryptor *     bdec_odd;
        /**< Blowfish decryptor class to be used in BLOWFISH_PERIOD_ODD
             period of the connection. */

    BlowfishEncryptor *     benc_even;
        /**< Blowfish encryptor class that will be used for scrambling
             messages in BLOWFISH_PERIOD_EVEN period of the connection. */

    BlowfishDecryptor *     bdec_even;
        /**< Blowfish decryptor class to be used in BLOWFISH_PERIOD_EVEN
             period of the connection. */
    
    std::string             other_side_public_key;
        /**< RSA public key of the other party. */
 
    std::string             blowfish_key_odd;
        /**< Blowfish key to be used by this connection in the odd phase. */

    std::string             blowfish_key_even;
        /**< Blowfish key to be used by this connection in the even phase. */

    SystemTime              blowfish_key_timestamp;
        /**< Time when the current blowfish key was generated and 
             we began to measure the new blowfish key timeout. */

    BlowfishKeyPeriod       blowfish_key_period;
        /**< Blowfish key recomputation and resend algorithm phase. */    
    
    bool                    server_side;
        /**< true if the node using this connection_info is the server
             for this connection. */
    
    int                     idle_ticks;
        /**< Idle ticks counter. It is periodicaly increased when nothing has
             arived. After receiving a message, it is set to zero. <br>
             @fixme Is there a reason for not removing @a idle_ticks? Maybe
                because of some statictics and tests? */

    SystemTime              last_message_time;
        /**< A time when the last message has been received on this
             connection.<br>
             This value is being used for monitoring whether the connection
             is idle or not - after some specified (via registry settings)
             time a ping message will be sent to find out whether the next
             party is still alive or not. */

    unsigned int            ping_messages_sent;
        /**< PING_MESSAGE is sent after idle_ticks value has reached 
             MAX_IDLE_TICKS. Its purpose is to check whether the other party
             is still alive. @a ping_messages_sent counts ping messages sent
             without any reply. When move than 3 PING_MESSAGEs has been sent
             without reply, the net module assumes that the connection has
             been broken. */

    NodeId                  node_id;
        /**< Who Am I connected to. Used mainly in connection creating. */

    bool                    permit_auto_close;
        /**< Whether this connection have auto connection close feature
             enabled or not. */

    bool                    udp_connected;
        /**< Whether UDP connected or not. */
    
    bool                    tcp_confirmed;
        /**< Whether TCP connection confirmed or not. */

    };  // class ConnectionInfo

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_CONNECTION_INFO_H
