/** @file massiv_checksum.h
 * CRC32 checksum, implemented using CryptoPP.
 */

/*
 * begin        : Fri Oct 11 2002
 * copyright    : (C) 2002 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_MASSIV_CHECKSUM_H
#define MASSIV_MASSIV_CHECKSUM_H

// Massiv

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_BITIOSTREAM_COMMON_H
#include "net/bitiostream_common.h"
#endif

// CryptoPP

#include "net/cryptopp_begin.h"
#include "cryptopp/crc.h"
#include "net/cryptopp_end.h"


namespace Massiv {
namespace Core {

namespace Test { class SocketsTest; }


/**
 * Class implementing the CRC32.
 * Usage of this checksum implementation requires the buffer to be 
 * byte-aligned.
 *
 * @a BufferType specifies a type of buffer the checksum should be computed
 * for. It should be some form of container derived from Array. At least,
 * it must implement size() and get_buffer() methods.
 *
 * @see Array
 *
 * @ingroup utility
 * @ingroup net
 */
template< class BufferType >
class CRC32
    {

public:

    typedef UInt32      ChecksumType;

public:

    /**
     * Constructor.
     *
     * @param buffer Specifies the buffer that will be bound with this
     *      checksum class.
     */
    CRC32
        ( 
        BufferType *    buffer 
        );

    /**
     * Copy constructor.
     */
    CRC32
        (
        const CRC32< BufferType > & source
        );

    /**
     * Destructor.
     */
    ~CRC32();

private:
    
    /**
     * Creates and initializes the 'digest' buffer. It will be used to
     * retrieve checksums from the CryptoPP CRC32 implementation.
     * 
     * @note Must not be called more than once!
     */
    void init_digest();

public:

    /**
     * Computes CRC32 for the whole buffer.
     * 
     * @return The relevant CRC32 checksum.
     */
    ChecksumType compute_checksum();

    /**
     * Checks whether the CRC32 checksum corresponds to the data in the
     * buffer. 
     *
     * @param crc The CRC32 value that is assumed to be correct for this
     *      buffer. It supposes the same buffer contents as there was at 
     *      the moment of computing the CRC first time (i.e. CRC has NOT 
     *      been appended to the buffer).
     *
     * @return true if the checksum is OK, false otherwise.
     */
    bool validate_checksum
        (
        ChecksumType        crc
        );

    /**
     * Checks whether the CRC32 checksum corresponds to data in 
     * the buffer. It supposes that the checksum is written in 4 last 
     * bytes of the buffer in little endian byte order (i.e. the least
     * significant byte is the first stored). 
     *
     * @return true if the checksum is OK, false otherwise.
     *
     * @note This routine doesn't remove the checksum value from the buffer!!
     * @note No double-words-alignment is needed.
     */
    bool validate_checksum();

    /**
     * Sets the buffer this instance is operating on to @a buffer.
     *
     * @param buffer Specifies the buffer to be bound with this instance.
     */
    void set_buffer
        (
        BufferType *    buffer
        );

private:

    /**
     * Computes checksum for only a part of the buffer.
     *
     * @param number Number of bytes in the buffer to be considered
     *      while computing CRC32.
     */
    ChecksumType compute_checksum
        (
        unsigned int        number
        );

private:

    CryptoPP::CRC32         crc32;
        /**< The CryptoPP implementation of the CRC32 checksum. */

    UInt8 *                 digest;
        /**< Buffer that receives the checksum from CryptoPP. */

private:

    BufferType *            buffer;
        /**< The buffer the checksum should be computed for. */

    }; // class CRC32


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_MASSIV_CHECKSUM_H
