/** @file massiv_rsa.cpp
 * RSA cipher.
 */

/*
 * begin        : Tue Dec 18 2001
 * original     : Wei Dai (file version)
 * modified     : (C) 2001 by Ondrej Pecta (string version), Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

// Massiv

#include "net/massiv_rsa.h"

// CryptoPP

#include "net/cryptopp_begin.h"
#include "cryptopp/sha.h"
#include "net/cryptopp_end.h"

// STD C++

#include <ctime>

namespace Massiv {
namespace Core {


RSACipher::RSACipher()
    {
    init_random_pool( Seed() );
    }


RSACipher::RSACipher
    (
    UInt32          private_key_length,
    const Seed &    seed
    )
    {
    // Init cipher.

    init_random_pool( seed );

    // Encode random data into @c keys.private_key.

    Decryptor               priv( random_pool, private_key_length );
    CryptoPP::StringSink    priv_key( keys.private_key );

    priv.DEREncode( priv_key );
    priv_key.MessageEnd();

    // Compute @c keys.public_key that is complementary to @c keys.private_key.

    Encryptor               pub( priv );
    CryptoPP::StringSink    pub_key( keys.public_key );

    pub.DEREncode( pub_key );
    pub_key.MessageEnd();
    }


RSACipher::RSACipher
    (
    const Keys &    keys,
    const Seed &    seed
    ) : keys( keys )
    {
    init_random_pool( seed );
    }


RSACipher::~RSACipher()
    {
    }


void RSACipher::init_random_pool
    (
    const Seed &    seed
    )
    {
    random_pool.Put( reinterpret_cast<const UInt8 *>( seed.c_str() ), seed.size() );
    }


UInt32 RSACipher::encrypt
    (
    const UInt8 *   message,
    UInt32          message_length
    )
    {
    return encrypt_other_public( message, message_length, keys.public_key );
    }


UInt32 RSACipher::encrypt_other_public
    (
    const UInt8 *       message,
    UInt32              message_length,
    const std::string & other_public_key
    )
    {
    CryptoPP::StringSource  key( other_public_key, true );
    Encryptor               encryptor( key );

    #if( CRYPTOPP_VERSION >= 500 )

        hope( message_length <= encryptor.FixedMaxPlaintextLength() );
        hope( encryptor.FixedCiphertextLength() <= MAX_BUFFER_LENGTH );

        encryptor.Encrypt
            (
            random_pool,
            message,
            message_length,
            output_buffer
            );

        return encryptor.FixedCiphertextLength();

    #else

        hope( message_length <= encryptor.MaxPlainTextLength() );
        hope( encryptor.CipherTextLength() <= MAX_BUFFER_LENGTH );

        encryptor.Encrypt
            (
            random_pool,
            message,
            message_length,
            output_buffer
            );

        return encryptor.CipherTextLength();

    #endif
    }


UInt32 RSACipher::decrypt
    (
    const UInt8 *   encrypted,
    UInt32          encrypted_length
    )
    {
    CryptoPP::StringSource      key( keys.private_key, true );
    Decryptor                   decryptor( key );

    #if( CRYPTOPP_VERSION >= 500 )

        CryptoPP::DecodingResult    result;

        hope( decryptor.FixedMaxPlaintextLength() <= MAX_BUFFER_LENGTH );

        if( encrypted_length < decryptor.FixedCiphertextLength() )
            {
            return 0;
            }

        result = decryptor.Decrypt
                    (
                    random_pool,
                    encrypted,
                    decryptor.FixedCiphertextLength(),
                    output_buffer
                    );

        if( result.isValidCoding )
            {
            return result.messageLength;
            }
        else
            {
            return 0;
            }

    #else

        hope( decryptor.MaxPlainTextLength() <= MAX_BUFFER_LENGTH );

        if( encrypted_length < decryptor.CipherTextLength() )
            {
            return 0;
            }

        return decryptor.Decrypt( encrypted, output_buffer );

    #endif
    }


UInt32 RSACipher::sign
    (
    const UInt8 *   message,
    UInt32          message_length
    )
    {
    CryptoPP::StringSource      key( keys.private_key, true );
    Signer                      signer( key );
    CryptoPP::SignerFilter *    filter;

    filter = new CryptoPP::SignerFilter
                (
                random_pool,
                signer,
                new CryptoPP::ArraySink
                    (
                    output_buffer,
                    MAX_BUFFER_LENGTH
                    )
                );

    // Filter message using SignerFilter into @c output_buffer.

    CryptoPP::StringSource  source
        (
        message,
        message_length,
        true,
        filter
        );

    return signer.SignatureLength();
    }


bool RSACipher::verify
    (
    const UInt8 *   message,
    UInt32          message_length,
    const UInt8 *   signature,
    UInt32          signature_length
    )
    {
    return verify_other_public
                (
                message,
                message_length,
                signature,
                signature_length,
                keys.public_key
                );
    }


bool RSACipher::verify_other_public
    (
    const UInt8 *       message,
    UInt32              message_length,
    const UInt8 *       signature,
    UInt32              signature_length,
    const std::string & other_public_key
    )
    {
    CryptoPP::StringSource      key( other_public_key, true );
    Verifier                    verifier( key );
    CryptoPP::VerifierFilter *  filter;
    UInt8                       result;

    if( signature_length != verifier.SignatureLength() )
        {
        return false;
        }

    filter = new CryptoPP::VerifierFilter
                (
                verifier,
                new CryptoPP::ArraySink( &result, 1 )
                );

    #if( CRYPTOPP_VERSION >= 500 )

        filter->Put( signature, verifier.SignatureLength() );

    #else

        filter->PutSignature( signature );

    #endif

    // Filter message using VerifierFilter into @c result, which will contain 
    // either 1 or 0 depending on the whether the verification succeeded or not.

    CryptoPP::StringSource source
        (
        message,
        message_length,
        true,
        filter
        );

    return result == 1;
    }


RSACipher::Seed RSACipher::generate_seed
    (
    UInt32 length
    )
    {
    CryptoPP::RandomPool    pool;
    std::string             seed;
    UInt32                  i;

    // Initialize random number generator.
    
    // FIXME:   This is insecure and should be reworked because the seed value
    //          is fully determined by the current time.

    std::srand( static_cast<unsigned int>( std::time( NULL ) ) );
    const unsigned int random_value = std::rand(); 
    pool.Put( reinterpret_cast<const UInt8 *>( &random_value ), sizeof( 
        random_value ) );
    
    // Generate @c length bytes from @c pool.
    
    seed.resize( length );

    for( i = 0; i < length; i++ )
        {
        seed[ i ] = pool.GenerateByte();
        }

    return seed;         
    }


const RSACipher::Keys & RSACipher::get_keys() const
    {
    return keys;
    }


const std::string & RSACipher::get_public_key() const
    {
    return keys.public_key;
    }

const std::string & RSACipher::get_private_key() const
    {
    return keys.private_key;
    }


const UInt8 * RSACipher::get_output_buffer() const
    {
    return output_buffer;
    }


UInt32 RSACipher::get_max_text_len_for_encoding
    (
    const std::string & other_public_key
    ) const
    {
    CryptoPP::StringSource  key( other_public_key, true );
    Encryptor               encryptor( key );
    
    #if( CRYPTOPP_VERSION >= 500 )
        return encryptor.FixedMaxPlaintextLength();
    #else
        return encryptor.MaxPlainTextLength();
    #endif
    }


void RSACipher::set_keys
    (
    const std::string & private_key,
    const std::string & public_key
    )
    {
    keys.private_key = private_key;
    keys.public_key = public_key;
    }


} // namespace Massiv::Core
} // namespace Massiv
