/** @file massiv_rsa.h
 * RSA cipher.
 */

/*
 * begin        : Tue Dec 18 2001
 * original     : Wei Dai (file version)
 * modified     : (C) 2001 by Ondrej Pecta (string version), Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_MASSIV_RSA_H
#define MASSIV_MASSIV_RSA_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

// CryptoPP

#include "net/cryptopp_begin.h"
#include "cryptopp/rsa.h"
#include "cryptopp/randpool.h"
#include "cryptopp/filters.h"
#include "net/cryptopp_end.h"

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup net
 * @{
 */

/**
 * RSACipher.
 *
 * Support for asymmetric encryption of messages.  
 *  
 * @todo Describe the class properly.
 */
class RSACipher
    {
    NO_COPY_CONSTRUCTOR( RSACipher );
    NO_ASSIGNMENT_OPERATOR( RSACipher );
    
public:

    typedef CryptoPP::RSAES_OAEP_SHA_Decryptor      Decryptor;
    typedef CryptoPP::RSAES_OAEP_SHA_Encryptor      Encryptor;
    typedef CryptoPP::RSASSA_PKCS1v15_SHA_Signer    Signer;
    typedef CryptoPP::RSASSA_PKCS1v15_SHA_Verifier  Verifier;

    enum Constants
        {
        MAX_BUFFER_LENGTH = 128,
            /**< Maximum message length. */
        };
        
    enum Parameters
        {
        DEFAULT_PRIVATE_KEY_LENGTH = 512,
            /**< Default lengths of private keys, used throughout the whole
                 system. */            
        DEFAULT_PRIVATE_KEY_SEED_LENGTH = 64
            /**< Default lengths of seeds used to generate private keys, used
                 throughout the whole system. */            
        };

public:

    /** Cipher keys. */
    struct Keys
        {
        friend class RSACipher;

    private:

        std::string public_key;
            /**< Public key. */
        std::string private_key;
            /**< Private key. */

        }; // Keys

    typedef std::string Seed;

public:

    /**
     * Initializes the cipher using a default seed and leaves the private and
     * public keys undefined.     
     */
    RSACipher();

    /**
     * Initializes the cipher using the given seed, generates a random private 
     * key of a specified length and then computes the complementary public key.
     * 
     * The value of the seed initializes the cipher's random number generator.
     * The seed directly affects what the value of the generated private key will
     * be.               
     */
    RSACipher
        (
        UInt32          private_key_length,
        const Seed &    seed
        );

    /**
     * Initializes the cipher using the given seed and sets the keys.
     * 
     * It is the responsibility of the caller to make sure the keys are valid.          
     */
    RSACipher
        (
        const Keys &    keys,
        const Seed &    seed
        );

    /**
     * Destructor.
     */
    ~RSACipher();

private:

    /**
     * Initializes random pool.
     */
    void init_random_pool
        (
        const Seed &    seed
        );

public:

    //
    // Single-buffer oriented operations.
    //

    /**
     * Encrypts @a message of length @a message_length into to a temporary 
     * buffer using the cipher's public key.
     *      
     * Returns the length of the encrypted message.
     */
    UInt32 encrypt
        (
        const UInt8 *   message,
        UInt32          message_length
        );

    /**
     * Encrypts @a message of length @a message_length into to a temporary 
     * buffer using the external @c other_public_key public key.
     *      
     * Returns the length of the encrypted message.
     */
    UInt32 encrypt_other_public
        (
        const UInt8 *       message,
        UInt32              message_length,
        const std::string & other_public_key
        );

    /**
     * Decrypts @a encrypted into a temporary buffer using the cipher's
     * private key.
     *          
     * Returns the length of the decrypted message.
     */
    UInt32 decrypt
        (
        const UInt8 *   encrypted,
        UInt32          encrypted_length
        );

    /**
     * Signs @a message of length @a message_length using the cipher's private
     * key and puts the signature into a temporary buffer.
     *      
     * Returns the length of the signature.
     */
    UInt32 sign
        (
        const UInt8 *   message,
        UInt32          message_length
        );

    /**
     * Verifies the message using the cipher's public key.
     *      
     * Checks whether @a signature is a valid signature of a given message.     
     */
    bool verify
        (
        const UInt8 *   message,
        UInt32          message_length,
        const UInt8 *   signature,
        UInt32          signature_length
        );

    /**
     * Verifies the message using an external public key.
     *
     * Checks whether @c signature is a valid signature of a given message,
     * that was signed by a private key complementary to the specified
     * @c other_public_key public key.               
     */
    bool verify_other_public
        (
        const UInt8 *       message,
        UInt32              message_length,
        const UInt8 *       signature,
        UInt32              signature_length,
        const std::string & other_public_key
        );

    /**
     * Returns the content of the temporary buffer.
     */
    const UInt8 * get_output_buffer() const;

    /**
     * Returns the maximum size of a plain text which can be encoded using this 
     * RSA cipher.
     *     
     * Depends on the key lengths.
     */
    UInt32  get_max_text_len_for_encoding
        (
        const std::string & other_public_key
        ) const;

public:

    //
    // Wrappers.
    //

    /**
     * Encrypts <first, last] range and places it to output.
     */
    template <class InputIterator, class OutputContainer>
    void encrypt_range
        (
        InputIterator           first,
        InputIterator           last,
        OutputContainer &       output
        )
        {
        int size;
        size = encrypt( copy_input( first, last, 0 ), last - first );
        copy_output( output, size );
        }

    /**
     * Encrypts <first, last] range with other public key and places it to output.
     */
    template <class InputIterator, class OutputContainer>
    void encrypt_range_other_public
        (
        InputIterator           first,
        InputIterator           last,
        OutputContainer &       output,
        std::string &           other_public
        )
        {
        int size;
        size = encrypt_other_public( copy_input( first, last, 0 ), last - first, 
            other_public );
        copy_output( output, size );
        }

    /**
     * Encrypts message into encrypted.
     */
    template <class InputContainer, class OutputContainer>
    void encrypt
        (
        const InputContainer &  message,
        OutputContainer &       encrypted
        )
        {
        encrypt_range( message.begin(), message.end(), encrypted );
        }

    /**
     * Encrypts message with other public key into encrypted.
     */
    template <class InputContainer, class OutputContainer>
    void encrypt
        (
        const InputContainer &  message,
        OutputContainer &       encrypted,
        std::string &           other_public
        )
        {
        encrypt_range_other_public( message.begin(), message.end(), encrypted, 
            other_public );
        }

    /**
     * Decrypts encrypted message and places it to decrypted.
     */
    template <class InputContainer, class OutputContainer>
    UInt32 decrypt
        (
        const InputContainer &  encrypted,
        OutputContainer &       decrypted
        )
        {
        const UInt8 *   input;
        UInt32          size;

        input = copy_input( encrypted.begin(), encrypted.end(), 0 );
        size = decrypt( input, encrypted.size() );
        copy_output( decrypted, size );
        return size;
        }

    /**
     * Signs the message.
     */
    template <class InputContainer, class OutputContainer>
    void sign
        (
        const InputContainer &  message,
        OutputContainer &       signature
        )
        {
        const UInt8 *   input;
        int             size;

        input = copy_input( message.begin(), message.end(), 0 );
        size = sign( input, message.size() );
        copy_output( signature, size );
        }

    /**
     * Verifies the message.
     */
    template <class MessageContainer, class SignatureContainer>
    bool verify
        (
        const MessageContainer &    message,
        const SignatureContainer &  signature
        )
        {
        return verify
            (
            copy_input( message.begin(), message.end(), 0 ),
            message.size(),
            copy_input( signature.begin(), signature.end(), 1 ),
            signature.size()
            );
        }

    /**
     * Verifies the message with other key.
     */
    template <class MessageContainer, class SignatureContainer>
    bool verify
        (
        const MessageContainer &    message,
        const SignatureContainer &  signature,
        std::string &               other_public
        )
        {
        return verify_other_public
            (
            copy_input( message.begin(), message.end(), 0 ),
            message.size(),
            copy_input( signature.begin(), signature.end(), 1 ),
            signature.size(),
            other_public
            );
        }

private:

    /**
     * Copies data from output buffer to the output container.
     */
    template <class OutputContainer>
    void copy_output
        (
        OutputContainer &   container,
        UInt32              size
        )
        {
        typename OutputContainer::iterator it;

        container.resize( size );

        for( it = container.begin(); it != container.end(); it++ )
            {
            *it = ( typename OutputContainer::value_type ) output_buffer[ it - 
                container.begin() ];
            }
        }

    /**
     * Copies data from input container to the input buffer.
     */
    template <class InputIterator>
    const UInt8 * copy_input
        (
        InputIterator   first,
        InputIterator   last,
        int             input_number
        )
        {
        UInt8 * buffer;
        int     i;

        hope( last - first <= MAX_BUFFER_LENGTH );

        i = 0;
        buffer = input_buffer[ input_number ];

        while( first < last )
            {
            buffer[ i ] = ( UInt8 ) *first;

            first++;
            i++;
            }

        return buffer;
        }

public:

    /**
     * Generates a seed of a given length.
     * 
     * Can be used to generate arbitrary random strings of a requested length.
     * 
     * @todo Move outside RSACipher?          
     */         
    static Seed generate_seed
        (
        UInt32 length
        );

    /**
     * Returns current keys.
     */
    const Keys & get_keys() const;

    /**
    * Returns current public key.
    */
    const std::string & get_public_key() const;

    /**
    * Returns current private key.
    */
    const std::string & get_private_key() const;

    /**
     * Sets keys.
     */
    void set_keys
        (
        const std::string & private_key,
        const std::string & public_key
        );

private:

    Keys                    keys;
        /**< Current pair of keys.
             Depending on operations performed, either the public key, the private
             key or both the public and private keys have to be filled and 
             valid. */
    CryptoPP::RandomPool    random_pool;
        /**< Random number pool. */
    UInt8                   output_buffer[ MAX_BUFFER_LENGTH ];
        /**< Temporary buffer used for storing resuts of cipher operations.
             Note that those operations that use output_buffer can not
             be implemented in other way (ie. writing directly to
             output container) because we do not know how long will be the
             resulting message until the operation is performed. */
    UInt8                   input_buffer[ 2 ][ MAX_BUFFER_LENGTH ];
        /**< Two input buffers. */

    }; // RSACipher

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_MASSIV_RSA_H
