/** @file net.cpp
 * Network layer interface.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta, Marek Svantner, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/net.h"
#include "net/connection_info_impl.h"
#include "utility/debug.h"
#include "global/global.h"
#include "registry/registry.h"
#include "registry/statistics.h"
#include "database/node_database.h"
#include "node/node_manager.h"
#include "logger/logger.h"
#include "utility/quote_string.h"
#include <stdexcept>
#include <sstream>
#include <iostream>
#include <typeinfo>


namespace Massiv {
namespace Core {

using namespace NetworkCommon;

static const char * msg_not_awaited_message =
    "Not awaited message in the incoming network message.";

static const char * msg_messages_number_mismatch =
    "Tcp buffer messages count mismatch in the incoming network message.";

// Implementation of Network::NetworkSettings
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Network::NetworkSettings::NetworkSettings()
    :   Settings( "Network" ),
        tcp_preallocated_buffers
            (
            access_integer
                (
                "tcp_preallocated_buffers",
                DEFAULT_TCP_PREALLOCATED_BUFFERS
                )
            ),
        udp_preallocated_buffers
            (
            access_integer
                (
                "udp_preallocated_buffers",
                DEFAULT_UDP_PREALLOCATED_BUFFERS
                )
            ),
        max_idle_seconds
            (
            access_integer
                (
                "max_idle_seconds",
                DEFAULT_MAX_IDLE_SECONDS
                )
            ),
        next_ping_delay
            (
            access_integer
                (
                "next_ping_delay",
                DEFAULT_NEXT_PING_DELAY
                )
            ),
        max_pings
            (
            access_integer
                (
                "max_pings",
                DEFAULT_MAX_PINGS
                )
            ),
        max_server_life
            (
            access_integer
                (
                "max_server_life",
                DEFAULT_MAX_SERVER_LIFE
                )
            ),
        blowfish_key_timeout
            (
            access_integer
                (
                "blowfish_key_timeout",
                DEFAULT_BLOWFISH_KEY_TIMEOUT
                )
            ),
        max_receive_bytes
            (
            access_integer
                (
                "max_receive_bytes",
                DEFAULT_MAX_RECEIVE_BYTES
                )
            ),
        blowfish_key_length
            (
            access_integer
                (
                "blowfish_key_length",
                DEFAULT_BLOWFISH_KEY_LENGTH
                )
            )
    {
    }  // Network::NetworkSettings::NetworkSettings()


// Implementation of Network
// ~~~~~~~~~~~~~~~~~~~~~~~~~

Network::Network()
    {
    arch_init();

    tcp_sendlist.clear();
    udp_sendlist.clear();
    async_connectionlist.clear();
    init_udp_list.clear();
    conn_list.clear();

    // Pre-allocate permanent TCP buffers
    for(
        unsigned int i = 0;
        i < ( unsigned ) settings.tcp_preallocated_buffers.read();
        i++
        )
        {
        TCPOBuffer tcpbuffer;
        tcpbuffer.permanent = true;
        tcp_permlist.push_front( tcpbuffer );
        };

    // Pre-allocate permantent UDP buffers
    for(
        unsigned int j = 0;
        j < ( unsigned ) settings.udp_preallocated_buffers.read();
        j++
        )
        {
        UDPOBuffer udpbuffer;
        udpbuffer.permanent = true;
        udp_permlist.push_front( udpbuffer );
        }

    well_known_port_socket = NULL;
    sockets_waiting_for_port.clear();
    my_node_id = NodeId();
    local_node_id_assigned = false;
    recvbuf_size = 0;
    last_tick_time = -1;
    last_tick_bytes = 0;
    bytes_delta = 0;
    time_delta = 0;
    }  // Network::Network()


Network::~Network()
    {
    std::list< TCPOBuffer >::iterator iter_tcp;
    std::list< UDPOBuffer >::iterator iter_udp;

    // Close all connections and stay silent.
    invalidate_bad_connections( true, true );

    arch_deinit();

    iter_tcp = tcp_permlist.begin();
    while( iter_tcp != tcp_permlist.end() )
        {
        iter_tcp->unbind();
        iter_tcp++;
        }

    iter_udp = udp_permlist.begin();
    while( iter_udp != udp_permlist.end() )
        {
        iter_udp->unbind();
        iter_udp++;
        }

    tcp_permlist.clear();
    tcp_sendlist.clear();
    udp_permlist.clear();
    udp_sendlist.clear();

    node_table.clear();

    // Clear sockets waiting for port, socket waiting for accept.
    if( well_known_port_socket )
        {
        well_known_port_socket->close();
        delete well_known_port_socket;
        }
    std::list< Socket * >::iterator sock_iter;

    sock_iter = sockets_waiting_for_port.begin();
    while( sock_iter != sockets_waiting_for_port.end() )
        {
        Socket * socket = *sock_iter;
        hope( socket );

        socket->close();
        delete socket;

        sock_iter++;
        }
    sockets_waiting_for_port.clear();
    }  // Network::~Network()


std::string Network::to_string() const
    {
    std::ostringstream                              ss;
    NodeHash::const_iterator                        n_it;
    std::list< NodeId >::const_iterator             a_it;
    std::list< TCPOBuffer >::const_iterator         tp_it;
    std::list< TCPOBuffer * >::const_iterator       ts_it;
    std::list< UDPOBuffer >::const_iterator         up_it;
    std::list< UDPOBuffer * >::const_iterator       us_it;
    std::list< ConnectionInfo * >::const_iterator   c_it;
    int                                             count;

    ss  << "* my_node_id = " << my_node_id 
        << " (assigned = " << local_node_id_assigned << ")\n";

    ss  << "* node_table:\n";

    for( n_it = node_table.begin(), count = 0; n_it != node_table.end(); 
        ++n_it, ++count )
        {
        const NodeId &      node_id = n_it->first;
        ConnectionInfo *    connection_info = n_it->second;

        ss << count << ": " << node_id << ":\n";

        if( connection_info )
            {
            ss << connection_info->to_string() << '\n';
            }
        else
            {
            ss << "NULL\n";
            }
        }

    ss << "* async_connectionlist:\n";

    for( a_it = async_connectionlist.begin(), count = 0; 
        a_it != async_connectionlist.end(); ++a_it, ++count )
        {
        ss << count << ": " << *a_it << '\n';
        }

    ss << "* tcp_permlist:\n";

    for( tp_it = tcp_permlist.begin(), count = 0; tp_it != tcp_permlist.end(); 
        ++tp_it, ++count )
        {
        ss << count << ": " << tp_it->to_string() << '\n';
        }
    
    ss << "* tcp_sendlist:\n";

    for( ts_it = tcp_sendlist.begin(), count = 0; ts_it != tcp_sendlist.end();
        ++ts_it, ++count )
        {
        TCPOBuffer * buffer = *ts_it;
        ss << count << ": " << buffer->to_string() << '\n';
        }
    
    ss << "* udp_permlist:\n";

    for( up_it = udp_permlist.begin(), count = 0; up_it != udp_permlist.end();
        ++up_it, ++count )
        {
        ss << count << ": " << up_it->to_string() << '\n';
        }
    
    ss << "* udp_sendlist:\n";

    for( us_it = udp_sendlist.begin(), count = 0; us_it != udp_sendlist.end();
        ++us_it, ++count )
        {
        UDPOBuffer * buffer = *us_it;
        ss << count << ": " << buffer->to_string() << '\n';
        }
    
    ss << "* conn_list:\n";

    for( c_it = conn_list.begin(), count = 0; c_it != conn_list.end();
        ++c_it, ++count )
        {
        ConnectionInfo * info = *c_it;
        hope( info != NULL );
        ss << count << ": " << info->to_string() << '\n';
        }
    
    ss << "* init_udp_list:\n";

    for( c_it = init_udp_list.begin(), count = 0; c_it != init_udp_list.end();
        ++c_it, ++count )
        {
        ConnectionInfo * info = *c_it;
        hope( info != NULL );
        ss << count << ": " << info->to_string() << '\n';
        }

    return ss.str();
    }


void Network::initialize()
    {
    NodeDatabase &  node_database = Global::node_database();
    NodeId          local_node_id = node_database.get_local_node_id();
    
    last_tick_time = Global::system_time();
    
    // Log.

    #ifdef OPTION_DEBUG

    Global::log_info
        (
        Status::FACILITY_NETWORK,
        Status::PRIORITY_LOW,
        to_string()
        );

    #endif

    std::ostringstream s;
    s << "Resetting network. New local node is " << local_node_id << ".";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        s.str()
        );
    
    // Change credentials.

    assign_local_node_id( local_node_id );
    rsa.set_keys
        (
        node_database.get_local_node_rsa_private_key(),
        node_database.get_local_node_rsa_public_key()
        );
    
    // Invalidate connections, do not inform node manager.

    invalidate_bad_connections( true, true );
    
    #ifdef OPTION_DEBUG

    Global::log_info
        (
        Status::FACILITY_NETWORK,
        Status::PRIORITY_LOW,
        to_string()
        );

    #endif
    
    // There should be no message to send.
    
    hope( tcp_sendlist.size() == 0 );
    hope( udp_sendlist.size() == 0 );
    }


void Network::assign_local_node_id
    (
    const NodeId & local_node_id
    )
    {
    hope
        ( 
        my_node_id.is_type( NodeId::TYPE_ANONYMOUS ) || 
        local_node_id.is_type( NodeId::TYPE_ANONYMOUS )
        );

    my_node_id             = local_node_id;
    local_node_id_assigned = true;
    }


const NodeId & Network::get_local_node_id()
    {
    if( !local_node_id_assigned )
        {
        panic( "Local node id not asigned yet!" );
        }
    return my_node_id;
    }


void Network::next_tick()
    {
    Statistics::MeasureSystemTime measure( net_stat.consumed_time );
    
    initiate_connections();
    accept();
    receive_all_messages();
    send_all_messages();
    invalidate_bad_connections();

    bytes_delta += net_stat.get_bytes_received() + net_stat.get_bytes_sent() - last_tick_bytes;
    time_delta += static_cast<Float32>
        ( 
        ( Global::system_time() - last_tick_time ).get_value()
        );

    // We need actual throughput per second.
    if( time_delta >= 1 )
        {
        net_stat.recompute_throughput( bytes_delta, time_delta );
        time_delta = 0;
        bytes_delta = 0;
        
        // After one second send init udps.
        init_udp();
        }

    last_tick_time = Global::system_time();
    last_tick_bytes = net_stat.get_bytes_received() + net_stat.get_bytes_sent();
    measure.stop();
    }


bool Network::is_connection_creating
    (
    const NodeId & node
    )
    {
    std::list< ConnectionInfo * >::iterator  iterp;

    iterp = conn_list.begin();
    while( iterp != conn_list.end() )
        {
        ConnectionInfo * conn_info = *iterp;
        hope( conn_info );
        
        if( conn_info->get_node_id() == node )
            {
            return true;           
            }
        iterp++;
        }
    
    NodeHash::iterator  iter;

    iter = node_table.find( node );
    try
        {
        if( iter != node_table.end() &&
            iter->second->get_connection_status() != ConnectionInfo::CONNECTED &&
            iter->second->get_connection_status() != ConnectionInfo::NOT_CONNECTED )
            {
            return true;
            }
        }
    catch( ... )
        {
        return true;
        }
    
    return false;
    }


bool Network::is_node_connected
    (
    const NodeId & node
    )
    {
    if( node == my_node_id )
        {
        return true;
        }

    NodeHash::iterator  iter;

    iter = node_table.find( node );
    try
        {
        if( iter != node_table.end() )
            {
            hope( iter->second );
            if( iter->second->get_connection_status() == ConnectionInfo::CONNECTED )
                {
                return true;
                }
            else
                {
                return false;
                }
            }
        else
            {
            return false;
            }
        }
    catch( ... )
        {
        return false;
        }
    }


void Network::invalidate_bad_connections
    (
    bool force_disconnect,
    bool deny_node_manager_info
    )
    {
    NodeHash::iterator  iter;
    std::list< ConnectionInfo * >::iterator  port_iter;

    iter = node_table.begin();
    while( iter != node_table.end() )
        {
        ConnectionInfo *    conn_info = iter->second;
        const NodeId &      node_id = iter->first;
        
        hope( conn_info );
        
        if( conn_info->get_connection_status() == ConnectionInfo::NOT_CONNECTED 
            || force_disconnect )
            {
            std::ostringstream  s;

            s << "Closing connection to node " << node_id << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
                s.str()
                );

            if( force_disconnect && 
                conn_info->get_connection_status() != ConnectionInfo::NOT_CONNECTED )
                {
                conn_info->set_disconnection_reason( ConnectionInfo::REQUEST_REASON );
                }

            switch( conn_info->get_disconnection_reason() )
                {
                case ConnectionInfo::REQUEST_REASON:
                    {
                    try
                        {
                        send_uncrypted_message
                            (
                            NODE_DISCONNECTED_MESSAGE,
                            conn_info->get_tcpsocket(),
                            VariantPointer( conn_info )
                            );
                        }
                    catch( ... )
                        {
                        // I don't care, maybe the connection doesn't exist, but it's better
                        // to inform other side, that connection is closed.
                        }

                    NodeManager & node_mgr = Global::node_manager();
                    
                    try
                        {
                        if( conn_info->should_inform_node_manager() && !deny_node_manager_info )
                            {
                            conn_info->set_inform_node_manager( false );
                            node_mgr.net_callback_connection_closed( node_id );
                            }
                        }
                    catch( ... )
                        {
                        if( !force_disconnect )
                            {
                            throw;
                            }
                        }

                    break;
                    }

                default:
                    {
                    if( conn_info->should_inform_node_manager() && !deny_node_manager_info )
                        {
                        conn_info->set_inform_node_manager( false );
                        NodeManager &       node_mgr = Global::node_manager();
                        node_mgr.net_callback_connection_lost( node_id );
                        }
                    break;
                    }
                }

            // Drop messages only when error or forced.
            // Otherwise only set connection info to NULL.
            // This guarantees all messages will be delivered.
            drop_all_to_server
                ( 
                node_id, 
                force_disconnect || 
                node_id.is_type( NodeId::TYPE_CLIENT ) ||
                conn_info->get_disconnection_reason() == ConnectionInfo::ERROR_REASON 
                );
            node_table.erase( iter++ );
            std::list< ConnectionInfo * >::iterator it;
            it = std::find( init_udp_list.begin(), init_udp_list.end(), conn_info );
            if( it != init_udp_list.end() )
                {
                hope( *it );
                init_udp_list.erase( it );
                }
            delete conn_info;
            net_stat.inc_connections_killed();

            }
        else
            {
            iter++;
            }
        }

    // Not introduced sockets cleanup.
    port_iter = conn_list.begin();
    while( port_iter != conn_list.end() )
        {
        ConnectionInfo * conn_info = *port_iter;
        
        hope( conn_info );
        if( conn_info->get_connection_status() == ConnectionInfo::NOT_CONNECTED || force_disconnect )
            {
            conn_list.erase( port_iter++ );
            delete conn_info;
            net_stat.inc_connections_killed();
            }
        else
            {
            port_iter++;
            }
        }

    // Delete messages prepared for sending and no connection exists yet.
    if( force_disconnect )
        {
        #ifdef OPTION_DEBUG
        
        if( tcp_sendlist.size() || udp_sendlist.size() )
            {
            std::ostringstream oss;
            oss << "Found " << tcp_sendlist.size() << " TCP and " << udp_sendlist.size();
            oss << " UDP not cleaned buffers.";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                oss.str()
                );
            }
        
        #endif

            {
            std::list< TCPOBuffer * >::iterator tcp_iter = tcp_sendlist.begin();
            while( tcp_iter != tcp_sendlist.end() )
                {
                TCPOBuffer * tcp_buf = *tcp_iter++;

                hope( tcp_buf->connection_info == NULL );
                drop( tcp_buf );
                }
            }

            {
            std::list< UDPOBuffer * >::iterator udp_iter = udp_sendlist.begin();
            while( udp_iter != udp_sendlist.end() )
                {
                UDPOBuffer * udp_buf = *udp_iter++;
                
                hope( udp_buf->connection_info == NULL );
                drop( udp_buf );
                }
            }
        }

    } // Network::invalidate_bad_connections()


int Network::receive_via_port_hash()
    {
    int                                     messages = 0;
    std::list< ConnectionInfo * >::iterator iter = conn_list.begin();

    while( iter != conn_list.end() )
        {
        ConnectionInfo *  conn_info = *iter;
        hope( conn_info );

        if( conn_info->get_connection_status() != ConnectionInfo::NOT_CONNECTED )
            {
            if( !send_ping_if_idle( conn_info ) )
                {
                break;
                }
            }

        switch( conn_info->get_connection_status() )
            {
            case ConnectionInfo::CONNECTED:
                {
                hope( 0 );
                break;
                }

            case ConnectionInfo::NOT_CONNECTED:
                {
                iter++;
                break;
                }

            case ConnectionInfo::NODE_ID_EXCHANGE_SERVER:
                {
                messages += receive_in_node_id_exchange_server_state( conn_info );
                iter++;
                break;
                }

            case ConnectionInfo::BLOWFISH_EXCHANGE_SERVER:
                {
                messages +=
                    receive_in_blowfish_exchange_server_state( conn_info );
                iter++;
                break;
                }
            case ConnectionInfo::UDP_CONNECTION_SETUP_SERVER:
                {
                messages +=
                    receive_in_udp_connection_setup_server_state( conn_info, iter );
                break;
                }
            
            case ConnectionInfo::TRYING_TO_DISCONNECT:
            case ConnectionInfo::BLOWFISH_EXCHANGE_CLIENT:
            case ConnectionInfo::UDP_PORT_RECEIVE_CLIENT:
                {
                throw ExBadConnectionStatus( "Unexpected connection status." );
                }
                
            } // end switch
        } // end while
  
    return messages;
    }  // Network::receive_via_port_hash()


int Network::receive_via_node_hash()
    {
    int                 messages = 0;
    NodeHash::iterator  iter;

    iter = node_table.begin();
    while( iter != node_table.end() )
        {
        ConnectionInfo *    conn_info = iter->second;
        
        hope( conn_info );
        
        if( conn_info->get_connection_status() != ConnectionInfo::NOT_CONNECTED )
            {
            if( !send_ping_if_idle( conn_info ) )
                {
                iter++;
                break;
                }
            }

        switch( conn_info->get_connection_status() )
            {
            case ConnectionInfo::NOT_CONNECTED:
                {
                //Some other massiv parts can also tell network, that connection's closed.
                //Not connected connections will be removed at the end of the net tick.
                iter++;
                break;
                }
            
            case ConnectionInfo::TRYING_TO_DISCONNECT:
            case ConnectionInfo::CONNECTED:
                {
                messages += receive_in_connected_state( conn_info );
                iter++;
                break;
                }
            
            case ConnectionInfo::BLOWFISH_EXCHANGE_CLIENT:
                {
                messages +=
                    receive_in_blowfish_exchange_client_state( conn_info );

                iter++;
                break;
                }

            case ConnectionInfo::UDP_PORT_RECEIVE_CLIENT:
                {
                messages +=
                    receive_in_udp_port_receive_client_state( conn_info );
                iter++;
                break;
                }

            case ConnectionInfo::NODE_ID_EXCHANGE_SERVER:
            case ConnectionInfo::BLOWFISH_EXCHANGE_SERVER:
            case ConnectionInfo::UDP_CONNECTION_SETUP_SERVER:
                {
                throw ExBadConnectionStatus( "Unexpected connection status." );
                }

            } // end switch

        } // end while

    return messages;
    }  // Network::receive_via_node_hash()


int Network::receive_via_sockets()
    {
    std::list< Socket * >::iterator iter_tcpsocket;
    int len;
    int messages = 0;

    // Try to receive something on "waiting" for assigned port sockets.
    // Client side.

    iter_tcpsocket = sockets_waiting_for_port.begin();

    while( iter_tcpsocket != sockets_waiting_for_port.end() )
        {
        Socket * waiting_socket = checked_cast<Socket *>( *iter_tcpsocket );

        hope( waiting_socket );

        if( waiting_socket->receive( recvbuf, recvbuf_size ) == true )
            {
            try
                {
                net_stat.inc_tcp_bytes_received( recvbuf_size );
                net_stat.inc_messages_received( 1 );

                // There should be only one message in this packet.

                UInt16 count = recvbuf.read_int( 16 );

                if( count != 1 )
                    {
                    throw ExMessagesNumberMismatch( msg_messages_number_mismatch );
                    }

                UInt32 tag = recvbuf.read_int( 32 );

                if( tag == SERVER_PORT_MESSAGE )
                    {
                    len = recvbuf.read_int( 32 );

                    if( len != 4 )
                        {
                        throw ExMessageLengthMismatch
                            ( "Wrong message length in incoming network message." );
                        }

                    // I have received a response from server!!!

                    NodeId sid = NodeId::create_from_uint32( recvbuf.read_int( 32 ) );
                    sockets_waiting_for_port.erase( iter_tcpsocket++ );
                    
                    NodeHash::iterator it = node_table.find( sid );
                    bool ok = false;
                    if( it == node_table.end() )
                        {
                        ok = true;
                        }
                    else
                        {
                        ConnectionInfo * conn = it->second;
                        if( conn->get_connection_status() == ConnectionInfo::NOT_CONNECTED 
                            && conn->get_disconnection_reason() == ConnectionInfo::REQUEST_REASON )
                            {
                            ok = true;
                            }
                        }


                    //hope( !is_node_connected( sid ) );

                    if( !ok ) // Do not create a new connection.
                        {
                        delete waiting_socket;
                        }
                    else
                        {
                        ConnectionInfo *    connection_info =
                            new ConnectionInfo( false );

                        net_stat.inc_connections_created( 1 );

                        connection_info->set_tcpsocket( waiting_socket );
                        connection_info->touch_last_message_time();

                        connection_info->set_connection_status
                            (
                            ConnectionInfo::BLOWFISH_EXCHANGE_CLIENT
                            );

                        connection_info->set_node_id( sid );

                        connection_info->set_other_side_public_key
                            (
                            get_key_from_node_id( sid )
                            );

                        // Send NodeId.

                        send_uncrypted_message
                            (
                            NODE_ID_MESSAGE,
                            connection_info->get_tcpsocket(),
                            VariantPointer()
                            );
                        
                        hope( node_table.find( sid ) == node_table.end() );

                        node_table.insert
                            (
                            Network::NodeHash::value_type
                                (
                                sid,
                                connection_info
                                )
                            );
                        }
                    }
                else
                    {
                    if( !process_error_message( tag ) )
                        {
                        throw ExMessageCorrupted( msg_not_awaited_message );
                        }
                    }

                messages++;
                }
            catch( ExConnectError & )
                {
                throw;
                }
            catch( ... )
                {
                sockets_waiting_for_port.erase( iter_tcpsocket++ );
                waiting_socket->close();

                net_stat.inc_bad_packets_received();
                net_stat.inc_connections_killed();
                }
            }
        else // Nothing received, getting older.
            {
            iter_tcpsocket++;

            SystemTime server_life_len =
                        Global::system_time() -
                        waiting_socket->get_server_start_timestamp();

            if( server_life_len.get_value() > settings.max_server_life )
                {
                // Too old.
                waiting_socket->close();
                sockets_waiting_for_port.remove( waiting_socket );
                net_stat.inc_connections_killed();
                
                throw ExConnectError( "No assigned port received! Connection failed." );
                }
            }
        } // end while

    return messages;
    }  // Network::receive_via_sockets()


bool Network::start_well_known_port_server
    (
    int port
    )
    {
    hope( well_known_port_socket == NULL );

    well_known_port_socket = create_tcp_socket();

    if( !well_known_port_socket->start_server( port ) )
        {
        delete well_known_port_socket;
        well_known_port_socket = NULL;

        std::ostringstream  s;
        s << "Cannot start well known server on port " << port << ".";

        throw ExWellKnownServerStartError( s.str() );
        }
    else
        {
        // Port printout.
        std::ostringstream  ss;
        ss << "DEBUG: TCP start_well_known_port_server() start on port " << port << ".";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST ),
            ss.str()
            );
        }

    return true;
    }


int Network::accept()
    {
    int accept1 = 0;

    if( well_known_port_socket )
        {
        accept1 = accept_on_well_known_port_server();
        net_stat.inc_connections_created( accept1 );
        }
    return( accept1 );
    }


void Network::initiate_connections()
    {
    std::list< NodeId >::iterator iter = async_connectionlist.begin();
    std::list< NodeId >::iterator pom;
    
    while( iter != async_connectionlist.end () )
        {
        pom = iter++;
        NodeId sid = *pom;
        async_connectionlist.erase( pom );
        
        if( is_node_connected( sid ) )
            {
            continue;
            }
        if( !initiate_connection_to_server( sid ) )
            {
            // @todo inform node manager about locked buffers.
    
            unsigned int counttcp = 0;
            unsigned int countudp = 0;
            
            std::list< TCPOBuffer * >::iterator tcpbuffer = tcp_sendlist.begin();
            std::list< UDPOBuffer * >::iterator udpbuffer = udp_sendlist.begin();

            std::list< TCPOBuffer * >::iterator tcp;
            std::list< UDPOBuffer * >::iterator udp;

            while( tcpbuffer != tcp_sendlist.end() )
                {
                tcp = tcpbuffer++;
                hope( *tcp );
                
                if( ( *tcp )->owner == sid )
                    {
                    #ifdef OPTION_DEBUG

                    {
                    std::ostringstream oss;
                    oss << "Found undeliverable TCP buffer for "
                        << ( *tcp )->owner << ".";

                    Global::log_info
                        (
                        Status::FACILITY_NETWORK,
                        Status::PRIORITY_LOWEST,
                        oss.str()
                        );
                    }
                    
                    #endif

                    ++counttcp;
                    }
                }
            while( udpbuffer != udp_sendlist.end() )
                {
                udp = udpbuffer++;
                hope( *udp );
                
                if( ( *udp )->owner == sid )
                    {
                    #ifdef OPTION_DEBUG

                    {
                    std::ostringstream oss;
                    oss << "Found undeliverable UDP buffer for "
                        << ( *udp )->owner << ".";

                    Global::log_info
                        (
                        Status::FACILITY_NETWORK,
                        Status::PRIORITY_LOWEST,
                        oss.str()
                        );
                    }
                    
                    #endif

                    ++countudp;
                    }
                } 
            
            {
            std::ostringstream oss;

            oss << "Async connection creation to node " << sid << " failed."
                << std::endl;
            oss << counttcp << " TCP and " << countudp
                << " UDP locked undeliverable buffers found.";
            
            Global::log
                (
                STATUS_DEBUG( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST ),
                oss.str()
                );
            }
                
            NodeManager & node_mgr = Global::node_manager();
            node_mgr.net_callback_cannot_initiate_connection( sid );
            }
        }
    }


void Network::async_initiate_connection_to_server
    (
    const NodeId & node
    )
    {
    // Check whether service is connecting. Service is always passive.
    if( my_node_id.is_type( NodeId::TYPE_SERVICE ) )
        {
        hope( node.is_type( NodeId::TYPE_SERVER ) );
        std::ostringstream oss;
        oss << "Service has stored a message for node " << node << "."; 
        oss << "Waiting for server connect.";

        Global::log
            (
            STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOW ),
            oss.str()
            );
        return;
        }

    std::list< NodeId >::iterator result =
        std::find
            (
            async_connectionlist.begin(),
            async_connectionlist.end(),
            node
            );

    NodeHash::iterator it = node_table.find( node );
    bool ok = false;
    if( it == node_table.end() )
        {
        ok = true;
        }
    else
        {
        ConnectionInfo * conn = it->second;
        if( conn->get_connection_status() == ConnectionInfo::NOT_CONNECTED )
            {
            ok = true;
            }
        }

    if( result == async_connectionlist.end() && ok )
        {
        std::ostringstream oss;
        oss << "Initiating async connection to node " << node << ".";

        Global::log
            (
            STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
            oss.str()
            );
        async_connectionlist.push_back( node );
        }
    }


bool Network::initiate_connection_to_server
    (
    ServerAddress & server_address
    )
    {
    TCPSocket * tcpsocket = create_tcp_socket();
    
    hope( tcpsocket );
    
    if( !tcpsocket->connect( server_address ) )
        {
        std::ostringstream  s;

        s   << "Cannot initiate connection to " << server_address.get_dns_name()
            << ":" << server_address.get_port() << ".";

        Global::log
            (
            STATUS_DEBUG( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST ),
            s.str()
            );

        delete tcpsocket;

        return false;
        }

    tcpsocket->server_address  = server_address;
    tcpsocket->touch_server_start_timestamp();
    sockets_waiting_for_port.push_front( tcpsocket );

    return true;
    }


bool Network::initiate_connection_to_server
    (
    const NodeId & node
    )
    {
    // It is possible only to connect to SERVER or SERVICE.
    hope( !node.is_type( NodeId::TYPE_CLIENT ) );
    hope( node != my_node_id );

    if( !is_node_connected( node ) )
        {
        #ifdef OPTION_DEBUG

        std::ostringstream oss;
        oss << "Connecting to node " << node << ".";

        Global::log
            (
            STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
            oss.str()
            );
        
        #endif

        NodeDatabase & node_db = Global::node_database();
        ServerAddress sa = node_db.get_node_server_address( node );

        return initiate_connection_to_server( sa );
        }

    return true;
    }


void Network::close_connection
    (
    const NodeId &  node
    )
    {
    NodeHash::iterator  iter;

    if( my_node_id.is_type( NodeId::TYPE_CLIENT ) &&
            node.is_type( NodeId::TYPE_SERVER ) )
        {
        iter = node_table.begin();
        
        while( iter != node_table.end() )
            {
            ConnectionInfo * conn_info = iter->second;
            hope( conn_info );
            
            conn_info->set_disconnection_reason( ConnectionInfo::REQUEST_REASON );
            conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
            iter++;
            }
        }
    else    
        {
        iter = node_table.find( node );
        
        if( iter != node_table.end() )
            {
            ConnectionInfo * conn_info = iter->second;
            hope( conn_info );

            conn_info->set_disconnection_reason( ConnectionInfo::REQUEST_REASON );
            conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
            }
        }
    }


bool Network::is_server_connection
    (
    const NodeId &  node
    )
    {
    return( my_node_id.is_type( NodeId::TYPE_SERVER ) ||  my_node_id.is_type( 
        NodeId::TYPE_SERVICE ) ) && node.is_type( NodeId::TYPE_SERVER );
    };


std::string Network::generate_blowfish_key
    (
    const std::string & other_public_key
    )
    {
    int max_key_len;
    int my_max_key_len;
    int len = settings.blowfish_key_length;

    max_key_len = rsa.get_max_text_len_for_encoding
        ( 
        other_public_key 
        );
    
    my_max_key_len = rsa.get_max_text_len_for_encoding
        ( 
        rsa.get_public_key()
        );
    
    if( max_key_len != my_max_key_len )
        {
        panic( "RSA ciphers with different strength used, please use the same "
            "bit strength!" );
        }

    if( len > max_key_len )
        {
        len = max_key_len;
        }    

    if( len == 0 )
        {
        panic( "RSA ciphers are too weak to encrypt the blowfish key!" );
        }

    return RSACipher::generate_seed( len );
    }


bool Network::is_tcp_packet_waiting
    (
    ConnectionInfo *    conn_info
    )
    {
    hope( conn_info );
    hope( conn_info->get_tcpsocket() );
    
    return conn_info->get_tcpsocket()->receive( recvbuf, recvbuf_size );
    }


bool Network::is_udp_datagram_waiting
    (
    ConnectionInfo *    conn_info
    )
    {
    hope( conn_info );
    hope( conn_info->get_udpsocket() );
    
    return conn_info->get_udpsocket()->receive( recvbuf, recvbuf_size );
    }


int Network::accept_on_well_known_port_server()
    {
    hope( well_known_port_socket != NULL );

    Socket *            newsocket;
    int                 accepted = 0;

    while( ( newsocket = well_known_port_socket->accept() ) != NULL )
        {
        BitOStream bos;

        std::ostringstream  s;

        s << "Accept on well known port.";

        Global::log
            (
            STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
            s.str()
            );

        write_packet_header( bos, SERVER_PORT_MESSAGE, 4 );
        bos.write_int( NodeId::convert_to_uint32( get_local_node_id() ), 32 );
        try
            {
            newsocket->send( bos );
            }
        catch( ExSendError & )
            {
            std::ostringstream  s;

            s << "Cannot send SERVER_PORT_MESSAGE to socket!";

            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_NETWORK,
                    Status::PRIORITY_HIGHEST
                    ),
                s.str()
                );
            newsocket->close();
            break;
            }
        net_stat.inc_tcp_bytes_sent( bos.get_buffer().size() );
        net_stat.inc_messages_sent( 1 );
        
        ConnectionInfo * connection_info =
            new ConnectionInfo( true );

        connection_info->set_tcpsocket( newsocket );
        connection_info->set_connection_status
            (
            ConnectionInfo::NODE_ID_EXCHANGE_SERVER
            );

        connection_info->touch_last_message_time();

        try
            {
            conn_list.push_back
                (
                connection_info
                );
             }
        catch( ... )
            {
            newsocket->close();
            delete newsocket;
            }

        accepted++;
        }

    return accepted;
    }  // Network::accept_on_well_known_port_server()


std::string Network::get_key_from_node_id
    (
    const NodeId & nodeid
    )
    {
    NodeDatabase &          node_database = Global::node_database();
    NodeDatabase::NodeData  node_data;

    #ifdef OPTION_DEBUG

    std::ostringstream oss;
    oss << "Asking for public key of node " << nodeid << ".";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        oss.str()
        );
    
    #endif

    node_data = node_database.get_node_data( nodeid );
    return node_data.get_rsa_public_key();
    }


bool Network::send_ping_if_idle
    (
    ConnectionInfo *    conn_info
    )
    {
    hope( conn_info );
    conn_info->inc_idle_ticks();

    if( conn_info->get_connection_status() != ConnectionInfo::CONNECTED && 
        conn_info->get_connection_status() != ConnectionInfo::TRYING_TO_DISCONNECT  )
        {
        if( Global::system_time() - conn_info->get_last_message_time() >
                conn_info->get_max_idle_seconds() )
            {
            // I am not in a CONNECTED state and i have timeout => kill connection
            // do not send any ping packet. If there's another connection, it's ok, becouse
            // redundant connection is closing.

            if( !( conn_info->get_node_id().is_null() ) )
                {
                std::ostringstream  s;

                if( !is_node_connected( conn_info->get_node_id() ) )
                    {
                    s << "Timeout during creating connection to ";
                    }
                else 
                    {
                    s << "Closing redundant connection to ";
                    }

                s << conn_info->get_node_id() << ".";
            
                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
                    s.str()
                    );
                }

            

            conn_info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
            conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );

            return false;
            }
        }
    else
        {
        if( Global::system_time() - conn_info->get_last_message_time() >
            conn_info->get_max_idle_seconds() +
            conn_info->get_max_idle_seconds() *
            ( conn_info->get_next_ping_delay() <<
            ( conn_info->get_ping_messages_sent() - 1 ) ) )
            {
            // Idle too long --> sending PING MESSAGE.
            if( conn_info->get_last_message_time() != SystemTime::INVALID )
                {
                net_stat.inc_ping_timeouts();
                if( conn_info->get_permit_auto_close() &&
                    conn_info->get_ping_messages_sent() < conn_info->get_max_pings() &&
                    my_node_id.get_type() != NodeId::TYPE_CLIENT &&
                    my_node_id.get_type() != NodeId::TYPE_ANONYMOUS )
                    {

                    std::ostringstream  s;

                    s   << "Requesting to close connection to node "
                        << conn_info->get_node_id() << ".";

                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                        s.str()
                        );
                    try
                        {
                        send_uncrypted_message
                            (
                            DISCONNECT_REQUEST_MESSAGE,
                            conn_info->get_tcpsocket(),
                            VariantPointer()
                            );
                        conn_info->set_connection_status( ConnectionInfo::TRYING_TO_DISCONNECT );
                        }
                    catch( ExSendError & )
                        {
                        conn_info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
                        conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
                        return false;
                        }
                    
                    conn_info->inc_ping_messages_sent();
                    }
                else
                if( conn_info->get_ping_messages_sent() <
                    conn_info->get_max_pings() )
                    {
                    std::ostringstream  s;

                    s << "Sending ping packet to node " << conn_info->get_node_id() << ".";

                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                        s.str()
                        );

                    try
                        {
                        send_uncrypted_message
                            (
                            PING_MESSAGE,
                            conn_info->get_tcpsocket(),
                            VariantPointer()
                            );
                        }
                    catch( ExSendError & )
                        {
                        conn_info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
                        conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
                        }

                    conn_info->inc_ping_messages_sent();
                    }
                else
                    {
                    // The other side isn't responding.

                    std::ostringstream  s;

                    s << "Closing connection due to timeout.";

                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
                        s.str()
                        );

                    conn_info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
                    conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );

                    return false;
                    }
                }
            }
        }

    return true;
    }  // Network::send_ping_if_idle()


bool Network::send_recomputed_blowfish_if_expired
    (
    const NodeId &      destination,
    ConnectionInfo *    conn_info
    )
    {
    hope( conn_info );
    bool result = false;

    if( conn_info->get_connection_status() != ConnectionInfo::CONNECTED )
        {
        return false;
        }

    if( (   conn_info->get_blowfish_key_period() !=
            ConnectionInfo::BLOWFISH_PERIOD_ODD ) &&
        (   conn_info->get_blowfish_key_period() !=
            ConnectionInfo::BLOWFISH_PERIOD_EVEN ) )
        {
        return false;
        }

    if( conn_info->is_server_side() )
        {
        if( settings.blowfish_key_timeout != SystemTime::INVALID )
            {
            SystemTime  diff_time =
                Global::system_time() -
                conn_info->get_blowfish_key_timestamp();

            if( diff_time.get_value() > settings.blowfish_key_timeout )
                {
                // Log exchange of the key.
                net_stat.inc_blowfish_expirations();
                
                std::ostringstream  ss;
                ss << "Initiating a runtime exchange of new blowfish key, period "
                   << conn_info->get_blowfish_key_period() << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
                    ss.str()
                    );

                // Recompute the blowfish key.
                std::string  key = generate_blowfish_key
                    (
                    conn_info->get_other_side_public_key()
                    );

                // Send the key.
                send_crypted_message
                    (
                    BLOWFISH_NEW_KEY_MESSAGE,
                    destination,
                    VariantPointer( &key )
                    );

                // Set the new key.
                conn_info->set_blowfish_key( key );

                // Set the new blowfish resend algorithm phase.
                if( conn_info->get_blowfish_key_period() ==
                    ConnectionInfo::BLOWFISH_PERIOD_ODD )
                    {
                    conn_info->set_blowfish_key_period
                        (
                        ConnectionInfo::BLOWFISH_PERIOD_EVEN_TRANSITION
                        );
                    }
                else  // blowfish_key_period == ConnInfo::BLOWFISH_PERIOD_EVEN
                    {
                    conn_info->set_blowfish_key_period
                        (
                        ConnectionInfo::BLOWFISH_PERIOD_ODD_TRANSITION
                        );
                    }

                result = true;
                }
            }
        }

    return result;
    }  // Network::recompute_blowfish_key_if_expired()


int Network::receive_in_node_id_exchange_server_state
    (
    ConnectionInfo *     conn_info
    )
    {
    int result = 0;

    hope( conn_info );
    try
        {
        if( is_tcp_packet_waiting( conn_info ) )
            {
        
            tcp_packet_received( conn_info, recvbuf_size );
            
            UInt16 count = recvbuf.read_int( 16 );
            if( count != 1 )
                {
                throw ExMessagesNumberMismatch( msg_messages_number_mismatch );
                }

            UInt32 tag = recvbuf.read_int( 32 );

            if( tag == NODE_ID_MESSAGE )
                {
                recvbuf.skip( 32 );
                
                NodeId  sid = NodeId::create_from_uint32( recvbuf.read_int( 32 ) );
                
                if( sid.is_type( NodeId::TYPE_ANONYMOUS ) )
                    {
                    sid = find_unique_anonymous_node_id();
                    hope( sid.is_type( NodeId::TYPE_ANONYMOUS ) );
                    }
                
                std::ostringstream  s;

                s << "Client introduced as " << sid << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_NORMAL ),
                    s.str()
                    );
                
                if( !sid.is_null() )
                    {
                    conn_info->set_node_id( sid );
                    
                    std::string key;
                    
                    try
                        {
                        key = get_key_from_node_id( conn_info->get_node_id() );
                        }
                    catch( ... ) // FIXME - define exception
                        {
                        send_uncrypted_message
                            (
                            NOT_AUTHORIZED_MESSAGE,
                            conn_info->get_tcpsocket(),
                            VariantPointer()
                            );
                            
                        throw ExConnectError( "Unknown client! Closing connection." );
                        }
                                        
                    conn_info->set_other_side_public_key
                        (
                        key
                        );

                    send_uncrypted_message
                        (
                        BLOWFISH_KEY_MESSAGE,
                        conn_info->get_tcpsocket(),
                        VariantPointer( conn_info )
                        );

                    // Set new blowfish cipher.

                    conn_info->set_connection_status
                        (
                        ConnectionInfo::BLOWFISH_EXCHANGE_SERVER
                        );                   
                   
                    }
                else
                    {
                    send_uncrypted_message
                        (
                        NOT_AUTHORIZED_MESSAGE,
                        conn_info->get_tcpsocket(),
                        VariantPointer()
                        );

                    throw ExConnectError( "Client not authorized to connect!" );
                    }
                }
            else
                {
                if( !process_error_message( tag ) )
                    {
                    // Bad packet received.
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    }
                }

            result++;
            }
        }
    catch( ... )
        {
        bad_packet_received( conn_info );
        }

    return result;
    }  // Network::receive_in_node_id_exchange_server_state()


int Network::receive_in_blowfish_exchange_client_state
    (
    ConnectionInfo *    conn_info
    )
    {
    int result = 0;

    hope( conn_info );

    // Only TCP connection at this time.
    try
        {
        if( is_tcp_packet_waiting( conn_info ) )
            {
            tcp_packet_received( conn_info, recvbuf_size );

            recvbuf.skip( 16 );

            UInt32 tag = recvbuf.read_int( 32 );

            if( tag == BLOWFISH_KEY_MESSAGE )
                {
                std::ostringstream  s;

                s   << "Received blowfish key from node "
                    << conn_info->get_node_id() << ".";

                Global::log
                    (
                    STATUS_DEBUG
                        (
                        Status::FACILITY_NETWORK,
                        Status::PRIORITY_HIGHEST
                        ),
                    s.str()
                    );
                            
                std::string decrypted_key;
                std::string crypted_key;

                recvbuf.skip( 32 );
                crypted_key = recvbuf.read_full_string();

                if( rsa.verify
                        (
                        crypted_key,
                        recvbuf.read_full_string(),
                        conn_info->get_other_side_public_key()
                        ) )
                    {
                    
                    if( rsa.decrypt( crypted_key, decrypted_key ) == 0 )
                        {
                        // Bad decryption.
                        
                        throw ExConnectError( "Authentization failed - cannot "
                            "decrypt blowfish cipher." );
                        };
                    
                    #ifdef OPTION_DEBUG

                    std::ostringstream  ss;
                    ss << "Received blowfish key " << quote_const_string( decrypted_key ) <<
                         " from node " << conn_info->get_node_id() << ".";
                    Global::log_debug( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST, ss.str() );
                    
                    #endif

                    // Succesfully verified the message.

                    conn_info->set_blowfish_key_odd( decrypted_key );
                    conn_info->set_blowfish_key_even( decrypted_key );

                    send_uncrypted_message
                        (
                        KEY_CONFIRMATION_MESSAGE,
                        conn_info->get_tcpsocket(),
                        VariantPointer( conn_info )
                        );

                    conn_info->set_connection_status
                        (
                        ConnectionInfo::UDP_PORT_RECEIVE_CLIENT
                        );

                    // Note: Don't use set_blowfish_key here - it would set
                    //  only the even-phase key because this connection is
                    //  in a BLOWFISH_PERIOD_ODD phase!!!
                    }
                else
                    {
                    // Bad signature.
                    throw ExMessageCorrupted( "Bad signature received." );
                    }
                }
            else
                {
                if( !process_error_message( tag ) )
                    {
                    // Bad packet received.
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    }
                }

            result++;
            }
        }
    catch( ExConnectError & )
        {
        throw;
        }
    catch( ... )
        {
        bad_packet_received( conn_info );
        }

    return result;
    }  // Network::receive_in_blowfish_exchange_client_state()


int Network::receive_in_blowfish_exchange_server_state
    (
    ConnectionInfo *                          conn_info
    )
    {
    int result = 0;
    
    hope( conn_info );

    // Only TCP connection at this time.
    try
        {
        if( is_tcp_packet_waiting( conn_info ) )
            {
            tcp_packet_received( conn_info, recvbuf_size );

            UInt16 count = recvbuf.read_int( 16 );
            if( count != 1 )
                {
                throw ExMessagesNumberMismatch( msg_messages_number_mismatch );
                }

            UInt32 tag = recvbuf.read_int( 32 );

            if( tag == KEY_CONFIRMATION_MESSAGE )
                {
                recvbuf.skip( 32 );

                if( recvbuf.read_full_string() == conn_info->get_blowfish_key_even() )
                    {
                    conn_info->touch_blowfish_key_timestamp();

                    // Create UDP socket & server.

                    UDPSocket * udpsocket = create_udp_socket();
                    hope( udpsocket );

                    if( !udpsocket->start_server( 0 ) )
                        {
                        panic( "Cannot bound port!" );
                        }

                    UInt16 free_port = udpsocket->get_bound_port();
                    if( free_port == 0 )
                        {
                        panic( "NULL port bound!" );                                
                        }
                        
                    conn_info->set_udpsocket( udpsocket );

                    conn_info->set_connection_status
                        (
                        ConnectionInfo::UDP_CONNECTION_SETUP_SERVER
                        );

                    // Send udp port message which also contains node_id of the client.
                    send_uncrypted_message
                        ( 
                        UDP_PORT_MESSAGE, 
                        conn_info->get_tcpsocket(), 
                        VariantPointer( &free_port ), 
                        STREAM_RELIABLE 
                        );
                    
                    }
                else
                    {
                    // Bad confirmation text received.
                    throw ExMessageCorrupted( "Corrupted body of the packet." );
                    }
                }
            else
                {
                if( !process_error_message( tag ) )
                    {
                    // Bad packet received.
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    }
                }

            result++;
            }
         }
    catch( ExConnectError & )
        {
        throw;
        }
    catch( ... )
        {
        net_stat.inc_bad_packets_received();
        conn_info->set_disconnection_reason
            (
            ConnectionInfo::ERROR_REASON
            );
        conn_info->set_connection_status
            (
            ConnectionInfo::NOT_CONNECTED
            );
        };
    return result;
    }  // Network::receive_in_blowfish_exchange_server_state()


int Network::receive_in_udp_port_receive_client_state
    (
    ConnectionInfo *        conn_info
    )
    {
    int result = 0;
    int len = 0;

    hope( conn_info );

    // Only TCP connection at this time.
    try
        {
        if( is_tcp_packet_waiting( conn_info ) )
            {
            tcp_packet_received( conn_info, recvbuf_size );

            UInt16 count = recvbuf.read_int( 16 );
            if( count != 1 )
                {
                throw ExMessagesNumberMismatch( msg_messages_number_mismatch );
                }

            UInt32 tag = recvbuf.read_int( 32 );

            if( tag == UDP_PORT_MESSAGE )
                {
                len = recvbuf.read_int( 32 );

                UInt32      node_id   = recvbuf.read_int( 32 );
                UInt32      udp_port  = recvbuf.read_int( 16 );

                // NodeId to be sure, I've got packet from proper server.
                NodeId  temporary_node_id = NodeId::create_from_uint32( node_id );
                hope( temporary_node_id == conn_info->get_node_id() );
                
                if( temporary_node_id.is_type( NodeId::TYPE_ANONYMOUS ) )
                    {
                    throw ExInvalidMessage
                        ( "Received UDP port message from anonymous node." );
                    }

                UDPSocket * udpsocket = create_udp_socket();
                hope( udpsocket );
                hope( conn_info->get_tcpsocket() );
                
                conn_info->set_udpsocket( udpsocket );
                ServerAddress sa
                    (
                    conn_info->get_tcpsocket()->server_address.get_dns_name(),
                    udp_port
                    );
                    {
                    // Port printout.
                    std::ostringstream ss;
                    ss << "DEBUG: UDP want to connect to " << temporary_node_id << " port " << udp_port << ".";
                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                        ss.str()
                        );
                    }
                if( !udpsocket->connect( sa ) )
                    {
                    std::ostringstream  s;

                    s << "Cannot ::connect() UDP CONNECT.";
                    throw ExSendError( s.str() );
                    }

                // Send initialization packet.
               
                init_udp_list.push_back( conn_info );
                
                send_uncrypted_message
                    ( 
                    TCP_CONNECT_MESSAGE, 
                    conn_info->get_tcpsocket(), 
                    VariantPointer() 
                    );

                NodeManager & node_mgr = Global::node_manager();
                
                if( node_mgr.net_callback_connection_established( temporary_node_id ) !=
                    NodeManager::RESULT_OK )
                    {
                    throw ExConnectError( "Node manager denied the connection." );
                    }

                conn_info->set_connection_status( ConnectionInfo::CONNECTED );
                conn_info->set_node_id( temporary_node_id );

                std::ostringstream  s;

                s   << "Connected to server nodeid " << temporary_node_id << " as " 
                    << my_node_id << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                    s.str()
                    );
                }
            else
                {
                if( !process_error_message( tag ) )
                    {
                    // Bad packet received.
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    }
                }

            result++;
            }
        }
    catch( ExConnectError & )
        {
        throw;
        }
    catch( ... )
        {
        bad_packet_received( conn_info );
        }

    return result;
    }  // Network::receive_in_node_id_receive_client_state()


int Network::receive_in_udp_connection_setup_server_state
    (
    ConnectionInfo *                           conn_info,
    std::list< ConnectionInfo * >::iterator &  iter
    )
    {
    int result = 0;
    int len = 0;
    
    NodeManager & node_mgr = Global::node_manager();
    hope( conn_info );
    
    std::list< ConnectionInfo * >::iterator pom = iter++;

    // Wait for UDP initialization packet.
    try
        {
        if( is_tcp_packet_waiting( conn_info ) )
            {
            NodeId sid = conn_info->get_node_id();

            tcp_packet_received( conn_info, recvbuf_size );

            UInt16 count = recvbuf.read_int( 16 );
            if( count != 1 )
                {
                std::ostringstream oss;
                oss << sid << " tag: " << recvbuf.read_int( 32 );
                oss << " len: " << recvbuf.read_int( 32 );
                oss << msg_messages_number_mismatch;
                throw ExMessagesNumberMismatch( oss.str() );
                throw ExMessagesNumberMismatch( msg_messages_number_mismatch );
                }

            UInt32 tag = recvbuf.read_int( 32 );
            if( tag == TCP_CONNECT_MESSAGE )
                {
                len = recvbuf.read_int( 32 );
                
                if( len != 0 )
                    {
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    } 
                
                // Insert into hash table.
                
                NodeHash::iterator iterh =
                        node_table.find( sid );

                if( iterh != node_table.end() 
                    && is_server_connection( sid ) )
                    {
                    std::ostringstream  s;

                        s   << "Found server connection to node "
                            << conn_info->get_node_id()
                            << " status: " << iterh->second->get_connection_status();

                        Global::log
                            (
                            STATUS_INFO
                                (
                                Status::FACILITY_NETWORK,
                                Status::PRIORITY_NORMAL
                                ),
                            s.str()
                            );


                    if( NodeId::Less()( sid, my_node_id ) )
                        {
                        hope( iterh != node_table.end() );

                        {
                        std::ostringstream  s;

                        s   << "Closing old connection to node "
                            << sid
                            << " and creating a new one." << my_node_id;

                        Global::log
                            (
                            STATUS_INFO
                                (
                                Status::FACILITY_NETWORK,
                                Status::PRIORITY_NORMAL
                                ),
                            s.str()
                            );
                        }

                        // All messages prepared for sending must be sent through new connection.
                        route_connection( sid, conn_info );

                        std::list< ConnectionInfo * >::iterator it;
                        it = std::find( init_udp_list.begin(), init_udp_list.end(), iterh->second );
        
                        if( it != init_udp_list.end() )
                            {
                            hope( *it );
                            init_udp_list.erase( it );
                            }

                        delete iterh->second;
                        node_table.erase( iterh );

                        {
                        std::ostringstream  s;

                        s   << "Old connection to node "
                            << sid
                            << " killed and new route prepared." << my_node_id;

                        Global::log
                            (
                            STATUS_INFO
                                (
                                Status::FACILITY_NETWORK,
                                Status::PRIORITY_NORMAL
                                ),
                            s.str()
                            );
                        }
                        }
                    else
                        {
                        std::ostringstream s;

                        s << "Connection with node " << sid;
                        s << " already exists and I am the slave!" << std::endl;
                        //throw ExConnectionExistsError( s.str() );
                        Global::log
                            (
                            STATUS_INFO
                                (
                                Status::FACILITY_NETWORK,
                                Status::PRIORITY_NORMAL
                                ),
                            s.str()
                            );

                        send_uncrypted_message
                            ( 
                            DISCONNECT_FORCE_MESSAGE, 
                            conn_info->get_tcpsocket(), 
                            VariantPointer()
                            );

                        return 1;
                        }
                    }
                
                                
                // If there is not cleaned connection in the same tick connection, clean it.
                // The other side will close it as redundant.
                if( iterh != node_table.end() && !is_server_connection( sid ) )
                    {
                    hope( sid.is_type( NodeId::TYPE_CLIENT ) );
                    send_uncrypted_message
                            ( 
                            ALREADY_CONNECTED_MESSAGE, 
                            conn_info->get_tcpsocket(), 
                            VariantPointer()
                            );
                    delete conn_info;
                    conn_list.erase( pom );
                    return 1;
                    }
                    
                hope( node_table.find( sid ) == node_table.end() );
                node_table.insert
                    (
                    Network::NodeHash::value_type( sid, conn_info )
                    );

                conn_list.erase( pom );
                
                // Node manager callback.    
                if( node_mgr.net_callback_connection_established( sid ) !=
                    NodeManager::RESULT_OK )
                    {
                    send_uncrypted_message
                        ( 
                        NOT_READY_TO_LOGIN_MESSAGE, 
                        conn_info->get_tcpsocket(), 
                        VariantPointer()
                        );
                    
                    // Log operation.
                    std::ostringstream s;
                    s   << "Node manager denied the connection to node " 
                        << sid << ".";

                    Global::log
                        (
                        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                        s.str()
                        );
                    
                    conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
                    return 1;
                    }
                
                conn_info->set_connection_status( ConnectionInfo::CONNECTED );
                
                std::ostringstream  s;
                s << "Connected to client node " << sid << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                    s.str()
                    );
                
                send_uncrypted_message
                    ( 
                    TCP_CONFIRMATION_MESSAGE, 
                    conn_info->get_tcpsocket(), 
                    VariantPointer()
                    );
                conn_info->set_tcp_confirmed( true );
                }
            else
            if( tag == DISCONNECT_READY_MESSAGE )
                {
                std::ostringstream  s;
                s << "Received acknowledgement to close connection to node " << sid << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                    s.str()
                    );

                conn_info->set_disconnection_reason( ConnectionInfo::REQUEST_REASON );
                send_uncrypted_message
                    ( 
                    NODE_DISCONNECTED_MESSAGE, 
                    conn_info->get_tcpsocket(), 
                    VariantPointer( conn_info )
                    );

                delete conn_info;
                conn_list.erase( pom );
                }
            else
                {
                if( !process_error_message( tag ) )
                    {
                    // Bad packet received.
                    throw ExMessageCorrupted( msg_not_awaited_message );
                    }
                }

            result++;
            }
        }
    catch( ExConnectError & )
        {
        throw;
        }
    catch( ... )
        {
        net_stat.inc_bad_packets_received();

        conn_info->set_disconnection_reason
            ( ConnectionInfo::ERROR_REASON );
        conn_info->set_connection_status
            ( ConnectionInfo::NOT_CONNECTED );

        }

    return result;
    }  // Network::receive_in_udp_connection_setup_server_state()


int Network::receive_in_connected_state
    (
    ConnectionInfo *        conn_info
    )
    {
    static NetIBuffer           buffer_head;
    int                         result = 0;

    hope( conn_info );

    // Receive TCP data.

    try
        {
        while( is_tcp_packet_waiting( conn_info ) )
            {
            tcp_packet_received( conn_info, recvbuf_size );

            if( !process_tcp_packet( recvbuf, conn_info ) )
                {
                // I've received system_message.
                std::ostringstream s;

                s   << "Received a system message from "
                    << conn_info->get_node_id() << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                    s.str()
                    );
                }
            
            result++;
            }
        }
    catch( ... )
        {
        bad_packet_received( conn_info );
        return 0;
        }

    // Receive UDP data.
    try
        {
        while( is_udp_datagram_waiting( conn_info ) )
            {
            udp_datagram_received( conn_info, recvbuf_size );
            result = decrypt_and_process_udp_datagram( recvbuf , conn_info );
            }
        }
    catch( ExMessageCorrupted )
        {
        net_stat.inc_bad_packets_received();
        // Bad udp packet is not killing the connection.
        // Just ignore the bad packet.
        }
    catch( ... )
        {
        bad_packet_received( conn_info );
        }

    const NodeId & sid = conn_info->get_node_id();
    send_recomputed_blowfish_if_expired( sid, conn_info );

    return result;
    }  // Network::receive_in_connected_state()


bool Network::control_connection_auto_close
    (
    const NodeId &  node,
    bool            auto_close
    )
    {
    NodeHash::iterator  iter;
    bool                old_state;

    iter = node_table.find( node );
    if( iter != node_table.end() )
        {
        ConnectionInfo * conn = iter->second;
        hope( conn );
        
        old_state = conn->get_permit_auto_close();
        conn->set_permit_auto_close( auto_close );
        }
    else
        {
        throw ExNoSuchConnection( "The connection not found." );
        }

    return old_state;
    } // Network::control_connection_auto_close


void Network::init_udp()
    {
    std::list< ConnectionInfo * >::iterator it = init_udp_list.begin();
    std::list< ConnectionInfo * >::iterator pom;
    
    while( it != init_udp_list.end() )
        {
        ConnectionInfo * conn_info = *it;
        hope( conn_info );
        hope( conn_info->get_udpsocket() );
        
        try
            {
            pom = it++; 
            if( !conn_info->is_udp_connected() && 
                conn_info->get_connection_status() == ConnectionInfo::CONNECTED )
                {
                std::ostringstream oss;
                oss << "Sending initial UDP message to " 
                    << conn_info->get_node_id() << ".";

                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOW ),
                    oss.str()
                    );

                send_uncrypted_message
                    ( 
                    UDP_INIT_MESSAGE, 
                    conn_info->get_udpsocket(),
                    VariantPointer(), 
                    NetworkCommon::STREAM_UNRELIABLE
                    );
                }
            else
                {
                init_udp_list.erase( pom );
                }
            }
        catch( ... )
            {
            std::ostringstream oss;
            oss << "Cannot send initial UDP message to " 
                << conn_info->get_node_id() << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_HIGH ),
                oss.str()
                );

            init_udp_list.erase( pom );
            }
        }
    }


NodeId Network::find_unique_anonymous_node_id()
    {
    // For anonymous nodes create unique node id.
    // The node id must be unique only inside local node.
    Int32   relative_id = 1;
    bool    found = false;
    NodeId  unique_node_id = NodeId::create_from_type_and_relative_id
        ( NodeId::TYPE_ANONYMOUS, relative_id );
    
    while( !found )
        {
        NodeHash::const_iterator it = node_table.find( unique_node_id );
        bool                     failed = false;
        
        if( it != node_table.end() )
            {
            failed = true;
            }
            
        std::list< ConnectionInfo * >::const_iterator it_list;
        
        if( !failed )
            {
            for( it_list = conn_list.begin(); it_list != conn_list.end(); ++it_list )
                {
                const ConnectionInfo * conn_info = *it_list;
                hope( conn_info );
                NodeId  conn_node_id = conn_info->get_node_id();
                if( conn_node_id == unique_node_id )
                    {
                    failed = true;
                    break;
                    }
                }
            }
        
        if( !failed )
            {
            found = true;
            }
        else
            {
            relative_id++;
            unique_node_id = NodeId::create_from_type_and_relative_id
                ( NodeId::TYPE_ANONYMOUS, relative_id );
            }
        }
    
    #ifdef OPTION_DEBUG

    std::ostringstream oss;
    oss << "Found unique node id " << unique_node_id << ".";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        oss.str()
        );
    
    #endif

    return unique_node_id;
    }

bool Network::is_safe_to_disconnect
    (
    const NodeId & node
    )
    {
       
    if( tcp_sendlist.size() != 0 )
        {
        std::list< TCPOBuffer >::iterator iter_tcp;

        iter_tcp = tcp_permlist.begin();
        while( iter_tcp != tcp_permlist.end() )
            {
            if( iter_tcp->owner == node )
                {
                return false;
                }
            iter_tcp++;
            }
        }

    if( udp_sendlist.size() != 0 )
        {
        std::list< UDPOBuffer >::iterator iter_udp;

        iter_udp = udp_permlist.begin();
        while( iter_udp != udp_permlist.end() )
            {
            if( iter_udp->owner == node )
                {
                return false;
                }
            iter_udp++;
            }
        }

    return true;
    }

} // namespace Massiv::Core
} // namespace Massiv
