/** @file net.h
 * Network layer interface.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Marek Vondrak, Ondrej Pecta, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NET_H
#define MASSIV_NET_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif

#ifndef MASSIV_NET_COMMON_H
#include "net/net_common.h"
#endif

#ifndef MASSIV_SOCKET_H
#include "net/socket.h"
#endif

#ifndef MASSIV_TCPOBUFFER_H
#include "net/tcpobuffer.h"
#endif

#ifndef MASSIV_UDPOBUFFER_H
#include "net/udpobuffer.h"
#endif

#ifndef MASSIV_BITISTREAM_H
#include "net/bitistream.h"
#endif

#ifndef MASSIV_CONNECTION_INFO_H
#include "net/connection_info.h"
#endif

#ifndef MASSIV_MASSIV_RSA_H
#include "net/massiv_rsa.h"
#endif

#ifndef MASSIV_NET_STATISTICS_H
#include "net/net_statistics.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#include <list>


namespace Massiv {
namespace Core {

namespace Test { class SocketsTest; };


/** @addtogroup net
 * @{
 */

/**
 * The network layer interface class.
 * Contains buffer locking/unlocking, packet crypting/decrypting etc.
 * Massiv uses both TCP and UDP connection to communicate between nodes.
 * The difference between reliable and unreliable packets is the
 * UDP buffer ability to contain more messages in one packet.
 * Each TCP packet has the following format:
 * - 10 bytes long header consisting of:
 *   bytes 0-1 : Number of messages - only 1 for TCP, later these bytes are
 *               used to determine correctly blowfish exchange phase.
 *   bytes 2-5 : Packet tag - makes difference between system and user messages.
 *               ( e.g. PING_MESSAGE, ALREADY_CONNECTED_MESSAGE etc. ).
 *   bytes 6-9 : Size of the contained message. The smallest TCP messages are
 *               system messages mostly having size equal to 0. ( PING_MESSAGE ).
 * - Message body.
 * 
 * Each UDP packet has the following format:
 * - 10 bytes long header consisting of:
 *   bytes 0-1 : Number of messages - can be >= 1.
 *   bytes 2-5 : Message tag - makes difference between system and user messages.
 *               ( e.g. PING_MESSAGE, ALREADY_CONNECTED_MESSAGE etc. ).
 *   bytes 6-9 : Size of the contained message (if the buffer contains more
 *               than one message, this is the size of the first one).
 * - Message body.
 * If number of messages is > 1 other messages follow in sequence:
 * 4 bytes - message tag.
 * 4 bytes - size of the body - x.
 * x bytes - message body. etc etc.
 * - Then follows 1 byte representing blowfish period ( odd x even ).
 * - At last a 4 byte CRC32 Internet checksum is added ( from uncrypted body ).
 * @note UDP packed is crypted almost entire ( except last 5 bytes ). 
 * That's the difference because TCP have header uncrypted and crypted is only 
 * the message body. The only way to cut received TCP message from the incoming 
 * stream is to analyze the header and to read required amount of bytes. That's 
 * why header is not crypted with the message body.
 */
class Network
    {
    friend class Massiv::Core::ConnectionInfo;
    friend class Massiv::Core::Test::SocketsTest;

public:

    /**
     * Default constructor.
     */
    Network();

    /**
     * Default destructor.
     */
    ~Network();

    /**
     * Dumps network state.
     */
    std::string to_string() const;

    /**
     * Initialize.
     * @pre NodeDatabse is already initiazed.
     */
    void initialize();

    /**
     * Assign local node id to the system.
     */
    void assign_local_node_id
        (
        const NodeId & local_node_id
        );

public:

    /**
     * Network-layer-specific message tags.
     */
    enum RegisteredMessage
        {
        SERVER_PORT_MESSAGE             = 0,
            /**< When a client wants to establish connection to a server,
                 it connects to server's "well-known-port". After accepting the
                 connection by server, the server creates another socket
                 (bound with another port) and tells client to connect to
                 this new port. A SERVER_PORT_MESSAGE contains the number
                 of this new port. */

        PING_MESSAGE                    = 1,
            /**< "Are you alive?" message. */

        PING_REPLY_MESSAGE              = 2,
            /**< "I am alive" message. */

        UDP_PORT_MESSAGE                = 5,
            /**< At the end of the connection exchange, server sends port,
                where udp server is sitting. */

        TCP_CONNECT_MESSAGE             = 6,
            /**< At the end of the connection exchange, client sends via TCP
                 this message to inform server, that TCP connection is ok
                 and UDP port received ok.*/

        TCP_CONFIRMATION_MESSAGE        = 7,
            /**< Server sends this message to client when TCP_CONNECT_MESSAGE
                 received.*/

        UDP_HANDSHAKE_MESSAGE           = 8,
            /**< Server sends this message to client when UDP_INIT_MESSAGE
                 received.*/

        UDP_INIT_MESSAGE                = 10,
            /**< When UDP not connected, client sends periodically
                 this message via UDP(!) to inform server about other side until
                 UDP_HANDSHAKE_MESSAGE received. */

        NODE_DISCONNECTED_MESSAGE       = 1000,
            /**< Message used when user legally disconnects from simulation. */

        NODE_ID_MESSAGE                 = 1001,
            /**< Message containing other side NodeID. */

        BLOWFISH_KEY_MESSAGE            = 1002,
            /**< Message containing blowfish key. */

        KEY_CONFIRMATION_MESSAGE        = 1003,
            /**< Message that is sent to confirm receiving message containing
                 either the RSA public key or blowfish key. */

        BLOWFISH_NEW_KEY_MESSAGE        = 1004,
            /**< Message initiating the new (recomputed) blowfish key
                 exchange.
                 @note Unlike BLOWFISH_KEY_MESSAGE, this message is being
                    encrypted using the actual blowfish key. */

        NEW_KEY_CONFIRMATION_MESSAGE    = 1005,
            /**< Acknowledgement to the BLOWFISH_NEW_KEY_MESSAGE.
                 @note The message is crypted using the new blowfish key. */

        DISCONNECT_REQUEST_MESSAGE      = 1006,
                /**< Server with greater NodeId can send this message
                 to other server to close temporary connection. */

        DISCONNECT_PERMIT_MESSAGE       = 1007,
                /**< Positive acknowledgement to the DISCONNECT_REQUEST_MESSAGE. */

        DISCONNECT_DENY_MESSAGE         = 1008,
                /**< Negative acknowledgement to the DISCONNECT_REQUEST_MESSAGE.
                     I have locked buffers for this server, cannot disconnect. */
        
        DISCONNECT_FORCE_MESSAGE         = 1009,
                /**< An order from one server to other when redundant 
                    connection detected to the DISCONNECT_REQUEST_MESSAGE. */
        
        DISCONNECT_READY_MESSAGE         = 1010,
                /**< Positive acknowledgement to the DISCONNECT_FORCE_MESSAGE. */
                
        ALREADY_CONNECTED_MESSAGE       = 2000,
            /**< When client wants to connect twice to the same server,
                 receives this message during key exchange.
                 @note The message is sent uncrypted. */

        NOT_AUTHORIZED_MESSAGE          = 2001,
            /**< Client not authorized to connect to simulation. */

        NOT_READY_TO_LOGIN_MESSAGE      = 2002,
            /**< Simulation is not ready to accept connection. */

        NET_MIN_USER_MESSAGE            = 10000,
            /**< Minimal number of the higher-layers-defined message.
                 Each message tag that is not the network-layer-specific
                 should be defined as
                    NET_MIN_USER_MESSAGE + some_positive_value.
                 Note that the message tag must be an 32-bit integer value!
                 */
        };  // enum Net::RegisteredMessage

    /**
     * Number of permanent TCP buffers.
     */
    enum
        {
        DEFAULT_TCP_PREALLOCATED_BUFFERS = 10
        };

    /**
     * Number of permanent UDP buffers.
     */
    enum
        {
        DEFAULT_UDP_PREALLOCATED_BUFFERS = 10
        };

    /**
     * Default number of seconds the connection can be idle for before
     * sending the PING message.
     */
    enum
        {
        DEFAULT_MAX_IDLE_SECONDS = 10
        };

    /**
     * Delay before the next PINGs are sent. See @a next_ping_delay.
     */
    enum
        {
        DEFAULT_NEXT_PING_DELAY = 2
        };

    /**
     * Maximum number of PING message sent before the connection will be
     * brought down.
     */
    enum
        {
        DEFAULT_MAX_PINGS = 3
        };

    /**
     * Maximal length of the server life measured seconds. This doesn't
     * concern the well-known port server, which is alive all the time.
     */
    enum
        {
        DEFAULT_MAX_SERVER_LIFE = 60
        };

    /**
     * Length of life of one instance of one specific blowfish key IN
     * SECONDS (unlike most other settings). After this timeout, the key
     * will be recomputed. Value of 0 means that the key will be NOT
     * recomputed.
     */
    enum
        {
        DEFAULT_BLOWFISH_KEY_TIMEOUT = 3600
        };

    /**
     * Number of bytes received in one Global::next_tick() on one connection.
     * Higher value increase network throughput in cost of lower performance
     * of other massiv parts.
     * Default value is 32kb per one connection per tick.
     */
    enum
        {
        DEFAULT_MAX_RECEIVE_BYTES = 32 * 1024
        };

    /**
     * Default size of the blowfish key in bytes. 
     * See also @c generate_blowfish_key.
     */
    enum
        {
        DEFAULT_BLOWFISH_KEY_LENGTH = 14
        };

public:

    /**
     * Network registry settings.
     */
    struct NetworkSettings : public Settings
        {
    public:

        /**
         * Constructor
         */
        NetworkSettings();

    public:

        Integer &   tcp_preallocated_buffers;
            /**< Number of preallocated TCP (reliable) buffers. */

        Integer &   udp_preallocated_buffers;
            /**< Number of preallocated UDP (unreliable) buffers. */

        Integer &   max_idle_seconds;
            /**< After this number of seconds of no network traffic on
                 a connection the first PING message will be sent. */

        Integer &   next_ping_delay;
            /**< The next PING message after the first will be sent after
                 further next_ping_delay * max_idle_seconds miliseconds. */

        Integer &   max_pings;
            /**< Maximum number of PING message sent before the connection
                 will be brought down. */

        Integer &   max_server_life;
            /**< Maximal length of the server life (not regarding the
                 well-known-port server, which is alive all the time).
                 Measured in seconds. */

        Integer &   blowfish_key_timeout;
            /**< Length of life of one instance of one specific blowfish key in
                 seconds. After this timeout, the key will be recomputed.
                 Value of 0 means that the key will be NOT recomputed. */

        Integer &   max_receive_bytes;
            /**< Maximum number of bytes received from a socket in one network
                 tick. Greater values increase network throughput but decrease
                 performance of other massiv parts. */

        Integer &   blowfish_key_length;
            /**< Length of the generated blowfish key. Used by server-side when 
                 generating blowfish key. If the other side have so short RSA
                 cipher length, that cannot encrypt generated blowfish key, 
                 length is set to maximum value, that the client-side can
                 encrypt.. */
                 
        }; // class Network::NetworkSettings

public:

    /**
     * Hash table for nodes and their connection_infos.
     */
    typedef ext::hash_map
        <
        NodeId,
        ConnectionInfo *,
        NodeId::Hash,
        NodeId::EqualKey
        > NodeHash;

    /**
     * Get node id associated with the local system.
     */
    const NodeId & get_local_node_id();

    /**
     * Runs next network tick. Being called from higher layers.
     */
    void next_tick();

    /**
     * Creates a new tcp or udp buffer for given destination server if it doesn't
     * already exist. Note that sometimes no new buffer has to be allocated
     * as permanent buffer may be returned, etc.<br>
     * Sets buffer status to locked & writes a message tag.
     * calls lock_reliable or lock_unreliable method. Depend on
     * @param type The type of buffer. Can be either 
     *      <tt>NetworkCommon::STREAM_RELIABLE</tt> (for buffers that will be
     *      sent via TCP) or <tt>NetworkCommon::STREAM_UNRELIABLE</tt> (for
     *      buffers that will be sent via UDP).
     * @param id NodeId of server that should receive the buffer's content
     *      in the future.
     * @param tcp_send_immediately If true, the tcp message will be sent
     *      immediately after unlocking the buffer. Otherwise, the buffer
     *      will be queued into the queue of buffers waiting for sending
     *      by send_all_messages().<br>
     * @return Pointer to a NetOBuffer.
     *
     * @exception ExNoSuchConnection Connection to the server given by @a id
     *      is not established.
     */
    NetOBuffer * lock
        (
        const NodeId &              id,
        NetworkCommon::StreamType   type,
        UInt32                      tag,
        bool                        tcp_send_immediately = false
        );

    /**
     * Check whether net output buffer to @a node can be locked.
     *
     * @return True if the network has direct (opened) connection
     *   to the @a node or the network can open a connection
     *   to the @a node automatically by itself.
     *   Always return true if the @a node is server node, because
     *   the network must be able to open connection to server node
     *   automatically everytime.
     */
    bool can_lock_buffer_to_node
        (
        const NodeId &            node,
        NetworkCommon::StreamType stream_type
        );

    /**
     * Check whether connection creating is pending.
     * @param NodeId node to check.
     * @return True when connection exists in port_table or node table.
     */    
    bool is_connection_creating
        (
        const NodeId & node
        );


    /**
     * Check whether a connection to node @node exists.
     * This connection _must_ have CONNECTED status - only
     * connections in node table can be CONNECTED.
     *
     * @return True if the network has pending direct (opened) connection
     *   to the @a node.
     */
    bool is_node_connected
        (
        const NodeId & node
        );

    /**
     * Creates a new tcp buffer for given destination server if it doesn't
     * already exist. Note that sometimes no new buffer has to be allocated
     * as permanent buffer may be returned, etc.<br>
     * Sets buffer status to locked & writes a message tag.
     *
     * @param id NodeId of server that should receive the buffer's content
     *      in the future.
     * @param tcp_send_immediately If true, the message will be sent
     *      immediately after unlocking the buffer. Otherwise, the buffer
     *      will be queued into the queue of buffers waiting for sending
     *      by send_all_messages().<br>
     * @return Pointer to a TCPOBuffer.
     *
     * @exception ExNoSuchConnection Connection to the server given by @a id
     *      is not established.
     */
    TCPOBuffer * lock_reliable
        (
        const NodeId &  id,
        UInt32          tag,
        bool            tcp_send_immediately = false
        );

    /**
     * Creates a new udp buffer for given destination server if it doesn't
     * already exist. Note that sometimes no new buffer has to be allocated
     * as permanent buffer may be returned, etc.<br>
     * Sets buffer status to locked & writes a message tag.
     *
     * @param id NodeId of server that should receive the buffer's content
     *      in the future.
     * @return Pointer to a UDPOBuffer.
     */
    UDPOBuffer * lock_unreliable
        (
        const NodeId &  id,
        UInt32          tag
        );

    /**
     * Unlocks the netbuffer. Calls either unlock_reliable() or
     * unlock_unreliable() according to the type of @a netbuffer.
     */
    void unlock
        (
        NetOBuffer *    netbuffer
        );

    /**
     * Sets a buffer status to NET_BUFFER_FLUSH_ALLOWED and unlocks the buffer.
     * If the buffer has been locked with tcp_send_immediately set to TRUE, it
     * will be immediately sent.
     */
    void unlock_reliable
        (
        TCPOBuffer *    netbuffer
        );

    /**
     * Sets a buffer status to @a flush_allowed and unlocks the buffer.
     * Settings flush_allowed to FALSE is important for packing more
     * messages into one UDP buffer, because system waits until flush(...)
     * is called before sending is initiated.
     * @param netbuffer The buffer to be unlocked.
     * @param flush_allowed See above.
     */
    void unlock_unreliable
        (
        UDPOBuffer *    netbuffer,
        bool            flush_allowed
        );

    /**
     * For explicit flushing of unreliable buffers.
     * It sets buffer's status to NET_BUFFER_FLUSH_ALLOWED, it will be
     * sent during the next net tick.
     * @note Reliable buffers flushing is automatic after unlocking, because 
     * they can contain only 1 message.
     */
    void flush
        (
        UDPOBuffer * netbuffer
        );

    /**
     * For previously locked TCPOBuffer buffer dropping - when we do not
     * want to send this message to the server.
     * @param netbuffer to drop down.
     */
    void drop
        (
        TCPOBuffer *  netbuffer
        );

    /**
     * For previously locked UDPOBuffer buffer dropping - when we do not
     * want to send this message ( these messages ) to the server.
     * @param netbuffer to drop down.
     */
    void drop
        (
        UDPOBuffer *  netbuffer
        );

    /**
     * Drops all previously locked tcp and udp buffers.
     */
    void drop_all();

    /**
     * Drops all previously locked tcp and udp buffers to specified server.
     * Used when servers are connecting simultaneously to each other.
     * @param nodeid NodeId to drop messages to.
     * @param really_drop False when we do not want really drop the message, 
     * but becouse we need to keep messages for further delivery on new
     * connection.
     */
    void drop_all_to_server
        (
        const NodeId & node,
        bool  really_drop = true
        );

    /**
     * Prepares permanent buffer for reusing or deletes dynamic buffer.
     * @param netbuffer Buffer to reset or delete.
     * @note called from send_message after message successfully sent.
     */
    void reset_buffer
        (
        NetOBuffer * netbuffer 
        );

    /**
     * When there are messages prepared for sending to a server and we are
     * waiting for server connection to become available. This function will
     * perform server connection lookup a and fill connection info needed
     * for sending given buffer.
     * @param buffer Netbuffer waiting for proper server connection.
     */
    void connection_lookup
        (
        NetOBuffer * buffer
        );

    /**
     * Sends all messages with NET_BUFFER_FLUSH_ALLOWED status set.
     * Should be called periodicaly.
     */
    void send_all_messages();

    /**
     * Tries to receive data from all pending connections.
     * Should be called periodicaly. 
     *
     * @return number of messages retrieved.
     */
    int receive_all_messages();
    
    /**
     * Invalidates all connection having NOT_CONNECTED status, which had before
     * status CONNECTED. Also informs node_manager about lost / closed connections.
     * @param force_disconnect true when disconnect even if the connection is ok.
     * @param deny_node_manager_info must be true when called not in next_tick(). 
     * @note Must be called at the end of the network::next_tick(). 
     */
    void invalidate_bad_connections
        (
        bool force_disconnect = false,
        bool deny_node_manager_info = false
        );
    
    /**
     * Creates win32/unix UDP socket.<br>
     * Implemented in arch/unix/net_init_unix.cpp or arch/win32/net_init_win32,
     * respectively.
     */
    UDPSocket * create_udp_socket();

    /**
     * Creates win32/unix TCP socket.
     * Implemented in arch/unix/net_init_unix.cpp or arch/win32/net_init_win32,
     * respectively.
     */
    TCPSocket * create_tcp_socket();

    /**
     * Starts well-known-port/repeating server on @a port.
     * The server listens periodicaly (i.e. the socket is non-blocking and
     * the server calls accept in each network tick). After accepting, new
     * port is created and the client is notified to initiate connection using
     * this new port.
     *
     * @param port The well-known port number.
     * @return true if successful, otherwise false.
     *
     * @note You should not call this routine with implicit @a port parameter,
     *      as this walks around the well-known port settings from
     *      the Registry.
     */
    bool start_well_known_port_server
        (
        int port
        );

    /**
     * Tries to accept new connections both on well-known-port socket and on
     * a set of "normal" waiting sockets.<br>
     * Call to this method is not blocking.
     * @return Number of connections accepted.
     */
    int accept();

    /**
     * Performs async server connection.
     * NodeId is inserted into async_connection list and connection initiated
     * in next tick.
     * @param NodeId node id of the server.
     */
    void async_initiate_connection_to_server
        (
        const NodeId & node
        );

    /**
     * Closes connection to specified NodeId.
     * @param nodeid to close connection to.
     */
    void close_connection
        (
        const NodeId & node
        );

    /**
     * Checks whether connection to node is a server connection.
     * @param nodeid to close connection to.
     * @return true when my_node_id is SERVER_TYPE and node too.
     */
    bool is_server_connection
        (
        const NodeId & node
        );

    /**
     * Controls auto close feature with specified node.
     * @param nodeid to control connection to.
     * @param auto_close specifies whether to permit auto close or not.
     * @return old state.
     */
    bool control_connection_auto_close
        (
        const NodeId &  node,
        bool            auto_close
        );

private:

    /**
     * Initializes a socket library.
     * (Calls WSAStartup() under win32; does nothing under Linux -
     * - nevertheless, should be always called for compatibility).<br>
     * Implemented in arch/unix/net_init_unix.cpp or arch/win32/net_init_win32,
     * respectively.
     */
    void arch_init();

    /**
     * Cleans up a socket library.
     * (Calls WSACleanup() under win32; does nothing under Linux -
     * - nevertheless, should be always called for compatibility).<br>
     * Implemented in arch/unix/net_init_unix.cpp or arch/win32/net_init_win32,
     * respectively.
     */
    void arch_deinit();

    /**
     * Checks whether received message isn't net system message( e.g. PING_MESSAGE,
     * PING_REPLY_MESSAGE, NODE_DISCONNECTED_MESSAGE etc. )
     * @param conn_info is a pointer to checked connection_info.
     * @param buffer is message body, buffer.tag is message tag.
     * @param period Blowfish period of the TCP message.
     * Called from receive_via_port_hash().
     * @return true when net system message.
     */
    bool process_net_system_message
        (
        ConnectionInfo *                  conn_info,
        NetIBuffer &                      buffer,
        ConnectionInfo::BlowfishKeyPeriod period
        );

    
    /**
     * Increases statistic counter's and prepares this connection for closing.
     * @param conn_info is a pointer to connection_info.
     */
    void bad_packet_received
        (
        ConnectionInfo * conn_info
        );

    /**
     * Receive data from connections registered in port_table
     * ( => connection_info exists, but key exchange is not finished).
     * Called by method receive_all_messages().
     * @return Number of messages retrieved.
     */
    int receive_via_port_hash();

    /**
     * Receive data from connections registered in node_table
     * ( => connection_info exists and key exchange is already finished ).
     * Called by method receive_all_messages().
     * @return Number of messages retrieved.
     */
    int receive_via_node_hash();

    /**
     * Receive data from sockets listed in not_introduced_sockets( server side )
     * and sockets_waiting_for_port( client side ). Connection_info doesn't
     * exist yet.
     * Called by method receive_all_messages().
     * @return Number of messages retrieved.
     */
    int receive_via_sockets();

    /**
     * Generates blowfish key for a connection with particular server.
     * @param sid Identifier of the server - the other party in the connection.
     * @param other_public_key Defines the public key, which will encrypt this key.
     * @return The generated key.
     * @note other_public_key is there for a reason of generating a key of maximal
     * length that can other side encrypt.
     */
    std::string generate_blowfish_key
        (
        const std::string & other_public_key
        );

    /**
     * Checks whether a new TCP message is present. When true, message
     * is received.
     * @return true, when some tcp packet on specified connection
     * received.
     * @param conn_info Connection info for connection to operate on.
     * @note When true, global BitIStream recvbuf points to this message.
     */
    bool is_tcp_packet_waiting
        (
        ConnectionInfo *    conn_info
        );

    /**
     * Checks whether a new UDP message is present. When true, message
     * is received.
     * @return true, when some udp datagram on specified connection
     * received.
     * @param conn_info Connection info for connection to operate on.
     * @note When true, global BitIStream recvbuf points to this message.
     */
    bool is_udp_datagram_waiting
        (
        ConnectionInfo *    conn_info
        );
    
    /**
     * Analyzes tcp messages header - number of messages and tag.
     * Decrypted message is sent to higher level
     * via SystemMessage::dispatch_system_message_from_packet.
     * @param recvbuf BitIStream containing received message.
     * @param conn_info Connection info for connection to operate on.
     * @return true when message is not system message ( eg. PING )
     * @throw ExMessagesNumberMismatch when more than 1 message
     * is contained in packet. Possibly corrupted header.
     */
    bool process_tcp_packet
        (
        BitIStream          recvbuf,
        ConnectionInfo *    conn_info
        );
    
    /**
     * When process_tcp_message returns true, message is decrypted
     * in this method.
     * @param recvbuf BitIStream containing received message.
     * @param period Blowfish period of the TCP message.
     * @return reference to NetIBuffer containing decrypted message.
     */
    NetIBuffer & decrypt_tcp_packet
        ( 
        BitIStream                        recvbuf,
        ConnectionInfo *                  conn_info,
        ConnectionInfo::BlowfishKeyPeriod period
        );

    /**
     * Analyzes and decrypts an UDP datagram, which can contain
     * more than one messages. Decrypted messages are sent to higher level
     * via SystemMessage::dispatch_system_message_from_packet.
     * @param recvbuf BitIStream containing received message.
     * @param conn_info Connection info for connection to operate on.
     * @return number of messages received.
     */
    int decrypt_and_process_udp_datagram
        ( 
        BitIStream                          recvbuf,
        ConnectionInfo *                    conn_info
        );

    /**
     * Tries to accept connection on to the well_known_port_server. Should
     * be called periodically.<br>
     * If a connection is accepted, finds a free port and creates new TCP
     * server listening on this port.
     * @return Number of connections accepted.
     */
    int accept_on_well_known_port_server();

    /**
     * Sends TCP message filled in @c netbuffer to destination node.
     * Used by @c send_all_messages(), ...
     */
    void send_message
        (
        TCPOBuffer * netbuffer
        );

    /**
     * Sends UDP message filled in @c netbuffer to destination node.
     * Used by @c send_all_messages(), ...
     */
    void send_message
        (
        UDPOBuffer * netbuffer
        );

    /**
     * Sends specific message to destination node.
     * @param message The message type to be sent.
     * @param destination Identifier of the destination node.
     * @param param Arbitrary (message-specific) data:
     *      <table>
     *          <tr><th>Message tag                 <th>param contents
     *          <tr><td>BLOWFISH_NEW_KEY_MESSAGE
     *              <td>Pointer to the blowfish key to be sent.
     *      </table>
     */
    void send_crypted_message
        (
        RegisteredMessage   message,
        const NodeId &      destination,
        VariantPointer      param
        );

    /**
     * Writes uncrypted message header to bitostream.
     * @param bos BitOStream to write to.
     * @param message_type RegisteredMessage to send( e.g.PING_MESSAGE ).
     * @param message_length Size of the message body.
     */
    void write_packet_header
        (
        BitOStream       & bos,
        RegisteredMessage  message_type, 
        UInt32             message_length 
        );

    /**
     * Used for sending messages when blowfish encryption is not active.
     * Similar to send_crypted_message, bud second @param is socket
     * we need to send this message to. (because we don't know the other party's
     * id!)
     * @param message Type of message.
     * @param socket The socket to be used for sending the message.
     * @param param Arbitrary (message-specific) data:
     *      <table>
     *          <tr><th>Message tag                     <th>param contents
     *          <tr><td>NODE_ID_MESSAGE
     *              <td>Pointer to the connection info structure.
     *          <tr><td>PING_MESSAGE                    <td>NULL
     *          <tr><td>PING_REPLY_MESSAGE              <td>NULL
     *          <tr><td>NODE_ID_MESSAGE
     *              <td>Pointer to node identifier (NodeId).
     *          <tr><td>BLOWFISH_KEY_MESSAGE
     *              <td>Pointer to the connection info structure.
     *          <tr><td>KEY_CONFIRMATION_MESSAGE        <td>NULL
     *          <tr><td>NEW_KEY_CONFIRMATION_MESSAGE    <td>NULL
     *      </table>
     */
    void send_uncrypted_message
        (
        RegisteredMessage          message,
        Socket *                   socket,
        VariantPointer             param,
        NetworkCommon::StreamType  stream = NetworkCommon::STREAM_RELIABLE
        );

    /**
     * @param nodeid of some node present in the system.
     * @return public_key of the client associated with the given @a nodeid.
     */
    std::string get_key_from_node_id
        (
        const NodeId & nodeid
        );

    /**
     * Increase idle ticks counter and checks whether the given connection
     * has been idle for "long" time; if so, it sends the ping message.
     * @param conn_info     Connection info for connection to operate on.
     * @return false if the PING_MESSAGE has already been sent for
     *      conn_info-\>max_pings times without reply from the other party.
     *      Otherwise returns true.
     */
    bool send_ping_if_idle
        (
        ConnectionInfo *    conn_info
        );

    /**
     * Checks whether the blowfish key has expired. If so, generates the
     * new key and sends it to the client.
     * @param destination   NodeId of the destination node.
     * @param conn_info     Connection info for connection to operate on.
     * @return true if the new blowfish has been sent.
     */
    bool send_recomputed_blowfish_if_expired
        (
        const NodeId &      destination,
        ConnectionInfo *    conn_info
        );

    /**
     * Receives and processes incoming messages from a connection in
     * NODE_ID_EXCHANGE_SERVER state.
     * @param conn_info     Connection info for connection to operate on.
     * @return 1 if a message has been accepted, otherwise 0.
     */
    int receive_in_node_id_exchange_server_state
        (
        ConnectionInfo *     conn_info
        );

    /**
     * Receives and processes incoming messages from a connection in
     * BLOWFISH_EXCHANGE_CLIENT state.
     * @param conn_info     Connection info for connection to operate on.
     * @return 1 if a message has been accepted, otherwise 0.
     */
    int receive_in_blowfish_exchange_client_state
        (
        ConnectionInfo *    conn_info
        );

    /**
     * Receives and processes incoming messages from a connection in
     * BLOWFISH_EXCHANGE_SERVER state.
     * @param conn_info     Connection info for connection to operate on.
     * @return 1 if a message has been accepted, otherwise 0.
     */
    int receive_in_blowfish_exchange_server_state
        (
        ConnectionInfo *        conn_info
        );

    /**
     * Receives and processes incoming messages from a connection in
     * UDP_PORT_RECEIVE_CLIENT state.
     * @param conn_info     Connection info for connection to operate on.
     * @return 1 if a message has been accepted, otherwise 0.
     */
    int receive_in_udp_port_receive_client_state
        (
        ConnectionInfo *        conn_info
        );

    /**
     * In this connection state server waits for UDP setup datagram
     * from client to create UDP-addressed connection.
     * @param conn_info     Connection info for connection to operate on.
     * @param iter          Iterator to the item in the connection list.
     * @return 1 if a message has been accepted, otherwise 0.
     */
    int receive_in_udp_connection_setup_server_state
        (
        ConnectionInfo *                            conn_info,
        std::list< ConnectionInfo * >::iterator &   iter
        );

    /**
     * Receives and processes incoming messages from a connection in
     * CONNECTED state.
     * @param conn_info     Connection info for connection to operate on.
     * @return Number of messages received.
     */
    int receive_in_connected_state
        (
        ConnectionInfo *        conn_info
        );

    /**
     * Processes the BLOWFISH_NEW_KEY_MESSAGE.
     * @param node_id_other Node id of the other party.
     * @param conn_info The connection_info to operate on.
     * @param period Blowfish period of the TCP message.
     * @param ibuffer Input buffer containing the BLOWFISH_NEW_KEY_MESSAGE.
     * 
     */
    void process_blowfish_new_key_message
        (
        const NodeId &                    node_id_other,
        ConnectionInfo *                  conn_info,
        NetIBuffer &                      ibuffer,
        ConnectionInfo::BlowfishKeyPeriod period
        );

    /**
     * Processes the NEW_KEY_CONFIRMATION_MESSAGE.
     * @param conn_info The connection_info to operate on.
     * @param ibuffer Input buffer containing the message.
     */
    void process_new_key_confirmation_message
        (
        ConnectionInfo *        conn_info,
        NetIBuffer &            buffer
        );

    /**
     * Handles error messages during creating connection.
     * @return true if handled error messsage otherwise false.
     */
    bool process_error_message
        (
        UInt32 tag
        );

    /**
     * Logs into staticstics, that TCP message received.
     * Sets timestamps of connection info to received time.
     * ( used in auto connection killing when no data received
     * for a specified time ).
     */
    void tcp_packet_received
        (
        ConnectionInfo * conn_info,
        UInt32           len
        );
    
    /**
     * Logs into staticstics, that UDP message received.
     * Sets timestamps of connection info to received time.
     * ( used in auto connection killing when no data received
     * for a specified time ).
     */
    void udp_datagram_received
        (
        ConnectionInfo * conn_info,
        UInt32           len
        );
    
    /**
     * Links @c buffer to ConnectionInfo object addressed by @c id or
     * if no such connection exists initiates the connection.
     * Throws ExNoSuchConnection if connection can not be initiated.
     *
     * A common part of lock_reliable, lock_unreliable methods.
     */
    void lock_buffer_helper
        (
        const NodeId &              id,
        NetOBuffer *                buffer,
        NetworkCommon::StreamType   type
        );
  
    /**
     * All prepared buffers for sending must be told to use new
     * connection. Used when 2 servers trying simultaneusly to connect.
     */
    void route_connection
        (
        const NodeId &   node,
        ConnectionInfo * conn_info
        );

    /**
     * Performs first phase of establishing connection to a server, i.e.
     * connects to server's well-known port and prepares for receiving new
     * port information.
     * @param server_address Address of the server.
     * @return true if successful, otherwise false.
     */
    bool initiate_connection_to_server
        (
        ServerAddress & server_address
        );
        
    /**
     * Performs connection initiation at the tick begin.
     * @Note executed in next_tick.
     */
    void initiate_connections();
    
    /**
     * Performs first phase of establishing connection to a server, i.e.
     * connects to server's well-known port and prepares for receiving new
     * port information.
     * @param nodeid of the server of the server.
     * @return true if successful, otherwise false.
     */
    bool initiate_connection_to_server
        (
        const NodeId & node
        );

    /**
     * Goes through init_udp_list and sends UDP init datagram to server
     * when udp not connected.
     * @note caled from Network::next_tick().
     */
    void init_udp();
        
    /**
     * When ANONYMOUS node is connecting to service, local unique node id 
     * is required to distinguish nodes.
     * @return local unique anonymous node id;
     */    
    NodeId find_unique_anonymous_node_id();

    /**
     * When auto-disconnection of two node occurs, we must ask
     * whether is safe to disconnect.
     * @param node NodeId to check.
     * @return true when no messages prepared to send;
     */    
    bool is_safe_to_disconnect
        (
        const NodeId & node
        );

public:

    NetStatistics               net_stat;
        /**< Network statistics. */

    NetworkSettings             settings;
        /**< Network settings. */

private:
    
    RSACipher                       rsa;
        /**< RSA cipher. Holds local node's private and public key. */

    std::list< NodeId >             async_connectionlist;
        /**< List of nodes connections should be initiated to. */

    std::list< TCPOBuffer >         tcp_permlist;
        /**< List of permanent TCP buffers. */

    std::list< TCPOBuffer * >       tcp_sendlist;
        /**< List of pointers to TCP buffers that will be sent out. */

    std::list< UDPOBuffer >         udp_permlist;
        /**< List of permanent UDP buffers. */

    std::list< UDPOBuffer * >       udp_sendlist;
        /**< List of pointers to UDP buffers that will be sent out. */

    BitIStream                      recvbuf;
        /**< Buffer for incoming packets. */

    UInt32                          recvbuf_size;
        /**< Received packet size. */

    bool                            local_node_id_assigned;
        /**< True if node id is was assigned to local system. */

    NodeId                          my_node_id;
        /**< Node id associated with local node. Valid only
             if @c local_node_id_assigned is true. */

    std::list< Socket * >           sockets_waiting_for_port;
        /**< List of sockets waiting for assigning port by the
             well-known-port server. */

    Socket *                        well_known_port_socket;
        /**< Socket associated with the well-known port. */

    NodeHash                        node_table;
        /**< Hash table of nodes and connection_infos bound with them. */

    std::list< ConnectionInfo * >   conn_list;
        /**< List of not authorized connection_infos. */

    std::list< ConnectionInfo * >   init_udp_list;
        /**< List of connections where UDP not connected. */

    UInt64                          bytes_delta;
        /**< How many bytes were received in last delta. */

    Float32                         time_delta;
        /**< Accumulate time difference. */

    SystemTime                      last_tick_time;
        /**< Last tick time. */

    UInt64                          last_tick_bytes;
        /**< Sum of all bytes processed till end of the last tick. */

    };  // class Network

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_NET_H
