/** @file net_buffers.cpp
 * Network buffer interface.
 */

/*
 * begin        : Sat Feb 15 2002
 * copyright    : (C) 2003, 2004 Ondrej Pecta, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/net.h"
#include "utility/debug.h"
#include "database/node_database.h"
#include "global/global.h"
#include <typeinfo>

namespace Massiv {
namespace Core {


#ifdef OPTION_DEBUG

/**
 * Define this to enable network buffers logging.
 */
//#define LOG_BUFFERS

#endif


NetOBuffer * Network::lock
    (
    const NodeId &            id,
    NetworkCommon::StreamType type,
    UInt32                    tag,
    bool                      tcp_send_immediately
    )
    {
    if( type == NetworkCommon::STREAM_RELIABLE )
        return lock_reliable( id, tag, tcp_send_immediately );
    else
        return lock_unreliable( id, tag );
    }


bool Network::can_lock_buffer_to_node
    (
    const NodeId &            node,
    NetworkCommon::StreamType stream_type
    )
    {
    #ifdef LOG_BUFFERS

    std::ostringstream oss;
    oss << "Checking whether it is possible to lock buffer to " << node << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
        
    #endif

    // Lookup the node in node database.
        {
        NodeDatabase &                  node_database = Global::node_database();
        const NodeDatabase::NodeData *  dest_node_data;

        // Look-up the node in node database.
        node_database.get_node_data( node, &dest_node_data );
        if( !dest_node_data )
            {
            return false;
            }
        }
    
    if( ( node.is_type( NodeId::TYPE_SERVER ) ) ||
        ( node.is_type( NodeId::TYPE_SERVICE ) ) ||
        ( is_node_connected( node ) ) )
        {
        return true;
        }
    
    return false;
    }
    

void Network::lock_buffer_helper
    (
    const NodeId &              id,
    NetOBuffer *                buffer,
    NetworkCommon::StreamType   type
    )
    {
    NodeHash::iterator iter;

    // Prepare.

    buffer->connection_info = NULL;
    
    if( !can_lock_buffer_to_node( id, type ) )
        {
        throw ExNoSuchConnection
            ( "Connection not established and can not initiate connection." );
        }

    buffer->owner = id;

    // Lookup and bind connection info.

    iter = node_table.find( id );

    if( iter != node_table.end() )
        {
        // There is some evidence of an existing connection.
        ConnectionInfo * conn_info = iter->second;
        hope( conn_info );
        
        if( conn_info->get_connection_status() == ConnectionInfo::CONNECTED )
            {
            buffer->connection_info = conn_info;
            }
        else
        if( conn_info->get_connection_status() == ConnectionInfo::NOT_CONNECTED )
            {
            #ifdef LOG_BUFFERS
            
            Global::log_info
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_LOW,
                "...connection info exists but need to reconnect."
                );

            #endif
            
            async_initiate_connection_to_server( id );
            }
        else
            {
            // The connection creation is pending, do not do anything.
            
            #ifdef LOG_BUFFERS
            
            Global::log_info
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_LOW,
                "...connection creation is in progress."
                );

            #endif
            }
        }
    else
        {
        // There is no evidence of existing connection.

        #ifdef LOG_BUFFERS
        
        Global::log_info
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_LOW,
            "...connection need be initiated first."
            );

        #endif

        async_initiate_connection_to_server( id );
        }
    }


TCPOBuffer * Network::lock_reliable
    (
    const NodeId &  id,
    UInt32          tag,
    bool            tcp_send_immediately
    )
    {
    hope( id != my_node_id );
    
    #ifdef LOG_BUFFERS

    {
    std::ostringstream ss;
    ss  << "Locking TCP buffer to " << id << " (tag = " << tag
        << ", tcp_send_immediately = " << tcp_send_immediately << ").";
    Global::log_info
        (
        Status::FACILITY_NETWORK,
        Status::PRIORITY_LOW,
        ss.str()
        );
    }

    #endif

    // Create new buffer or reuse an already created.
    
    TCPOBuffer *    buffer = NULL;
    bool            reused = false;

    std::list< TCPOBuffer >::iterator itertcp;
    itertcp = tcp_permlist.begin();
    while( itertcp != tcp_permlist.end() )
        {
        if( itertcp->status == NetOBuffer::NET_BUFFER_UNLOCKED )
            {
            buffer = &( *itertcp );
            hope( buffer );
            reused = true;
            break;
            }
        else
            {
            itertcp++;
            }
        }

    if( buffer == NULL )
        {
        buffer = new TCPOBuffer();
        buffer->permanent = false;
        }

    // Setup buffer.
    
    hope( buffer );

    try
        {
        lock_buffer_helper( id, buffer, NetworkCommon::STREAM_RELIABLE );
        
        buffer->send_immediately = tcp_send_immediately;
        buffer->init( tag );

        tcp_sendlist.push_back( buffer );
        buffer->self_iterator = tcp_sendlist.end();
        buffer->self_iterator--;
        }
    catch( Exception & )
        {
        if( !reused )
            {
            delete buffer;
            }

        throw;
        }

    return buffer;
    }  // Network::lock_reliable()


UDPOBuffer * Network::lock_unreliable
    (
    const NodeId &  id,
    UInt32          tag
    )
    {
    hope( id != my_node_id );
    
    #ifdef LOG_BUFFERS

    {
    std::ostringstream ss;
    ss << "Locking UDP buffer to " << id << " (tag = " << tag << ").";
    Global::log_info
        (
        Status::FACILITY_NETWORK,
        Status::PRIORITY_LOW,
        ss.str()
        );
    }

    #endif

    // Check whether there exists a suitable buffer, which is ready to be 
    // sent. If so, reuse it (our bitstream will be appended to the message).

    UDPOBuffer * buffer = NULL;

    std::list< UDPOBuffer * >::iterator iter_sendudp;
    iter_sendudp = udp_sendlist.begin();

    while( iter_sendudp != udp_sendlist.end() )
        {
        buffer = *iter_sendudp;
        hope( buffer );
        
        int max_udp_len = 
            NetworkCommon::MAX_WAITING_BYTES - 
            NetworkCommon::NET_BUFFER_HEAD_LEN_UDP - 
            NetworkCommon::NET_BUFFER_FOOT_LEN_UDP;
        
        if( buffer->owner == id &&
            buffer->status != NetOBuffer::NET_BUFFER_LOCKED && 
            buffer->get_buffer().size() < ( UInt32 ) max_udp_len )
            {
            #ifdef LOG_BUFFERS

            Global::log_info
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_LOW,
                "...appending to an existing buffer."
                );

            #endif

            buffer->msg_init( tag );
            return buffer;
            }
        else
            {
            iter_sendudp++;
            }
        }

    // Cannot reuse, try to find free permanent buffer or allocate a
    // new temporary buffer.

    std::list< UDPOBuffer >::iterator   iterudp;
    bool                                reused;

    buffer = NULL;
    reused = false;
    
    iterudp = udp_permlist.begin();

    while( iterudp != udp_permlist.end() )
        {
        if( iterudp->owner.is_null() )
            {
            buffer = &( *iterudp );
            hope( buffer );
            reused = true;
            break;
            }
        else
            {
            iterudp++;
            }
        }

    if( buffer == NULL )
        {
        buffer = new UDPOBuffer();
        buffer->permanent = false;
        }

    // Setup buffer.
    
    hope( buffer );

    try
        {
        lock_buffer_helper( id, buffer, NetworkCommon::STREAM_UNRELIABLE );
        buffer->init( tag );
        
        udp_sendlist.push_back( buffer );
        buffer->self_iterator = udp_sendlist.end();
        buffer->self_iterator--;
        }
    catch( Exception & )
        {
        if( !reused )
            {
            delete buffer;
            }

        throw;
        }
    
    return buffer;
    }  // Network::lock_unreliable()


void Network::unlock
    (
    NetOBuffer *    netbuffer
    )
    {
    hope( netbuffer );
    hope( netbuffer->status == NetOBuffer::NET_BUFFER_LOCKED );

    const std::type_info & info = typeid( *netbuffer );

    if( info == typeid( TCPOBuffer ) )
        {
        unlock_reliable( checked_cast<TCPOBuffer *>( netbuffer ) );
        }
    else
    if( info == typeid( UDPOBuffer ) )
        {
        unlock_unreliable( checked_cast<UDPOBuffer *>( netbuffer ), true );
        // TODO: decide whether to call unlock_unreliable()
        // with flush_allowed set to true or false.
        }
    else
        {
        panic( "Invalid type of netbuffer." );
        }
    }  // Network::unlock()


void Network::unlock_reliable
    (
    TCPOBuffer *  netbuffer
    )
    {
    hope( netbuffer );
    hope( netbuffer->status == NetOBuffer::NET_BUFFER_LOCKED );
    
    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Unlocking TCP buffer to " << netbuffer->owner << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif
    
    netbuffer->status = NetOBuffer::NET_BUFFER_FLUSH_ALLOWED;

    // Send the buffer if send_immediately is TRUE.

    if( netbuffer->send_immediately )
        {
        tcp_sendlist.erase( netbuffer->self_iterator );
        send_message( netbuffer );
        }
    }  // Network::unlock_reliable()


void Network::unlock_unreliable
    (
    UDPOBuffer *  netbuffer,
    bool          flush_allowed = true
    )
    {
    hope( netbuffer );
    hope( netbuffer->status == NetOBuffer::NET_BUFFER_LOCKED ); 

    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Unlocking UDP buffer to " << netbuffer->owner << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif
        
    if( flush_allowed ) 
        netbuffer->status = NetOBuffer::NET_BUFFER_FLUSH_ALLOWED;
    else
        netbuffer->status = NetOBuffer::NET_BUFFER_UNLOCKED;

    netbuffer->msg_flush();
    }


void Network::flush
    (
    UDPOBuffer *  netbuffer
    )
    {
    hope( netbuffer );

    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Flushing UDP buffer to " << netbuffer->owner << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif


    hope
        (
        netbuffer->status == NetOBuffer::NET_BUFFER_UNLOCKED ||
        netbuffer->status == NetOBuffer::NET_BUFFER_FLUSH_ALLOWED
        );

    netbuffer->status = NetOBuffer::NET_BUFFER_FLUSH_ALLOWED;
    }


void Network::drop
    (
    TCPOBuffer * netbuffer
    )
    {
    hope( netbuffer );

    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Dropping TCP buffer to " << netbuffer->owner << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif

    tcp_sendlist.erase( netbuffer->self_iterator );

    reset_buffer( netbuffer );
    }
    

void Network::drop
    (
    UDPOBuffer * netbuffer
    )
    {
    hope( netbuffer );

    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Dropping UDP buffer to " << netbuffer->owner << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif

    udp_sendlist.erase( netbuffer->self_iterator );

    reset_buffer( netbuffer );
    }


void Network::drop_all()
    {
    std::list< TCPOBuffer * >::iterator tcpbuffer = tcp_sendlist.begin();
    std::list< UDPOBuffer * >::iterator udpbuffer = udp_sendlist.begin();

    std::list< TCPOBuffer * >::iterator tcp;
    std::list< UDPOBuffer * >::iterator udp;
    
    while( tcpbuffer != tcp_sendlist.end() )
        {
        tcp = tcpbuffer++;
        hope( *tcp );
        
        drop( checked_cast<TCPOBuffer *>( *tcp ) );
        }
    while( udpbuffer != udp_sendlist.end() )
        {
        udp = udpbuffer++;
        hope( *udp );
        
        drop( checked_cast<UDPOBuffer *>( *udp ) );
        }
    }


void Network::drop_all_to_server
    (
    const NodeId & node,
    bool  really_drop
    )
    {
    #ifdef LOG_BUFFERS

    {
    std::ostringstream oss;
    oss << "Dropping all messages to " << node << ".";
    
    Global::log_debug
        ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
    }
        
    #endif

    std::list< TCPOBuffer * >::iterator tcpbuffer = tcp_sendlist.begin();
    std::list< UDPOBuffer * >::iterator udpbuffer = udp_sendlist.begin();

    std::list< TCPOBuffer * >::iterator tcp;
    std::list< UDPOBuffer * >::iterator udp;

    while( tcpbuffer != tcp_sendlist.end() )
        {
        tcp = tcpbuffer++;
        TCPOBuffer * buf = *tcp;
        hope( buf );
        
        if( buf->owner == node )
            {
            if( !really_drop )
                buf->connection_info = NULL; 
            else
                drop( checked_cast<TCPOBuffer *>( buf ) );
            }
        }
    while( udpbuffer != udp_sendlist.end() )
        {
        udp = udpbuffer++;
        UDPOBuffer * buf = *udp;
        hope( buf );

        if( buf->owner == node )
            {
            if( !really_drop )
                buf->connection_info = NULL; 
            else
                drop( checked_cast<UDPOBuffer *>( buf ) );
            }
        } 
    }


void Network::route_connection
    (
    const NodeId &   node,
    ConnectionInfo * conn_info
    )
    {
    hope( conn_info );
    std::list< TCPOBuffer * >::iterator tcpbuffer = tcp_sendlist.begin();
    std::list< UDPOBuffer * >::iterator udpbuffer = udp_sendlist.begin();

    std::list< TCPOBuffer * >::iterator tcp;
    std::list< UDPOBuffer * >::iterator udp;

    while( tcpbuffer != tcp_sendlist.end() )
        {
        tcp = tcpbuffer++;
        hope( *tcp );
        
        if( ( *tcp )->owner == node )
            ( *tcp )->connection_info = conn_info;
        }
    while( udpbuffer != udp_sendlist.end() )
        {
        udp = udpbuffer++;
        hope( *udp );
        
        if( ( *udp )->owner == node )
            ( *udp )->connection_info = conn_info;
        } 
    }


void Network::reset_buffer
    (
    NetOBuffer * netbuffer 
    )
    {
    hope( netbuffer );

    netbuffer->unbind();

    if( netbuffer->permanent )
        {
        netbuffer->status = NetOBuffer::NET_BUFFER_UNLOCKED;
        netbuffer->owner = NodeId();
        netbuffer->connection_info = NULL;
        const std::type_info & info = typeid( *netbuffer );
        if( info == typeid( UDPOBuffer ) )
            {
            UDPOBuffer * udp_buffer = checked_cast<UDPOBuffer *>( netbuffer );
            udp_buffer->count = 0;
            udp_buffer->msg_start = 0;
            }
        }
    else
        {
        delete netbuffer;
        }
    }

} // namespace Massiv::Core
} // namespace Massiv
