/** @file net_messages.cpp
 * Network message interface.
 */

/*
 * begin        : Fri Mar 15 2002
 * copyright    : (C) 2003 Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/debug.h"
#include "net/net.h"
#include "net/connection_info_impl.h"
#include "sysevent/system_msg.h"
#include "utility/quote_string.h" //~!
#include <iostream>
#include <sstream>

namespace Massiv {
namespace Core {


#ifdef OPTION_DEBUG

/**
 * Define this to enable logging of messages.
 */
//#define LOG_MESSAGES

#endif

static const char * msg_invalid_checksum =
    "Invalid checksum in the message.";

using namespace NetworkCommon;

int Network::receive_all_messages()
    {
    return
        receive_via_port_hash() +
        receive_via_node_hash() +
        receive_via_sockets();
    } // Network::receive_all_messages()


void Network::send_message
    (
    TCPOBuffer * netbuffer
    )
    {
    netbuffer->flush();

    // Log operation.
    
    #ifdef LOG_MESSAGES

    std::ostringstream s;

    s   << "Sending TCP message to node " << netbuffer->owner << ", size: "
        << netbuffer->crypted_buffer.size() << " bytes.";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        s.str()
        );

    #endif

    // Send message.

    try
        {
        ConnectionInfo *    connection_info = netbuffer->connection_info;
        Socket *            tcp_socket = connection_info->get_tcpsocket();

        hope( connection_info );
        hope( tcp_socket );
        
        tcp_socket->send( netbuffer->crypted_buffer );
        connection_info->touch_last_message_time();

        net_stat.inc_tcp_bytes_sent( netbuffer->crypted_buffer.size() );
        net_stat.inc_messages_sent( 1 );
        }
    catch( ... )
        {
        ConnectionInfo *    info = netbuffer->connection_info;
        hope( info );

        info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
        info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
        }

    reset_buffer( netbuffer );
    } // Network::send_message


void Network::send_message
    (
    UDPOBuffer * netbuffer
    )
    {
    netbuffer->flush();

    // Log operation.

    #ifdef LOG_MESSAGES

    std::ostringstream s;

    s   << "Sending UDP message to node " << netbuffer->owner << ", size: " 
        << netbuffer->crypted_buffer.size() << " bytes.";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        s.str()
        );

    #endif

    // Send message.

    try
        {
        ConnectionInfo *    connection_info = netbuffer->connection_info;
        Socket *            udp_socket = connection_info->get_udpsocket();

        hope( connection_info );
        hope( udp_socket );
        
        udp_socket->send( netbuffer->crypted_buffer );
        connection_info->touch_last_message_time();

        net_stat.inc_udp_bytes_sent( netbuffer->crypted_buffer.size() );
        net_stat.inc_messages_sent( 1 );
        }
    catch( ... )
        {
        ConnectionInfo *    info = netbuffer->connection_info;
        info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
        info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
        }

    reset_buffer( netbuffer );
    } // Network::send_message


void Network::connection_lookup
    (
    NetOBuffer * buffer
    )
    {
    hope( buffer );

    NodeHash::iterator iter = node_table.find( buffer->owner );

    if( is_node_connected( buffer->owner ) && iter != node_table.end() )
        {
        ConnectionInfo * conn_info = iter->second;
        
        hope( conn_info );
        
        if( conn_info->get_connection_status() ==
            ConnectionInfo::CONNECTED )
            {
            buffer->connection_info = conn_info;    
            }
        }
    } // Network::connection_lookup
    

void Network::send_all_messages()
    {
    std::list< TCPOBuffer * >::iterator tcp_iter = tcp_sendlist.begin();
    std::list< UDPOBuffer * >::iterator udp_iter = udp_sendlist.begin();

    // Send TCP messages.

    while( tcp_iter != tcp_sendlist.end() )
        {
        TCPOBuffer * tcpbuf = checked_cast<TCPOBuffer *>( *tcp_iter );
        hope( tcpbuf );

        if( tcpbuf->status == NetOBuffer::NET_BUFFER_FLUSH_ALLOWED )
            {
            ConnectionInfo * info = tcpbuf->connection_info;

            if( info == NULL )
                {
                // The message is properly locked, but connection doesn't exist.
                connection_lookup( tcpbuf );
                }
            else
            if( info->get_connection_status() == ConnectionInfo::CONNECTED )
                {
                if( info->is_tcp_confirmed() )
                    {
                    send_message( tcpbuf );

                    tcp_sendlist.erase( tcp_iter++ );
                    continue;
                    }
                }
            }

        ++tcp_iter;
        }

    // Send UDP messages only when connected.

    while( udp_iter != udp_sendlist.end() )
        {
        UDPOBuffer * udpbuf = checked_cast<UDPOBuffer *>( *udp_iter );
        hope( udpbuf );

        if( udpbuf->status == NetOBuffer::NET_BUFFER_FLUSH_ALLOWED )
            {
            ConnectionInfo * info = udpbuf->connection_info;

            if( info == NULL )
                {
                connection_lookup( udpbuf );
                }
            else
            if( info->get_connection_status() == ConnectionInfo::CONNECTED )
                {
                if( info->is_udp_connected() )
                    {
                    send_message( udpbuf );

                    udp_sendlist.erase( udp_iter++ );
                    continue;
                    }
                }
            }

        ++udp_iter;
        }

    #ifdef OPTION_DEBUG
    
    if( tcp_sendlist.size() == 0 )
        {
        std::list< TCPOBuffer >::iterator iter_tcp;

        iter_tcp = tcp_permlist.begin();
        while( iter_tcp != tcp_permlist.end() )
            {
            hope( iter_tcp->connection_info == NULL );
            hope( iter_tcp->owner == NodeId() );
            iter_tcp++;
            }
        }

    if( udp_sendlist.size() == 0 )
        {
        std::list< UDPOBuffer >::iterator iter_udp;

        iter_udp = udp_permlist.begin();
        while( iter_udp != udp_permlist.end() )
            {
            hope( iter_udp->connection_info == NULL );
            hope( iter_udp->owner == NodeId() );
            iter_udp++;
            }
        }

    #endif

    #ifdef OPTION_DEBUG
    
        {
        std::list< TCPOBuffer * >::iterator iter_tcp;

        iter_tcp = tcp_sendlist.begin();
        while( iter_tcp != tcp_sendlist.end() )
            {
            hope( ( *iter_tcp )->owner != NodeId::create_from_type_and_relative_id( "Server", 0 ) );
            iter_tcp++;
            }
        }

    
        {
        std::list< UDPOBuffer * >::iterator iter_udp;

        iter_udp = udp_sendlist.begin();
        while( iter_udp != udp_sendlist.end() )
            {
            hope( ( *iter_udp )->owner != NodeId::create_from_type_and_relative_id( "Server", 0 ) );
            iter_udp++;
            }
        }

    #endif
    }  // Network::send_all_messages()


void Network::send_crypted_message
    (
    RegisteredMessage   message,
    const NodeId &      destination,
    VariantPointer      param
    )
    {
    if( destination.is_null() )
        {
        return;
        }
    switch( message )
        {
        case BLOWFISH_NEW_KEY_MESSAGE:
            {
            TCPOBuffer * bufcrypto =
                checked_cast<TCPOBuffer *>
                    (
                    lock_reliable
                        (
                        destination,
                        BLOWFISH_NEW_KEY_MESSAGE,
                        true
                        )
                    );
                    
            hope( bufcrypto );
                    
            std::string key( *variant_cast<std::string *>( param ) );
            bufcrypto->write_full_string( key );

            unlock_reliable( bufcrypto );
            break;
            }

        default:
            panic( "Unexpected message parameter to send_crypted_message." );
        }
    }  // Network::send_crypted_message()


void Network::write_packet_header
    (
    BitOStream       & bos,
    RegisteredMessage  message_type, 
    UInt32             message_length 
    )
    {
    hope( bos.byte_size() == 0 );

    bos.write_int( 1, 16 );
    bos.write_int( message_type, 32 );
    bos.write_int( message_length, 32 );

    hope( bos.byte_size() == NET_BUFFER_HEAD_LEN_TCP );
    }


void Network::send_uncrypted_message
    (
    RegisteredMessage   message,
    Socket *            socket,
    VariantPointer      param,
    StreamType          stream
    )
    {
    BitOStream      bos;
    int             bytes_sent = 0;

    // Each message has the following format:
    //  <number_of_messages [ = always 1 ], UInt16>
    //  <message_type, UInt32>
    //  <message_length, UInt32>
    //  <message_data, message_length * UInt8>

    switch( message )
        {
        case UDP_INIT_MESSAGE:
        case NODE_ID_MESSAGE:
            {
            write_packet_header( bos, message, 4 );
            bos.write_int( NodeId::convert_to_uint32( get_local_node_id() ), 32 );
            break;
            }

        case KEY_CONFIRMATION_MESSAGE:
            {
            hope( !param.is_null() );
            ConnectionInfo *    conn_info =
                variant_cast<ConnectionInfo *>( param );

            write_packet_header( bos, message, 4 + conn_info->get_blowfish_key_even().size() );
            bos.write_full_string( conn_info->get_blowfish_key_even() );
            break;
            }

        case BLOWFISH_KEY_MESSAGE:
            {
            // Generate massiv blowfish_key for this nodeid.

            std::string         key;
            std::string         crypted_string;
            std::string         signed_string;
            ConnectionInfo *    connection_info =
                variant_cast<ConnectionInfo *>( param );

            key = generate_blowfish_key
                (
                connection_info->get_other_side_public_key()
                );

            #ifdef LOG_MESSAGES

            std::ostringstream  s;

            s << "Sending blowfish key " << quote_const_string( key ) << " to node " <<
                 connection_info->get_node_id() << " with len " << key.size() << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );
            
            #endif

            connection_info->set_blowfish_key_odd( key );
            connection_info->set_blowfish_key_even( key );
                // Note: Don't use set_blowfish_key here - it would set
                //  only the even-phase key because this connection is
                //  in a BLOWFISH_PERIOD_ODD phase!!!

            rsa.encrypt
                (
                key,
                crypted_string,
                connection_info->get_other_side_public_key()
                );

            // Sign the crypted blowfish key.

            rsa.sign( crypted_string, signed_string );

            write_packet_header( bos, message, crypted_string.size() + signed_string.size() + 8 );

            bos.write_full_string( crypted_string );
            bos.write_full_string( signed_string );

            break;
            }

        case NODE_DISCONNECTED_MESSAGE:
            {
            write_packet_header( bos, message, 2 );

            ConnectionInfo * conn_info = variant_cast<ConnectionInfo *>( param );
            bos.write_int( conn_info->get_disconnection_reason(), 16 );
            break;
            }

        case PING_REPLY_MESSAGE:
        case PING_MESSAGE:
        case UDP_HANDSHAKE_MESSAGE:
        case ALREADY_CONNECTED_MESSAGE:
        case NOT_AUTHORIZED_MESSAGE:
        case NOT_READY_TO_LOGIN_MESSAGE:
        case TCP_CONNECT_MESSAGE:
        case TCP_CONFIRMATION_MESSAGE:
        case DISCONNECT_DENY_MESSAGE:
        case DISCONNECT_PERMIT_MESSAGE:
        case DISCONNECT_REQUEST_MESSAGE:
        case DISCONNECT_FORCE_MESSAGE:
        case DISCONNECT_READY_MESSAGE:
            {
            hope( param.is_null() );
            write_packet_header( bos, message, 0 );
            break;
            }

        case NEW_KEY_CONFIRMATION_MESSAGE:
            {
            // Confirmation to BLOWFISH_NEW_KEY_MESSAGE

            hope( param.is_null() );
            write_packet_header( bos, message, 3 );
            bos.write_string( "OK" );
            break;
            }

        case UDP_PORT_MESSAGE:
            {
            hope( !param.is_null() );
            write_packet_header( bos, message, 6 );
            UInt16 free_port = *( variant_cast<UInt16 *>( param ) );
            bos.write_int( NodeId::convert_to_uint32( get_local_node_id() ), 32 );
            bos.write_int( free_port , 16 );
            break;
            }
        
        default:
            panic( "Unexpected message parameter to send_uncrypted_message." );
        }
    
    hope( socket );
        
    try
        {
        socket->send( bos );
        }
    catch( ExSendError & e )
        {
        std::ostringstream ss;

        ss  << "Can not send uncrypted message (type " << message << ").\n"
            <<  e.what();

        throw ExSendError( ss.str() );
        }
        
    if( stream == NetworkCommon::STREAM_RELIABLE )
        {
        net_stat.inc_tcp_bytes_sent( bytes_sent );
        }
    else
        {
        net_stat.inc_messages_sent( 1 );
        }

    net_stat.inc_messages_sent( 1 );
    bytes_sent = bos.byte_size();
    }  // Network::send_uncrypted_message()


bool Network::process_net_system_message
    (
    ConnectionInfo *                  conn_info,
    NetIBuffer &                      buffer,
    ConnectionInfo::BlowfishKeyPeriod period
    )
    {
    hope( conn_info );
    
    NodeId sid = conn_info->get_node_id();
    switch( buffer.tag )
        {
        case TCP_CONFIRMATION_MESSAGE:
            {
            conn_info->set_tcp_confirmed( true );

            #ifdef LOG_MESSAGES
            
            std::ostringstream  s;
            s << "TCP connection confirmed with " << sid << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );

            #endif

            return true;
            }

        case UDP_HANDSHAKE_MESSAGE:
            {
            conn_info->set_udp_connected( true );
            
            std::list< ConnectionInfo * >::iterator it;
            it = std::find( init_udp_list.begin(), init_udp_list.end(), conn_info );
            if( it != init_udp_list.end() )
                {
                hope( *it );
                init_udp_list.erase( it );
                }
                
            #ifdef LOG_MESSAGES
            
            std::ostringstream  s;
            s << "UDP handshake successfull with " << sid << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );

            #endif

            return true;
            }
        
        case PING_MESSAGE:
            {
            send_uncrypted_message
                (
                PING_REPLY_MESSAGE,
                conn_info->get_tcpsocket(),
                VariantPointer()
                );

            #ifdef LOG_MESSAGES

            std::ostringstream  s;
            s << "Ping reply message sent to node " << sid << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );

            #endif

            return true;
            }

        case PING_REPLY_MESSAGE:
            {
            return true;
            }

        case BLOWFISH_NEW_KEY_MESSAGE:
            {
            process_blowfish_new_key_message( sid, conn_info, buffer, period );
            return true;
            }

        case NEW_KEY_CONFIRMATION_MESSAGE:
            {
            process_new_key_confirmation_message( conn_info, buffer );
            return true;
            }

        case NODE_DISCONNECTED_MESSAGE:
            {
            buffer.skip( 32 );

            ConnectionInfo::DisconnectionReason reason =
                static_cast<ConnectionInfo::DisconnectionReason>
                ( buffer.read_int( 16 ) );

            if( reason == ConnectionInfo::REQUEST_REASON )
                {
                conn_info->set_disconnection_reason
                    (
                    ConnectionInfo::REQUEST_REASON
                    );
                }
            else
                {
                conn_info->set_disconnection_reason
                    (
                    ConnectionInfo::ERROR_REASON
                    );
                }

            conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );

            return true;
            }

        case DISCONNECT_FORCE_MESSAGE:
            {
            std::ostringstream  s;

            s << "Received order to close connection to node " << sid << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );
            
            send_uncrypted_message
                (
                DISCONNECT_READY_MESSAGE,
                conn_info->get_tcpsocket(),
                VariantPointer()
                );

            return true;
            }

        case DISCONNECT_READY_MESSAGE:
            {
            std::ostringstream  s;

            s << "Received acknowledgement to close connection to node " << sid << ".";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );

            close_connection( sid );
            return true;
            }

        case DISCONNECT_REQUEST_MESSAGE:
            {
            std::ostringstream  s;

            s << "Received request to close connection to node " << sid << ".";

            if( is_safe_to_disconnect( sid ) &&
                ( conn_info->get_permit_auto_close() ||
                my_node_id.get_type() == NodeId::TYPE_CLIENT ||
                my_node_id.get_type() == NodeId::TYPE_ANONYMOUS ) )
                {
                send_uncrypted_message
                    (
                    DISCONNECT_PERMIT_MESSAGE,
                    conn_info->get_tcpsocket(),
                    VariantPointer()
                    );
                conn_info->set_connection_status( ConnectionInfo::TRYING_TO_DISCONNECT );
                }
            else
                {
                send_uncrypted_message
                    (
                    DISCONNECT_DENY_MESSAGE,
                    conn_info->get_tcpsocket(),
                    VariantPointer()
                    );
                }

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );
            return true;
            }


        case DISCONNECT_PERMIT_MESSAGE:
            {
            std::ostringstream  s;
            if( conn_info->get_permit_auto_close() )
                {
                s << "Received permition to close connection to node " << sid;
                if( is_safe_to_disconnect( sid ) )
                    {
                    s << ".";
                    close_connection( sid );
                    }
                else
                    {
                    s << " but activity detected.";
                    conn_info->set_connection_status( ConnectionInfo::CONNECTED );
                    send_uncrypted_message
                        (
                        DISCONNECT_DENY_MESSAGE,
                        conn_info->get_tcpsocket(),
                        VariantPointer()
                        );
                    }
                }
            else
                {
                s << "Received permition to close connection to node " << sid;
                s << " but I cannot auto-close connection now.";
                conn_info->set_connection_status( ConnectionInfo::CONNECTED );
                send_uncrypted_message
                    (
                    DISCONNECT_DENY_MESSAGE,
                    conn_info->get_tcpsocket(),
                    VariantPointer()
                    );
                }
            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );
            return true;
            }

        case DISCONNECT_DENY_MESSAGE:
            {
            std::ostringstream  s;
            // Client cannot say Don't close my connection.
            if( ( sid.get_type() == NodeId::TYPE_CLIENT || sid.get_type() == NodeId::TYPE_ANONYMOUS ) &&
                conn_info->get_permit_auto_close() )
                {
                s << "Request to close connection to node " << sid << " denied,";
                s << " but the result is ignored because the node is client or "
                    "anonymous node.";
                close_connection( sid );
                }
            else
                {
                s << "Request denied to close connection to node " << sid << ".";
                conn_info->set_connection_status( ConnectionInfo::CONNECTED );
                }

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
                s.str()
                );

            return true;
            }

        case NOT_AUTHORIZED_MESSAGE:
            {
            std::ostringstream  s;

            s << "You are not authorized to connect to simulation.";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_HIGH ),
                s.str()
                );
            close_connection( sid );
            return true;
            }

        case NOT_READY_TO_LOGIN_MESSAGE:
            {
            std::ostringstream  s;

            s << "Simulation isn't ready to accept your login.";

            Global::log
                (
                STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_HIGH ),
                s.str()
                );
            close_connection( sid );
            return true;
            }

        default:
            {
            break;
            }
        }

    return false;
    }  // Network::process_net_system_message()


bool Network::process_error_message
    (
    UInt32           tag
    )
    {
    std::ostringstream oss;
        
    switch( tag )
        {
        case ALREADY_CONNECTED_MESSAGE:
            {
            // When server tries to connect more than once, it's ok, this situation
            // is possible when two servers are connecting to one another simultaneously.
            // When client tries to connect more than once, it's a ExConnectError exception.
            // ExConnectError exception raises exception to higher level.
            
            oss << "You are already connected!";
            if( my_node_id.get_type() == NodeId::TYPE_SERVER )
                throw ExConnectionExistsError( oss.str() );
            else
                throw ExConnectError( oss.str() );

            return true;
            }

        case NOT_AUTHORIZED_MESSAGE:
            {
            throw ExConnectError( "You are not authorized to connect to server." );
            }
        }
    return false;
    } // Network::process_error_message


void Network::process_blowfish_new_key_message
    (
    const NodeId &                    node_id_other,
    ConnectionInfo *                  conn_info,
    NetIBuffer &                      ibuffer,
    ConnectionInfo::BlowfishKeyPeriod message_period
    )
    {
    hope( conn_info );
    hope( !conn_info->is_server_side() );

    ConnectionInfo::BlowfishKeyPeriod period =
        conn_info->get_blowfish_key_period();

    unsigned int        length = ibuffer.read_int( 32 );

    // Decrypt the message body.

    BitIOStreamBuffer   crypted = ibuffer.get_buffer();
    BitIOStreamBuffer   decrypted;

    decrypted.resize( length );

    hope( conn_info->get_blowfish_decryptor_for_tcp( message_period ) );

    conn_info->get_blowfish_decryptor_for_tcp( message_period )->process_range
        (
        crypted.begin() + NET_BUFFER_HEAD_LEN_TCP,
        crypted.begin() + NET_BUFFER_HEAD_LEN_TCP + length,
        decrypted.begin()
        );

    ibuffer.bind( decrypted, false );

    // Read the key from the message.
    std::string new_key;
    new_key = ibuffer.read_full_string();

    // Set the new key.

    conn_info->set_blowfish_key( new_key );

    // Alter the blowfish resend algorithm period.

    if( period == ConnectionInfo::BLOWFISH_PERIOD_ODD )
        {
        conn_info->set_blowfish_key_period
            ( 
            ConnectionInfo::BLOWFISH_PERIOD_EVEN 
            );
        }
    else  // period == ConnectionInfo::BLOWFISH_PERIOD_EVEN
        {
        conn_info->set_blowfish_key_period
            ( 
            ConnectionInfo::BLOWFISH_PERIOD_ODD 
            );
        }

    // Send acknowledgement to the server.

    send_uncrypted_message
        (
        NEW_KEY_CONFIRMATION_MESSAGE,
        conn_info->get_tcpsocket(),
        VariantPointer()
        );
    
    net_stat.inc_blowfish_expirations();
    }  // Network::process_blowfish_new_key_message()


void Network::process_new_key_confirmation_message
    (
    ConnectionInfo *        conn_info,
    NetIBuffer &            ibuffer
    )
    {
    hope( conn_info );
    hope( conn_info->is_server_side() );

    ConnectionInfo::BlowfishKeyPeriod period =
        conn_info->get_blowfish_key_period();

    if( ( period != ConnectionInfo::BLOWFISH_PERIOD_EVEN_TRANSITION ) &&
        ( period != ConnectionInfo::BLOWFISH_PERIOD_ODD_TRANSITION ) )
        {
        throw ExWrongBlowfishPeriod
            ( 
            "Unexpected NEW_KEY_CONFIRMATION_MESSAGE received." 
            );
        }

    if( ibuffer.read_int( 32 ) != 3 )
        {
        throw ExMessageCorrupted
            ( 
            "Invalid length of the NEW_KEY_CONFIRMATION_MESSAGE." 
            );
        }

    if( ibuffer.read_string() == "OK" )
        {
        // Confirmation ok.
        if( period == ConnectionInfo::BLOWFISH_PERIOD_EVEN_TRANSITION )
            {
            conn_info->set_blowfish_key_period
                ( 
                ConnectionInfo::BLOWFISH_PERIOD_EVEN 
                );
            }
        else
            {
            // period == ConnectionInfo::BLOWFISH_PERIOD_ODD_TRANSITION
            conn_info->set_blowfish_key_period
                ( 
                ConnectionInfo::BLOWFISH_PERIOD_ODD 
                );
            }

        // Start new timeout.
        conn_info->touch_blowfish_key_timestamp();
        }
    else
        {
        // Bad confirmation text received.
        throw ExMessageCorrupted
            ( 
            "Corrupted body of the packet (NEW_KEY_CONFIRMATION_MESSAGE)." 
            );
        }
    }  // Network::process_new_key_confirmation_message()


bool Network::process_tcp_packet
    ( 
    BitIStream                          recvbuf,
    ConnectionInfo *                    conn_info
    )
    {
    hope( conn_info );
    
    static NetIBuffer   buffer_head;

    // There should be only one message in this packet.

    UInt16 period = recvbuf.read_int( 16 );
    buffer_head.tag = recvbuf.read_int( 32 );

    #ifdef LOG_MESSAGES

    std::ostringstream  s;

    s   << "Received a TCP message from node "
        << conn_info->get_node_id()
        << ", tag: "
        << buffer_head.tag
        << ", size: " << recvbuf_size
        << " bytes.";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        s.str()
        );

    #endif

    buffer_head.bind( recvbuf.get_buffer(), false, 48 );

    if( process_net_system_message( conn_info, buffer_head, 
            ( ConnectionInfo::BlowfishKeyPeriod ) period ) )
        {
        return false;
        }

    // Decrypting.

    NetIBuffer & buffer = decrypt_tcp_packet
        ( recvbuf, conn_info, ( ConnectionInfo::BlowfishKeyPeriod ) period );

    // Dispatch packet.

    SystemMessage::dispatch_system_message_from_packet
        (
        buffer,
        conn_info->get_node_id()
        );
    
    buffer_head.unbind();
    buffer.unbind();
    return true;
    }


NetIBuffer & Network::decrypt_tcp_packet
    ( 
    BitIStream                          recvbuf,
    ConnectionInfo *                    conn_info,
    ConnectionInfo::BlowfishKeyPeriod   period
    )
    {
    hope( conn_info );
    
    static BitIOStreamBuffer    crypted;
    BitIOStreamBuffer           decrypted;
    static NetIBuffer           buffer;

    int len = recvbuf.read_int( 32 );

    crypted = recvbuf.get_buffer();
    decrypted.resize( len );

    hope( conn_info->get_blowfish_decryptor_for_tcp( period ) );
    conn_info->get_blowfish_decryptor_for_tcp( period )->process_range
        (
        crypted.begin() + NET_BUFFER_HEAD_LEN_TCP,
        crypted.begin()  + NET_BUFFER_HEAD_LEN_TCP + len,
        decrypted.begin()
        );

    buffer.bind( decrypted, false );
    return buffer;
    } // Network::decrypt_tcp_packet


int Network::decrypt_and_process_udp_datagram
    ( 
    BitIStream          recvbuf,
    ConnectionInfo *    conn_info 
    )
    {
    hope( conn_info );
    
    BitIOStreamBuffer           crypted;
    BitIOStreamBuffer           decrypted;
    NetIBuffer                  buffer;
    int                         received_messages = 0;
    int                         position = 2;

    #ifdef LOG_MESSAGES

    std::ostringstream  ss;
    ss << "Receiving UDP buffer using blowfish key " <<
         quote_const_string( conn_info->get_blowfish_key_odd() ) <<
         " from node " << conn_info->get_node_id() << ".";
    Global::log_debug( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST, ss.str() );

    #endif

    // Check for special UDP init message...
        
    if( recvbuf.get_buffer().size() == 14 ) // Tohle je super.
        {
        if( conn_info->is_udp_connected() )
            {
            return 0;
            }
        int count = recvbuf.read_int( 16 );
        if( count != 1 )
            throw ExMessageCorrupted( "Corrupted initial UDP message! ( bad message count )" );
        int tag = recvbuf.read_int( 32 );
        if( tag != UDP_INIT_MESSAGE )
            {
            throw ExMessageCorrupted( "Corrupted initial UDP message! ( bad tag )" );
            }
        
        recvbuf.read_int( 32 ); // Size.
        
        NodeId sid = NodeId::create_from_uint32( recvbuf.read_int( 32 ) );
        
        if( !( my_node_id.is_type( NodeId::TYPE_SERVICE ) 
            && sid.is_type( NodeId::TYPE_ANONYMOUS ) ) )
            {
            if( sid != conn_info->get_node_id() )
                {
                std::ostringstream oss;
                oss << "Corrupted initial UDP message! ( bad nodeid " 
                    << sid << " awaited " << conn_info->get_node_id() << ")";
                throw ExMessageCorrupted( oss.str() );
                }
            }
        conn_info->set_udp_connected( true );
        send_uncrypted_message( UDP_HANDSHAKE_MESSAGE, conn_info->get_tcpsocket(), VariantPointer() );
        return 0;
        }

    // Rewind.
        
    recvbuf.seek( 0, BitIStream::WHENCE_START );
    
    if( recvbuf.get_buffer().size() % 8 != NET_BUFFER_FOOT_LEN_UDP )
        {
        std::ostringstream oss;
        oss << "Bad packet size! (" << recvbuf.get_buffer().size() << ")"; 
        throw ExMessageCorrupted( oss.str().c_str() );
        }
    
    // Get the blowfish resend algorithm phase.
        
    ConnectionInfo::BlowfishKeyPeriod   message_blowfish_period =
        ( ConnectionInfo::BlowfishKeyPeriod )
        ( recvbuf.get_buffer()[ recvbuf.get_buffer().size() - 5 ] & 0x01 );

    // Decrypt.
    
    crypted = recvbuf.get_buffer();

    int result_size = crypted.size() - 1;

    decrypted.resize( result_size );
        // We need to decrement crypted.size() because the flags field 
        // won't be copied into the decrypted buffer and thus this buffer
        // will be less.
    
    #ifdef LOG_MESSAGES

    ss.str( std::string() );
    std::size_t bs = crypted.size();
    ss << "Decrypting UDP buffer: period " << ( int ) message_blowfish_period <<
        ", crc: " << std::ios::hex << ( int ) crypted[ bs - 4 ] <<
        ( int ) crypted[ bs - 3 ] << ( int ) crypted[ bs - 2 ] << (int)crypted[ bs - 1 ];
    Global::log_debug( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST, ss.str() );
        
    #endif

    hope( conn_info->get_blowfish_decryptor_for_udp( message_blowfish_period ) );
    conn_info->get_blowfish_decryptor_for_udp( message_blowfish_period )->
        process_range
            (
            crypted.begin(),
            crypted.begin() + crypted.size() - NET_BUFFER_FOOT_LEN_UDP,
            decrypted.begin()
            );

    // Append checksum computed from the plain text by the sender node.    
    
    decrypted[ result_size - 4 ] = crypted[ crypted.size() - 4 ];
    decrypted[ result_size - 3 ] = crypted[ crypted.size() - 3 ];
    decrypted[ result_size - 2 ] = crypted[ crypted.size() - 2 ];
    decrypted[ result_size - 1 ] = crypted[ crypted.size() - 1 ];

    // Verify checksum.
    
    buffer.bind( decrypted );    
    
    if( !buffer.engine_checksum.validate_checksum() )
        {
        // Checksum is not correct, throwing the message away...
        buffer.unbind();
        throw ExMessageCorrupted( msg_invalid_checksum );
        }

    // How many messages do we have?

    buffer.bind_part( decrypted, false, 0, position );
    UInt16 count = buffer.read_int( 16 );
    buffer.unbind();

    #ifdef LOG_MESSAGES

    ss.str( std::string() );

    ss  << "Received an UDP packet ("
        << decrypted.size()
        << " bytes) from "
        << conn_info->get_node_id()
        << " carrying "
        << count
        << " messages.";

    Global::log
        (
        STATUS_INFO( Status::FACILITY_NETWORK, Status::PRIORITY_LOWEST ),
        ss.str()
        );

    #endif

    net_stat.inc_messages_received( count );

    // For each received message do:

    for( int i = 0; i < count ; i++ )
        {
        // Read the rest of the header.

        buffer.bind_part
            (
            decrypted,
            false,
            position,
            2 * sizeof( UInt32 )
            );

        UInt32 tag = buffer.read_int( 32 );
        UInt32 len = buffer.read_int( 32 );
        buffer.unbind();

        // Read the message.

        buffer.bind_part
            (
            decrypted,
            false,
            position + 2 * sizeof( UInt32 ),
            len
            );

        buffer.tag = tag;
        position += len + 2 * sizeof( UInt32 );

        // Store the message.

        SystemMessage::dispatch_system_message_from_packet
            (
            buffer,
            conn_info->get_node_id()
            );

        buffer.unbind();
        
        received_messages++;
        }
    return received_messages;
    }  //Network::decrypt_and_process_udp_datagram


void Network::tcp_packet_received
    (
    ConnectionInfo * conn_info,
    UInt32           len
    )
    {
    hope( conn_info );
    
    net_stat.inc_tcp_bytes_received( len );
    net_stat.inc_messages_received( 1 );
    conn_info->reset_idle_ticks();
    conn_info->touch_last_message_time();
    } // Network::tcp_packet_received


void Network::udp_datagram_received
    (
    ConnectionInfo *    conn_info,
    UInt32              len
    )
    {
    hope( conn_info );
    net_stat.inc_udp_bytes_received( len );
    conn_info->reset_idle_ticks();
    conn_info->touch_last_message_time();
    } // Network::udp_datagram_received


void Network::bad_packet_received
    (
    ConnectionInfo * conn_info
    )
    {
    hope( conn_info );
    
    net_stat.inc_bad_packets_received();
    conn_info->set_disconnection_reason( ConnectionInfo::ERROR_REASON );
    conn_info->set_connection_status( ConnectionInfo::NOT_CONNECTED );
    }


}; // namespace Massiv::Core
}; // namespace Massiv
