/** @file netobuffer.h
 * Bit-oriented network output stream buffer with "crypto" features.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NETOBUFFER_H
#define MASSIV_NETOBUFFER_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_CONNECTION_INFO_H
#include "net/connection_info.h"
#endif

#ifndef MASSIV_BITOSTREAM_H
#include "net/bitostream.h"
#endif


namespace Massiv {
namespace Core {


class Network;


/** @addtogroup net
 * @{
 */

/**
 * Bit-oriented network output stream. Contains abstract methods,
 * must be overriden.
 */
class NetOBuffer : public BitOStream
    {
    friend class Massiv::Core::Network;

public:
    
    /**
     * Default constructor.
     */
    NetOBuffer();
 
    /**
     * Default destructor.
     */
    virtual ~NetOBuffer();

private:

    /**
     * Initializes buffer - i.e. writes a message tag, reserves room
     * for the message length and for number of messages in buffer, ...
     */
    virtual void init
        ( 
        UInt32 tag  // buffer tag
        ) = 0;

    /**
     * Finishes buffer - i.e. writes message size into a reserved space
     * and encrypts it into crypted_buffer using blowfish cipher.
     */
    virtual void flush() = 0;

protected:

    /**
     * Dumps buffer state.
     */
    virtual std::string to_string() const;

public:

    enum BufferStatus 
        {
        NET_BUFFER_LOCKED =         0x01,
            /**< The buffer has been locked, i.e. someone has reserved it
                 for writing. */

        NET_BUFFER_UNLOCKED =       0x02,
            /**< The buffer is not locked, it may be locked. */

        NET_BUFFER_FLUSH_ALLOWED =  0x03,
            /**< Flushing the buffer is allowed. */

        NET_BUFFER_CRYPTED =        0x04
            /**< The buffer has been encrypted, it is ready to be sent. 
                 No one can lock the crypted buffer. */                 
        };  // enum BufferStatus


    BufferStatus        status;
        /**< Buffer status ( locked, unlocked, ready, etc ). */

    ConnectionInfo *    connection_info;
        /**< Information about the connection this buffer is being used by. */

    bool                permanent;
        /**< true if the buffer is permanent - i.e. pre-allocated. */

    NodeId              owner;
        /**< Buffer owner. */

protected:

    BitIOStreamBuffer    crypted_buffer;
        /**< Crypted output buffer. */

    };  // class NetOBuffer

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_NETOBUFFER_H
