/** @file serveraddress.h
 * Server address.
 */

/*
 * begin        : Tue May 15 2002
 * copyright    : (C) 2002 by Ondrej Pecta, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SERVERADDRESS_H
#define MASSIV_SERVERADDRESS_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


namespace Test { class SocketsTest; }


/** @addtogroup net
 * @{
 */

/**
 * Stores server address and port.
 * Implements the IP address resolving for a specified DNS name (but not vice
 * versa).
 */
class ServerAddress
    {
    friend class Massiv::Core::Test::SocketsTest;

public:

    /**
     * Constructor.
     * @param address DNS name or IP address in the textual form.
     *      Both forms of address will be automaticaly distinguished.
     * @param server_port The port the server listens on. It should be
     *      given in a HOST byte order.
     * @param caching Determines whether the IP address will be "cached"
     *      or not.
     * @note The textual form of IP address has form of a.b.c.d,
     *      where @a a to @d are decimal integer values 0-255.
     * @note No address resolving is being made in the constructor.
     * @note The @a name string will not be validated.
     */
    ServerAddress
        (
        std::string     address,
        UInt16          server_port,
        bool            caching = true
        );

    /**
     * Constructor.
     * @param ip IP address of the server. Format of this parameter is
     *      the same as the return value of inet_addr() of WinSock (or
     *      analogical library under Linux). You should never pass
     *      other value than the one got via calling this-\>get_ip()
     *      or this-\>assemble_ip().
     * @param server_port The port the server listens on. It should be
     *      given in a HOST byte order.
     * @param caching Determines whether the IP address will be "cached"
     *      or not.
     */
    ServerAddress
        (
        UInt32  ip,
        UInt16  server_port,
        bool    caching = true
        );

    /**
     * Default constructor
     */
    ServerAddress();

    /**
     * Destructor
     */
    ~ServerAddress();

    /**
     * Operator=.
     */
    ServerAddress & operator=
        (
        const ServerAddress & sa
        );

public:

    /**
     * @return Server port in the HOST byte order.
     */
    UInt16 get_port() const;

    /**
     * @return Server address (IP).
     * @note The returned IP should be handled as an opaque structure!
     */
    UInt32 get_ip() const;

    /**
     * @return Server address (DNS name).
     * @Note automatically sets dns name from assembled string from IP
     * when no dns filled and IP set.
     */
    std::string get_dns_name() const;

    /**
     * Sets port to value @a port.
     * @param port Port in the HOST byte order.
     */
    void set_port
        (
        UInt16 port
        );

    /**
     * Sets a new server address (IP).
     * @param ip The new ip address. Only values returned from
     *      this-\>get_port() or this-\>assemble_ip() should be
     *      used.
     */
    void set_ip
        (
        UInt32 ip
        );

    /**
     * Sets a new server address. No checks are done, arbitrary
     * string will be accepted in this step.
     * @param address The format of address is the same as in constructor
     *      - either DNS name or IP in textual form.
     */
    void set_addr
        (
        const std::string & address
        );

    /**
     * @param caching If true, the resolved IP address will be "cached".
     * @return The old mode of caching.
     */
    bool set_caching
        (
        bool caching
        );

    /**
     * Resolves an IP address.
     * @param dns_name Name address.
     * @return The resolved IP address.
     * @exception ExInvalidDnsName IP address could not be found for @a dns_name.
     * @note The returned value should be handled as an opaque structure!
     */
    static UInt32 resolve_ip_address
        (
        const std::string & dns_name
        );

    /**
     * Converts given string into the IP address.
     * @param string IP address in the textual form (see constructor for
     *      description).
     * @return IP address in format such as returned from inet_addr or
     *      IP_UNDEFINED if the string contains no IP address.
     * @note The returned value should be handled as an opaque structure!
     */
    static UInt32 string_to_ip
        (
        const std::string & string
        );

    /**
     * Converts saved UInt32 ip string into the IP address.
     * @return std::string dns address built from ip.
     */
    std::string ip_to_string();

    /**
     * Resolves an IP address (according to the @c dns_name member of this
     * class).
     * @return The resolved IP address in format such as if it was returned
     *      from inet_addr() (i.e. NETWORK byte order).
     * @exception ExInvalidDnsName IP address could not be found for DNS name
     *      specified by @a dns_name member of this class.
     * @note The returned value should be handled as an opaque structure!
     */
    UInt32 resolve_ip_address() const;

    /**
     * @return Separate fields of the IP address. This should be, because
     *      of the portability, the only way how to get these.
     * @note Implemented in arch.
     */
    void get_ip_fields
        (
        UInt8 &  field1,
        UInt8 &  field2,
        UInt8 &  field3,
        UInt8 &  field4
        );

    /**
     * Assembles IP address from its fields. No check of parameters validity
     * is done.
     * @return The assembled address in format such as if it had been returned
     *      from inet_addr() (i.e. NETWORK byte order).
     * @note Implemented in arch.
     */
    UInt32 assemble_ip
        (
        UInt8 field1,
        UInt8 field2,
        UInt8 field3,
        UInt8 field4
        );

    /**
     * Converts a 16-bit unsigned integer from the host byte order to
     * the network byte order.
     */
    static UInt16 host_to_network_short
        (
        UInt16 host
        );

    /**
     * Converts a 16-bit unsigned integer from the network byte order to
     * the host byte order.
     */
    static UInt16 network_to_host_short
        (
        UInt16 net
        );

public:

    /**
     * IP_UNDEFINED value should be assigned into @a ip if the ip address
     * is currently unknown (i.e. the ServerAddress instance has been
     * initialized through DNS name and IP address has not been resolved yet.
     */
    enum
        {
        IP_UNDEFINED = -1
        };

    /**
     * Analogy of IP_UNDEFINED for DNS name.
     */
    #define DNS_NAME_UNDEFINED  ""

private:

    UInt16          port;
        /**< Server port, stored in the NETWORK byte order. */

    std::string     dns_name;
        /**< Server name address (DNS name). Invalid DNS is signalled by an
             empty string. */

    mutable UInt32  ip;
        /**< Server IP address. An "invalid" IP value is IP_UNDEFINED.
             @note Stored in a format such as if the value had been returned
                   from the inet_addr() routine, i.e. the NETWORK order. For
                   portability nobody should explicitly count with this
                   order.
             @note We suppose IPv4 protocol, for IPv6 (IPng)
             the address would be 16 bytes long! */

    bool            caching;
        /**< If true, the IP address will not be resolved each time
             its needed. Instead, a DNS query will take place only
             when DNS name has been changed or when it's explicitly
             demanded. */

    };  // class ServerAddress

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_SERVERADDRESS_H
