/** @file socket.cpp
 * Socket interface.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/net.h"
#include "net/socket.h"
#include "net/bitostream.h"
#include <sstream>

namespace Massiv {
namespace Core {

using namespace NetworkCommon;


#ifdef OPTION_DEBUG

/**
 * Define this to enable logging of packet sending/receiving.
 */
//#define DEBUG_PACKETS

#endif


// Implementation of Socket
// ~~~~~~~~~~~~~~~~~~~~~~~~

Socket::Socket() :
    server_port( 0 ),
    server_start_timestamp( SystemTime::INVALID )
    {
    }


Socket::~Socket()
    {
    }


int Socket::get_tcp_server_port()
    {
    return server_port;
    }


void Socket::send
    (
    BitOStream & stream
    )
    {
    BitIOStreamBuffer buffer = stream.get_buffer();
    send( buffer );
    }


void Socket::send
    (
    BitIOStreamBuffer & buffer
    )
    {
    send( buffer.get_buffer(), buffer.size() );
    }


void Socket::touch_server_start_timestamp()
    {
    server_start_timestamp = Global::system_time();
    }


SystemTime Socket::get_server_start_timestamp()
    {
    return server_start_timestamp;
    }


// Implementation of TCPSocket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
TCPSocket::TCPSocket()
    :   body_len( 0 ),
        received_bytes( 0 )
    {
    }


void TCPSocket::close()
    {
    body_len = 0;
    received_bytes = 0;
    recvbuf.clear();
    }


void TCPSocket::send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    // Send data in non-blocking mode.
    // We are assuming that socket is in non-blocking mode now.

    int bytes_sent = internal_send( buf, len );
        
    if( bytes_sent == len )
        {
        #ifdef DEBUG_PACKETS

        std::ostringstream oss;
        oss << "TCP message (" << len << " bytes) was sent in non-blocking mode.";
        
        Global::log_debug
            ( Status::FACILITY_NETWORK, Status::PRIORITY_LOW, oss.str() );
        
        #endif

        return;
        }

    // Send rest of the data in blocking mode.
    
    #ifdef DEBUG_PACKETS

    {
    std::ostringstream oss;
    oss << "Sending TCP message (" << len - bytes_sent << " of " << len
        << " bytes) in blocking mode.";
    
    Global::log_warning
        ( Status::FACILITY_NETWORK, Status::PRIORITY_HIGH, oss.str() );
    }

    #endif

    set_non_block( false );

    try
        {
        while( bytes_sent < len )
            {
            std::ostringstream  oss;
            int                 sent_now;
            
            sent_now = internal_send
                (
                buf + bytes_sent,
                len - bytes_sent
                );

            if( !sent_now )
                {
                throw ExSendError( "Can not send TCP data even in blocking mode!" );
                }

            bytes_sent += sent_now;
            }
        }
    catch( Exception & )
        {
        set_non_block( true );
        throw;
        }
    
    set_non_block( true );
    }


bool TCPSocket::receive
    (
    BitIStream &  buf,
    UInt32 &      bytes_available
    )
    {
    Network &   net = Global::network();
    UInt32      old_received_bytes = received_bytes;
    UInt32      waiting_bytes = 0;
    UInt32      data_fetched = 0;
    
    waiting_bytes = get_pending_data_amount();
    bytes_available = 0;

    // If there is nothing to receive, quit.

    if( waiting_bytes == 0 )
        {
        return false;
        }
    
    // Read message header if it was not done yet.

    if( received_bytes < NET_BUFFER_HEAD_LEN_TCP && waiting_bytes > 0 )
        {
        if( received_bytes == 0 && waiting_bytes > 0 )
            {
            recvbuf.resize( NET_BUFFER_HEAD_LEN_TCP );
            }

        data_fetched = internal_receive
            (
            recvbuf.get_buffer() + received_bytes,
            NET_BUFFER_HEAD_LEN_TCP - received_bytes
            );

        if( data_fetched == 0 )
            {
            return false;
            }

        received_bytes += data_fetched;

        if( received_bytes == NET_BUFFER_HEAD_LEN_TCP )
            {
            // Parse header and determine message body length.

            static BitIStream header_stream;

            header_stream.bind( recvbuf, false, 0 );
            header_stream.skip( 48 );
            body_len = header_stream.read_int( 32 );
            header_stream.unbind();

            // Resize receive buffer so that that message body could be
            // read and appended.

            recvbuf.resize( body_len + NET_BUFFER_HEAD_LEN_TCP );
            
            // Log.

            #ifdef DEBUG_PACKETS

            std::ostringstream oss;

            oss << "Fetched TCP message header. Expecting message "
                << recvbuf.size() << " bytes long.";

            Global::log_debug
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_LOW,
                oss.str()
                );

            #endif
            }
        else
            {
            return false;
            }
        }
    
    // Read rest of the message, but not more than @c max_receive_bytes
    // bytes in this tick.
    
    waiting_bytes = get_pending_data_amount();

    while( waiting_bytes > 0 && received_bytes < recvbuf.size() )
        {
        UInt32 total_data_fetched = received_bytes - old_received_bytes;
        UInt32 max_bytes_to_read = recvbuf.size() - received_bytes;
        UInt32 bytes_to_read;
        
        if( total_data_fetched + max_bytes_to_read > 
            static_cast<UInt32>( net.settings.max_receive_bytes ) )
            {
            max_bytes_to_read = net.settings.max_receive_bytes - total_data_fetched;
            }

        if( waiting_bytes > max_bytes_to_read )
            {
            bytes_to_read = max_bytes_to_read;
            }
        else
            {
            bytes_to_read = waiting_bytes;
            }

        if( bytes_to_read == 0 )
            {
            break;
            }

        data_fetched = internal_receive
            (
            recvbuf.get_buffer() + received_bytes,
            bytes_to_read
            );

        if( data_fetched == 0 )
            {
            break;
            }

        received_bytes += data_fetched;
        waiting_bytes = get_pending_data_amount();
        }

    if( received_bytes == recvbuf.size() )
        {
        // The whole message has been received.
        // Return it to the caller.

        bytes_available = received_bytes;
        buf.bind( recvbuf, false, 0 );

        // Prepare for header of a next message.

        received_bytes = 0;
        
        // Log.
        
        #ifdef DEBUG_PACKETS

        Global::log_debug
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_LOW,
            "TCP message was fetched."
            );

        #endif

        return true;
        }

    // Still waiting for the rest of the message.
    // Signal what part of the message has been fetched so far.

    bytes_available = received_bytes;

    // Log.
    
    #ifdef DEBUG_PACKETS

    std::ostringstream oss;

    oss << "Still waiting for the rest of the TCP message. " << bytes_available
        << " of " << recvbuf.size() << " bytes available.";
    
    Global::log_debug
        (
        Status::FACILITY_NETWORK,
        Status::PRIORITY_LOW,
        oss.str()
        );

    #endif

    return false;
    }


// Implementation of UDPSocket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

UDPSocket::UDPSocket()
    {
    }


Socket * UDPSocket::accept()
    {
    return NULL;
    }


void UDPSocket::send
    (
    const UInt8 *   buf,
    int             len
    )
    {
    int bytes_sent = internal_send( buf, len );
    
    if( bytes_sent < len )
        {
        std::ostringstream oss;
        
        oss << "Only part of the datagram was sent! " << len - bytes_sent
            << " byte(s) were thrown away by the OS.";
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            oss.str()
            );
        }
    }


bool UDPSocket::receive
    (
    BitIStream &  buf,
    UInt32 &      bytes_available
    )
    {
    static BitIOStreamBuffer    recvbuf;
    unsigned long               waiting_bytes;
    unsigned long               data_fetched;

    data_fetched = 0;
    bytes_available = 0;
    
    // Peek message header.
    
    waiting_bytes = get_pending_data_amount();
    
    if( waiting_bytes >= NET_BUFFER_HEAD_LEN_UDP )
        {
        // Resize buffer so that the whole message would fit into it,
        // then read the data.

        recvbuf.resize( waiting_bytes );

        data_fetched = internal_receive
            (
            recvbuf.get_buffer(),
            recvbuf.size()
            );

        // Return read data to caller.

        bytes_available = data_fetched;
        recvbuf.resize( data_fetched );
        buf.bind( recvbuf, false, 0 );

        return true;
        }
    else
    if( waiting_bytes > 0 )
        {
        UInt8 buffer[ 1 ];

        // We have received some garbage.
        // Skip it.
        
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            "Throwing away truncated datagram!"
            );

        internal_receive( buffer, sizeof( buffer ) );

        return false;
        }

    // No message available.

    return false;
    }


} // namespace Massiv::Core
} // namespace Massiv
