/** @file socket.h
 * Socket interface.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SOCKET_H
#define MASSIV_SOCKET_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_GLOBAL_H
#include "global/global.h"
#endif

#ifndef MASSIV_BITISTREAM_H
#include "net/bitistream.h"
#endif

#ifndef MASSIV_SERVERADDRESS_H
#include "net/serveraddress.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif


namespace Massiv {
namespace Core {
namespace Test { class SocketsTest; }


/** @addtogroup net
 * @{
 */

/**
 * Simple berkeley socket wrapper that provides API for sending messages.
 * Everything must be packaged into messages, streams are not supported.
 * Messages are sent either reliably or unreliably, which depends
 * on what transport is selected (TCP or UDP).
 */
class Socket
    {
    friend class Massiv::Core::Test::SocketsTest;
    friend class Network;

public:

    /**
     * Constructor.
     */
    Socket();

    /**
     * Destructor.
     */
    virtual ~Socket();

public:

    /**
     * Connects to server, returns true if ok.
     */
    virtual bool connect
        (
        const ServerAddress & address
        ) = 0;

    /**
     * Closes socket.
     */
    virtual void close() = 0;

    /**
     * Starts server (i.e. prepares socket for calling accept()).
     * @c backlog determines maximum number of simulatenously connected
     * clients, use zero for unlimited count. @c backlog is valid
     * for TCP clients only.
     */
    virtual bool start_server
        ( 
        UInt16  port,
        int     backlog = 0
        ) = 0;

    /**
     * Accept connections.
     * Returns new socket or NULL on error.
     */
    virtual Socket * accept() = 0;
    
    /**
     * Enables blocking/nonblocking socket IO.
     */
    virtual bool set_non_block
        (
        bool    non_blocking
        ) = 0;

    /**
     * Returns TCP port, which has this socket accepted.
     * Used for creating UDP on the same socket as TCP.
     */
    int get_tcp_server_port();    

    /**
     * Sets the timestamp of the socket to the current time. It should
     * be used while the socket is in the waiting-for-accept state.
     */
    void touch_server_start_timestamp();

    /**
     * @return Timestamp of the waiting-for-accept socket. If this
     *         value is "too high", the socket should be closed.
     */
    SystemTime get_server_start_timestamp();

    /**
     * @return The number of bytes pending on the socket; for TCP socket, it
     * returns maximum number of bytes that can be actually read using
     * a single call to @c recv().
     */
    virtual int get_pending_data_amount() = 0;

    /**
     * Sends message of given length.
     * @param buf Buffer to be sent.
     * @param len Length of the buffer.
     * @note It is up to caller to properly format the message
     * (ie. header, ...).
     * Throws @c ExSendError if message can not be sent.
     */
    virtual void send
        (
        const UInt8 *   buf,
        int             len
        ) = 0;
    
    /**
     * Sends contents of @c stream.
     */
    void send
        (
        BitOStream &    stream
        );
    
    /**
     * Sends contents of @c buffer.
     */
    void send
        (
        BitIOStreamBuffer & buffer
        );
    
    /**
     * Receives a message that was sent by @c send().
     * @param buf A destination buffer.
     * @param bytes_available How many bytes have been fetched.
     * Returns true if the whole message was received.
     * If false is returned, subsequent calls to @c receive() will
     * fetch the rest of the message.
     * Throws @c ExReceiveError if message can not be received.
     */
    virtual bool receive
        ( 
        BitIStream &    buf,
        UInt32 &        bytes_available
        ) = 0;
    
protected:

    /**
     * Tries to send @c len bytes from @c buffer.
     * Returns number of bytes actualy read or throws an exception
     * if an error occurs.
     */
    virtual int internal_send
        (
        const UInt8 *   buf,
        int             len
        ) = 0;
    
    /**
     * Tries to receive @c len bytes from the socket.
     * Returns number of bytes actualy received or throws an exception
     * if an error occurs.
     */
    virtual int internal_receive
        (
        UInt8 * buf,
        int     len
        ) = 0;


protected:

    ServerAddress   server_address;
        /**< Don't ask me why, but I need know, who I am connecting to.
             Small hint - when I connect to well-known port and receive
             SERVER_PORT_MESSAGE, I must know, who I should make connection to. */
    UInt16          server_port;
        /**< Port bound with this socket. */
    SystemTime      server_start_timestamp;
        /**< When server was started on this socket. */

    };  // class Socket


/**
 * Socket implemented over TCP transport.
 */
class TCPSocket : public Socket
    {
protected:
    
    /**
     */
    TCPSocket();

    /**
     */
    virtual void close();

    /**
     */
    virtual void send
        (
        const UInt8 *   buf,
        int             len
        );
    
    /**
     */
    virtual bool receive
        ( 
        BitIStream &    buf,
        UInt32 &        bytes_available
        );

protected:
    
    BitIOStreamBuffer   recvbuf;
        /**< Because large amount of data cannot be received in one
             tick, each socket has to have own recvbuf for partially
             received data.
             It's allocated size signals expected size of the message
             (including header). */ 
    UInt32              body_len;
        /**< Length of the body of the message. 0 if the message header
             is not available at the socket yet. */
    UInt32              received_bytes;
        /**< Number of bytes received from the message so far. */
       
    }; // class TCPSocket


/**
 * Socket implemented over UDP transport.
 */
class UDPSocket : public Socket
    {
protected:
    
    /**
     */
    UDPSocket();
    
    /**
     */
    virtual Socket * accept();
    
    /**
     */
    virtual void send
        (
        const UInt8 *   buf,
        int             len
        );
    
    /**
     */
    virtual bool receive
        ( 
        BitIStream &    buf,
        UInt32 &        bytes_available
        );

public:

    /**
     * Returns port bound to the socket.
     */
    virtual UInt16 get_bound_port() = 0;

    /**
     * Sets address of the other side, the address where data are sent
     * by @c internal_send().
     * If not set manually, it will be automatically set by @c connect()
     * or @c receive() to the address of the other-side node.
     */
    virtual void set_destination_address
        (
        const ServerAddress & sa
        ) = 0;
            
    }; // class UDPSocket

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_SOCKET_H
