/** @file udpobuffer.cpp
 * Bit-oriented udp output stream buffer with crypting.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "net/udpobuffer.h"
#include "net/bitostream.h"
#include "net/connection_info_impl.h"
#include "utility/quote_string.h"
#include <sstream>


namespace Massiv {
namespace Core {

using namespace NetworkCommon;



UDPOBuffer::UDPOBuffer()
    :   engine_checksum( &crypted_buffer ),
        msg_start( 0 ),
        count( 0 )
    {
    // Reserve room in internal buffer - both of the should 
    // have the same size reserved.
    buffer.reserve
        ( NET_BUFFER_DEFAULT_LEN + NET_BUFFER_HEAD_LEN_UDP );

    crypted_buffer.reserve
        ( NET_BUFFER_DEFAULT_LEN + NET_BUFFER_HEAD_LEN_UDP );
    }


UDPOBuffer::UDPOBuffer
    (
    const UDPOBuffer &      sourcebuf
    ) : 
    NetOBuffer( sourcebuf ),
    engine_checksum( sourcebuf.engine_checksum ),
    msg_start( sourcebuf.msg_start ),
    count( sourcebuf.count )
    {
    owner =  sourcebuf.owner;
    engine_checksum.set_buffer( &buffer );
    }


UDPOBuffer::~UDPOBuffer()
    {
    }    


std::string UDPOBuffer::to_string() const
    {
    std::ostringstream ss;

    ss  << NetOBuffer::to_string()
        << "\ncount = " << count;

    return ss.str();
    }


void UDPOBuffer::init
    (
    UInt32  tag
    )
    {
    count = 0;

    // Reserve two bytes for msg counter.
    write_int( 0, 16 );
    
    // Complete init.
    msg_init( tag );
    }


void UDPOBuffer::msg_init
    (
    UInt32  tag
    )
    {
    hope( byte_size() > 0 ); 
    
    // New message => inc count.
    count++;

    align_to_byte();
       
    // Where I append.
    msg_start = byte_size();
    
    // Write message tag.
    write_int( tag, 32 );
    
    // Reserve space for message len.
    align_fill( 0, 4 );
    
    // Buffer is ready for bit stream.
    status = NET_BUFFER_LOCKED;
    }


void UDPOBuffer::flush()
    {
    // Status should be NET_BUFFER_FLUSH_ALLOWED.
    hope( status == NET_BUFFER_FLUSH_ALLOWED );
    hope( connection_info->get_blowfish_encryptor_for_udp() );
        
    unsigned int required_size;
    unsigned int crypted_size;

    #ifdef OPTION_DEBUG

    std::ostringstream  ss;
    ss << "Flushing UDP buffer using blowfish key " <<
         quote_const_string( connection_info->get_blowfish_key_odd() ) <<
         " to node " << connection_info->get_node_id() << ".";
    Global::log_debug( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST, ss.str() );

    #endif

    connection_info->get_blowfish_encryptor_for_udp()->align( buffer );
    required_size = byte_size() + sizeof( Crc32::ChecksumType ) + sizeof( byte );

    if( required_size <= MAX_WAITING_BYTES || count == 1 )
        {
        if( required_size > MAX_WAITING_BYTES )
            {
            Global::log_warning
                (
                Status::FACILITY_NETWORK,
                Status::PRIORITY_HIGH,
                "UDP packet too long, corruption possibility!"
                );
            }

        crypted_size = byte_size();
        }
    else
        {
        Global::log_warning
            (
            Status::FACILITY_NETWORK,
            Status::PRIORITY_HIGH,
            "Cutting last packet into two smaller because of very big packet length!"
            );
        
        // Maybe locking new buffer for the dropped message will be better solution.
        // Not tested, but should work. Clear this in case of problems.
        // Begin.
        Network & network = Global::network();

        NetIBuffer ibuffer;
        ibuffer.bind( buffer, false );
        ibuffer.skip( msg_start * 8 );
        UInt32 tag = ibuffer.read_int( 32 );
        UInt32 size = ibuffer.read_int( 32 );
        unsigned int i = 0;

        // Owner is the same as this buffer owner.
        UDPOBuffer * buf = network.lock_unreliable( owner, tag );

        for( i = 0; i < size; i++ )
            {
            buf->write_int( ibuffer.read_int( 8 ), 8 );
            }

        network.unlock( buf );
        // End.

        // Writes to the front of the buffer nr of msgs -1, cos of dropping the last.
        count -= 1;
        
        // Align.
        unsigned int padding = ( 8 - msg_start % 8 ) % 8;
        for( i = 0; i < padding; i++ )
            {
            buffer[ msg_start + i ] = ( UInt8 ) 0;
            }
        
        crypted_size = msg_start + padding;
        }
    
    // Writes to the front of the buffer nr of msgs.
    
    buffer[ 0 ] = ( UInt8 ) count;
    buffer[ 1 ] = ( UInt8 ) ( count >> 8 );

    hope( crypted_size % 8 == 0 );
    crypted_buffer.resize( crypted_size );
    hope( crypted_buffer.size() == crypted_size ); 
    
    // Write flags field (so far only blowfish key resend algorithm
    // phase flag is available )
    crypted_buffer.push_back
        ( ( UInt8 )( connection_info->get_blowfish_key_period() & 0x1 ) );

    // Write 32-bit CRC at the buffer's end.
    // ERROR! FIXME append_checksum() sometimes does bad crc!

    // append_checksum();

    // Slower solution becouse CRC32 constructor is slow.
    // Why it always initializes table? Static values will be better than
    // every time new generating.
    Crc32               crc_engine( &buffer );
    Crc32::ChecksumType crc = crc_engine.compute_checksum();
    append_checksum( crc );
    
    #ifdef OPTION_DEBUG

    ss.str( std::string() );
    std::size_t bs = crypted_buffer.size();
    ss << "Encrypting UDP buffer: period " << ( int ) crypted_buffer[ bs - 5 ] <<
        ", crc: " << std::ios::hex << ( int ) crypted_buffer[ bs - 4 ] <<
        ( int ) crypted_buffer[ bs - 3 ] << ( int ) crypted_buffer[ bs - 2 ] <<
        ( int ) crypted_buffer[ bs - 1 ];
    Global::log_debug( Status::FACILITY_NETWORK, Status::PRIORITY_HIGHEST, ss.str() );
    
    #endif

    // Whole buffer will be encrypted.
    connection_info->get_blowfish_encryptor_for_udp()->process_range
        ( 
        buffer.begin(), 
        buffer.begin() + crypted_size,
        crypted_buffer.begin() 
        );

    hope( crypted_buffer.size() % 8 == NET_BUFFER_FOOT_LEN_UDP );
        
    // DEBUG CRAP. Check whether computed checksum is really correct.
    // This is a validation of the checksum algorithm.
    #ifdef OPTION_DEBUG
        {
        BitIOStreamBuffer decrypted;
        decrypted.resize( crypted_buffer.size() - NET_BUFFER_FOOT_LEN_UDP );

        ConnectionInfo::BlowfishKeyPeriod   message_blowfish_period =
            ( ConnectionInfo::BlowfishKeyPeriod ) ( connection_info->get_blowfish_key_period() & 0x01 );   
        
        hope( connection_info->get_blowfish_decryptor_for_udp( message_blowfish_period ) );
        
        connection_info->get_blowfish_decryptor_for_udp( message_blowfish_period )->
            process_range
                (
                crypted_buffer.begin(),
                crypted_buffer.begin() + crypted_buffer.size() - NET_BUFFER_FOOT_LEN_UDP,
                decrypted.begin()
                );

        Crc32 old_crc_engine( &buffer );
        Crc32::ChecksumType crc_old = old_crc_engine.compute_checksum();
        
        Crc32::ChecksumType crc_check = crypted_buffer[ crypted_buffer.size() - 1 ];
        crc_check = ( crc_check << 8 ) | crypted_buffer[ crypted_buffer.size() - 2 ];
        crc_check = ( crc_check << 8 ) | crypted_buffer[ crypted_buffer.size() - 3 ];
        crc_check = ( crc_check << 8 ) | crypted_buffer[ crypted_buffer.size() - 4 ];
        
        hope( crc_check == crc_old );
        }
    #endif

    // Mark the buffer "ready to be sent".
    status = NET_BUFFER_CRYPTED;
    }


void UDPOBuffer::msg_flush()
    {
    // Status should be NET_BUFFER_UNLOCKED.
    hope( status != NET_BUFFER_LOCKED );
    
    // writes number of messages to the front of the buffer.
    // note: isn't needed here, it's always done by flush()
    buffer[0] = ( UInt8 ) count;
    buffer[1] = ( UInt8 ) ( count >> 8 );
    
    int size = byte_size() - NET_BUFFER_HEAD_LEN_UDP - msg_start + 2;

    buffer[ msg_start + 4 ] = ( UInt8 ) size;
    buffer[ msg_start + 5 ] = ( UInt8 ) ( size >> 8 );
    buffer[ msg_start + 6 ] = ( UInt8 ) ( size >> 16 );
    buffer[ msg_start + 7 ] = ( UInt8 ) ( size >> 24 );
    }


void UDPOBuffer::append_checksum
    (
    Crc32::ChecksumType     checksum
    )
    {
    crypted_buffer.push_back( checksum & 0x000000ff );
    crypted_buffer.push_back( ( checksum & 0x0000ff00 ) >> 8 );
    crypted_buffer.push_back( ( checksum & 0x00ff0000 ) >> 16 );
    crypted_buffer.push_back( ( checksum & 0xff000000 ) >> 24 );
    }


int UDPOBuffer::get_count_msgs() const
    {
    return count;
    }


} // namespace Massiv::Core
} // namespace Massiv
