/** @file udpobuffer.h
 * Bit-oriented udp output stream buffer with crypto
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_UDPOBUFFER_H
#define MASSIV_UDPOBUFFER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_MASSIV_CHECKSUM_H
#include "net/massiv_checksum.h"
#endif

#ifndef MASSIV_MASSIV_CHECKSUM_IMPL_H
#include "net/massiv_checksum_impl.h"
#endif

#ifndef MASSIV_NETOBUFFER_H
#include "net/netobuffer.h"
#endif



namespace Massiv {
namespace Core {

class Network;


/** @addtogroup net
 * @{
 */

/**
 * Implements a UDP buffer that can be sent via the UDP protocol by lower
 * network layers. The buffer can hold arbitrary number of messages that 
 * will be all send at a time. Number of messages is written in first 16
 * bits of the UDP buffer.
 *
 * Format of each message is: 
 * <ul>
 *  <li>tag of the message [32 bits] (determines a message type
 *      - several are reserved by the net layer, such as MESSAGE_PING, etc.).
 *  <li>length of the message [32 bits]
 *  <li>data [length specified by the length field]
 * </ul>
 */
class UDPOBuffer : public NetOBuffer
    {
    friend class Massiv::Core::Network;
    friend class Massiv::Core::Test::SocketsTest;

    typedef CRC32< BitIOStreamBuffer >  Crc32;

public:
    
    /**
     * Default constructor. Reserves default room in internal buffers.
     */
    UDPOBuffer( );
    
    /**
     * Copy constructor.
     */
    UDPOBuffer
        (
        const UDPOBuffer &      sourcebuf
        );

    /**
     * Default destructor.
     */
    virtual ~UDPOBuffer();
    
    /**
     */
    virtual std::string to_string() const;

private:

    /**
     * Initializes the buffer - i.e. reserves space for number of messages,
     * writes a first message tag and reserves space for the first message
     * length.
     * @param tag tag of the first message contained in the buffer.
     */
    virtual void init
        ( 
        UInt32 tag
        );

    /**
     * Writes message tag and reserves space for message length.
     * Should be called before appending new message to existing buffer.
     * msg_init of the first message is done automaticaly in the init()
     * method.
     * We should fit more messages into a simple UDP buffer because of
     * the efficiency. On the other hand, too much messages increases 
     * probability of receiving a message corrupted and the efficiency
     * can decrease.
     * @param tag The message tag.
     */
    void msg_init
        ( 
        UInt32 tag
        );
        
    /**
     * Finishes buffer - writes number of messages and encrypts it into 
     * crypted_buffer using the blowfish cipher from the appropriate 
     * connection_info class.
     */
    void flush();

    /**
     * Finishes message - writes the message size in the space reserved
     * before. 
     */
    void msg_flush();

public:

    /**
     * Computes a checksum of the buffer and appends it at the end.
     * Note that this method should be only used AFTER encrypting the
     * UDPOBuffer (because it operates on @a crypted_buffer).
     * @note This method is written to be used together with CRC32.
     *      For being used with another checksums, it might require
     *      to be rewrited.
     */
    inline void append_checksum();

    /**
     * Writes a given checksum to the buffer.
     * Note that this method should be only used AFTER encrypting the
     * UDPOBuffer (because it operates on @a crypted_buffer).
     * @note This method is written to be used together with CRC32.
     *      For being used with another checksums, it might require
     *      to be rewrited.
     */
    void append_checksum
        (
        Crc32::ChecksumType     checksum
        );

    /**
     * Returns number of messages contained in the buffer.
     * Used in network statistics.
     */
    int get_count_msgs() const;

private:
    
    Crc32                               engine_checksum;
        /**< The checksum "engine". */

    UInt32                              msg_start;
        /**< Offset of the beginning of the last message in the buffer -
             it is the place where the message length will be written after
             flushing the message. */

    UInt16                              count;
        /**< How many msgs is contained in the buffer. */

    std::list< UDPOBuffer * >::iterator self_iterator;
        /**< Iterator to buffer list in @c Network pointing to self. */

    };  // class UDPOBuffer


// UDPOBuffer inline methods implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

inline void UDPOBuffer::append_checksum()
    {
    append_checksum( engine_checksum.compute_checksum() );
    }

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_UDPOBUFFER_H
