/** @file node_action.h
 * Node actions.
 */

/*
 * begin        : Sat Oct 4 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_ACTION_H
#define MASSIV_NODE_ACTION_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Node actions is a way how to co-ordinate a sequence of commands
 * that must be executed on all server nodes.
 *
 * When dealing with a sequence of commands that must be executed
 * on all nodes of distributed system, there is always problem with
 * synchronization of the nodes. First problem is that two or more nodes
 * can initiate the sequence at same time and the system must properly
 * detect and solve the collision. Another problem is that often the
 * commands of the sequence must be executed one by one, each step
 * to be synchronized with other nodes.
 *
 * Node actions in massiv is a simple way how to execute a sequence of
 * command on all server nodes.
 *
 * Scenario of a node action:
 * First a node initiate the action. By initiating the action, system
 * contacts all other nodes and prepare all nodes for the action.
 * Then system selects coordinator of the action. Coordinator
 * is the node that will be sending commands to other nodes and will
 * coordinate the action. Nodes other that the coordinator are called
 * slave nodes.
 * If two nodes initiate the action at same time, system will detect
 * it and will properly select one coordinator from the nodes.
 * Once coordinator of the action is selected, it start executing
 * the action and sending comands to the nodes.
 * After all work is done, the coordinator notifies the system that
 * the action can be finished. The system contacts all nodes
 * and finish the action.
 * Once the action is finished, any node can initiate the action again.
 *
 * Class NodeAction itself is placeholder for node action related definitions
 * and base class for all node action objects.
 *
 * @note To define new action you must define new action id and implement
 *      node action object derived from NodeAction.
 */
class NodeAction
    {
    NO_COPY_CONSTRUCTOR( NodeAction );
    NO_ASSIGNMENT_OPERATOR( NodeAction );

public:
    
    /** Node action identification. */
    enum Id
        {
        NONE            = 0,
            /**< Not a valid action. */
        SERVER_QUIT     = 1,
            /**< Server quit action.
             * @see NodeActionServerQuit
             */
        STARTUP         = 2,
            /**< Action that manages startup sequence of server nodes.
             * @see NodeActionStarrtup
             */
        COUNT           = 3,
            /**< Number of actions. Not valid action. */
        }; // Id
    
    /** Next tick callback result. */
    enum NextTickResult
        {
        RESULT_CONTINUE,
            /**< Continue running the action. */
        RESULT_FINISH
            /**< Finish the action. */
        };
    
public:

    /**
     * Constructor.
     * Node action object is created if local node is selecte as coordinator
     * of the action.
     */
    NodeAction();
    
    /**
     * Destructor.
     * Node action object is destroyed iff its @c next_tick() callback
     * returns @c RESULT_FINISH or system is shutting down.
     */
    virtual ~NodeAction();
    
public:

    /**
     * Factory for creating node action objects.
     * @pre action_id is not @c NONE.
     * @pre action_id is not @c COUNT.
     */
    static NodeAction * create
        (
        Id action_id
        );

public:

    //
    // Callbacks.
    //

    /**
     * Action's next tick.
     * Periodically called by system.
     * @pre Local node is coordinator of the action.
     */
    virtual NextTickResult next_tick() = 0;

    }; // NodeAction

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_ACTION_H
