/** @file node_action_manager.cpp
 * Node action manager (implementation).
 */

/*
 * begin        : Sat Oct 4 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_action_manager.h"
#include "node/node_action_id.h"
#include "node/node_action_state.h"
#include "node/node_manager.h"
#include "utility/hash.h"
#include "database/node_database.h"
#include "global/global.h"
#include "sysevent/system_event_queue.h"
#include "sysevent/node_action_msg.h"
#include "system/system.h"
#include "utility/debug.h"
#include <memory>
#include <sstream>


namespace Massiv {
namespace Core {


// Node action manager internal structure.
class NodeActionManager::My
    {
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );
    
public:

    My() :
        num_servers( 0 )
        {
        int i;
        for( i = 0; i < NodeAction::COUNT; ++i )
            {
            node_action_counters[ i ] = 0;
            }
        }
    
    ~My()
        {
        }

public:

    void initialize()
        {
        // Determine number of server nodes.
        class NodeEnumerator : public NodeDatabase::Enumerator
            {
        public:
        
            NodeEnumerator() :
                num_nodes( 0 )
                {
                }
        
            virtual void operator()
                (
                const NodeDatabase::NodeData & data
                )
                {
                num_nodes++;
                }
            
            int num_nodes;
            
            }; // NodeEnumerator
        
        // Cleanup if reinitializing.
        node_action_states.clear();
        num_servers = 0;
        
        // Initialize.
        NodeDatabase & node_database = Global::node_database();
        NodeEnumerator enumerator;
        node_database.enumerate_nodes( NodeId::TYPE_SERVER, enumerator );
        num_servers = enumerator.num_nodes;
        }

    void next_tick()
        {
        NodeActionStates::iterator it;
        
        // Update each action.
        for( it = node_action_states.begin(); it != node_action_states.end(); ++it )
            {
            const NodeActionId & action_id = it->first;
            NodeActionState & action_state = it->second;
            
            // Update the action.
            update_action( action_id, action_state );
            }
        }

public:

    bool is_node_action_registered( const NodeActionId & action_id ) const
        {
        return node_action_states.find( action_id ) != node_action_states.end();
        }
    
    void register_node_action( const NodeActionId & action_id )
        {
        hope( !is_node_action_registered( action_id ) );
        NodeActionState state;
        node_action_states.insert( NodeActionStates::value_type( action_id, state ) );
        }
    
    NodeActionState & get_node_action_state( const NodeActionId & action_id )
        {
        NodeActionStates::iterator it = node_action_states.find( action_id );
        hope( it != node_action_states.end() );
        return it->second;
        }
    
    void unregister_node_action( const NodeActionId & action_id )
        {
        NodeActionStates::iterator it = node_action_states.find( action_id );
        hope( it != node_action_states.end() );
        node_action_states.erase( it );
        }
        
public:

    void pause_node_action_request_messages
        (
        const NodeId & node
        )
        {
        // Log.
            {
            std::ostringstream oss;
            oss << "Node action manager: Pausing node action request messages from node " << node << ".";
            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
            }
        
        unsigned int pause_flags = SystemMessage::PAUSE_NODE_ACTION_REQUEST;
        SystemEventQueue & system_event_queue = Global::system_event_queue();
        unsigned int pause_mask = system_event_queue.get_server_pause_mask( node );
        hope( !flags_on( pause_mask, pause_flags ) );
        pause_mask |= pause_flags;
        system_event_queue.set_server_pause_mask( node, pause_mask );
        }
    
    void unpause_node_action_request_messages()
        {
        class NodeEnumerator : public NodeDatabase::Enumerator
            {
        public:
        
            NodeEnumerator()
                {
                }
        
            virtual void operator()
                (
                const NodeDatabase::NodeData & data
                )
                {
                    {
                    std::ostringstream oss;
                    //oss << "Node action manager: Unpausing node action request messages from node " << data.get_node_id() << ".";
                    Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                    }
                
                const NodeId & node = data.get_node_id();
                unsigned int bits_to_disable = SystemMessage::PAUSE_NODE_ACTION_REQUEST;
                SystemEventQueue & system_event_queue = Global::system_event_queue();
                unsigned int pause_mask = system_event_queue.get_server_pause_mask( node );
                pause_mask &= ~( bits_to_disable );
                system_event_queue.set_server_pause_mask( node, pause_mask );
                }
            
            }; // ServerEnumerator
        
        NodeDatabase & node_database = Global::node_database();
        NodeEnumerator enumerator;
        node_database.enumerate_nodes( NodeId::TYPE_SERVER, enumerator );
        }
    
    void multicast_message_to_server_nodes( const SystemMessage & message, bool send_to_local_node  )
        {
        NodeManager & node_manager = Global::node_manager();
        node_manager.multicast_message_to_server_nodes( message, NetworkCommon::STREAM_RELIABLE, !send_to_local_node );
        }
        
private:

    void update_action( const NodeActionId & action_id, NodeActionState & action_state )
        {
        // Based on the action phase select what to do.
        switch( action_state.phase )
            {
            case NodeActionState::PHASE_NONE:
                break;
            
            case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
                {
                // Check if we have received replies from all nodes.
                hope( action_state.num_replies <= action_state.num_awaiting_replies );
                if( action_state.num_replies >= action_state.num_awaiting_replies )
                    {
                        {
                        std::ostringstream oss;
                        oss << "Node action " << action_id.get() << ":"
                            << " Local node is coordinator of the action."
                            << " Sending coordinator selected message to all other server nodes.";
                        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    // All replies received. Local node is coordinator of the action.
                    // Switch phase.
                    action_state.phase = NodeActionState::PHASE_COORDINATOR_RUNNING;
                    
                    // Notify all nodes that we are the action coordinator.
                    // We do not send the message to local node.
                    NodeActionCoordinatorSelectedMessage coordinator_message( action_id );
                    multicast_message_to_server_nodes( coordinator_message, false );
                        
                    // Because the message is not send to local node, we must increment
                    // outr action counter here.
                    node_action_counters[ action_id.get() ]++;
                    
                    // Begin action.
                    hope( !action_state.node_action_object );
                    action_state.node_action_object = NodeAction::create( action_id.get() );
                    }
                break;
                }
                
            case NodeActionState::PHASE_COORDINATOR_RUNNING:
                {
                NodeAction::NextTickResult result;
                hope( action_state.node_action_object != NULL );
                
                // Call action's callback.
                result = action_state.node_action_object->next_tick();
                
                switch( result )
                    {
                    case NodeAction::RESULT_CONTINUE:
                        break;
                        
                    case NodeAction::RESULT_FINISH:
                        {
                            {
                            std::ostringstream oss;
                            oss << "Node action " << action_id.get() << ":"
                                << " Coordinator is finishing the action."
                                << " Sending prepare finish message to all other server nodes.";
                            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                            }
                        
                        // Finish the action.
                        action_state.phase = NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY;
                        action_state.num_replies = 1; // Count local node.
                        
                        // Send prepare for finish message to all nodes.
                        // We do not send he message to local node.
                        NodeActionPrepareFinishMessage prepare_finish_message( action_id );
                        multicast_message_to_server_nodes( prepare_finish_message, false );
                        
                        // Finish action.
                        delete action_state.node_action_object;
                        action_state.node_action_object = NULL;
                        break;
                        }
                    }
                break;
                }
                
            case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
                {
                // Check if we have received replies from all nodes.
                hope( action_state.num_replies <= action_state.num_awaiting_replies );
                if( action_state.num_replies >= action_state.num_awaiting_replies )
                    {
                        {
                        std::ostringstream oss;
                        oss << "Node action " << action_id.get() << ":"
                            << " Prepare finish replies received from all server nodes."
                            << " Sending finish message to all server nodes.";
                        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    // All replies received. Switch self to slave mode because coordinator
                    // is not needed any more.
                    action_state.phase = NodeActionState::PHASE_WAITING_FOR_FINISH;
                    
                    // Send finish order to all nodes (including local node).
                    NodeActionFinishMessage finish_message( action_id );
                    multicast_message_to_server_nodes( finish_message, true );
                    }
                break;
                }
        
            case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
                // Coordinator not yet selected.
                break;
            
            case NodeActionState::PHASE_RUNNING:
            case NodeActionState::PHASE_WAITING_FOR_FINISH:
                // We are waiting for coordinator's orders.
                break;
            }
        }

private:

    typedef ext::hash_map
        <
        NodeActionId,
        NodeActionState,
        NodeActionId::Hash,
        NodeActionId::EqualKey
        > NodeActionStates;
    
private:

    NodeActionStates    node_action_states;
        /**< List of node actions in progress. */

public:

    int                 num_servers;
        /**< Number of server nodes. */
    UInt32              node_action_counters[ NodeAction::COUNT ];
        /**< Global counter for each node action. When sending node action request,
         * corresponding counter is included in the request message.
         * Once an action is finished, its global counter is incremented.
         * The counters are used to detect obsolete requests.
         */
    
    }; // NodeActionManager::My

    
//------------------------------------------------------------------------------


NodeActionManager::NodeActionManager() :
    my( NULL )
    {
    std::auto_ptr< My > my_ptr;
    
    my_ptr.reset( new My() );
    
    my = my_ptr.release();
    }


NodeActionManager::~NodeActionManager()
    {
    delete my;
    my = NULL;
    }


void NodeActionManager::initialize()
    {
    my->initialize();
    }


bool NodeActionManager::is_ready() const
    {
    return ( my->num_servers > 0 );
    }


void NodeActionManager::initiate_action
    (
    NodeAction::Id the_action_id
    )
    {
    hope( the_action_id != NodeAction::NONE );
    hope( the_action_id != NodeAction::COUNT );
    
    NodeActionId action_id( the_action_id );
    
    // Node action manager must be initialized.
    hope( is_ready() );
    
    // Currently actions can be initiated only on server nodes.
    hope( System::is_server() );
    
    // Log.
        {
        std::ostringstream oss;
        oss << "Initiating node action " << the_action_id << ".";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        }
    
    // Check if the action is already in progress...
    if( my->is_node_action_registered( action_id ) )
        {
        std::ostringstream oss;
        oss << "Node action " << the_action_id
            << " is already in progress. NodeActionManager::initiate_action() does nothing.";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        return;
        }
    
    // Begin the action.
    // Communication protocol requires action state to be registered now.
    my->register_node_action( action_id );
    NodeActionState & action_state = my->get_node_action_state( action_id );
    hope( action_state.phase == NodeActionState::PHASE_NONE );
    action_state.phase = NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY;
    action_state.num_replies = 1; // Count local node.
    action_state.num_awaiting_replies = my->num_servers;
    
    // Log.
        {
        std::ostringstream oss;
        oss << "Node action " << the_action_id << ":"
            << " Sending node action request messages to all other server nodes.";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        }
    
    // Send node action request to all server nodes.
    // We do not send the message to ourself.
    NodeActionRequestMessage request_message( action_id, my->node_action_counters[ the_action_id ] );
    my->multicast_message_to_server_nodes( request_message, false );
    }


NodeAction * NodeActionManager::get_node_action_object
    (
    NodeAction::Id the_action_id
    )
    {
    const NodeActionId action_id( the_action_id );
    
    if( !my->is_node_action_registered( action_id ) )
        {
        return NULL;
        }
    
    NodeActionState & action_state = my->get_node_action_state( action_id );
    return action_state.node_action_object; // Can be null.
    }


void NodeActionManager::next_tick()
    {
    my->next_tick();
    }


//------------------------------------------------------------------------------

// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_request_message
    (
    const NodeActionRequestMessage & request_message
    )
    {
    hope( request_message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( request_message.get_initiator() == request_message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = request_message.action_id;
    const NodeId & request_message_sender = request_message.get_sender();
    bool send_reply = false;
    SystemMessage::ExecuteResult result = SystemMessage::EXECUTE_OK;
    
    hope( action_id.get() >= 0 && action_id.get() < NodeAction::COUNT );
    
    if( my->node_action_counters[ action_id.get() ] != request_message.action_counter )
        {
        // Obsolete message. Ignore the message.
        std::ostringstream oss;
        oss << "Received obsolete node action " << action_id.get() << " request. Ignored.";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        return result;
        }
    
    // Register the action if not yet registered.
    if( !my->is_node_action_registered( action_id ) )
        {
        // First action request from a node. The node should not be local node because
        // if local node is action initiator, the action should be already registered.
        hope( request_message_sender != System::get_local_node_id() );
        my->register_node_action( action_id );
        NodeActionState & action_state = my->get_node_action_state( action_id );
        hope( action_state.phase == NodeActionState::PHASE_NONE );
        action_state.phase = NodeActionState::PHASE_WAITING_FOR_COORDINATOR;
        
        // Send reply to the node.
        send_reply = true;
        }
    else
        {
        // The requested action was already registered. There are several
        // possibilities:
        // 1/ Two or more nodes are initiators of the action (at same time).
        // 2/ Action coordinator of previous action has finished the previous
        //    action (but local node don't know that yet ) and another node
        //    initiated the action again.
        // 3/ This is obsolete request, because the action was started already
        //    some time ago but but the request was delayed on some connection.
        // Point 3/ is checked checked.
    
        // Get state of the action.
        NodeActionState & action_state = my->get_node_action_state( action_id );
    
        // Based on the action phase select what to do.
        switch( action_state.phase )
            {
            case NodeActionState::PHASE_NONE:
                panic( "Internal error. Unexpected node action phase." );
                break;
            
            case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
                {
                // Local node is initiator of the action. If request_message is
                // not for local node, two nodes at one time have initiated
                // the action and we must select which one will be coordinator.
                // Local node can't be message sender because initiator of the action
                // do not send request message to itself.
                hope( request_message_sender != System::get_local_node_id() );
                
                // Local node and another node have initiated the action at one time.
                // We must select which one will be coordinator of the action.
                if( NodeId::Less()( System::get_local_node_id(), request_message_sender ) )
                    {
                        {
                        std::ostringstream oss;
                        oss << "Node action " << action_id.get() << ":"
                            << " Action request collision with " << request_message_sender << "."
                            << " Local node won't be coordinator of the action.";
                        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    // The sender will be coordinator. Fix action state.
                    action_state.phase = NodeActionState::PHASE_WAITING_FOR_COORDINATOR;
                        
                    // Send reply to new coordinator.
                    send_reply = true;
                    }
                else
                    {
                    
                        {
                        std::ostringstream oss;
                        oss << "Node action " << action_id.get() << ":"
                            << " Action request collision with " << request_message_sender << "."
                            << " The " << request_message_sender << " won't be coordinator of the action.";
                        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    // Local node will be coordinator.
                    // Do not send reply to the node. The node will soon detect
                    // that we will be coordinator.
                    send_reply = false;
                    }
                break;
                } // case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY
                
            case NodeActionState::PHASE_COORDINATOR_RUNNING:
                panic( "Internal communication error. Received unexpected action request." );
                break;
                
            case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
                panic( "Internal communication error. Received unexpected action request." );
                break;
                
            case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
                {
                // It seems that two or more nodes have inititated the action at one time.
                // Just reply and leave the nodes to select coordinator.
                send_reply = true;
                break;
                }
                
            case NodeActionState::PHASE_RUNNING:
                panic( "Internal communication error. Received unexpected action request." );
                break;
                
            case NodeActionState::PHASE_WAITING_FOR_FINISH:
                {
                // We are waiting for finish message from coordinator but
                // another node already received finish message and now the node
                // is initiating the action again.
                // We must pause the message until we receive finish message
                // from old coordinator.
                my->pause_node_action_request_messages( request_message_sender );
                result = SystemMessage::EXECUTE_KEEP_MESSAGE_IN_QUEUE;
                break;
                }
            } // switch( action_state.phase )
        }
        
    // Send reply.
    if( send_reply )
        {
        NodeActionRequestReplyMessage reply_message( action_id, my->node_action_counters[ action_id.get() ] );
        reply_message.send( request_message_sender, NetworkCommon::STREAM_RELIABLE );
        }
    
    return result;
    }


// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_request_reply_message
    (
    const NodeActionRequestReplyMessage & message
    )
    {
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( message.get_initiator() == message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = message.action_id;

    hope( action_id.get() >= 0 && action_id.get() < NodeAction::COUNT );
    
    if( my->node_action_counters[ action_id.get() ] != message.action_counter )
        {
        // Obsolete message. Ignore the message.
        std::ostringstream oss;
        oss << "Received obsolete node action " << action_id.get() << " request reply. Ignored.";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        return SystemMessage::EXECUTE_OK;
        }
        
    // Verify action id.
    if( !my->is_node_action_registered( action_id ) )
        {
        panic( "Communication protocol error. Unexpected node action." );
        }
    
    // Get state of the action.
    NodeActionState & action_state = my->get_node_action_state( action_id );
    
    // Based on the action phase select what to do.
    switch( action_state.phase )
        {
        case NodeActionState::PHASE_NONE:
            panic( "Internal error. Unexpected node action phase." );
            break;
            
        case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
            // Local node is initiator of the actions.
            action_state.num_replies++;
            
                {
                std::ostringstream oss;
                oss << "Node action " << action_id.get() << ":"
                    << " Received request reply. "
                    << ( action_state.num_awaiting_replies - action_state.num_replies )
                    << " more to go.";
                Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            break;
        
        case NodeActionState::PHASE_COORDINATOR_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
        case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
        case NodeActionState::PHASE_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_FINISH:
            // Local node was initiator of the action some time ago
            // but it failed to be coordinator of the action.
            // We have received obsolete reply message. Ignore it.
            break;
        };
    
    return SystemMessage::EXECUTE_OK;
    }


// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_coordinator_selected_message
    (
    const NodeActionCoordinatorSelectedMessage & message
    )
    {
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( message.get_initiator() == message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = message.action_id;
    
    // Verify action id.
    if( !my->is_node_action_registered( action_id ) )
        {
        panic( "Communication protocol error. Unexpected node action." );
        }
    
    // Get state of the action.
    NodeActionState & action_state = my->get_node_action_state( action_id );
    
    // Based on the action phase select what to do.
    // Node: Coordinator doesn't send coordinator selected message to itself.
    switch( action_state.phase )
        {
        case NodeActionState::PHASE_NONE:
            panic( "Internal error. Unexpected node action phase." );
            break;
            
        case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
            action_state.phase = NodeActionState::PHASE_RUNNING;
            
                {
                std::ostringstream oss;
                oss << "Node action " << action_id.get() << ":"
                    << " Received coordinator selected message from " << message.get_sender() << ".";
                Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            // Increment action counter.
            my->node_action_counters[ action_id.get() ]++;
            break;
        
        case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
        case NodeActionState::PHASE_COORDINATOR_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
        case NodeActionState::PHASE_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_FINISH:
            panic( "Communication protocol error. Unexpected node action phase." );
            break;
        };
    
    return SystemMessage::EXECUTE_OK;
    }


// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_prepare_finish_message
    (
    const NodeActionPrepareFinishMessage & message
    )
    {
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( message.get_initiator() == message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = message.action_id;
    const NodeId & message_sender = message.get_sender();
    
    // Verify action id.
    if( !my->is_node_action_registered( action_id ) )
        {
        panic( "Communication protocol error. Unexpected node action." );
        }
    
    // Get state of the action.
    NodeActionState & action_state = my->get_node_action_state( action_id );
    
    // Based on the action phase select what to do.
    // Node: Coordinator doesn't send prepare finish message to itself.
    switch( action_state.phase )
        {
        case NodeActionState::PHASE_NONE:
            panic( "Internal error. Unexpected node action phase." );
            break;
            
        case NodeActionState::PHASE_RUNNING:
            {
            // Switch phase.
            action_state.phase = NodeActionState::PHASE_WAITING_FOR_FINISH;
            
            // Send reply to coordinator that we are ready to finish.
            NodeActionPrepareFinishReplyMessage reply_message( action_id );
            reply_message.send( message_sender, NetworkCommon::STREAM_RELIABLE );
            break;
            }
        
        case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
        case NodeActionState::PHASE_COORDINATOR_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
        case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
        case NodeActionState::PHASE_WAITING_FOR_FINISH:
            panic( "Communication protocol error. Unexpected node action phase." );
            break;
        };
    
    return SystemMessage::EXECUTE_OK;
    }


// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_prepare_finish_reply_message
    (
    const NodeActionPrepareFinishReplyMessage & message
    )
    {
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( message.get_initiator() == message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = message.action_id;
    
    // Verify action id.
    if( !my->is_node_action_registered( action_id ) )
        {
        panic( "Communication protocol error. Unexpected node action." );
        }
    
    // Get state of the action.
    NodeActionState & action_state = my->get_node_action_state( action_id );
    
    // Based on the action phase select what to do.
    switch( action_state.phase )
        {
        case NodeActionState::PHASE_NONE:
            panic( "Internal error. Unexpected node action phase." );
            break;

        case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
            action_state.num_replies++;
            break;

        case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
        case NodeActionState::PHASE_COORDINATOR_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
        case NodeActionState::PHASE_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_FINISH:
            panic( "Communication protocol error. Unexpected node action phase." );
            break;
        };
    
    return SystemMessage::EXECUTE_OK;
    }


// Done.
SystemMessage::ExecuteResult NodeActionManager::handle_node_action_finish_message
    (
    const NodeActionFinishMessage & message
    )
    {
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    hope( message.get_initiator() == message.get_sender() ); // Already verified.
    
    const NodeActionId action_id = message.action_id;
    
    // Verify action id.
    if( !my->is_node_action_registered( action_id ) )
        {
        panic( "Communication protocol error. Unexpected node action." );
        }
    
    // Get state of the action.
    NodeActionState & action_state = my->get_node_action_state( action_id );
    
    // Based on the action phase select what to do.
    switch( action_state.phase )
        {
        case NodeActionState::PHASE_NONE:
            panic( "Internal error. Unexpected node action phase." );
            break;
            
        case NodeActionState::PHASE_WAITING_FOR_FINISH:
            {
            // Remove the action.
            my->unregister_node_action( action_id );
            
            // Unpause waiting request messages (if any).
            my->unpause_node_action_request_messages();
            break;
            }

        case NodeActionState::PHASE_WAITING_FOR_REQUEST_REPLY:
        case NodeActionState::PHASE_COORDINATOR_RUNNING:
        case NodeActionState::PHASE_WAITING_FOR_PREPARE_FINISH_REPLY:
        case NodeActionState::PHASE_WAITING_FOR_COORDINATOR:
        case NodeActionState::PHASE_RUNNING:
            panic( "Communication protocol error. Unexpected node action phase." );
            break;
        };
    
    return SystemMessage::EXECUTE_OK;
    }


} // namespace Massiv::Core
} // namespace Massiv
