/** @file node_action_manager.h
 * Node action manager.
 */

/*
 * begin        : Sat Oct 4 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_ACTION_MANAGER_H
#define MASSIV_NODE_ACTION_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_NODE_ACTION_H
#include "node/node_action.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Node action manager manages running node actions and coordinate
 * the actions with other nodes.
 * See #NodeAction for more info.
 */
class MASSIV_EXPORTED NodeActionManager
    {
    NO_COPY_CONSTRUCTOR( NodeActionManager );
    NO_ASSIGNMENT_OPERATOR( NodeActionManager );

public:

    /** Constructor. */
    NodeActionManager();
    
    /** Destructor. */
    ~NodeActionManager();
    
    /** Initialize. */
    void initialize();
    
    /** Check whether the manager is ready to initiate node actions. */
    bool is_ready() const;
    
public:
    
    //
    // Actions.
    //
    
    /**
     * Inititate node action.
     * @pre Node action manager must be ready. See @c is_ready().
     * @pre Local node is server node.
     */
    void initiate_action
        (
        NodeAction::Id action_id
        );

    /**
     * Get node action object for specified action.
     * If specified action is not in progress or coordinator of the
     * action was not yet selected or local node
     * is not coordinator of the action, NULL is returned.
     */
    NodeAction * get_node_action_object
        (
        NodeAction::Id action_id
        );

public:
    
    /**
     * Node action manager's next tick.
     */
    void next_tick();
    
public:

    //
    // Message callbacks.
    //

    /**
     * Handle #NodeActionRequestMessage.
     * This is common callback for all ancestors of #NodeActionRequestMessage
     * and each of the ancestors must call this callback when executed.
     */
    SystemMessage::ExecuteResult handle_node_action_request_message
        (
        const NodeActionRequestMessage & message
        );
    
    /**
     * Handle #NodeActionRequestReplyMessage.
     */
    SystemMessage::ExecuteResult handle_node_action_request_reply_message
        (
        const NodeActionRequestReplyMessage & message
        );
    
    /**
     * Handle #NodeActionCoordinatorSelectedMessage.
     */
    SystemMessage::ExecuteResult handle_node_action_coordinator_selected_message
        (
        const NodeActionCoordinatorSelectedMessage & message
        );

    /**
     * Handle #NodeActionPrepareFinishMessage.
     */
    SystemMessage::ExecuteResult handle_node_action_prepare_finish_message
        (
        const NodeActionPrepareFinishMessage & message
        );
        
    /**
     * Handle #NodeActionPrepareFinishReplyMessage.
     */
    SystemMessage::ExecuteResult handle_node_action_prepare_finish_reply_message
        (
        const NodeActionPrepareFinishReplyMessage & message
        );

    /**
     * Handle #NodeActionFinishMessage.
     */
    SystemMessage::ExecuteResult handle_node_action_finish_message
        (
        const NodeActionFinishMessage & message
        );
        
private:

    /** Node action manager internal structure. */
    class My;
    
    My * my;
        /**< Internal structure. Always valid pointer. */
    
    }; // class NodeActionManager

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_ACTION_MANAGER_H
