/** @file node_action_startup.cpp
 * STARTUP node action object (implementation).
 */

/*
 * begin        : Mon Nov 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_action_startup.h"
#include "object/object_manager.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "node/node_action_manager.h"
#include "node/node_manager.h"
#include "time/time_manager.h"
#include "database/archive_database.h"
#include "utility/debug.h"


namespace Massiv {
namespace Core {


namespace {


void get_last_valid_archivation_id
    (
    StartupMessage & msg
    )
    {
    msg.archivation_id = Global::archive_database().get_latest_valid_archive();
    }


void enumerate_providers
    (
    StartupMessage & msg
    )
    {
    Global::object_manager().enumerate_local_providers( msg );
    }


void load_simulation
    (
    ArchivationId archivation_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::[anonymous]::load_simulation()" );
    
    if( archivation_id == ArchivationId() )
        {
        // Should never happed - verified elsewhere.
        panic( "Can't load simulation because of null archivation id." );
        }
    
    // Load simulation to memory.
    GlobalCallbacks::load_simulation( archivation_id );
    
    // Synchronize time on all server nodes.
    // FIXME: Is load_simulation() the right place for this?
        {
        typedef NodeManager::NodeIdList NodeIdList;
        TimeManager & time_manager = Global::time_manager();
        NodeIdList servers;
        NodeIdList::const_iterator it;
        int count = 0;
        
        Global::node_manager().enumerate_server_nodes( servers );
            
        // Register all servers to time manager.
        for( it = servers.begin(); it != servers.end(); ++it )
            {
            const NodeId & node_id = *it;
            if( node_id != Global::get_local_node_id() )
                {
                time_manager.add_node_for_synchronization( node_id );                    
                count++;
                }
            }
            
        // Log.
            {
            std::ostringstream oss;
            oss << "Enabling allow_synchronization with " << count
                << " server nodes.";
            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );

            time_manager.allow_synchronization();
            }
        }
    }


void unpause_simulation()
    {
    TimeManager & time_manager = Global::time_manager();
    time_manager.unpause();
    }


} // anonymous namespace


//------------------------------------------------------------------------------
// NodeActionStartup.
//------------------------------------------------------------------------------


NodeActionStartup::NodeActionStartup() :
    phase( PHASE_NONE )
    {
    }

NodeActionStartup::~NodeActionStartup()
    {
    }

NodeAction::NextTickResult NodeActionStartup::next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeActionStartup::next_tick()" );

    switch( phase )
        {
        case PHASE_NONE:
            {
            // Enumerate server nodes.
            NodeManager::NodeIdList nodes;
            Global::node_manager().enumerate_server_nodes( nodes );
            
            NodeManager::NodeIdList::const_iterator it;
            for( it = nodes.begin(); it != nodes.end(); ++it )
                {
                const NodeId & node_id = *it;
                servers.insert( Nodes::value_type( node_id, NodeState() ) );
                }
            
            // Ask servers what archive files they have.
            send_archivation_state_request();
            phase = PHASE_AWAITING_ARCHIVATION_STATE_REPLIES;
            log_phase();
            break;
            }
        
        case PHASE_AWAITING_ARCHIVATION_STATE_REPLIES:
            {
            if( archivation_state_replies_received_from_all_servers() )
                {
                // Tell servers to load simulation from selected archive file.
                send_load_simulation_request();
                phase = PHASE_AWAITING_LOAD_SIMULATION_REPLIES;
                log_phase();
                }
            break;
            }
        
        case PHASE_AWAITING_LOAD_SIMULATION_REPLIES:
            {
            if( load_archivation_replies_received_from_all_servers() )
                {
                // Tell servers to unpause simulation.
                register_remote_providers();
                send_unpause_simulation_request();
                phase = PHASE_SIMULATION_RUNNING;
                log_phase();
                }
            break;
            }
        
        case PHASE_SIMULATION_RUNNING:
            // Simulation is running. Nothing to do.
            break;
        };
    
    // Note: We never return RESULT_FINISH because once startup sequence
    // is done and simulation is running, it doesn't make sense to
    // be able to initiate startup sequence again. Servers must be
    // restarted to restart simulation.
    return RESULT_CONTINUE;
    }


bool NodeActionStartup::archivation_state_replies_received_from_all_servers() const
    {
    Nodes::const_iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeState & state = it->second;
        if( !state.received_archivation_state_reply )
            {
            return false;
            }
        }
    return true;
    }


bool NodeActionStartup::load_archivation_replies_received_from_all_servers() const
    {
    Nodes::const_iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeState & state = it->second;
        if( !state.received_load_simulation_reply )
            {
            return false;
            }
        }
    return true;
    }


void NodeActionStartup::send_archivation_state_request()
    {
    StartupMessage  msg( StartupMessage::ARCHIVATION_STATE_REQUEST );
    msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
    }


void NodeActionStartup::send_load_simulation_request()
    {
    // Select the lowest archivation id that is available on each server.
    hope( !servers.empty() );
    ArchivationId last_valid_archivation_id = servers.begin()->second.last_valid_archivation_id;
    
        {
        Nodes::const_iterator it;
        for( it = servers.begin(); it != servers.end(); ++it )
            {
            const NodeState & state = it->second;
            if( state.last_valid_archivation_id < last_valid_archivation_id )
                {
                last_valid_archivation_id = state.last_valid_archivation_id;
                }
            }
        }
    
    if( last_valid_archivation_id == ArchivationId() )
        {
        panic( "At least one server doesn't have any valid archive." );
        }
    
    StartupMessage msg( StartupMessage::LOAD_SIMULATION_REQUEST );
    msg.archivation_id = last_valid_archivation_id;
    msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
    }


void NodeActionStartup::send_unpause_simulation_request()
    {
    StartupMessage msg( StartupMessage::UNPAUSE_SIMULATION_REQUEST );
    msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
    }


void NodeActionStartup::register_remote_providers()
    {
    typedef NodeState::Providers Providers;
    Nodes::const_iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        const NodeState & state = it->second;
        Providers::const_iterator pit;
        
        // Do not register providers that are owned by local node.
        if( node_id == Global::get_local_node_id() )
            {
            continue;
            }
        
        for( pit = state.providers.begin(); pit != state.providers.end(); ++pit )
            {
            ProviderId provider_id = *pit;
            Global::object_manager().register_remote_provider( provider_id, node_id );
            }
        }
    }


void NodeActionStartup::log_phase()
    {
    std::string phase_name;
    
    switch( phase )
        {
        case PHASE_NONE:
            break;
        case PHASE_AWAITING_ARCHIVATION_STATE_REPLIES:
            phase_name = "AWAITING_ARCHIVATION_STATE_REPLIES";
            break;
        case PHASE_AWAITING_LOAD_SIMULATION_REPLIES:
            phase_name = "AWAITING_LOAD_SIMULATION_REPLIES";
            break;
        case PHASE_SIMULATION_RUNNING:
            phase_name = "SIMULATION_RUNNING";
            break;
        }
    
    std::ostringstream oss;
    oss << "Simulation startup coordinator phase is " << phase_name << ".";
    Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
    }


NodeActionStartup::NodeState & NodeActionStartup::get_node_state
    (
    const NodeId & node_id
    )
    {
    Nodes::iterator it = servers.find( node_id );
    hope( it != servers.end() );
    return it->second;
    }


//------------------------------------------------------------------------------
// Message handling.
//------------------------------------------------------------------------------


void NodeActionStartup::handle_archivation_state_reply_message
    (
    const StartupMessage & msg
    )
    {
    hope( msg.id.read() == StartupMessage::ARCHIVATION_STATE_REPLY );
    hope( phase == PHASE_AWAITING_ARCHIVATION_STATE_REPLIES );
    
    NodeState & node_state = get_node_state( msg.get_sender() );
    hope( !node_state.received_archivation_state_reply );
    node_state.received_archivation_state_reply = true;
    node_state.last_valid_archivation_id = msg.archivation_id;
    }


void NodeActionStartup::handle_load_simulation_reply
    (
    const StartupMessage & msg
    )
    {
    hope( msg.id.read() == StartupMessage::LOAD_SIMULATION_REPLY );
    hope( phase == PHASE_AWAITING_LOAD_SIMULATION_REPLIES );
    
    NodeState & node_state = get_node_state( msg.get_sender() );
    hope( !node_state.received_load_simulation_reply );
    node_state.received_load_simulation_reply = true;
    node_state.providers = msg.providers;
    }


void NodeActionStartup::handle_startup_message
    (
    const StartupMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeActionStartup::handle_startup_message()" );

    hope( msg.get_sender() == msg.get_initiator() ); // Already verified.
    hope( msg.get_sender().is_type( NodeId::TYPE_SERVER ) );
    const NodeId msg_sender = msg.get_sender();
    bool handled = false;
    
    switch( msg.id.read() )
        {
        case StartupMessage::ARCHIVATION_STATE_REQUEST:
            {
                {
                std::ostringstream oss;
                oss << "Sending ARCHIVATION_STATE_REPLY startup message to " << msg_sender << ".";
                Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            StartupMessage reply_msg( StartupMessage::ARCHIVATION_STATE_REPLY );
            get_last_valid_archivation_id( reply_msg );
            reply_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
            handled = true;
            break;
            }
            
        case StartupMessage::ARCHIVATION_STATE_REPLY:
            {
            NodeActionStartup & action_object = get_action_object();
            action_object.handle_archivation_state_reply_message( msg );
            handled = true;
            break;
            }
            
        case StartupMessage::LOAD_SIMULATION_REQUEST:
            {
                {
                std::ostringstream oss;
                oss << "Sending LOAD_SIMULATION_REPLY startup message to " << msg_sender << ".";
                Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            ArchivationId archivation_id = msg.archivation_id;
            StartupMessage reply_msg( StartupMessage::LOAD_SIMULATION_REPLY );
            load_simulation( archivation_id );
            enumerate_providers( reply_msg );
            reply_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
            handled = true;
            break;
            }
            
        case StartupMessage::LOAD_SIMULATION_REPLY:
            {
            NodeActionStartup & action_object = get_action_object();
            action_object.handle_load_simulation_reply( msg );
            handled = true;
            break;
            }
        
        case StartupMessage::UNPAUSE_SIMULATION_REQUEST:
            {
            unpause_simulation();
            handled = true;
            
            // Notify node manager that simulation is running.
            Global::node_manager().notify_simulation_is_running();
            break;
            }
        }
    
    if( !handled )
        {
        panic( "Invalid value od StartupMessage::id." );
        }
    }


NodeActionStartup & NodeActionStartup::get_action_object()
    {
    NodeActionManager & node_action_manager = Global::node_action_manager();
    NodeAction * action_object = node_action_manager.get_node_action_object( NodeAction::STARTUP );
    hope( action_object );
    NodeActionStartup * action_startup_object = checked_cast< NodeActionStartup * >( action_object );
    hope( action_startup_object );
    return *action_startup_object;
    }


} // namespace Core
} // namespace Massiv
