/** @file node_action_startup.h
 * STARTUP node action object.
 */

/*
 * begin        : Mon Nov 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_ACTION_STARTUP_H
#define MASSIV_NODE_ACTION_STARTUP_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODE_ACTION_H
#include "node/node_action.h"
#endif

#ifndef MASSIV_SIMULATION_SYNC_MSG_H
#include "sysevent/simulation_sync_msg.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Startup node action object.
 * Manages startup sequence of server nodes.
 *
 * NodeActionStartup object is created on the node that is
 * selected as coordinator of the action.
 * Once the object is created, it coordinates all server nodes
 * to load simulation state from an archive file and to unpause
 * the simulation.
 */
class NodeActionStartup : public NodeAction
    {
    NO_COPY_CONSTRUCTOR( NodeActionStartup );
    NO_ASSIGNMENT_OPERATOR( NodeActionStartup );
    
public:

    /** Startup phase. */
    enum Phase
        {
        PHASE_NONE,
            /**< Just created. */
        PHASE_AWAITING_ARCHIVATION_STATE_REPLIES,
            /**< Awaiting reply message to archivation state request from all server nodes. */
        PHASE_AWAITING_LOAD_SIMULATION_REPLIES,
            /**< Awaiting reply message to load simulation request from all server nodes. */
        PHASE_SIMULATION_RUNNING
            /**< Simulation is running. */
        }; // Phase
    
    /**
     * State of a server commanded by the action.
     */
    class NodeState
        {
    public:
        typedef StartupMessage::Providers Providers;
        
        /** Constructor. */
        NodeState() :
            received_archivation_state_reply( false ),
            received_load_simulation_reply( false )
            {
            }

    public:
        bool            received_archivation_state_reply;
            /**< True if archivation state reply message was received from the node. */
        ArchivationId   last_valid_archivation_id;
            /**< Last valid archivation id reported by the node. */
        bool            received_load_simulation_reply;
            /**< True if load simulation reply was received from the node. */
        Providers       providers;
            /**< List of providers owned by the node. */
        }; // NodeState
    
    /** NodeState table. */
    typedef ext::hash_map
        <
        NodeId,
        NodeState,
        NodeId::Hash,
        NodeId::EqualKey
        > Nodes;

public:

    /** Constructor. */
    NodeActionStartup();
    
    /** Destructor. */
    virtual ~NodeActionStartup();
    
public:

    // From NodeAction.
    virtual NextTickResult next_tick();
    
private:

    /** Check whether archivation state reply message was received from all server nodes. */
    bool archivation_state_replies_received_from_all_servers() const;
    
    /** Check whether load archivation reply message was received from all server nodes. */
    bool load_archivation_replies_received_from_all_servers() const;
    
    /** Send archivation state request message to all server nodes. */
    void send_archivation_state_request();
    
    /** Send load simulation request to all server nodes. */
    void send_load_simulation_request();
    
    /** Send unpause simulation request to all server nodes. */
    void send_unpause_simulation_request();
    
    /** Register remote providers owned by remote server nodes to object manager. */
    void register_remote_providers();
    
    /** Log current phase. */
    void log_phase();
    
public:

    /** Get node state of specified server. */
    NodeState & get_node_state
        (
        const NodeId & node_id
        );
    
    /** Handle archivation state reply message. */
    void handle_archivation_state_reply_message
        (
        const StartupMessage & msg
        );
    
    /** Hadle load simulation reply message. */
    void handle_load_simulation_reply
        (
        const StartupMessage & msg
        );
    
    /** Handle startup message. */
    static void handle_startup_message
        (
        const StartupMessage & msg
        );
    
    /**
     * Get NodeActionStartup object. Local node must
     * be coordinator of the action.
     */
    static NodeActionStartup & get_action_object();
    
private:

    Phase       phase;
        /**< Phase of the action. */
    Nodes       servers;
        /**< States of all server that are commanded by the action. */

    }; // class NodeActionStartup

/** @} */ // addtogroup node


} // namespace Core
} // namespace Massiv


#endif // MASSIV_NODE_ACTION_STARTUP_H
