/** @file node_action_state.h
 * Node action state structure.
 */

/*
 * begin        : Sat Oct 4 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_ACTION_STATE_H
#define MASSIV_NODE_ACTION_STATE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif


namespace Massiv {
namespace Core {


class NodeAction;


/** @addtogroup node
 * @{
 */

/**
 * Node action state structure.
 * Holds information about a node action in progress.
 * This structure is used by #NodeActionManager and there is no reason to use it
 * elsewhere.
 */
class NodeActionState
    {
public:

    /** Node action phase. */
    enum Phase
        {
        // Common:
        
        PHASE_NONE,
            /**< Default value when the state structure is created. */
        
        // Coordinator:
       
        PHASE_WAITING_FOR_REQUEST_REPLY,
            /**< Waiting for request reply message from all nodes. */
        PHASE_COORDINATOR_RUNNING,
            /**< Coordinator is running. */
        PHASE_WAITING_FOR_PREPARE_FINISH_REPLY,
            /**< Waiting for prepare finish reply from all nodes. */
        
        // Slave:
        
        PHASE_WAITING_FOR_COORDINATOR,
            /**< Waiting for coordinator selected message. */
        PHASE_RUNNING,
            /**< Running. Waiting for coordinator's orders. */
        PHASE_WAITING_FOR_FINISH
            /**< Waiting for finish message from coordinator. */
        
        }; // Phase
        
public:

    /** Constructor. */
    NodeActionState();
    
    /** Copy constructor. */
    NodeActionState
        (
        const NodeActionState & other
        );
    
    /** Destructor. */
    ~NodeActionState();
    
    /**
     * Operator=.
     * other's node_action_object is transferred to this object
     * (other's node_action_object is set to null).
     */
    NodeActionState & operator=
        (
        const NodeActionState & other
        );

public:

    Phase                           phase;
        /**< Current action phase. */
    int                             num_replies;
        /**< Number of received reply messages. Valid only
         * if phase is @c PHASE_WAITING_FOR_REQUEST_REPLY or
         * @c PHASE_WAITING_FOR_PREPARE_FINISH_REPLY.
         */
    int                             num_awaiting_replies;
        /**< How many reply messages we are waiting for. Valid
         * only if phase is @c PHASE_WAITING_FOR_REQUEST_REPLY or
         * @c PHASE_WAITING_FOR_PREPARE_FINISH_REPLY.
         */
    
    mutable NodeAction *            node_action_object;
        /**< Node action object. If not null, automatically
         * deleted in destructor.
         */
        
    }; // class NodeActionState

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_ACTION_STATE_H
