/** @file node_manager.cpp
 * Node manager implementation.
 */

/*
 * begin        : Wed Oct 17 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "node/node_manager.h"
#include "object/object_manager.h"
#include "object/archive_manager.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "system/system.h"
#include "utility/cache_impl.h"
#include "utility/debug.h"
#include "serializer/serializer.h"
#include "sysevent/system_event_queue.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "sysevent/forward_msg.h"
#include "time/time_manager.h"
#include "database/archive_database.h"
#include "database/node_database.h"
#include "net/net.h"
#include "node/node_manager_node.h"
#include "node/node_manager_server.h"
#include "node/node_manager_client.h"
#include "node/node_manager_anonymous.h"
#include "node/node_manager_service.h"

#include <sstream>


namespace Massiv {
namespace Core {


/*
 Table of reactions to "connection lost" event:

 S - server
 V - service
 C - client
 A - anonymous
 
 WHO column specifies the node that detected connection lost
 TO row specifies node to/from connection was lost
 
 error  - connection is thrown
 notify - core is notified, connection is closed
 kill   - node is disconnected from entire simulation
 "-"    - this combination should never happen

    \   |
     \to|
      \ |
 who   \|    S   |   V    |  C     |   A
 -------+--------------------------------
  S     | error  | notify | kill   |  -
 -------------------------------------------
  V     | notify | notify | notify | notify
 -------------------------------------------
  C     | error  | notify |   -    |  -
 -------------------------------------------
  A     |   -    | error  |   -    |  -
 -------------------------------------------

 */


//-----------------------------------------------------------------------------
// NodeManager implementation.
//-----------------------------------------------------------------------------


NodeManager::NodeManager() :
    initialized( false ),
    standalone_mode( false ),
    anonymous_mode( false ),
    node( NULL ),
    server( NULL ),
    client( NULL )
    {
    }


NodeManager::~NodeManager()
    {
    delete node;
    }


void NodeManager::initialize
    (
    bool the_standalone_mode
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::initialize()" );
    
    // Shutdown.
    delete node;
    node = NULL;
    server = NULL;
    client = NULL;
    initialized = false;
    standalone_mode = false;
    anonymous_mode = false;
    
    // Initialize.
    hope( !initialized );
    
    // Set common variables.
    standalone_mode = the_standalone_mode;
    
    // Log.
    if( standalone_mode )
        {
        std::ostringstream oss;
        oss << "Initializing node manager in standalone mode.";
        Global::log_info( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
        }
    
    // Set node manager to anonymous node.
    hope( !node );
    node = new Anonymous();
    anonymous_mode = true;
    
    // Done.
    initialized = true;
    }


void NodeManager::connect()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::connect()" );

    const NodeId       local_node_id = get_local_node_id();
    const NodeId::Type local_node_type = local_node_id.get_type();
    
    hope( initialized );
    hope( anonymous_mode );
    hope( node );
    delete node;
    node = NULL;
    hope( !server );
    hope( !client );
    
    // Create My structure.
    switch( local_node_type )
        {
        case NodeId::TYPE_SERVER:
            server = new Server();
            node = server;
            break;
        
        case NodeId::TYPE_CLIENT:
            client = new Client();
            node = client;
            break;
        
        case NodeId::TYPE_SERVICE:
            node = new Service();
            break;
        
        case NodeId::TYPE_ANONYMOUS:
            {
            std::ostringstream oss;
            oss << "Local node can't connect to simulation because "
                << "local node is running in anonymous mode. "
                << "Anonymous nodes can't connect to simulation.";
            throw ExNodeManagementError( oss.str() );
            }
        }

    hope( node );
    anonymous_mode = false;
    node->connect();
    }
    

void NodeManager::disconnect()
    {
    if( node )
        {
        node->disconnect();
        }
    }


void NodeManager::request_kill_node
    (
    const NodeId & node
    )
    {
    kill_node( node );
    }


NodeId NodeManager::get_local_node_id() const
    {
    NodeDatabase & node_database = Global::node_database();
    return node_database.get_local_node_id();
    }


NodeId NodeManager::get_local_system_forward_node() const
    {
    if( client )
        {
        return client->get_local_system_forward_node();
        }
    
    return NodeId();
    }


bool NodeManager::is_node_accessible
    (
    const NodeId & dest_node
    ) const
    {
    Network &       network = Global::network();
    NodeId::Type    dest_node_type = dest_node.get_type();

    if( dest_node == get_local_node_id() )
        {
        return true;
        }

    // Similar to Network::can_lock_buffer_to_node() but we
    // can't use it because we must check client nodes
    // in a special way. Client nodes can be accessible even if
    // net buffer can't be locked to them (no direct connection).
    // If there is not direct connection to a client but the client
    // is connected to simulation, the client is accessible via
    // message forwarding.
    
    // Lookup the node in node database.
        {
        NodeDatabase &                  node_database = Global::node_database();
        const NodeDatabase::NodeData *  dest_node_data;
                
        // Look-up the node in node database.
        node_database.get_node_data( dest_node, &dest_node_data );
        if( !dest_node_data )
            {
            return false;
            }
        }
        
    switch( dest_node_type )
        {
        case NodeId::TYPE_SERVER:
            return true;
            
        case NodeId::TYPE_CLIENT:
            if( !get_local_node_id().is_type( NodeId::TYPE_SERVER ) )
                {
                return network.can_lock_buffer_to_node( dest_node, NetworkCommon::STREAM_RELIABLE );
                }
            else
                {
                if( node )
                    {
                    return node->is_client_connected_to_simulation( dest_node );
                    }
                else
                    {
                    return false;
                    }
                }
            
        case NodeId::TYPE_SERVICE:
            return true;
        
        case NodeId::TYPE_ANONYMOUS:
            return network.can_lock_buffer_to_node( dest_node, NetworkCommon::STREAM_RELIABLE );

        } // switch( dest_node_type )
    
    return false;
    }


void NodeManager::next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::next_tick()" );
    
    if( node )
        {
        node->next_tick();
        }
    }

    
bool NodeManager::is_connected() const
    {
    if( node )
        {
        return node->is_connected();
        }
    else
        {
        return false;
        }
    }


//-----------------------------------------------------------------------------
// Misc.
//-----------------------------------------------------------------------------


void NodeManager::enumerate_archivation_nodes
    (
    NodeIdList & node_id_list
    )
    {
    hope( initialized && node );
    
    node_id_list.clear();
    
    Node::ServerTable::const_iterator it;
    for( it = node->servers.begin(); it != node->servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        node_id_list.push_back( node_id );
        }
    }


void NodeManager::send_message
    (
    const SystemMessage &     message_to_send,
    const NodeId &            dest_node,
    NetworkCommon::StreamType stream_type
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::send_message()" );
    Network &       network = Global::network();
    NodeId::Type    dest_node_type = dest_node.get_type();

    // Log.
        {
        std::ostringstream ss;
        ss  << "Sending '" << SystemMessage::type_to_string( message_to_send.get_type() ) << "' "
            << "to " << dest_node << ".";
        Global::log_debug
            (
            Status::FACILITY_NODE_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }

    if( !is_node_accessible( dest_node ) )
        {
        std::ostringstream oss;
        oss  << "Unable to send '" << SystemMessage::type_to_string( message_to_send.get_type() )
            << "' to " << dest_node << " because the destination node is not accesible by local node.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
        
        // Notify core.
        GlobalCallbacks::node_not_accessible( dest_node );
        return;
        }

    // Try send the message.
    switch( dest_node_type )
        {
        case NodeId::TYPE_SERVER:
            {
            // Network must be always able to lock net buffer to server nodes.
            hope( dest_node == get_local_node_id() || network.can_lock_buffer_to_node( dest_node, stream_type ) );
            do_send_message( message_to_send, dest_node, stream_type );
            break;
            }

        case NodeId::TYPE_CLIENT:
            {
            // Check if we can send the message directly to the destination node.
            if( dest_node == get_local_node_id() ||
                network.can_lock_buffer_to_node( dest_node, stream_type ) )
                {
                // We have opened connection to the client.
                do_send_message( message_to_send, dest_node, stream_type );
                }
            else
                {
                // There is no opened connection to the client node.
                // Local node must be server node.
                hope( server != NULL );
                // The client is connected to simulation.
                hope( server->is_client_connected_to_simulation( dest_node ) );
                // We will forward the message through forward node.
                // Messages can be forwarded to client nodes only.
                
                const NodeId forward_node = server->get_client_forward_node( dest_node );
                hope( forward_node.is_type( NodeId::TYPE_SERVER ) );
                
                if( message_to_send.get_type() == SystemMessage::FORWARD )
                    {
                    do_send_message( message_to_send, forward_node, stream_type );
                    }
                else
                    {
                    bool reliable = ( stream_type == NetworkCommon::STREAM_RELIABLE );
                    ForwardMessage  forward_message( dest_node, message_to_send, reliable );
                    do_send_message( forward_message, forward_node, stream_type );
                    }
                }
            break;
            }
            
        case NodeId::TYPE_SERVICE:
            {
            // Network must be always able to lock net buffer to service nodes.
            hope( dest_node == get_local_node_id() || network.can_lock_buffer_to_node( dest_node, stream_type ) );
            do_send_message( message_to_send, dest_node, stream_type );
            break;
            }

        case NodeId::TYPE_ANONYMOUS:
            {
            hope( dest_node == get_local_node_id() || network.can_lock_buffer_to_node( dest_node, stream_type ) );
            do_send_message( message_to_send, dest_node, stream_type );
            break;
            }
        } // switch( dest_node_type )
    }


void NodeManager::multicast_message
    (
    const SystemMessage &     message,
    const NodeIdList &        nodes,
    NetworkCommon::StreamType stream_type
    )
    {
    /** @todo FIXME CZ: Udelat podporu primo v Network?
     * Pak ale testovat prime konexe - nebo aspon ze to jsou
     * server nody.
     */
    NodeIdList::const_iterator  it;

    for( it = nodes.begin(); it != nodes.end(); ++it )
        {
        const NodeId & node_id = ( *it );
        message.send( node_id, stream_type );
        }
    }

    
void NodeManager::multicast_message_to_server_nodes
    (
    const SystemMessage &       message,
    NetworkCommon::StreamType   stream_type,
    bool                        donot_send_to_local_node_if_server
    )
    {
    hope( initialized && node && !anonymous_mode );

    const bool ignore_local_node = donot_send_to_local_node_if_server;
    const NodeId local_node_id = get_local_node_id();
    
    Node::ServerTable::const_iterator it;
    for( it = node->servers.begin(); it != node->servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        if( !ignore_local_node || node_id != local_node_id )
            {
            message.send( node_id, stream_type );
            }
        }
    }


void NodeManager::notify_simulation_is_running()
    {
    if( server )
        {
        server->notify_simulation_is_running();
        }
    }


void NodeManager::server_do_connect_attempt()
    {
    hope( server );
    hope( node );
    hope( get_local_node_id().get_type() == NodeId::TYPE_SERVER );

    server->initiate_connections_to_server_nodes();
    }

//-----------------------------------------------------------------------------
// Private.
//-----------------------------------------------------------------------------

void NodeManager::kill_node
    (
    const NodeId & other_node
    )
    {
    if( node )
        {
        node->kill_node( other_node );
        }
    }
    
    
void NodeManager::do_send_message
    (
    const SystemMessage &     message,
    const NodeId &            dest_node,
    NetworkCommon::StreamType stream_type
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::do_send_message()" );
    Network &   network = Global::network();

    // Caller must ensure that there is opened connection to the destination node.
    hope( dest_node == get_local_node_id() || network.can_lock_buffer_to_node( dest_node, stream_type ) );

    // Connection to local node doesn't go through network.
    if( dest_node == get_local_node_id() )
        {
        BitOStream                bos;
        BitIStream                bis;
        Serializer::Description   desc;

        message.bitstr_write( bos, desc );
        bis.bind( bos.get_buffer() );
        SystemMessage::dispatch_system_message_from_packet( bis, get_local_node_id() );

        return;
        }

    switch( stream_type )
        {
        case NetworkCommon::STREAM_RELIABLE:
            {
            Serializer::Description desc;
            TCPOBuffer * buffer = network.lock_reliable( dest_node, Network::NET_MIN_USER_MESSAGE );
            hope( buffer );
            
            message.bitstr_write( *buffer, desc );
            network.unlock_reliable( buffer );
            }
            break;

        case NetworkCommon::STREAM_UNRELIABLE:
            {
            Serializer::Description desc;
            UDPOBuffer * buffer = network.lock_unreliable( dest_node, Network::NET_MIN_USER_MESSAGE );
            hope( buffer );

            message.bitstr_write( *buffer, desc );
            network.unlock_unreliable( buffer, true );
            }
            break;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
