/** @file node_manager.h
 * Node manager.
 * Node manager stores information about nodes
 * connected to the simulation.
 */

/*
 * begin        : Wed Oct 17 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_MANAGER_H
#define MASSIV_NODE_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_NET_COMMON_H
#include "net/net_common.h"
#endif

#ifndef MASSIV_ARCHIVATION_ID_H
#include "object/archivation_id.h"
#endif

#ifndef MASSIV_CACHE_H
#include "utility/cache.h"
#endif

#ifndef MASSIV_MISC_CLASS_H
#include "utility/misc_class.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Node manager.
 * Node manager stores information about nodes
 * connected to the simulation.
 *
 * Every @c SystemMessage send must go through
 * the node manager.
 * System message can be sent only to nodes, which
 * are connected to the simulation.
 *
 * Node connected to the simulation has:
 * <ul>
 *   <li> Assigned unique @c NodeId.
 *   <li> Assigned unique @c ObjectId.
 * </ul>
 *
 */
class NodeManager
    {
public:

    /**
     * Result of connection established callback from network.
     */
    enum ConnectionEstablishedResult
        {
        RESULT_OK,
            /**< Connection was accepted. */
        RESULT_CLOSE_CONNECTION
            /**< Connection was not accepted. Network must
             * close the connection. */
        }; // ConnectionEstablishedResult
    
    /**
     * List of node ids.
     * Used to exchange data to/from node manager.
     */
    typedef SystemMessage::NodeIdList NodeIdList;

private:

    NO_COPY_CONSTRUCTOR( NodeManager );
    NO_ASSIGNMENT_OPERATOR( NodeManager );
    
public:

    class Node;
    class Server;
    class Client;
    class Service;
    class Anonymous;
    class NodeManagerSettings;

public:

    /**
     * Constructor.
     */
    NodeManager();

    /**
     * Destructor.
     */
    ~NodeManager();

    /**
     * Initialize the node manager.
     * @param standalone_mode Specified whether the node manager
     * should be initialized in standalone mode. This will force to
     * ignore all settings for remote nodes from the registry
     * and initializing the node manager like local node is the only
     * node in the simulation.
     * @pre Node database is already initialized.
     */
    void initialize
        (
        bool standalone_mode = false
        );
    
    /**
     * Connect to the simulation.
     * @pre The node manager is initialized.
     */
    void connect();
    
    /**
     * Disconnect from the simulation.
     */
    void disconnect();

public:

    /**
     * Run next node manager tick.
     */
    void next_tick();

    /**
     * If local node is client this function returns true if local node
     * is connected to simulation.
     * If local node is server this function returns true if simulation
     * is running.
     * If local node is service this function returns true if the service
     * is ready to accept requests.
     */
    bool is_connected() const;

    /**
     * Request to disconnect @a node from local node
     * in consequence of an error in communication with the @a node.
     * @note State of the node stored in node manager is not
     *      changed directly from this function. The node is only
     *      marked as disconecting (to prevent receiving any message
     *      from the node) and the kill request is scheduled to be
     *      processed as soon as possible.
     */
    void request_kill_node
        (
        const NodeId & node
        );

    /**
     * Get forward server node for local system.
     * @pre The local system is client node.
     */
    NodeId get_local_system_forward_node() const;
    
    /**
     * Check whether @a node is accessible by local node i.e. whether
     * a message can be sent to the @a node.
     * False is returned if the @a node is invalid node id or
     * there is no way how to deliver a message to the node.
     */
    bool is_node_accessible
        (
        const NodeId & node
        ) const;

public:

    //
    // Network callbacks.
    //

    /**
     * Callback which is called once a new connection is established to a node.
     * @warning This callback can be called only from Network::next_tick().
     */
    ConnectionEstablishedResult net_callback_connection_established
        (
        const NodeId & node_id
        );

    /**
     * Callback which is called once a existing connection to a node is closed
     * (but not lost), for example due to automatic closing of idle connections.
     * @warning This callback can be called only from Network::next_tick().
     */
    void net_callback_connection_closed
        (
        const NodeId & node_id
        );

    /**
     * Callback which is called if an existing connection to a node is lost.
     * @warning This callback can be called only from Network::next_tick().
     */
    void net_callback_connection_lost
        (
        const NodeId & node_id
        );
        
    /**
     * Callback which is called if cannot be restored or created connection to a node.
     * @warning This callback can be called only from Network::next_tick().
     */    
    void net_callback_cannot_initiate_connection
        (
        const NodeId & node_id
        );
        
public:

    /**
     * Handle AcceptClientMessage.
     */
    SystemMessage::ExecuteResult handle_accept_client_message
        (
        const AcceptClientMessage & msg
        );

    /**
     * Handle ClientAcceptedMessage.
     */
    SystemMessage::ExecuteResult handle_client_accepted_message
        (
        const ClientAcceptedMessage & msg
        );

    /**
     * Handle ClientReadyMessage.
     */
    SystemMessage::ExecuteResult handle_client_ready_message
        (
        const ClientReadyMessage & msg
        );

    /**
     * Handle ClientReadyReplyMessage.
     */
    SystemMessage::ExecuteResult handle_client_ready_reply_message
        (
        const ClientReadyReplyMessage & msg
        );

    /**
     * Handle KillClientMessage.
     */
    SystemMessage::ExecuteResult handle_kill_client_message
        (
        const KillClientMessage & msg
        );

    /**
     * Handle WelcomeMessage.
     */
    SystemMessage::ExecuteResult handle_welcome_message
        (
        const WelcomeMessage & msg
        );
        
    /**
     * Handle ForwardedMessage.
     */
    SystemMessage::ExecuteResult handle_forwarded_message
        (
        const ForwardedMessage & msg
        );
        
public:

    /**
     * Enumerate archivation nodes which are connected to the simulation.
     * @param node_id_list List where to store the archivation node ids.
     *        The list is cleared before any node ids are inserted into it.
     * @post If exception is thrown, the @a node_id_list is made empty.
     * @note Local node id is included in the @a node_id_list if
     * it is archivation node.
     */
    void enumerate_archivation_nodes
        (
        NodeIdList & node_id_list
        );
    
    // HACK FIXME
    void enumerate_server_nodes
        (
        NodeIdList & node_id_list
        )
        {
        enumerate_archivation_nodes( node_id_list );
        }

    /**
     * Send @a message to @a node.
     *
     * @note You should not use this function directly - 
     * use @c SystemMessage::send() instead.
     */
    void send_message
        (
        const SystemMessage &     message,
        const NodeId &            node,
        NetworkCommon::StreamType net_stream_type
        );

    /**
     * Send @a message to @a nodes.
     */
    void multicast_message
        (
        const SystemMessage &     message,
        const NodeIdList &        nodes,
        NetworkCommon::StreamType net_stream_type
        );
    
    /**
     * Send @a message to all server nodes.
     * @param donot_send_to_local_node_if_server If the local system
     *      is server node and this argument is true, the message
     *      is not send to local node.
     */
    void multicast_message_to_server_nodes
        (
        const SystemMessage &       message,
        NetworkCommon::StreamType   net_stream_type,
        bool                        donot_send_to_local_node_if_server = false
        );
        
public:

    /**
     * Notify node manager that simulation was successfully loaded
     * and is running.
     * This function is only for #NodeActionStartup purposes.
     * @pre Local node is server node.
     */
    void notify_simulation_is_running();
    
    /**
     * Attempts to connect to server nodes, which are not connected yet.
     * This function is used only in server starting procedure.
     * @pre Local node is server node.
     */
    void server_do_connect_attempt();

private:

    /**
     * Kill @a other_node from local system.
     */
    void kill_node
        (
        const NodeId & other_node
        );
    
    /**
     * Get local node id.
     */
    NodeId get_local_node_id() const;

    /**
     * Send a message to all server nodes.
     * @param ignore_local_node If true, the message is not send
     * to local node (in case the local node is server node).
     */
    void send_message_to_all_server_nodes
        (
        const SystemMessage &     message_to_send,
        NetworkCommon::StreamType stream_type,
        bool                      ignore_local_node = false
        ) const;

    /**
     * "Low-level" send @a message to @a node.
     */
    void do_send_message
        (
        const SystemMessage &     message,
        const NodeId &            node,
        NetworkCommon::StreamType net_stream_type
        );

private:

    bool            initialized;
        /**< True if the node manager was already initialized. */

    bool            standalone_mode;
        /**< True if the node manager is running in standalone node mode. */
    
    bool            anonymous_mode;
        /**< True if the node manager is running in anonymous mode. */
    
    Node *          node;
        /**< Node object that encapsulates node manager behaviour depending
         * on local node type. Null if the manager was not yet initialized.
         */
    
    Server *        server;
        /**< Same pointer as @c node. */
    
    Client *        client;
        /**< Same pointer as @c node. */

    }; // NodeManager

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_MANAGER_H
