/** @file node_manager_client.cpp
 * Client specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_client.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"


namespace Massiv {
namespace Core {


NodeManager::Client::Client() :
    phase( PHASE_NONE )
    {
    }


void NodeManager::Client::connect()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Client::connect()" );
    
    // Enumerate server nodes.
    initialize_server_node_table();

    // Select random server node and initiate connection to it.
    initiate_connection_to_random_server();
    
    // Switch phase.
    phase = PHASE_WAITING_FOR_WELCOME;
    
    // Log.
    log_phase();
    }


void NodeManager::Client::disconnect()
    {
    if( phase == PHASE_NONE )
        {
        return;
        }

    // Close connection to all server nodes.
    ServerTable::const_iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        Global::network().close_connection( node_id );
        }
    
    phase = PHASE_DISCONNECTING;
    }


bool NodeManager::Client::is_connected()
    {
    return phase == PHASE_RUNNING || phase == PHASE_DISCONNECTING;
    }


void NodeManager::Client::next_tick()
    {
    if( phase == PHASE_DISCONNECTING )
        {
        Network &                   network = Global::network();
        ServerTable::const_iterator it;
        bool                        done = true;

        for( it = servers.begin(); it != servers.end(); ++it )
            {
            const NodeId & node_id = it->first;
            /** @todo FIXME CZ: Vrati ta funkce false az kdyz je vsechno
             * spravne uzavreno, nebo ihned vraci false po te, co se
             * zavola close_connection() a tudiz jeste nemusi byt vsechno
             * v networku procisteno?
             */
            if( network.is_node_connected( node_id ) )
                {
                done = false;
                break;
                }
            }

        if( done )
            {
            phase = PHASE_NONE;
            }        
        }
    }


void NodeManager::Client::initiate_connection_to_random_server()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Client::initiate_connection_to_random_server()" );
    hope( !servers.empty() );
    const NodeId server_node_id = servers.begin()->first;
    
    Global::network().async_initiate_connection_to_server( server_node_id );
    
    forward_node = server_node_id;
    }


void NodeManager::Client::log_phase()
    {
    std::string phase_name;
    switch( phase )
        {
        case PHASE_NONE:
            phase_name = "NONE";
            break;
        case PHASE_WAITING_FOR_WELCOME:
            phase_name = "WAITING_FOR_WELCOME";
            break;
        case PHASE_RUNNING:
            phase_name = "RUNNING";
            break;
        case PHASE_DISCONNECTING:
            phase_name = "DISCONNECTING";
            break;
        }

    std::ostringstream oss;
    oss << "NodeManager::Client::phase is now " << phase_name << ".";
    Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
    }


} // namespace Massiv::Core
} // namespace Massiv
