/** @file node_manager_client.h
 * Client specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_MANAGER_CLIENT_H
#define MASSIV_NODE_MANAGER_CLIENT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODE_MANAGER_NODE_H
#include "node/node_manager_node.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Client encapsulates node manager's behaviour if local node is client.
 */
class NodeManager::Client : public NodeManager::Node
    {
    NO_COPY_CONSTRUCTOR( Client );
    NO_ASSIGNMENT_OPERATOR( Client );
    
public:
    
    /** Internal phase. */
    enum Phase
        {
        PHASE_NONE,
            /**< Initial phase. The client object has just been created but not initialized yet. */

        PHASE_WAITING_FOR_WELCOME,
            /**< Waiting for WelcomeMessage from a server node. */

        PHASE_RUNNING,
            /**< The client is running. */

        PHASE_DISCONNECTING,
            /**< Disconnecting from the simulation. */
        }; // Phase
        
public:

    /** Constructor. */
    Client();
    
public:

    // From NodeManager::Node.
    virtual void connect();
    
    // From NodeManager::Node.
    virtual void disconnect();
    
    // From NodeManager::Node.
    bool is_connected();
    
    // From NodeManager::Node.
    virtual void next_tick();

    // From NodeManager::Node.
    virtual ConnectionEstablishedResult net_callback_connection_established
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_connection_closed
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_connection_lost
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_cannot_initiate_connection
        (
        const NodeId & node_id
        );

public:

    /**
     * Handle WelcomeMessage.
     */
    SystemMessage::ExecuteResult handle_welcome_message
        (
        const WelcomeMessage & msg
        );

    /** Get forward server node for local system. */
    NodeId get_local_system_forward_node() const
        {
        return forward_node;
        }

private:

    /** Initiate connection to a server. */
    void initiate_connection_to_random_server();

    /** Log current phase. */
    void log_phase();

private:

    Phase               phase;
        /**< Internal phase. */

    NodeId              forward_node;
        /**< Forward server node for local node.
         * Forward node is the first server node the node manager creates
         * connection to.
         */

    }; // NodeManager::Client

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_MANAGER_CLIENT_H
