/** @file node_manager_client_callbacks.cpp
 * Client specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_client.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"


namespace Massiv {
namespace Core {


NodeManager::ConnectionEstablishedResult NodeManager::Client::net_callback_connection_established
    (
    const NodeId & node_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Client::net_callback_connection_established()" );
    
    // Check if not yet initialized...
    if( phase == PHASE_NONE )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id << ". The local node "
            << "is not yet ready.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }

    // Check anonymous and client node.
    if( node_id.is_type( NodeId::TYPE_ANONYMOUS ) )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id
            << " because anonymous and client nodes can't connect to client nodes.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }
        
    // Check service.
    if( node_id.is_type( NodeId::TYPE_SERVICE ) )
        {
        return RESULT_OK;
        }
    
    // Check server.
    hope( node_id.is_type( NodeId::TYPE_SERVER ) );
    hope( get_server_state( node_id ) != NULL ); // Server nodes already in the server table.
    return RESULT_OK;
    }


void NodeManager::Client::net_callback_connection_closed
    (
    const NodeId & node_id
    )
    {
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        if( node_id == forward_node && phase == PHASE_DISCONNECTING )
            {
            // The connection was closed because we are disconnecting. Do nothing.
            return;
            }
            
        if( node_id == forward_node )
            {
            std::ostringstream oss;
            oss << "Disconnected from simulation (connection closed by server node " << node_id << ").";
            throw ExConnectionLost( oss.str() );
            }
        
        // Auto-close feature. Do nothing because we are still connected to simulation through
        // our forward server node.
        }
    else
        {
        // Notify core.
        GlobalCallbacks::node_disconnected( node_id );
        }
    }


void NodeManager::Client::net_callback_connection_lost
    (
    const NodeId & node_id
    )
    {
    // Notify core.
    GlobalCallbacks::node_disconnected( node_id );
    
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        throw_connection_lost( node_id );
        }
    }


void NodeManager::Client::net_callback_cannot_initiate_connection
    (
    const NodeId & node_id
    )
    {
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        throw_connection_failed( node_id );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
