/** @file node_manager_node.cpp
 * Internal structure of node manager.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_node.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"
#include "system/system.h"
#include "sysevent/system_event_queue.h"


namespace Massiv {
namespace Core {


namespace {

// Helper classes.

class InitializeServerNodeTableEnumerator : public NodeDatabase::Enumerator
    {
    typedef NodeManager::Node::ServerTable  NodeTable;
    typedef NodeManager::Node::ServerState  NodeState;

public:
    InitializeServerNodeTableEnumerator
        (
        const NodeId &  the_local_node_id,
        NodeTable &     the_node_table,
        bool            the_ignore_remote_server_nodes
        ) :
        local_node_id( the_local_node_id ),
        node_table( the_node_table ),
        ignore_remote_server_nodes( the_ignore_remote_server_nodes )
        {
        }

    virtual void operator()
        (
        const NodeDatabase::NodeData & node_data
        )
        {
        const NodeId node_id = node_data.get_node_id();
        hope( node_id.get_type() == NodeId::TYPE_SERVER );
        hope( node_table.find( node_id ) == node_table.end() );
        const bool insert = ignore_remote_server_nodes ? ( local_node_id == node_id ) : true;
        if( insert )
            {
            NodeState state;
            state.phase = NodeManager::Node::REMOTE_SERVER_PHASE_NONE;
            node_table.insert( NodeTable::value_type( node_id, state ) );
            }
        }

private:
    const NodeId &  local_node_id;            
    NodeTable &     node_table;
    bool            ignore_remote_server_nodes;
    
    }; // InitializeServerNodeTableEnumerator

    
} // anonymous namespace


//-----------------------------------------------------------------------------
// NodeManager::Node
//-----------------------------------------------------------------------------


NodeManager::Node::Node()
    {
    }


NodeManager::Node::~Node()
    {
    }


bool NodeManager::Node::is_client_connected_to_simulation
    (
    const NodeId & node_id
    )
    {
    return false;
    }


void NodeManager::Node::kill_node
    (
    const NodeId & other_node
    )
    {
    Network & network = Global::network();

    /** @todo FIXME FIXME FIXME - Jak to funguje? Chceme kill. */
    network.close_connection( other_node );

    GlobalCallbacks::node_disconnected( other_node );
    }


NodeId NodeManager::Node::get_local_node_id() const
    {
    NodeDatabase & node_database = Global::node_database();
    return node_database.get_local_node_id();
    }


void NodeManager::Node::initialize_server_node_table
    (
    bool ignore_remote_server_nodes
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Node::initialize_server_node_table()" );
    hope( servers.empty() );
    
    // Enumerate server nodes and put them in the table.
    InitializeServerNodeTableEnumerator enumerator( get_local_node_id(), servers, ignore_remote_server_nodes );
    Global::node_database().enumerate_nodes( NodeId::TYPE_SERVER, enumerator );
    
    // If local node is server, make sure it was enumerated.
    if( System::is_server() && !is_registered_in_server_node_table( get_local_node_id() ) )
        {
        // Should never happen.
        std::ostringstream oss;
        oss << "Node manager internal error: Local server node " << get_local_node_id()
            << " was not found in server nodes!";
        panic( oss.str().c_str() );
        }
    }


bool NodeManager::Node::test_remote_phase_of_server_nodes
    (
    RemoteServerPhase test_phase
    ) const
    {
    ServerTable::const_iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const ServerState & state = it->second;
        if( state.phase != test_phase )
            {
            return false;
            }
        }
    return true;
    }


void NodeManager::Node::set_phase_of_server_nodes
    (
    RemoteServerPhase new_phase
    )
    {
    ServerTable::iterator it;
    for( it = servers.begin(); it != servers.end(); ++it )
        {
        ServerState & state = it->second;
        state.phase = new_phase;
        }
    }


void NodeManager::Node::pause_message_queues_of_server_nodes
    (
    unsigned int bits_to_enable
    ) const
    {
    SystemEventQueue &          system_event_queue = Global::system_event_queue();
    ServerTable::const_iterator it;

    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        unsigned int pause_mask = system_event_queue.get_server_pause_mask( node_id );
        pause_mask |= bits_to_enable;
        system_event_queue.set_server_pause_mask( node_id, pause_mask );
        }
    }


void NodeManager::Node::unpause_message_queues_of_server_nodes
    (
    unsigned int bits_to_disable
    ) const
    {
    SystemEventQueue &          system_event_queue = Global::system_event_queue();
    ServerTable::const_iterator it;

    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        unsigned int pause_mask = system_event_queue.get_server_pause_mask( node_id );
        pause_mask &= ~( bits_to_disable );
        system_event_queue.set_server_pause_mask( node_id, pause_mask );
        }
    }


bool NodeManager::Node::is_registered_in_server_node_table
    (
    const NodeId & node_id
    ) const
    {
    return servers.find( node_id ) != servers.end();
    }


NodeManager::Node::ServerState * NodeManager::Node::get_server_state
    (
    const NodeId & node_id
    )
    {
    ServerTable::iterator it = servers.find( node_id );
    if( it != servers.end() )
        {
        ServerState & state = it->second;
        return &state;
        }
    else
        {
        return NULL;
        }
    }


void NodeManager::Node::throw_connection_lost
    (
    const NodeId & node_id
    )
    {
    std::ostringstream oss;
    oss << "Connection lost to/from " << node_id << ".";
    throw ExConnectionLost( oss.str() );
    }


void NodeManager::Node::throw_connection_failed
    (
    const NodeId & node_id
    )
    {
    std::ostringstream oss;
    oss << "Failed to establish connection to " << node_id << ".";
    throw ExConnectionFailed( oss.str() );
    }


std::string NodeManager::Node::err_unexpected_message
    (
    const SystemMessage & msg
    )
    {
    const NodeId &          msg_sender = msg.get_sender();
    std::ostringstream      ss;
    ss  << "Unexpected " << SystemMessage::type_to_string( msg.get_type() )
        << " message received from node " << msg_sender << ".";
    return ss.str();
    }


} // namespace Massiv::Core
} // namespace Massiv
