/** @file node_manager_node.h
 * Internal structure of node manager.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_MANAGER_NODE_H
#define MASSIV_NODE_MANAGER_NODE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODE_MANAGER_H
#include "node/node_manager.h"
#endif

#ifndef MASSIV_NODE_MANAGER_SETTINGS_H
#include "node/node_manager_settings.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Node manager's internal data structure. Encapsulates behaviour
 * of node manager based on local node type.
 * @see NodeManager::Server, NodeManager::Client, NodeManager::Service,
 * NodeManager::Anonymous.
 */
class NodeManager::Node
    {
    NO_COPY_CONSTRUCTOR( Node );
    NO_ASSIGNMENT_OPERATOR( Node );
    
    friend class NodeManager;

public:

    /** Remote server phase. */
    enum RemoteServerPhase
        {
        REMOTE_SERVER_PHASE_NONE,
        REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED,
        REMOTE_SERVER_PHASE_RUNNING
        }; // RemoteServerPhase
    
    /** Disconnect reason. */
    enum DisconnectReason
        {
        CONNECTION_CLOSED_REASON,
        CONNECTION_LOST_REASON
        }; // DisconnectReason

    /** Server state. */
    struct ServerState
        {
        RemoteServerPhase   phase;
        }; // ServerState
    
    /** Server state table. */
    typedef ext::hash_map
        <
        NodeId,
        ServerState,
        NodeId::Hash,
        NodeId::EqualKey
        > ServerTable;

public:

    /** Constructor. */
    Node();
    
    /** Destructor. */
    virtual ~Node();

public:

    /**
     * Connect.
     */
    virtual void connect() = 0;
    
    /**
     * Disconnect.
     */
    virtual void disconnect() = 0;
    
    /**
     * Check whether the local node is connected to simulation.
     */
    virtual bool is_connected() = 0;
    
    /**
     * Tick.
     */
    virtual void next_tick() = 0;

    /**
     * Callback from network.
     * @see NodeManager.
     */
    virtual ConnectionEstablishedResult net_callback_connection_established
        (
        const NodeId & node_id
        ) = 0;

    /**
     * Callback from network.
     * @see NodeManager.
     */
    virtual void net_callback_connection_closed
        (
        const NodeId & node_id
        ) = 0;

    /**
     * Callback from network.
     * @see NodeManager.
     */
    virtual void net_callback_connection_lost
        (
        const NodeId & node_id
        ) = 0;

    /**
     * Callback from network.
     * @see NodeManager.
     */
    virtual void net_callback_cannot_initiate_connection
        (
        const NodeId & node_id
        ) = 0;

    /**
     * Check whether client node is connected to simulation.
     * Should be reimplemented only for server node, default implementation
     * returns false.
     */
    virtual bool is_client_connected_to_simulation
        (
        const NodeId & node_id
        );

    /**
     * Kill @a other_node from local system.
     * Should be reimplemented only for server node. Default implementation
     * only kills connection to the @a other_node and notifies core.
     */
    virtual void kill_node
        (
        const NodeId & other_node
        );

protected:

    /** Get local node id. */
    NodeId get_local_node_id() const;
    
    /**
     * Initialize @a servers table.
     * This functions should be called in @c connect() by derived class.
     * @param ignore_remote_server_nodes If true, remote server nodes won't
     *     be registered in the @a servers table. In other world, if local
     *     node is not server, the @a servers table will be empty; if local
     *     node is server, only local node will be registered in the table.
     * @pre @a servers table is empty.
     */
    void initialize_server_node_table
        (
        bool ignore_remote_server_nodes = false
        );
     
    /**
     * Check if phase of all server nodes stored in @c servers table
     * is @a test_phase.
     */
    bool test_remote_phase_of_server_nodes
        (
        RemoteServerPhase test_phase
        ) const;
        
    /**
     * Set phase of all server nodes stored in @c servers table
     * to @a new_phase
     */
    void set_phase_of_server_nodes
        (
        RemoteServerPhase new_phase
        );

    /**
     * Pause system message queues of all server nodes (registered
     * in @c servers table).
     * @param bits_to_enable Bits which will be set
     * from pause masks of the queues.
     */
    void pause_message_queues_of_server_nodes
        (
        unsigned int bits_to_enable
        ) const;

    /**
     * Unpause system message queues of all server nodes (registered
     * in @c servers table).
     * @param bits_to_disable Bits which will be removed
     * from pause masks of the queues.
     */
    void unpause_message_queues_of_server_nodes
        (
        unsigned int bits_to_disable
        ) const;

    /**
     * Check whether @a node_id is registered in @c servers table.
     */
    bool is_registered_in_server_node_table
        (
        const NodeId & node_id
        ) const;
    
    /** Get state of server node. Returns NULL the node is not registered in @c servers table. */
    ServerState * get_server_state
        (
        const NodeId & node_id
        );
    
    /** Throw @c ExConnectionLost exception. */
    void throw_connection_lost
        (
        const NodeId & node_id
        );
    
    /** Throw @c ExConnectionFailed exception. */
    void throw_connection_failed
        (
        const NodeId & node_id
        );

    /**
     * Error string.
     * Unexpected message received from some node.
     */
    std::string err_unexpected_message
        (
        const SystemMessage & msg
        );

public:

    ServerTable         servers;
        /**< Server table. Initialized at the beginning of connect().
         * Contains all server nodes even if there is no connection
         * to some of them. Server nodes are never removed from this table.
         */

    NodeManagerSettings settings;
        /**< Settings. */

    }; // NodeManager::Node

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_MANAGER_NODE_H
