/** @file node_manager_server.cpp
 * Server specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_server.h"
#include "utility/cache_impl.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"
#include "node/node_action.h"
#include "node/node_action_manager.h"


namespace Massiv {
namespace Core {


//-----------------------------------------------------------------------------
// NodeManager::Node::ClientState
//-----------------------------------------------------------------------------


NodeManager::Server::ClientState::ClientState() :
    phase( REMOTE_CLIENT_PHASE_NONE ),
    client_accepted_count( 0 ),
    was_account_notified( false ),
    was_registered_in_time_manager( false )
    {
    }


//-----------------------------------------------------------------------------
// NodeManager::Server
//-----------------------------------------------------------------------------


NodeManager::Server::Server() :
    phase( PHASE_NONE ),
    simulation_ready( false )
    {
    }


void NodeManager::Server::connect()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::connect()" );
    
    // Enumerate server nodes.
    initialize_server_node_table();
    // FIXME: Use standalone flag stored in NodeManager
    // to call initialize_server_node_table( true )?
    
    // Start waiting on well known port of the server.
    const ServerAddress sa = Global::node_database().get_node_server_address( get_local_node_id() );
    const int           port = sa.get_port();

    // Log.
        {
        std::ostringstream oss;
        oss  << "Starting server on port " << port << ".";
        Global::log_info( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
        }
        
    // Initialize network.
    if( !Global::network().start_well_known_port_server( port ) )
        {
        /** @todo FIXME */
        panic( "Network::start_well_known_port_server() failed." );
        }

    // Initialize connections to all servers.
    initiate_connections_to_server_nodes();
    
    // Pause all messages from any server until connections to all servers are established.
    pause_message_queues_of_server_nodes( SystemMessage::PAUSE_WAITING_FOR_INITIAL_CONNECTIONS );
    
    // Pause all simulation related messages until simulation is loaded.
    pause_message_queues_of_server_nodes( SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY );
    
    // Switch to connection phase.
    phase = PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS;

    // Log.
    log_phase();
    }


bool NodeManager::Server::is_connected()
    {
    if( phase != PHASE_RUNNING )
        {
        return false;
        }
    
    // Check if simulation is loaded and running...
    if( simulation_ready )
        {
        return true;
        }

    return false;
    }


void NodeManager::Server::disconnect()
    {
    phase = PHASE_NONE;
    }


void NodeManager::Server::next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::next_tick()" );

    switch( phase )
        {
        case PHASE_NONE:
            // The object should be initialized asap.
            break;

        case PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS:
            phase_waiting_for_server_to_server_connections();
            break;
        
        case PHASE_RUNNING:
            phase_running();
            break;
        }
    }


bool NodeManager::Server::is_client_connected_to_simulation
    (
    const NodeId & node_id
    )
    {
    ClientState * state = get_client_state( node_id );
    if( !state )
        {
        return false;
        }
    
    if( state->phase == REMOTE_CLIENT_PHASE_CONNECTED )
        {
        return true;
        }
    
    return false;
    }


void NodeManager::Server::kill_node
    (
    const NodeId & other_node
    )
    {
    NodeId::Type    other_node_type = other_node.get_type();
    
    if( other_node_type == NodeId::TYPE_CLIENT )
        {
        disconnect_client( other_node );
        }
    else if( other_node_type == NodeId::TYPE_SERVER )
        {
        std::ostringstream oss;
        oss << "Trying to kill server node " << other_node << ".";
        panic( oss.str().c_str() );
        }
    else
        {
        Super::kill_node( other_node );
        }
    }


NodeId NodeManager::Server::get_client_forward_node
    (
    const NodeId & client_node
    )
    {
    hope( client_node.is_type( NodeId::TYPE_CLIENT ) );
    hope( is_client_connected_to_simulation( client_node ) );
    ClientState * state = get_client_state( client_node );
    hope( state );
    return state->forward_node;
    }


void NodeManager::Server::notify_simulation_is_running()
    {
    hope( !simulation_ready );
    simulation_ready = true;
    
    // Unpause all simulation messages.
    unpause_message_queues_of_server_nodes( SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY );
    }


void NodeManager::Server::disconnect_client
    (
    const NodeId & other_node
    )
    {
    // Send kill client message to all server nodes (also to ourself).
    KillClientMessage   msg( other_node );
    msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
    
    ClientState * state = get_client_state( other_node );
    if( state )
        {
        state->phase = REMOTE_CLIENT_PHASE_DISCONNECTING;
        }
    }


void NodeManager::Server::do_kill_client
    (
    const NodeId & client_node
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::do_kill_client()" );
    ClientState *           node_state = get_client_state( client_node );
    bool                    notify_account_object = false;

    if( node_state && node_state->was_account_notified )
        {
        hope( node_state->forward_node == get_local_node_id() );
        node_state->was_account_notified = false;
        notify_account_object = true;
        }
    
    // Remove the client from time manager's data structures.
    if( node_state && node_state->was_registered_in_time_manager )
        {
        TimeManager & time_manager = Global::time_manager();
        time_manager.remove_node_from_synchronization( client_node );
        node_state->was_registered_in_time_manager = false;
        }

    // Remove the client from client node list.
    if( node_state )
        {
        // Remove the client from the client node list.
        clients.erase( client_node );
        node_state = NULL;
        }

    // Insert the client into "disconnected-clients" timeout cache.
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " will be able to connect again in "
            << settings.block_disconnected_client_timeout << " seconds.";
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );

        SystemTime  timeout = Global::system_time() + SystemTime( settings.block_disconnected_client_timeout );
        
        disconnected_clients.update( client_node, timeout, EmptyStruct() );
        }

    // Notify massiv system about the client's disconnection and notify
    // its account object if the local node is forward node of the client.
    GlobalCallbacks::client_disconnected( client_node, notify_account_object );
    
    // Kill connection to the client.
    Super::kill_node( client_node );
    }


void NodeManager::Server::phase_waiting_for_server_to_server_connections()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::phase_waiting_for_server_to_server_connections()" );
    
    // Wait until connections to all server nodes are established.
    // Note: Because automatic open/close connection feature can
    // be enabled, remote node manager phase PHASE_CONNECTION_ESTABLISHED
    // doesn't mean that there is opened connection, but it means that there
    // was successfully opened connection some time ago.
    if( test_remote_phase_of_server_nodes( REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED ) )
        {
        set_phase_of_server_nodes( REMOTE_SERVER_PHASE_RUNNING );

        // Unpause system message queues that were paused in connect().
        unpause_message_queues_of_server_nodes( SystemMessage::PAUSE_WAITING_FOR_INITIAL_CONNECTIONS );
        
        // Connections to all server nodes was successfully established.
        initiate_startup_node_action();

        // Switch to next phase.
        phase = PHASE_RUNNING;

        // Log.
        log_phase();
        }
    }


void NodeManager::Server::phase_running()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::phase_running()" );
    
    // Advance cache holding recently disconnected clients.
    disconnected_clients.advance( Global::system_time() );
    }


void NodeManager::Server::initiate_startup_node_action()
    {
    Global::node_action_manager().initiate_action( NodeAction::STARTUP );
    }


void NodeManager::Server::initiate_connections_to_server_nodes()
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::initiate_connections_to_server_nodes()" );
    ServerTable::iterator           it;

    for( it = servers.begin(); it != servers.end(); ++it )
        {
        const NodeId & node_id = it->first;
        ServerState & node_state = it->second;
        
        // Check if connection was already established some time ago.
        if( node_state.phase == REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED )
            {
            continue;
            }

        // Connection to local node doesn't go through network.
        if( node_id == get_local_node_id() )
            {
            node_state.phase = REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED;
            continue;
            }
        
        if( NodeId::Less()( get_local_node_id(), node_id ) )
            {
            // Log.
                {
                std::ostringstream oss;
                oss << "Initiating connection to remote server " << node_id << ".";
                Global::log_info( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            // Initiate connection to the node.
            Global::network().async_initiate_connection_to_server( node_id );
            }
        }
    }


NodeManager::ConnectionEstablishedResult NodeManager::Server::handle_connection_established_from_client
    (
    const NodeId & node_id
    )
    {
    hope( node_id.is_type( NodeId::TYPE_CLIENT ) );
    
    // Check if not yet initialized...
    if( phase != PHASE_RUNNING )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id << ". The local node "
            << "is not yet ready to accept clients.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }
    
    // Check if simuation is running...
    if( !is_connected() )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id << ". The local node "
            << "is not yet ready to accept clients.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }
    
    ClientState * node_state = get_client_state( node_id );
    if( node_state != NULL  )
        {
        // Client already connected.
        if( node_state->phase != REMOTE_CLIENT_PHASE_CONNECTED )
            {
            std::ostringstream oss;
            oss << "Denying connection to/from " << node_id
                << " because the client is trying to connect to simulation "
                << " through two or more server nodes at the same time.";
            Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );

            disconnect_client( node_id );
            return RESULT_CLOSE_CONNECTION;
            }
        else
            {
            return RESULT_OK;
            }
        }
    else
        {
        // Looks like first connection of the client (to simulation).
        SystemTime curr_time = Global::system_time();
        
        if( disconnected_clients.find( node_id, curr_time ) )
            {
            // The client was disconnected short while ago.
            // He must wait more time to connect again to the simulation.
            std::ostringstream oss;
            oss << "Denying connection to/from " << node_id
                << " because the client was disconnected recently.";
            Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
            return RESULT_CLOSE_CONNECTION;
            }
        else
            {
            ClientState         node_state;
        
            node_state.phase                    = REMOTE_CLIENT_PHASE_WAITING_FOR_ACCEPT_CLIENT_REPLIES;
            node_state.forward_node             = get_local_node_id();
            node_state.client_accepted_count    = 0; // Waiting for ClientAccepted messages.
            
            // Register the client.
            clients.insert( ClientTable::value_type( node_id, node_state ) );

            // Send AcceptClient message to all server nodes.
                {            
                AcceptClientMessage msg( node_id );
                msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
                }

            // Log.
                {
                std::ostringstream oss;
                oss << "Client " << node_id << " is trying to connect to the simulation. "
                    << "Waiting for ClientAccepted messages from all server nodes...";
                Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                }

            return RESULT_OK;
            }
        }
    }


void NodeManager::Server::log_phase()
    {
    const std::string phase_name = phase_to_string( phase );
    std::ostringstream oss;
    oss << "NodeManager::Server::phase is now " << phase_name << ".";
    Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
    }


std::string NodeManager::Server::phase_to_string
    (
    Phase phase
    )
    {
    switch( phase )
        {
        case PHASE_NONE:
            return "NONE";
        
        case PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS:
            return "WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS";
        
        case PHASE_RUNNING:
            return "RUNNING";
        }

    panic( "Invalid value of NodeManager::Server phase!" );
    return "UNKNOWN";
    }


NodeManager::Server::ClientState * NodeManager::Server::get_client_state
    (
    const NodeId & node_id
    )
    {
    ClientTable::iterator it = clients.find( node_id );
    if( it != clients.end() )
        {
        ClientState & client_state = it->second;
        return &client_state;
        }
    else
        {
        return NULL;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv

