/** @file node_manager_server.h
 * Server specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_MANAGER_SERVER_H
#define MASSIV_NODE_MANAGER_SERVER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODE_MANAGER_NODE_H
#include "node/node_manager_node.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Server encapsulates node manager's behaviour if local node is server.
 */
class NodeManager::Server : public NodeManager::Node
    {
    NO_COPY_CONSTRUCTOR( Server );
    NO_ASSIGNMENT_OPERATOR( Server );
    
    typedef NodeManager::Node Super;

public:

    /** Internal phase. */
    enum Phase
        {
        PHASE_NONE,
            /**< Initial phase. The server object has just been created but not initialized yet.
             */
        
        PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS,
            /**< The server is waiting until connections to all other server nodes
             * are established. During this phase the server nodes create crash-detection
             * links.
             */
        
        PHASE_RUNNING
            /**< The server is running. Connections to all other server nodes was
             * successfully established. Also crash-detection links among
             * server nodes was successfully established.
             */
        }; // Phase
    
    /** Remote client phase. */
    enum RemoteClientPhase
        {
        REMOTE_CLIENT_PHASE_NONE,
        REMOTE_CLIENT_PHASE_WAITING_FOR_ACCEPT_CLIENT_REPLIES,
            /**< We are waiting for ClientAcceptedMessage from all server nodes. */
        REMOTE_CLIENT_PHASE_AWAITING_CLIENT_READY_MESSAGE,
            /**< We are waiting for ClientReadyMessage from the forward node of the client. */
        REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES,
            /**< We are waiting for ClientReadyReplyMessage from all server nodes. */
        REMOTE_CLIENT_PHASE_CONNECTED,
            /**< The client node and it was successfully connected to the simulation.
             * Forward node id is properly set in the node state. */
        REMOTE_CLIENT_PHASE_DISCONNECTING
            /**< The remote node is beeing disconnected from local node.
             * Any pending messages to/from the remote node are ignored.
             */
        }; // RemoteClientPhase

    /** Client state. */
    struct ClientState
        {
    public:
        /** Constructor. */
        ClientState();
        
    public:
        RemoteClientPhase   phase;
            /**< Phase of the client node. */
        NodeId              forward_node;
            /**< Forward node of the client (always server node).
             * Forward node is used to forward messages to the client node
             * if we don't have opened direct connection to the client node
             * (and we can't initiate new one).
             * Forward node also manages the client connection to the simulation.
             */
        unsigned int        client_accepted_count;
            /**< Number of acks from server nodes. Used as counter to
             * count number of server that has replied to out request.
             *
             * If phase is REMOTE_CLIENT_PHASE_WAITING_FOR_ACCEPT_CLIENT_REPLIES,
             * the num_server_acks is used to count number of ClientAccepted messages.
             * If phase is REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES,
             * the num_server_acks is used to count number of ClientReadyReply messages.
             */
        bool                was_account_notified;
            /**< Whether a message was sent to the account object of the
             * client, notifying that the cilent was connected to the simulation.
             */
        bool                was_registered_in_time_manager;
            /**< Whether the node was registered in time manager
             * because of time synchronization.
             */

        }; // ClientState

    /** Client state table. */
    typedef ext::hash_map
        <
        NodeId,
        ClientState,
        NodeId::Hash,
        NodeId::EqualKey
        > ClientTable;

public:

    /** Constructor. */
    Server();

public:

    // From NodeManager::Node.
    virtual void connect();
    
    // From NodeManager::Node.
    virtual void disconnect();
    
    // From NodeManager::Node.
    bool is_connected();
    
    // From NodeManager::Node.
    virtual void next_tick();

    // From NodeManager::Node.
    virtual ConnectionEstablishedResult net_callback_connection_established
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_connection_closed
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_connection_lost
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual bool is_client_connected_to_simulation
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void net_callback_cannot_initiate_connection
        (
        const NodeId & node_id
        );

    // From NodeManager::Node.
    virtual void kill_node
        (
        const NodeId & other_node
        );

public:

    /**
     * Handle AcceptClientMessage.
     */
    SystemMessage::ExecuteResult handle_accept_client_message
        (
        const AcceptClientMessage & msg
        );

    /**
     * Handle ClientAcceptedMessage.
     */
    SystemMessage::ExecuteResult handle_client_accepted_message
        (
        const ClientAcceptedMessage & msg
        );

    /**
     * Handle ClientReadyMessage.
     */
    SystemMessage::ExecuteResult handle_client_ready_message
        (
        const ClientReadyMessage & msg
        );

    /**
     * Handle ClientReadyReplyMessage.
     */
    SystemMessage::ExecuteResult handle_client_ready_reply_message
        (
        const ClientReadyReplyMessage & msg
        );

    /**
     * Handle KillClientMessage.
     */
    SystemMessage::ExecuteResult handle_kill_client_message
        (
        const KillClientMessage & msg
        );

public:

    /**
     * Get forward node for specified cilent.
     * @pre The client node is connected.
     */
    NodeId get_client_forward_node
        (
        const NodeId & client_node
        );

    /**
     * Notify that simulation was loaded and is running.
     */
    void notify_simulation_is_running();

    /** Initiate connections to all server nodes. */
    void initiate_connections_to_server_nodes();


private:

    /** Disconnect client node from simulation. */
    void disconnect_client
        (
        const NodeId & other_node
        );

    /** Kill client node from local node. */
    void do_kill_client
        (
        const NodeId & client_node
        );

    /** Next tick for @c PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS phase. */
    void phase_waiting_for_server_to_server_connections();

    /** Next tick for @c PHASE_RUNNING phase. */
    void phase_running();
    
    /** Initiate @c NodeAction::STARTUP node action. */
    void initiate_startup_node_action();

    /**
     * Handle incoming connection from client node.
     */
    ConnectionEstablishedResult handle_connection_established_from_client
        (
        const NodeId & node_id
        );

    /** Log current server phase. */
    void log_phase();

    /** Convert server phase to string. */
    static std::string phase_to_string
        (
        Phase phase
        );

    /** Get client state. Return NULL if the client is not registsred in @c clients table. */
    ClientState * get_client_state
        (
        const NodeId & node_id
        );

private:

    /** Timeout cache holding node ids. */
    typedef Cache
            <
            NodeId,
            SystemTime,
            EmptyStruct,
            NodeId::Hash,
            NodeId::EqualKey,
            SystemTime::Less
            > NodeIdTimeoutCache;

private:

    Phase               phase;
        /**< Internal phase of the server. */
    
    NodeId              crash_detect_server;
        /**< Server node we have crash-detection link with.
         * There is a permanently opened connection to the server (the connection
         * has disabled auto-close feature). If the connection is lost,
         * we know that the server or the connection has crashed.
         */

    ClientTable         clients;
        /**< Client table. The table contains all client nodes connected
         * to simulation (even if the local node doesn't have opened connections
         * to the client nodes).
         */

    NodeIdTimeoutCache  disconnected_clients;
        /**< Timeout cache used to store node ids of some client nodes.
         * This cache is used to block incoming connections of the clients
         * which have been disconnected from the simulation a short
         * while ago.
         * Once a client is disconnected from the simulation, it is inserted
         * into this cache. 
         * The purpose of this cache is to give some time to the massiv
         * system to recover from the disconnection of the client
         * (the recovering includes processing of pending client-related
         * messages etc.).
         *
         * @note This cache is used to block only initial client connections
         *  ie. this cache will not block incoming connection if the
         *  client has its node state in @c client_nodes table.
         *  This will ensure that a client accepted by one server node
         *  will be accepted by all other server nodes even if they have
         *  the client in their @c disconnected_clients cache.
         */
    
    bool                simulation_ready;
        /**< True if simulation was loaded and is running. */

    }; // NodeManager::Server

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_MANAGER_SERVER_H
