/** @file node_manager_server_callbacks.cpp
 * Server specialization of NodeManager::Node.
 */
 
/*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_server.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"


namespace Massiv {
namespace Core {


NodeManager::ConnectionEstablishedResult NodeManager::Server::net_callback_connection_established
    (
    const NodeId & node_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::NodeManager::Server::net_callback_connection_established()" );
    
    // Check if not yet initialized...
    if( phase == PHASE_NONE )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id << ". The local node "
            << "is not yet ready.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }
    
    // Check anonymous node.
    if( node_id.is_type( NodeId::TYPE_ANONYMOUS ) )
        {
        std::ostringstream oss;
        oss << "Denying connection to/from " << node_id
            << " because anonymous nodes can't connect to server nodes.";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
        return RESULT_CLOSE_CONNECTION;
        }
    
    // Check service.
    if( node_id.is_type( NodeId::TYPE_SERVICE ) )
        {
        return RESULT_OK;
        }
    
    // Check client.
    if( node_id.is_type( NodeId::TYPE_CLIENT ) )
        {
        return handle_connection_established_from_client( node_id );
        }
    
    // Check server.
    hope( node_id.is_type( NodeId::TYPE_SERVER ) );
    hope( get_server_state( node_id ) != NULL ); // Server nodes already in the server table.
    
    if( phase == PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS )
        {
        // Log.
            {
            std::ostringstream oss;
            oss << "Handling connection establishes with " << node_id
                << " in WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS phase.";
            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
            }
            
        // We are waiting for connections from all server nodes.
        // Automatic open/close connection feature can be enabled.
        // Only connections to server and service nodes are permitted.
        
        // Check if this is the first connection to a (remote) server.
        bool first_connection = true;
        
            { // sob
            ServerTable::const_iterator it;
            for( it = servers.begin(); it != servers.end(); ++it )
                {
                const NodeId & server_id = it->first;
                const ServerState & state = it->second;
                if( server_id != get_local_node_id() && state.phase != REMOTE_SERVER_PHASE_NONE )
                    {
                    // There was already a connection established some time ago.
                    first_connection = false;
                    break;
                    }
                }
            } // eob

        // Log.
        if( first_connection )
            {
            std::ostringstream oss;
            oss << "Connection with " << node_id << " is the first connection to remote servers.";
            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
            }
        
        // Update server state of the server.
            {
            ServerState * state = get_server_state( node_id );
            hope( state );
            
            // Note: Automatic open/close connection can be enabled, so the phase of the remote node
            // can already be REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED.
            hope
                (
                state->phase == REMOTE_SERVER_PHASE_NONE ||
                state->phase == REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED
                );

            state->phase = REMOTE_SERVER_PHASE_CONNECTION_ESTABLISHED;
            }
            
        // Disable auto-close connection feature for the first server we have connection to.
        // This will create graph of connections which can't be automatically closed among server
        // nodes. If any server node crashes at least one another server node will detect it.
        if( first_connection )
            {
            hope( crash_detect_server == NodeId() );
            std::ostringstream oss;
            oss  << "Disabling auto-close feature for " << node_id << ".";
            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
                    
            Network & network = Global::network();
            network.control_connection_auto_close( node_id, false );
            crash_detect_server = node_id;
            }

        return RESULT_OK;
        }
    else
        {
        hope( phase == PHASE_RUNNING );
        return RESULT_OK;
        }
    }


void NodeManager::Server::net_callback_connection_closed
    (
    const NodeId & node_id
    )
    {
    switch( node_id.get_type() )
        {
        case NodeId::TYPE_CLIENT:
            {
            // If local node is forward node of the client then disconnect
            // the client from simulation.
            ClientState * state = get_client_state( node_id );
            if( state && state->forward_node == get_local_node_id() )
                {
                disconnect_client( node_id );
                }
            else
                {
                // We do NOT remove the client's node state because
                // the client is not disconnected from simulation, but
                // only from our server.
                }
            break;
            }
        
        case NodeId::TYPE_SERVER:
            {
            // Auto-close feature closed the connection.
            if( crash_detect_server == node_id )
                {
                std::ostringstream oss;
                oss << "Closed connection with auto-close-disabled to "
                    << node_id
                    << ". Servers are going down. Initiating local node shutdown.";
                Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGH, oss.str() );
                
                // FIXME: 
                // Initiate local node quit.
                GlobalCallbacks::shutdown_local_node();
                }
            break;
            }
        
        case NodeId::TYPE_SERVICE:
        case NodeId::TYPE_ANONYMOUS:
            // Notify core.
            GlobalCallbacks::node_disconnected( node_id );
            break;
        }
    }


void NodeManager::Server::net_callback_connection_lost
    (
    const NodeId & node_id
    )
    {
    switch( node_id.get_type() )
        {
        case NodeId::TYPE_CLIENT:
            disconnect_client( node_id );
            break;
        
        case NodeId::TYPE_SERVER:
            throw_connection_lost( node_id );
            break;
        
        case NodeId::TYPE_SERVICE:
        case NodeId::TYPE_ANONYMOUS:
            // Notify core.
            GlobalCallbacks::node_disconnected( node_id );
            break;
        }
    }


void NodeManager::Server::net_callback_cannot_initiate_connection
    (
    const NodeId & node_id
    )
    {
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        switch( phase )
            {
            case PHASE_NONE:
                break;
            case PHASE_WAITING_FOR_SERVER_TO_SERVER_CONNECTIONS:
                // We are testing whether other servers are running or not.
                // If not, just wait for them.
                break;
            case PHASE_RUNNING:
                throw_connection_failed( node_id );
                break;
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv

