/** @file node_manager_server_messages.cpp
 * Node manager message handling for server nodes.
 */

 /*
 * begin        : Sat Nov 8 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "node/node_manager_server.h"
#include "global/global.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/system_msg.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "net/net.h"


namespace Massiv {
namespace Core {


//------------------------------------------------------------------------------
// Server to server client connection/disconnection communication.
//------------------------------------------------------------------------------
    

SystemMessage::ExecuteResult NodeManager::Server::handle_accept_client_message
    (
    const AcceptClientMessage & msg
    )
    {
    const NodeId &  msg_sender  = msg.get_sender();
    const NodeId &  client_node = msg.get_client_node_id();

    hope( msg.get_initiator() == msg_sender );          // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );  // Already verified.

    /** @todo Message verification. */
    hope( !client_node.is_type( NodeId::TYPE_SERVER ) );
    
    ClientState * node_state = get_client_state( client_node );

    if( node_state )
        {
        // The client is already registered in our "client_nodes" database.
        switch( node_state->phase )
            {
            case REMOTE_CLIENT_PHASE_WAITING_FOR_ACCEPT_CLIENT_REPLIES:
                {
                // We have proclaimed us to be the forward node of the client.
                hope( node_state->forward_node == get_local_node_id() );
                    
                if( msg_sender == get_local_node_id() )
                    {
                    // We are the one who sent the message.
                    // Send reply to ourself.
                    ClientAcceptedMessage ca_msg( client_node );
                    ca_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
                    }
                else                    
                    {
                    // Another server node is trying to proclaim itself to be
                    // the forward node of the client.
                    // This means that the client is connecting to two
                    // server nodes at one time.
                    // Bad client - kill it.
                    disconnect_client( client_node );
                    }
                }
                break;
                    
            case REMOTE_CLIENT_PHASE_AWAITING_CLIENT_READY_MESSAGE:
                {
                // AcceptClientMessage received from a node but we have already
                // received AcceptClientMessage from some node.
                // This means "not-good" client, which is connecting to two
                // server nodes at one time.
                // Bad client - kill it.
                disconnect_client( client_node );
                }
                break;
                
            case REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES:
            case REMOTE_CLIENT_PHASE_CONNECTED:
                {
                // This is internal error! Should never happen.
                Global::log_warning
                    (
                    Status::FACILITY_NODE_MANAGEMENT,
                    Status::PRIORITY_HIGHEST,
                    err_unexpected_message( msg )
                    );
                    
                // Kill the client.
                disconnect_client( client_node );
                }
                break;
                
            case REMOTE_CLIENT_PHASE_DISCONNECTING:
                // The node is already beeing disconnected.
                // We have just send kill client message to all server nodes.
                // Nothing to do.
                break;

            default:
                panic( "Invalid value of client node state phase." );
                    
                // Kill the client.
                disconnect_client( client_node );
                break;
            }
        }
    else
        {
        // We don't have the client in our 'client_nodes' database.
        ClientState node_state;

        node_state.phase                    = REMOTE_CLIENT_PHASE_AWAITING_CLIENT_READY_MESSAGE;
        node_state.forward_node             = msg_sender;
        node_state.client_accepted_count    = 0;

        clients.insert( ClientTable::value_type( client_node, node_state ) );

        // Send ClientAccepted message to the forward node of the client.
            {
            ClientAcceptedMessage ca_msg( client_node );
            ca_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
            }
        }

    return SystemMessage::EXECUTE_OK;
    }


SystemMessage::ExecuteResult NodeManager::Server::handle_client_accepted_message
    (
    const ClientAcceptedMessage & msg
    )
    {
    DEBUG_ONLY( const NodeId &  msg_sender  = msg.get_sender(); )
    const NodeId &  client_node = msg.get_client_node_id();

    hope( msg.get_initiator() == msg_sender );          // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );  // Already verified.

    /** @todo Message verification. */
    hope( !client_node.is_type( NodeId::TYPE_SERVER ) );

    ClientState * node_state = get_client_state( client_node );
    if( node_state )
        {
        if( node_state->phase == REMOTE_CLIENT_PHASE_WAITING_FOR_ACCEPT_CLIENT_REPLIES )
            {
            hope( node_state->forward_node == get_local_node_id() );
            node_state->client_accepted_count++;
            hope( node_state->client_accepted_count <= servers.size() );

            // Check if we have received ClientAcceptedMessage from all server nodes.
            if( node_state->client_accepted_count >= servers.size() )
                {
                // All server nodes has accepted the client.
                node_state->phase = REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES;
                node_state->client_accepted_count = 0;

                // Send ClientReady message to all server nodes (including
                // to ourself).
                    {
                    ClientReadyMessage cr_msg( client_node );
                    cr_msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
                    }
                }
            }
        else
            {
            // Unexpected message. We are not forward node of the client any more.
            Global::log_warning
                (
                Status::FACILITY_NODE_MANAGEMENT,
                Status::PRIORITY_HIGHEST,
                err_unexpected_message( msg )
                );
                
            disconnect_client( client_node );
            }
        }
    else
        {
        // We don't have the client in client_nodes database
        // and we have received ClientAccepted message.
        // This means that the client was disconnected between
        // AcceptClient and ClientAccepted messages.
        
        disconnect_client( client_node );
        }
    
    return SystemMessage::EXECUTE_OK;
    }


SystemMessage::ExecuteResult NodeManager::Server::handle_client_ready_message
    (
    const ClientReadyMessage & msg
    )
    {
    const NodeId &  msg_sender  = msg.get_sender();
    const NodeId &  client_node = msg.get_client_node_id();

    hope( msg.get_initiator() == msg_sender );          // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );  // Already verified.

    /** @todo Message verification. */
    hope( !client_node.is_type( NodeId::TYPE_SERVER ) );

        {//sob
        ClientState * node_state = get_client_state( client_node );
        if( node_state )
            {
            switch( node_state->phase )
                {
                case REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES:
                    {
                    if( msg_sender == get_local_node_id() )
                        {
                        hope( node_state->forward_node == get_local_node_id() );
                        
                        // We are the sender.
                        // Send reply to ourself.
                        ClientReadyReplyMessage crr_msg( client_node );
                        crr_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
                        
                        // Do not call GlobalCallbacks::client_connected() here,
                        // it will be called in handle_client_ready_reply_message()
                        // because local node is forward node of the client.
                        }
                    else
                        {
                        disconnect_client( client_node );
                        }
                    }
                    break;

                case REMOTE_CLIENT_PHASE_AWAITING_CLIENT_READY_MESSAGE:
                    {
                    if( msg_sender == node_state->forward_node )
                        {
                        // Forward should not be local node. Forward node
                        // should have node_state->phase equal to
                        // REMOTE_CLIENT_PHASE_AWAITING_CLIENT_READY_REPLY_MESSAGE
                        // (see few lines above).
                        hope( msg_sender != get_local_node_id() );
                        
                        node_state->phase = REMOTE_CLIENT_PHASE_CONNECTED;
    
                        // Send reply to the forward node.
                            {
                            ClientReadyReplyMessage crr_msg( client_node );
                            crr_msg.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
                            }

                        // Log.
                            {
                            std::ostringstream ss;
                            ss  << "Client " << client_node << " proclaimed \"connected\" on local node "
                                << get_local_node_id() << ".";
                                
                            Global::log_debug
                                (
                                Status::FACILITY_NODE_MANAGEMENT,
                                Status::PRIORITY_NORMAL,
                                ss.str()
                                );
                            }
                            
                        // Notify local system.
                        // Local system is not forward node of the client,
                        // see few lines above.
                        GlobalCallbacks::client_connected( client_node, false );
                        }
                    else
                        {
                        disconnect_client( client_node );
                        }
                    }
                    break;

                default:
                    {
                    Global::log_warning
                        (
                        Status::FACILITY_NODE_MANAGEMENT,
                        Status::PRIORITY_HIGHEST,
                        err_unexpected_message( msg )
                        );
                    
                    disconnect_client( client_node );
                    }
                    break;
                }
            }
        else
            {
            disconnect_client( client_node );
            }
        }//eob
    
    return SystemMessage::EXECUTE_OK;
    }


SystemMessage::ExecuteResult NodeManager::Server::handle_client_ready_reply_message
    (
    const ClientReadyReplyMessage & msg
    )
    {
    DEBUG_ONLY( const NodeId &  msg_sender  = msg.get_sender(); )
    const NodeId &  client_node = msg.get_client_node_id();

    hope( msg.get_initiator() == msg_sender );          // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );  // Already verified.

    /** @todo Message verification. */
    hope( !client_node.is_type( NodeId::TYPE_SERVER ) );

        {//sob
        ClientState * node_state = get_client_state( client_node );
        if( node_state )
            {
            if( node_state->phase == REMOTE_CLIENT_PHASE_WAITING_FOR_CLIENT_READY_REPLIES )
                {
                hope( node_state->forward_node == get_local_node_id() );
                node_state->client_accepted_count++;
                hope( node_state->client_accepted_count <= servers.size() );

                if( node_state->client_accepted_count >= servers.size() )
                    {
                    // All server nodes has accepted the client.
                    node_state->phase = REMOTE_CLIENT_PHASE_CONNECTED;
                    node_state->client_accepted_count = 0;

                    // Log.
                        {
                        std::ostringstream ss;
                        ss  << "Client " << client_node << " connected to the simulation. "
                            << "Sending welcome message...";

                        Global::log_info
                            (
                            Status::FACILITY_NODE_MANAGEMENT,
                            Status::PRIORITY_NORMAL,
                            ss.str()
                            );
                        }
                    
                    // Disable auto-close connection feature for this client because local node
                    // is the forward node for the client.
                    // Because of this we will be able to detect the client crash and of course
                    // the forward node for any client must have all-time connection to the client
                    // to be able to send forward messages.
                        {
                        Network & network = Global::network();
                        network.control_connection_auto_close( client_node, false );
                        }
                    
                    // Send welcome message to the client.
                        {
                        WelcomeMessage  msg;
                        msg.send( client_node, NetworkCommon::STREAM_RELIABLE );
                        }
                        
                    // Register the client to the time manager.
                        {
                        TimeManager & time_manager = Global::time_manager();
                        time_manager.add_node_for_synchronization( client_node );
                        hope( !node_state->was_registered_in_time_manager );
                        node_state->was_registered_in_time_manager = true;
                        }
                        
                    // Notify client account object.
                        {
                        GlobalCallbacks::client_connected( client_node, true );
                        hope( !node_state->was_account_notified );
                        node_state->was_account_notified = true;
                        }
                    }
                }
            else
                {
                // Unexpected message. We are not forward node of the client any more.
                disconnect_client( client_node );
                }
            }
        else
            {
            disconnect_client( client_node );
            }
        }//eob
        
    return SystemMessage::EXECUTE_OK;
    }


SystemMessage::ExecuteResult NodeManager::Server::handle_kill_client_message
    (
    const KillClientMessage & msg
    )
    {
    DEBUG_ONLY( const NodeId &  msg_sender  = msg.get_sender(); )
    const NodeId & client_node = msg.get_client_node_id();

    hope( msg.get_initiator() == msg_sender );          // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );  // Already verified.

    do_kill_client( client_node );

    return SystemMessage::EXECUTE_OK;
    }


} // namespace Massiv::Core
} // namespace Massiv
