/** @file nodeid.cpp
 * Node id (serializable).
 */

/*
 * begin        : Thu Apr 11 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "node/nodeid.h"
#include "status/exception.h"
#include "serializer/serializer.h"
#include "serializer/stype_int.h"
#include "serializer/stype_string.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "serializer/stype_int.h"
#include <sstream>
#include <ostream>


namespace Massiv {
namespace Core {


NodeId NodeId::create_from_type_and_relative_id
    (
    UInt32   type,
    UInt32   relative_id
    )
    {
    NodeId  new_node_id;
    new_node_id.set_from_type_and_relative_id( type, relative_id );
    return new_node_id;
    }

    
NodeId NodeId::create_from_type_and_relative_id
    (
    const std::string & type_as_string,
    UInt32              relative_id
    )
    {
    UInt32 type;
    
    if( type_as_string == "Server" )
        {
        type = TYPE_SERVER;
        }
    else if( type_as_string == "Client" )
        {
        type = TYPE_CLIENT;
        }
    else if( type_as_string == "Service" )
        {
        type = TYPE_SERVICE;
        }
    else if( type_as_string == "Anonymous" )
        {
        type = TYPE_ANONYMOUS;
        }
    else
        {
        std::ostringstream ss;
        ss  << "Invalid NodeId type specified (" << type_as_string << ").";
        throw ExInvalidNodeId( ss.str() );
        }
    
    NodeId  new_node_id;
    new_node_id.set_from_type_and_relative_id( type, relative_id );
    return new_node_id;
    }


std::string NodeId::type_to_string
    (
    Type type
    )
    {
    switch( type )
        {
        case TYPE_SERVER:
            return "Server";

        case TYPE_CLIENT:
            return "Client";

        case TYPE_SERVICE:
            return "Service";
        
        case TYPE_ANONYMOUS:
            return "Anonymous";
        }

    return "(unknown node type)";
    }


std::string NodeId::to_string() const
    {
    std::ostringstream ss;
    ss << *this;
    return ss.str();
    }

    
std::string NodeId::to_identifier() const
    {
    if( operator==( NodeId() ) )
        {
        return "Unknown";
        }
    else
        {
        std::ostringstream ss;
        ss << type_to_string( get_type() ) << components.relative_id;
        return ss.str();
        }
    }


void NodeId::set_from_type_and_relative_id
    (
    UInt32   type,
    UInt32   relative_id
    )
    {
    // We must clear all 32 bits!!!
    clear();

    // Check validity.
    if( type < 0 || type >= TYPE_COUNT )
        {
        std::ostringstream ss;
        ss  << "Invalid NodeId type specified (" << type << ").";
        throw ExInvalidNodeId( ss.str() );
        }

    if( ( relative_id >> RELATIVE_ID_BITS_COUNT ) != 0 )
        {
        std::ostringstream ss;
        ss  << "Invalid NodeId relative id specified (" << relative_id << ").";
        throw ExInvalidNodeId( ss.str() );
        }

    components.type = type;
    components.relative_id = relative_id;
    }


void NodeId::bitstr_write
    (
    BitOStream &                        bos,
    const Serializer::Description &     desc
    ) const
    {
    SInt<UInt32, TYPE_BITS_COUNT>        type( components.type );
    SInt<UInt32, RELATIVE_ID_BITS_COUNT> relative_id( components.relative_id );
                                                                           
    type.bitstr_write( bos, desc );
    relative_id.bitstr_write( bos, desc );
    }


void NodeId::bitstr_read
    (
    BitIStream &                        bis,
    const Serializer::Description &     desc
    )
    {
    SInt<UInt32, TYPE_BITS_COUNT>        type;
    SInt<UInt32, RELATIVE_ID_BITS_COUNT> relative_id;

    type.bitstr_read( bis, desc );
    relative_id.bitstr_read( bis, desc );

    /** @todo FIXME Throw serialization exception. */
    set_from_type_and_relative_id( type, relative_id );
    }
    
    
void NodeId::text_write
    (
    TextWriter &                        tw,
    const Serializer::Description &     desc
    ) const
    {
    SInt< UInt32, RELATIVE_ID_BITS_COUNT > relative_id( components.relative_id );
    std::string type = type_to_string( get_type() );
    
    tw.line() << "[ ";
    
    if( operator==( NodeId() ) )
        {
        tw.line() << "Unknown";
        }
    else
        {
        tw.line() << type;
        tw.write_space();
        relative_id.text_write( tw, desc );
        }
    
    tw.line() << " ]";
    }


void NodeId::text_read
    (
    TextReader &                        tr,
    const Serializer::Description &     desc
    )
    {
    Type    type;
    SString look_ahead;
    SInt< UInt32, RELATIVE_ID_BITS_COUNT > relative_id;
    
    tr.require_string( "[" );
    
    // Read type.
    look_ahead.text_read( tr, desc );
    
    if( look_ahead == "Server" || look_ahead == "Unknown" )
        {
        type = TYPE_SERVER;
        }
    else if( look_ahead == "Client" )
        {
        type = TYPE_CLIENT;
        }
    else if( look_ahead == "Service" )
        {
        type = TYPE_SERVICE;
        }
    else if( look_ahead == "Anonymous" )
        {
        type = TYPE_ANONYMOUS;
        }
    else
        {
        throw ExUnexpectedValue( tr.unexpected_value( "'node type'", look_ahead ) );
        }
    
    if( look_ahead == "Unknown" )
        {
        relative_id = 0;
        }
    else
        {
        // Read relative id.
        relative_id.text_read( tr, desc );
        }
    
    components.type        = type;
    components.relative_id = relative_id;
    
    tr.require_string( "]" );
    }


std::ostream & operator<<
    (
    std::ostream &  s,
    const NodeId &  node_id
    )
    {
    Serializer::Description desc;
    TextWriter              tw( s );
    
    node_id.text_write( tw, desc );
    return s;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "node/nodeid_inline.h"
#endif
