/** @file nodeid.h
 * Node id (serializable).
 */

/*
 * begin        : Thu Apr 11 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_NODEID_H
#define MASSIV_NODEID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#include <string>
#include <iosfwd>


namespace Massiv {
namespace Core {


/** @addtogroup node
 * @{
 */

/**
 * Class that uniquely identifies node within the entire system.
 * NodeId consists of two parts - type and relative id.
 * Type is type of the node. Relative id is just a number.
 * NodeId is serializable.
 */
class MASSIV_EXPORTED NodeId
    {    
    friend class ObjectId;
    
public:

    /**
     * Node type.
     * @note Do not forget to update @c NodeId::type_to_name() and @c TYPE_BITS_COUNT if @c Type is updated.
     */
    enum Type
        {
        TYPE_SERVER         = 0,
            /**< Simulation server. Simulation server nodes perform simulation,
             * manages objects, perform archivation, etc. */
        TYPE_CLIENT         = 1,
            /**< Simulation client. Node that can connect to simulation
             * simulated by server nodes. Each client has its account object
             * in the simulation that is used as gateway between the client
             * and the simulation.
             */
        TYPE_SERVICE        = 2,
            /**< Service node. Service node provides some peer to peer services
             * like data download.
             */
        TYPE_ANONYMOUS      = 3
            /**< Anonymous node. Anonymous nodes have restricted rights.
             * They can only connect to service nodes. Thay can't connect
             * to simulation.
             * Anonymous nodes are ususally used to connect to data download
             * service and to download minimal required data.
             */
        }; // Type

    enum
        {
        TYPE_COUNT              = 4,
            /**< Number of node types. */
        TYPE_BITS_COUNT         = 2,
            /**< Number of bits required to store node type. */
        RELATIVE_ID_BITS_COUNT  = 20
            /**< Number of bits required to store relative node id.
             * Must be >= 1 and <= 32.
             */
        };

public:

    /**
     * NodeId hash function.
     */
    struct Hash
        {

        /** Hash function. */
        INLINE size_t operator()
            (
            const NodeId & node_id
            ) const throw();

        }; // struct Hash

    /**
     * NodeId equalkey function.
     */
    struct EqualKey
        {

        /** Equalkey function. */
        INLINE bool operator()
            (
            const NodeId & node_id1,
            const NodeId & node_id2
            ) const throw();

        }; // struct EqualKey

    /**
     * NodeId less function.
     * @note This function is compiler/architecture independent.
     */
    struct Less
        {

        /** Less function. */
        INLINE bool operator()
            (
            const NodeId & node_id1,
            const NodeId & node_id2
            ) const throw();

        }; // struct Less

    friend struct NodeId::Hash;
    friend struct NodeId::EqualKey;
    friend struct NodeId::Less;

public:

    /** Constructor. Constructs null node id (with @c TYPE_SERVER type). */
    INLINE NodeId();

    /**
     * Create NodeId from type and relative id numbers.
     * @throw ExInvalidNodeId Specified arguments are not valid NodeId components.
     */
    static NodeId create_from_type_and_relative_id
        (
        UInt32   type,
        UInt32   relative_id
        );

    /**
     * Create NodeId from type and relative id numbers.
     * @throw ExInvalidNodeId Specified arguments are not valid NodeId components.
     */
    static NodeId create_from_type_and_relative_id
        (
        const std::string & type_as_string,
        UInt32              relative_id
        );

    /** Copy constructor. */
    INLINE NodeId
        (
        const NodeId & node_id
        );

    /** Operator=. */
    INLINE NodeId & operator=
        (
        const NodeId & node_id
        );

    /** Operator==. */
    INLINE bool operator==
        (
        const NodeId & node_id
        ) const;

    /** Operator!=. */
    INLINE bool operator!=
        (
        const NodeId & node_id
        ) const;

    /**
     * Clear content of the id. Type is set to @c TYPE_SERVER
     * and relative id is set to zero.
     */
    INLINE void clear();

    /**
     * Check whether the node id is cleared.
     */
    INLINE bool is_null() const;

    /** Get node type. */
    INLINE Type get_type() const;

    /** Get node relative id. */
    INLINE UInt32 get_relative_id() const;

    /** Check whether the node id is specified @c type. */
    INLINE bool is_type
        (
        Type    type
        ) const;

    /** Convert node type to string. */
    static std::string type_to_string
        (
        Type    type
        );
    
    /**
     * Convert NodeId to string.
     * Outputs plain textual serialization of the id.
     */
    std::string to_string() const;
    
    /**
     * Convert NodeId to a C-style identifier, ie. no whitespaces will
     * be written.
     */
    std::string to_identifier() const;
    
public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &                        tw,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                        tr,
        const Serializer::Description &     desc
        );

public:

    /**
     * Return node id as 32-bit integer.
     * @warning This function for network purposes only. Do not use.
     */
    static INLINE UInt32 convert_to_uint32
        (
        const NodeId & node_id
        );

    /**
     * Create node id from 32-bit integer.
     * @warning This function for network purposes only. Do not use.
     */
    static INLINE NodeId create_from_uint32
        (
        UInt32 node_id_as_number
        );
        
private:

    /**
     * Set new NodeId content from type and relative id numbers.
     * @throw ExInvalidNodeId Specified arguments are not valid NodeId components.
     */
    void set_from_type_and_relative_id
        (
        UInt32   type,
        UInt32   relative_id
        );

public:

    /**
     * Write node id to output stream.
     */
    friend std::ostream & operator<<
        (
        std::ostream &  s,
        const NodeId &  node_id
        );

private:

    union
        {
        struct
            {
            UInt32       type : TYPE_BITS_COUNT;
                /**< Type of the node. */
            UInt32       relative_id : RELATIVE_ID_BITS_COUNT;
                /**< Relative id. */
            } components;
            /**< Identifier components. */
            
        UInt32           id;
            /**< Identifier as a number. */
        };

    }; // NodeId

/** @} */ // addtogroup node


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#ifndef MASSIV_NODEID_INLINE_H
#include "node/nodeid_inline.h"
#endif

#endif


#endif // MASSIV_NODEID_H
