/** @file archivation_id.h
 * Archivation unique identification.
 */

/*
 * begin        : Sun Oct 20 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_ARCHIVATION_ID_H
#define MASSIV_ARCHIVATION_ID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Archivation unique identification.
 *
 * @todo Dodelat - jak identifikovat archivace? Nejaky
 * prevod readable_string <=> int? Nebo jenom string?
 * Nebo string + poradove cislo?
 * Prevod id na jmeno souboru a zpet?
 *
 * @ingroup archivation
 */
class ArchivationId
    {
public:

    /**
     * Constructor.
     * The archivation id is set to invalid.
     */
    ArchivationId();

    /**
     * Copy constructor.
     */
    ArchivationId
        (
        const ArchivationId & archivation_id
        );

    /**
     * Operator=.
     */
    ArchivationId & operator=
        (
        const ArchivationId & arch_id
        );

    /**
     * Operator==.
     */
    bool operator==
        (
        const ArchivationId & arch_id
        ) const;

    /**
     * Operator!=.
     */
    bool operator!=
        (
        const ArchivationId & arch_id
        ) const;

    /**
     * Operator<.
     */
    bool operator<
        (
        const ArchivationId & arch_id
        ) const;

    /**
     * Operator<=.
     */
    bool operator<=
        (
        const ArchivationId & arch_id
        ) const;

    /**
     * Operator>.
     */
    bool operator>
        (
        const ArchivationId & arch_id
        ) const;

    /**
     * Operator>=.
     */
    bool operator>=
        (
        const ArchivationId & arch_id
        ) const;

public:

    /**
     * Make the archivation id invalid.
     */
    void invalidate();

    /**
     * Check whether the archivation id is invalid.
     */
    bool is_invalid() const;

    /**
     * Advance the archivation id to its next stage.
     */
    void increment();

    /**
     * Get order of the archivation id among all archivation ids.
     * The order is counted from zero (for invalid archivation id).
     */
    UInt32 get_order() const;

    /**
     * Get the archivation id as a string.
     * You can use this string as an archivation file name etc.
     * @return The archivation id as a string.
     */
    std::string to_string() const;

    /**
     * Create ArchivationId from a string.
     * @return Invalid ArchivationId if the @a name
     *         doesn't correspond to any valid ArchivationId.
     */
    static ArchivationId from_string
        (
        const std::string & name
        );

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

private:

    UInt32  id;
        /**< Archivation id as a number. */

    }; // ArchivationId


/**
 * Write ArchivationId to output stream.
 */
std::ostream & operator<<
    (
    std::ostream &          s,
    const ArchivationId &   arch_id
    );

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVATION_ID_H
