/** @file archive.cpp
 * Archive file.
 *
 * @todo !!!!!!!!!!!! PAKLIZE JE V ARCHIVU MIGRACNI SKUPINA BEZ GC ROOT OBJEKTU
 * MUZE SE SMAZAT, PROTOZE KDYBY BYLA V PAMETI, GC JI SMAZE !!!!!!!!!!!
 * Patrne kdyz ArchiveManager validuje (zavira) migracni skupiny, muze tohle
 * trivialne overit.
 */

/*
 * begin        : Fri Aug 23 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/archive.h"
#include "object/archive_internal.h"
#include "global/global.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "serializer/serializer.h"
#include "serializer/stype_int.h"
#include "status/status.h"
#include <sstream>


namespace Massiv {
namespace Core {


static const char g_archive_state_stream_name[] = "archive_state";


//------------------------------------------------------------------------------
    
// Implementation of ArchiveFile::InputStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ArchiveFile::InputStream::InputStream()
    {
    }

    
bool ArchiveFile::InputStream::open
    (
    ArchiveFile *       new_archive,
    const std::string & stream_name
    )
    {
    return Super::open( *new_archive, stream_name );
    }
    
    
bool ArchiveFile::InputStream::open
    (
    ArchiveFile *       new_archive,
    GroupId             group_id,
    const ObjectId &    object_id
    )
    {
    return open( new_archive, object_id_to_stream_name( group_id, object_id ) );
    }

    
// Implementation of ArchiveFile::OutputStream
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ArchiveFile::OutputStream::OutputStream() :
    archive( NULL ),
    linked( false )
    {    
    }


ArchiveFile::OutputStream::~OutputStream()
    {
    // Unlink the stream from its archive.
    if( linked )
        {
        hope( archive != NULL );
        hope( archive->linked_output_stream_count > 0 );
        archive->linked_output_stream_count--;
        linked = false;
        }
    }


bool ArchiveFile::OutputStream::open
    (
    ArchiveFile *       new_archive,
    const std::string & stream_name
    )
    {
    hope( new_archive != NULL );

    bool    failed;
    
    // Check if the output stream may be opened now.
    failed = !new_archive->can_open_output_stream();

    // Set fail bit if no output stream may be opened.
    if( failed )
        {
        this->setstate( ios_base::failbit );
        
        // We stays linked to the original archive.
        return false;
        }
    
    // Try open the stream.
    failed = !Super::open( *new_archive, stream_name );
    
    if( !failed )
        {
        // Unlink the original stream.
        if( linked )
            {
            hope( archive != NULL );
            hope( archive->linked_output_stream_count > 0 );
            archive->linked_output_stream_count--;
            linked = false;
            }
            
        // Relink to new_archive.
        archive = new_archive;
        archive->linked_output_stream_count++;
        linked = true;
        return true;
        }
    else
        {
        // Error. We stay linked to the original archive. The stream
        // may be closed now, but we don't mind.
        // Nothing to do.
        return false;
        }
    }


bool ArchiveFile::OutputStream::open
    (
    ArchiveFile *       new_archive,
    GroupId             group_id,
    const ObjectId &    object_id
    )
    {
    return open( new_archive, object_id_to_stream_name( group_id, object_id ) );
    }

    
//------------------------------------------------------------------------------

// Implementation of ArchiveFile
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


ArchiveFile::ArchiveFile
    (
    const std::string &     file_name,
    std::ios_base::openmode mode
    ) :
    Super( file_name, mode ),
    linked_output_stream_count( 0 )
    {
    }


ArchiveFile::ArchiveFile
    (
    const std::string &     file_name,
    const std::string &     archive_name
    ) :
    Super( file_name, archive_name ),
    linked_output_stream_count( 0 )
    {
    }


ArchiveFile::~ArchiveFile()
    {
    hope( linked_output_stream_count == 0 );
    }


std::string ArchiveFile::object_id_to_stream_name
    (
    GroupId          group_id,
    const ObjectId & object_id
    )
    {
    std::ostringstream  ss;
    
    // Group.
    ss << 'g' << group_id << '/' << object_id_to_plain_stream_name( object_id );
    
    return ss.str();
    }


std::string ArchiveFile::object_id_to_plain_stream_name
    (
    const ObjectId &    object_id
    )
    {
    // Because output of ObjectId::text_write() depends whether
    // ObjectInfo part of the ObjectId is availabe or not,
    // we must write ObjectId in a normalized way. We create copy of the
    // object_id and clear its optional part.
    std::ostringstream  ss;
    TextWriter          tw( ss );
    ObjectId            object_id_dup( object_id );
    
    // Clear optional part.
    object_id_dup.clear_optional_part();
    
    // Prefix.
    ss << 'o';
    
    // ObjectId content.
    object_id_dup.text_write( tw, Serializer::Description() );
    
    return ss.str();
    }


std::string ArchiveFile::group_id_to_stream_name
    (
    GroupId group_id
    )
    {
    std::ostringstream  ss;
    ss << 'g' << group_id << "/group_state";
    return ss.str();
    }


//------------------------------------------------------------------------------

// Implementation of Archive
// ~~~~~~~~~~~~~~~~~~~~~~~~~

Archive::Archive
    (
    const std::string & file_name,
    Mode                mode
    ) :
    Super( file_name, std::ios_base::in |
        ( mode == MODE_READ_ONLY ? std::ios_base::openmode( 0 ) : std::ios_base::out ) ),
    my( new My( mode ) )
    {
    my->archive = this;

    // Read information about objects stored in the archive.
    InputStream    state_stream;
    
    state_stream.open( this, g_archive_state_stream_name );

    if( !state_stream.good() )
        {
        std::ostringstream ss;
        ss << "Unable to open archive state stream. "
           << "The archive was not properly saved (validated). ";
        throw ExCorruptedArchive( ss.str() );
        }

    my->read_archive_state( state_stream );

    if( mode == MODE_READ_ONLY )
        {
        hope( my->read_only );
        }

    my->valid = true;
    }


Archive::Archive
    (
    const std::string & file_name,
    const std::string & archive_name
    ) :
    Super( file_name, archive_name ),
    my( new My( MODE_READ_WRITE ) )
    {
    my->archive = this;
    }


Archive::~Archive()
    {
    // If the archive is not valid or there are some opened groups,
    // damage the archive volume so it can't be opened again.
    if( !my->valid || !my->opened_groups.empty() )
        {
        remove( g_archive_state_stream_name );
        }
        
    // The archive is not validated here, because
    // the destructor can be called (for example) due to some exception.
    }

    
void Archive::validate()
    {
    // Check for opened groups.
    if( !my->opened_groups.empty() )
        {
        std::ostringstream ss;
        ss << "Unable to validate archive because there are "
           << my->opened_groups.size() << " opened groups left.";
        throw ExCanNotValidateArchive( ss.str() );
        }

    // Check for opened output streams.
    if( get_linked_output_stream_count() > 0 )
        {
        std::ostringstream ss;
        ss  << "Unable to validate archive because there are "
            << get_linked_output_stream_count()
            << " opened output streams left.";
        throw ExCanNotValidateArchive( ss.str() );
        }

    // Force validation if the archive is already valid.
    invalidate();

    try
        {
        OutputStream    state_stream;
        state_stream.open( this, g_archive_state_stream_name );
        if( !state_stream.good() )
            {
            std::ostringstream ss;
            ss << "Unable to create archive state stream.";
            throw ExCanNotCreateArchiveStream( ss.str() );
            }

        // Write archive state.
        my->write_archive_state( state_stream );
        }
    catch( ... )
        {
        remove( g_archive_state_stream_name ); // Make the archive invalid.
        throw;
        }

    // Lock the archive.
    my->read_only = true;
    my->valid     = true;
    }


void Archive::invalidate()
    {
    if( my->read_only )
        {
        std::ostringstream ss;
        ss << "Trying to write to locked archive. "
           << "The archive was validated and locked read only.";
        throw ExArchiveLocked( ss.str() );
        }

    if( my->valid )
        {
        my->report_invalidate();
        remove( g_archive_state_stream_name );
        my->valid = false;
        }
    }


bool Archive::is_valid() const
    {
    return my->valid;
    }
    
    
bool Archive::is_read_only() const
    {
    return my->read_only;
    }


void Archive::sync()
    {
    Super::flush();
    }
    
    
bool Archive::can_open_output_stream()
    {
    if( my->read_only )
        {
        return false;
        }
    else
        {
        // Invalidate the archive.
        invalidate();
        return true;
        }
    }


//------------------------------------------------------------------------------


Archive::GroupId Archive::create_group()
    {
    invalidate();

    GroupId         group_id = my->generate_group_id();
    My::GroupState  group_state;

    try 
        {
        // Register the group.
        my->registered_groups.insert
            (
            My::GroupInfoMap::value_type
                (
                group_id,
                My::GroupInfo()
                )
            );

        // Open the group.
        my->opened_groups.insert
            (
            My::GroupStateMap::value_type
                (
                group_id,
                group_state
                )
            );
        }
    catch( ... )
        {
        // Clean up.
        my->opened_groups.erase( group_id );
        my->registered_groups.erase( group_id );
        throw;
        }

    return group_id;
    }


bool Archive::validate_group
    (
    GroupId group_id
    )
    {
    invalidate();

    My::GroupStateMap::iterator group_state_it = my->opened_groups.find( group_id );

    if( !is_group_registered( group_id ) )
        {
        std::ostringstream ss;
        ss << "Trying to validate non-existing group. Group "
           << group_id << ".";
        throw ExArchiveGroupNotFound( ss.str() );
        }

    if( group_state_it == my->opened_groups.end() )
        {
        // Nothing to do.
        return true;
        }

    if( group_state_it->second.is_valid() )
        {
        my->validate_group( group_id, group_state_it->second );
        return true;
        }
    else
        {
        return false;
        }
    }


bool Archive::destroy_group
    (
    GroupId group_id
    )
    {
    My::GroupState                                  group_state;
    My::GroupInfoMap::iterator                      group_info_it;
    ObjectShared::ObjectIdList::iterator            object_id_it;
    ObjectShared::ObjectIdList                      objects;

    invalidate();

    if( !my->get_group_state( group_id, group_state ) )
        return false;

    group_info_it = my->registered_groups.find( group_id );
    hope( group_info_it != my->registered_groups.end() );

    // Close the group.
    my->opened_groups.erase( group_id );

    // Remove the group from the schedule.
    if( group_info_it->second.schedule_index.is_valid() )
        {
        my->group_schedule.pop( group_info_it->second.schedule_index );
        hope( !group_info_it->second.schedule_index.is_valid() );
        }

    // Unregister objects from the group.
    group_state.collect_objects( objects );
    for( object_id_it = objects.begin();
         object_id_it != objects.end();
         object_id_it++ )
        {
        const ObjectId & object_id = *object_id_it;
        // Unregister the object.
        my->object_to_group.erase( object_id );
        }

    // Unregister the group.
    my->registered_groups.erase( group_info_it );

    // Destroy group streams.
    remove( group_id_to_stream_name( group_id ) );

    for( object_id_it = objects.begin();
         object_id_it != objects.end();
         object_id_it++ )
        {
        const ObjectId & object_id = *object_id_it;
        // Destroy object stream.
        /** @todo Co se stane, kdyz je otevren stream? Spadne to? */
        my->remove_object_stream( group_id, object_id );
        }

    return true;
    }

    
void Archive::register_object
    (
    const ObjectId &            object_id,
    GroupId                     group_id,
    const STime *               object_swap_in_time
    )
    {
    My::ObjectIdToGroupMap::iterator    it;
    My::GroupInfoMap::iterator          group_info_it;
    My::GroupStateMap::iterator         group_state_it;
    
    // Only archivable objects can be stored in archive.
    hope( object_id.is_object_info_available() );
    hope( object_id.get_object_info().is_archivable() );

    // Note: Because group_id must be already opened group,
    //       we do not need to invalidate the archive here.

    // Check if the object is already registered.
    if( is_object_registered( object_id ) )
        {
        throw ExObjectAlreadyRegisteredInArchive( err_object_already_registered_string( object_id ) );
        }

    // Check if the group exist.
    group_info_it = my->get_group_info( group_id );

    // Check if the group is opened. See note above if changed.
    group_state_it = my->get_opened_group_state( group_id );

    // Register the object.
    it = my->object_to_group.insert
            (
            My::ObjectIdToGroupMap::value_type
                (
                object_id,
                group_id
                )
            ).first;

    // And now some exception catching... :-/
    try
        {
        // Insert the object to group state.
        if( group_state_it->second.add_object( object_id, object_swap_in_time, false ) )
            {
            // Reschedule the group.
            my->reschedule( group_state_it->second, group_info_it );
            }
        }
    catch( ... )
        {
        // Clean up. We do not need to reschedule the group.
        group_state_it->second.remove_object( object_id );
        my->object_to_group.erase( it );
        throw;
        }
    }

    
void Archive::collect_group_objects
    (
    GroupId                                 group_id,
    ObjectShared::ObjectIdList &            objects
    )
    {
    My::GroupStateMap::iterator group_state_it;

    // Check if the group is registered.
    my->get_group_info( group_id );

    // Check if the group is open.
    group_state_it = my->opened_groups.find( group_id );
    if( group_state_it != my->opened_groups.end() )
        {
        // The group is opened.
        group_state_it->second.collect_objects( objects );
        }
    else
        {
        // The group is not opened.
        My::GroupState  group_state;
        my->read_group_state( group_id, group_state );
        group_state.collect_objects( objects );
        }
    }


Archive::GroupId Archive::collect_group_objects
    (
    const ObjectId &                        object_id,
    ObjectShared::ObjectIdList &            objects
    )
    {
    GroupId                     group_id;
    My::GroupInfoMap::iterator  group_info_it;

    // Check if the object is registered.
    group_info_it = my->get_group_info_by_object( object_id );

    group_id = group_info_it->first;

    collect_group_objects( group_id, objects );

    return group_id;
    }


void Archive::read_extern_group_objects
    (
    std::istream &                  input_stream,
    ObjectShared::ObjectIdList &    objects
    )
    {
    My::GroupState          group_state;
    Serializer::Description desc;
    TextReader              tr( input_stream );
    
    tr.next_line_throw();
    try
        {
        // Read the group state.
        group_state.text_read( tr, desc );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << "Error reading archive group state from external stream. "
            << e.what();
        throw ExCorruptedArchive( ss.str() );
        }
    
    group_state.collect_objects( objects, false ); // Do not clear objects list.
    }

    
void Archive::validate_object
    (
    const ObjectId & object_id
    )
    {
    My::GroupInfoMap::iterator  group_info_it;
    My::GroupStateMap::iterator group_state_it;
    GroupId                     group_id;

    // Invalidate the archive.
    invalidate();
    
    // Report.
    my->report_validate_object( object_id );
    
    // Check if the object is registered.
    group_info_it = my->get_group_info_by_object( object_id );

    // Get group id.
    group_id = group_info_it->first;

    // Check if the group is opened.
    group_state_it = my->get_opened_group_state( group_id );

    // Get the group state.
    My::GroupState & group_state = group_state_it->second;
    
    // Validate the object.
    group_state.validate_object( object_id );

    // Validate the group if all objects from the group are already valid.
    if( group_state.is_valid() )
        {
        my->validate_group( group_id, group_state );
        }
    }


void Archive::copy_object_to
    (
    Archive &        dest_archive,
    const ObjectId & object_id
    )
    {
    My::GroupInfoMap::iterator      s_group_info_it;
    GroupId                         s_group_id;
    GroupId                         d_group_id;

    // Invalidate the destination archive.
    dest_archive.invalidate();

    // Report.
    my->report_copy_object( dest_archive, object_id );

    // Check if the object is registered in the source archive.
    s_group_info_it = my->get_group_info_by_object( object_id );

    // Get source group id.
    s_group_id = s_group_info_it->first;

    // Source group must be closed (valid).
    if( my->is_group_opened( s_group_id ) )
        {
        throw ExCanNotOpenArchiveStream( err_group_is_not_closed_string( s_group_id ) );
        }
    hope( is_object_valid( object_id ) );

    // Check if the object is registered in the destination archive.
    if( !dest_archive.is_object_registered( object_id ) )
        {
        My::GroupState                      s_group_state;
        My::ObjectStateMap::const_iterator  os_it;

        // Read source group state.
        my->read_group_state( s_group_id, s_group_state );
        hope( !s_group_state.get_object_state_map().empty() );
        hope( s_group_state.is_valid() );

        // Create new group in the destination archive.
        d_group_id = dest_archive.create_group();

        // Register all objects from the source group to the destination archive.
        for( os_it = s_group_state.get_object_state_map().begin(); os_it != s_group_state.get_object_state_map().end(); os_it++ )
            {
            const ObjectId &        oid = os_it->first;
            const My::ObjectState & os  = os_it->second;

            dest_archive.register_object
                (
                oid,
                d_group_id,
                os.has_swap_in_time() ? ( &os.swap_in_time ) : NULL
                );
            }
        }
    else
        {
        // Object is already registered in the destination archive.
        d_group_id = dest_archive.my->object_to_group.find( object_id )->second;
        }

    hope( dest_archive.is_object_registered( object_id ) );
    // hope( !dest_archive.is_object_valid( object_id ) );

    // Copy input object stream to output object stream.
        {
        InputStream  in;
        OutputStream out;

        in.open( this, s_group_id, object_id );
        out.open( &dest_archive, d_group_id, object_id );

        if( in.fail() )
            {
            std::ostringstream ss;
            ss  << debug_info_header_string()
                << "Unable to open archive object input stream. "
                << "Object: " << debug_object_info_string( s_group_id, object_id )
                << ".";
            /** @todo Throw another exception? Generic error open file? */
            throw ExCanNotOpenArchiveStream( ss.str() );
            }

        if( out.fail() )
            {
            std::ostringstream ss;
            ss  << dest_archive.debug_info_header_string()
                << "Unable to create archive output object stream. "
                << "Object: " << dest_archive.debug_object_info_string( d_group_id, object_id )
                << ".";
            throw ExCanNotCreateArchiveStream( ss.str() );
            }

        out << in.rdbuf();

        if( in.fail() )
            {
            std::ostringstream ss;
            ss  << debug_info_header_string()
                << "Error reading from object input stream."
                << " Object: " << debug_object_info_string( s_group_id, object_id )
                << ".";
            /** @todo FIXME Another exception? */
            throw ExArchivationError( ss.str() );
            }

        if( out.fail() )
            {
            std::ostringstream ss;
            ss  << dest_archive.debug_info_header_string()
                << "Error writing archive output object stream. "
                << "Object: " << dest_archive.debug_object_info_string( d_group_id, object_id )
                << ".";
            /** @todo FIXME Another exception? */
            throw ExArchivationError( ss.str() );
            }
        }

    dest_archive.validate_object( object_id );
    }


Archive::GroupId Archive::get_scheduled_group
    (
    const STime & time
    )
    {
    GroupId group_id = INVALID_GROUP_ID;

    if( my->group_schedule.is_empty() )
        return group_id;

    if( time <= my->group_schedule.get_key( 0 ) )
        return group_id;

    return my->group_schedule[ 0 ]->first;
    }


//------------------------------------------------------------------------------


void Archive::enumerate_registered_objects
    (
    ObjectShared::ObjectIdQueue & objects
    )
    {
    hope( objects.empty() );

    My::ObjectIdToGroupMap::const_iterator it;

    for( it = my->object_to_group.begin(); it != my->object_to_group.end(); it++ )
        {
        objects.push( it->first );
        }
    }

    
void Archive::enumerate_registered_groups
    (
    GroupIdList & groups
    )
    {
    hope( groups.empty() );

    My::GroupInfoMap::iterator it;

    groups.reserve( my->registered_groups.size() );

    /** @todo Use std::transform? */
    for( it = my->registered_groups.begin(); it != my->registered_groups.end(); it++ )
        {
        groups.push_back( it->first );
        }
    }

//------------------------------------------------------------------------------


bool Archive::is_group_registered
    (
    GroupId group_id
    ) const
    {
    return my->registered_groups.find( group_id ) != my->registered_groups.end();
    }


bool Archive::is_object_registered
    (
    const ObjectId & object_id
    ) const
    {
    return my->object_to_group.find( object_id ) != my->object_to_group.end();
    }


Archive::GroupId Archive::get_object_group_id
    (
    const ObjectId & object_id
    ) const
    {
    My::GroupInfoMap::iterator  group_info_it;
    GroupId                     group_id;
    
    // Check if the object is registered.
    group_info_it = my->get_group_info_by_object( object_id );

    // Get group id.
    group_id = group_info_it->first;
    
    return group_id;
    }
    
    
bool Archive::is_object_valid
    (
    const ObjectId & object_id
    ) const
    {
    My::ObjectIdToGroupMap::const_iterator  it = my->object_to_group.find( object_id );
    My::GroupStateMap::const_iterator       group_state_it;
    GroupId                                 group_id;

    if( it == my->object_to_group.end() )
        {
        return false;
        }

    group_id = it->second;
    
    hope( is_group_registered( group_id ) );

    group_state_it = my->opened_groups.find( group_id );
    if( group_state_it == my->opened_groups.end() )
        {
        return true; // The group is stored in the archive so it must be valid.
        }

    return group_state_it->second.is_object_valid( object_id );
    }
    

unsigned int Archive::get_registered_object_count() const
    {
    return my->object_to_group.size();
    }


unsigned int Archive::get_registered_group_count() const
    {
    return my->registered_groups.size();
    }


//------------------------------------------------------------------------------


void Archive::enable_report
    (
    unsigned int flags
    )
    {
    my->report_flags |= flags;
    }


void Archive::disable_report
    (
    unsigned int flags
    )
    {
    my->report_flags &= ~flags;
    }

    
//------------------------------------------------------------------------------


    //
    // Debug and error strings.
    //

std::string Archive::debug_info_string() const
    {
    /** @todo FIXME Report archive file. */
    return "(" + get_volume_name() + ")";
    }


std::string Archive::debug_info_header_string() const
    {
    return "Archive " + debug_info_string() + ": ";
    }


std::string Archive::debug_object_info_string
    (
    GroupId          group_id,
    const ObjectId & object_id
    ) const
    {
    std::ostringstream ss;
    
    ss  << object_id;
    
    if( group_id != INVALID_GROUP_ID )
        {
        ss  << " (stream "
            << object_id_to_stream_name( group_id, object_id )
            << ")";
        }

    return ss.str();
    }


std::string Archive::debug_group_info_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << "[" << group_id << "]" << " (stream "
        << group_id_to_stream_name( group_id )
        << ")";
    return ss.str();
    }
    
    
//------------------------------------------------------------------------------
    

    //
    // Error messages.
    //
    

std::string Archive::err_create_group_stream_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Unable to create (or write to) archive group stream. "
        << "Group " << debug_group_info_string( group_id ) << ".";
    return ss.str();
    }
    
    
std::string Archive::err_open_group_stream_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Unable to open archive group stream. "
        << "Group " << debug_group_info_string( group_id ) << ".";
    return ss.str();
    }


std::string Archive::err_read_group_state_string
    (
    GroupId             group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Error reading group state from archive group stream. "
        << "Group " << debug_group_info_string( group_id ) << ".";
    return ss.str();
    }


std::string Archive::err_group_not_registered_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Group " << debug_group_info_string( group_id )
        << " is not registered.";
    return ss.str();
    }


std::string Archive::err_group_not_opened_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Group " << debug_group_info_string( group_id )
        << " is not opened. "
        << "Group must opened before anyone can access its content.";
    return ss.str();
    }


std::string Archive::err_object_not_registered_string
    (
    const ObjectId & object_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Object " << debug_object_info_string( INVALID_GROUP_ID, object_id )
        << " is not registered in the archive. "
        << "Object must be registered in the archive before anyone "
        << "can store or access its state.";
    return ss.str();
    }


std::string Archive::err_object_already_registered_string
    (
    const ObjectId & object_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Object " << debug_object_info_string( INVALID_GROUP_ID, object_id )
        << " already registered in the archive."
        << "Object must be removed from the archive before it"
        << " can be registered again.";
    return ss.str();
    }


std::string Archive::err_group_is_not_closed_string
    (
    GroupId group_id
    ) const
    {
    std::ostringstream ss;
    ss  << debug_info_header_string()
        << "Group " << debug_group_info_string( group_id )
        << " is not closed (validated)."
        << " Operation requires valid group.";
    return ss.str();
    }
    
    
//------------------------------------------------------------------------------

    
static void write_group_snapshot
    (
    std::ostream &      s,
    Archive::GroupId    group_id,
    bool                opened,
    const std::string & name // stream name
    )
    {
    std::string color = opened ? "red" : "light yellow";

    s << name << " [ ";
    s << "style = filled "
      << "color = \"" << color << "\" ";
    s << " ];\n";

    s << "archive -> " << name << " ;\n";
    }


static void write_object_snapshot
    (
    std::ostream &      s,
    const std::string   object_name,
    const std::string & group_name,
    bool                valid
    )
    {
    std::string color = valid ? "light blue" : "red";

    s << object_name << " [ "
      << "style = filled "
      << "color = \"" << color << "\" "
      << " ];\n";

    s << group_name << " -> " << object_name << " ;\n";
    }


void Archive::save_graphviz_snapshot
    (
    std::ostream &  s
    )
    {
    s   << "digraph archive\n" /** @todo Archive (volume) name. */
        << "{\n";

    std::string  archive_color = is_valid() ? "light yellow" : "red";

    // Write main node.
    s   << "archive [ "
        << "style = filled "
        << "color = \"" << archive_color << "\" "
        << "label = \"Archive: "
            << my->registered_groups.size() << " groups, "
            << my->object_to_group.size() << " objects, "
            << my->opened_groups.size() << " opened groups, "
            << my->group_schedule.get_size() << " scheduled groups, "
            << ( is_read_only() ? "read-only" : "read-write" ) << " mode "
        << "\" "
        << " ];\n";

    // Write groups.
    My::GroupInfoMap::const_iterator group_info_it;
    for( group_info_it = my->registered_groups.begin();
         group_info_it != my->registered_groups.end();
         group_info_it++ )
        {
        GroupId group_id = group_info_it->first;
        write_group_snapshot
            (
            s,
            group_id,
            my->is_group_opened( group_id ),
            group_id_to_stream_name( group_id )
            );
        }

    // Write objects.
    My::ObjectIdToGroupMap::const_iterator o_it;
    for( o_it = my->object_to_group.begin(); o_it != my->object_to_group.end(); o_it++ )
        {
        const ObjectId & object_id = o_it->first;
        GroupId          group_id = o_it->second;
        bool             o_valid = true;

        if( my->is_group_opened( group_id ) )
            {
            My::GroupStateMap::iterator group_state_it = my->get_opened_group_state( group_id );
            o_valid = group_state_it->second.is_object_valid( object_id );
            }

        std::ostringstream ss;
        ss  << "\"" << object_id << "\"";

        write_object_snapshot
            (
            s,
            ss.str(),
            group_id_to_stream_name( group_id ),
            o_valid
            );
        }

    s << "}" << std::endl;
    }


} // namespace Massiv::Core
} // namespace Massiv
