/** @file archive.h
 * Archive file.
 */

/*
 * begin        : Fri Aug 23 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ARCHIVE_H
#define MASSIV_ARCHIVE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_COMPACT_VOLUME_H
#include "filesys/compact_volume.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * ArchiveFile is an extension of #CompactVolume.
 * It is designed for #Archive purposes only and you should never use ArchiveFile itself.
 *
 * ArchiveFile limits access to the volume by creating its own streams aka archive streams.
 * Only archive streams may be used to access data in the volume.
 * Archive streams has limited usage. Thay don't support opening the stream directly
 * by constructor. You must use @c open() member method instread. Also archive streams
 * do not support @c close() function. The stream object itself must be destroyed
 * to close the stream (and unlink the stream from the volume - see below).
 *
 * Archive streams are designed that the #ArchiveFile knows how many opened
 * output archive streams are linked to it at a time. Before an output archive
 * stream is opened, the #ArchiveFile calls @c can_open_output_stream() callback.
 * This function is implemeneted by #Archive itself and is used to check whether
 * write access to the archive is permitted.
 * Archives containing at least one linked output stream can't be validated.
 * See #Archive for more info.
 *
 * @see Archive.
 * @todo ArchiveFile can't use 'protected' inherit from CompactVolume
 *      but we want deny public access to all functions from CompactVolume.
 *      How to do that?
 *
 * @ingroup archivation
 */
class ArchiveFile : public CompactVolume
    {
    typedef CompactVolume       Super;

    NO_COPY_CONSTRUCTOR( ArchiveFile );
    NO_ASSIGNMENT_OPERATOR( ArchiveFile );

public:

    enum
        {
        INVALID_GROUP_ID = 0
            /**< Invalid group id. */
        };

    typedef UInt32      GroupId;
        /**< Group identification number. Uniquely identifies group stored in the archive.
         * Invalid group id is identified by @c INVALID_GROUP_ID value.
         */

    //
    // Archive streams.
    //

    //--------------------------------------------------------------------------

    /**
     * Generic input archive stream.
     * @note Archive streams do not support opening a stream in their constructors.
     *      Use @c open() member-function instead.
     */
    class InputStream : public Super::IFStream
        {
        typedef IFStream        Super;

        NO_COPY_CONSTRUCTOR( InputStream );
        NO_ASSIGNMENT_OPERATOR( InputStream );

    public:

        /** Constructor. */
        InputStream();

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            ArchiveFile *       archive,
            const std::string & stream_name
            );

        /**
         * Opens stream.
         * Returns true on success.
         * This function is only a shortcut for calling
         * @code
         * open( archive, ArchiveFile::object_id_to_stream_name( group_id, object_id ) )
         * @endcode
         */
        bool open
            (
            ArchiveFile *       archive,
            GroupId             group_id,
            const ObjectId &    object_id
            );

        }; // InputStream

    //--------------------------------------------------------------------------

    /**
     * Generic output archive stream.
     * @note Archive streams do not support opening a stream in their constructors.
     *      Use @c open() member-function instead.
     *
     * @warnig Creating output stream in an archive only calls @c can_open_output_stream()
     *      and links the stream to the archive. No more actions are performed.
     *      For example if you create output stream for an object, the group
     *      of the object is not opened nor it is invalidated! You have to
     *      set data structures of the archive to proper state before creating
     *      the stream.
     *      See #Archive for more info.
     */
    class OutputStream : public Super::OFStream
        {
        typedef OFStream        Super;

        NO_COPY_CONSTRUCTOR( OutputStream );
        NO_ASSIGNMENT_OPERATOR( OutputStream );

    public:

        /** Constructor. */
        OutputStream();

        /** Destructor. */
        ~OutputStream();

        /**
         * Opens stream.
         * Returns true on success.
         */
        bool open
            (
            ArchiveFile *       archive,
            const std::string & stream_name
            );

        /**
         * Opens streams identified by #ObjectId.
         * Returns true on success.
         * This function is only a shortcut for calling
         * @code
         * open( archive, ArchiveFile::object_id_to_stream_name( group_id, object_id ) )
         * @endcode
         */
        bool open
            (
            ArchiveFile *       archive,
            GroupId             group_id,
            const ObjectId &    object_id
            );

    private:

        ArchiveFile *   archive;
            /**< Archive to which the stream belongs to. */

        bool            linked;
            /**< True if the stream is linked to its archive.
             * Archives with at least one linked output stream can't be validated.
             */

        }; // OutputStream

    //--------------------------------------------------------------------------

    friend class OutputStream;

public:

    /**
     * Opens archive file.
     * @param file_name Describes existing file name.
     */
    ArchiveFile
        (
        const std::string &     file_name,
        std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out
        );

    /**
     * Creates empty archive file ("formats" the file).
     */
    ArchiveFile
        (
        const std::string &     file_name,
        const std::string &     archive_name
        );

    /**
     * Destructor.
     */
    virtual ~ArchiveFile();

public:

    /**
     * Check whether an output archive stream can be opened.
     * This is only callback which is called before an output archive stream
     * is opened and is implemenented by #Archive class.
     * @return True if the stream can be opened. False if the stream
     *      may not be opened.
     */
    virtual bool can_open_output_stream() = 0;

public:

    /**
     * Get number of output streams linked to this #ArchiveFile.
     */
    unsigned int get_linked_output_stream_count() const
        {
        return linked_output_stream_count;
        }

    /**
     * Converts #ObjectId to string which can be used as a stream name.
     * @param group_id Group id the object belongs to.
     * Returned string has two parts - group id part, identifying
     * group in which the object is stored, and object id part, identifying
     * the object. Object id part of the stream name is obtained
     * using @c object_id_to_plain_stream_name().
     */
    static std::string object_id_to_stream_name
        (
        GroupId             group_id,
        const ObjectId &    object_id
        );

    /**
     * Converts #ObjectId to string used as name of the stream for the object.
     * Returned name is not valid name of stream, because valid stream name
     * in the archive must contain group id part. Use @c object_id_to_stream_name()
     * to get full name of stream for an object.
     * Use @c object_id_to_plain_stream_name() only to get stream name
     * of the object relative to its group.
     */
    static std::string object_id_to_plain_stream_name
        (
        const ObjectId &    object_id
        );

    /**
     * Converts #GroupId to string which can be used as a stream name.
     */
    static std::string group_id_to_stream_name
        (
        GroupId     group_id
        );

private:

    unsigned int    linked_output_stream_count;
        /**< Number of linked output streams. */

    }; // ArchiveFile


//------------------------------------------------------------------------------


/**
 * Archive file.
 * Archive file contains groups of objects.
 * Every group has its unique group id.
 * Group can be opened or closed.
 * Closed group is stored in the archive and is valid (all objects
 * of the group are stored in the archive).
 * Opened group may not be valid (some objects of the group
 * may be stored, some not).
 *
 * Any write access to the archive marks the archive as invalid.
 * The only way to validate the archive is to use @c validate()
 * function. Due to archive consistency maintenance, when calling
 * @c validate() all groups must be closed and any write acces
 * to the archive from the validation will throw exception
 * (the archive is locked).
 * The only way to gain write acces to the archive after
 * its validation is to create new instance of Archive object.
 *
 * @todo Document group and object stream manipulation and validation.
 *
 * @todo If exception is thrown, we should somehow report archive file name.
 * @todo Do some statistic - how many groups in the archive, number of objects,
 * average objects per group, average object stream length, objects written
 * per second, ...
 * @todo CZ Chytat ExIoError pri zapisu a premapovavat je na
 * vyjimky odvozene od ExArchivationError?
 * @todo FIXME Jak testovat ze stream neexistuje?
 * @todo Pridat check_validity().
 * @todo Optimalizovat is_valid() testy, validate_object(), validate_group().
 *
 * @ingroup archivation
 */
class Archive : public ArchiveFile
    {
    typedef ArchiveFile Super;

public:

    /**
     * List of group ids.
     */
    typedef std::vector< GroupId >  GroupIdList;

    /**
     * Access mode to the archive.
     */
    enum Mode
        {
        MODE_READ_WRITE,
            /**< Both read and write operations are permitted. */

        MODE_READ_ONLY
            /**< Only read operations are permitted. The archive is locked.*/

        }; // Mode

    /**
     * Report flags.
     * Use @c enable_report() and @c disable_report() to enable/disable
     * specified reports.
     */
    enum ReportFlags
        {
        REPORT_INVALIDATE           = 0x0001,
            /**< Report archive invalidation. */

        REPORT_COPY_OBJECT          = 0x0002,
            /**< Report copying object to another archive. */

        REPORT_REMOVE_OBJECT_STREAM = 0x0004,
            /**< Report removing object stream from the archive. */

        REPORT_REMOVE_GROUP_STREAM  = 0x0008,
            /**< Report removing group stream from the archive. */

        REPORT_VALIDATE_OBJECT      = 0x0010,
            /**< Report object validation. */

        REPORT_VALIDATE_GROUP       = 0x0020,
            /**< Report group validation. */

        REPORT_ALL                  = 0xffff
            /**< Report everything. */

        }; // ReportFlags

public:

    /**
     * Opens archive file.
     * @param file_name Describes existing file name.
     * @param open_mode Open mode of the archive. If the archive is opened
     *          in @c MODE_READ_ONLY, it will be locked.
     */
    Archive
        (
        const std::string &     file_name,
        Mode                    open_mode
        );

    /**
     * Creates empty archive file ("formats" the file).
     */
    Archive
        (
        const std::string &     file_name,
        const std::string &     archive_name
        );

    /**
     * Destructor.
     * The archive is not validated automatically here.
     * To make the archive valid, call @c validate().
     */
    ~Archive();

public:

    /**
     * Validate the archive - proclaim the archive finished.
     * Any write access to the archive from now will throw exception.
     * The only way to get write access to the archive again is to destroy
     * this object and create new one.
     * This behaviour is due to archive consistency maintenance.
     *
     * All groups must be closed.
     * Writes archive state to the archive file.
     * The archive is not synced.
     */
    void validate();

    /**
     * Invalidate the archive. If the archive is in read only mode,
     * exception is thrown.
     */
    void invalidate();

    /**
     * Check whether the archive is valid.
     */
    bool is_valid() const;

    /**
     * Check whether the archive is in read only mode.
     * Only valid archives may be in read only mode.
     */
    bool is_read_only() const;

    /**
     * Flush internal buffers.
     * Opened groups are not synced.
     * @see Volume::sync().
     */
    void sync();

public:

    /**
     * #ArchiveFile callback.
     * If the archive is not in read only mode, it is invalidated
     * and the funtion returns true.
     */
    virtual bool can_open_output_stream();

public:

    /**
     * Create new empty group in the archive.
     * The group is opened.
     */
    GroupId create_group();

    /**
     * Validate @a group.
     * Does nothing if the group is already closed (and returns true).
     * Otherwise the group is validated and closed.
     * Returns true if the group was validated and closed.
     * @note Group is automatically validated once all objects from
     *  the group are validates. See @c validate_object().
     */
    bool validate_group
        (
        GroupId group
        );

    /**
     * Remove group from the archive.
     * Can be called on both opened and closed group.
     * All objects from the group are removed from the archive.
     * Returns whether the group was removed.
     */
    bool destroy_group
        (
        GroupId group_id
        );

    /**
     * Register object in the archive. The object is added
     * to the @a group.
     * The group must be opened (may change in the future).
     * @note Currently there is no way how to remove object
     *  from the archive except to destroy entire object's group.
     * @param object_swap_in_time Time when the object (and
     * its group) must be swapped in. Can be @ NULL.
     * @pre #ObjectInfo part of the @a object_id must be valid.
     */
    void register_object
        (
        const ObjectId &            object_id,
        GroupId                     group,
        const STime *               object_swap_in_time
        );

    /**
     * Get all object ids from one group.
     * Can be called on both opened and closed group.
     * Object ids from the group are stored in @a objects.
     * It is safe to call this function on read only archive.
     * @note Enumerated object ids always contain optional object info
     *      part of #ObjectId.
     */
    void collect_group_objects
        (
        GroupId                                 group_id,
        ObjectShared::ObjectIdList &            objects
        );

    /**
     * Get all object ids from one group. The group
     * is selected by one object from the group.
     * Can be called on both opened and closed group.
     * Object ids from the group are stored in @a objects.
     * It is safe to call this function on read only archive.
     * Returns group id of the collected group.
     * @note Enumerated object ids always contain optional object info
     *      part of #ObjectId.
     */
    GroupId collect_group_objects
        (
        const ObjectId &                        object_id,
        ObjectShared::ObjectIdList &            objects
        );

    /**
     * Get all object ids from one group stored in external stream.
     * Object ids from the group are stored in @a objects.
     * @note @a objects is not cleared.
     */
    static void read_extern_group_objects
        (
        std::istream &                  input_stream,
        ObjectShared::ObjectIdList &    objects
        );

    /**
     * Validate object.
     *
     * @pre The object is be registered in the archive.
     * @pre Group of the object is be opened.
     *
     * If all objects from the group of the object are already valid,
     * the entire group is automatically validated and closed.
     */
    void validate_object
        (
        const ObjectId & object_id
        );

    /**
     * Copy @a object from this (source) archive to another
     * (destination) @a dest_archive.
     *
     * @pre The object is registered in the source archive.
     * @pre The source group of the object is be closed.
     *
     * If @a object is not registered in the destination archive, new
     * group is created in the destination archive and all objects from
     * the source group of the @a object are registered in the new group.
     */
    void copy_object_to
        (
        Archive &        archive,
        const ObjectId & object
        );

    /**
     * Get first scheduled group for @a time.
     * If there is no group scheduled for the time,
     * @c INVALID_GROUP_ID is returned.
     */
    GroupId get_scheduled_group
        (
        const STime &   time
        );

public:

    /**
     * Enumerate all registered objects in the archive.
     * @a objects must be empty.
     */
    void enumerate_registered_objects
        (
        ObjectShared::ObjectIdQueue & objects
        );

    /**
     * Enumerate all registered groups in the archive.
     * @a groups must be empty.
     */
    void enumerate_registered_groups
        (
        GroupIdList & groups
        );

public:

    /**
     * Check if the @a group is registered in the archive.
     */
    bool is_group_registered
        (
        GroupId     group
        ) const;

    /**
     * Check if the @a object is registered in the archive.
     */
    bool is_object_registered
        (
        const ObjectId & object
        ) const;

    /**
     * Get group id of the group the object identified by @a object_id
     * belong to in this archive.
     * @pre The object is registered in the archive.
     */
    GroupId get_object_group_id
        (
        const ObjectId & object_id
        ) const;

    /**
     * Check whether the @a object is registered and valid.
     * @note Doesn't test if the entire group of the object
     * is valid.
     */
    bool is_object_valid
        (
        const ObjectId & object
        ) const;

    /**
     * Get number of registered objects.
     */
    unsigned int get_registered_object_count() const;

    /**
     * Get number of registered groups.
     */
    unsigned int get_registered_group_count() const;

public:

    /**
     * Enable report of specified events.
     */
    void enable_report
        (
        unsigned int report_flags
        );

    /**
     * Disable report of specified events.
     */
    void disable_report
        (
        unsigned int report_flags
        );

public:

    //
    // Debug and error strings.
    //

    /**
     * Returns string containing basic information
     * about the archive (file name, volume name).
     */
    std::string debug_info_string() const;

    /**
     * Returns string containing basic information
     * about the archive (file name, volume name).
     * The string is terminated by ": " characters.
     */
    std::string debug_info_header_string() const;

    /**
     * Returns debug string (object id and corresponding
     * stream name if @a group_id is valid).
     * @param group_id Group id the object belongs to.
     */
    std::string debug_object_info_string
        (
        GroupId          group_id,
        const ObjectId & object_id
        ) const;

    /**
     * Returns debug string (group id and corresponding
     * stream name).
     */
    std::string debug_group_info_string
        (
        GroupId group_id
        ) const;

public:

    //
    // Error messages.
    //

    /** Create error string. */
    std::string err_create_group_stream_string
        (
        GroupId group_id
        ) const;

    /** Create error string. */
    std::string err_open_group_stream_string
        (
        GroupId group_id
        ) const;

    /** Create error string. */
    std::string err_read_group_state_string
        (
        GroupId             group_id
        ) const;

    /** Create error string. */
    std::string err_group_not_registered_string
        (
        GroupId group_id
        ) const;

    /** Create error string. */
    std::string err_group_not_opened_string
        (
        GroupId group_id
        ) const;

    /** Create error string. */
    std::string err_object_not_registered_string
        (
        const ObjectId & object_id
        ) const;

    /** Create error string. */
    std::string err_object_already_registered_string
        (
        const ObjectId & object_id
        ) const;

    /** Create error string. */
    std::string err_group_is_not_closed_string
        (
        GroupId group_id
        ) const;

public:

    /**
     * Saves "snapshot" of current archive state.
     * The state is saved in a graph format to be examined
     * by GraphViz utility.
     * Groups and objects that weren't validated are drawn with red color.
     */
    void save_graphviz_snapshot
        (
        std::ostream & output_stream
        );

private:
    // Forward declaration of Archive internal structure.
    class My;

private:
    std::auto_ptr< My >     my;
        /**< Archive internal data structure. Always valid pointer. */

    }; // Archive

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVE_H
