/** @file archive_internal.cpp
 * #Archive internal data structures (implementation).
 */

/*
 * begin        : Fri Jun 20 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "object/archive_internal.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "serializer/serializer.h"
#include "serializer/stype_int.h"
#include "global/global.h"
#include "utility/debug.h"


namespace Massiv {
namespace Core {


// Implementation of Archive::My::ObjectState
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Archive::My::ObjectState::ObjectState() :
    swap_in_time_valid( false ),
    valid_stream( false )
    {
    }


Archive::My::ObjectState::ObjectState
    (
    const Archive::My::ObjectState & other
    ) :
    swap_in_time_valid( other.swap_in_time_valid ),
    swap_in_time( other.swap_in_time ),
    valid_stream( other.valid_stream )
    {
    }


Archive::My::ObjectState & Archive::My::ObjectState::operator=
    (
    const Archive::My::ObjectState & other
    )
    {
    swap_in_time_valid = other.swap_in_time_valid;
    swap_in_time       = other.swap_in_time;
    valid_stream       = other.valid_stream;
    return *this;
    }


Archive::My::ObjectState::ObjectState
    (
    const STime *       t,
    bool                v
    ) :
    valid_stream( v )
    {
    if( t )
        {
        swap_in_time       = ( *t );
        swap_in_time_valid = true;
        }
    else
        {
        swap_in_time_valid = false;
        }
    }


void Archive::My::ObjectState::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    tr.require_string( "swap_time" );

    // Read swap-in time validity flag.
    swap_in_time_valid.text_read( tr, desc );

    // Read swap-in time.
    if( swap_in_time_valid )
        {
        swap_in_time.text_read( tr, desc );
        }
    else
        {
        swap_in_time = STime(); // Not necessary.
        }

    // 'valid_stream' is ignored.
    valid_stream = true; // Group stored in the archive must be valid.
    }


void Archive::My::ObjectState::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    // 'valid_stream' is ignored.
    hope( valid_stream );
    
    tw.line() << "swap_time";

    // Write swap-in time validity flag.
    tw.write_space();
    swap_in_time_valid.text_write( tw, desc );
    
    // Write swap-in time.
    if( swap_in_time_valid )
        {
        tw.write_space();
        swap_in_time.text_write( tw, desc );
        }
    }
    
    
//------------------------------------------------------------------------------


// Implementation of Archive::GroupState
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Archive::My::GroupState::GroupState() :
    swap_in_time_valid( false )
    {
    }


Archive::My::GroupState::GroupState
    (
    const Archive::My::GroupState & other
    ) :
    swap_in_time_valid( other.swap_in_time_valid ),
    swap_in_time( other.swap_in_time ),
    objects( other.objects )
    {
    }


Archive::My::GroupState & Archive::My::GroupState::operator=
    (
    const Archive::My::GroupState & other
    )
    {
    swap_in_time_valid  = other.swap_in_time_valid;
    swap_in_time        = other.swap_in_time;
    objects             = other.objects;
    return *this;
    }


bool Archive::My::GroupState::add_object
    (
    const ObjectId &    object_id,
    const STime *       object_swap_in_time,
    bool                object_stream_valid
    )
    {
    ObjectState object_state( object_swap_in_time, object_stream_valid );

    return add_object( object_id, object_state );
    }


bool Archive::My::GroupState::add_object
    (
    const ObjectId &    object_id,
    const ObjectState & object_state
    )
    {
    // Insert object state.
    DEBUG_ONLY( bool r = ) objects.insert
                (
                ObjectStateMap::value_type
                    (
                    object_id,
                    object_state
                    )
                ).second;
    hope( r );
    
    // Update swap-in time.
    if( !object_state.has_swap_in_time() )
        {
        return false;
        }

    if( swap_in_time_valid )
        {
        if( object_state.swap_in_time < swap_in_time )
            {
            swap_in_time = object_state.swap_in_time;
            return true;
            }
        return false;
        }
    else
        {
        swap_in_time_valid = true;
        swap_in_time       = object_state.swap_in_time;
        return true;
        }
    }


void Archive::My::GroupState::remove_object
    (
    const ObjectId & object_id
    )
    {
    ObjectStateMap::const_iterator  it;
    bool                            new_time_valid = false;
    STime                           new_time;

    // Remove object state.
    objects.erase( object_id );

    // Update swap-in time.
    for( it = objects.begin(); it != objects.end(); it++ )
        {
        const ObjectState & state = it->second;
        
        // Update new swap-in time.
        if( state.has_swap_in_time() )
            {
            if( new_time_valid )
                {
                if( state.swap_in_time < new_time )
                    {
                    new_time = state.swap_in_time;
                    }
                }
            else
                {
                new_time = state.swap_in_time;
                new_time_valid = true;
                }
            }
        }

    swap_in_time_valid = new_time_valid;
    swap_in_time       = new_time;
    }


void Archive::My::GroupState::validate_object
    (
    const ObjectId & object_id
    )
    {
    ObjectStateMap::iterator it = objects.find( object_id );
    if( it == objects.end() )
        {
        return;
        }
    else
        {
        ObjectState & state = it->second;
        state.validate();
        }
    }


void Archive::My::GroupState::invalidate_object
    (
    const ObjectId & object_id
    )
    {
    ObjectStateMap::iterator it = objects.find( object_id );
    if( it == objects.end() )
        {
        return;
        }
    else
        {
        ObjectState & state = it->second;
        state.invalidate();
        }
    }

bool Archive::My::GroupState::is_valid() const
    {
    bool                            valid = true;   // Empty group is valid.
    ObjectStateMap::const_iterator  it;

    for( it = objects.begin(); it != objects.end(); it++ )
        {
        const ObjectState & state = it->second;
        if( !state.is_valid() )
            {
            valid = false;
            break;
            }
        }

    return valid;
    }


bool Archive::My::GroupState::is_object_valid
    (
    const ObjectId & object_id
    ) const
    {
    ObjectStateMap::const_iterator  it = objects.find( object_id );
    
    it = objects.find( object_id );
    if( it == objects.end() )
        {
        return false;
        }
    else
        {
        const ObjectState & state = it->second;
        return state.is_valid();
        }
    }


void Archive::My::GroupState::collect_objects
    (
    ObjectShared::ObjectIdList & out,
    bool clear
    )
    {
    ObjectStateMap::const_iterator it;

    if( clear )
        {
        out.clear();
        out.reserve( objects.size() );
        }

    for( it = objects.begin(); it != objects.end(); it++ )
        {
        const ObjectId &    object_id = it->first;
        
        out.push_back( object_id );
        }
    }


void Archive::My::GroupState::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    SInt< UInt32, 32 >      count;
    ObjectState             object_state;
    ObjectId                object_id;

    // Clear self.
    swap_in_time_valid = false;
    objects.clear();

    // Swap-in time of the entire group is not serialized
    // because it is computed from swap-in time of the objects
    // from the group.
    
    tr.require_string( "object_count" );

    // Read number of objects stored in the group.
    count.text_read( tr, desc );

    while( count-- )
        {
        tr.next_line_throw();
        
        // Read object id.
        object_id.text_read( tr, desc );
        
        // Check corrupted archive.
        if( objects.find( object_id ) != objects.end() )
            {
            std::ostringstream ss;
            ss  << "Corrupted stream of archive group. One object specified two or more times.";
            
            throw ExCorruptedArchive( ss.str() );
            }
        
        // Read object state.
        object_state.text_read( tr, desc );
        
        // Add the object to the group.
        add_object( object_id, object_state );
        }

    hope( is_valid() );
    }


void Archive::My::GroupState::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    ObjectStateMap::const_iterator it;

    hope( is_valid() ); // Only valid groups are stored in the archive.

    // Swap-in time of the entire group is not serialized
    // because it is computed from swap-in time of the objects
    // from the group.

    SInt< UInt32, 32 >  count = objects.size();
    
    tw.line() << "object_count";
    tw.write_space();
    
    // Store number of objects stored in the archive.
    count.text_write( tw, desc );
    
    for( it = objects.begin(); it != objects.end(); it++ )
        {
        const ObjectId &    object_id = it->first;
        const ObjectState & state = it->second;
        
        tw.end_line();
        object_id.text_write( tw, desc );       // ObjectId
        
        tw.write_space();
        state.text_write( tw, desc );           // ObjectState
        }
    }


//------------------------------------------------------------------------------


Archive::My::My
    (
    Mode    mode
    ) :
    archive( NULL ),
    next_group_id( 1 ),
    read_only( mode == MODE_READ_ONLY ),
    valid( false ),
    report_flags( 0 ) // Do not report anything.
    {
    }
    
    
void Archive::My::read_archive_state
    (
    InputStream &  file
    )
    {
    hope( registered_groups.empty() );
    hope( opened_groups.empty() );
    hope( object_to_group.empty() );
    hope( group_schedule.is_empty() );
    hope( next_group_id == 1 );

    try
        {
        TextReader              tr( file );
        Serializer::Description desc;

        // Get first line.
        tr.next_line_throw();
        
        // Read group ids of all registered groups and registerd them.
        read_groups_from_state_stream( tr, desc );
        
        // Register all objects and create global schedule.
        read_group_states_and_objects();
        
        // Do some verification.
        verify_archive();
        
        // Log.
        log_archive_open();
        }
    catch( std::exception & e )
        {
        // Clean. First clear group_schedule and object_to_group!
        group_schedule.clear();
        object_to_group.clear();
        registered_groups.clear();
        hope( opened_groups.empty() );
        next_group_id = 1;

        // Throw.
        std::ostringstream ss;
        ss << "Error reading archive state. "
           << e.what();
        throw ExCorruptedArchive( ss.str() );
        }
    catch( ... )
        {
        // Clean. First clear group_schedule and object_to_group!
        group_schedule.clear();
        object_to_group.clear();
        registered_groups.clear();
        hope( opened_groups.empty() );
        next_group_id = 1;

        // Throw.
        throw;
        }
    }


void Archive::My::read_groups_from_state_stream
    (
    TextReader &                tr,
    Serializer::Description &   desc
    )
    {
    hope( registered_groups.empty() );
    hope( opened_groups.empty() );
    hope( object_to_group.empty() );
    hope( group_schedule.is_empty() );
    
    SInt< UInt32, 32 >      group_count;
    UInt32                  count;
    GroupId                 group_id;
    
    // Read number of groups stored in the archive.
    tr.require_string( "group_count" );
    group_count.text_read( tr, desc );
    
    tr.next_line_throw();

    // Register all groups. Read count group ids.
    tr.require_string( "group_ids" );
    count = group_count;
    while( count-- )
        {
        group_id = text_read_group_id( tr, desc );

        if( group_id == INVALID_GROUP_ID )
            {
            throw ExCorruptedArchive( "Invalid group id in archive state stream." );
            }

        if( archive->is_group_registered( group_id ) )
            {
            std::ostringstream ss;
            ss << "Two or more groups with same group id '" << group_id << "'.";
            throw ExCorruptedArchive( ss.str() );
            }

        if( group_id >= next_group_id )
            {
            next_group_id = group_id + 1;
            hope( next_group_id != 0 );
            }

        // Register the group.
        registered_groups.insert
            (
            GroupInfoMap::value_type
                (
                group_id,
                GroupInfo()
                )
            );
        }

    hope( registered_groups.size() == group_count );
    }


void Archive::My::read_group_states_and_objects()
    {
    hope( opened_groups.empty() );
    hope( object_to_group.empty() );
    hope( group_schedule.is_empty() );
    
    GroupInfoMap::iterator          group_info_it;
    GroupState                      group_state;
    
    // For each registered group read its GroupState and register
    // all objects from the group.
    for( group_info_it = registered_groups.begin();
         group_info_it != registered_groups.end();
         group_info_it++ )
        {
        GroupId                                       group_id = group_info_it->first;
        ObjectShared::ObjectIdList                    objects;
        ObjectShared::ObjectIdList::const_iterator    it;

        // Read group state.
        read_group_state( group_id, group_state );

        // Register objects from the group.
        // Note: We do not catch exceptions here because
        // if any exception is thrown, object_to_group
        // and schedule will be cleared anyway (in the catch block
        // of the caller).
        
        // Enumerate objects from the group.
        group_state.collect_objects( objects );
        
        // Register each of the object from the group.
        for( it = objects.begin(); it != objects.end(); it++ )
            {
            const ObjectId &           object_id = *it;
            
            // Verify object id.
            if( !object_id.is_object_info_available() )
                {
                std::ostringstream ss;
                ss  << "Invalid object id " << object_id << ". The object id doesn't contain ObjectInfo part.";
                throw ExCorruptedArchive( ss.str() );
                }
            
            const ObjectInfo            object_info = object_id.get_object_info();
                
            if( !object_info.is_class_registered() )
                {
                std::ostringstream ss;
                ss  << "Invalid object id " << object_id 
                    << ". The class id stored in the object id is not supported by local node.";
                throw ExCorruptedArchive( ss.str() );
                }
            
            // Only archivable objects may be stored in archive.
            if( !object_info.is_archivable() )
                {
                std::ostringstream ss;
                ss  << "Invalid object id " << object_id
                    << ". The object is not archivable. Only archivable objects may be stored in archive file.";
                throw ExCorruptedArchive( ss.str() );
                }
            
            // Check if the object is already registered.
            if( archive->is_object_registered( object_id ) )
                {
                GroupId             other_group_id = object_to_group.find( object_id )->second;
                std::ostringstream  ss;
                ss  << "One object in two groups. Object id: " << object_id
                    << ", group1 id: " << other_group_id
                    << ", group2 id: " << group_id
                    << ".";
                throw ExCorruptedArchive( ss.str() );
                }

            // Insert the object in the map.
            object_to_group.insert
                (
                ObjectIdToGroupMap::value_type
                    (
                    object_id,
                    group_id
                    )
                );
            }

        // Insert the group in the schedule.
        reschedule( group_state, group_info_it );
        }
    }


void Archive::My::verify_archive()
    {
    unsigned int min_stream_count = 1 + registered_groups.size() + object_to_group.size();
    
    // There must be at least one archive state stream and one stream for each group
    // and one stream for each object.
    if( archive->get_stream_count() < min_stream_count )
        {
        std::ostringstream ss;
        ss  << "Mismatched file stream count. "
            << "Expected at least " << min_stream_count << " file streams, but "
            << archive->get_stream_count() << " is present in the archive. "
            << "Expected:"
            << " archive state stream count: " << 1
            << ", group stream count: " << registered_groups.size()
            << ", object stream count: " << object_to_group.size()
            << ".";
        throw ExCorruptedArchive( ss.str() );
        }
    
    // Note: We do not verify archive state stream because it is verified
    // by read_groups_from_state_stream().
    
    // Note: We do not verify group streams because they are verified
    // by read_group_states_and_objects().

    // Note: Verify object streams.
    ObjectIdToGroupMap::const_iterator it;
    for( it = object_to_group.begin(); it != object_to_group.end(); ++it )
        {
        const ObjectId &    object_id = it->first;
        GroupId             group_id = it->second;
        InputStream         stream;
        stream.open( archive, group_id, object_id );
        if( !stream.good() )
            {
            std::ostringstream ss;
            ss  << "Unable to open stream for object " << object_id << ".";
            throw ExCorruptedArchive( ss.str() );
            }
        }
    }


void Archive::My::log_archive_open()
    {
    std::ostringstream  ss;
    unsigned int        total_count = archive->get_stream_count();
    unsigned int        state_stream_count = 1;
    unsigned int        group_count = registered_groups.size();
    unsigned int        object_count = object_to_group.size();
    unsigned int        additional_count = total_count - ( state_stream_count + group_count + object_count );
    unsigned int        schedule_size = group_schedule.get_size();
    
    unsigned int        active_object_count = 0;
    unsigned int        inactive_object_count = 0;
    unsigned int        min_group_size = 0xffffffff;
    unsigned int        max_group_size = 0;
    
    GroupInfoMap::const_iterator    group_info_it;
    GroupState                      group_state;
    
    // Collect some object statistics.
    // For each registered group read its GroupState.
    for( group_info_it = registered_groups.begin();
         group_info_it != registered_groups.end();
         group_info_it++ )
        {
        GroupId     group_id = group_info_it->first;

        // Read group state.
        read_group_state( group_id, group_state );
        
        // Get objects from the group.
        const ObjectStateMap &          objects = group_state.get_object_state_map();
        ObjectStateMap::const_iterator  it;
        
        if( max_group_size < objects.size() ) max_group_size = objects.size();
        if( min_group_size > objects.size() ) min_group_size = objects.size();
        
        for( it = objects.begin(); it != objects.end(); ++it )
            {
            const ObjectState & state = it->second;
            
            if( state.has_swap_in_time() )
                {
                active_object_count++;
                }
            else
                {
                inactive_object_count++;
                }
            }
        }
    
    ss  << archive->debug_info_header_string()
        << "Successfully opened. "
        << "Statistics: "
        << total_count << " stream(s), "
        << state_stream_count << " archive state stream(s), "
        << group_count << " group(s), "
        << object_count << " object(s), "
        << additional_count << " additional stream(s), "
        << schedule_size << " scheduled group(s), "
        << ( group_count - schedule_size ) << " inactive group(s), "
        << active_object_count << " active object(s), "
        << inactive_object_count << " inactive object(s), "
        << ( group_count ? ( object_count / (float) group_count ) : 0 ) << " average group size, "
        << ( group_count ? min_group_size : 0 ) << " min group size, "
        << max_group_size << " max_group_size.";

    Global::log_info
        (
        Status::FACILITY_ARCHIVATION,
        Status::PRIORITY_NORMAL,
        ss.str()
        );
    }


void Archive::My::write_archive_state
    (
    OutputStream & file
    )
    {
    hope( opened_groups.empty() );
    hope( !read_only );

    TextWriter              tw( file );
    Serializer::Description desc;
    SInt< UInt32, 32 >      group_count;
    GroupInfoMap::iterator  group_info_it;

    // Get number of all groups.
    group_count = registered_groups.size();
    
    tw.line() << "group_count";
    tw.write_space();

    // Write number of groups stored in the archive.
    group_count.text_write( tw, desc );
    
    tw.end_line();
    
    tw.line() << "group_ids";

    // For each group write its group id...
    for( group_info_it = registered_groups.begin();
         group_info_it != registered_groups.end();
         group_info_it++ )
        {
        GroupId     group_id = group_info_it->first;
        
        tw.write_space();
        text_write_group_id( tw, desc, group_id );
        }
    }
    
    

    //
    // Read/write group state.
    //


void Archive::My::read_group_state
    (
    GroupId      group_id,
    GroupState & group_state
    )
    {
    InputStream    group_stream;
    group_stream.open( archive, ArchiveFile::group_id_to_stream_name( group_id ) );
    if( !group_stream.good() )
        {
        throw ExCorruptedArchive( archive->err_open_group_stream_string( group_id ) );
        }
    read_group_state( group_id, group_stream, group_state );
    }


void Archive::My::read_group_state
    (
    GroupId       group_id,
    InputStream & file,
    GroupState &  group_state
    )
    {
    Serializer::Description desc;
    TextReader              tr( file );
    tr.next_line_throw();
    try
        {
        // Read the group state.
        group_state.text_read( tr, desc );
        }
    catch( std::exception & e )
        {
        std::ostringstream ss;
        ss  << archive->err_read_group_state_string( group_id ) << ' ' << e.what();
        throw ExCorruptedArchive( ss.str() );
        }
    }


void Archive::My::write_group_state
    (
    GroupId            group_id,
    const GroupState & group_state
    )
    {
    archive->invalidate();

    OutputStream    group_stream;
    group_stream.open( archive, ArchiveFile::group_id_to_stream_name( group_id ) );
    if( !group_stream.good() )
        {
        throw ExCanNotCreateArchiveStream( archive->err_create_group_stream_string( group_id ) );
        }
    hope( !group_stream.fail() );

    write_group_state( group_id, group_stream, group_state );
    
    if( group_stream.fail() )
        {
        throw ExCanNotCreateArchiveStream( archive->err_create_group_stream_string( group_id ) );
        }
    }
    
    
void Archive::My::write_group_state
    (
    GroupId            group_id,
    OutputStream &     output_stream,
    const GroupState & group_state
    )
    {
    Serializer::Description desc;
    TextWriter              tw( output_stream );

    // Write the group state.
    group_state.text_write( tw, desc );
    }


Archive::GroupId Archive::My::generate_group_id()
    {
    hope( next_group_id != INVALID_GROUP_ID );
    GroupId r = next_group_id++;
    hope( !archive->is_group_registered( r ) );
    return r;
    }

    
bool Archive::My::is_group_opened
    (
    GroupId group_id
    ) const
    {
    return opened_groups.find( group_id ) != opened_groups.end();
    }


bool Archive::My::get_group_state
    (
    GroupId      group_id,
    GroupState & group_state
    )
    {
    GroupStateMap::iterator it = opened_groups.find( group_id );

    if( it != opened_groups.end() )
        {
        // Found. The group is opened.
        group_state = it->second;
        return true;
        }
    else
        {
        // The group is not opened.
        if( archive->is_group_registered( group_id ) )
            {
            read_group_state( group_id, group_state );
            return true;
            }
        else
            {
            // The group doesn't exist in the archive.
            return false;
            }
        }
    }


void Archive::My::reschedule
    (
    const GroupState &          group_state,
    GroupInfoMap::iterator      group_info_it
    )
    {
    GroupInfo & group_info  = group_info_it->second;

    if( group_state.has_swap_in_time() )
        {
        // Insert the group to the archive.
        if( group_info.schedule_index.is_valid() )
            {
            group_schedule.change_key( group_info.schedule_index, group_state.get_swap_in_time() );
            }
        else
            {
            group_schedule.push( group_state.get_swap_in_time(), group_info_it );
            }
        hope( group_info.schedule_index.is_valid() );
        }
    else
        {
        // Remove the group from the schedule.
        if( group_info.schedule_index.is_valid() )
            {
            group_schedule.pop( group_info.schedule_index );
            hope( !group_info.schedule_index.is_valid() );
            }
        }
    }


void Archive::My::invalidate_object
    (
    const ObjectId & object_id
    )
    {
    GroupInfoMap::iterator          group_info_it;
    GroupStateMap::iterator         group_state_it;
    GroupId                         group_id;

    archive->invalidate();

    // Check if the object is registered.
    group_info_it = get_group_info_by_object( object_id );

    // Get group id.
    group_id = group_info_it->first;

    // Open the group.
    group_state_it = open_group( group_id );

    // Invalidate the object.
    GroupState & group_state = group_state_it->second;

    if( group_state.is_object_valid( object_id ) )
        {
        /** @todo FIXME Check if there are some opened streams? */
        remove_object_stream( group_id, object_id );
        group_state.invalidate_object( object_id );
        }
    }


Archive::My::GroupStateMap::iterator Archive::My::open_group
    (
    GroupId group_id
    )
    {
    GroupStateMap::iterator    group_state_it;

    archive->invalidate();

    // Check if the group is already opened.
    group_state_it = opened_groups.find( group_id );
    if( group_state_it != opened_groups.end() )
        {
        return group_state_it;
        }

    // Check if the group is registered.
    get_group_info( group_id );

    GroupState  group_state;

    // Read the group from the archive stream.
    read_group_state( group_id, group_state );
    hope( group_state.is_valid() ); /** @todo FIXME Dat jako test soucasti read_group_state()? */

    remove_group_stream( group_id );

    group_state_it = opened_groups.insert
        (
        GroupStateMap::value_type
            (
            group_id,
            group_state
            )
        ).first;

    return group_state_it;
    }


void Archive::My::validate_group
    (
    GroupId            group_id,
    const GroupState & group_state
    )
    {
    hope( group_state.is_valid() );
    hope( is_group_opened( group_id ) );

    archive->invalidate();
    report_validate_group( group_id );

    // Write the group state.
    write_group_state( group_id, group_state );

    // Close the group.
    opened_groups.erase( group_id );
    }


void Archive::My::remove_group_stream
    (
    GroupId group_id
    )
    {
    archive->invalidate();
    report_remove_group_stream( group_id );
    archive->remove( ArchiveFile::group_id_to_stream_name( group_id ) );
    }


void Archive::My::remove_object_stream
    (
    GroupId          group_id,
    const ObjectId & object_id
    )
    {
    archive->invalidate();
    report_remove_object_stream( object_id );
    archive->remove( object_id_to_stream_name( group_id, object_id ) );
    }


Archive::My::GroupInfoMap::iterator Archive::My::get_group_info
    (
    GroupId group_id
    )
    {
    GroupInfoMap::iterator it = registered_groups.find( group_id );
    if( it == registered_groups.end() )
        {
        throw ExArchiveGroupNotFound( archive->err_group_not_registered_string( group_id ) );
        }
    return it;
    }


Archive::My::GroupInfoMap::iterator Archive::My::get_group_info_by_object
    (
    const ObjectId & object_id
    )
    {
    ObjectIdToGroupMap::iterator it = object_to_group.find( object_id );
    if( it == object_to_group.end() )
        {
        throw ExArchiveObjectNotFound( archive->err_object_not_registered_string( object_id ) );
        }
    
    GroupId                 group_id = it->second;
    GroupInfoMap::iterator  group_info_it = registered_groups.find( group_id );
    hope( group_info_it != registered_groups.end() );
    return group_info_it;
    }


Archive::My::GroupStateMap::iterator Archive::My::get_opened_group_state
    (
    GroupId group_id
    )
    {
    GroupStateMap::iterator it = opened_groups.find( group_id );
    if( it == opened_groups.end() )
        {
        throw ExArchiveGroupNotFound( archive->err_group_not_opened_string( group_id ) );
        }
    return it;
    }


Archive::My::GroupStateMap::iterator Archive::My::get_opened_group_state
    (
    const Archive::My::GroupInfoMap::iterator & group_info_it
    )
    {
    GroupId group_id = group_info_it->first;
    return get_opened_group_state( group_id );
    }
    
    
    //
    // Reports.
    //


void Archive::My::report_validate_object
    (
    const ObjectId & object_id
    )
    {
    if( report_flags & REPORT_VALIDATE_OBJECT )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Validating object " << object_id << ".";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void Archive::My::report_validate_group
    (
    GroupId group_id
    )
    {
    if( report_flags & REPORT_VALIDATE_GROUP )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Validating group " << archive->debug_group_info_string( group_id ) << ".";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void Archive::My::report_copy_object
    (
    Archive &        dest_archive,
    const ObjectId & object_id
    )
    {
    if( report_flags & REPORT_COPY_OBJECT )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Copying object " << object_id
            << " to destination archive "
            << dest_archive.debug_info_string()
            << ".";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void Archive::My::report_invalidate()
    {
    if( report_flags & REPORT_INVALIDATE )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Invalidating the archive. ";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void Archive::My::report_remove_group_stream
    (
    GroupId group_id
    )
    {
    if( report_flags & REPORT_REMOVE_GROUP_STREAM )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Removing group stream " << archive->debug_group_info_string( group_id ) << ".";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void Archive::My::report_remove_object_stream
    (
    const ObjectId & object_id
    )
    {
    if( report_flags & REPORT_REMOVE_OBJECT_STREAM )
        {
        std::ostringstream ss;
        ss  << archive->debug_info_header_string()
            << "Removing stream for object " << object_id << ".";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


    //
    // Miscenalleous.
    //

Archive::GroupId Archive::My::text_read_group_id
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    SInt< UInt32, 32 >  id;
    id.text_read( tr, desc );
    return id;
    }

    
void Archive::My::text_write_group_id
    (
    TextWriter &                    tw,
    const Serializer::Description & desc,
    GroupId                         group_id
    )
    {
    SInt< UInt32, 32 >  id = group_id;
    id.text_write( tw, desc );
    }


} // namespace Massiv::Core
} // namespace Massiv
