/** @file archive_internal.h
 * #Archive internal data structures.
 */

/*
 * begin        : Fri Jun 20 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ARCHIVE_INTERNAL_H
#define MASSIV_ARCHIVE_INTERNAL_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_ARCHIVE_H
#include "object/archive.h"
#endif

#ifndef MASSIV_HEAP_H
#include "utility/heap.h"
#endif

#ifndef MASSIV_HEAP_IMPL_H
#include "utility/heap_impl.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Internal structure of #Archive.
 *
 * Each group registered in the archive has one group info (#GroupInfo) which
 * is stored in group info map (#GroupInfoMap). Group info stores heap index
 * into group schedule (#GroupSchedule) and group schedule stores iterators
 * back to group info map.
 */
class Archive::My
    {
    
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );

public:

    //--------------------------------------------------------------------------

    /**
     * GroupInfo.
     * Currently #GroupInfo stores only position of a group inside group schedule
     * (heap index #HeapIndex). Each group registered in an archive has its own
     * #GroupInfo. Groups which are not scheduled in the schedule have invalid
     * heap index intheir #GroupInfo.
     */
    class GroupInfo
        {
        
    public:

        /** Constructor. */
        GroupInfo()
            {
            hope( !schedule_index.is_valid() );
            }

        /** Copy constructor. */
        GroupInfo
            (
            const GroupInfo & other
            ) :
            schedule_index( other.schedule_index )
            {
            }

        /** Operator=. */
        GroupInfo & operator=
            (
            const GroupInfo & other
            )
            {
            schedule_index = other.schedule_index;
            return *this;
            }

    public:

        HeapIndex   schedule_index;
            /**< Group schedule index. */

        }; // GroupInfo


    /**
     * Group info map.
     */
    typedef ext::hash_map
            <
            GroupId,
            GroupInfo
            > GroupInfoMap;

    //--------------------------------------------------------------------------

    /**
     * GroupScheduleIndexChanged monitors group position changes in group schedule.
     * If position of a group chages in the schedule, #GroupScheduleIndexChanged
     * updated @c schedule_index member of #GroupInfo of the group.
     */
    struct GroupScheduleIndexChanged
        {
        /** Called whenever group is moved in the group schedule. */
        void operator()
            (
            const GroupInfoMap::iterator & it,
            HeapIndex                       new_index
            ) const
            {
            GroupInfo & group_info = it->second;
            group_info.schedule_index = new_index;
            }

        }; // GroupScheduleIndexChanged


    /**
     * Group schedule heap.
     * Stores groups sorted by "swap-in" time (time when all objects
     * from a group must be swapped in).
     */
    typedef Heap
            <
            STime,                      // Swap in time.
            GroupInfoMap::iterator,     // Iterator to group info map.
            STime::Less,                // Time comparator.
            GroupScheduleIndexChanged   // Structure used to monitor changes in the heap.
            > GroupSchedule;
            
            
    //--------------------------------------------------------------------------

    /**
     * Object state.
     * Stores information about one object stored in a group.
     * Stores time when the object must be swapped-in and flag indicating
     * whether stream of the object in the archive is valid (ie whether
     * the object is properly stored in the archive).
     */
    class ObjectState
        {
    public:

        /** Constructor. */
        ObjectState();

        /** Copy constructor. */
        ObjectState
            (
            const ObjectState & other
            );

        /** Operator=. */
        ObjectState & operator=
            (
            const ObjectState & other
            );

    public:

        /**
         * Constructor.
         */
        ObjectState
            (
            const STime *       swap_in_time,
            bool                valid_stream
            );

    public:

        /**
         * Read from text stream.
         * @note Validity flag is set to true.
         */
        void text_read
            (
            TextReader &                    tr,
            const Serializer::Description & desc
            );

        /**
         * Write to text stream.
         * @note Validity flag is not written to the stream.
         * @pre Validity flag is set to true.
         */
        void text_write
            (
            TextWriter &                    tw,
            const Serializer::Description & desc
            ) const;

    public:

        /** Returns whether the object has swap-in time. */
        bool has_swap_in_time() const
            {
            return swap_in_time_valid;
            }

        /** Returns true if the stream of the object is valid. */
        bool is_valid() const
            {
            return valid_stream;
            }

        /** Validate. Only set valid stream flag to true. */
        void validate()
            {
            valid_stream = true;
            }

        /** Invalidate. Only set valid stream flag to false. */
        void invalidate()
            {
            valid_stream = false;
            }

    public:

        SBool           swap_in_time_valid;
            /**< True if the @c swap_in_time is valid. */

        STime           swap_in_time;
            /**< Time when the object must be swapped in. Valid only
             * if @c swap_in_time_valid is true.
             */

        bool            valid_stream;
            /**< Whether the object stream in the archive is valid. */
            
        }; // ObjectState

    /**
     * Object state map.
     */
    typedef ext::hash_map
            <
            ObjectId,
            ObjectState,
            ObjectId::Hash,
            ObjectId::EqualKey
            > ObjectStateMap;


    //--------------------------------------------------------------------------
    
    /**
     * Group state.
     * Stores #ObjectState for each object from the group.
     * If all object states are valid, the entire group state is marked valid.
     * If there is at least one invalid object state, the entire group is marked
     * invalid.
     * Empty group is marked valid.
     */
    class GroupState
        {
    public:

        /** Constructor. */
        GroupState();

        /** Copy constructor. */
        GroupState
            (
            const GroupState & other
            );

        /** Operator. */
        GroupState & operator=
            (
            const GroupState & other
            );

    public:

        /**
         * Add new object to the group.
         * Returns true if swap-in time of the group
         * has changed.
         * @pre The object is not already registered in the group.
         */
        bool add_object
            (
            const ObjectId &    object_id,
            const STime *       object_swap_in_time,
            bool                object_stream_valid
            );

        /**
         * Add new object to the group.
         * Returns true if swap-in time of the group
         * has changed.
         * @pre The object is not already registered in the group.
         */
        bool add_object
            (
            const ObjectId &    object_id,
            const ObjectState & object_state
            );

        /**
         * Remove object from the group.
         * It is safe to call this function if @a object
         * is not registered in the group.
         */
        void remove_object
            (
            const ObjectId & object
            );

        /**
         * Validate object from the group.
         */
        void validate_object
            (
            const ObjectId & object_id
            );

        /**
         * Invalidate object from the group.
         */
        void invalidate_object
            (
            const ObjectId & object_id
            );

        /**
         * Returns whether the group has swap-in time.
         * Groups with swap-in time are automatically
         * swapped in by the system when the simulation time
         * reaches the swap-in time.
         */
        bool has_swap_in_time() const
            {
            return swap_in_time_valid;
            }
            
        /**
         * Get swap-in time.
         */
        const STime & get_swap_in_time() const
            {
            hope( has_swap_in_time() );
            return swap_in_time;
            }

        /**
         * Returns true if the group is valid.
         * The group is valid if all its object states are valid.
         * Empty group is valid.
         */
        bool is_valid() const;

        /**
         * Returns true if the @a object from the group is valid.
         * If there isn't such object in the group, false is returned.
         */
        bool is_object_valid
            (
            const ObjectId & object_id
            ) const;
            
        /**
         * Enumerate all objects from the group.
         */
        void collect_objects
            (
            ObjectShared::ObjectIdList & objects,
            bool clear = true
            );

    public:

        /**
         * Read from text stream.
         */
        void text_read
            (
            TextReader &                    tr,
            const Serializer::Description & desc
            );

        /**
         * Write to text stream.
         * @pre The group is valid.
         */
        void text_write
            (
            TextWriter &                    tw,
            const Serializer::Description & desc
            ) const;
            
    public:
    
        /**
         * For debug purposes only.
         */
        const ObjectStateMap & get_object_state_map() const
            {
            return objects;
            }

    private:

        bool            swap_in_time_valid;
            /**< True if the @c swap_in_time is valid. */

        STime           swap_in_time;
            /**< Time when the group must be swapped in. Valid only
             * if @c swap_in_time_valid is true. Swap-in time of the
             * group is the oldest swap-in time of its objects.
             */

        ObjectStateMap  objects;
            /**< List of all objects in the group. */
            
        }; // GroupState
        
    //--------------------------------------------------------------------------

    /**
     * Group state map.
     */
    typedef ext::hash_map
            <
            GroupId,
            GroupState
            > GroupStateMap;
            
    /**
     * Map converting object id to its group.
     * The map also contains #ObjectInfo of the object.
     */
    typedef ext::hash_map
            <
            ObjectId,
            GroupId,
            ObjectId::Hash,
            ObjectId::EqualKey
            > ObjectIdToGroupMap;

    //--------------------------------------------------------------------------

public:

    /** Constructor. */
    My  (
        Mode        open_mode
        );
    
public:

    /**
     * Read archive state stream.
     * Archive must be just created and empty.
     */
    void read_archive_state
        (
        InputStream &  file
        );

private:

    /**
     * Read group ids of all groups from archive state stream
     * and register them.
     */
    void read_groups_from_state_stream
        (
        TextReader &                tr,
        Serializer::Description &   desc
        );

    /**
     * For each registered group read corresponding group state
     * and register all objects from the state to the archive.
     */
    void read_group_states_and_objects();

    /**
     * Verify streams in the archive after it has been opened.
     * This function is part of archive loading and should be
     * called only from read_archive_state().
     */
    void verify_archive();

    /**
     * Log miscenalleous info about the archive. This function is slow
     * and should be called only from read_archive_state().
     */
    void log_archive_open();

public:    

    /**
     * Write archive state stream. Writes list of all group ids registered
     * in the archive.
     * All groups must be closed.
     */
    void write_archive_state
        (
        OutputStream &  file
        );
        
public:

    //
    // Read/write group state.
    //

    /**
     * Read group state and store it in @a group_state.
     * Validity of the returned group state is not tested.
     * @note If exception is thrown, @a group_state
     * may be modified.
     */
    void read_group_state
        (
        GroupId         group_id,
        GroupState &    group_state
        );

    /**
     * Read group state and store it in @a group_state.
     * Validity of the returned group state is not tested.
     * @note If exception is thrown, @a group_state
     * may be modified.
     */
    void read_group_state
        (
        GroupId         group_id,
        InputStream &   input_stream,
        GroupState &    group_state
        );

    /**
     * Write group state.
     * Validity of the group is not tested.
     */
    void write_group_state
        (
        GroupId            group_id,
        const GroupState & group_state
        );

    /**
     * Write group state.
     * Validity of the group is not tested.
     */
    void write_group_state
        (
        GroupId            group_id,
        OutputStream &     output_stream,
        const GroupState & group_state
        );
        
public:

    /**
     * Generate new unique group id.
     */
    GroupId generate_group_id();

    /**
     * Check whether the @a group is opened.
     * False is returned if the group is not registered in the archive.
     */
    bool is_group_opened
        (
        GroupId group
        ) const;

    /**
     * Get group state.
     * If the group is closed, its state is read from the archive, but
     * the group is not opened.
     * Returns false if the group doesn't exist.
     * It is safe to call this function on read only archive.
     * @note If exception is thrown, @a group_state
     * may be modified.
     */
    bool get_group_state
        (
        GroupId      group_id,
        GroupState & group_state
        );

    /**
     * Reschedule group.
     */
    void reschedule
        (
        const GroupState &      group_state,
        GroupInfoMap::iterator  group_info_it
        );

    /**
     * Invalidate object.
     * Exception is thrown if the object is not registered.
     * The group of the object is opened.
     */
    void invalidate_object
        (
        const ObjectId & object_id
        );

    /**
     * Open group.
     * The group is invalidated.
     */
    GroupStateMap::iterator open_group
        (
        GroupId group_id
        );

    /**
     * Validate group.
     */
    void validate_group
        (
        GroupId            group_id,
        const GroupState & group_state
        );

    /**
     * Remove group stream from the archive volume.
     */
    void remove_group_stream
        (
        GroupId group_id
        );

    /**
     * Remove object stream from the archive volume.
     */
    void remove_object_stream
        (
        GroupId          group_id,
        const ObjectId & object_id
        );
        
    /**
     * Get group info.
     * Exception is thrown if the group is not registered.
     * Always return valid iterator.
     */
    GroupInfoMap::iterator get_group_info
        (
        GroupId group_id
        );

    /**
     * Get group info by object id from the group.
     * Exception is thrown if the object is not registered.
     * Always return valid iterator.
     */
    GroupInfoMap::iterator get_group_info_by_object
        (
        const ObjectId & object_id
        );

    /**
     * Get opened group state.
     * Exception is thrown if the group is not opened.
     * Always return valid iterator.
     * @note Doesn't check if the group is registered!
     */
    GroupStateMap::iterator get_opened_group_state
        (
        GroupId group_id
        );

    /**
     * Get opened group state.
     * Exception is thrown if the group is not opened.
     * Always return valid iterator.
     * @a group_info_it must be valid iterator.
     * @note Doesn't check if the group is registered!
     */
    GroupStateMap::iterator get_opened_group_state
        (
        const GroupInfoMap::iterator & group_info_it
        );

public:

    //
    // Reports.
    //

    /** Report. */
    void report_validate_object
        (
        const ObjectId & object_id
        );

    /** Report. */
    void report_validate_group
        (
        GroupId group_id
        );

    /** Report. */
    void report_copy_object
        (
        Archive &        dest_archive,
        const ObjectId & object_id
        );

    /** Report. */
    void report_invalidate();

    /** Report. */
    void report_remove_group_stream
        (
        GroupId group_id
        );

    /** Report. */
    void report_remove_object_stream
        (
        const ObjectId & object_id
        );

    /**
     * Destroy all opened groups.
     */
    void destroy_all_opened_groups();

public:

    //
    // Miscenalleous.
    //
    
    /**
     * Read #GroupId from stream.
     */
    static GroupId text_read_group_id
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );
    
    /**
     * Write #GroupId to stream.
     */
    static void text_write_group_id
        (
        TextWriter &                    tw,
        const Serializer::Description & desc,
        GroupId                         group_id
        );

public:

    Archive *               archive;
        /**< Archive which owns this My object. Always valid pointer.
         * ASAP instance of My is create, it must be set to corresponding
         * Archive object.
         */

    GroupInfoMap            registered_groups;
        /**< List of all groups registered in the archive. */

    GroupStateMap           opened_groups;
        /**< List of all opened groups. */

    ObjectIdToGroupMap      object_to_group;
        /**< Map converting (registered) object id to its group. 
         * The map also contains #ObjectInfo of the object.
         */

    GroupSchedule           group_schedule;
        /**< Group schedule. Stores groups sorted by "swap-in" time
         * (time when all objects from a group must be swapped in).
         */

    GroupId                 next_group_id;
        /**< Next available group id. */
        
    bool                    read_only;
        /**< If true any write access to the archive will throw exception. */

    bool                    valid;
        /**< Whether the archive is valid. Any write access (like opening
         * output archive stream) will invalidate the archive.
         */

    unsigned int            report_flags;
        /**< Report flags. */

    }; // Archive::My

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVE_INTERNAL_H
