/** @file archive_manager.cpp
 * Archive manager implementation.
 */

/*
 * begin        : Thu Oct 17 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/archive_manager.h"
#include "object/archive.h"
#include "object/object.h"
#include "object/object_common.h"
#include "object/object_manager.h"
#include "node/node_manager.h"
#include "object/collect_group.h"
#include "database/archive_database.h"
#include "utility/debug.h"
#include "sysevent/system_event_queue.h"
#include "sysevent/archivation_msg.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "serializer/serializer.h"
#include "system/system.h"
#include <fstream>
#include <iomanip> // IO manipulators


namespace Massiv {
namespace Core {

namespace {


/** Check whether @a arch_id is id of initial archive. */
bool is_initial_archive
    (
    const ArchivationId & arch_id
    )
    {
    ArchivationId initial_arch_id;    // Create invalid id...
    initial_arch_id.increment();      // ...and made lowest valid id from it.
    return arch_id == initial_arch_id;
    }

    
} // anonymous namespace


//-----------------------------------------------------------------------------
// CommonArchiveOperations implementation.
//-----------------------------------------------------------------------------

void CommonArchiveOperations::register_objects_to_archive
    (
    Archive *                        archive,
    const ObjectShared::ObjectList & object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::CommonArchiveOperations::register_objects_to_archive()" );
    hope( archive );

    // Create new archive group and register all objects to it.
    Archive::GroupId                         group_id;
    STime                                    event_time;
    ObjectShared::ObjectList::const_iterator it;

    // Create new archive group.
    group_id = archive->create_group();

    // Register all the objects to the group.
    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty *    object = ( *it );
        STime *             t;
        
        // Only archivable objects may be stored in an archive file.
        // Because swap file is also an archive file, only archivable
        // objects may be swapped-out.
        hope( object->is_archivable() );
        hope( !archive->is_object_registered( object->get_object_id() ) );

        if( object->has_scheduled_event() )
            {
            event_time = object->get_time_of_first_event();
            t = &event_time;
            }
        else
            {
            t = NULL;
            }

        archive->register_object( object->get_object_id(), group_id, t );
        }
    }

void CommonArchiveOperations::text_write_object_to_archive
    (
    Archive *                       archive,
    ObjectProperty *                object,
    const Serializer::Description & desc
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::CommonArchiveOperations::text_write_object_to_archive()" );
    hope( archive );
    hope( !object->is_replica() );

    const ObjectId & object_id = object->get_object_id();

    // The objects must be already registered in the archive.
    hope( archive->is_object_registered( object_id ) );

        { // Start of block - output stream must be destroyed befor object validation.
        Archive::OutputStream   out_stream;
        Archive::GroupId        group_id = archive->get_object_group_id( object_id );

        out_stream.open( archive, group_id, object_id );
        if( out_stream.fail() )
            {
            std::ostringstream ss;
            ss  << "Unable to create archive output object stream. "
                << "Archive: " << archive->debug_info_string() << ", "
                << "object: " << archive->debug_object_info_string( group_id, object_id )
                << ".";
            throw ExCanNotCreateArchiveStream( ss.str() );
            }

        TextWriter              tw( out_stream );
        
        // First write complete object id of the object.
        // This is used when packing unpacked archive.
        hope( object_id.is_object_info_available() );
        object_id.text_write( tw, desc );
        tw.end_line();

        // Write object's state.
        object->text_write( tw, desc );
        tw.end_line();

        if( out_stream.fail() )
            {
            std::ostringstream ss;
            ss  << "Error writing archive output object stream. "
                << "Archive: " << archive->debug_info_string() << ", "
                << "object: " << archive->debug_object_info_string( group_id, object_id )
                << ".";
            /** @todo FIXME Another exception? */
            throw ExArchivationError( ss.str() );
            }
        }

    archive->validate_object( object_id );
    }

//-----------------------------------------------------------------------------
// ArchiveManager::NodeState implementation.
//-----------------------------------------------------------------------------


ArchiveManager::NodeState::NodeState() :
    phase( PHASE_IDLE )
    {
    }


ArchiveManager::NodeState::~NodeState()
    {
    }


ArchiveManager::NodeState::NodeState
    (
    const NodeState & s
    ) :
    phase( s.phase )
    {
    }


ArchiveManager::NodeState & ArchiveManager::NodeState::operator=
    (
    const NodeState & s
    )
    {
    phase = s.phase;
    return *this;
    }


//-----------------------------------------------------------------------------
// ArchiveManagerSettings implementation.
//-----------------------------------------------------------------------------

namespace {

const Float32 DEFAULT_ARCHIVATION_DURATION =  1.0f * 60.0f; // one minute
const Float32 DEFAULT_ARCHIVATION_DELAY    = 10.0f * DEFAULT_ARCHIVATION_DURATION;

}

ArchiveManager::ArchiveManagerSettings::ArchiveManagerSettings() :
    Settings( "ArchiveManager" ),
    local_archivation_duration( access_float( "local_archivation_duration", DEFAULT_ARCHIVATION_DURATION ) ),
    auto_archivation_enabled( access_boolean( "auto_archivation_enabled", true ) ),
    auto_archivation_delay( access_float( "auto_archivation_delay", DEFAULT_ARCHIVATION_DELAY ) )
    {
    }


//-----------------------------------------------------------------------------
// ArchiveManager implementation.
//-----------------------------------------------------------------------------


ArchiveManager::ArchiveManager() :
    phase( PHASE_IDLE ),
    is_srpc_delayed_start_archivation_message( false ),
    archive( NULL ),
    total_object_to_archive_count( 0 ),
    total_archived_object_count( 0 ),
    auto_archivation_enabled( false ),
    started_node_count( 0 ),
    finished_node_count( 0 ),
    pending_archivation_start( false )
    {
    }


ArchiveManager::~ArchiveManager()
    {
    // Close archive file.
    if( archive )
        {
        delete archive;
        archive = NULL;
        }
    }
    
    
void ArchiveManager::initialize
    (
    const ArchivationId & latest_archivation_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::initialize()" );
    hope( phase == PHASE_IDLE );
    hope( !archivation_in_progress() );
    
    idle_phase_start_time = Global::system_time();
    archivation_id = latest_archivation_id;
    
        {
        std::ostringstream  ss;
        ArchivationId       arch_id = archivation_id;
        arch_id.increment();
        
        ss  << "ArchivationId of the next archivation will be \""
            << arch_id << "\".";
        
        Global::log_info
            (
            Status::FACILITY_ARCHIVATION,
            Status::PRIORITY_HIGH,
            ss.str()
            );
        }
    }

    
void ArchiveManager::enable_auto_archivation()
    {
    switch( Global::get_local_node_id().get_type() )
        {
        case NodeId::TYPE_SERVER:
            auto_archivation_enabled = true;
            break;
            
        case NodeId::TYPE_CLIENT:
        case NodeId::TYPE_SERVICE:
        case NodeId::TYPE_ANONYMOUS:
            // Unsupported.
            break;
        }
    }
    
    
void ArchiveManager::disable_auto_archivation()
    {
    auto_archivation_enabled = false;
    }
    
    
bool ArchiveManager::is_auto_archivation_enabled()
    {
    return auto_archivation_enabled;
    }
    

ArchiveManager::RequestStartArchivationResult ArchiveManager::request_start_archivation()
    {
    // Check if not server.
    if( !System::is_server() )
        {
        return RESULT_NODE_IS_NOT_SERVER;
        }
    
    // Check if archivation is already in progress.
    if( archivation_in_progress() )
        {
        return RESULT_ARCHIVATION_IN_PROGRESS;
        }
    
    // Check if some sync rpcs are in progress. Archivation can't
    // start if there are some pending sync rpcs.
    if( System::is_waiting_for_rpc_reply() )    
        {
        pending_archivation_start = true;
        return RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC;
        }
        
    // Check if the manager was initialized.
    if( archivation_id.is_invalid() )
        {
        // The manager was not initialized yet - we don't know
        // archivation id from which new archivations should begin.
        pending_archivation_start = true;
        return RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC;
        }
    
    // Start new archivation.
    start_archivation( false );
    
    return RESULT_ARCHIVATION_IN_PROGRESS;
    }


void ArchiveManager::request_start_initial_archivation()
    {
    hope( !archivation_in_progress() );
    hope( !System::is_waiting_for_rpc_reply() );
    
    start_archivation( true );
    }


ArchivationId ArchiveManager::get_current_archivation_id() const
    {
    return archivation_id;
    }


bool ArchiveManager::request_start_archivation
    (
    const ArchivationId & the_archivation_id
    )
    {
    // Check if archivation is already in progress.
    if( archivation_in_progress() )
        {
        return false;
        }
    
    /** @todo 
     * Pokud by SRPC melo byt trasnparentni pro uzivatele,
     * tak bychom meli poslat zpravu neco jako "request start archivation",
     * ktera se vyridi, jakmile klesneme v urovni zanoreni zpet na 0.
     * Na to se nemusi posilat zprava - staci si zaznamenat flazik
     * v archive manageru, podobne jako is_srpc_delayed_start_archivation_message.
     *
     * Komentar uz je mimo misu.
     * Pristup k archivaci primo pres archivation_id by melo byt
     * zpristupneno jen pres specialni funkci a nemelo by vyuzivat
     * pending_archivation_start.
     */
    if( System::is_waiting_for_rpc_reply() )
        {
        return false;
        }
    
    start_archivation( the_archivation_id );
    return true;
    }


void ArchiveManager::start_archivation
    (
    bool initial_archivation
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::start_archivation( bool )" );
    hope( phase == PHASE_IDLE );
    
    if( initial_archivation )
        {
        ArchivationId   arch_id;    // Create invalid id...
        arch_id.increment();        // ...and made lowest valid id from it.
        start_archivation( arch_id );
        }
    else
        {
        hope( !archivation_id.is_invalid() ); // Archivation manager must be initialized with latest valid archivation id.
        archivation_id.increment();
        start_archivation( archivation_id );
        }
    }


void ArchiveManager::start_archivation
    (
    const ArchivationId & arch_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::start_archivation( ArchivationId & )" );
    hope( phase == PHASE_IDLE );
    hope( !arch_id.is_invalid() );
    
    hope( !System::is_waiting_for_rpc_reply() );

    // Log.
        {
        std::ostringstream ss;
        ss  << "Starting new archivation...";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );
        }

    cleanup_archivation();
    
    // Clear pending requests.
    pending_archivation_start = false;

    ObjectManager &        object_manager = Global::object_manager();
    Archive *              swap = object_manager.get_swap();

    // Setup archivation id.
    archivation_id = arch_id;

    // Get all archivation nodes participating in the archivation.
    enumerate_archivation_nodes();

    // Move all objects which are stored in migration messages
    // in object manager to the memory so they can be archived.
    object_manager.prepare_for_archivation();

    // Enumerate swapped-out objects which must be archived.
    if( swap )
        {
        // Note: This will enumerate only archivable objects.
        swap->enumerate_registered_objects( objects_to_copy );
        }

    // Enumerate active objects which must be archived.
    object_manager.gc.enter_archive_mode( objects_to_archive );

    total_object_to_archive_count = objects_to_copy.size() + objects_to_archive.size();
    total_archived_object_count   = 0;

    // Setup archivation timeouts.
        {
        SystemTime curr_time = Global::system_time();
        Float64    duration  = settings.local_archivation_duration;
        
        // Override archivation duration for initial archive.
        if( is_initial_archive( arch_id ) )
            {
            duration = 1.0; // One second.
            }

        /** @todo FIXME Set default value. Log this. */
        hope( duration >= 0 );

        archivation_start_time  = curr_time;
        archivation_finish_time = curr_time + duration;
        }

    // Create archive file.
        {
        ArchiveDatabase & archive_database = Global::archive_database();
        
        archive = archive_database.create_regular_archive( archivation_id );
        archive->enable_report( Archive::REPORT_ALL );
        }

    // Store simulation state.
    object_manager.archive_simulation_state( archive );

    // Send archivation started status message to all archivation nodes.
        {
        ArchivationStatusMessage    msg( ArchivationStatusMessage::ARCHIVATION_STARTED );

        msg.set_archivation_id( archivation_id );
        send_status_message_to_archivation_nodes( msg );
        }

    // Everything is prepared now, let the archivation begin.
    phase = PHASE_LOCAL_ARCHIVATION;
    }


void ArchiveManager::archive_object_state
    (
    ObjectProperty * object
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::archive_object_state()" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( archive != NULL );
    hope( object != NULL );
    hope( !object->is_member_object() );
    hope( !object->is_replica() );
    hope( object->is_archivable() );
    hope( !archive->is_object_valid( object->get_object_id() ) );

    Serializer::Description     desc;
    desc.object_flags = Serializer::Description::OBJECT_ARCHIVE;

    // Register the object and its migration group to the archive file.
    register_object_to_archive( object );

    // The objects is now registered in the archive. Store its state.
    CommonArchiveOperations::text_write_object_to_archive( archive, object, desc );

    total_archived_object_count++;
    }


void ArchiveManager::archive_swap_in_object
    (
    const ObjectId & object_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::archive_swap_in_object()" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    archive_inactive_object( object_id );
    }


void ArchiveManager::schedule_object_for_archivation
    (
    const ObjectId & object_id
    )
    {
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( !archive->is_object_registered( object_id ) );

    objects_to_archive.push( object_id );
    total_object_to_archive_count++;
    }


void ArchiveManager::register_object_to_archive
    (
    ObjectProperty * object
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::register_object_to_archive()" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( object != NULL );
    hope( object->is_archivable() );

    if( archive->is_object_registered( object->get_object_id() ) )
        {
        // The object (and its migration group) is already
        // registered in the archive.
        }
    else
        {
        // The object is not registered in the archive yet.
        // Collect migration group and register all its objects in the archive.

        ObjectShared::ObjectList            objects;

        // Enumerate objects from the migration group.
            {
            GarbageCollector &              gc = Global::garbage_collector();
            CollectObjectGroupOperation     collect_objects( objects );

            gc.enumerate_migration_group( object, collect_objects );
            hope( !objects.empty() );

#ifdef OPTION_DEBUG
            // hope( std::find( objects.begin(), objects.end(), object ) != objects.end() );
            ObjectShared::ObjectList::iterator object_it;
            for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
                {
                ObjectProperty * o = ( *object_it );
                hope( !o->is_replica() );
                hope( o->is_archive_on_write() );
                hope( o->is_archivable() ); // Archivable and non-archivable objects may not
                                              // be in one migration group.
                }
#endif // OPTION_DEBUG
            }

        // Create new archive group and register all objects from
        // the migration group to the archive group.
        CommonArchiveOperations::register_objects_to_archive( archive, objects );
        }
    }


void ArchiveManager::next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::next_tick()" );
    // Note: This test must be done every tick, not only in idle phase tick!
    // The reason is that archivation can be initiated on local node while
    // there are some paused start archivation messages. Doing this test
    // only in idle phase would cause archivation to be never finished.
    if( is_srpc_delayed_start_archivation_message &&
        !( System::is_waiting_for_rpc_reply() )
        )
        {
        /** Unpause start archivation messages which were paused
         * because of pending sync rpc. We are now in base recursion level
         * of the main system loop - the archivation can be started now.
         */
        
        Global::log_debug
            (
            Status::FACILITY_ARCHIVATION,
            Status::PRIORITY_NORMAL,
            "Disabling PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED system message pause flag"
            " for all server nodes..."
            );
         
        // We can't use archivation_node_states, beause it is empty
        // in idle phase. We must enumerate all server nodes.
        
        NodeManager &           node_manager = Global::node_manager();
        SystemEventQueue &      system_event_queue = Global::system_event_queue();
        NodeManager::NodeIdList server_nodes;
        NodeManager::NodeIdList::const_iterator it;
        
        node_manager.enumerate_archivation_nodes( server_nodes );
        
        for( it = server_nodes.begin(); it != server_nodes.end(); ++it )
            {
            const NodeId & node_id = *it;
            unsigned int pause_mask = system_event_queue.get_server_pause_mask( node_id );
            pause_mask &= ~( SystemMessage::PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED );
            system_event_queue.set_server_pause_mask( node_id, pause_mask );
            }

        is_srpc_delayed_start_archivation_message = false;
        }
    
    switch( phase )
        {
        case PHASE_IDLE:
            phase_idle();
            break;
        case PHASE_LOCAL_ARCHIVATION:
            phase_local_archivation();
            break;
        case PHASE_FINISHING:
            phase_finishing();
            break;
        default:
            hope( false );
            break;
        }
    }


void ArchiveManager::phase_idle()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::phase_idle()" );
    if( !pending_archivation_start )
        {
        if( !auto_archivation_enabled )
            {
            return;
            }
        
        if( !settings.auto_archivation_enabled )
            {
            return;
            }
        }
        
    if( archivation_id.is_invalid() )
        {
        // Archive manager was not yet initialized - we don't know archivation id
        // of next archivation.
        return;
        }
    
    if( pending_archivation_start )
        {
        Global::log_info
            (
            Status::FACILITY_ARCHIVATION,
            Status::PRIORITY_NORMAL,
            "Pending archivation start: Initiatiating new archivation as requested some time ago."
            );
        
        switch( request_start_archivation() )
            {
            case RESULT_NODE_IS_NOT_SERVER:
                panic( "Local node is not server!?!" );
                break;
                
            case RESULT_ARCHIVATION_IN_PROGRESS:
                // Archivation started.
                hope( !pending_archivation_start );
                break;
            
            case RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC:
                // There are still some pending sync rpc calls.
                // Lets try luck in next tick.
                Global::log_info
                    (
                    Status::FACILITY_ARCHIVATION,
                    Status::PRIORITY_NORMAL,
                    "Still some pending sync rpc calls. Delaying archivation..."
                    );
                break;
            }

        return;
        }
    
    // Auto archivation.
    
    SystemTime curr_time = Global::system_time();
    Float64    delay = settings.auto_archivation_delay;
    SystemTime arch_time = idle_phase_start_time + delay ;
    
    if( curr_time >= arch_time )
        {
        Global::log_info
            (
            Status::FACILITY_ARCHIVATION,
            Status::PRIORITY_LOWEST,
            "Auto archivation - initiatiating new archivation."
            );
            
        switch( request_start_archivation() )
            {
            case RESULT_NODE_IS_NOT_SERVER:
                panic( "Local node is not server!?!" );
                break;

            case RESULT_ARCHIVATION_IN_PROGRESS:
                // Archivation started.
                break;
            
            case RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC:
                // There are still some pending sync rpc calls.
                // Lets try luck in next tick.
                break;
            }        
        }
    }

#ifdef OPTION_DEBUG
class VerifyObjectEnumerator : public GarbageCollector::ObjectOperation
    {
public:

    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        hope( object );
        hope( !object->is_member_object() );
        hope( object->is_registered() );
        GarbageCollector & gc = Global::garbage_collector();
        
        if( object->is_archive_on_write() )
            {
            std::ostringstream oss;
            bool is_registered_in_archive = archive->is_object_registered( object->get_object_id() );
            bool is_valid_in_archive = archive->is_object_valid( object->get_object_id() );
            
            // Some attributes we are interested in...
            oss << "Local archivation finished but object " << object->get_object_id()
                << " has archive_on_write set on! Details:"
                << "registered_in_archive " << ( is_registered_in_archive ? "yes" : "no" ) << ", "
                << "registered_and_valid_in_archive " << ( is_valid_in_archive ? "yes" : "no" ) << ", "
                << "archivable " << ( object->is_archivable() ? "yes" : "no" ) << ", "
                << "replica " << ( object->is_replica() ? "yes" : "no" ) << ", "
                << "finalizing " << ( object->is_finalizing() ? "yes" : "no" ) << ", "
                << "on_stack " << ( object->is_on_stack() ? "yes" : "no" ) << ". "
                << "\n";
                
            // Other attributes...
            gc.dump_migration_group( oss, object->get_object_id() );
            
            // Exit...
            panic( oss.str().c_str() );
            }
        }
    
    Archive * archive;

    }; // VerifyObjectEnumerator
#endif // OPTION_DEBUG

void ArchiveManager::phase_local_archivation()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::phase_local_archivation()" );
    bool finished;

    // Advance local archivation.
    finished = local_archivation_next_tick();

    // Check if we can switch to next phase.
    if( finished && ( started_node_count == archivation_node_states.size() ) )
        {
        // Local archivation is finished. Send archivation finished
        // status message to all archivation nodes.
        // Note: The message is also send to local node (if it is
        // an archivation node).
        ArchivationStatusMessage    msg( ArchivationStatusMessage::ARCHIVATION_FINISHED );

        msg.set_archivation_id( archivation_id );
        send_status_message_to_archivation_nodes( msg );

        // Switch to next phase.
        phase = PHASE_FINISHING;

        // Log.
            {
            std::ostringstream ss;
            ss  << "Local archivation finished and archivation started status message "
                << "received from every archivation node. "
                << "Archive manager is switching to finishing phase.";

            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_ARCHIVATION,
                    Status::PRIORITY_NORMAL
                    ),
                ss.str()
                );
            }
            
#ifdef OPTION_DEBUG
        // Verify that there are no objects with pending archive_on_write set on.
            {
            GarbageCollector & gc = Global::garbage_collector();
            VerifyObjectEnumerator enumerator;
            enumerator.archive = archive;
            
            gc.enumerate_all_objects( enumerator );               
            }
#endif // OPTION_DEBUG
        }
    else
        {
        // Not all of scheduled object are archived or not all of
        // archivation nodes have send archivation start status
        // message to us.
        }
    }


void ArchiveManager::phase_finishing()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::phase_finishing()" );
    if( finished_node_count == archivation_node_states.size() )
        {
        // All archivation nodes have finished their local archivations.
        // Global archivation is finished now.
        // We have nothing to do now.

        // Close archive file.
            {
            hope( total_object_to_archive_count == total_archived_object_count );

            // Validate archive file.
            archive->validate();

#ifdef OPTION_DEBUG
            // Save graphviz snapshot.
            if( false )
                {
                std::string   file_name = archivation_id.to_string() + ".dot";
                std::ofstream file( file_name.c_str() );
                archive->save_graphviz_snapshot( file );
                }
#endif // OPTION_DEBUG

            // Commit the archive.
            ArchiveDatabase & archive_database = Global::archive_database();
            archive_database.commit_regular_archive( archivation_id, archive );
            archive = NULL; // Archive database takes ovnership of the archive object.
            }

        // Unpause all system message queues which were paused during archivation finishing.
            {
            SystemEventQueue & system_event_queue = Global::system_event_queue();
            NodeStateMap::const_iterator it;
            for( it = archivation_node_states.begin(); it != archivation_node_states.end(); ++it )
                {
                const NodeId & node_id = it->first;
                unsigned int pause_mask = system_event_queue.get_server_pause_mask( node_id );
                pause_mask &= ~( SystemMessage::PAUSE_ARCHIVATION_FINISHING_START );
                system_event_queue.set_server_pause_mask( node_id, pause_mask );
                }
            }

        // And cleanup archivation stuff.
        cleanup_archivation();

        phase = PHASE_IDLE;
        idle_phase_start_time = Global::system_time();

        // Log.
            {
            std::ostringstream ss;
            ss  << "Archivation finished. Archivation finished status message "
                << "received from every archivation node. "
                << "Archive manager is switching to idle phase.";

            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_ARCHIVATION,
                    Status::PRIORITY_HIGH
                    ),
                ss.str()
                );
            }
        }
    else
        {
        // Not all of archivation nodes are finished,
        // we must wait...
        }
    }


bool ArchiveManager::local_archivation_next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::local_archivation_next_tick()" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( Global::system_time() >= archivation_start_time );
    hope( archivation_finish_time >= archivation_start_time );
    hope( total_archived_object_count <= total_object_to_archive_count );

    SystemTime   curr_time = Global::system_time();
    SystemTime   t_delta = curr_time - archivation_start_time;
    SystemTime   a_delta = archivation_finish_time - archivation_start_time;
    unsigned int count;
    unsigned int e_count; // Number of objects that must be archived till now.

    if( curr_time >= archivation_finish_time )
        {
        e_count = total_object_to_archive_count;
        }
    else
        {
        e_count = (unsigned int)
            (
            ( total_object_to_archive_count * t_delta.get_value() ) / a_delta.get_value()
            );
        }

    if( e_count > total_object_to_archive_count )
        e_count = total_object_to_archive_count;

    if( total_archived_object_count >= e_count )
        {
        // All objects that have to be archived till now was already archived.
        e_count = total_archived_object_count;
        }

    count = e_count - total_archived_object_count;

    // Archive count objects now.
    if( count > 0 )
        {
        archive_objects( count );
        }

    if( total_archived_object_count == total_object_to_archive_count )
        {
        return true;
        }
    else
        {
        return false;
        }
    }


void ArchiveManager::archive_objects
    (
    unsigned int count
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::archive_objects( unsigned int )" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( total_archived_object_count + count <= total_object_to_archive_count );

    // Log.
        {
        std::ostringstream ss;
        ss  << "Archivating " << count << " object(s) now...";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOWEST
                ),
            ss.str()
            );
        }

#ifdef OPTION_DEBUG
        // Save graphviz snapshot.
        if( false )
            {
            static int sub = 0;
            std::ostringstream ss;
            ss  << archivation_id.to_string() << "_"
                << std::setfill( '0' )
                << std::setw( 3 )
                << sub
                << std::setfill( ' ' )
                << ".dot";
            std::string   file_name = ss.str();
            std::ofstream file( file_name.c_str() );
            archive->save_graphviz_snapshot( file );
            sub++;
            }
#endif // OPTION_DEBUG

    // Because objects_to_archive and objects_to_copy can contain
    // dead objects or objects already archived, we use dest_count
    // to check whether count objects was really archived.

    ObjectId     object_id;
    unsigned int dest_count = total_archived_object_count + count;

    while( total_archived_object_count < dest_count )
        {
        hope( !objects_to_copy.empty() || !objects_to_archive.empty() );

        // First try copy objects from swap file to archive.
        if( !objects_to_copy.empty() )
            {
            object_id = objects_to_copy.front();
            objects_to_copy.pop();

            // Copy the object from swap to archive file.
            archive_inactive_object( object_id );
            }
        else
            {
            object_id = objects_to_archive.front();
            objects_to_archive.pop();

            // Store the object to archive file.
            archive_active_object( object_id );
            }
        hope( archive->is_object_valid( object_id ) );
        }
    }


void ArchiveManager::archive_active_object
    (
    const ObjectId & object_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::archive_active_object()" );
    ObjectManager &         object_manager = Global::object_manager();
    ObjectProperty *        object;

    object = object_manager.gc.find_local_object( object_id );

    // We _must_ test if the object is replica.
    if( object && !object->is_replica() )
        {
        // Note: The object may be already stored in the archive.
        hope( object->is_archivable() );
        object->archive_now();
        }
    else
        {
        // In case that the object is replica, it means
        // that the object was migrated to another node
        // (and archivated) and then replica was created
        // on local node. Strange, but may happen.
        }
    }


void ArchiveManager::archive_inactive_object
    (
    const ObjectId & object_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::archive_inactive_object()" );
    ObjectManager &        object_manager = Global::object_manager();
    Archive *              swap = object_manager.get_swap();

    hope( swap != NULL );

    // Check if the object is already archived.
    if( !archive->is_object_valid( object_id ) )
        {
        // Note: We do not know here if the object was registered in
        // the archive or not. But copy_object_to() will take care of it.
        swap->copy_object_to( *archive, object_id );
        total_archived_object_count++;
        }
    }


void ArchiveManager::finish_archivation_of_local_objects()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::finish_archivation_of_local_objects()" );
    hope( phase == PHASE_LOCAL_ARCHIVATION );
    hope( total_object_to_archive_count >= total_archived_object_count );

    unsigned int left_count = total_object_to_archive_count - total_archived_object_count;

    if( left_count > 0 )
        {
        archive_objects( left_count );
        }

    hope( total_object_to_archive_count == total_archived_object_count );
    /* Note: Archivation queues may be non-empty even if all scheduled
     * objects was archived! */

    // Log.
        {
        std::ostringstream ss;
        ss  << "Forced archivation of all local objects finished. "
            << "All " << total_archived_object_count
            << " owned objects was archived.";
        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_ARCHIVATION,
                Status::PRIORITY_LOW
                ),
            ss.str()
            );
        }
    }


void ArchiveManager::cleanup_archivation()
    {
    // Close archive file.
    if( archive )
        {
        delete archive;
        archive = NULL;
        }

    // Clear archivation queues.
    objects_to_archive = GarbageCollector::ObjectIdQueue();
    objects_to_copy    = ObjectShared::ObjectIdQueue();

    total_object_to_archive_count = 0;
    total_archived_object_count   = 0;

    archivation_node_states.clear();
    started_node_count  = 0;
    finished_node_count = 0;
    }


bool ArchiveManager::local_archivation_in_progress() const
    {
    return phase == PHASE_LOCAL_ARCHIVATION;
    }


bool ArchiveManager::archivation_in_progress() const
    {
    return phase != PHASE_IDLE;
    }


bool ArchiveManager::archivation_in_progress_on_node
    (
    NodeId node
    ) const
    {
    /* To make the archivation work properly, we may
     * not return archivation_in_progress() if node
     * is local node id.
     */
     
    hope( node.is_type( NodeId::TYPE_SERVER ) );

    NodeStateMap::const_iterator it = archivation_node_states.find( node );

    if( it == archivation_node_states.end() )
        {
        return false;
        }

    const NodeState & node_state = it->second;
    return node_state.phase != PHASE_IDLE;
    }

    
SystemMessage::ExecuteResult ArchiveManager::handle_archivation_status_message
    (
    const ArchivationStatusMessage & message
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ArchiveManager::handle_archivation_status_message()" );
    unsigned int            verify_result;

    // Already verified when dispatched.    
    hope( message.get_sender() == message.get_initiator() );
    hope( message.get_sender().get_type() == NodeId::TYPE_SERVER );

    // If we are waiting for ARCHIVATION_FINISHED messages from all archivation nodes,
    // and we have received ARCHIVATION_STARTED message from initiator,
    // and initiator already send ARCHIVATION_FINISHED to us,
    // then pause message from the initiator until all ARCHIVATION_FINISHED
    // are received.
    // Note: We must do this test before message verification.
    if( phase == PHASE_FINISHING && message.get_status() == ArchivationStatusMessage::ARCHIVATION_STARTED )
        {
        NodeStateMap::const_iterator it = archivation_node_states.find( message.get_initiator() );
        if( it != archivation_node_states.end() )
            {
            const NodeState & node_state = it->second;
            if( node_state.phase == PHASE_FINISHING )
                {
                SystemEventQueue & system_event_queue = Global::system_event_queue();
                unsigned int pause_mask = system_event_queue.get_server_pause_mask( message.get_initiator() );
                hope( !flags_on( pause_mask, SystemMessage::PAUSE_ARCHIVATION_FINISHING_START ) );
                pause_mask |= SystemMessage::PAUSE_ARCHIVATION_FINISHING_START;
                system_event_queue.set_server_pause_mask( message.get_initiator(), pause_mask );
                return SystemMessage::EXECUTE_KEEP_MESSAGE_IN_QUEUE;
                }
            }
        }

    verify_result = verify_archivation_status_message( message );
    if( verify_result != VERIFY_OK )
        {
        panic( "ArchivationStatusMessage verification failed!" );
        /** @todo FIXME */
        }

#ifdef OPTION_DEBUG
    if( archivation_in_progress() )
        {
        NodeStateMap::iterator  initiator_state_it;
        initiator_state_it = archivation_node_states.find( message.get_initiator() );
        hope( initiator_state_it != archivation_node_states.end() );
        }
#endif // OPTION_DEBUG

    switch( message.get_status() )
        {
        case ArchivationStatusMessage::ARCHIVATION_STARTED:
            {
            if( !archivation_in_progress() )
                {
                // We must delay the message if the system is processing
                // some sync rpc call right now. Archivation can be started
                // only in base recursion level of main system loop.
                if( System::is_waiting_for_rpc_reply() )
                    {
                    
                    {//sob
                    std::ostringstream ss;
                    ss  << "Enabling PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED system message pause flag"
                        << " for server node " << message.get_initiator() << ".";
                    Global::log_debug
                        (
                        Status::FACILITY_ARCHIVATION,
                        Status::PRIORITY_NORMAL,
                        ss.str()
                        );
                    }//eob
                    
                    SystemEventQueue & system_event_queue = Global::system_event_queue();
                    unsigned int pause_mask = system_event_queue.get_server_pause_mask( message.get_initiator() );
                    hope( !flags_on( pause_mask, SystemMessage::PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED ) );
                    pause_mask |= SystemMessage::PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED;
                    system_event_queue.set_server_pause_mask( message.get_initiator(), pause_mask );
                    
                    is_srpc_delayed_start_archivation_message = true;
                    
                    return SystemMessage::EXECUTE_KEEP_MESSAGE_IN_QUEUE;
                    }
                
                start_archivation();
                }

            NodeStateMap::iterator  initiator_state_it;

            initiator_state_it = archivation_node_states.find( message.get_initiator() );
            hope( initiator_state_it != archivation_node_states.end() );

            NodeState & initiator_state = initiator_state_it->second;
            hope( initiator_state.phase == PHASE_IDLE );

            initiator_state.phase = PHASE_LOCAL_ARCHIVATION;
            started_node_count++;

            break;
            }
        case ArchivationStatusMessage::ARCHIVATION_FINISHED:
            {
            hope( archivation_in_progress() );

            NodeStateMap::iterator  initiator_state_it;

            initiator_state_it = archivation_node_states.find( message.get_initiator() );
            hope( initiator_state_it != archivation_node_states.end() );

            NodeState & initiator_state = initiator_state_it->second;
            hope( initiator_state.phase == PHASE_LOCAL_ARCHIVATION );

            initiator_state.phase = PHASE_FINISHING;
            finished_node_count++;

            break;
            }
         default:
            panic( "BUG: Internal error in ArchivationStatusMessage verification!" );
            break;
        } // switch( message.get_status() )

    return SystemMessage::EXECUTE_OK;
    }


unsigned int ArchiveManager::verify_archivation_status_message
    (
    const ArchivationStatusMessage & message
    )
    {
    unsigned int                     flags = 0;
    NodeStateMap::const_iterator     msg_initiator_state_it;
    ArchivationStatusMessage::Status arch_status;

    // Already verified in SystemMessage::dispatch_system_message_from_packet().
    hope( !( message.get_sender().is_null() ) );
    hope( message.get_sender() == message.get_initiator() );
    hope( message.get_sender().is_type( NodeId::TYPE_SERVER ) );
    
    // Verify archivation id.
    if( message.get_archivation_id().is_invalid() )
        {
        flags |= VERIFY_INVALID_ARCHIVATION_ID;
        }
        
    if( archivation_in_progress() )
        {
        if( message.get_archivation_id() != archivation_id )
            {
            flags |= VERIFY_INVALID_ARCHIVATION_ID;
            }
        }
    else
        {
        if( message.get_archivation_id() <= archivation_id )
            {
            flags |= VERIFY_INVALID_ARCHIVATION_ID;
            }
        }
        
    // Try find message initiator in archivation_node_states.
    if( archivation_in_progress() )
        {
        msg_initiator_state_it = archivation_node_states.find( message.get_initiator() );
        }
    else
        {
        msg_initiator_state_it = archivation_node_states.end();
        }
        
    // Verify that the message initiator is participating in the archivation.
    if( archivation_in_progress() )
        {
        if( msg_initiator_state_it == archivation_node_states.end() )
            {
            flags |= VERIFY_UNKNOWN_ARCHIVATION_NODE;
            }
        }
    else
        {
        if( !message.get_initiator().is_type( NodeId::TYPE_SERVER ) )
            {
            flags |= VERIFY_UNKNOWN_ARCHIVATION_NODE;
            }
        }
        
    // Verify archivation status.
    arch_status = message.get_status();
    
    switch( arch_status )
        {
        case ArchivationStatusMessage::ARCHIVATION_STARTED:
            if( archivation_in_progress() )
                {
                if( msg_initiator_state_it != archivation_node_states.end() )
                    {
                    const NodeState & initiator_state = msg_initiator_state_it->second;
                    if( initiator_state.phase != PHASE_IDLE )
                        {
                        flags |= VERIFY_UNEXPECTED_ARCHIVATION_STATUS;
                        }
                    }
                }
            else
                {
                // Everything ok.
                // We know that the initiator is archivation node.
                }
            break;
        case ArchivationStatusMessage::ARCHIVATION_FINISHED:
            if( archivation_in_progress() )
                {
                if( msg_initiator_state_it != archivation_node_states.end() )
                    {
                    const NodeState & initiator_state = msg_initiator_state_it->second;
                    if( initiator_state.phase != PHASE_LOCAL_ARCHIVATION )
                        {
                        flags |= VERIFY_UNEXPECTED_ARCHIVATION_STATUS;
                        }
                    }
                }
            else
                {
                flags |= VERIFY_UNEXPECTED_ARCHIVATION_STATUS;
                }
            break;
        default:
            flags |= VERIFY_INVALID_ARCHIVATION_STATUS;
            break;
        };
        
    return flags;
    }


void ArchiveManager::send_status_message_to_archivation_nodes
    (
    const ArchivationStatusMessage &    message
    )
    {    
    NodeManager &               node_manager = Global::node_manager();
    NodeManager::NodeIdList     archivation_nodes;
 
    // Get archivation node list.
    // Note: We do not use NodeManager::enumerate_archivation_nodes()
    // because some archivation nodes could get connected during
    // archivation and these nodes is not waiting archivation status
    // message from us.
        {
        NodeStateMap::const_iterator it;

        archivation_nodes.reserve( archivation_node_states.size() );
        for( it = archivation_node_states.begin(); it != archivation_node_states.end(); ++it )
            {
            const NodeId & node_id = it->first;
            archivation_nodes.push_back( node_id );
            }
        }

    // Multicast the archivation status message to all archivation nodes.
    node_manager.multicast_message( message, archivation_nodes, NetworkCommon::STREAM_RELIABLE );
    }


void ArchiveManager::enumerate_archivation_nodes()
    {
    hope( phase == PHASE_IDLE );
    hope( archivation_node_states.empty() );

    NodeManager &                     node_manager = Global::node_manager();
    NodeManager::NodeIdList           archivation_nodes;
    NodeManager::NodeIdList::iterator it;

    archivation_node_states.clear();

    // Enumerate archivation nodes participating in the archivation.
    node_manager.enumerate_archivation_nodes( archivation_nodes );

    try
        {
        for( it = archivation_nodes.begin(); it != archivation_nodes.end(); ++it )
            {
            NodeId & node = ( *it );
            archivation_node_states.insert
                (
                NodeStateMap::value_type
                    (
                    node,
                    NodeState()
                    )
                );
            }
        }
    catch( ... )
        {
        archivation_node_states.clear();
        throw;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
