/** @file archive_manager.h
 * Archive manager.
 * Archive manager manages archivation
 * of objects on local node and synchronization with
 * other nodes.
 */

/*
 * begin        : Thu Oct 17 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ARCHIVE_MANAGER_H
#define MASSIV_ARCHIVE_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_ARCHIVATION_ID_H
#include "object/archivation_id.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h" // ArchivationStatusMessage
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Class which encapsulates common object archivation methods used by
 * ObjectManager and ArchiveManager.
 *
 * @ingroup archivation
 */
class CommonArchiveOperations
    {
    friend class ObjectManager;
    friend class ArchiveManager;

private:

    /**
     * Register objects from @a object_list to @a archive.
     * The objects are registered in new archive group.
     * @pre The objects are not yet registered in the archive.
     * @pre All of the objects are archiable.
     */
    static void register_objects_to_archive
        (
        Archive *                        archive,
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Serialize and validate one object to @a archive.
     *
     * Do not forget to initialize #desc.object_flags, it should be either
     * @c Serializer::Description::OBJECT_ARCHIVE or
     * @c Serializer::Description::OBJECT_SWAP.
     *
     * @pre The object is already registered in the @a archive.
     * @pre The object is not replica.
     *
     * @todo Cleanup stream error detection and exception throwing.
     */
    static void text_write_object_to_archive
        (
        Archive *                       archive,
        ObjectProperty *                object,
        const Serializer::Description & desc
        );

    }; // CommonArchiveOperation


/**
 * Archive manager.
 * Archive manager manages archivation
 * of objects on local node and synchronization with
 * other nodes.
 *
 * Archive manager is used only to store objects in some
 * archive file. To read objects from an archive, use
 * object manager.
 *
 * Only @e owned objects are archived.
 * @see ObjectManager
 *
 * Refering to local archivation means archivation of
 * local objects on some node. Local archivations are managed
 * by archive managers located on archivation nodes.
 *
 * Refering to global archivation means distributed
 * archivation of entire simulation. Global archivation
 * is managed by global archivation protocol. All archive
 * managers from archivation nodes participate together
 * to satisfy the global archivation protocol.
 *
 * @note During archivation swap file may not be closed.
 *
 * @ingroup archivation
 */
class MASSIV_EXPORTED ArchiveManager
    {
public:

    /**
     * Constructor.
     */
    ArchiveManager();

    /**
     * Destructor.
     */
    ~ArchiveManager();

    /**
     * Initialize the archive manager.
     * @param latest_archivation_id Latest valid archivation id
     *        on all server nodes. New archivation will gain this id
     *        incremented by one.
     */
    void initialize
        (
        const ArchivationId & latest_archivation_id
        );

public:

    /**
     * Enable auto archivation.
     *
     * @note This function will only enable tests during idle phase
     * of the manager whether new archivation should be automatically
     * initiated. The auto archivation must be also permitted via
     * settings to make the manager capable of automatically initiating
     * any archivation.
     * Even if auto archivation is enabled via settings, the manager will
     * not initiate new archivation until enable_auto_archivation() is called.
     * This ensures that new archivation is not initiated during system
     * startup or system shutdown.
     */
    void enable_auto_archivation();

    /**
     * Disable auto archivation.
     */
    void disable_auto_archivation();

    /**
     * Check whether auto archivation is enabled.
     */
    bool is_auto_archivation_enabled();

public:

    /**
     * Result of start archivation request.
     */
    enum RequestStartArchivationResult
        {
        RESULT_NODE_IS_NOT_SERVER,
            /**< Archivation can't be started because local node is not
             * server node.
             */
        RESULT_ARCHIVATION_IN_PROGRESS,
            /**< Archivation is running. Note there is no way how to detect
             * from the result whether the archivation was running before
             * the request. Use @c archivation_in_progress() before
             * calling the request to check whether new archivation
             * was started by the request.
             */
        RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC,
            /**< Start archivation request was delayed because currently
             * there are some pending sync rpc calls. You don't need
             * to ask to start archivation again, the archivation will
             * start as soon as possible, because archive manager has stored
             * flag inside and the archivation will be started as there
             * will be no pending sync rpc calls.
             */
        }; // RequestStartArchivationResult

    /**
     * Initiate new distributed global archivation of the simulation.
     * If archivation can't be initiated immediately, the archive manager
     * will store the request and will start the archivation as soon as it
     * will be possible.
     *
     * @note Global archivation can be started on two or more nodes
     *     indenpendently in one tick and everything will work fine
     *     if the nodes assign same @c ArchivationId to the archivation.
     */
    RequestStartArchivationResult request_start_archivation();

    /**
     * Start initial archivation.
     * This function should be used only when creating initial archive
     * of a server node.
     * @pre Archivation is not in progress.
     * @pre There are no pending sync rpc calls.
     */
    void request_start_initial_archivation();

    /**
     * Initiate new distributed global archivation of the simulation
     * with specified ArchivationId.
     * @return True if new archivation was started. False if new
     * archivation can't be initiated at the time (for example because
     * previous archivation is not finished yet).
     * @note If the system is processing sync rpc calls, the function
     * returns false.
     *
     * @warning This functions should be used only for
     *     archive packing.
     *
     * @todo FIXME: See notes inside. Rename this function and add
     *     many warnings.
     */
    bool request_start_archivation
        (
        const ArchivationId & archivation_id
        );

    /**
     * Get current archivation id used by the archive manager.
     */
    ArchivationId get_current_archivation_id() const;

private:

    /**
     * Initiate new distributed global archivation of the simulation.
     * @param initial_archivation Specifies whether inital archive
     * should be created. Initial archive is archive with lowest
     * valid ArchivationId.
     * @pre The manager is in idle phase - ie. no (local) archivation
     * is in progress.
     * @pre The system is not processing sync rpc calls.
     * @note Global archivation can be started on two or more nodes
     * indenpendently in one tick and everything will work fine
     * if the nodes assign same @c ArchivationId to the archivation.
     */
    void start_archivation
        (
        bool initial_archivation = false
        );

    /**
     * Start an archivation with specified @c ArchivationId.
     * @pre The manager is in idle phase - ie. no (local) archivation
     * is in progress.
     * @pre The system is not processing sync rpc calls.
     */
    void start_archivation
        (
        const ArchivationId & archivation_id
        );

public:

    //
    // For Object and ObjectManager purposes only!
    //

    /**
     * Store object state to archive file.
     * Object will be serialized using
     * @c Serializer::DescriptioN::OBJECT_ARCHIVE flags.
     *
     * @pre The manager is in local archivation phase.
     * @pre The function is called from @c Object::archive_now().
     * @pre The function is not called twice with same @a object
     * during one archivation.
     * @pre @a object is not @c NULL.
     * @pre @a object is not replica.
     * @pre @a object is not member object.
     */
    void archive_object_state
        (
        ObjectProperty * object
        );

    /**
     * Archive object which is about to be swapped-in.
     * This function is used to notify the archive manager
     * that the object will be swapped-in and so it must
     * be archived if it was not archived yet.
     *
     * Nothing is done if the object was already archived.
     * @pre Local archivation is in progress.
     * @pre Swap file must be open.
     * @pre The object is stored in the swap file.
     * @pre The object is archivable.
     */
    void archive_swap_in_object
        (
        const ObjectId & object_id
        );

    /**
     * Schedule object for archivation.
     * The object will be archived sometime in the future.
     *
     * Object manager uses this function to schedule objects
     * received from network for archivation while local
     * archivation is running.
     *
     * The archive manager will only
     * store the @a object_id and won't access the object itself
     * during this function.
     *
     * @pre Local archivation is in progress.
     * @pre The object was not scheduled for archivation yet.
     * @pre The object is owned (see @c ObjectManager).
     */
    void schedule_object_for_archivation
        (
        const ObjectId & object_id
        );

    /**
     * Register @a object and its migration group to the archive
     * file if it is not registered yet.
     * @pre The manager is in local archivation phase.
     * @pre @a object is not @c NULL.
     * @pre The @a object is archivable.
     */
    void register_object_to_archive
        (
        ObjectProperty * object
        );

public:

    /**
     * Run next archive manager tick.
     */
    void next_tick();

private:

    /**
     * Advance @c PHASE_IDLE phase.
     */
    void phase_idle();

    /**
     * Advance @c PHASE_LOCAL_ARCHIVATION phase.
     */
    void phase_local_archivation();

    /**
     * Advance @c PHASE_FINISHING phase.
     */
    void phase_finishing();

private:

    /**
     * Advance archivation of @e local objects.
     * @pre Phase of the archive manager is @c PHASE_LOCAL_ARCHIVATION.
     * @return True if all @e owned objects that was scheduled
     * for archivation was archived.
     *
     * @note Archivation queues may be non-empty even if all
     * objects scheduled for archivation was already archived.
     * @note Do not forget that receiving some message from another
     * node can schedule new objects to be archived.
     */
    bool local_archivation_next_tick();

    /**
     * Archive @a count objects.
     * @pre Phase of the archive manager is @c PHASE_LOCAL_ARCHIVATION.
     * @pre There are at least @a count objects scheduled for the
     * archivation which was not yet archived.
     */
    void archive_objects
        (
        unsigned int count
        );

    /**
     * Archive @e active object @a object_id.
     * Nothing is done if the object was already archived
     * or it is not active object any more.
     */
    void archive_active_object
        (
        const ObjectId & object_id
        );

    /**
     * Archive @e inactive object @a object_id.
     * Nothing is done if the object was already archived.
     */
    void archive_inactive_object
        (
        const ObjectId & object_id
        );

    /**
     * Finish archivation of all @e owned objects
     * that were scheduled for archivation.
     * @pre Phase of the archive manager is @c PHASE_LOCAL_ARCHIVATION.
     * @post @c total_object_to_archive_count is equal to @c total_archived_object_count.
     * @note Archivation queues may be non-empty even if all
     * objects scheduled for archivation was already archived.
     *
     * @note The archivation itself is not finished.
     * @note Do not forget that receiving some message from another
     * node can schedule new objects to be archived.
     */
    void finish_archivation_of_local_objects();

    /**
     * Cleanup archivation stuff.
     * Archive file is closed (but not validated).
     * Local archivation queues are made empty.
     * Archive object counts are cleared.
     * Distributed global archivation information is cleared.
     * The archive manager phase is left unchanged!
     * @warning Think twice before you call this function.
     */
    void cleanup_archivation();

public:

    /**
     * Check whether the phase of the archive manager is
     * @c PHASE_LOCAL_ARCHIVATION.
     */
    bool local_archivation_in_progress() const;

    /**
     * Check whether global archivation is running (ie.
     * whether tha phase of the archive manager is
     * @c PHASE_LOCAL_ARCHIVATION or @c PHASE_FINISHING).
     *
     * @note Returns true even if local archivation is finished,
     * so do not use this function to test if some objects
     * must be archived or not! Use @c archivation_in_progress_on_node()
     * or @c local_archivation_in_progress() instead.
     *
     * @note Do not use this function for testing
     * if objects from a migration message
     * received from local node must be archived!
     * Use @c archivation_in_progress_on_node
     * instead and pass local node as its argument!
     * This is because all incoming system messages are buffered
     * and thus the messages could be send after
     * start of current archivation.
     * @see archivation_in_progress_on_node()
     */
    bool archivation_in_progress() const;

    /**
     * Check whether archivation is running on @a node.
     *
     * @note Passing local node id as @a node is <b>not</b>
     * same as calling @c archivation_in_progress()!
     * Because incoming system messages are buffered
     * the archive manager must send "archivation start"
     * message to itself so the object managet can
     * properly handle incoming system messages.
     * @see archivation_in_progress
     *
     * @return True if the @a node is participating
     * in archivation and the manager has received message
     * from the @a node that it has started the archivation.
     * False otherwise.
     */
    bool archivation_in_progress_on_node
        (
        NodeId    node
        ) const;

public:

    /**
     * Handle ArchivationStatusMessage message.
     */
    SystemMessage::ExecuteResult handle_archivation_status_message
        (
        const ArchivationStatusMessage & message
        );

private:

    /**
     * ArchivationStatusMessage verification flags.
     */
    enum VerificationFlags
        {
        VERIFY_OK                           = 0x00000,
            /**< Successful ArchivationStatusMessage verification.
             * Defined only for convenience. */
        VERIFY_INVALID_ARCHIVATION_ID       = 0x00001,
            /**< ArchivationStatusMessage contains invalid ArchivationId.
             * If archivation is in progress, the ArchivationId must be same
             * as ArchivationId stored in the ArchiveManager.
             * If archivation is not in progress, the ArchivationId must be
             * bigger than ArchivationId stored in the ArchiveManager.
             */
        VERIFY_UNKNOWN_ARCHIVATION_NODE     = 0x00002,
            /**< ArchivationStatusMessage was received from a node, which is not
             * known as an archivation node. ArchiveManager "knows" only
             * the archivation nodes, which are enumerated during
             * @c start_archivation().
             * This flag is not set if archivation is not in progress.
             */
        VERIFY_INVALID_ARCHIVATION_STATUS   = 0x00004,
            /**< Invalid value of ArchivationStatusMessage::Status. */
        VERIFY_UNEXPECTED_ARCHIVATION_STATUS = 0x0008
            /**< Unexpected value of ArchivationStatusMessage::Status. */
        }; // VerificationFlags

    /**
     * Verify incoming ArchivationStatusMessage.
     * @return Combination of @c VerificationFlags. Use
     * comparison with @c VERIFY_OK to check if the verification was successful.
     */
    unsigned int verify_archivation_status_message
        (
        const ArchivationStatusMessage & message
        );

private:

    /**
     * Send archivation status message to all
     * archivation nodes which participates in current global
     * archivation.
     */
    void send_status_message_to_archivation_nodes
        (
        const ArchivationStatusMessage &    message
        );

    /**
     * Enumerate archivation nodes and push them
     * to @c archivation_node_states.
     * @pre Phase of the archive manager is @c PHASE_IDLE.
     * @pre @c archivation_node_states map is empty.
     */
    void enumerate_archivation_nodes();

public: // Make the phase public so doxygen will add it to the documentation.

    /**
     * Archive manager internal phase.
     */
    enum Phase
        {
        PHASE_IDLE              = 0,
            /**< Local archivation is not in progress.
             * The archive manager is waiting for
             * a reason to start an archivation.
             */
        PHASE_LOCAL_ARCHIVATION = 1,
            /**< Archivation of local objects is in progress.
             * Archive manager is storing objects scheduled for
             * the archivation to the archive file
             * and waiting until every local archivation on
             * another nodes is started (not finished!)
             * - ie. until the archive manager receives archivation
             * started status message from every archivation node.
             *
             * After local archivation is finished, the archive manager
             * sends archivation finished status message
             * to all archivation nodes and switch to
             * @c PHASE_FINISHING.
             */
        PHASE_FINISHING         = 2,
            /**< The archive manager is trying to finish the archivation.
             * It waits until it receives archivation finished status
             * message from all archivations nodes.
             */

        }; // Phase

private:

    /**
     * State of archivation on a node.
     */
    class NodeState
        {
    public:

        /** Constructor. */
        NodeState();

        /** Destructor. */
        ~NodeState();

        /** Copy constructor. */
        NodeState
            (
            const NodeState & node_state
            );

        /** Operator=. */
        NodeState & operator=
            (
            const NodeState & node_state
            );

    public:

        Phase       phase;
            /**< Phase of archive manager located on the node. */

        }; // NodeState

    /**
     * Map which associates @c NodeId with @c NodeState.
     */
    typedef ext::hash_map
                <
                NodeId,
                NodeState,
                NodeId::Hash,
                NodeId::EqualKey
                > NodeStateMap;

public:

    /**
     * ArchiveManager settings.
     */
    class ArchiveManagerSettings : public Settings
        {
    public:

        /** Constructor. */
        ArchiveManagerSettings();

    public:

        Float & local_archivation_duration;
            /**< How much time (in seconds) does the archive manager have
             * to archive all local objects.
             * @note The archive manager doesn't guarantee that all local
             * objects will be archived in the time (especially if the time
             * is small) because new local objects can be created
             * as a result of incoming migration message from another server.
             * @note This time affects only local archivation. Duration of global
             * distributed archivation depends on all servers perticipating
             * in the archivation.
             * @note Duration of initial archive creation is always fixed
             * (currently one second) and is not affected by this value.
             */

        Boolean & auto_archivation_enabled;
            /**< Whether automatic archivation is enabled.
             */

        Float & auto_archivation_delay;
            /**< How much time (in seconds) does the archive manager
             * waits in idle phase until it initiate new distributed archivation
             * if auto archivation is enabled.
             */

        }; // ArchiveManagerSettings.


private:

    Phase           phase;
        /**< Current archive manager phase. */

    bool            is_srpc_delayed_start_archivation_message;
        /**< True if there are some paused start archivation messages
         * which were paused because in their delivery time the system
         * was processing some sync rpc calls. During sync rpc call
         * archivation may not be started.
         */

    //
    // Local archivation stuff.
    //

    Archive *                           archive;
        /**< Archive file where all @e owned objects will be stored.
         * The pointer is valid only if @c phase is not @c PHASE_IDLE.
         */

    ObjectShared::ObjectIdQueue         objects_to_archive;
        /**< List of @e active objects to be archived.
         * New objects can be added to the queue during archivation. */

    ObjectShared::ObjectIdQueue         objects_to_copy;
        /**< List of @e inactive objects to be copied from swap to archive. */

    unsigned int        total_object_to_archive_count;
        /**< Total number of @e owned objects that must be archived.
         * This number can increase during archivation as new objects are
         * scheduled to be archived.
         * @note This number is not equal to ( objects_to_archive.size()
         * + objects_to_copy.size() ).
         */

    unsigned int        total_archived_object_count;
        /**< Total number of @e owned objects that was stored in the archive till now. */

    SystemTime          archivation_start_time;
        /**< Time when the archivation was started. */

    SystemTime          archivation_finish_time;
        /**< Time when archivation of @e local objects will be finished.
         * May change during archivation. */

    //
    // Global archivation stuff.
    //

    bool            auto_archivation_enabled;
        /**< Whether auto archivation is enabled. If true the manager
         * will automatically initiate distributed archivation once a time
         * (specified via settings).
         */

    SystemTime      idle_phase_start_time;
        /**< Time when idle phase started. This time is used by auto
         * archivation to detect whether new archivation should be
         * initiated.
         */

    ArchivationId   archivation_id;
        /**< Current archivation id. */

    NodeStateMap    archivation_node_states;
        /**< Map which stores archivation states of all nodes
         * participating in the archivation of the simulation.
         * @note Do not add any node to this list during archivation.
         * @note Local node is also included in the map (if it
         * is an archivation node).
         */

    unsigned int    started_node_count;
        /**< Number of archivation nodes which have started
         * their local archivations.
         */

    unsigned int    finished_node_count;
        /**< Number of archivation nodes which have finished
         * their local archivations.
         */

    //
    // Statistics and settings.
    //

    ArchiveManagerSettings      settings;
        /**< Archive manager settings. */

private:

    //
    // Archivation request.
    //

    bool        pending_archivation_start;
        /**< True whether there is pending request for archivation start. */

    }; // ArchiveManager

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVE_MANAGER_H
