/** @file class_manager.cpp
 * Class manager (implementation).
 */

/*
 * begin        : Sun May 12 2002
 * copyright    : (C) 2002 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/class_manager.h"
#include "utility/wildcards.h"
#include "status/exception.h"
#include "object/object.h"
#include "object/metaobject.h"
#include "object/lib/lib_exceptions_helper.h"
#include "global/global.h"
#include <sstream>
#include <algorithm>


namespace Massiv {
namespace Core {


ClassManager::ClassManager()
      : classes_initialized( false )
    {
    }


void ClassManager::register_class
    (
    MetaObject & metaobject
    )
    {
    hope( !classes_initialized );

    // Check that class is not registered yet.

    hope( from_type_info.find( &metaobject.get_class_type_info() ) == from_type_info.end() );
    hope( from_class_name.find( metaobject.get_class_name() ) == from_class_name.end() );

    // Register the class.

    from_class_id.insert
        (
        ClassIdToMetaObject::value_type
            ( metaobject.get_class_type_id(), &metaobject )
        );
    from_type_info.insert
        (
        TypeInfoToMetaObject::value_type
            ( &metaobject.get_class_type_info(), &metaobject )
        );
    from_class_name.insert
        (
        ClassNameToMetaObject::value_type
            ( metaobject.get_class_name(), &metaobject )
        );
    }


void ClassManager::initialize_classes()
    {
    hope( !classes_initialized );

    // Allow get_* methods.

    classes_initialized = true;

    // Initialize classes.

    ClassIdToMetaObject::iterator it;
    for( it = from_class_id.begin(); it != from_class_id.end(); ++it )
        {
        MetaObject * metaobject = it->second;
        hope( metaobject );
        metaobject->initialize_introspection();
        }
    for( it = from_class_id.begin(); it != from_class_id.end(); ++it )
        {
        MetaObject * metaobject = it->second;
        metaobject->initialize_inherited_introspection();
        }

    // Log.
    
    #ifdef OPTION_DEBUG

        {
        ClassManager::ClassNameToMetaObject::const_iterator it = 
            from_class_name.begin();
        std::ostringstream ss;

        ss << "The following classes were registered:\n";
        
        while( it != from_class_name.end() )
            {
            const MetaObject *  metaobject = it->second;
            const ClassTypeId & type_id = metaobject->get_class_type_id();

            ss << it->first << ' ' << type_id.dump() << '\n';
                
            ++it;
            }

        Global::log_info
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_LOW,
            ss.str()
            );
        }

    #endif
    }


const MetaObject * ClassManager::get_metaobject
    (
    ClassTypeId class_type_id
    ) const
    {
    const MetaObject * metaobject = get_metaobject_no_throw( class_type_id );
    
    if( !metaobject )
        {
        std::ostringstream ss;
        
        ss  << "No metaobject for class type id: " << class_type_id << ".\n"
            << "Does this node support " << class_type_id.get_kind() << "?";
        
        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }

    return metaobject;
    }


const MetaObject * ClassManager::get_metaobject
    (
    const std::type_info & class_type_info
    ) const
    {
    const MetaObject * metaobject = get_metaobject_no_throw( class_type_info );
    
    if( !metaobject )
        {
        std::ostringstream ss;
        
        // If client has access to std::type_info of that class,
        // the class should be available.

        ss  << "No metaobject for type info: " << class_type_info.name() << ".\n"
            << "Is IDL for this type properly defined?";

        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }
    
    return metaobject;
    }


const MetaObject * ClassManager::get_metaobject
    (
    const std::string & class_name
    ) const
    {
    const MetaObject * metaobject = get_metaobject_no_throw( class_name );
    
    if( !metaobject )
        {
        std::ostringstream ss;

        ss << "No metaobject for class name: " << class_name << ".";

        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }

    return metaobject;
    }


const ObjectFactory * ClassManager::get_factory
    (
    ClassTypeId class_type_id
    ) const
    {
    const ObjectFactory * factory = get_factory_no_throw( class_type_id );

    if( !factory )
        {
        std::ostringstream ss;
        
        ss  << "No factory for class type id: " << class_type_id << ".\n"
            << "Does this node support " << class_type_id.get_kind() << "?";
        
        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }

    return factory;
    }


const ObjectFactory * ClassManager::get_factory
    (
    const std::type_info & class_type_info
    ) const
    {
    const ObjectFactory * factory = get_factory_no_throw( class_type_info );
    
    if( !factory )
        {
        std::ostringstream ss;
        
        // If client has access to std::type_info of that class,
        // the class should be available.

        ss  << "No factory for type info: " << class_type_info.name() << ".\n"
            << "Is IDL for this type properly defined?";

        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }
    
    return factory;
    }


const ObjectFactory * ClassManager::get_factory
    (
    const std::string & class_name
    ) const
    {
    const ObjectFactory * factory = get_factory_no_throw( class_name );
    
    if( !factory )
        {
        std::ostringstream ss;

        ss << "No factory for class name: " << class_name << ".";

        RAISE_CORE_LIB_EXCEPTION( UndefinedClassTypeException, ss.str() );
        }

    return factory;
    }


const MetaObject * ClassManager::get_metaobject_no_throw
    (
    ClassTypeId class_type_id
    ) const
    {
    hope( classes_initialized );
    
    ClassIdToMetaObject::const_iterator it( from_class_id.find( class_type_id ) );
    if( it == from_class_id.end() )
        {
        return NULL;
        }
    return it->second;
    }


const MetaObject * ClassManager::get_metaobject_no_throw
    (
    const std::type_info & class_type_info
    ) const
    {
    hope( classes_initialized );

    TypeInfoToMetaObject::const_iterator it( from_type_info.find( &class_type_info ) );
    if( it == from_type_info.end() )
        {
        return NULL;
        }
    return it->second;
    }


const MetaObject * ClassManager::get_metaobject_no_throw
    (
    const std::string & class_name
    ) const
    {
    hope( classes_initialized );

    ClassNameToMetaObject::const_iterator it( from_class_name.find( class_name ) );
    if( it == from_class_name.end() )
        {
        return NULL;
        }
    return it->second;
    }


const ObjectFactory * ClassManager::get_factory_no_throw
    (
    ClassTypeId class_type_id
    ) const
    {
    const MetaObject * const metaobject = get_metaobject( class_type_id );
    if( metaobject )
        {
        return metaobject->get_object_factory();
        }
    return NULL;
    }


const ObjectFactory * ClassManager::get_factory_no_throw
    (
    const std::type_info & class_type_info
    ) const
    {
    const MetaObject * const metaobject = get_metaobject( class_type_info );
    if( metaobject )
        {
        return metaobject->get_object_factory();
        }
    return NULL;
    }


const ObjectFactory * ClassManager::get_factory_no_throw
    (
    const std::string & class_name
    ) const
    {
    const MetaObject * const metaobject = get_metaobject( class_name );
    if( metaobject )
        {
        return metaobject->get_object_factory();
        }
    return NULL;
    }


void ClassManager::find_classes
    (
    std::vector< std::string > &    result,
    const std::string &             pattern,
    bool                            include_subclasses
    ) const
    {
    ClassNameToMetaObject::const_iterator   it;

    if( include_subclasses )
        {
        std::vector<const MetaObject *> metaobjects;

        // Enumerate metaobjects of classes that match the selected pattern.

        for( it = from_class_name.begin(); it != from_class_name.end(); ++it )
            {
            const std::string & class_name = it->first;
            const MetaObject *  metaobject = it->second;
            
            if( Wildcards::match( pattern, class_name ) )
                {
                metaobjects.push_back( metaobject );
                }
            }
        
        std::sort( metaobjects.begin(), metaobjects.end() );

        // Report all subclasses of the enumerated metaobjects.

        for( it = from_class_name.begin(); it != from_class_name.end(); ++it )
            {
            const std::string & class_name = it->first;
            const MetaObject *  metaobject = it->second;

            // Test if @c metaobject describes a class that inherits from a
            // class in @c metaobjects.

            if( !std::binary_search( metaobjects.begin(), metaobjects.end(), 
                metaobject ) )
                {
                bool    parent_class_found = false;
                UInt32  i;

                for( i = 0; i < metaobjects.size(); i++ )
                    {
                    const std::type_info & class_type = 
                        metaobjects[ i ]->get_class_type_info();
                    
                    if( metaobject->inherits( class_type ) )
                        {
                        parent_class_found = true;
                        break;
                        }
                    }
                
                if( !parent_class_found )
                    {
                    continue;
                    }
                }

            // Report.
            
            result.push_back( class_name );
            }
        }
    else
        {
        // Report class names that match the selected pattern.

        for( it = from_class_name.begin(); it != from_class_name.end(); ++it )
            {
            const std::string & class_name = it->first;
            
            if( Wildcards::match( pattern, class_name ) )
                {
                result.push_back( class_name );
                }
            }
        }

    // Sort reported classes alphabetically.

    std::sort( result.begin(), result.end() );
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "class_manager_inline.h"
#endif
