/** @file class_manager.h
 * Class manager.
 */

/*
 * begin        : Sun May 12 2002
 * copyright    : (C) 2002 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CLASS_MANAGER_H
#define MASSIV_CLASS_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_CLASS_TYPE_ID_H
#include "object/class_type_id.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <typeinfo>
#include <string>
#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Structure that registers all known managed classes in the system and keeps
 * all metaobjects for the classes. Each class is assigned unique class type
 * id, which can be transmitted over network and is not platform specific
 * (unlike standard C++ type info).
 */
class MASSIV_EXPORTED ClassManager
    {
    NO_COPY_CONSTRUCTOR( ClassManager );
    NO_ASSIGNMENT_OPERATOR( ClassManager );

private:

    struct Hash
        {

        INLINE size_t operator()
            (
            const std::type_info * class_type_info
            ) const;

        }; // Hash

    struct EqualKey
        {

        INLINE bool operator()
            (
            const std::type_info * type_info1,
            const std::type_info * type_info2
            ) const;

        }; // EqualKey

    typedef ext::hash_map< ClassTypeId, MetaObject *, ClassTypeId::Hash, ClassTypeId::EqualKey > ClassIdToMetaObject;
    
    typedef ext::hash_map< const std::type_info *, MetaObject *, Hash, EqualKey > TypeInfoToMetaObject;

    typedef ext::hash_map< std::string, MetaObject * > ClassNameToMetaObject;

public:

    /**
     * Constructor.
     */
    ClassManager();

    /**
     * Register class described by given metaobject.
     *
     * Metaobject's information about class name and class type id should
     * be already initialized, introspection structures will be initialized
     * later.
     */
    void register_class
        (
        MetaObject & metaobject
        );

    /**
     * Initialize all registered metaobjects.
     *
     * No more classes can be registered once this method is called.
     */
    void initialize_classes();

public:

    // May-throw variants.

    /**
     * Get metaobject by class type id.
     * Throws UndefinedClassTypeException if @c class_type_id refers to an unknown
     * class.
     */
    const MetaObject * get_metaobject
        (
        ClassTypeId class_type_id
        ) const;

    /**
     * Get metaobject by class type info.
     * Throws UndefinedClassTypeException if metaobject can not be found.
     * This happens if there is no IDL description for @c class_type_info
     * or that description is invalid/unavailable (class kind).
     */
    const MetaObject * get_metaobject
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * Get metaobject by class name.
     * Throws UndefinedClassTypeException if @c class_name is uknown to the
     * current node.
     */
    const MetaObject * get_metaobject
        (
        const std::string & class_name
        ) const;

    /**
     * Get pointer to object factory.
     * May throw UndefinedClassTypeException.
     */
    const ObjectFactory * get_factory
        (
        ClassTypeId class_type_id
        ) const;

    /**
     * Get pointer to object factory.
     * May throw UndefinedClassTypeException.
     */
    const ObjectFactory * get_factory
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * Get pointer to object factory.
     * May throw UndefinedClassTypeException.
     */
    const ObjectFactory * get_factory
        (
        const std::string & class_name
        ) const;

    // No-throw variants.
    // Return NULL if metaobject/factory can not be returned.
    
    /**
     * No-throw variant of corresponding #get_metaobject().
     */
    const MetaObject * get_metaobject_no_throw
        (
        ClassTypeId class_type_id
        ) const;

    /**
     * No-throw variant of corresponding #get_metaobject().
     */
    const MetaObject * get_metaobject_no_throw
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * No-throw variant of corresponding #get_metaobject().
     */
    const MetaObject * get_metaobject_no_throw
        (
        const std::string & class_name
        ) const;

    /**
     * No-throw variant of corresponding #get_factory().
     */
    const ObjectFactory * get_factory_no_throw
        (
        ClassTypeId class_type_id
        ) const;

    /**
     * No-throw variant of corresponding #get_factory().
     */
    const ObjectFactory * get_factory_no_throw
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * No-throw variant of corresponding #get_factory().
     */
    const ObjectFactory * get_factory_no_throw
        (
        const std::string & class_name
        ) const;

public:

    /**
     * Find all classes matching given class name wildcard pattern.
     * If @c include_subclasses is true all subclasses of the enumerated
     * classes are added in the output. It is ensured that @c result will
     * be sorted alphabetically.
     */
    void find_classes
        (
        std::vector< std::string > &    result,
        const std::string &             pattern,
        bool                            include_subclasses = false
        ) const;

private:

    ClassIdToMetaObject     from_class_id;
        /**< Map ClassTypeId -> MetaObject. */

    TypeInfoToMetaObject    from_type_info;
        /**< Map std::type_info * -> MetaObject.
             Extra care must be taken if using shared libraries as std::type_info
             instances are no longer singletons. */

    ClassNameToMetaObject   from_class_name;
        /**< Map std::string -> MetaObject. */

    bool                    classes_initialized;
        /**< Classes initialize, no more registrations. */

    }; // class ClassManager

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object/class_manager_inline.h"
#endif


#endif // MASSIV_CLASS_MANAGER_H
