/** @file class_type_id.cpp
 * Class type id structure (implementation).
 */

/*
 * begin        : May 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/class_manager.h"
#include "object/metaobject.h"
#include "object/class_type_id.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "utility/quote_string.h"
#include "global/global.h"
#include "serializer/stype_int.h"
#include "serializer/stype_string.h"
#include "serializer/serializer.h"
#include "object/lib/lib_exceptions.h"

namespace Massiv {
namespace Core {


bool ClassTypeId::is_class_registered() const
    {
    return get_metaobject_no_throw() != NULL;
    }


const MetaObject * ClassTypeId::get_metaobject() const
    {
    ClassManager & class_manager = Global::class_manager();
    const MetaObject * const metaobject =
        class_manager.get_metaobject( *this );

    return metaobject;
    }


const ObjectFactory * ClassTypeId::get_factory() const
    {
    const MetaObject * const metaobject = get_metaobject();
    const ObjectFactory * const factory = metaobject->get_object_factory();
    
    return factory;
    }


const MetaObject * ClassTypeId::get_metaobject_no_throw() const
    {
    ClassManager & class_manager = Global::class_manager();
    return class_manager.get_metaobject_no_throw( *this );
    }


const ObjectFactory * ClassTypeId::get_factory_no_throw() const
    {
    const MetaObject * metaobject = get_metaobject_no_throw();
    
    if( !metaobject )
        {
        return NULL;
        }
    else
        {
        return metaobject->get_object_factory();
        }
    }


const std::type_info & ClassTypeId::get_type_info() const
    {
    return get_metaobject()->get_class_type_info();
    }


bool ClassTypeId::inherits
    (
    const std::type_info & class_type_info
    ) const
    {
    return get_metaobject()->inherits( class_type_info );
    }


bool ClassTypeId::inherits
    (
    const ClassTypeId & class_type_id
    ) const
    {
    return get_metaobject()->inherits( class_type_id.get_type_info() );
    }


std::string ClassTypeId::dump() const
    {
    std::ostringstream s;

    s << "[ kind = " << get_kind() << ", class_id = " << get_class_id() << " ]";
    return s.str();
    }


void ClassTypeId::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    SInt< UInt32, BITS_COUNT > s_id( id );
    s_id.bitstr_write( bos, desc );
    }


void ClassTypeId::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    SInt< UInt32, BITS_COUNT > s_id;
    s_id.bitstr_read( bis, desc );
    id = s_id;
    }


void ClassTypeId::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    const MetaObject * const metaobject = get_metaobject();
    tw.line() << quote_const_string( metaobject->get_class_name() );
    }
    

void ClassTypeId::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    ClassManager &      class_manager = Global::class_manager();
    SString             class_name;
    const MetaObject *  metaobject;

    // Read class name.
    
    // Note for future modifications:
    // ObjectInfo requires that if ClassTypeId::text_read() throws
    // ExUnexpectedValue(), it will not leave any unprocessed tokens
    // belonging to the ClassTypeId in the input stream.
    // Thus ObjectInfo can catch the exception and continue parsing
    // the input stream.

    class_name.text_read( tr, desc );
    
    // Find metaobject.
    
    try
        {
        metaobject = class_manager.get_metaobject( class_name );
        }
    catch( Lib::UndefinedClassTypeException & e )
        {
        throw ExUnexpectedValue
            (
            tr.unexpected_value( "'class name'", class_name ) +
            e.what()
            );
        }

    // Done.

    *this = metaobject->get_class_type_id();
    }


std::ostream & operator<<
    (
    std::ostream &      s,
    const ClassTypeId & id
    )
    {
    if( id.is_class_registered() )
        {
        TextWriter tw( s );
        id.text_write( tw, Serializer::Description() );
        }
    else
        {
        s << id.dump();
        }
    return s;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "class_type_id_inline.h"
#endif
