/** @file class_type_id.h
 * Class type id structure.
 */

/*
 * begin        : May 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CLASS_TYPE_ID_H
#define MASSIV_CLASS_TYPE_ID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_H
#include "serializer/serializer_fwd.h"
#endif


#include <iosfwd>
#include <typeinfo>


namespace Massiv {
namespace Core {


/**
 * Class kind flags.
 *
 * Managed classes of the same "class kind" are grouped into single
 * compilation/linkage unit. Each node "supports" classes of several kinds.
 * This means that the node contains all information required to instantiate
 * and provide information (access class metadata) about all classes of given
 * kinds, execute their methods, etc.. Such node must be linked with definition
 * of the class (code written by node "library" programmer) as well as code
 * generated by the @c factgen utility from class description written in IDL and
 * "class list" generated by the @c class_list utility.
 *
 * Core defines only single class kind: KIND_SHARED. All parents of a managed
 * class must be either @c KIND_SHARED or their kind must be the same as the
 * kind of the managed class. Every node must support @c KIND_SHARED classes.
 * All managed classes defined in massiv core are @c CLASS_SHARED.
 * Other class kinds are defined in src/core/object/object.idl.
 *
 * Classes of kind which is not "supported" by a node are called "alien" classes.
 * It's impossible to instantiate alient classes (or use them in any other way).
 * Node program must not link with any code using alien classes (including
 * definition of the classes ans sources generated from corresponding IDL).
 * However it's possible to have pointers of type "pointer to @c S", where
 * @c S is supported class, pointing to non-local object of type @c A, if
 * @c S is parent class of @c A (with limited type safety, see #ObjectPointer).
 *
 * Observe the following example. This is part of C++ declaration of a
 * @c KIND_SHARED class:
 *
 * @code
 * class MyInterface : public Object
 *     {
 *     virtual void f() = 0;
 *     };
 * @endcode
 *
 * @c KIND_SERVER class declared elsewhere:
 *
 * @code
 * class ServerOnlyObject : public MyInterface
 *     {
 *     virtual void f(); // ServerOnlyObject implements f().
 *     };
 * @endcode
 *
 * Part of program on node supporting only @c KIND_SHARED classes follows:
 *
 * @code
 * // Let it be initialized so that it refers to a ServerOnlyObject.
 * // ServerOnlyObject is an alien class to this node.
 * ObjectId object_id = ...;
 *
 * // No conversion validity is checked since object_id refers to an alien class.
 * Remote\< MyInterface \> ptr = object_id;
 *
 * // Calls MyInterface::f() on object_id, fails if object_id does not
 * // support MyInterface. This is tested by the server-side. See rpc.h for
 * // more information about RPC and what exceptions it might throw.
 * ptr-\>sync_f();
 * @endcode
 *
 * @ingroup objectid
 */
enum ClassKind
    {
    KIND_SHARED = 0,
        /**< Every node must support KIND_SHARED classes. */

    }; // enum ClassKind


/**
 * Identification of a managed class.
 * Comprises of the class kind (see #ClassKind) and class id which identifies
 * the class within group of all classes of given kind.
 *
 * @ingroup objectid
 */
class MASSIV_EXPORTED ClassTypeId
    {
public:

    /**
     * ClassTypeId hash function.
     */
    struct Hash
        {

        INLINE size_t operator()
            (
            const ClassTypeId & class_type_id
            ) const throw();

        }; // struct Hash

    /**
     * ClassTypeId equalkey function.
     */
    struct EqualKey
        {

        INLINE bool operator()
            (
            const ClassTypeId & class_type_id_1,
            const ClassTypeId & class_type_id_2
            ) const throw();

        }; // struct EqualKey

    /**
     * ClassTypeId less function.
     */
    struct Less
        {

        INLINE bool operator()
            (
            const ClassTypeId & class_type_id_1,
            const ClassTypeId & class_type_id_2
            ) const throw();

        }; // struct Less

    friend struct ClassTypeId::Hash;
    friend struct ClassTypeId::EqualKey;
    friend struct ClassTypeId::Less;
    friend class ClassManager;

public:

    enum
        {
        KIND_BITS_COUNT         = 2,
        CLASS_ID_BITS_COUNT     = 10,
        BITS_COUNT              = ( KIND_BITS_COUNT + CLASS_ID_BITS_COUNT ),
        };

public:

    /**
     * Default constructor.
     */
    INLINE ClassTypeId();

    /**
     * Constructor.
     */
    INLINE ClassTypeId
        (
        int     the_kind,
        int     the_class_id
        );

    /**
     * Copy constructor.
     */
    INLINE ClassTypeId
        (
        const ClassTypeId & other
        );

    /**
     * Operator=.
     */
    INLINE ClassTypeId & operator=
        (
        const ClassTypeId & other
        );

public:

    /**
     * Operator==.
     */
    INLINE bool operator==
        (
        const ClassTypeId & other
        ) const;

    /**
     * Operator!=.
     */
    INLINE bool operator!=
        (
        const ClassTypeId & other
        ) const;

public:

    /**
     * Are objects of this class supported by local node.
     *
     * @return @c true if this class type is registered to class manager,
     *         @c false otherwise. #get_metaobject() and #get_factory()
     *         will throw iff this returns @c false.
     */
    bool is_class_registered() const;

    /**
     * Returns metaobject for objects with this class id.
     * Throws UndefinedClassTypeException if there is no metaobject for this
     * class type on the current node.
     */
    const MetaObject * get_metaobject() const;

    /**
     * Returns object factory for objects with this class id.
     * Throws UndefinedClassTypeException if there is no factory for this
     * class type on the current node.
     */
    const ObjectFactory * get_factory() const;
    
    /**
     * No-throw variant of get_metaobject().
     * Returns NULL if there is no metaobject for this class type on
     * the current node.
     */
    const MetaObject * get_metaobject_no_throw() const;

    /**
     * No-throw variant of get_factory().
     * Returns NULL if there is no factory for this class type on the
     * current node.
     */
    const ObjectFactory * get_factory_no_throw() const;

    /**
     * Returns type information about object with this class id.
     * Throws UndefinedClassTypeException if there is no metadata for this
     * class type on the current node.
     */
    const std::type_info & get_type_info() const;

    /**
     * Checks whether objects with this class id
     * inherit class described by class_type_info (RTTI).
     * Throws UndefinedClassTypeException if there is no metadata for this
     * class type on the current node.
     */
    bool inherits
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * Checks whether objects with this class id
     * inherit object with another class id (RTTI).
     * Throws UndefinedClassTypeException if there is no metadata for this
     * or @c other class type on the current node.
     */
    bool inherits
        (
        const ClassTypeId & other
        ) const;

public:

    /**
     * Get class kind.
     */
    INLINE int get_kind() const;

    /**
     * Get class id unique within classes of the same kind.
     */
    INLINE int get_class_id() const;

    /**
     * Dump ClassTypeId's components.
     */
    std::string dump() const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     *
     * Written as class name, kind is NOT explicitly written.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

private:

    union
        {
        struct
            {
            UInt32  kind : KIND_BITS_COUNT;
                /**< Kind id. */

            UInt32  class_id : CLASS_ID_BITS_COUNT;
                /**< Class id unique within classes of the same kind. */

            } components;
                /**< Identifier components. */

        UInt32   id;
            /**< Identifier as a number. */
        };

    }; // class ClassTypeId


/**
 * Write class type id to stream.
 * If @c id refers to a known registered class, class name will be printed
 * instead.
 */
std::ostream & operator<<
    (
    std::ostream &      s,
    const ClassTypeId & id
    );


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "class_type_id_inline.h"
#endif


#endif // MASSIV_CLASS_TYPE_ID_H
