/** @file event.h
 * Event.
 */

/*
 * begin        : Tue Mar 19 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_EVENT_H
#define MASSIV_EVENT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif


namespace Massiv {
namespace Core {

namespace Test { class SerializerTest; }


/** @addtogroup object_management
 * @{
 */

/** @class Event
 * Event.
 * Every event is owned by one object.
 * Ownership of some event with execution time T means that
 * the owner of the event will be "delivered" (@see Object::delivered_to())
 * to the destination of the event in time T.
 * @note @c EventId associated with the event is stored outside of the Event class.
 */
class Event
    {
    friend class Test::SerializerTest;

public:

    /**
     * Constructor.
     */
    Event() :
        notify( false )
        {
        hope( execution_time.is_equal_to( STime::INVALID ) );
        hope( destination.is_null() );
        }

    /**
     * Copy constructor.
     */
     Event
         (
         const Event & e
         ) :
         execution_time( e.execution_time ),
         destination( e.destination ),
         notify( e.notify )
         {
         }

    /**
     * Constructor.
     */
    Event
        (
        const STime &       time_of_execution,
        const ObjectId &    destination_object,
        bool                the_notify
        ) :
        execution_time( time_of_execution ),
        destination( destination_object ),
        notify( the_notify )
        {
        }

    /**
     * Destructor.
     */
    ~Event()
        {
        }

    /**
     * Operator=.
     */
    Event & operator=
        (
        const Event & e
        )
        {
        execution_time       = e.execution_time;
        destination          = e.destination;
        notify               = e.notify;
        return *this;
        }

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

public:

    /**
     * Set content.
     */
    void set
        (
        const STime &       time_of_execution,
        const ObjectId &    destination_object,
        bool                the_notify
        )
        {
        execution_time    = time_of_execution;
        destination       = destination_object;
        notify            = the_notify;
        }

    /**
     * Get object id of the destination object of the event.
     */
    const ObjectId & get_destination() const
        {
        return destination;
        }

    /**
     * Get execution time of the event.
     */
    const STime & get_execution_time() const
        {
        return execution_time;
        }
    
    /**
     * Check whether owner of the objects should be notified
     * when the event is executed.
     */
    bool notify_owner() const
        {
        return notify;
        }

private:

    STime       execution_time;
        /**< Execution time of the event. */

    ObjectId    destination;
        /**< Destination (recipient) object of the event.
         * This is object id of the object that is passed
         * as recipient in @c Object::delivered_to().
         */
    
    SBool       notify;
        /**< Whether owner object of the events will be notified
         * when the event is executed.
         * See Object's @c migrate_to() and @c schedule_to() for more info.
         */

    }; // Event

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_EVENT_H
