/** @file event_handle.h
 * Event handle.
 */

/*
 * begin        : Wed May 15 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_EVENT_HANDLE_H
#define MASSIV_EVENT_HANDLE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_EVENTID_H
#include "object/eventid.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/** @class EventHandle
 * Event handle is handle to some scheduled event.
 * With event handle you can kill scheduled event.
 * Only public events can be killed.
 * Event handle is safe - you can call @c kill() on event handle
 * which points to event that was already executed (the event
 * itself is destroyed) or the object the event belongs to was destroyed.
 * Event handle is serializable.
 * @todo Add rescheduling?
 */
class MASSIV_EXPORTED EventHandle
    {
    friend class ObjectProperty;

public:

    /**
     * EventHandle hash functor.
     */
    struct Hash
        {
        std::size_t operator()
            (
            const EventHandle & handle
            );
        };

    friend struct Hash;

public:

    /**
     * Constructor.
     */
    EventHandle();

    /**
     * Copy constructor.
     */
    EventHandle
        (
        const EventHandle & event_handle
        );

    /**
     * Operator=.
     */
    EventHandle & operator=
        (
        const EventHandle & event_handle
        );

    /**
     * Destructor.
     */
    ~EventHandle();

public:

    /**
     * Kill the event this handle points to.
     * This function is safe, you can call it to dead events.
     * @note Nothing is done if kill() is called on event handle which
     * points to event scheduled for non-local object.
     */
    void kill();

    /**
     * Clear handle content.
     * The event this handle points to stays unchanged.
     */
    void clear();

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     * Writes object id and event id separated by space.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );
    
    /**
     * Dumps event handle.
     */
    friend std::ostream & operator<<
        (
        std::ostream &      stream,
        const EventHandle & handle
        );

private:

    /**
     * Set event handle content.
     */
    void set
        (
        const ObjectId &    object_id,
        const EventId &     event_id
        );

private:

    ObjectId        object_id;
        /**< Object the event belongs to. */

    EventId         event_id;
        /**< Identification of the scheduled event. */

    }; // EventHandle

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_EVENT_HANDLE_H
