/** @file eventid.h
 * Event identification.
 */

/*
 * begin        : Wed May 15 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_EVENTID_H
#define MASSIV_EVENTID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#include <cstdlib>
#include <iosfwd>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/** @class EventId
 * Event identification.
 * EventId can be private or public.
 * Private event id can't be transferred over network or saved to file.
 * See event id serialization.
 *
 * @note Id of some event is not unique within entire system.
 * EventId is unique only within one object.
 * So if somebody want to uniquely identify some event
 * in the system, he must use both @c ObjectId and @c EventId.
 *
 * @note Triple <source_object_id, destination_object_id, time>
 * does not uniquely identifies event.
 */
class EventId
    {
    friend class LocalSchedule;

public:

    /**
     * EventId hash function.
     */
    struct Hash
        {

        /** EventId hash function operator. */
        std::size_t operator()
            (
            const EventId & event_id
            ) const
            {
            return static_cast< std::size_t >( event_id.id );
            }

        }; // Hash

    /**
     * EventId equal key function.
     */
    struct EqualKey
        {

        /** EventId equal key function operator. */
        bool operator()
            (
            const EventId & event_id1,
            const EventId & event_id2
            ) const
            {
            return event_id1.id == event_id2.id;
            }

        }; // EqualKey

    /**
     * EventId less function.
     */
    struct Less
        {

        bool operator()
            (
            const EventId & event_id1,
            const EventId & event_id2
            ) const
            {
            return event_id1.id < event_id2.id;
            }

        }; // Less

    friend struct EventId::Hash;
    friend struct EventId::EqualKey;
    friend struct EventId::Less;

public:

    enum
        {
        NULL_ID = 0
        };

    /**
     * Default constructor.
     * Constructs private null id.
     */
    EventId()
        {
        set_private_null();
        }

    /**
     * Copy constructor.
     */
    EventId
        (
        const EventId & event_id
        ) :
        id( event_id.id )
        {
        }

    /**
     * Operator=.
     */
    EventId & operator=
        (
        const EventId & event_id
        )
        {
        id = event_id.id;
        return *this;
        }

    /**
     * Operator==.
     */
    bool operator==
        (
        const EventId & event_id
        ) const
        {
        return id == event_id.id;
        }

    /**
     * Operator!=.
     */
    bool operator!=
        (
        const EventId & event_id
        ) const
        {
        return id != event_id.id;
        }

    /**
     * Sets event id to private null id.
     */
    void set_private_null()
        {
        id = NULL_ID;
        }

    /**
     * Sets event id to public null id.
     */
    void set_public_null()
        {
        id = NULL_ID;
        components.is_public = 1;
        }

    /**
     * Checks whether the (relative) event id is null.
     * @note Doesn't check wheter the id is public or private.
     * Can return true even if the id is private or public.
     */
    bool is_null() const
        {
        return components.relative_id == 0;
        }

    /**
     * Checks whether the event id is private.
     * @note Doesn't check whether relative event id is null.
     * Can return true even if relative event id is null.
     */
    bool is_private() const
        {
        return components.is_public == 0;
        }

public:

    /**
     * Serialize to bitstream.
     * @note If the event id is private then relative
     * id is NOT serialized.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     * @note If the event stored in the stream
     * is private then relative id is set to null.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     * @note If the event id is private then relative
     * id is NOT serialized.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     * @note If the event stored in the stream
     * is private then relative id is set to null.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

    /**
     * Dumps event id.
     */
    friend std::ostream & operator<<
        (
        std::ostream &  stream,
        const EventId & event_id
        );

public:

    enum
        {
        RELATIVE_ID_BITS_COUNT = 40
            /**< Number of bits required to store relative event id. */
        };

private:

    union
        {
        struct
            {
            Int64   is_public : 1;
                /**< Set if the event id is public. */

            Int64   relative_id : RELATIVE_ID_BITS_COUNT;
                /**< Relative event id. */

            } components;
                /**< Identifier components. */

        Int64   id;
            /**< Identifier as a number. */
        };

    }; // EventId

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_EVENTID_H
