/** @file global_search_manager.cpp
 * Global search manager.
 */
 
/*
 * begin        : Wed Apr 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/global_search_manager.h"
#include "object/object_manager.h"
#include "database/node_database.h"
#include "global/global.h"
#include "utility/hash.h"
#include "pointer/objectid.h"
#include "sysevent/global_search_msg.h"
#include <sstream>


namespace Massiv {
namespace Core {


//------------------------------------------------------------------------------
// GlobalSearchManager::My
//------------------------------------------------------------------------------


struct GlobalSearchManager::My
    {
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );
    
public:
    /** Constructor. */
    My() :
        server_node_count( 0 ),
        new_session_id( 1 )
        {
        }
    
    /**
     * Global search session identification.
     * If changed do not forget to update global search messages.
     */
    typedef UInt32  SessionId;
    
    /**
     * Global search session.
     * Session represents result of global searching for one object.
     * The object id of the object we are searching for is not
     * stored in @c Session itself but in @c SessionMap.
     */
    class Session
        {
    public:
        /** Constructor. */
        Session() :
            id( 0 ),
            reply_count( 0 ),
            best_reply_object_status( ObjectStatus::UNKNOWN )
            {
            }
        
        /** Copy constructor. */
        Session
            (
            const Session & other
            ) :
            id( other.id ),
            reply_count( other.reply_count ),
            best_reply_object_status( other.best_reply_object_status ),
            best_reply_node( other.best_reply_node )
            {
            }
        
        /** Operator=. */
        Session & operator=
            (
            const Session & other
            )
            {
            id                       = other.id;
            reply_count              = other.reply_count;
            best_reply_object_status = other.best_reply_object_status;
            best_reply_node          = other.best_reply_node;
            return *this;
            }

    public:
        SessionId       id;
            /**< Session id.
             * @see GlobalSearchMessage, GlobalSearchReplyMessage
             */
        
        unsigned int    reply_count;
            /**< Number of replies we have received for this session.
             */
        
        ObjectStatus    best_reply_object_status;
            /**< The best object status we have received in reply to this
             * session.
             */

        NodeId          best_reply_node;
            /**< The node associated with the best object status we have
             * received in reply to this session.
             */
        }; // Session
    
    /** Global search session map. A session is identified by an object id. */
    typedef ext::hash_map
        <
        ObjectId,
        Session,
        ObjectId::Hash,
        ObjectId::EqualKey
        > SessionMap;
        
public:

    SessionMap      sessions;
        /**< Global search sessions. */
    
    unsigned int    server_node_count;
        /**< Number of server nodes perticipating in global search. */

    SessionId       new_session_id;
        /**< New available session id. */
    
    }; // GlobalSearchManager::My

    
//------------------------------------------------------------------------------
// GlobalSearchManager
//------------------------------------------------------------------------------
    

GlobalSearchManager::GlobalSearchManager() :
    my( NULL )
    {
    my = new My();
    }
    

GlobalSearchManager::~GlobalSearchManager()
    {
    delete my;
    my = NULL;
    }


void GlobalSearchManager::initialize()
    {
    // Cleanup if reinitializing.
    my->sessions.clear();
    my->server_node_count = 0;
    // new_session_id can stay unmodified.
    
    hope( my->new_session_id > 0 );
    
    // Determine number of server nodes.
    class NodeEnumerator : public NodeDatabase::Enumerator
        {
    public:
        
        NodeEnumerator() :
            num_nodes( 0 )
            {
            }
        
        virtual void operator()
            (
            const NodeDatabase::NodeData & data
            )
            {
            num_nodes++;
            }
            
        int num_nodes;
            
        }; // NodeEnumerator
        
    // Initialize.
    NodeDatabase & node_database = Global::node_database();
    NodeEnumerator enumerator;
    node_database.enumerate_nodes( NodeId::TYPE_SERVER, enumerator );
    my->server_node_count = enumerator.num_nodes;
    }

    
void GlobalSearchManager::initiate_search_for
    (
    const ObjectId & object
    )
    {
    My::SessionMap::iterator session_it;
    
        {
        std::ostringstream ss;
        ss  << "Initiating global search for object " << object << ".";
        Global::log_debug
            (
            Status::FACILITY_OBJECT_LOCALIZATION,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
    
    
    session_it = my->sessions.find( object );
    if( session_it != my->sessions.end() )
        {
        // There is already some pending session searching for the object.
        // Nothing to do.
        }
    else
        {
        // Create new session.
        My::Session     new_session;
        
        // Init.
        new_session.id = my->new_session_id++;
        
        // Insert into session map.
        my->sessions.insert
            (
            My::SessionMap::value_type
                (
                object,
                new_session
                )
            );
        
        // Contact all server nodes.
            {
            // Boovie: Note for future optimalization: The message is sent to
            // all server nodes thus it is also send to local system.
            // Care must be taken when someone wants to optimize this
            // not to send the message to local system!
            // In that case you must set reply_count of new session to 1
            // and you MUST check if the object is owned as it would be
            // done by the message! Otherwise you could notify object
            // manager about NOT_ACCESSIBLE object but the object could
            // be owned.
            // Also care must be taken if there is only one server node
            // running the simulation. In that case you must ensure that
            // the new session will not stay in session map forever 
            // (because no reply message will be sent).
            // If you think you can put the test at the end of
            // initiate_search_for() and notify object manager from this
            // function then think twice, because initiate_search_for()
            // is called from object manager's localization procol
            // and you shoud verify that you can notify the object manager
            // in the middle of handling the localization protocol!
            //
            // My opinion is that not sending the message to local system
            // will not save the world, so just leave it as it is.
            // Leave it clear, safe and bug-free.
            
            GlobalSearchMessage msg
                (
                object,
                new_session.id
                );
            
            // Send the message to all servers (including local node).
            msg.multicast_to_server_nodes( NetworkCommon::STREAM_RELIABLE );
            }
        }
    }
    
    
void GlobalSearchManager::cancel_search_for
    (
    const ObjectId & object
    )
    {
    My::SessionMap::iterator session_it;
    
    session_it = my->sessions.find( object );
    if( session_it != my->sessions.end() )
        {
        my->sessions.erase( session_it );
        
            {
            std::ostringstream ss;
            ss  << "Global search for " << object << " cancelled.";
            Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, ss.str() );
            }
        }
    }

    
void GlobalSearchManager::handle_global_search_message
    (
    const GlobalSearchMessage & msg
    )
    {
    ObjectManager &     object_manager = Global::object_manager();
    const NodeId &      msg_sender = msg.get_sender();
    
    ObjectStatus        object_status = ObjectStatus::UNKNOWN;
    NodeId              node;

    hope( msg.get_initiator() == msg_sender );         // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    
    if( object_manager.is_object_owned( msg.object ) )
        {
        // The object is owned ie. it is located on local system.
        object_status = ObjectStatus::ALIVE;
        node = Global::get_local_node_id();
        }
    else
        {
        // The object is not located on local system. Use localization
        // cache to get status of the object.
        object_manager.get_remote_object_status( msg.object, object_status, node );
        }
    
    // And finally send reply.
        {
        GlobalSearchReplyMessage msg_reply
            (
            msg.object,
            object_status,
            node,
            msg.session_id
            );
        
        msg_reply.send( msg_sender, NetworkCommon::STREAM_RELIABLE );
        }
    }
    
    
void GlobalSearchManager::handle_global_search_reply_message
    (
    const GlobalSearchReplyMessage & msg
    )
    {
    const NodeId &           msg_sender = msg.get_sender();
    const ObjectId &         object = msg.object;
    const ObjectStatus &     object_status = msg.object_status;
    const NodeId &           node = msg.node;
    
    My::SessionMap::iterator session_it;
    My::Session *            session = NULL;
    bool                     finish_the_session = false;

    hope( msg.get_initiator() == msg_sender );         // Already verified.
    hope( msg_sender.is_type( NodeId::TYPE_SERVER ) ); // Already verified.
    
    session_it = my->sessions.find( object );
    if( session_it == my->sessions.end() )
        {
        // The session the message is about is already gone or cancelled.
        // Ignore the message.
        return;
        }
    
    session = &( session_it->second );
    if( session->id != msg.session_id )
        {
        // The session the message is about is already gone or cancelled.
        // Ignore the message.
        return;
        }
    
    // Update the session.
    switch( object_status.get_value() ) 
        {
        case ObjectStatus::ALIVE:
            // Do not update best_reply_node in case that somebody is
            // trying to notify us from its localization cache that the
            // object is located on our local node.
            // The test "node != msg_sender" ensures that we will
            // not lost ALIVE status if the global search was initiated
            // on local node but the object was already owned.
            // Note: The condition "node != msg_sender" means
            //       "the object is not owned on the msg_sender".
            // Don't forget that GlobalSearchReplyMessage is also
            // sent by local node to itself if the local node is global search
            // initiator.
            if( session->best_reply_object_status == ObjectStatus::UNKNOWN )
                {
                session->best_reply_object_status = object_status;
                session->best_reply_node          = node;
                }
            else if( session->best_reply_object_status == ObjectStatus::NOT_ACCESSIBLE )
                {
                // Ignore the status. The object should not exist in the simulation.
                // Note: This case should never happen.
                    
                // Log.
                    {
                    std::ostringstream ss;
                    ss  << "Global search session for object " << object << " contains NOT_ACCESSIBLE status. Should never happen "
                        << "because sessions with NOT_ACCESSIBLE status are immediatelly closed.";
                    Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST, ss.str() );
                    }     
                }
            else
                {
                if( node == msg_sender )
                    {
                    // Sender is probably the owner of the object (the sender was owner
                    // when he sent the message).
                    session->best_reply_object_status = object_status;
                    session->best_reply_node          = node;
                    }
                else
                    {
                    // The received status is status from the sender's localization cache.
                    // It can be untrustworthy and we already have some untrustworthy alive status.
                    // We don't know whether the new status is better than the stored one.
                    // Ignore the new status.
                    }
                }
            
            // Force the session to be finished only in case that
            // the object is owned by the sender node.
            // Note: We can't finish the session if we receive ALIVE status
            // only from sender's localization cache because it could cause
            // errors in message delivery in case that there is a cycle
            // in localization caches of server nodes!
            if( node == msg_sender )
                {
                finish_the_session = true;
                }
            break;

        case ObjectStatus::NOT_ACCESSIBLE:
            session->best_reply_object_status = object_status;
            session->best_reply_node          = node;
            
            // Force the session to be finished.
            // Note: There is not similiar problem as with ALIVE status above
            // because NOT_ACCESSIBLE status can't be "obsolete".
            finish_the_session = true;
            break;
        
        case ObjectStatus::UNKNOWN:
            // Do not update the session.
            break;
        }
    session->reply_count++;
    
    // Check if we have received replies from all server nodes.
    // In that case the session can be finished.
    if( session->reply_count >= my->server_node_count )
        {
        finish_the_session = true;
        }

    if( finish_the_session )
        {
        // Check if the object exists somewhere in the simulation.
        if( session->best_reply_object_status == ObjectStatus::UNKNOWN )
            {
            // No server node owns the object. The object is not accessible
            // in the simulation.
            session->best_reply_object_status = ObjectStatus::NOT_ACCESSIBLE;
            session->best_reply_node          = Global::get_local_node_id();
            }
            
        // Log.
            {
            std::ostringstream ss;
            ss  << "Global search for " << object << " finished: "
                << "object status " << session->best_reply_object_status << ", "
                << "node " << session->best_reply_node << ". "
                << "Search finished after " << session->reply_count << " replies. ";
            Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, ss.str() );
            }
        
        // Notify object manager.
            {
            ObjectManager & object_manager = Global::object_manager();
            ObjectStatusMessage object_status_msg
                (
                object,
                session->best_reply_object_status,
                session->best_reply_node,
                false // Untrustworthy.
                );
            object_manager.update_remote_object_status_cache( object_status_msg );
            }
        
        // Remove the session from the session map.
        my->sessions.erase( session_it );
        session = NULL;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
