/** @file global_search_manager.h
 * Global search manager.
 */
 
/*
 * begin        : Wed Apr 10 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_GLOBAL_SEARCH_MANAGER_H
#define MASSIV_GLOBAL_SEARCH_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h" // ObjectId
#endif


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Global search manager.
 * The manager manages global search requests to find objects
 * in the distributed simulation.
 *
 * For each object we are searching for the manager creates
 * one session. The session is internal structure of the manager
 * and represents information about the object the manager gathered
 * from server nodes.
 * Once a new session is created, the manager contacts all server nodes
 * to gather information about the object. Once enough information
 * if gathered about the object the manager notifies object manager
 * about the status of the object and close the session.
 * Any received messages related to a closed or cancelled session are ignored.
 *
 * @see ObjectManager
 */
class GlobalSearchManager
    {
    NO_COPY_CONSTRUCTOR( GlobalSearchManager );
    NO_ASSIGNMENT_OPERATOR( GlobalSearchManager );
    
public:
    /** Constructor. */
    GlobalSearchManager();
    
    /** Destructor. */
    ~GlobalSearchManager();
    
    /**
     * Initialize.
     * @pre NodeManager is initialized.
     */
    void initialize();
    
public:
    /**
     * Initiate new global search for @a object.
     * If there is already unfinished session searching for
     * the object nothing is done. Otherwise new session is created.
     */
    void initiate_search_for
        (
        const ObjectId & object
        );
    
    /**
     * Cancel any global search (initiated from local system) for @a object.
     * Unfinished session searching for the object is cancelled. Object manager
     * will not be notified even if there is some partial information gathered
     * about the object.
     * If there is no opened session for the object this function does nothing.
     */
    void cancel_search_for
        (
        const ObjectId & object
        );

public:
    /**
     * Handle incoming GlobalSearchMessage.
     */
    void handle_global_search_message
        (
        const GlobalSearchMessage & message
        );
    
    /**
     * Handle incoming GlobalSearchReplyMessage.
     */
    void handle_global_search_reply_message
        (
        const GlobalSearchReplyMessage & message
        );
    
private:
    /** Global search manager internal structure. */
    struct My;
    
private:
    My *    my;
        /**< Global search manager internal data. Always non-null. */
    
    }; // GlobalSearchManager

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_GLOBAL_SEARCH_MANAGER_H
