/** @file account_object_interface.h
 * Account object interface.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ACCOUNT_OBJECT_INTERFACE_H
#define MASSIV_ACCOUNT_OBJECT_INTERFACE_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif



namespace Massiv {
namespace Core {
namespace Lib {


class NodeObjectInterface;

/**
 * AccountObjectInterface, defined by the core, is an interface that
 * must be implemented by AccountObject, defined by client-code.
 * AccountObject is associated with a particular client node and serves
 * as a proxy object to server nodes.
 * Client nodes are not allowed to directly manipulate objects stored
 * on server nodes except for AccountObject. All the actions that client
 * wants to do with remote objects must be done via AccountObject.
 * Technically, the only allowed object migration (ie. sending messages,
 * scheduling objects, RPC, etc.) from client node to an object placed
 * on a server node is to/via its AccountObject object.
 * @note AccountObject has assigned a special registered object_id that
 * allows client node to "talk" to it.
 *
 * AccountObjectInterface and its descendants have by default set 
 * @c simulation_startup_notify IDL class attribute to true.
 * Thus @c Object::object_updated() is called on them with
 * @c SIMULATION_STARTUP argument when simulation is beeing
 * restored from an archive.
 *
 * @ingroup object
 */
class MASSIV_EXPORTED AccountObjectInterface : public Object
    {
public:
    
    MASSIV_OBJECT( AccountObjectInterface );

    //
    // Callbacks from massiv core.
    //

    /**
     * Called by massiv core when client associated with this account
     * connects to the simulation.
     *
     * This function is called by massiv core using asynchronous rpc.
     *
     * @param node_object Points to a NodeObject that represents connected
     *      client node.
     */
    virtual void client_connected
        (
        const WeakPointer< NodeObjectInterface > &  node_object
        ) = 0;

    /**
     * Called by massiv core when client associated with this account
     * disconnects from the simulation.
     *
     * This function is called by massiv core using asynchronous rpc.
     */
    virtual void client_disconnected() = 0;
    
    /**
     * Called by massiv core to log system message(s).
     * In order to receive messages from @c sender node, account object must
     * be subscribed to @c LogDestinationClients logger destination on that
     * node.
     *
     * This function is called by massiv core using asynchronous rpc.
     * @param sender    sender node
     * @param messages  system messages
     */
    virtual void log_system_messages
        (
        const WeakPointer< NodeObjectInterface > &  sender,
        const PValueArray< SString > &              messages
        ) = 0;

    /**
     * Called by RPCObject when clients sends RPC request to account object.
     *
     * When delivering RPC request from client, RPCObject checks that callee
     * inherits AccountObjectInterface and then calls this method. Account
     * should check that request.node_object is expected client node, that
     * client does not try to call "private" method, etc.
     *
     * @return @c true if request passes the check, @c false if it should
     *         be discarded.
     */
    virtual bool check_client_rpc_request
        (
        const Pointer< RPCObject > & request
        ) const = 0;
    
    }; // AccountObjectInterface

    
} // namespace Massiv::Core::Lib
} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_CORE_LIB_ACCOUNT_OBJECT_INTERFACE_H
