/** @file lib_exceptions.h
 * Lib exceptions predefined by the core.
 * The identifiers always have "Exception" suffix whereas core exceptions
 * "Ex" prefix. This allows for simple discrimination between core
 * and lib (managed) exceptions.
 */

/*
 * begin        : Tue Oct 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LIB_EXCEPTIONS_H
#define MASSIV_LIB_EXCEPTIONS_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif



namespace Massiv {
namespace Core {
namespace Lib {


/** @defgroup ex_lib Library Exceptions
 * Managed exceptions (throwable managed objects).
 *
 * @ingroup exception
 * @{
 */

/**@def BASIC_CORE_LIB_EXCEPTION
 * Defines a Core::Lib exception class.
 * @param Name Name of the new exception class.
 * @param Base Name of the base exception class. Any throwable object.
 * @note IDL for the class must be written manually.
 */
#define BASIC_CORE_LIB_EXCEPTION( Name, Base ) \
class MASSIV_EXPORTED Name : public Base \
    { \
public: \
    MASSIV_OBJECT( Name ); \
    \
    Name() \
        { \
        initialize_object(); \
        } \
    \
    Name \
        ( \
        const std::string & message \
        ) : Base( message ) \
        { \
        initialize_object(); \
        } \
    \
    Name \
        ( \
        const Name & exception \
        ) : Base( exception )\
        { \
        initialize_object(); \
        } \
    };


/**@def BASIC_LIB_EXCEPTION
 * Defines a lib exception class, not exported from the Core::Lib.
 * The parameters are the same like those for @c BASIC_CORE_LIB_EXCEPTION.
 */
#define BASIC_LIB_EXCEPTION( Name, Base ) \
class Name : public Base \
    { \
public: \
    MASSIV_OBJECT( Name ); \
    \
    Name() \
        { \
        initialize_object(); \
        } \
    \
    Name \
        ( \
        const std::string & message \
        ) : Base( message ) \
        { \
        initialize_object(); \
        } \
    \
    Name \
        ( \
        const Name & exception \
        ) : Base( exception )\
        { \
        initialize_object(); \
        } \
    };


/**
 * Generic core exception.
 * Any core exception (Core::ExXXX) is translated to CoreException if
 * such an exception occurs and need be transmitted over network.
 */
BASIC_CORE_LIB_EXCEPTION( CoreException, ::Massiv::Core::Throwable );


/**
 * Generic managed exception used by %Massiv core runtime.
 */
BASIC_CORE_LIB_EXCEPTION( RuntimeException, ::Massiv::Core::Throwable );

// FIXME: document these exceptions briefly.

    BASIC_CORE_LIB_EXCEPTION( InvalidArgumentException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( CanNotAccessObjectException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( IllegalPointerConversionException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( ObjectNotOnLocalSystemException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( PointerPolicyViolationException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( RemoteCallFailedException, RuntimeException );
    
    BASIC_CORE_LIB_EXCEPTION( UndefinedClassTypeException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( InvalidEventException, RuntimeException );

    BASIC_CORE_LIB_EXCEPTION( AbstractClassException, RuntimeException );

/** @} */ // defgroup ex_lib


} // namespace Massiv::Core::Lib
} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_LIB_EXCEPTIONS_H
