/** @file node_object_interface.h
 * Node object interface.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_OBJECT_INTERFACE_H
#define MASSIV_NODE_OBJECT_INTERFACE_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif



namespace Massiv {
namespace Core {
namespace Lib {


class AccountObjectInterface;

/**
 * NodeObject is a special object that represents a node connected
 * to simulation. NodeObjectInterface, defined by the core, defines
 * interface that must be implemented by NodeObject.
 * NodeObject is implemented by client-code.
 *
 * Technically, each node is identified by its NodeId (for NodeManager
 * purposes) or node proxy ObjectId (to be visible to client code).
 * NodeObject's object id is equal to the proxy ObjectId.
 * This allows to migrate/replicate objects to a specific node
 * (by migrating/replicating the objects to NodeObject).
 * NodeObjectInterface is a base class for client-code-defined node objects.
 * Upon core startup NodeManager call-backs the code to create
 * (the only) NodeObject for the current node. The object is always
 * created with the same object_id (this is hard-wired).
 * 
 * NodeObject always resides on the node it represents, it is never replicated,
 * never migrates.
 * Since NodeObject represents a connected node it is not a persistent
 * object and is not archived.
 *
 * @ingroup object
 */
class MASSIV_EXPORTED NodeObjectInterface : public Object
    {
    MASSIV_OBJECT( NodeObjectInterface );
    
public:

    typedef NodeId::Type    NodeType;

public:

    /**
     * Returns node type.
     * To be implemented by client code.
     */
    virtual NodeType get_node_type() const = 0;
    
    /**
     * Create new account object.
     * Used by automatic account creation feature, see #AccountManager
     * for more info.
     * @return Pointer to newly created account object or null.
     * @note To disable automatic account creation, simply return null
     *      pointer.
     */
    virtual Pointer< AccountObjectInterface > create_account_object() = 0;

    }; // NodeObjectInterface


} // namespace Massiv::Core::Lib
} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_CORE_LIB_NODE_OBJECT_INTERFACE_H
