/** @file local_schedule.cpp
 * Local schedule maintained by object.
 */

/*
 * begin        : Wed May 15 2002
 * copyright    : (C) 2002 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "utility/heap_impl.h"
#include "utility/text_writer.h"
#include "object/local_schedule.h"
#include "serializer/stype_int.h"
#include "serializer/stype_vli.h"
#include "global/global.h"
#include "database/node_database.h"
         

namespace Massiv {
namespace Core {


LocalSchedule::LocalSchedule()
    {
    clear();
    }


LocalSchedule::~LocalSchedule()
    {
    clear();
    }


bool LocalSchedule::is_empty() const
    {
    return schedule.empty();
    }


int LocalSchedule::get_size() const
    {
    return (int) schedule.size();
    }


EventId LocalSchedule::add_new_public_event
    (
    const Event & e
    )
    {
    EventId eid = generate_public_event_id();
    insert_event( eid, e );
    return eid;
    }


void LocalSchedule::add_new_private_event
    (
    const Event & e
    )
    {
    EventId eid;
    hope( eid.is_private() );
    insert_event( eid, e );
    }


bool LocalSchedule::remove_event
    (
    const EventId & event_id,
    Event &         event_out
    )
    {
    EventSchedule::iterator it;

    // Private events are ignored.
    if( event_id.is_private() )
        {
        return false;
        }

    // Find the event.
    for( it = schedule.begin(); it != schedule.end(); it++ )
        {
        if( it->second.event_id == event_id )
            {
            // Remove the event from the schedule.
            event_out = it->second.event;
            schedule.erase( it );

            // Zero @c schedule_delay if schedule is empty.
            if( is_empty() )
                {
                schedule_delay = STime();
                }

            return true;
            }
        }

    return false;
    }


void LocalSchedule::remove_first_event
    (
    bool reset_delay
    )
    {
    hope( !is_empty() );

    // Remove first event from the schedule.
    schedule.erase( schedule.begin() );

    // Zero @c schedule_delay if @c reset_delay == true or schedule is empty.
    if( reset_delay || is_empty() )
        {
        schedule_delay = STime();
        }
    }


bool LocalSchedule::get_first_event
    (
    Event & event_out
    ) const
    {
    if( is_empty() )
        {
        return false;
        }

    const Event & e = schedule.begin()->second.event;
    event_out.set( e.get_execution_time() + schedule_delay, e.get_destination(), e.notify_owner() );
    return true;
    }


STime LocalSchedule::get_time_of_first_event() const
    {
    hope( !is_empty() );
    
    STime execution_time = schedule.begin()->second.event.get_execution_time();
    STime delayed_time = execution_time + schedule_delay;
    
    return delayed_time;
    }


void LocalSchedule::delay_schedule_processing
    (
    const STime &   time
    )
    {
    if( !is_empty() )
        {
        schedule_delay += time;
        }
    }


bool LocalSchedule::is_delayed() const
    {
    return schedule_delay != STime();
    }


void LocalSchedule::check_client_to_server_events
    (
    const NodeId &      client_node,
    const ObjectId &    object_id,
    const ObjectId &    destination
    ) const
    {
    hope( client_node.is_type( NodeId::TYPE_CLIENT ) );

    if( get_size() != 1 )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " sent object " << object_id << " with " <<
               ( get_size() ? "multiple" : "zero" ) << " scheduled events.";
        throw ExSerializationError( oss.str() );
        }

    const Item & item = schedule.begin()->second;
    const Event & event = item.event;
    const NodeDatabase::NodeData & client_data = Global::node_database().get_node_data( client_node );
    const ObjectId account_object = client_data.get_account_object_id();
    const STime max_delivery_time( STime::current_time() + 60.0 );

    // Destination of the event should be the same as destination of the migration message.
    if( event.get_destination() != destination )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " sent object " << object_id <<
               ", but destination of the object and destination of the migration"
               " message the object was received in differ.";
        throw ExSerializationError( oss.str() );
        }

    if( destination.is_null() )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " sent object " << object_id <<
               " with NULL migration destination.";
        throw ExSerializationError( oss.str() );
        }

    // Destination should be account object of the client.
    if( destination != account_object )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " sent object " << object_id <<
               " to " << destination << " which is different from its account object " <<
               account_object << ".";
        throw ExSerializationError( oss.str() );
        }

    // The objects should always be notified of the event (delivered_to(), delivery_failed()).
    if( !event.notify_owner() )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " sent object " << object_id <<
               " with disabled delivery notification.";
        throw ExSerializationError( oss.str() );
        }

    // The events should be executed in near future (the object won't live forever).
    if( event.get_execution_time() > max_delivery_time )
        {
        std::ostringstream oss;
        oss << "Client " << client_node << " scheduled delivery of object " << object_id <<
               " to " << event.get_execution_time() << ", which is much larger than" <<
               " current time " << STime::current_time() << ". We don't want to keep " <<
               " client objects on server nodes forever.";
        throw ExSerializationError( oss.str() );
        }
    }


void LocalSchedule::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    SVli<UInt32>                                    event_count( get_size() );
    SInt<Int64, EventId::RELATIVE_ID_BITS_COUNT>    next_id( next_public_event_id );
    EventSchedule::const_iterator                   it;
    
    // Write local schedule state.
    next_id.bitstr_write( bos, desc );

    // Write event count.
    event_count.bitstr_write( bos, desc );

    // Write scheduled events.
    for( it = schedule.begin(); it != schedule.end(); it++ )
        {
        it->second.event_id.bitstr_write( bos, desc );
        it->second.event.bitstr_write( bos, desc );
        }
    }


void LocalSchedule::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    SVli<UInt32>                                    event_count;
    SInt<Int64, EventId::RELATIVE_ID_BITS_COUNT>    next_id;
    Event                                           event;
    EventId                                         event_id;
    UInt32                                          i;

    clear();

    // Read local schedule state.
    next_id.bitstr_read( bis, desc );
    next_public_event_id = next_id;

    // Read event count.
    event_count.bitstr_read( bis, desc );

    // Read events.
    for( i = 0; i < event_count; i++ )
        {
        event_id.bitstr_read( bis, desc );
        event.bitstr_read( bis, desc );
        insert_event( event_id, event );
        }
    }


void LocalSchedule::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    SVli<UInt32>                                    event_count( get_size() );
    SInt<Int64, EventId::RELATIVE_ID_BITS_COUNT>    next_id( next_public_event_id );
    EventSchedule::const_iterator                   it;
    
    // Write local schedule state.
    next_id.text_write( tw, desc );
    tw.write_space();

    // Write event count.
    event_count.text_write( tw, desc );
    tw.write_space();

    // Write scheduled events.
    for( it = schedule.begin(); it != schedule.end(); it++ )
        {
        it->second.event_id.text_write( tw, desc );
        tw.write_space();
        it->second.event.text_write( tw, desc );
        tw.write_space();
        }
    }


void LocalSchedule::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    SVli<UInt32>                                    event_count;
    SInt<Int64, EventId::RELATIVE_ID_BITS_COUNT>    next_id;
    Event                                           event;
    EventId                                         event_id;
    UInt32                                          i;

    clear();

    // Read local schedule state.
    next_id.text_read( tr, desc );
    next_public_event_id = next_id;

    // Read event count.
    event_count.text_read( tr, desc );

    // Read events.
    for( i = 0; i < event_count; i++ )
        {
        event_id.text_read( tr, desc );
        event.text_read( tr, desc );
        insert_event( event_id, event );
        }
    }


EventId LocalSchedule::generate_public_event_id()
    {
    EventId     id;
    hope( sizeof( id.components ) == sizeof( id.id ) );

    id.components.is_public   = 1;
    id.components.relative_id = next_public_event_id++;
    hope( id.components.relative_id != 0 ); // id overflow
    next_public_event_id++;
    return id;
    }


void LocalSchedule::insert_event
    (
    const EventId & eid,
    const Event &   e
    )
    {
    schedule.insert
        (
        EventSchedule::value_type
            (
            e.get_execution_time(),
            Item( eid, e )
            )
        );
    }


void LocalSchedule::clear()
    {
    next_public_event_id = 1;
    schedule.clear();
    schedule_delay = STime();
    }


} // namespace Massiv::Core
} // namespace Massiv
