/** @file local_schedule.h
 * Local schedule maintained by object.
 */

/*
 * begin        : Mon Apr 29 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_LOCAL_SCHEDULE_H
#define MASSIV_LOCAL_SCHEDULE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_EVENT_H
#include "object/event.h"
#endif

#ifndef MASSIV_EVENTID_H
#include "object/eventid.h"
#endif

#include <map>


namespace Massiv {
namespace Core {

namespace Test { class SerializerTest; }


/** @addtogroup object_management
 * @{
 */

/** @class LocalSchedule
 * Event schedule for an @c Object. The schedule is named "local" because
 * it manages events for one particular object only.
 * Local schedule cooperates with @c GlobalSchedule in this way:
 * Global schedule is a heap of Objects with scheduled events ordered
 * by time of their first events (@c get_time_of_first_event()).
 * GlobalSchedule allows ObjectManager to effectively find out object
 * with the "most oldest" event.
 * 
 * Schedule processing for given Object must sometimes be delayed
 * (for "busy" objects, for example).
 * To guarantee that event ordering for that Object does not change
 * all scheduled events need be delayed. For that purposes @c schedule_delay
 * is defined, which is "global" for the whole schedule.
 * If @c schedule_delay is non-zero all items stored in @c schedule
 * will be treated as if they were scheduled to their actual time
 * plus @c schedule_delay. @c Item::Event::execution_time is not
 * affected by @c schedule_delay. When @c schedule gets empty
 * @c schedule_delay is automatically reset to zero. @c schedule_delay
 * can also be reset by @c remove_first_event() which has this semantics:
 * If schedule is delayed because the object was "busy" in the past
 * then the schedule may be undelayed when the object is no longer "busy"
 * (the object is certainly not "busy" when its first event is processed).
 * 
 * GlobalSchedule must be manually notified about local schedule changes
 * (@c ObjectManager::notify_of_local_schedule_change()).
 *
 * CZ Note: Boovie: LocalSchedule nepouziva zadnou strukturu pro rychle
 *   vyhledavani podle EventId. Hledani dle EventId je vyzadovano
 *   pouze ve remove_event(), ktere se vola pri "killovani" public eventy,
 *   k cemuz predpokladam nebude dochazet casto, a kdyz uz, tak bude
 *   LocalSchedule obsahovat pouze maly pocet udalosti.
 *   Predpokladam, ze vetsina objektu nebudou mit vubec zadne eventy (pripadne
 *   jednu jedinou, kdyz se budou presouvat na jine misto), takze pridavat
 *   do kazdeho objektu dalsi datovou strukturu mne prijde zbytecne
 *   zvyseni pametove narocnosti.
 */
class LocalSchedule
    {
    friend class Massiv::Core::Test::SerializerTest;

public:

    /**
     * Constructor.
     */
    LocalSchedule();

    /**
     * Destructor.
     */
    ~LocalSchedule();

public:

    /**
     * Check whether the local schedule is empty.
     */
    bool is_empty() const;

    /**
     * Get schedule size (number of events scheduled in the schedule).
     */
    int get_size() const;

    /**
     * Insert new public event into the local schedule.
     * Returned event id uniquely identifies inserted event
     * in this local schedule.
     */
    EventId add_new_public_event
        (
        const Event & event
        );

    /**
     * Insert new private event into the local schedule.
     */
    void add_new_private_event
        (
        const Event & event
        );

    /**
     * Remove specified event from the local schedule.
     * Nothing is done if @a event_id is private, even if
     * the event with this id exists.
     * Returns true if the event was removed. In that case
     * @a event_out contains removed event. Returns false
     * if no event was removed. In that case @a event_out
     * stays unchanged.
     */
    bool remove_event
        (
        const EventId & event_id,
        Event &         event_out
        );

    /**
     * Remove first scheduled event from the local schedule.
     * If @c reset_delay is set to true schedule delay is reset to zero.
     * This is useful for delaying of events of "busy" objects.
     * When an object is detected that it is no longer "busy" (thus its
     * first event is processed/removed), all its events can be undelayed.
     * Prior to calling this function there must be at least one scheduled event!
     */
    void remove_first_event
        (
        bool reset_delay
        );

    /**
     * Get first scheduled event from the local schedule.
     * Returns true if the event was copied to @a event_out.
     * Return false if there is no event in the schedule.
     * In that case @a event_out stays unchanged.
     */
    bool get_first_event
        (
        Event &         event_out
        ) const;

    /**
     * Get time of first scheduled event.
     * Prior to calling this function there must be at least one scheduled event!
     */
    STime get_time_of_first_event() const;

    /**
     * Adds @c time to current @c schedule_delay and forces schedule
     * processing to delay all scheduled events by @c schedule_delay.
     * If an event was scheduled to time @c t then it will be processed
     * at @c t + @c schedule_delay.
     */
    void delay_schedule_processing
        (
        const STime &   time
        );

    /**
     * Tests if local schedule is delayed.
     */
    bool is_delayed() const;

    /**
     * Check event schedule of client-to-server migration.
     *
     * @exception ExSerializationError Thrown if error is detected.
     */
    void check_client_to_server_events
        (
        const NodeId &      client_node,
        const ObjectId &    object_id,
        const ObjectId &    destination
        ) const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     * Writes events separated by spaces.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

private:

    /**
     * Generate new public event id.
     */
    EventId generate_public_event_id();

    /**
     * Insert event into the local schedule.
     */
    void insert_event
        (
        const EventId & event_id,
        const Event &   event
        );

    /**
     * Clear content.
     */
    void clear();

private:

    /**
     * Schedule item.
     */
    class Item
        {
    public:

        /** Constructor. */
        Item()
            {
            }

        /** Constructor. */
        Item
            (
            const EventId & eid,
            const Event &   e
            ) : event( e ),
                event_id( eid )
            {
            }

        /** Copy constructor. */
        Item
            (
            const Item & i
            ) : event( i.event ),
                event_id( i.event_id )
            {
            }

        /** Destructor. */
        ~Item()
            {
            }

        /** Operator=. */
        Item & operator=
            (
            const Item & i
            )
            {
            event    = i.event;
            event_id = i.event_id;
            return *this;
            }

    public:

        Event       event;
            /**< Event. */

        EventId     event_id;
            /**< Event id. */

        }; // Item

    /** @typedef EventSchedule
     * Scheduled events sorted by time.
     */
    typedef std::multimap
                <
                STime,
                Item,
                STime::Less
                > EventSchedule;

private:

    EventSchedule   schedule;
        /**< Scheduled events sorted by time. */
    Int64           next_public_event_id;
        /**< Next available public event id (relative). */
    STime           schedule_delay;
        /**< Schedule delay.
             @note The delay is not serialized because it affects
             local processing only. */

    }; // LocalSchedule

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_LOCAL_SCHEDULE_H
