/** @file metaobject.cpp
 * Metaobject base class (implementation).
 */

/*
 * begin        : Mar 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/global.h"
#include "utility/debug.h"
#include "utility/wildcards.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "serializer/serializer.h"
#include "object/object.h"
#include "object/rpc.h"
#include "object/class_manager.h"
#include "object/object_factory.h"
#include "object/metaobject.h"
#include "pointer/pointer_to_object.h"
#include "object/lib/lib_exceptions_helper.h"
#include <sstream>
#include <ostream>


namespace Massiv {
namespace Core {


#ifdef OPTION_DEBUG

/** @def DEBUG_PROPERTY_OFFSETS
 * Define this to enable debugging of direct access to properties using property
 * offset. You should always enable this when porting Massiv to new compiler.
 */
#define DEBUG_PROPERTY_OFFSETS

#endif


// class MetaObject

MetaObject::MetaObject()
      : class_type_info( NULL ),
        class_pointer_type_info( NULL ),
        offsets_initialized( false )
    {
    }


MetaObject::~MetaObject()
    {
    }


void MetaObject::initialize_inherited_introspection()
    {
    // Initialize info about inherited properties.

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        const ComponentInfo & component_info = *component_it;
        if( component_info.is_complete_component() )
            {
            continue;
            }
        hope( component_info.metaobject );
        const MetaObject & component_metaobject = *component_info.metaobject;
        const PropertyInfos & component_properties = component_metaobject.get_property_infos();

        // Add all properties defined in the component.
        // Skip all properties inherited into the component.

        PropertyInfoIterator property_it;
        for
            (
            property_it = component_properties.begin();
            property_it != component_properties.end();
            ++property_it
            )
            {
            const PropertyInfo & property_info = *property_it;
            if( property_info.component->is_complete_component() )
                {
                // Offsets of inherited properties is left relative to component
                // they are defined in. Offsets will be adjusted to correct ones
                // in #initialize_offsets()

                property_infos.push_back
                    (
                    PropertyInfo
                        (
                        property_info.name,
                        property_info.type,
                        component_it,
                        property_info.attributes,
                        property_info.component_offset
                        )
                    );
                }
            }
        }
    }


void MetaObject::initialize_offsets
    (
    const VariantPointer & object
    ) const
    {
    if( offsets_initialized )
        {
        return;
        }

    #ifdef OPTION_DEBUG
    hope( class_pointer_type_info );
    hope( object.type );
    hope( *class_pointer_type_info == *object.type );
    #endif

    // Initialize component offsets using generated upcast().

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        VariantPointer upcasted = upcast( object, component_it ).first;
        component_it->offset = reinterpret_cast< const UInt8 * >( upcasted.pointer ) -
            reinterpret_cast< const UInt8 * >( object.pointer );
        }

    // Adjust offsets of inherited properties.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        property_it->global_offset =
            property_it->component_offset + property_it->component->offset;
        }

    // Done.

    offsets_initialized = true;
    }


// inheritance

MetaObject::UpcastResult MetaObject::upcast
    (
    const VariantPointer &      object,
    ComponentInfoIterator       component_it
    ) const
    {
    hope( component_it != component_infos.end() );
    const std::size_t index = component_it - component_infos.begin();
    hope( index < component_infos.size() );
    return upcast( object, index );
    }


MetaObject::UpcastResult MetaObject::upcast
    (
    const VariantPointer &      object,
    const std::string &         component_name
    ) const
    {
    ComponentInfoIterator it;
    for( it = component_infos.begin(); it != component_infos.end(); ++it )
        {
        const ComponentInfo & component = *it;
        if( component.name == component_name )
            {
            return upcast( object, it );
            }
        }
    return UpcastResult( VariantPointer(), -1 );
    }


bool MetaObject::inherits
    (
    const std::type_info & other_type_info
    ) const
    {
    ComponentInfoIterator it;
    for( it = component_infos.begin(); it != component_infos.end(); ++it )
        {
        const MetaObject * const metaobject = it->metaobject;
        hope( metaobject );
        if( metaobject->get_class_type_info() == other_type_info )
            {
            return true;
            }
        }
    return false;
    }


// properties

Property * MetaObject::get_property
    (
    const VariantPointer &          object,
    std::size_t                     property_index
    ) const
    {
    hope( property_index < property_infos.size() );
    return get_property( object, property_infos.begin() + property_index );
    }


Property * MetaObject::get_property
    (
    const VariantPointer &          object,
    PropertyInfos::const_iterator   property_it
    ) const
    {
    // Warning: Nasty pointer hackery.

    #ifdef OPTION_DEBUG

    std::size_t property_index = property_it - property_infos.begin();
    hope( property_index < property_infos.size() );
    hope( object.type );
    hope( *object.type == *class_pointer_type_info );

    #endif

    // Upcast object to component containing this property.
    // Can't use component offset magic because object may be descendant of this managed class.
    // Offset magic probably could be used for non-virtual components, but who cares.

    VariantPointer upcasted = upcast( object, property_it->component ).first;
    UInt8 * const component_beginning = reinterpret_cast< UInt8 * >( upcasted.pointer );
    UInt8 * const property = component_beginning + property_it->component_offset;
    return reinterpret_cast< Property * >( property );
    }


Property * MetaObject::get_property_from_complete_object
    (
    const VariantPointer &  complete_object,
    PropertyInfoIterator    property_it
    ) const
    {
    // Warning: Nasty pointer hackery.

    UInt8 * const object_address = reinterpret_cast< UInt8 * >( complete_object.pointer );
    UInt8 * const fast_address = object_address + property_it->global_offset;

    #ifdef OPTION_DEBUG

    std::size_t property_index = property_it - property_infos.begin();
    hope( property_index < property_infos.size() );
    hope( complete_object.type );
    hope( *complete_object.type == *class_pointer_type_info );

    #endif

    #ifdef DEBUG_PROPERTY_OFFSETS

    // Check if precached property offset is right.

    VariantPointer upcasted = upcast( complete_object, property_it->component ).first;
    UInt8 * const component_beginning = reinterpret_cast< UInt8 * >( upcasted.pointer );
    UInt8 * const real_address = component_beginning + property_it->component_offset;

    if( real_address != fast_address )
        {
        std::ostringstream oss;

        oss << "GWAH: Fast property addressing error: real address = " <<
            real_address << ", fast address = " << fast_address << "\n    " <<
            "real component offset" << ( component_beginning - object_address ) <<
            ", cached component offset" << property_it->component->offset << "\n    " <<
            "property-component offset = " << property_it->component_offset <<
            ", property-global offset = " << property_it->global_offset << std::endl;

        Global::log( STATUS_ERROR( Status::FACILITY_OBJECT_MANAGEMENT ), oss.str() );

        return reinterpret_cast< Property * >( real_address );
        }

    #endif

    return reinterpret_cast< Property * >( fast_address );
    }


Property * MetaObject::get_property
    (
    const VariantPointer &      object,
    const std::string &         property_name
    ) const
    {
    PropertyInfoIterator result = property_infos.end();
    PropertyInfoIterator it;
    for( it = property_infos.begin(); it != property_infos.end(); ++it )
        {
        const PropertyInfo & property = *it;
        if( property_name == property.name )
            {
            if( result != property_infos.end() )
                {
                // Ambiguous get. Can't be distinguished from "no such property".

                return NULL;
                }
            result = it;
            }
        }

    if( result != property_infos.end() )
        {
        return get_property( object, result );
        }

    return NULL;
    }


void MetaObject::find_property_infos
    (
    PropertyInfoIterators &     result,
    const std::string &         pattern
    ) const
    {
    PropertyInfoIterator it;
    for( it = property_infos.begin(); it != property_infos.end(); ++it )
        {
        const PropertyInfo & property = *it;
        if( Wildcards::match( pattern, property.name ) )
            {
            result.push_back( it );
            }
        }
    }


void MetaObject::find_properties
    (
    PropertyPointers &          result,
    const PtrToObject &         object,
    const std::string &         pattern
    ) const
    {
    hope( object != NULL );
    find_properties
        (
        result,
        get_complete_object( object ),
        pattern
        );
    }


void MetaObject::find_properties
    (
    PropertyPointers &          result,
    const VariantPointer &      object,
    const std::string &         pattern
    ) const
    {
    PropertyInfoIterator it;
    for( it = property_infos.begin(); it != property_infos.end(); ++it )
        {
        const PropertyInfo & property = *it;
        if( Wildcards::match( pattern, property.name ) )
            {
            result.push_back( get_property( object, it ) );
            }
        }
    }


const MetaObject::PropertyInfo * MetaObject::find_property_info
    (
    const PtrToObject &         object,
    const Property *            property
    ) const
    {
    hope( object != NULL );
    return find_property_info
        (
        get_complete_object( object ),
        property
        );
    }


const MetaObject::PropertyInfo * MetaObject::find_property_info
    (
    const VariantPointer &      object,
    const Property *            property
    ) const
    {
    PropertyInfoIterator it;
    for( it = property_infos.begin(); it != property_infos.end(); ++it )
        {
        if( property == get_property( object, it ) )
            {
            // This may seem ugly, but it's actually safe.

            const PropertyInfo & property_info = *it;
            return &property_info;
            }
        }

    return NULL;
    }


bool MetaObject::get_property_info
    (
    const ObjectProperty *      object,
    const Property *            owned_property,
    Property::PropertyInfo &    info
    ) const
    {
    const PropertyInfo * const metaobject_property_info =
        find_property_info
            (
            PtrToObject( const_cast<ObjectProperty *>( object ) ),
            owned_property
            );
    if( metaobject_property_info )
        {
        info.property_name = metaobject_property_info->name;
        return true;
        }
    return false;
    }


void MetaObject::for_each_property_do
    (
    Property::PropertyOperation &   operation,
    ObjectProperty *                object
    ) const
    {
    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Call for_each_property_do on each property. Dirty pointer hackery.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        Property * const property = get_property_from_complete_object
            ( complete_object, property_it );
        property->for_each_property_do( operation );
        }
    }


void MetaObject::initialize_property_hierarchy
    (
    ObjectProperty *    object,
    bool                all_properties
    ) const
    {
    // Specialized version of for_each_property_do()...

    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Initialize property owners.
    // Depth-first traversal originated at @c object.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        Property * const property = get_property_from_complete_object
            ( complete_object, property_it );

        // Skip inherited properties if requested.

        if( !all_properties && !property_it->component->is_complete_component() )
            {
            // HACK/FIXME:
            //  Complete component's properties are always at the beginning
            //  of the list.

            break;
            }

        // Property::initialize_from_owner() ensures proper recursive
        // initialization of its properties if needed.

        property->initialize_from_owner( *object, property_it->attributes );
        }
    }


// methods

void MetaObject::call_method
    (
    const VariantPointer &      object,
    std::size_t                 method_index,
    TextReader &                tr
    ) const
    {
    std::auto_ptr< MethodPacket > packet( create_method_packet( method_index ) );
    hope( packet.get() );
    packet->text_read_arguments( tr, Serializer::Description() );
    packet->call_method( object );
    }


void MetaObject::call_method
    (
    const VariantPointer &      object,
    MethodInfos::const_iterator method_it,
    TextReader &                tr
    ) const
    {
    hope( method_it != method_infos.end() );
    const std::size_t index = method_it - method_infos.begin();
    hope( index < method_infos.size() );
    call_method( object, index, tr );
    }


bool MetaObject::call_method
    (
    const VariantPointer &      object,
    const std::string &         method_name,
    TextReader &                tr
    ) const
    {
    MethodInfos::const_iterator it;
    for( it = method_infos.begin(); it != method_infos.end(); ++it )
        {
        const MethodInfo & method = *it;
        if( method.name == method_name )
            {
            call_method( object, it, tr );
            return true;
            }
        }
    return false;
    }


std::auto_ptr< MethodPacket > MetaObject::remote_call_method
    (
    Remote< Object >            object,
    std::size_t                 method_index,
    TextReader &                tr
    ) const
    {
    std::auto_ptr< MethodPacket > packet( create_method_packet( method_index ) );

    if( !packet.get() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            RemoteCallFailedException,
            "Invalid method index."
            );
        }

    packet->text_read_arguments( tr, Serializer::Description() );
    Pointer< ResultObject > result = RPCObject::create_and_send
        (
        object,
        STime(),
        *packet.get(),
        get_class_type_id(),
        method_index,
        true,
        STime(),
        true
        );
    result->wait_for_reply();

    return result->create_results();
    }


std::auto_ptr< MethodPacket > MetaObject::remote_call_method
    (
    Remote< Object >            object,
    MethodInfos::const_iterator method_it,
    TextReader &                tr
    ) const
    {
    hope( method_it != method_infos.end() );
    const std::size_t index = method_it - method_infos.begin();
    if( index >= method_infos.size() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            RemoteCallFailedException,
            "Invalid method iterator."
            );
        }
    return remote_call_method( object, index, tr );
    }


std::auto_ptr< MethodPacket > MetaObject::remote_call_method
    (
    Remote< Object >            object,
    const std::string &         method_name,
    TextReader &                tr
    ) const
    {
    // Check methods in each component.

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        // Check methods in *component_it.

        const MetaObject * const component_metaobject = component_it->metaobject;
        hope( component_metaobject );
        const MethodInfos & component_methods = component_metaobject->get_method_infos();
        MethodInfoIterator method_it;
        for
            (
            method_it = component_methods.begin();
            method_it != component_methods.end();
            ++method_it
            )
            {
            if( method_it->name == method_name )
                {
                return component_metaobject->remote_call_method( object, method_it, tr );
                }
            }
        }

    RAISE_CORE_LIB_EXCEPTION
        (
        RemoteCallFailedException,
        "Invalid method name."
        );
    /*NOTREACHED*/
    }


void MetaObject::find_method_infos
    (
    MethodInfoIterators &       result,
    const std::string &         pattern
    ) const
    {
    // Check methods in each component.

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        // Check methods in *component_it.

        const MetaObject * const component_metaobject = component_it->metaobject;
        hope( component_metaobject );
        const MethodInfos & component_methods = component_metaobject->get_method_infos();
        MethodInfoIterator method_it;
        for
            (
            method_it = component_methods.begin();
            method_it != component_methods.end();
            ++method_it
            )
            {
            const MethodInfo & method = *method_it;
            if( Wildcards::match( pattern, method.name ) )
                {
                result.push_back( method_it );
                }
            }
        }
    }


void MetaObject::find_methods
    (
    MemberContexts &            result,
    const VariantPointer &      object,
    const std::string &         pattern
    ) const
    {
    // Check methods in each component.

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        // Check methods in *component_it.

        const MetaObject * const component_metaobject = component_it->metaobject;
        hope( component_metaobject );
        const MethodInfos & component_methods = component_metaobject->get_method_infos();
        MethodInfoIterator method_it;
        for
            (
            method_it = component_methods.begin();
            method_it != component_methods.end();
            ++method_it
            )
            {
            const MethodInfo & method = *method_it;
            if( Wildcards::match( pattern, method.name ) )
                {
                MemberContext context;
                context.object = upcast( object, component_it ).first;
                context.metaobject = component_metaobject;
                context.method_it = method_it;
                result.push_back( context );
                }
            }
        }
    }


int MetaObject::call_methods
    (
    const PtrToObject &         object,
    const std::string &         pattern,
    TextReader &                tr
    ) const
    {
    hope( object != NULL );
    return call_methods
        (
        get_complete_object( object ),
        pattern,
        tr
        );
    }


int MetaObject::call_methods
    (
    const VariantPointer &      object,
    const std::string &         pattern,
    TextReader &                tr
    ) const
    {
    int count = 0;

    // Check methods in each component.

    ComponentInfoIterator component_it;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        // Check methods in *component_it.

        const MetaObject * const component_metaobject = component_it->metaobject;
        hope( component_metaobject );
        const MethodInfos & component_methods = component_metaobject->get_method_infos();
        MethodInfoIterator method_it;
        for
            (
            method_it = component_methods.begin();
            method_it != component_methods.end();
            ++method_it
            )
            {
            const MethodInfo & method = *method_it;
            if( Wildcards::match( pattern, method.name ) )
                {
                component_metaobject->call_method
                    (
                    upcast( object, component_it ).first,
                    method_it,
                    tr
                    );
                count++;
                }
            }
        }

    return count;
    }


void MetaObject::call_method
    (
    const MemberContext &   context,
    TextReader &            tr
    )
    {
    hope( context.metaobject );
    hope( !context.object.is_null() );
    context.metaobject->call_method( context.object, context.method_it, tr );
    }


// serialization

void MetaObject::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc,
    const ObjectProperty *          object
    ) const
    {
    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Get object address.

    const UInt8 * const object_address = reinterpret_cast< const UInt8 * >( complete_object.pointer );

    // Write each property to bitstream.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        // Compute property pointer.

        const UInt8 * const property_address = object_address + property_it->global_offset;
        const Property * const property = reinterpret_cast< const Property * >( property_address );

        // Write property if it should be serialized.

        if( property->should_serialize( desc ) == Property::SSR_YES )
            {
            bos.write_int( 1, 1 );
            property->bitstr_write( bos, desc );
            }
        else
            {
            bos.write_int( 0, 1 );
            }
        }
    }


void MetaObject::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc,
    ObjectProperty *                object
    ) const
    {
    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Get object address.

    UInt8 * const object_address = reinterpret_cast< UInt8 * >( complete_object.pointer );

    // Read each property from bitstream.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        // Compute property pointer.

        UInt8 * const property_address = object_address + property_it->global_offset;
        Property * const property = reinterpret_cast< Property * >( property_address );

        // Read property.

        if( bis.read_int( 1 ) )
            {
            #ifdef OPTION_DEBUG
            try
                {
            #endif

                property->bitstr_read( bis, desc );

            #ifdef OPTION_DEBUG
                }
            catch( Exception & )
                {
                Global::log_warning
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_HIGHEST,
                    "Error while reading content of " +
                    property->get_info().property_name + "."
                    );
                throw;
                }
            #endif
            }
        }
    }


void MetaObject::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc,
    const ObjectProperty *          object
    ) const
    {
    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Get object address.

    const UInt8 * const object_address = reinterpret_cast< const UInt8 * >( complete_object.pointer );

    // Write intro.

    tw.line() << "object # " << class_name;
    tw.end_line();

    ComponentInfoIterator component_it = component_infos.end();
    const int indent = tw.change_indent( 4 );

    // Write each property to bitstream.

    PropertyInfoIterator property_it;
    for
        (
        property_it = property_infos.begin();
        property_it != property_infos.end();
        ++property_it
        )
        {
        // Compute property pointer.

        const UInt8 * const property_address = object_address + property_it->global_offset;
        const Property * const property = reinterpret_cast< const Property * >( property_address );

        // Skip properties that should not be serialized.

        if( property->should_serialize( desc ) != Property::SSR_YES )
            {
            continue;
            }

        // Change component?

        if( property_it->component != component_it )
            {
            tw.set_indent( indent - 2 );
            component_it = property_it->component;
            if( component_it->is_complete_component() )
                {
                tw.line() << ":m:";
                }
            else if( component_it->virtual_component )
                {
                hope( component_it->metaobject );
                tw.line() << ":v:" << component_it->metaobject->get_class_name();
                }
            else
                {
                tw.line() << "::" << component_it->name;
                }
            tw.end_line();
            tw.set_indent( indent );
            }

        // Write property.

        tw.line() << property_it->name << " ";
        tw.change_indent( 4 );
        property->text_write( tw, desc );
        tw.change_indent( -4 );
        tw.end_line();
        }

    // Done.

    tw.set_indent( indent - 4 );
    tw.line() << "@endobject";
    }


void MetaObject::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc,
    ObjectProperty *                object
    ) const
    {
    hope( object );
    hope( object->metaobject == this );
    hope( typeid( *object ) == *class_type_info );

    // Initialize offsets.

    VariantPointer complete_object = get_complete_object( object );
    initialize_offsets( complete_object );

    // Get object address.

    UInt8 * const object_address = reinterpret_cast< UInt8 * >( complete_object.pointer );

    // Beginning.

    tr.require_string( "object" );

    // Read lines.

    ComponentInfoIterator current_component = component_infos.end();
    std::string token;
    while( text_read_line( tr, token ) )
        {
        // Is token component name?

        if( ( token.size() >= 3 ) && ( token[ 0 ] == ':' ) )
            {
            // Change component.

            current_component = component_infos.end();
            if( token == ":m:" )
                {
                // Complete component.

                current_component = component_infos.begin();
                hope( current_component->is_complete_component() );
                }
            else if( ( token[ 1 ] == 'v' ) && ( token[ 2 ] == ':' ) )
                {
                // Virtual component.

                std::string name = token.substr( 3 );
                ComponentInfoIterator it;
                for( it = component_infos.begin(); it != component_infos.end(); ++it )
                    {
                    if( it->virtual_component )
                        {
                        hope( it->metaobject );
                        if( name == it->metaobject->get_class_name() )
                            {
                            current_component = it;
                            break;
                            }
                        }
                    }
                }
            else if( token[ 1 ] == ':' )
                {
                // Nonvirtual component.

                std::string name = token.substr( 2 );
                ComponentInfoIterator it;
                for( it = component_infos.begin(); it != component_infos.end(); ++it )
                    {
                    if( !it->virtual_component )
                        {
                        if( name == it->name )
                            {
                            current_component = it;
                            break;
                            }
                        }
                    }
                }

            if( current_component == component_infos.end() )
                {
                throw ExUnexpectedValue( tr.unexpected_value( "component name", token ) );
                }
            }
        else
            {
            // Not a component name. Should be property value.

            if( current_component == component_infos.end() )
                {
                throw ExUnexpectedValue( tr.unexpected_value( "component name", token ) );
                }

            PropertyInfoIterator it;
            for( it = property_infos.begin(); it != property_infos.end(); ++it )
                {
                if( ( it->component == current_component ) && ( it->name == token ) )
                    {
                    // Compute property pointer.

                    UInt8 * const property_address = object_address + it->global_offset;
                    Property * const property = reinterpret_cast< Property * >( property_address );

                    // Read property.

                    #ifdef OPTION_DEBUG
                    try
                        {
                    #endif

                        property->text_read( tr, desc );

                    #ifdef OPTION_DEBUG
                        }
                    catch( Exception & )
                        {
                        Global::log_warning
                            (
                            Status::FACILITY_OBJECT_MANAGEMENT,
                            Status::PRIORITY_HIGHEST,
                            "Error while reading content of " +
                            property->get_info().property_name + "."
                            );
                        throw;
                        }
                    #endif

                    break;
                    }
                }

            // Unknown property name?

            if( it == property_infos.end() )
                {
                std::string expected = "component ";
                if( current_component->is_complete_component() )
                    {
                    expected += ":m:";
                    }
                else
                    {
                    if( current_component->virtual_component )
                        {
                        hope( current_component->metaobject );
                        expected += ":v:";
                        expected += current_component->metaobject->get_class_name();
                        }
                    else
                        {
                        expected += "::";
                        expected += current_component->name;
                        }
                    }
                expected += " property name";
                throw ExUnexpectedValue( tr.unexpected_value( expected, token ) );
                }
            }
        }

    // Done. "@endobject" eaten by text_read_line().
    }


// other object operations

bool MetaObject::compare_objects
   (
   const ObjectProperty *,
   const ObjectProperty *
   ) const
   {
   std::string error = get_class_name();
   error += " is not comparable.";
   panic( error.c_str() );
   /*NOTREACHED*/
   }


void MetaObject::copy_objects
   (
   ObjectProperty *,
   const ObjectProperty *
   ) const
   {
   std::string error = get_class_name();
   error += " is not assignable.";
   panic( error.c_str() );
   }


// Helper methods.

VariantPointer MetaObject::get_complete_object
    (
    const ObjectProperty * object
    ) const
    {
    return downcast( const_cast<ObjectProperty *>( object ) );
    }


void MetaObject::text_read_object
    (
    TextReader & tr
    ) const
    {
    tr.require_string( "object" );
    }


bool MetaObject::text_read_line
    (
    TextReader &    tr,
    std::string &   result
    ) const
    {
    tr.next_line_throw();

    std::istream & is = tr.line();
    //hope( is );
    is >> result;
    if( is.fail() )
        {
        if( is.eof() )
            {
            throw ExUnexpectedEol( tr.unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( tr.unexpected_token( "string" ) );
            }
        }

    if( result == "@endobject" )
        {
        return false;
        }
    return true;
    }


void MetaObject::panic_upcast_index()
    {
    panic( "Upcast component index out of range." );
    }


// class MetaObject::ComponentInfo

MetaObject::ComponentInfo::ComponentInfo
    (
    const std::type_info &  component_type_info,
    bool                    the_virtual_component,
    int                     the_inherited_from,
    const std::string &     the_name
    ) : metaobject( NULL ),
        virtual_component( the_virtual_component ),
        inherited_from( the_inherited_from ),
        name( the_name ),
        offset( 0 )
    {
    // Find component's metaobject.

    ClassManager & class_manager = Global::class_manager();
    metaobject = class_manager.get_metaobject( component_type_info );
    }


std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::ComponentInfo &   info
    )
    {
    if( info.virtual_component )
        {
        stream << "virtual ";
        }
    if( info.is_complete_component() )
        {
        hope( info.metaobject );
        stream << info.metaobject->get_class_name();
        }
    else
        {
        stream << info.name;
        }
    return stream;
    }


// class MetaObject::DataInfo

MetaObject::DataInfo::DataInfo
    (
    const std::string &     the_name,
    const std::string &     the_type
    ) : name( the_name ),
        type( the_type )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                  stream,
    const MetaObject::DataInfo &    info
    )
    {
    stream << info.type << " " << info.name;
    return stream;
    }


// class MetaObject::PropertyInfo

MetaObject::PropertyInfo::PropertyInfo
    (
    const std::string &             the_name,
    const std::string &             the_type,
    ComponentInfoIterator           the_component,
    const Property::Attributes &    the_attributes,
    int                             the_component_offset
    ) : Super( the_name, the_type ),
        component( the_component ),
        attributes( the_attributes ),
        component_offset( the_component_offset ),
        global_offset( 0 )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::PropertyInfo &    info
    )
    {
    stream << info.type << " ";
    if( !info.component->is_complete_component() )
        {
        stream << info.component->name << "/";
        }
    stream << info.name;
    return stream;
    }


// class MetaObject::ArgumentInfo

MetaObject::ArgumentInfo::ArgumentInfo
    (
    const std::string &     the_name,
    const std::string &     the_type,
    Kind                    the_kind
    ) : Super( the_name, the_type ),
        kind( the_kind )
    {
    }


std::string MetaObject::ArgumentInfo::kind_string() const
    {
    switch( kind )
        {
    case ARGUMENT_IN:
        return "in";
    case ARGUMENT_OUT:
        return "out";
    case ARGUMENT_INOUT:
        return "inout";
        }
    return "unknown_argument_kind";
    }


std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::ArgumentInfo &    info
    )
    {
    stream << info.kind_string() << " " << info.type << " " << info.name;
    return stream;
    }


// class MetaObject::MethodInfo

MetaObject::MethodInfo::MethodInfo
    (
    const std::string &     the_name,
    const std::string &     the_result_type
    ) : name( the_name ),
        result_type( the_result_type )
    {
    }


std::ostream & operator<<
    (
    std::ostream &                              stream,
    const MetaObject::MethodInfo::Attributes &  attributes
    )
    {
    if( attributes.a_virtual || attributes.a_const )
        {
        stream << "< ";
        if( attributes.a_virtual )
            {
            stream << "virtual";
            }
        if( attributes.a_const )
            {
            if( attributes.a_virtual )
                {
                stream << ", ";
                }
            stream << "const";
            }
        stream << " >";
        }
    return stream;
    }


std::ostream & operator<<
    (
    std::ostream &                  stream,
    const MetaObject::MethodInfo &  info
    )
    {
    if( info.result_type.size() )
        {
        stream << info.result_type << " ";
        }

    stream << info.name << "(";

    MetaObject::ArgumentInfoIterator it;
    int argument_index = 0;
    for( it = info.arguments.begin(); it != info.arguments.end(); ++it, argument_index++ )
        {
        const MetaObject::ArgumentInfo & argument = *it;
        if( argument_index )
            {
            stream << ",";
            }
        stream << " " << argument;
        }

    if( argument_index )
        {
        stream << " ";
        }
    stream << ")";

    return stream;
    }


// global

std::ostream & operator<<
    (
    std::ostream &                  stream,
    const MetaObject::Attributes &  attributes
    )
    {
    bool first = true;

    #define ATTR if( first ) { first = false; stream << "< "; } else { stream << ", "; }
    #define BOOLATTR( n ) if( attributes.n ) { ATTR; stream << #n; }
    #define INTATTR( n ) ATTR; stream << #n " = " << attributes.n;

    BOOLATTR( abstract );
    BOOLATTR( assignable );
    BOOLATTR( client_server_migrate );
    BOOLATTR( comparable );
    BOOLATTR( has_strong_pointer );
    INTATTR( kind );
    BOOLATTR( no_archive );
    BOOLATTR( no_balance );
    BOOLATTR( no_migrate );
    BOOLATTR( node_object );
    BOOLATTR( root );
    BOOLATTR( simulation_startup_notify );
    BOOLATTR( throwable );
    BOOLATTR( tracked );
    BOOLATTR( value_type );

    #undef INTATTR
    #undef BOOLATTR
    #undef ATTR

    if( !first )
        {
        stream << " >";
        }
    return stream;
    }


std::ostream & operator<<
    (
    std::ostream &      stream,
    const MetaObject &  metaobject
    )
    {
    const MetaObject::ComponentInfos &  component_infos = metaobject.get_component_infos();
    const MetaObject::PropertyInfos &   property_infos = metaobject.get_property_infos();
    const MetaObject::MethodInfos &     method_infos = metaobject.get_method_infos();
    const std::string                   indent = "    ";

    stream << "class" << metaobject.get_attributes() << "\n";
    stream << indent << metaobject.get_class_name();

    // Write direct parents.

    MetaObject::ComponentInfoIterator component_it;
    bool first_parent = true;
    for
        (
        component_it = component_infos.begin();
        component_it != component_infos.end();
        ++component_it
        )
        {
        if( component_it->is_parent() )
            {
            if( first_parent )
                {
                stream << " : ";
                first_parent = false;
                }
            else
                {
                stream << ", ";
                }
            stream << *component_it;
            }
        }
    stream << "\n" << indent << "{\n";
    bool has_methods = false;

    // Methods.

    if( method_infos.size() )
        {
        MetaObject::MethodInfoIterator it;
        for( it = method_infos.begin(); it != method_infos.end(); ++it )
            {
            if( !has_methods )
                {
                has_methods = true;
                stream << indent << "// Methods.\n\n";
                }
            stream << indent << "method" << it->attributes << " " << *it << ";\n";
            }
        }

    // Inherited methods.

    component_it = component_infos.begin();
    hope( component_it != component_infos.end() );
    component_it++; // skip the most derived class component

    for
        (
        /* component_it now points to the second component */ ;
        component_it != component_infos.end();
        ++component_it
        )
        {
        const MetaObject * const component_metaobject = component_it->metaobject;
        hope( component_metaobject );
        const MetaObject::MethodInfos & component_methods =
                    component_metaobject->get_method_infos();
        MetaObject::MethodInfoIterator it;
        for(
            it = component_methods.begin();
            it != component_methods.end();
            ++it
            )
            {
            if( !has_methods )
                {
                has_methods = true;
                stream << indent << "// Methods.\n";
                }
            if( it == component_methods.begin() )
                {
                stream << "\n" << indent << "// In " << *component_it << ":\n";
                }
            stream << indent << "//  method" << it->attributes << " " << *it << ";\n";
            }
        }

    // Properties (including inherited properties).

    if( property_infos.size() )
        {
        if( has_methods )
            {
            stream << "\n";
            }
        stream << indent << "// Properties.\n";

        MetaObject::PropertyInfoIterator it;
        component_it = component_infos.end(); // current component.
        for( it = property_infos.begin(); it != property_infos.end(); ++it )
            {
            if( it->component != component_it )
                {
                component_it = it->component;
                stream << "\n";
                if( component_it != component_infos.begin() )
                    {
                    stream << indent << "// In " << *component_it << ":\n";
                    }
                }
            stream << indent;
            if( it->component != component_infos.begin() )
                {
                stream << "//  ";
                }
            stream << "property" << it->attributes << " " << *it << ";\n";
            }
        }

    stream << indent << "}";
    return stream;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "object/metaobject_inline.h"
#endif
