/** @file metaobject.h
 * Metaobject base class.
 */

/*
 * begin        : Mar 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_METAOBJECT_H
#define MASSIV_METAOBJECT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif


#include <typeinfo>
#include <string>
#include <vector>
#include <memory>
#include <iosfwd>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Metaobject base class.
 *
 * Metaobjects are used for introspection and serialization of managed objects
 * (classes which inherit Object). This class only defines the interface and
 * implements common methods. The rest is generated by factgen.pl and
 * classlist.pl.
 *
 * @note Whenever "this class" is used in documentation of MetaObject,
 *       it means "class described by this metaobject."
 *
 * @note All the pointers to properties/objects MetaObject operates upon
 *       (VariantPointers, Property *, ObjectProperty *) are aliasing
 *       short-lived pointers. Client must not store them into its classes
 *       and must ensure that core will not run meanwhile. The pointers
 *       might be invalidated otherwise.
 *       If this is too restrictive, one can use @c PtrToObject to pin
 *       the corresponding object.
 */
class MASSIV_EXPORTED MetaObject
    {
    NO_COPY_CONSTRUCTOR( MetaObject );
    NO_ASSIGNMENT_OPERATOR( MetaObject );

    /// Calls set_class_type_id()
    friend class ClassManager;

    //@{
    /// Object / ObjectProperty uses serializers and #initialize_property_attributes().
    friend class ObjectProperty;
    friend class Object;
    //@}

public:

    /**
     * Class attributes.
     */
    class MASSIV_EXPORTED Attributes
        {
    public:

        /** Default constructor. */
        INLINE Attributes();

    public:
        // Note: Keep attributes sorted by name.
        //       Update operator<< (near end-of-file) when adding attributes.

        bool        abstract;
            /**< True if objects of this class can't be instantiated. */

        bool        assignable;
            /**< True if the class implements operator=. */

        bool        client_server_migrate;
            /**< Can client migrate objects of this type to servers? */

        bool        comparable;
            /**< True if the class implements operator==. */

        bool        has_strong_pointer;
            /**< True iff the class contains strong pointers. */

        int         kind;
            /**< Class kind. See #ClassKind in src/core/object/class_type_id.h.
                 (inherited attribute) */

        bool        no_archive;
            /**< True if objects of the class should not be archived.
                 (inherited attribute) */

        bool        no_balance;
            /**< If true Balancer will not make any statistics regarding
                 data flow from objects of this class. */

        bool        no_migrate;
            /**< If true ObjectManager will refuse to migrate objects of this class.
                 (inherited attribute) */

        bool        node_object;
            /**< Is this class node object?
                 (inherited attribute) */

        bool        root;
            /**< True if objects of the class are GC roots.
                 (inherited attribute) */

        bool        simulation_startup_notify;
            /**< True if objects of this class should be notified
                 when simulation is restored from an archive. */

        bool        throwable;
            /**< Is this class inherited from Throwable?
                 (inherited attribute) */

        bool        tracked;
            /**< True if objects of the class are tracked by their object provider.
                 Useful when remote pointers to the objects are used often. */

        bool        value_type;
            /**< Is this class inherited from ValueType?
                 (inherited attribute) */

        }; // class Attributes

public:

    /**
     * Information about class "component" - class itself or one of classes
     * the class inherits.
     */
    class MASSIV_EXPORTED ComponentInfo;

    /**
     * Vector of component descriptions.
     */
    typedef std::vector< ComponentInfo > ComponentInfos;

    /**
     * Iterator pointing to component info.
     */
    typedef ComponentInfos::const_iterator ComponentInfoIterator;

    /**
     * Result of upcast methods.
     *
     * First element of the pair is pointer - result of the cast.
     * Second element is index of the component (index into vector of
     * components returned by #get_component_infos() of metaobject
     * which has been used to perform the cast).
     *
     * If upcast has failed, the pointer will be NULL and the index
     * will be -1.
     */
    typedef std::pair< VariantPointer, int > UpcastResult;

public:

    /**
     * Common base class for PropertyInfo a ArgumentInfo.
     */
    class MASSIV_EXPORTED DataInfo;

public:

    /**
     * Information about property defined in this component or inherited.
     */
    class MASSIV_EXPORTED PropertyInfo;

    /**
     * Vector of property descriptions.
     */
    typedef std::vector< PropertyInfo > PropertyInfos;

    /**
     * Vector of iterators pointing to property descriptions.
     */
    typedef std::vector< PropertyInfos::const_iterator > PropertyInfoIterators;

    /**
     * Iterator pointing to property description.
     */
    typedef PropertyInfos::const_iterator PropertyInfoIterator;

    /**
     * Vector of property pointers.
     */
    typedef std::vector< Property * > PropertyPointers;

public:

    /**
     * Information about method argument.
     */
    class MASSIV_EXPORTED ArgumentInfo;

    /**
     * Vector of argument infos.
     */
    typedef std::vector< ArgumentInfo > ArgumentInfos;

    /**
     * Iterator pointing to argument info.
     */
    typedef ArgumentInfos::const_iterator ArgumentInfoIterator;

public:

    /**
     * Information about method.
     */
    class MASSIV_EXPORTED MethodInfo;
    
    /**
     * Vector of method descriptions.
     */
    typedef std::vector< MethodInfo > MethodInfos;

    /**
     * Vector of iterators pointing to method descriptions.
     */
    typedef std::vector< MethodInfos::const_iterator > MethodInfoIterators;

    /**
     * Iterator pointing to a method description.
     */
    typedef MethodInfos::const_iterator MethodInfoIterator;

public:
    
    /**
     * Class member context information.
     *
     * @todo Get rid of this obscurity, currently needed for method introspection.
     */
    class MASSIV_EXPORTED MemberContext;
    
    /**
     * Vector of member contexts.
     */
    typedef std::vector< MemberContext > MemberContexts;

public:

    /**
     * Default constructor.
     */
    MetaObject();

    /**
     * Destructor.
     */
    virtual ~MetaObject();

// Initialization.
protected:

    /**
     * Initialize class introspection structures:
     * #component_infos, #property_infos and #method_infos.
     *
     * Only properties and methods directly defined in the class should be
     * added into #property_infos and #method_infos by this method.
     *
     * Called by ClassManager after all classes have been registered -
     * all metaobjects with basic class info will be available.
     */
    virtual void initialize_introspection() = 0;

    /**
     * Initialize introspection structures - inherited info.
     *
     * Adds information about inherited properties to #property_infos.
     *
     * Called by ClassManager after #initialize_introspection() has
     * been called for each registered class.
     */
    void initialize_inherited_introspection();

    /**
     * Initialize offset hacks in #component_infos and #property_infos.
     *
     * @param object Pointer to object of this class.
     *
     * @note Called by MetaObject when needed when user calls first method
     *       which requires the offsets. We could initialize the offsets
     *       sooner, but we would have to create an object of the managed
     *       class, which could have nasty side-effects.
     *
     *       Unfortunately this method will be usually called from @c const
     *       methods, so it's @c const too, even though it modified offset
     *       fields of component and property infos. That's why those fields
     *       are mutable, but "only for #initialize_offsets()".
     */
    void initialize_offsets
        (
        const VariantPointer & object
        ) const;

// Information about the class.
public:

    /**
     * Get class name.
     */
    INLINE const std::string & get_class_name() const;

    /**
     * Get standard C++ type information about the class.
     *
     * Returns <tt>typeid< class ></tt>.
     */
    INLINE const std::type_info & get_class_type_info() const;

    /**
     * Get standard C++ type information about pointer to the class.
     *
     * Returns <tt>typeid< class * ></tt>.
     */
    INLINE const std::type_info & get_class_pointer_type_info() const;

    /**
     * Get unique class type id.
     */
    INLINE const ClassTypeId & get_class_type_id() const;

    /**
     * Get object factory for this class.
     */
    INLINE const ObjectFactory * get_object_factory() const;

    /**
     * Convert <tt>ObjectProperty *</tt> to <tt>X *</tt>.
     *
     * This method may be called only if <tt>typeid( *object ) == typeid( X )</tt>.
     * Use #upcast() to get pointers to parents of @c X. This is required to
     * prevent problems with ambiguous casts when complex inheritance hierarchy
     * is used.
     */
    virtual VariantPointer downcast
        (
        ObjectProperty * object
        ) const = 0;
    
    /**
     * Equivalent to downcast().
     * @bug One of the variants should be removed, probably @c downcast().
     */
    VariantPointer get_complete_object
        (
        const ObjectProperty * object
        ) const;

// Class attributes.
public:

    /** Get structure with class attributes. */
    INLINE const Attributes & get_attributes() const;

    /** Is class abstract? */
    INLINE bool get_attribute_abstract() const;

    /** Is class assignable? */
    INLINE bool get_attribute_assignable() const;

    /** Can client migrate objects of this class to servers? */
    INLINE bool get_attribute_client_server_migrate() const;

    /** Is class comparable? */
    INLINE bool get_attribute_comparable() const;

    /** Does the class contain strong pointers? */
    INLINE bool get_attribute_has_strong_pointer() const;

    /** Get kind of this class. (iherited attribute) */
    INLINE int get_attribute_kind() const;
    
    /** Are objects of this class non-archivable? (inherited attribute) */
    INLINE bool get_attribute_no_archive() const;

    /** Get no-balance class attribute (inherited attribute)? */
    INLINE bool get_attribute_no_balance() const;
    
    /** Get no-migrate class attribute (inherited attribute)? */
    INLINE bool get_attribute_no_migrate() const;

    /** Is this node object class? (inherited attribute) */
    INLINE bool get_attribute_node_object() const;

    /** Is class GC root? (inherited attribute) */
    INLINE bool get_attribute_root() const;

    /** Notify objects of this class when simulation is restored from an archive? */
    INLINE bool get_attribute_simulation_startup_notify() const;

    /** Is this throwable class? (inherited attribute) */
    INLINE bool get_attribute_throwable() const;

    /** Are objects of this class tracked by provider? */
    INLINE bool get_attribute_tracked() const;

    /** Is this value type class? (inherited attribute) */
    INLINE bool get_attribute_value_type() const;

// Information about inheritance.
public:

    /**
     * Get information about class "components".
     *
     * The vector contains information about the managed class itself
     * (first element) and all classes it inherits. If a class is inherited
     * multiple times, it will be included in the vector multiple times.
     */
    INLINE const ComponentInfos & get_component_infos() const;

    /**
     * Upcast from pointer to this class to pointer to one of its components.
     *
     * @param object
     *      Pointer of type "pointer to the managed class", as returned by
     *      #downcast() or another #upcast().
     * @param component_index
     *      Valid index into component list returned by #get_component_infos().
     */
    virtual UpcastResult upcast
        (
        const VariantPointer &      object,
        std::size_t                 component_index
        ) const = 0;

    /**
     * Upcast from pointer to this class to pointer to one of its components.
     *
     * @param object
     *      Pointer of type "pointer to the managed class".
     * @param component_it
     *      Valid itarator into component list returned by #get_component_infos().
     */
    UpcastResult upcast
        (
        const VariantPointer &      object,
        ComponentInfoIterator       component_it
        ) const;

    /**
     * Upcast from pointer to this class to pointer to one of its components.
     *
     * @param object
     *      Pointer of type "pointer to the managed class".
     * @param component_name
     *      Component name.
     *
     * @return Result wukk contain NULL pointer if component of given name does not exist.
     *
     * @todo Allow unambiguous component name suffixes to be used instead of fully
     *       qualified component names.
     */
    UpcastResult upcast
        (
        const VariantPointer &      object,
        const std::string &         component_name
        ) const;

    /**
     * Checks whether managed class inherits another class described by its type info.
     */
    bool inherits
        (
        const std::type_info & other_type_info
        ) const;

// Information about properties.
public:

    /**
     * Get information about properties.
     *
     * The list contains descriptions of properties both defined in the managed
     * class and inherited.
     */
    INLINE const PropertyInfos & get_property_infos() const;

    /**
     * Get pointer to specified property of local object.
     *
     * @param object
     *      Pointer to the managed class, as returned by #downcast() or #upcast().
     * @param property_index
     *      Valid index into property list returned by #get_property_infos().
     */
    Property * get_property
        (
        const VariantPointer &      object,
        std::size_t                 property_index
        ) const;

    /**
     * Get pointer to specified property of local object.
     *
     * @param object
     *      Pointer to the managed class, as returned by #downcast() or #upcast().
     * @param property_it
     *      Valid iterator into property list returned by #get_property_infos().
     */
    Property * get_property
        (
        const VariantPointer &      object,
        PropertyInfoIterator        property_it
        ) const;
    
    /**
     * Get pointer to specified property of local object.
     *
     * @param complete_object
     *      Pointer to the most derived component of a managed class.
     * @param property_it
     *      Valid iterator into property list returned by #get_property_infos().
     *
     * The difference from the similar @c get_property() is that @c complete_object
     * must point to a complete object (its most derived component) whereas
     * @c object can point to any object's component.
     */
    Property * get_property_from_complete_object
        (
        const VariantPointer &      complete_object,
        PropertyInfoIterator        property_it
        ) const;

    /**
     * Get pointer to specified property of local object.
     *
     * @param object
     *      Pointer to the managed class, as returned by #downcast() or #upcast().
     * @param property_name
     *      Name of property.
     *
     * @return NULL if property with such name is not member of the class or if
     *         multiple properties with such name exist. Use #find_properties()
     *         to get multiple property pointers.
     */
    Property * get_property
        (
        const VariantPointer &      object,
        const std::string &         property_name
        ) const;

    /**
     * Get information about all properties matching given wildcard pattern.
     * If single property is inherited multiple times, pointer to its
     * description will be stored in @a result multiple times.
     *
     * @param result
     *      Push iterators pointing to found property infos to this vector.
     * @param pattern
     *      Wildcard pattern.
     */
    void find_property_infos
        (
        PropertyInfoIterators &     result,
        const std::string &         pattern
        ) const;

    /**
     * Get pointers to all properties matching given wildcard pattern.
     */
    void find_properties
        (
        PropertyPointers &          result,
        const PtrToObject &         object,
        const std::string &         pattern
        ) const;

    /**
     * Get pointers to all properties matching given wildcard pattern.
     */
    void find_properties
        (
        PropertyPointers &          result,
        const VariantPointer &      object,
        const std::string &         pattern
        ) const;

    /**
     * Find information about property given pointer to the property
     * and to the owning object.
     *
     * @return Pointer to property info if found or NULL.
     *
     * @note Returning pointer may look nasty but it's actually safe because
     *       PropertyInfo vectors are "constant" (never changed since initialized).
     */
    const PropertyInfo * find_property_info
        (
        const PtrToObject &         object,
        const Property *            property
        ) const;

    /**
     * Find information about property given pointer to the property
     * and to the owning object.
     */
    const PropertyInfo * find_property_info
        (
        const VariantPointer &      object,
        const Property *            property
        ) const;

    /**
     * Fill name of property @a owned_property that should
     * be owned by @c object.
     * Returns true if the factory "knows" about the property.
     *
     * @bug Obsolete method, #find_property_info() should be used instead.
     */
    bool get_property_info
        (
        const ObjectProperty *      object,
        const Property *            owned_property,
        Property::PropertyInfo &    info
        ) const;
    
    /**
     * Performs selected operation on every property owned by the object
     * @a object.
     *
     * @note @a object must be exactly of type managed by this metaobject.
     */
    void for_each_property_do
        (
        Property::PropertyOperation &   operation,
        ObjectProperty *                object
        ) const;

protected:
    
    /**
     * Initialize property ownership hierarchy and per-property
     * attributes.
     * Updates property owners and attributes of all owned properties or
     * properties defined in this class only (excluding inherited ones).
     * This depends on @a all_properties flag.
     */
    void initialize_property_hierarchy
        (
        ObjectProperty *    object,
        bool                all_properties = true
        ) const;

// Information about methods.
public:

    /**
     * Get information about methods callable via RPC.
     *
     * The list contains descriptions of methods newly introducted in this class.
     */
    INLINE const MethodInfos & get_method_infos() const;

    /**
     * Create new method arguments/results packet for method @a method_index.
     *
     * Returns NULL if @a method_index is invalid.
     */
    virtual std::auto_ptr< MethodPacket > create_method_packet
        (
        std::size_t                 method_index
        ) const = 0;

    /**
     * Call local method specified by index into list returned by #get_methods().
     * Method arguments will be read from given text reader using standard
     * serialization methods. Multiple arguments should be separated by
     * whitespace.
     */
    void call_method
        (
        const VariantPointer &      object,
        std::size_t                 method_index,
        TextReader &                tr
        ) const;

    /**
     * Call local method specified by valid iterator into list returned by #get_methods().
     * Method arguments will be read from given text reader.
     * @note If @c method_it points to a method defined by a parent class,
     * the call fails.
     */
    void call_method
        (
        const VariantPointer &      object,
        MethodInfos::const_iterator method_it,
        TextReader &                tr
        ) const;

    /**
     * Call local method specified by its name.
     * Method arguments will be read from given text reader.
     *
     * @return @c true on success, @c false if no such method exists in this class.
     */
    bool call_method
        (
        const VariantPointer &      object,
        const std::string &         method_name,
        TextReader &                tr
        ) const;

    /**
     * Call remote method specified by index into list returned by #get_methods().
     * Method arguments will be read from given text reader using standard
     * serialization methods. Multiple arguments should be separated by
     * whitespace.
     * Return method packet with results or throw exception if call failed.
     * The call is always synchronous.
     */
    std::auto_ptr< MethodPacket > remote_call_method
        (
        Remote< Object >            object,
        std::size_t                 method_index,
        TextReader &                tr
        ) const;
    
    /**
     * Call remote method specified by valid iterator into list returned by
     * #get_methods().
     * A @c remote_call_method() variant.
     */
    std::auto_ptr< MethodPacket > remote_call_method
        (
        Remote< Object >            object,
        MethodInfos::const_iterator method_it,
        TextReader &                tr
        ) const;
    
    /**
     * Call remote method specified by its name.
     * Unlike other #remote_call_method() methods, this one will search parents
     * of this class for method of given name if it's not implemented directly
     * by this class.
     */
    std::auto_ptr< MethodPacket > remote_call_method
        (
        Remote< Object >            object,
        const std::string &         method_name,
        TextReader &                tr
        ) const;

// Flat method API, accesses methods described by parent metaobjects.
public:

    /**
     * Get information about all methods matching given wildcard pattern.
     *
     * @param result
     *      Push iterators pointing to found method infos to this vector.
     * @param pattern
     *      Wildcard pattern.
     *
     * @bug @a result can be used to inspect method information only.
     *      Since returned method info iterators can point to methods
     *      described by parent metaobjects the iterators can not be
     *      passed to non-flat API without appropriate metaobject "cast".
     */
    void find_method_infos
        (
        MethodInfoIterators &       result,
        const std::string &         pattern
        ) const;

    /**
     * Get method contexts of all methods matching given wildcard pattern.
     *
     * @param result
     *      Push method contexts to this vector.
     *      The contexts specify what object component the methods belong to,
     *      object components's metaobjects and method iterators relative
     *      to that component. These iterators can be passed to non-flat
     *      introspection API.
     * @param object
     *      Object component that should be searched.
     *      It is assumed that @c object is described by @c this metaobject.
     *
     * @todo Improve standard method API (make it similar to property API)
     *       and get rid of this.
     */
    void find_methods
        (
        MemberContexts &            result,
        const VariantPointer &      object,
        const std::string &         pattern
        ) const;

    /**
     * Call all methods matching given wildcard pattern.
     *
     * See #find_method_infos and #call_method for info about the parameters.
     *
     * @return Number of methods called.
     */
    int call_methods
        (
        const PtrToObject &         object,
        const std::string &         pattern,
        TextReader &                tr
        ) const;

    /**
     * Call all methods matching given wildcard pattern.
     *
     * See #find_method_infos and #call_method for info about the parameters.
     *
     * @return Number of methods called.
     */
    int call_methods
        (
        const VariantPointer &      object,
        const std::string &         pattern,
        TextReader &                tr
        ) const;

    /**
     * Call method described by method @c context.
     *
     * @todo Get rid of this.
     */
    static void call_method
        (
        const MemberContext &       context,
        TextReader &                tr
        );

// Object serialization.
protected:

    /**
     * Serialize @a object to bitstream @a bos.
     *
     * For each property in well-defined order single bit is written: 1 if property
     * is dirty (in this case property value immediately follows), 0 otherwise.
     *
     * @bug Zero bit is written even if "should serialize to specified destination"
     *      condition is false. Replace boolean result of should_serialize by enum
     *      and test it even when reading to save few bits.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc,
        const ObjectProperty *          object
        ) const;

    /**
     * Read @a object from bitstream @a bis.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc,
        ObjectProperty *                object
        ) const;

    /**
     * Serialize @a object properties to text stream @a tw.
     *
     * Objects should be written this way:
     * @verbatim
object # class_name
  :flags:class_component
    property_name property_value
    ...
  ...
@endobject
@endverbatim
     * @c class_component is complete "inheritance path" for properties defined in
     * parents (ie. if @c B inherits @c A and @c C inherits @c B , inheritance path
     * for @c A in @c C is @c B/A ).
     *
     * "object" should be written to current line without any preceding whitespace.
     * "@endobject" should be written on a new line but without trailing newline
     * character (tw.end_line() should not be called after @endobject is written).
     *
     * For properties defined in most derived class, @c flags is set to @b m and
     * @c class_component is empty. For other components, the only allowed flag
     * is @b v, which denotes virtual inheritance. In that case @c class_component
     * is just name of the virtually inherited class.
     *
     * @todo Use unique suffixes of class_component instead of whole inheritance path.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc,
        const ObjectProperty *          object
        ) const;

    /**
     * Read @a object properties from text stream @a tr.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc,
        ObjectProperty *                object
        ) const;

// Other generated object operations.
protected:

    /**
     * Memberwise compare two objects of the same class.
     *
     * @note Default implementation panics with "uncomparable classes" error.
     *
     * @note Current implementation requires each managed class
     *       to implement operator==.
     */
    virtual bool compare_objects
        (
        const ObjectProperty *  object1,
        const ObjectProperty *  object2
        ) const;

    /**
     * Memberwise copy objects of the same class.
     *
     * @note Default implementation panics with "unassignable classes" error.
     *
     * @note Current implementation requires each managed class
     *       to implement operator=.
     */
    virtual void copy_objects
        (
        ObjectProperty *        destination,
        const ObjectProperty *  source
        ) const;

// Helper methods.
protected:

    /**
     * text_read() helper.
     * Throws exception unless string "object" is found on input.
     */
    void text_read_object
        (
        TextReader &    tr
        ) const;

    /**
     * text_read() helper.
     * Reads in next line and first word. Throws if something weird happens.
     * @return false if string "\@endobject" is read, true otherwise.
     */
    bool text_read_line
        (
        TextReader &    tr,
        std::string &   result
        ) const;

    /**
     * Panic with "Upcast component index out of range" error.
     */
    static NORETURN( void ) panic_upcast_index();

// Information initialized before registration.
protected:

    std::string                     class_name;
        /**< Class name. */

    const std::type_info *          class_type_info;
        /**< Type info about the managed class.
             Note that type infos must be compared by their operator==() 
             functions. When using shared libraries, std::type_info instances 
             need not be singleton objects. */        

    const std::type_info *          class_pointer_type_info;
        /**< Type info about a pointer to the managed class. 
             Note that type infos must be compared by their operator==() 
             functions. When using shared libraries, std::type_info instances 
             need not be singleton objects. */        

    ClassTypeId                     class_type_id;
        /**< Unique identifier of the managed class. */

    std::auto_ptr< ObjectFactory >  object_factory;
        /**< Object factory. */

    Attributes                      attributes;
        /**< Class attributes. */

// Information initialized by #initialize_introspection() and
// #initialize_inherited_introspection(). Nothing will be added to or removed
// from the vectors after metaobject initialization.
protected:

    ComponentInfos                  component_infos;
        /**< Information about all "components" including managed class itself,
             which will be the first element of this vector. */

    PropertyInfos                   property_infos;
        /**< Information about all properties, both defined in the class and inherited. */

    MethodInfos                     method_infos;
        /**< Information about methods that can be called via RPC defined in the class. */

    mutable bool                    offsets_initialized;
        /**< Have component and property offsets been initialized?

             @note Mutable because #initialize_offsets() is @c const. This
                   may be fixed in future. */

    }; // class MetaObject


class MASSIV_EXPORTED MetaObject::ComponentInfo
    {
    /// Only MetaObject can access #offset.
    friend class MetaObject;

public:

    /**
     * Constructor.
     *
     * Note that @a the_offset should be relative to pointer to the main
     * component, not to @a the_inherited_from component.
     */
    ComponentInfo
        (
        const std::type_info &  component_type_info,
        bool                    the_virtual_component,
        int                     the_inherited_from,
        const std::string &     the_name
        );

public:

    /** Is this the complete managed class component? */
    INLINE bool is_complete_component() const;

    /** Is this direct parent of the managed class? */
    INLINE bool is_parent() const;

public:

    const MetaObject *  metaobject;
        /**< Metaobject describing the component. */

    bool                virtual_component;
        /**< Is this component inherited virtually? */

    int                 inherited_from;
        /**< Index into #MetaObject::component_infos of component that directly
             inherits this component. -1 for the managed class itself, 0 for
             direct parent classes. */

    std::string         name;
        /**< Complete component name.
             Empty string for the class itself, complete inheritance path
             (class names separated by '/') for base classes. If class is
             inherited virtually, this will be one of the possible paths to
             use to cast to the virtual base. */

protected:

    mutable int         offset;
        /**< Pointer offset.

             When upcasting from pointer to the managed class to pointer to
             this component, move the pointer by this number of bytes (add) to
             get the correct pointer.

             The value is useful only when type of pointed object is the managed
             class, using @a offset will not work if pointed object inherits managed
             class (because of possible virtual inheritance problems).

             This value will be initialized by #initialize_offsets(), that's
             why it's mutable. This problem may be fixed in far far future. */
        
    }; // class MetaObject::ComponentInfo


/**
 * Dump ComponentInfo.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::ComponentInfo &   info
    );


class MASSIV_EXPORTED MetaObject::DataInfo
    {
public:

    /**
     * Constructor.
     */
    DataInfo
        (
        const std::string &     the_name,
        const std::string &     the_type
        );

public:

    std::string         name;
        /**< Property/argument name. */

    std::string         type;
        /**< Type in IDL syntax. */

    }; // class MetaObject::DataInfo


/**
 * Dump DataInfo.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::DataInfo &        info
    );


class MASSIV_EXPORTED MetaObject::PropertyInfo : public DataInfo
    {
    /// Only MetaObject can access offsets.
    friend class MetaObject;
   
    /// Superclass.
    typedef DataInfo Super;

public:

    /**
     * Constructor.
     */
    PropertyInfo
        (
        const std::string &             the_name,
        const std::string &             the_type,
        ComponentInfoIterator           the_component,
        const Property::Attributes &    the_attributes,
        int                             the_component_offset
        );

public:

    ComponentInfoIterator   component;
        /**< Component this property is defined in. */

    Property::Attributes    attributes;
        /**< Property attributes. */

protected:

    int                     component_offset;
        /**< Property offset within #component.

             When pointer to #component is moved by @a component_offset
             (in bytes), it will point to the property.

             This value is initialized by generated code. */

    mutable int             global_offset;
        /**< Property offset within complete component.
             
             When pointer to object of managed class is moved by @a global_offset
             (in bytes), it will point to the property.

             This value can't be used if pointed object inherits managed class.
             Use info from its real metaobject instead.

             This value is initialized when needed by #initialize_offsets(), and
             that's why it's mutable. May be fixed in future. */

    }; // class MetaObject::PropertyInfo


/**
 * Dump PropertyInfo.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::PropertyInfo &    info
    );


class MASSIV_EXPORTED MetaObject::ArgumentInfo : public DataInfo
    {
    /// Superclass.
    typedef DataInfo Super;

public:

    /**
     * Argument kind.
     */
    enum Kind
        {
        ARGUMENT_IN,    /**< "in" argument. */
        ARGUMENT_OUT,   /**< "out" argument. */
        ARGUMENT_INOUT  /**< "in/out" argument. */
        };

public:

    /**
     * Constructor.
     */
    ArgumentInfo
        (
        const std::string &     the_name,
        const std::string &     the_type,
        Kind                    the_kind
        );

public:

    /** Convert @a kind to string. */
    std::string kind_string() const;

public:

    Kind                kind;
        /**< Argument kind. */

    }; // class MetaObject::ArgumentInfo

    
/**
 * Dump ArgumentInfo.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                      stream,
    const MetaObject::ArgumentInfo &    info
    );


class MASSIV_EXPORTED MetaObject::MethodInfo
    {
public:

    /**
     * Method attributes.
     */
    struct MASSIV_EXPORTED Attributes
        {
    public:

        /**
         * Constructor.
         */
        INLINE Attributes();

    public:
        // Attribute names are prefixed by "a_" to prevent clashes with
        // reserved symbol names.
        // Do not forget to update operator<< when adding attributes.

        bool    a_const;
            /**< Value of 'const' attribute. */

        bool    a_virtual;
            /**< Value of 'virtual' attribute. */

        }; // struct Attributes

public:

    /**
     * Constructor.
     */
    MethodInfo
        (
        const std::string &     the_name,
        const std::string &     the_result_type = std::string()
        );

public:

    Attributes          attributes;
        /**< Method attributes. */

    std::string         name;
        /**< Method name. */

    ArgumentInfos       arguments;
        /**< Information about method arguments. */

    std::string         result_type;
        /**< Type of result in IDL syntax. Empty string if method does not return value. */

    }; // class MetaObject::MethodInfo


/**
 * Dump MethodInfo::Attributes in IDL style.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                              stream,
    const MetaObject::MethodInfo::Attributes &  info
    );


/**
 * Dump MethodInfo, without attributes, in IDL style.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                  stream,
    const MetaObject::MethodInfo &  info
    );


class MASSIV_EXPORTED MetaObject::MemberContext
    {
public:

    VariantPointer          object;
        /**< Object component that owns the described member. */
    const MetaObject *      metaobject;
        /**< Metaobject of the object component. */
    MethodInfoIterator      method_it;
        /**< MethodInfo, valid for a method member. */
    }; // class MetaObject::MemberContext


/**
 * Dump MetaObject::Attributes enclosed in @c \<, @c \>.
 *
 * Boolean attributes are written only if they are @c true, other attributes
 * are written in <tt>name = value</tt> style.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                  stream,
    const MetaObject::Attributes &  attributes
    );

    
/**
 * Write MetaObject to stream in IDL-like syntax.
 *
 * @note As all nice-behaved operator<< should do, trailing new-line will
 *       not be written.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &      stream,
    const MetaObject &  metaobject
    );

/** @} */ // addtogroup object_management


} // namespace Massiv::Core


namespace Generated {


/** @addtogroup object_management
 * @{
 */

/**
 * Generated metaobjects should be specializations of this class.
 *
 * Each generated metaobject should inherit ::Massiv::Core::MetaObject and
 * implement all pure virtual methods declared there.
 *
 * It should define the following static variables:
 * - <tt>const MetaObjectImpl< Class > * self</tt>. Pointer to the only instance
 *   of the metaobject.
 * - <tt>ClassTypeId class_id</tt>. Class type id. Value of this object
 *   is initialized in code generated by classlist.pl.
 *
 * @todo More static variables.
 *
 * @ingroup object_management
 */
template< typename Class >
class MetaObjectImpl; /* : public ::Massiv::Core::MetaObject
    {
    static const MetaObjectImpl< Class > * self
    static ::Massiv::Core::ClassTypeId class_id
    }; */


/**
 * Name of metaobject class for given object class.
 */
#define META_OBJECT( class ) ::Massiv::Generated::MetaObjectImpl< class >

/** @} */ // addtogroup object_management


} // namespace Massiv::Generated
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object/metaobject_inline.h"
#endif


#endif // MASSIV_METAOBJECT_H
