/** @file object.cpp
 * Object & ObjectProperty are tightly coupled classes.
 * Both implement "managed object" (each class implements part of it).
 * While ObjectProperty implements Property interface, holds all the
 * internal attributes and define private helper methods used by Object,
 * Object defines public API that can be used by client.
 * ObjectProperty is used exclusively by the core, Object is visible both to
 * the core and client.
 * The interface separation splits managed object into "core part of the
 * Object" and "client part of the Object".
 * ObjectProperty is simply accessed by "ObjectProperty *",
 * Object by @c ObjectPointer, @c PtrToObject or @c UpcallToObject.
 * "Object *" must not be used to access @c Object!!!.
 *
 * The core must ensure that if a client overridable method is active
 * (being called) on an Object, the Object instance is pinned.
 * Client code can potentially reenter system main loop, invoke GC, etc.
 * The core must thus know which objects are active (pinned) so that they
 * won't be deleted from the memory (because of load balancing, migration,
 * etc). ObjectPointer automatically pins the objects that are currently
 * in use (method call is in progress on them), so client code need not
 * care about Object pinning. Since core might not use ObjectPointers it
 * must pin Objects by its own mechanisms (@c PtrToObject, @c UpcallToObject).
 * Macro ENSURE_PINNED tests if Object is pinned and is provided for
 * "core correctness" self testing.
 *
 * Some of the upcalls called on an Object are not allowed to reenter
 * system main loop (for example as a result of synchronous RPC call).
 * In that case the core must lock current system tick prior
 * invoking the upcalls so that potential attempt to reenter the
 * core from the upcall will be detected (System::next_tick() would be
 * called while the previous tick was still locked).
 * To ensure that current tick is locked ENSURE_LOCKED macro is defined.
 *
 * Use @c UpcallToObject instead of @c PtrToObject when pointer is used
 * to upcall client/simulation code so that we would be able to trace
 * Object call stack.
 */

/*
 * begin        : Sun Dec 06 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "object/object.h"
#include "pointer/object_pointer.h"
#include "pointer/pointer_to_object.h"
#include "object/object_factory.h"
#include "object/class_manager.h"
#include "object/object_manager.h"
#include "replication/replica_server.h"
#include "system/tick_lock.h"
#include "system/system.h"
#include "object/lib/lib_exceptions_helper.h"
#include "node/node_manager.h"


namespace Massiv {
namespace Core {


#if( defined( OPTION_DEBUG ) )

/** Asserts if object is not pinned. */
#define ENSURE_PINNED() \
    if( !is_pinned() ) \
        panic( "Called a (virtual) method on an unpinned Object!" );

/**
 * Asserts if current system tick is not locked (and thus synchronous
 * RPC call may be invoked).
 */
#define ENSURE_LOCKED() \
    if( !System::is_tick_locked() ) \
        panic( "Current system tick must be locked!" );

#else

#define ENSURE_PINNED()
#define ENSURE_LOCKED()

#endif


namespace {

/**
 * This property operation writes object state into output stream.
 */
struct DebugDumpObject : public Property::PropertyOperation
    {

    DebugDumpObject
        (
        std::ostream &  stream
        ) : stream( stream )
        {
        }

    Result operator()
        (
        Property *  property
        )
        {
        Property::PropertyInfo info = property->get_info
            (
            Property::QUERY_TYPE |
            Property::QUERY_STATE |
            Property::QUERY_FULL_NAME_NO_ROOT
            );

        if( info.type == Property::TYPE_VALUE_TYPE ||
            info.type == Property::TYPE_POINTER ||
            info.type == Property::TYPE_CONTAINER )
            {
            stream << '\n' << info.property_name << " = " << info.state;
            return RESULT_DONE;
            }
        else
        if( info.type == Property::TYPE_OBJECT_PROPERTY )
            {
            return RESULT_RECURSE;
            }

        return RESULT_DONE;
        }

    std::ostream &  stream;
        /**< Stream to write to. */

    }; // struct DebugDumpObject

} // anonymous namespace
    

//
// Instance management.
//
    
Pointer<Object> Object::clone() const
    {
    ENSURE_PINNED();

    const ObjectFactory * factory = get_factory();
    hope( factory );

    return checked_cast<Object *>( factory->clone_object( this ) );
    }


bool Object::equals
    (
    WeakPointer<Object> object
    ) const
    {
    ENSURE_PINNED();
    
    return this == object;
    }


bool Object::memberwise_equals
    (
    Pointer<Object> object
    ) const
    {
    ENSURE_PINNED();

    PtrToObject other = object.dereference();

    // Objects must be of the same classes.
    
    hope( metaobject );
    
    #ifdef OPTION_DEBUG

    if( metaobject != other->metaobject )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            InvalidArgumentException,
            "memberwise_equals() - Objects must be of the same classes."
            );
        }

    #endif
    
    return metaobject->compare_objects( this, other );
    }


void Object::copy
    (
    Pointer<Object> object
    )
    {
    ENSURE_PINNED();

    PtrToObject other = object.dereference();

    // Objects must be of the same classes.

    hope( metaobject );
    
    #ifdef OPTION_DEBUG

    if( metaobject != other->metaobject )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            InvalidArgumentException,
            "copy() - Objects must be of the same classes."
            );
        }

    #endif

    metaobject->copy_objects( this, other );
    }


int Object::hash() const
    {
    ENSURE_PINNED();

    return 0;
    }


std::string Object::to_string() const
    {
    ENSURE_PINNED();

    std::ostringstream  stream;
    DebugDumpObject     write( stream );
    ObjectProperty *    object = const_cast<Object *>( this );

    object->for_each_owned_property_do( write );

    return stream.str();
    }


std::ostream & operator<<
    (
    std::ostream &          stream,
    const ObjectProperty &  object
    )
    {
    UpcallToObject upcall = const_cast<ObjectProperty *>( &object );
    stream << upcall->to_string();
    return stream;
    }


//
// RTTI.
//

bool Object::inherits
    (
    ClassTypeId class_type_id
    ) const
    {
    ClassManager & manager = Global::class_manager();
    const MetaObject * other_metaobject = manager.get_metaobject( class_type_id );
    return inherits( other_metaobject->get_class_type_info() );
    }


bool Object::inherits
    (
    const std::type_info & class_type_info
    ) const
    {
    hope( metaobject );
    return metaobject->inherits( class_type_info );
    }


const std::type_info & Object::get_type_info() const
    {
    hope( metaobject );
    return metaobject->get_class_type_info();
    }


//
// Object state.
//

void Object::name_object
    (
    const std::string & name
    )
    {
    GarbageCollector & gc = Global::garbage_collector();

    gc.name_object( get_object_id(), name );
    }


//
// Event scheduling.
//

EventHandle Object::schedule_to
    (
    WeakPointer<Object> destination,
    const STime &       time,
    bool                notify
    )
    {
    EventHandle h;

    ObjectProperty::migrate_to( destination.get_object_id(), time, notify, &h );

    return h;
    }


void Object::migrate_to
    (
    WeakPointer<Object>     destination,
    const STime &           time,
    bool                    notify
    )
    {
    ObjectProperty::migrate_to( destination.get_object_id(), time, notify, NULL );
    }


void Object::deliver_asap_to
    (
    WeakPointer<Object> destination,
    bool                can_deliver_to_replica
    )
    {
    // If object is not local, is inconsistent, object manager is not idle
    // (there are pending events) or there is little free stack space then
    // do not nest.
    // The two first conditions are needed because of @c deliver_asap_to()
    // semantics, the other two to increase system throughput or prevent
    // stack overflow.
    // This relates mainly to archivation and GC_DELAYED_UNPIN semantics
    // when system should be on the level zero of the system main loop
    // as much as possible (should not nest). Otherwise archivation can
    // not be started and/or objects active-unpinned (the latter may lead
    // to global event schedule blocking - objects remain conservatively
    // pinned and can not migrate). See GC_DELAYED_UNPIN.

    hope( !is_replica() );

    if( destination.is_local() && destination->is_consistent() &&
        Global::object_manager().is_idle() && System::have_enough_stack_space() )
        {
        if( can_deliver_to_replica || !destination->is_replica() )
            {
            Balancer & balancer = Global::balancer();
            
            // Log.

            if( Global::object_manager().get_settings().log_events )
                {
                Global::object_manager().log_delivered_to
                    (
                    get_object_id(),
                    destination->get_object_id(),
                    STime::current_time()
                    );
                }
            
            // Create strong pointers to the involved objects if possible.
            // This is needed so that it would be consistent with migration
            // delivery if a migration was scheduled instead.
            // See @c ObjectManager::dispatch_upcall().

            Pointer<Object> this_strong;
            Pointer<Object> destination_strong;

            this_strong = this;

            if( !destination->is_replica() )
                {
                destination_strong = destination;
                }

            // Deliver immediatelly.

            balancer.account_for_object_migration
                (
                this,
                destination.get_object_id()
                );

            UpcallToObject upcall = this;
            upcall.tag_upcall( "::Massiv::Core::Object::deliver_asap_to( ... )" );
            upcall->delivered_to( destination, STime::current_time() );

            return;
            }
        }
    
    // Schedule the migration.

    migrate_to( destination, STime::current_time() );
    }


//
// Replication.
//

void Object::replicate_to
    (
    WeakPointer<Object> destination_node,
    const STime &       timeout
    )
    {
    ENSURE_PINNED();

    const ObjectId & destination_object_id = destination_node.get_object_id();
    
    if( !destination_object_id.is_proxy_node_object_id() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            InvalidArgumentException,
            "destination_node must refer to a NodeObjectInterface object."
            );
        }

    NodeId node_id = destination_object_id.proxy_extract_node_id();

    // Check whether node_id is id of existing node and whether the node
    // is (still) connected to simulation.
    if( !Global::node_manager().is_node_accessible( node_id ) )
        {
        // FIXME: Call cancel_replication( node_id )?
        // Once a client is disconnected from the simulation, all replica
        // requests should be canceled through GlobalCallbacks::node_disconnected().
        // All new other requests will fail because is_node_accessible()
        // will return false. So we should call cancel_replication()
        // only if we are paranoid.
        return;
        }

    ObjectManager & object_manager = Global::object_manager();
    hope( object_manager.replica_server.get() );
    object_manager.replica_server->request_replication( get_object_id(), node_id, timeout );
    }


void Object::cancel_replication
    (
    WeakPointer<Object> destination_node
    )
    {
    ENSURE_PINNED();

    // Note that this will add entry to replication request table even if
    // there is none yet, but no replication update will be made. Time
    // test should probably be added to replica server.

    replicate_to( destination_node, STime() );
    }


//
// Object callbacks.
//

void Object::delivered_to
    (
    WeakPointer<Object> destination,
    const STime &       deliver_time
    )
    {
    ENSURE_PINNED();

    hope( !is_member_object() );
    hope( !is_replica() );
    
    hope( destination.is_local() );
    hope( !destination->is_member_object() );
    }


void Object::delivery_failed
    (
    DeliveryFailure failure,
    const STime &   deliver_time
    )
    {
    ENSURE_PINNED();

    hope( !is_member_object() );
    hope( !is_replica() );
    }


void Object::object_updated
    (
    UpdateReason
    )
    {
    // Synchronous RPC call is not allowed to be called from within
    // the method.

    ENSURE_PINNED();
    ENSURE_LOCKED();
    }


#undef ENSURE_PINNED
#undef ENSURE_LOCKED

} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "object_inline.h"
#endif
