/** @file object.h
 * Managed object, public API.
 */

/*
 * begin        : Sun Dec 06 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_H
#define MASSIV_OBJECT_H


#ifndef MASSIV_OBJECT_PROPERTY_H
#include "object/object_property.h"
#endif


namespace Massiv {
namespace Core {

    
/**
 * Managed object.
 *
 * Every client object managed by the core must be derived from @c Object.
 * @c Object defines public API that can be safely used by client and
 * implements basic @c Property-like functionality required by the core.
 * Managed objects are described by IDL (mainly inheritence hierarchy and its
 * data members, that can be any @c Property-derived classes) so that core
 * could automatically (without user's assistance) serialize, instantiate
 * or delete such instances.
 *
 * Any managed object must inherit Object exactly once (thus preferable
 * way for deriving client objects from @c Object is virtual inheritance),
 * have no user-defined constructors and destructors and must define
 * @c initialize() pseudoconstructor (see object instantiation).
 * The convention for Object is that methods explicitly marked "virtual"
 * can be overriden by client (these often serve as callbacks upcalled from
 * the core), while the others should not (regardless if they are already
 * virtual from ObjectProperty or not).
 *
 * Object instances are created by @c CreateObject template. This creates
 * a new instance of the specified class using its ObjectFactory, intializes it
 * and calls user-defined @c initialize() pseudo-constructor. The reason why
 * user can not simply define constructors for its managed classes is that core
 * needs to instantiate objects not only on behalf of user requests but also on
 * its own needs that must be transparent to the user (user should not be aware
 * of). An object can be instantiated many times during its global (simulation)
 * life time, even on more than one node at a time (see object replication).
 * Examples of such instantiations are delivery of object that migrated to the
 * local node, creating instance of an object read from the archive, object
 * replication, etc. @c initialize() pseudo-constructor is called by the core
 * once only, when it is instantiated for the first time (by @c CreateObject).
 * @c CreateObject creates "stand-alone objects" (they are allocated on the
 * heap, can be referenced by ObjectPointers and can migrate, be replicated,
 * etc.).
 *
 * @note    Stand-alone object instances can also be created directly by
 *          ObjectFactory. However this creates the object in a default
 *          state (@c initialize() won't be called). Also objects that
 *          are "embedded" to another objects as their member objects
 *          will be left in a default state. Owner object's @c initialize()
 *          should propagate the call to the member objects if needed.
 *
 * Simple temporary objects can be created by @c ObjectOnStack template.
 * This allocates the object on the stack, initializes it and calls user-defined
 * @c initialize() pseudo-constructor. Objects on the stack are not "stand-alone",
 * they can not be referenced by ObjectPointers, etc.
 * 
 * Each object is associated with @c MetaObject that accesses object
 * metadata (can be used for object introspection).
 *
 * Safe references to stand-alone object are represented by ObjectPointers.
 * User must not use native references to managed objects as core manages
 * lifetime of local instances automatically (for example the object can
 * migrate out of the local node or can be garbage collected). In order to
 * make this work core must know about what objects are referenced and that
 * is what ObjectPointers are for.
 *
 * Managed objects can have their members "private", "public" or "protected".
 * If they are "private" or "protected" the objects must properly declare their
 * friends so that core classes (meta objects and object factories) could access
 * them. For this purposes macro @c MASSIV_OBJECT is defined.
 *
 * Core implements concepts of value types, user-defined objects that can be
 * passed by value to RPC calls or instantiated on the stack more comfortably,
 * and throwables, objects that can be thrown by value and propagated to caller
 * during a RPC call. Both value types and throwables are special kind of managed
 * objects. See @c ValueType, @c Throwable.
 *
 * @ingroup object
 * @ingroup property
 */
class MASSIV_EXPORTED Object : public ObjectProperty
    {
private:
        
    friend class ObjectManager;
    friend class ObjectProperty;
    friend class ObjectFactory;
    friend class ReplicaClient;

public:

    //
    // Initialization.
    //

    using ObjectProperty::set_no_archive;
    
public:

    //
    // Instance management.
    //

    /**
     * Create new object and use copy values of all its properties.
     * May be redefined by client.
     */
    virtual Pointer<Object> clone() const;

    /**
     * Tests the object against @c object for equality.
     * By default reference-based comparison is performed, ie.
     * the two objects are equal iff they are the same instance
     * (this == object).
     * May be redefined by client.
     */
    virtual bool equals
        (
        WeakPointer<Object> object
        ) const;

    /**
     * Tests the object with @c object for memberwise equality.
     * The objects are compared on member-by-member basis.
     * operator==() is used for comparing object members.
     * May be redefined by client.
     */
    virtual bool memberwise_equals
        (
        Pointer<Object> object
        ) const;

    /**
     * Copies @c object to this.
     * By default memberwise-copy is performed by calling
     * operator=() on each member.
     * May be redefined by client.
     */
    virtual void copy
        (
        Pointer<Object> object
        );
    
    /**
     * Returns object's hash code.
     * May be redefined by client.
     */
    virtual int hash() const;
    
    /**
     * Converts object to string.
     * May be redefined by client.
     */
    virtual std::string to_string() const;

public:

    //
    // RTTI.
    //
    
    using ObjectProperty::get_object_id;

    /**
     * Checks if this object is derived from specified class (RTTI).
     */
    bool inherits
        (
        ClassTypeId class_type_id
        ) const;

    /**
     * Checks if this object is derived from specified class (RTTI).
     */
    bool inherits
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * Returns type info about object (RTTI).
     */
    const std::type_info & get_type_info() const;

public:

    //
    // Object state.
    //
    
    using ObjectProperty::is_pinned;
    using ObjectProperty::pin;
    using ObjectProperty::unpin;

    /**
     * Sets object name (for debugging & maintenance purposes).
     * Object will be registered to local naming database under this
     * name.
     * Only root objects can be named.
     */
    void name_object
        (
        const std::string & name
        );

public:

    //
    // Event scheduling API.
    //

    /**
     * Schedule the object to be delivered to the @a destination_object
     * at time @a time.
     * Returns event handle to the scheduled event.
     * When the migration is done (and @a notify is set to true),
     * @a delivered_to() will be called by the core if the delivery was
     * successfull, or @a delivery_failed() if the
     * the destination object was not found in simulation.
     *    
     * This function schedules the event as a public event.
     * The advantage of a public event is that it can be killed / cancelled.
     * On the other hand public event has much bigger overhead
     * when it is transfered over network.
     * So if you want to schedule some object but you do NOT
     * require EventHandle, use @c migrate_to() instead of @c schedule_to().
     *
     * @param notify If true, the object will be notified when the migration
     *      is done (by @c delivered_to() or by @c delivery_failed() ).
     *      If false, the object won't be notified of finished migration
     *      (neither @c delivered_to() not delivery_failed() will be called).
     *      Set @a notify to false to migrate the object without its consciousness.
     *
     * @todo Pripsat podminky - repliky, vzdalene
     * objekty, ...
     * @bug Killing scheduled events works only on events scheduled for
     *      the object itself, not for another objects.
     */
    EventHandle schedule_to
        (
        WeakPointer<Object> destination_object,
        const STime &       time,
        bool                notify = true
        );

    /**
     * Schedule the object to be delivered to @a destination
     * at time @a time.
     * Same as @c schedule_to() only does not return
     * event handle, so the event will be scheduled
     * as private event (can't be killed).
     * @see schedule_to().
     */
    void migrate_to
        (
        WeakPointer<Object> destination,
        const STime &       time,
        bool                notify = true
        );
    
    /**
     * Schedules the object to be delivered to @a destination
     * as soon as possible (ASAP).
     * The major difference from general @c migrate_to is that
     * the object is delivered _immediatelly_ if @a destination
     * is a local object instead of _scheduling_ the migration.
     *
     * If @c destination is a local object @c delivered_to will already
     * have been executed by the time when @c deliver_asap_to returns.
     * If the object is not local a migration to @a destination is scheduled
     * to current simulation time and @c delivered_to will be called when
     * the migration finishes.
     * It is better not to rely on this and simply assume that
     * @c delivered_to() _might_ or might _not_ already be called.
     *
     * @c can_deliver_to_replica determines whether the object can be
     * delivered to a consistent local replica of @a destination object.
     * If a migration is scheduled it is always delivered to the original
     * object.
     */
    void deliver_asap_to
        (
        WeakPointer<Object> destination,
        bool                can_deliver_to_replica = false
        );

public:

    //
    // Replication API.
    //

    /**
     * Request replication of the object to node specified by @a destination_node
     * proxy node object pointer until current simulation time exceeds @a timeout.
     */
    void replicate_to
        (
        WeakPointer<Object> destination_node,
        const STime &       timeout
        );

    /**
     * Cancel replication request.
     */
    void cancel_replication
        (
        WeakPointer<Object> destination_node
        );

protected:

    //
    // Client-redefinable callbacks.
    //

    /**
     * Delivery failure.
     */
    enum DeliveryFailure
        {
        DESTINATION_NOT_ACCESSIBLE,
            /**< Destination of the scheduled event is not accessible in the system. */
        EVENT_KILLED,
            /**< The scheduled event was killed. */
        }; // DeliveryFailure
    
    /**
     * Parameter of #object_updated() callback.
     */
    enum UpdateReason
        {
        INSTANCE_CREATED,
            //< New instance of object has been created.
        INSTANCE_MIGRATED,
            /**< Object migrated to the local system.
                 This includes objects swapped in from the swap or loaded
                 from the archive. */
        INSTANCE_FINALIZE,
            /**< Called automatically when GC triggers object deletion.
                 Client can intercept this and perform object-specific clean
                 up. However it is better not to rely on object finalization.
                 Most objects should not care. */
        REPLICA_CREATED,
            /**< New object replica has been created.
                 Note that @c REPLICA_UPDATE and @c REPLICA_UPDATED
                 will be "sent" too. */
        REPLICA_UPDATE,
            //< Object replica contents will be updated.
        REPLICA_UPDATED,
            //< Object replica contents has been updated.
        REPLICA_DESTROY,
            //< Object replica will be destroyed.
        SIMULATION_STARTUP
            /**< Starting simulation from an archive.
                 Object is notified with @c SIMULATION_STARTUP only
                 if its IDL class attribute @c simulation_startup_notify
                 is set to true.
                 Usually @c SIMULATION_STARTUP is used to cleanup
                 client connection related data that are valid during archivation
                 of the object but are not valid after the simulation is restored
                 from an archive.
                 See @c object_updated() for more info.
                 @see AccountObjectInterface. */
        }; // enum UpdateReason

    /**
     * Event delivery callback.
     *
     * Called whenever this object is delivered by some event to the destination
     * object of the event. The system guarantees that @c destination points
     * to a local object when the callback is invoked. However it is accessed
     * by a weak pointer so that deliveries to local object replicas by
     * @c deliver_asap_to() are possible.
     *
     * The core ensures that both @c destination (if does not point to a replica)
     * and @c this objects are strong-referenced at the time of the call.
     *
     * To be redefined by client.
     * Default implementation does nothing.
     *
     * @param deliver_time Time the event was scheduled to. May differ from
     *      current time (either by latency caused by transferring the object
     *      over network or by inaccuracy of distributed time synchronization).
     */
    virtual void delivered_to
        (
        WeakPointer<Object> destination,
        const STime &       deliver_time
        );

    /**
     * Delivery failure callback.
     * Called whenever this object can't be delivered by some event for some reason.
     * To be redefined by client.
     */
    virtual void delivery_failed
        (
        DeliveryFailure     failure,
        const STime &       deliver_time
        );
    
    /**
     * General-purpose "object state changed" callback.
     * May be overriden by client under a condition that the implementation
     * does not invoke a synchronous RPC call.
     * Default implementation does nothing.
     * Except very few cases (like debugging) client should ignore this
     * callback.
     *
     * Object is notified with @c SIMULATION_STARTUP reason only
     * if its IDL class attribute @c simulation_startup_notify
     * is set to true.
     * When simulation is being restored form an archive, the system
     * swaps-in all objects with @c simulation_startup_notify
     * class attribute set to on and calls @c object_updated()
     * on them.
     * When this callback is called with @c SIMULATION_STARTUP reason,
     * async rpc can be called, but it will not be executed
     * until the simulation startup is finished and the simulation
     * is running.
     *
     * See @c UpdateReason for more info.
     *
     * @warning Never call synchronous RPC from this function.
     */
    virtual void object_updated
        (
        UpdateReason reason
        );
    
    }; // Object

    
/**
 * Write object state to output stream.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &          stream,
    const ObjectProperty &  object
    );


/**
 * Use this macro in a managed object declaration to allow
 * meta objects and object factories to access object's
 * properties and methods.
 * The access is need for object serialization, replication, ...
 *
 * @bug Another name? MANAGED_OBJECT?
 */
#define MASSIV_OBJECT( Type ) \
    friend class ::Massiv::Generated::MetaObjectImpl< Type >; \
    friend class ::Massiv::Generated::ObjectFactoryImpl< Type >;


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object_inline.h"
#endif


#endif // MASSIV_OBJECT_H
