/** @file object_common.h
 * Object management common types.
 * Used to exchange data among object manager, garbage collector,
 * system messages and archive.
 */

/*
 * begin        : Tue Aug 27 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_COMMON_H
#define MASSIV_OBJECT_COMMON_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_OBJECT_INFO_H
#include "object/object_info.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <vector>
#include <queue>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Encapsulates types which are used to exchange data among
 * object manager, archive manager, garbage collector,
 * system messages and archive.
 */
namespace ObjectShared {

/**
 * List of object ids.
 */
typedef std::vector< ObjectId > ObjectIdList;

/**
 * Set of object ids.
 */
typedef ext::hash_set< ObjectId, ObjectId::Hash > ObjectIdSet;

/**
 * List of objects.
 */
typedef std::vector< ObjectProperty * > ObjectList;

/**
 * Object id queue.
 */
typedef std::queue< ObjectId > ObjectIdQueue;

/**
 * Algorithm that evaluates migration / replication groups
 * needs to treat property pointers as if they were bidirectional.
 * @c BackwardPointers is a list of "inversed" property pointers:
 * If property @c A references object @c B,
 * [ B, A ] is inserted into @c BackwardPointers.
 */
typedef ext::hash_multimap
        <
        ObjectId,
        Property *,
        ObjectId::Hash,
        ObjectId::EqualKey
        >   BackwardPointers;

/**
 * Backward pointer.
 * Always coupled with a (forward) property pointer.
 */
typedef BackwardPointers::iterator BackwardPointer;

/**
 * Range in @c BackwardPointers.
 */
typedef std::pair< BackwardPointer, BackwardPointer > BackwardPointersRange;

} // namespace ObjectShared

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_OBJECT_COMMON_H
