/** @file object_create.h
 * Helper templates used to create new object instances.
 */

/*
 * begin        : Sun Dec 12 2002
 * copyright    : (C) 2002, 2003 by Stepan Vondrak, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECT_CREATE_H
#define MASSIV_OBJECT_CREATE_H


#ifndef MASSIV_OBJECT_H
#include "object.h"
#endif

#ifndef MASSIV_CLASS_MANAGER_H
#include "class_manager.h"
#endif

#ifndef MASSIV_OBJECT_FACTORY_H
#include "object_factory.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Template that can be used to create "stand-alone" objects of a given type.
 * Since created object must be initialized by the core, simple
 * "new" is not sufficient.
 * "CreateObject< Type >( ... )" replaces "new Type( ... )" and ensures
 * proper initialization.
 *
 * Each constructor creates a "holder" object that holds a pointer to
 * a new instance of type @c Type class.
 * Holder object is default convertible to ObjectPointer so that it could be
 * used as an expression. You can also get ObjectId instead of pointer using
 * #id() method.
 *
 * Parameters to constructors are passed to @c Type::initialize(),
 * which is a replacement for constructor. If @c Type is a derived
 * class, it is @c Type's responsibility to initialize its base classes
 * and member objects.
 *
 * The object is created on the heap and can be referenced by ObjectPointers.
 * @pre ObjectFactory for Type must be registered to global ClassManager.
 *
 * @note    The object is created in two phases. In the first phase it
 *          is default-constructed using ObjectFactory. In the second phase
 *          @c initialize() is called on it.
 *
 * @note    Objects can also be instantiated on the stack. See @c ObjectOnStack
 *          template.
 */
template< class Type >
class CreateObject
    {
#if !defined( COMPILER_GCC ) || ( __GNUC__ < 3 ) 

    NO_COPY_CONSTRUCTOR( CreateObject );
    
    // CreateObject instances should never be copied.
    // However GCC 3.4.1 requires copy constructor to be accessible when
    // assigning CreateObject to an ObjectPointer. This is likely a compiler
    // bug.
        
#endif
        
    NO_ASSIGNMENT_OPERATOR( CreateObject );

public:

    // Create initialized object.
    // @note Arguments to constructors must be passed by reference so that
    // they could be delivered to @c Type::initialize() by reference too
    // if needed. Constructors must work as a "transparent bridge" to
    // @c Type::initialize() and must not modify the arguments in any way.

    CreateObject()
        {
        create_uninitialized();
        ptr->initialize();
        }

    template
        <
        typename T1
        >
    CreateObject
        (
        const T1 & arg1
        )
        {
        create_uninitialized();
        ptr->initialize( arg1 );
        }

    template
        <
        typename T1,
        typename T2
        >
    CreateObject        
        (
        const T1 & arg1,
        const T2 & arg2
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2 );
        }

    template
        <
        typename T1,
        typename T2,
        typename T3
        >
    CreateObject    
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3 );
        }

    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4
        >
    CreateObject
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3, arg4 );
        }
    
    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4,
        typename T5
        >
    CreateObject    
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4,
        const T5 & arg5
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3, arg4, arg5 );
        }
    
    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4,
        typename T5,
        typename T6
        >
    CreateObject    
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4,
        const T5 & arg5,
        const T6 & arg6
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3, arg4, arg5, arg6 );
        }

    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4,
        typename T5,
        typename T6,
        typename T7
        >
    CreateObject
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4,
        const T5 & arg5,
        const T6 & arg6,
        const T7 & arg7
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3, arg4, arg5, arg6, arg7 );
        }

    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4,
        typename T5,
        typename T6,
        typename T7,
        typename T8
        >
    CreateObject
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4,
        const T5 & arg5,
        const T6 & arg6,
        const T7 & arg7,
        const T8 & arg8
        )
        {
        create_uninitialized();
        ptr->initialize( arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8 );
        }

public:

    /**
     * Returns ObjectId of the created object.
     */
    const ObjectId & id() const
        {
        return ptr.get_object_id();
        }
    
    /**
     * Converts held pointer to a proxy object.
     * This allows to assign CreateObject instance to pointer of any
     * type.
     */
    operator ObjectProxy &() const
        {
        return ptr.convert();
        }

private:

    /**
     * Create uninitialized object.
     *
     * Calling create_unitialized() directly is not recommended, use
     * CreateObject's constructor instead as it will call object's
     * pseudo-constructor (initialize() method).
     */
    void create_uninitialized()
        {
        ClassManager &          class_manager = Global::class_manager();
        const ObjectFactory *   factory = class_manager.get_factory( typeid( Type ) );
        ObjectProperty *        object = factory->create_object();
        
        ptr = object->get_object_id();
        }

private:

    Pointer< Type > ptr;
        /**< Pointer to the created object. */

    }; // class CreateObject


/**
 * Template that encapsulates an object instance created on program stack.
 * Since created object must be initialized by the core, simple
 * "Type object_on_stack;" may not be sufficient (unless Type is a ValueType).
 *
 * "ObjectOnStack< Type > my_object( ... );" replaces "Type my_object( ... );".
 * ObjectOnStack serves as a handle to the created object which is accessed
 * by @c dereference() or @c operator-\>().
 *
 * Parameters to constructors are passed to @c Type::initialize(),
 * which is a replacement for constructor. If @c Type is a derived
 * class, it is @c Type's responsibility to initialize its base classes
 * and member objects.
 *
 * The object is created on the stack and can not be referenced by
 * ObjectPointers, can not migrate, etc. The object is semantically
 * equal to a member object of another stand-alone object, so the same
 * restrictions apply to objects on the stack as to member objects.
 *
 * @note    The object is first default constructed, initialized by the core
 *          and finally user-defined @c initialize() is called.
 *
 * @note    Objects that are intended to be used as "value objects" only, ie.
 *          instantiated on the stack, put into containers or passed as
 *          arguments to RPC methods, can be implemented as a "value type".
 *          This enforces additional requirements on object implementation but
 *          simplifies its usage. See @c Massiv::Core::ValueType.
 */
template< class Type >
class ObjectOnStack
    {
    NO_ASSIGNMENT_OPERATOR( ObjectOnStack );

public:
        
    // Constructors initialize the object to be owned by the stack root object
    // and automatically pass its arguments to @c Type::initialize()
    // pseudoconstructors.

    ObjectOnStack()
        {
        initialize_stack_object();
        object.initialize();
        }

    ObjectOnStack
        (
        const ObjectOnStack & other
        ) : object( other )
        {
        initialize_stack_object();
        }

    template
        <
        typename T1
        >
    ObjectOnStack
        (
        const T1 & arg1
        )
        {
        initialize_stack_object();
        object.initialize( arg1 );
        }
    
    template
        <
        typename T1,
        typename T2
        >
    ObjectOnStack
        (
        const T1 & arg1,
        const T2 & arg2
        )
        {
        initialize_stack_object();
        object.initialize( arg1, arg2 );
        }
    
    template
        <
        typename T1,
        typename T2,
        typename T3
        >
    ObjectOnStack
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3
        )
        {
        initialize_stack_object();
        object.initialize( arg1, arg2, arg3 );
        }
    
    template
        <
        typename T1,
        typename T2,
        typename T3,
        typename T4
        >
    ObjectOnStack
        (
        const T1 & arg1,
        const T2 & arg2,
        const T3 & arg3,
        const T4 & arg4
        )
        {
        initialize_stack_object();
        object.initialize( arg1, arg2, arg3, arg4 );
        }

public:

    /** Returns stored object. */
    operator Type &() const
        {
        return object;
        }    
    
    /** Returns stored object. */
    Type * operator->() const
        {
        return &object;
        }
    
    /** Returns stored object. */
    Type & dereference() const
        {
        return object;
        }    

private:
        
    /**
     * Initializes the object to be owned by a stack root object.
     */
    void initialize_stack_object()
        {
        object.initialize_from_owner( *object.get_root_owner(), Property::Attributes() );
        }

private:

    mutable Type object;
        /**< The stored object. */

    }; // class ObjectOnStack

/** @} */ // addtogroup object_management
    

} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_OBJECT_CREATE_H
