/** @file object_factory.h
 * Object factory.
 */

/*
 * begin        : Sun Dec 15 2001
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_FACTORY_H
#define MASSIV_OBJECT_FACTORY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * Object factory base class.
 *
 * Object factories are responsible for creation and initialization of
 * managed objects. Use factgen.pl to generate factories.
 */
class MASSIV_EXPORTED ObjectFactory
    {
    NO_COPY_CONSTRUCTOR( ObjectFactory );
    NO_ASSIGNMENT_OPERATOR( ObjectFactory );

    friend class ObjectProperty;

public:

    /**
     * Constructor.
     */
    ObjectFactory
        (
        const MetaObject & the_metaobject
        );

    /**
     * Destructor.
     */
    virtual ~ObjectFactory();

// Generated methods.
public:

    /**
     * Create new instance of object and initialize the object and its properties.
     *
     * Use this to create objects that are new to "the world".
     */
    virtual ObjectProperty * create_object() const = 0;

    /**
     * Create new instance of object with specified object id and initialize
     * the object and its properties.
     *
     * Use this to create objects with known object id such as migrated
     * objects, objects loaded from archive, etc. Created object will be owned
     * by this node. Use #create_replica to create object replicas.
     */
    virtual ObjectProperty * create_object
        (
        const ObjectId & object_id
        ) const = 0;

    /**
     * Create new instance of object replica with specified object id
     * and initialize the object and its properties.
     */
    virtual ObjectProperty * create_replica
        (
        const ObjectId & object_id
        ) const = 0;

    /**
     * Create new instance of object and copy its properties from given @a object.
     */
    virtual ObjectProperty * clone_object
        (
        const ObjectProperty * object
        ) const = 0;

    /**
     * Throw a copy-value of @c object by value.
     */
    virtual void raise_exception
        (
        const ObjectProperty * object
        ) const = 0;

// Helper methods.
protected:

    /**
     * Initialize new object.
     */
    void initialize_object
        (
        ObjectProperty &    object
        ) const;

    /**
     * Initialize object with given @a object_id.
     */
    void initialize_object
        (
        ObjectProperty &    object,
        const ObjectId &    object_id
        ) const;

    /**
     * Initialize object replica with given @a object_id.
     */
    void initialize_replica
        (
        ObjectProperty &    object,
        const ObjectId &    object_id
        ) const;

    /**
     * Return object-\>complete_object.
     */
    VariantPointer get_complete_object
        (
        const ObjectProperty * object
        ) const;

    /**
     * Throw "Can't create instance of abstract class." error.
     */
    NORETURN( void ) _abstract_instance() const;

protected:

    const MetaObject & metaobject;
        /**< Corresponding metaobject. */

    }; // class ObjectFactory

/** @} */ // addtogroup object_management


} // namespace Massiv::Core


namespace Generated {


/** @addtogroup object_management
 * @{
 */

/**
 * Generated factories should be specializations of this class.
 *
 * Each generated factory should inherit ::Massiv::Core::ObjectFactory and
 * implement all pure virtual methods declared there.
 *
 * It should define the following static variables:
 * - <tt>const ObjectFactoryImpl< Class > * self</tt>. Pointer to the only instance
 *   of the object factory.
 */
template< typename Class >
class ObjectFactoryImpl; /* : public ::Massiv::Core::ObjectFactory
    {
    static const ObjectFactoryImpl< Class > * self
    }; */


/**
 * Name of object factory class for given object class.
 */
#define OBJECT_FACTORY( class ) ::Massiv::Generated::ObjectFactoryImpl< class >

/** @} */ // addtogroup object_management


} // namespace Massiv::Generated
} // namespace Massiv


#endif // MASSIV_OBJECT_FACTORY_H
