/** @file object_info.h
 * ObjectInfo class.
 * ObjectInfo stores various information about an instance of an object
 * used to create the object during migration, unarchivation or swap-in.
 */

/*
 * begin        : Wed Jun 11 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_INFO_H
#define MASSIV_OBJECT_INFO_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_CLASS_TYPE_ID_H
#include "object/class_type_id.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif


namespace Massiv {
namespace Core {


/**
 * ObjectInfo holds type information about an object instance needed to
 * migrate, serialize & archive the object.
 * ObjectInfo is usually stored as an optional part of ObjectId, which
 * identifies the object instance.
 * 
 * When an object is migrated, archived or swapped-out the core
 * serializes @c ObjectInfo as a part of object's @c ObjectId so that it
 * would be able to recreate it from the stream.
 *
 * @note Archivable flag specifies whether the object is archivable.
 *      By default, the archivable flag is obtained from metaobject of the object.
 *      Archivable flag may also be changed at runtime but this must be
 *      done very carefully. There are some restrictions about enabling or
 *      disabling archivable flag at runtime.
 *      Currently the archivable flag can be enabled/disabled just after
 *      the object creation, before the object is assigned to any pointer.
 * @todo FIXME Jake presne podminky na zmeny archivable flaziku?
 *
 * @ingroup objectid
 */
class MASSIV_EXPORTED ObjectInfo
    {
public:

    /**
     * Text serialization requirements.
     */
    enum TextReadRequest
        {
        TEXT_READ_IGNORE_INVALID_CLASS_TYPE_ID,
            /**< When reading #ClassTypeId from input stream, ignore
             * if invalid class type id is read (ie. if the class is
             * unknown for local node.
             */
            
        TEXT_READ_THROW_IF_INVALID_CLASS_TYPE_ID
            /**< When reading #ClassTypeId from input stream, throw
             * an exception if invalid class type id is read.
             */
        
        }; // TextReadRequest

public:

    /**
     * Constructor.
     * Archivable flag is set to true (corresponds to default value
     * of "no_archive" class attribute).
     * @warning See note above about enabling/disabling archivable flag.
     */
    INLINE ObjectInfo();
    
    /** Copy constructor. */
    INLINE ObjectInfo
        (
        const ObjectInfo & other
        );
        
    /**
     * Constructor.
     * @warning See note above about enabling/disabling archivable flag.
     */
    INLINE ObjectInfo
        (
        const ClassTypeId & class_type_id,
        bool                archivable
        );
        
    /** Operator=. */
    INLINE ObjectInfo & operator=
        (
        const ObjectInfo & other
        );
        
public:

    /** Simple getter. Get #ClassTypeId of the object. */
    INLINE const ClassTypeId & get_class_type_id() const;
    
    /**
     * Check whether class type id stored in the #ObjectInfo is supported
     * by local node.
     * Calling this function is the same as calling
     * @code
     * get_class_type_id().is_class_registered()
     * @endcode
     * See @c ClassTypeId::is_class_registered() for more info.
     */
    INLINE bool is_class_registered() const;
    
    /** Simple getter. Get archivable flag. */
    INLINE bool is_archivable() const;
    
public:
    
    /** Simple setter. Set #ClassTypeId of the object. */
    INLINE void set_class_type_id
        (
        const ClassTypeId & class_type_id
        );
        
    /**
     * Simple setter. Set archivable flag.
     * @warning See note above about enabling/disabling archivable flag.
     */
    INLINE void set_archivable_flag
        (
        bool    is_archivable
        );
        
public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     * @param text_read_request Serialization requirement. If @a text_read_request
     *      is @c TEXT_READ_IGNORE_INVALID_CLASS_TYPE_ID and the input stream
     *      contains invalid (unsupported) class type id, the #ClassTypeId part of the
     *      #ObjectInfo will be set to default value.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc,
        TextReadRequest                 text_read_request = TEXT_READ_THROW_IF_INVALID_CLASS_TYPE_ID
        );
    
private:

    ClassTypeId     class_type_id;
        /**< #ClassTypeId of the object. */
        
    SBool           archivable;
        /**< Is the object archivable?
         * @warning See note above about enabling/disabling archivable flag.
         */
    
    }; // ObjectInfo


} // namespace Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object_info_inline.h"
#endif


#endif // MASSIV_OBJECT_INFO_H
