/** @file object_manager.cpp
 * Object manager maintains object providers and list of all known local
 * and replicated objects.
 */

/*
 * begin        : Wed Feb 20 2002
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "serializer/serializer.h"
#include "utility/debug.h"
#include "object/object.h"
#include "object/object_manager.h"
#include "object/object_provider.h"
#include "pointer/object_pointer.h"
#include "pointer/pointer_to_object.h"
#include "pointer/pointer_wrappers.h"
#include "pointer/garbage_collector.h"
#include "object/schedule_impl.h"
#include "utility/cache_impl.h"
#include "sysevent/object_status_msg.h"
#include "sysevent/migration_group_msg.h"
#include "object/event.h"
#include "object/archive.h"
#include "object/archive_manager.h"
#include "object/collect_group.h"
#include "time/time_manager.h"
#include "system/system.h"
#include "database/node_database.h"
#include "replication/replica_server.h"
#include "replication/replica_client.h"
#include "object/object_provider_state.h"
#include "sysevent/simulation_sync_msg.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "serializer/stype_int.h"
#include "filesys/volume_manager.h"
#include "filesys/file_system.h"
#include "filesys/file_iterator.h"
#include "object/object_factory.h"
#include "system/tick_lock.h"
#include <memory>
#include <sstream>


namespace Massiv {
namespace Core {


/*
 * List of caches and queues in object manager:
 * 1) Providers' caches.
 *    ObjectId => NodeId & object status
 * 2) Migration cache.
 *    ObjectId
 * 3) Waiting objects.
 * 4) Groups waiting for.
 * 5) Groups to process.
 * [ 6) Replicas ]
 *
 */

/*
 * POZOR - BEHEM ARCHIVACE NESMI DOJIT K ZADNEMU
 * ODSWAPOVANI OBJEKTU NA DISK!!!!
 * Jinak muze dojit k jeho archivaci, pokud se bude
 * naswapovavat, a to i kdyz do archivu nema patrit!!!!!
 */

/* @note Pokud je objekt ve swapu, NESMI se jakkoli vytvorit jeho
 * replika.
 * Replika objektu se nesmi vytvorit:
 * - Pokud je objekt ve swapu.
 * Muze se vytvorit je-li objekt waiting. Pri deliver() waiting objektu
 * se jeho replika znici.
 *
 * - Repliky nelze swapovat.
 *
 */

#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

/** @todo FIXME Temporary hack.
 * Time how long information stays in object migration cache.
 */
const UInt64    MIGRATION_OBJECT_CACHE_TIMEOUT = 10;

#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE


namespace {

/**
 * Deletes items from a @c container.
 * Assumes that stored items are pointers to objects owned by the container.
 */
template< typename T >
void clear_container
    (
    T & container
    )
    {
    typename T::iterator it( container.begin() );
    while( it != container.end() )
        {
        delete it->second;
        ++it;
        }
    container.clear();
    }

}
    

// ObjectManagerStatistics implementation.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectManager::ObjectManagerStatistics::ObjectManagerStatistics() :
    Statistics( "ObjectManagement", Status::FACILITY_OBJECT_MANAGEMENT, 
        Statistics::OPEN_NORMAL ),
    num_ticks( create_integer( "num_ticks" ) ),
    num_total_events( create_integer( "num_total_events" ) ),
    num_successful_events( create_integer( "num_successful_events" ) ),
    num_failed_events( create_integer( "num_failed_events" ) ),
    num_forwarded_events( create_integer( "num_forwarded_events" ) ),
    num_debug_random_local_migrations( create_integer( "num_debug_random_local_migrations" ) ),
    num_swapped_in_objects( create_integer( "num_swapped_in_objects" ) ),
    num_swapped_out_objects( create_integer( "num_swapped_out_objects" ) ),
    num_unpacked_objects( create_integer( "num_unpacked_objects" ) ),
    num_packed_objects( create_integer( "num_packed_objects" ) ),
    schedule_blocked( create_boolean( "schedule_blocked" ) ),
    too_busy( create_boolean( "too_busy" ) ),
    level_zero_time( create_float( "level_zero_time" ) )
    {
    }


// ObjectManagerSettings implementation.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectManager::ObjectManagerSettings::ObjectManagerSettings() :
    Settings( "ObjectManagement" ),
    debug_force_random_local_migration( access_boolean( "debug_force_random_local_migration", false ) ),
    debug_random_local_migration_probability( access_float( "debug_random_local_migration_probability", 0 ) ),
    debug_force_swap_out_of_objects_with_scheduled_events( access_boolean( "debug_force_swap_out_of_objects_with_scheduled_events", false ) ),
    max_schedule_delay( access_float( "max_schedule_delay", 4.0f ) ),
    busy_time_limit( access_float( "busy_time_limit", 0.25f ) ),
    log_events( access_boolean( "log_events", false ) ),
    log_process_migration_messages( access_boolean( "log_process_migration_messages", false ) )
    {
    }

                       
// ObjectManager::UpcallContext definition
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * If ObjectManager needs to delay-upcall client code it can do so
 * by filling @c UpcallContext so that it could be dispatched later.
 * This invocation scheme is defined to avoid recursion inside inner
 * parts of ObjectManager (client code may call-back ObjectManager).
 * @note The context automatically pins all objects that are directly
 * involved in the call so that the pointers would be valid at upcall
 * dispatch.
 */
class ObjectManager::UpcallContext
    {
private:

    NO_COPY_CONSTRUCTOR( UpcallContext );
    NO_ASSIGNMENT_OPERATOR( UpcallContext );

public:

    /** Upcall type. */
    enum Type
        {
        UPCALL_DELIVERED_TO,    /// Object::delivered_to().
        UPCALL_DELIVERY_FAILED, /// Object::delivery_failed().
        UPCALL_UNDEFINED        /// Undefined.
        };

    /** Constructor. */
    UpcallContext() :
        type( UPCALL_UNDEFINED ),
        param_source( NULL ),
        param_destination( NULL )
        {
        }

    /** Destructor. */
    ~UpcallContext()
        {
        clear();
        }

public:

    /** Initializes UPCALL_DELIVERED_TO context. */
    void delivered_to
        (
        ObjectProperty *    source,
        ObjectProperty *    destination,
        const STime &       time
        )
        {
        hope( type == UPCALL_UNDEFINED );
        type = UPCALL_DELIVERED_TO;
        
        param_source = source;
        param_destination = destination;
        param_time = time;
        }
        
    /** Initializes UPCALL_DELIVERY_FAILED context. */
    void delivery_failed
        (
        ObjectProperty *        source,
        ObjectDeliveryFailure   failure,
        const STime &           time
        )
        {
        hope( type == UPCALL_UNDEFINED );
        type = UPCALL_DELIVERY_FAILED;
        
        param_source = source;
        param_failure = failure;
        param_time = time;
        }

public:

    /** Tests if upcall context is filled and valid. */
    bool is_valid()
        {
        return type != UPCALL_UNDEFINED;
        }

    /** Clears upcall context. */
    void clear()
        {
        type = UPCALL_UNDEFINED;

        param_source = NULL;
        param_destination = NULL;
        }
        
public:
        
    /// Upcall parameters.

    Type                    type;
    PtrToObject             param_source;
    PtrToObject             param_destination;
    ObjectDeliveryFailure   param_failure;
    STime                   param_time;
    };


// ObjectManager implementation.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectManager::ObjectManager() :
    initialized( false ),
    simulation_enabled( false ),
    active_provider( NULL ),
    virtual_provider( NULL ),
    swap( NULL ),
    replica_client( new ReplicaClient() ),
    replica_server( new ReplicaServer() )
    {
    }


ObjectManager::~ObjectManager()
    {
    if( initialized )
        {
        shutdown();
        }
    }


void ObjectManager::initialize()
    {
    hope( !initialized );

    // Initialize Object manager.
        {
        // Reset statistics.
        statistics.reset();

        // Providers.
        active_provider  = NULL;
        virtual_provider = new ObjectProvider( ObjectProviderState() );
        }
    
    // Initialize GC, ReplicaServer, ReplicaClient and Balancer.
        {
        gc.initialize();
        replica_server->initialize();
        replica_client->initialize();
        balancer.initialize();
        }

    initialized = true;
    }


void ObjectManager::shutdown()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::shutdown()" );
    hope( initialized );

    // Close swap.
        {
        delete swap;
        swap = NULL;
        }
    
    // Clear internal state.
    // Must be done before GC is shut down.
        {
        simulation_enabled = false;
        global_schedule.clear();
        
        #ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
        migration_object_cache.clear();
        #endif

        waiting_objects.clear();
        groups_waiting_for.clear();
        groups_to_process.clear();
        }

    // Destroy providers.
        {
        clear_container( providers );
        active_provider = NULL;
        remote_provider_to_node.clear();
        }

    // Destroy virtual provider.
        {
        delete virtual_provider;
        virtual_provider = NULL;
        }

    // Shutdown Balancer, ReplicaServer, ReplicaClient and GC.
        {
        balancer.shutdown();
        replica_server->shutdown();
        replica_client->shutdown();
        gc.shutdown();
        }

    initialized = false;
    }


void ObjectManager::enable_simulation()
    {
    switch( Global::get_local_node_id().get_type() )
        {
        case NodeId::TYPE_SERVER:
        case NodeId::TYPE_CLIENT:
            simulation_enabled = true;
            break;
            
        case NodeId::TYPE_SERVICE:
        case NodeId::TYPE_ANONYMOUS:
            // Unsupported.
            break;
        }
    }
    
    
void ObjectManager::disable_simulation()
    {
    simulation_enabled = false;
    }
    
    
bool ObjectManager::is_simulation_enabled() const
    {
    return simulation_enabled;
    }

    
void ObjectManager::generate_id
    (
    ObjectId &          object_id,
    bool                track,
    const ClassTypeId & class_type_id,
    bool                no_archive
    )
    {
    DEBUG_ONLY( bool r = ) active_provider->generate_id( object_id, track, class_type_id, no_archive );
    hope( r );
    }


void ObjectManager::kill_event
    (
    const ObjectId & object_id,
    const EventId &  event_id
    )
    {
    hope( !event_id.is_private() );

    ObjectProperty * o;

    // Find registered object.
    o = gc.find_local_object( object_id );

    if( !o )
        {
        // Find swapped out object.
        o = try_to_swap_in( object_id );
        }

    if( o && !o->is_replica() )
        {
        Event e;
        
        if( o->remove_event( event_id, e ) )
            {
            // Notify the object only if the event has notify flag on.
            if( e.notify_owner() )
                {
                UpcallContext upcall;

                // Upcall Object::delivery_failed().
                upcall.delivery_failed
                    (
                    o,
                    Object::EVENT_KILLED,
                    e.get_execution_time()
                    );

                dispatch_upcall( upcall );
                }
            else
                {
                if( settings.log_events )
                    {
                    log_delivery_failed( o->get_object_id(),
                        Object::EVENT_KILLED, e.get_execution_time(),
                        true );
                    }
                }
            }
        }
    }


void ObjectManager::select_active_provider()
    {
    /** @todo Maybe somebody in the future can do here some heuristic for
     * selecting provider (provider with smallest persistent cache, provider
     * with smallest relative_id, ...).
     */

    // Select first provider from the list.
    ProviderList::iterator it( providers.begin() );
    hope( it != providers.end() );
    active_provider = it->second;

    hope( active_provider != NULL );
    hope( !active_provider->is_exhausted() );
    }


bool ObjectManager::is_local_provider
    (
    const ProviderId & provider_id
    )
    {
    return ( providers.find( provider_id ) != providers.end() );
    }


bool ObjectManager::is_remote_provider
    (
    const ProviderId & provider_id
    )
    {
    return ( remote_provider_to_node.find( provider_id ) != remote_provider_to_node.end() );
    }


void ObjectManager::register_local_provider
    (
    const ObjectProviderState & state
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::register_local_provider()" );
        {
        std::ostringstream ss;
        ss  << "Registering local object provider "
            << state.get_provider_id() << ".";
            
        Global::log_info
            (
            Status::FACILITY_OBJECT_LOCALIZATION,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
    
    hope( !state.get_provider_id().is_null() );
    
    if( providers.find( state.get_provider_id() ) != providers.end() )
        {
        std::ostringstream ss;
        ss  << "Trying to register local object provider " << state.get_provider_id()
            << " two or more times. "
            << "Seems that you have corrupted archives.";
        panic( ss.str().c_str() );
        }
        
    if( remote_provider_to_node.find( state.get_provider_id() ) != remote_provider_to_node.end() )
        {
        std::ostringstream ss;
        ss  << "Trying to register local object provider " << state.get_provider_id()
            << " that is already registered as remote provider. "
            << "Seems that you have corrupted archives.";
        panic( ss.str().c_str() );
        }

    std::auto_ptr<ObjectProvider>   p( new ObjectProvider( state ) );
    providers.insert( ProviderList::value_type( state.get_provider_id(), p.get() ) );
    p.release();

    select_active_provider();
    }
    
    
void ObjectManager::enumerate_local_providers
    (
    StartupMessage & msg
    )
    {
    ProviderList::const_iterator it;
    
    msg.providers.clear();
    
    for( it = providers.begin(); it != providers.end(); ++it )
        {
        const ProviderId & provider_id = it->first;
        msg.providers.push_back( provider_id );
        }
    }


void ObjectManager::register_remote_provider
    (
    const ProviderId & provider_id,
    const NodeId &     node_id
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::register_remote_provider()" );
        {
        std::ostringstream ss;
        ss  << "Registering remote object provider "
            << provider_id << " located on node "
            << node_id << ".";
            
        Global::log_info
            (
            Status::FACILITY_OBJECT_LOCALIZATION,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
    
    hope( !provider_id.is_null() );
    hope( !node_id.is_null() );
    hope( node_id != Global::get_local_node_id() );
    hope( node_id.is_type( NodeId::TYPE_SERVER ) );
    
    if( providers.find( provider_id ) != providers.end() )
        {
        std::ostringstream ss;
        ss  << "Trying to register remote object provider " << provider_id
            << " but the provider is already registered as local provider. "
            << "Seems that you have corrupted archives.";
        panic( ss.str().c_str() );
        }
    
    if( remote_provider_to_node.find( provider_id ) != remote_provider_to_node.end() )
        {
        std::ostringstream ss;
        ss  << "Trying to register remote object provider " << provider_id
            << " but the provider is already registered as remote provider. "
            << "Seems that you have corrupted archives.";
        panic( ss.str().c_str() );
        }

    remote_provider_to_node.insert
        (
        RemoteProviderToNodeMap::value_type
            (
            provider_id,
            node_id
            )
        );
    }


NodeId ObjectManager::get_provider_owner_node
    (
    const ProviderId & provider_id,
    bool               include_local_providers
    )
    {
    NodeId           null_node_id;
    ObjectProvider * p;
    RemoteProviderToNodeMap::const_iterator it;

    if( provider_id.is_null() )
        {
        return null_node_id;
        }

    p = get_provider( provider_id );

    if( p != virtual_provider ) // If the provider is local...
        {
        if( include_local_providers )
            {
            return Global::get_local_node_id();
            }
        else
            {
            return null_node_id;
            }
        }

    it = remote_provider_to_node.find( provider_id );
    if( it != remote_provider_to_node.end() ) // If the provider is remote...
        {
        return it->second;
        }
    else
        {
        return null_node_id;
        }
    }


void ObjectManager::global_notify_of_object_destruction
    (
    const ObjectId & object_id
    )
    {
    ObjectStatusMessage       status_msg
        (
        object_id,                      // Object id.
        ObjectStatus::NOT_ACCESSIBLE,   // Object was destroyed.
        Global::get_local_node_id(),    // Local node id.
        true                            // We have destroyed the object so mark the message as trustworthy.
        );

    // Update caches of local providers. Id the provider is remote then
    // cache of virtual provider is updated.
    update_remote_object_status_cache( status_msg );

    // Notify remote provider if the object is tracked.
    if( object_id.is_tracked_by_provider() )
        {
        NodeId    remote_sid( get_provider_owner_node( object_id.get_provider_id(), false ) );
        // Note: If the provider is local then remote_sid is null now!
        if( !remote_sid.is_null() )
            {
            status_msg.send( remote_sid, NetworkCommon::STREAM_RELIABLE );
            }
        }
    }


void ObjectManager::notify_of_local_schedule_change
    (
    ObjectProperty *    object
    )
    {
    hope( object != NULL );

    // Check whether the object is already in the global schedule.
    if( object->root_data->global_schedule_index.is_valid() )
        {
        // The object is in the global schedule.
        hope( global_schedule[ object->root_data->global_schedule_index ] == object );
        if( object->has_scheduled_event() )
            {
            // The object still has some scheduled events.
            hope( object->is_gc_root_object() );
            STime t = object->get_time_of_first_event();
            if( !t.is_equal_to( global_schedule.get_key( object->root_data->global_schedule_index ) ) )
                {
                global_schedule.change_key( object->root_data->global_schedule_index, t );
                }
            }
        else
            {
            // The object has no event.
            // If we have dynamically marked the object as gc root object,
            // convert the object back to normal.
            gc.make_gc_root( object, false );
            hope( !object->is_gc_root_object() || object->is_permanent_gc_root_object() );

            global_schedule.pop( object->root_data->global_schedule_index );
            hope( !object->root_data->global_schedule_index.is_valid() );
            }
        }
    else
        {
        // The object is not in the global schedule.
        if( object->has_scheduled_event() )
            {
            // The object has at least one event.
            // Make the object gc root object until all events are executed.
            gc.make_gc_root( object, true );
            global_schedule.push( object->get_time_of_first_event(), object );
            }
        else
            {
            // The object is not in the global schedule
            // and it has no event.
            // Strange but it is not fobidden.
            // In current implementation it can only happen if the
            // object has just been unpacked from a stream.
            hope( !object->is_gc_root_object() || object->is_permanent_gc_root_object() );
            }
        }
    }


void ObjectManager::remove_object_from_global_schedule
    (
    ObjectProperty * object
    )
    {
    hope( object->is_registered() );

    if( object->root_data->global_schedule_index.is_valid() )
        {
        // If we have dynamically marked the object as gc root object,
        // convert the object back to normal.
        gc.make_gc_root( object, false );

        // And remove the object from the global schedule.
        hope( global_schedule[ object->root_data->global_schedule_index ] == object );
        global_schedule.pop( object->root_data->global_schedule_index );
        }
    }


bool ObjectManager::is_object_waiting
    (
    const ObjectId & id
    ) const
    {
    return waiting_objects.find( id ) != waiting_objects.end();
    }


    //
    // Swap.
    //

void ObjectManager::open_swap
    (
    const std::string & archive_file
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::open_swap()" );
    DEBUG_ONLY( ArchiveManager & archive_manager = Global::archive_manager(); )

    hope( !archive_manager.archivation_in_progress() );
    hope( swap == NULL );
    hope( providers.empty() );

    // Log.
        {
        std::ostringstream ss;
        ss  << "Opening swap file \"" << archive_file << "\".";
        Global::log
            (
            STATUS_INFO
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );
        }

    swap = new Archive( archive_file, Archive::MODE_READ_WRITE );
    swap->enable_report( Archive::REPORT_INVALIDATE );

    restore_simulation_state( swap );
    notify_objects_in_swap_of_simulation_startup();
    }


void ObjectManager::create_empty_swap
    (
    const std::string & file_name
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::create_empty_swap()" );
    DEBUG_ONLY( ArchiveManager & archive_manager = Global::archive_manager(); )

    hope( !archive_manager.archivation_in_progress() );
    hope( swap == NULL );

    std::string volume_name = "Swap file";

    swap = new Archive( file_name, volume_name );
    swap->enable_report( Archive::REPORT_ALL );
    }
    

void ObjectManager::close_swap()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::close_swap()" );
    ArchiveManager & archive_manager = Global::archive_manager();

    if( !swap )
        {
        return;
        }

    // Closing swap file during archivation is not supported now.
    // To make it work during archivation, the archive manager
    // must be updated.
    if( archive_manager.archivation_in_progress() )
        {
        std::ostringstream ss;
        ss  << "Trying to close swap file during archivation."
            << " Swap file can't be closed while archivation is running."
            << " The request was ignored.";

        Global::log
            (
            STATUS_WARNING
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );

        return;
        }

    // Ok, right now swap file is opened and archivation is not
    // in progress.
    // Swap-in all objects stored in the swap file.
    swap_in_all_objects();

    // Close the swap file.
    // We do not validate the swap, because there is no reason to do that :)
    // From now, swap is just only another file that can be deleted.
    delete swap;
    swap = NULL;
    }


    //
    // Archivation.
    //


void ObjectManager::prepare_for_archivation()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::prepare_for_archivation()" );
    
    // Move all waiting migration messages to groups-to-process queue.
    refresh_all_waiting_migration_messages();
    hope( groups_waiting_for.empty() );

    // Create all objects from migration groups stored in groups-to-process queue.
    deliver_objects_from_migration_messages();

    hope( groups_waiting_for.empty() );
    hope( groups_to_process.empty() );
    hope( waiting_objects.empty() );
    }


void ObjectManager::archive_simulation_state
    (
    Archive * archive
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::archive_simulation_state()" );
    hope( archive );

    Archive::OutputStream   output_stream;
    std::string             stream_name = "simulation_state";
    output_stream.open( archive, stream_name );

    /** @todo FIXME Exceptions. */
    hope( !output_stream.fail() );

    TextWriter              tw( output_stream );
    Serializer::Description desc;   /** @todo FIXME Initialize. */

    // Store current simulation time.
    tw.line() << "simulation_time";
    tw.write_space();
    
        {
        STime   current_time = Massiv::System::time();
        current_time.text_write( tw, desc );
        }
    
    tw.end_line();

    // Store all local providers.
        {
        ProviderList::const_iterator    it;
        SUInt32                         count( providers.size() );
        ObjectProviderState             provider_state;

        // Write provider count.
        tw.line() << "object_provider_count";
        tw.write_space();
        count.text_write( tw, desc );

        // Write states of all providers..
        for( it = providers.begin(); it != providers.end(); ++it )
            {
            const ObjectProvider * provider = it->second;
            provider->get_state( provider_state );
            
            tw.end_line();
            tw.line() << "object_provider";
            tw.write_space();
            provider_state.text_write( tw, desc );
            }
        
        tw.end_line();
        }
    }


void ObjectManager::restore_simulation_state
    (
    Archive * archive
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::restore_simulation_state( Archive * )" );
    hope( archive );

    /** @todo FIXME Better asserts - no simulation in progress, ... */
    hope( providers.empty() );

    Archive::InputStream        input_stream;
    std::string                 stream_name = "simulation_state";
    input_stream.open( archive, stream_name );

    /** @todo FIXME Exceptions. */
    hope( !input_stream.fail() );
    
    restore_simulation_state( input_stream );
    }


void ObjectManager::restore_simulation_state
    (
    std::istream & input_stream
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::restore_simulation_state( std::istream )" );
    // This function is also called when merging unpacked archives
    // so there may be already some providers and remote providers
    // registered.

    /** @todo FIXME Better asserts - no simulation in progress, ... */

    TimeManager &           time_manager = Global::time_manager();
    TextReader              tr( input_stream );
    Serializer::Description desc;   /** @todo FIXME Initialize. */

    tr.next_line_throw();

    // Read current simulation time.
    tr.require_string( "simulation_time" );
        {
        STime   current_time;
        current_time.text_read( tr, desc );

        time_manager.set_time( current_time );
        }

    // Read all local providers.
        {
        SUInt32                         count;
        ObjectProviderState             provider_state;

        // Read provider count.
        tr.next_line_throw();
        tr.require_string( "object_provider_count" );
        count.text_read( tr, desc );

        hope( count >= 0 );

        // Read states of all providers.
        for( ; count > 0; count-- )
            {
            tr.next_line_throw();
            tr.require_string( "object_provider" );
            provider_state.text_read( tr, desc );
            register_local_provider( provider_state );
            }
        }

    hope( !input_stream.fail() );
    }

    
// Anonymous namespace for load_unpacked_archive().
namespace {
    
    struct UnpackedArchive
        {
        typedef std::vector< std::string >  NameList;
        typedef ObjectShared::ObjectIdList  ObjectIdList;
        typedef ObjectShared::ObjectList    ObjectList;
        
        std::string     directory;
            /**< Directory where the unpacked archive is stored.
             * Normalized, terminated by '/' character.
             */
        NameList        groups;
            /**< List of group directories in the archive.
             * Each string contains path to the group directory.
             * Each path is normalized, terminated by '/' character.
             */
        NameList        object_stream_names;
            /**< List of all object stream names.
             * Each string contains path to the object stream.
             * Each path is normalized, terminated by '/' character.
             */
        ObjectIdList    object_ids;
        ObjectList      objects;
        };
        
        
ObjectId load_unpacked_archive_read_object_id
    (
    const std::string & file_name
    )
    {
    VolumeManager::IFStream file( file_name.c_str() );
    
    if( file.fail() )
        {
        std::ostringstream oss;
        oss << "Unable to open object stream file '" << file_name << "'.";
        throw ExIOError( oss.str() );
        }
    
    TextReader  tr( file );
    ObjectId    object_id;
    
    tr.next_line();
    object_id.text_read( tr, Serializer::Description() );
    
    return object_id;
    }


} // anonymous namespace


void ObjectManager::load_unpacked_archive
    (
    const std::vector< std::string > & directories
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::load_unpacked_archive()" );
    
    hope( directories.size() > 0 );

    typedef UnpackedArchive::NameList       NameList;
    typedef UnpackedArchive::ObjectIdList   ObjectIdList;
    typedef UnpackedArchive::ObjectList     ObjectList;
    
    typedef std::vector< UnpackedArchive >  Archives;
    
    FileSystem &    nfs = Global::native_file_system();

    Archives        archives;   // List of unpacked archives to load into memory.
    std::size_t     num_directories = directories.size();
    std::size_t     i;
    
    archives.resize( num_directories );
    
    // Enumerate all directories.
    for( i = 0; i < num_directories; ++i )
        {
        std::string directory = directories[ i ];
        
        // Normalize the path.
        nfs.normalize_path( directory );
        if( directory.empty() )
            {
            directory = ".";
            directory += nfs.get_slash();
            }
        else if( directory[ directory.size() - 1 ] != nfs.get_slash() )
            {
            directory = directory + nfs.get_slash();
            }
        
        // Check if one directory is specified two or more times.
        Archives::const_iterator it;
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            const UnpackedArchive & archive = *it;
            if( archive.directory == directory )
                {
                std::ostringstream oss;
                oss << "Archive packing: Directory '" << directory << "' specified two or more times!";
                panic( oss.str().c_str() );
                }
            }
        
        archives[ i ].directory = directory;
        }

    // Log.
        {
        std::ostringstream ss;
        ss  << "Loading " << num_directories << " unpacked archives into memory from directories: ";
        for( i = 0; i < num_directories; ++i )
            {
            ss << '\n' << archives[ i ].directory;
            }

        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
    
    // Enumerate groups from each unpacked archive.
        {
        Archives::iterator it;
        
        // For each archive...
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            UnpackedArchive & archive = *it;
            FileIterator fs_it( nfs.open_directory( archive.directory, "g*" ) );
            
            // For each "g*" directory from the archive directory...
            for( ; fs_it.is_valid(); fs_it++ )
                {
                if( fs_it.is_directory() )
                    {
                    std::string path = archive.directory + fs_it.name();
                    
                    // Normalize the path.
                    nfs.normalize_path( path );
                    if( path.empty() )
                        {
                        path = ".";
                        path += nfs.get_slash();
                        }
                    else if( path[ path.size() - 1 ] != nfs.get_slash() )
                        {
                        path = path + nfs.get_slash();
                        }
                    
                    archive.groups.push_back( path );
                    }
                }
            }
        }
        
    // Read all objects from all groups.
        {
        Archives::iterator it;
        
        // For each archive...
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            UnpackedArchive & archive = *it;
            NameList::const_iterator it;
            
            // For each group from this archive...
            for( it = archive.groups.begin(); it != archive.groups.end(); ++it )
                {
                const std::string & group_directory = *it;
                
                // Log.
                    {
                    std::ostringstream ss;
                    ss  << "Reading group directory '" << group_directory << "'.";

                    Global::log_debug
                        (
                        Status::FACILITY_OBJECT_MANAGEMENT,
                        Status::PRIORITY_NORMAL,
                        ss.str()
                        );
                    }
                
                // Enumerate all object streams from the group directory...
                FileIterator fs_it( nfs.open_directory( group_directory, "o*" ) );
            
                // For each "o*" file from the group directory...
                for( ; fs_it.is_valid(); fs_it++ )
                    {
                    if( fs_it.is_file() )
                        {
                        std::string object_stream_name = group_directory + fs_it.name();
                        
                        // Register the object.
                        archive.object_stream_names.push_back( object_stream_name );
                        
                        // Read object id of the object.
                        ObjectId object_id = load_unpacked_archive_read_object_id( object_stream_name );
                        
                        archive.object_ids.push_back( object_id );
                        
                        hope( archive.object_stream_names.size() == archive.object_ids.size() );
                        }
                    }
                }
                
            // Log.
                {
                std::ostringstream ss;
                ss  << "Unpacked archive '" << archive.directory << "' contains "
                    << archive.object_ids.size() << " objects.";

                Global::log_debug
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_NORMAL,
                    ss.str()
                    );
                }
            }
        }
    
    // Create all objects from all archives.
        {
        Archives::iterator it;
        
        // For each archive...
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            UnpackedArchive & archive = *it;
            ObjectIdList::const_iterator it;
            
            // For each object from the archive...
            for( it = archive.object_ids.begin(); it != archive.object_ids.end(); ++it )
                {
                const ObjectId & object_id = *it;
                ClassTypeId      class_id;
                
                if( !object_id.is_object_info_available() )
                    {
                    std::ostringstream ss;
                    ss  << "Object info not available for object " << object_id
                        << " from unpacked archive '" << archive.directory << "'.";
                    throw ExCorruptedArchive( ss.str() );
                    }
            
                class_id = object_id.get_class_type_id();
            
                if( !class_id.is_class_registered() )
                    {
                    std::ostringstream ss;
                    ss  << "Unknown class id " << class_id << " of objects "
                        << object_id << " from unpacked archive '" << archive.directory << "'.";
                    throw ExCorruptedArchive( ss.str() );
                    }

                // Create the object.
                const ObjectFactory * factory = class_id.get_factory();
                
                // TODO: Check that object with same object id doesn't exist.
                ObjectProperty * object = factory->create_object( object_id );
            
                archive.objects.push_back( object );
                }
            }
        }

    // Read all of the objects from unpacked archives.
        {
        Archives::iterator it;
        
        // For each archive...
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            UnpackedArchive & archive = *it;
            ObjectList::iterator it;
            unsigned int j;
            Serializer::Description desc;
            desc.object_flags = Serializer::Description::OBJECT_SWAP;
            
            // For each object from this archive.
            for( j = 0, it = archive.objects.begin(); it != archive.objects.end(); ++it, ++j )
                {
                hope( j < archive.object_stream_names.size() );
                
                ObjectProperty * object = *it;
                ObjectId         object_id = object->get_object_id();
                ObjectId         read_object_id;
                std::string file_name = archive.object_stream_names[ j ];
                VolumeManager::IFStream file( file_name.c_str() );
                if( file.fail() )
                    {
                    std::ostringstream ss;
                    ss  << "Unable to open object file '" << file_name << "'.";
                    throw ExIOError( ss.str() );
                    }
                
                TextReader  tr( file );
                
                // Skip object id.
                tr.next_line();
                read_object_id.text_read( tr, desc );
                hope( read_object_id == object_id );
                
                tr.next_line();
                object->text_read( tr, desc );
                }
            }
        }
    
    // Ok, we have created all objects.
    // Now we must read simulation state of all archives.
    // We need two things from simulation states - current time
    // of the simulation and object providers associated with the archives.
    // Current time can be get from any archive (the time should be almost
    // same in all archives).
    // Object providers must be merged from all archives.
    
        {
        Archives::iterator it;
        
        // For each archive...
        for( it = archives.begin(); it != archives.end(); ++it )
            {
            UnpackedArchive & archive = *it;
            std::string file_name = archive.directory + "simulation_state";
            VolumeManager::IFStream file( file_name.c_str() );
            if( file.fail() )
                {
                std::ostringstream ss;
                ss  << "Unable to open simulation sate file '" << file_name << "'.";
                throw ExIOError( ss.str() );
                }
            
            // Read simulation state from the archive.
            // Note that simulation time will be overwritten but
            // previously registered object providers will not be erased.
            // So this function merges object providers for us at no cost.
            restore_simulation_state( file );
            }
        }
    }


void ObjectManager::notify_objects_in_swap_of_simulation_startup()
    {
    TAG_STACK_FRAME( "Massiv::Core::ObjectManager::notify_objects_in_swap_of_simulation_startup()" );

    hope( swap != NULL );
    
    Global::log_debug
        (
        Status::FACILITY_OBJECT_MANAGEMENT,
        Status::PRIORITY_NORMAL,
        "Notifying objects in swap of simulation startup."
        );
        
    // Note: First we will enumerate all objects with simulation_startup_notify
    // idl class attribute set to true stored in the swap.
    // Then we will swap-in all of the objects into memory and dispatch
    // the upcalls.
    // Speed of this function is not important.
    
    ObjectShared::ObjectIdQueue        objects_in_swap;
    ObjectShared::ObjectIdList  objects;
    std::vector< PtrToObject >  object_ptrs;
    
    // Enumerate all objects stored in the swap.
    swap->enumerate_registered_objects( objects_in_swap );
    
    // Enumerate objects with simulation_startup_notify attribute stored in the swap.
    while( !objects_in_swap.empty() )
        {
        ObjectId    object_id = objects_in_swap.front();
        objects_in_swap.pop();
        
        hope( object_id.is_object_info_available() );
        const ObjectInfo    object_info = object_id.get_object_info();
        hope( object_info.is_class_registered() );
        const ClassTypeId   class_type_id = object_info.get_class_type_id();
        const MetaObject *  metaobject = class_type_id.get_metaobject();
        hope( metaobject != NULL );

        if( metaobject->get_attribute_simulation_startup_notify() )
            {
            objects.push_back( object_id );
            }
        }
    
    // Log.
        {
        std::ostringstream ss;
        ss  << "Found " << objects.size()
            << " objects in the swap file to be notified of simulation startup.";

        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }
        
    // Log.
        {
        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            "Swapping-in all of the objects..."
            );
        }

    // Swap-in all of the objects into memory.
    // Note: We can't use swap_in( const ObjectShared::ObjectIdList & list )
    //      here because it requires the list to be one migration (archive) group!
        {
        ObjectShared::ObjectIdList::const_iterator it;
        for( it = objects.begin(); it != objects.end(); ++it )
            {
            const ObjectId &    object_id = *it;

            // Try to swap in the object. If the object is already
            // in memory, the try_to_swap_in() will return NULL
            // so we need to find the object asking GC.

            try_to_swap_in( object_id );

            ObjectProperty * object = gc.find_local_object( object_id );

            hope( object != NULL );
            object_ptrs.push_back( object );
            }
        }

    // Log.
        {
        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            "Upcalling objects..."
            );
        }
        
    // Upcall each object.
        {
        std::vector< PtrToObject >::iterator it;
        
        LOCK_SYSTEM();

        for( it = object_ptrs.begin(); it != object_ptrs.end(); ++it )
            {
            UpcallToObject upcall = it->get();
            
            upcall.tag_upcall
                ( "::Massiv::Core::Object::object_updated( SIMULATION_STARTUP )" );
            upcall->object_updated( Object::SIMULATION_STARTUP );
            }
        }
        
    // And finally swap-out all of the objects back to the swap file
    // because it is wasting of memory to leaving them active.
    
    // Log.
        {
        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            "Swapping-out all of the objects back to the swap..."
            );
        }
    
    // But first we have to destroy stack strong pointers before we can
    // swap-out the account objects.
    object_ptrs.clear();
    
        {
        ObjectShared::ObjectIdList::const_iterator it;
        for( it = objects.begin(); it != objects.end(); ++it )
            {
            const ObjectId & object_id = *it;
            try_to_swap_out( object_id );
            // Note: If you want sometime in the future check SwapOutResult,
            // don't forget that two account objects can be possibly in one
            // migration group (even if would be weird). In that case
            // try_to_swap_out() can return SWAP_OUT_RESULT_UNREGISTERED_OBJECT.
            }
        }
    }
    

Archive * ObjectManager::get_swap()
    {
    return swap;
    }


void ObjectManager::next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::next_tick()" );
    /** @todo FIXME
     * - Po urcitem case zmenit aktivniho providera (select_active_provider()).
     */

    // Process migration group messages that were generated between
    // next_tick()s. This includes for example messages scheduled
    // by some finished global search.
    process_migration_messages();
    
    // Advance garbage collector.
    gc.next_tick();

    // Advance balancer.
    balancer.next_tick();

    // Process migration group messages that were generated by GC.
    process_migration_messages();

    // Advance localization caches.
    advance_localization_caches();
    hope( groups_to_process.empty() );

    // Send replication updates.
    hope( replica_server.get() );
    replica_server->next_tick();

    // Process replication updates.
    hope( replica_client.get() );
    replica_client->next_tick();

    // Update statistics.
    statistics.num_ticks++;

    if( !System::inside_simulation() )
        {
        SystemTime time = SystemTime::current_time();
        
        if( statistics.level_zero_time + settings.busy_time_limit < 
            time.get_value() )
            {
            statistics.too_busy = true;
            }
        else
            {
            statistics.too_busy = false;
            }

        statistics.level_zero_time = static_cast<Float32>( time.get_value() );
        }
    }
    
    
void ObjectManager::simulation_next_tick()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::simulation_next_tick()" );
    // Note: In one tick simulation_next_tick() can be called
    // two or more times. Because of this we may not
    // count statistics.num_ticks in this function.
    
    // Process migration group messages that were left unprocessed.
    process_migration_messages();
    
    // Advance simulation.
    advance_simulation();
    
    // Process migration group messages that were left unprocessed.
    process_migration_messages();
    }


SystemTime ObjectManager::suggest_next_tick_time()
    {
    STime       current_simulation_time = Massiv::System::time();
    SystemTime  current_system_time = Massiv::System::system_time();
    STime       next_event_time = current_simulation_time;

    if( !global_schedule.is_empty() )
        {
        next_event_time = global_schedule.get_key( 0 );
        }

    if( next_event_time > current_simulation_time )
        {
        return current_system_time + next_event_time - current_simulation_time;
        }
    else
        {
        return SystemTime();
        }

    // TODO: Implement and utilize replica_server.suggest_next_tick_time().
    }


bool ObjectManager::is_idle()
    {
    STime current_simulation_time = Massiv::System::time();

    if( statistics.too_busy )
        {
        return false;
        }

    if( statistics.schedule_blocked )
        {
        return false;
        }

    if( !global_schedule.is_empty() && global_schedule.get_key( 0 ) < 
        current_simulation_time )
        {
        return false;
        }

    return true;
    }


bool ObjectManager::process_events
    (
    const STime &       curr_time,
    UpcallContext &     upcall
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::process_events()" );
    ObjectProperty *    source;
    ObjectProperty *    destination;
    Event               e;
    bool                r;

    // Assume that the schedule is not blocked. It will be later set to
    // true if a blocking is detected.
    statistics.schedule_blocked = false;

    // Clear upcall.
    upcall.clear();

    // Do we have any event to be processed now?
    while( !global_schedule.is_empty() && ( curr_time >= global_schedule.get_key( 0 ) ) )
        {
        // Get first object from global schedule.
        source = global_schedule[ 0 ];
        hope( !source->is_replica() );
        hope( !source->is_member_object() );
        hope( source->has_scheduled_event() );

        // Get first scheduled event.
        // The event is not removed from local schedule until it is fully processed.
        r = source->get_first_event( e );
        hope( r );
        hope( e.get_execution_time() == global_schedule.get_key( 0 ) );
        
        if( source->get_object_id().is_proxy_node_object_id() )
            {
            // Node objects may not migrate.
            // Report warning.
            
            // Log.
                {
                std::ostringstream ss;
                ss  << "Trying to execute event of node object! "
                    << "Node object " << source->get_object_id()
                    << ", destination object " << e.get_destination() << ". "
                    << "System will try to execute the event, but if the event "
                    << "will cause migration, the migration will fail.";
                Global::log_warning
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_HIGHEST,
                    ss.str()
                    );
                }
            
            // Try to execute the event. If the event will force
            // migration, the migration will fail.
            }

        // Look for the destination.
        // Find registered object.
        destination = gc.find_local_object( e.get_destination() );

        // Make sure the destination is not replica.
        if( destination && destination->is_replica() )
            {
            destination = NULL;
            }

        // Check if the destination is stored in swap.
        if( !destination )
            {
            // Find swapped out object.
            destination = try_to_swap_in( e.get_destination() );
            }

        hope( ( destination && !is_object_waiting( e.get_destination() ) ) || !destination );

#ifdef OPTION_DEBUG
        if( destination && settings.debug_force_random_local_migration.read() )
            {
            bool force_migration = std::rand() < (int)( RAND_MAX * settings.debug_random_local_migration_probability.read() );

            if( force_migration )
                {
                destination = NULL;
                statistics.num_debug_random_local_migrations++;
                }
            }
#endif // OPTION_DEBUG

        // Note: We don't need to test if the destination object id is proxy object id
        // for local node. If it is, on local node is NodeObject object
        // that has object id equal to proxy object id for local node.
        // Thus gc.find_local_object() will return != NULL if the NodeObject for local node
        // was already created.

        if( destination )
            {
            // Destination is local (and not replica) object is stored in the memory.
            hope( !destination->is_member_object() );

            // Process the event.
            // The event must be removed first!
            source->remove_first_event();   // This will call notify_of_schedule_change().

            statistics.num_successful_events++;
            
            // Notify the object only if the event has notify flag on.
            if( e.notify_owner() )
                {
                statistics.num_total_events++;

                // Upcall Object::delivered_to().
                upcall.delivered_to
                    (
                    source,
                    destination,
                    e.get_execution_time()
                    );
                    
                return true;
                }
            else
                {
                log_delivered_to( source->get_object_id(),
                    destination->get_object_id(), e.get_execution_time(),
                    true );
                }
            }
        else
            {
            // Destination object is not local or it is a replica or it is waiting in some migration queue.
            // Forward the source object (and its migration group) to the destination.

            ObjectStatus    o_status;
            NodeId          node_id;
            bool            dest_is_local_node_object = false;
            
            if( e.get_destination().is_proxy_node_object_id() &&
                e.get_destination().proxy_extract_node_id() == Global::get_local_node_id() )
                {
                // Destination is local node, but NodeObject was not created yet because
                // the NodeObject was not found by gc.find_local_object().
                dest_is_local_node_object = true;
                }

            // Check whether the destination object was destroyed.
            // If destination is proxy object id for some node and the
            // node is unknown for local node get_remote_object_status()
            // returns NOT_ACCESSIBLE.
            get_remote_object_status( e.get_destination(), o_status, node_id );
            if( o_status == ObjectStatus::NOT_ACCESSIBLE ||         // Destination is not accessible.
                dest_is_local_node_object ||                        // Destination is local node object (that doesn't exists).
                source->get_object_id().is_proxy_node_object_id()   // Source is node object, that is trying to migrate.
                )
                {
                // Notify the source object that it can't be delivered.
                // The event must be removed first!
                source->remove_first_event();   // This will call notify_of_schedule_change().

                statistics.num_failed_events++;
                
                // Notify the object only if the event has notify flag on.
                if( e.notify_owner() )
                    {
                    statistics.num_total_events++;
                
                    // Upcall Object::delivery_failed().
                    upcall.delivery_failed
                        (
                        source,
                        Object::DESTINATION_NOT_ACCESSIBLE,
                        e.get_execution_time()
                        );
                        
                    return true;
                    }
                else
                    {
                    log_delivery_failed( source->get_object_id(),
                        Object::DESTINATION_NOT_ACCESSIBLE, e.get_execution_time(),
                        true );
                    }
                }
            else
                {
                MigrationPackResult    result;
                
                // Note: Because we will directly process the message, we _MUST_
                // process all waiting groups _NOW_ (before destruction
                // of the objects from the message) !!!
                process_migration_messages();

                // Migration message.
                MigrationGroupMessage msg = MigrationGroupMessage
                    (
                    FindObjectMessage
                        (
                        e.get_destination(),            // Destination object.
                        0,                              // Hops count. Must be zero.
                        0                               // Provider hops count. Must be zero.
                        )
                    );
                
                // If the objects we are going to pack into the message are
                // nonarchivable, convert the message to nonarchivable migration
                // group message. This is required by sync RPC protocol because
                // migration messages containing sync RPC objects may not be
                // paused as a result delayed archivation.
                // Note: If source object is nonarchivable, then all objects
                // from its migration group are nonarchivable.
                if( !source->is_archivable() )
                    {
                    msg.make_nonarchivable();
                    }

                // Enumerate and serialize objects from the migration group.
                result = pack_migration_group_to_migration_message( source, msg );
                
                // Warning: source object and all objects from its migration group
                // was destroyed if the pack was successfull.
                
                switch( result )
                    {
                    case MIGRATION_PACK_RESULT_SUCCESS:
                        {
                        /* Directly process the message. Scheduling the message to be
                         * processed ASAP using add_migration_message_to_process_queue()
                         * would be only waste of time and memory.
                         */
                         
                        // Warning: source object was destroyed.
                        source = NULL;
                         
                        hope( groups_to_process.empty() ); // !!! See note above.
                        process_migration_group_message( msg );

                        statistics.num_forwarded_events++;

                        break;
                        }

                    case MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED:
                    case MIGRATION_PACK_RESULT_OBJECT_MAY_NOT_MIGRATE:
                        {
                        /* One of the objects is "busy" or can not migrate.
                           In order to not break global event ordering delay
                           the event (migration). Events scheduled after this
                           event will not be delivered until this event is
                           successfully dispatched (at next tick ObjectManager
                           will retry to dispatch it). */

                        if( source->get_time_of_first_event() == STime::asap() )
                            {
                            /* @c source was scheduled to be delivered ASAP.
                               Since it can not be delivered now, its local
                               schedule must be delayed by an epsilon time
                               so that another ASAP events will not be
                               blocked.
                               This is required by SRPC. */ 

                            STime epsilon( 0.00001f );
                            source->delay_schedule_processing( epsilon );
                            }

                        /* Issue a warning if @c source blocks global event schedule
                           too long. */

                        #ifdef OPTION_DEBUG

                        SystemTime      time = System::system_time();
                        SystemTime &    schedule_block_time = 
                            source->root_data->schedule_block_time;

                        if( schedule_block_time == SystemTime() )
                            {
                            /* This is the first time the @c source's current first
                               event has blocked the schedule. */

                            schedule_block_time = time;
                            }
                        else
                        if( schedule_block_time + settings.max_schedule_delay.read() < time )
                            {
                            std::ostringstream  ss;

                            ss  << "Event dispatching has been blocked by "
                                << source->get_object_id() << " for at least "
                                << time - schedule_block_time << " second(s) because "
                                << "the object could not migrate to destination object "
                                << e.get_destination() << ". ";

                            if( result == MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED )
                                {
                                ss  << "The schedule will be blocked until its "
                                    << "migration group is unpinned and unreferenced "
                                    << "from the stack. ";
                                }
                            else
                                {
                                ss  << "Migration group contains a no-migrate object. ";
                                }

                            ss  << "ASAP events will pass-through."
                                << "\nIf the migration group contains a RPC object there "
                                << "is a risk of deadlock!";

                            #if 0

                            ss  << "\nUpcalled objects (this thread only):\n"
                                << Global::garbage_collector().print_upcalled_objects()
                                << "\nPinned objects:\n";
                            
                            Global::garbage_collector().dump_pinned_objects( ss );

                            #endif

                            Global::log_warning
                                (
                                Status::FACILITY_OBJECT_MANAGEMENT,
                                Status::PRIORITY_HIGHEST,
                                ss.str()
                                );
                            }

                        #endif

                        /* Block the current and later events. */
                        
                        statistics.schedule_blocked = true;
                        return false;
                        }
                    
                    default:
                        panic( "Invalid value of MigrationResult!" );
                    }
                } // if( o_status == ObjectStatus::NOT_ACCESSIBLE ) else
            } // if( destination ) else

        statistics.num_total_events++;
        }

    return false;
    }


void ObjectManager::dispatch_upcall
    (
    UpcallContext & upcall
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::dispatch_upcall()" );
    
    UpcallToObject      call = upcall.param_source.get();
    bool                object_was_active = call->is_active_pinned();
    
    #ifdef GC_TAG_STACK

    std::string         tag;

    #endif

    // Upcall object.

    switch( upcall.type )
        {
        case UpcallContext::UPCALL_DELIVERED_TO:
            {
            // Create strong pointers to involved objects.
            // This is to keep alive not only the two objects but also the
            // objects reachable from these objects by strong pointers.

            Pointer<Object> source_strong = upcall.param_source;
            Pointer<Object> destination_strong = upcall.param_destination;
            
            // Touch involved objects.

            upcall.param_source->touch_object();
            upcall.param_destination->touch_object();
            
            // Upcall the object.

            if( settings.log_events )
                {
                log_delivered_to( upcall.param_source->get_object_id(),
                    upcall.param_destination->get_object_id(),
                    upcall.param_time
                    );
                }

            #ifdef GC_TAG_STACK
            
            {
            std::ostringstream ss;
            ss  << "::Massiv::Core::Object::delivered_to( "
                << upcall.param_destination->get_object_id() << ", "
                << upcall.param_time << " )";
            tag = ss.str();
            }

            call.tag_upcall( tag.c_str() );

            #endif

            call->delivered_to
                (
                upcall.param_destination,
                upcall.param_time
                );

            break;
            }

        case UpcallContext::UPCALL_DELIVERY_FAILED:
            {
            // Create strong pointers to involved objects.

            Pointer<Object> source_strong = upcall.param_source;

            // Touch involved objects.

            upcall.param_source->touch_object();
            
            if( settings.log_events )
                {
                log_delivery_failed( upcall.param_source->get_object_id(),
                    upcall.param_failure, upcall.param_time
                    );
                }

            #ifdef GC_TAG_STACK

            {
            std::ostringstream ss;
            ss  << "::Massiv::Core::Object::delivery_failed( "
                << upcall.param_failure << ", " << upcall.param_time << " )";
            tag = ss.str();
            }
            
            call.tag_upcall( tag.c_str() );

            #endif

            call->delivery_failed
                (
                upcall.param_failure,
                upcall.param_time
                );

            break;
            }

        default:
            panic( "Invalid UpcallContext." );
        }

    // If object had not been active before the upcall was dispatched, it
    // should not be active now too.
    // This allows to call a->migrate_to() from a->delivered_to() and let the
    // migration take place immediatelly.
    // This is required by SRPC (replies won't be blocked).

    if( !object_was_active )
        {
        call->ensure_active_unpinned();
        }

    // Clear context.
    
    upcall.clear();
    }


void ObjectManager::advance_simulation()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::advance_simulation()" );
    hope( groups_to_process.empty() );
    
    if( !simulation_enabled )
        {
        return;
        }

    STime           curr_time = Massiv::System::time();
    UpcallContext   upcall;
        
#ifdef OPTION_DEBUG
    {//sob
        if( settings.debug_force_swap_out_of_objects_with_scheduled_events )
            {
            // Swap-out all objects which have scheduled one or more events
            // for current time.
            while( !global_schedule.is_empty() && ( curr_time >= global_schedule.get_key( 0 ) ) )
                {
                ObjectProperty *    object = global_schedule[ 0 ];
                ObjectId            object_id = object->get_object_id();
                std::ostringstream  ss;
                SwapOutResult       result;

                result = swap_out_object( object );

                switch( result )
                    {
                    case SWAP_OUT_RESULT_SUCCESS:
                        ss  << "Object " << object_id << " successfully swapped-out.";
                        break;
                        
                    case SWAP_OUT_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because some objects from its migration group are pinned "
                            << "or stack referenced.";
                        break;
                        
                    case SWAP_OUT_RESULT_ARCHIVATION_IN_PROGRESS:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because archivation is in progress.";
                        break;
                        
                    case SWAP_OUT_RESULT_NO_SWAP_FILE:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because no swap file is opened.";
                        break;
                        
                    case SWAP_OUT_RESULT_IS_REPLICA:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because the object is replica.";
                        break;

                    case SWAP_OUT_RESULT_HAS_REPLICA_REQUESTS:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because the object is replicated to different nodes.";
                        break;
                    
                    case SWAP_OUT_RESULT_UNREGISTERED_OBJECT:
                        ss  << "Failed to swap-out object " << object_id << " "
                            << "because the object is not registered.";
                        break;
                        
                    case SWAP_OUT_RESULT_OBJECT_NOT_ARCHIVABLE:
                        ss  << "Failed to swap-out object "<< object_id << " "
                            << "because the object is not archivable.";
                        break;
                    };

                Global::log
                    (
                    STATUS_DEBUG
                        (
                        Status::FACILITY_OBJECT_MANAGEMENT,
                        Status::PRIORITY_NORMAL
                        ),
                    ss.str()
                    );

                if( result != SWAP_OUT_RESULT_SUCCESS )
                    {
                    // We can't swap-out more objects in current tick.
                    break;
                    }
                }
            }
    }//eob
#endif // OPTION_DEBUG

    // Swap in all objects that have some events scheduled for current tick.
    swap_in_objects( curr_time );

    while( true )
        {
        // Process events.
        process_events( curr_time, upcall );

        // If no upcall should be handled @c process_events()
        // finished its work completely.
        if( !upcall.is_valid() )
            {
            break;
            }
        
        // Dispatch the upcall.
        dispatch_upcall( upcall );
        }
    }


void ObjectManager::process_migration_messages()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::process_migration_messages()" );
    MigrationGroupMessage p;

    /* Get migration message if any is prepared.
     * Objects from the message are automatically
     * removed from waiting_objects set. */
    while( get_next_migration_group_to_process( p ) )
        {
        /* Process the message. */
        process_migration_group_message( p );
        }
    }


void ObjectManager::advance_localization_caches()
    {
    ProviderList::iterator  it;

#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
    // Advance migration cache.
    migration_object_cache.advance( Global::system_time() );
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

    // Advance localization caches of local providers.
    for( it = providers.begin(); it != providers.end(); it++ )
        {
        it->second->advance_cache();
        }

    // Advance localization cache of virtual provider.
    hope( virtual_provider != NULL );
    virtual_provider->advance_cache();
    }


void ObjectManager::update_find_object_message_for_sending
    (
    FindObjectMessage & p
    )
    {
    hope( (int) FindObjectMessage::HOPS_MAX_VALUE >= (int) MAX_FIND_OBJECT_MESSAGE_HOPS_COUNT );

    /* Check whether this is initial message send. In that case
     * refresh cached status (if any) of the object we are looking for. */
    if( p.hops == 0 )
        {
        refresh_remote_object_status_cache_timeout( p.destination );
        }

    if( p.hops < MAX_FIND_OBJECT_MESSAGE_HOPS_COUNT )
        {
        p.hops++;
        }
    else
        {
        /* Should never happen.
         * Reset hops count. We do not set it to zero because
         * zero is reserved for initial message send. */

        std::ostringstream ss;
        ss << "Internal error - Maximum object localization message hops count reached. "
           << "Should never happen. "
           << "Hops value: " << p.hops << ".";
           /** @todo Write message content? */

        Global::log
            (
            STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
            ss.str()
            );

        p.hops = 1;
        }
    }


void ObjectManager::add_migration_message_to_waiting_queue
    (
    const MigrationGroupMessage & p
    )
    {
    /* Add the message to waiting queue. */
    ObjectIdToMigrationGroupMultiMap::iterator it = groups_waiting_for.insert
        (
        ObjectIdToMigrationGroupMultiMap::value_type
            (
            p.destination,
            p
            )
        );

    try
        {
        /* Store information that all objects from the message are waiting. */
        
        /** @todo If an exception is throws from waiting_objects.insert( object_id ),
         * previous object ids stays in the waiting_objects.
         * Shoudl we first store all objects ids in a list, and then just use
         * waiting_objects.insert( list.begin(), list.end() )?
         */
        ObjectShared::ObjectIdList::const_iterator object_id_it;
        for( object_id_it = p.objects.begin();
             object_id_it != p.objects.end();
             ++object_id_it )
            {
            const ObjectId &            object_id = *object_id_it;
            waiting_objects.insert( object_id );
            }
        }
    catch( ... )
        {
        groups_waiting_for.erase( it );
        throw;
        }
    }


void ObjectManager::add_migration_message_to_process_queue
    (
    const MigrationGroupMessage & p
    )
    {
    /* Add the message to process queue. */
    groups_to_process.push_back( p );

    try
        {
        /* Store information that all objects from the message are waiting. */
        
        /** @todo If an exception is throws from waiting_objects.insert( object_id ),
         * previous object ids stays in the waiting_objects.
         * Shoudl we first store all objects ids in a list, and then just use
         * waiting_objects.insert( list.begin(), list.end() )?
         */
        ObjectShared::ObjectIdList::const_iterator object_id_it;
        for( object_id_it = p.objects.begin();
             object_id_it != p.objects.end();
             ++object_id_it )
            {
            const ObjectId &            object_id = *object_id_it;
            waiting_objects.insert( object_id );
            }
        }
    catch( ... )
        {
        groups_to_process.pop_back();
        throw;
        }
    }


void ObjectManager::_move_waiting_migration_messages_to_process_queue
    (
    ObjectIdToMigrationGroupMultiMap::iterator  first,
    ObjectIdToMigrationGroupMultiMap::iterator  last,
    size_t                                      count
    )
    {
    /* Move selected messages from waiting queue (groups_waiting_for)
     * to message-to-process queue (groups_to_process).
     * Note: Because ObjectIdToMigrationGroupMultiMap::value_type
     * is not same as type MigrationGroupMessage, we can't use simple
     * groups_to_process.insert( groups_to_process.end(), range.first, range.second ).
     *
     * Note: We do not use add_migration_message_to_process_queue() because
     *  we know that the objects in the messages are already marked as waiting
     *  and because this is much faster.
     */
    size_t  old_groups_to_process_size = groups_to_process.size();
    groups_to_process.resize( old_groups_to_process_size + count );
    try
        {
        // Boovie: Maybe std::transform will not throw exception
        // if MigrationGroup copy constructor and operator=
        // are both throw(). In that case the try/catch block
        // is useless. But who knows? Maybe compiler will.
        std::transform
            (
            first,                              // input begin
            last,                               // input end
            groups_to_process.end() - count,    // output
            SelectGroupFromObjectIdToMigrationGroupMultiMapValue() // unary functor
            );
        }
    catch( ... )
        {
        // Resize groups_to_process back to its original size.
        groups_to_process.resize( old_groups_to_process_size );
        throw;
        }

    /* Remove selected messages from waiting queue. */
    groups_waiting_for.erase( first, last );
    }


void ObjectManager::refresh_migration_messages_waiting_for
    (
    const ObjectId & object_id
    )
    {
    // Select all migration messages waiting for the specified object.
    std::pair
        <
        ObjectIdToMigrationGroupMultiMap::iterator,
        ObjectIdToMigrationGroupMultiMap::iterator
        > range
            (
            groups_waiting_for.equal_range( object_id )
            );
    size_t count = std::distance( range.first, range.second );

    if( count == 0 )
        {
        return;
        }

    _move_waiting_migration_messages_to_process_queue
        (
        range.first,
        range.second,
        count
        );
    }


void ObjectManager::refresh_all_waiting_migration_messages()
    {
    // Select all migration messages stored in groups_waiting_for.
    size_t count = groups_waiting_for.size();

    if( count == 0 )
        {
        return;
        }

    std::pair
        <
        ObjectIdToMigrationGroupMultiMap::iterator,
        ObjectIdToMigrationGroupMultiMap::iterator
        > range
            (
            groups_waiting_for.begin(),
            groups_waiting_for.end()
            );
    hope( range.first != range.second );

    _move_waiting_migration_messages_to_process_queue
        (
        range.first,
        range.second,
        count
        );
    }


void ObjectManager::deliver_objects_from_migration_messages()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::deliver_objects_from_migration_messages()" );
    MigrationGroupMessage msg;

    /* Get next migration message if any is prepared.
     * Objects from the message are automatically
     * removed from waiting_objects set. */
    while( get_next_migration_group_to_process( msg ) )
        {
        /* Unpack the message. */
        deliver_objects_from_migration_message( msg );
        }
    }


ObjectProvider * ObjectManager::get_provider
    (
    const ProviderId & pid
    )
    {
    ProviderList::iterator  it( providers.find( pid ) );
    ObjectProvider *        p;

    if( it != providers.end() )
        {
        p = it->second;       // Local provider.
        }
    else
        {
        p = virtual_provider; // Remote provider, return virtual_provider instead.
        }

    hope( p != NULL );
    return p;
    }


bool ObjectManager::get_next_migration_group_to_process
    (
    MigrationGroupMessage & p
    )
    {
    if( !groups_to_process.empty() )
        {
        ObjectShared::ObjectIdList::const_iterator it;

        /* Get first message. */
        p = groups_to_process.front();
        groups_to_process.pop_front();

        /* Remove all objects stored in the message from waiting_objects set.
         * Note: Stl (should) guarantee that next code will not throw exception. */
        it = p.objects.begin();
        while( it != p.objects.end() )
            {
            const ObjectId & object_id = *it;
            waiting_objects.erase( object_id );
            it++;
            }
        return true;
        }
    else
        {
        return false;
        }
    }


void ObjectManager::send_find_object_message
    (
    FindObjectMessage & msg,
    const NodeId &      node_id
    )
    {
    /* Prepare the message according to localization protocol. */
    update_find_object_message_for_sending( msg );

    /* Send the message. */
    msg.send( node_id, NetworkCommon::STREAM_RELIABLE );
    }


void ObjectManager::send_migration_message
    (
    MigrationGroupMessage & msg,
    const NodeId &          node_id
    )
    {
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
    bool       first_time = ( msg.hops == 0 ); // The message is send for the first time.
    SystemTime timeout = Global::system_time() + MIGRATION_OBJECT_CACHE_TIMEOUT;
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

    /* Prepare the message according to localization protocol. */
    update_find_object_message_for_sending( msg );

    /* Send the message. */
    msg.send( node_id, NetworkCommon::STREAM_RELIABLE );

    /* Update localization and migration caches. */
    ObjectShared::ObjectIdList::const_iterator it( msg.objects.begin() );
    ObjectShared::ObjectIdList::const_iterator end( msg.objects.end() );
    while( it != end )
        {
        const ObjectId & object_id = *it;
        ObjectStatusMessage os( object_id, ObjectStatus::ALIVE, node_id, true );
        update_remote_object_status_cache( os );
        
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE        
        // FIXME Doresit, kdy presne se objekty davaji do migration cache.
        if( first_time )
            {
            migration_object_cache.update( object_id, timeout, EmptyStruct() );
            }
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

        ++it;
        }
    }


bool ObjectManager::is_object_owned
    (
    const ObjectId & object_id
    )
    {
    ObjectProperty * object = gc.find_local_object( object_id );

    if( object && !object->is_replica() )
        {
        return true;
        }
    else
        {
        return is_object_waiting( object_id ) || is_swapped_out( object_id );
        }
    }


void ObjectManager::gc_unmanage_objects
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    ObjectShared::ObjectList::const_iterator it;

    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty * object = ( *it );
        hope( object );
        gc.unmanage_object( object );
        }
    }


bool ObjectManager::is_object_group_pinned_or_stack_referenced
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    ObjectShared::ObjectList::const_iterator it;
    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty * object = ( *it );
        hope( object );
        if( object->is_pinned() || object->is_stack_referenced() )
            {
            #ifdef OPTION_DEBUG
            
            Global::log_debug
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_LOW,
                "Object " + object->get_object_id().to_string() + " is " +
                ( object->is_pinned() ? " pinned." : " stack referenced." )
                );

            #endif

            return true;
            }
        }
    return false;
    }


void ObjectManager::log_delivered_to
    (
    const ObjectId & source,
    const ObjectId & destination,
    const STime &    deliver_time,
    bool             silent
    )
    {
    std::ostringstream oss;
    oss << source;
    if( silent )
        {
        oss << " silently";
        }
    oss << " delivered_to "
        << destination << " (deliver time " << deliver_time << ")";
    Global::log_debug( Status::FACILITY_OBJECT_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
    }


void ObjectManager::log_delivery_failed
    (
    const ObjectId &      source,
    ObjectDeliveryFailure failure, 
    const STime &         deliver_time,
    bool                  silent
    )
    {
    std::ostringstream oss;
    oss << source << " delivery";
    if( silent )
        {
        oss << " silently";
        }
    oss << " failed because ";
    
    switch( failure )
        {
        case Object::DESTINATION_NOT_ACCESSIBLE:
            oss << "DESTINATION_NOT_ACCESSIBLE";
            break;
        
        case Object::EVENT_KILLED:
            oss << "EVENT KILLED";
            break;
        }
    
    oss << " (deliver time " << deliver_time << ")";
    Global::log_debug( Status::FACILITY_OBJECT_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
    }


const ObjectManager::ObjectManagerStatistics & ObjectManager::get_statistics() const
    {
    return statistics;
    }


const ObjectManager::ObjectManagerSettings & ObjectManager::get_settings() const
    {
    return settings;
    }


} // namespace Massiv::Core
} // namespace Massiv
