/** @file object_manager.h
 * Object manager maintains object providers and list of all known local
 * and replicated objects.
 */

/*
 * begin        : Wed Feb 20 2002
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECT_MANAGER_H
#define MASSIV_OBJECT_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_PROVIDER_H
#include "object/object_provider.h"
#endif

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_GARBAGE_COLLECTOR_H
#include "pointer/garbage_collector.h"
#endif

#ifndef MASSIV_REPLICATION_FWD_H
#include "replication/replication_fwd.h"
#endif

#ifndef MASSIV_BALANCER_H
#include "balance/balancer.h"
#endif

#ifndef MASSIV_SCHEDULE_H
#include "object/schedule.h"
#endif

#ifndef MASSIV_MISC_CLASS_H
#include "utility/misc_class.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_MIGRATION_GROUP_MSG_H
#include "sysevent/migration_group_msg.h" // Required for migration queues.
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_STACK_ROOT_OBJECT_H
#include "object/stack_root_object.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#include <deque>
#include <queue>
#include <functional>
#include <memory>


// #define MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
// Define to enable migration cache.
// Note: Migration cache is obsolete. It is maintained only for
// debug purposes. Enabling migration cache has no effect on the massiv
// system.


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/** @class ObjectManager
 * Object manager maintains objects and objects providers,
 * manages object searching, conversions from objectid
 * to pointers and back. Only one active instance is allowed.
 *
 * @bug BUG: Pri kazdem write pristupu na objekt, se musi
 * zavolat write_requested() na objektu, hlave pri zpracovavani
 * udalosti na objektu a migraci apod, kde to nedela primo objekt.
 * Otestovat zda se to ve vsech relevantnich funkcich dela.
 * Problem je ve funkcich Objectu, napriklad collect by asi mel
 * volat write_requested() apod.
 * Nekde staci jen archive_now(), protoze write_requested()
 * zaspinuje repliky apod., coz ve vetsine pripadu neni nutne.
 *
 * There are three states of local object:
 * <ul>
 * <li> Active. The object is:
 *      <ul>
 *      <li> Registered in the garbage collector.
 *      <li> Not replicated.
 *      <li> Stored in the memory.
 *      </ul>
 * <li> Inactive. There are two states of inactive object.
 *      <ul>
 *      <li> Waiting.
 *          <ul>
 *          <li> Not registered in the garbage collector.
 *          <li> Not replicated (may change in the future).
 *          <li> Stored in waiting queues of object manager.
 *          </ul>
 *      <li> Swapped-out.
 *          <ul>
 *          <li> Not registered in the garbage collector.
 *          <li> Not replicated (may change in the future).
 *          <li> Stored in the swap.
 *          </ul>
 *      </ul>
 * <li> Replica. The object is:
 *      <ul>
 *      <li> Registered in the garbage collector.
 *      <li> Replicated.
 *      <li> Stored in the memory.
 *      </ul>
 * </ul>
 *
 * "Local object" means active, inactive or replica. <br>
 * "Registered object" means active or replica. <br>
 * "Owned object" means active or inactive. <br>
 * "Remote object" means not owned.
 *
 * Object manager caches:
 * <ul>
 * <li> Remote object status cache: Contains information on which
 *      node a remote object is (possibly) located.
 * <li> Migration object cache: Stores object ids of objects,
 *      which have migrated from the local node recently.
 *      Used by distributed global search protocol.
 * </ul>
 *
 * Cache rules:
 * <ul>
 * <li> If an object is owned - remote object cache for the object
 *  is empty and migration cache for the object is empty.
 * </ul>
 */
class MASSIV_EXPORTED ObjectManager
    {
    friend class ArchiveManager;
    friend class GlobalSearchManager;
    
    class UpcallContext;
    typedef Object::DeliveryFailure ObjectDeliveryFailure;
    
private:

    NO_COPY_CONSTRUCTOR( ObjectManager );
    NO_ASSIGNMENT_OPERATOR( ObjectManager );

public:

    /**
     * Constructor.
     */
    ObjectManager();

    /**
     * Destructor.
     */
    ~ObjectManager();
    
    /**
     * Performs additional initialization, mostly the work that can not
     * be done from the constructor because of access to core Globals.
     */
    void initialize();

    /**
     * Shutdowns object manager.
     */
    void shutdown();

public:

    /**
     * Enable simulation.
     * @note If simulation is disabled no event is executed.
     */
    void enable_simulation();
    
    /**
     * Disable event simulation.
     * @note If simulation is disabled no event is executed.
     */
    void disable_simulation();
    
    /**
     * Check whether event simulation is enabled.
     */
    bool is_simulation_enabled() const;

public:

    /**
     * Generates new object id (by active provider).
     * Returns new object_id.
     */
    void generate_id
        (
        ObjectId &          object_id,
        bool                track_by_provider,
        const ClassTypeId & class_type_id,
        bool                no_archive
        );

    /**
     * Kill event (with @a event_id) scheduled for object (with @a object_id).
     * This function is mainly for @c EventHandle purposes.
     * @see EventHandle::kill() for more details.
     */
    void kill_event
        (
        const ObjectId & object_id,
        const EventId &  event_id
        );

public:

    //
    // Providers.
    //

    /**
     * Force the object manager to (re)select one provider from local providers
     * and make him active.
     * Active provider is responsible for generating new object ids.
     * This function is automatically called when new provider is registered
     * or destroyed.
     */
    void select_active_provider();
    
    /**
     * Check whether @a provider_id is registered as local object provider.
     * @return True if the @a provider_id was previously registered
     *      by @a register_local_provider().
     */
    bool is_local_provider
        (
        const ProviderId & provider_id
        );
        
    /**
     * Check whether @a provider_id is registered as remote object provider.
     * @return True if the @a provider_id was previously registered
     *      by @a register_remote_provider().
     */
    bool is_remote_provider
        (
        const ProviderId & provider_id
        );

    /**
     * Register local object provider.
     */
    void register_local_provider
        (
        const ObjectProviderState & object_provider_state
        );
    
    /**
     * Enumerate local providers and push them into the
     * #StartupMessage message.
     * The provider list in the message is cleared at the beginning.
     */
    void enumerate_local_providers
        (
        StartupMessage & msg
        );

    /**
     * Register remote object provider.
     * @param node_id Node where the provider is located.
     */
    void register_remote_provider
        (
        const ProviderId & remote_provider_id,
        const NodeId &     node_id
        );

    /**
     * Get node id of the node which owns specified provider.
     * Returns @c NULL_ID if the provider doesn't exist in the system.
     * @param include_local_providers Whether valid node id should
     *  be returned when the provider is owned by local system.
     *  If @a include_local_providers is false, then this function
     *  returns @c NULL_ID if the provider is local. Otherwise
     *  local node id is returned.
     */
    NodeId get_provider_owner_node
        (
        const ProviderId &  provider_id,
        bool                include_local_providers
        );

public:

    //
    // Notification callbacks.
    //

    /**
     * Notify entire system of object destruction.
     * This function must be called whenever object
     * is permanently removed from simulation.
     * Object with @a object_id should already be unregistered.
     *
     * Caches of local providers are updated.
     * If the object is tracked then its provider is notified
     * of the object's destruction.
     */
    void global_notify_of_object_destruction
        (
        const ObjectId & object_id
        );

    /**
     * Notify the manager of object's local schedule change.
     * This function must be called whenever new event
     * is added or any event is deleted from the object.
     * @param object Object whose events were changed.
     */
    void notify_of_local_schedule_change
        (
        ObjectProperty *    object
        );

    /**
     * Remove @a object from global schedule.
     * Use to deactivate the @a object.
     *
     * @pre The object is registered.
     */
    void remove_object_from_global_schedule
        (
        ObjectProperty *    object
        );

    //
    // Message handling.
    //

    /**
     * Check wheter specified object is stored
     * in some waiting migration message.
     */
    bool is_object_waiting
        (
        const ObjectId & object_id
        ) const;

    /**
     * Check whether an object is owned.
     */
    bool is_object_owned
        (
        const ObjectId & object_id
        );

public:

    /**
     * Open the swap file.
     */
    void open_swap
        (
        const std::string & archive
        );

    /**
     * Create new empty swap and opens it.
     * @todo FIXME Temporary hack for debug purposes.
     */
    void create_empty_swap
        (
        const std::string & file_name
        );

    /**
     * Close the swap file.
     * Nothing is done if no swap file is opened
     * or archivation is running. Otherwise all
     * objects stored in the swap are loaded into
     * the memory and the swap itself is closed.
     *
     * Closing swap file during archivation is not supported now
     * (may change in the future).
     * If you need to load all objects from the swap (but not
     * close the swap itself), use @c swap_in_all_objects().
     *
     * @note Closing swap file during archivation is not supported now.
     * To make it work during archivation, the archive manager
     * must be updated.
     * @todo Make close_swap() working during archivation. But is there
     * any reason why to do that???
     */
    void close_swap();

    //
    // Archivation.
    //

private:

    /**
     * Prepare @e owned objects for archivation.
     * All objects stored in waiting migration messages
     * are extracted into memory so they can be
     * archived.
     * @note This function is only for @c ArchiveManager
     * purposes so do not use it.
     * @note CZ NOTE: Je dobre si uvedomit, ze tato funkce
     * nijak neovlivni simulaci sveta - vse bude (melo by :)
     * fungoval dal jak ma. Objekty jsou pouze vyhozeny
     * z wait front aby mohly byt zaarchivovany. Ale za chvili
     * budou opet presunuty do wait front.
     */
    void prepare_for_archivation();

    /**
     * Store all information (except objects) about simulation in the archive file.
     * This includes current simulation time and all states of local providers.
     */
    void archive_simulation_state
        (
        Archive * archive
        );

    /**
     * Restore all information (except objects) abou simulation from the archive file.
     * @see archive_simulation_state().
     */
    void restore_simulation_state
        (
        Archive * archive
        );

    /**
     * Restore all information (except objects) abou simulation from the input file.
     * @see archive_simulation_state().
     */
    void restore_simulation_state
        (
        std::istream & input_stream
        );

public:

    /**
     * Load unpacked archive.
     *
     * @param directories Array of directory names. Each directory
     *      holds one unpacked archive.
     * @pre @a directories is not empty.
     * @note Assumes that the unpacked archive is located on a native 
     *      filesystem (i.e. filesystem provided by the operating system,
     *      not Volume, CompressedVolume, etc.).
     */
    void load_unpacked_archive
        (
        const std::vector< std::string > & directories
        );

private:

    /**
     * This function enumarates all objects which have
     * @c simulation_startup_notify idl class attribute set to true
     * and calls @c Object::object_update() with 
     * @c Object::SIMULATION_STARTUP argument on each of them.
     * @pre Swap file is opened.
     */
    void notify_objects_in_swap_of_simulation_startup();

    /**
     * Get swap file.
     * @return @c NULL if swapping is not enabled,
     * otherwise pointer to swap archive file.
     * @note This function is only for @c ArchiveManager
     * purposes do not use it.
     */
    Archive * get_swap();

public:

    /**
     * Advance everything except simulation.
     * To advance simulation call @c simulation_next_tick().
     */
    void next_tick();
    
    /**
     * Advance simulation.
     */
    void simulation_next_tick();

    /**
     * Suggests when next tick should be processed.
     * Returns invalid time if "do not know".
     */
    SystemTime suggest_next_tick_time();

    /**
     * Tests if object manager is idle.
     * It is not idle if @c statistics.too_busy == true,
     * @c statistics.schedule_blocked == true or if there is a pending
     * migration scheduled to the past.
     */
    bool is_idle();

private:
    
    /**
     * Called by @c advance_simulation() to process all events that
     * are scheduled to @c curr_time.
     * If the processing needs to upcall client code, @c upcall
     * is filled and the processing stops.
     * The caller is then responsible for dispatching the @c upcall
     * and calling @c process_events() again to finish its work.
     */
    bool process_events
        (
        const STime &   curr_time,
        UpcallContext & upcall
        );

    /**
     * Dispatches @c upcall and invalidates it.
     */
    void dispatch_upcall
        (
        UpcallContext & upcall
        );

    /**
     * Advance simulation to current time.
     * All scheduled events till current time are executed.
     */
    void advance_simulation();

    /**
     * Process all pending object migration group messages
     * (i.e. all messsages that must be processed as soon
     * as possible ).
     * @note Do not forget that this function can do
     * almost everything - create or destroy objects,
     * schedule events, work with net subsystem, etc.
     * Do not expect that only  @c advance_simulation()
     * does "simulation logic".
     * On the other hand, you can assume that this function
     * do not execute any scheduled event ie. no Lib code
     * is called.
     */
    void process_migration_messages();

    /**
     * Advance localization caches.
     */
    void advance_localization_caches();

private:

    /**
     * Prepare FindObject message before sending according
     * to object localization protocol.
     */
    void update_find_object_message_for_sending
        (
        FindObjectMessage & p
        );

    /**
     * Add migration message to waiting queue where are stored
     * all migration messages waiting for some objects.
     */
    void add_migration_message_to_waiting_queue
        (
        const MigrationGroupMessage & message
        );

    /**
     * Add migration message to message-to-process queue.
     * The message will be processed as soon as possible.
     * All objects stored in the message are marked
     * as waiting.
     */
    void add_migration_message_to_process_queue
        (
        const MigrationGroupMessage & message
        );

    /**
     * Schedule migration messages that are stored in waiting queue
     * to be re-processed if they are waiting for object with @a object_id.
     */
    void refresh_migration_messages_waiting_for
        (
        const ObjectId & object_id
        );

    /**
     * Schedule all migration messages that are stored in waiting queue
     * to be re-processed.
     * For archivation purposes.
     */
    void refresh_all_waiting_migration_messages();

    /**
     * Create all objects from migration messages that
     * are scheduled in groups-to-process queue.
     * For archivation purposes.
     */
    void deliver_objects_from_migration_messages();

    /**
     * Get local object provider or virtual provider if the provider_id
     * belongs to remote provider.
     */
    ObjectProvider * get_provider
        (
        const ProviderId & provider_id
        );

    /**
     * Get next migration group message to process.
     * Objects stored in the returned message are automatically
     * removed from @c waiting_objects set.
     * Returns false if there is no message to process. In that
     * case @a output_message stays unchanged.
     * Returns true if @a output_message was filled
     * with migration group message.
     */
    bool get_next_migration_group_to_process
        (
        MigrationGroupMessage & ouput_message
        );

    /**
     * Send FindObject message to specified node.
     * The message is updated according to localization protocol.
     */
    void send_find_object_message
        (
        FindObjectMessage & message,
        const NodeId &      node_id
        );

    /**
     * Send MigrationMessage to specified node.
     * The message is updated according to localization protocol.
     */
    void send_migration_message
        (
        MigrationGroupMessage & message,
        const NodeId &          node_id
        );

    /**
     * Prevents GC from managing objects stored in the @a object_list.
     * @pre All pointers stored in the list are valid pointers to objects.
     */
    void gc_unmanage_objects
        (
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Check whether group of objects is pinned or stack-strong referenced.
     * @return True if some of the objects from the @a object_list
     *  are pinned or stack-strong referenced.
     */
    bool is_object_group_pinned_or_stack_referenced
        (
        const ObjectShared::ObjectList & object_list
        );

    //
    // Methods which manage various caches.
    // Implementation in object_manager_cache.cpp file.
    //

public:

    /**
     * Get cached status of remote object.
     * If @a object_id is proxy node object for local node,
     * @c ALIVE status is returned (even if the NodeObject representing
     * local node was not created yet - caller must handle this by itself).
     * If @a object_id is proxy node object for a remote node,
     * it will return @c ALIVE or @c NOT_ACCESIBLE status depending
     * on whether the node is accessible or not.
     * @note Doesn't check whethe the object is local.
     */
    void get_remote_object_status
        (
        const ObjectId &    object_id,
        ObjectStatus &      object_status,
        NodeId &            node_id
        );

    /**
     * Hint object manager about object location.
     */
    void hint_object_location
        (
        const ObjectId &    object_id,
        const NodeId &      node_id
        );

private:

    /**
     * Update remote object status cache.
     * @note Object manager can discard the status if it
     * detects some strange circumstances.
     * @note Only local cahe is updated. The information is not forwarded
     * to any remote provider even if the status is trustworthy.
     * @see global_notify_of_object_destruction().
     *
     * @warning Do not initiate global search from this function.
     *      See GlobalSearchManager.
     */
    void update_remote_object_status_cache
        (
        const ObjectStatusMessage & object_status_message
        );

    /**
     * Refresh cached object status timeout for one object.
     * Nothing is done if there is no cached status
     * for the object.
     */
    void refresh_remote_object_status_cache_timeout
        (
        const ObjectId & object_id
        );

    /**
     * Clear cached remote object status for one object.
     * Nothing is done if there is no cached status
     * for the object.
     * @note Same as calling @c update_remote_object_status_cache()
     *       with ObjectStatus::UNKNOWN.
     */
    void clear_remote_object_status_cache
        (
        const ObjectId & object_id
        );

    /**
     * Clear cached remote object status for objects from a list.
     */
    void clear_remote_object_status_cache
        (
        const ObjectShared::ObjectIdList & object_id_list
        );

#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
    /**
     * Clear migration object cache for one object.
     */
    void ObjectManager::clear_migration_object_cache
        (
        const ObjectId & object_id
        );

    /**
     * Clear migration object cache for objects from a list.
     */
    void ObjectManager::clear_migration_object_cache
        (
        const ObjectShared::ObjectIdList & object_id_list
        );
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

public:

    //
    // Methods which implement localization protocol.
    // It includes migration group and find object messages processing.
    // Implementation in object_manager_localization.cpp file.
    //

    /**
     * Handles incoming ObjectStatus message.
     * Should be called only from @c SystemMessage::execute().
     */
    void handle_incoming_object_status_message
        (
        ObjectStatusMessage & message
        );

    /**
     * Handles incoming FindObject message.
     * Should be called only from @c SystemMessage::execute().
     */
    void handle_incoming_find_object_message
        (
        FindObjectMessage & message
        );

    /**
     * Handles incoming MigrationGroup message.
     * Should be called only from @c SystemMessage::execute().
     */
    void handle_incoming_migration_group_message
        (
        MigrationGroupMessage & message
        );
        
private:
    
    /**
     * Preprocess migration group message received from client node.
     * It includes object id remapping, because objects on
     * client nodes are created using dummy object provider.
     */
    void preprocess_migration_group_message_from_client
        (
        MigrationGroupMessage & message
        );

private:

    // Implementation in object_manager_localization.cpp file.

    /**
     * Process FindObject message.
     */
    void process_find_object_message
        (
        FindObjectMessage & message
        );

    /**
     * Process MigrationGroup message.
     */
    void process_migration_group_message
        (
        MigrationGroupMessage & message
        );

private:

    // Implementation in object_manager_localization.cpp file.

    /**
     * Internal localization function. Replies to
     * localization message if the message destination object
     * is owned object.
     */
    void reply_to_localization_message_for_owned_object
        (
        const FindObjectMessage * message
        );

    /**
     * This is main function of localization and migration protocol.
     * Handles FindObject and MigrationGroup messages.
     * @param message_type Message type. Must be @c SystemMessage::FIND_OBJECT
     *      or SystemMessage::MIGRATION_GROUP
     *      or SystemMessage::MIGRATION_GROUP_NONARCHIVABLE.
     * @param find_object_message Valid poiter to FindObjectMessage if @a message_type
     *      is @c SystemMessage::FIND_OBJECT, otherwise @c NULL.
     * @param migration_group_message Valid pointer to MigrationGroupMessage if
     *      @a message_type is @c SystemMessage::MIGRATION_GROUP or
     *      @c SystemMessage::MIGRATION_GROUP_NONARCHIVABLE, otherwise @c NULL.
     */
    void process_localization_message
        (
        SystemMessage::Type     message_type,
        FindObjectMessage *     find_object_message,
        MigrationGroupMessage * migration_group_message
        );

private:

    //
    // Packing/unpacking objects to/from migration group messages,
    // replication group messages or swap file.
    //

    /**
     * Migration packing operation result.
     * Migration packing is operation in which objects from migration
     * group are packed (serialized) into migration group message.
     */
    enum MigrationPackResult
        {
        MIGRATION_PACK_RESULT_SUCCESS,
            /**< Migration was successfull. */
        MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED,
            /**< Migration failed because some objects from migration group are
             * pinned or stack-strong referenced.
             */
        MIGRATION_PACK_RESULT_OBJECT_MAY_NOT_MIGRATE
            /**< Migration failed because some objects from migration group
             * may not migrate (for example node objects representing local
             * node).
             */
        }; // MigrationResult

public:

    /**
     * Swap-out operation result.
     */
    enum SwapOutResult
        {
        SWAP_OUT_RESULT_SUCCESS,
            /**< Swap-out was successful. */
            
        SWAP_OUT_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED,
            /**< Swap-out failed because some objects from the migration group
             * are pinned or stack-strong referenced.
             */
             
        SWAP_OUT_RESULT_ARCHIVATION_IN_PROGRESS,
            /**< Swap-out failed because archivation is in progress. */
            
        SWAP_OUT_RESULT_NO_SWAP_FILE,
            /**< Swap-out failed because no swap file is opened. */
            
        SWAP_OUT_RESULT_IS_REPLICA,
            /**< Swap-out failed because the specified object is replica
             * and replicas can't be swapped-out.
             */

        SWAP_OUT_RESULT_HAS_REPLICA_REQUESTS,
            /**< Swap-out failed because the specified object is replicated
             * to different node and must remain in memory.
             */
        
        SWAP_OUT_RESULT_UNREGISTERED_OBJECT,
            /**< Swap-out failed because the specified object is not
             * accessible - the object is not registered in garbage
             * collector.
             */
        
        SWAP_OUT_RESULT_OBJECT_NOT_ARCHIVABLE
            /**< Swap-out failed because the specified object is not
             * archivable. Only archivable objects may be stored inside
             * an archive file and swap file is alse an archive file.
             */
        
        }; // SwapOutResult

private:

    //
    // Packing/unpacking to/from MigrationGroupMessage.
    // Implementation in object_manager_packing.cpp file.
    //

    /**
     * Enumerate migration group, pack its objects to
     * MigrationGroupMessage and finally destroy the objects.
     * @pre @a migration_message is empty - no objects are stored in it.
     * @note There is no guarantee that the @a migration_message
     * stays unchanged if an exception is throw.
     *
     * @return MigrationPackResult of the operation.
     * If the result is @c MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED
     * then the @a migration_message stays unchanged.
     */
    MigrationPackResult pack_migration_group_to_migration_message
        (
        ObjectProperty *        object_from_migration_group,
        MigrationGroupMessage & migration_message
        );

    /**
     * Pack migration group objects to MigrationGroupMessage
     * and finally destroy the objects.
     * @pre @a migration_message is empty - no objects are stored in it.
     * @note There is no guarantee that the @a migration_message
     * stays unchanged if an exception is throw.
     *
     * @return MigrationPackResult of the operation.
     * If the result is @c MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED
     * then the @a migration_message stays unchaned.
     */
    MigrationPackResult pack_migration_group_to_migration_message
        (
        const ObjectShared::ObjectList & objects_from_migration_group,
        MigrationGroupMessage &          migration_message
        );

    /**
     * Create all objects from the migration message @a message
     * and make them local. The objects are added to the simulation.
     *
     * @note Before calling this function make sure that
     * all objects from the message are not marked in @c waiting_objects
     * set.
     *
     * @throw ExMigrationGroupUnpackSerializationError The operation failed
     *  because there was a serialization error during unpacking objects from
     *  the migration group.
     *  Note that local system state is not changed - it is safe to continue
     *  simulation if this exception is thrown.
     * @throw ExMigrationGroupUnpackError The operation failed (for any reason).
     *  Note that local system state can be changed - remote providers can
     *  be notified, the objects can be archive etc. If this exception is thrown
     *  the system should shut down itself.
     */
    void deliver_objects_from_migration_message
        (
        const MigrationGroupMessage & message
        );

public:

    //
    // Object swapping.
    // Implementation in object_manager_packing.cpp file.
    //
    
    /**
     * Swap out an object (and its migration group) specified by its ObjectId.
     * All the objects from the migration group are destroyed.
     * @return SwapOutResult whether the swap-out was successfull or not.
     * @note The method can fail for many reasons - if no swap
     *  file is opened, archivation is in progress or some of the objects
     *  are pinned or stack-strong referenced.
     * @see SwapOutResult
     */
    SwapOutResult try_to_swap_out
        (
        const ObjectId & object_id
        );
        
private:

    /**
     * Swap out object (and its migration group).
     * All the objects from the migration group are destroyed.
     * @pre The @a object argument is not null.
     * @pre The object is not replica.
     * @return SwapOutResult whether the swap-out was successfull or not.
     * @note The method can fail for many reasons - if no swap
     *  file is opened, archivation is in progress or some of the objects
     *  are pinned or stack-strong referenced.
     * @see SwapOutResult
     */
    SwapOutResult swap_out_object
        (
        ObjectProperty * object
        );

    /**
     * Swap out objects to swap file.
     * The objects are registered in the swap as one (new) group
     * (ie. the objects should be one migration group).
     * The objects are destroyed.
     * @pre None of the objects is replica.
     * @pre All objects are archivable.
     * @return SwapOutResult whether the swap-out was successfull or not.
     * @note The method can fail for many reasons - if no swap
     *  file is opened, archivation is in progress or some of the objects
     *  are pinned or stack-strong referenced.
     * @see SwapOutResult
     */
    SwapOutResult swap_out_objects
        (
        const ObjectShared::ObjectList & object_list
        );

public:

    /**
     * Try load object from swap file.
     * Nothing is done if no swap file is opened or the requested object
     * is not stored in the swap. In these cases the function returns @c NULL.
     * If the object is properly swapped in, the pointer to the object
     * is returned (same as calling GarbageCollector::find_local_object()).
     * @note This function can be called at any time.
     */
    ObjectProperty * try_to_swap_in
        (
        const ObjectId & object_id
        );

    /**
     * Check whether the object is stored in the swap file.
     * If no swap file is opened, false is returned.
     * @note This function can be called at any time.
     */
    bool is_swapped_out
        (
        const ObjectId & object_id
        );

    /**
     * Swap in all objects that are stored in the swap.
     * If no swap file is opened nothing is done.
     * @note This function can be called at any time.
     */
    void swap_in_all_objects();

private:

    /**
     * Swap in all objects that are scheduled for the @a time.
     * @note This function is harmless and can be made public if needed.
     * @note This function can be called at any time.
     */
    void swap_in_objects
        (
        const STime &   time
        );

    /**
     * Load objects from swap file.
     * This is heart of all swap-in functions.
     * You should never call it directly.
     * @param object_ids List of objects which should be swapped-in.
     *      The list must contain all objects from one archive group
     *      or all objects from the swap.
     * @warning The objects are _NOT_ removed from the swap file!
     */
    void swap_in
        (
        const ObjectShared::ObjectIdList &    object_ids
        );

private:

    //
    // Misc methods used in swap-in and unpacking objects from migration messages.
    //

    /**
     * Destroy objects which are prepared for migration.
     * The objects are first unmanaged (from garbage collector)
     * and then destroyed one-by-one.
     * @pre None of the objects is pinned or stack strong referenced.
     */
    void pack_destroy_objects_prepared_for_migration
        (
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Instantiate objects.
     *
     * If some object already exist, they must be replicas. These replicas
     * will be destroyed. All replica inconsitencies resulting from that
     * actions will be determined.
     *
     * If an exception is thrown, the caller is responsible
     * for destroying objects created by this method!
     *
     * @pre @a object_list is empty.
     * @pre All object ids in @a object_id_list have valid optional part
     *      of #ObjectId.
     *
     * @note This method doesn't destroy created objects if
     * an exception is thrown.
     */
    void unpack_create_objects
        (
        const ObjectShared::ObjectIdList &  object_id_list,
        ObjectShared::ObjectList &          object_list
        );

    /**
     * Read objects' states from bit stream.
     */
    void unpack_bitstr_read_objects
        (
        const ObjectShared::ObjectList & object_list,
        BitIStream &                     bis,
        const Serializer::Description &  desc
        );

    /**
     * Schedule objects for archivation.
     */
    void unpack_schedule_objects_for_archivation
        (
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Notify remote providers of owned objects.
     */
    void unpack_notify_remote_providers_of_owned_objects
        (
        const ObjectShared::ObjectIdList & object_id_list
        );

    /**
     * Refresh all migration messages waiting for objects from @a object_id_list.
     */
    void unpack_refresh_migration_messages_waiting_for_objects
        (
        const ObjectShared::ObjectIdList & object_id_list
        );

    /**
     * Instantiate objects.
     * @note This method doesn't destroy created objects if
     * an exception is thrown.
     * If an exception is thrown, the caller is responsible
     * for destroying objects created by this method!
     * @pre @a object_list is empty.
     * @pre All object ids in @a object_id_list have valid optional
     *      part of #ObjectId.
     */
    void swap_in_create_objects
        (
        const ObjectShared::ObjectIdList &  object_id_list,
        ObjectShared::ObjectList &          object_list
        );

    /**
     * Read objects' states from text stream.
     */
    void swap_in_text_read_objects_from_swap
        (
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Clean-up method used if and exceptio is thrown during migration group
     * unpacking.
     */
    void unpack_exception_clean_up
        (
        const ObjectShared::ObjectList & object_list
        );

    /**
     * Clean-up method used if an exception is thrown during swap-in.
     */
    void swap_in_exception_clean_up
        (
        const ObjectShared::ObjectList & object_list
        );

private:

    //
    // Internal types.
    //

    enum
        {
        MAX_FIND_OBJECT_MESSAGE_HOPS_COUNT = 4
            /**< How many times can be FindObject message forwarded.
             * @todo Replace with config variable. */
        };

    enum
        {
        MAX_FIND_OBJECT_MESSAGE_PROVIDER_HOPS_COUNT = 1
            /**< How many times can be FindObject message forwarded
             * through object provider (if searching for tracked
             * object).
             * @todo Replace with config variable. */
        };

#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
    /** @typedef MigrationObjectCache
     * Migration object cache.
     */
    typedef Cache
                <
                ObjectId,
                SystemTime,
                EmptyStruct,
                ObjectId::Hash,
                ObjectId::EqualKey
                > MigrationObjectCache;
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

    /**
     * @c ObjectId set.
     */
    typedef ext::hash_set
                <
                ObjectId,
                ObjectId::Hash,
                ObjectId::EqualKey
                > ObjectIdHashSet;

    /**
     * Multimap which associates @c ObjectId with @c MigrationGroup.
     * There can be unlimited number of migration groups associated
     * with one object id.
     */
    typedef ext::hash_multimap
                <
                ObjectId,
                MigrationGroupMessage,
                ObjectId::Hash,
                ObjectId::EqualKey
                > ObjectIdToMigrationGroupMultiMap;

    /**
     * Unary functor. From @c ObjectIdToMigrationGroupMultiMap::value_type
     * returns @c MigrationGroupMessage.
     */
    struct SelectGroupFromObjectIdToMigrationGroupMultiMapValue
        {
        const MigrationGroupMessage & operator()
            (
            const ObjectIdToMigrationGroupMultiMap::value_type & value
            ) const
            {
            return value.second;
            }
        }; // SelectGroupFromObjectIdToMigrationGroupMultiMapValue

    /**
     * Double-ended queue of migration messages.
     * @note We do not use @c std::queue because we need
     * ability to use iterators.
     */
    typedef std::deque
                <
                MigrationGroupMessage
                > MigrationGroupDeque;

    /** @typedef ProviderList
     * List of local providers.
     */
    typedef ext::hash_map<
        ProviderId,
        ObjectProvider *,
        ProviderId::Hash,
        ProviderId::EqualKey >     ProviderList;

    /**
     * Maps remote provider to its owner node.
     * @note Local providers are not in this map.
     */
    typedef ext::hash_map
        <
        ProviderId,                 // Remote provider id.
        NodeId,                     // Node where the provider is located.
        ProviderId::Hash,
        ProviderId::EqualKey
        > RemoteProviderToNodeMap;

private:

    /**
     * ObjectManager statistics.
     * @todo Split into more statistics.
     */
    class ObjectManagerStatistics : public Statistics
        {
    public:

        /** Constructor. */
        ObjectManagerStatistics();

    public:

        Integer &   num_ticks;
            /**< Number of ticks the object manager has processed. */

        Integer &   num_total_events;
            /**< Total number of processed events. */
        Integer &   num_successful_events;
            /**< Number of successfully delivered events. */
        Integer &   num_failed_events;
            /**< Number of events which have failed to delivery
                 because destination object was not accessible in the system. */
        Integer &   num_forwarded_events;
            /**< How many times event delivery was forwarded to
                 migration protocol layer because destination object was
                 not locally accessible. */
        
        Integer &   num_debug_random_local_migrations;
            /**< How many times local migration was forced even thought
                 destination object was local.
                 @see ObjectManagerSettings::debug_force_random_local_migration. */

        Integer &   num_swapped_in_objects;
            /**< Number of swapped-in objects. */
        Integer &   num_swapped_out_objects;
            /**< Number of swapped-out objects. */
        Integer &   num_unpacked_objects;
            /**< Number of objects unpacked from migration messages.
                 Replicas are not included in the number. */
        Integer &   num_packed_objects;
            /**< Number of objects packed to migration messages.
                 Replicas are not included in the number. */

        Boolean &   schedule_blocked;
            /**< True if global event schedule is blocked by an object that could
                 not migrate. */
        Boolean &   too_busy;
            /**< True if we have been nested too long (time difference between
                 two ticks at level zero of the main loop was too big). */
        Float &     level_zero_time;
            /**< When we were at level zero of the main loop for the last time. */

        }; // ObjectManagerStatistics

public:

    /**
     * ObjectManager settings.
     */
    class ObjectManagerSettings : public Settings
        {
    public:

        /** Constructor. */
        ObjectManagerSettings();

    public:

        Boolean &   debug_force_random_local_migration;
            /**< Whether to enable random local migrations - object that
                 must be delivered to destination object is passed to
                 migration protocol layer even thought destination object
                 is local.
                 @note Ignored if the core is not compiled in debug mode. */
        Float &     debug_random_local_migration_probability;
            /**< Specifies probability of local migration
                 if debug_force_random_local_migration is true.
                 Should be in 0.0 to 1.0 range, other values may
                 cause random behavior. */
        Boolean &   debug_force_swap_out_of_objects_with_scheduled_events;
            /**< Whether to swap-out all objects which have scheduled
                 one or more events for current tick. The objects are
                 swapped-out before any event is processed in current tick. */
        Float &     max_schedule_delay;
            /**< Warn if global schedule has been blocked by a "busy" object
                 for at least @c max_schedule_delay seconds. */
        Float &     busy_time_limit;
            /**< Set @c statistics.too_busy to true if time difference
                 between ticks at level zero is greater that this limit. */
        Boolean &   log_events;
            /**< Log each @c delivered_to() and @c delivery_failed() events.
             * Should be used only for debug purposes because it can really
             * slow down entire system.
             */
        Boolean &   log_process_migration_messages;
            /**< Log each migration message that is beeing processed. */

        }; // ObjectManagerSettings.

public:

    /**
     * Get ObjectManager statistics.
     */
    const ObjectManagerStatistics & get_statistics() const;
    
    /**
     * Get ObjectManager settings.
     */
    const ObjectManagerSettings & get_settings() const;
    
public:

    /**
     * Log message that @a source object is delivered to @a destination object.
     * @param silent True if the event is silent ie. the @a source object
     *     is not notified of the event (see @c notify argument of @c Object::migrate_to()).
     */
    void log_delivered_to
        (
        const ObjectId & source,
        const ObjectId & destination,
        const STime &    deliver_time,
        bool             silent = false
        );

    /** 
     * Log message that delivery of @a source object has failed.
     * @param silent True if the event is silent ie. the @a source object
     *     is not notified of the event (see @c notify argument of @c Object::migrate_to()).
     */
    void log_delivery_failed
        (
        const ObjectId &      source,
        ObjectDeliveryFailure failure, 
        const STime &         deliver_time,
        bool                  silent = false
        );

private:

    //
    // Private. Think twice before you use some of these functions.
    //

    /**
     * Move migration messages from @c groups_waiting_for to @c groups_to_process.
     * @param count Number of elements in range [first, last).
     */
    void _move_waiting_migration_messages_to_process_queue
        (
        ObjectIdToMigrationGroupMultiMap::iterator  first,
        ObjectIdToMigrationGroupMultiMap::iterator  last,
        size_t                                      count
        );

private:

    //
    // State.
    //

    bool                                initialized;
        /**< Is object manager initialized? */

    //
    // Object scheduling.
    //
    
    bool                                simulation_enabled;
        /**< Whether simulation is enabled. If false, no event is
         * executed.
         */
    GlobalSchedule                      global_schedule;
        /**< List of objects with scheduled events. */

    //
    // Object searching (distributed).
    //

#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
    MigrationObjectCache                migration_object_cache;
        /**< Migration cache. Contains object that were recently moved to another node.
         * Used by global searching protocol.
         * @note Do not clear this cache for any reason except node (server) shutdown.
         */
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

    //
    // Waiting migration messages.
    //

    ObjectIdHashSet                     waiting_objects;
        /**< @c ObjectId list of objects from migration messages
         * that are stored in @c groups_waiting_for and @c groups_to_process.
         */
    ObjectIdToMigrationGroupMultiMap    groups_waiting_for;
        /**< List of migration messages that are waiting for some object.
         *
         * Object ids of objects from the migration messages that are stored
         * in @c groups_waiting_for are stored in @c waiting_objects set.
         *
         * There are several reasons why a migration message must be
         * waiting for some object:
         * <ul>
         * <li> Object manager pushes migration message in @c groups_waiting_for
         * if it has no info where the message destination object is located,
         * and it must do global search for the object. Until the
         * search is finished, the message is stored in @c groups_waiting_for
         * to be processed as soon as the search will be finished.
         * <li> Object manager pushes migration message in @c groups_waiting_for
         * if the message destination object is stored in some migration
         * group waiting to be processed in @c groups_to_process
         * queue. Until the migration group is processed, the message
         * is stored in @c groups_waiting_for.
         * <li> Object manager pushes migration message in @c groups_waiting_for
         * if the message destination object is stored in another
         * migration message which is already waiting in @c group_waiting_for.
         * </ul>
         *
         * Once a new object is created from migration message,
         * the object manager automatically moves migration groups
         * waiting for the created object from @c groups_waiting_for
         * to @c groups_to_process.
         */

    MigrationGroupDeque             groups_to_process;
        /**< List of all migration messages that are waiting
         * to be processed as soon as possible.
         *
         * Object managers pushes migration mesages to @c groups_to_process
         * if it can't process the messages immediately.
         *
         * Object ids of objects from the migration messages that are stored
         * in @c groups_to_process are stored in @c waiting_objects set.
         */

    //
    // Object providers.
    //

    ProviderList                    providers;
        /**< List of all providers that are owned by local system. */
    ObjectProvider *                active_provider;
        /**< Active provider (from @a providers) who currently generates new object ids. */
    ObjectProvider *                virtual_provider;
        /**< Provider (with null id) that represents other providers not owned by local system. */
    RemoteProviderToNodeMap         remote_provider_to_node;
        /**< Maps remote provider to its owner node. Local providers are not in this map. */

    //
    // Swap.
    //

    Archive *                       swap;
        /**< Swap file.
         * @note The swap can't be closed if archivation is in progress.
         * FIXME: Skutecne vhodne misto, kde tohle dokumentovat.
         */

    //
    // Statistics and settings.
    //

    ObjectManagerStatistics         statistics;
        /**< ObjectManager statistics. */
    ObjectManagerSettings           settings;
        /**< ObjectManager settings. */

public:

    //
    // Public subsystems.
    // Must be defined at the end of the class to ensure proper order of
    // construction/destruction.
    //
    
    StackRootObject                 stack_root;
        /**< Stack root object. */
    GarbageCollector                gc;
        /**< GarbageCollector object. */
    std::auto_ptr< ReplicaClient >  replica_client;
        /**< Replication client. Always non-NULL. */
    std::auto_ptr< ReplicaServer >  replica_server;
        /**< Replication server. Always non-NULL. */
    Balancer                        balancer;
        /**< Balancer. */

    }; // ObjectManager

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_OBJECT_MANAGER_H
