/** @file object_manager_cache.cpp
 * Implementation of ObjectManager's methods which
 * manage various caches.
 */

/*
 * begin        : Thu Nov 28 2002
 * copyright    : (C) 2002 Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "global/global.h"
#include "object/object.h"
#include "object/object_manager.h"
#include "object/object_common.h"
#include "sysevent/object_status_msg.h"
#include "utility/cache_impl.h"
#include "node/node_manager.h"
#include "system/system.h"
#include <memory>
#include <sstream>


namespace Massiv {
namespace Core {


void ObjectManager::get_remote_object_status
    (
    const ObjectId &    object_id,
    ObjectStatus &      object_status,
    NodeId &            node_id
    )
    {
    if( object_id.is_proxy_node_object_id() )
        {
        NodeId remote_node = object_id.proxy_extract_node_id();
        if( remote_node == Global::get_local_node_id() )
            {
            // Proxy object id for local node.
            // Set to ALIVE status (even if the NodeObject was not created yet -
            // caller must handle it by itself).
            object_status = ObjectStatus::ALIVE;
            node_id = remote_node;
            }
        else
            {
            // Proxy object id for remote node.
            // Check if the node is accessible.
            switch( remote_node.get_type() )
                {
                case NodeId::TYPE_SERVER:
                case NodeId::TYPE_CLIENT:
                    {
                    NodeManager & node_manager = Global::node_manager();
                    if( node_manager.is_node_accessible( remote_node ) )
                        {
                        object_status = ObjectStatus::ALIVE;
                        node_id = remote_node;
                        }
                    else
                        {
                        // The node is not valid node or it is not connected.
                        object_status = ObjectStatus::NOT_ACCESSIBLE;
                        node_id.clear();
                        }
                    break;
                    }

                case NodeId::TYPE_SERVICE:
                default:
                    // TODO: Log this.
                    object_status = ObjectStatus::NOT_ACCESSIBLE;
                    node_id.clear();
                    break;
                }
            }
        }
    else
        {
        // Get the provider which manages cached status for the object.
        // If the provider of the object is not local, virtual provider
        // is returned instead.
        ObjectProvider * object_provider = get_provider( object_id.get_provider_id() );

        hope( object_provider != NULL );
        object_provider->get_object_status( object_id, object_status, node_id );
        }
    }


void ObjectManager::hint_object_location
    (
    const ObjectId &    object_id,
    const NodeId &      node_id
    )
    {
    if( is_object_owned( object_id ) )
        {
        // Silently ignore any hints about local objects.

        return;
        }

    if( node_id == System::get_local_node_id() )
        {
        // Ignore hints about local node.

        return;
        }

    // Use the hint to update object status cache.
    // (FIXME: only if there is no info about the object?)

    update_remote_object_status_cache
        (
        ObjectStatusMessage
            (
            object_id,
            ObjectStatus( ObjectStatus::ALIVE ),
            node_id,
            false
            )
        );
    }


void ObjectManager::update_remote_object_status_cache
    (
    const ObjectStatusMessage & object_status_msg
    )
    {
    // Get the provider which manages cached status for the object.
    // If the provider of the object is not local, virtual provider
    // is returned instead.
    const ObjectId & object_id = object_status_msg.object;    
    
    if( object_id.is_proxy_node_object_id() )
        {
        // Somebody is trying to notify us of state of NodeObject.
        // This should never happen. NodeObjects are handled
        // in a special way by localization protocol.
        
        std::ostringstream ss;
        ss  << "Received object status notifying of status of proxy node object. Should never happen. "
            << "Sender " << object_status_msg.get_sender() << " (initiator: "
            << object_status_msg.get_initiator() << "). Status detail: "
            << object_status_msg;
    
        Global::log_warning
            (
            Status::FACILITY_OBJECT_LOCALIZATION,
            Status::PRIORITY_HIGHEST,
            ss.str()
            );
        
        // Ignore the message.
        return;
        }
    
    ObjectProvider * object_provider = get_provider( object_id.get_provider_id() );

    hope( object_provider != NULL );

    if( object_status_msg.object_status != ObjectStatus::UNKNOWN )
        {
        // Check whether the object is owned.
        bool    owned = is_object_owned( object_id );

        if( owned )
            {
            // The object is owned - do not update remote object cache.

#ifdef OPTION_DEBUG
                {
                /* Object should not be in remote or migration cache. */
                ObjectStatus    os;
                NodeId          si;
                get_remote_object_status( object_id, os, si );
                hope( os == ObjectStatus::UNKNOWN );
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE                
                hope( migration_object_cache.find( object_id, Global::system_time() ) == NULL );
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
                }
#endif // OPTION_DEBUG

            /* Check strange circumstances. */
            if( object_status_msg.object_status == ObjectStatus::NOT_ACCESSIBLE )
                {
                /* Note: This is not error, it is possible to happen. */
                std::ostringstream ss;
                ss << "Invalid NOT_ACCESSIBLE object status detected. "
                   << "Somebody is trying to notify us that object (id: "
                   << object_id << ") is not accessible, but the object is owned. "
                   << "The status was discarded. Status detail: "
                   << object_status_msg;

                Global::log
                    (
                    STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                    ss.str()
                    );
                }
            else
                {
                /* Note: This is not error, it is possible to happen. */
                std::ostringstream ss;
                ss << "Invalid ALIVE object status detected. "
                   << "Somebody is trying to notify us that object (id: "
                   << object_id << ") is alive on remote node, but the object is owned. "
                   << "The status was discarded. Status detail: "
                   << object_status_msg;

                Global::log
                    (
                    STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                    ss.str()
                    );
                }

            return; // Do not update remote cache.
            }
        else
            {
            /* The object is remote.
             * Check strange circumstances. */

            if( object_status_msg.object_status == ObjectStatus::ALIVE &&
                object_status_msg.node == Global::get_local_node_id() )
                {
                /* Somebody is trying to notify us that the object
                 * is owned, but it is not owned.
                 * Note: This is not error, it is possible to happen. */

                std::ostringstream ss;
                ss << "Invalid ALIVE object status detected. "
                   << "Somebody is trying to notify us that object (id: "
                   << object_id << ") is alive and owned, but it is not owned. "
                   << "The status was discarded. Status detail: "
                   << object_status_msg;

                Global::log
                    (
                    STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                    ss.str()
                    );

                /* Do not update cache to prevent internal error (cycle)
                 * in migration and find object message handling. */
                return;
                }
            }
        }

    object_provider->notify_of_object_status( object_status_msg );

    /* Finally refresh migration messages waiting for the object,
     * so they will be re-processed ASAP. */
    if( object_status_msg.object_status != ObjectStatus::UNKNOWN )
        {
        refresh_migration_messages_waiting_for( object_id );
        }
    }

    
void ObjectManager::refresh_remote_object_status_cache_timeout
    (
    const ObjectId & object_id
    )
    {
    // Get the provider which manages cached status for the object.
    // If the provider of the object is not local, virtual provider
    // is returned instead.
    ObjectProvider * object_provider = get_provider( object_id.get_provider_id() );

    hope( object_provider != NULL );
    object_provider->refresh_object_status( object_id );
    }


void ObjectManager::clear_remote_object_status_cache
    (
    const ObjectId & object_id
    )
    {
    // We will generate object status message which
    // will clear the cache.
    // To ensure that the cache will be cleared, the trustworthy
    // flag must be set.
    ObjectStatusMessage object_status_msg
                            (
                            object_id,
                            ObjectStatus::UNKNOWN, // clear the cache
                            NodeId(),              // null node id
                            true                   // trustworthy
                            );

    update_remote_object_status_cache( object_status_msg );
    }


void ObjectManager::clear_remote_object_status_cache
    (
    const ObjectShared::ObjectIdList & object_id_list
    )
    {
    ObjectShared::ObjectIdList::const_iterator it;
    
    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId &           object_id = *it;
        clear_remote_object_status_cache( object_id );
        }
    }

    
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE

void ObjectManager::clear_migration_object_cache
    (
    const ObjectId & object_id
    )
    {
    migration_object_cache.remove( object_id );
    }


void ObjectManager::clear_migration_object_cache
    (
    const ObjectShared::ObjectIdList & object_id_list
    )
    {
    ObjectShared::ObjectIdList::const_iterator it;

    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId & object_id = ( *it );
        clear_migration_object_cache( object_id );
        }
    }

#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE


} // namespace Massiv::Core
} // namespace Massiv
