/** @file object_manager_localization.cpp
 * Implementation of ObjectManager's methods which
 * implement localization protocol.
 * It includes migration group and find object messages
 * processing.
 */

/*
 * begin        : Tue Nov 23 2002
 * copyright    : (C) 2002 Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "object/object.h"
#include "object/object_manager.h"
#include "object/object_common.h"
#include "sysevent/object_status_msg.h"
#include "sysevent/migration_group_msg.h"
#include "utility/cache_impl.h"
#include "serializer/serializer.h"
#include "object/archive.h"
#include "object/archive_manager.h"
#include "object/collect_group.h"
#include "utility/debug.h"
#include "global/global.h"
#include "system/system.h"
#include "node/node_manager.h"
#include "global_search_manager.h"
#include <memory>
#include <sstream>


namespace Massiv {
namespace Core {


void ObjectManager::handle_incoming_object_status_message
    (
    ObjectStatusMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::handle_incoming_object_status_message()" );
    update_remote_object_status_cache( msg );
    }


void ObjectManager::handle_incoming_find_object_message
    (
    FindObjectMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::handle_incoming_find_object_message()" );
    process_find_object_message( msg );
    }


void ObjectManager::preprocess_migration_group_message_from_client
    (
    MigrationGroupMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::preprocess_migration_group_message_from_client()" );
    // Client can't migrate multiple objects.

    if( msg.objects.size() != 1 )
        {
        std::ostringstream ss;
        ss << "Client node sent " << msg.objects.size() << " objects in migration group.\n"
              "Only single-objects groups are allowed.";
        throw ExMigrationGroupUnpackError( ss.str() );
        }

    ObjectId &                  object_id   = msg.objects.front();
    if( !object_id.get_provider_id().is_dummy_provider_id() )
        {
        std::ostringstream ss;
        ss << "Client node sent object " << object_id << " with weird provider id.";
        throw ExMigrationGroupUnpackError( ss.str() );
        }
    if( !object_id.is_object_info_available() )
        {
        std::ostringstream ss;
        ss << "Client node sent object " << object_id << " without object info.";
        throw ExMigrationGroupUnpackError( ss.str() );
        }
    const bool                  is_tracked  = object_id.is_tracked_by_provider();
    const ObjectInfo            object_info = object_id.get_object_info();
    const bool                  no_archive = !object_info.is_archivable();
    const MetaObject * const    metaobject =
                object_id.get_class_type_id().get_metaobject_no_throw();
    if( !metaobject )
        {
        std::ostringstream ss;
        ss << "Client node sent object " << object_id << " with unsupported class type id.";
        throw ExMigrationGroupUnpackError( ss.str() );
        }

    if( !metaobject->get_attribute_client_server_migrate() )
        {
        std::ostringstream ss;
        ss << "Client node sent object " << object_id << ". Objects of type " <<
              object_id.get_class_type_id() << " can't migrate from clients to servers.";
        throw ExMigrationGroupUnpackError( ss.str() );
        }
    
    generate_id
        (
        object_id,
        is_tracked,
        object_info.get_class_type_id(),
        no_archive
        );
    }


void ObjectManager::handle_incoming_migration_group_message
    (
    MigrationGroupMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::handle_incoming_migration_group_message()" );
    // Note: groups_to_process may be non-empty here.
    
    /** @todo FIXME CZ: Je nutno pred zpracovanim tedle zpravy
     * nejdriv zpracovat vsechny v groups_to_process nebo ne?
     * Napr. advance_simulation() striktne zachovana poradi zpracovani
     * migracnich zprav podle groups_to_process, ale ted nevim jestli
     * to je k necemu potreba?
     * Pokud by to bylo potreba, tak misto deliver_objects_from_migration_message
     * nize by stacilo pushnout tuto zpravu do groups_to_process, takze
     * se zpracuje v neblizsich next_tick().
     * Problem je ale s deliver_objects_from_migration_message() nize,
     * pokud je potreba aby groups_to_process byly zpracovany pred touto
     * zpravou. Tak s tim nevim co delat...jedine zavolat 
     * process_migration_group_messages() na zacatku teto funkce...pak ale
     * bacha na vyjimky, ktere mohou vybublat vyse, protoze volajici
     * muze odchytavat nektere vyjimky a podle nich testovat, zda dana
     * zprava je "spatna", takze aby todle nezdetekovat, kdyz nejaka vybubla
     * z process_migration_group_messages(), coz by neznamenalo, ze tato
     * zprava je "spatna".
     *
     * Pozn. Moznost volat process_migration_group_messages()
     * na konci tedle funkce a na zacatek tedle funkce pridat
     * hope( groups_to_process.empty() ) nebude fungovat, protoze
     * mezi dvema handle_incoming_migration_group_message() se muzou
     * naschedulovat migracni zpravy do groups_to_process napriklad
     * ze zpravovani object status zpravy.
     *
     * Zatim bych rek ze to ale nevadi, ze tato zprava bude zpracovana
     * driv, nez ty v groups_to_process, nicemu by to vadit nemelo.
     * Poradi zprav je dulezite jen pro archivace, ale migracni zpravy
     * se stejnak obecne mohou predhanet libovolne.
     */

    ArchiveManager &        archive_manager = Global::archive_manager();
    const NodeId &          msg_sender = msg.get_sender();
    
    /** @todo FIXME Check if the message was received from client node.
     * In that case check that the destination is account object
     * and remap object ids of the objects stored in the message.
     */
    if( msg_sender.is_type( NodeId::TYPE_CLIENT ) )
        {
        // Objects created on the client has dummy object provider.
        // We must remap the object ids using valid object provider.
        /// @todo Verify hops count, destination, initiator, ...
        preprocess_migration_group_message_from_client( msg );        
        }
    else
        {
        hope( msg_sender.is_type( NodeId::TYPE_SERVER ) );
        
        /* Invalidate cached information about objects stored in the message. */
        clear_remote_object_status_cache( msg.objects );
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
        clear_migration_object_cache( msg.objects );
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
        }

    /* If archivation is in progress, we have to unpack the message
     * so the objects will be archived.
     * We take care only of objects from server nodes. We can't archive
     * objects from client nodes, because these objects have been
     * assigned new object ids, but archive file stores states of
     * providers from the beginning of the archivation!
     *
     * Note: It is not forbidden that message sender can be local node.
     *
     * Note: We do not need to test if message sender is local node,
     * because archivation_in_progress_on_node() takes care of this
     * correctly.
     */

    /* Note: The test must be done this way!!!
     * It will work even if the sender is local node.
     */
    if( msg_sender.is_type( NodeId::TYPE_SERVER ) &&
        archive_manager.local_archivation_in_progress() &&
        !archive_manager.archivation_in_progress_on_node( msg.get_sender() )
        )
        {
        // Local archivation is running...
        // and the objects was not archived on the sender.
        //
        // Create objects from the message so they will be archived.

        /* Note: The message is not passed to localization
         * protocol (process_migration_group_message()).
         * But it is not a bug. See doc/develop/communication_faq.txt
         * for more details.
         */
        deliver_objects_from_migration_message( msg );
        }
    else if( msg_sender.is_type( NodeId::TYPE_CLIENT ) )
        {
        // The message was received from a client.
        // Unpack the message so the data stored in the message
        // are verified and the client can be killed
        // if something is wrong. The verification includes
        // some rpc verification.
        // The unpack must be done on the first server which
        // received the message from the client so that there
        // will not migrate "bad" migration group messages
        // among server nodes. Also once a migration message
        // is accepted by one server, it can be unpacked
        // and repacked at any time and that will lost
        // sender and initiator ids of the original message.
        deliver_objects_from_migration_message( msg );
        }
    else
        {
        // Process the message.
        process_migration_group_message( msg );
        }
    }


void ObjectManager::process_find_object_message
    (
    FindObjectMessage & msg
    )
    {
    hope( msg.get_type() == SystemMessage::FIND_OBJECT );
    process_localization_message( SystemMessage::FIND_OBJECT, &msg, NULL );
    }


void ObjectManager::process_migration_group_message
    (
    MigrationGroupMessage & msg
    )
    {
    hope( msg.get_type() == SystemMessage::MIGRATION_GROUP ||
            msg.get_type() == SystemMessage::MIGRATION_GROUP_NONARCHIVABLE );
    process_localization_message( msg.get_type(), NULL, &msg );
    }


void ObjectManager::reply_to_localization_message_for_owned_object
    (
    const FindObjectMessage *   msg
    )
    {
    hope( msg != NULL );
    
    const NodeId &          local_node_id = Global::get_local_node_id();
    SystemMessage::Type     msg_type = msg->get_type();

    hope( msg_type == SystemMessage::FIND_OBJECT ||
            msg_type == SystemMessage::MIGRATION_GROUP ||
            msg_type == SystemMessage::MIGRATION_GROUP_NONARCHIVABLE );

    if( msg->get_initiator() == msg->get_sender() )
        {
        // The sender is initiator.
        if( msg->get_sender() != local_node_id )
            {
            // We have received the message from another node.
            switch( msg_type )
                {
                case SystemMessage::FIND_OBJECT:
                    {
                    /* Notify the searching (initiator) node that we have the object.
                     * Note: This is specific behavior of FindObject message. */
                    ObjectStatusMessage s( msg->destination, ObjectStatus::ALIVE, local_node_id, true );
                    s.send( msg->get_initiator(), NetworkCommon::STREAM_RELIABLE ); // Must be reliable.
                    }
                    break;
                case SystemMessage::MIGRATION_GROUP:
                case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                    /* Note: It is correct to do nothing here. We do NOT inform
                     * initiator node that we have the object, if the node
                     * is both initiator and sender. It is a behavior of object
                     * searching and cache update protocol.
                     * The initiator node should update its caches by itself.
                     */
                     break;
                default:
                    panic( "Invalid message type." );
                    break;
                }; // switch( msg_type )
            }
        else
            {
            /* msg->initiator == msg->sender == local_node_id.
             * The local system is trying to send message constructed locally
             * or the message was send from local system to local system
             * (should happen only in debug mode).
             *
             * CZ note: Teoreticky tohle muze take nastat, pokud
             * je v lokalni cachi jako cilovy node pro destination
             * tento node a soucasne tento node byl inicializator.
             * Ale tenhle pripad je osetren ve zpracovani ObjectStatus zprav
             * (v notify metodach) a take pri dotazovani cache je to testovano.
             */

            // Nothing to do here.
            }
        }
    else // if( msg->initiator() == msg->sender )
        {
        // The sender is not initiator.
        if( msg->get_sender() != local_node_id )
            {
            if( msg->get_initiator() != local_node_id )
                {
                /* Inform initiator node that we have the object
                 * so he doesn't need to forward other messages through sender any more. */
                 ObjectStatusMessage s( msg->destination, ObjectStatus::ALIVE, local_node_id, true );
                // Must be reliable because if the message is FindObject then
                // we have to send answer.
                /** @todo Ff the message is migration group, we ca use unreliable stream.
                 * But what is better? */
                s.send( msg->get_initiator(), NetworkCommon::STREAM_RELIABLE );
                }
            else
                {
                /* Owned object && msg->initiator == local_node_id.
                 * We was looking for some object, but during the search
                 * the object has become owned. So we do nothing here.
                 */
                }
            }
        else // if( msg->sender != local_node_id )
            {
            /* Localization error: msg->initiator != msg->sender == local_node_id.
             * Should never happen (I hope so).
             * But the object is owned, so no special action must be done,
             * only log this strange circumstance and invalidate
             * localization cache. */

            ObjectStatusMessage s( msg->destination, ObjectStatus::ALIVE, local_node_id, true );

            std::ostringstream ss;
            ss  << "Localization error: Unexpected localization message state while processing "
                << SystemMessage::type_to_string( msg_type )
                << " message for owned destination object (" << msg->destination << "). "
                << "Detected: initiator != sender && sender == local node. "
                << "Should never happen. Recovering from the error...";
            /** @todo Write message content? */
                
            Global::log
                (
                STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                ss.str()
                );

            /* Inform initiator node that we have the object. */
            s.send( msg->get_initiator(), NetworkCommon::STREAM_RELIABLE );

            /* Try to recover from the error. */
            clear_remote_object_status_cache( msg->destination );
            }         
        }
    }

            
void ObjectManager::process_localization_message
    (
    SystemMessage::Type     msg_type,
    FindObjectMessage *     find_object_msg,
    MigrationGroupMessage * migration_group_msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::process_localization_message()" );
    const NodeId            local_node_id = Global::get_local_node_id();
    FindObjectMessage *     msg;   // Pointer to the message. Always valid.
    const bool log_migration_message = ( migration_group_msg && settings.log_process_migration_messages.read() );
    
    switch( msg_type )
        {
        case SystemMessage::FIND_OBJECT:
            msg = find_object_msg;
            break;
        case SystemMessage::MIGRATION_GROUP:
        case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
            msg = migration_group_msg;
            break;
        default:
            panic( "Invalid message passed to ObjectManager::process_localization_message()!" );
            break;
        };

    hope( msg != NULL );
    hope( msg->get_type() == msg_type );

    hope( !msg->get_initiator().is_null() );
    hope( !msg->get_sender().is_null() );
    
    if( log_migration_message )
        {
        std::ostringstream oss;
        oss << "process_localization_message(): Processing: " << *migration_group_msg << ".";
        Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
        }

    // Note: True condition "msg.initiator == local_node_id" doesn't
    // have to indicate cycle in searching caches.
    // But everything will work fine even if there is a cycle.

    // Check whether the destination object is owned (ie. active or inactive).
    bool        dest_is_waiting = is_object_waiting( msg->destination );
    bool        dest_is_owned = dest_is_waiting;

    if( !dest_is_owned )
        {
        // The destination object is not waiting. Check if it is active or swapped-out.
        ObjectProperty * o = gc.find_local_object( msg->destination );
        dest_is_owned = ( o && !o->is_replica() ) || is_swapped_out( msg->destination );
        }
        
    // If the destination object is proxy node object id for local node
    // process the message like the object is owned.
    // Note: We MUST do this test here. If NodeObject representing local node
    // was created yet, dest_is_owned will be true (because the NodeObject has
    // same object id as proxy node object id for local node). But if NodeObject was
    // not created yet, dest_is_owned will be false. In that case,
    // we must deliver objects from the migration group because we must
    // call delivery_failed() on them.
    if( !dest_is_owned )
        {
        const ObjectId & dest_object_id = msg->destination;
        if( dest_object_id.is_proxy_node_object_id() &&
            local_node_id == dest_object_id.proxy_extract_node_id()
            )
            {
            // The message is for local node, but our NodeObject was not
            // created yet. Btw. it should never happen because NodeObject should
            // be created as soon as possible, but who knows. We will just
            // call delivery failed on these objects.
            // Unpack the objects so we can call delivery_failed() on them.
            dest_is_owned = true;
            }
        }

    if( dest_is_owned )
        {
        // The destination object is owned (active, waiting or swaped-out, but not replica).
        // Note: Condition "local_node_id == msg->get_initiator()" can be true.
        
        if( log_migration_message )
            {
            std::ostringstream oss;
            oss << "process_localization_message(): Destination object is owned.";
            Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
            }

#ifdef OPTION_DEBUG
        {
        // The destination object should not be in remote or migration cache.
        if( !msg->destination.is_proxy_node_object_id() )
            {
            ObjectStatus    os;
            NodeId          si;
            get_remote_object_status( msg->destination, os, si );
            hope( os == ObjectStatus::UNKNOWN );
            }
#ifdef MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
        hope( migration_object_cache.find( msg->destination, Global::system_time() ) == NULL );
#endif // MASSIV_OBJECT_MANAGER_MIGRATION_CACHE
        }
#endif // OPTION_DEBUG

        /* Before processing the message, satisfy localization protocol. */
        if( System::is_server() )
            {
            reply_to_localization_message_for_owned_object( msg );
            }

        switch( msg_type )
            {
            case SystemMessage::MIGRATION_GROUP:
            case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                if( dest_is_waiting )
                    {
                    if( log_migration_message )
                        {
                        std::ostringstream oss;
                        oss << "process_localization_message(): Destination object is waiting in some queue."
                            << " Putting the message to waiting queue.";
                        Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                        }
                        
                    // The destination is waiting in some queue.
                    add_migration_message_to_waiting_queue( *migration_group_msg );
                    }
                else
                    {
                    if( log_migration_message )
                        {
                        std::ostringstream oss;
                        oss << "process_localization_message(): Destination object is not waiting in any queue."
                            << " Unpacking objects from the message.";
                        Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    // Unpack objects from the message.
                    deliver_objects_from_migration_message( *migration_group_msg );
                    }
                break;
            default:
                // Nothing to do.
                break;
            };
        }
    else // if( dest_is_owned )
        {
        /* The object is not owned (but maybe we have its replica). */
        
        if( log_migration_message )
            {
            std::ostringstream oss;
            oss << "process_localization_message(): Destination object is not owned.";
            Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
            }

        ObjectStatus    dest_status; // Destination object status.
        NodeId          dest_node;   // On which node the destination object is (probably) located.
        bool            invalidate_cache = false;

        /* Check if we have some cached status of the destination object.
         * If destination is proxy node object, the get_remote_object_status()
         * will check if the node associated with the node object is
         * acessible. If it is, it will result ALIVE status, if not, it
         * will return NOT_ACCESSIBLE status.
         */
        get_remote_object_status( msg->destination, dest_status, dest_node );
        
        /* Check strange circumstances. */
        // Note: sender == initiator == sid reserved for sending
        // messages which was constructed locally.
        if( msg->get_sender() != msg->get_initiator() && msg->get_sender() == local_node_id )
            {
            std::ostringstream ss;
            ss  << "Localization error: Unexpected  localization message status while processing "
                << SystemMessage::type_to_string( msg_type )
                << " message for not owned object (" << msg->destination << "). "
                << "Detected: sender != initiator && sender == local node. "
                << "Should never happen. Recovering from the error...";
            /** @todo Write message content? */

            Global::log
                (
                STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                ss.str()
                );

            invalidate_cache = true;
            }

        if( dest_status == ObjectStatus::ALIVE )
            {
            // Note: dest_node may be proxy node object, but it is not
            // proxy node object for local node because it is tested at
            // the beginning of process_localization_message().
            
            if( dest_node == local_node_id )
                {
                /* Internal cache error. */

                std::ostringstream ss;
                ss  << "Localization cache error: Mismatched remote object localization cache content while processing "
                    << SystemMessage::type_to_string( msg_type )
                    << " message for not owned object (" << msg->destination << "). "
                    << "Cached status: Remote object located on local system. "
                    << "Should never happen. Recovering from the error...";
                /** @todo write message content, cache content? */

                Global::log
                    (
                    STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                    ss.str()
                    );

                invalidate_cache = true;
                }

            if( dest_node == msg->get_sender() )
                {
                /* Message from sender received, but the destination object should
                 * be located on the sender.
                 * Note that this doesn't mean that there must be cycle in localization
                 * caches. Often it means the dest object was on our node when
                 * sender sent the message, but before the message was delivered to
                 * us, the object has been sent to the sender.
                 * This often happen when load balancer performs migrations
                 * to lower cpu and memory load of local node or an objects
                 * migrates really often between two nodes.
                 *
                 * We do not invalidate localization cache because the probability
                 * that there is a cycle in localization caches is lesser then the probability
                 * that this is not error. If there is a cycle in the caches,
                 * it will be detected by hops and provider_hops.
                 */
                
                // Does anybody want to log this?
                /*
                std::ostringstream ss;
                ss  << "Detected possible cycle in remote object localization cache while processing "
                    << SystemMessage::type_to_string( msg_type )
                    << " message for remote object (" << msg->destination << "). "
                    << "Message received from expected object owner. ";

                Global::log
                    (
                    STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_LOW ),
                    ss.str()
                    );
                */
                }
            }

        if( msg->destination.is_tracked_by_provider() &&
            is_local_provider( msg->destination.get_provider_id() ) )
            {
            hope( (int) FindObjectMessage::PROVIDER_HOPS_MAX_VALUE >= (int) MAX_FIND_OBJECT_MESSAGE_PROVIDER_HOPS_COUNT );

            if( msg->provider_hops > MAX_FIND_OBJECT_MESSAGE_PROVIDER_HOPS_COUNT )
                {
                /* Invalidate provider cache and do broadcast search. */
                invalidate_cache = true;
                /** @todo Zresetovat provider_hops na 1?
                 * Asi jo. Stejnak to de asi do kytek, protoze to jedine
                 * co muzem delat je zresetovat lokalni cache, diky cemuz
                 * se udela broadcast search, ale ten pouzije vysledky
                 * z cizich cache, ktera muze byt taky spatne, tudiz
                 * se bude delat broadcast tak dlouho, doku se cache
                 * nevytimeouti...
                 * Tak to neni uplne pravda, viz. global search, ten totiz
                 * nepouzije vysledky z cizi cache, dokud si neni jisty,
                 * ze objekt neni nekde owned.
                 */
                }
            else
                {
                msg->provider_hops++;
                }
            msg->hops = 1; /* Reset hops count because provider's cache should be up to date. */
            }

        /* Do not allow packet to be forwarded too many times. */
        if( msg->hops >= MAX_FIND_OBJECT_MESSAGE_HOPS_COUNT )
            {
            dest_status = ObjectStatus::UNKNOWN;
            dest_node.clear();
            msg->hops = 1; /* Reset hops count. We do not set it to zero because
                            * zero is reserved for initial message send. */
            /* Note: We do not invalidate cache. */
            // FIXME Tohle doresit na schuzce - mozna by se mela cache
            // invalidovat...
            }

        /* Invalidate local cache if necessary. */
        if( invalidate_cache )
            {
            dest_status = ObjectStatus::UNKNOWN;
            dest_node.clear();
            clear_remote_object_status_cache( msg->destination );
            }

        if( dest_status == ObjectStatus::ALIVE )
            {
            if( log_migration_message )
                {
                std::ostringstream oss;
                oss << "process_localization_message(): Destination object found on node "
                    << dest_node << ". Sending the message to the node.";
                Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            switch( msg_type )
                {
                case SystemMessage::FIND_OBJECT:
                    send_find_object_message( *find_object_msg, dest_node );
                    break;
                case SystemMessage::MIGRATION_GROUP:
                case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                    send_migration_message( *migration_group_msg, dest_node );
                    break;
                default:
                    panic( "Invalid message type." );
                    break;
                }
            }
        else if( dest_status == ObjectStatus::NOT_ACCESSIBLE )
            {
            
            if( log_migration_message )
                {
                std::ostringstream oss;
                oss << "process_localization_message(): Destination object is not accessible in simulation."
                    << " Unpacking objects from the message to call delivery_failed().";
                Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                }
            
            /* Inform initiator node. */
            if( System::is_server() )
                {
                ObjectStatusMessage s( msg->destination, dest_status, local_node_id, false );
                s.send( msg->get_initiator(), NetworkCommon::STREAM_RELIABLE );
                }

            if( msg_type == SystemMessage::MIGRATION_GROUP ||
                msg_type == SystemMessage::MIGRATION_GROUP_NONARCHIVABLE )
                {
                /* Notify the objects from the messaget that thay can't
                 * be delivered. */
                deliver_objects_from_migration_message( *migration_group_msg );
                }

            /* Refresh cache for case that somebody will ask again. */
            /** @todo Pozor, musi se specialne resit pripad kdy se
             * pripojuje novy node do jiz bezici simulace,
             * musi se zinvalidovat vsechny NOT_ACCESSIBLE cached statusy
             * na jiz bezicich nodech o objektech, ktere jsou
             * na pripojujicim se nodu. */
            refresh_remote_object_status_cache_timeout( msg->destination );
            }
        else
            {
            /* Unknown object status.
             * Forward request to object provider or do global search.
             */
             
            if( log_migration_message )
                {
                std::ostringstream oss;
                oss << "process_localization_message(): We don't know what to do with message "
                    << "because we don't know anything about the destination.";
                Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                }

            /*
             * Note: If the provider is located on this node and
             * object is tracked, we must do broadcast search.
             */
            /*
             * Note: The provider_hops_count of the message was already processed.
             */

            // Get owner node of the provider.
            NodeId provider_node_id = get_provider_owner_node( msg->destination.get_provider_id(), false );
            // Note: If the provider is local then provider_node_id is null now!

            // If the destination is tracked and the provider exists...
            if( msg->destination.is_tracked_by_provider() &&
                !provider_node_id.is_null() )
                {
                if( log_migration_message )
                    {
                    std::ostringstream oss;
                    oss << "process_localization_message(): Sending the message to object provider "
                        << msg->destination.get_provider_id() << " to node " << provider_node_id
                        << ". The provider should know where the destination object id.";
                    Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                    }
                
                // Forward the message to the provider.
                switch( msg_type )
                    {
                    case SystemMessage::FIND_OBJECT:
                        send_find_object_message( *find_object_msg, provider_node_id );
                        break;
                    case SystemMessage::MIGRATION_GROUP:
                    case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                        send_migration_message( *migration_group_msg, provider_node_id );
                        break;
                    default:
                        panic( "Invalid message type." );
                        break;
                    }
                }
            else
                {
                if( System::is_client() )
                    {
                    // Client node can't do broadcast search.
                    // We will send the message to our forward node.
                    
                    if( log_migration_message )
                        {
                        std::ostringstream oss;
                        oss << "process_localization_message(): Because local node is client "
                            << "and client nodes can't initiate global search, we will send "
                            << "the message to our forward node.";
                        Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                        }
                    
                    NodeManager & node_manager = Global::node_manager();
                    
                    NodeId dest_node = node_manager.get_local_system_forward_node();
                    
                    switch( msg_type )
                        {
                        case SystemMessage::FIND_OBJECT:
                            send_find_object_message( *find_object_msg, dest_node );
                            break;
                        case SystemMessage::MIGRATION_GROUP:
                        case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                            send_migration_message( *migration_group_msg, dest_node );
                            break;
                        default:
                            panic( "Invalid message type." );
                            break;
                        }
                    }
                else
                    {
                    
                    if( log_migration_message )
                        {
                        std::ostringstream oss;
                        oss << "process_localization_message(): Initiating global search. "
                            << "Adding the message to waiting queue.";
                        Global::log_debug( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_NORMAL, oss.str() );
                        }

                    // Do global search for the object.
                    GlobalSearchManager & global_search_manager = Global::global_search_manager();
                    
                    // Store the message into waiting queue.
                    switch( msg_type )
                        {
                        case SystemMessage::MIGRATION_GROUP:
                        case SystemMessage::MIGRATION_GROUP_NONARCHIVABLE:
                            add_migration_message_to_waiting_queue( *migration_group_msg );
                            break;
                        
                        default:
                            panic( "Global search not implemented for FindObjectMessage!" );
                            break;
                        }
                    
                    // Start new global search for the message destination.
                    global_search_manager.initiate_search_for( msg->destination );
                    }
                }
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
