/** @file object_manager_packing.cpp
 * Implementation of ObjectManager's methods which
 * pack/unpack objects to/from migration group messages,
 * replication group messages or swap file.
 */

/*
 * begin        : Tue Nov 19 2002
 * copyright    : (C) 2002 Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "object/object.h"
#include "object/object_property_data.h"
#include "object/object_factory.h"
#include "object/object_manager.h"
#include "object/object_common.h"
#include "sysevent/object_status_msg.h"
#include "sysevent/migration_group_msg.h"
#include "serializer/serializer.h"
#include "object/archive.h"
#include "object/archive_manager.h"
#include "object/collect_group.h"
#include "object/object_property_data.h"
#include "system/system.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "system/tick_lock.h"
#include "replication/replica_client.h"
#include <memory>
#include <sstream>


#ifdef OPTION_DEBUG

/**
 * Define this to enable logging of migration group packing/unpacking.
 */
//#define LOG_PACKING

#endif


namespace Massiv {
namespace Core {


ObjectManager::MigrationPackResult ObjectManager::pack_migration_group_to_migration_message
    (
    ObjectProperty *        object_from_migration_group,
    MigrationGroupMessage & migration_msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::pack_migration_group_to_migration_message( ObjectProperty *, MigrationGroupMessage & )" );
    hope( !object_from_migration_group->is_replica() );
    
    ObjectShared::ObjectList    group_objects;
    CollectObjectGroupOperation collect_operation( group_objects );

    #ifdef LOG_PACKING

    Global::log_debug
        (
        Status::FACILITY_OBJECT_MANAGEMENT,
        Status::PRIORITY_NORMAL,
        "Packing migration group of " + 
        object_from_migration_group->get_object_id().to_string() + "."
        );

    #endif

    // Enumerate migration group.
    gc.enumerate_migration_group( object_from_migration_group, collect_operation );

    hope( !group_objects.empty() );
    hope( group_objects.front() == object_from_migration_group );

    return pack_migration_group_to_migration_message( group_objects, migration_msg );
    }


ObjectManager::MigrationPackResult ObjectManager::pack_migration_group_to_migration_message
    (
    const ObjectShared::ObjectList & group_objects,
    MigrationGroupMessage &          migration_msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::pack_migration_group_to_migration_message( ObjectIdList &, MigrationGroupMessage & )" );
    hope( !group_objects.empty() );
    hope( migration_msg.objects_count == 0 );
    hope( migration_msg.objects.empty() );
    hope( !migration_msg.destination.is_null() );

    // Check if some of the objects are pinned or stack referenced.
    if( is_object_group_pinned_or_stack_referenced( group_objects ) )
        {
        return MIGRATION_PACK_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED;
        }

    Serializer::Description     desc;
    desc.object_flags = Serializer::Description::OBJECT_MIGRATE;
    desc.destination_object = migration_msg.destination;
    BitOStream                  bos;  /** @todo FIXME At se bos stale nealokuje dynamicky pro kazdou zpravu... */
    ObjectShared::ObjectList::const_iterator    object_it;

    // Serialize objects from the migration group.
    for( object_it = group_objects.begin(); object_it != group_objects.end(); ++object_it )
        {
        ObjectProperty * object = ( *object_it );
        hope( !object->is_replica() );
        
        // Can object be migrated?
        if( object->get_metaobject()->get_attribute_no_migrate() )
            {
            return MIGRATION_PACK_RESULT_OBJECT_MAY_NOT_MIGRATE;
            }
        
        #ifdef LOG_PACKING
        
        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            "Packing " + object->get_object_id().to_string() +
            " (" + object->get_metaobject()->get_class_type_id().dump() + ")."
            );
        
        #endif

        object->bitstr_write( bos, desc );
        }
        
    #ifdef LOG_PACKING
    
    {
    std::ostringstream ss;

    ss << "Serialization consumed " << bos.byte_size() << " byte(s).";

    Global::log_debug
        (
        Status::FACILITY_OBJECT_MANAGEMENT,
        Status::PRIORITY_LOW,
        ss.str()
        );
    }

    #endif
    
    // Fill the migration message.
    migration_msg.objects_count = group_objects.size();
    migration_msg.message_stream.bind( bos.get_buffer() );
    migration_msg.objects.reserve( group_objects.size() );
    for( object_it = group_objects.begin(); object_it != group_objects.end(); ++object_it )
        {
        ObjectProperty * object = ( *object_it );
        migration_msg.objects.push_back( object->get_object_id() );
        }

    // And finally destroy all of the objects.
    pack_destroy_objects_prepared_for_migration( group_objects );

    statistics.num_packed_objects += ( int ) group_objects.size();

    return MIGRATION_PACK_RESULT_SUCCESS;
    }


//-----------------------------------------------------------------------------


void ObjectManager::deliver_objects_from_migration_message
    (
    const MigrationGroupMessage & msg
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::deliver_objects_from_migration_message()" );
    hope( msg.objects_count == msg.objects.size() );
    hope( msg.objects_count > 0 );

    ObjectShared::ObjectList    objects; // Objects from the message.

    ArchiveManager &            archive_manager = Global::archive_manager();
    Serializer::Description     desc;
    desc.object_flags = Serializer::Description::OBJECT_MIGRATE;
    desc.source_node = msg.get_sender();
    BitIStream                  message_stream = msg.message_stream; // Because 'msg' is const.

    /* Deliver objects from the message to their destination. */
    // + Zrusit repliky objektu z message. (ale viz. @bug v unpack_create_objects)
    // + Vytvorit objekty.
    // + Vyserializovat objekty z message.
    // + Vyserializovat eventy objektu a naschedulovat objekty do global schedule.
    // + Oznacit objekty k archivaci, pokud bezi.
    // + Informovat jejich providery - pouze remote, lokalni cache uz jsou prazdne.
    //   Problem - viz. nize.
    // + Zpracovat vsechny migracni zpravy, ktere cekaji
    //   na nektery z objektu, ktery byl vytvoren.
    // + Catchnout vyjimky (kdekoli) a smazat vytvorene objekty. Zde je problem,
    //   pokud destination neexistuje a zavolaji se callbacky, pokud se objekty
    //   znici, tak automaticky poslou notifikaci providerovi...

    #ifdef LOG_PACKING
    
        {
        std::ostringstream ss;
        ss << "Unpacking migration group:\n" << msg;
        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            ss.str()
            );
        }

    #endif

    try
        {
        // Create the objects stored in the message.
        unpack_create_objects( msg.objects, objects );

        // Extract the objects from the message stream.
        // Note that we can't enter objects to archive mode during this.
        unpack_bitstr_read_objects( objects, message_stream, desc );

        if( msg.get_sender().is_type( NodeId::TYPE_CLIENT ) &&
            Massiv::System::get_local_node_id().is_type( NodeId::TYPE_SERVER ) )
            {
            // Checked in preprocess_migration_group_message_from_client.
            hope( objects.size() == 1 );

            const ObjectProperty * const object = objects.front();
            hope( object );
            hope( object->root_data );
            object->root_data->events.check_client_to_server_events
                (
                msg.get_sender(),
                object->get_object_id(),
                msg.destination
                );
            }

        /* If archivation is in progress, we have to archive the objects.
         * We take care only of objects from server nodes. We can't archive
         * objects from client nodes, because these objects have been
         * assigned new object ids, but archive file stores states of
         * providers from the beginning of the archivation!
         *
         * Note: It is not forbidden that message sender can be local node.
         *
         * Note: We do not need to test if message sender is local node,
         * because archivation_in_progress_on_node() takes care of this
         * correctly.
         */
        /* Note: The test must be done this way!!! */
        if( msg.get_sender().is_type( NodeId::TYPE_SERVER ) &&
            archive_manager.local_archivation_in_progress() &&
            !archive_manager.archivation_in_progress_on_node( msg.get_sender() )
            )
            {
            // Local archivation is running and the objects was not archived on the sender.
            unpack_schedule_objects_for_archivation( objects );
            }

        // Notify remote providers that we have the objects.
        // Note: We notify only remote providers. Because the objects
        // are now local, cache of local providers is already updated
        // (as a part of handle_migration_message()).
        if( System::is_server() )
            {
            unpack_notify_remote_providers_of_owned_objects( msg.objects );
            }

        /* And finally process all messages that are waiting
         * for just created objects. */
        unpack_refresh_migration_messages_waiting_for_objects( msg.objects );

        statistics.num_unpacked_objects += ( int ) msg.objects.size();
        }
    catch( ExSerializationError & e )
        {
        unpack_exception_clean_up( objects );

            {
            std::ostringstream ss;
            ss  << "Migration group unpacking failed because of a serialization error. "
                << e.what() << "\nGroup content:\n" << msg;
            throw ExMigrationGroupUnpackSerializationError( ss.str() );
            }
        }
    catch( Exception & e )
        {
        unpack_exception_clean_up( objects );

            {
            std::ostringstream ss;
            ss  << "Migration group unpacking failed beause of an exception. "
                << e.what() << "\nGroup content:\n" << msg;
            throw ExMigrationGroupUnpackError( ss.str() );
            }
        }
    catch( ... )
        {
        unpack_exception_clean_up( objects );

            {
            std::ostringstream ss;
            ss  << "Migration group unpacking failed because of an unknown exception.\n"
                << "Group content:\n" << msg;
            throw ExMigrationGroupUnpackError( ss.str() );
            }
        }

    }


//-----------------------------------------------------------------------------


    //
    // Object swapping.
    //


ObjectManager::SwapOutResult ObjectManager::try_to_swap_out
    (
    const ObjectId & object_id
    )
    {
    ObjectProperty * object = gc.find_local_object( object_id );
    
    if( !object )
        {
        return SWAP_OUT_RESULT_UNREGISTERED_OBJECT;
        }

    return swap_out_object( object );
    }


ObjectManager::SwapOutResult ObjectManager::swap_out_object
    (
    ObjectProperty * object
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_out_object( ObjectProperty * )" );
    hope( object );
    
    // Replicas can't be swapped out.
    if( object->is_replica() )
        {
        return SWAP_OUT_RESULT_IS_REPLICA;
        }

    // Objects replicated to other nodes can't be swapped out.
    if( object->has_replica_requests() )
        {
        return SWAP_OUT_RESULT_HAS_REPLICA_REQUESTS;
        }
        
    // Only archivable objects may be stored inside an archive file
    // and because swap file is also an archive file, nonarchivable
    // objects can't be swapped out.
    if( !object->is_archivable() )
        {
        return SWAP_OUT_RESULT_OBJECT_NOT_ARCHIVABLE;
        }

    ObjectShared::ObjectList    object_list; // Entire migration group.
    CollectObjectGroupOperation collect_operation( object_list );

    // Enumerate migration group.
    gc.enumerate_migration_group( object, collect_operation );

    hope( !object_list.empty() );
    hope( object_list.front() == object );

    // Log.
        {
        std::ostringstream ss;
        ss << "Trying to swap-out " << object_list.size() << " object(s).";

        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );
        }

    return swap_out_objects( object_list );
    }


ObjectManager::SwapOutResult ObjectManager::swap_out_objects
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_out_objects()" );
    ArchiveManager & archive_manager = Global::archive_manager();

    // Check if swap file is opened.
    if( !swap )
        {
        return SWAP_OUT_RESULT_NO_SWAP_FILE;
        }

    // Swap-out must be disabled if archivation is in progress.
    // Otherwise one object could be archived two or more times
    // (not on local system but in the distributed system).
    // See develop docs.
    if( archive_manager.archivation_in_progress() )
        {
        return SWAP_OUT_RESULT_ARCHIVATION_IN_PROGRESS;
        }

    // Check if some of the objects are pinned or stack referenced.
    if( is_object_group_pinned_or_stack_referenced( object_list ) )
        {
        return SWAP_OUT_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED;
        }

    // Register the objects to the swap (into new archive group ).
    CommonArchiveOperations::register_objects_to_archive( swap, object_list );

    // Serialize the objects to the swap.
        {
        Serializer::Description                     desc;
        ObjectShared::ObjectList::const_iterator    it;
        std::ostringstream                          ss;
        
        desc.object_flags = Serializer::Description::OBJECT_SWAP;

        for( it = object_list.begin(); it != object_list.end(); ++it )
            {
            ObjectProperty * object = ( *it );
            hope( !object->is_replica() );
            hope( object->is_archivable() );
            
            // Log.
            ss.str( std::string() );
            
            ss  << "swap-out: " << object->get_object_id() 
                << " touch_time = " << object->get_touch_time()
                << " first_event_time = " << ( object->has_scheduled_event() ? 
                object->get_time_of_first_event() : STime() );

            Global::log
                (
                STATUS_INFO
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_LOWEST
                    ),
                ss.str()
                );

            // Write object.
            CommonArchiveOperations::text_write_object_to_archive( swap, object, desc );
            }
        }

    // And finally destroy all of the objects.
    pack_destroy_objects_prepared_for_migration( object_list );

    statistics.num_swapped_out_objects += ( int ) object_list.size();

    return SWAP_OUT_RESULT_SUCCESS;
    }


ObjectProperty * ObjectManager::try_to_swap_in
    (
    const ObjectId & object_id
    )
    {
    if( !swap )
        return NULL;

    if( !swap->is_object_registered( object_id ) )
        return NULL;

    Archive::GroupId              group_id;
    ObjectShared::ObjectIdList    ids;

    // Get the migration (archive) group.
    group_id = swap->collect_group_objects( object_id, ids );
    hope( !ids.empty() );

    swap_in( ids );
    swap->destroy_group( group_id );

    return gc.find_local_object( object_id );
    }
    

bool ObjectManager::is_swapped_out
    (
    const ObjectId & object_id
    )
    {
    return swap ? swap->is_object_registered( object_id ) : false;
    }


void ObjectManager::swap_in_all_objects()
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_in_all_objects()" );
    Archive::GroupIdList                    groups;
    Archive::GroupIdList::iterator          it;
    Archive::GroupId                        group_id;
    ObjectShared::ObjectIdList              ids;

    if( !swap )
        return;

    // Log.
        {
        std::ostringstream ss;
        ss  << "Swapping in all objects from the swap. "
            << swap->get_registered_object_count() << " objects in "
            << swap->get_registered_group_count() << " groups. ";
        Global::log
            (
            STATUS_INFO
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_HIGHEST
                ),
            ss.str()
            );
        }

    swap->enumerate_registered_groups( groups );

    for( it = groups.begin(); it != groups.end(); it++ )
        {
        group_id = ( *it );

        // Swap in the group.
        swap->collect_group_objects( group_id, ids );
        hope( !ids.empty() );

        swap_in( ids );
        swap->destroy_group( group_id );
        }
    }


void ObjectManager::swap_in_objects
    (
    const STime & time
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_in_objects( STime & )" );
    Archive::GroupId              group_id;
    ObjectShared::ObjectIdList    ids;

    if( !swap )
        return;

    while( 1 )
        {
        // Check if there is some group scheduled for the time.
        group_id = swap->get_scheduled_group( time );
        if( group_id == Archive::INVALID_GROUP_ID )
            break;

        // Log.
            {
            std::ostringstream ss;
            ss << "Swapping-in archive group " << swap->debug_group_info_string( group_id ) << ".";
            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_NORMAL
                    ),
                ss.str()
                );
            }

        // Swap in the group.
        swap->collect_group_objects( group_id, ids );
        hope( !ids.empty() );

        swap_in( ids );
        swap->destroy_group( group_id );
        }
    }


void ObjectManager::swap_in
    (
    const ObjectShared::ObjectIdList & ids
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_in( ObjectIdList & )" );
    hope( swap != NULL );

    ObjectShared::ObjectList     objects;

    /* Restore objects from the swap. */
    // + Vytvorit objekty (objekt nesmi lokalne existovat, a to ani jeho replika).
    // + Vyserializovat objekty ze swapu.
    // + Vyserializovat eventy objektu a naschedulovat objekty do global schedule.
    // + Zkopirovat do archivu pokud bezi archivace a objekty nebyly jeste zkopirovany.
    // + Provideri objektu nejsou informovani.
    // + Na zadny z vytvorenych objektu by nemel zadny jiny objekt cekat
    //   ve groups_waiting_for fronte. Nicmene ve fronte groups_to_process
    //   cekat muze.
    // + Catchnout vyjimky (kdekoli) a smazat vytvorene objekty. Zde je problem,
    //   pokud destination neexistuje a zavolaji se callbacky, pokud se objekty
    //   znici, tak automaticky poslou notifikaci providerovi...
    // + Objekty nejsou odstraneny ze swapu - to je povinosti volajiciho.


    // Log
        {
        std::ostringstream ss;
        ss  << "Trying to swap-in " << ids.size() << " object(s).";

        Global::log
            (
            STATUS_DEBUG
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_NORMAL
                ),
            ss.str()
            );
        }

    try 
        {
        // Create the objects. The objects are copied from the swap to archive
        // if archivation is in progress.
        swap_in_create_objects( ids, objects );

        // Extract the objects from the swap.
        swap_in_text_read_objects_from_swap( objects );

        // Log.
            {
            std::ostringstream ss;
            ss  << "Swapped in " << ids.size() << " object(s) from the swap. ";
            Global::log
                (
                STATUS_DEBUG
                    (
                    Status::FACILITY_OBJECT_MANAGEMENT,
                    Status::PRIORITY_LOWEST
                    ),
                ss.str()
                );
            }

        statistics.num_swapped_in_objects += ( int ) ids.size();
        }
    catch( std::exception & e )
        {
        swap_in_exception_clean_up( objects );

            {
            std::ostringstream ss;
            ss  << "Swap-in failed because of an exception. "
                << e.what();
            throw ExSwapInError( ss.str() );
            }
        }
    catch( ... )
        {
        swap_in_exception_clean_up( objects );

            {
            std::ostringstream ss;
            ss  << "Swap-in failed because of an unknown exception.";
            throw ExSwapInError( ss.str() );
            }
        }
    }


//-----------------------------------------------------------------------------


void ObjectManager::pack_destroy_objects_prepared_for_migration
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::pack_destroy_objects_prepared_for_migration()" );
    ObjectShared::ObjectList::const_iterator it;

    // First of all unmanage the objects.
    // This will prevent gc to destroy any object from the list
    // due to reference counting (while we will destroy the objects
    // one by one).
    // Note: Ref-counting is no more supported, but the code
    // stays here for future compatibility.
    gc_unmanage_objects( object_list );

    // Destroy object one by one.
    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty *    object = ( *it );

#ifdef OPTION_DEBUG
        ObjectId        object_id = object->get_object_id();
        ObjectStatus    object_status;
        NodeId          object_node;
#endif // OPTION_DEBUG

        // Destroy the object.
        hope( !object->is_finalizing() );
        hope( !object->is_stack_referenced() );
        object->delete_object( Object::DELETE_MIGRATED );

#ifdef OPTION_DEBUG
        hope( gc.find_local_object( object_id ) == NULL );
        hope( !is_object_waiting( object_id ) );
        if( !object_id.is_proxy_node_object_id() )
            {
            get_remote_object_status( object_id, object_status, object_node );
            hope( object_status == ObjectStatus::UNKNOWN );
            }
#endif // OPTION_DEBUG
        }
    }


void ObjectManager::unpack_create_objects
    (
    const ObjectShared::ObjectIdList &  object_id_list,
    ObjectShared::ObjectList &          object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::unpack_create_objects()" );
    hope( object_list.empty() );

    ObjectShared::ObjectIdList::const_iterator it;

    // We may have replicas of some objects in @a object_id_list. They must be
    // properly destroyed and replica consistency information must be updated.
    // Let replica_client handle it.

    hope( replica_client.get() );
    replica_client->handle_migration( object_id_list );

    object_list.reserve( object_id_list.size() );
    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId &            object_id = *it;
        hope( object_id.is_object_info_available() );
        const ObjectInfo &          object_info = object_id.get_object_info();
        hope( object_info.is_class_registered() );

        // Object should not exist. If we had its replica, replica_client should
        // have destroyed it.

        hope( !gc.find_local_object( object_id ) );
        hope( !is_swapped_out( object_id ) );

        // Create the object.
        
        const ObjectFactory * const object_factory = object_info.get_class_type_id().get_factory();
        ObjectProperty * const object = object_factory->create_object( object_id );
        hope( object );
        object_list.push_back( object );
        }
    }


void ObjectManager::unpack_bitstr_read_objects
    (
    const ObjectShared::ObjectList & object_list,
    BitIStream &                     bis,
    const Serializer::Description &  desc
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::unpack_bitstr_read_objects()" );
    ObjectShared::ObjectList::const_iterator    it;

    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty * object = ( *it );

        #ifdef LOG_PACKING

        Global::log_debug
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_NORMAL,
            "Reading content of " + object->get_object_id().to_string() +
            " (" + object->get_metaobject()->get_class_type_id().dump() + ")."
            );

        #endif

        // Note: The object is automatically scheduled to global schedule.
        object->bitstr_read( bis, desc );

        // Note that we can't enter the object to archive mode now.
        }
    }


void ObjectManager::unpack_schedule_objects_for_archivation
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    ArchiveManager & archive_manager = Global::archive_manager();
    ObjectShared::ObjectList::const_iterator it;

    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty * object = ( *it );

        if( object->is_archivable() )
            {
            object->enter_archive_mode();
            archive_manager.schedule_object_for_archivation( object->get_object_id() );
            }
        }
    }


void ObjectManager::unpack_notify_remote_providers_of_owned_objects
    (
    const ObjectShared::ObjectIdList & object_id_list
    )
    {
    ObjectShared::ObjectIdList::const_iterator  it;
    NodeId                                      local_node_id( Global::get_local_node_id() );

    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId & object_id = *it;
#ifdef OPTION_DEBUG
            {
            // Local objects should not be in localization cache.
            if( !object_id.is_proxy_node_object_id() )
                {
                ObjectStatus     o_status;
                NodeId           sid;
                get_remote_object_status( object_id, o_status, sid );
                hope( o_status == ObjectStatus::UNKNOWN );
                hope( sid.is_null() );
                }
            }
#endif // OPTION_DEBUG

        if( object_id.is_tracked_by_provider() ) // Notify of tracked objects only.
            {
            NodeId    remote_node_id( get_provider_owner_node( object_id.get_provider_id(), false ) );
            // Note: If the provider is local then remote_sid is null now!
            if( !remote_node_id.is_null() )
                {
                ObjectStatusMessage status_msg
                    (
                    object_id,              // Object id.
                    ObjectStatus::ALIVE,    // Object is local and alive.
                    local_node_id,          // Local node id.
                    true                    // Object is local so mark the message as trustworthy.
                    );

                status_msg.send( remote_node_id, NetworkCommon::STREAM_RELIABLE );
                }
            }
        }
    }


void ObjectManager::unpack_refresh_migration_messages_waiting_for_objects
    (
    const ObjectShared::ObjectIdList & object_id_list
    )
    {
    ObjectShared::ObjectIdList::const_iterator it;

    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId &           object_id = *it;
        hope( !is_object_waiting( object_id ) );
        refresh_migration_messages_waiting_for( object_id );
        }
    }


void ObjectManager::swap_in_create_objects
    (
    const ObjectShared::ObjectIdList &  object_id_list,
    ObjectShared::ObjectList &          object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_in_create_objects()" );
    hope( object_list.empty() );

    ArchiveManager & archive_manager = Global::archive_manager();
    ObjectShared::ObjectIdList::const_iterator  it;
    std::ostringstream                          ss;

    object_list.reserve( object_id_list.size() );
    for( it = object_id_list.begin(); it != object_id_list.end(); ++it )
        {
        const ObjectId &            object_id = *it;
        hope( object_id.is_object_info_available() );
        const ObjectInfo &          object_info = object_id.get_object_info();
        hope( object_info.is_class_registered() );

        // Log.
        ss.str( std::string() );
        ss << "swap-in: " << object_id;

        Global::log
            (
            STATUS_INFO
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_LOWEST
                ),
            ss.str()
            );

        // Object should not exist. Neither local nor replica.
        hope( !gc.find_local_object( object_id ) );
        
        const ObjectFactory *       factory = object_info.get_class_type_id().get_factory();
        ObjectProperty *            object = factory->create_object( object_id );

        object_list.push_back( object );
        hope( groups_waiting_for.count( object_id ) == 0 );

        // Archive the object.
        if( archive_manager.local_archivation_in_progress() ) // DO NOT USE archive_manager.archivation_in_progress()!!!!!!
            {
            hope( object_info.is_archivable() );
            archive_manager.archive_swap_in_object
                (
                object_id
                );
            }
        }
    }


void ObjectManager::swap_in_text_read_objects_from_swap
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::ObjectManager::swap_in_text_read_objects_from_swap()" );
    Serializer::Description                     desc;
    desc.object_flags = Serializer::Description::OBJECT_SWAP;
    ObjectShared::ObjectList::const_iterator    it;

    hope( swap );

    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty *            object = ( *it );
        const ObjectId &            object_id = object->get_object_id();
        ObjectId                    read_object_id;
        Archive::InputStream        input_stream;
        Archive::GroupId            group_id = swap->get_object_group_id( object_id );

        input_stream.open( swap, group_id, object_id );

        if( input_stream.fail() )
            {
            std::ostringstream ss;
            ss  << "Unable to open swap object input stream. "
                << "Archive: " << swap->debug_info_string() << ", "
                << "object: " << swap->debug_object_info_string( group_id, object_id )
                << ".";
            throw ExCanNotOpenArchiveStream( ss.str() );
            }

        TextReader       tr( input_stream );
        
        // Read complete object id part.
        tr.next_line();
        read_object_id.text_read( tr, desc );
        if( object_id != read_object_id )
            {
            std::ostringstream ss;
            ss  << "Corrupted swap file. Object id (" << read_object_id
                << ") read from object stream differs from object id ("
                << object_id << ") stored in the swap structures.";
            throw ExSwapInError( ss.str() );
            }

        // Read object's state. The object is automatically scheduled to global schedule.
        tr.next_line();
        object->text_read( tr, desc );

        /** @todo verify input_stream.fail() ? */
        }
    }


void ObjectManager::unpack_exception_clean_up
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    swap_in_exception_clean_up( object_list );
    }


void ObjectManager::swap_in_exception_clean_up
    (
    const ObjectShared::ObjectList & object_list
    )
    {
    ObjectShared::ObjectList::const_iterator it;

    // First of all unmanage the objects.
    // This will prevent gc to destroy any object from the list
    // due to reference counting (while we will destroy the objects
    // one by one).
    // Note: Ref-counting is no more supported, but the code
    // stays here for future compatibility.
    gc_unmanage_objects( object_list );

    // And now destroy the objects.
    for( it = object_list.begin(); it != object_list.end(); ++it )
        {
        ObjectProperty * object = ( *it );
        hope( object );
        hope( !object->is_stack_referenced() );
        object->delete_object( Object::DELETE_GROUP_CREATE_ERROR );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
