/** @file object_property.cpp
 * Managed object, private implementation.
 */

/*
 * begin        : Sun Dec 06 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

// FIXME: includes cleanup
#include "global/common.h"
#include "global/global.h"
#include "object/object_property.h"
#include "object/object_property_data.h"
#include "object/object.h"
#include "object/object_factory.h"
#include "object/class_manager.h"
#include "pointer/objectid.h"
#include "object/object_manager.h"
#include "pointer/object_pointer.h"
#include "pointer/pointer_to_object.h"
#include "property/property.h"
#include "object/event.h"
#include "object/local_schedule.h"
#include "pointer/garbage_collector.h"
#include "object/archive_manager.h"
#include "utility/heap_impl.h"
#include "replication/replication_flags.h"
#include "replication/replica_server.h"
#include "serializer/serializer.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "object/schedule.h"
#include "database/node_database.h"
#include "system/system.h"
#include "system/tick_lock.h"
#include "serializer/stype_int.h"
#include <memory>
#include <queue>
#include <sstream>


namespace Massiv {
namespace Core {


namespace {

enum GCFlags
    {
    GC_FLAGS_COLLECT_REQUESTED = 1,
        /**< Object was requested to be collected explicitly. */
    GC_FLAGS_DISPOSE_REQUESTED = 2,
        /**< Permanent GC root object was requested to be disposed. */
    GC_FLAGS_DYNAMIC_ROOT = 4,
        /**< Object is a dynamic GC root. */

    GC_FLAGS_BITS = 3
        /**< Number of bits required to serialize GC flags. */
    };

/** Get object GC flags that need be externalized. */
SInt<Int32, GC_FLAGS_BITS> get_gc_flags
    (
    const ObjectProperty * object
    )
    {
    int flags = 0;

    if( object->is_collect_pinned() )
        {
        flags |= GC_FLAGS_COLLECT_REQUESTED;
        }
    
    if( object->is_permanent_gc_root_object() && 
        object->is_gc_root_being_disposed() )
        {
        flags |= GC_FLAGS_DISPOSE_REQUESTED;
        }

    if( !object->is_permanent_gc_root_object() && 
        object->is_gc_root_object() )
        {
        flags |= GC_FLAGS_DYNAMIC_ROOT;
        }

    return flags;
    }

/**
 * Setup object state according to GC flags. Some state changes are irreversible.
 * For example if an object was requested to be explicitly collected, the request
 * will not be cancelled by another call to @c apply_gc_flags().
 */
void apply_gc_flags
    (
    ObjectProperty *    object,
    int                 flags
    )
    {
    GarbageCollector & gc = Global::garbage_collector();

    if( flags_on( flags, GC_FLAGS_COLLECT_REQUESTED ) )
        {
        // Ask GC to delete the object at a future GC tick. However we do not
        // want it to be deleted immediatelly as it is still used by the core.
        // That's why the object is pinned during the call to
        // @c GC::collect_object().

        object->pin();
        gc.collect_object( object );
        object->unpin();

        hope( object->is_collect_pinned() );
        }
    
    if( flags_on( flags, GC_FLAGS_DISPOSE_REQUESTED ) )
        {
        gc.dispose_gc_root( object );
        }
    
    if( flags_on( flags, GC_FLAGS_DYNAMIC_ROOT ) )
        {
        gc.make_gc_root( object, true );
        }
    else
        {
        gc.make_gc_root( object, false );
        }
    }

}
    

// class ObjectProperty

#ifdef GC_DELAYED_UNPIN

ObjectProperty::ActivationEra ObjectProperty::current_era( 1 );

#define INITIALIZER() \
    activation_era()

namespace {

    const int   MAX_ERA_TO_TIME_ITEMS = 1024;
        /**< Size of @c era_to_time buffer. */
    SystemTime  era_to_time[ MAX_ERA_TO_TIME_ITEMS ];
        /**< Cyclic buffer. Time of particular @c era is stored at
         @c era % MAX_ERA_TO_TIME_ITEMS position. */
}

#else

#define INITIALIZER() \
    activate_count( 0 )

#endif


//
// Constructors and destructors.
//

ObjectProperty::ObjectProperty()
      : metaobject( NULL ),
        root_data( NULL ),
        INITIALIZER()
    {
    }


ObjectProperty::ObjectProperty
    (
    const StackRootTag & tag
    ) : Property( tag ),
        metaobject( NULL ),
        root_data( NULL ),
        INITIALIZER()
    {
    // Initialize the object as a root object with a reserved object_id
    // The object is special and is not registered into GC / simulation.

    root_data = new RootData();
    }


ObjectProperty::ObjectProperty
    (
    const ObjectProperty & object
    ) : Property( object ),
        metaobject( NULL ),
        root_data( NULL ),
        INITIALIZER()
    {
    object.read_requested();
    }

#undef INITILIAZER


ObjectProperty::~ObjectProperty()
    {
    if( is_root_property() && !root_data->gc_shutdown )
        {
        hope( !is_archive_on_write() );
        
        if( is_root_property() && this != Global::get_stack_root_object() )
            {
            ensure_active_unpinned();
            hope( !is_pinned() );
            }
        }
    
    /* Destroy remaining stuff. */

    delete root_data;
    }


ObjectProperty & ObjectProperty::operator=
    (
    const ObjectProperty & object
    )
    {
    return *this;
    }


//
// Object state.
//

bool ObjectProperty::is_pinned() const
    {
    ObjectProperty * root = get_root_owner_object();

    /* Object is pinned if it is explicitly pinned. */

    if( !is_member_object() && root->root_data->gc_pin_count )
        {
        return true;
        }
    
    /* Object is pinned if it is created on the stack. */

    if( is_on_stack() )
        {
        return true;
        }

    /* Object is pinned if it is waiting for explict collect. */

    if( is_collect_pinned() )
        {
        return true;
        }

    /* Object is pinned if it might be active. */

    if( is_active_pinned() )
        {
        return true;
        }

    /* Object is pinned if it is referenced by an @c UpcallToObject. */

    if( is_upcall_pinned() )
        {
        return true;
        }
    
    /* Object is pinned if it is referenced by a @c PtrToObject. */

    if( is_ptr_pinned() )
        {
        return true;
        }

    /* Object is not pinned. */

    return false;
    }


bool ObjectProperty::is_upcall_pinned() const
    {
    return Global::garbage_collector().is_object_upcalled( this );
    }


bool ObjectProperty::is_permanent_gc_root_object() const
    {
    if( !is_replica() )
        {
        return get_metaobject()->get_attribute_root();
        }
    else
        {
        return false;
        }
    }


bool ObjectProperty::has_replica_requests() const
    {
    hope( Global::object_manager().replica_server.get() );
    return Global::object_manager().replica_server->has_replica_requests( get_object_id() );
    }
    

bool ObjectProperty::is_stack_referenced() const
    {
    ObjectProxyManager & manager = Global::object_proxy_manager();

    return flags_on
        (
        manager.classify_object( this ),
        ObjectProxy::CLASSIFY_STACK_STRONG_REFERENCED
        );
    }


//
// Extensions to Property.
//

void ObjectProperty::initialize_object
    (
    InitReason              reason,
    const ObjectId *        new_object_id
    )
    {
    ClassManager & class_manager = Global::class_manager();
    ObjectManager & object_manager = Global::object_manager();
    
    // NOTE: The initialization code must be able to reinitialize an object
    //       previously initialized by @c initialize_from_owner().

    // Access metaobject.

    hope( metaobject == NULL || metaobject->get_class_type_info() == typeid( *this ) );
    metaobject = class_manager.get_metaobject( typeid( *this ) );

    // Create root data object.

    hope( !root_data );
    root_data = new RootData();

    // Initialize basic property flags.

    owner = this;
    root_owner = this;
    replica = flags_on( reason, REPLICA );
    available = true; // object itself is always available
    archive_on_write = false;
    replication_flags = REPFLAGS_PROPERTY_DEFAULT;
    set_dirty_time();

    hope( !is_member_object() );
    
    // Setup property hierarchy.

    metaobject->initialize_property_hierarchy( this );
    
    // Generate object id unless specified.

    if( flags_on( reason, GENERATE_ID ) )
        {
        if( metaobject->get_attribute_node_object() )
            {
            // Create proxy node object id.

            NodeDatabase & node_database = Global::node_database();
            NodeId local_node_id = node_database.get_local_node_id();
            
            hope( metaobject->get_attribute_no_archive() );
            
            root_data->object_id = 
                ObjectId( local_node_id, metaobject->get_class_type_id() );
            }
        else
            {
            // Create general object id.

            const bool tracked = metaobject->get_attribute_tracked();
            object_manager.generate_id
                (
                root_data->object_id,
                tracked,
                metaobject->get_class_type_id(),
                metaobject->get_attribute_no_archive()
                );
            }
        }
    else
        {
        hope( flags_on( reason, LOAD_ID ) );
        hope( new_object_id );
        hope( new_object_id->is_object_info_available() );
        hope( new_object_id->get_class_type_id().is_class_registered() );
        hope( new_object_id->get_class_type_id() == metaobject->get_class_type_id() );
        root_data->object_id = *new_object_id;
        }
    
    // Register object.

    if( flags_on( reason, LOCAL_REGISTER ) )
        {
        object_manager.gc.register_object( this );
        if( flags_on( reason, REPLICA ) )
            {
            object_manager.gc.unmanage_object( this );
            }
        }
    }


namespace {

/**
 * Walks through a property hierarchy and tests if property owners
 * are set up correctly.
 */
struct TestHierarchy : public Property::PropertyOperation
    {

    /** Constructor. */
    TestHierarchy
        (
        Property * root_owner,
        Property * owner
        ) : root_owner( root_owner ),
            owner( owner )
        {
        }

    /** Check @c property. */
    virtual Result operator()
        (
        Property * property
        )
        {
        /* Check property. */

        hope( property->get_root_owner() == root_owner );
        hope( property->get_owner() == owner );

        /* Check owned properties recursively. */

        TestHierarchy operation( root_owner, property );
        property->for_each_owned_property_do( operation );

        /* Done. */

        return RESULT_DONE;
        }

    Property *  root_owner;
        /**< Root owner. */
    Property *  owner;
        /**< Owner. */

    };

} // anonymous namespace
    

void ObjectProperty::initialize_from_owner
    (
    Property &          direct_owner,
    const Attributes &  attributes
    )
    {
    // NOTE: The initialization code must be able to reinitialize an object
    //       partially initialized by a previous call to
    //       @c initialize_from_owner().

    if( &direct_owner != get_owner() || metaobject == NULL )
        {
        // Initialize metaobject.

        ClassManager & class_manager = Global::class_manager();
        metaobject = class_manager.get_metaobject( typeid( *this ) );

        // Initialize self.

        Property::initialize_from_owner( direct_owner, attributes );
        hope( is_member_object() );

        // Initialize owned properties.

        metaobject->initialize_property_hierarchy( this );
        }
    else
        {
        ClassManager &      class_manager = Global::class_manager();
        const MetaObject *  old_metaobject = metaobject;
        
        // Optimized version.
        // We are initializing the same object for the second time.
        // This can happen if @c initialize_from_owner() is called
        // from a constructor of an object for example.
        // Consider value type hierarchy:
        // ValueType <- A <- B.
        // Both A::A() and B::B() call ValueType::initialize_object() which in
        // turn calls @c initialize_from_owner(). At the time of first call
        // typeid( *this ) refers to A (since B has not been constructed yet).
        
        if( typeid( *this ) != metaobject->get_class_type_info() )
            {
            // Reinitialize metaobject.

            metaobject = class_manager.get_metaobject( typeid( *this ) );
            
            hope( metaobject->inherits( old_metaobject->get_class_type_info() ) );
            hope( &direct_owner == get_owner() );

            // Initialize new properties.
            // Inherited properties have already been initialized.

            metaobject->initialize_property_hierarchy( this, false );
            }

        // Self-test correctness.

        #ifdef OPTION_DEBUG

        TestHierarchy operation( direct_owner.get_root_owner(), &direct_owner );

        #endif
        }
    }


Property::PropertyType ObjectProperty::get_type() const
    {
    return Property::TYPE_OBJECT_PROPERTY;
    }


Property::PropertyInfo ObjectProperty::get_info
    (
    int query_info
    ) const
    {
    return Property::get_info( query_info );
    }


bool ObjectProperty::get_owned_info
    (
    const Property *            owned,
    Property::PropertyInfo &    info
    ) const
    {
    if( metaobject )
        {
        return metaobject->get_property_info( this, owned, info );
        }
    else
        {
        return false;
        }
    }


std::string ObjectProperty::generated_name() const
    {
    std::ostringstream  s;

    if( is_member_object() )
        {
        hope( metaobject );

        s   << "member_object$" << this 
            << "{" << metaobject->get_class_name() << "}";
        }
    else
        {
        s   << get_object_id();
        }

    return s.str();
    }


int ObjectProperty::memory_size() const
    {
    int size = Property::memory_size();
    
    size += sizeof( ObjectProperty ) - sizeof( Property );

    if( !is_member_object() )
        {
        size += sizeof( RootData );
        }

    return size;
    }


void ObjectProperty::for_each_owned_property_do
    (
    PropertyOperation & operation
    )
    {
    hope( metaobject );
    metaobject->for_each_property_do( operation, this );
    }


//
// Object state archiving.
//

void ObjectProperty::enter_archive_mode()
    {
    hope( !is_archive_on_write() );
    hope( !is_replica() );
    hope( is_archivable() );

    set_archive_on_write( true );
    }


void ObjectProperty::archive_now()
    {
    ArchiveManager &  archive_manager = Global::archive_manager();

    if( !is_archive_on_write() )
        return;

    hope( !is_replica() );
    hope( !is_member_object() );
    hope( is_archivable() );

    /* Archive object and leave archive mode. */

    archive_manager.archive_object_state( this );
    set_archive_on_write( false );
    }


void ObjectProperty::archive_migration_group_state()
    {
    ArchiveManager &  archive_manager = Global::archive_manager();

    if( !is_archive_on_write() )
        {
        return;
        }

    hope( !is_replica() );
    hope( !is_member_object() );

    /* Register the object and its migration group to current archive. */

    archive_manager.register_object_to_archive( this );
    }


// Getters for friends.

VariantPointer ObjectProperty::get_complete_object() const
    {
    hope( metaobject );
    return metaobject->get_complete_object( this );
    }


const ObjectFactory * ObjectProperty::get_factory() const
    {
    hope( metaobject );
    return metaobject->get_object_factory();
    }


//
// Private methods.
//

void ObjectProperty::delete_object
    (
    DeleteReason reason
    )
    {
    ObjectManager & object_manager = Global::object_manager();

    /* This must be root object. */
    hope( !is_member_object() );

    /* Object can not be pinned unless shutting down. */
    if( !flags_on( reason, SHUTTING_DOWN ) )
        {
        ensure_active_unpinned();
        hope( !is_pinned() );
        }

    /* "Being finalized" object may be deleted by GC only. */
    if( !flags_on( reason, GC_INITIATED ) )
        {
        hope( !is_finalizing() );
        }

    /* Store object's state if it was requested. */
    if( is_archive_on_write() )
        {
        if( !flags_on( reason, SHUTTING_DOWN ) )
            {
            archive_now();
            }
        }

    /* Notify object about finalization. */
    if( is_finalizing() )
        {
        hope( !is_replica() );
        hope( !flags_on( reason, SHUTTING_DOWN ) );
        
        LOCK_SYSTEM();

        {
        UpcallToObject upcall( this );
        upcall.tag_upcall
            ( "::Massiv::Core::Object::object_updated( INSTANCE_FINALIZE )" );
        upcall->object_updated( Object::INSTANCE_FINALIZE );
        }

        root_data->gc_finalizing = false;
        }

    /* Unregister object. */
    if( flags_on( reason, LOCAL_UNREGISTER ) )
        {
        // Remove the object from global schedule first
        // because remove_object_from_global_schedule()
        // requires the object to be still registered.

        object_manager.remove_object_from_global_schedule( this );
        object_manager.gc.unregister_object( this );
        
        // Update local object name database.

        object_manager.gc.undefine_local_name( get_object_id() );
        }

    /* Inform system that object was permanently deleted from simulation.
     * This notification must be called after object is unregistered due to
     * how the system caches are updated. */
    if( flags_on( reason, GLOBAL_DELETE ) )
        {
        if( !is_replica() )
            {
            object_manager.global_notify_of_object_destruction( get_object_id() );
            }
        }

    /* Clear flags that could prevent us from deleting the object if
       shutting down. */
    if( flags_on( reason, SHUTTING_DOWN ) )
        {
        if( is_pinned() )
            {
            std::ostringstream ss;

            ss << "Deleting pinned object " << get_object_id() << ".";

            Global::log_warning
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_HIGHEST,
                ss.str()
                );
            }

        root_data->gc_shutdown = true;
        }

    /* Bye bye little one. */
    
    delete this;
    }


void ObjectProperty::add_new_event
    (
    const Event &   e,
    EventHandle *   handle_out
    )
    {
    ObjectManager & object_manager = Global::object_manager();

    hope( !is_member_object() );
    hope( !is_replica() );

    /** @todo Can be replaced by archive_now()? */
    write_requested();

    if( handle_out )
        {
        EventHandle h;
        EventId eid = root_data->events.add_new_public_event( e );
        h.set( get_object_id(), eid );
        *handle_out = h;
        }
    else
        {
        root_data->events.add_new_private_event( e );
        }

    object_manager.notify_of_local_schedule_change( this );
    }


void ObjectProperty::migrate_to
    (
    const ObjectId &    destination,
    const STime &       time,
    bool                notify,
    EventHandle *       event_handle_out
    )
    {
    Balancer & balancer = Global::balancer();

    if( is_member_object() )
        {
        std::ostringstream ss;

        ss << "Trying to schedule member object as an event for object (id: "
           << destination << ").";

        RAISE_CORE_LIB_EXCEPTION( InvalidEventException, ss.str() );
        }

    if( is_replica() )
        {
        std::ostringstream ss;

        ss << "Trying to schedule object replica (id: "
           << get_object_id() << ") as an event for object (id: "
           << destination << ").";

        RAISE_CORE_LIB_EXCEPTION( InvalidEventException, ss.str() );
        }

    // Account for migration.
    balancer.account_for_object_migration( this, destination );

    // Note: write_requested() is not required here, because
    // it is called inside add_new_event().

    // Put event directly to our local schedule.
    Event e( time, destination, notify );
    add_new_event( e, event_handle_out );
    }


void ObjectProperty::remove_first_event()
    {
    ObjectManager & object_manager = Global::object_manager();

    hope( !is_member_object() );
    hope( !is_replica() );

    /** @todo Can be replaced by archive_now()? */
    write_requested();

    #ifdef OPTION_DEBUG
    root_data->schedule_block_time = SystemTime();
    #endif

    root_data->events.remove_first_event( true );
    object_manager.notify_of_local_schedule_change( this );
    }


bool ObjectProperty::remove_event
    (
    const EventId &   event_id,
    Event &           e
    )
    {
    ObjectManager &        object_manager = Global::object_manager();
    bool                   ret;

    hope( !is_member_object() );
    hope( !is_replica() );

    /** @todo Can be replaced by archive_now()? */
    write_requested();

    /** @todo Not exception safe. If exception is thrown by notify_of_local_schedule_change(),
     * the local schedule and 'e' should stay unchanged. But who minds... */
    ret = root_data->events.remove_event( event_id, e );
    if( ret )
        {
        object_manager.notify_of_local_schedule_change( this );
        }
    return ret;
    }


void ObjectProperty::delay_schedule_processing
    (
    const STime & time
    )
    {
    ObjectManager & object_manager = Global::object_manager();

    hope( !is_member_object() );
    hope( !is_replica() );

    root_data->events.delay_schedule_processing( time );
    object_manager.notify_of_local_schedule_change( this );
    }


bool ObjectProperty::is_schedule_delayed() const
    {
    hope( !is_member_object() );
    hope( !is_replica() );

    return root_data->events.is_delayed();
    }


void ObjectProperty::touch_object()
    {
    hope( !is_member_object() );
    
    root_data->touch_time = SystemTime::current_time();
    }


SystemTime ObjectProperty::get_touch_time() const
    {
    hope( !is_member_object() );
    
    #ifndef GC_DELAYED_UNPIN

    return root_data->touch_time;

    #else

    // Determine time when the object was touched explicitly.

    SystemTime explicit_touch_time = root_data->touch_time;
    
    // Try to determine time when the object was active-pinned
    // for the last time. Treat @c active_pin() as if it was touching
    // objects implicitly.

    if( current_era - activation_era < MAX_ERA_TO_TIME_ITEMS )
        {
        SystemTime implicit_touch_time = 
            era_to_time[ activation_era.get_value() % MAX_ERA_TO_TIME_ITEMS ];

        if( implicit_touch_time > explicit_touch_time )
            {
            return implicit_touch_time;
            }
        }

    return explicit_touch_time;

    #endif
    }


void ObjectProperty::advance_era()
    {
    #ifdef GC_DELAYED_UNPIN

    // Advance current era.

    current_era.advance();
    
    if( current_era == ActivationEra() )
        {
        // Correctly handle overflow.
        // ActivationEra::INVALID era is a reserved value that signals
        // that object is not active-pinned.
            
        current_era.advance();
        }
        
    // Associate era with SystemTime.

    era_to_time[ current_era.get_value() % MAX_ERA_TO_TIME_ITEMS ] = 
        SystemTime::current_time();

    #else

    panic( "Unsupported." );

    #endif
    }


//
// Serialization.
//

void ObjectProperty::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    ObjectManager & object_manager = Global::object_manager();

    // Get the flags.

    int flags = desc.object_flags;
    if( is_member_object() )
        {
        flags &= Serializer::Description::OBJECTF_PROPERTIES;
        }

    // Properties.

    if( flags_on( flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        hope( metaobject );
        metaobject->bitstr_write( bos, desc, this );
        }

    // Event queue.

    if( flags_on( flags, Serializer::Description::OBJECTF_EVENTS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );
        root_data->events.bitstr_write( bos, desc );
        }
    
    // Object name.

    if( flags_on( flags, Serializer::Description::OBJECTF_NAME ) )
        {
        hope( is_root_property() );
        root_data->object_name.bitstr_write( bos, desc );
        }

    // Object major version.

    if( flags_on( flags, Serializer::Description::OBJECTF_VERSION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );
        if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_WRITE_ZERO_VERSION ) )
            {
            const ObjectVersion zero_version( 0 );
            zero_version.bitstr_write( bos, desc );
            }
        else
            {
            root_data->major_version.bitstr_write( bos, desc );
            }
        }

    // Replication info.

    if( flags_on( flags, Serializer::Description::OBJECTF_REPLICATION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        hope( object_manager.replica_server.get() );
        object_manager.replica_server->handle_outgoing_migration( bos, desc, get_object_id() );
        }
    
    // GC flags.

    if( flags_on( flags, Serializer::Description::OBJECTF_GC_FLAGS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        get_gc_flags( this ).bitstr_write( bos, desc );
        }
    }


void ObjectProperty::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    ObjectManager & object_manager = Global::object_manager();

    // Get the flags.

    int flags = desc.object_flags;
    if( is_member_object() )
        {
        flags &= Serializer::Description::OBJECTF_PROPERTIES;
        }

    // Properties.

    if( flags_on( flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        hope( metaobject );
        metaobject->bitstr_read( bis, desc, this );
        }

    // Event queue.

    if( flags_on( flags, Serializer::Description::OBJECTF_EVENTS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );
        hope( root_data->events.is_empty() ); // The object should be "just created"
        hope( !is_archive_on_write() );       // and inited do default state.

        // Read the event schedule.

        root_data->events.bitstr_read( bis, desc );

        // Update global schedule.

        object_manager.notify_of_local_schedule_change( this );
        }
    
    // Object name.

    if( flags_on( flags, Serializer::Description::OBJECTF_NAME ) )
        {
        hope( is_root_property() );
        root_data->object_name.bitstr_read( bis, desc );
        
        // Update local object name database.

        object_manager.gc.define_local_name( get_object_id(), root_data->object_name );
        }

    // Object major version.

    if( flags_on( flags, Serializer::Description::OBJECTF_VERSION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        root_data->major_version.bitstr_read( bis, desc );
        if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_INCREASE_VERSION ) )
            {
            root_data->major_version++;
            }
        }
    
    // Replication info.

    if( flags_on( flags, Serializer::Description::OBJECTF_REPLICATION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        hope( object_manager.replica_server.get() );
        object_manager.replica_server->handle_incoming_migration( bis, desc, get_object_id() );
        }
    
    // GC flags.

    if( flags_on( flags, Serializer::Description::OBJECTF_GC_FLAGS ) )
        {
        SInt<Int32, GC_FLAGS_BITS> flags;

        hope( is_root_property() );
        hope( !is_replica() );

        flags.bitstr_read( bis, desc );
        apply_gc_flags( this, flags );
        }
    }


void ObjectProperty::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    // Get the flags.
    // OBJECTF_REPLICAS can be used only during migration (bitstream serialization).

    int flags = desc.object_flags;
    hope( !flags_on( flags, Serializer::Description::OBJECTF_REPLICATION ) );
    if( is_member_object() )
        {
        // Non-property are ignored by member objects.

        flags &= Serializer::Description::OBJECTF_PROPERTIES;
        }

    // Properties.

    bool use_new_line = false;
    if( flags_on( flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        hope( metaobject );

        metaobject->text_write( tw, desc, this );
        use_new_line = true;
        }

    // Event queue.

    if( flags_on( flags, Serializer::Description::OBJECTF_EVENTS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        if( use_new_line )
            {
            tw.end_line();
            }
        tw.line() << "events ";
        root_data->events.text_write( tw, desc );
        use_new_line = true;
        }
    
    // Object name.

    if( flags_on( flags, Serializer::Description::OBJECTF_NAME ) )
        {
        hope( is_root_property() );
        
        if( use_new_line )
            {
            tw.end_line();
            }
        
        tw.line() << "name ";
        root_data->object_name.text_write( tw, desc );
        use_new_line = true;
        }

    // Object major version.

    if( flags_on( flags, Serializer::Description::OBJECTF_VERSION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        if( use_new_line )
            {
            tw.end_line();
            }
        tw.line() << "version ";
        if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_WRITE_ZERO_VERSION ) )
            {
            const ObjectVersion zero_version( 0 );
            zero_version.text_write( tw, desc );
            }
        else
            {
            root_data->major_version.text_write( tw, desc );
            }
        use_new_line = true;
        }

    // GC flags.

    if( flags_on( flags, Serializer::Description::OBJECTF_GC_FLAGS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        if( use_new_line )
            {
            tw.end_line();
            }
        tw.line() << "gc_flags ";
        get_gc_flags( this ).text_write( tw, desc );
        use_new_line = true;
        }
    }


void ObjectProperty::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    ObjectManager & object_manager = Global::object_manager();

    // Get the flags.
    // OBJECTF_REPLICAS can be used only during migration (always bitstream serialization).

    int flags = desc.object_flags;
    hope( !flags_on( flags, Serializer::Description::OBJECTF_REPLICATION ) );
    if( is_member_object() )
        {
        // Non-property are ignored by member objects.

        flags &= Serializer::Description::OBJECTF_PROPERTIES;
        }

    // Properties.

    bool read_next_line = false;
    if( flags_on( flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        hope( metaobject );
        metaobject->text_read( tr, desc, this );
        read_next_line = true;
        }

    // Event queue.

    if( flags_on( flags, Serializer::Description::OBJECTF_EVENTS ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );
        hope( root_data->events.is_empty() ); // The object should be "just created"
        hope( !is_archive_on_write() );       // and inited to a default state.

        ObjectManager & object_manager = Global::object_manager();

        // Read the event schedule.

        if( read_next_line )
            {
            tr.next_line_throw();
            }
        tr.require_string( "events" );
        root_data->events.text_read( tr, desc );
        read_next_line = true;

        // Update global schedule.

        object_manager.notify_of_local_schedule_change( this );
        }
    
    // Object name.

    if( flags_on( flags, Serializer::Description::OBJECTF_NAME ) )
        {
        hope( is_root_property() );
        
        if( read_next_line )
            {
            tr.next_line_throw();
            }
        tr.require_string( "name" );
        root_data->object_name.text_read( tr, desc );
        read_next_line = true;
        
        // Update local object name database.

        object_manager.gc.define_local_name( get_object_id(), root_data->object_name );
        }

    // Object major version.

    if( flags_on( flags, Serializer::Description::OBJECTF_VERSION ) )
        {
        hope( is_root_property() );
        hope( !is_replica() );

        if( read_next_line )
            {
            tr.next_line_throw();
            }
        tr.require_string( "version" );
        root_data->major_version.text_read( tr, desc );
        read_next_line = true;
        }

    // GC flags.

    if( flags_on( flags, Serializer::Description::OBJECTF_GC_FLAGS ) )
        {
        SInt<Int32, GC_FLAGS_BITS> flags;

        hope( is_root_property() );
        hope( !is_replica() );

        if( read_next_line )
            {
            tr.next_line_throw();
            }
        tr.require_string( "gc_flags" );
        flags.text_read( tr, desc );

        // Setup object state according to GC flags.

        apply_gc_flags( this, flags );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "object_property_inline.h"
#endif
