/** @file object_property.h
 * Object property, base class for Object.
 */

/*
 * begin        : Sun Dec 06 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_PROPERTY_H
#define MASSIV_OBJECT_PROPERTY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_OBJECT_INFO_H
#include "object/object_info.h"
#endif

#ifndef MASSIV_EVENT_HANDLE_H
#include "object/event_handle.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif

#ifndef MASSIV_METAOBJECT_H
#include "object/metaobject.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif


namespace Massiv {
namespace Core {


/** @class ObjectProperty
 * Base class for Object, should be considered private to the core.
 *
 * ObjectProperty is responsible for implementation of @c Property interface
 * of any managed object (derived from @c Object) and common Object
 * services that are internal to the core and should be "hidden" from user.
 * It estabilishes linkage to object's metadata which is utilized for
 * object introspection, serialization implementation, etc.
 * Anything that needs be implemented to make an Object a Property is done
 * here.
 *
 * ObjectProperty instances can not be used on its own. They are always
 * instances of Object. While ObjectProperty defines and implements
 * internal Property interface and basic services for Object, Object defines
 * interface that is "visible" to users and that can be safely utilized by
 * them.
 *
 * The discrimination between ObjectProperty and Object is also due to
 * another reason. While ObjectProperty is internal to the core,
 * ObjectProperty can be referenced from the core by native references/pointers,
 * whereas Object must be pinned when its code is active and that's why
 * special references like ObjectPointer, PtrToObject and UpcallToObject are
 * defined (these ensure proper pinning, all with different semantics and
 * purpose).
 *
 * @todo
 *      Friend cleanup. Delegates?
 *
 * @ingroup property
 * @ingroup object
 */
class MASSIV_EXPORTED ObjectProperty : public Property
    {
    /**
     * Lot of friends. Once we have enough time Delegate objects should be implemented
     * that will allow access to subset of methods for various classes.
     */
    //@{
    friend class    Object;
    friend class    Property;
    friend class    MetaObject;
    friend class    ObjectProxy;
    friend class    ObjectFactory;
    friend class    ObjectManager;
    friend class    GarbageCollector;
    friend class    ArchiveManager;
    friend class    ReplicaClient; // delete_object, major_version
    friend class    ReplicaServer; // get_major_version
    friend class    CommonArchiveOperations;
    friend class    Balancer;
    friend class    PtrToObject;
    friend class    ActiveObjectProxy;
    friend struct   GlobalScheduleIndexChanged;
    friend struct   EnumerateReplicationGroup;
    friend struct   GatherMigrationGroupInfo;
    //@}

private:

    /**
     * Internal parameters describing what to do when
     * initializing objects.
     */
    enum InitParameters
        {
        GENERATE_ID             =   1,
            /**< Forces system to assign new object id to the initialized object. */
        LOAD_ID                 =   2,
            /**< Assigns known object id to the object. */
        REPLICA                 =   4,
            /**< Marks object properties as replicated. */
        READ_ONLY               =   8,
            /**< Marks object properties as read only. */
        LOCAL_REGISTER          =   16,
            /**< Registers object to local system. */
        }; // InitParameters

    /**
     * Internal parameters describing what to do when
     * deleting objects.
     */
    enum DeleteParameters
        {
        LOCAL_UNREGISTER        =   1,
            /**< Unregisters object from the local system. */
        GLOBAL_DELETE           =   2,
            /**< Informs remote systems about object destruction. */
        GC_INITIATED            =   4,
            /**< Delete was initiated by garbage collector. */
        SHUTTING_DOWN           =   8
            /**< GC is shutting down. */
        }; // DeleteParameters

    /**
     * What is the reason for initializing local object?
     */
    enum InitReason
        {
        INIT_NEW_INSTANCE       =   GENERATE_ID | LOCAL_REGISTER,
            /**< Object was created as a new instance. */
        INIT_REPLICA            =   LOAD_ID | LOCAL_REGISTER | REPLICA,
            /**< Object is replica. */
        INIT_MIGRATED           =   LOAD_ID | LOCAL_REGISTER,
            /**< Object migrated to the local system. */
        }; // InitReason

 public:
    /**
     * What is the reason for deleting local object?
     */
    enum DeleteReason
        {
        DELETE_INSTANCE         =   LOCAL_UNREGISTER | GLOBAL_DELETE,
            /**< Object is deleted from the whole system. */
        GC_DELETE_INSTANCE      =   DELETE_INSTANCE | GC_INITIATED,
            /**< Garbage collector initiated deletion. */
        DELETE_MIGRATED         =   LOCAL_UNREGISTER,
            /**< Object migrated from the local system.
                 Includes object swapping-out. */
        DELETE_REPLICA          =   LOCAL_UNREGISTER,
            /**< Delete object replica. */
        DELETE_GROUP_CREATE_ERROR = LOCAL_UNREGISTER,
            /**< Object is deleted as a result of group creation failure.
                 This includes swap-in failure and migration group unpacking
                 failure. */
        DELETE_SHUTTING_DOWN    =   GC_INITIATED | SHUTTING_DOWN
            /**< GC is shutting down. */
        }; // DeleteReason

public:

    /**
     * Object hash function.
     */
    struct Hash
        {
        INLINE size_t operator()
            (
            const ObjectProperty * object
            ) const;
        }; // struct Hash

    struct RootData;

protected:

    //
    // Constructors & destructors.
    //

    /**
     * Constructor.
     * Leaves object uninitialized.
     * @c initialize_object() or @c initialize_from_owner() is expected to
     * be called then.
     */
    ObjectProperty();

    /**
     * Internal constructor used by the core to create StackRootObject.
     * Once constructor returns, the object is fully initialized.
     */
    ObjectProperty
        (
        const StackRootTag & tag
        );

    /**
     * Copy constructor.
     * Leaves object uninitialized.
     * @c initialize_object() or @c initialize_from_owner() is expected to
     * be called then.
     */
    ObjectProperty
        (
        const ObjectProperty & object
        );

    /**
     * Destructor.
     * Must not be overriden by ObjectProperty subclasses.
     * That is because of the fact that at the time of object destruction the
     * object has already been unregistered from the system and is unavailable.
     * Moreover destruction invocation does not neccessarily mean that the
     * object is globally deleted from the distributed system.
     * The object may have only been swapped out, for example.
     * If a clean up action associated with global object destruction
     * (triggered by GC) is needed, override @c finalize().
     * However one should not rely on GC finalization (ie. when it takes place
     * and if ever).
     */
    virtual ~ObjectProperty();
    
    /**
     * Copies object.
     * Does nothing since there is nothing to copy.
     * However the definition allows compiler to generate default
     * (member-wise) operator= for ObjectProperty subclasses.
     */
    ObjectProperty & operator=
        (
        const ObjectProperty & object
        );
    
public:

    //
    // Object state.
    //
    
    /**
     * Checks if object is pinned to the local system.
     * Pinned object can not be deleted, migrated or swapped out.
     * GC does not treat pinned objects as GC roots.
     *
     * Object is also pinned if object is active (see @c active_pin()),
     * upcalled (= referenced by @c UpcallToObject), referenced by @c PtrToObject
     * or if client requested GC to collect it and it has not been collected
     * yet.
     * In that case Object is said to be active-pinned, upcall-pinned,
     * ptr-pinned or collect-pinned.
     * Object is also automatically pinned if it is created on the stack.
     *
     * Core pins objects when it needs to ensure that they won't be
     * deleted, migrated or swapped out.
     */
    bool is_pinned() const;

    /**
     * Pins object.
     * If called on a member object, the call is delegated to root owner.
     */
    INLINE void pin();

    /**
     * Unpins object.
     * If called on a member object, the call is delegated to root owner.
     */
    INLINE void unpin();
    
private:

    /**
     * Notifies object that it has been "activated".
     * That means that ObjectPointer with local semantics was dereferenced
     * (LocalProxy instance was created).
     * Pins the object.
     *
     * When object is deactivated @c active_unpin() must be called
     * (unless GC_DELAYED_UNPIN is defined).
     *
     * Works on root objects only. Used by ActiveObjectProxy.
     */
    INLINE void active_pin();

    /**
     * Notifies object that it has been "deactivated".
     * That means that the reference that caused object activation
     * was released (LocalProxy instances was deleted).
     * Unpins the object.
     *
     * If GC_DELAYED_UNPIN is defined all active-pinned objects are
     * automatically unpinned when it is known that no object can be active (see
     * @c GarbageCollector::upin_active_pinned_objects()), thus
     * calling @c active_unpin() at object "deactivation" is not required.
     * Objects are active-pinned conservatively then.
     *
     * Works on root objects only. Used by ActiveObjectProxy.
     */
    INLINE void active_unpin();
    
    /**
     * Asserts if object is not active-unpinned.
     * This method is used to express programmer's intention that object
     * should never be active-pinned (active) in this context.
     * Under normal cirmustances core keeps track of active objects but
     * this may be ineffecient. Because of this GC_DELAYED_UNPIN can be
     * enabled where objects are pinned conservatively and unpinned at
     * well-known places when it is known that no object can be active.
     * The consequence is that objects may be pinned for longer periods of
     * time. Note that pinned objects can not migrate and if someone waits
     * for the migration to take place (ie. SRPC) the code may even block
     * for ever (object will never be automatically unpinned). To avoid this,
     * call @c ensure_active_unpinned() which unpins the object if it is
     * conservatively pinned.
     * There are very few cases when this must be called and all the cases
     * are related to SRPC (ie. submitting a migration and waiting for the
     * migration).
     * The method is intended to be used by the core, client programmer
     * should not care.
     */
    INLINE void ensure_active_unpinned();
    
    /**
     * Notifies object that it is referenced by a PtrToObject instance.
     * Used by PtrToObject.
     */
    INLINE void ptr_pin();

    /**
     * Notifies object that the reference through PtrToObject has been
     * released. Used by PtrToObject.
     */
    INLINE void ptr_unpin();

public:

    /**
     * Checks if object is collect-pinned.
     */
    INLINE bool is_collect_pinned() const;

    /**
     * Checks if object is active-pinned.
     * If GC_DELAYED_UNPIN is defined, this is a conservative estimation
     * (ie. object can be conservatively active-pinned even if object is no
     * longer active).
     */
    INLINE bool is_active_pinned() const;

    /**
     * Checks if object is pinned by an UpcallToObject instance.
     */
    bool is_upcall_pinned() const;

    /**
     * Checks if object is pinned by a PtrToObject instance.
     */
    INLINE bool is_ptr_pinned() const;

    /**
     * Tests if the object is stack-strong referenced.
     */
    bool is_stack_referenced() const;

    /**
     * Is this Object instance GC root object?
     * Only registered objects can be GC roots.
     *
     * @note @c is_gc_root_object() can return false for permanent GC root
     *       objects if they are being disposed.
     */
    INLINE bool is_gc_root_object() const;

    /**
     * Is this Object instance permanent GC root object?
     * This is controlled on by-class basis via IDL "root" attribute.
     */
    bool is_permanent_gc_root_object() const;
    
    /**
     * Is this Object instance a permanent GC root object that is being
     * disposed?
     */
    INLINE bool is_gc_root_being_disposed() const;

    /**
     * Is this Object instance root object or member object?
     */
    INLINE bool is_member_object() const;

    /**
     * Is this Object registered into GC?
     */
    INLINE bool is_registered() const;
    
    /**
     * Is this Object being finalized?
     */
    INLINE bool is_finalizing() const;

    /**
     * Is this Object managed by the GC?
     */
    INLINE bool is_gc_managed() const;

    /**
     * Is this Object archivable?
     * @pre The object is not replica or member object.
     */
    INLINE bool is_archivable() const;

    /**
     * Is this object replicated to different nodes?
     */
    bool has_replica_requests() const;

    /**
     * Is this object in consistent state.
     *
     * @return @c false if object is inconsistent replica (see
     *          #RootData::consistent_replica for more info), @c true otherwise.
     */
    INLINE bool is_consistent() const;

    /**
     * Modify value of no_archive object flag.
     *
     * This must be done as soon as the object is created, #initialize() method
     * is the best place. This flag can be specified per-class in IDL.
     *
     * @todo Check that it's ok to change the flag? How?
     */
    INLINE void set_no_archive
        (
        bool no_archive
        );

    /**
     * Modify value of consistent_replica flag.
     *
     * Only ReplicaClient should change the flag.
     *
     * @pre Object is not memeber object (#root_data is valid).
     */
    INLINE void set_consistent_replica
        (
        bool consistent_replica
        );

public:

    //
    // Extensions to Property.
    // Private to the core.
    //

    /**
     * Initialize property flags and attributes from its owner.
     *
     * Never call this on root objects.
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        );

    /**
     * Return property type.
     */
    virtual PropertyType get_type() const;

    /**
     * Return information about the object.
     */
    virtual PropertyInfo get_info
        (
        int query_info = QUERY_FULL_NAME
        ) const;

    /**
     * Return information about owned property.
     */
    virtual bool get_owned_info
        (
        const Property *    owned,
        PropertyInfo &      info
        ) const;

    /**
     * Return object name.
     */
    virtual std::string generated_name() const;

    /**
     * Return how much memory object consumes.
     */
    virtual int memory_size() const;

    /**
     * Invoke @c operation on each owned property.
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        );
    
    /**
     * Serialize object to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read object from bitstream.
     *
     * Object will be automatically scheduled to global schedule if
     * @c OBJECTF_EVENTS object serialization flag is specified in @a desc.
     *
     * @pre If the object is not member object or replica then
     * local schedule is empty and the object is not in
     * @b archive_on_write state (the object should be "just created"
     * and inited to default state).
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize object to text stream.
     *
     * Object serialization uses the same rules as other text serializations.
     * It begins on current line and does not end with a new-line, but may (will)
     * contain new-line characters inside. It will look like this if all object
     * serialization flags are set:
     *
     * @verbatim
object
  ...
@endobject
events ...
name ...
version ...
replicas ...
gc_flags ...
@endverbatim
     */
    virtual void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read object from text stream.
     *
     * Object will be automatically scheduled to global schedule if
     * @c OBJECTF_EVENTS object serialization flag is specified in @a desc.
     *
     * @pre If the object is not member object or replica then
     * local schedule is empty and the object is not in
     * @b archive_on_write state (the object should be "just created"
     * and inited to default state).
     */
    virtual void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );
    
public:

    //
    // Getters for friends.
    //
    
    /**
     * Returns metaobject for this class.
     */
    INLINE const MetaObject * get_metaobject() const;

    /**
     * Returns object factory for this class.
     */
    const ObjectFactory * get_factory() const;

    /**
     * Returns pointer to complete object.
     * If this object property's dynamic type is X then pointer to X
     * is returned.
     */
    VariantPointer get_complete_object() const;
    
    /**
     * Get object major version.
     */
    INLINE ObjectVersion get_major_version() const;
    
    /**
     * Get #ObjectId of this object.
     * Asserts if object is a member object.
     * Optional part of #ObjectId is always valid.
     */
    INLINE const ObjectId & get_object_id() const;
    
    /**
     * Get #ObjectInfo of the object.
     */
    INLINE ObjectInfo get_object_info() const;
    
    /**
     * Get #ClassTypeId of the object.
     */
    INLINE ClassTypeId get_class_type_id() const;

    /**
     * Get object touch time.
     * If GC_DELAYED_UNPIN is defined, objects are implicitly touched
     * by @c active_pin().
     */
    SystemTime get_touch_time() const;

private:

    //
    // Private methods.
    //
    
    /**
     * Initialize "root" object (object with its own objectid, not member object).
     *
     * Called by ObjectFactory. It's up to the object to do all the necessary
     * steps to initialize all attributes of all owned properties, including
     * calls to #MetaObject::initialize_property_attributes, etc.
     */
    void initialize_object
        (
        InitReason              reason,
        const ObjectId *        new_object_id
        );

    /**
     * Deletes object.
     */
    void delete_object
        (
        DeleteReason reason = DELETE_INSTANCE
        );

    /**
     * Informs object that is is going to be archived.
     * Object "freezes" its current state so that it
     * could be later written to archive (archive_now()).
     * If object registers a write request and is in
     * archive mode it calls archive_now() automatically.
     */
    void enter_archive_mode();

    /**
     * Causes object that is already in archive mode to archive
     * its state immediatelly.
     * Leaves archive mode.
     * If the object is not in archive mode, method does nothing.
     */
    void archive_now();

    /**
     * Causes the migration group of the object to archive its state
     * immediatelly (state of a migration group contains only information
     * what objects belong to the group).
     * Nothing is done if state of the migration group
     * was already archived.
     * None of the objects from the migration group is archived.
     * The object doesn't leave archive mode.
     * If the object is not in archive mode, method does nothing.
     *
     * @note This method is intended for ObjectPointer management
     * purposes only to make archivation working during union of two
     * migration groups (also known as "archivation hack").
     * If two migration groups are going to be merged into one migration group,
     * archive_migration_group_state() must be called on both original groups
     * before the operation itself.
     *
     * @note archive_migration_group_state() is provided for performance
     * reasons only.
     * One can call archive_now() instead of archive_migration_group_state()
     * to fix up "archivation hack", but archive_now() will force
     * the object itself to be archived while archive_migration_group_state()
     * won't.
     */
    void archive_migration_group_state();

    /**
     * Add new event to local schedule.
     * If @a event_handle_out is not null then the event
     * is scheduled as public and @a event_handle_out is
     * filled with valid hanle.
     * If @a event_handle_out is null then the event
     * is scheduled as private and @a event_handle_out is
     * cleared (invalidated).
     * @note This function does not test if @a event
     * contains valid informations.
     */
    void add_new_event
        (
        const Event &       event,
        EventHandle *       event_handle_out
        );

    /**
     * Send this object to @a recipient_object.
     * @a event_handle_out specifies whether this object
     * should be scheduled as public or private event.
     * @param notify See Object's @c migrate_to() and @c schedule_to() for more info.
     */
    void migrate_to
        (
        const ObjectId &    recipient_object,
        const STime &       time,
        bool                notify,
        EventHandle *       event_handle_out
        );
    /**
     * Remove first scheduled event from object's local schedule.
     * Prior to calling this function there must be at least one schedule event!
     */
    void remove_first_event();

    /**
     * Remove scheduled event from object's local schedule.
     * Returns true if the event was removed from the schedule
     * and copied to @a event_out.
     * Returns false if the events doesn't exist (ot it is private event).
     * In that case @a event_out is left unchanged.
     */
    bool remove_event
        (
        const EventId & event_id,
        Event &         event_out
        );

    /**
     * Get first scheduled event from object's local schedule.
     * Returns true if the event was copied to @a event_out.
     * Returns false if there is no event in the local schedule.
     * In that case @a event_out is left unchanged.
     */
    INLINE bool get_first_event
        (
        Event & event_out
        ) const;

    /**
     * Check if the object has any scheduled event.
     */
    INLINE bool has_scheduled_event() const;

    /**
     * Get time of first scheduled event.
     * Prior to calling this function there must be at least one scheduled event!
     */
    INLINE STime get_time_of_first_event() const;

    /**
     * Delays schedule processing for this Object by @c time.
     * If the schedule is empty, nothing is done.
     */
    void delay_schedule_processing
        (
        const STime & time
        );

    /**
     * Tests if schedule processing for this Object is delayed.
     */
    bool is_schedule_delayed() const;

    /**
     * Touches object (updates object's touch time).
     * Call this to prevent object from being swapped out.
     */
    void touch_object();

    /**
     * Advances @c current_era and thus unpins all active-pinned
     * objects.
     * Asserts if GC_DELAYED_UNPIN is undefined.
     */
    static void advance_era();
    
private:

    const MetaObject *      metaobject;
        /**< Metaobject for this class. */
    RootData *              root_data;
        /**< Information about a root object.
             NULL if ObjectProperty is a member object of another object. */
    
    #ifndef GC_DELAYED_UNPIN
    
    UInt32                  activate_count;
        /**< How many times was object active-pinned? */
    
    #else
    
    typedef VirtualTime<>   ActivationEra;

    ActivationEra           activation_era;
        /**< What era was object activate-pinned in?
             Object is assumed to be active-pinned iff
             @c activation_era == @c current_era. */
    static ActivationEra    current_era;
        /**< Current activation era.
             Incremented to active-unpin all objects. */
    
    #endif

    }; // class ObjectProperty


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object_property_inline.h"
#endif


#endif // MASSIV_OBJECT_PROPERTY_H
