/** @file object_property_data.h
 * Internal classes used by ObjectProperty and the core.
 */

/*
 * begin        : Thu Jul 24 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_PROPERTY_DATA_H
#define MASSIV_OBJECT_PROPERTY_DATA_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_OBJECT_INFO_H
#include "object/object_info.h"
#endif

#ifndef MASSIV_LOCAL_SCHEDULE_H
#include "object/local_schedule.h"
#endif

#ifndef MASSIV_HEAP_INDEX_H
#include "utility/heap_index.h"
#endif

#ifndef MASSIV_STYPE_VLI_H
#include "serializer/stype_vli.h" // ObjectVersion
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h" // ObjectName
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif


namespace Massiv {
namespace Core {


/**
 * Information that describes a root ObjectProperty.
 *
 * @ingroup object
 * @ingroup property
 */
struct ObjectProperty::RootData
    {
public:
    
    /** Constructor. */
    RootData();

public:

    //
    // Basic attributes.
    //

    ObjectId                object_id;
        /**< #ObjectId of the object. Optional #ObjectInfo part of the #ObjectId
             is always valid. */
    ObjectName              object_name;
        /**< Optional object name. Mainly for debugging purposes.
             Managed by the Garbage Collector. */
    ObjectVersion           major_version;
        /**< Object "major version". It's increased every time the object migrates,
             therefore @c major_version.local_replication_timestamp never decreases.

             Major version is reset to zero when restarting server nodes from archive,
             let's just hope the 32bits used do not wrap around during server
             lifetime.

             More info can be found in description of replication and delta update
             protocols. */
public:

    //
    // Replication.
    //

    bool                    consistent_replica : 1;
        /**< Is replica in consistent state?
             Replica is in inconsistent state if it has pointer to object
             which is member of the same replication group but its state is
             from different update, or if it has pointer to another inconsistent
             replica.
             
             Unused unless object is replica. */

public:

    //
    // Garbage Collector attributes.
    //

    bool                    gc_registered : 1;
        /**< Is this object registered into GC and accessible
                through its @c object_id on the local system? */
    bool                    gc_finalizing : 1;
        /**< Is object being finalized? */
    bool                    gc_managed : 1;
        /**< Is GC allowed to delete object? */
    bool                    gc_root : 1;
        /**< Is this object GC root object? */
    bool                    gc_disposing_root : 1;
        /**< Is this object a permanent GC root object that should be disposed
             now? */
    bool                    gc_pointer_update_pending : 1;
        /**< Does object own pointer properties that recently changed
             and whose corresponding backward pointers need be updated? */
    bool                    gc_collect_pinned : 1;
        /**< Is this object collect-pinned? */
    bool                    gc_shutdown : 1;
        /**< Object is marked "shutdown" if GC is going to shutdown itself
             and will delete all objects. This flag disables additional checks
             that would prevent us from deleting the object (under general
             cirmustances).
             Filled by @c ObjectProperty::destroy(). */
    int                     gc_pin_count;
        /**< How many times was object pinned up by user? */
    int                     gc_ptr_pin_count;
        /**< How many times was object pinned up by PtrToObject? */
    GCEpoch                 gc_epoch;
        /**< Epoch when the object was reachable from GC root object
             for the last time. */
    VirtualTime<>           gc_enumeration_time;
        /**< Used by graph algorithm at @c GC::enumerate_replication_group().
             Time of last visit. */

public:

    //
    // Events.
    //

    LocalSchedule           events;
        /**< Local events scheduled for this object.
             for root objects only, NULL otherwise. */
    HeapIndex               global_schedule_index;
        /**< Schedule index to global schedule.
             For @c ObjectManager purposes only. */

    #ifdef OPTION_DEBUG
    
    SystemTime              schedule_block_time;
        /**< Time when ObjectManager attempted to process Object's first event
             for the first time but the object had been "busy". Object and the
             whole global schedule have been blocked at least since that time.
             @c SystemTime() implies no blocking. Automatically reset to
             @c SystemTime() by @c ObjectProperty::remove_first_event() which
             signalizes that the event no longer blocks the schedule. */

    #endif

public:

    //
    // Balancer attributes.
    //

    SystemTime              touch_time;
        /**< The last time the object (or another object from its migration
             group) was known to be "active".
             Since it is not updated automatically whenever Object is accessed
             (for example at pointer dereference time), the time need not be
             correct.
             Use @c ObjectProperty::touch_object() to set the time to the
             current system time.
             Treat it as a hint for memory optimizer only. */
    
    }; // struct ObjectProperty::RootData


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_OBJECT_PROPERTY_DATA_H
