/** @file object_inline.h
 * Object inline methods.
 */

/*
 * begin        : Sat Jan 18 2003
 * copyright    : (C) 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_HASH_FUNCTION_H
#include "utility/hash_function.h"
#endif

#ifndef MASSIV_OBJECT_PROPERTY_DATA_H
#include "object/object_property_data.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#if( !defined( GC_DELAYED_UNPIN ) && defined( OPTION_DEBUG ) )
#include "system/system.h"
#endif

  
namespace Massiv {
namespace Core {


// class ObjectProperty::Hash

size_t ObjectProperty::Hash::operator()
    (
    const ObjectProperty * object
    ) const
    {
    return ext::hash_int( reinterpret_cast<size_t>( object ) );
    }


// class ObjectProperty

// Object state.
    

void ObjectProperty::pin()
    {
    int & gc_pin_count = 
        checked_cast<ObjectProperty *>( get_root_owner() )->root_data->gc_pin_count;
    
    gc_pin_count++;
    }

    
void ObjectProperty::unpin()
    {
    int & gc_pin_count = 
        checked_cast<ObjectProperty *>( get_root_owner() )->root_data->gc_pin_count;

    hope( gc_pin_count > 0 );
    
    gc_pin_count--;
    }


void ObjectProperty::active_pin()
    {
    hope( !is_member_object() );

    #ifndef GC_DELAYED_UNPIN

    activate_count++;

        #ifdef OPTION_DEBUG
        
        if( is_replica() )
            {
            // Disable main loop recursion.
            System::lock_tick();
            }
        
        #endif

    #else

    activation_era = current_era;

    #endif
    }


void ObjectProperty::active_unpin()
    {
    hope( !is_member_object() );

    #ifndef GC_DELAYED_UNPIN

    activate_count--;

        #ifdef OPTION_DEBUG

        if( is_replica() )
            {
            // Enable main loop recursion.
            System::unlock_tick();
            }

        #endif

    #else

    activation_era.invalidate();
    hope( current_era > ActivationEra() );

    #endif
    }


void ObjectProperty::ensure_active_unpinned()
    {
    #ifdef GC_DELAYED_UNPIN

    // Programmer says that object should never be active-pinned here,
    // thus if it is conservatively active-pinned, active-unpin it.
    // Programmer must ensure that object is really not active
    // (this is checked if GC_DELAYED_UNPIN is undefined).

    if( is_active_pinned() )
        {
        active_unpin();
        }

    #endif

    #ifdef OPTION_DEBUG

    if( is_active_pinned() )
        {
        panic( ( std::string( "The core has detected that object " ) +
            get_info().property_name + " is active in a context where it "
            "is not allowed. This is either an application error or a core "
            "bug. See stack trace for more information." ).c_str() );
        }

    #endif
    }


void ObjectProperty::ptr_pin()
    {
    int & gc_ptr_pin_count = 
        checked_cast<ObjectProperty *>( get_root_owner() )->root_data->gc_ptr_pin_count;

    gc_ptr_pin_count++;
    }


void ObjectProperty::ptr_unpin()
    {
    int & gc_ptr_pin_count = 
        checked_cast<ObjectProperty *>( get_root_owner() )->root_data->gc_ptr_pin_count;

    hope( gc_ptr_pin_count > 0 );
    
    gc_ptr_pin_count--;
    }


bool ObjectProperty::is_collect_pinned() const
    {
    ObjectProperty * root = get_root_owner_object();
    return root->root_data->gc_collect_pinned;
    }


bool ObjectProperty::is_active_pinned() const
    {
    ObjectProperty * root = get_root_owner_object();
    
    /* Object is active-pinned if pinned by @c active_pin(). */

    #ifndef GC_DELAYED_UNPIN

    if( root->activate_count > 0 )
        {
        return true;
        }

    #else

    if( root->current_era == activation_era )
        {
        return true;
        }

    #endif

    return false;
    }


bool ObjectProperty::is_ptr_pinned() const
    {
    ObjectProperty * root = get_root_owner_object();
    
    return root->root_data->gc_ptr_pin_count > 0;
    }


bool ObjectProperty::is_gc_root_object() const
    {
    /* Only registered root objects can be GC roots. */

    hope( !is_member_object() );
    hope( !root_data->gc_root || ( root_data->gc_root && root_data->gc_registered ) );

    return root_data->gc_root;
    }


bool ObjectProperty::is_gc_root_being_disposed() const
    {
    hope( !is_member_object() );

    return root_data->gc_disposing_root;
    }


bool ObjectProperty::is_member_object() const
    {
    return !is_root_property();
    }


bool ObjectProperty::is_registered() const
    {
    hope( !is_member_object() );
    return root_data->gc_registered;
    }


bool ObjectProperty::is_finalizing() const
    {
    hope( !is_member_object() );
    return root_data->gc_finalizing;
    }


bool ObjectProperty::is_gc_managed() const
    {
    hope( !is_member_object() );
    return root_data->gc_managed;
    }
    
    
bool ObjectProperty::is_archivable() const
    {
    return get_object_info().is_archivable();
    }


bool ObjectProperty::is_consistent() const
    {
    ObjectProperty * const root = get_root_owner_object();
    if( root && root->is_replica() )
        {
        return root->root_data->consistent_replica;
        }
    return true;
    }


void ObjectProperty::set_no_archive
    (
    bool no_archive
    )
    {
    hope( !is_member_object() );
    root_data->object_id.set_no_archive( no_archive );
    }


void ObjectProperty::set_consistent_replica
    (
    bool consistent_replica
    )
    {
    hope( !is_member_object() );
    root_data->consistent_replica = consistent_replica;
    }


// Getters for friends.

const MetaObject * ObjectProperty::get_metaobject() const
    {
    return metaobject;
    }


ObjectVersion ObjectProperty::get_major_version() const
    {
    hope( !is_member_object() );
    return root_data->major_version;
    }


const ObjectId & ObjectProperty::get_object_id() const
    {
    hope( !is_member_object() );
    return root_data->object_id;
    }
    

ObjectInfo ObjectProperty::get_object_info() const
    {
    hope( !is_member_object() );
    return root_data->object_id.get_object_info();
    }
    
    
ClassTypeId ObjectProperty::get_class_type_id() const
    {
    hope( !is_member_object() );
    return root_data->object_id.get_class_type_id();
    }


// Private stuff. Strasnej bordel.

bool ObjectProperty::get_first_event
    (
    Event & e
    ) const
    {
    hope( !is_member_object() );
    hope( !is_replica() );

    // write_requested() is not required here.

    return root_data->events.get_first_event( e );
    }


bool ObjectProperty::has_scheduled_event() const
    {
    hope( !is_member_object() );
    return !root_data->events.is_empty();
    }


STime ObjectProperty::get_time_of_first_event() const
    {
    hope( !is_member_object() );
    hope( !root_data->events.is_empty() );
    return root_data->events.get_time_of_first_event();
    }

    
} // namespace Massiv::Core
} // namespace Massiv
