/** @file object_provider.cpp
 * Object provider generates new object ids from his pool and keeps track
 * of some remote objects.
 */

/*
 * begin        : Sun Dec 16 2001
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "object/object_provider.h"
#include "object/providerid.h"
#include "object/object_provider_state.h"
#include "node/nodeid.h"
#include "object/object_status.h"
#include "sysevent/object_status_msg.h"
#include "pointer/objectid.h"
#include "utility/cache.h"
#include "utility/cache_impl.h"
#include "utility/debug.h"
#include "global/global.h"
#include "time/system_time.h"
#include <sstream>


namespace Massiv {
namespace Core {


struct ObjectProvider::ObjectInfo
    {
    /** Constructor. */
    ObjectInfo() :
        status( ObjectStatus::UNKNOWN )
        {
        }

    /** Constructor. */
    ObjectInfo
        (
        const ObjectStatus &    _status,
        const NodeId &          _node_id
        ) : 
        status( _status ), 
        node_id( _node_id )
        {
        }

    /** Copy constructor. */
    ObjectInfo
        (
        const ObjectInfo &  info
        ) :
        status( info.status ),
        node_id( info.node_id )
        {
        }

    /** Operator=. */
    ObjectInfo & operator=
        (
        const ObjectInfo & info
        )
        {
        status  = info.status;
        node_id = info.node_id;
        return *this;
        }

    ObjectStatus        status;
        /**< Object status. */

    NodeId              node_id;
        /**< Node id associated with object status. */

    }; // ObjectProvider::ObjectInfo


struct ObjectProvider::My
    {
    NO_COPY_CONSTRUCTOR( My );
    NO_ASSIGNMENT_OPERATOR( My );
    
public:
    /** Constructor. */
    My  (
        const ObjectProviderState & _state
        ) :
        state( _state )
        {
        }

    /** Object status cache. */
    typedef Cache
            <
            ObjectId,
            SystemTime,
            ObjectInfo,
            ObjectId::Hash,
            ObjectId::EqualKey
            > ObjectCache;
            
    ObjectProviderState     state;
        /**< State of the object provider. */

    ObjectCache             cache;
        /**< Object status cache. */
    
    }; // ObjectProvider::My


// Implementation of ObjectProvider
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//@{
/** @todo FIXME Temporary hack.
 * Time how long information that object is not accessible stays in provider's cache.
 *
 * @bug This should be read from net settings.
 */
const Float64 OBJECT_NOT_ACCESSIBLE_TIMEOUT = 90.0;
const Float64 OBJECT_ALIVE_TIMEOUT          = 60.0;
//@}


ObjectProvider::ObjectProvider
    (
    const ObjectProviderState & s
    ) :
    my( new My( s ) )
    {
    }


ObjectProvider::~ObjectProvider()
    {
    }
    
    
void ObjectProvider::get_state
    (
    ObjectProviderState &     s
    ) const
    {
    s = my->state;
    }


bool ObjectProvider::generate_id
    (
    ObjectId &          object_id,
    bool                track,
    const ClassTypeId & class_type_id,
    bool                no_archive
    )
    {
    ObjectId    id;
    UInt64      relative_id;

    hope( sizeof( object_id.components ) == sizeof( object_id.id ) );

    relative_id = my->state.generate_provider_relative_object_id();
    if( relative_id == UINT64( 0 ) )
        {
        return false;
        }

    id.set_null();
    id.components.provider_id           = my->state.get_provider_id().get_id_value();
    id.components.provider_relative_id  = relative_id;
    id.components.tracked_by_provider   = track ? 1 : 0;
    id.components.class_kind            = class_type_id.get_kind();
    id.components.class_type_id         = class_type_id.get_class_id();
    id.components.no_archive            = no_archive;

    hope( id.components.provider_relative_id != INT64( 0 ) );

    object_id = id;
    return true;
    }


bool ObjectProvider::is_exhausted() const
    {
    return my->state.is_provider_relative_object_id_pool_exhausted();
    }


void ObjectProvider::notify_of_object_status
    (
    const ObjectStatusMessage & s
    )
    {
    ObjectInfo      o_info( s.object_status, s.node );
    SystemTime      current_time = Global::system_time();

    switch( s.object_status.get_value() )
        {
        case ObjectStatus::ALIVE:
            {

                /* Test collision between cached status and notified status.
                 * Collision is detected if cached status is NOT_ACCESSIBLE
                 * and notified status is ALIVE.
                 * That means that some time ago we have received status
                 * notifying us that the object was not accessible, but now
                 * somebody is trying to notify us that the object is alive.
                 *
                 * If we will sometime in the future support lazy simulation
                 * startup (one server start simulation while other servers
                 * are not running), the collision may happen if a server
                 * connects to running simulation. Other server may have
                 * cached objects located on the connected server as NOT_ACCESSIBLE.
                 * The collision must be handled as part of lazy simulation
                 * startup.
                 *
                 * But in current implementation, the collision may happen
                 * if ALIVE status is cached on some servers and the object
                 * is destroyed and some of the server uses the wrong ALIVE
                 * status to notify another server.
                 *
                 * In theory, the collision should never happen when the status
                 * message is marked trustworthy.
                 */
                {
                ObjectStatus    test_state;
                NodeId          test_node_id;

                get_object_status( s.object, test_state, test_node_id );
                if( test_state == ObjectStatus::NOT_ACCESSIBLE )
                    {
                    std::ostringstream ss;
                    ss << "Provider cache collision detected. "
                       << "Node (" << s.get_sender() << ") is trying to notify local system of ALIVE object status "
                       << "about object (" << s.object << ") located on node (" << s.node << "). "
                       << "Cached status of the object is NOT_ACCESSIBLE.";

                    Global::log
                        (
                        STATUS_WARNING( Status::FACILITY_OBJECT_LOCALIZATION, Status::PRIORITY_HIGHEST ),
                        ss.str()
                        );
                        
                    // Ignore the message.
                    return;
                    }
                }

            if( s.object.get_provider_id() == my->state.get_provider_id() && 
                s.object.is_tracked_by_provider() )
                {
                // Untrustworthy alive status is ignored for tracked objects
                // to ensure that searching protocol will work.
                if( s.trustworthy )
                    {
                    // Persistent cache (timeout is ignored).
                    my->cache.update( s.object, 0, o_info, true );
                    }
                }
            else
                {
                // Timeout cache.
                my->cache.update( s.object, current_time + OBJECT_ALIVE_TIMEOUT, o_info, false );
                }
            return;
            }

        case ObjectStatus::NOT_ACCESSIBLE:
            {
            // Timeout cache.
            my->cache.update( s.object, current_time + OBJECT_NOT_ACCESSIBLE_TIMEOUT, o_info, false );
            return;
            }

        case ObjectStatus::UNKNOWN:
            // Invalidate cache.
            my->cache.remove( s.object );
            return;

        default:
            panic( "Invalid object status value in notify_of_object_status" );
        };

    hope( false );
    }


void ObjectProvider::refresh_object_status
    (
    const ObjectId &    oid
    )
    {
    ObjectStatus                o_status;
    NodeId                      o_node;
    ObjectStatusMessage         s;

    get_object_status( oid, o_status, o_node );
    s.set( oid, o_status, o_node, true );
    notify_of_object_status( s ); 
    }


void ObjectProvider::get_object_status
    (
    const ObjectId &    id,
    ObjectStatus &      object_status,
    NodeId &            node_id
    )
    {
    ObjectInfo * cinfo = my->cache.find( id, Global::system_time() );

    if( cinfo )
        {
        object_status = cinfo->status;
        node_id       = cinfo->node_id;
        }
    else
        {
        object_status = ObjectStatus::UNKNOWN;
        node_id.clear();
        }
    }


void ObjectProvider::advance_cache()
    {
    my->cache.advance( Global::system_time() );
    }


void ObjectProvider::flush_cache()
    {
    my->cache.clear();
    }


} // namespace Massiv::Core
} // namespace Massiv
