/** @file object_provider.h
 * Object provider generates new object ids from his pool and keeps track
 * of some remote objects.
 */

/*
 * begin        : Sun Dec 16 2001
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECT_PROVIDER_H
#define MASSIV_OBJECT_PROVIDER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_PROVIDER_FWD_H
#include "object/object_provider_fwd.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#include <memory> // auto_ptr


namespace Massiv {
namespace Core {


/**
 * Object provider generates new object ids from his pool of available
 * object ids and keeps track of status of some objects using timeout cache.
 *
 * Event thought object provider can cache status of any object, object
 * manager directs the provider to cache only status of objects witch have
 * object id generated by the provider.
 *
 * @see ObjectProviderState
 *
 * @ingroup object_management
 * @ingroup objectid
 */
class ObjectProvider
    {
    NO_DEFAULT_CONSTRUCTOR( ObjectProvider );
    NO_COPY_CONSTRUCTOR( ObjectProvider );
    NO_ASSIGNMENT_OPERATOR( ObjectProvider );

public:
    /**
     * Constructor.
     */
    ObjectProvider
        (
        const ObjectProviderState & provider_state
        );
        
    /**
     * Destructor.
     */
    ~ObjectProvider();
        
    /**
     * Get current state of the provider.
     */
    void get_state
        (
        ObjectProviderState & output_provider_state
        ) const;

    /**
     * Generate new object id.
     * @return False if object id pool of the provider is exhausted. In that case
     * @a object_id is not filled with valid object id and every other call of generate_id()
     * will fail.
     */
    bool generate_id
        (
        ObjectId &          object_id,
        bool                track_by_provider,
        const ClassTypeId & class_type_id,
        bool                no_archive
        );

    /**
     * Check if id pool of the provider is exhausted.
     * @return True if the provider can't generate object id any more (next
     * call to generate_id() will fail), otherwise false.
     */
    bool is_exhausted() const;

    /**
     * Notify the provider of object status.
     * See @a ObjectStatusMessage about arguments' meaning.
     * Passing @c ObjectStatus::UNKNOWN as object status will invalidate
     * all information about the object.
     * @note This function doesn't test condition if the status
     * contains id of local object. Caller must do this test if required.
     */
    void notify_of_object_status
        (
        const ObjectStatusMessage &    status_info
        );

    /**
     * Refresh object status cache timeout.
     */
    void refresh_object_status
        (
        const ObjectId &    object_id
        );

    /**
     * Get cached status of object @a object_id.
     * If the provider knows nothing about the object then @a object_status is
     * set to @c ObjectStatus::UNKNOWN and @a node_id is set to null.
     */
    void get_object_status
        (
        const ObjectId &    object_id,
        ObjectStatus &      object_status,
        NodeId &            node_id
        );

    /**
     * Advance cache of the provider to current time.
     */
    void advance_cache();

    /**
     * Flush entire cache of the provider.
     */
    void flush_cache();

private:
    /** Object provider internal sctructure. */
    struct My;
    
    /** Object provider internal sctructure.
     * Stores object status of some remote object.
     */
    struct ObjectInfo;
    
private:
    std::auto_ptr< My >     my;
        /**< Object provider internal data. Always valid pointer. */

    }; // class ObjectProvider


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_OBJECT_PROVIDER_H
