/** @file object_provider_state.cpp
 * Object provider state.
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

 
#include "object/object_provider_state.h"
#include "serializer/stype_int.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"


namespace Massiv {
namespace Core {
 

ObjectProviderState::ObjectProviderState() :
    next_relative_object_id( 0 )
    {
    hope( provider_id.is_null() ); // Paranoia.
    }


ObjectProviderState::ObjectProviderState
    (
    const ProviderId &  _provider_id,
    UInt64              _next_relative_object_id
    ) :
    provider_id( _provider_id ),
    next_relative_object_id( 0 )
    {
    hope( ( _next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    next_relative_object_id = _next_relative_object_id;
    }


ObjectProviderState::ObjectProviderState
    (
    const ObjectProviderState & s
    ) :
    provider_id( s.provider_id ),
    next_relative_object_id( s.next_relative_object_id )
    {
    }
    

ObjectProviderState & ObjectProviderState::operator=
    (
    const ObjectProviderState & s
    )
    {
    provider_id             = s.provider_id;
    next_relative_object_id = s.next_relative_object_id;
    return *this;
    }


void ObjectProviderState::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    SInt<Int64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT> next_id( next_relative_object_id );
    
    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    
    provider_id.bitstr_write( bos, desc );
    next_id.bitstr_write( bos, desc );
    }


void ObjectProviderState::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    SInt<Int64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT> next_id;

    provider_id.bitstr_read( bis, desc );
    next_id.bitstr_read( bis, desc );
    
    next_relative_object_id = next_id;
    
    /// @todo Throw serialization exception.
    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    }

    
void ObjectProviderState::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    SInt<Int64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT> next_id( next_relative_object_id );

    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    
    tw.line() << "id";
    tw.write_space();
    provider_id.text_write( tw, desc );
    tw.write_space();
    tw.line() << "next_relative_id";
    tw.write_space();
    next_id.text_write( tw, desc );
    }


void ObjectProviderState::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    SInt<Int64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT> next_id;

    tr.require_string( "id" );
    provider_id.text_read( tr, desc );
    tr.require_string( "next_relative_id" );
    next_id.text_read( tr, desc );
    
    next_relative_object_id = next_id;

    /// @todo Throw serialization exception.
    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    }
    
    
const ProviderId & ObjectProviderState::get_provider_id() const
    {
    return provider_id;
    }


UInt64 ObjectProviderState::generate_provider_relative_object_id()
    {
    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    
    if( next_relative_object_id == UINT64( 0 ) )
        {
        return 0;
        }
    else
        {
        struct
            {
            UInt64  id : ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT;
            }   next_id;

        UInt64  result  = next_relative_object_id;
        
        // Increment next_relative_object_id to next available relative id.
        // This code will set next_relative_object_id to zero if the
        // pool of all available relative ids is exhausted.
        next_id.id      = result;
        next_id.id++; // This will wrap the id to zero if the pool is exhausted.
        next_relative_object_id = next_id.id;
        
        return result;
        }
    }

    
bool ObjectProviderState::is_provider_relative_object_id_pool_exhausted() const
    {
    hope( ( next_relative_object_id >> ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == UINT64( 0 ) );
    
    return ( next_relative_object_id == UINT64( 0 ) );
    }    


} // namespace Massiv::Core
} // namespace Massiv
