/** @file object_provider_state.h
 * Object provider state.
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECT_PROVIDER_STATE_H
#define MASSIV_OBJECT_PROVIDER_STATE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROVIDERID_H
#include "object/providerid.h"
#endif


namespace Massiv {
namespace Core {

namespace Test {
    class SerializerTest;
}


/**
 * Provider state stores everything necessary for archiving and unarchiving
 * object provider's state or transfering object provider over network.
 * It is also used by object provider to generate object provider
 * relative ids.
 *
 * Each object provider (resp. its state) contains pool of object provider
 * relative ids. Once the pool is exhausted, the provider is not able to
 * generate any relative id any more.
 *
 * The relative id pool is just a range of integers. Size of the range
 * is defined by ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT.
 * Relative id zero id used to represent that the pool is exhausted.
 *
 * @see ObjectProvider
 *
 * @ingroup object_management
 */
class ObjectProviderState
    {
    friend class Test::SerializerTest;
public:

    /**
     * Constructor.
     * Creating ObjectProvider with provider state inited with
     * this constructor will create provider with null id and
     * inability to generate any object id.
     */
    ObjectProviderState();

    /**
     * Constructor.
     *
     * @param provider_id ProviderId of the object provider.
     * @param next_relative_object_id Next available provider-relative object id.
     *      If set to zero the object provider inited with this state
     *      won't be able to generate any object id.
     *
     * @pre @a next_relative_object_id is valid provider-relative object id,
     * ie. condition "(@a next_relative_object_id >> @c ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) == 0"
     * must be true.
     */
    ObjectProviderState
        (
        const ProviderId &  provider_id,
        UInt64              next_relative_object_id
        );

    /**
     * Copy constructor.
     */
    ObjectProviderState
        (
        const ObjectProviderState & state
        );
        
    /**
     * Operator=.
     */
    ObjectProviderState & operator=
        (
        const ObjectProviderState & state
        );

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

public:

    /**
     * Get ProviderId stored in the object provider's state.
     */
    const ProviderId & get_provider_id() const;

    /**
     * Generate new provider-relative object id.
     * @return Zero if provider-relative object id pool is exhausted.
     */
    UInt64 generate_provider_relative_object_id();

    /**
     * Check whether provider-relative object id pool is exhausted.
     */
    bool is_provider_relative_object_id_pool_exhausted() const;

private:

    ProviderId              provider_id;
        /**< Identifier of the object  provider.*/

    UInt64                  next_relative_object_id;
        /**< Next available provider-relative object id.
         * Zero if provider-relative object id pool is exhausted.
         */

    }; // struct ObjectProviderState


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_OBJECT_PROVIDER_STATE_H
