/** @file object_status.h
 * Object status.
 * Used by object searching protocols.
 */

/*
 * begin        : Thu Apr 11 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECT_STATUS_H
#define MASSIV_OBJECT_STATUS_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#include <iosfwd>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup object_management
 * @{
 */

/**
 * ObjectStatus is used to store last known status of some object
 * ie. if the object exists some in the system or it has been
 * destroyed etc.
 * The status of objects is cached by object providers.
 *
 * Every status is associated with NodeId. For example if the status
 * is "object is alive" then associated node id says where 
 * the object is probably located.
 *
 * @see ObjectProvider.
 *
 */
class ObjectStatus
    {
public:

    /**
     * Status values.
     * When changed don't forget to update STATUS_COUNT.
     */
    enum Status
        {
        ALIVE               = 0,
            /**< Object exists somewhere in the system.
             * Associated node id specifies las known location of the object.
             */

        NOT_ACCESSIBLE      = 1,
            /**< Object is not accessible in the system. (The object can
             * be deleted or node where the object is located is not
             * accessible.) Associated node id specifies node which
             * destroyed the object, or which done global search and
             * didn't find the object.
             */

        UNKNOWN             = 2,
            /**< We don't know anything about the object.
             * Associated node id is null.
             * This status is also used to invalidate cached status.
             */
        }; // Status

    enum
        {
        STATUS_COUNT        = 3,
            /**< Internal. Number of status values. For checking status validity
             * only. When changed don't forget to update STATUS_BITS_COUNT.
             */

        STATUS_BITS_COUNT   = 2
            /**< Number of bits required to store the status. */
        };

public:

    /** Constructor. */
    INLINE ObjectStatus();

    /** Constructor. */
    INLINE ObjectStatus
        (
        Status  s
        );

    /** Copy constructor. */
    INLINE ObjectStatus
        (
        const ObjectStatus & s
        );

    /** Operator=. */
    INLINE ObjectStatus & operator=
        (
        const ObjectStatus & s
        );

    /** Operator==. */
    INLINE bool operator==
        (
        const ObjectStatus & s
        ) const;

    /** Operator!=. */
    INLINE bool operator!=
        (
        const ObjectStatus & s
        ) const;

    /** Set. */
    INLINE void set
        (
        Status _status
        );

    /** Get status value. */
    INLINE Status get_value() const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description &
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description &
        );

public:

    /**
     * Convert to string.
     */
    std::string to_string() const;

private:
    Status      status;
        /**< Status value. */

    }; // ObjectStatus

    
/**
 * Write status to output string stream.
 */
std::ostream & operator<<
    (
    std::ostream &          s,
    const ObjectStatus &    status
    );

/** @} */ // addtogroup object_management


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "object/object_status_inline.h"
#endif


#endif // MASSIV_OBJECT_STATUS_H
