/** @file providerid.h
 * Object provider id.
 */

/*
 * begin        : Thu Apr 11 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_PROVIDERID_H
#define MASSIV_PROVIDERID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_OBJECT_FWD_h
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#include <iosfwd>


namespace Massiv {
namespace Core {

namespace Test {
    class ObjectProviderTest;
}


/**
 * ProviderId is an identification number that uniquely identifies object
 * provider within the entire system.
 *
 * There are two special provider ids.
 * <ul>
 *  <li> Virtual provider id. Object provider with virtual provider id
 *      never exists. Virtual provider id is used to generate and detect
 *      proxy node object ids (see ObjectId). Also it is used
 *      by ObjectManager's caches when the object manager needs to
 *      cache some information indexed by object provider which is not
 *      owned by local system.
 *  <li> Dummy provider id. Object provider with dummy provider id is
 *      used to create objects on client nodes, because normal object provider
 *      can be owned only by server nodes.
 * </ul>
 *
 * @see ObjectProvider
 *
 * @ingroup objectid
 * @ingroup object_management
 */
class ProviderId
    {
    friend class Test::ObjectProviderTest;

public:

    enum
        {
        PROVIDER_ID_BITS_COUNT = 9,
            /**< Number of bits required to store provider id.
             * Must be >= 1.
             * See ObjectId for upper limit, but it can't be bigger
             * than 32 (size of int under 32-bit architectures).
             *
             * Number (2^PROVIDER_ID_BITS_COUNT-1-k) specifies maximum number
             * of unique object providers in the entire system, where k is
             * number of special object provider ids.
             */
        PROVIDER_RELATIVE_ID_BITS_COUNT = 32
            /** Number of bits required to store provider relative id.
             * Provider relative id us used to identify objects relative
             * to one object provider.
             * Provider relative id is not used directly by ProviderId class,
             * but by another classes like ObjectProvider or ObjectId.
             * See ObjectId for limits.
             *
             * Number (2^PROVIDER_RELATIVE_ID_BITS_COUNT-1) specifies maximum
             * number of unique object ids any provider can generate.
             */
        };

public:

    /**
     * ProviderId hash functor.
     */
    struct Hash
        {
        /** Hash function. */
        INLINE size_t operator()
            (
            const ProviderId & provider_id
            ) const throw();

        }; // Hash

    /**
     * ProviderId equalkey functor.
     */
    struct EqualKey
        {
        /** Equalkey function. */
        INLINE bool operator()
            (
            const ProviderId & provider_id1,
            const ProviderId & provider_id2
            ) const throw();

        }; // EqualKey

    /**
     * ProviderId less functor.
     */
    struct Less
        {
        /** Less function. */
        INLINE bool operator()
            (
            const ProviderId & provider_id1,
            const ProviderId & provider_id2
            ) const throw();

        }; // Less

    friend struct ProviderId::Hash;
    friend struct ProviderId::EqualKey;
    friend struct ProviderId::Less;

public:

    enum
        {
        NULL_ID = 0
            /**< Null id. */
        };

    /** Constructor. Constructs null id. */
    INLINE ProviderId();

    /**
     * Create @a ProviderId from a number.
     * @pre The @a new_id_value is valid provide id,
     * ie. "(@a new_id_value >> @c PROVIDER_ID_BITS_COUNT ) == 0"
     * condition is true.
     */
    static ProviderId create_from_number
        (
        UInt32  new_id_value
        );
    
    /** Copy constructor. */
    INLINE ProviderId
        (
        const ProviderId & provider_id
        );

    /** Operator=. */
    INLINE ProviderId & operator=
        (
        const ProviderId & provider_id
        );

    /** Operator==. */
    INLINE bool operator==
        (
        const ProviderId & provider_id
        ) const;

    /** Operator!=. */
    INLINE bool operator!=
        (
        const ProviderId & provider_id
        ) const;

    /** Set to null. */
    INLINE void set_null();

    /** Check whether is null. */
    INLINE bool is_null() const;
    
public:

    /**
     * Create virtual provider id.
     */
    static ProviderId create_virtual_provider_id();
    
    /**
     * Create dummy provider id.
     */
    static ProviderId create_dummy_provider_id();

    /**
     * Returns true if the provider id is virtual provider id.
     */
    INLINE bool is_virtual_provider_id() const;
    
    /**
     * Returns true if the provider id is dummy provider id.
     */
    INLINE bool is_dummy_provider_id() const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description &
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description &
        );

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

public:

    /**
     * Get id value as number.
     */
    INLINE UInt32 get_id_value() const;

private:

    /**
     * Set new id value from a number.
     */
    void set_from_number
        (
        UInt32 new_id
        );
        
private:

    /**
     * Special provider id values.
     * @note For internal purposes only.
     */
    enum SpecialProviderId
        {
        VIRTUAL_PROVIDER_ID     = 0,
            /**< Virtual provider id value. */
        DUMMY_PROVIDER_ID       = ( ( 1 << PROVIDER_ID_BITS_COUNT ) - 1 )
            /**< Dummy provider id value. */
        }; // SpecialProviderId

private:

    UInt32      id;
        /**< Identifier as a number. */

    }; // ProviderId

    
/**
 * Write provider id to output stream.
 */
std::ostream & operator<<
    (
    std::ostream &      s,
    const ProviderId &  provider_id
    );


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "providerid_inline.h"
#endif


#endif // MASSIV_PROVIDERID_H
