/** @file rpc.cpp
 * Remote procedure call-related stuff (implementation).
 */

/*
 * begin        : May 2003
 * copyright    : (C) 2003 by Stepan Vondrak, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "object/rpc.h"
#include "utility/debug.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "system/system.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "serializer/serializer.h"
#include "serializer/stype_string.h"
#include "object/metaobject.h"
#include "object/object_create.h"
#include "object/object_manager.h"
#include "object/class_manager.h"
#include "replication/replica_server.h"
#include "database/node_database.h"
#include "object/throwable.h"
#include "object/lib/lib_exceptions_helper.h"
#include "object/lib/lib_exceptions.h"


namespace Massiv {
namespace Core {

namespace {

/**
 * Since we can not use std::auto_ptr to incomplete types, simple C++ pointers
 * are used instead and reset_ptr( ptr, new_ptr ) replaces ptr.reset( new_ptr ).
 */
template< typename T >
void reset_ptr( T * & ptr, T * new_ptr = NULL )
    {
    if( ptr != new_ptr )
        {
        delete ptr;
        ptr = new_ptr;
        }
    }

} // anonymous namespace


// class MethodPacket

void MethodPacket::bitstr_write_arguments
    (
    BitOStream &,
    const Serializer::Description &
    ) const
    {
    }


void MethodPacket::bitstr_read_arguments
    (
    BitIStream &,
    const Serializer::Description &
    )
    {
    }


void MethodPacket::text_write_arguments
    (
    TextWriter &,
    const Serializer::Description &
    ) const
    {
    }


void MethodPacket::text_read_arguments
    (
    TextReader &,
    const Serializer::Description &
    )
    {
    }


void MethodPacket::bitstr_write_results
    (
    BitOStream &,
    const Serializer::Description &
    ) const
    {
    }


void MethodPacket::bitstr_read_results
    (
    BitIStream &,
    const Serializer::Description &
    )
    {
    }


void MethodPacket::text_write_results
    (
    TextWriter &,
    const Serializer::Description &
    ) const
    {
    }


void MethodPacket::text_read_results
    (
    TextReader &,
    const Serializer::Description &
    )
    {
    }


std::string MethodPacket::get_argument_value
    (
    int index
    ) const
    {
    RAISE_CORE_LIB_EXCEPTION( InvalidArgumentException, "Invalid argument index." );
    }


// class RPCObject

RPCObject::RPCObject()
      : data( NULL )
    {
    }


RPCObject::RPCObject
    (
    const RPCObject & other
    ) : Super( other ),
        data( NULL ),
        class_type_id( other.class_type_id ),
        method_index( other.method_index ),
        client_to_server( other.client_to_server ),
        type( other.type ),
        request_destination( other.request_destination ),
        reply_destination( other.reply_destination ),
        node_object( other.node_object ),
        replica_timeout( other.replica_timeout )
    {
    if( other.data != NULL )
        {
        reset_ptr( data, new BitIStream( *other.data ) );
        }
    }


RPCObject::~RPCObject()
    {
    delete data;
    }


void RPCObject::initialize
    (
    const MethodPacket &        packet,
    ClassTypeId                 the_class_id,
    int                         the_method_index,
    WeakPointer< Object >       the_request_destination,
    WeakPointer< ResultObject > the_reply_destination,
    const STime &               the_replica_timeout,
    bool                        no_archive
    )
    {
    // Set no_archive flag.

    set_no_archive( no_archive );

    // Initialize data.

    BitOStream bos;
    packet.bitstr_write_arguments( bos, Serializer::Description() );
    reset_ptr( data, new BitIStream( bos.get_buffer() ) );

    // Set client_to_server if local node is client and request destination
    // is not local object.

    if( Massiv::System::get_local_node_id().is_type( NodeId::TYPE_CLIENT ) &&
        !the_request_destination.is_local() )
        {
        client_to_server = true;
        }

    // Copy parameters.
    // Do not copy if equal to default-constructed. This should reduce
    // size of serialized object a bit.

    class_type_id = the_class_id;
    method_index = the_method_index;
    type = TYPE_REQUEST;
    if( the_request_destination != null )
        {
        request_destination = the_request_destination;
        }
    if( the_reply_destination != null )
        {
        reply_destination = the_reply_destination;
        }
    if( the_replica_timeout != STime() )
        {
        replica_timeout = the_replica_timeout;
        }

    // Send address of this node in the request if #reply_destination is valid,
    // #replica_timeout is valid, or #client_to_server is true.

    if( ( reply_destination != null ) || ( replica_timeout.read() != STime() ) ||
        client_to_server.read() )
        {
        node_object = ObjectId( Massiv::System::get_local_node_id() );
        }
    }


RPCObject & RPCObject::operator=
    (
    const RPCObject & other
    )
    {
    Super::operator=( other );

    if( other.data != NULL )
        {
        reset_ptr( data, new BitIStream( *other.data ) );
        }
    else
        {
        reset_ptr( data );
        }
    class_type_id = other.class_type_id;
    method_index = other.method_index;
    type = other.type;
    request_destination = other.request_destination;
    reply_destination = other.reply_destination;
    node_object = other.node_object;
    replica_timeout = other.replica_timeout;

    return *this;
    }


Pointer< ResultObject > RPCObject::create_and_send
    (
    WeakPointer< Object >       destination,
    const STime &               time,
    const MethodPacket &        packet,
    ClassTypeId                 class_type_id,
    int                         method_index,
    bool                        require_reply,
    const STime &               replica_timeout,
    bool                        no_archive
    )
    {
    // Allocate result object.

    Pointer< ResultObject > result_object;
    if( require_reply )
        {
        // If destination is client and the call is not local, require_reply
        // must be false.

        if( destination.get_object_id().is_proxy_node_object_id() &&
            destination.get_object_id() != Massiv::System::get_local_node_id() )
            {
            // Problems with node_id type being SERVICE should be solved elsewhere,
            // no migration at all is allowed to such nodes.

            if( destination.get_object_id().proxy_extract_node_id().is_type( NodeId::TYPE_CLIENT ) )
                {
                // This will crash if class_type_id or method_index is invalid, but who cares...

                std::ostringstream ss;
                const MetaObject * const metaobject =
                            Global::class_manager().get_metaobject( class_type_id );
                ss << "Tried to request reply to RPC request from client node.\n" <<
                      "This is not allowed for security reasons.\n" <<
                      "Interface " << class_type_id << ", method " <<
                      metaobject->get_method_infos()[ method_index ].name << std::endl;
                RAISE_CORE_LIB_EXCEPTION( RemoteCallFailedException, ss.str() );
                }
            }

        // Allocate result object.

        result_object = CreateObject< ResultObject >( no_archive );
        }

    // Allocate and initialize RPC request.

    Pointer< RPCObject > rpc_object
        (
        CreateObject< RPCObject >
            (
            packet,
            class_type_id,
            method_index,
            destination,
            result_object,
            replica_timeout,
            no_archive
            )
        );

    // Check RPC object.
    // FIXME: this may throw ExSerializationError to lib code.

    #ifdef OPTION_DEBUG
    rpc_object->check_properties();
    #endif

    // Send RPC request.

    if( !require_reply )
        {
        rpc_object->migrate_to( destination, time );
        }
    else
        {
        result_object->rpc_handle = rpc_object->schedule_to( destination, time );
        }

    // Ensure that RPC object is active-unpinned so that it could migrate
    // immediatelly.

    Global::garbage_collector().ensure_object_active_unpinned( rpc_object );

    return result_object;
    }


void RPCObject::create_and_send_to_replicas
    (
    WeakPointer< Object >       object,
    const STime &               time,
    const MethodPacket &        packet,
    ClassTypeId                 class_type_id,
    int                         method_index
    )
    {
    // Get list of nodes to replicate to.

    std::vector< NodeId > nodes;
    ObjectManager & object_manager = Global::object_manager();
    hope( object_manager.replica_server.get() );
    object_manager.replica_server->gather_replication_nodes( nodes, object.get_object_id() );

    // Create RPCObjects and migrate them to the nodes.

    Pointer< RPCObject > master_rpc_object;
    std::vector< NodeId >::const_iterator node_it;
    NodeDatabase & node_database = Global::node_database();
    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        const NodeId & node_id = *node_it;

        // Create RPC object to send.

        Pointer< RPCObject > rpc_object;
        if( master_rpc_object == null )
            {
            // Create master RPC object.

            master_rpc_object = CreateObject< RPCObject >
                ( packet, class_type_id, method_index, object, null, STime(), false );
            master_rpc_object->type = TYPE_REPLICA_REQUEST;
            rpc_object = master_rpc_object;
            }
        else
            {
            // Clone master RPC object.

            rpc_object = master_rpc_object->clone().convert();
            }

        // Set destination node object pointer.

        const NodeDatabase::NodeData & node_data = node_database.get_node_data( node_id );
        rpc_object->node_object = node_data.get_proxy_node_object_id();

        // Check RPC object.
        // FIXME: this may throw ExSerializationError to lib code.

        #ifdef OPTION_DEBUG
        rpc_object->check_properties();
        #endif

        // Send it.

        rpc_object->migrate_to( rpc_object->node_object, time );
        }
    }


std::string RPCObject::type_string() const
    {
    switch( type )
        {
        case TYPE_REQUEST:
            return "request";
        case TYPE_REPLICA_REQUEST:
            return "replica_request";
        case TYPE_REPLY:
            return "reply";
        case TYPE_DELIVERY_FAILURE:
            return "delivery_failure";
        case TYPE_INTERFACE_FAILURE:
            return "unsupported_interface_failure";
        case TYPE_EXCEPTION_FAILURE:
            return "exception_failure";
        }
    return "unknown_type";
    }


std::auto_ptr< MethodPacket > RPCObject::create_packet() const
    {
    if( data == NULL )
        {
        return std::auto_ptr< MethodPacket >( NULL );
        }

    // Find metaobject.

    const MetaObject * const metaobject = class_type_id.get_metaobject();

    // Create packet.

    std::auto_ptr< MethodPacket > packet( metaobject->create_method_packet( method_index ) );
    hope( packet.get() );

    // Initialize packet.

    BitIStream stream( *data );
    stream.rewind();
    if( is_request() )
        {
        packet->bitstr_read_arguments( stream, Serializer::Description() );
        }
    else
        {
        hope( type == TYPE_REPLY );
        packet->bitstr_read_results( stream, Serializer::Description() );
        }

    return packet;
    }


void RPCObject::delivered_to
    (
    WeakPointer< Object >,
    const STime &
    )
    {
    // check_properties() is done during serialization, so we can safely assume
    // that contents of the message are not complete bogus.

    hope( ( type == TYPE_REQUEST ) || !client_to_server );

    // Failure?

    if( is_failure() )
        {
        if( reply_destination.is_local() )
            {
            reply_destination->fail( type, exception );
            }
        else
            {
            // Weird, reply_destination is not local.

            std::ostringstream ss;
            ss << "Can't deliver RPC failure because reply_destination is not local.\n" <<
                "Details: " << to_string();
            Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
            }
        }
    else
        {
        // Request or reply - class id, method index and data are valid.

        // Request or reply?

        if( is_request() )
            {
            // Find metaobject.

            const MetaObject * const metaobject = class_type_id.get_metaobject();

            // Check if request destination is local.

            if( !request_destination.is_local() )
                {
                std::ostringstream ss;
                ss << "Can't deliver RPC request because request_destination is not local.\n" <<
                    "Details: " << to_string();
                Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_LOW, ss.str() );
                return;
                }

            // If this is client-to-server RPC request, request_destination must be the
            // account object. Ask it if the request is ok.

            if( client_to_server )
                {
                try
                    {
                    Pointer< Lib::AccountObjectInterface > account( request_destination.convert() );
                    if( !account->check_client_rpc_request( this ) )
                        {
                        // FIXME: kill connection to the client.

                        std::ostringstream ss;
                        ss << "RPC request from client node rejected by account object.\n" <<
                            "Details: " << to_string();
                        Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
                        return;
                        }
                    }
                catch( ... )
                    {
                    // FIXME: kill connection to the client.

                    std::ostringstream ss;
                    ss << "RPC request from client node rejected, because either object " <<
                        request_destination.get_object_id() <<
                        "\ndoes not implement check_client_rpc_request(), or the call has failed.\n"
                        "Details: " << to_string();
                    Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
                    return;
                    }
                }

            // Start replication if requested.

            if( ( type == TYPE_REQUEST ) && ( replica_timeout.read() != STime() ) )
                {
                request_destination->replicate_to( node_object, STime::current_time() + replica_timeout );
                }

            // Create method packet.

            std::auto_ptr< MethodPacket > packet
                ( metaobject->create_method_packet( method_index ) );
            hope( packet.get() );

            // Read arguments.

            hope( data != NULL );
            packet->bitstr_read_arguments( *data, Serializer::Description() );

            // Call the method.
            // If an exception is caught remap it to a throwable object
            // allocated on the heap and link to the RPC Object so that
            // it could migrate together with the RPC object back
            // to @c reply_destination.
            // Do not serialize the exception to @c data (as it is done
            // with out & inout arguments) because the actually serialized
            // type is not fixed. It will simplify implementation a lot
            // (ObjectManager handles dynamic object creation for us).

            ResultObject::State     call_result;
            Pointer< Throwable >    throwable;
            bool                    rethrow;

            rethrow = false;

            try
                {
                if( packet->call_method( request_destination ) )
                    {
                    call_result = ResultObject::STATE_SUCCESS;
                    }
                else
                    {
                    call_result = ResultObject::STATE_UNSUPPORTED_INTERFACE;
                    }
                }
            catch( Throwable & t )
                {
                if( reply_destination != null )
                    {
                    throwable = t.clone().convert();
                    }

                call_result = ResultObject::STATE_EXCEPTION;
                }
            catch( std::exception & e )
                {
                // We have caught a non-managed exception.
                // The exception will not only be propagated to the caller node
                // but it will also be rethrown on this node once the RPCObject
                // with the exception is sent to the caller.

                Lib::CoreException t( e.what() + std::string( "\n(remapped from " ) +
                    typeid( e ).name() + std::string( ")" ) );

                throwable = t.clone().convert();
                rethrow = true;

                call_result = ResultObject::STATE_EXCEPTION;
                }

            if( throwable != null )
                {
                // Check if type of this and callee node differ. If they do,
                // only TYPE_SHARED class kinds can be used (just to be safe)
                // and throwables of other kinds must be remapped.

                if( reply_destination != null )
                    {
                    hope( node_object != null );
                    hope( node_object.get_object_id().is_proxy_node_object_id() );
                    const NodeId destination_node_id =
                        node_object.get_object_id().proxy_extract_node_id();
                    const NodeId & local_node_id = System::get_local_node_id();
                    if( destination_node_id.get_type() != local_node_id.get_type() )
                        {
                        // Find metaobject for throwable and remap it if necessary.

                        const MetaObject * const throwable_meta = throwable->get_metaobject();
                        hope( throwable_meta );
                        if( throwable_meta->get_attribute_kind() != KIND_SHARED )
                            {
                            Lib::CoreException t( throwable->what() +
                                std::string( "\n(remapped from " ) +
                                throwable_meta->get_class_name() +
                                std::string( ")" ) );
                            throwable = t.clone().convert();
                            }
                        }
                    }
                
                // Set no_archive flag to match that of this RPCObject.

                throwable->set_no_archive( !is_archivable() );
                }

            // Send results back if requested.

            if( reply_destination != null )
                {
                hope( type == TYPE_REQUEST );

                // All nodes are able to send back results. However, results should never
                // be requested from client nodes, because those can't be trusted. This is
                // checked in #create_and_send(), it's not client's problem.

                // Successful call?

                if( call_result == ResultObject::STATE_SUCCESS )
                    {
                    BitOStream bos;

                    // Send back results.

                    packet->bitstr_write_results( bos, Serializer::Description() );
                    reset_ptr( data, new BitIStream( bos.get_buffer() ) );
                    type = TYPE_REPLY;
                    }
                else
                if( call_result == ResultObject::STATE_EXCEPTION )
                    {
                    // Send back exception.
                    
                    reset_ptr( data );
                    exception = throwable;
                    type = TYPE_EXCEPTION_FAILURE;
                    }
                else
                if( call_result == ResultObject::STATE_UNSUPPORTED_INTERFACE )
                    {
                    // Send back interface failure.

                    reset_ptr( data );
                    type = TYPE_INTERFACE_FAILURE;
                    }

                // Send it.

                send_reply();
                }
            else
                {
                // Log call failure.

                switch( call_result )
                    {
                    case ResultObject::STATE_EXCEPTION:
                        {
                        std::ostringstream ss;

                        ss  << "RPC failed - exception while executing the method.\n" <<
                            "Details: " << to_string();
                        
                        Global::log_warning
                            ( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
                        break;
                        }
                    
                    case ResultObject::STATE_UNSUPPORTED_INTERFACE:
                        {
                        std::ostringstream ss;

                        ss  << "RPC failed - object does not supported requested interface.\n" <<
                            "Details: " << to_string();
                        
                        Global::log_warning
                            ( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
                        break;
                        }
                    
                    default:
                        {    
                        break;
                        }
                    }
                }

            // Rethrow the caught non-managed exception on the local node
            // if requested.

            if( throwable != null && rethrow )
                {
                throw Exception( throwable->what() );
                }
            }
        else
            {
            hope( type == TYPE_REPLY );

            // Deliver results.

            if( reply_destination.is_local() )
                {
                reply_destination->deliver_result( this );
                }
            else
                {
                // Weird, reply_destination is not local.

                std::ostringstream ss;
                ss << "Can't deliver RPC results because reply_destination is not local.\n" <<
                    "Details: " << to_string();
                Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
                }
            }
        }
    }


void RPCObject::delivery_failed
    (
    DeliveryFailure     failure,
    const STime &
    )
    {
    if( failure == Object::EVENT_KILLED )
        {
        // The call was cancelled.
        // Log this on DEBUG.

        #ifdef OPTION_DEBUG
        std::ostringstream ss;
        ss << "RPC delivery cancelled. Warning: Cancelling is not reliable if objects migrate."
              "\nDetails: " << to_string();
        Global::log_debug( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
        #endif

        return;
        }

    if( ( type == TYPE_REQUEST ) && ( reply_destination != null ) )
        {
        // RPC request, wants reply. No check "error if this is client" is done,
        // see similar case in #delivered_to() for more info.

        // Log this on DEBUG.

        #ifdef OPTION_DEBUG
        std::ostringstream ss;
        ss << "Delivery of RPC request failed. Sending exception back to reply_destination."
              "\nDetails: " << to_string();
        Global::log_debug( Status::FACILITY_RPC, Status::PRIORITY_NORMAL, ss.str() );
        #endif

        // Modify RPCObject and send failure back to result object.

        type = TYPE_DELIVERY_FAILURE;
        reset_ptr( data );
        send_reply();
        }
    else
        {
        // Other delivery failure - log only.

        std::ostringstream ss;
        ss  << "Delivery of RPC " << type_string() << " failed.\n" <<
            "Details: " << to_string();
        Global::log_warning( Status::FACILITY_RPC, Status::PRIORITY_HIGH, ss.str() );
        }
    }


int RPCObject::memory_size() const
    {
    int size = Super::memory_size();
    size += sizeof( class_type_id ) + sizeof( data );
    if( data != NULL )
        {
        size += data->byte_size();
        }
    return size;
    }


std::string RPCObject::to_string() const
    {
    std::ostringstream ss;
    ss << "RPC " << type_string() << " (id: " << get_object_id() << ")";
    if( client_to_server )
        {
        ss << "; client to server";
        }
    ss << "\nrequest destination = " << request_destination.get_object_id();
    ss << "\nreply destination = " << reply_destination.get_object_id();
    ss << "\nnode object = " << node_object.get_object_id();
    ss << "\ntype = " << type_string();

    ClassManager & class_manager = Global::class_manager();
    const MetaObject * const metaobject = class_manager.get_metaobject_no_throw( class_type_id );
    
    if( metaobject )
        {
        const MetaObject::MethodInfos & method_infos = metaobject->get_method_infos();
        ss << "\nmethod = ";
        if( method_index >= 0 && method_index < static_cast< int >( method_infos.size() ) )
            {
            ss << method_infos[ method_index ];
            }
        else
            {
            ss << "<invalid method index>";
            }
        ss << "\ninterface = " << metaobject->get_class_name();

        if( is_request() )
            {
            std::auto_ptr< MethodPacket > packet( create_packet() );
            ss << "\narguments: ";
            TextWriter tw( ss );
            packet->text_write_arguments( tw, Serializer::Description() );
            }
        else
        if( type == TYPE_REPLY )
            {
            std::auto_ptr< MethodPacket > packet( create_packet() );
            ss << "\nresults: ";
            TextWriter tw( ss );
            packet->text_write_results( tw, Serializer::Description() );
            }
        else
        if( type == TYPE_EXCEPTION_FAILURE )
            {
            ss << "\nexception: " << exception;
            }
        }
    else
        {
        ss << "\nwarning: interface " << class_type_id << " is alien to this "
            "node. Can't determine method name and arguments/results.";
        }

    return ss.str();
    }


void RPCObject::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    // Write Object.

    Super::bitstr_write( bos, desc );

    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        // Class type id and method index.

        class_type_id.bitstr_write( bos, desc );
        method_index.bitstr_write( bos, desc );

        // Copy data.
        // FIXME: bitistream can't be copied to bitostream without seeking.

        if( has_data() )
            {
            hope( data != NULL );
            BitIStream stream( *data );
            stream.rewind();
            bos.align_to_byte();
            bos.write_bitistream( stream );
            }
        }
    }


void RPCObject::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    // Read Object.

    Super::bitstr_read( bis, desc );

    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        // Read class type id and method index.

        class_type_id.bitstr_read( bis, desc );
        method_index.bitstr_read( bis, desc );

        // Check RPCObject.

        check_properties( desc.source_node );

        // Read data.

        if( has_data() )
            {
            bis.align_to_byte();
            reset_ptr( data, new BitIStream( bis.tail() ) );
            }
        }
    }


void RPCObject::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    // Write Object.

    Super::text_write( tw, desc );

    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        // Properties already written, so we must begin a new line.

        tw.end_line();

        // Write class type id.

        tw.line() << "class ";
        class_type_id.text_write( tw, desc );

        // Write method name. Should never throw, interface should not be alien.
        
        ClassManager & class_manager = Global::class_manager();
        const MetaObject * const metaobject = class_manager.get_metaobject( class_type_id );

        tw.end_line();
        tw.line() << "method " << metaobject->get_method_infos()[ method_index ].name;

        // Convert bitstream to text stream and write it.

        if( has_data() )
            {
            tw.end_line();
            tw.line() << "data ";

            // Create method packet from data and write it to text stream.

            std::auto_ptr< MethodPacket > packet( create_packet() );
            if( is_request() )
                {
                packet->text_write_arguments( tw, Serializer::Description() );
                }
            else
            if( type == TYPE_REPLY )    
                {
                packet->text_write_results( tw, Serializer::Description() );
                }
            }
        }
    }


void RPCObject::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    hope( desc.source_node.is_null() );

    // Read Object.

    Super::text_read( tr, desc );

    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        // Properties already read, so we must begin a new line.

        tr.next_line_throw();

        // Read class type id.

        tr.require_string( "class" );
        class_type_id.text_read( tr, desc );

        {//sob

            // Read method name.

            tr.next_line_throw();
            tr.require_string( "method" );
            SString method_name;
            method_name.text_read( tr, desc );

            // Find the method.

            ClassManager & class_manager = Global::class_manager();
            const MetaObject * const metaobject = class_manager.get_metaobject( class_type_id );
            const MetaObject::MethodInfos & method_infos = metaobject->get_method_infos();
            MetaObject::MethodInfos::const_iterator it;
            for( it = method_infos.begin(); it != method_infos.end(); ++it )
                {
                if( it->name == method_name )
                    {
                    break;
                    }
                }

            if( it == method_infos.end() )
                {
                // No such method.

                std::ostringstream ss;
                ss << "Class interface " << class_type_id <<
                      " has not method called " << method_name;
                throw ExSerializationError( ss.str() );
                }

            method_index = it - method_infos.begin();

        }//eob

        // Check RPCObject.

        check_properties();

        // Read text stream and convert it to bitstream.

        if( has_data() )
            {
            tr.next_line_throw();
            tr.require_string( "data" );

            // Create new RPC packet.

            const MetaObject * const metaobject = class_type_id.get_metaobject();

            std::auto_ptr< MethodPacket > packet
                ( metaobject->create_method_packet( method_index ) );
            hope( packet.get() );

            // Read data and convert it to bitstream.

            BitOStream bos;
            if( is_request() )
                {
                packet->text_read_arguments( tr, Serializer::Description() );
                packet->bitstr_write_arguments( bos, Serializer::Description() );
                }
            else
            if( type == TYPE_REPLY )    
                {
                packet->text_read_results( tr, Serializer::Description() );
                packet->bitstr_write_results( bos, Serializer::Description() );
                }
            reset_ptr( data, new BitIStream( bos.get_buffer() ) );
            }
        }
    }


void RPCObject::send_reply()
    {
    hope
        (
        type == TYPE_REPLY ||
        type == TYPE_DELIVERY_FAILURE ||
        type == TYPE_INTERFACE_FAILURE ||
        type == TYPE_EXCEPTION_FAILURE
        );
    hope( node_object != null );
    hope( node_object.get_object_id().is_proxy_node_object_id() );

    WeakPointer< Object > node_object_backup = node_object;
    request_destination = null;
    node_object = null;
    if( client_to_server )
        {
        client_to_server = false;
        }

    #ifdef OPTION_DEBUG
    check_properties();
    #endif

    // Ensure that throwable object is active-unpinned so that RPC object's
    // migration group can migrate immediatelly.

    if( exception != null )
        {
        Global::garbage_collector().ensure_object_active_unpinned( exception );
        }

    // If reply_destination has "dummy" (client-private) object id, send
    // reply directly to the node_object. Otherwise use node_object as
    // a hint for localization protocol.

    if( reply_destination.get_object_id().get_provider_id().is_dummy_provider_id() )
        {
        migrate_to( node_object_backup, STime() );
        }
    else
        {
        Global::object_manager().hint_object_location
            (
            reply_destination.get_object_id(),
            node_object_backup.get_object_id().proxy_extract_node_id()
            );
        migrate_to( reply_destination, STime() );
        }
    }


void RPCObject::check_properties
    (
    const NodeId & source_node
    ) const
    {
    // If RPCObject has been sent from client node and this is server node,
    // check if #client_to_server is set properly.

    if( !source_node.is_null() &&
        source_node.is_type( NodeId::TYPE_CLIENT ) &&
        Massiv::System::get_local_node_id().is_type( NodeId::TYPE_SERVER ) )
        {
        if( !client_to_server.read() )
            {
            throw ExSerializationError( "Client tried to send RPCObject to server node "
                "with client_to_server set to false. Details: " + to_string() );
            }

        // If reply is requested, node_object must be equal to source_node.

        if( reply_destination != null )
            {
            if( !node_object.get_object_id().is_proxy_node_object_id() ||
                node_object.get_object_id().proxy_extract_node_id() != source_node )
                {
                std::ostringstream ss;
                ss << "Bad destination node object id " << node_object.get_object_id() <<
                     " of RPC request sent from client node " << source_node <<
                     ". Details: " << to_string();
                throw ExSerializationError( ss.str() );
                }
            }

        // From now on we can trust value of client_to_server...
        }

    // Check type.

    if( ( type != TYPE_REQUEST ) && ( type != TYPE_REPLICA_REQUEST ) &&
        ( type != TYPE_REPLY ) && ( type != TYPE_DELIVERY_FAILURE ) &&
        ( type != TYPE_INTERFACE_FAILURE ) && ( type != TYPE_EXCEPTION_FAILURE ) )
        {
        std::ostringstream ss;
        ss << "Invalid RPCObject type " << type << ". Details\n:" << to_string();
        throw ExSerializationError( ss.str() );
        }

    if( client_to_server.read() && ( type != TYPE_REQUEST ) )
        {
        std::ostringstream ss;
        ss  << "Invalid client-to-server RPCObject type " << type_string()
            << ". Details: " << to_string();
        throw ExSerializationError( ss.str() );
        }

    // Check if class type id is supported by this node.

    ClassManager & class_manager = Global::class_manager();
    const MetaObject * const metaobject = class_manager.get_metaobject_no_throw( class_type_id );
    if( !metaobject )
        {
        std::ostringstream ss;
        ss  << "RPC interface " << class_type_id << " is alien. Details: "
            << to_string();
        throw ExSerializationError( ss.str() );
        }

    // Check method index.

    const int method_count = metaobject->get_method_infos().size();
    if( ( method_index < 0 ) || ( method_index >= method_count ) )
        {
        std::ostringstream ss;
        ss  << "Invalid RPC method index " << method_index << ", interface "
            << class_type_id << " has " << method_count << " method(s). "
            << "Details: " << to_string();
        throw ExSerializationError( ss.str() );
        }

    // Check pointers.

    if( is_request() )
        {
        // Request destination must be set.

        if( request_destination == null )
            {
            std::ostringstream ss;
            ss << "request_destination is not set in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        // Request destination can't be "dummy" (client-private).

        if( request_destination.get_object_id().get_provider_id().is_dummy_provider_id() )
            {
            std::ostringstream ss;
            ss << "request_destinaion is private (dummy) object id in RPC " << type_string() <<
                ".\nDetails: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        // Exception must be null.

        if( exception != null )
            {
            std::ostringstream ss;
            ss << "exception must be null in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        if( type == TYPE_REQUEST )
            {
            // Check node object.

            if( ( reply_destination != null ) || ( replica_timeout.read() != STime() ) ||
                client_to_server.read() )
                {
                // Node object must be set.

                if( node_object == null )
                    {
                    std::ostringstream ss;
                    ss << "node_object is not set in RPC request with reply or replica request.\n"
                        "Details: " << to_string();
                    throw ExSerializationError( ss.str() );
                    }

                // Node object must be proxy node object id.

                if( !node_object.get_object_id().is_proxy_node_object_id() )
                    {
                    std::ostringstream ss;
                    ss << "node_object is not proxy node object id in RPC request.\n"
                        "Details: " << to_string();
                    throw ExSerializationError( ss.str() );
                    }

                // Node object must be client node if ResultObject is on client.

                if  (
                    reply_destination != null &&
                    reply_destination.get_object_id().get_provider_id().is_dummy_provider_id() &&
                    !node_object.get_object_id().proxy_extract_node_id().is_type
                                ( NodeId::TYPE_CLIENT )
                    )
                    {
                    std::ostringstream ss;
                    ss << "node_object is not client node in RPC request with reply to client.\n"
                        "Details: " << to_string();
                    throw ExSerializationError( ss.str() );
                    }
                }
            else
                {
                // Node object must be NULL.

                if( node_object != null )
                    {
                    std::ostringstream ss;
                    ss << "node_object must be NULL in RPC request without reply and replica\n"
                        "request if it's not sent from client to server node.\n"
                        "Details: " << to_string();
                    throw ExSerializationError( ss.str() );
                    }
                }
            }
        else if( type == TYPE_REPLICA_REQUEST )
            {
            // Replies to RPC to replicas are not allowed.

            if( reply_destination != null )
                {
                std::ostringstream ss;
                ss << "reply_destination must be NULL in RPC to replicas.\n"
                    "Details: " << to_string();
                throw ExSerializationError( ss.str() );
                }

            // Node object must be specified.

            if( node_object == null )
                {
                std::ostringstream ss;
                ss << "node_object is not set in RPC to replicas.\n"
                    "Details: " << to_string();
                throw ExSerializationError( ss.str() );
                }
            if( !node_object.get_object_id().is_proxy_node_object_id() )
                {
                std::ostringstream ss;
                ss << "node_object is not proxy node object id in RPC to replicas.\n"
                    "Details: " << to_string();
                throw ExSerializationError( ss.str() );
                }
            }
        }
    else if( is_reply() )
        {
        // Request destination must be NULL.

        if( request_destination != null )
            {
            std::ostringstream ss;
            ss << "request_destination is set in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        // Exception must be null iff type != TYPE_EXCEPTION_FAILURE.

        if( ( type != TYPE_EXCEPTION_FAILURE ) && ( exception != null ) )
            {
            std::ostringstream ss;
            ss << "exception must be null in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        if( ( type == TYPE_EXCEPTION_FAILURE ) && ( exception == null ) )
            {
            std::ostringstream ss;
            ss << "exception must not be null in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        // Node object must be NULL.

        if( node_object != null )
            {
            std::ostringstream ss;
            ss << "node_object is set in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }

        // Reply destination must be set.

        if( reply_destination == null )
            {
            std::ostringstream ss;
            ss << "reply_destination is not set in RPC " << type_string() << ".\n"
                "Details: " << to_string();
            throw ExSerializationError( ss.str() );
            }
        }
    }


// class ResultObject

void ResultObject::initialize
    (
    bool no_archive
    )
    {
    set_no_archive( no_archive );
    state = STATE_UNKNOWN;
    hope( result == null );
    }


std::string ResultObject::to_string() const
    {
    std::ostringstream ss;

    ss << "ResultObject [ state = ";

    switch( state )
        {
        case STATE_UNKNOWN:
            {
            ss << "STATE_UNKNOWN ]";
            break;
            }
        case STATE_DELIVERY_FAILED:
            {
            ss << "STATE_DELIVERY_FAILED ]";
            break;
            }
        case STATE_UNSUPPORTED_INTERFACE:
            {
            ss << "STATE_UNSUPPORTED_INTERFACE ]";
            break;
            }
        case STATE_EXCEPTION:
            {
            ss << "STATE_EXCEPTION ]:\n";
            ss << *exception.dereference();
            break;
            }
        case STATE_CANCELLED:
            {
            ss << "STATE_CANCELLED ]";
            break;
            }
        case STATE_SUCCESS:
            {
            std::auto_ptr< MethodPacket >   packet( create_results() );
            TextWriter                      tw( ss );

            ss << "STATE_SUCCESS ]:\n";
            packet->text_write_results( tw, Serializer::Description() );
            break;
            }
        default:
            {
            panic( "Invalid state of ResultObject." );
            }
        }

    return ss.str();
    }


void ResultObject::wait_for_reply()
    {
    // Wait until rpc result is delivered.

    System::wait_for_rpc_reply( this );

    // Throw an exception if there was an error.

    switch( state )
        {
        case STATE_SUCCESS:
            {
            break;
            }
        case STATE_DELIVERY_FAILED:
            {
            RAISE_CORE_LIB_EXCEPTION
                ( RemoteCallFailedException, "Failed to deliver RPC request." );
            }
        case STATE_UNSUPPORTED_INTERFACE:
            {
            RAISE_CORE_LIB_EXCEPTION
                (
                IllegalPointerConversionException,
                "Remote object does not implement the requested interface."
                );
            }
        case STATE_EXCEPTION:
            {
            exception->raise();
            }
        case STATE_CANCELLED:
            {
            RAISE_CORE_LIB_EXCEPTION
                ( RemoteCallFailedException, "The call was cancelled." );
            }
        case STATE_UNKNOWN:
            {
            panic( "Remote call reply received but state is still UNKNOWN." );
            }
        }
    }


void ResultObject::cancel()
    {
    if( state != STATE_UNKNOWN )
        {
        // RPC call already completed.

        return;
        }
        
    rpc_handle.kill();
    state = STATE_CANCELLED;
    }


std::auto_ptr< MethodPacket > ResultObject::create_results() const
    {
    if( state != STATE_SUCCESS )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            RemoteCallFailedException,
            "Results are not available."
            );
        }

    hope( result != null );
    hope( result->is_reply() );

    return result->create_packet();
    }


void ResultObject::deliver_result
    (
    Pointer< RPCObject > the_result
    )
    {
    if( state == STATE_CANCELLED )
        {
        // Already cancelled.

        return;
        }

    hope( result == null );
    hope( the_result != null );
    hope( the_result->is_reply() );

    state = STATE_SUCCESS;
    result = the_result;
    }


void ResultObject::fail
    (
    RPCObject::Type         failure,
    Pointer< Throwable >    the_exception
    )
    {
    if( state == STATE_CANCELLED )
        {
        // Already cancelled.

        return;
        }
    
    switch( failure )
        {
        case RPCObject::TYPE_DELIVERY_FAILURE:
            {
            state = STATE_DELIVERY_FAILED;
            break;
            }
        case RPCObject::TYPE_INTERFACE_FAILURE:
            {
            state = STATE_UNSUPPORTED_INTERFACE;
            break;
            }
        case RPCObject::TYPE_EXCEPTION_FAILURE:
            {
            exception = the_exception;
            state = STATE_EXCEPTION;
            break;
            }
        default:
            {
            panic( "Invalid failure type." );
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "rpc_inline.h"
#endif
