/** @file rpc.h
 * Remote procedure call-related stuff.
 */

/*
 * begin        : May 2003
 * copyright    : (C) 2003 by Stepan Vondrak, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_RPC_H
#define MASSIV_RPC_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_OBJECT_FACTORY_H
#include "object/object_factory.h"
#endif

#ifndef MASSIV_PROP_INT_H
#include "property/prop_int.h"
#endif

#ifndef MASSIV_PROP_BOOL_H
#include "property/prop_bool.h"
#endif

#ifndef MASSIV_PROP_POINTER_H
#include "property/prop_pointer.h"
#endif

#ifndef MASSIV_PROP_ENUM_H
#include "property/prop_enum.h"
#endif

#ifndef MASSIV_PROP_TIME_H
#include "property/prop_time.h"
#endif

#ifndef MASSIV_STYPE_VLI_H
#include "serializer/stype_vli.h"
#endif

#ifndef MASSIV_PROP_EVENT_HANDLE_H
#include "property/prop_event_handle.h"
#endif


namespace Massiv {
namespace Core {


/**
 * Structure holding method arguments and results which can be serialized
 * to/from any kind of stream.
 *
 * Default implementation of serializations is "no arguments" and "no results".
 *
 * Implementation of method packet for each method defined in IDL is generated
 * by factgen (see MethodPacketImpl and #METHOD_PACKET macro). It will contain
 * public variable for each method argument (with the same name) and variable
 * called @c _result if method returns a value.
 *
 * @ingroup rpc
 */
class MASSIV_EXPORTED MethodPacket
    {
    NO_COPY_CONSTRUCTOR( MethodPacket );
    NO_ASSIGNMENT_OPERATOR( MethodPacket );

public:

    /**
     * Constructor.
     */
    INLINE MethodPacket();

    /**
     * Destructor.
     */
    virtual INLINE ~MethodPacket();
    
public:

    // Arguments.

    /** Serialize arguments to bitstream. */
    virtual void bitstr_write_arguments
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /** Read arguments from bitstream. */
    virtual void bitstr_read_arguments
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /** Serialize arguments to text stream (on single line). */
    virtual void text_write_arguments
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /** Read arguments from text stream. */
    virtual void text_read_arguments
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

public:

    // Results.

    /** Serialize results to bitstream. */
    virtual void bitstr_write_results
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /** Read results from bitstream. */
    virtual void bitstr_read_results
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /** Serialize results to text stream. */
    virtual void text_write_results
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /** Read results from text stream. */
    virtual void text_read_results
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

public:

    // Generic access to properties.

    /**
     * Get value of @a index-th argument.
     *
     * Use @a index -1 to get value of result.
     * Examine metaobject to determine argument count, names and types.
     * Note that only some arguments will be initialized - values of
     * @c in and @c inout arguments will be valid if this is argument packet,
     * and result, @c inout and @c out arguments will be initialized if this
     * is result packet.
     *
     * @note Default implementation throws Massiv::Core::Lib::InvalidArgumentException.
     */
    virtual std::string get_argument_value
        (
        int index
        ) const;

public:

    /**
     * Call the method using pointer to the component this method has been
     * defined in.
     *
     * If class @c A has method @c foo() and class @c B inherits @c class A,
     * the pointer must be of type <tt>A *</tt> even if the object is actually
     * instance of @C B. Use MetaObject for class @c B to upcast pointer
     * to @c B to pointer to @c A.
     *
     * "in" and "inout" argument variables should be initialized before the
     * call, method can modify "inout", "out" and @c _result variables.
     */
    virtual void call_method
        (
        const VariantPointer & object
        ) = 0;

    /**
     * Call the method using object pointer to Object.
     *
     * @c object must point to a local object. However weak pointer is used
     * so that it could point to an object replica. If it points to a
     * non-replica then the core ensures that @c object is stack-strong-referenced
     * during the call.
     *
     * Returns @c true on success, @c false if pointer conversion throws
     * Massiv::Core::Lib::IllegalPointerConversionException exception.
     */
    virtual bool call_method
        (
        const WeakPointer< Object > & object
        ) = 0;

    }; // class MethodPacket


/**
 * Object keeping bitstream with RPC arguments or results.
 *
 * It inherits Object and can be migrated or archived. It's a bit hacky,
 * because RPC data are not kept as properties but in a single bitstream
 * and @c RPCObject extends standard serialization of @c Object.
 *
 * RPC object must never be sent to or via node which does not support interface
 * specified by #class_type_id. Otherwise @c RPCObject serialization (#text_read()
 * or #bitstr_read()) throws an excepction which should be handled by the object
 * manager - client nodes will be disconnected, server nodes should never do that.
 *
 * In the past "clean" RPC object was generated for each method, it has been
 * replaced by @c RPCPacket.
 *
 * @ingroup rpc
 * @ingroup object
 */
class MASSIV_EXPORTED RPCObject : public Object
    {
    MASSIV_OBJECT( RPCObject )
    
    /// Superclass.
    typedef Object Super;

public:

    /**
     * Type of RPC object.
     */
    enum Type
        {
        TYPE_REQUEST,           ///< RPC request.
        TYPE_REPLICA_REQUEST,   ///< Request sent to object replicas.
        TYPE_REPLY,             ///< Reply to successful REQUEST.
        TYPE_DELIVERY_FAILURE,  ///< Reply to failed REQUEST. Delivery failed.
        TYPE_INTERFACE_FAILURE, ///< Reply to failed REQUEST. Wrong interface failure.
        TYPE_EXCEPTION_FAILURE  ///< Reply to failed REQUEST. Exception during execution.
        };
     
public:

    /**
     * Default constructor.
     */
    RPCObject();

    /**
     * Copy constructor.
     */
    RPCObject
        (
        const RPCObject & other
        );

    /**
     * Destructor.
     */
    ~RPCObject();

    /**
     * Initialize object with RPC request.
     *
     * @param packet
     *      Packet with arguments.
     * @param the_class_id
     *      Id of class the method to call is defined in.
     * @param the_method_index
     *      Index of method within class specified by @a the_class_id.
     * @param the_request_destination
     *      Destination of the request.
     * @param the_reply_destination
     *      If non-null, reply #RPCObject will be sent to this object.
     * @param the_replica_timeout
     *      For how many seconds should callee object be replicated to caller node.
     * @param no_archive
     *      Object attribute no_archive will be set to this value. Use @c true
     *      for synchoronous RPC
     *
     * To initialize request to replicas, change @a type of the initialized
     * object.
     */
    void initialize
        (
        const MethodPacket &        packet,
        ClassTypeId                 the_class_id,
        int                         the_method_index,
        WeakPointer< Object >       the_request_destination,
        WeakPointer< ResultObject > the_reply_destination,
        const STime &               the_replica_timeout,
        bool                        no_archive
        );

    /**
     * Copy RPC object.
     */
    RPCObject & operator=
        (
        const RPCObject & other
        );

public:

    /**
     * Create and send RPC request to @a destination.
     *
     * Delivery will be scheduled at time specified by @a time.
     *
     * If @a require_reply is @c true, new result object will be returned.
     * @a require_reply must be @c false if @a destination is object owned
     * by client node.
     *
     * If @a no_archive is @c true, RPCObject containing the request and
     * ResultObject will be non-archivable. This flag should be used only
     * for synchronous RPC.
     *
     * See #initialize() for more info about @a packet, @a class_type_id,
     * @a method_index and @ replica_timeout arguments.
     */
    static Pointer< ResultObject > create_and_send
        (
        WeakPointer< Object >       destination,
        const STime &               time,
        const MethodPacket &        packet,
        ClassTypeId                 class_type_id,
        int                         method_index,
        bool                        require_reply,
        const STime &               replica_timeout,
        bool                        no_archive
        );

    /**
     * Create and send RPC request to replicas of @a object.
     *
     * Request will be delivered to all nodes which replicate @a object
     * from local node. This means that @a object should be local object.
     *
     * Delivery will be scheduled at time specified by @a time.
     *
     * Specified method should be @c const.
     *
     * Replies from replicas are not and will not be supported.
     */
    static void create_and_send_to_replicas
        (
        WeakPointer< Object >       object,
        const STime &               time,
        const MethodPacket &        packet,
        ClassTypeId                 class_type_id,
        int                         method_index
        );

public:

    /** Is this RPC request (REQUEST or REPLICA_REQUEST)? */
    INLINE bool is_request() const;

    /** Is this RPC reply (REPLY or is_failure())? */
    INLINE bool is_reply() const;

    /** Is this RPC failure reply (*_FAILURE)? */
    INLINE bool is_failure() const;

    /** Are @c data associated with this RPC? */
    INLINE bool has_data() const;

    /** Convert @a type to string. */
    std::string type_string() const;

public:

    /**
     * Create new method packet with method arguments/results.
     */
    std::auto_ptr< MethodPacket > create_packet() const;

protected:

    /**
     * Deliver RPC object.
     */
    virtual void delivered_to
        (
        WeakPointer< Object >   destination,
        const STime &           deliver_time
        );

    /**
     * Delivery failure callback.
     */
    virtual void delivery_failed
        (
        DeliveryFailure failure,
        const STime &   deliver_time
        );

    /**
     * Return how much memory the object occupies.
     */
    virtual int memory_size() const;

    /**
     * Dumps RPCObject.
     */
    virtual std::string to_string() const;
    
// Extended serialization.
protected:

    /**
     * Serialize object to bitstream.
     *
     * Appends @a class_type_id, @a method_index and @a data bitstream to
     * standard @c Object serialization.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read object from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize object to text stream.
     *
     * Appends lines:
     * - @c class @a class_name_derived_from_class_type_id
     * - @c method @a name_of_method_derived_from_method_index
     * - @c data followed by @a data bitstream converted to text stream using @c RPCPacket.
     */
    virtual void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read object from text stream.
     */
    virtual void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

private:

    /**
     * Send RPCObject back to #reply_destination.
     *
     * #data and #type must be modified before calling this method.
     * The method will clear #request_destination and #node_object and
     * migrate self to correct destination.
     */
    void send_reply();

    /**
     * Check RPCObject attributes after they have been read from bitstream
     * or text stream, before reading argument/result data.
     *
     * @param source_node Set when calling #check_properties() from
     *      bitstr_write() during migration message unpacking.
     *
     * @exception ExSerializationError Thrown if error is detected.
     */
    void check_properties
        (
        const NodeId & source_node = NodeId()
        ) const;

protected:

    BitIStream *                    data;
        /**< Arguments, results or failures serialized to bitstream.
             Only "simple" types can be transmitted. Strong pointers
             must not be marshalled into @c data (the referenced objects
             will not migrate together with the RPC object).

             @note This can not be std::auto_ptr because BitIStream
             is an incomplete type here. */

    ClassTypeId                     class_type_id;
        /**< Id of class which defines method to call.
             Note that this will be different from class id of destination
             object if method is inherited. */

    SVli< Int32 >                   method_index;
        /**< Index into table of methods of class specified by @a class_type_id. */

public:

    PBool                           client_to_server;
        /**< @c true iff this RPCObject is request sent from client node to
             server node. It must be set correctly, otherwise RPCObject will
             be discarded by bitstr_write() on the first server it's sent
             to. That means that value of #client_to_server can be always
             truested on server nodes once RPCObject passes that test. */

    PEnum< Type >                   type;
        /**< Type of RPC object. */

    PWeakPointer< Object >          request_destination;
        /**< Send call request to this object.
             The object is expected to implement interface @c class_type_id. */

    PWeakPointer< ResultObject >    reply_destination;
        /**< Send reply RPCObject to this ResultObject.
             If #type is @c REQUEST, this is @c NULL unless caller wants
             a reply. Always @c NULL if #type is @c REPLICA_REQUEST. */

    PWeakPointer< Object >          node_object;
        /**< Pointer to node object.

             If #type is @c REQUEST, #node_object is pointer to caller node.
             It will be valid (non-NULL) iff:
             - #reply_destination is valid. If #reply_destination is private
               (owned by client node, objects can't be migrated there directly),
               #node_object must be used as migration destination. Otherwise
               #node_object is just a hint for object localization protocols.
             - #replica_timeout is valid. Callee object will be replicated to
               this node.
             - #client_to_server is @c true.

             If #type is @c REPLICA_REQUEST, #node_object is pointer to node
             object which probably has replica of #request_destination. The
             request must be sent to the node. If there is no such replica on
             the node, the request can be silently dropped.

             If #type is something else, #node_object is not used and must be
             @c NULL. */

    PTime                           replica_timeout;
        /**< If valid, callee object should be replicated to caller node
             for given number of seconds. */

    PPointer< Throwable >           exception;
        /**< Exception object. Describes an exception leaked from the method
             called. Always allocated as a stand-alone object outside the RPC
             Object. This allows to examine the exception without throwing
             it and simplifies exception propagation implementation.

             Note that @c exception can refer to _any_ throwable object
             (unlike in/out arguments where the types are always fixed). */

    }; // class RPCObject


/**
 * Destination of RPC result.
 *
 * ResultObject is created for each RPC call. When RPCObject with remote call
 * results returns successfully, @a state of #ResultObject will be changed from
 * @a STATE_UNKNOWN to @a STATE_SUCCESS.
 *
 * @ingroup rpc
 * @ingroup object
 */
class MASSIV_EXPORTED ResultObject : public Object
    {
public:

    MASSIV_OBJECT( ResultObject );
    friend class RPCObject;

    /**
     * Result object state.
     */
    enum State
        {
        STATE_UNKNOWN,
            /**< Waiting for reply. */

        STATE_SUCCESS,
            /**< RPCObject with results returned. */

        STATE_DELIVERY_FAILED,
            /**< Remote call failed (delivery of request failed). */

        STATE_UNSUPPORTED_INTERFACE,
            /**< Remote object does not support requested interface. */

        STATE_EXCEPTION,
            /**< Exception during method execution. */

        STATE_CANCELLED
            /**< Call cancelled. */

        }; // enum State

public:

    /**
     * Initialize result object.
     *
     * @param no_archive Value of the no_archive flag.
     */
    void initialize
        (
        bool no_archive
        );

    /**
     * Dumps ResultObject.
     */
    virtual std::string to_string() const;

    /**
     * Wait until state changes from STATE_UNKNOWN.
     *
     * Returns on STATE_SUCCESS, throws Massiv::Core::RemoteCallFailedException if state changed
     * to STATE_DELIVERY_FAILED or STATE_CANCELLED, Massiv::Core::IllegalPointerConversionException
     * if state changed to STATE_UNSUPPORTED_INTERFACE, a throwable object if
     * state changed to STATE_EXCEPTION.
     *
     * This method is used internally to implement synchronous RPC. It should
     * be never called by RPC user. It might always timeout under certain
     * circumstances.
     */
    void wait_for_reply();
    
    /**
     * Cancel RPC call.
     */
    void cancel();

    /**
     * Create packet with method results.
     *
     * You must cast it to correct packet type to access the results.
     * Use #METHOD_PACKET macro to get packet type.
     */
    std::auto_ptr< MethodPacket > create_results() const;

private:

    /**
     * Deliver RPCObject with results.
     */
    void deliver_result
        (
        Pointer< RPCObject >    the_result
        );

    /**
     * Remote call failed.
     */
    void fail
        (
        RPCObject::Type         failure,
        Pointer< Throwable >    the_exception
        );

public:

    PEnum< State >              state;
        /**< State of result object. */

    PPointer< RPCObject >       result;
        /**< RPC object with results if @a state is @c STATE_SUCCESS. */

    PPointer< Throwable >       exception;
        /**< Caught exception if @a state is @c STATE_EXCEPTION. */

private:

    PEventHandle                rpc_handle;
        /**< Handle to the RPC object allowing to cancel RPC call. */

    }; // class ResultObject


} // namespace Massiv::Core


namespace Generated {


/**
 * MethodTags structure.
 *
 * This structure should be specialized for each class and it should contain
 * definition of single structure per each method defined in the class:
 *
 * @code
 * struct foo { enum index = N; };
 * @endcode
 *
 * where @c foo is name of the method and @c index is method index within the
 * class. These tag structures will be used to specialize
 * #::Massiv::Generated::MethodPacket class template.
 *
 * @note It's probably not necessary to specialize this structure for classes
 *       with no methods, default "implementation" is ok for them.
 *
 * @ingroup rpc
 */
template< typename Class >
struct MethodTags;


/**
 * Method arguments/results packet.
 *
 * This class should be specialized for each method defined in IDL (using
 * corresponding tag structure - see #MethodTags). It should inherit
 * #::Massiv::Core::MethodPacket and override its virtual methods as
 * needed.
 *
 * @ingroup rpc
 */
template< typename Tag >
class MethodPacketImpl; /* : public ::Massiv::Core::MethodPacket
    {
    }; */


} // namespace Massiv::Generated


namespace Core {
// back to core; actually useless for buch of #defines, but it seem nicer to me :-)


/**
 * Name of method packet class for method @a method defined in class @a class.
 *
 * @ingroup rpc
 */
#define METHOD_PACKET( class, method ) ::Massiv::Generated::MethodPacketImpl \
    < ::Massiv::Generated::MethodTags< class >::method >


/**
 * Index of @a method within @a class.
 *
 * @ingroup rpc
 */
#define METHOD_INDEX( class, method ) \
    ::Massiv::Generated::MethodTags< class >::method::index


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#ifndef MASSIV_RPC_INLINE_H
#include "object/rpc_inline.h"
#endif

#endif


#endif // MASSIV_RPC_H
