/** @file stack_root_object.h
 * Special object that owns all stack properties.
 */

/*
 * begin        : Mon Jun 21 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STACK_ROOT_OBJECT_H
#define MASSIV_STACK_ROOT_OBJECT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_PROPERTY_H
#include "object/object_property.h"
#endif


namespace Massiv {
namespace Core {


/**
 * Root property that owns all properties that reside on program stack.
 * The property is an ObjectProperty so that any stack Property could assume
 * that get_root_owner_object() always returns non-NULL.
 * StackRootObject is not a container property because it can not enumerate
 * owned properties (because of effeciency).
 * The object is not registered to the GC/ObjectManager and should not be
 * visible to client code (one can not have a pointer to it).
 * Any property is by default initialized to be owned by @c StackRootObject.
 * Since the property may be dirty, @c StackRootObject must be dirty too.
 * To ensure this we dirty the root object whenever replica time advances.
 * This is done by @c replica_time_advanced().
 *
 * @ingroup object_management
 * @ingroup property
 */
class MASSIV_EXPORTED StackRootObject : public ObjectProperty
    {
    NO_COPY_CONSTRUCTOR( StackRootObject );
    NO_ASSIGNMENT_OPERATOR( StackRootObject );

public:

    /**
     * Constructor.
     */
    StackRootObject();

    /**
     * Call-back invoked by ReplicaServer when current replica time
     * advances to make @c StackRootObject dirty.
     */
    void replica_time_advanced();

protected:

    /**
     * Unsupported operation.
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        );

    /**
     * Returns object name.
     */
    virtual std::string generated_name() const;

    /**
     * Unsupported operation.
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        );
    
    /**
     * Unsupported operation.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Unsupported operation.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Unsupported operation.
     */
    virtual void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Unsupported operation.
     */
    virtual void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

    }; // StackRootObject


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STACK_ROOT_OBJECT_H
