/** @file throwable.cpp
 * Throwable implementation.
 */

/*
 * begin        : Sat Oct 25 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "object/throwable.h"
#include "object/object_factory.h"
#include "serializer/serializer.h"
#include "serializer/stype_string.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "pointer/garbage_collector.h"
#include "global/global.h"


namespace Massiv {
namespace Core {


Throwable::Throwable()
    {
    }


Throwable::Throwable
    (
    const std::string & message
    ) : ExLibError( message )
    {
    }


Throwable::Throwable
    (
    const Throwable & value
    ) : ValueType( value ),
        ExLibError( value.what() )
    {
    }


Throwable::~Throwable() throw()
    {
    }


void Throwable::raise()
    {
    get_factory()->raise_exception( this );
    }


void Throwable::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    ValueType::bitstr_write( bos, desc );

    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        SString message( what() );
        SString stack( get_call_stack_history() );

        message.bitstr_write( bos, desc );
        stack.bitstr_write( bos, desc );
        }
    }


void Throwable::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    ValueType::bitstr_read( bis, desc );
    
    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        SString message;
        SString stack;
        
        message.bitstr_read( bis, desc );
        description = message;

        stack.bitstr_read( bis, desc );
        call_stack_history = stack;
        }
    }


void Throwable::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    ValueType::text_write( tw, desc );
    
    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        SString message( what() );
        SString stack( get_call_stack_history() );
        
        // Properties already written, so we must begin a new line.

        tw.end_line();
        
        // Write exception description.

        tw.line() << "what ";
        message.text_write( tw, desc );

        // Write call stack history.

        tw.end_line();
        tw.line() << "stack ";
        stack.text_write( tw, desc );
        }
    }


void Throwable::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    ValueType::text_read( tr, desc );
    
    if( flags_on( desc.object_flags, Serializer::Description::OBJECTF_PROPERTIES ) )
        {
        SString message;
        SString stack;

        // Properties already read, so we must begin a new line.

        tr.next_line_throw();

        // Read exception description.

        tr.require_string( "what" );
        message.text_read( tr, desc );
        description = message;

        // Read call stack history.

        tr.next_line_throw();
        tr.require_string( "stack" );
        stack.text_read( tr, desc );
        call_stack_history = stack;
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
