/** @file throwable.h
 * Throwable.
 */

/*
 * begin        : Sat Oct 25 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_THROWABLE_H
#define MASSIV_THROWABLE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_VALUE_TYPE_H
#include "object/value_type.h"
#endif


namespace Massiv {
namespace Core {


/**
 * Throwable is a base class for all managed exceptions.
 *
 * Unlike general exceptions, throwables are described by IDL and can
 * be thrown accross node boundaries. This is utilized by SRPC.
 * Throwable is a value type (see @c ValueType for implementation
 * requirements), that should be used for purposes of exception handling
 * only. Throwables are thrown by standard "throw" statement or @c raise(),
 * always by value.
 *
 * @note General core exceptions, caught by RPC dispatch code, are mapped
 * to @c Lib::CoreException so that they could be re-raised on the caller's
 * node. Lib code should throw and catch throwables only.
 *
 * @ingroup object
 * @ingroup exception
 */
class MASSIV_EXPORTED Throwable : public ValueType, public ExLibError
    {
public:
    
    MASSIV_OBJECT( Throwable );
    
    /**
     * Default constructor.
     */
    Throwable();

    /**
     * Constructor.
     */
    Throwable
        (
        const std::string & message
        );

    /**
     * Copy constructor.
     * Must be called by throwable type's copy constructor.
     */
    Throwable
        (
        const Throwable & value
        );

    /**
     * Destructor.
     */
    virtual ~Throwable() throw();

    /**
     * Throws a copy-value of self.
     */
    void raise();

protected:
    
    /**
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     */
    virtual void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     */
    virtual void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

    }; // class Throwable


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_THROWABLE_H
