/** @file value_type.h
 * Value type object.
 */

/*
 * begin        : Mon Jun 21 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VALUE_TYPE_H
#define MASSIV_VALUE_TYPE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_OBJECT_FACTORY_H
#include "object/object_factory.h"
#endif


namespace Massiv {
namespace Core {


/**
 * "Value type" is any managed object derived from ValueType class that
 * correctly implements ValueType interface and conforms to the enforced
 * implementation.
 *
 * ValueTypes are special "value objects" that can be easily created on program
 * stack or as member objects of another objects (or container properties).
 * Value types should not be instantiated as "stand-alone" objects.
 * Their purpose is to give client an opportunity to easily implement its own
 * "property-like" structures which are defined outside the core.
 * For example simple integer pair can be a value type.
 * Value types are like Properties (can be used wherever Property can be used,
 * can be instantiated like Properties) and like Objects (they must be
 * described by the IDL, are automatically serializable, ...).
 *
 * The difference from basic Object is that they can not be created
 * by @c CreateObject or @c ObjectOnStack templates because value types
 * have no @c initialize() pseudo-constructors. Value types define
 * "true" constructors instead.
 *
 * In order to value type be instantiable in a property-like fashion
 * (by simply calling value type's constructor) there are additional
 * requirements on value type implementation (in comparison to general Object).
 * Any created Object must be initialized. General objects are automatically
 * initialized by @c CreateObject or @c ObjectOnStack templates.
 * Value types must be initialized manually by a code in its constructor!
 * In particular any constructor must call @c initialize_object() as its
 * first command.
 *
 * Value type must implement its own copy constructor that ensures that
 * ValueType's copy constructor is called.
 * Value type's default constructor can call @c default_constructed()
 * (see @c Property for more info).
 * Value type must not have a destructor.
 * Value types can be inherited from another value types.
 * Value types used as in/out arguments in RPC calls must not contain
 * strong pointers (because of "simplified" marshalling).
 * It is possible to "embed" value types into another value types like
 * member objects are embedded into parent objects.
 *
 * @note Since ValueType is like a Property the restrictions on Property
 *       apply automatically to any value type.
 *
 * @code
 * class MyPair : public ValueType
 *  {
 *  MyPair()
 *      {
 *      initialize_object();
 *      default_constructed();
 *      }
 *  MyPair( int a, int b ) :
 *      a( a ), b( b )
 *      {
 *      initialize_object();
 *      }
 *  MyPair( const MyPair & pair ) :
 *      ValueType( pair ),
 *      a( pair.a ), b( pair.b )
 *      {
 *      initialize_object();
 *      }
 *  PInt32  a;
 *  PInt32  b;
 *  };
 * class Derived : public MyPair
 *  {
 *  Derived() :
 *      MyPair()
 *      {
 *      initialize_object();
 *      default_constructed();
 *      }
 *  Derived( int a, int b, int c ) :
 *      MyPair( a, b ), c( c )
 *      {
 *      initialize_object();
 *      }
 *  Derived( const Derived & derived ) :
 *      MyPair( derived ),
 *      c( derived.c )
 *      {
 *      initialize_object();
 *      }
 *  PInt32  c;
 *  };
 *
 * {
 * MyPair pair; // Property-like instantiation.
 * PArray<MyPair> pair_array; // Value-types can be inserted into containers.
 * pair_array.push_back( pair );
 * }
 * @endcode
 *
 * @ingroup object
 */
class MASSIV_EXPORTED ValueType : public Object
    {
    MASSIV_OBJECT( ValueType );
    
protected:

    /**
     * Default constructor.
     */
    ValueType();

    /**
     * Copy constructor.
     * Must be called by value type's copy constructor.
     */
    ValueType
        (
        const ValueType & value
        );

    /**
     * Initializes the object.
     * Must be called by any value type constructor manually as its
     * first command.
     */
    void initialize_object();

    }; // class ValueType


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_VALUE_TYPE_H
