/** @file rstring.cpp
 * Reference counted String implementation.
 */

/*
 * begin        : Tue Nov 6 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "utility/hash.h"
#include "other/rstring.h"
#include <cstdarg>
#include <cstdio>


namespace Massiv {
namespace Core {

// String::StringBody implementation

String::StringBody::StringBody
    (
    const Element * text
    ) : string_buffer( text )
    {
    }

String::StringBody::StringBody
    (
    const StdString & s
    ) : string_buffer( s )
    {
    }

// String::ElementReference implementation

String::ElementReference::ElementReference
    (
    String &               s,
    Position position
    ) : string( s ), position( position )
    {
    }

String::ElementReference & String::ElementReference::operator=
    (
    Element        e
    )
    {
    string.make_own_buffer();
    string.get_std_string().operator[]( position ) = e;
    return *this;
    }

// String::Hasher implementation

size_t String::Hasher::operator()
    (
    const String & key
    ) const
    {
    ext::hash<std::string> hasher;
    return hasher( key.get_const_std_string() );
    }

// String::EqualTo implementation

bool String::EqualTo::operator()
    (
    const String & s1,
    const String & s2
    ) const
    {
    return s1->compare( s2 ) == 0;
    }

// String::Less implementation

bool String::Less::operator()
    (
    const String & s1,
    const String & s2
    ) const
    {
    return s1->compare( s2 ) < 0;
    }

// String implementation

String::String() :
    body( NULL )
    {
    static String __empty_String( "" );
    body = __empty_String.body;
    }

String::String
    (
    const Element * text
    )
    {
    body = new StringBody( text );
    }

String::String
    (
    const StdString & s
    )
    {
    body = new StringBody( s );
    }

void String::make_own_buffer
    (
    ) const
    {
    if( body->get_reference_count() > 1 )
        {
        body = new StringBody( get_const_std_string() );
        }
    }

String::StdString & String::get_std_string
    (
    )
    {
    make_own_buffer();
    return body->string_buffer;
    }

String & String::operator+=
    (
    const String &     s
    )
    {
    get_std_string().operator+=( s.get_const_std_string() );
    return *this;
    }

String & String::operator+=
    (
    const StdString &   s
    )
    {
    get_std_string().operator+= ( s );
    return *this;
    }

String & String::operator+=
    (
    const Element *     s
    )
    {
    get_std_string().operator+=( s );
    return *this;
    }

String & String::operator+=
    (
    Element             c
    )
    {
    get_std_string().operator+=( c );
    return *this;
    }

String::ElementReference String::operator[]
    (
    Position            pos
    )
    {
    return ElementReference( *this, pos );
    }

String String::printf
    (
    const String::Element * format,
    ...
    )
    {
    va_list arg_list;
    char    buffer[1024]; /* FIXME */

    va_start(arg_list, format);
    vsprintf(buffer, format, arg_list);
    va_end(arg_list);

    return String( buffer );
    }

static bool match
    (
    const String::Element * pattern,
    const String::Element * string
    )
    {
    switch( pattern[0] )
        {
        case '\0':
            {
            return string[0] == '\0';
            }

        case '?':
            {
            /* '?': matches 1 element */

            if( string[0] == '\0' )
                    return false;

            return match( ++pattern, ++string );
            }

        case '*':
            {
            /* '*': matches arbitrary number of elements (including zero) */

            pattern++;

            while( string[0] != '\0' )
                {
                if( match( pattern, string ) )
                        return true;

                string++;
                }

            /* we are at the end of the string */

            return pattern[0] == '\0';
            }

        default:
            {
            /* element in pattern must match corresponding element in the string */

            if( pattern[0] != string[0] )
                    return false;

            return match( ++pattern, ++string );
            }
        };
    }

bool String::match_pattern
    (
    const Element * pattern
    ) const
    {
    return match( pattern, get_const_std_string().begin() );
    }


} // namespace Massiv::Core
} // namespace Massiv
