/** @file rstring.h
 * Reference counted string.
 */

/*
 * begin        : Tue Nov 6 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_RSTRING_H
#define MASSIV_RSTRING_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "global/reference.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @class String
 * Completely obsolete class - do not use.
 *
 * String with reference-counted buffers.
 * The buffer is implemented as std::string. To take advantage of copy-on-write
 * method, the following changes were made to string interface in comparison to
 * std::string from STL:
 * - Const methods from std::string are accesible using operator-\>() instead of
 *   operator.().
 * - Non-const methods from std::string are accesible using get_std_string(),
 *   which assumes that string will be modified so private copy is created
 *
 * @warning We assume that std::string is not reference-counted. String has been
 * tested only with STLPort.
 *
 * @todo Other operators (when needed). sprintf()-like method (or iostreams?).
 */
class String
    {
public:

    typedef char            Element;
        /**< String element. May be changed to wide char in future. */

    typedef std::string     StdString;
        /**< Implementation of string body uses standard std::string. */

    typedef StdString::size_type Position;
        /**< Position in string. */

private:

    /**
     * Shared string buffer.
     */
    class StringBody : public SharedObject
        {
    public:

        /**
         * Constructor.
         */
        StringBody
            (
            const Element *     text
            );

        /**
         * Constructor.
         */
        StringBody
            (
            const StdString &   s
            );

    public:

        StdString   string_buffer;
            /**< String buffer. */

        }; // class StringBody

public:

    /**
     * Object representing reference to string buffer.
     * Performs copy on write.
     */
    class ElementReference
        {
    public:

        /**
         * Constructor.
         */
        ElementReference
            (
            String &            s,
            Position            position
            );

        /**
         * Read element from the position.
         */
        operator Element () const;

        /**
         * Write element to the position.
         * Perform copy on write.
         */
        ElementReference & operator=
            (
            Element             e
            );

    public:

        String &                string;     /**< String the element points to.  */
        Position                position;   /**< Position in the string.        */
        }; // class ElementReference

public:

        /**
         * @todo aby se String mohl pouzit jako klic ve strukturach v STL,
         *       je treba definovat dalsi tridy popisujici chovani klice;
         *       nenapsat to jako ext::hash<>, std::equal_to<>, std::less<> ?
         */

        /**
         * Definition of hasher class for String.
         * Implemented as a wrapper for ext::hash<std::string>.
         */
        class Hasher
            {
        public:

            /**
             * Hash function.
             */
            size_t operator()
                (
                const String & key
                ) const;

            };

        /**
         * Definition of key equal_to comparator class for String.
         */
        class EqualTo
            {
        public:

            /**
             * String compare function.
             */
            bool operator()
                (
                const String & s1,
                const String & s2
                ) const;

            };

        /**
         * Definition of key less comparator class for String.
         */
        class Less
            {
        public:

            /**
             * String compare function.
             */
            bool operator()
                (
                const String & s1,
                const String & s2
                ) const;

            };

public:

    /**
     * Default constructor.
     */
    String();

    /**
     * Create String from null-terminated array of character elements.
     */
    String
        (
        const Element *         text
        );

    /**
     * Create String from standard std::string.
     */
    String
        (
        const StdString &       s
        );

    /**
     * Copy constructor.
     */
    String
        (
        const String &          s
        );

    /**
     * Destructroy string.
     */
    ~String();

public:

    /**
     * Copy string.
     */
    String & operator=
        (
        const String &         s
        );

    /**
     * Get const pointer to std::string.
     * All const functions from STL string are available via this. See STL documentation
     * for more info.
     */
    const StdString * operator->
        (
        ) const;

    /**
     * Cast to const std::string.
     */
    operator const StdString &
        (
        ) const;

    /**
     * Get string element.
     */
    Element operator[]
        (
        Position pos
        ) const;

    /**
     * Get object representing reference to string element.
     * This makes the following code possible:
     * @code
     * String a, b, c;
     * a = b = c = "test";
     * a[0] = b[2];
     * @endcode
     *
     * String buffers of @a b and @a c remain shared and @a a gets its own buffer.
     */
    ElementReference operator[]
        (
        Position            pos
        );

    /**
     * Append string.
     * @param s String to append to @a this.
     */
    String & operator+=
        (
        const String &      s
        );

    /**
     * Append standard std::string.
     * @param s String to append to @a this.
     */
    String & operator+=
        (
        const StdString &   s
        );

    /**
     * Append null-terminated string.
     * @param s String to append to @a this.
     */
    String & operator+=
        (
        const Element *     s
        );

    /**
     * Append single character.
     * @param c Character to append to @a this.
     */
    String & operator+=
        (
        Element             c
        );

public:

    /**
     * Return formated string.
     */
    static String printf
        (
        const Element *     format,
        ...
        );

    /**
     * Return true if string matches the pattern.
     * Supported wildcards are '*', '?'.
     */
    bool match_pattern
        (
        const Element *     pattern
        ) const;

public:

    /**
     * Get const pointer to std::string.
     */
    const StdString & get_const_std_string
        (
        ) const;

    /**
     * Get pointer to std::string.
     */
    StdString & get_std_string
        (
        );

public:

    /**
     * Make sure string buffer is not shared.
     */
    void make_own_buffer
        (
        ) const;

private:

    mutable Reference<StringBody>   body;
        /**< Reference to shared string buffer. */

    }; // class String


// Implementation

inline String::ElementReference::operator String::Element () const
    {
    return string.get_const_std_string().operator[]( position );
    }

inline String::String
    (
    const String & s
    ) : body( NULL )
    {
    body = s.body;
    }

inline String::~String()
    {
    }

inline String & String::operator=
    (
    const String & s
    )
    {
    body = s.body;
    return *this;
    }

inline const String::StdString * String::operator->
    (
    ) const
    {
    return &get_const_std_string();
    }

inline String::Element String::operator[]
    (
    Position        pos
    ) const
    {
    // Note: STLPort's operator[] returns const_reference, g++-3 returns charT.
    return get_const_std_string()[pos];
    }

inline const String::StdString & String::get_const_std_string
    (
    ) const
    {
    return body->string_buffer;
    }

inline String::operator const String::StdString &
    (
    ) const
    {
    return get_const_std_string();
    }


} // namespace Massiv::Core
} // namespace Massiv


#endif // _RSTRING_H_
