/** @file union_find.cpp
 * Union-find.
 */

/*
 * begin        : Sat Mar 17 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "other/union_find.h"
#include <iosfwd>

namespace Massiv {
namespace Core {


// Implementation of UnionFindBase::Group
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UnionFindBase::Group::Group()
    :   count( 0 ),
        head( NULL )
    {
    }


UnionFindBase::Group::Group
    (
    const Group & g
    ) : count( 0 ),
        head( NULL )
    {
    hope( !g.head && !g.count );
    }


UnionFindBase::Group::~Group()
    {
    if( head )
        {
        delete head;
        }
    }


void UnionFindBase::Group::init
    (
    UnionFindBase &     uf,
    VariantPointer      group_iterator,
    BaseElement &       e
    )
    {
    GroupList::iterator & group =
        *variant_cast<GroupList::iterator *>( group_iterator );

    if( !head )
        {
        head = new BaseElement( uf );
        }

    hope( head->is_unlinked() );
    hope( &*group == this );

    head->group = group;
    e.link_to( *head );
    }


bool UnionFindBase::Group::empty() const
    {
    if( !count )
        {
        hope( head->is_unlinked() );
        return true;
        }

    return false;
    }


int UnionFindBase::Group::get_count() const
    {
    return count;
    }


void UnionFindBase::Group::place_to_group()
    {
    count++;
    }


void UnionFindBase::Group::remove_from_group()
    {
    count--;
    }


// Implementation of UnionFindBase::BaseElement
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UnionFindBase::BaseElement::BaseElement
    (
    UnionFindBase & uf
    ) : union_find( uf )
    {
    init_default();
    }


UnionFindBase::BaseElement::BaseElement
    (
    const BaseElement & e
    ) : union_find( e.union_find )
    {
    init_default();
    union_find.merge( *this, const_cast<BaseElement &>( e ) );
    }


UnionFindBase::BaseElement::~BaseElement()
    {
    /* Unlink element. */

    unlink();

    /* Unregister element. */

    union_find.element_count--;
    }


void UnionFindBase::BaseElement::init_default()
    {
    /* Initialize. */

    next = prev = this;
    link_to_called = false;

    /* Place element into its own trivial group. */

    group = union_find.groups.end();

    /* Register element. */

    union_find.element_count++;
    }


bool UnionFindBase::BaseElement::is_unlinked() const
    {
    if( next == this )
        {
        hope( prev == this );
        hope( group == union_find.groups.end() );

        return true;
        }

    return false;
    }


void UnionFindBase::BaseElement::link_to
    (
    BaseElement & e
    )
    {
    hope( &union_find == &e.union_find );
    hope( !union_find.is_trivial_group( e ) );

    unlink();

    next = e.next;
    next->prev = this;
    prev = &e;
    prev->next = this;

    group = e.group;

    union_find.place_to_group( *this );
    link_to_called = true;
    }


void UnionFindBase::BaseElement::unlink()
    {
    /* Unlink. */

    next->prev = prev;
    prev->next = next;

    next = prev = this;

    if( link_to_called )
        {
        union_find.remove_from_group( *this );
        link_to_called = false;
        }

    /* Place to own group. */

    group = union_find.groups.end();

    /* NOTE: Careful, union_find.groups must still be alive! */
    }


// Implementation of UnionFindBase::BaseIterator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UnionFindBase::BaseIterator::BaseIterator
    (
    BaseElement * e
    ) : e( e )
    {
    }


UnionFindBase::BaseIterator::BaseIterator
    (
    ) : e( NULL )
    {
    }


UnionFindBase::BaseIterator::BaseIterator
    (
    const BaseIterator & it
    ) : e( it.e )
    {
    }


UnionFindBase::BaseIterator & UnionFindBase::BaseIterator::assign
    (
    const BaseIterator & it
    )
    {
    e = it.e;
    return *this;
    }


UnionFindBase::BaseIterator & UnionFindBase::BaseIterator::operator++()
    {
    hope( e );

    if( e == e->next )
        {
        /* This is iterator to trivial group. */

        e = NULL;
        }
    else
        {
        /* This is iterator to non-trivial group. */

        hope( e != e->group->head );
        e = e->next;
        }

    return *this;
    }


UnionFindBase::BaseIterator UnionFindBase::BaseIterator::operator++( int )
    {
    BaseIterator old = *this;
    operator++();
    return old;
    }


UnionFindBase::BaseElement & UnionFindBase::BaseIterator::get_reference()
    {
    hope( e );
    return *e;
    }


bool UnionFindBase::BaseIterator::operator==
    (
    const BaseIterator & it
    ) const
    {
    return e == it.e;
    }


bool UnionFindBase::BaseIterator::operator!=
    (
    const BaseIterator & i
    ) const
    {
    return !( *this == i );
    }


// Implementation of UnionFindBase::
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UnionFindBase::UnionFindBase
    (
    ) : element_count( 0 )
    {
    }


UnionFindBase::~UnionFindBase()
    {
    hope( !get_user_element_count() );

    /*
     * NOTE: All BaseElements must be deleted before
     *       UnionFindBase destruction.
     *       At this point the only remaining elements
     *       are the list heads in this-\>groups.
     *       Note that BaseElement::~BaseElement references
     *       this-\>groups so all elements must be deleted
     *       before this-\>groups destruction.
     *       This is achieved by manual groups.clear().
     */

    groups.clear();
    hope( !element_count );

    /* All BaseElement instances have been deleted. */
    }


bool UnionFindBase::same_groups
    (
    BaseElement & e1,
    BaseElement & e2
    )
    {
    if( e1.group == e2.group )
        {
        if( is_trivial_group( e1 ) )
            {
            hope( is_trivial_group( e2 ) );

            return &e1 == &e2;
            }
        else
            {
            hope( !is_trivial_group( e2 ) );

            return true;
            }
        }

    return false;
    }


void UnionFindBase::merge
    (
    BaseElement & e1,
    BaseElement & e2
    )
    {
    if( same_groups( e1, e2 ) )
        return;

    /* Handle trivial groups. */

    if( is_trivial_group( e1 ) )
        {
        if( is_trivial_group( e2 ) )
            {
            create_group( e1 );
            e2.link_to( e1 );
            }
        else
            {
            e1.link_to( e2 );
            }
        }
    else
        {
        if( is_trivial_group( e2 ) )
            {
            e2.link_to( e1 );
            }
        else
            {
            BaseElement *   t;
            BaseElement *   head;
            Group *         first_group;

            /* Handle non-trivial groups. */

            if( count( e1 ) > count( e2 ) )
                {
                merge( e2, e1 );
                return;
                }

            /* Change element "ownership" from group e1 to group e2. */

            first_group = &( *e1.group );
            head = first_group->head;
            t = head->next;

            while( t != head )
                {
                remove_from_group( *t );
                t->group = e2.group;
                place_to_group( *t );

                t = t->next;
                }

            /*
             * Connect two cyclic lists with head into one cyclic list with head.
             * This is done in two steps:
             * 1) unlink head from the first list -> we get cyclic list without head
             * 2) connect cyclic list without head to the second cyclic list with head.
             */

            hope( t == head );

            t->unlink();

            e1.next->prev = e2.prev;
            e2.prev->next = e1.next;

            e1.next = &e2;
            e2.prev = &e1;

            hope( first_group->empty() );
            }
        }
    }


void UnionFindBase::unlink
    (
    BaseElement & e
    )
    {
    if( e.is_unlinked() )
        return;

    e.unlink();
    }


int UnionFindBase::count
    (
    BaseElement & e
    )
    {
    return is_trivial_group( e ) ? 1 : e.group->get_count();
    }


void UnionFindBase::create_group
    (
    BaseElement & e
    )
    {
    GroupList::iterator it( groups.begin() );

    // FIXME: optimize?

    while( it != groups.end() )
        {
        if( it->empty() )
            {
            /* Reuse empty group. */

            break;
            }

        it++;
        }

    if( it == groups.end() )
        {
        /* Allocate new group. */

        groups.push_front( Group() );
        it = groups.begin();
        }

    /* Initialize / reuse group. */

    it->init( *this, VariantPointer( &it ), e );
    }


bool UnionFindBase::is_trivial_group
    (
    BaseElement & e
    )
    {
    return e.group == groups.end();
    }


void UnionFindBase::place_to_group
    (
    BaseElement & e
    )
    {
    hope( !is_trivial_group( e ) );
    e.group->place_to_group();
    }


void UnionFindBase::remove_from_group
    (
    BaseElement & e
    )
    {
    hope( !is_trivial_group( e ) );
    e.group->remove_from_group();
    }


int UnionFindBase::get_user_element_count()
    {
    /**
     * Return total number of elements minus
     * number of list heads in non-trivial groups.
     */
    return element_count - groups.size();
    }


UnionFindBase::BaseIterator UnionFindBase::begin
    (
    BaseElement & e
    )
    {
    if( !is_trivial_group( e ) )
        {
        return BaseIterator( e.group->head->next );
        }
    else
        {
        return BaseIterator( &e );
        }
    }


UnionFindBase::BaseIterator UnionFindBase::end
    (
    BaseElement & e
    )
    {
    if( !is_trivial_group( e ) )
        {
        return BaseIterator( e.group->head );
        }
    else
        {
        return BaseIterator( NULL );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
