/** @file union_find.h
 * Union-find.
 */

/*
 * begin        : Sat Mar 17 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_UNION_FIND_H
#define MASSIV_UNION_FIND_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif

#include <list>


namespace Massiv {
namespace Core {


/**@class UnionFindBase
 * This class manages groups on a given set of items. The basic
 * operations are group merging, enumerating items belonging
 * to a specified group and testing whether two items belong to
 * the same group.
 * UnionFindBase is NOT a container.
 * It does not store any elements (they are stored outside UnionFindBase instance,
 * UnionFindBase does not own them).
 *
 * Implementation notes:
 *  - there are two types of managed groups:
 *      - groups that have allocated their entries in @c UnionFindBase::groups
 *        (non-trivial groups), these groups can hold >= 1 elements
 *        (@c Group::head serves as element list head, holds no data)
 *      - groups that have no entries in @c UnionFindBase::groups
 *        (trivial groups), these groups hold exactly one element
 *        (@c Group::head serves as storage for the stored element)
 */
class UnionFindBase
    {
public:

    class BaseElement;
    class BaseIterator;
    class Group;

    friend class BaseElement;
    friend class BaseIterator;

    /**@class Group
     * Structure that holds information about non-trivial group.
     * Contains number of items in the group and element list head.
     * If the number of items is zero then the group is ready
     * to reuse.
     */
    class Group
        {
        friend class UnionFindBase;
        friend class BaseIterator;

    public:

        /**
         * Default constructor.
         * Leaves object in an uninitialized state,
         * @c init() is expected to be called then.
         */
        Group();

        /**
         * Used by STL when inserting object into @c GroupList.
         * @c g must be default-constructed object.
         */
        Group
            (
            const Group & g
            );

        /**
         * Destructor.
         */
        ~Group();

    private:

        /**
         * Initializes yet empty or default-constructed group.
         * Resulting group will be non-trivial and will contain element @c e.
         * @c group_iterator is expected to be pointer to GroupList::iterator
         * that references this group.
         */
        void init
            (
            UnionFindBase &     uf,
            VariantPointer      group_iterator,
            BaseElement &       e
            );

        /**
         * Checks if is empty.
         */
        bool empty() const;

        /**
         * Returns number of elements in the group.
         */
        int get_count() const;

        /**
         * Registers new element into the group.
         */
        void place_to_group();

        /**
         * Unregisters element from the group.
         */
        void remove_from_group();

    private:

        int             count;
            /**< Number of items in the group. */
        BaseElement *   head;
            /**< Group element list head (tail). */

        }; // Group

    typedef std::list<Group>        GroupList;

    /**@class BaseElement
     * Double-linked cyclic list element.
     */
    class BaseElement
        {
        friend class UnionFindBase;
        friend class BaseIterator;
        friend class Group;

    public:

        /**
         * Registers element into UnionFindBase and
         * places it into its own trivial group.
         */
        BaseElement
            (
            UnionFindBase & uf
            );

        /**
         * Copies the element and places it into the
         * same group as e.
         */
        BaseElement
            (
            const BaseElement & e
            );

        /**
         * Unlinks the element and destroys it.
         * UnionFindBase instance must still exist!
         */
        ~BaseElement();

    protected:

        /**
         * Initializes element.
         * Called from constructors.
         */
        void init_default();

        /**
         * Checks if this element is unlinked.
         */
        bool is_unlinked() const;

        /**
         * Links element to non-trivial group where e belongs to.
         */
        void link_to
            (
            BaseElement & e
            );

        /**
         * Unlinks the element and places it
         * into its own trivial group.
         */
        void unlink();

    protected:

        BaseElement *       next;
            /**< Next element in the list. */
        BaseElement *       prev;
            /**< Previous element in the list. */
        GroupList::iterator group;
            /**< Group this element belongs to (iterator to UnionFindBase.groups).
                 This iterator is never invalidated. */
        UnionFindBase &     union_find;
            /**< Manager. */
        bool                link_to_called : 1;
            /**< Was element linked using link_to()? */

        }; // BaseElement

    /**@class BaseIterator.
     * Iterator that enumerates elements within a group.
     * Implementation notes:
     *  - non-trivial groups are handled as double linked lists:
     *    list head serves as end(), head-\>next serves as begin()
     *  - trivial groups are handled in a special way:
     *    pointer to the only element in the group serves as begin(),
     *    NULL pointer serves as end()
     */
    class BaseIterator
        {
        friend class UnionFindBase;

    public:

        /**
         * Default constructor.
         */
        BaseIterator
            (
            );

        /**
         * Private constructor.
         * Called by UnionFindBase::begin()
         */
        BaseIterator
            (
            BaseElement * e
            );

        /**
         * Copy constructor.
         */
        BaseIterator
            (
            const BaseIterator & it
            );

        /**
         * Assign iterator.
         */
        BaseIterator & assign
            (
            const BaseIterator & it
            );

    public:

        /**
         * Move iterator, preincrement.
         */
        BaseIterator & operator++();

        /**
         * Move iterator, postincrement.
         */
        BaseIterator operator++( int );

    public:

        /**
         * Get reference to item iterator points to.
         */
        BaseElement & get_reference();

        /**
         * Compare iterators.
         */
        bool operator==
            (
            const BaseIterator & it
            ) const;

        /**
         * Compare iterators.
         */
        bool operator!=
            (
            const BaseIterator & i
            ) const;

    private:

        BaseElement * e;
            /**< Pointer to current element. */

        }; // BaseIterator

public:

    typedef BaseElement     Element;
    typedef BaseIterator    Iterator;

public:

    /**
     * Constructor.
     */
    UnionFindBase
        (
        );

    /**
     * Destructor.
     * UnionFindBase must not be deleted when it still manages
     * any elements (they must have been deleted before).
     */
    ~UnionFindBase();

    /**
     * Checks whether e1, e2 belongs to the same group.
     */
    bool same_groups
        (
        BaseElement & e1,
        BaseElement & e2
        );

    /**
     * Merges groups described by e1, e2.
     */
    void merge
        (
        BaseElement & e1,
        BaseElement & e2
        );

    /**
     * Unlinks the element from the group and
     * places it into its own trivial group.
     */
    void unlink
        (
        BaseElement & e
        );

    /**
     * Returns number of elements in group where e belongs to.
     */
    int count
        (
        BaseElement & e
        );

    /**
     * Return iterator pointing to the first element in the group specified by e.
     */
    BaseIterator begin
        (
        BaseElement & e
        );

    /**
     * Return iterator pointing at the end of the group specified by e.
     */
    BaseIterator end
        (
        BaseElement & e
        );

private:

    /**
     * Creates non-trivial group that it will contain element e.
     */
    void create_group
        (
        BaseElement & e
        );

    /**
     * Checks if e belongs to trivial group.
     */
    bool is_trivial_group
        (
        BaseElement & e
        );

    /**
     * Register e into its group.
     */
    void place_to_group
        (
        BaseElement & e
        );

    /**
     * Unregisters e from its group.
     */
    void remove_from_group
        (
        BaseElement & e
        );

    /**
     * Returns number of managed user elements.
     */
    int get_user_element_count();

private:

    GroupList   groups;
        /**< List of non-trivial groups. */
    int         element_count;
        /**< Number of elements currently managed
             (including list heads in non-trivial groups). */

    }; // UnionFindBase


/**@class UnionFind
 * UnionFind extends UnionFindBase by these features:
 *  - Element contains Type
 *  - fully operational Iterator that dereferences to Type
 *@code
 * UnionFind<UFItem>            uf;         // Declare UnionFind instance.
 * UnionFind<MyType>::Element   a( uf );    // Register item to managed set.
 * UnionFind<MyType>::Element   b( uf );    // Register item to managed set.
 * UnionFind<MyType>::Element   c( uf );    // Register item to managed set.
 *
 * uf.merge( a, b );                        // Merge groups containing a, b into one group.
 * hope( uf.same_group( a, b ) );         // a, b should belong to the same group.
 *@endcode
 *@todo out-of-class implementace
 */
template <class Type>
class UnionFind : public UnionFindBase
    {
public:

    /**@class Element
     */
    class Element : public Type, public UnionFindBase::BaseElement
        {
    public:

        /**
         * Registers the element into UnionFind.
         */
        Element
            (
            UnionFind<Type> &   uf,
            const Type &        value = Type()
            ) : Type( value ),
                BaseElement( uf )
            {
            }

        }; // Element

    /**@class Iterator
     */
    class Iterator : public UnionFindBase::BaseIterator
        {
    public:

        /**
         * Default constructor.
         */
        Iterator()
            {
            };

        /**
         * Initialization by BaseIterator.
         * BaseIterator it must be compatible with this Iterator.
         * That means that BaseElement referenced by it must be convertible
         * to UnionFind<Type>::Element.
         */
        Iterator
            (
            const BaseIterator & it
            ) : BaseIterator( it )
            {
            }

        /**
         * Returns reference to pointed item.
         */
        Type & operator*()
            {
            return static_cast<Element &>( get_reference() );
            }

        /**
         * Returns pointer to pointed item.
         */
        Type * operator->()
            {
            return &operator*();
            }

        /**
         * Assigns iterator.
         * BaseIterator must be compatible with this Iterator.
         */
        Iterator & operator=
            (
            const BaseIterator & it
            )
            {
            assign( it );
            return *this;
            }

        }; // Iterator

    }; // UnionFind


}; // namespace Massiv::Core
}; // namespace Massiv


#endif // _UNION_FIND_H_
