/** @file garbage_collector.cpp
 * Garbage collector basic implementation.
 */

/*
 * begin        : Wed Jun 1 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "global/global.h"
#include "utility/debug.h"
#include "object/object.h"
#include "object/object_property_data.h"
#include "pointer/garbage_collector.h"
#include "pointer/pointer_to_object.h"
#include "pointer/pointer_base.h"
#include "object/lib/lib_exceptions_helper.h"
#include "system/system.h"


namespace Massiv {
namespace Core {


namespace {

/** Clears queue. */
template <typename Queue>
void clear_queue
    (
    Queue & queue
    )
    {
    while( !queue.empty() )
        {
        queue.pop();
        }
    }

}    

// GarbageCollector::GCStatistics implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GarbageCollector::GCStatistics::GCStatistics() :
    Statistics( "ObjectManagement/GC", Status::FACILITY_OBJECT_MANAGEMENT, 
        Statistics::OPEN_NORMAL ),
    phase( create_integer( "phase", PHASE_IDLE ) ),
    num_objects( create_integer( "num_objects", 0 ) ),
    num_finalized( create_integer( "num_finalized", 0 ) ),
    num_explicitly_collected( create_integer( "num_explicitly_collected", 0 ) ),
    num_processed( create_integer( "num_processed", 0 ) ),
    consumed_time( create_float( "consumed_time", 0 ) ),
    force_run( create_boolean( "force_run", false ) ),
    object_count_limit( create_integer( "object_count_limit", 64 ) ),
    last_run( create_float( "last_run", 0.0f ) ),
    unpin_time( create_float( "unpin_time", 0.0f ) )
    {
    }


// GarbageCollector::GCSettings implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GarbageCollector::GCSettings::GCSettings() :
    Settings( "ObjectManagement/GC" ),
    can_run_any_time( access_boolean( "can_run_any_time", false ) ),
    max_iterations_per_tick( access_integer( "max_iterations_per_tick", 256 ) ),
    max_run_period( access_float( "max_run_period", 30.0f ) ),
    limit_delta( access_integer( "limit_delta", 50 ) ),
    max_limit_delta_fraction( access_float( "max_limit_delta_fraction", 2.0f ) ),
    thrashing_ratio( access_float( "thrashing_ratio", 0.8f ) ),
    thrashing_coef( access_float( "thrashing_coef", 0.2f ) ),
    lazy_ratio( access_float( "lazy_ratio", 0.5f ) ),
    lazy_coef( access_float( "lazy_coef", 1.2f ) ),
    max_unpin_time_delta( access_float( "max_unpin_time_delta", 2.0f ) )
    {
    }


// GarbageCollector implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GarbageCollector::GarbageCollector() :
    initialized( false ),
    migration_groups_frozen( false )
    {
    }


GarbageCollector::~GarbageCollector()
    {
    if( initialized )
        {
        shutdown();
        }
    }


void GarbageCollector::initialize()
    {
    hope( !initialized );

    statistics.phase = PHASE_IDLE;
    define_local_name( ObjectId(), "null" );
    
    initialized = true;
    }


void GarbageCollector::shutdown()
    {
    /* Switch to PHASE_SHUTTING_DOWN. */

    hope( statistics.phase != PHASE_SWEEP_LOCK &&
        statistics.phase != PHASE_SWEEP_FINALIZE );

    statistics.phase = PHASE_SHUTTING_DOWN;

    /* Do the same cleanup actions like if we unregistered all objects.
       This includes invalidating all ObjectPointers (unlinking ObjectProxies
       from registered objects) which is needed for safe object deletion. */
    
    root_objects.clear();
    update_backward_pointers();
    proxy_manager.invalidate_all_pointers();
    
    /* Delete all objects now. Since we are shutting down the objects can be
       deleted immediatelly regardless of their current state (pin, ...). */

    RegisteredObjects::iterator it = registered_objects.begin();

    while( it != registered_objects.end() )
        {
        ObjectProperty * object = it->second;

        if( object->is_gc_managed() )
            {
            /* It is okay to delete the object now. Note also that
               @c registered_objects can not change meanwhile so iterators
               will stay valid. */

            object->delete_object( Object::DELETE_SHUTTING_DOWN );
            }
        else
            {
            std::ostringstream ss;

            /* Unmanaged object can not be deleted by GC. */

            ss  << object->get_object_id() << " was left in the memory although "
                << "GC is shutting down. The object could not be deleted because it "
                   "is 'unmanaged'.";

            Global::log_warning
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_HIGHEST,
                ss.str()
                );
            }

        ++it;
        }

    /* Reset state. Leave @c proxy_manager and @c backward_pointers
       initialized because they are used by (potentially existing)
       ObjectPointer instances. These do not have to be reset. */

    registered_objects.clear();
    root_objects.clear();
    clear_queue( pointer_update_pending_objects );
    clear_queue( collect_pinned_objects );
    epoch = GCEpoch();
    clear_queue( explore_pending );
    names_to_objects.clear();
    objects_to_names.clear();
    object_to_migration_group.clear();
    migration_groups_frozen = false;

    initialized = false;
    }


ObjectProperty * GarbageCollector::find_local_object
    (
    const ObjectId &    object_id
    )
    {
    /* Search local objects. */

    RegisteredObjects::iterator it( registered_objects.find( object_id ) );

    if( it != registered_objects.end() )
        {
        return it->second;
        }
    else
        {
        return NULL;
        }
    }


void GarbageCollector::register_object
    (
    ObjectProperty * object
    )
    {
    bool insert_result;

    hope( object );
    hope( !object->is_member_object() );

    if( object->is_registered() )
        return;

    hope( statistics.phase != PHASE_SWEEP );
    hope( statistics.phase != PHASE_SWEEP_LOCK );
    hope( statistics.phase != PHASE_SHUTTING_DOWN );

    /* Register object. */

    RegisteredObjects::value_type value( object->get_object_id(), object );

    insert_result = registered_objects.insert( value ).second;

    hope( insert_result );

    object->root_data->gc_registered = true;

    /* Validate pointers. */

    proxy_manager.validate_pointers( object );

    /* Register permanent GC root object. */

    if( object->is_permanent_gc_root_object() )
        internal_make_gc_root( object, true );
    }


void GarbageCollector::unregister_object
    (
    ObjectProperty * object
    )
    {
    /**
     * Important note:
     * GC must be able to unregister any object _any_ time.
     * If it stores native pointers to Object (ObjectProperty *) in its
     * internal data structures then it must be able to safely
     * remove such items when particular object is being unregistered.
     * Pinned objects can not be unregistered.
     */

    hope( object );
    hope( !object->is_member_object() );

    if( !object->is_registered() )
        return;

    hope( statistics.phase != PHASE_SWEEP_LOCK );
    hope( statistics.phase != PHASE_SHUTTING_DOWN );
    hope( !migration_groups_frozen );

    #ifndef GC_MARK_PHASE_SPANNED

    hope( statistics.phase != PHASE_MARK );
    hope( explore_pending.empty() );

    #endif

    hope( !object->is_finalizing() );
    object->ensure_active_unpinned();
    hope( !object->is_pinned() );

    /* Unregister GC root object. */

    if( object->is_gc_root_object() )
        internal_make_gc_root( object, false );

    /* Update backward pointers. */

    update_backward_pointers();

    /* Invalidate pointers. */

    object->root_data->gc_registered = false;
    proxy_manager.invalidate_pointers( object );

    /* Unregister object. */

    RegisteredObjects::iterator it;

    it = registered_objects.find( object->get_object_id() );
    registered_objects.erase( it );
    }


void GarbageCollector::dispose_gc_root
    (
    ObjectProperty * object
    )
    {
    hope( object != NULL );
    hope( !object->is_member_object() );

    if( !object->is_permanent_gc_root_object() )
        return;

    if( !object->is_replica() )
        object->archive_now();
    
    object->root_data->gc_disposing_root = true;
    internal_make_gc_root( object, false );
    }


void GarbageCollector::dispose_gc_root
    (
    const CommonPointerBase & pointer
    )
    {
    ObjectProxy *       proxy = pointer.proxy;
    ObjectProperty *    object = proxy->get_object();

    #ifdef OPTION_DEBUG

    if( !object )
        {
        std::ostringstream s;

        s << "Can not dispose remote object " << proxy->get_object_id() << ".";
        RAISE_CORE_LIB_EXCEPTION( ObjectNotOnLocalSystemException, s.str() );
        }

    #endif

    hope( object );
    dispose_gc_root( object );
    }


void GarbageCollector::collect_object
    (
    ObjectProperty * object
    )
    {
    /**
     * Explicit (possibly client's) request to collect object.
     * The object is deleted immediatelly unless it is already
     * "being finalized" or it is an object replica.
     */

    hope( object != NULL );
    hope( !object->is_member_object() );
    
    hope( statistics.phase != PHASE_SHUTTING_DOWN );

    if( !object->is_registered() )
        return;

    if( !object->is_gc_managed() )
        return;

    if( object->is_finalizing() )
        return;

    if( object->is_replica() )
        return;
    
    if( object->is_pinned() )
        {
        /**
         * Object is pinned and can not be collected now.
         * Make it collect-pinned so that it could be collected later
         * (on next GC tick), if possible.
         */

        if( object->is_on_stack() )
            {
            RAISE_CORE_LIB_EXCEPTION
                (
                CanNotAccessObjectException,
                "Can not collect object on the stack."
                );
            }

        if( !object->is_collect_pinned() )
            {
            object->archive_now();
            object->root_data->gc_collect_pinned = true;
            collect_pinned_objects.push( object );
            }

        return;
        }

    /* Collect object. */

    hope( statistics.phase != PHASE_SWEEP_LOCK );

    object->delete_object( Object::GC_DELETE_INSTANCE );
    statistics.num_explicitly_collected++;
    }


void GarbageCollector::collect_object
    (
    const CommonPointerBase & pointer
    )
    {
    ObjectProxy *       proxy = pointer.proxy;
    ObjectProperty *    object = proxy->get_object();

    #ifdef OPTION_DEBUG

    if( !object )
        {
        std::ostringstream s;

        s << "Can not collect remote object " << proxy->get_object_id() << ".";
        RAISE_CORE_LIB_EXCEPTION( ObjectNotOnLocalSystemException, s.str() );
        }

    #endif

    hope( object );
    collect_object( object );
    }


void GarbageCollector::process_collect_pinned_objects()
    {
    int count = collect_pinned_objects.size();

    while( count-- > 0 )
        {
        ObjectProperty * object = collect_pinned_objects.front();

        /**
         * Since the object is present in the queue it is collect-pinned.
         * Thus @c object must still be valid and could not have been
         * unregistered from the GC meanwhile.
         */
        
        hope( !object->is_member_object() );
        hope( object->is_collect_pinned() );
        
        /**
         * The object will now be officialy removed from the queue
         * and collect-unpinned. This will allow to retry to
         * collect the object. However the object may still be pinned
         * and in that case it would become collect-pinned again.
         * Those objects will be processed on next call to
         * @c process_collect_pinned_objects().
         */
        
        collect_pinned_objects.pop();
        object->root_data->gc_collect_pinned = false;

        collect_object( object );
        }
    }


void GarbageCollector::unpin_active_pinned_objects()
    {
    /**
     * In order to be able to unpin all active-pinned objects we must
     * be sure that none of those objects is still active.
     * If @c System::inside_simulation() == false then no client
     * code can run and thus no object can be active.
     */

    if( System::inside_simulation() )
        {
        return;
        }

    /**
     * In order to prevent GC from deleting objects that are "active"
     * (there is a native reference to that objects) but unreferenced
     * by ObjectPointers any active object must be pinned.
     * Object is "activated" (becomes active) at ObjectPointer dereference
     * (more correctly when LocalProxy is created) and
     * "deactivated" when the dereferenced pointer is released
     * (when LocalProxy is deleted).
     * The object pinning is handled by ObjectProperty::active_pin()
     * and ObjectProperty::active_unpin() methods.
     * 
     * Since the pinning and unpinning may introduce quite significant
     * overhead, GC_DELAYED_UNPIN mode can be enabled.
     * In this mode object is pinned at its "activation" and unpinned
     * later, when it is a priori known that no object can be active.
     * The pinning is handled in this way:
     *  Object is active-pinned iff "current activation era" == era which
     *  the object was activated in
     *  @c active_pin() sets object "activation era" to current activation era
     *  @c active_unpin() is not required to be called
     *  @c unpin_active_pinned_objects() changes current activation era
     *  and thus automatically unpins all objects
     * The main drawback of this method is that objects are pinned
     * conservatively, for a longer time than it might be needed, thus
     * preventing objects from migrating, etc. This may also cause problems
     * when program waits for object migration (currently SRPC only),
     * this must be handled in a special way (see
     * @c ObjectProperty::ensure_active_unpinned() /
     * GC::ensure_object_active_unpinned()).
     */

    #ifdef GC_DELAYED_UNPIN

    /* Change current activation era and thus active-unpin all objects. */

    ObjectProperty::advance_era();

    #endif
    
    #ifdef OPTION_DEBUG
    
    RegisteredObjects::iterator it = registered_objects.begin();
    
    /* No object can be active if @c System::inside_simulation() is false. */

    while( it != registered_objects.end() )
        {
        ObjectProperty * object = it->second;
        
        if( object->is_active_pinned() )
            {
            panic
                (
                "System::inside_simulation() says that simulation code is "
                "not active but we have active-pinned objects!"
                );
            }
        
        ++it;
        }

    #endif

    /* Update statistics. */

    #ifdef GC_DELAYED_UNPIN

    SystemTime time = System::system_time();

    if( statistics.unpin_time + settings.max_unpin_time_delta < time.get_value() &&
        registered_objects.size() != 0 )
        {
        std::ostringstream ss;

        ss  << "Objects were active-pinned for "
            << time - SystemTime( statistics.unpin_time )
            << " second(s).";

        Global::log_warning
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_HIGHEST,
            ss.str()            
            );
        }

    statistics.unpin_time = static_cast<Float32>( time.get_value() );

    #endif
    }


void GarbageCollector::ensure_object_active_unpinned
    (
    const CommonPointerBase & pointer
    )
    {
    ObjectProxy *       proxy = pointer.proxy;
    ObjectProperty *    object = proxy->get_object();
    
    hope( object );
    object->ensure_active_unpinned();
    }


void GarbageCollector::unmanage_object
    (
    ObjectProperty * object
    )
    {
    hope( object );
    hope( !object->is_member_object() );
    hope( object->is_registered() );
    hope( !object->is_finalizing() );

    object->root_data->gc_managed = false;
    }


void GarbageCollector::make_gc_root
    (
    ObjectProperty *    object,
    bool                mode
    )
    {
    if( !mode && object->is_permanent_gc_root_object() )
        return;

    internal_make_gc_root( object, mode );
    }


void GarbageCollector::internal_make_gc_root
    (
    ObjectProperty *    object,
    bool                mode
    )
    {
    hope( object != NULL );
    hope( !object->is_member_object() );
    hope( object->is_registered() );

    /**
     * Dynamically changing "GC-rootness" of the object.
     * Equals to registering/unregistering object
     * to/from @c root_objects.
     */

    if( mode )
        {
        /* Object is becoming GC root. */

        if( !object->is_gc_root_object() && !object->is_gc_root_being_disposed() )
            {
            if( statistics.phase == PHASE_MARK )
                {
                mark_object( object );
                }

            if( !object->is_replica() )
                {
                object->archive_now();
                }

            root_objects.insert( object );
            object->root_data->gc_root = true;
            }
        }
    else
        {
        /* Object is becoming regular object. */

        if( object->is_gc_root_object() )
            {
            if( !object->is_replica() )
                {
                object->archive_now();
                }
            
            root_objects.erase( root_objects.find( object ) );
            object->root_data->gc_root = false;
            }
        }
    }


void GarbageCollector::enter_archive_mode
    (
    ObjectIdQueue & objects_to_archive
    )
    {
    /* Do not delete any object here! */

    hope( objects_to_archive.empty() );

    RegisteredObjects::iterator it;

    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        ObjectProperty * object = it->second;

        /* DO NOT push object replicas to @c objects_to_archive queue! */

        if( !object->is_replica() && object->is_archivable() )
            {
            object->enter_archive_mode();
            objects_to_archive.push( object->get_object_id() );
            }
        }
    }


bool GarbageCollector::is_object_upcalled
    (
    const ObjectProperty * object
    )
    {
    return UpcallToObject::is_object_referenced( object );
    }


ObjectProperty * GarbageCollector::who
    (
    bool skip_replicas
    )
    {
    return UpcallToObject::get_top_most_object( skip_replicas );
    }


std::string GarbageCollector::print_upcalled_objects()
    {
    return UpcallToObject::dump_list();
    }


void GarbageCollector::next_tick()
    {
    Statistics::MeasureSystemTime measure( statistics.consumed_time );

    /* Can run? */

    if( !settings.can_run_any_time && System::inside_simulation() )
        {
        // Prevent GC from running if objects can be active.
        // Note that active objects can not be deleted but are not treated
        // as GC roots.
        
        return;
        }

    /* Run GC tick. */
    
    unpin_active_pinned_objects();
    process_collect_pinned_objects();

    switch( statistics.phase )
        {
        case PHASE_IDLE:
            {
            /**
             * GC does not run (ie. object-pointer graph is not
             * being traversed).
             */

            if( statistics.num_objects >= statistics.object_count_limit ||
                statistics.last_run + settings.max_run_period <= 
                System::system_time().get_value() ||
                statistics.force_run.read() )
                {
                start_next_run();
                }

            break;
            }

        case PHASE_MARK:
            {
            /**
             * Mark phase can last one or more ticks.
             * At the last tick of mark phase sweep phase is
             * processed.
             */

            mark_phase();
            break;
            }

        case PHASE_SWEEP:
            {
            /**
             * Sweep phase must immediately follow mark phase.
             * Client code must not run inbetween them.
             * Transition from mark phase to sweep phase,
             * sweep phase processing and transition to idle phase
             * must take place in one GC tick.
             */

            panic( "Sweep phase lasts > 0 ticks." );
            }

        case PHASE_SWEEP_LOCK:
            {
            /**
             * Sweep lock phase is sweep phase subphase.
             * Sweep phase must be processed on behalf of the last
             * tick of mark phase.
             */

            panic( "Sweep lock phase lasts > 0 ticks." );
            }

        case PHASE_SWEEP_FINALIZE:
            {
            /**
             * Object's finalization invoked next GC tick.
             * Finalizer is already running.
             * Do nothing.
             */

            break;
            }

        case PHASE_SHUTTING_DOWN:
            {
            /**
             * GC::shutdown() invoked next GC tick.
             * This is forbidden.
             */

            panic( "GC::shutdown() interacts with the simulation. Forbidden." );
            }
        }

    /* Update statistics. */

    statistics.num_objects = registered_objects.size();
    }


void GarbageCollector::force_run()
    {
    if( is_running() )
        return;

    statistics.force_run = true;
    }


bool GarbageCollector::is_running() const
    {
    return statistics.phase != PHASE_IDLE;
    }


GarbageCollector::Phase GarbageCollector::get_phase() const
    {
    return static_cast<Phase>( statistics.phase.read() );
    }


const GarbageCollector::GCStatistics & GarbageCollector::get_statistics() const
    {
    return statistics;
    }


const GarbageCollector::GCSettings & GarbageCollector::get_settings() const
    {
    return settings;
    }


} // namespace Massiv::Core
} // namespace Massiv
