/** @file garbage_collector.h
 * Garbage collector.
 */

/*
 * begin        : Wed Jun 1 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_GARBAGE_COLLECTOR_H
#define MASSIV_GARBAGE_COLLECTOR_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#ifndef MASSIV_POINTER_PARAMETERS_H
#include "pointer/pointer_parameters.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_OBJECT_PROXY_H
#include "pointer/object_proxy.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#include <set>
#include <queue>
#include <string>


namespace Massiv {
namespace Core {


/** @class GarbageCollector
 * Garbage collector for local objects present on the local system.
 *
 * GC works in GC runs. At each run (that can be launched any time)
 * it determines what managed objects are reachable from local
 * GC root objects via strong pointers. Unreferenced objects,
 * objects referenced by weak pointers or unreachable objects
 * are classified as garbage. For complete list of supported
 * pointer types and their relationship with GC consult @ref pointer.
 * Each GC run consists of several phases. The two major ones are
 * "mark" phase, where GC walks through object-pointer graph, and
 * "sweep" phase, where GC deletes garbage-classified objects.
 * Notice that strong pointers can never point to remote objects
 * thus the whole GC mechanism can be reduced to local object/pointer
 * analysis.
 *
 * Notice that migration group (each maximal group of strong-connected
 * objects) without a GC root object is deleted on next GC run, because it
 * is unreachable.
 *
 * Object replicas are registered but not managed by GC.
 * ObjectPointer replicas are always treated as weak pointers.
 *
 * @ingroup pointer
 * @ingroup object_management
 */
class MASSIV_EXPORTED GarbageCollector
    {
    friend class Object;
    friend class ObjectProxy;
    friend class Balancer;
    friend class RPCObject;

public:

    //
    // Data structures.
    //

    typedef ObjectShared::ObjectIdQueue         ObjectIdQueue;
    typedef ObjectShared::BackwardPointers      BackwardPointers;
    typedef ObjectShared::BackwardPointer       BackwardPointer;
    typedef ObjectShared::BackwardPointersRange BackwardPointersRange;
    typedef UInt32                              MigrationGroupId;

    /** List of registered objects. */
    typedef ext::hash_map
                <
                ObjectId,
                ObjectProperty *,
                ObjectId::Hash,
                ObjectId::EqualKey
                >   RegisteredObjects;

    /** Queue of Objects. */
    typedef std::queue
                <
                ObjectProperty *
                >   ObjectQueue;

    /** List of registered GC root objects. */
    typedef ext::hash_set
                <
                ObjectProperty *,
                ObjectProperty::Hash
                >   RootObjects;

    /** Map translating local object name to an ObjectId. */
    typedef ext::hash_map
                <
                std::string,
                ObjectId
                >   NameToObjectIdMap;

    /** Map translating ObjectId to a local name. */
    typedef ext::hash_map
                <
                ObjectId,
                std::string,
                ObjectId::Hash,
                ObjectId::EqualKey
                >   ObjectIdToNameMap;

    /** Map translating Object to its MigrationGroupId. */
    typedef ext::hash_map
                <
                ObjectProperty *,
                MigrationGroupId,
                ObjectProperty::Hash
                >   ObjectToGroupId;

public:

    //
    // Statistics & settings.
    //

    enum Phase
        {
        PHASE_IDLE              =   0,
            /**< GC is not running. */
        PHASE_MARK              =   1,
            /**< GC is performing mark phase. */
        PHASE_SWEEP             =   2,
            /**< GC is performing sweep phase. */
        PHASE_SWEEP_LOCK        =   3,
            /**< Sweep phase subphase. GC data is locked during that phase. */
        PHASE_SWEEP_FINALIZE    =   4,
            /**< Sweep phase subphase. GC is finalizing garbage-classified objects. */
        PHASE_SHUTTING_DOWN     =   5
            /**< Shutting down. */
        }; // Phase

    /**
     * Garbage collector statistics.
     * Stored in "Statistics/ObjectManagement/GC" node.
     */
    class GCStatistics : public Statistics
        {
    public:

        /**
         * Constructor.
         */
        GCStatistics();

    public:

        //
        // Statistics attributes.
        //

        Integer &   phase;
            /**< Current GC phase. */
        Integer &   num_objects;
            /**< Number of registered objects. */
        Integer &   num_finalized;
            /**< Number of objects collected during last GC run. */
        Integer &   num_explicitly_collected;
            /**< Number of objects that were explicitly collected by GC::collect_object(). */
        Integer &   num_processed;
            /**< Number of objects processed during sweep phase. */
        Float &     consumed_time;
            /**< How much system time did GC consume? */
        Boolean &   force_run;
            /**< Force next GC run at the next tick. */
        Integer &   object_count_limit;
            /**< Start next GC run if object count reaches this limit. */
        Float &     last_run;
            /**< When the last GC run finished. */
        Float &     unpin_time;
            /**< When active-pinned objects were unpinned by
                 @c unpin_active_pinned_objects() for the last time. */

        }; // GCStatistics

    /**
     * Garbage collector settings.
     * Stored in "Settings/ObjectManagement/GC" node.
     */
    class GCSettings : public Settings
        {
    public:

        /**
         * Constructor.
         */
        GCSettings();

    public:

        //
        // GC settings.
        //
         
        Boolean &   can_run_any_time;
            /**< Set this to true to allow GC to run even when objects are
                 or might be active. Since active objects can not be deleted
                 and are not treated as GC roots this may cause that migration
                 groups will not be deleted atomically. */

        //
        // GC tuning.
        //

        Integer &   max_iterations_per_tick;
            /**< Maximum number of internal GC iterations within a GC tick. */
        Float &     max_run_period;
            /**< Maximum time difference between two GC runs. */
        Integer &   limit_delta;
            /**< Limit delta added to @c statistics.num_processed to get absolute
                 @c object_count_limit. */
        Float &     max_limit_delta_fraction;
            /**< Maximum limit delta relative to @c statistics.num_processed. */
        Float &     thrashing_ratio;
            /**< GC is in "thrashing mode" if @c statistics.num_finalized >
                 thrashing_ratio * statistics.num_processed. */
        Float &     thrashing_coef;
            /**< @c limit_delta adaptation coefficient for "thrashing" mode. */
        Float &     lazy_ratio;
            /**< GC is in "lazy mode" if @c statistics.num_finalized <
                 lazy_ratio * statistics.num_processed. */
        Float &     lazy_coef;
            /**< @c limit_delta adaptation coefficient for "lazy" mode. */
        Float &     max_unpin_time_delta;
            /**< Maximum time difference between two successfull calls
                 to @c unpin_active_pinned_objects(). If the limit is overflown
                 a warning is reported. */

        }; // GCSettings

public:

    /**
     * Constructor.
     * @c initialize() must be called to initialize the instance.
     */
    GarbageCollector();

    /**
     * Destructor.
     */
    ~GarbageCollector();

    /**
     * Initializes GC.
     * Must be called after GC has been successfully registered to Globals.
     */
    void initialize();

    /**
     * Shutdowns GC.
     * Resets GC instance state. @c initialize() can be called to reinitialize
     * the instance.
     */
    void shutdown();

public:

    //
    // Local object management.
    //

    /**
     * Returns pointer to object @c object_id.
     * Object won't be locked in the local memory,
     * so the pointer may become invalid soon.
     * To avoid dereferencing invalid pointers,
     * use ObjectPointer instead.
     */
    ObjectProperty * find_local_object
        (
        const ObjectId & object_id
        );

    /**
     * Registers object.
     * Object becomes accessible by its @c object_id.
     * When registering a migration group replica the group
     * must be refreshed then.
     */
    void register_object
        (
        ObjectProperty * object
        );

    /**
     * Unregisters object.
     * Object becomes inaccessible but remains in memory
     * and must be deleted manually.
     * It is guaranteed that GC will not delete @c object, however
     * strong-referenced objects by @c object may be deleted.
     * @note Before unregistering the object do not forget
     * to remove the object from global schedule (see ObjectManager).
     */
    void unregister_object
        (
        ObjectProperty * object
        );

    /**
     * Turns permanent GC root @c object to regular non-root object.
     * At a future GC run, the whole migration group of @c object will
     * be deleted atomically unless it contains an object that has been
     * activated by an access through a weak pointer (this problem can be
     * avoided by setting @c settings.can_run_any_time to false) or a pinned
     * object.
     * Once disposed there is no way to promote the object back to a GC root.
     */
    void dispose_gc_root
        (
        ObjectProperty * object
        );
    
    /**
     * Turns permanent GC root @c object to regular non-root object.
     * A variant that accepts any ObjectPointer.
     */
    void dispose_gc_root
        (
        const CommonPointerBase & pointer
        );

    /**
     * Proclaims object garbage.
     * If called on object replica, the request is ignored.
     * @note GC collects the object immediatelly if possible.
     */
    void collect_object
        (
        ObjectProperty * object
        );
    
    /**
     * Collects object referenced by an object pointer.
     * A variant that accepts any ObjectPointer.
     */
    void collect_object
        (
        const CommonPointerBase & pointer
        );

    /**
     * Prevents GC from managing @c object.
     * @c object remains in memory and is accessible by ObjectPointers.
     * @c object must be deleted manually by @c Object::delete_object().
     * @c object is not treated as a GC root.
     */
    void unmanage_object
        (
        ObjectProperty * object
        );

    /**
     * This method allows to dynamically mark/unmark object as a GC root.
     * Object is turned to GC root object if @c mode is true.
     *
     * Does not allow to turn permanent GC root objects to non GC roots.
     */
    void make_gc_root
        (
        ObjectProperty *    object,
        bool                mode
        );

    /**
     * Prepares objects for archivation.
     * All registered objects enter archive mode.
     * Returns list of objects that should be archived
     * (@c objects_to_archive).
     * @c objects_to_archive must be empty when called.
     */
    void enter_archive_mode
        (
        ObjectIdQueue & objects_to_archive
        );

    /**
     * Tests if @c object is upcalled from the core.
     * The test is reduced to simple UpcallToObject to @c object existance
     * test.
     * Use @c object = NULL to test if there is at least one non-NULL
     * UpcallToObject instance.
     */
    bool is_object_upcalled
        (
        const ObjectProperty * object
        );
    
    /**
     * Returns object that is currently running its (simulation) code or
     * what object was upcalled from the core for the last time.
     * This works as long as UpcallToObjects are used to upcall these objects
     * from the core and upcall instances are created/destroyed in LIFO order.
     * Whenever UpcallToObject is created, its instance is chained into a list
     * of active pointers which can be later traversed to determine this
     * kind of information. This technique simulates native C++ stack
     * traversal. Direct local calls are not detected.
     * Also direct local calls via ObjectPointers are not detected
     * (LocalProxies are not chained because of dereference effeciency).
     * If these conditions are not satisfied, the method will return
     * incorrect result.
     *
     * If @c skip_replicas is set to true then object replicas on the top
     * of the stack will be skipped.
     */
    ObjectProperty * who
        (
        bool skip_replicas = false
        );

    /**
     * Determines what objects are currently upcalled from the core.
     * The same notes/restrictions as to @c who() apply.
     */
    std::string print_upcalled_objects();
    
public:

    //
    // Object & pointer enumerators.
    //

    /**
     * Functor that performs operation on an object.
     */
    struct ObjectOperation
        {

        /**
         * Default constructor.
         */
        ObjectOperation();

        /**
         * Destructor.
         */
        virtual ~ObjectOperation();

        /**
         * Operation.
         * @note Do not manipulate with @a object in this function
         * as it can affect object enumeration!
         * Store @a object to some set for later processing.
         */
        virtual void operator()
            (
            ObjectProperty * object
            ) = 0;

        }; // ObjectOperation

    /**
     * Functor that performs operation on a pointer property.
     */
    struct PointerOperation
        {

        /**
         * Default constructor.
         */
        PointerOperation();

        /**
         * Destructor.
         */
        virtual ~PointerOperation();

        /**
         * Operation.
         */
        virtual void operator()
            (
            PointerInfo * info
            ) = 0;

        }; // PointerOperation

    /**
     * Direction of pointer enumeration.
     */
    enum PointerDirection
        {
        POINTER_FORWARD,        /**< Enumerate objects pointed to from current object. */
        POINTER_BACKWARD,       /**< Enumerate objects pointing to current object. */
        POINTER_BIDIRECTIONAL,  /**< Enumerate both forward and backward pointers. */
        }; // enum PointerDirection

    /**
     * Enumerates pointer properties owned by @c object that match given
     * <@c pointer_flags_yes, @c pointer_flags_no> pointer flags mask.
     */
    void enumerate_pointers
        (
        ObjectProperty *    object,
        PointerOperation &  operation,
        int                 pointer_flags_yes = 0,
        int                 pointer_flags_no = 0
        );

    /**
     * Enumerates pointer properties that reference @c object and match
     * given <@c pointer_flags_yes, @c pointer_flags_no> pointer flags mask.
     */
    void enumerate_backward_pointers
        (
        ObjectProperty *    object,
        PointerOperation &  operation,
        int                 pointer_flags_yes = 0,
        int                 pointer_flags_no = 0
        );

    /**
     * Enumerates objects referenced by @c object by pointer properties
     * matching <@c repflags_yes, @c repflags_no> replication flags mask.
     */
    void enumerate_referenced_objects
        (
        ObjectProperty *    object,
        int                 repflags_yes,
        int                 repflags_no,
        ObjectOperation &   operation
        );

    /**
     * Enumerates objects that reference @c object by pointer properties
     * matching <@c repflags_yes, @c repflags_no> replication flags mask.
     */
    void enumerate_backward_referenced_objects
        (
        ObjectProperty *    object,
        int                 repflags_yes,
        int                 repflags_no,
        ObjectOperation &   operation
        );

    /**
     * Enumerates replication group for <@c repflags_yes, @c repflags_no>.
     * Determines what objects are reachable from @c object
     * by pointer properties matching @c repflags_yes and
     * @c repflags_no.
     *
     * By default pointer properties are treated as if they were bidirectional.
     * If @c POINTER_FORWARD @a direction is used, pointers are treated
     * in standard way, and as backward if @c POINTER_BACKWARD  @a direction
     * is used.
     *
     * Only objects in consistent state will be enumerated.
     *
     * @pre @a object is in consistent state.
     */
    void enumerate_replication_group
        (
        ObjectProperty *    object,
        int                 repflags_yes,
        int                 repflags_no,
        ObjectOperation &   operation,
        PointerDirection    direction = POINTER_BIDIRECTIONAL
        );

    /**
     * Enumerates migration group for @c object.
     */
    void enumerate_migration_group
        (
        ObjectProperty *    object,
        ObjectOperation &   operation
        );

    /**
     * Enumerates all registered object replicas, including inconsistent replicas.
     */
    void enumerate_replicas
        (
        ObjectOperation &   operation
        );

    /**
     * Enumerates all registered objects, including both consistent and inconsistent replicas.
     */
    void enumerate_all_objects
        (
        ObjectOperation &   operation
        );

    /**
     * Freezes current migration groups state so that it could be
     * inspected by @c get_migration_group_id().
     * Until unfrozen (@c unfreeze_migration_groups) migration groups
     * can not change.
     */
    void freeze_migration_groups();

    /**
     * Unfreezes migration groups.
     */
    void unfreeze_migration_groups();

    /**
     * Returns migration group id of @c object.
     * The returned id is well-defined and consistent. Two objects
     * from the same migration group always have the same MigrationGroupId.
     * Objects from different groups always have different MigrationGroupIds.
     * MigrationGroupIds are generated in an ascending order starting with
     * zero.
     * In order to be able to call this migration groups must have been
     * frozen (see @c freeze_migration_groups()).
     */
    MigrationGroupId get_migration_group_id
        (
        ObjectProperty *    object
        );
    
    /**
     * Returns migration group id of @c object.
     * If @c object refers to a not yet explored migration group
     * (ie. a new migration group id will be returned), @c operation
     * will be invoked on all objects of the "new" migration group.
     */
    MigrationGroupId get_migration_group_id
        (
        ObjectProperty *    object,
        ObjectOperation &   operation
        );

public:

    //
    // Running GC.
    //

    /**
     * Runs next GC tick.
     */
    void next_tick();

    /**
     * Forces GC to start next GC run on next tick.
     * If GC is already running nothing is done.
     */
    void force_run();

    /**
     * Is GC running?
     */
    bool is_running() const;

    /**
     * Returns current GC phase.
     */
    Phase get_phase() const;

    /**
     * Returns GC statistics.
     */
    const GCStatistics & get_statistics() const;
    
    /**
     * Returns GC settings.
     */
    const GCSettings & get_settings() const;

public:

    //
    // Debug.
    //

    /**
     * Saves "snapshot" of current state of managed objects
     * present on the local system.
     * The state is saved in a graph format to be examined
     * by GraphViz utility.
     */
    void save_memory_snapshot
        (
        const std::string & file_name
        );

    /**
     * Dumps list of registered objects plus basic object information.
     * Only objects whose class matches @c pattern are printed.
     * If @c include_subclasses is set to true subclasses of the matched
     * classes are printed too.
     */
    void dump_registered_objects
        (
        std::ostream &      stream,
        const std::string & pattern,
        bool                include_subclasses
        );
    
    /**
     * Dumps replication group for @c object_id. The group is determined
     * by @c repflags_yes, @c repflags_no replication flags masks.
     * See @c enumerate_replication_group for more information on
     * replication groups.
     */
    void dump_replication_group
        (
        std::ostream &      stream,
        const ObjectId &    object_id,
        int                 repflags_yes,
        int                 repflags_no
        );
    
    /**
     * Dumps migration group for @c object_id.
     * See @c enumerate_migration_group for more information on
     * migration groups.
     */
    void dump_migration_group
        (
        std::ostream &      stream,
        const ObjectId &    object_id
        );

    /**
     * Dumps list of pinned objects together with reasons why the objects
     * were pinned.
     */
    void dump_pinned_objects
        (
        std::ostream &      stream
        );
    
    /**
     * Defines local name mapping for @c object_id.
     * If name collision occurs the mapping will not be defined.
     * Called by Object.
     */
    void define_local_name
        (
        const ObjectId &    object_id,
        const std::string & local_name
        );

    /**
     * Undefines local name mapping for @c object_id.
     * Called by Object.
     */
    void undefine_local_name
        (
        const ObjectId &    object_id
        );
    
    /**
     * Names @c object_id by @c name.
     * If @c name is empty the object name is cleared.
     */
    void name_object
        (
        const ObjectId &    object_id,
        const std::string & name
        );

    /**
     * Returns @c names_to_objects.
     */
    const NameToObjectIdMap & get_names_to_objects() const;

    /**
     * Returns @c objects_to_names.
     */
    const ObjectIdToNameMap & get_objects_to_names() const;

    /**
     * Has object already been marked during current GC epoch?
     * For object reachability debugging only.
     * @todo should be private
     */
    bool is_object_marked
        (
        ObjectProperty *    object
        );
    
private:

    //
    // Internal stuff.
    //
    
    /**
     * Internal version of make_gc_root().
     * Allows to turn permanent GC root objects to non GC roots, which is
     * required by @c unregister_object() and @c dispose_gc_root().
     */
    void internal_make_gc_root
        (
        ObjectProperty *    object,
        bool                mode
        );
    
    /**
     * Registers backward pointer that is "inverse" to
     * @c property -\> @c object_id pointer property. Fills @c bp.
     */
    void link_backward_pointer
        (
        const ObjectId &    object_id,
        Property *          property,
        BackwardPointer &   bp
        );

    /**
     * Unregisters backward pointer that was previously registered
     * by @c link_backward_pointer(). Invalidates @c bp.
     */
    void unlink_backward_pointer
        (
        BackwardPointer &   bp
        );

    /**
     * Relinks backward pointer @c bp.
     * Corresponding pointer property is assumed to be pointing
     * to @c object_id now. Changes @c bp.
     */
    void relink_backward_pointer
        (
        const ObjectId &    object_id,
        BackwardPointer &   bp
        );

    /**
     * Schedules backward pointer paired with @c property to be
     * updated later (by @c update_backward_pointer()).
     * Prior reading @c backward_pointers @c update_backward_pointers()
     * must be called.
     */
    void lazy_relink_backward_pointer
        (
        Property *          property
        );

    /**
     * Updates (relinks) backward pointer @c bp immediatelly.
     * Corresponding pointer property is assumed to be pointing to
     * @c object_id now. Changes @c bp.
     */
    void update_backward_pointer
        (
        const ObjectId &    object_id,
        BackwardPointer &   bp
        );

    /**
     * Updates all backward pointers that are scheduled to be updated.
     */
    void update_backward_pointers();

    /**
     * Updates backward pointers that are paired with pointer properties
     * owned by @c object.
     */
    void update_backward_pointers
        (
        ObjectProperty * object
        );

    /**
     * Tests if @c backward_pointers are up-to-date.
     */
    bool are_backward_pointers_up_to_date();

    /**
     * Processes objects from @c collect_pinned_objects
     * and collects them. Still pinned objects remain in the queue.
     */
    void process_collect_pinned_objects();

    /**
     * Unpins all active-pinned objects that have been active-pinned in the
     * current "activation era".
     */
    void unpin_active_pinned_objects();

    /**
     * Ensures that object referenced by @c pointer (any ObjectPointer) is
     * active-unpinned (not active).
     * If object is conservatively active-pinned it will be active-unpinned.
     * Asserts otherwise.
     * See @c ObjectProperty::ensure_active_unpinned().
     */
    void ensure_object_active_unpinned
        (
        const CommonPointerBase & pointer
        );

    /**
     * Marks object reachable from GC root objects.
     * If GC is not performing mark phase, method does nothing.
     * @c object may be NULL.
     */
    void mark_object
        (
        ObjectProperty * object
        );

    /**
     * Marks GC root objects and initiates object-pointer graph traversal.
     * Called at the beginning of mark phase.
     */
    void mark_root_objects();

    /**
     * Marks objects referenced from @c object through strong pointers.
     * Called by the mark algorithm (from within @c mark_phase()).
     */
    void mark_referenced_objects
        (
        ObjectProperty * object
        );

    /**
     * Examines what objects are referenced from GC roots
     * and marks those objects.
     */
    void mark_phase();

    /**
     * Determines what objects were not marked during mark phase
     * and proclaims those objects garbage.
     */
    void sweep_phase();

    /**
     * Schedules next GC run.
     */
    void schedule_next_run();

    /**
     * Starts next GC run.
     */
    void start_next_run();

private:

    //
    // State.
    //

    bool                initialized;
        /**< Is GC initialized? */
    bool                migration_groups_frozen;
        /**< Are migration groups frozen? */

    //
    // Object & pointer list.
    //

    RegisteredObjects   registered_objects;
        /**< List of all objects that are currently present
             on the local system. */
    RootObjects         root_objects;
        /**< List of registered GC root objects.
             Always subset of @c registered_objects. */
    BackwardPointers    backward_pointers;
        /**< List of "inversed" property pointers.
             When there is a pointer property @c A that reference object @c B,
             [ B, A ] is registered here. Each "inversed" pointer maintains
             a link to its corresponding (forward) pointer property (@c A).
             Under GC_LAZY_POINTERS mode not all entries have to be
             up-to-date, because system does "delayed" @c backward_pointers
             updates. @c backward_pointers entries always point
             to valid properties (@c A). */
    ObjectQueue         pointer_update_pending_objects;
        /**< List of objects whose pointer properties recently changed but
             their paired backward pointers have not been updated yet.
             Prior accessing @c backward_pointers @c update_backward_pointers()
             must be called. This ensures that @c backward_pointers will
             be up-to-date and @c pointer_update_pending_objects empty.
             The list is always empty if not in GC_LAZY_POINTERS mode. */
    ObjectQueue         collect_pinned_objects;
        /**< List of objects that were requested to be collected (by calling
             @c collect_object()) but still were not collected because they were
             pinned. GC ensures that these objects are not finalized nor
             unregistered. */
    ObjectToGroupId     object_to_migration_group;
        /**< Cache object -> its migration group. Valid if migration groups
             are frozen only. */

    //
    // Object reachability.
    //

    GCEpoch             epoch;
        /**< Current GC epoch (= number of GC run). */
    
    #ifdef GC_MARK_PHASE_SPANNED
    
    ObjectIdQueue       explore_pending;
        /**< Breadth-First search queue. */

    #else
    
    ObjectQueue         explore_pending;
        /**< Breadth-First search queue. Client code can not run during
             mark phase in this mode thus no objects can be unregistered
             from GC meanwhile -> ObjectQueue can be used instead. */

    #endif

    //
    // Object naming.
    //

    NameToObjectIdMap   names_to_objects;
        /**< List of local object names and their mapping to objects.
             Updated automatically when an object registers/unregisters
             to/from the GC.
             @note The database is managed by the local GC only,
             the update process does not cooperate with another nodes. */
    ObjectIdToNameMap   objects_to_names;
        /**< List of objects that have their names defined.
             Synchronized with @c names_to_objects. */

    //
    // Statistics & settings.
    //

    GCStatistics        statistics;
        /**< Statistics. */
    GCSettings          settings;
        /**< Settings. */

public:

    //
    // Object proxy manager.
    //

    ObjectProxyManager  proxy_manager;
        /**< ObjectProxy manager. Defined at the end of the class so that
             it would be deleted first when GC shuts down. */

    }; // GarbageCollector


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_GARBAGE_COLLECTOR_H
