/** @file garbage_collector_debug.cpp
 * Garbage collector: debugging.
 */

/*
 * begin        : Wed Dec 25 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "pointer/garbage_collector.h"
#include "pointer/pointer_base.h"
#include "object/object.h"
#include "object/object_property_data.h"
#include "object/class_manager.h"
#include "global/global.h"
#include "utility/quote_string.h"
#include <algorithm>
#include <iosfwd>
#include <fstream>
#include <iomanip>


namespace Massiv {
namespace Core {


//
// Object - pointer graph.
// GraphViz output.
//
    
namespace {

#define __endl  "\\n\\" << '\n'

void write_edge
    (
    std::ostream &      str,
    const ObjectId &    object_id,
    PointerInfo *       info
    );


/**
 * Dumps properties.
 */
struct GCDumpObjectProperties : public Property::PropertyOperation
    {

    /**
     * Constructor.
     */
    GCDumpObjectProperties
        (
        std::ostream &  stream
        ) : stream( stream )
        {
        }

    /**
     * Writes state of the supplied property to the output stream
     * if it is not a @c TYPE_POINTER.
     */
    virtual Result operator()
        (
        Property *  property
        )
        {
        Property::PropertyInfo  info;

        info = property->get_info
                        (
                        Property::QUERY_FULL_NAME_NO_ROOT |
                        Property::QUERY_TYPE |
                        Property::QUERY_STATE
                        );

        if( info.type == Property::TYPE_VALUE_TYPE ||
            info.type == Property::TYPE_CONTAINER )
            {
            stream  << info.property_name
                    << " = "
                    << info.state
                    << __endl;

            return RESULT_DONE;
            }
        else
        if( info.type == Property::TYPE_OBJECT_PROPERTY )
            {
            return RESULT_RECURSE;
            }

        return RESULT_DONE;
        }

    std::ostream &  stream;
        /**< Stream to write to. */

    }; // GCDumpObjectProperties


struct GCDumpOwnedPointers : public GarbageCollector::PointerOperation
    {

    /**
     * Constructor.
     */
    GCDumpOwnedPointers
        (
        std::ostream &      stream,
        const ObjectId &    object_id
        ) : stream( stream ),
            object_id( object_id )
        {
        }

    /**
     * Writes graph edge.
     */
    virtual void operator()
        (
        PointerInfo * info
        )
        {
        write_edge( stream, object_id, info );
        }

    std::ostream &  stream;
        /**< Stream to write to. */
    ObjectId        object_id;
        /**< Object being scanned. */

    }; // GCDumpOwnedPointers


void write_object_id
    (
    std::ostream &      str,
    const ObjectId &    object_id
    )
    {
    str << "\"" << object_id << "\"";
    }


void write_object
    (
    std::ostream &      str,
    ObjectProperty *    object
    )
    {
    GCDumpObjectProperties dump( str );

    str << object->get_object_id() << __endl;

    object->for_each_owned_property_do( dump );
    }


void write_edges
    (
    std::ostream &      str,
    const ObjectId &    object_id,
    ObjectProperty *    object,
    GarbageCollector &  gc
    )
    {
    GCDumpOwnedPointers dump( str, object_id );

    gc.enumerate_pointers( object, dump );
    }


void write_vertex
    (
    std::ostream &      str,
    const ObjectId &    object_id,
    ObjectProperty *    object,
    GarbageCollector &  gc
    )
    {
    write_object_id( str, object_id );
    str << " [";

    if( !object )
        {
        /* Remote object. */

        str << "shape = box, style = \"filled\", color = \"light blue\"";
        }
    else
        {
        std::string style = "filled";

        /* Local object. */

        if( !object->is_gc_root_object() )
            {
            /* Regular object. */

            std::string color;

            if( !gc.is_object_marked( object ) )
                {
                /* Not marked yet. */

                color = "light yellow";
                }
            else
                {
                /* Marked already. */

                color = "yellow";
                }

            str << "shape = \"box\", style = \"" << style << "\", color = \"" << color << "\"";
            }
        else
            {
            /* GC Root. */

            str << "shape = \"box\", style = \"" << style << "\", color = \"red\"";
            }

        str << ", label = \\" << '\n';

        str << "\"";
        write_object( str, object );
        str << "\"";
        }

    str << "]";
    str << '\n';
    }


void write_edge
    (
    std::ostream &      str,
    const ObjectId &    object_id,
    PointerInfo *       info
    )
    {
    const ObjectId & referenced_id = info->ptr_get_object_id();
    const Property * property = info->ptr_get_property();

    if( referenced_id.is_null() )
        {
        return;
        }

    write_object_id( str, object_id );
    str << " -> ";
    write_object_id( str, referenced_id );
    str << " [";

    str << "label = \"";
    str << property->get_info( Property::QUERY_FULL_NAME_NO_ROOT ).property_name;
    str << "\"";

    if( flags_on( info->ptr_get_pointer_flags(), WEAK_POINTER ) )
        {
        str << ", color = \"blue\"";
        }

    str << "]";
    str << '\n';
    }


void write_stack_edge
    (
    std::ostream &      str,
    const ObjectId &    object_id,
    bool                strong
    )
    {
    if( object_id.is_null() )
        {
        return;
        }

    str << "stack -> ";
    write_object_id( str, object_id );

    if( strong )
        {
        str << " [label = \"stack-strong\"]";
        }
    else
        {
        str << " [label = \"stack-weak\", color = \"blue\"]";
        }

    str << '\n';
    }


void write_pin_edge
    (
    std::ostream &      str,
    const ObjectId &    object_id
    )
    {
    hope( !object_id.is_null() );

    str << "pin -> ";
    write_object_id( str, object_id );
    str << " [label = \"pin\"]";
    str << '\n';
    }


} // anonymous namespace


void GarbageCollector::save_memory_snapshot
    (
    const std::string & file_name
    )
    {
    std::ofstream str( file_name.c_str() );

    /* Write header. */

    str  << "digraph objects\n"
         << "{\n";

    /* Write special graph vertices. */

    str  << "stack [label = \"program stack\", shape = \"box\", style = \"filled\", "
         << "color = \"green\"]\n";

    str  << "pin [label = \"pin\", shape = \"box\", style = \"filled\", "
         << "color = \"green\"]\n";

    /* Write objects on the heap including pointers they are linked with. */

    RegisteredObjects::iterator it;

    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        ObjectId            object_id;
        ObjectProperty *    object;
        int                 classification;

        /* Examine registered object. */

        object_id = it->first;
        object = it->second;

        /* Write graph vertex. */

        write_vertex( str, object_id, object, *this );

        /* Write outgoing edges. */

        write_edges( str, object_id, object, *this );

        /* Write special ingoing edges. */

        classification = proxy_manager.classify_object( object );

        if( flags_on( classification, ObjectProxy::CLASSIFY_STACK_STRONG_REFERENCED ) )
            {
            write_stack_edge( str, object_id, true );
            }

        if( flags_on( classification, ObjectProxy::CLASSIFY_STACK_WEAK_REFERENCED ) )
            {
            write_stack_edge( str, object_id, false );
            }

        if( object->is_pinned() )
            {
            write_pin_edge( str, object_id );
            }
        }

    /* Done. */

    str << "}" << std::endl;
    }


#undef __endl


//
// Registered objects dump.
//
    
namespace {

enum Parameters
    {
    OBJECT_ID_TEXT_LENGTH = 40,
    OBJECT_TAG_LENGTH =     4,
    MEMORY_SIZE_LENGTH =    8,
    MAX_NUM_OBJECTS =       256
    };


void dump_header
    (
    std::ostream &  stream
    )
    {
    stream.setf( std::ios_base::left );
    
    stream << std::setw( OBJECT_ID_TEXT_LENGTH );
    stream << "ObjectId";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "GC";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Rot";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Rep";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "SWk";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "SSg";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "PWk";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "PSg";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Pin";

    stream << std::setw( MEMORY_SIZE_LENGTH );
    stream << "Memory";

    stream << '\n';
    
    stream << std::setw( OBJECT_ID_TEXT_LENGTH );
    stream << "--------";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "--";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "---";

    stream << std::setw( MEMORY_SIZE_LENGTH );
    stream << "------";
    
    stream << '\n';

    stream.unsetf( std::ios_base::left );
    }


void dump_object
    (
    std::ostream &      stream,
    ObjectProperty *    object
    )
    {
    const ObjectId &    object_id = object->get_object_id();
    int                 classification;
    std::string         textual_object_id;

    stream.setf( std::ios_base::left );

    /* Write ObjectId. */
    
    stream << std::setw( OBJECT_ID_TEXT_LENGTH );
    stream << object_id.to_string().substr( 0, OBJECT_ID_TEXT_LENGTH - 1 );

    /* Write object basic info. */
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << object->is_permanent_gc_root_object();
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << object->is_gc_root_object();
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << object->is_replica();

    /* Write object references info. */
    
    classification = Global::object_proxy_manager().classify_object( object );

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << flags_on( classification, ObjectProxy::CLASSIFY_STACK_WEAK_REFERENCED );
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << flags_on( classification, ObjectProxy::CLASSIFY_STACK_STRONG_REFERENCED );
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << flags_on( classification, ObjectProxy::CLASSIFY_WEAK_REFERENCED );
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << flags_on( classification, ObjectProxy::CLASSIFY_STRONG_REFERENCED );
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << object->is_pinned();

    /* Write object memory size. */
    
    stream << std::setw( MEMORY_SIZE_LENGTH );
    stream << object->memory_size();

    stream.unsetf( std::ios_base::left );
    stream << '\n';
    }


void dump_footer
    (
    std::ostream &  stream,
    int             match_count,
    int             total_count
    )
    {
    stream << match_count << " of " << total_count << " object(s) dumped.";
    stream << std::endl;
    }


struct GCDumpObjects : public GarbageCollector::ObjectOperation
    {

    /**
     * Constructor.
     */
    GCDumpObjects
        (
        std::ostream & stream
        ) : stream( stream )
        {
        }

    /**
     * Dumps object.
     */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        dump_object( stream, object );
        }

    std::ostream & stream;
        /**< Stream to write to. */

    }; // GCDumpObjects


} // anonymous namespace    


void GarbageCollector::dump_registered_objects
    (
    std::ostream &      stream,
    const std::string & pattern,
    bool                include_subclasses
    )
    {
    RegisteredObjects::iterator     it;
    std::vector<std::string>        classes;
    UInt32                          match_count;

    /* Initialize. */

    Global::class_manager().find_classes( classes, pattern, include_subclasses );

    /* Write header. */

    dump_header( stream );

    /* Write object list. */

    match_count = 0;

    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        ObjectId            object_id;
        ObjectProperty *    object;
        const MetaObject *  metaobject;

        /* Examine registered object. */

        object_id = it->first;
        object = it->second;
        metaobject = object->get_metaobject();

        /* Should this object be printed? */

        if( !std::binary_search( classes.begin(), classes.end(), 
            metaobject->get_class_name() ) )
            {
            continue;
            }
        
        /* Dump object. */

        ++match_count;

        if( match_count < MAX_NUM_OBJECTS )
            {
            dump_object( stream, object );
            }
        else
            {
            break;
            }
        }

    /* Write footer. */

    dump_footer( stream, match_count, registered_objects.size() );
    }


void GarbageCollector::dump_replication_group
    (
    std::ostream &      stream,
    const ObjectId &    object_id,
    int                 repflags_yes,
    int                 repflags_no
    )
    {
    GCDumpObjects       operation( stream );
    ObjectProperty *    object = find_local_object( object_id );

    if( object )
        {
        dump_header( stream );
        enumerate_replication_group( object, repflags_yes, repflags_no, operation );
        }
    else
        {
        stream << "Object " << object_id << " not present on the local system.";
        stream << std::endl;
        }
    }


void GarbageCollector::dump_migration_group
    (
    std::ostream &      stream,
    const ObjectId &    object_id
    )
    {
    dump_replication_group
        (
        stream,
        object_id,
        REPMASK_MIGRATE_YES,
        REPMASK_MIGRATE_NO
        );
    }


void GarbageCollector::dump_pinned_objects
    (
    std::ostream &      stream
    )
    {
    RegisteredObjects::iterator it;
    int print_count;
    int total_count;

    print_count = 0;
    total_count = 0;
    
    /* Dump header. */
    
    stream.setf( std::ios_base::left );
    
    stream << std::setw( OBJECT_ID_TEXT_LENGTH );
    stream << "ObjectId";

    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Act";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Up";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Ptr";
    
    stream << std::setw( OBJECT_TAG_LENGTH );
    stream << "Col\n";

    /* Examine objects. */

    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        const ObjectId &    object_id = it->first;
        ObjectProperty *    object = it->second;

        /* Should dump? */

        if( !object->is_pinned() )
            {
            continue;
            }

        total_count++;    

        if( print_count > MAX_NUM_OBJECTS )
            {
            continue;
            }
        else
            {
            print_count++;
            }

        /* Dump. */
        
        stream << std::setw( OBJECT_ID_TEXT_LENGTH );
        stream << object_id.to_string().substr( 0, OBJECT_ID_TEXT_LENGTH - 1 );

        stream << std::setw( OBJECT_TAG_LENGTH );
        stream << ( object->is_active_pinned() ? 'A' : ' ' );
        
        stream << std::setw( OBJECT_TAG_LENGTH );
        stream << ( object->is_upcall_pinned() ? 'U' : ' ' );
        
        stream << std::setw( OBJECT_TAG_LENGTH );
        stream << ( object->is_ptr_pinned() ? 'P' : ' ' );
        
        stream << std::setw( OBJECT_TAG_LENGTH );
        stream << ( object->is_collect_pinned() ? 'C' : ' ' );

        stream << '\n';
        }

    /* Dump footer. */

    stream << print_count << " of " << total_count << " object(s) dumped.";
    }


//
// Naming.
//

namespace {

bool is_simple_string
    (
    const std::string & s
    )
    {
    std::ostringstream ss;
    ss << quote_const_string( s );
    return ss.str() == s;
    }

} // anonymous namespace
    
void GarbageCollector::define_local_name
    (
    const ObjectId &    object_id,
    const std::string & local_name
    )
    {
    bool success;
    
    if( !local_name.size() )
        {
        undefine_local_name( object_id );
        return;
        }

    if( !is_simple_string( local_name ) )
        {
        throw ExObjectManagementError( "Illegal object naming." );
        }
    
    undefine_local_name( object_id );

    success = names_to_objects.insert
        ( NameToObjectIdMap::value_type( local_name, object_id ) ).second;

    if( success )
        {
        objects_to_names.insert
            ( ObjectIdToNameMap::value_type( object_id, local_name ) );
        }
    else
        {
        std::ostringstream  ss;

        ss  << "Naming for " << object_id <<
            " was not set on the local node because the name conflicts.";

        Global::log
            (
            STATUS_WARNING
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_HIGH
                ),
            ss.str()
            );
        }
    }


void GarbageCollector::undefine_local_name
    (
    const ObjectId &    object_id
    )
    {
    NameToObjectIdMap::iterator name_it;
    ObjectIdToNameMap::iterator object_it;
    
    object_it = objects_to_names.find( object_id );

    if( object_it != objects_to_names.end() )
        {
        const std::string & previous_name = object_it->second;
        
        name_it = names_to_objects.find( previous_name );
        hope( name_it != names_to_objects.end() );

        names_to_objects.erase( name_it );
        objects_to_names.erase( object_it );
        }
    }


void GarbageCollector::name_object
    (
    const ObjectId &    object_id,
    const std::string & name
    )
    {
    ObjectProperty * object;

    /**
     * Object naming is implemented in this way:
     *  1) Each Object has its @c object_name
     *  2) Each node (GC) defines local name mapping. When an Object
     *     gets registered to that node its mapping is updated
     *     according to object's @c object_name. When an Object is
     *     unregistered the mapping for that object is undefined.
     *  3) If a name collision of local name mapping occurs
     *     the object that caused the collision will be left
     *     unmapped.
     */

    /* Change local name mapping. */

    define_local_name( object_id, name );
    
    /* Locate the object and change its name. */

    object = find_local_object( object_id );
    
    if( object )
        {
        object->root_data->object_name = name;
        }
    else
        {
        std::ostringstream  ss;

        ss  << "Naming for a remote object " << object_id
            << " was changed locally only.";

        Global::log
            (
            STATUS_WARNING
                (
                Status::FACILITY_OBJECT_MANAGEMENT,
                Status::PRIORITY_HIGH
                ),
            ss.str()
            );
        }
    }


const GarbageCollector::NameToObjectIdMap &
    GarbageCollector::get_names_to_objects() const
    {
    return names_to_objects;
    }


const GarbageCollector::ObjectIdToNameMap &
    GarbageCollector::get_objects_to_names() const
    {
    return objects_to_names;
    }


} // namespace Massiv::Core
} // namespace Massiv
