/** @file garbage_collector_enumerator.cpp
 * Garbage collector: pointer & object enumeration.
 */

/*
 * begin        : Sat Jan 4 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "pointer/garbage_collector.h"
#include "pointer/pointer_base.h"
#include "object/object_property_data.h"


namespace Massiv {
namespace Core {


#ifdef OPTION_DEBUG

/** @def ENUMERATION_LOG_LEVEL
 * Level of enumeration logging. Possible values:
 *  0   ... no logging at all
 *  1   ... migration group logging
 *  2   ... migration group and pointer enumeration logging
 */
//#define ENUMERATION_LOG_LEVEL   2

#endif


#ifndef ENUMERATION_LOG_LEVEL

#define ENUMERATION_LOG_LEVEL   0

#endif


namespace {


#if ENUMERATION_LOG_LEVEL > 0

/**
 * Helper class that allows for buffered logging (Global::log() can not
 * be used to log individual messages because of poor performance).
 *
 * Maintains a global @c attached_log that serves as an intermediate buffer
 * where log messages are written to. When done, the buffer is flushed
 * and its content sent to the global logger.
 *
 * To enable logging create an instance of BufferedLogger on the stack.
 * The logging will be enabled until the instance goes out of its scope.
 */
class BufferedLogger
    {
public:
    
    /** Attaches new log stream to the BufferedLogger. */
    BufferedLogger()
        {
        previous_log = attached_log;
        attached_log = &log;
        }

    /** Detaches current log stream from the BufferedLogger. */
    ~BufferedLogger()
        {
        Global::log_info
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_LOW,
            attached_log->str()
            );
        
        attached_log = previous_log;
        }

public:
    
    static std::ostringstream * attached_log;
        /**< Stream the messages are written to. May be NULL if no logging
             is enabled. */

private:

    std::ostringstream *        previous_log;
        /**< Previous attached log. */
    std::ostringstream          log;
        /**< Log buffer. */
    };


std::ostringstream * BufferedLogger::attached_log = NULL;

#else // ENUMERATION_LOG_LEVEL == 0

class BufferedLogger { };

#endif


} // anonymous namespace    


// Implementation of GarbageCollector::ObjectOperation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GarbageCollector::ObjectOperation::ObjectOperation()
    {
    }


GarbageCollector::ObjectOperation::~ObjectOperation()
    {
    }


// Implementation of GarbageCollector::PointerOperation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

GarbageCollector::PointerOperation::PointerOperation()
    {
    }


GarbageCollector::PointerOperation::~PointerOperation()
    {
    }


// Implementation of GarbageCollector
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * This property operation checks whether the enumerated property
 * is a pointer property matching specified pointer flags. If yes then
 * supplied pointer operation is invoked on the pointer property.
 */
struct EnumeratePointers : public Property::PropertyOperation
    {
    /** Constructor. */
    EnumeratePointers
        (
        GarbageCollector::PointerOperation &    operation,
        int                                     pointer_flags_yes,
        int                                     pointer_flags_no
        ) : operation( operation ),
            pointer_flags_yes( pointer_flags_yes ),
            pointer_flags_no( pointer_flags_no )
        {
        }

    /** Property operation -> pointer operation filter. */
    virtual Result operator()
        (
        Property * property
        )
        {
        PointerInfo * pointer_info = property->get_pointer_info();

        if( pointer_info != NULL )
            {
            int pointer_flags = pointer_info->ptr_get_pointer_flags();

            if( flags_on( pointer_flags, pointer_flags_yes, pointer_flags_no ) )
                {
                #if ENUMERATION_LOG_LEVEL > 1

                if( BufferedLogger::attached_log )
                    {
                    *BufferedLogger::attached_log
                        << "...enumerating pointer to "
                        << ( ( pointer_info->ptr_get_object() != NULL || 
                           pointer_info->ptr_get_object_id().is_null() ) ?
                           "local " : "remote " )
                        << pointer_info->ptr_get_object_id()
                        << " (" << property->get_info().property_name << ")\n";
                    }

                #endif

                operation( pointer_info );
                }
            else
                {
                #if ENUMERATION_LOG_LEVEL > 1

                if( BufferedLogger::attached_log )
                    {
                    *BufferedLogger::attached_log
                        << "...skipping pointer to "
                        << ( ( pointer_info->ptr_get_object() != NULL || 
                           pointer_info->ptr_get_object_id().is_null() ) ?
                           "local " : "remote " )
                        << pointer_info->ptr_get_object_id()
                        << " (" << property->get_info().property_name << ")\n";
                    }

                #endif
                }

            return RESULT_DONE;
            }

        return RESULT_RECURSE;
        }

    GarbageCollector::PointerOperation &    operation;
        /**< Pointer operation that is invoked on pointer properties. */
    int                                     pointer_flags_yes;
        /**< Pointer flags that must be set. */
    int                                     pointer_flags_no;
        /**< Pointer flags that must not be set. */
    }; // EnumeratePointers


/**
 * This pointer operation checks whether enumerated pointer property
 * matches given pointer replication masks (@c repflags_yes / repflags_no).
 * If yes then it calls object operation on the object referenced
 * by the pointer.
 */
struct EnumerateReferencedObjects : public GarbageCollector::PointerOperation
    {
    /** Constructor. */
    EnumerateReferencedObjects
        (
        int                                 repflags_yes,
        int                                 repflags_no,
        GarbageCollector::ObjectOperation & operation
        ) : repflags_yes( repflags_yes ),
            repflags_no( repflags_no ),
            operation( operation )
        {
        }

    /** Pointer operation -> object operation filter. */
    virtual void operator()
        (
        PointerInfo * info
        )
        {
        if( flags_on( info->ptr_get_replication_flags(), repflags_yes, repflags_no ) )
            {
            ObjectProperty * object = info->ptr_get_object();

            if( object )
                {
                hope( !object->is_on_stack() );
                
                operation( object );
                }
            }
        else
            {
            #if ENUMERATION_LOG_LEVEL > 0
            
            if( BufferedLogger::attached_log )
                {
                *BufferedLogger::attached_log
                    << "...pointer does not match requested ptr_repflags\n";
                }

            #endif
            }
        }

    int                                 repflags_yes;
        /**< Which replication flags must be set to true. */
    int                                 repflags_no;
        /**< Which replication flags must be set to false. */
    GarbageCollector::ObjectOperation & operation;
        /**< Object operation invoked on those objects that
             are referenced by pointer properties matching given
             replication flags mask. */
    }; // EnumerateReferencedObjects


/**
 * The backward enumerator differs from @c EnumerateReferencedObjects
 * in the way how it handles filtered pointer properties.
 * It enumerates pointer owner objects instead of pointer-referenced
 * objects.
 */
struct EnumerateBackwardReferencedObjects : public EnumerateReferencedObjects
    {
    /** Constructor. */
    EnumerateBackwardReferencedObjects
        (
        int                                 repflags_yes,
        int                                 repflags_no,
        GarbageCollector::ObjectOperation & operation
        ) : EnumerateReferencedObjects( repflags_yes, repflags_no, operation )
        {
        }

    /** Pointer operation -> object operation filter. */
    virtual void operator()
        (
        PointerInfo * info
        )
        {
        if( flags_on( info->ptr_get_replication_flags(), repflags_yes, repflags_no ) )
            {
            const Property * property = info->ptr_get_property();
            ObjectProperty * object = property->get_root_owner_object();

            hope( object );
            
            if( object->is_on_stack() )
                {
                // The pointer is owned by the stack root object.
                // Stack root object is never enumerated.

                return;
                }

            operation( object );
            }
        else
            {
            #if ENUMERATION_LOG_LEVEL > 0
            
            if( BufferedLogger::attached_log )
                {
                *BufferedLogger::attached_log
                    << "...backward pointer does not match requested ptr_repflags\n";
                }

            #endif
            }
        }

    }; // EnumerateBackwardReferencedObjects


/**
 * Context for @c GC::enumerate_replication_group() and
 * @c GC::enumerate_referenced_objects().
 *
 * Maintains queue of objects that need be explored.
 * In each turn @c pop() pops object from the queue and invokes external
 * @c operation on it (enumerates @c object).
 * Then @c object's neighborhood is pushed to the queue (the context is used
 * as an object operation for @c GC::enumerate_(backward)_referenced_objects()).
 *
 * Objects in inconsistent state will not be enumerated.
 */
struct EnumerateReplicationGroup : public GarbageCollector::ObjectOperation
    {
    /** Constructor. */
    EnumerateReplicationGroup
        (
        ObjectProperty *                    object,
        GarbageCollector::ObjectOperation & operation
        ) : operation( operation )
        {
        hope( object );
        hope( object->is_consistent() );
        current_time.advance();
        
        push( object );
        }

    /**
     * Pushes @c object to the queue unless @c object was already explored
     * by this @c EnumerateReplicationGroup instance.
     * This is used to push object's neighborhood objects to the queue
     * by calling @c GC::enumerate_referenced_objects() and/or
     * @c GC::enumerate_backward_referenced_objects().
     */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        hope( !object->is_member_object() );

        if( object->is_consistent() &&
            !object->root_data->gc_enumeration_time.is_equal_to( current_time ) )
            {
            #if ENUMERATION_LOG_LEVEL > 0

            if( BufferedLogger::attached_log )
                {    
                *BufferedLogger::attached_log << "...visiting " 
                    << object->get_object_id() << '\n';
                }

            #endif
            
            push( object );
            }
        else
            {
            #if ENUMERATION_LOG_LEVEL > 0

            if( BufferedLogger::attached_log )
                {    
                *BufferedLogger::attached_log << "...skipping visited " 
                    << object->get_object_id() << '\n';
                }

            #endif
            }
        }

    /** Pushes object to the queue. */
    void push
        (
        ObjectProperty * object
        )
        {
        object->root_data->gc_enumeration_time = current_time;
        active_objects.push( object );
        }

    /** Pops object from the queue. */
    ObjectProperty * pop()
        {
        ObjectProperty * object;

        if( active_objects.empty() )
            {
            object = NULL;
            }
        else
            {
            object = active_objects.front();
            active_objects.pop();
            
            #if ENUMERATION_LOG_LEVEL > 0

            if( BufferedLogger::attached_log )
                {
                *BufferedLogger::attached_log << "...enumerating " 
                    << object->get_object_id() << '\n';
                }

            #endif

            operation( object );
            }

        return object;
        }

    static VirtualTime<>                current_time;
        /**< Enumeration time. */
    GarbageCollector::ObjectQueue       active_objects;
        /**< Queue of objects that need be explored. */
    GarbageCollector::ObjectOperation & operation;
        /**< Object operation invoked on enumerated objects. */
    }; // EnumerateReplicationGroup


VirtualTime<> EnumerateReplicationGroup::current_time;


/**
 * PointerOperation that updates backward pointer that is coupled
 * with enumerated pointer property.
 */
struct UpdateBackwardPointers : public GarbageCollector::PointerOperation
    {

    /**
     * Updates backward pointer corresponding to pointer property
     * described by @c info.
     */
    virtual void operator()
        (
        PointerInfo * info
        )
        {
        info->update_backward_pointer();
        }

    }; // UpdateBackwardPointers


/**
 * Registers enumerated objects to migration group cache.
 * Assumes that enumerated objects belong to the same
 * migration group with @c group_id id.
 * Calls external @c operation on all enumerated objects.
 */
struct UpdateMigrationGroupCache : public GarbageCollector::ObjectOperation
    {
    /** Constructor. */
    UpdateMigrationGroupCache
        (
        GarbageCollector::ObjectToGroupId & cache,
        GarbageCollector::ObjectOperation & operation
        ) : cache( cache ),
            operation( operation )
        {
        }    

    /** Updates @c cache. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        cache.insert( GarbageCollector::ObjectToGroupId::value_type( object, group_id ) );
        operation( object );
        }

    static GarbageCollector::MigrationGroupId   group_id;
        /**< Id of the migration group being enumerated now. */
    GarbageCollector::ObjectToGroupId &         cache;
        /**< Cache. */
    GarbageCollector::ObjectOperation &         operation;
        /**< External operation called on all enumerated objects. */
    
    }; // UpdateMigrationGroupCache


GarbageCollector::MigrationGroupId UpdateMigrationGroupCache::group_id;


/** Object operation that does nothing. */
struct EmptyObjectOperation : public GarbageCollector::ObjectOperation
    {
    
    /** Empty operation. */
    virtual void operator()
        (
        ObjectProperty * object
        )
        {
        }
    
    }; // EmptyObjectOperation


void GarbageCollector::enumerate_pointers
    (
    ObjectProperty *    object,
    PointerOperation &  operation,
    int                 pointer_flags_yes,
    int                 pointer_flags_no
    )
    {
    EnumeratePointers enumerator
        ( operation, pointer_flags_yes, pointer_flags_no );

    // FIXME:
    //  Use metaobject to enumerate pointer properties
    //  (walk only through those properties that are either pointers or
    //   can own pointers).

    object->for_each_owned_property_do( enumerator );
    }


void GarbageCollector::enumerate_backward_pointers
    (
    ObjectProperty *    object,
    PointerOperation &  operation,
    int                 pointer_flags_yes,
    int                 pointer_flags_no
    )
    {
    const ObjectId &            object_id = object->get_object_id();
    BackwardPointersRange       range( backward_pointers.equal_range( object_id ) );
    BackwardPointers::iterator  it;

    hope( are_backward_pointers_up_to_date() );

    for( it = range.first; it != range.second; ++it )
        {
        Property *      property = it->second;
        PointerInfo *   pointer_info = property->get_pointer_info();
        int             pointer_flags = pointer_info->ptr_get_pointer_flags();
        
        hope( pointer_info->is_backward_pointer_up_to_date() );

        if( !flags_on( pointer_flags, pointer_flags_yes, pointer_flags_no ) )
            {
            #if ENUMERATION_LOG_LEVEL > 0

            if( BufferedLogger::attached_log )
                {
                *BufferedLogger::attached_log
                    << "...skipping backward pointer to local "
                    << property->get_root_owner_object()->get_object_id()
                    << " (" << property->get_info().property_name << ")\n";
                }

            #endif
            
            continue;
            }
        
        #if ENUMERATION_LOG_LEVEL > 0

        if( BufferedLogger::attached_log )
            {
            *BufferedLogger::attached_log
                << "...enumerating backward pointer to local "
                << property->get_root_owner_object()->get_object_id()
                << " (" << property->get_info().property_name << ")\n";
            }

        #endif

        operation( pointer_info );
        }
    }


void GarbageCollector::enumerate_referenced_objects
    (
    ObjectProperty *    object,
    int                 repflags_yes,
    int                 repflags_no,
    ObjectOperation &   operation
    )
    {
    EnumerateReferencedObjects enumerator
        (
        repflags_yes,
        repflags_no,
        operation
        );

    enumerate_pointers( object, enumerator );
    }


void GarbageCollector::enumerate_backward_referenced_objects
    (
    ObjectProperty *    object,
    int                 repflags_yes,
    int                 repflags_no,
    ObjectOperation &   operation
    )
    {
    EnumerateBackwardReferencedObjects enumerator
        (
        repflags_yes,
        repflags_no,
        operation
        );

    enumerate_backward_pointers( object, enumerator );
    }


void GarbageCollector::enumerate_replication_group
    (
    ObjectProperty *    object,
    int                 repflags_yes,
    int                 repflags_no,
    ObjectOperation &   operation,
    PointerDirection    direction
    )
    {
    hope( object );
    hope( object->is_consistent() );
    
    if( direction == POINTER_BACKWARD || direction == POINTER_BIDIRECTIONAL )
        {
        update_backward_pointers();
        }
    
    #if ENUMERATION_LOG_LEVEL > 0

    if( BufferedLogger::attached_log )
        {    
        *BufferedLogger::attached_log 
            << "Enumerating replication group of "
            << object->get_object_id() << " with [ repflags_yes = "
            << repflags_yes << ", repflags_no = " << repflags_no
            << ", direction = ";

        switch( direction )
            {
            case POINTER_BACKWARD:
                *BufferedLogger::attached_log << "POINTER_BACKWARD";
                break;
            
            case POINTER_FORWARD:
                *BufferedLogger::attached_log << "POINTER_FORWARD";
                break;
            
            case POINTER_BIDIRECTIONAL:
                *BufferedLogger::attached_log << "POINTER_BIDIRECTIONAL";
                break;

            default:    
                panic( "Unsupported pointer direction." );
            }
        
        *BufferedLogger::attached_log << " ].\n";
        }

    #endif
    
    EnumerateReplicationGroup enumerator( object, operation );

    while( ( object = enumerator.pop() ) != NULL )
        {
        /* Push @c object's neighborhood objects to the queue. */

        if( direction == POINTER_FORWARD || direction == POINTER_BIDIRECTIONAL )
            {
            enumerate_referenced_objects
                (
                object,
                repflags_yes,
                repflags_no,
                enumerator
                );
            }

        if( direction == POINTER_BACKWARD || direction == POINTER_BIDIRECTIONAL )
            {
            enumerate_backward_referenced_objects
                (
                object,
                repflags_yes,
                repflags_no,
                enumerator
                );
            }
        }
    }


void GarbageCollector::enumerate_migration_group
    (
    ObjectProperty *    object,
    ObjectOperation &   operation
    )
    {
    #if ENUMERATION_LOG_LEVEL > 0
    
    BufferedLogger log_this;
    
    #endif

    enumerate_replication_group
        (
        object,
        REPMASK_MIGRATE_YES,
        REPMASK_MIGRATE_NO,
        operation
        );
    }


void GarbageCollector::enumerate_replicas
    (
    ObjectOperation &   operation
    )
    {
    RegisteredObjects::iterator it;
    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        ObjectProperty * const object = it->second;
        hope( object );
        if( object->is_replica() )
            {
            operation( object );
            }
        }
    }

    
void GarbageCollector::enumerate_all_objects
    (
    ObjectOperation &   operation
    )
    {
    RegisteredObjects::iterator it;
    for( it = registered_objects.begin(); it != registered_objects.end(); ++it )
        {
        ObjectProperty * const object = it->second;
        hope( object );
        operation( object );
        }
    }


void GarbageCollector::freeze_migration_groups()
    {
    hope( !migration_groups_frozen );
    hope( object_to_migration_group.empty() );
    
    update_backward_pointers();
    hope( are_backward_pointers_up_to_date() );

    migration_groups_frozen = true;
    UpdateMigrationGroupCache::group_id = 0;
    }


void GarbageCollector::unfreeze_migration_groups()
    {
    hope( migration_groups_frozen );
    object_to_migration_group.clear();

    migration_groups_frozen = false;
    }


GarbageCollector::MigrationGroupId GarbageCollector::get_migration_group_id
    (
    ObjectProperty *    object
    )
    {
    EmptyObjectOperation operation;
    
    return get_migration_group_id( object, operation );
    }


GarbageCollector::MigrationGroupId GarbageCollector::get_migration_group_id
    (
    ObjectProperty *    object,
    ObjectOperation &   operation
    )
    {
    hope( migration_groups_frozen );

    /* Lookup cache first. */

    ObjectToGroupId::const_iterator it = object_to_migration_group.find( object );
    
    if( it != object_to_migration_group.end() )
        {
        return it->second;
        }

    /* Migration group of @c object is not present in the cache yet,
       enumerate it and update the cache. */

    UpdateMigrationGroupCache my_operation( object_to_migration_group, operation );
    
    enumerate_migration_group( object, my_operation );

    /* Return id of the enumerated migration group and generate new id for the
       next migration group (this is the id that will be returned next time). */

    return my_operation.group_id++;
    }


void GarbageCollector::link_backward_pointer
    (
    const ObjectId &    object_id,
    Property *          property,
    BackwardPointer &   bp
    )
    {
    hope( !migration_groups_frozen );
    bp = backward_pointers.insert( BackwardPointers::value_type( object_id, property ) );
    }


void GarbageCollector::unlink_backward_pointer
    (
    BackwardPointer &   bp
    )
    {
    hope( !migration_groups_frozen );
    backward_pointers.erase( bp );
    }


void GarbageCollector::relink_backward_pointer
    (
    const ObjectId &    object_id,
    BackwardPointer &   bp
    )
    {
    update_backward_pointer( object_id, bp );
    }


void GarbageCollector::lazy_relink_backward_pointer
    (
    Property *  property
    )
    {
    /**
     * Delay backward pointer update.
     * At later processing each pointer property, whose corresponding
     * backward pointer is not up-to-date, is examined and the backward
     * pointer is updated. The major trick is that backward pointers
     * are not updated unless it is really needed.
     * To simplify the task we store a list of objects
     * (@c pointer_update_pending_objects) that own / might own
     * those properties (@c property) instead of storing explicit
     * list of the properties. It is easier to detect that an object
     * is being unregistered from the local system than that a property
     * is being deleted.
     * @c update_backward_pointers() scans this list and does the update.
     */

    Property *                  property_owner = property->get_root_owner();
    ObjectProperty *            object = checked_cast<ObjectProperty *>( property_owner );
    ObjectProperty::RootData *  root_data = object->root_data;

    // FIXME: gc_pointer_update_pending might be moved to ObjectProperty
    //        for performance reasons.

    if( !root_data->gc_pointer_update_pending )
        {
        root_data->gc_pointer_update_pending = true;
        pointer_update_pending_objects.push( object );
        }
    }


void GarbageCollector::update_backward_pointer
    (
    const ObjectId &    object_id,
    BackwardPointer &   bp
    )
    {
    Property * property = bp->second;

    unlink_backward_pointer( bp );
    link_backward_pointer( object_id, property, bp );
    }


void GarbageCollector::update_backward_pointers()
    {
    ObjectProperty * object;
    
    if( are_backward_pointers_up_to_date() )
        {
        return;
        }
    
    #if ENUMERATION_LOG_LEVEL > 1

    if( BufferedLogger::attached_log )
        {
        *BufferedLogger::attached_log << "Updating backward pointers:\n";
        }
            
    #endif

    while( !pointer_update_pending_objects.empty() )
        {
        object = pointer_update_pending_objects.front();
        pointer_update_pending_objects.pop();
        
        /**
         * Conservatively update all backward pointers that
         * are coupled with property pointers owned by @c object.
         * Here we assume that @c object is still registered to the local
         * system. To ensure this, call @c update_backward_pointers()
         * whenever an object is going to be unregistered.
         */

        update_backward_pointers( object );
        }
    
    #ifdef OPTION_DEBUG
    
    /* Test if all backward pointers are really up-to-date now. */

    BackwardPointers::iterator  it = backward_pointers.begin();

    while( it != backward_pointers.end() )
        {
        Property *      property = it->second;
        PointerInfo *   pointer_info = property->get_pointer_info();
    
        hope( pointer_info->is_backward_pointer_up_to_date() );
        ++it;
        }

    #endif
    }


void GarbageCollector::update_backward_pointers
    (
    ObjectProperty * object
    )
    {
    UpdateBackwardPointers operation;
    
    if( object->is_on_stack() )
        {
        // We are not able to update backward pointers paired with
        // pointer properties owned by StackRootObject.
        // Backward pointers of that pointer properties are always
        // not up-to-date.
        // It is okay because we are not interested in pointer properties
        // owned by StackRootObject.

        return;
        }

    enumerate_pointers( object, operation );
    object->root_data->gc_pointer_update_pending = false;
    }


bool GarbageCollector::are_backward_pointers_up_to_date()
    {
    return pointer_update_pending_objects.empty();
    }


} // namespace Massiv::Core
} // namespace Massiv
