/** @file garbage_collector_mark_sweep.cpp
 * Garbage collector: mark & sweep.
 */

/*
 * begin        : Wed Dec 25 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "pointer/garbage_collector.h"
#include "pointer/object_proxy.h"
#include "pointer/pointer_base.h"
#include "object/object.h"
#include "object/object_property_data.h"
#include "utility/debug.h"
#include "system/system.h"
#include "global/global.h"
#include <sstream>


namespace Massiv {
namespace Core {

namespace {

/**
 * Pointer operation that marks objects referenced by the enumerated pointers.
 * Assumes that only strong pointers are enumerated.
 */
struct MarkStrongReferencedObjects : public GarbageCollector::PointerOperation
    {
    
    /** Marks referenced object. */
    virtual void operator()
        (
        PointerInfo * info
        )
        {
        info->mark_referenced_object();
        }

    };

} // anonymous namespace
    

bool GarbageCollector::is_object_marked
    (
    ObjectProperty * object
    )
    {
    hope( !object->is_member_object() );
    return !object->root_data->gc_epoch.is_older_than( epoch );
    }


void GarbageCollector::mark_object
    (
    ObjectProperty * object
    )
    {
    if( statistics.phase != PHASE_MARK )
        {
        return;
        }

    if( !object )
        {
        return;
        }

    if( is_object_marked( object ) )
        {
        return;
        }

    /* Object has not been marked at this epoch yet. Mark it now. */

    object->root_data->gc_epoch = epoch;
    
    /* Mark objects referenced by strong pointers owned by @c object
       at a future call to @c mark_phase() and @c mark_referenced_objects(). */
    
    if( !object->get_metaobject()->get_attribute_has_strong_pointer() )
        {
        return;
        }
    
    #ifdef GC_MARK_PHASE_SPANNED

    explore_pending.push( object->get_object_id() );

    #else
    
    explore_pending.push( object );
    
    #endif
    }


void GarbageCollector::mark_root_objects()
    {
    RootObjects::iterator   it;
    ObjectProperty *        object;

    for( it = root_objects.begin(); it != root_objects.end(); ++it )
        {
        object = *it;
        mark_object( object );
        }
    }


void GarbageCollector::mark_referenced_objects
    (
    ObjectProperty * object
    )
    {
    MarkStrongReferencedObjects operation;
    enumerate_pointers( object, operation, 0, WEAK_POINTER );
    }


void GarbageCollector::mark_phase()
    {
    ObjectProperty *    object;
    int                 num_objects;

    /* Walk through the [object, pointer] graph. */

    num_objects = 0;

    while( !explore_pending.empty() )
        {
        #ifdef GC_MARK_PHASE_SPANNED

        /* Limit number of iterations. */

        if( ++num_objects > settings.max_iterations_per_tick.read() )
            {
            break;
            }

        #endif

        /* Get object to explore references from. */

        #ifdef GC_MARK_PHASE_SPANNED
        
        ObjectId object_id;

        object_id = explore_pending.front();
        explore_pending.pop();

        object = find_local_object( object_id );

        if( !object )
            {
            /* Object was unregistered from GC meanwhile. */

            continue;
            }

        #else

        object = explore_pending.front();
        explore_pending.pop();

        #endif
        
        /* Explore references from a marked (visited) object. */
        
        mark_referenced_objects( object );
        }

    /* Mark phase finished? */

    if( explore_pending.empty() )
        {
        /*
         * This is the last tick of mark phase.
         * No delay between mark phase and sweep phase processing
         * is allowed. Process sweep phase immediately.
         */

        statistics.phase = PHASE_SWEEP;
        sweep_phase();

        return;
        }
    }


void GarbageCollector::sweep_phase()
    {
    RegisteredObjects::iterator rit;
    ObjectProperty *            object;
    ObjectQueue                 garbage;

    /*
     * Examine all registered objects and determine
     * what objects were not marked during mark phase.
     * Those objects are unreachable from GC root objects
     * and are inserted into @c garbage (unless they are pinned).
     * Sweep phase can not be spanned among multiple GC ticks.
     * Moreover it must be processed just after mark phase
     * finishes (ie. no pointer changes can occur inbetween
     * mark phase and sweep phase).
     */

    statistics.phase = PHASE_SWEEP_LOCK;

    for( rit = registered_objects.begin(); rit != registered_objects.end(); ++rit )
        {
        object = rit->second;

        if( !object->is_gc_managed() )
            {
            continue;
            }

        if( object->is_pinned() )
            {
            #ifdef OPTION_DEBUG

            if( !settings.can_run_any_time )    
                {
                object->ensure_active_unpinned();
                }

            #endif

            continue;
            }
        
        if( object->is_stack_referenced() )
            {
            continue;
            }

        if( !is_object_marked( object ) )
            {
            /*
             * We can not delete object immediatelly, because
             * object's finalize method or object destructor may
             * cause another objects to be deleted or created,
             * thus @c registered_objects may be modified
             * in an unpredictable way.
             */

            garbage.push( object );
            object->root_data->gc_finalizing = true;
            continue;
            }
        }

    /* Finalize objects. */
    
    std::string finalization_errors;
    int         error_count;

    statistics.phase = PHASE_SWEEP_FINALIZE;
    statistics.num_processed = registered_objects.size();
    statistics.num_finalized = garbage.size();

    error_count = 0;
    
    while( !garbage.empty() )
        {
        object = garbage.front();
        garbage.pop();

        /*
         * 1) Objects from @c garbage are marked "being finalized".
         * This ensures that these objects will only be deleted
         * by the finalizer.
         * 2) If @c garbage contains an object replica then
         * it also contains the corresponding migration group.
         * That is because object replicas can not be pinned and
         * if they are referenced from program stack or other objects
         * by proper pointers then the whole migration group is marked.
         */

        hope( object->is_finalizing() );
        hope( !object->is_gc_root_object() );
        hope( !object->is_stack_referenced() );

        /* Finalize object. */

        try
            {
            object->delete_object( Object::GC_DELETE_INSTANCE );
            }
        catch( Exception & e )
            {
            finalization_errors += '\n';
            finalization_errors += e.what();

            error_count++;
            }
        }

    statistics.num_objects = registered_objects.size();

    /* Tidy proxy objects. */

    proxy_manager.tidy_proxy_objects();

    /* Schedule next GC run. */

    statistics.phase = PHASE_IDLE;
    schedule_next_run();

    /* Propagate finalization errors. */

    if( error_count )
        {
        throw ExObjectFinalizationError
            (
            "The following errors were caught during object finalization:" + 
            finalization_errors
            );
        }
    }


void GarbageCollector::schedule_next_run()
    {
    std::ostringstream ss;

    /* Schedule next GC run. */

    if( statistics.num_finalized >
        statistics.num_processed * settings.thrashing_ratio.read() )
        {
        /* GC is in thrashing mode. Decrease @c limit_delta. */

        settings.limit_delta = 
            static_cast<Int32>( settings.limit_delta * settings.thrashing_coef );
        }
    else
    if( statistics.num_finalized <=
        statistics.num_processed * settings.lazy_ratio.read() )
        {
        /* GC is in lazy mode. Increase @c limit_delta. */

        settings.limit_delta = 
            static_cast<Int32>( ( settings.limit_delta + 1 ) * settings.lazy_coef );
        }
    
    if( settings.limit_delta > 
        settings.max_limit_delta_fraction * ( statistics.num_objects.read() + 1 ) )
        {
        /* Bound limit delta. */

        settings.limit_delta = 
            static_cast<Int32>( settings.max_limit_delta_fraction * 
            statistics.num_objects.read() );
        }
    
    if( settings.limit_delta == 0 )
        {
        /* Bound limit delta. */

        settings.limit_delta++;
        }

    statistics.object_count_limit = 
        statistics.num_objects + settings.limit_delta;

    statistics.last_run = 
        static_cast<Float32>( System::system_time().get_value() );

    /* Log GC state. */

    ss  << "Finished GC run.\n"
        << "Details:\n"
        << statistics;

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );
    }


void GarbageCollector::start_next_run()
    {
    /* Log GC state. */

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_OBJECT_MANAGEMENT,
            Status::PRIORITY_LOW
            ),
        "Starting new GC run."
        );

    /*
     * Reset statistics.
     * Start mark phase.
     */
    
    statistics.num_finalized = 0;
    statistics.num_explicitly_collected = 0;
    statistics.num_processed = 0;
    statistics.consumed_time = 0.0f;
    statistics.force_run = false;
    
    statistics.phase = PHASE_MARK;

    /*
     * Initiate next GC run.
     * This involves unmarking all objects marked at previous GC run
     * (done by advancing GC epoch) and marking those objects
     * that serve as GC root objects for next GC run
     * (GC root objects and stack-referenced objects).
     */

    epoch.advance();

    /* Mark GC root objects. */

    mark_root_objects();

    /* Mark stack-referenced objects. */

    proxy_manager.mark_stack_referenced_objects();

    /*
     * Finish mark phase (and thus the whole GC run) if
     * !GC_MARK_PHASE_SPANNED.
     */

    #ifndef GC_MARK_PHASE_SPANNED

    mark_phase();
    hope( statistics.phase == PHASE_IDLE );

    #endif
    }


} // namespace Massiv::Core
} // namespace Massiv
