/** @file object_pointer.h
 * Pointer to Object.
 */

/*
 * begin        : Sun Dec 15 2001
 * copyright    : (C) 2001, 2002, 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_POINTER_H
#define MASSIV_OBJECT_POINTER_H

#ifndef MASSIV_CONSTRAINTS_H
#include "global/constraints.h"
#endif

#ifndef MASSIV_POINTER_BASE_H
#include "pointer/pointer_base.h"
#endif

#ifndef MASSIV_DEFAULT_CAST_H
#include "pointer/default_cast.h"
#endif

#ifndef MASSIV_POINTER_WRAPPERS_H
#include "pointer/pointer_wrappers.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup pointer
 * @{
 */

/**@class ObjectPointer
 * @c ObjectPointer represents a safe pointer to a component
 * of @c Object with id @c object_id. @c Base parameter
 * describes pointer characteristics and its type.
 *
 * The implementation prevents client from dereferencing pointers
 * that point to non-existing objects (either because they were
 * destroyed or simply unregistered from @c ObjectManager) or
 * are illegal on that object. See implementation details for refinement!
 *
 * @c Base::PointerFlags parameter defines type of the pointer.
 * Its an ORed combination of @c PointerFlags
 * (defined in @c pointer_parameters.h). Property pointers
 * define replication flags that can be used to describe
 * migration and replication groups (groups of objects connected by
 * by property pointers of given type).
 * Not all flag combinations together with replication flags are allowed.
 * This is controlled by pointer policies.
 *
 * Objects managed by the system are garbage-collected.
 * GC detects what objects are reachable through strong pointers
 * from special objects called GC roots and program stack.
 * Unreachable objects are automatically deleted.
 * See @c GarbageCollector.
 *
 * @c Base::PointerCast specifies what component of object with
 * given object id is referenced by this pointer. PointerCast is a functor
 * object that casts general "ObjectProperty *" to "Base::PointerType *".
 * Use this to resolve @c Base::PointerType * ambiguity in multiple
 * inheritance hierarchy.
 * See @c CastObject for more information on use of @c Base::PointerCast
 * parameter.
 *
 * @c Base::PointerProxy defines semantics of operator-\>(). @c operator-\>()
 * never returns native pointer to the referenced object, active proxy object
 * is returned instead. Active proxy object either directly accesses
 * referenced object or translates method invocation requests to RPC calls.
 * The proxy is returned by value and thus is coupled with the corresponding
 * operator-\>() call.
 * See @c ActiveObjectProxy.
 *
 * ObjectPointer is implemented as a light-weight object that holds a 
 * reference to a proxy object that represents the "referenced" object.
 * Most of the operations that ObjectPointer defines are implemented
 * as direct calls to the proxy object or @c ObjectProxyManager when proxy
 * object needs be replaced by a new one (pointer value changes).
 * Since there may be more pointer "kinds" with different semantics,
 * most of the pointer functionality (the code that actually "does something")
 * is moved into @c Base class. ObjectPointer then serves as a wrapper
 * to @c Base with the same external interface.
 *
 * There are two major pointer kinds: "stack pointers" (pointers on
 * program stack, implemented by @c StackPointerBase; they are not properties)
 * and "pointer properties" (implemented by @c PropertyPointerBase, can also
 * be instantiated on the stack as any other property).
 * This separation to stack and property pointers is purely because of
 * implementation effeciency.
 */
template <class Base>
class ObjectPointer : public Base
    {
public:

    typedef Base                            Super;
    typedef typename Base::PointerType      PointerType;
    typedef typename Base::PointerCast      PointerCast;
    typedef typename Base::PointerProxy     PointerProxy;
    typedef typename Base::PointerTraits    PointerTraits;
    typedef ObjectPointer<Base>             ThisObjectPointer;

    /// Stack pointer with the same characteristics as this pointer.
    typedef ObjectPointer
                <
                StackPointerBase<PointerTraits>
                >   StackCounterpart;

    /// Property pointer with the same characteristics as this pointer.
    typedef ObjectPointer
                <
                PropertyPointerBase<PointerTraits>
                >   PropertyCounterpart;

public:
    
    using Base::get_pointer_flags;
    using Base::is_weak;

protected:
    
    using Base::default_constructed;
    using Base::manager;
    using Base::set_proxy;
    using Base::replace_proxy;
    using Base::before_read;
    using Base::before_write;
    using Base::proxy;
    using Base::cast;
    using Base::check_strong_pointer_object_id;

public:

    //
    // Constructors & destructors.
    //

    /**
     * Macro that is used by the constructors to initialize ObjectPointer.
     * If pointer initialization fails, the pointer is reinitialized
     * to point to @c NULL_ID and the exception is re-thrown.
     */

    #define INIT( initialization ) \
        try \
            { \
            initialization; \
            } \
        catch( ... ) \
            { \
            set_proxy( manager().get_null_proxy( get_pointer_flags() ) ); \
            default_constructed(); \
            throw; \
            }

    #ifdef OPTION_DEBUG

    #define INIT_MUST_SUCCEED( initialization ) \
        { \
        try \
            { \
            initialization; \
            } \
        catch( ... ) \
            { \
            panic( "Unexpected exception in ObjectPointer's constructor." ); \
            } \
        }

    #else

    #define INIT_MUST_SUCCEED( initialization ) \
        initialization;

    #endif

    /**
     * Default constructor.
     * Initializes pointer to point to @c NULL_ID.
     */
    ObjectPointer() : Super()
        {
        set_proxy( manager().get_null_proxy( get_pointer_flags() ) );
        default_constructed();
        }

    /**
     * Constructor.
     * Initializes pointer to point to @c object_id.
     */
    ObjectPointer
        (
        const ObjectId & object_id
        ) : Super()
        {
        INIT( set_proxy( manager().get_proxy( object_id, cast, get_pointer_flags() ) ) );
        }

    /**
     * Constructor.
     * Initializes pointer to point to the same object as @c ptr.
     * Performs implicit pointer conversion (consists of pointer upcast
     * and pointer flags change ("pointer mutation")).
     */
    template <class OtherBase>
    ObjectPointer
        (
        const ObjectPointer<OtherBase> & ptr
        ) : Super()
        {
        // Check cast (compile-time).

        typedef typename OtherBase::PointerType OtherPointerType;
        check_default_cast<PointerType, OtherPointerType>();

        // Assign.

        INIT( set_proxy( upcast_from( &ptr.convert(), ptr.get_cast(), 
            ptr.get_pointer_flags() ) ) );
        }

    /**
     * Copy constructor.
     * Calls pointer base's copy constructor.
     * This ensures that pointer properties are correctly copied and
     * initialized (Property class).
     */
    ObjectPointer
        (
        const ThisObjectPointer & ptr
        ) : Super( ptr )
        {
        ptr.before_read();
        INIT_MUST_SUCCEED( set_proxy( manager().get_duplicated_proxy(
            ptr.proxy, get_pointer_flags() ) ) );
        }

    /**
     * Constructor.
     * Initializes pointer from @c native_pointer.
     */
    ObjectPointer
        (
        const PointerType * native_pointer
        ) : Super()
        {
        #ifndef OPTION_DEBUG
        
        INIT_MUST_SUCCEED( set_proxy( manager().get_proxy_from_native_pointer(
            native_pointer, cast, get_pointer_flags() ) ) );

        #else

        INIT( set_proxy( manager().get_proxy_from_native_pointer( native_pointer,
            cast, get_pointer_flags() ) ) );

        #endif
        }

    /**
     * Constructor.
     * Initializes pointer to point to the object referenced by @c other_proxy.
     * Performs explit pointer conversion.
     */
    ObjectPointer
        (
        ObjectProxy & other_proxy
        ) : Super()
        {
        INIT( set_proxy( manager().get_converted_proxy( &other_proxy, cast, 
            get_pointer_flags() ) ) );
        }

    /**
     * Constructor.
     * Initializes pointer to point to @c NULL_ID.
     */
    ObjectPointer
        (
        const NullKeyword & null
        ) : Super()
        {
        set_proxy( manager().get_null_proxy( get_pointer_flags() ) );
        }

    #undef INIT
    #undef INIT_MUST_SUCCEED

    /**
     * Destructor.
     */
    ~ObjectPointer()
        {
        }

public:

    //
    // Operator=.
    //

    /** Operator=. */
    ThisObjectPointer & operator=
        (
        const ObjectId & object_id
        )
        {
        before_write();
        replace_proxy( manager().get_proxy( object_id, cast, 
            get_pointer_flags() ) );

        return *this;
        }

    /** Operator=. */
    template <class OtherBase>
    ThisObjectPointer & operator=
        (
        const ObjectPointer<OtherBase> & ptr
        )
        {
        // Check cast (compile-time).

        typedef typename OtherBase::PointerType OtherPointerType;
        check_default_cast<PointerType, OtherPointerType>();

        // Assign.

        before_write();
        replace_proxy( upcast_from( &ptr.convert(), ptr.get_cast(), 
            ptr.get_pointer_flags() ) );

        return *this;
        }

    /** Operator=. */
    ThisObjectPointer & operator=
        (
        const ThisObjectPointer & ptr
        )
        {
        before_write();
        ptr.before_read();
        replace_proxy( manager().get_duplicated_proxy( ptr.proxy, 
            get_pointer_flags() ) );

        return *this;
        }

    /** Operator=. */
    ThisObjectPointer & operator=
        (
        const PointerType * native_pointer
        )
        {
        before_write();
        replace_proxy( manager().get_proxy_from_native_pointer
            ( native_pointer, cast, get_pointer_flags() ) );

        return *this;
        }

    /** Operator=. */
    ThisObjectPointer & operator=
        (
        ObjectProxy & other_proxy
        )
        {
        before_write();
        replace_proxy( manager().get_converted_proxy( &other_proxy, cast, 
            get_pointer_flags() ) );

        return *this;
        }

    /** Operator=. */
    ThisObjectPointer & operator=
        (
        const NullKeyword & null
        )
        {
        set_null();
        return *this;
        }

public:

    //
    // Comparing pointers.
    //

    // Comparing with native pointers.

    /** Compares pointers. */
    friend bool operator==
        (
        const PointerType *         native_pointer,
        const ThisObjectPointer &   ptr
        )
        {
        if( native_pointer != NULL )
            {
            if( !ptr.is_local() )
                {
                return false;
                }
            else
                {
                return variant_cast<PointerType *>
                    ( ptr.proxy->get_component() ) == native_pointer;
                }
            }
        
        return ptr.is_null();
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const PointerType *         native_pointer,
        const ThisObjectPointer &   ptr
        )
        {
        return !( native_pointer == ptr );
        }

    /** Compares pointers. */
    friend bool operator==
        (
        const ThisObjectPointer &   ptr,
        const PointerType *         native_pointer
        )
        {
        return ( native_pointer == ptr );
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const ThisObjectPointer &   ptr,
        const PointerType *         native_pointer
        )
        {
        return !( ptr == native_pointer );
        }

    // Comparing with stack counterparts.

    /** Compares pointers. */
    friend bool operator==
        (
        const ThisObjectPointer &   ptr1,
        const StackCounterpart &    ptr2
        )
        {
        return compare( &ptr1.convert(), &ptr2.convert() );
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const ThisObjectPointer &   ptr1,
        const StackCounterpart &    ptr2
        )
        {
        return !( ptr1 == ptr2 );
        }

    // Comparing with property counterparts.

    /** Compares pointers. */
    friend bool operator==
        (
        const ThisObjectPointer &   ptr1,
        const PropertyCounterpart & ptr2
        )
        {
        return compare( &ptr1.convert(), &ptr2.convert() );
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const ThisObjectPointer &   ptr1,
        const PropertyCounterpart & ptr2
        )
        {
        return !( ptr1 == ptr2 );
        }

    // Comparing with general ObjectPointer of any type
    // and characteristics, actually object ids are compared.

    /**
     * Compares this pointer with general @c ptr.
     * Returns true if the two pointers reference the same object
     * (object component, pointer type & characteristics are ignored).
     */
    template <class OtherBase>
    bool operator==
        (
        const ObjectPointer<OtherBase> & ptr
        )
        {
        return get_object_id() == ptr.get_object_id();
        }
    
    /** Compares this pointer with general @c ptr. */
    template <class OtherBase>
    bool operator!=
        (
        const ObjectPointer<OtherBase> & ptr
        )
        {
        return !operator==( ptr );
        }

    // Comparing with @c null.

    /** Compares pointer with @c null. */
    friend bool operator==
        (
        const ThisObjectPointer &   ptr,
        const NullKeyword &         null
        )
        {
        return ptr.is_null();
        }

    /** Compares pointer with @c null. */
    friend bool operator==
        (
        const NullKeyword &         null,
        const ThisObjectPointer &   ptr
        )
        {
        return ptr.is_null();
        }

    /** Compares pointer with @c null. */
    friend bool operator!=
        (
        const ThisObjectPointer &   ptr,
        const NullKeyword &         null
        )
        {
        return !ptr.is_null();
        }

    /** Compares pointer with @c null. */
    friend bool operator!=
        (
        const NullKeyword &         null,
        const ThisObjectPointer &   ptr
        )
        {
        return !ptr.is_null();
        }

public:

    //
    // Dereferencing.
    // Returning active proxy objects.
    //

    /** Dereference. */
    PointerProxy operator->()
        {
        before_read();
        return PointerProxy( proxy );
        }

    /** Dereference. */
    PointerProxy operator->() const
        {
        before_read();
        return PointerProxy( proxy );
        }

    /** Returns active object proxy. */
    PointerProxy dereference()
        {
        before_read();
        return PointerProxy( proxy );
        }

    /** Returns active object proxy. */
    PointerProxy dereference() const
        {
        before_read();
        return PointerProxy( proxy );
        }

public:

    //
    // State.
    //

    /** Is pointer pointing to @c NULL_ID? */
    bool is_null() const
        {
        before_read();
        return proxy->is_null();
        }

    /** Set pointer to @c NULL_ID. */
    void set_null()
        {
        before_write();
        replace_proxy( manager().get_null_proxy( get_pointer_flags() ) );
        }

    /**
     * Does pointer point to an unambiguous component of a local object?
     * Can pointer be dereferenced locally (without an error)?
     */
    bool is_local() const
        {
        before_read();
        return proxy->get_object() != NULL && !proxy->is_ambiguous();
        }

    /** Returns object id of the referenced object. */
    const ObjectId & get_object_id() const
        {
        before_read();
        return proxy->get_object_id();
        }

    /**
     * Returns object proxy.
     * Assigning the proxy to another ObjectPointer invokes
     * the conversion.
     * Conversion validity is checked at run-time.
     * @code
     * other_pointer = pointer.convert();
     * @endcode
     */
    ObjectProxy & convert() const
        {
        before_read();
        return *proxy;
        }

    /**
     * Returns cast object that characterizes this pointer.
     * Defined mainly for internal purposes where ObjectPointer
     * needs to access cast object of an ObjectPointer of different
     * type.
     * Since cast object is a static property of a pointer @c before_read()
     * should not be called.
     */
    const PointerCast & get_cast() const
        {
        return cast;
        }

public:

    //
    // RTTI.
    //

    /**
     * Does referenced object inherit class described by @c class_type_info?
     * Throws UndefinedClassTypeException if object's dynamic type is unknown.
     */
    bool inherits
        (
        const std::type_info & class_type_info
        ) const
        {
        return proxy->inherits( class_type_info );
        }

    /** Returns type info of requested object component. */
    const std::type_info & get_type_info() const
        {
        return typeid( PointerType );
        }

    /**
     * Returns type info of referenced object.
     * Throws UndefinedClassTypeException if object's dynamic type is unknown.
     */
    const std::type_info & get_object_type_info() const
        {
        return proxy->get_object_type_info();
        }

public:

    //
    // Serialization.
    //

    // Stoupik's note:
    // These methods will be virtual if Base is a Property and non-virtual otherwise.

    /** Writes pointer state. */
    INLINE void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const
        {
        get_object_id().bitstr_write( bos, desc );
        }

    /** Restores pointer state. */
    INLINE void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        )
        {
        ObjectId id;
        id.bitstr_read( bis, desc );
        if( !is_weak() )
            {
            check_strong_pointer_object_id( desc, id );
            }
        operator=( id );
        }

    /** Writes pointer state. */
    INLINE void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const
        {
        get_object_id().text_write( tw, desc );
        }

    /** Restores pointer state. */
    INLINE void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        )
        {
        ObjectId id;
        id.text_read( tr, desc );
        operator=( id );
        }

private:

    //
    // Private stuff.
    //

    /**
     * Returns linked proxy that is a result of pointer upcast conversion from
     * @c other_proxy (with @c OtherPointerCast type and @c other_flags pointer
     * flags) to this pointer type. Allows for pointer flags change ("pointer
     * mutation").
     */
    template <class OtherPointerCast>
    ObjectProxy * upcast_from
        (
        ObjectProxy *               other_proxy,
        const OtherPointerCast &    other_cast,
        int                         other_flags
        )
        {
        return manager().get_upcasted_proxy( other_proxy, cast, other_flags, 
            get_pointer_flags() );
        }

    /**
     * Returns linked proxy that is a result of pointer upcast conversion from
     * @c other_proxy (with @c PointerCast type and @c other_flags pointer flags)
     * to this pointer type. Allows for pointer flags change ("pointer mutation").
     * Special version where the two pointers are characterized by the
     * same cast object and no conversion is needed.
     *
     * @note
     *      This used to be "explicit specialization" of generic @c upcast_from.
     *      According to C++ standard, member template specializations must be
     *      defined in namespace scope, ie. out of class. The declaration would be
     *      <tt> template< Base > template<> ObjectPointer< Base >::upcast_from... </tt>
     *      but such declaration is not allowed (enclosing templates must be explicitly
     *      specialized too). So I changed it to non-template method and I really hope
     *      that all compilers will choose this version when pointer cast object is
     *      PointerCast.
     */
    ObjectProxy * upcast_from
        (
        ObjectProxy *       other_proxy,
        const PointerCast & other_cast,
        int                 other_flags
        )
        {
        return manager().get_proxy_with_different_flags( other_proxy,
            other_flags, get_pointer_flags() );
        }

    /**
     * Compares two pointers that are linked to @c proxy1 and proxy2.
     * Both the pointers are assumed to have the same type & cast
     * object (pointer flags may be different).
     * @note before_read() must have already been called on both
     * the pointers since this method accesses pointer bases directly.
     */
    static bool compare
        (
        const ObjectProxy * proxy1,
        const ObjectProxy * proxy2
        )
        {
        #ifdef OPTION_DEBUG

        const ObjectId & id1 = proxy1->get_object_id();
        const ObjectId & id2 = proxy2->get_object_id();

        // Because the two pointers have the same characteristics,
        // they must share the same proxy object if they are linked
        // to the same object_id.

        if( id1 == id2 )
            {
            hope( proxy1 == proxy2 );
            }
        else
            {
            hope( proxy1 != proxy2 );
            }

        #endif

        return proxy1 == proxy2;
        }

    }; // ObjectPointer

/** @} */ // addtogroup pointer


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_OBJECT_POINTER_H
