/** @file object_proxy.cpp
 * Object proxy implementation.
 */

/*
 * begin        : Wed Dec 25 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "pointer/object_proxy.h"
#include "object/object.h"
#include "pointer/garbage_collector.h"
#include "pointer/pointer_base.h"
#include "object/lib/lib_exceptions_helper.h"
#include "object/lib/lib_exceptions.h"
#include "object/object_manager.h"
#include "replication/replica_client.h"
#include "utility/debug.h"
#include "global/global.h"
#include <sstream>
#include <string>


namespace Massiv {
namespace Core {

namespace {

/** Tests if the two cast objects describe the same pointer cast. */
bool same_casts
    (
    const CastObject & cast1,
    const CastObject & cast2
    )
    {
    #ifndef OPTION_SHARED_LIBRARIES
    
    // If core is linked statically, each cast object (allocated as a
    // global static object, defined by corresponding PointerBase<Cast>) is
    // a singleton.
    // Simple address compare is sufficient.

    return &cast1 == &cast2;

    #else

    // There may be multiple cast object instances describing the same
    // pointer cast. If they are the same type, they must describe the
    // same cast.

    return typeid( cast1 ) == typeid( cast2 );

    #endif
    }

} // anonymous namespace
    

// Implementation of ObjectProxy
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectProxy::ObjectProxy
    (
    const ObjectId &    object_id,
    const CastObject &  cast,
    ObjectProperty *    object
    ) : object_id( object_id ),
        cast( cast ),
        object( object )
    {
    int i;

    for( i = 0; i < POINTER_FLAGS_COUNT; i++ )
        {
        reference_count[ i ] = 0;
        }

    link_proxy( object );
    }


ObjectProxy::~ObjectProxy()
    {
    // ObjectProxy should not be referenced when it is being deleted.
    // The only exception is @c null_proxy, which is a global object and that
    // is deleted automatically at program exit.

    hope( !is_referenced() || this == &ObjectProxyManager::null_proxy );
    }


void ObjectProxy::link_proxy
    (
    ObjectProperty * o
    )
    {
    if( o )
        {
        // Link proxy with Object.

        hope( object_id == o->get_object_id() );

        object = o;
        component = cast( object );
        
        // If @c component is NULL then @c object could not be casted
        // to the requested type using @c cast object.
        // This implies that @c object either does not inherit the
        // requested class, or it is ambiguous (multiple inheritance).
        // The inheritance check is normally performed at pointer assignment.
        // However if type info was not available at that time (the object
        // may have been remote), no check was performed.
        // The invalid/ambiguous assignment will be reported when the
        // corresponding pointer is dereferenced.
        }
    else
        {
        // Unlink proxy.

        object = NULL;
        component.set_null();
        }
    }


void ObjectProxy::link_backward_pointer
    (
    Property *          property,
    BackwardPointer &   bp
    ) RELEASE_NO_THROW()
    {
    // Property pointer @c property is changing,
    // inform referenced object about possible migration group change.

    changing_migration_group();

    // Link backward pointer.

    gc().link_backward_pointer( object_id, property, bp );
    }


void ObjectProxy::unlink_backward_pointer
    (
    BackwardPointer &   bp
    ) RELEASE_NO_THROW()
    {
    // Property pointer is changing, inform referenced object
    // about possible migration group change.

    changing_migration_group();

    // Unlink backward pointer.

    gc().unlink_backward_pointer( bp );
    }


void ObjectProxy::relink_backward_pointer
    (
    ObjectProxy *       new_proxy,
    BackwardPointer &   bp
    ) RELEASE_NO_THROW()
    {
    // Property pointer is changing, inform referenced object
    // and the new object about possible migration group change.

    // TODO: migracni skupiny zpracovavat jen pokud bezi archivace

    changing_migration_group();
    new_proxy->changing_migration_group();

    // Relink backward pointer.

    gc().relink_backward_pointer( new_proxy->object_id, bp );
    }


void ObjectProxy::lazy_relink_backward_pointer
    (
    ObjectProxy *       new_proxy,
    Property *          property
    ) RELEASE_NO_THROW()
    {
    // Property pointer is changing, inform referenced object
    // and the new object about possible migration group change.

    // TODO: migracni skupiny zpracovavat jen pokud bezi archivace

    changing_migration_group();
    new_proxy->changing_migration_group();

    // Lazy-relink backward pointer.

    gc().lazy_relink_backward_pointer( property );
    }


void ObjectProxy::update_backward_pointer
    (
    BackwardPointer &   bp
    )
    {
    gc().update_backward_pointer( object_id, bp );
    }


void ObjectProxy::changing_migration_group() RELEASE_NO_THROW()
    {
    ObjectProperty * object = get_object();

    if( object )
        {
        object->archive_migration_group_state();
        }
    }


void ObjectProxy::changing_pointer_owner
    (
    Property *  property,
    Property *  new_owner
    ) RELEASE_NO_THROW()
    {
    ObjectProperty * object;

    // Property pointer owner is changing, inform linked object,
    // old owner and new owner about possible migration group change.
    
    // TODO: migracni skupiny zpracovavat jen pokud bezi archivace

    object = get_object();

    if( object )
        {
        // Notify linked object.
        object->archive_migration_group_state();
        }
    
    object = property->get_root_owner_object();

    if( object )
        {
        // Notify the old owner.
        object->archive_migration_group_state();
        }

    object = new_owner->get_root_owner_object();
    
    if( object )
        {
        // Notify the new owner.
        object->archive_migration_group_state();
        }
    }


void ObjectProxy::mark_object() RELEASE_NO_THROW()
    {
    gc().mark_object( get_object() );
    }


bool ObjectProxy::inherits
    (
    const std::type_info & class_type_info
    ) const
    {
    if( !object_id.is_object_info_available() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            UndefinedClassTypeException,
            "Type info is not available."
            );
        }
    else
        {
        return object_id.get_class_type_id().inherits( class_type_info );
        }
    }


const std::type_info & ObjectProxy::get_object_type_info() const
    {
    if( !object_id.is_object_info_available() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            UndefinedClassTypeException,
            "Type info is not available."
            );
        }
    else
        {
        return object_id.get_class_type_id().get_type_info();
        }
    }


int ObjectProxy::classify() const RELEASE_NO_THROW()
    {
    int classification = 0;
    int pointer_flags;

    for( pointer_flags = 0; pointer_flags < POINTER_FLAGS_COUNT; pointer_flags++ )
        {
        if( reference_count[ pointer_flags ] )
            {
            /* Proxy is linked with ObjectPointer of type @c pointer_flags. */

            if( pointer_flags & STACK_POINTER )
                {
                if( pointer_flags & WEAK_POINTER )
                    {
                    classification |= CLASSIFY_STACK_WEAK_REFERENCED;
                    }
                else
                    {
                    classification |= CLASSIFY_STACK_STRONG_REFERENCED;
                    }
                }
            else
                {
                if( pointer_flags & WEAK_POINTER )
                    {
                    classification |= CLASSIFY_WEAK_REFERENCED;
                    }
                else
                    {
                    classification |= CLASSIFY_STRONG_REFERENCED;
                    }
                }
            }
        }

    return classification;
    }


bool ObjectProxy::is_referenced() const RELEASE_NO_THROW()
    {
    int count = 0;
    int i;

    for( i = 0; i < POINTER_FLAGS_COUNT; i++ )
        {
        count += reference_count[ i ];
        }

    return count != 0;
    }


GarbageCollector & ObjectProxy::gc()
    {
    return Global::garbage_collector();
    }


// Implementation of ObjectProxyManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectId             ObjectProxyManager::null_id;
DefaultCast<Object>  ObjectProxyManager::null_cast;
ObjectProxy          ObjectProxyManager::null_proxy( null_id, null_cast, NULL );


/**
 * List of ObjectProxy objects.
 * Each proxy is identified by <@c object_id, @c cast> pair,
 * no duplicate entries are allowed.
 * Since we need to do "projections" (ie. find all ObjectProxies
 * for given @c object_id), the data structure has to be indexed
 * by @c object_id.
 */
class ObjectProxyManager::ProxyList : public
        ext::hash_multimap
            <
            ObjectId,
            ObjectProxy *,
            ObjectId::Hash,
            ObjectId::EqualKey
            >
        {
        };


typedef std::pair
    <
    ObjectProxyManager::ProxyList::iterator,
    ObjectProxyManager::ProxyList::iterator
    > ProxyListRange;


ObjectProxyManager::ObjectProxyManager()
    {
    proxy_objects = new ProxyList();
    }


ObjectProxyManager::~ObjectProxyManager()
    {
    tidy_proxy_objects();
    
    if( !proxy_objects->empty() )
        {
        std::ostringstream  ss;
        ProxyList::iterator it;
        ObjectProxy *       proxy;

        /* Can not delete @c proxy_objects because there are ObjectPointers
           linked to them. */
        
        ss << "Can not properly delete ObjectProxyManager because there are "
              "still pointers to managed objects. Have you deleted all global "
              "ObjectPointers?\n"
              "Referenced objects:";

        it = proxy_objects->begin();

        while( it != proxy_objects->end() )
            {
            proxy = it->second;

            hope( proxy->is_referenced() );
            
            ss  << "\n" << proxy->get_object_id() << " [ classify() = " 
                << proxy->classify() << " ]";
            
            ++it;
            }

        panic( ss.str().c_str() );
        }

    delete proxy_objects;
    }


ObjectProxy * ObjectProxyManager::get_proxy
    (
    const ObjectId &    object_id,
    const CastObject &  cast,
    int                 pointer_flags
    )/* throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    {
    ObjectProxy * p;

    /* Requesting @c NULL_ID object? */

    if( object_id.is_null() )
        {
        return get_null_proxy( pointer_flags );
        }

    /* Get proxy. */

    GarbageCollector &  gc = Global::garbage_collector();
    ObjectProperty *    object = gc.find_local_object( object_id );

    can_assign( object_id, object, cast, pointer_flags );
    p = find_proxy( object_id, cast );

    if( !p )
        {
        try
            {
            hope( !object_id.is_null() );

            p = new ObjectProxy( object_id, cast, object );
            proxy_objects->insert( ProxyList::value_type( object_id, p ) );
            }
        catch( ... )
            {
            panic( "GC error." );
            }
        }

    /* Link proxy with ObjectPointer. */

    p->link_pointer( pointer_flags );

    return p;
    }


ObjectProxy * ObjectProxyManager::get_converted_proxy
    (
    ObjectProxy *       proxy,
    const CastObject &  cast,
    int                 pointer_flags
    )/* throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    {
    /* Test validity. */

    can_assign( proxy->get_object_id(), proxy->get_object(), cast,
        pointer_flags );
    
    /* Get proxy. */

    return get_converted_proxy_no_checks( proxy, cast, pointer_flags );
    }


ObjectProxy * ObjectProxyManager::get_converted_proxy_no_checks
    (
    ObjectProxy *       proxy,
    const CastObject &  cast,
    int                 pointer_flags
    ) throw()
    {
    const ObjectId &    object_id = proxy->get_object_id();
    ObjectProxy *       p;
    
    if( !same_casts( proxy->get_cast(), cast ) )
        {
        /* Requesting @c NULL_ID object? */

        if( object_id.is_null() )
            {
            return get_null_proxy( pointer_flags );
            }

        /* Find proxy. */

        p = find_proxy( object_id, cast );

        if( !p )
            {
            try
                {
                hope( !object_id.is_null() );

                p = new ObjectProxy( object_id, cast, proxy->get_object() );
                proxy_objects->insert( ProxyList::value_type( object_id, p ) );
                }
            catch( ... )
                {
                panic( "GC error." );
                }
            }
        }
    else
        {
        /* Can use the same proxy. */

        p = proxy;
        }

    /* Link proxy with ObjectPointer. */

    p->link_pointer( pointer_flags );

    return p;
    }


ObjectProxy * ObjectProxyManager::get_proxy_from_native_pointer
    (
    const ObjectProperty *  object,
    const CastObject &      cast,
    int                     pointer_flags
    )
    #ifdef OPTION_DEBUG
    /*throw( Lib::PointerPolicyViolationException )*/
    #else
    throw()
    #endif    
    {
    ObjectProxy * p;

    /* Requesting @c NULL_ID object? */

    if( !object )
        {
        return get_null_proxy( pointer_flags );
        }

    #ifdef OPTION_DEBUG
    
    if( object->is_on_stack() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            PointerPolicyViolationException,
            "Can not create ObjectPointer to an object on the stack."
            );
        }

    if( object->is_member_object() )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            PointerPolicyViolationException,
            "Can not create ObjectPointer to a member object."
            );
        }

    #endif

    /* Get proxy. */

    const ObjectId & object_id = object->get_object_id();

    #ifdef OPTION_DEBUG
    
    try
        {
        // Can only fail because of pointer policy violation since the
        // conversion validity has been checked in compile-time by ObjectPointer
        // and the referenced object is always local (we have a native
        // pointer to it).
        
        can_assign( object_id, object, cast, pointer_flags );
        }
    catch( Lib::PointerPolicyViolationException & )
        {
        throw;
        }
    catch( Lib::IllegalPointerConversionException & e )
        {
        std::ostringstream ss;

        ss << "Incorrect metaobject for " << object_id << ". "
            "It says that the following conversion is not valid, but "
            "it is: " << e.what();
        
        panic( ss.str().c_str() );
        }
    catch( ... )
        {
        panic( "Internal error in ObjectProxyManager::get_proxy_from_native_pointer()." );
        }

    #endif

    p = find_proxy( object_id, cast );

    if( !p )
        {
        try
            {
            hope( !object_id.is_null() );

            p = new ObjectProxy( object_id, cast, const_cast<ObjectProperty *>( object ) );
            proxy_objects->insert( ProxyList::value_type( object_id, p ) );
            }
        catch( ... )
            {
            panic( "GC error." );
            }
        }

    /* Link proxy with ObjectPointer. */

    p->link_pointer( pointer_flags );

    return p;
    }


ObjectProxy * ObjectProxyManager::find_proxy
    (
    const ObjectId &    object_id,
    const CastObject &  cast
    )
    {
    ProxyListRange  range( proxy_objects->equal_range( object_id ) );
    ObjectProxy *   proxy;
    ProxyList::iterator it;

    for( it = range.first; it != range.second; ++it )
        {
        proxy = it->second;

        if( same_casts( proxy->get_cast(), cast ) )
            {
            return proxy;
            }
        }

    return NULL;
    }


void ObjectProxyManager::can_assign
    (
    const ObjectId &        object_id,
    const ObjectProperty *  object,
    const CastObject &      cast,
    int                     pointer_flags
    )
    #ifdef OPTION_DEBUG
    /*throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    #else
    /*throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException
        )*/
    #endif     
    {
    const std::type_info & pointer_type = cast.get_target_type_info();

    #ifdef OPTION_DEBUG

    if( object != NULL )
        {
        hope( object->get_object_id() == object_id );
        }

    #endif

    if( !object_id.is_null() )
        {
        /**
         * Interface test.
         * Test if @c object_id inherits @c pointer_type.
         * If type information of the object identified by @c object_id
         * is not available or can not be interpreted by local node,
         * the test is not performed.
         */

        const MetaObject * metaobject = 
            object_id.is_object_info_available() ?
            object_id.get_class_type_id().get_metaobject_no_throw() :
            NULL;

        if( metaobject )
            {
            if( !metaobject->inherits( pointer_type ) )
                {
                std::ostringstream s;

                s   << "Can not assign \""
                    << metaobject->get_class_type_info().name()
                    << " *\" to \""
                    << pointer_type.name() << " *\".";

                RAISE_CORE_LIB_EXCEPTION( IllegalPointerConversionException, s.str() );
                }
            }
            
        /* Object presence and policy test. */

        can_assign( object_id, object, pointer_flags );
        }
    }


void ObjectProxyManager::can_assign
    (
    const ObjectId &        object_id,
    const ObjectProperty *  object,
    int                     pointer_flags
    )
    #ifdef OPTION_DEBUG
    /*throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    #else
    /*throw
        (
        Lib::IllegalPointerConversionException,
        Lib::ObjectNotOnLocalSystemException
        )*/
    #endif
    {
    if( !flags_on( pointer_flags, WEAK_POINTER ) && !object && 
        !object_id.is_null() )
        {
        std::ostringstream s;
        
        s   << "Can not assign remote object " << object_id
            << " to a strong pointer.";
        
        RAISE_CORE_LIB_EXCEPTION( ObjectNotOnLocalSystemException, s.str() );
        }

    #ifdef OPTION_DEBUG

    if( !flags_on( pointer_flags, WEAK_POINTER ) && object &&
        object->is_replica() )
        {
        std::ostringstream s;

        s   << "Can not assign replica of " << object_id
            << " to a strong pointer.";

        RAISE_CORE_LIB_EXCEPTION( PointerPolicyViolationException, s.str() );
        }

    #endif
    }


void ObjectProxyManager::test_pointer_property_policy
    (
    Property *      pointer,
    ObjectProxy *   proxy
    )
    #ifdef OPTION_DEBUG
    /*throw( Lib::PointerPolicyViolationException )*/
    #else
    throw()
    #endif
    {
    #ifdef OPTION_DEBUG

    hope( pointer );
    hope( proxy );
    ObjectProperty *    owner_object = pointer->get_root_owner_object();
    ObjectProperty *    referenced_object = proxy->get_object();
    const ObjectId &    referenced_object_id = proxy->get_object_id();
    PointerInfo *       info = pointer->get_pointer_info();
    hope( info );
    int                 pointer_flags = info->ptr_get_pointer_flags();
    int                 ptr_replication_flags = info->ptr_get_replication_flags();

    /* Pointer property policies. */

    if( proxy == &null_proxy )
        {
        return;
        }

    hope( owner_object );

    if( owner_object->is_on_stack() )
        {
        // The pointer is either on program stack or has not been fully
        // initialized yet. In the latter case the policy must be retested
        // later when pointer gets initialized.

        return;
        }

    if( pointer->is_replica() )
        {
        // Replicas have all ptr_repflags set "correctly", but they are used
        // only for object enumeration. The following check should be performed,
        // but it can't be, because replicas may be inconsistent (see ReplicaClient):

        if( pointer->is_consistent() )
            {
            int repflags_yes;
            int repflags_no;
            Global::object_manager().replica_client->get_local_repflags( repflags_yes, repflags_no );
            if( flags_on( ptr_replication_flags, repflags_yes, repflags_no ) )
                {
                // ptr_repflags match repflags of local node, therefor pointed object
                // must be local and it must be replica (belonging to the same
                // replication group).

                // WARNING: This is not actually a pointer policy because application
                //          programmer can not change/write to replicas. They are updated
                //          automatically by the replication update protocol. This check is
                //          internal to the core and is intended to catch replication
                //          update protocol bugs.

                if( !referenced_object->is_replica() )
                    {
                    std::ostringstream s;

                    s << "REPLICATION PROTOCOL BUG: Pointer replica with ptr_repflags "
                      << "matching repflags of local node can not point to a local object "
                      << referenced_object_id
                      << ".";

                    RAISE_CORE_LIB_EXCEPTION( PointerPolicyViolationException, s.str() );
                    }
                }
            }
        }
    else
        {
        if( !flags_on( pointer_flags, WEAK_POINTER ) )
            {
            hope( flags_on( ptr_replication_flags, REPFLAGS_MIGRATE ) );
            }

        if( flags_on( ptr_replication_flags, REPFLAGS_MIGRATE ) )
            {
            if( !referenced_object && !referenced_object_id.is_null() )
                {
                std::ostringstream s;

                s   << "Pointer property with REPFLAGS_MIGRATE pointer replication "
                    << "flags bit set can not point to a remote object "
                    << referenced_object_id
                    << ".";

                RAISE_CORE_LIB_EXCEPTION( PointerPolicyViolationException, s.str() );
                }

            if( referenced_object && referenced_object->is_replica() )
                {
                std::ostringstream s;

                s   << "Pointer property with REPFLAGS_MIGRATE pointer replication "
                    << "flags bit set can not point to an object replica "
                    << referenced_object_id
                    << ".";

                RAISE_CORE_LIB_EXCEPTION( PointerPolicyViolationException, s.str() );
                }

            if( referenced_object &&
                referenced_object->is_archivable() ^ owner_object->is_archivable() )
                {
                std::ostringstream s;

                s   << "Merging archivable and non-archivable migration groups. "
                    << owner_object->get_object_id()
                    << " tries to link to "
                    << referenced_object_id
                    << ".";

                RAISE_CORE_LIB_EXCEPTION( PointerPolicyViolationException, s.str() );
                }
            }
        }

    #endif // OPTION_DEBUG
    }


void ObjectProxyManager::validate_pointers
    (
    ObjectProperty * object
    )
    {
    ProxyListRange      range( proxy_objects->equal_range( object->get_object_id() ) );
    ObjectProxy *       proxy;
    ProxyList::iterator it;

    /* Link ObjectProxies for @c object with @c object. */

    for( it = range.first; it != range.second; ++it )
        {
        proxy = it->second;
        proxy->link_proxy( object );
        }
    }


void ObjectProxyManager::invalidate_pointers
    (
    ObjectProperty * object
    )
    {
    ProxyListRange      range( proxy_objects->equal_range( object->get_object_id() ) );
    ObjectProxy *       proxy;
    ProxyList::iterator it;

    /* Unlink ObjectProxies for @c object from @c object. */

    for( it = range.first; it != range.second; ++it )
        {
        proxy = it->second;
        proxy->link_proxy( NULL );
        }
    }


void ObjectProxyManager::invalidate_all_pointers()
    {
    ObjectProxy *       proxy;
    ProxyList::iterator it;

    /* Unlink all ObjectProxies from local objects. */

    for( it = proxy_objects->begin(); it != proxy_objects->end(); ++it )
        {
        proxy = it->second;
        proxy->link_proxy( NULL );
        }
    }


int ObjectProxyManager::classify_object
    (
    const ObjectProperty * object
    )
    {
    ProxyListRange      range( proxy_objects->equal_range( object->get_object_id() ) );
    ObjectProxy *       proxy;
    ProxyList::iterator it;
    int                 classification;

    /* Classify object according to its proxy objects. */

    classification = 0;

    for( it = range.first; it != range.second; ++it )
        {
        proxy = it->second;
        classification |= proxy->classify();
        }

    return classification;
    }


void ObjectProxyManager::tidy_proxy_objects()
    {
    ProxyList::iterator it;
    ObjectProxy *       proxy;

    it = proxy_objects->begin();

    while( it != proxy_objects->end() )
        {
        proxy = it->second;

        if( !proxy->is_referenced() )
            {
            delete proxy;
            proxy_objects->erase( it++ );
            }
        else
            {
            ++it;
            }
        }
    }


void ObjectProxyManager::mark_stack_referenced_objects()
    {
    ProxyList::iterator it;
    ObjectProxy *       proxy;

    for( it = proxy_objects->begin(); it != proxy_objects->end(); ++it )
        {
        proxy = it->second;

        if( proxy->classify() & ObjectProxy::CLASSIFY_STACK_STRONG_REFERENCED )
            {
            proxy->mark_object();
            }
        }
    }


// Implementation of ActiveObjectProxy
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ActiveObjectProxy::err_null_object()
    {
    RAISE_CORE_LIB_EXCEPTION
        (
        CanNotAccessObjectException,
        "Dereferencing NULL_ID pointer."
        );
    }


void ActiveObjectProxy::err_not_local()
    {
    RAISE_CORE_LIB_EXCEPTION
        (
        ObjectNotOnLocalSystemException,
        "Dereferencing pointer to remote object."
        );
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "pointer/object_proxy_inline.h"
#endif
