/** @file object_proxy.h
 * Object proxy.
 */

/*
 * begin        : Thu Dec 25 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_PROXY_H
#define MASSIV_OBJECT_PROXY_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_POINTER_PARAMETERS_H
#include "pointer/pointer_parameters.h"
#endif

#ifndef MASSIV_DEFAULT_CAST_H
#include "pointer/default_cast.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_RPC_STUBS_H
#include "pointer/rpc_stubs.h"
#endif

#ifndef MASSIV_LIB_EXCEPTIONS_HELPER_H
#include "object/lib/lib_exceptions_helper.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup pointer
 * @{
 */

/** @class ObjectProxy
 * Each ObjectPointer is linked with an @c ObjectProxy that
 * serves as a proxy object for referenced object, holding most of the
 * information associated with an ObjectPointer. 
 *
 * ObjectPointer never references the object directly.
 * This indirection allows to dynamically validate/invalidate
 * ObjectPointers by simple linking/unlinking object proxies
 * to/from corresponding objects.
 *
 * ObjectPointer's dereferencing mechanism returns "active proxies"
 * that either directly access linked objects or translate method
 * calls to RPC calls. See @c ActiveObjectProxy.
 */
class MASSIV_EXPORTED ObjectProxy
    {
public:

    typedef ObjectShared::BackwardPointer   BackwardPointer;

    /**
     * Classification flags.
     * Used by @c classify().
     */
    enum Classification
        {
        CLASSIFY_STRONG_REFERENCED          =   1,
            /**< Proxy is referenced by a (local) property-strong pointer. */
        CLASSIFY_STACK_STRONG_REFERENCED    =   2,
            /**< Proxy is referenced by a (local) stack-strong pointer. */
        CLASSIFY_WEAK_REFERENCED            =   4,
            /**< Proxy is referenced by a (local) property-weak pointer. */
        CLASSIFY_STACK_WEAK_REFERENCED      =   8,
            /**< Proxy is referenced by a (local) stack-weak pointer. */
        }; // ObjectIdClassification

public:

    //
    // Constructors & destructors.
    //

    /** Constructor. */
    ObjectProxy
        (
        const ObjectId &    object_id,
        const CastObject &  cast,
        ObjectProperty *    object
        );

    /** Destructor. */
    ~ObjectProxy();

    NO_COPY_CONSTRUCTOR( ObjectProxy );
    NO_ASSIGNMENT_OPERATOR( ObjectProxy );

public:

    //
    // Proxy queries.
    //

    /** Returns @c object_id of bound object. */
    INLINE const ObjectId & get_object_id() const RELEASE_NO_THROW();

    /**< Tests if the bound object is @c NULL_ID. */
    INLINE bool is_null() const RELEASE_NO_THROW();

    /** Returns pointer to the bound object. */
    INLINE ObjectProperty * get_object() const RELEASE_NO_THROW();

    /** Returns the component of the bound object. */
    INLINE const VariantPointer & get_component() const RELEASE_NO_THROW();
    
    /** Tests if the component of the bound object is not ambiguous. */
    INLINE bool is_ambiguous() const RELEASE_NO_THROW();

    /** Returns cast object. */
    INLINE const CastObject & get_cast() const RELEASE_NO_THROW();

    /**
     * Tests if linked object inherits from @c class_type_info.
     * If dynamic object type is unknown UndefinedClassTypeException is
     * thrown.
     */
    bool inherits
        (
        const std::type_info & class_type_info
        ) const;

    /**
     * Returns type information of the referenced object.
     * If dynamic object type is unknown to the current node,
     * type information is unavailable UndefinedClassTypeException
     * is thrown.
     */
    const std::type_info & get_object_type_info() const;

    /**
     * Classifies proxy object.
     * Returns ORed flag combination defined at @c Classification.
     * Note that classification of proxy object partially classifies
     * linked object. To fully classify an object, use
     * @c ObjectProxyManager::classify_object().
     */
    int classify() const RELEASE_NO_THROW();

    /** Tests if proxy is referenced by ObjectPointer. */
    bool is_referenced() const RELEASE_NO_THROW();

public:

    //
    // ObjectProxy & ObjectPointer linking.
    //

    /**
     * Links proxy with Object.
     * Caches pointer to component, validates all ObjectPointers
     * that share this @c ObjectProxy.
     * If @c o is @c NULL then proxy is unlinked from the object
     * and corresponding ObjectPointers are invalidated.
     */
    void link_proxy
        (
        ObjectProperty * o
        );

    /** Links ObjectPointer of type @c pointer_flags with proxy. */
    INLINE void link_pointer
        (
        int pointer_flags
        );

    /** Unlinks ObjectPointer of type @c pointer_flags from proxy. */
    INLINE void unlink_pointer
        (
        int pointer_flags
        );

    /**
     * Creates and links backward pointer.
     * The pointer is "inverse" to pointer property described
     * by @c property. Returns @c bp.
     * Wrapper for GarbageCollector::link_backward_pointer().
     */
    void link_backward_pointer
        (
        Property *          property,
        BackwardPointer &   bp
        ) RELEASE_NO_THROW();

    /**
     * Unlinks pointer created by @c link_backward_pointer().
     * Invalidates @c bp.
     * Wrapper for GarbageCollector::unlink_backward_pointer().
     */
    void unlink_backward_pointer
        (
        BackwardPointer &   bp
        ) RELEASE_NO_THROW();

    /**
     * Unlinks backward pointer created by @c link_backward_pointer() or
     * @c relink_backward_pointer() from proxy and links it
     * to @c new_proxy. Fills @c bp.
     * Wrapper for GarbageCollector::relink_backward_pointer().
     */
    void relink_backward_pointer
        (
        ObjectProxy *       new_proxy,
        BackwardPointer &   bp
        ) RELEASE_NO_THROW();

    /**
     * Lazy-relinks backward pointer created by @c link_backward_pointer()
     * or lazy_relink_backward_pointer() from proxy and links it to
     * @c new_proxy.
     * Wrapper for GarbageCollector::lazy_relink_backward_pointer().
     */
    void lazy_relink_backward_pointer
        (
        ObjectProxy *       new_proxy,
        Property *          property
        ) RELEASE_NO_THROW();

    /**
     * Updates backward pointer.
     * Wrapper for GarbageCollector::update_backward_pointer().
     */
    void update_backward_pointer
        (
        BackwardPointer &   bp
        );

public:

    //
    // ObjectPointer helper methods.
    //

    /**
     * Called by ObjectPointer when migration group for linked object
     * might change.
     */
    void changing_migration_group() RELEASE_NO_THROW();

    /**
     * Called by ObjectPointer when pointer owner changes.
     * The implementation uses this to notify the new owner, old owner
     * and linked object about potential change of their migration
     * groups.
     * @c property determines the pointer property being processed.
     */
    void changing_pointer_owner
        (
        Property *  property,
        Property *  new_owner
        ) RELEASE_NO_THROW();

    /** Marks linked object reachable from GC root objects. */
    void mark_object() RELEASE_NO_THROW();

private:

    //
    // Private.
    //

    /** Returns Garbage Collector. */
    static GarbageCollector & gc();

private:

    //
    // ObjectProxy to Object linkage.
    //

    ObjectId            object_id;
        /**< ObjectId of bound object. */
    const CastObject &  cast;
        /**< Cast object that specifies what object component
             is referenced. */
    ObjectProperty *    object;
        /**< Pointer to @c Object. Link between @c ObjectProxy and
             corresponding local @c Object.
             Equals to NULL if object is not registered on the local
             system. */
    VariantPointer      component;
        /**< Pointer to Object's component specified by @c cast. */

private:

    //
    // ObjectPointer to ObjectProxy linkage.
    //

    int reference_count[ POINTER_FLAGS_COUNT ];
        /**< Number of @c ObjectPointers that share this @c ObjectProxy.
             Classified according to their pointer flags.
             This allows to detect what pointer types reference particular
             objects. */

    }; // ObjectProxy

/**
 * Simple class that manages ObjectProxy objects, used by ObjectPointer and
 * Garbage Collector.
 *
 * Maintains a list of ObjectProxies, manages its allocation to
 * ObjectPointers and ensures consistent use of ObjectPointers by verifying
 * so called pointer policies. There are two kinds of pointer policies:
 * basic policies applicable to both stack and pointer properties and
 * extended policies applicable to pointer properties only.
 *
 * Basic policies relate to the WEAK_POINTER flag and what objects can be
 * referenced by object pointers. Extended policies test for proper use
 * of pointers with respect to implicit definition of migration and
 * replication groups.
 *
 * If a policy violation is detected, one of the following exceptions will
 * be thrown: IllegalPointerConversionException, ObjectNotOnLocalSystemException
 * and general PointerPolicyViolationException. The last one is thrown
 * on OPTION_DEBUG mode only.
 */
class MASSIV_EXPORTED ObjectProxyManager
    {
public:

    friend class ObjectProxy;

    class ProxyList;

public:

    /** Constructor. */
    ObjectProxyManager();

    /** Destructor. */
    ~ObjectProxyManager();

public:

    //
    // Accessing ObjectProxy.
    //

    /**
     * Returns proxy for <@c object_id, @c cast>.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     *
     * Throws IllegalPointerConversionException if @c object_id can not be
     * casted using @c cast object.
     * Throws ObjectNotOnLocalSystemException if pointer_flags specifies a
     * strong pointer and @c object_id refers to a remote object.
     * Throws PointerPolicyViolation if pointer to requested object can
     * not be created because of policy violation.
     *
     * When the returned proxy is going to be forgotten (the pointer instance
     * is going to be thrown away) @c release_proxy() must be called.
     *
     * @note If object's type info is not available, no conversion test
     * validity is checked!
     *
     * @note This is the most general version of @c get_proxy(). The other
     * variants are used in special cases only like pointer upcasting, etc.
     * These variants are highly optimized but rely on additional
     * preconditions.
     */
    ObjectProxy * get_proxy
        (
        const ObjectId &    object_id,
        const CastObject &  cast,
        int                 pointer_flags
        ) /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;
    
    /**
     * Returns object proxy that differs from @c proxy in @c cast
     * parameter and @c pointer_flags. 
     * Used for converting ObjectPointers only.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     */
    ObjectProxy * get_converted_proxy
        (
        ObjectProxy *       proxy,
        const CastObject &  cast,
        int                 pointer_flags
        ) /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;
    
    /**
     * Returns object proxy that differs from @c proxy in @c cast
     * parameter and @c pointer_flags.
     * A @c get_converted_proxy() variant that does not do runtime checks.
     */
    ObjectProxy * get_converted_proxy_no_checks
        (
        ObjectProxy *       proxy,
        const CastObject &  cast,
        int                 pointer_flags
        ) throw();
    
    /**
     * Returns object proxy that differs from @c proxy in @c cast
     * parameter and @c pointer_flags. @c old_pointer_flags refers to flags
     * that were used to get @c proxy.
     *
     * This version supports type-safe upcasting together with pointer flags
     * change only, ie. @c cast must refer to a parent class of the object
     * represented by @c proxy. Conversion validity (except for pointer flags
     * change) must be checked by ObjectPointer which can be done in
     * compile-time.
     *
     * Used for upcasting ObjectPointers only.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     */
    INLINE ObjectProxy * get_upcasted_proxy
        (
        ObjectProxy *       proxy,
        const CastObject &  cast,
        int                 old_pointer_flags,
        int                 pointer_flags
        ) /*throw
            (
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;

    /**
     * Returns object proxy for @c object.
     * Used for converting native pointers to ObjectPointers.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     *
     * @note IllegalPointerConversionException is never thrown because the validity
     * is checked in compile-time. ObjectNotOnLocalSystemException is also not
     * thrown because the object must be local if we have a native pointer to
     * it. However PointerPolicyViolation can be thrown on debug mode.
     */
    ObjectProxy * get_proxy_from_native_pointer
        (
        const ObjectProperty *  object,
        const CastObject &      cast,
        int                     pointer_flags
        )
        #ifdef OPTION_DEBUG
        /*throw( Lib::PointerPolicyViolationException )*/;
        #else
        throw();
        #endif    

    /**
     * Returns proxy that differs from @c proxy in pointer_flags only.
     * @c old_pointer_flags refers to the old pointer flags that were used
     * to get @c proxy, @c pointer_flags refers to the new requested flags.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     */
    INLINE ObjectProxy * get_proxy_with_different_flags
        (
        ObjectProxy *       proxy,
        int                 old_pointer_flags,
        int                 pointer_flags
        ) /*throw
            (
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;
    
    /**
     * Duplicates "handle" to @c proxy.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     */
    INLINE ObjectProxy * get_duplicated_proxy
        (
        ObjectProxy *       proxy,
        int                 pointer_flags
        ) RELEASE_NO_THROW();

    /**
     * Returns proxy to @c NULL_ID object.
     * Links proxy with ObjectPointer of type @c pointer_flags.
     * @note Returned proxy is "universal" - used for any type
     * of cast object.
     */
    INLINE ObjectProxy * get_null_proxy
        (
        int                 pointer_flags
        ) RELEASE_NO_THROW();

    /**
     * Releases @c proxy returned by @c get_proxy().
     * Unlinks ObjectPointer of type @c pointer_flags from the proxy.
     * Each call to a @c get_proxy() must be paired with a subsequent
     * call to @c release_proxy().
     */
    INLINE void release_proxy
        (
        ObjectProxy *       proxy,
        int                 pointer_flags
        ) RELEASE_NO_THROW();
    
public:

    //
    // Pointer policies.
    //

    /**
     * Checks if we can safely assign @c object_id to ObjectPointer
     * with <@c pointer_flags, @c cast> characteristics, verifies basic
     * pointer policies.
     *
     * If the test fails IllegalPointerConversionException,
     * ObjectNotOnLocalSystemException or PointerPolicyViolationException
     * exceptions are thrown.
     * Called automatically by @c get_proxy() variants as needed.
     *
     * @note Pointer property specific policies are further tested
     *       by @c test_pointer_property_policy().
     *
     * @note PointerPolicyViolationException is thrown in OPTION_DEBUG mode
     *       only so that it is consistent with @c test_pointer_property_policy().
     *
     * @note @c object is passed as additional parameter for effeciency reasons.
     *       If object with given @c object_id is local @c object must be
     *       valid pointer to such object and if it is remote @c object must
     *       be NULL.
     */
    static void can_assign
        (
        const ObjectId &        object_id,
        const ObjectProperty *  object,
        const CastObject &      cast,
        int                     pointer_flags
        )
        #ifdef OPTION_DEBUG
        /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;
        #else
        /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException
            )*/;
        #endif     
    
    /**
     * Checks if ObjectPointer with specified @c pointer_flags can
     * reference @c object with specified @c object_id. Use NULL @c object
     * if @c object_id is remote.
     *
     * Called automatically by @c get_proxy() variants or more general
     * @c can_assign() as needed.
     */
    static void can_assign
        (
        const ObjectId &        object_id,
        const ObjectProperty *  object,
        int                     pointer_flags
        )
        #ifdef OPTION_DEBUG
        /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException,
            Lib::PointerPolicyViolationException
            )*/;
        #else
        /*throw
            (
            Lib::IllegalPointerConversionException,
            Lib::ObjectNotOnLocalSystemException
            )*/;
        #endif     

    /**
     * Tests whether @c proxy (that was returned by one of @c get_proxy()
     * variants and passed @c can_assign() test) can be assigned to @c pointer
     * property, verifies extended pointer policies.
     * 
     * If assignment is rejected an exception is thrown.
     * The test is performed in OPTION_DEBUG mode only.
     * To be called by @c PropertyPointerBase when pointer value / type
     * changes.
     *
     * @note @c null_proxy must not be rejected.
     *
     * @note Pointer policy tested here is defined for pointer properties only
     * and is called just before the pointer property is updated.
     * @c can_assign() defines and tests more general policies that apply to
     * both stack pointers and pointer properties.
     */
    static void test_pointer_property_policy
        (
        Property *      pointer,
        ObjectProxy *   proxy
        )
        #ifdef OPTION_DEBUG
        /*throw( Lib::PointerPolicyViolationException )*/;
        #else
        throw();
        #endif
    
    /**
     * Tests if @c new_pointer_flags are less restrictive than
     * @c pointer_flags.
     *
     * If yes then pointer with @c pointer_flags and some @c cast cast object
     * can be converted to a pointer with @c new_pointer_flags and @c new_cast
     * cast object without a run-time test if @c new_cast refers to the same
     * component or a superclass of the component referenced by @c cast.
     *
     * @note This must respect pointer policy.
     */
    static INLINE bool can_reduce_pointer_flags
        (
        int pointer_flags,
        int new_pointer_flags
        );
    
public:

    //
    // Garbage collector helper methods.
    //

    /**
     * Validates ObjectPointers pointing to @c object.
     */
    void validate_pointers
        (
        ObjectProperty * object
        );

    /**
     * Invalidates ObjectPointers pointing to @c object.
     */
    void invalidate_pointers
        (
        ObjectProperty * object
        );

    /**
     * Invalidates all ObjectPointers.
     */
    void invalidate_all_pointers();

    /**
     * Determines what types of local ObjectPointers reference @c object.
     * The result is an ORed flag combination defined
     * at @c ProxyObject::Classification.
     */
    int classify_object
        (
        const ObjectProperty * object
        );

    /**
     * Deletes unused (unreferenced) proxy objects.
     */
    void tidy_proxy_objects();

    /**
     * Marks stack-strong referenced objects.
     */
    void mark_stack_referenced_objects();

private:

    //
    // Private stuff.
    //

    /**
     * Finds requested proxy object.
     * Returns NULL if such proxy is not registered.
     */
    ObjectProxy * find_proxy
        (
        const ObjectId &    object_id,
        const CastObject &  cast
        );

private:

    ProxyList *                 proxy_objects;
        /**< List of registered ObjectProxies. */
    static ObjectId             null_id;
        /**< @c NULL_ID. */
    static DefaultCast<Object>  null_cast;
        /**< Cast object used by @c null_proxy. */
    static ObjectProxy          null_proxy;
        /**< Null proxy. */

    }; // ObjectProxyManager


/** @class ActiveObjectProxy
 * Active proxy is a proxy object that is always residing on program stack.
 *
 * It is returned by ObjectPointer::operator-\>() or
 * ObjectPointer::get_proxy(), always by value.
 * Proxy object can further be dereferenced via overloaded
 * operator-\>() which in fact defines semantics of ObjectPointer
 * dereferencing.
 *
 * @c LocalProxy is used for accessing members of the local bound object.
 * @c RemoteProxy is used for remote invocation of methods on the bound
 * remote or local object.
 *
 * Active proxy's life time corresponds to life time of "pointer"
 * returned by ObjectPointer::operator-\>().
 * Active proxy can be stored and safely manipulated by client.
 * This can be used for reducing ObjectPointer::operator-\>()
 * overhead if it is dereferenced multiple times within the same context.
 */
class MASSIV_EXPORTED ActiveObjectProxy
    {
protected:

    /** Raises "Dereferencing NULL_ID object." exception. */
    void err_null_object();

    /** Raises "Dereferencing pointer to remote object." exception. */
    void err_not_local();

    /** Active-pins object. */
    INLINE void active_pin
        (
        ObjectProperty * object
        );
    
    /** Active-unpins object. */
    INLINE void active_unpin
        (
        ObjectProperty * object
        );

    }; // ActiveObjectProxy


/** @class LocalProxy
 * LocalProxy represents an active link to a local object.
 *
 * The proxy is used for determining time period for which
 * the bound object is being "dereferenced". During that period
 * object can not be deleted or migrated because client may touch
 * its data via native pointer returned by @c operator-\>().
 * To ensure that object will not be deleted the object is pinned up
 * until the destruction of the active proxy.
 *
 * Client must not store native pointers to Iface. ObjectPointers or
 * LocalProxies can be used instead.
 */
template <class Iface>
class LocalProxy : public ActiveObjectProxy
    {
public:

    /** Activates proxy. */
    LocalProxy
        (
        ObjectProxy * proxy
        )
        {
        #ifdef OPTION_DEBUG
        if( !proxy->get_object() )
            {
            if( proxy->is_null() )
                err_null_object();
            else
                err_not_local();
            }
        #endif

        component = variant_cast<Iface *>( proxy->get_component() );
        active_pin( component );
        }

    /** Copies active proxy. */
    LocalProxy
        (
        const LocalProxy & proxy
        ) : component( proxy.component )
        {
        active_pin( component );
        }

    #ifndef GC_DELAYED_UNPIN

    /** Deactivates proxy. */
    ~LocalProxy()
        {
        active_unpin( component );
        }

    #endif

    /**
     * Accesses component of @c Object.
     * Until the program releases returned "Iface *" proxy object remains 
     * active.
     */
    Iface * operator->() const
        {
        return component;
        }
        
    /**
     * Accesses component of @c Object.
     * The returned "Iface *" must be released when the proxy becomes inactive.
     */
    operator Iface *() const
        {
        return component;
        }
    
    /**
     * Accesses component of @c Object.
     * The returned "Iface &" must be released when the proxy becomes inactive.
     */
    operator Iface &() const
        {
        return *component;
        }
    
    /**
     * Accesses component of @c Object.
     * The returned variant pointer must be released when the proxy becomes
     * inactive.
     */
    operator VariantPointer() const
        {
        return VariantPointer( component );
        }
    
    NO_ASSIGNMENT_OPERATOR( LocalProxy<Iface> );

protected:

    Iface * component;
        /**< Native pointer to bound object's component. */

    }; // LocalProxy


/** @class RemoteProxy
 * RemoteProxy represents an active link to a remote object with interface
 * @c Iface.
 *
 * The proxy in cooperation with RPCStubs class translates local method calls
 * to RPC calls.
 */
template <class Iface>
class RemoteProxy :
    public ActiveObjectProxy,
    public Massiv::Generated::RPCStubs<Iface>
    {
private:

    typedef Massiv::Generated::RPCStubs<Iface> Stubs;

    using Stubs::_bind;

public:

    /** Activates proxy. */
    RemoteProxy
        (
        ObjectProxy * proxy
        )
        {
        #ifdef OPTION_DEBUG
        if( proxy->is_null() )
            err_null_object();
        #endif
        _bind( proxy->get_object_id() );
        }

    /** Copies active proxy. */
    RemoteProxy
        (
        const RemoteProxy & proxy
        )
        {
        }

    /** Deactivates proxy. */
    ~RemoteProxy()
        {
        }

    /**
     * Accesses methods on @c object_id exposed through @c Iface.
     * Returned @c Stubs does the advertised calls translation.
     */
    Stubs * operator->()
        {
        return this;
        }

    /**
     * Accesses const methods on @c object_id exposed through @c Iface.
     * Returned @c Stubs does the advertised calls translation.
     */
    const Stubs * operator->() const
        {
        return this;
        }

    }; // RemoteProxy

/** @} */ // addtogroup pointer


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#ifndef MASSIV_OBJECT_PROXY_INLINE_H
#include "object_proxy_inline.h"
#endif

#endif

#endif // MASSIV_OBJECT_PROXY_H
