/** @file object_proxy_inline.h
 * Object proxy inline methods.
 */

/*
 * begin        : Sat Jan 18 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_OBJECT_PROPERTY_H
#include "object/object_property.h"
#endif

#ifndef MASSIV_LIB_EXCEPTIONS_HELPER_H
#include "object/lib/lib_exceptions_helper.h"
#endif

#ifdef OPTION_DEBUG

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#endif

namespace Massiv {
namespace Core {


// Inline implementation of ObjectProxy
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

const ObjectId & ObjectProxy::get_object_id() const RELEASE_NO_THROW()
    {
    return object_id;
    }

bool ObjectProxy::is_null() const RELEASE_NO_THROW()
    {
    ObjectProxy * null_proxy = &ObjectProxyManager::null_proxy;

    #ifdef OPTION_DEBUG

    // There is only one proxy that is linked to @c NULL_ID object.
    // That proxy is called @c null_proxy.

    if( object_id.is_null() )
        {
        hope( this == null_proxy );
        }
    else
        {
        hope( this != null_proxy );
        }

    #endif

    return this == null_proxy;
    }

ObjectProperty * ObjectProxy::get_object() const RELEASE_NO_THROW()
    {
    return object;
    }

const VariantPointer & ObjectProxy::get_component() const RELEASE_NO_THROW()
    {
    #ifdef OPTION_DEBUG

    if( is_ambiguous() )
        {
        // In a normal case conversion validity is checked at assign-time.
        // If type information was not available (the referenced object was
        // remote, for example) even an illegal cast could have not been
        // rejected. The object is now local and the requested component
        // is unavailable or ambiguous.

        RAISE_CORE_LIB_EXCEPTION
            (
            IllegalPointerConversionException,
            std::string( "Conversion from \"" ) + 
            typeid( *object ).name() + " *\" to \"" +
            component.get_type_info().name() +
            "\" is illegal or ambiguous."
            );
        }

    #endif

    return component;
    }

bool ObjectProxy::is_ambiguous() const RELEASE_NO_THROW()
    {
    return object != NULL && component.is_null();
    }

const CastObject & ObjectProxy::get_cast() const RELEASE_NO_THROW()
    {
    return cast;
    }

void ObjectProxy::link_pointer
    (
    int pointer_flags
    )
    {
    hope( pointer_flags >= 0 && pointer_flags < POINTER_FLAGS_COUNT );
    reference_count[ pointer_flags ]++;

    #ifdef GC_MARK_PHASE_SPANNED

    // Mark referenced object to prevent GC from deleting it.
    // Since mark phase may be spanned among multiple GC ticks
    // each tick can "see" different memory image
    // (some pointers have changed since previous tick).
    // This could result in fake detection of unreachable objects.
    // To avoid this, we conservatively call @c GC::mark_object()
    // whenever we link ObjectPointer to an object (via proxy object).
    // Referenced object is marked as if it was reachable
    // from a GC root.

    mark_object();

    #endif
    }

void ObjectProxy::unlink_pointer
    (
    int pointer_flags
    )
    {
    hope( pointer_flags >= 0 && pointer_flags < POINTER_FLAGS_COUNT );
    reference_count[ pointer_flags ]--;
    }


// Inline implementation of ObjectProxyManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ObjectProxy * ObjectProxyManager::get_upcasted_proxy
    (
    ObjectProxy *       proxy,
    const CastObject &  cast,
    int                 old_pointer_flags,    
    int                 pointer_flags
    ) /*throw
        (
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    {
    if( can_reduce_pointer_flags( old_pointer_flags, pointer_flags ) )
        {
        #ifdef OPTION_DEBUG
        
        try
            {
            return get_converted_proxy( proxy, cast, pointer_flags );
            }
        catch( ... )
            {
            panic
                ( 
                "ObjectPointer upcast failed because either: "
                "1) Can not upcast proxy to a component specified by the cast "
                "object. Probably error in the IDL."
                "2) Pointer flags used to retrieve previous proxy could not be "
                "reduced to requested pointer flags. This is an internal error."
                );
            /*NOTREACHED*/
            }

        #else

        // Upcast must always succeed.

        return get_converted_proxy_no_checks( proxy, cast, pointer_flags );

        #endif
        }
    else
        {
        // Must perform checked conversion.

        return get_converted_proxy( proxy, cast, pointer_flags );
        }
    }

ObjectProxy * ObjectProxyManager::get_proxy_with_different_flags
    (
    ObjectProxy *       proxy,
    int                 old_pointer_flags,
    int                 pointer_flags
    ) /*throw
        (
        Lib::ObjectNotOnLocalSystemException,
        Lib::PointerPolicyViolationException
        )*/
    {
    if( can_reduce_pointer_flags( old_pointer_flags, pointer_flags ) )
        {
        #ifdef OPTION_DEBUG

        try
            {
            can_assign( proxy->get_object_id(), proxy->get_object(), 
                pointer_flags );
            }
        catch( ... )
            {
            panic
                (
                "ObjectPointer copy failed. "
                "Pointer flags used to retrieve previous proxy could not be "
                "reduced to requested pointer flags. "
                "This is an internal error."
                );
            }

        #endif
        }
    else
        {
        can_assign( proxy->get_object_id(), proxy->get_object(), pointer_flags );
        }
    
    /* Link proxy with ObjectPointer. */

    proxy->link_pointer( pointer_flags );
    
    return proxy;
    }

ObjectProxy * ObjectProxyManager::get_duplicated_proxy
    (
    ObjectProxy *       proxy,
    int                 pointer_flags
    ) RELEASE_NO_THROW()
    {
    /* Link proxy with ObjectPointer. */

    proxy->link_pointer( pointer_flags );
    
    return proxy;
    }

ObjectProxy * ObjectProxyManager::get_null_proxy
    (
    int                 pointer_flags
    ) RELEASE_NO_THROW()
    {
    /* Link proxy with ObjectPointer. */

    null_proxy.link_pointer( pointer_flags );

    return &null_proxy;
    }

void ObjectProxyManager::release_proxy
    (
    ObjectProxy *       proxy,
    int                 pointer_flags
    ) RELEASE_NO_THROW()
    {
    proxy->unlink_pointer( pointer_flags );
    }

bool ObjectProxyManager::can_reduce_pointer_flags
    (
    int pointer_flags,
    int new_pointer_flags
    )
    {
    if( new_pointer_flags == pointer_flags )
        {
        return true;
        }
    else
    if( flags_on( new_pointer_flags, WEAK_POINTER ) )
        {
        return true;
        }

    return false;
    }


// Inline implementation of ActiveObjectProxy
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ActiveObjectProxy::active_pin
    (
    ObjectProperty * object
    )
    {
    object->active_pin();
    }

void ActiveObjectProxy::active_unpin
    (
    ObjectProperty * object
    )
    {
    object->active_unpin();
    }


} // namespace Massiv::Core
} // namespace Massiv
