/** @file objectid.cpp
 * ObjectId implementation.
 */

/*
 * begin        : Sun Dec 16 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "pointer/objectid.h"
#include "global/global.h"
#include "utility/text_writer.h"
#include "utility/text_reader.h"
#include "serializer/serializer.h"
#include "serializer/stype_int.h"
#include "serializer/stype_string.h"
#include "pointer/garbage_collector.h"
#include "utility/debug.h"


namespace Massiv {
namespace Core {


ObjectId::ObjectId
    (
    const NodeId &  node_id,
    ClassTypeId     class_type_id    
    )
    {
    hope( ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT >= ( sizeof( NodeId ) * 8 ) );

    // Clear id.

    set_null();

    // Initialize id.
    
    components.provider_id = 0;
    components.provider_relative_id = NodeId::convert_to_uint32( node_id );
    components.tracked_by_provider = 0;
    components.class_kind = class_type_id.get_kind();
    components.class_type_id = class_type_id.get_class_id();
    
    if( is_object_info_available() )
        {
        components.no_archive = 1; // NodeObjects must have "no_archive" attribute.
        }
    else
        {
        // ObjectInfo is not available. No one should access no_archive flag
        // because no_archive flag is part of ObjectInfo.
        // So we don't need to care of its value.
        components.no_archive = 0;
        }

    // Proxy objects may not be tracked.
    hope( components.provider_id == 0 );
    hope( !components.tracked_by_provider );
    }


NodeId ObjectId::proxy_extract_node_id() const
    {
    hope( is_proxy_node_object_id() );
    NodeId node_id = NodeId::create_from_uint32( components.provider_relative_id );
    return node_id;
    }


void ObjectId::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    // During bit serialization, optional part stays unchanged.
    SUInt64 s_id( id );
    s_id.bitstr_write( bos, desc );
    }


void ObjectId::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    // During bit serialization, optional part stays unchanged.
    SUInt64 s_id;
    s_id.bitstr_read( bis, desc );
    id = s_id;
    }


void ObjectId::text_write
    (
    TextWriter &                    tw,
    const Serializer::Description & desc
    ) const
    {
    tw.line() << "[ ";

    if( is_null() )
        {
        // Let's just hope no-one defines class called "null" in root namespace.

        tw.line() << "null";
        }
    else
        {
        // Serialize the components.

        SInt<UInt64, ProviderId::PROVIDER_ID_BITS_COUNT>
            provider_id( components.provider_id );
            
        SInt<UInt64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT>
            provider_relative_id( components.provider_relative_id );
            
        SInt<UInt64, 1>
            tracked_by_provider( components.tracked_by_provider );

        provider_id.text_write( tw, desc );

        tw.write_space();
        provider_relative_id.text_write( tw, desc );

        tw.write_space();
        tracked_by_provider.text_write( tw, desc );
        
        // Write optional ObjectInfo part.

        if( is_object_info_available() )    
            {
            // Optional part is available, but we don't know if its class
            // is supported by local node.

            ObjectInfo object_info = get_object_info();
            
            if( object_info.is_class_registered() )
                {
                // The class is supported by local node.

                tw.write_space();
                object_info.text_write( tw, desc );
                }
            }
        }

    // Done.

    tw.line() << " ]";
    }
    

void ObjectId::text_read
    (
    TextReader &                    tr,
    const Serializer::Description & desc
    )
    {
    SString         look_ahead;

    set_null();

    // Begin.

    look_ahead.text_read( tr, desc );

    if( look_ahead == "[" )
        {
        std::istream::pos_type position = tr.line().tellg();
        look_ahead.text_read( tr, desc );

        if( look_ahead != "null" )
            {
            tr.line().seekg( position );
            tr.line().clear();

            // Id components.

            SInt<UInt64, ProviderId::PROVIDER_ID_BITS_COUNT>            provider_id;
            SInt<UInt64, ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT>   provider_relative_id;
            SInt<UInt64, 1>                                             tracked_by_provider;

            provider_id.text_read( tr, desc );
            provider_relative_id.text_read( tr, desc );
            tracked_by_provider.text_read( tr, desc );
            
            // Read optional ObjectInfo part.
            
            position = tr.line().tellg();
            look_ahead.text_read( tr, desc );
            
            if( look_ahead != "]" )
                {
                // Optional part is available.
                
                tr.line().seekg( position );
                tr.line().clear();
                
                ObjectInfo  object_info;
                object_info.text_read
                    (
                    tr,
                    desc,
                    ObjectInfo::TEXT_READ_IGNORE_INVALID_CLASS_TYPE_ID
                    );
                
                // End.

                tr.require_string( "]" );
                
                components.class_kind = object_info.get_class_type_id().get_kind();
                components.class_type_id = object_info.get_class_type_id().get_class_id();
                components.no_archive = !object_info.is_archivable();
                }
            else
                {
                // Optional part is not available.

                hope( components.class_kind == 0 );
                hope( components.class_type_id == 0 );
                hope( components.no_archive == 0 );
                }

            components.provider_id = provider_id;
            components.provider_relative_id = provider_relative_id;
            components.tracked_by_provider = tracked_by_provider;

#ifdef OPTION_DEBUG
            // If object info part was read, it must contain a registered class.
            if( is_object_info_available() )
                {
                hope( get_object_info().is_class_registered() );
                }
#endif // OPTION_DEBUG
            }
        else
            {
            // End.
            tr.require_string( "]" );
            }
        }
    else
        {
        // Lookup local object names database.

        GarbageCollector & gc = Global::garbage_collector();
        const GarbageCollector::NameToObjectIdMap & map = gc.get_names_to_objects();
        GarbageCollector::NameToObjectIdMap::const_iterator it = map.find( look_ahead );
        
        if( it == map.end() )
            {
            throw ExUnexpectedValue( tr.unexpected_value( "'[' or 'object name'", look_ahead ) );
            }

        operator=( it->second );
        }
    }


int ObjectId::memory_size() const
    {
    return sizeof( *this );
    }
    

int ObjectId::bitstream_size() const
    {
    return 64;
    }
    
    
std::string ObjectId::to_string() const
    {
    std::ostringstream ss;
    ss << *this;
    return ss.str();
    }


std::ostream & operator<<
    (
    std::ostream &    s,
    const ObjectId &  object_id
    )
    {
    // Lookup local object name.

    GarbageCollector & gc = Global::garbage_collector();
    const GarbageCollector::ObjectIdToNameMap & map = gc.get_objects_to_names();
    GarbageCollector::ObjectIdToNameMap::const_iterator it = map.find( object_id );

    if( it != map.end() )
        {
        // Print object name.

        s << it->second;
        }
    else
        {       
        // Print general object id.

        TextWriter tw( s );
        object_id.text_write( tw, Serializer::Description() );
        }

    return s;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "pointer/objectid_inline.h"
#endif
