/** @file objectid.h
 * Object id management.
 */

/*
 * begin        : Sun Dec 16 2001
 * copyright    : (C) 2001, 2002 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_OBJECTID_H
#define MASSIV_OBJECTID_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_OBJECT_INFO_H
#include "object/object_info.h"
#endif

#ifndef MASSIV_PROVIDERID_H
#include "object/providerid.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#include <iosfwd>


namespace Massiv {
namespace Core {


/**@class ObjectId
 * Class that uniquely identifies an object within the simulated world.
 * ObjectId consists of two parts. The first part is mandatory, the second
 * part is optional.
 *
 * The first part of ObjectId contains information that is needed
 * to properly localize the object in the simulation.
 * It consists of provider id, object id relative to the provider and
 * tracked-by-provider flag.
 * Provider id identifies a "provider", the system component that generated
 * the object id. Provider-relative id is the id that was actualy generated
 * by the provider. This seperation allows providers to generate object ids
 * independently on each other.
 * The tracked-by-provider flags specifies whether the provider monitors
 * location of the object in the simulation.
 *
 * The second part of ObjectId (#ObjectInfo) is optional and contains
 * additional information about the object like #ClassTypeId (type
 * information) and other miscellanneous flags.
 * See #ObjectInfo for more details.
 * Since we support class kinds and separate compilation the optional part
 * may not be interpreted by all nodes properly (node does not understand
 * type information of an alien class, for example). If such node can not
 * interpret the additional information it either discards/truncates it or
 * leaves it uninterpreted. The truncation may happen during textual
 * serialization only (the node is not able to read/write ClassTypeId in a
 * textual format). Bitstream serialization leaves the information
 * uninterpreted, no information loss happens.
 *
 * To access #ObjectInfo part use @c get_object_info() method. Note that
 * the info may not be interpretable by current node.
 *
 * Since #ObjectInfo part of the ObjectId is optional, its hash, equalkey
 * and less functors ignore it and use mandatory part of the ObjectId only.
 *
 * ObjectId is serializable.
 *
 * There is a database of named objects that allows to resolve object ids
 * by their "names".  The database is managed by the Garbage Collector
 * and is local to every node. The database is not a "naming service"
 * (object id resolution does not cooperate with another nodes).
 * Object names can be used in textual ObjectId representations
 * (the textual serializer looks up the database) under a condition
 * that the node, that will be reading the textual ObjectId, will be able
 * to resolve object id by searching its local database.
 *
 * @ingroup pointer
 * @ingroup objectid
 */
class MASSIV_EXPORTED ObjectId
    {
    friend class ObjectProvider;

public:

    /**
     * ObjectId hash function.
     * Ignores optional object info part of ObjectId.
     */
    struct MASSIV_EXPORTED Hash
        {

        INLINE size_t operator()
            (
            const ObjectId & object_id
            ) const throw();

        }; // struct Hash

    /**
     * ObjectId equalkey function.
     * Ignores optional object info part of ObjectId.
     */
    struct MASSIV_EXPORTED EqualKey
        {

        INLINE bool operator()
            (
            const ObjectId & object_id1,
            const ObjectId & object_id2
            ) const throw();

        }; // struct EqualKey

    /**
     * ObjectId less function.
     * Ignores optional object info part of ObjectId.
     */
    struct MASSIV_EXPORTED Less
        {

        INLINE bool operator()
            (
            const ObjectId & object_id1,
            const ObjectId & object_id2
            ) const throw();

        }; // struct Less

    friend struct ObjectId::Hash;
    friend struct ObjectId::EqualKey;
    friend struct ObjectId::Less;

public:

    enum
        {
        NULL_ID = 0
            /**< Id of a "NULL" object. */
        };

public:

    /**
     * Default constructor.
     * This constructs @c NULL_ID object id.
     */
    INLINE ObjectId();
    
    /**
     * Create ObjectId of the node object that represents node @c node_id.
     * See @c NodeObjectInterface.
     * @param class_type_id Optional parameter. Specifies #ClassTypeId of
     *      the node object. If not specified, created ObjectId will not
     *      contain the optional #ObjectInfo part.
     * @note If class type id is specified, "no-archive" flag is set to true,
     *      because node objects can't be archivable objects.
     */
    ObjectId
        (
        const NodeId &  node_id,
        ClassTypeId     class_type_id = ClassTypeId()
        );

    /**
     * Copy constructor.
     */
    INLINE ObjectId
        (
        const ObjectId & object_id
        );

    /**
     * Operator=.
     */
    INLINE ObjectId & operator=
        (
        const ObjectId & object_id
        );

public:

    /**
     * Operator==.
     */
    INLINE bool operator==
        (
        const ObjectId & object_id
        ) const;

    /**
     * Operator!=.
     */
    INLINE bool operator!=
        (
        const ObjectId & object_id
        ) const;

public:

    /**
     * Sets to null.
     */
    INLINE void set_null();

    /**
     * Checks whether is null (@c NULL_ID).
     */
    INLINE bool is_null() const;
    
    /**
     * Clear optional #ObjectInfo part
     * (@c is_object_info_available() will return false).
     */
    INLINE void clear_optional_part();

    /**
     * Set #no_archive flag to given value.
     *
     * This value must not be changed once object is fully initialized.
     * So change it in @c initialize() method and no later.
     */
    INLINE void set_no_archive
        (
        bool no_archive
        );

public:

    /**
     * Returns true if location of the object is monitored by its provider.
     */
    INLINE bool is_tracked_by_provider() const;
    
    /**
     * Returns true if optional #ObjectInfo part of the #ObjectId
     * is available.
     * @warning Even if #ObjectInfo is available, it doesn't imply
     *      that the information can be interpreted by current node.
     */
    INLINE bool is_object_info_available() const;
    
    /**
     * Get #ClassTypeId stored in the #ObjectId.
     * @note If #ObjectInfo is not available, returned ClassTypeId
     *      will refer to a "void" class. To test whether #ObjectInfo
     *      is available, use @c is_object_info_available().
     * @warning Even if #ObjectInfo is available, it doesn't imply
     *      that class type id stored in the #ObjectInfo is supported
     *      by local node. Caller must check this by itself.
     */
    INLINE ClassTypeId get_class_type_id() const;
    
    /**
     * Get #ObjectInfo stored in the #ObjectId.
     * @pre #ObjectInfo is available. To detect whether #ObjectInfo
     *      is available, use @c is_object_info_available() function.
     * @warning Even if #ObjectInfo is available, it doesn't imply
     *      that the information can be interpreted by current node.
     */
    INLINE ObjectInfo get_object_info() const;
    
public:

    /**
     * Returns provider id.
     */
    INLINE ProviderId get_provider_id() const;
    
    /**
     * Returns true if the object is proxy node object id.
     */
    INLINE bool is_proxy_node_object_id() const;
    
    /**
     * Extract node id from the proxy object id.
     * @pre The object id is proxy node object id.
     */
    NodeId proxy_extract_node_id() const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Serialize to text stream.
     * ObjectId is always written in the most general way,
     * never looks up local object name database.
     * Truncates the optional part if current node is not able to interpret
     * it.
     */
    void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from text stream.
     * Accepts local object names.
     * If the optional part is present and can not be interpeted by current
     * node it is discarded.
     */
    void text_read
        (
        TextReader &                    tr,
        const Serializer::Description & desc
        );

    /**
     * Guess memory size occupied by the object id.
     */
    int memory_size() const;
    
    /**
     * Guess bitstream serialization size.
     */
    int bitstream_size() const;
    
    /**
     * Return object id as a string.
     * The result is exactly what operator<< would write to a stream.
     */
    std::string to_string() const;
        
private:

    /**
     * Get mandatory part of the #ObjectId as @c UInt64.
     * Optional object info part is ignored.
     * Returned value is used by hash, equalkey and less functors.
     */
    INLINE UInt64 get_mandatory_part_as_uint64() const;

private:

    union
        {
        struct
            {
            
            // Mandatory part.
            
            UInt64      provider_id : ProviderId::PROVIDER_ID_BITS_COUNT;
                /**< Id of the provider which generated this id. */
                
            UInt64      provider_relative_id : ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT;
                /**< Provider-relative object id. */
                
            UInt64      tracked_by_provider : 1;
                /**< Set if location of the object is monitored by its provider. */
            
            // Optional part. If it is not available, all components must be zero.
            
            UInt64      class_kind : ClassTypeId::KIND_BITS_COUNT;
                /**< Class kind. Optional part. */
                
            UInt64      class_type_id : ClassTypeId::CLASS_ID_BITS_COUNT;
                /**< Id of class type, unique within all class of the same kind. Optional part. */
            
            UInt64      no_archive : 1;
                /**< Set if this object should not be archived. Optional part.
                 * @note It's "no" archive, because we want standard value to be zero.
                 * @note See #ObjectInfo about in what cases no_archive flag may be changed at runtime.
                 * @note Yes, this is optional part of object id. Optional part is lost only when
                 *       archiving alien object. Objects with #no_archive set will never be archived.
                 *       Objects read from archive will always have #no_archive set to 0. So this
                 *       value is optional, but it's always set correctly.
                 */

            }   components;
            /**< Identifier components. */

        UInt64   id;
            /**< Identifier as a number. */
        };

    }; // ObjectId

    
/**
 * Write object id to output stream.
 * Translates object id to local object name if possible.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &      s,
    const ObjectId &    object_id
    );


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "pointer/objectid_inline.h"
#endif


#endif // MASSIV_OBJECTID_H
