/** @file objectid_inline.h
 * ObjectId implementation.
 */

/*
 * begin        : Sun Dec 16 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_HASH_FUNCTION_H
#include "utility/hash_function.h"
#endif


namespace Massiv {
namespace Core {


// struct ObjectId::Hash

size_t ObjectId::Hash::operator()
    (
    const ObjectId & object_id
    ) const throw()
    {
    // We must ignore optional part.
    UInt64  mandatory_part = object_id.get_mandatory_part_as_uint64();
    return ext::hash_int( mandatory_part );
    }


// struct ObjectId::EqualKey

bool ObjectId::EqualKey::operator()
    (
    const ObjectId & object_id1,
    const ObjectId & object_id2
    ) const throw()
    {
    // We must ignore optional part.
    UInt64  mandatory_part1 = object_id1.get_mandatory_part_as_uint64();
    UInt64  mandatory_part2 = object_id2.get_mandatory_part_as_uint64();
    
    return mandatory_part1 == mandatory_part2;
    }


// struct ObjectId::Less

bool ObjectId::Less::operator()
    (
    const ObjectId & object_id1,
    const ObjectId & object_id2
    ) const throw()
    {
    // We must ignore optional part.
    UInt64  mandatory_part1 = object_id1.get_mandatory_part_as_uint64();
    UInt64  mandatory_part2 = object_id2.get_mandatory_part_as_uint64();
    
    return mandatory_part1 < mandatory_part2;
    }


// class ObjectId
// ~~~~~~~~~~~~~~


ObjectId::ObjectId()
    {
    set_null();
    }


ObjectId::ObjectId
    (
    const ObjectId & object_id
    ) : id( object_id.id )
    {
    }


ObjectId & ObjectId::operator=
    (
    const ObjectId & object_id
    )
    {
    id = object_id.id;
    return *this;
    }


bool ObjectId::operator==
    (
    const ObjectId & object_id
    ) const
    {
    return EqualKey()( *this, object_id );
    }


bool ObjectId::operator!=
    (
    const ObjectId & object_id
    ) const
    {
    return !EqualKey()( *this, object_id );
    }


void ObjectId::set_null()
    {
    id = NULL_ID;
    }


bool ObjectId::is_null() const
    {
    return id == NULL_ID;
    }

    
void ObjectId::clear_optional_part()
    {
    components.class_kind = 0;
    components.class_type_id = 0;
    components.no_archive = 0;
    }


void ObjectId::set_no_archive
    (
    bool no_archive
    )
    {
    components.no_archive = no_archive ? 1 : 0;
    }


bool ObjectId::is_tracked_by_provider() const
    {
    return components.tracked_by_provider != 0;
    }
    

bool ObjectId::is_object_info_available() const
    {
    // Note: We can't use ClassTypeId::is_class_registered() to detect
    // whether ObjectInfo is available because even if ObjectInfo is
    // available, it may refer to a class unsupported by local node.

    // Note: Do not use components.no_archive in this function!
    // See ObjectId( const NodeId & ).

    return components.class_kind != 0 || components.class_type_id != 0;
    }

    
ClassTypeId ObjectId::get_class_type_id() const
    {
    ClassTypeId class_type_id( components.class_kind, components.class_type_id );
    return class_type_id;
    }


ObjectInfo ObjectId::get_object_info() const
    {
    hope( is_object_info_available() );
    return ObjectInfo( get_class_type_id(), components.no_archive == 0 );
    }
    

ProviderId ObjectId::get_provider_id() const
    {
    return ProviderId::create_from_number( components.provider_id );
    }
    
    
bool ObjectId::is_proxy_node_object_id() const
    {
    return !is_null() && components.provider_id == 0;
    }

    
UInt64 ObjectId::get_mandatory_part_as_uint64() const
    {
    UInt64 result;

    // Stoupik's note:
    // Those casts are required by ICC, otherwise the "shift count too large"
    // warning is produced. ICC probably treats components.xxx as int, not
    // as int64.

    result =
        ( static_cast< UInt64 >( components.provider_id ) ) |
        ( static_cast< UInt64 >( components.provider_relative_id ) <<
            ( ProviderId::PROVIDER_ID_BITS_COUNT ) ) |
        ( static_cast< UInt64 >( components.tracked_by_provider ) <<
            ( ProviderId::PROVIDER_ID_BITS_COUNT + ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) );
    
    return result;
    
    /*
    union
        {
        struct
            {
            UInt64      provider_id : ProviderId::PROVIDER_ID_BITS_COUNT;
            UInt64      provider_relative_id : ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT;
            UInt64      tracked_by_provider : 1;
            } components;
            
        Uint64 id;
        } mandatory_part;
    
    mandatory_part.id = 0;
    mandatory_part.components.provider_id = components.provider_id;
    mandatory_part.components.provider_relative_id = components.provider_relative_id;
    mandatory_part.components.tracked_by_provider = components.tracked_by_provider;
    
    return mandatory_part.id;
    */
    }


} // namespace Massiv::Core
} // namespace Massiv
