/** @file pointer_base.h
 * Object pointer base classes.
 */

/*
 * begin        : Sun Dec 15 2001
 * copyright    : (C) 2001, 2002, 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_POINTER_BASE_H
#define MASSIV_POINTER_BASE_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_POINTER_PARAMETERS_H
#include "pointer/pointer_parameters.h"
#endif

#ifndef MASSIV_OBJECT_PROXY_H
#include "pointer/object_proxy.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_GLOBAL_H
#include "global/global.h"
#endif

#ifdef OPTION_DEBUG

#ifndef MASSIV_DEBUG_H
#include "utility/debug.h"
#endif

#endif

#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup pointer
 * @{
 */

/**
 * Common pointer base that stores pointer-type neutral data.
 * See @c PointerBase<Cast> which adds pointer-type specific attributes.
 *
 * The inheritance hierarchy for ObjectPointer looks like this:
 *
 * CommonPointerBase [ pointer type neutral base ] <-
 * PointerBase<Cast> [ pointer type specific base ] <-
 * {Stack | Property}PointerBase<Traits> [ pointer traits specific base ] <-
 * ObjectPointer<Base> <-
 * PPointer | PWeakPointer | PRemote | Pointer | WeakPointer | Remote
 * [ shorthands for ObjectPointers with predefined Bases ]
 *
 * Since CommonPointerBase is the only class in the hierarchy that is not
 * a template, core uses CommonPointerBase in its interfaces where any
 * ObjectPointer should be accepted.
 */
class MASSIV_EXPORTED CommonPointerBase
    {
private:

    NO_COPY_CONSTRUCTOR( CommonPointerBase );
    
    friend class GarbageCollector;

protected:
        
    /**
     * Default constructor.
     * Subclass is expected to initialize @c proxy to something valid.
     */
    CommonPointerBase() : proxy( NULL )
        {
        }

    /**
     * Destructor.
     * Subclass is expected to release @c proxy.
     */
    ~CommonPointerBase()
        {
        }

protected:

    /**
     * Throw serialization exception if @a desc is description of
     * client to server migration and @a object_id is non-null.
     *
     * This hacky method is called from ObjectPointer::bitstr_read()
     * if it's a strong pointer.
     */
    static void check_strong_pointer_object_id
        (
        const Serializer::Description & desc,
        const ObjectId &                object_id
        );
    
    /**
     * Returns object proxy manager.
     */
    static ObjectProxyManager & manager()
        {
        return Global::object_proxy_manager();
        }

protected:
        
    ObjectProxy * proxy;
        /**< Proxy object. */
    
    }; // CommonPointerBase

/**
 * Pointer base implementation, with pointer-type neutral code in
 * @c CommonPointerBase.
 * The implementation is further refined by @c StackPointerBase
 * and @c PropertyPointerBase, which are derived from
 * @c PointerBase.
 * ObjectPointer stores @c proxy (in its CommonPointerBase) which is a pointer
 * to proxy object that represents component (defined via @c cast parameter)
 * of the referenced object. Proxy object can further be linked to the
 * referenced object, if the object is registered on the local system.
 * Each proxy is identified by @c object_id of the referenced object
 * and its component (@c cast parameter).
 * There is a special "null proxy" that represents @c NULL_ID object.
 * Null proxy can be shared by all @c NULL_ID ObjectPointers regardless
 * of their @c cast parameters.
 */
template <class Cast>
class PointerBase : public CommonPointerBase
    {
protected:

    static Cast cast;
        /**< Cast functor. This should be a global singleton object.
             @warning If core is linked as a DLL, @c PointerBase<Cast>::cast
             may be linked to the target address space multiple times, because
             we do not export/import it. Thus in a DLL @c cast is no longer
             a singleton object. */

    }; // PointerBase

template <class Cast>
Cast PointerBase<Cast>::cast;

/**
 * This class defines implementation refinement for ObjectPointers
 * placed on program stack.
 * @c Traits parameter specifies pointer characteristics (pointer flags,
 * cast object, dereference semantics), it is either @c WeakPointerTraits
 * or @c StrongPointerTraits.
 */
template <class Traits>
class StackPointerBase : public PointerBase<typename Traits::Cast>
    {
public:

    typedef typename Traits::Type       PointerType;
    typedef typename Traits::Cast       PointerCast;
    typedef typename Traits::ProxyType  PointerProxy;
    typedef Traits                      PointerTraits;
    typedef StackPointerBase<Traits>    ThisBase;

    enum
        {
        PointerFlags = Traits::Flags
        };
        
protected:
    
    using PointerBase<PointerCast>::cast;
    using PointerBase<PointerCast>::manager;
    using PointerBase<PointerCast>::proxy;

protected:

    //
    // Implementation.
    //

    /** Constructor. */
    StackPointerBase()
        {
        }

    /**
     * Copy constructor.
     * Does the same initialization as default constructor.
     */
    StackPointerBase
        (
        const ThisBase & base
        )
        {
        }

    /** Destructor. */
    ~StackPointerBase()
        {
        manager().release_proxy( proxy, get_pointer_flags() );
        }

public:

    /** Returns pointer flags. */
    int get_pointer_flags() const
        {
        return PointerFlags | STACK_POINTER;
        }

    /** Stack pointer is always on program stack. */
    bool is_on_stack() const
        {
        return true;
        }

    /** Is this pointer weak? */
    bool is_weak() const
        {
        return get_pointer_flags() & WEAK_POINTER;
        }

protected:

    /** Called by ObjectPointer before writing to @c StackPointerBase. */
    void before_write()
        {
        }

    /** Called by ObjectPointer before reading from @c StackPointerBase. */
    void before_read() const
        {
        }

    /**
     * Called by ObjectPointer to notify that the pointer was
     * default-constructed.
     */
    void default_constructed()
        {
        }

    /** Called by ObjectPointer to replace object proxy. */
    void replace_proxy
        (
        ObjectProxy * new_proxy
        )
        {
        manager().release_proxy( proxy, get_pointer_flags() );
        proxy = new_proxy;
        }

    /** Called by ObjectPointer to initialize its first link to object proxy. */
    void set_proxy
        (
        ObjectProxy * new_proxy
        )
        {
        hope( !proxy );
        proxy = new_proxy;
        }

public:

    /** Dumps stack pointer value. */
    friend std::ostream & operator<<
        (
        std::ostream &      stream,
        const ThisBase &    base
        )
        {
        stream << base.proxy->get_object_id();

        if( base.is_weak() )
            {
            stream << " [ weak ]";
            }

        return stream;
        }
    
    /** Returns printable pointer state. */
    std::string to_string() const
        {
        std::ostringstream s;
        
        s << *this;

        return s.str();
        }

    }; // StackPointerBase

/**
 * Abstract interface implemented by @c PropertyPointerBase.
 * Accesses information about a property pointer.
 */
class PointerInfo
    {
public:

    /** Does nothing. Defined just to get rid of GCC warnings. */
    virtual ~PointerInfo();

    /** Returns @c object_id of the referenced object. */
    virtual const ObjectId & ptr_get_object_id() const = 0;

    /** Returns pointer to referenced object. */
    virtual ObjectProperty * ptr_get_object() const = 0;

    /** Returns @c pointer_flags. */
    virtual int ptr_get_pointer_flags() const = 0;

    /** Returns @c ptr_replication_flags. */
    virtual int ptr_get_replication_flags() const = 0;

    /** Returns pointer to corresponding property. */
    virtual const Property * ptr_get_property() const = 0;

    /** Updates/relinks backward pointer. Invoked by GC. */
    virtual void update_backward_pointer() = 0;

    /**
     * Tests if backward pointer is up-to-date (properly synchronized
     * with the (forward) pointer property.
     * Invoked by GC to test lazy pointer updates consistency.
     */
    virtual bool is_backward_pointer_up_to_date() const = 0;

    /**
     * Marks referenced object. Invoked by GC when it is running
     * mark phase.
     */
    virtual void mark_referenced_object() = 0;

    }; // PointerInfo

/**
 * This class defines implementation refinement for ObjectPointer
 * properties. Pointer properties store replication flags,
 * that determine when referenced object should be replicated/migrated.
 * Pointer properties are "monitored" which allows to enumerate
 * objects connected by pointers having specific replication flags.
 * Each pointer property is paired / coupled with a backward pointer
 * (whose value must be synchronized with pointer property's value),
 * so that the enumeration could treat pointers as bidirectional.
 * @c Traits parameter specifies pointer characteristics,
 * it is either @c WeakPointerTraits or @c StrongPointerTraits.
 */
template <class Traits>
class PropertyPointerBase :
    public PointerBase<typename Traits::Cast>,
    public PointerInfo,
    public Property
    {
public:

    typedef typename Traits::Type           PointerType;
    typedef typename Traits::Cast           PointerCast;
    typedef typename Traits::ProxyType      PointerProxy;
    typedef Traits                          PointerTraits;
    typedef PropertyPointerBase<Traits>     ThisBase;

    enum
        {
        PointerFlags = Traits::Flags
        };

protected:
    
    using PointerBase<PointerCast>::cast;
    using PointerBase<PointerCast>::manager;
    using PointerBase<PointerCast>::proxy;

protected:

    //
    // Implementation.
    //

    /** Constructor. */
    PropertyPointerBase() :
        flags( PointerFlags | STACK_POINTER ),
        ptr_replication_flags( PointerFlags & WEAK_POINTER ? 0 : REPFLAGS_MIGRATE )
        {
        }

    /**
     * Copy constructor.
     * Does the same initialization as default constructor.
     * Calls Property copy constructor.
     */
    PropertyPointerBase
        (
        const ThisBase & base
        ) : Property( base ),
            flags( PointerFlags | STACK_POINTER ),
            ptr_replication_flags( PointerFlags & WEAK_POINTER ? 0 : REPFLAGS_MIGRATE )
        {
        }

    /** Destructor. */
    ~PropertyPointerBase()
        {
        // Pointer destruction may change migration group
        // of pointer owner object (1) and referenced object (2).
        // (1) was already archived if it was in archive-on-write mode.
        // (2) is informed about possible group changed by
        // @c unlink_backward_pointer().

        proxy->unlink_backward_pointer( backward_pointer );
        manager().release_proxy( proxy, get_pointer_flags() );
        }

public:

    /** Returns pointer flags. */
    int get_pointer_flags() const
        {
        return flags;
        }

    /**
     * Is this pointer weak?
     * Pointer replica is always weak.
     */
    bool is_weak() const
        {
        return get_pointer_flags() & WEAK_POINTER;
        }

protected:

    /** Called by ObjectPointer before writing to @c PropertyPointerBase. */
    void before_write()
        {
        Property::write_requested();
        }

    /** Called by ObjectPointer before reading from @c PropertyPointerBase. */
    void before_read() const
        {
        Property::read_requested();
        }
    
    /**
     * Called by ObjectPointer to notify that the pointer was
     * default-constructed.
     */
    void default_constructed()
        {
        Property::default_constructed();
        }

    /**
     * Called by ObjectPointer to replace object proxy when
     * pointer value changes.
     * Synchronizes corresponding backward pointer.
     * Called by ObjectPointer's operator=().
     */
    void replace_proxy
        (
        ObjectProxy * new_proxy
        )
        {
        test_policy( new_proxy );

        if( proxy == new_proxy )
            {
            // Pointer value does not change.
            // Release the new proxy.

            manager().release_proxy( new_proxy, get_pointer_flags() );
            return;
            }

        // Changing pointer value may change migration group
        // of pointer owner object (1), referenced object (2)
        // and the new object (3).
        // (1) was already notified about the change by calling
        // @c write_requested() at @c before_write().
        // (2) and (3) are notified by @c (lazy_)relink_backward_pointer().
        // If the objects are in archive-on-write mode
        // they must be archived now. This is done automatically
        // when backward pointer is relinked.

        #ifdef GC_LAZY_POINTERS
        proxy->lazy_relink_backward_pointer( new_proxy, this );
        #else
        proxy->relink_backward_pointer( new_proxy, backward_pointer );
        #endif

        manager().release_proxy( proxy, get_pointer_flags() );
        proxy = new_proxy;
        }

    /**
     * Called by ObjectPointer to initialize its first link to object proxy.
     * It is assumed that no proxy is currently linked.
     * Initializes corresponding backward pointer.
     * Called from ObjectPointer's contructors.
     * @note Pointer property is currently initialized "on stack".
     */
    void set_proxy
        (
        ObjectProxy * new_proxy
        )
        {
        hope( !proxy );
        
        test_policy( new_proxy );

        // Setting proxy changes migration group of pointer owner object (1)
        // and referenced object (2).
        // (1) need not be informed as it is StackRootObject.
        // (2) is informed by @c link_backward_pointer().

        proxy = new_proxy;
        proxy->link_backward_pointer( this, backward_pointer );
        }

public:

    //
    // Property extensions.
    // No read/write handlers get called.
    //

    /**
     * Initialize property flags from its owner and set per-property
     * attributes. @c Attributes::ptr_repflags is handled here.
     *
     * Pointers require special handling of #replica flag.
     * ObjectPointer replicas are always weak (ie. static pointer
     * type is ignored). This allows to define replication groups
     * as subsets of migration groups without violating the rule
     * that strong pointers must point to local objects only.
     * Only weak pointers can point to object replicas.
     *
     * @note Since this may change pointer semantics, the policy must
     *       be retested.
     *
     * @note Since this may change pointer owner, the change possibly results
     *       in migration group change.
     *
     * @note Because the pointer will be referencing the same object as before,
     *       backward pointer need not be updated.
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        )
        {
        int old_flags = get_pointer_flags();

        // Notify referenced object, the old owner and the new owner that
        // their migration groups may change.

        proxy->changing_pointer_owner( this, &direct_owner );

        // Initialize owner and basic per-property attributes.

        Property::initialize_from_owner( direct_owner, attributes );
        
        // Update pointer state.
        
        if( is_replica() )
            {
            // All pointer replicas are considered weak, but ptr_repflags may still
            // have REPFLAGS_MIGRATE bit set.

            hope( !is_on_stack() );
                
            flags |= WEAK_POINTER;
            flags &= ~STACK_POINTER;
                
            ptr_replication_flags = attributes.ptr_repflags;
            }
        else
        if( is_on_stack() )
            {
            flags |= STACK_POINTER;
            
            // Use default @c ptr_replication_flags according to static
            // pointer type.

            ptr_replication_flags = PointerFlags & WEAK_POINTER ? 0 : REPFLAGS_MIGRATE;
            }
        else
            {
            flags &= ~STACK_POINTER;
            
            // Set @c ptr_replication_flags according to @c attributes.

            ptr_replication_flags = attributes.ptr_repflags;
            
            #ifdef OPTION_DEBUG

            if( !is_weak() && !flags_on( ptr_replication_flags, REPFLAGS_MIGRATE ) )
                {
                panic( ( std::string( "Strong pointer property " ) + 
                    get_info().property_name + " has an incorrect ptr_repflags "
                    "IDL attribute. At least REPFLAGS_MIGRATE must be set." ) .c_str() );
                }

            #endif
            }

        proxy->link_pointer( get_pointer_flags() );
        proxy->unlink_pointer( old_flags );

        // Test.

        retest_policy();
        }

    /** Returns TYPE_POINTER. */
    virtual Property::PropertyType get_type() const
        {
        return Property::TYPE_POINTER;
        }

    /** Returns pointer info. */
    virtual PointerInfo * get_pointer_info()
        {
        return this;
        }

    /** Returns printable pointer state. */
    virtual std::string to_string() const
        {
        std::ostringstream s;
        
        s << proxy->get_object_id() << " [ "
          << ( is_weak() ? "weak, " : "" ) << "repflags = "
          << ptr_replication_flags << " ]";

        return s.str();
        }

    /** Returns memory size consumed by the pointer. */
    virtual int memory_size() const
        {
        return sizeof( ThisBase );
        }

    /** Computes pointer hash code. */
    virtual int hash() const
        {
        ObjectId::Hash hasher;
        return hasher( proxy->get_object_id() );
        }

    /** Marks referenced object. Invoked by GC. */
    virtual void mark_referenced_object()
        {
        hope( !is_weak() );
        proxy->mark_object();
        }

public:

    //
    // PointerInfo implementation.
    //

    /**
     */
    virtual const ObjectId & ptr_get_object_id() const
        {
        return proxy->get_object_id();
        }

    /**
     */
    virtual ObjectProperty * ptr_get_object() const
        {
        return proxy->get_object();
        }

    /**
     */
    virtual int ptr_get_pointer_flags() const
        {
        return get_pointer_flags();
        }

    /**
     */
    virtual int ptr_get_replication_flags() const
        {
        return ptr_replication_flags;
        }

    /**
     */
    virtual const Property * ptr_get_property() const
        {
        return this;
        }

    /**
     */
    virtual void update_backward_pointer()
        {
        proxy->update_backward_pointer( backward_pointer );
        }

    /**
     */
    virtual bool is_backward_pointer_up_to_date() const
        {
        hope( backward_pointer->second == static_cast<const Property *>( this ) );
        
        if( backward_pointer->first == proxy->get_object_id() )
            {
            return true;
            }
        else
        if( get_root_owner() == Global::get_stack_root_object() )
            {
            // If pointer property is owned by StackRootObject then its backward
            // pointer is never updated in GC_LAZY_POINTERS mode.
            // That is because StackRootObject is not able to effeciently
            // enumerate all owned pointer properties and perform the delayed
            // update.
            // Backward pointers referencing StackRootObject (= pointer properties
            // owned by StackRootObject) are ignored anyway so this is okay.

            return true;
            }

        return false;
        }

private:

    /**
     * Tests pointer policy for @c new_proxy.
     * To be called by @c set_proxy(), @c replace_proxy().
     * Releases @c new_proxy if violation is detected.
     */
    void test_policy
        (
        ObjectProxy * new_proxy
        )
        {
        #ifdef OPTION_DEBUG
        try
            {
            manager().test_pointer_property_policy( this, new_proxy );
            }
        catch( ... )
            {
            manager().release_proxy( new_proxy, get_pointer_flags() );
            throw;
            }
        #endif
        }

    /**
     * Retests pointer policy for current @c proxy.
     * To be called when pointer characteristics changes or pointer property
     * gets fully initialized.
     */
    void retest_policy()
        {
        #ifdef OPTION_DEBUG
        manager().test_pointer_property_policy( this, proxy );
        #endif
        }

private:

    int                             flags;
        /**< Current pointer flags. */
    int                             ptr_replication_flags;
        /**< Pointer replication flags.
             If pointer is replica, they should be used only when enumeration
             replication groups and ignored otherwise. */
    ObjectShared::BackwardPointer   backward_pointer;
        /**< Backward pointer registration. */

    }; // PropertyPointerBase

/** @} */ // addtogroup pointer

    
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_POINTER_BASE_H
