/** @file pointer_fwd.h
 * Pointer subsystem forward declarations.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_POINTER_FWD_H
#define MASSIV_POINTER_FWD_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_VIRTUAL_TIME_H
#include "utility/virtual_time.h"
#endif

#ifndef MASSIV_NULL_KEYWORD_H
#include "pointer/null_keyword.h"
#endif


namespace Massiv {
namespace Core {


//
// Pointer subsystem implementation setup.
// Allows to balance between "fast pointers" and "fast GC runs", does not
// change basic pointer / GC semantics.
//

#ifndef OPTION_OPTIMIZE

/** @def GC_MARK_PHASE_SPANNED
 * Can GC mark phase last more than one GC tick?
 * If enabled, writes to pointers (pointer changes) would be slower, but
 * system would not be blocked when running mark phase.
 *
 * GC_MARK_PHASE_SPANNED
 *      - <tt>+</tt> no system blocking when GC scans for reachable objects
 *      - <tt>-</tt> pointer writes penalty
 *
 * !GC_MARK_PHASE_SPANNED
 *      - <tt>-</tt> the whole GC run is performed at one GC tick (thus system may
 *        be blocked for a while)
 *      - <tt>+</tt> faster pointer writes
 *
 * @ingroup pointer
 */
#define GC_MARK_PHASE_SPANNED

#endif

/** @def GC_DELAYED_UNPIN
 * Objects are pinned when pointers to them are dereferenced to prevent
 * GC from deleting them (there may be "unreferenced" active objects,
 * that are referenced by native pointers only).
 *
 * This define determines when objects are unpinned. Either at the end
 * of "activation era" (typically next GC tick, objects are pinned
 * conservatively until then) or immediatelly once the native reference is 
 * released.
 * The negative property of longer pin times (that are often conservatively
 * overestimated) are unwanted migration stalls (objects can not migrate
 * if they are pinned).
 * Non-delayed unpinning adds a non-zero overhead as core needs to know
 * when the last native reference was released.
 *
 * GC_DELAYED_UNPIN:
 *      - <tt>+</tt> faster pointer dereferencing
 *      - <tt>-</tt> migration stalls
 *
 * !GC_DELAYED_UNPIN:
 *      - <tt>-</tt> slower pointer dereferencing
 *      - <tt>+</tt> no migration stalls
 *
 * @ingroup pointer
 */
#define GC_DELAYED_UNPIN

/** @def GC_LAZY_POINTERS
 * Pointer properties are treated as bidirectional when enumerating migration
 * groups of objects. This define controls when to update corresponding
 * "backward pointer" paired with a ("forward") pointer property. Either
 * immediatelly, when pointer property is changed, or later, when migration
 * group enumeration takes place.
 *
 * Should always be on unless debugging migration group enumeration or the
 * lazy algorithm itself.
 * On OPTION_DEBUG GC automatically does additional consistency tests,
 * which validates the lazy algorithm.
 *
 * GC_LAZY_POINTERS
 *      - <tt>+</tt> faster pointer property changes
 *      - <tt>-</tt> need to update backward pointers at some time (which is a blocking 
 *        operation, however fast enough not to be perceived)
 *
 * !GC_LAZY_POINTERS
 *      - <tt>-</tt> slower pointer property changes
 *      - <tt>+</tt> no blocking updates
 *
 * @ingroup pointer
 */
#define GC_LAZY_POINTERS

#ifndef OPTION_OPTIMIZE

/** @def GC_TAG_STACK
 * Define this to enable core to lib upcall tagging which can be used for
 * debugging purposes.
 *
 * @ingroup pointer
 */
#define GC_TAG_STACK

#endif

//
// Forward declarations.
//

// default_cast.h

class Object;

template <class Type>
class DefaultCast;

// object_proxy.h

class ObjectProxy;

template <class Iface>
class LocalProxy;

template <class Iface>
class RemoteProxy;

// pointer_parameters.h

class CastObject;

template <class PtrType, class PtrCast, class PtrProxy>
struct WeakPointerTraits;

template <class PtrType, class PtrCast, class PtrProxy>
struct StrongPointerTraits;

// pointer_wrappers.h

template
    <
    class   Type,
    class   Cast    = DefaultCast<Type>
    >
class Pointer;

template
    <
    class   Type,
    class   Cast    = DefaultCast<Type>
    >
class WeakPointer;

template
    <
    class   Type,
    class   Cast    = DefaultCast<Type>
    >
class PPointer;

template
    <
    class   Type,
    class   Cast    = DefaultCast<Type>
    >
class PWeakPointer;

template
    <
    class   Type
    >
class Remote;

template
    <
    class   Type
    >
class PRemote;

// pointer_base.h

class CommonPointerBase;

template <class Cast>
class PointerBase;    

template <class Traits>
class StackPointerBase;

template <class Traits>
class PropertyPointerBase;

class PointerInfo;

// object_pointer.h

template <class Base>
class ObjectPointer;

// pointer_to_object.h

class PtrToObject;
class UpcallToObject;

// garbage_collector.h

typedef VirtualTime<Int32> GCEpoch;

class GarbageCollector;


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_POINTER_FWD_H
