/** @file pointer_parameters.h
 * Object pointer parameters.
 */

/*
 * begin        : Tue Apr 23 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_POINTER_PARAMETERS_H
#define MASSIV_POINTER_PARAMETERS_H

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup pointer
 * @{
 */

//
// Pointer flags.
//

/**@enum PointerFlags
 * Common pointer flags.
 */
enum PointerFlags
    {
    WEAK_POINTER        = 0x01,
        /**< Weak pointer.
             Garbage collector can destroy referenced object.
             Weak pointers are allowed to point to remote objects. */
    STACK_POINTER       = 0x02,
        /**< This pointer is placed on program stack.
             Program stack is treated as implicit GC root object. */

    POINTER_FLAGS_BITS  = 2,
        /**< Number of bits required to store @c PointerFlags. */
    POINTER_FLAGS_COUNT = 1 << POINTER_FLAGS_BITS
        /**< Maximum number of pointer flags combinations. */
    }; // PointerFlags

//
// Cast object.
//

/**
 * @c CastObject is an interface for a functor object whose purpose is
 * to convert pointer to @c ObjectProperty to a pointer to some derived
 * class.
 *
 * We use cast objects instead of hard-coded dynamic_casts/static_casts
 * so that client could implement its own conversion mechanisms and
 * resolve possible ambiguities in multiple inheritance hierarchies if needed.
 * For example dynamic_cast will fail if target type is ambigous on that
 * object instance. The ambiguity can be resolved by performing the conversion
 * in more steps while each step is unique (ie. downcast to a derived class,
 * then upcast to parent classes as needed).
 *
 * Used by ObjectPointer to specify components of referenced objects.
 *
 * @code
 * class A : Object { };
 * class B1 : public A { };
 * class B2 : public A { };
 * class C : public B1, B2 { };
 *  // C inherits A twice (through B1 and B2),
 *  // dynamic_cast to A will fail if called on an instance of C.
 *
 * class AThroughB1Cast : public CastObject
 *  {
 *  virtual VariantPointer operator()( ObjectProperty * object ) const
 *  { return static_cast<A *>( static_cast<B1 *>( dynamic_cast<C *>( object ) ) ); }
 *  virtual const std::type_info & get_target_type_info() const
 *  { return typeid( A ); }
 *  };
 * @endcode
 */
class CastObject
    {
public:

    /**
     * Destructor.
     *
     * Defined just to get rid of annoying GCC warnings.
     */
    virtual ~CastObject();

    /**
     * Performs the cast.
     * Returns NULL pointer if conversion was ambiguous.
     */
    virtual VariantPointer operator()
        (
        ObjectProperty * object
        ) const = 0;

    /**
     * Returns type of the pointer returned from operator()
     * ( !! typeid( T ) !! ).
     */
    virtual const std::type_info & get_target_type_info() const = 0;

    }; // CastObject

//
// Pointer traits.
// 

/**
 * Weak pointer characteristics.
 * Serves as @c Traits parameter for StackPointerBase/PropertyPointerBase.
 */
template <class PtrType, class PtrCast, class PtrProxy>
struct WeakPointerTraits
    {
    typedef PtrType     Type;
    typedef PtrCast     Cast;
    typedef PtrProxy    ProxyType;
    
    enum
        {
        Flags = WEAK_POINTER
        };
    }; // WeakPointerTraits

/**
 * Strong pointer characteristics.
 * Serves as @c Traits parameter for StackPointerBase/PropertyPointerBase.
 */
template <class PtrType, class PtrCast, class PtrProxy>
struct StrongPointerTraits
    {
    typedef PtrType     Type;
    typedef PtrCast     Cast;
    typedef PtrProxy    ProxyType;

    enum
        {
        Flags = 0
        };    
    }; // StrongPointerTraits

/** @} */ // addtogroup pointer


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_POINTER_PARAMETERS_H
