/** @file pointer_to_object.cpp
 * Pointer to Object.
 * Notes on UpcallToObject:
 *  UpcallToObject purpose is to get a list of "upcalled" objects and
 *  determine the "top-most" one (with respect to the current thread).
 *  Because of this the registrations must be handled on a per-thread
 *  basis.
 *  Registration code simply registers the instance to the current
 *  thread's registration list (its back is denoted by @c last).
 *  Since the registrations/unregistrations are expected in LIFO order
 *  (relatively to the current thread), both can be handled effeciently.
 */

/*
 * begin        : Sat Jun 8 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "pointer/pointer_to_object.h"
#include "pointer/object_proxy.h"
#include "object/object.h"
#include "utility/debug.h"
#include <sstream>


namespace Massiv {
namespace Core {


namespace {

const int           max_threads = 16;
    /**< Maximum number of simulation threads. */
int                 num_threads = 1;
    /**< Current number of simulation threads. */
int                 current_thread_index = 0;
    /**< Current simulation thread index. */    
UpcallToObject *    last_by_thread[ max_threads ];
    /**< This is a copy of @c UpcallToObject::last attributes specific to
         particular threads. For current thread, @c UpcallToObject::last is
         the up-to-date value (corresponding slot in @c last_by_thread is
         nulled). @c last_by_thread is updated on thread switches. */

}        


UpcallToObject * UpcallToObject::last = NULL;


bool UpcallToObject::is_object_referenced
    (
    const ObjectProperty * object
    )
    {
    // Find all living UpcallToObject instances and test if there
    // is an instance that references a component of @c object.
    // Need to search registration lists of all simulation threads.
    // The "component test" reduces to simple object root comparison.

    ObjectProperty *    my_root = object ? object->get_root_owner_object() : NULL;
    int                 i;

    // TODO: If this was too slow (mainly if called repeatedly),
    //       pin & unpin objects explicitly from UpcallToObject constructors,
    //       destructors and operator=.

    for( i = 0; i < num_threads; i++ )
        {
        UpcallToObject * upcall;

        // Get the last item registered in the list specific for thread @c i.

        if( i != current_thread_index )
            {
            upcall = last_by_thread[ i ];
            }
        else
            {
            upcall = last;
            }

        // Search the list.

        while( upcall != NULL )
            {
            ObjectProperty * referenced = upcall->object;
            ObjectProperty * referenced_root;

            if( referenced != NULL )
                {
                referenced_root = referenced->get_root_owner_object();
                }
            else
                {
                referenced_root = NULL;
                }

            if( referenced_root == my_root || 
                ( referenced_root != NULL && my_root == NULL ) )
                {
                return true;
                }
            
            upcall = upcall->prev;
            }
        }

    return false;
    }


ObjectProperty * UpcallToObject::get_top_most_object
    (
    bool skip_replicas
    )
    {
    UpcallToObject * upcall = last;

    if( !skip_replicas )
        {
        if( upcall )
            {
            return upcall->object;
            }
        else
            {
            return NULL;
            }
        }
    else
        {
        while( upcall != NULL )
            {
            ObjectProperty * referenced = upcall->object;

            if( referenced != NULL && !referenced->is_replica() )
                {
                return referenced;
                }

            upcall = upcall->prev;
            }

        return NULL;
        }
    }


std::string UpcallToObject::dump_list()
    {
    UpcallToObject *    upcall = last;
    std::ostringstream  ss;
    
    while( upcall != NULL )
        {
        ObjectProperty * referenced = upcall->object;
        
        if( referenced != NULL )
            {
            ss << referenced->get_root_owner_object()->get_object_id();

            #ifdef GC_TAG_STACK
            
            ss << " : " << ( upcall->tag ? upcall->tag : "???" ) << '\n';
            
            #endif
            }

        upcall = upcall->prev;
        }

    return ss.str();
    }


void UpcallToObject::simulation_thread_resumed
    (
    int simulation_thread_index
    )
    {
    // Save @c last of the current thread's registration list.
    
    last_by_thread[ current_thread_index ] = last;
    
    // Update current thread context.

    current_thread_index = simulation_thread_index;

    if( current_thread_index >= num_threads )
        {
        num_threads = current_thread_index + 1;
        
        if( num_threads > max_threads )
            {
            panic( "Too many simulation threads." );
            }
        }

    // Restore @c last for the new thread.

    last = last_by_thread[ current_thread_index ];
    last_by_thread[ current_thread_index ] = NULL;
    }


} // namespace Massiv::Core
} // namespace Massiv

#ifndef ENABLE_INLINE
#include "pointer_to_object_inline.h"
#endif
