/** @file pointer_to_object.h
 * Pointer to Object.
 */

/*
 * begin        : Sat Jun 8 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_POINTER_TO_OBJECT_H
#define MASSIV_POINTER_TO_OBJECT_H

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#include <string>


namespace Massiv {
namespace Core {
    

/** @addtogroup pointer
 * @{
 */

/**
 * Base class for @c PtrToObject and @c UpcallToObject.
 * Both pointer types implement a pointer to Object.
 *
 * Used by the core when it needs to access/upcall @c Object.
 * The difference from native "Object *" is that @c Object is pinned
 * until the pointer is released.
 * The core must not use "Object *" to access @c Object.
 *
 * @note Using @c ObjectProperty is correct as long as it is used
 * to call @c ObjectProperty methods only and if caller ensures that
 * referenced object is not deleted.
 *
 * @note One can convert @c ObjectPointer to @c Object to @c XXXToObject
 * by the following code sequence.
 * @c XXXToObject ptr( my_object_pointer.dereference() );
 *
 * @c PtrToObject and @c UpcallToObject differ in pinning implementation,
 * whether they are "registered" and their usage restrictions.
 * Both @c PtrToObject and @c UpcallToObject unpin objects immediatelly,
 * there is no "delayed unpin" semantics. This is the major difference
 * from LocalProxy<Object>.
 */
class MASSIV_EXPORTED PtrToObjectBase
    {
public:
        
    /** Dereferences pointer. */
    INLINE Object * operator->() const;
    
    /** Dereferences pointer. */
    INLINE Object & operator*() const;

    /** Dereferences pointer. */
    INLINE operator Object *() const;

    /** Dereferences pointer. */
    INLINE Object * get() const;

    /** Compares pointers. */
    INLINE bool operator==
        (
        const Object * other
        ) const;
    
    /** Compares pointers. */
    INLINE bool operator!=
        (
        const Object * other
        ) const;

protected:

    Object *  object;
        /**< Native pointer to Object. @c object need not be a root object. */
    
    }; // PtrToObjectBase


/**
 * Simple pointer to Object with semantics similar to LocalProxy<Object>.
 * Unlike LocalProxy<Object>, PtrToObject is more flexible.
 * It allows to point to NULL, member objects of another objects and is
 * assignable.
 *
 * Referenced objects are pinned explicitly and unpinned when PtrToObjects
 * no longer reference them. There is no "delayed unpin" semantics.
 * That is because PtrToObjects are used by core and core needs pin
 * objects "precisely".
 *
 * PtrToObject is not "registered" (see UpcallToObject, which
 * is "registered").
 */
class MASSIV_EXPORTED PtrToObject : public PtrToObjectBase
    {
public:
        
    /** Default constructor. */
    INLINE PtrToObject();

    /** Constructor. */
    INLINE PtrToObject
        (
        ObjectProperty * property
        );
    
    /** Constructor. */
    INLINE PtrToObject
        (
        const LocalProxy<Object> & proxy
        );

    /** Copy constructor. */
    INLINE PtrToObject
        (
        const PtrToObject & ptr
        );

    /** Destructor. */
    INLINE ~PtrToObject();

    /** Operator=. */
    INLINE PtrToObject & operator=
        (
        const PtrToObject & ptr
        );

    }; // PtrToObject


/**
 * Advanced pointer to object that registers all active pointer instances
 * and chains them into a LIFO list. This allows core to determine not only
 * what Objects are referenced but also the reference order (ie. what
 * Object was referenced for the last time). Core uses this to upcall
 * Objects and keep track of active instances (call stack).
 *
 * @c UpcallToObjects should be used consistently in LIFO order (with
 * respect to the current thread) to upcall client code from the core only,
 * so that we could then answer to these questions (direct local calls are
 * ignored):
 *  - what Object is currently running its code?
 *  - what is the call stack?
 *  ...
 *
 * @note    In order to be able to identify the call stack, each simulation
 *          thread (instance of main loop) maintains its separate list of
 *          pointer instances registered to that thread. Otherwise LIFO
 *          ordering (with respect to the current thread) could not have been
 *          ensured.
 *
 * Object is implicitly pinned if it is referenced by an @c UpcallToObject
 * (see @c ObjectProperty::is_upcall_pinned()).
 */
class MASSIV_EXPORTED UpcallToObject : public PtrToObjectBase
    {
private:
        
    friend class GarbageCollector;
    friend class GlobalCallbacks;

public:
        
    /** Default constructor. */
    INLINE UpcallToObject();

    /** Constructor. */
    INLINE UpcallToObject
        (
        ObjectProperty * property
        );
    
    /** Constructor. */
    INLINE UpcallToObject
        (
        const LocalProxy<Object> & proxy
        );
    
    /** Copy constructor. */
    INLINE UpcallToObject
        (
        const UpcallToObject & ptr
        );

    /** Destructor. */
    INLINE ~UpcallToObject();

    /** Operator=. */
    INLINE UpcallToObject & operator=
        (
        const UpcallToObject & ptr
        );

    /**
     * Associates this upcall with a specified @c new_tag (for example method
     * name and its arguments, etc).
     * @c dump_list() is able to print these tags.
     * If @c GC_TAG_STACK is not defined, the method does nothing.
     */
    INLINE void tag_upcall
        (
        const char * new_tag
        );

private:

    /**
     * Tests if @c object is referenced by a @c UpcallToObject instance.
     * Objects are not notified about the existing references explicitly,
     * registration lists of all simulation threads are searched instead.
     * If @c object is NULL then true is returned if there is at least one
     * non-NULL UpcallToObject instance.
     */
    static bool is_object_referenced
        (
        const ObjectProperty * object
        );
    
    /**
     * Returns object referenced by the last registered UpcallToObject
     * instance (with respect to the current simulation thread).
     * If @c skip_replicas is true then object replicas registered on the
     * top of the stack will be skipped.
     */
    static ObjectProperty * get_top_most_object
        (
        bool skip_replicas = false
        );

    /**
     * Dumps current thread's UpcallToObject registration list.
     */
    static std::string dump_list();

    /**
     * Called by the core when simulation thread @c simulation_thread_index
     * was resumed.
     * Switches @c last to the new thread-specific value (@c last is
     * thread-local, the implementation ensures proper swapping of
     * static @c last on simulation thread switches which occurs on well-defined
     * points only; two simulation threads never run in parallel).
     */
    static void simulation_thread_resumed
        (
        int simulation_thread_index
        );

private:

    UpcallToObject *        prev;
        /**< Previous pointer in the list of active pointer instances.
             The list is specific to each simulation thread. */
    static UpcallToObject * last;
        /**< The last pointer instance registered to the current thread's
             registration list.
             @c simulation_thread_resumed() updates @c last so that it would
             always point to the back of the current thread's registration
             list. */
    
    #ifdef GC_TAG_STACK

    const char *            tag;
        /**< Upcall tag printed by @c dump_list() when call-stack is printed. */

    #endif

    }; // UpcallToObject

/** @} */ // addtogroup pointer


} // namespace Massiv::Core
} // namespace Massiv

#ifdef ENABLE_INLINE
#include "pointer_to_object_inline.h"
#endif

#endif // MASSIV_POINTER_TO_OBJECT_H
