/** @file pointer_wrappers.h
 * ObjectPointer wrappers.
 */

/*
 * begin        : Sun May 19 2002
 * copyright    : (C) 2001, 2002 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_POINTER_WRAPPERS_H
#define MASSIV_POINTER_WRAPPERS_H

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif


namespace Massiv {
namespace Core {

//
// ObjectPointer wrappers.
//

/** @name ObjectPointer Macros
 * This is intended to wrap ObjectPointer to shorten the spelling,
 * it should not extend its functionality in any way.
 * It could be reduced to simple template typedefs according to
 * C++/0x standard.
 *
 * @ingroup pointer
 */
//@{

#define WRAP_LOCAL_POINTER( WrapperName, Base, Traits ) \
class WrapperName : \
    public ObjectPointer< Base< Traits< Type, Cast, LocalProxy<Type> > > > \
    { \
private: \
    typedef ObjectPointer< Base< Traits< Type, Cast, LocalProxy<Type> > > > Super; \
public: \
    WrapperName() { } \
    ~WrapperName() { } \
    WrapperName( ObjectProxy & proxy ) : Super( proxy ) { } \
    template <class Ptr> \
    WrapperName( const Ptr & ptr ) : Super( ptr ) { } \
    template <class Ptr> \
    WrapperName & operator=( const Ptr & ptr ) { Super::operator=( ptr ); return *this; } \
    WrapperName & operator=( ObjectProxy & proxy ) { Super::operator=( proxy ); return *this; } \
public: \
    Remote< Type > remote() const { return *this; } \
    };

#define WRAP_REMOTE_POINTER( WrapperName, Base, Traits ) \
class WrapperName : \
    public ObjectPointer< Base< Traits< Type, DefaultCast<Type>, RemoteProxy<Type> > > > \
    { \
private: \
    typedef ObjectPointer< Base< Traits< Type, DefaultCast<Type>, RemoteProxy<Type> > > > Super; \
public: \
    WrapperName() { } \
    ~WrapperName() { } \
    WrapperName( ObjectProxy & proxy ) : Super( proxy ) { } \
    template <class Ptr> \
    WrapperName( const Ptr & ptr ) : Super( ptr ) { } \
    template <class Ptr> \
    WrapperName & operator=( const Ptr & ptr ) { Super::operator=( ptr ); return *this; } \
    WrapperName & operator=( ObjectProxy & proxy ) { Super::operator=( proxy ); return *this; } \
    };

//@}


/** @class Pointer
 * Stack strong pointer.
 *
 * @ingroup pointer
 */
template
    <
    class Type,
    class Cast
    >
WRAP_LOCAL_POINTER( Pointer, StackPointerBase, StrongPointerTraits );

/** @class WeakPointer
 * Stack weak pointer.
 *
 * @ingroup pointer
 */
template
    <
    class Type,
    class Cast
    >
WRAP_LOCAL_POINTER( WeakPointer, StackPointerBase, WeakPointerTraits );

/** @class Remote
 * Stack RPC pointer.
 *
 * @ingroup pointer
 * @ingroup rpc
 */
template
    <
    class Type
    >
WRAP_REMOTE_POINTER( Remote, StackPointerBase, WeakPointerTraits );

/** @class PPointer
 * Strong property pointer.
 *
 * @ingroup pointer
 * @ingroup property
 */
template
    <
    class Type,
    class Cast
    >
WRAP_LOCAL_POINTER( PPointer, PropertyPointerBase, StrongPointerTraits );

/** @class PWeakPointer
 * Weak property pointer.
 *
 * @ingroup pointer
 * @ingroup property
 */
template
    <
    class Type,
    class Cast
    >
WRAP_LOCAL_POINTER( PWeakPointer, PropertyPointerBase, WeakPointerTraits );

/** @class PRemote
 * RPC property pointer.
 *
 * @ingroup pointer
 * @ingroup property
 * @ingroup rpc
 */
template
    <
    class Type
    >
WRAP_REMOTE_POINTER( PRemote, PropertyPointerBase, WeakPointerTraits );

#undef WRAP_LOCAL_POINTER
#undef WRAP_REMOTE_POINTER

} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_POINTER_WRAPPERS_H
