/** @file rpc_stubs.cpp
 * RPC flags and stub class (implementation).
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "pointer/rpc_stubs.h"
#include "pointer/garbage_collector.h"
#include "global/global.h"
#include "system/system_context.h"
#include "object/lib/lib_exceptions_helper.h"
#include <sstream>


namespace Massiv {
namespace Core {


void RPCStubsBase::_tried_to_call_non_const_method()
    {
    RAISE_CORE_LIB_EXCEPTION
        (
        RemoteCallFailedException,
        "Tried to call non-const method on object replica."
        );
    }


void RPCStubsBase::_unexpected_argument_class_type
    (
    const std::type_info &  expected,
    const std::type_info &  passed
    )
    {
    std::ostringstream oss;
    oss << "Tried to pass object of type '" << passed.name() <<
           "' as value type argument '" << expected.name() << "'. " <<
           "Value types are passed by value, types must be the same.";
    RAISE_CORE_LIB_EXCEPTION
        (
        RemoteCallFailedException,
        oss.str()
        );
    }


STime RPCStubsBase::_get_absolute_time() const
    {
    if( _flags & RPC_TIMED )
        {
        return _time;
        }
    else if( _flags & RPC_DELAYED )
        {
        return STime( STime::current_time() + _time );
        }
    return STime();
    }


ObjectProperty * RPCStubsBase::_local_srpc_call() const
    {
    // Can't use RPC_ALLOW_LOCAL_REPLICA_CALL for non-const methods.

    if( _flags & RPC_ALLOW_LOCAL_REPLICA_CALL )
        {
        RAISE_CORE_LIB_EXCEPTION
            (
            RemoteCallFailedException,
            "Tried to call non-const method with RPC_ALLOW_LOCAL_REPLICA_CALL flag set."
            );
        }
    
    // Test if optimization is enabled globally via settings.

    #ifndef OPTION_OPTIMIZE

    SystemContext::SRPCSettings & settings = 
        Global::system_context().srpc_settings;

    if( !settings.optimize_local_calls )
        {
        return NULL;
        }

    #endif

    // Can't optimize the call if it's scheduled to the future.

    if( ( _flags & RPC_DELAYED ) || ( _time != STime::asap() ) )
        {
        return NULL;
        }

    // Find the object and return it unless it's replica.

    ObjectProperty * const object = Global::garbage_collector().find_local_object( _object_id );
    if( object && !object->is_replica() )
        {
        return object;
        }

    // Can't call locally.

    return NULL;
    }


ObjectProperty * RPCStubsBase::_local_const_srpc_call() const
    {
    // Test if optimization is enabled globally via settings.

    #ifndef OPTION_OPTIMIZE

    SystemContext::SRPCSettings & settings = 
        Global::system_context().srpc_settings;

    if( !settings.optimize_local_calls )
        {
        return NULL;
        }

    #endif
    
    // Can't optimize the call if it's scheduled to the future.

    if( ( _flags & RPC_DELAYED ) || ( _time != STime::asap() ) )
        {
        return NULL;
        }

    // Find the object.

    ObjectProperty * const object = Global::garbage_collector().find_local_object( _object_id );
    if( object )
        {
        if
            (
            !object->is_replica() ||
            ( object->is_consistent() && ( _flags & RPC_ALLOW_LOCAL_REPLICA_CALL ) )
            )
            {
            // Object is not replica or it's consistent and local replica calls are allowed.

            return object;
            }
        }

    // Can't call locally.

    return NULL;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "rpc_stubs_inline.h"
#endif
