/** @file rpc_stubs.h
 * RPC flags and stub class.
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_RPC_STUBS_H
#define MASSIV_RPC_STUBS_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#include <typeinfo>


namespace Massiv {
namespace Core {


/**
 * Remote call flags.
 *
 * @ingroup rpc
 */
enum RPCFlags
    {

    RPC_TIMED       = 0x0001,
        /**< Time is absolute. */

    RPC_DELAYED     = 0x0002,
        /**< Time is relative to current time. */

    RPC_RESULTS     = 0x0004,
        /**< Reply with results should be sent back to newly created #Massiv::Core::ResultObject. */

    RPC_NO_ARCHIVE  = 0x0008,
        /**< Mark object used to transfer RPC arguments as unarchivable.
             This flag must be used for synchronous RPC requests. It should
             never be set otherwise. */

    RPC_ALLOW_LOCAL_REPLICA_CALL = 0x0010,
        /**< Call replica method if replica of the object is available.
             This flag may be set only for synchronous RPC requests. Called method
             must be @c const. */
        
    RPC_REPLICAS    = 0x0020,
        /**< Send request to replicas of the object instead of object itself.
             Results from call to replicas can't be requested. Called method
             must be @c const. If the object is not local, no requests will be
             sent (probably). Requests will be delivered at most once to each
             replica. Nothing else is guaranteed. */

    RPC_DEFAULT     = ( RPC_TIMED ),
        /**< Default RPC flags. */

    RPC_SYNC_YES    = ( RPC_RESULTS | RPC_NO_ARCHIVE ),
        /**< Flags always set during synchronous RPC. */

    RPC_SYNC_NO     = ( RPC_DELAYED | RPC_REPLICAS ),
        /**< Flags that must not be set during synchronous RPC. */

    }; // enum RPCFlags


/**
 * Each Generated::RPCStubs<> class must be derived from
 * this class.
 *
 * @ingroup rpc
 * @ingroup pointer
 */
class MASSIV_EXPORTED RPCStubsBase
    {
protected:

    /**
     * Default Constructor.
     */
    INLINE RPCStubsBase();

protected:

    /**
     * Bind stub class to @c id.
     */
    INLINE void _bind
        (
        const ObjectId & id
        );

    /**
     * Set #_flags and _time.
     */
    INLINE void _param
        (
        int             flags,
        const STime &   time
        );

    /**
     * Check/set #_flags for synchronous RPC.
     */
    INLINE void _sync();

protected:

    /**
     * Throw exception because user tried to use RPC_REPLICAS or RPC_ALLOW_LOCAL_REPLICA_CALL
     * flags while calling a non-const method.
     */
    static NORETURN( void ) _tried_to_call_non_const_method();

    /**
     * Throw exception because user tried to pass object of different type as
     * method argument.
     */
    static NORETURN( void ) _unexpected_argument_class_type
        (
        const std::type_info &  expected,
        const std::type_info &  passed
        );

    /**
     * Fail with error if types do not match.
     */
    static INLINE void _check_class_argument_type
        (
        const std::type_info &  expected,
        const std::type_info &  passed
        );

    /**
     * Get absolute time computed from @a _flags and @a _time.
     */
    STime _get_absolute_time() const;

    /**
     * Check if non-const method can be called locally.
     *
     * @return Pointer to local object if method can be called. Object must not
     *         be replica. @c NULL otherwise.
     *
     * @exception RemoteCallFailedException Thrown if
     *            @c RPC_ALLOW_LOCAL_REPLICA_CALL is set.
     */
    ObjectProperty * _local_srpc_call() const;

    /**
     * Check if const method can be called locally.
     *
     * @return Pointer to local object if method can be called. If @c RPC_ALLOW_LOCAL_REPLICA_CALL
     *         is set, object must be consistent (ie. either consistent replica or non-replica),
     *         otherwise it must not be replica. @c NULL otherwise.
     */
    ObjectProperty * _local_const_srpc_call() const;

    /**
     * Is RPC_RESULTS flag set?
     */
    INLINE bool _require_reply() const;

    /**
     * Is RPC_NO_ARCHIVE flag set?
     */
    INLINE bool _no_archive() const;

protected:

    ObjectId    _object_id;
        /**< Target object for RPC call. */

    int         _flags;
        /**< #RPCFlags. */

    STime       _time;
        /**< Migration time or delay. */

    STime       _replica_timeout;
        /**< Request replication of called object to callee node for
             #_replica_timeout seconds. */

    }; // class RPCStubsBase


} // namespace Massiv::Core


namespace Generated { // namespace Massiv::Generated

/**
 * RPCStubs defines and implements methods that have the same
 * signatures as corresponding @c Iface. The implementation
 * of each method marshalls its arguments and invokes a remote call.
 * RPCStubs is specialized for each supported @c Iface and must
 * be derived from RPCStubsBase.
 *
 * Remote calls may be perfomed in many ways - asynchronous scheduled to
 * specific simulation time, asynchronous to object replicas, synchronous,
 * they may trigger object replication, etc. Each RPCStubs class should
 * implement several methods that set call options and two methods per
 * IDL-defined method: @c async_* version (which takes only @c in and @c inout
 * parameters and performs asynchronous call with optional reply request) and
 * @c sync_* version (with full signature; it uses async_* version to perform
 * the call and then waits for the results).
 *
 * @ingroup rpc
 * @ingroup pointer
 */
template <class Iface>
class RPCStubs;

} // namespace Massiv::Generated

} // namespace Massiv


#ifdef ENABLE_INLINE
#include "rpc_stubs_inline.h"
#endif

#endif // MASSIV_RPC_STUBS_H
