/** @file attribute.h
 * Managed attributes.
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ATTRIBUTE_H
#define MASSIV_ATTRIBUTE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_DEBUG_STORAGE_H
#include "utility/debug_storage.h"
#endif

#include <string>
#include <ostream>


namespace Massiv {
namespace Core {


/** @addtogroup attribute_types
 * @{
 */

/** @class DefaultAttributeEventHandlers
 * This class defines default implementation of
 * ManagedAttribute callbacks. See ManagedAttribute.
 */
class MASSIV_EXPORTED DefaultAttributeEventHandlers
    {
public:

    /** Called by ManagedAttribute from its default constructor. */
    void default_constructed()
        {
        }

    /** Called by ManagedAttribute before attribute change. */
    void write_requested()
        {
        }

    /** Called by ManagedAttribute before attribute read. */
    void read_requested() const
        {
        }

    }; // DefaultAttributeEventHandlers


/** @class ManagedAttribute
 * ManagedAttribute encapsulates Type into object that
 * keeps track modifications of the value it represents.
 * Value can be read through read() method and written
 * through write() method. To simplify use of the attribute,
 * operator=() and default conversion to Type were implemented.
 *
 * AttributeBase is a class that is inherited by ManagedAttribute.
 * This class may be provided by the client. The purpose is to
 * store client-specific data related to the attribute as well
 * as to implement various callbacks called by ManagedAttribute.
 * See DefaultAttributeEventHandlers for information on which callbacks
 * must be defined and what their purpose is.
 */
template <class Type, class AttributeBase = DefaultAttributeEventHandlers>
class ManagedAttribute : public AttributeBase
    {
private:

    typedef ManagedAttribute<Type, AttributeBase>   ThisAttribute;

public:

    /// Constant reference to Type.
    typedef const Type &    ConstReferenceType;

    /// Reference to Type.
    typedef Type &          ReferenceType;

    using AttributeBase::default_constructed;
    using AttributeBase::write_requested;
    using AttributeBase::read_requested;

public:

    /** Default constructor. */
    ManagedAttribute() : value( Type() )
        {
        default_constructed();
        }

    /** Initializes managed attribute from native attribute. */
    ManagedAttribute
        (
        ConstReferenceType v
        ) : value( v )
        {
        }

    /** Copy constructor. */
    ManagedAttribute
        (
        const ManagedAttribute & m
        ) : AttributeBase( m ),
            value( m.read() )
        {
        }

public:

    /** Reads value of the attribute. */
    ConstReferenceType read() const
        {
        read_requested();
        return value;
        }

    /** Modifies value of the attribute. */
    ThisAttribute & write
        (
        ConstReferenceType new_value
        )
        {
        write_requested();
        value = new_value;

        return *this;
        }

public:

    /**
     * Attribute can be used as r-value in expressions
     * so "read" operator is provided.
     * There should not be defined another conversions
     * to avoid ambiguities.
     */
    operator ConstReferenceType() const
        {
        return read();
        }

    /** Alias for write(). */
    ThisAttribute & operator=
        (
        ConstReferenceType value
        )
        {
        return write( value );
        }

    /** Predefined operator=() for any managed attribute. */
    ThisAttribute & operator=
        (
        const ThisAttribute & m
        )
        {
        write( m.read() );
        return *this;
        }

protected:

    /**
     * Returns writeable reference to the value of the attribute.
     * This allows to directly modify value stored in the attribute.
     */
    ReferenceType write()
        {
        write_requested();
        return value;
        }

private:

    Type value;
        /**< Current value of the attribute. */

    }; // ManagedAttribute


/** @class NumericAttribute
 * NumericAttribute extends ManagedAttribute in that way that
 * it overloads arithmetic operators so that the attribute
 * can be used in expressions as NumericType.
 */
template <class NumericType, class AttributeBase = DefaultAttributeEventHandlers>
class NumericAttribute : public ManagedAttribute<NumericType, AttributeBase>
    {
private:

    typedef ManagedAttribute<NumericType, AttributeBase>    Super;
    typedef NumericAttribute<NumericType, AttributeBase>    ThisAttribute;

public:
    
    using Super::write_requested;
    using Super::read_requested;
    using Super::read;
    using Super::write;

public:

    // l-values

    /** Constructor. */
    NumericAttribute() : Super()
        {
        }

    /** Constructs and initializes numeric attribute to value provided. */
    NumericAttribute
        (
        NumericType value
        ) : Super( value )
        {
        }

    /** Copy constructor. */
    NumericAttribute
        (
        const ThisAttribute & attribute
        ) : Super( attribute )
        {
        }

    /** Assigns value to the attribute. */
    NumericAttribute & operator=
        (
        NumericType value
        )
        {
        write( value );
        return *this;
        }

    /** Operator+=. */
    NumericAttribute & operator+=
        (
        NumericType value
        )
        {
        write( read() + value );
        return *this;
        }

    /** Operator-=. */
    NumericAttribute & operator-=
        (
        NumericType value
        )
        {
        write( read() - value );
        return *this;
        }

    /** Operator*=. */
    NumericAttribute & operator*=
        (
        NumericType value
        )
        {
        write( read() * value );
        return *this;
        }

    /** Operator/=. */
    NumericAttribute & operator/=
        (
        NumericType value
        )
        {
        write( read() / value );
        return *this;
        }

    // r-values

    /** Operator++, preincrement. */
    NumericType operator++()
        {
        return write( read() + 1 );
        }

    /** Operator++, postincrement. */
    NumericType operator++( int )
        {
        NumericType value = read();
        write( value + 1 );
        return value;
        }

    /** Operator--, predecrement. */
    NumericType operator--()
        {
        return write( read() - 1 );
        }

    /** Operator--, postdecrement. */
    NumericType operator--( int )
        {
        NumericType value = read();
        write( value - 1 );
        return value;
        }

    /** Operator-. */
    NumericType operator-()
        {
        return -read();
        }

    }; // NumericAttribute


/** @class IntegralAttribute
 * IntegralAttribute extends NumericAttribute by implementation
 * of operators that may be used on integral types only.
 */
template <class IntegralType, class AttributeBase = DefaultAttributeEventHandlers>
class IntegralAttribute : public NumericAttribute<IntegralType, AttributeBase>
    {
private:

    typedef NumericAttribute<IntegralType, AttributeBase>   Super;
    typedef IntegralAttribute<IntegralType, AttributeBase>  ThisAttribute;

public:

    using Super::write_requested;
    using Super::read_requested;
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    IntegralAttribute()
        {
        }

    /** Initializes attribute from the value. */
    IntegralAttribute
        (
        IntegralType value
        ) : Super( value )
        {
        }

    /** Copy contructor. */
    IntegralAttribute
        (
        const ThisAttribute & attribute
        ) : Super( attribute )
        {
        }

    /** Operator=. */
    IntegralAttribute & operator=
        (
        IntegralType value
        )
        {
        write( value );
        return *this;
        }

    /** Operator&=. */
    IntegralAttribute & operator&=
        (
        IntegralType value
        )
        {
        write( read() & value );
        return *this;
        }

    /** Operator|=. */
    IntegralAttribute & operator|=
        (
        IntegralType value
        )
        {
        write( read() | value );
        return *this;
        }

    /** Operator^=. */
    IntegralAttribute & operator^=
        (
        IntegralType value
        )
        {
        write( read() ^ value );
        return *this;
        }

    /** Operator%=. */
    IntegralAttribute & operator%=
        (
        IntegralType value
        )
        {
        write( read() % value );
        return *this;
        }

    /** Operator!. */
    IntegralType operator!()
        {
        return !read();
        }

    }; // IntegralAttribute


/** @class StringAttribute
 * Simple attribute wrapper for std::string.
 * Some standard string methods like insert, append, etc. are implemented.
 * Both iterator (@c Iterator) and const iterator (@c ConstIterator)
 * are implemented.
 * Use of this class (PString is derived form this class)
 * is shown in source file test_serializer.cpp.
 */
template <class AttributeBase = DefaultAttributeEventHandlers>
class StringAttribute :
    public ManagedAttribute<std::string, AttributeBase>,
    public DebugStorage
    {
private:

    typedef ManagedAttribute<std::string, AttributeBase>    Super;
    typedef StringAttribute<AttributeBase>                  ThisAttribute;

public:

    typedef std::string::size_type                          SizeType;
    typedef std::string::difference_type                    DifferenceType;
    typedef std::string::reference                          ReferenceType;
    typedef std::string::const_reference                    ConstReferenceType;
    typedef std::string::value_type                         ValueType;

    typedef SafeRandomAccessIterator
                <
                SequenceIteratorTraits<ThisAttribute>
                >   Iterator;

    typedef SafeRandomAccessIterator
                <
                SequenceConstIteratorTraits<ThisAttribute>
                >   ConstIterator;

    using Super::write_requested;
    using Super::read_requested;
    using Super::read;
    using Super::write;

public:

    /** Constructor. */
    StringAttribute()
        {
        }

    /** Initializes attribute from the value. */
    StringAttribute
        (
        const std::string & s
        ) : Super( s )
        {
        }

    /** Copy constructor. */
    StringAttribute
        (
        const ThisAttribute & s
        ) : Super( s )
        {
        }

    /** Operator=. */
    ThisAttribute & operator=
        (
        const std::string & s
        )
        {
        write( s );
        return *this;
        }

    /** Destructor. */
    ~StringAttribute()
        {
        }

    /** Returns iterator pointing to the beginning of the string. */
    Iterator begin()
        {
        return Iterator( this, Iterator::BeginTag() );
        }

    /** Returns const iterator pointing to the beginning of the string. */
    ConstIterator begin() const
        {
        return ConstIterator( const_cast<ThisAttribute *>( this ), ConstIterator::BeginTag() );
        }

    /** Returns iterator pointing to the end of the string. */
    Iterator end()
        {
        return begin() + size();
        }

    /** Returns const iterator pointing to the end of the string. */
    ConstIterator end() const
        {
        return begin() + size();
        }

    /** Returns character at position @c pos. */
    ConstReferenceType at
        (
        SizeType pos
        ) const
        {
        DebugStorage::range_error( pos >= 0 && pos < size(), "at()" );
        return read().at( pos );
        }

    /**
     * Returns reference to character at position @c pos.
     * Assumes that character will be modified.
     */
    ReferenceType write_at
        (
        SizeType pos
        )
        {
        DebugStorage::range_error( pos >= 0 && pos < size(), "write_at()" );
        return write().at( pos );
        }

    /**
     * Returns const reference to character at position @c pos.
     * To modify get writeable reference use @c write_at() or operator().
     */
    ConstReferenceType operator[]
        (
        SizeType pos
        ) const
        {
        return at( pos );
        }

    /**
     * Returns reference to character at position @c pos.
     * Assumes that character will be modified.
     */
    ReferenceType operator()
        (
        SizeType pos
        )
        {
        return write_at( pos );
        }

    /** Appends @ chr to the string. */
    ThisAttribute & push_back
        (
        ConstReferenceType chr
        )
        {
        invalidate_iterators();

        write().push_back( chr );
        return *this;
        }

    /** Appends string @c s to the string. */
    ThisAttribute & append
        (
        const std::string & s
        )
        {
        invalidate_iterators();

        write().append( s );
        return *this;
        }

    /** Appends string @c to the string. */
    ThisAttribute & operator+=
        (
        const std::string & s
        )
        {
        return append( s );
        }

    /** Inserts @c s at position @c pos. */
    ThisAttribute & insert
        (
        SizeType            pos,
        const std::string & s
        )
        {
        invalidate_iterators();

        write().insert( pos, s );
        return *this;
        }

    /** Compares strings. */
    int compare
        (
        const std::string & s
        ) const
        {
        return read().compare( s );
        }

    /** Compares strings. */
    friend bool operator==
        (
        const ThisAttribute &   s1,
        const ThisAttribute &   s2
        )
        {
        return !s1.compare( s2.read() );
        }

    /** Compares strings. */
    friend bool operator!=
        (
        const ThisAttribute &   s1,
        const ThisAttribute &   s2
        )
        {
        return !operator==( s1, s2 );
        }

    /** Compares strings. */
    friend bool operator==
        (
        const ThisAttribute &   s1,
        const std::string &     s2
        )
        {
        return !s1.compare( s2 );
        }

    /** Compares strings. */
    friend bool operator!=
        (
        const ThisAttribute &   s1,
        const std::string &     s2
        )
        {
        return !operator==( s1, s2 );
        }

    /** Finds substring @c s in string starting at position @c pos. */
    SizeType find
        (
        const std::string & s,
        SizeType            pos = 0
        ) const
        {
        return read().find( s, pos );
        }

    /** Returns current length of the string. */
    SizeType size() const
        {
        return read().size();
        }

    }; // class StringAttribute


/**
 * Write string to stream.
 */
template< class AttributeBase >
std::ostream & operator<<
    (
    std::ostream &                              stream,
    const StringAttribute< AttributeBase > &    str
    )
    {
    stream << str.read();
    return stream;
    }


/** @class BooleanAttribute
 * Simple attribute wrapper for bool.
 */
template <class AttributeBase = DefaultAttributeEventHandlers>
class BooleanAttribute : public ManagedAttribute<bool, AttributeBase>
    {
private:

    typedef ManagedAttribute<bool, AttributeBase>   Super;
    typedef BooleanAttribute<AttributeBase>         ThisAttribute;

public:

    using Super::read;
    using Super::write;

public:

    /** Constructor. */
    BooleanAttribute()
        : Super()
        {
        }

    /** Initializes attribute from the value. */
    BooleanAttribute
        (
        bool b
        ) : Super( b )
        {
        }

    /** Copy constructor. */
    BooleanAttribute
        (
        const ThisAttribute & b
        ) : Super( b )
        {
        }

    /** Operator=. */
    BooleanAttribute & operator=
        (
        bool b
        )
        {
        write( b );
        return *this;
        }

    /** Operator|=. */
    BooleanAttribute & operator|=
        (
        bool b
        )
        {
        write( read() | b );
        return *this;
        }

    /** Operator&=. */
    BooleanAttribute & operator&=
        (
        bool b
        )
        {
        write( read() & b );
        return *this;
        }

    /** Operator!. */
    bool operator!()
        {
        return !read();
        }

    /** Flips value (true <-> false). */
    BooleanAttribute & flip()
        {
        write( !read() );
        return *this;
        }

    }; // BooleanAttribute

/** @} */ // addtogroup attribute_types


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ATTRIBUTE_H
