/** @file color.h
 * RGBA floating-point color.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_COLOR_H
#define MASSIV_COLOR_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#ifndef MASSIV_STYPE_FLOAT_H
#include "serializer/stype_float.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup template_types
 * @{
 */

/**
 * Class for storing color components.
 */
class MASSIV_EXPORTED ColorData
    {
public:

    struct FloatQTraits
        {
        static const Float32 min_value;
        static const Float32 max_value;
        enum { bits = 8 };
        };

public:

    typedef Float32                             Component;
    typedef SFloatQ< Component, FloatQTraits >  InternalComponent;

public:

    InternalComponent   comp[ 4 ];
        /**< Components of color. */

    }; // ColorData.


/**
 * Serializable RGBA color.
 *
 * Standard values for all the components are in the < 0.0, 1.0 > range.
 * Values outside the range are allowed, but in current implementation they
 * will be clamped during lossy serialization (replication).
 *
 * Alpha equal to 0.0 should be interpreted as transparent, alpha equal to
 * 1.0 as opaque.
 *
 * @c Base must be either #DefaultAttributeHandlers or #Property.
 */
template< class Base >
class MASSIV_EXPORTED Color : public Base
    {
private:

    /// This class.
    typedef Color< Base >                   Self;

    /// Superclass.
    typedef Base                            Super;

    /// Internal representation of the color component.
    typedef ColorData::InternalComponent    InternalComponent;

    /// To make MSVC6 happy.
    typedef Serializer::Description         SerializerDescription;

protected:

    using Base::default_constructed;
    using Base::read_requested;
    using Base::write_requested;    

public:

    /// Color component.
    typedef ColorData::Component            Component;
    
    /// Corresponding native type.
    typedef ColorData                       NativeType;

    /// Corresponding property type.
    typedef PColor                          PropertyType;

    /// Corresponding stype.
    typedef SColor                          SerializableType;

public:

    /**
     * Basic simple named colors.
     * Each component is either 0.0f or 1.0f.
     */
    enum Name
        {
        BLACK,
        WHITE,
        RED,
        GREEN,
        BLUE,
        YELLOW,
        MAGENTA,
        CYAN,
        };
        
public:

    /**
     * Default constructor.
     *
     * Sets color to transparent black.
     */
    INLINE Color();

    /**
     * Create simple color from name.
     */
    Color
        (
        Name        name,
        Component   alpha = 1.0f
        );

    /**
     * Create color from specified components.
     */
    INLINE Color
        (
        Component   red,
        Component   green,
        Component   blue,
        Component   alpha = 1.0f
        );

    /**
     * Create color from specified color data.
     */
    INLINE Color
        (
        const ColorData & cd
        );
        
    /**
     * Copy constructor.
     */
    INLINE Color
        (
        const Color & other
        );

    /**
     * Assignment operator.
     */
    INLINE Color & operator=
        (
        const Color & other
        );
        
    /**
     * Assignment operator.
     */
    INLINE Self & operator=
        (
        const ColorData & cd
        );
        
    /**
     * Cast to ColorData.
     */
    INLINE operator const ColorData & () const;
    

public:

    //@{
    /** Getter. */
    INLINE Component red() const;

    INLINE Component green() const;

    INLINE Component blue() const;

    INLINE Component alpha() const;
    //@}

    /** SColor/PColor conversion. */
    INLINE const ColorData & read() const;
        
    //@{
    /** Setter. */
    INLINE void set_red
        (
        Component red
        );

    INLINE void set_green
        (
        Component green
        );

    INLINE void set_blue
        (
        Component blue
        );

    INLINE void set_alpha
        (
        Component alpha
        );
    //@}

public:

    /**
     * Compare colors for exact equality (no epsilons).
     */
    INLINE bool operator==
        (
        const Color & other
        ) const;

    /**
     * Compare color for inequality.
     */
    INLINE bool operator!=
        (
        const Color & other
        ) const;

    /**
     * Lexicographically compare two colors.
     */
    INLINE bool operator<
        (
        const Color & other
        ) const;

    // TODO: <=, >, >= ?

public:

    /**
     * Add colors, including alpha channel.
     */
    INLINE Color & operator+=
        (
        const Color & other
        );

    /**
     * Subtract colors, including alpha channel.
     */
    INLINE Color & operator-=
        (
        const Color & other
        );

    /**
     * Multiply all components by scalar value.
     */
    INLINE Color & operator*=
        (
        Component scalar
        );

    /**
     * Multiply two colors, including alpha channel.
     */
    INLINE Color & operator*=
        (
        const Color & other
        );

public:

    /**
     * Add colors, including alpha channel.
     */
    INLINE Color operator+
        (
        const Color & other
        ) const;

    /**
     * Subtract colors, including alpha channel.
     */
    INLINE Color operator-
        (
        const Color & other
        ) const;

    /**
     * Multiply all components by scalar value.
     */
    INLINE Color operator*
        (
        Component scalar
        ) const;

    /**
     * Multiply colors, including alpha channel.
     */
    INLINE Color operator*
        (
        const Color & other
        ) const;

public:

    /**
     * Add RGB components, don't change alpha.
     */
    INLINE Color & add_rgb
        (
        const Color & other
        );

    /**
     * Subtract RGB components, don't change alpha.
     */
    INLINE Color & subtract_rgb
        (
        const Color & other
        );

    /**
     * Multiply RGB components by scalar, don't change alpha.
     */
    INLINE Color & multiply_rgb
        (
        Component scalar
        );

    /**
     * Multiply RGB components, don't change alpha.
     */
    INLINE Color & multiply_rgb
        (
        const Color & other
        );

public:

    /**
     * Scale all components by scalar around center value (0.5).
     */
    INLINE Color & scale
        (
        Component scalar
        );

    /**
     * Scale RGB components by scalar around center value (0.5).
     */
    INLINE Color & scale_rgb
        (
        Component scalar
        );
        
public:

    /**
     * Compute color intensity from RGB components using standard 3:6:1 weights.
     */
    INLINE Component intensity() const;

    
    // TODO: more standard color operations

public:

    /**
     * Serialize to bitstream.
     */
    INLINE void bitstr_write
        (
        BitOStream &                    bos,
        const SerializerDescription &   desc
        ) const;
        
    /**
     * Read from bitstream.
     */
    INLINE void bitstr_read
        (
        BitIStream &                    bis,
        const SerializerDescription &   desc
        );

    /**
     * Serialize to text stream.
     */
    INLINE void text_write
        (
        TextWriter &                    tw,
        const SerializerDescription &   desc
        ) const;

    /**
     * Read from text stream.
     */
    INLINE void text_read
        (
        TextReader &                    tr,
        const SerializerDescription &   desc
        );

public:

    /**
     * Convert color to string.
     */
    std::string to_string() const;

    /**
     * Compute hash code.
     *
     * Warning: return constant 0, see sources.
     */
    INLINE int hash() const;

    /**
     * Compute occupied memory size.
     */
    INLINE int memory_size() const;
    
    /**
     * Guess bitstream serialization size (in bits).
     */
    INLINE int bitstream_size() const;

private:

    ColorData rgba;
        /**< Color components. */

    }; // class Color


/**
 * Multiply all components by scalar value.
 */
template< class Base >
MASSIV_EXPORTED inline Color< Base > operator*
    (
    typename Color< Base >::Component   scalar,
    const Color< Base > &                 color
    )
    {
    return color * scalar;
    }

/** @} */ // addtogroup attribute_types


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "color_inline.h"
#endif


#endif // MASSIV_COLOR_H
