/** @file orientation.cpp
 * Heading-pitch-bank orientation (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "property/orientation.h"
#include "property/property.h"
#include "serializer/serializer.h"
#include "serializer/stype_vector.h"
#include <sstream>


namespace Massiv {
namespace Core {


const Float32 OrientationData::HeadingTraits::min_value = 0.0f;
const Float32 OrientationData::HeadingTraits::max_value = 2.0f * MATHF_PI;

const Float32 OrientationData::PitchBankTraits::min_value = -MATHF_PI;
const Float32 OrientationData::PitchBankTraits::max_value = +MATHF_PI;


template< class Base >
void Orientation< Base >::set_from_forward_vector
    (
    const SVector3 & forward_vector
    )
    {
    hope( forward_vector.is_normalized() );

    // atan2 returns 0 for vector ( 1, 0 ) and PI / 2 for vector ( 0, 1 ).
    // We want 0 for vector ( 0, 1 ) and PI / 2 for vector ( -1, 0 ).

    const Float32 x = forward_vector.y();
    const Float32 y = -forward_vector.x();
    set_heading( FloatTraits< Float32 >::atan2( y, x ) );

    // Level vector gives us zero pitch, "up" vector corresponds to PI / 2.

    set_pitch( FloatTraits< Float32 >::safe_asin( forward_vector.z() ) );

    // Bank is always zero.

    set_bank( 0.0f );
    }


template< class Base >
void Orientation< Base >::set_from_up_vector
    (
    const SVector3 & up_vector
    )
    {
    hope( up_vector.is_normalized() );

    // There are many ways to choose orientation heading from up vector. We
    // will use heading of the up vector as if it were forward vector, and
    // zero heading if up.z is +1 or -1. It's easy to determine pitch for such
    // heading. Note that this means that the orientation will always point
    // downwards.

    // Observe the picture:
    //
    // a) up.z > 0.0               b) up.z < 0.0
    //
    // z ^                         z ^
    //   .   up                      .
    //   .  /                        .
    //   . /                         .
    //   ./                          .
    //   +-.......> x,y              +-.......> x,y
    //     --                       /  --
    //       --                    /     --
    //         -- fwd             /        -- up
    //                          fwd

    const Float32 x = up_vector.y();
    const Float32 y = -up_vector.x();
    if( FloatTraits< Float32 >::epsilon_zero( x ) &&
        FloatTraits< Float32 >::epsilon_zero( y ) )
        {
        set_heading( 0.0f );
        set_pitch( 0.0f );
        }
    else
        {
        // See #set_from_forward_vector().

        set_heading( FloatTraits< Float32 >::atan2( y, x ) );

        // We want zero pitch if z == 1.0, -PI / 2 if z == 0, -PI if z == -1.0.
        // If z < 0, the orientation will point in the opposite direction
        // of up vector (heading-wise) and it will be "upside down"
        // (pitch < -PI / 2).

        const Float32 z = up_vector.z();
        set_pitch( FloatTraits< Float32 >::safe_asin( z ) - MATHF_PI_2 );
        }

    // Bank is always zero.

    set_bank( 0.0f );
    }


// To rotate a point, multiply one of the following matrices by a
// a column vector containing the point coordinates from the right:

// Heading matrix:
// (  cos h   -sin h        0 )
// (  sin h    cos h        0 )
// (      0        0        1 )

// Pitch matrix:
// (      1        0        0 )
// (      0    cos p   -sin p )
// (      0    sin p    cos p )

// Bank matrix:
// (  cos b        0    sin b )
// (      0        1        0 )
// ( -sin b        0    cos b )

// Vectors returned by the following methods are columns of matrix computed
// as composition of heading, pitch and bank matrix.
// Let's hope these are right...

template< class Base >
SVector3 Orientation< Base >::right_vector() const
    {
    read_requested();

    const Float32 cos_h = AngleTraits::cos( data.h );
    const Float32 sin_h = AngleTraits::sin( data.h );
    const Float32 cos_p = AngleTraits::cos( data.p );
    const Float32 sin_p = AngleTraits::sin( data.p );
    const Float32 cos_b = AngleTraits::cos( data.b );
    const Float32 sin_b = AngleTraits::sin( data.b );

    return SVector3
        (
         cos_h * cos_b - sin_h * sin_p * sin_b,
         sin_h * cos_b + cos_h * sin_p * sin_b,
        -cos_p * sin_b
        );
    }


template< class Base >
SVector3 Orientation< Base >::forward_vector() const
    {
    read_requested();

    const Float32 cos_h = AngleTraits::cos( data.h );
    const Float32 sin_h = AngleTraits::sin( data.h );
    const Float32 cos_p = AngleTraits::cos( data.p );
    const Float32 sin_p = AngleTraits::sin( data.p );

    return SVector3
        (
        -sin_h * cos_p,
         cos_h * cos_p,
         sin_p
        );
    }


template< class Base >
SVector3 Orientation< Base >::up_vector() const
    {
    read_requested();

    const Float32 cos_h = AngleTraits::cos( data.h );
    const Float32 sin_h = AngleTraits::sin( data.h );
    const Float32 cos_p = AngleTraits::cos( data.p );
    const Float32 sin_p = AngleTraits::sin( data.p );
    const Float32 cos_b = AngleTraits::cos( data.b );
    const Float32 sin_b = AngleTraits::sin( data.b );

    return SVector3
        (
        cos_h * sin_b + sin_h * sin_p * cos_b,
        sin_h * sin_b - cos_h * sin_p * cos_b,
        cos_p * cos_b
        );
    }


template< class Base >
void Orientation< Base >::rotation_base
    (
    SVector3 & right_vector,
    SVector3 & forward_vector,
    SVector3 & up_vector
    ) const
    {
    read_requested();

    // copy-paste from methods above...

    const Float32 cos_h = AngleTraits::cos( data.h );
    const Float32 sin_h = AngleTraits::sin( data.h );
    const Float32 cos_p = AngleTraits::cos( data.p );
    const Float32 sin_p = AngleTraits::sin( data.p );
    const Float32 cos_b = AngleTraits::cos( data.b );
    const Float32 sin_b = AngleTraits::sin( data.b );

    right_vector =  SVector3
        (
         cos_h * cos_b - sin_h * sin_p * sin_b,
         sin_h * cos_b + cos_h * sin_p * sin_b,
        -cos_p * sin_b
        );
    forward_vector = SVector3
        (
        -sin_h * cos_p,
         cos_h * cos_p,
         sin_p
        );
    up_vector = SVector3
        (
         cos_h * sin_b + sin_h * sin_p * cos_b,
         sin_h * sin_b - cos_h * sin_p * cos_b,
         cos_p * cos_b
        );
    }


template< class Base >
void Orientation< Base >::rotate
    (
    SVector3 & point
    ) const
    {
    SVector3 r, f, u;
    rotation_base( r, f, u );
    const SVector3 p = point;

    point.set_x( p.x() * r.x() + p.y() * f.x() + p.z() * u.x() );
    point.set_y( p.x() * r.y() + p.y() * f.y() + p.z() * u.y() );
    point.set_z( p.x() * r.z() + p.y() * f.z() + p.z() * u.z() );
    }


template< class Base >
void Orientation< Base >::inverse_rotate
    (
    SVector3 & point
    ) const
    {
    SVector3 r, f, u;
    rotation_base( r, f, u );
    point = SVector3( point * r, point * f, point * u );
    }


template< class Base >
void Orientation< Base >::bitstr_write
    (
    BitOStream &                    bos,
    const SerializerDescription &   desc
    ) const
    {
    read_requested();
    data.h.bitstr_write( bos, desc );
    data.p.bitstr_write( bos, desc );
    data.b.bitstr_write( bos, desc );
    }


template< class Base >
void Orientation< Base >::bitstr_read
    (
    BitIStream &                    bis,
    const SerializerDescription &   desc
    )
    {
    write_requested();
    data.h.bitstr_read( bis, desc );
    data.p.bitstr_read( bis, desc );
    data.b.bitstr_read( bis, desc );
    }


template< class Base >
std::string Orientation< Base >::to_string() const
    {
    std::ostringstream oss;
    oss << "( " << heading() << ", " << pitch() << ", " << bank() << " )";
    return oss.str();
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "orientation_inline.h"
#endif


// Instantiate templates - stype and property.

namespace Massiv {
namespace Core {

template class MASSIV_EXPORTED Orientation<>;
template class MASSIV_EXPORTED Orientation< Property >;

} // namespace Massiv::Core
} // namespace Massiv
