/** @file orientation.h
 * Heading-pitch-bank orientation.
 *
 * Documentation and method names of all orientation-related classes assume
 * the following world-space coordinate system:
 *
 * The world-space coordinate system is right handed, with @a x axis pointing
 * "east", @a y axis pointing "north" and @a z axis pointing "up".
 *
 * Whenever speaking about clockwise or counter-clockwise rotations about
 * an axis, we assume that we're looking towards the axis. Default rotation
 * direction is counter-clockwise. If you are lost, use right-hand rule:
 * Imagine you grab the axis in your right hand, with your thumb pointing in
 * the direction of the axis. Then direction of your fingers corresponds to
 * counter-clockwise rotation about the axis.
 *
 * Another example. Imagine you're standing in our world looking north,
 * @a x axis points to your right, @a y axis points forward and @a z axis points
 * up. This is our default entity orientation. To rotate your body to the left,
 * rotate counter-clockwise about the @a z axis. To look up, rotate your head
 * counter-clockwise about the @a x axis.
 *
 * If different coordinate system is used (especially if it's left-handed),
 * be careful when using these classes.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_ORIENTATION_H
#define MASSIV_ORIENTATION_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_MATH_H
#include "utility/math.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_STYPE_FWD_H
#include "serializer/stype_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#ifndef MASSIV_STYPE_FLOAT_H
#include "serializer/stype_float.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup template_types
 * @{
 */

/**
 * Class for storing orientation components.
 */
class MASSIV_EXPORTED OrientationData
    {
public:

    /**
     * Quantize < 0, 2 * PI > to 10 bits.
     */
    struct HeadingTraits
        {
        static const Float32 min_value;
        static const Float32 max_value;
        enum { bits = 10 };
        };

    /**
     * Quantize < -PI, PI > to 10 bits.
     */
    struct PitchBankTraits
        {
        static const Float32 min_value;
        static const Float32 max_value;
        enum { bits = 10 };
        };

public:

    /**
     * Component type used in the API.
     */
    typedef Float32                             Angle;

    /**
     * Component type used to represent and serialize heading.
     */
    typedef SFloatQ< Angle, HeadingTraits >     Heading;

    /**
     * Component type used to represent and serialize pitch.
     */
    typedef SFloatQ< Angle, PitchBankTraits >   Pitch;

    /**
     * Component type used to represent and serialize bank.
     */
    typedef SFloatQ< Angle, PitchBankTraits >   Bank;

public:

    Heading     h;
        /**< Heading. */

    Pitch       p;
        /**< Pitch. */

    Bank        b;
        /**< Bank. */
        
    }; // OrientationData.


/**
 * Serializable orientation based on head, pitch and bank angles.
 *
 * All three angles specify counter-clockwise rotation about the axes.
 * These three rotations are applied in heading-pitch-bank order.
 * @a Heading angle specifies rotation about the @a z axis, it's always in
 * range < 0, 2 * PI ).
 * @a Pitch angle specifies rotation about object's @a x axis (world @a x axis
 * after it has been rotated by @a heading). It's always in range < -PI, PI ).
 * @a Bank angle specifies rotation about object's @a y axis (world @a y axis
 * after it has been rotated by @a heading and @a pitch), it's always in range
 * < -PI, PI ).
 *
 * When using our standard world coordinate system (see orientation.h),
 * increasing @a heading will rotate an object to the left, increasing @a pitch
 * will pitch an object up towards the @a z axis ("look up"), increasing
 * @a bank will bank an object to the right (tilt right).
 *
 * Sometimes @a heading is called yaw, and @a bank is called roll.
 *
 * See orientation.h for more info about standard coordinate systems
 * and rotations.
 *
 * Although angle-based representation of orientation is easy to understand,
 * it has many disadvantages, such as gimbal lock problem (when @a pitch is
 * PI / 2, @a heading and @a bank represent rotation about the same axis),
 * simply interpolating the angles may give you weird results, etc. However,
 * it's good enough for us now. Quaternions are too complex and tricky.
 *
 * @todo More operations. Add/subtract/interpolate.
 */
template< class Base >
class MASSIV_EXPORTED Orientation : public Base
    {
private:

    /// This class.
    typedef Orientation< Base >     Self;

    /// Superclass.
    typedef Base                    Super;

    /// To make MSVC6 happy.
    typedef Serializer::Description SerializerDescription;

protected:
    
    using Base::default_constructed;
    using Base::read_requested;
    using Base::write_requested;    

public:

    /**
     * Component type used in the API.
     *
     * Angle in radians.
     */
    typedef OrientationData::Angle  Angle;

    /**
     * FloatTraits for Angle.
     */
    typedef FloatTraits< Angle >    AngleTraits;

    /**
     * Corresponding native type.
     */
    typedef OrientationData         NativeType;

    /**
     * Corresponding property type.
     */
    typedef POrientation            PropertyType;

    /**
     * Corresponding stype.
     */
    typedef SOrientation            SerializableType;

public:
    
    /**
     * Default constructor.
     */
    INLINE Orientation();

    /**
     * Constructor.
     */
    INLINE Orientation
        (
        Angle the_heading,
        Angle the_pitch,
        Angle the_bank
        );

    /**
     * Create Orientation from specified orientation data.
     */
    INLINE Orientation
        (
        const OrientationData & data
        );

    /**
     * Copy constructor.
     */
    INLINE Orientation
        (
        const Orientation & other
        );

    /**
     * Assignment operator.
     */
    INLINE Orientation & operator=
        (
        const Orientation & other
        );

    /**
     * Assignment operator.
     */
    INLINE Self & operator=
        (
        const OrientationData & data
        );
        
    /**
     * Cast to OrientationData.
     */
    INLINE operator const OrientationData & () const;

    /**
     * Compare orientations.
     */
    INLINE bool operator==
        (
        const Orientation & orientation
        ) const;
    
    /**
     * Compare orientations.
     */
    INLINE bool operator!=
        (
        const Orientation & orientation
        ) const;

public:

    /**
     * Initialize orientation from normalized forward vector.
     *
     * Vector pointing to the north (for example (0, 1, 0)) corresponds to
     * zero heading, vector pointing to the west (for example (-1, 0, 0))
     * corresponds to PI / 2 heading, etc. Bank will be always set to 0.
     */
    void set_from_forward_vector
        (
        const SVector3 & forward_vector
        );

    /**
     * Initialize orientation from normalized up vector.
     *
     * Heading is set from the @a up_vector heading (see #set_from_forward_vector())
     * if it's pointing upwards and from opposite vector if it's pointing downwards.
     * Bank will be always set to 0.
     *
     * You should use #set_from_forward_vector() instead of this if possible, because
     * it's more natural to determine "up" from "forward" than "forward" from "up".
     */
    void set_from_up_vector
        (
        const SVector3 & up_vector
        );
        
public:

    //@{
    /** Getter. */
    INLINE Angle heading() const;

    INLINE Angle pitch() const;

    INLINE Angle bank() const;
    //@}

    /** SOrientation/POrientation conversion. */
    INLINE const OrientationData & read() const;

    //@{
    /** Setter. */
    INLINE void set_heading
        (
        Angle new_heading
        );

    INLINE void set_pitch
        (
        Angle new_pitch
        );

    INLINE void set_bank
        (
        Angle new_bank
        );
    //@}

public:

    /**
     * Get ( 1, 0, 0 ) vector transformed by the orientation.
     */
    SVector3 right_vector() const;

    /**
     * Get ( 0, 1, 0 ) vector transformed by the orientation.
     */
    SVector3 forward_vector() const;

    /**
     * Get ( 0, 0, 1 ) vector transformed by the orientation.
     */
    SVector3 up_vector() const;

    /**
     * Get all three directional vectors.
     *
     * These vectors form orthogonal base of unit vectors. 3x3 matrix
     * containing @a right, @a forward and @a up vectors in the first, second
     * and third column is rotation matrix - multiply column vector by the
     * matrix (matrix on the left, vector on the right) to apply rotation
     * specified by the orientation to the vector. Because the matrix is
     * orthogonal, its inverse is equal to transposed matrix - place the
     * vectors in matrix rows and multiply the matrix by column vector to apply
     * inverse rotation to the vector.
     */
    void rotation_base
        (
        SVector3 & right_vector,
        SVector3 & forward_vector,
        SVector3 & up_vector
        ) const;

    /**
     * Apply rotation to given point.
     */
    void rotate
        (
        SVector3 & point
        ) const;

    /**
     * Apply inverse rotation to given point.
     */
    void inverse_rotate
        (
        SVector3 & point
        ) const;

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &                    bos,
        const SerializerDescription &   desc
        ) const;

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &                    bis,
        const SerializerDescription &   desc
        );

    /**
     * Serialize to text stream.
     */
    INLINE void text_write
        (
        TextWriter &                    tw,
        const SerializerDescription &   desc
        ) const;

    /**
     * Read from text stream.
     */
    INLINE void text_read
        (
        TextReader &                    tr,
        const SerializerDescription &   desc
        );

public:

    /**
     * Convert orientation to string.
     */
    std::string to_string() const;

    /**
     * Compute hash code.
     *
     * Warning: returns constant 0, see sources.
     */
    INLINE int hash() const;

    /**
     * Compute occupied memory size.
     */
    INLINE int memory_size() const;

    /**
     * Guess bitstream serialization size (in bits).
     */
    INLINE int bitstream_size() const;

private:

    /**
     * Clamp-periodic to < 0, 2 * PI ).
     */
    static INLINE Angle normalize_heading
        (
        Angle heading
        );

    /**
     * Clamp-periodic to < -PI, PI ).
     */
    static INLINE Angle normalize_pitch
        (
        Angle pitch
        );

    /**
     * Clamp-periodic to < -PI, PI ).
     */
    static INLINE Angle normalize_bank
        (
        Angle bank
        );

private:

    OrientationData data;
        /**< The orientation data. */

    }; // class Orientation

/** @} */ // addtogroup attribute_types


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "orientation_inline.h"
#endif


#endif // MASSIV_ORIENTATION_H
