/** @file orientation_inline.h
 * Heading-pitch-bank orientation (inline methods).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_TEXT_WRITER_H
#include "utility/text_writer.h"
#endif


namespace Massiv {
namespace Core {


template< class Base >
Orientation< Base >::Orientation()
    {
    data.h = 0.0f;
    data.p = 0.0f;
    data.b = 0.0f;
    
    default_constructed();
    }


template< class Base >
Orientation< Base >::Orientation
    (
    Angle the_heading,
    Angle the_pitch,
    Angle the_bank
    ) 
    {
    data.h = normalize_heading( the_heading );
    data.p = normalize_pitch( the_pitch );
    data.b = normalize_bank( the_bank );
    }


template< class Base >
Orientation< Base >::Orientation
    (
    const OrientationData & d
    )
    {
    data.h = d.h;
    data.p = d.p;
    data.b = d.b;
    }


template< class Base >
Orientation< Base >::Orientation
    (
    const Orientation & other
    ) : Super( other )
    {
    other.read_requested();
    data.h = other.data.h;
    data.p = other.data.p;
    data.b = other.data.b;
    }


template< class Base >
Orientation< Base > & Orientation< Base >::operator=
    (
    const Orientation & other
    )
    {
    write_requested();
    other.read_requested();
    data.h = other.data.h;
    data.p = other.data.p;
    data.b = other.data.b;
    return *this;
    }


template< class Base >
Orientation< Base > & Orientation< Base >::operator=
    (
    const OrientationData & d
    )
    {
    write_requested();

    data.h = d.h;
    data.p = d.p;
    data.b = d.b;

    return *this;
    }


template< class Base >        
Orientation< Base >::operator const OrientationData & () const
    {
    return read();
    } 


template< class Base >        
bool Orientation< Base >::operator==
    (
    const Orientation & orientation
    ) const
    {
    return  data.h == orientation.data.h &&
            data.p == orientation.data.p &&
            data.b == orientation.data.b;
    }


template< class Base >        
bool Orientation< Base >::operator!=
    (
    const Orientation & orientation
    ) const
    {
    return !operator==( orientation );
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::heading() const
    {
    read_requested();
    return data.h;
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::pitch() const
    {
    read_requested();
    return data.p;
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::bank() const
    {
    read_requested();
    return data.b;
    }


template< class Base >
const OrientationData & Orientation< Base >::read() const
    {
    read_requested();
    return data;
    };


template< class Base >
void Orientation< Base >::set_heading
    (
    Angle new_heading
    )
    {
    write_requested();
    data.h = normalize_heading( new_heading );
    }


template< class Base >
void Orientation< Base >::set_pitch
    (
    Angle new_pitch
    )
    {
    write_requested();
    data.p = normalize_pitch( new_pitch );
    }


template< class Base >
void Orientation< Base >::set_bank
    (
    Angle new_bank
    )
    {
    write_requested();
    data.b = normalize_bank( new_bank );
    }


template< class Base >
void Orientation< Base >::text_write
    (
    TextWriter &                    tw,
    const SerializerDescription &   desc
    ) const
    {
    read_requested();
    data.h.text_write( tw, desc );
    tw.write_space();
    data.p.text_write( tw, desc );
    tw.write_space();
    data.b.text_write( tw, desc );
    }


template< class Base >
void Orientation< Base >::text_read
    (
    TextReader &                    tr,
    const SerializerDescription &   desc
    )
    {
    write_requested();
    data.h.text_read( tr, desc );
    data.p.text_read( tr, desc );
    data.b.text_read( tr, desc );
    }


template< class Base >
int Orientation< Base >::hash() const
    {
    // Only insane people would use Orientation as a hash key.
    // If they want to do so, they should change this method.

    return 0;
    }


template< class Base >
int Orientation< Base >::memory_size() const
    {
    return sizeof( *this );
    }


template< class Base >
int Orientation< Base >::bitstream_size() const
    {
    return data.h.bitstream_size() +
        data.p.bitstream_size() +
        data.b.bitstream_size();
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::normalize_heading
    (
    Angle heading
    )
    {
    return AngleTraits::clamp_periodic( heading, 0.0f, 2.0f * MATHF_PI );
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::normalize_pitch
    (
    Angle pitch
    )
    {
    return AngleTraits::clamp_periodic( pitch, -MATHF_PI, MATHF_PI );
    }


template< class Base >
typename Orientation< Base >::Angle Orientation< Base >::normalize_bank
    (
    Angle bank
    )
    {
    return normalize_pitch( bank );
    }


} // namespace Massiv::Core
} // namespace Massiv
