/** @file prop_array.h
 * Array property.
 */

/*
 * begin        : Tue Nov 16 2001
 * copyright    : (C) 2001, 2002, 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_ARRAY_H
#define MASSIV_PROP_ARRAY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_GLOBAL_H
#include "global/global.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_DEBUG_STORAGE_H
#include "utility/debug_storage.h"
#endif

#ifndef MASSIV_SERIALIZER_H
#include "serializer/serializer.h"
#endif

#ifndef MASSIV_SERIALIZER_INT_H
#include "serializer/serializer_int.h"
#endif

#ifndef MASSIV_SERIALIZER_VLI_H
#include "serializer/serializer_vli.h"
#endif

#ifndef MASSIV_SERIALIZER_BOOL_H
#include "serializer/serializer_bool.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h"
#endif

#ifndef MASSIV_TEXT_WRITER_H
#include "utility/text_writer.h"    // FIXME: nejak to zrusit
#endif

#ifndef MASSIV_TEXT_READER_H
#include "utility/text_reader.h"    // FIXME: nejak to zrusit
#endif

#ifdef OPTION_DEBUG
#include "utility/debug.h"
#endif

#include <vector>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/**
 * Abstract interface implemented by array property.
 * Accesses array elements and provides information about an array.
 */
class ArrayInfo
    {
public:

    /** Does nothing. Defined just to get rid of GCC warnings. */
    virtual ~ArrayInfo();

    /** Returns array size. */
    virtual std::size_t array_get_size() const = 0;

    /** Clears array. */
    virtual void array_clear() = 0;
    
    /** Resizes array. */
    virtual void array_resize
        (
        std::size_t new_size
        ) = 0;

    /** Returns element stored at @c index. */
    virtual Property * array_get_element
        (
        std::size_t index
        ) = 0;

    }; // ArrayInfo

/**
 * Base class for PValueArray or PArray.
 * Defines array-like storage for a serializable @c Type.
 */
template <class Type>
class PArrayBase : public Property, public ArrayInfo, public DebugStorage
    {
protected:

    typedef Serializer::Description             Description;
    typedef PArrayBase<Type>                    ThisBase;
    typedef std::vector<Type>                   Storage;

    typedef typename Storage::size_type         SizeType;
    typedef typename Storage::difference_type   DifferenceType;
    typedef typename Storage::value_type        ValueType;
    typedef typename Storage::reference         ReferenceType;
    typedef typename Storage::const_reference   ConstReferenceType;
    typedef typename Storage::const_iterator    StorageConstIterator;

    enum Constants
        {
        MAX_PRINT_ELEMENTS  =   8,
            /**< Maximum number of elements that are dumped by @c to_string(). */
        MAX_ELEMENTS_PER_LINE = 16
            /**< Maximum number of elements that are serialized to a single line. */
        };

protected:

    //
    // Constructors.
    //

    /** Default constructor. */
    PArrayBase() : max_size( 0 )
        {
        default_constructed();
        }

    /**
     * Copy constructor.
     * @c storage is expected to be copied by subclass.
     */
    PArrayBase
        (
        const ThisBase & base
        ) : Property( base ),
            max_size( base.max_size )
        {
        }

    /** Destructor. */
    ~PArrayBase()
        {
        }

public:

    //
    // Array interface.
    //

    /** Returns number of elements stored in the array. */
    SizeType size() const
        {
        read_requested();
        return storage.size();
        }

    /** Tests if the array is empty. */
    bool empty() const
        {
        return !size();
        }

public:

    //
    // Property extensions.
    //

    /**
     */
    virtual std::string to_string() const
        {
        std::string output;
        SizeType    count;
        SizeType    index;

        count = size();

        if( count > MAX_PRINT_ELEMENTS )
            {
            count = MAX_PRINT_ELEMENTS;
            }

        output = "{ ";

        for( index = 0; index < count; index++ )
            {
            if( index )
                {
                output += ", ";
                }

            output += storage[ index ].to_string();
            }

        if( size() > count )
            {
            output += ", ...";
            }

        output += " }";

        return output;
        }
    
    /**
     */
    virtual int memory_size() const
        {
        SizeType    accumulator = 0;
        SizeType    index = 0;

        while( index < storage.size() )
            {
            accumulator += storage[ index ].memory_size();
            ++index;
            }

        accumulator += sizeof( *this );

        return accumulator;
        }

    /**
     */
    virtual int bitstream_size() const
        {
        SizeType    accumulator = 0;
        SizeType    index = 0;

        while( index < storage.size() )
            {
            accumulator += storage[ index ].bitstream_size();
            ++index;
            }

        return accumulator * 6 / 5;
        }
    
    /**
     */
    virtual PropertyType get_type() const
        {
        return Property::TYPE_CONTAINER;
        }
    
    /**
     */
    ArrayInfo * get_array_info()
        {
        return this;
        }

protected:

    //
    // Internal implementation.
    //
    
    /** Clears array. */
    void internal_clear()
        {
        write_requested();
        invalidate_iterators();

        resize_time = Global::replica_time();
        storage.clear();
        }

    /** Resizes array. */
    void internal_resize
        (
        SizeType            new_size,
        ConstReferenceType  filler
        )
        {
        SizeType old_size = storage.size();
        
        // Test size.

        if( new_size == old_size )
            {
            invalidate_iterators();
            return;
            }

        DebugStorage::range_error( new_size >= 0, "resize()" );

        // Notify about state change.
        // Dirty the whole property hierachy.

        write_requested();
        invalidate_iterators();

        // Resize.

        resize_time = Global::replica_time();
        storage.resize( new_size, filler );
        
        if( max_size < new_size )
            {
            max_size = new_size;
            }
        }
    
    /**
     * Deletes element at position @c index.
     * The last array item remains invalid and must be deleted manually.
     */
    void internal_erase
        (
        SizeType    index
        )
        {
        SizeType i;

        DebugStorage::range_error( index >= 0 && index < size(), "erase()" );
        
        write_requested();
        invalidate_iterators();
        
        for( i = index + 1; i < storage.size(); i++ )    
            {
            storage[ i - 1 ] = storage[ i ];
            }
        }
    
    /**
     * Inserts element at position @c index. The elements from [ index, end )
     * range will be shifted one element right to make room for @c filler.
     * The last array item will be overwritten.
     */
    void internal_insert
        (
        SizeType            index,
        ConstReferenceType  filler = Type()
        )
        {
        SizeType i;

        write_requested();
        invalidate_iterators();

        for( i = storage.size() - 1; i > index; i-- )
            {
            storage[ i ] = storage[ i - 1 ];
            }

        storage[ i ] = filler;
        }
    
    /**
     * Returns number of elements (@c Derived::PAGE_SIZE aligned) that are dirty
     * relative to @c time.
     * The scanning begins at element @c it. Once function returns, @c it
     * points either to the end of @c storage or to the first element of the
     * "closest" clean page.
     *
     * If @a compare_times is @c false, all elements are assumed dirty.
     *
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     */
    template <class Derived>
    UInt32 dirty_count
        (
        StorageConstIterator &  it,
        bool                    compare_times,
        const ReplicaTime &     time,
        const Derived &         downcasted
        ) const
        {
        UInt32 count = 0;
        
        // Count.

        #ifdef OPTION_DEBUG

        if( it != storage.end() )
            {
            hope( ( it - storage.begin() ) % Derived::PAGE_SIZE == 0 );
            }

        #endif

        if( compare_times )
            {
            while( it != storage.end() )
                {
                const ReplicaTime & dirty_time = 
                    downcasted.get_element_dirty_time( it );

                if( !time.is_older_than( dirty_time ) )
                    {
                    // This page is clean relative to @c time.
                    break;
                    }

                if( Derived::IS_VALUE_ARRAY && it - storage.begin() + 
                    Derived::PAGE_SIZE <= static_cast<Int32>( storage.size() ) )
                    {
                    // PValueArray ensures that @c downcasted.get_element_dirty_time()
                    // always returns the same value for all the elements belonging
                    // to the same page.

                    count += Derived::PAGE_SIZE;
                    it += Derived::PAGE_SIZE;
                    }
                else
                    {
                    // Must check all elements in the page.

                    ++count;
                    ++it;
                    }
                }
            }
        else
            {
            StorageConstIterator storage_end = storage.end();
            count = storage_end - it;
            it = storage_end;
            }

        // Align.
        
        count = Derived::PAGE_SIZE * ( ( count + Derived::PAGE_SIZE - 1 ) /
            Derived::PAGE_SIZE );
        hope( count % Derived::PAGE_SIZE == 0 );

        return count;
        }

    /**
     * Returns number of elements (@c Derived::PAGE_SIZE aligned) that are clean
     * relative to @c time.
     * The scanning begins at element @c it. Once function returns, @c it
     * points either to the end of @c storage or to the first element of the
     * "closest" dirty page.
     *
     * If @a compare_times is @c false, all elements are assumed dirty.
     *
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     */
    template <class Derived>
    UInt32 clean_count
        (
        StorageConstIterator & it,
        bool                   compare_times,
        const ReplicaTime &    time,
        const Derived &        downcasted
        ) const
        {
        UInt32 count = 0;

        // Count.

        #ifdef OPTION_DEBUG

        if( it != storage.end() )
            {
            hope( ( it - storage.begin() ) % Derived::PAGE_SIZE == 0 );
            }

        #endif

        if( compare_times )
            {
            while( it != storage.end() )
                {
                const ReplicaTime & dirty_time = 
                    downcasted.get_element_dirty_time( it );

                if( time.is_older_than( dirty_time ) )
                    {
                    // This page is dirty relative to @c time.
                    break;
                    }
                
                if( Derived::IS_VALUE_ARRAY && it - storage.begin() + 
                    Derived::PAGE_SIZE <= static_cast<Int32>( storage.size() ) )
                    {
                    // PValueArray ensures that @c downcasted.get_element_dirty_time()
                    // always returns the same value for all the elements belonging
                    // to the same page.

                    count += Derived::PAGE_SIZE;
                    it += Derived::PAGE_SIZE;
                    }
                else
                    {
                    // Must check all elements in the page.

                    ++count;
                    ++it;
                    }
                }
            }
        
        // Align.

        count = Derived::PAGE_SIZE * ( ( count + Derived::PAGE_SIZE - 1 ) /
            Derived::PAGE_SIZE );
        hope( count % Derived::PAGE_SIZE == 0 );

        return count;
        }

    /**
     * Serializes to bitstream.
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     */
    template <class Derived>
    void internal_bitstr_write
        (
        BitOStream &        bos,
        const Description & desc,
        const Derived &     downcasted
        ) const
        {
        const bool compare_times = desc.compare_times;
        const ReplicaTime & time = desc.ack_time;

        // Write container structure info.

        if( time.is_older_than( resize_time ) )
            {
            // Array resized since @c time.

            Serializer::BitStrBool::write( bos, true );
            Serializer::BitStrVli::write( bos, size() );
            }
        else
            {
            Serializer::BitStrBool::write( bos, false );
            }

        // Write container elements.

        StorageConstIterator    storage_it = storage.begin();
        SizeType                index = 0;

        while( true )
            {
            UInt32      dirty;
            UInt32      clean;

            // Write elements stored in dirty pages.

            dirty = dirty_count( storage_it, compare_times, time, downcasted );
            Serializer::BitStrVli::write( bos, dirty / Derived::PAGE_SIZE );

            if( index + dirty > size() )
                {
                dirty = size() - index;
                }

            while( dirty-- > 0 )
                {
                downcasted.bitstr_write( index, bos, desc );
                index++;
                }

            // Skip clean elements.

            clean = clean_count( storage_it, compare_times, time, downcasted );
            Serializer::BitStrVli::write( bos, clean / Derived::PAGE_SIZE );

            if( !clean )
                {
                break;
                }

            if( index + clean > size() )
                {
                clean = size() - index;
                }

            index += clean;
            }
        }

    /**
     * Reads from bitstream.
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     */
    template <class Derived>
    void internal_bitstr_read
        (
        BitIStream &        bis,
        const Description & desc,
        Derived &           downcasted
        )
        {
        bool        need_resize;
        UInt32      new_size;
        SizeType    index;

        // Read container structure info.

        Serializer::BitStrBool::read( bis, need_resize );

        if( need_resize )
            {
            Serializer::BitStrVli::read( bis, new_size );
            downcasted.resize( new_size );
            }

        // Read container elements.

        index = 0;

        while( true )
            {
            UInt32      dirty;
            UInt32      clean;

            // Read dirty elements.

            Serializer::BitStrVli::read( bis, dirty );
            dirty *= Derived::PAGE_SIZE;

            if( index + dirty > storage.size() )
                {
                dirty = storage.size() - index;
                }

            while( dirty-- > 0 )
                {
                downcasted.bitstr_read( index, bis, desc );
                index++;
                }

            // Skip clean elements.

            Serializer::BitStrVli::read( bis, clean );

            if( !clean )
                {
                break;
                }

            clean *= Derived::PAGE_SIZE;

            if( index + clean > storage.size() )
                {
                clean = storage.size() - index;
                }

            index += clean;
            }
        }

    /**
     * Serializes to text stream.
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     *
     * First token in serialized array is either @a resize or @a noresize.
     * @a Resize is followed by new number of array elements. Element updates
     * follow, each starting on a new line. Update may begin with:
     * - @a element. This is update of element from a dirty page. It's followed
     *   by element serialization defined in @a downcasted.
     * - @a elements, followed by a count of element updates. It's directly
     *   followed by elements serialization.
     *   @c elements is used to reduce serialization overhead when "element"
     *   consumes much more space than the actual element serialization.
     * - @a skip, followed by a positive number. Given number of elements are
     *   not dirty, so no updates for them are written.
     * - @a endarray, followed by a number. End of array serialization.
     *   The number has the same meaning as in @a skip.
     *
     * Number of @a element updates plus sum of counts specified by @a elements
     * and @a skip and @a endarray will be always equal to array size.
     */
    template <class Derived>
    void internal_text_write
        (
        TextWriter &        tw,
        const Description & desc,
        const Derived &     downcasted
        ) const
        {
        const bool compare_times = desc.compare_times;
        const ReplicaTime & time = desc.ack_time;

        // Write container structure info.

        if( time.is_older_than( resize_time ) )
            {
            // Array resized since @c time.

            tw.line() << "resize ";
            const SUInt32 s = size();
            s.text_write( tw, desc );
            }
        else
            {
            tw.line() << "noresize";
            }
        tw.end_line();

        // Write container elements.

        StorageConstIterator    storage_it = storage.begin();
        SizeType                index = 0;

        while( true )
            {
            // Write elements stored in dirty pages.

            SUInt32 dirty_elements = dirty_count
                ( storage_it, compare_times, time, downcasted );

            if( index + dirty_elements > size() )
                {
                dirty_elements = size() - index;
                }

            if( !Derived::IS_VALUE_ARRAY )
                {
                // Write each dirty element on a new line that is prefixed
                // by "element" keyword.

                while( dirty_elements-- > 0 )
                    {
                    tw.line() << "element ";
                    tw.change_indent( 4 );
                    downcasted.text_write( index, tw, desc );
                    tw.change_indent( -4 );
                    tw.end_line();
                    index++;
                    }
                }
            else
                {
                // Write "elements <count>" prefix and append element
                // serialization to the same line.
                // If there are too many dirty elements so that they would
                // not fit to a single line, more lines will be written.

                while( dirty_elements > 0 )
                    {
                    SUInt32 count = dirty_elements;

                    if( count > MAX_ELEMENTS_PER_LINE )
                        {
                        count = MAX_ELEMENTS_PER_LINE;
                        }

                    tw.line() << "elements ";
                    count.text_write( tw, desc );
                    dirty_elements -= count;

                    while( count-- > 0 )
                        {
                        tw.write_space();
                        downcasted.text_write( index, tw, desc );
                        index++;
                        }

                    tw.end_line();
                    }
                }

            // Count clean elements.

            SUInt32 clean_elements = clean_count
                ( storage_it, compare_times, time, downcasted );

            if( index + clean_elements >= size() )
                {
                // Run of clean elements reaches end of the array.
                // Write number of clean elements to make things even more safe.

                const SUInt32 end_elements = size() - index;
                tw.line() << "endarray ";
                end_elements.text_write( tw, desc );
                break;
                }
    
            // Skip clean elements.

            tw.line() << "skip ";
            clean_elements.text_write( tw, desc );
            tw.end_line();

            index += clean_elements;
            }
        }

    /**
     * Reads from text stream.
     * @c downcasted is expected to be downcasted *this that points to
     * either PValueArray or PArray.
     */
    template <class Derived>
    void internal_text_read
        (
        TextReader &        tr,
        const Description & desc,
        Derived &           downcasted
        )
        {
        SString token;

        // Read container structure info.

        token.text_read( tr, desc );
        if( token == "resize" )
            {
            SUInt32 new_size;
            new_size.text_read( tr, desc );
            downcasted.resize( new_size );
            }
        else if( token != "noresize" )
            {
            throw ExUnexpectedToken( tr.unexpected_token( "resize or noresize" ) );
            }

        // Read container elements.

        const SizeType element_count = size();
        SizeType index = 0;
        while( true )
            {
            // Read entry type.

            tr.next_line_throw();
            token.text_read( tr, desc );

            // End of array serialization.

            if( token == "endarray" )
                {
                SUInt32 end_count;
                end_count.text_read( tr, desc );
                index += end_count;
                break;
                }

            // Dirty element.

            else if( token == "element" )
                {
                if( index >= element_count )
                    {
                    throw ExUnexpectedToken( tr.unexpected_token( "endarray" ) );
                    }

                downcasted.text_read( index, tr, desc );
                index++;
                }

            // Dirty elements.

            else if( token == "elements" )
                {
                SUInt32 dirty_count;
                dirty_count.text_read( tr, desc );

                if( index + dirty_count > element_count )
                    {
                    std::ostringstream ss;
                    ss << "integer between 0 and " << ( element_count - index );
                    throw ExUnexpectedValue( tr.unexpected_value( ss.str(), dirty_count ) );
                    }

                while( dirty_count-- > 0 )
                    {
                    downcasted.text_read( index, tr, desc );
                    index++;
                    }
                }

            // Skip clean elements.

            else if( token == "skip" )
                {
                SUInt32 skip_count;
                skip_count.text_read( tr, desc );

                if( index + skip_count > element_count )
                    {
                    std::ostringstream ss;
                    ss << "integer between 0 and " << ( element_count - index );
                    throw ExUnexpectedValue( tr.unexpected_value( ss.str(), skip_count ) );
                    }
                else
                    {
                    index += skip_count;
                    }
                }

            // Invalid entry.

            else
                {
                throw ExUnexpectedToken( tr.unexpected_token( "element, skip or endarray" ) );
                }
            }

        if( index != element_count )
            {
            std::ostringstream ss;
            ss << "Invalid number of array element updates: Expected " <<
                element_count << ", read " << index << ".\nLine " <<
                tr.line_number() << ": " << tr.str();
            throw ExSerializationError( ss.str() );
            }
        }

protected:

    Storage             storage;
        /**< Storage that holds stored items. */
    SizeType            max_size;
        /**< Maximum array size. It is actually maximum of @c storage.size() with
             respect to array's lifetime on this node. This value is used when
             creating new default-constructed elements. Elements beyond the
             (max_size - 1)-th element don't have to be marked dirty - their contents 
             won't be replicated to replica clients. This value is not serialized.

             @note Arrays must not be swapped out when someone requests their
                   replicas - #max_size would be forgotten.*/
    ReplicaTime         resize_time;
        /**< Time of the last resize. */
    }; // PArrayBase

/**
 * Simple array property.
 *
 * The array is able to hold lightweight serializable @c STypes organized
 * into pages. When an element gets dirty the page will be made dirty too
 * (the whole page will be serialized then).
 *
 * @note @c SType must not be a property.
 * @note To create a two-dimensional array use property array of arrays
 *  of @c STypes.
 *
 * <h3>Notes on PValueArray implementation</h3>
 *
 * Since @c SType is expected not to be a Property we have to manually
 * call write_requested() / read_requested() when accessing stored
 * elements, also pages need be made dirty too. 
 * We do not have to implement "property extensions" here, because
 * we are not working with properties.
 */
template <class SType>
class PValueArray : public PArrayBase<SType>
    {
public:

    typedef Serializer::Description                 Description;
    typedef PValueArray<SType>                      ThisArray;
    typedef PArrayBase<SType>                       Super;
    
    typedef std::vector<ReplicaTime>                DirtyTimes;
    typedef DirtyTimes::const_iterator              DirtyTimesConstIterator;

    typedef typename Super::ValueType               ValueType;
    typedef typename Super::SizeType                SizeType;
    typedef typename Super::DifferenceType          DifferenceType;
    typedef typename Super::ReferenceType           ReferenceType;
    typedef typename Super::ConstReferenceType      ConstReferenceType;
    typedef typename Super::Storage                 Storage;
    typedef typename Super::StorageConstIterator    StorageConstIterator;

    typedef SafeRandomAccessIterator
                <
                SequenceIteratorTraits<ThisArray>
                >   Iterator;

    typedef SafeRandomAccessIterator
                <
                SequenceConstIteratorTraits<ThisArray>
                >   ConstIterator;

    enum Constants
        {
        PAGE_SIZE = 8,
        IS_VALUE_ARRAY = true
        };

    using Super::size;
    using Super::empty;

protected:

    using Super::internal_clear;
    using Super::internal_resize;
    using Super::internal_erase;
    using Super::internal_insert;
    using Super::dirty_count;
    using Super::clean_count;
    using Super::internal_bitstr_write;    
    using Super::internal_bitstr_read;
    using Super::internal_text_write;
    using Super::internal_text_read;
    using Super::storage;
    using Super::max_size;
    using Super::resize_time;
    using Super::read_requested;
    using Super::write_requested;

public:

    //
    // Constructors.
    //

    /** Default constructor. */
    PValueArray()
        {
        }

    /** Copy constructor. */
    PValueArray
        (
        const ThisArray & array
        ) : Super( array )
        {
        if( !array.is_default_constructed() )
            {
            operator=( array );
            }
        }

    /** Constructor. */
    PValueArray
        (
        const Storage & array
        )
        {
        operator=( array );
        }

    /** Destructor. */
    ~PValueArray()
        {
        }

public:

    //
    // Array interface.
    //

    /** Returns iterator pointing to the beginning of the array. */
    Iterator begin()
        {
        typedef typename Iterator::BeginTag BeginTag;
        return Iterator( this, BeginTag() );
        }

    /** Returns const iterator pointing to the beginning of the array. */
    ConstIterator begin() const
        {
        typedef typename ConstIterator::BeginTag BeginTag;
        return ConstIterator( const_cast<ThisArray *>( this ), BeginTag() );
        }

    /** Returns iterator pointing to the end of the array. */
    Iterator end()
        {
        return begin() + size();
        }

    /** Returns const iterator pointing to the end of the array. */
    ConstIterator end() const
        {
        return begin() + size();
        }

    /** Returns const reference to a stored item. */
    ConstReferenceType operator[]
        (
        SizeType index
        ) const
        {
        return at( index );
        }

    /** Returns reference to a stored item. */
    ReferenceType operator()
        (
        SizeType index
        )
        {
        return write_at( index );
        }

    /** Returns const reference to a stored item. */
    ConstReferenceType at
        (
        SizeType index
        ) const
        {
        DebugStorage::range_error( index >= 0 && index < size(), "at()" );

        read_requested();
        return storage[ index ];
        }

    /** Returns reference to a stored item. */
    ReferenceType write_at
        (
        SizeType index
        )
        {
        DebugStorage::range_error( index >= 0 && index < size(), "write_at()" );

        write_requested();
        page_dirty_times[ index / PAGE_SIZE ] = Global::replica_time();

        return storage[ index ];
        }
    
    /**
     * Writes @c value to the array. Makes the array dirty only iff
     * @c value differs from the current value.
     */
    void write_at
        (
        SizeType            index,
        ConstReferenceType  value
        )
        {
        // FIXME: To prevent operator!=() ambiguity each SType should
        // define compare() method.

        /*if( at( index ) != value )*/
            {
            write_at( index ) = value;
            }
        }

    /** Clears array. */
    void clear()
        {
        internal_clear();
        page_dirty_times.clear();
        }

    /** Resizes array. */
    void resize
        (
        SizeType            new_size,
        ConstReferenceType  filler,
        bool                filler_is_default_constructed
        )
        {
        SizeType old_size = storage.size();
        SizeType old_max_size = max_size;
        SizeType page = old_size / PAGE_SIZE;
        
        Super::internal_resize( new_size, filler );

        // Initialize dirty times for the new pages.
        // All pages between old last page and page containing ( max_size - 1)-th
        // must be marked as dirty. We don't know whether replication clients
        // have these pages or if they have been already deleted.

        page_dirty_times.resize( ( new_size + PAGE_SIZE - 1 ) / PAGE_SIZE );
        
        if( old_max_size )
            {
            SizeType old_max_page = ( old_max_size - 1 ) / PAGE_SIZE;
            while( page <= old_max_page && page < page_dirty_times.size() )
                {
                page_dirty_times[ page ] = resize_time;
                page++;
                }
            }

        // If the remaining elements in the array are default-constructed,
        // we must leave pages clean too.

        if( !filler_is_default_constructed )
            {
            while( page < page_dirty_times.size() )
                {
                page_dirty_times[ page ] = resize_time;
                page++;
                }
            }
        }
    
    /** Resizes array. */
    void resize
        (
        SizeType            new_size,
        ConstReferenceType  filler
        )
        {
        resize( new_size, filler, false );
        }
    
    /** Resizes array. */
    void resize
        (
        SizeType new_size
        )
        {
        resize( new_size, SType(), true );
        }
    
    /** Deletes element at position @c index. */
    void erase
        (
        SizeType index
        )
        {
        SizeType i;

        Super::internal_erase( index );
        
        for( i = index / PAGE_SIZE; i < page_dirty_times.size(); i++ )
            {
            page_dirty_times[ i ] = Global::replica_time();
            }

        pop_back();
        }
    
    /** Inserts element at position @c index. */
    void insert
        (
        SizeType            index,
        ConstReferenceType  filler = SType()
        )
        {
        SizeType i;

        DebugStorage::range_error( index >= 0 && index <= size(), "insert()" );

        resize( size() + 1 );
        Super::internal_insert( index, filler );
        
        for( i = index / PAGE_SIZE; i < page_dirty_times.size(); i++ )
            {
            page_dirty_times[ i ] = Global::replica_time();
            }
        }

    /** Inserts @c element at the end of the array. */
    void push_back
        (
        ConstReferenceType  element
        )
        {
        resize( size() + 1, element );
        }

    /** Deletes the last element from the array. */
    void pop_back()
        {
        resize( size() - 1 );
        }

    /** Copies array elements. */
    ThisArray & operator=
        (
        const ThisArray & array
        )
        {
        SizeType i;

        write_requested();
        resize( array.size() );

        for( i = 0; i < array.size(); i++ )
            {
            storage[ i ] = array[ i ];
            }

        for( i = 0; i < page_dirty_times.size(); i++ )
            {
            page_dirty_times[ i ] = Global::replica_time();
            }

        return *this;
        }
    
    /** Copies array elements from std::vector. */
    ThisArray & operator=
        (
        const Storage & array
        )
        {
        SizeType i;

        write_requested();
        resize( array.size() );

        for( i = 0; i < array.size(); i++ )
            {
            storage[ i ] = array[ i ];
            }

        for( i = 0; i < page_dirty_times.size(); i++ )
            {
            page_dirty_times[ i ] = Global::replica_time();
            }

        return *this;
        }

    /** Compares arrays. */
    friend bool operator==
        (
        const ThisArray &   array1,
        const ThisArray &   array2
        )
        {
        SizeType index;

        if( array1.size() != array2.size() )
            {
            return false;
            }

        for( index = 0; index < array1.size(); index++ )
            {
            if( array1[ index ] != array2[ index ] )
                {
                return false;
                }
            }

        return true;
        }

    /** Compares arrays. */
    friend bool operator!=
        (
        const ThisArray &   array1,
        const ThisArray &   array2
        )
        {
        return !operator==( array1, array2 );
        }

public:

    //
    // Property extensions.
    //
    
    /**
     * Guesses memory size occupied by the array.
     */
    virtual int memory_size() const
        {
        return Super::memory_size() + page_dirty_times.size() * 
            sizeof( ReplicaTime );
        }

    /**
     * Serializes to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        internal_bitstr_write( bos, desc, *this );
        }

    /**
     * Reads from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        internal_bitstr_read( bis, desc, *this );
        }

    /**
     * Serializes to text stream.
     */
    virtual void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        internal_text_write( tw, desc, *this );
        }

    /**
     * Reads from text stream.
     */
    virtual void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        internal_text_read( tr, desc, *this );
        }
        
public:

    //
    // Upcalls from PArrayBase.
    //
    
    /**
     * Returns dirty time of elemented pointed by @c it.
     */
    const ReplicaTime & get_element_dirty_time
        (
        StorageConstIterator it
        ) const
        {
        return page_dirty_times[ ( it - storage.begin() ) / PAGE_SIZE ];
        }

    /**
     * Serializes array element to bitstream.
     */
    void bitstr_write
        (
        SizeType            index,
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        at( index ).bitstr_write( bos, desc );
        }

    /**
     * Reads array element from bitstream.
     */
    void bitstr_read
        (
        SizeType            index,
        BitIStream &        bis,
        const Description & desc
        )
        {
        ValueType value;
        value.bitstr_read( bis, desc );
        write_at( index, value );
        }

    /**
     * Serializes array element to text stream.
     */
    void text_write
        (
        SizeType            index,
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        at( index ).text_write( tw, desc );
        }

    /**
     * Reads array element from text stream.
     */
    void text_read
        (
        SizeType            index,
        TextReader &        tr,
        const Description & desc
        )
        {
        ValueType value;
        value.text_read( tr, desc );
        write_at( index, value );
        }

public:

    //
    // ArrayInfo implementation.
    //
    
    /** Returns array size. */
    virtual std::size_t array_get_size() const
        {
        return size();
        }

    /** Clears array. */
    virtual void array_clear()
        {
        clear();
        }
    
    /** Resizes array. */
    virtual void array_resize
        (
        std::size_t new_size
        )
        {
        resize( new_size );
        }

    /** Not implemented by PValueArray. */
    virtual Property * array_get_element
        (
        std::size_t index
        )
        {
        return NULL;
        }

private:
    
    mutable DirtyTimes  page_dirty_times;
        /**< Each element stored in the array belongs to a page that
             has assigned a "dirty" time, time of the last modification. */

    }; // PValueArray

/**
 * Array property of property types with a STL-like interface.
 *
 * A simple container property that allows to store any Property-derived
 * types ranging from simple integer properties to complex Objects or other
 * containers.
 *
 * @note To create a two-dimensional array use property array of arrays
 *  of @c PTypes.
 *
 * @note Stored elements can not be referenced by ObjectPointers even if they
 *  were Objects.
 *
 * <h3>Notes on PArray implementation</h3>
 *
 * The major difference from PValueArray is that properties have to be
 * handled in a special way (initializations, attribute propagations, ...).
 * There is no need to manually call write_requested() / read_requested()
 * when accessing array elements because this is called automatically
 * by the stored elements if they are modified / read.
 * Because of the same reason no explicit @c page_dirty_times need be
 * maintained.
 */
template <class PType>
class PArray : public PArrayBase<PType>
    {
public:

    typedef Serializer::Description                 Description;
    typedef Property::Attributes                    Attributes;
    typedef Property::PropertyOperation             PropertyOperation;
    typedef Property::PropertyInfo                  PropertyInfo;

    typedef PArray<PType>                           ThisArray;
    typedef PArrayBase<PType>                       Super;

    typedef typename Super::ValueType               ValueType;
    typedef typename Super::SizeType                SizeType;
    typedef typename Super::DifferenceType          DifferenceType;
    typedef typename Super::ReferenceType           ReferenceType;
    typedef typename Super::ConstReferenceType      ConstReferenceType;
    typedef typename Super::Storage                 Storage;
    typedef typename Super::StorageConstIterator    StorageConstIterator;

    typedef SafeRandomAccessIterator
                <
                SequenceIteratorTraits<ThisArray>
                >   Iterator;

    typedef SafeRandomAccessIterator
                <
                SequenceConstIteratorTraits<ThisArray>
                >   ConstIterator;

    enum Constants
        {
        PAGE_SIZE = 1,
        IS_VALUE_ARRAY = false
        };

    using Super::size;
    using Super::empty;

protected:

    using Super::internal_clear;
    using Super::internal_resize;
    using Super::internal_erase;
    using Super::internal_insert;
    using Super::dirty_count;
    using Super::clean_count;
    using Super::internal_bitstr_write;    
    using Super::internal_bitstr_read;
    using Super::internal_text_write;
    using Super::internal_text_read;
    using Super::storage;
    using Super::max_size;
    using Super::resize_time;
    using Super::read_requested;
    using Super::write_requested;

public:

    //
    // Constructors.
    //

    /** Default constructor. */
    PArray()
        {
        }

    /** Copy constructor. */
    PArray
        (
        const ThisArray & array
        ) : Super( array )
        {
        if( !array.is_default_constructed() )
            {
            operator=( array );
            }
        }

    /** Constructor. */
    PArray
        (
        const Storage & array
        )
        {
        operator=( array );
        }

    /** Destructor. */
    ~PArray()
        {
        }

public:

    //
    // Array interface.
    //

    /** Returns iterator pointing to the beginning of the array. */
    Iterator begin()
        {
        typedef typename Iterator::BeginTag BeginTag;
        return Iterator( this, BeginTag() );
        }

    /** Returns const iterator pointing to the beginning of the array. */
    ConstIterator begin() const
        {
        typedef typename ConstIterator::BeginTag BeginTag;
        return ConstIterator( const_cast<ThisArray *>( this ), BeginTag() );
        }

    /** Returns iterator pointing to the end of the array. */
    Iterator end()
        {
        return begin() + size();
        }

    /** Returns const iterator pointing to the end of the array. */
    ConstIterator end() const
        {
        return begin() + size();
        }

    /** Returns const reference to a stored item. */
    ConstReferenceType operator[]
        (
        SizeType    index
        ) const
        {
        return at( index );
        }

    /** Returns reference to a stored item. */
    ReferenceType operator[]
        (
        SizeType    index
        )
        {
        return write_at( index );
        }

    /** Returns reference to a stored item. */
    ReferenceType operator()
        (
        SizeType    index
        )
        {
        return operator[]( index );
        }

    /** Returns const reference to a stored item. */
    ConstReferenceType at
        (
        SizeType    index
        ) const
        {
        DebugStorage::range_error( index >= 0 && index < size(), "at()" );
        return storage[ index ];
        }

    /** Returns reference to a stored item. */
    ReferenceType write_at
        (
        SizeType    index
        )
        {
        DebugStorage::range_error( index >= 0 && index < size(), "write_at()" );
        return storage[ index ];
        }
    
    /** Clears array. */
    void clear()
        {
        internal_clear();
        }

    /** Resizes array. */
    void resize
        (
        SizeType            new_size,
        ConstReferenceType  filler = PType()
        )
        {
        SizeType old_size = storage.size();
        SizeType old_max_size = max_size;
        SizeType capacity = storage.capacity();
        SizeType index = old_size;

        Super::internal_resize( new_size, filler );

        // Initialize new items.

        if( capacity < new_size )
            {
            // There was not enough space in the storage. @c resize() caused storage
            // to reallocate. The whole array / stored properties have to be
            // reinitialized.

            index = 0;
            }

        while( index < new_size )
            {
            // If new elements are default-constructed, they are not dirty.
            // That's wrong for elements at indices < old_max_size. We must explicitly
            // make those elements dirty.

            if( filler.is_default_constructed() && index < old_max_size )
                {
                storage[ index ].set_dirty_time();
                }

            #ifdef OPTION_DEBUG

            if( index >= old_size )
                {
                if( index >= old_max_size && filler.is_default_constructed() )
                    {
                    hope( storage[ index ].is_default_constructed() );
                    }
                else
                    {
                    hope( storage[ index ].get_dirty_time() == Global::replica_time() );
                    }
                }

            #endif

            storage[ index ].initialize_from_owner( *this, stored_attributes );
            index++;
            }
        }
    
    /** Deletes element at position @c index. */
    void erase
        (
        SizeType    index
        )
        {
        internal_erase( index );
        pop_back();
        }
    
    /** Inserts element at position @c index. */
    void insert
        (
        SizeType            index,
        ConstReferenceType  filler = PType()
        )
        {
        DebugStorage::range_error( index >= 0 && index <= size(), "insert()" );

        resize( size() + 1 );
        internal_insert( index, filler );
        }

    /** Inserts @c element at the end of the array. */
    void push_back
        (
        ConstReferenceType  element
        )
        {
        resize( size() + 1, element );
        }

    /** Deletes the last element from the array. */
    void pop_back()
        {
        resize( size() - 1 );
        }

    /** Copies array elements. */
    ThisArray & operator=
        (
        const ThisArray & array
        )
        {
        SizeType i;

        write_requested();
        resize( array.size() );

        for( i = 0; i < array.size(); i++ )
            {
            storage[ i ] = array[ i ];
            hope( storage[ i ].get_dirty_time() == Global::replica_time() );
            }

        return *this;
        }
    
    /** Copies array elements from std::vector. */
    ThisArray & operator=
        (
        const Storage & array
        )
        {
        SizeType i;

        write_requested();
        resize( array.size() );
        
        for( i = 0; i < array.size(); i++ )
            {
            storage[ i ] = array[ i ];
            hope( storage[ i ].get_dirty_time() == Global::replica_time() );
            }

        return *this;
        }

    /** Compares arrays. */
    friend bool operator==
        (
        const ThisArray &   array1,
        const ThisArray &   array2
        )
        {
        SizeType index;

        if( array1.size() != array2.size() )
            {
            return false;
            }

        for( index = 0; index < array1.size(); index++ )
            {
            if( array1[ index ] != array2[ index ] )
                {
                return false;
                }
            }

        return true;
        }

    /** Compares arrays. */
    friend bool operator!=
        (
        const ThisArray &   array1,
        const ThisArray &   array2
        )
        {
        return !operator==( array1, array2 );
        }

public:

    //
    // Property extensions.
    // Note: no read/write_requested() should be called from these methods.
    //
    
    /**
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        )
        {
        SizeType index;

        stored_attributes = attributes;
        Property::initialize_from_owner( direct_owner, attributes );

        for( index = 0; index < storage.size(); index++ )
            {
            storage[ index ].initialize_from_owner( *this, attributes );
            }
        }

    /**
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        )
        {
        SizeType index;

        for( index = 0; index < storage.size(); index++ )
            {
            storage[ index ].for_each_property_do( operation );
            }
        }
    
    /**
     */
    virtual bool get_owned_info
        (
        const Property *    owned,
        PropertyInfo &      info
        ) const
        {
        SizeType index;

        for( index = 0; index < storage.size(); index++ )
            {
            if( &storage[ index ] == owned )
                {
                std::ostringstream ss;

                ss << "element[ " << index << " ]";

                info.property_name = ss.str();
                return true;
                }
            }

        return false;
        }
    
public:

    //
    // Serialization.
    //

    /**
     * Serializes to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        internal_bitstr_write( bos, desc, *this );
        }

    /**
     * Reads from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        internal_bitstr_read( bis, desc, *this );
        }

    /**
     * Serializes to text stream.
     */
    virtual void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        internal_text_write( tw, desc, *this );
        }

    /**
     * Reads from text stream.
     */
    virtual void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        internal_text_read( tr, desc, *this );
        }
    
public:

    //
    // Upcalls from PArrayBase.
    //
    
    /**
     * Returns dirty time of elemented pointed by @c it.
     */
    const ReplicaTime & get_element_dirty_time
        (
        StorageConstIterator it
        ) const
        {
        return it->get_dirty_time();
        }
    
    /**
     * Serializes array element to bitstream.
     */
    void bitstr_write
        (
        SizeType            index,
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        if( storage[ index ].should_serialize( desc ) == Property::SSR_YES )
            {
            Serializer::BitStrBool::write( bos, true );
            storage[ index ].bitstr_write( bos, desc );
            }
        else
            {
            Serializer::BitStrBool::write( bos, false );
            }
        }

    /**
     * Reads array element from bitstream.
     */
    void bitstr_read
        (
        SizeType            index,
        BitIStream &        bis,
        const Description & desc
        )
        {
        bool should_read;
        
        Serializer::BitStrBool::read( bis, should_read );

        if( should_read )
            {
            storage[ index ].bitstr_read( bis, desc );
            }
        }

    /**
     * Serializes array element to text stream.
     *
     * Note that even if this element belongs to dirty page (otherwise this
     * method would not be called by #internal_text_write()), we may still
     * decide not to write the element. Thus the format of element update
     * will be either "element true <serialized-element>" or "element false".
     */
    void text_write
        (
        SizeType            index,
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        if( storage[ index ].should_serialize( desc ) == Property::SSR_YES )
            {
            Serializer::TextBool::write( tw, true );
            tw.write_space();
            storage[ index ].text_write( tw, desc );
            }
        else
            {
            Serializer::TextBool::write( tw, false );
            }
        }

    /**
     * Reads array element from text stream.
     */
    void text_read
        (
        SizeType            index,
        TextReader &        tr,
        const Description & desc
        )
        {
        bool should_read;

        Serializer::TextBool::read( tr, should_read );

        if( should_read )
            {
            storage[ index ].text_read( tr, desc );
            }
        }

public:

    //
    // ArrayInfo implementation.
    //
    
    /** Returns array size. */
    virtual std::size_t array_get_size() const
        {
        return size();
        }

    /** Clears array. */
    virtual void array_clear()
        {
        clear();
        }
    
    /** Resizes array. */
    virtual void array_resize
        (
        std::size_t new_size
        )
        {
        resize( new_size );
        }

    /** Returns element stored at @c index. */
    virtual Property * array_get_element
        (
        std::size_t index
        )
        {
        return &write_at( index );
        }

private:

    Attributes  stored_attributes;
        /**< Initialization attributes that are propagated to new elements
             when array resizes. */

    }; // PArray

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROP_ARRAY_H
