/** @file prop_dictionary.h
 * Dictionary property.
 */

/*
 * begin        : Sat Apr 5 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_DICTIONARY_H
#define MASSIV_PROP_DICTIONARY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_PROP_ARRAY_H
#include "property/prop_array.h"
#endif

#ifndef MASSIV_PROP_PAIR_H
#include "property/prop_pair.h"
#endif

#ifndef MASSIV_DEBUG_STORAGE_H
#include "utility/debug_storage.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h"
#endif

#ifndef MASSIV_TEXT_WRITER_H
#include "utility/text_writer.h"    // FIXME: nejak to zrusit
#endif

#ifndef MASSIV_TEXT_READER_H
#include "utility/text_reader.h"    // FIXME: nejak to zrusit
#endif

#ifndef MASSIV_LIB_EXCEPTIONS_HELPER_H
#include "object/lib/lib_exceptions_helper.h"
#endif

#include <utility>


#ifdef COMPILER_ICC
// Set level of warning 654: partial override of overloaded function to 4.
// Unfortunately this is useless, because the warning is generated when
// the template is instantiated, not when it's defined. FIXME.
#pragma warning( 4 654 )
#endif


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/**
 * Abstract interface implemented by a dictionary property.
 * Accesses dictionary elements and provides information about a dictionary.
 * Defined mainly for debugging purposes.
 */
class DictionaryInfo
    {
public:

    /** Does nothing. Defined just to get rid of GCC warnings. */
    virtual ~DictionaryInfo();

    /** Returns dictionary size. */
    virtual std::size_t dictionary_get_size() const = 0;

    /** Clears dictionary. */
    virtual void dictionary_clear() = 0;
    
    /** Returns property associated with @c key. */
    virtual Property * dictionary_look_up
        (
        const Property * key
        ) = 0;

    /**
     * Returns property associated with key that was serialized into
     * @c tr.
     */
    virtual Property * dictionary_look_up
        (
        TextReader & tr
        ) = 0;

    }; // DictionaryInfo

/**
 * Abstract interface implemented by a set property.
 * Accesses set elements and provides information about a set.
 * Defined mainly for debugging purposes.
 */
class SetInfo
    {
public:

    /** Does nothing. Defined just to get rid of GCC warnings. */
    virtual ~SetInfo();
        
    /** Returns set size. */
    virtual std::size_t set_get_size() const = 0;

    /** Clears set. */
    virtual void set_clear() = 0;
    
    }; // SetInfo

/**
 * Default hash key functor for a property type.
 */
template <class PropertyType>
class DefaultPropertyHashKey
    {
public:

    /** Computes hash code for @c property. */
    std::size_t operator()
        (
        const PropertyType & property
        ) const
        {
        return property.hash();
        }
    }; // DefaultPropertyHashKey

/**
 * Default equal key functor for a property type.
 */
template <class PropertyType>
class DefaultPropertyEqualKey
    {
public:

    /** Compares properties. */
    bool operator()
        (
        const PropertyType & property1,
        const PropertyType & property2
        ) const
        {
        return property1 == property2;
        }
    }; // DefaultPropertyEqualKey

/**
 * Traits parameter for a dictionary property.
 * Used by @c PHashTableBase.
 */
template
    <
    class Key,
    class Data,
    class HashKey,
    class EqualKey
    >
class DictionaryTraits
    {
protected:

    /// Entries of the hash table.
    typedef PPair<Key, Data>    Entry;
    /// Entry key.
    typedef Key                 KeyType;
    /// Associated data.
    typedef Data                DataProperty;

    /** Type returned by hash table iterator. */
    struct IteratorValue
        {
    public:

        /** Constructs iterator value from an @c Entry. */
        IteratorValue
            (
            Entry & entry
            ) : first( entry.first ),
                second( entry.second )
            {
            }

        /**
         * Allows iterators to return the same object by both the operator*()
         * and operator-\>().
         */
        IteratorValue * operator->()
            {
            return this;
            }

    public:
            
        const KeyType &     first;
            /**< Key. */
        DataProperty &      second;
            /**< Data. */
        };

protected:

    /** Extracts key-part from the stored entry. */
    static const KeyType & get_key
        (
        const Entry & entry
        )
        {
        return entry.first;
        }

protected:

    HashKey     hash_key;
        /**< Hash key functor. Operates on @c KeyType. */
    EqualKey    equal_key;
        /**< Equal key functor. Operates on @c KeyType. */
    }; // DictionaryTraits

/**
 * Traits parameter for a set property.
 * Used by @c PHashTableBase.
 */
template
    <
    class Key,
    class HashKey,
    class EqualKey
    >
class SetTraits
    {
protected:

    /// Entries of the hash table.
    typedef Key             Entry;
    /// Entry key.
    typedef Key             KeyType;

    /// Type returned by hash table iterator. */
    typedef const Entry &   IteratorValue;

protected:

    /** Extracts key-part from the stored entry. */
    static const KeyType & get_key
        (
        const Entry & entry
        )
        {
        return entry;
        }

protected:

    HashKey     hash_key;
        /**< Hash key functor. */
    EqualKey    equal_key;
        /**< Equal key functor. */
    }; // SetTraits

/**
 * Base class for a simple dictionary/set property with a very limited
 * set of supported features.
 *
 * The property implements a hash table holding @c Traits::Entry
 * objects. Each entry contains @c Traits::KeyProperty component that serves
 * as @c Entry key. @c Traits::KeyProperty is extracted from @c Traits::Entry
 * by @c Traits::get_key() method. @c Traits::hash_key() and
 * @c Traits::equal_key() are used to do key-specific operations:
 * compute hash code and compare two keys for equality.
 * @c Traits::IteratorValue defines an object type that is returned by
 * a hash table iterator when it is dereferenced.
 *
 * Unlike a full-featured STL-like hash (multi) map/set the following
 * restrictions apply:
 * 1) Each collision chain is represented by an array, which means
 * that the insert()/erase() operations may relocate stored elements
 * (they can be moved in the array). Iterators are invalidated by both
 * insert & erase operations.
 * 2) Entry must be a property.
 * These restrictions allow to implement the dictionary as a single property
 * ("primitive type"), that might work more effeciently compared to
 * general hash map/set.
 * Full-featured hash (multi) map/set can obviously be implemented as a
 * managed data structure consisting of Objects linked by ObjectPointers.
 */
template <class Traits>
class PHashTableBase : 
    public Property,
    public Traits,
    public DebugStorage
    {
public:

    typedef std::size_t SizeType;

    enum Parameters
        {
        MAX_PRINT_ITEMS =       8,
            /**< Maximum number of items printed by @c to_string(). */
        DEFAULT_TABLE_SIZE =    57
            /**< Default number of buckets. */
        };

protected:

    typedef typename Traits::Entry          Entry;
    typedef typename Traits::KeyType        KeyType;
    typedef typename Traits::IteratorValue  IteratorValue;

    typedef PArray<Entry>                   CollisionChain;
    typedef PArray<CollisionChain>          Buckets;
    typedef Serializer::Description         Description;
    typedef DebugStorage::DebugIterator     DebugIterator;
    
    struct BeginTag
        {
        };

    struct EndTag
        {
        };

public:

    /**
     * Simple iterator used to enumerate elements stored in the table.
     * @c IteratorValue determines type returned by operator-\> or
     * operator*().
     * Since there may be more iterators, each returning different object
     * types (for example only part containing entry's key can be returned),
     * @c IteratorValue is defined.
     * @c IteratorValue must provide constructor that takes "const Entry &"
     * and constructs value object that represents the entry, and properly
     * define @c operator-\>() (so that @c IteratorValue could be returned
     * by value from @c Iterator::operator-\>()).
     */
    class Iterator : public DebugIterator
        {
    public:

        friend class PHashTableBase<Traits>;

        /** Default constructor. */
        Iterator() :
            table( NULL ),
            chain_index( 0 ),
            position( 0 )
            {
            register_iterator( table );
            }

        /** Initializes iterator to point to the beginning of the sequence. */
        Iterator
            (
            PHashTableBase *    table,
            const BeginTag &
            ) : table( table ),
                chain_index( 0 ),
                position( 0 )
            {
            register_iterator( table );

            Buckets & buckets = table->buckets;

            while( chain_index < buckets.size() && buckets[ chain_index ].empty() )
                {
                chain_index++;
                }

            // @c chain_index == buckets.size() && position == 0 represents end().
            }
        
        /** Initializes iterator to point to the end of the sequence. */
        Iterator
            (
            PHashTableBase *    table,
            const EndTag &
            ) : table( table ),
                chain_index( table->buckets.size() ),
                position( 0 )
            {
            register_iterator( table );
            }
        
        /** Initializes iterator to point to a specific entry. */
        Iterator
            (
            PHashTableBase *    table,
            SizeType            chain_index,
            SizeType            position
            ) : table( table ),
                chain_index( chain_index ),
                position( position )
            {
            register_iterator( table );
            }

        /** Copy constructor. */
        Iterator
            (
            const Iterator & it
            ) : table( it.table ),
                chain_index( it.chain_index ),
                position( it.position )
            {
            register_iterator( table );
            }

        /** Destructor. */
        ~Iterator()
            {
            unregister_iterator( table );
            }

        /** Dereference. */
        IteratorValue operator*()
            {
            return IteratorValue( get_entry() );
            }
        
        /** Dereference. */
        IteratorValue operator->()
            {
            return operator*();
            }

        /** Returns key of the referenced entry. */
        const KeyType & get_key() const
            {
            return table->get_key( get_entry() );
            }

        /** Operator=. */
        Iterator & operator=
            (
            const Iterator & it
            )
            {
            unregister_iterator( table );

            it.check_validity();
            table = it.table;
            chain_index = it.chain_index;
            position = it.position;
            DebugIterator::operator=( it );

            register_iterator( table );
            
            return *this;
            }

        /** Moves iterator. Preincrement. */
        Iterator & operator++()
            {
            Buckets & buckets = table->buckets;
                
            if( chain_index >= buckets.size() )
                {
                // We are at end().

                DebugStorage::iterator_position_error();
                }

            if( ++position >= buckets[ chain_index ].size() )
                {
                // Advance to the next collision chain.

                position = 0;
                chain_index++;

                while( chain_index < buckets.size() && buckets[ chain_index ].empty() )
                    {
                    chain_index++;
                    }
                }

            return *this;
            }

        /** Moves iterator. Postincrement. */
        Iterator operator++( int )
            {
            Iterator it = *this;
            operator++();
            return it;
            }

        /** Compares iterators. */
        bool operator==
            (
            const Iterator & it
            ) const
            {
            DebugStorage::incompatible_iterator_error( table == it.table );
            return chain_index == it.chain_index && position == it.position;
            }
        
        /** Compares iterators. */
        bool operator!=
            (
            const Iterator & it
            ) const
            {
            return !operator==( it );
            }

    private:

        /** Returns entry that is referenced by the iterator. */
        Entry & get_entry() const
            {
            check_validity();
            
            if( !table || chain_index >= table->buckets.size() ||
                position >= table->buckets[ chain_index ].size() )
                {
                DebugStorage::iterator_position_error();
                }
                
            return table->buckets[ chain_index ][ position ];
            }

    #ifndef COMPILER_MSVC
    // The above friend seems not to work under MSVC6.
    private:
    #endif
            
        PHashTableBase *    table;
            /**< Hash table. */
        SizeType            chain_index;
            /**< Current chain index. */
        SizeType            position;
            /**< Current position within the chain. */
        }; // Iterator

    friend class Iterator;
    
    typedef std::pair<Iterator, Iterator>   KeyRange;

protected:

    //
    // Constructors.
    //

    /** Default constructor. */
    PHashTableBase()
        {
        // Initialize owned properties.

        buckets.initialize_from_owner( *this, Attributes() );
        
        // Initialize table.
        
        clear();

        // Notify Property that it is default-constructed.
        
        // Note: @c buckets is not "absolutely" default-constructed
        //       (because it is not empty), it is default-constructed
        //       in PHashTableBase context.

        buckets.default_constructed();
        default_constructed();
        }

    /** Copy constructor. */
    PHashTableBase
        (
        const PHashTableBase & table
        ) : Property( table )
        {
        // Initialize owned properties.

        buckets.initialize_from_owner( *this, Attributes() );

        // Initialize table.
        
        if( !table.is_default_constructed() )
            {
            num_items = table.num_items;
            buckets = table.buckets;
            }
        }

    /** Operator=. */
    PHashTableBase & operator=
        (
        const PHashTableBase & table
        )
        {
        num_items = table.num_items;
        buckets = table.buckets;
        
        return *this;
        }

public:

    //
    // Property extensions.
    //
    
    /**
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        )
        {
        Property::initialize_from_owner( direct_owner, attributes );
        buckets.initialize_from_owner( *this, attributes );
        }

    /**
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        )
        {
        buckets.for_each_property_do( operation );
        }
    
    /**
     */
    virtual std::string to_string() const
        {
        std::string s;
        SizeType    chain_index = 0;
        SizeType    count = 0;

        s += "{ ";

        while( chain_index < buckets.size() )
            {
            const CollisionChain &  chain = buckets[ chain_index ];
            SizeType                position = 0;

            while( position < chain.size() )
                {
                if( count )
                    {
                    s += ", ";
                    }

                s += chain[ position ].to_string();

                if( ++count > MAX_PRINT_ITEMS )
                    {
                    s += ", ... }";
                    return s;
                    }

                position++;
                }

            chain_index++;
            }

        s += " }";
        return s;
        }
    
    /**
     */
    virtual PropertyType get_type() const
        {
        return Property::TYPE_CONTAINER;
        }
    
    /**
     */
    virtual bool get_owned_info
        (
        const Property *    owned,
        PropertyInfo &      info
        ) const
        {
        if( &buckets == owned )
            {
            info.property_name = "buckets";
            return true;
            }

        return false;
        }
    
    /**
     */
    virtual void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        num_items.bitstr_write( bos, desc );
        buckets.bitstr_write( bos, desc );
        }

    /**
     */
    virtual void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        // Since we are changing @c num_items, which is not a Property,
        // @c write_requested() must be called here.
        // Properties call @c write_requested() automatically from
        // their @c xxx_read() if their state is modified.

        write_requested();

        num_items.bitstr_read( bis, desc );
        buckets.bitstr_read( bis, desc );
        }

    /**
     */
    virtual void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        tw.line() << "items ";
        num_items.text_write( tw, desc );
        tw.end_line();

        tw.line() << "buckets ";
        tw.change_indent( 4 );
        buckets.text_write( tw, desc );
        tw.change_indent( -4 );
        }

    /**
     */
    virtual void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        // @c write_requested() called for the same reason like in
        // @c bitstr_read().

        write_requested();

        tr.require_string( "items" );
        num_items.text_read( tr, desc );

        tr.next_line_throw();
        tr.require_string( "buckets" );
        buckets.text_read( tr, desc );
        }

public:

    /** Returns table size. */
    SizeType size() const
        {
        read_requested();
        return num_items;
        }

    /** Tests if the table is empty. */
    bool empty() const
        {
        return !size();
        }

    /** Clears the table. */
    void clear
        (
        int table_size = DEFAULT_TABLE_SIZE
        )
        {
        write_requested();
        invalidate_iterators();

        num_items = 0;
        buckets.clear();
        buckets.resize( table_size );
        }

    /** Returns iterator pointing to the beginning of the table. */
    Iterator begin()
        {
        return Iterator( this, BeginTag() );
        }
    
    /** Returns iterator pointing to the end of the table. */
    Iterator end()
        {
        return Iterator( this, EndTag() );
        }
    
    /** Erases element pointed by the iterator. */
    void erase
        (
        const Iterator & it
        )
        {
        // Probe the iterator.

        DebugStorage::incompatible_iterator_error( this == it.table );
        it.check_validity();

        // Erase element.

        buckets[ it.chain_index ].erase( it.position );
        num_items--;

        invalidate_iterators();
        }
    
    /** Erases first entry with @c key. Returns true if the entry was erased. */
    bool erase
        (
        const KeyType & key
        )
        {
        CollisionChain &    chain = get_chain( key );
        SizeType            index = 0;

        invalidate_iterators();

        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                chain.erase( index );
                num_items--;
                return true;
                }

            index++;
            }

        return false;
        }
    
    /**
     * Tries to find an entry containing @c key.
     * Returns NULL if no entry is found.
     */
    const Entry * search
        (
        const KeyType & key
        ) const
        {
        const CollisionChain &  chain = get_chain( key );
        SizeType                index = 0;

        while( index < chain.size() )
            {
            const Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                return &entry;
                }

            index++;
            }

        return NULL;
        }
    
    /**
     * Tries to find an entry containing @c key.
     * Returns NULL if no entry is found.
     */
    Entry * search
        (
        const KeyType & key
        )
        {
        CollisionChain &    chain = get_chain( key );
        SizeType            index = 0;

        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                return &entry;
                }

            index++;
            }

        return NULL;
        }

    /** Returns iterator pointing to an entry with @c key. */
    Iterator find
        (
        const KeyType & key
        )
        {
        SizeType            chain_index = get_chain_index( key );
        CollisionChain &    chain = buckets[ chain_index ];
        SizeType            index = 0;

        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                return Iterator( this, chain_index, index );
                }

            index++;
            }

        return end();
        }

    /** Tests if @c key is stored in the table. */
    bool contains
        (
        const KeyType & key
        ) const
        {
        return search( key ) != NULL;
        }
        
    /** Compares tables. */
    friend bool operator==
        (
        const PHashTableBase & d1,
        const PHashTableBase & d2
        )
        {
        SizeType i;

        if( d1.buckets.size() != d2.buckets.size() )
            {
            return false;
            }

        for( i = 0; i < d1.buckets.size(); i++ )
            {
            const CollisionChain &  chain1 = d1.buckets[ i ];
            const CollisionChain &  chain2 = d2.buckets[ i ];
            SizeType                j, k;

            if( chain1.size() != chain2.size() )
                {
                return false;
                }

            // Compare chain1 and chain2 (treat them as sets).
            // Since the chains can not be ordered they can not be
            // compared by simple chain1 == chain2.

            // Test if chain1 is a subset of chain2.

            for( j = 0; j < chain1.size(); j++ )
                {
                bool found = false;

                for( k = 0; k < chain2.size(); k++ )
                    {
                    if( d1.equal_key( PHashTableBase::get_key( chain1[ j ] ),
                                PHashTableBase::get_key( chain2[ k ] ) ) )
                        {
                        found = true;
                        break;
                        }
                    }

                if( !found )
                    {
                    return false;
                    }
                }
            
            // Test if chain2 is a subset of chain1.

            for( j = 0; j < chain2.size(); j++ )
                {
                bool found = false;

                for( k = 0; k < chain1.size(); k++ )
                    {
                    if( d1.equal_key( PHashTableBase::get_key( chain1[ k ] ),
                                PHashTableBase::get_key( chain2[ j ] ) ) )
                        {
                        found = true;
                        break;
                        }
                    }

                if( !found )
                    {
                    return false;
                    }
                }
            }

        return true;
        }
    
    /** Compares tables. */
    friend bool operator!=
        (
        const PHashTableBase & d1,
        const PHashTableBase & d2
        )
        {
        return !operator==( d1, d2 );
        }

protected:
    
    /** Inserts @c entry into the table. Allows key duplicities. */
    void insert_multi
        (
        const Entry &   element
        )
        {
        CollisionChain &    chain = get_chain( get_key( element ) );
        SizeType            index = 0;
        
        invalidate_iterators();
        
        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), get_key( element ) ) )
                {
                // Insert the new pair to this position.
                // This ensures that all the pairs with this key
                // will be clustered in just one chain range.

                chain.insert( index, element );
                num_items++;

                return;
                }

            index++;
            }

        chain.push_back( element );
        num_items++;
        }

    /** Returns iterator range for elements matching @c key. */
    KeyRange equal_range
        (
        const KeyType & key
        )
        {
        SizeType            chain_index = get_chain_index( key );
        CollisionChain &    chain = buckets[ chain_index ];
        SizeType            position = 0;

        while( position < chain.size() )
            {
            Entry & entry = chain[ position ];

            if( equal_key( get_key( entry ), key ) )
                {
                Iterator start( this, chain_index, position );
                Iterator stop( start );
                Iterator end( this->end() );

                // All the pairs with the same key are clustered
                // in just one chain range.

                while( equal_key( stop.get_key(), key ) )
                    {
                    ++stop;
                    
                    if( stop == end )
                        {
                        break;
                        }
                    }

                return KeyRange( start, stop );
                }

            position++;
            }

        return KeyRange( Iterator(), Iterator() );
        }

    /** Returns index of collision chain that should contain @c key. */
    SizeType get_chain_index
        (
        const KeyType & key
        ) const
        {
        return hash_key( key ) % buckets.size();
        }

    /** Returns collision chain that should contain @c key. */
    CollisionChain & get_chain
        (
        const KeyType & key
        )
        {
        return buckets[ get_chain_index( key ) ];
        }
    
    /** Returns collision chain that should contain @c key. */
    const CollisionChain & get_chain
        (
        const KeyType & key
        ) const
        {
        return buckets[ get_chain_index( key ) ];
        }

protected:

    SInt32      num_items;
        /**< Number of stored items. */
    Buckets     buckets;
        /**< Buckets. */
    }; // PHashTableBase

/**
 * Base class for a dictionary property.
 * Uses PHashTableBase to implement hash table holding <@c KeyProperty,
 * @c DataProperty> pairs.
 * @c KeyProperty and @c DataProperty must be property types.
 * @c HashKey and @c EqualKey functors operate on @c KeyProperty only.
 */
template
    <
    class KeyProperty,
    class DataProperty,
    class HashKey,
    class EqualKey
    >
class PDictionaryBase :
    public PHashTableBase
        <
        DictionaryTraits<KeyProperty, DataProperty, HashKey, EqualKey>
        >,
    public DictionaryInfo
    {
private:
    
    /// Hash table.
    typedef PHashTableBase
        <
        DictionaryTraits<KeyProperty, DataProperty, HashKey, EqualKey>
        >   Table;
        
    /// Types defined in superclasses.
    //@{
public:

    typedef typename Table::SizeType            SizeType;

protected:

    typedef typename Table::Entry               Entry;
    typedef typename Table::CollisionChain      CollisionChain;
    typedef typename Table::Buckets             Buckets;

public:

    typedef typename Table::Iterator            Iterator;
    //@}

public:

    using Table::size;
    using Table::clear;

public:

    //
    // Property extensions.
    //
    
    /**
     */
    virtual DictionaryInfo * get_dictionary_info()
        {
        return this;
        }

public:
        
    //
    // DictionaryInfo implementation.
    //
    
    /**
     */
    virtual std::size_t dictionary_get_size() const
        {
        return size();
        }

    /**
     */
    virtual void dictionary_clear()
        {
        clear();
        }

    /**
     */
    virtual Property * dictionary_look_up
        (
        TextReader & tr
        )
        {
        Serializer::Description desc;
        KeyProperty             key;

        key.text_read( tr, desc );
        return static_cast<DictionaryInfo *>( this )->dictionary_look_up( &key );
        }
    
    }; // PDictionaryBase

/**
 * Simple dictionary property.
 *
 * Associates @c KeyProperties with @c DataProperties.
 * Does not allow to store pairs with duplicate keys into the dictionary.
 *
 * For further information on how the property is implemented and
 * what restrictions apply to @c KeyProperty and @c DataProperty
 * consult @c PDictionaryBase.
 */
template
    <
    class KeyProperty,
    class DataProperty,
    class HashKey,
    class EqualKey
    >
class PDictionary :
    public PDictionaryBase<KeyProperty, DataProperty, HashKey, EqualKey>
    {
private:

    /// Superclass.
    typedef PDictionaryBase<KeyProperty, DataProperty, HashKey, EqualKey> Super;

    /// Types defined in superclass.
    //@{
public:

    typedef typename Super::SizeType            SizeType;

protected:

    typedef typename Super::Entry               Entry;
    typedef typename Super::CollisionChain      CollisionChain;
    typedef typename Super::Buckets             Buckets;

public:

    typedef typename Super::Iterator            Iterator;
    //@}

protected:

    using Super::invalidate_iterators;
    using Super::num_items;

public:
    
    /**
     * Simple indexer class returned by operator[].
     */
    class Indexer
        {
    public:

        /** Constructor. */
        Indexer
            (
            PDictionary &   dictionary,
            const KeyProperty & key
            ) : dictionary( dictionary ),
                key( key )
            {
            }

        /** Performs insert. */
        Indexer & operator=
            (
            const DataProperty & data
            )
            {
            dictionary.insert( key, data );
            return *this;
            }

        /** Performs look up. */
        operator const DataProperty &()
            {
            return dictionary.look_up( key );
            }

    private:
            
        PDictionary &   dictionary;
            /**< Dictionary. */
        KeyProperty     key;
            /**< Key property. */
        }; // Indexer

public:
    
    /** Returns data associated with @c key. */
    const DataProperty & look_up
        (
        const KeyProperty & key
        ) const
        {
        const Entry * entry = search( key );

        if( entry )
            {
            return entry->second;
            }
        else
            {
            RAISE_CORE_LIB_EXCEPTION
                (
                CanNotAccessObjectException, 
                "Key is not associated with any data."
                );
            }
        }
    
    /** Performs @c key look up. */
    const DataProperty & operator[]
        (
        const KeyProperty & key
        ) const
        {
        return look_up( key );
        }

    /** Returns indexer class for @c key. */
    Indexer operator[]
        (
        const KeyProperty & key
        )
        {
        return Indexer( *this, key );
        }
    
    /** Associates @c data with @c key. */
    void insert
        (
        const KeyProperty &     key,
        const DataProperty &    data
        )
        {
        CollisionChain &    chain = get_chain( key );
        SizeType            index = 0;
        
        invalidate_iterators();
        
        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                entry.second = data;
                return;
                }

            index++;
            }

        chain.push_back( Entry( key, data ) );
        num_items++;
        }

public:
    
    //
    // DictionaryInfo implementation.
    //
    
    /**
     */
    virtual Property * dictionary_look_up
        (
        const Property * key
        )
        {
        Entry * entry = search( checked_cast<const KeyProperty &>( *key ) );
        return entry ? &entry->second : NULL;
        }

    }; // PDictionary


/**
 * Simple multi-dictionary property.
 *
 * Associates @c KeyProperties with @c DataProperties.
 * Allows to store pairs with duplicate keys into the dictionary.
 * Each set of pairs with the same key form just one "cluster" (range
 * of the collision chain).
 */
template
    <
    class KeyProperty,
    class DataProperty,
    class HashKey,
    class EqualKey
    >
class PMultiDictionary :
    public PDictionaryBase<KeyProperty, DataProperty, HashKey, EqualKey>
    {
private:

    /// Superclass.
    typedef PDictionaryBase<KeyProperty, DataProperty, HashKey, EqualKey> Super;

    /// Types defined in superclass.
    //@{
public:

    typedef typename Super::SizeType    SizeType;
    typedef typename Super::Iterator    Iterator;
    typedef typename Super::KeyRange    KeyRange;

protected:

    typedef typename Super::Entry       Entry;
    //@}

public:
    
    /** Inserts [ @c key, @c data ] pair into the dictionary. */
    void insert
        (
        const KeyProperty &     key,
        const DataProperty &    data
        )
        {
        Super::insert_multi( Entry( key, data ) );
        }

    /** Returns dictionary range for elements matching @c key. */
    KeyRange equal_range
        (
        const KeyProperty & key
        )
        {
        return Super::equal_range( key );
        }

public:
            
    //
    // DictionaryInfo implementation.
    //
    
    /** Not implemented by PMultiDictionary. */
    virtual Property * dictionary_look_up
        (
        const Property * key
        )
        {
        return NULL;
        }

    }; // PMultiDictionary

/**
 * Base class for a set property.
 *
 * Uses PHashTableBase to implement hash table holding @c KeyProperty
 * elements.
 */
template
    <
    class KeyProperty,
    class HashKey,
    class EqualKey
    >
class PSetBase :
    public PHashTableBase< SetTraits<KeyProperty, HashKey, EqualKey> >,
    public SetInfo
    {
private:
    
    /// Hash table.
    typedef PHashTableBase< SetTraits<KeyProperty, HashKey, EqualKey> > Table;
        
    /// Types defined in superclasses.
    //@{
public:

    typedef typename Table::SizeType            SizeType;

protected:

    typedef typename Table::Entry               Entry;
    typedef typename Table::CollisionChain      CollisionChain;
    typedef typename Table::Buckets             Buckets;

public:

    typedef typename Table::Iterator            Iterator;
    //@}

public:

    using Table::size;
    using Table::clear;

public:

    //
    // Property extensions.
    //
    
    /**
     */
    virtual SetInfo * get_set_info()
        {
        return this;
        }

public:
        
    //
    // SetInfo implementation.
    //
    
    /**
     */
    virtual std::size_t set_get_size() const
        {
        return size();
        }

    /**
     */
    virtual void set_clear()
        {
        clear();
        }

    }; // PSetBase

/**
 * Simple set property.
 *
 * Stores @c KeyProperties, does not allow duplicities.
 *
 * For further information on how the property is implemented and
 * what restrictions apply to @c KeyProperty consult @c PSetBase.
 */
template
    <
    class KeyProperty,
    class HashKey,
    class EqualKey
    >
class PSet : public PSetBase<KeyProperty, HashKey, EqualKey>
    {
private:

    /// Superclass.
    typedef PSetBase<KeyProperty, HashKey, EqualKey> Super;

    /// Types defined in superclass.
    //@{
public:

    typedef typename Super::SizeType            SizeType;
    typedef typename Super::Iterator            Iterator;

protected:

    typedef typename Super::Entry               Entry;
    typedef typename Super::CollisionChain      CollisionChain;
    typedef typename Super::Buckets             Buckets;
    //@}

protected:

    using Super::invalidate_iterators;
    using Super::num_items;

public:
    
    /** Inserts @c key into the set. */
    void insert
        (
        const KeyProperty & key
        )
        {
        CollisionChain &    chain = get_chain( key );
        SizeType            index = 0;
        
        invalidate_iterators();
        
        while( index < chain.size() )
            {
            Entry & entry = chain[ index ];

            if( equal_key( get_key( entry ), key ) )
                {
                return;
                }

            index++;
            }

        chain.push_back( Entry( key ) );
        num_items++;
        }

    }; // PSet

/**
 * Simple multi-set property.
 *
 * Stores @c KeyProperties, allows duplicities.
 * Elements with the same key are clustered to a single key range.
 */
template
    <
    class KeyProperty,
    class HashKey,
    class EqualKey
    >
class PMultiSet : public PSetBase<KeyProperty, HashKey, EqualKey>
    {
private:

    /// Superclass.
    typedef PSetBase<KeyProperty, HashKey, EqualKey> Super;

public:

    /// Types defined in superclass.
    //@{
    typedef typename Super::SizeType    SizeType;
    typedef typename Super::Iterator    Iterator;
    typedef typename Super::KeyRange    KeyRange;
    //@}

public:
    
    /** Inserts @c key into the multi-set. */
    void insert
        (
        const KeyProperty & key
        )
        {
        Super::insert_multi( key );
        }

    /** Returns iterator range for elements matching @c key. */
    KeyRange equal_range
        (
        const KeyProperty & key
        )
        {
        return Super::equal_range( key );
        }

    }; // PMultiSet

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#ifdef COMPILER_ICC
// Set level of warning 654: partial override of overloaded function back to 1.
#pragma warning( 1 654 )
#endif


#endif // MASSIV_PROP_DICTIONARY_H
