/** @file prop_enum.h
 * Enumeration property.
 */

/*
 * begin        : Mon Apr 7 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_ENUM_H
#define MASSIV_PROP_ENUM_H


#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_SERIALIZER_VLI_H
#include "serializer/serializer_vli.h"
#endif

#include <sstream>


namespace Massiv {
namespace Core {
        

/** @addtogroup property
 * @{
 */

/**
 * Enumeration property.
 * @c EnumType is expected to a be an enumeration type.
 * @note @c EnumType values should have defined their mapping to integers
 * explicitly so that they could be safely serialized.
 */
template <class EnumType>
class PEnum : public ManagedAttribute<EnumType, Property>
    {
private:

    typedef ManagedAttribute<EnumType, Property>    Super;
    typedef PEnum<EnumType>                         ThisProperty;
    typedef Serializer::Description                 Description;

public:
    
    using Super::read;
    using Super::write;

public:
        
    /** Default constructor. */
    PEnum()
        {
        }

    /** Constructor. */
    PEnum
        (
        EnumType value
        ) : Super( value )
        {
        }

    /** Copy constructor. */
    PEnum
        (
        const ThisProperty & value
        ) : Super( value )
        {
        }

    /** Operator=. */
    ThisProperty & operator=
        (
        EnumType value
        )
        {
        Super::operator=( value );
        return *this;
        }

    /** Operator=. */
    ThisProperty & operator=
        (
        const ThisProperty & value
        )
        {
        Super::operator=( value.read() );  
        return *this;
        }

    #ifdef COMPILER_ICC

    /*
     * ICC-related hack.
     *
     * ICC/Linux 7.0.075 won't compile the @c if condition in the following code:
     * @code
     * enum E { VAL = 0 };
     * PEnum< E > e;
     * if( e == VAL ) { }
     * @endcode
     *
     * It says:
     *   more than one operator "==" matches these operands:
     *   built-in operator "enum == enum"
     *   built-in operator "enum == enum"
     * (yes, the two lines above are exactly the same). I really don't understand
     * that error.
     */
    /** Compare with @c EnumType. */
    bool operator==
        (
        EnumType other
        ) const
        {
        return read() == other;
        }

    /** Compare with @c EnumType. */
    bool operator!=
        (
        EnumType other
        ) const
        {
        return read() != other;
        }

    #endif // COMPILER_ICC

public:
    
    /**
     */
    virtual std::string to_string() const
        {
        std::ostringstream s;

        s << read();
        return s.str();
        }

    /**
     */
    virtual int hash() const
        {
        return read();
        }
    
    /**
     */
    virtual int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     */
    virtual int bitstream_size() const
        {
        return 8;
        }

    /**
     */
    virtual void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        Serializer::BitStrVli::write( bos, read() );
        }

    /**
     */
    virtual void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        int temp;
        Serializer::BitStrVli::read( bis, temp );
        write( static_cast<EnumType>( temp ) );
        }

    /**
     */
    virtual void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextVli::write( tw, read() );
        }

    /**
     */
    virtual void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        int temp;
        Serializer::TextVli::read( tr, temp );
        write( static_cast<EnumType>( temp ) );
        }

    }; // PEnum

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROP_ENUM_H
