/** @file prop_float.h
 * Floating-point property.
 */

/*
 * begin        : Fri Feb 15 2001
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_FLOAT_H
#define MASSIV_PROP_FLOAT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_SERIALIZER_FLOAT_H
#include "serializer/serializer_float.h"
#endif

#ifndef MASSIV_SERIALIZER_H
#include "serializer/serializer.h"
#endif

#include <sstream>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/** @class PFloatQ
 * Floating-point property.
 *
 * Quantizes specified range to integer (1 to 32 bits) during lossy bitstream
 * serialization. See SFloatQ for more information about the quantization and
 * @a Traits structure.
 */
template< class Type, class Traits >
class PFloatQ : public NumericAttribute< Type, Property >
    {
private:

    /// Superclass.
    typedef NumericAttribute< Type, Property >      Super;

    /// This class.
    typedef PFloatQ< Type, Traits >                 ThisProperty;

    /// Hack to prevent msvc6 internal errors.
    typedef Serializer::Description                 Description;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE PFloatQ()
        {
        }

    /** Constructor. */
    INLINE PFloatQ
        (
        Type value
        ) : Super( value )
        {
        }

    /** Copy constructor. */
    INLINE PFloatQ
        (
        const ThisProperty & property
        ) : Super( property )
        {
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        Type value
        )
        {
        Super::operator=( value );
        return *this;
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        const ThisProperty & property
        )
        {
        Super::operator=( property.read() );
        return *this;
        }

public:

    /**
     */
    virtual std::string to_string() const
        {
        std::ostringstream s;

        s << read();
        return s.str();
        }
    
    /**
     */
    virtual int hash() const
        {
        return static_cast<int>( read() );
        }
    
    /**
     */
    virtual int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     */
    virtual int bitstream_size() const
        {
        return Traits::bits;
        }

public:

    /**
     * Serialize to bitstream.
     */
    virtual INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        Serializer::BitStrFloat::write
            (
            bos,
            desc,
            read(),
            Traits::min_value,
            Traits::max_value,
            Traits::bits
            );
        }

    /**
     * Read from bitstream.
     */
    virtual INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        Type temp;
        Serializer::BitStrFloat::read
            (
            bis,
            desc,
            temp,
            Traits::min_value,
            Traits::max_value,
            Traits::bits
            );
        write( temp );
        }

    /**
     * Serialize to text stream.
     */
    virtual INLINE void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextFloat::write( tw, read() );
        }

    /**
     * Read from text stream.
     */
    virtual INLINE void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        Type temp;
        Serializer::TextFloat::read( tr, temp );
        write( temp );
        }

    }; // class PFloatQ


/** @class PFloat
 * Floating-point property.
 * @a Type should be Float32 or Float64.
 * No transformations are applied during serialization.
 */
template <class Type>
class PFloat : public NumericAttribute<Type, Property>
    {
private:

    typedef NumericAttribute<Type, Property>    Super;
    typedef PFloat<Type>                        ThisProperty;
    typedef Serializer::Description             Description;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE PFloat()
        {
        }

    /** Constructor. */
    INLINE PFloat
        (
        Type value
        ) : Super( value )
        {
        }

    /** Copy constructor. */
    INLINE PFloat
        (
        const ThisProperty & property
        ) : Super( property )
        {
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        Type value
        )
        {
        Super::operator=( value );
        return *this;
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        const ThisProperty & property
        )
        {
        Super::operator=( property.read() );
        return *this;
        }

public:

    /**
     */
    virtual std::string to_string() const
        {
        std::ostringstream s;

        s << read();
        return s.str();
        }

    /**
     */
    virtual int hash() const
        {
        return static_cast<int>( read() );
        }
    
    /**
     */
    virtual int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     */
    virtual int bitstream_size() const
        {
        return sizeof( Type ) * 8;
        }

public:

    /**
     * Serialize to bitstream.
     */
    virtual INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        Serializer::BitStrFloat::write( bos, read() );
        }

    /**
     * Read from bitstream.
     */
    virtual INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        Type temp;
        Serializer::BitStrFloat::read( bis, temp );
        write( temp );
        }

    /**
     * Serialize to text stream.
     */
    virtual INLINE void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        Serializer::TextFloat::write( tw, read() );
        }

    /**
     * Read from text stream.
     */
    virtual INLINE void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        Type temp;
        Serializer::TextFloat::read( tr, temp );
        write( temp );
        }

    }; // class PFloat

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROP_FLOAT_H
