/** @file prop_int.h
 * Integral property.
 */

/*
 * begin        : Tue Nov 16 2001
 * copyright    : (C) 2001 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_INT_H
#define MASSIV_PROP_INT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_SERIALIZER_INT_H
#include "serializer/serializer_int.h"
#endif

#include <sstream>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/** @class PInt
 * Integral property.
 * @a Type should be one of Int32, UInt32, Int64, UInt64.
 * @a bits should >= 1 and <= 32 (Int32, UInt32) or <= 64 (Int64, UInt64).
 * Always truncated to specified number of bits (even when migrating/archiving).
 */
template <class Type, int bits>
class PInt : public IntegralAttribute<Type, Property>
    {
private:

    typedef IntegralAttribute<Type, Property>   Super;
    typedef PInt<Type, bits>                    ThisProperty;
    typedef Serializer::Description             Description;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE PInt()
        {
        }

    /** Constructor. */
    INLINE PInt
        (
        Type v
        ) : Super( v )
        {
        }

    /** Copy constructor. */
    INLINE PInt
        (
        const ThisProperty & property
        ) : Super( property )
        {
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        Type value
        )
        {
        Super::operator=( value );
        return *this;
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        const ThisProperty & property
        )
        {
        Super::operator=( property.read() );
        return *this;
        }

public:

    /**
     */
    virtual std::string to_string() const
        {
        std::ostringstream s;

        s << read();
        return s.str();
        }

    /**
     */
    virtual int hash() const
        {
        return static_cast<int>( read() );
        }
    
    /**
     */
    virtual int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     */
    virtual int bitstream_size() const
        {
        return bits;
        }

public:

    /**
     * Serialize to bitstream.
     */
    virtual INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        // Note: Truncating even if !lossy
        Serializer::BitStrInt::write( bos, read(), bits );
        }

    /**
     * Read from bitstream.
     */
    virtual INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        Type temp;
        Serializer::BitStrInt::read( bis, temp, bits );
        write( temp );
        }

    /**
     * Serialize to text stream.
     */
    virtual INLINE void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextInt::write( tw, read(), bits );
        }

    /**
     * Read from text stream.
     */
    virtual INLINE void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        Type temp;
        Serializer::TextInt::read( tr, temp, bits );
        write( temp );
        }

    }; // PInt

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROP_INT_H
