/** @file prop_pair.h
 * Pair property.
 */

/*
 * begin        : Mon Sep 15 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_PAIR_H
#define MASSIV_PROP_PAIR_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_TEXT_READER_H
#include "utility/text_reader.h" // FIXME: nejak to zrusit
#endif

#ifndef MASSIV_TEXT_WRITER_H
#include "utility/text_writer.h" // FIXME: nejak to zrusit
#endif

#ifndef MASSIV_SERIALIZER_BOOL_H
#include "serializer/serializer_bool.h"
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/**
 * Property pair, defined in a STL-like fashion.
 *
 * Both pair components, @c FirstProperty and @c SecondProperty, must
 * be property types.
 */
template <class FirstProperty, class SecondProperty>
class PPair : public Property
    {
private:

    typedef Serializer::Description Description;

public:

    //
    // Constructors.
    //

    /** Default constructor. */
    PPair()
        {
        if( first.is_default_constructed() && second.is_default_constructed() )
            {
            default_constructed();
            }
        
        first.initialize_from_owner( *this, Attributes() );
        second.initialize_from_owner( *this, Attributes() );
        }

    /** Constructor. */
    PPair
        (
        const FirstProperty &   the_first,
        const SecondProperty &  the_second
        ) : first( the_first ),
            second( the_second )
        {
        first.initialize_from_owner( *this, Attributes() );
        second.initialize_from_owner( *this, Attributes() );
        }

    /** Copy constructor. */
    PPair
        (
        const PPair & pair
        ) : Property( pair ),
            first( pair.first ),
            second( pair.second )
        {
        first.initialize_from_owner( *this, Attributes() );
        second.initialize_from_owner( *this, Attributes() );
        }

    /** Operator=. */
    PPair & operator=
        (
        const PPair & pair
        )
        {
        first = pair.first;
        second = pair.second;

        return *this;
        }

public:

    //
    // Property extensions.
    //
    
    /**
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        )
        {
        Property::initialize_from_owner( direct_owner, attributes );
        
        first.initialize_from_owner( *this, attributes );
        second.initialize_from_owner( *this, attributes );
        }

    /**
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        )
        {
        first.for_each_property_do( operation );
        second.for_each_property_do( operation );
        }
    
    /**
     */
    virtual bool get_owned_info
        (
        const Property *    owned,
        PropertyInfo &      info
        ) const
        {
        if( &first == owned )
            {
            info.property_name = "first";
            return true;
            }
        
        if( &second == owned )
            {
            info.property_name = "second";
            return true;
            }

        return false;
        }
    
    /**
     */
    virtual std::string to_string() const
        {
        std::string s;

        s += first.to_string();
        s += " : ";
        s += second.to_string();

        return s;
        }
    
    /**
     */
    virtual void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        // Write first component.

        if( first.should_serialize( desc ) == Property::SSR_YES )
            {
            Serializer::BitStrBool::write( bos, true );
            first.bitstr_write( bos, desc );
            }
        else
            {
            Serializer::BitStrBool::write( bos, false );
            }
        
        // Write second component.

        if( second.should_serialize( desc ) == Property::SSR_YES )
            {
            Serializer::BitStrBool::write( bos, true );
            second.bitstr_write( bos, desc );
            }
        else
            {
            Serializer::BitStrBool::write( bos, false );
            }
        }

    /**
     */
    virtual void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        bool should_read;

        // Read first component.

        Serializer::BitStrBool::read( bis, should_read );
        
        if( should_read )
            {
            first.bitstr_read( bis, desc );
            }

        // Read second component.

        Serializer::BitStrBool::read( bis, should_read );
        
        if( should_read )
            {
            second.bitstr_read( bis, desc );
            }
        }

    /**
     */
    virtual void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        // Write first component.

        if( first.should_serialize( desc ) == Property::SSR_YES )
            {
            tw.line() << "first ";
            first.text_write( tw, desc );
            }
        else
            {
            tw.line() << "skip_first";
            }
        tw.end_line();
        
        // Write second component.

        if( second.should_serialize( desc ) == Property::SSR_YES )
            {
            tw.line() << "second ";
            second.text_write( tw, desc );
            }
        else
            {
            tw.line() << "skip_second";
            }
        }

    /**
     */
    virtual void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        SString token;

        // Read first component.

        token.text_read( tr, desc );
        if( token == "first" )
            {
            first.text_read( tr, desc );
            }
        else if( token != "skip_first" )
            {
            throw ExUnexpectedToken( tr.unexpected_token( "'first' or 'skip_first'" ) );
            }

        // Read second component.

        tr.next_line_throw();
        token.text_read( tr, desc );
        if( token == "second" )
            {
            second.text_read( tr, desc );
            }
        else if( token != "skip_second" )
            {
            throw ExUnexpectedToken( tr.unexpected_token( "'second' or 'skip_second'" ) );
            }
        }

public:
            
    FirstProperty   first;
        /**< First component. */
    SecondProperty  second;
        /**< Second component. */

    }; // class PPair

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_PROP_PAIR_H
