/** @file prop_time.h
 * Time property.
 */

/*
 * begin        : Fri Mar 22 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_PROP_TIME_H
#define MASSIV_PROP_TIME_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/** @class PVirtualTime
 * Virtual time property.
 * @a VTime should be a serializable class derived from @c VirtualTime<>
 * (ie. @c STime).
 *
 * @ingroup time
 */
template <class VTime>
class MASSIV_EXPORTED PVirtualTime : public ManagedAttribute<VTime, Property>
    {
private:

    typedef ManagedAttribute<VTime, Property>   Super;
    typedef PVirtualTime<VTime>                 ThisProperty;
    typedef Serializer::Description             Description;

public:
    
    using Super::read;
    using Super::write;

public:

    /**
     * Default constructor.
     */
    INLINE PVirtualTime();

    /**
     * Constructor.
     */
    INLINE PVirtualTime
        (
        const VTime & time
        );

    /**
     * Copy constructor.
     */
    INLINE PVirtualTime
        (
        const ThisProperty & property
        );

    /**
     * Operator=.
     */
    INLINE PVirtualTime & operator=
        (
        const VTime & time
        );

    /**
     * Operator=.
     */
    INLINE PVirtualTime & operator=
        (
        const ThisProperty & property
        );

    /**
     * Operator+=.
     */
    INLINE PVirtualTime & operator+=
        (
        const VTime & time
        );

    /**
     * Operator-=.
     */
    INLINE PVirtualTime & operator-=
        (
        const VTime & time
        );

    /**
     * Advances virtual time by unit time.
     */
    INLINE void advance();

    /**
     * Resets time to invalid value.
     */
    INLINE void invalidate();

    /**
     * Compares two times.
     */
    INLINE bool is_equal_to
        (
        const VTime & time
        ) const;

    /**
     * Checks whether this time is older than @a time.
     */
    INLINE bool is_older_than
        (
        const VTime & time
        ) const;
    
public:

    /**
     */
    virtual std::string to_string() const;

    /**
     */
    virtual int hash() const;
    
    /**
     */
    virtual int memory_size() const;

    /**
     */
    virtual int bitstream_size() const;

public:

    /**
     * Serialize to bitstream.
     */
    virtual INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        );

    /**
     * Serialize to text stream.
     */
    virtual INLINE void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const;

    /**
     * Read from text stream.
     */
    virtual INLINE void text_read
        (
        TextReader &        tr,
        const Description & desc
        );

    }; // PVirtualTime

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "property/prop_time_inline.h"
#endif


#endif // MASSIV_PROP_TIME_H
