/** @file prop_vli.h
 * Integral property.
 */

/*
 * begin        : Mon Mar 3 2003
 * copyright    : (C) 2001 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROP_VLI_H
#define MASSIV_PROP_VLI_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_SERIALIZER_VLI_H
#include "serializer/serializer_vli.h"
#endif

#include <sstream>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/** @class PVli
 * Variable length integral property.
 * @a Type should be one of Int32, UInt32, Int64, UInt64.
 * Never truncated, always full size.
 */
template <class Type>
class PVli : public IntegralAttribute<Type, Property>
    {
private:

    typedef IntegralAttribute<Type, Property>   Super;
    typedef PVli<Type>                          ThisProperty;
    typedef Serializer::Description             Description;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE PVli()
        {
        }

    /** Constructor. */
    INLINE PVli
        (
        Type v
        ) : Super( v )
        {
        }

    /** Copy constructor. */
    INLINE PVli
        (
        const ThisProperty & property
        ) : Super( property )
        {
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        Type value
        )
        {
        Super::operator=( value );
        return *this;
        }

    /** Operator=. */
    INLINE ThisProperty & operator=
        (
        const ThisProperty & property
        )
        {
        Super::operator=( property.read() );
        return *this;
        }

public:

    /**
     */
    virtual std::string to_string() const
        {
        std::ostringstream s;

        s << read();
        return s.str();
        }

    /**
     */
    virtual int hash() const
        {
        return static_cast<int>( read() );
        }
    
    /**
     */
    virtual int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     */
    virtual int bitstream_size() const
        {
        return sizeof( Type ) * 8;
        }

public:

    /**
     * Serialize to bitstream.
     */
    virtual INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        Serializer::BitStrVli::write( bos, read() );
        }

    /**
     * Read from bitstream.
     */
    virtual INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        Type temp;
        Serializer::BitStrVli::read( bis, temp );
        write( temp );
        }

    /**
     * Serialize to text stream.
     */
    virtual INLINE void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextVli::write( tw, read() );
        }

    /**
     * Read from text stream.
     */
    virtual INLINE void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        Type temp;
        Serializer::TextVli::read( tr, temp );
        write( temp );
        }

    }; // PVli

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROP_VLI_H
