/** @file property.cpp
 * Property implementation.
 */

/*
 * begin        : Tue Nov 30 2001
 * copyright    : (C) 2001, 2002, 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "global/global.h"
#include "utility/debug.h"
#include "property/property.h"
#include "object/object.h"
#include "serializer/serializer.h"
#include "status/exception.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "object/lib/lib_exceptions_helper.h"
#include "object/object_manager.h"
#include "replication/replica_client.h"
#include <sstream>


namespace Massiv {
namespace Core {


Property::~Property()
    {
    }


void Property::initialize_from_owner
    (
    Property &          direct_owner,
    const Attributes &  attributes
    )
    {
    // Initialize ownership.

    owner = &direct_owner;
    hope( owner != this );

    // Initialize attributes from owner.

    root_owner = direct_owner.root_owner;
    replica = direct_owner.replica;
    archive_on_write = direct_owner.archive_on_write;
    
    #ifdef OPTION_DEBUG
    
    if( owner->is_default_constructed() & !is_default_constructed() )
        {
        PropertyInfo    info = get_info( QUERY_FULL_NAME | QUERY_TYPE );
        std::string     message;

        message = "Inconsistent use of default_constructed().\n"
                  "Owner property tells that it is default-constructed, "
                  "but child (= this) property is not. "
                  "The property is not implemented correctly with "
                  "respect to the property rules. Does your copy constructor "
                  "call Property copy constructor?\n" +
                  info.property_name + " (" + info.type_name + ")";

        panic( message.c_str() );
        }

    if( owner->dirty_time.is_older_than( dirty_time ) )
        {
        PropertyInfo    info = get_info( QUERY_FULL_NAME | QUERY_TYPE );
        std::string     message;
        
        message = "Child property is dirty, but parent property is not.\n" +
                  info.property_name + " (" + info.type_name + ")";

        panic( message.c_str() );
        }

    #endif

    // Check that no more than REPFLAGS_BITS are used.

    hope( only_lower_bits( attributes.repflags, REPFLAGS_BITS ) );
    hope( only_lower_bits( attributes.ptr_repflags, REPFLAGS_BITS ) );

    // Every property/member object must migrate.

    hope( flags_on( attributes.repflags, REPFLAGS_MIGRATE ) );

    // If pointer defines a replication group (ptr_repflags != 0), it must define
    // migration group too (ptr_repflags must be at least REPFLAGS_MIGRATE).

    hope( !attributes.ptr_repflags || 
        flags_on( attributes.ptr_repflags, REPFLAGS_MIGRATE ) );

    replication_flags = attributes.repflags;
    // attributes.ptr_repflags are ignored by ordinary properties.

    // Initialize #available flag.

    if( replica )
        {
        // Replica property is available if it should be replicated to this node.

        ReplicaClient & replica_client = *Global::object_manager().replica_client;
        int repflags_yes;
        int repflags_no;
        replica_client.get_local_repflags( repflags_yes, repflags_no );
        available = flags_on( replication_flags, repflags_yes, repflags_no );
        }
    else
        {
        // All properties are available if object is owned by local node.

        available = true;
        }
    }


// Property information.

Property::PropertyType Property::get_type() const
    {
    return TYPE_VALUE_TYPE;
    }


PointerInfo * Property::get_pointer_info()
    {
    return NULL;
    }


ArrayInfo * Property::get_array_info()
    {
    return NULL;
    }


DictionaryInfo * Property::get_dictionary_info()
    {
    return NULL;
    }


SetInfo * Property::get_set_info()
    {
    return NULL;
    }


Property::PropertyInfo Property::get_info
    (
    int query_info
    ) const
    {
    PropertyInfo        info;

    /* Initialize. */

    info.type = TYPE_UNKNOWN;

    /* Property type. */

    if( flags_on( query_info, QUERY_TYPE ) )
        {
        info.type = get_type();
        info.type_name = typeid( *this ).name();
        }

    /* Property state. */

    if( flags_on( query_info, QUERY_STATE ) )
        {
        if( can_read() )
            {
            info.state = to_string();
            }
        else
            {
            info.state = "<can not read>";
            }
        }

    /* Property name. */

    if( flags_on( query_info, QUERY_NAME ) )
        {
        const Property *    property = this;
        PropertyInfo        other_info;
        std::string         qualified_name;

        while( true )
            {
            if( property->is_root_property() ||
                !property->owner->get_owned_info( property, other_info ) )
                {
                /* This is a root property or owner property does not know
                   anything about @c property. */
                
                qualified_name = property->generated_name() + qualified_name;
                }
            else
                {
                /* We have retrieved property name. */

                qualified_name = other_info.property_name + qualified_name;
                }
            
            if( !flags_on( query_info, QUERY_FULL_NAME ) )
                {
                break;
                }
            
            if( property->is_root_property() )
                {
                break;
                }

            if( property->owner->is_root_property() && 
                flags_on( query_info, QUERY_FULL_NAME_NO_ROOT ) )
                {
                break;
                }

            property = property->owner;
            qualified_name = std::string( "." ) + qualified_name;
            }

        info.property_name = qualified_name;
        }

    return info;
    }


bool Property::get_owned_info
    (
    const Property *            owned,
    Property::PropertyInfo &    info
    ) const
    {
    return false;
    }


std::string Property::to_string() const
    {
    return "<unknown>";
    }


std::string Property::generated_name() const
    {
    std::ostringstream  s;

    s << "property$" << this;

    return s.str();
    }


int Property::hash() const
    {
    return 0;
    }


// Property operations.

void Property::for_each_owned_property_do
    (
    PropertyOperation & operation
    )
    {
    }


// Helper functions.

namespace {

/**
 * Property operation used by Property::memory_size().
 */
struct CalculateMemorySize : public Property::PropertyOperation
    {
    CalculateMemorySize() : accumulator( 0 )
        {
        }

    virtual Result operator()
        (
        Property * property
        )
        {
        accumulator += property->memory_size();
        return RESULT_DONE;
        }
    
    int accumulator;
        /**< Size accumulator. */

    }; // struct CalculateMemorySize

/**
 * Property operation used by Property::bitstream_size().
 */
struct CalculateBitStreamSize : public Property::PropertyOperation
    {
    CalculateBitStreamSize() : accumulator( 0 )
        {
        }

    virtual Result operator()
        (
        Property * property
        )
        {
        accumulator += property->bitstream_size();
        return RESULT_DONE;
        }

    int accumulator;
        /**< Size accumulator. */

    }; // struct CalculateBitStreamSize

#ifdef OPTION_DEBUG

/**
 * Property operation used by @c default_constructed() to check
 * that all owned properties are default-constructed too.
 */
struct CheckDefaultConstructed : public Property::PropertyOperation
    {
    CheckDefaultConstructed
        (
        Property * originator
        ) : originator( originator )
        {
        }

    virtual Result operator()
        (
        Property * property
        )
        {
        if( !property->is_default_constructed() )
            {
            Property::PropertyInfo  originator_info = originator->get_info();
            Property::PropertyInfo  my_info = property->get_info();
            std::string             message;

            message = "default_constructed() was called on a property that "
                      "owns not default-constructed properties.\n"
                      "Property: " + originator_info.property_name + "\n"
                      "Owned, not default-constructed, property: " + 
                      my_info.property_name;

            panic( message.c_str() );
            }

        return RESULT_RECURSE;
        }
    
    Property * originator;
        /**< The property @c default_constructed() was called on. */

    }; // struct CheckDefaultConstructed 

#endif

} // anonymous namespace


int Property::memory_size() const
    {
    CalculateMemorySize operation;
    
    const_cast<Property *>( this )->for_each_owned_property_do( operation );

    return operation.accumulator + sizeof( Property );
    }


int Property::bitstream_size() const
    {
    CalculateBitStreamSize operation;
    
    const_cast<Property *>( this )->for_each_owned_property_do( operation );

    return operation.accumulator;
    }


#ifdef OPTION_DEBUG

void Property::default_constructed()
    {
    // This is a debugging version.
    // The unchecked version is implemented inline.

    // Check whether all owned properties are default-constructed.
    // If not, current property can not be default-constructed.

    CheckDefaultConstructed operation( this );
    for_each_owned_property_do( operation );

    // Perform the operation.

    set_invalid_dirty_time();
    }

#endif


// Write/read handlers.

void Property::write_requested()
    {
    /* Test @c dirty_time consistency. */

    hope( !get_root_owner()->get_dirty_time().is_older_than( get_dirty_time() ) );
    
    /* Archive object. */

    if( root_owner->is_archive_on_write() )
        {
        root_owner->cast_to_object()->archive_now();
        }

    /* Test write permissions. */
    
    #ifdef OPTION_DEBUG

    if( !can_write() )
        {
        std::ostringstream s;

        s   << "Can not write to property \""
            << get_info().property_name
            << "\".";

        RAISE_CORE_LIB_EXCEPTION( CanNotAccessObjectException, s.str() );
        }

    #endif
    
    /*
     * Propagate the request and dirty property owners.
     * If the property is available and it has been already "write_request-ed"
     * in the same replication tick there is no need to propagate the request
     * for the second time.
     */

    #ifdef LAZY_PROPERTY_DIRTYING

    if( available && ( get_dirty_time() == Global::replica_time() ) )
        {
        return;
        }
        
    #endif
    
    /* Dirty property and its owners, mark them as available. */
    
    Property * property = this;
    Property * prev;
    
    do
        {
        hope( !property->get_owner()->get_dirty_time().is_older_than
            ( property->get_dirty_time() ) );

        property->set_dirty_time();
        property->available = true;

        prev = property;
        property = property->owner;
        }
    while( property != prev );
    }


bool Property::is_consistent() const
    {
    ObjectProperty * const object = get_root_owner_object();
    if( object )
        {
        return object->is_consistent();
        }
    return true;
    }


// Property state manipulation.

bool Property::can_write() const
    {
    if( replica && !Global::replica_writes_allowed_hack() )
        {
        return false;
        }

    if( is_root_property() )
        {
        return true;
        }
    else
        {
        return owner->can_write();
        }
    }


bool Property::can_read() const
    {
    if( !available )
        {
        return false;
        }

    if( is_root_property() )
        {
        return true;
        }
    else
        {
        return owner->can_read();
        }
    }


void Property::assign
    (
    const Property & property
    )
    {
    Serializer::Description desc;

    if( typeid( *this ) == typeid( property ) )
        {
        /* Since property types match bitstream serialization can be used. */
        
        static BitIOStreamBuffer    buffer;
        static BitOStream           output;
        static BitIStream           input;

        buffer.clear();

        output.bind( buffer );
        property.bitstr_write( output, desc );
        
        input.bind( buffer );
        bitstr_read( input, desc );
        }
    else
        {
        /* Property types do not match -> use textstream serialization which
           allows for value conversion if the two types are "compatible". */

        std::stringstream   buffer;
            
        /* Write @c property to a text stream. */    
            {
            TextWriter  output( buffer );
            property.text_write( output, desc );
            }

        /* Read @c this from the text stream. */
            {
            TextReader  input( buffer );
            
            input.next_line();
            text_read( input, desc );
            }
        }
    }


void Property::set
    (
    const std::string & string
    )
    {
    Serializer::Description desc;
    std::istringstream      buffer( string );
    TextReader              input( buffer );

    input.next_line();
    text_read( input, desc );
    }


std::string Property::get() const
    {
    Serializer::Description desc;
    std::ostringstream      buffer;
    TextWriter              output( buffer );

    text_write( output, desc );
    return buffer.str();
    }


// Object property.

ObjectProperty * Property::cast_to_object()
    {
    return checked_cast<ObjectProperty *>( this );
    }


const ObjectProperty * Property::cast_to_object() const
    {
    return checked_cast<const ObjectProperty *>( this );
    }


// global

std::ostream & operator<<
    (
    std::ostream &                  stream,
    const Property::Attributes &    attributes
    )
    {
    // Warning: this must be kept in sync with object.idl!

    bool first = true;
    if( attributes.repflags != 0x0f )
        {
        first = false;
        stream << "< repflags = " << attributes.repflags;
        }
    if( attributes.ptr_repflags )
        {
        if( first )
            {
            stream << "< ";
            first = false;
            }
        else
            {
            stream << ", ";
            }
        stream << "< ptr_repflags = " << attributes.ptr_repflags;
        }
    if( !first )
        {
        stream << " >";
        }
    return stream;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "property/property_inline.h"
#endif
