/** @file property.h
 * Properties.
 */

/*
 * begin        : Tue Nov 30 2001
 * copyright    : (C) 2001, 2002, 2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROPERTY_H
#define MASSIV_PROPERTY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_REPLICATION_FLAGS_H
#include "replication/replication_flags.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup property
 * @{
 */

/**@class Property
 * A property is any object that inherits from Property class and
 * conforms to Property interface.
 *
 * Property holds information about a property hierarchy (what Property
 * owns the current one) and replication-related data (replication masks,
 * time of the last modification, etc.).
 * Properties can either be simple value types (like integers) or
 * more complex ones, like containers (own other properties) or objects.
 * Properties "monitor" their changes and either accept the change or not.
 *
 * Any property is:
 *  - serializable
 *  - replicable
 *
 * The following must be implemented when deriving a property:
 *  - constructors
 *  - operator=
 *  - the serialization interface
 *  - the enumeration interface to access owned properties
 *
 * Property hierarchy:
 *  When a new property is created it must properly register to an existing
 *  hierarchy, or build-up a root of the new hierarchy.
 *  By-default any property is owned by a StackRootObject, which is a
 *  special property that owns all properties on program stack.
 *  If the property should belong to a different (non-stack) hierarchy,
 *  it must be reinitialized.
 *  When a property value changes it informs all owners about the change,
 *  so that they could update their "dirty time" (time of the last change).
 *
 *  Currently there are the following hierarchies with these root properties:
 *  - StackRootObject:
 *      owns all default-initialized properties and properties allocated
 *      on the stack
 *  - any Object allocated on the heap ("stand-alone" object):
 *      owns its properties and thus makes its own hierarchy
 *
 * In order to keep track of property changes, any property must be
 * implemented like this:
 *  - default constructor calls @c default_constructed() to inform Property
 *    that it should not be dirty
 *  - copy constructor does not explicitly make the property dirty,
 *    if needed it is done by Property's copy constructor (which _must_ be
 *    called); when copying a non-default-constructed property the new
 *    property will automatically be dirty
 *  - any other constructor does not explicitly make the property dirty,
 *    because it is done automatically by the default Property constructor
 *  - any property change (for example by operator=) must notify property
 *    owners about the change
 *    (@c write_requested() must be called)
 *  - any access to the value held by the property must notify property
 *    owners about the request
 *    (@c read_requested() must be called)
 *
 * @code
 * class SampleProperty : Property
 *     {
 * public:
 *     SampleProperty() { default_constructed(); }
 *     SampleProperty( int i ) : value( i ) { }
 *     SampleProperty( const SampleProperty & p ) : 
 *         value( p.value ), Property( p ) { }
 *     void set( int i ) { write_requested(); value = i; }
 *     int get() const { read_requested(); return value; }
 *     ...
 * protected:
 *     int value;
 *     };
 * @endcode
 */
class MASSIV_EXPORTED Property
    {
    friend class ObjectProperty;

public:

    /**
     * Property type.
     */
    enum PropertyType
        {
        TYPE_UNKNOWN            =   0,
            /**< Unknown type. */
        TYPE_VALUE_TYPE         =   1,
            /**< Simple value type (default). */
        TYPE_OBJECT_PROPERTY    =   2,
            /**< Object. */
        TYPE_POINTER            =   3,
            /**< Pointer. */
        TYPE_CONTAINER          =   4,
            /**< Container. */

        TYPE_COUNT
        }; // PropertyType

    /**
     * PropertyInfo
     */
    struct PropertyInfo
        {
        PropertyType    type;
            /**< Property type class. */
        std::string     type_name;
            /**< Property type (exact type). */
        std::string     property_name;
            /**< Name of property. */
        std::string     state;
            /**< Printable state of the property. */
        }; // PropertyInfo

    /**
     * QueryPropertyInfo.
     */
    enum QueryPropertyInfo
        {
        QUERY_TYPE                  =   1,
            /**< Retrieve type information. */
        QUERY_STATE                 =   2,
            /**< Retrieve property state. */
        QUERY_NAME                  =   4,
            /**< Retrieve property name. */
        QUERY_FULL_NAME             =   8 | QUERY_NAME,
            /**< Retrieve fully qualified property name. */
        QUERY_FULL_NAME_NO_ROOT     =   16 | QUERY_FULL_NAME
            /**< Retrieve fully qualified property name while
                 ignoring root owner object property. */
        }; // QueryPropertyInfo

public:

    /**
     * Per-property attributes.
     *
     * This structure hold values of all property attributes defined by IDL.
     */
    struct MASSIV_EXPORTED Attributes
        {
    public:

        /**
         * Constructor.
         */
        INLINE Attributes();

        /**
         * Reset attributes to default values.
         */
        INLINE void reset();

    public:
        // Do not forget to update operator<< near eof when adding attributes.

        /**
         * Replication flags.
         * See @c src/core/replication/replication_flags.h and
         * @c src/core/object/object.idl for more info.
         */
        //@{
        int     repflags;
        int     ptr_repflags;
        //@}

        }; // struct Attributes

protected:

    struct StackRootTag
        {
        };

private:

    /**
     * Operator=.
     * There is nothing to copy because Property class holds no "value".
     * Making the operator private forces the implementation / derived
     * class to define it.
     * The operator=() defined by the derived class is expected
     * to call @c write_requested().
     */
    Property & operator=
        (
        const Property & property
        );

public:

    //
    // Instance management.
    //

    /**
     * Default constructor.
     *
     * Property is default-initialized as an available stack property.
     * If property is intended to be owned by another property or
     * to be a root property it must be properly reinitialized using
     * @c initialize_from_owner() or ObjectProperty::initialize_object()
     * method.
     * @note Property is automatically marked dirty in order to avoid
     * bugs in replication updates if someone forgot to implement
     * its property constructors properly. To save some bits during
     * serialization call @c default_constructed() from your default
     * constructor to notify Property that it is really default-constructed
     * and that it is not dirty.
     */
    INLINE Property();

    /**
     * One-purpose constructor that is used by the core when StackRootObject
     * is being constructed. The initialization must be handled in a special
     * way, thus default constructor can not be used.
     */
    INLINE Property
        (
        const StackRootTag &
        );

    /**
     * Copy constructor.
     * Must be called from copy constructors defined in the derived classes!
     *
     * Property is default-initialized as a stack property holding a
     * copy value of another property.
     * If property is intended to be owned by another property or
     * to be a root property it must be properly reinitialized using
     * @c initiliaze_xxx() methods.
     * @note If @c property was default-constructed (and has not been modified
     * since it was created) the new property won't be marked dirty too,
     * because it is actually a copy of a "default value".
     * If it is placed in the container it has the semantics "it is there
     * but not dirty, because it is default".
     */
    INLINE Property
        (
        const Property & property
        );

    /**
     * Destructor.
     */
    virtual ~Property();

    /**
     * Initialize property flags from its owner and set per-property
     * attributes. Set up property so that it will be owned by @c direct_owner.
     *
     * This method should be called by root object for each its owned property
     * as soon as the root object is initialized. Container properties should
     * propagate this call to both existing and newly created owned properties.
     *
     * It initializes basic property values from the owner property.
     *
     * @param direct_owner
     *      Owner of this property.
     *      Never equal to @c *this, @c initialize_from_owner will never be
     *      called for root properties (not owned by anyone).
     *
     * @note Nor @c direct_owner nor the property are made dirty.
     */
    virtual void initialize_from_owner
        (
        Property &          direct_owner,
        const Attributes &  attributes
        );

public:

    //
    // Property information.
    //

    /**
     * Returns type of the property.
     */
    virtual PropertyType get_type() const;

    /**
     * Returns pointer info class that describes information
     * about a pointer property.
     * Returns @c NULL if this property is not a pointer.
     */
    virtual PointerInfo * get_pointer_info();

    /**
     * Returns array info class that describes information about an array
     * property (PArray, PValueArray).
     * Returns @c NULL if this property is not an array container.
     */
    virtual ArrayInfo * get_array_info();
    
    /**
     * Returns dictionary info class that describes information about a
     * dictionary property (PDictionary, PMultiDictionary).
     * Returns @c NULL if this property is not a dictionary container.
     */
    virtual DictionaryInfo * get_dictionary_info();
    
    /**
     * Returns set info class that describes information about a
     * set property (PSet, PMultiSet).
     * Returns @c NULL if this property is not a dictionary container.
     */
    virtual SetInfo * get_set_info();

    /**
     * Returns more thorough information on property
     * including its current (printable) state.
     * May be used for debugging.
     * @c query_info (an ORed combination of QueryPropertyInfo)
     * describes what information should be retrieved.
     */
    virtual PropertyInfo get_info
        (
        int query_info = QUERY_FULL_NAME
        ) const;

    /**
     * Fills information about property @c owned that this property owns.
     * Returns true if the information was filled in.
     * Note that owner may know nothing about @c owned property.
     * When invoked by system, @c info is partially filled already,
     * the implementation must fill (short) @c property_name.
     */
    virtual bool get_owned_info
        (
        const Property *    owned,
        PropertyInfo &      info
        ) const;

    /**
     * Returns (printable) property state.
     * Used for debugging purposes.
     */
    virtual std::string to_string() const;

    /**
     * Returns auto-generated property name.
     * This is a helper function for @c get_info().
     */
    virtual std::string generated_name() const;

    /**
     * Computes property hash code.
     */
    virtual int hash() const;

public:

    //
    // Property operations.
    //

    /**
     * Functor that performs operations on a property.
     */
    class MASSIV_EXPORTED PropertyOperation
        {
    protected:

        enum Result
            {
            RESULT_DONE     =   0,
                /**< Operation completed. */
            RESULT_RECURSE  =   1
                /**< Invoke operation recursively on owned properties. */
            };

        friend class Property;

    public:

        /**
         * Default constructor.
         */
        PropertyOperation()
            {
            }

        /**
         * Destructor.
         */
        virtual ~PropertyOperation()
            {
            }

        /**
         * User-defined operation called on a property.
         * Returned result defines what to do next
         * (ie. should invoke operation on owned properties?).
         */
        virtual Result operator()
            (
            Property *      property
            ) = 0;

        }; // PropertyOperation

    /**
     * Invokes the operation on each owned property
     * by calling for_each_property_do() on them.
     * Must be redefined in property descendant because
     * Property does not maintain list of owned properties.
     * @note operation is NOT invoked on this property.
     */
    virtual void for_each_owned_property_do
        (
        PropertyOperation & operation
        );

    /**
     * Invokes operation on the property and owned possible properties
     * if requested (@c PropertyOperation::Result).
     */
    INLINE void for_each_property_do
        (
        PropertyOperation & operation
        );

public:

    //
    // Helper functions.
    //

    /**
     * Returns how much memory the property and its children occupy.
     * The calculation need not be accurate.
     */
    virtual int memory_size() const;

    /**
     * Guesses bitstream serialization size (in bits).
     * The calculation need not be accurate.
     */
    virtual int bitstream_size() const;
        
public:

    //
    // Property state manipulation.
    //

    /**
     * Result of @c should_serialize() method.
     */
    enum ShouldSerializeResult
        {
        SSR_YES,
            /**< Yes, serialize this one. */
        SSR_REPFLAGS_DONT_MATCH,
            /**< No, replication flags do not match. */
        SSR_DEFAULT_CONSTRUCTED,
            /**< Property is default-constructed, it has not been modified yet. */
        SSR_NOT_DIRTY,
            /**< No, property is not dirty. */
        };

    /**
     * Checks whether property should be serialized.
     *
     * @param desc  Description of serialization.
     * @result      See @c Property::ShouldSerializeResult.
     */
    INLINE ShouldSerializeResult should_serialize
        (
        const Serializer::Description & desc
        ) const;

    /**
     * Checks whether property is dirty relatively to @a time,
     * ie. if #dirty_time is greater than @a time.
     */
    INLINE bool is_dirty
        (
        const ReplicaTime & time
        ) const;

    /**
     * Returns dirty time.
     */
    INLINE const ReplicaTime & get_dirty_time() const;

    /**
     * Sets dirty time to current replication time.
     */
    INLINE void set_dirty_time();

    /**
     * Sets dirty time to INVALID.
     */
    INLINE void set_invalid_dirty_time();
    
    /**
     * Tells Property that the property was default-constructed so that
     * it should reset its dirty time to @c INVALID.
     * To be called from subclass's constructors only.
     * @note This is an optimization hint only. If someone forgets to
     * call this nothing bad happens.
     * @pre @c default_constructed() was called on all properties that were
     * owned by this property.
     * @note The property need not be "absolutely-default-constructed".
     * @c default_constructed() just tells that the property value is
     * default in given context. For example 3 may be a default value for
     * an integer property of MyClass.
     */
#ifndef OPTION_DEBUG
    INLINE void default_constructed();
#else
    void default_constructed();
#endif // OPTION_DEBUG

    /**
     * Tests if the property was constructed by a default constructor
     * or copied from a @c is_default_constructed() property and has
     * not been changed since then.
     */
    INLINE bool is_default_constructed() const;

    /**
     * Checks if this property is a replica.
     */
    INLINE bool is_replica() const;

    /**
     * Checks if root owner object is in consistent state.
     *
     * @return @c false if root owner object is inconsistent replica,
     *         @c true otherwise (even if root owner is not object, or if it's
     *         not object replcia).
     */
    bool is_consistent() const;

    /**
     * Checks whether property can be written to.
     */
    bool can_write() const;

    /**
     * Checks whether property can be read.
     */
    bool can_read() const;

    /**
     * Sets arhive-on-write state.
     * Can be called on root owner object property only.
     */
    INLINE void set_archive_on_write
        (
        bool state
        );

    /**
     * Reads current archive-on-write state.
     */
    INLINE bool is_archive_on_write();
    
    /**
     * Tests if the property resides on the program stack.
     */
    INLINE bool is_on_stack() const;

public:

    //
    // Serialization.
    //

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description &
        ) const = 0;

    /**
     * Read from bitstream.
     * If property state will be changed, @c write_requested() must be
     * called before the change (as usually...).
     * @note Unless it is certain that the state will be modified
     * @c write_requested() should not be called (ie. it should be called
     * just before the actual write operation).
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description &
        ) = 0;

    /**
     * Serialize to text stream.
     *
     * Serialization should begin at current write position (no initial
     * whitespace) and no trailing whitespace (including new-line characters)
     * should be written.
     */
    virtual void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description &
        ) const = 0;

    /**
     * Read from text stream.
     * If property state will be changed, @c write_requested() must be
     * called before the change (as usually...).
     * @note Until it is sure that the state will be modified
     * @c write_requested() should not be called (ie. it should be called
     * just before the actual write operation).
     */
    virtual void text_read
        (
        TextReader &                    tr,
        const Serializer::Description &
        ) = 0;

    /**
     * Get @c property state and assign it to @c this.
     * The properties need not be of the same type. If they are incompatible
     * the assignment fails though.
     * Uses serialization to perform the task.
     */
    void assign
        (
        const Property & property
        );

    /**
     * Set property state.
     * Uses serialization to perform the task.
     */
    void set
        (
        const std::string & string
        );
    
    /**
     * Get property state in the format acceptable by @c set().
     * Uses serialization to perform the task.
     */
    std::string get() const;
    
public:

    //
    // Object property.
    //

    /**
     * Returns pointer to direct owner property.
     */
    INLINE Property * get_owner() const;

    /**
     * Returns pointer to root owner property.
     */
    INLINE Property * get_root_owner() const;

    /**
     * Returns root object property that owns this property.
     * Returns NULL if root owner property is not
     * @c TYPE_OBJECT_PROPERTY.
     */
    INLINE ObjectProperty * get_root_owner_object() const;

    /**
     * Checks whether the property is root property.
     */
    INLINE bool is_root_property() const;

    /**
     * Checks whether this property is a root property and
     * its type matches @c TYPE_OBJECT_PROPERTY.
     */
    INLINE bool is_root_owner_object() const;
    
    /**
     * Casts this property to Object.
     */
    ObjectProperty * cast_to_object();

    /**
     * Casts this property to const Object.
     */
    const ObjectProperty * cast_to_object() const;

protected:

    //
    // Write/read handlers.
    //

    /**
     * Must be called before write to the attribute / property.
     * The implementation may reject the write if the property
     * is read-only.
     * Propagates the request into property owners.
     */
    void write_requested();

    /**
     * Must be called before read from the attribute / property.
     * The implementation may reject the read if the property
     * is unavailable.
     * Propagates the request into property owners.
     *
     * Must not be called from helper methods invoked by the core, such as
     * @c initialize_from_owner(), because these methods may be called
     * even if the property is "unavailable".
     * @c read_requested() should be called only if the read is processed
     * on behalf of an user/client request.
     *
     * @bug Should not be inline if both OPTION_OPTIMIZE and
     *      OPTION_DEBUG are defined.
     */
    INLINE void read_requested() const;

private:

    Property *          owner;
        /**< Pointer to property that owns this property.
             @c owner is equal to @c this for root objects. */
    Property *          root_owner;
        /**< Owner property can also be owned by another property.
             Root owner is the root of ownership tree. */
    ReplicaTime         dirty_time;
        /**< Time of last modification of the property.
             Should be @b always more recent than ReplicaTime::INVALID unless
             the property is default constructed (see #default_constructed()).
             As type name suggests, this timestamp is used when replicating
             properties by the delta-update protocol. It's updated to current
             replica time even if this property is owned by object replica
             (when the replica is modified), but #dirty_time of replicas is
             never used by the core and may be freely modified by game library.
             For example, it may be used to check if replica property has been
             updated. */
    bool                available : 1;
        /**< Is this property valid?
             @c True for all properties if object is owned by local node,
             may be @c false for some properties of replicated objects. */
    bool                replica : 1;
        /**< Is this property replica? */
    bool                archive_on_write : 1;
        /**< Perform archive-on-write on this entity? */
    int                 replication_flags : REPFLAGS_BITS;
        /**< IDL/game-defined replication flags.
             See serializer/replication_flags.h for more info. */

    }; // class Property


/**
 * Write property attributes in IDL style.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                  stream,
    const Property::Attributes &    attributes
    );


/**
 * Write property state to output stream.
 */
MASSIV_EXPORTED INLINE std::ostream & operator<<
    (
    std::ostream &      stream,
    const Property &    property
    );

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "property/property_inline.h"
#endif


#endif // MASSIV_PROPERTY_H
