/** @file property_fwd.h
 * Forward declarations of properties.
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_PROPERTY_FWD_H
#define MASSIV_PROPERTY_FWD_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STYPE_FWD_H
#include "serializer/stype_fwd.h"
#endif


namespace Massiv {
namespace Core {

// Property subsystem settings.
    
/**
 * If enabled properties are dirtied only if their @c dirty_time would
 * change. This can be used to optimize out unneccessary write request
 * propagations.
 *
 * @ingroup property
 */
#define LAZY_PROPERTY_DIRTYING

// attribute.h
    
class DefaultAttributeEventHandlers;

// property.h

class Property;

// prop_bool.h

class PBool;

// prop_float.h

template< class Type, class Traits >
class PFloatQ;

template< class Type >
class PFloat;

// prop_int.h

template< class Type, int bits >
class PInt;

// prop_string.h

class PString;

// prop_time.h

template< class VTime >
class PVirtualTime;

// prop_vli.h

template< class Type >
class PVli;

// prop_enum.h

template < class EnumType >
class PEnum;

// vector23.h
// forward declaration is in stype_fwd.h

// color.h
// forward declaration is in stype_fwd.h

// orientation.h
// forward declaration is in stype_fwd.h

// prop_array.h

class ArrayInfo;

template < class SType >
class PValueArray;

template < class PType >
class PArray;

// prop_pair.h

template < class FirstProperty, class SecondProperty >
class PPair;

// prop_dictionary.h

class DictionaryInfo;
class SetInfo;

template < class PropertyType >
class DefaultPropertyHashKey;

template < class PropertyType >
class DefaultPropertyEqualKey;

template
    <
    class KeyProperty,
    class DataProperty,
    class HashKey = DefaultPropertyHashKey< KeyProperty >, 
    class EqualKey = DefaultPropertyEqualKey< KeyProperty >
    >
class PDictionary;

template
    <
    class KeyProperty,
    class DataProperty,
    class HashKey = DefaultPropertyHashKey< KeyProperty >, 
    class EqualKey = DefaultPropertyEqualKey< KeyProperty >
    >
class PMultiDictionary;

template
    <
    class KeyProperty,
    class HashKey = DefaultPropertyHashKey< KeyProperty >, 
    class EqualKey = DefaultPropertyEqualKey< KeyProperty >
    >
class PSet;

template
    <
    class KeyProperty,
    class HashKey = DefaultPropertyHashKey< KeyProperty >, 
    class EqualKey = DefaultPropertyEqualKey< KeyProperty >
    >
class PMultiSet;

// Shortcut typedefs.


/** @addtogroup property
 * @{
 */

/** @name Integral Properties with Fixed Bitsize.
 * All bits will be written during serialization.
 */
//@{
typedef PInt< Int8, 8 >         PInt8;      /**< Singed 8-bit integer. */
typedef PInt< UInt8, 8 >        PUInt8;     /**< Unsigned 8-bit integer. */
typedef PInt< Int16, 16 >       PInt16;     /**< Signed 16-bit integer. */
typedef PInt< UInt16, 16 >      PUInt16;    /**< Unsigned 16-bit integer. */
typedef PInt< Int32, 32 >       PInt32;     /**< Signed 32-bit integer. */
typedef PInt< UInt32, 32 >      PUInt32;    /**< Unsigned 32-bit integer. */
typedef PInt< Int64, 64 >       PInt64;     /**< Signed 64-bit integer. */
typedef PInt< UInt64, 64 >      PUInt64;    /**< Unsigned 64-bit integer. */
//@}


/** @name Integral Properties with Variable-length Bitstream Encoding.
 * Integral property, uses variable-length encoding when serializing to bitstream.
 *
 * @todo use PInt32, etc. typedefs for VLInts? Rename old ints to something
 *       stupid like BitInt?
 */
//@{
typedef PVli< Int32 >           PVlInt32;   /**< Signed 32-bit integer. */
typedef PVli< UInt32 >          PVlUInt32;  /**< Unsigned 32-bit integer. */
typedef PVli< Int64 >           PVlInt64;   /**< Signed 64-bit integer. */
typedef PVli< UInt64 >          PVlUInt64;  /**< Unsigned 64-bit integer. */
//@}


/** @name Floating-point Properties. */
//@{
typedef PFloat< Float32 >       PFloat32;   /**< Single precision (32-bit) float. */
typedef PFloat< Float64 >       PFloat64;   /**< Double precition (64-bit) float. */
//@}


/**
 * Property representing simulation time in seconds.
 * Simulation time is synchronized between all nodes connected to simulation.
 *
 * @ingroup time
 */
typedef PVirtualTime< STime >   PTime;


/** @name Math Vector Properties. */
//@{
typedef Vector23< 2, SFloat32, Property >   PVector2;   /**< 2D single precision vector. */
typedef Vector23< 3, SFloat32, Property >   PVector3;   /**< 3D single precision vector. */
//@}


/** Color property. */
typedef Color< Property >                   PColor;


/** Angular orientation property. */
typedef Orientation< Property >             POrientation;

/** @} */ // addtogroup property


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_PROPERTY_FWD_H
