/** @file property_inline.h
 * Property (inline methods)
 */

/*
 * begin        : Tue Nov 30 2001
 * copyright    : (C) 2001-2003 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SERIALIZER_H
#include "serializer/serializer.h"
#endif

#ifndef MASSIV_GLOBAL_H
#include "global/global.h"
#endif


#ifdef OPTION_DEBUG
#ifndef MASSIV_LIB_EXCEPTIONS_HELPER_H
#include "object/lib/lib_exceptions_helper.h"
#endif

#include <sstream>
#endif

#include <ostream>


namespace Massiv {
namespace Core {


// struct Property::Attributes

Property::Attributes::Attributes()
      : repflags( REPFLAGS_PROPERTY_DEFAULT ),
        ptr_repflags( REPFLAGS_POINTER_DEFAULT )
    {
    }


void Property::Attributes::reset()
    {
    repflags = REPFLAGS_PROPERTY_DEFAULT;
    ptr_repflags = REPFLAGS_POINTER_DEFAULT;
    }


// class Property

Property::Property()
      : owner( Global::get_stack_root_object() ),
        root_owner( Global::get_stack_root_object() ),
        available( true ),
        replica( false ),
        archive_on_write( false ),
        replication_flags( REPFLAGS_PROPERTY_DEFAULT )
    {
    // Since this is the most general constructor we should make
    // the property dirty (just to be safe...).
    // Subclass's default constructor may call @c default_constructed() to
    // undo @c set_dirty_time() (default-constructed properties need not
    // be dirty).

    set_dirty_time();
    }


Property::Property
    (
    const StackRootTag &
    ) : available( true ),
        replica( false ),
        archive_on_write( false ),
        replication_flags( REPFLAGS_PROPERTY_DEFAULT )    
    {
    owner = this;
    root_owner = this;
    }


Property::Property
    (
    const Property & property
    ) : owner( Global::get_stack_root_object() ),
        root_owner( Global::get_stack_root_object() ),
        available( true ),
        replica( false ),
        archive_on_write( false ),
        replication_flags( REPFLAGS_PROPERTY_DEFAULT )
    {
    // Make the new property dirty only iff the copied property
    // is not default-constructed.
    
    if( !property.is_default_constructed() )
        {
        set_dirty_time();
        }
    }


// Property operations.

void Property::for_each_property_do
    (
    PropertyOperation & operation
    )
    {
    PropertyOperation::Result result;

    result = operation( this );

    if( result == PropertyOperation::RESULT_RECURSE )
        {
        for_each_owned_property_do( operation );
        }
    }


// Replication state manipulation.

Property::ShouldSerializeResult Property::should_serialize
    (
    const Serializer::Description & desc
    ) const
    {
    hope( flags_on( desc.repflags_yes, REPFLAGS_MIGRATE ) ||
            ( !desc.repflags_yes && !desc.repflags_no ) );

    // Replication/migration of replicas is not allowed.

    hope( !replica );
    hope( available );

    // Test replication flags.

    if( !desc.replication_flags_match( replication_flags ) )
        {
        return SSR_REPFLAGS_DONT_MATCH;
        }

    // Test if property is default-constructed.

    if( is_default_constructed() )
        {
        return SSR_DEFAULT_CONSTRUCTED;
        }

    // Test dirty stamp.

    if( desc.compare_times && !is_dirty( desc.ack_time ) )
        {
        return SSR_NOT_DIRTY;
        }

    return SSR_YES;
    }


bool Property::is_dirty
    (
    const ReplicaTime & time
    ) const
    {
    return ( dirty_time > time );
    }


const ReplicaTime & Property::get_dirty_time() const
    {
    return dirty_time;
    }


void Property::set_dirty_time()
    {
    dirty_time = Global::replica_time();
    }


void Property::set_invalid_dirty_time()
    {
    dirty_time = ReplicaTime();
    }


#ifndef OPTION_DEBUG

void Property::default_constructed()
    {
    set_invalid_dirty_time();
    }

#endif


bool Property::is_default_constructed() const
    {
    return dirty_time.is_equal_to( ReplicaTime() );
    }


bool Property::is_replica() const
    {
    return replica;
    }


void Property::set_archive_on_write
    (
    bool state
    )
    {
    hope( is_root_owner_object() );
    archive_on_write = state;
    }


bool Property::is_archive_on_write()
    {
    return archive_on_write;
    }


bool Property::is_on_stack() const
    {
    return root_owner == Global::get_stack_root_object();
    }


// Object property.

Property * Property::get_owner() const
    {
    return owner;
    }


Property * Property::get_root_owner() const
    {
    return root_owner;
    }


ObjectProperty * Property::get_root_owner_object() const
    {
    if( root_owner->get_type() == TYPE_OBJECT_PROPERTY )
        {
        return root_owner->cast_to_object();
        }
    else
        {
        return NULL;
        }
    }


bool Property::is_root_property() const
    {
    return owner == this;
    }


bool Property::is_root_owner_object() const
    {
    return is_root_property() && get_type() == TYPE_OBJECT_PROPERTY;
    }


// Write/read handlers.

void Property::read_requested() const
    {
    #ifdef OPTION_DEBUG
        
    if( !can_read() )
        {
        std::ostringstream s;

        s   << "Can not read from property \""
            << get_info().property_name
            << "\".";

        RAISE_CORE_LIB_EXCEPTION( CanNotAccessObjectException, s.str() );
        }

    #endif
    }


// Debugging.

std::ostream & operator<<
    (
    std::ostream &      stream,
    const Property &    property
    )
    {
    stream << property.to_string();
    return stream;
    }


} // namespace Massiv::Core
} // namespace Massiv
