/** @file vector23.h
 * 2D, 3D numeric vector.
 */

/*
 * begin        : Sun Feb 23 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VECTOR23_H
#define MASSIV_VECTOR23_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_MATH_H
#include "utility/math.h"
#endif

#ifndef MASSIV_STYPE_FLOAT_H
#include "serializer/stype_float.h"
#endif

#ifndef MASSIV_PROPERTY_H
#include "property/property.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_TEXT_WRITER_H
#include "utility/text_writer.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup template_types
 * @{
 */

/** @class Vector23Data
 * Class representing components of vector.
 * @c NumDimensions is supposed to be either 2 or 3.
 * @c Type must be a light-weight serializable type (SType),
 * not a property.
 */
template <int NumDimensions, class SType>
class Vector23Data
    {
public:

    SType   comp[ NumDimensions ];
        /**< Components of vector. */

    }; // Vector23Data.


/** @class Vector23
 * 2D or 3D vector of light-weight serializable numeric types.
 * The class implements basic vector operations as they are known
 * from the linear algebra.
 *
 * @param NumDimensions Either 2 or 3.
 * @param SType Lightweight numeric serializable type, preferably
 *              floating-point type. These methods can't be used if @a SType
 *              is integral:
 *              - epsilon_equals()
 *              - epsilon_zero()
 *              - is_normalized()
 *              - magnitude()
 *              - normalize()
 */
template
    <
    int     NumDimensions,
    class   SType,
    class   Base
    >
class Vector23 : public Base
    {
private:

    /// Structure containing vector data.
    typedef Vector23Data<NumDimensions, SType>              VectorData;

    /// Self.
    typedef Vector23<NumDimensions, SType, Base>            ThisVector;

    /// Superclass.
    typedef Base                                            Super;

    /// Native component.
    typedef typename SType::NativeType                      NativeComponent;

    /// To make MSVC6 happy.
    typedef Serializer::Description                         Description;

    enum Constants
        {
        N = NumDimensions
            /**< Shorthand for @c NumDimensions. */
        };

public:
    
    /// Corresponding native type.
    typedef VectorData                                      NativeType;

    /// Corresponding property type.
    typedef Vector23<NumDimensions, SType, Property>        PropertyType;
    
    /// Corresponding stype.
    typedef Vector23<NumDimensions, SType>                  SerializableType;

protected:
    
    using Base::default_constructed;
    using Base::read_requested;
    using Base::write_requested;

public:

    /** Default constructor. Zeroes the vector. */
    Vector23()
        {
        default_constructed();
        }

    /** Constructor for 2D vector. */
    Vector23
        (
        NativeComponent comp0,
        NativeComponent comp1
        )
        {
        hope( N == 2 );
        
        d.comp[ 0 ] = comp0;
        d.comp[ 1 ] = comp1;
        }

    /** Constructor for 3D vector. */
    Vector23
        (
        NativeComponent comp0,
        NativeComponent comp1,
        NativeComponent comp2
        )
        {
        hope( N == 3 );

        d.comp[ 0 ] = comp0;
        d.comp[ 1 ] = comp1;
        d.comp[ 2 ] = comp2;
        }

    /** Copy constructor. */
    Vector23
        (
        const ThisVector &  v
        ) : Super( v )
        {
        int i;

        v.read_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] = v.d.comp[ i ];
            }
        }

    /** Constructor. */
    Vector23
        (
        const VectorData & vd
        )
        {
        int i;
        
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] = vd.comp[ i ];
            }
        }

public:

    /** Operator=. */
    ThisVector & operator=
        (
        const ThisVector &  v
        )
        {
        int i;

        v.read_requested();
        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] = v.d.comp[ i ];
            }

        return *this;
        }

    /** Operator=. */
    ThisVector & operator=
        (
        const VectorData & vd
        )
        {
        int i;

        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] = vd.comp[ i ];
            }

        return *this;
        }

public:

    /** Return vector content. */
    const VectorData & read() const
        {
        read_requested();
        return d;
        }
        
    /** Cast to VectorData. */
    operator const VectorData & () const
        {
        return read();
        }

    /** Get component. */
    NativeComponent operator[]
        (
        int i
        ) const
        {
        hope( i < N );
        read_requested();
        return d.comp[ i ];
        }

    /** Get @a x component. */
    NativeComponent x() const
        {
        return ( *this )[ 0 ];
        }

    /** Get @a y component. */
    NativeComponent y() const
        {
        return ( *this )[ 1 ];
        }

    /** Get @a z component. */
    NativeComponent z() const
        {
        return ( *this )[ 2 ];
        }

public:

    /** Zero all components. */
    void set_zero()
        {
        int i;
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] = static_cast< NativeComponent >( 0 );
            }
        }

    /** Set component. */
    void set
        (
        int         i,
        NativeComponent  value
        )
        {
        hope( i < N );
        write_requested();
        d.comp[ i ] = value;
        }

    /** Set @a x component. */
    void set_x
        (
        NativeComponent value
        )
        {
        set( 0, value );
        }

    /** Set @a y component. */
    void set_y
        (
        NativeComponent value
        )
        {
        set( 1, value );
        }

    /** Set @a z component. */
    void set_z
        (
        NativeComponent value
        )
        {
        set( 2, value );
        }

public:

    /** Compare vectors for exact equality. */
    bool operator==
        (
        const VectorData & vd
        ) const
        {
        int i;

        read_requested();

        for( i = 0; i < N; i++ )
            {
            if( d.comp[ i ] != vd.comp[ i ] )
                return false;
            }

        return true;
        }

    /** Operator!=. */
    bool operator!=
        (
        const VectorData & vd
        ) const
        {
        return !operator==( vd );
        }

    /** Is this vector shorter than @a vd? */
    bool operator<
        (
        const VectorData & vd
        ) const
        {
        ThisVector temp( vd );
        return square_magnitude() < temp.square_magnitude();
        }

    /** Is this vector shorter or the same length as @a vd? */
    bool operator<=
        (
        const VectorData & vd
        ) const
        {
        ThisVector temp( vd );
        return square_magnitude() <= temp.square_magnitude();
        }

    /**
     * Greater function.
     * Return true if every component is greater than an
     * incident component of @param vd.
     */
    bool greater
        (
        const VectorData & vd
        ) const
        {
        int i;

        for( i = 0; i < N; i++ )
            {
            if( d.comp[ i ] <= vd.comp[ i ] )
                return false;
            }

        return true;
        }

    /**
     * Return true if every component is equal to zero.
     */
    bool is_zero() const
        {
        int i;
        for( i = 0; i < N; i++ )
            {
            if( d.comp[ i ] != static_cast< NativeComponent >( 0 ) )
                {
                return false;
                }
            }
        return true;
        }

    /**
     * Epsilon-compare two vectors.
     */
    bool epsilon_equals
        (
        const VectorData & other
        ) const
        {
        read_requested();
        int i;
        for( i = 0; i < N; i++ )
            {
            if( !FloatTraits< NativeComponent >::epsilon_equals( d.comp[ i ], other.comp[ i ] ) )
                {
                return false;
                }
            }
        return true;
        }

    /**
     * Epsilon-compare to zero vector.
     */
    bool epsilon_zero() const
        {
        read_requested();
        int i;
        for( i = 0; i < N; i++ )
            {
            if( !FloatTraits< NativeComponent >::epsilon_zero( d.comp[ i ] ) )
                {
                return false;
                }
            }
        return true;
        }

    /**
     * Is vector length epsilon-equal to 1.0?
     */
    bool is_normalized() const
        {
        return FloatTraits< NativeComponent >::epsilon_equals
            ( square_magnitude(), static_cast< NativeComponent>( 1 ) );
        }

public:

    /** Add vectors. */
    ThisVector & operator+=
        (
        const VectorData & vd
        )
        {
        int i;

        write_requested();
        
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] += vd.comp[ i ];
            }
        
        return *this;
        }

    /** Subtract vectors. */
    ThisVector & operator-=
        (
        const VectorData & vd
        )
        {
        int i;

        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] -= vd.comp[ i ];
            }

        return *this;
        }

    /** Multiply vector components by @a scalar. */
    ThisVector & operator*=
        (
        NativeComponent scalar
        )
        {
        int i;

        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] *= scalar;
            }

        return *this;
        }

    /** Divide vector components by @a scalar. */
    ThisVector & operator/=
        (
        NativeComponent scalar
        )
        {
        int i;

        hope( scalar != static_cast< NativeComponent >( 0 ) );
        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] /= scalar;
            }

        return *this;
        }

    /** Add vectors. */
    ThisVector operator+
        (
        const VectorData & vd
        ) const
        {
        ThisVector  temp( *this );
        temp += vd;
        return temp;
        }

    /** Subtract vectors. */
    ThisVector operator-
        (
        const VectorData & vd
        ) const
        {
        ThisVector  temp( *this );
        temp -= vd;
        return temp;
        }

    /** Get vector multiplied by @a scalar. */
    ThisVector operator*
        (
        NativeComponent scalar
        ) const
        {
        ThisVector temp( *this );
        temp *= scalar;
        return temp;
        }

    /** Get vector divided @a scalar. */
    ThisVector operator/
        (
        NativeComponent scalar
        ) const
        {
        ThisVector temp( *this );
        temp /= scalar;
        return temp;
        }

    /** Square magnitude. */
    NativeComponent square_magnitude() const
        {
        read_requested();
        
        NativeComponent q( static_cast< NativeComponent >( 0 ) );
        int             i;

        for( i = 0; i < N; i++ )
            {
            q += d.comp[ i ] * d.comp[ i ];
            }

        return q;
        }

    /** Magnitude. */
    NativeComponent magnitude() const
        {
        return FloatTraits< NativeComponent >::sqrt( square_magnitude() );
        }

    /** Normalize vector. */
    ThisVector & normalize()
        {
        *this /= magnitude();
        return *this;
        }

    /** Dot product. */
    static NativeComponent dot_product
        (
        const VectorData & vd1,
        const VectorData & vd2
        )
        {
        NativeComponent q( static_cast< NativeComponent >( 0 ) );
        int             i;
        
        for( i = 0; i < N; i++ )
            {
            q += vd1.comp[ i ] * vd2.comp[ i ];
            }

        return q;
        }

    /** Dot product. */
    NativeComponent operator*
        (
        const VectorData & other
        ) const
        {
        return dot_product( *this, other );
        }

    /** Cross product. */
    ThisVector & cross_product
        (
        const VectorData & vd1,
        const VectorData & vd2
        )
        {
        hope( N == 3 );
        write_requested();

        d.comp[ 0 ] = vd1.comp[ 1 ] * vd2.comp[ 2 ] - vd1.comp[ 2 ] * vd2.comp[ 1 ];
        d.comp[ 1 ] = vd1.comp[ 2 ] * vd2.comp[ 0 ] - vd1.comp[ 0 ] * vd2.comp[ 2 ];
        d.comp[ 2 ] = vd1.comp[ 0 ] * vd2.comp[ 1 ] - vd1.comp[ 1 ] * vd2.comp[ 0 ];
        
        return *this;
        }

    /** Cross product. */
    ThisVector operator%
        (
        const VectorData & other
        ) const
        {
        ThisVector result;
        result.cross_product( *this, other );
        return result;
        }

    /** Multiply components by corresponding components of @a other. */
    ThisVector & scale
        (
        const VectorData & other
        )
        {
        write_requested();
        int i;
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] *= other.comp[ i ];
            }
        return *this;
        }

    /** Divide components by corresponding components of @a other. */
    ThisVector & inverse_scale
        (
        const VectorData & other
        )
        {
        write_requested();
        int i;
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ] /= other.comp[ i ];
            }
        return *this;
        }

    /** Unary operator+. */
    ThisVector operator+() const
        {
        ThisVector temp( *this );
        return temp;
        }

    /** Unary operator-. */
    ThisVector operator-() const
        {
        ThisVector temp;
        temp -= *this;
        return temp;
        }

public:
    
    /** Serialize to bitstream. */
    void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        int i;
        
        for( i = 0; i < N; i++ )
            {
            d.comp[ i ].bitstr_write( bos, desc );
            }
        }

    /** Read from bitstream. */
    void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        int i;

        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ].bitstr_read( bis, desc );
            }
        }

    /** Serialize to text stream. */
    void text_write
        (
        TextWriter &        tw,
        const Description & desc
        ) const
        {
        int i;

        for( i = 0; i < N; i++ )
            {
            if( i != 0 )
                {
                tw.write_space();
                }
            d.comp[ i ].text_write( tw, desc );
            }
        }

    /** Read from text stream. */
    void text_read
        (
        TextReader &        tr,
        const Description & desc
        )
        {
        int i;

        write_requested();

        for( i = 0; i < N; i++ )
            {
            d.comp[ i ].text_read( tr, desc );
            }
        }

    /** Print vector state. */
    std::string to_string() const
        {
        std::ostringstream  ss;
        int                 i;

        ss << "( ";

        for( i = 0; i < N - 1; i++ )
            {
            ss << d.comp[ i ] << ", ";
            }

        ss << d.comp[ i ] << " )";

        return ss.str();
        }
    
    /** Compute hash code. */
    int hash() const
        {
        int code = 0;
        int i;

        for( i = 0; i < N; i++ )
            {
            code ^= d.comp[ i ].hash();
            }

        return code;
        }
    
    /** Compute occupied memory size. */
    int memory_size() const
        {
        return sizeof( *this );
        }
    
    /** Guess bitstream serialization size (in bits). */
    int bitstream_size() const
        {
        int size = 0;
        int i;

        for( i = 0; i < N; i++ )
            {
            size += d.comp[ i ].bitstream_size();
            }

        return size;
        }

private:

    VectorData  d;
        /**< Vector components. */

    }; // Vector23


/** Multiply @a vector components by @a scalar. */
template< int NumDimensions, class SType, class Base >
MASSIV_EXPORTED inline Vector23< NumDimensions, SType, Base > operator*
    (
    SType                                           scalar,
    const Vector23< NumDimensions, SType, Base > &  vector
    )
    {
    return vector * scalar;
    }

/** @} */ // addtogroup attribute_types
    

} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_VECTOR23_H
