/** @file config_variable.cpp
 * Config variables that are being stored in the Registry.
 * @see config_variable.h
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/debug.h"
#include "utility/types.h"
#include "registry/conversion.h"
#include "registry/config_variable.h"
#include "registry/registry.h"
#include <stdexcept>
#include <typeinfo>
#include <sstream>

namespace Massiv {
namespace Core {


// Implementation of ConfigVariable
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


ConfigVariable::ConfigVariable()
    {
    }


ConfigVariable::~ConfigVariable()
    {
    }


const SystemTime & ConfigVariable::get_modify_time() const
    {
    return modify_time;
    }


bool ConfigVariable::was_modified
    (
    const SystemTime & time
    ) const
    {
    return modify_time > time;
    }


ConfigVariable::CVarType ConfigVariable::get_type() const
    {
    const std::type_info & info = typeid( *this );
    CVarType type;

    if( info == typeid( ConfigVariableInteger ) )
        type = CVAR_TYPE_INTEGER;
    else
    if( info == typeid( ConfigVariableLargeInteger ) )
        type = CVAR_TYPE_LARGE_INTEGER;
    else
    if( info == typeid( ConfigVariableBoolean ) )
        type = CVAR_TYPE_BOOLEAN;
    else
    if( info == typeid( ConfigVariableFloat ) )
        type = CVAR_TYPE_FLOAT;
    else
    if( info == typeid( ConfigVariableString ) )
        type = CVAR_TYPE_STRING;
    else
        type = CVAR_TYPE_UNKNOWN;

    return type;
    }


ConfigVariableInteger & ConfigVariable::as_integer()
    {
    try
        {
        return checked_cast<ConfigVariableInteger &>( *this );
        }
    catch( std::bad_cast & )
        {
        throw ExCVarTypeMismatch( "Can not cast to ConfigVariableInteger." );
        }
    }


ConfigVariableLargeInteger & ConfigVariable::as_large_integer()
    {
    try
        {
        return checked_cast<ConfigVariableLargeInteger &>( *this );
        }
    catch( std::bad_cast & )
        {
        throw ExCVarTypeMismatch( "Can not cast to ConfigVariableLargeInteger." );
        }
    }


ConfigVariableBoolean & ConfigVariable::as_boolean()
    {
    try
        {
        return checked_cast<ConfigVariableBoolean &>( *this );
        }
    catch( std::bad_cast & )
        {
        throw ExCVarTypeMismatch( "Can not cast to ConfigVariableBoolean." );
        }
    }


ConfigVariableFloat & ConfigVariable::as_float()
    {
    try
        {
        return checked_cast<ConfigVariableFloat &>( *this );
        }
    catch( std::bad_cast & )
        {
        throw ExCVarTypeMismatch( "Can not cast to ConfigVariableFloat." );
        }
    }


ConfigVariableString & ConfigVariable::as_string()
    {
    try
        {
        return checked_cast<ConfigVariableString &>( *this );
        }
    catch( std::bad_cast & )
        {
        throw ExCVarTypeMismatch( "Can not cast to ConfigVariableString." );
        }
    }


void ConfigVariable::revert_to_default()
    {
    write_string( default_value );
    }


template <class T>
static T read_as
    (
    const ConfigVariable * c
    )
    {
    ConfigVariable * cvar = const_cast<ConfigVariable *>( c );

    switch( cvar->get_type() )
        {
        case ConfigVariable::CVAR_TYPE_INTEGER:
            return ManagedConversion<T, Int32>( cvar->as_integer().read() );

        case ConfigVariable::CVAR_TYPE_LARGE_INTEGER:
            return ManagedConversion<T, Int64>( cvar->as_large_integer().read() );

        case ConfigVariable::CVAR_TYPE_BOOLEAN:
            return ManagedConversion<T, bool>( cvar->as_boolean().read() );

        case ConfigVariable::CVAR_TYPE_FLOAT:
            return ManagedConversion<T, Float32>( cvar->as_float().read() );

        case ConfigVariable::CVAR_TYPE_STRING:
            return ManagedConversion<T, std::string>( cvar->as_string().read() );

        default:
            panic( "Invalid config variable type in read_as." );
        }

    hope( false );

    return T();
    }


template <class T>
static void write_as
    (
    ConfigVariable  * cvar,
    const T         & value
    )
    {
    switch( cvar->get_type() )
        {
        case ConfigVariable::CVAR_TYPE_INTEGER:
            cvar->as_integer().write( ManagedConversion<Int32, T>( value ) );
            return;

        case ConfigVariable::CVAR_TYPE_LARGE_INTEGER:
            cvar->as_large_integer().write( ManagedConversion<Int64, T>( value ) );
            return;

        case ConfigVariable::CVAR_TYPE_BOOLEAN:
            cvar->as_boolean().write( ManagedConversion<bool, T>( value ) );
            return;

        case ConfigVariable::CVAR_TYPE_FLOAT:
            cvar->as_float().write( ManagedConversion<Float32, T>( value ) );
            return;

        case ConfigVariable::CVAR_TYPE_STRING:
            cvar->as_string().write( ManagedConversion<std::string, T>( value ) );
            return;

        default:
            panic( "Invalid config variable type in write_as." );
        }

    hope( false );
    }


Int32 ConfigVariable::read_integer() const
    {
    return read_as<Int32>( this );
    }


Int64 ConfigVariable::read_large_integer() const
    {
    return read_as<Int64>( this );
    }


bool ConfigVariable::read_boolean() const
    {
    return read_as<bool>( this );
    }


Float32 ConfigVariable::read_float() const
    {
    return read_as<Float32>( this );
    }


std::string ConfigVariable::read_string() const
    {
    return read_as<std::string>( this );
    }


void ConfigVariable::write_integer
    (
    Int32 value
    )
    {
    write_as( this, value );
    }


void ConfigVariable::write_large_integer
    (
    Int64 value
    )
    {
    write_as( this, value );
    }


void ConfigVariable::write_boolean
    (
    bool value
    )
    {
    write_as( this, value );
    }


void ConfigVariable::write_float
    (
    Float32 value
    )
    {
    write_as( this, value );
    }


void ConfigVariable::write_string
    (
    const std::string & s
    )
    {
    write_as( this, s );
    }


const std::string & ConfigVariable::get_name() const
    {
    return name;
    }


void ConfigVariable::setup
    (
    const std::string   & variable_name,
    const std::string   & value,
    bool                  revert_to_default
    )
    {
    name = variable_name;
    default_value = value;

    if( revert_to_default )
        this->revert_to_default();
    }


void ConfigVariable::setup_default
    (
    const std::string   & value,
    bool                  revert_to_default
    )
    {
    default_value = value;

    if( revert_to_default )
        this->revert_to_default();
    }


bool ConfigVariable::match
    (
    const std::string   & variable_name,
    CVarType              type
    ) const
    {
    return name == variable_name && type == get_type();
    }


bool ConfigVariable::match_default_value
    (
    const std::string   & value
    ) const
    {
    return default_value == value;
    }


} // namespace Massiv::Core
} // namespace Massiv
