/** @file config_variable.h
 * Config variables that are being stored in the Registry.
 * They can be handled as a common primitive types, but
 * moreover contain callbacks called upon their change, etc. They also 
 * make the base class for classes that can be stored in the Registry.
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CONFIG_VARIABLE_H
#define MASSIV_CONFIG_VARIABLE_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup registry
 * @{
 */

/**
 * ConfigVariable.
 * Base class for specialized variables Integer, Boolean, Float, String.
 * ConfigVariable must implement external interface of 
 * @c DefaultAttributeEventHandlers.
 */
class MASSIV_EXPORTED ConfigVariable
    {
    friend class Registry;

protected:

    /**
     * Constructor.
     * Called internally by ManagedAttribute.
     */
    ConfigVariable
        (
        );

    /**
     * Destructor.
     * Virtual to  make ConfigVariable to be a polymorphic type.
     */
    virtual ~ConfigVariable();

    /** @name Config variable callbacks */
    //@{

    /**
     * Called by ManagedAttribute's default constructor.
     */
    inline void default_constructed();

    /**
     * Called by ManagedAttribute when value is going to be modified.
     */
    inline void write_requested();

    /**
     * Called by ManagedAttribute when value is going to be read.
     */
    inline void read_requested() const;

    //@}

public:

    /**
     * Type of config variable.
     */
    enum CVarType
        {
        CVAR_TYPE_UNKNOWN       =   0,
        CVAR_TYPE_INTEGER       =   1,
        CVAR_TYPE_LARGE_INTEGER =   2,
        CVAR_TYPE_BOOLEAN       =   3,
        CVAR_TYPE_FLOAT         =   4,
        CVAR_TYPE_STRING        =   5,
        CVAR_TYPE_COUNT
        };

    /** @name Predefined types of config variables */
    //@{

    typedef IntegralAttribute<Int32, ConfigVariable>    CVarInteger;
    typedef IntegralAttribute<Int64, ConfigVariable>    CVarLargeInteger;
    typedef NumericAttribute<Float32, ConfigVariable>   CVarFloat;
    typedef StringAttribute<ConfigVariable>             CVarString;
    typedef BooleanAttribute<ConfigVariable>            CVarBoolean;

    //@}

public:

    /**
     * Get modification time of the variable.
     * @return The last modification time.
     */
    const SystemTime & get_modify_time() const;

    /**
     * Check if someone wrote to the attribute since @a time.
     * Shortcut for <tt>get_modify_time() \> time</tt>.
     */
    bool was_modified
        (
        const SystemTime & time
        ) const;

    /**
     * Set value to default.
     */
    void revert_to_default();

public:

    /** @name Methods for reading from config variables with conversions */
    //@{

    /**
     * Reads integer from the variable. Perform conversions if neccessary.
     * @return The value of the variable.
     */
    Int32 read_integer() const;

    /**
     * Reads large integer from the variable. Perform conversions if neccessary.
     * @return The value of the variable.
     */
    Int64 read_large_integer() const;

    /**
     * Reads boolean from the variable. Perform conversions if neccessary.
     * @return The value of the variable.
     */
    bool read_boolean() const;

    /**
     * Reads float from the variable. Perform conversions if neccessary.
     * @return The value of the variable.
     */
    Float32 read_float() const;

    /**
     * Reads string from the variable. Perform conversions if neccessary.
     * @return The value of the variable.
     */
    std::string read_string() const;

    //@}

public:

    /** @name Methods for writing into config variables with conversions */
    //@{

    /**
     * Writes integer to the variable. Perform conversions if neccessary.
     * @param value The value to be written into the variable.
     */
    void write_integer
        (
        Int32 value
        );

    /**
     * Writes large integer to the variable. Perform conversions if neccessary.
     * @param value The value to be written into the variable.
     */
    void write_large_integer
        (
        Int64 value
        );

    /**
     * Writes boolean to the variable. Perform conversions if neccessary.
     * @param value The value to be written into the variable.
     */
    void write_boolean
        (
        bool value
        );

    /**
     * Writes float to the variable. Perform conversions if neccessary.
     * @param value The value to be written into the variable.
     */
    void write_float
        (
        Float32 value
        );

    /**
     * Writes string to the variable. Perform conversions if neccessary.
     * @param value The value to be written into the variable.
     */
    void write_string
        (
        const std::string & s
        );

    //@}

public:

    /**
     * Returns variable type.
     * @see CVarType
     */
    CVarType get_type() const;

    /**
     * Returns variable name.
     * Variable must have been placed into the registry.
     */
    const std::string & get_name() const;

    /** @name Methods for accessing value of config variables */
    //@{

    /**
     * Returns reference to ConfigVariableInteger.
     * ConfigVariable must be Int32.
     */
    CVarInteger & as_integer();

    /**
     * Returns reference to ConfigVariableLargeInteger.
     * ConfigVariable must be Int64.
     */
    CVarLargeInteger & as_large_integer();

    /**
     * Returns reference to ConfigVariableBoolean.
     * ConfigVariable must be bool.
     */
    CVarBoolean & as_boolean();

    /**
     * Returns reference to ConfigVariableFloat.
     * ConfigVariable must be float.
     */
    CVarFloat & as_float();

    /**
     * Returns reference to ConfigVariableString.
     * ConfigVariable must be string.
     */
    CVarString & as_string();

    //@}

public:

    /**
     * Called by Registry when variable is created.
     * @param variable_name Name of the variable.
     * @param value The default value to be set for the variable.
     * @param revert_to_default If true, the variable value will be
     *      set to the default value.
     */
    void setup
        (
        const std::string   & variable_name,
        const std::string   & value,
        bool                  revert_to_default = true
        );

    /**
     * Sets default value for the variable. 
     * @param value The variable value.
     * @param revert_to_default If true, the variable value will be
     *      set to the default value.
     */
    void setup_default
        (
        const std::string   & value,
        bool                  revert_to_default = true
        );

    /**
     * Checks whether this variable matches the given criteria.
     * @param variable_name The variable name.
     * @param type The variable type.
     * @return true if this variable matches the given name and type, 
     *      false otherwise.
     */
    bool match
        (
        const std::string   & variable_name,
        CVarType              type
        ) const;

    /**
     * Checks whether this variable matches the default value criteria.
     * @param value The default value.
     * @return true if this variable matches the given default value.
     */
    bool match_default_value
        (
        const std::string   & value
        ) const;

private:

    mutable SystemTime  modify_time;
        /**< Set automatically to current time on write to the variable. */
    std::string         default_value;
        /**< Default value. */
    std::string         name;
        /**< Variable name. */

    }; // ConfigVariable


typedef ConfigVariable::CVarInteger         ConfigVariableInteger;
typedef ConfigVariable::CVarLargeInteger    ConfigVariableLargeInteger;
typedef ConfigVariable::CVarFloat           ConfigVariableFloat;
typedef ConfigVariable::CVarString          ConfigVariableString;
typedef ConfigVariable::CVarBoolean         ConfigVariableBoolean;


// Inline implementation of ConfigVariable
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

inline void ConfigVariable::default_constructed()
    {
    }

inline void ConfigVariable::write_requested()
    {
    modify_time = SystemTime::current_time();
    }

inline void ConfigVariable::read_requested() const
    {
    }

/** @} */ // addtogroup registry


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_CONFIG_VARIABLE_H
