/** @file conversion.h
 * Conversions among primitive types.
 */

/*
 * begin        : Tue Nov 29 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CONVERSION_H
#define MASSIV_CONVERSION_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#include <string>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * General conversion among types that can form attribute type.
 * Use @c ManagedConversion<TargetType, SourceType>( source_type_value )
 * to convert value to @c TargetType.
 *
 * Example:
 * 
 * @code
 * Int32    x_int32 = 123;
 * Float64  x_float64 = ManagedConversion<Float64, Int32>( x );
 * @endcode
 */
template <class TargetType, class SourceType>
class ManagedConversion
    {
public:

    /**
     * Default constructor.
     * Performs the conversion.
     * Should be specialized for @c TargetType, @c SourceType
     * where neccessary.
     */
    ManagedConversion
        (
        const SourceType & v
        )
        {
        value = static_cast<TargetType>( v );
        }

    /** Returns converted value. */
    operator TargetType()
        {
        return value;
        }

private:

    /** Returns string that represents "true". */
    const char * string_true()
        {
        return "true";
        }

    /** Returns string that represents "false". */
    const char * string_false()
        {
        return "false";
        }

protected:

    TargetType value;
        /**< Converted value. */
    }; // ManagedConversion


//
// Conversions from Int32.
//

template<>
inline ManagedConversion<bool, Int32>::ManagedConversion
    (
    const Int32 & v
    ) : value( v != 0 )
    {
    }

template<>
inline ManagedConversion<std::string, Int32>::ManagedConversion
    (
    const Int32 & v
    )
    {
    std::ostringstream s;

    s << v;
    value = s.str();
    }

//
// Conversions from Int64.
//

template<>
inline ManagedConversion<bool, Int64>::ManagedConversion
    (
    const Int64 & v
    ) : value( v != 0 )
    {
    }

template<>
inline ManagedConversion<std::string, Int64>::ManagedConversion
    (
    const Int64 & v
    )
    {
    std::ostringstream s;

    s << v;
    value = s.str();
    }

//
// Conversions from bool.
//

template<>
inline ManagedConversion<std::string, bool>::ManagedConversion
    (
    const bool & v
    ) : value( v ? string_true() : string_false() )
    {
    }

//
// Conversions from float.
//

template<>
inline ManagedConversion<bool, Float32>::ManagedConversion
    (
    const Float32 & v
    ) : value( v != 0.0 )
    {
    }

template<>
inline ManagedConversion<std::string, Float32>::ManagedConversion
    (
    const Float32 & v
    )
    {
    std::ostringstream s;

    s << v;
    value = s.str();
    }

//
// Conversions from std::string.
//

template<>
inline ManagedConversion<Int32, std::string>::ManagedConversion
    (
    const std::string & v
    )
    {
    std::istringstream ss( v );
    ss >> value;
    if( ss.fail() )
        {
        throw ExIOError( "Error converting string to integer." );
        }
    }

template<>
inline ManagedConversion<Int64, std::string>::ManagedConversion
    (
    const std::string & v
    )
    {
    std::istringstream ss( v );
    ss >> value;
    if( ss.fail() )
        {
        throw ExIOError( "Error converting string to large integer." );
        }
    }

template<>
inline ManagedConversion<Float32, std::string>::ManagedConversion
    (
    const std::string & v
    )
    {
    std::istringstream ss( v );
    ss >> value;
    if( ss.fail() )
        {
        throw ExIOError( "Error converting string to float." );
        }
    }

template<>
inline ManagedConversion<bool, std::string>::ManagedConversion
    (
    const std::string & v
    )
    {
    if( v == string_true() )
        {
        value = true;
        }
    else if( v == string_false() )
        {
        value = false;
        }
    else
        {
        std::istringstream ss( v );
        Int32 i;
        ss >> i;
        if( ss.fail() )
            {
            throw ExIOError( "String value is not one of 'true', 'false', integer." );
            }
        value = ( i ? true : false );
        }
    }

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_CONVERSION_H
