/** @file registry.cpp
 * The registry provides a mechanism of configuration files handling.
 * @see registry.h
 */

/*
 * begin        : Thu Apr 18 2002
 * copyright    : (C) 2002 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "utility/debug.h"
#include "utility/types.h"
#include "utility/wildcards.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "utility/quote_string.h"
#include "registry/conversion.h"
#include "registry/config_variable.h"
#include "registry/registry.h"
#include <stdexcept>
#include <string>
#include <sstream>
#include <fstream>
#include <list>
#include <cstdio>

/**@def MAX_ALIAS_DEPTH
 * Maximal length of chain of aliases (one alias might refer to another one,
 * and so on - up to MAX_ALIAS_DEPTH references).
 */
#define MAX_ALIAS_DEPTH     64

namespace Massiv {
namespace Core {

static const char * INVALID_PATH_CHARACTERS = "!()[] \t=\"'";

// Implementation of Registry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::Registry()
    {
    root_node = new RegistryNode( this );
    }


Registry::~Registry()
    {
    delete root_node;
    }


ConfigVariable & Registry::create_variable
    (
    const std::string &         fullname,
    ConfigVariable::CVarType    type,
    const std::string &         default_value
    )
    {
    hope( root_node );

    std::string just_path( fullname );
    std::string just_name;

    split_node_path_rev( just_path, just_name );

    return create_variable( just_name, just_path, type, default_value );
    }  // Registry::create_variable()


ConfigVariable & Registry::create_variable
    (
    const std::string &         name,
    const std::string &         path,
    ConfigVariable::CVarType    type,
    const std::string &         default_value
    )
    {
    hope( root_node );

    RegistryNode * regnode = create_node_path( path );

    return regnode->create_variable( name, type, default_value );
    }  // Registry::create_variable()


ConfigVariable & Registry::create_variable
    (
    const std::string &         name,
    RegistryNode * const        regnode,
    ConfigVariable::CVarType    type,
    const std::string &         default_value
    )
    {
    return regnode->create_variable( name, type, default_value );
    }  // Registry::create_variable()


void Registry::create_symbolic_link
    (
    const std::string & alias,
    const std::string & link_to,
    bool                check_target
    )
    {
    hope( root_node );

    std::string alias_just_path( alias );
    std::string alias_just_name;

    split_node_path_rev( alias_just_path, alias_just_name );

    RegistryNode * regnode = create_node_path( alias_just_path );
    hope( regnode );

    regnode->create_symbolic_link( alias_just_name, link_to, check_target );
    }  // Registry::create_symbolic_link()


void Registry::create_symbolic_link
    (
    const std::string & aliasname,
    const std::string & aliaspath,
    const std::string & link_to,
    bool                check_target
    )
    {
    hope( root_node );

    RegistryNode * regnode = create_node_path( aliaspath );
    hope( regnode );

    regnode->create_symbolic_link( aliasname, link_to, check_target );
    }  // Registry::create_symbolic_link()


void Registry::create_symbolic_link
    (
    const std::string &     aliasname,
    RegistryNode * const    regnode,
    const std::string &     link_to,
    bool                    check_target
    )
    {
    regnode->create_symbolic_link( aliasname, link_to, check_target );
    }  // Registry::create_symbolic_link()


std::string Registry::resolve_symbolic_link
    (
    const std::string & fullalias
    ) const
    {
    std::string     a( fullalias );
    std::string     alias_just_name;
    int             depth = 0;
    bool            found = false;

    RegistryNode *  regnode;

    /* go through the oriented graph of links */
    do
        {
        /* find a RegistryNode that contains the alias */
        split_node_path_rev( a, alias_just_name );
        regnode = find_node_for_path( a );

        LinkList::const_iterator it( regnode->links.find( alias_just_name ) );
        if( it != regnode->links.end() )
            {
            a = it->second;
                // now a contains the target of alias (with path)
            found = true;
            }
        else
            {
            /* alias not found */
            a += ("/" + alias_just_name);
            break;
            }

        if( ++depth > MAX_ALIAS_DEPTH )
            {
            std::ostringstream s;
            s << "alias loop detected while resolving \"" << fullalias << "\"";
            throw ExCVarAliasLoop( s.str() );
            }
        }
    while( true );

    if( !found )
        {
        a = fullalias;
        }

    return a;
    }  // Registry::resolve_symbolic_link()


std::string Registry::resolve_symbolic_link
    (
    const std::string &     plainalias,
    RegistryNode * const    regnode
    ) const
    {
    std::string              result;
    LinkList::const_iterator it( regnode->links.find( plainalias ) );
    VariableList::const_iterator itv;

    if( it != regnode->links.end() )
        {
        /* alias name found */
        result = resolve_symbolic_link( it->second );
        }
    else
    if( itv = regnode->variables.find( plainalias ),
        itv != regnode->variables.end() )
        {
        /* variable name found */
        result = plainalias;
        }
    else
        {
        /* name not found between aliases nor variables*/
        std::ostringstream s;

        s << "Plain alias name \"" << plainalias << "\" not found.";

        throw ExAliasNotFound( s.str() );
        }

    return result;
    }  // Registry::resolve_symbolic_link()


bool Registry::find_variable
    (
    const std::string::value_type *     pattern,
    SearchContext &                     context
    ) const
    {
    return find_variable( pattern, context, *root_node );
    }  // Registry::find_variable()


bool Registry::find_variable
    (
    const std::string::value_type *     pattern,
    SearchContext &                     context,
    RegistryNode &                      search_from
    ) const
    {
    hope( root_node );

    bool state = false;
        /* indicates whether next alias or variable was already found. */

    bool empty_pattern = ( ( pattern == 0 ) || ( pattern[ 0 ] == '\0' ) );
        /* empty_pattern indicates that the caller is interested in any variable. */

    if( (context.get_type() == SearchContext::RRT_VOID)
     || (context.get_type() == SearchContext::RRT_NOT_FOUND ) )
        {
        /* empty context - prepare first search */
        context.regnode = &search_from;
        }


    /* find node that contains some available alias or variable */
    while( true )
        {
        while( context.next_in_node() )
            {
            if( empty_pattern || 
                Wildcards::match( pattern, context.get_name().c_str() ) 
              )
                {
                state = true;
                break;
                }
            }

        if ( state || !context.traverse_nodes() )
            break;
        }

    if( context.get_type() == SearchContext::RRT_NOT_FOUND )
        {
        /* found */
        return false;
        }
    else
        {
        /* not found */
        return true;
        }
    }  // Registry::find_variable()


void Registry::initialize_search_context
    (
    SearchContext &     search_context
    )
    {
    initialize_search_context( search_context, *root_node );
    }  // Registry::initialize_search_context()


void Registry::initialize_search_context
    (
    SearchContext &     search_context,
    RegistryNode &      node
    )
    {
    search_context.type             = SearchContext::RRT_VOID;
    search_context.it_subnodes      = node.subnodes.begin();
    search_context.node_names_stack.erase
        (
        search_context.node_names_stack.begin(),
        search_context.node_names_stack.end()
        );

    if( &node != root_node )
        {
        search_context.regnode = &node;
        search_context.prepare_context( root_node );
        }
    }  // Registry::initialize_search_context()


ConfigVariable & Registry::access_variable
    (
    const std::string & fullname
    )
    {
    ConfigVariable *    cvar;

    std::string varname    = resolve_symbolic_link( fullname );
    RegistryNode * regnode = find_node_for_variable( varname );
    hope( regnode );

    std::string         just_path( varname );
    std::string         just_name;

    split_node_path_rev( just_path, just_name );

    VariableList::const_iterator it( regnode->variables.find( just_name ) );
    if( it == regnode->variables.end() )
        {
        /* variable not found */
        std::ostringstream s;
        s << "Variable \"" << varname << "\" not found.";
        throw ExCVarNotFound( s.str() );
        }

    cvar = it->second;

    return *cvar;
    }  // Registry::access_variable()


ConfigVariable & Registry::access_variable
    (
    const std::string &     name,
    RegistryNode *          regnode
    )
    {
    ConfigVariable *    cvar;

    std::string varname     = resolve_symbolic_link( name, regnode );

    std::string     just_path( varname );
    std::string     just_name;

    split_node_path_rev( just_path, just_name );

    if( !just_path.empty() )
        {
        regnode = find_node_for_path( just_path );
        }

    VariableList::const_iterator it( regnode->variables.find( just_name ) );
    if( it == regnode->variables.end() )
        {
        /* variable not found */
        std::ostringstream s;
        s << "Variable \"" << varname << "\" not found.";
        throw ExCVarNotFound( s.str() );
        }

    cvar = it->second;

    return *cvar;
    }  // Registry::access_variable()


Registry::SearchContext Registry::open_registry_node
    (
    const std::string & path
    )
    {
    SearchContext   sc( this );

    RegistryNode *  regnode = find_node_for_path( path );

    /* node was found (otherwise, find_node_for_path has thrown exception. */
    sc.change_context( *regnode );
    sc.type = SearchContext::RRT_VOID;

    return sc;
    }  // Registry::open_registry_node()


bool Registry::search_registry_node
    (
    const std::string::value_type * pattern,
    SearchContext &                 context
    ) const
    {
    hope( root_node );
    hope( context.regnode );
        /* we assume that the context was previously initialized by open_registry_node */

    context.next_in_node();

    if( pattern[0] != '\0' )
        {
        /* pattern matching */

        while( context.get_type() != SearchContext::RRT_NOT_FOUND )
            {
            if( Wildcards::match( pattern, context.get_name().c_str() ) )
                break;

            context.next_in_node();
            }
        }

    if( context.get_type() == SearchContext::RRT_NOT_FOUND )
        return false;
    else
        return true;
    }  // Registry::search_registry_node()


void Registry::load_config_file
    (
    const std::string & filename
    )
    {
    load_config_file( filename, *root_node );
    }  // Registry::load_config_file()


void Registry::load_config_file
    (
    const std::string & filename,
    RegistryNode &      node
    )
    {
    /* open file */
    std::ifstream ifs_config( filename.c_str() );

    if( !ifs_config )
        {
        std::ostringstream s;
        s << "Cannot open file \"" << filename << "\".";
        throw ExCanNotOpenConfigFile( s.str() );
        }

    /* load config file */
    load_config_file
        ( 
        ifs_config, 
        node, 
        filename
        );
    }  // Registry::load_config_file()


void Registry::load_config_file
    (
    std::istream &      istr,
    const std::string & loading_path
    )
    {
    load_config_file( istr, *root_node, loading_path );
    }  // Registry::load_config_file()


void Registry::load_config_file
    (
    std::istream &      istr,
    RegistryNode &      node,
    const std::string & loading_path
    )
    {
    loading_context_node = &node;
    load_included_config_file( istr, node, loading_path );
    }  // Registry::load_config_file()


void Registry::load_included_config_file
    (
    std::istream &      istr,
    RegistryNode &      node,
    const std::string & loading_path
    )
    {
    TextReader trd( istr );

    while( trd.next_line() )
        {
        // Get one line of the config file
        std::string             line = trd.str();

        // Ensure that the line contains niether DOS nor Linux endline
        std::string::size_type  last_pos =  line.find_last_not_of( "\n\r" );

        if( last_pos != std::string::npos )
            {
            line.resize( last_pos + 1 );
            }
        else
            {
            line.clear();
            continue;
            }

        // One config file line processing
        process_line_of_config_file
                            ( 
                            line, 
                            trd.line_number(), 
                            node,
                            loading_path 
                            );
        }
    }  // Registry::load_included_config_file()


void Registry::save_config_file
    (
    const std::string &         filename,
    const std::string &         path
    )
    {
    save_config_file( filename, *root_node, path );
    }  // Registry::save_config_file()


void Registry::save_config_file
    (
    const std::string &         filename,
    Registry::RegistryNode &    node,
    const std::string &         path
    )
    {
    /* open file */
    std::ofstream ofs_config( filename.c_str() );

    if( !ofs_config )
        {
        std::ostringstream s;
        s << "Cannot open file \"" << filename << "\".";
        throw ExCanNotOpenConfigFile( s.str() );
        }

    /* save config file */
    save_config_file( ofs_config, node, path );    
    }  // Registry::save_config_file()


void Registry::save_config_file
    (
    std::ostream &              ofilestr,
    const std::string &         path
    )
    {
    save_config_file( ofilestr, *root_node, path );
    }  // Registry::save_config_file()


void Registry::save_config_file
    (
    std::ostream &              ofilestr,
    Registry::RegistryNode &    node,
    const std::string &         path
    )
    {
    SearchContext       sc( this );
    initialize_search_context( sc, node );

    RegistryNode *      actual_node = root_node;
        /* Registry node the variables are actually being saved from. */

    ConfigVariable *    cvar;

    TextWriter          twr( ofilestr );

    if( &node == root_node )
        {
        twr.line() << "[ ]";
        twr.end_line();
        }

    while( find_variable( 0, sc, node ) )
        {
        if( sc.get_node() != actual_node )
            {
            /* write new node tag */
            twr.end_line();
            twr.end_line();
            twr.line() << "[ ";
            
            if( !path.empty() )
                {
                twr.line() << path << '/';
                }
            
            twr.line() << sc.get_path() << " ]";
            twr.end_line();

            actual_node = sc.get_node();
            }

        cvar = &sc.get_variable();

        twr.line() << sc.get_name() << " : ";

        if( sc.get_type() == SearchContext::RRT_ALIAS )
            {
            twr.line() << "symlink = " << sc.get_alias_target();
            }
        else
        if( sc.get_type() == SearchContext::RRT_VARIABLE )
            {
            switch( cvar->get_type() )
                {
                case ConfigVariable::CVAR_TYPE_INTEGER:
                    {
                    twr.line() << "integer = " << cvar->read_string();
                    break;
                    }
                case ConfigVariable::CVAR_TYPE_LARGE_INTEGER:
                    {
                    twr.line() << "largeint = " << cvar->read_string();
                    break;
                    }
                case ConfigVariable::CVAR_TYPE_FLOAT:
                    {
                    twr.line() << "float = " << cvar->read_string();
                    break;
                    }
                case ConfigVariable::CVAR_TYPE_STRING:
                    {
                    twr.line() << "string = ";
                    twr.line() << quote_const_string( cvar->read_string(), true );
                    break;
                    }
                case ConfigVariable::CVAR_TYPE_BOOLEAN:
                    {
                    twr.line() << "boolean = " << cvar->read_string();
                    break;
                    }
                default:
                    panic( "Invalid config variable type in registry" );
                    break;
                }
            }

        twr.end_line();
        }
    }  // Registry::save_config_file()


Registry::RegistryNode * Registry::find_node_for_variable
    (
    const std::string & fullname
    ) const
    {
    std::string just_variable;  /* noninteresting, just for technical reasons */
    std::string just_path( fullname );
    std::string node_name;

    split_node_path_rev( just_path, just_variable);

    return find_node_for_path( just_path );
    }  // Registry::find_node_for_variable()


Registry::RegistryNode * Registry::find_node_for_path
    (
    const std::string & path
    ) const
    {
    std::string just_path = path;
    std::string node_name;

    RegistryNode *  node_pointer = root_node;

    while ( !just_path.empty() && node_pointer )
        {
        split_node_path( just_path, node_name );

        if ( !node_name.empty() )
            {
            node_pointer = node_pointer->find_subnode( node_name );
            }
        }

    if( !node_pointer )
        {
        std::ostringstream s;
        s << "Node " << path << " not found.";
        throw ExRegistryNodeNotFound( s.str() );
        }

    return node_pointer;
    }  // Registry::find_node_for_path()



Registry::RegistryNode * Registry::create_node_path
    (
    const std::string & path
    )
    {
    return root_node->create_node_path( path );
    }  // Registry::create_node_path()


void Registry::split_node_path  // static method
    (
    std::string & path,
    std::string & node
    )
    {
    node.erase( node.begin(), node.end() );

    std::string::size_type offset = path.find( "/" );

    if ( offset < std::string::npos )
        {
        /* '/' was found */

        node.insert( 0, path, 0, offset );
            // copies the first node name into node

        path.erase( 0, offset + 1 );
            // erases the first node name, including '/',  from path
        }
    else
        {
        /* '/' was not found */

        node = path;
        path.erase( path.begin(), path.end() );
        }
    }  // Registry::split_node_path()


void Registry::split_node_path_rev  // static method
    (
    std::string & path,
    std::string & variable
    )
    {
    variable.erase( variable.begin(), variable.end() );

    std::string::size_type offset = path.rfind( "/" );

    if ( offset < std::string::npos )
        {
        /* '/' was found */

        variable.insert( 0, path, offset + 1, path.length() - offset );
            // copies the variable name into variable

        path.erase( offset );
            // erases the variable name, including its '/', from the path
        }
    else
        {
        /* '/' was not found */

        variable.swap( path );
        }
    }  // Registry::split_node_path_rev()


Registry::Iterator Registry::begin
    (
    const std::string & path,
    const std::string & pattern
    )
    {
    try
        {
        return Iterator( this, path, pattern );
        }
    catch( ExRegistryNodeEmpty & )
        {
        return Iterator();
        }
    }


Registry::Iterator Registry::end()
    {
    return Iterator();
    }


Registry::ConstIterator Registry::begin
    (
    const std::string & path,
    const std::string & pattern
    ) const
    {
    try
        {
        // FIXME: workaround const_cast<>
        return ConstIterator( const_cast<Registry *>( this ), path, pattern );
        }
    catch( ExRegistryNodeEmpty & )
        {
        return ConstIterator();
        }
    }


Registry::ConstIterator Registry::end() const
    {
    return ConstIterator();
    }


Registry::SubnodeIterator Registry::begin_subnodes
    (
    const std::string & path,
    const std::string & pattern
    )
    {
    return SubnodeIterator( this, path, pattern );
    }


Registry::SubnodeIterator Registry::end_subnodes()
    {
    return SubnodeIterator();
    }


Registry::ConstSubnodeIterator Registry::begin_subnodes
    (
    const std::string & path,
    const std::string & pattern
    ) const
    {
    return ConstSubnodeIterator( const_cast<Registry *>( this ), path, pattern );
    }


Registry::ConstSubnodeIterator Registry::end_subnodes() const
    {
    return ConstSubnodeIterator();
    }



// Auxiliary routines for process_line_of_config_file...
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

enum E_NUM_TYPE
    {
    VV_UNKNOWN,
    VV_INT,
    VV_FLOAT,
    };


/**
 * Throws ExConfigFileSyntaxError.
 * @param message Description of the error.
 * @param filename Name of file in which the error occured; this will be
 *      combined with the message string.
 * @param line_number Number of line in the file in which the error
 *      occured; this will be combined with message string and filename
 *      string.
 */
static NORETURN( void ) throw_syntax
    ( 
    std::string     message, 
    std::string     filename,
    long            line_number 
    )
    {
    std::ostringstream s;

    s << "Config file syntax: file " << filename << ": line " << line_number
      << ": " << message;

    throw ExConfigFileSyntaxError( s.str() );
    }


/*
 * Checks syntax of the number.
 * @param filename Name of the configuration file being processed at the 
 *      moment.
 * @param line_number Number of line being just processed.
 * @exception ExConfigFileSyntaxError
 * @return VV_INT, VV_FLOAT or VV_UNKNOWN.
 */
static E_NUM_TYPE check_number_syntax
    (
    const std::string & str_number,
    const std::string & filename,
    long                line_number
    )
    {
    std::string::size_type   pos = 0;
    bool is_float = false;

    if( str_number[ 0 ] == '-' )
        {
        if( str_number.size() < 2 )
            throw_syntax( "invalid number syntax", filename, line_number );

        pos++;
        }

    pos = str_number.find_first_not_of( "0123456789", pos );

    if( pos != std::string::npos )
        {
        if( str_number[ pos ] != '.' )
            throw_syntax( "invalid number syntax", filename, line_number );
        else
            {
            is_float = true;

            pos = str_number.find_first_not_of( "0123456789", ++pos );

            if( pos != std::string::npos )
                throw_syntax( "invalid number syntax", filename, line_number );
            }

        }

    if( is_float )
        return VV_FLOAT;
    else
        return VV_INT;
    }  // check_number_syntax()


/*
 * Checks syntax of the boolean.
 * @param filename Name of the configuration file being processed at the 
 *      moment.
 * @param line_number Number of line being just processed.
 * @exception ExConfigFileSyntaxError
 * @return true or false according to the string.
 */
static bool check_boolean_syntax
    (
    const std::string & str_bool,
    const std::string & filename,
    long                line_number
    )
    {
    bool result;

    if( str_bool == "false" )
        result = false;
    else
    if( str_bool == "true" )
        result = true;
    else
        throw_syntax( "invalid identifier", filename, line_number );

    return result;
    }  // check_boolean_syntax()


/*
 * Checks syntax of the string.
 * @param filename Name of the configuration file being processed at the 
 *      moment.
 * @param line_number Number of line being just processed.
 * @exception ExConfigFileSyntaxError
 */
static void check_string_syntax
    (
    const std::string & str,
    const std::string & filename, 
    long                line_number
    )
    {
    if( ( str[ 0 ] != '"' ) || ( str[ str.size() - 1 ] != '"' ) )
        throw_syntax( "invalid string syntax", filename, line_number );
    }  // check_string_syntax()


/*
 * Check type syntax.
 * @param filename Name of the configuration file being processed at the 
 *      moment.
 * @param line_number Number of line being just processed.
 * @exception ExConfigFileSyntaxError
 */
static ConfigVariable::CVarType check_type_syntax
    (
    const std::string & str,
    const std::string & filename, 
    long                line_number,
    bool &              is_alias
    )
    {
    ConfigVariable::CVarType    result;

    is_alias = false;

    if( str == "integer" )
        result = ConfigVariable::CVAR_TYPE_INTEGER;
    else
    if( str == "largeint" )
        result = ConfigVariable::CVAR_TYPE_LARGE_INTEGER;
    else
    if( str == "float" )
        result = ConfigVariable::CVAR_TYPE_FLOAT;
    else
    if( str == "boolean" )
        result = ConfigVariable::CVAR_TYPE_BOOLEAN;
    else
    if( str == "string" )
        result = ConfigVariable::CVAR_TYPE_STRING;
    else
    if( str == "symlink" )
        {
        result = ConfigVariable::CVAR_TYPE_UNKNOWN;
        is_alias = true;
        }
    else
        throw_syntax( "invalid variable type", filename, line_number );

    return result;
    }  // Registry::check_type_syntax()


// ... and back to Registry implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

std::string Registry::get_path_from_filename
    (
    std::string             filename
    )
    {
    std::string             result;
    std::string::size_type  pos = filename.find_last_of( "/" );
    
    if( pos != std::string::npos )
        {
        result = filename.substr( 0, pos );
        if( result[ result.size() - 1 ] != '/' )
            {
            result += '/';
            }
        }
    else
        {
        result.erase();
        }

    return result;
    }  // Registry::get_path_from_filename()


void Registry::process_line_of_config_file
    (
    std::string &       line,
    long                line_number,
    RegistryNode &      node,
    const std::string & loading_path
    )
    {
    std::string::size_type begpos = line.find_first_not_of( " \t" );
    std::string::size_type endpos;

    // Check empty line.
    if( begpos == std::string::npos )
        return;

    if( line[ begpos ] == '!' )
        {
        /* process include of another configuration file */
        process_include( line, line_number, node, begpos, loading_path );
        }
    else
    if( line[ begpos ] == '[' )
        {
        /* new context definition */
        begpos = line.find_first_not_of( " [\t", begpos );
        
        if( begpos == std::string::npos )
            {
            throw_syntax( 
                "invalid syntax, expected node name", 
                loading_path,
                line_number
                );
            }

        if( line[ begpos ] != ']' )
            {
            endpos = line.find_last_not_of( " \t" );

            if( line[ endpos ] != ']' )
                {
                throw_syntax( "']' expected", loading_path, line_number );
                }

            endpos = line.find_last_not_of( " ]\t", endpos );

            std::string loading_path_context = 
                        line.substr( begpos, endpos - begpos + 1 );

            if( line[ endpos ] == '/' )  
                {
                loading_path_context.erase( loading_path_context.length() - 1 );
                }

            loading_context_node = 
                        node.create_node_path( loading_path_context );

            hope( loading_context_node );
            }
        else
            {
            loading_context_node = &node;
            }
        }
    else
        {
        /* process variable definition */
        std::string                 var_name;
        std::string                 var_value;
        ConfigVariable::CVarType    var_type = ConfigVariable::CVAR_TYPE_UNKNOWN;

        /* extract varname field */
        endpos = line.find_first_of( ": \t", begpos );

        if( endpos == std::string::npos )
            {
            throw_syntax
                ( 
                "missing variable definition", 
                loading_path, 
                line_number 
                );
            }

        var_name = line.substr( begpos, endpos - begpos );
        check_path_syntax( var_name, loading_path, line_number );

        /* extract vartype field */
        if( line[ endpos ] != ':' )  
            {
            begpos = line.find_first_of( ":", endpos );
            if( begpos != std::string::npos )
                begpos = line.find_first_not_of( " \t:", begpos );
            }
        else
            {
            begpos = line.find_first_not_of( " \t", endpos + 1 );
            }

        if( begpos == std::string::npos )  
            {
            throw_syntax
                ( 
                "missing ':' or type identifier", 
                loading_path,
                line_number 
                );
            }

        endpos = line.find_first_of( " \t=", begpos );
        if( endpos == std::string::npos ) 
            {
            throw_syntax( "missing '='", loading_path, line_number );
            }

        bool    is_alias;
            /* true iff the current line of configuration file defines an 
               alias. */

        var_type =
            check_type_syntax( 
                    line.substr( begpos, endpos - begpos ),
                    loading_path,
                    line_number,
                    is_alias
                    );

        /* extract varvalue field */
        if( line[ endpos ] != '=' )  
            {
            begpos = line.find_first_of( "=", endpos );
            if( begpos != std::string::npos )
                begpos = line.find_first_not_of( " \t=", begpos );
            }
        else
            {
            begpos = line.find_first_not_of( " \t", endpos + 1 );
            }

        if( begpos == std::string::npos )  
            {
            throw_syntax
                ( 
                "missing '=' or variable value", 
                loading_path,
                line_number 
                );
            }

        /* string value has to be handled a special way... */
        // It would not be necessary to do all this ugly stuff if sane parsing
        // using istream formatted extractors was used...
        if( line[ begpos ] == '"' )
            {
            std::string::iterator       it_end = line.begin() + begpos + 1;
            bool                        was_backslash = false;

            while( it_end != line.end() )
                {
                if( *it_end == '\\' )
                    {
                    if( was_backslash )
                        {
                        // Double backslash.
                        
                        was_backslash = false;
                        }
                    else
                        {
                        // First backslash.
                        
                        was_backslash = true;
                        }
                    }
                else
                if( *it_end == '"' )
                    {
                    if( was_backslash )
                        {
                        // Escaped quote.

                        was_backslash = false;
                        }
                    else
                        {
                        // Unescaped double quote - end of string.

                        break;
                        }
                    }
                else
                    {
                    was_backslash = false;
                    }

                it_end++;
                }  // while

            if( ( it_end == line.end() ) || ( *it_end != '"' ) )
                {
                throw_syntax
                    ( 
                    "wrong string syntax, final quote not found", 
                    loading_path,
                    line_number 
                    );             
                }

            endpos = it_end - line.begin() + 1;
            }
        else
            {
            /* Variable value is not a string. */
            endpos = line.find_first_of( " \t", begpos );
            }

        std::string::size_type  line_len = line.length();

        if( ( endpos == std::string::npos ) || ( endpos > line_len ) )  
            {
            endpos = line_len;
            }

        var_value = line.substr( begpos, endpos - begpos );

        process_variable_value_string(
                                var_name,
                                var_value,
                                var_type,
                                is_alias,
                                *loading_context_node,
                                line,
                                loading_path,
                                line_number
                                );
        }
    }  // Registry::process_line_of_config_file()


void Registry::process_include
    ( 
    std::string &           line, 
    long                    line_number, 
    RegistryNode &          node,
    std::string::size_type  pos,
    const std::string &     loading_path
    )
    {
    std::string::size_type  begpos;
    std::string::size_type  endpos;
    std::string::size_type  line_len = line.size();

    begpos = line.find( "!include", pos );

    if( begpos != pos )
        {
        throw_syntax( "invalid include directive", loading_path, line_number );
        }

    begpos += 8;    // 8 == strlen( "!include" )
    
    begpos = line.find_first_not_of( " \t", begpos );

    if( ( begpos == std::string::npos ) || ( begpos >= line_len ) || 
        ( line[ begpos ] != '<' ) )
        {
        throw_syntax
            ( 
            "missing '<' in include directive", 
            loading_path,
            line_number 
            );
        }

    begpos++;
    
    endpos = line.find( '>', begpos );

    if( endpos == std::string::npos )
        {
        throw_syntax
            ( 
            "missing '>' in include directive", 
            loading_path,
            line_number 
            );
        }

    endpos--;

    if( begpos >= endpos + 1 )
        {
        throw_syntax
            ( 
            "missing configuration file path and name in include directive", 
            loading_path,
            line_number 
            );
        }

    std::string     loading_path_only( get_path_from_filename( loading_path ) );

    /* Open the 'nested' configuration file. */
    
    std::string filename
        ( 
        loading_path_only + line.substr( begpos, endpos - begpos + 1 ) 
        );

    /* Convert backslashes to slashes. */

    std::string::size_type      index = 0;
    while( ( index = filename.find( "\\", index ) ) != std::string::npos )
        {
        filename.replace( index, 1, "/" );
        }

    std::ifstream ifs_config( filename.c_str() );

    if( !ifs_config )
        {
        std::ostringstream s;
        s << "Cannot open file \"" << filename << "\".";
        throw ExCanNotOpenConfigFile( s.str() );
        }

    load_included_config_file( ifs_config, node, filename );
    }  // Registry::process_include()


void Registry::process_variable_value_string
    (
    std::string &               var_name,
    std::string &               var_value,
    ConfigVariable::CVarType    var_type,
    bool                        is_alias,
    RegistryNode &              node,
    const std::string &         line,
    const std::string &         filename,
    long                        line_number
    )
    {
    if( var_value.empty() )
        {
        throw_syntax
            ( 
            "undefined variable or alias value", 
            filename, 
            line_number 
            );
        }

    std::string var_just_path_partial( var_name );
    std::string var_just_name;
    split_node_path_rev( var_just_path_partial, var_just_name );
    
    RegistryNode *  relative_node = &node;

    if( !var_just_path_partial.empty() )
        {
        relative_node = node.create_node_path( var_just_path_partial );
            // Append path from the variable/alias definition line to 
            // the currect relative path.
        }    

    if( is_alias )
        {
        check_path_syntax( var_value, filename, line_number );
        relative_node->create_symbolic_link( 
                    var_just_name, 
                    var_value, 
                    false 
                    );
        // todo: set the link to the new variable
        }
    else
        {
        if( var_value[ 0 ] == '"' )
            {
            check_string_syntax( var_value, filename, line_number );

            if( ( var_type != ConfigVariable::CVAR_TYPE_STRING )
             && ( var_type != ConfigVariable::CVAR_TYPE_UNKNOWN ) )
                {
                throw_syntax
                    ( 
                    "invalid variable type", 
                    filename,
                    line_number 
                    );
                }

            // Decode the string.

            std::istringstream iss( var_value );
            hope( iss.good() );
            std::string decoded_value;
            iss >> quote_string( decoded_value );
            if( iss.fail() )
                {
                throw_syntax
                    ( 
                    "failed to decode string variable value", 
                    filename,
                    line_number 
                    );
                }

            relative_node->create_variable
                (
                var_just_name,
                ConfigVariable::CVAR_TYPE_STRING,
                decoded_value
                );
            }
        else
        if( ( ( var_value[ 0 ] >= '0' ) && ( var_value[ 0 ] <= '9' ) )
           || ( var_value[ 0 ] == '-' ) )
            {
            switch( check_number_syntax( var_value, filename, line_number ) )
                {
            case VV_INT:
                if( ( var_type == ConfigVariable::CVAR_TYPE_INTEGER )
                 || ( var_type == ConfigVariable::CVAR_TYPE_LARGE_INTEGER ) )
                    {
                    relative_node->create_variable( 
                        var_just_name,
                        var_type, 
                        var_value
                        );
                    }
                else
                if( var_type == ConfigVariable::CVAR_TYPE_FLOAT )
                    {
                    relative_node->create_variable
                        (
                        var_just_name,
                        ConfigVariable::CVAR_TYPE_FLOAT,
                        var_value
                        );
                    }
                else
                    {
                    throw_syntax
                        ( 
                        "invalid variable type", 
                        filename,
                        line_number 
                        );
                    }
                break;
            case VV_FLOAT:
                if( (var_type != ConfigVariable::CVAR_TYPE_FLOAT )
                 && ( var_type != ConfigVariable::CVAR_TYPE_UNKNOWN ) )
                    {
                    throw_syntax
                        ( 
                        "invalid variable type", 
                        filename,
                        line_number 
                        );
                    }

                relative_node->create_variable
                    (
                    var_just_name,
                    ConfigVariable::CVAR_TYPE_FLOAT,
                    var_value
                    );
                break;
            default:
                break;
                }
            }
        else
            {
            check_boolean_syntax( var_value, filename, line_number );
            if( ( var_type != ConfigVariable::CVAR_TYPE_BOOLEAN )
             && ( var_type != ConfigVariable::CVAR_TYPE_UNKNOWN ) )
                {
                throw_syntax
                    ( 
                    "invalid variable type", 
                    filename,
                    line_number 
                    );
                }

            relative_node->create_variable
                (
                var_just_name,
                ConfigVariable::CVAR_TYPE_BOOLEAN,
                var_value
                );
            }
        }
    }  // Registry::process_variable_value_string()


void Registry::check_name_syntax
    (
    const std::string & name
    )
    {
    bool is_correct = true;

    if( !name.empty() )
        {
        std::string::const_iterator     it( name.begin() );

        while( it != name.end() )
            {
            do
                {
                const char * aux_invalid = INVALID_PATH_CHARACTERS;

                while( *aux_invalid != '\0' )
                    {
                    if( *aux_invalid++ == *it )
                        {
                        is_correct = false;
                        break;
                        }
                    }

                if( !is_correct || ( *it < 32 ) || ( *it == '/' ) )
                    {
                    is_correct = false;
                    break;
                    }
                } while( false );

            if( !is_correct )  break;
            it++;
            }  // while( it != path.end() )
        }  // if( !name.empty() )
    else
        {
        is_correct = false;  // empty node, variable or alias name is invalid
        }

    if( !is_correct )
        {
        std::ostringstream s;
        s << "Invalid variable, node or alias name syntax in \"" << name;
        s << "\".";
        throw ExWrongSyntax( s.str() );
        }
    }  // Registry::check_name_syntax()


void Registry::check_name_syntax
    (
    const std::string & name,
    const std::string & filename,
    long                line_number
    )
    {
    try
        {
        check_name_syntax( name );
        }
    catch( ExWrongSyntax & )
        {
        throw_syntax( "invalid path syntax", filename, line_number );
        }
    }  // Registry::check_name_syntax()


void Registry::check_path_syntax
    (
    const std::string & path
    )
    {
    bool is_correct = true;

    if( !path.empty() )
        {
        std::string::const_iterator     it( path.begin() );
        bool                            was_slash = false;
            /* true iff the previous compared character from path was '/' */

        if( ( path[ 0 ] == '/' ) || ( path[ path.size() - 1 ] == '/' ) )
            {
            is_correct = false;
            }
        else
            {
            while( it != path.end() )
                {
                do
                    {
                    const char * aux_invalid = INVALID_PATH_CHARACTERS;

                    while( *aux_invalid != '\0' )
                        {
                        if( *aux_invalid++ == *it )
                            {
                            is_correct = false;
                            break;
                            }
                        }

                    if( !is_correct || ( *it < 32 ) )
                        {
                        is_correct = false;
                        break;
                        }

                    if( *it == '/' )
                        {
                        if( was_slash )
                            {
                            is_correct = false;
                            break;
                            }
                        else
                            {
                            was_slash = true;
                            break;
                            }
                        }                

                    was_slash = false;
                    } while( false );

                if( !is_correct )  break;
                it++;
                }  // while( it != path.end() )
            }  // else
        }  // if( path.empty() )

    if( !is_correct )
        {
        std::ostringstream s;
        s << "invalid path syntax in \"" << path << "\"";
        throw ExWrongSyntax( s.str() );
        }
    }  // Registry::check_path_syntax()


void Registry::check_path_syntax
    (
    const std::string & path,
    const std::string & filename,
    long                line_number
    )
    {
    try
        {
        check_path_syntax( path );
        }
    catch( ExWrongSyntax & )
        {
        throw_syntax( "invalid path syntax", filename, line_number );
        }
    }  // Registry::check_path_syntax()


// Implementation of Registry::RegistryNode
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::RegistryNode::~RegistryNode()
    {
    SubnodeList::iterator it( subnodes.begin() );

    while ( it != subnodes.end() )
        {
        delete (it++)->second;
        }
    }  // Registry::RegistryNode::~RegistryNode()


ConfigVariable & Registry::RegistryNode::create_variable
    (
    const std::string &         name,
    ConfigVariable::CVarType    type,
    const std::string &         default_value
    )
    {
    VariableList::const_iterator vit( variables.find( name ) );
    LinkList::const_iterator lit( links.find( name ) );
    ConfigVariable *cvar;

    if( vit != variables.end() )
        {
        /* variable already defined, check type */

        cvar = &access_variable( name );
        hope( cvar );

        if( !cvar->match( name, type ) )
            {
            std::ostringstream s;
            s << "variable \"" << name << "\" already defined, type does not match";
            throw ExCVarTypeMismatch( s.str() );
            }

        if( !cvar->match_default_value( default_value ) )
            {
            /* don't create new variable, just alter the existing */
            cvar->setup_default( default_value, true );
            }
        else
            {
            /* default value of the new variable is the same as the one of the
             * old --> just revert the variable to default. */
            cvar->revert_to_default();
            }

        return *vit->second;
        }

    if( lit != links.end() )
        {
        /* link already defined */

        std::ostringstream s;
        s << "name \"" << name << "\" is a symbolic link";
        throw ExCVarDuplicateName( s.str() );
        }

    switch( type )
        {
        case ConfigVariable::CVAR_TYPE_INTEGER:
            cvar = new ConfigVariableInteger();
            break;

        case ConfigVariable::CVAR_TYPE_LARGE_INTEGER:
            cvar = new ConfigVariableLargeInteger();
            break;

        case ConfigVariable::CVAR_TYPE_BOOLEAN:
            cvar = new ConfigVariableBoolean();
            break;

        case ConfigVariable::CVAR_TYPE_FLOAT:
            cvar = new ConfigVariableFloat();
            break;

        case ConfigVariable::CVAR_TYPE_STRING:
            cvar = new ConfigVariableString();
            break;

        default:
            panic( "Invalid config variable type in create_variable" );
        }

    cvar->setup( name, default_value, true );
    variables.insert( VariableList::value_type( name, cvar ) );

    return *cvar;
    } // Registry::RegistryNode::create_variable()


ConfigVariable & Registry::RegistryNode::create_variable
    (
    const std::string &         name,
    const std::string &         path,
    ConfigVariable::CVarType    type,
    const std::string &         default_value
    )
    {
    RegistryNode * regnode = create_node_path( path );
    hope( regnode );

    return regnode->create_variable( name, type, default_value );
    }  // Registry::RegistryNode::create_variable()


void Registry::RegistryNode::create_symbolic_link
    (
    const std::string & alias,
    const std::string & link_to,
    bool                check_target
    )
    {
    LinkList::iterator  it( links.find( alias ) );

    bool                do_insert = true;
        /* false indicates that the symbolic link already exists
           in the registry and refers to the same target - we
           have just to check whether the target exists (in case
           @a check_target is true. */


    if( it != links.end() )
        {
        if( it->second != link_to )
            {
            /* alias already defined, points to another target than 
             * specified by link_to parameter. */

            std::ostringstream s;
            s << "Alias " << alias << " already defined.";
            throw ExCVarDuplicateName( s.str() );
            }
        else
            {
            do_insert = false;  // just check existence of target...
            }
        }

    if( do_insert && ( variables.find( alias ) != variables.end() ) )
        {
        /* alias is real variable name */
        
        std::ostringstream s;
        s << "Config variable " << alias;
        s << " already defined.";
        throw ExCVarDuplicateName( s.str() );
        }


    if( check_target )
        {
        /* check if a target of the link is defined */

        std::string link_to_just_path( link_to );
        std::string link_to_just_name;

        split_node_path_rev( link_to_just_path, link_to_just_name );

        registry_pointer->find_node_for_path( link_to_just_path );
            // gets a pointer to the node that possibly contains the target; ignore the result

        try
            {
            registry_pointer->access_variable( link_to );
            }
        catch( ExCVarNotFound & )
            {
            std::ostringstream s;
            s << "Alias target " << link_to << " undefined.";
            throw ExCVarNotFound( s.str() );
            };
        }

    if( do_insert )
        {
        links.insert( LinkList::value_type( alias, link_to ) );
        }
    }  // Registry::RegistryNode::create_symbolic_link()


void Registry::RegistryNode::create_symbolic_link
    (
    const std::string &     aliasname,
    const std::string &     aliaspath,
    const std::string &     link_to,
    bool                    check_target
    )
    {
    RegistryNode * regnode = create_node_path( aliaspath );
    hope( regnode );

    regnode->create_symbolic_link( aliasname, link_to, check_target );
    }  // Registry::RegistryNode::create_symbolic_link()


Registry::RegistryNode * Registry::RegistryNode::create_subnode
    (
    const std::string & name
    )
    {
    RegistryNode *  new_node = 0;

    SubnodeList::const_iterator snl_it( subnodes.find( name ) );
    if ( snl_it == subnodes.end() )
        {
        /* subnode was not already created */

        new_node = new RegistryNode( registry_pointer );

        /* put the new subnode into subnodes list of this node */
        subnodes.insert( SubnodeList::value_type( name, new_node ) );
        }

    return new_node;
    }  // Registry::RegistryNode::create_subnode()


Registry::RegistryNode * Registry::RegistryNode::create_node_path
    (
    const std::string & path
    )
    {
    std::string path_tmp( path );
    std::string node_name;

    RegistryNode *  node_pointer = this;
    RegistryNode *  tmp_node_pointer;

    check_path_syntax( path );

    while ( !path_tmp.empty() && node_pointer )
        {
        split_node_path( path_tmp, node_name );

        if ( !node_name.empty() )
            {
            tmp_node_pointer = node_pointer->find_subnode( node_name );

            if ( tmp_node_pointer == 0 )
                {
                /* this registry node was not constructed so far */

                tmp_node_pointer = new RegistryNode( registry_pointer );

                node_pointer->subnodes.insert
                    (
                    SubnodeList::value_type
                        (
                        node_name,
                        tmp_node_pointer
                        )
                    );

                node_pointer = tmp_node_pointer;
                }
            else
                {
                /* the node is already constructed */

                node_pointer = tmp_node_pointer;
                }
            }
        }

    return node_pointer;
    }  // Registry::RegistryNode::create_node_path()


Registry::RegistryNode * Registry::RegistryNode::find_subnode
    (
    const std::string & name
    ) const
    {
    SubnodeList::const_iterator snl_it( subnodes.find( name ) );
    RegistryNode *  regnode = 0;

    if ( snl_it != subnodes.end() )
        {
        regnode = snl_it->second;
        }

    return regnode;
    }  // Registry::RegistryNode::find_subnode()


ConfigVariable & Registry::RegistryNode::access_variable
    (
    const std::string & name
    )
    {
    VariableList::const_iterator it( variables.find( name ) );
    ConfigVariable *cvar;

    if( it == variables.end() )
        {
        std::ostringstream s;
        s << "variable resolved to \"" << name << "\" does not exist";
        throw ExCVarNotFound( s.str() );
        }

    cvar = it->second;

    return *cvar;
    } // Registry::RegistryNode::access_variable()


// Implementation of Registry::SearchContext
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::SearchContext::SearchContext
    (
    Registry *              registry_pointer,
    RegistryNode &          regnode,
    const std::string &     name
    ) : cvar( NULL ),
        registry_pointer( registry_pointer ),
        regnode( NULL )
    {
    change_context( regnode, name );
    node_names_stack.erase
        (
        node_names_stack.begin(),
        node_names_stack.end()
        );
    }  // Registry::SearchContext::SearchContext()


void Registry::SearchContext::change_context
    (
    RegistryNode &          regnode
    )
    {
    this->regnode = &regnode;

    if( !regnode.links.empty() )
        {
        type = RRT_ALIAS;
        it_links = regnode.links.begin();
        }
    else
    if ( !regnode.variables.empty() )
        {
        type = RRT_VARIABLE;
        it_vars = regnode.variables.begin();
        }
    else
        {
        std::ostringstream s;
        s << "Registry node empty (" << get_path() << ")";
        throw ExRegistryNodeEmpty( s.str() );
        }

    cvar = 0;
    }  // Registry::SearchContext::change_context()


void Registry::SearchContext::change_context
    (
    RegistryNode &          regnode,
    const std::string &     name
    )
    {
    do
        {
        if( name.compare( "" ) )
            {
            /* name is not empty */

            this->regnode = &regnode;

            /* search for name between aliases */
            it_links = regnode.links.find( name );
            if( it_links != regnode.links.end() )
                {
                /* name found */
                type = RRT_ALIAS;
                break;
                }

            /* search for name between variables */
            it_vars = regnode.variables.find( name );
            if( it_vars != regnode.variables.end() )
                {
                /* name found */
                type = RRT_VARIABLE;
                break;
                }
            else
                {
                std::ostringstream s;
                s << "Registry node empty (" << get_path() << ")";
                throw ExRegistryNodeEmpty( s.str() );
                }
            }
        else
            {
            /* name is empty */
            change_context( regnode );
            }
        } while( false );

        cvar = 0;
    }  // Registry::SearchContext::change_context()


std::string Registry::SearchContext::get_name() const
    {
    std::string name( "" );

    switch( type )
        {
    case RRT_VARIABLE:
        name = it_vars->first;
        break;
    case RRT_ALIAS:
        name = it_links->first;
        break;
    default:
        break;
        }

    return name;
    }  // Registry::SearchContext::access_variable()


Registry::RegistryNode * Registry::SearchContext::get_node() const
    {
    return regnode;
    }  // Registry::SearchContext::get_node()


ConfigVariable & Registry::SearchContext::get_variable()
    {
    if( !cvar )
        {
        if( get_type() == RRT_ALIAS )
            {
            std::string varname =
                registry_pointer->resolve_symbolic_link( 
                                                it_links->first, 
                                                get_node() 
                                                );

            cvar = &registry_pointer->access_variable( varname );
            }
        else
            {
            cvar = it_vars->second;
            }
        }

    return *cvar;
    }   // Registry::SearchContext::get_variable()


std::string Registry::SearchContext::get_alias_target() const
    {
    std::string result;

    if( type == RRT_ALIAS )
        {
        result = it_links->second;
        }
    else
        {
        result = "";
        }

    return result;
    }   // Registry::SearchContext::get_alias_target()


std::string Registry::SearchContext::get_path() const
    {
    std::string result;

    std::list< std::string >::const_iterator it( node_names_stack.begin() );

    if( !node_names_stack.empty() )
        {
        result = *it++;
        }

    while( it != node_names_stack.end() )
        {
        result += "/" + *it++;
        }

    return result;
    }  // Registry::SearchContext::get_path()


void Registry::SearchContext::prepare_context
    (
    RegistryNode *          root_node
    )
    {
    RegistryNode *          myregnode = root_node;
    SubnodeList::iterator   myit_subnodes = root_node->subnodes.begin();

    while( myregnode != regnode )
        {
        bool go_forward = true;
        bool was_found  = false;

        if( myit_subnodes == myregnode->subnodes.end() )
            {
            /* no next subnode available - turn back (as far as there it is
            posible to descent) and descent */

            go_forward = false;

            while( !go_forward )
                {
                if( !regnodes_stack.empty() )
                    {
                    myregnode = regnodes_stack.top();
                    regnodes_stack.pop();

                    hope( !it_subnodes_stack.empty() );
                    myit_subnodes = it_subnodes_stack.top();
                    it_subnodes_stack.pop();

                    node_names_stack.pop_back();

                    if( myit_subnodes != myregnode->subnodes.end() )
                        go_forward = true;
                    }
                else
                    {
                    go_forward = false;
                    type = RRT_NOT_FOUND;
                    break;
                    }
                }
            }

        if( go_forward )
            {
            /* there is a subnode available - descent in a registry tree */

            was_found = true;

            regnodes_stack.push( myregnode );

            myregnode = myit_subnodes->second;

            node_names_stack.push_back( myit_subnodes->first );

            myit_subnodes++;
                /* prepare it_subnodes for the next search */

            it_subnodes_stack.push( myit_subnodes );
                /* push it_subnodes to the stack for this level */

            myit_subnodes = myregnode->subnodes.begin();
            }
        }

    //restoring start state. Only node_names_stack is filled.
    regnode = NULL;
    type    = SearchContext::RRT_VOID;
    
    while( it_subnodes_stack.size() )
        {
        it_subnodes_stack.pop();
        }
    
    while( regnodes_stack.size() )
        {
        regnodes_stack.pop();
        }

    }

bool Registry::SearchContext::traverse_nodes()
    {
    bool go_forward = true;
    bool was_found  = false;

    if( it_subnodes == regnode->subnodes.end() )
        {
        /* no next subnode available - turn back (as far as there it is
           posible to descent) and descent */

        go_forward = false;

        while( !go_forward )
            {
            if( !regnodes_stack.empty() )
                {
                regnode = regnodes_stack.top();
                regnodes_stack.pop();

                hope( !it_subnodes_stack.empty() );
                it_subnodes = it_subnodes_stack.top();
                it_subnodes_stack.pop();

                node_names_stack.pop_back();

                if( it_subnodes != regnode->subnodes.end() )
                    go_forward = true;
                }
            else
                {
                go_forward = false;
                type = RRT_NOT_FOUND;
                break;
                }
            }
        }

    if( go_forward )
        {
        /* there is a subnode available - descent in a registry tree */

        was_found = true;

        regnodes_stack.push( regnode );

        regnode = it_subnodes->second;

        node_names_stack.push_back( it_subnodes->first );

        it_subnodes++;
            /* prepare it_subnodes for the next search */

        it_subnodes_stack.push( it_subnodes );
            /* push it_subnodes to the stack for this level */

        it_subnodes = regnode->subnodes.begin();
        }

    if( !was_found )  type = RRT_NOT_FOUND;

    cvar = 0;
    return was_found;
    }  // Registry::SearchContext::traverse_nodes()


bool Registry::SearchContext::next_in_node()
    {
    bool result = false;

    do
        {
        if( ( type == RRT_VOID ) || ( type == RRT_NOT_FOUND ) )
            {
            /* start search  from beginning */

            it_links = regnode->links.begin();

            if( it_links != regnode->links.end() )
                {
                /* the alias was available */

                type = RRT_ALIAS;
                result = true;
                break;
                }

            it_vars = regnode->variables.begin();

            if( it_vars != regnode->variables.end() )
                {
                /* the variable was available */

                type = RRT_VARIABLE;
                result = true;
                break;
                }
            }

        if( type == RRT_ALIAS )
            {
            /* search among aliases */

            it_links++;
            if( it_links != regnode->links.end() )

                {
                /* next alias was available */

                result = true;
                break;
                }
            else
                {
                /* next alias not found, variables will be tried */

                type = RRT_VOID;
                }
            }

        if( ( type == RRT_VOID ) || ( type == RRT_NOT_FOUND )
         || ( type == RRT_VARIABLE ) )
            {
            if( type == RRT_VARIABLE )
                {
                /* continue search among variables */
                it_vars++;
                }
            else
                {
                /* search among variables from begining */
                it_vars = regnode->variables.begin();
                }

            if( it_vars != regnode->variables.end() )
                {
                /* next variable was available */

                type = RRT_VARIABLE;
                result = true;
                break;
                }
            }
        } while( false );

    if( !result ) type = RRT_NOT_FOUND;

    cvar = 0;
    return result;
    }  // Registry::SearchContext::next_in_node()


// Implementation of Registry::ConstIterator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::ConstIterator::ConstIterator()
    :   context( NULL ),
        pattern()
    {
    /* Context is initialized to RRT_VOID. */
    }


Registry::ConstIterator::ConstIterator
    (
    Registry *          registry,
    const std::string & path,
    const std::string & pattern
    ) : context( registry->open_registry_node( path ) ),
        pattern( pattern )
    {
    operator++();
    }


Registry::ConstIterator & Registry::ConstIterator::operator++()
    {
    Registry * registry;

    registry = context.registry_pointer;
    hope( registry );

    /* Find next variable. */

    registry->search_registry_node( pattern.c_str(), context );

    if( context.get_type() == SearchContext::RRT_NOT_FOUND )
        {

        /* Initialize context to RRT_VOID. */

        registry->initialize_search_context( context );
        }

    return *this;
    }


Registry::ConstIterator Registry::ConstIterator::operator++( int )
    {
    ConstIterator it = *this;
    operator++();
    return it;
    }


const ConfigVariable & Registry::ConstIterator::operator*()
    {
    return context.get_variable();
    }


const ConfigVariable * Registry::ConstIterator::operator->()
    {
    return &operator*();
    }


std::string Registry::ConstIterator::get_name() const
    {
    return context.get_name();
    }


bool Registry::ConstIterator::operator==
    (
    const ConstIterator & it
    ) const
    {
    if( context.get_type() != it.context.get_type() )
        return false;

    if( context.get_type() == SearchContext::RRT_VOID )
        return true;

    if( pattern != it.pattern )
        return false;

    return get_name() == it.get_name();
    }


bool Registry::ConstIterator::operator!=
    (
    const ConstIterator & it
    ) const
    {
    return !operator==( it );
    }


Registry::ConstIterator & Registry::ConstIterator::operator=
    (
    const ConstIterator & it
    )
    {
    context = it.context;
    pattern = it.pattern;

    return *this;
    }


// Implementation of Registry::Iterator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::Iterator::Iterator() : Super()
    {
    }


Registry::Iterator::Iterator
    (
    Registry *          registry,
    const std::string & path,
    const std::string & pattern
    ) : Super( registry, path, pattern )
    {
    }


Registry::Iterator & Registry::Iterator::operator++()
    {
    Super::operator++();
    return *this;
    }


Registry::Iterator Registry::Iterator::operator++( int )
    {
    Iterator it = *this;
    operator++();
    return it;
    }


ConfigVariable & Registry::Iterator::operator*()
    {
    return context.get_variable();
    }


ConfigVariable * Registry::Iterator::operator->()
    {
    return &operator*();
    }


Registry::Iterator & Registry::Iterator::operator=
    (
    const Iterator & it
    )
    {
    Super::operator=( it );
    return *this;
    }


// Implementation of Registry::ConstSubnodeIterator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::ConstSubnodeIterator::ConstSubnodeIterator() 
    : registry( NULL ),
      node( NULL ),
      it_sub(),
      pattern( "*" )
    {
    }


Registry::ConstSubnodeIterator::ConstSubnodeIterator
    (
    Registry *          registry,
    const std::string & path,
    const std::string & pattern
    ) : registry( registry ),
        pattern( pattern )
    {
    node = registry->find_node_for_path( path );
    it_sub = node->subnodes.begin();

    while( ( it_sub != node->subnodes.end() ) && 
        !Wildcards::match( pattern.c_str(), it_sub->first.c_str() ) )
        {
        it_sub++;
        }
    }


Registry::ConstSubnodeIterator & Registry::ConstSubnodeIterator::operator++()
    {
    hope( registry );
    it_sub++;

    while( ( it_sub != node->subnodes.end() ) && 
        !Wildcards::match( pattern.c_str(), it_sub->first.c_str() ) )
        {
        it_sub++;
        }

    return *this;
    }


Registry::ConstSubnodeIterator Registry::ConstSubnodeIterator::operator++( int )
    {
    ConstSubnodeIterator it = *this;
    operator++();
    return it;
    }


const std::string & Registry::ConstSubnodeIterator::operator*()
    {
    return it_sub->first;
    }


const std::string * Registry::ConstSubnodeIterator::operator->()
    {
    return &it_sub->first;
    }


const Registry::RegistryNode * 
Registry::ConstSubnodeIterator::get_subnode() const
    {
    return it_sub->second;
    }


bool Registry::ConstSubnodeIterator::operator==
    (
    const ConstSubnodeIterator & it
    ) const
    {
    bool is_this_end = !registry || !node;

    if( !is_this_end )
        {
        is_this_end = ( it_sub == node->subnodes.end() );
        }

    bool is_it_end   = !it.registry || !it.node;
    
    if( !is_it_end )
        {
        is_it_end = ( it.it_sub == it.node->subnodes.end() );
        }

    if( is_this_end && is_it_end )
        return true;

    if( !is_this_end && !is_it_end )
        {
        return it_sub == it.it_sub;
        }
    
    return false;
    }


bool Registry::ConstSubnodeIterator::operator!=
    (
    const ConstSubnodeIterator & it
    ) const
    {
    return !operator==( it );
    }


Registry::ConstSubnodeIterator & Registry::ConstSubnodeIterator::operator=
    (
    const ConstSubnodeIterator & it
    )
    {
    registry = it.registry;
    it_sub   = it.it_sub;
    pattern  = it.pattern;
    node     = it.node;

    return *this;
    }


// Implementation of Registry::SubnodeIterator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry::SubnodeIterator::SubnodeIterator() : Super()
    {
    }


Registry::SubnodeIterator::SubnodeIterator
    (
    Registry *          registry,
    const std::string & path,
    const std::string & pattern
    ) : Super( registry, path, pattern )
    {
    }


Registry::SubnodeIterator & Registry::SubnodeIterator::operator++()
    {
    Super::operator++();
    return *this;
    }


Registry::SubnodeIterator Registry::SubnodeIterator::operator++( int )
    {
    SubnodeIterator it = *this;
    operator++();
    return it;
    }


std::string Registry::SubnodeIterator::operator*()
    {
    return it_sub->first;
    }


const std::string * Registry::SubnodeIterator::operator->()
    {
    return &it_sub->first;
    }


Registry::RegistryNode * Registry::SubnodeIterator::get_subnode()
    {
    return it_sub->second;
    }


Registry::SubnodeIterator & Registry::SubnodeIterator::operator=
    (
    const SubnodeIterator & it
    )
    {
    Super::operator=( it );
    return *this;
    }


} // namespace Massiv::Core
} // namespace Massiv
