/** @file registry.h
 * The Registry provides a mechanism of configuration files handling. It
 * is able to load/save them and internaly store the configuration values
 * in a special objects - instances of ConfigVariable.
 *
 * The structure of the registry is hierarchical - the full variable name
 * is specified by @em path, in a way similar to file systems (for 
 * example <tt>/monster/troll/default_strength</tt>). This enables to
 * logically split configuration of several modules into various registry
 * subtrees and potentially save it into different configuration files.
 *
 * Note that the core modules don't usually manipulate with the Registry 
 * directly; instead, they make use of the Massiv::Settings class. Instances
 * of this class hold references to the specific config variables and thus 
 * avoid looking the variables up round and round and make it all more 
 * efficient.
 *
 * Moreover, the registry is used by Statistics. These objects also stores 
 * their variables in the registry tree.
 */

/*
 * begin        : Thu Apr 18 2002
 * copyright    : (C) 2002 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_REGISTRY_H
#define MASSIV_REGISTRY_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#ifndef MASSIV_CONFIG_VARIABLE_H
#include "registry/config_variable.h"
#endif

#include <list>
#include <map>
#include <stack>
#include <string>

namespace Massiv {
namespace Core {

class ConfigVariable;
namespace Test { class RegistryTest; }


/** @addtogroup registry
 * @{
 */

/**
 * Hierarchical database that stores 
 * @link ConfigVariable config variables @endlink and enables saving
 * and loading of @link configuration_syntax configuration files @endlink.
 * 
 * Variables are identified by a fully-qualified path; the "directories"
 * in the registry are called nodes. Separated nodes in the path are
 * separated using '/'. Valid node, variable or alias name can't contain
 * the following characters: ()[] ='"/ tab and all characters having ASCII
 * code lower than 32 (non-printable characters).
 *
 * Registry uses the friend nested class Registry::RegistryNode to build
 * the hierarchical registry structure. Each RegistryNode manages a list
 * of variables stored in this node, and similarily a list of aliases and
 * a list of sub-nodes.
 */
class MASSIV_EXPORTED Registry
    {
public:

    class RegistryNode;
    class SearchContext;
    class Iterator;
    class ConstIterator;
    class SubnodeIterator;
    class ConstSubnodeIterator;

    friend class RegistryNode;
    friend class SearchContext;
    friend class Massiv::Core::Test::RegistryTest;

    typedef Iterator                iterator;
    typedef ConstIterator           const_iterator;
    typedef SubnodeIterator         subnode_iterator;
    typedef ConstSubnodeIterator    const_subnode_iterator;

public:

    /**
     * Constructor.
     */
    Registry();

    NO_COPY_CONSTRUCTOR( Registry );
    NO_ASSIGNMENT_OPERATOR( Registry );

public:

    /**
     * Destructor.
     */
    virtual ~Registry();

public:

    /** @name Variable/links manipulation methods */
    //@{

    /**
     * Creates a new variable in the registry node specified by its path.
     *
     * @param name  Fully-qualified variable name
     * @param type  Type of the variable, see CVarType.
     * @param default_value Default (and initial) value.
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     * @return %Reference to the inserted ConfigVariable.
     */
    ConfigVariable & create_variable
        (
        const std::string &         fullname,
        ConfigVariable::CVarType    type,
        const std::string &         default_value
        );

    /**
     * Creates a new variable in the registry node specified by its path.
     *
     * @param name  Plain variable name, without path.
     * @param path  Path of the variable in the registry.
     * @param type  Type of the variable, see CVarType.
     * @param default_value Default (and initial) value.
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     * @return %Reference to the inserted ConfigVariable.
     */
    ConfigVariable & create_variable
        (
        const std::string &         name,
        const std::string &         path,
        ConfigVariable::CVarType    type,
        const std::string &         default_value
        );

    /**
     * Creates a new variable in the given registry node.
     *
     * @param name     Plain variable name, without path.
     * @param regnode  %Pointer to the node that is to contain
     *                 this variable.
     * @param type     Type of the variable, see CVarType.
     * @param default_value Default (and initial) value.
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     * @return %Reference to the inserted ConfigVariable.
     */
    ConfigVariable & create_variable
        (
        const std::string &         name,
        RegistryNode * const        regnode,
        ConfigVariable::CVarType    type,
        const std::string &         default_value
        );

public:

    /**
     * Creates symbolic link to existing variable within the registry.
     *
     * @param fullalias   Fully-qualified symbolic link name
     * @param link_to     Fully-qualified link-target name. The target may be
     *                    either a variable name or another alias name. link_to
     *                    value is restricted the same way as alias is.
     * @param check_target If true, the routine checks if the target is defined.
     * @exception ExCVarDuplicateName The name already exists in the registry.
     * @exception ExCVarNotFound      A target of the alias not defined.
     * @exception ExRegistryNodeNotFound
     * @note If the alias of the specified name already exists in the registry,
     *      but with the same target (@a link_to), no exception will be thrown
     *      (except that the target was not found and @a check_target was
     *      true) and the method will be considered a void operation.
     */
    void create_symbolic_link
        (
        const std::string &     fullalias,
        const std::string &     link_to,
        bool                    check_target = true
        );

    /**
     * Creates symbolic link to existing variable within the registry.
     *
     * @param aliasname   Plain symbolic link name, without path.
     * @param aliaspath   Path of the alias name.
     * @param link_to     Fully-qualified link-target name. The target may be
     *                    either a variable name or another alias name. 
     *                    @a link_to
     *                    value is restricted the same way as alias is.
     * @param check_target If true, the routine checks if the target is defined.
     * @exception ExCVarDuplicateName The name already exists in the registry.
     * @exception ExCVarNotFound      A target of the alias not defined.
     * @exception ExRegistryNodeNotFound
     * @note If the alias of the specified name already exists in the registry,
     *      but with the same target (@a link_to), no exception will be thrown
     *      (except that the target was not found and @a check_target was
     *      true) and the method will be considered a void operation.
     */
    void create_symbolic_link
        (
        const std::string &     aliasname,
        const std::string &     aliaspath,
        const std::string &     link_to,
        bool                    check_target = true
        );

    /**
     * Creates symbolic link to existing variable within the registry.
     *
     * @param aliasname   Fully-qualified symbolic link name
     * @param regnode     %Pointer to the node to store the link.
     * @param link_to     Fully-qualified link-target name. The target may be
     *                    either a variable name or another alias name. link_to
     *                    value is restricted the same way as alias is.
     * @param check_target If true, the routine checks if the target is defined.
     * @exception ExCVarDuplicateName The name already exists in the registry.
     * @exception ExCVarNotFound      A target of the alias not defined.
     * @exception ExRegistryNodeNotFound
     * @note This function was added for clean and effective loading of
     *       configuration file.
     *
     * @note If the alias of the specified name already exists in the registry,
     *      but with the same target (@a link_to), no exception will be thrown
     *      (except that the target was not found and @a check_target was
     *      true) and the method will be considered a void operation.
     */
     void create_symbolic_link
        (
        const std::string &     aliasname,
        RegistryNode * const    regnode,
        const std::string &     link_to,
        bool                    check_target = true
        );

    /**
     * Searches for the variable name pointed to by an alias or by a chain of
     * aliases.
     * If alias points to another alias then name is resolved recursively.
     *
     * @param fullalias   Fully-qualified symbolic link or variable name.
     * @exception ExAliasNotFound The name of the alias not found.
     * @exception ExCVarAliasLoop Too long chain of symbolic links found
     *            (this prevents the alias loop). The chain length constraint
     *            is given by MAX_ALIAS_DEPTH constant.
     * @exception ExRegistryNodeNotFound
     * @return Name of the resolved variable with path.
     * @note This method doesn't check whether the resolved name is really
     *       present in the registry.
     */
    std::string resolve_symbolic_link
        (
        const std::string & fullalias
        ) const;

    /**
     * Searches for the variable name pointed to by an alias or by a chain of
     * aliases.
     * If alias points to another alias then name is resolved recursively.
     *
     * @param plainalias    Plain alias name (without path)
     * @param regnode       %Pointer to the node, where the alias is stored.
     * @exception ExAliasNotFound The name of the alias not found.
     * @exception ExCVarAliasLoop Too long chain of symbolic links found
     *            (this prevents the alias loop). The chain length constraint
     *            is given by MAX_ALIAS_DEPTH constant.
     * @exception ExRegistryNodeNotFound
     * @return Name of the resolved variable with path.
     * @note This method doesn't check whether the resolved name is really
     *       present in the registry.
     */
    std::string resolve_symbolic_link
        (
        const std::string & plainalias,
        RegistryNode * const  regnode
        ) const;

    //@}

    /** @name Registry searching methods */
    //@{

    /**
     * Searches for variable name or alias in the registry. It is necessary
     * to call @a initialize_search_context before the first use of
     * the @a context.
     *
     * @param pattern  Pattern of the variable searched for. Pattern
     *                 can contain special characters '*' and '?'.
     *                 '*' represents arbitrary number of arbitrary
     *                 characters, whereas '?' represents just one
     *                 arbitrary character. Note that @a pattern describes
     *                 really just a VARIABLE NAME, not a path!
     * @param context  Here the method returns the search context;
     *                 this can be stored by caller a used again
     *                 in this place, when the caller wants to
     *                 continue his search.
     * @return true if the variable was found, otherwise false.
     * @see initialize_search_context()
     * @see find_variable( const std::string::value_type*, SearchContext&, RegistryNode& )
     */
    bool find_variable
        (
        const std::string::value_type *     pattern,
        SearchContext &                     context
        ) const;

    /**
     * Searches for variable name or alias in the registry. It is necessary
     * to call @a initialize_search_context() before the first use of
     * the @a context.
     *
     * @param pattern  Pattern of the variable searched for. Pattern
     *                 can contain special characters '*' and '?'.
     *                 '*' represents arbitrary number of arbitrary
     *                 characters, whereas '?' represents just one
     *                 arbitrary character. Note that @a pattern describes
     *                 really just a VARIABLE NAME, not a path!
     * @param context  Here the method returns the search context;
     *                 this can be stored by caller a used again
     *                 in this place, when the caller wants to
     *                 continue his search.
     * @param search_from Specifies a node its subtree should be searched.
     *
     * @return true if the variable was found, otherwise false.
     *
     * @invariant The returned context always references the last variable 
     *    or alias found. If none was found, the context is invalid (its
     *    type is RRT_NOT_FOUND) and should not be used.
     * @invariant Each search starts
     *    -# at the variable or alias next to the one that was referenced
     *       by context before this search (if context's type wasn't one of
     *       RRT_VOID or RRT_NOT_FOUND). The 'next' variable or alias is
     *       taken in the following order:
     *       - next alias
     *       - next variable, if no further alias is present in the node
     *       - the same order in the next node, if no further alias nor
     *         variable is present in the node.
     *    -# at the first alias in the node (if context's type was one of
     *       RRT_VOID or RRT_NOT_FOUND). In case there is no alias in the
     *       node, first variable is considered. If there is even no variable,
     *       the next node is taken.
     *
     * @see SearchContext
     */
    bool find_variable
        (
        const std::string::value_type *     pattern,
        SearchContext &                     context,
        RegistryNode &                      search_from
        ) const;

    /**
     * Initializes search context to RRT_VOID. The context will be
     * "associated" with the whole registry tree.
     *
     * @param search_context Search context.
     */
    void initialize_search_context
        (
        SearchContext &  search_context
        );

    /**
     * Initializes search context to RRT_VOID. The context will be
     * "associated" only with a subtree of a @a node.
     *
     * @param search_context Search context.
     * @param node Specifies the registry node the search context will be
     *        "associated" with.
     */
    void initialize_search_context
        (
        SearchContext &     search_context,
        RegistryNode &      node
        );

    /**
     * Accesses variable by name or its alias.
     *
     * @param fullname Fully-qualified variable or symbolic link name.
     * @exception ExCVarNotFound  The variable was not found.
     * @exception ExAliasNotFound The name of the alias not found.
     * @exception ExCVarAliasLoop Too long chain of symbolic links found
     *            (this prevents the alias loop). The chain length constraint
     *            is given by MAX_ALIAS_DEPTH constant.
     * @exception ExRegistryNodeNotFound The %registry node doesn't exist.
     * @return %Reference to the variable if found.
     */
    ConfigVariable & access_variable
        (
        const std::string & fullname
        );

    /**
     * Accesses variable by name or its alias.
     *
     * @param name Variable or symbolic link name relative to the given
     *      @a regnode.
     * @param regnode %Registry node the variable is given relative to.
     * @exception ExCVarNotFound  The variable was not found.
     * @exception ExAliasNotFound The name of the alias not found.
     * @exception ExCVarAliasLoop Too long chain of symbolic links found
     *            (this prevents the alias loop). The chain length constraint
     *            is given by MAX_ALIAS_DEPTH constant.
     * @exception ExRegistryNodeNotFound The %registry node doesn't exist.
     * @return %Reference to the variable if found.
     */
    ConfigVariable & access_variable
        (
        const std::string &     name,
        RegistryNode *          regnode
        );

    /**
     * Opens a registry node. After the node is opened, it is possible
     * to search for variable within this node using the method
     * @a search_registry_node().
     *
     * @param path Path to the node.
     * @exception ExRegistryNodeNotFound
     * @return SearchContext object that can be stored by caller and
     *         used later in function @a search_registry_node().
     * @see search_registry_node
     * @note No close operation is needed.
     */
    SearchContext open_registry_node
        (
        const std::string & path
        );

    /**
     * Similar to @c find_variable(), but the search is restricted just
     * to the node given by @a context.
     *
     * @param pattern  Pattern of the variable searched for. Pattern
     *                 can contain special characters '*' and '?'.
     *                 '*' represents arbitrary number of arbitrary
     *                 characters, whereas '?' represents just one
     *                 arbitrary character. Note that @a pattern describes
     *                 really just a <em>variable name</em>, not a path!
     * @param context  Here the method returns the search context;
     *                 this can be stored by caller a used again
     *                 in this place, when the caller wants to
     *                 continue his search.
     * @return true if the variable or alias was found, otherwise false
     * @see open_registry_node
     */
    bool search_registry_node
        (
        const std::string::value_type * pattern,
        SearchContext &                 context
        ) const;

    //@}

public:

    /** @name Registry loading and saving methods */
    //@{

    /**
     * Loads registry from a configuration file. Format of the file
     * is described in @ref configuration_syntax.
     *
     * @param filename  Name of the file with path.
     * @exception ExCannotOpenConfigFile
     * @exception ExConfigFileSyntaxError
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     *
     * @note Symbolic link targets are not checked during the load!
     *
     * @note None of variables already present in the registry will 
     *      be deleted. 
     *
     * @note Variables that are present in the registry and that
     *      are contained in the configuration file with another
     *      value (but the same type) will be assigned both the default
     *      and actual value to the value specified in the configuration
     *      file. Unlike variables, symbolic links cannot be altered to
     *      refer to another target.
     */
    void load_config_file
        (
        const std::string & filename
        );

    /**
     * Loads registry from a configuration file. Format of the file
     * is described in @ref configuration_syntax.
     *
     * @param filename  Name of the file with path.
     * @param node      All the stuff that is about to be loaded will
     *                  be placed relatively to this node (into its 
     *                  subtree).
     * @exception ExCannotOpenConfigFile
     * @exception ExConfigFileSyntaxError
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     *
     * @note Symbolic link targets are not checked during the load!
     *
     * @note None of variables already present in the registry will 
     *      be deleted. 
     *
     * @note Variables that are present in the registry and that
     *      are contained in the configuration file with another
     *      value (but the same type) will be assigned both the default
     *      and actual value to the value specified in the configuration
     *      file. Unlike variables, symbolic links cannot be altered to
     *      refer to another target.
     */
    void load_config_file
        (
        const std::string & filename,
        RegistryNode &      node
        );

    /**
     * Loads registry from a stream. Input format is described in
     * @ref configuration_syntax.
     *
     * @param istr      Input stream.
     * @param loading_path %Path to the configuration file currently
     *      being loaded. Needed because of the @c !include directive.
     * @exception ExConfigFileSyntaxError
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     *
     * @note Symbolic link targets are not checked during the load!
     *
     * @note None of variables already present in the registry will 
     *      be deleted. 
     *
     * @note Variables that are present in the registry and that
     *      are contained in the configuration file with another
     *      value (but the same type) will be assigned both the default
     *      and actual value to the value specified in the configuration
     *      file. Unlike variables, symbolic links cannot be altered to
     *      refer to another target.
     */
    void load_config_file
        (
        std::istream &      istr,
        const std::string & loading_path = ""
        );

    /**
     * Loads registry from a stream. Input format is described in
     * @ref configuration_syntax.
     *
     * @param istr      Input stream.
     * @param node      All the stuff that is about to be loaded will
     *                  be placed in a subtree of registry specified
     *                  by this node.
     * @param loading_path Path to the configuration file currently
     *      being loaded. Needed because of !include directive.
     *
     * @exception ExConfigFileSyntaxError
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     *
     * @note Symbolic link targets are not checked during the load!
     *
     * @note None of variables already present in the registry will 
     *      be deleted. 
     *
     * @note Variables that are present in the registry and that
     *      are contained in the configuration file with another
     *      value (but the same type) will be assigned both the default
     *      and actual value to the value specified in the configuration
     *      file. Unlike variables, symbolic links cannot be altered to
     *      refer to another target.
     */
    void load_config_file
        (
        std::istream &      istr,
        RegistryNode &      node,
        const std::string & loading_path = ""
        );

    /**
     * Loads nested configuration (skips loading initialization) file 
     * from a stream. Input format is described in
     * @ref configuration_syntax. This method differs from
     * @c load_config_file in fact that this one doesn't initialize
     * loading_context_node and thus can be used even for loading
     * 'nested' (included) configuration files.
     *
     * @param istr      Input stream.
     * @param node      All the stuff that is about to be loaded will
     *                  be placed in a subtree of registry specified
     *                  by this node.
     * @param loading_path Path to the configuration file currently
     *      being loaded. Needed because of !include directive.
     * @exception ExConfigFileSyntaxError
     * @exception ExCVarTypeMismasch The variable of the same name
     *            already exists, but has different type.
     * @exception ExCVarDuplicateName The symbolic link of the
     *            same name already exists.
     *
     * @note Symbolic link targets are not checked during the load!
     *
     * @note None of variables already present in the registry will 
     *      be deleted. 
     *
     * @note Variables that are present in the registry and that
     *      are contained in the configuration file with another
     *      value (but the same type) will be assigned both the default
     *      and actual value to the value specified in the configuration
     *      file. Unlike variables, symbolic links cannot be altered to
     *      refer to another target.
     */
    void load_included_config_file
        (
        std::istream &      istr,
        RegistryNode &      node,
        const std::string & loading_path = ""
        );

    /**
     * Saves registry to a configuration file. Format of the file
     * is described in @ref configuration_syntax.
     *
     * @param filename  Name of the file with path.
     * @param path Specifies a path associated with the root of the
     *      hierarchy being saved. Next time, this specific configuration 
     *      will be loaded under the node given by this @c path. Note
     *      that the path may not be '/'-terminated!
     *
     * @exception ExCannotOpenConfigFile
     */
    void save_config_file
        (
        const std::string &         filename,
        const std::string &         path = ""
        );

    /**
     * Saves a registry subtree given by @a node to a file. Output format
     * is described in @ref configuration_syntax.
     *
     * @param filename  Name of the file with path.
     * @param node      Specifies a node its subtree will be saved.
     * @param path Specifies a path associated with the root of the
     *      hierarchy being saved. Next time, this specific configuration 
     *      will be loaded under the node given by this @c path. Note
     *      that the path may not be '/'-terminated!
     */
    void save_config_file
        (
        const std::string &         filename,
        Registry::RegistryNode &    node,
        const std::string &         path = ""
        );

    /**
     * Saves registry to a stream. Output format is described in
     * @ref configuration_syntax.
     *
     * @param ostr      Output stream.
     * @param path Specifies a path associated with the root of the
     *      hierarchy being saved. Next time, this specific configuration 
     *      will be loaded under the node given by this @c path. Note
     *      that the path may not be '/'-terminated!
     */
    void save_config_file
        (
        std::ostream &              ofilestr,
        const std::string &         path = ""
        );

    /**
     * Saves a registry subtree given by @a node to a stream. Output format 
     * is described in @ref configuration_syntax.
     *
     * @param ostr      Output stream.
     * @param node      Specifies a node its subtree will be saved.
     * @param path Specifies a path associated with the root of the
     *      hierarchy being saved. Next time, this specific configuration 
     *      will be loaded under the node given by this @c path. Note
     *      that the path may not be '/'-terminated!
     */
    void save_config_file
        (
        std::ostream &              ofilestr,
        Registry::RegistryNode &    node,
        const std::string &         path = ""
        );

    //@}

public:

    /** @name Path manipulation methods */
    //@{

    /**
     * Searches for the RegistryNode specified by full name (path + plain
     * name) of a variable that might be contained within. Note that the
     * result doesn't depend on whether the variable actually exists.
     *
     * @param fullname Fully-qualified variable or symbolic link name.
     * @exception ExRegistryNodeNotFound The node doesn't exist.
     * @returns %Reference to the registry node that would contain or
     *          contains the variable.
     */
    RegistryNode * find_node_for_variable
        (
        const std::string & fullname
        ) const;

    /**
     * Searches for the RegistryNode specified by its path.
     *
     * @param path Path of the desired node. Contains no variable or alias name.
     * @exception ExRegistryNodeNotFound The node doesn't exist.
     * @return %Reference to the registry node specified by the given path.
     */
    RegistryNode * find_node_for_path
        (
        const std::string & path
        ) const;

    /**
     * Creates all of the nodes in the specified path. Makes no changes for
     * nodes that already exist.
     *
     * @param path Path of nodes to be created in the registry.
     * @return The pointer to the lowest-level RegistryNode object created.
     * @note  In case of lack of memory the path may stay partially constructed
     *        after the exception is thrown.
     */
    RegistryNode * create_node_path
        (
        const std::string & path
        );

public:

    /**
     * Wrenches the first component (node) of the path in the registry.
     *
     * @param path Input:  the original path; must be syntactically correct.
     *             Output: the path without the first node and following '/'.
     * @param node Output: The first node of the path. Null if the path is empty
     *             or the name of some variable or alias.
     * @note There won't be done any change to the registry structure.
     */
    static void split_node_path
        (
        std::string & path,
        std::string & node
        );

    /**
     * Splits the fully-qualified registry name to the path and the
     * variable name.
     *
     * @param path Input:  Syntactically correct fully-qualified variable name.
     *             Output: The path to the variable.
     * @param variable Output: The wrenched name of the variable.
     */
    static void split_node_path_rev
        (
        std::string & path,
        std::string & variable
        );

    //@}

public:

    /** @name Iterator related methods */
    //@{

    /**
     * Returns iterator pointing to the first variable in
     * the enumeration of variables matching @c pattern
     * that are stored in the node identified by @c path.
     */
    Iterator begin
        (
        const std::string & path,
        const std::string & pattern = "*"
        );

    /**
     * Returns iterator pointing 'beyond' the end of enumeration of variables.
     */
    Iterator end();

    /**
     * Returns const iterator pointing to the first variable in
     * the enumeration of variables matching @c pattern that are stored in
     * the node identified by @c path.
     */
    ConstIterator begin
        (
        const std::string & path,
        const std::string & pattern = "*"
        ) const;

    /**
     * Returns const iterator pointing 'beyond' the end of enumeration.
     */
    ConstIterator end() const;

    /**
     * Returns iterator pointing to the first subnode in the
     * enumeration of subnodes matching @c pattern that are stored
     * in the node identified by @c path.
     */
    SubnodeIterator begin_subnodes
        (
        const std::string & path,
        const std::string & pattern = "*"
        );

    /**
     * Returns iterator pointing 'beyond' the end of enumeration of subnodes.
     */
    SubnodeIterator end_subnodes();

    /**
     * Returns const iterator pointing to the first subnode in the
     * enumeration of subnodes matching @c pattern that are stored
     * in the node identified by @c path.
     */
    ConstSubnodeIterator begin_subnodes
        (
        const std::string & path,
        const std::string & pattern = "*"
        ) const;

    /**
     * Returns iterator pointing 'beyond' the end of enumeration of subnodes.
     */
    ConstSubnodeIterator end_subnodes() const;

    //@}

protected:

    /** @name Configuration files processing implementation methods */
    //@{

    /**
     * Removes name of file from @a filename, concludes the path with '/'.
     * In case the @c filename doesn't contain a path, @c actual_loading_path
     * will remain empty. Returns the resultant string.
     *
     * @param filename %File name with path.
     * @return Plain name of the file.
     */
    std::string get_path_from_filename
        (
        std::string         filename
        );

    /**
     * Processes one line of configuration file.
     *
     * @param line
     * @param line_number Number of the line, one-based.
     * @param node      All the stuff that is about to be loaded will
     *                  be placed in a subtree of registry specified
     *                  by this node.
     * @param loading_path Path to the configuration file currently
     *      being loaded. Needed because of !include directive.
     * @exception ExConfigFileSyntaxError
     */
    void process_line_of_config_file
        (
        std::string &       line,
        long                line_number,
        RegistryNode &      node,
        const std::string & loading_path = ""
        );

    /**
     * Processes one line of configuration file containing include directive.
     * @param line  The particular line of config file.
     * @param line_number Number of the line.
     * @param node Actual registry node - the following variables definitions
     *      will be inserted relatively to this node.
     * @param pos Position where '!include' clause should begin.
     * @param loading_path Path to the configuration file currently
     *      being loaded. Needed because of !include directive.
     */
    void process_include
        ( 
        std::string &           line, 
        long                    line_number, 
        RegistryNode &          node,
        std::string::size_type  pos,
        const std::string &     loading_path = ""
        );

    /**
     * Processes variable value from configuration file line, i.e.
     * 
     * @param var_name  Name of the variable.
     * @param var_value Value of the variable (to be processed).
     * @param var_type  Type of the variable.
     * @param is_alias  true if the parameters define a new alias.
     * @param node      All the stuff that is about to be loaded will
     *                  be placed in a subtree of registry specified
     *                  by this node.
     * @param line      The specific line of the configuration file.
     * @param filename  Name of file being processed (relatively to the
     *                  actual directory).
     * @param line_number Number of the line, one-based.
     * @exception ExConfigFileSyntaxError
     */
    void process_variable_value_string
        (
        std::string &               var_name,
        std::string &               var_value,
        ConfigVariable::CVarType    var_type,
        bool                        is_alias,
        RegistryNode &              node,
        const std::string &         line,
        const std::string &         filename,
        long                        line_number
        );

    /**
     * Checks the syntax of a node, variable or alias name.
     *
     * @param name Name of a node, variable or alias.
     * @exception ExWrongSyntax
     */
    static void check_name_syntax
        (
        const std::string & name
        );

    /**
     * Checks the syntax of a node, variable or alias name. If the
     * syntax is wrong, @a ExConfigFileSyntaxError exception is thrown
     * with the message formed using the @a line_number and @a filename.
     *
     * @param name Name of a node, variable or alias.
     * @param filename Name of the file being processed.
     * @param line_number Number of the line being processed.
     * @exception ExConfigFileSyntaxError
     */
    static void check_name_syntax
        (
        const std::string & name,
        const std::string & filename,
        long                line_number
        );

    /**
     * Checks the syntax of a path in registry. The path can be given
     * with or without the name of variable.
     *
     * @param path  Path in the registry.
     * @exception ExWrongSyntax
     */
    static void check_path_syntax
        (
        const std::string & path
        );

    /**
     * Checks the syntax of a path in registry. The path can be given
     * with or without the name of variable. If the syntax is wrong,
     * @a ExConfigFileSyntaxError exception is thrown with the message
     * formed using the @a line_number and @a filename.
     *
     * @param path  Path in the registry.
     * @param filename Name of the file being processed.
     * @param line_number Line number in the configuration file.
     * @exception ExConfigFileSyntaxError
     */
    static void check_path_syntax
        (
        const std::string & path,
        const std::string & filename,
        long                line_number
        );

    //@}

protected:

    RegistryNode *  loading_context_node;
        /**< @link RegistryNode Registry node @endlink relative to which 
             new variables are actualy
             loaded from the registry from configuration file or from
             a stream. Even this node itself is relative to a node 
             that has been user-defined as the root of all data currently 
             being loaded. */

protected:

    RegistryNode *  root_node;
        /**< Root node of the registry hierarchy. */

public:

    typedef std::map<std::string, ConfigVariable *> VariableList;
    typedef std::map<std::string, std::string>      LinkList;
    typedef std::map<std::string, RegistryNode *>   SubnodeList;


    /**
     * Represents a node in the hierarchy of the 
     * @link Registry registry @endlink.
     * Includes a list of variables, aliases and subdirectories.
     *
     * @ingroup registry
     */
    class MASSIV_EXPORTED RegistryNode
        {
        friend class Registry;
    public:

        /**
         * Constructor
         */
        RegistryNode
            (
            Registry *  registry
            ) : registry_pointer( registry )
            {
            }

        /**
         * Destructor
         */
        ~RegistryNode();

    public:

        /**
         * Creates a new variable in this registry node.
         *
         * @param name  Plain variable name; without path, even relative.
         * @param type  Type of the variable, see CVarType.
         * @param default_value Default (and initial) value.
         * @exception ExCVarTypeMismasch The variable of the same name
         *            already exists, but has different type.
         * @exception ExCVarDuplicateName The symbolic link of the
         *            same name already exists.
         * @return %Reference to the inserted ConfigVariable.
         */
        ConfigVariable & create_variable
            (
            const std::string &         name,
            ConfigVariable::CVarType    type,
            const std::string &         default_value
            );

        /**
         * Creates a new variable in the registry node specified by its path.
         *
         * @param name  Plain variable name, without path.
         * @param path  Path of the variable in the registry taken
         *              <em>relatively to this registry node</em>
         * @param type  Type of the variable, see CVarType.
         * @param default_value Default (and initial) value.
         * @exception ExCVarTypeMismasch The variable of the same name
         *            already exists, but has different type.
         * @exception ExCVarDuplicateName The symbolic link of the
         *            same name already exists.
         * @return %Reference to the inserted ConfigVariable.
         */
        ConfigVariable & create_variable
            (
            const std::string &         name,
            const std::string &         path,
            ConfigVariable::CVarType    type,
            const std::string &         default_value
            );

        /**
         * Creates symbolic link to existing variable. The variable may range in
         * any of existing nodes of the registry.
         *
         * @param alias   Plain symbolic link name, without path
         * @param link_to Link target path/name relative to this node. The
         *                target may be either a variable name or another
         *                alias name. @a link_to value is restricted the same
         *                way as alias is.
         * @param check_target If true, the routine checks if the target is defined.
         * @exception ExCVarDuplicateName The name already exists in the registry.
         * @exception ExCVarNotFound      A target of the alias not defined (and
         *                @a check_target is @c true.
         * @exception ExRegistryNodeNotFound
         * @note If the alias of the specified name already exists in the registry,
         *      but with the same target (@a link_to), no exception will be thrown
         *      (except that the target was not found and @a check_target was
         *      true) and the method will be considered a void operation.
         *
         * @see Registry::create_symbolic_link()
         */
        void create_symbolic_link
            (
            const std::string & alias,
            const std::string & link_to,
            bool                check_target
            );

        /**
         * Creates symbolic link to existing variable.
         *
         * @param aliasname   Plain symbolic link name, without path.
         * @param aliaspath   Path of the alias name taken 
         *                    <em>relatively to this registry node</em>.
         * @param link_to     Fully-qualified link-target name. The target may 
         *                    be either a variable name or another alias name. 
         *                    @a link_to value is restricted the same way as 
         *                    alias is.
         * @param check_target If true, the routine checks if the target is defined.
         * @exception ExCVarDuplicateName The name already exists in the registry.
         * @exception ExCVarNotFound      A target of the alias not defined (and
         *                    @a check_target is @c true.
         * @exception ExRegistryNodeNotFound
         * @note If the alias of the specified name already exists in the registry,
         *      but with the same target (@a link_to), no exception will be thrown
         *      (except that the target was not found and @a check_target was
         *      true) and the method will be considered a void operation.
         *
         * @see Registry::create_symbolic_link()
         */
        void create_symbolic_link
            (
            const std::string &     aliasname,
            const std::string &     aliaspath,
            const std::string &     link_to,
            bool                    check_target = true
            );

        /**
         * Creates subnode of this node.
         * Makes no changes to nodes that already exists; creating existing
         * item yields no error.
         *
         * @param name Name of the subnode.
         * @return %Pointer to the constructed subnode object.
         */
        RegistryNode * create_subnode
            (
            const std::string & name
            );

        /**
         * Creates all of the nodes in the specified path under this registry 
         * node. Makes no changes for nodes that already exist.
         *
         * @param path Path of nodes to be created in the registry.
         * @return %Reference to the lowest-level RegistryNode object created.
         * @note  In case of lack of memory the path may stay partially 
         *        constructed after the exception is thrown.
         */
        RegistryNode * create_node_path
            (
            const std::string & path
            );

        /**
         * Looks for the pointer to the specified subnode.
         *
         * @param name Name of the subnode.
         * @return The pointer to the subnode object. @c NULL if the subnode 
         *         was not found.
         */
        RegistryNode * find_subnode
            (
            const std::string & name
            ) const;

        /**
         * Accesses variable by name.
         *
         * @param name Plain variable name, without path. No alias name!!!
         * @exception ExCVarNotFound The variable was not found.
         * @exception ExAliasNotFound The name of the alias not found.
         * @exception ExCVarAliasLoop Too long chain of symbolic links found
         *            (this prevents the alias loop). The chain length 
         *            constraint is given by @c MAX_ALIAS_DEPTH constant.
         * @exception ExRegistryNodeNotFound
         * @return %Reference to the variable if found; otherwise throws 
         *      @c ExCVarNotFound.
         * @note Symbolic links are not resolved in this routine!!!
         */
        ConfigVariable & access_variable
            (
            const std::string & name
            );

    protected:

        Registry *          registry_pointer;
            /**< %Pointer to the registry object responsible for this node. */

    public:

        SubnodeList         subnodes;
            /**< %List of subnodes (subdirectories of a node in the registry). */

        VariableList        variables;
            /**< %List of all variables in this node of registry. */

        LinkList            links;
            /**< %List of all aliases in this node of registry. */

        };  // class RegistryNode


    /**
     * Represents a pointer into the 
     * @link Registry registry @endlink. Because it isn't possible to
     * delete from registry, this pointer is valid even after arbitrary
     * number of operations over the registry.
     *
     * SearchContext should be used to keep context for example during
     * search through the registry.
     *
     * @ingroup registry
     */
    class MASSIV_EXPORTED SearchContext
        {
        friend class Registry;
        friend class ConstIterator;

    public:

        /**
         * Determines whether the search context actually points to
         * a variable or an alias.
         */
        enum SearchContextType
            {
            RRT_VOID,
                /**< The pointer actually doesn't point to anything. */
            RRT_NOT_FOUND,
                /**< No variable nor alias was found during the last search. */
            RRT_VARIABLE,
                /**< Points to variable. */
            RRT_ALIAS,
                /**< Points to alias. */
            };

    public:

        /**
         * Constructor.
         *
         * @param registry_pointer %Pointer to the registry object
         *        bound with this context.
         */
        SearchContext
            (
            Registry *  registry_pointer
            ) : cvar( NULL ),
                registry_pointer( registry_pointer ),
                regnode( NULL ),
                type( RRT_VOID )
            {
            }

        /**
         * Constructor.
         *
         * @param registry_pointer %Pointer to the registry object
         *        bound with this context.
         * @param regnode %Pointer to registry node to be pointed to.
         * @param name    Fully-qualified name of the variable.
         */
        SearchContext
            (
            Registry *              registry_pointer,
            RegistryNode &          regnode,
            const std::string &     name
            );

        /**
         * Changes the context to the first name in the node.
         *
         * @param regnode %Pointer to register node to be pointed at.
         * @exception ExRegistryNodeEmpty
         */
        void change_context
            (
            RegistryNode &          regnode
            );

        /**
         * Changes the context to another name in the registry.
         *
         * @param regnode %Pointer to register node to be pointed at.
         * @param name    Fully-qualified name of the variable.
         * @exception ExRegistryNodeEmpty
         */
        void change_context
            (
            RegistryNode &          regnode,
            const std::string &     name
            );

        /** @name Methods for obtaining information about the referenced entity */
        //@{

        /**
         * @return Name of the variable/alias referenced by this context.
         */
        std::string get_name() const;

        /**
         * @return Path to the variable/alias referenced by this context.
         * @note The path is valid only when the context is used by function
         *       Registry::find_variable. Don't call this routine after
         *       this context has been processed with
         *       Registry::search_registry_node()
         */
        std::string get_path() const;

        /**
         * @return %Node referenced by this context.
         */
        RegistryNode * get_node() const;

        /**
         * @return %Reference to the referenced config variable. Symbolic
         *         links are resolved first.
         */
        ConfigVariable & get_variable();

        /**
         * @return The target of the alias. If the referenced entity
         *         is not an alias, an empty string is returned.
         */
        std::string get_alias_target() const;

        /**
         * @return Type of the search context.
         */
        SearchContextType get_type() const
            {
            return type;
            }

        //@}

    protected:

        /** @name Methods for traversing the registry */
        //@{

        /**
         * Prepares the context in the case when search context doesn't
         * include full path.
         * @param root_node %Node to complete path to.
         */
        void prepare_context
            (
            RegistryNode * root_node
            );

        /**
         * Changes the context to reference the next registry
         * node (the ordering of nodes in the registry tree is the
         * pre-order). Changes just the regnode and it_subnodes members
         * (and the stacks)
         * and thus the SearchContext class may stay in an inconsistent
         * state after a call to this method.
         *
         * @return false if the traverse has ended, otherwise true.
         */
        bool traverse_nodes();

        /**
         * Tries to change the context to reference the next alias
         * or variable within the actual node.
         * Aliase names are considered before variables.
         *
         * @return true if successful, otherwise false.
         * @note Sets type member to RRT_NOT_FOUND if there is no
         *       alias nor variable available in the node.
         */
        bool next_in_node();

        //@}

    protected:

        ConfigVariable *        cvar;
            /**< %Pointer to the referenced config variable. */

        Registry *              registry_pointer;
            /**< %Pointer to the registry object this context is bound with. */

    protected:

        RegistryNode *          regnode;
            /**< %Pointer to the referenced RegistryNode. */

        VariableList::iterator  it_vars;
            /**< %Iterator through a list of variables of referenced node. */

        LinkList::iterator      it_links;
            /**< %Iterator through a list of links of referenced node. */

        SubnodeList::iterator   it_subnodes;
            /**< %Iterator through a list of subnodes; it points to a subnode
                 that has to be searched through as the next. */

        SearchContextType       type;
            /**< Determines whether the registry pointer points to
                 a variable or a link. */

    protected:

        std::stack< SubnodeList::iterator >   it_subnodes_stack;
            /**< A helper stack for the registry tree traversing algorithm. */

        std::stack< RegistryNode * >          regnodes_stack;
            /**< Keeps the stack of regnodes for returning up while traversing
                 the registry tree. */

        std::list< std::string >              node_names_stack;
            /**< Stack of nodes names. */

        }; // class SearchContext

public:

    /**
     * Const iterator enables enumerating variables stored in the 
     * @link Registry registry @endlink in a given node.
     * Uses @c open_registry_node() & @c search_registry_node().
     *
     * @ingroup registry
     */
    class MASSIV_EXPORTED ConstIterator
        {
    public:

        /** Default constructor. Corresponds to @c end(). */
        ConstIterator();

        /** 
         * Constructor. 
         * 
         * Corresponds to @c begin(). 
         * 
         * @param registry A pointer to the @link Registry registry @endlink 
         *      object this iterator is bound with.
         * @param path The path in the @link Registry registry @endlink
         *      that specifies a @link RegistryNode registry node @endlink
         *      to bind this iterator with.
         * @param pattern The iterator will ignore plain names that don't 
         *      match this wildcards pattern (@c * and @c ? can be used
         *      as a special symbols).
         */
        ConstIterator
            (
            Registry *          registry,
            const std::string & path,
            const std::string & pattern
            );

        /** 
         * Moves iterator (preincrement) to the next available variable
         * or sets it to end() if no more variables are available.
         *
         * @return %Reference to the moved iterator.
         */
        ConstIterator & operator++();

        /** 
         * Moves iterator (postincrement) to the next availabla variable
         * or sets it to end() if no more variables are available.
         *
         * @return The moved iterator.
         */
        ConstIterator operator++( int );

        /** 
         * Dereferences iterator. 
         * @return %Reference to the dereferenced config variable.
         */
        const ConfigVariable & operator*();

        /** 
         * Dereferences iterator. 
         * @return %Pointer to the dereferenced config variable.
         */
        const ConfigVariable * operator->();

        /** 
         * Gets the name of the variable pointed to by this iterator.
         * @returns The variable name. 
         */
        std::string get_name() const;

        /** 
         * Compares iterators. 
         * 
         * @param it The iterator to be compared with @c this.
         * @return true if both iterators are equal, false otherwise.
         */
        bool operator==
            (
            const ConstIterator & it
            ) const;

        /** 
         * Compares iterators. 
         *
         * @param it The iterator to be compared with @c this.
         * @return true if both iterators are not equal, false otherwise.
         */
        bool operator!=
            (
            const ConstIterator & it
            ) const;

        /** 
         * Assigns iterators. 
         *
         * @param it The iterator to be assigned into @c this.
         * @return %Reference to @c this.
         */
        ConstIterator & operator=
            (
            const ConstIterator & it
            );

    protected:

        SearchContext   context;
            /**< Search context. */
        std::string     pattern;
            /**< Search pattern (the iterator will ignore all plain names
                 that don't match this wildcards pattern). */
        }; // ConstIterator

    /**
     * %Iterator enables enumerating variables stored in the 
     * @link Registry registry @endlink in a given 
     * @link RegistryNode node @endlink.
     *
     * @ingroup registry
     */
    class MASSIV_EXPORTED Iterator : public ConstIterator
        {
    private:

        typedef ConstIterator   Super;

    public:

        /** Default constructor. Corresponds to @c end(). */
        Iterator();

        /** 
         * Constructor. 
         * 
         * Corresponds to @c begin(). 
         * 
         * @param registry A pointer to the @link Registry registry @endlink 
         *      object this iterator is bound with.
         * @param path The path in the @link Registry registry @endlink
         *      that specifies a @link RegistryNode registry node @endlink
         *      to bind this iterator with.
         * @param pattern The iterator will ignore plain names that don't 
         *      match this wildcards pattern (@c * and @c ? can be used
         *      as a special symbols).        
         */
        Iterator
            (
            Registry *          registry,
            const std::string & path,
            const std::string & pattern
            );

        /** 
         * Moves iterator (preincrement) to the next available variable
         * or sets it to end() if no more variables are available.
         *
         * @return %Reference to the moved iterator.
         */
        Iterator & operator++();

        /** 
         * Moves iterator (postincrement) to the next available variable
         * or sets it to end() if no more variables are available.
         *
         * @return The moved iterator.
         */
        Iterator operator++( int );

        /** 
         * Dereferences iterator. 
         * @return %Reference to the dereferenced config variable.
         */
        ConfigVariable & operator*();

        /** 
         * Dereferences iterator. 
         * @return %Pointer to the dereferenced config variable.
         */
        ConfigVariable * operator->();

        /** 
         * Assigns iterators. 
         * 
         * @param it The iterator to be assigned into @c this.
         * @return %Reference to @c this.
         */
        Iterator & operator=
            (
            const Iterator & it
            );

        }; // Iterator

    /**
     * Constant subnode iterator enables enumerating subnodes of a given
     * @link RegistryNode node @endlink.
     * 
     * @ingroup registry
     */
    class MASSIV_EXPORTED ConstSubnodeIterator
        {
    public:

        /** Default constructor. Corresponds to @c end_subnodes(). */
        ConstSubnodeIterator();

        /** 
         * Constructor. 
         *
         * Corresponds to @c begin_subnodes(). 
         *
         * @param registry A pointer to the @link Registry registry @endlink 
         *      object this iterator is bound with.
         * @param path The path in the @link Registry registry @endlink
         *      that specifies a @link RegistryNode registry node @endlink
         *      to bind this iterator with.
         * @param pattern The iterator will ignore plain subnode names that 
         *      don't match this wildcards pattern (@c * and @c ? can be used
         *      as a special symbols).        
         */
        ConstSubnodeIterator
            (
            Registry *          registry,
            const std::string & path,
            const std::string & pattern
            );

        /** 
         * Moves iterator (preincrement) to the next available subnode
         * or sets it to end_subnodes() if no more variables are available.
         *
         * @return %Reference to the moved iterator.
         */
        ConstSubnodeIterator & operator++();

        /** 
         * Moves iterator (postincrement) to the next available subnode
         * or sets it to end_subnodes() if no more variables are available.
         *
         * @return The moved iterator.
         */
        ConstSubnodeIterator operator++( int );

        /** 
         * Dereferences iterator. 
         * @return Name of the referenced subnode.
         */
        const std::string & operator*();

        /** 
         * Dereferences iterator.
         * @return %Pointer to string containing name of the referenced subnode.
         */
        const std::string * operator->();

        /**
         * Gets the referenced subnode.
         * @return %Pointer to the referenced subnode.
         */
        const RegistryNode * get_subnode() const;

        /** 
         * Compares iterators. 
         *
         * @param it The iterator to be compared with @c this one.
         * @return true if both iterators are equal, false otherwise.
         */
        bool operator==
            (
            const ConstSubnodeIterator & it
            ) const;

        /** 
         * Compares iterators. 
         * 
         * @param it The iterator to be compared with @c this one.
         * @return true if both iterators are not equal, false otherwise.
         */
        bool operator!=
            (
            const ConstSubnodeIterator & it
            ) const;

        /** 
         * Assigns iterator. 
         *
         * @param it The iterator to be assigned into @c this.
         * @return %Reference to @c this.
         */
        ConstSubnodeIterator & operator=
            (
            const ConstSubnodeIterator & it
            );

    protected:

        Registry *              registry;
            /**< %Pointer to the registry object bound with this iterator. */

        RegistryNode *          node;
            /**< Referenced registry node. */

        SubnodeList::iterator   it_sub;
            /**< %Iterator over subnodes in a specified node. */

        std::string             pattern;
            /**< Pattern for filtering subnodes. Subnodes that don't match
                 this patter won't be enumerated. */
        
        }; // ConstSubnodeIterator

    /**
     * %Subnode iterator enables enumerating subnodes of a given 
     * @link RegistryNode node @endlink.
     *
     * @ingroup registry
     */
    class MASSIV_EXPORTED SubnodeIterator : public ConstSubnodeIterator
        {
    private:

        typedef ConstSubnodeIterator   Super;

    public:

        /** Default constructor. Corresponds to @c end_subnodes(). */
        SubnodeIterator();

        /** 
         * Constructor. 
         *
         * Corresponds to @c begin_subnodes(). 
         *
         * @param registry A pointer to the @link Registry registry @endlink 
         *      object this iterator is bound with.
         * @param path The path in the @link Registry registry @endlink
         *      that specifies a @link RegistryNode registry node @endlink
         *      to bind this iterator with.
         * @param pattern The iterator will ignore plain subnode names that 
         *      don't match this wildcards pattern (@c * and @c ? can be used
         *      as a special symbols).        
         */
        SubnodeIterator
            (
            Registry *          registry,
            const std::string & path,
            const std::string & pattern
            );

        /** 
         * Moves iterator (preincrement) to the next available subnode
         * or sets it to end_subnodes() if no more variables are available.
         *
         * @return %Reference to the moved iterator.
         */
        SubnodeIterator & operator++();

        /** 
         * Moves iterator (postincrement) to the next available subnode
         * or sets it to end_subnodes() if no more variables are available.
         *
         * @return The move iterator.
         */
        SubnodeIterator operator++( int );

        /**
         * Dereferences iterator.
         * @return Name of the referenced subnode.
         */
        std::string operator*();

        /**
         * Dereferences iterator.
         * @return %Pointer to string containing name of the referenced subnode.
         */
        const std::string * operator->();

        /**
         * Gets the referenced subnode.
         * @return %Pointer to the referenced subnode.
         */
        RegistryNode * get_subnode();

        /** 
         * Assigns iterator. 
         *
         * @param it The iterator to be assigned into @c this.
         * @return %Reference to @c this.
         */
        SubnodeIterator & operator=
            (
            const SubnodeIterator & it
            );

        }; // SubnodeIterator

    }; // class Registry

/** @} */ // addtogroup registry


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_REGISTRY_H
