/** @file statistics.cpp
 * Statistics make use of the Registry for storing values that are being
 * watched and evaluated during the system is running.
 * @see statistics.h
 */

/*
 * begin        : Mon Nov 11 2002
 * copyright    : (C) 2002, 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/global.h"
#include "utility/debug.h"
#include "utility/hash.h"
#include "registry/statistics.h"
#include "registry/conversion.h"
#include <ostream>
#include <sstream>


namespace Massiv {
namespace Core {


// Implementation of StatsAndSettingsBase
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

StatsAndSettingsBase::StatsAndSettingsBase()
    : node( NULL )
    {
    }


void StatsAndSettingsBase::init
    (
    const std::string & registry_path
    )
    {
    Registry & registry = Global::registry();

    hope( !node );

    path = registry_path;
    node = registry.create_node_path( path );

    hope( node );
    }


StatsAndSettingsBase::Iterator StatsAndSettingsBase::begin()
    {
    Registry & registry = Global::registry();
    return registry.begin( path );
    }


StatsAndSettingsBase::Iterator StatsAndSettingsBase::end()
    {
    Registry & registry = Global::registry();
    return registry.end();
    }


StatsAndSettingsBase::ConstIterator StatsAndSettingsBase::begin() const
    {
    Registry & registry = Global::registry();
    return registry.begin( path );
    }


StatsAndSettingsBase::ConstIterator StatsAndSettingsBase::end() const
    {
    Registry & registry = Global::registry();
    return registry.end();
    }


void StatsAndSettingsBase::reset()
    {
    Iterator it( begin() );

    while( it != end() )
        {
        it->revert_to_default();
        it++;
        }
    }


std::ostream & operator<<
    (
    std::ostream &                  stream,
    const StatsAndSettingsBase &    s
    )
    {
    StatsAndSettingsBase::ConstIterator it( s.begin() );

    while( it != s.end() )
        {
        stream << it.get_name() << " = " << it->read_string() << '\n';
        it++;
        }

    stream.flush();
    return stream;
    }


const std::string & StatsAndSettingsBase::get_registry_path() const
    {
    return path;
    }


// Implementation of Statistics::MeasureSystemTime
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Statistics::MeasureSystemTime::MeasureSystemTime
    (
    Float & time_accumulator,
    bool    start_now
    ) : accumulator( time_accumulator ),
        start_time( INVALID_TIME )
    {
    if( start_now )
        {
        start();
        }
    }


Statistics::MeasureSystemTime::~MeasureSystemTime()
    {
    stop();
    }


void Statistics::MeasureSystemTime::start()
    {
    start_time = Global::system_time();
    }


void Statistics::MeasureSystemTime::stop()
    {
    if( start_time != INVALID_TIME )
        {
        Float32 delta;

        delta = static_cast<Float32>( ( Global::system_time() - start_time ).get_value() );
        accumulator += delta;

        start_time = INVALID_TIME;
        }
    }


// Definition of Statistics::My
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * Class that holds information global to all Statistics classes.
 * Always exists as a singleton object referenced by @c Statistics::my.
 */
class Statistics::My
    {
public:

    /** Constructor. */
    My() : unique_counter( 0 )
        {
        }

public:
        
    typedef ext::hash_multimap<std::string, std::string> UnusedStats;

    UnusedStats unused_statistics;
        /**< List of unused statistics. Multimap that associates statistics
             original paths with a generated ones. */
    int         unique_counter;
        /**< Unique counter. It is used for generating of unique names
             for statistics nodes. */

    }; // Statistics::My


// Implementation of Statistics
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Statistics::My * Statistics::my = NULL;


Statistics::Statistics
    (
    const char *        registry_path,
    Status::Facility    facility,
    OpenMode            mode
    ) : facility( facility ),
        mode( mode ),
        original_path( registry_path )
    {
    std::ostringstream stream;
    std::string generated_path;

    // Initialize.

    if( !my )
        {
        my = new My();
        }
    
    stream << "Statistics/" << registry_path;

    // Create registry node.

    if( mode == OPEN_NORMAL || mode == OPEN_OVERWRITE )
        {
        generated_path = stream.str();
        recycled = false;
        }
    else
    if( mode == OPEN_UNIQUE )
        {
        // Generate statistics name.

        stream << "$" << ++my->unique_counter;
        generated_path = stream.str();

        recycled = false;
        }
    else
    if( mode == OPEN_RECYCLED )
        {
        // Try to recycle an existing statistics stored in the registry.

        My::UnusedStats::iterator it = 
            my->unused_statistics.find( original_path );
        
        if( it != my->unused_statistics.end() )
            {
            // Recycle first statistics whose original path equals
            // to original_path.

            generated_path = it->second;
            my->unused_statistics.erase( it );
            
            recycled = true;
            }
        else
            {
            // No appropriate statistics to recycle.
            // Generate a new statistics name.

            stream << "$recycleable" << ++my->unique_counter;
            generated_path = stream.str();

            recycled = false;
            }
        }
    else
        {
        panic( "Bad openmode." );
        }

    init( generated_path );
    }


Statistics::~Statistics()
    {
    if( mode == OPEN_RECYCLED )
        {
        // Mark registry entries as unused and ready to be recycled.
        
        my->unused_statistics.insert
            ( My::UnusedStats::value_type( original_path, path ) );
        }
    }


Statistics::Integer & Statistics::create_integer
    (
    const char *    name,
    Int32           default_value
    )
    {
    if( !recycled )    
        {
        try
            {
            return node->create_variable
                (
                name,
                ConfigVariable::CVAR_TYPE_INTEGER,
                ManagedConversion<std::string, Int32>( default_value )
                ).as_integer();
            }
        catch( ExCVarDuplicateName & )
            {
            if( mode == OPEN_OVERWRITE )
                {
                return node->access_variable( name ).as_integer();
                }
                
            throw;
            }
        }
    else
        {
        return node->access_variable( name ).as_integer();
        }
    }


Statistics::LargeInteger & Statistics::create_large_integer
    (
    const char *    name,
    Int64           default_value
    )
    {
    if( !recycled )
        {
        try
            {
            return node->create_variable
                (
                name,
                ConfigVariable::CVAR_TYPE_LARGE_INTEGER,
                ManagedConversion<std::string, Int64>( default_value )
                ).as_large_integer();
            }
        catch( ExCVarDuplicateName & )
            {
            if( mode == OPEN_OVERWRITE )
                {
                return node->access_variable( name ).as_large_integer();
                }
                
            throw;
            }
        }
    else
        {
        return node->access_variable( name ).as_large_integer();
        }
    }


Statistics::Float & Statistics::create_float
    (
    const char *    name,
    Float32         default_value
    )
    {
    if( !recycled )
        {
        try
            {
            return node->create_variable
                (
                name,
                ConfigVariable::CVAR_TYPE_FLOAT,
                ManagedConversion<std::string, Float32>( default_value )
                ).as_float();
            }
        catch( ExCVarDuplicateName & )
            {
            if( mode == OPEN_OVERWRITE )
                {
                return node->access_variable( name ).as_float();
                }
                
            throw;
            }
        }
    else
        {
        return node->access_variable( name ).as_float();
        }
    }


Statistics::Boolean & Statistics::create_boolean
    (
    const char *    name,
    bool            default_value
    )
    {
    if( !recycled )    
        {
        try
            {
            return node->create_variable
                (
                name,
                ConfigVariable::CVAR_TYPE_BOOLEAN,
                ManagedConversion<std::string, bool>( default_value )
                ).as_boolean();
            }
        catch( ExCVarDuplicateName & )
            {
            if( mode == OPEN_OVERWRITE )
                {
                return node->access_variable( name ).as_boolean();
                }
                
            throw;
            }
        }
    else
        {
        return node->access_variable( name ).as_boolean();
        }
    }


Statistics::String & Statistics::create_string
    (
    const char *    name,
    const char *    default_value
    )
    {
    if( !recycled )
        {    
        try
            {
            return node->create_variable
                (
                name,
                ConfigVariable::CVAR_TYPE_STRING,
                default_value
                ).as_string();
            }
        catch( ExCVarDuplicateName & )
            {
            if( mode == OPEN_OVERWRITE )
                {
                return node->access_variable( name ).as_string();
                }
                
            throw;
            }
        }
    else
        {
        return node->access_variable( name ).as_string();
        }
    }


void Statistics::log() const
    {
    std::ostringstream ss;

    ss  << "Statistics:\n"
        << *this;

    Global::log
        (
        STATUS_INFO
            (
            facility,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );
    }


// Implementation of Settings
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

Settings::Settings
    (
    const char *    registry_path
    )
    {
    init( std::string( "Settings/" ) + registry_path );
    }


Settings::Integer & Settings::access_integer
    (
    const char *    name,
    Int32           default_value
    )
    {
    Integer * integer;

    try
        {
        integer = &node->access_variable( name ).as_integer();
        }
    catch( ExCVarNotFound & )
        {
        integer = &node->create_variable
            (
            name,
            ConfigVariable::CVAR_TYPE_INTEGER,
            ManagedConversion<std::string, int>( default_value )
            ).as_integer();
        }

    return *integer;
    }


Settings::LargeInteger & Settings::access_large_integer
    (
    const char *    name,
    Int64           default_value
    )
    {
    LargeInteger * integer;

    try
        {
        integer = &node->access_variable( name ).as_large_integer();
        }
    catch( ExCVarNotFound & )
        {
        integer = &node->create_variable
            (
            name,
            ConfigVariable::CVAR_TYPE_LARGE_INTEGER,
            ManagedConversion<std::string, Int64>( default_value )
            ).as_large_integer();
        }

    return *integer;
    }


Settings::Float & Settings::access_float
    (
    const char *    name,
    Float32         default_value
    )
    {
    Float * f;

    try
        {
        f = &node->access_variable( name ).as_float();
        }
    catch( ExCVarNotFound & )
        {
        f = &node->create_variable
            (
            name,
            ConfigVariable::CVAR_TYPE_FLOAT,
            ManagedConversion<std::string, float>( default_value )
            ).as_float();
        }

    return *f;
    }


Settings::Boolean & Settings::access_boolean
    (
    const char *    name,
    bool            default_value
    )
    {
    Boolean * boolean;

    try
        {
        boolean = &node->access_variable( name ).as_boolean();
        }
    catch( ExCVarNotFound & )
        {
        boolean = &node->create_variable
            (
            name,
            ConfigVariable::CVAR_TYPE_BOOLEAN,
            ManagedConversion<std::string, bool>( default_value )
            ).as_boolean();
        }

    return *boolean;
    }


Settings::String & Settings::access_string
    (
    const char *    name,
    const char *    default_value
    )
    {
    String * s;

    try
        {
        s = &node->access_variable( name ).as_string();
        }
    catch( ExCVarNotFound & )
        {
        s = &node->create_variable
            (
            name,
            ConfigVariable::CVAR_TYPE_STRING,
            default_value
            ).as_string();
        }

    return *s;
    }


} // namespace Massiv::Core
} // namespace Massiv
