/** @file statistics.h
 * Statistics and settings make usage of the Registry more simple and 
 * efficient.
 *
 * Statistics make use of the Registry for storing values that are being
 * watched and evaluated during the system is running.
 *
 * Settings can keep references to some variables in the Registry and thus
 * optimalize access to them (registry search mechanisms don't have to be
 * launched each time the variable is needed).
 * 
 */

/*
 * begin        : Mon Nov 11 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_STATISTICS_H
#define MASSIV_STATISTICS_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REGISTRY_H
#include "registry/registry.h"
#endif

#ifndef MASSIV_CONFIG_VARIABLE_H
#include "registry/config_variable.h"
#endif

#ifndef MASSIV_STATUS_H
#include "status/status.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#include <string>
#include <iosfwd>


namespace Massiv {
namespace Core {


/** @addtogroup registry
 * @{
 */

/**
 * Base class used by both @c Statistics and @c Settings.
 * 
 * Manages one of the registry nodes - all the variables within can
 * be accessed directly and thus more efficiently.
 *
 * Defines attribute iterators and various helper functions used
 * mainly for debugging purposes.
 */
class MASSIV_EXPORTED StatsAndSettingsBase
    {
public:

    typedef ConfigVariableInteger       Integer;
    typedef ConfigVariableLargeInteger  LargeInteger;
    typedef ConfigVariableFloat         Float;
    typedef ConfigVariableBoolean       Boolean;
    typedef ConfigVariableString        String;

    typedef Registry::Iterator          Iterator;
    typedef Registry::ConstIterator     ConstIterator;

protected:

    /**
     * Default constructor.
     * @c init() is supposed to be called after construction of the object.
     */
    StatsAndSettingsBase();

    /** 
     * Creates/opens registry node. Must be called after construction of
     * the object is complete.
     * @param registry_path Specifies the node.
     */
    void init
        (
        const std::string & registry_path
        );

public:

    /** @name Iterator related methods */
    //@{

    /** Returns iterator pointing to the first attribute of managed node. */
    Iterator begin();

    /** Returns iterator pointing to the end of attribute enumeration. */
    Iterator end();

    /** Returns const iterator pointing to the first attribute of managed node. */
    ConstIterator begin() const;

    /** Returns const iterator pointing to the end of attribute enumeration. */
    ConstIterator end() const;

    //@}

public:

    /** Resets all attributes contained within the managed node to default. */
    void reset();

    /** Returns the managed node path. */
    const std::string & get_registry_path() const;

protected:

    Registry::RegistryNode *    node;
        /**< %Registry node that holds the attributes. */
    std::string                 path;
        /**< %Registry path for the @c node. */

    }; // StatsAndSettingsBase

    
/**
 * Writes attributes and their contents to stream.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &                  stream,
    const StatsAndSettingsBase &    s
    );

    
/**
 * Base class for all statistics.
 * The class simplifies usage of registry which holds gathered
 * data. Statistics are stored in the registry under @c %Statistics node.
 *
 * @code
 *  // Every statistics gathered in the system should be
 *  // implemented according to this pattern.
 *
 *  /// This class gathers some statistics.
 *  class MyClass
 *      {
 *  public:
 *
 *      /// Statistics class represents link to registry.
 *      class MyStatistics : public Statistics
 *          {
 *      public:
 *          MyStatistics() :
 *              // MyStatistics will be accessible under
 *              // Statistics/MyStatistics node in the registry.
 *              Statistics( "MyStatistics" ),
 *              // Allocate attributes in the registry.
 *              my_integer( create_integer( "my_integer" ),
 *              my_float( create_float( "my_float" ),
 *              ...
 *              {
 *              }
 *
 *      public:
 *          // Attributes that will be stored in the registry.
 *          Integer &   my_integer;
 *          Float &     my_float;
 *          ...
 *          }
 *
 *      /// Returns read-only statistics.
 *      const MyStatistics & get_statistics() const
 *          {
 *          return statistics;
 *          }
 *
 *      /// Method that updates statistics.
 *      void function()
 *          {
 *          // MyClass writes directly to statistics attributes!
 *          // That ensures that current statistics state is visible
 *          // in the registry.
 *          // There is no need to "mirror" statistics attributes
 *          // to "native" C++ attributes and do manual mirroring / updating
 *          // of the attributes in the registry.
 *
 *          statistics.my_integer++;
 *          }
 *
 *      /// Some method that wills to measure system time delta.
 *      void compute()
 *          {
 *          Statistics::MeasureSystemTime measure( statistics.my_float );
 *          // Lengthy computation follows.
 *          ...
 *          // System time difference is now added to statistics.my_float.
 *          }
 *
 *  private:
 *
 *      MyStatistics statistics;
 *      };
 * @endcode
 */
class MASSIV_EXPORTED Statistics : public StatsAndSettingsBase
    {
private:

    class My;

public:

    /**
     * Helper class for measuring system time deltas.
     * The class measures time delta since construction of
     * the object until its destruction. Computed difference
     * is added to a @c %Float attribute stored in @c Statistics.
     * If one wants to temporarily turn time measuring off/on
     * @c stop() & @c start() can be called.
     */
    class MeasureSystemTime
        {
    private:

        enum Constants
            {
            INVALID_TIME = -1
                /**< Invalid @c start_time. */
            }; // Constants

    public:

        /**
         * Constructor.
         * @c start_now determines wheter @c start() will be called
         * from the constructor.
         */
        MeasureSystemTime
            (
            Float & time_accumulator,
            bool    start_now = true
            );

        /**
         * Destructor.
         * Calls @c stop().
         */
        ~MeasureSystemTime();

        /**
         * Starts measuring system time.
         */
        void start();

        /**
         * Stops measuring system time.
         * Adds the computed time difference from the time when @c start()
         * was called to the time accumulator.
         */
        void stop();

    private:

        Float &     accumulator;
            /**< Attribute the computed time delta will be added to. */
        SystemTime  start_time;
            /**< Initial system time. */

        }; // MeasureSystemTime

protected:

    /**
     * Statistics open mode.
     * See Statistics constructor.
     */
    enum OpenMode
        {
        OPEN_NORMAL,
            /**< Opens registry node specified by @c registry_path. */
        OPEN_UNIQUE,
            /**< Unique suffix is appended to @c registry_path. */
        OPEN_RECYCLED,
            /**< Extends @c OPEN_UNIQUE in that way that it tries to attach
                 Statistics object to an existing registry node that is not
                 used.
                 Statistics is unused if corresponding Statistics
                 class no longer exists but its entries still remain in the
                 registry. If no such node is found a new one will be
                 allocated.
                 Only statistics that were opened as @c OPEN_RECYCLED
                 can by recycled. */
        OPEN_OVERWRITE
            /**< Opens registry node as if @c OPEN_NORMAL was specified.
                 If statistics entries were already allocated they would
                 be overwritten. */
        }; // OpenMode

    /**
     * Initializes statistics.
     * Creates registry node that will hold the statistics.
     * @param registry_path The path in the registry where the specific
     *      statistics variables should be stored. It must be specified 
     *      relatively to the @c %Statistics node.
     */
    Statistics
        (
        const char *        registry_path,
        Status::Facility    facility = Status::FACILITY_IO,
        OpenMode            mode = OPEN_UNIQUE
        );

    /**
     * Destructor.
     */
    ~Statistics();

protected:

    /** @name Attribute creation methods */
    //@{

    /** Creates integer attribute. */
    Integer & create_integer
        (
        const char *    name,
        Int32           default_value = 0
        );

    /** Creates large integer attribute. */
    LargeInteger & create_large_integer
        (
        const char *    name,
        Int64           default_value = 0
        );

    /** Creates float attribute. */
    Float & create_float
        (
        const char *    name,
        Float32         default_value = 0
        );

    /** Creates boolean attribute. */
    Boolean & create_boolean
        (
        const char *    name,
        bool            default_value = false
        );

    /** Creates string attribute. */
    String & create_string
        (
        const char *    name,
        const char *    default_value = ""
        );

    //@}

public:

    /** 
     * Logs information about all statistic variables managed by this class.
     */
    void log() const;

private:

    Status::Facility    facility;
        /**< Facility bound with this statistics class. */
    OpenMode            mode;
        /**< Open mode. */
    std::string         original_path;
        /**< Original path. */
    bool                recycled;
        /**< Is this statistics a new one or a recycled one? */
    static My *         my;
        /**< Statistics context. */

    }; // Statistics

/**
 * Base class for various settings.
 * The class simplifies usage of registry which holds the settings.
 * Settings are stored in the registry under @c %Settings %node.
 * @c Settings class is intended to be used in the same sense
 * as @c Statistics class.
 */
class MASSIV_EXPORTED Settings : public StatsAndSettingsBase
    {
protected:

    /**
     * Opens registry node specified by @c registry_path
     * and accesses the settings.
     * @c registry_path must be specified relatively to the <tt>%Settings</tt>
     * registry node.
     */
    Settings
        (
        const char *    registry_path
        );

protected:

    /** @name Variable access methods */
    //@{

    /**
     * Accesses integer attribute.
     * If attribute is not defined in the registry then the attribute
     * will be created and initialized to @c default_value.
     * @param name Fully-qualified variable name.
     * @param default_value The default value of the variable.
     */
    Integer & access_integer
        (
        const char *    name,
        Int32           default_value = 0
        );

    /** 
     * Access large integer.
     * If attribute is not defined in the registry then the attribute
     * will be created and initialized to @c default_value.
     * @param name Fully-qualified variable name.
     * @param default_value The default value of the variable. 
     */
    LargeInteger & access_large_integer
        (
        const char *    name,
        Int64           default_value = 0
        );

    /**
     * Accesses float attribute. 
     * If attribute is not defined in the registry then the attribute
     * will be created and initialized to @c default_value.
     * @param name Fully-qualified variable name.
     * @param default_value The default value of the variable. 
     */
    Float & access_float
        (
        const char *    name,
        Float32         default_value = 0
        );

    /** 
     * Accesses boolean attribute. 
     * If attribute is not defined in the registry then the attribute
     * will be created and initialized to @c default_value.
     * @param name Fully-qualified variable name.
     * @param default_value The default value of the variable. 
     */
    Boolean & access_boolean
        (
        const char *    name,
        bool            default_value = false
        );

    /** 
     * Accesses boolean attribute. 
     * If attribute is not defined in the registry then the attribute
     * will be created and initialized to @c default_value.
     * @param name Fully-qualified variable name.
     * @param default_value The default value of the variable. 
     */
    String & access_string
        (
        const char *    name,
        const char *    default_value = ""
        );

    //@}

    }; // Settings

/** @} */ // addtogroup registry

    
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STATISTICS_H
