/** @file replica_client.cpp
 * Replication client.
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "replication/replica_client.h"
#include "replication/replica_group_set.h"
#include "replication/replica_manager.h"
#include "object/object.h"
#include "object/object_property_data.h"
#include "object/object_factory.h"
#include "object/object_manager.h"
#include "object/class_manager.h"
#include "pointer/pointer_to_object.h"
#include "database/node_database.h"
#include "serializer/serializer.h"
#include "system/system.h"
#include "system/tick_lock.h"


namespace Massiv {
namespace Core {


// Note: I still don't know how priorities of these messages should be
// defined. Regular messages may pollute the log, but they divide it
// into nice sections. The best solution would probably be:
//   if( current section contains irregular messages )
//       { boost its priority; }

/**
 * Status of replication logs that happen regularly such as "received update".
 */
#define REGULAR_STATUS Status::FACILITY_REPLICATION, Status::PRIORITY_LOW

/**
 * Status of replication logs that are not serious but do not happen often,
 * such as "destroying replica".
 */
#define INFREQUENT_STATUS Status::FACILITY_REPLICATION, Status::PRIORITY_NORMAL


namespace {


/**
 * Get statistics name from node id.
 */
std::string statistics_name
    (
    const NodeId & node_id
    )
    {
    std::ostringstream ss;
    ss << "Replication/Client/" << node_id.to_identifier();
    return ss.str();
    }


/**
 * Dump object replication state and id of @a object to @a s.
 */
void dump_object_info
    (
    std::ostream &      s,
    const ObjectId &    object_id
    )
    {
    GarbageCollector & gc = Global::garbage_collector();
    ObjectProperty * const object = gc.find_local_object( object_id );
    if( object )
        {
        if( object->is_replica() )
            {
            if( object->is_consistent() )
                {
                s << "(CR)"; // consistent replica
                }
            else
                {
                s << "(IR)"; // inconsistent replica
                }
            }
        else
            {
            s << "(O)"; // local object
            }
        }
    else
        {
        s << "(X)"; // non-existing object
        }
    s << " " << object_id;
    }


}// anonymous namespace


// class ReplicaClient::NodeInfo::MyStatistics

ReplicaClient::NodeInfo::MyStatistics::MyStatistics
    (
    const NodeId & node_id
    ) : Statistics
            (
            statistics_name( node_id ).c_str(),
            Status::FACILITY_REPLICATION,
            Statistics::OPEN_NORMAL
            ),
        updates_received    ( create_integer( "updates_received",   0 ) ),
        updates_ignored     ( create_integer( "updates_ignored",    0 ) ),
        acks_sent           ( create_integer( "acks_sent",          0 ) )
    {
    }


ReplicaClient::NodeInfo::MyStatistics::~MyStatistics()
    {
    }


// class ReplicaClient::NodeInfo

ReplicaClient::NodeInfo::NodeInfo
    (
    const NodeId & the_node_id
    ) : node_id( the_node_id ),
        statistics( new MyStatistics( the_node_id ) )
    {
    }


ReplicaClient::NodeInfo::NodeInfo
    (
    const NodeInfo & other
    ) : node_id( other.node_id ),
        update_time( other.update_time ),
        message( other.message ),
        statistics( new MyStatistics( other.node_id ) )
    {
    }


ReplicaClient::NodeInfo & ReplicaClient::NodeInfo::operator=
    (
    const NodeInfo & other
    )
    {
    node_id = other.node_id;
    update_time = other.update_time;
    message = other.message;
    statistics.reset( new MyStatistics( other.node_id ) );
    return *this;
    }


// class ReplicaClient::MyStatistics

ReplicaClient::MyStatistics::MyStatistics()
      : Statistics
            (
            "Replication/Client",
            Status::FACILITY_REPLICATION,
            Statistics::OPEN_NORMAL
            ),
        updates_received        ( create_integer( "updates_received",       0 ) ),
        updates_ignored         ( create_integer( "updates_ignored",        0 ) ),
        acks_sent               ( create_integer( "acks_sent",              0 ) ),
        replica_count           ( create_integer( "replica_count",          0 ) ),
        replica_count_max       ( create_integer( "replica_count_max",      0 ) ),
        inconsistent_count      ( create_integer( "inconsistent_count",     0 ) ),
        inconsistent_count_max  ( create_integer( "inconsistent_count_max", 0 ) )
    {
    }


ReplicaClient::MyStatistics::~MyStatistics()
    {
    }


// class ReplicaClient

ReplicaClient::ReplicaClient()
      : initialized( false )
    {
    }


ReplicaClient::~ReplicaClient()
    {
    if( initialized )
        {
        shutdown();
        }
    }


void ReplicaClient::initialize()
    {
    hope( !initialized );
    
    initialized = true;
    }


void ReplicaClient::shutdown()
    {
    ReplicaManager & replica_manager = Global::replica_manager();
    GarbageCollector & gc = Global::garbage_collector();

    hope( initialized );

    // Enumerate all replicas on this node.

    ReplicationGroupSet replicas( 0,0 ); // repflags ignored by add_registered_replicas
    replicas.add_registered_replicas();

    // Update statistics.

    update_replica_count( 0 );
    update_inconsistent_count( 0 );

    // Call-backs called on Object replicas must not reenter core.
    // Lock the system.

    LOCK_SYSTEM();

    // Notify replica manager of replica destructions.

    ObjectIdSet::const_iterator it;
    for( it = replicas.object_ids.begin(); it != replicas.object_ids.end(); ++it )
        {
        const ObjectId & object_id = *it;
        ObjectProperty * const object = gc.find_local_object( object_id );
        hope( object && object->is_replica() );
        replica_manager.replica_updated( ReplicaManager::REPLICA_DESTROY, object );
        }

    // Destroy all replicas on this node.

    for( it = replicas.object_ids.begin(); it != replicas.object_ids.end(); ++it )
        {
        const ObjectId & object_id = *it;
        ObjectProperty * const object = gc.find_local_object( object_id );
        hope( object && object->is_replica() );
        destroy_replica( object );
        }

    // Reset state.

    nodes.clear();
    initialized = false;
    }


void ReplicaClient::next_tick()
    {
    // Process all recent update messages.

    NodeInfo::HashMap::iterator it;
    for( it = nodes.begin(); it != nodes.end(); ++it )
        {
        NodeInfo & node_info = it->second;
        hope( node_info.statistics.get() );
        if( node_info.message == NULL )
            {
            // No update for this node.

            continue;
            }
        const NodeId & server_node = it->first;

        // Remove message from the info.

        Reference< ReplicationUpdateMessage > message = node_info.message;
        node_info.message = NULL;
        hope( server_node == message->get_initiator() );

        // Debug info.

        #ifdef DEBUG_REPLICATION
        if( settings.debug )
            {
            std::ostringstream ss;
            ss << "Processing replication update from node " << server_node <<
                  ", replica_time = " << message->replica_time << ".";
            Global::log_debug( REGULAR_STATUS, ss.str() );
            }
        #endif // DEBUG_REPLICATION

        // Check contents of update message.

        if( !check_replication_update_message( *message, server_node, node_info ) )
            {
            // Ignore this message.

            node_info.statistics->updates_ignored++;
            statistics.updates_ignored++;
            return;
            }

        // Warning: From now on, any exception is fatal and node must be shut down.

        // Notify replica manager of replication update.

        notify_replica_manager_of_update( server_node, message->simulation_time );

        // Update consistency of objects - mark objects that are not part of
        // the update but can reach updated objects as inconsistent.

        check_replication_update_consistency( *message );

        // Notify replica manager of replica destruction.

        notify_replica_manager_of_replica_destruction( *message );

        // Destroy replicas.

        destroy_replicas( message->objects_to_destroy, server_node );

        // Update [create] object replicas and mark them as consistent.

        Global::allow_replica_writes( true );
        update_replicas( *message, server_node );
        Global::allow_replica_writes( false );

        // Notify replica manager of replica updates.

        notify_replica_manager_of_replica_updates( *message );

        // Create reply message.

        ReplicationAckMessage ack_message;
        ack_message.replica_time = message->replica_time;

        // Send ack.

        node_info.statistics->acks_sent++;
        statistics.acks_sent++;
        ack_message.send( server_node, message->tcp_ack ?
                    NetworkCommon::STREAM_RELIABLE : NetworkCommon::STREAM_UNRELIABLE );

        // Update update_time.

        node_info.update_time = message->replica_time;
        }
    }


void ReplicaClient::get_local_repflags
    (
    int & repflags_yes,
    int & repflags_no
    )
    {
    const NodeId & local_node_id = System::get_local_node_id();
    settings.get_repflags( repflags_yes, repflags_no, local_node_id );
    }


void ReplicaClient::handle_replication_update_message
    (
    ReplicationUpdateMessage & message
    )
    {
    // Find corresponding node info.

    const NodeId & server_node = message.get_initiator();
    NodeInfo & node_info = access_node( server_node );
    hope( node_info.statistics.get() );
    node_info.statistics->updates_received++;
    statistics.updates_received++;

    // Debug info.

    #ifdef DEBUG_REPLICATION
    if( settings.debug )
        {
        std::ostringstream ss;
        ss << "Received replication update from node " << server_node << ": replica_time = " <<
              message.replica_time << ", simulation time " << message.simulation_time <<
              ", " <<  message.objects_to_update.size() << " update(s), " <<
              ", " <<  message.unchanged_objects.size() << " unchanged, " <<
              message.objects_to_destroy.size() << " destroy(s).";
        Global::log_debug( REGULAR_STATUS, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Check if there is already a message from the node.

    if( node_info.message != NULL )
        {
        // Either @a message or node_info.message will be dropped.

        node_info.statistics->updates_ignored++;
        statistics.updates_ignored++;

        // Is this message newer?

        const ReplicationUpdateMessage & other_msg = *node_info.message;
        if( other_msg.replica_time >= message.replica_time )
            {
            hope( other_msg.replica_time > node_info.update_time );

            // Discard the message.

            #ifdef DEBUG_REPLICATION
            if( settings.debug )
                {
                std::ostringstream ss;
                ss << "Discarding replication update from node " << server_node <<
                      " with timestamp " << other_msg.replica_time << ", update with timestamp " <<
                      message.replica_time << " has been already received.";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            return;
            }
        }
    else
        {
        // Check message order.

        if( node_info.update_time >= message.replica_time )
            {
            #ifdef DEBUG_REPLICATION
            if( settings.debug )
                {
                std::ostringstream ss;
                ss << "Received obsolete update with timestamp " << node_info.update_time <<
                    " from node " << server_node << ". Most recent update was " <<
                    node_info.update_time << ".";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            node_info.statistics->updates_ignored++;
            statistics.updates_ignored++;
            return;
            }
        }

    #ifdef DEBUG_REPLICATION
    // Replace old message?

    if( settings.debug.read() && node_info.message != NULL )
        {
        std::ostringstream ss;
        ss << "Discarding old replication update from node " << server_node <<
              " with timestamp " << node_info.message->replica_time <<
              ", update with timestamp " << message.replica_time << " received.";
        Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Remember the message.

    node_info.message = &message;
    }


void ReplicaClient::handle_migration
    (
    const ObjectIdList & objects
    )
    {
    // Notify replica manager of migration.

    notify_replica_manager_of_migration();

    // Make set from @a objects.

    ObjectIdSet object_set;
    ObjectIdList::const_iterator it;
    for( it = objects.begin(); it != objects.end(); ++it )
        {
        object_set.insert( *it );
        }

    // Update consistency info.

    update_consistency( object_set );

    // Notify manager of replica destruction.

    ReplicaManager & replica_manager = Global::replica_manager();
    GarbageCollector & gc = Global::garbage_collector();
    for( it = objects.begin(); it != objects.end(); ++it )
        {
        ObjectProperty * const object = gc.find_local_object( *it );
        if( object && object->is_replica() )
            {
            replica_manager.replica_updated( ReplicaManager::REPLICA_DESTROY, object );
            }
        }

    // Destroy replicas.

    for( it = objects.begin(); it != objects.end(); ++it )
        {
        ObjectProperty * const object = gc.find_local_object( *it );
        if( object && object->is_replica() )
            {
            destroy_replica( object );
            }
        }
    }


bool ReplicaClient::check_replication_update_message
    (
    const ReplicationUpdateMessage &    message,
    const NodeId &                      node_id,
    const NodeInfo &                    node_info
    )
    {
    // Check objects to update and unchanged objects.

    if( !check_objects_to_update( message.objects_to_update, node_id, node_info ) )
        {
        return false;
        }
    if( !check_objects_to_update( message.unchanged_objects, node_id, node_info ) )
        {
        return false;
        }

    // Check objects to destroy.

    if( !check_objects_to_destroy( message.objects_to_destroy, node_id, node_info ) )
        {
        return false;
        }

    return true;
    }


bool ReplicaClient::check_objects_to_update
    (
    const ReplicationUpdateMessage::ObjectList &    objects,
    const NodeId &                                  node_id,
    const NodeInfo &                                node_info
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;

    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        const ObjectId & object_id = object_it->object_id;
        const ObjectVersion update_version = object_it->major_version;

        if( !object_id.is_object_info_available() )
            {
            std::ostringstream  ss;
            ss << "Received replica update without class type id specified.";
            Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );

            return false;
            }

        // Check if class type id is known.
        // Can't use class_type_id.get_metaobject() - it returns non-NULL or
        // throws an exception.

        const ClassTypeId class_type_id = object_id.get_class_type_id();
        ClassManager & class_manager = Global::class_manager();
        if( !class_manager.get_metaobject_no_throw( class_type_id ) )
            {
            std::ostringstream ss;
            ss << "Received replica update with unknown class type id: kind = " <<
                class_type_id.get_kind() << ", id = " << class_type_id.get_class_id() << ".";
            Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );

            return false;
            }
        hope( class_type_id.is_class_registered() );

        // Try to find the object.

        ObjectProperty * const object = gc.find_local_object( object_id );
        if( object )
            {
            // Object must be replica with lower or equal major version
            // and the same class.

            if( !object->is_replica() )
                {
                #ifdef DEBUG_REPLICATION
                if( settings.debug )
                    {
                    std::ostringstream ss;
                    ss << "Received replica update for local non-replica object " <<
                        object_id << " from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION

                return false;
                }

            const ObjectVersion local_version = object->get_major_version();
            if( local_version > update_version )
                {
                #ifdef DEBUG_REPLICATION
                if( settings.debug )
                    {
                    std::ostringstream ss;
                    ss << "Received replica update with old major version (" <<
                        update_version << " < " << local_version << ") of object " <<
                        object->get_object_id() << " from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION

                return false;
                }

            const ClassTypeId & local_class_type_id = object->get_class_type_id();
            if( local_class_type_id != class_type_id )
                {
                std::ostringstream ss;
                ss << "Object class type id mismatch in replica update of object " <<
                    object->get_object_id() << " from node " << node_id <<
                    ": current class is " << local_class_type_id <<
                    ", class in update message is " << class_type_id << ".";
                Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );

                return false;
                }
            }
        else
            {
            // Look into swap.

            if( object_manager.is_swapped_out( object_id ) )
                {
                #ifdef DEBUG_REPLICATION
                if( settings.debug )
                    {
                    std::ostringstream ss;
                    ss << "Received replica update for local non-replica object " <<
                        object_id << " (swapped out) from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION

                return false;
                }

            // FIXME: Is it possible to get class type id of swapped objects?
            }
        }

    // Ok.

    return true;
    }


bool ReplicaClient::check_objects_to_destroy
    (
    const ReplicationUpdateMessage::ObjectList &    objects,
    const NodeId &                                  node_id,
    const NodeInfo &                                node_info
    )
    {
    // No need to check objects to destroy...

    // FIXME: Maybe not true. We must ACK obsolete destroy request, but
    // should we apply updates and destroys in packet containing obsolete
    // destroys? I think it's safe but I'm not 100% sure.

    return true;
    }


void ReplicaClient::check_replication_update_consistency
    (
    const ReplicationUpdateMessage & message
    )
    {
    GarbageCollector & gc = Global::object_manager().gc;

    // Gather set of object ids in @a message.

    ObjectIdSet message_objects;

    // All updated/unchanged objects. They have already been checked by
    // check_objects_to_update() and are proper replicas, so add them all.

    ReplicationUpdateMessage::ObjectList::const_iterator list_it;
    const ReplicationUpdateMessage::ObjectList & updated = message.objects_to_update;
    for( list_it = updated.begin(); list_it != updated.end(); ++list_it )
        {
        message_objects.insert( list_it->object_id );
        }
    const ReplicationUpdateMessage::ObjectList & unchanged = message.unchanged_objects;
    for( list_it = unchanged.begin(); list_it != unchanged.end(); ++list_it )
        {
        message_objects.insert( list_it->object_id );
        }

    // Add objects to destroy that are really replicas.

    const ReplicationUpdateMessage::ObjectList & destroyed = message.objects_to_destroy;
    for( list_it = destroyed.begin(); list_it != destroyed.end(); ++list_it )
        {
        const ObjectId & object_id = list_it->object_id;
        ObjectProperty * const object = gc.find_local_object( object_id );
        if( object && object->is_replica() )
            {
            message_objects.insert( object_id );
            }
        }

    // Update consistency info.

    update_consistency( message_objects );
    }


void ReplicaClient::update_consistency
    (
    const ObjectIdSet & objects
    )
    {
    GarbageCollector & gc = Global::object_manager().gc;

    // Get replication flags of local node.

    int repflags_yes, repflags_no;
    get_local_repflags( repflags_yes, repflags_no );

    // Enumerate object backwards-reachable from @a objects.
    // We must temporarily mark the objects as consistent to make the enumeration
    // work. However, they must be switched back to their previous state - we must
    // leave inconsistent objects in inconsistent state until they are updated.

    ObjectIdSet::const_iterator set_it;
    ReplicationGroupSet local_objects( repflags_yes, repflags_no );
    for( set_it = objects.begin(); set_it != objects.end(); ++set_it )
        {
        ObjectProperty * const object = gc.find_local_object( *set_it );
        if( object )
            {
            const bool is_inconsistent = !object->is_consistent();
            if( is_inconsistent )
                {
                object->set_consistent_replica( true );
                }
            local_objects.add_referencee_objects( object );
            if( is_inconsistent )
                {
                object->set_consistent_replica( false );
                }
            }
        }

    // "@a local_objects (-) @a objects" is set of inconsitent objects, where
    // (-) is "subtract set" operation.

    ObjectIdSet inconsistent_objects;
    for
        (
        set_it = local_objects.object_ids.begin();
        set_it != local_objects.object_ids.end();
        ++set_it
        )
        {
        if( objects.find( *set_it ) == objects.end() )
            {
            inconsistent_objects.insert( *set_it );
            }
        }

    if( inconsistent_objects.empty() )
        {
        return;
        }

    // Switch objects from @a inconsistent_objects to inconsistent state.

    std::size_t non_replica_count = 0;
    for( set_it = inconsistent_objects.begin(); set_it != inconsistent_objects.end(); ++set_it )
        {
        ObjectProperty * const object = gc.find_local_object( *set_it );
        hope( object );

        // Do not touch non-replica objects. I don't think there should be such objects,
        // but I'm not 100% sure, so remember that this happened.

        if( !object->is_replica() )
            {
            non_replica_count++;
            }
        else if( object->is_consistent() )
            {
            #ifdef DEBUG_REPLICATION
            if( settings.debug )
                {
                std::ostringstream ss;
                ss << "Object replica " << *set_it << " became inconsistent.";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            object->set_consistent_replica( false );
            update_inconsistent_count( statistics.inconsistent_count + 1 );
            }
        }

    // Log this weird case really well.

    if( non_replica_count )
        {
        std::ostringstream ss;
        ss << "Found " << non_replica_count << " non-replica inconsistent object(s)\n" <<
              "This probably should not happen. Please send part of replication.log containing\n" <<
              "this message and ~500 lines before it to stoupik and backup all logs from all\n" <<
              "nodes and be ready to provide them as well. Thanks.\n" <<
              "Updated objects: ";
        for( set_it = objects.begin(); set_it != objects.end(); ++set_it )
            {
            if( set_it != objects.begin() )
                {
                ss << ", ";
                }
            dump_object_info( ss, *set_it );
            }
        ss << "\nInconsistent objects: ";
        for( set_it = inconsistent_objects.begin(); set_it != inconsistent_objects.end(); ++set_it )
            {
            if( set_it != inconsistent_objects.begin() )
                {
                ss << ", ";
                }
            dump_object_info( ss, *set_it );
            }
        Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGHEST, ss.str() );
        }
    }


void ReplicaClient::destroy_replicas
    (
    const ReplicationUpdateMessage::ObjectList &    objects,
    const NodeId &                                  node_id
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;

    // Call-backs called on Object replicas must not reenter core.
    // Lock the system.

    LOCK_SYSTEM();

    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the object to destroy.
        
        const ObjectId & object_id = object_it->object_id;
        const ObjectVersion update_version = object_it->major_version;
        ObjectProperty * const object = gc.find_local_object( object_id );

        if( object )
            {
            // Object found.

            if( object->is_replica() )
                {
                // Replica found.

                const ObjectVersion local_version = object->get_major_version();
                if( local_version > update_version )
                    {
                    #ifdef DEBUG_REPLICATION
                    if( settings.debug )
                        {
                        std::ostringstream ss;
                        ss << "Received replica destroy request with old major version (" <<
                            update_version << " < " << local_version << ") of object " <<
                            object->get_object_id() << " from node " << node_id << ".";
                        Global::log_debug( Status::FACILITY_REPLICATION,
                                    Status::PRIORITY_HIGH, ss.str() );
                        }
                    #endif // DEBUG_REPLICATION
                    }
                else
                    {
                    // Ok, destroy the replica.

                    #ifdef DEBUG_REPLICATION
                    if( settings.debug )
                        {
                        std::ostringstream ss;
                        ss << "Destroying replica of object " << object_id <<
                            " as requested by node " << node_id << ".";
                        Global::log_debug( INFREQUENT_STATUS, ss.str() );
                        }
                    #endif // DEBUG_REPLICATION

                    destroy_replica( object );
                    }
                }
            else
                {
                // Local object found.
                
                #ifdef DEBUG_REPLICATION
                if( settings.debug )
                    {
                    std::ostringstream ss;
                    ss << "Received replica destroy request for local non-replica object " <<
                        object_id << " from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION
                }
            }
        else
            {
            // Object not found.

            #ifdef DEBUG_REPLICATION

            if( settings.debug )
                {
                if( object_manager.is_swapped_out( object_id ) )
                    {
                    // Object is swapped out.

                    std::ostringstream ss;
                    ss << "Received replica destroy request for local non-replica object " <<
                        object_id << " (swapped out) from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                else
                    {
                    // No such object known.
                    // Normal priority - this may happen because of high latency.

                    std::ostringstream ss;
                    ss << "Received replica destroy request for non-existing " <<
                        object_id << " from node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_NORMAL, ss.str() );
                    }
                }

            #endif // DEBUG_REPLICATION
            }
        }
    }


void ReplicaClient::destroy_replica
    (
    ObjectProperty * object
    )
    {
    hope( object );
    hope( object->is_replica() );

    { // sob
    // Call-backs called on Object replicas must not reenter core. Lock the system.

    LOCK_SYSTEM();

    // Inform object about its destruction.

    UpcallToObject upcall( object );
    upcall.tag_upcall
        ( "::Massiv::Core::Object::object_updated( Object::REPLICA_DESTROY )" );
    upcall->object_updated( Object::REPLICA_DESTROY );
    } // eob

    update_replica_count( statistics.replica_count - 1 );
    if( !object->is_consistent() )
        {
        update_inconsistent_count( statistics.inconsistent_count - 1 );
        }

    object->delete_object( Object::DELETE_REPLICA );
    }


void ReplicaClient::update_replicas
    (
    const ReplicationUpdateMessage &    message,
    const NodeId &                      node_id
    )
    {
    // Call-backs called on Object replicas must not reenter core.
    // Lock the system.

    LOCK_SYSTEM();

    const ReplicationUpdateMessage::ObjectList & objects = message.objects_to_update;
    const ReplicationUpdateMessage::ObjectList & unchanged = message.unchanged_objects;
    ReplicationUpdateMessage::ObjectList::const_iterator object_it;

    // First we must create all replicas before thay are read from bitstream.
    // This is because when reading a strong pointer from bitstream, the
    // object (replica) the pointer points to must already exists.
    // If allocation throws, uninitialized replicas may remain allocated.
    // This should be easy to fix...

    create_new_replicas( objects, node_id );
    create_new_replicas( unchanged, node_id );

    // Update object replicas.

    Serializer::Description desc;
    desc.ack_time = message.replica_time; // only informational value
    int repflags_yes, repflags_no;
    get_local_repflags( repflags_yes, repflags_no );
    desc.repflags_yes = repflags_yes;
    desc.repflags_no = repflags_no;
    desc.lossy = true;
    desc.object_flags = Serializer::Description::OBJECT_REPLICATE;

    // Must copy the bitstream because @a message is const.

    BitIStream stream = message.objects;

    // Now read the updates from the bitstream, make all objects from the update consistent.
    // Note: It this throws, we're screwed...

    update_replicas( objects, node_id, &stream, &desc );
    update_replicas( unchanged, node_id, NULL, NULL );
    }


void ReplicaClient::create_new_replicas
    (
    const ReplicationUpdateMessage::ObjectList &    objects,
    const NodeId &                                  node_id
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;

    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the object.

        const ObjectId & object_id = object_it->object_id;
        hope( object_id.is_object_info_available() );
        const ClassTypeId class_type_id = object_id.get_class_type_id();
        hope( class_type_id.is_class_registered() );

        ObjectProperty * const object = gc.find_local_object( object_id );

        // Create new object if replica does not exist.

        if( !object )
            {
            #ifdef DEBUG_REPLICATION
            if( settings.debug )
                {
                std::ostringstream ss;
                ss << "Creating replica of object " << object_id <<
                    " as requested by node " << node_id << ".";
                Global::log_debug( INFREQUENT_STATUS, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            hope( !object_manager.is_swapped_out( object_id ) );
            const ObjectFactory * const object_factory = class_type_id.get_factory();
            hope( object_factory );
            object_factory->create_replica( object_id );
            // Factory "sends" REPLICA_CREATED automatically
            update_replica_count( statistics.replica_count + 1 );
            }
        }
    }


void ReplicaClient::update_replicas
    (
    const ReplicationUpdateMessage::ObjectList &    objects,
    const NodeId &                                  node_id,
    BitIStream *                                    stream,
    Serializer::Description *                       desc
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;

    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the object.

        const ObjectId & object_id = object_it->object_id;
        const ObjectVersion update_version = object_it->major_version;
        hope( object_id.is_object_info_available() );
        const ClassTypeId class_type_id = object_id.get_class_type_id();
        hope( class_type_id.is_class_registered() );

        ObjectProperty * const object = gc.find_local_object( object_id );

        // These checks were already done in check_replication_update_message().

        hope( object );
        hope( object->is_replica() );
        DEBUG_ONLY( const ObjectVersion local_version = object->get_major_version(); )
        hope( local_version <= update_version );
        hope( object->get_class_type_id() == class_type_id );

        if( stream && desc )
            {
            // Inform the object that it's going to be updated.

            UpcallToObject upcall( object );
            upcall.tag_upcall
                ( "::Massiv::Core::Object::object_updated( Object::REPLICA_UPDATE )" );
            upcall->object_updated( Object::REPLICA_UPDATE );

            // Read object from bitstream.

            object->ensure_active_unpinned();
            object->bitstr_read( *stream, *desc );

            #ifdef DEBUG_REPLICATION
            if( settings.debug )
                {
                std::ostringstream ss;
                ss << "Contents of replica of object " << object_id <<
                    " updated from node " << node_id << ".";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_LOWEST, ss.str() );
                }
            #endif // DEBUG_REPLICATION
            }

        // Make the object consistent.

        make_replica_consistent( object );

        // Assign major version.

        object->root_data->major_version = update_version;

        if( stream && desc )
            {
            // Inform the object that it has been updated.

            UpcallToObject upcall( object );

            upcall.tag_upcall
                ( "::Massiv:::Core::Object::object_updated( Object::REPLICA_UPDATED )" );
            upcall->object_updated( Object::REPLICA_UPDATED );
            }
        }
    }


void ReplicaClient::make_replica_consistent
    (
    ObjectProperty * object
    )
    {
    hope( object );
    hope( object->is_replica() );

    // Make the object consistent.

    if( !object->is_consistent() )
        {
        update_inconsistent_count( statistics.inconsistent_count - 1 );
        object->set_consistent_replica( true );

        #ifdef DEBUG_REPLICATION
        if( settings.debug )
            {
            std::ostringstream ss;
            ss << "Object replica " << object->get_object_id() << " became consistent.";
            Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
            }
        #endif // DEBUG_REPLICATION
        }
    }


void ReplicaClient::notify_replica_manager_of_update
    (
    const NodeId &  node,
    const STime &   simulation_time
    )
    {
    ReplicaManager & replica_manager = Global::replica_manager();
    replica_manager.replication_message_received( node, simulation_time );
    }


void ReplicaClient::notify_replica_manager_of_migration()
    {
    ReplicaManager & replica_manager = Global::replica_manager();
    replica_manager.migration_message_received();
    }


void ReplicaClient::notify_replica_manager_of_replica_updates
    (
    const ReplicationUpdateMessage &    message
    )
    {
    ReplicaManager & replica_manager = Global::replica_manager();
    GarbageCollector & gc = Global::garbage_collector();
    const ReplicationUpdateMessage::ObjectList & objects = message.objects_to_update;
    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    
    // To 'object_group' we will enumerate all objects from replication groups
    // of updated replicas.

    int repflags_yes, repflags_no;
    get_local_repflags( repflags_yes, repflags_no );
    ReplicationGroupSet object_group( repflags_yes, repflags_no );
    ObjectIdSet::const_iterator object_id_it;
    
    // For each updated replica...

    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the replica.

        const ObjectId & object_id = object_it->object_id;
        ObjectProperty * const object = gc.find_local_object( object_id );
        
        // The object must exist and must be replica
        // because this function is called after update_replicas().

        hope( object );
        hope( object->is_replica() );
        
        // Enumerate replication group of the replica on local node.

        object_group.add_replication_group( object );
        }
    
    // We will first notify of updated replicas because
    // 'objects' is std::vector<> so we can't do fast search.

    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the replica.

        const ObjectId & object_id = object_it->object_id;
        ObjectProperty * const object = gc.find_local_object( object_id );

        // Update only replicas.

        if( object && object->is_replica() )
            {
            // Content of the replica was updated.

            replica_manager.replica_updated
                (
                ReplicaManager::REPLICA_UPDATED,
                object
                );
            }
        
        // Remove the replica from 'object_group'.

        object_group.object_ids.erase( object_id );
        }

    // Now notify rest of replicas.
    // For each replica in 'object_group' notify replica_manager that
    // another replica from its replication group was updated.

    for( object_id_it = object_group.object_ids.begin();
         object_id_it != object_group.object_ids.end();
         ++object_id_it )
        {
        // Try to find the object.

        const ObjectId & object_id = *object_id_it;
        ObjectProperty * const object = gc.find_local_object( object_id );
        
        // Update only replicas.

        if( object && object->is_replica() )
            {
            // Content of the replica was not updated, but content of
            // another replica from its replication group was updated.

            replica_manager.replica_updated
                (
                ReplicaManager::REPLICA_GROUP_UPDATED,
                object
                );
            }
        }
    }


void ReplicaClient::notify_replica_manager_of_replica_destruction
    (
    const ReplicationUpdateMessage &    message
    )
    {
    ReplicaManager & replica_manager = Global::replica_manager();
    GarbageCollector & gc = Global::garbage_collector();
    const ReplicationUpdateMessage::ObjectList & objects = message.objects_to_destroy;

    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        // Try to find the object to destroy.
        // Note: the conditions must be the same as in #destroy_replicas().

        const ObjectId & object_id = object_it->object_id;
        const ObjectVersion update_version = object_it->major_version;
        ObjectProperty * const object = gc.find_local_object( object_id );

        if( object && object->is_replica() )
            {
            const ObjectVersion local_version = object->get_major_version();
            if( local_version <= update_version )
                {
                // destroy_replicas() will destroy this object, notify the manager.

                replica_manager.replica_updated( ReplicaManager::REPLICA_DESTROY, object );
                }
            }
        }
    }


void ReplicaClient::update_replica_count
    (
    std::size_t     new_count
    )
    {
    // msvc6 can't compile without explicit casting.
    const Int32 new_count_i32 = static_cast< Int32 >( new_count );
    if( new_count_i32 > statistics.replica_count_max )
        {
        statistics.replica_count_max = new_count_i32;
        }
    statistics.replica_count = new_count_i32;
    }


void ReplicaClient::update_inconsistent_count
    (
    std::size_t     new_count
    )
    {
    // msvc6 can't compile without explicit casting.
    const Int32 new_count_i32 = static_cast< Int32 >( new_count );
    if( new_count_i32 > statistics.inconsistent_count_max )
        {
        statistics.inconsistent_count_max = new_count_i32;
        }
    statistics.inconsistent_count = new_count_i32;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "replica_client_inline.h"
#endif
