/** @file replica_client.h
 * Replication client.
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICA_CLIENT_H
#define MASSIV_REPLICA_CLIENT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#ifndef MASSIV_REPLICATION_FWD_H
#include "replication/replication_fwd.h"
#endif

#ifndef MASSIV_REPLICATION_SETTINGS_H
#include "replication/replication_settings.h"
#endif

#ifndef MASSIV_REPLICATION_MSG_H
#include "sysevent/replication_msg.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * Replication client.
 */
class ReplicaClient
    {
    NO_COPY_CONSTRUCTOR( ReplicaClient );
    NO_ASSIGNMENT_OPERATOR( ReplicaClient );

private:

    /**
     * Information about single replication server node.
     *
     * @todo No NodeInfo cleanup is done. Keep count of replicas from the
     *       server node and remove NodeInfo after reasonable timeout when
     *       it drops to zero.
     */
    class NodeInfo
        {
        NO_DEFAULT_CONSTRUCTOR( NodeInfo );

    public:

        /**
         * Information about updates from single node.
         */
        class MyStatistics : public Statistics
            {
        public:

            /**
             * Constructor.
             */
            MyStatistics
                (
                const NodeId & node_id
                );

            /**
             * Destructor.
             */
            ~MyStatistics();

        public:

            Integer &   updates_received;
                /**< Number of updates received, both valid and invalid. */

            Integer &   updates_ignored;
                /**< Number of invalid updates received. */

            Integer &   acks_sent;
                /**< Number of ack packets sent. */

            /* Note:
             * Because object replica does not remember which node created/updated
             * its content, we can't remember per-node replica count - object replica
             * may be created and destroyed by different nodes.
             */

            }; // class MyStatistics
    
    public:

        /**
         * Hash map from NodeId to NodeInfo.
         */
        typedef ext::hash_map< NodeId, NodeInfo,
                    NodeId::Hash, NodeId::EqualKey > HashMap;

    public:

        /**
         * Constrctor.
         */
        explicit NodeInfo
            (
            const NodeId & the_node_id
            );

        /**
         * Copy constructor.
         */
        NodeInfo
            (
            const NodeInfo & other
            );

        /**
         * Assignment operator.
         */
        NodeInfo & operator=
            (
            const NodeInfo & other
            );

    public:

        NodeId                                  node_id;
            /**< Id of node this info is about. */

        ReplicaTime                             update_time;
            /**< Replica timestamp of the most recent update from the server. */

        Reference< ReplicationUpdateMessage >   message;
            /**< Most recent update message from the node that has not been
                 processed yet. */

        std::auto_ptr< MyStatistics >           statistics;
            /**< Replication statistics. */

        }; // class NodeInfo

public:

    /**
     * Information replication updates.
     */
    class MyStatistics : public Statistics
        {
    public:

        /**
         * Constructor.
         */
        MyStatistics();

        /**
         * Destructor.
         */
        ~MyStatistics();

    public:

        Integer &   updates_received;
            /**< Total number of updates received. */

        Integer &   updates_ignored;
            /**< How many updates, of #updates_received, were ignored. */

        Integer &   acks_sent;
            /**< Total number of acks sent. */

        Integer &   replica_count;
            /**< Total number of replicas. */

        Integer &   replica_count_max;
            /**< Maximal value of #replica_count. */

        Integer &   inconsistent_count;
            /**< Current number of inconsitent replicas. */

        Integer &   inconsistent_count_max;
            /**< Maximal value of #inconsistent_count. */

        }; // class MyStatistics

public:

    /** List of object ids. */
    typedef ObjectShared::ObjectIdList ObjectIdList;

    /** Set of object ids. */
    typedef ObjectShared::ObjectIdSet ObjectIdSet;

public:

    /**
     * Constructor.
     * @c initialize() is expected to be called to initialize the instance.
     */
    ReplicaClient();

    /**
     * Destructor.
     */
    ~ReplicaClient();

public:

    /**
     * Initializes ReplicaClient.
     */
    void initialize();

    /**
     * Shutdowns ReplicaClient.
     * Resets instance state so that it could be reused by calling
     * @c initialize().
     */
    void shutdown();

    /**
     * Process replication updates received in this tick.
     */
    void next_tick();

public:

    /**
     * Get replication flags of local node.
     */
    void get_local_repflags
        (
        int & repflags_yes,
        int & repflags_no
        );

public:

    /**
     * Handle incoming replication update message.
     *
     * Remember it if it's most recent known update from the node and it passes,
     * #check_replication_update_message(), it will be processed in #next_tick().
     */
    void handle_replication_update_message
        (
        ReplicationUpdateMessage & message
        );

    /**
     * Handle migration message.
     *
     * Destroy all replicas of objects in the message and mark all members of
     * their replication groups that have pointers to them as inconsistent.
     */
    void handle_migration
        (
        const ObjectIdList & objects
        );

private:

    /**
     * Check contents of update message @a message from node @a node_id.
     *
     * @return @c true if it's ok to apply the requested changes to object
     *         replicas, @c false otherwise.
     */
    bool check_replication_update_message
        (
        const ReplicationUpdateMessage &    message,
        const NodeId &                      node_id,
        const NodeInfo &                    node_info
        );

    /**
     * Check list of updated/still alive objects from update message.
     *
     * @return @c true if it's ok to apply the requested changes to object
     *         replicas, @c false otherwise.
     */
    bool check_objects_to_update
        (
        const ReplicationUpdateMessage::ObjectList &    message,
        const NodeId &                                  node_id,
        const NodeInfo &                                node_info
        );

    /**
     * Check list of objects to destroy from update message.
     *
     * @return @c true if it's ok to apply the requested changes to object
     *         replicas, @c false otherwise.
     */
    bool check_objects_to_destroy
        (
        const ReplicationUpdateMessage::ObjectList &    message,
        const NodeId &                                  node_id,
        const NodeInfo &                                node_info
        );

private:

    /**
     * Update consistency of local replica groups as if @a message updates were applied.
     *
     * @note Consistency of objects to update will not be changed.
     */
    void check_replication_update_consistency
        (
        const ReplicationUpdateMessage & message
        );

    /**
     * Update inconsistency flags of all replicas affected by update of @a objects.
     *
     * Replica group consistency defined by an example: Assume there is replication
     * group consisting of objects @c A, @c B, @c C and @c D on @a server1 and
     * @a client has replicas of those objects. Then:
     * -# The replication group is split into two groups on @a server1.
     * -# Objects @c C and @c D migrate to @a server2. Objects @c A and @c B
     *    stay on @a server1. There is a pointer from @c C or @c D to @c A or @c B.
     * -# In next replication update @a client receives update of objects @c A
     *    and @c B from @a server1.
     * .
     * In this case we say that replicas of objects @c C and @c D are in inconsistent
     * state on @a client. This can't be prevented on replication servers and it
     * can cause nasty hard-to-detect-and-debug problems. On @a client objects @c C and @c D
     * still seem to be part of <tt>A-B-C-D</tt> replication group, but that's no longer
     * true. Pointers from replicas of @c C and @c D to replicas ofr @c A and @c B can't be
     * trusted, because replicas of @c A and @c B are more recent and their contents may
     * be unexpected. Replicas of @c A or @c B (or even both) may be even deleted if @a server2
     * deleted the objects - pointer from @c C or @c D will point to non-existing object
     * in that case. Replication group containing new replicas of objects @c A and @c B will
     * be consistent (because updates from single node always contain complete replication groups).
     * Note that if update of @c C and @c D is received first and neither @c A nor @c B has
     * pointer to one of them, replicas of @c A and @c B on @a client will remain consistent.
     *
     * How inconsistencies are detected: Using current state of replicas, enumerate
     * all replicas that have pointer (direct or via other objects) to object
     * in @a objects set. All such objects are inconsistent.
     *
     * This method detects which objects became inconsistent (objects that can reach
     * an object in the @a objetcs set) and marks them as inconsistent. It does not
     * change consistency of the @a objects in the set.
     */
    void update_consistency
        (
        const ObjectIdSet & objects
        );

private:

    /**
     * Destroy replicas given by list @a objects.
     */
    void destroy_replicas
        (
        const ReplicationUpdateMessage::ObjectList &    objects,
        const NodeId &                                  node_id
        );

    /**
     * Destroy given replica @a object.
     */
    void destroy_replica
        (
        ObjectProperty * object
        );

    /**
     * Update replicas from update @a message and mark them as consistent.
     *
     * @pre Message has been checked and update information is valid.
     *
     * @bug Not exception safe. If serialization fails, replicas
     *      will be left in terrible mess. Note that this is not as easy to fix
     *      as it may seem - we can't just delete replicas, we must revert them
     *      previous state. That sucks. This bug will @b not be fixed anytime
     *      soon. It's not such a big problem, because:
     *      - Client nodes should trust server nodes. If replication from
     *        server to client contains nonsense, it means that there is a bug
     *        in replication protocol and it's ok if client exists.
     *      - The same goes for server-to-server replication. In this case, whole
     *        simulation will be terminated.
     *      - Client to server replication is forbidden, clients can't crash
     *        servers this way.
     */
    void update_replicas
        (
        const ReplicationUpdateMessage &    message,
        const NodeId &                      node_id
        );

    /**
     * Create new replicas of @a objects that do not exist yet.
     */
    void create_new_replicas
        (
        const ReplicationUpdateMessage::ObjectList &    objects,
        const NodeId &                                  node_id
        );

    /**
     * Update contents of @a objects replicas.
     *
     * If @a objects contains list of unchanged objects, @a stream and @a desc
     * will be @c NULL.
     */
    void update_replicas
        (
        const ReplicationUpdateMessage::ObjectList &    objects,
        const NodeId &                                  node_id,
        BitIStream *                                    stream,
        Serializer::Description *                       desc
        );

    /**
     * Mark replica of @a object consistent.
     */
    void make_replica_consistent
        (
        ObjectProperty * object
        );

private:

    /**
     * Notify #ReplicaManager that replica updates have been received.
     */
    void notify_replica_manager_of_update
        (
        const NodeId &                      node,
        const STime &                       simulation_time
        );

    /**
     * Notify #ReplicaManager that migration message has been received.
     */
    void notify_replica_manager_of_migration();

    /**
     * Notify #ReplicaManager of replica updates.
     *
     * @pre Replicas from the message have been updated by @c update_replicas().
     */
    void notify_replica_manager_of_replica_updates
        (
        const ReplicationUpdateMessage &    message
        );

    /**
     * Notify #ReplicaManager of destruction of replicas.
     *
     * @pre Replicas from the message have not been destroyed yet.
     */
    void notify_replica_manager_of_replica_destruction
        (
        const ReplicationUpdateMessage &    message
        );

private:

    /**
     * Get reference to node info.
     *
     * Creates correctly initialized node info if it does not exist.
     */
    INLINE NodeInfo & access_node
        (
        const NodeId &      node_id
        );

    /**
     * Insert and initialize new node info.
     *
     * @pre No node info for given node if exists.
     */
    INLINE NodeInfo::HashMap::iterator insert_node
        (
        const NodeId &      node_id
        );

private:

    /**
     * Update "number of replicas" statistics.
     */
    void update_replica_count
        (
        std::size_t     new_count
        );

    /**
     * Update "number of inconsitent replicas" statistics.
     */
    void update_inconsistent_count
        (
        std::size_t     new_count
        );

private:

    bool                    initialized;
        /**< Is ReplicaClient instance initialized? */

    ReplicationSettings     settings;
        /**< Settings. */

    NodeInfo::HashMap       nodes;
        /**< Information about replication server nodes. */

    MyStatistics            statistics;

    }; // class ReplicaClient

/** @} */ // addtogroup replication


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "replica_client_inline.h"
#endif


#endif // MASSIV_REPLICA_CLIENT_H
