/** @file replica_manager.cpp
 * Replica manager (implementation).
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "replication/replica_manager.h"
#include "object/class_manager.h"
#include "node/nodeid.h"
#include "serializer/stype_time.h"
#include "global/global.h"


namespace Massiv {
namespace Core {


/**
 * Internal data structure for #ReplicaManager.
 */
class ReplicaManager::My
    {
public:
    
    /**
     * Callback context. Stores all information required to
     * invoke the callback.
     */
    class CallbackContext
        {
    public:

        /** Hasher. */
        struct Hash
            {
            inline size_t operator()
                (
                const CallbackContext & context
                ) const throw()
                {
                return reinterpret_cast< size_t >( context.callback );
                }

            }; // struct Hash

        /** Compare Callbacks. */
        struct EqualKey
            {
            inline bool operator()
                (
                const CallbackContext & context1,
                const CallbackContext & context2
                ) const throw()
                {
                return context1.callback == context2.callback;
                }

            }; // struct EqualKey

    public:
    
        /** Constructor. */
        CallbackContext() :
            callback( NULL ),
            tag( 0 )
            {
            }
        
        /** Constructor. */
        CallbackContext
            (
            Callback        the_callback,
            int             the_tag,
            VariantPointer  the_data
            ) :
            callback( the_callback ),
            tag( the_tag ),
            data( the_data )
            {
            }
        
        /** Copy constructor. */
        CallbackContext
            (
            const CallbackContext & other
            ) :
            callback( other.callback ),
            tag( other.tag ),
            data( other.data )
            {
            }
        
        /** Destructor. */
        ~CallbackContext()
            {
            }
            
        /** Operator=. */
        CallbackContext & operator=
            (
            const CallbackContext & other
            )
            {
            callback = other.callback;
            tag = other.tag;
            data = other.data;
            return *this;
            }
        
    public:
    
        Callback        callback;
            /**< The callback itself. */
        int             tag;
            /**< User defined tag associated with the callback. */
        VariantPointer  data;
            /**< User defined data associated with the callback. */
        
        }; // CallbackContext
    
public:

    /**
     * Check whether callback stored in @a context matches
     * another callback specified as additional arguments.
     */
    bool match_context
        (
        const CallbackContext & context,
        Callback                callback,
        int                     tag,
        VariantPointer          data,
        bool                    match_tag,
        bool                    match_data
        ) const;


    // Note: MSVC6 is unable to compile and specialize
    // template function
    //
    // template< class MultiMap >
    // bool unregister_callback
    //     (
    //     MultiMap &       multimap,
    //     Callback         callback,
    //     int              tag,
    //     VariantPointer   data,
    //     bool             match_tag,
    //     bool             match_data
    //     );
    // so we must define unregister_object_id_callback()
    // and unregister_class_id_callback() below.
    
    /**
     * Unregister one callback from @a message_callbacks.
     * @return True whether one callback was found and removed,
     *      false if no matching callback was found.
     */
    bool unregister_message_callback
        (
        Callback        callback,
        int             tag,
        VariantPointer  data,
        bool            match_tag,
        bool            match_data
        );

    /**
     * Unregister one callback from @a object_id_callbacks.
     * @return True whether one callback was found and removed,
     *      false if no matching callback was found.
     */
    bool unregister_object_id_callback
        (
        Callback        callback,
        int             tag,
        VariantPointer  data,
        bool            match_tag,
        bool            match_data
        );
        
    /**
     * Unregister one callback from @a class_id_callbacks.
     * @return True whether one callback was found and removed,
     *      false if no matching callback was found.
     */
    bool unregister_class_id_callback
        (
        Callback        callback,
        int             tag,
        VariantPointer  data,
        bool            match_tag,
        bool            match_data
        );
        
public:

    /** Hash multiset for per message callbacks. */
    typedef ext::hash_multiset
        <
        CallbackContext,
        CallbackContext::Hash,
        CallbackContext::EqualKey
        > MessageCallbacks;

    /** Hash multimap for per object callbacks. */
    typedef ext::hash_multimap
        <
        ObjectId,
        CallbackContext,
        ObjectId::Hash,
        ObjectId::EqualKey
        > ObjectIdCallbacks;
    
    /** Hash multimap for per class callbacks. */
    typedef ext::hash_multimap
        <
        ClassTypeId,
        CallbackContext,
        ClassTypeId::Hash,
        ClassTypeId::EqualKey
        > ClassIdCallbacks;
    
public:

    MessageCallbacks    message_callbacks;
        /**< List of callbacks to call when replication message has been received. */

    ObjectIdCallbacks   object_id_callbacks;
        /**< List of per object callbacks. */

    ClassIdCallbacks    class_id_callbacks;
        /**< List of per class callbacks. */

    NodeId              source_node_id;
        /**< Id of node which sent the most recent update. */

    STime               source_node_time;
        /**< Simulation time of node which sent the most recent update. */
    
    }; // My


// ReplicaManager::My
// ~~~~~~~~~~~~~~~~~~
    

bool ReplicaManager::My::match_context
    (
    const CallbackContext & context,
    Callback                callback,
    int                     tag,
    VariantPointer          data,
    bool                    match_tag,
    bool                    match_data
    ) const
    {
    bool  callback_matches = ( context.callback == callback );
    bool  tag_matches = ( context.tag == tag );
    bool  data_matches = ( context.data == data );
    bool  matches = true;
            
    if( !callback_matches )
        matches = false;
            
    if( match_tag && !tag_matches )
        matches = false;
            
    if( match_data && !data_matches )
        matches = false;
    
    return matches;
    }


bool ReplicaManager::My::unregister_message_callback
    (
    Callback        callback,
    int             tag,
    VariantPointer  data,
    bool            match_tag,
    bool            match_data
    )
    {
    MessageCallbacks::iterator it;
    for( it = message_callbacks.begin(); it != message_callbacks.end(); ++it )
        {
        const CallbackContext & context = *it;
        if( match_context( context, callback, tag, data, match_tag, match_data ) )
            {
            message_callbacks.erase( it );
            return true;
            }
        }

    return false;
    }


bool ReplicaManager::My::unregister_object_id_callback
    (
    Callback        callback,
    int             tag,
    VariantPointer  data,
    bool            match_tag,
    bool            match_data
    )
    {
    bool erased = false;
    ObjectIdCallbacks::iterator it = object_id_callbacks.begin();

    while( it != object_id_callbacks.end() )
        {
        const CallbackContext & context = it->second;
        if( match_context( context, callback, tag, data, match_tag, match_data ) )
            {
            object_id_callbacks.erase( it );
            erased = true;
            break;
            }
        else
            {
            ++it;
            }
        }

    return erased;
    }


bool ReplicaManager::My::unregister_class_id_callback
    (
    Callback        callback,
    int             tag,
    VariantPointer  data,
    bool            match_tag,
    bool            match_data
    )
    {
    bool erased = false;
    ClassIdCallbacks::iterator it = class_id_callbacks.begin();
    
    while( it != class_id_callbacks.end() )
        {
        const CallbackContext & context = it->second;
        if( match_context( context, callback, tag, data, match_tag, match_data ) )
            {
            class_id_callbacks.erase( it );
            erased = true;
            break;
            }
        else
            {
            ++it;
            }
        }

    return erased;
    }


// ReplicaManager
// ~~~~~~~~~~~~~~


ReplicaManager::ReplicaManager() :
    my( new My() )
    {
    }


ReplicaManager::~ReplicaManager()
    {
    }

    
void ReplicaManager::register_message_callback
    (
    Callback            callback,
    int                 tag,
    VariantPointer      data
    )
    {
    hope( callback != NULL );
    my->message_callbacks.insert
        (
        My::CallbackContext
            (
            callback,
            tag,
            data
            )
        );
    }


void ReplicaManager::register_callback
    (
    const ObjectId &    object_id,
    Callback            callback,
    int                 tag,
    VariantPointer      data
    )
    {
    hope( callback != NULL );

    // Create callback context.
    My::CallbackContext context
        (
        callback,
        tag,
        data
        );

    // Register the context.
    my->object_id_callbacks.insert
        (
        My::ObjectIdCallbacks::value_type
            (
            object_id,
            context
            )
        );
    }


void ReplicaManager::register_callback
    (
    const ClassTypeId & class_id,
    Callback            callback,
    int                 tag,
    VariantPointer      data
    )
    {
    hope( callback != NULL );
    
    // Create callback context.
    My::CallbackContext context
        (
        callback,
        tag,
        data
        );
    
    // Register the context.
    my->class_id_callbacks.insert
        (
        My::ClassIdCallbacks::value_type
            (
            class_id,
            context
            )
        );
    }


void ReplicaManager::register_callback
    (
    const std::type_info &  type_info,
    Callback                callback,
    int                     tag,
    VariantPointer          data
    )
    {
    ClassManager & class_manager = Global::class_manager();
    const MetaObject * metaobject = class_manager.get_metaobject( type_info );
    hope( metaobject );
    
    register_callback
        (
        metaobject->get_class_type_id(),
        callback,
        tag,
        data
        );
    }

    
void ReplicaManager::unregister_callback
    (
    Callback        callback,
    int             tag,
    VariantPointer  data,
    bool            match_tag,
    bool            match_data
    )
    {
    // Unregister from per message callbacks.
    while( my->unregister_message_callback( callback, tag, data, match_tag, match_data ) )
        {
        }

    // Unregister from per object callbacks.
    while( my->unregister_object_id_callback( callback, tag, data, match_tag, match_data ) )
        {
        }
    
    // Unregister from per class callbacks.
    while( my->unregister_class_id_callback( callback, tag, data, match_tag, match_data ) )
        {
        }
    }


const NodeId & ReplicaManager::get_source_node_id() const
    {
    return my->source_node_id;
    }


const STime & ReplicaManager::get_source_node_time() const
    {
    return my->source_node_time;
    }


void ReplicaManager::replication_message_received
    (
    const NodeId &  node,
    const STime &   simulation_time
    )
    {
    my->source_node_id = node;
    my->source_node_time = simulation_time;

    My::MessageCallbacks::const_iterator it;
    for( it = my->message_callbacks.begin(); it != my->message_callbacks.end(); ++it )
        {
        const My::CallbackContext & context = *it;
        context.callback( UPDATE_RECEIVED, null, context.tag, context.data );
        }
    }


void ReplicaManager::migration_message_received()
    {
    my->source_node_id = NodeId();
    my->source_node_time = STime();

    My::MessageCallbacks::const_iterator it;
    for( it = my->message_callbacks.begin(); it != my->message_callbacks.end(); ++it )
        {
        const My::CallbackContext & context = *it;
        context.callback( MIGRATION_RECEIVED, null, context.tag, context.data );
        }
    }


void ReplicaManager::replica_updated
    (
    UpdateReason            reason,
    ObjectProperty *        replica
    )
    {
    hope( replica );
    hope( replica->is_replica() );
    const ObjectId object_id = replica->get_object_id();
    hope( object_id.is_object_info_available() );
    const ClassTypeId class_id = replica->get_object_info().get_class_type_id();
    hope( class_id.is_class_registered() );

    WeakPointer< Object > replica_ptr( replica->get_object_id() );

    // Lookup per object callbacks.
        {
        std::pair< My::ObjectIdCallbacks::const_iterator,
                   My::ObjectIdCallbacks::const_iterator > p =
                   my->object_id_callbacks.equal_range( object_id );
        My::ObjectIdCallbacks::const_iterator it;

        for( it = p.first; it != p.second; ++it )
            {
            // Invoke callback.
            const My::CallbackContext & context = it->second;
            context.callback( reason, replica_ptr, context.tag, context.data );
            }
        }

    // Lookup per class callbacks.
        {
        std::pair< My::ClassIdCallbacks::const_iterator,
                   My::ClassIdCallbacks::const_iterator > p =
                   my->class_id_callbacks.equal_range( class_id );
        My::ClassIdCallbacks::const_iterator it;

        for( it = p.first; it != p.second; ++it )
            {
            // Invoke callback.
            const My::CallbackContext & context = it->second;
            context.callback( reason, replica_ptr, context.tag, context.data );
            }
        }
    }

    
} // namespace Massiv::Core
} // namespace Massiv
