/** @file replica_manager.h
 * Replica manager.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICA_MANAGER_H
#define MASSIV_REPLICA_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_OBJECT_H
#include "object/object.h"
#endif

#ifndef MASSIV_OBJECT_POINTER_H
#include "pointer/object_pointer.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * ReplicaManager monitors updates of replicas.
 * User can register a callback that is invoked everytime
 * specified replicas are updated.
 *
 * When registering a callback, user must select what replicas
 * the callback will be monitoring. There are two posibilities -
 * monitoring per object or per class.
 * If #ObjectId is specified, the callback is invoked only
 * if replica of the object identified by the object id is updated.
 * If #ClassTypeId is specified, the callback is invoked on
 * every updated replica with same class type id as the one specified.
 *
 * One callback can be registered more than once.
 * There can be unlimited number of callbacks registered
 * for monitoring one #ObjectId or one #ClassTypeId.
 */
class MASSIV_EXPORTED ReplicaManager
    {
    NO_COPY_CONSTRUCTOR( ReplicaManager );
    NO_ASSIGNMENT_OPERATOR( ReplicaManager );

    /// Replica client may call replica_updated() and replication_message_received().
    friend class ReplicaClient;

public:

    /**
     * Reason of invoking #Callback callback.
     */
    enum UpdateReason
        {
        UPDATE_RECEIVED,
            /**< Replication message has been received. Replica pointer will be @c NULL. */

        MIGRATION_RECEIVED,
            /**< Migration message has been received. Replica pointer will be @c NULL. */

        REPLICA_UPDATED,
            /**< Replica content has been updated.
                 This callback will be called after all updates have been applied - all
                 objects in replication group are already up-to-date. */
        
        REPLICA_GROUP_UPDATED,
            /**< Content of another replica from replication group
                 of the replica was updated. The replica itself was not
                 updated.
                 
                 This UpdateReason gives ability to replicas to monitor changes in
                 their replication groups. */

        REPLICA_DESTROY,
            /**< Replica will be destroyed.
                 This callback is called before any objects in the replication group
                 are destroyed. */

        }; // enum UpdateReason

    /**
     * Static replica updated callback.
     * Called when replica state has changed.
     *
     * @param reason Why the callback was invoked.
     * @param replica Pointer to the replica. Always non-null and pointing
     *      to a local object unless @a reason is @c UPDATE_RECEIVED.
     *      If @a reason is @c REPLICA_DESTROY, @a replica points
     *      to the replica that will be destroyed asap.
     * @param user_tag User defined tag specified when the monitor object
     *      was registered to #ReplicaManager.
     * @param user_data User data specified when the monitor object
     *      was registered to #ReplicaManager.
     *
     * @warning It is forbidden to call synchronous RPC from this function.
     */
    typedef void ( * Callback )
        (
        UpdateReason                reason,
        WeakPointer< Object >       replica,
        int                         user_tag,
        VariantPointer              user_data
        );
    
public:

    /**
     * Constructor.
     */
    ReplicaManager();
    
    /**
     * Destructor.
     */
    ~ReplicaManager();
    
public:
    
    /**
     * Register callback that should be invoked when valid replication
     * update message has been received.
     */
    void register_message_callback
        (
        Callback            callback,
        int                 user_tag,
        VariantPointer      user_data
        );

    /**
     * Register callback that should be invoked when replica
     * of object identified by @a object_id has changed.
     * @param object_id #ObjectId of the object whose replica the callback
     *      monitors.
     * @param callback Callback to invoke when replica is updated.
     * @param user_tag See #Callback.
     * @param user_data See #Callback.
     */
    void register_callback
        (
        const ObjectId &    object_id,
        Callback            callback,
        int                 user_tag,
        VariantPointer      user_data
        );
    
    /**
     * Register callback that shoud be invoked when replica
     * of #ClassTypeId same as @a class_id has changed.
     * @param class_id #ClassTypeId of objects to monitor.
     * @param callback Callback to invoke when replica is updated.
     * @param user_tag See #Callback.
     * @param user_data See #Callback.
     */
    void register_callback
        (
        const ClassTypeId & class_id,
        Callback            callback,
        int                 use_tag,
        VariantPointer      use_data
        );
        
    /**
     * Register callback that shoud be invoked when replica
     * of @c std::type_info same as @a type_info has changed.
     * @param type_info Type of objects to monitor.
     * @param callback Callback to invoke when replica is updated.
     * @param user_tag See #Callback.
     * @param user_data See #Callback.
     */
    void register_callback
        (
        const std::type_info &  type_info,
        Callback                callback,
        int                     user_tag,
        VariantPointer          user_data
        );

    /**
     * Unregister all object callbacks that match @a callback.
     *
     * @param callback Callback to remove.
     * @param tag If @a match_tag is true, only callbacks that have been
     *      registered with this @a tag are removed. If @a match_tag
     *      is false, @a tag is ignored.
     * @param data If @a match_data is true, only callbacks that have been
     *      registered with this @a data are removed. If @a match_data
     *      is false, @a data is ignored.
     */
    void unregister_callback
        (
        Callback        callback,
        int             tag = 0,
        VariantPointer  data = VariantPointer(),
        bool            match_tag = false,
        bool            match_data = false
        );

public:

    /**
     * Get id of node which sent current update.
     *
     * This method can be called only from the callback. Source node id
     * is currently not available for migration updates.
     */
    const NodeId & get_source_node_id() const;

    /**
     * Get simulation time of source node at the moment it sent the update.
     *
     * Simulation time is not available for migration updates.
     */
    const STime & get_source_node_time() const;

// public for ReplicaClient
protected:

    /**
     * Method called by #ReplicaClient when update message has been received.
     *
     * @param node Id of node which sent the update.
     * @param simulation_time Simulation time of node which sent the update at
     *                        the moment the update was sent.
     */
    void replication_message_received
        (
        const NodeId &  node,
        const STime &   simulation_time
        );

    /**
     * Method called by #ReplicaClient when migration message has been received.
     */
    void migration_message_received();

    /**
     * Method called by #ReplicaClient when replica has been updated or will
     * be destroyed.
     */
    void replica_updated
        (
        UpdateReason        reason,
        ObjectProperty *    replica
        );

private:

    // Internal data structure.
    class My;
    
    std::auto_ptr< My > my;
        /**< Internal data structure. Always non-null. */
    
    }; // ReplicaManager

/** @} */ // addtogroup replication


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_REPLICA_MANAGER_H
