/** @file replica_server.cpp
 * Replication server.
 */

/*
 * begin        : Sat Feb 01 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "replication/replica_server.h"
#include "replication/replica_group_set.h"
#include "global/global.h"
#include "utility/debug.h"
#include "utility/text_writer.h"
#include "system/system.h"
#include "net/net.h"
#include "serializer/serializer.h"
#include "sysevent/replication_msg.h"
#include "object/object.h"
#include "object/object_manager.h"
#include "object/stack_root_object.h"
#include "database/node_database.h"
#include "balance/balancer.h"
#include "object/class_manager.h"


namespace Massiv {
namespace Core {


// Note: I still don't know how priorities of these messages should be
// defined. Reguilar messages may pollute the log, but they divide it
// into nice sections. The best solution would probably be:
//   if( current section contains irregular messages )
//       { boost its priority; }

/**
 * Status of replication logs that happen regularly such as "sending update".
 */
#define REGULAR_STATUS Status::FACILITY_REPLICATION, Status::PRIORITY_LOW

/**
 * Status of replication logs that are not serious but do not happen often,
 * such as "destroying replica".
 */
#define INFREQUENT_STATUS Status::FACILITY_REPLICATION, Status::PRIORITY_NORMAL


/** @def OPTIMIZE_UNCHANGED_REPLICAS
 * If defined, replication updates will be smaller, but the algorithm to
 * send the optimized updates is more complex - slower and potentionally buggy.
 */
#define OPTIMIZE_UNCHANGED_REPLICAS


// class ReplicaServer::NodeInfo

ReplicaServer::NodeInfo::NodeInfo
    (
    const NodeId & the_node_id
    ) : node_id( the_node_id ),
        statistics( new ReplicaServerNodeStatistics( the_node_id ) ),
        pending_updates( node_id ),
        waiting_for_tcp_ack( false )
    {
    hope( statistics.get() );
    pending_updates.set_statistics( statistics.get() );
    }


ReplicaServer::NodeInfo::NodeInfo
    (
    const NodeInfo & other
    ) : node_id( other.node_id ),
        statistics( new ReplicaServerNodeStatistics( other.node_id ) ),
        requests( other.requests ),
        pending_updates( other.pending_updates ),
        versions( other.versions ),
        waiting_for_tcp_ack( other.waiting_for_tcp_ack ),
        wait_for_ack_time( other.wait_for_ack_time ),
        update_time( other.update_time )
    {
    hope( statistics.get() );
    pending_updates.set_statistics( statistics.get() );
    }


ReplicaServer::NodeInfo & ReplicaServer::NodeInfo::operator=
    (
    const NodeInfo & other
    )
    {
    node_id = other.node_id;
    statistics.reset( new ReplicaServerNodeStatistics( node_id ) );
    requests = other.requests;
    pending_updates = other.pending_updates;
    pending_updates.set_statistics( statistics.get() );
    versions = other.versions;
    waiting_for_tcp_ack = other.waiting_for_tcp_ack;
    wait_for_ack_time = other.wait_for_ack_time;
    update_time = other.update_time;
    return *this;
    }


// class ReplicaServer

ReplicaServer::ReplicaServer() :
    initialized( false )
    {
    }


ReplicaServer::~ReplicaServer()
    {
    if( initialized )
        {
        shutdown();
        }
    }


void ReplicaServer::initialize()
    {
    hope( !initialized );

    // @c replica_time must not be invalid.

    advance_replica_time();

    initialized = true;
    }


void ReplicaServer::shutdown()
    {
    hope( initialized );

    // Clear all nodes.
    // Note: nodes.clear() would not update the statistics.

    NodeInfo::HashMap::const_iterator it;
    while( ( it = nodes.begin() ) != nodes.end() )
        {
        erase_node( it->first );
        }

    // Reset replica update timestamp.

    replica_time = ReplicaTime();

    initialized = false;
    }


void ReplicaServer::request_replication
    (
    const ObjectId &    object,
    const NodeId &      destination_node,
    const STime &       timeout
    )
    {
    #ifdef DEBUG_REPLICATION
    if( settings.debug )
        {
        std::ostringstream ss;
        ss << "Requesting replication of object " << object << " to node " << destination_node <<
            " until simulation time " << timeout << ".";
        Global::log_debug( INFREQUENT_STATUS, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Get table of replica requests.

    NodeInfo & node_info = access_node( destination_node );
    ReplicaRequests & requests = node_info.requests;

    // Add the request.

    const std::pair< ReplicaRequests::iterator, bool > result =
                requests.insert( ReplicaRequests::value_type( object, ReplicaRequest() ) );
    ReplicaRequest & request = result.first->second;
    if( result.second )
        {
        // New request.

        update_request_count( node_info, requests.size(), requests.size() + 1 );
        }
    request.timeout = timeout;
    }


void ReplicaServer::node_disconnected
    (
    const NodeId &      node
    )
    {
    #ifdef DEBUG_REPLICATION
    if( settings.debug )
        {
        std::ostringstream ss;
        ss << "Forgetting all information about replication to node " << node << ".";
        Global::log_debug( INFREQUENT_STATUS, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Clear information inside the node_info, but do not destroy it, because
    // this method may be called when someone is iterating over #nodes hash.

    clear_node( node );
    }


namespace {

/**
 * Returns metaobject for specified @c object_id or returns NULL if type
 * information can not be retrieved or interpreted on the local node.
 */
const MetaObject * get_metaobject
    (
    const ObjectId & object_id
    )
    {
    if( !object_id.is_object_info_available() )
        {
        return NULL;
        }
    else
        {
        return object_id.get_class_type_id().get_metaobject_no_throw();
        }
    }

/**
 * Tests if the class name of an object identified by @c object_id is one of
 * the names present in @c classes list. Assumes that the name match if it
 * can not be retrieved from @c object_id.
 */
bool match_class_name
    (
    const std::vector< std::string > &  classes,
    const ObjectId &                    object_id
    )
    {
    const MetaObject * metaobject = get_metaobject( object_id );

    if( metaobject != NULL && !std::binary_search( classes.begin(),
        classes.end(), metaobject->get_class_name() ) )
        {
        return false;
        }

    return true;
    }

} // anonymous namespace


void ReplicaServer::gather_replicated_objects
    (
    ObjectShared::ObjectIdSet & objects,
    const std::string &         class_name_pattern,
    bool                        include_subclasses
    ) const
    {
    NodeInfo::HashMap::const_iterator   node_it;
    std::size_t                         count;
    std::vector< std::string >          classes;
    const std::size_t                   max_report_count = 128;

    Global::class_manager().find_classes( classes, class_name_pattern,
        include_subclasses );

    for( node_it = nodes.begin(), count = 0; node_it != nodes.end(); ++node_it )
        {
        const NodeInfo & node_info = node_it->second;

        // Check requests.

        const ReplicaRequests & requests = node_info.requests;
        ReplicaRequests::const_iterator req_it;
        for( req_it = requests.begin(); req_it != requests.end(); ++req_it )
            {
            const ObjectId & object_id = req_it->first;

            if( !match_class_name( classes, object_id ) )
                {
                continue;
                }

            if( ++count > max_report_count )
                {
                return;
                }

            objects.insert( object_id );
            }

        // Check versions.

        const ReplicaVersions & versions = node_info.versions;
        ReplicaVersions::const_iterator ver_it;
        for( ver_it = versions.begin(); ver_it != versions.end(); ++ver_it )
            {
            const ObjectId & object_id = ver_it->first;

            if( !match_class_name( classes, object_id ) )
                {
                continue;
                }

            if( ++count > max_report_count )
                {
                return;
                }

            objects.insert( object_id );
            }
        }
    }


void ReplicaServer::gather_replication_nodes
    (
    std::vector< NodeId > &     result,
    const ObjectId &            object
    ) const
    {
    NodeInfo::HashMap::const_iterator   node_it;
    std::size_t                         count;
    const std::size_t                   max_report_count = 128;

    for( node_it = nodes.begin(), count = 0; node_it != nodes.end(); ++node_it )
        {
        const NodeId & node_id = node_it->first;
        const NodeInfo & node_info = node_it->second;

        if( node_info.requests.find( object ) != node_info.requests.end() )
            {
            if( ++count > max_report_count )
                {
                return;
                }

            result.push_back( node_id );
            }
        }
    }


void ReplicaServer::dump_object_info
    (
    std::ostream &      s,
    const ObjectId &    object
    ) const
    {
    const std::size_t max_report_count = 8;

    // Dump requests.

    NodeInfo::HashMap::const_iterator node_it;
    bool has_requests = false;
    std::size_t count = 0;

    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        const NodeId & node_id = node_it->first;
        const NodeInfo & node_info = node_it->second;
        const ReplicaRequests::const_iterator req_it = node_info.requests.find( object );
        if( req_it != node_info.requests.end() )
            {
            if( has_requests )
                {
                s << ", ";
                }
            else
                {
                s << "Requests to: ";
                has_requests = true;
                }

            if( ++count > max_report_count )
                {
                s << "...";
                break;
                }

            s << node_id << " (until " << req_it->second.timeout << ")";
            }
        }
    if( !has_requests )
        {
        s << "No replication requests";
        }
    s << "\n";

    // Dump replica info.

    bool has_versions = false;
    count = 0;

    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        const NodeId & node_id = node_it->first;
        const NodeInfo & node_info = node_it->second;
        const ReplicaVersions::const_iterator ver_it = node_info.versions.find( object );
        if( ver_it != node_info.versions.end() )
            {
            if( has_versions )
                {
                s << ", ";
                }
            else
                {
                s << "Replicas on: ";
                has_versions = true;
                }

            if( ++count > max_report_count )
                {
                s << "...";
                break;
                }

            s << node_id;
            const ReplicaVersion::State state = ver_it->second.state;
            if( state != ReplicaVersion::STATE_REPLICATING )
                {
                s << " (" << ReplicaVersion::state_string( state ) << ")";
                }
            }
        }
    if( !has_versions )
        {
        s << "No replica versions";
        }
    }


void ReplicaServer::next_tick()
    {
    // Test if properly initialized.

    hope( replica_time != ReplicaTime() );

    // For each node, check if update should be sent.

    bool advance_time = false;
    std::vector< NodeId > empty_infos;
    const SystemTime empty_time( SystemTime::current_time() + 20.0 );
    NodeInfo::HashMap::iterator node_it;
    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        const NodeId & node_id = node_it->first;
        NodeInfo & node_info = node_it->second;

        // If update should be sent, send it.

        if( should_update( node_id, node_info ) )
            {
            // Send update.

            if( update( node_id, node_info ) )
                {
                advance_time = true;
                }
            }

        // Remember empty node infos.

        if( node_info.is_empty() && node_info.update_time > empty_time  )
            {
            empty_infos.push_back( node_id );
            }
        }

    // Remove empty node infos.

    std::vector< NodeId >::const_iterator empty_node_it;
    for( empty_node_it = empty_infos.begin(); empty_node_it != empty_infos.end(); ++empty_node_it )
        {
        const NodeId & node_id = *empty_node_it;
        erase_node( node_id );
        }

    // If at least one update has been sent, subsequent writes must be done with
    // increased timestamp.

    if( advance_time )
        {
        advance_replica_time();
        }
    }


void ReplicaServer::handle_replication_ack_message
    (
    const ReplicationAckMessage & message
    )
    {
    statistics.acks_received++;

    // Find corresponding node info.

    const NodeId & client_node = message.get_initiator();
    NodeInfo::HashMap::iterator node_info_it = nodes.find( client_node );
    if( node_info_it == nodes.end() )
        {
        // No node info -> no version info -> should not receive any acks.

        #ifdef DEBUG_REPLICATION
        if( settings.debug )
            {
            std::ostringstream ss;
            ss << "Received ACK message, but no info for client node " <<
                client_node << " exists.";
            Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
            }
        #endif // DEBUG_REPLICATION

        statistics.acks_ignored++;
        return;
        }

    #ifdef DEBUG_REPLICATION
    if( settings.debug )
        {
        std::ostringstream ss;
        ss << "Received replication ACK from node " << client_node << ", stamp " <<
            message.replica_time << ".";
        Global::log_debug( REGULAR_STATUS, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Process ack by backlog.

    NodeInfo & node_info = access_node( client_node );
    const std::size_t old_version_count = node_info.versions.size();
    if( !node_info.pending_updates.ack( message, node_info.versions, settings.debug ) )
        {
        statistics.acks_ignored++;
        }
    update_replica_count( node_info, old_version_count, node_info.versions.size() );

    // If updating via TCP, allow more updated if expected ACK has been received.

    if( node_info.waiting_for_tcp_ack )
        {
        if( message.replica_time >= node_info.wait_for_ack_time )
            {
            // Got the ACK - allow more updates.

            if( message.replica_time > node_info.wait_for_ack_time )
                {
                // This should never happen - we do not send out updates once,
                // switched to TCP and ack is received.

                std::ostringstream ss;
                ss << "Received more recent update ACK with timestamp " <<
                      message.replica_time.get_value() << " from node " << client_node <<
                      " than expected TCP update ACK " << node_info.wait_for_ack_time << ".";
                Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                }

            node_info.waiting_for_tcp_ack = false;
            }
        }
    }


void ReplicaServer::advance_replica_time()
    {
    replica_time.advance();

    // Notify StackRootObject about the time advance.
    // We must ensure that its dirty time is always equal to the
    // current replica time.

    Global::object_manager().stack_root.replica_time_advanced();
    }


bool ReplicaServer::has_replica_requests
    (
    const ObjectId & object_id
    ) const
    {
    // Check each node for object replication request.

    NodeInfo::HashMap::const_iterator node_it;
    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        // Try to find object replication request.

        const NodeInfo & node_info = node_it->second;
        const ReplicaRequests & requests = node_info.requests;
        ReplicaRequests::const_iterator request_it = requests.find( object_id );
        if( request_it != requests.end() )
            {
            return true;
            }
        }

    return false;
    }


void ReplicaServer::handle_outgoing_migration
    (
    BitOStream &                        bos,
    const Serializer::Description &     desc,
    const ObjectId &                    object_id
    )
    {
    // Check if migration destination is client node.

    hope( flags_on( desc.object_flags, Serializer::Description::OBJECTF_REPLICATION ) );
    bool destination_client = false;
    const ObjectId & destination_id = desc.destination_object;
    hope( !destination_id.is_null() );
    if( destination_id.is_proxy_node_object_id() &&
        destination_id.proxy_extract_node_id().is_type( NodeId::TYPE_CLIENT ) )
        {
        destination_client = true;
        }

    // Do not transfer replication responsibility to clients.

    if( !destination_client )
        {
        #ifdef DEBUG_REPLICATION
        bool logged = false;
        std::ostringstream ss;
        #endif

        // Check each node for object replication request and object version info.

        NodeInfo::HashMap::iterator node_it;
        for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
            {
            // Try to find object replication request and replica version info.

            const NodeId & node_id = node_it->first;
            NodeInfo & node_info = node_it->second;
            ReplicaRequests & requests = node_info.requests;
            ReplicaRequests::iterator request_it = requests.find( object_id );
            ReplicaVersions & versions = node_info.versions;
            ReplicaVersions::iterator version_it = versions.find( object_id );

            // Send info regarding node_id?

            if( request_it != requests.end() || version_it != versions.end() )
                {
                // Remember logging info about it.

                #ifdef DEBUG_REPLICATION
                if( settings.debug )
                    {
                    if( !logged )
                        {
                        ss << "Transferring replication responsibilities for object " << object_id <<
                            " which migrates to " << destination_id << ": ";
                        logged = true;
                        }
                    else
                        {
                        ss << ", ";
                        }
                    ss << "node " << node_id;
                    if( request_it != requests.end() )
                        {
                        ss << " (request timeout " << request_it->second.timeout << ")";
                        }
                    if( version_it != versions.end() )
                        {
                        ss << " has replica";
                        }
                    }
                #endif // DEBUG_REPLICATION
                }

            // Write node id.

            hope( !node_id.is_null() );
            node_id.bitstr_write( bos, desc );

            // Write replication request info and remove the request.

            if( request_it != requests.end() )
                {
                const ReplicaRequest & request = request_it->second;
                bos.write_int( 1, 1 );
                request.timeout.bitstr_write( bos, desc );
                update_request_count( node_info, requests.size(), requests.size() - 1 );
                requests.erase( request_it );
                }
            else
                {
                bos.write_int( 0, 1 );
                }

            // May the node have replica of the object? Remove the info.

            if( version_it != versions.end() )
                {
                bos.write_int( 1, 1 );
                update_replica_count( node_info, versions.size(), versions.size() - 1 );
                versions.erase( version_it );
                }
            else
                {
                bos.write_int( 0, 1 );
                }
            }

        // Log it.

        #ifdef DEBUG_REPLICATION
        if( settings.debug.read() && logged )
            {
            Global::log_debug( INFREQUENT_STATUS, ss.str() );
            }
        #endif
        }

    // Terminate info by NULL node id.

    NodeId null_id;
    hope( null_id.is_null() );
    null_id.bitstr_write( bos, desc );
    }


void ReplicaServer::handle_incoming_migration
    (
    BitIStream &                        bis,
    const Serializer::Description &     desc,
    const ObjectId &                    object_id
    )
    {
    while( 1 )
        {
        // Read replication info.

        NodeId node_id;
        node_id.bitstr_read( bis, desc );
        if( node_id.is_null() )
            {
            break;
            }

        // Request replication of object_id to node_id?

        if( bis.read_int( 1 ) )
            {
            STime timeout;
            timeout.bitstr_read( bis, desc );
            request_replication( object_id, node_id, timeout );
            }

        // Is there replica of object_id on node_id?

        if( bis.read_int( 1 ) )
            {
            // Object must already exists, with correct major version.

            GarbageCollector & gc = Global::object_manager().gc;
            ObjectProperty * const object = gc.find_local_object( object_id );
            hope( object );
            const ObjectVersion major_version = object->get_major_version();
            hope( major_version >= 1 );

            // Switch version info state to "creating". That ensures that if the
            // object should replicated to node_id, it will, and otherwise it
            // will switch to "destroying" state and will be properly destroyed.

            NodeInfo & node_info = access_node( node_id );
            ReplicaVersions & versions = node_info.versions;
            const std::size_t old_version_count = versions.size();
            ReplicaVersion & version_info = versions[ object_id ];
            version_info.set_state_creating( replica_time, major_version );
            update_replica_count( node_info, old_version_count, versions.size() );
            }
        }
    }


void ReplicaServer::text_write_requests
    (
    TextWriter &                        tw,
    const Serializer::Description &     desc,
    const ObjectId &                    object_id,
    bool                                remove_requests
    )
    {
    // Check each node for object replication request.

    NodeInfo::HashMap::iterator node_it;
    for( node_it = nodes.begin(); node_it != nodes.end(); ++node_it )
        {
        // Try to find object replication request.

        const NodeId & node_id = node_it->first;
        NodeInfo & node_info = node_it->second;
        ReplicaRequests & requests = node_info.requests;
        ReplicaRequests::iterator request_it = requests.find( object_id );
        if( request_it != requests.end() )
            {
            // Serialize the request.

            hope( !node_id.is_null() );
            node_id.text_write( tw, desc );
            tw.write_space();

            const ReplicaRequest & request = request_it->second;
            request.timeout.text_write( tw, desc );
            tw.write_space();

            // Optionally remove the request.

            if( remove_requests )
                {
                requests.erase( request_it );
                }
            }
        }

    // Write NULL id.

    NodeId null_id;
    hope( null_id.is_null() );
    null_id.text_write( tw, desc );
    }


void ReplicaServer::text_read_requests
    (
    TextReader &                        tr,
    const Serializer::Description &     desc,
    const ObjectId &                    object_id
    )
    {
    while( 1 )
        {
        // Read the request.

        NodeId node_id;
        node_id.text_read( tr, desc );
        if( node_id.is_null() )
            {
            break;
            }

        // FIXME: check node database if node_id is ok.

        STime timeout;
        timeout.text_read( tr, desc );

        // Insert request into the table.

        request_replication( object_id, node_id, timeout );
        }
    }


bool ReplicaServer::should_update
    (
    const NodeId &      node_id,
    const NodeInfo &    node_info
    ) const
    {
    // Check if we're waiting for an ACK.

    if( node_info.waiting_for_tcp_ack )
        {
        return false;
        }

    // Check update time.

    Float32 min_period, max_period;
    settings.get_update_periods( min_period, max_period, node_id );
    const Float32 update_period = node_info.pending_updates.update_period
                ( min_period, max_period );
    const SystemTime current_time( SystemTime::current_time() );
    if( current_time < node_info.update_time + update_period )
        {
        return false;
        }

    // Ok - do update.

    return true;
    }


namespace {


/**
 * Gather set of objects to replicate to node given by @a node_info
 * to @a replication_group_set.
 *
 * As a side effect invalid requests will be removed from replica request list.
 * These include requests for non-local objects and timeout requests.
 *
 * If @a cleanup_only is @c true, no enumeration will be done, only the cleanup.
 */
void gather_objects_to_replicate
    (
    ReplicationGroupSet &       replication_group_set,
    ReplicaServer::NodeInfo &   node_info,
    bool                        cleanup_only
    )
    {
    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;
    const STime & current_time = Massiv::System::time();

    // Gather set of objects to replicate.

    ObjectShared::ObjectIdList invalid_requests;
    ReplicaRequests & requests = node_info.requests;
    ReplicaRequests::const_iterator request_it;
    for( request_it = requests.begin(); request_it != requests.end(); ++request_it )
        {
        const ObjectId & object_id = request_it->first;
        const ReplicaRequest & request = request_it->second;

        // Find the object.

        ObjectProperty * object = gc.find_local_object( object_id );
        if( !object )
            {
            object = object_manager.try_to_swap_in( object_id );
            }

        // Enumerate its replication group if object replication request is valid.

        if( object && !object->is_replica() && request.timeout > current_time )
            {
            if( !cleanup_only )
                {
                replication_group_set.add_replication_group( object );
                }
            }
        else
            {
            // Remove this request.

            invalid_requests.push_back( object_id );
            }
        }

    // Request cleanup.

    ObjectShared::ObjectIdList::const_iterator invalid_request_it;
    for( invalid_request_it = invalid_requests.begin();
        invalid_request_it != invalid_requests.end();
        ++invalid_request_it )
        {
        requests.erase( *invalid_request_it );
        }
    }


/**
 * Fill lists of objects to update (those in @a replication_group_set)
 * and to destroy (those not in @a replication_group_set but existing
 * on node given by @a node_info) into @a message and serialize replicas
 * of objects to update.
 *
 * Note: @a repliaction_group_set will be modified.
 */
void fill_replication_update_message
    (
    ReplicationUpdateMessage &  message,
    ReplicationGroupSet &       replication_group_set,
    const NodeId &              node_id,
    ReplicaServer::NodeInfo &   node_info,
    int                         repflags_yes,
    int                         repflags_no,
    bool                        debug
    )
    {
    typedef ReplicationGroupSet::ObjectIdSet ObjectIdSet;

    ObjectManager & object_manager = Global::object_manager();
    GarbageCollector & gc = object_manager.gc;
    ReplicaServer & replica_server = *object_manager.replica_server;
    Balancer & balancer = object_manager.balancer;
    const ReplicaTime & replica_time = replica_server.get_replica_time();

    // We will directly modify the object id set...

    ObjectIdSet & objects_to_replicate = replication_group_set.object_ids;

    // This method is very tricky. We start with two objects sets:
    // - objects_to_replicate - objects that should be replicated to the client
    // - node_info.versions - we know something about their state on the client
    // We want to:
    // - create new replicas of objects in objects_to_replicate but not in versions
    // - destroy replicas of objects in versions but not in objects_to_replicate
    // - objects in both sets are the tricky ones; we know that client has object
    //   replica from local or destroyid it recently (but we haven't received
    //   ack about that yet); so we:
    //   - update contents of objects that are dirty
    //
    // However, this is not sufficient. Replication client must update information
    // about replica constistency. To do that, it must know which replicas from this
    // server are up-to-date. We simply send object ids of all object that were in
    // both sets, but were not dirty.
    //
    // Note that this could be probably optimized - we don't have to send all such
    // object ids, only those that are required for successful consistency information
    // update. In the past we were actually sending object ids only of those objects
    // that were in replication groups of some of the updated objects. But the code
    // was really messy and I did't trust it much, so I reimplemented it in suboptimal
    // but currect way. The optimized version would look like this:
    // "For each object from original objects_to_replicate that has been neither
    // newly creaty nor updated, add it to unchanged_objects."
    // If this rule was applied, replication groups where no objects are new
    // or dirty would not be included in unchanged_objects. That seems to be ok;
    // the objects should not be in inconsistent state on the client, because they
    // should be in exactly the same state on the client as on the server, ie.
    // complete replication group with no dangling replication-group-defining
    // pointers.
    //
    // This optimization is performed iff OPTIMIZE_UNCHANGED_REPLICAS is defined.

    #ifdef OPTIMIZE_UNCHANGED_REPLICAS

    // Set of all potentionally unchanged_objects.
    ObjectIdSet         unchanged_objects;

    // Objects belonging to groups with at least one update.
    ReplicationGroupSet updated_groups( repflags_yes, repflags_no );

    #endif // OPTIMIZE_UNCHANGED_REPLICAS

    // Prepare serialization.
    // There are two Serializer::Descriptions:
    // - "desc" is used to serialize the object and its properties. Repflags are
    //   set to repflags of destination node.
    // - "object_desc" is used only to test if object is dirty. It contains default
    //   repflags, because repflags are specified per-property, root object's
    //   repflags are always default.

    BitOStream bos;
    Serializer::Description object_desc;
    object_desc.compare_times = true;
    object_desc.current_time = replica_time;
    object_desc.lossy = true;
    object_desc.object_flags = Serializer::Description::OBJECT_REPLICATE;
    Serializer::Description desc( object_desc );
    desc.repflags_yes = repflags_yes;
    desc.repflags_no = repflags_no;

    // Iterate over replica version table - we know something about state of those
    // replica on the replication client.

    ReplicaVersions & versions = node_info.versions;
    ReplicaVersions::iterator version_it;
    for( version_it = versions.begin(); version_it != versions.end(); ++version_it )
        {
        const ObjectId & incomplete_object_id = version_it->first;
        ReplicaVersion & version = version_it->second;

        // Should we updatethis object?

        const ObjectIdSet::iterator replica_it = objects_to_replicate.find( incomplete_object_id );
        if( replica_it == objects_to_replicate.end() )
            {
            // No - destroy the object.

            #ifdef DEBUG_REPLICATION
            if( debug )
                {
                std::ostringstream ss;
                ss << "Destroying replica of object " << incomplete_object_id << " on node " << node_id << ".";
                Global::log_debug( INFREQUENT_STATUS, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            if( !version.is_state_destroying() )
                {
                hope( version.is_state_living() );
                version.set_state_destroying( replica_time );
                }
            message.add_object_to_destroy( incomplete_object_id, version.major_version );
            }
        else
            {
            // Update this object.

            ObjectProperty * const object = gc.find_local_object( incomplete_object_id );
            hope( object );
            hope( !object->is_replica() );

            // Maybe object_id doesn't contain optional ObjectInfo part of
            // ObjectId, but ReplicationUpdateMessage::add_object_to_update()
            // requires the optional part to be valid. So we will ask
            // the object itself for the right object id.

            const ObjectId & full_object_id = object->get_object_id();

            // Note: Since change to migration handling we should not have any records
            // about replicas with different major versions.

            hope( object->get_major_version() == version.major_version );

            // If the replica is in destroying state it means that:
            // - this node created the replica
            // - then the replication request timed out or the object left replication groups
            //   to update, so this node sent replica destroy request
            // - but before the destruction was acked, the object became part of replication
            //   group to update
            // Just switch state back to creating. This ensures that all properties that are
            // not default-constructed become dirty, and all update acks older than current
            // replica_time will be ignored.

            const bool recreate = version.is_state_destroying();
            if( recreate )
                {
                version.set_state_creating( replica_time, object->get_major_version() );
                }

            // Is the object dirty?

            object_desc.ack_time = version.minor_version;
            desc.ack_time = version.minor_version;
            if( object->should_serialize( object_desc ) == Property::SSR_YES )
                {
                // It's dirty, write object update.

                #ifdef DEBUG_REPLICATION

                if( debug )
                    {
                    if( recreate )
                        {
                        std::ostringstream ss;
                        ss << "Re-creating replica of object " << full_object_id <<
                            " on node " << node_id << ".";
                        Global::log_debug( INFREQUENT_STATUS, ss.str() );
                        }
                    else
                        {
                        std::ostringstream ss;
                        ss << "Updating replica of object " << full_object_id <<
                            " on node " << node_id << ".";
                        Global::log_debug( Status::FACILITY_REPLICATION,
                            Status::PRIORITY_LOWEST, ss.str() );
                        }
                    }

                #endif // DEBUG_REPLICATION

                // Serialize the object.

                const Int32 previous_stream_size = bos.byte_size();
                object->bitstr_write( bos, desc );
                message.add_object_to_update( full_object_id, version.major_version );
                balancer.account_for_object_replication
                    ( object, node_id, bos.byte_size() - previous_stream_size );

                #ifdef OPTIMIZE_UNCHANGED_REPLICAS
                updated_groups.add_replication_group( object );
                #endif
                }
            else
                {
                // Not dirty, just record its object id.
                // This should really be optimized somehow (when I'm able to prove
                // that such optimization is really ok).

                #ifdef OPTIMIZE_UNCHANGED_REPLICAS

                unchanged_objects.insert( full_object_id );

                #else // OPTIMIZE_UNCHANGED_REPLICAS

                #ifdef DEBUG_REPLICATION
                if( debug )
                    {
                    std::ostringstream ss;
                    ss << "Sending id of unchanged object " << full_object_id <<
                        " to node " << node_id << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION,
                        Status::PRIORITY_LOWEST, ss.str() );
                    }
                #endif // DEBUG_REPLICATION

                message.add_unchanged_object( full_object_id, object->get_major_version() );

                #endif // OPTIMIZE_UNCHANGED_REPLICAS
                }

            // Remove object from the set. See below.

            objects_to_replicate.erase( replica_it );
            }
        }

    // Contents of objects_to_replicate at this point:
    //   objects to update with no known version on the client
    // So create new replica version entry for each object remaining in objects_to_replicate
    // and send creation request of the replica.

    ObjectIdSet::const_iterator replica_it;
    for( replica_it = objects_to_replicate.begin();
        replica_it != objects_to_replicate.end(); ++replica_it )
        {
        const ObjectId & incomplete_object_id = *replica_it;
        ObjectProperty * const object = gc.find_local_object( incomplete_object_id );
        hope( object );
        hope( !object->is_replica() );

        // Maybe object_id doesn't contain optional #ObjectInfo part of
        // ObjectId, but ReplicationUpdateMessage::add_object_to_update()
        // requires the optional part to be valid. So we will ask
        // the object itself for the right object id.

        const ObjectId & full_object_id = object->get_object_id();

        #ifdef DEBUG_REPLICATION
        if( debug )
            {
            std::ostringstream ss;
            ss << "Creating replica of object " << full_object_id << " on node " << node_id << ".";
            Global::log_debug( INFREQUENT_STATUS, ss.str() );
            }
        #endif // DEBUG_REPLICATION

        // Create new replica version entry.

        ReplicaVersion version( replica_time, object->get_major_version() );
        DEBUG_ONLY( std::pair< ReplicaVersions::iterator ) DEBUG_COMMA DEBUG_ONLY( bool > result = )
                    versions.insert( ReplicaVersions::value_type( full_object_id, version ) );
        hope( result.second );

        // Write object update.

        object_desc.ack_time = version.minor_version;
        desc.ack_time = version.minor_version;
        hope( object->should_serialize( object_desc ) == Property::SSR_YES );

        const Int32 previous_stream_size = bos.byte_size();
        object->bitstr_write( bos, desc );
        message.add_object_to_update( full_object_id, version.major_version );
        balancer.account_for_object_replication
            ( object, node_id, bos.byte_size() - previous_stream_size );
        }

    #ifdef OPTIMIZE_UNCHANGED_REPLICAS

    // For each object in both unchanged_objects and updated_groups:

    for( replica_it = unchanged_objects.begin(); replica_it != unchanged_objects.end(); ++replica_it )
        {
        const ObjectId & incomplete_object_id = *replica_it;
        if( updated_groups.object_ids.find( incomplete_object_id ) != updated_groups.object_ids.end() )
            {
            // Find the object.

            ObjectProperty * const object = gc.find_local_object( incomplete_object_id );
            hope( object );
            hope( !object->is_replica() );

            // Maybe object_id doesn't contain optional #ObjectInfo part of
            // ObjectId, but ReplicationUpdateMessage::add_object_to_update()
            // requires the optional part to be valid. So we will ask
            // the object itself for the right object id.

            const ObjectId & full_object_id = object->get_object_id();

            // Add it to list of unchanged objects.

            #ifdef DEBUG_REPLICATION
            if( debug )
                {
                std::ostringstream ss;
                ss << "Sending id of unchanged object " << full_object_id <<
                    " to node " << node_id << ".";
                Global::log_debug( Status::FACILITY_REPLICATION,
                    Status::PRIORITY_LOWEST, ss.str() );
                }
            #endif // DEBUG_REPLICATION

            message.add_unchanged_object( full_object_id, object->get_major_version() );
            }
        }

    #endif // OPTIMIZE_UNCHANGED_REPLICAS

    // Bind bitostream to message.

    message.objects.bind( bos.get_buffer() );
    }

} // anonymous namespace


bool ReplicaServer::update
    (
    const NodeId &      node_id,
    NodeInfo &          node_info
    )
    {
    // Get repflags of destination node.

    int repflags_yes, repflags_no;
    settings.get_repflags( repflags_yes, repflags_no, node_id );

    // Check if @a node_id is local node.

    const bool is_local_node = ( node_id == System::get_local_node_id() );

    // Gather set of objects to replicate or just perform replica request
    // cleanup if @a is_local_node is @c true.

    const std::size_t old_request_count = node_info.requests.size();
    ReplicationGroupSet replication_group_set( repflags_yes, repflags_no );
    gather_objects_to_replicate( replication_group_set, node_info, is_local_node );
    update_request_count( node_info, old_request_count, node_info.requests.size() );

    // If @a node_id is local, we're done - no messages are sent.
    // FIXME: do the same if node is not connected.

    if( is_local_node )
        {
        return false;
        }

    // Create replication update message.

    const ReplicaTime & replica_time = Global::replica_time();
    ReplicationUpdateMessage message;
    message.replica_time = replica_time;
    message.simulation_time = STime::current_time();
    message.tcp_ack = false;

    // Fill the message.

    const std::size_t old_version_count = node_info.versions.size();
    fill_replication_update_message( message, replication_group_set, node_id,
                node_info, repflags_yes, repflags_no, settings.debug );
    update_replica_count( node_info, old_version_count, node_info.versions.size() );

    // Do not send the message if empty.

    if( message.is_empty() )
        {
        return false;
        }

    // Determine whether UDP or TCP should be used.

    NetworkCommon::StreamType stream = NetworkCommon::STREAM_UNRELIABLE;
    if( message.guess_size() > settings.max_udp_update_size.read() )
        {
        message.tcp_ack = true;
        stream = NetworkCommon::STREAM_RELIABLE;
        node_info.waiting_for_tcp_ack = true;
        node_info.wait_for_ack_time = replica_time;
        }

    // Remember update in backlog.

    node_info.pending_updates.add_update( message );

    // Log.

    #ifdef DEBUG_REPLICATION
    if( settings.debug )
        {
        std::ostringstream ss;
        ss << "Sending replication update via " << ( message.tcp_ack ? "TCP" : "UDP" ) <<
            ": stamp " << message.replica_time << ", time " << message.simulation_time <<
            ", " << message.objects_to_update.size() << " object update(s), " <<
            message.unchanged_objects.size() << " unchanged object(s), " <<
            message.objects_to_destroy.size() << " object destroy(s).";
        Global::log_debug( REGULAR_STATUS, ss.str() );
        }
    #endif // DEBUG_REPLICATION

    // Send the update.

    message.send( node_id, stream );

    // Remember time of this update.

    node_info.update_time = SystemTime::current_time();

    // Update statistics.

    hope( node_info.statistics.get() );
    //node_info.statistics->updates_sent++; updated by backlog automatically
    statistics.updates_sent++;
    if( message.tcp_ack )
        {
        node_info.statistics->updates_sent_tcp++;
        statistics.updates_sent_tcp++;
        }

    return true;
    }


ReplicaServer::NodeInfo::HashMap::iterator ReplicaServer::insert_node
    (
    const NodeId &      node_id
    )
    {
    // Insert new node info.

    std::pair< NodeInfo::HashMap::iterator, bool > result =
        nodes.insert( NodeInfo::HashMap::value_type( node_id, NodeInfo( node_id ) ) );
    hope( result.second );
    update_node_count();

    return result.first;
    }


void ReplicaServer::clear_node
    (
    const NodeId &      node_id
    )
    {
    const NodeInfo::HashMap::iterator node_it = nodes.find( node_id );
    if( node_it != nodes.end() )
        {
        NodeInfo & node_info = node_it->second;

        // Update statistics - request and replica count.

        update_request_count( node_info, node_info.requests.size(), 0 );
        update_replica_count( node_info, node_info.versions.size(), 0 );

        // Clear requests and versions.

        node_info.requests.clear();
        node_info.versions.clear();

        // Clear backlog. This will update statistics automatically.

        node_info.pending_updates.clear();

        // Drop waiting_for_tcp. Leaving it on would block the replication altogether.

        node_info.waiting_for_tcp_ack = false;
        }
    }


void ReplicaServer::erase_node
    (
    const NodeId &      node_id
    )
    {
    // Clear the node info.

    clear_node( node_id );

    // Destroy it and update statistics.

    nodes.erase( node_id );
    update_node_count();
    }


void ReplicaServer::update_node_count()
    {
    const Int32 count = nodes.size();
    if( count > statistics.node_count_max )
        {
        statistics.node_count_max = count;
        }
    statistics.node_count = count;
    }


void ReplicaServer::update_request_count
    (
    NodeInfo &      node_info,
    std::size_t     old_count,
    std::size_t     new_count
    )
    {
    // Update node statistics.

    hope( node_info.statistics.get() );
    const Int32 new_count_i32 = static_cast< Int32 >( new_count );
    if( new_count_i32 > node_info.statistics->request_count_max )
        {
        node_info.statistics->request_count_max = new_count_i32;
        }
    node_info.statistics->request_count = new_count_i32;

    // Update total statistics.

    const Int32 old_count_i32 = static_cast< Int32 >( old_count );
    const Int32 total_count = statistics.request_count + new_count_i32 - old_count_i32;
    if( total_count > statistics.request_count_max )
        {
        statistics.request_count_max = total_count;
        }
    statistics.request_count = total_count;
    }


void ReplicaServer::update_replica_count
    (
    NodeInfo &      node_info,
    std::size_t     old_count,
    std::size_t     new_count
    )
    {
    // Update node statistics.

    hope( node_info.statistics.get() );
    const Int32 new_count_i32 = static_cast< Int32 >( new_count );
    if( new_count_i32 > node_info.statistics->replica_count_max )
        {
        node_info.statistics->replica_count_max = new_count_i32;
        }
    node_info.statistics->replica_count = new_count_i32;

    // Update total statistics.

    const Int32 old_count_i32 = static_cast< Int32 >( old_count );
    const Int32 total_count = statistics.replica_count + new_count_i32 - old_count_i32;
    if( total_count > statistics.replica_count_max )
        {
        statistics.replica_count_max = total_count;
        }
    statistics.replica_count = total_count;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "replica_server_inline.h"
#endif
