/** @file replica_server.h
 * Replication server.
 */

/*
 * begin        : Sat Feb 01 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICA_SERVER_H
#define MASSIV_REPLICA_SERVER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REPLICATION_FWD_H
#include "replication/replication_fwd.h"
#endif

#ifndef MASSIV_REPLICA_REQUEST_H
#include "replication/replica_request.h"
#endif

#ifndef MASSIV_REPLICA_UPDATE_H
#include "replication/replica_update.h"
#endif

#ifndef MASSIV_REPLICA_VERSION_H
#include "replication/replica_version.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_REPLICA_SERVER_STATS_H
#include "replication/replica_server_stats.h"
#endif

#ifndef MASSIV_REPLICATION_SETTINGS_H
#include "replication/replication_settings.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#include <memory>


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * Replication server.
 */
class MASSIV_EXPORTED ReplicaServer
    {
    NO_COPY_CONSTRUCTOR( ReplicaServer );
    NO_ASSIGNMENT_OPERATOR( ReplicaServer );

public:

    /**
     * Information about single replication client node.
     */
    class MASSIV_EXPORTED NodeInfo
        {
        NO_DEFAULT_CONSTRUCTOR( NodeInfo );

    public:

        /**
         * Hash map from NodeId to NodeInfo.
         */
        typedef ext::hash_map< NodeId, NodeInfo,
            NodeId::Hash, NodeId::EqualKey > HashMap;

    public:

        /** Constrctor. */
        explicit NodeInfo
            (
            const NodeId & the_node_id
            );

        /**
         * Copy constructor.
         *
         * @note It is not recommended to copy node infos with non-empty
         *       #requests or #versions.
         */
        NodeInfo
            (
            const NodeInfo & other
            );

        /**
         * Assignment operator.
         *
         * @note It is not recommended to copy node infos with non-empty
         *       #requests or #versions.
         */
        NodeInfo & operator=
            (
            const NodeInfo & other
            );

    public:

        /**
         * Is node info empty? (are both #requests and #versions empty?)
         */
        INLINE bool is_empty() const;

    public:

        NodeId                  node_id;
            /**< Id of node this info is about. */

        std::auto_ptr< ReplicaServerNodeStatistics > statistics;
            /**< Statistics regarding replication to the node. */

        ReplicaRequests         requests;
            /**< Replicas requested by the node. */

        ReplicaUpdateBacklog    pending_updates;
            /**< Information about updates that have not been acked yet. */

        ReplicaVersions         versions;
            /**< States of replicas on the node. */

        bool                    waiting_for_tcp_ack;
            /**< @c true if the most recent update has been sent via TCP instead
                 of UDP and the server is waiting for ack of that update. */

        ReplicaTime             wait_for_ack_time;
            /**< Replica time of the most recent update sent to client node if
                 #waiting_for_tcp_ack is @c true, undefined otherwise. */

        SystemTime              update_time;
            /**< System time of the most recent replication update has been sent. */

        }; // class NodeInfo

public:

    /** 
     * Constructor.
     * @c initialize() must be called then.
     */
    ReplicaServer();

    /**
     * Destructor.
     */
    ~ReplicaServer();

    /**
     * Initialize ReplicaServer.
     * The work can not be done by the constructor because the implementation
     * requires access to properly initialized Globals.
     */
    void initialize();

    /**
     * Shutdowns ReplicaServer.
     * Resets instance state so that it could be reused and reinitialized
     * (by @c initialize()).
     */
    void shutdown();

public:

    /**
     * Create new replication request.
     *
     * @param object
     *      Id of object to replicate. Must be local object.
     * @param destination_node
     *      Id of node to replicate @a object to.
     * @param timeout
     *      Stop replication when simulation time reaches @a timeout.
     *      Use @c STime() to stop replication.
     *
     * @pre @a object_id and @a destination_node are valid ids. This should
     *      be checked by Object "builtins".
     */
    void request_replication
        (
        const ObjectId &    object,
        const NodeId &      destination_node,
        const STime &       timeout
        );

    /**
     * Inform replica server about disconnection of @a node from the simulation.
     *
     * Replication to the node will be stopped and all info about the
     * node will be forgotten.
     *
     * This method @b must @b not be called when connection between two nodes
     * is closed because network thought it would be a good idea, but both nodes
     * remain part of the simulation.. That would result in replication errors
     * if @a node has replicas of objects owned by this node.
     */
    void node_disconnected
        (
        const NodeId &      node
        );

public:

    /**
     * Gather ids of objects that are replicated (either their replica exists
     * somewhere or replication request exists) and their class name matches
     * @c class_name_pattern or they inherit from a parent class whose name
     * matches the pattern (if @c include_subclasses is true).
     * There is a preset limit on maximum number of objects returned.
     */
    void gather_replicated_objects
        (
        ObjectShared::ObjectIdSet & objects,
        const std::string &         class_name_pattern,
        bool                        include_subclasses
        ) const;

    /**
     * Push ids of all nodes that request replicas of @a object to @a result.
     * There is a preset limit on maximum number of nodes returned.
     */
    void gather_replication_nodes
        (
        std::vector< NodeId > &     result,
        const ObjectId &            object
        ) const;

    /**
     * Print information about replication of @a object to stream @a s.
     *
     * Information will be written to two lines.
     * - "Requests to: " and set of node ids followed by request timeouts.
     * - "Replicas on: " and set of node ids followed by object state.
     * .
     * Line with no info will not be printed. Trailing new-line will not be
     * written to the stream.
     *
     * There is a preset limit on maximum number of entries returned.
     */
    void dump_object_info
        (
        std::ostream &      s,
        const ObjectId &    object
        ) const;

public:

    /**
     * Send replication updates if needed.
     */
    void next_tick();

    /**
     * Handle incoming replication ack message.
     */
    void handle_replication_ack_message
        (
        const ReplicationAckMessage & message
        );

public:

    /**
     * Get current replication timestamp.
     */
    INLINE const ReplicaTime & get_replica_time() const;

    /**
     * Advance replication timestamp.
     *
     * @note Calling this method is semi-hack and should be probably done only
     *       in self-tests. Replication timestamp advances automatically when
     *       #next_tick() is called.
     */
    void advance_replica_time();

public:

    /**
     * Check if there are any replication requests of object @a object_id.
     */
    bool has_replica_requests
        (
        const ObjectId & object_id
        ) const;

    /**
     * Handle outgoing migration of @a object_id described by @a desc.
     *
     * When object migrates from server node @c A to server node @c B, server @c B
     * becomes responsible for object's replication and replicas. This means that
     * along the object the following information to must be transferred to @c B:
     * - Requests of object's replication.
     * - List of nodes that may have replicas of the object.
     * .
     * Server @c A will forget all replication requests and information about
     * replicas of the object. That means that server @c A will never try to destroy
     * replicas of objects that migrate away. Server @c B is responsible for that.
     *
     * When object migrates to client node, replication requests are not sent along
     * with it (as client's are not allowed to replicate their objects) and replicas
     * of the object will be destroyed by the server.
     */
    void handle_outgoing_migration
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc,
        const ObjectId &                    object_id
        );

    /**
     * Handle incoming migration of @a object described by @a desc.
     *
     * This method reads information about replication written to @a bis by
     * #handle_outgoing_miration(), registers replication requests and updates
     * replica version info.
     *
     * It's expected that @a bis contains no replication requests if this
     * method is called on client node.
     */
    void handle_incoming_migration
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc,
        const ObjectId &                    object_id
        );

private:

    /**
     * Write requests of object @a object_id replicas to text stream.
     * Forget the requests if @a remove_requests if @c true.
     *
     * @bug No longer used, remove it once everybody is sure it's not needed.
     */
    void text_write_requests
        (
        TextWriter &                        tw,
        const Serializer::Description &     desc,
        const ObjectId &                    object_id,
        bool                                remove_requests
        );

    /**
     * Read requests of object @a object_id replicas from text stream
     * and insert them into the request table.
     *
     * @bug No longer used, remove it once everybody is sure it's not needed.
     */
    void text_read_requests
        (
        TextReader &                        tr,
        const Serializer::Description &     desc,
        const ObjectId &                    object_id
        );

private:

    /**
     * Check if updates should be sent to given node.
     */
    bool should_update
        (
        const NodeId &      node_id,
        const NodeInfo &    node_info
        ) const;

    /**
     * Send replica update to given node.
     *
     * @return @c true if update has been sent, @c false if there were no dirty
     *         objects to send.
     */
    bool update
        (
        const NodeId &      node_id,
        NodeInfo &          node_info
        );
        
private:

    /**
     * Get reference to node info.
     *
     * Creates correctly initialized node info if it does not exist.
     */
    INLINE NodeInfo & access_node
        (
        const NodeId &      node_id
        );

    /**
     * Insert and initialize new node info.
     *
     * @pre No node info for given node if exists.
     */
    NodeInfo::HashMap::iterator insert_node
        (
        const NodeId &      node_id
        );

    /**
     * Clear replication requests, info about replica versions and update backlog
     * for given node.
     *
     * This method will @b not destroy the node info.
     */
    void clear_node
        (
        const NodeId &      node_id
        );

    /**
     * Erase given node.
     *
     * Never destroy node infos directly. Call this method to update statistics properly.
     * This method is no-op if the node does not exist.
     */
    void erase_node
        (
        const NodeId &      node_id
        );

private:

    /**
     * Update node count statistics to current <tt>nodes.size()</tt>.
     */
    void update_node_count();

    /**
     * Update node replication request count statistics.
     *
     * This method may be called both before and after the requests are modified.
     */
    void update_request_count
        (
        NodeInfo &      node_info,
        std::size_t     old_count,
        std::size_t     new_count
        );

    /**
     * Update number of replicas on given node statistics.
     *
     * This method may be called both before and after the versions are modified.
     */
    void update_replica_count
        (
        NodeInfo &      node_info,
        std::size_t     old_count,
        std::size_t     new_count
        );
        
private:

    bool                        initialized;
        /**< Is ReplicaServer initialized? */

    ReplicaTime                 replica_time;
        /**< Current replication timestamp.
             Should be increased whenever update is sent to a replication client. */

    ReplicationSettings         settings;
        /**< Settings. */

    ReplicaServerStatistics     statistics;
        /**< Replication statistics. */

    NodeInfo::HashMap           nodes;
        /**< Information about replication client nodes. */

    }; // class ReplicaServer

/** @} */ // addtogroup replication

        
} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "replica_server_inline.h"
#endif


#endif // MASSIV_REPLICA_SERVER_H
