/** @file replica_update.cpp
 * Replication update information (implementation).
 */

/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "replication/replica_update.h"
#include "replication/replica_version.h"
#include "replication/replica_server_stats.h"
#include "sysevent/replication_msg.h"
#include "global/global.h"
#include <sstream>


namespace Massiv {
namespace Core {


namespace {

    
/**
 * Return <tt>factor * a + ( 1 - factor ) * b</tt>
 */
Float32 lerp
    (
    Float32 a,
    Float32 b,
    Float32 factor
    )
    {
    return factor * a + ( 1.0f - factor ) * b;
    }


/**
 * Lerp used to update ping.
 */
Float32 lerp_ping
    (
    Float32 a,
    Float32 b
    )
    {
    return lerp( a, b, 0.8f );
    }


/**
 * Lerp used to update packet loss.
 */
Float32 lerp_loss
    (
    Float32 a,
    Float32 b
    )
    {
    return lerp( a, b, 0.8f );
    }

    
} // anonymous namespace


// class ReplicaUpdateInfo

void ReplicaUpdateInfo::apply
    (
    ReplicaVersions &   versions,
    const NodeId &      node_id,
    bool                debug
    ) const
    {
    ObjectList::const_iterator object_it;
    for (
        object_it = updated_objects.begin();
        object_it != updated_objects.end();
        ++object_it
        )
        {
        versions.ack( *object_it, replica_time, ReplicaVersions::ACK_UPDATE, node_id, debug );
        }

    for
        (
        object_it = destroyed_objects.begin();
        object_it != destroyed_objects.end();
        ++object_it
        )
        {
        versions.ack( *object_it, replica_time, ReplicaVersions::ACK_DESTROY, node_id, debug );
        }
    }


void ReplicaUpdateInfo::initialize
    (
    const ReplicationUpdateMessage & message
    )
    {
    hope( updated_objects.empty() && destroyed_objects.empty() );

    // Copy timestamp.

    replica_time = message.replica_time;

    // Copy list of updated objects.

    const ReplicationUpdateMessage::ObjectList & update_objs = message.objects_to_update;
    ReplicationUpdateMessage::ObjectList::const_iterator object_it;
    for( object_it = update_objs.begin(); object_it != update_objs.end(); ++object_it )
        {
        updated_objects.push_back( object_it->object_id );
        }

    // Copy list of destroyed objects.

    const ReplicationUpdateMessage::ObjectList & destroy_objs = message.objects_to_destroy;
    for( object_it = destroy_objs.begin(); object_it != destroy_objs.end(); ++object_it )
        {
        destroyed_objects.push_back( object_it->object_id );
        }

    // Set system time.

    update_time = SystemTime::current_time();
    }


// class ReplicaUpdateBacklog

ReplicaUpdateBacklog::ReplicaUpdateBacklog
    (
    const NodeId &                  the_node_id
    ) : node_id( the_node_id ),
        backlog_size( 0 ),
        stats( NULL )
    {
    }


ReplicaUpdateBacklog::ReplicaUpdateBacklog
    (
    const ReplicaUpdateBacklog & other
    ) : node_id( other.node_id ),
        backlog( other.backlog ),
        backlog_size( other.backlog_size ),
        stats( other.stats )
    {
    hope( stats );
    }


ReplicaUpdateBacklog & ReplicaUpdateBacklog::operator=
    (
    const ReplicaUpdateBacklog & other
    )
    {
    node_id = other.node_id;
    backlog = other.backlog;
    backlog_size = other.backlog_size;
    stats = other.stats;
    hope( stats );
    return *this;
    }


void ReplicaUpdateBacklog::set_statistics
    (
    ReplicaServerNodeStatistics * statistics
    )
    {
    hope( statistics );
    stats = statistics;
    }


void ReplicaUpdateBacklog::clear()
    {
    hope( stats );
    backlog.clear();
    backlog_size = 0;
    stats->backlog_length = 0;
    }


void ReplicaUpdateBacklog::add_update
    (
    const ReplicationUpdateMessage & message
    )
    {
    hope( is_empty() || backlog.back().replica_time < message.replica_time );
    hope( stats );

    // Delete oldest entry if backlog is full.

    if( backlog_size == MAX_BACKLOG_LENGTH )
        {
        std::ostringstream ss;
        ss << "Replication update info backlog for node " << node_id << " is full. "
              "Discarding entry " << SystemTime::current_time() - backlog.front().update_time <<
              " seconds old, with replica time " << backlog.front().replica_time << ".";
        Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
        backlog.pop_front();
        backlog_size--;
        stats->packet_loss = lerp_loss( stats->packet_loss, 1.0f );
        stats->backlog_overflows++;
        }

    // Add new entry.

    backlog.push_back( ReplicaUpdateInfo() );
    backlog_size++;
    ReplicaUpdateInfo & info = backlog.back();
    info.initialize( message );

    // Update statistics.

    stats->backlog_length = static_cast< Int32 >( backlog_size );
    if( stats->backlog_length > stats->backlog_length_max )
        {
        stats->backlog_length_max = stats->backlog_length;
        }
    if( stats->packet_loss > stats->packet_loss_max )
        {
        stats->packet_loss_max = stats->packet_loss;
        }
    stats->updates_sent++;
    }


bool ReplicaUpdateBacklog::ack
    (
    const ReplicationAckMessage &   message,
    ReplicaVersions &               versions,
    bool                            debug
    )
    {
    hope( stats );
    stats->acks_received++;

    // Find info about the update in backlog.
    
    const ReplicaTime & ack_time = message.replica_time;
    Backlog::const_iterator it = backlog.begin();
    while( ( it != backlog.end() ) && ( it->replica_time < ack_time ) )
        {
        ++it;
        }

    if( ( it == backlog.end() ) || ( it->replica_time != ack_time ) )
        {
        std::ostringstream ss;
        ss << "Received replication update ACK from node " << node_id << " with replica time " <<
              ack_time << ", but there is no information about the update in backlog. ";
        Status::Priority priority;
        if( it == backlog.begin() )
            {
            ss << "Probably ACK of old forgotten update.";
            priority = Status::PRIORITY_LOW;
            }
        else
            {
            ss << "Bogus replica time - older updates still have not been acked.";
            priority = Status::PRIORITY_HIGHEST;
            }
        Global::log_warning( Status::FACILITY_REPLICATION, priority, ss.str() );
        stats->acks_ignored++;

        return false;
        }

    // Apply the ack.

    hope( it->replica_time == ack_time );
    it->apply( versions, node_id, debug );

    // Update ping.

    const Float32 ack_ping = static_cast< Float32 >
                ( ( SystemTime::current_time() - it->update_time ).get_value() );
    stats->ping = lerp_ping( stats->ping, ack_ping );

    // Remove backlog entries up to @a it and update packet_loss.

    while( backlog.front().replica_time != ack_time )
        {
        std::ostringstream ss;
        ss << "Removing entry " << backlog.front().replica_time << " from backlog for node " <<
              node_id << ", because ACK of newer update has been received. ACK of the entry has "
              "been probably lost.";
        Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_LOW, ss.str() );
        backlog.pop_front();
        backlog_size--;
        stats->packet_loss = lerp_loss( stats->packet_loss, 1.0f );
        }
    hope( backlog.front().replica_time == ack_time );
    backlog.pop_front();
    backlog_size--;
    stats->packet_loss = lerp_loss( stats->packet_loss, 0.0f );

    // Update statistics.

    stats->backlog_length = static_cast< Int32 >( backlog_size );
    if( stats->packet_loss > stats->packet_loss_max )
        {
        stats->packet_loss_max = stats->packet_loss;
        }
    if( stats->ping > stats->ping_max )
        {
        stats->ping_max = stats->ping;
        }

    return true;
    }


Float32 ReplicaUpdateBacklog::update_period
    (
    Float32 min_period,
    Float32 max_period
    ) const
    {
    hope( stats );

    // Update as fast as possible if there are no pending updates.

    if( is_empty() )
        {
        stats->update_period = min_period;
        return min_period;
        }

    // Compute time elapsed since oldest unacked update has been sent.

    const Float32 wait_time = static_cast< Float32 >
                ( ( SystemTime::current_time() - backlog.front().update_time ).get_value() );

    // Compute period factors:
    // - Packet loss factor is relatively small. I don't think that high packet
    //   loss always mean that bandwidth is saturated.
    // - I assume that high pings mean low bandwidth. Do not slow down until ping
    //   is larger than 150ms, always use max_period if ping is above 1150ms.
    // - Weird wait factor means "if oldest update has not been acked for
    //   wwf * ping seconds, something weird happened. High periods will be used
    //   in that case, probably max_period unless both ping and packet loss are low.

    const Float32 loss_factor =
                ( stats->packet_loss > 0.1f ) ? 0.3f * ( stats->packet_loss - 0.1f ) : 0.0f;
    const Float32 ping_factor =
                ( stats->ping > 0.15f ) ? ( stats->ping - 0.15f ) : 0.0f;
    const Float32 weird_wait_factor =
                2.0f + ( ( loss_factor > 0.2f ) ? 10.0f * ( loss_factor - 0.2f ) : 0.0f );
    stats->wait_limit = weird_wait_factor * stats->ping;
    const Float32 wait_factor =
                ( wait_time > stats->wait_limit ) ? ( wait_time - stats->ping ) : 0.0f;

    // Compute update period.

    const Float32 factor = loss_factor + ping_factor + wait_factor;
    if( factor <= 0.0f )
        {
        stats->update_period = min_period;
        }
    else if( factor >= 1.0f )
        {
        stats->update_period = max_period;
        }
    else
        {
        stats->update_period = lerp( min_period, max_period, factor );
        }
    return stats->update_period;
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "replica_update_inline.h"
#endif
