/** @file replica_update.h
 * Replication update information.
 */

/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICA_UPDATE_H
#define MASSIV_REPLICA_UPDATE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_REPLICATION_FWD_H
#include "replication/replication_fwd.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#include <list>
#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * Information about single replication update remembered by server that sent the update.
 */
class ReplicaUpdateInfo
    {
public:

    /**
     * List of object ids.
     */
    typedef std::vector< ObjectId > ObjectList;

public:

    /**
     * Initialize update information from update message.
     *
     * @note List of @a unchanged_objects is not remembered, because it's not
     *       needed - there was no change between last known ack and current
     *       update, we can ignore acks of such objects.
     */
    void initialize
        (
        const ReplicationUpdateMessage & message
        );

    /**
     * Apply update information to information about state of replicas.
     *
     * Parameters @a node_id and @a debug are used only to generate debugging log messages.
     */
    void apply
        (
        ReplicaVersions &   versions,
        const NodeId &      node_id,
        bool                debug
        ) const;
    
public:

    ReplicaTime     replica_time;
        /**< Replica time of the update. */

    SystemTime      update_time;
        /**< System time when the update has been sent. */

    ObjectList      updated_objects;
        /**< Objects updated in the update. */

    ObjectList      destroyed_objects;
        /**< Objects destroyed in the update. */
    
    }; // class ReplicaUpdateInfo


/**
 * Replication update backlog for single node.
 * Cyclic buffer of ReplicaUpdateInfo objects.
 */
class ReplicaUpdateBacklog
    {
    NO_DEFAULT_CONSTRUCTOR( ReplicaUpdateBacklog );

private:

    /**
     * Maximum backlog length.
     */
    enum
        {
        MAX_BACKLOG_LENGTH = 64
        };

public:

    /**
     * List of update infos, sorted by ReplicaUpdateInfo::replica_time.
     */
    typedef std::list< ReplicaUpdateInfo > Backlog;

public:

    /**
     * Constructor.
     *
     * Constructed backlog will @b not be usable until #set_statistics() is called.
     */
    explicit ReplicaUpdateBacklog
        (
        const NodeId & the_node_id
        );

    /**
     * Copy constructor.
     */
    ReplicaUpdateBacklog
        (
        const ReplicaUpdateBacklog & other
        );

    /**
     * Assignment operator.
     */
    ReplicaUpdateBacklog & operator=
        (
        const ReplicaUpdateBacklog & other
        );

    /**
     * Set pointer to statistics to use.
     *
     * This method must be called after constructor.
     */
    void set_statistics
        (
        ReplicaServerNodeStatistics * statistics
        );

    /**
     * Clear backlog.
     *
     * @post is_empty() will return @c true. Associated node id and statistics
     *       will not be changed.
     */
    void clear();

public:

    /**
     * Add information about replication update to backlog.
     */
    void add_update
        (
        const ReplicationUpdateMessage & message
        );

    /**
     * Handle ACK message.
     *
     * @return @c true if ack has been processed, @c false if it has been ignored
     *         because it was old or invalid.
     */
    bool ack
        (
        const ReplicationAckMessage &   message,
        ReplicaVersions &               versions,
        bool                            debug
        );

    /**
     * Determine replication update period.
     *
     * @return Time in seconds between two consective replication updates to this node.
     */
    Float32 update_period
        (
        Float32 min_period,
        Float32 max_period
        ) const;

    /**
     * Is backlog empty?
     */
    INLINE bool is_empty() const;

private:

    NodeId                          node_id;
        /**< Id of destination node. */

    Backlog                         backlog;
        /**< List of update infos, sorted by replica time. */

    std::size_t                     backlog_size;
        /**< Current backlog size. */

    ReplicaServerNodeStatistics *   stats;
        /**< Statistics of replica updating. */

    }; // class ReplicaUpdateBacklog

/** @} */ // addtogroup replication


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "replica_update_inline.h"
#endif


#endif // MASSIV_REPLICA_UPDATE_H
