/** @file replica_version.cpp
 * Structure remembering versions of replicas on a single node.
 */

/*
 * begin        : Fri Jan 31 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "property/property_fwd.h" // must be included soon because of msvc6
#include "replication/replica_version.h"
#include "utility/debug.h"
#include "global/global.h"
#include <sstream>


namespace Massiv {
namespace Core {


// class ReplicaVersion

void ReplicaVersion::ack_update
    (
    const ReplicaTime & time,
    const NodeId &      node_id,
    bool                debug
    )
    {
    switch( state )
        {
        case STATE_CREATING:
        case STATE_REPLICATING:
            {
            hope( create_time != ReplicaTime::INVALID );
            if( time < create_time )
                {
                // Update ACK from "previous life".

                #ifdef DEBUG_REPLICATION
                if( debug )
                    {
                    std::ostringstream ss;
                    ss << "Received obsolete update ACK from " << node_id << " with timestamp " <<
                          time << " for replica with create_time " << create_time.get_value() << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION
                }
            else
                {
                // Expected update ACK.

                minor_version = time;
                state = STATE_REPLICATING;
                }
            break;
            }

        case STATE_DESTROYING:
            {
            // Update ACK for dying replica. It should be obsolete ack.

            hope( destroy_time != ReplicaTime::INVALID );
            if( time >= destroy_time )
                {
                // Should never happen. We never sent such update.
                // FIXME: Exception instead of warning? Get objectid somewhere.

                std::ostringstream ss;
                ss << "Received unexpected update ACK from " << node_id << " with timestamp " <<
                      time << " for dying replica with destroy_time " << destroy_time << ".";
                Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGHEST, ss.str() );
                }
            else
                {
                // May happen and we could update #version, but it would be
                // pretty useless because this replica will either change
                // to STATE_CREATING or will be destroyed.

                #ifdef DEBUG_REPLICATION
                if( debug )
                    {
                    std::ostringstream ss;
                    ss << "Received obsolete update ACK from " << node_id << " with timestamp " <<
                          time << " for dying replica with destroy_time " << destroy_time << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION
                }
            break;
            }

        default:
            {
            panic( "Unexpected replica state." );
            }
        }
    }


bool ReplicaVersion::ack_destroy
    (
    const ReplicaTime & time,
    const NodeId &      node_id,
    bool                debug
    )
    {
    switch( state )
        {
        case STATE_CREATING:
            {
            hope( create_time != ReplicaTime::INVALID );
            if( time < create_time )
                {
                // Obsolete death ACK.

                #ifdef DEBUG_REPLICATION
                if( debug )
                    {
                    std::ostringstream ss;
                    ss << "Received obsolete death ACK from " << node_id << " with timestamp " <<
                          time << " for 'creating' replica with create_time " << create_time << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION
                }
            else
                {
                // Unexpected death ACK.

                std::ostringstream ss;
                ss << "Received unexpected death ACK from " << node_id << " with timestamp " <<
                      time << " for 'creating' replica with create_time " << create_time << ".";
                Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGHEST, ss.str() );
                }
            break;
            }

        case STATE_REPLICATING:
            {
            // Invalid death ACK.

            std::ostringstream ss;
            ss << "Received unexpected death ACK from " << node_id << " with timestamp " <<
                  time << " for replica with create_time " << create_time << ".";
            Global::log_warning( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGHEST, ss.str() );
            break;
            }

        case STATE_DESTROYING:
            {
            hope( destroy_time != ReplicaTime::INVALID );
            if( time < destroy_time )
                {
                // Obsolete death ACK.

                #ifdef DEBUG_REPLICATION
                if( debug )
                    {
                    std::ostringstream ss;
                    ss << "Received obsolete death ACK from " << node_id << " with timestamp " <<
                          time << " for 'dying' replica with destroy_time " << destroy_time << ".";
                    Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_HIGH, ss.str() );
                    }
                #endif // DEBUG_REPLICATION
                }
            else
                {
                // Death acked.

                return true;
                }
            break;
            }

        default:
            {
            panic( "Unexpected replica state." );
            }
        }

    return false;
    }


std::string ReplicaVersion::state_string
    (
    State state
    )
    {
    switch( state )
        {
        case STATE_CREATING:
            return "creating";
        case STATE_REPLICATING:
            return "replicating";
        case STATE_DESTROYING:
            return "destroying";
        case STATE_INVALID:
            break;
        }
    return "invalid_state";
    }


// class ReplicaVersions

void ReplicaVersions::ack
    (
    const ObjectId &    object_id,
    const ReplicaTime & time,
    AckType             type,
    const NodeId &      node_id,
    bool                debug
    )
    {
    hope( type == ACK_UPDATE || type == ACK_DESTROY );
    
    // Find the version.

    iterator version_it = find( object_id );
    if( version_it == end() )
        {
        // ACK of non-existing entry. Both may happen when object migrates
        // away, so log them with low priority.

        if( type == ACK_UPDATE )
            {
            #ifdef DEBUG_REPLICATION
            if( debug )
                {
                std::ostringstream ss;
                ss << "Received invalid update ACK from " << node_id <<
                      " for non-existing object version, object id " << object_id << ".";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_LOW, ss.str() );
                }
            #endif // DEBUG_REPLICATION
            }
        else
            {
            #ifdef DEBUG_REPLICATION
            if( debug )
                {
                std::ostringstream ss;
                ss << "Received invalid destroy ACK from " << node_id <<
                      " for non-existing object version, object id " << object_id << ".";
                Global::log_debug( Status::FACILITY_REPLICATION, Status::PRIORITY_LOW, ss.str() );
                }
            #endif // DEBUG_REPLICATION
            }
        }
    else
        {
        ReplicaVersion & version = version_it->second;

        // Process the ack.

        if( type == ACK_UPDATE )
            {
            version.ack_update( time, node_id, debug );
            }
        else
            {
            const bool destroy = version.ack_destroy( time, node_id, debug );
            if( destroy )
                {
                erase( version_it );
                }
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "replica_version_inline.h"
#endif
