/** @file replica_version.h
 * Structure remembering versions of replicas on a single node.
 */

/*
 * begin        : Fri Jan 31 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICA_VERSION_H
#define MASSIV_REPLICA_VERSION_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_REPLICATION_FWD_H
#include "replication/replication_fwd.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_REPLICATION_FLAGS_H
#include "replication/replication_flags.h"
#endif

#ifndef MASSIV_STYPE_VLI_H
#include "serializer/stype_vli.h" // ObjectVersion
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * Information about replica version.
 */
class ReplicaVersion
    {

public:

    /**
     * Hash map from ObjectId to ReplicaVersion.
     */
    typedef ext::hash_map< ObjectId, ReplicaVersion,
        ObjectId::Hash, ObjectId::EqualKey > HashMap;

    /**
     * State of object replication.
     */
    enum State
        {
        STATE_CREATING,     ///< Waiting for ACK of first replica updates.
        STATE_REPLICATING,  ///< Using standard replication delta-update protocol.
        STATE_DESTROYING,   ///< Waiting for ACK of replica destruction.
        STATE_INVALID,      ///< Invalid replica state. Never used.
        }; // enum State

public:

    /**
     * Construct invalid replica version info.
     */
    INLINE ReplicaVersion();

    /**
     * Construct replica version with #create_time set to given @a time.
     * and object #major_version set to @a version.
     */
    INLINE ReplicaVersion
        (
        const ReplicaTime & time,
        ObjectVersion       version
        );

public:

    /**
     * Is state @c STATE_CREATING or @c STATE_REPLICATING?
     */
    INLINE bool is_state_living() const;

    /**
     * Is state @c STATE_DESTROYING?
     */
    INLINE bool is_state_destroying() const;

    /**
     * Switch state to @c STATE_CREATING.
     * Sets #create_time to given @a time, #major_version to @a version.
     * and #minor_version to @c INVALID (nothing is assumed about object
     * state on replication destination).
     *
     * @note The only way to switch state to @c STATE_REPLICATING is by
     *       receiving an update ack.
     *
     * @note Minor version must be set to @c INVALID to ensure that complete
     *       replica will be sent in next update. This is actually the only
     *       difference between #STATE_CREATING and #STATE_REPLICATING, they
     *       could be simply merged into single state, but I think that two
     *       different states make things a bit easier to understand.
     */
    INLINE void set_state_creating
        (
        const ReplicaTime & time,
        ObjectVersion       version
        );
    
    /**
     * Switch state to @c STATE_DESTROYING.
     */
    INLINE void set_state_destroying
        (
        const ReplicaTime & time
        );

public:

    /**
     * ACK of replica update received.
     */
    void ack_update
        (
        const ReplicaTime & time,
        const NodeId &      node_id,
        bool                debug
        );

    /**
     * ACK of replica destruction received.
     *
     * @return
     *      @c true if replica version information can be removed from the
     *      cache, @c false otherwise.
     */
    bool ack_destroy
        (
        const ReplicaTime & time,
        const NodeId &      node_id,
        bool                debug
        );

public:

    /**
     * Convert State to string.
     */
    static std::string state_string
        (
        State state
        );

public:

    State           state;
        /**< Replica state. */

    ReplicaTime     minor_version;
        /**< Acked value of replica minor version. It means that destination node's
             replica is not older than that sent when local replication timestamp has
             been equal to #minor_version. */

    ObjectVersion   major_version;
        /**< Replica major version. Unlike #minor_version, this is not acked value
             but value sent in most recent replica update (which should be equal to
             major version when most recent switch to @c STATE_CREATING happened).
             It's used when sending "destroy object" update. */

    ReplicaTime     create_time;
        /**< When state changed to STATE_CREATING. */

    ReplicaTime     destroy_time;
        /**< When state changed to STATE_DESTROYING. */

    }; // class ReplicaVersion

    
/**
 * ReplicaVersions remembers version of replicas on single node.
 *
 * @note Due to changes to delta-update protocol, this structure is no longer
 * a "cache" - it never forgets information about replica version unless
 * replica is destroyed.
 */
class ReplicaVersions : public ReplicaVersion::HashMap
    {
public:

    /**
     * Type of ACK.
     */
    enum AckType
        {
        ACK_UPDATE,
        ACK_DESTROY
        }; // enum AckType

public:

    /** Default constructor */
    INLINE ReplicaVersions();

public:

    /**
     * Convert AckType to string.
     */
    INLINE std::string ack_name
        (
        AckType             type
        ) const;

    /**
     * Handle object update/destruction ack.
     *
     * @param object_id Id of acked object.
     * @param time Timestamp of the ack.
     * @param type Type of ack - ACK_UPDATE or ACK_DESTROY.
     * @param node_id Id of node that sent this ack.
     * @param debug Log debugging messages?
     */
    void ack
        (
        const ObjectId &    object_id,
        const ReplicaTime & time,
        AckType             type,
        const NodeId &      node_id,
        bool                debug
        );

    }; // class ReplicaVersions

/** @} */ // addtogroup replication


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "replica_version_inline.h"
#endif


#endif // MASSIV_REPLICA_VERSION_H
