/** @file replication_flags.h
 * Replication flags.
 */

/*
 * begin        : Sat May 18 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICATION_FLAGS_H
#define MASSIV_REPLICATION_FLAGS_H


namespace Massiv {
namespace Core {


/** @addtogroup replication
 * @{
 */

/**
 * Replication flags.
 * These are used to select subset of properties when migrating or replicating
 * an object and to define migration/replication object groups.
 *
 * Destination of replication is internally specified by nodeid (special object
 * pointer in "script"). Property will be replicated only if its flags
 * (@c repflags in IDL) match replication masks defined by destination node.
 * These masks specify which bits in replication flags must be set to 1, which
 * must be set to 0, and which are ignored.
 *
 * Migration/replication group is defined in similar way. Objects @b A and @b B
 * belong to the same group iff replication flags (called @c ptr_repflags in
 * IDL) of object pointer owned by @b A and pointing to @b B match  replication
 * masks of destination node.
 *
 * REPFLAGS_MIGRATE is the only hardwired flag. Every property should have
 * this bit set. When migrating, replication masks specified per-node are
 * ignored, @c REPMASK_MIGRATE_YES and @c REPMASK_MIGRATE_NO are used instead.
 * These masks accept every property with @c REPFLAGS_MIGRATE set, ie. every
 * property. Use @c REPFLAGS_MIGRATE flag to define migration groups. Objects
 * in these groups will always be owned by the same server, all pointers with
 * @c ptr_repflags set to <tt>REPFLAGS_MIGRATE | something</tt> will be always
 * local.
 *
 * Warning: Replication group must be ALWAYS subset of migration group (for
 * obvious reasons), so @c ptr_repflags may be either zero or
 * @c REPFLAGS_MIGRATE bit must be set to 1.
 *
 * Replication of replicas is not allowed, otherwise rules for "well-defined"
 * replication masks would be way too complex.
 */
enum ReplicationFlags
    {
    REPFLAGS_NONE               = 0x00,
        /**< No bit set. */
    REPFLAGS_MIGRATE            = 0x01,
        /**< Migrate this property/... */
    REPFLAGS_BIT1               = 0x02,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT2               = 0x04,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT3               = 0x08,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT4               = 0x10,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT5               = 0x20,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT6               = 0x40,
        /**< Gamelib-defined meaning. */
    REPFLAGS_BIT7               = 0x80,
        /**< Gamelib-defined meaning. */

    REPFLAGS_BITS               = 8,
        /**< Size of replication flags.
             Useful when flags are stored in a bitfield. */

    REPFLAGS_PROPERTY_DEFAULT   = REPFLAGS_MIGRATE,
        /**< Default for properties. */
    REPFLAGS_POINTER_DEFAULT    = REPFLAGS_NONE,
        /**< Default for pointers. */

    REPMASK_MIGRATE_YES         = REPFLAGS_MIGRATE,
        /**< Mask used instead of @c repflags_yes when migrating. */
    REPMASK_MIGRATE_NO          = REPFLAGS_NONE,
        /**< Mask used instead of @c repflags_no when migrating. */

    }; // enum ReplicationFlags

/** @} */ // addtogroup replication


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_REPLICATION_FLAGS_H
