/** @file replication_settings.cpp
 * Replication settings.
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "replication/replication_settings.h"
#include "utility/debug.h"
#include "replication/replication_flags.h"
#include <sstream>


namespace Massiv {
namespace Core {


ReplicationSettings::ReplicationSettings()
      : Settings( "ObjectManagement/Replication" ),
        #ifdef OPTION_DEBUG
        debug( access_boolean( "debug", true ) ),
        #else
        debug( access_boolean( "debug", false ) ),
        #endif
        server_repflags_yes( access_integer( "server_repflags_yes", REPMASK_MIGRATE_YES ) ),
        server_repflags_no( access_integer( "server_repflags_no", REPMASK_MIGRATE_NO ) ),
        client_repflags_yes( access_integer( "client_repflags_yes", REPMASK_MIGRATE_YES ) ),
        client_repflags_no( access_integer( "client_repflags_no", REPMASK_MIGRATE_NO ) ),
        min_client_update_period( access_float( "min_client_update_period", 0.2f ) ),
        max_client_update_period( access_float( "max_client_update_period", 1.0f ) ),
        min_server_update_period( access_float( "min_server_update_period", 0.5f ) ),
        max_server_update_period( access_float( "max_server_update_period", 2.0f ) ),
        max_udp_update_size( access_integer( "max_udp_update_size", 1500 ) )
    {
    }


void ReplicationSettings::get_repflags
    (
    int &           repflags_yes,
    int &           repflags_no,
    const NodeId &  node
    ) const
    {
    if( node.is_type( NodeId::TYPE_SERVER ) )
        {
        repflags_yes = server_repflags_yes;
        repflags_no = server_repflags_no;
        }
    else if( node.is_type( NodeId::TYPE_CLIENT ) )
        {
        repflags_yes = client_repflags_yes;
        repflags_no = client_repflags_no;
        }
    else
        {
        // Panic. Such nodes should not be part of the simulation.

        std::ostringstream ss;
        ss << "Can't determine repflags for node " << node << ", unsupported type.";
        panic( ss.str().c_str() );
        /*NOTREACHED*/
        }
    }


void ReplicationSettings::get_update_periods
    (
    Float32 &       min_period,
    Float32 &       max_period,
    const NodeId &  node
    ) const
    {
    if( node.is_type( NodeId::TYPE_SERVER ) )
        {
        min_period = min_server_update_period;
        max_period = max_server_update_period;
        }
    else if( node.is_type( NodeId::TYPE_CLIENT ) )
        {
        min_period = min_client_update_period;
        max_period = max_client_update_period;
        }
    else
        {
        // Panic. Such nodes should not be part of the simulation.

        std::ostringstream ss;
        ss << "Can't determine update periods for node " << node << ", unsupported type.";
        panic( ss.str().c_str() );
        /*NOTREACHED*/
        }
    }

    
} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "replication_settings_inline.h"
#endif
