/** @file serializer.h
 * Serialization.
 */

/*
 * begin        : Thu Feb 21 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SERIALIZER_H
#define MASSIV_SERIALIZER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_REPLICATION_FLAGS_H
#include "replication/replication_flags.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup serialization
 * @{
 */

/**
 * Serialization methods and structures.
 */
namespace Serializer {


/**
 * Serialization description.
 *
 * This structure is argument of bitstream and textual serialization functions.
 * It's used to select which object attributes should be serialized.
 *
 * When writing, the owning object decides which owned objects should be
 * serialized and then it calls their @c bitstr_write() or @c text_write()
 * method. Simple data types should ignore the #Description in their
 * serializers and simply write themselves to specified stream. There are
 * several reasons for this:
 * - Many serializable types (all STypes) can't decide whether they should
 *   be serialized or not. Information about replication flags and dirty
 *   times is not stored in STypes, it's known only by their owner.
 * - Complex serializable types (managed objects, arrays, etc.) must know
 *   which owned objects will be written to the stream during their
 *   serialization.
 */
struct MASSIV_EXPORTED Description
    {
public:

    /**
     * Flags describing which object attributes should be serialized.
     */
    enum ObjectFlags
        {
        OBJECTF_PROPERTIES              = 0x0001,
            ///< Serialize all properties recursively.
        OBJECTF_EVENTS                  = 0x0002,
            /**< Serialize event queue.
                 Global schedule will be updated when reading events. */
        OBJECTF_NAME                    = 0x0004,
            ///< Serialize object name.
        OBJECTF_VERSION                 = 0x0008,
            ///< Serialize major object version.
        OBJECTF_INCREASE_VERSION        = 0x0010 | OBJECTF_VERSION,
            ///< Increase major version when reading.
        OBJECTF_WRITE_ZERO_VERSION      = 0x0020 | OBJECTF_VERSION,
            /**< Semi-hack - same as @c OBJECTF_VERSION when reading,
                 but always write zero version. */
        OBJECTF_REPLICATION             = 0x0040,
            /**< Serialize replication info.
                 This flag may be set only during migration, #destination_object
                 must be set. #ReplicaServer::handle_outgoing_migration() will be
                 called when writing and #ReplicaServer::handle_incoming_miration()
                 when reading. See documentation of those methods for more info. */
        OBJECTF_GC_FLAGS                = 0x0080,
            /**< Serialize object flags utilized by the Garbage Collector. */

        // Standard flag combinations.

        OBJECT_DEFAULT          = OBJECTF_PROPERTIES,
            ///< Basic serialization. Used when serializing member objects.

        OBJECT_MIGRATE          = OBJECTF_PROPERTIES | OBJECTF_EVENTS |
                                  OBJECTF_NAME | OBJECTF_INCREASE_VERSION |
                                  OBJECTF_REPLICATION | OBJECTF_GC_FLAGS,
            ///< Object migration.

        OBJECT_ARCHIVE          = OBJECTF_PROPERTIES | OBJECTF_EVENTS |
                                  OBJECTF_NAME | OBJECTF_WRITE_ZERO_VERSION |
                                  OBJECTF_GC_FLAGS,
            /**< Object archivation.
                 @note Objects serialized to archive must be readable using
                       @c OBJECT_SWAP flags. That's why object version is
                       written as zero - it's not required by archive, but it's
                       required in real swap. */

        OBJECT_SWAP             = OBJECTF_PROPERTIES | OBJECTF_EVENTS |
                                  OBJECTF_NAME | OBJECTF_VERSION |
                                  OBJECTF_GC_FLAGS,
            /**< Object swapping.
                 @note Object major version must be swapped because it's stored
                       in the #Object, replication requests are stored in replica
                       server structures and is not swapped.
                 @todo Do not allow swapping of objects with replica requests. */

        OBJECT_REPLICATE        = OBJECTF_PROPERTIES,
            /**< Object replication.
                 @note Object major version is sent separately.
                 @note Object name is not serialized. */

        }; // enum ObjectFlags

public:

    bool                compare_times;
        /**< Should dirty time be compared to #ack_time?

             If @c true, both #ack_time and #current_time are valid. These times are used
             when receiver of the object serialization already has its older version.

             #compare_times is set to @c true only when writing replication updates. */

    ReplicaTime         ack_time;
        /**< ACK timestamp used during replication.

             When writing, #ack_time will be equal to timestamp of the most recent acked
             update sent to replication client. This means that all properties with dirty
             time lower or equal to #ack_time are up-to-date on the client. Only properties
             with dirty time greater than #ack_time should be serialized.

             When reading, this is set to update timestamp and is pretty useless.

             #ack_time should be @c INVALID unless replicating. */

    ReplicaTime         current_time;
        /**< Current replica time of local node.

             Valid iff #compare_times is @c true. Property dirty time is guaranteed to be
             lower or equal to #current_time. All future writes to the property will be done
             with timestamp greater than #current_time. This can be used to do lazy updates
             of dirty timestamps of complex data structures (see PArray). */

    int                 repflags_yes : REPFLAGS_BITS;
        /**< Only properties with these flags set to 1 will be replicated.
             See @c ReplicationFlags in @c replication_flags.h for more info. */

    int                 repflags_no : REPFLAGS_BITS;
        /**< Only properties with these flags set to 0 will be replicated. */

    bool                lossy : 1;
        /**< Is serialization lossy?
             Some information may be lost during bitstream serialization if true (floats
             can be quantized, etc.). Textual serialization should be always lossless. */

    int                 object_flags;
        /**< Flags describing which parts of root objects should be serialized.
             Use values defined in ObjectFlags enum.
             Ignored by non-root object (member objects, objects owned by
             containers) and non-object properties. */

    NodeId              source_node;
        /**< Node id of source node.
             Set only when bitstr_read() is called when unpacking migration
             message. #Source_node is set form SystemMessage::sender, not
             SystemMessage::initiator. */

    ObjectId            destination_object;
        /**< Object id of destination object if this is migration write serialization. */

public:

    /**
     * Default constructor.
     * Sets @c time to @c INVALID_TIME, replication masks to migration defaults
     * (see @c ReplicationFlags for more info), @c lossy to false.
     */
    INLINE Description();

    /**
     * Check if replication flags match.
     */
    INLINE bool replication_flags_match
        (
        int repflags
        ) const;

    }; // struct Description


} // namespace Massiv::Core::Serializer

/** @} */ // addtogroup serialization


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "serializer/serializer_inline.h"
#endif


#endif // MASSIV_SERIALIZER_H
