/** @file serializer_float.cpp
 * Floating-point serializers.
 */

/*
 * begin        : Sun Apr 21 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "utility/math.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "serializer/serializer.h"
#include "serializer/serializer_float.h"
#include "serializer/bitint_masks.h"
#include "net/bitistream.h"
#include "net/bitostream.h"


namespace Massiv {
namespace Core {
namespace Serializer {


void BitStrFloat::read
    (
    BitIStream &        bis,
    const Description & desc,
    Float32 &           result,
    Float32             min_value,
    Float32             max_value,
    int                 bits
    )
    {
    if( !desc.lossy )
        {
        read( bis, result );
        return;
        }

    hope( bits > 0 );
    hope( bits <= 32 );
    hope( max_value > min_value );

    const Float32 qmax = static_cast< Float32 >( uint32_masks[ bits ] );

    result = static_cast< Float32 >
        ( bis.read_int( bits ) * ( max_value - min_value ) / qmax + min_value );
    }


void BitStrFloat::read
    (
    BitIStream &        bis,
    const Description & desc,
    Float64 &           result,
    Float64             min_value,
    Float64             max_value,
    int                 bits
    )
    {
    if( !desc.lossy )
        {
        read( bis, result );
        return;
        }

    hope( bits > 0 );
    hope( bits <= 32 );
    hope( max_value > min_value );

    const Float64 qmax = static_cast< Float64 >( uint32_masks[ bits ] );

    result = static_cast< Float64 >
        ( bis.read_int( bits ) * ( max_value - min_value ) / qmax + min_value );
    }


void BitStrFloat::write
    (
    BitOStream &        bos,
    const Description & desc,
    Float32             value,
    Float32             min_value,
    Float32             max_value,
    int                 bits
    )
    {
    if( !desc.lossy )
        {
        write( bos, value );
        return;
        }

    hope( bits > 0 );
    hope( bits <= 32 );
    hope( max_value > min_value );

    const UInt32  qmax_int = uint32_masks[ bits ];
    const Float32 qmax_float = static_cast< Float32 >( qmax_int );

    // round( ( x - min ) * qmax / ( max - min ) )
    const Float32 v = FloatTraits< Float32 >::round
        ( ( value - min_value ) * qmax_float / ( max_value - min_value ) );

    UInt32 i;
    if( v <= 0.0f )
        {
        i = 0;
        }
    else if( v >= qmax_float )
        {
        i = qmax_int;
        }
    else
        {
        i = static_cast< UInt32 >( v );
        hope( i <= qmax_int );
        }

    bos.write_int( i, bits );
    }


void BitStrFloat::write
    (
    BitOStream &        bos,
    const Description & desc,
    Float64             value,
    Float64             min_value,
    Float64             max_value,
    int                 bits
    )
    {
    if( !desc.lossy )
        {
        write( bos, value );
        return;
        }

    hope( bits > 0 );
    hope( bits <= 32 );
    hope( max_value > min_value );

    const UInt32  qmax_int = uint32_masks[ bits ];
    const Float64 qmax_float = static_cast< Float64 >( qmax_int );

    // round( ( x - min ) * qmax / ( max - min ) )
    const Float64 v = FloatTraits< Float64 >::round
        ( ( value - min_value ) * qmax_float / ( max_value - min_value ) );

    UInt32 i;
    if( v <= 0.0 )
        {
        i = 0;
        }
    else if( v >= qmax_float )
        {
        i = qmax_int;
        }
    else
        {
        i = static_cast< UInt32 >( v );
        hope( i <= qmax_int );
        }

    bos.write_int( i, bits );
    }


void TextFloat::read
    (
    TextReader &    tr,
    Float32 &       value
    )
    {
    std::istream &  is = tr.line();
    //hope( is );
    is >> value;
    if( is.fail() )
        {
        if( is.eof() )
            throw ExUnexpectedEol( tr.unexpected_eol() );
        else
            throw ExUnexpectedToken( tr.unexpected_token( "floating-point number" ) );
        }
    }


void TextFloat::read
    (
    TextReader &    tr,
    Float64 &       value
    )
    {
    std::istream &  is = tr.line();
    //hope( is );
    is >> value;
    if( is.fail() )
        {
        if( is.eof() )
            throw ExUnexpectedEol( tr.unexpected_eol() );
        else
            throw ExUnexpectedToken( tr.unexpected_token( "floating-point number" ) );
        }
    }


} // namespace Massiv::Core::Serializer
} // namespace Massiv::Core
} // namespace Massiv


#ifndef EXPOSE_SERIALIZERS
#include "serializer/serializer_float_inline.h"
#endif
