/** @file stype_enum.h
 * Serializable lightweight enumeration type.
 */

/*
 * begin        : Mon Apr 7 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STYPE_ENUM_H
#define MASSIV_STYPE_ENUM_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STYPE_FWD_H
#include "serializer/stype_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_VLI_H
#include "serializer/serializer_vli.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#include <string>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup stype
 * @{
 */

/** @class SEnum
 * Serializable enumeration type.
 * @c EnumType is expected to be a native enumeration type with explicitly
 * defined mapping to integers (due to safe serialization).
 */
template <class EnumType>
class SEnum
    {
private:

    typedef Serializer::Description Description;    // Because of MSVC6.

public:
    
    /// Corresponding native type.
    typedef EnumType        NativeType;
    /// Corresponding property type.
    typedef PEnum<EnumType> PropertyType;

public:

    /** Default constructor. */
    SEnum()
        : value( static_cast<EnumType>( 0 ) )
        {
        }

    /** Constructor. */
    SEnum
        (
        EnumType v
        ) : value( v )
        {
        }

    /** Conversion to EnumType. */
    operator EnumType() const
        {
        return read();
        }

    /** Conversion to integer. */
    operator int() const
        {
        return read();
        }
    
    /** Operator=. */
    SEnum & operator=
        (
        EnumType v
        )
        {
        return write( v );
        }

    /** Returns value. */
    EnumType read() const
        {
        return value;
        }

    /** Changes value. */
    SEnum & write
        (
        EnumType v
        )
        {
        value = v;
        return *this;
        }

public:

    /**
     * Serialize to bitstream.
     */
    void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        Serializer::BitStrVli::write( bos, read() );
        }

    /**
     * Read from bitstream.
     */
    void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        int temp;
        Serializer::BitStrVli::read( bis, temp );
        write( static_cast<EnumType>( temp ) );
        }

    /**
     * Serialize to text stream.
     */
    void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextVli::write( tw, read() );
        }

    /**
     * Read from text stream.
     */
    void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        int temp;
        Serializer::TextVli::read( tr, temp );
        write( static_cast<EnumType>( temp ) );
        }
    
    /**
     * Print SType state (debug).
     */
    std::string to_string() const
        {
        std::ostringstream ss;
        ss << read();
        return ss.str();
        }

    /**
     * Return hash code.
     */
    int hash() const
        {
        return static_cast<int>( read() );
        }

    /**
     * Compute memory size consumed by SType.
     */
    int memory_size() const
        {
        return sizeof( *this );
        }

    /**
     * Guess bitstream serialization size (in bits).
     */
    int bitstream_size() const
        {
        return 8;
        }

private:

    EnumType    value;
        /**< Current value. */

    }; // class SEnum

/** @} */ // addtogroup stype


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STYPE_ENUM_H
