/** @file stype_float.h
 * Serializable lightweight floats. Wrappers around *Float* native types.
 */

/*
 * begin        : Sat Mar 16 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STYPE_FLOAT_H
#define MASSIV_STYPE_FLOAT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STYPE_FWD_H
#include "serializer/stype_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_FLOAT_H
#include "serializer/serializer_float.h"
#endif


#include <string>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup stype
 * @{
 */

/** @class SFloatBase
 * Common floating-point stuff. Boring. Private - use either SFloat<...> or Float.
 * @todo math.h-like methods? (abs(), ...)
 */
template <class Type>
class SFloatBase
    {
private:

    typedef Serializer::Description Description;    // Because of MSVC6.

public:

    /** Constructor. */
    INLINE SFloatBase()
        : value( Type() )
        {
        }

    /** Copy contructor. */
    INLINE SFloatBase
        (
        const SFloatBase & f
        ) : value( f.value )
        {
        }

    /** Constructor. */
    INLINE SFloatBase
        (
        Type v
        ) : value( v )
        {
        }

public:

    /** Write value. */
    INLINE void write
        (
        Type v
        )
        {
        value = v;
        }

    /** Read value. */
    INLINE Type read() const
        {
        return value;
        }

    /** Operator=. */
    INLINE SFloatBase & operator=
        (
        Type v
        )
        {
        write( v );
        return *this;
        }

    /** Read value. */
    INLINE operator Type() const
        {
        return read();
        }
    
public:

    // l-values

    /** Operator+=. */
    INLINE SFloatBase & operator+=
        (
        Type v
        )
        {
        value += v;
        return *this;
        }

    /** Operator-=. */
    INLINE SFloatBase & operator-=
        (
        Type v
        )
        {
        value -= v;
        return *this;
        }

    /** Operator*=. */
    INLINE SFloatBase & operator*=
        (
        Type v
        )
        {
        value *= v;
        return *this;
        }

    /** Operator/=. */
    INLINE SFloatBase & operator/=
        (
        Type v
        )
        {
        value /= v;
        return *this;
        }

public:

    // r-values

    /** Operator++, preincrement. */
    INLINE Type operator++()
        {
        return ++value;
        }

    /** Operator++, postincrement. */
    INLINE Type operator++( int )
        {
        return value++;
        }

    /** Operator--, predecrement. */
    INLINE Type operator--()
        {
        return --value;
        }

    /** Operator--, postdecrement. */
    INLINE Type operator--( int )
        {
        return value--;
        }

public:

    /**
     * Serialize to text stream.
     */
    INLINE void text_write
        (
        TextWriter &                    tw,
        const Description &
        ) const
        {
        Serializer::TextFloat::write( tw, read() );
        }

    /**
     * Read from text stream.
     */
    INLINE void text_read
        (
        TextReader &                    tr,
        const Description &
        )
        {
        Type temp;
        Serializer::TextFloat::read( tr, temp );
        write( temp );
        }
    
    /**
     * Print SType state (debug).
     */
    INLINE std::string to_string() const
        {
        std::ostringstream ss;
        ss << read();
        return ss.str();
        }

    /**
     * Return hash code.
     */
    INLINE int hash() const
        {
        return static_cast<int>( read() );
        }

    /**
     * Compute memory size consumed by SType.
     */
    INLINE int memory_size() const
        {
        return sizeof( *this );
        }

private:

    Type value;
        /**< Value. */

    }; // class SFloatBase


/** @class SFloatQ
 * Serializable floating-point.
 * Quantizes specified range to integer (1 to 32 bits) during lossy bitstream
 * replication. It behaves the same way as SFloat during losless or text
 * serialization - no quantization is done.
 *
 * The range and number of bits to quantize to is specified by @a Traits
 * structure. It should define three values: @a min_value, @a max_value and
 * @a bits. Values lower or equal to @a min_value will be transformed to 0,
 * values greater or equal to @a max_value will be transformed to (2 ^ @a bits - 1).
 *
 * @a Type should be Float32 or Float64.
 * @a Traits::bits should be >= 1 and <= 32.
 * @a Traits::min_value should be < @a Traits::max_value.
 *
 * Example - quantize angles between -PI and +PI to 10 bits:
 *
 * @code
 * struct MyFloatQTraits
 *     {
 *     static const Float32 min_value;
 *     static const Float32 max_value;
 *     enum { bits = 10 };
 *     };
 *
 * const Float32 MyFloatQTraits::min_value = -MATHF_PI;
 * const Float32 MyFloatQTraits::max_value = MATHF_PI;
 *
 * typedef SFloatQ< Float32, MyFloatQTraits > MyFloatQ;
 * @endcode
 */
template< class Type, class Traits >
class SFloatQ : public SFloatBase< Type >
    {
private:

    /// Superclass.
    typedef SFloatBase< Type >      Super;

    /// Hack to prevent msvc6 internal errors.
    typedef Serializer::Description Description;

public:
    
    /// Corresponding native type.
    typedef Type                        NativeType;

    /// Corresponding property type.
    typedef PFloatQ< Type, Traits >     PropertyType;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE SFloatQ()
        {
        }

    /** Constructor. */
    INLINE SFloatQ
        (
        Type v
        ) : Super( v )
        {
        }

    /** Operator=. */
    INLINE SFloatQ & operator=
        (
        Type v
        )
        {
        Super::operator=( v );
        return *this;
        }

public:

    /**
     * Serialize to bitstream.
     */
    INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description & desc
        ) const
        {
        Type value = read();
        Serializer::BitStrFloat::write
            (
            bos,
            desc,
            value,
            Traits::min_value,
            Traits::max_value,
            Traits::bits
            );
        }

    /**
     * Read from bitstream.
     */
    INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description & desc
        )
        {
        Type temp;
        Serializer::BitStrFloat::read
            (
            bis,
            desc,
            temp,
            Traits::min_value,
            Traits::max_value,
            Traits::bits
            );
        write( temp );
        }
    
    /**
     * Guess bitstream serialization size (in bits).
     */
    INLINE int bitstream_size() const
        {
        return Traits::bits;
        }

    }; // class SFloatQ


/** @class SFloat
 * Serializable floating-point.
 * @a Type should be Float32 or Float64.
 * No transformations applied.
 */
template <class Type>
class SFloat : public SFloatBase<Type>
    {
private:

    typedef SFloatBase<Type>        Super;
    typedef Serializer::Description Description;    // Because of MSVC6.

public:
    
    /// Corresponding native type.
    typedef Type            NativeType;
    /// Corresponding property type.
    typedef PFloat<Type>    PropertyType;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE SFloat()
        {
        }

    /** Constructor. */
    INLINE SFloat
        (
        Type v
        ) : SFloatBase<Type>( v )
        {
        }

    /** Operator=. */
    INLINE SFloat & operator=
        (
        Type v
        )
        {
        SFloatBase<Type>::operator=( v );
        return *this;
        }

public:

    /**
     * Serialize to bitstream.
     */
    INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        Serializer::BitStrFloat::write( bos, read() );
        }

    /**
     * Read from bitstream.
     */
    INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        Type temp;
        Serializer::BitStrFloat::read( bis, temp );
        write( temp );
        }
    
    /**
     * Guess bitstream serialization size (in bits).
     */
    INLINE int bitstream_size() const
        {
        return sizeof( Type ) * 8;
        }

    }; // class SFloat

/** @} */ // addtogroup stype


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STYPE_FLOAT_H
