/** @file stype_fwd.h
 * Forward declarations of light-weight serializable types (STypes).
 * STypes are either used as (internal) core types that need be serializable
 * or as light-weight serializable elements held by container properties
 * instead of corresponding element properties (@see PValueArray).
 * Note that any property is serializable and can be used in the same way as
 * the corresponding stype. However that is ineffecient since property
 * is always "heavier" than a SType.
 * Not all properties have to be paired with a SType. This is in those
 * cases where property lightening has no use (like PValueArray).
 * STypes have to implement additional interface, similar to Properties:
 *  @c memory_size(), @c bitstream_size(), @c print_state()
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STYPE_FWD_H
#define MASSIV_STYPE_FWD_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

// We must include this monstrosity only for DefaultAttributeEventHandlers
// TODO: split attribute.h, move it together with color.h and vector23.h to
// another directory.
#ifndef MASSIV_ATTRIBUTE_H
#include "property/attribute.h"
#endif


namespace Massiv {
namespace Core {


// stype_bool.h

class SBool;

// stype_float.h

template< class Type, class Traits >
class SFloatQ;

template< class Type >
class SFloat;

// stype_int.h

template< class Type, int bits >
class SInt;

// stype_vli.h

template< class Type >
class SVli;

// stype_string.h

class SString;

// stype_time.h

class STime;

// stype_enum.h

template< class EnumType >
class SEnum;

// property/vector23.h

template
    <
    int NumDimensions,
    class SType = SFloat< Float32 >,
    class Base  = DefaultAttributeEventHandlers
    >
class Vector23;

// property/color.h

template< class Base = DefaultAttributeEventHandlers >
class Color;

// property/orientation.h

template< class Base = DefaultAttributeEventHandlers >
class Orientation;


// Shortcut typedefs.

/** @addtogroup stype
 * @{
 */

/** @name Integral STypes with Fixed Bitsize.
 * All bits will be written during serialization.
 */
//@{
typedef SInt< Int32, 32 >       SInt32;     /**< Signed 32-bit integer. */
typedef SInt< UInt32, 32 >      SUInt32;    /**< Unsigned 32-bit integer. */
typedef SInt< Int64, 64 >       SInt64;     /**< Signed 64-bit integer. */
typedef SInt< UInt64, 64 >      SUInt64;    /**< Unsigned 64-bit integer. */
//@}


/** @name Integral STypes with Variable-length Bitstream Encoding.
 * Integral SType, uses variable-length encoding when serializing to bitstream.
 */
//@{
typedef SVli< Int32 >           SVlInt32;   /**< Signed 32-bit integer. */
typedef SVli< UInt32 >          SVlUInt32;  /**< Unsigned 32-bit integer. */
typedef SVli< Int64 >           SVlInt64;   /**< Signed 64-bit integer. */
typedef SVli< UInt64 >          SVlUInt64;  /**< Unsigned 64-bit integer. */
//@} 


/** @name Floating-pointe STypes. */
//@{
typedef SFloat< Float32 >       SFloat32;   /**< Single precision (32-bit) float. */
typedef SFloat< Float64 >       SFloat64;   /**< Double precision (64-bit) float. */
//@}


/** @name Math Vector STypes. */
//@{
typedef Vector23< 2, SFloat32 > SVector2;   /**< 2D single precision vector. */
typedef Vector23< 3, SFloat32 > SVector3;   /**< 3D single precision vector. */
//@}


/** Color SType. */
typedef Color<>                 SColor;


/** Angular orientation SType. */
typedef Orientation<>           SOrientation;

/** @} */ // addtogroup stype


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STYPE_FWD_H
