/** @file stype_int.h
 * Serializable lightweight integers. Wrappers around *Int* native types.
 */

/*
 * begin        : Sat Mar 16 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STYPE_INT_H
#define MASSIV_STYPE_INT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STYPE_FWD_H
#include "serializer/stype_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#ifndef MASSIV_SERIALIZER_INT_H
#include "serializer/serializer_int.h"
#endif

#include <string>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup stype
 * @{
 */

/** @class SIntBase
 * Common integer stuff. Boring. Private - use either SInt<...> or Int.
 */
template <class Type>
class SIntBase
    {
private:

    typedef Serializer::Description Description;    // Because of MSVC6.

public:

    /** Constructor. */
    INLINE SIntBase() :
        value( Type() )
        {
        }

    /** Copy contructor. */
    INLINE SIntBase
        (
        const SIntBase & i
        ) : value( i.value )
        {
        }

    /** Constructor. */
    INLINE SIntBase
        (
        Type v
        ) : value( v )
        {
        }

public:

    /** Write value. */
    INLINE void write
        (
        Type v
        )
        {
        value = v;
        }

    /** Read value. */
    INLINE Type read() const
        {
        return value;
        }

    /** Operator=. */
    INLINE SIntBase & operator=
        (
        Type v
        )
        {
        write( v );
        return *this;
        }

    /** Read value. */
    INLINE operator Type() const
        {
        return read();
        }
    
public:

    // l-values

    /** Operator+=. */
    INLINE SIntBase & operator+=
        (
        Type v
        )
        {
        value += v;
        return *this;
        }

    /** Operator-=. */
    INLINE SIntBase & operator-=
        (
        Type v
        )
        {
        value -= v;
        return *this;
        }

    /** Operator*=. */
    INLINE SIntBase & operator*=
        (
        Type v
        )
        {
        value *= v;
        return *this;
        }

    /** Operator/=. */
    INLINE SIntBase & operator/=
        (
        Type v
        )
        {
        value /= v;
        return *this;
        }

    /** Operator%=. */
    INLINE SIntBase & operator%=
        (
        Type v
        )
        {
        value %= v;
        return *this;
        }

    /** Operator&=. */
    INLINE SIntBase & operator&=
        (
        Type v
        )
        {
        value &= v;
        return *this;
        }

    /** Operator|=. */
    INLINE SIntBase & operator|=
        (
        Type v
        )
        {
        value |= v;
        return *this;
        }

    /** Operator^=. */
    INLINE SIntBase & operator^=
        (
        Type v
        )
        {
        value ^= v;
        return *this;
        }

    /** Operator<<=. */
    INLINE SIntBase & operator<<=
        (
        int c
        )
        {
        value <<= c;
        return *this;
        }

    /** Operator>>=. */
    INLINE SIntBase & operator>>=
        (
        int c
        )
        {
        value >>= c;
        return *this;
        }

public:

    // r-values

    /** Operator++, preincrement. */
    INLINE Type operator++()
        {
        return ++value;
        }

    /** Operator++, postincrement. */
    INLINE Type operator++( int )
        {
        return value++;
        }

    /** Operator--, predecrement. */
    INLINE Type operator--()
        {
        return --value;
        }

    /** Operator--, postdecrement. */
    INLINE Type operator--( int )
        {
        return value--;
        }

public:
            
    /**
     * Print SType state (debug).
     */
    INLINE std::string to_string() const
        {
        std::ostringstream ss;
        ss << read();
        return ss.str();
        }

    /**
     * Return hash code.
     */
    INLINE int hash() const
        {
        return static_cast<int>( read() );
        }

    /**
     * Compute memory size consumed by SType.
     */
    INLINE int memory_size() const
        {
        return sizeof( *this );
        }

private:

    Type value;
        /**< Value. */

    }; // class SIntBase


/** @class SInt
 * Serializable integer.
 * @a Type should be one of Int32, UInt32, Int64, UInt64.
 * @a bits should >= 1 and <= 32 (Int32, UInt32) or <= 64 (Int64, UInt64).
 * Always truncated to specified number of bits (even when migrating/archiving).
 */
template <class Type, int bits>
class SInt : public SIntBase<Type>
    {
private:

    typedef SIntBase<Type>          Super;
    typedef Serializer::Description Description;    // Because of MSVC6.

public:

    /// Corresponding native type.
    typedef Type                NativeType;
    /// Corresponding property type.
    typedef PInt<Type, bits>    PropertyType;

public:
    
    using Super::read;
    using Super::write;

public:

    /** Default constructor. */
    INLINE SInt()
        {
        }

    /** Constructor. */
    INLINE SInt
        (
        Type v
        ) : SIntBase<Type>( v )
        {
        }

    /** Operator=. */
    INLINE SInt & operator=
        (
        Type v
        )
        {
        SIntBase<Type>::operator=( v );
        return *this;
        }

public:

    /**
     * Serialize to bitstream.
     */
    INLINE void bitstr_write
        (
        BitOStream &        bos,
        const Description &
        ) const
        {
        // Note: Truncating even if !lossy
        Serializer::BitStrInt::write( bos, read(), bits );
        }

    /**
     * Read from bitstream.
     */
    INLINE void bitstr_read
        (
        BitIStream &        bis,
        const Description &
        )
        {
        Type temp;
        Serializer::BitStrInt::read( bis, temp, bits );
        write( temp );
        }

    /**
     * Serialize to text stream.
     */
    INLINE void text_write
        (
        TextWriter &        tw,
        const Description &
        ) const
        {
        Serializer::TextInt::write( tw, read(), bits );
        }

    /**
     * Read from text stream.
     */
    INLINE void text_read
        (
        TextReader &        tr,
        const Description &
        )
        {
        Type temp;
        Serializer::TextInt::read( tr, temp, bits );
        write( temp );
        }

    /**
     * Guess bitstream serialization size.
     */
    INLINE int bitstream_size() const
        {
        return bits;
        }
    
    }; // class SInt

/** @} */ // addtogroup stype


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_STYPE_INT_H
