/** @file stype_time.h
 * Serializable "simulation" ("game") time.
 */

/*
 * begin        : Fri Apr 5 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_STYPE_TIME_H
#define MASSIV_STYPE_TIME_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_VIRTUAL_TIME_H
#include "utility/virtual_time.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_PROPERTY_FWD_H
#include "property/property_fwd.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup stype
 * @{
 */

/** @class STime
 * Serializable simulation time.
 * Lightweight version of PTime.
 *
 * @ingroup time
 */
class MASSIV_EXPORTED STime : public VirtualTime<Float64>
    {
public:

    typedef VirtualTime<Float64>    Super;
    typedef Float64                 TimeType;
    typedef Super::Less             Less;
    typedef Super::EqualTo          EqualTo;

public:

    /// Corresponding native type.
    typedef TimeType    NativeType;
    /// Corresponding property type.
    typedef PTime       PropertyType;

public:

    /** Default constructor. */
    INLINE STime();

    /** Constructor. */
    INLINE STime
        (
        TimeType time
        );

    /** Constructor. */
    INLINE STime
        (
        const Super & time
        );

    /** Copy constructor. */
    INLINE STime
        (
        const STime & t
        );

    /** Operator=. */
    INLINE STime & operator=
        (
        const Super & t
        );

    /** Compares times. */
    INLINE bool operator==
        (
        const Super & time
        ) const;
    
    /** Compares times. */
    INLINE bool operator!=
        (
        const Super & time
        ) const;

    /** Returns current time. */
    static STime current_time();

    /**
     * Returns time that represents "ASAP" (as soon as possible),
     * equal to STime().
     */
    static INLINE STime asap();
    
    /** Returns time that represents one minute. */
    static INLINE STime minute();
    
    /** Returns time that represents one hour. */
    static INLINE STime hour();
    
    /** Returns time that represents one day. */
    static INLINE STime day();
    
    /** Returns time that represents one week. */
    static INLINE STime week();

public:

    /**
     * Serialize to bitstream.
     */
    INLINE void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description &
        ) const;

    /**
     * Read from bitstream.
     */
    INLINE void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description &
        );

    /**
     * Serialize to text stream.
     * Writes TimeType using integer serializer.
     */
    INLINE void text_write
        (
        TextWriter &                    tw,
        const Serializer::Description &
        ) const;

    /**
     * Read from text stream.
     */
    INLINE void text_read
        (
        TextReader &                    tr,
        const Serializer::Description &
        );
    
    /**
     * Print SType state (debug).
     */
    INLINE std::string to_string() const;

    /**
     * Return hash code.
     */
    INLINE int hash() const;

    /**
     * Compute memory size consumed by SType.
     */
    INLINE int memory_size() const;

    /**
     * Guess bitstream serialization size (in bits).
     */
    INLINE int bitstream_size() const;

    }; // class STime

/** @} */ // addtogroup stype


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "stype_time_inline.h"
#endif


#endif // MASSIV_STYPE_TIME_H
