/** @file exception.cpp
 * Exception class implementation.
 * @see exception.h
 */

/*
 * begin        : 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "status/exception.h"
#include "global/global.h"
#include "thread/thread.h"
#include "system/system.h"
#include <ostream>
#include <sstream>
#include <ctime>


namespace Massiv {
namespace Core {


#ifdef OPTION_DEBUG

/** @def EXCEPTION_STACK_BACKTRACE
 * Log stack backtrace along with exception info whenever exception is thrown.
 * This should be always disabled unless @c OPTION_DEBUG is defined.
 * Warning: enabling this may generate a lot of garbage in logs.
 */
#define EXCEPTION_STACK_BACKTRACE

#endif // OPTION_DEBUG


namespace {

static StackFrameEntry *    g_last_stack_frame_entry[ ThreadManager::MAX_THREADS ];
    /**< The last stack-frame that was tagged in the given thread. */

}
    
    
// Implementation of Exception
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

Exception::Exception
    (
    const std::string & message,
    int                 flags
    ) : facility( Status::FACILITY_ANY ),
        description( message ),
#ifdef EXCEPTION_STACK_BACKTRACE
        call_stack_history
            (
            flags_on( flags, NO_STACK_TRACE ) ?
            "" : System::print_stack_trace()
            ),
#endif
        flags( flags )
    {
    log_this();
    }


Exception::Exception
    (
    const Status::Facility  facility,
    const std::string &     message,
    int                     flags
    ) : facility( facility ),
        description( message ),
#ifdef EXCEPTION_STACK_BACKTRACE
        call_stack_history( System::print_stack_trace() ),
#endif
        flags( flags )
    {
    log_this();
    }


Exception::Exception
    (
    const Exception &   exception
    ) : facility( exception.facility ),
        description( exception.description ),
        call_stack_history( exception.call_stack_history ),
        flags( exception.flags )
    {
    }


Exception::~Exception() throw()
    {
    }


const char * Exception::what() const throw()
    {
    return description.c_str();
    }


Status::Facility Exception::get_facility() const
    {
    return facility;
    }


const std::string & Exception::get_call_stack_history() const
    {
    return call_stack_history;
    }


void Exception::log_this()
    {
    if( !Global::g_log_exceptions )
        {
        // Exception logging must be disabled during the "first" phase
        // of system startup when logger can not be used yet.

        return;
        }
    
    if( flags_on( flags, NO_LOG ) )
        {
        // Do not log this exception.

        return;
        }

    ::time_t            current_time;
    std::ostringstream  ss;

    time( &current_time );
    ss  << "*** Exception on " << ::ctime( &current_time ) << '\n'
        << description << '\n';
    
    #ifdef EXCEPTION_STACK_BACKTRACE

    ss  << "Call-stack history:\n" << call_stack_history;

    #endif

    Global::log
        (
        Status( facility, Status::TYPE_ERROR, ( flags & LOW_PRIORITY ) ?
            Status::PRIORITY_LOW : Status::PRIORITY_HIGHEST ),
        ss.str()
        );
    }


// Implementation of StackFrameEntry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
StackFrameEntry::StackFrameEntry
    (
    const char * tag
    ) throw() : 
        tag( tag )
    {
    if( !Global::stack_tracer_available() )
        {
        UInt32 index = Thread::get_current_thread_index();
        prev = g_last_stack_frame_entry[ index ];
        g_last_stack_frame_entry[ index ] = this;
        }
    }


StackFrameEntry::~StackFrameEntry() throw()
    {
    if( !Global::stack_tracer_available() )
        {
        UInt32 index = Thread::get_current_thread_index();
        g_last_stack_frame_entry[ index ] = prev;
        }
    }


std::string StackFrameEntry::print_registered_frames()
    {
    if( !Global::stack_tracer_available() )
        {
        UInt32              index = Thread::get_current_thread_index();
        StackFrameEntry *   entry = g_last_stack_frame_entry[ index ];
        std::ostringstream  ss;

        // Print registered StackFrameEntries.

        std::size_t count = 0;
        while( entry != NULL && count < 30 )
            {
            ss << entry->tag << '\n';
            entry = entry->prev;
            count++;
            }

        if( count == 30 )
            {
            ss << "...more stack frame entries follow, stopped at 30th...";
            }
        else
            {
            ss << "<root>";
            }

        return ss.str();
        }
    else
        {
        return "<not_available>";
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
