/** @file exception.h
 * Exception base class and the exception hierarchy definition.
 */

/*
 * begin        : 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_EXCEPTION_H
#define MASSIV_EXCEPTION_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATUS_H
#include "status/status.h"
#endif

#include <stdexcept>


namespace Massiv {
namespace Core {


/** @addtogroup exception
 * @{
 */

/**
 * Generic exception.
 * Root for massiv-defined exception hierarchy.
 */
class MASSIV_EXPORTED Exception : public std::exception
    {
private:
        
    NO_ASSIGNMENT_OPERATOR( Exception );

public:

    /**
     * Exception flags.
     */
    enum Flags
        {
        LOW_PRIORITY    = 0x01,
            /**< Log this exception with low priority. */

        NO_LOG          = 0x02,
            /**< Do not log this exception at all. */

        NO_STACK_TRACE  = 0x04
            /**< Do not do stack trace. */

        }; // enum Flags

public:

    /**
     * Create exception.
     * Exception message will be automatically sent to the logger.
     *
     * @param message The error message.
     * @param flags Combination of Flags.
     *
     * @note You should rather use the second constructor to define
     * facility.
     */
    Exception
        (
        const std::string & message,
        int                 flags = 0x0
        );

    /**
     * Create exception.
     * Exception message will be automatically sent to the logger.
     *
     * @param facility Defines facility this exception relates to.
     * @param message The error message.
     * @param logging_disabled If true, the logger will ignore this 
     *      exception.
     */
    Exception
        (
        const Status::Facility  facility,
        const std::string &     message,
        int                     flags = 0x0
        );

    /**
     * Copy constructor.
     */
    Exception
        (
        const Exception &   exception
        );

    /**
     * Destructor.
     */
    virtual ~Exception() throw();
    
public:

    /**
     * Textual exception identification.
     * @return Returns pointer to char * string containing the
     * identification of the exception.
     */
    virtual const char * what() const throw();

    /**
     * Returns exception facility.
     */
    Status::Facility get_facility() const;
    
    /**
     * Returns call-stack history relative to the time when
     * the exception was thrown.
     * This can be used to localize the source of the exception.
     */
    const std::string & get_call_stack_history() const;

private:

    /**
     * Sends the exception to logger.
     */
    void log_this();
    
protected:

    Status::Facility    facility;
        /**< Exception facility. */
    std::string         description;
        /**< Exception description. */
    std::string         call_stack_history;
        /**< Call-stack history at the time when the exception was created. */
    int                 flags;
        /**< Combination of Flags. */
    }; // class Exception


/** 
 * StackFrameEntry is a helper class that aids debugging exceptions
 * and program control-flow.
 * The instances should always be created on the program stack.
 * Each StackFrameEntry holds client-supplied string that "tags" the frame.
 * Since active StackFrameEntries are linked together we are able to
 * determine which stack frames are active and print their tags.
 * This is used to trace back call-stack history when an exception is thrown,
 * for example.
 * StackFrameEntries should be used to tag "major checkpoints" only
 * (as they are not free).
 *
 * If target architecture supports native stack tracing (see StackTracer)
 * under given compilation mode (ie. if OPTION_DEBUG is defined),
 * @c StackFrameEntries are ignored.
 *
 * @code
 * void f()
 *  {
 *  StackFrameEntry frame( "::f()" );
 *  throw Exception( "Something went wrong!" );
 *  }
 * void g()
 *  {
 *  StackFrameEntry frame( "::g()" );
 *  g();
 *  }
 * void main()
 *  {
 *  StackFrameEntry frame( "::main()" );
 *  try
 *      {
 *      g();
 *      }
 *  catch( Exception & e )
 *      {
 *      std::cout << e.get_call_stack_history();
 *          // ::f()
 *          // ::g()
 *          // ::main()
 *      }
 *  }
 * @endcode
 */
class MASSIV_EXPORTED StackFrameEntry
    {
public:

    /**
     * Creates StackFrameEntry and associates it with the current
     * frame. 
     * 
     * @param tag A pointer to a constant C-style string which will
     *      tag the stack frame. The string must stay valid until
     *      the frame deactivation (the tag is not copied).
     */   
    StackFrameEntry
        (
        const char * tag
        ) throw();

    /**
     * Deassociates entry from the current stack frame.
     */
    ~StackFrameEntry() throw();
    
    /**
     * Prints registered stack frames.
     * Only stack frames that were tagged by @c StackFrameEntry are
     * displayed.     
     */
    static std::string print_registered_frames();

private:
        
    const char *                tag;
        /**< Pointer to a user-defined tag. */
    StackFrameEntry *           prev;
        /**< Previous stack-frame that was tagged.
             NULL if this is the top-most frame. */
    }; // class StackFrameEntry


/**@def TAG_STACK_FRAME
 * Tags stack frame by @c tag. @c tag is supposed to be a string literal.
 */
#define TAG_STACK_FRAME( tag ) \
    ::Massiv::Core::StackFrameEntry stack_frame( tag )


/**@def MASSIV_CORE_GENERIC_EXCEPTION
 * Defines an 'facility-generic' exception class. This exception is
 * derived directly from @c Exception class.
 *
 * @param Name New generic exception name.
 * @param Facility Facility the new exception relates to.
 *
 * @note Without the constructor with 2 parameters the macro will
 *  not work for the inheritance hierarchy depths greater than 1.
 *
 * @note Since additional facilities can not be defined outside the core
 *       the class can be exported.
 */
#define MASSIV_CORE_GENERIC_EXCEPTION( Name, Facility ) \
class MASSIV_EXPORTED Name : public ::Massiv::Core::Exception \
    { \
public: \
    Name() : ::Massiv::Core::Exception \
                                ( \
                                ::Massiv::Core::Status::Facility, \
                                std::string(), \
                                0x0 \
                                ) \
        { \
        } \
    Name \
        ( \
        const std::string & message, \
        int                 flags = 0x0 \
        ) : ::Massiv::Core::Exception \
                                ( \
                                ::Massiv::Core::Status::Facility, \
                                message, \
                                flags \
                                ) \
        { \
        } \
    };


/**@def MASSIV_CORE_EXCEPTION
 * Defines an exception class implemented by the core.
 * @param Name Name of the new exception class.
 * @param Base Name of the base exception class. All exceptions defined
 *      using this macro should be derived (directly or indirectly)
 *      from an generic exception (defined by @c MASSIV_GENERIC_EXCEPTION
 *      macro).
 */
#define MASSIV_CORE_EXCEPTION( Name, Base ) \
    class MASSIV_EXPORTED Name : public Base \
        { \
    public: \
        Name() : Base( std::string(), 0x0 ) \
            { \
            } \
        Name \
            ( \
            const std::string & message, \
            int                 flags = 0x0 \
            ) : Base( message, flags ) \
            { \
            } \
        };


/**@def MASSIV_CORE_EXCEPTION_FLAGS
 * Defines an exception class implemented by the core with non-zero default flags.
 *
 * @param Name Name of the new exception class.
 * @param Base Name of the base exception class. All exceptions defined
 *      using this macro should be derived (directly or indirectly)
 *      from an generic exception (defined by @c MASSIV_GENERIC_EXCEPTION
 *      macro).
 * @param the_flags Combination of ::Massiv::Core::Exception::Flags.
 */
#define MASSIV_CORE_EXCEPTION_FLAGS( Name, Base, the_flags ) \
    class MASSIV_EXPORTED Name : public Base \
        { \
    public: \
        Name() : Base( std::string(), the_flags ) \
            { \
            } \
        Name \
            ( \
            const std::string & message, \
            int                 flags = the_flags \
            ) : Base( message, flags ) \
            { \
            } \
        };


/**@def MASSIV_EXCEPTION
 * Defines an exception class implemented outside the core.
 * The parameters are the same like those used by @c MASSIV_CORE_EXCEPTION.
 */
#define MASSIV_EXCEPTION( Name, Base ) \
    class Name : public Base \
        { \
    public: \
        Name() : Base( std::string(), 0x0 ) \
            { \
            } \
        Name \
            ( \
            const std::string & message, \
            int                 flags = 0x0 \
            ) : Base( message, flags ) \
            { \
            } \
        };


/**@def MASSIV_EXCEPTION_FLAGS
 * Defines an exception class implemented outside the core with non-zero default flags.
 *
 * The parameters are the same like those used by @c MASSIV_CORE_EXCEPTION_FLAGS.
 */
#define MASSIV_EXCEPTION_FLAGS( Name, Base, the_flags ) \
    class Name : public Base \
        { \
    public: \
        Name() : Base( std::string(), the_flags ) \
            { \
            } \
        Name \
            ( \
            const std::string & message, \
            int                 flags = the_flags, \
            ) : Base( message, flags ) \
            { \
            } \
        };


//
// Massiv specific exception subhierarchy.
//

/** @defgroup ex_common Common Errors
 * Covers all errors that don't fit into another more specific group.
 */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExCommonError, FACILITY_GENERIC );

    MASSIV_CORE_EXCEPTION( ExOutOfRange, ExCommonError );

    MASSIV_CORE_EXCEPTION( ExIteratorError, ExCommonError );
//@}
    

/** @defgroup ex_global Global System Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExSystemError, FACILITY_SYSTEM );

    MASSIV_CORE_EXCEPTION( ExSystemInitializationError, ExSystemError );
    
    MASSIV_CORE_EXCEPTION( ExSystemNodeConfigError, ExSystemError );
//@}


/** @defgroup ex_io I/O Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExIOError, FACILITY_IO );
    
    MASSIV_CORE_EXCEPTION( ExSerializationError, ExIOError );
    
        MASSIV_CORE_EXCEPTION( ExUnexpectedEol, ExSerializationError );

        MASSIV_CORE_EXCEPTION( ExUnexpectedEof, ExSerializationError );

        MASSIV_CORE_EXCEPTION( ExUnexpectedToken, ExSerializationError );

        MASSIV_CORE_EXCEPTION( ExUnexpectedValue, ExSerializationError );

    MASSIV_CORE_EXCEPTION( ExCommandLineError, ExIOError );

        MASSIV_CORE_EXCEPTION( ExUnexpectedCommandLineSwitch, ExCommandLineError );

        MASSIV_CORE_EXCEPTION( ExMissingCommandLineArgument, ExCommandLineError );

        MASSIV_CORE_EXCEPTION( ExMissingQuote, ExCommandLineError );

        MASSIV_CORE_EXCEPTION( ExMissingSwitch, ExCommandLineError );
//@}
        

/** @defgroup ex_logger Logger Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExLoggerError, FACILITY_LOGGER );

    MASSIV_CORE_EXCEPTION( ExCanNotOpenDestination, ExLoggerError );

    MASSIV_CORE_EXCEPTION( ExInvalidDestinationName, ExLoggerError );

    MASSIV_CORE_EXCEPTION( ExLoggerInitError, ExLoggerError );

        MASSIV_CORE_EXCEPTION( ExInvalidPriorityString, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExInvalidTypeString, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExInvalidFacilityString, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExInvalidSelectorElement, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExUnexpectedLoggerSetting, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExRuleRegistrationFailed, ExLoggerInitError );

        MASSIV_CORE_EXCEPTION( ExEmptySelector, ExLoggerInitError );
//@}


/** @defgroup ex_net Network Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExNetworkError, FACILITY_NETWORK );

    MASSIV_CORE_EXCEPTION( ExMessageCorrupted, ExNetworkError );

        MASSIV_CORE_EXCEPTION( ExMessageLengthMismatch, ExMessageCorrupted );

        MASSIV_CORE_EXCEPTION( ExMessagesNumberMismatch, ExMessageCorrupted );

    MASSIV_CORE_EXCEPTION( ExInvalidMessage, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExCanNotInitiateConnectionToServer, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExNoSuchConnection, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExConnectError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExSendError, ExNetworkError );
        
    MASSIV_CORE_EXCEPTION( ExReceiveError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExBadConnectionStatus, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExInvalidDnsName, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExClockError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExNetworkInitError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExWellKnownServerStartError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExConnectionExistsError, ExNetworkError );

    MASSIV_CORE_EXCEPTION( ExWrongBlowfishPeriod, ExNetworkError );
//@}


/** @defgroup ex_object Object Management Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExObjectManagementError, FACILITY_OBJECT_MANAGEMENT );

    MASSIV_CORE_EXCEPTION( ExMigrationGroupUnpackError, ExObjectManagementError );

    MASSIV_CORE_EXCEPTION( ExMigrationGroupUnpackSerializationError, ExMigrationGroupUnpackError );

    MASSIV_CORE_EXCEPTION( ExSwapInError, ExObjectManagementError );
    
    MASSIV_CORE_EXCEPTION( ExInvalidNodeId, ExObjectManagementError );

    MASSIV_CORE_EXCEPTION( ExObjectFinalizationError, ExObjectManagementError );
//@}
    

/** @defgroup ex_archive Archivation Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExArchivationError, FACILITY_ARCHIVATION );

    MASSIV_CORE_EXCEPTION( ExCorruptedArchive, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExCanNotOpenArchiveStream, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExCanNotCreateArchiveStream, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExArchiveLocked, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExCanNotValidateArchive, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExArchiveGroupNotFound, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExObjectAlreadyRegisteredInArchive, ExArchivationError );

    MASSIV_CORE_EXCEPTION( ExArchiveObjectNotFound, ExArchivationError );
//@}


/** @defgroup ex_registry Registry and Config Variable Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExRegistryError, FACILITY_REGISTRY );

    MASSIV_CORE_EXCEPTION( ExConfigFileErr, ExRegistryError );

        MASSIV_CORE_EXCEPTION( ExCanNotOpenConfigFile, ExConfigFileErr );

        MASSIV_CORE_EXCEPTION( ExCannotWriteConfigFile, ExConfigFileErr );

        MASSIV_CORE_EXCEPTION( ExConfigFileSyntaxError, ExConfigFileErr );

    MASSIV_CORE_EXCEPTION( ExCVarAliasLoop, ExRegistryError );

    MASSIV_CORE_EXCEPTION( ExCVarDuplicateName, ExRegistryError );

    MASSIV_CORE_EXCEPTION_FLAGS( ExCVarNotFound, ExRegistryError, NO_LOG );

        MASSIV_CORE_EXCEPTION_FLAGS( ExAliasNotFound, ExCVarNotFound, NO_LOG );

    MASSIV_CORE_EXCEPTION( ExCVarTypeMismatch, ExRegistryError );

    MASSIV_CORE_EXCEPTION( ExIllegalAlias, ExRegistryError );

    MASSIV_CORE_EXCEPTION( ExIllegalRegistryNode, ExRegistryError );

    MASSIV_CORE_EXCEPTION_FLAGS( ExRegistryNodeEmpty, ExRegistryError, NO_LOG );

    MASSIV_CORE_EXCEPTION_FLAGS( ExRegistryNodeNotFound, ExRegistryError, NO_LOG );

    MASSIV_CORE_EXCEPTION( ExWrongSyntax, ExRegistryError );
//@}


/** @defgroup ex_sync Synchronization Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExSynchronizationError, FACILITY_SYNCHRONIZATION );

    MASSIV_CORE_EXCEPTION( ExMutexError, ExSynchronizationError );

        MASSIV_CORE_EXCEPTION( ExMutexAbandoned, ExMutexError );

        MASSIV_CORE_EXCEPTION( ExMutexAlreadyExists, ExMutexError );

        MASSIV_CORE_EXCEPTION( ExMutexNotOwned, ExMutexError );

    MASSIV_CORE_EXCEPTION( ExCriticalSectionError, ExSynchronizationError );

        MASSIV_CORE_EXCEPTION( ExCriticalSectionCanNotCreate, ExCriticalSectionError );

        MASSIV_CORE_EXCEPTION( ExCriticalSectionCanNotEnter, ExCriticalSectionError );

        MASSIV_CORE_EXCEPTION( ExCriticalSectionCanNotLeave, ExCriticalSectionError );

        MASSIV_CORE_EXCEPTION( ExCriticalSectionCanNotDestroy, ExCriticalSectionError );

    MASSIV_CORE_EXCEPTION( ExSemaphoreError, ExSynchronizationError );

        MASSIV_CORE_EXCEPTION( ExSemaphoreCanNotCreate, ExSemaphoreError );

        MASSIV_CORE_EXCEPTION( ExSemaphoreCanNotDecrease, ExSemaphoreError );

        MASSIV_CORE_EXCEPTION( ExSemaphoreCanNotIncrease, ExSemaphoreError );

        MASSIV_CORE_EXCEPTION( ExSemaphoreCanNotDestroy, ExSemaphoreError );
//@}


/** @defgroup ex_thread Threading Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExThreadError, FACILITY_THREAD );

    MASSIV_CORE_EXCEPTION( ExThreadCanNotCreate, ExThreadError );

    MASSIV_CORE_EXCEPTION( ExThreadCanNotJoin, ExThreadError );
//@}


/** @defgroup ex_data Data Management Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExDataManagementError, FACILITY_DATA_MANAGEMENT );

    MASSIV_CORE_EXCEPTION( ExNotReady, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExNotMasterDataServer, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExUsedDataObjectHashTableNotEmpty, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExCanNotCreateDataDescriptionFile, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExDataFileNotExist, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExCanNotRenameDataDescriptionFile, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExDataObjectNotRegistered, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExNoAvailableDataNode, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExCanNotOpenDataFile, ExDataManagementError );

    // Stoupik: LOW_PRIORITY because data manager usually throws these on database change.
    MASSIV_CORE_EXCEPTION_FLAGS( ExDataNodeNotExist, ExDataManagementError, LOW_PRIORITY );

    // Stoupik: LOW_PRIORITY because data manager usually throws these on database change.
    MASSIV_CORE_EXCEPTION_FLAGS( ExDataObjectNotExist, ExDataManagementError, LOW_PRIORITY );

    MASSIV_CORE_EXCEPTION( ExDataFileAlreadyUsed, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExDataNodeAlreadyExist, ExDataManagementError );

    MASSIV_CORE_EXCEPTION( ExDeprecatedDataNode, ExDataManagementError );
//@}


/** @defgroup ex_time Time Management Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExTimeManagementError, FACILITY_TIME_MANAGEMENT );

    MASSIV_CORE_EXCEPTION( ExNodeAlreadyExistInSynchronizationHashTable, ExTimeManagementError );

    MASSIV_CORE_EXCEPTION( ExNodeNotExistInSynchronizationHashTable, ExTimeManagementError );
//@}


/** @defgroup ex_fs File System Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExFileSystemError, FACILITY_FILE_SYSTEM );

    MASSIV_CORE_EXCEPTION( ExUnrecognizedImageFile, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExIntegrityError, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExInvalidFsIterator, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExCanNotChangeDirectory, ExFileSystemError );

    MASSIV_CORE_EXCEPTION_FLAGS( ExCanNotFindFile, ExFileSystemError, LOW_PRIORITY | NO_STACK_TRACE );

    MASSIV_CORE_EXCEPTION( ExCanNotOpenFile, ExFileSystemError );

    MASSIV_CORE_EXCEPTION_FLAGS( ExNotADirectory, ExFileSystemError, LOW_PRIORITY | NO_STACK_TRACE );

    MASSIV_CORE_EXCEPTION_FLAGS( ExNotAFile, ExFileSystemError, LOW_PRIORITY | NO_STACK_TRACE );

    MASSIV_CORE_EXCEPTION( ExPathTooLong, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExErrorEnumeratingDirectory, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExCanNotGetFileInformation, ExFileSystemError );

    MASSIV_CORE_EXCEPTION_FLAGS( ExDirectoryNotFound, ExFileSystemError, LOW_PRIORITY | NO_STACK_TRACE );

    MASSIV_CORE_EXCEPTION( ExCanNotCreateDirectory, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExCanNotDeleteDirectory, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExDirectoryNotEmpty, ExCanNotDeleteDirectory );

    MASSIV_CORE_EXCEPTION_FLAGS( ExCanNotDeleteFile, ExFileSystemError, LOW_PRIORITY | NO_STACK_TRACE );

    MASSIV_CORE_EXCEPTION( ExCanNotRenameFile, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExCanNotCreateFile, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExUnrecognizedVolumeImage, ExFileSystemError );

    MASSIV_CORE_EXCEPTION( ExCanNotOpenVolumeImage, ExFileSystemError );
//@}


/** @defgroup ex_volume Volume Manager Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExVolumeManagerError, FACILITY_VOLUME_MANAGER );

    MASSIV_CORE_EXCEPTION( ExVolumeNotMounted, ExVolumeManagerError );
    
    MASSIV_CORE_EXCEPTION( ExVolumeAlreadyMounted, ExVolumeManagerError );
//@}


/** @defgroup ex_lru LRUCache Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExLRUCacheError, FACILITY_LRU_CACHE );

    MASSIV_CORE_EXCEPTION( ExOutOfCacheMemory, ExLRUCacheError );
//@}


/** @defgroup ex_nodedata Node Database Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExNodeDatabaseError, FACILITY_NODE_DATABASE );

    MASSIV_CORE_EXCEPTION( ExNodeNotFoundInNodeDatabase, ExNodeDatabaseError );
//@}

        
/** @defgroup ex_archivedata Archive Database Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExArchiveDatabaseError, FACILITY_ARCHIVE_DATABASE );

    MASSIV_CORE_EXCEPTION( ExArchiveDatabaseInitializationError, ExArchiveDatabaseError );
//@}


/** @defgroup ex_nodemgmt Node Management Errors */
//@{
MASSIV_CORE_GENERIC_EXCEPTION( ExNodeManagementError, FACILITY_NODE_MANAGEMENT );

    MASSIV_CORE_EXCEPTION( ExUnexpectedSystemMessage, ExNodeManagementError );

    MASSIV_CORE_EXCEPTION( ExConnectionLost, ExNodeManagementError );
    
    MASSIV_CORE_EXCEPTION( ExConnectionFailed, ExNodeManagementError );
//@}


/**
 * Library exception.
 * Inherited by Massiv::Core::Throwable, do not use directly.
 */
MASSIV_CORE_GENERIC_EXCEPTION( ExLibError, FACILITY_LIB );

/** @} */ // addtogroup exception

            
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_EXCEPTION_H
