/** @file status.h
 * Status declaration.
 */

/*
 * begin        : Tue Nov 6 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_STATUS_H
#define MASSIV_STATUS_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <string>

namespace Massiv {
namespace Core {


/**
 * Status - wraps information about result of some action.
 * Status is used for logging various messages (LogMessage, Logger). 
 * Don't use it to return values from functions, methods, etc.
 *
 * @ingroup logger
 * @ingroup exception
 */
class MASSIV_EXPORTED Status
    {
public:

    /**
     * Facility specifies subsystem that produced the status.
     * Do not forget to update Status::facility_string() when adding new 
     * facility.
     */
    enum Facility
        {
        FACILITY_ANY        = 0,
            /**< Unknown facility. Used by logger to match any LogMessage
                 Never use this in LogMessage status. */

        FACILITY_GENERIC,
            /**< Generic facility. Used for all things which don't fit
             * into another more specific facility.
             * For example errors from generic purpose data structures etc.
             */

        FACILITY_SYSTEM,
            /**< Global system stuff (startup, initialization, etc.). */
                 
        FACILITY_IO,                    /**< I/O. */
        FACILITY_LOGGER,                /**< Logger.*/
        FACILITY_NETWORK,               /**< Network subsystem. */
        FACILITY_REGISTRY,              /**< Registry. */
        FACILITY_OBJECT_MANAGEMENT,     /**< General object management. */
        FACILITY_OBJECT_LOCALIZATION,   /**< Distributed object localization. */
        FACILITY_ARCHIVATION,           /**< Object archivation. */
        FACILITY_FILE_SYSTEM,           /**< File system. */
        FACILITY_VOLUME_MANAGER,        /**< Volume manager. */
        FACILITY_LRU_CACHE,             /**< LRU cache. */
        FACILITY_SYNCHRONIZATION,       /**< Synchronization. */
        FACILITY_THREAD,                /**< Threading. */
        FACILITY_DATA_MANAGEMENT,       /**< Data manager. */
        FACILITY_TIME_MANAGEMENT,       /**< Time manager. */
        FACILITY_NODE_DATABASE,         /**< Node database facility. */
        FACILITY_ARCHIVE_DATABASE,      /**< Archive database facility. */
        FACILITY_NODE_MANAGEMENT,       /**< Node management, synchronization and communication. */
        FACILITY_REPLICATION,           /**< Object replication. */
        FACILITY_RPC,                   /**< Remote procedure call. */
        FACILITY_BALANCER,              /**< Balancer. */
        FACILITY_LIB,                   /**< Lib code. */
        
        FACILITY_TEST,                  /**< Self-tests. */
        
        FACILITY_COUNT
        };

    /**
     * Status type.
     * Messages of type <=TYPE_WARNING should be sent directly to logger as
     * LogMessages. If type is TYPE_ERROR, throw them as corresponding exceptions.
     * Order of types is important in logger rules (@see Logger).
     * @note Don't forget to change Logger when adding new values!
     */
    enum Type
        {
        TYPE_ANY            = 0,
            /**< Unknown type. Do not use. */
        TYPE_DEBUG          = 1,
            /**< Debugging message. */
        TYPE_INFO           = 2,
            /**< Informational message. */
        TYPE_WARNING        = 3,
            /**< Warning. */
        TYPE_ERROR          = 4,
            /**< Critical error or exception. */
        TYPE_COUNT
        };

    /**
     * Status priority.
     * Should be always set to @c PRIORITY_HIGHEST for status of type 
     * @c TYPE_ERROR.
     * 
     * @note Don't forget to change Logger when adding new values!
     */
    enum Priority
        {
        PRIORITY_LOWEST     = 0,
        PRIORITY_LOW        = 1,
        PRIORITY_NORMAL     = 2,
        PRIORITY_HIGH       = 3,
        PRIORITY_HIGHEST    = 4,
        PRIORITY_COUNT
        };

public:

    /** @name Various conversion methods */
    //@{

    /**
     * Convert Facility to string.
     * @param facility The facility.
     * @return The facility name.
     */
    static std::string facility_string
        (
        Facility facility
        );

    /**
     * Convert string describing facility to Facility. Returns FACILITY_COUNT 
     * if the string is not recognized.
     * @param fac_str Name of the facility.
     * @return The facility value.
     */
    static Facility facility_from_string
        (
        const std::string &     fac_str
        );

    /**
     * Convert Type to string.
     * @param type The type.
     * @return The type name.
     */
    static std::string type_string
        (
        Type type
        );

    /**
     * Convert string describing type to Type. Returns @c TYPE_COUNT if the
     * string is not recognized.
     * @param type_str Name of the type.
     * @return The type value.
     */
    static Type type_from_string
        (
        const std::string &     type_str
        );

    /**
     * Convert Priority to string.
     * @param priority The priority value.
     * @return The priority 'name'.
     */
    static std::string priority_string
        (
        Priority priority
        );

    /**
     * Convert string describing priority to Priority. Returns PRIORITY_COUNT
     * if the string is not recognized.
     * @param pri_str 'Name' of the priority.
     * @return The priority value.
     */
    static Priority priority_from_string
        (
        const std::string &     pri_str
        );

    /**
     * Convert Priority to single character.
     * @param priority The priority value.
     * @return The single character that specifies the priority.
     */
    static char priority_char
        (
        Priority priority
        );

    //@}

public:

    union
        {
        struct
            {
            int priority : 4;
            int type : 4;
            int facility : 8;
            int reserved : 16;  // unusable padding
            }   c;

        int     v;
        } value;
            /**< Status value. */

public:

    /**
     * Default constructor. Creates unknown/undefined status.
     */
    Status();

    /**
     * Status constructor.
     * @param facility  Status facility.
     * @param type      Status type.
     * @param priority  Status priority.
     */
    Status
        (
        Facility    facility,
        Type        type,
        Priority    priority
        );

    /**
     * Copy constructor.
     */
    Status( const Status & s );

public:

    /**
     * Create debug status.
     * @param facility Status facility.
     * @param priority Status priority.
     */
    static INLINE Status debug
        (
        Facility    facility,
        Priority    priority = PRIORITY_NORMAL
        );

    /**
     * Create info status.
     * @param facility Status facility.
     * @param priority Status priority.
     */
    static INLINE Status info
        (
        Facility    facility,
        Priority    priority = PRIORITY_NORMAL
        );

    /**
     * Create warning status.
     * @param facility Status facility.
     * @param priority Status priority.
     */
    static INLINE Status warning
        (
        Facility    facility,
        Priority    priority = PRIORITY_NORMAL
        );

    /**
     * Create error status.

     */
    static INLINE Status error
        (
        Facility    facility,
        Priority    priority = PRIORITY_HIGHEST
        );

public:

    /**
     * Assignment operator.
     */
    Status & operator=( const Status & s );

    /**
     * Compare status.
     */
    bool operator==( const Status & s ) const;

    /**
     * Compare status.
     */
    bool operator!=( const Status & s ) const;

public:

    /**
     * Get status description.
     * @param message   Custom message to include in result.
     * @return          String describing the status.
     */
    std::string get_description
        (
        const std::string &     message
        ) const;

    }; // class Status


/** @def STATUS_DEBUG
 * Create a debugging status.
 * @param f     Facility.
 * @param p     Priority.
 */
#define STATUS_DEBUG( f, p ) \
    Massiv::Core::Status( f, Status::TYPE_DEBUG, p )


/** @def STATUS_INFO
 * Create an informational status.
 * @param f     Facility.
 * @param p     Priority.
 */
#define STATUS_INFO( f, p ) \
    Massiv::Core::Status( f, Status::TYPE_INFO, p )


/** @def STATUS_WARNING
 * Create a warning status.
 * @param f     Facility.
 * @param p     Priority.
 */
#define STATUS_WARNING( f, p ) \
    Massiv::Core::Status( f, Status::TYPE_WARNING, p )


/** @def STATUS_ERROR
 * Create an error/exception status.
 * @param f     Facility.
 */
#define STATUS_ERROR( f ) \
    Massiv::Core::Status( f, Status::TYPE_ERROR, Status::PRIORITY_HIGHEST )


} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "status_inline.h"
#endif


#endif // MASSIV_STATUS_H
