/** @file archivation_msg.h
 * Archivation messages.
 */

/*
 * begin        : Sat Oct 19 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_ARCHIVATION_MSG_H
#define MASSIV_ARCHIVATION_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_ARCHIVATION_ID_H
#include "object/archivation_id.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * Base class of all archivation messages.
 * Stores archivation id that uniquely identitifies an archivation.
 *
 * @ingroup archivation
 */
class ArchivationMessage : public SystemMessage
    {
    typedef SystemMessage       Super;

protected:

    /**
     * Constructor.
     */
    ArchivationMessage();

    /**
     * Copy constructor.
     */
    ArchivationMessage
        (
        const ArchivationMessage & msg
        );

    /**
     * Operator=.
     */
    ArchivationMessage & operator=
        (
        const ArchivationMessage & msg
        );

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    /**
     * Get archivation id.
     */
    const ArchivationId & get_archivation_id() const
        {
        return archivation_id;
        }

    /**
     * Set archivation id.
     */
    void set_archivation_id
        (
        const ArchivationId & _archivation_id
        )
        {
        archivation_id = _archivation_id;
        }

private:

    ArchivationId   archivation_id;
        /**< @c ArchivationId of the archivation this
         * message is about.
         */

    }; // ArchivationMessage


/**
 * Archivation status message.
 * Used by distributed archivation protocol to notify
 * of an archivation progress on some node.
 *
 * @ingroup archivation
 */
class ArchivationStatusMessage : public ArchivationMessage
    {
    typedef ArchivationMessage  Super;

public:

    /**
     * Archivation status.
     */
    enum Status
        {
        ARCHIVATION_STARTED     = 0,
            /**< Local archivation has been started on a node. */
        ARCHIVATION_FINISHED    = 1,
            /**< Local archivation has been finished on a node. */
        STATUS_BITS             = 1
            /**< Number of bits required to store the @c Status. */
        }; // Status

public:

    /**
     * Constructor.
     * The status is set to @c ARCHIVATION_STARTED, but
     * do not rely on that and use another constructor instead.
     * This constructor should be used only if you want to
     * read the message from a stream.
     */
    ArchivationStatusMessage();

    /** Constructor */
    ArchivationStatusMessage
        (
        Status  archivation_status
        );

    /** Copy constructor. */
    ArchivationStatusMessage
        (
        const ArchivationStatusMessage & msg
        );

    /** Operator=. */
    ArchivationStatusMessage & operator=
        (
        const ArchivationStatusMessage & msg
        );

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

public:

    /**
     * Get archivation status.
     */
    Status get_status() const
        {
        return status;
        }

    /**
     * Set archivation status.
     */
    void set_status
        (
        Status _status
        )
        {
        status = _status;
        }

private:

    Status      status;
        /**< Archivation status. */

    }; // ArchivationStatusMessage

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_ARCHIVATION_MSG_H
