/** @file client_connection_msg.h
 * Client connection messages.
 */

/*
 * begin        : Mon Feb 03 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_CLIENT_CONNECTION_MSG_H
#define MASSIV_CLIENT_CONNECTION_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * This message is sent from server node to client node once it is
 * successfully connected to the simulation.
 */
class WelcomeMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:
    /** Constructor. */
    WelcomeMessage()
        {
        }

    /** Copy constructor. */
    WelcomeMessage
        (
        const WelcomeMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    WelcomeMessage & operator=
        (
        const WelcomeMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();
        
    }; // WelcomeMessage


/**
 * ClientConnectionMessage.
 * Base message for client connection (server side) related messages.
 */
class ClientConnectionMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:
    /** Constructor. */
    ClientConnectionMessage();

    /** Constructor. */
    ClientConnectionMessage
        (
        const NodeId &  client_node_id
        );
        
    /** Copy constructor. */
    ClientConnectionMessage
        (
        const ClientConnectionMessage & msg
        );

    /** Operator=. */
    ClientConnectionMessage & operator=
        (
        const ClientConnectionMessage & msg
        );

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    /**
     * Get client node id.
     */
    const NodeId & get_client_node_id() const
        {
        return client_node_id;
        }

    /**
     * Set client node id.
     */
    void set_client_node_id
        (
        const NodeId & _client_node_id
        )
        {
        client_node_id = _client_node_id;
        }

private:
    NodeId              client_node_id;
        /**< NodeId of the client node which is connecting
         * to the simulation.
         */

    }; // ClientConnectionMessage

        
/**
 * AcceptClientMessage.
 * This message is used to notify server nodes
 * that a client is trying to connect to the
 * simulation.
 * The server node which will send this message
 * will be marked as forward node of the client
 * (see NodeManager for more details).
 * Every server node replies to this message
 * with @c ClientAcceptedMessage.
 */
class AcceptClientMessage : public ClientConnectionMessage
    {
    typedef ClientConnectionMessage   Super;

public:
    /** Constructor. */
    AcceptClientMessage()
        {
        }

    /** Constructor. */
    AcceptClientMessage
        (
        const NodeId &  _client_node_id
        ) :
        Super( _client_node_id )
        {
        }

    /** Copy constructor. */
    AcceptClientMessage
        (
        const AcceptClientMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    AcceptClientMessage & operator=
        (
        const AcceptClientMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

    }; // AcceptClientMessage


/**
 * ClientAcceptedMessage.
 * Reply message to AcceptClientMessage.
 */
class ClientAcceptedMessage : public ClientConnectionMessage
    {
    typedef ClientConnectionMessage   Super;

public:
    /** Constructor. */
    ClientAcceptedMessage()
        {
        }

    /** Constructor. */
    ClientAcceptedMessage
        (
        const NodeId &  _client_node_id
        ) :
        Super( _client_node_id )
        {
        }

    /** Copy constructor. */
    ClientAcceptedMessage
        (
        const ClientAcceptedMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    ClientAcceptedMessage & operator=
        (
        const ClientAcceptedMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

    }; // ClientAcceptedMessage


/**
 * ClientReadyMessage.
 * Message used by server nodes to notify
 * that client was successfully connected
 * to the simulation.
 */
class ClientReadyMessage : public ClientConnectionMessage
    {
    typedef ClientConnectionMessage     Super;

public:
    /** Constructor. */
    ClientReadyMessage()
        {
        }

    /** Constructor. */
    ClientReadyMessage
        (
        const NodeId &  _client_node_id
        ) :
        Super( _client_node_id )
        {
        }

    /** Copy constructor. */
    ClientReadyMessage
        (
        const ClientReadyMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    ClientReadyMessage & operator=
        (
        const ClientReadyMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

    }; // ClientReadyMessage


/**
 * ClientReadyReplyMessage.
 * Message used by server nodes to reply
 * to ClientReadyMessage.
 */
class ClientReadyReplyMessage : public ClientConnectionMessage
    {
    typedef ClientConnectionMessage     Super;

public:
    /** Constructor. */
    ClientReadyReplyMessage()
        {
        }

    /** Constructor. */
    ClientReadyReplyMessage
        (
        const NodeId &  _client_node_id
        ) :
        Super( _client_node_id )
        {
        }

    /** Copy constructor. */
    ClientReadyReplyMessage
        (
        const ClientReadyReplyMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    ClientReadyReplyMessage & operator=
        (
        const ClientReadyReplyMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

    }; // ClientReadyReplyMessage


/**
 * This message is used to disconnect client node from simulation.
 */
class KillClientMessage : public ClientConnectionMessage
    {
    typedef ClientConnectionMessage     Super;

public:
    /** Constructor. */
    KillClientMessage()
        {
        }

    /** Constructor. */
    KillClientMessage
        (
        const NodeId &  _client_node_id
        ) :
        Super( _client_node_id )
        {
        }

    /** Copy constructor. */
    KillClientMessage
        (
        const KillClientMessage & msg
        ) :
        Super( msg )
        {
        }

    /** Operator=. */
    KillClientMessage & operator=
        (
        const KillClientMessage & msg
        )
        {
        Super::operator=( msg );
        return *this;
        }

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();

    }; // KillClientMessage

/** @} */ // addtogroup sysevent

                
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_CLIENT_CONNECTION_MSG_H
