/** @file data_msg.cpp
 * Data message.
 */

/*
 * begin        : Wed Jan 22 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "sysevent/data_msg.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "global/global.h"
#include "database/node_database.h"
#include <cstring>

namespace Massiv {
namespace Core {


// **** DataVersionRequestMessage start ****

DataVersionRequestMessage::DataVersionRequestMessage()
    {
    }


DataVersionRequestMessage::DataVersionRequestMessage
    (
    const DataVersionRequestMessage &      msg
    ) : Super( msg )
    {
    }


DataVersionRequestMessage & DataVersionRequestMessage::operator=
    (
    const DataVersionRequestMessage &      msg
    )
    {
    Super::operator=( msg );
    return *this;
    }


SystemMessage::Type DataVersionRequestMessage::get_type() const
    {
    return SystemMessage::DATA_VERSION_REQUEST;
    }


SystemMessage::ExecuteResult DataVersionRequestMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_version_request_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_server();
        data_manager.handle_data_version_request_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataVersionRequestMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    }


void DataVersionRequestMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );
    }


// **** DataVersionRequestMessage end ****
// **** DataRequestMessage start ****

DataRequestMessage::DataRequestMessage()
    {
    }

DataRequestMessage::DataRequestMessage
    (
    const std::string &             _data_file_name,
    const DataVersion &             _data_version,
    int                             _download_speed
    ) : data_file_name( _data_file_name ),
        data_version( _data_version ),
        download_speed( _download_speed )
    {
    }


DataRequestMessage::DataRequestMessage
    (
    const DataRequestMessage &      msg
    ) : Super( msg ),
        data_file_name( msg.data_file_name ),
        data_version( msg.data_version ),
        download_speed( msg.download_speed )
    {
    }


DataRequestMessage & DataRequestMessage::operator=
    (
    const DataRequestMessage &      msg
    )
    {
    Super::operator=( msg );
    data_file_name = msg.data_file_name;
    data_version = msg.data_version;
    download_speed = msg.download_speed;
    return *this;
    }


SystemMessage::Type DataRequestMessage::get_type() const
    {
    return SystemMessage::DATA_REQUEST;
    }


SystemMessage::ExecuteResult DataRequestMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_request_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_server();
        data_manager.handle_data_request_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataRequestMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    
    data_file_name.bitstr_write( bos, desc );
    data_version.bitstr_write( bos );
    download_speed.bitstr_write( bos , desc);
    }


void DataRequestMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );

    data_file_name.bitstr_read( bis, desc );
    data_version.bitstr_read( bis );
    download_speed.bitstr_read( bis, desc );
    }

// **** DataRequestMessage end ****
// **** DataVersionMessage start ****

DataVersionMessage::DataVersionMessage()
    {
    }

DataVersionMessage::DataVersionMessage
    (
    const DataVersion &             _data_version,
    const MagicNumberType &         _magic_number
    ) :
    data_version( _data_version ),
    magic_number( _magic_number)
    {
    }


DataVersionMessage::DataVersionMessage
    (
    const DataVersionMessage &      msg
    ) : 
    Super( msg ),
    data_version( msg.data_version ),
    magic_number( msg.magic_number )
    {
    }


DataVersionMessage & DataVersionMessage::operator=
    (
    const DataVersionMessage &      msg
    )
    {
    Super::operator=( msg );
    data_version = msg.data_version;
    magic_number = msg.magic_number;
    return *this;
    }


SystemMessage::Type DataVersionMessage::get_type() const
    {
    return SystemMessage::DATA_VERSION;
    }


SystemMessage::ExecuteResult DataVersionMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_version_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_client();
        data_manager.handle_data_version_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataVersionMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    
    data_version.bitstr_write( bos );
    magic_number.bitstr_write( bos, desc );
    }


void DataVersionMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );

    data_version.bitstr_read( bis );
    magic_number.bitstr_read( bis, desc );
    }

// **** DataVersionMessage end ****
// **** DataMessage start ****

DataMessage::DataMessage() :
    buffer_size( 0 ),
    offset( 0 ),
    buffer( NULL ),
    end( false )
    {
    }

DataMessage::DataMessage
    (
    int                     _buffer_size,
    const std::string &     _data_file_name
    ) : buffer_size( _buffer_size ),
        offset( 0 ),
        end( false ),
        data_file_name( _data_file_name )
    {
    //hope( buffer_size != 0 ); FIXME
    buffer = new UInt8[ buffer_size ];
    hope( buffer != NULL );
    hope( _data_file_name.size() );
    }


DataMessage::DataMessage
    (
    const DataMessage &      msg
    ) : Super( msg ),
        buffer_size( msg.buffer_size ),
        offset( msg.offset ),
        end( msg.end ),
        data_file_name( msg.data_file_name )
    {
    hope( buffer_size != 0 );
    buffer = new UInt8[ buffer_size ];
    hope( buffer != NULL );
    memmove( buffer, msg.buffer, buffer_size );
    }


DataMessage & DataMessage::operator=
    (
    const DataMessage &      msg
    )
    {
    Super::operator=( msg );

    buffer_size = msg.buffer_size;
    hope( buffer_size != 0 );
    offset = msg.offset;
    end = msg.end;
    data_file_name = msg.data_file_name;
    
    delete buffer;
    buffer = new UInt8[ buffer_size ];
    hope( buffer != NULL );
    memmove( buffer, msg.buffer, buffer_size );

    return *this;
    }

DataMessage::~DataMessage()
    {
    delete buffer;
    }



SystemMessage::Type DataMessage::get_type() const
    {
    return SystemMessage::DATA;
    }


SystemMessage::ExecuteResult DataMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_client();
        data_manager.handle_data_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    
    buffer_size.bitstr_write( bos, desc );
    offset.bitstr_write( bos, desc );
    end.bitstr_write( bos, desc );
    data_file_name.bitstr_write( bos, desc );
    
    bos.write_raw( buffer, buffer_size );
    }


void DataMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );

    buffer_size.bitstr_read( bis, desc );
    offset.bitstr_read( bis, desc );
    end.bitstr_read( bis, desc );
    data_file_name.bitstr_read( bis, desc );

    buffer = new UInt8[ buffer_size ];
    hope( buffer != NULL );

    bis.read_raw( buffer, buffer_size );
    }

// **** DataMessage end ****
// **** DataConfirmationMessage start ****

DataConfirmationMessage::DataConfirmationMessage()
    {
    }

DataConfirmationMessage::DataConfirmationMessage
    (
    int                     _offset,
    const std::string &     _data_file_name
    ) :
    offset( _offset ),
    data_file_name( _data_file_name )
    {
    }


DataConfirmationMessage::DataConfirmationMessage
    (
    const DataConfirmationMessage &      msg
    ) :
    Super( msg ),
    offset( msg.offset ),
    data_file_name( msg.data_file_name )
    {
    }


DataConfirmationMessage & DataConfirmationMessage::operator=
    (
    const DataConfirmationMessage &      msg
    )
    {
    Super::operator=( msg );
    offset = msg.offset;
    data_file_name = msg.data_file_name;
    return *this;
    }


SystemMessage::Type DataConfirmationMessage::get_type() const
    {
    return SystemMessage::DATA_CONFIRMATION;
    }


SystemMessage::ExecuteResult DataConfirmationMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        static DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_confirmation_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        static DataManager &      data_manager = Global::data_manager_test_server();
        data_manager.handle_data_confirmation_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataConfirmationMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    
    offset.bitstr_write( bos, desc );
    data_file_name.bitstr_write( bos, desc );

    std::ostringstream  ss;
    ss << "DataConfirmationMessage::bitstr_write: data_file_name= " << data_file_name << "\n";
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );
    }


void DataConfirmationMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );

    offset.bitstr_read( bis, desc );
    data_file_name.bitstr_read( bis, desc );
    
    std::ostringstream  ss;
    ss << "DataConfirmationMessage::bitstr_read: data_file_name= " << data_file_name << "\n";
    Global::log
        (
        STATUS_INFO( Status::FACILITY_DATA_MANAGEMENT, Status::PRIORITY_LOWEST ),
        ss.str()
        );
    }

// **** DataConfirmationMessage end ****
// **** DataErrorMessage start ****

DataErrorMessage::DataErrorMessage()
    {
    }

DataErrorMessage::DataErrorMessage
    (
    DataManager::DataMessageErrorType      _error_type
    ) :
    error_type( _error_type )
    {
    }


DataErrorMessage::DataErrorMessage
    (
    const DataErrorMessage &      msg
    ) :
    Super( msg ),
    error_type( msg.error_type )
    {
    }


DataErrorMessage & DataErrorMessage::operator=
    (
    const DataErrorMessage &      msg
    )
    {
    Super::operator=( msg );
    error_type = msg.error_type;
    return *this;
    }


SystemMessage::Type DataErrorMessage::get_type() const
    {
    return SystemMessage::DATA_ERROR;
    }


SystemMessage::ExecuteResult DataErrorMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_error_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_client();
        data_manager.handle_data_error_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataErrorMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    SInt<UInt32, DataManager::ERROR_TYPE_BIT_COUNT> error( error_type );
    Super::bitstr_write( bos, desc );
    
    error.bitstr_write( bos, desc );
    }


void DataErrorMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    SInt<UInt32, DataManager::ERROR_TYPE_BIT_COUNT> error;

    Super::bitstr_read( bis, desc );
    
    error.bitstr_read( bis, desc );
    error_type = static_cast< DataManager::DataMessageErrorType >( error.read() );

    if( error_type >= DataManager::ERROR_TYPE_COUNT )
        {
        throw ExSerializationError( "Value of data message error type is greater than ERROR_TYPE_COUNT" );
        }
    }

// **** DataErrorMessage end ****
// **** DataUpdateRequestMessage start ****

DataUpdateRequestMessage::DataUpdateRequestMessage()
    {
    }


DataUpdateRequestMessage::DataUpdateRequestMessage
    (
    const DataUpdateRequestMessage &      msg
    ) : Super( msg )
    {
    }


DataUpdateRequestMessage & DataUpdateRequestMessage::operator=
    (
    const DataUpdateRequestMessage &      msg
    )
    {
    Super::operator=( msg );
    return *this;
    }


SystemMessage::Type DataUpdateRequestMessage::get_type() const
    {
    return SystemMessage::DATA_UPDATE_REQUEST;
    }


SystemMessage::ExecuteResult DataUpdateRequestMessage::execute()
    {
    NodeDatabase &   node_database = Global::node_database();
    if( get_initiator() != node_database.get_local_node_id() )
        {
        DataManager &      data_manager = Global::data_manager();
        data_manager.handle_data_update_request_message( *this );
        }
    else
        {
        // Message is sent during local DataManager test.
        DataManager &      data_manager = Global::data_manager_test_server();
        data_manager.handle_data_update_request_message( *this );
        }
    return SystemMessage::EXECUTE_OK;
    }


void DataUpdateRequestMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    Super::bitstr_write( bos, desc );
    }


void DataUpdateRequestMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    Super::bitstr_read( bis, desc );
    }


// **** DataUpdateRequestMessage end ****


} // namespace Massiv::Core
} // namespace Massiv
