/** @file data_msg.h
 * Data message.
 */

/*
 * begin        : Wed Jan 22 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_DATA_MSG_H
#define MASSIV_DATA_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_BITIOSTREAM_FWD_H
#include "net/bitiostream_fwd.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif

#ifndef MASSIV_STYPE_STRING_H
#include "serializer/stype_string.h"
#endif

#ifndef MASSIV_DATA_MANAGER_H
#include "data/data_manager.h"
#endif


namespace Massiv {
namespace Core {


typedef DataManager::DataVersion         DataVersion;
typedef DataManager::MagicNumberType     MagicNumberType;
typedef DataManager::SMagicNumberType    SMagicNumberType;


/** @addtogroup sysevent
 * @{
 */

/**
 * Data version request message.
 *
 * @ingroup data_manager
 */
class DataVersionRequestMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataVersionRequestMessage();

    /**
     * Copy constructor.
     */
    DataVersionRequestMessage
        (
        const DataVersionRequestMessage &      msg
        );

    /**
     * Operator=.
     */
    DataVersionRequestMessage & operator=
        (
        const DataVersionRequestMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    // No attributes.

    }; // DataVersionRequestMessage


/**
 * Data request message.
 *
 * @ingroup data_manager
 */
class DataRequestMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataRequestMessage();

    /** 
     * Constructor. 
     */
    DataRequestMessage
        (
        const std::string &             _data_file_name,
        const DataVersion &             _data_version,
        int                             _download_speed
        );

    /**
     * Copy constructor.
     */
    DataRequestMessage
        (
        const DataRequestMessage &      msg
        );

    /**
     * Operator=.
     */
    DataRequestMessage & operator=
        (
        const DataRequestMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:


    SString         data_file_name;
        /**< Data file name of file to be downloaded. */

    DataVersion     data_version;
        /**< Version of data file. */

    SInt32          download_speed;
        /**< Requested download speed. */

    }; // DataRequestMessage


/**
 * Data version message.
 *
 * @ingroup data_manager
 */
class DataVersionMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataVersionMessage();

    /** 
     * Constructor. 
     */
    DataVersionMessage
        (
        const DataVersion &             _data_version,
        const MagicNumberType &         _magic_number
        );

    /**
     * Copy constructor.
     */
    DataVersionMessage
        (
        const DataVersionMessage &      msg
        );

    /**
     * Operator=.
     */
    DataVersionMessage & operator=
        (
        const DataVersionMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    DataVersion                 data_version;
        /**< Version od data description. */

    SMagicNumberType            magic_number;
        /**< Data magic number. */
        
    }; // DataVersionMessage


/**
 * Data message.
 *
 * @ingroup data_manager
 */
class DataMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataMessage();

    /** 
     * Constructor. 
     */
    DataMessage
        (
        int                     _buffer_size,
        const std::string &     _data_file_name
        );

    /**
     * Copy constructor.
     */
    DataMessage
        (
        const DataMessage &      msg
        );

    /**
     * Operator=.
     */
    DataMessage & operator=
        (
        const DataMessage &       msg
        );

    /**
     * Destructor.
     */
    ~DataMessage();

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    SInt32                      buffer_size;
        /**< Size of buffer. */
    
    SInt32                      offset;
        /**< Offset in destination stream. */

    UInt8 *                     buffer;
        /**< Buffer for data. */

    SBool                       end;
        /**< True when this is a last data message of current data file transfer. */
    
    SString                     data_file_name;
        /**< Data file name. */
        
    }; // DataMessage



/**
 * Data confirmation message.
 *
 * @ingroup data_manager
 */
class DataConfirmationMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataConfirmationMessage();

    /** 
     * Constructor. 
     */
    DataConfirmationMessage
        (
        int                     _offset,
        const std::string &     _data_file_name
        );

    /**
     * Copy constructor.
     */
    DataConfirmationMessage
        (
        const DataConfirmationMessage &      msg
        );

    /**
     * Operator=.
     */
    DataConfirmationMessage & operator=
        (
        const DataConfirmationMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    SInt32                      offset;
        /**< Offset in destination stream. */

    SString                     data_file_name;
        /**< Data file name. */

    }; // DataConfirmationMessage



/**
 * Data error message.
 *
 * @ingroup data_manager
 */
class DataErrorMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataErrorMessage();

    /** 
     * Constructor. 
     */
    DataErrorMessage
        (
        DataManager::DataMessageErrorType      _error_type
        );

    /**
     * Copy constructor.
     */
    DataErrorMessage
        (
        const DataErrorMessage &      msg
        );

    /**
     * Operator=.
     */
    DataErrorMessage & operator=
        (
        const DataErrorMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    DataManager::DataMessageErrorType      error_type;
        /**< Data message error type. */

    }; // DataErrorMessage


/**
 * Data update request message.
 *
 * @ingroup data_manager
 */
class DataUpdateRequestMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** 
     * Constructor. 
     */
    DataUpdateRequestMessage();

    /**
     * Copy constructor.
     */
    DataUpdateRequestMessage
        (
        const DataUpdateRequestMessage &      msg
        );

    /**
     * Operator=.
     */
    DataUpdateRequestMessage & operator=
        (
        const DataUpdateRequestMessage &       msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    // No attributes.

    }; // DataUpdateRequestMessage

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_DATA_MSG_H
