/** @file forward_msg.cpp
 * Forward messages used to forward system messages
 * to client nodes.
 */

/*
 * begin        : Wed Oct 23 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "sysevent/forward_msg.h"
#include "serializer/serializer.h"
#include "utility/debug.h"
#include "global/global.h"
#include "net/bitostream.h"
#include "node/node_manager.h"


namespace Massiv {
namespace Core {


//-----------------------------------------------------------------------------
// ForwardMessage implementation.
//-----------------------------------------------------------------------------


ForwardMessage::ForwardMessage
    (
    const NodeId &        dest,
    const SystemMessage & fwd_msg,
    bool                  _reliable
    ) :
    destination_node( dest ),
    message_to_forward( &fwd_msg ),
    reliable( _reliable )
    {
    }


ForwardMessage::ForwardMessage
    (
    const ForwardMessage & msg
    ) :
    Super( msg ),
    destination_node( msg.destination_node ),
    message_to_forward( msg.message_to_forward ),
    reliable( msg.reliable )
    {
    }


ForwardMessage & ForwardMessage::operator=
    (
    const ForwardMessage & msg
    )
    {
    Super::operator=( msg );
    destination_node   = msg.destination_node;
    message_to_forward = msg.message_to_forward;
    reliable           = msg.reliable;
    return *this;
    }


void ForwardMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    // Write super class.
    Super::bitstr_write( bos, desc );

    // Write destination node.
    destination_node.bitstr_write( bos, desc );
    
    // Write reliable flag.
    reliable.bitstr_write( bos, desc );

    // Align to byte boundary.
    bos.align_to_byte();

    // Write the forwarded message itself.
    hope( message_to_forward != NULL );
    message_to_forward->bitstr_write( bos, desc );
    }


void ForwardMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    // ForwardMessage can't be read from a stream.
    // Use ForwardedMessage instead!
    panic( "ForwardMessage can't be read from a stream!" );
    }


SystemMessage::Type ForwardMessage::get_type() const
    {
    return SystemMessage::FORWARD;
    }


SystemMessage::ExecuteResult ForwardMessage::execute()
    {
    // ForwardMessage can't be executed.
    // Use ForwardedMessage instead!
    panic( "ForwardMessage can't be executed!" );
    return SystemMessage::EXECUTE_OK;
    }


//-----------------------------------------------------------------------------
// ForwardedMessage implementation.
//-----------------------------------------------------------------------------


ForwardedMessage::ForwardedMessage() :
    reliable( true )
    {
    }


ForwardedMessage::ForwardedMessage
    (
    const ForwardedMessage & msg
    ) :
    Super( msg ),
    destination_node( msg.destination_node ),
    forwarded_message( msg.forwarded_message ),
    reliable( msg.reliable )
    {
    }


ForwardedMessage & ForwardedMessage::operator=
    (
    const ForwardedMessage & msg
    )
    {
    Super::operator=( msg );
    destination_node  = msg.destination_node;
    forwarded_message = msg.forwarded_message;
    reliable          = msg.reliable;
    return *this;
    }


void ForwardedMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    // Write super class.
    Super::bitstr_write( bos, desc );

    // Write destination node.
    destination_node.bitstr_write( bos, desc );
    
    // Write reliable flag.
    reliable.bitstr_write( bos, desc );

    // Align the stream to byte boundary.
    bos.align_to_byte();

    // Write the forwarded message.
    // We must create a copy of message_stream because this->bitstr_write() is const.
        {
        BitIStream  s( forwarded_message );
        s.rewind();
        bos.write_bitistream( s );
        }
    }


void ForwardedMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    // Read super class.
    Super::bitstr_read( bis, desc );

    // Read destination node.
    destination_node.bitstr_read( bis, desc );
    
    // Read reliable flag.
    reliable.bitstr_read( bis, desc );

    // Align the stream to byte boundary.
    bis.align_to_byte();

    // Store the message stream.
    forwarded_message = bis.tail();
    }


SystemMessage::Type ForwardedMessage::get_type() const
    {
    return SystemMessage::FORWARD;
    }


SystemMessage::ExecuteResult ForwardedMessage::execute()
    {
    NodeManager & node_manager = Global::node_manager();
    return node_manager.handle_forwarded_message( *this );
    }


}; // namespace Massiv::Core
}; // namespace Massiv
