/** @file forward_msg.h
 * Forward messages used to forward system messages
 * to client nodes.
 */

/*
 * begin        : Wed Oct 23 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_FORWARD_MSG_H
#define MASSIV_FORWARD_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_SERIALIZER_FWD_H
#include "serializer/serializer_fwd.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif

#ifndef MASSIV_BITISTREAM_H
#include "net/bitistream.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * Forward message.
 * Forward message is used to forward system messages
 * to client nodes.
 *
 * @note @c ForwardMessage is used for sending forwarded
 * messages, @c ForwardedMessage is used for receiving and
 * resending forwarded messages.
 *
 * @note @c ForwardMessage has same type as @c ForwardedMessage.
 * @see ForwardedMessage
 *
 * @warning When calling copy constructor
 * from ancestor of ForwardMessage, make sure
 * that the compiler will not call
 * explicit copy construtor!
 *
 * @warning If the forwarded message contains migration group message,
 * the message is <b>not</b> captured by archivation. It is one of the reasons
 * why destination of forwarded message can't be server node.
 * Before forwarding migration group message, the objects stored in the
 * migration message must be removed from the simulation.
 */
class ForwardMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /**
     * Constructor.
     * @note Only reference to @a message_to_forward is stored in the ForwardMessage,
     * so do not release @a message_to_forward until the ForwardMessage is destroyed.
     */
    explicit ForwardMessage
        (
        const NodeId &        destination_node,
        const SystemMessage & message_to_forward,
        bool                  reliable
        );

    /**
     * Copy constructor.
     * @warning When calling copy constructor
     * from ancestor of ForwardMessage, make sure
     * that the compiler will call this constructor
     * (not explicit copy construtor)!
     */
    ForwardMessage
        (
        const ForwardMessage & message
        );

    /** Operator=. */
    ForwardMessage & operator=
        (
        const ForwardMessage & message
        );

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     * @warning <b>Do not use</b> this function.
     * It does nothing. Use @c ForwardedMessage instead.
     * @c ForwardMessage can't be read from a stream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /**
     * @warning <b>Do not use</b> this function.
     * It does nothing. Use @c ForwardedMessage instead.
     * @c ForwardMessage can't be executed.
     */
    virtual ExecuteResult execute();

private:

    NodeId                destination_node;
        /**< Destination node the forwarded message must be send to.
         * @note A message can be forwarded only to client node.
         */

    const SystemMessage * message_to_forward;
        /** Reference to system message that must be forwarded. */
    
    SBool                 reliable;
        /**< Specifies whether the forwarded message uses reliable
         * or unreliable transfer protocol.
         */

    }; // ForwardMessage


/**
 * Forward message.
 * Forward message is used to forward system messages
 * to client nodes.
 *
 * @note @c ForwardMessage is used for sending forwarded
 * messages, @c ForwardedMessage is used for receiving and
 * resending forwarded messages.
 *
 * @note @c ForwardMessage has same type as @c ForwardedMessage.
 * @see ForwardMessage
 *
 * @warning If the forwarded message contains migration group message,
 * the message is <b>not</b> captured by archivation. It is one of the reasons
 * why destination of forwarded message can't be server node.
 * Before forwarding migration group message, the objects stored in the
 * migration message must be removed from the simulation.
 */
class ForwardedMessage : public SystemMessage
    {
    typedef SystemMessage   Super;

public:

    /** Constructor. */
    ForwardedMessage();

    /** Copy constructor. */
    ForwardedMessage
        (
        const ForwardedMessage & message
        );

    /** Operator=. */
    ForwardedMessage & operator=
        (
        const ForwardedMessage & message
        );

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    //
    // From SystemMessage.
    //

    /** */
    virtual SystemMessage::Type get_type() const;

    /** */
    virtual ExecuteResult execute();
    
public:
    /**
     * Get destination node.
     */
    const NodeId & get_destination_node() const
        {
        return destination_node;
        }
    
    /**
     * Check whether the forwarded message uses reliable transfer protocol.
     */
    bool use_reliable_stream() const
        {
        return reliable;
        }

public:

    NodeId          destination_node;
        /**< Destination node the forwarded message must be send to.
         * @note A message can be forwarded only to client node.
         */

    BitIStream      forwarded_message;
        /**< Serialized system message stored in the @c ForwardedMessage. */
    
    SBool           reliable;
        /**< Specifies whether the forwarded message uses reliable
         * or unreliable transfer protocol.
         */

    }; // ForwardedMessage

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_FORWARD_MSG_H
