/** @file migration_group_msg.cpp
 * Migration group message.
 */

/*
 * begin        : Tue Apr 30 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/common.h"
#include "global/global.h"
#include "sysevent/migration_group_msg.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "object/object_manager.h"


namespace Massiv {
namespace Core {


MigrationGroupMessage::MigrationGroupMessage() :
    objects_count( 0 ),
    nonarchivable( false )
    {
    }


MigrationGroupMessage::MigrationGroupMessage
    (
    const FindObjectMessage & find_msg
    ) :
    Super( find_msg ),
    objects_count( 0 ),
    nonarchivable( false )
    {
    }


MigrationGroupMessage::MigrationGroupMessage
    (
    const MigrationGroupMessage & msg
    ) :
    Super( msg ),
    objects_count( msg.objects_count ),
    objects( msg.objects ),
    message_stream( msg.message_stream ),
    nonarchivable( msg.nonarchivable )
    {
    }


MigrationGroupMessage & MigrationGroupMessage::operator=
    (
    const MigrationGroupMessage & msg
    )
    {
    Super::operator=( msg );
    objects_count  = msg.objects_count;
    objects        = msg.objects;
    message_stream = msg.message_stream;
    nonarchivable  = msg.nonarchivable;
    return *this;
    }


SystemMessage::Type MigrationGroupMessage::get_type() const
    {
    if( nonarchivable )
        {
        return SystemMessage::MIGRATION_GROUP_NONARCHIVABLE;
        }
    else
        {
        return SystemMessage::MIGRATION_GROUP;
        }
    }


SystemMessage::ExecuteResult MigrationGroupMessage::execute()
    {
    ObjectManager &      object_manager = Global::object_manager();
    object_manager.handle_incoming_migration_group_message( *this );

    return SystemMessage::EXECUTE_OK;
    }


void MigrationGroupMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    unsigned int i;

    // Note: We can't serialize entire original message_stream
    // because some content of this packet could be changed.
    
    hope( objects_count > 0 );
    hope( objects_count == objects.size() );

    // Write super class.
    Super::bitstr_write( bos, desc );
    
    // Write object count.
    objects_count.bitstr_write( bos, desc );
    
    // Write object info of every object.
    for( i = 0; i < objects_count; i++ )
        {
        objects[ i ].bitstr_write( bos, desc );
        }

    // Align output stream to byte boundary.
    bos.align_to_byte();

    // Write serialized objects.
    // We must create a copy of message_stream because this->bitstr_write() is const.
    BitIStream      s( message_stream );
    s.rewind();
    bos.write_bitistream( s );
    }


void MigrationGroupMessage::bitstr_read
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    unsigned int        i;
    ObjectId            object_id;

    objects.clear();

    // Read super class.
    Super::bitstr_read( bis, desc );

    // Read object count.
    objects_count.bitstr_read( bis, desc );
    
    // Check object count.
    if( objects_count <= 0 )
        {
        throw ExSerializationError( "Empty migration group message." );
        }
    
    objects.reserve( objects_count );

    // Read object info for each object.
    for( i = 0; i < objects_count; i++ )
        {
        object_id.bitstr_read( bis, desc );
        objects.push_back( object_id );
        }

    // Align the stream to byte boundary.
    bis.align_to_byte();

    // Store rest of the message stream.
    message_stream = bis.tail();
    }


std::ostream & operator<<
    (
    std::ostream & s,
    const MigrationGroupMessage & msg
    )
    {
    s   << "sender " << msg.get_sender()
        << ", initiator " << msg.get_initiator()
        << ", destination " << msg.destination
        << ", hops count " << msg.hops
        << ", provider hops count " << msg.provider_hops
        << ", archivable " << ( msg.nonarchivable ? "no" : "yes" )
        << ", objects " << msg.objects_count
        << " (";
    
    ObjectShared::ObjectIdList::const_iterator it;
    for( it = msg.objects.begin(); it != msg.objects.end(); ++it )
        {
        const ObjectId & object_id = *it;
        s   << " " << object_id;
        }
    
    s   << ")";
    return s;
    }


}; // namespace Massiv::Core
}; // namespace Massiv
