/** @file migration_group_msg.h
 * Migration group message.
 */

/*
 * begin        : Tue Apr 30 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_MIGRATION_GROUP_MSG_H
#define MASSIV_MIGRATION_GROUP_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_BITIOSTREAM_FWD_H
#include "net/bitiostream_fwd.h"
#endif

#ifndef MASSIV_BITISTREAM_H
#include "net/bitistream.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_OBJECT_COMMON_H
#include "object/object_common.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#ifndef MASSIV_FIND_OBJECT_MSG_H
#include "sysevent/find_object_msg.h"
#endif

#include <vector>
#include <iosfwd>


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * MigrationGroup message packet.
 * Used for object migrating.
 *
 * There are two variants of migration group messages -
 * archivable and nonarchivable.
 * Both of the variants uses same class MigrationGroupMessage,
 * but have different types - MIGRATION_GROUP and MIGRATION_GROUP_NONARCHIVABLE.
 * Each of the type has different pause masks (see SystemMessage). This feature
 * is used by archivation and sync RPC mechanism.
 *
 * To create nonarchivable migration group, create MigrationGroupMessage
 * object and then call make_nonarchivable() method. Once the nonarchivable
 * flag is set to true, it can't be set back to false.
 *
 * Once the message is made nonarchivable, it changes its type
 * from MIGRATION_GROUP to MIGRATION_GROUP_NONARCHIVABLE
 * (the nonarchivable flag becomes part of the message type).
 * This behaviour is required to make working distributed archivation with
 * synch RPC mechanism and to make archivable and nonarchivable migration
 * messages to use one class (MigrationGroupMessage).
 *
 * @warning When calling copy constructor
 * from ancestor of MigrationGroup, make sure
 * that the compiler will not call
 * explicit copy construtor!
 */
class MigrationGroupMessage : public FindObjectMessage
    {
    typedef FindObjectMessage     Super;

public:

    /**
     * Constructor. 
     */
    MigrationGroupMessage();

    /** 
     * Constructor.
     */
    explicit MigrationGroupMessage
        (
        const FindObjectMessage & find_object_message
        );

    /**
     * Copy constructor.
     * @note Buffer of @a message_stream is not duplicated but stays shared.
     *
     * @warning When calling copy constructor
     * from ancestor of MigrationGroup, make sure
     * that the compiler will call this constructor
     * (not explicit copy construtor)!
     */
    MigrationGroupMessage
        (
        const MigrationGroupMessage & msg
        );

    /**
     * Operator=.
     * @note Buffer of @a message_stream is not duplicated but stays shared.
     */
    MigrationGroupMessage & operator=
        (
        const MigrationGroupMessage & msg
        );

public:

    //
    // From SystemMessage.
    //

    /**
     *
     */
    virtual SystemMessage::Type get_type() const;

    /**
     *
     */
    virtual SystemMessage::ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );
        
public:

    /**
     * Convert the message to nonarchivable.
     */
    void make_nonarchivable()
        {
        nonarchivable = true;
        }

public:

    SInt< UInt32, 32 >                      objects_count;
        /**< Number of objects stored in the group message. */

    ObjectShared::ObjectIdList              objects;
        /**< Object ids of objects stored in the group message. 
         * First object is the object that forced the migration group
         * to migrate.
         * @todo There is no more the requirement that the message must
         * containt at least one object. Replace asserts with conditions etc.
         * Does enybody need empty migration group messages?
         */

    BitIStream                              message_stream;
        /**< Stored message stream. Contains serialized objects. */
    
    bool                                    nonarchivable;
        /**< Whether this migration group is nonarchivable.
         * Once this flag is set to true, it can't be set back to false.
         * This flags is not serialized and thus is not transfered over
         * network. Once the message is made nonarchivable, it changes
         * its type from MIGRATION_GROUP to MIGRATION_GROUP_NONARCHIVABLE
         * and so the nonarchivable flag becomes part of the message type.
         */

    }; // MigrationGroupMessage


/**
 * Write information about migration message to output message.
 */
std::ostream & operator<<
    (
    std::ostream & s,
    const MigrationGroupMessage & msg
    );

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_MIGRATION_GROUP_MSG_H
