/** @file node_action_msg.h
 * Various node action messages.
 */

/*
 * begin        : Sat Oct 4 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_NODE_ACTION_MSG_H
#define MASSIV_NODE_ACTION_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_NODE_ACTION_ID_H
#include "node/node_action_id.h"
#endif

#ifndef MASSIV_STYPE_INT_H
#include "serializer/stype_int.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * Base class for node action messages.
 * Holds identification of an node action.
 *
 * @ingroup node
 */
class NodeActionMessage : public SystemMessage
    {
    /// Superclass.
    typedef SystemMessage Super;
    
public:

    /** Constructor. */
    NodeActionMessage();
    
    /** Constructor. */
    NodeActionMessage( const NodeActionId & action_id );
    
    /** Copy constructor. */
    NodeActionMessage( const NodeActionMessage & other );
    
    /** Operator=. */
    NodeActionMessage & operator=( const NodeActionMessage & other );
    
public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );
    
public:
    NodeActionId    action_id;
        /**< Node action identification. */

    }; // NodeActionMessage


/**
 * Node action request message.
 *
 * @ingroup node
 */
class NodeActionRequestMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

    /** Constructor. */
    NodeActionRequestMessage()
        {
        }
        
    /** Constructor. */
    NodeActionRequestMessage( const NodeActionId & the_action_id, UInt32 counter ) :
        Super( the_action_id ),
        action_counter( counter )
        {
        }
    
    /** Copy constructor. */
    NodeActionRequestMessage( const NodeActionRequestMessage & other ) :
        Super( other ),
        action_counter( other.action_counter )
        {
        }
        
    /** Operator=. */
    NodeActionRequestMessage & operator=( const NodeActionRequestMessage & other )
        {
        Super::operator=( other );
        action_counter = other.action_counter;
        return *this;
        }
        
public:

    // From SystemMessage.
    virtual ExecuteResult execute();
    
    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:
    
    SInt< UInt32, 32 >  action_counter;
        /**< Action counter used to identify obsolete messages. */

    }; // NodeActionRequestMessage


/**
 * NodeActionRequestReplyMessage.
 *
 * @ingroup node
 */
class NodeActionRequestReplyMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

    /** Constructor. */
    NodeActionRequestReplyMessage()
        {
        }
        
    /** Constructor. */
    NodeActionRequestReplyMessage( const NodeActionId & the_action_id, UInt32 counter ) :
        Super( the_action_id ),
        action_counter( counter )
        {
        }

    /** Copy constructor. */
    NodeActionRequestReplyMessage( const NodeActionRequestReplyMessage & other ) :
        Super( other ),
        action_counter( other.action_counter )
        {
        }

    /** Operator=. */
    NodeActionRequestReplyMessage & operator=( const NodeActionRequestReplyMessage & other )
        {
        Super::operator=( other );
        action_counter = other.action_counter;
        return *this;
        }

public:

    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

    // From SystemMessage.
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:
    
    SInt< UInt32, 32 >  action_counter;
        /**< Action counter used to identify obsolete messages. */

    }; // NodeActionRequestReplyMessage


/**
 * NodeActionCoordinatorSelectedMessage.
 *
 * @ingroup node
 */
class NodeActionCoordinatorSelectedMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

    /** Constructor. */
    NodeActionCoordinatorSelectedMessage()
        {
        }

    /** Constructor. */
    NodeActionCoordinatorSelectedMessage( const NodeActionId & the_action_id ) :
        Super( the_action_id )
        {
        }

    /** Copy constructor. */
    NodeActionCoordinatorSelectedMessage( const NodeActionCoordinatorSelectedMessage & other ) :
        Super( other )
        {
        }

    /** Operator=. */
    NodeActionCoordinatorSelectedMessage & operator=( const NodeActionCoordinatorSelectedMessage & other )
        {
        Super::operator=( other );
        return *this;
        }
    
public:

    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

    // From SystemMessage.
    virtual ExecuteResult execute();
    
    }; // NodeActionCoordinatorSelectedMessage
    

/**
 * NodeActionPrepareFinishMessage.
 *
 * @ingroup node
 */
class NodeActionPrepareFinishMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

    /** Constructor. */
    NodeActionPrepareFinishMessage()
        {
        }

    /** Constructor. */
    NodeActionPrepareFinishMessage( const NodeActionId & the_action_id ) :
        Super( the_action_id )
        {
        }

    /** Copy constructor. */
    NodeActionPrepareFinishMessage( const NodeActionPrepareFinishMessage & other ) :
        Super( other )
        {
        }

    /** Operator=. */
    NodeActionPrepareFinishMessage & operator=( const NodeActionPrepareFinishMessage & other )
        {
        Super::operator=( other );
        return *this;
        }
        
public:

    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

    // From SystemMessage.
    virtual ExecuteResult execute();
    
    }; // NodeActionPrepareFinishMessage


/**
 * NodeActionPrepareFinishReplyMessage.
 *
 * @ingroup node
 */
class NodeActionPrepareFinishReplyMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

    /** Constructor. */
    NodeActionPrepareFinishReplyMessage()
        {
        }

    /** Constructor. */
    NodeActionPrepareFinishReplyMessage( const NodeActionId & the_action_id ) :
        Super( the_action_id )
        {
        }

    /** Copy constructor. */
    NodeActionPrepareFinishReplyMessage( const NodeActionPrepareFinishReplyMessage & other ) :
        Super( other )
        {
        }

    /** Operator=. */
    NodeActionPrepareFinishReplyMessage & operator=( const NodeActionPrepareFinishReplyMessage & other )
        {
        Super::operator=( other );
        return *this;
        }
        
public:

    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

    // From SystemMessage.
    virtual ExecuteResult execute();
    
    }; // NodeActionPrepareFinishReplyMessage


/**
 * NodeActionFinishMessage.
 *
 * @ingroup node
 */
class NodeActionFinishMessage : public NodeActionMessage
    {
    /// Superclass.
    typedef NodeActionMessage Super;
    
public:

     /** Constructor. */
    NodeActionFinishMessage()
        {
        }

    /** Constructor. */
    NodeActionFinishMessage( const NodeActionId & the_action_id ) :
        Super( the_action_id )
        {
        }

    /** Copy constructor. */
    NodeActionFinishMessage( const NodeActionFinishMessage & other ) :
        Super( other )
        {
        }

    /** Operator=. */
    NodeActionFinishMessage & operator=( const NodeActionFinishMessage & other )
        {
        Super::operator=( other );
        return *this;
        }
   
public:

    // From SystemMessage.
    virtual SystemMessage::Type get_type() const;

    // From SystemMessage.
    virtual ExecuteResult execute();
    
    }; // NodeActionFinishMessage

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_NODE_ACTION_MSG_H
