/** @file replication_msg.cpp
 * Replication-related messages (implementation).
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "sysevent/replication_msg.h"
#include "object/object_manager.h"
#include "replication/replica_client.h"
#include "replication/replica_server.h"
#include "serializer/stype_int.h"
#include "net/bitistream.h"
#include "net/bitostream.h"


namespace Massiv {
namespace Core {


// class ReplicationUpdateMessage

ReplicationUpdateMessage::ReplicationUpdateMessage()
      : tcp_ack( false )
    {
    }


ReplicationUpdateMessage::ReplicationUpdateMessage
    (
    const ReplicationUpdateMessage & other
    ) : Super( other ),
        tcp_ack( other.tcp_ack ),
        replica_time( other.replica_time ),
        simulation_time( other.simulation_time ),
        objects_to_update( other.objects_to_update ),
        unchanged_objects( other.unchanged_objects ),
        objects_to_destroy( other.objects_to_destroy ),
        objects( other.objects )
    {
    }


ReplicationUpdateMessage & ReplicationUpdateMessage::operator=
    (
    const ReplicationUpdateMessage & other
    )
    {
    Super::operator=( other );
    tcp_ack = other.tcp_ack;
    replica_time = other.replica_time;
    simulation_time = other.simulation_time;
    objects_to_update = other.objects_to_update;
    unchanged_objects = other.unchanged_objects;
    objects_to_destroy = other.objects_to_destroy;
    objects = other.objects;
    return *this;
    }


SystemMessage::Type ReplicationUpdateMessage::get_type() const
    {
    return SystemMessage::REPLICATION_UPDATE;
    }


SystemMessage::ExecuteResult ReplicationUpdateMessage::execute()
    {
    ObjectManager & object_manager = Global::object_manager();
    hope( object_manager.replica_client.get() );
    object_manager.replica_client->handle_replication_update_message( *this );
    return EXECUTE_OK;
    }


void ReplicationUpdateMessage::bitstr_write
    (
    BitOStream &                        bos,
    const Serializer::Description &     desc
    ) const
    {   
    hope( !is_empty() );

    // Serialize superclass.

    Super::bitstr_write( bos, desc );

    // Write tcp_ack.

    tcp_ack.bitstr_write( bos, desc );

    // Write times.

    replica_time.bitstr_write( bos, desc );
    simulation_time.bitstr_write( bos, desc );

    // Write lists of objects to update and to destroy.

    const SVlUInt32 update_count( objects_to_update.size() );
    const SVlUInt32 unchanged_count( unchanged_objects.size() );
    const SVlUInt32 destroy_count( objects_to_destroy.size() );
    update_count.bitstr_write( bos, desc );
    unchanged_count.bitstr_write( bos, desc );
    destroy_count.bitstr_write( bos, desc );

    ObjectList::const_iterator object_it;
    for( object_it = objects_to_update.begin();
        object_it != objects_to_update.end(); ++object_it )
        {
        object_it->bitstr_write( bos, desc );
        }
    for( object_it = unchanged_objects.begin();
        object_it != unchanged_objects.end(); ++object_it )
        {
        object_it->bitstr_write( bos, desc );
        }
    for( object_it = objects_to_destroy.begin();
        object_it != objects_to_destroy.end(); ++object_it )
        {
        object_it->bitstr_write( bos, desc );
        }

    // Write serialized replicas.
    // Note: must do copy because @c this is const and @c write_bitistream
    //       requires non-const parameter. Stupid.

    BitIStream stream( objects );
    stream.rewind();
    bos.align_to_byte();
    bos.write_bitistream( stream );
    }


void ReplicationUpdateMessage::bitstr_read
    (
    BitIStream &                        bis,
    const Serializer::Description &     desc
    )
    {
    // Read superclass.

    Super::bitstr_read( bis, desc );

    // Read tcp_ack.

    tcp_ack.bitstr_read( bis, desc );

    // Read times.

    replica_time.bitstr_read( bis, desc );
    simulation_time.bitstr_read( bis, desc );

    // Read list sizes.

    SVlUInt32 update_count;
    SVlUInt32 unchanged_count;
    SVlUInt32 destroy_count;
    update_count.bitstr_read( bis, desc );
    unchanged_count.bitstr_read( bis, desc );
    destroy_count.bitstr_read( bis, desc );
    // FIXME: throw if counts are == 0?

    // Read object lists.

    unsigned i;
    objects_to_update.clear();
    for( i = 0; i < update_count; i++ )
        {
        ObjectInfo object_info;
        object_info.bitstr_read( bis, desc );
        objects_to_update.push_back( object_info );
        }

    unchanged_objects.clear();
    for( i = 0; i < unchanged_count; i++ )
        {
        ObjectInfo object_info;
        object_info.bitstr_read( bis, desc );
        unchanged_objects.push_back( object_info );
        }
    
    objects_to_destroy.clear();
    for( i = 0; i < destroy_count; i++ )
        {
        ObjectInfo object_info;
        object_info.bitstr_read( bis, desc );
        objects_to_destroy.push_back( object_info );
        }

    // Read serialized replicas.

    bis.align_to_byte();
    objects = bis.tail();
    }


int ReplicationUpdateMessage::guess_size() const
    {
    int bit_count = 0;

    // Size of SystemMessage. FIXME: Use Super::get_size() instead.

    bit_count += SystemMessage::TYPE_BITS_COUNT;
    bit_count += NodeId::TYPE_BITS_COUNT;
    bit_count += NodeId::RELATIVE_ID_BITS_COUNT;

    // tcp_ack.

    bit_count += 1;

    // ReplicaTime and STime.

    bit_count += 8 * sizeof( ReplicaTime::TimeType );
    bit_count += 8 * sizeof( STime::TimeType );

    // List sizes.

    bit_count += 3 * 16;

    // Object lists.

    bit_count += objects_to_update.size() * ( 64 + 16 ); // object id + object version
    bit_count += unchanged_objects.size() * ( 64 + 16 );
    bit_count += objects_to_destroy.size() * ( 64 + 16 );

    // Done.

    return ( bit_count + 7 ) / 8 + objects.byte_size();
    }


// class ReplicationAckMessage

ReplicationAckMessage::ReplicationAckMessage()
    {
    }


ReplicationAckMessage::ReplicationAckMessage
    (
    const ReplicationAckMessage & other
    ) : Super( other ),
        replica_time( other.replica_time )
    {
    }


ReplicationAckMessage & ReplicationAckMessage::operator=
    (
    const ReplicationAckMessage & other
    )
    {
    Super::operator=( other );
    replica_time = other.replica_time;
    return *this;
    }


SystemMessage::Type ReplicationAckMessage::get_type() const
    {
    return SystemMessage::REPLICATION_ACK;
    }


SystemMessage::ExecuteResult ReplicationAckMessage::execute()
    {
    ObjectManager & object_manager = Global::object_manager();
    hope( object_manager.replica_server.get() );
    object_manager.replica_server->handle_replication_ack_message( *this );
    return EXECUTE_OK;
    }


void ReplicationAckMessage::bitstr_write
    (
    BitOStream &                        bos,
    const Serializer::Description &     desc
    ) const
    {
    // Serialize superclass.

    Super::bitstr_write( bos, desc );

    // Write time.

    replica_time.bitstr_write( bos, desc );
    }


void ReplicationAckMessage::bitstr_read
    (
    BitIStream &                        bis,
    const Serializer::Description &     desc
    )
    {
    // Read superclass.

    Super::bitstr_read( bis, desc );

    // Read time.

    replica_time.bitstr_read( bis, desc );
    }


} // namespace Massiv::Core
} // namespace Massiv


#ifndef ENABLE_INLINE
#include "sysevent/replication_msg_inline.h"
#endif
