/** @file replication_msg.h
 * Replication-related messages.
 */

/*
 * begin        : Feb 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REPLICATION_MSG_H
#define MASSIV_REPLICATION_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_MSG_H
#include "sysevent/system_msg.h"
#endif

#ifndef MASSIV_OBJECTID_H
#include "pointer/objectid.h"
#endif

#ifndef MASSIV_REPLICA_TIME_H
#include "replication/replica_time.h"
#endif

#ifndef MASSIV_OBJECT_FWD_H
#include "object/object_fwd.h"
#endif

#ifndef MASSIV_STYPE_BOOL_H
#include "serializer/stype_bool.h"
#endif

#ifndef MASSIV_STYPE_VLI_H
#include "serializer/stype_vli.h" // ObjectVersion
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#ifndef MASSIV_BITISTREAM_H
#include "net/bitistream.h"
#endif

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * Message containing object replicas.
 *
 * @ingroup replication
 */
class ReplicationUpdateMessage : public SystemMessage
    {
    /// Superclass.
    typedef SystemMessage   Super;

public:

    /**
     * Per-object info.
     */
    class ObjectInfo
        {

    public:

        /**
         * Constructor.
         */
        INLINE ObjectInfo();

        /**
         * Constructor.
         */
        INLINE ObjectInfo
            (
            const ObjectId &    the_object_id,
            ObjectVersion       the_major_version
            );

    public:

        /**
         * Write to bitstream.
         */
        INLINE void bitstr_write
            (
            BitOStream &                        bos,
            const Serializer::Description &     desc
            ) const;

        /**
         * Read from bitstream.
         */
        INLINE void bitstr_read
            (
            BitIStream &                        bis,
            const Serializer::Description &     desc
            );
            
    public:
    
        ObjectId                object_id;
            /**< Object id. */
        ObjectVersion           major_version;
            /**< Object major version. */

        }; // class ObjectInfo

    /**
     * List of object infos.
     */
    typedef std::vector< ObjectInfo > ObjectList;

public:

    /**
     * Constructor.
     */
    ReplicationUpdateMessage();

    /**
     * Copy constructor.
     */
    ReplicationUpdateMessage
        (
        const ReplicationUpdateMessage & other
        );

    /**
     * Assignment operator.
     */
    ReplicationUpdateMessage & operator=
        (
        const ReplicationUpdateMessage & other
        );

public:

    /**
     * Get message type.
     */
    virtual SystemMessage::Type get_type() const;

    /**
     * Process message.
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     *
     * @pre Message is non-empty, ie. objects_to_update or
     *      objects_to_destroy is non-empty.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    /**
     * Is the message empty? (ie. both #objects_to_update and
     * #objects_to_destroy are empty)
     */
    INLINE bool is_empty() const;

    /**
     * Try to guess message size in bytes.
     *
     * Note that this is just a guess - it does not include size of data added
     * by network layer and it may be wrong if SystemMessage or serializers are
     * updated. Unfortunately we must compute this number before the message is
     * written to bitstream.
     */
    int guess_size() const;

    /**
     * Add object to list of objects to update.
     * @pre The @a object_id has valid #ObjectInfo part.
     */
    INLINE void add_object_to_update
        (
        const ObjectId &    object_id,
        ObjectVersion       major_version
        );

    /**
     * Add object to list of objects that are alive but have not changed since
     * last acked update.
     * @pre The @a object_id has valid #ObjectInfo part.
     */
    INLINE void add_unchanged_object
        (
        const ObjectId &    object_id,
        ObjectVersion       major_version
        );

    /**
     * Add object to list of objects to destroy.
     */
    INLINE void add_object_to_destroy
        (
        const ObjectId &    object_id,
        ObjectVersion       major_version
        );

public:

    SBool           tcp_ack;
        /**< Send ReplicationAckMessage reply via TCP? */

    ReplicaTime     replica_time;
        /**< Replication timestamp of sender node. */

    STime           simulation_time;
        /**< Simulation time of the sending node. */

    ObjectList      objects_to_update;
        /**< List of objects to update. */

    ObjectList      unchanged_objects;
        /**< List of living objects that have not changed since last update. */

    ObjectList      objects_to_destroy;
        /**< List of objects to destroy.
             @note Class type ids of the objects to destroy are invalid. */

    BitIStream      objects;
        /**< Serialized replicas of objects to update. */

    }; // class ReplicationUpdateMessage

    
/**
 * Ack as a reply to ReplicationUpdateMessage.
 *
 * @ingroup replication
 */
class ReplicationAckMessage : public SystemMessage
    {
    /// Superclass.
    typedef SystemMessage   Super;

public:

    /**
     * Constructor.
     */
    ReplicationAckMessage();

    /**
     * Copy constructor.
     */
    ReplicationAckMessage
        (
        const ReplicationAckMessage & other
        );

    /**
     * Assignment operator.
     */
    ReplicationAckMessage & operator=
        (
        const ReplicationAckMessage & other
        );

public:

    /**
     * Get message type.
     */
    virtual SystemMessage::Type get_type() const;

    /**
     * Process message.
     */
    virtual ExecuteResult execute();

public:

    /**
     * Serialize to bitstream.
     *
     * @pre Message is non-empty, ie. updated_objects or
     *      destroyed_objects is non-empty.
     */
    virtual void bitstr_write
        (
        BitOStream &                        bos,
        const Serializer::Description &     desc
        ) const;

    /**
     * Read from bitstream.
     */
    virtual void bitstr_read
        (
        BitIStream &                        bis,
        const Serializer::Description &     desc
        );

public:

    ReplicaTime     replica_time;
        /**< Equal to #RepliactionUpdateMessage::replica_time of acked message. */

    }; // class ReplicationAckMessage

/** @} */ // addtogroup sysevent
    

} // namespace Massiv::Core
} // namespace Massiv


#ifdef ENABLE_INLINE
#include "sysevent/replication_msg_inline.h"
#endif


#endif // MASSIV_REPLICATION_MSG_H
