/** @file system_event_queue.cpp
 * System event queue.
 */

/*
 * begin        : Sat Jun 27 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "sysevent/system_event_queue.h"
#include "sysevent/system_msg.h"
#include "database/node_database.h"
#include "global/global.h"
#include "utility/debug.h"
#include "node/node_manager.h"
#include <sstream>


namespace Massiv {
namespace Core {

namespace {

/**
 * Called when an exception is thrown while executing a message.
 */
void message_execute_exception
    (
    SystemMessage & msg,
    std::exception & e
    )
    {
    std::ostringstream oss;
    oss << "Exception during executing " << SystemMessage::type_to_string( msg.get_type() )
        << " message from " << msg.get_sender() << " (initiator "
        << msg.get_initiator() << "). " << e.what();
    
    // Kill the node.
    Global::node_manager().request_kill_node( msg.get_sender() );
    }

} // anonymous namespace


/**
 * Class used to initialize SystemEventQueue.
 */
class SystemEventQueueServerNodeEnumerator : public NodeDatabase::Enumerator
    {
public:
    SystemEventQueueServerNodeEnumerator
        (
        SystemEventQueue & _queue
        ) :
        queue( _queue )
        {
        }
        
    virtual void operator()
        (
        const NodeDatabase::NodeData & node_data
        )
        {
        NodeId  node_id = node_data.get_node_id();
        bool    result;
        
        hope( node_id.get_type() == NodeId::TYPE_SERVER );

        // Try to insert new empty SystemMessageQueue for the node.
        result = queue.node_messages.insert
                    (
                    SystemEventQueue::NodeMessageQueueHash::value_type
                        (
                        node_id,
                        SystemMessageQueue()
                        )
                    ).second;
                    
        if( !result )
            {
            // The 'node_messages' already contain a SystemMessageQueue for
            // this node. We leave the the SystemMessageQueue untouched.
            // See update_server_node_queues().
            // Nothing to do here.
            }
        }

public:

    SystemEventQueue &  queue;
        
    }; // ServerNodeEnumerator


// Implementation of SystemEventQueue
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SystemEventQueue::SystemEventQueue()
    {
    }


SystemEventQueue::~SystemEventQueue()
    {
    }


void SystemEventQueue::update_server_node_queues()
    {
    // We will enumerate all server nodes and create system message queues
    // for each of them. We may not overwrite existing queues, because
    // 'node_messages' can already contain some messages.
    // SystemEventQueueServerNodeEnumerator takes care of it.
    // Note: 'events' and 'common_messages' stays unchanged.
    
    NodeDatabase &                          node_database = Global::node_database();
    SystemEventQueueServerNodeEnumerator    enumerator( *this );
    
    // Enumerate all server nodes and create one SystemMessageQueue for each of them.
    node_database.enumerate_nodes( NodeId::TYPE_SERVER, enumerator );
    }


void SystemEventQueue::reset()
    {
    events = EventQueue();
    common_messages = SystemMessageQueue(); // This will clear pause mask.
    node_messages.clear();
    update_server_node_queues();
    }


void SystemEventQueue::push_back_event
    (
    const SystemEvent & e
    )
    {
    events.push( e );
    }


void SystemEventQueue::push_back_message
    (
    const Reference< SystemMessage > & system_msg
    )
    {
    hope( !system_msg.is_null() );

    const NodeId &                  msg_sender = system_msg->get_sender();
    NodeMessageQueueHash::iterator  it = node_messages.find( msg_sender );
    if( it != node_messages.end() )
        {
        hope( msg_sender.get_type() == NodeId::TYPE_SERVER );
        
        // The message was received from server node.
        // Push it to corresponding system message queue.
        SystemMessageQueue & queue = it->second;
        queue.push_back( system_msg );
        }
    else
        {
        hope( msg_sender.get_type() != NodeId::TYPE_SERVER );
        
        // The message was received from non server node.
        // Push it to common message queue.
        common_messages.push_back( system_msg );
        }
    }


void SystemEventQueue::set_server_pause_mask
    (
    const NodeId &  server_node_id,
    unsigned int    new_pause_mask
    )
    {
    hope( server_node_id.get_type() == NodeId::TYPE_SERVER );
    NodeMessageQueueHash::iterator it = node_messages.find( server_node_id );
    hope( it != node_messages.end() );
    SystemMessageQueue & queue = it->second;
    queue.set_pause_mask( new_pause_mask );
    }


unsigned int SystemEventQueue::get_server_pause_mask
    (
    const NodeId & server_node_id
    ) const
    {
    hope( server_node_id.get_type() == NodeId::TYPE_SERVER );
    NodeMessageQueueHash::const_iterator it = node_messages.find( server_node_id );
    hope( it != node_messages.end() );
    const SystemMessageQueue & queue = it->second;
    return queue.get_pause_mask();
    }


void SystemEventQueue::execute_all_events()
    {
    TAG_STACK_FRAME( "::Massiv::Core::SystemEventQueue::execute_all_events()" );
    
    // Execute all messages received from server nodes.
    execute_server_messages();

    // Execute all messages received from non server nodes.
    execute_non_server_messages();

    // Execute all events from local node.
    execute_events();
    }


void SystemEventQueue::execute_server_messages()
    {
    NodeMessageQueueHash::iterator it;

    for( it = node_messages.begin(); it != node_messages.end(); ++it )
        {
        SystemMessageQueue &        queue = it->second;
        
        // Execute all unpaused messages.
        while( !queue.empty() )
            {
            Reference< SystemMessage >  msg = queue.front();
            SystemMessage::ExecuteResult result;
        
            try 
                {
                result = msg->execute();
                }
            catch( std::exception & e )
                {
                message_execute_exception( *msg, e );
                result = SystemMessage::EXECUTE_OK;
                }

            switch( result )
                {
                case SystemMessage::EXECUTE_OK:
                    // If set_pause_mask() was called from inside execute(),
                    // we don't know if the message is still in front of the queue.
                    // If set_pause_mask() was called, there are two possibilities:
                    // 1/ The message is one of the messages that were paused ie.
                    //    the message is now in paused queues.
                    // 2/ Some messages were unpaused so our message (maybe) is not
                    //    in front of the queue any more.
                    // Case 1/ is serious error in the implementation of the
                    // execute() method and should never happen.
                    if( queue.front() == msg )
                        {
                        queue.pop_front();
                        }
                    else
                        {
                        // We must remove the message from the queue.
                        // This is slow, but should not happen very often.
                        const NodeId & node_id = it->first;
                        
                        // Log.
                            {
                            std::ostringstream oss;
                            oss << "System message '"
                                << SystemMessage::type_to_string( msg->get_type() )
                                << "' received from " << node_id
                                << " has paused/unpaused system message queue "
                                << "of the node. The message is not in front "
                                << "of the queue any more. The message will be removed from the queue now.";
                            Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, oss.str() );
                            }

                        queue.remove_message( msg.get() );
                        }
                    break;
                case SystemMessage::EXECUTE_KEEP_MESSAGE_IN_QUEUE:
                    // Do not pop the message from the queue.
                    // The message itself must ensure that we will not
                    // stay in neverending loop (mostly by pausing/unpausing
                    // the sender queue).
                    break;
                default:
                    panic( "Invalid value returned from SystemMessage::execute()." );
                    break;
                }
            }
        }
    }


void SystemEventQueue::execute_non_server_messages()
    {
    SystemMessageQueue & queue = common_messages;
    while( !queue.empty() )
        {
        Reference< SystemMessage > msg = queue.front();
        SystemMessage::ExecuteResult result;
        
        try 
            {
            result = msg->execute();
            }
        catch( std::exception & e )
            {
            message_execute_exception( *msg, e );
            result = SystemMessage::EXECUTE_OK;
            }
        
        switch( result )
            {
            case SystemMessage::EXECUTE_OK:
                queue.pop_front();
                break;
                
            case SystemMessage::EXECUTE_KEEP_MESSAGE_IN_QUEUE:
                // Because queue for non server nodes can't be paused or unpaused,
                // returning this value from SystemMessage::execute() is internal error.
                    {
                    std::ostringstream oss;
                    oss << "EXECUTE_KEEP_MESSAGE_IN_QUEUE received from SystemMessage::execute() while processing "
                        << SystemMessage::type_to_string( msg->get_type() ) << " message from "
                        << msg->get_sender() << ". Probably internal error in massiv core.";
                    Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
                    // FIXME: Kill the node?
                    }
                // Ignore (otherwise we will get to neverending loop).
                queue.pop_front();
                break;
                
            default:
                panic( "Invalid value returned from SystemMessage::execute()." );
                break;
            }
        }
    }


void SystemEventQueue::execute_events()
    {
    while( !events.empty() )
        {
        SystemEvent e = events.front();
        events.pop();
        e.execute();
        }
    }
        

} // namespace Massiv::Core
} // namespace Massiv
