/** @file system_event_queue.h
 * System event queue.
 */

/*
 * begin        : Sat Jun 27 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_SYSTEM_EVENT_QUEUE_H
#define MASSIV_SYSTEM_EVENT_QUEUE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_H
#include "sysevent/system_event.h"
#endif

#ifndef MASSIV_SYSTEM_MESSAGE_QUEUE_H
#include "sysevent/system_message_queue.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#include <queue>


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * System event queue.
 *
 * System event queue manages events which come from local system node or remote
 * system nodes.
 *
 * System event received from remote nodes can be only system message
 * (see @c SystemMessage).
 * System event received from local node can be system message, user input,
 * message from operating system etc.
 *
 * System event queue just simple takes one event by one and executes
 * them. See @c SystemEvent for more inforamtion.
 *
 * System messages received from remote server nodes are inserted to
 * system message queues (see @c SystemMessageQueue). Every server node
 * has its own system message queue. System message queue support
 * message pausing. Only server nodes has system message queues
 * because message pausing is available only for server nodes, not for
 * client or other nodes. Mesage pausing is ability to pause some
 * system messages from some node (until the messages are unpaused).
 * For more information about message pausing see @c SystemMessageQueue
 * and @c SystemMessage.
 *
 * Messages from nodes that are not servers are inserted into
 * one common message queue. These messages can not be paused.
 *
 * System events from local system node other than system messages are inserted
 * into its own queue.
 *
 * @note If the local node is server node, it has its own system message queue.
 *
 * @todo FIXME Optimize. Current implementation in every tick pools
 * every system message queue if there are some new messages available.
 * This is good if there is small number of server nodes.
 * In the future, system message queues should somehow notify system event queue
 * that they have some new messages so the system event queue will pool only
 * non-empty message queues. Note: This will not require to modify system message queue
 * because only system event queue pushs and pops data to system message queues.
 */
class SystemEventQueue
    {
    // Class used to initialize SystemEventQueue.
    // Defined in system_event_queue.cpp.
    friend class SystemEventQueueServerNodeEnumerator;
    
public:

    /** Constructor. */
    SystemEventQueue();

    /** Destructor. */
    ~SystemEventQueue();

    /**
     * This function enumaretes all server nodes and create #SystemMessageQueue
     * for each of them.
     *
     * This function can be called more than once.
     * In that case system message queues which are already created for some
     * nodes stays unachanged.
     *
     * @warning Once the server nodes are enumerated, their system message queues
     *      are never removed from this #SystemEventQueue.
     *      If a system message queue exists for some server node and this function
     *      is called but the server nodes is not enumerated again, the system message
     *      queue for this node stays in this #SystemEventQueue with all messages
     *      it contains.
     */
    void update_server_node_queues();
    
    /**
     * Wipe-out all messages from all queues and clear all pause masks of server queues.
     */
    void reset();

public:

    /**
     * Insert system event in the queue.
     */
    void push_back_event
        (
        const SystemEvent & event
        );

    /**
     * Insert system message in the queue.
     * @pre @a system_message is valid reference to system message.
     * @note Dependent of the message sender the message
     * is pushed to system message queue associated with the sender.
     */
    void push_back_message
        (
        const Reference< SystemMessage > & system_message
        );

    /**
     * Set new pause mask of server message queue.
     * @pre @a server_node_id is server node (type is @c NodeId::TYPE_SERVER).
     * @see SystemMessageQueue.
     */
    void set_server_pause_mask
        (
        const NodeId & server_node_id,
        unsigned int   new_pause_mask
        );

    /**
     * Get pause mask of server message queue.
     * @pre @a server_node_id is server node (type is @c NodeId::TYPE_SERVER).
     * @see SystemMessageQueue.
     */
    unsigned int get_server_pause_mask
        (
        const NodeId & server_node_id
        ) const;

    /**
     * Execute all events waiting in the queue.
     * Each event is first removed from the queue and then executed.
     */
    void execute_all_events();

private:

    /**
     * Execute all messages received from server nodes.
     */
    void execute_server_messages();

    /**
     * Execute all messages received from non server nodes.
     */
    void execute_non_server_messages();

    /**
     * Execute all events from local node.
     */
    void execute_events();

private:

    /**
     * Queue of system events.
     */
    typedef std::queue< SystemEvent >   EventQueue;

    /**
     * Hash map which associates @c NodeId with one @c SystemMessageQueue.
     * @note Once the hash map is created and initialized, no message queues are added or
     * removed from it. Due to this, storing SystemMessageQueue by value in the hash
     * map is not ineffecient.
     */
    typedef ext::hash_map
                <
                NodeId,
                SystemMessageQueue,
                NodeId::Hash,
                NodeId::EqualKey
                > NodeMessageQueueHash;

private:

    EventQueue              events;
        /**< Queue of pending events which were inserted by push_back_event().
         */

    SystemMessageQueue      common_messages;
        /**< Queue of system messages received from non-server nodes
         * which were inserted by push_back_message().
         * @note Pause mask is never changed for this queue.
         */

    NodeMessageQueueHash    node_messages;
        /**< List of system message queues for server nodes inserted
         * by push_back_message().
         * Each server node has one system message queue in this list.
         */

    }; // SystemEventQueue

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SYSTEM_EVENT_QUEUE_H
