/** @file system_message_queue.cpp
 * System message queue.
 */

/*
 * begin        : Mon Jan 13 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

 
#include "sysevent/system_message_queue.h"
#include "sysevent/system_msg.h"
#include "utility/debug.h"


namespace Massiv {
namespace Core {


// Implementation of SystemMessageQueue::MessageContainer::Less.
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

bool SystemMessageQueue::MessageContainer::Less::operator()
    (
    const MessageContainer & container1,
    const MessageContainer & container2
    ) const
    {
    return container1.order_id < container2.order_id;
    }


// Implementation of SystemMessageQueue::MessageContainer.
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SystemMessageQueue::MessageContainer::MessageContainer() :
    order_id( 0 )
    {
    }


SystemMessageQueue::MessageContainer::MessageContainer
    (
    unsigned int                        _order_id,
    const Reference< SystemMessage > &  _system_msg
    ) :
    order_id( _order_id ),
    system_msg( _system_msg )
    {
    }
        

SystemMessageQueue::MessageContainer::MessageContainer
    (
    const MessageContainer & c
    ) :
    order_id( c.order_id ),
    system_msg( c.system_msg )
    {
    }


SystemMessageQueue::MessageContainer::~MessageContainer()
    {
    }


SystemMessageQueue::MessageContainer & SystemMessageQueue::MessageContainer::operator=
    (
    const MessageContainer & c
    )
    {
    order_id = c.order_id;
    system_msg = c.system_msg;
    return *this;
    }     


// Implementation of SystemMessageQueue.
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SystemMessageQueue::SystemMessageQueue() :
    new_order_id( 0 ),
    pause_mask( 0 )
    {
    }


SystemMessageQueue::SystemMessageQueue
    (
    const SystemMessageQueue & queue
    ) :
    unpaused_queue( queue.unpaused_queue ),
    paused_queue( queue.paused_queue ),
    new_order_id( queue.new_order_id ),
    pause_mask( queue.pause_mask )
    {
    }

    
SystemMessageQueue::~SystemMessageQueue()
    {
    }


SystemMessageQueue & SystemMessageQueue::operator=
    (
    const SystemMessageQueue & queue
    )
    {
    unpaused_queue = queue.unpaused_queue;
    paused_queue = queue.paused_queue;
    new_order_id = queue.new_order_id;
    pause_mask = queue.pause_mask;
    return *this;
    }


void SystemMessageQueue::push_back
    (
    const Reference< SystemMessage > & system_msg
    )
    {
    hope( !system_msg.is_null() );

    // Create message container and assign unique order-id to it.
    MessageContainer    container( new_order_id, system_msg );
    new_order_id++;
    hope( new_order_id != 0 ); // Owerflow should never happen, unsigned int is soooo big.

    // Insert the container into one of the subqueues.
    if( is_message_paused( system_msg.get() ) )
        {
        paused_queue.push_back( container );
        }
    else
        {
        unpaused_queue.push_back( container );
        }
    }


Reference< SystemMessage > SystemMessageQueue::front()
    {
    hope( !unpaused_queue.empty() );
    MessageContainer container = unpaused_queue.front();
    return container.system_msg;
    }


void SystemMessageQueue::pop_front()
    {
    hope( !unpaused_queue.empty() );
    unpaused_queue.pop_front();
    }


unsigned int SystemMessageQueue::size() const
    {
    return unpaused_queue.size();
    }


bool SystemMessageQueue::empty() const
    {
    return unpaused_queue.empty();
    }
        

void SystemMessageQueue::set_pause_mask
    (
    unsigned int new_pause_mask
    )
    {
    if( pause_mask == new_pause_mask )
        return;
    
    MessageSet                              paused_set;
    MessageSet                              unpaused_set;
    std::pair< MessageSet::iterator, bool>  ret;

    pause_mask = new_pause_mask;

    // Move messages from paused subqueue to paused and unpaused sets.
    while( !paused_queue.empty() )
        {
        MessageContainer container = paused_queue.front();
        paused_queue.pop_front();
        if( is_message_paused( container.system_msg.get() ) )
            {
            ret = paused_set.insert( container );
            }
        else
            {
            ret = unpaused_set.insert( container );
            }
        hope( ret.second ); // Order-id is unique.
        }

    // Move messages from unpaused subqueue to paused and unpaused sets.
    while( !unpaused_queue.empty() )
        {
        MessageContainer container = unpaused_queue.front();
        unpaused_queue.pop_front();
        if( is_message_paused( container.system_msg.get() ) )
            {
            ret = paused_set.insert( container );
            }
        else
            {
            ret = unpaused_set.insert( container );
            }
        hope( ret.second ); // Order-id is unique.
        }

    // Recreate subqueues.
    // fixme: Would be paused_queue.insert( paused_queue.begin(), paused_set.begin(), paused_set.end() ) faster?
    paused_queue   = MessageQueue( paused_set.begin(), paused_set.end() );
    unpaused_queue = MessageQueue( unpaused_set.begin(), unpaused_set.end() );
    }


unsigned int SystemMessageQueue::get_pause_mask() const
    {                                       
    return pause_mask;
    }


bool SystemMessageQueue::is_message_paused
    (
    const SystemMessage * system_msg
    )
    {
    hope( system_msg );

    SystemMessage::Type msg_type = system_msg->get_type();
    unsigned int        msg_pause_mask = SystemMessage::get_message_pause_mask( msg_type );

    // If at least one bit is set in both message pause mask and queue pause mask
    // then the message is paused.
    if( ( msg_pause_mask & pause_mask ) != 0 )
        {
        return true;
        }
    else
        {
        return false;
        }
    }    


void SystemMessageQueue::remove_message
    (
    SystemMessage * msg
    )
    {
    hope( msg );
    int i;
    for( i = 0; i < 2; ++i )
        {
        MessageQueue & queue = ( i == 0 ? unpaused_queue : paused_queue );
        MessageQueue::iterator it;
        for( it = queue.begin(); it != queue.end(); ++it )
            {
            MessageContainer & container = *it;
            if( container.system_msg.get() == msg )
                {
                queue.erase( it );
                return;
                }
            }
        }
    
    panic( "::Massiv::Core::SystemMessageQueue::remove_message() failed - specified message not found!" );
    }


} // namespace Massiv::Core
} // namespace Massiv
