/** @file system_message_queue.h
 * System message queue.
 */

/*
 * begin        : Mon Jan 13 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SYSTEM_MESSAGE_QUEUE_H
#define MASSIV_SYSTEM_MESSAGE_QUEUE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#include <deque>
#include <set>


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * SystemMessageQueue stores ordered list of system messages.
 * The messages are ordered in order they are inserted in the queue.
 *
 * SystemMessageQueue has ability to pause and unpause some messages.
 * Paused messages are selected by pause mask.
 * Every SystemMessageQueue has its own pause mask.
 * Every system message has associated bits from the mask which force
 * the message to be paused.
 *
 * SystemMessageQueue has its own pause mask which is used to detect
 * which messages must be paused or unpaused.
 * Value of 0 means that all messages will be unpaused.
 *
 * The main feature of SystemMessageQueue is that it keeps
 * order of messages which are being paused or unpaused.
 *
 * For example lets have messages (in this order) @b A, @b B, @b C,
 * @b D, @b E and @b F. <br>
 * Now if we pause @a B, @b C and @b E, the queue will contain
 * two groups (subqueues) of messages. Paused subqueue will contain
 * (in this order) @b B, @b C and @b E. Unpaused subqueue will contain
 * (in this order) @b A, @b D and @b F. <br>
 * Now if we pause message @b D, the queue will contain
 * in paused subqueue @b B, @b C, @b D and @b E and in unpaused
 * subqueue @b A and @b F. <br>
 * Now if we unpause all messages, the queue will contain
 * only unpaused subqueue which contains all messages
 * (in this order) @b A, @b B, @b C, @b D, @b E and @b F.
 *
 * @note The ability of keeping right order of the messages
 * are required by various massiv protocols (for example
 * archivation protocol) to work in (mostly obscure but not
 * limited to) all possible situations.
 */
class SystemMessageQueue
    {
public:

    /**
     * Default constructor.
     */
    SystemMessageQueue();

    /**
     * Copy constructor.
     */
    SystemMessageQueue
        (
        const SystemMessageQueue & queue
        );

    /**
     * Destructor.
     */
    ~SystemMessageQueue();

    /**
     * Operator=.
     */
    SystemMessageQueue & operator=
        (
        const SystemMessageQueue & queue
        );

public:

    /**
     * Insert system message to the queue.
     * @pre @a system_message is valid reference to a system message.
     */
    void push_back
        (
        const Reference< SystemMessage > & system_message
        );

    /**
     * Get first waiting (unpaused) system message from the queue.
     * The message itself is not removed from the queue.
     * Only unpaused message is returned.
     * @pre Unpaused subqueue must be non-empty.
     * @note Changing pause mask between front() and pop_front()
     * can change first unpaused message in the queue so take
     * care of it.
     */
    Reference< SystemMessage > front();

    /**
     * Removes first (unpaused) system message from the queue.
     * @pre Unpaused subqueue must be non-empty.
     * @note Changing pause mask between front() and pop_front()
     * can change first unpaused message in the queue so take
     * care of it.
     */
    void pop_front();

    /**
     * Get number of message in unpaused subqueue.
     * @note This function ignores paused subqueue. So this
     * function can return zero even if the paused subqueue
     * is non-empty.
     */
    unsigned int size() const;

    /**
     * Check whether the unpaused subqueue is empty.
     * @note This function ignores paused subqueue. So this
     * function can return true even if the paused subqueue
     * is non-empty.
     */
    bool empty() const;

    /**
     * Set new pause mask of the queue.
     * @note This operation is fast if new_pause_mask is same as current pause mask (ie. nothing
     * is done).
     * @note This operation may be slow.
     */
    void set_pause_mask
        (
        unsigned int new_pause_mask
        );

    /**
     * Get current pause mask of the queue.
     */
    unsigned int get_pause_mask() const;
    
    /**
     * Remove specified message from the queue.
     * @pre msg != NULL
     * @pre The queue contains the message.
     */
    void remove_message
        (
        SystemMessage * msg
        );

private:

    /**
     * Check whether the message is paused.
     * Return value is computed based on the current queue pause mask
     * and message pause mask.
     * @pre @a system_message is valid pointer.
     */
    bool is_message_paused
        (
        const SystemMessage * system_message
        );

private:

    /**
     * Message container is used to store system message.
     * It also contains order-id of the message. This order-id
     * is used to keep right order if the message in the queue.
     */
    class MessageContainer
        {
    public:

        /**
         * Less comparator.
         */
        struct Less
            {
            bool operator()
                (
                const MessageContainer & container1,
                const MessageContainer & container2
                ) const;
            }; // Less
            
    public:

        /** Default constructor. */
        MessageContainer();

        /** Constructor. */
        MessageContainer
            (
            unsigned int                       order_id,
            const Reference< SystemMessage > & system_message
            );

        /** Copy constructor. */
        MessageContainer
            (
            const MessageContainer & container
            );

        /** Destructor. */
        ~MessageContainer();

        /** Operator=. */
        MessageContainer & operator=
            (
            const MessageContainer & container
            );
            
    public:

        unsigned int                order_id;
            /**< Order-id used to keep right order of container in the queue.
             * Every message stored in the queue has associated unique order-id. */

        Reference< SystemMessage >  system_msg;
            /**< System message stored in the container. */
        
        }; // MessageContainer

private:

    /**
     * MessageQueue is used to store system messages.
     * @note We don't use here some sorted container
     * because this should be faster because sorting
     * is required only during pausing/unpasing of messages
     * and this should happen rarely.
     * Also message queue should contain only small number of messages
     * because one message queue is associated with only one server node
     * (see @c SystemEventQueue).
     */
    typedef std::deque
                <
                MessageContainer
                > MessageQueue;

    /**
     * MessageSet is used to store system message in
     * order by their order-id (see @c MessageContainer).
     * This container is used mainly for union of paused
     * and unpaused messages into one @C MessageQueue.
     */
    typedef std::set
                <                
                MessageContainer,
                MessageContainer::Less
                > MessageSet;

private:

    MessageQueue    unpaused_queue;
        /**< Subqueue which stores messages which are not paused. */

    MessageQueue    paused_queue;
        /**< Subqueue which stores messages which are paused. */

    unsigned int    new_order_id;
        /**< New order id value. This value will be assigned to new message
         * inserted to the queue. */

    unsigned int    pause_mask;
        /**< Current pause mask for the queue. All system messages
         * which have at least one bit set which is also set in the @c pause_mask
         * are paused. */

    }; // SystemMessageQueue

/** @} */ // addtogroup sysevent


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SYSTEM_MESSAGE_QUEUE_H
