/** @file system_msg.cpp
 * System message common stuff.
 */

/*
 * begin        : Sat Jun 27 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "sysevent/system_msg.h"
#include "net/bitistream.h"
#include "net/bitostream.h"
#include "serializer/serializer.h"
#include "serializer/serializer_int.h"
#include "global/global.h"
#include "utility/debug.h"
#include "sysevent/object_status_msg.h"
#include "sysevent/find_object_msg.h"
#include "sysevent/migration_group_msg.h"
#include "sysevent/replication_msg.h"
#include "sysevent/archivation_msg.h"
#include "node/node_manager.h"
#include "sysevent/forward_msg.h"
#include "sysevent/time_synch_msg.h"
#include "sysevent/simulation_sync_msg.h"
#include "sysevent/client_connection_msg.h"
#include "sysevent/global_search_msg.h"
#include "sysevent/data_msg.h"
#include "sysevent/system_event_queue.h"
#include "sysevent/account_msg.h"
#include "sysevent/node_action_msg.h"
#include "sysevent/balancer_msg.h"
#include "status/status.h"
#include "system/system.h"


namespace Massiv {
namespace Core {


/**
 * Verification types.
 */
enum VerificationType
    {
    VT_SENDER_NODE_TYPE,
        /**< Verifies whether sender node of a message can be sender
         * of the message.
         */
        
    VT_RECEIVER_NODE_TYPE,
        /**< Verifies whether receiver node of a message can be receiver
         * of the message.
         */
        
    VT_INITIATOR_NODE_TYPE,
        /**< Verifies whether initiator node of a message can be initiator
         * of the message.
         */

    VT_SENDER_RECEIVER_RELATIONSHIP
        /**<  Sender-initiator (SI) relationship verification.
         * SI verification is used to verify requirements for relationship
         * between message initiator and message sender.
         */
    }; // VerificationType


/**
 * Sender-initiator (SI) relationship verification results.
 */
enum SIVerificationResult
    {
    SI_INITIATOR_EQUAL_TO_SENDER,
        // Message initiator must be same as message sender.
    SI_INITIATOR_EQUAL_TO_SENDER_EQUAL_TO_LOCAL_NODE,
        // Message initiator must be same as message sender and
        // it must be local node.
    SI_ANY_INITIATOR,
        // Message initiator doesn't have to be same as message sender
        // and it can be any node type.
    }; // SIVerificationResult


/**
 * Flags for each node type used by sri verification.
 */
enum NodeTypeFlags
    {
    NTF_SERVER          = 0x01,
        /**< Flag for server node type. */

    NTF_CLIENT          = 0x02,
        /**< Flag for client node type. */

    NTF_SERVICE         = 0x04,
        /**< Flag for service node type. */
    
    NTF_ANONYMOUS       = 0x08
        /**< Flag for anonymous node type. */

    }; // SRIVerificationNodeType


/**
 * Factory that creates system message object.
 */
template< class Type >
struct SystemMessageFactory
    {
    static SystemMessage * create()
        {
        return new Type;
        }
    }; // SystemMessageFactory
    

/**
 * Factory that creates nonarchivable MigrationGroupMessage.
 * Nonarchivable migration group messages are handles in a special way,
 * for more info see documentation of MigrationGroupMessage.
 */
struct NonarchivableMigrationGroupFactory
    {
    static SystemMessage * create()
        {
        MigrationGroupMessage * msg = new MigrationGroupMessage();
        msg->make_nonarchivable();
        hope( msg->get_type() == SystemMessage::MIGRATION_GROUP_NONARCHIVABLE );
        return msg;
        }
    }; // NonarchivableMigrationGroupFactory

    
typedef SystemMessage * CreateSystemMessage();
    
    
/**
 * Structure stored in system message table for each message type.
 * Contains miscenalleous data used for message verification,
 * message creation etc.
 */
struct SystemMessageTableItem
    {
public:

    /** Constructor. */
    SystemMessageTableItem() :
        create( NULL ),
        sender_flags( 0 ),
        initiator_flags( 0 ),
        receiver_flags( 0 ),
        required_si_verification_result( SI_INITIATOR_EQUAL_TO_SENDER ),
        pause_mask( 0 )
        {
        }
        
public:

    CreateSystemMessage *   create;
        /**< Function that creates system message of this type. */
    
    std::string             name;
        /**< Name of the message type. */
    
    unsigned int            sender_flags;
        /**< These flags specifies which node type can send the message. */
        
    unsigned int            initiator_flags;
        /**< These flags specifies which node type can be initiator of the message. */
        
    unsigned int            receiver_flags;
        /**< These flags specifies which node type can receive the message. */
        
    SIVerificationResult    required_si_verification_result;
        /**< SI verification stuff. */
        
    unsigned int            pause_mask;
        /**< Value returned by SystemMessage::get_message_pause_mask() function.
         * Every value specifies pause mask - bits which are set
         * will force the message to be paused in system message queue.
         */

    }; // SystemMessageTableItem


typedef std::vector< SystemMessageTableItem > SystemMessageTable;

//-----------------------------------------------------------------------------

static SystemMessageTable   msg_table;
static bool                 msg_table_inited = false;

//-----------------------------------------------------------------------------


void SystemMessage::initialize_system_message_table()
    {
    hope( msg_table.empty() );
    hope( !msg_table_inited );
    
    msg_table.resize( TYPE_COUNT );
    
        {
        int i;
        for( i = 0; i < TYPE_COUNT; ++i )
            {
            hope( !msg_table[ i ].create );
            msg_table[ i ].name = "RESERVED";
            }
        }
    
    // PAUSE_WAITING_FOR_SIMULATION_READY
    // Because simulation is started asap _first_ StartSimulationMessage
    // is received, PAUSE_WAITING_FOR_SIMULATION_READY will delay the message
    // until local simulation is ready to go.
    
    msg_table[ OBJECT_STATUS ].create = SystemMessageFactory< ObjectStatusMessage >::create;
    msg_table[ OBJECT_STATUS ].name = "ObjectStatus";
    msg_table[ OBJECT_STATUS ].sender_flags    = NTF_SERVER;
    msg_table[ OBJECT_STATUS ].initiator_flags = NTF_SERVER;
    msg_table[ OBJECT_STATUS ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ OBJECT_STATUS ].required_si_verification_result = SI_ANY_INITIATOR;
    msg_table[ OBJECT_STATUS ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ FIND_OBJECT ].create = SystemMessageFactory< FindObjectMessage >::create;
    msg_table[ FIND_OBJECT ].name = "FindObject";
    msg_table[ FIND_OBJECT ].sender_flags    = NTF_SERVER | NTF_CLIENT;
    msg_table[ FIND_OBJECT ].initiator_flags = NTF_SERVER | NTF_CLIENT;
    msg_table[ FIND_OBJECT ].receiver_flags  = NTF_SERVER;
    msg_table[ FIND_OBJECT ].required_si_verification_result = SI_ANY_INITIATOR;
    msg_table[ FIND_OBJECT ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    // When updated don't forget to update MIGRATION_GROUP_NONARCHIVABLE message.
    msg_table[ MIGRATION_GROUP ].create = SystemMessageFactory< MigrationGroupMessage >::create;
    msg_table[ MIGRATION_GROUP ].name = "MigrationGroup";
    msg_table[ MIGRATION_GROUP ].sender_flags    = NTF_SERVER | NTF_CLIENT;
    msg_table[ MIGRATION_GROUP ].initiator_flags = NTF_SERVER | NTF_CLIENT;
    msg_table[ MIGRATION_GROUP ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ MIGRATION_GROUP ].required_si_verification_result = SI_ANY_INITIATOR;
    msg_table[ MIGRATION_GROUP ].pause_mask = 0 | 
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY |
        SystemMessage::PAUSE_ARCHIVATION_FINISHING_START |
                // Pause all migration groups until current archivations is finished.
        SystemMessage::PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED;
                // Pause all "start-new-archivation" messages until pending synd
                // rpc is finished.
                
    
    // Nonarchivable migration group message has special factory because
    // nonarchivable and archivable migration messages uses same class even if
    // the messages have different types.
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].create = NonarchivableMigrationGroupFactory::create;
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].name = "MigrationGroup(nonarchivable)";
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].sender_flags    =
        msg_table[ MIGRATION_GROUP ].sender_flags;
        
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].initiator_flags =
        msg_table[ MIGRATION_GROUP ].initiator_flags;
        
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].receiver_flags  =
        msg_table[ MIGRATION_GROUP ].receiver_flags;
        
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].required_si_verification_result =
        msg_table[ MIGRATION_GROUP ].required_si_verification_result;
        
    msg_table[ MIGRATION_GROUP_NONARCHIVABLE ].pause_mask = 0 | 
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY |
        SystemMessage::PAUSE_ARCHIVATION_FINISHING_START;
                // Pause all migration groups until current archivations is finished.
                // Note: Because this is nonarchivable migration group, we do not
                // set PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED flag.
    
    msg_table[ REPLICATION_UPDATE ].create = SystemMessageFactory< ReplicationUpdateMessage >::create;
    msg_table[ REPLICATION_UPDATE ].name = "ReplicationUpdate";
    msg_table[ REPLICATION_UPDATE ].sender_flags    = NTF_SERVER;
    msg_table[ REPLICATION_UPDATE ].initiator_flags = NTF_SERVER;
    msg_table[ REPLICATION_UPDATE ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ REPLICATION_UPDATE ].required_si_verification_result = SI_ANY_INITIATOR;
    msg_table[ REPLICATION_UPDATE ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ REPLICATION_ACK ].create = SystemMessageFactory< ReplicationAckMessage >::create;
    msg_table[ REPLICATION_ACK ].name = "ReplicationAck";
    msg_table[ REPLICATION_ACK ].sender_flags    = NTF_SERVER | NTF_CLIENT;
    msg_table[ REPLICATION_ACK ].initiator_flags = NTF_SERVER | NTF_CLIENT;
    msg_table[ REPLICATION_ACK ].receiver_flags  = NTF_SERVER;
    msg_table[ REPLICATION_ACK ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER; // fixme - is this correct?
    msg_table[ REPLICATION_ACK ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;
    
    msg_table[ ARCHIVATION_STATUS ].create = SystemMessageFactory< ArchivationStatusMessage >::create;
    msg_table[ ARCHIVATION_STATUS ].name = "ArchivationStatus";
    msg_table[ ARCHIVATION_STATUS ].sender_flags    = NTF_SERVER;
    msg_table[ ARCHIVATION_STATUS ].initiator_flags = NTF_SERVER;
    msg_table[ ARCHIVATION_STATUS ].receiver_flags  = NTF_SERVER;
    msg_table[ ARCHIVATION_STATUS ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ARCHIVATION_STATUS ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY |
        SystemMessage::PAUSE_ARCHIVATION_FINISHING_START |
                // Pause all "start-new-archivation" messages until current archivation
                // is finished.
        SystemMessage::PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED;
                // Pause all "start-new-archivation" messages until pending synd
                // rpc is finished.
              

    // ForwardMessage and ForwardedMessage
    // ForwardMessage is used only for inital send. Once received, only ForwardedMessage
    // is used.
    msg_table[ FORWARD ].create = SystemMessageFactory< ForwardedMessage >::create;
    msg_table[ FORWARD ].name = "Forward";
    msg_table[ FORWARD ].sender_flags    = NTF_SERVER;
    msg_table[ FORWARD ].initiator_flags = NTF_SERVER;
    msg_table[ FORWARD ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ FORWARD ].required_si_verification_result = SI_ANY_INITIATOR;
    msg_table[ FORWARD ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ TIME_SYNCHRONIZATION ].create = SystemMessageFactory< TimeSynchronizationMessage >::create;
    msg_table[ TIME_SYNCHRONIZATION ].name = "TimeSynchronization";
    msg_table[ TIME_SYNCHRONIZATION ].sender_flags    = NTF_SERVER;
    msg_table[ TIME_SYNCHRONIZATION ].initiator_flags = NTF_SERVER;
    msg_table[ TIME_SYNCHRONIZATION ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ TIME_SYNCHRONIZATION ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ TIME_SYNCHRONIZATION ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;
        
    msg_table[ STARTUP ].create = SystemMessageFactory< StartupMessage >::create;
    msg_table[ STARTUP ].name = "Startup";
    msg_table[ STARTUP ].sender_flags    = NTF_SERVER;
    msg_table[ STARTUP ].initiator_flags = NTF_SERVER;
    msg_table[ STARTUP ].receiver_flags  = NTF_SERVER;
    msg_table[ STARTUP ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ STARTUP ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_INITIAL_CONNECTIONS;
                // Delay NodeReadyMessage until all connections are established.

    /** @todo FIXME Pro boovieho: Procistit sender, initiator a receiver flagy pro vsechny
     * data zpravy. Nevim jak funguje data server, takze sem vsude nafrkal
     * NTF_SERVICE aby to jelo.
     */

    msg_table[ DATA_VERSION_REQUEST ].create = SystemMessageFactory< DataVersionRequestMessage >::create;
    msg_table[ DATA_VERSION_REQUEST ].name = "DataVersionRequest";
    msg_table[ DATA_VERSION_REQUEST ].sender_flags    = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_VERSION_REQUEST ].initiator_flags = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_VERSION_REQUEST ].receiver_flags  = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_VERSION_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_VERSION_REQUEST ].pause_mask = 0 |
        0; /** @todo FIXME */
    
    msg_table[ DATA_REQUEST ].create = SystemMessageFactory< DataRequestMessage >::create;
    msg_table[ DATA_REQUEST ].name = "DataRequest";
    msg_table[ DATA_REQUEST ].sender_flags    = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_REQUEST ].initiator_flags = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_REQUEST ].receiver_flags  = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_REQUEST ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ DATA_VERSION ].create = SystemMessageFactory< DataVersionMessage >::create;
    msg_table[ DATA_VERSION ].name = "DataVersion";
    msg_table[ DATA_VERSION ].sender_flags    = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_VERSION ].initiator_flags = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_VERSION ].receiver_flags  = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_VERSION ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_VERSION ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ DATA ].create = SystemMessageFactory< DataMessage >::create;
    msg_table[ DATA ].name = "Data";
    msg_table[ DATA ].sender_flags    = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA ].initiator_flags = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA ].receiver_flags  = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ DATA_CONFIRMATION ].create = SystemMessageFactory< DataConfirmationMessage >::create;
    msg_table[ DATA_CONFIRMATION ].name = "DataConfirmation";
    msg_table[ DATA_CONFIRMATION ].sender_flags    = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_CONFIRMATION ].initiator_flags = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_CONFIRMATION ].receiver_flags  = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_CONFIRMATION ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_CONFIRMATION ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ DATA_ERROR ].create = SystemMessageFactory< DataErrorMessage >::create;
    msg_table[ DATA_ERROR ].name = "DataError";
    msg_table[ DATA_ERROR ].sender_flags    = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_ERROR ].initiator_flags = NTF_SERVER | NTF_SERVICE;
    msg_table[ DATA_ERROR ].receiver_flags  = NTF_SERVER | NTF_CLIENT | NTF_SERVICE | NTF_ANONYMOUS;
    msg_table[ DATA_ERROR ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_ERROR ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ DATA_UPDATE_REQUEST ].create = SystemMessageFactory< DataUpdateRequestMessage >::create;
    msg_table[ DATA_UPDATE_REQUEST ].name = "DataUpdateRequest";
    msg_table[ DATA_UPDATE_REQUEST ].sender_flags    = NTF_SERVER;
    msg_table[ DATA_UPDATE_REQUEST ].initiator_flags = NTF_SERVER;
    msg_table[ DATA_UPDATE_REQUEST ].receiver_flags  = NTF_SERVICE;
    msg_table[ DATA_UPDATE_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ DATA_UPDATE_REQUEST ].pause_mask = 0 |
        0; /** @todo FIXME */

    msg_table[ ACCEPT_CLIENT ].create = SystemMessageFactory< AcceptClientMessage >::create;
    msg_table[ ACCEPT_CLIENT ].name = "AcceptClient";
    msg_table[ ACCEPT_CLIENT ].sender_flags    = NTF_SERVER;
    msg_table[ ACCEPT_CLIENT ].initiator_flags = NTF_SERVER;
    msg_table[ ACCEPT_CLIENT ].receiver_flags  = NTF_SERVER;
    msg_table[ ACCEPT_CLIENT ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ACCEPT_CLIENT ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;
    
    msg_table[ CLIENT_ACCEPTED ].create = SystemMessageFactory< ClientAcceptedMessage >::create;
    msg_table[ CLIENT_ACCEPTED ].name = "ClientAccepted";
    msg_table[ CLIENT_ACCEPTED ].sender_flags    = NTF_SERVER;
    msg_table[ CLIENT_ACCEPTED ].initiator_flags = NTF_SERVER;
    msg_table[ CLIENT_ACCEPTED ].receiver_flags  = NTF_SERVER;
    msg_table[ CLIENT_ACCEPTED ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ CLIENT_ACCEPTED ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;
    
    msg_table[ CLIENT_READY ].create = SystemMessageFactory< ClientReadyMessage >::create;
    msg_table[ CLIENT_READY ].name = "ClientReady";
    msg_table[ CLIENT_READY ].sender_flags    = NTF_SERVER;
    msg_table[ CLIENT_READY ].initiator_flags = NTF_SERVER;
    msg_table[ CLIENT_READY ].receiver_flags  = NTF_SERVER;
    msg_table[ CLIENT_READY ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ CLIENT_READY ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ CLIENT_READY_REPLY ].create = SystemMessageFactory< ClientReadyReplyMessage >::create;
    msg_table[ CLIENT_READY_REPLY ].name = "ClientReadyReply";
    msg_table[ CLIENT_READY_REPLY ].sender_flags    = NTF_SERVER;
    msg_table[ CLIENT_READY_REPLY ].initiator_flags = NTF_SERVER;
    msg_table[ CLIENT_READY_REPLY ].receiver_flags  = NTF_SERVER;
    msg_table[ CLIENT_READY_REPLY ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ CLIENT_READY_REPLY ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ KILL_CLIENT ].create = SystemMessageFactory< KillClientMessage >::create;
    msg_table[ KILL_CLIENT ].name = "KillClient";
    msg_table[ KILL_CLIENT ].sender_flags    = NTF_SERVER;
    msg_table[ KILL_CLIENT ].initiator_flags = NTF_SERVER;
    msg_table[ KILL_CLIENT ].receiver_flags  = NTF_SERVER;
    msg_table[ KILL_CLIENT ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ KILL_CLIENT ].pause_mask = 0 |
        0;

    msg_table[ WELCOME ].create = SystemMessageFactory< WelcomeMessage >::create;
    msg_table[ WELCOME ].name = "Welcome";
    msg_table[ WELCOME ].sender_flags    = NTF_SERVER;
    msg_table[ WELCOME ].initiator_flags = NTF_SERVER;
    msg_table[ WELCOME ].receiver_flags  = NTF_CLIENT;
    msg_table[ WELCOME ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ WELCOME ].pause_mask = 0 |
        0;
            // Welcome message can be sent only to client, so no pause mask is required here.

    msg_table[ GLOBAL_SEARCH ].create = SystemMessageFactory< GlobalSearchMessage>::create;
    msg_table[ GLOBAL_SEARCH ].name = "GlobalSearch";
    msg_table[ GLOBAL_SEARCH ].sender_flags    = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH ].initiator_flags = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH ].receiver_flags  = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ GLOBAL_SEARCH ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ GLOBAL_SEARCH_REPLY ].create = SystemMessageFactory< GlobalSearchReplyMessage >::create;
    msg_table[ GLOBAL_SEARCH_REPLY ].name = "GlobalSearchReply";
    msg_table[ GLOBAL_SEARCH_REPLY ].sender_flags    = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH_REPLY ].initiator_flags = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH_REPLY ].receiver_flags  = NTF_SERVER;
    msg_table[ GLOBAL_SEARCH_REPLY ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ GLOBAL_SEARCH_REPLY ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;

    msg_table[ TIME_MAGIC_CONSTANT ].create = SystemMessageFactory< TimeMagicConstantMessage >::create;
    msg_table[ TIME_MAGIC_CONSTANT ].name = "TimeMagicConstant";
    msg_table[ TIME_MAGIC_CONSTANT ].sender_flags    = NTF_SERVER;
    msg_table[ TIME_MAGIC_CONSTANT ].initiator_flags = NTF_SERVER;
    msg_table[ TIME_MAGIC_CONSTANT ].receiver_flags  = NTF_SERVER | NTF_CLIENT;
    msg_table[ TIME_MAGIC_CONSTANT ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ TIME_MAGIC_CONSTANT ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY;
        
    msg_table[ ACCOUNT_REGISTER_REQUEST ].create = SystemMessageFactory< AccountRegisterRequestMessage >::create;
    msg_table[ ACCOUNT_REGISTER_REQUEST ].name = "AccountRegisterRequest";
    msg_table[ ACCOUNT_REGISTER_REQUEST ].sender_flags    = NTF_ANONYMOUS;
    msg_table[ ACCOUNT_REGISTER_REQUEST ].initiator_flags = NTF_ANONYMOUS;
    msg_table[ ACCOUNT_REGISTER_REQUEST ].receiver_flags  = NTF_SERVICE;
    msg_table[ ACCOUNT_REGISTER_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ACCOUNT_REGISTER_REQUEST ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY; // FIXME

    msg_table[ ACCOUNT_REGISTER_RESPOND ].create = SystemMessageFactory< AccountRegisterRespondMessage >::create;
    msg_table[ ACCOUNT_REGISTER_RESPOND ].name = "AccountRegisterRespond";
    msg_table[ ACCOUNT_REGISTER_RESPOND ].sender_flags    = NTF_SERVICE;
    msg_table[ ACCOUNT_REGISTER_RESPOND ].initiator_flags = NTF_SERVICE;
    msg_table[ ACCOUNT_REGISTER_RESPOND ].receiver_flags  = NTF_ANONYMOUS;
    msg_table[ ACCOUNT_REGISTER_RESPOND ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ACCOUNT_REGISTER_RESPOND ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY; // FIXME

    msg_table[ ACCOUNT_CREATE_REQUEST ].create = SystemMessageFactory< AccountCreateRequestMessage >::create;
    msg_table[ ACCOUNT_CREATE_REQUEST ].name = "AccountCreateRequest";
    msg_table[ ACCOUNT_CREATE_REQUEST ].sender_flags    = NTF_SERVICE;
    msg_table[ ACCOUNT_CREATE_REQUEST ].initiator_flags = NTF_SERVICE;
    msg_table[ ACCOUNT_CREATE_REQUEST ].receiver_flags  = NTF_SERVER;
    msg_table[ ACCOUNT_CREATE_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ACCOUNT_CREATE_REQUEST ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY; // FIXME

    msg_table[ ACCOUNT_CREATE_RESPOND ].create = SystemMessageFactory< AccountCreateRespondMessage >::create;
    msg_table[ ACCOUNT_CREATE_RESPOND ].name = "AccountCreateRespond";
    msg_table[ ACCOUNT_CREATE_RESPOND ].sender_flags    = NTF_SERVER;
    msg_table[ ACCOUNT_CREATE_RESPOND ].initiator_flags = NTF_SERVER;
    msg_table[ ACCOUNT_CREATE_RESPOND ].receiver_flags  = NTF_SERVICE;
    msg_table[ ACCOUNT_CREATE_RESPOND ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ ACCOUNT_CREATE_RESPOND ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_SIMULATION_READY; // FIXME
    
    //
    // Node action messages.
    // Most of node action messages have empty pause mask because
    // these messages may not be sent until a node action request message
    // is received. Thus we need set pause mask only for node action request messages.
    //
    
    msg_table[ NODE_ACTION_REQUEST ].create = SystemMessageFactory< NodeActionRequestMessage >::create;
    msg_table[ NODE_ACTION_REQUEST ].name = "NodeActionRequest";
    msg_table[ NODE_ACTION_REQUEST ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_REQUEST ].pause_mask = 0 |
        SystemMessage::PAUSE_WAITING_FOR_INITIAL_CONNECTIONS |
        SystemMessage::PAUSE_NODE_ACTION_REQUEST;
        // Note: PAUSE_WAITING_FOR_SIMULATION_READY can't be used for NODE_ACTION_REQUEST
        // because node action is used to make the simulation ready.
        // Server nodes use STARTUP node action to coordinate simulation startup.
        // Startup node action is initiated once connections to all server nodes are
        // established, thus NODE_ACTION_REQUEST uses PAUSE_WAITING_FOR_INITIAL_CONNECTIONS
        // (it will pause startup node action message until all connections are properly established).
        // Once simulation is ready, all servers will disable PAUSE_WAITING_FOR_SIMULATION_READY flag
        // from message queues that will 

    msg_table[ NODE_ACTION_REQUEST_REPLY ].create = SystemMessageFactory< NodeActionRequestReplyMessage >::create;
    msg_table[ NODE_ACTION_REQUEST_REPLY ].name = "NodeActionRequestReply";
    msg_table[ NODE_ACTION_REQUEST_REPLY ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST_REPLY ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST_REPLY ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_REQUEST_REPLY ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_REQUEST_REPLY ].pause_mask = 0;
        
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].create = SystemMessageFactory< NodeActionCoordinatorSelectedMessage >::create;
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].name = "NodeActionCoordinatorSelected";
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_COORDINATOR_SELECTED ].pause_mask = 0;

    msg_table[ NODE_ACTION_PREPARE_FINISH ].create = SystemMessageFactory< NodeActionPrepareFinishMessage >::create;
    msg_table[ NODE_ACTION_PREPARE_FINISH ].name = "NodeActionPrepareFinish";
    msg_table[ NODE_ACTION_PREPARE_FINISH ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_PREPARE_FINISH ].pause_mask = 0;

    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].create = SystemMessageFactory< NodeActionPrepareFinishMessage >::create;
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].name = "NodeActionPrepareFinishReply";
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_PREPARE_FINISH_REPLY ].pause_mask = 0;
    
    msg_table[ NODE_ACTION_FINISH ].create = SystemMessageFactory< NodeActionFinishMessage >::create;
    msg_table[ NODE_ACTION_FINISH ].name = "NodeActionFinish";
    msg_table[ NODE_ACTION_FINISH ].sender_flags    = NTF_SERVER;
    msg_table[ NODE_ACTION_FINISH ].initiator_flags = NTF_SERVER;
    msg_table[ NODE_ACTION_FINISH ].receiver_flags  = NTF_SERVER;
    msg_table[ NODE_ACTION_FINISH ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ NODE_ACTION_FINISH ].pause_mask = 0;

    //
    // Balancer messages.
    //
    
    msg_table[ RESOURCE_LOAD_MESSAGE ].create = SystemMessageFactory< ResourceLoadMessage >::create;
    msg_table[ RESOURCE_LOAD_MESSAGE ].name = "ResourceLoadMessage";
    msg_table[ RESOURCE_LOAD_MESSAGE ].sender_flags    = NTF_SERVER;
    msg_table[ RESOURCE_LOAD_MESSAGE ].initiator_flags = NTF_SERVER;
    msg_table[ RESOURCE_LOAD_MESSAGE ].receiver_flags  = NTF_SERVER;
    msg_table[ RESOURCE_LOAD_MESSAGE ].required_si_verification_result = SI_INITIATOR_EQUAL_TO_SENDER;
    msg_table[ RESOURCE_LOAD_MESSAGE ].pause_mask = 0;
    
    msg_table_inited = true;
    }

//-----------------------------------------------------------------------------

static unsigned int node_to_type_flags
    (
    const NodeId & node
    )
    {
    switch( node.get_type() )
        {
        case NodeId::TYPE_SERVER:
            return NTF_SERVER;
            
        case NodeId::TYPE_CLIENT:
            return NTF_CLIENT;
            
        case NodeId::TYPE_SERVICE:
            return NTF_SERVICE;
        
        case NodeId::TYPE_ANONYMOUS:
            return NTF_ANONYMOUS;
        }
        
    panic( "Invalid node type." );
    return 0;
    }

    
static bool verify_message_type
    (
    SystemMessage::Type msg_type
    )
    {
    return ( msg_type >= 0 && static_cast<int>( msg_type ) < SystemMessage::
        TYPE_COUNT );
    }


static bool verify_sender_node_type
    (
    SystemMessage::Type     msg_type,
    unsigned int            node_type_flags
    )
    {
    hope( msg_type >= 0 || msg_type < (int) SystemMessage::TYPE_COUNT );
    unsigned int flags = msg_table[ msg_type ].sender_flags;
    return ( ( flags & node_type_flags ) == node_type_flags );
    }


static bool verify_receiver_node_type
    (
    SystemMessage::Type     msg_type,
    unsigned int            node_type_flags
    )
    {
    hope( msg_type >= 0 || msg_type < (int) SystemMessage::TYPE_COUNT );
    unsigned int flags = msg_table[ msg_type ].receiver_flags;
    return ( ( flags & node_type_flags ) == node_type_flags );
    }


static bool verify_initiator_node_type
    (
    SystemMessage::Type     msg_type,
    unsigned int            node_type_flags
    )
    {
    hope( msg_type >= 0 || msg_type < (int) SystemMessage::TYPE_COUNT );
    unsigned int flags = msg_table[ msg_type ].initiator_flags;
    return ( ( flags & node_type_flags ) == node_type_flags );
    }


static bool verify_sender_initiator_relationship
    (
    SystemMessage::Type     msg_type,
    const NodeId &          sender,
    const NodeId &          initiator
    )
    {
    hope( msg_table_inited );
    hope( msg_type >= 0 && msg_type < (int) SystemMessage::TYPE_COUNT );
    
    bool failed = true;
    
    switch( msg_table[ msg_type ].required_si_verification_result )
        {
        case SI_INITIATOR_EQUAL_TO_SENDER:
            failed = ( sender != initiator );
            break;
            
        case SI_INITIATOR_EQUAL_TO_SENDER_EQUAL_TO_LOCAL_NODE:
            failed = ( ( sender != initiator ) || ( sender != Global::get_local_node_id() ) );
            break;
        
        case SI_ANY_INITIATOR:
            failed = false;
            break;
        }
        
    return !failed;
    }

//-----------------------------------------------------------------------------

/*
 * Report some verification failure.
 * @param receiver Should be local node id.
 */
static void report_verification_failure
    (
    VerificationType    verification_type,
    SystemMessage::Type msg_type,
    const NodeId *      sender,
    const NodeId *      initiator,
    const NodeId *      receiver
    )
    {
    std::string         verification_name;
    std::ostringstream  ss;
    
    switch( verification_type )
        {
        case VT_SENDER_NODE_TYPE:
            verification_name = "SENDER_NODE_TYPE";
            break;
        
        case VT_RECEIVER_NODE_TYPE:
            verification_name = "RECEIVER_NODE_TYPE";
            break;
            
        case VT_INITIATOR_NODE_TYPE:
            verification_name = "INITIATOR_NODE_TYPE";
            break;
            
        case VT_SENDER_RECEIVER_RELATIONSHIP:
            verification_name = "SENDER_RECEIVER_RELATIONSHIP";
            break;
        };
        
    ss  << "System message verification " << verification_name << " failed. "
        << "Message type " << SystemMessage::type_to_string( msg_type );
        
    ss  << ", sender ";
    if( sender )
        {
        ss  << ( *sender );
        }
    else
        {
        ss  << "unknown";
        }
        
    ss  << ", initiator ";
    if( initiator )
        {
        ss  << ( *initiator );
        }
    else
        {
        ss  << "unknown";
        }
        
    ss  << ", receiver";
    if( receiver )
        {
        ss  << ( *receiver );
        }
    else
        {
        ss  << "unknown";
        }
    
    ss  << ".";

    Global::log
        (
        STATUS_WARNING
            (
            Status::FACILITY_NETWORK, /** @todo Vlastni facility? */
            Status::PRIORITY_HIGHEST
            ),
        ss.str()
        );
    }


static void recover_from_invalid_message
    (
    const NodeId &  sender,
    const NodeId &  receiver // local node
    )
    {
    NodeManager & node_manager = Global::node_manager();
    node_manager.request_kill_node( sender );
    }


//-----------------------------------------------------------------------------
// SystemMessage implementation.
//-----------------------------------------------------------------------------


SystemMessage::SystemMessage()
    {
    sender    = Global::get_local_node_id();
    initiator = Global::get_local_node_id();
    }


SystemMessage::~SystemMessage()
    {
    }


SystemMessage::SystemMessage
    (
    const SystemMessage & msg
    ) :
    Super( msg ),
    sender( msg.sender ),
    initiator( msg.initiator )
    {
    }


SystemMessage & SystemMessage::operator=
    (
    const SystemMessage & msg
    )
    {
    Super::operator=( msg );
    sender    = msg.sender;
    initiator = msg.initiator;
    return *this;
    }


void SystemMessage::bitstr_write
    (
    BitOStream &                    bos,
    const Serializer::Description & desc
    ) const
    {
    /* Sender is not transferred over network. It is for local
     * purposes only.
     */
    
    SInt<Int32, TYPE_BITS_COUNT> t( get_type() );
    hope( t >= 0 && t < TYPE_COUNT );
    
    t.bitstr_write( bos, desc );

    // Write initiator.
    initiator.bitstr_write( bos, desc );
    }


void SystemMessage::bitstr_read
    (
    BitIStream & bis,
    const Serializer::Description & desc
    )
    {
    // Message type should be already extracted from
    // the stream and sender is not transferred over network.

    // Read initiator.
    initiator.bitstr_read( bis, desc );
    }


SystemMessage::Type SystemMessage::bitstr_read_type
    (
    BitIStream &                    bis,
    const Serializer::Description & desc
    )
    {
    SInt<Int32, TYPE_BITS_COUNT> t;

    t.bitstr_read( bis, desc );

    /** @todo FIXME - Safe serialization - throw exception if t is invalid. */
    hope( t >= 0 && t < TYPE_COUNT );

    return Type( t.read() );
    }


void SystemMessage::send
    (
    NodeId                      node,
    NetworkCommon::StreamType   stream_type
    ) const
    {
    NodeManager & node_manager = Global::node_manager();

    node_manager.send_message( *this, node, stream_type );
    }


void SystemMessage::multicast
    (
    const NodeIdList &          nodes,
    NetworkCommon::StreamType   stream_type
    )
    {
    NodeManager & node_manager = Global::node_manager();

    node_manager.multicast_message( *this, nodes, stream_type );
    }
                                                

void SystemMessage::multicast_to_server_nodes
    (
    NetworkCommon::StreamType   stream_type,
    bool                        donot_send_to_local_node_if_server
    )
    {
    NodeManager & node_manager = Global::node_manager();

    node_manager.multicast_message_to_server_nodes
        (
        *this,
        stream_type,
        donot_send_to_local_node_if_server
        );
    }


std::string SystemMessage::type_to_string
    (
    Type message_type
    )
    {
    hope( msg_table_inited );

    if( message_type >= 0 && message_type < (int) TYPE_COUNT )
        return msg_table[ message_type ].name;
        
    return "unknown_message_type";
    }

    
void SystemMessage::dispatch_system_message_from_packet
    (
    BitIStream &    bis,
    NodeId          msg_sender
    )
    {
    TAG_STACK_FRAME( "::Massiv::Core::SystemMessage::dispatch_system_message_from_packet()" );
    hope( msg_table_inited );

    /** @todo FIXME Safe serialization - jak handlovat chyby?
     * jake presne vyjimky se mohou objevit? Rozliseni
     * vyjimek zpusobenych serializaci, fatal vyjimky, ... */

    SystemEventQueue &   system_event_queue = Global::system_event_queue();
    Serializer::Description     desc;
    Type                        msg_type;
    Reference< SystemMessage >  system_msg;
    NodeId                      msg_receiver = Global::get_local_node_id();

    hope( !msg_sender.is_null() );

    // Read message type.
    msg_type = bitstr_read_type( bis, desc );
    
    // Verify message type.
    if( !verify_message_type( msg_type ) )
        {
        std::ostringstream oss;
        oss << "Received invalid message type " << msg_type
            << " from " << msg_sender << ".";
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }
    
    // Log.
        {
        std::ostringstream ss;
        ss  << "Received " << type_to_string( msg_type ) << " (" << msg_type
            << ") message from " << msg_sender << ".";
        /** @todo Another facility? */
        Global::log_debug( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_NORMAL, ss.str() );
        }

    // Create instance of the system message.
        {
        CreateSystemMessage * csm;
        csm = msg_table[ msg_type ].create;
        if( !csm )
            {
            std::ostringstream oss;
            oss << "Received invalid message type " << msg_type
                << " from " << msg_sender << ". The message type is reserved.";
            Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
            recover_from_invalid_message( msg_sender, msg_receiver );
            return;
            }
        system_msg = ( *csm )();
        hope( !system_msg.is_null() );
        hope( system_msg->get_type() == msg_type );
        }
    
    // Verify message sender node type.
    if( !verify_sender_node_type( msg_type, node_to_type_flags( msg_sender ) ) )
        {
        report_verification_failure
            (
            VT_SENDER_NODE_TYPE,
            msg_type,
            &msg_sender,
            NULL,
            &msg_receiver
            );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }

    // Verify message received node type.
    if( !verify_receiver_node_type( msg_type, node_to_type_flags( msg_receiver ) ) )
        {
        report_verification_failure
            (
            VT_RECEIVER_NODE_TYPE,
            msg_type,
            &msg_sender,
            NULL,
            &msg_receiver
            );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }

    // Set message sender.
    system_msg->set_sender( msg_sender );

    try 
        {
        // Read the message from the stream.
        system_msg->bitstr_read( bis, desc );
        }
    catch( std::exception & e )
        {
        std::ostringstream oss;
        oss << "Error reading content of " << type_to_string( msg_type )
            << " message received from " << msg_sender << ". " << e.what();
        Global::log_warning( Status::FACILITY_NODE_MANAGEMENT, Status::PRIORITY_HIGHEST, oss.str() );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }
    
    // Because anonymous nodes share same NodeId, the NodeId
    // inside system_msg->initiator is not valid NodeId.
    // We set it to msg_sender because msg_sender is unique
    // NodeId (unique for local node only) assigned by Network for
    // the anonymous node.
    if( system_msg->get_initiator().is_type( NodeId::TYPE_ANONYMOUS ) )
        {
        system_msg->set_initiator( msg_sender );
        }
    
    // Verify message initiator node type.
    if( !verify_initiator_node_type( msg_type, node_to_type_flags( system_msg->get_initiator() ) ) )
        {
        report_verification_failure
            (
            VT_INITIATOR_NODE_TYPE,
            msg_type,
            &msg_sender,
            &system_msg->get_initiator(),
            &msg_receiver
            );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }
        
    // Verify sender initiator reliationship.
    if( !verify_sender_initiator_relationship( msg_type, msg_sender, system_msg->get_initiator() ) )
        {
        report_verification_failure
            (
            VT_SENDER_RECEIVER_RELATIONSHIP,
            msg_type,
            &msg_sender,
            &system_msg->get_initiator(),
            &msg_receiver
            );
        recover_from_invalid_message( msg_sender, msg_receiver );
        return;
        }

    // todo: Otestovat zda dany node je pripojen do simulace.

    // Insert the message into the system event queue.
    system_event_queue.push_back_message( system_msg );

    // Boost.
    // This allows for quick responses if node receives messages.
    System::boost();
    }


unsigned int SystemMessage::get_message_pause_mask
    (
    Type    msg_type
    )
    {
    hope( msg_table_inited );
    hope( msg_type >= 0 && msg_type < (int) SystemMessage::TYPE_COUNT );
    
    return msg_table[ msg_type ].pause_mask;
    }


} // namespace Massiv::Core
} // namespace Massiv
