/** @file system_msg.h
 * System message common stuff.
 */

/*
 * begin        : Sat Jun 27 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_SYSTEM_MSG_H
#define MASSIV_SYSTEM_MSG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_BITIOSTREAM_FWD_H
#include "net/bitiostream_fwd.h"
#endif

#ifndef MASSIV_NET_COMMON_H
#include "net/net_common.h" // StreamType
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"  // SharedObject
#endif

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup sysevent
 * @{
 */

/**
 * System message base class.
 * Contains message type, message sender and message initiator.
 * Every system message has its type. This type is used
 * to identify system message when reading from a stream.
 * Due to this, system message type serialization is
 * not processed as usual.
 *
 * Message sender is the node that forwarded the message to local node.
 *
 * Message initiator is the node from which the message originates.
 * @note Initiator is automatically set to local node
 * when the message object is created.
 *
 * Every ancestor must redefine @a get_type() function.
 * To read message type from a stream, use static
 * method xxx_read_type() (where xxx is serialization prefix).
 *
 * To create new system message A:
 * <ol>
 *   <li> Add new message type in @c SystemMessage::Type.
 *        Update @c SystemMessage::TYPE_BITS_COUNT.
 *   <li> Don't forget to typedef Super, as it is typedefed in super
 *        classes (for case that somebody has forgot to define it as private).
 *   <li> Create constructor, copy constructor and operator=.
 *        Do not forget to properly create (update) super class in
 *        copy constructor and operator=.
 *   <li> Create virtual function get_type(). It must return type of A.
 *   <li> Create virtual function execute().
 *   <li> Create virtual functions bitstr_read() and bitstr_write().
 *        Do not forget to serialize super class.
 *   <li> Update SystemMessage::initialize_system_message_table().
 *        You must properly specify all attributes of @c SystemMessageTableItem
 *        for the message type.
 *   <li> Update system_event_fwd.h file.
 * </ol>
 *
 * @note SystemMessage is SharedObject so one doesn't need
 * to take care of message deletion in SystemEvent. This is
 * the only reason why system message is SharedObject.
 * On the other hand, system message objects can still be created
 * on the stack as "normal" objects.
 *
 * @warning Function initialize_system_message_table() must be called before
 * any other function from SystemMessage is called.
 */
class SystemMessage : public SharedObject
    {
    typedef SharedObject        Super;

public:
    
    /**
     * System message type.
     *
     * @note When adding new message type you can change order of existing
     *      types but you must keep all types inside 0 to TYPE_COUNT range.
     */
    enum Type
        {
        //
        // Simulation messages.
        //
        OBJECT_STATUS                   = 0,
            /**< ObjectStatusMessage. */
        FIND_OBJECT                     = 1,
            /**< FindObjectMessage. */
        MIGRATION_GROUP                 = 2,
            /**< MigrationGroupMessage. */
        MIGRATION_GROUP_NONARCHIVABLE   = 3,
            /**< MigrationGroupMessage (nonarchivable). */
        REPLICATION_UPDATE              = 4,
            /**< ReplicationUpdateMessage. */
        REPLICATION_ACK                 = 5,
            /**< ReplicationAckMessage. */
        ARCHIVATION_STATUS              = 6,
            /**< ArchivationStatusMessage. */
        FORWARD                         = 7,
            /**< ForwardMessage and ForwardedMessage. */
        TIME_SYNCHRONIZATION            = 8,
            /**< TimeSynchronizationMessage. */

        //
        // Startup synchronization messages.
        //
        STARTUP                 = 9,
            /**< StartupMessage. */
        RESERVED_10             = 10,
        RESERVED_11             = 11,
        RESERVED_12             = 12,
        RESERVED_13             = 13,
        
        //
        // Data messages.
        //
        DATA_VERSION_REQUEST    = 14,
            /**< DataVersionRequestMessage. */
        DATA_REQUEST            = 15,
            /**< DataRequestMessage. */
        DATA_VERSION            = 16,
            /**< DataVersionMessage. */
        DATA                    = 17,
            /**< DataMessage. */
        DATA_CONFIRMATION       = 18,
            /**< DataConfirmationMessage. */
        DATA_ERROR              = 19,
            /**< DataErrorMessage. */
        DATA_UPDATE_REQUEST     = 20,
            /**< DataUpdateReqeuestMessage. */

        //
        // Client connection messages.
        //
        ACCEPT_CLIENT           = 21,
            /**< AcceptClientMessage. */
        CLIENT_ACCEPTED         = 22,
            /**< ClientAcceptedMessage. */
        CLIENT_READY            = 23,
            /**< ClientReadyMessage. */
        CLIENT_READY_REPLY      = 24,
            /**< ClientReadyReplyMessage. */
        KILL_CLIENT             = 25,
            /**< KillClientMessage. */
        WELCOME                 = 26,
        
        //
        // Node management.
        //
        RESERVED_27             = 27,
        
        //
        // Global search.
        //
        GLOBAL_SEARCH           = 28,
            /**< GlobalSearchMessage. */
        GLOBAL_SEARCH_REPLY     = 29,
            /**< GlobalSearchReplyMessage. */

        TIME_MAGIC_CONSTANT     = 30,
            /**< TimeMagicConstantMessage. */

        //
        // Account management messages.
        //
        ACCOUNT_REGISTER_REQUEST    = 31,
            /**< AccountRegisterRequestMessage. */
        ACCOUNT_REGISTER_RESPOND    = 32,
            /**< AccountCreateRespondMessage. */
        ACCOUNT_CREATE_REQUEST      = 33,
            /**< AccountCreateRequestMessage. */
        ACCOUNT_CREATE_RESPOND      = 34,
            /**< AccountCreateRespondMessage. */
            
        //
        // Node action messages.
        //
        NODE_ACTION_REQUEST                 = 35,
            /**< NodeActionRequestMessage. */
        NODE_ACTION_REQUEST_REPLY           = 36,
            /**< NodeActionRequestReplyMessage. */
        NODE_ACTION_COORDINATOR_SELECTED    = 37,
            /**< NodeActionCoordinatorSelectedMessage. */
        NODE_ACTION_PREPARE_FINISH          = 38,
            /**< NodeActionPrepareFinishMessage. */
        NODE_ACTION_PREPARE_FINISH_REPLY    = 39,
            /**< NodeActionPrepareFinishReplyMessage. */
        NODE_ACTION_FINISH                  = 40,
            /**< NodeActionFinishMessage. */

        //
        // Balancer messages.
        //
        RESOURCE_LOAD_MESSAGE   = 41
            /**< ResourceLoadMessage. */
        
        }; // Type

    enum
        {
        TYPE_COUNT              = 42,
            /**< Number of system message types. */
        
        TYPE_BITS_COUNT         = 6
            /**< Number of bits required to store message type. 
             * Must be >= 1 and <= 31. */
        };

    /**
     * Value which can be returned by @c SystemMessage::execute().
     * @see @c SystemEventQueue.
     */
    enum ExecuteResult
        {
        EXECUTE_OK,
            /**< Execution done. The message should be removed from message queue. */
        EXECUTE_KEEP_MESSAGE_IN_QUEUE
            /**< Do not remove the message from the queue. The @c SystemMessage::execute()
             * must ensure that @c SystemEventQueue will not fall into neverending loop
             * (mostly by pausing/unpausing system message queue).
             *
             * This value can be returned @b ONLY and @b ONLY if the message sender
             * is server node.
             *
             * @see @c SystemMessageQueue.
             */
        }; // ExecuteResult


    /**
     * Pause flags. This flags are combined together to create pause mask.
     * Zero pause mask means that no message will be paused.
     * @note Messages are paused per (server) node.
     * @see SystemMessageQueue and SystemEventQueue.
     */
    enum PauseFlags
        {
        PAUSE_ARCHIVATION_FINISHING_START                   = 0x0001,
            /**< Flag used by archivation manager if archivation start request is received
             * while the archivation manager is still finishing current archivation.
             *
             * This flag is used if archivation start message is received
             * from node, which already send archivation finish message, but
             * not all of archivation finish messages was received from all server
             * nodes so new archivation can't be started.
             *
             * If this flag is set, all migration and archivation status messages
             * are paused.
             */
        PAUSE_WAITING_FOR_INITIAL_CONNECTIONS               = 0x0002,
            /**< Flag used by node manager if it is waiting until all connections to server
             * nodes are established.
             *
             * This flag is used if the node manager receives node ready message from a server
             * node but not all connections to other nodes was established.
             * All messages from the node are paused until all the connections are established.
             */
        PAUSE_ARCHIVATION_WAITING_FOR_SYNC_RPC_FINISHED     = 0x0004,
            /**< Flag used by archivation manager if archivation start request is received
             * while local system has some pending synchronous RPC. In that case
             * new archivation can't be started and the start archivation request
             * must be paused. Also all (archivable) migration group messages are
             * paused from the node which sent the start archivation request.
             *
             * If this flag is set, all migration and archivation status messages
             * are paused. Nonarchivable migration messages are not paused.
             */
        PAUSE_NODE_ACTION_REQUEST                           = 0x0008,
            /**< Flag used by node action manager to pause all node action request
             * messages.
             *
             * This flag is used when finishing a node action and local node
             * has not yet finished the action but another node requested
             * the action again (because the node has already finished the action).
             * In that case local node pauses all action request messages
             * until it finishes the pending action.
             */
        PAUSE_WAITING_FOR_SIMULATION_READY                  = 0x0010
            /**< Flag used by node manager if it is waiting until simulation is
             * is ready.
             * In other worlds: Messages that has this flag set, won't be
             * delivered until simulation state is loaded from an archive file
             * and the simulation is unpaused.
             */
        
        }; // PauseFlags

    /**
     * Node id list.
     * Used by when multicasting messages.
     */
    typedef std::vector< NodeId > NodeIdList;

protected:

    /**
     * Constructor.
     * Sender is set to local node id.
     * Initiator is set to local node id.
     */
    SystemMessage();

    /**
     * Destructor.
     */
    virtual ~SystemMessage();

    /**
     * Copy constructor.
     */
    SystemMessage
        (
        const SystemMessage & msg
        );

    /**
     * Operator=.
     */
    SystemMessage & operator=
        (
        const SystemMessage & msg
        );
            
public:

    /**
     * Get system message type.
     * Every ancestor must redefine this function.
     */
    virtual Type get_type() const = 0;

    /**
     * Execute the system message.
     * The sender attribute must be set prior calling this function.
     *
     * @note This function is called by @c SystemEventQueue.
     */
     virtual ExecuteResult execute() = 0;

     /**
      * Convert type to string.
      */
     static std::string type_to_string
        (
        Type message_type
        );

public:

    /**
     * Serialize to bitstream.
     * @note Must be virtual.
     */
    virtual void bitstr_write
        (
        BitOStream &                    bos,
        const Serializer::Description & desc
        ) const;

    /**
     * Read from bitstream.
     * @note Both message type and message sender
     * are not extracted from the stream.
     * Both type and sender are left unchanged so they
     * can be set before serialization.
     * @note Must be virtual.
     */
    virtual void bitstr_read
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

    /**
     * Read message type from input stream.
     */
    static Type bitstr_read_type
        (
        BitIStream &                    bis,
        const Serializer::Description & desc
        );

public:

    /**
     * Send the system message to specified node.
     */
    void send
        (
        NodeId                    node,
        NetworkCommon::StreamType stream_type
        ) const;
    
    /**
     * Send @a message to @a nodes.
     */
    void multicast
        (
        const NodeIdList &        nodes,
        NetworkCommon::StreamType stream_type
        );
                                                    
    /**
     * Send @a message to all server nodes.
     * @param donot_send_to_local_node_if_server If the local system
     *      is server node and this argument is true, the message
     *      is not send to local node.
     */
    void multicast_to_server_nodes
        (
        NetworkCommon::StreamType stream_type,
        bool                      donot_send_to_local_node_if_server = false
        );
                                                                                             
    /**
     * Read system message from specified stream and insert
     * it as a system event into global system event queue.
     * This function is used by Network to process received
     * network packets.
     * @param bis Stream containing serialized system message.
     * @param sender Node from which the stream was received.
     */
     static void dispatch_system_message_from_packet
        (
        BitIStream &        bis,
        NodeId              sender
        );

    /**
     * Get message pause mask by its type.
     * @see @c SystemEventQueue and @c SystemMessageQueue for more inforamtion.
     */
    static unsigned int get_message_pause_mask
        (
        Type    message_type
        );

public:

    /**
     * Get message sender.
     */
    const NodeId & get_sender() const
        {
        return sender;
        }

    /**
     * Set message sender.
     */
    void set_sender
        (
        const NodeId & _sender
        )
        {
        sender = _sender;
        }

    /**
     * Get initiator node.
     */
    const NodeId & get_initiator() const
        {
        return initiator;
        }

    /**
     * Set the initiator.
     * @note Initiator is automatically set to local node
     * when the message object is created.
     */
    void set_initiator
        (
        const NodeId & _initiator
        )
        {
        initiator = _initiator;
        }
        
public:

    /**
     * Initialize tables used for miscenalleous purposes like
     * message verification etc.
     * This function must be called before any other function
     * from SystemMessage is called.
     */
    static void initialize_system_message_table();

private:

    NodeId          sender;
        /**< Specifies node from which the message was received.
         * Must be set to local node id if the message
         * is constructed locally (before sending). Constructor
         * autmoatically does it.
         * This attribute is not transferred over network!
         * So it is not affected by serialization.
         *
         * @note If the message is constructed locally, setting
         * the sender to local node id is not only required
         * by localization protocol but also by archivation protocol.
         * Setting node id to null would cause random archivation crashes.
         * So do not change this behavior in the future.
         */

    NodeId          initiator;
        /**< Specifies node from which the message originates. */

    }; // SystemMessage

/** @} */ // addtogroup sysevent

    
} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SYSTEM_MSG_H
