/** @file system.cpp
 * Massiv local system api.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "system/system.h"
#include "pointer/garbage_collector.h"
#include "node/node_manager.h"
#include "database/well_known_object_id_database.h"
#include "database/archive_database.h"
#include "object/archive_manager.h"
#include "object/object_manager.h"
#include "object/object_provider_state.h"
#include "utility/debug.h"
#include "thread/stack_tracer.h"
#include "filesys/compact_volume.h"

namespace Massiv {

using namespace Massiv::Core;

System::Phase System::g_system_phase = System::PHASE_UNINITIALIZED;


void System::download_data()
    {
    hope( g_system_phase == PHASE_INITIALIZED );

    // Switch phase.
    g_system_phase = PHASE_DOWNLOADING_DATA;
    
    try
        {
        // Download data.
        download_data_loop();
        
        // Switch phase.
        g_system_phase = PHASE_DATA_DOWNLOADED;
        }
    catch( Exception & )
        {
        // Undo work.
        g_system_phase = PHASE_INITIALIZED;

        throw;
        }
    }


void System::force_gc()
    {
    GarbageCollector & gc = Global::garbage_collector();

    if( gc.get_settings().can_run_any_time )
        {
        if( gc.is_running() )
            {
            return;
            }

        gc.force_run();

        while( !gc.is_running() )
            {
            gc.next_tick();
            }

        while( gc.is_running() )
            {
            gc.next_tick();
            }
        }
    else
        {
        gc.force_run();
        }
    }


void System::collect_object
    (
    const Core::CommonPointerBase & pointer
    )
    {
    Global::garbage_collector().collect_object( pointer );
    }


void System::dispose_gc_root
    (
    const Core::CommonPointerBase & pointer
    )
    {
    Global::garbage_collector().dispose_gc_root( pointer );
    }


std::string System::print_stack_trace()
    {
    if( Global::g_print_stack_traces )
        {
        if( Global::stack_tracer_available() )
            {
            std::ostringstream ss;
            Global::stack_tracer().dump_stack( ss );
            return ss.str();
            }
        else
            {
            return StackFrameEntry::print_registered_frames();
            }
        }
    else
        {
        return "<not_available>";
        }
    }


const ObjectId & System::get_naming_service_object_id()
    {
    static ObjectId id;
    
    id = Global::well_known_object_id_database().get_naming_service_object();
    
    return id;
    }


void System::create_initial_object_provider
    (
    UInt32  id
    )
    {
    ObjectManager &     object_manager = Global::object_manager();
    UInt32              provider_id_num = id;
    ProviderId          provider_id = ProviderId::create_from_number( provider_id_num );
    UInt64              next_object_id( 1 );
    
    hope( !provider_id.is_virtual_provider_id() );
    hope( !provider_id.is_dummy_provider_id() );
    ObjectProviderState provider_state( provider_id, next_object_id );
    
    object_manager.register_local_provider( provider_state );
    }


void System::create_initial_archive()
    {
    ArchiveManager & archive_manager = Global::archive_manager();

    hope( g_system_phase == PHASE_INITIALIZED );

    if( archive_manager.archivation_in_progress() )
        {
        panic( "Some archivation in progress while trying to create initial archive!" );
        }

    // Start initial archivation.        
    archive_manager.request_start_initial_archivation();

    while( archive_manager.archivation_in_progress() )
        {
        next_tick();
        }
    }


void System::unpack_archive
    (
    const std::string & serialized_archivation_id,
    const std::string & directory
    )
    {
    hope( g_system_phase == PHASE_INITIALIZED );

    ArchiveDatabase &   archive_database = Global::archive_database();
    ArchivationId       archivation_id;
    
    // Extract archivation id.

    archivation_id = ArchivationId::from_string( serialized_archivation_id );
    
    if( archivation_id.is_invalid() )
        {
        panic( "Invalid archivation id specified." );
        }

    // Unpack.

    CompactVolume volume
        (
        archive_database.archivation_id_to_regular_path_file_name( archivation_id ),
        std::ios_base::in
        );

    volume.unpack( directory, ".", "*" );
    }


void System::pack_archives
    (
    const std::vector< std::string > & directories,
    const std::string & serialized_archivation_id
    )
    {
    hope( g_system_phase == PHASE_INITIALIZED );
    hope( !is_enabled( CAPABILITY_SIMULATION ) );

    ObjectManager &     object_manager = Global::object_manager();
    ArchiveManager &    archive_manager = Global::archive_manager();
    ArchivationId       archivation_id;
    
    // Extract archivation id.

    archivation_id = ArchivationId::from_string( serialized_archivation_id );
    
    if( archivation_id.is_invalid() )
        {
        panic( "Invalid archivation id specified." );
        }
    
    if( directories.size() <= 0 )
        {
        panic( "At least one unpacked archive must be specified." );
        }
    
    // Load unpacked archive to memory.

    object_manager.load_unpacked_archive( directories );
    
    if( archive_manager.archivation_in_progress() )
        {
        panic( "Some archivation in progress while trying to pack archive!" );
        }
    
    // Archive all objects.

    if( !archive_manager.request_start_archivation( archivation_id ) )
        {
        panic( "Start archivation failed!" );
        }

    while( archive_manager.archivation_in_progress() )
        {
        next_tick();
        }
    }


void System::enable
    (
    Capability cap
    )    
    {
    ObjectManager &     object_manager = Global::object_manager();
    ArchiveManager &    archive_manager = Global::archive_manager();
    Balancer &          balancer = Global::balancer();

    switch( cap )
        {
        case CAPABILITY_SIMULATION:
            object_manager.enable_simulation();
            break;
            
        case CAPABILITY_AUTO_ARCHIVATION:
            archive_manager.enable_auto_archivation();
            break;

        case CAPABILITY_LOAD_BALANCING:
            balancer.enable_load_balancing();
            break;
        }
    }
    
    
void System::disable
    (
    Capability cap
    )    
    {
    ObjectManager &     object_manager = Global::object_manager();
    ArchiveManager &    archive_manager = Global::archive_manager();
    Balancer &          balancer = Global::balancer();

    switch( cap )
        {
        case CAPABILITY_SIMULATION:
            object_manager.disable_simulation();
            break;
            
        case CAPABILITY_AUTO_ARCHIVATION:
            archive_manager.disable_auto_archivation();
            break;
        
        case CAPABILITY_LOAD_BALANCING:
            balancer.disable_load_balancing();
            break;
        }
    }


bool System::is_enabled    
    (
    Capability cap
    )    
    {
    ObjectManager &     object_manager = Global::object_manager();
    ArchiveManager &    archive_manager = Global::archive_manager();
    Balancer &          balancer = Global::balancer();

    switch( cap )
        {
        case CAPABILITY_SIMULATION:
            return object_manager.is_simulation_enabled();
            
        case CAPABILITY_AUTO_ARCHIVATION:
            return archive_manager.is_auto_archivation_enabled();

        case CAPABILITY_LOAD_BALANCING:
            return balancer.is_load_balancing_enabled();
        }
    
    return false;
    }


} // namespace Massiv
