/** @file system.h
 * Massiv local system api.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SYSTEM_H
#define MASSIV_SYSTEM_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_POINTER_FWD_H
#include "pointer/pointer_fwd.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#include <vector>
#include <iosfwd>


namespace Massiv {

namespace Core {

class ObjectId;
class STime;
class SystemTime;
class ResultObject;
class CommonPointerBase;
class VariantPointer;

namespace Lib {

class NodeObjectInterface;

} // namespace Core::Lib

} // namespace Core


/**
 * Application interface to local massiv system.
 *
 * @ingroup global
 */
class MASSIV_EXPORTED System
    {
public:
        
    NO_DEFAULT_CONSTRUCTOR( System );
    NO_COPY_CONSTRUCTOR( System );
    NO_ASSIGNMENT_OPERATOR( System );

public:

    /**
     * Internal phase of the system.
     */
    enum Phase
        {
        PHASE_UNINITIALIZED,
            /**< System is not intialized. */
        PHASE_INITIALIZED,
            /**< The system is initialized. */
        PHASE_DOWNLOADING_DATA,
            /**< The system is downloding data. */
        PHASE_DATA_DOWNLOADED,
            /**< Minimum data has been downloaded. */
        PHASE_CONNECTING,
            /**< The system is connecting to the simulation. */
        PHASE_CONNECTED,
            /**< The system is connected to simulation. */
        PHASE_RUNNING,
            /**< Main loop is running. */
        PHASE_DISCONNECTING,
            /**< The system is diconnecting from the simulation. */
        PHASE_SHUTTING_DOWN
            /**< The system is shutting down. */
        }; // Phase
    
    /**
     * Class registration callback.
     */
    typedef void RegisterClasses();
    
    /**
     * Create node object callback.
     *
     * Create node object representing local system.
     * The main purpose of this function is to create
     * node object for local system. However it can also
     * create other objects as well.
     *
     * @todo FIXME CZ: Nepresne! Upravit kdy se to presne vola.
     * Muze se odsimulovat nejaka zprava driv nez se to zavola?
     * Tzn. jestli se muze zavolat ObjectManager::next_tick().
     *
     * If the local system is server, this function is called after
     * simulation has been restored from archive.
     *
     * If the local system is client, this function is called after
     * the system is properly connected to the simulation, just
     * before returning from System::connect().
     * 
     * @warning Do not call any fuction from @c System.
     */
    typedef void CreateNodeObject();
    
    /**
     * Next tick callback.
     *
     * Run next tick on the node.
     * This callback is periodically called from massiv core.
     */
    typedef void NextTick();

    /**
     * System start up info.
     * Describes what subsystems should be started.
     */
    struct MASSIV_EXPORTED StartUpInfo
        {
    public:

        /**
         * Default constructor.
         * Sets @c node_type to @c NodeId::TYPE_ANONYMOUS, all other
         * fields are cleared.
         */
        StartUpInfo();

        /**
         * Initializes @c StartUpInfo according to a "predefined profile"
         * for given @c node_type.
         * The other parameters are simply copied to the corresponding
         * attributes in the structure.
         */
        StartUpInfo
            (
            Core::NodeId::Type  node_type,
            const std::string & conf_file,
            RegisterClasses *   register_classes = NULL,
            CreateNodeObject *  create_node_object = NULL,
            NextTick  *         next_tick = NULL
            );

    public:
            
        Core::NodeId::Type  node_type;
            /**< Node type. */
        bool                standalone_node;
            /**< Standalone node?
                 If set to true all node-related configuration will be
                 ignored. Node will be started "stand alone". */
        bool                base_services;
            /**< Start base services (registry & logger)? */
        bool                load_config;
            /**< Load config? */
        bool                threading;
            /**< Start threading services? */
        bool                file_systems;
            /**< Start file systems srevices? */
        bool                networking;
            /**< Start networking? */
        bool                helper_subsystems;
            /**< Start simulation-helper services? */
        bool                data_management;
            /**< Start data-management services? */
        bool                object_management;
            /**< Start object-management services? */
        bool                system_context;
            /**< Initialize system context? */
        bool                quiet_start;
            /**< Disable logger during the system initialization. This is
                 a hack done because of some extern utilities - to make
                 them run not so verbatim. */
        bool                save_registry_on_exit;
            /**< Should registry content be saved on system shutdown? */
        std::string         massiv_conf_file_name;
            /**< Path to config file massiv.conf. If empty, MASSIV_CONFPATH
                 environment variable will be opened to localize configuration
                 file. Ignored if @c load_config is false. */
        bool                single_threaded_srpc;
            /**< Single threaded SRPC (SRPC threading model setup)? */
        RegisterClasses *   register_classes_callback;
            /**< The callback that will be automatically registered when
                 the structure is passed to @c System::initialize(). */
        CreateNodeObject *  create_node_object_callback;
            /**< The callback that will be automatically registered when
                 the structure is passed to @c System::initialize(). */
        NextTick *          next_tick_callback;
            /**< The callback that will be automatically registered when
                 the structure is passed to @c System::initialize(). */
        }; // StartUpInfo
    
    /**
     * Massiv core capabilities.
     *
     * Not every node type supports each capability.
     * If a capability is not supported by a node type, enable()
     * and disable() have no effect and is_enabled() returns false
     * for the capability.
     *
     * Main purpose of the capabilities is to disable various subsystems
     * until simulation is properly started or local node is properly
     * connected to the simulation.
     *
     * @note Even if a capability is enabled via enable(), its functionality
     *  may be disabled via settings.
     *  For example if AUTO_ARCHIVATION is disabled, archive manager
     *  will not automatically initiate new archivation even if auto archivation
     *  is permitted via settings. On the other hand if AUTO_ARCHIVATION
     *  is enabled but auto archivation is not permitted via settings, the
     *  archive manager will not automatically initiate new archivation.
     */
    enum Capability
        {
        CAPABILITY_SIMULATION,
            /**< Simulation.
             * If simulation is not enabled, object manager won't execute
             * any event even if simulation time is not paused.
             * Managed by ObjectManager.
             *
             * Supported nodes: Server, client.
             */
        CAPABILITY_AUTO_ARCHIVATION,
            /**< Automatic archivation.
             * If auto archivation is enabled, local server node will
             * automatically initiate distributed archivation once a time.
             * Managed by ArchiveManager.
             *
             * Supported nodes: Server.
             */
        CAPABILITY_LOAD_BALANCING
            /**< Automatic load balancing.
             * If enabled Balancer will run.
             *
             * Supported nodes: Server, client.
             */
             
        // todo:
        // Automatic data download from data server.
        
        }; // Capability

public:

    //
    // System callbacks.
    //

    /**
     * Register class registration callback.
     * This must be done before the system initialization.
     * @note System::initialize() will do the registration for you if the
     * callback was filled to StartUpInfo structure.
     */
    static void register_register_classes
        (
        RegisterClasses * callback
        );
    
    /**
     * Register create node object callback.
     * This must be done before the system initialization.
     * @note System::initialize() will do the registration for you if the
     * callback was filled to StartUpInfo structure.
     */
    static void register_create_node_object
        (
        CreateNodeObject * callback
        );
    
    /**
     * Register next tick callback.
     * This must be done before the system initialization.
     * @note System::initialize() will do the registration for you if the
     * callback was filled to StartUpInfo structure.
     */
    static void register_next_tick
        (
        NextTick * callback
        );
        
public:

    //
    // Main system API.
    //
    
    /**
     * Get current system phase.
     */
    static Phase get_phase();

    /**
     * Initialize the system.
     * @c info describes what subsystems should be started (or restared
     * if they are already running).
     *
     * @post If system phase was @c PHASE_UNINITIALIZED the phase will be
     * changed to @c PHASE_INITIALIZED.
     */
    static void initialize
        (
        const StartUpInfo & info
        );
        
    /**
     * Download minimum data required to start/connect to simulation.
     *
     * @pre System phase is @c PHASE_INITIALIZED.
     * @post System phase is @c PHASE_DATA_DOWNLOADED.
     */
    static void download_data();
    
    /**
     * Connect to simulation using given node credentials (identity) stored
     * in @c my_config.
     *
     * Initializes core with node data stored in @a my_config.
     * Until @c System::connect() is called, node operates in an anonymous
     * mode and has assigned a special anonymous #NodeId.
     * While in anonymous mode, only connections to service nodes can
     * be estabilished.
     * This allows to download minimum data prior connecting to the
     * simulation.
     *
     * Server nodes automatically load simulation state from local archives.
     * Initial archive must be created in a special way, when server node
     * is not connected to simulation. See @c create_initial_archive().
     *
     * @param @c my_config Input stream containing local node info.
     *      Content of the info has the same format like registry config file.
     *      It must contain: <br>
     *      <ul>
     *      <li> String variable @c node_id containing my #NodeId (identity).
     *      <li> String variable @c rsa_private_key containing rsa private key
     *           of my node.
     *      <li> String variable @c rsa_public_key containing rsa public key
     *           of my node.
     *      </ul>
     *      Other variables are ignored.
     *
     * @throw ExSystemNodeConfigError Invalid @c my_config.
     *
     * @pre System phase is @c PHASE_DATA_DOWNLOADED.
     * @post System phase is @c PHASE_CONNECTED.
     */
    static void connect
        (
        std::istream &  my_config
        );
    
    /**
     * Run simulation or the main system loop.
     *
     * @pre System phase is @c PHASE_CONNECTED.
     * @post System phase is @c PHASE_CONNECTED.
     */
    static void run();
    
    /**
     * Disconnect from the simulation.
     * Node is switched back to anonymous mode.
     *
     * @pre System phase is @c PHASE_CONNECTED.
     * @post System phase is @c PHASE_DATA_DOWNLOADED.
     */
    static void disconnect();

    /**
     * Shutdown the system.
     *
     * @post System phase @c PHASE_UNINITIALIZED.
     */
    static void shutdown();

    /**
     * Check whether the system is successfully connected to the simulation.
     * @return True if system phase is @c PHASE_CONNECTED or @c PHASE_RUNNING.
     */
    static bool is_connected();

    /**
     * Request quit from current system loop ie.
     * from @c download_data(), @c connect(), @c run() or @c disconnect().
     *
     * @note The request quit flag is cleared at the beginning
     * of connect(), run() and disconnect().
     */
    static void request_quit();

public:

    //
    // Running.
    //

    /**
     * "Locks" current Massiv tick.
     * Until the tick is "unlocked" current tick can not finish.
     * Any call to @c System::next_tick() will fail until the
     * tick is unlocked.
     * Tick locking can be used to prevent client from issuing a synchronous
     * RPC call (ie. reentering the core and starting next tick).
     */
    static void lock_tick();

    /**
     * "Unlock" current Massiv tick.
     */
    static void unlock_tick();

    /**
     * Tests if current tick is "locked".
     */
    static bool is_tick_locked();

    /**
     * Wait for rpc call result.
     * Blocks until RPC call result is returned to @a result_object.
     * Doesn't block deliveries of migration messages.
     * This applies also to the blocked objects (= objects that wait for
     * a RPC reply).
     * Throws RemoteCallFailedException if RPC can not be delivered or system
     * rejects the call.
     * @pre result_object is valid pointer to #ResultObject.
     * @warning Should be called only by #ResultObject or it could always
     * timeout in certain cases.
     */
    static void wait_for_rpc_reply
        (
        const Core::Pointer<Core::ResultObject> & result_object
        );
    
    /**
     * Check if the system is waiting for at least one rpc call result.
     */
    static bool is_waiting_for_rpc_reply();
    
    /**
     * Check if client code is running.
     */
    static bool inside_simulation();
    
    /**
     * Tells scheduler that it should not sleep at the end of the current
     * tick. Sets sleep time to zero.
     */
    static void boost();
    
    /**
     * Tests if there is enough stack space to upcall client code.
     * The stack limit is set via registry (Settings/System/SRPC).
     */
    static bool have_enough_stack_space();

public:

    //
    // Time.
    //

    /**
     * Get system time.
     */
    static const Core::SystemTime & system_time();

    /**
     * Get simulation time.
     * @warning Returns same value until next_tick() is called.
     */
    static const Core::STime & time();

    /**
     * Get total sleep time.
     */
    static const Core::SystemTime & total_sleep_time();
    
public:

    //
    // Miscellanneous.
    //

    /**
     * Forces GC to start next GC run.
     * Waits for GC to complete the run (if possible).
     */
    static void force_gc();

    /**
     * Forces GC to collect object referenced by @c pointer.
     * If the object can not be deleted immediatelly, it will be deleted at
     * the next GC tick.
     */
    static void collect_object
        (
        const Core::CommonPointerBase & pointer
        );
    
    /**
     * Tells GC that the permanent GC root object referenced @c by pointer
     * should be disposed. The object will be demoted to a non root object and
     * deleted at a future GC run together with its migration group if possible.
     */
    static void dispose_gc_root
        (
        const Core::CommonPointerBase & pointer
        );
    
    /**
     * Prints call stack of the current thread.
     * Either uses native StackTracer (if supported by the architecture) or
     * dumps list of manually registered stack frame entries (#StackFrameEntry).
     */
    static std::string print_stack_trace();

public:
    
    //
    // Well known object ids.
    //
    
    /**
     * Get #ObjectId of naming service object.
     * Returns null object id if naming service object id is not
     * registered in the well known object id database.
     *
     * Core holds only object id of the naming service object
     * and doesn't care of interface of the object.
     * Interface and implementation of naming service object is
     * completely on Lib side.
     *
     * Example:
     * @code
     * WeakPointer< Object > object = System::get_naming_service_object();
     * if( object != null )
     *     {
     *     Remote< NamingService > service = object.convert();
     *     // Do whatever you want...NamingService is interface
     *     // implemented in Lib.
     *     }
     * @endcode
     */
    static const Core::ObjectId & get_naming_service_object_id();

public:

    //
    // Local node information.
    //

    /**
     * Get node id of the local system.
     * @note Returns anonymous node id if node is not connected to simulation.
     */
    static const Core::NodeId & get_local_node_id();

    /**
     * Check if node id of the local system is server node id.
     * @note Returns false if node is not connected to simulation.
     */
    static bool is_server();

    /**
     * Check if node id of the local system is client node id.
     * @note Returns false if node is not connected to simulation.
     */
    static bool is_client();
    
public:

    //
    // Archive management.
    //

    /**
     * Create new object provider.
     * Object providers are used to generate objects ids assigned to objects.
     * The caller should ensure that the id is a valid provider id
     * (see ProviderId) and it is unique in the system.
     * By default, if simulation is started from archive on a
     * @c NodeId::TYPE_SERVER node, object providers are automatically
     * restored from the archive.
     * On @c NodeId::TYPE_CLIENT nodes, local "dummy" providers are
     * automatically created on @c connect().
     * The only case when program needs to call this is when server is not
     * connected to simulation and its task is to create an initial archive.
     */
    static void create_initial_object_provider
        (
        Core::UInt32  id
        );

    /**
     * Create initial archive.
     * To be called from server nodes that did not connect to simulation.
     *
     * Initial archive is typically created in the following way:
     *  1) System::initialize( NodeId::TYPE_SERVER )
     *  [ Starts server in an anonymous mode. ]
     *  2) System::create_initial_object_provider()
     *  [ Creates object provider so that we will be able to create objects. ]
     *  3) Create "initial" objects (typically some sort of global objects)
     *  [ Events (RPC calls, for example) should be scheduled to that objects
     *    so that something will actually happen when the server is started
     *    from this archive. This way, the initial archive logic can be
     *    implemented as a regular server lib code dispatched at server
     *    restart. ]
     *  4) System::create_initial_archive()
     *  5) System::shutdown()
     *
     * @note    Simulation is not running while in anonymous mode, thus
     *          all object migrations, RPC calls, etc. will be dispatched
     *          once the server is restarted from the (initial) archive.
     * @pre     System phase is @c PHASE_INITIALIZED.
     */
    static void create_initial_archive();
    
    /**
     * Export archive streams stored in a specified archive file denoted by
     * by its @c serialized_archivation_id to a @c directory residing on the
     * external file system.
     */
    static void unpack_archive
        (
        const std::string & serialized_archivation_id,
        const std::string & directory
        );

    /**
     * Merge exported archive streams to one archive volume.
     * Create an archive volume from one or more unpacked archives.
     *
     * @note Archives can be unpacked via filesys utilities or VolumeManager
     *       like regular filesys volumes. Since archive volumes contain
     *       additional "metadata", unpacked archives can not be simply
     *       packed via filesys because the metadata will be missing/invalid.
     *       @c pack_archives() ensures proper reconstrution of the
     *       metadata even if the archive streams are modified
     *       Can last quite a long time because all objects must be loaded
     *       into memory (the metadata will be reconstructed directly
     *       from object states).
     *
     * @param directories Array of directory names. Each directory
     *      holds one unpacked archive.
     * @param serialized_archivation_id Textual ArchivationId specifies
     *      the archive file to which the archive fill will be packed.
     *      ArchivationId has format "massiv_X.archive", where
     *      X is a number.
     *
     * @pre @a directories is not empty.
     */
    static void pack_archives
        (
        const std::vector< std::string > &  directories,
        const std::string &                 serialized_archivation_id
        );
    
public:

    //
    // Capabilities.
    //

    /**
     * Enable core capability.
     * Nothing is done if the capability is already enabled
     * or the capability is not supported by local node.
     */
    static void enable
        (
        Capability capability
        );
        
    /**
     * Disable core capability.
     * Nothing is done if the capability is already disabled
     * or the capability is not supported by local node.
     */
    static void disable
        (
        Capability capability
        );
    
    /**
     * Check whether an capability is enabled.
     * Returns false if the capability is not supported by local node.
     */
    static bool is_enabled
        (
        Capability capability
        );

private:
    
    //
    // Internal methods.
    //

    /**
     * Fires callbacks registered by @c register_classes().
     */
    static void fire_register_classes_callbacks();

    /**
     * Fires callbacks registered by @c register_create_node_object().
     */
    static void fire_create_node_object_callbacks();

    /**
     * Fires callbacks registered by @c register_next_tick().
     */
    static void fire_next_tick_callbacks();

    /**
     * Connect loop.
     */
    static void connect_loop();
    
    /**
     * Disconnect loop.
     */
    static void disconnect_loop();

    /**
     * Download data loop.
     */
    static void download_data_loop();
    
    /**
     * Simulation run loop thread entry point.
     * @c arg refers to corresponding thread context object.
     */
    static int run_loop
        (
        Core::VariantPointer arg
        );
    
public:
        
    /**
     * Generic loop.
     * This loop only waits until quit is requested.
     * Until then it calls @c next_tick().
     * @pre System is not connected to simulation. @c run() must be used
     *      instead.
     */
    static void generic_loop();
    
    /**
     * Was quit requested?
     */
    static bool quit_requested();

private:
    
    /**
     * Runs next Massiv tick.
     */
    static void next_tick();
    
    /**
     * Starts tick.
     */
    static void do_start_tick();
    
    /**
     * Does simulation tick.
     */
    static void do_simulation_tick();

    /**
     * Finishes tick.
     */
    static void do_finish_tick();

    /**
     * Handles SRPC replies.
     * Used by thread-based waiting scheme.
     */
    static void handle_srpc_replies();

    /**
     * Clears request quit flag.
     */
    static void clear_request_quit_flag();

private:

    static Phase g_system_phase;
        /**< System phase. */

    }; // System


} // namespace Massiv


#endif // MASSIV_SYSTEM_H
