/** @file system_callbacks.cpp
 * System callbacks.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "system/system.h"
#include "global/global.h"
#include "object/class_manager.h"
#include "object/object_property.h"
#include "object/lib/node_object_interface.h"
#include "pointer/object_pointer.h"
#include <vector>
#include <sstream>

namespace Massiv {

using namespace Massiv::Core;


namespace {


/**
 * List of class registration callbacks.
 */
typedef std::vector< System::RegisterClasses * > ClassRegistrationCallbacks;

/**
 * List of create node callbacks.
 */
typedef std::vector< System::CreateNodeObject * > CreateNodeObjectCallbacks;

/**
 * List of next tick callbacks.
 */
typedef std::vector< System::NextTick * > NextTickCallbacks;


ClassRegistrationCallbacks  g_class_registration_callbacks;
    /**< Class registration callbacks. */

CreateNodeObjectCallbacks   g_create_node_object_callbacks;
    /**< Create node object callbacks. */

NextTickCallbacks           g_next_tick_callbacks;
    /**< Next tick callbacks. */


template< typename Callbacks >
void fire_callbacks
    (
    const Callbacks & callbacks
    )
    {
    typename Callbacks::const_iterator it;

    for( it = callbacks.begin(); it != callbacks.end(); ++it )
        {
        typename Callbacks::value_type callback = *it;
        callback();
        }
    }    

template< typename Callbacks >
bool find_callback
    (
    const Callbacks &               callbacks,
    typename Callbacks::value_type  callback
    )
    {
    typename Callbacks::const_iterator it;

    for( it = callbacks.begin(); it != callbacks.end(); ++it )
        {
        if( *it == callback )
            {
            return true;
            }
        }
    
    return false;
    }    


} // anonymous namespace


void System::register_register_classes
    (
    RegisterClasses * callback
    )
    {
    if( !find_callback( g_class_registration_callbacks, callback ) )
        {
        g_class_registration_callbacks.push_back( callback );
        }
    }
    

void System::register_create_node_object
    (
    CreateNodeObject * callback
    )
    {
    if( !find_callback( g_create_node_object_callbacks, callback ) )
        {
        g_create_node_object_callbacks.push_back( callback );
        }
    }

    
void System::register_next_tick
    (
    NextTick * callback
    )
    {
    if( !find_callback( g_next_tick_callbacks, callback ) )
        {
        g_next_tick_callbacks.push_back( callback );
        }
    }


void System::fire_register_classes_callbacks()
    {
    fire_callbacks( g_class_registration_callbacks );
    }


void System::fire_create_node_object_callbacks()
    {
    fire_callbacks( g_create_node_object_callbacks );

    #ifdef OPTION_DEBUG

    Pointer<Lib::NodeObjectInterface> node_object = get_local_node_id();

    if( !node_object.is_local() )
        {
        panic( "create_node_object_callback() did not create a node object for "
            "the local node." );
        }

    #endif
    }


void System::fire_next_tick_callbacks()
    {
    fire_callbacks( g_next_tick_callbacks );
    }


} // namespace Massiv
