/** @file system_connect.cpp
 * System connect/disconnect API.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "system/system.h"
#include "object/object_manager.h"
#include "node/node_manager.h"
#include "database/node_database.h"
#include "global/global_callbacks.h"
#include "global/global.h"
#include "object/global_search_manager.h"
#include "node/node_action_manager.h"
#include "utility/debug.h"
#include <sstream>

namespace Massiv {

using namespace Massiv::Core;

namespace {


/**
 * Local node settings.
 * Must be kept in sync with structure used by NodeDatabase.
 */
class LocalNodeSettings : public Settings
    {
public:

    /** Constructor. */
    LocalNodeSettings() :
        Settings( "NodeDatabase/local_node" ),
        node_id( access_string( "node_id" ) ),
        rsa_public_key( access_string( "rsa_public_key" ) ),
        rsa_private_key( access_string( "rsa_private_key" ) )
        {
        }

public:
        
    String &    node_id;
        /**< Node id. */
    String &    rsa_public_key;
        /**< Public key. */
    String &    rsa_private_key;
        /**< Private key. */
    };
    

/**
 * Reads local node settings from @c my_config and imports it to NodeDatabase.
 * The config is expected to have been saved in a "registry format".
 */
void install_local_node_config
    (
    std::istream & my_config
    )
    {
    Registry &          registry = Global::registry();
    NodeDatabase &      node_database = Global::node_database();
    LocalNodeSettings   settings;

    // Read config into registry.

    try
        {
        Registry::RegistryNode * node = 
            registry.create_node_path( "Settings/NodeDatabase/local_node" );
        
        registry.load_config_file( my_config, *node );
        }
    catch( ExRegistryError & e )
        {
        std::ostringstream ss;
        ss  << "Error reading local node config. " << e.what();
        throw ExSystemNodeConfigError( ss.str() );
        }
            
    // Feed node database from the data stored in the registry.

    try
        {
        node_database.load_local_node_config();
        }
    catch( ExNodeDatabaseError & e )
        {
        std::ostringstream ss;
        ss  << "Error reading local node config. " << e.what();
        throw ExSystemNodeConfigError( ss.str() );
        }
    
    // Clear "secret" entries in the registry. Once they have been read
    // they can be discarded. Note that the entries are backed up by the
    // external @c my_config so their destruction is okay.

    settings.rsa_private_key = std::string();
    }


/**
 * Uninstalls local node config.
 */
void uninstall_local_node_config()
    {
    LocalNodeSettings settings;
    
    settings.node_id = std::string();
    settings.rsa_public_key = std::string();
    settings.rsa_private_key = std::string();
    }


} // anonymous namespace


void System::connect
    (
    std::istream &  my_config
    )
    {
    hope( g_system_phase == PHASE_DATA_DOWNLOADED );
    
    // Initiate connect.
        {
        // Log.
        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Initiating connect."
            );
        
        // Clear request quit flag.
        clear_request_quit_flag();
        
        // Switch phase.
        g_system_phase = PHASE_CONNECTING;
        }
    
    // Install @c my_config, which contains new node credentials.
    // Since now the node will not be anonymous.
        {
        // Install config.
        install_local_node_config( my_config );
        
        try
            {
            // Switch node database to full mode.
            Global::node_database().switch_to_mode( NodeDatabase::MODE_FULL );
            
            try
                {
                // Notify that local node identity has changed.
                GlobalCallbacks::identity_changed();
                
                // Reinitialize node action manager.
                // Node action manager requires to enumerate all server nodes
                // in initialize() and this is the place where we know that
                // node database contains all server nodes.
                // We can't reinitialize node action manager each time
                // when node database is updated, because node action manager
                // can't be reinitialized while the node is connected.
                if( Global::g_node_action_manager )
                    {
                    Global::g_node_action_manager->initialize();
                    }
                    
                // Reinitialize global search manager.
                // It must be done here because of the same reason as node
                // action manager.
                if( Global::g_global_search_manager )
                    {
                    Global::g_global_search_manager->initialize();
                    }
                }
            catch( Exception & )
                {
                panic( "Fatal error while preparing for entering connect loop." );
                }
            }
        catch( Exception & )
            {
            // Undo work.

            try
                {
                uninstall_local_node_config();
                g_system_phase = PHASE_DATA_DOWNLOADED;
                }
            catch( Exception & )
                {
                panic( "Fatal error while recovering from a previous error." );
                }
                
            throw;
            }
        }
    
    // Enter connect loop.
        {
        try
            {
            connect_loop();
            }
        catch( Exception & )
            {
            // Undo work.
            
            try
                {
                // Reset credentials.
                uninstall_local_node_config();
                
                // Switch node database back to simple mode.
                Global::node_database().switch_to_mode( NodeDatabase::MODE_SIMPLE );
                
                // Notify that local node identity has changed.
                GlobalCallbacks::identity_changed();
                
                // Switch phase.
                g_system_phase = PHASE_DATA_DOWNLOADED;
                }
            catch( Exception & )
                {
                panic( "Fatal error recovering from an exception that occured "
                    "while in connect loop." );
                }

            throw;
            }
        }
    
    // Connected.
        {
        try
            {
            // Notify massiv core that we are connected.
            GlobalCallbacks::system_connected();

            // Create node object representing local system in the simulation.
            // Created for both server and client nodes.
            if( is_server() || is_client() )
                {
                System::fire_create_node_object_callbacks();
                }
            }
        catch( Exception & )
            {
            panic( "Could not switch to PHASE_CONNECTED because core/lib failed "
                "to handle system callback." );
            }
        }
        
    // Done.
        {
        // Log.
        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Connected."
            );
        
        // Switch phase.
        g_system_phase = PHASE_CONNECTED;
        }
    }


void System::disconnect()
    {
    hope( g_system_phase == PHASE_CONNECTED );
        
    // Initiate disconnect.
        {
        // Log.
        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Initiating disconnect."
            );
        
        try
            {
            // Clear request quit flag.
            clear_request_quit_flag();
            
            // Notify massiv core that we disconnect has been initiated.
            GlobalCallbacks::system_disconnecting();

            // Switch phase.
            g_system_phase = PHASE_DISCONNECTING;
            }
        catch( Exception & )
            {
            panic( "Error while initiating disconnect." );
            }
        }
    
    // Enter disconnect loop.
        {
        try
            {
            disconnect_loop();
            }
        catch( Exception & )
            {
            // Undo work.

            try
                {
                // Notify massiv core that we are still connected.
                GlobalCallbacks::system_connected();

                // Switch phase.
                g_system_phase = PHASE_CONNECTED;
                }
            catch( Exception & )
                {
                panic( "Disconnect failed and could not reset state back to "
                    "PHASE_CONNECTED." );
                }
            
            throw;
            }
        }

    // Reset credentials back to anonymous.
        {
        try
            {
            // Reset credentials.
            uninstall_local_node_config();
            
            // Switch node database back to simple mode.
            Global::node_database().switch_to_mode( NodeDatabase::MODE_SIMPLE );
            
            // Notify that local node identity has changed.
            GlobalCallbacks::identity_changed();
            }
        catch( Exception & )
            {
            panic( "Could not reset node credentials back to anonymous." );
            }
        }
    
    // Reinitialize object management, clean up after we have disconnected
    // from the simulation (ie. delete managed objects, ...).
        {
        ObjectManager & object_manager = Global::object_manager();

        try
            {
            object_manager.shutdown();
            object_manager.initialize();
            }
        catch( Exception & )
            {
            panic( "Object manager was unable to reinitialize." );
            }
        }
    
    // Done.
        {
        // Log.
        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Disconnected."
            );
        
        // Switch phase.
        g_system_phase = PHASE_DATA_DOWNLOADED;
        }
    }


const Core::NodeId & System::get_local_node_id()
    {
    static NodeId node_id;
    
    // We can't cache node id, because it
    // may be changed at run-time.
    
    node_id = Global::get_local_node_id();
    
    return node_id;
    }


bool System::is_server()
    {
    const NodeId & node_id = get_local_node_id();
    return node_id.is_type( NodeId::TYPE_SERVER );
    }


bool System::is_client()
    {
    const NodeId & node_id = get_local_node_id();
    return node_id.is_type( NodeId::TYPE_CLIENT );
    }


} // namespace Massiv
