/** @file system_context.h
 * System context.
 */

/*
 * begin        : Fri Dec 5 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SYSTEM_CONTEXT_H
#define MASSIV_SYSTEM_CONTEXT_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif


namespace Massiv {

/**
 * System context.
 *
 * @ingroup unclassified
 */
class MASSIV_EXPORTED SystemContext
    {
public:
        
    /**
     * Special flags that can be used to tweak core.
     * The flags can not be set from the registry and are statically
     * specified at system startup by @c System::initialize().
     */
    enum Flags
        {
        SINGLE_THREADED_SRPC    =   1
            /**< Use single thread for SRPC processing. */    
        }; // Flags

    /** Constructor. */
    SystemContext
        (
        int flags
        );

public:

    /**
     * Scheduler settings.
     */
    class MASSIV_EXPORTED SchedulerSettings : public Core::Settings
        {
    public:
            
        /** Constructor. */
        SchedulerSettings();

    public:

        Float & tick_delta_low_limit;
            /**< If time delta between two ticks is smaller than
                 @c tick_delta_low_limit, sleep time will increase.
                 This is the evidence that there were very little work
                 to do in the previous tick. */
        Float & tick_delta_high_limit;
            /**< If time delta between two ticks is larger than
                 @c tick_delta_low_limit, sleep time will decrease. */
        Float & min_sleep_time;
            /**< Minimum sleep time. */
        Float & max_sleep_time;
            /**< Maximum sleep time. */
        Float & sleep_time_increment_constant;
            /**< Constant added to @c sleep_time if idle. */
        Float & sleep_time_decrement_constant;
            /**< Constant subtracted from @c sleep_time if loaded. */

        };
    
    /**
     * SRPC settings.
     */
    class MASSIV_EXPORTED SRPCSettings : public Core::Settings
        {
    public:

        /** Constructor. */
        SRPCSettings();

    public:
            
        Integer &   num_threads;
            /**< Number of simulation threads.
                 Ignored (treated as if 1) if @c SINGLE_THREADED_SRPC flag
                 is set. */
        Float &     reply_timeout;
            /**< When to cancel SRPC call if reply was not delivered yet? */
        Boolean &   optimize_local_calls;
            /**< Optimize local calls by calling local objects directly?
                 If OPTION_OPTIMIZE is defined then the actual value is ignored
                 and is treated as if it was true.
                 Should always be true unless debugging. */
        Integer &   min_stack_size;
            /**< Minimum stack size reserved for the core. Reject SRPC calls
                 if there is less stack space than the preset limit. */
        Float &     statistics_period;
            /**< Time period between two time marks when SRPC statistics are
                 reevaluated. See @c srpc_statistics.time_mark. */

        }; // SRPCSettings

    /**
     * Scheduler statistics.
     */
    class MASSIV_EXPORTED SchedulerStatistics : public Core::Statistics
        {
    public:

        /** Constructor. */
        SchedulerStatistics();

    public:
        
        Float &         sleep_time;
            /**< Current sleep constant. */
        Boolean &       tick_boosted;
            /**< Is current tick boosted? If yes, then there will be no sleep
                 at the end of the tick. */
        Float &         total_sleep_time;
            /**< Total sleep time. */
        Float &         total_core_time;
            /**< Total time consumed by the core. */
        Float &         total_client_time;
            /**< Total time consumed by client code. */

        }; // SchedulerStatistics

    /**
     * SRPC statistics.
     * Local calls are not accounted unless @c optimize_local_calls is false.
     */
    class MASSIV_EXPORTED SRPCStatistics : public Core::Statistics
        {
    public:

        /** Constructor. */
        SRPCStatistics();

    public:

        LargeInteger &  total_calls;
            /**< Total number of SRPC calls dispatched. */
        Float &         total_wait_time;
            /**< Total time spent while waiting for SRPC replies. */
        Float &         time_mark;
            /**< System time since "recent" statistics have been gathering.
                 Reset each @c srpc_settings.statistics_period seconds. */
        Integer &       num_recent_calls;
            /**< Number of recent calls. Measured since the last time mark. */
        Float &         recent_wait_time;
            /**< Recent wait time. Measured since the last time mark. */
        Float &         average_calls_per_second;
            /**< Average number of SRPC calls per second. */
        Float &         average_response_time;
            /**< Average time spent while waiting for a SRPC reply. */

        }; // SRPCStatistics

public:    

    SchedulerSettings   scheduler_settings;
        /**< Scheduler settings. */
    SRPCSettings        srpc_settings;
        /**< SRPC settings. */
    SchedulerStatistics scheduler_statistics;
        /**< Scheduler statistics. */
    SRPCStatistics      srpc_statistics;
        /**< SRPC statistics. */
    int                 flags;
        /**< Combination of @c Flags. */

    }; // SystemContext


} // namespace Massiv

#endif // MASSIV_SYSTEM_CONTEXT_H
