/** @file system_init.cpp
 * System init/shutdown API.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "global/global.h"
#include "logger/logger.h"
#include "logger/logdest_cout.h"
#include "object/object_manager.h"
#include "object/global_search_manager.h"
#include "pointer/garbage_collector.h"
#include "replication/replica_server.h"
#include "object/class_manager.h"
#include "node/node_manager.h"
#include "node/node_action_manager.h"
#include "object/archive_manager.h"
#include "object/stack_root_object.h"
#include "sysevent/system_event_queue.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "database/archive_database.h"
#include "database/well_known_object_id_database.h"
#include "net/net.h"
#include "replication/replica_time.h"
#include "replication/replica_manager.h"
#include "registry/registry.h"
#include "object/object_provider.h"
#include "utility/debug.h"
#include "time/system_time.h"
#include "data/data_manager.h"
#include "filesys/volume_manager.h"
#include "filesys/native_file_system.h"
#include "filesys/file_ack_manager.h"
#include "thread/thread.h"
#include "thread/synch.h"
#include "thread/semaph.h"
#include "system/system.h"
#include "system/system_context.h"
#include "global/global_callbacks.h"
#include "data/account_manager.h"
#include "data/path_manager.h"
#include "thread/stack_tracer.h"
#include <vector>
#include <memory>


namespace Massiv {

using namespace Massiv::Core;


namespace {

const char *    g_env_var_config_file = "MASSIV_CONFPATH";
    /**< Name of the environment variable containing full path to the
         configuration file. */
bool            g_save_registry_on_exit = false;
    /**< Should registry content be saved on core shutdown. */

/** According to node type returns corresponding type of data manager. */
DataManager::Type get_data_manager_type
    (
    NodeId::Type node_type
    )
    {
    switch( node_type )
        {
        case NodeId::TYPE_SERVER:
            return DataManager::DATA_SERVER;
        case NodeId::TYPE_CLIENT:
            return DataManager::DATA_CLIENT;
        case NodeId::TYPE_SERVICE:
            return DataManager::MASTER_DATA_SERVICE;
        default:
            panic( "Invalid node_type." );    
        }
    }


/** Returns name of the configuration file. */
std::string get_config_file()
    {
    char * envvar = getenv( g_env_var_config_file );

    if( envvar != NULL )
        {
        return envvar;
        }
    else
        {
        return "massiv.conf";
        }
    }


/** Saves registry. */
void try_saving_registry
    (
    const std::string & config_file_name
    )
    {
    Registry & registry = Global::registry();

    try 
        {
        registry.save_config_file( config_file_name );
        }
    catch( Exception & e )
        {
        std::ostringstream ss;
        ss  << "Unable to save registry to \"" << config_file_name << "\" file. "
            << e.what();

        Global::log_warning
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_HIGH,
            ss.str()
            );
        }
    }


/** Asserts if requested subsystem is registered. */
template< typename T >
void must_not_run
    (
    T * ptr
    )
    {
    if( ptr )
        {
        std::ostringstream ss;
        ss << typeid( T ).name() << " can not be restarted!";
        panic( ss.str().c_str() );
        }
    }


/** Asserts if requested subsystem is not registered. */
template< typename T >
void require_subsystem
    (
    T * ptr
    )
    {
    if( !ptr )
        {
        std::ostringstream ss;
        ss << typeid( T ).name() << " must be started before!";
        panic( ss.str().c_str() );
        }
    }


/** Unregisters and deletes the subsystem. */
template< typename T >
void destroy_subsystem
    (
    T * & ptr
    )
    {
    delete ptr;
    ptr = NULL;
    }

}


// Implementation of System::StartUpInfo
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

System::StartUpInfo::StartUpInfo() :
    node_type( Core::NodeId::TYPE_ANONYMOUS ),
    standalone_node( false ),
    base_services( false ),
    load_config( false ),
    threading( false ),
    file_systems( false ),
    networking( false ),
    helper_subsystems( false ),
    data_management( false ),
    object_management( false ),
    system_context( false ),
    quiet_start( false ),
    save_registry_on_exit( false ),
    massiv_conf_file_name(),
    single_threaded_srpc( false ),
    register_classes_callback( NULL ),
    create_node_object_callback( NULL ),
    next_tick_callback( NULL )
    {
    }


System::StartUpInfo::StartUpInfo
    (
    Core::NodeId::Type  node_type,
    const std::string & conf_file,
    RegisterClasses *   register_classes,
    CreateNodeObject *  create_node_object,
    NextTick  *         next_tick
    ) : node_type( node_type ),
        massiv_conf_file_name( conf_file ),
        register_classes_callback( register_classes ),
        create_node_object_callback( create_node_object ),
        next_tick_callback( next_tick )
    {
    switch( node_type )
        {
        case Core::NodeId::TYPE_ANONYMOUS:
            {
            standalone_node = true;
            base_services = true;
            load_config = false;
            threading = true;
            file_systems = true;
            networking = false;
            helper_subsystems = false;
            data_management = false;
            object_management = false;
            system_context = true;
            quiet_start = false;
            save_registry_on_exit = false;
            single_threaded_srpc = false;

            break;
            }
        
        case Core::NodeId::TYPE_SERVER:
            {
            standalone_node = false;
            base_services = true;
            load_config = true;
            threading = true;
            file_systems = true;
            networking = true;
            helper_subsystems = true;
            data_management = true;
            object_management = true;
            system_context = true;
            quiet_start = false;
            save_registry_on_exit = true;
            single_threaded_srpc = false;

            break;
            }

        case Core::NodeId::TYPE_CLIENT:
            {
            standalone_node = false;
            base_services = true;
            load_config = true;
            threading = true;
            file_systems = true;
            networking = true;
            helper_subsystems = true;
            data_management = true;
            object_management = true;
            system_context = true;
            quiet_start = false;
            save_registry_on_exit = true;
            single_threaded_srpc = true;

            break;
            }

        case Core::NodeId::TYPE_SERVICE:
            {
            standalone_node = false;
            base_services = true;
            load_config = true;
            threading = true;
            file_systems = true;
            networking = true;
            helper_subsystems = true;
            data_management = true;
            object_management = true;
            system_context = true;
            quiet_start = false;
            save_registry_on_exit = true;
            single_threaded_srpc = false;

            break;
            }

        default:
            {
            panic( "No startup profile for given node type defined." );
            }
        }
    }


// Implementation of System
// ~~~~~~~~~~~~~~~~~~~~~~~~

void System::initialize
    (
    const StartUpInfo & info
    )
    {
    std::ostringstream  log;

    try
        {
        // Preconditions.
            {
            hope( !System::is_tick_locked() );
            hope( !System::is_waiting_for_rpc_reply() );
            }
        
        // Register callbacks.
            {
            if( info.register_classes_callback != NULL )
                {
                register_register_classes( info.register_classes_callback );
                }
            
            if( info.create_node_object_callback != NULL )
                {
                register_create_node_object( info.create_node_object_callback );
                }    
            
            if( info.next_tick_callback != NULL )
                {
                register_next_tick( info.next_tick_callback );
                }
            }

        // Instantiate global objects.
            {
            // Registry & logger.
            if( info.base_services )    
                {
                hope( Global::g_log_exceptions == false );

                must_not_run( Global::g_registry );
                must_not_run( Global::g_logger );
                must_not_run( Global::g_system_context );
                
                log << "\nbase_services";
                
                // Registry.
                Global::g_registry = new Registry();

                // Logger.
                Global::g_logger = new Logger();

                if( info.quiet_start ) 
                    {
                    Global::g_logger->enable( false );
                    }

                // Enable logging of exceptions.
                Global::g_log_exceptions = true;
                }

            // Load config. Must be done ASAP.
            if( info.load_config )
                {
                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );

                log << "\nconfig";

                // Read registry.
                    {
                    std::string config = info.massiv_conf_file_name;

                    if( !config.size() )
                        {
                        config = get_config_file();
                        }
                    
                    Global::g_registry->load_config_file
                        ( PathManager::resolve_path_to_config( config ) );
                    }
                }
            
            // Threading.
            if( info.threading )    
                {
                must_not_run( Global::g_semaphore_manager );
                must_not_run( Global::g_thread_manager );
                must_not_run( Global::g_critical_section_manager );
                must_not_run( Global::g_stack_tracer );
                hope( Global::g_print_stack_traces == false );

                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );

                log << "\nthreading";
                
                // Semaphore manager.
                Global::g_semaphore_manager = new SemaphoreManager();
                
                // Critical section manager.
                Global::g_critical_section_manager = new CriticalSectionManager();
                
                // Stack tracer.
                Global::g_stack_tracer = StackTracer::create();

                // Thread manager.
                Global::g_thread_manager = new ThreadManager();
                
                // Enable stack traces.
                Global::g_print_stack_traces = true;
                }
            
            // File systems.
            if( info.file_systems )    
                {
                must_not_run( Global::g_native_file_system );
                must_not_run( Global::g_volume_manager );
                must_not_run( Global::g_file_ack_manager );
                must_not_run( Global::g_path_manager );

                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );
                
                log << "\nfile_systems";

                // Native file system.
                Global::g_native_file_system = NativeFileSystem::instantiate();
                
                // Path manager.
                Global::g_path_manager = new PathManager();

                // Volume manager.
                Global::g_volume_manager = new VolumeManager();
                
                // File ack manager.
                Global::g_file_ack_manager = new FileAckManager();
                }

            // Network.
            if( info.networking )    
                {
                must_not_run( Global::g_network );
                must_not_run( Global::g_node_database );
                must_not_run( Global::g_system_event_queue );
                must_not_run( Global::g_node_manager );
                must_not_run( Global::g_node_action_manager );

                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );
                
                log << "\nnetworking";

                // Network.
                Global::g_network = new Network();
            
                // Node database.
                Global::g_node_database = new NodeDatabase();
                
                // System event queue.
                Global::g_system_event_queue = new SystemEventQueue();
                
                // Node manager.
                Global::g_node_manager = new NodeManager();
                
                // Node action manager.
                Global::g_node_action_manager = new NodeActionManager();
                }
                
            // Helper subsystems.
            if( info.helper_subsystems )    
                {
                must_not_run( Global::g_class_manager );
                must_not_run( Global::g_global_search_manager );
                must_not_run( Global::g_replica_manager );
                must_not_run( Global::g_archive_database );
                must_not_run( Global::g_archive_manager );
                must_not_run( Global::g_well_known_object_id_database );
                must_not_run( Global::g_time_manager );
                
                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );
                require_subsystem( Global::g_network );
                
                log << "\nsimulation_helper_subsystems";

                // Class manager.
                Global::g_class_manager = new ClassManager();
                
                // Global search manager.
                Global::g_global_search_manager = new GlobalSearchManager();
                
                // Replica manager.
                Global::g_replica_manager = new ReplicaManager();

                // Archive database.
                Global::g_archive_database = new ArchiveDatabase();
                
                // Archive manager.
                Global::g_archive_manager = new ArchiveManager();
                
                // Well known object id database.
                Global::g_well_known_object_id_database = new WellKnownObjectIdDatabase();
                
                // Time manager.
                Global::g_time_manager = new TimeManager();
                }

            // Data management.    
            if( info.data_management )
                {
                must_not_run( Global::g_data_manager );
                must_not_run( Global::g_account_manager );
                must_not_run( Global::g_data_manager_test_client );
                must_not_run( Global::g_data_manager_test_server );
                
                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );
                require_subsystem( Global::g_network );

                log << "\ndata_management";

                // Data manager.
                Global::g_data_manager = new DataManager();

                // Account manager.
                Global::g_account_manager = new AccountManager();
                    
                // For testing purposes.
                Global::g_data_manager_test_client = new DataManager();
                Global::g_data_manager_test_server = new DataManager();
                }
            
            // Object management.
            if( info.object_management )
                {        
                must_not_run( Global::g_object_manager );
                must_not_run( Global::g_stack_root_object );
                must_not_run( Global::g_balancer );
                must_not_run( Global::g_replica_time_ptr );
                must_not_run( Global::g_replica_server );
                must_not_run( Global::g_proxy_manager );
                must_not_run( Global::g_gc );
                
                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );
                require_subsystem( Global::g_network );

                log << "\nobject_management";

                // Object manager.
                Global::g_object_manager = new ObjectManager();
                
                // Garbage collector.
                Global::g_gc = &Global::g_object_manager->gc;

                // Object proxy manager.
                Global::g_proxy_manager = &Global::g_gc->proxy_manager;

                // Replica server.
                Global::g_replica_server = Global::g_object_manager->replica_server.get();

                // Replica time.
                Global::g_replica_time_ptr = &Global::g_replica_server->get_replica_time();

                // Balancer.
                Global::g_balancer = &Global::g_object_manager->balancer;
                
                // Stack root object.
                Global::g_stack_root_object = &Global::g_object_manager->stack_root;
                }
            
            // System context.
            if( info.system_context )
                {
                must_not_run( Global::g_system_context );
                
                require_subsystem( Global::g_registry );
                require_subsystem( Global::g_logger );

                log << "\nsystem_context";

                // System context.
                Global::g_system_context = new SystemContext
                    (
                    info.single_threaded_srpc ? 
                    SystemContext::SINGLE_THREADED_SRPC :
                    0
                    );
                }
            }

        // Initialize objects.
            {
            // Logger.
            if( info.base_services && info.load_config && info.file_systems )
                {
                // Initialize logger from the registry.
                Global::g_logger->initialize_from_registry();
                }
            
            // Networking.
            if( info.networking )    
                {
                // Initialize node database.
                // Must be initialized ASAP to make get_local_node_id() working.
                // The database is initialized with anonymous local node data.
                // Switching node database will automatically initialize (update)
                // system message queueus (Global::g_system_event_queue).
                Global::g_node_database->switch_to_mode( NodeDatabase::MODE_SIMPLE );
                
                // Initialize system message tables.
                SystemMessage::initialize_system_message_table();
                
                // Initialize network.
                Global::g_network->initialize();

                // Initialize node manager.
                Global::g_node_manager->initialize( info.standalone_node );
                }
            
            // Helper subsystems.
            if( info.helper_subsystems )    
                {
                hope( info.node_type != NodeId::TYPE_ANONYMOUS );

                // Instantiate metaobjects and factories.
                fire_register_classes_callbacks();
                
                // Initialize class manager.
                Global::g_class_manager->initialize_classes();
                
                // Global search manager is initialized in connect().
                
                // Initialize archive database.
                // Must be initialized before initialization of node manager.
                Global::g_archive_database->initialize( info.node_type );
                
                // Note: Archive manager will be initialized once connections
                //       to all server nodes are established by node manager.
                }

            // Data management.
            if( info.data_management )    
                {
                // Initialize data maanger.
                Global::g_data_manager->initialize( get_data_manager_type( info.node_type ) );
                }
            
            // Object management.
            if( info.object_management )    
                {
                // Initialize object manager.
                Global::g_object_manager->initialize();
                }
            }

        // Done.
            {
            if( g_system_phase == PHASE_UNINITIALIZED )
                {
                // System was initialized.
                g_system_phase = PHASE_INITIALIZED;

                // Callback.
                GlobalCallbacks::system_initialized();
                
                // Log.
                Global::log_info
                    (
                    Status::FACILITY_SYSTEM,
                    Status::PRIORITY_NORMAL,
                    "System was successfully initialized. "
                    "The following subsystems were started:" + log.str()
                    );
                }
            else
                {
                // Log.
                Global::log_info
                    (
                    Status::FACILITY_SYSTEM,
                    Status::PRIORITY_NORMAL,
                    "The following subsystems were (re)started:" + log.str()
                    );
                }

            g_save_registry_on_exit = info.save_registry_on_exit;
            }
        }
    catch( Exception & e )
        {
        System::shutdown();
        throw ExSystemInitializationError( e.what() );
        }
    }


void System::shutdown()
    {
    Phase old_phase = g_system_phase;

    // Prepare.
        {
        // Preconditions.
        hope( !System::is_tick_locked() );
        hope( !System::is_waiting_for_rpc_reply() );
        }
    
    // Initiate shutdown.
        {
        // Log.
        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Initiating system shutdown."
            );

        // Switch phase.
        g_system_phase = PHASE_SHUTTING_DOWN;
        }
        
    // Shutdown.
    try
        {
        // Notify core.
            {
            if( old_phase != PHASE_UNINITIALIZED )
                {
                // Notify.
                GlobalCallbacks::system_shutdown();

                // Save registry state to config file.
                if( g_save_registry_on_exit && Global::g_path_manager != NULL )
                    {
                    try_saving_registry
                        (
                        Global::path_manager().get_work_directory() + 
                        "/massiv.conf.shutdown"
                        );
                    }
                }
            }   

        // Destroy globals.
            {
            // System context.
                {
                destroy_subsystem( Global::g_system_context );
                }

            // Object management.
                {
                // Destroy object manager.
                    {
                    destroy_subsystem( Global::g_object_manager );
                    }
                
                // Clear cached pointers.
                    {
                    Global::g_stack_root_object = NULL;
                    Global::g_balancer = NULL;
                    Global::g_replica_time_ptr = NULL;
                    Global::g_replica_server = NULL;
                    Global::g_proxy_manager = NULL;
                    Global::g_gc = NULL;
                    }
                }

            // Helper subsystems subsystems.
                {
                destroy_subsystem( Global::g_class_manager );
                destroy_subsystem( Global::g_global_search_manager );
                destroy_subsystem( Global::g_replica_manager );
                destroy_subsystem( Global::g_archive_manager );
                destroy_subsystem( Global::g_archive_database );
                destroy_subsystem( Global::g_well_known_object_id_database );
                destroy_subsystem( Global::g_time_manager );
                destroy_subsystem( Global::g_account_manager );
                }

            // Networking.
                {
                destroy_subsystem( Global::g_network );
                destroy_subsystem( Global::g_node_action_manager );
                destroy_subsystem( Global::g_node_manager );
                destroy_subsystem( Global::g_node_database );
                destroy_subsystem( Global::g_system_event_queue );
                }
            
            // Data object management.
                {
                destroy_subsystem( Global::g_data_manager );
                destroy_subsystem( Global::g_data_manager_test_server );
                destroy_subsystem( Global::g_data_manager_test_client );
                }
            
            // File systems.
                {
                destroy_subsystem( Global::g_path_manager );
                destroy_subsystem( Global::g_file_ack_manager );
                destroy_subsystem( Global::g_volume_manager );
                destroy_subsystem( Global::g_native_file_system );
                }

            // Threading.
                {
                Global::g_print_stack_traces = false;
                destroy_subsystem( Global::g_critical_section_manager );
                destroy_subsystem( Global::g_semaphore_manager );
                destroy_subsystem( Global::g_thread_manager );
                destroy_subsystem( Global::g_stack_tracer );
                }

            // Logger & registry.
                {
                Global::g_log_exceptions = false;
                destroy_subsystem( Global::g_logger );
                destroy_subsystem( Global::g_registry );
                }
            }
        
        // Done.
            {
            g_system_phase = System::PHASE_UNINITIALIZED;
            }
        }
    catch( Exception & e )
        {
        panic( ( std::string( "Failed to shutdown the system. " ) + 
            e.what() ).c_str() );
        }
    }


} // namespace Massiv
