/** @file system_loops.cpp
 * System loops.
 */

/*
 * begin        : Tue Feb 24 2003
 * copyright    : (C) 2003 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/global.h"
#include "logger/logger.h"
#include "object/object_manager.h"
#include "object/throwable.h"
#include "pointer/garbage_collector.h"
#include "pointer/object_pointer.h"
#include "node/node_manager.h"
#include "node/node_action_manager.h"
#include "object/archive_manager.h"
#include "object/rpc.h"
#include "database/archive_database.h"
#include "sysevent/system_event_queue.h"
#include "time/time_manager.h"
#include "database/node_database.h"
#include "utility/debug.h"
#include "time/system_time.h"
#include "data/data_manager.h"
#include "data/account_manager.h"
#include "thread/thread.h"
#include "system/system.h"
#include "system/system_context.h"
#include "global/global_callbacks.h"
#include "net/net.h"
#include "filesys/file_ack_manager.h"
#include "registry/statistics.h"


/** @def LOG_SCHEDULER
 * Define this to enable scheduler logging.
 */
//#define LOG_SCHEDULER


namespace Massiv {

using namespace Massiv::Core;


namespace {    

STime           g_simulation_time;
    /**< Simulation time. */
bool            g_request_quit = false;
    /**< Request quit flag. */
int             g_tick_lock_count = 0;
    /**< How many times current tick was "locked". */
int             g_inside_simulation_tick_count = 0;
    /**< Are we inside Object Manager's simulation tick? */


/**
 * Sleep for a time to reduce wasting of CPU cycles if we have nothing
 * to do.
 */
void sleep()
    {
    ObjectManager &                         object_manager = Global::object_manager();
    SystemContext::SchedulerSettings &      settings = Global::system_context().scheduler_settings;
    SystemContext::SchedulerStatistics &    statistics = Global::system_context().scheduler_statistics;
    
    static Float64  previous_time = 0.0f;
    Float64         current_time = SystemTime::current_time().get_value();
    Float64         delta = current_time - previous_time - statistics.sleep_time;
    Float64         suggested_next_tick_time;
    
    // According to @c delta, which measures current CPU load
    // (time difference between two calls to @c next_tick() minus
    // current sleep time), adapt sleep time.

    if( delta < settings.tick_delta_low_limit )
        {
        statistics.sleep_time += settings.sleep_time_increment_constant;
        }
    else
    if( delta > settings.tick_delta_high_limit )
        {
        statistics.sleep_time -= settings.sleep_time_decrement_constant;
        }
        
    if( statistics.sleep_time > settings.max_sleep_time )
        {
        statistics.sleep_time = settings.max_sleep_time;
        }
    else
    if( statistics.sleep_time < settings.min_sleep_time )
        {
        statistics.sleep_time = settings.min_sleep_time;
        }

    // Query subsystems and ask them to suggest when next tick should
    // be processed.

    suggested_next_tick_time = object_manager.suggest_next_tick_time().get_value();

    if( suggested_next_tick_time > current_time &&
        suggested_next_tick_time - current_time < statistics.sleep_time )
        {
        statistics.sleep_time = 
            static_cast<Float32>( suggested_next_tick_time - current_time );
        }
    
    // Sleep.

    #ifdef LOG_SCHEDULER
    
    static Float32 previous_sleep_time;

    if( statistics.sleep_time != previous_sleep_time )
        {
        std::ostringstream ss;

        ss << "Sleep time changed to " << statistics.sleep_time * 1000.0f << " ms.";

        Global::log_debug
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_LOW,
            ss.str()
            );

        previous_sleep_time = statistics.sleep_time;
        }

    #endif
    
    if( !statistics.tick_boosted )
        {
        previous_time = current_time;
        
        Thread::sleep( static_cast<Int32>( statistics.sleep_time * 1000.0f ) );
        statistics.total_sleep_time += statistics.sleep_time;
        }
    else
        {
        statistics.tick_boosted = false;
        }

    // Log time differences between two ticks on the "level zero" of the
    // main loop.

    #ifdef LOG_SCHEDULER
    
    if( !System::inside_simulation() )
        {
        std::ostringstream  ss;
        
        ss << "System/Scheduler statistics:\n" << statistics;

        Global::log_info
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_LOW,
            ss.str()
            );
        }

    #endif
    }


} // anonymous namespace


//
// Tick handling.
//

void System::request_quit()
    {
    g_request_quit = true;
    }


void System::clear_request_quit_flag()
    {
    g_request_quit = false;
    }


void System::lock_tick()
    {
    g_tick_lock_count++;
    }


void System::unlock_tick()
    {
    g_tick_lock_count--;
    hope( g_tick_lock_count >= 0 );
    }


bool System::is_tick_locked()
    {
    return g_tick_lock_count != 0;
    }


void System::boost()
    {
    SystemContext::SchedulerStatistics &    statistics = 
        Global::system_context().scheduler_statistics;
    SystemContext::SchedulerSettings &      settings = 
        Global::system_context().scheduler_settings;

    statistics.tick_boosted = true;
    statistics.sleep_time = settings.min_sleep_time;
    }


void System::next_tick()
    {
    // Start new tick.
    do_start_tick();

    // Finish current simulation tick.
    do_simulation_tick();
    
    // Finish current tick.
    do_finish_tick();

    // Handle SRPC replies.
    handle_srpc_replies();

    // Sleep.
    sleep();
    }


void System::do_start_tick()
    {
    NodeDatabase &      node_database       = Global::node_database();
    ObjectManager &     object_manager      = Global::object_manager();
    SystemEventQueue &  system_event_queue  = Global::system_event_queue();
    Network &           network             = Global::network();
    ArchiveManager &    archive_manager     = Global::archive_manager();
    NodeManager &       node_manager        = Global::node_manager();
    NodeActionManager & node_action_manager = Global::node_action_manager();
    DataManager &       data_manager        = Global::data_manager();
    TimeManager &       time_manager        = Global::time_manager();
    FileAckManager &    file_ack_manager    = Global::file_ack_manager();
    ArchiveDatabase &   archive_database    = Global::archive_database();
    Logger &            logger              = Global::logger();
    VolumeManager &     volume_manager      = Global::volume_manager();
    AccountManager &    account_manager     = Global::account_manager();
    
    // Test if can start a new tick.
    if( System::is_tick_locked() )
        {
        panic
            (
            "Previous tick has not finished yet or a new one can not be "
            "started in this context."
            );
        }

    Statistics::MeasureSystemTime   clock
        ( Global::system_context().scheduler_statistics.total_core_time );

    try
        {
        // Callback - we are beginning new tick.
        GlobalCallbacks::next_tick_start();
        
        // Set up current simulation time.
        time_manager.next_tick();
        g_simulation_time = time_manager.get_time();
        
        // Push all incoming network packets in the system queue.
        network.next_tick();

        // Execute all scheduled system events.
        system_event_queue.execute_all_events();
        
        // Advance archivation.
        archive_manager.next_tick();
        
        // Advance node database.
        node_database.next_tick();
        
        // Advance node management.
        node_manager.next_tick();
        
        // Advance node action manager.
        node_action_manager.next_tick();

        // Advance data management.
        data_manager.next_tick();
        
        // Advance non-simulation stuff like GC and localization caches.
        object_manager.next_tick();
        
        // Remove obsolete archives.
        archive_database.next_tick();
        
        // Advance file ack manager.
        file_ack_manager.next_tick();
            
        // Advance file systems.
        volume_manager.next_tick();
        
        // Advance account manager.
        account_manager.next_tick();

        // Flush logger.
        logger.flush();
        }
    catch( Exception & e )
        {
        std::ostringstream ss;
        
        ss << "Non-simulation exception caught: " << e.what()
           << "\nTrying to propagate the error and quit from simulation properly.";

        Global::log_warning
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_HIGHEST,
            ss.str()
            );

        request_quit();
        throw;
        }
    }


void System::do_simulation_tick()
    {
    ObjectManager &                 object_manager = Global::object_manager();
    Statistics::MeasureSystemTime   clock
        ( Global::system_context().scheduler_statistics.total_client_time );

    try
        {
        g_inside_simulation_tick_count++;

        // Advance simulation.
        object_manager.simulation_next_tick();
        
        g_inside_simulation_tick_count--;
        }
    catch( Throwable & e )
        {
        g_inside_simulation_tick_count--;

        // This is a simulation related error leaked from an upcall
        // issued by ObjectManager.
        // Ignore the error.
        
        Global::log_warning
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_HIGHEST,
            std::string( "Ignoring a managed exception leaked from the "
            "application: " ) + e.what()
            );
        }
    catch( Exception & e )
        {
        g_inside_simulation_tick_count--;

        // This error can not not be simply accounted to any managed object.
        
        // What to do next?
        // 1) ignore the error
        // 2) rethrow
        // 3) request quit from the current loop
        //    [ @c next_tick() may still be called until the loop is 
        //      cancelled! ]
        
        Global::log_warning
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_HIGHEST,
            std::string( "Non-managed exception leaked from the application: " ) +
            e.what() + "\nTrying to propagate the error and quit from simulation "
            " properly."
            );

        request_quit();
        throw;
        }
    }


void System::do_finish_tick()
    {
    // Test if current tick can be finished.
    if( System::is_tick_locked() )
        {
        panic( "Current tick is locked and must not be finished now." );
        }
    
    // Callback - we are finishing current tick.
        {
        GlobalCallbacks::next_tick_finish();
        
        try
            {
            lock_tick();
            System::fire_next_tick_callbacks();
            unlock_tick();
            }
        catch( Exception & e )
            {
            std::ostringstream ss;

            ss  << "Exception leaked from system next tick callback. Can not continue." 
                << e.what();

            panic( ss.str().c_str() );
            }
        }
    }


bool System::quit_requested()
    {
    return g_request_quit;
    }


bool System::inside_simulation()
    {
    if( g_inside_simulation_tick_count > 0 )
        {
        return true;
        }

    if( Global::garbage_collector().is_object_upcalled( NULL ) )
        {
        return true;
        }

    return false;
    }


//
// System loops.
//
    
void System::download_data_loop()
    {
    DataManager & data_manager = Global::data_manager();

    // Log.
    Global::log_debug
        (
        Status::FACILITY_SYSTEM,
        Status::PRIORITY_NORMAL,
        "Downloading data from data service node..."
        );
    
    // Wait until data manager downloads all required data.
    while( !data_manager.is_ready_for_use() )
        {
        if( System::quit_requested() )
            {
            throw ExSystemError( "Quit requested in download_data_loop()." );
            }
        System::next_tick();
        }
    }


void System::connect_loop()
    {
    NodeManager & node_manager = Global::node_manager();
    SystemTime    start_time = System::system_time();

    // Log.
    Global::log_debug
        (
        Status::FACILITY_SYSTEM,
        Status::PRIORITY_NORMAL,
        "Connecting to simulation..."
        );
        
    // Start connecting.
    node_manager.connect();    
    
    // Wait until connected.
    while( !node_manager.is_connected() )
        {
        // Every second. Better do not configure via settings.
        if( System::system_time() - 1 > 
            start_time && is_server() )
            {
            start_time = System::system_time();
            node_manager.server_do_connect_attempt();
            }
        if( System::quit_requested() )
            {
            throw ExSystemError( "Quit requested in connect_loop()." );
            }
        System::next_tick();
        }
    }


void System::disconnect_loop()
    {
    NodeManager & node_manager = Global::node_manager();

    // Log.
    Global::log_debug
        (
        Status::FACILITY_SYSTEM,
        Status::PRIORITY_NORMAL,
        "Disconnecting..."
        );
        
    // Start disconnecting.
    node_manager.disconnect();
    
    // Wait until disconnected.
    while( node_manager.is_connected() )
        {
        if( System::quit_requested() )
            {
            throw ExSystemError( "Quit requested in disconnect_loop()." );
            }
        System::next_tick();
        }
    }


void System::generic_loop()
    {
    hope( !is_connected() );

    clear_request_quit_flag();
    
    while( !quit_requested() )
        {
        System::next_tick();
        }
    }


//
// System phase.
//

bool System::is_connected()
    {
    return g_system_phase == PHASE_CONNECTED || g_system_phase == PHASE_RUNNING;
    }


System::Phase System::get_phase()
    {
    return g_system_phase;
    }
    

//
// Time.
//

const SystemTime & System::system_time()
    {
    return Global::system_time();
    }


const STime & System::time()
    {
    return g_simulation_time;
    }


const SystemTime & System::total_sleep_time()
    {
    SystemContext::SchedulerStatistics & statistics = 
        Global::system_context().scheduler_statistics;
    static SystemTime time;
    
    time = statistics.total_sleep_time;
    return time;
    }


} // namespace Massiv
