/** @file system_run.cpp
 * System run, SRPC handling.
 *
 * Waiting for SRPC replies uses one of the following methods. The method
 * selection depends on the number of available simulation threads:
 *
 *  - Stack-based waiting:
 *      When caller blocks (waits for RPC reply) system will run a new
 *      instance of the main loop. This is known as main loop nesting.
 *      Main loop at the level zero denotes the original loop (no one waits
 *      for SRPC reply), level one the first nested loop, etc.
 *      The main drawback is LIFO order of SPRC request processing.
 *      If caller A at level zero blocks and waits for SRPC result and there
 *      is another request at level one, A has to wait until the call
 *      at level one completes.
 *      The advantage is implementation simplicity.
 *
 *  - Thread-based waiting:
 *      When caller blocks the nested loop will be started in a different
 *      thread (if there is any, otherwise stack-based waiting will be used).
 *      This allows to resume execution of the caller as soon as the reply is
 *      received without the need for waiting for completion of the nested
 *      loops.
 *      Note that two simulation threads never run in parallel so locking is
 *      not required.
 */

/*
 * begin        : Tue Sep 29 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/global.h"
#include "object/object_fwd.h"
#include "pointer/object_pointer.h"
#include "pointer/garbage_collector.h"
#include "object/rpc.h"
#include "thread/thread.h"
#include "thread/semaph.h"
#include "registry/statistics.h"
#include "system/system.h"
#include "system/system_context.h"
#include "global/global_callbacks.h"
#include "utility/debug.h"
#include "utility/variant_pointer.h"


#ifdef OPTION_DEBUG

/** @def LOG_SRPC_CALLS
 * Define this to log SRPC processing.
 */
//#define LOG_SRPC_CALLS

#endif


namespace Massiv {

using namespace Massiv::Core;


namespace {    

/**
 * Structure that holds context of a pending SRPC call.
 * ResultObject and previous SRPC call in the same stack of pending calls.
 */
struct CallContext
    {
public:

    /** Constructor. */
    CallContext() : previous_call( NULL )
        {
        }

    /** Dumps call context. */
    friend std::ostream & operator<<
        (
        std::ostream &      stream,
        const CallContext & call_context
        )
        {
        stream  << "result = " << call_context.result_object->to_string() << "\n"
                << "time = " << call_context.time;
        
        return stream;
        }

public:

    Pointer<ResultObject>   result_object;
        /**< RPC result object. */
    SystemTime              time;
        /**< Submit time. */
    CallContext *           previous_call;
        /**< Previous call in the stack of pending SRPC calls associated
             with the same thread. */
    }; // CallContext


/**
 * Context associated with a simulation thread running @c run_loop().
 * Maintains a stack of pending SRPC calls blocked in this thread.
 */
struct ThreadContext
    {
public:
    
    /** Constructor. */
    ThreadContext() : 
        run( Global::semaphore_manager().create( 0 ) ),
        raise_exception( false ),
        last_call( NULL )
        {
        }    

    /** Destructor. */
    ~ThreadContext()
        {
        delete run;
        }

public:
        
    Reference<Thread>   thread;
        /**< Thread object. */
    Semaphore *         run;
        /**< Signalled when given thread should run. */
    std::string         exception;
        /**< Exception string. */
    bool                raise_exception;
        /**< Raise exception with @c exception message once the thread
             is resumed? */
    CallContext *       last_call;
        /**< Context of a SRPC call on top of the stack of pending SRPC calls
             associated with this thread. This is the call the thread waits
             until completes. */
             
    }; // ThreadContext


Int32           g_num_threads = 0;
    /**< Number of simulation threads available.
         A copy of @c SystemContext::SRPCSettings::num_threads. */
ThreadContext * g_thread_contexts = NULL;
    /**< Thread contexts array including the original thread allocated at
         slot 0. @c g_num_threads items. */
ThreadContext * g_current_thread = NULL;
    /**< Context of the current thread. Only one thread runs at a time. */
ThreadContext * g_original_thread = NULL;
    /**< Context of the original thread. */
Int32           g_pending_srpc_count = 0;
    /**< Number of pending SRPC calls. */
bool            g_quit_initiated = false;
    /**< Quit from run_loop()? Are threads shutting down? */


/** 
 * Initializes SRPC structures and launches simulation threads.
 * @c run_loop points to @c System::run_loop() (passed as a parameter
 * to avoid friend declarations).
 */
void launch_srpc_threads
    (
    Thread::ThreadRoutine run_loop
    )
    {
    SystemContext &                 context = Global::system_context();
    SystemContext::SRPCSettings &   settings = context.srpc_settings;
    SystemContext::SRPCStatistics & statistics = context.srpc_statistics;
    ThreadManager &                 thread_manager = Global::thread_manager();
    int                             i;
    
    if( flags_on( context.flags, SystemContext::SINGLE_THREADED_SRPC ) )
        {
        settings.num_threads = 1;
        }

    if( settings.num_threads < 1 )
        {
        panic( "Invalid SRPCSettings/num_threads!" );
        }

    hope( g_thread_contexts == NULL );

    // Initialize.
    
    statistics.reset();

    g_num_threads = settings.num_threads;
    g_thread_contexts = new ThreadContext[ g_num_threads ]();
    g_quit_initiated = false;

    // Launch additional threads.
    // They will automatically block at the beginning of @c run_loop()
    // and wait for run request.
    
    for( i = 1; i < g_num_threads; i++ )
        {
        g_thread_contexts[ i ].thread =
            thread_manager.create( run_loop, VariantPointer( 
                &g_thread_contexts[ i ] ) );
        }
    
    // Allow current (= original) thread to enter @c run_loop.
    // Current thread will be the only running thread.
    
    g_original_thread = &g_thread_contexts[ 0 ];
    g_original_thread->run->up();
    g_current_thread = g_original_thread;
    }


/** Shutdowns simulation threads. */
void shutdown_srpc_threads()
    {
    int i;
    
    // Initiate quit from the run loops.
    
    g_quit_initiated = true;
    System::request_quit();

    // Cancel all pending calls.

    for( i = 0; i < g_num_threads; i++ )
        {
        CallContext * call_context = g_thread_contexts[ i ].last_call;

        while( call_context != NULL )
            {
            call_context->result_object->cancel();
            call_context = call_context->previous_call;
            }
        }

    // Correctly close all threads.
    // Each thread is awaken to shutdown itself.
    
    if( g_thread_contexts != NULL )
        {
        for( i = 1; i < g_num_threads; i++ )
            {
            ThreadContext & thread_context = g_thread_contexts[ i ];

            thread_context.run->up();
            thread_context.thread->join();
            }
        }

    // Deinitialize.
    
    delete [] g_thread_contexts;

    g_current_thread = NULL;
    g_original_thread = NULL;
    g_thread_contexts = NULL;
    g_num_threads = 0;
    g_quit_initiated = false;
    }


/**
 * Finds simulation thread that can be awaken and is not waiting for SRPC
 * result.
 * Returns NULL if no such thread is available.
 */
ThreadContext * allocate_thread()
    {
    int i;

    for( i = 0; i < g_num_threads; i++ )
        {
        ThreadContext & thread_context = g_thread_contexts[ i ];

        if( thread_context.last_call != NULL )
            {
            // Can not allocate thread that waits for a SRPC result.
            // Note that current thread will alredy have been "blocked" by
            // the time when @c allocate_thread() is called.

            continue;
            }

        return &thread_context;
        }

    // No thread available.

    return NULL;
    }


/**
 * Performs actions when thread is resumed.
 * Program must call this whenever thread is resumed.
 */
void thread_resumed
    (
    ThreadContext & context
    )
    {
    // Update context.

    g_current_thread = &context;

    // Log.

    #ifdef LOG_SRPC_CALLS

    std::ostringstream ss;

    ss << "Thread " << g_current_thread - g_thread_contexts << " was resumed.";
    
    Global::log_debug
        (
        Status::FACILITY_RPC,
        Status::PRIORITY_LOWEST,
        ss.str()
        );

    #endif

    // Boost self.

    System::boost();

    // Notify core.

    GlobalCallbacks::simulation_thread_resumed
        ( g_current_thread - g_thread_contexts );
    
    // Raise queued exception if requested.

    if( context.raise_exception )
        {
        context.raise_exception = false;
        throw ExSystemError( context.exception );
        }
    }


/**
 * Switches to @c target_thread, blocks self.
 * Cooperative multitasking like, two threads never run in parallel.
 * Thread switch always happens at a well-defined (safe) point in the program.
 */
void switch_to_thread
    (
    ThreadContext & target_thread
    )
    {
    ThreadContext & current_thread = *g_current_thread;

    // Avoid deadlocks.
    // If client code was upcalled and did not wait for a SRPC reply then
    // there would be a risk that system would never switch back to the
    // current thread.
    // Client code must either be not upcalled in the current thread or
    // must be waiting for a SRPC reply.

    hope( !g_quit_initiated );
    hope( Global::garbage_collector().who() == NULL || 
          g_current_thread->last_call != NULL );

    // Update context.

    g_current_thread = NULL;

    // Wake up target thread. Block self.

    target_thread.run->up();
    current_thread.run->down();
    
    thread_resumed( current_thread );
    }


/** Switches to target thread and throws exception in its context. */
void raise_exception
    (
    ThreadContext &     target_thread,
    const std::string & message
    )
    {
    hope( !g_quit_initiated );

    if( g_current_thread == &target_thread )
        {
        // Raise exception in the current thread.

        throw ExSystemError( message );
        }
    else
        {
        // Queue exception.

        hope( !target_thread.raise_exception );

        target_thread.raise_exception = true;
        target_thread.exception = message;
        
        // Switch to target thread.

        switch_to_thread( target_thread );
        }
    }


/** Logs internal state. */
void log_all()
    {
    std::ostringstream  ss;
    int                 i;

    ss << "Number of pending SRPC calls = " << g_pending_srpc_count;

    for( i = 0; i < g_num_threads; i++ )
        {
        ThreadContext & thread_context = g_thread_contexts[ i ];
        CallContext *   call_context = thread_context.last_call;

        ss << "\nThread " << i << ": ";

        while( call_context != NULL )
            {
            ss << "\n-- call --:\n" << *call_context;
            call_context = call_context->previous_call;
            }
        }
    
    Global::log_debug
        (
        Status::FACILITY_RPC,
        Status::PRIORITY_LOWEST,
        ss.str()
        );
    }


} // anonymous namespace


int System::run_loop
    (
    VariantPointer arg
    )
    {
    ThreadContext & thread_context = *variant_cast<ThreadContext *>( arg );
    
    // Wait until we are asked to run.
    
    thread_context.run->down();
    thread_resumed( thread_context );

    // Run loop.
    
    while( !g_quit_initiated )
        {
        // Do next tick.
        
        try
            {
            next_tick();
            }
        catch( Exception & e )
            {
            // This error is not related to simulation.
            // Signal it to the original thread unless we are already
            // quitting.

            if( !g_quit_initiated )
                {
                raise_exception( *g_original_thread, e.what() );
                }
            }
        
        // If @c quit_requested(), we are the original thread and all SRPC
        // calls have been completed/cancelled then quit the loop.
        // This initiates quit (@c g_quit_initiated). Quit request is ignored
        // otherwise.
        
        if( quit_requested() && g_current_thread == g_original_thread &&
            !is_waiting_for_rpc_reply() )
            {
            break;
            }
        }

    return 0;
    }


void System::handle_srpc_replies()
    {
    SystemContext::SRPCSettings &   settings = Global::system_context().srpc_settings;
    SystemTime                      current_time = SystemTime::current_time();
    int                             i;

    // Can run?

    if( g_system_phase != PHASE_RUNNING )
        {
        return;
        }
    
    if( g_quit_initiated )
        {
        // If quit is initiated the original thread may be waiting for us
        // to complete thus we can not switch to it (that would lead to a
        // deadlock) or to any other thread.
        // @c shutdown_srpc_threads() wakes up the thread one by one and waits
        // their completion as needed.
        
        return;
        }

    // Test what SRPCs associated with simulation threads have completed
    // and find out what thread has been waiting for the longest time.
    // Note that only top-most calls need be inspected.

    ThreadContext * thread_to_resume = NULL;
    SystemTime      max_wait_time;
    
    for( i = 0; i < g_num_threads; i++ )
        {
        ThreadContext & thread_context = g_thread_contexts[ i ];
        CallContext *   call_context = thread_context.last_call;
        SystemTime      timeout = settings.reply_timeout.read();
        SystemTime      wait_time;

        if( call_context == NULL )
            {
            // Thread i does not wait for a SRPC completion.

            continue;
            }

        wait_time = current_time - call_context->time;

        if( wait_time > timeout || quit_requested() )
            {
            // We are waiting too long, cancel the call.
            
            Global::log_info
                (
                Status::FACILITY_RPC,
                Status::PRIORITY_HIGHEST,
                std::string( "Cancelling SRPC call because " ) +
                ( quit_requested() ? "quit was requested" : "of timeout" ) +
                ". Details:\n" + call_context->result_object->to_string()
                );

            call_context->result_object->cancel();
            }

        if( call_context->result_object->state != ResultObject::STATE_UNKNOWN )
            {
            // Received a reply for the SRPC the thread i waits for.
            // Remember caller's thread so that we could switch to it unless
            // there is another thread that can be resumed and that has been
            // waiting longer.

            if( wait_time >= max_wait_time )
                {
                max_wait_time = wait_time;
                thread_to_resume = &thread_context;
                }
            }
        }
        
    // Switch to the elected thread. Suspend self.

    if( thread_to_resume != NULL )
        {
        #ifdef LOG_SRPC_CALLS

        std::ostringstream ss;
        
        ss  << "Switching to thread " << thread_to_resume - g_thread_contexts
            << " as its last SRPC call completed and the thread waits for the"
            << " longest time. "
            << *thread_to_resume->last_call;
        
        Global::log_debug
            (
            Status::FACILITY_RPC,
            Status::PRIORITY_LOWEST,
            ss.str()
            );

        #endif
        
        if( g_current_thread != thread_to_resume )
            {
            switch_to_thread( *thread_to_resume );
            }

        return;
        }

    // If there are no pending SRPCs and we are not the original thread,
    // "switch" to it. This ensures that the other threads run only
    // iff they are either completing the last simulation tick
    // or the original thread waits for SRPC.

    if( !g_pending_srpc_count && g_current_thread != g_original_thread )
        {
        // Log.

        #ifdef LOG_SRPC_CALLS

        Global::log_debug
            (
            Status::FACILITY_RPC,
            Status::PRIORITY_LOWEST,
            "Switching to original thread because no SRPCs are pending."
            );

        #endif

        // Switch to original thread.

        switch_to_thread( *g_original_thread );
        return;
        }
    }


void System::wait_for_rpc_reply
    (
    const Pointer<ResultObject> &   result_object
    )
    {
    SystemContext::SRPCSettings &   settings = Global::system_context().srpc_settings;
    SystemContext::SRPCStatistics & statistics = Global::system_context().srpc_statistics;
    Float64                         time1 = SystemTime::current_time().get_value();

    hope( g_system_phase == PHASE_RUNNING );

    // Test if can wait for the call.
        {
        if( is_tick_locked() )
            {
            RAISE_CORE_LIB_EXCEPTION
                (
                RemoteCallFailedException,
                "System prohibits calling synchronous RPC in this context!"
                );
            }

        if( !have_enough_stack_space() )
            {
            Global::log_warning
                (
                Status::FACILITY_RPC,
                Status::PRIORITY_HIGHEST,
                "Cancelling the call to prevent stack overflow. Details:\n" +
                result_object->to_string()
                );

            result_object->cancel();
            return;
            }

        if( quit_requested() )
            {
            Global::log_warning
                (
                Status::FACILITY_RPC,
                Status::PRIORITY_HIGHEST,
                "Cancelling the call because quit was requested. Details:\n" +
                result_object->to_string()
                );
            
            result_object->cancel();
            return;
            }

        hope( !g_quit_initiated );
        }

    // Spawn new instance of the main loop and wait for reply.
        {
        CallContext call_context;
        
        // Register call.

        call_context.result_object = result_object;
        call_context.time = SystemTime::current_time();
        call_context.previous_call = g_current_thread->last_call;
        
        g_current_thread->last_call = &call_context;
        g_pending_srpc_count++;

        // Log.

        #ifdef LOG_SRPC_CALLS
        
        std::ostringstream ss;

        ss  << "Thread " << g_current_thread - g_thread_contexts
            << " waits for a SRPC reply.";

        Global::log_debug
            (
            Status::FACILITY_RPC,
            Status::PRIORITY_LOWEST,
            ss.str()
            );

        log_all();

        #endif

        // Wait for the reply.

        try
            {
            // Do not sleep at the end of the tick.

            boost();

            // Finish current tick.
            // Warning! This calls @c handle_srpc_replies() and may switch
            // thread. By now current SRPC call must have already been
            // registered or system might not switch back to this thread.

            do_finish_tick();
            
            // Allocate simulation thread that will run new instance of the
            // main loop.

            ThreadContext * simulation_thread = allocate_thread();
            
            if( simulation_thread == NULL )
                {
                // No simulation thread available.
                // Dispatch the call using stack-based semantics, simple
                // @c next_tick() recursion.
                        
                // Wait until RPC result is delivered.
                // System can switch to a different thread meanwhile if
                // such a thread waits for SRPC completion and the reply
                // was received.

                while( result_object->state == ResultObject::STATE_UNKNOWN )
                    {
                    System::next_tick();
                    }
                }
            else
                {
                // Thread-based waiting.
                // Enter main loop ran by @c simulation_thread.
                // Because the loop will be ran within a context of a different
                // thread we will be able to resume current thread as soon as
                // the reply is received.

                if( result_object->state == ResultObject::STATE_UNKNOWN )
                    {
                    switch_to_thread( *simulation_thread );
                    }

                // Test if we have been awaken correctly.
                
                hope( result_object->state != ResultObject::STATE_UNKNOWN );
                }
            
            // Start new tick.

            do_start_tick();
            }
        catch( Exception & )    
            {
            // Unregister call.

            g_current_thread->last_call = call_context.previous_call;
            g_pending_srpc_count--;
            
            // Cancel call.

            result_object->cancel();

            // Rethrow.

            throw;
            }

        // Unregister call.

        g_current_thread->last_call = call_context.previous_call;
        g_pending_srpc_count--;

        // Log.
        
        #ifdef LOG_SRPC_CALLS
        
        log_all();

        #endif
        }

    // Update statistics.
        {
        Float64 time2 = SystemTime::current_time().get_value();
        Float32 delta = static_cast<Float32>( time2 - time1 );

        statistics.total_calls++;
        statistics.num_recent_calls++;

        statistics.total_wait_time += delta;
        statistics.recent_wait_time += delta;
        
        if( statistics.time_mark + settings.statistics_period < time2 )
            {
            statistics.average_calls_per_second = static_cast<Float32>(
                statistics.num_recent_calls / ( time2 - statistics.time_mark ) );
            statistics.average_response_time = statistics.recent_wait_time / 
                statistics.num_recent_calls;

            statistics.num_recent_calls = 0;
            statistics.recent_wait_time = 0;
            statistics.time_mark = static_cast<Float32>( time2 );
            }
        }
    }


void System::run()
    {
    TAG_STACK_FRAME( "::Massiv::System::run()" );
    hope( g_system_phase == PHASE_CONNECTED );
    
    // Initialize.
        {
        // Clear quit flag.
        
        clear_request_quit_flag();

        // Launch additional threads.
        
        launch_srpc_threads( &System::run_loop );

        // Switch phase.
        
        g_system_phase = PHASE_RUNNING;
        }

    // Enter @c run_loop().
        {
        // Log.
        
        Global::log_debug
            (
            Status::FACILITY_SYSTEM,
            Status::PRIORITY_NORMAL,
            "Running main loop..."
            );

        // Run loop.
        
        try
            {
            run_loop( VariantPointer( g_original_thread ) );
            }
        catch( Exception & )
            {
            // Undo work.

            try
                {
                // Shutdown threads.

                shutdown_srpc_threads();
                
                // Switch phase.

                g_system_phase = PHASE_CONNECTED;
                }
            catch( Exception & )
                {
                panic( "Could not recover from an exception leaked from the main "
                    "run loop." );
                }
            
            throw;
            }
        }

    // Deinitialize.
        {
        // Shutdown threads.
        
        try
            {
            shutdown_srpc_threads();
            }
        catch( Exception & )
            {
            panic( "Unable to shutdown SRPC threads." );
            }
        
        // Switch phase.
        
        g_system_phase = PHASE_CONNECTED;
        }
    }

    
bool System::is_waiting_for_rpc_reply()
    {
    return g_pending_srpc_count > 0;
    }


bool System::have_enough_stack_space()
    {
    SystemContext::SRPCSettings & settings = Global::system_context().srpc_settings;

    return Thread::get_remaining_stack_size() >= 
        static_cast<UInt32>( settings.min_stack_size );
    }


} // namespace Massiv
