/** @file test.h
 * Base testing class.
 */

/*
 * begin        : Tue Nov 27 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_TEST_H
#define MASSIV_TEST_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <string>


namespace Massiv {
namespace Core {
namespace Test {


/** @def test_condition_
 * Test if condition is true. Note that @this must point to object of
 * class compatible with @a Test.
 */
#define test_condition_( cond ) \
    do { this->test_condition( ( cond ), #cond, __FILE__, __LINE__ ); } while( 0 )

/** @def test_fail_
 * Test failed.
 */
#define test_fail_( msg ) \
    do { this->test_fail( msg, __FILE__, __LINE__ ); } while( 0 )

/** @def test_success_
 * Test succeeded.
 */
#define test_success_() \
    do { this->test_success(); } while( 0 )


class Test
    {
public:

    /**
     * Constructor.
     * @param name Test name.
     */
    Test
        (
        const std::string & name
        );

    /**
     * Destructor.
     *
     * @todo Better report.
     */
    virtual ~Test();

public:

    /**
     * Run test
     */
    virtual void test() = 0;

    /**
     * Set #fail_exception to @c true.
     */
    void exception_failed();

protected:

    /**
     * Test condition.
     */
    void test_condition
        (
        bool                result,
        const std::string & message,
        const char *        file,
        int                 line
        );

    /**
     * Test failed.
     *
     * @todo Use output specified by user, possibly logger.
     */
    void test_fail
        (
        const std::string & message,
        const char *        file,
        int                 line
        );

    /**
     * Test succeeded.
     */
    void test_success();

private:

    std::string     name;
        /**< Test name. */

    int             success_count;
        /**< Number of successful tests. */

    int             fail_count;
        /**< Number of failed tests. */

    bool            fail_exception;
        /**< Unexpected exception thrown during tests. */

    };


}; // namespace Massiv::Core::Test
}; // namespace Massiv::Core
}; // namespace Massiv


#endif // _TEST_H_
