/** @file test_all.cpp
 * Main tester.
 */

/*
 * begin        : Tue Nov 27 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_all.h"
#include "status/exception.h"
#include "global/global.h"
#include "system/system.h"
#include "logger/logger.h"
#include "registry/registry.h"
#include "test/test_logger.h"
#include "test/test_bitiostream.h"
#include "test/test_bitiostream_speed.h"
#include "test/test_crypto.h"
#include "test/test_container.h"
#include "test/test_serializer.h"
#include "test/test_objects.h"
#include "test/test_heap.h"
#include "test/test_cache.h"
#include "test/test_union_find.h"
#include "test/test_array.h"
#include "test/test_registry.h"
#include "test/test_sockets.h"
#include "test/test_architecture.h"
#include "test/test_filesys.h"
#include "test/test_fs_enumeration.h"
#include "test/test_volume_enumeration.h"
#include "test/test_volume_manager.h"
#include "test/test_thread.h"
#include "test/test_factgen.h"
#include "test/test_dataobject.h"
#include "test/test_time.h"
#include "test/test_object_provider.h"
#include "test/test_vector.h"
#include "test/class_list.h"
#include <typeinfo>
#include <iostream>
#include <sstream>


namespace Massiv {
namespace Core {
namespace Test {


/** @class TestFactory
 * Factory that creates testing object.
 */
template <class Type>
struct TestFactory
    {
    static Test * create_test()
        {
        return new Type;
        }
    };

typedef Test * CreateTest();


/**
 * Basic test of classes required by Test itself.
 */
static CreateTest * self_tests[] =
    {
    TestFactory<ArchitectureTest>::create_test,
    TestFactory<LoggerTest>::create_test,
    NULL
    };


/**
 * All standard tests.
 * @todo Divide this into several groups.
 */
static CreateTest * all_tests[] =
    {
    TestFactory<BitIOStreamTest>::create_test,
    TestFactory<CryptoTest>::create_test,
    TestFactory<VectorTest>::create_test,
    TestFactory<ContainerTest>::create_test,
    TestFactory<SerializerTest>::create_test, 
    TestFactory<HeapTest>::create_test,
    TestFactory<CacheTest>::create_test,
    TestFactory<UnionFindTest>::create_test,
    TestFactory<ArrayTest>::create_test,
    TestFactory<FactoryGeneratorTest>::create_test,
    TestFactory<RegistryTest>::create_test,
    TestFactory<FsEnumerationTest>::create_test,
    TestFactory<VolumeEnumerationTest>::create_test,
    TestFactory<VolumeManagerTest>::create_test,
    TestFactory<ThreadTest>::create_test,
    TestFactory<DataObjectTest>::create_test,
    NULL
    };


/**
 * Tests that take a lot of time to complete, especially on DEBUG.
 *
 * Some of them (BitIOStreamSpeedTest) are benchmarks, so it's ok, but the rest
 * should be fixed or made into separate programs.
 */
static CreateTest * long_tests[] =
    {
    TestFactory<FileSystemTest>::create_test,
    TestFactory<BitIOStreamSpeedTest>::create_test,
    TestFactory<TimeTest>::create_test,
    NULL
    };


/**
 * Tests that fail.
 * Move them here until things are fixed.
 */
static CreateTest * fail_tests[] =
    {
    TestFactory<ObjectManagementTest>::create_test,
    TestFactory<SocketsTest>::create_test,
    NULL
    };


static void test_group
    (
    CreateTest ** ct
    )
    {
    Test *          test;

    hope( ct );
    while( *ct )
        {
        test = NULL;
        try
            {
            test = (*ct)();
            hope( test );
            test->test();
            delete test;
            }
        catch( std::exception & e )
            {
            std::cout << "Unhandled exception leaked from " <<
                        typeid( *test ).name() << std::endl;
            std::cout << e.what() << std::endl;
            test->exception_failed();
            delete test;
            }
        catch( ... )
            {
            std::cout << "Unhandled exception leaked from " <<
                        typeid( *test ).name() << std::endl;
            test->exception_failed();
            delete test;
            }
        ct++;
        }
    }


void test_all()
    {
    try 
        {
        std::cout << "Initializing Massiv core..." << std::endl;

        System::StartUpInfo startup_info
            (
            NodeId::TYPE_ANONYMOUS,
            "",
            &Massiv::Generated::register_classes
            );
        startup_info.load_config = false;
        startup_info.quiet_start = true;
        System::initialize( startup_info );
        Global::logger().enable( true );

        test_group( self_tests );
        }
    catch( Exception & e )
        {
        std::cout << "Unhandled exception." << std::endl;
        std::cout << e.what() << std::endl;

        return;
        }
    catch( std::exception & e )
        {
        std::cout << "Unhandled exception." << std::endl;
        std::cout << e.what() << std::endl;

        return;
        }
    catch( ... )
        {
        std::cout << "Unhandled exception. " << std::endl;

        return;
        }

    test_group( all_tests );
    test_group( long_tests );
    test_group( fail_tests );

    System::shutdown();
    }


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
