/** @file test_architecture.cpp
 * Architecture, OS and compiler miscellaneous tests.
 */

/*
 * begin        : Thu May 16 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_architecture.h"
#include "utility/types.h"
#include "utility/hash.h"
#include <map>

using namespace Massiv::Core;


namespace Massiv {
namespace Core {
namespace Test {


ArchitectureTest::ArchitectureTest() :
    Test( "architecture" )
    {
    }

ArchitectureTest::~ArchitectureTest()
    {
    }


struct TestHashItem
    {

    TestHashItem
        (
        int v
        )
        {
        int i;
        for( i = 0; i < 10; i++ )
            {
            array[ i ] = v;
            }
        }

    int array[10];
    };


void ArchitectureTest::test()
    {
    typedef ext::hash_map
                <
                int,
                TestHashItem
                > TestHash;

    typedef std::multimap
                <
                int,
                int
                > TestMultiMap;

    bool                b;
    Int64               i64;
    UInt64              u64;
    int                 count = 100 * 1000;
    int                 i;
    int                 v = count / 2;
    TestHash            h;
    TestHash::iterator  it;

    /* Type size test. */
    test_condition_( sizeof( Int8 )    == 1 );
    test_condition_( sizeof( Int16 )   == 2 );
    test_condition_( sizeof( Int32 )   == 4 );
    test_condition_( sizeof( Int64 )   == 8 );
    test_condition_( sizeof( UInt8 )   == 1 );
    test_condition_( sizeof( UInt16 )  == 2 );
    test_condition_( sizeof( UInt32 )  == 4 );
    test_condition_( sizeof( UInt64 )  == 8 );
    test_condition_( sizeof( Float32 ) == 4 );
    test_condition_( sizeof( Float64 ) == 8 );

    i64 =  INT64( 0x00000f0f00000000 );
    u64 = UINT64( 0x00000f0f00000000 );

    test_condition_( ( i64 >> 32 ) == 0xf0f );
    test_condition_( ( u64 >> 32 ) == 0xf0f );

    /* Bit boolean test. */
    bit_bool = true;
    b = bit_bool;
    test_condition_( bit_bool == 1 );
    test_condition_( b );
    bit_bool = false;
    b = bit_bool;
    test_condition_( bit_bool == 0 );
    test_condition_( !b );

    // Integer wrapping test.
        {
        UInt64  big = ( UINT64( 1 ) << 50 );
        struct
            {
            UInt64  value : 50;
            } number;

        test_condition_( ( big >> 50 ) == UINT64( 1 ) );

        number.value = big;
        test_condition_( number.value == UINT64( 0 ) );

        big = number.value;
        test_condition_( big == UINT64( 0 ) );

        big = ( UINT64( 1 ) << 50 ) - 1;
        number.value = big;
        number.value++;
        test_condition_( number.value == UINT64( 0 ) );

        big = number.value;
        test_condition_( big == UINT64( 0 ) );
        }

    /* ext::hash_map test. Tests if iterator into hash_map stays
     * valid during insertion and deletion in/from the hash_map.
     */
    it = h.insert( TestHash::value_type( v, v ) ).first;
    test_condition_( it != h.end() );

    for( i = 0; i < count; i++ )
        {
        h.insert( TestHash::value_type( i, TestHashItem( i ) ) );
        }
    test_condition_( it == h.find( v ) );

    for( i = 0; i < count; i++ )
        {
        hope( h.find( i ) != h.end() );
        if( i != v )
            {
            h.erase( h.find( i ) );
            }
        }
    test_condition_( it == h.find( v ) );
    test_condition_( it->second.array[0] == v );

    /* std::multimap test. Tests if order of elements
     * with same key is same as their insertion-order.
     */
        {
        TestMultiMap    m;
        int             keys[ 9 ]  = { 0, 0, 0, 2, 2, 2, 1, 1, 1 };
        int             items[ 9 ] = { 1, 2, 3, 4, 5, 6, 7, 8, 9 }; // Insertion order.
        int             pop_order[ 9 ] = { 1, 2, 3, 7, 8, 9, 4, 5, 6 };
        int             i;

        // Insert all items in the map.
        for( i = 0; i < 9; i++ )
            {
            m.insert( TestMultiMap::value_type( keys[ i ], items[ i ] ) );
            }

        // Test insertion order.
        for( i = 0; i < 3; i++ ) // For each key...
            {
            std::pair
                <
                TestMultiMap::const_iterator,
                TestMultiMap::const_iterator
                > range = m.equal_range( i );
            int   start[3] = { 0, 6, 3 };
            int   count = 0;
            hope( range.first != range.second );
            for( ; range.first != range.second; range.first++, count++ )
                {
                hope( count < 3 );
                test_condition_(  range.first->second == items[ start[ i ] + count ] );
                }
            test_condition_( count == 3 );
            }

        // Remove all items.
        for( i = 0; i < 9; i++ )
            {
            // Pop minimum.
            hope( m.begin() != m.end() );
            test_condition_( m.begin()->second == pop_order[ i ] );
            m.erase( m.begin() );
            }
        }
    }


}; // namespace Massiv::Core::Test
}; // namespace Massiv::Core
}; // namespace Massiv
