/** @file test_bitiostream.cpp
 * Bitstream testing.
 */

/*
 * begin        : Tue Nov 27 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_bitiostream.h"
#include "net/bitostream.h"
#include "net/bitistream.h"
#include <string>

namespace Massiv {
namespace Core {
namespace Test {


BitIOStreamTest::BitIOStreamTest() :
    Test( "bitiostreams" )
    {
    }


BitIOStreamTest::~BitIOStreamTest()
    {
    }


void BitIOStreamTest::test()
    {
    test_a();
    test_b();
    }


void BitIOStreamTest::test_a()
    {
    std::string         s = "cau";
    UInt32              i32 = 1;
    UInt64              i64 = UINT64( 0xfecba98765432100 );
    Float32             f32 = 0.12345f;
    Float64             f64 = 0.123456789;
    int                 bitlen1;
    int                 bitlen2;

    BitIOStreamBuffer   buffer1;
    BitIOStreamBuffer   buffer2;
    BitOStream          bos1( 1000 );
    BitOStream          bos2;
    BitIStream          bis1;

    bitlen1 = 0;
    test_condition_( bos1.bit_size() == 0 );
    test_condition_( bos1.byte_size() == 0 );

    buffer1 = bos1.get_buffer();

    // write to bos1
    bos1.write_int( 1, 1 );
    test_condition_( bos1.bit_size() == 1 );
    bos1.align_to_byte();
    test_condition_( bos1.bit_size() == 8 );
    bos1.align_fill( 0, 4 );
    test_condition_( bos1.byte_size() == 5 );

    test_condition_(  buffer1[1] == 0 );
    test_condition_(  buffer1[4] == 0 );
    bos1.align_fill( 1, 4 );
    test_condition_( bos1.byte_size() == 9 );
    test_condition_(  buffer1[6] == 1 );
    test_condition_(  buffer1[7] == 1 );

    // write integer
    bos1.write_int( i32, 1 );   // 1 bit
    test_condition_( bos1.bit_size() == bitlen1 + 9 * 8 + 1 );
    test_condition_( bos1.byte_size() == 1 + 9);
    bitlen1 = bos1.bit_size();

    // write string
    bos1.write_string( s );     // 4 * 7 = 28 bits
    test_condition_( bos1.bit_size() == bitlen1 + 4 * 7 );
    test_condition_( bos1.byte_size() == ( bitlen1 + 4 * 7 - 1 ) / 8 + 1 );
    test_condition_( buffer1 == bos1.get_buffer() );
    test_condition_( buffer1[8] >= i32 );
    test_condition_( buffer1[ buffer1.size() - 1 ] == '\0' );
    bitlen1 = bos1.bit_size();

    // write to bos2
    bos2.bind( buffer1, true ); // Duplicate buffer.
    buffer2 = bos2.get_buffer();
    test_condition_( buffer1 == buffer2 );                   // conatins same data
    test_condition_( buffer1.size() == ( bitlen1 - 1 ) / 8 + 1 );
    bitlen2 = bos2.bit_size();
    test_condition_( bitlen2 == ( bos1.byte_size() * 8 ) );  // but bitlen1 != bitlen2

    // write int
    bos2.write_int64( i64, 8 ); // 8 bits
    test_condition_( bos2.bit_size() == bitlen2 + 8 );
    test_condition_( bos2.byte_size() == ( bitlen2 + 8 - 1 ) / 8 + 1 );
    test_condition_( buffer2[ buffer2.size() - 1 ] == ( i64 & 0xff ) );
    bitlen2 = bos2.bit_size();

    // write string
    bos2.write_string( s );     // 4 * 7 = 28 bits
    test_condition_( bos2.bit_size() == bitlen2 + 4 * 7 );
    test_condition_( bos2.byte_size() == ( bitlen2 + 4 * 7 - 1 ) / 8 + 1 );
    test_condition_( buffer2 == bos2.get_buffer() );
    test_condition_( buffer2[ buffer2.size() - 1 ] == '\0' );
    bitlen2 = bos2.bit_size();

    bos2.write_float( f32 );
    bos2.write_int( i32, 32 );
    bos2.write_float64( f64 );
    bos2.write_int64( i64, 64 );
    test_condition_( bos2.bit_size() == bitlen2 + 32 + 32 + 64 + 64 );
    bitlen2 = bos2.bit_size();

    test_condition_( buffer1 != buffer2 );
    bos1.reset();
    bos2.reset();
    test_condition_( buffer1.size() == ( bitlen1 - 1 ) / 8 + 1);
    test_condition_( buffer2.size() == ( bitlen2 - 1 ) / 8 + 1);

    // test BitIStream
    bis1.bind( buffer1 );
    test_condition_( buffer1 == bis1.get_buffer() );    // contains same data

    test_condition_( bis1.read_int( 1 ) == 1 );
    test_condition_( bis1.read_int( 7 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );


    test_condition_( bis1.read_int( 1 ) == ( i32 & 0x01 ) );
    test_condition_( bis1.read_string() == s );
    buffer1[0] = 0;
    test_condition_( bis1.get_buffer() == buffer1 );    // both buffer1 and bis1 modified

    bis1.unbind();
    test_condition_( bis1.get_buffer().size() == 0 );

    bis1.bind( buffer2, true ); // Duplicate buffer.
    test_condition_( bis1.get_buffer() == buffer2 );    // contains same data
    buffer2[0] = 2;
    test_condition_( bis1.get_buffer() != buffer2 );    // buffer2 modified but bos1 not

    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 32 ) == 0 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );
    test_condition_( bis1.read_int( 8 ) == 1 );

    test_condition_( bis1.read_int( 1 ) == ( i32 & 0x01 ) );
    test_condition_( bis1.read_string() == s );

    bis1.read_int( 32 - 29 );   // Skip hole.

    test_condition_( bis1.read_int64( 8 ) == ( i64 & 0xff ) );
    test_condition_( bis1.read_string() == s );
    test_condition_( bis1.read_float() == f32 );
    test_condition_( bis1.read_int( 32 ) == i32 );
    test_condition_( bis1.read_float64() == f64 );
    test_condition_( bis1.read_int64( 64 ) == i64 );
    }  // BitIOStreamTest::test_a()


void BitIOStreamTest::test_b()
    {

    #define TRY_OUT_OF_RANGE( command ) \
        try \
            { \
            command; \
            test_condition_( false ); \
            } \
        catch( ExOutOfRange & ) \
            { \
            test_condition_( true ); \
            }


    #define TRY_END_OF_STREAM( command ) \
        try \
            { \
            command; \
            test_condition_( false ); \
            } \
        catch( ExUnexpectedEof & ) \
            { \
            test_condition_( true ); \
            }


    BitIOStreamBuffer   buffer1;
    BitIOStreamBuffer   buffer2;
    
    BitOStream          bos( 1000 );
    
    unsigned int        pattern[] = { 76, 112, 240, 124, 124, 240, 112, 76 };
        /* binary pattern:
         *  pattern[ 0 ] = 01001100
         *  pattern[ 1 ] = 01110000
         *  pattern[ 2 ] = 11110000
         *  pattern[ 3 ] = 01111100
         *  pattern[ 4 ] = 01111100
         *  pattern[ 5 ] = 11110000
         *  pattern[ 6 ] = 01110000
         *  pattern[ 7 ] = 01001100
         *
         *  Actually, in the stream contents of each byte is turned over.
         *  Therefore, the stream will begin like this:
         *     00110010 00001110 00001111 00111110 etc.
         */

    const unsigned int  num_items = sizeof( pattern ) / sizeof( unsigned int );
    unsigned i;
    
    /* write pattern into buffer */
    for( i = 0; i < num_items; i++ )
        {
        bos.write_int( pattern[ i ], 8 );
        }
    
    /* check the pattern */
    BitIStream          bis( bos.get_buffer(), true, 0 );
    test_condition_( bis.get_buffer().size() == 8 ); // 8 bytes
    test_condition_( bis.get_position() == 0 );
    test_condition_( bis.count_remaining_bits() == 8 * 8 ); // 8 bytes, each 8 bits

    for( i = 0; i < num_items; i++ )
        {
        test_condition_( bis.read_int( 8 ) == pattern[ i ] );
        }

    /* test "read-after-end-of-stream" exception */
    TRY_END_OF_STREAM( bis.read_int( 8 ) );

    /* test of rewind() */
    bis.rewind();
    for( i = 0; i < num_items; i++ )
        {
        test_condition_( bis.read_int( 8 ) == pattern[ i ] );
        }

    /* test of seek() */
    bis.seek( 1, BitIOStream::WHENCE_START );
    test_condition_( bis.read_int( 7 ) == 38 );

    bis.seek( 11, BitIOStream::WHENCE_START );
    test_condition_( bis.read_int( 7 ) == 14 );

    bis.seek( -2, BitIOStream::WHENCE_CURRENT );
    test_condition_( bis.read_int( 8 ) == pattern[ 2 ] );

    bis.seek( 10, BitIOStream::WHENCE_CURRENT );
    test_condition_( bis.read_int( 3 ) == 7 );

    bis.seek( -31, BitIOStream::WHENCE_CURRENT );
    test_condition_( bis.read_int( 15 ) == 16833 );
    
    bis.seek( -6, BitIOStream::WHENCE_END );
    test_condition_( bis.read_int( 6 ) == 19 );

    /* test of skip() */
    bis.seek( 3, BitIOStream::WHENCE_START );
    test_condition_( bis.read_int( 4 ) == 9 );
    
    bis.skip( 11 );
    test_condition_( bis.read_int( 5 ) == 28 );

    /* skiping and seeking out of range */
    TRY_OUT_OF_RANGE( bis.skip( 42 ) );
    TRY_OUT_OF_RANGE( bis.skip( -24 ) );
    TRY_OUT_OF_RANGE( bis.seek( -1, BitIOStream::WHENCE_START ) );
    TRY_OUT_OF_RANGE( bis.seek( 1, BitIOStream::WHENCE_END ) );
    TRY_OUT_OF_RANGE( bis.seek( 65, BitIOStream::WHENCE_START ) );
    TRY_OUT_OF_RANGE( bis.seek( -65, BitIOStream::WHENCE_END ) );

    /* changing offset */
    test_condition_( bis.change_offset( 8 ) == 0 );
    bis.seek( 2, BitIOStream::WHENCE_START );
    test_condition_( bis.read_int( 11 ) == 1052 );

    TRY_OUT_OF_RANGE( bis.skip( -14 ) );

    /* changing stream end */
    test_condition_( bis.change_stream_end( 8 ) == 7 * 8 );
    bis.seek( 3, BitIOStream::WHENCE_START );

    TRY_END_OF_STREAM( bis.read_int( 6 ) );

    bis.seek( -5, BitIOStream::WHENCE_END );
    test_condition_( bis.read_int( 5 ) == 14 );

    /* head() */
    bis.change_stream_end( 7 * 8 ); 
    bis.seek( 13, BitIOStream::WHENCE_START );
    BitIStream bis1;
    bis1 = bis.head( false );

    test_condition_( bis1.data + 2 == bis1.end );
    test_condition_( *(bis1.end) == 124 );
    test_condition_( bis1.read_int( 6 ) == 48 );

    TRY_END_OF_STREAM( bis1.read_int( 11 ) );

    BitIStream bis2 = bis.head( true );
    test_condition_( bis2.data + 2 == bis2.end );
    test_condition_( bis2.read_int( 6 ) == 48 );
    test_condition_( bis2.read_int( 10 ) == 961 );

    TRY_END_OF_STREAM( bis2.read_int( 1 ) );

    /* tail() */
    bis.seek( 3, BitIOStream::WHENCE_START );
    BitIStream bis3 = bis.tail( false );

    for( i = 2; i < 8; i++ )
        {
        test_condition_( bis3.read_int( 8 ) == pattern[ i ] );
        }

    TRY_END_OF_STREAM( bis3.read_int( 1 ) );

    BitIStream bis4( bis.tail( true ) );

    for( i = 2; i < 8; i++ )
        {
        test_condition_( bis4.read_int( 8 ) == pattern[ i ] );
        }

    TRY_END_OF_STREAM( bis4.read_int( 1 ) );

    /* test of istream::align_to_byte(), get_position() 
       and count_remaining_bits() */
    bis.change_offset( 0 );
    bis.change_stream_end( 64 );
    bis.rewind();

    test_condition_( bis.get_position() == 0 );
    test_condition_( bis.count_remaining_bits() == 64 );
    bis.read_int( 8 );
    test_condition_( bis.get_position() == 8 );
    test_condition_( bis.count_remaining_bits() == 56 );
    bis.align_to_byte();
    test_condition_( bis.get_position() == 8 );
    bis.read_int( 3 );
    test_condition_( bis.get_position() == 11 );
    test_condition_( bis.count_remaining_bits() == 53 );
    bis.align_to_byte();
    test_condition_( bis.get_position() == 16 );
    test_condition_( bis.count_remaining_bits() == 48 );        

    /* test of ostream::write_bitistream() */
    BitOStream  bos2;
    bos2.write_int( 13, 4 );
    bis.rewind();
    bis.skip( 9 );
    bos2.write_bitistream( bis, true ); // byte-aligned
    
    BitIStream  bis5;
    bis5.bind( bos2.get_buffer(), false, 0 );

    test_condition_( bis5.read_int( 4 ) == 13 );
    bis5.skip( 4 );

    for( i = 0; i < 7; i++ )
        {
        test_condition_( bis5.read_int( 8 ) == pattern[ i + 1 ] );
        }

    TRY_END_OF_STREAM( bis5.read_int( 1 ) );

    bos2.reset( 100 );
    bos2.write_int( 13, 4 );
    bis.rewind();
    bis.skip( 4 );
    bos2.write_bitistream( bis, false ); // no byte-alignment

    bis5.bind( bos2.get_buffer(), true, 0 );

    test_condition_( bis5.read_int( 4 ) == 13 );
    
    for( i = 0; i < 7; i++ )
        {
        unsigned int t_value = ( pattern[ i ] & 0xF0 ) >> 4;
        t_value |= ( pattern[ i + 1 ] & 0x0F ) << 4;
        test_condition_( bis5.read_int( 8 ) == t_value );
        }

    test_condition_( bis5.read_int( 4 ) == 4 );
    TRY_END_OF_STREAM( bis5.read_int( 1 ) );

    #undef TRY_OUT_OF_RANGE
    #undef TRY_END_OF_STREAM

    }  // BitIOStreamTest::seek_test()


}; // namespace Massiv::Core::Test
}; // namespace Massiv::Core
}; // namespace Massiv
