/** @file test_bitiostream_speed.cpp
 * Bitstream speed testing.
 */

/*
 * begin        : Tue Feb Nov 18 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_bitiostream_speed.h"
#include "net/bitostream.h"
#include "net/bitistream.h"
#include "global/global.h"
#include <iostream>

namespace Massiv {
namespace Core {
namespace Test {


namespace {

const SystemTime SUBTEST_DURATION = 5.0;
const unsigned int CHUNK_SIZE = 20 * 1024;

} // anonymous namespace


BitIOStreamSpeedTest::BitIOStreamSpeedTest() :
    Super( "bitiostream_speed" ),
    bits_written( 0 ),
    subtest_duration( SUBTEST_DURATION )
    {
    }

    
BitIOStreamSpeedTest::~BitIOStreamSpeedTest()
    {
    }


void BitIOStreamSpeedTest::test()
    {
    test_uint32_bits_write( 1, CHUNK_SIZE );
    test_uint32_bits_write( 2, CHUNK_SIZE );
    test_uint32_bits_write( 8, CHUNK_SIZE );
    test_uint32_bits_write( 16, CHUNK_SIZE );
    test_uint32_bits_write( 32, CHUNK_SIZE );
    test_uint32_random_bits_write( CHUNK_SIZE );
    }
    
void BitIOStreamSpeedTest::test_uint32_bits_write
    (
    unsigned int bits,
    unsigned int chunk_bytes
    )
    {
    hope( bits > 0 && bits <= 32 );
    
        {
        std::cout << "Writing " << bits << " bit(s) of UInt32 to BitOStream "
            << "for at least " << subtest_duration.get_value() << " seconds."
            << std::endl
            << "Each dot represents " << chunk_bytes << " bytes written."
            << std::endl;
        }
        
        {
        bits_written = 0;
        start_time   = Global::system_time();
        while( ( Global::system_time() - start_time ) < subtest_duration )
            {
            const unsigned int iterations = ( 8 * chunk_bytes + bits - 1 ) / bits;

            /* Stoupik FIXME !!!
             * Even though I moved bitostream inside the while loop, this test
             * still allocates more and more memory (gcc-3.2, release build).
             * Is it bug in bitstreams, dynamic shared arrays, libstdc++ ?
             */
            BitOStream bos;
            
            for( unsigned int i = 0; i < iterations; i++ )
                {
                bos.write_int( bits_written, bits );
                }
            std::cout << '.';
            bits_written += iterations * bits;
            }
        stop_time = Global::system_time();
        std::cout << std::endl;
        
        print_statistic();
        }
    }


void BitIOStreamSpeedTest::test_uint32_random_bits_write
    (
    unsigned int chunk_bytes
    )
    {
    int bits;
    
        {
        std::cout << "Writing random bits of UInt32 to BitOStream "
            << "for at least " << subtest_duration.get_value() << " seconds."
            << std::endl
            << "Each dot represents " << chunk_bytes << " bytes written."
            << std::endl;
        }
        
        {
        BitOStream  bos;
        bits_written = 0;
        bits = 8;
        start_time   = Global::system_time();
        while( ( Global::system_time() - start_time ) < subtest_duration )
            {
            unsigned int start_bits = bits_written;
            while( bits_written - start_bits < chunk_bytes * 8 )
                {
                bos.write_int( bits_written, bits );
                bits_written += bits;
                bits = bits + 11;
                bits %= 33;
                }
            std::cout << '.';
            }
        stop_time = Global::system_time();
        
        hope( (unsigned int) bos.bit_size() == bits_written );
        std::cout << std::endl;
        print_statistic();
        }
    }
    
    
void BitIOStreamSpeedTest::print_statistic()
    {
    const double    duration = ( stop_time - start_time ).get_value();
    double          bytes_written = bits_written / 8.0;
    double          kbytes_written = bytes_written / 1024.0;
    
    std::cout << "Results: "
        << "Duration " << duration << " seconds, "
        << kbytes_written << " KBytes written, "
        << ( kbytes_written / duration ) << " KBytes per second."
        << std::endl;
    }
   

} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
