/** @file test_container.cpp
 * Testing of all container properties (PArray, PDictionary, PSet).
 */

/*
 * begin        : Thu Mar 11 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_container.h"
#include "property/prop_array.h"
#include "serializer/stype_int.h"
#include "property/prop_string.h"
#include "property/prop_dictionary.h"
#include "object/object_manager.h"
#include "replication/replica_server.h"
#include <cstdio>

using namespace Massiv::Core;


namespace Massiv {
namespace Core {
namespace Test {


ContainerTest::ContainerTest() :
    Test( "containers" )
    {
    }


ContainerTest::~ContainerTest()
    {
    }


void ContainerTest::test()
    {
    test_value_array();
    test_property_array();
    test_dictionary();
    test_set();
    }


void ContainerTest::test_value_array()
    {
    typedef PValueArray<SInt32>     ArrayType;
    typedef PArray<ArrayType>       Array2DType;

    // PValueArray<SInt32> test.

    ArrayType   vect;

    test_condition_( vect.size() == 0 );
    test_condition_( vect.empty() );

    vect.push_back( SInt32( 0 ) );
    vect.push_back( SInt32( 1 ) );
    vect.push_back( SInt32( 2 ) );
    vect.push_back( SInt32( 3 ) );
    test_condition_( vect.size() == 4 );
    test_condition_( !vect.empty() );

    ArrayType::Iterator it = vect.begin();
    test_condition_( *it == SInt32( 0 ) );
    it++;
    test_condition_( *it == SInt32( 1 ) );
    ArrayType::ReferenceType ir = *it;
    it++;
    test_condition_( *it == SInt32( 2 ) );
    it++;
    test_condition_( *it == SInt32( 3 ) );
    it++;
    test_condition_( it == vect.end() );
    ir = -1;
    test_condition_( vect[ 1 ] == SInt32( -1 ) );
    test_condition_( vect[ 3 ] == SInt32( 3 ) );
    it = vect.begin();
    it = it + 2;
    test_condition_( *it == SInt32( 2 ) );
    
    vect.pop_back();
    test_condition_( vect.size() == 3 );
    vect.clear();
    test_condition_( vect.size() == 0 );
    vect.resize( 300 );
    test_condition_( vect.size() == 300 );
    vect.push_back( SInt32( 0x0d13de4d ) );
    it = vect.begin() + 299;
    it++;
    test_condition_( it != vect.end() );
    test_condition_( *it == SInt32( 0x0d13de4d ) );
    it++;
    test_condition_( it == vect.end() );

    // PArray< PValueArray<SInt32> > test.

    Array2DType vect2d;
    const int   num_rows = 4;
    const int   num_columns = 4;
    int         i, j;

    vect2d.resize( num_rows );

    for( i = 0; i < num_rows; i++ )
        {
        vect2d[ i ].resize( num_columns );

        for( j = 0; j < num_columns; j++ )
            {
            vect2d[ i ]( j ) = i * j;
            }
        }

    for( i = 0; i < num_rows; i++ )
        {
        for( j = 0; j < num_columns; j++ )
            {
            test_condition_( vect2d[ i ][ j ] == i * j );
            }
        }
    }


void ContainerTest::test_property_array()
    {
    typedef PArray<PString>     ArrayType;
    typedef PArray<ArrayType>   Array2DType;

    ObjectManager & object_manager = Global::object_manager();
    ReplicaServer & replica_server = *object_manager.replica_server;
    ReplicaTime replica_time = replica_server.get_replica_time();

    // PArray<PString> test.

    ArrayType   vect;

    vect.push_back( PString( "one" ) );
    vect.push_back( PString( "two" ) );
    vect.push_back( PString( "three" ) );
    vect.push_back( PString( "four" ) );

    test_condition_( vect.size() == 4 );

    ArrayType::ConstIterator it = vect.begin() + 1;

    test_condition_( *it == PString( "two" ) );
    test_condition_( !vect.is_dirty( replica_time ) );

    replica_server.advance_replica_time();

    vect[ 1 ]( 0 ) = 'T';
    test_condition_( *it == PString( "Two" ) );
    test_condition_( vect.is_dirty( replica_time ) );

    // PArray< PArray<PString> > test.

    Array2DType vect2d;
    const int   num_rows = 4;
    const int   num_columns = 4;
    int         i, j;

    vect2d.resize( num_rows );

    for( i = 0; i < num_rows; i++ )
        {
        vect2d[ i ].resize( num_columns );

        for( j = 0; j < num_columns; j++ )
            {
            vect2d[ i ][ j ] = vect[ i ];
            vect2d[ i ][ j ] += " * ";
            vect2d[ i ][ j ] += vect[ j ];
            }
        }

    replica_server.advance_replica_time();
    replica_time = replica_server.get_replica_time();
    replica_server.advance_replica_time();

    test_condition_( !vect2d.is_dirty( replica_time ) );

    for( i = 0; i < num_rows; i += 2 )
        {
        vect2d[ i ][ i ] = "modified";
        vect2d[ i ][ i ]( 0 ) = 'M';

        test_condition_( vect2d[ i ][ i ] == PString( "Modified" ) );
        }

    test_condition_( vect2d.is_dirty( replica_time ) );

    for( i = 0; i < num_rows - 1; i += 2 )
        {
        test_condition_( vect2d[ i ].is_dirty( replica_time ) );
        test_condition_( !vect2d[ i + 1 ].is_dirty( replica_time ) );
        }
    }


void ContainerTest::test_dictionary()
    {
    typedef PDictionary
                <
                PString,
                PString,
                ext::hash<std::string>,
                std::equal_to<std::string>
                >   StringToStringMap;
    
    typedef PMultiDictionary
                <
                PString,
                PString,
                ext::hash<std::string>,
                std::equal_to<std::string>
                >   StringToStringMultiMap;
    
    ObjectManager &             object_manager = Global::object_manager();
    ReplicaServer &             replica_server = *object_manager.replica_server;
    ReplicaTime                 replica_time = replica_server.get_replica_time();
    
    // Dictionary test.

    StringToStringMap           map;
    StringToStringMap::Iterator m_it;

    m_it = map.begin();
    test_condition_( m_it == map.end() );

    map.insert( PString( "one" ),   PString( "1" ) );
    map.insert( PString( "two" ),   PString( "2" ) );
    map.insert( PString( "three" ), PString( "3" ) );
    map.insert( PString( "four" ),  PString( "4" ) );

    test_condition_( map.size() == 4 );

    m_it = map.begin();
    m_it++;
    m_it++;
    m_it++;
    m_it++;
    test_condition_( m_it == map.end() );

    test_condition_( map.look_up( PString( "one" ) ) == PString( "1" ) );
    test_condition_( map.look_up( PString( "two" ) ) == PString( "2" ) );
    test_condition_( map.look_up( PString( "three" ) ) == PString( "3" ) );
    test_condition_( map.look_up( PString( "four" ) ) == PString( "4" ) );

    map.erase( PString( "two" ) );
    test_condition_( map.size() == 3 );
    test_condition_( !map.contains( PString( "two" ) ) );
    
    map[ PString( "hello" ) ] = PString( "world" );
    test_condition_( map[ PString( "hello" ) ] == PString( "world" ) );

    replica_server.advance_replica_time();
    replica_time = replica_server.get_replica_time();
    replica_server.advance_replica_time();

    test_condition_( !map.is_dirty( replica_time ) );
    map.insert( PString( "one" ), PString( "one" ) );
    test_condition_( map.is_dirty( replica_time ) );

    replica_server.advance_replica_time();
    replica_time = replica_server.get_replica_time();
    replica_server.advance_replica_time();

    test_condition_( !map.is_dirty( replica_time ) );
    m_it = map.begin();
    m_it->second = PString( "???" );
    test_condition_( map.is_dirty( replica_time ) );

    // PMultiDictionary test.

    StringToStringMultiMap              multi_map;
    StringToStringMultiMap::KeyRange    mm_range;
    StringToStringMultiMap::Iterator    mm_it;
    int                                 count;

    multi_map.insert( PString( "one" ),   PString( "1" ) );
    multi_map.insert( PString( "two" ),   PString( "2" ) );
    multi_map.insert( PString( "three" ), PString( "3" ) );
    multi_map.insert( PString( "four" ),  PString( "4" ) );

    multi_map.insert( PString( "multi-key" ), PString( "first" ) );
    multi_map.insert( PString( "multi-key" ), PString( "second" ) );
    multi_map.insert( PString( "multi-key" ), PString( "third" ) );

    mm_range = multi_map.equal_range( PString( "multi-key" ) );
    
    for( mm_it = mm_range.first, count = 0; mm_it != mm_range.second; ++mm_it )
        {
        count++;
        }

    test_condition_( count == 3 );

    multi_map.erase( mm_range.first );
    mm_range = multi_map.equal_range( PString( "multi-key" ) );
    
    for( mm_it = mm_range.first, count = 0; mm_it != mm_range.second; ++mm_it )
        {
        count++;
        }

    test_condition_( count == 2 );
    
    mm_range = multi_map.equal_range( PString( "!!!" ) );
    
    for( mm_it = mm_range.first, count = 0; mm_it != mm_range.second; ++mm_it )
        {
        count++;
        }

    test_condition_( count == 0 );
    }


void ContainerTest::test_set()
    {
    typedef PSet
                <
                PString,
                ext::hash<std::string>,
                std::equal_to<std::string>
                >   StringSet;
    
    typedef PMultiSet
                <
                PString,
                ext::hash<std::string>,
                std::equal_to<std::string>
                >   StringMultiSet;

    // Since sets and dictionaries share common hash table implementation,
    // which has been tested by @c test_dictionary(), no extra tests are
    // needed.

    // Set test.

    StringSet           set;
    StringSet::Iterator it;

    set.insert( PString( "str1" ) );
    set.insert( PString( "str2" ) );
    set.insert( PString( "str3" ) );

    test_condition_( set.size() == 3 );

    set.erase( PString( "str2" ) );
    test_condition_( set.size() == 2 );

    it = set.find( PString( "str3" ) );
    test_condition_( it != set.end() );
    set.erase( it );

    test_condition_( set.size() == 1 );
    set.insert( PString( "str1" ) );
    test_condition_( set.size() == 1 );

    it = set.begin();
    test_condition_( it == set.find( PString( "str1" ) ) );
    test_condition_( ( *it )[ 0 ] == 's' );

    // Multi-set test.

    StringMultiSet              multi_set;
    StringMultiSet::KeyRange    ms_range;
    StringMultiSet::Iterator    ms_it;
    int                         count;

    multi_set.insert( PString( "hello" ) );
    multi_set.insert( PString( "1234" ) );
    multi_set.insert( PString( "1234" ) );
    multi_set.insert( PString( "1234" ) );
    multi_set.insert( PString( "!2!4" ) );

    ms_range = multi_set.equal_range( PString( "1234" ) );
    
    for( ms_it = ms_range.first, count = 0; ms_it != ms_range.second; ++ms_it )
        {
        count++;
        }

    test_condition_( count == 3 );
    
    ms_range = multi_set.equal_range( PString( "4321" ) );
    
    for( ms_it = ms_range.first, count = 0; ms_it != ms_range.second; ++ms_it )
        {
        count++;
        }

    test_condition_( count == 0 );
    }


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
