/** @file test_crypto.cpp
 * cryptography testing.
 */

/*
 * begin        : Tue Nov 27 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_crypto.h"
#include "net/massiv_blowfish.h"
#include "net/massiv_rsa.h"
#include <cstdio>
#include <iostream>
#include <vector>
#include <set>


namespace Massiv {
namespace Core {
namespace Test {


CryptoTest::CryptoTest() :
    Test( "cryptography" )
    {
    }


CryptoTest::~CryptoTest()
    {
    }


void CryptoTest::test()
    {
    std::string         message( "This is going to be crypted.1234" );
    std::string         crypted_string;
    std::string         decrypted_string;
    std::vector<UInt8>  crypted_vector;
    std::vector<UInt8>  decrypted_vector;

    // RSA test.

    RSACipher::Seed     rsa_seed( "SEED 12345 SEED 1234" );
    RSACipher           rsa( 1024, rsa_seed );

    // String test.

    rsa.encrypt( message, crypted_string);
    rsa.decrypt( crypted_string, decrypted_string );
    test_condition_( decrypted_string == message );

    RSACipher           rsa2( rsa.get_keys(), rsa_seed );

    rsa2.decrypt( crypted_string, decrypted_string );
    test_condition_( decrypted_string == message );

    // Range test.

    rsa.encrypt_range( message.begin() + 4, message.begin() + 4 + 8, crypted_string );
    rsa.decrypt( crypted_string, decrypted_string );
    test_condition_( decrypted_string == " is goin" );

    // Cross-container test.

    rsa.decrypt( crypted_string, decrypted_vector );
    test_condition_
        (
        decrypted_vector[0] == ' ' &&
        decrypted_vector[1] == 'i' &&
        decrypted_vector[2] == 's' &&
        decrypted_vector[3] == ' ' &&
        decrypted_vector[4] == 'g' &&
        decrypted_vector[5] == 'o' &&
        decrypted_vector[6] == 'i' &&
        decrypted_vector[7] == 'n'
        );

    // Sign-verify test.

    std::string         signature;

    rsa.sign( message, signature );
    test_condition_( rsa.verify( message, signature ) );

    // Blowfish test.

    BlowfishEncryptor   benc( "key" );
    BlowfishDecryptor   bdec( "key" );

    // String test.

    benc.align( message );
    benc.process( message, crypted_string );
    bdec.process( crypted_string, decrypted_string );
    test_condition_( decrypted_string == message );

    // Range test.

    crypted_string.resize( 8 );
    benc.process_range( message.begin() + 4, message.begin() + 4 + 8, crypted_string.begin() );
    bdec.process( crypted_string, decrypted_string );
    test_condition_( decrypted_string == " is goin" );

    // Cross-container test.

    bdec.process( crypted_string, decrypted_vector );
    test_condition_
        (
        decrypted_vector[0] == ' ' &&
        decrypted_vector[1] == 'i' &&
        decrypted_vector[2] == 's' &&
        decrypted_vector[3] == ' ' &&
        decrypted_vector[4] == 'g' &&
        decrypted_vector[5] == 'o' &&
        decrypted_vector[6] == 'i' &&
        decrypted_vector[7] == 'n'
        );
    RSACipher::Seed     rsa_seed2( message.c_str() );
    RSACipher           rsa3( 1024, rsa_seed2 );
    
    std::string rsa_public = rsa.get_public_key();
    std::string rsa3_public = rsa3.get_public_key();

    std::string crypted, decrypted2, signed_message;
   
    test_condition_( rsa2.get_public_key() == rsa_public );
    test_condition_( rsa_public != rsa3_public );

    rsa.encrypt( message, crypted, rsa3_public );
    test_condition_( message != crypted );
    
    rsa.sign( crypted, signed_message );
    
    test_condition_( rsa3.verify( crypted, signed_message, rsa_public ) );
    rsa3.decrypt( crypted, decrypted2 );
    test_condition_( decrypted2 == message );
    }

} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
