/** @file test_filesys.cpp
 * Filesystem testing.
 */

/*
 * begin        : Fri Sep 13 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "test/test_filesys.h"


namespace Massiv {
namespace Core {
namespace Test {


#define SOURCE_TREE_DIRECTORY   "../../../../../src/core/"

static const char * const files_to_test[] =
    {
    SOURCE_TREE_DIRECTORY "filesys/basic_volume.cpp",
    SOURCE_TREE_DIRECTORY "filesys/block_pool.cpp",
    SOURCE_TREE_DIRECTORY "filesys/compact_volume.cpp",
    SOURCE_TREE_DIRECTORY "filesys/compressed_stream.cpp",
    SOURCE_TREE_DIRECTORY "filesys/simple_block_pool.cpp",
    SOURCE_TREE_DIRECTORY "filesys/stream_pool.cpp",
    SOURCE_TREE_DIRECTORY "filesys/volume.cpp",
    SOURCE_TREE_DIRECTORY "filesys/basic_volume.h",
    SOURCE_TREE_DIRECTORY "filesys/block_pool.h",
    SOURCE_TREE_DIRECTORY "filesys/compact_volume.h",
    SOURCE_TREE_DIRECTORY "filesys/compressed_stream.h",
    SOURCE_TREE_DIRECTORY "filesys/simple_block_pool.h",
    SOURCE_TREE_DIRECTORY "filesys/stream_pool.h",
    SOURCE_TREE_DIRECTORY "filesys/volume.h",
    SOURCE_TREE_DIRECTORY "test/main.cpp",

    SOURCE_TREE_DIRECTORY "test/test.cpp",
    SOURCE_TREE_DIRECTORY "test/test_all.cpp",
    SOURCE_TREE_DIRECTORY "test/test_architecture.cpp",
    SOURCE_TREE_DIRECTORY "test/test_array.cpp",
    SOURCE_TREE_DIRECTORY "test/test_bitiostream.cpp",
    SOURCE_TREE_DIRECTORY "test/test_cache.cpp",
    SOURCE_TREE_DIRECTORY "test/test_container.cpp",
    SOURCE_TREE_DIRECTORY "test/test_crypto.cpp",
    SOURCE_TREE_DIRECTORY "test/test_filesys.cpp",
    SOURCE_TREE_DIRECTORY "test/test_heap.cpp",
    SOURCE_TREE_DIRECTORY "test/test_logger.cpp",
    SOURCE_TREE_DIRECTORY "test/test_objects.cpp",
    SOURCE_TREE_DIRECTORY "test/test_registry.cpp",
    SOURCE_TREE_DIRECTORY "test/test_serializer.cpp",
    SOURCE_TREE_DIRECTORY "test/test_sockets.cpp",
    SOURCE_TREE_DIRECTORY "test/test_union_find.cpp",
    SOURCE_TREE_DIRECTORY "test/test.h",
    SOURCE_TREE_DIRECTORY "test/test_all.h",
    SOURCE_TREE_DIRECTORY "test/test_architecture.h",
    SOURCE_TREE_DIRECTORY "test/test_array.h",
    SOURCE_TREE_DIRECTORY "test/test_bitiostream.h",
    SOURCE_TREE_DIRECTORY "test/test_cache.h",
    SOURCE_TREE_DIRECTORY "test/test_container.h",
    SOURCE_TREE_DIRECTORY "test/test_crypto.h",
    SOURCE_TREE_DIRECTORY "test/test_filesys.h",
    SOURCE_TREE_DIRECTORY "test/test_heap.h",
    SOURCE_TREE_DIRECTORY "test/test_logger.h",
    SOURCE_TREE_DIRECTORY "test/test_objects.h",
    SOURCE_TREE_DIRECTORY "test/test_registry.h",
    SOURCE_TREE_DIRECTORY "test/test_serializer.h",
    SOURCE_TREE_DIRECTORY "test/test_sockets.h",
    SOURCE_TREE_DIRECTORY "test/test_union_find.h",

    NULL
    };


template <class Base>
struct TestPool : public Base
    {
    typedef typename Base::BlockIndex   BlockIndex;

    TestPool
        (
        bool            restore,
        const char *    image
        )
        : stream
            (
            restore ? image : create_file( image ),
            std::ios_base::in | std::ios_base::out | std::ios_base::binary
            )
        {
        if( restore )
            restore_state();
        else
            format_pool();
        }

    ~TestPool()
        {
        close_pool();
        }

    virtual void load_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        )
        {
        BasicVolume::read_from_stream( stream, ptr, index * block_size, block_size );
        }

    virtual void save_block
        (
        BlockIndex  index,
        UInt8 *     ptr
        )
        {
        BasicVolume::write_to_stream( stream, ptr, index * block_size, block_size );
        }

    static const char * create_file
        (
        const char * file
        )
        {
        std::ofstream s( file );
        return file;
        }

    std::fstream    stream;
    };


typedef TestPool<BlockPool>     TestBlockPool;
typedef TestPool<StreamPool>    TestStreamPool;


FileSystemTest::FileSystemTest()
    : Test( "filesystem" )
    {
    }


FileSystemTest::~FileSystemTest()
    {
    }


void FileSystemTest::test()
    {
    test_cache();
    test_block_pool();
    test_stream_pool();
    test_compressed_stream();
    test_volume();
    test_compact_volume();
    }


void FileSystemTest::test_cache()
    {
    LRUCache::LRUItem               item[ 4 ];
    LRUCache                        cache( 2 );
    const LRUCache::LRUStatistics * stats;
    int                             i;

    // Initialize.

    for( i = 0; i < 4; i++ )
        cache.register_item( &item[ i ] );

    stats = &cache.get_statistics();

    test_condition_( stats->memory_size == 2 );
    test_condition_( stats->free_memory == 2 );
    test_condition_( stats->num_items == 4 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 0 );
    test_condition_( stats->num_cached_items == 0 );
    test_condition_( stats->num_cache_out_candidates == 0 );

    // Test 1.

    item[ 0 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 1 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 1 );
    test_condition_( stats->num_cached_items == 1 );
    test_condition_( stats->num_cache_out_candidates == 0 );

    // Test 2.

    item[ 0 ].unlock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 1 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 1 );
    test_condition_( stats->num_cached_items == 1 );
    test_condition_( stats->num_cache_out_candidates == 1 );

    // Test 3.

    item[ 1 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 2 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 1 );

    // Test 4.

    item[ 2 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 3 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 0 );

    // Test 5.

    item[ 1 ].unlock();
    item[ 2 ].unlock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 0 );
    test_condition_( stats->num_misses == 3 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 2 );

    // Test 6.

    item[ 1 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 1 );
    test_condition_( stats->num_misses == 3 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 1 );

    // Test 7.

    item[ 1 ].unlock();
    item[ 2 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 2 );
    test_condition_( stats->num_misses == 3 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 1 );

    // Test 8.

    item[ 3 ].lock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 2 );
    test_condition_( stats->num_misses == 4 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 0 );

    // Test 9.

    item[ 3 ].unlock();
    item[ 2 ].unlock();

    stats = &cache.get_statistics();

    test_condition_( stats->free_memory == 0 );
    test_condition_( stats->num_hits == 2 );
    test_condition_( stats->num_misses == 4 );
    test_condition_( stats->num_cached_items == 2 );
    test_condition_( stats->num_cache_out_candidates == 2 );

    // Finalize.

    cache.unregister_all();

    stats = &cache.get_statistics();

    test_condition_( stats->num_items == 0 );
    test_condition_( stats->free_memory == 2 );
    test_condition_( stats->num_hits == 2 );
    test_condition_( stats->num_misses == 4 );
    test_condition_( stats->num_cached_items == 0 );
    test_condition_( stats->num_cache_out_candidates == 0 );
    }


void FileSystemTest::test_block_pool()
    {
    TestBlockPool::BlockIndex   chain[ 3 ];
    TestBlockPool::BlockIndex   block;

    // Test.

        {
        TestBlockPool   pool( false, "test_block_pool.image" );

        test_condition_( pool.num_free_blocks == 0 );

        // Create/erase test.

        chain[ 0 ] = pool.create_chain( 10 );
        test_condition_( pool.num_free_blocks == 0 );

        pool.erase_chain( chain[ 0 ] );
        test_condition_( pool.num_free_blocks == 10 );

        chain[ 0 ] = pool.create_chain( 3 );
        test_condition_( pool.num_free_blocks == 7 );

        chain[ 1 ] = pool.create_chain( 2 );
        test_condition_( pool.num_free_blocks == 5 );

        chain[ 2 ] = pool.create_chain( 10 );
        test_condition_( pool.num_free_blocks == 0 );

        pool.erase_chain( chain[ 1 ] );
        test_condition_( pool.num_free_blocks == 2 );

        pool.erase_chain( chain[ 0 ] );
        test_condition_( pool.num_free_blocks == 5 );

        pool.erase_chain( chain[ 2 ] );
        test_condition_( pool.num_free_blocks == 15 );

        // Truncate/append test.

        chain[ 0 ] = pool.create_chain( 20 );
        test_condition_( pool.num_free_blocks == 0 );

        pool.append_blocks( chain[ 0 ], 10 );
        test_condition_( pool.num_free_blocks == 0 );

        pool.erase_chain( chain[ 0 ] );
        test_condition_( pool.num_free_blocks == 20 + 10 );

        chain[ 0 ] = pool.create_chain( 50 );
        test_condition_( pool.num_free_blocks == 0 );

        block = pool.get_block( chain[ 0 ], 49 );

        test_condition_( block != TestBlockPool::INVALID_BLOCK );
        test_condition_( pool.get_next_block( block ) == TestBlockPool::INVALID_BLOCK );

        block = pool.get_block( chain[ 0 ], 10 );
        pool.erase_chain( block ); // Truncate chain[ 0 ].
        test_condition_( pool.num_free_blocks == 40 );

        pool.erase_chain( chain[ 0 ] );
        test_condition_( pool.num_free_blocks == 50 );

        // Create test.

        chain[ 0 ] = pool.create_chain( 3 );
        test_condition_( pool.num_free_blocks == 47 );

        chain[ 1 ] = pool.create_chain( 2 );
        test_condition_( pool.num_free_blocks == 45 );

        chain[ 2 ] = pool.create_chain( 10 );
        test_condition_( pool.num_free_blocks == 35 );

        pool.close_pool();
        }

    // Restore test.

        {
        TestBlockPool   pool( true, "test_block_pool.image" );

        test_condition_( pool.get_chain_length( chain[ 0 ] ) == 3 );
        test_condition_( pool.get_chain_length( chain[ 1 ] ) == 2 );
        test_condition_( pool.get_chain_length( chain[ 2 ] ) == 10 );
        
        pool.close_pool();
        }
    }


void FileSystemTest::test_stream_pool()
    {
    TestStreamPool::StreamDescriptor    descriptor[ 3 ];

    // Test.

        {
        TestStreamPool  pool( false, "test_stream_pool.image" );

            // Create streams.

            {
            pool.create_stream( descriptor[ 0 ] );
            pool.create_stream( descriptor[ 1 ] );
            pool.create_stream( descriptor[ 2 ] );
            }

            // Create & write test.

            {
            TestStreamPool::OStream stream1( descriptor[ 0 ], std::ios_base::out );
            int                     i;

            for( i = 0; i < 10; i++ )
                {
                stream1 << "hello!" << std::endl;
                test_condition_( stream1.good() );
                }
            }

            // Copy test.

            {
            TestStreamPool::OStream stream2( descriptor[ 1 ], std::ios_base::out );
            TestStreamPool::IStream stream1( descriptor[ 0 ], std::ios_base::in );

            stream2 << stream1.rdbuf();
            test_condition_( stream1.good() );
            test_condition_( stream2.good() );
            }

            // Read, write & seek test.

            {
            TestStreamPool::IOStream stream3( descriptor[ 2 ], std::ios_base::in | std::ios_base::out );
            std::string s;

            stream3 << "hello!";
            test_condition_( stream3.good() );

            stream3 >> s;
            test_condition_( s == "hello!" );
            test_condition_( stream3.rdstate() == std::ios_base::eofbit );
            stream3.clear();

            stream3.seekg( 0, std::ios_base::beg );
            test_condition_( stream3.good() );
            stream3 >> s;
            test_condition_( s == "hello!" );
            test_condition_( stream3.rdstate() == std::ios_base::eofbit );
            stream3.clear();

            stream3.seekp( 1, std::ios_base::beg );
            test_condition_( stream3.good() );
            stream3 << "hello!";
            test_condition_( stream3.good() );

            stream3.seekg( 0, std::ios_base::beg );
            test_condition_( stream3.good() );
            stream3 >> s;
            test_condition_( s == "hhello!" );
            test_condition_( stream3.rdstate() == std::ios_base::eofbit );
            stream3.clear();
            }

        pool.close_pool();
        }
    }


void FileSystemTest::test_compressed_stream()
    {
    // Compress.

        {
        std::ofstream raw_output( "test_compressed.gzip", std::ios_base::out | std::ios_base::binary );
        CompressedStream::OStream   compressor( raw_output );

        compressor << "This_is_a_test1!" << std::endl;
        compressor << "This_is_a_test2!" << std::endl;
        compressor << "This_is_a_test3!" << std::endl;
        compressor << "This_is_a_test4!" << std::endl;
        }

    // Decompress.

        {
        std::ifstream raw_input( "test_compressed.gzip", std::ios_base::in | std::ios_base::binary );
        CompressedStream::IStream   decompressor( raw_input );
        std::string                 decompressed;

        decompressor >> decompressed;
        test_condition_( decompressed == "This_is_a_test1!" );

        decompressor >> decompressed;
        test_condition_( decompressed == "This_is_a_test2!" );

        decompressor >> decompressed;
        test_condition_( decompressed == "This_is_a_test3!" );

        decompressor >> decompressed;
        test_condition_( decompressed == "This_is_a_test4!" );
        }
    }


void FileSystemTest::copy_to_volume
    (
    BasicVolume &       volume,
    const std::string & file_name,
    int                 creation_flags
    )
    {
    std::ifstream           input( file_name.c_str(), std::ios_base::in | std::ios_base::binary );
    BasicVolume::OFStream   output
        (
        volume,
        file_name,
        std::ios_base::out | std::ios_base::binary,
        creation_flags
        );

    test_condition_( input.good() );
    test_condition_( output.good() );

    output << input.rdbuf();
    test_condition_( output.good() );
    }


void FileSystemTest::verify
    (
    BasicVolume &       volume,
    const std::string & file_name
    )
    {
    std::ifstream           input1( file_name.c_str(), std::ios_base::in | std::ios_base::binary );
    BasicVolume::IFStream   input2( volume, file_name, std::ios_base::in | std::ios_base::binary );
    bool                    difference;

    test_condition_( input1.good() );
    test_condition_( input2.good() );

    difference = false;

    // FIXME: optimize this

    while( input1.good() && input2.good() )
        {
        char    c1;
        char    c2;

        input1.read( &c1, sizeof( char ) );
        input2.read( &c2, sizeof( char ) );

        difference |= ( c1 != c2 );
        }

    test_condition_( flags_on( input1.rdstate(), std::ios_base::eofbit ) );
    test_condition_( flags_on( input2.rdstate(), std::ios_base::eofbit ) );
    test_condition_( !difference );
    }


void FileSystemTest::copy_to_volume
    (
    BasicVolume &   volume,
    int             creation_flags
    )
    {
    const char * const * files;

    for( files = files_to_test; *files != NULL; files++ )
        {
        copy_to_volume( volume, *files, creation_flags );
        }
    }


void FileSystemTest::verify
    (
    BasicVolume &   volume
    )
    {
    const char * const * files;

    for( files = files_to_test; *files != NULL; files++ )
        {
        verify( volume, *files );
        }
    }


void FileSystemTest::test_volume()
    {
    // Format test.

        {
        Volume  volume1( "test_volume1.image", "VOLUME_LABEL1" );
        copy_to_volume( volume1, 0 );

        Volume  volume2( "test_volume2.image", "VOLUME_LABEL2" );
        copy_to_volume( volume2, BasicVolume::STREAM_COMPRESSED );
        }

    // Restore test.

        {
        Volume  volume1( "test_volume1.image", std::ios_base::in );
        verify( volume1 );

        Volume  volume2( "test_volume2.image", std::ios_base::in );
        verify( volume2 );
        }
    }


void FileSystemTest::test_compact_volume()
    {
    // Format test.

        {
        CompactVolume   volume1( "test_compact_volume1.image", "VOLUME_LABEL1" );
        copy_to_volume( volume1, 0 );

        CompactVolume   volume2( "test_compact_volume2.image", "VOLUME_LABEL2" );
        copy_to_volume( volume2, BasicVolume::STREAM_COMPRESSED );
        }

    // Restore test.

        {
        CompactVolume   volume1( "test_compact_volume1.image", std::ios_base::in );
        verify( volume1 );

        CompactVolume   volume2( "test_compact_volume2.image", std::ios_base::in );
        verify( volume2 );
        }
    }


#undef SOURCE_TREE_DIRECTORY


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
