/** @file test_fs_enumeration.cpp
 * Test of interface for enumeration of files and directories in filesystem.
 */

/*
 * begin        : Sun Feb 02 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "filesys/file_system.h"
#include "test/test_fs_enumeration.h"
#include "global/global.h"
#include <iostream>
#include <fstream>
#include <string>

namespace Massiv {
namespace Core {
namespace Test {


FsEnumerationTest::FsEnumerationTest()
    :   Test( "file system enumeration"),
        native_fs( Global::native_file_system() )
    {
    }  // FsEnumerationTest::FsEnumerationTest()


FsEnumerationTest::~FsEnumerationTest()
    {
    }  // FsEnumerationTest::~FsEnumerationTest()


void FsEnumerationTest::test()
    {
    init();

    try
        {
        go_through_test();
        descent_ascent_test();
        enter_leave_test();
        is_in_root_test();
        directory_manipulation_test();
        deinit();
        }
    catch( ... )
        {
        // Set the current directory to the old one.
        
        native_fs.set_current_directory( old_curr_dir );
        throw;
        }
    }  // FsEnumerationTest::test()


void FsEnumerationTest::init()
    {
    old_curr_dir = native_fs.get_current_directory();
    
    try
        {
        std::string ned( "non-existing-directory" );
        native_fs.set_current_directory( ned );
        test_condition_( false );
        }
    catch( ExCanNotChangeDirectory )
        {
        test_condition_( true );
        }
    catch( ... )
        {
        test_condition_( false );
        }

    std::string     path( "../..\\../../../src/core/test/test_hierarchy" );
    native_fs.set_current_directory( path );
    test_dir = native_fs.get_current_directory();
    }  // FsEnumerationTest::init()


void FsEnumerationTest::deinit()
    {
    native_fs.set_current_directory( old_curr_dir );
    test_condition_( old_curr_dir == native_fs.get_current_directory() );
    }  // FsEnumerationTest::deinit()


void FsEnumerationTest::go_through_test()
    {
    FileIterator it = get_iterator_for_current( std::string( "t*" ) );
    FileIterator it_aux = get_iterator_for_current( std::string( "t*" ) );

    test_condition_( it == it_aux );

    perform_go_through_test( it );
    test_condition_( it != it_aux );
    it.rewind();
    test_condition_( it == it_aux );

    perform_go_through_test( it );
    test_condition_( it != it_aux );
    it_aux.unbind();
    test_condition_( it == it_aux );    // Both the iterators are invalid now.
    }  // FsEnumerationTest::go_through_test()



void FsEnumerationTest::perform_go_through_test
    (
    FileIterator it
    )
    {
    bool    flags[ 3 ];
    bool    success = true;
    int     i;

    for( i = 0; i < 3; i++ )    flags[ i ] = false;

    test_condition_( !it.is_in_root() );

    i = 0;
    while( it.is_valid() )
        {
        if( ( it.name() == "tdir1" ) && !flags[ i ] )
            {
            flags[ i++ ] = true;
            test_condition_( it.is_directory() );
            test_condition_( !it.is_file() );
            test_condition_( it.size() == 0 );
            }
        else
        if( ( it.name() == "tdir2" ) && !flags[ i ] )  
            {
            flags[ i++ ] = true;
            test_condition_( it.is_directory() );
            test_condition_( !it.is_file() );
            }
        else
        if( ( it.name() == "tfile1.txt" ) && !flags[ i ] )  
            {
            flags[ i++ ] = true;
            test_condition_( it.is_file() );
            test_condition_( !it.is_directory() );
            test_condition_( it.size() == 1 );
            }
        else
            {
            success = false;
            break;
            }
        
        it++;
        }

    if( success )
        {
        for( i = 0; i < 3; i++ )
            {
            if( flags[ i ] == false )
                {
                success = false;
                break;
                }
            }
        }

    if( success )
        {
        test_condition_( true );
        }
    else
        {
        test_condition_( false );
        }
    }  // FsEnumerationTest::perform_go_through_test()


FileIterator FsEnumerationTest::get_iterator_for_current
    (
    std::string     mask
    )
    {
    return native_fs.open_current_directory( mask );
    }  // FsEnumerationTest::get_iterator_for_current()


void FsEnumerationTest::descent_ascent_test()
    {
    FileIterator it = get_iterator_for_current( std::string( "t*" ) );
    test_condition_( it.is_valid() );

    while( it.is_valid() && ( it.name() != "tdir2" ) )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( !it.is_directory() );
    test_condition_( it.is_file() );
    test_condition_( it.size() == 3 );
    test_condition_( it.name() == "tfile211.txt" );

    it.ascent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    it.ascent();

    perform_go_through_test( it );
    it.rewind();

    while( it.is_valid() && !it.is_file() )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    try
        {
        it.descent();
        test_condition_( 0 );
        }
    catch( ExNotADirectory )
        {
        test_condition_( 1 );
        }
    catch( ... )
        {
        test_condition_( 0 );
        }

    try
        {
        it.enter();
        test_condition_( 0 );
        }
    catch( ExNotADirectory )
        {
        test_condition_( 1 );
        }
    catch( ... )
        {
        test_condition_( 0 );
        }
    }  // FsEnumerationTest::descent_ascent_test()


void FsEnumerationTest::enter_leave_test()
    {
    FileIterator it = get_iterator_for_current( std::string( "t*" ) );
    test_condition_( it.is_valid() );

    while( it.is_valid() && ( it.name() != "tdir2" ) )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    FileIterator it_enter = it.enter();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    test_condition_( it_enter.is_valid() );
    test_condition_( !it_enter.is_directory() );
    test_condition_( it_enter.is_file() );
    test_condition_( it_enter.size() == 3 );
    test_condition_( it_enter.name() == "tfile211.txt" );

    FileIterator it_leave = it_enter.leave();

    test_condition_( it_enter.is_valid() );
    test_condition_( !it_enter.is_directory() );
    test_condition_( it_enter.is_file() );
    test_condition_( it_enter.size() == 3 );
    test_condition_( it_enter.name() == "tfile211.txt" );

    test_condition_( it_leave.is_valid() );
    test_condition_( it_leave.is_directory() );
    test_condition_( !it_leave.is_file() );
    test_condition_( it_leave.size() == 0 );
    test_condition_( it_leave.name() == "tdir21" );
    }  // FsEnumerationTest::enter_leave_test()


void FsEnumerationTest::is_in_root_test()
    {
    std::string     old_current_dir = native_fs.get_current_directory();
    FileIterator    it = get_iterator_for_current( std::string( "*" ) );

    std::string     old_dir = "";
    std::string     new_dir = it.get_absolute_directory();

    while( old_dir != new_dir )
        {
        it.ascent();
        old_dir = new_dir;
        new_dir = it.get_absolute_directory();
        }

    test_condition_( it.is_in_root() );

    native_fs.set_current_directory( old_current_dir );
    }  // FsEnumerationTest::is_in_root_test()


void FsEnumerationTest::directory_manipulation_test()
    {
    native_fs.set_current_directory( test_dir );
    
    native_fs.create_directory( "a/bb\\ccc" );
    FileIterator it( native_fs.open_current_directory( "a" ) );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a", "bb" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb", "ccc" );
    test_condition_( it.is_valid() );

    std::ofstream fstream( "a/bb/file", std::ios_base::out );
    if( !fstream )
        {
        std::cout << "Cannot create a/bb/file." << std::endl;
        native_fs.delete_directory( "a", FileSystem::DELETE_ALL_RECURSIVE );
        return;
        }
    fstream.close();

    native_fs.create_directory( "a/bb/ccc/dddd/eeeee" );
    it = native_fs.open_directory( "a/bb/ccc", "dddd" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb/ccc/dddd", "eeeee" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb", "file" );
    test_condition_( it.is_valid() );

    try
        {
        native_fs.delete_directory( "a", FileSystem::DELETE_EMPTY );
        test_condition_( false );
        }
    catch( ExDirectoryNotEmpty )
        {
        test_condition_( true );
        }
    catch( ExCanNotDeleteDirectory )
        {
        test_condition_( true );
        }
    catch( ... )
        {
        test_condition_( false );
        }

    it = native_fs.open_current_directory( "a" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a", "bb" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb", "ccc" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb/ccc", "dddd" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb/ccc/dddd", "eeeee" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb", "file" );
    test_condition_( it.is_valid() );

    try
        {
        native_fs.delete_directory( "a", FileSystem::DELETE_DIRS_RECURSIVE );
        test_condition_( false );
        }
    catch( ExDirectoryNotEmpty )
        {
        test_condition_( true );
        }
    catch( ExCanNotDeleteDirectory )
        {
        test_condition_( true );
        }
    catch( ... )
        {
        test_condition_( false );
        }

    it = native_fs.open_current_directory( "a" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a", "bb" );
    test_condition_( it.is_valid() );
    it = native_fs.open_directory( "a/bb", "ccc" );
    test_condition_( !it.is_valid() );
    
    #define EX_DIR_NOT_FOUND( command ) \
        try  \
            {  \
            command;  \
            test_condition_( false );  \
            }  \
        catch( ExDirectoryNotFound )  \
            {  \
            test_condition_( true );  \
            }
    
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb/ccc", "dddd" ) );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb/ccc/dddd", "eeeee" ) );
    it = native_fs.open_directory( "a/bb", "file" );
    test_condition_( it.is_valid() );

    it.unbind();
    native_fs.delete_directory( "a", FileSystem::DELETE_ALL_RECURSIVE );

    it = native_fs.open_current_directory( "a" );
    test_condition_( !it.is_valid() );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a", "bb" ) );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb", "ccc" ) );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb/ccc", "dddd" ) );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb/ccc/dddd", "eeeee" ) );
    EX_DIR_NOT_FOUND( native_fs.open_directory( "a/bb", "file" ) );

    #undef TEST_EX_DIRECTORY_NOT_FOUND
    }  // FsEnumerationTest::directory_manipulation_test()


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
