/** @file test_logger.cpp
 * Logger testing.
 */

/*
 * begin        : Tue Nov 27 2001
 * copyright    : (C) 2001 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "test/test_logger.h"
#include "global/global.h"
#include "logger/logdest.h"
#include "logger/logger.h"
#include <iostream>


namespace Massiv {
namespace Core {
namespace Test {


/** 
 * Testing log destination.
 */
class LogDestinationTest : public LogDestination
    {
public:

    /**
     * Constructor.
     */
    LogDestinationTest() :
            tests( 0 ),
            tsets( 0 )
        {
        }

    /**
     * Destructor.
     */
    ~LogDestinationTest()
        {
        }

public:

    int tests;
        /**< Number of "test" messages received. */
    int tsets;
        /**< Number of "tset" messages received. */

    /**
     * Reset @a tests and @a tsets to zero.
     */
    void reset()
        {
        tests = tsets = 0;
        }

protected:

    /**
     * Get short name describing destination.
     * Returns "test".
     */
    virtual std::string get_name() const
        {
        return "test";
        }

    /**
     * Get short description of destination.
     */
    virtual std::string get_description() const
        {
        return "Testing destination.";
        }

    /**
     * Match destination name.
     * Matches "test" (-\>success) and "tset" (-\>failure).
     */
    virtual bool match_name
        (
        const std::string & name
        )
        {
        if( name == "test" || name == "tset" )
            {
            return true;
            }
        return false;
        }

    /**
     * Log message.
     */
    virtual void log
        (
        const std::string & name,
        const LogMessage &
        )
        {
        if( name == "test" )
            tests++;
        else if( name == "tset" )
            tsets++;
        else
            throw "Unsupported destination name.";
        }

    /**
     * Flush destination.
     */
    virtual void flush()
        {
        }

    /**
     * Flush destination hard.
     */
    virtual void flush_hard()
        {
        }

    }; // class LogDestinationTest


LoggerTest::LoggerTest() :
        Test( "logger" )
    {
    }


LoggerTest::~LoggerTest()
    {
    }


void LoggerTest::test()
    {
    using namespace Massiv::Core;

    std::string                 str;
    Logger::RuleIterator        ri1, ri2;
    LogDestinationTest *        logdest = new LogDestinationTest;

    // Selector: *.*.>=lowest
    Logger::SelectorElement     se_all
        ( Logger::SelectorElement::PRIORITY_GEQ,
          Status::FACILITY_ANY, Status::TYPE_ANY, Status::PRIORITY_LOWEST );
    Logger::Selector            s_all;
    s_all.insert( se_all );

    // Selector: test.>=warning.>=low, *.*.<=high
    Logger::SelectorElement     se_adv1a
        ( Logger::SelectorElement::TYPE_GEQ | Logger::SelectorElement::PRIORITY_GEQ,
          Status::FACILITY_TEST, Status::TYPE_WARNING, Status::PRIORITY_LOW );
    Logger::SelectorElement     se_adv1b
        ( Logger::SelectorElement::PRIORITY_LEQ,
          Status::FACILITY_ANY, Status::TYPE_ANY, Status::PRIORITY_HIGH );
    Logger::Selector            s_adv1;
    s_adv1.insert( se_adv1a );
    s_adv1.insert( se_adv1b );

    // Selector: !logger.*.=lowest, !test,=error,<=highest
    Logger::SelectorElement     se_adv2a
        ( Logger::SelectorElement::NEGATE | Logger::SelectorElement::PRIORITY_EQ,
          Status::FACILITY_LOGGER, Status::TYPE_ANY, Status::PRIORITY_LOWEST );
    Logger::SelectorElement     se_adv2b
        ( Logger::SelectorElement::NEGATE | Logger::SelectorElement::TYPE_EQ |
          Logger::SelectorElement::PRIORITY_LEQ,
          Status::FACILITY_TEST, Status::TYPE_ERROR, Status::PRIORITY_HIGHEST );
    Logger::Selector            s_adv2;
    s_adv2.insert( se_adv2a );
    s_adv2.insert( se_adv2b );

    // Message
    LogMessage                  message
        ( STATUS_DEBUG( Status::FACILITY_TEST, Status::PRIORITY_LOWEST ), "msg" );

    // Init logger
    Logger & logger = ( Global::logger() );
    logger.clear_rules();

    // Simple insert works?
    test_condition_( ( ri1 = logger.insert_rule( s_all ) ) != logger.end_rule() );
    if( ri1 == logger.end_rule() )
        return;
    ri2 = ri1;
    test_condition_( ri1 == ri2 );
    ++ri2;
    test_condition_( ri1 != ri2 );
    test_condition_( ri2 == logger.end_rule() );

    // Registration before usage
    if( !logger.register_destination( logdest ) )
        {
        delete logdest;
        throw ExLoggerError( "Cannot register 'logdest'." );
        }

    logger.insert_destination( ri1, "test" );
    logger.log( message );
    test_condition_( logdest->tests == 1 && logdest->tsets == 0 );

    // Rule name
    str = ri1.to_string();
    test_condition_( str == "*.*.>=lowest -> test" );

    // Simple destination insert/erase
    logger.insert_destination( ri1, "bullshit" );
    logger.erase_destination( ri1, "test" );
    logger.insert_destination( ri1, "tset" );
    logger.log( message );
    test_condition_( logdest->tests == 1 && logdest->tsets == 1 );
    logger.insert_destination( ri1, "test" );
    logger.log( message );
    test_condition_( logdest->tests == 2 && logdest->tsets == 2 );
    logdest->reset();

    // Re-registration
    logger.unregister_destination( logdest );
    logdest = new LogDestinationTest;
        // logger.unregister_destination deleted the object

    logger.log( message );
    test_condition_( logdest->tests == 0 && logdest->tsets == 0 );

    if( !logger.register_destination( logdest ) )
        {
        delete logdest;
        throw ExLoggerError( "Cannot register 'logdest'." );
        }

    logger.log( message );
    test_condition_( logdest->tests == 1 && logdest->tsets == 1 );

    // Clear
    logdest->reset();
    logger.clear_rules();
    logger.log( message );
    test_condition_( logdest->tests == 0 && logdest->tsets == 0 );

    // Advanced rule 1
    test_condition_( ( ri1 = logger.insert_rule( s_adv1 ) ) != logger.end_rule() );
    logger.insert_destination( ri1, "test" );
    logger.insert_destination( ri1, "TEST" ); // Case sensitive right now; TEST<test - FIXME?
    str = ri1.to_string();
    test_condition_( str == "*.*.<=high, test.>=warning.>=low -> TEST, test" );
     // Selector elements are sorted, * first

    // Advanced rule 2
    test_condition_( ( ri2 = logger.insert_rule( s_adv2 ) ) != logger.end_rule() );
    logger.insert_destination( ri2, "tset" );
    str = ri2.to_string();
    test_condition_( str == "!logger.*.=lowest, !test.=error.<=highest -> tset" );

    // Test 'em
    message.status = STATUS_WARNING( Status::FACILITY_TEST, Status::PRIORITY_LOW );
    logger.log( message );
    test_condition_( logdest->tests == 1 && logdest->tsets == 1 );
    message.status = STATUS_ERROR( Status::FACILITY_TEST );
    logger.log( message );
    test_condition_( logdest->tests == 1 && logdest->tsets == 1 );
    message.status = 
        Status( Status::FACILITY_TEST, Status::TYPE_ERROR, Status::PRIORITY_LOW );

    logger.log( message );
    test_condition_( logdest->tests == 2 && logdest->tsets == 1 );
    message.status = STATUS_INFO( Status::FACILITY_TEST, Status::PRIORITY_LOWEST );
    logger.log( message );
    test_condition_( logdest->tests == 2 && logdest->tsets == 2 );
    message.status = STATUS_INFO( Status::FACILITY_LOGGER, Status::PRIORITY_LOW );
    logger.log( message );
    test_condition_( logdest->tests == 2 && logdest->tsets == 3 );
    message.status = STATUS_INFO( Status::FACILITY_LOGGER, Status::PRIORITY_LOWEST );
    logger.log( message );
    test_condition_( logdest->tests == 2 && logdest->tsets == 3 );

    // Malformed rules
    // FIXME

    // Done
    logdest->reset();
    logger.clear_rules();
    logger.unregister_destination( logdest );
    logdest = NULL;
    }


}; // namespace Massiv::Core::Test
}; // namespace Massiv::Core
}; // namespace Massiv
