/** @file test_registry.cpp
 * Registry tests.
 */

/*
 * begin        : Fri Apr 26 2002
 * copyright    : (C) 2002 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_registry.h"
#include "global/global.h"
#include "registry/registry.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include <iostream>
#include <fstream>


namespace Massiv {
namespace Core {
namespace Test {


// RegistryTest implementation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

RegistryTest::RegistryTest()
    : Test( "registry")
    {
    registry = new Registry();
    }


RegistryTest::~RegistryTest()
    {
    delete registry;
    }


void RegistryTest::test()
    {
    hope( registry );

    test_syntax_checks();
    test_inserts();
    test_aliases();
    test_one_node_search();
    test_registry_search();
    test_wrong_strings();
    test_config_file();
    }  // RegistryTest::test()


// Registry tests implementations
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void RegistryTest::test_syntax_checks()
    {
    test_correct_path_syntax( "" );
    test_correct_path_syntax( "correct_string_9_$#/~" );
    test_wrong_path_syntax( "/wrong/syntax" );
    test_wrong_path_syntax( "wrong/syntax/" );
    test_wrong_path_syntax( "wrong//syntax" );
    test_wrong_path_syntax( "wrong/syntax/\t" );
    test_wrong_path_syntax( "wrong(syntax" );
    test_wrong_path_syntax( "wrong syntax" );

    test_correct_name_syntax( "correct_name_0_$#~" );
    test_wrong_name_syntax( "" );
    test_wrong_name_syntax( "wrong/syntax" );
    test_wrong_name_syntax( "wrong/syntax/\tx" );
    test_wrong_name_syntax( "wrong(syntax" );
    test_wrong_name_syntax( "wrong syntax" );
    }  // RegistryTest::test_syntax_checks()


void RegistryTest::test_inserts()
    {
    /* fill registry with sample data */
    registry->create_variable
        ( 
        "name", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "root_node" 
        );

    registry->create_variable
        ( 
        "sub0/name", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "child100" 
        );

    registry->create_variable
        (
        "sub1/sub0/name",
        ConfigVariable::CVAR_TYPE_STRING,
        "child210"
        );

    registry->create_variable
        (
        "sub1/name",
        ConfigVariable::CVAR_TYPE_STRING,
        "child200"
        );

    registry->create_variable 
        (
        "sub1/sub1/name",
        ConfigVariable::CVAR_TYPE_STRING,
        "child211"
        );

    registry->create_variable
        (
        "sub1/sub0/boolvar",
        ConfigVariable::CVAR_TYPE_BOOLEAN,
        "true"
        );

    registry->create_variable
        (
        "sub1/sub0/floatvar",
        ConfigVariable::CVAR_TYPE_FLOAT,
        "3.14"
        );

    registry->create_variable 
        (
        "sub1/sub1/intvar",
        ConfigVariable::CVAR_TYPE_INTEGER,
        "37"
        );

    registry->create_variable
        (
        "sub1/sub1/large_int_var",
        ConfigVariable::CVAR_TYPE_LARGE_INTEGER,
        "666666666666666"
        );
    
        
    /* positive checks */    
    ConfigVariable * cvar;
    
    cvar = &registry->access_variable( "name" );
    test_condition_( cvar->read_string() == "root_node" );

    cvar = &registry->access_variable( "sub1/sub1/name" );
    test_condition_( cvar->read_string() == "child211" );

    cvar = &registry->access_variable( "sub0/name" );
    test_condition_( cvar->read_string() == "child100" );

    cvar = &registry->access_variable( "sub1/name" );
    test_condition_( cvar->read_string() == "child200" );

    cvar = &registry->access_variable( "sub1/sub0/name" );
    test_condition_( cvar->read_string() == "child210" );

    cvar = &registry->access_variable( "sub1/sub0/boolvar" );
    test_condition_( cvar->read_boolean() == true );

    cvar = &registry->access_variable( "sub1/sub0/floatvar" );
    test_condition_( cvar->read_float() == (float)3.14 );

    cvar = &registry->access_variable( "sub1/sub1/intvar" );
    test_condition_( cvar->read_integer() == 37 );

    cvar = &registry->access_variable( "sub1/sub1/large_int_var" );
    test_condition_( cvar->read_large_integer() == INT64( 666666666666666 ) );
    
    /* negative checks 
     * tries to access a non-defined variable in an existing node,
     * tests whether an exception is properly thrown. */
    try
        {
        cvar = &registry->access_variable( "sub0/nondef" );
        test_condition_( false );
        }
    catch( Exception )
        {
        test_condition_( true );
        }
    
    /* tries to access a non-defined variable in a non-defined node, 
     * tests whether an excpetion is properly thrown. */
    try
        {
        cvar = &registry->access_variable( "nondef/nondef" );
        test_condition_( false );
        }
    catch( Exception )
        {
        test_condition_( true );
        }

    registry->create_variable
        (
        "sub0/xvar",
        ConfigVariable::CVAR_TYPE_INTEGER,
        "13"
        );

    registry->create_variable
        (
        "boolvar",
        ConfigVariable::CVAR_TYPE_BOOLEAN,
        "false"
        );

    /* creating variable that is already created with 
       another type */
    try
        {
        registry->create_variable
            (
            "sub0/xvar",
            ConfigVariable::CVAR_TYPE_BOOLEAN,
            "true"
            );
        test_condition_( false );
        }
    catch( Exception )
        {
        test_condition_( true );
        }

    cvar = &registry->access_variable( "sub0/xvar" );
    test_condition_( cvar->read_integer() == 13 );

    cvar = &registry->access_variable( "boolvar" );
    test_condition_( cvar->read_boolean() == false );

    /* changing previously inserted variable */
    cvar->write_boolean( true );

    cvar = &registry->access_variable( "boolvar" );
    test_condition_( cvar->read_boolean() == true );

    /* creating existing variable with another default value */
    registry->create_variable
        (
        "boolvar",
        ConfigVariable::CVAR_TYPE_BOOLEAN,
        "true"
        );

    test_condition_( cvar->read_boolean() == true );

    registry->create_variable
        (
        "boolvar",
        ConfigVariable::CVAR_TYPE_BOOLEAN,
        "false"
        );

    test_condition_( cvar->read_boolean() == false );


    
    Registry::RegistryNode *    regnode;
    regnode = registry->find_node_for_variable( "sub1/sub0/name" );

    cvar = &registry->access_variable( "name", regnode );
    test_condition_( cvar->read_string() == "child210" );
    }  // test_inserts()


void RegistryTest::test_aliases()
    {
    /* note: the registry still contains data from test_inserts() */

    /* add sample aliases into the registry */
    registry->create_symbolic_link( "link_intvar", "sub1/sub1/intvar" );

    /* continue adding ordinary aliases... */
    registry->create_symbolic_link
        ( 
        "sub1/sub0/link_link_intvar",
        "link_intvar"
        );

    registry->create_symbolic_link
        (
        "sub0/link_link_link_intvar",
        "sub1/sub0/link_link_intvar"
        );

    /* check links */
    ConfigVariable *    cvar;

    cvar = &registry->access_variable( "link_intvar" );
    test_condition_( cvar->read_integer() == 37 );
        /* single link from root to its descendant */

    cvar = NULL;
    Registry::RegistryNode *    regnode;
    regnode = registry->find_node_for_path( "sub1/sub0" );

    cvar = &registry->access_variable( "link_link_intvar", regnode );
    test_condition_( cvar->read_integer() == 37 );

    cvar = &registry->access_variable( "sub1/sub0/link_link_intvar" );
    test_condition_( cvar->read_integer() == 37 );
        /* double link */

    cvar = &registry->access_variable( "sub0/link_link_link_intvar" );
    test_condition_( cvar->read_integer() == 37 );
        /* tripple link */

    /* test patologic cases */
    /* create link that already exists */
    try
        {
        registry->create_symbolic_link( "link_intvar", "sub1/name" );
        test_condition_( false );
        }
    catch( Exception )
        {
        test_condition_( true );
        }

    /* create link to non-defined target */
    try
        {
        registry->create_symbolic_link( "nondef_target", "sub1/nondef" );
        test_condition_( false );
        }
    catch( Exception )
        {
        test_condition_( true );
        }


    try
        {
        registry->create_symbolic_link( "link_intvar", "sub1/sub1/intvar" );
            /* Creating duplicate link - has the same target as the one
             * already defined and thus an exception should not be thrown. */

        test_condition_( true );
        }
    catch( ExCVarDuplicateName )
        {
        test_condition_( false );
        }

    try
        {
        registry->create_symbolic_link( "link_intvar", "sub1/noname" );
            /* In this case an exception will be shown, becase we are trying
             * to create a symbolic link that already exists with a differect
             * target. */

        test_condition_( false );
        }
    catch( ExCVarDuplicateName )
        {
        test_condition_( true );
        }

    
    /* Now we will add an alias to undefined variable without "check target
     * existence" option. Then we will add an identic alias, but with the
     * mentioned option 'on'. We will check whether the second operation will
     * throw exception correctly.
     * Note that creating a symbolic link to an undefined variable is 
     * generally unsafe and thus it should be avoided. */
    registry->create_symbolic_link
                            ( 
                            "link_to_undefined", 
                            "sub1/sub1/sub1/undefined", 
                            false 
                            );

   
    try
        {
        registry->create_symbolic_link
                                (
                                "link_to_undefined",
                                "sub1/sub1/sub1/undefined",
                                true
                                );

        test_condition_( false );
        }
    catch( ExRegistryNodeNotFound )
        {
        test_condition_( true );
        }
    catch( ExCVarNotFound )
        {
        test_condition_( true );
        }

    /* now we must define the target - otherwise a exception would
     * be thrown later... */
    registry->create_variable
                        (
                        "sub1/sub1/sub1/undefined",
                        ConfigVariable::CVAR_TYPE_INTEGER,
                        "0"
                        );
    }  // test_aliases()


void RegistryTest::test_one_node_search()
    {
    /* note: the registry still contains data from test_aliases() */

    /* prepare data */
    registry->create_variable
        ( 
        "sub1/s001test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test001" 
        );

    registry->create_variable
        ( 
        "sub1/s002test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test002" 
        );

    registry->create_variable
        ( 
        "sub1/s003test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test003" 
        );

    registry->create_variable
        ( 
        "sub1/s010test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test010"
        );

    registry->create_variable
        ( 
        "s010test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test010bad" 
        );

    registry->create_variable
        ( 
        "sub1/sub0/s010test", 
        ConfigVariable::CVAR_TYPE_STRING, 
        "test010bad" 
        );

    registry->create_symbolic_link
        (
        "sub1/link001_floatvar",
        "sub1/sub0/floatvar"
        );

    registry->create_symbolic_link
        (
        "sub1/link010_floatvar",
        "sub1/sub0/floatvar"
        );

    registry->create_symbolic_link
        (
        "sub1/link100_floatvar",
        "sub1/sub0/floatvar"
        );

    Registry::SearchContext sc( registry );
    sc = registry->open_registry_node( "sub1" );
    bool result;

    result = registry->search_registry_node( "*?e", sc );
    test_condition_( result );
    test_condition_( sc.get_name() == "name" );

    result = registry->search_registry_node( "*?e", sc );
    test_condition_( !result );

    sc = registry->open_registry_node( "sub1" );
    result = registry->search_registry_node( "l*10_f*", sc );
    test_condition_( result );
    test_condition_( sc.get_name() == "link010_floatvar" );

    result = registry->search_registry_node( "s0?0t??*", sc );
    test_condition_( result );
    test_condition_( sc.get_name() == "s010test" );

    registry->create_symbolic_link
        (
        "justlinks/link_boolvar", 
        "sub1/sub0/boolvar" 
        );

    sc = registry->open_registry_node( "justlinks" );
    result = registry->search_registry_node( "", sc );
    test_condition_( result );
    test_condition_( sc.get_name() == "link_boolvar" );

    result = registry->search_registry_node( "", sc );
    test_condition_( !result );
    }  // test_one_node_search()


void RegistryTest::test_registry_search()
    {
    /* note: the registry now contains data from test_aliases() */

    bool result;

    Registry::SearchContext     sc( registry );
    registry->initialize_search_context( sc );
    
    result = registry->find_variable( "s0?0**", sc );
    test_condition_( result );
    test_condition_( sc.get_type() != Registry::SearchContext::RRT_NOT_FOUND );
    test_condition_( sc.get_name() == "s010test" );

    ConfigVariable * cvar;

    cvar = &registry->access_variable( sc.get_path() + '/' + sc.get_name() );
    test_condition_( cvar->read_string() == "test010bad" );

    result = registry->find_variable( "s0?0**", sc );
    test_condition_( result );
    test_condition_( sc.get_type() != Registry::SearchContext::RRT_NOT_FOUND );
    test_condition_( sc.get_name() == "s010test" );

    cvar = &registry->access_variable( sc.get_path() + '/' + sc.get_name() );
    test_condition_( cvar->read_string() == "test010" );

    result = registry->find_variable( "s0?0**", sc );
    test_condition_( result );
    test_condition_( sc.get_type() != Registry::SearchContext::RRT_NOT_FOUND );
    test_condition_( sc.get_name() == "s010test" );

    cvar = &registry->access_variable( sc.get_path() + '/' + sc.get_name() );
    test_condition_( cvar->read_string() == "test010bad" );

    result = registry->find_variable( "s0?0**", sc );
    test_condition_( !result );
    }  // RegistryTest::test_registry_search()


void RegistryTest::test_wrong_strings()
    {
    std::string str( "test_wrong_strings/str1:string=\"blabla" );

    try
        {
        registry->process_line_of_config_file( str, 1, *registry->root_node );
        test_condition_( 0 );
        }
    catch( ... )
        {
        test_condition_( 1 );
        }

    /* The following should be considered OK. */
    str = "[ test_wrong_strings ]";
    registry->process_line_of_config_file( str, 1, *registry->root_node );

    str = "str_wrong1:string=\"blabla\"";

    registry->process_line_of_config_file( str, 10, *registry->root_node );

    ConfigVariable *    cvar;
    cvar = &registry->access_variable( "test_wrong_strings/str_wrong1" );
    test_condition_( cvar->read_string() == "blabla" );
    }  // RegistryTest::test_wrong_strings()


void RegistryTest::test_config_file()
    {
    /* Loads config file and tests whether all the stuff has been loaded
     * correctly. Then loads the new config file as a subtree of registry
     * and checks also what has been loaded. */

    registry->load_config_file( 
                    std::string( "../../../../../src/core/test/massiv.conf" ) 
                    );

    /* test loaded values */
    ConfigVariable *    cvar;

    cvar = &registry->access_variable( "str_with_space" );
    test_condition_( cvar->read_string() == "a b" );

    cvar = &registry->access_variable( "str_with_quotes0" );
    test_condition_( cvar->read_string() == "a\"b" );

    cvar = &registry->access_variable( "str_with_quotes1" );
    {
    std::string str_aux( "\"a\"" );
    test_condition_( cvar->read_string() == str_aux );
    }

    cvar = &registry->access_variable( "str_with_bslash" );
    test_condition_( cvar->read_string() == "\\\\\\\\abc\\\\\\def\\\\" );

    cvar = &registry->access_variable( "sub0/loaded1" );
    test_condition_( cvar->read_string() == "blabla" );

    cvar = &registry->access_variable( "sub0/conf/loaded2" );
    test_condition_( cvar->read_boolean() == false );

    cvar = &registry->access_variable( "sub0/conf/loaded3" );
    test_condition_( cvar->read_float() == ( float ) 3.14 );

    cvar = &registry->access_variable( "conf/loaded4" );
    test_condition_( cvar->read_string() == "child100" );

    cvar = &registry->access_variable( "sub1/sub0/conf/loaded5" );
    test_condition_( cvar->read_integer() == 13 );

    cvar = &registry->access_variable( "sub1/sub0/conf/loaded6" );
    test_condition_( cvar->read_float() == ( float ) 17 );

    cvar = &registry->access_variable( "sub1/sub0/loaded7" );
    test_condition_( cvar->read_large_integer() == INT64( 555555555555555 ) );

    /* test values loaded from massiv.incl included from massiv.conf */
    cvar = &registry->access_variable( "var_included" );
    test_condition_( cvar->read_string() == "var_included" );

    /* test values loaded from test_hierarchy/massiv.incl included from 
       massiv.conf and massiv.incl */
    cvar = &registry->access_variable( "sub0/var_included2" );
    test_condition_( cvar->read_string() == "var_included2" );
    cvar = &registry->access_variable( "sub1/var_included2" );
    test_condition_( cvar->read_string() == "var_included2" );

    
    /* Prepare a variable that should be overwritten by the following loading
     * of the massiv_nested.conf file. */
    registry->create_variable
        ( 
        "sub0/nested/sub_loaded_existing", 
        ConfigVariable::CVAR_TYPE_INTEGER, 
        "7" 
        );

    /* Create nested_node node. Under this the massiv_nested.conf 
     * configuration file will be loaded. */
    Registry::RegistryNode *  nested_node = 
                                registry->create_node_path( "sub0/nested" );

    registry->load_config_file(
            std::string( "../../../../../src/core/test/massiv_nested.conf" ),
            *nested_node
            );

    cvar = &registry->access_variable( "sub0/nested/sub_loaded1" );
    test_condition_( cvar->read_integer() == 1 );

    cvar = &registry->access_variable( "sub0/nested/sub_loaded_existing" );
    test_condition_( cvar->read_integer() == 13 );
    
    cvar = &registry->access_variable( "sub0/nested/sub/sub_loaded2" );
    test_condition_( cvar->read_float() == ( float ) 2.0 );

    cvar = &registry->access_variable( "sub0/nested/sub/sub_link_loaded1" );
    test_condition_( cvar->read_string() == "blabla" );

    cvar = &registry->access_variable( "sub0/nested/sub/sub/sub/sub_loaded3" );
    test_condition_( cvar->read_string() == "three" );

    cvar = &registry->access_variable( "sub0/nested/sub/sub_loaded4" );
    test_condition_( cvar->read_boolean() == false );

    cvar = &registry->access_variable( "sub0/nested/sub_loaded5" );
    test_condition_( cvar->read_large_integer() == ( Int64 ) 5 );

    /* save of config file */
    registry->save_config_file( std::string("massiv.conf.tmp") );

    /* load again to the new registry */
    Registry    new_registry;

    new_registry.load_config_file( std::string("massiv.conf.tmp") );

    cvar = &new_registry.access_variable( "sub1/sub1/name" );
    test_condition_( cvar->read_string() == "child211" );

    cvar = &new_registry.access_variable( "sub1/sub1/intvar" );
    test_condition_( cvar->read_integer() == 37 );

    cvar = &new_registry.access_variable( "sub1/sub0/conf/loaded5" );
    test_condition_( cvar->read_integer() == 13 );

    cvar = &new_registry.access_variable( "sub1/sub0/conf/loaded6" );
    test_condition_( cvar->read_float() == (float)17 );
    test_condition_( cvar->get_type() == ConfigVariable::CVAR_TYPE_FLOAT );

    cvar = &new_registry.access_variable( "sub1/sub0/link_link_intvar" );
    test_condition_( cvar->read_integer() == 37 );

    cvar = &new_registry.access_variable( "sub1/s001test" );
    test_condition_( cvar->read_string() == "test001" );

    cvar = &new_registry.access_variable( "sub1/link100_floatvar" );
    test_condition_( cvar->read_float() == (float)3.14 );
    test_condition_( cvar->get_type() == ConfigVariable::CVAR_TYPE_FLOAT );

    cvar = &new_registry.access_variable( "s010test" );
    test_condition_( cvar->read_string() == "test010bad" );

    cvar = &new_registry.access_variable( "boolvar" );
    test_condition_( cvar->read_boolean() == false );
    }  // test_config_file()


void RegistryTest::test_correct_path_syntax
    (
    std::string     path
    )
    {
    try
        {
        Registry::check_path_syntax( path );
        test_condition_( true );
        }
    catch( ExWrongSyntax & )
        {
        test_condition_( false );
        }
    }  // test_correct_path_syntax()


void RegistryTest::test_wrong_path_syntax
    (
    std::string     path
    )
    {
    try
        {
        Registry::check_path_syntax( path );
        test_condition_( false );
        }
    catch( ExWrongSyntax & )
        {
        test_condition_( true );
        }
    }  // test_wrong_path_syntax()


void RegistryTest::test_correct_name_syntax
    (
    std::string     name
    )
    {
    try
        {
        Registry::check_name_syntax( name );
        test_condition_( true );
        }
    catch( ExWrongSyntax & )
        {
        test_condition_( false );
        }
    }  // test_correct_name_syntax()


void RegistryTest::test_wrong_name_syntax
    (
    std::string     name
    )
    {
    try
        {
        Registry::check_name_syntax( name );
        test_condition_( false );
        }
    catch( ExWrongSyntax & )
        {
        test_condition_( true );
        }
    }  // test_wrong_name_syntax()


}; // namespace Massiv::Core::Test
}; // namespace Massiv::Core
}; // namespace Massiv
