/** @file test_serializer.cpp
 * Testing of all serializers.
 * @note ObjectId/ObjectPointer/PPointer, EventHandle/PEventHandle serializers are not tested.
 *       It can't be easily done without all the object stuff (factories, etc.).
 */

/*
 * begin        : Thu Feb 21 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "test/test_serializer.h"
#include "serializer/serializer.h"
#include "property/prop_int.h"
#include "property/prop_vli.h"
#include "property/prop_bool.h"
#include "property/prop_float.h"
#include "property/vector23.h"
#include "property/prop_time.h"
#include "property/prop_string.h"
#include "property/prop_vector.h"
#include "property/prop_array.h"
#include "property/prop_dictionary.h"
#include "serializer/stype_time.h"
#include "serializer/stype_int.h"
#include "object/object_status.h"
#include "object/providerid.h"
#include "node/nodeid.h"
#include "net/net.h"
#include "sysevent/find_object_msg.h"
#include "sysevent/migration_group_msg.h"
#include "sysevent/object_status_msg.h"
#include "sysevent/time_synch_msg.h"
#include "object/local_schedule.h"
#include "time/time_manager.h"
#include "sysevent/system_event_queue.h"
#include "global/global.h"
#include "object/object_manager.h"
#include "replication/replica_server.h"
#include "object/object_provider_state.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include <sstream>


namespace Massiv {
namespace Core {
namespace Test {


namespace {

struct FQ32_2
    {
    static const Float32 min_value;
    static const Float32 max_value;
    enum { bits = 2 };
    };
const Float32 FQ32_2::min_value = 0.0f;
const Float32 FQ32_2::max_value = 1.0f;

struct FQ32_32
    {
    static const Float32 min_value;
    static const Float32 max_value;
    enum { bits = 32 };
    };
const Float32 FQ32_32::min_value = -8.0f;
const Float32 FQ32_32::max_value = 4.0f;

struct FQ64_2
    {
    static const Float64 min_value;
    static const Float64 max_value;
    enum { bits = 2 };
    };
const Float64 FQ64_2::min_value = 0.0;
const Float64 FQ64_2::max_value = 1.0;

struct FQ64_32
    {
    static const Float64 min_value;
    static const Float64 max_value;
    enum { bits = 32 };
    };
const Float64 FQ64_32::min_value = -128.0;
const Float64 FQ64_32::max_value = 128.0;

} // anonymous namespace


SerializerTest::SerializerTest() :
    Test( "serializers" )
    {
    }


SerializerTest::~SerializerTest()
    {
    }


void SerializerTest::test()
    {
    test_bitstream();
    test_text();

    test_events();
    test_system_messages();
    test_array();
    test_dictionary();
    test_set();
    }


void SerializerTest::test_bitstream()
    {
    // Bitstream serializers. All the types are simple, dirty timestamps are not
    // used at all.

    BitOStream                          bos;
    int                                 bitlen;
    Serializer::Description             desc;
    desc.repflags_yes = REPMASK_MIGRATE_YES;
    desc.repflags_no = REPMASK_MIGRATE_NO;
    desc.lossy = true;

    PVli<UInt64>                        vui64( UINT64( 0xfedcba9876543210 ) );
    PVli<Int32>                         vi32( -3 );
    PVli<Int64>                         vi64( INT64( 0xfffffff ) );
    PVli<UInt32>                        vui32( 0x87654321 );

    PInt<Int32, 7>                      i32( 0xff );
    PInt<UInt32, 32>                    ui32( 0x87654321 );
    PInt<Int64, 15>                     i64( INT64( 0xffff ) );
    PInt<UInt64, 64>                    ui64( UINT64( 0xfedcba9876543210 ) );
    PBool                               bt( true );
    PBool                               bf( false );
    PFloat<Float32>                     f32( 8.0f );
    PFloat<Float64>                     f64( 16384.0 );
    PFloatQ<Float32, FQ32_2>            f32q2_0a( 0.0f );
    PFloatQ<Float32, FQ32_2>            f32q2_0b( 0.15f );
    PFloatQ<Float32, FQ32_2>            f32q2_1a( 0.25f );
    PFloatQ<Float32, FQ32_2>            f32q2_1b( 0.4f );
    PFloatQ<Float32, FQ32_2>            f32q2_2a( 0.6f );
    PFloatQ<Float32, FQ32_2>            f32q2_2b( 0.75f );
    PFloatQ<Float32, FQ32_2>            f32q2_3a( 0.85f );
    PFloatQ<Float32, FQ32_2>            f32q2_3b( 1.0f );
    PFloatQ<Float32, FQ32_32>           f32q32_min( -57.3f );  // -> 0
    PFloatQ<Float32, FQ32_32>           f32q32_mid( -2.0f );  // -> near 2^31
    PFloatQ<Float32, FQ32_32>           f32q32_max( 132.7f ); // -> neg 0
    PFloatQ<Float32, FQ32_2>            f64q2_0( 0.15f );
    PFloatQ<Float64, FQ64_2>            f64q2_3( 0.85 );
    PFloatQ<Float64, FQ64_32>           f64q32_min( -500.0 );
    PFloatQ<Float64, FQ64_32>           f64q32_max( 500.0 );
    STime                               stime( 1234 * 1000 );
    PTime                               ptime( 4321 * 1000 );
    ProviderId                          pid1 = ProviderId::create_from_number( ( 1 << ProviderId::PROVIDER_ID_BITS_COUNT ) - 1 );
    UInt64                              next_id = ( UINT64( 1 ) << ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT ) - 1;
    ObjectProviderState                 pstate( pid1, next_id );
    NodeId                              sid1 = NodeId::create_from_type_and_relative_id( NodeId::TYPE_CLIENT, ( ( 1 << NodeId::RELATIVE_ID_BITS_COUNT ) - 1 ) );
    ObjectStatus                        ostatus_alive( ObjectStatus::ALIVE );
    ObjectStatus                        ostatus_not_accessible( ObjectStatus::NOT_ACCESSIBLE );
    ObjectStatus                        ostatus_unknown( ObjectStatus::UNKNOWN );
    PString                             pstring( "Kobyla ma maly bok" );
    PString                             pstring2( "" );
    PVector2                            pvector2( 789, -123 );
    PVector3                            pvector3( 1, 2, 123456 );
    int                                 raw_buffer_len = 456;
    UInt8 *                             raw_buffer = new UInt8[ raw_buffer_len ];
    hope( raw_buffer );
    
    /** StringAttribute / StringAttribute::iterator test */

    PString::Iterator it1 =  pstring.begin();
    PString::Iterator it2 =  pstring.end();

    pstring.insert( 6, "(krava)" );
    // Now iterators it1 and it2 are not valid.

    it1 =  pstring.begin();
    it2 =  pstring.end();
    // Now are valid again.

    // Let's reverse string using function std::reverse().
    std::reverse( pstring.begin(), pstring.end() );

    test_condition_( *--it2 == 'K' );
    test_condition_( *it1 == 'k' );

    it2 -= pstring.size() - 1;

    test_condition_( it2 == it1 );  // Both at the beginning.

    *it1 = 'X';     // write_requested() called.

    test_condition_( *it2 == 'X' );  // read_requested() called.


    // Write
    // ~~~~~
    test_condition_( bos.bit_size() == 0 );
    bitlen = 0;

    // variable length integers
    vui64.bitstr_write( bos, desc );
    vi64.bitstr_write( bos, desc );
    vui32.bitstr_write( bos, desc );
    vi32.bitstr_write( bos, desc );

    test_condition_( bos.bit_size() == bitlen + 128 + 57 + 64 + 5 );
    bitlen = bos.bit_size();

    // integers
    i32.bitstr_write( bos, desc );
    ui32.bitstr_write( bos, desc );
    i64.bitstr_write( bos, desc );
    ui64.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 7 + 32 + 15 + 64 );
    bitlen = bos.bit_size();

    // booleans
    bt.bitstr_write( bos, desc );
    bf.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 2 );
    bitlen = bos.bit_size();

    // floats
    f32.bitstr_write( bos, desc );
    f64.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 32 + 64 );
    bitlen = bos.bit_size();

    // quantized floats
    f32q2_0a.bitstr_write( bos, desc );
    f32q2_0b.bitstr_write( bos, desc );
    f32q2_1a.bitstr_write( bos, desc );
    f32q2_1b.bitstr_write( bos, desc );
    f32q2_2a.bitstr_write( bos, desc );
    f32q2_2b.bitstr_write( bos, desc );
    f32q2_3a.bitstr_write( bos, desc );
    f32q2_3b.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 8 * 2 );
    bitlen = bos.bit_size();
    f32q32_min.bitstr_write( bos, desc );
    f32q32_mid.bitstr_write( bos, desc );
    f32q32_max.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 3 * 32 );
    bitlen = bos.bit_size();
    f64q2_0.bitstr_write( bos, desc );
    f64q2_3.bitstr_write( bos, desc );
    f64q32_min.bitstr_write( bos, desc );
    f64q32_max.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 2 * 2 + 2 * 32 );
    bitlen = bos.bit_size();

    // time
    stime.bitstr_write( bos, desc );
    ptime.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 2 * 64 );
    bitlen = bos.bit_size();

    // provider id and state
    pid1.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + ProviderId::PROVIDER_ID_BITS_COUNT );
    bitlen = bos.bit_size();
    pstate.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + ProviderId::PROVIDER_ID_BITS_COUNT + ProviderId::PROVIDER_RELATIVE_ID_BITS_COUNT );
    bitlen = bos.bit_size();

    // node id
    sid1.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + NodeId::TYPE_BITS_COUNT + NodeId::RELATIVE_ID_BITS_COUNT );
    bitlen = bos.bit_size();

    // object status
    ostatus_alive.bitstr_write( bos, desc );
    ostatus_not_accessible.bitstr_write( bos, desc );
    ostatus_unknown.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 3 * ObjectStatus::STATUS_BITS_COUNT );
    bitlen = bos.bit_size();

    // string
    pstring.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + ( int )( pstring.size() + 1 ) * 7 );
    bitlen = bos.bit_size();
    pstring2.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + ( int )( pstring2.size() + 1 ) * 7 );
    bitlen = bos.bit_size();

    // vector23
    pvector2.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 2 * 32 );
    bitlen = bos.bit_size();
    pvector3.bitstr_write( bos, desc );
    test_condition_( bos.bit_size() == bitlen + 3 * 32 );
    bitlen = bos.bit_size();

    // Raw binary data
    bos.write_raw( raw_buffer, raw_buffer_len );
    bitlen += 8 - ( bitlen % 8 ); // Write raw alignes to byte boundary.
    test_condition_( bos.bit_size() == bitlen + 8 * raw_buffer_len );
    bitlen = bos.bit_size();

    // Read
    // ~~~~

    BitIStream bis( bos.get_buffer() );

    // integers
    i32 = 0; ui32 = 0; i64 = INT64( 0 ); ui64 = UINT64( 0 ); vui64 = UINT64( 0 ); vi32 = 0; vi64 = INT64( 0 ); vui32 = 0;
    vui64.bitstr_read( bis, desc );         test_condition_( vui64 == UINT64( 0xfedcba9876543210 ) );
    vi64.bitstr_read( bis, desc );          test_condition_( vi64 == INT64( 0xfffffff ) );
    vui32.bitstr_read( bis, desc );         test_condition_( vui32 == 0x87654321 );    
    vi32.bitstr_read( bis, desc );          test_condition_( vi32 == -3);
    
        
    i32.bitstr_read( bis, desc );           test_condition_( i32 == ( 0xff & 0x7f ) );
    ui32.bitstr_read( bis, desc );          test_condition_( ui32 == 0x87654321 );
    i64.bitstr_read( bis, desc );           test_condition_( i64 == ( INT64( 0xffff ) & INT64( 0x7fff ) ) );
    ui64.bitstr_read( bis, desc );          test_condition_( ui64 == UINT64( 0xfedcba9876543210 ) );

    // booleans
    bt = false; bf = true;
    bt.bitstr_read( bis, desc );            test_condition_( bt == true );
    bf.bitstr_read( bis, desc);             test_condition_( bf == false );

    // floats
    f32 = 0.0f; f64 = 0.0;
    f32.bitstr_read( bis, desc );           test_condition_( f32 == 8.0f );
    f64.bitstr_read( bis, desc );           test_condition_( f64 == 16384.0 );

    // quantized floats
    SInt<UInt32, 2>     qi2;
    SInt<UInt32, 32>    qi32;
    f32q2_0a.bitstr_read( bis, desc );      test_condition_( f32q2_0a == 0.0f );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 0 );
    f32q2_1a.bitstr_read( bis, desc );      test_condition_( fabs( f32q2_1a - 0.3333f ) < 0.001f );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 1 );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 2 );
    f32q2_2b.bitstr_read( bis, desc );      test_condition_( fabs( f32q2_2b - 0.6666f ) < 0.001f );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 3 );
    f32q2_3b.bitstr_read( bis, desc );      test_condition_( f32q2_3b == 1.0f );
    qi32.bitstr_read( bis, desc );          test_condition_( qi32 == 0 );
    qi32.bitstr_read( bis, desc );          test_condition_( qi32 >= (UInt32) ( ( 1U << 31 ) - 1 ) );
                                            test_condition_( qi32 <= (UInt32) ( ( 1U << 31 ) + 1 ) );
    qi32.bitstr_read( bis, desc );          test_condition_( qi32 == (UInt32) ~0 );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 0 );
    qi2.bitstr_read( bis, desc );           test_condition_( qi2 == 3 );
    f64q32_min.bitstr_read( bis, desc );    test_condition_( fabs( f64q32_min - -128.0 ) < 0.0001 );
    f64q32_max.bitstr_read( bis, desc );    test_condition_( fabs( f64q32_max - 128.0 ) < 0.0001 );

    // time
    stime.bitstr_read( bis, desc );         test_condition_( stime.is_equal_to( STime( 1234 * 1000 ) ) );
    ptime.bitstr_read( bis, desc );         test_condition_( ptime.is_equal_to( STime( 4321 * 1000 ) ) );

    // provider id and state
    ProviderId          pid2;
    ObjectProviderState pstate2;
    pid2.bitstr_read( bis, desc );          test_condition_( pid2 == pid1 );
    pstate2.bitstr_read( bis, desc );       test_condition_( pstate2.get_provider_id() == pid1 );
                                            test_condition_( pstate2.next_relative_object_id == next_id);
    // node id
    NodeId    sid2;
    sid2.bitstr_read( bis, desc );          test_condition_( sid2 == sid1 );

    // object status
    ObjectStatus ostatus1;
    ObjectStatus ostatus2( ObjectStatus::ALIVE );
    ostatus1.bitstr_read( bis, desc );      test_condition_( ostatus1 == ObjectStatus::ALIVE );
    ostatus1.bitstr_read( bis, desc );      test_condition_( ostatus1 == ObjectStatus::NOT_ACCESSIBLE );
    ostatus2.bitstr_read( bis, desc );      test_condition_( ostatus2 == ObjectStatus::UNKNOWN );

    // string
    PString     pstring_r1;
    pstring_r1.bitstr_read( bis, desc );
    test_condition_( pstring.compare( pstring_r1 ) == 0 );
    PString     pstring_r2;
    pstring_r2.bitstr_read( bis, desc );
    test_condition_( pstring2.compare( pstring_r2 ) == 0 );

    // vector23
    PVector2    pvector2_r; 
    pvector2_r.bitstr_read( bis, desc );
    test_condition_( pvector2 == pvector2_r );
    PVector3    pvector3_r; 
    pvector3_r.bitstr_read( bis, desc );
    test_condition_( pvector3 == pvector3_r );

    // Raw binary data
    UInt8 * raw_buffer_read = new UInt8[ raw_buffer_len ];
    hope( raw_buffer_read );
    std::memset( raw_buffer_read, 0, raw_buffer_len ); // Zero all.
    bis.read_raw( raw_buffer_read, raw_buffer_len );
    test_condition_( std::memcmp( raw_buffer_read, raw_buffer, raw_buffer_len ) == 0 );

    // Cleanup.

    delete [] raw_buffer;
    delete [] raw_buffer_read;
    }


void SerializerTest::test_text()
    {
    std::ostringstream                  ostream;
    TextWriter                          tw( ostream );
    Serializer::Description             desc;
    desc.repflags_yes = REPMASK_MIGRATE_YES;
    desc.repflags_no = REPMASK_MIGRATE_NO;
    desc.lossy = false;
    
    PVli<Int32>                         vi32( 0xff );
    PInt<Int32, 7>                      i32( 0xff );
    PInt<UInt32, 32>                    ui32( 0x87654321 );
    PInt<Int64, 15>                     i64( INT64( 0xffff ) );
    PInt<UInt64, 64>                    ui64( UINT64( 0xfedcba9876543210 ) );
    PBool                               bt( true );
    PBool                               bf( false );
    PFloat<Float32>                     f32( 8.0f );
    PFloat<Float64>                     f64( 16384.0 );
    PFloatQ<Float32, FQ32_2>            f32q( -8.0f );
    PFloatQ<Float64, FQ64_2>            f64q( 0.5 );
    STime                               stime( 1234 * 1000 );
    PTime                               ptime( 4321 * 1000 );

    PString                             pstring_w1( "Dog\n\t eat dog" );
    PString                             pstring_w2( "" ); // Empty string.
    PString                             pstring_w3( "Cat\n\t eat cat." );

    // Write
    // ~~~~~

    tw.line() << "# Comment";
    tw.end_line();
    tw.change_indent( 2 );
    test_condition_( tw.change_indent( 0 ) == 2 );
    tw.end_line(); // empty line
    test_condition_( tw.line_number() == 3 );
    test_condition_( tw.change_indent( -4 ) == 0 );

    // variable length integers
    vi32.text_write( tw, desc );
    tw.end_line();

    // integers
    i32.text_write( tw, desc );
    tw.write_space();
    ui32.text_write( tw, desc );
    tw.write_space();
    i64.text_write( tw, desc );
    tw.write_space();
    ui64.text_write( tw, desc );
    tw.end_line();

    // booleans
    tw.end_line();
    tw.change_indent( 2 );
    tw.line() << "# Indented";
    tw.end_line();
    bt.text_write( tw, desc );
    tw.write_space();
    bf.text_write( tw, desc );
    tw.end_line();

    // floats
    tw.end_line();
    tw.change_indent( 2 );
    tw.line() << "# Indented more";
    tw.end_line();
    f32.text_write( tw, desc );
    tw.write_space();
    f64.text_write( tw, desc );
    tw.end_line();
    f32q.text_write( tw, desc );
    tw.write_space();
    f64q.text_write( tw, desc );
    tw.end_line();

    // time
    tw.end_line();
    tw.set_indent( 0 );
    tw.line() << "# Not indented";
    tw.end_line();
    stime.text_write( tw, desc );
    tw.end_line();
    ptime.text_write( tw, desc );
    tw.end_line();

    // string
    tw.end_line();
    pstring_w1.text_write( tw, desc );
    tw.write_space();
    pstring_w2.text_write( tw, desc );
    tw.end_line();
    pstring_w3.text_write( tw, desc );
    tw.end_line();

    // direct hack tests
    tw.end_line();
    tw.line() << "true or false";
    tw.end_line();


    // Read
    // ~~~~

    std::istringstream                  istream( ostream.str() );
    TextReader                          tr( istream );

    // integers
    i32 = 0; ui32 = 0; i64 = INT64( 0 ); ui64 = UINT64( 0 );  vi32 = 0 ;
    test_condition_( tr.next_line() );
    vi32.text_read( tr, desc );             test_condition_( vi32 == ( 0xff ) );

    test_condition_( tr.next_line() );
    i32.text_read( tr, desc );              test_condition_( i32 == ( 0xff & 0x7f ) );
    ui32.text_read( tr, desc );             test_condition_( ui32 == 0x87654321 );
    i64.text_read( tr, desc );              test_condition_( i64 == ( INT64( 0xffff ) & INT64( 0x7fff ) ) );
    ui64.text_read( tr, desc );             test_condition_( ui64 == UINT64( 0xfedcba9876543210 ) );

    // booleans
    bt = false; bf = true;
    test_condition_( tr.next_line() );
    bt.text_read( tr, desc );               test_condition_( bt == true );
    bf.text_read( tr, desc );               test_condition_( bf == false );

    // floats
    f32 = 0.0f; f64 = 0.0; f32q = 0.0f; f64q = 0.0;
    test_condition_( tr.next_line() );
    f32.text_read( tr, desc );              test_condition_( f32 == 8.0f );
    f64.text_read( tr, desc );              test_condition_( f64 == 16384.0 );
    test_condition_( tr.next_line() );
    f32q.text_read( tr, desc );             test_condition_( f32q == -8.0f );
    f64q.text_read( tr, desc );             test_condition_( f64q == 0.5 );

    // time
    test_condition_( tr.next_line() );
    stime.text_read( tr, desc );            test_condition_( stime.is_equal_to( STime( 1234 * 1000 ) ) );
    test_condition_( tr.next_line() );
    ptime.text_read( tr, desc );            test_condition_( ptime.is_equal_to( STime( 4321 * 1000 ) ) );

    // string
    PString     pstring_r1;
    PString     pstring_r2;
    PString     pstring_r3;
    test_condition_( tr.next_line() );
    pstring_r1.text_read( tr, desc );
    pstring_r2.text_read( tr, desc );
    test_condition_( tr.next_line() );
    pstring_r3.text_read( tr, desc );
    test_condition_( pstring_r1.compare( pstring_w1 ) == 0 );
    test_condition_( pstring_r2.compare( pstring_w2 ) == 0 );
    test_condition_( pstring_r3.compare( pstring_w3 ) == 0 );

    bt = false; bf = true;
    test_condition_( tr.next_line() );
    try
        {
        i32.text_read( tr, desc ); // should throw ExUnexpectedToken, and read nothing
        }
    catch( Exception & ex )
        {
        test_condition_( typeid( ex ) == typeid( ExUnexpectedToken ) );
        tr.line().clear();
        }
    bt.text_read( tr, desc );
    test_condition_( bt == true );
    try
        {
        bf.text_read( tr, desc ); // should throw ExUnexpectedValue, and read "or"
        }
    catch( Exception & ex )
        {
        test_condition_( typeid( ex ) == typeid( ExUnexpectedValue ) );
        tr.line().clear();
        }
    bf.text_read( tr, desc );
    test_condition_( bf == false );
    try
        {
        bt.text_read( tr, desc );
        }
    catch( Exception & ex )
        {
        test_condition_( typeid( ex ) == typeid( ExUnexpectedEol ) );
        tr.line().clear();
        }
    test_condition_( !tr.next_line() );
    }


void SerializerTest::test_events()
    {
    // Bitstream serializers.
    BitOStream                          bos;
    BitIStream                          bis;
    Serializer::Description             desc;
    desc.repflags_yes = REPMASK_MIGRATE_YES;
    desc.repflags_no = REPMASK_MIGRATE_NO;
    desc.lossy = true;

    // Text serializers.
    std::ostringstream                  ostream;
    TextWriter                          tw( ostream );

    LocalSchedule                       schedule;
    LocalSchedule                       oscheduleb;
    LocalSchedule                       oschedulet;
    Event                               e;
    Event                               oeb;
    Event                               oet;
    STime                               stime1;
    STime                               stime2;
    ObjectId                            oid1; // FIXME initialize to non-zero
    ObjectId                            oid2; // FIXME initialize to non-zero
    EventId                             eid;
    EventId                             oeidb;
    EventId                             oeidt;

    test_condition_( schedule.is_empty() );
    test_condition_( schedule.get_size() == 0 );
    test_condition_( eid.is_private() );
    test_condition_( eid.is_null() );
    test_condition_( !schedule.remove_event( EventId(), e ) );

    stime1 = STime( 1.0 );
    stime2 = STime( 2.0 );
    test_condition_( stime1.is_older_than( stime2 ) );

    // Write some event.
    e = Event( stime1, oid1, true );
    e.bitstr_write( bos, desc );
    e.text_write( tw, desc );

    tw.end_line();

    // Write some event id.
    eid = schedule.add_new_public_event( e );
    eid.bitstr_write( bos, desc );
    eid.text_write( tw, desc );

    tw.end_line();

    // Write and schedule another event.
    e = Event( stime2, oid2, false );
    e.bitstr_write( bos, desc );
    e.text_write( tw, desc );

    tw.end_line();

    schedule.add_new_private_event( e );

    // Write the local schedule.
    schedule.bitstr_write( bos, desc );
    schedule.text_write( tw, desc );

    test_condition_( !schedule.is_empty() );
    test_condition_( schedule.get_size() == 2 );
    test_condition_( schedule.get_time_of_first_event().is_equal_to( stime1 ) );
    test_condition_( !eid.is_private() );
    test_condition_( !eid.is_null() );

    //
    // Read everything back.
    //

    std::istringstream                  istream( ostream.str() );
    TextReader                          tr( istream );
    bis.bind( bos.get_buffer() );

    // Read first event.
    test_condition_( tr.next_line() );
    oeb.bitstr_read( bis, desc );
    oet.text_read( tr, desc );

    test_condition_( oeb.execution_time.is_equal_to( stime1 ) );
    test_condition_( oeb.destination == oid1 );
    test_condition_( oeb.notify_owner() );

    test_condition_( oet.execution_time.is_equal_to( stime1 ) );
    test_condition_( oet.destination == oid1 );
    test_condition_( oet.notify_owner() );

    // Read event id.
    test_condition_( tr.next_line() );
    oeidb.bitstr_read( bis, desc );
    oeidt.text_read( tr, desc );

    test_condition_( !oeidb.is_private() );
    test_condition_( !oeidb.is_null() );
    test_condition_( oeidb == eid );

    test_condition_( !oeidt.is_private() );
    test_condition_( !oeidt.is_null() );
    test_condition_( oeidt == eid );

    // Read second event.
    test_condition_( tr.next_line() );
    oeb.bitstr_read( bis, desc );
    oet.text_read( tr, desc );

    test_condition_( oeb.execution_time.is_equal_to( stime2 ) );
    test_condition_( oeb.destination == oid2 );
    test_condition_( !oeb.notify_owner() );

    test_condition_( oet.execution_time.is_equal_to( stime2 ) );
    test_condition_( oet.destination == oid2 );
    test_condition_( !oeb.notify_owner() );

    // Read local schedule.
    test_condition_( tr.next_line() );
    oscheduleb.bitstr_read( bis, desc );
    oschedulet.text_read( tr, desc );

    test_condition_( !oscheduleb.is_empty() );
    test_condition_( oscheduleb.get_size() == 2 );
    test_condition_( oscheduleb.get_time_of_first_event().is_equal_to( stime1 ) );
    test_condition_( oscheduleb.remove_event( eid, e ) );
    test_condition_( e.execution_time.is_equal_to( stime1 ) );
    test_condition_( e.notify_owner() );
    test_condition_( e.destination == oid1 );
    test_condition_( oscheduleb.get_size() == 1 );
    test_condition_( oscheduleb.get_time_of_first_event().is_equal_to( stime2 ) );
    test_condition_( oscheduleb.get_first_event( e ) );
    oscheduleb.remove_first_event( true );
    test_condition_( e.execution_time.is_equal_to( stime2 ) );
    test_condition_( e.destination == oid2 );
    test_condition_( !e.notify_owner() );
    test_condition_( oscheduleb.is_empty() );

    test_condition_( !oschedulet.is_empty() );
    test_condition_( oschedulet.get_size() == 2 );
    test_condition_( oschedulet.get_time_of_first_event().is_equal_to( stime1 ) );
    test_condition_( oschedulet.remove_event( eid, e ) );
    test_condition_( e.execution_time.is_equal_to( stime1 ) );
    test_condition_( e.destination == oid1 );
    test_condition_( e.notify_owner() );
    test_condition_( oschedulet.get_size() == 1 );
    test_condition_( oschedulet.get_time_of_first_event().is_equal_to( stime2 ) );
    test_condition_( oschedulet.get_first_event( e ) );
    oschedulet.remove_first_event( true );
    test_condition_( e.execution_time.is_equal_to( stime2 ) );
    test_condition_( e.destination == oid2 );
    test_condition_( !e.notify_owner() );
    test_condition_( oschedulet.is_empty() );
    }


void SerializerTest::test_system_messages()
    {
#if 0
    // Bitstream serializers.
    BitOStream                          bos;
    BitIStream                          bis;
    Serializer::Description             desc;
    desc.repflags_yes = REPMASK_MIGRATE_YES;
    desc.repflags_no = REPMASK_MIGRATE_NO;
    desc.lossy = true;

    FindObjectMessage                   ifomp;
    FindObjectMessage                   ofomp;

    MigrationGroupMessage               imgmp;
    MigrationGroupMessage               omgmp;

    ObjectStatusMessage                 iosmp;
    ObjectStatusMessage                 oosmp;

    TimeSynchronizationMessage          itsmp;
    TimeSynchronizationMessage          otsmp;

    ObjectId                            oid1; // FIXME initialize to non-zero
    ObjectId                            oid2; // FIXME initialize to non-zero
    NodeId                              sid1;
    NodeId                              sid2;
    UInt32                              hops_count;
    UInt32                              provider_hops_count;
    ObjectStatus                        ostatus;
    STime                               stime;
    STime                               massiv_time;
    bool                                synchronized;
    Float64                             sync_interval;
    
    SystemMessage::Type                 msg_type;

    // Init common stuff.
    sid1 = NodeId::create_from_type_and_relative_id( NodeId::TYPE_SERVER, ( 0x4321fedc & ( ( 1 << NodeId::RELATIVE_ID_BITS_COUNT ) - 1 ) ) );
    sid2 = NodeId::create_from_type_and_relative_id( NodeId::TYPE_CLIENT, ( 0x98765abd & ( ( 1 << NodeId::RELATIVE_ID_BITS_COUNT ) - 1 ) ) );
    hops_count = FindObjectMessage::HOPS_MAX_VALUE;
    provider_hops_count = FindObjectMessage::PROVIDER_HOPS_MAX_VALUE;
    ostatus.set( ObjectStatus::ALIVE );
    stime = STime( 1.0 );
    massiv_time = Global::system_time();
    synchronized = true;
    sync_interval = 123.456;
    
    //
    // Initialize messages.
    //

    ifomp = FindObjectMessage
        (
        oid1, // destination object
        hops_count,
        provider_hops_count
        );
    ifomp.set_sender( sid1 );
    ifomp.set_initiator( sid2 );

    imgmp = MigrationGroupMessage
        (
        ifomp
        );
    imgmp.objects_count = 1;
    imgmp.objects.push_back( oid2 );
    /** @todo imgmp.message_stream test? */

    iosmp = ObjectStatusMessage
        (
        oid1,
        ostatus,
        sid1,
        true
        );

    itsmp = TimeSynchronizationMessage
        (
        synchronized,
        massiv_time,
        sync_interval
        );

    // Write the messages.
    ifomp.bitstr_write( bos, desc );
    imgmp.bitstr_write( bos, desc );
    iosmp.bitstr_write( bos, desc );
    itsmp.bitstr_write( bos, desc );

    // Initialize bitistream.
    bis.bind( bos.get_buffer() );

    // Read the messages back.
    msg_type = SystemMessage::bitstr_read_type( bis, desc );
    test_condition_( msg_type == SystemMessage::FIND_OBJECT );
    ofomp.bitstr_read( bis, desc );

    msg_type = SystemMessage::bitstr_read_type( bis, desc );
    test_condition_( msg_type == SystemMessage::MIGRATION_GROUP );
    omgmp.bitstr_read( bis, desc );

    msg_type = SystemMessage::bitstr_read_type( bis, desc );
    test_condition_( msg_type == SystemMessage::OBJECT_STATUS );
    oosmp.bitstr_read( bis, desc );

    msg_type = SystemMessage::bitstr_read_type( bis, desc );
    test_condition_( msg_type == SystemMessage::TIME_SYNCHRONIZATION );
    otsmp.bitstr_read( bis, desc );

    //
    // Test messages.
    //

    // FindObjectMessage.
    // Sender is not serialized.
    test_condition_( ofomp.get_sender() == Global::network().get_local_node_id() );
    test_condition_( ofomp.destination == oid1 );
    test_condition_( ofomp.get_initiator() == sid2 );
    test_condition_( ofomp.hops == hops_count );
    test_condition_( ofomp.provider_hops == provider_hops_count );

    // MigrationGroupMessage.
    // Sender is not serialized.
    test_condition_( omgmp.get_sender() == Global::network().get_local_node_id() );
    test_condition_( omgmp.destination == oid1 );
    test_condition_( omgmp.get_initiator() == sid2 );
    test_condition_( omgmp.hops == hops_count );
    test_condition_( omgmp.provider_hops = provider_hops_count );
    test_condition_( omgmp.objects_count == 1 );
    test_condition_( omgmp.objects.size() == 1 );
    if( omgmp.objects.size() > 0 )
        {
        test_condition_( omgmp.objects[0] == oid2 );
        }

    // ObjectStatus message.
    test_condition_( oosmp.object == oid1 );
    test_condition_( oosmp.object_status == ostatus );
    test_condition_( oosmp.node == sid1 );
    test_condition_( oosmp.trustworthy == true );

    // TimeSynchronization message.
    test_condition_( otsmp.synchronized == synchronized );
    test_condition_( otsmp.massiv_time == massiv_time );
    test_condition_( otsmp.sync_interval == sync_interval );

#endif
    }


void SerializerTest::test_array()
    {
    PArray< PInt32 >            array;
    PArray< PInt32 >            after_init;
    PArray< PInt32 >            first_modify;
    PArray< PInt32 >            second_modify;

    PArray< PArray< PString > > array2d;
    PArray< PArray< PString > > after_init2d;
    PArray< PArray< PString > > first_modify2d;
    PArray< PArray< PString > > second_modify2d;

    BitOStream                  bos;
    std::ostringstream          text_output;
    TextWriter                  tw( text_output );
    Serializer::Description     desc;
    int                         i;

    ReplicaTime                 time1;
    ReplicaTime                 time2;
    ReplicaTime                 time3;
    ReplicaTime                 time4;
    ReplicaTime                 time5;
    ReplicaTime                 time6;

    ObjectManager &             object_manager = Global::object_manager();
    ReplicaServer &             replica_server = *object_manager.replica_server;

    // Write
    // ~~~~~

    // When sending delta-updates of arrays, we assume that the (virtual) client
    // has acked the most recent update of the array. This should result in
    // the most optimal updates. The times written to desc.ack_time could be
    // lower, but higher times would lead to corruption.

    // Initialize array (tick 1).

    desc.compare_times = true;
    desc.ack_time = ReplicaTime(); // client does not have the array
    desc.current_time = time1 = replica_server.get_replica_time();

    for( i = 0; i < 101; i++ )
        {
        array.push_back( PInt32( i ) );
        }

    after_init = array;

    array.bitstr_write( bos, desc );
    array.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Modify array (tick 2).

    desc.ack_time = time1; // client acked first update
    desc.current_time = time2 = replica_server.get_replica_time();

    array[ 50 ] = 0;
    array[ 51 ] = 0;
    array[ 58 ] = 0;
    array[ 99 ] = 0;

    first_modify = array;

    array.bitstr_write( bos, desc );
    array.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Modify array (tick 3).

    desc.ack_time = time2; // client acked second update
    desc.current_time = time3 = replica_server.get_replica_time();

    array[ 78 ] = 0;

    second_modify = array;

    array.bitstr_write( bos, desc );
    array.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Initialize 2D array (tick 4).

    desc.ack_time = ReplicaTime(); // client does not have array2d
    desc.current_time = time4 = replica_server.get_replica_time();

    array2d.resize( 4 );

    array2d[ 0 ].resize( 4, PString( "even row" ) );
    array2d[ 1 ].resize( 4, PString( "odd row" ) );
    array2d[ 2 ] = array2d[ 0 ];
    array2d[ 3 ] = array2d[ 1 ];

    after_init2d = array2d;

    array2d.bitstr_write( bos, desc );
    array2d.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Modify 2D array (tick 5).

    desc.ack_time = time4; // client acked first update of array2d
    desc.current_time = time5 = replica_server.get_replica_time();

    array2d[ 2 ][ 0 ] = PString( "first" );
    array2d[ 2 ][ 1 ] = PString( "second" );
    array2d[ 2 ][ 2 ] = PString( "third" );
    array2d[ 2 ][ 3 ] = PString( "fourth" );

    first_modify2d = array2d;

    array2d.bitstr_write( bos, desc );
    array2d.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Modify 2D array (tick 6).

    desc.ack_time = time5; // client acked second update of array2d
    desc.current_time = time6 = replica_server.get_replica_time();

    array2d[ 2 ][ 0 ]( 0 ) = 'F';

    second_modify2d = array2d;

    array2d.bitstr_write( bos, desc );
    array2d.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Read
    // ~~~~

    PArray< PInt32 >            bis_restored;
    PArray< PInt32 >            tr_restored;
    PArray< PArray< PString > > bis_restored2d;
    PArray< PArray< PString > > tr_restored2d;
    BitIStream                  bis;
    std::istringstream          text_input( text_output.str() );
    TextReader                  tr( text_input );

    bis.bind( bos.get_buffer() );

    // Integer array.

    desc.current_time = ReplicaTime();
    desc.ack_time = time1; // does not really matter, only info
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == after_init );

    desc.ack_time = time2;
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == first_modify );

    desc.ack_time = time3;
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == second_modify );

    // 2D string array.

    desc.ack_time = time4;
    bis_restored2d.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored2d.text_read( tr, desc );

    test_condition_( bis_restored2d == tr_restored2d );
    test_condition_( tr_restored2d == after_init2d );

    desc.ack_time = time5;
    bis_restored2d.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored2d.text_read( tr, desc );

    test_condition_( bis_restored2d == tr_restored2d );
    test_condition_( tr_restored2d == first_modify2d );

    desc.ack_time = time6;
    bis_restored2d.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored2d.text_read( tr, desc );

    test_condition_( bis_restored2d == tr_restored2d );
    test_condition_( tr_restored2d == second_modify2d );
    }


void SerializerTest::test_dictionary()
    {
    typedef PDictionary
        <
        PString,
        PString,
        ext::hash<std::string>,
        std::equal_to<std::string>
        >  StringToStringMap;

    StringToStringMap           map;
    StringToStringMap           after_init;
    StringToStringMap           first_modify;
    StringToStringMap           second_modify;

    StringToStringMap::Iterator it;

    BitOStream                  bos;
    std::ostringstream          text_output;
    TextWriter                  tw( text_output );
    Serializer::Description     desc;
    int                         i;
    
    ReplicaTime                 time1;
    ReplicaTime                 time2;
    ReplicaTime                 time3;

    ObjectManager &             object_manager = Global::object_manager();
    ReplicaServer &             replica_server = *object_manager.replica_server;
    
    // Write
    // ~~~~~

    // Initialize dictionary (tick 1).

    desc.compare_times = true;
    desc.ack_time = ReplicaTime();
    desc.current_time = time1 = replica_server.get_replica_time();

    for( i = 0; i < 20; i++ )
        {
        std::ostringstream ss;

        ss << "Item #" << i;
        map.insert( ss.str(), ss.str() );
        }

    after_init = map;

    map.bitstr_write( bos, desc );
    map.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();

    // Modify dictionary (tick 2).

    desc.ack_time = time1;
    desc.current_time = time2 = replica_server.get_replica_time();

    map.erase( PString( "Item #5" ) );
    map.insert( PString( "Item #10" ), PString( "Hello" ) );

    it = map.begin();

    for( i = 0; i < 17; i++ )
        {
        ++it;
        }

    it->second = PString( "The quick brown fox jumps over the lazy dog." );

    first_modify = map;

    map.bitstr_write( bos, desc );
    map.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();
    
    // Modify dictionary (tick 3).

    desc.ack_time = time2;
    desc.current_time = time3 = replica_server.get_replica_time();

    map.insert( PString( "Associate" ), PString( "This!" ) );
    second_modify = map;

    map.bitstr_write( bos, desc );
    map.text_write( tw, desc );
    tw.end_line();

    replica_server.advance_replica_time();
    
    // Read
    // ~~~~

    StringToStringMap                           bis_restored;
    StringToStringMap                           tr_restored;
    BitIStream                                  bis;
    std::istringstream                          text_input( text_output.str() );
    TextReader                                  tr( text_input );

    bis.bind( bos.get_buffer() );

    desc.current_time = ReplicaTime();
    desc.ack_time = time1;
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == after_init );

    desc.ack_time = time2;
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == first_modify );

    desc.ack_time = time3;
    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == second_modify );
    }


void SerializerTest::test_set()
    {
    typedef PSet
        <
        PString,
        ext::hash<std::string>,
        std::equal_to<std::string>
        >  StringSet;

    StringSet                   set;

    BitOStream                  bos;
    std::ostringstream          text_output;
    TextWriter                  tw( text_output );
    Serializer::Description     desc;
    int                         i;
    
    // Since both PDictionary & PSet share the same serialization mechanism,
    // no extra tests are neccessary. Unlike test_array() and test_dictionary()
    // this is "migration" test (compare_times is false).

    // Write
    // ~~~~~

    for( i = 0; i < 20; i++ )
        {
        std::ostringstream ss;

        ss << "Item #" << i;
        set.insert( ss.str() );
        }

    set.bitstr_write( bos, desc );
    set.text_write( tw, desc );
    tw.end_line();

    // Read
    // ~~~~
    
    StringSet               bis_restored;
    StringSet               tr_restored;
    BitIStream              bis;
    std::istringstream      text_input( text_output.str() );
    TextReader              tr( text_input );

    bis.bind( bos.get_buffer() );

    bis_restored.bitstr_read( bis, desc );
    test_condition_( tr.next_line() );
    tr_restored.text_read( tr, desc );

    test_condition_( bis_restored == tr_restored );
    test_condition_( tr_restored == set );
    }


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
