/** @file test_time.cpp
 * Time synchronization test.
 */

/*
 * begin        : Tue Feb 4 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "test/test_time.h"
#include "time/time_manager.h"
#include "thread/thread.h"
#include "thread/synch.h"
#include "time/system_time.h"
#include "global/global.h"
#include <iostream>
#include <sstream>
#include <cstdlib>  // for std::rand() function
#include <fstream>


using namespace Massiv::Core::Test;
using namespace Massiv::Core;


namespace Massiv {
namespace Core {
namespace Test {


struct TimeTest::SynchronizeThreadData
    {
    friend class TimeTest;

    SynchronizeThreadData
        (        
        TimeManagerArray *      tma,
        const SystemTime &      system_time,
        const NodeId &          nodeid,
        int                     num_ticks
        );

    TimeManagerArray *      time_manager_array;
        /**< Reference to TimeTest::time_manager_array. */
    SystemTime              init_time_value;
        /**< Initial system time to simulate eariler TimeManager start. */
    NodeId                  node_id;
        /**< NodeId to identify TimeManager in time_manager_array. */
    int                     num_ticks_alive;
        /**< How many ticks should TimeManager be alive. */

    }; // TimeTest::SynchronizeThreadData


CriticalSection *   TimeTest::tma_cs;
CriticalSection *   TimeTest::cout_cs;


#define TIME_TEST_BUF_SIZE 2

TimeTest::TimeTest() :
    Test( "time" )
    {
    CriticalSectionManager & csm = Global::critical_section_manager();
    tma_cs = csm.create();
    cout_cs = csm.create();
    }


TimeTest::~TimeTest()
    {
    delete tma_cs;
    delete cout_cs;
    }


void TimeTest::test_system_time()
    {
    SystemTime sys_time = SystemTime::current_time();
    Thread::sleep( 1000 );   // Waiting 1 second.
    sys_time = SystemTime::current_time() - sys_time;

    // Really should be less than 2s.
    test_condition_( ( sys_time > 0.5 ) && ( sys_time < 2 ) );
    }


int TimeTest::add_time_manager_to_array
    (
    TimeManagerArray &  tm_array,
    TimeManager *       time_manager
    )
    {
    tma_cs->enter();  // time_manager is one of pointers in time_manager_array.

    tm_array.push_back( time_manager );

    iter_tm iter = tm_array.begin();
    while( iter != tm_array.end() )
        {
        TimeManager * tm = *iter;
        if( tm != time_manager )
            {
            tm->lock();
            tm->add_node_for_synchronization( time_manager->my_node_id );
            time_manager->add_node_for_synchronization( tm->my_node_id );
            tm->unlock();
            }
        ++iter;
        }

    tma_cs->leave();

    return 0;
    }


int TimeTest::delete_time_manager_from_array
    (
    TimeManagerArray &  tm_array,
    TimeManager *       time_manager
    )
    {
    tma_cs->enter();

    iter_tm iter = tm_array.begin();
    while( iter != tm_array.end() )
        {
        TimeManager * tm = *iter;
        if( tm != time_manager )
            {
            tm->lock();
            tm->remove_node_from_synchronization( time_manager->my_node_id );
            tm->unlock();
            }
        ++iter;
        }

    tm_array.remove( time_manager );

    tma_cs->leave();

    return 0;
    }


int TimeTest::send_synchronization_info
    (
    TimeManagerArray &  tm_array,
    TimeManager *       time_manager
    )
    {
    TimeManager::SynchronizationInfo synch_info;  // my synchronization info

    time_manager->lock();
    synch_info = time_manager->get_local_synchronization_info();
    synch_info.massiv_time = time_manager->get_time();
    time_manager->unlock();

    // I need to simulate message delaying.
    // Simulate delay between 5 - 1000 ms.

    int random_number = std::rand();
    random_number %= 995;
    random_number += 5;
    Thread::sleep( random_number );

    tma_cs->enter();

    iter_tm iter = tm_array.begin();
    while( iter != tm_array.end() )
        {
        TimeManager * tm = *iter;
        if( tm != time_manager )
            {
            tm->lock();
            tm->set_remote_synchronization_info( time_manager->my_node_id, synch_info );
            tm->unlock();
            }
        ++iter;
        }

    tma_cs->leave();

    return 0;
    }


UInt32 TimeTest::get_node_id()
    {
    CriticalSectionManager & csm = Global::critical_section_manager();
    static CriticalSection *    ni_cs = csm.create();
    static UInt32               node_id = 1000;

    ni_cs->enter();
    UInt32 nid = node_id++;
    ni_cs->leave();

    return nid;
    }


int TimeTest::synchronize_thread
    (
    VariantPointer arg
    )
    {
    SynchronizeThreadData * synch_tdata =
        variant_cast<SynchronizeThreadData *>( arg );
    TimeManagerArray &      time_manager_array =
        *synch_tdata->time_manager_array;

    // I want to generate the same numbers. Set random number generator seed.
    srand( ( UInt32 )( synch_tdata->init_time_value.get_value() ) );

    TimeManager * tm = new TimeManager( synch_tdata->node_id, synch_tdata->init_time_value );

    // Time counting for this time manager initialized now.

    tm->allow_synchronization();
    TimeTest::add_time_manager_to_array( time_manager_array, tm );

    // Now other time threads and their time managers do know each other.

    cout_cs->enter();
    std::cout << "TimeManager( NodeId = " << synch_tdata->node_id << " ) created." << std::endl;
    cout_cs->leave();

    int num_ticks = synch_tdata->num_ticks_alive;  // Number of ticks of a TimeManager.
    bool write_once_synchronized = true;

    while( num_ticks-- )
        {
        tm->lock();
        tm->next_tick();

        // We can't send local synchronization information using tm->send_synchronize_info()
        // in this simulation.
        // Solution: instead of that we call TimeTest::send_synchronization_info()
        // every x ticks.
        tm->unlock();
        if( ( !tm->is_time_synchronized() ) ||   // Time was not synchronized yet.
            ( tm->should_synchronize_time() ) )
            {
            TimeTest::send_synchronization_info( time_manager_array, tm );
            tm->set_last_server_sync_system_time();
            }
        if( tm->is_time_synchronized() && write_once_synchronized )
            {
            write_once_synchronized = false;
            cout_cs->enter();
            std::cout << "TimeManager( NodeId = " << synch_tdata->node_id << " ) synchronized!!!" << std::endl;
            cout_cs->leave();
            }

        Thread::sleep( 0 );   // Leave processor.
        Thread::sleep( 40 );  // 25x/second
        }

    TimeTest::delete_time_manager_from_array( time_manager_array, tm );

    cout_cs->enter();
    std::cout << "TimeManager( NodeId = " << synch_tdata->node_id << " ) finished." << std::endl;
    cout_cs->leave();

    return 0;
    }


int TimeTest::synchronize_thread_screen_report
    (
    VariantPointer arg
    )
    {
    SynchronizeThreadData *     synch_tdata = variant_cast<SynchronizeThreadData *>( arg );
    TimeManagerArray &          time_manager_array = *synch_tdata->time_manager_array;

    std::ofstream               ofs( "synch.dat" );
    std::ostringstream          os;

    os << "Massiv time synchronization of particular TimeManagers (each TimeManager simulated by one thread)" << std::endl;
    os << "Hint(in czech): - kazdy radkovy blok je prehled casu vsech TimeManageru v jednom okamziku" << std::endl;
    os << "                - sloupce tvori jednotlive TimeManagery" << std::endl;
    os << "                - neni zaznam pro dany TimeManager => (vlakno) jeste neexistuje" << std::endl;
    os << "Struktura zaznamu:  101.084   - aktualni 'massiv time' daneho TimeManageru" << std::endl;
    os << "(priklad)           0.015     - aktualni 'increment' k soucasnemu massiv casu" << std::endl;
    os << "                    10       - aktualni 'sync_interval'" << std::endl;
    os << "**1\t\t2\t\t3\t\t4\t\t5**" << std::endl
       << "------------------------------------------------------------------------"
       << std::endl;

    ofs << os.str();
    ofs << std::endl;

    // Wait for the first TimeManager.

    while( time_manager_array.empty() )
        {
        Thread::sleep( 100 );
        }

    while( !time_manager_array.empty() )
        {
        os.str( "" );  // temporary

        tma_cs->enter();

        iter_tm itt = time_manager_array.begin();
        while( itt != time_manager_array.end() )
            {
            TimeManager * tm = *itt;
            tm->lock();
            os << tm->get_time().get_value();
            tm->unlock();
            os << "\t\t";
            ++itt;
            }
        os << std::endl;

        itt = time_manager_array.begin();
        while( itt != time_manager_array.end() )
            {
            TimeManager * tm = *itt;
            tm->lock();
            TimeManager::SynchronizationInfo si = tm->get_local_synchronization_info();
            tm->unlock();
            os << "" << si.increment;
            os << "\t\t";
            ++itt;
            }
        os << std::endl;

        itt = time_manager_array.begin();
        while( itt != time_manager_array.end() )
            {
            TimeManager * tm = *itt;
            tm->lock();
            TimeManager::SynchronizationInfo si = tm->get_local_synchronization_info();
            tm->unlock();
            os << "" << ( int )tm->time_sync_interval.get_value();
            os << "\t\t";
            ++itt;
            }
        os << std::endl;

        tma_cs->leave();

        cout_cs->enter();
        std::cout << os.str() << std::endl;
        cout_cs->leave();

        ofs << os.str();
        ofs << std::endl;

        Thread::sleep( 1000 );
        }
    return 0;
    }


void TimeTest::test_time_synchronization_algorithm()
    {
    cout_cs->enter();
    std::cout
        << "Time synchronization: See file 'synch.dat' for synchronization report."
        << std::endl;
    cout_cs->leave();

    SynchronizeThreadData * synch_screendata =
        new SynchronizeThreadData
                (
                &time_manager_array,
                SystemTime( 0 ),
                NodeId(),
                0
                );

    ThreadManager & tm = Global::thread_manager();
    Reference<Thread> synchro_screen = tm.create
        (
        TimeTest::synchronize_thread_screen_report,
        VariantPointer( synch_screendata )
        );

    typedef     std::pair< int, int > _p;
    // Initial values for time manager threads. first - system time, second - live ticks.
    typedef     std::vector< _p > _v;
    _v v;
    
    std::cout << "Sleeping 10 seconds..." << "\n";
    Thread::sleep( 10000 ); //sleep 10 seconds - to enable zero time for time managers
    v.push_back( _p( 5, 5000 ) );
    v.push_back( _p( 8, 5000 ) );
    v.push_back( _p( 7, 5000 ) );
    v.push_back( _p( 2, 5000 ) );
    v.push_back( _p( 9, 5000 ) );
    v.push_back( _p( 5, 5000 ) );

    typedef _v::iterator iter_v;

    int     NUM_THREADS = 4;
    iter_v  iti = v.begin();
    int     i;
    for( i = 0; i < NUM_THREADS; i++)
        {
        if( iti == v.end() )
            {
            iti = v.begin();
            }

        SynchronizeThreadData * synch_tdata1 =
            new SynchronizeThreadData
                    (
                    &time_manager_array,
                    SystemTime( iti->first ),
                    NodeId
                        (
                        NodeId::create_from_type_and_relative_id
                            (
                            NodeId::TYPE_SERVER,
                            get_node_id()
                            )
                        ),
                    iti->second
                    );
        ++iti;

        Reference<Thread> synchro = tm.create
            (
            TimeTest::synchronize_thread,
            VariantPointer( synch_tdata1 )
            );

        //Thread::sleep( 10000 );
        }

    synchro_screen->join();
    delete synch_screendata;
    // We are not waiting for other threads.
    }


void TimeTest::test()
    {
    test_system_time();
    test_time_manager();
    test_time_synchronization_algorithm();
    }


TimeTest::SynchronizeThreadData::SynchronizeThreadData
    (
    TimeManagerArray *      tma,
    const SystemTime &      system_time,
    const NodeId &          nodeid,
    int                     num_ticks
    ) :
    time_manager_array( tma ),
    init_time_value( system_time ),
    node_id( nodeid ),
    num_ticks_alive( num_ticks )
    {
    }


void TimeTest::test_time_manager()
    {
    // Test global time manager (like in real simulation).
    TimeManager &    time_manager = Global::time_manager();

    // Some functionality tests.
    STime       time1;
    STime       time2;
    STime       time3;
    time_manager.set_time( 123.456 ); // Reinitialize. New time will be 123.456 s.
    test_condition_( time_manager.is_paused() );

    time1 = time_manager.get_time();

    Thread::sleep( 500 );
    time_manager.next_tick();
    time2 = time_manager.get_time();
    test_condition_( time1 == time2 );

    try
        {
        time_manager.pause(); // Should throw exception.
        test_condition_( 0 );
        }
    catch( ... )
        {
        test_condition_( 1 );
        }

    time_manager.unpause();
    test_condition_( !time_manager.is_paused() );

    Thread::sleep( 500 );
    time2 = time_manager.get_time();
    test_condition_( time1 == time2 ); // Time is still the same.

    Thread::sleep( 500 );
    time_manager.next_tick();
    time2 = time_manager.get_time();
    test_condition_( time1 <= time2 );

    time_manager.pause();
    test_condition_( time_manager.is_paused() );
    time2 = time_manager.get_time();

    Thread::sleep( 500 );
    time_manager.next_tick();
    time1 = time_manager.get_time();
    test_condition_( time1 == time2 );

    time_manager.unpause();

    Thread::sleep( 500 );
    time2 = time_manager.get_time();
    test_condition_( time1 == time2 );
    time_manager.next_tick();
    time1 = time_manager.get_time();
    test_condition_( time1 >= time2 );

    time_manager.next_tick();
    time1 = time_manager.get_time();
    Thread::sleep( 500 );
    time2 = time_manager.get_time();
    test_condition_( time1 == time2 );

    // Start time synchronization with virtual massiv nodes.
    time_manager.set_time( 0 ); // Reinitialize and pause.
    // Next lines of code are commented because node manager throws exception when
    // time synchronization message is sent to not connected nodes.
    // But initial time synchronization process should look like this.
    /*
    time_manager.add_node_for_synchronization( NodeId( 123 ) );
    time_manager.add_node_for_synchronization( NodeId( 124 ) );
    time_manager.allow_synchronization();

    while( !time_manager.is_time_synchronized() )
        {
        time_manager.next_tick();
        }

    time_manager.unpause(); // Run time counting.
    */

    time_manager.set_time( 0 ); // Reinitialize.
    }


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
