/** @file test_vector.cpp
 * Testing of Vector23.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "test/test_vector.h"
#include "serializer/stype_vector.h"
#include "property/prop_vector.h"
#include <sstream>


namespace Massiv {
namespace Core {
namespace Test {


VectorTest::VectorTest() :
    Test( "vector23" )
    {
    }


VectorTest::~VectorTest()
    {
    }


void VectorTest::test()
    {
    SVector2 v2_1;
    PVector2 v2_2( 2, -1 );

    // -=
    v2_1 -= v2_2;
    test_condition_( v2_1 == PVector2( -2, 1 ) );
    v2_2 = v2_1;
    test_condition_( v2_2 == v2_1 );

    // +=
    SVector3 v3_1( 1, 1, 1 );
    SVector3 v3_2( 2, -1, 3 );
    v3_1 += v3_2;
    test_condition_( v3_1 == SVector3( 3, 0, 4 ) );

    PVector3 pv1( 1, 1, 2 );
    SVector3 pv2( 1, 2, 4 );
    PVector3 pv3;

    pv1 += pv2;
    test_condition_( pv1 == PVector3( 2, 3, 6 ) );

    // *=
    pv1 *= 2;
    test_condition_( pv1 == PVector3( 4, 6, 12 ) );

    // /=
    pv1 /= -2;
    test_condition_( pv1 == PVector3( -2, -3, -6 ) );

    // unary +
    pv2 = pv1;
    pv1 = +pv2;
    test_condition_( pv1 == pv2 );

    // unary -
    pv2 = PVector3 ( 0, 0, 0) - pv1;
    pv1 = -pv1;
    test_condition_( pv1 == pv2 );

    // operator+
    pv3 += pv1;
    pv3 += pv2;
    pv1 = pv1 + pv2;
    test_condition_( pv1 == pv3 );
    test_condition_( pv1 != pv2 );

    // operator-
    pv3 = pv1;
    pv3 -= pv2;
    pv1 = pv1 - pv2;
    test_condition_( pv1 == pv3 );

    // operator*
    pv3 = pv1 + pv1 + pv1 + pv1;
    Float32 sc = 2;
    pv1 = pv1 * sc + pv1 * sc;
    test_condition_( pv1 == pv3 );

    // operator/
    pv3 = pv1;
    pv3 /= sc;
    pv1 = pv1 / sc;
    test_condition_( pv1 == pv3 );

    // square magnitude.
    PVector3 pv4( 2, 3, -4 );
    Float32 q = pv4.square_magnitude();
    test_condition_( q == 29 );

    // magnitude.
    PVector2 pv5( -4, 3 );
    q = ( float )pv5.magnitude();
    test_condition_( q == 5 );

    // dot_product.
    q = SVector2::dot_product( SVector2( 2, 3 ), PVector2( -2, 4 ) ) ;
    test_condition_( q == 8 );
    test_condition_( SVector3( 2, 3, 4 ) * SVector3( 4, 3, 2 ) == ( 8 + 9 + 8 ) );

    // cross_product.
    PVector3 cp;
    cp.cross_product( SVector3( 1, 0, 0 ), PVector3( 0, 1, 0 ) );
    test_condition_( cp == SVector3( 0, 0, 1 ) );
    std::string str = cp.to_string();
    test_condition_( str == "( 0, 0, 1 )" );
    test_condition_( SVector3( 3, 0, 0 ) % SVector3( 4, 0, 0 ) == SVector3( 0, 0, 0 ) );
    test_condition_( SVector3( 3, 0, 0 ) % SVector3( 0, 4, 0 ) == SVector3( 0, 0, 12 ) );
    test_condition_( SVector3( 3, 0, 0 ) % SVector3( 0, 0, 4 ) == SVector3( 0, -12, 0 ) );
    test_condition_
        (
        SVector3( 2, 3, 4 ) %
        SVector3( 3, 2, 1 ) ==
        SVector3
            (
            ( 3 * 1 ) - ( 4 * 2 ),
            ( 4 * 3 ) - ( 2 * 1 ),
            ( 2 * 2 ) - ( 3 * 3 )
            )
        );

    // <
    PVector3 xx1( 1, 2, 3 );
    SVector3 xx2( 4, -5, -6 );
    test_condition_( xx1 < xx2 );
    test_condition_( xx1 <= xx2 );

    // greater
    PVector2 xxx1( 1, 2 );
    SVector2 xxx2( 3, 4 );
    test_condition_( xxx2.greater( xxx1 ) );

    xxx2.set_zero();
    test_condition_( xxx2.is_zero() );

    // Normalization.

    SVector3 normalized = SVector3( 345.6f, -234.8f, 0.98478f ).normalize();
    test_condition_( normalized.is_normalized() );
    test_condition_( FloatTraits< Float32 >::epsilon_equals( normalized.magnitude(), 1.0f ) );
    }


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
