/** @file test_volume_enumeration.cpp
 * Test of interface for enumeration of files and directories in
 * Massiv::Core::BasicVolume filesystem. These tests are written
 * separately from normal filesystem enumerator tests, although
 * both of them are quite similar. Otherwise the tests would be
 * very opaque.
 */

/*
 * begin        : Wed Feb 26 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "filesys/file_system.h"
#include "test/test_filesys.h"
#include "test/test_volume_enumeration.h"
#include "global/global.h"
#include <iostream>
#include <string>
#include <typeinfo>

namespace Massiv {
namespace Core {
namespace Test {


static const char * const test_hierarchy_files[] =
    {
    "tfile1.txt",
    "massiv.incl",
    "tdir1/tfile11.txt",
    "tdir2/tdir21/tfile211.txt",
    NULL
    };


VolumeEnumerationTest::VolumeEnumerationTest()
    :   Test( "volume file system enumeration"),
        native_fs( Global::native_file_system() ),
        old_current_directory( native_fs.get_current_directory() )
    {
    try
        {
        native_fs.set_current_directory
            (
            "../../../../../src/core/test/test_hierarchy"
            );

        volume = new Volume
            (
            old_current_directory + "test_volume_iterators.image",
            "VOLUME_TEST_ITERATORS"
            );

        // Copy 'test_hierarchy' to volume.
        const char * const * files;

        for( files = test_hierarchy_files; *files != NULL; files++ )
            {
            copy_to_volume( *volume, *files );
            }

        inited_successfuly = true;
        }
    catch( ExCanNotChangeDirectory )
        {
        std::cout << "Warning: VolumeEnumerationTest initialization failed"
                  << std::endl << "         Make sure you launch test_core "
                  << "from correct directory relatively " << std::endl
                  << "         to core sources" << std::endl;

        inited_successfuly = false;
        }
    }  // VolumeEnumerationTest::VolumeEnumerationTest()


VolumeEnumerationTest::~VolumeEnumerationTest()
    {
    delete volume;
    native_fs.set_current_directory( old_current_directory );
    }  // VolumeEnumerationTest::~VolumeEnumerationTest()


void VolumeEnumerationTest::test()
    {
    try
        {
        if( inited_successfuly )
            {
            go_through_test();
            descent_ascent_test();
            enter_leave_test();
            is_in_root_test();
            misc_tests();
            pack_unpack_test();
            }
        }
    catch( ... )
        {
        // Set the current directory to the old one.
        
        native_fs.set_current_directory( old_current_directory );
        throw;
        }
    }  // VolumeEnumerationTest::test()


void VolumeEnumerationTest::go_through_test()
    {
    FileIterator it = volume->open_directory( "/", "t*" );
    FileIterator it_aux = volume->open_directory( "/", "t*" );

    test_condition_( it == it_aux );

    perform_go_through_test( it );
    test_condition_( it != it_aux );
    it.rewind();
    test_condition_( it == it_aux );

    perform_go_through_test( it );
    test_condition_( it != it_aux );
    it_aux.unbind();
    test_condition_( it == it_aux );    // Both iterators are now invalid.
    }  // VolumeEnumerationTest::go_through_test()


void VolumeEnumerationTest::perform_go_through_test
    (
    FileIterator    it
    )
    {
    bool    flags[ 3 ];
    bool    success = true;
    int     i;

    for( i = 0; i < 3; i++ )    flags[ i ] = false;

    test_condition_( it.is_in_root() );

    i = 0;
    while( it.is_valid() )
        {
        if( ( it.name() == "tdir1" ) && !flags[ i ] )
            {
            flags[ i++ ] = true;
            test_condition_( it.is_directory() );
            test_condition_( !it.is_file() );
            test_condition_( it.size() == 0 );
            }
        else
        if( ( it.name() == "tdir2" ) && !flags[ i ] )
            {
            flags[ i++ ] = true;
            test_condition_( it.is_directory() );
            test_condition_( !it.is_file() );
            }
        else
        if( ( it.name() == "tfile1.txt" ) && !flags[ i ] )
            {
            flags[ i++ ] = true;
            test_condition_( it.is_file() );
            test_condition_( !it.is_directory() );
            test_condition_( it.size() == 1 );
            }
        else
            {
            success = false;
            break;
            }

        it++;
        }

    if( success )
        {
        for( i = 0; i < 3; i++ )
            {
            if( flags[ i ] == false )
                {
                success = false;
                break;
                }
            }
        }

    if( success )
        {
        test_condition_( true );
        }
    else
        {
        test_condition_( false );
        }
    }  // VolumeEnumerationTest::perform_go_through_test()


void VolumeEnumerationTest::descent_ascent_test()
    {
    FileIterator it;
    it = volume->open_directory( "/", "t*" );

    test_condition_( it.is_valid() );

    while( it.is_valid() && ( it.name() != "tdir2" ) )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( !it.is_directory() );
    test_condition_( it.is_file() );
    test_condition_( it.size() == 3 );
    test_condition_( it.name() == "tfile211.txt" );

    it.ascent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    it.ascent();

    perform_go_through_test( it );
    it.rewind();

    while( it.is_valid() && !it.is_file() )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    try
        {
        it.descent();
        test_condition_( 0 );
        }
    catch( ExNotADirectory & )
        {
        test_condition_( 1 );
        }
    catch( ... )
        {
        test_condition_( 0 );
        }

    try
        {
        it.enter();
        test_condition_( 0 );
        }
    catch( ExNotADirectory & )
        {
        test_condition_( 1 );
        }
    catch( ... )
        {
        test_condition_( 0 );
        }
    }  // VolumeEnumerationTest::descent_ascent_test()


void VolumeEnumerationTest::enter_leave_test()
    {
    FileIterator it = volume->open_directory( "/", "t*" );

    test_condition_( it.is_valid() );

    while( it.is_valid() && ( it.name() != "tdir2" ) )
        {
        it++;
        }

    test_condition_( it.is_valid() );

    it.descent();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    FileIterator it_enter = it.enter();

    test_condition_( it.is_valid() );
    test_condition_( it.is_directory() );
    test_condition_( !it.is_file() );
    test_condition_( it.size() == 0 );
    test_condition_( it.name() == "tdir21" );

    test_condition_( it_enter.is_valid() );
    test_condition_( !it_enter.is_directory() );
    test_condition_( it_enter.is_file() );
    test_condition_( it_enter.size() == 3 );
    test_condition_( it_enter.name() == "tfile211.txt" );

    FileIterator it_leave = it_enter.leave();

    test_condition_( it_enter.is_valid() );
    test_condition_( !it_enter.is_directory() );
    test_condition_( it_enter.is_file() );
    test_condition_( it_enter.size() == 3 );
    test_condition_( it_enter.name() == "tfile211.txt" );

    test_condition_( it_leave.is_valid() );
    test_condition_( it_leave.is_directory() );
    test_condition_( !it_leave.is_file() );
    test_condition_( it_leave.size() == 0 );
    test_condition_( it_leave.name() == "tdir21" );
    }  // VolumeEnumerationTest::enter_leave_test()


void VolumeEnumerationTest::is_in_root_test()
    {
    FileIterator    it = volume->open_directory( "tdir2/tdir21", "t*" );
    std::string     old_dir = "";
    std::string     new_dir = it.get_absolute_directory();

    while( old_dir != new_dir )
        {
        it.ascent();
        old_dir = new_dir;
        new_dir = it.get_absolute_directory();
        }

    test_condition_( it.is_in_root() );

    Reference< BasicVolume::IteratorImpl > it_impl =
        reference_cast< BasicVolume::IteratorImpl >( it.get_impl() );

    test_condition_( it_impl->node == &volume->root_directory );
    }  // VolumeEnumerationTest::is_in_root_test()


void VolumeEnumerationTest::copy_to_volume
    (
    BasicVolume &           volume,
    const std::string &     file_name
    )
    {
    std::ifstream           input
        (
        file_name.c_str(),
        std::ios_base::in | std::ios_base::binary
        );

    BasicVolume::OFStream   output
        (
        volume,
        file_name,
        std::ios_base::out | std::ios_base::binary,
        0
        );

    test_condition_( input.good() );
    test_condition_( output.good() );

    output << input.rdbuf();
    test_condition_( output.good() );
    }  // VolumeEnumerationTest::copy_to_volume()


void VolumeEnumerationTest::misc_tests()
    {
    volume->set_current_directory( "/tdir2/tdir21" );
    FileIterator it = volume->open_current_directory( "t*" );
    int count = 0;

    while( it.is_valid() )
        {
        count++;
        test_condition_( it.name() == "tfile211.txt" );
        test_condition_( it.is_file() );
        it++;
        }

    test_condition_( count == 1 );

    volume->set_current_directory( "/" );
    test_condition_( volume->current_directory_node == &volume->root_directory );

    it = volume->open_directory( "/tdir2/tdir21", "t*" );

    count = 0;
    while( it.is_valid() )
        {
        count++;
        test_condition_( it.name() == "tfile211.txt" );
        test_condition_( it.is_file() );
        it++;
        }

    test_condition_( count == 1 );

    volume->set_current_directory( "tdir2" );
    it = volume->open_current_directory( "t*" );

    count = 0;
    while( it.is_valid() )
        {
        count++;
        test_condition_( it.name() == "tdir21" );
        it++;
        }

    test_condition_( count == 1 );

    volume->set_current_directory( ".." );
    it = volume->open_current_directory( "t*" );
    test_condition_( it.is_in_root() );
    }  // VolumeEnumerationTest::misc_tests()


void VolumeEnumerationTest::pack_unpack_test()
    {
    // Phase 1: Unpack volume into test_iterators/src, pack this into
    //          another volume and unpack it again into test_iterators/dest.

    native_fs.set_current_directory( old_current_directory );

    /*try
        {
        native_fs.delete_directory
            (
            "test_iterators/src",
            FileSystem::DELETE_ALL_RECURSIVE
            );
        }
    catch( ... )
        {
        }

    try
        {
        native_fs.delete_directory
            (
            "test_iterators/dest",
            FileSystem::DELETE_ALL_RECURSIVE
            );
        }
    catch( ... )
        {
        }*/

    native_fs.create_directory( "test_iterators/src" );
    native_fs.create_directory( "test_iterators/dest" );

    test_condition_( volume->get_stream_count() > 1 );
    volume->unpack( "test_iterators/src", "", "t*", true );
    test_condition_( volume->get_stream_count() == 1 );

    Volume  aux_volume
        (
        old_current_directory + "test_volume_iterators1.image",
        "VOLUME_TEST_ITERATORS1"
        );

    aux_volume.pack( "", "test_iterators/src", "*", 0 );
    aux_volume.unpack( "test_iterators/dest", "", "*" );

    // Phase 2: Check whether there are correct files in test_iterators/dest.
    
    FileIterator it = native_fs.open_directory( "test_iterators/dest", "*" );

    while( it.is_valid() && !it.is_file() )
        {
        it++;
        }

    test_condition_( it.is_valid() && it.is_file() );
    test_condition_( it.name() == "tfile1.txt" );
    test_condition_( it.size() == 1 );

    it = native_fs.open_directory( "test_iterators/dest/tdir1", "*" );

    while( it.is_valid() && !it.is_file() )
        {
        it++;
        }

    test_condition_( it.is_valid() && it.is_file() );
    test_condition_( it.name() == "tfile11.txt" );
    test_condition_( it.size() == 2 );

    it = native_fs.open_directory( "test_iterators/dest/tdir2/tdir21", "*" );

    while( it.is_valid() && !it.is_file() )
        {
        it++;
        }

    test_condition_( it.is_valid() && it.is_file() );
    test_condition_( it.name() == "tfile211.txt" );
    test_condition_( it.size() == 3 );
    }  // VolumeEnumerationTest::pack_unpack_test()


#undef SOURCE_TREE_DIRECTORY


} // namespace Massiv::Core::Test
} // namespace Massiv::Core
} // namespace Massiv
