/** @file test_volume_manager.cpp
 * Test of volume manager.
 */

/*
 * begin        : Sun 22 Jun 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "test/test_volume_manager.h"
#include "filesys/volume_manager.h"
#include "filesys/file_system.h"
#include "status/exception.h"
#include <iostream>
#include <sstream>
#include <string>

namespace Massiv {
namespace Core {
namespace Test {


VolumeManagerTest::VolumeManagerTest()
    : Test( "volume manager" )
    {
    }


VolumeManagerTest::~VolumeManagerTest()
    {
    }


void VolumeManagerTest::test()
    {
    VolumeManager::format_volume( "volume1.image", "volume1" );
    VolumeManager::format_volume( "volume2.image", "volume2" );
    VolumeManager::format_volume( "volume3.image", "volume3" );

    // Explicit qualification test.

        {
        VolumeManager::OFStream stream;

        stream.open( "volume1://my_file1.txt" );
        stream << "ahoj1";
        stream.close();
        
        stream.open( "volume2://my_file2.txt" );
        stream << "ahoj2";
        stream.close();
        
        stream.open( "native://my_file3.txt" );
        stream << "ahoj3";
        stream.close();
        }

        {
        VolumeManager::IFStream stream;
        std::string s;

        stream.open( "my_file1.txt" );
        s.clear();
        stream >> s;
        stream.close();
        test_condition_( s == "ahoj1" );

        stream.open( "my_file2.txt" );
        s.clear();
        stream >> s;
        stream.close();
        test_condition_( s == "ahoj2" );

        stream.open( "my_file3.txt" );
        s.clear();
        stream >> s;
        stream.close();
        test_condition_( s == "ahoj3" );

        try
            {
            stream.open( "undefined://my_stream.txt" );
            stream.close();
            test_condition_( false );
            }
        catch( ExVolumeNotMounted & )
            {
            test_condition_( !stream.is_open() );
            test_condition_( true );
            }                           
        }

    // Stream lookup test.

        {
        VolumeManager::OFStream stream;

        stream.open( "volume1://file.txt" );
        stream << "volume1";
        stream.close();

        stream.open( "native://file.txt" );
        stream << "native";
        stream.close();
        
        stream.open( "volume1://directory/stream1.txt" );
        stream << "volume1";
        stream.close();
        
        stream.open( "directory/stream2.txt" );
        stream << "should_be_created_on_volume1";
        stream.close();

        stream.open( "file.txt" );
        stream << "native!";
        stream.close();
        }

        {
        VolumeManager::IFStream stream;
        std::string s;

        stream.open( "file.txt" );
        s.clear();
        stream >> s;
        stream.close();
        test_condition_( s == "native!" );
        
        stream.open( "volume1://directory/stream2.txt" );
        s.clear();
        stream >> s;
        stream.close();
        test_condition_( s == "should_be_created_on_volume1" );
        }

    // Iterator test.
        
        {
        VolumeManager::OFStream stream1( "volume1://images/vtest_volume1/vtest_nice1.txt" );
        VolumeManager::OFStream stream2( "volume2://images/vtest_volume2/vtest_nice2.txt" );
        VolumeManager::OFStream stream3( "volume3://images/vtest_volume3/vtest_nice3.txt" );
        VolumeManager::OFStream stream4( "native://images/vtest_native/vtest_nice4.txt" );
        }

        {
        FileSystem &    fs = VolumeManager::file_system();
        FileIterator    it;
        
        it = fs.open_directory( "images", "vtest_*" );
        test_condition_( it.name() == "vtest_native" );

        it.descent();
        test_condition_( it.name() == "vtest_nice4.txt" );
        it.ascent();

        test_condition_( it.name() == "vtest_native" );
        it.next();
        test_condition_( it.name() == "vtest_volume1" );
        
        it.descent();
        test_condition_( it.name() == "vtest_nice1.txt" );
        it.ascent();
        
        test_condition_( it.name() == "vtest_native" );
        it.next();
        test_condition_( it.name() == "vtest_volume1" );
        it.next();
        test_condition_( it.name() == "vtest_volume2" );
        
        it.descent();
        test_condition_( it.name() == "vtest_nice2.txt" );
        it.ascent();
        
        test_condition_( it.name() == "vtest_native" );
        it.next();
        test_condition_( it.name() == "vtest_volume1" );
        it.next();
        test_condition_( it.name() == "vtest_volume2" );
        it.next();
        test_condition_( it.name() == "vtest_volume3" );
        
        it.descent();
        test_condition_( it.name() == "vtest_nice3.txt" );
        it.ascent();
        
        test_condition_( it.name() == "vtest_native" );
        it.next();
        test_condition_( it.name() == "vtest_volume1" );
        it.next();
        test_condition_( it.name() == "vtest_volume2" );
        it.next();
        test_condition_( it.name() == "vtest_volume3" );
        it.next();
        test_condition_( !it.is_valid() );
        }

    // Done.

    VolumeManager::flush();
    }


}  // namespace Test
}  // namespace Core
}  // namespace Massiv
