/** @file stack_tracer_unix.cpp
 * Stack trace implementation for Unix.
 * @see stack_tracer.h
 */

/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Steapn Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "thread/stack_tracer.h"
#include <iostream>


#ifdef SYSTEM_LINUX


/** @def DEMANGLE_BACKTRACE
 * Try to demangle strings provided by backtrace_symbols and transform them
 * to nicer messages.
 */
#define DEMANGLE_BACKTRACE


#include <execinfo.h>

#ifdef DEMANGLE_BACKTRACE
#include <cxxabi.h>
#endif


namespace Massiv {
namespace Core {

/** 
 * %Stack tracer implementation for glibc-based systems.
 */
class StackTracerGlibc : public StackTracer
    {
public:

    /**
     * Dumps the current thread's stack.
     *
     * @param stream The stream the dump will be written into.
     * @todo C++ symbol demangling.
     */
    virtual void dump_stack
        (
        std::ostream & stream
        )
        {
        const std::size_t trace_limit = 50;
        void * backtrace_array[ trace_limit ];
        std::size_t size = ::backtrace( backtrace_array, trace_limit );
        char ** strings = ::backtrace_symbols( backtrace_array, size );
        if( strings )
            {
            // Skip first entry (self) in stack backtrace.
            
            bool first = true;
            std::size_t i;
            for( i = 1; i < size; i++ )
                {
                if( first )
                    {
                    first = false;
                    }
                else
                    {
                    stream << "\n";
                    }
                stream << i << ": ";
                #ifdef DEMANGLE_BACKTRACE

                // Find string components. Expected string syntax:
                // libname(funcname+offset) [address]

                const std::string mangled( strings[ i ] );
                const std::string::size_type lib_end = mangled.find_first_of( "(" );
                std::string::size_type func_begin = std::string::npos;
                std::string::size_type func_end = std::string::npos;
                std::string::size_type offs_begin = std::string::npos;
                std::string::size_type offs_end = std::string::npos;
                std::string::size_type rest_begin = std::string::npos;
                if( lib_end != std::string::npos )
                    {
                    func_begin = lib_end + 1;
                    func_end = mangled.find_first_of( ")+", func_begin );
                    }
                if( func_end != std::string::npos )
                    {
                    if( mangled[ func_end ] == '+' )
                        {
                        offs_begin = func_end + 1;
                        offs_end = mangled.find_first_of( ")", offs_begin );
                        }
                    if( offs_end == std::string::npos )
                        {
                        rest_begin = func_end + 1;
                        }
                    else
                        {
                        rest_begin = offs_end + 1;
                        }
                    }

                // If library and function names have been found, print nice output.
                // Otherwise print original mangled string.

                if( lib_end != std::string::npos && func_end != std::string::npos )
                    {
                    // Library name.

                    stream << std::string( mangled, 0, lib_end ) << ": ";

                    // Demangle function name.
                    // Note: For some reason __cxa__demangle demangled "main" to "unsigned long".
                    //       So use it only for symbols which look like C++ mangled symbols.

                    std::string mangled_func( mangled, func_begin, func_end - func_begin );
                    char * demangled_func = NULL;
                    int status = -1;
                    if( mangled_func.size() > 2 && mangled_func[ 0 ] == '_' && mangled_func[ 1 ] == 'Z' )
                        {
                        demangled_func = abi::__cxa_demangle( mangled_func.c_str(), NULL, NULL, &status );
                        }
                    if( status == 0 )
                        {
                        assert( demangled_func ); // intentionally assert() instead of hope()
                        stream << demangled_func;
                        free( demangled_func );
                        }
                    else
                        {
                        assert( !demangled_func ); // intentionally assert() instead of hope()
                        stream << mangled_func;
                        }

                    // Optional offset.

                    if( offs_end != std::string::npos )
                        {
                        stream << " + " << std::string( mangled, offs_begin, offs_end - offs_begin );
                        }

                    // Rest of mangled string.

                    if( rest_begin != std::string::npos )
                        {
                        stream << " " << std::string( mangled, rest_begin );
                        }
                    }
                else
                    {
                    stream << mangled;
                    }

                #else // DEMANGLE_BACKTRACE

                stream << strings[ i ];

                #endif // DEMANGLE_BACKTRACE
                }
            free( strings );

            if( size == trace_limit )
                {
                stream << "\n... more entries follow ...";
                }
            }
        }

    /** Installs our signal handler for the current thread. */
    virtual void install_parachute()
        {
        // Do nothing. Default singal handlers are good enough for us.
        }

    }; // class StackTracerGlibc


StackTracer * StackTracer::create()
    {
    return new StackTracerGlibc();
    }

} // namespace Massiv::Core
} // namespace Massiv


#else // system

namespace Massiv {
namespace Core {

// Not implemented on non-glibc systems.

StackTracer * StackTracer::create()
    {
    return NULL;
    }

} // namespace Massiv::Core
} // namespace Massiv

#endif // system

