/** @file synch_unix.cpp
 * Unix-specific critical section class.
 * @see synch_unix.h
 */

/* 
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "status/exception.h"
#include "thread/arch/unix/synch_unix.h"
#include <pthread.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <cerrno>
#include <iostream>
#include <sstream>

namespace Massiv {
namespace Core {


// Define union semun if not defined in sys/sem.h
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#if !defined( __GNU_LIBRARY__ ) || defined( _SEM_SEMUN_UNDEFINED )
union semun
    {
    int                    val;
    struct semid_ds *      buf;
    unsigned short int *   array;
    struct seminfo *       __buf;
    };
#endif


// Error messages
// ~~~~~~~~~~~~~~
static const char * err_cannot_create  = 
    "Couldn't create critical section; errno ";

static const char * err_cannot_destroy = 
    "Couldn't destroy critical section; errno ";

static const char * err_cannot_enter   =
    "Couldn't enter critical section; errno ";

static const char * err_cannot_leave   =
    "Couldn't leave critical section; errno ";


// Implementation of CriticalSectionManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CriticalSection * CriticalSectionManager::create
    (
    bool            owner,
    const char *    name
    )
    {
    return new CriticalSectionUnix( owner, name );
    }


// Implementation of CriticalSectionUnix
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CriticalSectionUnix::CriticalSectionUnix
    (
    bool            owner,
    const char *    name
    )
    {
    int             key = 0;

    this->owner = owner;

    if( name )
        {
        // Construct the key of the semaphore from the name. The key is not
        // created using ftok(), because this would bring some restrictions
        // for name and would require creation of a temporary file.

        named = true;

        for( const char * pc = name; *pc != '\0'; pc++ )  key += *pc;
        if( key == IPC_PRIVATE )  key++;

        if( owner )
            {
            // If semaphore already existed, we must detect this and
            // change owner to false.
            semaphore_id = semget( ( key_t ) key, 1, 0666 | IPC_CREAT | IPC_EXCL );
            if( semaphore_id == -1 )
                {
                owner = false;
                this->owner = owner;
                semaphore_id = semget( ( key_t ) key, 1, 0666 );
                }
            }
        else
            {
            semaphore_id = semget( ( key_t ) key, 1, 0666 );
            }

        if( semaphore_id == -1 )
            {
            std::ostringstream str;
            str << err_cannot_create << errno;
            throw ExCriticalSectionCanNotCreate( str.str() );
            }
        else
        if( owner )
            {
            // Initialize the semaphore value
            union semun sem_union;
            sem_union.val = 1;
    
            if( semctl( semaphore_id, 0, SETVAL, sem_union ) == -1 )
                {
                std::ostringstream str;
                str << err_cannot_create << errno;
                throw ExCriticalSectionCanNotCreate( str.str() );
                }
            }
        }
    else
        {
        named = false;

        #ifdef __USE_UNIX98
        pthread_mutexattr_init( &attr );
        if( pthread_mutexattr_settype( &attr, PTHREAD_MUTEX_RECURSIVE ) != 0 )
            {
            std::ostringstream str;
            str << err_cannot_create << errno;
            throw ExCriticalSectionCanNotCreate( str.str() );
            }
        pthread_mutex_init( &mutex, &attr );
        #else
        pthread_mutex_init( &mutex, NULL );
        #endif
        }
    }  // CriticalSectionUnix::CriticalSectionUnix()


CriticalSectionUnix::~CriticalSectionUnix()
    {
    if( named )
        {
        if( owner )
            {
            union semun sem_union;
            hope( semaphore_id != -1 );
            if( semctl( semaphore_id, 0, IPC_RMID, sem_union ) == -1 )
                {
                std::ostringstream str;
                str << err_cannot_destroy << errno;
                throw ExCriticalSectionCanNotDestroy( str.str() );
                }
            }
        }
    else
        {
        if( pthread_mutex_destroy( &mutex ) != 0 )
            {
            std::ostringstream str;
            str << err_cannot_destroy << errno;
            throw ExCriticalSectionCanNotDestroy( str.str() );
            }
        if( pthread_mutexattr_destroy( &attr ) != 0 )
            {
            std::ostringstream str;
            str << err_cannot_destroy << errno;
            throw ExCriticalSectionCanNotDestroy( str.str() );
            }
        }
    }


void CriticalSectionUnix::enter()
    {
    if( named )
        {
        hope( semaphore_id != -1 );

        struct sembuf sem_buf;

        sem_buf.sem_num = 0;
        sem_buf.sem_op = -1;
        sem_buf.sem_flg = SEM_UNDO;

        if( semop( semaphore_id, &sem_buf, 1 ) == -1 )
            {
            std::ostringstream str;
            str << err_cannot_enter << errno;
            throw ExCriticalSectionCanNotEnter( str.str() );
            }
        }
    else
        {
        if( pthread_mutex_lock( &mutex ) != 0 )
            {
            std::ostringstream str;
            str << err_cannot_enter << errno;
            throw ExCriticalSectionCanNotEnter( str.str() );
            }
        }
    }  // CriticalSectionUnix::enter()


bool CriticalSectionUnix::try_enter()
    {
    if( named )
        {
        hope( semaphore_id != -1 );

        struct sembuf sem_buf;

        sem_buf.sem_num = 0;
        sem_buf.sem_op = -1;
        sem_buf.sem_flg = SEM_UNDO | IPC_NOWAIT;

        if( semop( semaphore_id, &sem_buf, 1 ) == -1 )
            {
            if( errno != EAGAIN )
                {
                std::ostringstream str;
                str << err_cannot_enter << errno;
                throw ExCriticalSectionCanNotEnter( str.str() );
                }
            else
                {
                return false;
                }
            }
        }
    else
        {
        // @todo: implementace pro nepojmenovany zamek
        throw ExCommonError( "try_lock() not implemented yet for unnamed locks." );
        }
    return true;
    }  // CriticalSectionUnix::try_enter()


void CriticalSectionUnix::leave()
    {
    if( named )
        {
        hope( semaphore_id != -1 );

        struct sembuf sem_buf;

        sem_buf.sem_num = 0;
        sem_buf.sem_op = 1;
        sem_buf.sem_flg = SEM_UNDO;

        if( semop( semaphore_id, &sem_buf, 1 ) == -1 )
           {
           std::ostringstream str;
           str << err_cannot_leave << errno;
           throw ExCriticalSectionCanNotLeave( str.str() );
           }
        }
    else
        {
        if( pthread_mutex_unlock( &mutex ) != 0 )
            {
            std::ostringstream str;
            str << err_cannot_leave << errno;
            throw ExCriticalSectionCanNotLeave( str.str() );
            }
        }
    }  // CriticalSectionUnix::leave()


} // namespace Massiv::Core
} // namespace Massiv
