/** @file thread_unix.cpp
 * Interface for Unix threads.
 * @see thread_unix.h
 */

/*
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "thread/arch/unix/thread_unix.h"
#include "status/exception.h"
#include "utility/debug.h"
#include "thread/thread.h"
#include "global/global.h"
#include <pthread.h>
#include <unistd.h>
#include <errno.h>

namespace Massiv {
namespace Core {


// Implementation of ThreadUnix
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ThreadUnix::ThreadUnix
    (
    ThreadRoutine   thread_routine,
    VariantPointer  arg
    ) : Thread( thread_routine, arg )
    {
    // Launch thread.

    switch( pthread_create( &thread_id, NULL, thread_launcher, this ) != 0 )
        {
        case 0:
            break;
        case EAGAIN:
            throw ExThreadCanNotCreate( "Cannot create thread; not enough system resources." );
        default:
            panic( "Unexpected value returned from pthread_create()." );
        }
    }


ThreadUnix::~ThreadUnix()
    {
    }


int ThreadUnix::join()
    {
    void * result;

    switch( pthread_join( thread_id, &result ) != 0 )
        {
        case 0:
            break;
        case ESRCH:
            throw ExThreadCanNotJoin( "Cannot join thread; wrong thread id." );
        case EINVAL:
            throw ExThreadCanNotJoin( "Cannot join thread; another thread has been detached." );
        case EDEADLK:
            throw ExThreadCanNotJoin( "Cannot join thread; thread id refers to the calling thread." );
        default:
            panic( "Unexpected value returned from pthread_join()." );
        }

    return reinterpret_cast<int>( result );
    }


void * ThreadUnix::thread_launcher
    (
    void * context
    )
    {
    ThreadUnix * this_thread = reinterpret_cast<ThreadUnix *>( context );
    void * exit_status;

    Global::thread_manager().thread_starting();
    exit_status = reinterpret_cast<void *>( this_thread->run() );
    Global::thread_manager().thread_stopping();

    return exit_status;
    }


// Implementation of Thread
// ~~~~~~~~~~~~~~~~~~~~~~~~

void Thread::exit
    (
    int exit_code
    )
    {
    Global::thread_manager().thread_stopping();
    pthread_exit( reinterpret_cast<void *>( exit_code ) );
    }


void Thread::sleep
    (
    int miliseconds
    )
    {
    ::usleep( miliseconds * 1000 );  // To get nanoseconds.
    }


UInt32 Thread::get_remaining_stack_size()
    {
    #if defined( SYSTEM_LINUX ) && defined( ARCH_X86 )

    // Not-so-portable stuff:
    // - phtread_getattr_np() is glibc-specific.
    // - stack growth direction is processor-specific (this is probably defined somewhere?).

    // Get information about current thread.

    pthread_attr_t attr;
    pthread_getattr_np( pthread_self(), &attr );
    void * stack_base;
    std::size_t stack_size;
    pthread_attr_getstack( &attr, &stack_base, &stack_size );

    // On x86, stack_base < stack_pointer.

    void * stack_pointer = &attr;
    return reinterpret_cast< UInt8 * >( stack_pointer ) -
           reinterpret_cast< UInt8 * >( stack_base );
    
    #elif defined( COMPILER_CYGWIN )

    // FIXME: This needs be implemented.

    return 256 * 1024;//~!

    #else

    #error "Need to implement this."

    #endif
    }


Thread::ThreadId Thread::get_current_thread_id()
    {
    ThreadId id = static_cast<ThreadId>( ::pthread_self() );
    hope( id != 0 );
    return id;
    }


Thread::ThreadIndex Thread::get_current_thread_index()
    {
    // TODO: Use TSD here.
    return Global::thread_manager().find_thread_index( get_current_thread_id() );
    }


// Implementation of ThreadManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reference<Thread> ThreadManager::create
    (
    Thread::ThreadRoutine   start_routine,
    VariantPointer          arg
    )
    {
    Thread * t;

    t = new ThreadUnix( start_routine, arg );

    // Wait until thread starts.
    // This ensures that Thread object (@c t) still exists
    // when the thread physically starts.

    launcher_semaphore->down();

    return t;
    }


} // namespace Massiv::Core
} // namespace Massiv
