/** @file stack_tracer_win32.cpp
 * Stack trace implementation (MSVC/Win32).
 * Based on a public domain "StackWalk" program
 * (http://www.mvps.org/win32/misc/stackwalk.html).
 */

/*
 * begin        : Sat Dec 7 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "thread/stack_tracer.h"
#include <iostream>
#include <iomanip>
#include <cstring>
#include <sstream>
#include <stdexcept>

#include <process.h>
#include <windows.h>
#include <winnt.h>
#include <imagehlp.h>


namespace Massiv {
namespace Core {


#if defined( COMPILER_MSVC )


// Definition of StackTracerMSVC
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * Stack tracer implementation for MSVC/Win32. Uses Microsoft proprietary helper
 * libraries to access symbol information and walk stack. Relies on MSVC-generated 
 * stack frames.
 * This should be written independently on other parts of the core.
 */
class StackTracerMSVC : public StackTracer
    {
public:

    enum Parameters
        {
        MAX_SYMBOL_NAME_LENGTH =    256,
            /**< Maximum symbol name length. */
        MAX_FRAMES =                256
            /**< Maximum number of frames to dump. */
        };

    /**
     * Earlier versions of the imagehlp.dll (for example that one shipped with
     * NT4) do not implement @c ::SymGetLineFromAddr(). To prevent link failures
     * at program start try to resolve the symbol dynamically using
     * @c ::GetProcAddress() and if that fails fallback to a simpler mode
     * that does not utilize this feature.
     * The signature was copied from the MS PSDK.
     */
    typedef BOOL ( __stdcall * SymGetLineFromAddrPtr )
        (
        IN  HANDLE          process_handle,
        IN  DWORD           address,
        OUT PDWORD          displacement,
        OUT PIMAGEHLP_LINE  line
        );

public:
            
    /**
     * Initializes image helper library and loads symbols.
     * Throws std::runtime_error exception if initialization failed.
     */
    StackTracerMSVC() :
        thread_ready( INVALID_HANDLE_VALUE ),
        request_ready( INVALID_HANDLE_VALUE ),
        thread_to_dump( INVALID_HANDLE_VALUE ),
        handling_exception( false ),
        dump_done( INVALID_HANDLE_VALUE ),
        stream_to_dump_to( NULL ),
        closing( false ),
        sym_get_line_from_addr( NULL ),
        symbol( NULL )
        {
        self = this;

        // Initialize symbol handler stuff and load symbols.
        // May not work under Windows9x.

        if( !::SymInitialize
            (
            ::GetCurrentProcess(),
            const_cast<char *>( get_symbol_search_paths().c_str() ),  
            true
            ) )
            {
            throw std::runtime_error( "SymInitialize() failed." );
            }

        if( !::SymSetOptions( SYMOPT_LOAD_LINES | SYMOPT_DEFERRED_LOADS ) )
            {
            throw std::runtime_error( "SymSetOptions() failed." );
            }

        // Resolve "extras" not implemented by all versions of the
        // image helper DLL.

        sym_get_line_from_addr = reinterpret_cast<SymGetLineFromAddrPtr>( 
            ::GetProcAddress( ::GetModuleHandle( "imagehlp.dll" ), 
            "SymGetLineFromAddr" ) );

        // Initialize buffers.
        
        symbol = reinterpret_cast<IMAGEHLP_SYMBOL *>( new char[ 
            sizeof( IMAGEHLP_SYMBOL ) + MAX_SYMBOL_NAME_LENGTH ] );
        
        memset( symbol, 0, sizeof( IMAGEHLP_SYMBOL ) + MAX_SYMBOL_NAME_LENGTH );
        symbol->SizeOfStruct = sizeof( IMAGEHLP_SYMBOL );
        symbol->MaxNameLength = MAX_SYMBOL_NAME_LENGTH;

        // Create stack dumper thread.

        thread_ready = ::CreateSemaphore( NULL, 0, 1, NULL );
        request_ready = ::CreateSemaphore( NULL, 0, 1, NULL );
        dump_done = ::CreateSemaphore( NULL, 0, 1, NULL );

        _beginthread( stack_dumper_thread, 0, this );
        }

    /** Unloads symbols and shutdowns. */
    ~StackTracerMSVC()
        {
        // Stop stack dumper thread.
        // Prepare stack dump request for that thread.

        ::WaitForSingleObject( thread_ready, INFINITE );
        closing = true;
        ::ReleaseSemaphore( request_ready, 1, NULL );

        // Free resources.

        ::SymCleanup( ::GetCurrentProcess() );
        ::CloseHandle( thread_ready );
        ::CloseHandle( request_ready );
        ::CloseHandle( dump_done );

        thread_ready = INVALID_HANDLE_VALUE;
        request_ready = INVALID_HANDLE_VALUE;
        dump_done = INVALID_HANDLE_VALUE;
        closing = false;

        sym_get_line_from_addr = NULL;
        
        delete [] reinterpret_cast<char *>( symbol );
        symbol = NULL;
        }

    /** Dumps current thread's stack. */
    virtual void dump_stack
        (
        std::ostream &  stream
        )
        {
        dump_stack( NULL, stream );
        }
    
    /** Hooks our Win32 SEH translator for the current thread if necessary. */
    void install_parachute()
        {
        #if ( _MSC_VER < 1400 )
        
        // On earlier versions of MSVC (prior to MSVC 8), Win32 SEH exceptions 
        // are by default mapped to untyped C++ exceptions that can be caught by 
        // the catch( ... ) construct. 
        // We do not like such a behaviour and want our program to crash 
        // immediatelly if a SEH exception occurs. To do that, we hook a SEH
        // translator. 
        
        ::_set_se_translator( win32_seh_translator );
        
        #else
        
        // Win32 SEH exceptions are not mapped to C++ exceptions anymore.
        // We can not use the translator under the synchronous C++ exception model
        // and so we are also unable to do a stack trace when the program crashes.
        // This could possibly be worked around though (for example by writing a 
        // signal handler)... FIXME 
        
        #endif
        }

private:
    
    /**
     * Thread that once asked dumps stack trace of another thread that
     * was suspended before.
     */
    static void __cdecl stack_dumper_thread
        (
        void * arg
        )
        {
        StackTracerMSVC * object = reinterpret_cast<StackTracerMSVC *>( arg );

        while( true )
            {
            // Make self ready.

            ::ReleaseSemaphore( object->thread_ready, 1, NULL );

            // Wait until request for stack dumper thread is ready.

            ::WaitForSingleObject( object->request_ready, INFINITE );
            
            if( object->closing )
                {
                break;
                }

            // Handle the request.

            if( !object->handling_exception )
                {
                // Dump the stack from thread's current context. We are assuming
                // that the last operation the thread has made was a call to
                // dump_stack().

                CONTEXT current_context;

                memset( &current_context, 0, sizeof( current_context ) );
                current_context.ContextFlags = CONTEXT_FULL;

                if( ::GetThreadContext( object->thread_to_dump, &current_context ) )
                    {
                    object->dump_stack
                        (
                        object->thread_to_dump,
                        current_context,
                        *object->stream_to_dump_to,
                        2
                        );
                    }
                }
            else
                {
                // Dump the stack from thread's specified specified context.
                // We are assuming that a SEH exception was thrown at this context
                // and the exception is currently being handled by that thread.
                // Stack dump will originate at the place of the exception.

                object->dump_stack
                    (
                    object->thread_to_dump,
                    object->exception_context,
                    *object->stream_to_dump_to,
                    0
                    );
                }
            
            // Wake up @c thread_to_dump.
            
            ::ReleaseSemaphore( object->dump_done, 1, NULL );
            ::CloseHandle( object->thread_to_dump );

            object->thread_to_dump = INVALID_HANDLE_VALUE;
            object->handling_exception = false;
            object->stream_to_dump_to = NULL;
            }
        }
    
    /**
     * Dumps current thread's stack starting from a specified context.
     * If @c context is NULL, the current context will be used. Otherwise
     * it is assumed that an exception is being handled now and the exception
     * was thrown at the specified @c context.
     */
    void dump_stack
        (
        CONTEXT *       context,
        std::ostream &  stream
        )
        {
        // Wait until stack dumper thread is ready.
        // Prepare stack dump request for that thread.

        ::WaitForSingleObject( thread_ready, INFINITE );
        assert( thread_to_dump == INVALID_HANDLE_VALUE ); // intentionally assert() instead of hope()
        ::DuplicateHandle
            (
            ::GetCurrentProcess(),
            ::GetCurrentThread(),
            ::GetCurrentProcess(),
            &thread_to_dump,
            0,
            false,
            DUPLICATE_SAME_ACCESS
            );
        
        if( context != NULL )
            {
            memcpy( &exception_context, context, sizeof( exception_context ) );
            handling_exception = true;
            }
        else
            {
            handling_exception = false;
            }

        stream_to_dump_to = &stream;
        
        // Wake up stack dumper thread, suspend self until dump is done.

        ::ReleaseSemaphore( request_ready, 1, NULL );
        ::WaitForSingleObject( dump_done, INFINITE );
        }
    
    /**
     * Dumps stack of a suspended thread identified by @c thread_handle to a
     * @c stream. The thread must be suspended in @c context or an exception
     * that was thrown from the @c context is currently being handled.
     * Must be properly synchronized (can not run in parallel).
     */
    void dump_stack
        (
        HANDLE          thread_handle,
        const CONTEXT & context,
        std::ostream &  stream,
        int             skip_first_n_frames
        )
        {
        CONTEXT             thread_context;
        STACKFRAME          stack_frame;
        int                 frame_number;
        IMAGEHLP_LINE       line;
        IMAGEHLP_MODULE     module;
        unsigned long       offset_from_symbol;

        // Initialize.

        thread_context = context;

        memset( &stack_frame, 0, sizeof( stack_frame ) );
        stack_frame.AddrPC.Offset = thread_context.Eip;
        stack_frame.AddrPC.Mode = AddrModeFlat;
        stack_frame.AddrFrame.Offset = thread_context.Ebp;
        stack_frame.AddrFrame.Mode = AddrModeFlat;
        
        frame_number = 0;

        memset( &line, 0, sizeof( line ) );
        line.SizeOfStruct = sizeof( line );

        memset( &module, 0, sizeof( module ) );
        module.SizeOfStruct = sizeof( module );
        
        // Walk through stack.

        while( true )
            {
            // Walk to previous stack entry.

            if( ++frame_number > MAX_FRAMES )
                {
                stream << "...";
                break;
                }

            if( !::StackWalk
                (
                #ifdef ARCH_X86
                IMAGE_FILE_MACHINE_I386,
                #else
                #error "Need to implement this."
                #endif    
                ::GetCurrentProcess(),
                thread_handle,
                &stack_frame,
                &thread_context,
                NULL,
                &::SymFunctionTableAccess,
                &::SymGetModuleBase,
                NULL
                ) )
                {
                break;
                }

            // Print it.

            if( frame_number > skip_first_n_frames )
                {
                stream << std::setiosflags( std::ios_base::left ) << std::setw( 3 ) 
                    << frame_number - skip_first_n_frames << ": ";
                
                if( stack_frame.AddrPC.Offset != 0 )
                    {
                    if( ::SymGetSymFromAddr
                        (
                        ::GetCurrentProcess(),
                        stack_frame.AddrPC.Offset,
                        &offset_from_symbol,
                        symbol
                        ) )
                        {
                        char                undecorated_name[ 1024 ];
                        std::ostringstream  symbol_stream;

                        if( ::UnDecorateSymbolName
                            (
                            symbol->Name,
                            undecorated_name,
                            sizeof( undecorated_name ),
                            UNDNAME_NAME_ONLY
                            ) )
                            {
                            symbol_stream << undecorated_name;
                            }

                        if( offset_from_symbol != 0 )
                            {
                            symbol_stream << " + 0x" << std::hex 
                                << static_cast<UInt32>( offset_from_symbol ) 
                                << std::dec;
                            }

                        stream << std::setiosflags( std::ios_base::left ) 
                            << std::setw( 40 ) << symbol_stream.str() 
                            << std::setw( 0 );
                        }
                    else
                        {
                        stream << "0x" << std::setw( 8 ) << std::setfill( '0' )
                            << std::hex << static_cast<UInt32>( 
                            stack_frame.AddrPC.Offset ) << std::dec
                            << std::setw( 30 ) << std::setfill( ' ' ) << ' ';
                        }

                    if( sym_get_line_from_addr != NULL && sym_get_line_from_addr( 
                        ::GetCurrentProcess(), stack_frame.AddrPC.Offset,
                        &offset_from_symbol, &line ) )
                        {
                        stream << " file: " << line.FileName << " (" 
                            << std::dec << line.LineNumber << ")";
                        }
                    else
                    if( ::SymGetModuleInfo
                        (
                        ::GetCurrentProcess(),
                        stack_frame.AddrPC.Offset,
                        &module
                        ) )
                        {
                        stream << " " << module.ModuleName 
                            << " + 0x" << std::hex << static_cast<UInt32>( 
                            stack_frame.AddrPC.Offset - module.BaseOfImage ) 
                            << std::dec;
                        }
                    }

                stream << '\n';
                }

            // Stack bottom?

            if( stack_frame.AddrReturn.Offset == 0 )
                {
                break;
                }
            }
        }

    /**
     * Our SEH translator function.          
     * Invoked by the C++ run-time when a SEH exception is raised by a thread 
     * that has got the translator registered. Called within the context of the
     * thread where the exception was raised.          
     */
    static void _cdecl win32_seh_translator
        (
        unsigned int        error_code,
        PEXCEPTION_POINTERS pointers
        )
        {
        PEXCEPTION_RECORD   exception = pointers->ExceptionRecord;
        PCONTEXT            context = pointers->ContextRecord;

        // Dump basic information.

        std::cerr << "Win32 SEH exception 0x" << std::hex << error_code
            << std::dec << " raised:";
            
        if( exception->NumberParameters )
            {
            unsigned int i;
            
            std::cerr << "\nParameters: ";

            for( i = 0; i < exception->NumberParameters; i++ )
                {
                if( i )
                    {
                    std::cerr << ", ";
                    }
                std::cerr << exception->ExceptionInformation[ i ];
                }
            }

        std::cerr << "\nTrying to do a stack-trace:\n";
        std::cerr.flush();

        // Try to do a stack trace.

        self->dump_stack( context, std::cerr );

        // Done.

        std::cerr.flush();
        std::abort();
        }

    /**
     * Returns list of paths where image helper library should look to
     * when trying to load symbols. Paths are separated by semicolons.
     */
    std::string get_symbol_search_paths() const
        {
        std::string paths;
        char        buffer[ 1024 ];
        int         size;

        // Add current directory.

        size = ::GetCurrentDirectory
            (
            sizeof( buffer ),
            buffer
            );

        if( size && size < sizeof( buffer ) )
            {
            paths += buffer;
            }

        // Add directory with executable.
        
        size = ::GetModuleFileName
            (
            NULL,
            buffer,
            sizeof( buffer )
            );

        if( size > 0 )
            {
            char * p = buffer + size - 1;
            char * prev_p = p;
                
            // Cut off module name, leave complete path.

            while( p >= buffer )
                {
                if( *p == '/' || *p == '\\' )
                    {
                    *p = '\0';
                    break;
                    }
                if( *p == ':' )
                    {
                    p = prev_p;
                    *p = '\0';
                    break;
                    }
                p--;
                }

            if( paths.size() )
                {
                paths += ';';
                }

            paths += buffer;
            }
        
        // Add content of selected environment variables.

        const char * variables[] =
            {
            "PATH",
            "SystemRoot",
            "_NT_SYMBOL_PATH",
            "_NT_ALTERNATE_SYMBOL_PATH"
            };
        int i;

        for( i = 0; i < sizeof( variables ) / sizeof( variables[ 0 ] ); i++ )
            {
            size = ::GetEnvironmentVariable
                (
                variables[ i ],
                buffer,
                sizeof( buffer )
                );

            if( size && size < sizeof( buffer ) )
                {
                if( paths.size() )
                    {
                    paths += ';';
                    }
                paths += buffer;
                }
            }

        // Done.

        return paths;
        }

private:
    
    HANDLE                      thread_ready;
        /**< Semaphore used to synchronize stack dumps.
             When signalled, thread is ready to accept new request. */
    HANDLE                      request_ready;
        /**< Semaphore used to synchronize stack dumps.
             When signalled, request for the stack dumper thread is ready. */
    HANDLE                      dump_done;
        /**< Semaphore used to synchronize stack dumps.
             When signalled, stack dump has finished. */
    HANDLE                      thread_to_dump;
        /**< Handle of the thread whose stack should be dumped. */
    CONTEXT                     exception_context;
        /**< If thread_to_dump is currently handling an exception,
             @c exception_context can hold context of the place where the
             exception was thrown so that the stack dump from this place could
             be done. */
    bool                        handling_exception;
        /**< Is @c thread_to_dump currently handling an exception that was raised
             in @c exception_context? */
    std::ostream *              stream_to_dump_to;
        /**< Stream to dump the stack to. */
    bool                        closing;
        /**< Should stack dumper thread be stopped? */
    SymGetLineFromAddrPtr       sym_get_line_from_addr;
        /**< Pointer to @c ::SymGetLineFromAddr(). NULL if not implemented
             by the helper DLL. */
    IMAGEHLP_SYMBOL *           symbol;
        /**< Preallocated "symbol" buffer. */
    static StackTracerMSVC *    self;
        /**< Pointer to self. */

    }; // StackTracerMSVC


StackTracerMSVC * StackTracerMSVC::self = NULL;


// Implementation of StackTracer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

StackTracer * StackTracer::create()
    {
    try
        {
        return new StackTracerMSVC();
        }
    catch( std::runtime_error & )
        {
        return NULL;
        }
    }


#else // !COMPILER_MSVC


// StackTracer have not been implemented for non-MSVC compilers under Win32.

StackTracer * StackTracer::create()
    {
    return NULL;
    }


#endif


} // namespace Massiv::Core
} // namespace Massiv
