/** @file synch_win32.cpp
 * Win32-specific critical section class.
 * @see synch_win32.h
 */

/* 
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista, Marek Svantner
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "utility/debug.h"
#include "status/exception.h"
#include "thread/arch/win32/synch_win32.h"
#include <iostream>
#include <sstream>


namespace Massiv {
namespace Core {


// Implementation of CriticalSectionManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CriticalSection * CriticalSectionManager::create
    (
    bool            owner,
    const char *    name
    )
    {
    return new CriticalSectionWin32( owner, name );
    }


// Implementation of CriticalSectionWin32
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

CriticalSectionWin32::CriticalSectionWin32
    (
    bool            owner,
    const char *    name
    )
    {
    if( !name || owner )
        {
        mutex = ::CreateMutex( NULL, FALSE, name );
        }
    else
        {
        mutex = ::OpenMutex( MUTEX_ALL_ACCESS, FALSE, name );
        }

    if( mutex == NULL )
        {
        std::ostringstream  str;
        str << "Couldn't create critical section: " << ::GetLastError();
        throw ExCriticalSectionCanNotCreate( str.str() );
        }
    }


CriticalSectionWin32::~CriticalSectionWin32()
    {
    if( ::CloseHandle( mutex ) == 0 )
        {
        std::ostringstream  str;
        str << "Couldn't destroy the critical section: " << ::GetLastError();
        throw ExCriticalSectionCanNotDestroy( str.str() );
        }
    }


void CriticalSectionWin32::enter()
    {
    switch( ::WaitForSingleObject( mutex, INFINITE ) )
        {
    case WAIT_OBJECT_0:
        // ok, we entered the critical section
        break;

    case WAIT_ABANDONED:
        // Owner of the critical section terminated without unlocking it,
        // ownership of the mutex is granted to us, but the mutex is 
        // unlocked. Fixme: another reaction than an exception?
        throw ExCriticalSectionCanNotCreate( "Got hold of abandoned mutex." );
        break;

    case WAIT_FAILED:
        {
        std::ostringstream  str;
        str << "Couldn't enter critical section: " << ::GetLastError();
        throw ExCriticalSectionCanNotCreate( str.str() );
        break;
        }

    default:
        panic( "Unexpected value returned from WaitForSingleObject." );
        break;
        }
    }


bool CriticalSectionWin32::try_enter()
    {
    switch( ::WaitForSingleObject( mutex, 0 ) )
        {
        case WAIT_OBJECT_0:
            // ok, we entered the critical section
            break;

        case WAIT_ABANDONED:
            // Owner of the critical section terminated without unlocking it,
            // ownership of the mutex is granted to us, but the mutex is 
            // unlocked. Fixme: another reaction than an exception?
            throw ExCriticalSectionCanNotCreate( "Got hold of abandoned mutex." );
            break;

        case WAIT_FAILED:
            {
            std::ostringstream  str;
            str << "Couldn't enter critical section: " << ::GetLastError();
            throw ExCriticalSectionCanNotCreate( str.str() );
            break;
            }
        case WAIT_TIMEOUT:
            {
            // CanNot enter critical section.
            return false;
            }
        default:
            panic( "Unexpected value returned from WaitForSingleObject." );
            break;
        }
    return true;
    }


void CriticalSectionWin32::leave()
    {
    if( ::ReleaseMutex( mutex ) == 0 )
        {   
        std::ostringstream  str;
        str << "Couldn't leave critical section: " << ::GetLastError();
        throw ExCriticalSectionCanNotLeave( str.str() );
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
