/** @file thread_win32.cpp
 * Interface for Win32 threads.
 * @see thread_win32.h
 */

/*
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "status/exception.h"
#include "thread/arch/win32/thread_win32.h"
#include "utility/debug.h"
#include "thread/semaph.h"
#include "thread/thread.h"
#include "global/global.h"
#include <process.h>
#include <windows.h>
#include <eh.h>
#include <sstream>


namespace Massiv {
namespace Core {

namespace {

__declspec( thread ) Thread::ThreadId current_thread_index;
    /**< Current thread index. A per thread global variable. */

}
    

// Implementation of ThreadWin32
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ThreadWin32::ThreadWin32
    (
    ThreadRoutine   thread_routine,
    VariantPointer  arg
    ) : Thread( thread_routine, arg )
    {
    unsigned int handle;

    // Launch thread.

    // Important note: This must be done via C library call,
    // not direct Win32 API. That would lead to memory leaks or
    // program malfunction (C/C++ library, exceptions, etc.).

    handle = _beginthreadex
        (
        NULL,
        0,
        thread_launcher,
        this,
        0,
        &thread_id
        );

    if( handle == 0 )
        {
        std::ostringstream  ss;
        ss << "Cannot create thread; error: " << ::GetLastError();
        throw ExThreadCanNotCreate( ss.str() );
        }

    thread_handle = reinterpret_cast<HANDLE>( handle );
    }


ThreadWin32::~ThreadWin32()
    {
    ::CloseHandle( thread_handle );
    }


int ThreadWin32::join()
    {
    unsigned long exit_code;

    switch( ::WaitForSingleObject( thread_handle, INFINITE ) )
        {
        case WAIT_OBJECT_0:
            // The state of the specified object is signaled.
            break;
        case WAIT_ABANDONED:
            // The specified object is a mutex object that was not released by the
            // thread that owned the mutex object before the owning thread terminated.
            // Ownership of the mutex object is granted to the calling thread, and the
            // mutex is set to nonsignaled.
            throw ExThreadCanNotJoin( "Cannot join thread; returned value of WaitForSingleObject() == WAIT_ABANDONED" );
        case WAIT_TIMEOUT:
            // The time-out interval elapsed, and the object's state is nonsignaled.
            throw ExThreadCanNotJoin( "Cannot join thread; returned value of WaitForSingleObject() == WAIT_TIMEOUT" );
        case WAIT_FAILED:
            {
            std::ostringstream  ss;
            ss << "Cannot join thread; returned value of WaitForSingleObject() == WAIT_FAILED; error: " << GetLastError();
            throw ExThreadCanNotJoin( ss.str() );
            }
        default:
            panic( "Unexpected value returned from WaitForSingleObject()." );
        }

    if( ::GetExitCodeThread( thread_handle, &exit_code ) == 0 )
        {
        std::ostringstream  ss;
        ss << "Cannot join thread; returned value of GetExitCodeThread() == 0; error: " << GetLastError();
        throw ExThreadCanNotJoin( ss.str() );
        }

    return exit_code;
    }


unsigned __stdcall ThreadWin32::thread_launcher
    (
    void * context
    )
    {
    ThreadWin32 *   this_thread = reinterpret_cast<ThreadWin32 *>( context );
    unsigned        exit_code;

    current_thread_index = Global::thread_manager().thread_starting();
    exit_code = this_thread->run();
    Global::thread_manager().thread_stopping();

    return exit_code;
    }


// Implementation of Thread
// ~~~~~~~~~~~~~~~~~~~~~~~~

void Thread::exit
    (
    int exit_code
    )
    {
    // Stop current thread.
    // C library _endthreadex() ensures that thread-allocated resources
    // will be properly freed.
    // Win32's ExitThread() must not be called!

    Global::thread_manager().thread_stopping();
    _endthreadex( exit_code );
    }


void Thread::sleep
    (
    int miliseconds
    )
    {
    ::Sleep( miliseconds );
    }


UInt32 Thread::get_remaining_stack_size()
    {
    #ifdef ARCH_X86

    ::MEMORY_BASIC_INFORMATION info;

    // Get information about a stack region.
    // Use @c &info as an approximation of the current stack pointer.
    
    if( ::VirtualQuery( &info, &info, sizeof( info ) ) != sizeof( info ) )
        {
        return 0;
        }
    
    // @c info.BaseAddress points to the page where @c info is allocated
    // ("current stack pointer").
    // @c info.AllocationBase points to the stack limit.
    // @c info.RegionSize holds current stack size (number of bytes to the end
    // of the region, the bottom of the stack).
    // This relies on top-to-down stack layout.

    return reinterpret_cast<int>( info.BaseAddress ) - 
        reinterpret_cast<int>( info.AllocationBase ); 

    #else

    #error "Need to implement this."

    #endif
    }


Thread::ThreadId Thread::get_current_thread_id()
    {
    ThreadId id = ::GetCurrentThreadId();
    hope( id != 0 );
    return id;
    }


Thread::ThreadIndex Thread::get_current_thread_index()
    {
    return current_thread_index;
    }


// Implementation of ThreadManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reference<Thread> ThreadManager::create
    (
    Thread::ThreadRoutine   start_routine,
    VariantPointer          arg
    )
    {
    Thread * t;

    t = new ThreadWin32( start_routine, arg );

    // Wait until thread starts.
    // This ensures that Thread object (@c t) still exists
    // when thread physicially starts.

    launcher_semaphore->down();

    return t;
    }


} // namespace Massiv::Core
} // namespace Massiv
