/** @file request_queue.cpp
 * %Request queue implementation.
 * @see request_queue.h
 */

/*
 * begin        : Sat Oct 24 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "thread/request_queue.h"
#include "global/global.h"
#include <algorithm>

namespace Massiv {
namespace Core {


// Implementation of RequestQueue::Request
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RequestQueue::Request::Request()
    {
    queue = NULL;
    priority = 0;

    can_cancel = false;
    pending = false;
    result = RESULT_PENDING;
    completed = NULL;
    }


RequestQueue::Request::~Request()
    {
    delete completed;
    }


void RequestQueue::Request::cancel()
    {
    hope( queue );
    queue->cancel_request( this );
    }


void RequestQueue::Request::wait_for_completion()
    {
    hope( completed );
    completed->down();
    }


void RequestQueue::Request::signal_error
    (
    const Status &          status,
    const std::string &     message,
    bool                    save_context
    )
    {
    hope( queue );
    queue->signal_error( status, message, save_context ? this : NULL );
    }


RequestQueue::Request::ProcessingResult RequestQueue::Request::process()
    {
    return RESULT_DONE;
    }


void RequestQueue::Request::cancel_routine()
    {
    }


// Implementation of RequestQueue::StopThreadRequest
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RequestQueue::Request::ProcessingResult RequestQueue::StopThreadRequest::process()
    {
    return RESULT_STOP;
    }


// Implementation of RequestQueue::Error
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RequestQueue::Error::Error()
    {
    }


RequestQueue::Error::Error
    (
    const Status &      status,
    const std::string & message,
    Request *           request
    ) : message( LogMessage( status, message ) ),
        request( request )
    {
    }


// Implementation of RequestQueue
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RequestQueue::RequestQueue
    (
    int max_requests
    )
    {
    CriticalSectionManager & csm = Global::critical_section_manager();
    section = csm.create();
    SemaphoreManager & sm = Global::semaphore_manager();
    free_slots = sm.create( max_requests );
    full_slots = sm.create( 0 );

    stopping = false;
    ThreadManager & tm = Global::thread_manager();
    worker_thread = tm.create( thread, VariantPointer( this ) );
    }


RequestQueue::~RequestQueue()
    {
    // Send stop request.
    // Do not wait for completion of the request since
    // we may already be stopping.
    // Wait until worker thread finishes instead.

    stop( false );
    worker_thread->join();

    // Destroy.

    delete section;
    delete free_slots;
    delete full_slots;
    }


void RequestQueue::stop
    (
    bool wait
    )
    {
    RequestPtr r;

    // Send stop request.
    // Its priority is set up so that it will be the last processed
    // request.

    r = new StopThreadRequest();

    queue_request
        (
        r.get(),
        MINIMUM_PRIORITY - 1,
        false,
        wait,
        STOP_REQUEST
        );

    // Wait for completion.

    if( wait )
        {
        r->wait_for_completion();
        }
    }


void RequestQueue::queue_request
    (
    Request *   r,
    Priority    priority,
    bool        can_cancel,
    bool        signal_completion
    )
    {
    queue_request
        (
        r,
        priority,
        can_cancel,
        signal_completion,
        PUBLIC_REQUEST
        );
    }


bool RequestQueue::are_requests_pending()
    {
    Lock lock( section );

    return queue.size() > 0;
    }


bool RequestQueue::are_errors_pending()
    {
    Lock lock( section );

    return errors.size() > 0;
    }


RequestQueue::Error RequestQueue::get_last_error()
    {
    Lock    lock( section );
    Error   e;

    hope( errors.size() );

    e = errors.top();
    errors.pop();

    return e;
    }


void RequestQueue::log_errors()
    {
    Lock lock( section );

    while( errors.size() )
        {
        Error & e = errors.top();

        Global::log( e.message );

        errors.pop();
        }
    }


void RequestQueue::queue_request
    (
    Request *   r,
    Priority    priority,
    bool        can_cancel,
    bool        signal_completion,
    int         request_id
    )
    {
    // Update semaphore value.

    free_slots->down();

    try
        {
        Lock    lock( section );
        int     i;

        // If we are stopping, public requests can not be sent.

        if( request_id == PUBLIC_REQUEST )
            {
            hope( !stopping );
            hope( priority >= MINIMUM_PRIORITY );
            }
        else
        if( request_id == STOP_REQUEST )
            {
            // If we are already stopping the other stop request
            // will never be completed.

            hope( !stopping || ( stopping && !signal_completion ) );

            if( stopping )
                {
                return;
                }

            stopping = true;
            }

        // Initialize request.

        hope( !r->pending );

        r->queue = this;
        r->priority = priority;
        r->can_cancel = can_cancel;

        if( signal_completion )
            {
            SemaphoreManager & sm = Global::semaphore_manager();
            r->completed = sm.create( 0 );
            }

        // Register request.

        queue.push_back( r );

        for( i = queue.size() - 2; i >= 0; i-- )
            {
            if( queue[ i ]->priority < queue[ i + 1 ]->priority )
                {
                queue[ i ].swap( queue[ i + 1 ] );
                }
            else
                {
                break;
                }
            }

        r->pending = true;
        }
    catch( ... )
        {
        // Undo queue state.

        free_slots->up();
        complete_request( r, Request::RESULT_ERROR );

        return;
        }

    // Update semaphore value.

    full_slots->up();
    }


RequestQueue::RequestPtr RequestQueue::dequeue_request()
    {
    RequestPtr r;

    // Update semaphore value.

    full_slots->down();

    {
    Lock lock( section );

    // Unregister request.

    hope( queue.size() > 0 );

    r = queue.front();
    queue.pop_front();

    // Update request.

    r->can_cancel = false;
    r->pending = false;
    }

    // Update semaphore value.

    free_slots->up();

    return r;
    }


void RequestQueue::cancel_request
    (
    Request * r
    )
    {
    // Locked section.

    {
    Lock            lock( section );
    Queue::iterator it;

    if( !r->can_cancel || !r->pending )
        {
        // Request can not be cancelled because either:
        //      1) request is being processed now
        //      2) request was completed already
        //      3) it is not allowed

        return;
        }

    // Find request registration.

    for( it = queue.begin(); it != queue.end(); it++ )
        {
        if( r == *it )
            break;
        }

    hope( it != queue.end() );
    queue.erase( it );
    }

    // Complete request.

    complete_request( r, Request::RESULT_CANCELLED );

    // Update semaphore values.

    free_slots->up();
    full_slots->down();
    }


void RequestQueue::complete_request
    (
    Request *                   r,
    Request::ProcessingResult   result
    )
    {
    Lock lock( section );

    r->can_cancel = false;
    r->pending = false;
    r->result = result;

    if( result == Request::RESULT_CANCELLED )
        r->cancel_routine();

    if( r->completed )
        r->completed->up();
    }


RequestQueue::Request::ProcessingResult RequestQueue::process_next_request()
    {
    Request::ProcessingResult result;
    RequestPtr r;

    // Process request.

    r = dequeue_request();
    result = r->process();

    // Complete request.

    complete_request( r.get(), result );

    return result;
    }


int RequestQueue::thread
    (
    VariantPointer arg
    )
    {
    RequestQueue * queue = variant_cast<RequestQueue *>( arg );

    // Thread loop.

    while( queue->process_next_request() != Request::RESULT_STOP )
        {
        }

    return 0;
    }


void RequestQueue::signal_error
    (
    const Status &      status,
    const std::string & message,
    Request *           request
    )
    {
    Lock lock( section );

    errors.push( Error( status, message, request ) );
    }


} // namespace Massiv::Core
} // namespace Massiv
