/** @file request_queue.h
 * %Request queue implements a priority queue of requests and a worker
 * thread that processes them asynchronously.
 */

/*
 * begin        : Sat Oct 24 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_REQUEST_QUEUE_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#ifndef MASSIV_THREAD_H
#include "thread/thread.h"
#endif

#ifndef MASSIV_SYNCH_H
#include "thread/synch.h"
#endif

#ifndef MASSIV_SEMAPH_H
#include "thread/semaph.h"
#endif

#ifndef MASSIV_LOGMESSAGE_H
#include "logger/logmessage.h"
#endif

#include <deque>
#include <stack>

namespace Massiv {
namespace Core {


/** @addtogroup thread
 * @{
 */

/** 
 * Class implements a priority queue of requests.
 *
 * Each request is a class derived from @c RequestQueue::Request.
 * RequestQueue launches worker thread that takes queued requests
 * and processes them asynchronously.
 *
 * Requests with the same priority are processed in FIFO order.
 * Requests should be referenced by @c RequestPtr (which ensures
 * that request object will not be deleted, because the Request
 * instances are reference-counted).
 * %RequestQueue is thread-safe.
 *
 * Example (note that in the reality <tt>RequestQueue::Request</tt> 
 * would be replaced by some suitable descendant):
 * @code
 * RequestQueue                q( 3 );
 * RequestQueue::RequestPtr    ptr1, ptr2, ptr3;
 *
 * q.queue_request( ptr1 = new RequestQueue::Request() );   // 1
 * q.queue_request( ptr2 = new RequestQueue::Request() );   // 2
 * q.queue_request( ptr3 = new RequestQueue::Request() );   // 3
 *
 * ptr3.set_null();
 *     // this enables to deallocate the third request
 *
 * ptr1->cancel();
 * ptr2->wait_for_completion();
 * @endcode
 */
class RequestQueue
    {
private:

    NO_COPY_CONSTRUCTOR( RequestQueue );
    NO_ASSIGNMENT_OPERATOR( RequestQueue );

public:

    typedef int Priority;

    /**
     * %Parameters of the queue.
     */
    enum Parameters
        {
        DEFAULT_PRIORITY = 0,
            /**< Default request priority. */
        MINIMUM_PRIORITY = 0,
            /**< Minimum request priority. */
        }; // Parameters

    /**
     * %Constants.
     */
    enum Constants
        {
        PUBLIC_REQUEST,
            /**< Identification of public request. */
        STOP_REQUEST
            /**< Identification of StopThreadRequest. */
        }; // Constants

    /** 
     * Base class for %RequestQueue-managed requests.
     */
    class Request : public SharedObject
        {
    private:

        friend class RequestQueue;

    public:

        /**
         * Constants that indicate the result of processing a request.
         */
        enum ProcessingResult
            {
            RESULT_PENDING,
                /**< Request is still pending. */
            RESULT_DONE,
                /**< Request has been processed successfully. */
            RESULT_CANCELLED,
                /**< Request has been cancelled. */
            RESULT_ERROR,
                /**< Request could not be processed. */
            RESULT_STOP
                /**< Worker thread has been asked to stop. */
            }; // ProcessingResult

    public:

        /** Default constructor. */
        Request();

        /** Destructor. */
        virtual ~Request();

        /** Cancels the request. */
        void cancel();

        /** Waits until the operation completes. */
        void wait_for_completion();

        /**
         * Puts an %error into the @link RequestQueue::errors error stack 
         * @endlink.
         * The %error can be examined by the main thread by calling
         * @c RequestQueue:get_last_error() afterwards.
         *
         * @param status A status of the %error.
         * @param message A textual message to describe the %error.
         * @param save_context Determines whether the constructed @c Error 
         *      object points to this request object or not.
         */
        void signal_error
            (
            const Status &          status,
            const std::string &     message,
            bool                    save_context = true
            );

    public:

        /** @name Request callbacks */
        //@{

        /**
         * Called by the request queue's worker thread when it chooses to 
         * process the request.
         *
         * @return The result of the processing.
         */
        virtual ProcessingResult process();

        /**
         * Called by @c RequestQueue when the request is cancelled.
         */
        virtual void cancel_routine();

        //@}

    private:

        /** 
         * @name Attributes that require serialized access
         * These attributes are being accessed by RequestQueue and thus
         * each access must be locked by queue-\>lock.
         */
        //@{

        mutable RequestQueue *      queue;
            /**< %Request queue. */
        mutable Priority            priority;
            /**< %Request priority. */
        mutable bool                can_cancel : 1;
            /**< Can %request be cancelled? */
        mutable bool                pending : 1;
            /**< %Request is waiting for being processed. */
        mutable ProcessingResult    result;
            /**< Current processing result. */
        mutable Semaphore *         completed;
            /**< %Semaphore is blocked until request is completed.
                 @c completed may be NULL if client did not plan to wait
                 for completion. */

        //@}

        }; // class Request

    friend  class Request;
    typedef Reference<Request, MTReferenceTraits>   RequestPtr;
        /**< %Pointer to %request %type. */

    /** 
     * Internal request that requests to stop the worker thread
     * request queue worker thread.
     */
    class StopThreadRequest : public Request
        {
    protected:

        /**
         */
        virtual ProcessingResult process();

        }; // StopThreadRequest

    /** 
     * Class describing an %error condition that occured during
     * processing of a request.
     *
     * %Errors are stored it @c ErrorStack so that they can be
     * examined by the main thread.
     */
    class Error
        {
    public:

        /** Default constructor. */
        Error();

        /** 
         * Constructor. 
         *
         * @param status Status of the %error.
         * @param message The textual message describing the %error.
         * @param request The request concerned.
         */
        Error
            (
            const Status &      status,
            const std::string & message,
            Request *           request
            );

    public:

        LogMessage      message;
            /**< %Log message @endlink describing the %error. */
        RequestPtr      request;
            /**< Optional pointer to the request that is relevant
                 for this %error. */
        }; // Error

    typedef std::deque<RequestPtr>      Queue;
        /**< Queue for requests. */
    typedef std::stack<Error>           ErrorStack;
        /**< %Stack of @link Error Errors @endlink. */

public:

    //
    // Start & stop.
    //

    /**
     * Constructor.
     *
     * Initializes request queue.
     *
     * @param max_requests Maximum number of pending requests.
     */
    RequestQueue
        (
        int max_requests
        );

    /**
     * Destructor.
     * Waits for completion of all pending requests.
     */
    ~RequestQueue();

    /**
     * Makes queue reject new requests.
     *
     * @param wait true instructs %RequestQueue to wait until all pending 
     *      requests are completed. Note that the method should not be
     *      called multiple times if @c %wait is true.
     */
    void stop
        (
        bool wait = true
        );

public:

    //
    // Request queueing.
    //

    /**
     * Queues a requestto be processed later by a worker thread.
     *
     * @param r The request.
     * @param priority The priority of the request.
     * @param can_cancel If true, it will be possible to cancel the request.
     * @param signal_completion If true, it will be possible to call
     *      <tt>r-\>wait_for_completion()</tt>.
     */
    void queue_request
        (
        Request *   r,
        Priority    priority = DEFAULT_PRIORITY,
        bool        can_cancel = true,
        bool        signal_completion = true
        );

    /** Tests if there are pending requests. */
    bool are_requests_pending();

public:

    /** @name Error handling */
    //@{

    /** 
     * Tests if there are any %errors stored on the 
     * @link Errors error stack @endlink. 
     */
    bool are_errors_pending();

    /**
     * Gets information about the last %error.
     *
     * @return Last %error from the @link errors error stack @endlink.
     * @note Fails if no %error is available in the stack.
     */
    Error get_last_error();

    /**
     * Sends %errors from the @link Errors error stack @endlink to logger 
     * and empties the error stack.
     */
    void log_errors();

    //@}

private:

    //
    //
    // Private.

    /** 
     * Pushes request into the queue. 
     *
     * @param r %Pointer to the request.
     * @param priority The priority of the request.
     * @param can_cancel If true, it will be possible to cancel the request.
     * @param signal_completion If true, it will be possible to call
     *      <tt>r-\>wait_for_completion()</tt>.
     * @param request_id Type of requests (see 
     *      @link RequestQueue::Constants constants @endlink).
     */
    void queue_request
        (
        Request *   r,
        Priority    priority,
        bool        can_cancel,
        bool        signal_completion,
        int         request_id
        );

    /** 
     * Pops the first request from the queue. 
     * @return %Pointer to the request.
     */
    RequestPtr dequeue_request();

    /** 
     * Cancels request. 
     * @param r A %pointer to the request to be canceled.
     */
    void cancel_request
        (
        Request * r
        );

    /** 
     * Completes request. 
     *
     * @param r A %pointer to the request to be completed now.
     * @param result A result of the processing.
     */
    void complete_request
        (
        Request *                   r,
        Request::ProcessingResult   result
        );

    /** 
     * Processes next pending request. 
     * @param A result of the processing.
     */
    Request::ProcessingResult process_next_request();

    /** Worker thread routine. */
    static int thread
        (
        VariantPointer arg
        );

    /** 
     * Puts an %error to the @link RequestQueue::errors error stack @endlink.
     * The %error can be examined by the main thread by calling
     * @c RequestQueue:get_last_error() afterwards.
     *
     * @param status A @link Status status @endlink of the %error.
     * @param message A textual message to describe the %error.
     * @param request A %pointer to the relevant %request %object. 
     */
    void signal_error
        (
        const Status &      status,
        const std::string & message,
        Request *           request
        );

private:

    CriticalSection *           section;
        /**< Main critical section. */
    Queue                       queue;
        /**< Request queue. */
    Semaphore *                 free_slots;
        /**< Blocks when number of free slots of the @c queue drops to zero. */
    Semaphore *                 full_slots;
        /**< Blocks when number of full slots of the @c queue drops to zero. */
    Reference<Thread>           worker_thread;
        /**< Worker thread reference. Processes requests. */
    ErrorStack                  errors;
        /**< %Error stack. */
    bool                        stopping : 1;
        /**< Are we rejecting new requests? */

    }; // RequestQueue

/** @} */ // addtogroup thread


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_REQUEST_QUEUE_H
