/** @file semaph.h
 * Platform independent semaphore interface.
 */

/*
 * begin        : Tue Nov 1 2002
 * copyright    : (C) 2002 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_SEMAPH_H
#define MASSIV_SEMAPH_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup thread
 * @{
 */

/** 
 * Platform-independent Dijkstra semaphore.
 *
 * Use @link SemaphoreManager semaphore manager @endlink to create semaphore
 * instances.
 */
class MASSIV_EXPORTED Semaphore
    {
public:

    /**
     * Destructor.
     */
    virtual ~Semaphore();

    /**
     * If the count equals zero, %down() blocks and wait for another 
     * thread/process to increase the semaphore count.
     */
    virtual void down() = 0;

    /**
     * Increases the semaphore count by 1.
     *
     * If the count had been 0, another thread waiting in the down() routine
     * may continue.
     */
    virtual void up() = 0;

    }; // Semaphore


/** 
 * Semaphore management (use this for creating semaphores - see the example
 * below).
 * 
 * @code
 * {
 * Semaphore * sem = Global::semaphore_manager().create( 4 );
 *     {
 *     sem->down();
 *     // Work 'in' the guarded section.
 *     sem->up();
 *     }
 * delete sem;
 * }
 * @endcode
 */
class MASSIV_EXPORTED SemaphoreManager
    {
public:

    /**
     * Constructor.
     */
    SemaphoreManager();

    /**
     * Creates and returns a new instance of the Semaphore class.
     * @param init_count - initial semaphore count.
     */
    Semaphore * create
        (
        int init_count
        );

    }; // SemaphoreManager

/** @} */ // addtogroup thread


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SEMAPH_H
