/** @file shmem.h
 * Platform-independent shared memory interface.
 */

/* 
 * begin        : Thu Mar 10 2003
 * copyright    : (C) 2003 by Martin Havlista
 */


/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_SHMEM_H
#define MASSIV_SHMEM_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#include <fstream>

namespace Massiv {
namespace Core {


/** @addtogroup thread
 * @{
 */

/** 
 * Provides platform independent inter-process memory sharing.
 *
 * @code
 * std::string      memory_name = "MY_SHARED_MEMORY";
 * UInt32           size = 1000;
 *
 * SharedMemory *   shmem = new SharedMemory( size, memory_name );
 *     // creating the shared memory system object, etc.
 *
 * char *           pointer = static_cast< char * >( shmem.get_memory() );
 *     // getting a pointer to the actual shared memory
 *
 * // work with the memory...
 *
 * delete shmem;
 * @endcode
 */
class SharedMemory
    {
public:
        
    /**
     * Constructor.
     *
     * @param size Size of the memory.
     * @param name Name of the shared memory. This is needed for a process
     *      to access a shared memory created by another one. If this name
     *      is the same as the name of some other synchronization primitive
     *      (semaphore, mutex, etc.), the constructor might throw an 
     *      exception (platform-dependent behaviour). To avoid this, always
     *      choose system-unique names for shared memory.
     *
     * @exception ExCommonError The shared memory could not be created.
     */
    SharedMemory
        (
        UInt32                  size,
        const std::string &     name
        );

    /**
     * Destructor.
     * @exception ExCommonError Could not destroy the shared memory.
     * @bug Exception in the destructor.
     */
    ~SharedMemory();
    
    /**
     * Gets the shared memory pointer.
     * @returns %Pointer to the shared memory.
     */
    void * get_memory() const;
            
private:

    void *          memory;            
        /**< Raw pointer to the shared memory. */
    std::string     map_name;
        /**< Name of the shared memory. */
    UInt32          _size;
        /**< Size of allocated memory. */
    bool            owner;
        /**< True when this instance is owner of shared memory and is responsible
             for destroying. Unix only. */
    int             shmid;
        /**< Shared memory identifier. Unix only. */
    }; // SharedMemory

/** @} */ // addtogroup thread


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SHMEM_H
