/** @file synch.h
 * Platform-independent interface for synchronization primitives 
 * (@link CriticalSection critical section @endlink
 * and @link Lock lock @endlink).
 */

/*
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_SYNCH_H
#define MASSIV_SYNCH_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif


namespace Massiv {
namespace Core {


/** @addtogroup thread
 * @{
 */

/** 
 * Class implementing a platform-independent critical section.
 *
 * @note Critical section can be created using the 
 * @link CriticalSectionManager critical section manager @endlink.
 */
class MASSIV_EXPORTED CriticalSection
    {
public:

    /**
     * Constructor.
     */
    CriticalSection();

    /**
     * Destructor.
     *
     * @exception ExCriticalSectionCannotDestroy
     */
    virtual ~CriticalSection();

    NO_COPY_CONSTRUCTOR( CriticalSection );
    NO_ASSIGNMENT_OPERATOR( CriticalSection );

public:

    /**
     * Enters the critical section. 
     *
     * Blocks the current thread until the critical section is available.
     *
     * @exception ExCriticalSectionCannotEnter
     */
    virtual void enter() = 0;

    /**
     * Tries to enter the critical section. 
     *
     * @return false if critical section is not free (in this case, the
     *      calling thread has not entered the section), true otherwise
     *      (the calling thread has entered the section).
     */
    virtual bool try_enter() = 0;

    /**
     * Leaves critical section.
     *
     * It will be available for another 'client' then.
     *
     * @exception ExCriticalSectionCannotLeave
     */
    virtual void leave() = 0;

    }; // CriticalSection


/** 
 * Simple class that aids using @c CriticalSection objects.
 *
 * It is used as a temporary object whose constructor enters the section
 * and destructor leaves the section.
 * The main benefit to standard enter() / leave() paradigm is that
 * exception handling semantics automatically unlocks the section
 * if an exception is thrown.
 *
 * See @link CriticalSectionManager critical section manager @endlink
 * for example.
 */
class MASSIV_EXPORTED Lock
    {
public:

    /**
     * Constructor enters critical section.
     *
     * The calling thread will remain in the section till destruction
     * of the %Lock object.
     *
     * @param section The critical section to be entered.
     * @exception ExCriticalSectionCannotEnter
     */
    Lock
        (
        CriticalSection * section
        );

    /**
     * Destructor leaves critical section.
     * @exception ExCriticalSectionCannotLeave
     */
    ~Lock();

    NO_COPY_CONSTRUCTOR( Lock );
    NO_ASSIGNMENT_OPERATOR( Lock );

private:

    CriticalSection * section;
        /** Locked critical section. */
    }; // Lock


/** 
 * Manages critical sections as in the examples below:
 *
 * @code
 * CriticalSectionManager & csm = Global::critical_section_manager();
 * CriticalSection * cs = csm::create();
 * 
 * cs->enter();
 * // Work in the critical section.
 * cs->leave();
 *
 * delete cs;
 * @endcode
 *
 * More advantageous usage of critical sections exercises the @c Lock
 * class, as it will be automaticaly unlock the section if an exception
 * is thrown, etc.:
 *
 * @code
 * CriticalSectionManager & csm = Global::critical_section_manager();
 * CriticalSection * cs = csm::create();
 * 
 *     {
 *     Lock lock( cs );
 *     // Work in the critical section.
 *     }  // critical section is automaticaly abandoned at the end of lock's scope
 *
 * delete cs;
 * @endcode
 */
class MASSIV_EXPORTED CriticalSectionManager
    {
public:

    /**
     * Constructor.
     */
    CriticalSectionManager();

    /**
     * Creates and returns a new instance of CriticalSection.
     *
     * On Win32 system creates CriticalSectionWin32.
     * On Unix system creates CriticalSectionUnix.
     *
     * @param owner true indicates that this instance of critical section
     *      will be responsible for destroying the low-level primitives
     *      used for critical section implementation. Moreover, true causes
     *      that an exception will be thrown in case the synchronization
     *      primitives used to implement the critical section are not
     *      created yet. You do not need to use this parameter in case
     *      of unnamed critical section.
     * @param name If not NULL, critical section will be used via named
     *      version of some synchronization primitives that can be shared
     *      among processes.
     *
     * @exception ExCriticalSectionCannotCreate
     *
     * @note Successful creating of a new instance depends on creating some
     *      system-dependent synchronization primitives used for
     *      implementation of the critical section.
     */
    CriticalSection * create
        (
        bool            owner = true,
        const char *    name = 0
        );

    }; // CriticalSectionManager

/** @} */ // addtogroup thread


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_SYNCH_H
