/** @file thread.cpp
 * Platform-independent interface for thread manipulation.
 * @see thread.h
 */

/*
 * begin        : Fri Jan 17 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "thread/thread.h"
#include "global/global.h"
#include "utility/debug.h"
#include "global/global_callbacks.h"
#include <iostream>
#include <stdexcept>
#include <cstdlib>


namespace Massiv {
namespace Core {


// Implementation of Thread
// ~~~~~~~~~~~~~~~~~~~~~~~~

Thread::Thread
    (
    ThreadRoutine   thread_routine,
    VariantPointer  arg
    ) : routine( thread_routine ),
        routine_args( arg )
    {
    }


Thread::~Thread()
    {
    }


int Thread::run()
    {
    try
        {
        // Read all the information needed to start thread routine.

        ThreadRoutine   r = routine;
        VariantPointer  args = routine_args;

        routine = NULL;
        routine_args.set_null();

        // Notify ThreadManager that the thread has started.
        // Since now Thread object (this) must be treated
        // inaccessible as it may be deleted by client
        // (the physical thread might still be running though).

        launcher_semaphore()->up();

        // Call thread routine.

        return r( args );
        }
    catch( std::exception & e )
        {
        // Note that Thread object is inaccessible.

        std::cerr   << "Unhandled exception leaked from a thread." << std::endl
                    << e.what() << std::endl;
        abort();
        }
    catch( ... )
        {
        // Note that Thread object is inaccessible.

        std::cerr   << "Unhandled exception leaked from a thread." << std::endl;
        abort();
        }
    /*NOTREACHED*/
    }


Semaphore * Thread::launcher_semaphore()
    {
    return Global::thread_manager().launcher_semaphore;
    }


// Implementation of ThreadManager
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ThreadManager::ThreadManager()
    {
    int i;

    // Allocate synchronization primitives.
    
    launcher_semaphore = Global::semaphore_manager().create( 0 );
    array_cs = Global::critical_section_manager().create();

    // Clear thread id array.
    
    for( i = 0; i < MAX_THREADS; i++ )
        {
        thread_id_array[ i ] = 0;
        }

    // Register current thread.

    thread_starting();
    }


ThreadManager::~ThreadManager()
    {
    delete launcher_semaphore;
    delete array_cs;
    }


Thread::ThreadIndex ThreadManager::thread_starting()
    {
    int i;

    // Register thread.

    array_cs->enter();
    for( i = 0; i < MAX_THREADS; i++ )
        {
        if( thread_id_array[ i ] == 0 )
            {
            thread_id_array[ i ] = Thread::get_current_thread_id();
            hope( thread_id_array[ i ] != 0 );

            break;
            }
        }
    array_cs->leave();

    if( i == MAX_THREADS )
        {
        panic( "Too many threads." );
        }

    // Notify core.

    GlobalCallbacks::thread_starting();

    // Done.

    return i;
    }


void ThreadManager::thread_stopping()
    {
    // Notify core.

    GlobalCallbacks::thread_stopping();

    // Unregister thread.

    array_cs->enter();
    thread_id_array[ Thread::get_current_thread_index() ] = 0;
    array_cs->leave();
    }


Thread::ThreadIndex ThreadManager::find_thread_index
    (
    Thread::ThreadId thread_id
    ) const
    {
    int i;
    
    array_cs->enter();

    for( i = 0; i < MAX_THREADS; i++ )
        {
        if( thread_id_array[ i ] == thread_id )
            {
            array_cs->leave();
            return i;
            }
        }

    panic( "Such a thread is not registered." );
    }


} // namespace Massiv::Core
} // namespace Massiv
