/** @file thread.h
 * Platform-independent interface for thread manipulation.
 */

/*
 * begin        : Tue May 14 2002
 * copyright    : (C) 2002 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_THREAD_H
#define MASSIV_THREAD_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#ifndef MASSIV_SEMAPH_H
#include "thread/semaph.h"
#endif

#ifndef MASSIV_SYNCH_H
#include "thread/synch.h"
#endif

#ifndef MASSIV_VARIANT_POINTER_H
#include "utility/variant_pointer.h"
#endif

namespace Massiv {
namespace Core {


/** @addtogroup thread
 * @{
 */

/** 
 * Platform independent thread interface.
 *
 * Currently, only Win32 and Unix threads are supported.
 * %Thread object might hold additional information about a thread
 * or define methods to control the thread.
 *
 * %Thread object's life-time is determined by client's references
 * to it. When there is no @link Reference reference @endlink to the 
 * object it is deleted.
 * However the physical thread might still be running.
 * That is because the running thread DOES NOT reference "its"
 * %Thread object and has no access to it.
 *
 * Threads can be launched using the @link ThreadManager thread 
 * manager @endlink.
 */
class MASSIV_EXPORTED Thread : public SharedObject
    {
public:

    /**
     * Prototype of a thread function.
     */
    typedef int ( *ThreadRoutine )( VariantPointer arg );
    
    /**
     * ThreadId.
     * Always non-zero.
     */
    typedef UInt32 ThreadId;

    /**
     * Thread index.
     */
    typedef UInt32 ThreadIndex;

public:

    /** @name Current thread management */
    //@{

    /**
     * Terminates execution of the calling thread.
     * 
     * @param exit_code Defines thread exit code that is returned
     *      by join().
     */
    static void exit
        (
        int exit_code
        );

    /**
     * Blocks thread execution temporarily.
     * @param miliseconds Number of miliseconds the thread will be paused.
     */
    static void sleep
        (
        int miliseconds
        );

    /**
     * Returns remaining stack size in bytes with respect to the calling thread.
     */
    static UInt32 get_remaining_stack_size();
    
    /**
     * Get current thread id, a non-zero identification of the calling thread.
     * @return %Thread id of the calling thread.
     */
    static ThreadId get_current_thread_id();
    
    /**
     * Gets an index of the current thread ranging from 0 to MAX_THREADS - 1.
     *
     * @return Position of the current thread in the
     *      @link ThreadManager::thread_id_array thread id array @endlink.
     */
    static ThreadIndex get_current_thread_index();

    //@}

public:

    /**
     * Constructor.
     * 
     * @param thread_routine The thread routine.
     * @param arg Bundle of arguments for the thread routine (@a arg will 
     *      be passed to the thread routine).
     *
     * @note Doesn't actually launch any threads.
     */
    Thread
        (
        ThreadRoutine   thread_routine,
        VariantPointer  arg
        );

    /**
     * Destructor.
     */
    virtual ~Thread();

    /**
     * Suspends execution of the calling thread until the thread
     * identified by this instance of class Thread exits.
     * 
     * @return The exit code of the thread the method is waiting for
     *      (see @link Thread::exit() exit() @endlink).
     */
    virtual int join() = 0;

protected:

    /**
     * Starts client's thread routine.
     * Called within a context of @c this thread. Once returned the object
     * need not be alive any more.
     * 
     * @return The thread's exit code (see @link Thread::exit() exit() 
     *      @endlink).
     */
    int run();

protected:

    /**
     * Gets global @link ThreadManager::launcher_semaphore launcher 
     * semaphore @endlink from the @link ThreadManager 
     * thread manager @endlink.
     *
     * @return The launcher semaphore pointer.
     */
    static Semaphore * launcher_semaphore();

private:

    //
    // Used by thread launcher.
    // Synchronized by ThreadManager::launcher_semaphore.
    //

    ThreadRoutine   routine;
        /**< %Thread routine. */
    VariantPointer  routine_args;
        /**< %Thread routine arguments. */

    }; // Thread


/** 
 * %ThreadManager provides functionality for launching and handling threads
 * properly (it exercises additional consistency checking, etc.)
 *
 * @code
 * int function
 *     (
 *     VariantPointer arg
 *     )
 *     {
 *     int number = *variant_cast<int *>( arg );
 *
 *     if( number == 1 )
 *         {
 *         Thread::exit( 123 );
 *         }
 *
 *     return 0;
 *     }
 * ...
 * int arg = 1;
 * Reference<Thread> thread = Global::thread_manager().create( function, &arg );
 * int result = thread->join();  // result == 123
 * @endcode
 */
class MASSIV_EXPORTED ThreadManager
    {
public:

    friend class Thread;

    enum Parameters
        {
        MAX_THREADS = 256
            /**< Maximum number of running threads. */
        }; // enum.

    /**
     * Constructor.
     */
    ThreadManager();

    /**
     * Destructor.
     */
    ~ThreadManager();

public:

    //
    // Public API.
    //

    /**
     * Creates a new thread to be executed.
     *
     * Returns new @c Thread object instance.
     * Resources for thread are not deallocated until all references
     * to the thread object are released.
     * Waits until the new thread starts (to ensure that the 'client'
     * doesn't delete the %Thread object prior to this).
     *
     * Multiple calls to @c create() must be synchronized by client.
     */
    Reference<Thread> create
        (
        Thread::ThreadRoutine   thread_routine,
        VariantPointer          args
        );

public:

    //
    // Private API used by architecture-specific thread implementation.
    //

    /**
     * Called by the thread implementation when a new thread is started.
     *
     * This ensures proper thread registration and notifies other subsystems
     * about the new thread. Must be called within the context of that thread.
     * 
     * Returns thread's index.
     */
    Thread::ThreadIndex thread_starting();

    /**
     * Called by the thread implemenation when the current thread is stopping.
     * Unregisters the thread and performs cleanup.
     */
    void thread_stopping();

    /**
     * Translates thread id to a thread index.
     */
    Thread::ThreadIndex find_thread_index
        (
        Thread::ThreadId thread_id
        ) const;

private:

    Semaphore *         launcher_semaphore;
        /**< %Semaphore which is used to synchronize access to a Thread object
             when a new thread is being launched. It is used to keep thread
             object alive until physical thread starts. */

    CriticalSection *   array_cs;
        /**< %Critical section used to synchronize access to
             @c thread_id_array. */
        
    Thread::ThreadId    thread_id_array[ MAX_THREADS ];
        /**< %Thread id array. Contains thread ids of running threads. */
            
    }; // ThreadManager

/** @} */ // addtogroup thread


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_THREAD_H
