/** @file time_manager.cpp
 * Simulation time synchronization.
 */

/*
 * begin        : Fri Aug 15 2002
 * copyright    : (C) 2002 by Petr Tovarys, Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "time/time_manager.h"
#include "serializer/stype_fwd.h"
#include "global/global.h"
#include "system/system.h"
#include "sysevent/time_synch_msg.h"
#include <math.h>
#include <sstream>


namespace Massiv {
namespace Core {


#define TM_INCREMENT_PERCENTAGE             20
    /**< When massiv time needs to be modified (slowed down, sped up), this 
         value says how fast will be the modification done.
         Example: massiv time should be expanded by value 'x'. This value
         will be changed to x +(-) x*(TM_INCREMENT_PERCENTAGE / 100) */         
#define TM_REAL_TIME_ACCURACY               3
    /**< Synchronizing with real time doesnt have to be much accurate.
         This value is a tolerance in seconds. */
#define TM_MC_TOLERANCE                     0.1
    /**< Magic constant tolerance. If magic constant value is not between
         <1 - TM_MC_TOLERANCE, 1 + TM_MC_TOLERANCE>, warning message is printed out
         to console and the value clamped to the range. */
#define TM_INCREMENT_TOLERANCE              3
    /**< This value manages in which way is done speeding up of massiv time.
         If local_synch_info.increment is greater than this value, whole value
         of local_synch_info.increment is added to massiv_time. If not, 
         local_synch_info.increment is gradually added to massiv time to achieve
         smart speeding up of massiv time. */
#define TM_INIT_SYNC_INTERVAL               1
    /**< Initial time synchronization interval. Value is in seconds. */
#define TM_MAX_TIME_SYNC_INTERVAL           60*60
    /**< Maximum time synchronization interval in seconds. */
#define TM_MIN_TIME_SYNC_INTERVAL           10
    /**< Smallest time synchronization interval in seconds. */
#define TM_CLIENT_TIME_SYNC_INTERVAL        60
    /**< Client time synchronization interval. */
#define TM_CLIENT_SLOWDOWN_TIME_VALUE       -0.5
    /**< Client slow down time value. If difference between clients massiv time 
         and its servers massiv time if greater than this value, clients massiv
         time is slowed down. Value is in seconds. */
#define TM_CLIENT_NUM_SLOWDOWN              3
    /**< Clients slow down attempts before clients massiv time is really slowed down. */

TimeManager::TimeManager
    (
    bool            init_paused
    )
    {
    initialize( init_paused );
    }


TimeManager::TimeManager
    (
    const NodeId & node_id
    ) :
    zero_system_time( Global::system_time() ),
    synchronization_allowed( false ),
    paused( false ),
    minimum_time_sync_value( TM_MIN_TIME_SYNC_INTERVAL ),
    test( true ),
    my_node_id( node_id )
    {
    time_sync_interval = time_manager_settings.time_synchronization_interval;
    }


TimeManager::TimeManager
    (
    const NodeId &          node_id,
    const SystemTime &      initial_time_value
    ) :
    zero_system_time( Global::system_time() - initial_time_value ),
    synchronization_allowed( false ),
    paused( false ),
    minimum_time_sync_value( TM_MIN_TIME_SYNC_INTERVAL ),
    test( true ),
    my_node_id( node_id )
    {
    time_sync_interval = time_manager_settings.time_synchronization_interval;
    CriticalSectionManager & csm = Global::critical_section_manager();
    cs = csm.create();
    }


TimeManager::~TimeManager()
    {
    if( test )
        {
        delete cs;
        }
    }


void TimeManager::initialize
    (
    bool init_paused
    )
    {
    current_time = STime();
    zero_system_time = Global::system_time();
    system_time = SystemTime();
    last_server_sync_system_time = SystemTime();
    last_client_sync_system_time = SystemTime();
    sync_info_hash_server.clear();
    sync_info_hash_client.clear();
    local_synch_info = SynchronizationInfo();    
    synchronization_allowed = false;
    paused = init_paused;
    pause_start = SystemTime();
    pause_end = SystemTime();
    pause_delta = SystemTime();
    real_time_sync = RealTimeSync();    
    minimum_time_sync_value = TM_MIN_TIME_SYNC_INTERVAL;
    time_sync_interval = SystemTime();
    test = false;
    time_sync_interval = time_manager_settings.time_synchronization_interval;
    }


void TimeManager::next_tick()
    {
    const SystemTime old_system_time = system_time;
    system_time = compute_system_time( false );
    const SystemTime addition = system_time - old_system_time;

    current_time = compute_simulation_time( addition );

    if( synchronization_allowed )
        {
        if( should_synchronize_time() )
            {
            synchronize_time();
            send_synchronization_info();
            }
        }
    if( sync_info_hash_client.size() )
        {
        synchronize_clients();
        }

    // Actual simulation time printout.
    static STime    old_time = System::time();
    STime           curr_time = get_time();
    bool    write =  curr_time.get_value() - static_cast< int >( curr_time.get_value() ) < 0.05;
    if( ( curr_time - old_time ) > 0.95 && ( write || ( curr_time - old_time ) > 1 ) )
        {
        old_time = curr_time;
        std::ostringstream  ss;
        ss << "[ Simulation time: " << ( curr_time.get_value() ) << " ]";
        Global::log
            (
            STATUS_INFO( Status::FACILITY_TIME_MANAGEMENT, Status::PRIORITY_LOWEST ),
            ss.str()
            );
        }

    // Temporary tests and server time print-out
    /*if( System::is_server() )
        {
            {
            static SystemTime   old_time = system_time;
            SystemTime          act_time = system_time;
            const STime         time = get_time();
            bool    write =  time.get_value() - static_cast< int >( time.get_value() ) < 0.1;
            if( act_time - old_time >= 0.8 && write )
                {
                old_time = act_time;            
                std::ostringstream  ss;
                ss << "Time [ " << time.get_value() << " ]";
                Global::log
                    (
                    STATUS_INFO( Status::FACILITY_TIME_MANAGEMENT, Status::PRIORITY_LOWEST ),
                    ss.str()
                    );
                }
            }*/
        // FIXME Magic constant test. Presunout test do time testu.
/*        const NodeId & node_id = Global::get_local_node_id();
        if( node_id == NodeId::create_from_type_and_relative_id( 0, 2 ) )
            {
                {
                static bool b1 = true;
                if( get_time() > 20 && b1 )
                    {
                    b1 = false;
                    synchronize_with_real_time( Global::system_time().get_value() );
                    }
                }
                {
                static bool b2 = true;
                if( get_time() > 30 && b2 )
                    {
                    b2 = false;
                    synchronize_with_real_time( Global::system_time().get_value() + 5 );
                    }
                }
                {
                static bool b3 = true;
                if( get_time() > 50 && b3 )
                    {
                    b3 = false;
                    synchronize_with_real_time( Global::system_time().get_value() + 5 );
                    }
                }
            }*/
    }


SystemTime TimeManager::compute_system_time
    (
    bool        ignore_current_pause_state
    ) const
    {
    if( !ignore_current_pause_state && paused )
        {
        return pause_start;
        }

    const SystemTime    real_system_time = Global::system_time();
    hope( real_system_time >= zero_system_time );

    SystemTime    sys_time = real_system_time - zero_system_time;
    hope( sys_time >= pause_delta );

    return sys_time - pause_delta;
    }


STime TimeManager::compute_simulation_time
    (
    const SystemTime &  addition
    )
    {
    STime   new_current_time = current_time;
    Float64 add = addition.get_value();
    
    // Modify value by time_speed_constant
    add *= ( Float64 )time_manager_settings.time_speed_constant;
    
    // Modify value by real_time_sync.magic_constant
    add *= real_time_sync.magic_constant;

    if( local_synch_info.increment < TM_INCREMENT_TOLERANCE )
        {
        // local_synch_info.increment is small - it can be added gradually.
        Float64 percentage = add * TM_INCREMENT_PERCENTAGE / 100;
        if( local_synch_info.increment > percentage )
            {
            // Increase
            add += percentage;
            local_synch_info.increment -= percentage;
            if( local_synch_info.increment < percentage )
                {
                add += local_synch_info.increment;
                local_synch_info.increment = 0.0;
                }
            }
        if( local_synch_info.increment < -percentage )
            {
            // Decrease
            add -= percentage;
            local_synch_info.increment += percentage;
            if( local_synch_info.increment > -percentage )
                {
                add += local_synch_info.increment;
                local_synch_info.increment = 0.0;
                }
            }
        }
    else
        {
        // local_synch_info.increment is big - whole value must be added in one moment.
        // This cause a jump in time to the future.
        add += local_synch_info.increment;
        local_synch_info.increment = 0.0;
        }

    new_current_time += add;
    hope( new_current_time >= current_time );
    return new_current_time;
    }


STime TimeManager::get_time() const
    {
    return current_time;
    }
    

void TimeManager::add_node_for_synchronization
    (
    const NodeId &      node_id
    )
    {
    if( node_id == Global::get_local_node_id() )
        {
        throw ExTimeManagementError
            ( "Cannot add local node into synchronization hash table." );
        }
    SynchronizationInfo      synch_info; // Blank synchronization info.

    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        // Server.
        if( !sync_info_hash_server.insert
                (
                SynchronizationInfoHash::value_type( node_id, synch_info )
                ).second
          )
            {
            std::ostringstream  ss;
            ss << "Server node (node id " << node_id << ") already exists in synchronization hash table.";
            throw ExNodeAlreadyExistInSynchronizationHashTable( ss.str() );
            }
        }
    else
        {
        // Client.
        if( !sync_info_hash_client.insert
                (
                SynchronizationInfoHash::value_type( node_id, synch_info )
                ).second
          )
            {
            std::ostringstream  ss;
            ss << "Client node (node id " << node_id << ") already exists in synchronization hash table.";
            throw ExNodeAlreadyExistInSynchronizationHashTable( ss.str() );
            }
        // Send time sync message immediately to client.
        TimeSynchronizationMessage msg
            (
            local_synch_info.synchronized,
            get_time(),
            time_sync_interval.get_value()
            );
        msg.send( node_id, NetworkCommon::STREAM_RELIABLE );
        TimeMagicConstantMessage msg2( real_time_sync.magic_constant );
        msg2.send( node_id, NetworkCommon::STREAM_RELIABLE );
        }
    }


bool TimeManager::remove_node_from_synchronization
    (
    const NodeId &      node_id
    )
    {
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        // Server.
        SynchronizationInfoHash::iterator   iter = sync_info_hash_server.find( node_id );
        if( iter == sync_info_hash_server.end() )
            {
            return false;
            }
        sync_info_hash_server.erase( iter );
        }
    else
        {
        // Client.
        SynchronizationInfoHash::iterator   iter = sync_info_hash_client.find( node_id );
        if( iter == sync_info_hash_client.end() )
            {
            return false;
            }
        sync_info_hash_client.erase( iter );
        }    
    return true;
    }


bool TimeManager::is_time_synchronized() const
    {
    if( local_synch_info.synchronized  )
        return true;
    return false;
    }
    

void TimeManager::send_synchronization_info()
    {
    if( test )
        {
        return;
        }
    // Client can send sync message too, but server handle this message
    // differently than the same message from server node.
    TimeSynchronizationMessage msg
        (
        local_synch_info.synchronized,
        get_time(),
        time_sync_interval.get_value()
        );

    if( System::is_server() )
        {
        // Local node is server.
        SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
        while( it != sync_info_hash_server.end() )
            {
            // Send message to server nodes.
            msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
            ++it;
            }
/*        it = sync_info_hash_client.begin();
        while( it != sync_info_hash_client.end() )
            {
            // Send message to client nodes.
            msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
            ++it;
            }*/
        }
    else
        {
        // Local node is client
        /*hope( sync_info_hash_server.size() == 1 );
        SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
        msg.send( it->first, NetworkCommon::STREAM_RELIABLE );*/
        }
        
    set_last_server_sync_system_time();
    }


bool TimeManager::synchronize_time()
    {
    int         num = 0; // Number of nodes used to compute synchronized massiv time.
    Float64     tmp_time = 0;

    // Try to compute synchronized massiv time from synchronized nodes first.
    // If found at least one synchronized node, massiv time will
    // be synchronized for local node.
    SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
    while( it != sync_info_hash_server.end() )
        {
        SynchronizationInfo & si = it->second;
        if( ( si.valid ) && ( si.synchronized ) )
            {
            num++;
            Float64 value = si.massiv_time.get_value() - si.deliver_time.get_value();
            if( value < 0 && System::is_server() )
                {
                value = 0.0;
                }
            if( test && value < 0 )
                {
                value = 0.0;
                }
            tmp_time += value;
            si.valid = false;   // Information can be used just once.
            }
        ++it;
        }
    if( num != 0 )
        {
        local_synch_info.increment = tmp_time / num;  // Arithmetic mean.
        local_synch_info.synchronized = true;
        }
    else
        {
        if( !local_synch_info.synchronized )
            {
            // No synchronized node was found - let's compute with non-synchronized
            // valid information.
            // I have to find the biggest value of nonsynchronized data && every
            // node has to have valid information in hash table. Then I can
            // say I have computed synchronized massiv time.
            num = 0;
            unsigned int    num_valid = 0; // Number of valid items in hash table.
            unsigned int    num_server = 0; // Number of servers in hash table.
            SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
            while( it != sync_info_hash_server.end() )
                {
                num_server++;
                SynchronizationInfo & si = it->second;
                Float64 tmp_time2 = 0;
                if( si.valid )
                    {
                    num_valid++;
                    tmp_time2 = si.massiv_time.get_value() - si.deliver_time.get_value();
                    if( tmp_time2 > tmp_time )
                        {
                        // Looking for the biggest time.
                        tmp_time = tmp_time2;
                        }
                    }
                ++it;
                }
            // I can say i'm synchronized when i have valid information from all servers.
            if( num_valid == num_server )
                {
                local_synch_info.increment = tmp_time;
                local_synch_info.synchronized = true;
                num = 1;   // synchronized with one server
                }
            }
        }
    if( num )
        {
        adapt_time_manager_synchronization_values();
        return true;
        }
    return false;
    }


void TimeManager::adapt_time_manager_synchronization_values()
    {        
    if( System::is_server() )
        {
        // Only server's time synchronization interval is changed during simulation.
        // Client's time synchronization interval is fixed so it doesnt need to be
        // changed.
        if( local_synch_info.increment == 0 )
            {
            // Time synchronization interval can be increased.
            if( time_sync_interval < TM_MAX_TIME_SYNC_INTERVAL )
                {
                Float64 temp = time_sync_interval.get_value();
                temp *= 2;
                time_sync_interval = temp;
                if( time_sync_interval > TM_MAX_TIME_SYNC_INTERVAL )
                    {
                    time_sync_interval = TM_MAX_TIME_SYNC_INTERVAL;
                    }
                }
            }
        else
            {
            // Time synchronization interval must be decreased.
            if( time_sync_interval > TM_MIN_TIME_SYNC_INTERVAL )
                {
                Float64 temp = time_sync_interval.get_value();
                temp /= 2;
                time_sync_interval = temp;
                if( time_sync_interval < TM_MIN_TIME_SYNC_INTERVAL )
                    {
                    time_sync_interval = TM_MIN_TIME_SYNC_INTERVAL;
                    }
                }
            }
        find_minimal_time_sync_interval();
        }
    else
        {
        // Client. See if we should slow down massiv time.
        if( local_synch_info.increment < TM_CLIENT_SLOWDOWN_TIME_VALUE )
            {
            // Client should be slowed down. This could be caused by network delay.
            // So we check if this value have came TM_CLIENT_NUM_SLOWDOWN times.
            local_synch_info.num_slow_down++;
            if( local_synch_info.num_slow_down != TM_CLIENT_NUM_SLOWDOWN )
                {
                local_synch_info.increment = 0.0;
                }
            else
                {
                // Value of local_synch_info.increment is leaved, so client will be
                // slowed down.
                local_synch_info.num_slow_down = 0;
                }
            }
        else
            {
            if( local_synch_info.increment < 0 )
                {
                // Normal negative value of local_synch_increment.
                local_synch_info.num_slow_down = 0;
                local_synch_info.increment = 0.0;
                }
            }
        }
    }


TimeManager::SynchronizationInfo TimeManager::get_local_synchronization_info()
    {
    local_synch_info.sync_interval = time_sync_interval;
    return local_synch_info;
    }

    
void TimeManager::set_remote_synchronization_info
    (
    const NodeId &                  node_id,
    const SynchronizationInfo &     synchronization_info
    )
    {
    SynchronizationInfoHash::iterator   iter;
    if( node_id.is_type( NodeId::TYPE_SERVER ) )
        {
        iter = sync_info_hash_server.find( node_id );
        if( iter == sync_info_hash_server.end() )
            {
            std::ostringstream  ss;
            ss << "Server node (node id " << node_id << ") doesn't exist in synchronization hash table";
            throw ExNodeNotExistInSynchronizationHashTable( ss.str() );
            }
        }
    else
        {
        iter = sync_info_hash_client.find( node_id );
        if( iter == sync_info_hash_client.end() )
            {
            std::ostringstream  ss;
            ss << "Client node (node id " << node_id << ") doesn't exist in synchronization hash table";
            throw ExNodeNotExistInSynchronizationHashTable( ss.str() );
            }
        }

    SynchronizationInfo & synch_info = iter->second;
    synch_info = synchronization_info;

    synch_info.deliver_time = get_time();
    synch_info.valid = true;
    
    if( test )
        {
        // No message handling in test mode - so searching for minimal time sync value
        // mus be done here.
        find_minimal_time_sync_interval();
        }
    }


void TimeManager::allow_synchronization()
    {
    if( test )
        {
        synchronization_allowed = true;
        return;
        }
    synchronization_allowed = true;
    }


void TimeManager::set_time
    (
    const STime & new_time
    )
    {
    if( test )
        return;

    zero_system_time = Global::system_time();
    system_time.invalidate();
    last_server_sync_system_time.invalidate();
    sync_info_hash_server.clear();
    sync_info_hash_client.clear();
    local_synch_info = SynchronizationInfo();
    synchronization_allowed = false;
    paused = true;
    pause_start.invalidate();
    pause_end.invalidate();
    pause_delta.invalidate();

    current_time = new_time;
    }


TimeManager::SynchronizationInfo::SynchronizationInfo() :
    valid( false ),
    synchronized( false ),
    num_slow_down( 0 ),
    sync_interval( TM_MIN_TIME_SYNC_INTERVAL ),
    increment( 0.0 )
    {
    }


TimeManager::SynchronizationInfo::SynchronizationInfo
    (
    bool                    _synchronized,
    const SystemTime &      _massiv_time,
    const SystemTime &      _sync_interval
    ) :
    valid( true ),
    synchronized( _synchronized ),
    num_slow_down( 0 ),
    sync_interval( _sync_interval ),
    massiv_time( _massiv_time ),
    increment( 0.0 )
    {
    }


void TimeManager::lock()
    {
    if( !test )
        throw ExCommonError( "Time manager is not in test mode. Call of lock() is not allowed." );
    cs->enter();
    }


void TimeManager::unlock
    (
    )
    {
    if( !test )
        throw ExCommonError( "Time manager is not in test mode. Call of unlock() is not allowed." );
    cs->leave();
    }


void TimeManager::handle_time_synchronization_message
    (
    const TimeSynchronizationMessage &  time_synch_msg
    )
    {
    TimeManager::SynchronizationInfo synch_info
        (
        time_synch_msg.synchronized,
        time_synch_msg.massiv_time,
        time_synch_msg.sync_interval
        );
    set_remote_synchronization_info
        (
        time_synch_msg.get_initiator(),
        synch_info
        );
    if( System::is_server() )
        {
        find_minimal_time_sync_interval();
        }
    }


void TimeManager::handle_time_magic_constant_message
    (
    const TimeMagicConstantMessage &  time_mc_msg
    )
    {
    real_time_sync.magic_constant = time_mc_msg.magic_constant;
    hope( TM_MC_TOLERANCE < 1 );
    if( real_time_sync.magic_constant < 1 - TM_MC_TOLERANCE )
        {
        std::ostringstream  ss;
        ss  << "time_magic_constant = " << real_time_sync.magic_constant 
            << ", the value is too low.";
        Global::log
            (
            STATUS_WARNING( Status::FACILITY_TIME_MANAGEMENT, Status::
                PRIORITY_LOWEST ),
            ss.str()
            );
        real_time_sync.magic_constant = 1 - TM_MC_TOLERANCE;
        }
    if( real_time_sync.magic_constant > 1 + TM_MC_TOLERANCE )
        {
        std::ostringstream  ss;
        ss  << "time_magic_constant = " << real_time_sync.magic_constant 
            << ", the value is too high.";
        Global::log
            (
            STATUS_WARNING( Status::FACILITY_TIME_MANAGEMENT, Status::
                PRIORITY_LOWEST ),
            ss.str()
            );
        real_time_sync.magic_constant = 1 + TM_MC_TOLERANCE;
        }
    hope( real_time_sync.magic_constant > 0 );
    // Resend message to client nodes.
    TimeMagicConstantMessage msg( real_time_sync.magic_constant );
    if( System::is_server() )
        {
        SynchronizationInfoHash::iterator it = sync_info_hash_client.begin();
        while( it != sync_info_hash_client.end() )
            {
            msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
            ++it;
            }
        }
    }

bool TimeManager::should_synchronize_time() const
    {
    if( !is_time_synchronized() )
        {
        // Time is not synchronized - send message every TM_INIT_SYNC_INTERVAL second.        
        static SystemTime   last_time = Global::system_time() - TM_INIT_SYNC_INTERVAL;
        SystemTime          act_time = Global::system_time();
        if( act_time - last_time >= TM_INIT_SYNC_INTERVAL )
            {
            last_time = act_time;
            return true;
            }        
        }
    // @a minimum_time_sync_value is never changed for 
    // a client and its value is always TM_MIN_TIME_SYNC_INTERVAL
    if( system_time - last_server_sync_system_time > minimum_time_sync_value )
        {
        return true;
        }
    return false;
    }


void TimeManager::set_last_server_sync_system_time()
    {
    last_server_sync_system_time = system_time;
    }
    

void TimeManager::pause()
    {
    if( paused )
        {
        throw ExTimeManagementError( "Cannot pause() time manager twice." );
        }
    pause_start = compute_system_time();
    system_time = pause_start;
    paused = true;

    real_time_sync.valid = false;
    }


void TimeManager::unpause()
    {
    if( !paused )
        {
        throw ExTimeManagementError( "Cannot unpause() not paused time manager." );
        }
    pause_end = compute_system_time();
    hope( pause_end >= pause_start );
    pause_delta += pause_end - pause_start;
    paused = false;
    }


bool TimeManager::is_paused() const
    {
    return paused;
    }


TimeManager::TimeManagerSettings::TimeManagerSettings() :
    Settings( "TimeManager" ),
    time_synchronization_interval( access_integer( "init_time_sync_interval", TM_MIN_TIME_SYNC_INTERVAL ) ),
    time_speed_constant( access_float( "time_speed_constant", 1.0 ) )
    {
    }


void TimeManager::synchronize_with_real_time
    (
    Float64     real_time
    )
    {
    hope( System::is_server() );
    if( real_time_sync.valid )
        {
        // Last real time sync was successful, so now we can compute magic constant of massiv time
        // to run with the biggest accuracy with real time.
        Float64     massiv_time = get_time().get_value();
        Float64     massiv_time_delta = massiv_time - real_time_sync.massiv_time.get_value();
        massiv_time_delta /= ( Float64 )time_manager_settings.time_speed_constant;
        hope( massiv_time_delta > 0 );
        
        Float64     real_time_delta = real_time - real_time_sync.real_time;
        hope( real_time_delta > 0 );
        Float64     delta = massiv_time_delta - real_time_delta;
        
        if( fabs( delta ) >= TM_REAL_TIME_ACCURACY )
            {
            real_time_sync.magic_constant = real_time_delta / ( massiv_time_delta / real_time_sync.magic_constant );
            real_time_sync.massiv_time = get_time();
            real_time_sync.real_time = real_time;
            // Send message with magic constant to other nodes.
            TimeMagicConstantMessage msg( real_time_sync.magic_constant );
            SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
            while( it != sync_info_hash_server.end() )
                {
                msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
                ++it;
                }
            it = sync_info_hash_client.begin();
            while( it != sync_info_hash_client.end() )
                {
                msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
                ++it;
                }
            }
        }
    else
        {
        real_time_sync.massiv_time = get_time();
        real_time_sync.real_time = real_time;
        real_time_sync.valid = true;
        }
    }
    
    
TimeManager::RealTimeSync::RealTimeSync() :
    real_time( 0.0 ),
    valid( false ),
    magic_constant( 1.0 )
    {
    }


void TimeManager::find_minimal_time_sync_interval()
    {
    minimum_time_sync_value = time_sync_interval; // Local value.
    SynchronizationInfoHash::iterator it = sync_info_hash_server.begin();
    while( it != sync_info_hash_server.end() )
        {
        const SynchronizationInfo & si = it->second;
        if( si.sync_interval < minimum_time_sync_value )
            {
            minimum_time_sync_value = si.sync_interval;
            }
        ++it;
        }
    }


void TimeManager::synchronize_clients()
    {
    hope( System::is_server() );

    if( system_time - last_client_sync_system_time >= TM_CLIENT_TIME_SYNC_INTERVAL )
        {
        last_client_sync_system_time = system_time;
        TimeSynchronizationMessage msg
            (
            local_synch_info.synchronized,
            get_time(),
            time_sync_interval.get_value()
            );
        SynchronizationInfoHash::iterator it = sync_info_hash_client.begin();
        while( it != sync_info_hash_client.end() )
            {
            msg.send( it->first, NetworkCommon::STREAM_RELIABLE );
            ++it;
            }
        }
    }
    

} // namespace Massiv::Core
} // namespace Massiv
