/** @file time_manager.h
 * Simulation time synchronization.
 */

/*
 * begin        : Fri Aug 15 2002
 * copyright    : (C) 2002 by Petr Tovarys, Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_TIME_MANAGER_H
#define MASSIV_TIME_MANAGER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_STYPE_TIME_H
#include "serializer/stype_time.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#ifndef MASSIV_NODEID_H
#include "node/nodeid.h"
#endif

#ifndef MASSIV_SYNCH_H
#include "thread/synch.h"
#endif

#ifndef MASSIV_SYSTEM_EVENT_FWD_H
#include "sysevent/system_event_fwd.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif


namespace Massiv {
namespace Core {

namespace Test {
    class TimeTest;
}


/** @addtogroup time
 * @{
 */

/**
 * %Time manager synchronizes simulation time among nodes.
 *
 * %Time manager will be in paused state after creation and time 
 * synchronization will be disabled.
 *
 * Local node is running as single massiv server
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * @code
 * static TimeManager &    time_manager = Global::time_manager();
 * time_manager.set_time( 10 ); // 10 seconds.
 * time_manager.unpause();
 * @endcode
 *
 * Local node is running with other nodes to synchronize time with
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * @code
 * static TimeManager &    time_manager = Global::time_manager();
 * time_manager.set_time( 10.0 ); // 10 seconds.
 * time_manager.add_node_for_synchronization( NodeId( 123 ) );
 * time_manager.add_node_for_synchronization( NodeId( 124 ) );
 * time_manager.allow_synchronization();
 * time_manager.unpause(); 
 * @endcode
 * FIXME: prepsat priklad - je zastaraly
 */
class TimeManager
    {
    friend class ObjectManager;
    friend class Test::TimeTest;

public:

    /**
     * Contain time synchronization information about one massiv server.
     * Each TimeManager has hash table of instances of this class and uses this array
     * to compute its synchronization information (@param TimeManager::local_synch_info).
     */
    class SynchronizationInfo
        {

        friend class TimeManager;
        friend class Test::TimeTest;

    public:

        /** Constructor. */
        SynchronizationInfo();

        /** Constructor. */
        SynchronizationInfo
            (
            bool                    _synchronized,
            const SystemTime &      _massiv_time,
            const SystemTime &      _sync_interval
            );

    private:

        bool            valid;
            /**< Zero if data are not valid. Not used in local node. */

        bool            synchronized;
            /**< True when massiv time is succesfully synchronized with other massiv servers. */

        STime           deliver_time;
            /**< Massiv time of a !local! server when information was delivered.
                 Function which stores information here sets this attribute. */
        
        int             num_slow_down;
            /**< Used by a client only. 
                 When @a increment is less than TM_CLIENT_SLOWDOWN_TIME_VALUE,
                 this value is icreased by one. 
                 When value is equal to TM_CLIENT_NUM_SLOWDOWN, client's massiv time
                 is slowed down.
                 Network delay can cause that a value of @a increment will be nagative 
                 and too small, so it will indicate, that client's massiv time 
                 should be slowed down. So we count number of attempts to slow down 
                 massiv time, and if a value is equal to TM_CLIENT_NUM_SLOWDOWN,
                 it's high probability that client's massiv time really should be slowed
                 down to be synchronized with its server. */
            
        SystemTime      sync_interval;
            /**< Time synchronization interval for a server node. This value is increased
                 when massiv time is not sped up and is descreased when massiv time
                 must be sped up. */

        STime           massiv_time;
            /**< Massiv time of a server. */

        Float64         increment;
            /**< Increment of massiv time computed by @method synchronize_time().
                 It's called increment, but it can be also negative value.
                 Server:
                 -when zero, massiv time is ok. 
                 -when positive, this value should be added to local massiv time 
                 to be synchronized with other massiv servers.
                 -this value is never negative for a server node.
                 Client:
                 -when negative closed to zero or zero, massiv time is ok. 
                 -when negative and less than TM_CLIENT_SLOWDOWN_TIME_VALUE,
                  client's massiv time is a candidate to be slowed down.
                 -when positive, this value should be added to local massiv time 
                 to be synchronized with other massiv servers. */
        
        };  // class SynchronizationInfo

    typedef ext::hash_map
                <
                NodeId,
                SynchronizationInfo,
                NodeId::Hash,
                NodeId::EqualKey
                > SynchronizationInfoHash;

    /**
     * Struct RealTimeSync.
     * This structure si used to modify speed of massiv time to be as fast 
     * as real time.
     */
    struct RealTimeSync
        {
        
        /** Constructor. */
        RealTimeSync();
        
        STime       massiv_time;
            /**< Massiv time which corresponds to real_time. */
            
        Float64     real_time;
            /**< Real time which corresponds to massiv_time. */
        
        bool        valid;
            /**< True when information was once stored here. */
            
        Float64     magic_constant;
            /**< Magic constant is ratio real time to massiv time.
                 massiv time delta * magic_constant = real time delta
                 Default value is 1.*/
            
        }; // RealTimeSync

    /**
     * Constructor. Default settings is paused.
     */
    TimeManager
        (
        bool init_paused = true
        );


    /**
     * FOR TESTING PURPOSES.
     * Constructor with nodeId.
     */
    TimeManager
        (
        const NodeId & node_id
        );

    /**
     * FOR TESTING PURPOSES.
     * Constructor with nodeId and initial value of systemtime.
     */
    TimeManager
        (
        const NodeId &      node_id,
        const SystemTime &  initial_time_value
        );

    /**
     * Destructor.
     */
    ~TimeManager();

    /**
     * Initialize time manager.     
     * @param FIXME
     */
    void initialize
        (
        bool init_paused = true
        );
    
    
public:

    /**
     * Runs next tick.
     * If @attribute synchronization_allowed is true, it tries to synchronize time
     * (@method synchronize_time() ) every call of next_tick() till massiv time is
     * synchronized (@attribute time_synchronized is true) AND it sends local synchronization
     * information (@method send_synchronization_info() ) to other massiv nodes.
     * When synchronized (@param time_synchronized is true), it calls above methods
     * every @param TimeManager::SYNCHRONIZE_TICK_COUNT ticks.
     */
    void next_tick();


    /**
     * Get current simulation(=massiv) time.
     * @note Simulation time stays unchanged until @c next_tick() is called
     * (may change in the future).
     */
    STime get_time() const;


    /**
     * Set new simulation time.
     * Reinitializes whole time manager and time manager is paused.
     * After this method time manager must be initialized again.
     * Used by ObjectManager to restore simulation time
     * when restoring simulation from an archive file.
     */
    void set_time
        (
        const STime & time
        );

    /**
     * Pause time manager. Call @method unpause() to continue time counting.
     */
    void pause();


    /**
     * Unpause time manager. Call @method pause() to pause time counting.
     */
    void unpause();


    /**
     * Return true if time manager is paused.
     */
    bool is_paused() const;

    /**
     * Main synchronization function. Algorithm may change in the future.
     * Tries to compute synchronized massiv time. If computed, synchronized massiv time
     * and other computed synchronization information are stored
     * in @a local_synch_info, specially @a local_synch_info.increment is the most 
     * important information which was computed.
     * Sets @at time_synchronized to true if synchronized massiv time is computed.
     * Return true when synchronized massiv time was computed.
     */
    bool synchronize_time();


    /**
     * Send local synchronization information (stored in class @attribute local_synch_info)
     * in to other massiv servers.
     */
    void send_synchronization_info();


    /**
     * Returns local synchronization information (@attribute local_synch_info).
     * This information was computed by @method compute_synchronization_data().
     */
    SynchronizationInfo get_local_synchronization_info();


    /**
     * Sets synchronization information of one massiv server into hash table
     * (@attribute synchronization_info_hash).
     * This data was delivered by system message.
     */
    void set_remote_synchronization_info
        (
        const NodeId & node_id,
        const SynchronizationInfo & synchronization_info
        );


    /**
     * Add massiv node to hash table of massiv nodes to synchronize with.
     */
    void add_node_for_synchronization
        (
        const NodeId & node_id
        );

    /**
     * Removes massiv node from synchronization hash table.
     * Returns true when node was removed.
     */
    bool remove_node_from_synchronization
        (
        const NodeId & node_id
        );

    /**
     * Returns true if time is synchronized.
     */
    bool is_time_synchronized() const;


    /**
     * Allows synchronization with other massiv servers.
     */
    void allow_synchronization();


    /**
     * Handle time synchronization message.
     * Adds information from message to synchronization hash table.
     */
    void handle_time_synchronization_message
        (
        const TimeSynchronizationMessage &  time_synch_msg
        );

    
    /**
     * Handle time magic constant message.
     * Magic constant was computed by some massiv server. 
     * Every massiv node (even clients) will receive this constant.
     */
    void handle_time_magic_constant_message
        (
        const TimeMagicConstantMessage &  time_mc_msg
        );

public: // @todo: temporary public for testing purposes

    /**
     * Return true when it is time to synchronize massiv time.
     */
    bool should_synchronize_time() const;

    /**
     * Set system time when last server synchronization was done.
     */
    void set_last_server_sync_system_time();
    
private:

    /**
     * Return system time of running (non-paused) time manager. Counted from zero.
     * Note: time spent in pause state is abated from returned value.
     * If @param ignore_current_pause_state is false, time when time manager was
     * paused is returned. Else time value within current pause is added to
     * returned value.
     */
    SystemTime compute_system_time
        (
        bool        ignore_current_pause_state = true
        ) const;


    /**
     * Compute simulation time.
     * @param addition - value to add to current_time.
     */
    STime compute_simulation_time
        (
        const SystemTime &  addition
        );

    /**
     * Modify massiv time to be as fast as real time.
     * Purpose of this method is to compute magic constant, and after this
     * send magic constant message to other massiv nodes.
     * @param real_time - real_time value. This value can be obtained from NTP server.
     *  
     * When it is called for the first time, it just stores information to
     * @a real_time_sync. When it is called for the second time, it can compute
     * ratio between real time and massiv time = magic constant.
     *
     * When it is not possible to use NTP server, or another real time synchronization
     * technique, massiv time will be as fast as the fastest hardware clock of some
     * massiv server.
     *
     * Typical use is to run the simulation, and call this method. After some hours (days)
     * it is necessary to call this method again -\> and magic constant will be computed
     * and sent to other massiv nodes included clients. Note that time manager must not be
     * paused between calls of this method. Pausing of time manager will cause invalidation
     * of @a real_time_sync.
     */
    void synchronize_with_real_time
        (
        Float64     real_time
        );

    /**
     * This function finds a minimal value of time synchronization interval among all
     * servers in simulation.
     */
    void find_minimal_time_sync_interval();
    
    /**
     * Send time synchronization message to clients.
     */
    void synchronize_clients();

    /**
     * Called in @method synchronize_time() and adapt some synchronization values
     *
     * Server: 
     * Every server computes its own time synchronization interval value every
     * call of @method synchronize_time(). This value depends on value of @a
     * local_synch_info.increment which is set after a call of method synchronize_time().
     * If @a local_synch_info.increment is zero, time synchronization interval is increased
     * If @a local_synch_info.increment is positive value, time synchronization interval
     * is reduced to achieve more accurate time synchronization.
     *
     * Ideal situation is when all servers are synchronized with zero 
     * local_synch_info.increment, so time synchronization interval value can be 
     * the greatest value - TM_MAX_TIME_SYNC_INTERVAL.
     *
     * Client:
     * Massiv time is managed by massiv servers hardware clock. 
     * However clients hardware clock can be faster than the fastest clock of any massiv
     * server. This fact leads to slowing down of clients massiv time.
     * So in a client case this method decides whether to slow down client massiv time.
     */
    void adapt_time_manager_synchronization_values();    

public: 

    /**
     * TimeManager settings.
     */
    class TimeManagerSettings : public Settings
        {
        friend class Test::TimeTest;
    
    public:

        /** Constructor. */
        TimeManagerSettings();

    public:

        Integer &           time_synchronization_interval;
            /**< Interval between time synchronization messages. */
        
        Float &             time_speed_constant;
            /**< FOR TESTING PURPOSES. */

        }; // TimeManagerSettings.

private:

    STime                       current_time;
        /**< Current simulation time. This time doesn't change during one tick. */

    SystemTime                  zero_system_time;
        /**< Starting system time. */

    SystemTime                  system_time;
        /**< System time of time manager. Doesn't change during one tick.
             Counted from zero. */

    SystemTime                  last_server_sync_system_time;
        /**< Last system time when @method synchronize_time() was called. */

    SystemTime                  last_client_sync_system_time;
        /**< Last system time when time sync message was sent to clients. */

    SynchronizationInfoHash     sync_info_hash_server;
        /**< List of server synchronization information.
             One item for one massiv server. */

    SynchronizationInfoHash     sync_info_hash_client;
        /**< List of client synchronization information.
             One item for one massiv client connected to this server. */

    SynchronizationInfo         local_synch_info;
        /**< Local server stores here its computed synchronization information. */

    bool                        synchronization_allowed;
        /**< When set to true, time manager will synchronize time with other massiv servers. */

    bool                        paused;
        /**< True when time manager is paused. */

    SystemTime                  pause_start;
        /**< System time when time manager was paused. */

    SystemTime                  pause_end;
        /**< System time when time manager was unpaused. */

    SystemTime                  pause_delta;
        /**< Sum of system time spent in paused state. */

    TimeManagerSettings         time_manager_settings;
        /**< Time manager settings. */

    RealTimeSync                real_time_sync;
        /**< Real time synchronization information. */
        
    SystemTime                  minimum_time_sync_value;
        /**< Smallest value of time sync interval values of all servers. */
        
    SystemTime                  time_sync_interval;
        /**< Interval between time synchronization messages. */
    
public:

    bool                        test;
        /**< True when node manager is created in test mode. */

    NodeId                      my_node_id;
        /**< FOR TESTING PURPOSES. It will be removed later. */

    CriticalSection *           cs;
        /**< FOR TESTING PURPOSES. It will be removed later. */

    void lock();
        /**< FOR TESTING PURPOSES. It will be removed later. */
    void unlock();
        /**< FOR TESTING PURPOSES. It will be removed later. */

    }; // TimeManager

/** @} */ // addtogroup time


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_TIME_MANAGER_H
