/** @file array.h
 * Arrays.
 */

/*
 * begin        : Tue Apr 09 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_ARRAY_H
#define MASSIV_ARRAY_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#ifndef MASSIV_DEBUG_STORAGE_H
#include "utility/debug_storage.h"
#endif

#include <memory>

namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * DefaultArrayHandlers.
 */
class DefaultArrayHandlers
    {
public:

    typedef std::size_t SizeType;

    /**
     * Called by storage when item at @a position is going to be accessed.
     */
    void read_requested
        (
        SizeType position
        ) const
        {
        }

    /**
     * Called by storage when item at @a position will be modified.
     */
    void write_requested
        (
        SizeType position
        )
        {
        }

    /**
     * Called by storage when array size changes.
     */
    void size_changed
        (
        SizeType old_size,
        SizeType new_size
        )
        {
        }

    }; // DefaultArrayHandlers

/** @class StaticStorage.
 * Storage for type @a T of maximum capacity @a max_items.
 */
template <class T, int max_items, class H = DefaultArrayHandlers>
class StaticStorage : public DebugStorage, public H
    {
    /**
     * Defines "T * items" pointing to the native array.
     */
    #define items   reinterpret_cast<T *>( raw_storage )

public:

    typedef T                               ValueType;
    typedef StaticStorage<T, max_items, H>  Storage;
    typedef unsigned int                    SizeType;
    typedef int                             DifferenceType;
    typedef Storage *                       StoragePointer;
    typedef ValueType &                     ReferenceType;
    typedef const ValueType &               ConstReferenceType;

protected:
    
    using H::read_requested;
    using H::write_requested;
    using H::size_changed;

public:

    //
    // Storage interface implementation.
    //

    /**
     * Resizes the container.
     */
    void resize
        (
        SizeType    new_size,
        const T &   value = T()
        )
        {
        SizeType old_size = size();

        invalidate_iterators();
        range_error( new_size >= 0 && new_size <= max_items, "resize()" );

        if( new_size > num_items )
            {
            std::uninitialized_fill( items + num_items, items + new_size, value );
            }
        else
            {
            // Stoupik: no std::destroy() in c++ standard, it's STL extension
            //std::destroy( items + new_size, items + num_items );

            T * item = items + new_size;
            for( SizeType index = new_size; index < num_items; index++, item++ )
                {
                item->~T();
                }
            }

        num_items = new_size;

        size_changed( old_size, new_size );
        }

    /**
     * Reserves at least specified storage capacity.
     */
    void reserve
        (
        SizeType size
        )
        {
        invalidate_iterators();
        range_error( size <= max_items, "reserve()" );
        }

public:

    //
    // Common implementation.
    //

    #define COMMON_IMPLEMENTATION() \
        SizeType size() const \
            { \
            return num_items; \
            } \
        \
        SizeType max_size() const \
            { \
            return max_items; \
            } \
        \
        T * get_buffer() \
            { \
            if( !items ) \
                buffer_error(); \
            return items; \
            } \
        \
        const T * get_buffer() const \
            { \
            if( !items ) \
                buffer_error(); \
            return items; \
            }

    #define ASSIGN_IMPLEMENTATION() \
        template <class T2> \
        Storage & operator= \
            ( \
            const T2 & array \
            ) \
            { \
            SizeType i; \
            SizeType new_size = array.size(); \
            resize( new_size ); \
            for( i = 0; i < new_size; i++ ) \
                { \
                write_at( i ) = array.at( i ); \
                } \
            return *this; \
            }

    #define AT_IMPLEMENTATION() \
        ReferenceType write_at \
            ( \
            SizeType    index \
            ) \
            { \
            range_error( index >= 0 && index < size(), "write_at()" ); \
            write_requested( index ); \
            return get_buffer()[ index ]; \
            } \
        \
        ConstReferenceType at \
            ( \
            SizeType    index \
            ) const \
            { \
            range_error( index >= 0 && index < size(), "at()" ); \
            read_requested( index ); \
            return get_buffer()[ index ]; \
            }

    #define OPERATOR_IMPLEMENTATION() \
        ReferenceType operator() \
            ( \
            SizeType    index \
            ) \
            { \
            return write_at( index ); \
            } \
        \
        ConstReferenceType operator[] \
            ( \
            SizeType    index \
            ) const \
            { \
            return at( index ); \
            }

    #define CLEAR_IMPLEMENTATION() \
        void clear() \
            { \
            resize( 0 ); \
            }

    #define POSITION_IMPLEMENTATION() \
        SizeType reference_position \
            ( \
            ConstReferenceType item \
            ) const \
            { \
            return &item - get_buffer(); \
            } \

public:

    COMMON_IMPLEMENTATION();
    ASSIGN_IMPLEMENTATION();
    AT_IMPLEMENTATION();
    OPERATOR_IMPLEMENTATION();
    CLEAR_IMPLEMENTATION();
    POSITION_IMPLEMENTATION();

    #undef items

public:

    /**
     * Default constructor.
     */
    StaticStorage() : num_items( 0 )
        {
        }

    /**
     * Copy constructor.
     */
    template <class A>
    StaticStorage
        (
        const A & array
        ) : num_items( 0 )
        {
        operator=( array );
        }

    /**
     * Destructor.
     */
    ~StaticStorage()
        {
        clear();
        }

private:

    SizeType        num_items;
        /**< Current number of valid items in the array. */
    mutable char    raw_storage[ max_items * sizeof( T ) ];
        /**< Array storage. */

    }; // StaticStorage


/** @class UnmanagedStorage
 * UnmanagedStorage works as a proxy storage for normal C/C++ array.
 * Storage for type @a T whose buffer is allocated outside
 * the storage and is not managed by the storage.
 * External buffer must remain valid until storage deletion.
 * Storage can not be resized because we do not have control over
 * the storage (can not erase or insert items).
 */
template <class T, class H = DefaultArrayHandlers>
class UnmanagedStorage : public DebugStorage, public H
    {
public:

    typedef T                       ValueType;
    typedef UnmanagedStorage<T, H>  Storage;
    typedef unsigned int            SizeType;
    typedef int                     DifferenceType;
    typedef Storage *               StoragePointer;
    typedef ValueType &             ReferenceType;
    typedef const ValueType &       ConstReferenceType;

protected:
    
    using H::read_requested;
    using H::write_requested;
    using H::size_changed;

public:

    //
    // Storage interface implementation.
    //

    /**
     */
    void resize
        (
        SizeType    size,
        const T &   value = T()
        )
        {
        invalidate_iterators();
        range_error( false, "resize()" );
        }

    /**
     */
    void reserve
        (
        SizeType    size
        )
        {
        invalidate_iterators();
        range_error( size <= max_items, "reserve()" );
        }

    /**
     * Binds the storage to external @a buffer.
     * Note! Do not use this storage for sharing
     * buffers among arrays, use SharedStorage instead (which is safe).
     */
    void bind
        (
        T *         buffer,
        SizeType    valid_items
        )
        {
        invalidate_iterators();

        num_items = valid_items;
        max_items = valid_items;
        items = buffer;
        }

    /**
     * Unbinds the storage from the external buffer.
     */
    void unbind()
        {
        bind( NULL, 0 );
        }

    // Note that operator= won't work even if sizes match, but it's required
    // by normalized array. clear() is defined for completeness, but it won't
    // work either. I think that resize() could be implemented, but Markoid
    // is probably too lazy and it's not required anywhere yet.

    COMMON_IMPLEMENTATION();
    ASSIGN_IMPLEMENTATION();
    AT_IMPLEMENTATION();
    OPERATOR_IMPLEMENTATION();
    CLEAR_IMPLEMENTATION();
    POSITION_IMPLEMENTATION();

public:

    /**
     * Default constructor.
     */
    UnmanagedStorage() :
        num_items( 0 ),
        max_items( 0 ),
        items( NULL )
        {
        }

    /**
     * Destructor.
     */
    ~UnmanagedStorage()
        {
        }

private:

    SizeType    num_items;
        /**< Current number of valid items in the array. */
    SizeType    max_items;
        /**< Maximum number of items in the array. Equals to num_items. */
    T *         items;
        /**< Buffer. */

    }; // UnmanagedStorage

    
/** @class DynamicStorage.
 * Storage for type @a T of potentially unlimited capacity.
 */
template <class T, class H = DefaultArrayHandlers>
class DynamicStorage : public DebugStorage, public H
    {
public:

    typedef T                       ValueType;
    typedef DynamicStorage<T, H>    Storage;
    typedef unsigned int            SizeType;
    typedef int                     DifferenceType;
    typedef Storage *               StoragePointer;
    typedef ValueType &             ReferenceType;
    typedef const ValueType &       ConstReferenceType;

protected:
    
    using H::read_requested;
    using H::write_requested;
    using H::size_changed;

public:

    //
    // Storage interface implementation.
    //

    /**
     */
    void resize
        (
        SizeType    new_size,
        const T &   value = T()
        )
        {
        SizeType old_size = size();

        invalidate_iterators();

        if( new_size > num_items )
            {
            grow( new_size, max_items, value );
            }
        else
            {
            shrink( new_size );
            }

        size_changed( old_size, new_size );
        }

    /**
     */
    void reserve
        (
        SizeType size
        )
        {
        invalidate_iterators();
        grow( num_items, size, T() );
        }

    COMMON_IMPLEMENTATION();
    ASSIGN_IMPLEMENTATION();
    AT_IMPLEMENTATION();
    OPERATOR_IMPLEMENTATION();
    CLEAR_IMPLEMENTATION();
    POSITION_IMPLEMENTATION();

public:

    /**
     * Default constructor.
     */
    DynamicStorage() :
        num_items( 0 ),
        max_items( 0 ),
        items( NULL )
        {
        }

    /**
     * Copy constructor.
     */
    template <class A>
    DynamicStorage
        (
        const A & array
        ) : num_items( 0 ),
            max_items( 0 ),
            items( NULL )
        {
        operator=( array );
        }

    /**
     * Destructor.
     */
    ~DynamicStorage()
        {
        destroy();
        }

private:

    /**
     * Grows the array.
     */
    void grow
        (
        SizeType    new_num_items,
        SizeType    new_max_items,
        const T &   value
        )
        {
        if( new_max_items < max_items )
            new_max_items = max_items;

        if( new_num_items < num_items )
            new_num_items = num_items;

        if( new_max_items < new_num_items )
            new_max_items = new_num_items;

        if( new_max_items > max_items )
            {
            char *  raw_storage;
            T *     new_items;

            if( new_max_items < 2 * max_items )
                new_max_items = 2 * max_items + 1;

            /* Allocate new array of T without T::T() being called on allocated items. */

            raw_storage = new char[ new_max_items * sizeof( T ) ];
            new_items = reinterpret_cast<T *>( raw_storage );

            /**
             * Call copy constructors for existing items and
             * default constructors for new items.
             * Leave items <new_num_items, new_max_items) unitialized.
             */

            std::uninitialized_copy( items, items + num_items, new_items );
            std::uninitialized_fill( new_items + num_items, new_items + new_num_items, value );

            /* Replace old array of T with the new array. */

            destroy();

            items = new_items;
            num_items = new_num_items;
            max_items = new_max_items;
            }
        else
            {
            std::uninitialized_fill( items + num_items, items + new_num_items, value );
            num_items = new_num_items;
            }
        }

    /**
     * Shrinks the array and deletes the items behind the new end of the array.
     */
    void shrink
        (
        SizeType new_num_items
        )
        {
        if( new_num_items > num_items )
            return;

        // Stoupik: std::destroy is not part of the standard.
        //std::destroy( items + new_num_items, items + num_items );

        T * item = items + new_num_items;
        for( SizeType index = new_num_items; index < num_items; index++, item++ )
            {
            item->~T();
            }

        num_items = new_num_items;
        }

    /**
     * Completely deletes whole array.
     */
    void destroy()
        {
        char * raw_storage;

        if( items )
            {
            shrink( 0 );

            raw_storage = reinterpret_cast<char *>( items );
            delete [] raw_storage;

            items = NULL;
            num_items = 0;
            max_items = 0;
            }
        }

private:

    SizeType    num_items;
        /**< Current number of valid items in the array. */
    SizeType    max_items;
        /**< Current number of allocated items in the array.
             Items in range <num_items, max_items) are always uninitialized. */
    T *         items;
        /**< Array storage. */

    }; // DynamicStorage

/** @class SharedStorage.
 * This storage works as a proxy storage for storage @a S.
 * Storage allows us to create array instances that safely share
 * the same data items.
 * Note that those operations that invalidate iterators also invalidate
 * iterators from other arrays that share that same storage.
 * This is achieved by registering iterators both to the proxy storage
 * and the referenced shared storage.
 */
template <class S, class H = DefaultArrayHandlers>
class SharedStorage : public DebugStorage, public H
    {
private:

    struct SharedS : public SharedObject, public S
        {
        };

public:

    typedef typename S::ValueType       ValueType;
    typedef typename S::SizeType        SizeType;
    typedef typename S::DifferenceType  DifferenceType;
    typedef SharedStorage<S, H>         Storage;
    typedef Reference<SharedS>          StoragePointer;
    typedef ValueType                   T;
    typedef ValueType &                 ReferenceType;
    typedef const ValueType &           ConstReferenceType;

protected:
    
    using H::read_requested;
    using H::write_requested;
    using H::size_changed;

public:

    /**
     * Default constructor.
     */
    SharedStorage()
        {
        }

    /**
     * Copy constructor.
     */
    SharedStorage
        (
        const Storage & array
        ) : storage( array.storage )
        {
        }

    /**
     * Destructor.
     */
    ~SharedStorage()
        {
        }

public:

    //
    // Storage interface implementation.
    //

    /**
     */
    SizeType size() const
        {
        if( !storage.is_null() )
            return storage->size();
        else
            return 0;
        }

    /**
     */
    SizeType max_size() const
        {
        if( !storage.is_null() )
            return storage->max_size();
        else
            return 0;
        }

    /**
     */
    void resize
        (
        SizeType    new_size,
        const T &   value = T()
        )
        {
        SizeType old_size = size();

        if( storage.is_null() )
            bind_storage( new SharedS() );

        storage->resize( new_size, value );

        size_changed( old_size, new_size );
        }

    /**
     */
    void reserve
        (
        SizeType size
        )
        {
        if( storage.is_null() )
            bind_storage( new SharedS() );

        storage->reserve( size );
        }

    /**
     */
    void clear()
        {
        if( !storage.is_null() )
            storage->clear();
        }

    /**
     */
    T * get_buffer()
        {
        if( !storage.is_null() )
            return storage->get_buffer();
        else
            return NULL;
        }

    /**
     */
    const T * get_buffer() const
        {
        if( !storage.is_null() )
            return storage->get_buffer();
        else
            return NULL;
        }

    /**
     */
    Storage & operator=
        (
        const Storage & array
        )
        {
        bind_storage( array.storage );
        return *this;
        }

    /**
     */
    ReferenceType write_at
        (
        SizeType    index
        )
        {
        range_error( !storage.is_null(), "write_at()" );
        write_requested( index );
        return storage->write_at( index );
        }

    /**
     */
    ConstReferenceType at
        (
        SizeType    index
        ) const
        {
        range_error( !storage.is_null(), "at()" );
        read_requested( index );
        return storage->at( index );
        }

    /**
     */
    ReferenceType operator()
        (
        SizeType    index
        )
        {
        range_error( !storage.is_null(), "operator()" );
        write_requested( index );
        return storage->operator()( index );
        }

    /**
     */
    ConstReferenceType operator[]
        (
        SizeType    index
        ) const
        {
        range_error( !storage.is_null(), "operator[]" );
        read_requested( index );
        return storage->operator[]( index );
        }

    /**
     */
    SizeType reference_position
        (
        ConstReferenceType item
        ) const
        {
        range_error( !storage.is_null(), "reference_position()" );
        return storage->reference_position( item );
        }

protected:

    //
    // Debugging.
    //

    /**
     */
    void register_iterator
        (
        const DebugIterator & iterator
        ) const
        {
        DebugStorage::register_iterator( iterator );

        if( !storage.is_null() )
            storage->register_iterator( iterator );
        }

    /**
     */
    void unregister_iterator
        (
        const DebugIterator & iterator
        ) const
        {
        DebugStorage::unregister_iterator( iterator );

        if( !storage.is_null() )
            storage->unregister_iterator( iterator );
        }

    /**
     */
    void invalidate_iterators() const
        {
        DebugStorage::invalidate_iterators();

        if( !storage.is_null() )
            storage->invalidate_iterators();
        }

private:

    /**
     * Binds this proxy storage to the storage @a s.
     */
    void bind_storage
        (
        Reference<SharedS>  s
        )
        {
        transfer_iterators( storage.get(), s.get() );
        storage = s;
        }

private:

    Reference<SharedS>  storage;
        /**< Array storage. */

    }; // SharedStorage

#undef COMMON_IMPLEMENTATION
#undef ASSIGN_IMPLEMENTATION
#undef AT_IMPLEMENTATION
#undef OPERATOR_IMPLEMENTATION
#undef CLEAR_IMPLEMENTATION
#undef POSITION_IMPLEMENTATION

/** @class Array
 * Vector-like array with the aim to give programmer
 * ability to access container's internal buffer directly.
 * Moreover the buffer is stored as C/C++ array.
 * Manipulating the buffer is potentially dangerous
 * and must not be combined with operations
 * on the array (ie. until the buffer is "released"
 * back to the array object, array object can not
 * perform any operations).
 * The following methods invalidate iterators:
 * resize(), reserve(), push_back(), operator=()
 *
 * @a ArrayStorage describes what type of items
 * array holds and what allocation strategy is used.
 * It is one of StaticStorage, UnmanagedStorage, DynamicStorage,
 * SharedStorage.
 *
 * We also implement a mechanism of event handlers / call-backs that
 * are invoked on various events, such as array resizing,
 * array item change, ...
 *
 * The major difference from STL-vector is that we
 * define operator[] and operator() so that we can distinguish
 * among array reads and writes.
 * Operator[] returns ConstReferenceType, operator() returns
 * ReferenceType.
 * Similarly we define at() and write_at().
 *
 * @code
 * typedef StaticStorage<int, 10>   S1; // Storage for 10 integers.
 * typedef DynamicStorage<int>      S2; // Storage for integers.
 * typedef SharedStorage<S1>        S3; // Shared storage for 10 integers.
 * typedef SharedStorage<S2>        S4; // Shared storage for integers.
 * typedef UnmanagedStorage<int>    S5; // Storage with external buffer.
 *
 * int          c_array[] = { 1, 2, 3, 4, 5 };
 * Array<S1>    a;
 * Array<S3>    b, c;
 * Array<S5>    d;
 *
 * b.push_back( 1 );
 * b.push_back( 2 );
 * c = b;   // Share array content.
 *
 * hope( b[ 0 ] == 1 );
 * c(0) = 2;
 * hope( b[ 0 ] == 2 );
 *
 * a = b;   // Copy content of array b into array a.
 *
 * d.bind( c_array, 5 );   // Bind storage to external buffer.
 * @endcode
 */
template <class ArrayStorage>
class Array : public ArrayStorage
    {
public:

    typedef typename ArrayStorage::ValueType            ValueType;
    typedef typename ArrayStorage::SizeType             SizeType;
    typedef typename ArrayStorage::DifferenceType       DifferenceType;
    typedef typename ArrayStorage::StoragePointer       StoragePointer;
    typedef typename ArrayStorage::ReferenceType        ReferenceType;
    typedef typename ArrayStorage::ConstReferenceType   ConstReferenceType;

    using ArrayStorage::size;
    using ArrayStorage::at;
    using ArrayStorage::write_at;
    using ArrayStorage::resize;

public:

    //
    // Iterators.
    //

    typedef SafeRandomAccessIterator
                <
                SequenceIteratorTraits<ArrayStorage>
                >   SafeIterator;

    typedef SafeRandomAccessIterator
                <
                SequenceConstIteratorTraits<ArrayStorage>
                >   SafeConstIterator;

    #ifdef OPTION_LIBDEBUG

    typedef SafeIterator        Iterator;
    typedef SafeConstIterator   ConstIterator;

    #else

    typedef ValueType *         Iterator;
    typedef const ValueType *   ConstIterator;

    #endif // OPTION_LIBDEBUG

public:

    //
    // Array operations.
    //

    /**
     * Default constructor.
     */
    Array()
        {
        }

    /**
     * Copy constructor.
     */
    template <class A>
    Array
        (
        const A &   array
        ) : ArrayStorage( array )
        {
        }

    /**
     * Operator=.
     */
    template <class A>
    Array & operator=
        (
        const A &   array
        )
        {
        ArrayStorage::operator=( array );
        return *this;
        }

    /**
     */
    bool empty()
        {
        return !size();
        }

    /**
     */
    ReferenceType front()
        {
        return write_at( 0 );
        }

    /**
     */
    ConstReferenceType front() const
        {
        return at( 0 );
        }

    /**
     */
    ReferenceType back()
        {
        return write_at( size() - 1 );
        }

    /**
     */
    ConstReferenceType back() const
        {
        return at( size() - 1 );
        }

    /**
     */
    void push_back
        (
        const ValueType &   value
        )
        {
        resize( size() + 1, value );
        }

    /**
     */
    void pop_back()
        {
        range_error( !empty(), "pop_back()" );
        resize( size() - 1 );
        }

    /**
     */
    Iterator begin()
        {
        #ifdef OPTION_LIBDEBUG

        return SafeIterator( this, SafeIterator::BeginTag() );

        #else

        return size() ? &front() : NULL;

        #endif
        }

    /**
     */
    ConstIterator begin() const
        {
        #ifdef OPTION_LIBDEBUG

        Array * ptr = const_cast<Array *>( this );

        return SafeConstIterator( ptr, SafeConstIterator::BeginTag() );

        #else

        return size() ? &front() : NULL;

        #endif
        }

    /**
     */
    Iterator end()
        {
        return begin() + size();
        }

    /**
     */
    ConstIterator end() const
        {
        return begin() + size();
        }

    /**
     */
    SizeType iterator_position
        (
        ConstIterator   iterator
        ) const
        {
        return iterator - begin();
        }

    /**
     */
    SizeType iterator_position
        (
        Iterator        iterator
        )
        {
        return iterator - begin();
        }

    /**
     */
    template <class S>
    bool operator==
        (
        const Array<S> &    array
        ) const
        {
        if( size() != array.size() )
            return false;

        ConstIterator                       it1( begin() );
        typename Array<S>::ConstIterator    it2( array.begin() );

        while( it1 != end() )
            {
            if( *it1 != *it2 )
                return false;

            it1++;
            it2++;
            }

        return true;
        }

    /**
     */
    template <class S>
    bool operator!=
        (
        const Array<S> &    array
        ) const
        {
        return !operator==( array );
        }

    }; // Array

/** @class NormalizedArray
 * Extends the array by defining at() & operator[] in STL-fashion.
 * Note that compiler may choose "write" versions of those methods
 * even when it is not needed.
 * @param S array storage.
 */
template <class S>
class NormalizedArray : public Array<S>
    {
public:

    typedef Array<S>                            Super;

    typedef typename Super::ValueType           ValueType;
    typedef typename Super::SizeType            SizeType;
    typedef typename Super::DifferenceType      DifferenceType;
    typedef typename Super::StoragePointer      StoragePointer;
    typedef typename Super::ReferenceType       ReferenceType;
    typedef typename Super::ConstReferenceType  ConstReferenceType;

    typedef typename Super::Iterator            Iterator;
    typedef typename Super::ConstIterator       ConstIterator;

    using Super::size;
    using Super::write_at;
    using Super::resize;

public:

    /**
     * Default constructor.
     */
    NormalizedArray()
        {
        }

    /**
     * Copy constructor.
     */
    template <class A>
    NormalizedArray
        (
        const A &   array
        ) : Super( array )
        {
        }

    /**
     * Operator=.
     */
    template <class A>
    NormalizedArray & operator=
        (
        const A &   array
        )
        {
        Super::operator=( array );
        return *this;
        }

public:

    //
    // Extensions to array interface.
    //

    ConstReferenceType at
        (
        SizeType    index
        ) const
        {
        return Super::at( index );
        }

    ReferenceType at
        (
        SizeType    index
        )
        {
        return Super::write_at( index );
        }
    
    ConstReferenceType operator[]
        (
        SizeType    index
        ) const
        {
        return Super::operator[]( index );
        }

    ReferenceType operator[]
        (
        SizeType    index
        )
        {
        return Super::operator()( index );
        }

    }; // NormalizedArray

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_ARRAY_H
