/** @file bitint_traits.h
 * Helper defintions and functions for integers with specified bitlength.
 * This header may look a bit stupid. It's because of MSVC and its (nonexistent)
 * partial specialization support.
 */

/*
 * begin        : Thu Feb 21 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BITINT_TRAITS_H
#define MASSIV_BITINT_TRAITS_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup type_traits
 * @{
 */

/** @struct Int_Traits
 * Integer stuff.
 * @param Type Either IntXX or UIntXX, where XX is 8, 16, 32 or 64.
 */
template<class Type>
struct Int_Traits
    {
    typedef void Unsigned_Type;
        /**< Unsigned integer type with the same bitlength as Type. */

    typedef void Signed_Type;
        /**< Signed integer type with the same bitlength as Type. */

    enum
        {
        bit_length = 0
            /**< Length of Type in bits. */
        };

    };

struct Int_Traits_8
    {
    typedef UInt8   Unsigned_Type;
    typedef Int8    Signed_Type;
    enum { bit_length = 8 };
    };
template<>
struct Int_Traits<Int8> : public Int_Traits_8
    {
    };
template<>
struct Int_Traits<UInt8> : public Int_Traits_8
    {
    };

struct Int_Traits_16
    {
    typedef UInt16  Unsigned_Type;
    typedef Int16   Signed_Type;
    enum { bit_length = 16 };
    };
template<>
struct Int_Traits<Int16> : public Int_Traits_16
    {
    };
template<>
struct Int_Traits<UInt16> : public Int_Traits_16
    {
    };

struct Int_Traits_32
    {
    typedef UInt32  Unsigned_Type;
    typedef Int32   Signed_Type;
    enum { bit_length = 32 };
    };
template<>
struct Int_Traits<Int32> : public Int_Traits_32
    {
    };
template<>
struct Int_Traits<UInt32> : public Int_Traits_32
    {
    };

struct Int_Traits_64
    {
    typedef UInt64  Unsigned_Type;
    typedef Int64   Signed_Type;
    enum { bit_length = 64 };
    };
template<>
struct Int_Traits<Int64> : public Int_Traits_64
    {
    };
template<>
struct Int_Traits<UInt64> : public Int_Traits_64
    {
    };


/** @struct BitInt_Traits
 * Helpers for integers with specified bitlength.
 * @param Type Either IntXX or UIntXX, where XX is 8, 16, 32 or 64.
 */
template<class Type, int bits>
struct BitInt_Traits
    {
    typedef typename Int_Traits<Type>::Unsigned_Type    Unsigned_Type;
        /**< Unsigned integer type with the same bitlength as Type. */

    typedef typename Int_Traits<Type>::Signed_Type      Signed_Type;
        /**< Signed integer type with the same bitlength as Type. */

    enum
        {
        bit_length = Int_Traits<Type>::bit_length
            /**< Length of Type in bits. */
        };

    enum
        {
        is_big_enough = ( bits <= bit_length ? 1 : 0 )
            /**< Can type hold specified number of bits? */
        };

    /**
     * Return bitmask = ( 1 << bits ) - 1.
     */
    static Unsigned_Type mask()
        {
        return static_cast<Unsigned_Type>
            ( ( static_cast<Unsigned_Type>( 1 ) << bits ) - static_cast<Unsigned_Type>( 1 ) );
        }

    }; // struct BitInt_Traits

template<>
inline UInt8 BitInt_Traits<Int8,8>::mask()
    {
    return ~ (UInt8) 0;
    }

template<>
inline UInt8 BitInt_Traits<UInt8,8>::mask()
    {
    return ~ (UInt8) 0;
    }

template<>
inline UInt16 BitInt_Traits<Int16,16>::mask()
    {
    return ~ (UInt16) 0;
    }

template<>
inline UInt16 BitInt_Traits<UInt16,16>::mask()
    {
    return ~ (UInt16) 0;
    }

template<>
inline UInt32 BitInt_Traits<Int32,32>::mask()
    {
    return ~ (UInt32) 0;
    }

template<>
inline UInt32 BitInt_Traits<UInt32,32>::mask()
    {
    return ~ (UInt32) 0;
    }

template<>
inline UInt64 BitInt_Traits<Int64,64>::mask()
    {
    return ~ (UInt64) 0;
    }

template<>
inline UInt64 BitInt_Traits<UInt64,64>::mask()
    {
    return ~ (UInt64) 0;
    }

/** @} */ // addtogroup type_traits


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_BITINT_TRAITS_H
