/** @file cache.h
 * Generic timeout cache.
 */

/*
 * begin        : Thu Feb 21 2002
 * copyright    : (C) 2001 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_CACHE_H
#define MASSIV_CACHE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HEAP_H
#include "utility/heap.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <algorithm>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * CacheTraits defines interface between client and
 * internal cache implementation. This is the only way
 * how client can react on cache-generated events.
 * To create your own traits, derive from @c CacheTraits
 * or define a new class that implements the same interface.
 */
template
    <
    class Key,
    class Data
    >
struct CacheTraits
    {
public:

    /**
     * Reasons why an item is removed from @c Cache.
     */
    enum RemoveReason
        {
        REASON_EXPLICIT_REMOVE,
            /**< The item was removed on client's request. */
        REASON_TIMEOUTED
            /**< The item timeouted. */
        }; // RemoveReason

public:

    /**
     * Called by Cache when inserting a new item.
     */
    static void inserting
        (
        Key &   key,
        Data &  content
        )
        {
        }

    /**
     * Called by Cache when removing an item.
     */
    static void removing
        (
        Key &           key,
        Data &          content,
        RemoveReason    reason
        )
        {
        }

    }; // CacheTraits


/** @class Cache
 * Generic timeout cache class.
 * Cache is associative container that associates objects of type @a Key
 * with object of type @a Data. Each cache element (key and data) has
 * associated (absolute) time of death (timeout) of type @a Time.
 * Cache elements whose timeouts are less than current time
 * (specified in @a find() and @a advance()) are automatically
 * removed from the cache.
 *
 * @todo Improve allocation of CacheElements (recycle old...).
 */
template
    <
    class Key,
    class Time,
    class Data,
    class HashKeyFcn = ext::hash<Key>,
    class EqualKey = std::equal_to<Key>,
    class TimeLess = std::less<Time>,
    class Traits = CacheTraits<Key, Data>
    >
class Cache
    {

public:

    /**
     * Destructor.
     */
    ~Cache();

    /**
     * Find cache element whose key is @a key.
     * @param key Searching key.
     * @param time Current time. (Ignored if founded element is persistent.)
     * @return Pointer to element content or NULL if element was not found.
     */
    Data * find
        (
        const Key &     key,
        const Time &    time
        );

    /**
     * Update cache element timeout.
     * Nothing is done if the element is not found in the cache or it is persistent.
     * @return True if element was found in the cache, false otherwise.
     */
    bool update_timeout
        (
        const Key &     key,
        const Time &    new_timeout,
        bool            increase_only = false
        );

    /**
     * Update cache element content.
     * Nothing is done if the element is not found in the cache.
     * @return True if element was found in the cache, false otherwise.
     */
    bool update_content
        (
        const Key &     key,
        const Data &    new_content
        );

    /**
     * Insert new cache element in the cache or change element's timeout,
     * content and persistency.
     * If element with @a key is not found, new element is inserted.
     * If element is found, its timeout, content and persistency are changed.
     * @note If element is persistent, @a new_timeout is ignored.
     */
    void update
        (
        const Key &     key,
        const Time &    new_timeout,
        const Data &    new_content,
        bool            persistent = false
        );

    /**
     * Remove cache element from the cache.
     */
    void remove
        (
        const Key &     key
        );

    /**
     * Remove all cache elements from the cache.
     */
    void clear();

    /**
     * Update all elements in the cache.
     * All elements whose timeout is lower than @a time
     * are removed from the cache.
     * @param time Current time.
     */
    void advance
        (
        const Time &    time
        );

    /**
     * Check if the cache is empty.
     */
    bool is_empty() const;

private:

    /** @class CacheElement
     * Container which holds key, heap index and content of one cache element.
     * Elements with invalid @a heap_index ( @c !heap_index.is_valid() ) are persistent.
     */
    struct CacheElement
        {

        /** Constructor. */
        CacheElement
            (
            const Key &  _hash_key,
            const Data & _content
            ) :
            hash_key( _hash_key ),
            content( _content )
            {
            }

        Key         hash_key;
            /**< Hash key of this element. */
        HeapIndex   heap_index;
            /**< Index of this element in cache heap. */
        Data        content;
            /**< Content of the element. */

        }; // struct CacheElement

    /** @struct IndexChanged
     * Monitors cache element position changes in the heap.
     */
    struct IndexChanged : public std::binary_function<CacheElement *, HeapIndex, void>
        {

        /**
         * Called whenever cache element changes location in cache heap.
         */
        void operator()
            (
            CacheElement *  e,
            HeapIndex       new_index
            ) const
            {
            // Boovie note: I was unable to move this implementation to cache_impl.h
            // without getting linkage errors. Somebody can try...
            e->heap_index = new_index;
            }

        }; // struct IndexChanged

    /** @typedef CacheHeap
     * Cache heap which stores cached elements sorted by @a Time.
     */
    typedef Heap<Time, CacheElement *, TimeLess, IndexChanged>   CacheHeap;

    /** @typedef CacheHash
     * Cache hash for fast finding elements by @a Key.
     */
    typedef ext::hash_map<Key, CacheElement *, HashKeyFcn, EqualKey>    CacheHash;

    /** @typedef CacheHashIterator
     * Iterator to a @c CacheHash map.
     */
    typedef typename CacheHash::iterator CacheHashIterator;

private:

    CacheHeap           heap;
        /**< Cache heap. */

    CacheHash           hash;
        /**< Cache hash. */

    }; // class Cache

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_CACHE_H
