/** @file cache_impl.h
 * Inline implementation of cache. See cache.h file.
 */

/*
 * begin        : Mon Apr 29 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_CACHE_IMPL_H
#define MASSIV_CACHE_IMPL_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_CACHE_H
#include "utility/cache.h"
#endif

#ifndef MASSIV_HEAP_IMPL_H
#include "utility/heap_impl.h"
#endif

#include <memory> // auto_ptr<>


namespace Massiv {
namespace Core {

// Implementation of Cache
// ~~~~~~~~~~~~~~~~~~~~~~~

#define CT  Cache<Key, Time, Data, HashKeyFcn, EqualKey, TimeLess, Traits>


template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline CT::~Cache()
    {
    clear();
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline Data * CT::find
    (
    const Key &     key,
    const Time &    time
    )
    {
    CacheHashIterator   it( hash.find( key ));
    CacheElement *      e;

    if( it == hash.end() )
        {
        return NULL;
        }

    e = ( *it ).second;

    // If the element is not persistent, check if it is still valid.

    if( e->heap_index.is_valid() && TimeLess()( heap.get_key( e->heap_index ), time ) )
        {
        hash.erase( it );
        heap.pop( e->heap_index );
        hope( !e->heap_index.is_valid() );

        Traits::removing( e->hash_key, e->content, Traits::REASON_TIMEOUTED );
        delete e;

        return NULL;
        }

    return &( e->content );
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline bool CT::update_timeout
    (
    const Key &     key,
    const Time &    new_timeout,
    bool            increase_only
    )
    {
    CacheHashIterator it( hash.find( key ));

    if( it == hash.end() )
        {
        return false;
        }

    // Change key only if the element is not persistent.

    if( ( *it ).second->heap_index.is_valid() )
        {
        heap.change_key( ( *it ).second->heap_index, new_timeout, increase_only );
        }

    return true;
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline bool CT::update_content
    (
    const Key &     key,
    const Data &    new_content
    )
    {
    CacheHashIterator it( hash.find( key ));

    if( it == hash.end() )
        {
        return false;
        }

    ( *it ).second->content = new_content;
    return true;
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline void CT::update
    (
    const Key &     key,
    const Time &    new_timeout,
    const Data &    new_content,
    bool            persistent
    )
    {
    CacheHashIterator   it( hash.find( key ) );
    CacheElement *      e;

    if( it == hash.end() )
        {
        std::auto_ptr< CacheElement > ep( new CacheElement( key, new_content ) );
        typedef typename CacheHash::value_type CacheHashValueType;

        e  = ep.get();
        hope( !e->heap_index.is_valid() );
        it = hash.insert( CacheHashValueType( key, e ) ).first;

        if( !persistent )
            {
            try
                {
                heap.push( new_timeout, e );
                }
            catch( ... )
                {
                hash.erase( it );
                throw;
                }
            }

        ep.release();

        Traits::inserting( e->hash_key, e->content );
        }
    else
        {
        e = ( *it ).second;
        e->content = new_content;

        if( !e->heap_index.is_valid() )
            {
            if( !persistent )
                {
                heap.push( new_timeout, e );
                }
            }
        else
            {
            if( persistent )
                {
                heap.pop( e->heap_index );
                hope( !e->heap_index.is_valid() );
                }
            else
                {
                heap.change_key( e->heap_index, new_timeout );
                }
            }
        }
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline void CT::remove
    (
    const Key &     key
    )
    {
    CacheHashIterator it( hash.find( key ) );

    if( it != hash.end() )
        {
        CacheElement * e = ( *it ).second;

        hash.erase( it );
        if( e->heap_index.is_valid() )
            {
            heap.pop( e->heap_index );
            hope( !e->heap_index.is_valid() );
            }

        Traits::removing( e->hash_key, e->content, Traits::REASON_EXPLICIT_REMOVE );
        delete e;
        }
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline void CT::clear()
    {
    while( !heap.is_empty() )
        {
        CacheElement * e;

        hope( heap[ 0 ]->heap_index == 0 );
        heap.pop( 0, &e );
        hope( !e->heap_index.is_valid() );
        hash.erase( e->hash_key );

        Traits::removing( e->hash_key, e->content, Traits::REASON_EXPLICIT_REMOVE );
        delete e;
        }

    // Destroy persistent elements.

        {
        CacheHashIterator it( hash.begin() );
        while( it != hash.end() )
            {
            CacheElement * e;

            e = ( *it ).second;
            hope( !e->heap_index.is_valid() );

            Traits::removing( e->hash_key, e->content, Traits::REASON_EXPLICIT_REMOVE );
            delete e;

            ++it;
            }

        hash.clear();
        }
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline void CT::advance
    (
    const Time &    time
    )
    {
    while( !heap.is_empty() && TimeLess()( heap.get_key( 0 ), time ) )
        {
        CacheElement * e;

        hope( heap[ 0 ]->heap_index == 0 );
        heap.pop( 0, &e );
        hash.erase( e->hash_key );
        hope( !e->heap_index.is_valid() );

        Traits::removing( e->hash_key, e->content, Traits::REASON_TIMEOUTED );
        delete e;
        }
    }

template <class Key, class Time, class Data, class HashKeyFcn, class EqualKey, class TimeLess, class Traits>
inline bool CT::is_empty() const
    {
    // Note heap.get_size() is not same as hash.size().

    return hash.empty();
    }

#undef CT

} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_CACHE_IMPL_H
