/** @file debug.h
 * Debugging macros - #panic() and #hope().
 */

/*
 * begin        : Sat Oct 12 2002
 * copyright    : (C) 2002-2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_DEBUG_H
#define MASSIV_DEBUG_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <cassert>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * Internal function called by @c panic macro.
 * Never returns.
 */
NORETURN( void ) MASSIV_EXPORTED debug_panic
    (
    const char * const  file,
    const unsigned      line,
    const char * const  function,
    const char * const  message
    ) throw();


/**
 * Internal function called by improved @c hope macro.
 * Never returns.
 */
NORETURN( void ) MASSIV_EXPORTED debug_assert
    (
    const char * const  file,
    const unsigned      line,
    const char * const  function,
    const char * const  message
    ) throw();


/**
 * Report error message and exit.
 */
#define panic( m ) Massiv::Core::debug_panic( __FILE__, __LINE__, CURRENT_FUNCTION_NAME, m )


/** @def hope
 * Improved assert-like macro.
 * It shows more information than assert and gets logged.
 */

#undef hope

#ifdef OPTION_DEBUG

#if defined( COMPILER_GCC ) && ( __GNUC__ >= 3 ) && defined( SYSTEM_LINUX )

#define hope( e ) \
    ( \
    __ASSERT_VOID_CAST( __builtin_expect( !!( e ), 1 ) ? \
        0 : \
        ( Massiv::Core::debug_assert( __FILE__, __LINE__, CURRENT_FUNCTION_NAME, __STRING( e ) ), 0 ) ) \
    )

#else // compiler

#define hope( e ) \
    ( void )( ( e ) || ( Massiv::Core::debug_assert( __FILE__, __LINE__, CURRENT_FUNCTION_NAME, #e ), 0 ) )

#endif // compiler

#else // OPTION_DEBUG

#define hope( e ) ( static_cast< void >( 0 ) )

#endif // OPTION_DEBUG


/**
 * Shortcut for <tt>hope( e && msg )</tt>.
 */
#define hope_msg( e, msg ) hope( ( e ) && ( msg ) )

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_DEBUG_H
