/** @file debug_storage.h
 * Debug storage.
 */

/*
 * begin        : Tue Apr 09 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_DEBUG_STORAGE_H
#define MASSIV_DEBUG_STORAGE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <iterator>

#ifdef OPTION_DEBUG

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#endif

#ifdef OPTION_LIBDEBUG

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#endif


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class DebugStorage
 * Debug storage is a storage / container that registers
 * iterators pointing to it and checks their proper use.
 * Prevents client from dereferencing invalidated iterator.
 */
class DebugStorage
    {
    //
    // Debug iterator.
    //

#ifdef OPTION_LIBDEBUG

    public:

        /**
         * Base class for safe iterators pointing to the storage.
         */
        struct DebugIterator
            {
            friend class DebugStorage;

        protected:

            struct Hash
                {
                size_t operator()
                    (
                    const DebugIterator * const it
                    ) const
                    {
                    return reinterpret_cast<size_t>( it );
                    }
                };

            struct EqualKey
                {
                bool operator()
                    (
                    const DebugIterator * const it1,
                    const DebugIterator * const it2
                    ) const
                    {
                    return it1 == it2;
                    }
                };

        protected:

            /** Constructor. */
            DebugIterator() : valid( true )
                {
                }

        protected:

            /** Invalidates the iterator. */
            void invalidate() const
                {
                valid = false;
                }

            /**
             * Registers iterator to the storage.
             * Iterator may be registered to multiple storages.
             */
            void register_iterator
                (
                const DebugStorage * storage
                ) const
                {
                if( storage )
                    {
                    storage->register_iterator( *this );
                    }
                }

            /** Unregisters iterator. */
            void unregister_iterator
                (
                const DebugStorage * storage
                ) const
                {
                if( storage )
                    {
                    storage->unregister_iterator( *this );
                    }
                }
        
        public:

            /** Checks whether the iterator is still valid. */
            void check_validity() const
                {
                if( !valid )
                    {
                    DebugStorage::invalidated_iterator_error();
                    }
                }

        private:

            mutable bool    valid;
                /**< Is iterator still valid? */

            }; // DebugIterator

        friend struct DebugIterator;

    protected:

        /** Constructor. */
        DebugStorage()
            {
            }

        /** Destructor. */
        virtual ~DebugStorage()
            {
            if( iterators.size() )
                {
                DebugStorage::storage_delete_error();
                }
            }

    public:

        /** Registers iterator. */
        virtual void register_iterator
            (
            const DebugIterator & iterator
            ) const
            {
            bool inserted;

            inserted = iterators.insert( &iterator ).second;
            hope( inserted );
            }

        /** Unregisters iterator. */
        virtual void unregister_iterator
            (
            const DebugIterator & iterator
            ) const
            {
            IteratorList::iterator it( iterators.find( &iterator ) );

            hope( it != iterators.end() );
            iterators.erase( it );
            }

        /** Invalidates all iterators pointing to this storage. */
        virtual void invalidate_iterators() const
            {
            IteratorList::iterator it( iterators.begin() );

            while( it != iterators.end() )
                {
                (*it)->invalidate();
                it++;
                }
            }

        /**
         * Transfers registration of these iterators that are
         * also registered to @a s1 to storage @a s2.
         */
        virtual void transfer_iterators
            (
            const DebugStorage * s1,
            const DebugStorage * s2
            )
            {
            IteratorList::iterator it( iterators.begin() );

            while( it != iterators.end() )
                {
                const DebugIterator * iterator = *it;

                iterator->invalidate();

                if( s1 )
                    s1->unregister_iterator( *iterator );

                if( s2 )
                    s2->register_iterator( *iterator );

                it++;
                }
            }

    protected:

        typedef ext::hash_set
                    <
                    const DebugIterator *,
                    DebugIterator::Hash,
                    DebugIterator::EqualKey
                    >    IteratorList;

        mutable IteratorList    iterators;
            /**< List of iterators pointing to this storage. */

#else // !OPTION_LIBDEBUG
    
    public:

        struct DebugIterator
            {
            void invalidate() const                                 { }
            void check_validity() const                             { }
            void register_iterator( const DebugStorage * ) const    { }
            void unregister_iterator( const DebugStorage * ) const  { }
            };

        void register_iterator( const DebugIterator & ) const       { }
        void unregister_iterator( const DebugIterator & ) const     { }
        void invalidate_iterators() const                           { }
        void transfer_iterators( const DebugStorage *, const DebugStorage * ) { }

#endif // OPTION_LIBDEBUG

    //
    // Error handling.
    //

#ifdef OPTION_DEBUG

    public:

        /** Signals range error. */
        static void range_error
            (
            int             condition,
            const char *    message
            )
            {
            if( !condition )
                {
                throw ExOutOfRange( message );
                }
            }

        /** Signals iterator incompatibility error. */
        static void incompatible_iterator_error
            (
            int condition
            )
            {
            if( !condition )
                {
                throw ExIteratorError( "Incompatible iterator used." );
                }
            }

        /** Signals iterator error. */
        static void invalidated_iterator_error()
            {
            throw ExIteratorError( "Invalidated iterator dereferenced." );
            }

        /** Signals iterator error. */
        static void storage_delete_error()
            {
            throw ExIteratorError
                ( "Deleting storage that is still referenced by iterator(s)." );
            }

        /** Signals iterator error. */
        static void iterator_position_error()
            {
            throw ExIteratorError( "Iterator points outside the storage." );
            }

        /** Signals buffer error. */
        static void buffer_error()
            {
            throw ExCommonError( "Buffer error." );
            }
                                                                
#else // !OPTION_DEBUG

    public:

        static void range_error( int , const char * )   { }
        static void incompatible_iterator_error( int )  { }
        static void invalidated_iterator_error()        { }
        static void storage_delete_error()              { }
        static void iterator_position_error()           { }
        static void buffer_error()                      { }

#endif // OPTION_DEBUG

    }; // DebugStorage

/**
 * SafeRandomAccessIterator @c Traits implementation.
 * @c Sequence is supposed to be a container class that
 * indexes its elements in an ordinary way plus
 * implements @c at() and @c write_at() for accessing
 * elements at requested positions.
 * @c Array is an example of a sequence.
 * @c SafeRandomAccessIterator< SequenceConstIteratorTraits<Sequence> >
 * defines safe const iterator that can be used for reading elements
 * stored in @c Sequence.
 */
template <class Sequence>
class SequenceConstIteratorTraits
    {
public:

    typedef Sequence                                Storage;
    typedef Sequence *                              StoragePointer;
    typedef typename Sequence::SizeType             SizeType;
    typedef typename Sequence::DifferenceType       DifferenceType;
    typedef typename Sequence::ConstReferenceType   ReferenceType;
    typedef typename Sequence::ValueType            ValueType;
    typedef SequenceConstIteratorTraits<Sequence>   Traits;
    typedef SequenceConstIteratorTraits<Sequence>   ConstTraits;

public:

    /** Sets iterator context to point to @c begin(). */
    void begin
        (
        StoragePointer
        )
        {
        position = 0;
        }

    /** Sets iterator context to point to @c end(). */
    void end
        (
        StoragePointer storage
        )
        {
        position = storage->size();
        }

    /** Moves iterator. */
    void move
        (
        StoragePointer  storage,
        DifferenceType  difference
        )
        {
        position += difference;

        // @c size() is valid position (@c end()).

        if( position < 0 || position > storage->size() )
            {
            DebugStorage::iterator_position_error();
            }
        }

    /** Dereferences iterator. */
    ReferenceType dereference
        (
        StoragePointer  storage
        ) const
        {
        check_dereference( storage );
        return storage->at( position );
        }

    /** Computes iterator difference. */
    DifferenceType difference
        (
        const Traits & context
        ) const
        {
        return position - context.position;
        }

protected:

    /** Checks if iterator can be dereferenced. */
    void check_dereference
        (
        StoragePointer  storage
        ) const
        {
        // Position @c size() can not be dereferenced.

        if( position < 0 || position >= storage->size() )
            {
            DebugStorage::iterator_position_error();
            }
        }

protected:

    mutable SizeType    position;
        /**< Current iterator position.
             Position @c size() corresponds to @c end(). */

    }; // SequenceConstIteratorTraits

/**
 * @c SafeRandomAccessIterator< SequenceIteratorTraits<Sequence> >
 * defines safe iterator that can be used for modifying elements
 * stored in @c Sequence.
 */
template <class Sequence>
class SequenceIteratorTraits : public SequenceConstIteratorTraits<Sequence>
    {
public:

    typedef typename Sequence::ReferenceType        ReferenceType;
    typedef SequenceIteratorTraits<Sequence>        Traits;
    typedef SequenceConstIteratorTraits<Sequence>   ConstTraits;

    typedef SequenceConstIteratorTraits<Sequence>   Super;
    typedef typename Super::Storage                 Storage;
    typedef typename Super::StoragePointer          StoragePointer;
    typedef typename Super::SizeType                SizeType;
    typedef typename Super::DifferenceType          DifferenceType;
    typedef typename Super::ValueType               ValueType;

protected:
    
    using Super::position;

public:

    /** Dereferences iterator. */
    ReferenceType dereference
        (
        StoragePointer  storage
        )
        {
        check_dereference( storage );
        return storage->write_at( position );
        }

    }; // SequenceIteratorTraits

/**
 * Safe iterator pointing to a DebugStorage.
 * @c Traits parameter defines semantics of the iterator.
 * For required interface that @c Traits must implement consult
 * @c SequenceIteratorTraits.
 * @c Traits for @c Iterator should be derived from @c Traits (ConstTraits)
 * for @c ConstIterator.
 * This ensures that conversion from @c Iterator to @c ConstIterator
 * is possible.
 */
template <class Traits>
class SafeRandomAccessIterator :
    public DebugStorage::DebugIterator,
    public std::iterator<std::random_access_iterator_tag, typename Traits::ValueType>
    {
public:

    typedef DebugStorage::DebugIterator             Super;

    typedef typename Traits::ConstTraits            ConstTraits;
    typedef SafeRandomAccessIterator<Traits>        Iterator;
    typedef SafeRandomAccessIterator<ConstTraits>   ConstIterator;

    typedef typename Traits::Storage                Storage;
    typedef typename Traits::StoragePointer         StoragePointer;
    typedef typename Traits::SizeType               SizeType;
    typedef typename Traits::DifferenceType         DifferenceType;
    typedef typename Traits::ReferenceType          ReferenceType;

    class BeginTag
        {
        };

    class EndTag
        {
        };

public:

    //
    // Iterator construction.
    //

    /**
     * Internal constructor.
     * Iterator will point to @c begin().
     */
    SafeRandomAccessIterator
        (
        StoragePointer storage,
        BeginTag
        ) : storage( storage )
        {
        hope( storage );
        register_iterator( storage );

        context.begin( storage );
        }

    /**
     * Internal constructor.
     * Iterator will point to @c end().
     */
    SafeRandomAccessIterator
        (
        StoragePointer storage,
        EndTag
        ) : storage( storage )
        {
        hope( storage );
        register_iterator( storage );

        context.end( storage );
        }

    /** Default constructor. */
    SafeRandomAccessIterator()
        : storage( NULL )
        {
        register_iterator( storage );
        }

    /** Copy constructor. */
    SafeRandomAccessIterator
        (
        const Iterator & iterator
        ) : storage( iterator.storage ),
            context( iterator.context )
        {
        register_iterator( storage );
        }

    /** Internal constructor. */
    SafeRandomAccessIterator
        (
        StoragePointer      storage,
        const ConstTraits & context
        ) : storage( storage ),
            context( context )
        {
        register_iterator( storage );
        }

    /** Destructor. */
    ~SafeRandomAccessIterator()
        {
        unregister_iterator( storage );
        }

    /** Assigns iterator. */
    Iterator & operator=
        (
        const Iterator & iterator
        )
        {
        unregister_iterator( storage );

        iterator.check_validity();
        storage = iterator.storage;
        context = iterator.context;
        Super::operator=( iterator );

        register_iterator( storage );

        return *this;
        }

public:

    //
    // Dereferencing.
    //

    /** Returns reference to item iterator points to. */
    ReferenceType operator*()
        {
        check_validity();
        return context.dereference( storage );
        }

public:

    //
    // Movement.
    //

    /** Moves iterator, preincrement. */
    Iterator & operator++()
        {
        context.move( storage, 1 );
        return *this;
        }

    /** Moves iterator, postincrement. */
    Iterator operator++
        (
        int
        )
        {
        Iterator old = *this;
        operator++();
        return old;
        }

    /** Moves iterator, predecrement. */
    Iterator & operator--()
        {
        context.move( storage, -1 );
        return *this;
        }

    /** Moves iterator, postdecrement. */
    Iterator operator--
        (
        int
        )
        {
        Iterator old = *this;
        operator--();
        return old;
        }

    /** Moves iterator. */
    Iterator & operator+=
        (
        DifferenceType count
        )
        {
        context.move( storage, count );
        return *this;
        }

    /** Moves iterator. */
    Iterator & operator-=
        (
        DifferenceType count
        )
        {
        return operator+=( -count );
        }

    /** Moves iterator. */
    Iterator operator+
        (
        DifferenceType n
        ) const
        {
        Iterator it2 = *this;
        it2 += n;
        return it2;
        }

    /** Moves iterator. */
    Iterator operator-
        (
        DifferenceType n
        ) const
        {
        Iterator it2 = *this;
        it2 -= n;
        return it2;
        }

    /** Computes difference between iterators. */
    DifferenceType operator-
        (
        const ConstIterator & iterator
        ) const
        {
        check_compatibility( iterator );
        return context.difference( iterator.get_context() );
        }

public:

    //
    // Compares.
    //

    /** Compares iterators. */
    bool operator==
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator == 0;
        }

    /** Compares iterators. */
    bool operator!=
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator != 0;
        }

    /** Compares iterators. */
    bool operator<
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator < 0;
        }

    /** Compares iterators. */
    bool operator<=
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator <= 0;
        }

    /** Compares iterators. */
    bool operator>
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator > 0;
        }

    /** Compares iterators. */
    bool operator>=
        (
        const Iterator & iterator
        ) const
        {
        return *this - iterator >= 0;
        }

public:

    //
    // Other.
    //

    /** Converts @c Iterator to @c ConstIterator. */
    operator ConstIterator() const
        {
        return ConstIterator( storage, static_cast<const ConstTraits &>( context ) );
        }

    /** Returns pointer to the storage. */
    StoragePointer get_storage() const
        {
        return storage;
        }

    /** Returns iterator context. */
    const Traits & get_context() const
        {
        return context;
        }

private:

    //
    // Debugging.
    //

    /** Checks if the two iterators point to the same storage. */
    void check_compatibility
        (
        const ConstIterator & iterator
        ) const
        {
        DebugStorage::incompatible_iterator_error( storage == iterator.get_storage() );
        }

private:

    StoragePointer      storage;
        /**< Pointer to storage. */
    Traits              context;
        /**< Iterator context. */

    }; // SafeRandomAccessIterator

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_DEBUG_STORAGE_H
