/** @file hash.h
 * Standard C++ library hash_map/set extension wrapper.
 * The wrapper creates 'ext' namespace to contain SGI-style hash containers.
 */

/*
 * begin        : 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_HASH_H
#define MASSIV_HASH_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_HASH_FUNCTION_H
#include "utility/hash_function.h"
#endif

#include <cstdlib>


#ifdef OPTION_STLPORT

    // STLPort hash_map can be included as standard header.

    #include <hash_map>
    #include <hash_set>

    // Merge std and ext.
    
    #ifndef ext
    #define ext std
    #endif

#else // OPTION_STLPORT

    // Standard system/compiler-defined library is used.

    #if defined( COMPILER_GCC )

        // GCC with libstdc++.
        // Only 3.0 (libstdc++3 in Debian ) and 3.2 or newer
        // (libstdc++5 in Debian) versions are supported.

        #if ( __GNUC__ < 3 )

            #error "STLPort must be used with GCC older than 3.0."

        #elif ( __GNUC__ == 3 ) && ( __GNUC_MINOR__ == 0 )

            // GCC 3.0. "hash_map" is in "ext" subdirectory.

            #include <ext/hash_map>
            #include <ext/hash_set>

            // Hash for std::string must be implemented.

            #include <string>

            namespace std {

                template<> struct hash<std::string>
                    {
                    inline size_t operator()( const std::string & string ) const
                        {
                        return ext::hash_string( string.c_str() );
                        }
                    };

            } // namespace std

            // Extensions are in std namespace.

            namespace ext {

                using std::hash_map;
                using std::hash_set;
                using std::hash_multimap;
                using std::hash_multiset;
                using std::hash;

            } // namespace ext

        #elif ( __GNUC__ == 3 ) && ( __GNUC_MINOR__ == 1 )

            #error "GCC 3.1 is not supported."

        #elif ( ( __GNUC__ == 3 ) && ( __GNUC_MINOR__ >= 2 ) ) || ( __GNUC__ == 4 )

            // GCC 3.2 or newer. Tested with 3.2, 3.3 and 4.3, more recent 
            // versions may be different again.

            // Include "hash_map" and "hash_set" from the "ext" subdirectory.

            #if ( __GNUC__ >= 4 )
                // GCC 4.3 considers these headers deprecated and suggests that
                // tr1 containers unordered_set and unordered_map should be
                // used instead.
                // Unfortunatelly, these containers have different rules for
                // iterator invalidation and hence can not be used as a simple 
                // replacement for SGI's hash_set and hash_map (nor std::set
                // and std::map). For now, we temporarily disable deprecation 
                // warnings so that compilation will go smoothly, but in the long
                // run, we should consider implementing our own hash containers 
                // with SGI semantics. That should not be that hard anyway.

                #ifdef __DEPRECATED
                    #undef __DEPRECATED
                    #define RESTORE_DEPRECATION
                #endif
            #endif

            #include <ext/hash_map>
            #include <ext/hash_set>

            #ifdef RESTORE_DEPRECATION
                #define __DEPRECATED
                #undef RESTORE_DEPRECATION
            #endif

            // Hash for std::string must be implemented.

            #include <string>

            namespace __gnu_cxx {

                template<> struct hash<std::string>
                    {
                    inline size_t operator()( const std::string & string ) const
                        {
                        return ext::hash_string( string.c_str() );
                        }
                    };

            } // namespace __gnu_cxx

            // Extensions are (correctly) not in "std" namespace.

            namespace ext {

                using __gnu_cxx::hashtable;
                using __gnu_cxx::hash;
                using __gnu_cxx::hash_map;
                using __gnu_cxx::hash_multimap;
                using __gnu_cxx::hash_set;
                using __gnu_cxx::hash_multiset;

            } // namespace ext

        #else

            #error "Unsupported gcc version."

        #endif // gcc version

    #elif defined( COMPILER_ICC )

        // ICC (like MSVC7) uses Dinkumware STD C++ library.
        // Hash maps and friend can be hacked the same way like it was
        // done for MSVC. Not done though.

        #error "STLPort must be used with ICC."

    #elif defined( COMPILER_MSVC )

        #if ( _MSC_VER < 1300 )

            #error "STLPort must be used with MSVC older than 7.0."

        #else

            #include <hash_map>
            #include <hash_set>
            #include <string>

            #if ( _MSC_VER < 1310 )
            // MSVC 7.0 defines hash containers in "std" namespace.
            #define hash_namespace  std
            #else
            // MSVC 7.1, 8 and 9 define hash containers in "stdext" namespace.
            #define hash_namespace  stdext
            #endif

            // SGI-style wrappers for Dinkumware-style hash containers.

            namespace ext {

                template <class Key>
                struct hash
                    {
                    inline size_t operator()( const Key & key ) const
                        {
                        return static_cast<size_t>( key );
                        }
                    };

                template <>
                struct hash<std::string>
                    {
                    inline size_t operator()( const std::string & key ) const
                        {
                        return ext::hash_string( key.c_str() );
                        }
                    };

                template <class Key, class HashKey, class EqualKey>
                struct HashTraits
                    {
                public:

                    enum
                        {
                        bucket_size = 32,
                        min_buckets = 128
                        };

                    inline size_t operator()( const Key & key ) const
                        {
                        return hash_key( key );
                        }

                    inline bool operator()( const Key & key1, const Key & key2 ) const
                        {
                        return !equal_key( key1, key2 );
                        }

                private:

                    HashKey     hash_key;
                    EqualKey    equal_key;
                    };

                template
                    <
                    class   Key,
                    class   HashKey = hash<Key>,
                    class   EqualKey = std::equal_to<Key>
                    >
                class hash_set : public hash_namespace::hash_set< Key, HashTraits<Key, HashKey, EqualKey> >
                    {
                    };

                template
                    <
                    class   Key,
                    class   HashKey = hash<Key>,
                    class   EqualKey = std::equal_to<Key>
                    >
                class hash_multiset : public hash_namespace::hash_multiset< Key, HashTraits<Key, HashKey, EqualKey> >
                    {
                    };

                template
                    <
                    class   Key,
                    class   Data,
                    class   HashKey = hash<Key>,
                    class   EqualKey = std::equal_to<Key>
                    >
                class hash_map : public hash_namespace::hash_map< Key, Data, HashTraits<Key, HashKey, EqualKey> >
                    {
                    };

                template
                    <
                    class   Key,
                    class   Data,
                    class   HashKey = hash<Key>,
                    class   EqualKey = std::equal_to<Key>
                    >
                class hash_multimap : public hash_namespace::hash_multimap< Key, Data, HashTraits<Key, HashKey, EqualKey> >
                    {
                    };

            } // namespace ext

            #undef hash_namespace

        #endif // MSVC version

    #else

        #error "Unknown standard library is used. Don't know how to include hash extensions."

    #endif // compiler

#endif // OPTION_STLPORT


#endif // MASSIV_HASH_H
