/** @file heap.h
 * Generic heap.
 */

/*
 * begin        : Fri Feb 22 2002
 * copyright    : (C) 2001 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_HEAP_H
#define MASSIV_HEAP_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HEAP_INDEX_H
#include "utility/heap_index.h"
#endif

#ifndef MASSIV_MISC_CLASS_H
#include "utility/misc_class.h"
#endif

#include <vector>
#include <algorithm>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class Heap
 * Generic heap. Stores elements of type @a Data which are ordered by associated keys of type
 * @a Key. Two keys are compared using function object of type @a KeyCompareFnc.
 * @a KeyCompareFnc is a binary predicate which compares two objects, returning true
 * if the first precedes the second. This predicate must satisfy the standard mathematical
 * definition of a strict weak ordering.
 * Stored element can keep track of its position in the heap by specifying binary function
 * object @a HeapIndexChangedFnc which is called everytime the element's position in the
 * heap is changed. Its first parameter is stored element's data and is of type @a Data,
 * second parameter is new element's index and must be of type @c HeapIndex.
 * @note @c HeapIndexChangedFnc is called on element which is being removed from the heap
 * using @c pop() with index value @c HeapIndex::INVALID.
 * @note Elements in the heap are stored in ascending order ie. element with minimal key
 * has zero index.
 */
template <
    class Key,
    class Data,
    class KeyCompareFnc = std::less<Key>,
    class HeapIndexChangedFnc = VoidBinaryFunction<Data, HeapIndex> >
class Heap
    {

public:

    /**
     * Destructor.
     */
    ~Heap();

public:

    /**
     * Add new element to the heap.
     * @param key Key associated to the element.
     * @param x Element's data.
     */
    void push
        (
        const Key &     key,
        const Data &    x
        );

    /**
     * Remove element from the heap.
     * @param index Index of element to remove. Specify zero to remove element with minimal key.
     * @param removed_data Pointer to @a Data where to store removed element. Can be @c NULL.
     */
    void pop
        (
        HeapIndex   index           = 0,
        Data *      removed_data    = NULL
        );

    /**
     * Change key associated with element.
     * @param index Index of element.
     * @param new_key New key value.
     * @param increase_only Increase key only.
     */
    void change_key
        (
        HeapIndex           index,
        const Key &         new_key,
        bool                increase_only = false
        );

    /**
     * Get key associted with element.
     * @param index Index of element.
     */
    Key get_key
        (
        HeapIndex           index
        ) const;

    /**
     * Check if the heap is empty.
     */
    bool is_empty
        (
        void
        ) const;

    /**
     * Get heap size (number of elements stored in the heap).
     */
    int get_size
        (
        void
        ) const;

    /**
     * Return const reference to element at position @a index.
     */
    const Data & operator[]
        (
        HeapIndex   index
        ) const;

    /**
     * Return reference with write access to element at position @a index.
     */
    Data & operator[]
        (
        HeapIndex   index
        );

    /**
     * Clears heap content.
     */
    void clear();

private:

    /**
     * Heap manipulation function.
     * Move element with @a index up the heap.
     */
    void up
        (
        HeapIndex   index
        );

    /**
     * Heap manipulation function.
     * Move element with @a index down the heap.
     */
    void down
        (
        HeapIndex   index
        );

private:

    /** @typedef HeapElement
     * Type of data stored in the heap. Contains associated key and element's data.
     */
    typedef std::pair<Key, Data>        HeapElement;

    /** @typedef HeapArray
     * Heap array.
     */
    typedef std::vector<HeapElement>    HeapArray;

    HeapArray           heap_array;
        /**< Heap array. */

    }; // class Heap

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_HEAP_H
