/** @file heap_impl.h
 * Inline implementation of heap. See heap.h file.
 */

/*
 * begin        : Mon Apr 29 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_HEAP_IMPL_H
#define MASSIV_HEAP_IMPL_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_HEAP_H
#include "utility/heap.h"
#endif



namespace Massiv {
namespace Core {

// Inline implementation of Heap
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


#define HT  Heap<Key, Data, KeyCompareFnc, HeapIndexChangedFnc>


template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline HT::~Heap()
    {
    clear();
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::push
    (
    const Key &     key,
    const Data &    x
    )
    {
    HeapIndex       i( heap_array.size() );

    heap_array.push_back( HeapElement( key, x ) );
    HeapIndexChangedFnc()( (*(heap_array.end() - 1)).second, i );
    up( i );
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::pop
    (
    HeapIndex   index,
    Data *      d
    )
    {
    unsigned int i = index.get_value();

    hope( i >= 0 && i < heap_array.size() );

    typename HeapArray::iterator    ii( heap_array.begin() + i );
    typename HeapArray::iterator    last( heap_array.end() - 1 );
    bool                            go_down;

    if( /* heap_array.size() <= 1 || */ i == ( heap_array.size() - 1 ) )
        {
        HeapIndexChangedFnc()( (*last).second, HeapIndex::INVALID );
        if( d )
            {
            (*d) = (*last).second;
            }
        heap_array.pop_back();
        return;
        }

    go_down = KeyCompareFnc()( (*ii).first, (*last).first );

    HeapIndexChangedFnc()( (*ii).second, HeapIndex::INVALID );
    if( d )
        {
        (*d) = (*ii).second;
        }
    (*ii) = (*last);
    HeapIndexChangedFnc()( (*ii).second, i );
    heap_array.pop_back();

    if( go_down )
        {
        down( i );
        }
    else
        {
        up( i );
        }
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::change_key
    (
    HeapIndex       index,
    const Key &     k,
    bool            increase_only
    )
    {
    unsigned int i = index.get_value();

    hope( i >= 0 && i < heap_array.size() );

    typename HeapArray::iterator    ii( heap_array.begin() + i );
    Key                             old_key = (*ii).first;

    if( KeyCompareFnc()( old_key, k ) )
        {
        (*ii).first = k;
        down( i );
        }
    else
        {
        if( !increase_only )
            {
            (*ii).first = k;
            up( i );
            }
        }
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::up
    (
    HeapIndex   index
    )
    {
    unsigned int i = index.get_value();

    hope( i >= 0 && i < heap_array.size() );

    HeapElement                     t;  // Travels outside heap.
    typename HeapArray::iterator    ii; // Points to hole.

    if( i == 0 )
        {
        return;
        }

    ii = heap_array.begin() + i;
    t = (*ii);

    while( 1 )
        {
        if( i == 0 )
            {
            break;
            }

        unsigned int                    j  = ( i - 1 ) / 2;  // Parent index.
        typename HeapArray::iterator    ij = heap_array.begin() + j;

        if( !KeyCompareFnc()( t.first, (*ij).first ))
            {
            break;
            }
        (*ii) = (*ij);  // Swap hole and parent.
        HeapIndexChangedFnc()( (*ii).second, i );   // Inform parent about his index change.
        i = j;
        ii = ij;
        }

    (*ii) = t;
    HeapIndexChangedFnc()( (*ii).second, i );
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::down
    (
    HeapIndex   index
    )
    {
    unsigned int i = index.get_value();

    hope( i >= 0 && i < heap_array.size() );

    HeapElement                     t;  // Travels outside heap.
    typename HeapArray::iterator    ii; // Points to hole.

    ii = heap_array.begin() + i;
    t = (*ii);

    while( 1 )
        {
        unsigned int                    j  = 2 * i + 2;  // Second child;
        typename HeapArray::iterator    ij;

        if( j >= heap_array.size() )
            {
            --j;
            if( j < heap_array.size() )
                {
                ij = heap_array.begin() + j;
                if( KeyCompareFnc()( (*ij).first, t.first ) )
                    {
                    (*ii) = (*ij);
                    HeapIndexChangedFnc()( (*ii).second, i );
                    i = j;
                    ii = ij;
                    }
                }
            break;
            }

        ij = heap_array.begin() + j;

        // Select minimum.
        if( KeyCompareFnc()( (*(ij - 1)).first, (*ij).first ) )
            {
            --j;
            --ij;
            }

        if( !KeyCompareFnc()( (*ij).first, t.first ) )
            {
            break;
            }

        (*ii) = (*ij);
        HeapIndexChangedFnc()( (*ii).second, i );
        i = j;
        ii = ij;
        }

    (*ii) = t;
    HeapIndexChangedFnc()( (*ii).second, i );
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline Key HT::get_key
    (
    HeapIndex   index
    ) const
    {
    unsigned int i = index.get_value();
    hope( i >= 0 && i < heap_array.size() );
    return heap_array[ i ].first;
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline bool HT::is_empty
    (
    void
    ) const
    {
    return heap_array.empty();
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline int HT::get_size
    (
    void
    ) const
    {
    return (int) heap_array.size();
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline const Data & HT::operator[]
    (
    HeapIndex   index
    ) const
    {
    unsigned int i = index.get_value();
    hope( i >= 0 && i < heap_array.size() );
    return heap_array[ i ].second;
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline Data & HT::operator[]
    (
    HeapIndex index
    )
    {
    unsigned int i = index.get_value();
    hope( i >= 0 && i < heap_array.size() );
    return heap_array[ i ].second;
    }

template <
    class Key,
    class Data,
    class KeyCompareFnc,
    class HeapIndexChangedFnc>
inline void HT::clear()
    {
    while( !is_empty() )
        {
        pop( 0 );
        }
    }


#undef HT



}; // namespace Massiv::Core
}; // namespace Massiv


#endif // _HEAP_IMPL_H_
