/** @file heap_index.h
 * Heap index.
 */

/*
 * begin        : Fri Feb 22 2002
 * copyright    : (C) 2001 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_HEAP_INDEX_H
#define MASSIV_HEAP_INDEX_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif



namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class HeapIndex
 * Index into heap.
 * @note @c HeapIndex is not intended to behave like integer. Its only purpose
 * is to hold some "index" which is passed to and from @c Heap functions.
 */
class HeapIndex
    {

public:

    enum
        {
        INVALID = static_cast<unsigned int>( -1 )
        };

public:

    /**
     * Constructor.
     */
    HeapIndex()
        {
        index = static_cast<unsigned int>( INVALID );
        }

    /**
     * Copy constructor.
     */
    HeapIndex
        (
        const HeapIndex & i
        )
        {
        index = i.index;
        }

    /**
     * Constructor.
     */
    HeapIndex
        (
        unsigned int i
        )
        {
        index = i;
        }

    /**
     * Operator=.
     */
    HeapIndex & operator=
        (
        const HeapIndex & i
        )
        {
        index = i.index;
        return *this;
        }

    /**
     * Get index value.
     */
    unsigned int get_value() const
        {
        return index;
        }

    /**
     * Make index invalid.
     */
    void set_invalid()
        {
        index = static_cast<unsigned int>( INVALID );
        }

    /**
     * Check index validity.
     * @note Doesn't check if the index points to valid heap element, but only tests if the index
     * contains value which is acceptable as index into some heap.
     */
    bool is_valid() const
        {
        return index != static_cast<unsigned int>( INVALID );
        }

    /**
     * Check if the index is equal to another index.
     */
    bool is_equal_to
        (
        const HeapIndex & i
        ) const
        {
        return index == i.index;
        }

private:

    unsigned int    index;
        /**< Held index. */

    }; // HeapIndex


inline bool operator==
    (
    const HeapIndex & i1,
    const HeapIndex & i2
    )
    {
    return i1.is_equal_to( i2 );
    }


inline bool operator!=
    (
    const HeapIndex & i1,
    const HeapIndex & i2
    )
    {
    return !i1.is_equal_to( i2 );
    }

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_HEAP_INDEX_H
