/** @file lru_cache.cpp
 * LRUCache.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/lru_cache.h"
#include "status/exception.h"
#include "registry/registry.h"
#include "global/global.h"
#include "utility/debug.h"
#include <cstdlib>
#include <sstream>


namespace Massiv {
namespace Core {


namespace {


#ifdef OPTION_DEBUG

/**
 * Monitor that checks if LRUCache instance is not reentered form a callback.
 * See INSIDE_CALLBACK() macro.
 */
class CallbackMonitor
    {
public:
        
    /** Locks cache. */
    CallbackMonitor
        (
        LRUCache & cache
        ) : cache( cache )
        {
        hope( !cache.inside_callback );
        cache.inside_callback = true;
        }
    
    /** Unlocks cache. */
    ~CallbackMonitor()
        {
        cache.inside_callback = false;
        }

    LRUCache & cache;
        /**< Cache. */
    };

/**
 * "Call" this prior invoking a guarded callback.
 * This ensures that it will be registered that the callback is active
 * and unregistered when the callback finishes.
 */
#define INSIDE_CALLBACK() CallbackMonitor __monitor( *this );

/**
 * Call this at the beginning of any public LRUCache method so that it
 * would be tested that the cache is not reentered from a callback.
 */
#define PREVENT_REENTRANCY() \
    if( inside_callback ) \
        { \
        panic( "LRUCache reentered from a callback." ); \
        }

#else

#define INSIDE_CALLBACK()
#define PREVENT_REENTRANCY()

#endif


}
    

// Double-linked cyclic list manipulators
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#define init( list, item ) \
    ( item )->list##_next = item; \
    ( item )->list##_prev = item;

#define insert_head( list, item, head ) \
    ( item )->list##_next = ( head )->list##_next; \
    ( item )->list##_next->list##_prev = item; \
    ( head )->list##_next = item; \
    ( item )->list##_prev = head;

#define remove( list, item ) \
    ( item )->list##_next->list##_prev = ( item )->list##_prev; \
    ( item )->list##_prev->list##_next = ( item )->list##_next; \
    ( item )->list##_next = item; \
    ( item )->list##_prev = item;


// Implementation of LRUCache::LRUStatistics
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

LRUCache::LRUStatistics::LRUStatistics()
    :   Statistics( "LRUCache", Status::FACILITY_LRU_CACHE, Statistics::OPEN_RECYCLED ),
        cache_name( create_string( "cache_name" ) ),
        memory_limit( create_integer( "memory_limit" ) ),
        memory_size( create_integer( "memory_size" ) ),
        free_memory( create_integer( "free_memory" ) ),
        num_items( create_integer( "num_items" ) ),
        num_hits( create_integer( "num_hits" ) ),
        num_misses( create_integer( "num_misses" ) ),
        num_cached_items( create_integer( "num_cached_items" ) ),
        num_cache_out_candidates( create_integer( "num_cache_out_candidates" ) ),
        adaptive( create_boolean( "adaptive" ) ),
        current_tick( create_large_integer( "current_tick" ) ),
        cache_in_time( create_large_integer( "cache_in_time" ) )
    {
    cache_name = get_registry_path();
    }


// Implementation of LRUCache::LRUItem
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

LRUCache::LRUItem::LRUItem()
    {
    init( registration, this );
    init( lru, this );

    lock_count = 0;
    cached_in = false;
    cache_in_time = 0;
    cache_out_time = 0;

    cache = NULL;
    }


LRUCache::LRUItem::~LRUItem()
    {
    hope( !cache );
    }


void LRUCache::LRUItem::lock()
    {
    hope( cache );
    cache->lock( this );
    }


void LRUCache::LRUItem::unlock()
    {
    hope( cache );
    cache->unlock( this );
    }


LRUCache * LRUCache::LRUItem::get_cache()
    {
    return cache;
    }


int LRUCache::LRUItem::memory_size()
    {
    return 1;
    }


void LRUCache::LRUItem::cache_in()
    {
    }


void LRUCache::LRUItem::cache_out()
    {
    }


void LRUCache::LRUItem::flush()
    {
    }


bool LRUCache::LRUItem::is_dirty()
    {
    return false;
    }


std::string LRUCache::LRUItem::get_name()
    {
    std::ostringstream ss;

    ss << "LRUItem$" << this;

    return ss.str();
    }


// Implementation LRUCache
// ~~~~~~~~~~~~~~~~~~~~~~~

LRUCache::LRUCache
    (
    int     memory_limit,
    bool    adaptive
    )
    {
    statistics.adaptive = adaptive;
    statistics.num_items = 0;

    if( adaptive )
        {
        statistics.memory_limit = memory_limit;
        statistics.memory_size = 0;
        statistics.free_memory = 0;
        }
    else
        {
        statistics.memory_limit = memory_limit;
        statistics.memory_size = memory_limit;
        statistics.free_memory = memory_limit;
        }

    statistics.num_cached_items = 0;
    statistics.num_cache_out_candidates = 0;

    reset_statistics();
    
    #ifdef OPTION_DEBUG

    inside_callback = false;

    #endif
    }


LRUCache::~LRUCache()
    {
    hope( !statistics.num_items );
    }


void LRUCache::register_item
    (
    LRUItem *   item
    )
    {
    PREVENT_REENTRANCY();

    hope( item );
    hope( !item->cache );

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream  ss;

    ss  << statistics.cache_name << ": "
        << item->get_name() << " - registering.";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOWEST
            ),
        ss.str()
        );

    #endif

    insert_head( registration, item, &head );
    item->cache = this;

    statistics.num_items++;
    }


void LRUCache::unregister_item
    (
    LRUItem *   item
    )
    {
    PREVENT_REENTRANCY();

    hope( item );
    hope( item->cache == this );
    
    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream  ss;

    ss  << statistics.cache_name << ": "
        << item->get_name() << " - unregistering.";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOWEST
            ),
        ss.str()
        );

    #endif

    item->lock_count = 0;
    cache_out( item );

    remove( registration, item );
    item->cache = NULL;

    statistics.num_items--;
    }


void LRUCache::unregister_all()
    {
    PREVENT_REENTRANCY();

    while( statistics.num_items )
        {
        unregister_item( head.registration_next );
        }
    }


void LRUCache::flush
    (
    int flush_flags
    )
    {
    LRUItem * item;
    
    PREVENT_REENTRANCY();
    item = head.registration_next;

    while( item != &head )
        {
        const bool is_locked = item->lock_count > 0;
            
        if( !( is_locked && flags_on( flush_flags, FLUSH_UNLOCKED ) ) )
            {
            flush( item );
            }

        item = item->registration_next;
        }
    }


void LRUCache::lock
    (
    LRUItem *   item
    )
    {
    PREVENT_REENTRANCY();

    hope( item );
    hope( item->cache == this );

    statistics.current_tick++;

    if( !item->lock_count++ )
        {
        if( item->cached_in )
            {
            remove( lru, item );
            statistics.num_cache_out_candidates--;

            statistics.num_hits++;
            }
        else
            {
            cache_in( item );

            statistics.num_misses++;
            }
        }
    else
        {
        statistics.num_hits++;
        }
    }


void LRUCache::unlock
    (
    LRUItem *   item
    )
    {
    PREVENT_REENTRANCY();

    hope( item );
    hope( item->lock_count > 0 );
    hope( item->cache == this );

    if( !--item->lock_count )
        {
        if( item->cached_in )
            {
            insert_head( lru, item, &head );
            statistics.num_cache_out_candidates++;
            }
        }
    }


const LRUCache::LRUStatistics & LRUCache::get_statistics() const
    {
    return statistics;
    }


void LRUCache::reset_statistics()
    {
    statistics.num_hits = 0;
    statistics.num_misses = 0;
    }


void LRUCache::set_cache_name
    (
    const std::string & name
    )
    {
    statistics.cache_name = name;
    }


bool LRUCache::cache_in
    (
    LRUItem *   item
    )
    {
    int size;

    hope( item );

    if( item->cached_in )
        {
        return false;
        }

    size = item->memory_size();

    /* Adapt cache size. */

    if( statistics.adaptive )
        {
        if( statistics.current_tick - item->cache_out_time < THRASHING_PERIOD )
            {
            /* The item was recently cached out, try to increase cache size. */

            grow_cache( size * ( 256 + GROW_FACTOR ) / 256 );
            }
        }

    /* Allocate memory. */

    while( statistics.free_memory < size )
        {
        bool success;

        if( !statistics.num_cache_out_candidates )
            {
            /* Cache is full and nothing can be cached out. */

            if( grow_cache( size ) )
                {
                continue;
                }

            std::ostringstream  ss;

            ss << statistics.cache_name << ": "
               << "Out of cache memory while trying to cache in "
               << item->get_name() << ".\n"
               << size << " memory unit(s) needed.\n"
               << "Details:\n"
               << statistics;

            throw ExOutOfCacheMemory( ss.str() );
            }

        success = cache_out( head.lru_prev );
        hope( success );
        }

    /* Log operation. */

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream  ss;

    ss  << statistics.cache_name << ": "
        << item->get_name() << " - caching in.";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOWEST
            ),
        ss.str()
        );

    #endif

    /* Cache item into memory. */

    hope( item->cache == this );

        {
        INSIDE_CALLBACK();
        item->cache_in();
        }

    item->cached_in = true;
    item->cache_in_time = statistics.current_tick;

    statistics.free_memory -= size;
    statistics.num_cached_items++;

    return true;
    }


bool LRUCache::cache_out
    (
    LRUItem *   item
    )
    {
    hope( item );
    hope( !item->lock_count );
    hope( item->cache == this );

    if( !item->cached_in )
        {
        return false;
        }

    /* Flush item. */

    flush( item );

    /* Log operation. */

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream ss;

    ss  << statistics.cache_name << ": "
        << item->get_name() << " - caching out.";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOWEST
            ),
        ss.str()
        );

    #endif

    /* Remove item from memory. */

    hope( item->cache == this );

        {
        INSIDE_CALLBACK();
        item->cache_out();
        }

    item->cached_in = false;
    item->cache_out_time = statistics.current_tick;

    statistics.free_memory += item->memory_size();
    statistics.num_cached_items--;

    /* Remove item from LRU chain. */

    remove( lru, item );
    statistics.num_cache_out_candidates--;

    /* Adapt cache size. */

    if( statistics.adaptive )
        {
        if( statistics.current_tick - item->cache_in_time > LAZY_PERIOD &&
            statistics.current_tick - statistics.cache_in_time > LAZY_PERIOD )
            {
            /* Both the item and the cache are just too lazy... */

            shrink_cache( statistics.free_memory * SHRINK_FACTOR / 256 );
            }
        }

    return true;
    }


bool LRUCache::flush
    (
    LRUItem *   item
    )
    {
    hope( item );

    if( !item->cached_in )
        {
        return false;
        }

    if( !item->is_dirty() )
        {
        return false;
        }
    
    /* Log operation. */

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream ss;

    ss  << statistics.cache_name << ": "
        << item->get_name() << " - flushing.";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOWEST
            ),
        ss.str()
        );

    #endif

    /* Flush item. */

    hope( item->cache == this );

        {
        INSIDE_CALLBACK();
        item->flush();
        }

    return true;
    }


bool LRUCache::grow_cache
    (
    int size
    )
    {
    if( !statistics.adaptive )
        {
        return false;
        }

    if( statistics.memory_size + size > statistics.memory_limit )
        {
        return false;
        }

    /* Grow cache. */

    statistics.memory_size += size;
    statistics.free_memory += size;

    /* Log operation. */
    
    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream  ss;

    ss  << statistics.cache_name << ": "
        << "Cache size grown to " << statistics.memory_size
        << " memory unit(s).";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );

    #endif

    return true;
    }


bool LRUCache::shrink_cache
    (
    int size
    )
    {
    if( !statistics.adaptive )
        {
        return false;
        }

    if( statistics.free_memory < size )
        {
        return false;
        }

    /* Shrink cache. */

    statistics.free_memory -= size;
    statistics.memory_size -= size;

    /* Log operation. */

    #ifndef LIMITED_LRU_CACHE_LOGGING

    std::ostringstream ss;

    ss  << statistics.cache_name << ": "
        << "Cache size reduced to " << statistics.memory_size
        << " memory unit(s).";

    Global::log
        (
        STATUS_INFO
            (
            Status::FACILITY_LRU_CACHE,
            Status::PRIORITY_LOW
            ),
        ss.str()
        );

    #endif

    return true;
    }


#undef init
#undef insert_head
#undef remove

} // namespace Massiv::Core
} // namespace Massiv
