/** @file lru_cache.h
 * LRUCache.
 */

/*
 * begin        : Fri 28 Jun 2 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_LRU_CACHE_H
#define MASSIV_LRU_CACHE_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @def LIMITED_LRU_CACHE_LOGGING
 * Define this to limit LRU Cache logging.
 */
#define LIMITED_LRU_CACHE_LOGGING

/** @class LRUCache
 * Cache of cacheable items.
 * @see LRUItem.
 */
class LRUCache
    {
private:

    /**
     * Cache parameters.
     */
    enum Parameters
        {
        THRASHING_PERIOD =  8,
            /**< The cache is in "trashing" mode when it periodically
                 caches the same objects in & out within a time period
                 shorter than @c THRASHING_PERIOD. */
        GROW_FACTOR =       1024,
            /**< Cache grow factor. Determines how much cache grows
                 when it is in thrashing mode. */
        LAZY_PERIOD =       1024,
            /**< The cache gets "lazy" when it did not cached in any
                 object within a time period longer than @c LAZY_PERIOD. */
        SHRINK_FACTOR =     24
            /**< Cache shrink factor. Determines how much cache shrinks
                 when it is in lazy mode. */
        }; // Parameters

public:

    /**
     * Cacheable item. Item managed by LRUCache.
     * Each LRUItem is (logically) either "in memory" or "cached out".
     * Transitions among these states are signalled by invoking
     * cache_in() or cache_out() by the caching mechanism.
     * These methods are client-defined callbacks that typically
     * perform memory allocation or deallocation of additional data
     * that the item holds. Amount of memory required to place item
     * into memory should be reported by memory_size().
     * Prior accessing the additional data, item must be locked (lock()).
     * This ensures that the item will be "in memory".
     * Item won't be "cached out" until it is unlocked (unlock()).
     * LRUCache, analyzing usage statistics of registered items and
     * their memory requirements, chooses what items will be placed
     * "in memory" and what items will be "cached out".
     *
     * @warning LRUItem callbacks must not re-enter the cache.
     */
    class LRUItem
        {
        friend class LRUCache;

    public:

        //
        // Main.
        //

        /**
         * Constructor.
         */
        LRUItem();

        /**
         * Destructor.
         */
        virtual ~LRUItem();

        /**
         * Locks object in memory.
         */
        void lock();

        /**
         * Unlocks object.
         */
        void unlock();

        /**
         * Returns pointer to cache.
         */
        LRUCache * get_cache();

    protected:

        //
        // Cache callbacks.
        //

        /**
         * Returns how much "memory units" the item requires.
         * Each item can require different amount of memory.
         */
        virtual int memory_size();

        /**
         * Called by LRUCache when it caches the item into memory.
         */
        virtual void cache_in();

        /**
         * Called by LRUCache when it caches the item out of memory.
         * Dirty items are flushed before they are cached out.
         */
        virtual void cache_out();

        /**
         * Called by LRUCache to flush a dirty item.
         */
        virtual void flush();

        /**
         * Called by LRUCache to test if the item is dirty.
         * The "dirty flag" is utilized by @c LRUCache::flush(), which
         * flushes dirty items only.
         */
        virtual bool is_dirty();

        /**
         * Returns printable name of the item.
         */
        virtual std::string get_name();

    private:

        LRUItem *   lru_next;
            /**< Next item in LRU chain. */
        LRUItem *   lru_prev;
            /**< Previous item in LRU chain. */
        LRUItem *   registration_next;
            /**< Next item in item registration list. */
        LRUItem *   registration_prev;
            /**< Next item in item registration list. */
        int         lock_count;
            /**< How many times was item locked in memory. */
        bool        cached_in : 1;
            /**< Is item cached in? */
        Int64       cache_out_time;
            /**< When the item was cached out for the last time. */
        Int64       cache_in_time;
            /**< When the item was cached in for the last time. */
        LRUCache *  cache;
            /**< Pointer to cache that registers this item. */
        }; // LRUItem

    friend class LRUItem;

    /**
     * LRU Cache statistics.
     */
    struct LRUStatistics : public Statistics
        {
    public:

        /**
         * Constructor.
         */
        LRUStatistics();

    public:

        String &        cache_name;
            /**< Printable name of the monitored cache. */
        Integer &       memory_limit;
            /**< Maximum amount of memory units available to the cache. */
        Integer &       memory_size;
            /**< Current number of memory units available to the cache. */
        Integer &       free_memory;
            /**< Number of free memory units (relative to @c memory_size). */
        Integer &       num_items;
            /**< Number of registered items. */
        Integer &       num_hits;
            /**< Number of cache hits. */
        Integer &       num_misses;
            /**< Number of cache misses. */
        Integer &       num_cached_items;
            /**< Number of items that are "in memory". */
        Integer &       num_cache_out_candidates;
            /**< Number of items that are "in memory" and unlocked.
                 This is number of items linked in LRU chain. */
        Boolean &       adaptive;
            /**< Are we working in an adaptive mode? */
        LargeInteger &  current_tick;
            /**< Internal cache tick. Tick is increment by @c lock() operation. */
        LargeInteger &  cache_in_time;
            /**< When cache_in() was called for the last time. */

        }; // LRUStatistics

    /**
     * Flags used by @c LRUCache::flush().
     */
    enum FlushFlags
        {
        FLUSH_UNLOCKED =        1,
            /**< Flush unlocked items only. */
        DEFAULT_FLUSH_FLAGS =   0
            /**< Default flush flags. All items are flushed. */
        }; // FlushFlags

public:

    /**
     * Constructor.
     * LRUCache can work in two modes:
     *  1) non-adaptive:
     *  memory_size == memory_limit
     *  2) adaptive:
     *  memory_size <= memory_limit,
     *  caching mechanism optimizes @c memory_size.
     */
    LRUCache
        (
        int  memory_limit,
        bool adaptive = false
        );

    /**
     * Destructor.
     * All items must be unregistered prior destruction of the cache.
     */
    ~LRUCache();

    /**
     * Registers item into the cache.
     */
    void register_item
        (
        LRUItem *   item
        );

    /**
     * Unregisters item from the cache.
     */
    void unregister_item
        (
        LRUItem *   item
        );

    /**
     * Unregisters all registered items.
     */
    void unregister_all();

    /**
     * Flushes cache.
     * @c flush_flags (an ORed @c FlushFlag combination) determines
     * what items should be flushed.
     */
    void flush
        (
        int flush_flags = DEFAULT_FLUSH_FLAGS
        );

    /**
     * Locks object in memory.
     * Object will not be cached out until it is unlocked.
     */
    void lock
        (
        LRUItem *   item
        );

    /**
     * Unlocks object.
     * Object may be cached out.
     */
    void unlock
        (
        LRUItem *   item
        );

    /**
     * Returns cache statistics.
     */
    const LRUStatistics & get_statistics() const;

    /**
     * Resets cache statistics.
     */
    void reset_statistics();

    /**
     * Names the cache.
     * Used for debugging purposes, logging.
     */
    void set_cache_name
        (
        const std::string & name
        );

private:

    /**
     * Places item into memory.
     * Returns true if object was not "in memory".
     */
    bool cache_in
        (
        LRUItem *   item
        );

    /**
     * Removes item from the memory.
     * Returns true if object was in memory.
     */
    bool cache_out
        (
        LRUItem *   item
        );

    /**
     * Flushes the item.
     * Returns true if object was dirty.
     */
    bool flush
        (
        LRUItem *   item
        );

    /**
     * Grows current cache size (@c memory_size ) by @c size units.
     * Returns true if succeeded.
     */
    bool grow_cache
        (
        int size
        );

    /**
     * Shrinks current cache size (@c memory_size ) by @c size units.
     * Returns true if succeeded.
     */
    bool shrink_cache
        (
        int size
        );

private:

    LRUItem         head;
        /**< Head of item registration list and LRU chain.
             LRU chain is a list of may-be-cached-out candidates. */
    LRUStatistics   statistics;
        /**< Cache statistics. */

public:

    #ifdef OPTION_DEBUG

    bool            inside_callback;
        /**< True if inside a callback. Defined in order to detect
             reentrancy and signal an error (debug mode only). */

    #endif

    }; // LRUCache

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_LRU_CACHE_H
