/** @file math.h
 * Include this header instead of cmath.
 *
 * It includes cmath and properly imports functions to namespace std.
 * Additional mathematic functions will be defined in namespace Massiv::Core.
 *
 * You can use FloatTraits template to access information about given
 * floating-point type and call mathematics functions with specified precisiion.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_MATH_H
#define MASSIV_MATH_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#include <cmath>


// Import functions to namespace std if they are not there.
// The following functions should be defined in namespace std:
//   acos    cos     fmod    modf    tan
//   asin    cosh    frexp   pow     tanh
//   atan    exp     ldexp   sin
//   atan2   fabs    log     sinh
//   ceil    floor   log10   sqrt

// STLPort/MSVC does not import math symbols into namespace std.
// Note that for only "double" versions are imported. I'm too
// lazy to write std::floor( float ) using ::floorf(), etc.
// You should use #FloatTraits instead of direct calls to math functions.
#if defined( OPTION_STLPORT ) && defined( COMPILER_MSVC )

namespace std {

using ::acos;
using ::asin;
using ::atan;
using ::atan2;
using ::ceil;
using ::cos;
using ::cosh;
using ::exp;
using ::fabs;
using ::floor;
using ::fmod;
using ::frexp;
using ::ldexp;
using ::log;
using ::log10;
using ::modf;
using ::pow;
using ::sin;
using ::sinh;
using ::sqrt;
using ::tan;
using ::tanh;

} // namespace std

#endif


/** @addtogroup utility
 * @{
 */

/** @def MATH_CSTYLE
 * If compiler (C++ library) does not implement @c float versions of math functions in
 * <tt>namespace std</tt>, use C-style functions instead.
 *
 * C-style:
 * @code
 * double ::floor( double );
 * float ::floorf( float );
 * @endcode
 *
 * C++-style:
 * @code
 * double std::floor( double );
 * float std::floor( float );
 * @endcode
 *
 * If MATH_CSTYLE is defined, it's expected that @c double versions of math functions are
 * defined in <tt>namespace std</tt> too.
 *
 * User of math.h can safely ignore this. Use Massiv::Core::FloatTraits instead.
 */

// MSVC 6.0 and MSVC 7.0
// Stoupik FIXME: Pokud je to problem i v MSVC 7.1, opravte si to, jinak smazte
// tento comment.
#if defined( COMPILER_MSVC ) && ( _MSC_VER <= 1300 )

#define MATH_CSTYLE

#endif


/** @name Useful double precision math constants.
 * Sometimes similar constants are defined in cmath header, but it's not
 * specified in C++ standard, so we define them again here.
 */
//@{
#define MATH_E              2.7182818284590452354   ///< e
#define MATH_LOG2E          1.4426950408889634074   ///< log_2( e )
#define MATH_LOG10E         0.43429448190325182765  ///< log_10( e )
#define MATH_LN2            0.69314718055994530942  ///< log_e( 2 )
#define MATH_LN10           2.30258509299404568402  ///< log_e( 10 )
#define MATH_PI             3.14159265358979323846  ///< pi
#define MATH_PI_2           1.57079632679489661923  ///< pi / 2
#define MATH_PI_4           0.78539816339744830962  ///< pi / 4
#define MATH_1_PI           0.31830988618379067154  ///< 1 / pi
#define MATH_2_PI           0.63661977236758134308  ///< 2 / pi, @b not 2 * pi
#define MATH_2_SQRTPI       1.12837916709551257390  ///< 2 / sqrt( pi )
#define MATH_SQRT2          1.41421356237309504880  ///< sqrt( 2 )
#define MATH_SQRT1_2        0.70710678118654752440  ///< 1 / sqrt( 2 )
//@}


/** @name Useful single precision math constants.
 * Sometimes similar constants are defined in cmath header, but it's not
 * specified in C++ standard, so we define them again here.
 */
//@{
#define MATHF_E             2.7182818284590452354f  ///< e
#define MATHF_LOG2E         1.4426950408889634074f  ///< log_2( e )
#define MATHF_LOG10E        0.43429448190325182765f ///< log_10( e )
#define MATHF_LN2           0.69314718055994530942f ///< log_e( 2 )
#define MATHF_LN10          2.30258509299404568402f ///< log_e( 10 )
#define MATHF_PI            3.14159265358979323846f ///< pi
#define MATHF_PI_2          1.57079632679489661923f ///< pi / 2
#define MATHF_PI_4          0.78539816339744830962f ///< pi / 4
#define MATHF_1_PI          0.31830988618379067154f ///< 1 / pi
#define MATHF_2_PI          0.63661977236758134308f ///< 2 / pi, @b not 2 * pi
#define MATHF_2_SQRTPI      1.12837916709551257390f ///< 2 / sqrt( pi )
#define MATHF_SQRT2         1.41421356237309504880f ///< sqrt( 2 )
#define MATHF_SQRT1_2       0.70710678118654752440f ///< 1 / sqrt( 2 )
//@}


/** @name Near-zero floating-point values. */
//@{
#define FLOAT32_EPSILON     0.00001f
#define FLOAT64_EPSILON     0.0000001
//@}

/** @} */ // addtogroup utility


// Define helper math structures and functions.

namespace Massiv {
namespace Core {


/** @addtogroup type_traits
 * @{
 */

/**
 * Basic traits for floating-point types.
 */
template< typename T >
struct BasicFloatTraits;


/**
 * BasicFloatTraits for Float32.
 */
template<>
struct MASSIV_EXPORTED BasicFloatTraits< Float32 >
    {
    /// The type.
    typedef Float32 FloatType;
    /// Paired integer type.
    typedef Int32 IntegerType;

public:

    /// Get epsilon value.
    static FloatType epsilon()
        {
        return FLOAT32_EPSILON;
        }

public:

    static FloatType floor
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::floorf( x );
        #else
        return std::floor( x );
        #endif
        }

    static FloatType ceil
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::ceilf( x );
        #else
        return std::ceil( x );
        #endif
        }

    static FloatType abs
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::fabsf( x );
        #else
        return std::fabs( x );
        #endif
        }

    static FloatType mod
        (
        FloatType x,
        FloatType y
        )
        {
        #ifdef MATH_CSTYLE
        return ::fmodf( x, y );
        #else
        return std::fmod( x, y );
        #endif
        }

public:

    static FloatType sin
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::sinf( x );
        #else
        return std::sin( x );
        #endif
        }

    static FloatType cos
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::cosf( x );
        #else
        return std::cos( x );
        #endif
        }

    static FloatType tan
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::tanf( x );
        #else
        return std::tan( x );
        #endif
        }

    static FloatType asin
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::asinf( x );
        #else
        return std::asin( x );
        #endif
        }

    static FloatType acos
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::acosf( x );
        #else
        return std::acos( x );
        #endif
        }

    static FloatType atan2
        (
        FloatType y,
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::atan2f( y, x );
        #else
        return std::atan2( y, x );
        #endif
        }

public:

    static FloatType exp
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::expf( x );
        #else
        return std::exp( x );
        #endif
        }

    static FloatType log
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::logf( x );
        #else
        return std::log( x );
        #endif
        }

    static FloatType pow
        (
        FloatType x,
        FloatType y
        )
        {
        #ifdef MATH_CSTYLE
        return ::powf( x, y );
        #else
        return std::pow( x, y );
        #endif
        }

    static FloatType sqrt
        (
        FloatType x
        )
        {
        #ifdef MATH_CSTYLE
        return ::sqrtf( x );
        #else
        return std::sqrt( x );
        #endif
        }

    }; // struct BasicFloatTraits< Float32 >


/**
 * BasicFloatTraits for Float64.
 */
template<>
struct MASSIV_EXPORTED BasicFloatTraits< Float64 >
    {
    /// The type.
    typedef Float64 FloatType;
    /// Paired integer type.
    typedef Int32 IntegerType;

public:

    /// Get epsilon value.
    static FloatType epsilon()
        {
        return FLOAT64_EPSILON;
        }

public:

    static FloatType floor
        (
        FloatType x
        )
        {
        return std::floor( x );
        }

    static FloatType ceil
        (
        FloatType x
        )
        {
        return std::ceil( x );
        }
    
    static FloatType abs
        (
        FloatType x
        )
        {
        return std::fabs( x );
        }

    static FloatType mod
        (
        FloatType x,
        FloatType y
        )
        {
        return std::fmod( x, y );
        }

public:

    static FloatType sin
        (
        FloatType x
        )
        {
        return std::sin( x );
        }

    static FloatType cos
        (
        FloatType x
        )
        {
        return std::cos( x );
        }

    static FloatType tan
        (
        FloatType x
        )
        {
        return std::tan( x );
        }

    static FloatType asin
        (
        FloatType x
        )
        {
        return std::asin( x );
        }

    static FloatType acos
        (
        FloatType x
        )
        {
        return std::acos( x );
        }

    static FloatType atan2
        (
        FloatType y,
        FloatType x
        )
        {
        return std::atan2( y, x );
        }

public:

    static FloatType exp
        (
        FloatType x
        )
        {
        return std::exp( x );
        }

    static FloatType log
        (
        FloatType x
        )
        {
        return std::log( x );
        }

    static FloatType pow
        (
        FloatType x,
        FloatType y
        )
        {
        return std::pow( x, y );
        }

    static FloatType sqrt
        (
        FloatType x
        )
        {
        return std::sqrt( x );
        }

    }; // struct BasicFloatTraits< Float64 >


/**
 * Information about floating-point type.
 */
template< typename T >
struct MASSIV_EXPORTED FloatTraits : public BasicFloatTraits< T >
    {
public:
    
    /// Superclass.
    typedef BasicFloatTraits< T > Super;

    /// The type.
    typedef typename Super::FloatType   FloatType;
    /// Paired integer type.
    typedef typename Super::IntegerType IntegerType;

    using Super::epsilon;
    using Super::floor;
    using Super::ceil;
    using Super::abs;
    using Super::acos;

public:

    /// Epsilon^2.
    static FloatType epsilon_squared()
        {
        return epsilon() * epsilon();
        }
    
    /// Test signness.
    static bool is_positive
        (
        FloatType x
        )
        {
        // FIXME: Tady by stalo za to ukrast Honzikuv kod z Quark.

        return x >= static_cast< FloatType>( 0 );
        }
    
    /// Test signness.
    static bool is_negative
        (
        FloatType x
        )
        {
        return !is_positive( x );
        }

    /// Round.
    static FloatType round
        (
        FloatType x
        )
        {
        if( is_positive( x ) )
            {
            return floor( x + 0.5f );
            }
        else
            {
            return ceil( x - 0.5f );
            }
        }
    
    /// Truncate to integer.
    static IntegerType to_integer
        (
        FloatType x
        )
        {
        #ifndef OPTION_OPTIMIZE

        if( is_positive( x ) )
            {
            return static_cast< IntegerType >( floor( x ) );
            }
        else
            {
            return static_cast< IntegerType >( ceil( x ) );
            }

        #else

        // We silently hope that FPU is set up so that it truncates
        // fractional part when converting a floating point value to
        // an integer.

        // FIXME: Tady by stalo za to ukrast Honzikuv kod z Quark.

        return static_cast< IntegerType >( x );

        #endif
        }
    
    /// Floor.
    static IntegerType floor_int
        (
        FloatType x
        )
        {
        return static_cast< IntegerType >( std::floor( x ) );
        }
    
    /// Ceil.
    static IntegerType ceil_int
        (
        FloatType x
        )
        {
        return static_cast< IntegerType >( std::ceil( x ) );
        }

    /// Distance.
    static FloatType distance
        (
        FloatType x,
        FloatType y
        )
        {
        return abs( x - y );
        }

    /// Epsilon-equals.
    static bool epsilon_equals
        (
        FloatType x,
        FloatType y
        )
        {
        return distance( x, y ) < epsilon();
        }

    /// Epsilon-equals to zero.
    static bool epsilon_zero
        (
        FloatType x
        )
        {
        return epsilon_equals( x, static_cast< FloatType >( 0 ) );
        }
    
// more complex functions, used to be global functions. clamp() still is.
public:

    /**
     * Convert angle in degrees to angle in radians.
     */
    static INLINE FloatType degrees_to_radians
        (
        FloatType degrees
        )
        {
        return degrees * static_cast< FloatType >( MATH_PI ) / static_cast< FloatType >( 180 );
        };

    /**
     * Convert angle in radians to angle in degrees.
     */
    static INLINE FloatType radians_to_degrees
        (
        FloatType radians
        )
        {
        return radians * static_cast< FloatType >( 180 ) / static_cast< FloatType >( MATH_PI );
        };

    /**
     * Arc cosine that accepts values slightly out of ( -1, 1 ) range.
     */
    static INLINE FloatType safe_acos
        (
        FloatType x
        )
        {
        hope( x > static_cast< FloatType >( -1 ) - epsilon() );
        hope( x < static_cast< FloatType >( 1 ) + epsilon() );

        if( x < static_cast< FloatType >( -1 ) )
            {
            return acos( static_cast< FloatType >( -1 ) );
            }
        if( x > 1.0f )
            {
            return acos( static_cast< FloatType >( 1 ) );
            }
        return acos( x );
        };


    /**
     * Arc sine that accepts values slightly out of ( -1, 1 ) range.
     */
    static INLINE FloatType safe_asin
        (
        FloatType x
        )
        {
        hope( x > static_cast< FloatType >( -1 ) - epsilon() );
        hope( x < static_cast< FloatType >( 1 ) + epsilon() );

        if( x < static_cast< FloatType >( -1 ) )
            {
            return asin( static_cast< FloatType >( -1 ) );
            }
        if( x > 1.0f )
            {
            return asin( static_cast< FloatType >( 1 ) );
            }
        return asin( x );
        };

    /**
     * Generic "modulo" function.
     *
     * Returns @a result in range < @a minimum, @a maximum ) for which the following
     * holds true:
     * @code
     * result == value + K * ( maximum - minimum )
     * @endcode
     * for some integral @a K.
     *
     * @todo Better function name. This is not "clamp".
     */
    static INLINE FloatType clamp_periodic
        (
        FloatType value,
        FloatType minimum,
        FloatType maximum
        )
        {
        const FloatType delta = maximum - minimum;
        hope( delta > epsilon() );

        const FloatType period_count = floor( ( value - minimum ) / delta );
        const FloatType result = value - period_count * delta;

        // Because of floating-point precision problems, result might be slightly
        // below minimum or above maximum.

        if( ( result < minimum ) || ( result >= maximum ) )
            {
            return minimum;
            }
        return result;
        };

    }; // struct FloatTraits

/** @} */ // addtogroup type_traits


/**
 * Clamp given @a value to range < @a minimum, @a maximum >.
 *
 * @ingroup utility
 */
template< typename T >
inline T clamp
    (
    T value,
    T minimum,
    T maximum
    )
    {
    if( value < minimum )
        {
        return minimum;
        }
    if( value > maximum )
        {
        return maximum;
        }
    return value;
    }


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_MATH_H
