/** @file quote_string.cpp
 * These functions/classes offer "symmetric" std:string reading/writing.
 * See quote_string.h for more info.
 */

/*
 * begin        : Mar 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/quote_string.h"
#include <iostream>


namespace Massiv {
namespace Core {


namespace {


/**
 * Convert 4bit number to hexadecimal character.
 */
char encode_digit
    (
    unsigned int d
    )
    {
    static char table[] = "0123456789abcdef";

    hope( d <= 15 );
    return table[ d ];
    }


/**
 * Convert hexadecimal character to 4bit number.
 *
 * @return Number outside 0..15 range if @a c is not hexadecimal character.
 */
unsigned int decode_digit
    (
    char c
    )
    {
    if( c >= 'a' && c <= 'f' )
        {
        return c - 'a' + 10;
        }
    if( c >= 'A' && c <= 'F' )
        {
        return c - 'A' + 10;
        }
    if( c >= '0' && c <= '9' )
        {
        return c - '0';
        }

    return static_cast< unsigned int >( -1 );
    }


/**
 * Determine if string must be quoted.
 *
 * @return
 *      @c true is #write_quoted_string() must be used to write the string
 *      to be able to read it back by #QuoteStringCommon::read_string().
 *      @c false if standard @c operator<< may be used to write the string.
 */
bool should_quote
    (
    const std::string & str
    )
    {
    // Empty string must be quoted.

    if( str.empty() )
        {
        return true;
        }

    // Check all characters.

    std::string::const_iterator it;
    for( it = str.begin(); it != str.end(); ++it )
        {
        const char c = *it;
        if( c <= 32 || c >= 127 )
            {
            // Nonprintable or weird character.

            return true;
            }
        if( c == '"' || c == '\\' || c == '#' )
            {
            // Character that may cause problems under some circumstances.

            return true;
            }
        }

    // Ok.
    
    return false;
    }


/**
 * Write quoted string to @c ostream.
 */
void write_quoted_string
    (
    std::ostream &      os,
    const std::string & str
    )
    {
    std::ostream::sentry sentry( os );
    if( sentry )
        {
        // Write initial double quote.

        os.put( '"' );

        // Write the string.

        std::string::const_iterator it;
        for( it = str.begin(); it != str.end(); ++it )
            {
            const char c = *it;
            const unsigned int d = static_cast< unsigned char >( c );

            // Escape sequences.

            if( c == '"' )
                {
                os.write( "\\\"", 2 );
                }
            else if( c == '\\' )
                {
                os.write( "\\\\", 2 );
                }
            else if( c == '\n' )
                {
                os.write( "\\n", 2 );
                }
            else if( c == '\r' )
                {
                os.write( "\\r", 2 );
                }
            else if( c == '\t' )
                {
                os.write( "\\t", 2 );
                }

            // "Nice" ASCII characters.

            else if( d >= 32 && d < 127 )
                {
                os.put( c );
                }

            // The rest.

            else
                {
                os.write( "\\x", 2 );
                os.put( encode_digit( ( d >> 4 ) & 0x0f ) );
                os.put( encode_digit( d & 0x0f ) );
                }
            }

        // Write trailing double quote.

        os.put( '"' );
        }
    }


/**
 * Read quoted string from @c istream.
 */   
void read_quoted_string
    (
    std::istream &  is,
    std::string &   str
    )
    {
    // Consume the quote.

    std::istream::char_type c = is.get();
    hope( c == '"' );

    // Parse the string.

    while( 1 )
        {
        c = is.get();
        if( is.fail() )
            {
            break;
            }

        if( c == '"' )
            {
            // The end.

            break;
            }

        if( c == '\\' )
            {
            // Escape sequence.

            const std::istream::char_type e = is.get();
            if( is.fail() )
                {
                break;
                }

            switch( e )
                {
                case '"':
                    str.push_back( '"' );
                    break;
                case '\\':
                    str.push_back( '\\' );
                    break;
                case 'n':
                    str.push_back( '\n' );
                    break;
                case 'r':
                    str.push_back( '\r' );
                    break;
                case 't':
                    str.push_back( '\t' );
                    break;
                case 'x':
                    {
                    const std::istream::char_type hc = is.get();
                    const std::istream::char_type lc = is.get();
                    const unsigned int hd = decode_digit( hc );
                    const unsigned int ld = decode_digit( lc );
                    if( hd >= 16 || ld >= 16 )
                        {
                        is.setstate( std::ios_base::failbit );
                        }
                    if( is.fail() )
                        {
                        break;
                        }
                    str.push_back( static_cast< char >( ( hd << 4 ) | ld ) );
                    break;
                    }
                }
            }
        else
            {
            // Standard character.

            str.push_back( c );
            }
        }
    }


} // anonymous namespace


void QuoteStringCommon::write
    (
    std::ostream &      os,
    const std::string & str,
    bool                force_quotes
    )
    {
    if( force_quotes || should_quote( str ) )
        {
        write_quoted_string( os, str );
        }
    else
        {
        os << str;
        }
    }


void QuoteStringCommon::read
    (
    std::istream &      is,
    std::string &       str
    )
    {
    std::istream::sentry sentry( is );
    if( sentry )
        {
        // Starts with a quote?

        const std::istream::char_type c = is.peek();
        if( c == '"' )
            {
            // Yes - read quoted string.

            str.clear();
            read_quoted_string( is, str );
            }
        else
            {
            // No - read standard string.

            is >> str;
            }
        }
    }


ConstQuoteString quote_const_string
    (
    const std::string & str,
    bool                force_quotes
    )
    {
    return ConstQuoteString( str, force_quotes );
    }


QuoteString quote_string
    (
    std::string &       str,
    bool                force_quotes
    )
    {
    return QuoteString( str, force_quotes );
    }
    

std::ostream & operator<<
    (
    std::ostream &              os,
    const ConstQuoteString &    str
    )
    {
    QuoteStringCommon::write( os, str.str, str.force_quotes );
    return os;
    }


std::ostream & operator<<
    (
    std::ostream &              os,
    const QuoteString &         str
    )
    {
    QuoteStringCommon::write( os, str.str, str.force_quotes );
    return os;
    }


std::istream & operator>>
    (
    std::istream &              is,
    QuoteString                 str
    )
    {
    QuoteStringCommon::read( is, str.str );
    return is;
    }


} // namespace Massiv::Core
} // namespace Massiv
