/** @file quote_string.h
 * These functions/classes offer "symmetric" std:string reading/writing.
 * That means that any string written to stream using operator>> can be
 * read back by single call to operator<<.
 *
 * C-like escape sequences are used to encode common special characters,
 * which makes nearly all written strings human-readable. The following
 * escape sequences are used when writing and recognized when reading:
 * - <tt>\\"</tt> - double quote.
 * - <tt>\\\\</tt> - backslash.
 * - <tt>\\n</tt> - new line.
 * - <tt>\\r</tt> - carriage return.
 * - <tt>\\t</tt> - horizontal tab.
 * - <tt>\\xNN</tt> - byte with hexadecimal value NN.
 *
 * Basic usage is simple:
 * @code
 * std::ostream os;
 * const std::string cs;
 * os << Massiv::Core::quote_const_string( s );
 *
 * std::istream is;
 * std::string s;
 * is >> Massiv::Core::quote_string( s );
 * @endcode
 */

/*
 * begin        : Mar 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_QUOTE_STRING_H
#define MASSIV_QUOTE_STRING_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <iosfwd>
#include <string>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * Common public string reading and writing methods.
 */
class MASSIV_EXPORTED QuoteStringCommon
    {
    NO_DEFAULT_CONSTRUCTOR( QuoteStringCommon );
    NO_COPY_CONSTRUCTOR( QuoteStringCommon );
    NO_ASSIGNMENT_OPERATOR( QuoteStringCommon );

public:

    /**
     * Write string to given ostream.
     *
     * If @a force_quotes is @c false and string @a str contains only
     * "nice" non-whitespace characters, standard @c operator<< will be used
     * to write the string. Otherwise string will be quoted and escape
     * sequences will be used for special characters.
     */
    static void write
        (
        std::ostream &      os,
        const std::string & str,
        bool                force_quotes
        );

    /**
     * Read string from given istream.
     *
     * If token starts with a '"' character, quoted string will be read.
     * Otherwise standard @c std::string @c operator<< will be used to read
     * the string.
     */
    static void read
        (
        std::istream &      is,
        std::string &       str
        );

    }; // class QuoteStringCommon


/**
 * Class that keeps @c const reference to @c std::string.
 * It supports formatted ostream output.
 */
class MASSIV_EXPORTED ConstQuoteString
    {
public:

    /** Constructor. */
    ConstQuoteString
        (
        const std::string & the_str,
        bool                the_force_quotes
        ) : str( the_str ),
            force_quotes( the_force_quotes )
        
        {
        }

public:

    const std::string & str;
        /**< Reference to bound string. */

    bool                force_quotes;
        /**< Quote string even if not necessary when writing. */

    }; // class ConstQuoteString

    
/**
 * Class that keeps reference to @c std::string.
 * It supportes formatted iostream input and output.
 */
class QuoteString
    {
public:

    /** Constructor. */
    QuoteString
        (
        std::string &   the_str,
        bool            the_force_quotes
        ) : str( the_str ),
            force_quotes( the_force_quotes )
        {
        }

public:

    std::string &   str;
        /**< Reference to bound string. */

    bool            force_quotes;
        /**< Quote string even if not necessary when writing. */

    }; // class QuoteString


/**
 * Bind given <tt>const std::string &</tt> to ConstQuoteString, which can be
 * used to write quoted string to @c std::ostream.
 * If @a force_quotes is @c false (default value), string will be written
 * unquoted if it contains only "nice" non-whitespace characters.
 */
ConstQuoteString MASSIV_EXPORTED quote_const_string
    (
    const std::string & str,
    bool                force_quotes = false
    );


/**
 * Bind given <tt>std::string &</tt> to QuoteString, which can be
 * used to read and write quoted string to @c std::ostream / from @c std::istream.
 * If @a force_quotes is @c false (default value), string will be written
 * unquoted if it contains only "nice" non-whitespace characters.
 */
QuoteString MASSIV_EXPORTED quote_string
    (
    std::string &       str,
    bool                force_quotes = false
    );


/**
 * Write quoted string to given @c std::ostream.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &              os,
    const ConstQuoteString &    str
    );


/**
 * Write quoted string to given @c std::ostream.
 */
MASSIV_EXPORTED std::ostream & operator<<
    (
    std::ostream &              os,
    const QuoteString &         str
    );


/**
 * Read quoted string from given @c std::istream.
 */
MASSIV_EXPORTED std::istream & operator>>
    (
    std::istream &              is,
    QuoteString                 str
    );

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_QUOTE_STRING_H
