/** @file reference.cpp
 * Reference.
 */

/*
 * begin        : Tue Nov 6 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/reference.h"
#include "thread/synch.h"
#include "thread/atomic.h"

namespace Massiv {
namespace Core {


// Implementation of SharedObject
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SharedObject::SharedObject() :
    reference_count( 0 )
    {
    }


SharedObject::SharedObject
    (
    const SharedObject & other
    ) : reference_count( 0 ) // Must not be copied from the other object!
    {
    }


SharedObject::~SharedObject()
    {
    hope( reference_count == 0 );

    // If this fails:
    // The object was allocated on the stack. It became out of its
    // scope so it is being destroyed now but someone still has
    // reference to it.
    // Or someone is trying to delete native pointer stored in the reference.
    }
    
    
SharedObject & SharedObject::operator=
    (
    const SharedObject & other
    )
    {
    // Do NOT copy reference_count from @c other!
    return *this;
    }


int SharedObject::get_reference_count() const
    {
    return reference_count;
    }


// Implementation of MTReferenceTraits
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void MTReferenceTraits::add_reference
    (
    const SharedObject * native_ptr
    )
    {
    if( native_ptr )
        {
        Atomic::increment( native_ptr->reference_count );
        }
    }


void MTReferenceTraits::remove_reference
    (
    SharedObject * native_ptr
    )
    {
    if( native_ptr )
        {
        if( !Atomic::decrement( native_ptr->reference_count ) )
            {
            delete native_ptr;

            return;
            }
        }
        
    }


void MTReferenceTraits::change_reference
    (
    const SharedObject *    native_ptr1,
    SharedObject *          native_ptr2
    )
    {
    if( native_ptr1 )
        {
        Atomic::increment( native_ptr1->reference_count );
        }

    if( native_ptr2 )
        {
        if( !Atomic::decrement( native_ptr2->reference_count ) )
            {
            delete native_ptr2;

            return;
            }
        }
    }


} // namespace Massiv::Core
} // namespace Massiv
