/** @file reference.h s
 * Reference-counted object.
 */

/*
 * begin        : Tue Nov 6 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef MASSIV_REFERENCE_H
#define MASSIV_REFERENCE_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <cstdlib>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class SharedObject
 * Reference-counted object base class.
 * For information about shared objects, their usage
 * and references to them see Reference<>.
 * To make your object "shareable", the object must
 * inherit from SharedObject and have its destructor
 * protected, so that user can not destroy
 * it by "delete object".
 */
class MASSIV_EXPORTED SharedObject
    {
private:

    friend class ReferenceTraits;
    friend class MTReferenceTraits;

protected:

    /** Constructor. */
    SharedObject();
    
    /** Copy constructor. */
    SharedObject
        (
        const SharedObject & other
        );

    /** Destructor. */
    virtual ~SharedObject();
    
    /** Operator=. */
    SharedObject & operator=
        (
        const SharedObject & other
        );

public:

    /** Get number of references to this object. */
    int get_reference_count() const;

private:

    mutable int reference_count;
        /**< Number of references to this object. */

    }; // SharedObject

/**
 * Default reference traits.
 * Implements simple reference counting.
 */
class MASSIV_EXPORTED ReferenceTraits
    {
public:

    /** Adds reference to shared object. */
    static void add_reference
        (
        const SharedObject * native_ptr
        )
        {
        if( native_ptr )
            {
            native_ptr->reference_count++;
            }
        }

    /** Removes reference from shared object. */
    static void remove_reference
        (
        SharedObject * native_ptr
        )
        {
        if( native_ptr && !--native_ptr->reference_count )
            {
            delete native_ptr;
            }
        }

    /** Changes reference to shared object from native_ptr2 to native_ptr1. */
    static void change_reference
        (
        const SharedObject *    native_ptr1,
        SharedObject *          native_ptr2
        )
        {
        add_reference( native_ptr1 );
        remove_reference( native_ptr2 );
        }
    }; // ReferenceTraits

/**
 * Multithreaded reference traits.
 * Implements thread-safe reference counting.
 */
class MASSIV_EXPORTED MTReferenceTraits
    {
public:

    /** Adds reference to shared object. */
    static void add_reference
        (
        const SharedObject * native_ptr
        );

    /** Removes reference from shared object. */
    static void remove_reference
        (
        SharedObject * native_ptr
        );

    /** Changes reference to shared object. */
    static void change_reference
        (
        const SharedObject *    native_ptr1,
        SharedObject *          native_ptr2
        );
    }; // MTReferenceTraits

/** @class Reference
 * Reference to shared (reference-counted) object.
 * Reference object substitutes native pointer to SharedObject or @c Type.
 * The implementation automatically notifies the object
 * about adding or removing references to it
 * so that the object can track its usage.
 * If there's no reference to shared object,
 * the object is automatically garbage collected.
 * On the other hand this prevents user from
 * "deleting" object that is in use (= still referenced) by another object.
 * To remove a reference to shared object,
 * set the reference to NULL or let it destroy itself when
 * it is going out of its scope.
 * @c Traits parameter determines implementation of Reference.
 * It can either be @c ReferenceTraits or @c MTReferenceTraits.
 *
 * @code
 * Reference<MyObject> r1, r2;
 *
 * r1 = new MyObject();     // bind reference to new shared object
 * r2 = r1;                 // make another reference to shared object
 * r1 = NULL;               // remove first reference to shared object
 * r2 = NULL;               // remove second reference to shared object
 *                          // and since no other reference points to the object
 *                          // destroy it
 * @endcode
 *
 * You do not have to set references to NULL when you are leaving a function
 * or destroying object that references another object because
 * it is done automatically.
 *
 * @code
 * void f( Reference<MyObject> r )
 *      {
 *      // do some computation on object referenced by r
 *      // ...
 *      r = NULL;           // this is not neccessary
 *      }
 *
 * void g()
 *      {
 *      Reference<MyObject> r = new MyObject();
 *      f( r );
 *      r = NULL;           // this is not neccessary
 *      }
 * @endcode
 *
 * Notes about implementation:
 *
 * - this simple garbage collector is based on
 *   instant reference counting and thus can not
 *   handle cycles in references properly;
 *   this results in memory leaks (objects get not destroyed):
 * - user must ensure that cycles will not occur
 * - because of implementation, the compiler can not always
 *   perform automatic type casting, use reference_cast<> when needed.
 */
template <class Type, class Traits = ReferenceTraits>
class Reference
    {
private:

    typedef Reference<Type, Traits> ThisReference;

public:

    //
    // Constructors & destructors.
    //

    /** Constructor. */
    Reference
        (
        Type * ptr = NULL
        )
        {
        native_ptr = ptr;
        Traits::add_reference( native_ptr );
        }

    /** Copy constructor. */
    Reference
        (
        const ThisReference & r
        )
        {
        native_ptr = r.native_ptr;
        Traits::add_reference( native_ptr );
        }

    /** Destructor. */
    ~Reference()
        {
        Traits::remove_reference( native_ptr );
        }

public:

    //
    // Assigning.
    //

    /** Operator=. */
    ThisReference & operator=
        (
        const Type * ptr
        )
        {
        Traits::change_reference( ptr, native_ptr );
        native_ptr = const_cast<Type *>( ptr );

        return *this;
        }

    /** Operator=. */
    ThisReference & operator=
        (
        const ThisReference & r
        )
        {
        return operator=( r.native_ptr );
        }

public:

    //
    // Dereferencing.
    //

    /** Operator*. */
    Type & operator*() const
        {
        hope( native_ptr );
        return *native_ptr;
        }

    /** Operator->. */
    Type * operator->() const
        {
        hope( native_ptr );
        return native_ptr;
        }
    
    /** Returns native pointer. */
    Type * get() const
        {
        return native_ptr;
        }

public:

    //
    // State.
    //

    /** Sets to NULL. */
    void set_null()
        {
        Traits::remove_reference( native_ptr );
        native_ptr = NULL;
        }

    /** Checks if is NULL. */
    bool is_null() const
        {
        return native_ptr == NULL;
        }

    /**
     * Swaps pointer values.
     * Corresponds to @c std::swap( *this, r ).
     */
    void swap
        (
        ThisReference & r
        )
        {
        Type * old_ptr;

        old_ptr = native_ptr;
        native_ptr = r.native_ptr;
        r.native_ptr = old_ptr;
        }

public:

    //
    // Comparing.
    //

    /** Compares pointers. */
    friend bool operator==
        (
        const ThisReference &   r,
        const Type *            ptr
        )
        {
        return r.native_ptr == ptr;
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const ThisReference &   r,
        const Type *            ptr
        )
        {
        return !operator==( r, ptr );
        }
    
    /** Compares pointers. */
    friend bool operator==
        (
        const Type *            ptr,
        const ThisReference &   r
        )
        {
        return r.native_ptr == ptr;
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const Type *            ptr,
        const ThisReference &   r
        )
        {
        return !operator==( ptr, r );
        }
    
    /** Compares pointers. */
    friend bool operator==
        (
        const ThisReference &   r1,
        const ThisReference &   r2
        )
        {
        return r1.native_ptr == r2.native_ptr;
        }

    /** Compares pointers. */
    friend bool operator!=
        (
        const ThisReference &   r1,
        const ThisReference &   r2
        )
        {
        return !operator==( r1, r2 );
        }

public:

    /**
     * Hash functor.
     */
    struct Hash
        {
        size_t operator()
            (
            ThisReference & ptr
            )
            {
            return reinterpret_cast<size_t>( ptr.native_ptr );
            }
        }; // Hash

private:

    Type * native_ptr;
        /**< Native pointer. */

    }; // Reference


/**
 * Reference casting.
 * Use reference_cast<Type>( r ) to cast reference @a r to type @a Type.
 * SourceType must be derived from SharedObject.
 * In debug mode we verify if conversion is valid.
 */
template
    <
    class TargetType,
    class SourceType, 
    class Traits
    >
inline Reference<TargetType, Traits> reference_cast
    (
    const Reference<SourceType, Traits> & cast_from
    )
    {
    return Reference<TargetType, Traits>
        (
        checked_cast<TargetType *>( cast_from.operator->() )
        );
    }

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_REFERENCE_H
