/** @file text_reader.cpp
 * Simple line-oriented text file parser.
 */

/*
 * begin        : Fri Apr 12 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/text_reader.h"
#include "utility/quote_string.h"


namespace Massiv {
namespace Core {


TextReader::TextReader
    (
    std::istream & stream
    ) : stream( stream ),
        current_line_number( 0 ),
        line_offset( 0 )
    {
    }


std::string TextReader::str() const
    {
    return line_string;
    }


int TextReader::line_number() const
    {
    return current_line_number;
    }


bool TextReader::next_line()
    {
    line_offset = stream.tellg();

    while( std::getline( stream, line_string ) )
        {
        current_line_number++;
//#ifdef SYSTEM_UNIX
        // FIXME Haf: kdyz ctu na unixu textovy soubor, ktery ma 
        // ukoncenou radku dosovym new-linem, tak v promenne
        // line_string je na konci znak '\r', coz je problem.
        if( line_string.size() )
            {
            std::string::iterator it = line_string.end() - 1;
            if( *it == '\r' )
                {
                line_string.erase( it );
                }
            }
//#endif
        std::string::size_type pos = line_string.find_first_not_of( " \t" );
        
        if( pos != std::string::npos && line_string[ pos ] != '#' )
            {
            line_stream.str( line_string );
            line_stream.clear();
            line_stream.seekg( 0 );
            line_stream.unsetf( std::ios_base::basefield );
            
            return true;
            }

        line_offset = stream.tellg();
        }

    return false;
    }


void TextReader::next_line_throw()
    {
    if( !next_line() )
        {
        throw ExUnexpectedEof( unexpected_eof() );
        }
    }


size_t TextReader::current_position()
    {
    return line_offset + line().tellg();
    }


void TextReader::require_string
    (
    const std::string & expected
    )
    {
    std::string s = read_string();
    if( s != expected )
        {
        throw ExUnexpectedValue( unexpected_value( expected, s ) );
        }
    }

    
std::string TextReader::read_string()
    {
    std::string s;

    line() >> s;
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "string" ) );
            }
        }

    return s;
    }


std::string TextReader::read_quote_string()
    {
    std::string s;

    line() >> quote_string( s );
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "quoted string" ) );
            }
        }

    return s;
    }


int TextReader::read_int()
    {
    int i;

    line() >> i;
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "int" ) );
            }
        }

    return i;
    }


Int64 TextReader::read_int64()
    {
    Int64 i;

    line() >> i;
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "int" ) );
            }
        }

    return i;
    }


Float32 TextReader::read_float()
    {
    Float32 f;

    line() >> f;
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "float" ) );
            }
        }

    return f;
    }


Float64 TextReader::read_float64()
    {
    Float64 f;

    line() >> f;
    if( line().fail() )
        {
        if( line().eof() )
            {
            throw ExUnexpectedEol( unexpected_eol() );
            }
        else
            {
            throw ExUnexpectedToken( unexpected_token( "float" ) );
            }
        }

    return f;
    }


std::string TextReader::unexpected_eol()
    {
    std::stringstream   ss;
    ss << "Unexpected end of line. " <<
       "Line " << line_number() << ": " << str();
    return ss.str();
    }


std::string TextReader::unexpected_eof
    (
    const std::string &     fileid
    )
    {
    std::stringstream   ss;
    ss << "Unexpected end of file " << fileid << ".\n";
    return ss.str();
    }


std::string TextReader::unexpected_token
    (
    const std::string &     expected
    )
    {
    std::stringstream   ss;
    ss << "Unexpected token. Expected " << expected << ". " <<
       "Line " << line_number() << ", column " << line().tellg() << ": " << str();
    return ss.str();
    }


// Instantiate @c TextReader::unexpected_value() methods for "common" argument
// types so that they could be used outside the core.
// This must be done because TextReader is MASSIV_EXPORTED.
// Under MSVC6 explicit class template method instantiation does not work,
// so it must be hacked this way:

#ifdef COMPILER_MSVC

namespace {

void force_instantiation
    (
    TextReader &        tr,
    const Int32 &       i,
    const Float32 &     f,
    const std::string & s
    )
    {
    tr.unexpected_value( s, i );
    tr.unexpected_value( s, f );
    tr.unexpected_value( s, s );
    }

} // anonymous namespace

#else

template MASSIV_EXPORTED std::string TextReader::unexpected_value
    ( 
    const std::string &,
    const Int32 &
    );

template MASSIV_EXPORTED std::string TextReader::unexpected_value
    ( 
    const std::string &,
    const Float32 &
    );

template MASSIV_EXPORTED std::string TextReader::unexpected_value
    ( 
    const std::string &,
    const std::string &
    );

#endif


} // namespace Massiv::Core
} // namespace Massiv
