/** @file text_reader.h
 * Simple line-oriented text file parser.
 */

/*
 * begin        : Fri Apr 12 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_TEXT_READER_H
#define MASSIV_TEXT_READER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#ifndef MASSIV_EXCEPTION_H
#include "status/exception.h"
#endif

#include <iostream>
#include <sstream>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class TextReader
 * Line-oriented text file parser.
 * TextReader is a simple wrapper around istream classes. It automatically skips
 * empty lines and lines containing comments (beginning with a '#' character).
 *
 * @ingroup serialization
 */
class MASSIV_EXPORTED TextReader
    {
public:

    /**
     * Initialize TextReader.
     * @param stream Stream to read from.
     * @todo Always set exceptions( badbit ) here?
     *
     * @note First line of the stream is not automatically read.
     *  You must call next_line() by yourself.
     */
    TextReader
        (
        std::istream & stream
        );

public:

    /**
     * Get reference to line input stream.
     * You should not keep this reference after call to next_line()
     * (it would work now, but may break in the future).
     */
    inline std::istream & line()
        {
        return line_stream;
        }

    /**
     * Return current line as string.
     */
    std::string str() const;

    /**
     * Get current line number, counting from 1.
     */
    int line_number() const;

    /**
     * Read in next line.
     * Lines containing only whitespace or beginning with a '#' character are skipped.
     * @return true on success, false if eof has been reached. Leaves eofbit set.
     */
    bool next_line();

    /**
     * Read in next line.
     * Lines containing only whitespace or beginning with a '#' character are skipped.
     *
     * @exception ExUnexpectedEof if next line is not available.
     */
    void next_line_throw();

    /**
     * Returns current position within the stream.
     */
    size_t current_position();

public:

    /**
     * Read string from current line and check if it's the @a expected string.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type std::string can't be read.
     * @exception ExUnexpectedValue if string is not equal to @a expected.
     */
    void require_string
        (
        const std::string & expected
        );

    /**
     * Read string from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type @c std::string can't be read.
     */
    std::string read_string();

    /**
     * Read quoted string from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type quoted string can't be read.
     */
    std::string read_quote_string();

    /**
     * Read integer from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type @c int can't be read.
     */
    int read_int();

    /**
     * Read 64bit integer from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type @c int can't be read.
     */
    Int64 read_int64();

    /**
     * Read float from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type @c int can't be read.
     */
    Float32 read_float();

    /**
     * Read double-precision float from current line.
     *
     * @exception ExUnexpectedEol if end-of-line is reached.
     * @exception ExUnexpectedToken if value of type @c int can't be read.
     */
    Float64 read_float64();

public:
    // Note: These methods are not const, because they may use the line stream.
    // Const references to TextReader are pretty useless, so this should
    // not be a problem.

    /**
     * Create message for "unexpected end of line" exception.
     */
    std::string unexpected_eol();

    /**
     * Create message for "unexpected end of file" exception.
     * @param fileid Either empty string or string starting with a space.
     */
    std::string unexpected_eof
        (
        const std::string &     fileid = std::string()
        );

    /**
     * Create message for "unexpected token" exception.
     */
    std::string unexpected_token
        (
        const std::string &     expected
        );

    /**
     * Create message for "unexpected value" exception.
     */
    template< typename T >
    std::string unexpected_value
        (
        const std::string &     expected,
        const T &               found
        )
        {
        std::stringstream   ss;
        ss << "Unexpected value. Expected " << expected << ", found " << found << ".\n" <<
            "Line " << line_number() << ", column " << line().tellg() << ": " << str();
        return ss.str();
        }
    
private:

    std::istream &      stream;
        /**< Input stream. */

    std::istringstream  line_stream;
        /**< Line stream. */

    std::string         line_string;
        /**< Line string.
             @note line_stream.str() behaves really weird in gcc-2.95, so we keep its copy. */

    int                 current_line_number;
        /**< Current line number. */
    
    size_t              line_offset;
        /**< Current line offset within @c stream. */

    }; // class TextReader

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_TEXT_READER_H
