/** @file text_writer.h
 * Simple line-oriented text file writer.
 */

/*
 * begin        : Fri Apr 12 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_TEXT_WRITER_H
#define MASSIV_TEXT_WRITER_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <iostream>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class TextWriter
 * Line-oriented text file writer.
 * TextWriter is a wrapper around ostream classes.
 * Just a simple indentation stuff (spaces, no tabs).
 * @todo Remember state (begin line, end line), check weird usage.
 *
 * @ingroup serialization
 */
class MASSIV_EXPORTED TextWriter
    {
public:

    /**
     * Initialize TextWriter.
     * @param stream Stream to write to.
     */
    TextWriter
        (
        std::ostream & stream
        );

public:

    /**
     * Get current line stream.
     *
     * Writes indentation spaces if this is first call to #line() since
     * #end_line() call.
     *
     * @return Stream to write to.
     */
    std::ostream & line();

    /**
     * End current line.
     *
     * Stream returned by #line() becomes invalid.
     *
     * You may call #end_line() mutiple times to write empty lines with no
     * indentation whitespace.
     */
    void end_line();

public:

    /**
     * Get current line number, counting from 1. Increased when by #end_line().
     */
    int line_number();

    /**
     * Is current line "empty"?
     *
     * @b iff #line() has not been called for current line, even if nothing
     * has been written to the stream.
     */
    bool is_line_empty();

public:

    /**
     * Change indentation.
     * @param delta     Indent (positive delta) or unindent (negative delta).
     * @return          Current indentation.
     */
    int change_indent
        (
        int delta
        );

    /**
     * Set indentation.
     * @param indent    Indenatation level to set. Must be >= 0.
     */
    void set_indent
        (
        int indent
        );

    /**
     * Write whitespace to stream.
     */
    void write_space
        (
        int count = 1
        );

private:

    std::ostream &      stream;
        /**< Output stream. */

    int                 current_line_number;
        /**< Current line number. */

    int                 indentation;
        /**< Indentation. Always >= 0. */

    bool                indented;
        /**< Has indentation been written to current line? */

    }; // class TextWriter

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_TEXT_WRITER_H
