/** @file variant_pointer.h
 * Variant pointer.
 */

/*
 * begin        : Sun Dec 15 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VARIANT_POINTER_H
#define MASSIV_VARIANT_POINTER_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#include <cstdlib>
#include <typeinfo>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/** @class VariantPointer
 * Pointer whose type can change during run-time.
 * This is a replacement for dangerous void *.
 * One often needs to pass a pointer to some object,
 * whose interface should accept ANY pointer.
 * Most commonly @c void * is passed.
 * This is not a good approach because conversion
 * to void * causes the pointer to lose type information
 * that may be needed during its reconstruction or conversion
 * back to some type.
 *@code
 *
 *  class B : public A { };
 *
 *  A *     ap1, ap2;
 *  B *     bp;
 *  void *  ptr;
 *
 *  ptr = bp;   // Lose of type information
 *  ap1 = reinterpret_cast<A *>( ptr ); // No pointer value change made
 *  ap2 = static_cast<A *>( bp );
 *  hope( ap1 == ap2 );   // This fails because ap1 != ap2
 *
 *@endcode
 *
 * We provide a special class that can store pointer of ANY
 * type and which knows what type it currently is.
 * To variant_cast the stored value use template variant_cast<>().
 * Please note that this class does NOT perform any
 * pointer conversions and is only a restricted wrapper for
 * void *. See restrictions for variant_cast<>().
 *
 *@code
 *
 *  VariantPointer  ptr;
 *  A *             a_ptr;
 *  B *             b_ptr;
 *
 *  ptr = a_ptr;
 *  a_ptr = variant_cast<A *>( ptr );   // Ok, ptr holds A *
 *  ptr = b_ptr;
 *  a_ptr = variant_cast<A *>( ptr );   // Error, ptr holds B *
 *                                      // so variant_cast<B *> is allowed only
 *@endcode
 */
class VariantPointer
    {
public:

    struct Hash
        {

        std::size_t operator()
            (
            const VariantPointer & ptr
            ) const
            {
            return reinterpret_cast< std::size_t >( ptr.pointer );
            }

        }; // Hash

public:

    /**
     * Default constructor.
     */
    VariantPointer() :
        pointer( NULL )
        {
        #ifdef OPTION_DEBUG

        type = NULL;

        #endif
        }

    /**
     * Constructor.
     * Constructs a NULL pointer of type @c type.
     */
    VariantPointer
        (
        const std::type_info & type
        ) : pointer( NULL )
        {
        #ifdef OPTION_DEBUG

        this->type = &type;

        #endif
        }

    /**
     * Constructor.
     */
    template <class T>
    explicit VariantPointer
        (
        T * pointer
        ) : pointer( pointer )
        {
        #ifdef OPTION_DEBUG

        type = &typeid( T * );

        #endif
        }

    /**
     * Constructor.
     * Constructs VariantPointer pointing to @c pointer.
     * Assumes that @c pointer is of type @c type.
     */
    VariantPointer
        (
        void *                  pointer,
        const std::type_info &  type
        ) : pointer( pointer )
        {
        #ifdef OPTION_DEBUG

        this->type = &type;

        #endif
        }

    /**
     * Copy constructor.
     */
    VariantPointer
        (
        const VariantPointer & ptr
        ) : pointer( ptr.pointer )
        {
        #ifdef OPTION_DEBUG

        type = ptr.type;

        #endif
        }

    /**
     * Operator=.
     */
    VariantPointer & operator=
        (
        const VariantPointer & ptr
        )
        {
        pointer = ptr.pointer;

        #ifdef OPTION_DEBUG

        type = ptr.type;

        #endif

        return *this;
        }

    /**
     * Operator=.
     */
    template <class T>
    VariantPointer & operator=
        (
        T * ptr
        )
        {
        pointer = ptr;

        #ifdef OPTION_DEBUG

        type = &typeid( T * );

        #endif

        return *this;
        }

    /**
     * Operator==.
     */
    bool operator==
        (
        const VariantPointer & ptr
        ) const
        {
        #ifdef OPTION_DEBUG

        if( type != ptr.type )
            return false;

        #endif

        return pointer == ptr.pointer;
        }

    #ifdef OPTION_DEBUG

    /**
     * Returns current type of the pointer.
     * Debug mode only.
     */
    const std::type_info & get_type_info() const
        {
        hope( type );
        return *type;
        }

    #endif

    /**
     * Checks if is null.
     */
    bool is_null() const
        {
        return pointer == NULL;
        }

    /**
     * Sets to null. Preserves pointer type.
     */
    void set_null()
        {
        pointer = NULL;
        }

public:

    void * pointer;
        /**< Pointer value. */

    #ifdef OPTION_DEBUG

    const std::type_info * type;
        /**< Current pointer type. */

    #endif

    }; // VariantPointer


/**
 * Reads value of the variant pointer.
 * @param T must match type of pointer that is
 *        currently stored in the variant pointer.
 *        No pointer conversions are performed.
 */
template <class T>
T variant_cast
    (
    const VariantPointer & ptr
    )
    {
    #ifdef OPTION_DEBUG

    hope( ptr.type );
    hope( typeid( T ) == *ptr.type );

    #endif

    return reinterpret_cast<T>( ptr.pointer );
    }

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_VARIANT_POINTER_H
