/** @file virtual_time.h
 * Virtual time template.
 */

/*
 * begin        : Sun Dec 01 2001
 * copyright    : (C) 2001 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_VIRTUAL_TIME_H
#define MASSIV_VIRTUAL_TIME_H


#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

#include <ostream>


namespace Massiv {
namespace Core {


/** @addtogroup utility
 * @{
 */

/**
 * Virtual time implementation.
 *
 * Virtual time is a value that is gradually increased during the
 * simulation, but has nothing in common with the real time.
 *
 * Serves as a base class for @c SystemTime.
 *
 * @todo UInt32 or Int64? Beware of wrap-around.
 *
 * @ingroup time
 */
template <class T = UInt32>
class VirtualTime
    {
public:

    /**
     * Type used to represent the time.
     */
    typedef T TimeType;

    enum
        {
        INVALID = 0
            /**< The oldest time possible is considered invalid. */
        };

public:

    /**
     * Virtual time Less comparator.
     */
    struct Less
        {
    public:

        /**
         * Virtual time compare function.
         */
        bool operator()
            (
            const VirtualTime<T> & t1,
            const VirtualTime<T> & t2
            ) const
            {
            return t1.is_older_than( t2 );
            }

        }; // Less

    /**
     * Virtual time @c EqualTo comparator.
     */
    struct EqualTo
        {
    public:

        /**
         * Virtual time compare function.
         */
        bool operator()
            (
            const VirtualTime<T> & t1,
            const VirtualTime<T> & t2
            ) const
            {
            return t1.is_equal_to( t2 );
            }

        }; // EqualTo

public:

    /**
     * Default constructor.
     */
    VirtualTime();

    /**
     * Constructor.
     * @param time The time value to initialize this instance.
     */
    VirtualTime
        (
        const T & time
        );

    /**
     * Copy constructor.
     */
    VirtualTime
        (
        const VirtualTime & t
        );

    /**
     * Operator=.
     */
    VirtualTime & operator=
        (
        const VirtualTime & t
        );

    /**
     * Operator +=.
     */
    VirtualTime & operator+=
        (
        const VirtualTime & t
        );

    /**
     * Operator -=.
     */
    VirtualTime & operator-=
        (
        const VirtualTime & t
        );

    /**
     * Binary operator+.
     */
    VirtualTime operator+
        (
        const VirtualTime & t
        ) const;

    /**
     * Binary operator-.
     */
    VirtualTime operator-
        (
        const VirtualTime & t
        ) const;

    /** @name Comparison operators */
    //@{

    /**
     * Operator==.
     * @return true if both times are equal.
     */
    bool operator==
        (
        const VirtualTime & t
        ) const;

    /**
     * Operator!=.
     * @return true if both times are not equal.
     */
    bool operator!=
        (
        const VirtualTime & t
        ) const;

    /**
     * Binary operator<.
     * 
     * @return true if the time given by this instance is older than @a t.
     */
    bool operator<
        (
        const VirtualTime & t
        ) const;

    /**
     * Binary operator>.
     *
     * @return true if @a t is older than the time given by this instance time.
     */
    bool operator>
        (
        const VirtualTime & t
        ) const;

    /**
     * Binary operator<=.
     *
     * @return true if the time given by this instance is older or equal 
     *      than @a t.
     */
    bool operator<=
        (
        const VirtualTime & t
        ) const;

    /**
     * Binary operator>=.
     *
     * @return true if @a t is older or equal than the time given by this 
     *      instance time.
     */
    bool operator>=
        (
        const VirtualTime & t
        ) const;

    //@}

public:

    /**
     * Advances virtual time by unit time.
     */
    void advance();

    /**
     * Resets the time to invalid value.
     */
    void invalidate();

    /**
     * Compares two virtual times.
     *
     * @param t Time to compare this instance with.
     * @return true if both times are equal, false otherwise.
     */
    bool is_equal_to
        (
        const VirtualTime & t
        ) const;

    /**
     * Checks whether this time is older than @a t.
     * 
     * @return true if the time specified by this instance is 'older' than
     *      @a t; false otherwise.
     *
     * @note @c INVALID time is the oldest time ever.
     */
    bool is_older_than
        (
        const VirtualTime & t
        ) const;

    /**
     * Checks validity of this time.
     *
     * @return true if this instance of time is invalid.
     */
    bool is_invalid() const;

    /**
     * Checks validity of this time.
     * 
     * @return true if this instance of time is valid.
     */
    bool is_valid() const;

public:

    /**
     * Get internal value of time.
     *
     * @return The internal time value.
     * @note Intended mainly for serializers usage.
     * @note Synonymous to @c read().
     */
    const T & get_value() const;

    /**
     * Get internal value of time.
     *
     * @return The internal time value.
     * @note Intended mainly for serializers usage.
     * @note Synonymous to @c get_value().
     */
    const T & read() const;

protected:

    /**
     * Set internal value of time.
     * 
     * @param t The new time value.
     * @note Intended mainly for serializers usage.
     */
    void set_value
        (
        const T & t
        );

private:

    T time;
        /**< Time value. */

    }; // VirtualTime


#define VT  VirtualTime<T>

template <class T>
inline VT::VirtualTime()
    {
    invalidate();
    }

template <class T>
inline VT::VirtualTime
    (
    const T & time
    ) : time( time )
    {
    }

template <class T>
inline VT::VirtualTime
    (
    const VT & t
    ) : time( t.time )
    {
    }

template <class T>
inline VT & VT::operator=
    (
    const VT & t
    )
    {
    time = t.time;
    return *this;
    }

template <class T>
inline VT & VT::operator+=
    (
    const VT & t
    )
    {
    time += t.time;
    return *this;
    }

template <class T>
inline VT & VT::operator-=
    (
    const VT & t
    )
    {
    if( t.is_older_than( *this ) )
        {
        time -= t.time;
        }
    else
        {
        invalidate();
        }
    return *this;
    }

template <class T>
inline VT VT::operator+
    (
    const VT & t
    ) const
    {
    VT  rt( *this );
    rt += t;
    return rt;
    }

template <class T>
inline VT VT::operator-
    (
    const VT & t
    ) const
    {
    VT rt( *this );
    rt -= t;
    return rt;
    }

template <class T>
inline bool VT::operator==
    (
    const VT & t
    ) const
    {
    return time == t.time;
    }

template <class T>
inline bool VT::operator!=
    (
    const VT & t
    ) const
    {
    return time != t.time;
    }

template <class T>
inline bool VT::operator<
    (
    const VT & t
    ) const
    {
    return time < t.time;
    }

template <class T>
inline bool VT::operator>
    (
    const VT & t
    ) const
    {
    return time > t.time;
    }

template <class T>
inline bool VT::operator<=
    (
    const VT & t
    ) const
    {
    return time <= t.time;
    }

template <class T>
inline bool VT::operator>=
    (
    const VT & t
    ) const
    {
    return time >= t.time;
    }

template <class T>
inline void VT::advance()
    {
    time++;
    }

template <class T>
inline void VT::invalidate()
    {
    time = INVALID;
    }

template <class T>
inline bool VT::is_equal_to
    (
    const VT & t
    ) const
    {
    return time == t.time;
    }

template <class T>
inline bool VT::is_older_than
    (
    const VT & t
    ) const
    {
    return time < t.time;
    }

template <class T>
inline bool VT::is_invalid() const
    {
    return time == INVALID;
    }

template <class T>
inline bool VT::is_valid() const
    {
    return !is_invalid();
    }

template <class T>
std::ostream & operator<<
    (
    std::ostream &          stream,
    const VirtualTime<T> &  t
    )
    {
    static VirtualTime<T>   invalid;
    
    if( t == invalid )
        {
        stream << "[ invalid ]";
        }
    else
        {
        stream << "[ " << t.get_value() << " ]";
        }
    
    return stream;
    }

template <class T>
inline const T & VT::get_value() const
    {
    return time;
    }

template <class T>
inline const T & VT::read() const
    {
    return get_value();
    }

template <class T>
inline void VT::set_value
    (
    const T & t
    )
    {
    time = t;
    }


#undef VT

/** @} */ // addtogroup utility


} // namespace Massiv::Core
} // namespace Massiv


#endif // MASSIV_VIRTUAL_TIME_H
