/** @file client_global.cpp
 * Global client stuff.
 */

/*
 * begin        : Sun Mar 9 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core_base.h"
#include "client_global.h"
#include "tag_database.h"
#include "input/input_manager.h"
#include "input/input_layers.h"
#include "gui/client_full_console.h"
#include "gui/client_chat_console.h"
#include "gui/game_interface.h"
#include "gui/editor.h"
#include "gui/terrain_window.h"
#include "media_repository.h"
#include "game/game_context.h"
#include "game/sector_manager.h"
#include "game/entity_manager.h"
#include "game/player_manager.h"
#include "game/game_input.h"
#include "sdl/client_sdl.h"
#include "renderer/renderer.h"
#include "renderer/render_manager.h"
#include "renderer/pass_manager.h"
#include "renderer/font.h"
#include "renderer/icon.h"
#include "renderer/text_rectangle.h"
#include "renderer/camera.h"
#include "renderer/mesh_entity.h" // hack
#include "renderer/render_objects.h" // hack
#include "renderer/layers.h"
#include "gui/widget_input_dispatcher.h"
#include <memory> // auto_ptr


namespace Demo {
namespace Client {


Global::Objects *    Global::objects = NULL;

// Settings
bool                 Global::rendering_disabled = false;
UInt32               Global::chat_console_height;
UInt32               Global::window_width;
UInt32               Global::window_height;
UInt32               Global::window_bpp;
UInt32               Global::account_timeout;


void Global::initialize
    (
    UInt32 the_window_width,
    UInt32 the_window_height,
    UInt32 the_window_bpp,
    UInt32 the_chat_console_height,
    UInt32 the_account_timeout
    )
    {
    chat_console_height = the_chat_console_height;
    window_bpp = the_window_bpp;
    window_height = the_window_height;
    window_width = the_window_width;
    account_timeout = the_account_timeout;

    try
        {
        if( !SDL::is_sdl_initialized() )
            {
            SDL::initialize();
            }
        if( !SDL::is_video_mode_opened() )
            {
            SDL::open_video_mode( window_width, window_height, window_bpp );
            }

        // Initialize the objects.

        objects = new Objects;
        objects->initialize();
        }
    catch( ... )
        {
        shutdown();
        throw;
        }
    }


void Global::shutdown
    (
    bool reset_sdl
    )
    {
    if( objects != NULL )
        {
        delete objects;
        objects = NULL;
        }
    if( reset_sdl )
        {
        SDL::close_video_mode();
        SDL::shutdown();
        }
    }


void Global::disable_rendering()
    {
    if( !rendering_disabled )
        {
        rendering_disabled = true;

        // Note: Current renderer is not deinitialized.
        // In the future we may want to shutdown the renderer and close the application
        // main window, but that would require the whole GUI to be able to shutdown
        // and initialize without affect the functionality - lot of hacks in game managers,
        // consoles and inventory.
        }
    }


void Global::enable_rendering()
    {
    if( rendering_disabled )
        {
        // FIXME: initialize video and renderer

        rendering_disabled = false;
        }
    }


Renderer::TextureManager & Global::get_texture_manager()
    {
    hope( objects );
    hope( objects->renderer.get() );
    return Renderer::Global::get_texture_manager();
    }


void Global::set_message
    (
    const std::string & text
    )
    {
    hope( objects );
    hope( objects->message.get() );
    Renderer::TextRectangle & msg = *objects->message;

    msg.print_line( text );
    msg.set_base_offset( SVector2
        (
        ( SDL::get_screen_width() - msg.get_base_size().x() ) / 2,
        ( SDL::get_screen_height() - msg.get_base_size().y() ) / 2
        ) );
    }


Global::Objects::Objects()
    {
    // Note: Wa can't initialize media repository in this constructor
    // because Global::objects is NULL, so other objects are not able
    // to use the repository anywhere in this constructor.

    renderer.reset( new Renderer::Global );
    media_repository.reset( new MediaRepository() );
    console_font.reset( new Renderer::Font( "font/console" ) );
    message.reset( new Renderer::TextRectangle( console_font.get() ) );
    background.reset( new Renderer::Icon( "image/background" ) );
    cursor.reset( new Renderer::Icon( "image/cursor" ) );
    full_console.reset( new FullConsole );
    chat_console.reset( new ChatConsole );
    console_switcher.reset( new Console::ConsoleSwitcher );
    model_database.reset( new TagDatabase( "model_database" ) );
    player_manager.reset( new PlayerManager );
    sector_manager.reset( new SectorManager );
    entity_manager.reset( new EntityManager );
    game_input.reset( new GameInput );
    intersection_entity.reset( new Renderer::MeshEntity );
    input_manager.reset( new InputManager );
    widget_input_dispatcher.reset( new WidgetInputDispatcher() );
    terrain_window.reset( new TerrainWindow() );
    game_interface.reset( new GameInterface() );
    editor.reset( new Editor() );
    game_context.reset( new GameContext() );
    }


Global::Objects::~Objects()
    {
    game_context.reset();
    media_repository.reset();
    editor.reset();
    game_interface.reset();
    terrain_window.reset();
    widget_input_dispatcher.reset();
    input_manager.reset();
    game_input.reset();
    intersection_entity.reset();
    entity_manager.reset();
    sector_manager.reset();
    player_manager.reset();
    model_database.reset();
    console_switcher.reset();
    full_console.reset();
    chat_console.reset();
    cursor.reset();
    background.reset();
    message.reset();
    console_font.reset();
    renderer.reset();
    }


void Global::Objects::initialize()
    {
    // MediaRepository.

    hope( media_repository.get() );
    media_repository->initialize();

    // Generic message in topmost pass.

    hope( message.get() );
    message->add_animation_object( new Renderer::OverlayColorFactorAnimation );
    message->register_to_global_manager( Renderer::PASS_TOPMOST, Renderer::LAYER_TOPMOST );

    // Background wallpaper.

    hope( background.get() );
    background->set_base_size( SVector2
        (
        static_cast< Float32 >( SDL::get_screen_width() ),
        static_cast< Float32 >( SDL::get_screen_height() )
        ) );
    background->set_base_color( SColor( 0.6f, 0.6f, 0.6f, 1.0f ) );
    //background->register_to_global_manager( Renderer::PASS_BACKGROUND, Renderer::LAYER_WALLPAPER );

    // Cursor.

    hope( cursor.get() );
    cursor->add_animation_object( new Renderer::OverlayMousePositionAnimation );
    cursor->set_base_size( SVector2
        (
        16,
        16
        ) );
    cursor->register_to_global_manager( Renderer::PASS_TOPMOST, Renderer::LAYER_TOPMOST );

    // Widget input dispatcher.

    hope( widget_input_dispatcher.get() );
    widget_input_dispatcher->register_to_global_input_manager( InputLayers::LAYER_UI );
    widget_input_dispatcher->register_to_global_manager( Renderer::PASS_UI, Renderer::LAYER_UI );

    // Terrain window.

    hope( terrain_window.get() );
    terrain_window->initialize();

    // Console::ConsoleSwitcher
    hope( console_switcher.get() );
    console_switcher->initialize();
    console_switcher->register_to_global_input_manager( InputLayers::LAYER_CONSOLE );

    // FullConsole.

    hope( full_console.get() );
    full_console->initialize();
    full_console->set_dispatcher_as_parent();

    // ChatConsole.

    hope( chat_console.get() );
    chat_console->set_dispatcher_as_parent();
    chat_console->initialize();

    // Game managers.

    hope( player_manager.get() );
    player_manager->initialize();
    hope( sector_manager.get() );
    sector_manager->initialize();
    hope( entity_manager.get() );
    entity_manager->initialize();
    hope( game_input.get() );
    game_input->register_to_global_input_manager( InputLayers::LAYER_GAME );

    // Intersection entity. Hack.

    {//sob
        const Float32 height = 1.0f;
        const Float32 radius = 0.05f;
        const std::size_t vertex_count = 8;

        hope( intersection_entity.get() );

        // Create and initialize pyramid mesh.

        intersection_entity->allocate( 1 );
        Renderer::Mesh & mesh = intersection_entity->meshes[ 0 ];
        mesh.set_vertex_count( vertex_count + 1 );
        mesh.set_triangle_count( 2 * vertex_count - 2 );
        Renderer::Objects::initialize_vertex( mesh, 0, SVector3( 0.0f, 0.0f, height ) );
        Renderer::Objects::initialize_circle_vertices( mesh, 1, radius, 0.0f, vertex_count );
        Renderer::Objects::initialize_pyramid_triangles( mesh, 0, 0, vertex_count );

        // Initialize render objects.

        intersection_entity->initialize_render_objects();
        Renderer::PipelineState & pipeline_state =
                    intersection_entity->render_objects[ 0 ].pipeline_states[ 0 ];
        pipeline_state.color = SColor( 0.3f, 0.6f, 1.0f );
    }//eob
    intersection_entity->register_to_global_manager( Renderer::PASS_WORLD );

    // Game interface.

    hope( game_interface.get() );
    game_interface->initialize();

    // Editor.

    hope( editor.get() );
    editor->initialize();
    }


} // namespace Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "client_global_inline.h"
#endif
