/** @file client_global.h
 * Global client stuff.
 */

/*
 * begin        : Sun Mar 9 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef CLIENT_GLOBAL_H
#define CLIENT_GLOBAL_H

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef MASSIV_CORE_MANAGEMENT_H
#include "core_management.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_CONSOLE_H
#include "gui/client_console.h"
#endif


#include <memory>


namespace Demo {
namespace Client {


class FullConsole;
class ChatConsole;
class InputManager;
class GameInterface;
class MediaRepository;
class GameContext;
class GameInput;
class PlayerManager;
class SectorManager;
class EntityManager;
class TagDatabase;
class WidgetInputDispatcher;
class Editor;
class TerrainWindow;


/**
 * Client global stuff.
 *
 * @todo We can't add every interesting font, image, icon or text to Global.
 *       Implement nice repositiories/databases for these objects?
 */
class Global
    {
public:

    /**
     * Initialize.
     * @param window_width width of client window.
     * @param window_height height of client window.
     * @param window_bpp color depth of client window.
     * @param chat_console_height height of chat console in lines.
     * @param account_timeout Max allowed time to link with an account.
     * @Note also initialize global settings. Zero window_bpp means fullscreen.
     */
    static void initialize
        (
        UInt32 window_width,
        UInt32 window_height,
        UInt32 window_bpp,
        UInt32 chat_console_height,
        UInt32 account_timeout
        );

    /**
     * Shutdown.
     * @param reset_sdl wherher to reset video mode or not.
     */
    static void shutdown
        (
        bool reset_sdl = true
        );

    /**
     * Disable rendering.
     *
     * Accesing any renderer-related structures or the main window is
     * forbidden while the rendering is disabled.
     */
    static void disable_rendering();

    /**
     * Enable rendering.
     */
    static void enable_rendering();

// getters
public:

    /**
     * Is client initialized?
     */
    static INLINE bool is_initialized();

    //@{
    /**
     * Getter.
     * It's an error to call this function if the global object does not exist.
     */
    static INLINE Renderer::Global & get_renderer();

    static Renderer::TextureManager & get_texture_manager();

    static INLINE Renderer::Font & get_console_font();

    static INLINE Renderer::TextRectangle & get_message();

    static INLINE FullConsole & get_full_console();

    static INLINE ChatConsole & get_chat_console();

    static INLINE Console::ConsoleSwitcher & get_console_switcher();

    static INLINE TagDatabase & get_model_database();

    static INLINE PlayerManager & get_player_manager();

    static INLINE SectorManager & get_sector_manager();

    static INLINE EntityManager & get_entity_manager();

    static INLINE Renderer::MeshEntity & get_intersection_entity();

    static INLINE GameInput & get_game_input();

    static INLINE InputManager & get_input_manager();

    static INLINE GameInterface & get_game_interface();

    static INLINE MediaRepository & get_media_repository();

    static INLINE GameContext & get_game_context();

    static INLINE WidgetInputDispatcher & get_widget_input_dispatcher();

    static INLINE Editor & get_editor();

    static INLINE TerrainWindow & get_terrain_window();
    //@}

    /**
     * Update message.
     *
     * @todo Get rid of this.
     */
    static void set_message
        (
        const std::string & text
        );

private:

    /**
     * Global objects.
     */
    struct Objects
        {
        /** Allocate global objects. */
        Objects();

        /** Destroy global objects. */
        ~Objects();

        /** Initialize global objects. */
        void initialize();

    public:

        std::auto_ptr< Renderer::Global >           renderer;
            /**< The renderer. */

        std::auto_ptr< Renderer::Font >             console_font;
            /**< Simple console font.
                 I doubt we will have more fonts in the future, currently
                 every text object uses this font. */

        std::auto_ptr< Renderer::TextRectangle >    message;
            /**< Message to show when there is no console. Get rid of this. */

        std::auto_ptr< Renderer::Icon >             background;
            /**< Background picture. */

        std::auto_ptr< Renderer::Icon >             cursor;
            /**< Cursor picture. */

        std::auto_ptr< FullConsole >                full_console;
            /**< Global full console. */

        std::auto_ptr< ChatConsole >                chat_console;
            /**< Global chat console. */

        std::auto_ptr< Console::ConsoleSwitcher >   console_switcher;
            /**< Global console switcher. */

        std::auto_ptr< TagDatabase >                model_database;
            /**< Entity tags -> model data object name mappings. */

        std::auto_ptr< PlayerManager >              player_manager;
            /**< Player manager. */

        std::auto_ptr< SectorManager >              sector_manager;
            /**< Sector manager. */

        std::auto_ptr< EntityManager >              entity_manager;
            /**< Client entity manager. */

        std::auto_ptr< Renderer::MeshEntity >       intersection_entity;
            /**< Raycast "demo" hack. Remove this. */

        std::auto_ptr< GameInput >                  game_input;
            /**< Class that handles non-GUI keyboard input. */

        std::auto_ptr< InputManager >               input_manager;
            /**< Input (mouse and keyboard) manager. */

        std::auto_ptr< GameInterface >              game_interface;
            /**< Game interface. */

        std::auto_ptr< MediaRepository >            media_repository;
            /**< Media repository. */

        std::auto_ptr< GameContext >                game_context;
            /**< Game context. */

        std::auto_ptr< WidgetInputDispatcher >      widget_input_dispatcher;
            /**< Widget input dispatcher. */

        std::auto_ptr< Editor >                     editor;
            /**< Editor. */

        std::auto_ptr< TerrainWindow >              terrain_window;
            /**< Terrain window. */

        }; // struct Objects

    static Objects * objects;
        /**< Global objects.
             Stupid wrapper, we don't want to have static auto_ptrs. */

public:

    static bool                 rendering_disabled;
            /**< Is rendering disabled? */

    static UInt32               chat_console_height;
            /**< Chat console height ( lines ). */

    static UInt32               window_width;
            /**< Chat console height ( lines ). */

    static UInt32               window_height;
            /**< Chat console height ( lines ). */

    static UInt32               window_bpp;
            /**< Window bits per pixel. 0 for windowed mode. */

    static UInt32               account_timeout;
            /**< Max allowed time to link with an account (seconds). */

    }; // Global


} // namespace Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "client_global_inline.h"
#endif


#endif // CLIENT_GLOBAL_H
