/** @file client_main.cpp
 * Demo client.
 */

/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

// Core

#include "core.h"
#include "core_generated.h"
#include "cmdline/cmdline_parser.h"
#include "net/massiv_blowfish.h"
#include "net/massiv_checksum.h"
#include "net/massiv_checksum_impl.h"
#include "net/bitiostream_common.h"
#include "data/path_manager.h"
#include "data/account_manager.h"

// Client

#include "config.h"
#include "client_main.h"
#include "client_global.h"
#include "client_exception.h"
#include "gui/client_full_console.h"
#include "gui/client_chat_console.h"
#include "sdl/client_sdl.h"
#include "renderer/opengl.h"
#include "renderer/renderer.h"
#include "renderer/text_rectangle.h"
#include "renderer/mesh_entity.h" // hack
#include "renderer/terrain_settings.h"
#include "gui/login_dialog.h"
#include "gui/message_box.h"
#include "game/game_context.h"
#include "game/player_manager.h"
#include "game/sector_manager.h"
#include "game/entity_manager.h"
#include "game/game_input.h"
#include "gui/game_interface.h"
#include "gui/editor.h"
#include "gui/terrain_window.h"

// Lib

#include "lib/shared/client_node_interface.h"
#include "lib/client/class_list.h"
#include "lib/client/client_node.h"
#include "lib/shared/account_object_rpc.h"

// STD C++

#include <iosfwd>
#include <fstream>


namespace Demo {
namespace Client {

INJECT_CORE_SYMBOLS();

using namespace Demo::Lib;
using namespace Massiv;


namespace {

std::auto_ptr< LoginDialog >    g_login_dialog;
    /**< Login dialog. Created and destroyed in #login_dialog_loop(). */

std::auto_ptr< MessageBox >     g_message_box;
    /**< Message box. Created and destroyed in #login_dialog_loop(). */

std::string                     g_local_node_config;
    /**< Local node config. HACK. */

bool next_tick(); // Forward declaration.

/**
 * System will call this to create NodeObject.
 */
void create_node_object_callback()
    {
    Pointer< Demo::Lib::ClientNode >   client_node;

    // Create our node object reresenting local none.
    // The node object will receive messages from simulation.
    std::cout << "Creating node object..." << std::endl;
    client_node = CreateObject< Demo::Lib::ClientNode >();
    std::cout << "**** ObjectId of my node object: " << client_node.get_object_id() << std::endl;
    }


/**
 * Next tick callback.
 */
void next_tick_callback()
    {
    if( g_login_dialog.get() && !g_message_box.get() )
        {
        g_login_dialog->next_tick();
        }

    if( g_message_box.get() )
        {
        const MessageBox::Result result = g_message_box->get_result();
        if( result != MessageBox::RESULT_NONE )
            {
            g_message_box.reset();
            if( g_login_dialog.get() )
                {
                g_login_dialog->set_focus();
                }
            }
        }

    if( next_tick() )
        {
        System::request_quit();
        }
    }


/**
 * Next tick callback, client not initialized yet.
 */
void simple_next_tick()
    {
    static System::Phase old_phase = System::PHASE_DISCONNECTING; // bogus value
    System::Phase system_phase = System::get_phase();

    if( system_phase != old_phase )
        {
        old_phase = system_phase;
        switch( system_phase )
            {
            case System::PHASE_INITIALIZED:
                std::cout << "Massiv core initialzied." << std::endl;
                break;

            case System::PHASE_DOWNLOADING_DATA:
                std::cout << "Downloading required data..." << std::endl;
                break;

            case System::PHASE_DATA_DOWNLOADED:
                std::cout << "Data downloaded." << std::endl;
                break;

            default:
                std::cout << "!Unexpected phase " << system_phase << "." << std::endl;
            }
        }
    }


/**
 * Returns true if the application should quit.
 */
bool next_tick()
    {
    if( !Global::is_initialized() )
        {
        // Still initializing massiv core. Limited functionality.

        simple_next_tick();
        return false;
        }

    if( SDL::process_events() )
        {
        return true;
        }

    if( Global::get_game_input().quit_requested() )
        {
        return true;
        }

    static System::Phase    old_phase = System::PHASE_INITIALIZED;
    System::Phase           system_phase = System::get_phase();

    static SystemTime       old_time;
    static bool             old_time_valid = false;
    const SystemTime &      current_time = System::system_time();
    if( !old_time_valid )
        {
        old_time = current_time;
        old_time_valid = true;
        }
    const SystemTime        time_delta = current_time - old_time;
    old_time = current_time ;

    // Limited work if rendering is disabled.

    if( Global::rendering_disabled )
        {
        if( system_phase == System::PHASE_RUNNING )
            {
            // Update consoles.

            Global::get_full_console().next_tick();
            Global::get_chat_console().next_tick();

            // Update managers.

            Global::get_player_manager().update();
            Global::get_sector_manager().update();
            Global::get_entity_manager().update();

            // Be really lazy.

            Thread::sleep( 250 );

            // Refresh the blackness.

            glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
            glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
            Renderer::Global::flip();
            }

        return false;
        }

    // Rendering is enabled - full next_tick update.

    if( ( system_phase == System::PHASE_RUNNING ) && Global::get_full_console().get_should_be_visible() )
        {
        Global::get_message().require_color_factor_animation().
            set_target_factor( SColor( 1.0f, 1.0f, 1.0f, 0.0f ) );
        }
    else
        {
        if( system_phase != old_phase )
            {
            switch( system_phase )
                {
                // FIXME: this always reprints the message, which is not necessary.
                case System::PHASE_UNINITIALIZED:
                case System::PHASE_INITIALIZED:
                case System::PHASE_DOWNLOADING_DATA:
                    panic( "Client should not be initialized yet." );
                    break;

                case System::PHASE_DATA_DOWNLOADED:
                    Global::set_message( "" ); // Login dialog is on screen.
                    break;

                case System::PHASE_CONNECTING:
                case System::PHASE_CONNECTED:
                    Global::set_message( "Connecting..." );
                    break;

                case System::PHASE_RUNNING:
                    Global::set_message( "" );
                    break;

                case System::PHASE_DISCONNECTING:
                    Global::set_message( "Disconnecting..." );
                    break;

                case System::PHASE_SHUTTING_DOWN:
                    Global::set_message( "Shutting down..." );
                    break;
                }
            old_phase = system_phase;
            }

        Global::get_message().require_color_factor_animation().
            set_target_factor( SColor( 1.0f, 1.0f, 1.0f, 1.0f ) );
        }

    // Clear. Hacky. Should this be performed by pass manager?

    Renderer::PipelineState clear_state;
    clear_state.buffer_mask.color = true;
    clear_state.buffer_mask.depth = true;
    Renderer::Global::apply_state( clear_state );

    const Renderer::TerrainSettings & terrain_settings = Renderer::Global::get_terrain_settings();
    if( terrain_settings.enable_fog )
        {
        // Clear framebuffer to fog color.

        glClearColor
            (
            terrain_settings.fog_red,
            terrain_settings.fog_green,
            terrain_settings.fog_blue,
            0.0f
            );
        }
    else
        {
        // FIXME: enable background picture in this case?

        glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
        }
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    if( system_phase == System::PHASE_RUNNING )
        {
        // Update consoles.

        Global::get_full_console().redraw();
        Global::get_full_console().next_tick();

        Global::get_chat_console().redraw();
        Global::get_chat_console().next_tick();

        // Update editor.

        Global::get_editor().next_tick();

        // Update managers.

        Global::get_player_manager().update();
        Global::get_sector_manager().update();
        Global::get_entity_manager().update();

        // Update entity lights. Reserve single light for "sun" and single light
        // for special effects.

        Global::get_entity_manager().update_lights( Renderer::MAX_LIGHT_SOURCES - 2 );

        // Update camera that tracks the player.

        Global::get_terrain_window().update_camera();

        // Raycast testing. (temporary hack)

        SectorManager::RayIntersection sector_intersection;
        EntityManager::RayIntersection entity_intersection;
        sector_intersection.layer_id = entity_intersection.layer_id =
                    Global::get_player_manager().get_visible_layer();
        Renderer::CirclingCamera & camera = Global::get_player_manager().get_camera();
        int mouse_x, mouse_y;
        Global::get_input_manager().get_mouse_position( mouse_x, mouse_y );
        camera.cursor_ray( sector_intersection.origin, sector_intersection.direction,
                    mouse_x, mouse_y );
        entity_intersection.origin = sector_intersection.origin;
        entity_intersection.direction = sector_intersection.direction;
        Renderer::MeshEntity & intersection_entity = Global::get_intersection_entity();
#if 1
        intersection_entity.set_visible( false );
#else // 1
        const bool sector_found = Global::get_sector_manager().cast_ray( sector_intersection );
        const bool entity_found = Global::get_entity_manager().cast_ray( entity_intersection );

        if( sector_found || entity_found )
            {
            SVector3 position;
            SVector3 normal;
            if( sector_found && ( !entity_found ||
                entity_intersection.distance > sector_intersection.distance ) )
                {
                position = sector_intersection.intersection;
                normal = sector_intersection.normal;
                }
            else
                {
                hope( entity_found );
                position = entity_intersection.intersection;
                normal = entity_intersection.normal;
                }

            SOrientation orientation;
            orientation.set_from_up_vector( normal );
            intersection_entity.set_orientation( orientation );
            intersection_entity.set_position( position );
            intersection_entity.set_visible( true );
            }
        else
            {
            intersection_entity.set_visible( false );
            }
#endif // 1

        }

    // Render frame.

    Renderer::Global::render( time_delta );

    // Show everything on screen.

    Renderer::Global::flip();

    // Done.

    return false;
    }


} // anonymous namespace


Main::Main()
      : settings( NULL )
    {
    }


Main::~Main()
    {
    if( settings )
        {
        delete settings;
        }
    }


//
// Callbacks from ClientNode object.
//

void Main::welcome_to_simulation
    (
    WeakPointer< AccountObject > account
    )
    {
    GameContext & game_context = Global::get_game_context();

    hope( account != null );
    hope( Global::is_initialized() );
    hope( !game_context.is_linked_with_account() );

    // Log.
    std::cout << "Connected to simulation through account object "
        << account.get_object_id() << ".\n"
        << "Welcome to simulation." << std::endl;

    // Store the pointer in GameContext.
    game_context.link_with_account( account );
    }


void Main::player_selected
    (
    WeakPointer< ClientEntity > player
    )
    {
    GameContext & game_context = Global::get_game_context();
    GameInterface & game_interface = Global::get_game_interface();
    PlayerManager & player_manager = Global::get_player_manager();
    Console::ConsoleSwitcher & console_switcher = Global::get_console_switcher();

    // Log.
    std::cout << "Selected player object " << player.get_object_id()
        << "." << std::endl;

    hope( Global::is_initialized() );
    hope( game_context.is_linked_with_account() || player == NULL );

    // Store the pointer in GameContext.
    game_context.player_selected( player );

    // Prepare for monitoring replicas of the player and request
    // replication of the player.
    game_interface.player_selected( player );
    player_manager.set_player( player );

    // Enable or disable chat mode.
    console_switcher.enable( player != null );
    }


void Main::initialize_client()
    {
    // Minimal required data was downloaded from service data node.
    // We are ready to initialize renderer, input and other client subsystems.

    // Initialize client.
    std::cout << "Initializing client..." << std::endl;
    hope( !Global::is_initialized() );
    Global::initialize
        (
        settings->window_width,
        settings->window_height,
        settings->window_bpp,
        settings->chat_console_height,
        settings->account_timeout
        );
    }


void Main::run
    (
    int             argc,
    const char **   argv
    )
    {
    bool        quit = false;
    bool        show_error = false;

    std::string error_text;

    // Initialize core.
    if( !initialize_system( argc, argv ) )
        {
        return;
        }

    try
        {
        // Download data.
        System::download_data();
        }
    catch( Exception & )
        {
        quit = true;
        }

    while( !quit )
        {
        // Initialize client.
        initialize_client();

        Global::enable_rendering();

        try
            {
            quit = !login_dialog_loop( error_text, show_error );

            show_error = false;

            if( !quit )
                {
                std::istringstream my_config( g_local_node_config );

                // Connect.
                System::connect( my_config );

                // Run.
                try
                    {
                    System::run();
                    }
                catch( Exception & ex )
                    {
                    error_text = ex.what();
                    show_error = true;
                    }

                // Deselect player.
                player_selected( null );

                // Remove pointer to account.
                    {
                    GameContext & game_context = Global::get_game_context();
                    game_context.link_with_account( null );
                    }

                // Disconnect.
                System::disconnect();
                }

            }
        catch( Exception & ex )
            {
            error_text = ex.what();
            show_error = true;
            }

        // Shutdown client.
        Global::shutdown( quit );
        }

    // Shutdown core.
    System::shutdown();
    }


bool Main::login_dialog_loop
    (
    const std::string & error,
    bool                show_error
    )
    {
    static std::string              login_name = "default";
    std::string                     login_password;
    std::string                     error_text;
    bool                            should_quit = false;
    bool                            done = false;

    while( !done )
        {
        bool    try_load_config_file = false;

        try
            {
            // Create and initialize login dialog.

            g_login_dialog.reset( new LoginDialog() );
            g_login_dialog->initialize( login_name, login_password, error_text );

            // Create testing modal message box.

            if( show_error )
                {
                g_message_box.reset( new MessageBox() );
                g_message_box->set_dispatcher_as_parent();
                g_message_box->set_message( error,
                            Renderer::TextRectangle::ALIGN_CENTER );
                g_message_box->do_layout();
                g_message_box->center_object();
                BlockerWidget * const blocker = g_message_box->switch_to_modal_mode();
                hope( blocker );
                blocker->set_base_color( SColor( 0.0f, 0.0f, 0.0f, 1.0f ) );
                blocker->set_opacity( 0.3f );
                }

            // Wait until user selects what to do.

            System::generic_loop();

            // Check login dialog result.

            switch( g_login_dialog->get_result() )
                {
                case LoginDialog::RESULT_OK:
                    login_name      = g_login_dialog->get_login_name();
                    login_password  = g_login_dialog->get_login_password();
                    try_load_config_file = true;
                    break;

                case LoginDialog::RESULT_CANCEL:
                    should_quit = true;
                    done = true;
                    break;
                }
            }
        catch( ... )
            {
            // Cleanup.

            g_message_box.reset();
            g_login_dialog.reset();
            throw;
            }

        // Destroy login dialog and message box.

        g_message_box.reset();
        g_login_dialog.reset();

        if( !done && try_load_config_file )
            {
            // Try loading local node config.
            g_local_node_config = read_local_node_config( login_name, login_password );

            // Everything ready.
            done = true;
            }
        }

    return !should_quit;
    }


bool Main::initialize_system
    (
    int                     argc,
    const char * const *    argv
    )
    {
    std::string config_file_name = CONFIG_MASSIV_CONF_PATH; // "./massiv.conf";

    // Parse command line.

    {//sob
    Massiv::CmdLineParser cmd_line( argc, argv, "h?", "npc" );
    bool parse = true;
    while( parse )
        {
        std::string parameter;
        const char option = cmd_line.get_next( parameter );
        switch( option )
            {
            case '-':
            case '\0':
                parse = false;
                break;

            case 'c':
                config_file_name = parameter;
                break;

            case 'h':
            case '?':
                std::cout << "Usage:\n" <<
                    argv[ 0 ] << " [options]\n\n" <<
                    "Options:\n" <<
                    "  -h (-?)          This help\n" <<
                    "  -c <file name>   Override default config file name.\n" <<
                    std::endl;
                return false;

            default:
                panic( "Invalid command line option." );
            }
        }
    }//eob

    // Initialize core system.

    std::cout << "Initializing system..." << std::endl;

    System::initialize
        (
        System::StartUpInfo
            (
            Massiv::Core::NodeId::TYPE_CLIENT,
            config_file_name,
            &Massiv::Generated::register_classes,
            &create_node_object_callback,
            &next_tick_callback
            )
        );

    // Read client settings.

    settings = new ClientSettings();

    // Check client settings

    check_client_settings();

    return true;
    }


void Main::check_client_settings()
    {
    std::cout << "Validating client settings..." << std::endl;
    if( settings->account_timeout < 5 )
        {
        settings->account_timeout = Main::DEFAULT_ACCOUNT_TIMEOUT;
        std::cout << "Account timeout automatically set to ";
        std::cout << Main::DEFAULT_ACCOUNT_TIMEOUT << ". " << std::endl;
        }
    if( settings->window_width < 320 )
        {
        settings->window_width = Main::DEFAULT_WINDOW_WIDTH;
        std::cout << "Window width automatically set to ";
        std::cout << Main::DEFAULT_WINDOW_WIDTH << "." << std::endl;
        }
    if( settings->window_height < 200 )
        {
        settings->window_height = Main::DEFAULT_WINDOW_HEIGHT;
        std::cout << "Window height automatically set to ";
        std::cout << Main::DEFAULT_WINDOW_HEIGHT << "." << std::endl;
        }
    }


void Main::decrypt_string
    (
    const std::string & input,
    const std::string & password,
    std::string &       decrypted
    )
    {
    if( input.size() < 13 )
        {
        std::ostringstream ss;
        ss  << "File cannot be a massiv login file!";
        throw ExBadLoginFile( ss.str() );
        }

    AccountManager::decrypt_string
        (
        input,
        password,
        decrypted
        );
    }


std::string Main::read_local_node_config
    (
    const std::string & name,
    const std::string & password
    )
    {
    std::string         config_file_name = Massiv::Core::Global::path_manager().get_login_directory() + "/" + name;
    std::ifstream       file( config_file_name.c_str(), std::ios_base::binary );
    std::ostringstream  oss;
    std::string         decrypted;

    oss << file.rdbuf();
    if( !file.good() )
        {
        std::ostringstream ss;
        ss  << "Error reading '" << config_file_name << "' file.";
        throw Massiv::Core::ExIOError( ss.str() );
        }
    else
        {
        std::istringstream iss( oss.str() );
        std::string pom;
        iss >> quote_string( pom );
        decrypt_string( pom, password, decrypted );
        }

    return decrypted;
    }


// Implemetation of Main::ClientSettings
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Main::ClientSettings::ClientSettings()
    :   Settings( "Client" ),
        chat_console_height
            (
            access_integer
                (
                "chat_console_height",
                Main::DEFAULT_CHAT_CONSOLE_HEIGHT
                )
            ),
        window_width
            (
            access_integer
                (
                "window_width",
                Main::DEFAULT_WINDOW_WIDTH
                )
            ),
        window_height
            (
            access_integer
                (
                "window_height",
                Main::DEFAULT_WINDOW_HEIGHT
                )
            ),
        window_bpp
            (
            access_integer
                (
                "window_bpp",
                Main::DEFAULT_WINDOW_BPP
                )
            ),
        account_timeout
            (
            access_integer
                (
                "account_timeout",
                Main::DEFAULT_ACCOUNT_TIMEOUT
                )
            )
    {
    }


} // namespace Client
} // namespace Demo
