/** @file client_main.h
 * Demo client.
 */

/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

 
#ifndef DEMO_CLIENT_MAIN_H
#define DEMO_CLIENT_MAIN_H

#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef MASSIV_CORE_MANAGEMENT_H
#include "core_management.h"
#endif

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_ACCOUNT_OBJECT_H
#include "lib/shared/account_object.h"
#endif

#ifndef DEMO_CLIENT_ENTITY_H
#include "lib/shared/client_entity.h"
#endif

namespace Demo {
namespace Client {

INJECT_CORE_SYMBOLS();


/**
 * Main. Static class that encapsulates program's main loop,
 * initialization, shutdown, callbacks from ClientNode etc.
 */
class Main
    {
    NO_COPY_CONSTRUCTOR( Main );
    NO_ASSIGNMENT_OPERATOR( Main );
    
public:

    /**
     * Constructor.
     */
    Main();

    /**
     * Destructor.
     */
    ~Main();
    
public:

    /**
     * Client registry settings.
     */
    struct ClientSettings : public Massiv::Core::Settings
        {
    public:

        /**
         * Constructor
         */
        ClientSettings();

    public:

        Integer &           chat_console_height;
            /**< Chat console height ( lines ). */
            
        Integer &           window_width;
            /**< Chat console height ( lines ). */
            
        Integer &           window_height;
            /**< Chat console height ( lines ). */
        
        Integer &           window_bpp;
            /**< Window bits per pixel. 0 for windowed mode. */
        
        Integer &           account_timeout;
            /**< Max allowed time to link with an account( seconds ). */
        };
        
    enum 
        {    
        DEFAULT_CHAT_CONSOLE_HEIGHT = 10
            /**< Default chat console height (lines). */
        };
    
    enum
        {
        DEFAULT_WINDOW_HEIGHT = 600
            /**< Default chat console height (pixels). */
        };
        
    enum
        {
        DEFAULT_WINDOW_WIDTH = 800
            /**< Default chat console width (pixels). */
        };
    
    enum
        {
        DEFAULT_WINDOW_BPP = 0
            /**< Default window bpp (bits per pixels). */
        };

    enum
        {
        DEFAULT_ACCOUNT_TIMEOUT = 30
            /**< Default max allowed time to link with an account (seconds). */
        };
            
public:

    //
    // Callbacks from ClientNode object.
    //

    static void welcome_to_simulation
        (
        WeakPointer< Lib::AccountObject > account
        );
    
    static void player_selected
        (
        WeakPointer< Lib::ClientEntity > player
        );

public:

    /**
     * Main function.
     * Initialize system and run the main loop.
     * @note This function can't have "main" name, because
     * "main" is #define defined by SDL.
     */
    void run
        (
        int             argc,
        const char **   argv
        );

private:

    /**
     * Initializes system. ( singletons etc. )
     */
    bool initialize_system
        (
        int                     argc,
        const char * const *    argv
        );
    
    /**
     * Initializes client window.
     */    
    void initialize_client();
    
    /**
     * Opens login dialog.
     * @param error_text string to show in messagebox.
     * @param show_error whether to show messgebox or not.
     * @return true if LOGIN button clicked.
     */
    bool login_dialog_loop
        (
        const std::string & error_text,
        bool                show_error
        );

    /**
     * Checks client settings in massiv.settings.conf.
     * Sets to default values when invalid detected.     
     */
    void check_client_settings();

    /**
     * Read local node config file.
     * @param input String to decrypt.
     * @param password Password to use for decrypting.
     * @param decrypted String containing decrypted text.
     */  
    void decrypt_string
        (
        const std::string & input, 
        const std::string & password, 
        std::string &       decrypted
        );
    
    /**
    * Read local node config file.
    * @param name Name of config file (without path).
    * @param password Password to access the config file.
    * @return String containing content of the local node config file.
    */    
    std::string read_local_node_config
        (
        const std::string & name,
        const std::string & password
        );

public:

    ClientSettings * settings;        

    }; // Main


} // namespace Client
} // namespace Demo


#endif // DEMO_CLIENT_MAIN_H
