/** @file editor_edit_entity_panel.cpp
 * Editor edit entity panel (implementation).
 */

/*
 * begin        : Sat Dec 16 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "editor/editor_edit_entity_panel.h"
#include "gui/widget_layout.h"
#include "gui/editor.h"
#include "renderer/renderer.h"
#include "lib/shared/client_entity.h"
#include "gui/client_full_console.h"
#include "game/game_context.h"
#include "lib/shared/account_object_rpc.h"
#include <sstream>


namespace Demo {
namespace Client {


const Float32 BORDER_WIDTH = 16.0f;
const Float32 BORDER_HEIGHT = 16.0f;
const Float32 PADDING_X = 8.0f;
const Float32 PADDING_Y = 8.0f;

static const char EE_POSITION[] = "Position";
static const char EE_ORIENTATION[] = "Orientation";
static const char EE_MODEL_TAGS[] = "Model tags";
static const char EE_INTERVAL[] = "Interval";

EditorEditEntityPanel::EditorEditEntityPanel()
    {
    // Init widget.
    set_border_size( BORDER_WIDTH, BORDER_HEIGHT );
    set_focusable( WINDOW_FOCUSABLE );
    set_dragable( true );
    set_auto_bringable_to_front( true );

    Float32 width = 300.0f;
    Float32 height = 20.0f;

    // Create widgets.

    hint_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    hint_label->set_parent_overlay_object( this );

    entity_name_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    entity_name_label->set_parent_overlay_object( this );

    entity_name_edit_box.reset( new EditBox() );
    entity_name_edit_box->set_parent_widget( this );
    entity_name_edit_box->set_base_size( SVector2( width, height ) ); // Fixme, ugly.
    entity_name_edit_box->do_layout();
    entity_name_edit_box->set_max_text_length_from_size();

    entity_properties_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    entity_properties_label->set_parent_overlay_object( this );

    entity_properties_list_box.reset( new ListBox() );
    entity_properties_list_box->set_parent_widget( this );

    value_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    value_label->set_parent_overlay_object( this );

    value_edit_box.reset( new EditBox() );
    value_edit_box->set_parent_widget( this );
    value_edit_box->set_base_size( SVector2( width, height ) ); // Fixme, ugly.
    value_edit_box->do_layout();
    value_edit_box->set_max_text_length_from_size();

    apply_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    apply_button->set_parent_widget( this );
    apply_button->set_text( "Apply" );
    apply_button->set_size_from_text();
    apply_button->do_layout();
    apply_button->set_message_receiver( this ); // Receive apply click.

    entity_properties_list_box->set_message_receiver( this );

    hint_value_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    hint_value_label->set_parent_overlay_object( this );

    setup_edit_entity_text_labels( "" );
    setup_entity_properties_list_box( Demo::Lib::ClientEntity::TYPE_DEFAULT, false );
    }


void EditorEditEntityPanel::do_layout()
    {
    // Do layout.

    // Make list box same width as edit box and label.
        {
        const Float32 lbw = entity_properties_list_box->get_current_size().x();
        Float32 value = 200.0f;
        if( lbw < value )
            {
            entity_properties_list_box->set_base_size( SVector2( value,
                entity_properties_list_box->get_current_size().y() ) );
            entity_properties_list_box->do_layout();
            }
        }

    WidgetTableLayout layout( BORDER_WIDTH, BORDER_HEIGHT, PADDING_X, PADDING_Y );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_CENTER );
    layout.put_object( hint_label.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( entity_name_label.get() );
    layout.put_object( entity_name_edit_box.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( entity_properties_label.get() );
    layout.put_object( entity_properties_list_box.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( value_label.get() );
    layout.put_object( value_edit_box.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( apply_button.get() );
    layout.put_object( hint_value_label.get() );

    const SVector2 new_size = layout.do_layout();
    set_base_size( new_size );
    }


void EditorEditEntityPanel::button_clicked
    (
    Button * button
    )
    {
    if( button == apply_button.get() )
        {
        EditEntityOptions opt = get_edit_entity_options();
        Remote< Lib::AccountObject > account = Global::get_game_context().get_account();
        if( account == null )
            {
            return;
            }
        if( entity_properties_list_box->get_selected_item().name == EE_POSITION )
            {
            // Move entity.
            account->async_edit_entity_command
                (
                opt.entity_name,
                "translate_entity",
                opt.value
                );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_ORIENTATION )
            {
            // Rotate entity.
            Float32 value;
                {
                std::istringstream  ss( opt.value );
                ss >> value;
                }
            std::ostringstream  ss;
            ss << Massiv::Core::FloatTraits< Float32 >::degrees_to_radians( value ) << " 0 0";
            account->async_edit_entity_command
                (
                opt.entity_name,
                "rotate_entity",
                ss.str()
                );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_MODEL_TAGS )
            {
            // Set model tags.
            account->async_edit_entity_command
                (
                opt.entity_name,
                "set_model_tags",
                opt.value
                );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_INTERVAL )
            {
            // Set interval.
            hope( actual_client_entity != null );
            Pointer< Lib::ClientEntity > ce = actual_client_entity.convert();
            /*
            if( ce->type == Demo::Lib::ClientEntity::TYPE_COLOR_TIMER )
                {
                account->async_edit_entity_command
                    (
                    opt.entity_name,
                    "setup_color_timer",
                    opt.value
                    );
                }
             */
            }
        }
    }


void EditorEditEntityPanel::list_box_item_selected
    (
    ListBox *   list_box
    )
    {
    if( list_box == entity_properties_list_box.get() )
        {
        if( list_box->is_item_selected() )
            {
            set_property_value();
            }
        }
    }


void EditorEditEntityPanel::set_property_value()
    {
    if( actual_client_entity != null )
        {
        value_edit_box->set_text( "" );
        hint_value_label->print_line( "                                           " );
        apply_button->enable();

        if( entity_properties_list_box->get_selected_item().name == "None" )
            {
            apply_button->disable();
            }
        if( entity_properties_list_box->get_selected_item().name == EE_POSITION )
            {
            // Set actual player position.
            WeakPointer< Lib::ClientEntity > ce = actual_client_entity.convert();
            std::ostringstream  ss;
            ss << ce->current_position.origin.x() << " " << ce->current_position.origin.y();
            value_edit_box->set_text( ss.str() );
            hint_value_label->print_line( "Click with right button to set new position" );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_ORIENTATION )
            {
            // Set actual player orientation.
            WeakPointer< Lib::ClientEntity > ce = actual_client_entity.convert();
            std::ostringstream  ss;
            ss << Massiv::Core::FloatTraits< Float32 >::radians_to_degrees( ce->current_position.orientation.heading() );
            value_edit_box->set_text( ss.str() );
            hint_value_label->print_line( "Set float value in degrees <0 .. 360>      " );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_MODEL_TAGS )
            {
            // Set actual player position.
            WeakPointer< Lib::ClientEntity > ce = actual_client_entity.convert();
            std::ostringstream  ss;
            ss << ce->model_tags;
            value_edit_box->set_text( ss.str() );
            hint_value_label->print_line( "Set model tags (monster, player, ... )     " );
            }
        if( entity_properties_list_box->get_selected_item().name == EE_INTERVAL )
            {
            // Set time interval.
            WeakPointer< Lib::ClientEntity > ce = actual_client_entity.convert();
            value_edit_box->set_text( "" ); // FIXME: Cannot get value.
            hint_value_label->print_line( "Set time interval in seconds (e.g. 2.75)   " );
            }
        }
    }


EditEntityOptions EditorEditEntityPanel::get_edit_entity_options()
    {
    EditEntityOptions opt;

    opt.entity_name = entity_name_edit_box->get_text();
    opt.value = value_edit_box->get_text();

    return opt;
    }


void EditorEditEntityPanel::set_position_coordinates
    (
    Float32     x,
    Float32     y
    )
    {
    if( entity_properties_list_box->get_selected_item().name == EE_POSITION )
        {
        // We can set position value.
        std::ostringstream  ss;
        ss << x << " " << y;
        value_edit_box->set_text( ss.str() );
        }
    }


void EditorEditEntityPanel::set_entity_intersection
    (
    const EntityManager::RayIntersection & entity_intersection
    )
    {
    EntityManager::EntityPointer client_entity = entity_intersection.entity;
    actual_client_entity = client_entity;
    const Massiv::Core::ObjectId entity_object_id = client_entity->entity.get_object_id();

        {
        std::ostringstream  ss;
        ss << entity_object_id; //entity.get_object_id();
        entity_name_edit_box->set_text( ss.str() );
        }
    //int property_id_before = entity_properties_list_box->get_selected_item().id;
    //setup_entity_properties_list_box( entity_intersection.entity->type, false );
    value_edit_box->set_text( "" );
    //if( entity_properties_list_box->select_item_by_id( property_id_before ) )
        {
        set_property_value(); // Set the same property value for this entity.
        }
    }


bool EditorEditEntityPanel::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( key == SDLK_ESCAPE )
        {
        // Hide editor.

        disable_focus();
        Global::get_editor().set_visibility( false );
        return true;
        }
    return false;
    }


void EditorEditEntityPanel::setup_entity_properties_list_box
    (
    const Demo::Lib::ClientEntity::Type     type,
    bool                                    init
    )
    {
    typedef ListBox::Item       Item;
    typedef Lib::ClientSector   Sector;
    int         id = 0;
    std::string name;

    entity_properties_list_box->clear();

    id = -1;
    name = "None";
    entity_properties_list_box->push( Item( id, name ) );

    if( init )
        {
        entity_properties_list_box->set_base_size( entity_properties_list_box->get_ideal_size() );
        entity_properties_list_box->do_layout();
        return;
        }

    // Common entity part.

    id = 0;
    name = EE_POSITION;
    entity_properties_list_box->push( Item( id, name ) );

    id = 1;
    name = EE_ORIENTATION;
    entity_properties_list_box->push( Item( id, name ) );

    id = 2;
    name = EE_MODEL_TAGS;
    entity_properties_list_box->push( Item( id, name ) );

    switch( type )
        {
        case Demo::Lib::ClientEntity::TYPE_MONSTER:
            {
            break;
            }
        case Demo::Lib::ClientEntity::TYPE_PLAYER:
            {
            break;
            }
        /*case Demo::Lib::ClientEntity::TYPE_COLOR_TIMER:
            {
            id = 3;
            name = EE_INTERVAL;
            entity_properties_list_box->push( Item( id, name ) );
            break;
            }*/
        default:
            {
            break;
            }
        }

    entity_properties_list_box->set_base_size( entity_properties_list_box->get_ideal_size() );
    entity_properties_list_box->do_layout();
    }


void EditorEditEntityPanel::setup_edit_entity_text_labels
    (
    const std::string &     entity_name
    )
    {
    // Common part.

    hint_label->print_line( "Click  entity with right button to edit its properties" );

    entity_name_label->print_line( "Entity name:    " );
    entity_name_edit_box->set_text( "" );

    entity_properties_label->print_line( "Select property:" );

    value_label->print_line( "Value:          " );
    value_edit_box->set_text( "" );
    }

} // namespace Demo::Client
} // namespace Demo
