/** @file edit_elevation_panel.cpp
 * Modify elevation panel (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "editor/editor_elevation_panel.h"
#include "gui/editor.h"
#include <sstream>


namespace Demo {
namespace Client {


ModifyElevationPanel::ModifyElevationPanel()
    {
    set_border_size( 16.0f, 16.0f );
    set_focusable( WINDOW_FOCUSABLE );
    set_dragable( true );
    set_auto_bringable_to_front( true );

    // Create widgets.

    map_kind_list_box.reset( new ListBox() );
    map_kind_list_box->set_parent_widget( this );

    operation_list_box.reset( new ListBox() );
    operation_list_box->set_parent_widget( this );

    function_list_box.reset( new ListBox() );
    function_list_box->set_parent_widget( this );

    radius_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    radius_label->set_parent_overlay_object( this );
    radius_label->print_line( "Radius:   " );

    strength_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    strength_label->set_parent_overlay_object( this );
    strength_label->print_line( "Strength: " );

    radius_edit_box.reset( new EditBox() );
    radius_edit_box->set_parent_widget( this );

    strength_edit_box.reset( new EditBox() );
    strength_edit_box->set_parent_widget( this );

    align_to_grid_button.reset( new Button( Button::TYPE_SWITCH ) );
    align_to_grid_button->set_parent_widget( this );

    inverse_strength_button.reset( new Button( Button::TYPE_SWITCH ) );
    inverse_strength_button->set_parent_widget( this );

    // Do layout.

    Float32 px = 10.0f; // Padding x.
    Float32 py = 10.0f; // Padding y.
    Float32 w = 260.0f; // Width.
    Float32 cw = w - 2.0f * px; // Client width.
    Float32 fh = 20.0f; // Font height.
    Float32 x;
    Float32 y;
    Float32 tx; // Temporary.

    x = px;
    y = py + py + py;
    map_kind_list_box->set_base_offset( SVector2( x, y ) );
    map_kind_list_box->set_base_size( SVector2( cw, fh ) );
    map_kind_list_box->do_layout();

    y += fh + py;
    operation_list_box->set_base_offset( SVector2( x, y ) );
    operation_list_box->set_base_size( SVector2( cw, fh ) );
    operation_list_box->do_layout();

    y += fh + py;
    function_list_box->set_base_offset( SVector2( x, y ) );
    function_list_box->set_base_size( SVector2( cw, fh ) );
    function_list_box->do_layout();

    y += fh + py;
    radius_label->set_base_offset( SVector2( x, y ) );

    tx = radius_label->get_current_size().x();
    x += tx;
    radius_edit_box->set_base_offset( SVector2( x, y ) );
    radius_edit_box->set_base_size( SVector2( cw - tx, fh ) );
    radius_edit_box->set_max_text_length_from_size();
    radius_edit_box->set_text( "1" );
    x -= tx;

    y += fh + py;
    strength_label->set_base_offset( SVector2( x, y ) );

    tx = strength_label->get_current_size().x();
    x += tx;
    strength_edit_box->set_base_offset( SVector2( x, y ) );
    strength_edit_box->set_base_size( SVector2( cw - tx, fh ) );
    strength_edit_box->set_max_text_length_from_size();
    strength_edit_box->set_text( "0.0" );
    x -= tx;

    y += fh + py;
    align_to_grid_button->set_base_offset( SVector2( x, y ) );
    align_to_grid_button->set_text( "Align to Grid" );
    align_to_grid_button->set_size_from_text();
    align_to_grid_button->set_base_size( SVector2
        ( align_to_grid_button->get_current_size().x(), fh ) );
    align_to_grid_button->push();

    tx = align_to_grid_button->get_current_size().x() + px;
    x += tx;
    inverse_strength_button->set_base_offset( SVector2( x, y ) );
    inverse_strength_button->set_text( "Inverse Str." );
    inverse_strength_button->set_size_from_text();
    inverse_strength_button->set_base_size( SVector2
        ( inverse_strength_button->get_current_size().x(), fh ) );
    x -= tx;

    y += fh + py;
    set_base_size( SVector2( w, y ) );
    do_layout();

    // Fill list boxes.

    setup_list_boxes();
    }


ModifyElevationOptions ModifyElevationPanel::get_modify_elevation_options()
    {
    ModifyElevationOptions opt;

    hope( map_kind_list_box->is_item_selected() );
    opt.map = static_cast< ModifyElevationOptions::MapKind >( map_kind_list_box->get_selected_item().id );

    hope( operation_list_box->is_item_selected() );
    opt.operation = static_cast< ModifyElevationOptions::MapOperation >( operation_list_box->get_selected_item().id );

    hope( function_list_box->is_item_selected() );
    opt.function = static_cast< ModifyElevationOptions::MapFunction >( function_list_box->get_selected_item().id );

        {
        opt.radius = 10.0f;
        std::istringstream ss( radius_edit_box->get_text() );
        ss >> opt.radius;
        }

        {
        opt.strength = 10.0f;
        std::istringstream ss( strength_edit_box->get_text() );
        ss >> opt.strength;
        }

    opt.align_to_grid = align_to_grid_button->is_down();

    if( inverse_strength_button->is_down() )
        {
        opt.strength = - opt.strength;
        }

    return opt;
    }


bool ModifyElevationPanel::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( key == SDLK_ESCAPE )
        {
        // Hide editor.

        disable_focus();
        Global::get_editor().set_visibility( false );
        return true;
        }
    return false;
    }


void ModifyElevationPanel::setup_list_boxes()
    {
    typedef ListBox::Item       Item;
    typedef Lib::ClientSector   Sector;
    int         id;
    std::string name;

    // Map kind.

    id = Sector::KIND_ELEVATION_MAP;
    name = "Elevation Map";
    map_kind_list_box->push( Item( id, name ) );

    id = Sector::KIND_DELTA_ELEVATION_MAP;
    name = "Delta Elevation Map";
    map_kind_list_box->push( Item( id, name ) );

    // Operation.

    id = Sector::OPERATION_SET;
    name = "Set";
    operation_list_box->push( Item( id, name ) );

    id = Sector::OPERATION_ADD;
    name = "Add";
    operation_list_box->push( Item( id, name ) );

    id = Sector::OPERATION_MULTIPLY;
    name = "Multiply";
    operation_list_box->push( Item( id, name ) );

    // Function.

    id = Sector::FUNCTION_CIRCLE;
    name = "Circle";
    function_list_box->push( Item( id, name ) );

    id = Sector::FUNCTION_FUZZY_CIRCLE;
    name = "Fuzzy Circle";
    function_list_box->push( Item( id, name ) );

    id = Sector::FUNCTION_GAUSSIAN;
    name = "Gaussian";
    function_list_box->push( Item( id, name ) );
    }


} // namespace Demo::Client
} // namespace Demo
