/** @file editor_material_panel.cpp
 * Editor material panel (implementation).
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "editor/editor_material_panel.h"
#include "gui/widget_layout.h"
#include "gui/editor.h"
#include "renderer/renderer.h"
#include "renderer/terrain_material.h"
#include <sstream>


namespace Demo {
namespace Client {


namespace {

const Float32 BORDER_WIDTH = 16.0f;
const Float32 BORDER_HEIGHT = 16.0f;
const Float32 PADDING_X = 8.0f;
const Float32 PADDING_Y = 8.0f;

} // anonymous namespace


EditorMaterialPanel::EditorMaterialPanel()
    {
    // Init widget.
    set_border_size( BORDER_WIDTH, BORDER_HEIGHT );
    set_focusable( WINDOW_FOCUSABLE );
    set_dragable( true );
    set_auto_bringable_to_front( true );

    // Create widgets.

    material_list_box.reset( new ListBox() );
    material_list_box->set_parent_widget( this );
    // Size of material_list_box is set in setup_list_boxes().

    radius_label.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    radius_label->set_parent_overlay_object( this );
    radius_label->print_line( "Radius:" );

    radius_edit_box.reset( new EditBox() );
    radius_edit_box->set_parent_widget( this );
    radius_edit_box->set_base_size( SVector2( 100.0f, 20.0f ) ); // Fixme, ugly.
    radius_edit_box->set_text( "10" );
    radius_edit_box->do_layout();
    radius_edit_box->set_max_text_length_from_size();

    align_to_grid_button.reset( new Button( Button::TYPE_SWITCH ) );
    align_to_grid_button->set_parent_widget( this );
    align_to_grid_button->set_text( "Align to Grid" );
    align_to_grid_button->set_size_from_text();
    align_to_grid_button->do_layout();

    // Fill list boxes.
    setup_list_boxes();
    }


void EditorMaterialPanel::do_layout()
    {
    // Do layout.

    // Make list box same width as edit box and label.
        {
        const Float32 rlw = radius_label->get_current_size().x();
        const Float32 rebw = radius_edit_box->get_current_size().x();
        const Float32 mlbw = material_list_box->get_current_size().x();
        if( rlw + PADDING_X + rebw > mlbw )
            {
            material_list_box->set_base_size( SVector2( rlw + PADDING_X + rebw,
                material_list_box->get_current_size().y() ) );
            material_list_box->do_layout();
            }
        else
            {
            radius_edit_box->set_base_size( SVector2( mlbw - PADDING_X - rlw,
                radius_edit_box->get_current_size().y() ) );
            radius_edit_box->do_layout();
            }
        }

    WidgetTableLayout layout( BORDER_WIDTH, BORDER_HEIGHT, PADDING_X, PADDING_Y );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( material_list_box.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( radius_label.get() );
    layout.put_object( radius_edit_box.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_LEFT );
    layout.put_object( align_to_grid_button.get() );

    const SVector2 new_size = layout.do_layout();
    set_base_size( new_size );
    }


ModifyMaterialOptions EditorMaterialPanel::get_modify_material_options()
    {
    ModifyMaterialOptions opt;

    if( !material_list_box->is_item_selected() )
        {
        // No material in material database?
        opt.material_id = 0;
        }
    else
        {
        opt.material_id = material_list_box->get_selected_item().id;
        }

        {
        opt.radius = 10.0f;
        std::istringstream ss( radius_edit_box->get_text() );
        ss >> opt.radius;
        }

    opt.align_to_grid = align_to_grid_button->is_down();

    return opt;
    }


void EditorMaterialPanel::refresh
    (
    bool force
    )
    {
    Renderer::TerrainMaterialDatabase & terrain_material_database = Renderer::Global::get_terrain_material_database();

    if( force || last_time_materials_refreshed < terrain_material_database.get_update_time() )
        {
        setup_material_list_box();
        do_layout();
        last_time_materials_refreshed = SystemTime::current_time();
        }
    }


bool EditorMaterialPanel::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( key == SDLK_ESCAPE )
        {
        // Hide editor.

        disable_focus();
        Global::get_editor().set_visibility( false );
        return true;
        }
    return false;
    }


void EditorMaterialPanel::setup_material_list_box()
    {
    typedef ListBox::Item       Item;
    Renderer::TerrainMaterialDatabase & terrain_material_database = Renderer::Global::get_terrain_material_database();
    const int   max_num_materials = ( 2 << ( Lib::ClientSector::MATERIAL_ID_NUM_BITS - 1 ) );
    int         id;
    bool        default_added = false;
    int         prev_selected = -1;

    if( material_list_box->is_item_selected() )
        {
        prev_selected = material_list_box->get_selected_item().id;
        }

    material_list_box->clear();

    for( id = 0; id < max_num_materials; ++id )
        {
        const Renderer::TerrainMaterialDatabase::MaterialInfo & material_info = terrain_material_database.get_material( id );

        if( material_info.name == "default" )
            {
            if( default_added )
                {
                continue;
                }
            else
                {
                default_added = true;
                }
            }

        material_list_box->push( Item( id, material_info.name ) );
        }

    // Reselect previously selected material.
    if( prev_selected >= 0 )
        {
        material_list_box->select_item_by_id( prev_selected );
        }

    material_list_box->set_base_size( material_list_box->get_ideal_size() );
    material_list_box->do_layout();
    }


void EditorMaterialPanel::setup_list_boxes()
    {
    setup_material_list_box();
    }


} // namespace Demo::Client
} // namespace Demo
