/** @file entity_manager.cpp
 * Manager of client entities (implementation).
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "game/entity_manager.h"
#include "client_global.h"
#include "tag_database.h"
#include "renderer/renderer.h"
#include "renderer/pass_manager.h"
#include "renderer/passes.h"
#include "renderer/model_entity.h"
#include "renderer/render_objects.h"
#include "renderer/terrain_settings.h"
#include "game/player_manager.h"
#include "game/sector_manager.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/client_light.h"
#include "lib/shared/client_sector.h"


namespace Demo {
namespace Client {


/** @def PREDICTION_DEBUG
 * Level of prediction debugging messages:
 * - 0: Do not log prediction debugging info.
 * - 1: Log resets.
 * - 2: Log some info.
 * - 3: Log all info.
 */
#define PREDICTION_DEBUG 1


namespace {


//@{
/// Import stuff from ClientSector.
const Float32 SECTOR_X_SIZE = static_cast< Float32 >( ::Demo::Lib::ClientSector::SECTOR_X_SIZE );
const Float32 SECTOR_Y_SIZE = static_cast< Float32 >( ::Demo::Lib::ClientSector::SECTOR_Y_SIZE );
const int NUM_X_TILES = ::Demo::Lib::ClientSector::NUM_X_TILES;
const int NUM_Y_TILES = ::Demo::Lib::ClientSector::NUM_Y_TILES;
const Float32 TILE_X_SIZE = SECTOR_X_SIZE / static_cast< Float32 >( NUM_X_TILES );
const Float32 TILE_Y_SIZE = SECTOR_Y_SIZE / static_cast< Float32 >( NUM_Y_TILES );
//@}


/// Speed of smooth orientation changes.
const Float32 ROTATION_SPEED_FACTOR = 3.0f;


// Linear movement functions:
//   X( t ) = origin + velocity * ( t - zero_time )

// zero_time is remembered as separate paramater, because
//   origin - velocity * zero_time
// could become too big and imprecise for large zero_time and velocity.


/**
 * Compute origin and velocity of linear movement given two points and times.
 */
void get_lm_velocity
    (
    SVector2 &                  velocity,
    const SVector2 &            x0,
    const EntityManager::Time & t0,
    const SVector2 &            x1,
    const EntityManager::Time & t1
    )
    {
    // Note: Must subtract .get_value(), because Time can't be negative.

    const EntityManager::Time::TimeType td = t1.get_value() - t0.get_value();
    if( ::Massiv::Core::FloatTraits< EntityManager::Time::TimeType >::epsilon_zero( td ) )
        {
        velocity = SVector2();
        }
    else
        {
        velocity = ( x1 - x0 ) / static_cast< Float32 >( td );
        }
    }


/**
 * Compute linear movement position.
 */
void get_lm_position
    (
    SVector2 &                  position,
    const SVector2 &            x0,
    const EntityManager::Time & t0,
    const SVector2 &            v,
    const EntityManager::Time & t
    )
    {
    // Note: Must subtract .get_value(), because Time can't be negative.

    position = x0 + v * static_cast< Float32 >( t.get_value() - t0.get_value() );
    }


/**
 * Periodic-clamp @a angle into 0-2pi range.
 */
Float32 normalize_angle
    (
    Float32 angle
    )
    {
    return ::Massiv::Core::FloatTraits< Float32 >::clamp_periodic
        (
        angle,
        0.0f,
        2.0f * MATHF_PI
        );
    }


/**
 * Change @a current_angle towards @a final_angle.
 */
Float32 move_angle
    (
    Float32 current_angle,
    Float32 final_angle,
    Float32 max_delta
    )
    {
    // Normalize angles.

    current_angle = normalize_angle( current_angle );
    final_angle = normalize_angle( final_angle );

    // Determine rotation direction and distance.

    const Float32 distance = final_angle - current_angle;
    bool increase;
    if( distance >= 0.0f )
        {
        // --- 0 --- ca --- fa --- 2pi ---

        if( distance <= MATHF_PI )
            {
            // Increase ca towards fa.

            increase = true;
            }
        else
            {
            // Decrease ca to wrap-around towards fa.

            current_angle += 2.0f * MATHF_PI;
            increase = false;
            }
        }
    else
        {
        // --- 0 --- fa --- ca --- 2pi ---

        if( -distance <= MATHF_PI )
            {
            // Decrease ca towards fa.

            increase = false;
            }
        else
            {
            // Increase ca to wrap-around towards fa.

            current_angle -= 2.0f * MATHF_PI;
            increase = true;
            }
        }

    // Apply change.

    if( increase )
        {
        const Float32 delta = final_angle - current_angle;
        hope( delta >= 0.0f );
        if( delta < max_delta )
            {
            return final_angle;
            }
        else
            {
            return normalize_angle( current_angle + max_delta );
            }
        }
    else
        {
        const Float32 delta = current_angle - final_angle;
        hope( delta >= 0.0f );
        if( delta < max_delta )
            {
            return final_angle;
            }
        else
            {
            return normalize_angle( current_angle - max_delta );
            }
        }
    /*NOTREACHED*/
    }


/**
 * Compute light source position and influence.
 */
void get_light_params
    (
    SVector3 &                          position,
    Float32 &                           influence,
    const EntityManager::EntityInfo &   entity,
    const SVector3 &                    player_position
    )
    {
    EntityManager::LightPointer light = entity.light;
    hope( !light.is_null() );

    // Compute position.

    position = light->get_offset();
    entity.orientation.rotate( position );
    position += entity.position;

    // Compute distance from player.

    const Float32 distance = ( player_position - position ).magnitude();

    // Compute attenuation.

    const SVector3 & af = light->get_attenuation();
    const Float32 attenuation = 1.0f / ( distance * ( distance * af.z() + af.y() ) + af.x() );

    // Multiply by diffuse + ambient intensity.

    const SColor & ambient = light->get_ambient_intensity();
    const SColor & diffuse = light->get_diffuse_intensity();
    influence = attenuation *
                ( ambient.red() + ambient.green() + ambient.blue() +
                  diffuse.red() + diffuse.green() + diffuse.blue() );
    }


} // anonymous namespace


// class EntityManager::MyStatistics

EntityManager::MyStatistics::MyStatistics()
      : ::Massiv::Core::Statistics( "Client/EntityManager" ),
        entity_count( create_integer( "entity_count", 0 ) )
    {
    }


// class EntityManager::MySettings

EntityManager::MySettings::MySettings()
      : ::Massiv::Core::Settings( "Client/EntityManager" ),
        prediction( access_boolean( "prediction", true ) ),
        prediction_limit( access_float( "prediction_limit", 0.5f ) )
    {
    }


// class EntityManager::EntityInfo

EntityManager::EntityInfo::EntityInfo
    (
    EntityPointer the_entity
    ) : entity( the_entity ),
        light( the_entity->light ),
        should_be_visible( false ),
        history_base_index( 0 ),
        history_size( 0 ),
        prediction_valid( false )
    {
    update_visual( false );
    }


EntityManager::EntityInfo::EntityInfo
    (
    const EntityInfo & other
    ) : entity( other.entity ),
        light( other.light ),
        visual( other.visual ),
        model_name( other.model_name ),
        layer_id( other.layer_id ),
        should_be_visible( other.should_be_visible ),
        position( other.position ),
        orientation( other.orientation ),
        sector_id( other.sector_id ),
        // msvc6 is unable to initialize history using "history( other.history )"
        history_base_index( other.history_base_index ),
        history_size( other.history_size ),
        prediction_valid( other.prediction_valid ),
        predicted_destination( other.predicted_destination ),
        predicted_destination_time( other.predicted_destination_time ),
        predicted_velocity( other.predicted_velocity ),
        final_orientation( other.final_orientation ),
        previous_move_time( other.previous_move_time ),
        previous_model_tags( other.previous_model_tags )
    {
    int i;
    for( i = 0; i < POSITION_HISTORY_LENGTH; ++i )
        {
        history[ i ] = other.history[ i ];
        }
    update_visual( false );
    }


void EntityManager::EntityInfo::update
    (
    Time &              next_update_time,
    const Time &        update_time,
    const Time &        current_time,
    Float32             prediction_limit,
    const ObjectId &    visible_layer,
    bool                force
    )
    {
    // Use new replica data if it has been updated or the update was forced.

    if( update_time.is_valid() || force )
        {
        // Update pointer to light source.

        light = entity->light;

        // Update #visual.

        update_visual( force );

        // Copy data we do not predict.

        ::Demo::Lib::EntityPosition & entity_position = entity->current_position;
        // May be changed in prediction, this is basic guess...
        sector_id = entity->current_sector.get_object_id();

        // Reset entity prediction if layer has changed.

        const ObjectId old_layer_id = layer_id;
        layer_id = entity_position.layer.get_object_id();
        if( layer_id != old_layer_id )
            {
            history[ 0 ].position = entity_position.origin;
            history[ 0 ].time = update_time;
            history_base_index = 0;
            history_size = 1;
            prediction_valid = false;
            }

        // Update position history if replica position has been updated.

        else if( entity_position.origin.get_dirty_time().is_valid() )
            {
            if( history_size == POSITION_HISTORY_LENGTH )
                {
                history_base_index++;
                }
            else
                {
                history_size++;
                }
            const std::size_t index = get_history_index( 0 );
            history[ index ].position = entity_position.origin;
            history[ index ].time = update_time;
            }

        // Clear replica position dirty timestamp.

        entity_position.origin.set_invalid_dirty_time();

        // Remember final orientation.

        final_orientation = entity_position.orientation;
        }

    // Layer and visibility.
    // This must be done in every update(), we have no visible_layer_changed flag.

    should_be_visible = ( layer_id == visible_layer );
    if( !should_be_visible )
        {
        // Never draw entity in invisible layer. Otherwise visibility of the visual
        // is determined by #move() method.

        if( visual.get() )
            {
            visual->set_visible( false );
            }
        }

    // Update prediction information.
    // Currently all entities use "generic" prediction.
    // Better waypoint prediction should be implemented in the future.

    update_generic_prediction( next_update_time, update_time.is_valid(),
        current_time, prediction_limit );
    hope( prediction_valid );
    }


void EntityManager::EntityInfo::move
    (
    bool            use_prediction,
    const Time &    current_time
    )
    {
    // Time delta.

    const Float32 time_delta = static_cast< Float32 >
                ( ( current_time - previous_move_time ).get_value() );
    previous_move_time = current_time;

    // Determine current position.

    SVector2 position2d;
    if( use_prediction )
        {
        hope( prediction_valid );
        get_lm_position( position2d, predicted_destination, predicted_destination_time,
                    predicted_velocity, current_time );
        }
    else
        {
        position2d = entity->current_position.origin;
        }

    // Determine current orientation.

    const Float32 max_angle_delta = ROTATION_SPEED_FACTOR * time_delta;
    if( use_prediction )
        {
        orientation.set_heading( move_angle( orientation.heading(),
                    final_orientation.heading(), max_angle_delta ) );
        orientation.set_pitch( move_angle( orientation.pitch(),
                    final_orientation.pitch(), max_angle_delta ) );
        orientation.set_bank( move_angle( orientation.bank(),
                    final_orientation.bank(), max_angle_delta ) );
        }
    else
        {
        orientation = final_orientation;
        }

    // Determine current sector.
    // If get_sector() fails, use previous value sector_id - either from previous move()
    // or from position copied from most recent replica update.

    const SectorManager & sector_manager = Global::get_sector_manager();
    sector_manager.get_sector( sector_id, layer_id, position2d.x(), position2d.y() );

    // Determine height and update visual.

    Float32 height;
    if( sector_manager.get_height( height, sector_id, position2d.x(), position2d.y() ) )
        {
        // We are able to determine entity position.
        // Update visual position and make it visible if it's in currently visible layer.

        position = SVector3( position2d.x(), position2d.y(), height );
        if( visual.get() )
            {
            visual->set_position( position );
            visual->set_orientation( orientation );
            if( should_be_visible )
                {
                visual->set_visible( true );
                }
            }
        }
    else
        {
        // We can't determine entity position.
        // Hide visual if it's in visible layer. Otherwise it's already invisible.

        if( should_be_visible )
            {
            if( visual.get() )
                {
                visual->set_visible( false );
                }
            }
        }
    }


void EntityManager::EntityInfo::update_fog()
    {
    if( visual.get() )
        {
        Renderer::RenderObject::PipelineStates & states = visual->render_object.pipeline_states;
        Renderer::RenderObject::PipelineStates::iterator it;
        for( it = states.begin(); it != states.end(); ++it )
            {
            Renderer::Global::get_terrain_settings().apply_fog( *it );
            }
        }
    }


void EntityManager::EntityInfo::update_visual
    (
    bool force
    )
    {
    EntityManager & entity_manager = Global::get_entity_manager();

    if( entity->is_visual( entity_manager.is_admin_mode() ) )
        {
        // Create/update visual representation.
        // Determine current tag set and model name.
        // Use empty model name if an database match can't be found (damaged or
        // incomplete database). Default model (simple box) will be created for
        // such models.

        std::string new_model_name;
        bool        model_name_changed = true;

        try
            {
            TagDatabase & model_database = Global::get_model_database();
            const std::string model_tags = entity->get_model_tags();

            // If the tags have changed, the model has no visual object yet
            // or the update was forced, lookup the database. Note that the
            // database can not be refreshed here. It must be done on a
            // "central" place so that visual object updates could be forced
            // for all entities.

            if( model_tags != previous_model_tags || !visual.get() || force )
                {
                const TagSet tag_set( model_tags );
                TagDatabase::Match match = model_database.match( tag_set );
                if( model_database.match_found( match ) )
                    {
                    new_model_name = model_database.get_value( match, "model" );
                    }
                previous_model_tags = model_tags;
                }
            else
                {
                model_name_changed = false;
                }
            }
        catch( std::exception & )
            {
            // Ignore any database errors, use dummy model.
            }

        // Destroy old #visual if model name has changed.

        if( model_name_changed && model_name != new_model_name )
            {
            visual.reset();
            model_name = new_model_name;
            }

        // If there is no #visual, create it.

        if( !visual.get() )
            {
            visual.reset( new Renderer::ModelEntity( model_name ) );
            visual->register_to_global_manager( Renderer::PASS_WORLD );

            // New entities are invisible until updated.

            visual->set_visible( false );
            }
        }
    else
        {
        // Destroy visual representation.

        visual.reset();
        }
    }


void EntityManager::EntityInfo::update_generic_prediction
    (
    Time &          next_update_time,
    bool            force_update,
    const Time &    current_time,
    Float32         prediction_limit
    )
    {
    // This function assumes that entity motion is linear and tries to estimate
    // linear motion parameters which will be used for position prediction.

    // First we try to guess parameters of "real" entity motion (assuming it's
    // linear) from history of replica position. This is quite tricky, because
    // the time in PositionTime structure is equal to simulation time when
    // replica update has been sent, not to simulation time when physics
    // set the position. So we try to find the oldest history entry which
    // seems to belong to current linear motion and use that to estimate entity
    // speed to minimize errors.

    // "Real" motion parameters are stored in real_origin, real_zero_time and
    // real_velocity.

    // If currently predicted entity position is too far away from position
    // extrapolated from "real" motion, entity has probably teleported (or
    // previous prediction failed miserably). In that case we use parameters
    // from "real" motion directly as prediction parameters (this will
    // "teleport" the entity to "real" path immediately).

    // Otherwise we setup prediction parameters so that the entity will move
    // from its current position and reach the "real" path in "near" future.
    // This can be quite tricky, more information is below (where value of
    // predicted_destination_time is determined.

    hope( history_size >= 1 );

    // If prediction is valid, entity has not been modified and #predicted_destination_time
    // has not been reached yet, do not update prediction info.

    if( prediction_valid && !force_update && current_time < predicted_destination_time )
        {
        return;
        }

    // Estimate "real" linear motion parameters of recent entity movement.

    SVector2 real_origin;
    Time real_zero_time;
    SVector2 real_velocity;
    const std::size_t i0 = get_history_index( 0 );
    Float32 cross_delta = prediction_limit; // described below...
    bool stopped = true;

    // If there is only one element in position history or the most recent
    // element is older than prediction_limit, assume entity does not move.

    if( entity->stopped.read() ||
        ( history_size == 1 ) ||
        ( history[ i0 ].time + prediction_limit < current_time ) )
        {
        // No movement.

#if PREDICTION_DEBUG >= 2

        {//sob
        std::ostringstream oss;
        oss << "Entity " << entity.get_object_id() << " prediction: No motion ";
        if( entity->stopped )
            {
            oss << "(stopped).";
            }
        else if( history_size == 1 )
            {
            oss << "(empty position history).";
            }
        else
            {
            oss << "(no update for " << ( current_time - history[ i0 ].time ) << " seconds).";
            }

        using namespace Massiv::Core;
        Massiv::Core::Global::log_debug( Status::FACILITY_GENERIC,
                    Status::PRIORITY_NORMAL, oss.str() );
        } //eob

#endif // PREDICTION_DEBUG >= 2

        real_origin = history[ i0 ].position;
        real_zero_time = history[ i0 ].time;
        hope( real_velocity.is_zero() );
        hope( stopped );
        }
    else
        {
        // Use most recent position and oldest position on current motion line
        // (estimated from two most recent positions) to estimate motion line and speed.
        // FIXME: this could use some improvement (LR?)

        const std::size_t i1 = get_history_index( 1 );

        // Compute time delta between the two updates and cross delta.

        Float32 update_delta = static_cast< Float32 >( history[ i0 ].time.get_value() -
                    history[ i1 ].time.get_value() );
        cross_delta = 2.0f * update_delta;

        const SVector2 & p0 = history[ i0 ].position;
        const SVector2 & p1 = history[ i1 ].position;
        if( p0.epsilon_equals( p1 ) )
            {
            // Equal positions - no movement.

#if PREDICTION_DEBUG >= 2

            {//sob
            std::ostringstream oss;
            oss << "Entity " << entity.get_object_id() << " prediction: No motion "
                        "(two most recent positions are the same)";

            using namespace Massiv::Core;
            Massiv::Core::Global::log_debug( Status::FACILITY_GENERIC,
                        Status::PRIORITY_NORMAL, oss.str() );
            } //eob

#endif // PREDICTION_DEBUG >= 2

            real_origin = p0;
            real_zero_time = history[ i0 ].time;
            hope( real_velocity.is_zero() );
            hope( stopped );
            }
        else
            {
            // How many history elements were part of the linear motion?

            const Float32 recent_speed = ( ( p0 - p1 ) / update_delta ).magnitude();
            const Float32 max_distance = 2.0f * recent_speed;

            const SVector2 n = SVector2( p1.y() - p0.y(), p0.x() - p1.x() ).normalize();
            const Float32 d = -p0 * n;
            std::size_t count;
            for( count = 2; count < history_size; count++ )
                {
                const std::size_t index = get_history_index( count );
                const SVector2 & position = history[ index ].position;
                const Float32 perpendicular_distance = position * n + d;
                if( perpendicular_distance > 0.01f ) // seems like a nice number to me
                    {
                    // Not on the line.

                    break;
                    }
                const Float32 distance = ( position - p0 ).magnitude();
                if( distance > max_distance )
                    {
                    // Too far away.

                    break;
                    }
                }

            // Compute motion parameters.

            const std::size_t i2 = get_history_index( count - 1 );
            real_origin = history[ i2 ].position;
            real_zero_time = history[ i2 ].time;
            get_lm_velocity( real_velocity, real_origin, real_zero_time, p0, history[ i0 ].time );
            stopped = false;

#if PREDICTION_DEBUG >= 2

            {//sob
            std::ostringstream oss;
            oss << "Entity " << entity.get_object_id() << " prediction: using " <<
                        count << " positions from history.";

#if PREDICTION_DEBUG >= 3
            oss << "\norigin = " << real_origin.to_string() <<
                        ", zero_time = " << real_zero_time <<
                        ", velocity = " << real_velocity.to_string();
#endif // PREDICTION_DEBUG >= 3

            using namespace Massiv::Core;
            Massiv::Core::Global::log_debug( Status::FACILITY_GENERIC,
                        Status::PRIORITY_NORMAL, oss.str() );
            } //eob

#endif // PREDICTION_DEBUG >= 2
            }
        }

    // Check whether current predicted position (if prediction is valid)
    // if reasonably near to "real" current position. If not, entity teleported
    // or prediction error is too big.

    bool reset_prediction = !prediction_valid;
    SVector2 predicted_position; // valid iff prediction_valid

    if( prediction_valid )
        {
        // Check where entity should be according to estimated "real" path and
        // where it currently is according to previous path estimate.

        SVector2 real_position;
        get_lm_position( real_position, real_origin, real_zero_time,
                    real_velocity, current_time );
        get_lm_position( predicted_position, predicted_destination, predicted_destination_time,
                    predicted_velocity, current_time );

        // If those two positions are too far away, set reset_position to true,
        // which will "teleport" the entity directly to "real" path.

        const Float32 prediction_error = ( predicted_position - real_position ).square_magnitude();
        const Float32 zero_error = TILE_X_SIZE * TILE_Y_SIZE;
        if( ( prediction_error > zero_error ) &&
            ( prediction_error > ( real_velocity * 1.0f ).square_magnitude() ) )
            {
            reset_prediction = true;

#if PREDICTION_DEBUG >= 1

            {//sob
            std::ostringstream oss;
            oss << "Entity " << entity.get_object_id() << " prediction: resetting "
                        "because prediction error is too big.";

            using namespace Massiv::Core;
            Massiv::Core::Global::log_debug( Status::FACILITY_GENERIC,
                        Status::PRIORITY_NORMAL, oss.str() );
            } //eob

#endif // PREDICTION_DEBUG >= 1
            }
        }
    else
        {
        hope( reset_prediction );

#if PREDICTION_DEBUG >= 1

        {//sob
        std::ostringstream oss;
        oss << "Entity " << entity.get_object_id() << " prediction: resetting "
                    "because current prediction is invalid.";

        using namespace Massiv::Core;
        Massiv::Core::Global::log_debug( Status::FACILITY_GENERIC,
                    Status::PRIORITY_NORMAL, oss.str() );
        } //eob

#endif // PREDICTION_DEBUG >= 1
        }

    // Finally compute movement parameters. Start with #predicted_destination_time.
    // Next prediction update will happen at that time, or when entity replica is
    // changed. This time must be chosen carefully, espcially when reset_prediction
    // is false. In that case entity will move from current position to reach "real"
    // path at #predicted_destination_time. If #predicted_destination_time is too
    // close to current time, movement will be jerky, if it's high, it might "smooth"
    // "real" path too much, stride too far away and prediction error might became
    // too large (entity would be teleported to "real" path in that case). Moreover,
    // if entity is stopped, we want to move directly to its real position and stop
    // there.

    // If prediction is valid, entity is stopped and #predicted_destination_time is
    // in the future, do not change it (this ensures that we reach stopped position
    // in finite time). Otherwise set it to current_time + cross_delta. cross_delta
    // is computed above and is equal to 2 * estimated time between two updates, or
    // to prediction_limit if time between two updates is unknown. This still not
    // perfect (actually it's not even ok, but I don't care anymore...). Note that the
    // last condition should be always false, see first "if" far above, but I left
    // it here just to be safe.

    if( !prediction_valid || !stopped || predicted_destination_time <= current_time )
        {
        predicted_destination_time = current_time + cross_delta;
        }

    // Update next_update_time.

    if( next_update_time > predicted_destination_time )
        {
        next_update_time = predicted_destination_time;
        }

    // Now determine position we should reach at #predicted_destination_time. This point
    // is always on the "real" path.

    get_lm_position( predicted_destination, real_origin, real_zero_time,
                real_velocity, predicted_destination_time );

    // If current prediction is invalid or current position is too far away from
    // "real" position, teleport to "real" path immediately. Otherwise move
    // smoothly from current position to reach #predicted_destination at
    // #predicted_destination_time.

    if( reset_prediction )
        {
        predicted_velocity = real_velocity;
        }
    else
        {
        get_lm_velocity( predicted_velocity, predicted_position, current_time,
                    predicted_destination, predicted_destination_time );
        }

    prediction_valid = true;
    }


// class EntityManager

EntityManager::EntityManager()
      : registered_light_count( 0 ),
        admin_mode( false )
    {
    }


EntityManager::~EntityManager()
    {
    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.unregister_callback( &entity_replica_callback );
    }


void EntityManager::initialize()
    {
    // Create world rendering pass.

    Renderer::PassManager & pass_manager = Renderer::Global::get_pass_manager();
    pass_manager.create_render_manager( Renderer::PASS_WORLD );

    // Monitor client entity updates.

    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.register_callback
        (
        typeid( ::Demo::Lib::ClientEntity ),
        &entity_replica_callback,
        0,
        ::Massiv::Core::VariantPointer( this )
        );
    }


void EntityManager::update()
    {
    // Do nothing if rendering is disabled.

    if( Global::rendering_disabled )
        {
        return;
        }

    // Determine current time.

    const Time current_time( Time::current_time() );

    // Refresh the model database.

    TagDatabase & model_database = Global::get_model_database();
    const bool    model_database_updated = model_database.refresh();

    // Create entries for new entities and update prediction info if replication
    // update message has been received or next_update_time has been reached.
    // Also update entity visuals of all entities if the update was forced, the
    // terrain settings have changed or the model database was updated.

    if( Renderer::Global::get_terrain_settings().settings_changed( settings_change_time ) )
        {
        settings_change_time = SystemTime::current_time();
        force_entity_update();
        }

    if( model_database_updated )
        {
        force_entity_update();
        }

    if( !updated_entities.empty() || current_time >= next_update_time )
        {
        const bool update_was_forced = !next_update_time.is_valid();

        // Guesstimate next update time upper bound.

        next_update_time = current_time + 1.0f;

        // Create new entities.

        UpdateTimes::const_iterator it;
        for( it = updated_entities.begin(); it != updated_entities.end(); ++it )
            {
            const ObjectId & object_id = it->first;
            EntityInfos::iterator eit = entities.find( object_id );
            if( eit == entities.end() )
                {
                eit = entities.insert( EntityInfos::value_type
                            ( object_id, EntityInfo( object_id ) ) ).first;
                }
            }

        // Update all entities (including those not mentioned in recent replica update).

        EntityInfos::iterator eit;
        for( eit = entities.begin(); eit != entities.end(); ++eit )
            {
            const ObjectId & object_id = eit->first;
            EntityInfo & info = eit->second;

            Time update_time;
            it = updated_entities.find( object_id );
            if( it != updated_entities.end() )
                {
                update_time = it->second;
                }

            info.update( next_update_time, update_time, current_time,
                        settings.prediction_limit, visible_layer, 
                        update_was_forced );

            // Always update fog entity parameters. Because entity models are updated
            // automatically, we don't know when the pipeline state may change. Note
            // that any entity may (will) render non-fogged in the frame it reloads
            // it's model. The right thing to do would be to merge fogging parameters
            // from render manager settings in a way similar to lighting.

            info.update_fog();
            }

        // Done.

        updated_entities.clear();
        }

    // Move entities.

    EntityInfos::iterator eit;
    for( eit = entities.begin(); eit != entities.end(); ++eit )
        {
        EntityInfo & info = eit->second;
        info.move( settings.prediction, current_time );

        // Report new position to player manager.

        if( player.get_object_id() == eit->first )
            {
            Global::get_player_manager().move_player( info.position, info.orientation );
            }
        }

    // Update statistics.

    statistics.entity_count = entities.size();
    }


void EntityManager::set_visible_layer
    (
    const ObjectId & new_visible_layer
    )
    {
    if( visible_layer == new_visible_layer )
        {
        return;
        }
    visible_layer = new_visible_layer;

    // Force update of all entities.

    force_entity_update();
    }


void EntityManager::set_player
    (
    EntityPointer new_player
    )
    {
    player = new_player;

    const EntityInfos::const_iterator eit = entities.find( player.get_object_id() );
    if( eit != entities.end() )
        {
        const EntityInfo & info = eit->second;
        Global::get_player_manager().move_player( info.position, info.orientation );
        }
    }


void EntityManager::update_lights
    (
    std::size_t light_count
    )
    {
    struct LightInfo
        {
        LightPointer    light;
        SVector3        position;
        Float32         influence;
        };
    LightInfo linfos[ Renderer::MAX_LIGHT_SOURCES ];

    if( light_count > Renderer::MAX_LIGHT_SOURCES )
        {
        light_count = Renderer::MAX_LIGHT_SOURCES;
        }

    // Determine player position.

    SVector3 player_position;
    EntityInfos::const_iterator eit;
    eit = entities.find( player.get_object_id() );
    if( eit != entities.end() )
        {
        player_position = eit->second.position;
        }

    // Find light_count most influental source.
    // Influence = intensity at player position (pretty stupid but should be enough).

    std::size_t used_lights = 0;
    for( eit = entities.begin(); eit != entities.end(); ++eit )
        {
        const EntityInfo & entity = eit->second;

        // Skip non-light entities and entities from invisible layer.

        if( entity.light.is_null() || ( entity.layer_id != visible_layer ) )
            {
            continue;
            }

        // Determine light source position and influence.

        SVector3 position;
        Float32 influence;
        get_light_params( position, influence, entity, player_position );

        // Put it into set of most influential lights.

        std::size_t index = used_lights;
        if( index >= light_count )
            {
            // Compare it with least influential "visible" light.

            for( index = 0; index < light_count; index++ )
                {
                if( linfos[ index ].influence < influence )
                    {
                    break;
                    }
                }

            if( index == light_count )
                {
                // Not visible.

                continue;
                }
            }
        else
            {
            used_lights++;
            }
        linfos[ index ].light = entity.light;
        linfos[ index ].position = position;
        linfos[ index ].influence = influence;
        }

    // Update lights.

    std::size_t i;
    for( i = 0; i < used_lights; i++ )
        {
        Renderer::Light & light = lights[ i ];
        const LightInfo & li = linfos[ i ];
        light.set_position( li.position );
        light.set_infinite( false );
        light.set_ambient_intensity( li.light->get_ambient_intensity() );
        light.set_diffuse_intensity( li.light->get_diffuse_intensity() );
        light.set_specular_intensity( li.light->get_specular_intensity() );
        light.set_attenuation( li.light->get_attenuation() );
        }

    // Register/unregister lights.

    for( i = registered_light_count; i < used_lights; i++ )
        {
        lights[ i ].register_to_global_manager( Renderer::PASS_WORLD );
        }
    for( i = used_lights; i < registered_light_count; i++ )
        {
        lights[ i ].unregister_from_manager();
        }
    registered_light_count = used_lights;
    }


void EntityManager::force_entity_update()
    {
    next_update_time = Time();
    }


void EntityManager::toggle_admin_mode()
    {
    admin_mode = !admin_mode;
    force_entity_update();
    }


bool EntityManager::is_admin_mode() const
    {
    return admin_mode;
    }


bool EntityManager::cast_ray
    (
    RayIntersection & intersection
    )
    {
    Float32 best_distance;
    EntityInfos::const_iterator best_it = entities.end();
    EntityInfos::iterator it;
    for( it = entities.begin(); it != entities.end(); ++it )
        {
        // Skip entities from different layer.

        EntityInfo & entity = it->second;
        if( entity.layer_id != intersection.layer_id )
            {
            continue;
            }

        // Skip entities with no visual representation.

        if( !entity.visual.get() )
            {
            continue;
            }
        Renderer::ModelEntity & model = *entity.visual;

        // Transform ray into entity-space.
        // TODO: Implement transform structure to do these things?

        SVector3 origin = intersection.origin;
        origin -= model.get_position();
        model.get_orientation().inverse_rotate( origin );
        origin.inverse_scale( model.get_scale() );

        SVector3 destination = intersection.origin + intersection.direction;
        destination -= model.get_position();
        model.get_orientation().inverse_rotate( destination );
        destination.inverse_scale( model.get_scale() );

        // Intersect entity.

        Float32 distance;
        SVector3 normal;
        if( model.cast_ray( distance, normal, origin, destination - origin ) )
            {
            if( best_it == entities.end() || distance < best_distance )
                {
                best_it = it;
                best_distance = distance;
                intersection.normal = normal;
                intersection.entity = entity.entity;
                }
            }
        }

    // Fill in intersection structure if intersectio has been found.

    if( best_it != entities.end() )
        {
        // Compute intersection point.

        intersection.intersection = intersection.origin + intersection.direction * best_distance;
        intersection.distance = best_distance;

        // Transform intersection normal.

        hope( best_it->second.visual.get() );
        const Renderer::ModelEntity & model = *best_it->second.visual;
        intersection.normal.inverse_scale( model.get_scale() );
        model.get_orientation().rotate( intersection.normal );
        intersection.normal.normalize();
        }

    return ( best_it != entities.end() );
    }


void EntityManager::entity_replica_callback
    (
    ::Massiv::Core::ReplicaManager::UpdateReason            reason,
    ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
    int                                                     user_tag,
    ::Massiv::Core::VariantPointer                          user_data
    )
    {
    EntityManager * const self = ::Massiv::Core::variant_cast< EntityManager * >( user_data );
    switch( reason )
        {
        case ::Massiv::Core::ReplicaManager::UPDATE_RECEIVED:
        case ::Massiv::Core::ReplicaManager::MIGRATION_RECEIVED:
            {
            panic( "Unexpected reason." );
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_UPDATED:
            {
            self->entity_updated
                (
                replica.convert(),
                ::Massiv::Core::Global::replica_manager().get_source_node_time()
                );
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_GROUP_UPDATED:
            {
            // We don't care about anything pointing from/to entity.
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_DESTROY:
            {
            self->destroy_entity( replica.convert() );
            break;
            }
        }
    }


void EntityManager::entity_updated
    (
    EntityPointer   entity,
    const Time &    time
    )
    {
    updated_entities.insert( UpdateTimes::value_type( entity.get_object_id(), time ) );
    }


void EntityManager::destroy_entity
    (
    EntityPointer entity
    )
    {
    // Remove all info about the entity.

    const ObjectId & object_id = entity.get_object_id();
    updated_entities.erase( object_id );
    entities.erase( object_id );
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "entity_manager_inline.h"
#endif
