/** @file entity_manager.h
 * Manager of client entities.
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_ENTITY_MANAGER_H
#define CLIENT_ENTITY_MANAGER_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_LIGHT_H
#include "renderer/light.h"
#endif

#ifndef CLIENT_MODEL_ENTITY_H
#include "renderer/model_entity.h"
#endif

#ifndef MASSIV_REPLICA_MANAGER_H
#include "replication/replica_manager.h"
#endif

#ifndef MASSIV_STATISTICS_H
#include "registry/statistics.h"
#endif

#ifndef MASSIV_SYSTEM_TIME_H
#include "time/system_time.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * Manager of client entity replicas.
 *
 * @todo Improve prediction:
 * - Use simulation time instead of delivery system time.
 * - Support replicas from multiple servers (per-entity simulation time of last update).
 * - More information about motion in ClientEntity: velocity vector, more motion types,
 *   waypoints?
 */
class EntityManager
    {
    NO_COPY_CONSTRUCTOR( EntityManager );
    NO_ASSIGNMENT_OPERATOR( EntityManager );

public:

    /**
     * Entity manager statistics.
     * Stored in "Statistics/Client/EntityManager" node.
     */
    class MyStatistics : public ::Massiv::Core::Statistics
        {
    public:

        /**
         * Constructor.
         */
        MyStatistics();

    public:

        Integer &   entity_count;
            /**< Number of managed entities. */

        // TODO: prediction and replication statistics

        }; // class MyStatistics

        
    /**
     * Entity manager settings.
     * Stored in "Settings/Client/EntityManager" node.
     */
    class MySettings : public ::Massiv::Core::Settings
        {
    public:

        /**
         * Constructor.
         */
        MySettings();

    public:

        Boolean &   prediction;
            /**< Use prediction. */

        Float &     prediction_limit;
            /**< Entity will move using predicted path for up to #prediction_limit seconds,
                 then it will stop. */

        }; // class MySettings

public:

    /// ObjectId.
    typedef ::Massiv::Core::ObjectId ObjectId;

    /// Weak pointer to ClientEntity.
    typedef ::Massiv::Core::WeakPointer< ::Demo::Lib::ClientEntity > EntityPointer;

    /// Weak pointer to ClientLight.
    typedef ::Massiv::Core::WeakPointer< ::Demo::Lib::ClientLight > LightPointer;

    /// Simulation time.
    typedef STime Time;

public:

    /**
     * All client-side entity-related information.
     */
    class EntityInfo
        {
    private:

        /**
         * Position and time.
         */
        struct PositionTime
            {
        public:

            /**
             * Default constructor.
             */
            INLINE PositionTime();

            /**
             * Constructor.
             */
            INLINE PositionTime
                (
                const SVector2 &    the_position,
                const Time &        the_time
                );

        public:

            SVector2    position;
                /**< Entity position. */

            Time        time;
                /**< Simulation time of replica update.
                     Replica of entity at #position has been sent when server node simulation
                     time was #time. This means that #time is rough estimate - simulation times
                     may be out of sync and there is non-constant delay between position update
                     by entity physics and replication of the entity. */
            };

        enum
            {
            POSITION_HISTORY_LENGTH = 10,
                /**< Number of remembered PositionTime samples in position prediction history. */
            };

    public:

        /**
         * Default constructor.
         */
        INLINE EntityInfo();

        /**
         * Constructor.
         */
        explicit EntityInfo
            (
            EntityPointer the_entity
            );

        /**
         * Copy constructor.
         */
        EntityInfo
            (
            const EntityInfo & other
            );

        /**
         * Destructor.
         */
        INLINE ~EntityInfo();

    public:

        /**
         * Update entity information (orientation, prediction info, light) and
         * visibility from entity replica.
         *
         * This method will be called on all entities whenever at least on entity
         * replica is updated or @c current_time is greater than or equal to
         * @a next_update_time computed in previous update.
         * It will also be called when the update was forced explicitly. In that
         * case @c force is set to true.
         *
         * @retval next_update_time
         *      Time of the next update of all entities. Entity should changed this
         *      value if it wants to be updated sooner.
         * @param update_time
         *      Simulation time of most recent update, INVALID if entity replica has
         *      not been updated since last call to this method.
         * @param current_time
         *      Current simulation time.
         * @param prediction_limit
         *      Assume something "weird" happened if most recent update is older than
         *      @a prediction_limit seconds.
         * @param force
         *      Was entity updated forced explicitly?
         */
        void update
            (
            Time &              next_update_time,
            const Time &        update_time,
            const Time &        current_time,
            Float32             prediction_limit,
            const ObjectId &    visible_layer,
            bool                force
            );

        /**
         * Move entity using prediction info or current replica position.
         */
        void move
            (
            bool            use_prediction,
            const Time &    current_time
            );

        /**
         * Update fog settings of entity visual from global terrain settings.
         */
        void update_fog();

    private:

        /**
         * Update #visual.
         *
         * #Update_visual() determines model tag set and model data object name for
         * current entity state. If model name is different from current #model_name,
         * new #visual will be created. It may destroy #visual if current entity state
         * is "non-visual entity". If it creates new #visual, it will be invisible.
         *
         * The method tries to optimize access to the database and caches the previous
         * result. If the model database is updated, all entities must be forced
         * to update their visual objects as well (by setting @c force to true).
         */
        void update_visual
            (
            bool force
            );

        /**
         * Update prediction information for generic moving entity.
         */
        void update_generic_prediction
            (
            Time &          next_update_time,
            bool            force_update,
            const Time &    current_time,
            Float32         prediction_limit
            );

    private:

        /**
         * Get index into to the #history cyclic buffer pointing to @a index-th
         * most recent element.
         */
        INLINE std::size_t get_history_index
            (
            std::size_t index
            ) const;

    public:

        EntityPointer                                   entity;
            /**< Pointer to entity replica. */

        LightPointer                                    light;
            /**< Pointer to light replica. May be @c NULL. */

        mutable std::auto_ptr< Renderer::ModelEntity >  visual;
            /**< Visual representation of the entity.
                 It may be @c NULL if entity has no visual representation. */

        std::string                                     model_name;
            /**< Model name used to create current #visual. */

    public:

        ObjectId        layer_id;
            /**< Id of layer this entity is in. */

        bool            should_be_visible;
            /**< Entity should be visible if @c true, but the visual may be invisible if
                 we are unable to determine current position. Visual is always invisible
                 if #should_be_visible is @c false. */

    public:

        SVector3        position;
            /**< Current entity position. */

        SOrientation    orientation;
            /**< Current entity orientation. */

        ObjectId        sector_id;
            /**< Current entity sector. */

    private:

        /** @name Position history.
         * History of entity replica position. It's used to gues current motion.
         */
        //@{
        PositionTime    history[ POSITION_HISTORY_LENGTH ];
            /**< History of entity position - cyclic buffer.
                 Used by position prediction. */

        std::size_t     history_base_index;
            /**< Index of the oldest element in #history.
                 Index of the second oldest element (if it exists) is #history_bas_index + 1
                 (modulo cyclic buffer length). */

        std::size_t     history_size;
            /**< Number of valid elements in #history. */
        //@}

    private:

        /** @name Predicted movement.
         * Parameters of predicted linear movement.
         * Entity is expected to move along the line specified by #predicted_destination
         * point and #predicted_velocity vector, and reach #predicted_destination when
         * simulation time is equal to #predicted_destination_time. Predicted movement
         * parameters are updated whenever entity replica is changed or #predicted_destination
         * is reached.
         */
        //@{
        bool            prediction_valid;
            /**< Is predicted movement valid? */

        SVector2        predicted_destination;
            /**< Prediction destination coordinates. */

        Time            predicted_destination_time;
            /**< Entity should reach #predicted_destination at this time. */

        SVector2        predicted_velocity;
            /**< Velocity of predicted movement. */
        //@}

    private:

        SOrientation    final_orientation;
            /**< Rotate #orientation towards #final_orientation. */

        Time            previous_move_time;
            /**< Time of previous move(). */

    private:

        std::string     previous_model_tags;
            /**< Previous model tags. */

        }; // class EntityInfo

private:

    /// Map from object id to entity info.
    typedef ext::hash_map< ObjectId, EntityInfo, ObjectId::Hash > EntityInfos;

    /// Map from object id to simulation time.
    typedef ext::hash_map< ObjectId, Time, ObjectId::Hash > UpdateTimes;

public:

    /**
     * Parameters and results of #cast_ray().
     */
    struct RayIntersection
        {
    // in
    public:
        ObjectId        layer_id;
            /** Object id of layer. */

        SVector3        origin;
            /**< Ray origin. */

        SVector3        direction;
            /**< Ray direction. */

    // out
    public:

        SVector3        intersection;
            /**< Intersection point. */

        Float32         distance;
            /**< Distance of #intersection point from #origin, in units specified
                 by length of #direction vector. */

        SVector3        normal;
            /**< Normal of intersected triangle. */

        EntityPointer   entity;
            /**< Pointer to intersected entity. */

        }; // struct RayIntersection

public:

    /**
     * Constructor.
     */
    EntityManager();

    /**
     * Destructor.
     */
    ~EntityManager();

    /**
     * Initialize entity manager.
     */
    void initialize();

    /**
     * Update entities.
     *
     * This should be called after #SectorManager::update().
     */
    void update();

public:

    /**
     * Get id of visible layer.
     */
    INLINE const ObjectId & get_visible_layer() const;

    /**
     * Set id of visible layer.
     *
     * This method should be called only by #PlayerManager::update().
     * Visibility of entities will be changed by next call to #update().
     */
    void set_visible_layer
        (
        const ObjectId & new_visible_layer
        );

    /**
     * Set pointer to current player entity.
     */
    void set_player
        (
        EntityPointer new_player
        );

public:

    /**
     * Find, update and register @a light_count most "visible" lights.
     */
    void update_lights
        (
        std::size_t light_count
        );

    /**
     * Forces update of all entities, including their visuals.
     *
     * Must be called whenever the model database is updated. Otherwise,
     * static entities (decorations), that do not move, would have their
     * visuals never updated.
     */
    void force_entity_update();
    
    /**
     * Toggles admin mode.
     */
    void toggle_admin_mode();

    /**
     * Tests if admin mode is enabled.
     */
    bool is_admin_mode() const;

public:

    /**
     * Intersect ray with entities in given layer.
     *
     * The following attributes of @a intersection should be initialized when
     * calling this method:
     * - layer_id:  Object id of layer to search.
     * - origin:    Ray origin.
     * - direction: Ray direction.
     *
     * If intersection is found, @c true will be returned and @a intersection
     * updated (see #RayIntersection). Otherwise @c false will be returned.
     */
    bool cast_ray
        (
        RayIntersection & intersection
        );

private:

    /**
     * Entity replica update callback.
     */
    static void entity_replica_callback
        (
        ::Massiv::Core::ReplicaManager::UpdateReason            reason,
        ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
        int                                                     user_tag,
        ::Massiv::Core::VariantPointer                          user_data
        );

    /**
     * Inform entity manager about new entity or entity replica update.
     */
    void entity_updated
        (
        EntityPointer   entity,
        const Time &    time
        );

    /**
     * Inform entity manager that entity replica will be destroyed.
     */
    void destroy_entity
        (
        EntityPointer entity
        );

private:

    MyStatistics        statistics;
        /**< Statistics. */

    MySettings          settings;
        /**< Settings. */

private:

    ObjectId            visible_layer;
        /**< Id of visible layer. */

    EntityInfos         entities;
        /**< Replicated entities. */

    UpdateTimes         updated_entities;
        /**< Object ids of updated client entities and times of their updates. */

    EntityPointer       player;
        /**< Player entity.
             EntityManager will inform PlayerManager about current predicted
             position of this entity. */

    Renderer::Light     lights[ Renderer::MAX_LIGHT_SOURCES ];
        /**< Light sources nearest to #player. */

    std::size_t         registered_light_count;
        /**< Number of #lights registered to render manager. */

    Time                next_update_time;
        /**< Update entities at this time (or when any replica is updates). */

    SystemTime          settings_change_time;
        /**< Time of last change of terrain settings. */
    
    bool                admin_mode;
        /**< Is admin mode enabled? If true some invisible entities will be
             drawn too. */

    }; // class EntityManager


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "entity_manager_inline.h"
#endif


#endif // CLIENT_ENTITY_MANAGER_H
