/** @file game_context.h
 * Game context.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef CLIENT_GAME_CONTEXT_H
#define CLIENT_GAME_CONTEXT_H 1


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif


namespace Demo {
namespace Client {

INJECT_CORE_SYMBOLS();

/**
 * GameContext object holds miscenalleous information about
 * connection of the client to game simulation
 * including pointer to client's account object or currently
 * selected player.
 */
class GameContext
    {
    NO_COPY_CONSTRUCTOR( GameContext );
    NO_ASSIGNMENT_OPERATOR( GameContext );

public:

    /**
     * Constructor.
     */
    GameContext();
    
public:

    /**
     * Is the game context linked with account object of the client?
     * If returns true, @c get_account_object() will return non-null pointer.
     */
    bool is_linked_with_account() const;

    /**
     * Link the game context with specified account object.
     * Pass @c null to unlink current account object.
     */
    void link_with_account
        (
        WeakPointer< Lib::AccountObject > account
        );
    
    /**
     * Get account object linked with the game context.
     * This function always return non-null pointer.
     * If there is no account object linked with the game context,
     * the function calls @c panic() (may throw an exception
     * in future).
     */
    Remote< Lib::AccountObject > require_account() const;

    /**
     * Get account object linked with the game context.
     * This function returns null pointer if the game context is not
     * linked with any account object.
     */
    Remote< Lib::AccountObject > get_account() const;
    
public:

    /**
     * Is some player object selected?
     */
    bool is_player_selected() const;
    
    /**
     * Set selected player pointer to @a player.
     * @param player Pointer to player that was selected.
     *      Null if player was deselected.
     */
    void player_selected
        (
        WeakPointer< Lib::ClientEntity > player
        );
    
    /**
     * Get pointer to currently selected player.
     * Returns null pointer if no player is selected.
     */
    WeakPointer< Lib::ClientEntity > get_selected_player() const;
    
private:

    PRemote< Lib::AccountObject >       account;
        /**< Account object the client is linked with. */
    
    PWeakPointer< Lib::ClientEntity >   selected_player;
        /**< Selected player. Null if no player is selected. */

    }; // GameContext


} // namespace Client
} // namespace Demo

#endif // CLIENT_GAME_CONTEXT_H
