/** @file game_input.cpp
 * Handler of keyboard input not eaten by GUI (implementation).
 */


/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "game/game_input.h"
#include "input/input_manager.h"
#include "gui/client_chat_console.h"
#include "gui/client_full_console.h"
#include "gui/game_interface.h"
#include "gui/editor.h"
#include "renderer/renderer.h"
#include "renderer/terrain_settings.h"
#include "client_global.h"
#include "lib/shared/account_object.h"
#include "lib/shared/account_object_rpc.h"
#include "game/game_context.h"


namespace Demo {
namespace Client {


GameInput::GameInput()
      : should_quit( false )
    {
    }


GameInput::~GameInput()
    {
    }


bool GameInput::quit_requested() const
    {
    return should_quit;
    }


bool GameInput::key_up
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    return false;
    }


bool GameInput::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    // Note: camera controls are handled by CameraControl.

    switch( key )
        {
        case 'w':
            {
            // Enables/disables draw_wireframe.

            bool new_value = !Renderer::Global::get_terrain_settings().draw_wireframe;
            Renderer::Global::get_terrain_settings().draw_wireframe = new_value;

            FullConsole & console = Global::get_full_console();
            std::ostringstream oss;
            oss << "Renderer: draw_wireframe set to " << new_value << ".";

            console.print( oss.str() );

            return true;
            }

        case 'f':
            {
            // Enables/disables fog.

            bool new_value = !Renderer::Global::get_terrain_settings().enable_fog;
            Renderer::Global::get_terrain_settings().enable_fog = new_value;

            FullConsole & console = Global::get_full_console();
            std::ostringstream oss;
            oss << "Renderer: enable_fog set to " << new_value << ".";

            console.print( oss.str() );

            return true;
            }

        case 'x':
            {
            // Enables/disables textures.

            bool new_value = !Renderer::Global::get_terrain_settings().enable_texture;
            Renderer::Global::get_terrain_settings().enable_texture = new_value;

            FullConsole & console = Global::get_full_console();
            std::ostringstream oss;
            oss << "Renderer: enable_texture set to " << new_value << ".";

            console.print( oss.str() );

            return true;
            }

        case 'l':
            {
            // Enables/disables light.

            bool new_value = !Renderer::Global::get_terrain_settings().enable_light;
            Renderer::Global::get_terrain_settings().enable_light = new_value;

            FullConsole & console = Global::get_full_console();
            std::ostringstream oss;
            oss << "Renderer: enable_light set to " << new_value << ".";

            console.print( oss.str() );

            return true;
            }

        case 'm':
            {
            // Enables/disables tile marks.

            bool new_value = !Renderer::Global::get_terrain_settings().draw_marks;
            Renderer::Global::get_terrain_settings().draw_marks = new_value;

            FullConsole & console = Global::get_full_console();
            std::ostringstream oss;
            oss << "Renderer: draw_marks set to " << new_value << ".";

            console.print( oss.str() );

            return true;
            }

        case 'p':
            {
            // Enables/disables path finding for the selected player.

            Massiv::Core::Remote< Lib::AccountObject > account =
                Global::get_game_context().get_account();

            if( account != null )
                {
                Lib::Console::Environment env;
                account->async_process_account_command
                    ( "/toggle_path_finding", env );
                }

            break;
            }

        case 't':
            {
            // Set focus to chat console.

            ChatConsole & console = Global::get_chat_console();
            console.set_focus();
            return true;
            }

        case 'i':
            {
            // Toggle visibility of inventory.

            GameInterface & game_interface = Global::get_game_interface();
            game_interface.toggle_panel( GameInterface::PANEL_INVENTORY );
            return true;
            }

        case 'a':
            {
            // Toggle visibility of abilities.

            GameInterface & game_interface = Global::get_game_interface();
            game_interface.toggle_panel( GameInterface::PANEL_ABILITIES );
            return true;
            }

        case 'e':
            {
            // (Show) and set focus to editor.

            Editor & editor = Global::get_editor();
            if( editor.is_visible() )
                {
                editor.update_visibility();
                }
            else
                {
                editor.set_visibility( true );
                }
            return true;
            }

        case 'h':
            {
            // Show/hide help.

            if( !help_message.get() )
                {
                create_help_message();
                }
            else
                {
                help_message->set_visibility( !help_message->is_visible() );
                }
            if( help_message->is_visible() )
                {
                help_message->set_focus();
                }
            return true;
            }

        case 'c':
            {
            // Show/hide credits message.

            if( !credits_message.get() )
                {
                create_credits_message();
                }
            else
                {
                credits_message->set_visibility( !credits_message->is_visible() );
                }
            if( credits_message->is_visible() )
                {
                credits_message->set_focus();
                }
            return true;
            }

        case 'q':
            {
            // Show quit requester.

            if( !quit_requester.get() )
                {
                create_quit_requester();
                }
            BlockerWidget * const blocker = quit_requester->switch_to_modal_mode();
            hope( blocker );
            blocker->set_base_color( SColor( 0.1f, 0.1f, 0.3f, 1.0f ) );
            blocker->set_opacity( 0.3f );
            return true;
            }

        case 'r':
            {
            if( mod & KMOD_CTRL )
                {
                // Show "disable rendering" requester.

                if( !disable_rendering_requester.get() )
                    {
                    create_disable_rendering_requester();
                    }
                BlockerWidget * const blocker = disable_rendering_requester->switch_to_modal_mode();
                hope( blocker );
                blocker->set_base_color( SColor( 0.1f, 0.1f, 0.3f, 1.0f ) );
                blocker->set_opacity( 0.3f );
                }
            return true;
            }

        case 'v':
            {
            // Show/hide invisible entities (admin mode).

            Global::get_entity_manager().toggle_admin_mode();
            return true;
            }

        default:
            break;
        }

    return false;
    }


void GameInput::message_box_action
    (
    MessageBox * message_box
    )
    {
    hope( message_box );
    if( message_box ==  quit_requester.get() )
        {
        const MessageBox::Result result = quit_requester->get_result();
        if( result == MessageBox::RESULT_OK )
            {
            should_quit = true;
            }
        quit_requester->switch_to_modeless_mode();
        quit_requester->set_visibility( false );
        }
    else if( message_box ==  disable_rendering_requester.get() )
        {
        const MessageBox::Result result = disable_rendering_requester->get_result();
        disable_rendering_requester->switch_to_modeless_mode();
        disable_rendering_requester->set_visibility( false );
        if( result == MessageBox::RESULT_OK )
            {
            Global::disable_rendering();
            }
        }
    else if( message_box == help_message.get() )
        {
        help_message->set_visibility( false );
        }
    else if( message_box == credits_message.get() )
        {
        credits_message->set_visibility( false );
        }
    message_box->clear_result();
    }


void GameInput::create_credits_message()
    {
    credits_message.reset( new MessageBox() );
    Renderer::TextRectangle & text = credits_message->get_message_object();
    text.set_columns_rows( 51, 11 );
    text.set_size_from_font();

    // Print everything in neutral color.

    text.set_text_color( SColor( 1.0f, 1.0f, 1.0f ) );
    //                  0         1         2         3         4         5         6         7
    //                  01234567890123456789012345678901234567890123456789012345678901234567890123456789
    //xt.print( 0,  0, "                           Massiv Demo Credits                          " );
    text.print( 0,  2, "" );
    text.print( 18,  3, "- stoupik@users.sourceforge.net" );
    text.print( 18,  4, "- markoid@users.sourceforge.net" );
    text.print( 18,  5, "- boovie@users.sourceforge.net" );
    text.print( 18,  6, "- octa@users.sourceforge.net" );
    text.print( 18,  7, "- marekus@users.sourceforge.net" );
    text.print( 18,  8, "- hafik@users.sourceforge.net" );


    // Colors.

    text.set_text_color( SColor( 1.0f, 0.8f, 0.2f ) );
    text.print( 16,  0, "Massiv Demo Credits" );

    text.set_text_color( SColor( 1.0f, 0.8f, 0.4f ) );
    text.print( 16,  1, "Copyright 2000-2004" );

    text.set_text_color( SColor( 0.2f, 1.0f, 0.4f ) );
    text.print( 0, 3, "Stepan Vondrak" );
    text.print( 0, 4, "Marek Vondrak" );
    text.print( 0, 5, "Petr Tovarys" );
    text.print( 0, 6, "Ondrej Pecta" );
    text.print( 0, 7, "Marek Svantner" );
    text.print( 0, 8, "Martin Havlista" );

    text.set_text_color( SColor( 0.0f, 0.5f, 1.0f ) );
    text.print( 19, 10, "MFF UK Praha" );

    credits_message->set_message_receiver( this );
    credits_message->do_layout();
    credits_message->center_object();
    credits_message->set_dispatcher_as_parent();
    }


void GameInput::create_help_message()
    {
    help_message.reset( new MessageBox() );
    Renderer::TextRectangle & text = help_message->get_message_object();
    text.set_columns_rows( 72, 19 );
    text.set_size_from_font();

    // Print everything in neutral color.

    text.set_text_color( SColor( 1.0f, 1.0f, 1.0f ) );
    //                  0         1         2         3         4         5         6         7
    //                  01234567890123456789012345678901234567890123456789012345678901234567890123456789
    //xt.print( 0,  0, "                         Massiv Demo Quick Help                         " );
    text.print( 0,  1, "" );
    text.print( 0,  2, " t ... Activate chat console." );
    text.print( 0,  3, " ~ ... Activate full-screen console." );
    text.print( 0,  4, " i ... Show/hide inventory panel." );
    text.print( 0,  5, " a ... Show/hide abilities panel." );
    text.print( 0,  6, " e ... Activate editor. Press escape to close active editor window." );
    text.print( 0,  7, "       Only privileged users are allowed to modify terrain and entities." );
    text.print( 0,  8, " p ... Enable/disable path finding." );
    text.print( 0,  9, " w ... Toggles wireframe." );
    text.print( 0,  10, " x ... Toggles textures." );
    text.print( 0,  11, " l ... Toggles light." );
    text.print( 0,  12, " f ... Toggles fog." );
    text.print( 0,  13, " m ... Toggles entity mark." );
    text.print( 0,  14, " c ... Show credits." );
    text.print( 0,  15, " v ... Show/hide invisible entities." );
    text.print( 0,  16, " h ... Show/hide this help." );
    text.print( 0,  17, " q ... Quit." );
    text.print( 0, 17, "" );
    text.print( 0, 18, " Use arrows to rotate camera when main window has focus." );

    // Colors.

    text.set_text_color( SColor( 1.0f, 0.8f, 0.2f ) );
    text.print( 0,  0, "                         Massiv Demo Quick Help                         " );
    text.set_text_color( SColor( 0.2f, 1.0f, 0.4f ) );
    text.print( 1, 2, "t" );
    text.print( 1, 3, "~" );
    text.print( 1, 4, "i" );
    text.print( 1, 5, "a" );
    text.print( 1, 6, "e" );
    text.print( 1, 8, "p" );
    text.print( 1, 9, "w" );
    text.print( 1, 10, "x" );
    text.print( 1, 11, "l" );
    text.print( 1, 12, "f" );
    text.print( 1, 13, "m" );
    text.print( 1, 14, "c" );
    text.print( 1, 15, "v" );
    text.print( 1, 16, "h" );
    text.print( 1, 17, "q" );

    help_message->set_message_receiver( this );
    help_message->do_layout();
    help_message->center_object();
    help_message->set_dispatcher_as_parent();
    }


void GameInput::create_quit_requester()
    {
    quit_requester.reset( new MessageBox( MessageBox::FLAG_TWO_BUTTONS, "  Yes  ", "  No  " ) );
    quit_requester->set_message( "  Do you really want to quit?  \n\n" );
    quit_requester->set_message_receiver( this );
    quit_requester->do_layout();
    quit_requester->center_object();
    quit_requester->set_dispatcher_as_parent();
    }


void GameInput::create_disable_rendering_requester()
    {
    disable_rendering_requester.reset( new MessageBox( MessageBox::FLAG_TWO_BUTTONS, "  Yes  ", "  No  " ) );
    disable_rendering_requester->set_message( "  Do you really want to disable screen updates?  \n\n  Press CTRL+R to enable it again  \n\n" );
    disable_rendering_requester->set_message_receiver( this );
    disable_rendering_requester->do_layout();
    disable_rendering_requester->center_object();
    disable_rendering_requester->set_dispatcher_as_parent();
    }


} // namespace Client
} // namespace Demo
