/** @file player_manager.cpp
 * Manager of active player entity (implementation).
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "game/player_manager.h"
#include "client_global.h"
#include "renderer/renderer.h"
#include "renderer/pass_manager.h"
#include "renderer/passes.h"
#include "renderer/camera.h"
#include "game/entity_manager.h"
#include "game/sector_manager.h"
#include "lib/shared/client_entity.h"


namespace Demo {
namespace Client {


PlayerManager::PlayerManager()
    {
    }


PlayerManager::~PlayerManager()
    {
    // Stop replica tracking.

    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.unregister_callback( &player_replica_callback );

    // Unregister world-pass camera.

    Renderer::PassManager & pass_manager = Renderer::Global::get_pass_manager();
    if( pass_manager.has_render_manager( Renderer::PASS_WORLD ) )
        {
        pass_manager.get_render_manager( Renderer::PASS_WORLD ).set_camera( NULL );
        }
    }


void PlayerManager::initialize()
    {
    // Create camera in world rendering pass.

    Renderer::PassManager & pass_manager = Renderer::Global::get_pass_manager();
    Renderer::RenderManager & render_manager =
                pass_manager.create_render_manager( Renderer::PASS_WORLD );
    camera.reset( new Renderer::CirclingCamera );
    camera->set_latitude( MATHF_PI_4 );
    camera->set_longitude( 0.0f );
    camera->set_distance( 5.0f );
    render_manager.set_camera( camera.get() );
    }


void PlayerManager::set_player
    (
    EntityPointer new_player
    )
    {
    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();

    // Unregister old callback.

    replica_manager.unregister_callback( &player_replica_callback );
    set_next_visible_layer( ObjectId() );

    // Register new callback.

    player = new_player;
    if( !player.is_null() )
        {
        replica_manager.register_callback
            (
            player.get_object_id(),
            &player_replica_callback,
            0,
            ::Massiv::Core::VariantPointer( this )
            );
        if( player.is_local() )
            {
            set_next_visible_layer( player->current_position.layer.get_object_id() );
            }
        }

    // Inform entity manager about the change.

    Global::get_entity_manager().set_player( player );
    }


void PlayerManager::update()
    {
    // Switch layers?

    if( next_layer_id != visible_layer_id )
        {
        std::ostringstream ss;
        ss << "Switching active layer from " << visible_layer_id << " to " << next_layer_id << ".";
        Massiv::Core::Global::log_info( Massiv::Core::Status::FACILITY_GENERIC,
            Massiv::Core::Status::PRIORITY_NORMAL, ss.str() );
        
        visible_layer_id = next_layer_id;
        Global::get_sector_manager().set_visible_layer( visible_layer_id );
        Global::get_entity_manager().set_visible_layer( visible_layer_id );
        }

    // SectorManager::update() will update sectors and rebuild terrains.

    // EntityManager::update() will move the entities, including player, and
    // report new player position back.
    }


void PlayerManager::move_player
    (
    const SVector3 &        new_position,
    const SOrientation &    new_orientation
    )
    {
    position = new_position;
    orientation = new_orientation;

    hope( camera.get() );
    camera->set_target( position );
    }


void PlayerManager::player_replica_callback
    (
    ::Massiv::Core::ReplicaManager::UpdateReason            reason,
    ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
    int                                                     user_tag,
    ::Massiv::Core::VariantPointer                          user_data
    )
    {
    PlayerManager * const self = ::Massiv::Core::variant_cast< PlayerManager * >( user_data );

    switch( reason )
        {
        case ::Massiv::Core::ReplicaManager::UPDATE_RECEIVED:
        case ::Massiv::Core::ReplicaManager::MIGRATION_RECEIVED:
            {
            panic( "Unexpected reason." );
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_UPDATED:
            {
            PlayerManager::EntityPointer entity = replica.convert();
            const PlayerManager::ObjectId & layer_id =
                entity->current_position.layer.get_object_id();
            self->set_next_visible_layer( layer_id );
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_GROUP_UPDATED:
            {
            // We don't care about anything pointing from/to player.
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_DESTROY:
            {
            self->set_next_visible_layer( PlayerManager::ObjectId() );
            break;
            }
        }
    }


void PlayerManager::set_next_visible_layer
    (
    const ObjectId & next_visible_layer
    )
    {
    next_layer_id = next_visible_layer;
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "player_manager_inline.h"
#endif
