/** @file player_manager.h
 * Manager of active player entity.
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_PLAYER_MANAGER_H
#define CLIENT_PLAYER_MANAGER_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef MASSIV_REPLICA_MANAGER_H
#include "replication/replica_manager.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * Manager of active player client entity.
 *
 * It keeps track of current layer, camera position and orientation, etc.
 */
class PlayerManager
    {
    NO_COPY_CONSTRUCTOR( PlayerManager );
    NO_ASSIGNMENT_OPERATOR( PlayerManager );

private:

    /// ObjectId.
    typedef ::Massiv::Core::ObjectId ObjectId;

    /// Weak pointer to ClientEntity.
    typedef ::Massiv::Core::WeakPointer< ::Demo::Lib::ClientEntity > EntityPointer;

public:

    /**
     * Constructor.
     */
    PlayerManager();

    /**
     * Destructor.
     */
    ~PlayerManager();

public:

    /**
     * Initialize player manager.
     */
    void initialize();

    /**
     * Set pointer to current player entity.
     *
     * PlayerManager will start monitoring player replica and track current
     * layer id. You can call #get_visible_layer() to determine current layer.
     *
     * It will also inform EntityManager about current player id and request
     * player position change notifications from EntityManager.
     *
     * Use @c null @a new_player if there is no active player entity.
     */
    void set_player
        (
        EntityPointer new_player
        );

public:

    /**
     * Update player manager.
     *
     * #update() may be called once per tick or less often, it must be followed by
     * calls to #SectorManager::update() and #EntityManager::update().
     *
     * @note This method will only update current layer id. Player position will be
     *       updated by the entity manager and reported back to player manager.
     */
    void update();

public:

    /**
     * Get pointer to current player entity.
     */
    INLINE EntityPointer get_player() const;

    /**
     * Get current player's position.
     */
    INLINE const SVector3 & get_player_position() const;
    
    /**
     * Get current player's orientation.
     */
    INLINE const SOrientation & get_player_orientation() const;
    
    /**
     * Get id of visible layer.
     */
    INLINE const ObjectId & get_visible_layer() const;

    /**
     * Get reference to current camera.
     */
    INLINE Renderer::CirclingCamera & get_camera();

public:

    /**
     * Inform player manager about new player entity position.
     *
     * This method should be called only by EntityManager. @a New_position
     * and @a new_orientation may be predicted values.
     */
    void move_player
        (
        const SVector3 &        new_position,
        const SOrientation &    new_orientation
        );

private:

    /**
     * Player replica update callback.
     */
    static void player_replica_callback
        (
        ::Massiv::Core::ReplicaManager::UpdateReason            reason,
        ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
        int                                                     user_tag,
        ::Massiv::Core::VariantPointer                          user_data
        );
        
    /**
     * Set id of layer that should be made visible by next #update().
     */
    void set_next_visible_layer
        (
        const ObjectId & next_visible_layer
        );

private:

    EntityPointer                               player;
        /**< Pointer to player entity. */

    ObjectId                                    visible_layer_id;
        /**< Id of currently visible layer. */

    ObjectId                                    next_layer_id;
        /**< Id of layer to make visible by #update() call. */

private:

    SVector3                                    position;
        /**< Player layer-relative position. */

    SOrientation                                orientation;
        /**< Player orientation. */

private:

    std::auto_ptr< Renderer::CirclingCamera >   camera;
        /**< Camera. */

    }; // class PlayerManager


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "player_manager_inline.h"
#endif


#endif // CLIENT_PLAYER_MANAGER_H
