/** @file sector_manager.cpp
 * Manager of sector replicas and terrains (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>

#ifdef COMPILER_MSVC
namespace std { using ::abs; using ::size_t; } // FIXME: VERY NASTY MSVC6-related HACK
#endif

#include "game/sector_manager.h"
#include "renderer/renderer.h"
#include "renderer/terrain_settings.h"
#include "renderer/terrain_material.h"
#include "renderer/pass_manager.h"
#include "renderer/passes.h"
#include "renderer/light.h" // hack - should be in LightManager
#include "client_global.h"
#include "raycast.h"
#include "lib/shared/client_sector.h"
#include "lib/shared/client_entity.h"


namespace Demo {
namespace Client {

namespace {


//@{
/// Import stuff from ClientSector.
const Float32 SECTOR_X_SIZE = static_cast< Float32 >( ::Demo::Lib::ClientSector::SECTOR_X_SIZE );
const Float32 SECTOR_Y_SIZE = static_cast< Float32 >( ::Demo::Lib::ClientSector::SECTOR_Y_SIZE );
const int NUM_X_TILES = ::Demo::Lib::ClientSector::NUM_X_TILES;
const int NUM_Y_TILES = ::Demo::Lib::ClientSector::NUM_Y_TILES;
const Float32 TILE_X_SIZE = SECTOR_X_SIZE / static_cast< Float32 >( NUM_X_TILES );
const Float32 TILE_Y_SIZE = SECTOR_Y_SIZE / static_cast< Float32 >( NUM_Y_TILES );
//@}


/**
 * Get client sector vertex height.
 */
Float32 get_sector_height
    (
    const SectorManager::SectorPointer &    sector,
    int                                     x,
    int                                     y
    )
    {
    hope( x >= 0 );
    hope( x < NUM_X_TILES );
    hope( y >= 0 );
    hope( y < NUM_Y_TILES ) ;
    const int index = y * NUM_X_TILES + x;
    return sector->elevation_map[ index ] + sector->delta_elevation_map[ index ];
    }


/**
 * Get client sector vertex material.
 */
Int32 get_sector_material
    (
    const SectorManager::SectorPointer &    sector,
    int                                     x,
    int                                     y
    )
    {
    hope( x >= 0 );
    hope( x < NUM_X_TILES );
    hope( y >= 0 );
    hope( y < NUM_Y_TILES ) ;
    const int index = y * NUM_X_TILES + x;
    return sector->material_map[ index ];
    }


/**
 * Get client sector tile flags.
 */
Int32 get_sector_tile_flags
    (
    SectorManager::SectorPointer    sector,
    int                             x,
    int                             y
    )
    {
    hope( x >= 0 );
    hope( x < NUM_X_TILES );
    hope( y >= 0 );
    hope( y < NUM_Y_TILES ) ;
    const int index = y * NUM_X_TILES + x;
    return sector->tile_flag_map[ index ];
    }


/**
 * Pointer to sector and its 8 neighbors.
 *
 * First index is in south-\>north direction (!), second index west-\>east,
 * sector itself is at index (1, 1).
 */
typedef SectorManager::SectorPointer SectorNeighborhood[ 3 ][ 3 ];

/**
 * Convert sector-relative vertex/tile address that may be "slightly"
 * out-of-bounds to valid address.
 *
 * If address points to non-existing neighbor (@a neighbors pointer is @c NULL),
 * address of "nearest" vertex/tile will be returned. "Nearest address" search
 * is consistent. It will be always the same for the same address, it does not
 * matter relatively to which sector the address has been specified.
 *
 * @retval neighbor_x   @a x index into @a neighbors.
 * @retval neighbor_y   @a y index into @a neighbors.
 * @retval tile_x       Tile/vertex @a x coordinate relative to returned neighbor.
 * @retval tile_y       Tile/vertex @a y coordinate relative to returned neighbor.
 * @param  neighbors    Pointers to all sector replicas in 3x3 area.
 * @param  x            Tile/vertex @a x coordinate relative to <tt>neighbors[ 1 ][ 1 ]</tt>.
 * @param  y            Tile/vertex @a y coordinate relative to <tt>neighbors[ 1 ][ 1 ]</tt>.
 */
void get_neighbordhood_address
    (
    std::size_t &           neighbor_x,
    std::size_t &           neighbor_y,
    int &                   tile_x,
    int &                   tile_y,
    SectorNeighborhood &    neighbors,
    int                     x,
    int                     y
    )
    {
    // Check inputs. "slightly" out of bounds is defined as +/-2 vertices/tiles.

    hope( !neighbors[ 1 ][ 1 ].is_null() );
    hope( x >= -2 );
    hope( x < NUM_X_TILES + 2 );
    hope( y >= -2 );
    hope( y < NUM_Y_TILES + 2 );

    // Get valid sector address.

    std::size_t nx = 1;
    std::size_t ny = 1;
    if( x < 0 )
        {
        nx--;
        x += NUM_X_TILES;
        }
    else if( x >= NUM_X_TILES )
        {
        nx++;
        x -= NUM_X_TILES;
        }
    if( y < 0 )
        {
        ny--;
        y += NUM_Y_TILES;
        }
    else if( y >= NUM_Y_TILES )
        {
        ny++;
        y -= NUM_Y_TILES;
        }

    hope( x >= 0 );
    hope( x < NUM_X_TILES );
    hope( y >= 0 );
    hope( y < NUM_Y_TILES );

    // If the sector does not exist, search its neighbors.
    // The order in which we try the neighbors is fixed and the same for
    // all addresses (too-far neighbors are skipped). This means that the
    // result may not be the nearest available address, but it's easy and
    // consistent. Edge-neighbors are preffered to corner-neighbors. (1, 1)
    // sector should be always tested too, this search must succeed.

    if( neighbors[ ny ][ nx ].is_null() )
        {
        // We try neighbors in fixed order. Only "near" neighbors are checked,
        // ( 1, 1 ) sector must be always checked.

        // 1. Only "slightly" out-of-bounds addresses are allowed in input:
        // nx == 0      =>      x >= NUM_X_TILES - 2
        // nx == 2      =>      x < 2

        // 2. If NUM_X_TILES >= 4:
        // x >= NUM_X_TILES - 2     =>      x >= 2
        // x < 2                    =>      x < NUM_X_TILES - 2

        // => We try ( nx - 1 ) iff ( x < 2 ) [always if nx == 2, never if nx == 0],
        //    and ( nx + 1 ) iff ( x >= NUM_X_TILES - 2 ) [always if nx == 0, never if nx == 2].
        // Similarly for y axis.

        // Make sure we won't try to move out of the 3x3 neighborhood.
        hope( ( nx != 0 ) || ( x >= 2 ) );
        hope( ( nx != 2 ) || ( x < NUM_X_TILES - 2 ) );
        hope( ( ny != 0 ) || ( y >= 2 ) );
        hope( ( ny != 2 ) || ( y < NUM_Y_TILES -2 ) );

        if( ( x < 2 ) && !neighbors[ ny ][ nx - 1 ].is_null() )
            {
            // W neighbor.

            nx--;
            x = NUM_X_TILES - 1;
            }
        else if( ( y < 2 ) && !neighbors[ ny - 1 ][ nx ].is_null() )
            {
            // S neighbor.

            ny--;
            y = NUM_Y_TILES - 1;
            }
        else if( ( x >= NUM_X_TILES - 2 ) && !neighbors[ ny ][ nx + 1 ].is_null() )
            {
            // E neighbor.

            nx++;
            x = 0;
            }
        else if( ( y >= NUM_Y_TILES - 2 ) && !neighbors[ ny + 1 ][ nx ].is_null() )
            {
            // N neighbor.

            ny++;
            y = 0;
            }
        else if( ( x < 2 ) && ( y < 2 ) && !neighbors[ ny - 1 ][ nx - 1 ].is_null() )
            {
            // SW neighbor.

            nx--;
            ny--;
            x = NUM_X_TILES - 1;
            y = NUM_Y_TILES - 1;
            }
        else if( ( x >= NUM_X_TILES - 2 ) && ( y < 2 ) && !neighbors[ ny - 1 ][ nx + 1 ].is_null() )
            {
            // SE neighbor.

            nx++;
            ny--;
            x = 0;
            y = NUM_Y_TILES - 1;
            }
        else if( ( x >= NUM_X_TILES - 2 ) && ( y >= NUM_Y_TILES - 2 ) && !neighbors[ ny + 1 ][ nx + 1 ].is_null() )
            {
            // NE neighbor.

            nx++;
            ny++;
            x = 0;
            y = 0;
            }
        else if( ( x < 2 ) && ( y >= NUM_Y_TILES - 2 ) && !neighbors[ ny + 1 ][ nx - 1 ].is_null() )
            {
            // NW neighbor.

            nx--;
            ny++;
            x = NUM_X_TILES - 1;
            y = 0;
            }
        }

    // Done.

    hope( !neighbors[ ny ][ nx ].is_null() );
    neighbor_x = nx;
    neighbor_y = ny;
    tile_x = x;
    tile_y = y;
    }


/**
 * Get vertex height in client sector neighborhood.
 */
Float32 get_neighborhood_height
    (
    SectorNeighborhood &    neighbors,
    int                     x,
    int                     y
    )
    {
    std::size_t nx, ny;
    int vx, vy;
    get_neighbordhood_address( nx, ny, vx, vy, neighbors, x, y );
    return get_sector_height( neighbors[ ny ][ nx ], vx, vy );
    }


/**
 * Get vertex material in client sector neighborhood.
 */
Int32 get_neighborhood_material
    (
    SectorNeighborhood &    neighbors,
    int                     x,
    int                     y
    )
    {
    std::size_t nx, ny;
    int vx, vy;
    get_neighbordhood_address( nx, ny, vx, vy, neighbors, x, y );
    return get_sector_material( neighbors[ ny ][ nx ], vx, vy );
    }


/**
 * Compute nearest intersection of ray and sector heightmap.
 *
 * @a Intersection should contain ray origin and direction and intersection
 * info if @a best_found is @c true. If nearer intersection is found in this
 * sector, @a best_found will be set to @c true and @a intersection updated.
 * Nearest vertex info will NOT be updated by this funtion.
 */
void intersect_sector
    (
    bool &                              best_found,
    SectorManager::RayIntersection &    intersection,
    const SectorManager::SectorInfo &   sector
    )
    {
    hope( !best_found || intersection.distance >= 0.0f );

    const Renderer::Heightmap & heightmap = *sector.heightmap;
    const int x_tiles = heightmap.x_size - 1;
    const int y_strips = heightmap.y_size - 1;
    const Float32 x_origin = static_cast< Float32 >( sector.address.x ) * SECTOR_X_SIZE;
    const Float32 y_origin = static_cast< Float32 >( sector.address.y ) * SECTOR_Y_SIZE;

    int y;
    for( y = 0; y < y_strips; y++ )
        {
        int x;
        for( x = 0; x < x_tiles; x++ )
            {
            #define VERTEX( x, y ) \
                x_origin + ( x ) * TILE_X_SIZE, \
                y_origin + ( y ) * TILE_Y_SIZE, \
                heightmap.get_height( ( x ), ( y ) )

            const SVector3 v0( VERTEX(     x,     y ) );
            const SVector3 v1( VERTEX( x + 1,     y ) );
            const SVector3 v2( VERTEX(     x, y + 1 ) );
            const SVector3 v3( VERTEX( x + 1, y + 1 ) );

            #undef VERTEX

            Float32 distance;
            SVector3 normal;
            bool new_best = false;
            if( Raycast::intersect_triangle( distance, normal, intersection.origin,
                intersection.direction, v0, v1, v2 ) )
                {
                hope( distance >= 0.0f );
                if( !best_found || ( distance < intersection.distance ) )
                    {
                    intersection.distance = distance;
                    intersection.normal = normal;
                    new_best = true;
                    best_found = true;
                    }
                }
            if( Raycast::intersect_triangle( distance, normal, intersection.origin,
                intersection.direction, v1, v3, v2 ) )
                {
                if( !best_found || ( distance < intersection.distance ) )
                    {
                    intersection.distance = distance;
                    intersection.normal = normal;
                    new_best = true;
                    best_found = true;
                    }
                }
            if( new_best )
                {
                intersection.intersection = intersection.origin +
                    intersection.direction * intersection.distance;
                intersection.tile_sector = sector.address;
                intersection.tile_x = x;
                intersection.tile_y = y;
                }
            }
        }
    }


} // anonymous namesapce


// class SectorManager::SectorAddress

bool SectorManager::SectorAddress::is_neighbor
    (
    const SectorAddress &   other,
    int                     distance
    ) const
    {
    if( ( layer_id == other.layer_id ) &&
        ( std::abs( x - other.x ) <= distance ) &&
        ( std::abs( y - other.y ) <= distance ) )
        {
        return true;
        }
    return false;
    }


// class SectorManager::SectorInfo

SectorManager::SectorInfo::SectorInfo
    (
    SectorPointer sector
    ) : client_sector( sector ),
        heightmap( new Renderer::Heightmap( Renderer::Global::get_terrain_settings() ) ),
        terrain( new Renderer::Terrain )
    {
    // Copy address.

    const ::Demo::Lib::SectorAddress & sector_address = sector->sector_address;
    address.layer_id = sector_address.layer.get_object_id();
    address.x = sector_address.column;
    address.y = sector_address.row;

    // Initialize heightmap.

    heightmap->x_size = NUM_X_TILES + 1;
    heightmap->y_size = NUM_Y_TILES + 1;
    heightmap->elements.resize( ( heightmap->x_size + 2 ) * ( heightmap->y_size + 2 ) );
    heightmap->low_height = ::Demo::Lib::ClientSector::min_height;
    heightmap->high_height = ::Demo::Lib::ClientSector::max_height;
    heightmap->texture_scale = 1.0f;

    // Initialize terrain.

    terrain->set_scale( SVector3( TILE_X_SIZE, TILE_Y_SIZE, 1.0f ) );
    terrain->set_position( SVector3
        (
        static_cast< Float32 >( address.x ) * SECTOR_X_SIZE,
        static_cast< Float32 >( address.y ) * SECTOR_Y_SIZE,
        0.0f
        ) );
    terrain->register_to_global_manager( Renderer::PASS_WORLD );
    }


// class SectorManager

SectorManager::SectorManager()
    {
    }


SectorManager::~SectorManager()
    {
    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.unregister_callback( &sector_replica_callback );
    }


void SectorManager::initialize()
    {
    // Create world rendering pass.

    Renderer::PassManager & pass_manager = Renderer::Global::get_pass_manager();
    Renderer::RenderManager & render_manager =
        pass_manager.create_render_manager( Renderer::PASS_WORLD );
    render_manager.set_lighting( true );

    // Monitor client sector updates.

    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.register_callback
        (
        typeid( ::Demo::Lib::ClientSector ),
        &sector_replica_callback,
        0,
        ::Massiv::Core::VariantPointer( this )
        );

    // Hacks follow...

    // Create directional "sun" light source.

    sky_light.reset( new Renderer::Light );
    sky_light->set_position( SVector3( 0.0f, -1.0f, 0.3f ) );
    sky_light->set_infinite( true );
    sky_light->set_ambient_intensity( SColor( 0.3f, 0.2f, 0.1f ) );
    sky_light->set_diffuse_intensity( SColor( 0.8f, 0.75f, 0.6f ) );
    sky_light->register_to_global_manager( Renderer::PASS_WORLD );
    }


void SectorManager::update()
    {
    // Do nothing if rendering is disabled.

    if( Global::rendering_disabled )
        {
        return;
        }

    // Create entries for new sectors.

    ObjectIdSet::const_iterator it;
    for( it = updated_sectors.begin(); it != updated_sectors.end(); ++it )
        {
        if( sectors.find( *it ) == sectors.end() )
            {
            SectorInfo & info =
                sectors.insert( SectorInfos::value_type( *it, SectorInfo( *it ) ) ).first->second;
            sectors_by_address.insert( AddressMap::value_type( info.address, *it ) );
            if( info.address.layer_id == visible_layer )
                {
                info.terrain->set_visible( true );
                }
            else
                {
                info.terrain->set_visible( false );
                }
            }
        }

    // Check update time of material database and settings.

    Renderer::TerrainMaterialDatabase & material_database =
                Renderer::Global::get_terrain_material_database();
    material_database.refresh();
    const bool materials_changed = ( material_database.get_update_time() > rebuild_time );
    const bool settings_changed = Renderer::Global::get_terrain_settings().
                settings_changed( rebuild_time );

    if( materials_changed || settings_changed )
        {
        // Rebuild all terrains.

        SectorInfos::iterator sit;
        for( sit = sectors.begin(); sit != sectors.end(); ++sit )
            {
            SectorInfo & sector = sit->second;
            prepare_heightmap( sector );
            hope( sector.terrain.get() );
            sector.terrain->build( *sector.heightmap );
            }

        // Done.

        rebuild_time = ::Massiv::Core::SystemTime::current_time();
        }
    else
        {
        // Determine which terrains should be rebuilt - updated sectors and their neighbors.

        ObjectIdSet rebuild_terrains;
        for( it = updated_sectors.begin(); it != updated_sectors.end(); ++it )
            {
            const SectorInfo & sector_info = sectors.find( *it )->second;
            const SectorAddress & address = sector_info.address;

            SectorInfos::const_iterator sit;
            for( sit = sectors.begin(); sit != sectors.end(); ++sit )
                {
                const SectorInfo & other_info = sit->second;
                if( address.is_neighbor( other_info.address, 1 ) )
                    {
                    rebuild_terrains.insert( sit->first );
                    }
                }
            }

        // Rebuild terrains.

        for( it = rebuild_terrains.begin(); it != rebuild_terrains.end(); ++it )
            {
            const SectorInfos::iterator sit = sectors.find( *it );
            hope( sit != sectors.end() );
            SectorInfo & sector = sit->second;
            prepare_heightmap( sector );
            hope( sector.terrain.get() );
            sector.terrain->build( *sector.heightmap );
            }
        }

    // Done.

    updated_sectors.clear();
    }


void SectorManager::set_visible_layer
    (
    const ObjectId & new_visible_layer
    )
    {
    if( visible_layer == new_visible_layer )
        {
        return;
        }
    visible_layer = new_visible_layer;

    // Update visibility of all terrains.

    SectorInfos::iterator sit;
    for( sit = sectors.begin(); sit != sectors.end(); ++sit )
        {
        SectorInfo & sector = sit->second;
        if( sector.address.layer_id == visible_layer )
            {
            sector.terrain->set_visible( true );
            }
        else
            {
            sector.terrain->set_visible( false );
            }
        }
    }


bool SectorManager::cast_ray
    (
    RayIntersection & intersection
    ) const
    {
    // FIXME: This implementation is really lame and slow.

    // Find best intersection.

    bool best_found = false;
    SectorInfos::const_iterator sit;
    for( sit = sectors.begin(); sit != sectors.end(); ++sit )
        {
        const SectorInfo & info = sit->second;

        // Skip sectors from different layers.

        if( info.address.layer_id != intersection.layer_id )
            {
            continue;
            }

        // Check triangles in the sector.

        intersect_sector( best_found, intersection, info );
        }

    // Set results.

    if( best_found )
        {
        // Determine nearest vertex.

        const Float32 tile_x = intersection.intersection.x() -
            intersection.tile_sector.x * SECTOR_X_SIZE -
            intersection.tile_x * TILE_X_SIZE;
        const Float32 tile_y = intersection.intersection.y() -
            intersection.tile_sector.y * SECTOR_Y_SIZE -
            intersection.tile_y * TILE_Y_SIZE;

        const int vertex_x_index = intersection.tile_x +
            ( ( tile_x > 0.5f * TILE_X_SIZE ) ? 1 : 0 );
        const int vertex_y_index = intersection.tile_y +
            ( ( tile_y > 0.5f * TILE_Y_SIZE ) ? 1 : 0 );

        intersection.vertex_sector = intersection.tile_sector;
        if( vertex_x_index >= NUM_X_TILES )
            {
            intersection.vertex_sector.x++;
            intersection.vertex_x = vertex_x_index - NUM_X_TILES;
            }
        else
            {
            intersection.vertex_x = vertex_x_index;
            }
        if( vertex_y_index >= NUM_Y_TILES )
            {
            intersection.vertex_sector.y++;
            intersection.vertex_y = vertex_y_index - NUM_Y_TILES;
            }
        else
            {
            intersection.vertex_y = vertex_y_index;
            }
        }

    return best_found;
    }


bool SectorManager::get_height
    (
    Float32 &           height,
    const ObjectId &    sector_id,
    Float32             x,
    Float32             y
    ) const
    {
    // Find the sector.

    const SectorInfos::const_iterator sit = sectors.find( sector_id );
    if( sit == sectors.end() )
        {
        return false;
        }

    // Determine point location.

    const SectorInfo & info = sit->second;
    const Renderer::Heightmap & heightmap = *info.heightmap;
    const Float32 sector_x = x - info.address.x * SECTOR_X_SIZE;
    const Float32 sector_y = y - info.address.y * SECTOR_Y_SIZE;
    const Float32 tile_x = sector_x / TILE_X_SIZE;
    const Float32 tile_y = sector_y / TILE_Y_SIZE;
    const int index_x = ::Massiv::Core::FloatTraits< Float32 >::floor_int( tile_x );
    const int index_y = ::Massiv::Core::FloatTraits< Float32 >::floor_int( tile_y );

    // Out-of-sector position?

    if( index_x < 0 )
        {
        if( index_y < 0 )
            {
            height = heightmap.get_height( 0, 0 );
            }
        else if( index_y >= NUM_Y_TILES )
            {
            height = heightmap.get_height( 0, NUM_Y_TILES );
            }
        else
            {
            height = heightmap.get_height( 0, index_y );
            }
        }
    else if( index_x >= NUM_X_TILES )
        {
        if( index_y < 0 )
            {
            height = heightmap.get_height( NUM_X_TILES, 0 );
            }
        else if( index_y >= NUM_Y_TILES )
            {
            height = heightmap.get_height( NUM_X_TILES, NUM_Y_TILES );
            }
        else
            {
            height = heightmap.get_height( NUM_X_TILES, index_y );
            }
        }
    else
        {
        if( index_y < 0 )
            {
            height = heightmap.get_height( index_x, 0 );
            }
        else if( index_y >= NUM_Y_TILES )
            {
            height = heightmap.get_height( index_x, NUM_Y_TILES );
            }
        else
            {
            const Float32 h00 = heightmap.get_height(     index_x,     index_y );
            const Float32 h10 = heightmap.get_height( 1 + index_x,     index_y );
            const Float32 h01 = heightmap.get_height(     index_x, 1 + index_y );
            const Float32 h11 = heightmap.get_height( 1 + index_x, 1 + index_y );
            const Float32 fx = tile_x - static_cast< Float32 >( index_x );
            const Float32 fy = tile_y - static_cast< Float32 >( index_y );
            height = ( 1.0f - fy ) * ( ( 1.0f - fx ) * h00 + fx * h10 ) +
                                fy * ( ( 1.0f - fx ) * h01 + fx * h11 );
            }
        }

    return true;
    }


bool SectorManager::get_sector
    (
    ObjectId &          sector_id,
    const ObjectId &    layer_id,
    Float32             x,
    Float32             y
    ) const
    {
    const int sector_x = ::Massiv::Core::FloatTraits< Float32 >::floor_int( x / SECTOR_X_SIZE );
    const int sector_y = ::Massiv::Core::FloatTraits< Float32 >::floor_int( y / SECTOR_Y_SIZE );
    const AddressMap::const_iterator it = sectors_by_address.find
                ( SectorAddress( layer_id, sector_x, sector_y ) );
    if( it == sectors_by_address.end() )
        {
        return false;
        }

    sector_id = it->second.get_object_id();
    return true;
    }


void SectorManager::sector_replica_callback
    (
    ::Massiv::Core::ReplicaManager::UpdateReason            reason,
    ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
    int                                                     user_tag,
    ::Massiv::Core::VariantPointer                          user_data
    )
    {
    SectorManager * const self = ::Massiv::Core::variant_cast< SectorManager * >( user_data );
    switch( reason )
        {
        case ::Massiv::Core::ReplicaManager::UPDATE_RECEIVED:
        case ::Massiv::Core::ReplicaManager::MIGRATION_RECEIVED:
            {
            panic( "Unexpected reason." );
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_UPDATED:
            {
            self->sector_updated( replica.convert() );
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_GROUP_UPDATED:
            {
            // We don't care about objects in sector's replication group.
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_DESTROY:
            {
            self->destroy_sector( replica.convert() );
            break;
            }
        }
    }


void SectorManager::sector_updated
    (
    SectorPointer sector
    )
    {
    // Update updated_sectors set. Terrains will be rebuilt by #update().

    updated_sectors.insert( sector.get_object_id() );
    }


void SectorManager::destroy_sector
    (
    SectorPointer sector
    )
    {
    // Remove all info about the sector.

    const ObjectId & object_id = sector.get_object_id();
    updated_sectors.erase( object_id );
    const SectorInfos::iterator sit = sectors.find( object_id );
    if( sit != sectors.end() )
        {
        sectors_by_address.erase( sit->second.address );
        sectors.erase( sit );
        }
    }


void SectorManager::prepare_heightmap
    (
    const SectorInfo & sector
    )
    {
    hope( sector.heightmap.get() );
    Renderer::Heightmap & heightmap = *sector.heightmap;
    const int x_size = NUM_X_TILES + 1;
    const int y_size = NUM_Y_TILES + 1;

    const SectorPointer client_sector( sector.client_sector );
    int x, y;

    // Copy values from main area.

    for( y = 0; y < y_size - 1; y++ )
        {
        for( x = 0; x < x_size - 1; x++ )
            {
            heightmap.set_height( x, y, get_sector_height( client_sector, x, y ) );
            heightmap.set_material( x, y, get_sector_material( client_sector, x, y ) );
            heightmap.set_tile_flags( x, y, get_sector_tile_flags( client_sector, x, y ) );
            }
        }

    // Copy values from sector neighborhood.
    // 1 vertex wide borders are needed in west and south directions (for terrain builder),
    // 2 vertices wide borders are needed in north and east directions (one row/column for
    // border vertices and one for terrain builder).

    // If neighbor sector is not available, we must guess the height. This is quite tricky,
    // because the guess must be consistent. This is implemented by
    // get_neighbordhood_address().

    SectorNeighborhood neighbors;

    neighbors[ 0 ][ 0 ] = get_neighbor_sector( sector, -1, -1 );
    neighbors[ 0 ][ 1 ] = get_neighbor_sector( sector,  0, -1 );
    neighbors[ 0 ][ 2 ] = get_neighbor_sector( sector,  1, -1 );
    neighbors[ 1 ][ 0 ] = get_neighbor_sector( sector, -1,  0 );
    neighbors[ 1 ][ 1 ] = client_sector;
    neighbors[ 1 ][ 2 ] = get_neighbor_sector( sector,  1,  0 );
    neighbors[ 2 ][ 0 ] = get_neighbor_sector( sector, -1,  1 );
    neighbors[ 2 ][ 1 ] = get_neighbor_sector( sector,  0,  1 );
    neighbors[ 2 ][ 2 ] = get_neighbor_sector( sector,  1,  1 );

    // West and east borders.

    for( y = 0; y < y_size - 1; y++ )
        {
        heightmap.set_height(         -1, y, get_neighborhood_height( neighbors,         -1, y ) );
        heightmap.set_height( x_size - 1, y, get_neighborhood_height( neighbors, x_size - 1, y ) );
        heightmap.set_height(     x_size, y, get_neighborhood_height( neighbors,     x_size, y ) );
        heightmap.set_material( x_size - 1, y, get_neighborhood_material( neighbors, x_size - 1, y ) );
        }

    // South and north borders.

    for( x = 0; x < x_size - 1; x++ )
        {
        heightmap.set_height( x,         -1, get_neighborhood_height( neighbors, x,         -1 ) );
        heightmap.set_height( x, y_size - 1, get_neighborhood_height( neighbors, x, y_size - 1 ) );
        heightmap.set_height( x,     y_size, get_neighborhood_height( neighbors, x,     y_size ) );
        heightmap.set_material( x, y_size - 1, get_neighborhood_material( neighbors, x, y_size - 1 ) );
        }

    // Corner heights.

    heightmap.set_height(         -1,         -1, get_neighborhood_height( neighbors,         -1,         -1 ) );
    heightmap.set_height( x_size - 1,         -1, get_neighborhood_height( neighbors, x_size - 1,         -1 ) );
    heightmap.set_height(     x_size,         -1, get_neighborhood_height( neighbors,     x_size,         -1 ) );
    heightmap.set_height(         -1, y_size - 1, get_neighborhood_height( neighbors,         -1, y_size - 1 ) );
    heightmap.set_height(         -1,     y_size, get_neighborhood_height( neighbors,         -1,     y_size ) );
    heightmap.set_height( x_size - 1, y_size - 1, get_neighborhood_height( neighbors, x_size - 1, y_size - 1 ) );
    heightmap.set_height( x_size - 1,     y_size, get_neighborhood_height( neighbors, x_size - 1,     y_size ) );
    heightmap.set_height(     x_size, y_size - 1, get_neighborhood_height( neighbors,     x_size, y_size - 1 ) );
    heightmap.set_height(     x_size,     y_size, get_neighborhood_height( neighbors,     x_size,     y_size ) );

    // Corner materials.

    heightmap.set_material( x_size - 1, y_size - 1,
                get_neighborhood_material( neighbors, x_size - 1, y_size - 1 ) );
    }


SectorManager::SectorPointer SectorManager::get_neighbor_sector
    (
    const SectorInfo &  sector_info,
    int                 x_delta,
    int                 y_delta
    ) const
    {
    SectorAddress address( sector_info.address );
    address.x += x_delta;
    address.y += y_delta;
    SectorManager::AddressMap::const_iterator it = sectors_by_address.find( address );
    if( it == sectors_by_address.end() )
        {
        return ::Massiv::Core::null;
        }
    return it->second;
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "sector_manager_inline.h"
#endif
