/** @file sector_manager.h
 * Manager of sector replicas and terrains.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_SECTOR_MANAGER_H
#define CLIENT_SECTOR_MANAGER_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_TERRAIN_H
#include "renderer/terrain.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#ifndef MASSIV_REPLICA_MANAGER_H
#include "replication/replica_manager.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * Manager of client sector replicas and terrain objects.
 *
 * @todo Get rid of light hacks. Implement proper LightManager.
 */
class SectorManager
    {
    NO_COPY_CONSTRUCTOR( SectorManager );
    NO_ASSIGNMENT_OPERATOR( SectorManager );

public:

    /// ObjectId.
    typedef ::Massiv::Core::ObjectId ObjectId;

    /// Weak pointer to ClientSector.
    typedef ::Massiv::Core::WeakPointer< ::Demo::Lib::ClientSector > SectorPointer;

    /// Weak pointer to ClientEntity.
    typedef ::Massiv::Core::WeakPointer< ::Demo::Lib::ClientEntity > EntityPointer;
    
public:

    /**
     * Simplified sector address structure.
     */
    class SectorAddress
        {
    public:

        /**
         * Hasher.
         */
        struct Hash
            {
            INLINE std::size_t operator()
                (
                const SectorAddress & address
                ) const throw();
            }; // struct Hash

    public:

        /**
         * Default constructor.
         */
        INLINE SectorAddress();

        /**
         * Constructor.
         */
        INLINE SectorAddress
            (
            const ObjectId &    the_layer_id,
            const int           the_x,
            const int           the_y
            );

        /**
         * Destructor.
         */
        INLINE ~SectorAddress();

    public:

        /**
         * Compare addresses.
         */
        INLINE bool operator==
            (
            const SectorAddress & other
            ) const;
        
        /**
         * Is this address neighbor of given address?
         */
        bool is_neighbor
            (
            const SectorAddress &   other,
            int                     distance
            ) const;

    public:

        ObjectId    layer_id;
            /**< Object id of Layer.
                 Can be used only as unique layer id, Layer object itself is
                 never replicated to client. */

        int         x;
            /**< X position index. */

        int         y;
            /**< Y position index. */

        }; // class SectorAddress

public:

    /**
     * All client-side sector-related information.
     */
    class SectorInfo
        {
    public:

        /**
         * Default constructor.
         */
        INLINE SectorInfo();

        /**
         * Constructor.
         */
        SectorInfo
            (
            SectorPointer sector
            );

        /**
         * Copy constructor.
         */
        INLINE SectorInfo
            (
            const SectorInfo & other
            );

        /**
         * Destructor.
         */
        INLINE ~SectorInfo();

    public:

        SectorPointer                                   client_sector;
            /**< Pointer to ClientSector. */

        SectorAddress                                   address;
            /**< Sector address. */

        mutable std::auto_ptr< Renderer::Heightmap >    heightmap;
            /**< Heightmap used to rebuild terrains and perform ray casts. */

        mutable std::auto_ptr< Renderer::Terrain >      terrain;
            /**< Owning pointer to terrain. */

        }; // class SectorInfo

private:

    /// Set of object ids.
    typedef ext::hash_set< ObjectId, ObjectId::Hash > ObjectIdSet;

    /// Map from sector address to sector pointer.
    typedef ext::hash_map< SectorAddress, SectorPointer, SectorAddress::Hash > AddressMap;

    /// Map from object id to sector info.
    typedef ext::hash_map< ObjectId, SectorInfo, ObjectId::Hash > SectorInfos;

public:

    /**
     * Parameters and results of #cast_ray().
     */
    struct RayIntersection
        {
    // in
    public:
        ObjectId        layer_id;
            /** Object id of layer. */

        SVector3        origin;
            /**< Ray origin. */

        SVector3        direction;
            /**< Ray direction. */

    // out
    public:

        SVector3        intersection;
            /**< Intersection point. */
            
        Float32         distance;
            /**< Distance of #intersection point from #origin, in units specified
                 by length of #direction vector. */

        SVector3        normal;
            /**< Normal of intersected triangle. */

        SectorAddress   tile_sector;
            /**< Address of sector that owns intersected tile. */

        int             tile_x;
            /**< X index of intersected tile. */

        int             tile_y;
            /**< Y index of intersected tile. */

        SectorAddress   vertex_sector;
            /**< Address of sector that owns vertex nearest to the intersection.
                 If the intersection is near eastern or nothern border of sector,
                 #vertex_sector may be different from #tile_sector. There is no
                 guarantee that sector with such address really exists in that case. */

        int             vertex_x;
            /**< X index of vertex nearest to the intersection. */

        int             vertex_y;
            /**< Y index of vertex nearest to the intersection. */
        
        }; // struct RayIntersection

public:

    /**
     * Constructor.
     */
    SectorManager();

    /**
     * Destructor.
     */
    ~SectorManager();

    /**
     * Initialize sector manager.
     *
     * @note This contains temporary hacks - it creates light in world render pass.
     */
    void initialize();

public:

    /**
     * Update and rebuild terrains from updated client sectors.
     */
    void update();   

public:

    /**
     * Get id of visible layer.
     */
    INLINE const ObjectId & get_visible_layer() const;

    /**
     * Set id of visible layer.
     *
     * This will immediately affect visibility of terrains.
     * It will be automatically called by #PlayerManager::update().
     */
    void set_visible_layer
        (
        const ObjectId & new_visible_layer
        );

public:

    /**
     * Intersect ray with sectors in given layer.
     *
     * The following attributes of @a intersection should be initialized when
     * calling this method:
     * - layer_id:  Object id of layer to search.
     * - origin:    Ray origin.
     * - direction: Ray direction.
     *
     * If intersection is found, @c true will be returned and @a intersection
     * updated (see #RayIntersection). Otherwise @c false will be returned.
     */
    bool cast_ray
        (
        RayIntersection & intersection
        ) const;

    /**
     * Get terrain height at given point.
     *
     * @retval height Terrain height at @a x, @a y.
     * @param sector_id Id of sector to examine.
     * @param x X coordinate in the layer (not sector).
     * @param y Y coordinate in the layer (not sector).
     * @return @c true on success, @c false if replica of the sector does not exist.
     */
    bool get_height
        (
        Float32 &           height,
        const ObjectId &    sector_id,
        Float32             x,
        Float32             y
        ) const;

    /**
     * Get id of sector for given layer coordinates.
     *
     * @return @c true if @a sector_id has been set, @c false if no sector exists
     *         at given coordinates.
     */
    bool get_sector
        (
        ObjectId &          sector_id,
        const ObjectId &    layer_id,
        Float32             x,
        Float32             y
        ) const;

// hacks
public:

    ///hack
    Renderer::Light & get_sky_light()
        {
        hope( sky_light.get() );
        return *sky_light;
        }

private:

    /**
     * Sector replica update callback.
     */
    static void sector_replica_callback
        (
        ::Massiv::Core::ReplicaManager::UpdateReason            reason,
        ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
        int                                                     user_tag,
        ::Massiv::Core::VariantPointer                          user_data
        );

    /**
     * Inform terrain manager about new sector replica or replica update.
     */
    void sector_updated
        (
        SectorPointer sector
        );

    /**
     * Inform terrain manager that sector replica will be destroyed.
     */
    void destroy_sector
        (
        SectorPointer sector
        );

private:

    /**
     * Initialize #heightmap from given sector and its neighbors.
     */
    void prepare_heightmap
        (
        const SectorInfo & sector
        );

    /**
     * Get sector at coordinates specified relatively to given sector.
     */
    SectorPointer get_neighbor_sector
        (
        const SectorInfo &  sector_info,
        int                 x_delta,
        int                 y_delta
        ) const;

private:

    ObjectId            visible_layer;
        /**< Id of layer to draw. */

    SectorInfos         sectors;
        /**< Replicated sectors - infos and terrains. */

    AddressMap          sectors_by_address;
        /**< Map from sector address to sector pointer. */

    ObjectIdSet         updated_sectors;
        /**< Object ids of updated client sectors. */

    SystemTime          rebuild_time;
        /**< Time of last rebuild-all-sectors.
             All sectors are rebuilt whenever material database or renderer
             settings change. */

// hacks
private:

    std::auto_ptr< Renderer::Light >            sky_light;
        /**< Hacked directional light. */

    }; // class SectorManager


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "sector_manager_inline.h"
#endif


#endif // CLIENT_SECTOR_MANAGER_H
