/** @file button.cpp
 * Simple graphic button with text label (implementation).
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/button.h"
#include "client_global.h"
#include "media_repository.h"


namespace Demo {
namespace Client {



Button::Button
    (
    ButtonType          the_type
    ) :
    type( the_type ),
    message_receiver( NULL ),
    down( false )
    {
    MediaRepository & media = Global::get_media_repository();
    
    text.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    text->set_parent_overlay_object( this );
    
    background_icon.reset( media.create_icon( "gui/panel_background" ) );
    white.reset( new Renderer::Icon( "texture/white" ) );
    }


void Button::set_text
    (
    const std::string & new_text
    )
    {
    text->print_line( new_text );
    }


void Button::set_size_from_text()
    {
    SVector2 text_size = text->get_current_size();
    text_size += SVector2( 8.0f + 8.0f, 4.0f + 4.0f );
    set_base_size( text_size );
    }


void Button::render()
    {
    // Update text.
        {
        SVector2 text_size = text->get_current_size();
        SVector2 size = get_current_size();
        SVector2 pos = SVector2(
                  ( size.x() - text_size.x() ) / 2.0f,
                  ( size.y() - text_size.y() ) / 2.0f );
                  
        if( is_down() )
            {
            pos += SVector2( 1.0f, 1.0f );
            }

        text->set_base_offset( pos );

        
        Float32 alpha = ( is_enabled() ? get_current_color().alpha() : get_current_color().alpha() * 0.4f );
        SColor color( get_current_color() );
        color.set_alpha( color.alpha() * alpha );
        
        text->set_base_color( color );
        }
    
    render_background();
    render_frame();
    }


bool Button::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    if( Super::mouse_button_down( button, mod ) )
        {
        return true;
        }

    if( button == SDL_BUTTON_LEFT )
        {
        down = !down;
        clicked( true );
        return true;
        }
    
    return false;
    }

    
void Button::clicked
    (
    bool notify_receiver
    )
    {
    if( notify_receiver )
        {
        if( down || type == TYPE_SWITCH )
            {
            if( message_receiver )
                {
                message_receiver->button_clicked( this );
                }
            }
        }
    
    if( type == TYPE_ONE_SHOT )
        {
        down = false;
        }
    }


void Button::render_background()
    {
    hope( background_icon.get() );
    
    Float32             width = get_current_size().x();
    Float32             height = get_current_size().y();
    Float32             texture_width = 256.0f;
    Float32             texture_height = 256.0f;

    // Compute uv coordinates to create repeatable pattern.
    Float32             uv_h_size = width / texture_width;
    Float32             uv_v_size = height / texture_height;

    background_icon->set_origin( TOP_LEFT );
    background_icon->set_base_offset( screenspace_translation().svector() );
    background_icon->set_base_size( get_current_size() );
    background_icon->set_base_color( get_current_color() );
    if( is_down() )
        {
        background_icon->set_uv( - 1.0f / texture_width, - 1.0f / texture_height,
            uv_h_size - 1.0f / texture_width, uv_v_size - 1.0f / texture_height );
        }
    else
        {
        background_icon->set_uv( 0.0f, 0.0f, uv_h_size, uv_v_size );
        }
    background_icon->render();
    
    // Highlight the button.
    if( is_enabled() && is_mouse_on() )
        {
        white->set_base_offset( get_top_left_corner_on_screen() );
        white->set_base_size( get_current_size() );
        white->set_base_color( SColor( 1.0f, 1.0f, 1.0f, 0.4f ) );
        white->render();
        }
    }


void Button::render_frame()
    {
    Float32 alpha = ( is_enabled() ? get_current_color().alpha() : get_current_color().alpha() * 0.4f );
    const SColor    dark( 0.0f, 0.0f, 0.0f, 0.6f * alpha );
    const SColor    light( 1.0f, 1.0f, 1.0f, 0.6f * alpha );
    const SVector2  top_left = screenspace_translation().svector();
    const SVector2  size = get_current_size();
    
    render_hline( is_up() ? light : dark, top_left.x(), top_left.x() + size.x(), top_left.y() );
    render_vline( is_up() ? light : dark, top_left.x(), top_left.y(), top_left.y() + size.y() );
    render_hline( is_up() ? dark : light, top_left.x(), top_left.x() + size.x(), top_left.y() + size.y() - 1.0f );
    render_vline( is_up() ? dark : light, top_left.x() + size.x() - 1.0f, top_left.y(), top_left.y() + size.y() );
    }


void Button::render_hline
    (
    const SColor &  color,
    Float32         x1,
    Float32         x2,
    Float32         y
    )
    {
    const SVector2 top_left( x1, y );
    const SVector2 size( x2 - x1, 1.0f );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


void Button::render_vline
    (
    const SColor &  color,
    Float32         x,
    Float32         y1,
    Float32         y2
    )
    {
    const SVector2 top_left( x, y1 );
    const SVector2 size( 1.0f, y2 - y1 );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "gui/button_inline.h"
#endif
