/** @file button.h
 * Simple graphic button with text label.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_BUTTON_H
#define CLIENT_BUTTON_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_TEXT_RECTANGLE_H
#include "renderer/text_rectangle.h"
#endif

#ifndef CLIENT_ICON_H
#include "renderer/icon.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * Simple graphic button with text label.
 * Button is #Widget.
 * As #InputObject it handles mouse clicks and switch self up and down.
 * As #OverlayObject it is able to draw self.
 *
 * Button generates messages when the button is pressed up or
 * down (depending on its @c ButtonType). The messages are
 * sent to #ButtonMessageReceiver, which is registered in the button
 * via @c set_message_receiver().
 */
class Button : public Widget
    {
    typedef Widget  Super;
    
    NO_COPY_CONSTRUCTOR( Button );
    NO_ASSIGNMENT_OPERATOR( Button );

public:

    /**
     * Type of button.
     */
    enum ButtonType
        {
        TYPE_SWITCH,
            /**< The button is switchable. */

        TYPE_ONE_SHOT
            /**< When the button is pressed down, it is automatically
             * set up.
             */

        }; // ButtonType
    
    /**
     * Receiver of messages from the button.
     */
    class ButtonMessageReceiver
        {
    public:

        /**
         * Destructor.
         */
        virtual ~ButtonMessageReceiver()
            {
            }

        /**
         * Called when user clicks the button. If type of the button is
         * @c TYPE_SWITCH, the callback is called when the user
         * pressed the button down or up. If the type is @a TYPE_ONE_SHOT,
         * the callback is called only when the user pressed the button down.
         */
        virtual void button_clicked
            (
            Button * button
            )
            {
            }
        
        }; // ButtonMessageReceiver
        
public:

    /**
     * Constructor.
     */
    Button
        (
        ButtonType          type
        );

public:

    /**
     * Is the button up?
     */
    INLINE bool is_up() const;
    
    /**
     * Is the button down?
     */
    INLINE bool is_down() const;
    
    /**
     * Push the button up or down.
     * Nothing is done if the button is already in requested state (up or down).
     * @param notify_receiver Whether the message receiver should be
     *      notified of this click. If button type is @c TYPE_ONE_SHOT,
     *      the receiver will be notified only if down is true.
     */
    INLINE void push
        (
        bool    down = true,
        bool    notify_receiver = false
        );

public:

    /**
     * Set new text that will be drawn inside the button.
     */
    void set_text
        (
        const std::string & new_text
        );
    
    /**
     * Set size of the button from the text specified by @a set_text().
     */
    void set_size_from_text();

    /**
     * Set #ButtonMessageReceiver of the button.
     * If @a message_receiver is @c NULL, message notifying will be
     * disabled.
     */
    INLINE void set_message_receiver
        (
        ButtonMessageReceiver * message_receiver
        );

public:

    // Renderer::OverlayObject
    virtual void render();
    
    // InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );

private:

    /**
     * Called when the button was clicked.
     */
    void clicked
        (
        bool notify_receiver
        );

    /**
     * Render background of the button.
     */
    void render_background();
    
    /**
     * Render frame of the button.
     */
    void render_frame();

    /** Render horizontal line. */
    void render_hline
        (
        const SColor &  color,
        Float32         x1,
        Float32         x2,
        Float32         y
        );

    /** Render vertical line. */
    void render_vline
        (
        const SColor &  color,
        Float32         x,
        Float32         y1,
        Float32         y2
        );

private:

    ButtonType                          type;
        /**< Type of the button. */

protected:

    ButtonMessageReceiver *             message_receiver;
        /**< Message receiver. */
        
private:

    bool                                down;
        /**< Is the button down or up? */
    
    std::auto_ptr< Renderer::TextRectangle >    text;
        /**< Text drawn inside the button. */

    std::auto_ptr< Renderer::Icon >     background_icon;
        /**< Background texture. */
    
    std::auto_ptr< Renderer::Icon >     white;
        /**< White texture used to draw lines. */

    }; // Button


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "gui/button_inline.h"
#endif


#endif // CLIENT_BUTTON_H
