/** @file camera_control.h
 * Camera control.
 */

/*
 * begin        : Fri Nov 14 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/camera_control.h"
#include "gui/widget_input_dispatcher.h"
#include "client_global.h"

namespace Demo {
namespace Client {


// Implementation of CameraControl
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

namespace {

const Float32   min_distance = 4.0f;
const Float32   max_distance = 100.0f;
const Float32   min_latitude = MATHF_PI / 8.0f;
const Float32   max_latitude = MATHF_PI / 2.0f;

const Float32   max_longitude_speed = MATHF_PI * 2.0f;
const Float32   max_distance_speed = 40.0f;
const Float32   max_longitude_acceleration = MATHF_PI;
const Float32   max_distance_acceleration = 30.0f;

const Float32   longitude_acceleration_coef = MATHF_PI * 4.0f;
const Float32   distance_acceleration_coef = 30.0f;

const Float32   longitude_friction_coef = 4.0f;
const Float32   distance_friction_coef = 4.0f;

const Float32   distance_coef = 0.1f;

} // anonymous namespace


CameraControl::CameraControl() :
    longitude( 0 ),
    longitude_speed( 0.0f ),
    longitude_acceleration( 0.0f ),
    distance( 3.0f ),
    distance_speed( 0.0f ),
    distance_acceleration( 0.0f ),
    latitude( 0.0f )
    {
    }


void CameraControl::next_tick()
    {
    InputManager &  input_manager = Global::get_input_manager();
    SystemTime      current_time = Massiv::System::system_time();

    Float32 left_duration = 0.0f;
    Float32 right_duration = 0.0f;
    Float32 up_duration = 0.0f;
    Float32 down_duration = 0.0f;

    // If there is no widget with keyboard focus, interpret arrows as camera movement.

    WidgetInputDispatcher * const dispatcher = &Global::get_widget_input_dispatcher();
    hope( dispatcher );
    if( !dispatcher->get_focus_widget() )
        {
        const InputManager::KeyState &  left = input_manager.get_key_state( SDLK_LEFT );
        const InputManager::KeyState &  right = input_manager.get_key_state( SDLK_RIGHT );
        const InputManager::KeyState &  up = input_manager.get_key_state( SDLK_UP );
        const InputManager::KeyState &  down = input_manager.get_key_state( SDLK_DOWN );

        left_duration = get_key_press_duration( left, current_time );
        right_duration = get_key_press_duration( right, current_time );
        up_duration = get_key_press_duration( up, current_time );
        down_duration = get_key_press_duration( down, current_time );
        }

    Float32 tick_duration =
        static_cast<Float32>( ( current_time - last_tick ).get_value() );

    // Update camera parameters.

    longitude += longitude_speed * tick_duration +
        longitude_acceleration * tick_duration * tick_duration * 0.5f;

    distance += distance_speed * tick_duration +
        distance_acceleration * tick_duration * tick_duration * 0.5f;

    longitude_speed += longitude_acceleration * tick_duration;
    distance_speed += distance_acceleration * tick_duration;

    longitude_speed = Massiv::Core::clamp
        ( longitude_speed, -max_longitude_speed, max_longitude_speed );

    distance_speed = Massiv::Core::clamp
        ( distance_speed, -max_distance_speed, max_distance_speed );

    distance = Massiv::Core::clamp
        ( distance , min_distance, max_distance );

    // Update accelerations according to keyboard impulses.

    longitude_acceleration += ( right_duration - left_duration ) *
        longitude_acceleration_coef;

    distance_acceleration += ( down_duration - up_duration ) *
        distance_acceleration_coef;

    longitude_acceleration = Massiv::Core::clamp
        (
        longitude_acceleration,
        -max_longitude_acceleration,
        max_longitude_acceleration
        );

    distance_acceleration = Massiv::Core::clamp
        (
        distance_acceleration,
        -max_distance_acceleration,
        max_distance_acceleration
        );

    // Apply friction.

    longitude_acceleration *= ::Massiv::Core::FloatTraits<Float32>::exp
        ( -longitude_friction_coef * tick_duration );

    distance_acceleration *= ::Massiv::Core::FloatTraits<Float32>::exp
        ( -distance_friction_coef * tick_duration );

    longitude_speed *= ::Massiv::Core::FloatTraits<Float32>::exp
        ( -longitude_friction_coef * tick_duration );

    distance_speed *= ::Massiv::Core::FloatTraits<Float32>::exp
        ( -distance_friction_coef * tick_duration );

    // Compute latitude as a function of distance.

    latitude = Massiv::Core::clamp
        ( distance * distance_coef, min_latitude, max_latitude );

    // Done.

    last_tick = current_time;
    }


void CameraControl::apply_control
    (
    Renderer::CirclingCamera &  camera,
    Float32                     longitude_offset
    ) const
    {
    camera.set_latitude( latitude );
    camera.set_longitude( longitude + longitude_offset );
    camera.set_distance( distance );
    }


Float32 CameraControl::get_key_press_duration
    (
    const InputManager::KeyState &  state,
    const SystemTime &              current_time
    )
    {
    if( state.press_time < state.release_time && state.release_time < last_tick )
        {
        // The key was released in a past tick.

        return 0.0f;
        }
    else
    if( state.press_time < state.release_time && state.release_time > last_tick )
        {
        // The key was pressed for a portion of the current tick.

        if( state.press_time < last_tick )
            {
            return static_cast<Float32>
                ( ( state.release_time - last_tick ).get_value() );
            }
        else
            {
            return static_cast<Float32>
                ( ( state.release_time - state.press_time ).get_value() );
            }
        }
    else
    if( state.press_time > state.release_time && state.down )
        {
        // The key was pressed during the whole current tick and has
        // not been released yet.

        return static_cast<Float32>
            ( ( current_time - last_tick ).get_value() );
        }

    return 0.0f;
    }


} // namespace Client
} // namespace Demo
