/** @file client_console.cpp
 * Client console implementation.
 */


/*
 * begin        : Thu Mar 13 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "gui/client_console.h"
#include "sdl/client_sdl.h"
#include "input/input_layers.h"
#include "input/input_manager.h"
#include "renderer/font.h"
#include "renderer/text_rectangle.h"
#include "renderer/layers.h"
#include "game/game_context.h"
#include "lib/client/client_node_rpc.h"
#include "lib/shared/account_object_rpc.h"

// Consoles.
#include "gui/client_full_console.h"
#include "gui/client_chat_console.h"

#include <cstring>
#include <sstream>

namespace Demo {
namespace Client {

namespace {

const unsigned max_key_buffer_size = 128;
    /**< Maximum number of buffered keys. */

}


// static initialization.
Console::History    Console::commands( 100 );
std::string         Console::current_line( "" );


// Implementation of Console::ConsoleSwitcher
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Console::ConsoleSwitcher::ConsoleSwitcher
    (
    ) : enabled( false ), 
        chat_visible( false )
    {
    }


Console::ConsoleSwitcher::~ConsoleSwitcher
    (
    )
    {
    }


bool Console::ConsoleSwitcher::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( !enabled )
        {
        return false;
        }
    else
        {
        if( key == SDLK_BACKQUOTE )
            {
            chat_visible = !chat_visible;
            Demo::Client::Global::get_chat_console().show( chat_visible );
            Demo::Client::Global::get_full_console().show( !chat_visible );
            if( !chat_visible )
                {
                Demo::Client::Global::get_full_console().bring_to_front();
                Demo::Client::Global::get_full_console().set_focus();
                }
            return true;
            }
        }
        
        
    return false;
    }


void Console::ConsoleSwitcher::initialize
    (
    )
    {
    }


void Console::ConsoleSwitcher::enable
    (
    bool new_enabled
    )
    {
    enabled = new_enabled;
    if( enabled )
        {
        key_down( SDLK_BACKQUOTE, ( SDLMod ) 0  );
        }
    }


// Implementation of Console::History
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
Console::History::History
    (
    int max_size
    ) : max_size( max_size ),
        current_position( 0 ),
        write_filename()
    {
    }

void Console::History::push
    (
    const ColorString & s
    )
    {
    if( log.size() > max_size )
        {
        log.pop_front();
        }
        
    log.push_back( s );

    VolumeManager::OFStream file;
    PathManager &           pm = Massiv::Core::Global::path_manager();

    if( write_filename.size() > 0 )
        {
        try
            {
            file.open( pm.get_work_directory() + "\\" + write_filename, std::ios_base::app );
            }
        catch( ... )
            {
            // Be silent.
            }
        
        if( file )
            {
            file << s.to_string() << std::endl;
            }
        else
            {
            std::ostringstream oss;
            oss << "Cannot write console output to file: " << write_filename;
            oss << ". Disabling file output.";
            write_filename = "";
            push( s );
            }
        }
    }


Console::History::Iterator Console::History::begin()
    {
    return log.begin() += current_position;
    }


Console::History::Iterator Console::History::end()
    {
    return log.end();
    }


int Console::History::size()
    {
    return log.size();
    }


void Console::History::change_position
    (
    int delta
    )
    {
    set_position( current_position + delta );
    }


void Console::History::set_position
    (
    int position
    )
    {
    if( position < 0 )
        {
        position = 0;
        }
        
    current_position = position;

    if( current_position > log.size() )
        {
        if( log.size() == 1 )
            {
            current_position = 0;    
            }
        else
            {
            current_position = log.size() - 2;    
            }
        }
    }


int Console::History::get_position()
    {
    return current_position;
    }


bool Console::History::is_front()
    {
    return current_position == 0;
    }


bool Console::History::is_back()
    {
    return current_position == log.size() - 1;
    }


const ColorString & Console::History::current()
    {
    return log.at( current_position );
    }


void Console::History::clear_log()
    {
    log.clear();
    }


void Console::History::set_output_file
    (
    const std::string & filename
    )
    {
    write_filename = filename;
    }


// Implementation of Console::KeyInfo
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Console::KeyInfo::KeyInfo
    (
    SDLKey  key,
    SDLMod  mod
    ) : key( key ),
        mod( mod )
    {
    }


// Implementation of Console
// ~~~~~~~~~~~~~~~~~~~~~~~~~
    
Console::Console()
      : display( 1000 ),
        num_lines( 0 ),
        num_columns( 0 ),
        cursor_position( 0 ),
        history_browsed( false ),
        line_modified( false ),
        should_be_visible( false ),
        loose_focus_on_submit( false ),
        prompt( "[unknown]> " ),
        cell_width( 0 ),
        cell_height( 0 ),
        redraw_display( true ),
        redraw_command( true ),
        resizing( false ),
        waiting_for_auto_completion( false ),
        waiting_for_command_completion( false ),
        completion_index( 0 )
    {
    }


void Console::print
    (
    const ColorString & color_str,
    History           & history
    )
    {
    std::string s = color_str.to_string();
    std::string::size_type startpos = 0;
    std::string::size_type pos = 0;
    ColorString str( color_str );
    ColorString line;

    // Split @c s to "lines" and push each line to the display log.
    while( ( pos = s.find( "\n", startpos, 1 ) ) != s.npos )
        {
        line = str.substr( startpos, pos - startpos );
        
        while( line.size() > ( unsigned int ) num_columns )
            {
            history.push( line.substr( 0, num_columns ) );
            line = line.substr( num_columns );
            }
            
        startpos = pos + 1;
        history.push( line );
        }

    // Push last line fragment to the display log.

    line = str.substr( startpos );
    while( line.size() > ( unsigned int ) num_columns )
        {
        history.push( line.substr( 0, num_columns ) );
        line = line.substr( num_columns );
        }

    history.push( line );
    focus_to_current();
    redraw_display = true;
    }


void Console::print_in_color
    (
    const std::auto_ptr< Renderer::TextRectangle > & where,
    unsigned int                                     row,
    const ColorString &                              text,
    unsigned int                                     column
    )
    {
    unsigned int written = column;
    
    const ColorString::ColorMap color_map = text.get_color_map();
    ColorString::ColorMap::const_iterator it = color_map.begin();
    ColorString::ColorMap::const_iterator it_next;
    const std::string str = text.to_string();
    
    
    while( it != color_map.end() )
        {
        if( it->first >= str.size() )
            {
            return;
            }
            
        it_next = it;
        ++it_next;
        where->set_text_color( it->second );
        if( it_next != color_map.end() )
            {
            where->print( written, row, str.substr( it->first, it_next->first - it->first ) );
            written += it_next->first - it->first;
            }
        else
            {
            where->print( written, row, str.substr( it->first ) );
            }
        ++it;        
        }
    }


void Console::print
    (
    const ColorString & s
    )
    {
    print( s, display );
    }


bool Console::key_up
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( InputObject::key_up( key, mod ) )
        return true;
    
    if( !is_visible() )
        {
        return false;
        }
    else
        {
        // Console eats all keys.
        return true;
        }
    }


bool Console::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( InputObject::key_down( key, mod ) )
        return true;

    // Invisible console doesn't eat any key.
    if( !should_be_visible )
        {
        return false;
        }
    
    // Console eats all keys unless blocked.
    bool event_eaten = true;

    switch( key )
        {
        case SDLK_ESCAPE:
            {
            set_current( std::string() );
            discard_command_completions();

            cursor_position = 0;
            break;
            }

        case SDLK_UP:
            {
            if( mod & KMOD_SHIFT )
                {
                // Scroll up by a line.
                display.change_position( -1 );
                redraw_display = true;
                }
            else
                {
                if( is_blocked() )
                    {
                    event_eaten = false;
                    break;
                    }

                if( ( !commands.is_front() || !history_browsed ) 
                     && commands.size() > 0 )
                    {
                    // Replace current command by a previous one.
                    if( !history_browsed )
                        {
                        history_browsed = true;
                        }
                    else
                        {
                        commands.change_position( -1 );
                        }

                    if( commands.size() )
                        {
                        set_current( commands.current().to_string() );
                        discard_command_completions();
                        }
                    }
                }

            break;
            }

        case SDLK_DOWN:
            {
            if( mod & KMOD_SHIFT )
                {
                // Scroll down by a line.
                display.change_position( 1 );
                
                redraw_display = true;
                }
            else
                {
                if( is_blocked() )
                    {
                    event_eaten = false;
                    break;
                    }

                if( !commands.is_back() && commands.size() > 0 )
                    {
                    // Replace current command by a next one.
                    if( !history_browsed )
                        {
                        history_browsed = true;
                        }

                    commands.change_position( 1 );
                    set_current( commands.current().to_string() );
                    discard_command_completions();
                    }
                }

            break;
            }

        case SDLK_PAGEUP:
            {
            if( mod & KMOD_CTRL )
                {    
                display.set_position( 0 );
                }
            else
                {
                // Scroll up by a page.
                display.change_position( -num_lines + 3 );
                }
            
            redraw_display = true;
            break;
            }

        case SDLK_PAGEDOWN:
            {
            if( mod & KMOD_CTRL )
                {    
                // Scroll to the bottom.
                display.set_position( display.size() - display_text->get_rows() + 2 );
                }
            else
                {
                // Scroll down by a page.
                display.change_position( num_lines - 3 );
                }
                
            redraw_display = true;
            break;
            }

        case SDLK_HOME:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }

            // Go to the beginning of the current line.
            cursor_position = 0;
            
            discard_command_completions();
            focus_to_current();

            break;
            }

        case SDLK_END:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
            
            // Go to the end of the current line.
            cursor_position = current_line.size();

            discard_command_completions();
            focus_to_current();

            break;
            }

        case SDLK_LEFT:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
            
            if( mod & KMOD_CTRL )
                {
                // Move cursor left by one word.
                if( cursor_position > 0 )
                    {
                    cursor_position--;
                    }

                while( cursor_position > 0 && current_line[ cursor_position ] == ' ' )
                    {
                    cursor_position--;
                    }

                while( cursor_position > 0 && current_line[ cursor_position - 1 ] != ' ' )
                    {
                    cursor_position--;
                    }
                
                discard_command_completions();
                focus_to_current();
                }
            else
                {
                if( cursor_position > 0 )
                    {
                    // Move cursor left.
                    cursor_position--;
                    
                    discard_command_completions();
                    focus_to_current();
                    }
                }

            break;
            }

        case SDLK_RIGHT:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
            
            if( mod & KMOD_CTRL )
                {
                // Move cursor right by one word.
                if( cursor_position < current_line.size() )
                    {
                    cursor_position++;
                    }

                while
                    (
                    cursor_position < current_line.size() &&
                    current_line[ cursor_position ] != ' '
                    )
                    {
                    cursor_position++;
                    }

                while
                    (
                    cursor_position < current_line.size() &&
                    current_line[ cursor_position ] == ' '
                    )
                    {
                    cursor_position++;
                    }
                
                discard_command_completions();
                focus_to_current();
                }
            else
                {
                if( cursor_position < current_line.size() )
                    {
                    // Move cursor right.
                    cursor_position++;
                    
                    discard_command_completions();
                    focus_to_current();
                    }
                }
            
            break;
            }

        case SDLK_BACKSPACE:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
            
            if( mod & KMOD_CTRL )
                {
                // TODO: Smazat posledni slovo
                }
            else
                {
                // Delete character before cursor.
                if( cursor_position > 0 )
                    {
                    current_line.erase( --cursor_position, 1 );
                    
                    discard_command_completions();
                    focus_to_current();
                    
                    line_modified = true;
                    redraw_command = true;
                    }
                }

            break;
            }

        case SDLK_DELETE:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
            
            // Delete character at the cursor.
            if( cursor_position < current_line.size() )
                {
                current_line.erase( cursor_position, 1 );
                
                discard_command_completions();
                focus_to_current();
                
                line_modified = true;
                redraw_command = true;
                }

            break;
            }

        case SDLK_RETURN:
        case SDLK_KP_ENTER:
            {
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }
        
            // Return the current line.
            if( current_line[ 0 ] == '/' )
                {
                submit_current();
                discard_command_completions();
                }
            else
                {
                submit_chat();
                }

            set_current( std::string() );
            focus_to_current();
            if( loose_focus_on_submit )
                {
                disable_focus();
                }

            break;
            }

        case SDLK_TAB:
            {
            static bool last_without_shift = true;

            if( mod & KMOD_ALT )
                {
                // WIN32 hack:
                //  Do not treat "Alt+Tab" (Window switch) as equivalent
                //  to "Tab".

                break;
                }

            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }

            if( command_completions.size() && !( mod & KMOD_SHIFT ) )
                {
                // Cycle forward within possible command completions.
                if( !last_without_shift )
                    completion_index = ( completion_index + 2 ) % command_completions.size();

                set_current( command_completions[ completion_index ] );

                completion_index = ( completion_index + 1 ) % command_completions.size();
                line_modified = true;
                last_without_shift = true;
                }
            else if( command_completions.size() && ( mod & KMOD_SHIFT ) )
                {
                // Cycle backward within possible command completions.
                if( last_without_shift )
                    completion_index = ( completion_index - 2 + command_completions.size() ) % command_completions.size();
                
                set_current( command_completions[ completion_index ] );

                completion_index = ( completion_index - 1 + command_completions.size() ) % command_completions.size();
                line_modified = true;
                last_without_shift = false;
                }
            else
                {
                // Retrieve all posible command completions.
                if( current_line.size() )
                    {
                    if( current_line[ 0 ] != '/' )    
                        {
                        set_current( "/" + current_line );
                        }
                        auto_complete_current();
                    
                    }
                else
                    {
                    auto_complete_current();
                    }
                }

            break;
            }

        default:
            {
            bool    shift = ( mod & KMOD_SHIFT ) != 0;
            bool    caps = ( mod & KMOD_CAPS ) != 0;
            char    ch;
            
            if( is_blocked() )
                {
                event_eaten = false;
                break;
                }

            if( mod & ( KMOD_ALT | KMOD_CTRL ) )
                {
                break;
                }
                
            if( visible )
                {
                ch = SDL::key_to_ascii_character( key, caps, shift );
                
                if( ch == 0 )
                    {
                    break;
                    }

                // Insert character.
                if( cursor_position < current_line.size() )
                    {
                    current_line.insert( cursor_position, 1, ch );
                    }
                else
                    {
                    current_line.append( 1, ch );
                    }

                cursor_position++;
            
                discard_command_completions();
                focus_to_current();

                line_modified = true;
                redraw_command = true;
                }
            } // case default
        } // switch( key )

    // Buffer unprocessed key.
    if( !event_eaten )
        {
        hope( is_blocked() );
        
        if( key_buffer.size() < max_key_buffer_size )
            {
            key_buffer.push( KeyInfo( key, mod ) );
            }
        }
    
    return true;
    }


void Console::set_prompt
    (
    const std::string & s
    )
    {
    prompt = s;
    redraw_command = true;
    }
    

bool Console::is_blocked()
    {
    return waiting_for_command_completion || waiting_for_auto_completion;
    }


void Console::next_tick()
    {
    if( waiting_for_command_completion )
        {
        // Test if last command completed.

        hope( !command_result.is_null() );
        if( command_result->state != ::Massiv::Core::ResultObject::STATE_UNKNOWN )
            {
            command_result = ::Massiv::Core::null;
            command_completed();
            }
        }

    if( waiting_for_auto_completion )
        {
        // Test if last auto-complete request completed.

        hope( !auto_complete_result.is_null() );
        if( auto_complete_result->state == ::Massiv::Core::ResultObject::STATE_SUCCESS )
            {
            typedef METHOD_PACKET( ::Demo::Lib::ClientNode, auto_complete_console_command ) Packet;
            std::auto_ptr<Packet> packet( checked_cast<Packet *>( auto_complete_result->create_results().release() ) );
            
            auto_complete_done( packet->_result );
            auto_complete_result = ::Massiv::Core::null;
            }
        else
        if( auto_complete_result->state != ::Massiv::Core::ResultObject::STATE_UNKNOWN )
            {
            ::Massiv::Core::PValueArray< ::Massiv::Core::SString > void_result;

            auto_complete_done( void_result );
            auto_complete_result = ::Massiv::Core::null;
            }
        }

    process_buffered_commands();
    
    if( !is_blocked() )
        {
        // Replay all buffered keys.

        while( !key_buffer.empty() )
            {
            const KeyInfo & key_info = key_buffer.front();
            
            key_down( key_info.key, key_info.mod );
            key_buffer.pop();

            if( is_blocked() )
                {
                // We are blocked again.
                // Leave rest of the keys in the buffer.
                
                break;
                }
            }
        }
    }


void Console::command_completed()
    {
    // The last submitted command completed.
    waiting_for_command_completion = false;
    redraw_command = true;
    }


void Console::auto_complete_done
    (
    const ::Massiv::Core::PValueArray< ::Massiv::Core::SString > & result
    )
    {
    std::size_t i;

    // The last auto-complete request completed.
    waiting_for_auto_completion = false;
    redraw_command = true;

    command_completions.clear();
    
    for( i = 0; i < result.size(); i++ )
        {
        command_completions.push_back( result[ i ] );
        }
    
    // Suggest first possible completion.
    if( command_completions.size() )
        {
        if( command_completions.size() > 1 )
            {
            std::ostringstream ss;
            ss << "-=Possibilities=-\n";

            std::string common_prefix = command_completions[ 0 ];

            for( i = 0; i < command_completions.size(); i++ )
                {
                if( command_completions[ i ] != common_prefix &&
                    common_prefix.size() > 0 )
                    {
                    std::string pom;
                    if( common_prefix.size() > command_completions[ i ].size() )
                        {
                        pom = command_completions[ i ];
                        common_prefix = common_prefix.substr( 0, command_completions[ i ].size() );
                        }
                    else
                        {
                        pom = command_completions[ i ].substr( 0, common_prefix.size() );
                        }

                    while( pom != common_prefix )
                        {
                        pom = pom.substr( 0, pom.size() - 1 );
                        common_prefix = common_prefix.substr( 0, common_prefix.size() - 1 );
                        }

                    }
                
                if( i <= 50 )
                    {
                    ss << command_completions[ i ] << "\n";
                    }
                else
                    {
                    ss << "... and more\n";
                    break;
                    }
                }
            
            print( ss.str() ); 
            set_current( common_prefix );
            }
        else
            {
            // Disable completion cycling because there is only one completion.
            set_current( command_completions.front() + " " );
            discard_command_completions();
            }
        line_modified = true;
        }
    }


void Console::discard_command_completions()
    {
    command_completions.clear();
    completion_index = 0;
    }


void Console::focus_to_current()
    {
    if( display.size() - display.get_position() > display_text->get_rows() - 2 )
        {
        display.set_position( display.size() - display_text->get_rows() + 2 );
        if( sb.get() )
            {
            sb->set_scroll_position( display.get_position() );
            }
        }
    }


void Console::set_current
    (
    const std::string & s
    )
    {
    current_line = s;
    cursor_position = current_line.size();
    line_modified = false;
    redraw_command = true;
    }


void Console::submit_current()
    {
    ColorString color_line( current_line );
    color_line.set_color( SColor( 0.9f, 0.9f, 0.9f ), 0 );
    
    // Update console.
    if( line_modified )
        {
        // New command.
        commands.push( current_line );
        print( color_line );
        
        // Add command to history.
        commands.set_position( commands.size() - 1 );
        }
    else
        {
        // Recalling old command from the history.
        print( color_line );
        }

    history_browsed = false;
    focus_to_current();

    if( current_line.size() )
        {
        // Send the command to the local ClientNode object.
        Massiv::Core::Remote<Demo::Lib::ClientNode> my_node = 
            Massiv::System::get_local_node_id();

        command_result =
            my_node->param( ::Massiv::Core::RPC_RESULTS ).async_process_console_command
                ( current_line );

        waiting_for_command_completion = true;
        }
    }


void Console::submit_chat()
    {
    if( !current_line.size() )
        {
        return;    
        }
    
    // Update console.
    if( line_modified )
        {
        // New command.
        commands.push( current_line );
        
        // Add command to history.
        commands.set_position( commands.size() - 1 );
        }

    history_browsed = false;

    // Send chat message to my account object.
    Massiv::Core::Remote<Demo::Lib::AccountObject> 
        my_account = Global::get_game_context().require_account();

    my_account->async_say( current_line );
    }


void Console::auto_complete_current()
    {
    std::string partial_command;
    
    completion_index = 0;

    // Determine what part of current line will be sent to auto-completion.
    partial_command = current_line.substr( 0, cursor_position );

    // Send auto-complete request to the local ClientNode object.
    Massiv::Core::Remote<Demo::Lib::ClientNode> my_node = 
        Massiv::System::get_local_node_id();
    
    auto_complete_result = my_node->param
        ( ::Massiv::Core::RPC_RESULTS ).async_auto_complete_console_command( partial_command );
    
    waiting_for_auto_completion = true;
    redraw_command = true;
    }


void Console::clear_display()
    {
    display.set_position( 0 );
    display.clear_log();
    
    print_welcome();
    
    redraw_display = true;
    }


void Console::show
    (
    bool visible
    )
    {
    should_be_visible = visible;
    redraw_display = true;
    redraw_command = true;
    cursor_position = current_line.size();
    }


void Console::redraw_command_line
    (
    unsigned int        rows,
    std::string  &      command
    )
    {
    if( !( visible && redraw_command ) )
        {
        return;
        }
    // Resize overlay if row count has changed.

    int printable_rows = rows;
       
    if( rows != static_cast<UInt32>( command_text->get_rows() ) )
        {
        redraw_display = true;
        printable_rows = ( ( int ) rows > num_lines - 2 ) ? num_lines - 2: rows;
        command_text->set_base_offset
            ( 
            SVector2
                ( 
                command_text->get_base_offset().x(),
                static_cast< Float32 >
                    ( 
                    ( num_lines - 2 ) * cell_height - ( printable_rows - 1 ) * cell_height 
                    )
                )
            );
        command_text->set_columns_rows( num_columns, printable_rows );
        command_text->set_base_size( SVector2
            (
            static_cast< Float32 >( cell_width * num_columns ),
            static_cast< Float32 >( cell_height * printable_rows )
            ) );
        }
    else
        {
        command_text->clear();
        }

    // Print the text.

    std::string::size_type position = 0;
    int y = 0;
    
    for( int i = rows; i > printable_rows; i-- )
        {
        command = command.substr( num_columns );
        }
    
    while( position < command.size() )
        {
        std::string::size_type end = position + num_columns;
        if( end > command.size() )
            {
            end = command.size();
            }
        command_text->print( 0, y, command.substr( position, end - position ) );
        position = end;
        y++;
        }

    redraw_command = false;

    // Set command-line color.

    if( is_blocked() )
        {
        const SColor color( 1.0f, 0.5f, 0.5f );
        command_text->set_base_color( color );
        cursor_text->set_base_color( color );
        }
    else
        {
        const SColor color( 1.0f, 1.0f, 1.0f );
        command_text->set_base_color( color );
        cursor_text->set_base_color( color );
        }
    }


void Console::process_buffered_commands()
    {
    if( is_blocked() )
        {
        return;
        }    
    if( !buffered_command_list.size() )
        {
        // No command to process.
        return;
        }
    std::string command = buffered_command_list.front();
    buffered_command_list.pop_front();

    // Send the command to the local ClientNode object.
    Massiv::Core::Remote<Demo::Lib::ClientNode> my_node = 
        Massiv::System::get_local_node_id();

    command_result =
        my_node->param( ::Massiv::Core::RPC_RESULTS ).async_process_console_command( command );

    waiting_for_command_completion = true;
    }
    
    
void Console::push_front_command_to_buffered_command_list
    (
    const std::string & command
    )
    {
    buffered_command_list.push_front( command );
    }


void Console::push_back_command_to_buffered_command_list
    (
    const std::string & command
    )
    {
    buffered_command_list.push_back( command );
    }


bool Console::command_line_click()
    {
    SVector2 position = Global::get_input_manager().get_mouse_position(); 
    Float32 top_offset = get_base_offset().y() + command_text->get_base_offset().y();
    Float32 bottom_offset = get_base_offset().y() + 
                            command_text->get_base_offset().y()+ 
                            command_text->get_base_size().y();
                            
    if( ( position.y() >= top_offset ) && 
        ( position.y() <= bottom_offset ) 
        )
        {
        return true;
        }
    return false;
    }


void Console::set_cursor_position()
    {
    hope( command_line_click() );
    
    SVector2 position = Global::get_input_manager().get_mouse_position(); 
    UInt32 new_x = static_cast< UInt32>
            ( 
            position.x() / cell_width
            );
    UInt32 new_y = static_cast< UInt32>
            ( 
            ( position.y() - get_base_offset().y() - command_text->get_base_offset().y() ) / cell_height
            );
    Int32 new_cursor_position = new_x - prompt.size();
    while( new_y > 0 )
        {
        new_cursor_position += num_columns;
        new_y--;
        }
    if( new_cursor_position > ( Int32 ) current_line.size() )
        {
        new_cursor_position = current_line.size();
        }
    if( new_cursor_position < 0 )
        {
        new_cursor_position = 0;
        }
    
    cursor_position = new_cursor_position;
    }


bool Console::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    if( Widget::mouse_button_down( button, mod ) )
        {
        return true;
        }
        
    if( !should_be_visible )
        {
        return false;
        }
        
    if( button == SDL_BUTTON_LEFT )
        {
        set_focus();
        redraw_command = true;
        
        if( top_border_click() )
            {
            resizing = true;
            }
        else
            {
            if( command_line_click() )
                {
                set_cursor_position();
                }
            }
        return true;
        }
    else
    if( ( button == SDL_BUTTON_WHEELUP ) || ( button == SDL_BUTTON_WHEELDOWN ) )
        {
        set_focus();
        const SDLKey key = ( button == SDL_BUTTON_WHEELDOWN ) ? SDLK_DOWN : SDLK_UP;
        int line_count = ( mod & KMOD_SHIFT ) ? 15 : 5;
        if( line_count >= num_lines )
            {
            line_count = num_lines - 1;
            }
        while( line_count > 0 )
            {
            key_down( key, static_cast<SDLMod>( KMOD_SHIFT ) );
            line_count--;
            }
        return true;
        }

    return false;
    }
    

void Console::mouse_button_grab_lost
    (
    MouseButtonGrab grab
    )
    {
    resizing = false;
    }


void Console::scroll_bar_position_changed
    (
    ScrollBar * scroll_bar
    )
    {
    display.set_position( scroll_bar->get_scroll_position() );
    }


void Console::redraw_displays
    (
    unsigned int rows
    )
    {
    History::Iterator   it = display.begin();
    int                 y = 0;
    bool                sb_visible = false;

    display_text->clear();
    
    if( ( !display.is_front() && display.size() > 1 ) || rows > 1 )
        {
        sb_visible = true;
        }
        
    for( int i = rows; i > 1 ; i-- )
        {
        if( it != display.end() )
            {
            print_in_color( display_text, y, ColorString( " " ) );
            y++;
            ++it;
            }
        } 
    
    while( ( it != display.end() ) && ( y < display_text->get_rows() - 2 ) )
        {
        ColorString s = *it;

        y++;
        ++it;
        
        if( s.size() && ( y < display_text->get_rows() - 1 || it == display.end() ) )
            {
            print_in_color( display_text, y - 1, s );
            }
        }
    
    if( it != display.end() && rows == 1 )
        {
        sb_visible = true;
        }
        
    if( sb_visible )
        {
        int size = display.size();
        if( size - 2 != sb->get_scroll_range().max_position )
            {
            sb->set_scroll_range( ScrollBar::ScrollRange( 0, size - 2 ) );
            focus_to_current();
            }
        
        sb->set_scroll_position( display.get_position(), false );
        }
    else
        {
        sb->set_scroll_range( ScrollBar::ScrollRange( 0, 0 ) );
        }
                
    redraw_display = false;
    }


void Console::redraw()
    {
    hope( display_text.get() );
    hope( command_text.get() );
    hope( cursor_text.get() );

    // Compute how many lines the command spans.

    std::string command = prompt + current_line;

    // Is the command prompt shorter than one line?
    hope( command.size() - current_line.size() < ( unsigned int ) num_columns );
    
    const unsigned int rows =
            (
            command.size() ?
                ( command.size() + num_columns ) / num_columns
                :
                1
            );
    
    hope( rows > 0 );
    
    // Redraw display?

    redraw_displays( rows );
            
    // Redraw command line?
    redraw_command_line( rows, command );

    // Animations
    redraw_animations( rows );     
    }


void Console::set_write_filename
    (
    const std::string & filename
    )
    {
    display.set_output_file( filename );
    }


} // namespace Client
} // namespace Demo
