/** @file client_console.h
 * Client console.
 */

/*
 * begin        : Thu Mar 13 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_CONSOLE_H
#define CLIENT_CONSOLE_H

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_INPUT_OBJECT_H
#include "input/input_object.h"
#endif

#ifndef CLIENT_CLIENT_SDL_H
#include "sdl/client_sdl.h"
#endif

#ifndef CLIENT_OPENGL_H
#include "renderer/opengl.h"
#endif

#ifndef CLIENT_WIDGET_INPUT_DISPATCHER_H
#include "gui/widget_input_dispatcher.h"
#endif

#ifndef CLIENT_INPUT_BUTTON_H
#include "gui/icon_button.h"
#endif

#ifndef CLIENT_PANEL_BORDER_H
#include "gui/panel_border.h"
#endif

#ifndef COLOR_STRING_H
#include "gui/color_string.h"
#endif

#ifndef CLIENT_SCROLL_BAR_H
#include "gui/scroll_bar.h"
#endif

#include <memory>
#include <deque>
#include <string>
#include <vector>

namespace Demo {
namespace Client {


/**
 * Console.
 */
class Console :
    public Widget,
    public ScrollBar::ScrollBarMessageReceiver
    {
public:

    /**
     * Console switcher.
     */
    class ConsoleSwitcher : public InputObject
        {
    public:

        ConsoleSwitcher();

        ~ConsoleSwitcher();

    public:

        // InputManager.
        virtual bool key_down
            (
            SDLKey  key,
            SDLMod  mod
            );

        void initialize
            (
            );

        void enable
            (
            bool new_enabled = true
            );

    private:

        bool    enabled;

        bool    chat_visible;
        };

    /**
     * Simple history.
     */
    class History
        {
    public:

        typedef std::deque<ColorString> Log;
        typedef Log::const_iterator     Iterator;

        /**
         * Constructor.
         */
        History
            (
            int max_size
            );


        /**
         * Pushes @c s into log.
         * Old entries are thrown away.
         */
        void push
            (
            const ColorString & s
            );

        /**
         * Returns iterator pointing to the element at @c current_position.
         */
        Iterator begin();

        /**
         * Returns iterator pointing to the end of the log.
         */
        Iterator end();

        /**
         * Returns current history size.
         */
        int size();

        /**
         * Changes @c current_position.
         */
        void change_position
            (
            int delta
            );

        /**
         * Sets @c current_position.
         */
        void set_position
            (
            int position
            );

        /**
         * Returns @c current_position.
         */
        int get_position();

        /**
         * Tests if @c current_position points to the beginning of the log
         * (the oldest entry).
         */
        bool is_front();

        /**
         * Tests if @c current_position points to the end of the log
         * (the newest entry).
         */
        bool is_back();

        /**
         * Returns entry at @c current_position.
         */
        const ColorString & current();

        /**
         * Clears all data.
         */
        void clear_log();

        /**
         * Sets output filename.
         */
        void set_output_file
            (
            const std::string & filename
            );

    private:

        Log     log;
            /**< Log. */
        UInt32  max_size;
            /**< Maximum log size. */
        UInt32  current_position;
            /**< Current position within the log. */
        std::string write_filename;
        /**< Filename to write output to. */

        }; // History

    /**
     * Key info.
     * Used to replay key presses when console unblocks.
     */
    class KeyInfo
        {
    public:

        /** Constructor. */
        KeyInfo
            (
            SDLKey  key,
            SDLMod  mod
            );

    public:

        SDLKey  key;
            /**< Key. */
        SDLMod  mod;
            /**< Key modifier. */
        }; // KeyInfo

    typedef std::queue<KeyInfo>         KeyBuffer;
    typedef std::vector<std::string>    CommandCompletions;

public:

    /**
     * Constructor.
     */
    Console();

    /**
     * Prints string to the console.
     * @param string ColorString to print.
     */
    void print
        (
        const ColorString & s
        );

public:

    // InputManager.
    virtual bool key_up
        (
        SDLKey  key,
        SDLMod  mod
        );

    // InputManager.
    virtual bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );

    // InputManager.
    virtual void mouse_button_grab_lost
        (
        MouseButtonGrab grab
        );

    // From ScrollBar::ScrollBarMessageReceiver
    virtual void scroll_bar_position_changed
        (
        ScrollBar * scroll_bar
        );

public:

    /**
     * Redraws console.
     */
    void redraw();

    /**
     * Initializes console.
     */
    virtual void initialize() = 0;

protected:

    /**
     * Prints welcome message to the first line of console.
     */
    virtual void print_welcome() = 0;

    /**
     * Redraws chat console or console displays. Called from redraw.
     * @param rows Rows reserverd for command line.
     */
    void redraw_displays
        (
        unsigned int rows
        );

    /**
     * Redraw animations. Called from redraw.
     * @param rows Rows reserverd for command line.
     */
    virtual void redraw_animations
        (
        int rows
        ) = 0;

     /**
      * Returns true when mouse clicked on panel top border.
      */
     virtual bool top_border_click() = 0;

public:

    /**
     * Sets console prompt.
     */
    void set_prompt
        (
        const std::string & s
        );

    /**
     * Sets console visibility.
     */
    void show
        (
        bool visible = true
        );

    /**
     * Is console blocked (waiting for response)?
     */
    bool is_blocked();

    /**
     * Should command line be visible?
     */
    inline bool get_should_be_visible() const
        {
        return should_be_visible;
        }

    /**
     * Clears display.
     */
    void clear_display();

    /**
     * Advances console "logic".
     * Must be called periodically.
     */
    void next_tick();

protected:

    // InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );

    /**
     * Redraws command line. Called from redraw.
     * @param rows Rows reserverd for command line.
     * @param command String containing command prompt.
     */
    void redraw_command_line
        (
        unsigned int        rows,
        std::string &       command
        );

    /**
     * Prints string to the specified history on console.
     */
    void print
        (
        const ColorString & s,
        History           & history
        );

    /**
     * Prints string in color to the screen.
     * @param where is a rectangle to draw to.
     * @param row to draw to.
     * @param text to draw.
     * @param column Optional column where to start.
     */
    void print_in_color
        (
        const std::auto_ptr< Renderer::TextRectangle > &    where,
        unsigned int                                        row,
        const ColorString &                                 text,
        unsigned int                                        column = 0
        );

    /**
     * Last command completed callback.
     */
    void command_completed();

    /**
     * Last auto-complete request completed callback.
     */
    void auto_complete_done
        (
        const ::Massiv::Core::PValueArray< ::Massiv::Core::SString > & result
        );

    /**
     * Discards current @c command_completions.
     */
    void discard_command_completions();

    /**
     * Focuses display to the current line.
     */
    void focus_to_current();

    /**
     * Sets current line.
     */
    void set_current
        (
        const std::string & s
        );

    /**
     * Submits current command.
     */
    void submit_current();

    /**
     * Submits current text.
     */
    void submit_chat();

    /**
     * Auto-completes current command.
     */
    void auto_complete_current();

    /**
     * Process first command from buffered command list.
     */
    void process_buffered_commands();

    /**
     * Sets new cursor position according to current mouse pos.
     */
    void set_cursor_position();

    /**
     * Returns true when mouse clicked into command_text area.
     */
    bool command_line_click();

public:

    /**
     * Push front command to list of buffered commands.
     */
    void push_front_command_to_buffered_command_list
        (
        const std::string & command
        );

    /**
     * Push back command to list of buffered commands.
     */
    void push_back_command_to_buffered_command_list
        (
        const std::string & command
        );

    /**
     * Sets filename to write output to.
     */
    void set_write_filename
        (
        const std::string & filename
        );

protected:

    //
    // Console display.
    //

    History                 display;
        /**< History log of displayed lines. */

    static History          commands;
        /**< History of submitted commands. */

    int                     num_lines;
        /**< Number of lines that can be displayed on the screen. */

    int                     num_columns;
        /**< Number of columns that can be displayed on the screen. */

    static std::string      current_line;
        /**< Current line. */

    UInt32                  cursor_position;
        /**< Index into @c current_line. */

    bool                    history_browsed : 1;
        /**< Did user browse command history since the last command submit? */

    bool                    line_modified : 1;
        /**< Did we modify current line? */

    bool                    should_be_visible : 1;
        /**< Should be console visible? */

    bool                    loose_focus_on_submit : 1;
        /**< Should console loose focus whenever command is submitted? */

    std::string             prompt;
        /**< Command line prompt. */

protected:

    //
    // Overlay objects.
    //
    int                                         cell_width;
        /**< Width of TextRectangle cell. */

    int                                         cell_height;
        /**< Height of TextRectangle cell. */

    std::auto_ptr< Renderer::TextRectangle >    display_text;
        /**< Currently visible part of #display. */

    std::auto_ptr< Renderer::TextRectangle >    command_text;
        /**< Command line overlay object. */

    std::auto_ptr< Renderer::TextRectangle >    cursor_text;
        /**< Cursor overlay object. */

    std::auto_ptr< PanelBorder >                panel_border;
        /**< Background panel. */

    std::auto_ptr< ScrollBar >                  sb;
        /**< Optional scroll bar. */

    bool                                        redraw_display : 1;
        /**< Display has changed and must be redrawn. */

    bool                                        redraw_command : 1;
        /**< Command line has changed and must be redrawn. */

    bool                                        resizing;
        /**< Whether console is resizing. */

private:

    //
    // State.
    //

    KeyBuffer               key_buffer;
        /**< Buffers events that occurs when console is blocked.
             The buffer is replayed when console unblocks. */

protected:

    //
    // Command processing.
    //

    bool                        waiting_for_auto_completion : 1;
        /**< Is the console waiting for auto-complete result? */

    bool                        waiting_for_command_completion : 1;
        /**< Is the console waiting for command completion? */

    CommandCompletions          command_completions;
        /**< Auto-complete result. List of possible command completions. */

    int                         completion_index;
        /**< Index into @c command_completions. */

    std::deque< std::string >   buffered_command_list;
        /**< List of the buffered commands used for batch execution. */

protected:

    //
    // Pending RPC.
    //

    ::Massiv::Core::Pointer< ::Massiv::Core::ResultObject > command_result;
        /**< Result object from the last submitted command. */

    ::Massiv::Core::Pointer< ::Massiv::Core::ResultObject > auto_complete_result;
        /**< Result object from the last auto-complete request. */

    }; // Console

} // namespace Client
} // namespace Demo

#endif // CLIENT_CONSOLE_H
