/** @file client_full_console.cpp
 * Full console implementation.
 */


/*
 * begin        : Thu Mar 13 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "gui/client_full_console.h"
#include "gui/client_chat_console.h"
#include "sdl/client_sdl.h"
#include "renderer/layers.h"
#include "renderer/font.h"
#include <cstring>
#include <sstream>
namespace Demo {
namespace Client {


// Implementation of FullConsole
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
FullConsole::FullConsole()
      : Console()
    {
    set_focusable( CLICK_FOCUSABLE );
    }


void FullConsole::initialize()
    {
    set_base_size_from_screen();
    set_auto_bringable_to_front( true );

    // Get font.
    Renderer::Font & font = Global::get_console_font();

    // Compute sizes.

    font.get_texture(); // initializes font info as a side effect
    cell_width = font.get_character_width();
    cell_height = font.get_character_height();
    num_columns = ( SDL::get_screen_width() - 2 * cell_width ) / cell_width;
    num_lines = SDL::get_screen_height() / cell_height;

    
    // Create overlay objects.

    const SColor color_speed( 1.0f, 1.0f, 1.0f, 1.0f );
    const SColor initial_factor( 1.0f, 1.0f, 1.0f, 0.0f );

    // My size.
    
    set_base_size
        (
        SVector2
            ( 
            static_cast< Float32 >( SDL::get_screen_width() ),
            static_cast< Float32 >( SDL::get_screen_height() )
            )
        );
        
    set_base_offset
        (
        SVector2
            (
            0.0f, 
            0.0f
            ) 
        );

    add_animation_object( new Renderer::OverlayOffsetAnimation
        (
        SVector2( 1.0f, static_cast< Float32 >( SDL::get_screen_height() ) ),
        get_base_offset()
        ) );

    add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( color_speed, initial_factor ) );
    
    // Create panels.

    panel_border.reset( new PanelBorder
        (
        "gui/panel_corner",
        "gui/panel_hborder",
        "gui/panel_vborder",
        "gui/panel_background",
        0.0f,
        48.0f
        ) );
        
    panel_border->set_base_size_from_screen();
    panel_border->set_parent_widget( this );
        
    panel_border->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( color_speed, initial_factor ) );
        
    panel_border->set_type( Widget::TRANSPARENT_WIDGET );
    
    //scroll bar
    sb.reset( new ScrollBar( ScrollBar::VERTICAL ) );
    sb->set_base_offset
        ( 
        SVector2
            ( 
            static_cast< Float32 > ( SDL::get_screen_width() - 2 * cell_width ),
            static_cast< Float32 > ( cell_height ) 
            ) 
        );
    sb->set_base_size
        (
        SVector2
            ( 
            static_cast< Float32 > ( 2 * cell_width ), 
            static_cast< Float32 > ( panel_border->get_base_size().y() - 2 * cell_height )
            )
        );
    sb->do_layout();
    sb->set_scroll_range( ScrollBar::ScrollRange( 0, 100 ) );
    sb->set_scroll_position( 0 );
    sb->set_scroll_line_size( 1 );
    sb->set_scroll_page_size( 10 );
    sb->set_parent_widget( this );
    sb->set_message_receiver( this );
    
    // Create rectangles.
    
    display_text.reset( new Renderer::TextRectangle( &font, num_columns, num_lines - 1 ) );
    display_text->set_base_offset( SVector2( 0.0f, static_cast< Float32 >( SDL::get_screen_height() ) ) );
    
    display_text->add_animation_object( new Renderer::OverlayOffsetAnimation
        (
        SVector2( 1.0f, static_cast< Float32 >( SDL::get_screen_height() ) ),
        display_text->get_base_offset()
        ) );
    display_text->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( color_speed, initial_factor ) );
    display_text->set_parent_overlay_object( this );

    command_text.reset( new Renderer::TextRectangle( &font, num_columns ) );
    command_text->set_base_offset
        ( SVector2( 0.0f, static_cast< Float32 >( SDL::get_screen_height() - 2 * cell_height ) ) );

    command_text->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( color_speed, initial_factor ) );
    command_text->set_parent_overlay_object( this );

    cursor_text.reset( new Renderer::TextRectangle( &font ) );
    cursor_text->set_base_offset
        ( SVector2( 0.0f, static_cast< Float32 >( SDL::get_screen_height() + 4 ) ) );
    cursor_text->add_animation_object( new Renderer::OverlayOffsetAnimation
        (
        SVector2
            (
            static_cast< Float32 >( SDL::get_screen_width() ), 
            static_cast< Float32 >( SDL::get_screen_height() / 8 ) 
            ),
        cursor_text->get_base_offset()
        ) );
    cursor_text->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( color_speed, initial_factor ) );
    cursor_text->print( 0, 0, "_" );
    cursor_text->set_parent_overlay_object( this );

    redraw_display = true;
    redraw_command = true;

    // Done just prepare both consoles.
    clear_display();
    }
    

void FullConsole::redraw_animations
    (
    int rows
    )
    {
    // Animate alpha.

    const SColor target_color_factor( 1.0f, 1.0f, 1.0f, ( should_be_visible ) ? 1.0f : 0.0f );
   
    require_color_factor_animation().set_target_factor( target_color_factor );
   
    display_text->require_color_factor_animation().set_target_factor( target_color_factor );
    panel_border->require_color_factor_animation().set_target_factor( target_color_factor );
        
    command_text->require_color_factor_animation().set_target_factor( target_color_factor );
    cursor_text->require_color_factor_animation().set_target_factor( target_color_factor );

    // Animate positions.

    const Float32 screen_height = static_cast< Float32 >( SDL::get_screen_height() );
    Float32 command_y = screen_height;
    Float32 display_y = 0.0f;
    Float32 position = 0.0f;
    
    if( should_be_visible )
        {
        command_y -= command_text->get_base_size().y() + 20.0f;

        display_y = static_cast< Float32 >( -cell_height * ( command_text->get_rows() - 2 ) );
        position = 0.0f;
        }
    else
        {
        display_y = -display_text->get_base_size().y();
        position = -panel_border->get_base_size().y();
        }

    require_offset_animation().set_target_offset( SVector2( 0.0f, position ) );
    display_text->require_offset_animation().set_target_offset( SVector2( 0.0f, display_y ) );
    
    // Cursor position.
    
    int cursor_offset;
    
    cursor_offset = cursor_position + prompt.size();
    
    const int printable_rows = ( rows > num_lines - 2 ) ? num_lines - 2: rows;
        
    const int cursor_row = cursor_offset / command_text->get_columns() - ( rows - printable_rows );
    const int cursor_column = cursor_offset % command_text->get_columns();
    
    cursor_text->require_offset_animation().set_target_offset
        (
        SVector2
            (
            static_cast< Float32 >( cell_width * cursor_column ),
            static_cast< Float32 >( command_text->get_current_offset().y() + cursor_row * cell_height )
            )
        );
    }
    
    
void FullConsole::print_welcome()
    {
    ColorString color_str;
    
    color_str.set_color( SColor( 1.0f, 0.8f, 0.2f ) );
    color_str = "Welcome to console window. Use ";
    color_str.set_color( SColor( 0.2f, 1.0f, 0.4f ) );
    color_str += "/help";
    color_str.set_color( SColor( 1.0f, 0.8f, 0.2f ) );
    color_str += " for help on console commands.";
                
    print( color_str );
    }


bool FullConsole::top_border_click()
    {
    return false;
    }
    
} // namespace Client
} // namespace Demo
