/** @file edit_box.cpp
 * Simple edit box (implementation).
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "gui/edit_box.h"
#include "client_global.h"
#include "input/input_manager.h"
#include "renderer/font.h"


namespace Demo {
namespace Client {


EditBox::EditBox
    (
    Type    the_type
    ) :
    type( the_type ),
    cursor_position( 0 ),
    max_text_length( 1000 ),
    message_receiver( NULL )
    {
    set_focusable( TAB_CLICK_FOCUSABLE );
    text_rectangle.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    text_rectangle->set_parent_overlay_object( this );
    
    white.reset( new Renderer::Icon( "texture/white" ) );
    }


void EditBox::set_text
    (
    const std::string & new_text
    )
    {
    text = new_text.substr( 0, max_text_length );
    cursor_position = text.size();
    print_text();
    }


std::string EditBox::get_text() const
    {
    return text;
    }


void EditBox::set_max_text_length_from_size()
    {
    SVector2    size = get_current_size();
    Float32     char_width = get_character_width();
    Float32     indexf = size.x() / char_width;
    Int32       index = Massiv::Core::FloatTraits< Float32 >::to_integer( indexf );
    
    // Leave one place for cursor at the end of the text.
    index--;
    
    if( index < 0 )
        {
        index = 1;
        }
    
    max_text_length = static_cast< std::string::size_type >( index );
    
    // Update text.
    text = text.substr( 0, max_text_length );
    fix_cursor_position();
    print_text();
    }


void EditBox::render()
    {
    // Update text rectangle position.
        {
        Float32 y = ( get_current_size().y() - text_rectangle->get_current_size().y() ) / 2.0f;
        text_rectangle->set_base_offset( SVector2( 0.0f, y ) );
        }
    
    render_background();
    render_frame();
    if( has_focus() )
        {
        render_cursor();
        }
    }


bool EditBox::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    if( Super::mouse_button_down( button, mod ) )
        {
        return true;
        }

    if( button == SDL_BUTTON_LEFT )
        {
        // We already have focus. User clicked to select cursor position.
        
        // Select cursor position from the click.
        SVector2    top_left = get_top_left_corner_on_screen();
        SVector2    mouse_pos = Global::get_input_manager().get_mouse_position();
        SVector2    delta = mouse_pos - top_left;
        Float32     dx = delta.x();
        Float32     char_width = get_character_width();
        Float32     indexf = dx / char_width;
        Int32       index = Massiv::Core::FloatTraits< Float32 >::to_integer( indexf );
        
        cursor_position = static_cast< std::string::size_type >( index );
        fix_cursor_position();
        
        return true;
        }
    
    return false;
    }


bool EditBox::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( Super::key_down( key, mod ) )
        {
        return true;
        }
    
    bool                    shift = false;
    bool                    caps = false;
    
    std::string::value_type c;
    
    if( ( mod & KMOD_SHIFT ) != 0 ) shift = !shift;
    if( ( mod & KMOD_CAPS ) != 0 ) caps = !caps;
    
    c = SDL::key_to_ascii_character( key, caps, shift );
    
    if( c != 0 )
        {
        put_char( c );
        return true;
        }
        
    switch( key )
        {
        case SDLK_BACKSPACE:
            erase_previous_char();
            return true;

        case SDLK_DELETE:
            erase_char();
            return true;
        
        case SDLK_LEFT:
            cursor_left();
            return true;
        
        case SDLK_RIGHT:
            cursor_right();
            return true;
        
        case SDLK_HOME:
            cursor_begin();
            return true;
        
        case SDLK_END:
            cursor_end();
            return true;

        default: // to get rid of tons of warnings...
            break;
        }
    
    return false;
    }


void EditBox::render_background()
    {
    SVector2    top_left = get_top_left_corner_on_screen();
    SVector2    size = get_current_size();
    
    if( has_focus() )
        {
        white->set_base_color( SColor( 0.7f, 0.65f, 0.50f, 0.5f ) );
        }
    else
        {
        white->set_base_color( SColor( 0.0f, 0.0f, 0.0f, 0.6f ) );
        }
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->render();
    }


void EditBox::render_frame()
    {
    const SColor    dark( 0.0f, 0.0f, 0.0f, 0.6f * get_current_color().alpha() );
    const SColor    light( 1.0f, 1.0f, 1.0f, 0.6f * get_current_color().alpha() );
    const SVector2  top_left = screenspace_translation().svector();
    const SVector2  size = get_current_size();
    bool            focus = has_focus();
    
    focus = true;
    
    render_hline( focus ? dark : light, top_left.x(), top_left.x() + size.x(), top_left.y() );
    render_vline( focus ? dark : light, top_left.x(), top_left.y(), top_left.y() + size.y() );
    render_hline( focus ? light : dark, top_left.x(), top_left.x() + size.x(), top_left.y() + size.y() - 1.0f );
    render_vline( focus ? light : dark, top_left.x() + size.x() - 1.0f, top_left.y(), top_left.y() + size.y() );
    }


void EditBox::render_hline
    (
    const SColor &  color,
    Float32         x1,
    Float32         x2,
    Float32         y
    )
    {
    const SVector2 top_left( x1, y );
    const SVector2 size( x2 - x1, 1.0f );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


void EditBox::render_vline
    (
    const SColor &  color,
    Float32         x,
    Float32         y1,
    Float32         y2
    )
    {
    const SVector2 top_left( x, y1 );
    const SVector2 size( 1.0f, y2 - y1 );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


void EditBox::render_cursor()
    {
    static SystemTime   last_time = Massiv::System::system_time();
    static bool         visible = true;
    SystemTime          current_time = Massiv::System::system_time();
    
    if( ( last_time + 0.2 ).is_older_than( current_time ) )
        {
        last_time = current_time;
        visible = !visible;
        }
    
    if( !visible )
        {
        return;
        }

    Float32     char_width = get_character_width();
    Float32     char_height = get_character_height();
    Float32     offset = char_width * static_cast< Float32 >( cursor_position );
    SVector2    size = SVector2( char_width, 2.0f ); // char_height
    SVector2    top_left = get_top_left_corner_on_screen() + 
        SVector2( offset, get_current_size().y() / 2.0f + char_height / 2.0f - size.y() );
    
    white->set_base_color( SColor( 1.0f, 1.0f, 1.0f, 0.6f ) );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->render();
    }


void EditBox::print_text()
    {
    if( type == TYPE_PASSWORD )
        {
        std::string t = text;
        std::string::iterator it;
        for( it = t.begin(); it != t.end(); ++it )
            {
            ( *it ) = '*';
            }
        text_rectangle->print_line( t );
        }
    else
        {
        text_rectangle->print_line( text );
        }
    }


void EditBox::fix_cursor_position()
    {
    if( cursor_position < 0 )
        {
        cursor_position = 0;
        }
    else if( cursor_position > text.size() )
        {
        cursor_position = text.size();
        }
    }


void EditBox::put_char
    (
    std::string::value_type c
    )
    {
    if( text.size() >= max_text_length )
        {
        return;
        }

    std::string::iterator it = text.begin() + cursor_position;
    text.insert( it, c );
    cursor_position++;
    print_text();
    }


void EditBox::erase_char()
    {
    std::string::iterator it = text.begin() + cursor_position;
    if( it != text.end() )
        {
        text.erase( it );
        print_text();
        }
    }


void EditBox::erase_previous_char()
    {
    if( cursor_position > 0 )
        {
        std::string::iterator it = text.begin() + ( cursor_position - 1 );
        cursor_left();
        text.erase( it );
        print_text();
        }
    }


void EditBox::cursor_left()
    {
    if( cursor_position > 0 )
        {
        cursor_position--;
        }
    }


void EditBox::cursor_right()
    {
    if( cursor_position < text.size() )
        {
        cursor_position++;
        }
    }


void EditBox::cursor_begin()
    {
    cursor_position = 0;
    }


void EditBox::cursor_end()
    {
    cursor_position = text.size();
    }


Float32 EditBox::get_character_width() const
    {
    Renderer::Font & font = Global::get_console_font();
    return static_cast< Float32 >( font.get_character_width() );
    }


Float32 EditBox::get_character_height() const
    {
    Renderer::Font & font = Global::get_console_font();
    return static_cast< Float32 >( font.get_character_height() );
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "gui/edit_box_inline.h"
#endif
