/** @file edit_box.h
 * Simple edit box.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_EDIT_BOX_H
#define CLIENT_EDIT_BOX_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#ifndef CLIENT_TEXT_RECTANGLE_H
#include "renderer/text_rectangle.h"
#endif

#ifndef CLIENT_ICON_H
#include "renderer/icon.h"
#endif

#include <memory>
#include <string>


namespace Demo {
namespace Client {


/**
 * Simple edit box.
 */
class EditBox : public Widget
    {
    typedef Widget  Super;
    
    NO_COPY_CONSTRUCTOR( EditBox );
    NO_ASSIGNMENT_OPERATOR( EditBox );

public:

    /** Edit box type. */
    enum Type
        {
        TYPE_STANDARD,
            /**< Standard edit box. */
        TYPE_PASSWORD
            /**< Edit box used for passwords. Uses '*' character
             * to draw edit box text.
             */
        }; // Type

    /** Constructor. */
    EditBox
        (
        Type    type = TYPE_STANDARD
        );
    
    /**
     * Receiver of messages from the edit box.
     */
    class EditBoxMessageReceiver
        {
    public:
    
        }; // EditBoxMessageReceiver
        
public:

    /** Set text to edit box. */
    void set_text
        (
        const std::string & new_text
        );
    
    /** Get text from the edit box. */
    std::string get_text() const;
    
    /** Set max length of text the edit box can contain from size of the edit box. */
    void set_max_text_length_from_size();
    
public:        
    
    /**
     * Set #EditBoxMessageReceiver of the edit box.
     * If @a message_receiver is @c NULL, message notifying will be
     * disabled.
     */
    INLINE void set_message_receiver
        (
        EditBoxMessageReceiver * message_receiver
        );

public:

    // From Renderer::OverlayObject
    virtual void render();
    
    // From InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );

    // From InputObject
    virtual bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );
        
public:

    /** Get character width of the edit box. */
    Float32 get_character_width() const;
    
    /** Get character height of the edit box. */
    Float32 get_character_height() const;

private:

    /** Render background. */
    void render_background();
    
    /** Render frame. */
    void render_frame();
    
    /** Render horizontal line. */
    void render_hline
        (
        const SColor &  color,
        Float32         x1,
        Float32         x2,
        Float32         y
        );
    
    /** Render vertical line. */
    void render_vline
        (
        const SColor &  color,
        Float32         x,
        Float32         y1,
        Float32         y2
        );
    
    /** Render cursor. */
    void render_cursor();
    
    /** Put @a text to @a text_rectangle. */
    void print_text();

    /** Fix @a cursor_position so its value is >= 0 and <= text.size(). */
    void fix_cursor_position();
    
    /** Put character to current cursor position. */
    void put_char
        (
        std::string::value_type c
        );
    
    /** Erase character at cursor position. */
    void erase_char();
    
    /** Erase character before cursor position. */
    void erase_previous_char();
    
    /** Move cursor left. */
    void cursor_left();
    
    /** Move cursor right. */
    void cursor_right();
    
    /** Move cursor to the beginning of the text. */
    void cursor_begin();
    
    /** Move cursor to the end of the text. */
    void cursor_end();

private:

    Type                    type;
        /**< Edit box type. */

    std::string::size_type  cursor_position;
        /**< Edit box cursor position.
         * Always >= 0 and <= text.size().
         */
    
    std::string::size_type  max_text_length;
        /**< Max text length. */

    std::string             text;
        /**< Edit box text. */

    std::auto_ptr< Renderer::TextRectangle >    text_rectangle;
        /**< Edit box text rectnagle. */

    std::auto_ptr< Renderer::Icon >             white;
        /**< White texture used to draw lines and background. */

protected:

    EditBoxMessageReceiver * message_receiver;
        /**< Message receiver. */

    }; // Button


} // namespace Demo::Client
} // namespalice Demo

#ifdef ENABLE_INLINE
#include "gui/edit_box_inline.h"
#endif

#endif // CLIENT_EDIT_BOX_H
