/** @file editor.cpp
 * Editor (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "client_global.h"
#include "gui/editor.h"
#include "gui/widget_input_dispatcher.h"
#include "gui/client_full_console.h"
#include "game/game_context.h"
#include "lib/shared/account_object.h"
#include "lib/shared/console.h"
#include "lib/shared/account_object_rpc.h"
#include "lib/client/client_node.h"
#include "lib/client/client_node_rpc.h"
#include "lib/shared/client_sector.h"
#include <sstream>
#include <iostream>


namespace Demo {
namespace Client {


Editor::Editor() :
    mode( MODE_NONE ),
    visible( true )
    {
    // We can't create widgets now, because media repository
    // is not initialized yet.
    }


Editor::~Editor()
    {
    }


void Editor::initialize()
    {
    Widget * parent_widget = &Global::get_widget_input_dispatcher();
    hope( parent_widget );

    // Create widgets.

    editor_mode_panel.reset( new EditorModePanel() );
    editor_mode_panel->set_base_offset( SVector2( 64.0f, 0.0f ) );
    editor_mode_panel->set_parent_widget( parent_widget );
    editor_mode_panel->do_layout();

    modify_elevation_panel.reset( new ModifyElevationPanel() );
    modify_elevation_panel->set_base_offset( SVector2( 100.0f, 100.0f ) );
    modify_elevation_panel->set_parent_widget( parent_widget );

    editor_material_panel.reset( new EditorMaterialPanel() );
    editor_material_panel->set_base_offset( SVector2( 100.0f, 100.0f ) );
    editor_material_panel->set_parent_widget( parent_widget );
    editor_material_panel->do_layout();

    editor_add_entity_panel.reset( new EditorAddEntityPanel() );
    editor_add_entity_panel->set_base_offset( SVector2( 100.0f, 100.0f ) );
    editor_add_entity_panel->set_parent_widget( parent_widget );
    editor_add_entity_panel->do_layout();

    editor_edit_entity_panel.reset( new EditorEditEntityPanel() );
    editor_edit_entity_panel->set_base_offset( SVector2( 100.0f, 100.0f ) );
    editor_edit_entity_panel->set_parent_widget( parent_widget );
    editor_edit_entity_panel->do_layout();

    editor_delete_entity_panel.reset( new EditorDeleteEntityPanel() );
    editor_delete_entity_panel->set_base_offset( SVector2( 100.0f, 100.0f ) );
    editor_delete_entity_panel->set_parent_widget( parent_widget );
    editor_delete_entity_panel->do_layout();

    mode = MODE_NONE;

    set_visibility( false );
    }


void Editor::next_tick()
    {
    editor_material_panel->refresh();
    }


void Editor::set_visibility
    (
    bool the_visible
    )
    {
    visible = the_visible;
    update_visibility();
    }


bool Editor::is_visible() const
    {
    return visible;
    }


void Editor::set_mode
    (
    Mode new_mode
    )
    {
    mode = new_mode;
    update_visibility();
    }


void Editor::update_visibility()
    {
    editor_mode_panel->set_visibility( visible );

    const bool elevation_visible = visible && mode == MODE_MODIFY_ELEVATION;
    const bool material_visible = visible && mode == MODE_MODIFY_MATERIAL;
    const bool add_entity_visible = visible && mode == MODE_ADD_ENTITY;
    const bool edit_entity_visible = visible && mode == MODE_EDIT_ENTITY;
    const bool delete_entity_visible = visible && mode == MODE_DELETE_ENTITY;

    modify_elevation_panel->set_visibility( elevation_visible );
    if( elevation_visible )
        {
        modify_elevation_panel->set_focus();
        }
    editor_material_panel->set_visibility( material_visible );
    if( material_visible )
        {
        editor_material_panel->set_focus();
        }
    editor_add_entity_panel->set_visibility( add_entity_visible );
    if( add_entity_visible )
        {
        editor_add_entity_panel->set_focus();
        }
    editor_edit_entity_panel->set_visibility( edit_entity_visible );
    if( edit_entity_visible )
        {
        editor_edit_entity_panel->set_focus();
        }
    editor_delete_entity_panel->set_visibility( delete_entity_visible );
    if( delete_entity_visible )
        {
        editor_delete_entity_panel->set_focus();
        }
    }


bool Editor::handle_sector_click
    (
    SDLMod mod,
    const SectorManager::RayIntersection & sector_intersection
    )
    {
    bool result = false;

    switch( mode )
        {
        case MODE_NONE:
            break;

        case MODE_MODIFY_ELEVATION:
            send_modify_elevation_command( mod, sector_intersection );
            result = true;
            break;

        case MODE_MODIFY_MATERIAL:
            send_modify_material_command( sector_intersection );
            result = true;
            break;

        case MODE_ADD_ENTITY:
            send_add_entity_command( sector_intersection );
            result = true;
            break;

        case MODE_EDIT_ENTITY:
            // Modify position x and position y.
            editor_edit_entity_panel->set_position_coordinates( sector_intersection.intersection.x(), sector_intersection.intersection.y() );
            break;

        case MODE_DELETE_ENTITY:
            break;

        default:
            hope( false );
        }

    return result;
    }


bool Editor::handle_entity_click
    (
    SDLMod mod,
    const EntityManager::RayIntersection & entity_intersection
    )
    {
    bool result = false;

    switch( mode )
        {
        case MODE_NONE:
            break;

        case MODE_MODIFY_ELEVATION:
            break;

        case MODE_MODIFY_MATERIAL:
            break;

        case MODE_ADD_ENTITY:
            break;

        case MODE_EDIT_ENTITY:
            editor_edit_entity_panel->set_entity_intersection( entity_intersection );
            result = true;
            break;

        case MODE_DELETE_ENTITY:
            editor_delete_entity_panel->set_entity_intersection( entity_intersection );
            result = true;
            break;

        default:
            hope( false );
        }

    return result;
    }


void Editor::send_modify_elevation_command
    (
    SDLMod mod,
    const SectorManager::RayIntersection & sector_intersection
    )
    {
    ModifyElevationOptions opt = modify_elevation_panel->get_modify_elevation_options();
    SVector2 center = SVector2( sector_intersection.intersection.x(), sector_intersection.intersection.y() );

    // Align to tile grid.
    if( opt.align_to_grid )
        {
        Float32 x = center.x();
        Float32 y = center.y();
        Float32 xalign = Lib::ClientSector::SECTOR_X_SIZE / static_cast< Float32 >( Lib::ClientSector::NUM_X_TILES );
        Float32 yalign = Lib::ClientSector::SECTOR_Y_SIZE / static_cast< Float32 >( Lib::ClientSector::NUM_Y_TILES );

        x = FloatTraits< Float32 >::round( x / xalign ) * xalign;
        y = FloatTraits< Float32 >::round( y / yalign ) * yalign;

        center = SVector2( x, y );
        }

    // Inverse the operation if requested.
    if( mod & KMOD_SHIFT )
        {
        opt.strength = -opt.strength;
        }

    // Generate command.
    std::ostringstream oss;

    oss << "/modify_elevation "
        << center.x() << " "
        << center.y() << " "
        << Lib::ClientSector::map_kind_to_string( opt.map ) << " "
        << Lib::ClientSector::map_function_to_string( opt.function ) << " "
        << Lib::ClientSector::map_operation_to_string( opt.operation ) << " "
        << opt.radius << " "
        << opt.strength;

    // Send modify elevation command to account.
    Remote< Lib::AccountObject > account = Global::get_game_context().get_account();
    if( account != null )
        {
        Demo::Lib::Console::Environment    env;
        account->async_process_account_command( oss.str(), env );
        }
    }


void Editor::send_modify_material_command
    (
    const SectorManager::RayIntersection & sector_intersection
    )
    {
    ModifyMaterialOptions opt = editor_material_panel->get_modify_material_options();
    SVector2 center = SVector2( sector_intersection.intersection.x(), sector_intersection.intersection.y() );

    // Align to tile grid.
    if( opt.align_to_grid )
        {
        Float32 x = center.x();
        Float32 y = center.y();
        Float32 xalign = Lib::ClientSector::SECTOR_X_SIZE / static_cast< Float32 >( Lib::ClientSector::NUM_X_TILES );
        Float32 yalign = Lib::ClientSector::SECTOR_Y_SIZE / static_cast< Float32 >( Lib::ClientSector::NUM_Y_TILES );

        x = FloatTraits< Float32 >::round( x / xalign ) * xalign;
        y = FloatTraits< Float32 >::round( y / yalign ) * yalign;

        center = SVector2( x, y );
        }

    // Send modify elevation command to account.
    Remote< Lib::AccountObject > account = Global::get_game_context().get_account();
    if( account != null )
        {
        account->async_modify_material_command
            (
            center,
            opt.radius,
            opt.material_id
            );
        }
    }


void Editor::send_add_entity_command
    (
    const SectorManager::RayIntersection & sector_intersection
    )
    {
    AddEntityOptions opt = editor_add_entity_panel->get_add_entity_options();
    if( opt.entity_class == "None" )
        {
        return;
        }
    SVector2 center = SVector2( sector_intersection.intersection.x(), sector_intersection.intersection.y() );

    // Set correct model tag.
    std::string model_tags = opt.model_tags;

    std::string custom_options = editor_add_entity_panel->get_add_entity_custom_options_string();

    Remote< Lib::AccountObject > account = Global::get_game_context().get_account();
    if( account == null )
        {
        return;
        }

    std::ostringstream  ss;
    ss << "Demo::Lib::" << opt.entity_class << " "          // Entity class name.
       << center.x() << " " << center.y() << " "            // Origin.
       << Massiv::Core::FloatTraits< Float32 >::degrees_to_radians( opt.orientation ) << " 0 0 " // Orientation.
       << sector_intersection.layer_id << " "               // Layer id.
       << ( opt.solid ? "true" : "false" ) << " "           // Is entity solid?
       << opt.bound_x << " " << opt.bound_y << " "          // Bound box.
       << ( opt.axis_aligned ? "true" : "false" ) << " "    // Bbox aligned.
       << quote_const_string( model_tags ) << " "           // Model tag.
       << custom_options;                                   // Custom parameter.

    account->async_create_entity_command
        (
        ss.str()
        );
    }


} // namespace Demo::Client
} // namespace Demo
