/** @file editor.h
 * Editor.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_EDITOR_H
#define CLIENT_EDITOR_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_EDITOR_MODE_PANEL_H
#include "editor/editor_mode_panel.h"
#endif

#ifndef CLIENT_EDITOR_MATERIAL_PANEL_H
#include "editor/editor_material_panel.h"
#endif

#ifndef CLIENT_EDITOR_ADD_ENTITY_PANEL_H
#include "editor/editor_add_entity_panel.h"
#endif

#ifndef CLIENT_EDITOR_EDIT_ENTITY_PANEL_H
#include "editor/editor_edit_entity_panel.h"
#endif

#ifndef CLIENT_EDITOR_DELETE_ENTITY_PANEL_H
#include "editor/editor_delete_entity_panel.h"
#endif

#ifndef CLIENT_EDITOR_ELEVATION_PANEL_H
#include "editor/editor_elevation_panel.h"
#endif

#ifndef CLIENT_ENTITY_MANAGER_H
#include "game/entity_manager.h"
#endif

#ifndef CLIENT_SECTOR_MANAGER_H
#include "game/sector_manager.h"
#endif


namespace Demo {
namespace Client {


/**
 * Editor.
 */
class Editor
    {
    NO_COPY_CONSTRUCTOR( Editor );
    NO_ASSIGNMENT_OPERATOR( Editor );
    
public:

    /**
     * Editor mode.
     * When changed, @c update EditorModePanel::setup_list_boxes().
     */
    enum Mode
        {
        MODE_NONE,
        MODE_MODIFY_ELEVATION,
        MODE_MODIFY_MATERIAL,
        MODE_ADD_ENTITY,
        MODE_EDIT_ENTITY,
        MODE_DELETE_ENTITY
        }; // Mode
    
public:

    /** Constructor. */
    Editor();
    
    /** Destructor. */
    ~Editor();
    
    /** Initialize. */
    void initialize();
    
    /** Editor's next tick. */
    void next_tick();
    
public:

    /** Show/hide editor. */
    void set_visibility
        (
        bool visible
        );
    
    /** Is the editor visible? */
    bool is_visible() const;

    /**
     * Update visibility of editor widgets and set keyboard input focus to
     * current edit dialog.
     */
    void update_visibility();
    
public:

    /** Switch editor to specified mode. */
    void set_mode
        (
        Mode new_mode
        );
    
public:

    /**
     * Handle mouse click to a sector.
     * @param mod Current state of keyboard modifiers.
     * @param sector_intersection Sector intersection info.
     * @return True if the event was processed (the event was "eaten"),
     *     false otherwise.
     * @note Called by #TerrainWindow.
     */
    bool handle_sector_click
        (
        SDLMod mod,
        const SectorManager::RayIntersection & sector_intersection
        );
    
    /**
     * Handle mouse click to a entity.
     * @param mod Current state of keyboard modifiers.
     * @param sector_intersection Entity intersection info.
     * @return True if the event was processed (the event was "eaten"),
     *     false otherwise.
     * @note Called by #TerrainWindow.
     * @return True if the event was processed, false otherwise.
     */
    bool handle_entity_click
        (
        SDLMod mod,
        const EntityManager::RayIntersection & entity_intersection
        );

private:

    /**
     * Send modify elevation command to account object.
     * If shift is pressed the operation has inverse effect.
     */
    void send_modify_elevation_command
        (
        SDLMod mod,
        const SectorManager::RayIntersection & sector_intersection
        );

    /** Send modify material command to account object. */
    void send_modify_material_command
        (
        const SectorManager::RayIntersection & sector_intersection
        );

    /** Send add entity command to account object. */
    void send_add_entity_command
        (
        const SectorManager::RayIntersection & sector_intersection
        );

private:

    Mode    mode;
        /**< Current editor mode. */
        
    std::auto_ptr< EditorModePanel >        editor_mode_panel;
        /**< Editor mode panel. */

    std::auto_ptr< ModifyElevationPanel >   modify_elevation_panel;
        /**< Modify elevation panel. */
    
    std::auto_ptr< EditorMaterialPanel >    editor_material_panel;
        /**< Material panel. */
    
    std::auto_ptr< EditorAddEntityPanel >   editor_add_entity_panel;
        /**< Add entity panel. */

    std::auto_ptr< EditorEditEntityPanel >  editor_edit_entity_panel;
        /**< Edit entity panel. */

    std::auto_ptr< EditorDeleteEntityPanel >  editor_delete_entity_panel;
        /**< Delete entity panel. */

    bool    visible;
        /**< Is editor visible? */
    
    }; // Editor


} // namespace Demo::Client
} // namespace Demo

#endif // CLIENT_EDITOR_H
