/** @file game_interface.cpp
 * Game interface (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "gui/game_interface.h"
#include "gui/widget_input_dispatcher.h"
#include "global/global.h"
#include "renderer/passes.h"
#include "renderer/layers.h"
#include "renderer/overlay_animation.h"
#include "input/input_layers.h"
#include "client_global.h"
#include "media_repository.h"
#include "lib/shared/client_item.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/account_object_rpc.h"
#include "replication/replica_manager.h"
#include "game/game_context.h"
#include "game/sector_manager.h"
#include "game/player_manager.h"
#include "input/input_manager.h"
#include "renderer/camera.h"
#include <sstream>


namespace Demo {
namespace Client {

// anonymous namespace
namespace {


/**
 * Send player move action to its account object.
 */
void do_action
    (
    const Demo::Lib::PlayerMoveAction & action
    )
    {
    // Send the action.
    Global::get_game_context().require_account()->async_handle_move_action( action );
    }


/**
 * Send player item action to its account object.
 */
void do_action
    (
    const Demo::Lib::PlayerItemAction & action
    )
    {
    // Send the action.
    Global::get_game_context().require_account()->async_handle_item_action( action );
    }


} // anonymous namespace


// GameInterface::CursorContext
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


void GameInterface::CursorContext::set_panel_slot_context
    (
    PanelId                 the_panel_id,
    int                     the_slot_index,
    const std::string &     icon_name,
    WeakPointer< Object >   the_slot_object
    )
    {
    context = PANEL_SLOT_CONTEXT;
    panel_id = the_panel_id;
    slot_index = the_slot_index;
    slot_object = the_slot_object;

    change_icon( icon_name );
    }


void GameInterface::CursorContext::change_icon
    (
    const std::string & the_icon_name
    )
    {
    if( ( icon_name != the_icon_name ) || !icon.get() )
        {
        MediaRepository & media = Global::get_media_repository();

        icon_name = the_icon_name;

        icon.reset( media.create_icon( icon_name ) );
        icon->add_animation_object( new Renderer::OverlayMousePositionAnimation );
        icon->add_animation_object( new Renderer::OverlayOffsetDeltaAnimation( SVector2( 8.0f, 8.0f ) ) );
        icon->set_base_size( SVector2( 32.0f, 32.0f ) );
        icon->register_to_global_manager( Renderer::PASS_TOPMOST, Renderer::LAYER_TOPMOST );
        }
    }


// GameInterface
//~~~~~~~~~~~~~~


GameInterface::GameInterface() :
    visible( true )
    {
    // We can't create icons now, because media repository
    // is not initialized yet.
    }


GameInterface::~GameInterface()
    {
    // Cancel monitoring player replica.
    unregister_player_replica_callback();
    }


void GameInterface::initialize()
    {
    Widget * parent_widget = &Global::get_widget_input_dispatcher();
    hope( parent_widget );

    inventory_button.reset
        (
        new IconButton
            (
            IconButton::TYPE_SWITCH,
            "gui/inventory_button_up",
            "gui/inventory_button_down",
            "gui/inventory_button_up_mouse_on",
            "gui/inventory_button_down_mouse_on"
            )
        );

    ability_button.reset
        (
        new IconButton
            (
            IconButton::TYPE_SWITCH,
            "gui/ability_button_up",
            "gui/ability_button_down",
            "gui/ability_button_up_mouse_on",
            "gui/ability_button_down_mouse_on"
            )
        );

    player_portrait.reset
        (
        new IconButton
            (
            IconButton::TYPE_ONE_SHOT,
            "image/portrait1",
            "image/portrait1",
            "image/portrait1",
            "image/portrait1"
            )
        );

    player_ability_info.reset( new Renderer::TextRectangle() );

    inventory_panel.reset
        (
        new SlotPanel( "Inventory" )
        );

    ability_panel.reset
        (
        new SlotPanel( "Abilities" )
        );

    // Panel buttons.

    hope( inventory_button.get() );
    inventory_button->set_origin( Renderer::OverlayObject::TOP_RIGHT );
    inventory_button->set_base_offset( SVector2( 0.0f, 0.0f ) );
    inventory_button->set_base_size( SVector2( 32.0f, 32.0f ) );
    inventory_button->set_parent_widget( parent_widget );
    inventory_button->set_message_receiver( this );

    hope( ability_button.get() );
    ability_button->set_origin( Renderer::OverlayObject::TOP_RIGHT );
    ability_button->set_base_offset( SVector2( 30.0f, 0.0f ) );
    ability_button->set_base_size( SVector2( 32.0f, 32.0f ) );
    ability_button->set_parent_widget( parent_widget );
    ability_button->set_message_receiver( this );

    hope( player_portrait.get() );
    player_portrait->set_origin( Renderer::OverlayObject::TOP_LEFT );
    player_portrait->set_base_offset( SVector2( 0.0f, 0.0f ) );
    player_portrait->set_base_size( SVector2( 64.0f, 64.0f ) );
    player_portrait->set_parent_widget( parent_widget );
    player_portrait->set_message_receiver( this );

    player_ability_info->set_font( &Global::get_console_font() );
    player_ability_info->set_parent_overlay_object( parent_widget );
    player_ability_info->set_base_offset( SVector2( 0.0f, 64.0f ) );

    inventory_panel->set_origin( Renderer::OverlayObject::TOP_RIGHT );
    inventory_panel->set_base_offset( SVector2( 0.0f, 32.0f ) );
    inventory_panel->set_base_size( SVector2( 250.0f, 200.0f ) );
    inventory_panel->set_parent_widget( parent_widget );
    inventory_panel->set_message_receiver( this );
    inventory_panel->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( SColor( 2.0f, 2.0f, 2.0f, 2.0f ) ) );

    ability_panel->set_origin( Renderer::OverlayObject::TOP_RIGHT );
    ability_panel->set_base_offset( SVector2( 0.0f, 32.0f ) );
    ability_panel->set_base_size( SVector2( 180.0f, 170.0f ) );
    ability_panel->set_parent_widget( parent_widget );
    ability_panel->set_message_receiver( this );
    ability_panel->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( SColor( 2.0f, 2.0f, 2.0f, 2.0f ) ) );

    update_panel_animations();

    // Hide self.
    set_visibility( false );
    }


void GameInterface::player_selected
    (
    WeakPointer< Lib::ClientEntity > player
    )
    {
    // Hide self.
    set_visibility( false );

    // Clear all player related stuff.
    clear_mirrored_data();

    // Cancel monitoring of previous player.
    unregister_player_replica_callback();

    // Link to new player.
    if( player != null )
        {
        ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();

        // Show self.
        set_visibility( true );

        // Request replication of the player.
        Global::get_game_context().require_account()->async_request_player_replication( player );

        // Start monitoring replicas of the player.
        replica_manager.register_callback
            (
            player.get_object_id(),
            &player_replica_callback,
            0,
            ::Massiv::Core::VariantPointer( this )
            );
        }
    }


void GameInterface::clear_mirrored_data()
    {
    cursor_context.clear();
    inventory_mirror.clear();
    inventory_panel->change_slot_table_size( 0, 0 );
    ability_panel->change_slot_table_size( 0, 0 );
    }


void GameInterface::update_player_mirror
    (
    WeakPointer< Lib::ClientEntity > replica
    )
    {
    if( !replica.is_local() )
        {
        clear_mirrored_data();
        return;
        }

    bool update_inventory = true; // Update inventory mirror and inventory panel?
    bool update_abilities = true; // Update ability panel?

    if( !replica->is_consistent() )
        {
        // Leave previous state.
        update_inventory = false;
        }

    if( update_inventory )
        {
        inventory_mirror.update( replica->inventory );

        // Resize inventory panel.
        if( inventory_mirror.slot_columns != inventory_panel->get_slot_columns() ||
            inventory_mirror.slot_rows != inventory_panel->get_slot_rows() )
            {
            inventory_panel->change_slot_table_size
                (
                inventory_mirror.slot_columns,
                inventory_mirror.slot_rows
                );
            }

        // Update icons in inventory panel.
            {
            MediaRepository & media = Global::get_media_repository();

            unsigned int i;
            for( i = 0; i < inventory_mirror.slots.size(); ++i )
                {
                const ItemMirror & item_mirror = inventory_mirror.slots[ i ];
                SlotPanel::Slot & slot = inventory_panel->get_slot( i );

                if( !item_mirror.is_empty() )
                    {
                    std::string icon_name = item_mirror.get_icon_name();
                    if( slot.icon_name != icon_name )
                        {
                        slot.icon.reset( media.create_icon( icon_name ) );
                        slot.icon_name = icon_name;
                        }

                    slot.set_title( item_mirror.get_description() );

                    if( item_mirror.equipped )
                        {
                        slot.highlight = true;
                        }
                    else
                        {
                        slot.highlight = false;
                        }
                    }
                else
                    {
                    // Clear the slot.
                    slot.clear();
                    }
                }
            }
        }

    // Resize ability panel.
        {
        ability_panel->change_slot_table_size( 3, 2 );
        ability_panel->set_slot_padding( 16.0f, 16.0f );
        }

    // Update icons in ability panel.
    if( update_abilities )
        {
        MediaRepository & media = Global::get_media_repository();

        int i;
        for( i = 0; i < Lib::ABILITY_COUNT; ++i )
            {
            hope( ability_panel->is_valid_slot_index( i ) );
            Lib::AbilityId ability = static_cast< Lib::AbilityId >( i );
            SlotPanel::Slot &            slot = ability_panel->get_slot( i );

            if( replica->supports_ability( ability ) )
                {
                std::string icon_name = "item/icon_unknown";
                std::string ability_text;

                    {
                    std::ostringstream ss;
                    ss  << replica->get_ability_value( ability ) << "/"
                        << replica->get_ability_upper_limit( ability );
                    ability_text = ss.str();
                    }

                switch( ability )
                    {
                    case Lib::ABILITY_HEALTH:
                        icon_name = "item/icon_red_potion";
                        break;

                    case Lib::ABILITY_SPEED:
                        icon_name = "item/icon_boots";
                        break;

                    default:
                        break;
                    }

                // Set slot title.
                slot.set_title( Lib::ClientEntity::ability_to_string( ability ) );

                // Set slot icon.
                if( slot.icon_name != icon_name )
                    {
                    slot.icon.reset( media.create_icon( icon_name ) );
                    slot.icon_name = icon_name;
                    }

                // Set slot text.
                if( slot.get_text() != ability_text )
                    {
                    slot.set_text( ability_text );
                    }
                }
            else
                {
                // Clear the slot.
                slot.clear();
                }
            }
        }

    // Updare cursor context.
    switch( cursor_context.context )
        {
        case CursorContext::EMPTY_CONTEXT:
            break;

        case CursorContext::PANEL_SLOT_CONTEXT:
            {
            switch( cursor_context.panel_id )
                {
                case PANEL_INVENTORY:
                    {
                    if( !inventory_mirror.is_valid_slot_index( cursor_context.slot_index ) )
                        {
                        cursor_context.clear();
                        }
                    else
                        {
                        const ItemMirror & item_mirror = inventory_mirror.slots[ cursor_context.slot_index ];
                        WeakPointer< Object > item_ptr = item_mirror.item_replica;
                        if( item_ptr != cursor_context.slot_object )
                            {
                            cursor_context.clear();
                            }
                        else
                            {
                            // Update item icon.
                            hope( !item_mirror.is_empty() );
                            cursor_context.change_icon( item_mirror.get_icon_name() );
                            }
                        }
                    break;
                    }

                case PANEL_ABILITIES:
                    break;
                }
            break;
            }
        }

    // Update player_ability_info.
        {
        std::string health;
        std::string speed;
        std::string::size_type length;

            {
            std::ostringstream ss;
            ss  << "health " << replica->get_ability_value( Lib::ABILITY_HEALTH )
                << "/" << replica->get_ability_upper_limit( Lib::ABILITY_HEALTH );
            health = ss.str();
            }

            {
            std::ostringstream ss;
            ss  << "speed  " << replica->get_ability_value( Lib::ABILITY_SPEED )
                << "/" << replica->get_ability_upper_limit( Lib::ABILITY_SPEED );
            speed = ss.str();
            }

        length = health.size();
        if( length < speed.size() ) length = speed.size();
        hope( length > 0 );

        player_ability_info->clear();
        player_ability_info->set_columns_rows( length, 2 );
        player_ability_info->set_size_from_font();
        player_ability_info->print( 0, 0, health );
        player_ability_info->print( 0, 1, speed );
        }
    }


void GameInterface::set_visibility
    (
    bool the_visible
    )
    {
    visible = the_visible;
    update_visibility();
    }


bool GameInterface::is_visible() const
    {
    return visible;
    }


void GameInterface::toggle_panel
    (
    PanelId panel_id
    )
    {
    bool is_button_down;
    if( panel_id == PANEL_INVENTORY )
        {
        is_button_down = inventory_button->is_down();
        }
    else
        {
        hope( panel_id == PANEL_ABILITIES );
        is_button_down = ability_button->is_down();
        }

    show_hide_panel( panel_id, !is_button_down );
    }


void GameInterface::button_clicked
    (
    Button * button
    )
    {
    if( button == inventory_button.get() )
        {
        bool is_down = inventory_button->is_down();
        show_hide_panel( PANEL_INVENTORY, is_down );
        }
    else if( button == ability_button.get() )
        {
        bool is_down = ability_button->is_down();
        show_hide_panel( PANEL_ABILITIES, is_down );
        }
    else if( button == player_portrait.get() )
        {
        handle_player_portrait_clicked();
        }
    }


void GameInterface::slot_panel_close
    (
    SlotPanel * panel
    )
    {
    PanelId panel_id = panel_to_id( panel );
    show_hide_panel( panel_id, false );
    }


void GameInterface::slot_panel_slot_clicked
    (
    SlotPanel * panel,
    int         slot_index
    )
    {
    PanelId panel_id = panel_to_id( panel );

    switch( panel_id )
        {
        case PANEL_INVENTORY:
            handle_inventory_slot_clicked( slot_index );
            break;

        case PANEL_ABILITIES:
            handle_ability_slot_clicked( slot_index );
            break;
        }
    }


bool GameInterface::handle_sector_click
    (
    SDLMod mod,
    const SectorManager::RayIntersection & sector_intersection
    )
    {
    // Clicked with an empty cursor or with an item?

    if( cursor_context.context == CursorContext::PANEL_SLOT_CONTEXT )
        {
        // Clicked with an ability or an item.

        if( cursor_context.panel_id == PANEL_INVENTORY )
            {
            // Clicked with an item.

            do_action_inventory_item_to_sector( cursor_context.slot_index, sector_intersection );
            cursor_context.clear();
            }
        else
            {
            // Clicked with an ability. Clear cursor contex.

            cursor_context.clear();
            }
        }
    else
        {
        // Send player action to server.

        Demo::Lib::PlayerMoveAction action;
        action.target_position = SVector2( sector_intersection.intersection.x(), sector_intersection.intersection.y() );
        do_action( action );
        }

    return true;
    }


bool GameInterface::handle_entity_click
    (
    SDLMod mod,
    const EntityManager::RayIntersection & entity_intersection
    )
    {
    // Clicked with an empty cursor or with an item?

    if( cursor_context.context == CursorContext::PANEL_SLOT_CONTEXT )
        {
        // Clicked with an ability or an item.

        if( cursor_context.panel_id == PANEL_INVENTORY )
            {
            // Clicked with an item.

            do_action_inventory_item_to_entity( cursor_context.slot_index, entity_intersection );
            cursor_context.clear();
            }
        else
            {
            // Clicked with an ability. Do nothing.

            cursor_context.clear();
            }
        }
    else
        {
        // Clicked with empty cursor. Walk towards the entity.

        const SVector3 player_position_3d = ::Demo::Client::Global::get_player_manager().get_player_position();
        const SVector2 player_position = SVector2( player_position_3d.x(), player_position_3d.y() );
        const SVector2 entity_position = entity_intersection.entity->current_position.origin.read();
        const Float32 ENTITY_RADIUS = 0.5f; // estimated entity radius.
        SVector2 entity_to_player = player_position - entity_position;
        if( !entity_to_player.epsilon_zero() )
            {
            if( entity_to_player.magnitude() > ENTITY_RADIUS )
                {
                entity_to_player.normalize();
                entity_to_player *= ENTITY_RADIUS;

                Demo::Lib::PlayerMoveAction action;
                action.target_position = entity_position + entity_to_player;
                do_action( action );
                }
            }
        }
    return false;
    }


void GameInterface::handle_inventory_slot_clicked
    (
    int slot_index
    )
    {
    bool empty_cursor_context = false;
    ItemMirror * item_mirror = NULL; // NULL if the slot is empty.

    // Get slot content.
    hope( inventory_mirror.slot_columns == inventory_panel->get_slot_columns() );
    hope( inventory_mirror.slot_rows == inventory_panel->get_slot_rows() );

        {
        ItemMirror & item = inventory_mirror.slots[ slot_index ];
        if( !item.is_empty() )
            {
            item_mirror = &item;
            }
        }

    // Analyze cursor context.
    switch( cursor_context.context )
        {
        case CursorContext::EMPTY_CONTEXT:
            empty_cursor_context = true;
            break;

        case CursorContext::PANEL_SLOT_CONTEXT:
            break;
        }

    // Do action.
    if( empty_cursor_context )
        {
        if( !item_mirror )
            {
            // Empty slot. Nothing to do.
            }
        else
            {
            // Move the slot to cursor context.
            cursor_context.set_panel_slot_context
                (
                PANEL_INVENTORY,
                slot_index,
                item_mirror->get_icon_name(),
                item_mirror->item_replica
                );
            }
        }
    else
        {
        if( cursor_context.context == CursorContext::PANEL_SLOT_CONTEXT )
            {
            if( cursor_context.panel_id == PANEL_INVENTORY )
                {
                // Clicked with selected item from inventory panel to anothet item slot.
                do_action_inventory_item_to_inventory
                    (
                    cursor_context.slot_index, // source item
                    slot_index                 // destination slot
                    );

                cursor_context.clear();
                }
            }
        }
    }


void GameInterface::handle_ability_slot_clicked
    (
    int slot_clicked
    )
    {
    cursor_context.clear();
    }


void GameInterface::handle_player_portrait_clicked()
    {
    if( cursor_context.context == CursorContext::PANEL_SLOT_CONTEXT )
        {
        switch( cursor_context.panel_id )
            {
            case PANEL_INVENTORY:
                do_action_inventory_item_to_player( cursor_context.slot_index );
                cursor_context.clear();
                break;

            case PANEL_ABILITIES:
                // Nothing to do.
                break;
            }
        }
    }


void GameInterface::do_action_inventory_item_to_inventory
    (
    int source_slot_index,
    int target_slot_index
    )
    {
    Demo::Lib::PlayerItemAction action;
    action.source = Demo::Lib::PlayerItemAction::SOURCE_INVENTORY;
    action.source_slot_index = source_slot_index;
    action.source_object = inventory_mirror.slots[ source_slot_index ].item_replica;
    action.target = Demo::Lib::PlayerItemAction::TARGET_INVENTORY;
    action.target_slot_index = target_slot_index;
    action.target_object = inventory_mirror.slots[ target_slot_index ].item_replica;
    do_action( action );
    }


void GameInterface::do_action_inventory_item_to_player
    (
    int slot_index
    )
    {
    Demo::Lib::PlayerItemAction action;
    action.source = Demo::Lib::PlayerItemAction::SOURCE_INVENTORY;
    action.source_slot_index = slot_index;
    action.source_object = inventory_mirror.slots[ slot_index ].item_replica;
    action.target = Demo::Lib::PlayerItemAction::TARGET_ENTITY;
    action.target_object = ::Demo::Client::Global::get_player_manager().get_player()->entity;
    do_action( action );
    }


void GameInterface::do_action_inventory_item_to_sector
    (
    int                                     slot_index,
    const SectorManager::RayIntersection &  sector_intersection
    )
    {
    Demo::Lib::PlayerItemAction action;
    action.source = Demo::Lib::PlayerItemAction::SOURCE_INVENTORY;
    action.source_slot_index = slot_index;
    action.source_object = inventory_mirror.slots[ slot_index ].item_replica;
    action.target = Demo::Lib::PlayerItemAction::TARGET_WORLD;
    action.target_position = SVector2
        (
        sector_intersection.intersection.x(),
        sector_intersection.intersection.y()
        );
    do_action( action );
    }


void GameInterface::do_action_inventory_item_to_entity
    (
    int                                     slot_index,
    const EntityManager::RayIntersection &  entity_intersection
    )
    {
    Demo::Lib::PlayerItemAction action;
    action.source = Demo::Lib::PlayerItemAction::SOURCE_INVENTORY;
    action.source_slot_index = slot_index;
    action.source_object = inventory_mirror.slots[ slot_index ].item_replica;
    action.target = Demo::Lib::PlayerItemAction::TARGET_ENTITY;
    action.target_object = entity_intersection.entity->entity;
    do_action( action );
    }


void GameInterface::update_panel_animations()
    {
    bool inv_butt_down = inventory_button->is_down();
    bool abil_butt_down = ability_button->is_down();

        {
        const SVector2 size = inventory_panel->get_current_size();
        const SColor target_color_factor( 1.0f, 1.0f, 1.0f, inv_butt_down ? 0.7f : 0.0f );
        inventory_panel->require_color_factor_animation().set_target_factor( target_color_factor );
        }

        {
        const SVector2 size = ability_panel->get_current_size();
        const SColor target_color_factor( 1.0f, 1.0f, 1.0f, abil_butt_down ? 0.7f : 0.0f );
        ability_panel->require_color_factor_animation().set_target_factor( target_color_factor );
        }
    }


void GameInterface::show_hide_panel
    (
    PanelId panel_id,
    bool    show
    )
    {
    switch( panel_id )
        {
        case PANEL_INVENTORY:
            {
            inventory_button->push( show, false );
            if( show )
                {
                ability_button->push( !show, false );
                }
            break;
            }

        case PANEL_ABILITIES:
            {
            ability_button->push( show, false );
            if( show )
                {
                inventory_button->push( !show, false );
                }
            break;
            }
        }

    update_panel_animations();
    }


GameInterface::PanelId GameInterface::panel_to_id
    (
    const SlotPanel * panel
    ) const
    {
    hope( panel );
    if( panel == inventory_panel.get() )
        {
        return PANEL_INVENTORY;
        }
    else if( panel == ability_panel.get() )
        {
        return PANEL_ABILITIES;
        }
    else
        {
        panic( "Invalid panel." );
        }
    }


void GameInterface::player_replica_callback
    (
    ::Massiv::Core::ReplicaManager::UpdateReason            reason,
    ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
    int                                                     user_tag,
    ::Massiv::Core::VariantPointer                          user_data
    )
    {
    GameInterface * const self = ::Massiv::Core::variant_cast< GameInterface * >( user_data );

    switch( reason )
        {
        case ::Massiv::Core::ReplicaManager::UPDATE_RECEIVED:
            {
            panic( "Unexpected reason." );
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_UPDATED:
            {
            WeakPointer< Lib::ClientEntity > player = replica.convert();
            self->update_player_mirror( player );
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_GROUP_UPDATED:
            {
            // Maybe inventory has changed. Force update.
            WeakPointer< Lib::ClientEntity > player = replica.convert();
            self->update_player_mirror( player );
            break;
            }

        case ::Massiv::Core::ReplicaManager::REPLICA_DESTROY:
            {
            self->clear_mirrored_data();
            break;
            }

        default:
            {
            break;
            }
        }
    }


void GameInterface::unregister_player_replica_callback()
    {
    ::Massiv::Core::ReplicaManager & replica_manager = ::Massiv::Core::Global::replica_manager();
    replica_manager.unregister_callback( &player_replica_callback );
    }


void GameInterface::update_visibility()
    {
    inventory_button->set_visibility( visible );
    ability_button->set_visibility( visible );
    player_portrait->set_visibility( visible );
    player_ability_info->set_visibility( visible );
    inventory_panel->set_visibility( visible );
    ability_panel->set_visibility( visible );
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "gui/game_interface_inline.h"
#endif
