/** @file game_interface.h
 * Game interface.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_GAME_INTERFACE_H
#define CLIENT_GAME_INTERFACE_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_ICON_BUTTON_H
#include "icon_button.h"
#endif

#ifndef CLIENT_SLOT_PANEL_H
#include "gui/slot_panel.h"
#endif

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#ifndef CLIENT_ITEM_SLOT_CONTAINER_MIRROR_H
#include "game/item_slot_container_mirror.h"
#endif

#ifndef MASSIV_REPLICA_MANAGER_H
#include "replication/replica_manager.h"
#endif

#ifndef CLIENT_ENTITY_MANAGER_H
#include "game/entity_manager.h"
#endif

#ifndef CLIENT_SECTOR_MANAGER_H
#include "game/sector_manager.h"
#endif


namespace Demo {
namespace Client {

INJECT_CORE_SYMBOLS();


/**
 * Game interface.
 */
class GameInterface :
    public Button::ButtonMessageReceiver,
    public SlotPanel::SlotPanelMessageReceiver
    {
    NO_COPY_CONSTRUCTOR( GameInterface );
    NO_ASSIGNMENT_OPERATOR( GameInterface );

public:

    /** Identifies panel on the screen. */
    enum PanelId
        {
        PANEL_INVENTORY,
        PANEL_ABILITIES
        }; // PanelId

    /**
     * CursorContext holds information about user actions with a mouse.
     * For example if the user clicks some item in inventory,
     * CursorContext stores information that the item was selected.
     * Mouse click handler which handles next mouse click can examine the
     * CursorContext and determine what the user want to do.
     */
    class CursorContext
        {
        NO_COPY_CONSTRUCTOR( CursorContext );
        NO_ASSIGNMENT_OPERATOR( CursorContext );

    public:
        /**
         * Context type.
         */
        enum Context
            {
            EMPTY_CONTEXT,
                /**< CursorContext doesn't contain any information. */

            PANEL_SLOT_CONTEXT
                /**< CursorContext identifies a slot of some game inerface panel.
                 * Valid attributes are @a panel_id, @a slot_index, @a slot_object.
                 */
            };

    public:

        /**
         * Default constructor. @a context is set to EMPTY_CONTEXT. Other
         * values stay uninitialized.
         */
        INLINE CursorContext();

        /**
         * Clear context content. @a context is set to EMPTY_CONTEXT,
         * @a icon is destroyed. Other values stay unchanged.
         */
        INLINE void clear();

        /**
         * Set context content.
         */
        void set_panel_slot_context
            (
            PanelId                 panel_id,
            int                     slot_index,
            const std::string &     icon_name,
            WeakPointer< Object >   slot_item
            );

    public:

        /**
         * Change icon drawn beside cursor.
         */
        void change_icon
            (
            const std::string & icon_name
            );

    public:
        Context                 context;
            /**< Context type. */

        PanelId                 panel_id;
            /**< Identifies game interface panel. */

        int                     slot_index;
            /**< Identifies slot index in the panel indentified by @a panel_id. */

        WeakPointer< Object >   slot_object;
            /**< Object stored in the slot. Used to detect invalid cursor context
             * when corresponding panel is updated by a replica.
             */

    public:
        std::auto_ptr< Renderer::Icon > icon;
            /**< Icon that is drawn beside cursor. */

        std::string                     icon_name;
            /**< Icon template name. */

        }; // CursorContext

public:

    /**
     * Constructor.
     */
    GameInterface();

    /**
     * Desctructor.
     */
    ~GameInterface();

    /**
     * Initialize.
     */
    void initialize();

    /**
     * Called when @a player is selected by client.
     * Prepares for monitoring replicas of the player and requests
     * replication of the player.
     */
    void player_selected
        (
        WeakPointer< Lib::ClientEntity > player
        );

    /**
     * Clear all mirrored data from simulation and related stuff.
     */
    void clear_mirrored_data();

public:

    /** Update player state from its @a replica. */
    void update_player_mirror
        (
        WeakPointer< Lib::ClientEntity > replica
        );

public:

    /** Show/hide. */
    void set_visibility
        (
        bool visible
        );

    /** Is visible? */
    bool is_visible() const;

    /** Toggle visibility of panel @a panel_id and hide other panels. */
    void toggle_panel
        (
        PanelId panel_id
        );

public:

    // IconButton::MessageReceiver
    virtual void button_clicked
        (
        Button * button
        );

    // SlotPanel::SlotPanelMessageReceiver
    virtual void slot_panel_close
        (
        SlotPanel * panel
        );

    // SlotPanel::SlotPanelMessageReceiver
    virtual void slot_panel_slot_clicked
        (
        SlotPanel * panel,
        int         slot_index
        );

public:

    /**
     * Handle mouse click to a sector.
     * @param mod Current state of keyboard modifiers.
     * @param sector_intersection Sector intersection info.
     * @return True if the event was processed (the event was "eaten"),
     *     false otherwise.
     * @note Called by #TerrainWindow.
     */
    bool handle_sector_click
        (
        SDLMod mod,
        const SectorManager::RayIntersection & sector_intersection
        );

    /**
     * Handle mouse click to a entity.
     * @param mod Current state of keyboard modifiers.
     * @param sector_intersection Entity intersection info.
     * @return True if the event was processed (the event was "eaten"),
     *     false otherwise.
     * @note Called by #TerrainWindow.
     * @return True if the event was processed, false otherwise.
     */
    bool handle_entity_click
        (
        SDLMod mod,
        const EntityManager::RayIntersection & entity_intersection
        );

private:

    /** Handle mouse click to inventory panel. */
    void handle_inventory_slot_clicked
        (
        int slot_index
        );

    /** Handle mouse click to ability panel. */
    void handle_ability_slot_clicked
        (
        int slot_clicked
        );

    /** Handle mouse click to player's portrait. */
    void handle_player_portrait_clicked();

private:

    /** Send player item action manipulating with two slots from inventory panel. */
    void do_action_inventory_item_to_inventory
        (
        int source_slot_index,
        int target_slot_index
        );

    /** Send player item action - using inventory item on player. */
    void do_action_inventory_item_to_player
        (
        int slot_index
        );

    void do_action_inventory_item_to_sector
        (
        int                                     slot_index,
        const SectorManager::RayIntersection &  sector_intersection
        );

    void do_action_inventory_item_to_entity
        (
        int                                     slot_index,
        const EntityManager::RayIntersection &  entity_intersection
        );

private:

    /** Update animations for all panels. */
    void update_panel_animations();

    /** Show or hide interface panels. */
    void show_hide_panel
        (
        PanelId panel_id,
        bool    show
        );

    /** Convert panel pointer to corresponding #PanelId. */
    PanelId panel_to_id
        (
        const SlotPanel * panel
        ) const;

private:

    /**
     * Player replica update callback.
     */
    static void player_replica_callback
        (
        ::Massiv::Core::ReplicaManager::UpdateReason            reason,
        ::Massiv::Core::WeakPointer< ::Massiv::Core::Object >   replica,
        int                                                     user_tag,
        ::Massiv::Core::VariantPointer                          user_data
        );

    /**
     * Unregister player replica callback from replica manager.
     */
    void unregister_player_replica_callback();

private:

    /** Update visibility of game widgets. */
    void update_visibility();

private:
    CursorContext                   cursor_context;

    ItemSlotContainerMirror         inventory_mirror;

    std::auto_ptr< IconButton >     inventory_button;
    std::auto_ptr< IconButton >     ability_button;
    std::auto_ptr< IconButton >     player_portrait;
        /**< Portrait icon of the player. We use #IconButton class so
         * we are able to handle clicks to the portrait.
         */

    std::auto_ptr< Renderer::TextRectangle >    player_ability_info;
        /**< Text containing some of player's abilities, drawn
         * below player portrait.
         */

    std::auto_ptr< SlotPanel >      inventory_panel;
    std::auto_ptr< SlotPanel >      ability_panel;

    bool                            visible;
        /**< Is visible? */

    }; // GameInterface


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "gui/game_interface_inline.h"
#endif


#endif // CLIENT_GAME_INTERFACE_H
