/** @file list_box.cpp
 * Simple list box (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "gui/list_box.h"
#include "client_global.h"
#include "input/input_manager.h"
#include "renderer/font.h"


namespace Demo {
namespace Client {


ListBox::ListBox() :
    selected_item_index( static_cast< ItemIndex >( -1 ) ),
    message_receiver( NULL )
    {
    selected_item.reset( new Button( Button::TYPE_ONE_SHOT ) );
    selected_item->set_parent_widget( this );
    selected_item->set_message_receiver( this );
    
    prev_item_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    prev_item_button->set_text( "<" );
    prev_item_button->set_parent_widget( this );
    prev_item_button->set_message_receiver( this );
    
    next_item_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    next_item_button->set_text( ">" );
    next_item_button->set_parent_widget( this );
    next_item_button->set_message_receiver( this );
    
    select_item_popup_menu.reset( new PopupMenu() );
    select_item_popup_menu->initialize();
    select_item_popup_menu->set_message_receiver( this );
    
    update_buttons();
    }


ListBox::~ListBox()
    {
    }


void ListBox::do_layout()
    {
    Super::do_layout();

    SVector2    size = get_current_size();
    Float32     size_x = size.x();
    Float32     size_y = size.y();
    Float32     size_min = ( size_x < size_y ? size_x : size_y );
    
    selected_item->set_base_offset( SVector2( size_min, 0.0f ) );
    selected_item->set_base_size( SVector2( size_x - 2.0f * size_min, size_min ) );
    selected_item->do_layout();
    
    prev_item_button->set_base_offset( SVector2( 0.0f, 0.0f ) );
    prev_item_button->set_base_size( SVector2( size_min, size_min ) );
    prev_item_button->do_layout();
    
    next_item_button->set_base_offset( SVector2( size_x - size_min, 0.0f ) );
    next_item_button->set_base_size( SVector2( size_min, size_min ) );
    next_item_button->do_layout();
    
    // Make width of the popup menu same width as selected_item button.
        {
        Float32 new_width = selected_item->get_current_size().x();
        Float32 width = select_item_popup_menu->get_current_size().x();
        Float32 height = select_item_popup_menu->get_current_size().y();
        if( new_width > width )
            {
            select_item_popup_menu->set_base_size( SVector2( new_width, height ) );
            select_item_popup_menu->do_layout();
            }
        }
    }


SVector2 ListBox::get_ideal_size() const
    {
    const Float32 font_char_width = 8.0f; // FIXME
    const Float32 bsize = 20.0f; // Button size. Should be font char height + some padding.
    
    Float32 w = 0;
    Float32 h = bsize;
    
    w += bsize + bsize; // Prev and Next buttons.
    
    // Determine optimal size of selected_item.
        {
        Items::const_iterator it;
        Float32 max_width = 0.0f;
        for( it = items.begin(); it != items.end(); ++it )
            {
            const Item & item = *it;
            const Float32 name_width = static_cast< Float32 >( item.name.size() ) * font_char_width;
            max_width = ( max_width < name_width ? name_width : max_width );
            }
        // Add some left and right padding.
        max_width += 8.0f;
        w += max_width;
        }
    
    return SVector2( w, h );
    }


void ListBox::clear()
    {
    select_item_popup_menu->clear();
    items.clear();
    do_select_item( 0 );
    update_buttons();
    notify_of_item_selected();
    }


ListBox::ItemIndex ListBox::push
    (
    const Item & item
    )
    {
    ItemIndex result = items.size();
    items.push_back( item );
    if( result == 0 )
        {
        do_select_item( result );
        }
    update_buttons();
    // Register the item to popup menu.
        {
        int id = static_cast< int >( result );
        PopupMenu::Item popup_item( id, item.name );
        select_item_popup_menu->push( popup_item );
        }

    if( result == 0 )
        {
        notify_of_item_selected();
        }
    
    return result;
    }


bool ListBox::is_item_selected() const
    {
    return ( selected_item_index < items.size() );
    }


const ListBox::Item & ListBox::get_selected_item() const
    {
    hope( is_item_selected() );
    return items[ selected_item_index ];
    }


bool ListBox::select_item_by_id
    (
    int id
    )
    {
    ItemIndex index;
    Items::const_iterator it;
    for( it = items.begin(), index = 0; it != items.end(); ++it, ++index )
        {
        const Item & item = *it;
        if( item.id == id )
            {
            do_select_item( index );
            update_buttons();
            return true;
            }
        }
    return false;
    }


void ListBox::set_message_receiver
    (
    ListBoxMessageReceiver * the_message_receiver
    )
    {
    message_receiver = the_message_receiver;
    }


void ListBox::button_clicked
    (
    Button * button
    )
    {
    select_item_popup_menu->set_visibility( false );
    
    if( !is_item_selected() && !items.empty() )
        {
        do_select_item( 0 );
        update_buttons();
        notify_of_item_selected();
        }
    
    if( button == selected_item.get() )
        {
        if( items.size() > 1 )
            {
            SVector2    position = get_top_left_corner_on_screen();
            SVector2    size = get_current_size();
            Float32     size_x = size.x();
            Float32     size_y = size.y();
            Float32     size_min = ( size_x < size_y ? size_x : size_y );
            select_item_popup_menu->set_base_offset( position + SVector2( size_min, 0.0f /* size_min */ ) );
            select_item_popup_menu->set_visibility( true );
            }
        }
    else if( button == prev_item_button.get() )
        {
        if( selected_item_index < items.size() &&
            selected_item_index > 0 )
            {
            do_select_item( selected_item_index - 1 );
            update_buttons();
            notify_of_item_selected();
            }
        }
    else if( button == next_item_button.get() )
        {
        if( selected_item_index < items.size() &&
            ( selected_item_index + 1 ) < items.size() )
            {
            do_select_item( selected_item_index + 1 );
            update_buttons();
            notify_of_item_selected();
            }
        }
    }


void ListBox::popup_menu_item_clicked
    (
    PopupMenu * popup_menu,
    int item_id
    )
    {
    hope( popup_menu == select_item_popup_menu.get() );
    ItemIndex index = static_cast< ItemIndex >( item_id );
    if( index < items.size() )
        {
        do_select_item( index );
        update_buttons();
        notify_of_item_selected();
        }
    }


void ListBox::do_select_item
    (
    ItemIndex index
    )
    {
    if( index >= items.size() )
        {
        selected_item_index = static_cast< ItemIndex >( -1 );
        set_text( "" );
        }
    else
        {
        selected_item_index = index;
        set_text( items[ index ].name );
        }
    }


void ListBox::set_text
    (
    const std::string & text
    )
    {
    selected_item->set_text( text );
    }


void ListBox::update_buttons()
    {
    if( items.size() <= 1 )
        {
        next_item_button->disable();
        prev_item_button->disable();
        selected_item->disable();
        return;
        }

    selected_item->enable();
        
    if( !is_item_selected() )
        {
        // Should never happen. If nonempty, there is always
        // an item selected.
        next_item_button->enable();
        prev_item_button->enable();
        return;
        }
    
    if( selected_item_index == 0 )
        {
        next_item_button->enable();
        prev_item_button->disable();
        return;
        }
    
    if( selected_item_index == ( items.size() - 1 ) )
        {
        next_item_button->disable();
        prev_item_button->enable();
        return;
        }
    
    next_item_button->enable();
    prev_item_button->enable();
    }


void ListBox::notify_of_item_selected()
    {
    if( message_receiver )
        {
        message_receiver->list_box_item_selected( this );
        }
    }


} // namespace Demo::Client
} // namespace Demo
