/** @file list_box.h
 * Simple list box.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_LIST_BOX_H
#define CLIENT_LIST_BOX_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#ifndef CLIENT_POPUP_MENU_H
#include "gui/popup_menu.h"
#endif

#include <memory>
#include <string>


namespace Demo {
namespace Client {


/**
 * Simple list box.
 */
class ListBox :
    public Widget,
    public Button::ButtonMessageReceiver,
    public PopupMenu::PopupMenuMessageReceiver
    {
    typedef Widget  Super;
    
    NO_COPY_CONSTRUCTOR( ListBox );
    NO_ASSIGNMENT_OPERATOR( ListBox );
    
public:
    
    typedef std::size_t ItemIndex;

public:

    /** List box item. */
    struct Item
        {
        Item() :
            id( 0 )
            {
            }
        
        Item
            (
            int the_id,
            const std::string & the_name
            ) :
            id( the_id ),
            name( the_name )
            {
            }
            
        int             id;
            /**< User defined integer value. */
        std::string     name;
            /**< Item name. */
        }; // Item
    
    /**
     * Receiver of messages from the list box.
     */
    class ListBoxMessageReceiver
        {
    public:
    
        /** Destructor. */
        virtual ~ListBoxMessageReceiver()
            {
            }
        
        /**
         * Called when an item from the list box is selected.
         * Also called when selected item is removed from the list box.
         * Use @a is_item_selected() to detect whether new item is
         * selected or previously selected item was deselected.
         */
        virtual void list_box_item_selected
            (
            ListBox *   list_box
            )
            {
            }

        }; // ListBoxMessageReceiver

public:

    /** Constructor. */
    ListBox();
    
    /** Destructor. */
    ~ListBox();
    
public:

    // From superclass.
    virtual void do_layout();
    
    /** Get optimal size of the list box. */
    SVector2 get_ideal_size() const;

public:

    /** Clear list box content. All items are removed. */
    void clear();
    
    /**
     * Insert @a item at the end of the list box. 
     * @return Index of the inserted item.
     */
    ItemIndex push
        (
        const Item & item
        );
    
    /** Check whether an item is selected. */
    bool is_item_selected() const;
    
    /**
     * Get selected item.
     * @pre An item is selected.
     * @note Insering new item or clearing content of the list box
     *     invalidates returned reference.
     */
    const Item & get_selected_item() const;
    
    /**
     * Select item by its id.
     * @return false if there is no item with specified id, else true is returned.
     */
    bool select_item_by_id
        (
        int id
        );
    
    /**
     * Set #ListBoxMessageReceiver of the list box.
     * If @a message_receiver is @c NULL, message notifying will be
     * disabled.
     */
    void set_message_receiver
        (
        ListBoxMessageReceiver * message_receiver
        );
    
public:

    // From ButtonMessageReceiver.
    virtual void button_clicked
        (
        Button * button
        );
    
    // From PopupMenuMessageReceiver.
    virtual void popup_menu_item_clicked
        (
        PopupMenu * popup_menu,
        int item_id
        );

private:

    /**
     * Select item at @a index position.
     * If @a index is out of range, currently selected item is deselected.
     */
    void do_select_item
        (
        ItemIndex   index
        );

    /**
     * Print @a text to @a selected_item button.
     */
    void set_text
        (
        const std::string & text
        );
    
    /**
     * Update states of buttons.
     */
    void update_buttons();

    /**
     * Notify list box message receiver that item was selected or deselected.
     */
    void notify_of_item_selected();

private:

    typedef std::vector< Item > Items;
    
    Items                       items;
        /**< List of items. */
    ItemIndex                   selected_item_index;
        /**< Index of selected item. */
    std::auto_ptr< Button >     selected_item;
        /**< Button containing name of selected item. When
         * clicked, popup menu is shown.
         */
    std::auto_ptr< Button >     next_item_button;
        /**< Button used to select next item. */
    std::auto_ptr< Button >     prev_item_button;
        /**< Button used to select previous item. */
    std::auto_ptr< PopupMenu >  select_item_popup_menu;
        /**< Popup menu for selecting list box items. */
    
    ListBoxMessageReceiver *    message_receiver;
        /**< List box message receiver. */

    }; // ListBox


} // namespace Demo::Client
} // namespalice Demo


#endif // CLIENT_LIST_BOX_H
