/** @file login_dialog.h
 * Login dialog.
 */

/*
 * begin        : Wed Jul 9 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef CLIENT_LOGIN_DIALOG_H
#define CLIENT_LOGIN_DIALOG_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_WIDGET_INPUT_DISPATCHER_H
#include "gui/widget_input_dispatcher.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#ifndef CLIENT_PANEL_BORDER_H
#include "gui/panel_border.h"
#endif

#ifndef CLIENT_TEXT_RECTANGLE_H
#include "renderer/text_rectangle.h"
#endif

#ifndef CLIENT_EDIT_BOX_H
#include "gui/edit_box.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * LoginDialog. Used to ask user for login name and password.
 * Login name corresponds to local node config file, password is used
 * to access content of the config file. The password is not the password
 * used to connect to the massiv simulation. To connect to the simulation
 * system uses rsa private key stored in the local node config file.
 *
 * @note Login name should be name of config file relative to "login" directory.
 */
class LoginDialog :
    public PanelBorder,
    public Button::ButtonMessageReceiver,
    public EditBox::EditBoxMessageReceiver
    {
    NO_COPY_CONSTRUCTOR( LoginDialog );
    NO_ASSIGNMENT_OPERATOR( LoginDialog );

    /// Superclass.
    typedef PanelBorder  Super;
    
public:

    /** Login dialog result. */
    enum Result
        {
        RESULT_OK,
            /**< User has specified login name and password. */
            
        RESULT_CANCEL
            /**< The dialog was cancelled. */

        }; // Result
        
    /** Login dialog mode. */
    enum Mode
        {
        MODE_NONE,
            /**< Not initialized yet. */
        MODE_ASK_LOGIN,
            /**< Asking login name and password. */
        MODE_ASK_NEW_ACCOUNT,
            /**< Asking information to create new account. */
        MODE_WAITING_ACCOUNT_CREATION
            /**< Waiting for creation of new account. */
        }; // Mode

public:

    /** Constructor. */
    LoginDialog();
    
    /**
     * Initialize the dialog.
     * @param login_name Text to show in login name edit box.
     * @param login_password Text to show in password edit box.
     * @param error_text Error text to draw. Leave empty for default.
     */
    void initialize
        (
        const std::string & login_name,
        const std::string & login_password,
        const std::string & error_text = ""
        );

    /** Dialog's next tick. */
    void next_tick();
    
    /** Get dialog result. */
    Result get_result() const;

    /**
     * Get login name.
     * @pre Result of the dialog is @c RESULT_OK.
     */
    std::string get_login_name() const;
    
    /**
     * Get login password.
     * @pre Result of the dialog is @c RESULT_OK.
     */
    std::string get_login_password() const;

public:

    // From IconButton::MessageReceiver.
    virtual void button_clicked
        (
        Button * button
        );

    // From InputObject.
    virtual bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );

private:

    /**
     * Try commit login name and password from edit boxes
     * to @a login_name and @a login_password.
     */
    bool commit_input_data();
    
    /** Try send account creation request to account manager. */
    void request_account_creation();
    
    /** Set @c MODE_ASK_LOGIN mode. */
    void set_ask_login_mode
        (
        const std::string & the_login_name,
        const std::string & the_login_password
        );

    /** Set @c MODE_ASK_NEW_ACCOUNT mode. */
    void set_ask_new_account_mode();
    
    /** Set @c MODE_WAITING_ACCOUNT_CREATION mode. */
    void set_waiting_account_creation_mode();
    
    /** Set error text shown at bottom of the dialog. */
    void set_error_text
        (
        const std::string & text
        );

    /** Set info text. */
    void set_info_text
        (
        const std::string & the_info_text
        );

    /** Set new title. */
    void set_title
        (
        const std::string & the_title
        );

private:

    Mode                        mode;
        /**< Login dialog mode. */

    Result                      result;
        /**< Dialog result. */
    
    std::string                 login_name;
        /**< Login name. */
    
    std::string                 login_password;
        /**< Login password. */
        
    //
    // Gui items.
    //

    std::auto_ptr< Renderer::Icon >             picture;
        /**< Login picture. */
        
    std::auto_ptr< Renderer::TextRectangle >    title;
        /**< Title. */
    
    std::auto_ptr< Renderer::TextRectangle >    login_name_label;
        /**< Login name label. */
    
    std::auto_ptr< EditBox >                    login_name_edit_box;
        /**< Login name edit box. */
    
    std::auto_ptr< Renderer::TextRectangle >    login_password_label;
        /**< Login password label. */
    
    std::auto_ptr< EditBox >                    login_password_edit_box;
        /**< Login password edit box. */
        
    std::auto_ptr< Renderer::TextRectangle >    login_password_again_label;
        /**< Login password label. */
    
    std::auto_ptr< EditBox >                    login_password_again_edit_box;
        /**< Second login password edit box for verification of password
         * when creating new account.
         */
    
    std::auto_ptr< Button >                     ok_button;
        /**< Ok button. */
    
    std::auto_ptr< Button >                     cancel_button;
        /**< Cancel button. */
        
    std::auto_ptr< Button >                     create_account_button;
        /**< Button to switch to @c MODE_ASK_NEW_ACCOUNT mode. */
    
    std::auto_ptr< Renderer::TextRectangle >    error_text;
        /**< Text rectangle containing error text. */
    
    std::auto_ptr< Renderer::TextRectangle >    info_text;
        /**< Info text. */
        
    }; // LoginDialog


} // namespace Demo::Client
} // namespace Demo

#endif // CLIENT_LOGIN_DIALOG_H
