/** @file message_box.cpp
 * Simple message box (implementation).
 */

/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "client_global.h"
#include "gui/message_box.h"
#include "gui/widget_layout.h"


namespace Demo {
namespace Client {


namespace {

const Float32 BORDER_WIDTH = 16.0f;
const Float32 BORDER_HEIGHT = 16.0f;
const Float32 PADDING_X = 16.0f;
const Float32 PADDING_Y = 16.0f;

} // anonymous namespace


MessageBox::MessageBox
    (
    int                 flags,
    const std::string & ok,
    const std::string & cancel
    ) : result( RESULT_NONE ),
        message_receiver( NULL )
    {
    // Init widget.

    set_border_size( BORDER_WIDTH, BORDER_HEIGHT );
    set_dragable( true );
    set_auto_bringable_to_front( true );

    // Create widgets.

    message.reset( new Renderer::TextRectangle( &Global::get_console_font() ) );
    message->set_parent_overlay_object( this );

    ok_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    ok_button->set_text( ok );
    ok_button->set_parent_widget( this );
    ok_button->set_message_receiver( this );

    if( flags & FLAG_TWO_BUTTONS )
        {
        cancel_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
        cancel_button->set_text( cancel );
        cancel_button->set_parent_widget( this );
        cancel_button->set_message_receiver( this );
        }

    set_focusable( WINDOW_FOCUSABLE );
    }


MessageBox::~MessageBox()
    {
    }


void MessageBox::set_message
    (
    const std::string &                 text,
    Renderer::TextRectangle::Alignment  alignment
    )
    {
    hope( message.get() );
    message->print_text( text, alignment );
    }


Renderer::TextRectangle & MessageBox::get_message_object()
    {
    hope( message.get() );
    return *message;
    }


void MessageBox::set_ok_text
    (
    const std::string & text
    )
    {
    hope( ok_button.get() );
    ok_button->set_text( text );
    }


void MessageBox::set_cancel_text
    (
    const std::string & text
    )
    {
    if( cancel_button.get() )
        {
        cancel_button->set_text( text );
        }
    }


void MessageBox::set_message_receiver
    (
    MessageBoxReceiver * the_message_receiver
    )
    {
    message_receiver = the_message_receiver;
    }


BlockerWidget * MessageBox::switch_to_modal_mode()
    {
    if( !blocker_widget.get() )
        {
        blocker_widget.reset( new BlockerWidget() );
        }
    blocker_widget->set_base_size_from_screen();
    blocker_widget->bring_to_front();
    set_visibility( true );
    bring_to_front();
    set_focus();
    
    return blocker_widget.get();
    }


void MessageBox::switch_to_modeless_mode()
    {
    blocker_widget.reset();
    }


void MessageBox::do_layout()
    {
    // Layout child objects.

    hope( message.get() );
    hope( ok_button.get() );
    ok_button->set_size_from_text();
    ok_button->do_layout();

    bool has_cancel = false;
    if( cancel_button.get() )
        {
        cancel_button->set_size_from_text();
        cancel_button->do_layout();
        has_cancel = true;
        }

    // Layout message box.

    WidgetTableLayout layout( BORDER_WIDTH, BORDER_HEIGHT, PADDING_X, PADDING_Y );

    layout.next_row();
    layout.next_cell( has_cancel ? 1 : 1, WidgetTableLayout::HALIGN_CENTER );
    layout.put_object( message.get() );

    layout.next_row();
    layout.next_cell( 1, WidgetTableLayout::HALIGN_CENTER );
    layout.put_object( ok_button.get() );

    if( has_cancel )
        {
        //layout.next_cell( 1, WidgetTableLayout::HALIGN_CENTER );
        layout.put_object( cancel_button.get() );
        }

    const SVector2 new_size = layout.do_layout();
    set_base_size( new_size );
    }


MessageBox::Result MessageBox::get_result() const
    {
    return result;
    }


void MessageBox::clear_result()
    {
    result = RESULT_NONE;
    }


void MessageBox::button_clicked
    (
    Button * button
    )
    {
    hope( button );
    if( result == RESULT_NONE )
        {
        if( button == ok_button.get() )
            {
            result = RESULT_OK;
            if( message_receiver )
                {
                message_receiver->message_box_action( this );
                }
            }
        else if( button == cancel_button.get() )
            {
            result = RESULT_CANCEL;
            if( message_receiver )
                {
                message_receiver->message_box_action( this );
                }
            }
        }
    }


bool MessageBox::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    if( key == SDLK_TAB )
        return true;

    // Handle the event by super class.
    if( Super::key_down( key, mod ) )
        {
        return true;
        }

    if( result == RESULT_NONE )
        {
        switch( key )
            {
            case SDLK_ESCAPE:
                {
                result = RESULT_CANCEL;
                if( message_receiver )
                    {
                    message_receiver->message_box_action( this );
                    }
                return true;
                }

            case SDLK_RETURN:
            case SDLK_KP_ENTER:
                {
                result = RESULT_OK;
                if( message_receiver )
                    {
                    message_receiver->message_box_action( this );
                    }
                return true;
                }

            default:
                break;
            }
        }

    return false;
    }


} // namespace Demo::Client
} // namespace Demo
