/** @file message_box.h
 * Simple message box.
 */

/*
 * begin        : Dec 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MESSAGE_BOX_H
#define CLIENT_MESSAGE_BOX_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_PANEL_BORDER_H
#include "gui/panel_border.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#ifndef CLIENT_BLOCKER_WIDGET_H
#include "gui/blocker_widget.h"
#endif

#ifndef CLIENT_TEXT_RECTANGLE_H
#include "renderer/text_rectangle.h"
#endif


namespace Demo {
namespace Client {


/**
 * Simple message box.
 *
 * Its size is set automatically in #do_layout().
 */
class MessageBox :
    public PanelBorder,
    public Button::ButtonMessageReceiver
    {
    NO_COPY_CONSTRUCTOR( MessageBox );
    NO_ASSIGNMENT_OPERATOR( MessageBox );

    /// Superclass.
    typedef PanelBorder Super;

public:

    /**
     * Receiver of messages from the message box.
     */
    class MessageBoxReceiver
        {
    public:

        /**
         * Destructor.
         */
        virtual ~MessageBoxReceiver()
            {
            }

        /**
         * Called when user clocks message box button or presses enter or escape key.
         *
         * Call <tt>message_box-\>get_result()</tt> to determine action taken.
         * Call <tt>message_box-\>clear_result()</tt> if you want to receive more
         * messages from the message box.
         */
        virtual void message_box_action
            (
            MessageBox * message_box
            )
            {
            }

        }; // class MessageBoxReceiver

public:

    /**
     * Message box mode flags.
     */
    enum Flags
        {
        FLAG_TWO_BUTTONS    = 0x01,
            /**< Two buttons ("ok" and "cancel") instead of single "ok" button. */
        };

    /**
     * Message box result.
     */
    enum Result
        {
        RESULT_NONE,
            /**< No result yet. */

        RESULT_OK,
            /**< First button clicked or enter pressed. */

        RESULT_CANCEL,
            /**< Second button clicked or escape pressed. */
        };

public:

    /**
     * Constructor.
     *
     * @param flags Combination of Flags.
     */
    explicit MessageBox
        (
        int                 flags = 0,
        const std::string & ok = "  OK  ",
        const std::string & cancel = "Cancel"
        );

    /**
     * Destructor.
     */
    virtual ~MessageBox();

public:

    /**
     * Set message box text.
     *
     * Text may contain new-line characters, see Renderer::TextRectangle::print_text()
     * for more information.
     */
    void set_message
        (
        const std::string &                 text,
        Renderer::TextRectangle::Alignment  alignment = Renderer::TextRectangle::ALIGN_LEFT
        );

    /**
     * Get reference to message object.
     *
     * Use this instead of #set_text() if you want to create ultracool text
     * message with fancy colors and bump mapped, radiosity lighted, reflective,
     * transparent trinkets.
     */
    Renderer::TextRectangle & get_message_object();

    /**
     * Change text on "ok" button.
     */
    void set_ok_text
        (
        const std::string & text
        );

    /**
     * Change text on "cancel" button.
     */
    void set_cancel_text
        (
        const std::string & text
        );

    /**
     * Set receiver of action messages from the message box.
     *
     * If @a message_receiver is @c NULL, message notifying will be disabled.
     * You should use #get_result() to determine message box in the handler.
     * Note that message box ignores its input unless it's in RESULT_NONE state,
     * call #clear_result() to receive more events.
     */
    void set_message_receiver
        (
        MessageBoxReceiver * message_receiver
        );

    /**
     * Switch message box to modal mode.
     *
     * It creates fullscreen BlockerWidget to block user clicks to other regions
     * of the screen. Message box will be in modal mode until #switch_to_modeless_mode()
     * is called or it's destroyed.
     *
     * Do not hide modal message box, switch them to modeless mode first. Otherwise
     * the blocker widget will remain on screen.
     *
     * @return Pointer to blocker. Do not resize/destroy it. However, you
     *         can change its color and opacity. It will be destroyed automatically
     *         with the message box.
     */
    BlockerWidget * switch_to_modal_mode();

    /**
     * Switch message box to modeless mode.
     *
     * This method destroys blocker widget if it exists.
     */
    void switch_to_modeless_mode();

    /**
     * Call this method after all message box texts are set to layout the box.
     */
    virtual void do_layout();

    /**
     * Get message box result.
     *
     * @return Message box result. May be RESULT_CANCEL even if message box has
     *         only single button and user pressed escape.
     */
    Result get_result() const;

    /**
     * Set message box result to RESULT_NONE.
     */
    void clear_result();

public:

    /**
     * Callback from Button::MessageReceiver.
     */
    virtual void button_clicked
        (
        Button * button
        );

    /**
     * Handle keyboard input.
     */
    virtual bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );

private:

    std::auto_ptr< Renderer::TextRectangle >    message;
        /**< Message box text. */
        
    std::auto_ptr< Button >                     ok_button;
        /**< Primary "ok" button. */

    std::auto_ptr< Button >                     cancel_button;
        /**< Secondary "cancel" button.
             Will be @c NULL unless @c FLAG_TWO_BUTTONS has been set. */

    std::auto_ptr< BlockerWidget >              blocker_widget;
        /**< Blocker widget. Used only in modal mode. */

    Result                                      result;
        /**< Message box result. */

    MessageBoxReceiver *                        message_receiver;
        /**< Receiver of message box actions. */

    }; // class MessageBox


} // namespace Demo::Client
} // namespace Demo


#endif // CLIENT_MESSAGE_BOX_H
