/** @file panel_border.cpp
 * Decorative panel border (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/panel_border.h"
#include "client_global.h"
#include "media_repository.h"


// Animate borders of the panel if some of its children has focus.
#define ANIMATE_BORDERS_IF_FOCUS


namespace Demo {
namespace Client {


PanelBorder::PanelBorder
    (
    const std::string & the_corner_icon,
    const std::string & the_hborder_icon,
    const std::string & the_vborder_icon,
    const std::string & the_background_icon,
    Float32 the_border_width,
    Float32 the_border_height
    ) :
    border_width( the_border_width ),
    border_height( the_border_height )
    {
    MediaRepository & media = Global::get_media_repository();
    
    corner_icon.reset( media.create_icon( the_corner_icon ) );
    hborder_icon.reset( media.create_icon( the_hborder_icon ) );
    vborder_icon.reset( media.create_icon( the_vborder_icon ) );
    background_icon.reset( media.create_icon( the_background_icon ) );
    }


void PanelBorder::set_border_size
    (
    Float32 width,
    Float32 height
    )
    {
    border_width = width;
    border_height = height;
    }


void PanelBorder::render()
    {
    render_background();
    render_borders();
    render_corners();
    }

    
void PanelBorder::render_background()
    {
    hope( background_icon.get() );
    
    Float32             width = get_current_size().x();
    Float32             height = get_current_size().y();
    Float32             texture_width = 256.0f;
    Float32             texture_height = 256.0f;

    // Compute uv coordinates to create repeatable pattern.
    Float32             uv_h_size = width / texture_width;
    Float32             uv_v_size = height / texture_height;

    background_icon->set_base_offset( screenspace_translation().svector() );
    background_icon->set_base_size( get_current_size() );
    background_icon->set_base_color( get_current_color() );
    background_icon->set_uv( 0.0f, 0.0f, uv_h_size, uv_v_size );
    background_icon->render();
    }


void PanelBorder::render_borders()
    {
    hope( hborder_icon.get() );
    hope( vborder_icon.get() );
    Renderer::Icon *    icon = NULL;
    Float32             width = get_current_size().x();
    Float32             height = get_current_size().y();
    Float32             texture_length = 64.0f;
    
    // Compute uv coordinates to create repeatable pattern.
    Float32             uv_h_size = width / texture_length;
    Float32             uv_v_size = height / texture_length;
    
    // Compute screen position of four corners.
    SVector2            tl_corner = screenspace_translation().svector();
    SVector2            tr_corner = tl_corner + SVector2( width, 0.0f );
    SVector2            bl_corner = tl_corner + SVector2( 0.0f, height );
    SVector2            br_corner = tl_corner + SVector2( width, height );
    
    // Draw horizontal borders.
    icon = hborder_icon.get();
    
    icon->set_origin( OverlayObject::TOP_LEFT );
    icon->set_base_color( get_current_color() );
    icon->set_base_size( SVector2( width, border_height ) );
    
    Float32             anim_uv_offset;
    if( has_my_child_focus() )
        {
#ifdef ANIMATE_BORDERS_IF_FOCUS
        const SystemTime curr_system_time = SystemTime::current_time();
        const Float64 offset = Massiv::Core::FloatTraits< Float64 >::mod( curr_system_time.read() / 6.0, 1.0 );
        anim_uv_offset = static_cast< Float32 >( offset );
#else
        anim_uv_offset = 0.0f;
#endif
        }
    else
        {
        anim_uv_offset = 0.0f;
        }
    
    icon->set_uv( anim_uv_offset + 0.0f, 0.0f, anim_uv_offset + uv_h_size, 1.0f );
    
    render_corner( icon, tl_corner, false, false );
    render_corner( icon, bl_corner, false, true );
    
    // Draw vertical borders.
    icon = vborder_icon.get();
    
    icon->set_origin( OverlayObject::TOP_LEFT );
    icon->set_base_color( get_current_color() );
    icon->set_base_size( SVector2( border_width, height ) );
    
    icon->set_uv( 0.0f, anim_uv_offset + 0.0f, 1.0f, anim_uv_offset + uv_v_size );
    
    render_corner( icon, tl_corner, false, false );
    render_corner( icon, tr_corner, true, false );
    }
    

void PanelBorder::render_corners()
    {
    hope( corner_icon.get() );
    Renderer::Icon *    icon = corner_icon.get();
    Float32             width = get_current_size().x();
    Float32             height = get_current_size().y();

    // Compute screen position of four corners.
    SVector2            tl_corner = screenspace_translation().svector();
    SVector2            tr_corner = tl_corner + SVector2( width, 0.0f );
    SVector2            bl_corner = tl_corner + SVector2( 0.0f, height );
    SVector2            br_corner = tl_corner + SVector2( width, height );
    
    icon->set_origin( OverlayObject::TOP_LEFT );
    icon->set_base_size( SVector2( border_width, border_height ) );
    icon->set_base_color( get_current_color() );
    
    // Render four corners.
    render_corner( icon, tl_corner, false, false );
    render_corner( icon, tr_corner, true, false );
    render_corner( icon, bl_corner, false, true );
    render_corner( icon, br_corner, true, true );
    }
    

void PanelBorder::render_corner
    (
    Renderer::Icon *    icon,
    const SVector2 &    screen_position,
    bool                flip_x,
    bool                flip_y
    )
    {
    hope( icon );

    // Original uv.
    SVector2    origin_top_left_uv;
    SVector2    origin_bottom_right_uv;
    
    // Flipped uv.
    SVector2    top_left_uv;
    SVector2    bottom_right_uv;
    
    // Final screen position.
    SVector2    pos = screen_position;
    
    if( flip_x || flip_y )
        {
        // Store original uv.
        origin_top_left_uv = icon->get_top_left_uv().svector();
        origin_bottom_right_uv = icon->get_bottom_right_uv().svector();
        top_left_uv = icon->get_top_left_uv().svector();
        bottom_right_uv = icon->get_bottom_right_uv().svector();
        }
    
    if( flip_x )
        {
        // Flip texture coordinate.
        Float32 t;
        t = top_left_uv.x();
        top_left_uv.set_x( bottom_right_uv.x() );
        bottom_right_uv.set_x( t );
        
        // Translate left.
        pos.set_x( pos.x() - icon->get_current_size().x() );
        }
    
    if( flip_y )
        {
        // Flip texture coordinate.
        Float32 t;
        t = top_left_uv.y();
        top_left_uv.set_y( bottom_right_uv.y() );
        bottom_right_uv.set_y( t );
        
        // Translate up.
        pos.set_y( pos.y() - icon->get_current_size().y() );
        }
    
    if( flip_x || flip_y )
        {
        // Set new flipped uv.
        icon->set_uv( top_left_uv.x(), top_left_uv.y(),
            bottom_right_uv.x(), bottom_right_uv.y() );
        }
    
    icon->set_base_offset( pos );
    icon->render();
    
    if( flip_x || flip_y )
        {
        // Restore original uv.
        icon->set_uv( origin_top_left_uv.x(), origin_top_left_uv.y(),
            origin_bottom_right_uv.x(), origin_bottom_right_uv.y() );
        }
    }


} // namespace Demo::Client
} // namespace Demo
