/** @file panel_border.h
 * Decorative panel border.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_PANEL_BORDER_H
#define CLIENT_PANEL_BORDER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_ICON_H
#include "renderer/icon.h"
#endif

#include <memory>


namespace Demo {
namespace Client {


/**
 * PanelBorder is simple widget object. When drawn,
 * it fills its screen rectangle with background texture
 * and decorative border.
 */
class PanelBorder : public Widget
    {
    NO_COPY_CONSTRUCTOR( PanelBorder );
    NO_ASSIGNMENT_OPERATOR( PanelBorder );

public:

    /**
     * Constructor.
     * Icon parameters are names of icon templates from media repository.
     * @param corner_icon Image to draw at corners.
     * @param hborder_icon Image containing horizontal border.
     *      The texture must be repeatable in horizontal direction.
     *      The icon must occupy entire texture.
     * @param hborder_icon Image containing vertical border.
     *      The texture must be repeatable in vertical direction.
     *      The icon must occupy entire texture.
     * @param background_icon Background texture.
     *      The texture must be repeatable in both directions.
     *      The icon must occupy entire texture.
     */
    PanelBorder
        (
        const std::string & corner_icon = "gui/panel_corner",
        const std::string & hborder_icon = "gui/panel_hborder",
        const std::string & vborder_icon = "gui/panel_vborder",
        const std::string & background_icon = "gui/panel_background",
        Float32 border_width = 48.0f,
        Float32 border_height = 48.0f
        );
        
public:

    /**
     * Set border width and height.
     */
    void set_border_size
        (
        Float32 width,
        Float32 height
        );
        
public:

    // Renderer::OverlayObject
    virtual void render();

private:

    /**
     * Render background texture.
     */
    void render_background();
    
    /**
     * Draw four borders.
     */
    void render_borders();
    
    /**
     * Draw four corners.
     */
    void render_corners();

    /**
     * Helper function. Draws @a icon at @a screen_position.
     * Origin, size and color of the @a icon must be already set.
     * @param flip_x If true, the icon is flipped around
     *      @a screen_position in horizontal direction.
     * @param flip_u If true, the icon is flipped around
     *      @a screen_position in vertical direction.
     */
    void render_corner
        (
        Renderer::Icon *    icon,
        const SVector2 &    screen_position,
        bool                flip_x,
        bool                flip_y
        );

private:

    std::auto_ptr< Renderer::Icon >     corner_icon;
        /**< Corner image. */

    std::auto_ptr< Renderer::Icon >     hborder_icon;
        /**< Horizontal border texture. */
        
    std::auto_ptr< Renderer::Icon >     vborder_icon;
        /**< Vertical border texture. */
        
    std::auto_ptr< Renderer::Icon >     background_icon;
        /**< Background texture. */
        
    Float32                             border_width;
        /**< Border width. */
        
    Float32                             border_height;
        /**< Border height. */

    }; // PanelBorder


} // namespace Demo::Client
} // namespace Demo


#endif // CLIENT_PANEL_BORDER_H
