/** @file popup_menu.cpp
 * Simple popup menu (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/popup_menu.h"
#include "client_global.h"
#include "input/input_layers.h"
#include "renderer/layers.h"
#include "gui/widget_input_dispatcher.h"


namespace Demo {
namespace Client {


PopupMenu * PopupMenu::active_popup_menu = NULL;


PopupMenu::PopupMenu() :
    padding_x( 8.0f ),
    padding_y( 8.0f )
    {
    set_auto_bringable_to_front( true );
    set_border_size( padding_x, padding_y );
    set_visibility( false );
    }


PopupMenu::~PopupMenu()
    {
    if( active_popup_menu && active_popup_menu == this )
        {
        active_popup_menu = NULL;
        }
    }


void PopupMenu::initialize()
    {
    set_parent_widget( &Global::get_widget_input_dispatcher() );
    }


void PopupMenu::do_layout()
    {
    Super::do_layout();
    // Do not call update_size() because size should not be changes
    // from do_layout() method.
    // We only resize all buttons to same width.
    update_buttons();
    }


void PopupMenu::set_visibility
    (
    bool visible
    )
    {
    // TODO: Should be set_visibility() made virtual? Is there anybody
    // calling set_visibility() on PopupMenu using pointer to OverlayObject
    // base class?
    if( visible )
        {
        if( active_popup_menu && active_popup_menu != this )
            {
            active_popup_menu->set_visibility( false );
            hope( active_popup_menu == NULL );
            }
        Super::set_visibility( visible );
        active_popup_menu = this;
        bring_to_front();
        }
    else
        {
        Super::set_visibility( visible );
        if( active_popup_menu == this )
            {
            active_popup_menu = NULL;
            }
        }
    }


void PopupMenu::clear()
    {
    items.clear();
    update_size();
    }


PopupMenu::ItemIndex PopupMenu::push
    (
    const Item & the_item
    )
    {
    ItemIndex result = items.size();
    Item item = the_item;
    
    SVector2 button_position = SVector2( padding_x, padding_y + get_button_area_height() );
    
    item.button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    item.button->set_text( item.name );
    item.button->set_parent_widget( this );
    item.button->set_message_receiver( this );
    item.button->set_size_from_text();
    item.button->set_base_offset( button_position );
    item.button->do_layout();

    items.push_back( item );
    update_size();
    return result;
    }


void PopupMenu::set_message_receiver
    (
    PopupMenuMessageReceiver * the_message_receiver
    )
    {
    message_receiver = the_message_receiver;
    }


void PopupMenu::button_clicked
    (
    Button * button
    )
    {
    // Find the button that was clicked.
    Items::iterator it;
    Item * found_item = NULL;
    for( it = items.begin(); it != items.end(); ++it )
        {
        Item & item = *it;
        if( item.button.get() == button )
            {
            found_item = &item;
            break;
            }
        }
    
    if( found_item )
        {
        if( message_receiver )
            {
            message_receiver->popup_menu_item_clicked( this, found_item->id );
            }
        set_visibility( false );
        }
    }


Float32 PopupMenu::update_buttons()
    {
    Float32 w = 0.0f;
    Items::iterator it;
    
    // Get maximum width of a button.
    for( it = items.begin(); it != items.end(); ++it )
        {
        Item & item = *it;
        Float32 iw = item.button->get_current_size().x();
        if( iw > w )
            {
            w = iw;
            }
        }
    
    // Popup menu width can only grow.
    Float32 min_width = get_current_size().x() - 2.0f * padding_x;
    if( w < min_width )
        {
        w = min_width;
        }
    
    // Resize all buttons.
    for( it = items.begin(); it != items.end(); ++it )
        {
        Item & item = *it;
        item.button->set_base_size( SVector2( w, item.button->get_base_size().y() ) );
        item.button->do_layout();
        }

    return w;
    }


Float32 PopupMenu::get_button_area_height() const
    {
    Float32 h = 0.0f;
    Items::const_iterator it;
    for( it = items.begin(); it != items.end(); ++it )
        {
        const Item & item = *it;
        h += item.button->get_current_size().y();
        }
    return h;
    }


void PopupMenu::update_size()
    {
    Float32 buttons_width = update_buttons();
    Float32 buttons_height = get_button_area_height();
    SVector2 size = SVector2( 2.0f * padding_x + buttons_width, 2.0f * padding_y + buttons_height );
    set_base_size( size );
    }
    


} // namespace Demo::Client
} // namespace Demo
