/** @file popup_menu.h
 * Simple popup menu.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_POPUP_MENU_H
#define CLIENT_POPUP_MENU_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#ifndef CLIENT_PANEL_BORDER_H
#include "gui/panel_border.h"
#endif

#include <memory>
#include <string>


namespace Demo {
namespace Client {


class WidgetInputDispatcher;


/**
 * Simple popup menu.
 *
 * Only one popup menu can be visible at a time. Before a popup menu
 * is made visible, the active popup menu is automatically hidden.
 *
 * @warning Popup menus should not be linked as child widgets using @c set_parent_widget().
 *     Use @c initialize() method instead.
 *
 * @note You don't need to take care of PopupMenu's size. PopupMenu
 *     automatically resizes itself to ideal proportions based on its
 *     content.
 * @note If you wanna resize the popup menu, you should never make the size
 *     of the popup menu smaller than the automatic size. You should only
 *     extend width of the popup menu size. Do not forget to call @c do_layout()
 *     after the resizement.
 */
class PopupMenu :
    public PanelBorder,
    public Button::ButtonMessageReceiver
    {
    typedef PanelBorder Super;
    
    NO_COPY_CONSTRUCTOR( PopupMenu );
    NO_ASSIGNMENT_OPERATOR( PopupMenu );

    friend class WidgetInputDispatcher; // Access to active_popup_menu.
    
public:
    
    typedef std::size_t ItemIndex;

public:

    /** #Popup menu item. */
    struct Item
        {
        friend class PopupMenu;

    public:
    
        Item() :
            id( 0 )
            {
            }
        
        Item
            (
            int the_id,
            const std::string & the_name
            ) :
            id( the_id ),
            name( the_name )
            {
            }
        
        Item
            (
            const Item & other
            ) :
            id( other.id ),
            name( other.name )
            {
            button.reset( other.button.release() );
            }
        
        Item & operator=
            (
            const Item & other
            )
            {
            id = other.id;
            name = other.name;
            button.reset( other.button.release() );
            return *this;
            }
    
    public:
            
        int             id;
            /**< User defined integer value. */
        std::string     name;
            /**< Item name. */
        
    private:
    
        mutable std::auto_ptr< Button > button;
            /**< Button used to draw the item. For #PopupMenu purposes only. */

        }; // Item
        
    /** Popup menu message receiver. */
    class PopupMenuMessageReceiver
        {
    public:

        /**
         * Destructor.
         */
        virtual ~PopupMenuMessageReceiver()
            {
            }

        /**
         * Called whem user clicks on one of the items on the popup menu.
         * @param item_id User defined identifiaction if the item specified
         *     by the user when the item was inserted to the popup menu.
         */
        virtual void popup_menu_item_clicked
            (
            PopupMenu * popup_menu,
            int item_id
            )
            {
            }
        
        }; // ButtonMessageReceiver
    

public:

    /** Constructor. */
    PopupMenu();
    
    /** Destructor. */
    ~PopupMenu();
    
    /**
     * Initialize the popup menu.
     * Popup menus should not be linked as child widgets using @c set_parent_widget().
     * Instead the popup menus should be initialized with this method.
     */
    void initialize();
    
public:

    // From superclass.
    virtual void do_layout();

public:

    // Overriden method from OverlayObject.
    void set_visibility
        (
        bool visible
        );

public:

    /** Clear popup menu content. All items are removed. */
    void clear();
    
    /**
     * Insert @a item at the end of the popup menu. 
     * @return Index of the inserted item.
     */
    ItemIndex push
        (
        const Item & item
        );
        
    /**
     * Set #PopupMenuMessageReceiver of the popup menu.
     * If @a message_receiver is @c NULL, message notifying will be
     * disabled.
     */
    void set_message_receiver
        (
        PopupMenuMessageReceiver * message_receiver
        );

public:

    // From ButtonMessageReceiver.
    virtual void button_clicked
        (
        Button * button
        );
        
private:

    /** Update width of all buttons. Returns width of are occupied by buttons. */
    Float32 update_buttons();
    
    /** Get height of area occupied by buttons. */
    Float32 get_button_area_height() const;
    
    /** Update size of the popup menu according to area occupied by buttons. */
    void update_size();

private:

    typedef std::vector< Item > Items;
    
    Items                       items;
        /**< List of items. */
        
protected:
    
    PopupMenuMessageReceiver *  message_receiver;
        /**< Message receiver. */
    Float32 padding_x;
        /**< Horizontal padding between popup menu border and button area. */
    Float32 padding_y;
        /**< Vertical padding between popup menu border and button area. */
        
private:

    static PopupMenu *          active_popup_menu;
        /**< Aliasing pointer to active popup menu. */

    }; // PopupMenu


} // namespace Demo::Client
} // namespalice Demo


#endif // CLIENT_POPUP_MENU_H
