/** @file scroll_bar.cpp
 * Simple scroll bar (implementation).
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/scroll_bar.h"
#include "client_global.h"
#include "media_repository.h"
#include "input/input_manager.h"


namespace Demo {
namespace Client {


// ScrollBar::Box
//~~~~~~~~~~~~~~~

ScrollBar::Box::Box() :
    light_color( 1.0f, 1.0f, 1.0f, 0.6f ),
    dark_color( 0.0f, 0.0f, 0.0f, 0.6f ),
    highlight_color( 1.0f, 1.0f, 1.0f, 0.3f ),
    inverted( false ),
    highlight( false )
    {
    MediaRepository & media = Global::get_media_repository();
    
    white.reset( new Renderer::Icon( "texture/white" ) );
    background.reset( media.create_icon( "gui/panel_background" ) );
    }


void ScrollBar::Box::render
    (
    const SVector2 & screen_position,
    const SVector2 & size,
    Float32          alpha
    )
    {
    // Render background pattern.
        {
        SColor c = SColor( 1.0f, 1.0f, 1.0f, alpha );
        background->set_base_offset( screen_position );
        background->set_base_size( size );
        background->set_base_color( c );
        background->set_pattern_uv();
        background->render();
        }
    
    // Highlight.
    if( highlight )
        {
        SColor c = highlight_color;
        c *= alpha;
        white->set_base_offset( screen_position );
        white->set_base_size( size );
        white->set_base_color( c );
        white->render();
        }
    
    // Render frame.
        {
        const SVector2 & top_left = screen_position;
        SColor c1 = light_color;
        SColor c2 = dark_color;
        c1 *= alpha;
        c2 *= alpha;
        render_hline( inverted ? c2 : c1, top_left.x(), top_left.x() + size.x(), top_left.y() );
        render_vline( inverted ? c2 : c1, top_left.x(), top_left.y(), top_left.y() + size.y() );
        render_hline( inverted ? c1 : c2, top_left.x(), top_left.x() + size.x(), top_left.y() + size.y() - 1.0f );
        render_vline( inverted ? c1 : c2, top_left.x() + size.x() - 1.0f, top_left.y(), top_left.y() + size.y() );
        }
    }


void ScrollBar::Box::render_hline
    (
    const SColor &  color,
    Float32         x1,
    Float32         x2,
    Float32         y
    )
    {
    const SVector2 top_left( x1, y );
    const SVector2 size( x2 - x1, 1.0f );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


void ScrollBar::Box::render_vline
    (
    const SColor &  color,
    Float32         x,
    Float32         y1,
    Float32         y2
    )
    {
    const SVector2 top_left( x, y1 );
    const SVector2 size( 1.0f, y2 - y1 );
    white->set_base_offset( top_left );
    white->set_base_size( size );
    white->set_base_color( color );
    white->render();
    }


// ScrollBar
//~~~~~~~~~~


ScrollBar::ScrollBar
    (
    Type the_type
    ) :
    message_receiver( NULL ),
    type( the_type ),
    scroll_position( 0 ),
    scroll_range( 0, 0 ),
    scroll_page_size( 1 ),
    scroll_line_size( 1 ),
    scroll_box_draging( false )
    {
    line_up_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    line_up_button->set_parent_widget( this );
    line_up_button->set_message_receiver( this );
    line_up_button->set_text( "-" );
    
    line_down_button.reset( new Button( Button::TYPE_ONE_SHOT ) );
    line_down_button->set_parent_widget( this );
    line_down_button->set_message_receiver( this );
    line_down_button->set_text( "+" );
    
    background_box.inverted = true;
    background_box.highlight = true;
    background_box.highlight_color = SColor( 0.0f, 0.0f, 0.0f, 0.3f );
    }


ScrollBar::~ScrollBar()
    {
    }


void ScrollBar::do_layout()
    {
    Super::do_layout();
    
    const Float32 bsize = get_button_size();
    
    line_up_button->set_base_offset( SVector2( 0.0f, 0.0f ) );
    line_up_button->set_base_size( SVector2( bsize, bsize ) );
    line_up_button->do_layout();
    
    line_down_button->set_base_offset( ( type == HORIZONTAL ?
        SVector2( get_current_size().x() - bsize, 0.0f ) :
        SVector2( 0.0f, get_current_size().y() - bsize ) ) );
    line_down_button->set_base_size( SVector2( bsize, bsize ) );
    line_down_button->do_layout();
    }


void ScrollBar::line_up
    (
    bool notify
    )
    {
    set_scroll_position( scroll_position - scroll_line_size, notify );
    }


void ScrollBar::line_down
    (
    bool notify
    )
    {
    set_scroll_position( scroll_position + scroll_line_size, notify );
    }


void ScrollBar::page_up
    (
    bool notify
    )
    {
    set_scroll_position( scroll_position - scroll_page_size, notify );
    }

    
void ScrollBar::page_down
    (
    bool notify
    )
    {
    set_scroll_position( scroll_position + scroll_page_size, notify );
    }


int ScrollBar::get_scroll_position() const
    {
    return scroll_position;
    }


void ScrollBar::set_scroll_position
    (
    int new_scroll_position,
    bool notify
    )
    {
    scroll_position = new_scroll_position;
    clamp_scroll_position();
    
    if( notify && message_receiver )
        {
        message_receiver->scroll_bar_position_changed( this );
        }
    }


ScrollBar::ScrollRange ScrollBar::get_scroll_range() const
    {
    return scroll_range;
    }


void ScrollBar::set_scroll_range
    (
    const ScrollRange & new_scroll_range,
    bool notify
    )
    {
    scroll_range = new_scroll_range;
    clamp_scroll_position();
    if( notify && message_receiver )
        {
        message_receiver->scroll_bar_position_changed( this );
        }
    }


int ScrollBar::get_scroll_page_size() const
    {
    return scroll_page_size;
    }


void ScrollBar::set_scroll_page_size
    (
    int new_scroll_page_size
    )
    {
    scroll_page_size = new_scroll_page_size;
    }


int ScrollBar::get_scroll_line_size() const
    {
    return scroll_line_size;
    }


void ScrollBar::set_scroll_line_size
    (
    int new_scroll_line_size
    )
    {
    scroll_line_size = new_scroll_line_size;
    }

    
void ScrollBar::button_clicked
    (
    Button * button
    )
    {
    if( button == line_up_button.get() )
        {
        line_up();
        }
    else if( button == line_down_button.get() )
        {
        line_down();
        }
    }


void ScrollBar::mouse_button_grab_lost
    (
    MouseButtonGrab grab
    )
    {
    if( grab == LEFT_MOUSE_BUTTON_GRAB )
        {
        scroll_box_draging = false;
        }
    }


bool ScrollBar::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    if( Super::mouse_button_down( button, mod ) )
        {
        return true;
        }
    
    // Check if clicked on scroll box.
    do  {
        if( button != SDL_BUTTON_LEFT )
            {
            break;
            }

        const SVector2 screen_position = get_top_left_corner_on_screen();
        const Float32  bsize = get_button_size();
        const SVector2 sb_screen_position = ( type == HORIZONTAL ?
            SVector2( screen_position.x() + bsize + get_scroll_box_offset(), screen_position.y() ) :
            SVector2( screen_position.x(), screen_position.y() + bsize + get_scroll_box_offset() ) );
        const SVector2 mouse_position = Global::get_input_manager().get_mouse_position();
    
        if( mouse_position.x() < sb_screen_position.x() ||
            mouse_position.x() >= sb_screen_position.x() + bsize  ||
            mouse_position.y() < sb_screen_position.y() ||
            mouse_position.y() >= sb_screen_position.y() + bsize )
            {
            // Out of scroll box.
            if( mouse_position.y() < sb_screen_position.y() )
                {
                page_up( true );
                }
            if( mouse_position.y() >= sb_screen_position.y() + bsize )
                {
                page_down( true );
                }    
            break;
            }
        
        scroll_box_draging = true;
        scroll_box_drag_begin_mouse_position = mouse_position;
        scroll_box_drag_begin_offset = get_scroll_box_offset();
        return true;
        }
    while( false );

    return false;
    }


bool ScrollBar::mouse_move
    (
    int    dx,
    int    dy,
    SDLMod mod
    )
    {
    if( scroll_box_draging && mouse_button_grabbed( LEFT_MOUSE_BUTTON_GRAB ) )
        {
        const SVector2 mouse_position = Global::get_input_manager().get_mouse_position();
        const SVector2 mouse_delta = mouse_position - scroll_box_drag_begin_mouse_position;
        const Float32 new_offset = ( type == HORIZONTAL ?
            scroll_box_drag_begin_offset + mouse_delta.x() :
            scroll_box_drag_begin_offset + mouse_delta.y() );
        set_scroll_box_offset( new_offset );
        return true;
        }

    return false;
    }


void ScrollBar::render()
    {
    const SVector2 screen_position = get_top_left_corner_on_screen();
    const SVector2 size = get_current_size();
    const Float32 alpha = get_current_color().alpha();
    const Float32 bsize = get_button_size();
    
    // Render background area.
        {
        const SVector2 bg_size = ( type == HORIZONTAL ?
            SVector2( size.x() - 2.0f * bsize, size.y() ) :
            SVector2( size.x(), size.y() - 2.0f * bsize ) );
        const SVector2 bg_screen_position = ( type == HORIZONTAL ?
            SVector2( screen_position.x() + bsize, screen_position.y() ) :
            SVector2( screen_position.x(), screen_position.y() + bsize ) );
        background_box.render( bg_screen_position, bg_size, alpha );
        }
    
    // Render scroll box.
        {
        const SVector2 sb_size = SVector2( bsize, bsize );
        const SVector2 sb_screen_position = ( type == HORIZONTAL ?
            SVector2( screen_position.x() + bsize + get_scroll_box_offset(), screen_position.y() ) :
            SVector2( screen_position.x(), screen_position.y() + bsize + get_scroll_box_offset() ) );
        scroll_box.render( sb_screen_position, sb_size, alpha );
        }
    }


void ScrollBar::clamp_scroll_position()
    {
    scroll_position = ( scroll_position < scroll_range.min_position ?
        scroll_range.min_position : scroll_position );
    scroll_position = ( scroll_position > scroll_range.max_position ?
        scroll_range.max_position : scroll_position );
    }


Float32 ScrollBar::get_button_size() const
    {
    const SVector2 size = get_current_size();
    const Float32 result = ( type == HORIZONTAL ? size.y() : size.x() );
    return ( result < 1.0f ? 1.0f : result );
    }


Float32 ScrollBar::get_scroll_box_area_size() const
    {
    const SVector2 size = get_current_size();
    const Float32 bsize = get_button_size();
    Float32 result = ( type == HORIZONTAL ? size.x() : size.y() );
    
    result -= bsize; // up button
    result -= bsize; // down button
    
    return ( result < 1.0f ? 1.0f : result );
    }


Float32 ScrollBar::get_scroll_box_offset() const
    {
    const Float32 bsize = get_button_size();
    const Float32 sb_area_size = get_scroll_box_area_size();
    const Float32 s_pos = static_cast< Float32 >( scroll_position - scroll_range.min_position );
    const Float32 s_delta = static_cast< Float32 >( scroll_range.max_position - scroll_range.min_position );
    if( s_delta <= 0.0f )
        {
        return 0.0f;
        }
    const Float32 result = ( ( sb_area_size - bsize ) * s_pos ) / s_delta;
    return result;
    }


void ScrollBar::set_scroll_box_offset
    (
    Float32 new_offset
    )
    {
    const Float32 bsize = get_button_size();
    const Float32 sb_area_size = get_scroll_box_area_size();
    if( sb_area_size - bsize <= 0.0f )
        {
        set_scroll_position( scroll_range.min_position );
        return;
        }
    const Float32 s_delta = static_cast< Float32 >( scroll_range.max_position - scroll_range.min_position );
    const Float32 s_pos = ( new_offset * s_delta ) / ( sb_area_size - bsize ) + scroll_range.min_position;
    const int new_pos = static_cast< int >( Massiv::Core::FloatTraits< Float32 >::round( s_pos ) );
    set_scroll_position( new_pos );
    }


void ScrollBar::set_message_receiver
    (
    ScrollBarMessageReceiver * receiver
    )
    {
    message_receiver = receiver;
    }
        

} // namespace Demo::Client
} // namespalice Demo
