/** @file scroll_bar.h
 * Simple scroll bar.
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_SCROLL_BAR_H
#define CLIENT_SCROLL_BAR_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_BUTTON_H
#include "gui/button.h"
#endif

#include <memory>
#include <string>


namespace Demo {
namespace Client {


/**
 * Simple scroll bar.
 *
 * Usage:
 * @code
 * std::auto_ptr< ScrollBar > sb( new ScrollBar( ScrollBar::HORIZONTAL ) );
 * sb->set_parent_widget( parent );
 * // Do not set message receiver now if it is not yet initialized
 * // so the commands below will not invoke the receiver's callbacks.
 * sb->set_base_size( scroll_box_size );
 * sb->do_layout();
 * sb->set_scroll_range( ScrollBar::ScrollRange( 0, 100 ) );
 * sb->set_scroll_position( 20 );
 * sb->set_scroll_line_size( 1 );
 * sb->set_scroll_page_size( 10 );
 * sb->set_message_receiver( message_receiver ); // Set message receiver at the end.
 * @endcode
 */
class ScrollBar :
    public Widget,
    public Button::ButtonMessageReceiver
    {
    typedef Widget  Super;
    
    NO_COPY_CONSTRUCTOR( ScrollBar );
    NO_ASSIGNMENT_OPERATOR( ScrollBar );

public:

    /** Scroll bar type. */
    enum Type
        {
        HORIZONTAL,
        VERTICAL
        }; // Type

    /** Scroll range of scroll box. */
    struct ScrollRange
        {
        /** Constructor. */
        ScrollRange
            (
            int the_min_position,
            int the_max_position
            ) :
            min_position( the_min_position ),
            max_position( the_max_position )
            {
            }
        
        int min_position;
            /**< Minimum scroll box position. */
        int max_position;
            /**< Maximum scroll box position. */
        }; // ScrollRange

    /** Scroll bar message receiver. */
    class ScrollBarMessageReceiver
        {
    public:

        /** Destructor. */
        virtual ~ScrollBarMessageReceiver()
            {
            }

        /**
         * Called when scroll box position has chaged.
         */
        virtual void scroll_bar_position_changed
            (
            ScrollBar * scroll_bar
            )
            {
            }
        
        }; // ScrollBarMessageReceiver

    /**
     * Overlay object that is drawn as box with light and dark borders.
     * Used to draw scroll box and background.
     */
    class Box
        {
        NO_COPY_CONSTRUCTOR( Box );
        NO_ASSIGNMENT_OPERATOR( Box );

    public:
    
        /** Constructor. */
        Box();
        
        /** Render the box. */
        void render
            (
            const SVector2 & screen_position,
            const SVector2 & size,
            Float32          alpha
            );

        /** Render horizontal line. */
        void render_hline
            (
            const SColor &  color,
            Float32         x1,
            Float32         x2,
            Float32         y
            );

        /** Render vertical line. */
        void render_vline
            (
            const SColor &  color,
            Float32         x,
            Float32         y1,
            Float32         y2
            );

    private:
    
        std::auto_ptr< Renderer::Icon > white;
            /**< White texture used to draw lines and highlight rectangle. */
        std::auto_ptr< Renderer::Icon > background;
            /**< Background texture. */
    
    public:
        SColor                          light_color;
            /**< Color used for light parts of the box borders. */
        SColor                          dark_color;
            /**< Color used for dark parts of the box borders. */
        SColor                          highlight_color;
            /**< Color used for highlighting the box. */
        bool                            inverted;
            /**< If true, light parts of the box borders will be drawn by @c dark_color and
             * dark parts will be drawn by @c light_color. */
        bool                            highlight;
            /**< If true, the box will be highlighted. */

        }; // Box

public:

    /** Constructor. */
    ScrollBar
        (
        Type type
        );
    
    /** Destructor. */
    ~ScrollBar();
    
public:

    // From superclass.
    virtual void do_layout();
    
public:

    /** Scroll one line up. */
    void line_up
        (
        bool notify_message_receiver = true
        );

    /** Scroll one line down. */
    void line_down
        (
        bool notify_message_receiver = true
        );
        
    /** Scroll one page up. */
    void page_up
        (
        bool notify_message_receiver = true
        );

    /** Scroll one page down. */
    void page_down
        (
        bool notify_message_receiver = true
        );
    
    /** Get current scroll position. */
    int get_scroll_position() const;
    
    /** Set scroll position. */
    void set_scroll_position
        (
        int new_scroll_position,
        bool notify_message_receiver = true
        );

    /** Get current scroll range. */
    ScrollRange get_scroll_range() const;
    
    /** Set scroll range. */
    void set_scroll_range
        (
        const ScrollRange & new_scroll_range,
        bool notify_message_receiver = true
        );

    /** Get current page size. */
    int get_scroll_page_size() const;
    
    /** Set page size. */
    void set_scroll_page_size
        (
        int new_scroll_page_size
        );

    /** Get current line size. */
    int get_scroll_line_size() const;
    
    /** Set line size. */
    void set_scroll_line_size
        (
        int new_scroll_line_size
        );

    /** Set message receiver. */
    void set_message_receiver
        (
        ScrollBarMessageReceiver * receiver
        );
        
public:

    // From ButtonMessageReceiver.
    virtual void button_clicked
        (
        Button * button
        );

    // From Widget.
    virtual void mouse_button_grab_lost
        (
        MouseButtonGrab grab
        );

    // From Widget.
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );

    // From Widget.
    virtual bool mouse_move
        (
        int    dx,
        int    dy,
        SDLMod mod
        );

    // From OverlayObject.
    virtual void render();
    
private:
    
    /** Make scroll_position to be in scroll range. */
    void clamp_scroll_position();
    
    /** Get width/height of a button (line-up, line-down or scroll box). */
    Float32 get_button_size() const;
    
    /** Get size of area where scroll box can move. */
    Float32 get_scroll_box_area_size() const;
    
    /** Get offset of scroll box in scroll box area. */
    Float32 get_scroll_box_offset() const;
    
    /**
     * Set new scroll box offset in scroll box area. Current scroll
     * position is updated as well.
     */
    void set_scroll_box_offset
        (
        Float32 new_offset
        );

private:

    ScrollBarMessageReceiver *  message_receiver;
        /**< Scroll bar message receiver. */
    Type                        type;
        /**< Scroll bar type. */
    int                         scroll_position;
        /**< Current scroll position. */
    ScrollRange                 scroll_range;
        /**< Current scroll range of scroll box. */
    int                         scroll_page_size;
        /**< Current scroll page size. */
    int                         scroll_line_size;
        /**< Current scroll line size. */
    std::auto_ptr< Button >     line_up_button;
        /**< Line up button. */
    std::auto_ptr< Button >     line_down_button;
        /**< Line down button. */
    Box                         background_box;
        /**< Background of the scroll bar. */
    Box                         scroll_box;
        /**< Scroll box. */
    bool                        scroll_box_draging;
        /**< If true, user is deaging the scroll box. */
    SVector2                    scroll_box_drag_begin_mouse_position;
        /**< Mouse position when scroll box draging started. */
    float                       scroll_box_drag_begin_offset;
        /**< Scroll box offset when scroll box draging started. */

    }; // ScrollBar


} // namespace Demo::Client
} // namespalice Demo


#endif // CLIENT_SCROLL_BAR_H
