/** @file slot_panel.cpp
 * Slot panel (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/slot_panel.h"
#include "media_repository.h"
#include "input/input_manager.h"
#include "client_global.h"
#include "renderer/font.h"


namespace Demo {
namespace Client {


// SlotPanel::Slot
//~~~~~~~~~~~~~~~~


void SlotPanel::Slot::set_text
    (
    const std::string & the_text
    )
    {
    text_rectangle.reset( new Renderer::TextRectangle() );
    text_rectangle->set_font( &Global::get_console_font() );
    text_rectangle->print_line( the_text );
    }


// SlotPanel
//~~~~~~~~~~


SlotPanel::SlotPanel
    (
    const std::string & the_title
    ) :
    PanelBorder
        (
        "gui/panel_corner",
        "gui/panel_hborder",
        "gui/panel_vborder",
        "gui/panel_background"
        ),
    rows( 0 ),
    cols( 0 ),
    selected_slot_index( -1 ),
    slot_hpadding( 8.0f ),
    slot_vpadding( 8.0f ),
    message_receiver( NULL )
    {
    hope( cols >= 0 );
    hope( rows >= 0 );
    
    MediaRepository & media = Global::get_media_repository();
    
    set_dragable( true );
    set_auto_bringable_to_front( true );
    
    // Title.
    // Title is registered as child overlay object so its offset is relative
    // to the panel.
    
    title.reset( new Renderer::TextRectangle() );
    title->set_font( &Global::get_console_font() );
    title->print_line( the_title );
    title->set_parent_overlay_object( this );
    
    // Slot images.
    
    slot_background_icon.reset( media.create_icon( "gui/slot_panel/slot_background" ) );
    slot_highlight_icon.reset( media.create_icon( "gui/slot_panel/slot_highlight" ) );
    
    // Close button.
    
    close_button.reset
        (
        new IconButton
            (
            IconButton::TYPE_ONE_SHOT,
            "gui/close_button_up",
            "gui/close_button_down",
            "gui/close_button_up_mouse_on",
            "gui/close_button_down_mouse_on"
            )
        );
    
    close_button->set_origin( Renderer::OverlayObject::TOP_RIGHT );
    close_button->set_base_offset( SVector2( 0.0f, 0.0f ) );
    close_button->set_base_size( SVector2( 28.0f, 28.0f ) );
    close_button->set_parent_widget( this );
    close_button->set_message_receiver( this );
    
    // Slot title.

    slot_title.reset( new Renderer::TextRectangle() );
    slot_title->set_font( &Global::get_console_font() );
    slot_title->set_parent_overlay_object( this );
    slot_title->add_animation_object( new Renderer::OverlayColorFactorAnimation
        ( SColor( 4.0f, 4.0f, 4.0f, 4.0f ) ) );
    
    // Slot table.
    
    slots.resize( cols * rows );
    update_slot_screen_regions();
    }
    

void SlotPanel::change_slot_table_size
    (
    int the_columns,
    int the_rows
    )
    {
    hope( the_columns >= 0 );
    hope( the_rows >= 0 );
    
    hope( ( the_columns == 0 && the_rows == 0 ) || ( the_columns != 0 && the_rows != 0 ) );
    
    cols = the_columns;
    rows = the_rows;
    
    slots.resize( cols * rows );
    update_slot_screen_regions();
    }


void SlotPanel::set_slot_padding
    (
    Float32     horizontal_padding,
    Float32     vertical_padding
    )
    {
    slot_hpadding = horizontal_padding;
    slot_vpadding = vertical_padding;
    }


void SlotPanel::button_clicked
    (
    Button * button
    )
    {
    if( button == close_button.get() )
        {
        // Panel close request.
        if( message_receiver )
            {
            message_receiver->slot_panel_close( this );
            }
        }
    }


bool SlotPanel::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    // Because slots are not widgets and panel draging is handled
    // in super class, we will first try to handle the click by the slots.
    // If no slot is clicked, we will send the click to super class.

    if( button == SDL_BUTTON_LEFT )
        {
        // Check if clicked on some slot.
        int x, y;
        for( y = 0; y < rows; ++y )
            {
            for( x = 0; x < cols; ++x )
                {
                if( is_mouse_on_slot( x, y ) )
                    {
                    if( message_receiver )
                        {
                        int slot_index = x + y * cols;
                        hope( is_valid_slot_index( slot_index ) );
                        message_receiver->slot_panel_slot_clicked( this, slot_index );
                        }
                    
                    return true;
                    }
                }
            }
        }
        
    if( PanelBorder::mouse_button_down( button, mod ) )
        {
        return true;
        }
    
    return false;
    }


void SlotPanel::render()
    {
    // Render background and border.
    PanelBorder::render();
    
    // Render slots.
    if( size_dirty )
        {
        update_slot_screen_regions();
        }
    render_slots();
    
    // Update close button.
    if( color_dirty )
        {
        close_button->set_base_color( get_current_color() );
        }
    
    // Update title.
        {
        // Title is registered as child overlay object so its offset is relative
        // to the panel.

        if( size_dirty )
            {
            SVector2 title_size = title->get_current_size();
            SVector2 size = get_current_size();
            
            title->set_base_offset( SVector2
                ( ( size.x() - title_size.x() ) / 2.0f, 32.0f ) );
            }
        
        if( color_dirty )
            {
            title->set_base_color( get_current_color() );
            }
        }
    
    // Update slot_title and slot_title_text.
        {
        int x, y;
        bool found = false;
        
        for( y = 0; y < rows; ++y )
            {
            for( x = 0; x < cols; ++x )
                {
                if( is_mouse_on_slot( x, y ) )
                    {
                    found = true;
                    break;
                    }
                }

            if( found )
                {
                break;
                }
            }
        
        if( found )
            {
            int slot_index = x + y * cols;
            hope( is_valid_slot_index( slot_index ) );
            Slot & slot = get_slot( slot_index );
            
            // Update slot_title color animation.
            if( slot.get_title().empty() )
                {
                SColor target_color_factor = get_current_color();
                target_color_factor.set_alpha( 0.0f );
                slot_title->require_color_factor_animation().set_target_factor( target_color_factor );
                }
            else
                {
                SColor target_color_factor = get_current_color();
                slot_title->require_color_factor_animation().set_target_factor( target_color_factor );
                }
            
            if( slot_title_text != slot.get_title() && !slot.get_title().empty() )
                {
                slot_title->print_line( slot.get_title() );
                slot_title_text = slot.get_title();
                }
            }
        else
            {
            SColor target_color_factor = get_current_color();
            target_color_factor.set_alpha( 0.0f );
            slot_title->require_color_factor_animation().set_target_factor( target_color_factor );
            }
        }
    
    // Update slot_title.
    if( slot_title->is_visible() )
        {
        SVector2 title_size = slot_title->get_current_size();
        SVector2 size = get_current_size();
            
        slot_title->set_base_offset( SVector2
            ( ( size.x() - title_size.x() ) / 2.0f,
              ( size.y() - 32.0f ) ) );
        
        /*
        if( color_dirty )
            {
            slot_title->set_base_color( get_current_color() );
            }
        */
        }
    }


void SlotPanel::render_slots()
    {
    int x, y;
    for( y = 0; y < rows; ++y )
        {
        for( x = 0; x < cols; ++x )
            {
            render_slot( x, y );
            }
        }
    }


void SlotPanel::render_slot
    (
    int x,
    int y
    )
    {
    int slot_index = x + y * cols;
    hope( is_valid_slot_index( slot_index ) );
    Slot & slot = get_slot( slot_index );
    bool selected = ( selected_slot_index == slot_index );
    SVector2 slot_size = slot.screen_bottom_right - slot.screen_top_left;
    Renderer::Icon * icon = NULL;
    bool mouse_on = is_mouse_on_slot( x, y );
    
    // Draw background.
    icon = slot_background_icon.get();
    if( icon )
        {
        icon->set_base_color( get_current_color() );
        icon->set_base_offset( slot.screen_top_left );
        icon->set_base_size( slot_size );
        icon->render();
        }
    
    // Draw slot icon.
    icon = slot.icon.get();
    if( icon )
        {
        icon->set_base_color( get_current_color() );
        icon->set_base_offset( slot.screen_top_left );
        icon->set_base_size( slot_size );
        icon->render();
        }
    
    // Draw highlighting.
    icon = slot_highlight_icon.get();
    if( icon && ( slot.highlight || selected || mouse_on ) )
        {
        icon->set_base_color( get_current_color() );
        icon->set_base_offset( slot.screen_top_left );
        icon->set_base_size( slot_size );
        icon->render();
        }
    
    // Draw slot text.
    if( slot.text_rectangle.get() )
        {
        Renderer::TextRectangle * tr = slot.text_rectangle.get();
        tr->set_size_from_font(); // Restore original size.
        Float32 max_w = slot_size.x() + slot_hpadding;
        Float32 w = tr->get_current_size().x();
        Float32 x;
        if( w > max_w )
            {
            tr->set_base_size( SVector2( max_w, slot_vpadding ) );
            x = slot_size.x() / 2.0f - max_w / 2.0f;
            }
        else
            {
            tr->set_base_size( SVector2( w, slot_vpadding ) );
            x = slot_size.x() / 2.0f - w / 2.0f;
            }
        SVector2 delta( x, slot_size.y() );
        
        tr->set_base_color( get_current_color() );
        tr->set_base_offset( slot.screen_top_left + delta );
        tr->render();
        }
    }


bool SlotPanel::is_mouse_on_slot
    (
    int x,
    int y
    ) const
    {
    int slot_index = x + y * cols;
    hope( is_valid_slot_index( slot_index ) );
    const Slot & slot = get_slot( slot_index );
    InputManager & input_manager = Global::get_input_manager();
    SVector2 mouse_pos = input_manager.get_mouse_position();
    
    if( mouse_pos.x() < slot.screen_top_left.x() ||
        mouse_pos.y() < slot.screen_top_left.y() ||
        mouse_pos.x() >= slot.screen_bottom_right.x() ||
        mouse_pos.y() >= slot.screen_bottom_right.y() )
        {
        return false;
        }
    else
        {
        return true;
        }
    }


void SlotPanel::get_working_screen_region
    (
    SVector2 & top_left,
    SVector2 & bottom_right
    ) const
    {
    SVector2 size = get_current_size();
    SVector2 screen_top_left = screenspace_translation().svector();
    Float32  margin_width = 32.0f;
    Float32  margin_height = 32.0f;
    
    top_left = SVector2
        (
        screen_top_left.x() + margin_width,
        screen_top_left.y() + margin_height
        );
    
    bottom_right = SVector2
        (
        screen_top_left.x() + size.x() - margin_width,
        screen_top_left.y() + size.y() - margin_height
        );
    }


void SlotPanel::get_slot_table_screen_region
    (
    SVector2 & top_left,
    SVector2 & bottom_right
    ) const
    {
    get_working_screen_region( top_left, bottom_right );
    
    // Some place for panel title.
    top_left.set_y( top_left.y() + 32.0f );
    
    // Remove extra slot padding.
    bottom_right.set_x( bottom_right.x() + slot_hpadding );
    bottom_right.set_y( bottom_right.y() + slot_vpadding );
    }


void SlotPanel::update_slot_screen_regions()
    {
    int         x, y;
    SVector2    slot_table_top_left;
    SVector2    slot_table_bottom_right;
    Float32     slot_table_width;
    Float32     slot_table_height;
    Float32     slot_width;
    Float32     slot_height;
    
    if( cols <= 0 || rows <= 0 )
        {
        hope( slots.empty() );
        return;
        }

    // Get slot table region.
    get_slot_table_screen_region( slot_table_top_left, slot_table_bottom_right );
    
    // Compute slot table size.
    slot_table_width = slot_table_bottom_right.x() - slot_table_top_left.x();
    slot_table_height = slot_table_bottom_right.y() - slot_table_top_left.y();
    
    // Compute slot size.
    slot_width = slot_table_width / static_cast< Float32 >( cols );
    slot_height = slot_table_height / static_cast< Float32 >( rows );
    
    // Compute screen region for each slot...
    for( y = 0; y < rows; ++y )
        {
        for( x = 0; x < cols; ++x )
            {
            int slot_index = x + y * cols;
            Slot & slot = get_slot( slot_index );
            
            Float32 xf = static_cast< Float32 >( x );
            Float32 yf = static_cast< Float32 >( y );
            
            slot.screen_top_left = SVector2
                (
                slot_table_top_left.x() + xf * slot_width,
                slot_table_top_left.y() + yf * slot_height
                );
            
            slot.screen_bottom_right = SVector2
                (
                slot_table_top_left.x() + ( xf + 1.0f ) * slot_width - slot_hpadding,
                slot_table_top_left.y() + ( yf + 1.0f ) * slot_height - slot_vpadding
                );
            }
        }
    }
    
    
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "gui/slot_panel_inline.h"
#endif
