/** @file slot_panel.h
 * Slot panel.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_SLOT_PANEL_H
#define CLIENT_SLOT_PANEL_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif

#ifndef CLIENT_ICON_BUTTON_H
#include "gui/icon_button.h"
#endif

#ifndef CLIENT_PANEL_BORDER_H
#include "gui/panel_border.h"
#endif

#ifndef CLIENT_TEXT_RECTANGLE_H
#include "renderer/text_rectangle.h"
#endif


namespace Demo {
namespace Client {


/**
 * Slot panel.
 * Slot panels contains slots. Number of slots is specified in
 * constructor.
 * Each slot can contain icon which is drawn when the slot is beeing
 * rendered.
 * One slot can be selected. Selected slot is highlighted during drawing.
 */
class SlotPanel : public PanelBorder, public Button::ButtonMessageReceiver
    {
    NO_COPY_CONSTRUCTOR( SlotPanel );
    NO_ASSIGNMENT_OPERATOR( SlotPanel );

public:

    /**
     * Receiver of messages from the panel.
     */
    class SlotPanelMessageReceiver
        {
    public:

        /**
         * Destructor.
         */
        virtual ~SlotPanelMessageReceiver()
            {
            }

        /**
         * Called when user clicks close button on the panel.
         */
        virtual void slot_panel_close
            (
            SlotPanel * panel
            )
            {
            }
        
        /**
         * Called when user clicks a slot on the panel.
         */
        virtual void slot_panel_slot_clicked
            (
            SlotPanel * panel,
            int         slot_index
            )
            {
            }

        }; // SlotPanelMessageReceiver

public:

    /**
     * Structure holding miscenalleous data of one slot.
     */
    class Slot
        {
        friend class SlotPanel;

    public:
        /** Constructor. */
        INLINE Slot();
        
        /** Destructor. */
        INLINE ~Slot();

        /** Copy constructor. */
        INLINE Slot
            (
            const Slot & other
            );
        
        /** Copy constructor. */
        INLINE Slot & operator=
            (
            const Slot & other
            );

    public:
    
        /** Get slot text. */
        INLINE const std::string & get_text() const;
        
        /** Set slot text. */
        void set_text
            (
            const std::string & the_text
            );
        
        /** Get slot title. */
        INLINE const std::string & get_title() const;
        
        /** Set slot title. */
        INLINE void set_title
            (
            const std::string & the_title
            );
        
        /** Clear slot content. */
        INLINE void clear();

    public:
        mutable std::auto_ptr< Renderer::Icon >     icon;
            /**< Image to draw at the slot. May be @c NULL. */
        
        std::string                                 icon_name;
            /**< Icon template name. */
        
        bool                                        highlight;
            /**< Highlight the slot. */
    
    private:
        mutable std::auto_ptr< Renderer::TextRectangle >    text_rectangle;
            /**< Text to draw below the slot. */
        
        std::string                                         text;
            /**< Text inside @a text_rectangle. */
        
        std::string                                         title;
            /**< Slot title. Visible if mouse is on the slot. */

    public:
        SVector2    screen_top_left;
            /**< Top left corner position of the slot on screen. */

        SVector2    screen_bottom_right;
            /**< Bottom right corner position of the slot on screen. */

        }; // Slot

public:
    
    /**
     * Constructor.
     */
    SlotPanel
        (
        const std::string & title
        );
    
    /**
     * Change slot table size.
     */
    void change_slot_table_size
        (
        int columns,
        int rows
        );
    
    /**
     * Set slot padding.
     */
    void set_slot_padding
        (
        Float32     horizontal_padding,
        Float32     vertical_padding
        );
    
public:

    /**
     * Get number of columns.
     */
    INLINE int get_slot_columns() const;
    
    /**
     * Get number of rows.
     */
    INLINE int get_slot_rows() const;

    /**
     * Mark slot at @a slot_index selected. If there is another slot
     * selected, it is deselected first.
     * If @a slot_index is not valid slot index, only currently selected
     * slot is deselected.
     * @param slot_index Index of slot to select. Pass < 0 to deselect
     *      currently selected slot.
     */
    INLINE void select_slot
        (
        int slot_index
        );
    
    /**
     * Get index of currently selected slot.
     * @return < 0 if no slot is selected or valid slot index.
     */
    INLINE int get_selected_slot_index() const;
    
public:

    /**
     * Get reference to slot at @a slot_index.
     * @param slot_index Must be valid slot index.
     */
    INLINE const Slot & get_slot
        (
        int slot_index
        ) const;

    /**
     * Get reference to slot at @a slot_index.
     * @param slot_index Must be valid slot index.
     */
    INLINE Slot & get_slot
        (
        int slot_index
        );
        
    /**
     * Get pointer to slot at @a slot_index.
     * If @a slot_index is not valid slot index, @c NULL is returned.
     */
    INLINE Slot * get_slot_ptr
        (
        int slot_index
        );

    /**
     * Get pointer to slot at @a slot_index.
     * If @a slot_index is not valid slot index, @c NULL is returned.
     */
    INLINE const Slot * get_slot_ptr
        (
        int slot_index
        ) const;

public:

    /**
     * Check whether @a slot_index is valid slot index.
     */
    INLINE bool is_valid_slot_index
        (
        int slot_index
        ) const;

public:

    /**
     * Set #SlotPanelMessageReceiver of the panel.
     * If @a message_receiver is @c NULL, message notifying will be
     * disabled.
     */
    INLINE void set_message_receiver
        (
        SlotPanelMessageReceiver * message_receiver
        );

public:

    // IconButton
    virtual void button_clicked
        (
        Button * button
        );

    // InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );

    // OverlayObject
    virtual void render();

private:

    /**
     * Render all slots.
     */
    void render_slots();
    
    /**
     * Render one slot.
     */
    void render_slot
        (
        int x,
        int y
        );
    
    /**
     * Check whether mouse cursor is on top of specified slot.
     */
    bool is_mouse_on_slot
        (
        int x,
        int y
        ) const;

    /**
     * Get region of the panel on screen excluding border area of the panel.
     */
    void get_working_screen_region
        (
        SVector2 & top_left,
        SVector2 & bottom_right
        ) const;
    
    /**
     * Get region of area occupied by slots on screen.
     */
    void get_slot_table_screen_region
        (
        SVector2 & top_left,
        SVector2 & bottom_right
        ) const;
    
    /**
     * Update screen regions of all slots.
     */
    void update_slot_screen_regions();
    
private:

    /** Array of slots. */
    typedef std::vector< Slot > Slots;

private:

    std::auto_ptr< Renderer::TextRectangle > title;
        /**< Panel title. */

    int         rows;
        /**< Number of slot rows. Always >= 1. */

    int         cols;
        /**< Number of slot columns. Always >= 1. */
    
    Slots       slots;
        /**< Array of slots. */

    int         selected_slot_index;
        /**< Index of selected slot. Either < 0 if no slot is selected
         * or valid slot index.
         */
    
    Float32     slot_hpadding;
        /**< Horizontal padding between two slots. */
    
    Float32     slot_vpadding;
        /**< Vertical padding between two slots. */
    
    std::auto_ptr< Renderer::Icon >             slot_background_icon;
        /**< Image used to draw slot backgrounds. */

    std::auto_ptr< Renderer::Icon >             slot_highlight_icon;
        /**< Image used to draw highlighted slots. */

    std::auto_ptr< IconButton >                 close_button;
        /**< Close button of the panel. */
    
    std::auto_ptr< Renderer::TextRectangle >    slot_title;
        /**< Visible title of slot under mouse cursor. */
    
    std::string                                 slot_title_text;
        /**< Text stored inside @a slot_title. */
    
private:

    SlotPanelMessageReceiver *   message_receiver;
        /**< Message receiver for messages from the slot panel. */

    }; // SlotPanel


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "gui/slot_panel_inline.h"
#endif


#endif // CLIENT_SLOT_PANEL_H
