/** @file widget.cpp
 * Widget (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/widget.h"
#include "client_global.h"
#include "gui/widget_input_dispatcher.h"
#include "input/input_manager.h"
#include <algorithm>


namespace Demo {
namespace Client {


Widget::Widget
    (
    Type            widget_type,
    bool            is_widget_input_dispatcher
    ) :
    type( widget_type ),
    behavior( DEFAULT_BEHAVIOR ),
    parent( NULL ),
    dispatcher( is_widget_input_dispatcher ),
    dragable( false ),
    auto_bringable_to_front( false ),
    draging( false )
    {
    }

    
Widget::~Widget()
    {
    // Unlink self from parent widget.
    set_parent_widget( NULL );

    // Unlink all child widgets.
    while( !children.empty() )
        {
        Widget * child = children[ 0 ];
        hope( child );
        child->set_parent_widget( NULL );
        }
    }


void Widget::do_layout()
    {
    // TODO: Move this function to OverlayObject? Without it
    // animating size of complex widgets is not possible.
    }


SVector2 Widget::get_top_left_corner_on_screen() const
    {
    const OpenGL::Vector2 top_left = screenspace_translation();
    return top_left.svector();
    }


SVector2 Widget::get_bottom_right_corner_on_screen() const
    {
    const SVector2  top_left = get_top_left_corner_on_screen();
    const SVector2  size( get_current_size() );
    return top_left + size;
    }


bool Widget::contains_screen_point
    (
    const SVector2 & point
    ) const
    {
    // Note: Visibility is ignored by this function.

    bool inside = false;
    
    switch( type )
        {
        case SOLID_WIDGET:
            {
            const SVector2  top_left = get_top_left_corner_on_screen();
            const SVector2  bottom_right = get_bottom_right_corner_on_screen();
    
            // Check if the widget itself contains the point.
            if( point.x() < top_left.x() ||
                point.x() >= bottom_right.x() ||
                point.y() < top_left.y() ||
                point.y() >= bottom_right.y() )
                {
                inside = false;
                }
            else
                {
                inside = true;
                }
            
            break;
            }
        
        case TRANSPARENT_WIDGET:
            inside = false;
            break;
        }
    
    return inside;
    }


bool Widget::is_mouse_on() const
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        return dispatcher->is_mouse_on( this );
        }
    else
        {
        InputManager & input_manager = Global::get_input_manager();
        return contains_screen_point( input_manager.get_mouse_position() );
        }
    }


void Widget::set_type
    (
    Type new_type
    )
    {
    type = new_type;
    }


void Widget::set_parent_widget
    (
    Widget *    new_parent
    )
    {
    // Set parent overlay object.
    set_parent_overlay_object( new_parent );
    
    // Unlink from current parent.
    if( parent )
        {
        // If this widget or some of its child widgets has focus, disable the focus
        // so our widget dispatcher will not hold pointer to the widget
        // with focus (because maybe the widget is going to be destroyed asap).
        disable_focus();
        release_mouse_button_grab( LEFT_MOUSE_BUTTON_GRAB );
        release_mouse_button_grab( RIGHT_MOUSE_BUTTON_GRAB );
        parent->unlink_child_widget( this );
        parent = NULL;
        }

    // Link to new parent.
    if( new_parent )
        {
        new_parent->link_child_widget( this );
        parent = new_parent;
        }
    }


void Widget::set_dispatcher_as_parent()
    {
    Widget * const parent_widget = &Global::get_widget_input_dispatcher();
    hope( parent_widget );
    set_parent_widget( parent_widget );
    }


bool Widget::is_my_child_widget
    (
    Widget * widget
    ) const
    {
    if( widget == this )
        {
        return false;
        }

    while( widget && widget != this )
        {
        widget = widget->parent;
        }
    return ( widget == this );
    }


Widget * Widget::find_parent
    (
    const Widget * superparent
    )
    {
    Widget * widget = this;
    while( widget && widget->parent != superparent )
        {
        widget = widget->parent;
        }
    return widget;
    }


bool Widget::has_focus() const
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        return dispatcher->has_focus( this );
        }
    else
        {
        return false;
        }
    }


bool Widget::has_my_child_focus() const
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        Widget * focus_widget = dispatcher->get_focus_widget();
        if( focus_widget )
            {
            return is_my_child_widget( focus_widget );
            }
        else
            {
            return false;
            }
        }
    else
        {
        return false;
        }
    }


void Widget::set_focus()
    {
    hope( is_focusable() );
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        dispatcher->set_focus( this );
        }
    }


void Widget::disable_focus()
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        dispatcher->disable_focus( this );
        }
    }


void Widget::set_focusable
    (
    int the_focusable
    )
    {
    hope( Massiv::Core::flags_on( the_focusable, 0, ~FOCUSABLE_MASK ) );
    behavior = ( behavior & ~FOCUSABLE_MASK ) | the_focusable;
    }


bool Widget::mouse_button_grabbed
    (
    MouseButtonGrab grab
    ) const
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        return dispatcher->mouse_button_grabbed( this, grab );
        }
    else
        {
        return false;
        }
    }

    
void Widget::release_mouse_button_grab
    (
    MouseButtonGrab grab
    )
    {
    WidgetInputDispatcher * dispatcher = get_my_dispatcher();
    hope( dispatcher != this );
    if( dispatcher )
        {
        dispatcher->release_mouse_button_grab( this, grab );
        }
    }


void Widget::enable()
    {
    enable_input();
    }


void Widget::disable()
    {
    disable_input();
    }


bool Widget::is_enabled() const
    {
    return is_input_enabled();
    }


bool Widget::is_enabled_from_parents() const
    {
    const Widget * w;
    for( w = this; w; w = w->parent )
        {
        if( !w->is_enabled() )
            {
            return false;
            }
        }
    return true;
    }


bool Widget::is_visible_from_parents() const
    {
    const Widget * w;
    for( w = this; w; w = w->parent )
        {
        if( !w->is_visible() )
            {
            return false;
            }
        }
    return true;
    }


void Widget::set_dragable
    (
    bool drag
    )
    {
    dragable = drag;
    }
    

bool Widget::is_dragable() const
    {
    return dragable;
    }


void Widget::set_auto_bringable_to_front
    (
    bool bringable
    )
    {
    auto_bringable_to_front = bringable;
    }


bool Widget::is_auto_bringable_to_front() const
    {
    return auto_bringable_to_front;
    }


void Widget::bring_to_front()
    {
    Widget * parent_widget = get_parent();
    
    if( !parent_widget )
        {
        return;
        }
    
    // Relink parent overlay object.
    set_parent_overlay_object( NULL );
    set_parent_overlay_object( parent_widget );
    
    // Relink self.
    parent_widget->unlink_child_widget( this );
    parent = NULL;
    parent_widget->link_child_widget( this );
    parent = parent_widget;
    }


void Widget::mouse_button_grab_lost
    (
    MouseButtonGrab grab
    )
    {
    if( grab == LEFT_MOUSE_BUTTON_GRAB )
        {
        draging = false;
        }
    }


bool Widget::mouse_button_down
    (
    int     button,
    SDLMod  mod
    )
    {
    if( InputObject::mouse_button_down( button, mod ) )
        {
        return true;
        }
    
    if( !is_dragable() )
        {
        return false;
        }
    
    if( button == SDL_BUTTON_LEFT )
        {
        SVector2 drag_begin_mouse_position = Global::get_input_manager().get_mouse_position();
        drag_delta = drag_begin_mouse_position - get_top_left_corner_on_screen();
        draging = true;
        return true;
        }
        
    return false;
    }
    
    
bool Widget::mouse_move
    (
    int    dx,
    int    dy,
    SDLMod mod
    )
    {
    if( is_dragable() && mouse_button_grabbed( LEFT_MOUSE_BUTTON_GRAB ) && draging )
        {
        // We use current mouse position instead of dx and dy because
        // mouse move events are sent even if the cursor is not moving (if the
        // cursor is at border of screen) so our widget will not move outside
        // screen.
        SVector2 mouse_position = Global::get_input_manager().get_mouse_position();
        SVector2 target_screen_position = mouse_position - drag_delta;
        SVector2 offset = inverse_screenspace_translation( target_screen_position );
        set_base_offset( offset );
        
        return true;
        }

    return false;
    }


WidgetInputDispatcher * Widget::get_my_dispatcher() const
    {
    Widget * w = NULL;
    
    // Find first widget input dispatcher.
    for( w = parent; w; w = w->parent )
        {
        if( w->is_dispatcher() )
            {
            break;
            }
        }
    
    if( w )
        {
        WidgetInputDispatcher * dispatcher;
        dispatcher = dynamic_cast< WidgetInputDispatcher * >( w );
        hope( dispatcher );
        return dispatcher;
        }
    else
        {
        return NULL;
        }
    }


bool Widget::is_dispatcher() const
    {
    return dispatcher;
    }


void Widget::render()
    {
    }


void Widget::link_child_widget
    (
    Widget * child
    )
    {
    hope( child );
    hope( !child->get_parent() );
    hope( std::find( children.begin(), children.end(), child ) == children.end() );
    
    children.push_back( child );
    }
    

void Widget::unlink_child_widget
    (
    Widget * child
    )
    {
    hope( child );
    hope( child->get_parent() == this );
    
    Children::iterator child_it = std::find( children.begin(), children.end(), child );
    hope( child_it != children.end() );
    
    children.erase( child_it );
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "gui/widget_inline.h"
#endif
