/** @file widget.h
 * Widget.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_WIDGET_H
#define CLIENT_WIDGET_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_GUI_FWD_H
#include "gui/gui_fwd.h"
#endif

#ifndef CLIENT_OVERLAY_OBJECT_H
#include "renderer/overlay_object.h"
#endif

#ifndef CLIENT_INPUT_OBJECT_H
#include "input/input_object.h"
#endif

#include <vector>


namespace Demo {
namespace Client {
    

/**
 * Widget.
 * Widget is #InputObject and #OverlayObject in one.
 * Widget is able to receive input events, occupies some (by default
 * rectangular) area on screen and is renderable.
 *
 * Widget can have child widgets. Use @c set_parent_widget() method
 * to set child/parent relationship.
 * Child widget can have only one parent. Parent widget can have
 * unlimited number of child widgets.
 * Area occupied by child widget on screen is not limited by area
 * of its parent. Child widget may be outside its parent on screen.
 *
 * Child/parent relationship is used when input event is dispatched
 * by #WidgetInputDispatcher. See info below.
 * Parent widgets are rendered before rendering of their child widgets,
 * because child widgets are automatically registered as child overlay objects
 * of their parent widget. See #OverlayObject for more info.
 *
 * Widget itself should not be registered as input object in #InputManager
 * because #InputManager is not able to dispatch input events
 * to its input objects depending on their screen positions and widget
 * children.
 * Use #WidgetInputDispatcher instead to dispatch input events
 * to widgets depending on their screen positions and widget children.
 *
 * See #WidgetInputDispatcher how input events are dispatched
 * when a widget have children widgets.
 *
 * If widget is (not necessary direct) child of #WidgetInputDispatcher,
 * it may have set keyboard focus. Widget with keyboard focus
 * will receive keybord events directly as the event arrives
 * into the #WidgetInputDispatcher.
 *
 * @warning Widgets may not be destroyed, linked or unlinked from their
 *      parents during input event processing!
 */
class Widget : public InputObject, public Renderer::OverlayObject
    {
    NO_COPY_CONSTRUCTOR( Widget );
    NO_ASSIGNMENT_OPERATOR( Widget );

    friend class WidgetInputDispatcher; // Access to children.
    
public:
    
    /**
     * Widget type.
     */
    enum Type
        {
        SOLID_WIDGET,
            /**< Solid widget. See @c contains_screen_point() for more info. */
            
        TRANSPARENT_WIDGET,
            /**< Transparent widget. See @c contains_screen_point() for more info.
             * @note Transparent widget stays renderable. Transparency in this
             *      context only means that the widget is intended only
             *      as holder for child widgets and it will not block mouse input
             *      events (see #WidgetInputDispatcher for more info).
             */
             
        DEFAULT_TYPE = SOLID_WIDGET
            /**< Default widget type. */

        }; // Type
    
    /** Widget behavior flags. */
    enum BehaviorFlags
        {
        FOCUSABLE           = 0x01,
            /**< The widget is focusable. */

        TAB_FOCUSABLE       = 0x02 | FOCUSABLE,
            /**< Can gain focus when user presses TAB key (TAB key iterates through visible
                 tab-focusable widgets). */

        CLICK_FOCUSABLE     = 0x04 | FOCUSABLE,
            /**< Can gain focus when user clicks the widget. */

        WINDOW_FOCUSABLE    = 0x08 | CLICK_FOCUSABLE,
            /**< When the widget should gain focus, it will try to find focusable
                 child and focus it instead. If there is no such child, the widget
                 will gain focus. All WINDOW_FOCUSABLE widgets are automatically
                 CLICK_FOCUSABLE. */

        CLICK_LOSE_FOCUS    = 0x10 | CLICK_FOCUSABLE,
            /**< When clicked, the widget disables focus from currently focused
                 widget but does not gain it itself. */

        TAB_CLICK_FOCUSABLE = TAB_FOCUSABLE | CLICK_FOCUSABLE,
            /**< Can gain focus by tabs and clicks. */

        FOCUSABLE_MASK      = TAB_FOCUSABLE | CLICK_FOCUSABLE |
                              WINDOW_FOCUSABLE | CLICK_LOSE_FOCUS,
            /**< All "focusable" bits. */

        DEFAULT_BEHAVIOR    = 0
            /**< Default behavior flag combination. */

        }; // BehaviorFlags

    /** Mouse button grab. */
    enum MouseButtonGrab
        {
        LEFT_MOUSE_BUTTON_GRAB,
            /**< If user clicks left mouse on the widget, the widget automatically
             * grabs mouse until the user release the button. While in grab mode,
             * all mouse move input events are send directly to the widget.
             */
        RIGHT_MOUSE_BUTTON_GRAB
            /**< If user clicks right mouse on the widget, the widget automatically
             * grabs mouse until the user release the button. While in grab mode,
             * all mouse move input events are send directly to the widget.
             */
        }; // MouseButtonGrab

public:

    /**
     * Constructor.
     * @param is_widget_input_dispatcher Set to true only and only if
     *      the widget is convertible to #WidgetInputDispatcher.
     */
    Widget
        (
        Type    widget_type = DEFAULT_TYPE,
        bool    is_widget_input_dispatcher = false
        );
    
    /**
     * Destructor.
     * The widget unlinks self from its parent widget
     * and unlinks all child widgets.
     *
     * @warning Do not destroy widgets while input event processing
     *      is in progress! See @c set_parent_widget().
     */
    virtual ~Widget();
    
public:

    /**
     * Arrange self and child widgets.
     * Main purpose of this method is tell the widget that
     * its size has changed and it should rearrange its child widgets
     * based on the new size.
     *
     * Everytime changing size of a widget, @c do_layout() should be
     * called asap so the widget is able to rearrange its child widgets.
     *
     * Typical do_layout() method looks like this code:
     * @code
     * void MyWidget::do_layout()
     *     {
     *     SuperClass::do_layout();
     *     child_widget->set_base_offset( SVector2( x, y ) );
     *     child_widget->set_base_size( SVector2( width, height ) );
     *     child_widget->do_layout();
     *     }
     * @endcode
     *
     * There are two types of widgets - widgets that arranges self
     * to fit in given size and widgets that resize self based on
     * arrangement of its child widgets.
     * The widgets that resize self based on child widgets can call
     * @c set_base_size() from @c do_layout() and caller don't need to
     * set any size before calling @c do_layout().
     * The widgets of the first type on the other hand will only
     * rearrange self and its child widgets to fit to current size
     * so the caller should set size of these widgets before
     * calling @c do_layout().
     * See documentation of each widget to check what is the type the widget.
     */
    virtual void do_layout();
    
public:

    /**
     * Get current top left corner of the widget's area on screen.
     * Returned point is relative to top left corner of screen.
     *
     * Shortcut for <tt>OverlayObject::screenspace_translation()</tt>.
     */
    SVector2 get_top_left_corner_on_screen() const;

    /**
     * Get current bottom right corner of the widget's area on screen.
     * Returned point is relative to top left corner of screen.
     *
     * Shortcut for <tt>OverlayObject::screenspace_translation()
     * + OverlayObject::get_current_size()</tt>.
     */
    SVector2 get_bottom_right_corner_on_screen() const;
    
    /**
     * Check whether the widget contains @a point in screen coordinates.
     *
     * If type of the widget is @c SOLID_WIDGET, the point is tested against
     * current rectangle occupied by the overlay object.
     *
     * If type of the widget is @c TRANSPARENT_WIDGET, the function
     * returns false.
     *
     * In both cases, areas of child widgets are ignored by this function.
     *
     * @param point Point in screen coordinates relative to upper left
     *      corner of the screen.
     * @note Ancestors can redefine this function to support non-rectangular
     *      shapes or holes in the widget etc.
     * @warning The test is done no matter whether the widget is
     *      visible or whether it is overlapped by other (non-child) widgets etc.
     */
    virtual bool contains_screen_point
        (
        const SVector2 & point
        ) const;

    /**
     * Check whether mouse is above the widget.
     */
    bool is_mouse_on() const;

public:

    /**
     * Sets new type of widget.
     * @param Type new type of the widget.
     */
    void set_type
        (
        Type    new_type
        );

    /**
     * Link the widget to @a parent widget as its child.
     * The widget is also automatically linked as child overlay object
     * of the parent.
     * @param parent Parent widget. Pass @a NULL to unlink the widget
     *      from its current parent.
     * @warning Do not link/unlink widgets while input event processing
     *      is in progress!
     */
    void set_parent_widget
        (
        Widget *    parent
        );

    /**
     * Make widget "standalone window", ie. set input dispatcher as its parent.
     * Shortcut for <tt>set_parent_widget( &Global::get_widget_input_dispatcher() );</tt>.
     */
    void set_dispatcher_as_parent();

    /**
     * Check whether @a widget is (not necessary direct) child of this widget.
     */
    bool is_my_child_widget
        (
        Widget * widget
        ) const;
    
    /** 
     * Get parent widget.
     * @return Current parent widget or @c NULL.
     */
    INLINE Widget * get_parent() const;

    /**
     * Find (grand)parent of this widget that is child of @a superparent
     * Use @c NULL @a superparent to find greatest grandparent.
     *
     * @return Pointer to widget that is ancestor of @a this and direct child of
     *         @a superparent or @c NULL if no such widget exists.
     */
    Widget * find_parent
        (
        const Widget * superparent
        );
    
    /**
     * Is the widget focusable?
     */
    INLINE bool is_focusable() const;

    /**
     * Is the widget tab-focusable?
     */
    INLINE bool is_tab_focusable() const;

    /**
     * Is the widget click-focusable?
     */
    INLINE bool is_click_focusable() const;

    /**
     * Is the widget window-focusable?
     */
    INLINE bool is_window_focusable() const;

    /**
     * Disable focus on click?
     */
    INLINE bool is_click_lose_focus() const;

    /**
     * Check whether the widget has keyboard focus.
     */
    bool has_focus() const;
    
    /**
     * Check whether some of child widgets has focus.
     */
    bool has_my_child_focus() const;

    /**
     * Set keyboard focus to this widget.
     * The widget should be (not necessary direct) child of some
     * #WidgetInputDispatcher, otherwise nothing is done
     * and @c set_focus() has no effect.
     */
    void set_focus();
    
    /**
     * Disable keybord focus from this widget or its child widgets.
     * The widget should be (not necessary direct) child of some
     * #WidgetInputDispatcher, otherwise nothing is done
     * and @c disable_focus() has no effect.
     */
    void disable_focus();
    
    /**
     * Change focusability of the widget.
     *
     * @param focusable Combination of *FOCUSABLE BehaviorFlags.
     */
    void set_focusable
        (
        int focusable
        );
    
    /**
     * Check whether the widget has grabbed mouse button.
     * While mouse button is grabbed, the widget will receive
     * all mouse move events until corresponding mouse buton is release.
     */
    bool mouse_button_grabbed
        (
        MouseButtonGrab grab
        ) const;
    
    /**
     * Release mouse button grab from this widget or its child widgets.
     */
    void release_mouse_button_grab
        (
        MouseButtonGrab grab
        );

public:

    /**
     * Enable the widget.
     * Disabled widgets won't receive input events.
     */
    void enable();
    
    /**
     * Disable the widget.
     */
    void disable();
    
    /**
     * Check whether the widget is enabled.
     */
    bool is_enabled() const;
    
    /**
     * Check if all widgets above the widget (including the widget itself)
     * in parent/child hiearchy are enabled.
     */
    bool is_enabled_from_parents() const;
    
    /**
     * Check if all widgets above the widget (including the widget itself)
     * in parent/child hiearchy are visible.
     */
    bool is_visible_from_parents() const;
    
public:

    /**
     * Change dragability of the widget.
     * If the widget is made dragable, the widget can be dragged by
     * pressing left mouse button and moving the mouse.
     */
    void set_dragable
        (
        bool dragable
        );
    
    /**
     * Check whether the widget is dragable.
     */
    bool is_dragable() const;
    
    /**
     * Change auto-bringable-to-front feature of the widget.
     * If the widget is made auto bringable to front, the widget
     * is automatically placed above all other widgets if user
     * click the widget with mouse.
     */
    void set_auto_bringable_to_front
        (
        bool bringable
        );
    
    /**
     * Check whether the widget has auto-bringable-to-front feature on.
     */
    bool is_auto_bringable_to_front() const;
    
    /**
     * Place the widget above all other widgets.
     */
    void bring_to_front();
    
public:

    /**
     * Callback that is invoked when the widget is losing mouse button grab.
     */
    virtual void mouse_button_grab_lost
        (
        MouseButtonGrab grab
        );
        
    // InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );
        
    // InputObject
    virtual bool mouse_move
        (
        int    dx,
        int    dy,
        SDLMod mod
        );

public:
    
    // OverlayObject
    // Empty function.
    virtual void render();

private:

    /**
     * Get #WidgetInputDispatcher this widget is (not necessary direct) child of.
     */
    WidgetInputDispatcher * get_my_dispatcher() const;
    
    /**
     * Check whether this widget id #WidgetInputDispatcher.
     * If returns true, the widget is convertible to #WidgetInputDispatcher class.
     */
    bool is_dispatcher() const;

private:

    /**
     * Link @a child widget.
     */
    void link_child_widget
        (
        Widget * child
        );
    
    /**
     * Unlink @a child widget.
     */
    void unlink_child_widget
        (
        Widget * child
        );

private:

    /**
     * Array of aliasing pointers to child widgets.
     */
    typedef std::vector< Widget * > Children;

private:
    
    Type            type;
        /**< Widget type. */
    
    int             behavior;
        /**< Combination of BehaviorFlags flags. */

    Widget *        parent;
        /**< Aliasing pointer to parent widget. */
    
    Children        children;
        /**< Array of aliasing pointers to child widgets. */
    
    bool            dispatcher;
        /**< True whether the widget is convertible to #WidgetInputDispatcher. */
    
    bool            dragable;
        /**< True if the widget is dragable. */
        
    bool            auto_bringable_to_front;
        /**< True if auto-bring-to-front feature is enabled. */
    
    bool            draging;
        /**< True if the widget is beeing draged. */
    
    SVector2        drag_delta;
        /**< Delta between cursor position and top left corner of the widget
         * when dragging.
         */

    }; // Widget


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "gui/widget_inline.h"
#endif


#endif // CLIENT_WIDGET_H
