/** @file widget_input_dispatcher.cpp
 * Widget input dispatcher (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/widget_input_dispatcher.h"
#include "gui/popup_menu.h"
#include "client_global.h"
#include "input/input_manager.h"
#include <vector>
#include <queue>
#include <algorithm>


namespace Demo {
namespace Client {


WidgetInputDispatcher::WidgetInputDispatcher() :
    Super( TRANSPARENT_WIDGET, true ),
    focus_widget( NULL ),
    left_mouse_button_grab( false ),
    left_mouse_button_grab_widget( NULL ),
    right_mouse_button_grab( false ),
    right_mouse_button_grab_widget( NULL )
    {
    set_base_size_from_screen();
    }


WidgetInputDispatcher::~WidgetInputDispatcher()
    {
    // Because dispatcher is used inside ~Widget() to unlink
    // child widgets, we must unlink all child widgets right now
    // so Widget's destructor will not try to unlink any
    // child widget and will not try to access WidgetInputDispatcher
    // part of the object (which is not available inside ~Widget() any more).
    while( !children.empty() )
        {
        Widget * w = children.front();
        w->set_parent_widget( NULL );
        }
    }


//-----------------------------------------------------------------------------


bool WidgetInputDispatcher::is_mouse_on
    (
    const Widget * widget
    )
    {
    if( !widget )
        {
        return false;
        }
    
    return get_widget_under_mouse() == widget;
    }


bool WidgetInputDispatcher::has_focus
    (
    const Widget * widget
    )
    {
    hope( widget );
    return widget == focus_widget;
    }


void WidgetInputDispatcher::set_focus
    (
    Widget * widget
    )
    {
    hope( widget );
    hope( is_my_child_widget( widget ) );

    // If this widget is WINDOW_FOCUSABLE, try to pass focus to its child.

    Widget * widget_to_focus = widget;
    if( Massiv::Core::flags_on( widget->behavior, WINDOW_FOCUSABLE ) )
        {
        Widgets widgets;
        enumerate_focusable_widgets( widgets, FOCUSABLE, widget );

        // If focus_widget is child of widget, do not change focus.

        if( focus_widget != widget )
            {
            Widgets::iterator it = std::find( widgets.begin(), widgets.end(), focus_widget );
            if( it != widgets.end() )
                {
                // Child of widget is already focused - do not change focus.
                return;
                }
            }
        Widgets::iterator it;
        for( it = widgets.begin(); it != widgets.end(); ++it )
            {
            if( *it != widget )
                {
                widget_to_focus = *it;
                break;
                }
            }
        }

    hope( widget_to_focus );
    hope( widget_to_focus->is_focusable() );
    focus_widget = widget_to_focus;
    }


void WidgetInputDispatcher::disable_focus
    (
    Widget * widget
    )
    {
    hope( widget );
    hope( is_my_child_widget( widget ) );
    
    if( focus_widget && ( widget == focus_widget ||
        widget->is_my_child_widget( focus_widget ) ) )
        {
        focus_widget = NULL;
        }
    }


Widget * WidgetInputDispatcher::get_focus_widget()
    {
    return focus_widget;
    }


bool WidgetInputDispatcher::mouse_button_grabbed
    (
    const Widget * widget,
    MouseButtonGrab grab
    )
    {
    hope( widget );

    switch( grab )
        {
        case Widget::LEFT_MOUSE_BUTTON_GRAB:
            if( left_mouse_button_grab && left_mouse_button_grab_widget == widget )
                return true;
            break;
        case Widget::RIGHT_MOUSE_BUTTON_GRAB:
            if( right_mouse_button_grab && right_mouse_button_grab_widget == widget )
                return true;
            break;
        }
    return false;
    }


void WidgetInputDispatcher::release_mouse_button_grab
    (
    Widget * widget,
    MouseButtonGrab grab
    )
    {
    hope( widget );
    hope( is_my_child_widget( widget ) );
    
    switch( grab )
        {
        case Widget::LEFT_MOUSE_BUTTON_GRAB:
            if( left_mouse_button_grab && ( left_mouse_button_grab_widget == widget ||
                widget->is_my_child_widget( left_mouse_button_grab_widget ) )
                )
                {
                left_mouse_button_grab = false;
                left_mouse_button_grab_widget = NULL;
                }
            break;
        case Widget::RIGHT_MOUSE_BUTTON_GRAB:
            if( right_mouse_button_grab && ( right_mouse_button_grab_widget == widget ||
                widget->is_my_child_widget( right_mouse_button_grab_widget ) )
                )
                {
                right_mouse_button_grab = false;
                right_mouse_button_grab_widget = NULL;
                }
            break;
        }
    }


//-----------------------------------------------------------------------------
    
    
bool WidgetInputDispatcher::key_down
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    // Handle pending requests.
    execute_pending_requests();
    
    // Handle the event by super class.
    if( Super::key_down( key, mod ) )
        {
        // Handle pending requests.
        execute_pending_requests();
        return true;
        }
    
    bool result = false;
    bool eaten = false;

    {//sob
    Widget * w = focus_widget;
    if( w )
        {
        if( w->is_visible_from_parents() )
            {
            result = true;
            while( w && w != this && !eaten )
                {
                if( w->is_visible_from_parents() && w->is_enabled_from_parents() )
                    {
                    eaten = w->key_down( key, mod );
                    }
                w = w->get_parent();
                }
            }
        else
            {
            w->disable_focus();
            }
        }
    }//eob
        
    // If TAB was pressed, move to next or previous tab-focusable widget.
    if( !eaten && key == SDLK_TAB && focus_widget )
        {
        Widgets     widgets;            // Focusable, visible widgets.

        result = true;

        Widget * const parent = focus_widget->find_parent( this );
        hope( parent && "Focused widget is not child of global input dispatcher" );
        enumerate_focusable_widgets( widgets, TAB_FOCUSABLE, parent );

        if( !widgets.empty() )
            {
            // Try find widget with focus in the list.
            Widgets::iterator it = std::find( widgets.begin(), widgets.end(), focus_widget );
            if( it != widgets.end() )
                {
                // Focus widget found. Pass focus to next widget.
                
                if( ( mod & KMOD_SHIFT ) == 0 )
                    {
                    // Move to next widget.
                    it++;
                    if( it == widgets.end() )
                        {
                        it = widgets.begin();
                        }
                    }
                else
                    {
                    // Move to previous widget.
                    if( it == widgets.begin() )
                        {
                        it = widgets.end();
                        }
                    it--;
                    }

                Widget * widget = *it;
                widget->set_focus();
                }
            else
                {
                hope( focus_widget );
                focus_widget->disable_focus();
                }
            }
        else
            {
            // There are no visible focusable widgets at all.
            hope( focus_widget );
            focus_widget->disable_focus();
            }
        }
   
    // Handle pending requests.
    execute_pending_requests();
    
    return result;
    }


bool WidgetInputDispatcher::key_up
    (
    SDLKey  key,
    SDLMod  mod
    )
    {
    // Handle pending requests.
    execute_pending_requests();
    
    // Handle the event by super class.
    if( Super::key_up( key, mod ) )
        {
        // Handle pending requests.
        execute_pending_requests();
        return true;
        }
    
    bool result = false;

    bool eaten = false;
    Widget * w = focus_widget;
    if( w )
        {
        result = true;
        while( w && w != this && !eaten )
            {
            if( w->is_visible_from_parents() && w->is_enabled_from_parents() )
                {
                eaten = w->key_up( key, mod );
                }
            w = w->get_parent();
            }
        }
    
    // Handle pending requests.
    execute_pending_requests();
    
    return result;
    }


bool WidgetInputDispatcher::mouse_button_down
    (
    int                 button,
    SDLMod              mod
    )
    {
    Widget * w = NULL;
    
    // Handle pending requests.
    execute_pending_requests();
    
    if( !is_visible() )
        {
        return false;
        }
    
    // Handle the event by super class.
    if( Super::mouse_button_down( button, mod ) )
        {
        // Handle pending requests.
        execute_pending_requests();
        return true;
        }
    
    bool result = false;
    
    // Remove left mouse grab.
    if( button == SDL_BUTTON_LEFT && left_mouse_button_grab )
        {
        Widget * w = left_mouse_button_grab_widget;
        left_mouse_button_grab = false;
        left_mouse_button_grab_widget = NULL;
        if( w )
            {
            w->mouse_button_grab_lost( Widget::LEFT_MOUSE_BUTTON_GRAB );
            }
        }
    
    // Remove right mouse grab.
    if( button == SDL_BUTTON_RIGHT && right_mouse_button_grab )
        {
        Widget * w = right_mouse_button_grab_widget;
        right_mouse_button_grab = false;
        right_mouse_button_grab_widget = NULL;
        if( w )
            {
            w->mouse_button_grab_lost( Widget::RIGHT_MOUSE_BUTTON_GRAB );
            }
        }
    
    // Get visible widget under mouse.
    w = get_widget_under_mouse();
    
    if( w )
        {
        // Eat the event. We don't care whether the event will or won't be
        // eaten by the widget itself because the event is blocked by the widget.
        result = true;
    
        // If clicked outside active popup menu, hide the menu.
        if( PopupMenu::active_popup_menu &&
            PopupMenu::active_popup_menu != w &&
            !PopupMenu::active_popup_menu->is_my_child_widget( w ) )
            {
            PopupMenu::active_popup_menu->set_visibility( false );
            hope( PopupMenu::active_popup_menu == NULL );
            }

        // Try handle the event by the widget.
        // We ignore the result of the callback because we don't care of it.
        if( w->is_enabled_from_parents() )
            {
            if( w->is_auto_bringable_to_front() )
                {
                w->bring_to_front();
                }
            w->mouse_button_down( button, mod );
            }
    
        if( button == SDL_BUTTON_LEFT )
            {
            // Set grab to catch corresponding button down event in the future.
            left_mouse_button_grab = true;
            left_mouse_button_grab_widget = w;
            }
    
        if( button == SDL_BUTTON_RIGHT )
            {
            // Set focus to catch corresponding button down event in the future.
            right_mouse_button_grab = true;
            right_mouse_button_grab_widget = w;
            }
        }
    
    if( w && w->is_click_focusable() && w != focus_widget && w->is_enabled_from_parents() )
        {
        if( w->is_click_lose_focus() )
            {
            if( focus_widget )
                {
                focus_widget->disable_focus();
                }
            }
        else
            {
            w->set_focus();
            }
        }
    
    // Handle pending requests.
    execute_pending_requests();
    
    return result;
    }


bool WidgetInputDispatcher::mouse_button_up
    (
    int                 button,
    SDLMod              mod
    )
    {
    // Handle pending requests.
    execute_pending_requests();
    
    // Do not ignore the event if the dispatcher is not visible
    // because we receives the event because we have some time ago
    // handled mouse button down event.
    // if( !is_visible() )
    //     {
    //     return false;
    //     }
    
    // Handle the event by super class.
    if( Super::mouse_button_up( button, mod ) )
        {
        // Handle pending requests.
        execute_pending_requests();
        return true;
        }
    
    Widget * w = NULL;
    bool result = false;
    
    // Left mouse grab.
    if( button == SDL_BUTTON_LEFT && left_mouse_button_grab )
        {
        w = left_mouse_button_grab_widget; // Note: Can be NULL.
        left_mouse_button_grab = false;
        left_mouse_button_grab_widget = NULL;
        result = true; // Eat the event even if w is NULL.
        }
    
    // Right mouse grab.
    if( button == SDL_BUTTON_RIGHT && right_mouse_button_grab )
        {
        w = right_mouse_button_grab_widget; // Note: Can be NULL.
        right_mouse_button_grab = false;
        right_mouse_button_grab_widget = NULL;
        result = true; // Eat the event even if w is NULL.
        }
    
    // If grab widget is not visible, we only ignore the event (do not
    // remove mouse grab from the widget!).
    
    // Handle the event by grab widget.
    if( w )
        {
        // Handle the event. Ignore the event if the widget is not visible.
        if( w->is_visible_from_parents() && w->is_enabled_from_parents() )
            {
            // Ignore result!
            w->mouse_button_up( button, mod );
            }
        
        // Notify of grab lost.
        if( button == SDL_BUTTON_LEFT )
            {
            w->mouse_button_grab_lost( Widget::LEFT_MOUSE_BUTTON_GRAB );
            }
        if( button == SDL_BUTTON_RIGHT )
            {
            w->mouse_button_grab_lost( Widget::RIGHT_MOUSE_BUTTON_GRAB );
            }
        }
    
    // If the event was not eaten, try handle it by widget under mouse.
    if( !result )
        {
        w = get_widget_under_mouse();
        if( w )
            {
            if( w->is_enabled_from_parents() )
                {
                w->mouse_button_up( button, mod );
                }
            result = true;
            }
        }
    
    // Handle pending requests.
    execute_pending_requests();

    return result;
    }

    
bool WidgetInputDispatcher::mouse_move
    (
    int     dx,
    int     dy,
    SDLMod  mod
    )
    {
    // Handle pending requests.
    execute_pending_requests();
    
    if( !is_visible() )
        {
        return false;
        }
    
    // Handle the event by super class.
    if( Super::mouse_move( dx, dy, mod ) )
        {
        // Handle pending requests.
        execute_pending_requests();
        return true;
        }
        
    bool result = false;

    // Left mouse grab.
    if( left_mouse_button_grab )
        {
        Widget * w = left_mouse_button_grab_widget;
        if( w && w->is_visible_from_parents() && w->is_enabled_from_parents() )
            {
            w->mouse_move( dx, dy, mod );
            }
        result = true; // Eat the event.
        }
    
    // Right mouse grab.
    if( right_mouse_button_grab )
        {
        Widget * w = right_mouse_button_grab_widget;
        if( w && w->is_visible_from_parents() && w->is_enabled_from_parents() )
            {
            w->mouse_move( dx, dy, mod );
            }
        result = true; // Eat the event.
        }
    
    // If the mouse is not grabbed, try handle the event by widget under mouse.
    if( !result )
        {
        Widget * w;
        w = get_widget_under_mouse();
        if( w )
            {
            if( w->is_enabled_from_parents() )
                {
                w->mouse_move( dx, dy, mod );
                }
            result = true;
            }
        }

    // Handle pending requests.
    execute_pending_requests();

    return result;
    }


void WidgetInputDispatcher::execute_pending_requests()
    {
    }


void WidgetInputDispatcher::enumerate_focusable_widgets
    (
    Widgets &   widgets,
    int         behavior_mask,
    Widget *    parent
    )
    {
    hope( behavior_mask & FOCUSABLE );

    // Enumerate focusable, visible widgets.
    typedef std::queue< Widget * > WidgetQueue;
        
    WidgetQueue widgets_to_examine; // Queue of widgets to examine.
    
    // Initialize.
    widgets.clear();
    widgets_to_examine.push( parent ? parent : this );
        
    // Enumerate all visible, focusable widgets.
    while( !widgets_to_examine.empty() )
        {
        Widget * widget = widgets_to_examine.front();
        widgets_to_examine.pop();
            
        // Stop enumeration if reached another widget input dispatcher.
        // The input dispatcher handles its focus widget by itself.
        if( widget->is_dispatcher() && this != widget )
            {
            // TODO: Print warning. There should be only one global dispatcher.
            continue;
            }
            
        // We don't need to use is_visible_from_parents() or is_enabled_from_parents()
        // because we are enumerating widgets from parents to children.
        if( !widget->is_visible() || !widget->is_enabled() )
            {
            continue;
            }
            
        if( Massiv::Core::flags_on( widget->behavior, behavior_mask ) )
            {
            widgets.push_back( widget );
            }
            
        // Enumerate children of the widget.
        Children::iterator child_it;
        for( child_it = widget->children.begin(); child_it != widget->children.end(); ++child_it )
            {
            Widget * child = *child_it;
            hope( child );
            widgets_to_examine.push( child );
            }
        }
    }


Widget * WidgetInputDispatcher::get_widget_under_mouse()
    {
    if( !is_visible() )
        {
        return NULL;
        }

    return get_child_widget_under_mouse( this );
    }


Widget * WidgetInputDispatcher::get_child_widget_under_mouse
    (
    Widget * widget
    )
    {
    hope( widget );
    
    const SVector2 mouse_pos = Global::get_input_manager().get_mouse_position();
    
    // Note: Area occupied by child widget on screen is not limited by area
    // of its parent. Child widget may be outside its parent on screen.
    // So we must iterate through entire widget hiearchy.
    
        {
        // We iterate in reverse order because child overlay objects are drawn
        // in order they are registered to their parent (last drawn object
        // is top most visible on screen so it should receive input events
        // before other child widgets).
        Children::reverse_iterator  child_it;
        
        for( child_it = widget->children.rbegin(); child_it != widget->children.rend(); ++child_it )
            {
            Widget * child = *child_it;
            hope( child );
            if( child->is_visible() )
                {
                Widget * w = get_child_widget_under_mouse( child );
                if( w )
                    {
                    return w;
                    }
                if( child->contains_screen_point( mouse_pos ) )
                    {
                    return child;
                    }
                }
            }
        }

    return NULL;
    }


} // namespace Demo::Client
} // namespace Demo
