/** @file widget_input_dispatcher.h
 * Widget input dispatcher.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_WIDGET_INPUT_DISPATCHER_H
#define CLIENT_WIDGET_INPUT_DISPATCHER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_WIDGET_H
#include "gui/widget.h"
#endif


namespace Demo {
namespace Client {


/**
 * WidgetInputDispatcher dispatches input events to its child widgets.
 *
 * #Widget itself should not be registered as input object in #InputManager
 * because #InputManager is not able to dispatch input events
 * to its input objects depending on their screen positions and widget
 * children.
 * Instead of registering widgets in #InputManager, WidgetInputDispatcher
 * should be registered in the #InputManager and the widgets should be
 * linked to the WidgetInputDispatcher as child widgets (this is possible
 * because WidgetInputDispatcher is #Widget itself).
 *
 * When an input event arrives, WidgetInputDispatcher tries to
 * handle the event by sending the event to its child widgets.
 * 
 * Inputs events are not send to any widget which is not visible or
 * have disabled input including its child widgets - if parent widget
 * is not visible or have disabled input, no event is sent
 * to its child widgets, even if the child widgets are visible
 * or have enabled input.
 *
 * Mouse clicks and mouse move events are sent only to widget
 * at mouse position. Whether a widget is at mouse position or not is
 * detected by calling @c Widget::is_mouse_on() method.
 * Once @c Widget::is_mouse_on() returns true, the event will be eaten
 * even if the widget itself doesn't process the event. This
 * behaviour is called "event blocking".
 *
 * When processing mouse button down, the widget under mouse will
 * grab the mouse. The mouse stays grabbed until corresponding
 * mouse button up event. While mouse is grabbed, the grab widget receives all
 * mouse move events even if the mouse is not above the widget.
 * The widgets will alse receive the final mouse button up events
 * and WidgetInputDispatcher will call @c Widget::mouse_grab_lost()
 * method on the widget.
 * Mouse is grabbed separately for left and right mouse button.
 * If mouse is not grabbed, mouse move events are sent to widgets
 * under the mouse.
 *
 * Keyboard events are sent only to one one explicitly selected widget.
 * This widget is called widget with keyboard focus.
 * If widget with keybord focus is set,
 * Use @c Widget::set_focus() and @c Widget::disable_focus()
 * to set keybord focus to a widget.
 *
 * WidgetInputDispatcher automatically handle TAB key to iterate
 * through visible focusable widgets.
 */
class WidgetInputDispatcher : public Widget
    {
    typedef Widget  Super;
    
    NO_COPY_CONSTRUCTOR( WidgetInputDispatcher );
    NO_ASSIGNMENT_OPERATOR( WidgetInputDispatcher );
    
public:

    /**
     * Constructor.
     * @note WidgetInputDispatcher is created as #Widget with
     *      @c Widget::TRANSPARENT_WIDGET type.
     */
    WidgetInputDispatcher();
    
    /** Destructor. */
    virtual ~WidgetInputDispatcher();
    
public:
    
    // InputObject
    virtual bool key_up
        (
        SDLKey  key,
        SDLMod  mod
        );
        
    // InputObject
    virtual bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );
    
    // InputObject
    virtual bool mouse_button_up
        (
        int     button,
        SDLMod  mod
        );

    // InputObject
    virtual bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );
    
    // InputObject
    virtual bool mouse_move
        (
        int    dx,
        int    dy,
        SDLMod mod
        );
        
public:

    /**
     * Check whether mouse is above the widget.
     */
    bool is_mouse_on
        (
        const Widget * widget
        );

    /**
     * Check whether @a widget has keyboard focus.
     */
    bool has_focus
        (
        const Widget * widget
        );

    /**
     * Set keyboard focus to @a widget.
     * @pre The widget is (not neccessary direct) child of the dispatcher.
     * @note Do not use this function directly, use @c Widget::set_focus()
     *  instead.
     */
    void set_focus
        (
        Widget * widget
        );

    /**
     * Disable keybord focus from @a widget or its child widgets.
     * @pre The widget is (not neccessary direct) child of the dispatcher.
     * @note Do not use this function directly, use @c Widget::disable_focus()
     *  instead.
     */
    void disable_focus
        (
        Widget * widget
        );

    /**
     * Get pointer to widget with keyboard focus.
     */
    Widget * get_focus_widget();
    
    /**
     * Check whether the widget has grabbed mouse button.
     * While mouse button is grabbed, the widget will receive
     * all mouse move events until corresponding mouse buton is release.
     */
    bool mouse_button_grabbed
        (
        const Widget * widget,
        MouseButtonGrab grab
        );

    /**
     * Release mouse button grab fro @a widget or its child widgets.
     * @pre The widget is (not neccessary direct) child of the dispatcher.
     */
    void release_mouse_button_grab
        (
        Widget * widget,
        MouseButtonGrab grab
        );

private:

    typedef std::vector< Widget * > Widgets;

    void execute_pending_requests();
    
    /**
     * Enumerate focusable, visible and enabled widgets.
     * @param widgets Ouput widget list. Cleared at the beginning.
     * @param behavior_mask All given behavior flags of widget must be set
     *                      to add it to @a widgets.
     * @param parent If set, enumerate only child widgets of @a parent including
     *               @a parent itself.
     */
    void enumerate_focusable_widgets
        (
        Widgets &   widgets,
        int         behavior_mask,
        Widget *    parent
        );
    
    /**
     * Get visible child widget under mouse.
     */
    Widget * get_widget_under_mouse();
    
    /**
     * Get visible child widget under mouse.
     */
    Widget * get_child_widget_under_mouse
        (
        Widget * widget
        );

private:

    Widget *    focus_widget;
        /**< Current widget with keyboard focus. */
    bool        left_mouse_button_grab;
        /**< If true, left mouse button was pressed on some child widgets
         * some time ago and the button was not released yet.
         * @c left_mouse_button_grab_widget specifies the widget.
         */
    Widget *    left_mouse_button_grab_widget;
        /**< Widget on which left mouse button was pressed if
         * @c left_mouse_button_grab is true.
         * Note that even if @c left_mouse_button_grab is true,
         * @c left_mouse_button_grab_widget can be null (for example
         * because the widget was destroyed but the mouse button was not yet
         * released).
         */
    bool        right_mouse_button_grab;
        /**< If true, right mouse button was pressed on some child widgets
         * some time ago and the button was not released yet.
         * @c right_mouse_button_grab_widget specifies the widget.
         */
    Widget *    right_mouse_button_grab_widget;
        /**< Widget on which right mouse button was pressed if
         * @c right_mouse_button_grab is true.
         * Note that even if @c right_mouse_button_grab is true,
         * @c right_mouse_button_grab_widget can be null (for example
         * because the widget was destroyed but the mouse button was not yet
         * released).
         */

    }; // WidgetInputDispatcher


} // namespace Demo::Client
} // namespace Demo


#endif // CLIENT_WIDGET_INPUT_DISPATCHER_H
