/** @file widget_layout.cpp
 * Helper classes to simplify positioning of child widgets (implementation).
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "gui/widget_layout.h"


namespace Demo {
namespace Client {


namespace {

Float32 fmax
    (
    Float32 f1,
    Float32 f2
    )
    {
    return ( f1 < f2 ? f2 : f1 );
    }

} // anonymous namespace


WidgetTableLayout::WidgetTableLayout
    (
    Float32 the_border_width,
    Float32 the_border_height,
    Float32 the_padding_width,
    Float32 the_padding_height,
    const SVector2 & the_position
    ) :
    border_width( the_border_width ),
    border_height( the_border_height ),
    padding_width( the_padding_width ),
    padding_height( the_padding_height ),
    position( the_position )
    {
    }


void WidgetTableLayout::next_row()
    {
    // Finish previous cell.
    if( !rows.empty() )
        {
        Row & row = rows[ rows.size() - 1 ];
        if( !row.cells.empty() )
            {
            int prev_col_span = row.cells[ row.cells.size() - 1 ].col_span;
            hope( prev_col_span > 0 );
            for( prev_col_span--; prev_col_span; prev_col_span-- )
                {
                row.cells.push_back( Cell() );
                }
            }
        }
    
    rows.push_back( Row() );
    }


void WidgetTableLayout::next_cell
    (
    int     col_span,
    HAlign  halign,
    VAlign  valign
    )
    {
    hope( !rows.empty() );
    hope( col_span > 0 );
    
    Row & row = rows[ rows.size() - 1 ];
    
    // Finish previous cell.
    if( !row.cells.empty() )
        {
        int prev_col_span = row.cells[ row.cells.size() - 1 ].col_span;
        hope( prev_col_span > 0 );
        for( prev_col_span--; prev_col_span; prev_col_span-- )
            {
            row.cells.push_back( Cell() );
            }
        }
    
    row.cells.push_back( Cell( col_span, halign, valign ) );
    }


void WidgetTableLayout::put_object
    (
    OverlayObject * object
    )
    {
    if( !object )
        {
        return;
        }
    
    hope( !rows.empty() );
    Row & row = rows[ rows.size() - 1 ];
    hope( !row.cells.empty() );
    Cell & cell = row.cells[ row.cells.size() - 1 ];
    cell.objects.push_back( CellObject( object ) );
    }


void WidgetTableLayout::compute_cell_object_height
    (
    CellObject & cell_object
    )
    {
    cell_object.height = cell_object.object->get_current_size().y();
    }


void WidgetTableLayout::compute_cell_object_width
    (
    CellObject & cell_object
    )
    {
    cell_object.width = cell_object.object->get_current_size().x();
    }


void WidgetTableLayout::compute_cell_height
    (
    Cell & cell
    )
    {
    CellObjects::iterator it;
    Float32 height = 0.0f;
    
    // Compute height of the cell - maximum of heights of its objects.
    for( it = cell.objects.begin(); it != cell.objects.end(); ++it )
        {
        CellObject & cell_object = *it;
        compute_cell_object_height( cell_object );
        height = fmax( height, cell_object.height );
        }

    cell.height = height;
    }


void WidgetTableLayout::compute_cell_min_width
    (
    Cell & cell
    )
    {
    CellObjects::iterator it;
    Float32 min_width = 0.0f;
    
    for( it = cell.objects.begin(); it != cell.objects.end(); ++it )
        {
        CellObject & cell_object = *it;
        compute_cell_object_width( cell_object );
        min_width += cell_object.width;
        
        // Add horizontal padding.
        if( ( it + 1 ) != cell.objects.end() )
            {
            min_width += padding_width;
            }
        }

    cell.min_width = min_width;
    }


void WidgetTableLayout::compute_row_height
    (
    Row & row
    )
    {
    Cells::iterator it;
    Float32 height = 0.0f;
    
    // Compute height of the row - maximum of heights of its cells.
    for( it = row.cells.begin(); it != row.cells.end(); ++it )
        {
        Cell & cell = *it;
        compute_cell_height( cell );
        height = fmax( height, cell.height );
        }

    row.height = height;
    }


Float32 WidgetTableLayout::compute_min_width_of_column
    (
    std::size_t column_index
    )
    {
    Rows::iterator it;
    Float32 column_width = 0.0f;
    
    for( it = rows.begin(); it != rows.end(); ++it )
        {
        Row & row = *it;
        // Check if the column is present in this row.
        if( column_index < row.cells.size() )
            {
            Cell & cell = row.cells[ column_index ];
            compute_cell_min_width( cell );
            column_width = fmax( column_width, cell.min_width );
            }
        }
    
    return column_width;
    }


void WidgetTableLayout::set_width_of_column
    (
    std::size_t column_index,
    Float32     column_width
    )
    {
    Rows::iterator it;
    
    for( it = rows.begin(); it != rows.end(); ++it )
        {
        Row & row = *it;
        // Check if the column is present in this row.
        if( column_index < row.cells.size() )
            {
            Cell & cell = row.cells[ column_index ];
            cell.width = column_width;
            }
        }
    }


void WidgetTableLayout::do_cell_layout
    (
    Cell &              cell,
    Float32             height,
    const SVector2 &    top_left
    )
    {
    hope( cell.min_width <= cell.width );
    hope( cell.height <= height );
    
    Float32 x = 0.0f;
    
    if( cell.halign == HALIGN_CENTER )
        {
        x += ( cell.width - cell.min_width ) / 2.0f;
        }
    else if( cell.halign == HALIGN_RIGHT )
        {
        x += cell.width - cell.min_width;
        }
    
    CellObjects::iterator it;
    for( it = cell.objects.begin(); it != cell.objects.end(); ++it )
        {
        CellObject & cell_object = *it;
        Float32 y = 0.0f;
        
        if( cell.valign == VALIGN_CENTER )
            {
            y += ( height - cell_object.height ) / 2.0f;
            }
        else if( cell.valign == VALIGN_BOTTOM )
            {
            y += height - cell_object.height;
            }
        
        cell_object.object->set_base_offset( top_left + SVector2( x, y ) );
        x += cell_object.width;
        
        // Add padding.
        x += padding_width;
        }
    }


void WidgetTableLayout::do_row_layout
    (
    Row &               row,
    const SVector2 &    top_left
    )
    {
    Cells::iterator it;
    SVector2 cell_top_left = top_left;
    for( it = row.cells.begin(); it != row.cells.end(); ++it )
        {
        Cell & cell = *it;
        do_cell_layout( cell, row.height, cell_top_left );
        cell_top_left += SVector2( cell.width, 0.0f );
        // Add padding.
        if( ( it + 1 ) != row.cells.end() )
            {
            cell_top_left += SVector2( padding_width, 0.0f );
            }
        }
    }


SVector2 WidgetTableLayout::do_layout()
    {
    Float32 table_width = 0.0f;
    Float32 table_height = 0.0f;
    std::size_t num_columns = 0;
    
    // Finish previous cell.
    if( !rows.empty() )
        {
        Row & row = rows[ rows.size() - 1 ];
        if( !row.cells.empty() )
            {
            int prev_col_span = row.cells[ row.cells.size() - 1 ].col_span;
            hope( prev_col_span > 0 );
            for( prev_col_span--; prev_col_span; prev_col_span-- )
                {
                row.cells.push_back( Cell() );
                }
            }
        }
    
    // Get num of columns.
        {
        Rows::iterator it;
        for( it = rows.begin(); it != rows.end(); ++it )
            {
            Row & row = *it;
            if( num_columns < row.cells.size() )
                {
                num_columns = row.cells.size();
                }
            }
        }
    
    // Compute width of each column.
        {
        std::size_t column_index;
        for( column_index = 0; column_index < num_columns; ++column_index )
            {
            const Float32 column_width = compute_min_width_of_column( column_index );
            set_width_of_column( column_index, column_width );
            table_width += column_width;
            // Add padding.
            if( column_index + 1 != num_columns )
                {
                table_width += padding_width;
                }
            }
        }
    
    // Compute width of cells with col_span > 1.
        {
        Rows::iterator it;
        for( it = rows.begin(); it != rows.end(); ++it )
            {
            Row & row = *it;
            Cells::iterator cell_it;
            for( cell_it = row.cells.begin(); cell_it != row.cells.end(); ++cell_it )
                {
                Cell & cell = *cell_it;
                int col_span = cell.col_span;
                hope( col_span >= 1 );
                for( col_span--; col_span; col_span-- )
                    {
                    // Eat width of the cell that is invisible.
                    ++cell_it;
                    hope( cell_it != row.cells.end() ); 
                    Cell & invis_cell = *cell_it;
                    hope( invis_cell.col_span == 1 );
                    hope( invis_cell.objects.empty() );
                    cell.width += invis_cell.width;
                    }
                }
            }
        }
    
    // Compute height of each row.
        {
        Rows::iterator it;
        for( it = rows.begin(); it != rows.end(); ++it )
            {
            Row & row = *it;
            compute_row_height( row );
            table_height += row.height;
            // Add padding.
            if( ( it + 1 ) != rows.end() )
                {
                table_height += padding_height;
                }
            }
        }
    
    // Do layout of the table.
        {
        SVector2 top_left = position + SVector2( border_width, border_height );
        Rows::iterator it;
        for( it = rows.begin(); it != rows.end(); ++it )
            {
            Row & row = *it;
            do_row_layout( row, top_left );
            top_left += SVector2( 0.0f, row.height + padding_height );
            }
        }
    
    // Add borders.
    table_width += border_width * 2.0f;
    table_height += border_height * 2.0f;
    
    return SVector2( table_width, table_height );
    }


} // namespace Demo::Client
} // namespace Demo
