/** @file widget_layout.h
 * Helper classes to simplify positioning of child widgets.
 */

/*
 * begin        : Nov 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_WIDGET_LAYOUT_H
#define CLIENT_WIDGET_LAYOUT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OVERLAY_OBJECT_H
#include "renderer/overlay_object.h"
#endif

#include <vector>


namespace Demo {
namespace Client {


/**
 * WidgetTableLayout is helper object that can be used
 * to set positions of child widgets of a parent widget in
 * such way that the widgets will be placed in an invisible
 * table. Position of each of the child widgets is set
 * relative to top left corner of the table.
 *
 * Table is divided into several rows. Each row has several cells.
 * One cell can contain several widgets.
 * Each cell is expanded so all objects registered in the cell
 * will fit in the cell.
 *
 * Usage:
 * @code
 * // Index of a cell is pair [r,c], where r stands for row
 * // and c stands for column of the table.
 * WidgetTableLayout layout( 4.0f, 4.0f, 2.0f, 2.0f );
 *
 * // Begin first row.
 * layout.next_row();
 * layout.next_cell();
 *
 * // Child 1 and 2 in one cell at index [0,0].
 * layout.put_object( child_widget1 );
 * layout.put_object( child_widget2 );
 *
 * // Cell [1,0] will be empty. It will be automatically expanded
 * // so its width will be same as width of cell [0,0].
 * layout.next_row();
 * layout.next_cell();
 *
 * // Child 3 in cell [1,1].
 * layout.next_cell();
 * layout.put_object( child_widget3 );
 *
 * // Once all child widgets are registered, do layout of the table.
 * SVector2 size = layout.do_layout();
 *
 * // Set size of the parent widget to size of the table.
 * parent_widget->set_base_size( size );
 *
 * @endcode
 */
class WidgetTableLayout
    {
    NO_COPY_CONSTRUCTOR( WidgetTableLayout );
    NO_ASSIGNMENT_OPERATOR( WidgetTableLayout );

    typedef Renderer::OverlayObject OverlayObject;

public:

    /**< Vertical alignment. */
    enum VAlign
        {
        VALIGN_TOP,
        VALIGN_CENTER,
        VALIGN_BOTTOM
        };
    
    /**< Horizontal alignment. */
    enum HAlign
        {
        HALIGN_LEFT,
        HALIGN_CENTER,
        HALIGN_RIGHT
        };

public:

    /**
     * Constructor.
     * @param position Position of top left corner of the table
     * relative to top left cordern of parent widget.
     */
    WidgetTableLayout
        (
        Float32 border_width,
        Float32 border_height,
        Float32 padding_width,
        Float32 padding_height,
        const SVector2 & position = SVector2( 0.0f, 0.0f )
        );

public:

    /** Begin new table row. */
    void next_row();
    
    /**
     * Begin new table cell.
     * @param col_span How many columns the cell occupies. Must be greater than zero.
     * @param halign Horizontal alignment of objects in the cell.
     * @param valign Vertical alignment of objects in the cell.
     */
    void next_cell
        (
        int     col_span = 1,
        HAlign  halign = HALIGN_CENTER,
        VAlign  valign = VALIGN_CENTER
        );
    
    /**
     * Put @a object in current cell.
     * @warning The object should not be deleted until the WidgetTableLayout
     * object is destroyed.
     */
    void put_object
        (
        OverlayObject * object
        );
        
    /**
     * Do layout of the table.
     * Position of each registered object is set based on the table layout
     * relative to top left corner of the table.
     * @return Width of the table (including borders).
     */
    SVector2 do_layout();

private:

    /** Object in table cell. */
    struct CellObject
        {
        CellObject() :
            object( NULL ),
            width( 0.0f ),
            height( 0.0f )
            {
            }

        CellObject
            (
            OverlayObject * the_object
            ) :
            object( the_object ),
            width( 0.0f ),
            height( 0.0f )
            {
            }

        OverlayObject * object;     ///< Overlay object.
        Float32         width;      ///< Width of the overlaty object.
        Float32         height;     ///< Height of the overlay object.
        }; // CellObject
    
    typedef std::vector< CellObject > CellObjects;

    /** Table cell. */
    struct Cell
        {
        Cell() :
            col_span( 1 ),
            halign( HALIGN_CENTER ),
            valign( VALIGN_CENTER ),
            width( 0.0f ),
            height( 0.0f ),
            min_width( 0.0f )
            {
            }

        Cell(
            int the_col_span,
            HAlign the_halign,
            VAlign the_valign
            ) :
            col_span( the_col_span ),
            halign( the_halign ),
            valign( the_valign ),
            width( 0.0f ),
            height( 0.0f ),
            min_width( 0.0f )
            {
            }
        
        CellObjects     objects;    ///< Cell objects in the cell.
        int             col_span;   ///< How many columns this cell occupies.
        HAlign          halign;     ///< Horizontal alignment of the cell.
        VAlign          valign;     ///< Vertical alignment of the cell.
        Float32         width;      ///< Final width of the cell.
        Float32         height;     ///< Height of objects in the cell.
        Float32         min_width;  ///< Width of objects in the cell.
        }; // Cell

    typedef std::vector< Cell > Cells;

    /** Table row. */
    struct Row
        {
        Row() :
            height( 0.0f )
            {
            }
        
        Cells           cells;  ///< Cells in the row.
        Float32         height; ///< Final row height.
        }; // Row

    typedef std::vector< Row > Rows;

private:

    void compute_cell_object_height
        (
        CellObject & cell_object
        );

    void compute_cell_object_width
        (
        CellObject & cell_object
        );
        
    void compute_cell_height
        (
        Cell & cell
        );
        
    void compute_cell_min_width
        (
        Cell & cell
        );
        
    void compute_row_height
        (
        Row & row
        );
        
    Float32 compute_min_width_of_column
        (
        std::size_t column_index
        );
        
    void set_width_of_column
        (
        std::size_t column_index,
        Float32     column_width
        );
        
    void do_cell_layout
        (
        Cell &              cell,
        Float32             height,
        const SVector2 &    top_left
        );
        
    void do_row_layout
        (
        Row &               row,
        const SVector2 &    top_left
        );

private:

    Float32         border_width;
    Float32         border_height;
    Float32         padding_width;
    Float32         padding_height;
    SVector2        position;
    Rows            rows;

    }; // WidgetTableLayout


} // namespace Demo::Client
} // namespace Demo


#endif // CLIENT_WIDGET_LAYOUT_H
