/** @file input_manager.cpp
 * Mouse and keyboard manager (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "input/input_manager.h"
#include "input/input_object.h"
#include "global/global.h"

#include "client_global.h" // REMOVE
#include "gui/terrain_window.h" // REMOVE


namespace Demo {
namespace Client {


// InputManager::Event
// ~~~~~~~~~~~~~~~~~~~

InputManager::Event::Event() :
    type( TYPE_NONE ),
    mod( KMOD_NONE ),
    key( SDLK_UNKNOWN ),
    mouse_button( 0 ),
    dx( 0 ),
    dy( 0 )
    {
    }


void InputManager::Event::set_key_up
    (
    SDLKey              _key,
    SDLMod              _mod,
    const SystemTime &  _release_time,
    const SystemTime &  _press_time
    )
    {
    type = TYPE_KEY_UP;
    key = _key;
    mod = _mod;
    time = _release_time;
    complement_time = _press_time;
    }


void InputManager::Event::set_key_down
    (
    SDLKey              _key,
    SDLMod              _mod,
    const SystemTime &  _release_time,
    const SystemTime &  _press_time
    )
    {
    type = TYPE_KEY_DOWN;
    key = _key;
    mod = _mod;
    time = _press_time;
    complement_time = _release_time;
    }


void InputManager::Event::set_mouse_move
    (
    int                 _dx,
    int                 _dy,
    SDLMod              _mod,
    const SystemTime &  _time
    )
    {
    type = TYPE_MOUSE_MOVE;
    dx = _dx;
    dy = _dy;
    mod = _mod;
    time = _time;
    }


void InputManager::Event::set_mouse_button_up
    (
    int                 _mouse_button,
    SDLMod              _mod,
    const SystemTime &  _release_time,
    const SystemTime &  _press_time
    )
    {
    type = TYPE_MOUSE_BUTTON_UP;
    mouse_button = _mouse_button;
    mod = _mod;
    time = _release_time;
    complement_time = _press_time;
    }


void InputManager::Event::set_mouse_button_down
    (
    int                 _mouse_button,
    SDLMod              _mod,
    const SystemTime &  _release_time,
    const SystemTime &  _press_time
    )
    {
    type = TYPE_MOUSE_BUTTON_DOWN;
    mouse_button = _mouse_button;
    mod = _mod;
    time = _press_time;
    complement_time = _release_time;
    }


// InputManager
// ~~~~~~~~~~~~

InputManager::InputManager() :
    left_mouse_button_grab_object( NULL ),
    right_mouse_button_grab_object( NULL )
    {
    keys.resize( SDLK_LAST );
    mouse_buttons.resize( MOUSE_BUTTON_LAST );
    }


InputManager::~InputManager()
    {
    // We never keep empty layers, so layers.empty() <=> no objects registered.

    if( !layers.empty() )
        {
        // Unregister objects.

        int object_count = 0;
        Layers::iterator layer_it;
        while( ( layer_it = layers.begin() ) != layers.end() )
            {
            Objects & objects = layer_it->second;
            const Objects::iterator object_it = objects.begin();
            hope( object_it != objects.end() );
            InputObject * const registered_object = *object_it;
            hope( registered_object );
            registered_object->unregister_from_input_manager();
            object_count++;
            }

        // Log warning.

        {//sob
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << object_count << " objects still registered in ~InputManager.";
        ::Massiv::Core::Global::log_warning
            (
            Status::FACILITY_GENERIC,
            Status::PRIORITY_NORMAL,
            oss.str()
            );
        }//eob

        }
    }


SVector2 InputManager::get_mouse_position() const
    {
    int xx, yy;
    get_mouse_position( xx, yy );

    Float32 x = static_cast< Float32 >( xx );
    Float32 y = static_cast< Float32 >( yy );

    return SVector2( x, y );
    }


void InputManager::get_mouse_position
    (
    int & x,
    int & y
    ) const
    {
    const int max_x = SDL::get_screen_width() - 1;
    const int max_y = SDL::get_screen_height() - 1;
    int mouse_x, mouse_y;
    SDL_GetMouseState( &mouse_x, &mouse_y );

    // Clip to screen.
    mouse_x = ( mouse_x < 0 ? 0 : ( mouse_x > max_x ? max_x : mouse_x ) );
    mouse_y = ( mouse_y < 0 ? 0 : ( mouse_y > max_y ? max_y : mouse_y ) );

    x = mouse_x;
    y = mouse_y;
    }


bool InputManager::is_key_down
    (
    SDLKey key
    ) const
    {
    if( key >= 0 && key < SDLK_LAST )
        {
        return keys[ key ].down;
        }
    else
        {
        return false;
        }
    }


bool InputManager::is_mouse_button_down
    (
    int mouse_button
    ) const
    {
    MouseButton b = sdl_to_native_mouse_button( mouse_button );

    if( b != MOUSE_BUTTON_LAST )
        {
        return mouse_buttons[ b ].down;
        }
    else
        {
        return false;
        }
    }


const InputManager::KeyState & InputManager::get_key_state
    (
    SDLKey key
    ) const
    {
    hope( key >= 0 && key < SDLK_LAST );
    return keys[ key ];
    }


const InputManager::MouseButtonState & InputManager::get_mouse_button_state
    (
    int mouse_button
    ) const
    {
    MouseButton b = sdl_to_native_mouse_button( mouse_button );

    hope( b != MOUSE_BUTTON_LAST );
    return mouse_buttons[ b ];
    }


void InputManager::key_up
    (
    const SDL_keysym & keysym
    )
    {
    SDLKey  key = keysym.sym;
    SDLMod  mod = keysym.mod;

    if( key >= 0 && key < SDLK_LAST )
        {
        KeyState & state = keys[ key ];

        state.down = false;
        state.release_time = Massiv::System::system_time();
        }

    // Dispatch event.
    if( !Global::rendering_disabled )
        {
        Event   event;
        event.set_key_up( key, mod, keys[ key ].release_time, keys[ key ].press_time );
        dispatch_event( event );
        }
    }


void InputManager::key_down
    (
    const SDL_keysym & keysym
    )
    {
    SDLKey  key = keysym.sym;
    SDLMod  mod = keysym.mod;

    if( key >= 0 && key < SDLK_LAST )
        {
        KeyState & state = keys[ key ];

        state.down = true;
        state.press_time = Massiv::System::system_time();
        }

    // Dispatch event.
    if( !Global::rendering_disabled )
        {
        Event   event;
        event.set_key_down( key, mod, keys[ key ].release_time, keys[ key ].press_time );
        dispatch_event( event );

        // FIXME: Nechutny hack. Test kamery.
        //Global::get_terrain_window().update_camera(); //~!
        }
    else
        {
        if( ( key == 'r' ) && ( mod & KMOD_CTRL ) )
            {
            Global::enable_rendering();
            }
        }
    }


void InputManager::mouse_move
    (
    int dx,
    int dy
    )
    {
    if( Global::rendering_disabled )
        {
        return;
        }

    SDLMod mod = SDL_GetModState();

    // Dispatch the event.
    Event   event;
    event.set_mouse_move( dx, dy, mod, Massiv::System::system_time() );
    dispatch_event( event );
    }


void InputManager::mouse_button_up
    (
    int button
    )
    {
    MouseButton b = sdl_to_native_mouse_button( button );
    SDLMod      mod = SDL_GetModState();

    if( b != MOUSE_BUTTON_LAST )
        {
        MouseButtonState & state = mouse_buttons[ b ];

        state.down = false;
        state.release_time = Massiv::System::system_time();
        }

    // Dispatch event.
    if( !Global::rendering_disabled )
        {
        Event   event;
        event.set_mouse_button_up
            (
            button,
            mod,
            mouse_buttons[ b ].release_time,
            mouse_buttons[ b ].press_time
            );
        dispatch_event( event );
        }
    }


void InputManager::mouse_button_down
    (
    int button
    )
    {
    MouseButton b = sdl_to_native_mouse_button( button );
    SDLMod      mod = SDL_GetModState();

    if( b != MOUSE_BUTTON_LAST )
        {
        MouseButtonState & state = mouse_buttons[ b ];

        state.down = true;
        state.press_time = Massiv::System::system_time();
        }

    // Dispatch event.
    if( !Global::rendering_disabled )
        {
        Event   event;
        event.set_mouse_button_down
            (
            button,
            mod,
            mouse_buttons[ b ].release_time,
            mouse_buttons[ b ].press_time
            );
        dispatch_event( event );
        }
    }


void InputManager::register_object
    (
    InputObject *   object,
    int             layer
    )
    {
    hope( object );

    // Register the object.
    Objects & objects = layers[ layer ];
    objects.push_back( object );
    }


void InputManager::unregister_object
    (
    InputObject *   object,
    int             layer
    )
    {
    hope( object );

    if( left_mouse_button_grab_object == object )
        {
        left_mouse_button_grab_object = NULL;
        }

    if( right_mouse_button_grab_object == object )
        {
        right_mouse_button_grab_object = NULL;
        }

    // Get the layer.
    Layers::iterator layer_it = layers.find( layer );
    hope( layer_it->first == layer );
    Objects & objects = layer_it->second;

    // Unregister the object.
    Objects::iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        InputObject * const registered_object = *object_it;
        hope( registered_object );
        if( registered_object == object )
            {
            objects.erase( object_it );

            // Remove layer if it's empty.
            if( objects.empty() )
                {
                layers.erase( layer_it );
                }

            // Done.
            return;
            }
        }

    panic( "Object not registered in unregister_object()." );
    }


bool InputManager::dispatch_event
    (
    const Event & event
    )
    {
    Layers::iterator layer_it;
    bool             eaten = false; // Was the event eaten?

    if( event.type == Event::TYPE_MOUSE_BUTTON_UP &&
        event.mouse_button == SDL_BUTTON_LEFT &&
        left_mouse_button_grab_object )
        {
        eaten = true;
        dispatch_event_to_object( left_mouse_button_grab_object, event );
        left_mouse_button_grab_object = NULL;
        }

    if( event.type == Event::TYPE_MOUSE_BUTTON_UP &&
        event.mouse_button == SDL_BUTTON_RIGHT &&
        right_mouse_button_grab_object )
        {
        eaten = true;
        dispatch_event_to_object( right_mouse_button_grab_object, event );
        right_mouse_button_grab_object = NULL;
        }

    if( event.type == Event::TYPE_MOUSE_MOVE &&
        ( left_mouse_button_grab_object || right_mouse_button_grab_object ) )
        {
        eaten = true;
        if( left_mouse_button_grab_object )
            {
            dispatch_event_to_object( left_mouse_button_grab_object, event );
            }
        if( right_mouse_button_grab_object )
            {
            dispatch_event_to_object( right_mouse_button_grab_object, event );
            }
        }

    if( !eaten )
        {
        // For each layer...
        for( layer_it = layers.begin(); layer_it != layers.end(); ++layer_it )
            {
            Objects & objects = layer_it->second;
            Objects::iterator object_it;

            // ...for each object in the layer...
            for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
                {
                InputObject * const object = *object_it;
                hope( object );
                if( object->is_input_enabled() )
                    {
                    eaten = dispatch_event_to_object( object, event );
                    if( eaten )
                        {
                        return eaten;
                        }
                    }
                }
            }
        }

    return eaten;
    }


bool InputManager::dispatch_event_to_object
    (
    InputObject *   object,
    const Event &   event
    )
    {
    hope( object );
    bool eaten = false;

    switch( event.type )
        {
        case Event::TYPE_NONE:
            break;

        case Event::TYPE_KEY_UP:
            eaten = object->key_up( event.key, event.mod );
            break;

        case Event::TYPE_KEY_DOWN:
            eaten = object->key_down( event.key, event.mod );
            break;

        case Event::TYPE_MOUSE_MOVE:
            eaten = object->mouse_move( event.dx, event.dy, event.mod );
            break;

        case Event::TYPE_MOUSE_BUTTON_UP:
            eaten = object->mouse_button_up( event.mouse_button, event.mod );
            break;

        case Event::TYPE_MOUSE_BUTTON_DOWN:
            eaten = object->mouse_button_down( event.mouse_button, event.mod );
            if( eaten )
                {
                // Grab the mouse.
                if( event.mouse_button == SDL_BUTTON_LEFT )
                    {
                    left_mouse_button_grab_object = object;
                    }
                if( event.mouse_button == SDL_BUTTON_RIGHT )
                    {
                    right_mouse_button_grab_object = object;
                    }
                }
            break;
        }

    return eaten;
    }


InputManager::MouseButton InputManager::sdl_to_native_mouse_button
    (
    int button
    ) const
    {
    switch( button )
        {
        case SDL_BUTTON_LEFT:
            return MOUSE_BUTTON_LEFT;

        case SDL_BUTTON_MIDDLE:
            return MOUSE_BUTTON_MIDDLE;

        case SDL_BUTTON_RIGHT:
            return MOUSE_BUTTON_RIGHT;

        case SDL_BUTTON_WHEELUP:
            return MOUSE_BUTTON_WHEEL_UP;

        case SDL_BUTTON_WHEELDOWN:
            return MOUSE_BUTTON_WHEEL_DOWN;

        default:
            return MOUSE_BUTTON_LAST;
        }
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "input/input_manager_inline.h"
#endif
