/** @file input_manager.h
 * Mouse and keyboard manager.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_INPUT_MANAGER_H
#define CLIENT_INPUT_MANAGER_H


#ifndef MASSIV_CORE_H
#include "core.h"
#endif

#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_INPUT_FWD_H
#include "input/input_fwd.h"
#endif

#ifndef CLIENT_SDL_H
#include "sdl/client_sdl.h"
#endif


#include <vector>
#include <map>


namespace Demo {
namespace Client {


/**
 * Keyboard and mouse manager.
 *
 * Input manager keeps state of keyboard and mouse buttons
 * and mouse position on screen.
 * It also dispatches input events to registered input objects.
 *
 * When an input event arrives, the manager calls corresponding
 * callback on registered input objects until one of the callbacks
 * returns true ie. until the event is processed by an input object.
 * Objects in lower input layers are handled first.
 *
 * Registered objects are not owned by the manager and should be unregistered
 * before the manager is destroyed.
 *
 * If rendering is globally disabled the entity manager accepts only single
 * input - press of the ctrl+r combination which enables the rendering again.
 */
class InputManager
    {
    NO_COPY_CONSTRUCTOR( InputManager );
    NO_ASSIGNMENT_OPERATOR( InputManager );

public:

    struct KeyState;
    struct MouseButtonState;

    /**
     * Constructor.
     */
    InputManager();

    /**
     * Destructor.
     *
     * All objects should be unregistered before destruction of the manager.
     * If some objects are still registered, they will be unregistered
     * automatically but warning will be logged.
     */
    ~InputManager();

public:

    /**
     * Get mouse position on screen in pixels.
     */
    SVector2 get_mouse_position() const;

    /**
     * Get mouse position on screen in pixels.
     */
    void get_mouse_position
        (
        int & x,
        int & y
        ) const;

    /**
     * Check whether specified key is down or up.
     * @param key Symbolic key name. One from @c SDLK_* enum values.
     */
    bool is_key_down
        (
        SDLKey key
        ) const;

    /**
     * Check whether specified mouse button is down or up.
     * @param mouse_button Mouse button index. One from @c SDL_BUTTON_* values.
     */
    bool is_mouse_button_down
        (
        int mouse_button
        ) const;

    /**
     * Return key state of a given key.
     */
    const KeyState & get_key_state
        (
        SDLKey key
        ) const;

    /**
     * Return key state of a given key.
     */
    const MouseButtonState & get_mouse_button_state
        (
        int mouse_button
        ) const;

public:

    //@{
    /**
     * Input callback called by @c SDL.
     */
    void key_up
        (
        const SDL_keysym & keysym
        );

    void key_down
        (
        const SDL_keysym & keysym
        );

    void mouse_move
        (
        int dx,
        int dy
        );
    //@}

    //@{
    /**
     * Input callback called by @c SDL.
     * @param button One of @c SDL_BUTTON_* values. Other values are ignored.
     */
    void mouse_button_up
        (
        int button
        );

    void mouse_button_down
        (
        int button
        );
    //@}

protected:

    /// #InputObject can register/unregister itself to/from the manager.
    friend class InputObject;

    /**
     * Register input object.
     */
    void register_object
        (
        InputObject *   object,
        int             layer
        );

    /**
     * Unregister input object.
     */
    void unregister_object
        (
        InputObject *   object,
        int             layer
        );

public:

    /**
     * Keyboard key state.
     */
    struct KeyState
        {
    public:
        KeyState() :
            down( false )
            {
            }

    public:
        bool        down;
            /**< Is the key up or down? */
        SystemTime  press_time;
            /**< When the key was pressed for the last time. */
        SystemTime  release_time;
            /**< When the key was released for the last time. */

        }; // KeyState

    /**
     * Mouse button state.
     */
    struct MouseButtonState
        {
    public:
        MouseButtonState() :
            down( false )
            {
            }

    public:
        bool        down;
            /**< Is the button up or down? */
        SystemTime  press_time;
            /**< When the button was pressed for the last time. */
        SystemTime  release_time;
            /**< When the button was released for the last time. */

        }; // MouseButtonState

private:

    /**
     * Symbolic names for mouse buttons. Only for internal purposes
     * for mapping @c SDL_BUTTON_* values to array indices.
     */
    enum MouseButton
        {
        MOUSE_BUTTON_LEFT         = 0,
        MOUSE_BUTTON_MIDDLE       = 1,
        MOUSE_BUTTON_RIGHT        = 2,
        MOUSE_BUTTON_WHEEL_UP     = 3,
        MOUSE_BUTTON_WHEEL_DOWN   = 4,
        MOUSE_BUTTON_LAST         = 5
        }; // MouseButton

    /**
     * Input event.
     */
    class Event
        {
    public:
        enum Type
            {
            TYPE_NONE,
            TYPE_KEY_UP,
            TYPE_KEY_DOWN,
            TYPE_MOUSE_MOVE,
            TYPE_MOUSE_BUTTON_UP,
            TYPE_MOUSE_BUTTON_DOWN
            }; // Type

    public:
        /** Constructor. */
        Event();

    public:
        //@{
        /**
         * Set event type and content for specified event.
         */
        void set_key_up
            (
            SDLKey              key,
            SDLMod              mod,
            const SystemTime &  release_time,
            const SystemTime &  press_time
            );

        void set_key_down
            (
            SDLKey              key,
            SDLMod              mod,
            const SystemTime &  release_time,
            const SystemTime &  press_time
            );

        void set_mouse_move
            (
            int                 dx,
            int                 dy,
            SDLMod              mod,
            const SystemTime &  time
            );

        void set_mouse_button_up
            (
            int                 mouse_button,
            SDLMod              mod,
            const SystemTime &  release_time,
            const SystemTime &  press_time
            );

        void set_mouse_button_down
            (
            int                 mouse_button,
            SDLMod              mod,
            const SystemTime &  release_time,
            const SystemTime &  press_time
            );
        //@}

    public:
        Type        type;
            /**< Type of the event. */
        SDLMod      mod;
            /**< Keyboard modifiers. */
        SDLKey      key;
            /**< Key which was pressed down or up. */
        int         mouse_button;
            /**< Mouse button which was pressed down or up. */
        int         dx;
            /**< Horizontal mouse movement delta. */
        int         dy;
            /**< Vertical mouse movement delta. */
        SystemTime  time;
            /**< Event time. */
        SystemTime  complement_time;
            /**< Time of a complementary event. For example if @c type is
                 @c TYPE_KEY_UP then @c complementary event is @c TYPE_KEY_DOWN. */

        }; // Event

    /**
     * Vector of #KeyState structures.
     */
    typedef std::vector< KeyState > Keys;

    /**
     * Vector of #MouseButtonState structires.
     */
    typedef std::vector< MouseButtonState > MouseButtons;

    /**
     * Vector of aliasing #InputObject pointers.
     */
    typedef std::vector< InputObject * > Objects;

    /**
     * Map from layer index to #Objects.
     */
    typedef std::map< int, Objects > Layers;

private:

    /**
     * Dispatch @a event.
     * Objects in lower input layers are processed first.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    bool dispatch_event
        (
        const Event & event
        );

    /**
     * Dispatch @a event to @a object.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    bool dispatch_event_to_object
        (
        InputObject *   object,
        const Event &   event
        );

    /**
     * Convert SDL symbolic mouse button name to #MouseButton name.
     * @param button One of @c SDL_BUTTON_* values.
     * @return #MouseButton name of the mouse button. If @a button
     *      is unknown symbolic mouse button name,@c MOUSE_BUTTON_LAST
     *      is returned.
     */
    MouseButton sdl_to_native_mouse_button
        (
        int button
        ) const;

private:

    Keys                        keys;
        /**< Current keyboard state. */

    MouseButtons                mouse_buttons;
        /**< Current mouse state. */

    Layers                      layers;
        /**< Layers with registered input objects. */

    InputObject *               left_mouse_button_grab_object;
        /**< Object that has eaten left mouse button down event. The object
         * will receive all mouse move events until left mouse button
         * is released.
         */
    InputObject *               right_mouse_button_grab_object;
        /**< Object that has eaten right mouse button down event. The object
         * will receive all mouse move events until right mouse button
         * is released.
         */

    }; // InputManager



} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "input/input_manager_inline.h"
#endif


#endif // CLIENT_INPUT_MANAGER_H
