/** @file input_object.h
 * Input object.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_INPUT_OBJECT_H
#define CLIENT_INPUT_OBJECT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_INPUT_FWD_H
#include "input/input_fwd.h"
#endif

#ifndef CLIENT_SDL_H
#include "sdl/client_sdl.h"
#endif


namespace Demo {
namespace Client {


/**
 * Input object is able to receive keyboard and mouse input events.
 * Input objects should be registered #InputManager to automatically
 * receive input events.
 *
 * Only enabled input object can receive input events.
 * By default, input object is created enabled.
 *
 */
class InputObject
    {
public:

    /**
     * Constructor.
     * The input object is automatically enabled.
     */
    INLINE InputObject();
    
    /**
     * Destructor.
     */
    virtual ~InputObject();

public:

    /**
     * Check whether the input object is enabled or disabler.
     * Only enabled input objects can receive input events.
     */
    INLINE bool is_input_enabled() const;
    
    /**
     * Enable the input object.
     */
    INLINE void enable_input();

    /**
     * Disable the input object.
     */
    INLINE void disable_input();
    
    /**
     * Enable or disable the input object.
     */
    INLINE void enable_disable_input
        (
        bool enable
        );

public:

    /**
     * Register the input object to input manager.
     *
     * You can re-register the object. Use @c NULL as @a new_manager
     * to unregister the object.
     */
    void register_to_input_manager
        (
        InputManager *  new_manager,
        int             new_layer
        );

    /**
     * Register the input object to global input manager.
     */
    void register_to_global_input_manager
        (
        int new_layer
        );

    /**
     * Unregister the input object from the manager.
     *
     * Shortcut for <tt>register_to_input_manager( NULL, 0 )</tt>.
     */
    INLINE void unregister_from_input_manager();

public:
    
    /**
     * Handle key up event.
     * Default implementation does nothing and returns false.
     * @param key Symbolic key name. One from @c SDLK_* enum values.
     * @param mod Keyboard modifiers. Combination of @c KMOD_* flags.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    virtual INLINE bool key_up
        (
        SDLKey  key,
        SDLMod  mod
        );
        
    /**
     * Handle key down event.
     * Default implementation does nothing and returns false.
     * @param key Symbolic key name. One from @c SDLK_* enum values.
     * @param mod Keyboard modifiers. Combination of @c KMOD_* flags.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    virtual INLINE bool key_down
        (
        SDLKey  key,
        SDLMod  mod
        );
    
    /**
     * Handle mouse button up event.
     * Default implementation does nothing and returns false.
     * @param button Mouse button index. One from @c SDL_BUTTON_* values.
     * @param mod Keyboard modifiers. Combination of @c KMOD_* flags.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    virtual INLINE bool mouse_button_up
        (
        int     button,
        SDLMod  mod
        );

    /**
     * Handle mouse button down event.
     * Default implementation does nothing and returns false.
     * @param button Mouse button index. One from @c SDL_BUTTON_* values.
     * @param mod Keyboard modifiers. Combination of @c KMOD_* flags.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    virtual INLINE bool mouse_button_down
        (
        int     button,
        SDLMod  mod
        );
    
    /**
     * Handle mouse movement event.
     * Default implementation does nothing and returns false.
     * @param dx Horizontal movement delta.
     * @param dy Vertical movement delta.
     * @param mod Keyboard modifiers. Combination of @c KMOD_* flags.
     * @return True if the event was processed (the event was "eaten").
     *      False otherwise.
     */
    virtual INLINE bool mouse_move
        (
        int    dx,
        int    dy,
        SDLMod mod
        );

private:

    bool                input_enabled;
        /**< Is the input object enabled or disabled? */
        
private:

    InputManager *      manager;
        /**< Manager this object is registered to. It may be NULL. */

    int                 layer;
        /**< Input layer index.
         * 0 is the layer which is processed first when new event arrives.
         * @c MAX_INT is the opposite layer.
         * Ignored unless registered to manager.
         */

    }; // InputObject


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "input/input_object_inline.h"
#endif

    
#endif // CLIENT_INPUT_OBJECT_H
