/** @file media_repository.h
 * Media repository.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MEDIA_REPOSITORY_H
#define CLIENT_MEDIA_REPOSITORY_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


namespace Demo {
namespace Client {


/**
 * MediaRepository hols templates for media objects.
 * The templates can be used to create new media objects.
 * Each template is identified by its name.
 */
class MediaRepository
    {
    NO_COPY_CONSTRUCTOR( MediaRepository );
    NO_ASSIGNMENT_OPERATOR( MediaRepository );
    
public:

    /**
     * Constructor.
     */
    MediaRepository();
    
    /**
     * Initialize.
     */
    void initialize();
    
public:

    /**
     * Create new icon from icon template.
     * Ownership of the returned icon is transferred to the caller.
     */
    Renderer::Icon * create_icon
        (
        const std::string & icon_template_name
        );
        
public:
    
    /**
     * IconTemplate holds infomation about one icon.
     */
    class IconTemplate
        {
    public:
    
        /** Constructor. */
        INLINE IconTemplate();
        
        /** Constructor. */
        INLINE IconTemplate
            (
            const std::string & texture_name,
            Float32             u1,
            Float32             v1,
            Float32             u2,
            Float32             v2
            );
        
    public:
        std::string     texture_name;
            /**< Icon texture name. */
            
        SVector2        top_left_uv;
            /**< Icon top left texture coordinates. */
            
        SVector2        bottom_right_uv;
            /**< Icon bottom right texture coordinates. */

        }; // IconTemplate
    
    /**
     * Hash of icon templates in the repository.
     */
    typedef ext::hash_map
        <
        std::string,
        IconTemplate
        > IconTemplates;

public:

    /**
     * Register new icon template to the repository.
     */
    void register_icon_template
        (
        const std::string & icon_template_name,
        const IconTemplate & icon_template
        );

private:

    /**
     * Throw exception if a media object is not found in the repository.
     */
    void error_media_not_registered
        (
        const std::string & media_type,
        const std::string & media_name
        );

    /**
     * Throw exception if a media object is already registered in the repository.
     */
    void error_media_already_registered
        (
        const std::string & media_type,
        const std::string & media_name
        );

private:

    IconTemplates   icon_templates;
        /**< Registered icon templates. */

    }; // MediaRepository


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "media_repository_inline.h"
#endif


#endif // CLIENT_MEDIA_REPOSITORY_H
