/** @file raycast.cpp
 * Generic raycasting (implementation).
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "raycast.h"


namespace Demo {
namespace Client {
namespace Raycast {


bool intersect_triangle
    (
    Float32 &           distance,
    SVector3 &          normal,
    const SVector3 &    origin,
    const SVector3 &    direction,
    const SVector3 &    v0,
    const SVector3 &    v1,
    const SVector3 &    v2
    )
    {
    // Check if the ray lies inside cone specified by @a origin and edges of the triangle.
    // Note that this test will always fail if triangle is backfacing.

    if( ( ( ( v1 - origin ) % ( v0 - origin ) ) * direction < 0.0f ) ||
        ( ( ( v2 - origin ) % ( v1 - origin ) ) * direction < 0.0f ) ||
        ( ( ( v0 - origin ) % ( v2 - origin ) ) * direction < 0.0f ) )
        {
        return false;
        }

    // Compute point of intersection of the ray with triangle plane.

    // Plane equation: n * x + d = 0.
    // Note: normalization is NOT required, but we could have problems with
    // small triangles otherwise (coplanar test below could skip the triangle).
    
    SVector3 n = ( ( v1 - v0 ) % ( v2 - v0 ) );
    if( n.epsilon_zero() )
        {
        // Degenerated triangle.

        return false;
        }
    n.normalize();
    const Float32 d = -( n * v0 );

    // 1a. intersection * n + d = 0
    // 1b. intersection = origin + direction * distance
    //  2. ( origin + direction * distance ) * n + d = 0
    //  3. distance = -( d + origin * n ) / ( direction * n )

    const Float32 direction_dot_n = direction * n;
    if( ::Massiv::Core::FloatTraits< Float32 >::epsilon_zero( direction_dot_n ) )
        {
        // If ( direction * n ) == 0, the ray and the plane are coplanar.
        // This should happen only in really weird cases, usually the first
        // test above rejects the ray.

        return false;
        }
    else
        {
        distance = -( d + origin * n ) / ( direction_dot_n );
        normal = n;
        }

    // Ignore intersection if it's "behind" the origin.

    return ( distance >= 0.0f );
    }


bool intersect_triangle_with_normal
    (
    Float32 &           distance,
    const SVector3 &    origin,
    const SVector3 &    direction,
    const SVector3 &    v0,
    const SVector3 &    v1,
    const SVector3 &    v2,
    const SVector3 &    normal
    )
    {
    // Check if the ray lies inside cone specified by @a origin and edges of the triangle.
    // Note that this test will always fail if triangle is backfacing.

    if( ( ( ( v1 - origin ) % ( v0 - origin ) ) * direction < 0.0f ) ||
        ( ( ( v2 - origin ) % ( v1 - origin ) ) * direction < 0.0f ) ||
        ( ( ( v0 - origin ) % ( v2 - origin ) ) * direction < 0.0f ) )
        {
        return false;
        }

    // Compute point of intersection of the ray with triangle plane.

    // Plane equation: normal * x + d = 0.
    // Warning: if normal is really short, the coplanar test below could reject the triangle.

    const Float32 d = -( normal * v0 );

    const Float32 direction_dot_normal = direction * normal;
    if( ::Massiv::Core::FloatTraits< Float32 >::epsilon_zero( direction_dot_normal ) )
        {
        // The ray and the plane are coplanar.

        return false;
        }
    else
        {
        distance = -( d + origin * normal ) / ( direction_dot_normal );
        }

    // Ignore intersection if it's "behind" the origin.

    return ( distance >= 0.0f );
    }


} // namespace Demo::Client::Raycast
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "raycast_inline.h"
#endif
