/** @file camera.cpp
 * 3D camera (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/camera.h"
#include "renderer/opengl.h"
#include "sdl/client_sdl.h"


namespace Demo {
namespace Client {
namespace Renderer {


// class Camera

Camera::Camera()
      : projection( PERSPECTIVE ),
        fov( MATHF_PI_2 ),
        zoom( 16.0f )
    {
    }


Camera::~Camera()
    {
    }


void Camera::cursor_ray
    (
    SVector3 &  origin,
    SVector3 &  direction,
    int         x_position,
    int         y_position
    ) const
    {
    SVector3 position, right, forward, up;
    get_camera_transformation( position, right, forward, up );
    const Float32 screen_x_size = static_cast< Float32 >( SDL::get_screen_width() );
    const Float32 screen_y_size = static_cast< Float32 >( SDL::get_screen_height() );
    
    if( projection == PERSPECTIVE )
        {
        // Origin at camera position.

        origin = position;

        // Compute direction point at imaginary view plane at unit distance from camera.

        const Float32 projective_zoom = 2.0f *
            ::Massiv::Core::FloatTraits< Float32 >::tan( 0.5f * fov );
        const Float32 x = projective_zoom *
            ( static_cast< Float32 >( x_position ) / screen_x_size - 0.5f );
        const Float32 y = projective_zoom * screen_y_size / screen_x_size *
            ( static_cast< Float32 >( y_position ) / screen_y_size - 0.5f );
        const SVector3 point = position + forward + right * x - up * y;
        direction = point - position;
        }
    else
        {
        hope( projection == PARALLEL );

        // Compute origin.

        const Float32 x = zoom *
            ( static_cast< Float32 >( x_position ) / screen_x_size - 0.5f );
        const Float32 y = zoom * screen_y_size / screen_x_size *
            ( static_cast< Float32 >( y_position ) / screen_y_size - 0.5f );
        origin = position + right * x - up * y;

        // Direction is camera forward vector.

        direction = forward;
        }
    }


void Camera::apply_projection() const
    {
    // FIXME: these numbers are "randomly" chosen.

    static const GLdouble near_distance = 0.5;
    static const GLdouble far_distance = 1000.0;

    // Compute height-width aspect ratio.
    // FIXME: we assume pixels are square.

    const GLdouble aspect_ratio =
        static_cast< GLdouble >( SDL::get_screen_height() ) /
        static_cast< GLdouble >( SDL::get_screen_width() );

    // Set projection.
    
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    if( projection == PERSPECTIVE )
        {
        // Set frustum parameters.

        using namespace ::Massiv::Core;
        const GLdouble projective_zoom = FloatTraits< GLdouble >::tan( fov / 2.0 );
        glFrustum
            (
            -near_distance * projective_zoom,
             near_distance * projective_zoom,
            -near_distance * aspect_ratio * projective_zoom,
             near_distance * aspect_ratio * projective_zoom,
             near_distance,
             far_distance
            );
        }
    else
        {
        // Set ortho parameters.

        hope( projection == PARALLEL );
        const GLdouble ortho_zoom = zoom / 2.0;
        glOrtho
            (
            -ortho_zoom,
             ortho_zoom,
            -aspect_ratio * ortho_zoom,
             aspect_ratio * ortho_zoom,
            -far_distance,
             far_distance
            );
        }

    // Done.

    glMatrixMode( GL_MODELVIEW );
    }


void Camera::apply_transformation
    (
    bool translate
    ) const
    {
    // Assume matrix mode is MODELVIEW...

    glPushMatrix();

    // Note that the first specified transformation will be applied to the
    // objects as the last one, so read this code from bottom to top.

    // Object is is our standard camera-space now: X to the right, Y forward
    // and Z up. OpenGL camera-space is X right, Y up, Z backward. This
    // rotation is equal to the axis flipping.

    glRotatef( -90.0f, 1.0f, 0.0f, 0.0f );

    // Apply camerat transformation to the object. Modelview matrix transforms
    // the object, not the camera, so we must use inverse transformation:
    // - Translate the object by -camera_position.
    // - Apply inverse of camera rotation to the object.

    SVector3 position, right, forward, up;
    get_camera_transformation( position, right, forward, up );

    // Place base vectors into matrix rows. Note that OpenGL uses column-major
    // ordering of matrix elements, so it looks transposed in the source.

    const GLfloat matrix[ 16 ] =
        {
        right.x(), forward.x(), up.x(), 0.0f,
        right.y(), forward.y(), up.y(), 0.0f,
        right.z(), forward.z(), up.z(), 0.0f,
             0.0f,        0.0f,   0.0f, 1.0f
        };
    glMultMatrixf( matrix );

    // Inverse translation.

    if( translate )
        {
        glTranslatef( -position.x(), -position.y(), -position.z() );
        }
    }


void Camera::revert_transformation() const
    {
    glPopMatrix();
    }


// class CirclingCamera

CirclingCamera::CirclingCamera()
      : latitude( 0.0f ),
        longitude( 0.0f ),
        distance( 1.0f )
    {
    }


CirclingCamera::~CirclingCamera()
    {
    }


void CirclingCamera::get_camera_transformation
    (
    SVector3 &  position,
    SVector3 &  right,
    SVector3 &  forward,
    SVector3 &  up
    ) const
    {
    // Compute orientation.

    SOrientation orientation( longitude, -latitude, 0.0f );
    orientation.rotation_base( right, forward, up );

    // Compute position.

    position = target - forward * distance;
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "camera_inline.h"
#endif
