/** @file camera.h
 * 3D camera.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_CAMERA_H
#define CLIENT_CAMERA_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Generic camera interface.
 *
 * Each camera may use either perspective (default) or parallel projection.
 * When using perspective projection, horizontal field of view (angle in
 * radians) can be specified. Default is PI/2 (90 degrees). Zoom value (for
 * parallel projection, default is 16.0) specifies how many world-units can
 * fit horizontally on the screen.
 *
 * Classes that inherit #Camera should define API for position and orientation
 * changes and implement #get_camera_transformation().
 */
class Camera
    {
    NO_COPY_CONSTRUCTOR( Camera );
    NO_ASSIGNMENT_OPERATOR( Camera );

public:

    /**
     * Camera projection mode.
     */
    enum Projection
        {
        PERSPECTIVE,
        PARALLEL
        };

public:

    /**
     * Constructor.
     */
    Camera();

    /**
     * Destructor.
     */
    virtual ~Camera();

public:

    //@{
    /** Getter. */
    INLINE Projection get_projection() const;

    INLINE Float32 get_fov() const;

    INLINE Float32 get_zoom() const;
    //@}

    //@{
    /** Setter. */
    INLINE void set_projection
        (
        Projection new_projection
        );

    INLINE void set_fov
        (
        Float32 new_fov
        );

    INLINE void set_zoom
        (
        Float32 new_zoom
        );
    //@}

public:

    /**
     * Compute camera position and orientation.
     *
     * This method computes camera position (@a position) and orientation
     * (orthogonal normalized vectors @a right @a forward and @a up) in
     * world-space. See src/core/property/orientation.h for more info about
     * standard world-space coordinate system.
     */
    virtual void get_camera_transformation
        (
        SVector3 &  position,
        SVector3 &  right,
        SVector3 &  forward,
        SVector3 &  up
        ) const = 0;

    /**
     * Compute ray @a origin and @a direction from cursor screen-space position.
     *
     * @note @a direction will not be normalized.
     */
    void cursor_ray
        (
        SVector3 &  origin,
        SVector3 &  direction,
        int         x_position,
        int         y_position
        ) const;

public:

    /**
     * Apply camera projection settings to OpenGL projection matrix.
     *
     * Current projection matrix will be replaced by the new one. Matrix
     * mode will be changed back to modelview on return.
     */
    void apply_projection() const;

    /**
     * Apply camera transformation settings to OpenGL modelview matrix.
     *
     * This transformation will transform object from world-space to
     * camera-space. If @a translate if @c false, zero vector will be used
     * instead of real camera translation. This can be used to render objects
     * placed in "infinity" (such as sky backdrop).
     *
     * Then rotation transforming object from our standard camera-space
     * coordinate system to OpenGL eye-space coordinate system will be applied.
     *
     * This method assumes that current matrix mode is set to modelview.
     * Current modelview matrix will be pushed. Use #revert_transformation()
     * to pop it back.
     */
    void apply_transformation
        (
        bool translate = true
        ) const;

    /**
     * Pop matrix from modelview matrix stack.
     */
    void revert_transformation() const;

private:

    Projection  projection;
        /**< Current projection mode. */

    Float32     fov;
        /**< Field of view in radians. */

    Float32     zoom;
        /**< Parallel zoom. */

    }; // class Camera


/**
 * Camera circling aroung given target.
 *
 * This camera always looks at given @a target from point at a circle specified
 * by @a latitude, @a longitude and @a distance.
 *
 * If @a latitude is 0.0, camera will at the same level as the target, if
 * @a latitude is PI / 2, camera will be directly above the target and if
 * @a latitude is -PI / 2, camera will be directly below the target.
 *
 * If @a longitude is 0.0, camera will be looking  to the north (along the
 * @a y axis), if it's PI / 2, camera will be looking to the west,
 * PI -\> to the south, 3 * PI / 2 (or -PI / 2) -\> to the east.
 *
 * @a Distance should be positive number. If it's negative, the camera will
 * be looking away from the target and its position will be the same as if
 * ( - @a latitude, PI + @a longitude, - @a distance ) parameters were used.
 */
class CirclingCamera : public Camera
    {
    NO_COPY_CONSTRUCTOR( CirclingCamera );
    NO_ASSIGNMENT_OPERATOR( CirclingCamera );

public:

    /**
     * Constructor.
     */
    CirclingCamera();

    /**
     * Destructor.
     */
    virtual ~CirclingCamera();

public:

    /**
     */
    virtual void get_camera_transformation
        (
        SVector3 &  position,
        SVector3 &  right,
        SVector3 &  forward,
        SVector3 &  up
        ) const;

public:

    //@{
    /** Getter. */
    INLINE const SVector3 & get_target() const;

    INLINE Float32 get_latitude() const;

    INLINE Float32 get_longitude() const;

    INLINE Float32 get_distance() const;
    //@}

    //@{
    /** Setter. */
    INLINE void set_target
        (
        const SVector3 & new_target
        );

    INLINE void set_latitude
        (
        Float32 new_latitude
        );

    INLINE void set_longitude
        (
        Float32 new_longitude
        );

    INLINE void set_distance
        (
        Float32 new_distance
        );
    //@}

private:

    SVector3    target;
        /**< Target position. */

    Float32     latitude;
        /**< Latitude in radians. */

    Float32     longitude;
        /**< Longitude in radians. */

    Float32     distance;
        /**< Distance from the #target. */

    }; // class CirclingCamera


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "camera_inline.h"
#endif


#endif // CLIENT_CAMERA_H
