/** @file font.cpp
 * Fixed-width font (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/font.h"
#include "global/global.h"
#include "utility/text_reader.h"
#include "data/data_manager.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"


namespace Demo {
namespace Client {
namespace Renderer {


Font::Font
    (
    const std::string & name
    ) : font_version( 0 ),
        texture_version( 0 ),
        texture_rows( 0 ),
        texture_columns( 0 ),
        first_character( 0 ),
        character_count( 0 ),
        texture_width( 0 ),
        texture_height( 0 ),
        character_width( 0 ),
        character_height( 0 )
    {
    // Get data object for the font.
    
    Massiv::Core::DataManager & data_manager = Massiv::Core::Global::data_manager();
    font_data_object.reset( data_manager.get_data_object( name ) );

    // Load the font.

    load();
    hope( font_version == 1 );
    }


Font::~Font()
    {
    }


TextureDataObject::Reference Font::get_texture()
    {
    // Check update of font description.

    hope( font_data_object.get() );
    if( font_data_object->is_data_updated() )
        {
        // Reload description. This always bumps version and refreshes texture.

        load();
        }
    else
        {
        // Check if texture has been updated.

        refresh();
        }

    // Tell data manager to download new version if available.
    // FIXME: this function probably does something else...
    font_data_object->download_actual_version();

    hope( !texture_data_object.is_null() );
    return texture_data_object;
    }


void Font::get_uv
    (
    char        c,
    GLfloat &   top_left_u,
    GLfloat &   top_left_v,
    GLfloat &   bottom_right_u,
    GLfloat &   bottom_right_v
    ) const
    {
    if( !is_valid( c ) )
        {
        top_left_u = top_left_v = bottom_right_u = bottom_right_v = 0.0f;
        }
    else
        {
        const CharInfo & info =
            character_infos[ static_cast< unsigned char >( c ) - first_character ];
        hope( info.valid );
        const int row = info.row;
        const int column = info.column;
        const GLfloat width = static_cast< GLfloat >( character_width ) / texture_width;
        const GLfloat height = static_cast< GLfloat >( character_height ) / texture_height;
        
        top_left_u = column * width;
        top_left_v = row * height;
        bottom_right_u = ( column + 1 ) * width;
        bottom_right_v = ( row + 1 ) * height;
        }
    }


void Font::load()
    {
    using namespace Massiv::Core;

    hope( font_data_object.get() );
    font_data_object->refresh();
    font_data_object->open();

    std::vector< std::string > strings;
    int min_char = 255;
    int max_char = 0;
    try
        {
        // Parse font description.

        TextReader tr( font_data_object->data );
        std::string token;

        // Tag.

        tr.next_line_throw();
        tr.require_string( "font" );

        // Texture data object name.

        tr.next_line_throw();
        token = tr.read_quote_string();
        texture_data_object = Global::get_texture_manager().get_texture_data_object( token );
        hope( !texture_data_object.is_null() );
        texture_version = 0; // to force refresh

        // Number of columns and rows.

        tr.next_line_throw();
        texture_columns = tr.read_int();
        if( ( texture_columns < 1 ) || ( texture_columns > 255 ) )
            {
            throw ExUnexpectedValue( tr.unexpected_value
                ( "integer in range < 1, 255 >", texture_columns ) );
            }
        tr.next_line_throw();
        texture_rows = tr.read_int();
        if( ( texture_rows < 1 ) || ( texture_rows > 255 ) )
            {
            throw ExUnexpectedValue( tr.unexpected_value
                ( "integer in range < 1, 255 >", texture_rows ) );
            }

        // Allocate temporary strings.

        strings.resize( texture_rows );

        // Read character->cell mappings. Determine range of used characters.

        int y;
        for( y = 0; y < texture_rows; y++ )
            {
            tr.next_line_throw();
            strings[ y ] = std::string( tr.str(), 0, texture_columns );
            std::string::size_type x;
            for( x = 0; x < strings[ y ].size(); x++ )
                {
                const int c = static_cast< unsigned char >( strings[ y ][ x ] );
                hope( c >= 0 && c <= 255 );
                if( c == ' ' )
                    {
                    continue;
                    }
                if( c < min_char )
                    {
                    min_char = c;
                    }
                if( c > max_char )
                    {
                    max_char = c;
                    }
                }
            }
        }
    catch( ... )
        {
        font_data_object->close();
        throw;
        }

    font_data_object->close();

    // Initialize character infos.

    character_infos.clear();
    if( min_char >= max_char )
        {
        // Weird, no characters. Log this?

        first_character = 0;
        character_count = 0;
        }
    else
        {
        first_character = min_char;
        character_count = max_char - min_char + 1;
        character_infos.resize( character_count );

        int y;
        for( y = 0; y < texture_rows; y++ )
            {
            std::string::size_type x;
            for( x = 0; x < strings[ y ].size(); x++ )
                {
                const int c = static_cast< unsigned char >( strings[ y ][ x ] );
                if( c == ' ' )
                    {
                    continue;
                    }
                const int index = c - first_character;
                hope( index >= 0 );
                hope( index < character_count );
                character_infos[ index ].valid = true;
                character_infos[ index ].row = static_cast< UInt8 >( y );
                character_infos[ index ].column = static_cast< UInt8 >( x );
                }
            }
        }

    // Refresh texture info. This increases font version.

    hope( texture_version == 0 );
    refresh();
    hope( texture_version > 0 );
    }


void Font::refresh()
    {
    hope( !texture_data_object.is_null() );
    if( texture_version != texture_data_object->get_version() )
        {
        hope( texture_version < texture_data_object->get_version() );

        // Get texture info.

        Texture::Reference texture = texture_data_object->get_texture( get_texture_flags() );
        texture_width = texture->get_width();
        texture_height = texture->get_height();
        character_width = texture->get_valid_width() / texture_columns;
        character_height = texture->get_valid_height() / texture_rows;
        
        // Update version.

        texture_version = texture_data_object->get_version();
        font_version++;
        }
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "font_inline.h"
#endif
