/** @file font.h
 * Fixed-width font.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_FONT_H
#define CLIENT_FONT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_H
#include "renderer/opengl.h"
#endif

#ifndef CLIENT_TEXTURE_DATA_OBJECT_H
#include "renderer/texture_data_object.h"
#endif

#include <vector>


namespace Massiv {
namespace Core {

class DataObject;

}
}


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Fixed-width font.
 *
 * Object containing font information and font image. It can be used to
 * print to #TextRectangle object for example.
 *
 * When rendering the font, you should assume that its RGB is premultiplied
 * by alpha (Because intensity is converted to alpha if font texture has no
 * alpha channel, images containing alpha channel should premultiply RGB
 * too. This may change in the future).
 *
 * Because Font stores its data in data object, its contents and attributes
 * may change at "any time". Each method that may modify font attributes
 * documents this.
 *
 * "Immediate" usage of #Font should look like this:
 * - Get texture data object using #get_texture().
 * - Get required information about the font using various getters.
 * - Draw something using the texture data object, use #get_uv() to retreive
 *   texture coordinates of individual characters.
 *
 * Vertex array usage of #Font should look like this:
 * - Create vertex array and index array with character cells.
 * - Remember font version, initialize it to 0.
 * - Whenever cell contents change, set version back to 0.
 * - Whenever you draw the text:
 *   - Get texture data object using #get_texture().
 *   - Check font version, and if it's different from the version stored
 *     with the array, recompute array texture coordinates. You may have
 *     to recompute vertex coordinates if character sizes change.
 *     Remember new version.
 *   - Draw the array.
 *
 * In reality you should use classes such as #TextRectangle instead of
 * rendering the font characters yourself.
 */
class Font
    {
    NO_COPY_CONSTRUCTOR( Font );
    NO_ASSIGNMENT_OPERATOR( Font );
    
private:

    /**
     * Information about single character.
     */
    class CharInfo
        {
    public:

        /** Default constructor. */
        INLINE CharInfo();

    public:

        bool    valid;
            /**< Is this character valid? */

        UInt8   row;
            /**< Texture row. */

        UInt8   column;
            /**< Texture column. */

        }; // class CharInfo

public:

    /**
     * Constructor.
     *
     * @param name Font data object name.
     */
    Font
        (
        const std::string & name
        );

    /**
     * Destructor.
     */
    ~Font();

public:

    /**
     * Get font texture data object.
     *
     * If update of font description or font texture is detected, font version
     * will be increased - any change to font attributes is possible in that
     * case.
     */
    TextureDataObject::Reference get_texture();

    /**
     * Get flags that should be passed to #TextureDataObject::get_texture().
     */
    static INLINE int get_texture_flags();

    /**
     * Get font version.
     *
     * Font version is an integer value starting at 1, which is increased
     * whenever font attributes are changed because of font description
     * or font texture update.
     */
    INLINE int get_version() const;

public:

    /**
     * Check if character @a c is valid - ie. it has image data.
     */
    INLINE bool is_valid
        (
        char        c
        ) const;

    /**
     * Get character texture coordinates.
     *
     * All coordinates will be set to 0 if character is invalid. Note that
     * it's not generally safe to draw such character and expect "space" to
     * be drawn.
     *
     * This method does not check data updates, it may return bogus
     * coordinates unless #get_texture() has been called in the same
     * data manager update tick.
     */
    void get_uv
        (
        char        c,
        GLfloat &   top_left_u,
        GLfloat &   top_left_v,
        GLfloat &   bottom_right_u,
        GLfloat &   bottom_right_v
        ) const;

public:

    /**
     * Get character width in texels.
     */
    INLINE int get_character_width();

    /**
     * Get character height in texels.
     */
    INLINE int get_character_height();

private:

    /**
     * (Re)load font.
     *
     * Always reloads font description. Reloads texture data if it has been
     * updated. Font version will be always increased.
     */
    void load();

    /**
     * Refresh information from texture data object and increase font version
     * if texture version has changed.
     */
    void refresh();

private:

    std::auto_ptr< Massiv::Core::DataObject >   font_data_object;
        /**< Font description data object. */

    int                                         font_version;
        /**< Font version.
             Increased whenever font description or font texture is updated. */

    TextureDataObject::Reference                texture_data_object;
        /**< Texture data object. */

    int                                         texture_version;
        /**< Version of #texture_data_object. */

// info "cached" from description
private:

    int         texture_rows;
        /**< Number of rows in font texture. */

    int         texture_columns;
        /**< Number of columns in font texture. */

    std::vector< CharInfo > character_infos;
        /**< Information about characters. */

    int         first_character;
        /**< ASCII code of first character info. */

    int         character_count;
        /**< Size of #character_infos. */

// info "cached" from texture
private:

    int         texture_width;
        /**< Texture width in texels.
             This includes invalid region if image has been extended. */

    int         texture_height;
        /**< Texture height in texels.
             This includes invalid region if image has been extended. */

    int         character_width;
        /**< Character width in texels. */

    int         character_height;
        /**< Character height in texels. */

    }; // class Font


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "font_inline.h"
#endif


#endif // CLIENT_FONT_H
