/** @file light.h
 * Light source.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_LIGHT_H
#define CLIENT_LIGHT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_PIPELINE_STATE_H
#include "renderer/pipeline_state.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Light source.
 *
 * Light sources can be either directional at infinite distance or omnidirectional.
 * Set #infinity to @c true (default) to make light directional. #Position specifies
 * vector from world origin to the light (vector opposite to the light direction)
 * if light is directional.
 *
 * Position is specified in world coordinates. Default position is (0, -1, 0).
 * Default value ambient intensity is "black" (0, 0, 0).
 * Default values of both diffuse and specular intensity are "white" (1, 1, 1).
 * Default value of attenuation is (1, 0, 0).
 *
 * See PipelineState::Lighting::Light and OpenGL documentation for more information
 * about vertex lighting.
 */
class Light
    {
    NO_COPY_CONSTRUCTOR( Light );
    NO_ASSIGNMENT_OPERATOR( Light );

public:

    /**
     * Constructor.
     */
    Light();

    /**
     * Destructor.
     */
    ~Light();

public:

    //@{
    /** Getter. */
    INLINE const SVector3 & get_position() const;

    INLINE bool is_infinite() const;

    INLINE const SColor & get_ambient_intensity() const;

    INLINE const SColor & get_diffuse_intensity() const;

    INLINE const SColor & get_specular_intensity() const;

    INLINE const SVector3 & get_attenuation() const;
    //@}

    //@{
    /** Setter. */
    INLINE void set_position
        (
        const SVector3 & new_position
        );

    INLINE void set_infinite
        (
        bool new_infinite
        );

    INLINE void set_ambient_intensity
        (
        const SColor & new_ambient_intensity
        );

    INLINE void set_diffuse_intensity
        (
        const SColor & new_diffuse_intensity
        );

    INLINE void set_specular_intensity
        (
        const SColor & new_diffuse_intensity
        );

    INLINE void set_attenuation
        (
        const SVector3 & new_attenuation
        );
    //@}
     
public:

    /**
     * Register light to render manager.
     *
     * It's allowed to register already registered light. It will be
     * unregistered from current manager first.
     */
    void register_to_manager
        (
        RenderManager * new_manager
        );

    /**
     * Register light to global render manager owned by pass manager.
     *
     * RenderManager must already exist for the pass. It will NOT be created
     * automatically.
     */
    void register_to_global_manager
        (
        int pass_index
        );

    /**
     * Unregister light from the manager.
     *
     * Shortcut for <tt>register_to_manager( NULL )</tt>.
     */
    INLINE void unregister_from_manager();

public:

    /**
     * Convert Light to PipelineState::Lighting::Light (enabled).
     *
     * If @a entity is not @c NULL, light will be transformed by inverse
     * transformation of the @a entity - it's coordinates will be in
     * object-space of the @a entity.
     */
    void set_pipeline_light
        (
        PipelineState::Lighting::Light &    light,
        const RenderEntity *                entity
        ) const;

private:

    SVector3        position;
        /**< Light position. */

    bool            infinite;
        /**< If @c true, light's position is in inifinity in the direction
             specified by @a position. */

    SColor          ambient_intensity;
        /**< Ambient intensity of the light. */

    SColor          diffuse_intensity;
        /**< Diffuse intensity of the light. */

    SColor          specular_intensity;
        /**< Specular intensity of the light. */

    SVector3        attenuation;
        /**< Light intensity attenuation factors. */

private:

    RenderManager * manager;
        /**< Manager this light is registered to. It may be NULL. */

    }; // class Light


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "light_inline.h"
#endif


#endif // CLIENT_LIGHT_H
