/** @file mesh.cpp
 * 3D triangle mesh (implementation).
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/mesh.h"
#include "renderer/vertex_array.h"
#include "raycast.h"


namespace Demo {
namespace Client {
namespace Renderer {


namespace {


/**
 * Resize vector of owning pointers to vectors and allocate/delete the inner vectors.
 *
 * @param streams Vector to resize.
 * @param stream_count New size of @a streams.
 * @param element_count Number of elements in each stream.
 */
template< typename StreamType >
void resize_streams
    (
    std::vector< StreamType * > &   streams,
    std::size_t                     stream_count,
    std::size_t                     element_count
    )
    {
    const std::size_t size = streams.size();
    if( size < stream_count )
        {
        // Allocate new streams.

        streams.resize( stream_count );
        std::size_t i;
        for( i = size; i < stream_count; i++ )
            {
            streams[ i ] = new StreamType;
            streams[ i ]->resize( element_count );
            }
        }
    else if( size > stream_count )
        {
        // Destroy some streams.

        std::size_t i;
        for( i = stream_count; i < size; i++ )
            {
            delete streams[ i ];
            streams[ i ] = NULL;
            }
        streams.resize( stream_count );
        }
    }


/**
 * Resize all vectors owned by given vector.
 *
 * @param streams Owning vector.
 * @param element_count Number of elements in each stream.
 */
template< typename StreamType >
void resize_elements
    (
    std::vector< StreamType * > &   streams,
    std::size_t                     element_count
    )
    {
    std::size_t i;
    for( i = 0; i < streams.size(); i++ )
        {
        hope( streams[ i ] );
        streams[ i ]->resize( element_count );
        }
    }

    
} // anonymous namespace


Mesh::Mesh()
      : vertex_count( 0 ),
        triangle_count( 0 )
    {
    // Allocate single empty vertex stream.

    set_vertex_stream_count( 1 );
    }


Mesh::~Mesh()
    {
    // Destroy all streams.

    set_vertex_stream_count( 0 );
    set_normal_stream_count( 0 );
    set_texcoord_stream_count( 0 );
    set_color_stream_count( 0 );
    }


void Mesh::clear()
    {
    // Destroy all streams, single vertex stream.

    set_vertex_stream_count( 1 );
    set_normal_stream_count( 0 );
    set_texcoord_stream_count( 0 );
    set_color_stream_count( 0 );

    // Zero vertices/triangle.

    set_vertex_count( 0 );
    set_triangle_count( 0 );
    }


void Mesh::set_vertex_stream_count
    (
    std::size_t count
    )
    {
    resize_streams< Vectors >( vertex_coordinates, count, vertex_count );
    resize_streams< Vectors >( triangle_normals, count, triangle_count );
    }


void Mesh::set_normal_stream_count
    (
    std::size_t count
    )
    {
    resize_streams< Vectors >( vertex_normals, count, vertex_count );
    }


void Mesh::set_texcoord_stream_count
    (
    std::size_t count
    )
    {
    resize_streams< TexCoords >( texture_coordinates, count, vertex_count );
    }


void Mesh::set_color_stream_count
    (
    std::size_t count
    )
    {
    resize_streams< Colors >( vertex_colors, count, vertex_count );
    }


void Mesh::set_vertex_count
    (
    std::size_t count
    )
    {
    vertex_count = count;
    resize_elements< Vectors >( vertex_coordinates, vertex_count );
    resize_elements< Vectors >( vertex_normals, vertex_count );
    resize_elements< TexCoords >( texture_coordinates, vertex_count );
    resize_elements< Colors >( vertex_colors, vertex_count );
    }


void Mesh::set_triangle_count
    (
    std::size_t count
    )
    {
    triangle_count = count;
    triangles.resize( triangle_count );
    resize_elements< Vectors >( triangle_normals, triangle_count );
    }


void Mesh::update_triangle_normals
    (
    std::size_t stream_index
    )
    {
    hope( stream_index < vertex_coordinates.size() );
    hope( triangle_normals.size() == vertex_coordinates.size() );
    hope( vertex_coordinates[ stream_index ] );
    hope( triangle_normals[ stream_index ] );
    const Vectors & vertices = *vertex_coordinates[ stream_index ];
    Vectors & normals = *triangle_normals[ stream_index ];
    hope( normals.size() == triangles.size() );

    Triangles::const_iterator t_it;
    Vectors::iterator n_it;
    for
        (
        t_it = triangles.begin(), n_it = normals.begin();
        t_it != triangles.end();
        ++t_it, ++n_it )
        {
        hope( t_it->v0 < vertices.size() );
        hope( t_it->v1 < vertices.size() );
        hope( t_it->v2 < vertices.size() );
        const SVector3 normal = ( ( vertices[ t_it->v1 ] - vertices[ t_it->v0 ] ) %
                                  ( vertices[ t_it->v2 ] - vertices[ t_it->v0 ] ) );
        if( normal.epsilon_zero() )
            {
            // Degenerated triangle.

            n_it->set_zero();
            }
        else
            {
            *n_it = normal;
            n_it->normalize();
            }
        }
    }


void Mesh::update_normals
    (
    std::size_t vertex_stream_index,
    std::size_t normal_stream_index
    )
    {
    // Update triangle normals.

    update_triangle_normals( vertex_stream_index );

    hope( vertex_stream_index < triangle_normals.size() );
    hope( normal_stream_index < vertex_normals.size() );
    hope( triangle_normals[ vertex_stream_index ] );
    hope( vertex_normals[ normal_stream_index ] );
    const Vectors & tn = *triangle_normals[ vertex_stream_index ];
    Vectors & vn = *vertex_normals[ normal_stream_index ];
    hope( tn.size() == triangles.size() );

    // Reset vertex normals.

    Vectors::iterator vn_it;
    for( vn_it = vn.begin(); vn_it != vn.end(); ++vn_it )
        {
        vn_it->set_zero();
        }

    // Cumulate triangle normals. Each normal has the same weight.
    
    Triangles::const_iterator t_it;
    Vectors::const_iterator tn_it;
    for
        (
        t_it = triangles.begin(), tn_it = tn.begin();
        t_it != triangles.end();
        ++t_it, ++tn_it
        )
        {
        hope( t_it->v0 < vn.size() );
        hope( t_it->v1 < vn.size() );
        hope( t_it->v2 < vn.size() );
        vn[ t_it->v0 ] += *tn_it;
        vn[ t_it->v1 ] += *tn_it;
        vn[ t_it->v2 ] += *tn_it;
        }

    // Normalize vertex normals.

    for( vn_it = vn.begin(); vn_it != vn.end(); ++vn_it )
        {
        if( vn_it->epsilon_zero() )
            {
            vn_it->set_zero();
            }
        else
            {
            vn_it->normalize();
            }
        }
    }


bool Mesh::cast_ray
    (
    Float32 &           distance,
    SVector3 &          normal,
    const SVector3 &    origin,
    const SVector3 &    direction,
    std::size_t         stream_index
    ) const
    {
    hope( stream_index < vertex_coordinates.size() ) ;
    hope( vertex_coordinates.size() == triangle_normals.size() );
    hope( vertex_coordinates[ stream_index ] );
    hope( triangle_normals[ stream_index ] );
    const Vectors & vertices = *vertex_coordinates[ stream_index ];
    const Vectors & normals = *triangle_normals[ stream_index ];
    hope( normals.size() == triangles.size() );

    bool best_found = false;
    Triangles::const_iterator t_it;
    Vectors::const_iterator n_it;
    for
        (
        t_it = triangles.begin(), n_it = normals.begin();
        t_it != triangles.end();
        ++t_it, ++n_it
        )
        {
        Float32 current_distance;
        hope( t_it->v0 < vertices.size() );
        hope( t_it->v1 < vertices.size() );
        hope( t_it->v2 < vertices.size() );
        if
            (
            !n_it->is_zero() &&
            Raycast::intersect_triangle_with_normal
                (
                current_distance,
                origin,
                direction,
                vertices[ t_it->v0 ],
                vertices[ t_it->v1 ],
                vertices[ t_it->v2 ],
                *n_it
                )
            )
            {
            hope( current_distance >= 0.0f );
            if( !best_found || ( current_distance < distance ) )
                {
                distance = current_distance;
                normal = *n_it;
                best_found = true;
                }
            }
        }

    return best_found;
    }


void Mesh::copy_vertex_coordinates
    (
    VertexArray &   array,
    std::size_t     array_stream,
    std::size_t     mesh_stream
    ) const
    {
    VertexArray::Vector3Stream dest_stream;
    array.lock_vectors3( dest_stream, array_stream );
    hope( dest_stream.size() == vertex_count );

    hope( mesh_stream < vertex_coordinates.size() );
    hope( vertex_coordinates[ mesh_stream ] );
    const Vectors & src_stream = *vertex_coordinates[ mesh_stream ];
   
    std::size_t i;
    for( i = 0; i < vertex_count; i++ )
        {
        dest_stream[ i ] = src_stream[ i ];
        }
    
    array.unlock_vectors3( array_stream );
    }


void Mesh::copy_vertex_normals
    (
    VertexArray &   array,
    std::size_t     array_stream,
    std::size_t     mesh_stream
    ) const
    {
    VertexArray::Vector3Stream dest_stream;
    array.lock_vectors3( dest_stream, array_stream );
    hope( dest_stream.size() == vertex_count );

    hope( mesh_stream < vertex_normals.size() );
    hope( vertex_normals[ mesh_stream ] );
    const Vectors & src_stream = *vertex_normals[ mesh_stream ];

    std::size_t i;
    for( i = 0; i < vertex_count; i++ )
        {
        dest_stream[ i ] = src_stream[ i ];
        }

    array.unlock_vectors3( array_stream );
    }


void Mesh::copy_texture_coordinates
    (
    VertexArray &   array,
    std::size_t     array_stream,
    std::size_t     mesh_stream
    ) const
    {
    VertexArray::Vector2Stream dest_stream;
    array.lock_vectors2( dest_stream, array_stream );
    hope( dest_stream.size() == vertex_count );

    hope( mesh_stream < texture_coordinates.size() );
    hope( texture_coordinates[ mesh_stream ] );
    const TexCoords & src_stream = *texture_coordinates[ mesh_stream ];

    std::size_t i;
    for( i = 0; i < vertex_count; i++ )
        {
        dest_stream[ i ] = src_stream[ i ];
        }

    array.unlock_vectors2( array_stream );
    }


void Mesh::copy_vertex_colors
    (
    VertexArray &   array,
    std::size_t     array_stream,
    std::size_t     mesh_stream
    ) const
    {
    VertexArray::ColorStream dest_stream;
    array.lock_colors( dest_stream, array_stream );
    hope( dest_stream.size() == vertex_count );

    hope( mesh_stream < vertex_colors.size() );
    hope( vertex_colors[ mesh_stream ] );
    const Colors & src_stream = *vertex_colors[ mesh_stream ];

    std::size_t i;
    for( i = 0; i < vertex_count; i++ )
        {
        dest_stream[ i ] = src_stream[ i ];
        }

    array.unlock_colors( array_stream );
    }


void Mesh::allocate_array
    (
    VertexArray &   array,
    std::size_t     vector2_stream_count,
    std::size_t     vector3_stream_count,
    std::size_t     color_stream_count
    ) const
    {
    array.set_stream_count( VertexArray::TYPE_VECTOR2, vector2_stream_count );
    array.set_stream_count( VertexArray::TYPE_VECTOR3, vector3_stream_count );
    array.set_stream_count( VertexArray::TYPE_COLOR, color_stream_count );
    array.allocate( vertex_count );
    }


void Mesh::allocate_array
    (
    VertexArray &   array
    ) const
    {
    allocate_array
        (
        array,
        texture_coordinates.size(),
        vertex_coordinates.size() + vertex_normals.size(),
        vertex_colors.size()
        );
    }


void Mesh::initialize_array
    (
    VertexArray &   array
    ) const
    {
    std::size_t i;

    const std::size_t vertex_streams = vertex_coordinates.size();
    for( i = 0; i < vertex_streams; i++ )
        {
        copy_vertex_coordinates( array, i, i );
        }

    const std::size_t normal_streams = vertex_normals.size();
    for( i = 0; i < normal_streams; i++ )
        {
        copy_vertex_normals( array, i + vertex_streams, i );
        }

    const std::size_t texcoord_streams = texture_coordinates.size();
    for( i = 0; i < texcoord_streams; i++ )
        {
        copy_texture_coordinates( array, i, i );
        }

    const std::size_t color_streams = vertex_colors.size();
    for( i = 0; i < color_streams; i++ )
        {
        copy_vertex_colors( array, i, i );
        }
    }


void Mesh::default_array_selection
    (
    VertexArray &   array
    ) const
    {
    // Must have been allocated by #allocate_array()...
    
    hope( array.get_stream_count( VertexArray::TYPE_VECTOR2 ) == texture_coordinates.size() );
    hope( array.get_stream_count( VertexArray::TYPE_VECTOR3 ) ==
                vertex_coordinates.size() + vertex_normals.size() );
    hope( array.get_stream_count( VertexArray::TYPE_COLOR ) == vertex_colors.size() );

    // Select streams.

    if( !vertex_coordinates.empty() )
        {
        array.select( VertexArray::VERTEX_ARRAY, VertexArray::TYPE_VECTOR3, 0 );
        }
    if( !vertex_normals.empty() )
        {
        array.select( VertexArray::NORMAL_ARRAY, VertexArray::TYPE_VECTOR3,
                    vertex_coordinates.size() );
        }
    if( !vertex_colors.empty() )
        {
        array.select( VertexArray::COLOR_ARRAY, VertexArray::TYPE_COLOR, 0 );
        }

    std::size_t i;
    for( i = 0; i < texture_coordinates.size() && i < MAX_TEXTURE_UNITS; i++ )
        {
        array.select( VertexArray::texcoord_array_target( i ), VertexArray::TYPE_VECTOR2, i );
        }
    }


void Mesh::create_array
    (
    VertexArray &   array
    ) const
    {
    allocate_array( array );
    initialize_array( array );
    default_array_selection( array );
    }


void Mesh::initialize_index_array
    (
    RenderObject::Geometry & geometry
    ) const
    {
    geometry.initialize_index_array( GL_TRIANGLES, 3 * triangle_count );
    geometry.lock();

    Triangles::const_iterator it;
    std::size_t i = 0;
    for( it = triangles.begin(); it != triangles.end(); ++it )
        {
        geometry.indices[ i++ ] = it->v0;
        geometry.indices[ i++ ] = it->v1;
        geometry.indices[ i++ ] = it->v2;
        }

    geometry.unlock();
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "mesh_inline.h"
#endif
