/** @file mesh.h
 * 3D triangle mesh.
 */

/*
 * begin        : Sep 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MESH_H
#define CLIENT_MESH_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_RENDER_OBJECT_H
#include "renderer/render_object.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * 3D triangle mesh.
 *
 * Mesh is simple representation of 3D object made of triangles. It should be
 * easy to create and modify and can be converted into vertex arrays
 * (see #VertexArray) and render objects (see #RenderObject). Namespace
 * #Demo::Client::Renderer::Objects offers functions to create simple meshes.
 *
 * Basic mesh consists of single set of vertex coordinates and triangles. More
 * information about vertices can be stored in the mesh - vertex normals,
 * texture coordinates and vertex colors.
 *
 * Mesh data are stored in @a streams. Several types of vertex streams can be
 * allocated: vertex coordinates (#vertex_coordinates, at least on such stream
 * must exist), vertex normals (#vertex_normals), texture coordinates
 * (#texture_coordinates) and vertex colors (#vertex_colors). All vertex streams
 * contain the same number of elements. There are two types of triangle streams:
 * triangles themselves (#triangles, exactly one such stream always exists) and
 * triangle normals (#triangle_normals).
 *
 * Existing (allocated) vertex data can be accessed directly. However, all
 * operations that might resize the arrays are forbidden, call #set_vertex_count()
 * instead. The same rules applies for triangle structures, use
 * #set_triangle_count() to resize them. Stream arrays should not be reallocated
 * directly either, use #set_*_count() method instead.
 *
 * @note Texture coordinates are specified per vertex, not three coordinates
 *       per triangle. Discontinuities in texturing result in disconitnuitites
 *       in vertex normals and colors.
 *
 * @todo Element setters/getters instead of public streams?
 */
class Mesh
    {
public:

    /**
     * Vector (array) of 3D vectors - coordinates or normals.
     */
    typedef std::vector< SVector3 >     Vectors;

    /**
     * Vector (array) of 2D vectors - texture coordinates.
     */
    typedef std::vector< SVector2 >     TexCoords;

    /**
     * Vector of colors.
     */
    typedef std::vector< SColor >       Colors;

    /**
     * Triangle - vertex inidices in counter-clockwise order.
     */
    struct Triangle
        {
        //@{
        /// Vertex index.
        std::size_t v0;
        std::size_t v1;
        std::size_t v2;
        //@}
        };

public:

    /**
     * Vertex streams.
     */
    typedef std::vector< Vectors * >    VertexStreams;

    /**
     * Normal streams.
     */
    typedef std::vector< Vectors * >    NormalStreams;

    /**
     * Texture coordinate streams.
     */
    typedef std::vector< TexCoords * >  TexCoordStreams;

    /**
     * Color streams.
     */
    typedef std::vector< Colors * >     ColorStreams;

    /**
     * Triangles.
     */
    typedef std::vector< Triangle >     Triangles;

public:

    /**
     * Default constructor.
     *
     * Creates mesh with single vertex coordinate stream.
     */
    Mesh();

    /**
     * Destructor.
     */
    ~Mesh();

public:

    /**
     * Clear mesh to initial state - single coordinate stream, no vertices/triangles.
     */
    void clear();

public:

    /**
     * Set number of vertex coordinate streams.
     */
    void set_vertex_stream_count
        (
        std::size_t count
        );

    /**
     * Set number of vertex normal streams.
     */
    void set_normal_stream_count
        (
        std::size_t count
        );

    /**
     * Set number of vertex texture coordinate streams.
     */
    void set_texcoord_stream_count
        (
        std::size_t count
        );

    /**
     * Set number of vertex color streams.
     */
    void set_color_stream_count
        (
        std::size_t count
        );

    //@{
    /** Get number of streams of given type. */
    INLINE std::size_t get_vertex_stream_count() const;
    INLINE std::size_t get_normal_stream_count() const;
    INLINE std::size_t get_texcoord_stream_count() const;
    INLINE std::size_t get_color_stream_count() const;
    //@}

public:

    /**
     * Set number of vertices.
     */
    void set_vertex_count
        (
        std::size_t count
        );

    /**
     * Set number of triangles.
     */
    void set_triangle_count
        (
        std::size_t count
        );

    /** Get number of vertices. */
    INLINE std::size_t get_vertex_count() const;
    
    /** Get number of trianlges. */
    INLINE std::size_t get_triangle_count() const;

public:

    /**
     * Recompute triangle normals for given vertex stream.
     *
     * If triangle is degenerated, its normal will be zero vector.
     */
    void update_triangle_normals
        (
        std::size_t stream_index
        );

    /**
     * Recompute triangle and vertex normals for given vertex stream.
     *
     * Triangle normals will be update as well, because they are needed to
     * compute vertex normals. Any operation that invalidates triangle normals
     * invalidates vertex normals as well, so updating vertex normals without
     * previous update to triangle normals makes no sense.
     *
     * @param vertex_stream_index Which set of vertices (and triangle normals) should
     *                            be used to compute vertex normals.
     * @param normal_stream_index Which set of vertex normals should be modified.
     */
    void update_normals
        (
        std::size_t vertex_stream_index,
        std::size_t normal_stream_index
        );

public:

    /**
     * Compute intersection of ray with the mesh.
     *
     * The ray is specified by @a origin and @a direction. The set of triangles
     * to test (vertex coordinates to use) is specified by @a stream_index.

     * If the ray intersects the mesh, #cast_ray() will return @c true,
     * @a distance will be set to distance from @a origin to intersection point
     * in units specified by length of @a direction
     * (ie. <tt>intersection = origin + direction * distance</tt>) and @a normal will
     * be set to normal of intersected triangle. @a Distance will be always positive.
     * @c false will be returned if the ray does not intersect the mesh.
     *
     * @pre Triangle normals are valid.
     */
    bool cast_ray
        (
        Float32 &           distance,
        SVector3 &          normal,
        const SVector3 &    origin,
        const SVector3 &    direction,
        std::size_t         stream_index
        ) const;

public:

    /**
     * Copy vertex coordinates to given vertex array stream.
     *
     * @param array Array to update.
     * @param array_stream Index of vector3 array stream to update.
     * @param mesh_stream Index of vertex coordinate stream to copy from.
     */
    void copy_vertex_coordinates
        (
        VertexArray &   array,
        std::size_t     array_stream,
        std::size_t     mesh_stream
        ) const;

    /**
     * Copy vertex normal to given vertex array stream.
     *
     * @param array Array to update.
     * @param array_stream Index of vector3 array stream to update.
     * @param mesh_stream Index of vertex normal stream to copy from.
     */
    void copy_vertex_normals
        (
        VertexArray &   array,
        std::size_t     array_stream,
        std::size_t     mesh_stream
        ) const;

    /**
     * Copy texture coordinates to given vertex array stream.
     *
     * @param array Array to update.
     * @param array_stream Index of vector2 array stream to update.
     * @param mesh_stream Index of texture coordinate stream to copy from.
     */
    void copy_texture_coordinates
        (
        VertexArray &   array,
        std::size_t     array_stream,
        std::size_t     mesh_stream
        ) const;

    /**
     * Copy vertex colors to given vertex array stream.
     *
     * @param array Array to update.
     * @param array_stream Index of color array stream to update.
     * @param mesh_stream Index of color stream to copy from.
     */
    void copy_vertex_colors
        (
        VertexArray &   array,
        std::size_t     array_stream,
        std::size_t     mesh_stream
        ) const;

    /**
     * Allocate vertex array streams.
     *
     * Simple helper method to allocate vertex array streams with
     * #vertex_count elements.
     *
     * @pre Array is free.
     */
    void allocate_array
        (
        VertexArray &   array,
        std::size_t     vector2_stream_count,
        std::size_t     vector3_stream_count,
        std::size_t     color_stream_count
        ) const;

    /**
     * Allocate vertex array streams.
     *
     * This method will allocate single vector2 stream for each mesh texture
     * coordinate stream, single vector3 stream for each vertex coordinate
     * and vertex normal stream and single color stream for each mesh color
     * stream.
     *
     * @pre Array is free.
     */
    void allocate_array
        (
        VertexArray &   array
        ) const;

    /**
     * Fill streams of array allocated by #allocate_array() with vertex data
     * from the mesh. All vertex coordinate streams will be copied to vector3
     * streams, followed by all vertex normal streams.
     */
    void initialize_array
        (
        VertexArray &   array
        ) const;

    /**
     * Setup default array stream selection.
     *
     * For each non-empty array of streams, first stream will be selected to
     * corresponding array. If multiple texture coordinate streams exist, they
     * will be selected to different texture units, up to maximum number of
     * units or existing texture coordinate streams.
     *
     * @pre Array has been allocated by #allocate_array().
     */
    void default_array_selection
        (
        VertexArray &   array
        ) const;

    /**
     * Call #allocate_array(), #initialize_array() and #default_array_selection().
     */
    void create_array
        (
        VertexArray &   array
        ) const;

public:

    /**
     * Initialize render object geometry to seperate triangle index array
     * from mesh triangles.
     */
    void initialize_index_array
        (
        RenderObject::Geometry & geometry
        ) const;

private:

    std::size_t         vertex_count;
        /**< Number of elements in vertex streams. */

    std::size_t         triangle_count;
        /**< Number of triangles. */

public:

    VertexStreams       vertex_coordinates;
        /**< Streams containing vertex coordinates.
             At least one stream must exists. */

    NormalStreams       vertex_normals;
        /**< Streams containing vertex normals. */

    TexCoordStreams     texture_coordinates;
        /**< Streams containing per-vertex texture coordinates. */

    ColorStreams        vertex_colors;
        /**< Streams containing per-vertex colors. */

    Triangles           triangles;
        /**< Triangles. */

    NormalStreams       triangle_normals;
        /**< Triangle normals for each vertex coordinate stream.
             The same number of streams as #vertex_coordinates, each stream
             has the same number of elements as #triangles. */

    }; // class Mesh


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "mesh_inline.h"
#endif


#endif // CLIENT_MESH_H
