/** @file mesh_entity.h
 * Mesh-based render entity.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MESH_ENTITY_H
#define CLIENT_MESH_ENTITY_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_SIMPLE_ENTITY_H
#include "renderer/simple_entity.h"
#endif

#ifndef CLIENT_MESH_H
#include "renderer/mesh.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Mesh-based render entity.
 *
 * This class adds array of Mesh objects to #SimpleEntity. The render objects
 * are created from these meshes.
 */
class MeshEntity : public SimpleEntity
    {
    NO_COPY_CONSTRUCTOR( MeshEntity );
    NO_ASSIGNMENT_OPERATOR( MeshEntity );

    /// Superclass.
    typedef SimpleEntity Super;

public:

    /**
     * Constructor.
     */
    MeshEntity();

    /**
     * Destructor.
     */
    virtual ~MeshEntity();

public:

    /**
     * Allocate given number of render objects and meshes.
     *
     * @pre No render objects are allocated.
     */
    virtual void allocate
        (
        std::size_t object_count = 1
        );

    /**
     * Initialize render objects from mesh objects.
     *
     * Meshes do not support materials/pipeline states, the pipeline state
     * will be initialized to default with z-buffering enabled.
     */
    void initialize_render_objects();

public:

    /**
     * Compute intersection of ray with the entity.
     *
     * Calls #Mesh::cast_ray() for each entity mesh and returns the nearest
     * intersection.
     *
     * @retvat distance Distance along the ray from the @a origin to the intersection.
     * @retvat normal Normal of the intersected surface (triangle).
     * @param origin Ray origin in entity-space (!).
     * @param direction Ray direction in entity-space (!).
     * @param stream_index Index of vertex coordinate and triangle normal streams to use.
     * @return @c true if ray intersects the entity, @c false otherwise.
     *
     * @pre Meshes have valid triangle normals.
     */
    bool cast_ray
        (
        Float32 &           distance,
        SVector3 &          normal,
        const SVector3 &    origin,
        const SVector3 &    direction,
        std::size_t         stream_index = 0
        ) const;

public:

    Mesh *  meshes;
        /**< Array of mesh objects.
             Number of mesh and render objects is always the same. */

    }; // class MeshEntity


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "mesh_entity_inline.h"
#endif


#endif // CLIENT_RENDER_ENTITY_H
