/** @file model.cpp
 * 3D model object - mesh + render jobs (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/model.h"
#include "renderer/render_objects.h"
#include "renderer/model_md2.h"


namespace Demo {
namespace Client {
namespace Renderer {


namespace {

/**
 * Check if @a data_object contains model data in supported format.
 *
 * @a Data_object will be opened, but the stream might be in any state.
 * (@c !good, undefined read position). You probably want to start with
 * <tt>stream.clear()</tt> and <tt>stream.seekg( ... )</tt>.
 *
 * This function should never throw. Return @c true if supported format is
 * detected, @c false otherwise.
 */
typedef bool Detect
    (
    ::Massiv::Core::DataObject & data_object
    );


/**
 * Load @a model from @a data_object.
 *
 * This function will be called only after successful call to corresponding
 * #Detect method, with no modification to @a data_object. @a Model will be cleared
 * (single vertex coordinate stream, no vertices, no triangles). It should throw
 * an exception or return @c false on error.
 */
typedef bool Load
    (
    Model &                         model,
    ::Massiv::Core::DataObject &    data_object
    );


/**
 * Detect/Load pair.
 */
struct Loader
    {
    Detect *    detect;
        /**< Detect function. */

    Load *      load;
        /**< Load function. */

    }; // struct Loader


/**
 * Array of all loaders, terminated by <tt>NULL, NULL</tt> #Loader.
 */
const Loader loader[] =
    {
    { &ModelLoad::detect_md2, &ModelLoad::load_md2 },
    { NULL, NULL }
    };


} // anonymous namespace


Model::Model()
      : version( 0 )
    {
    }


Model::~Model()
    {
    }


void Model::clear()
    {
    mesh.clear();
    render_template.clear();
    }


bool Model::load
    (
    ::Massiv::Core::DataObject & data_object
    )
    {
    clear();

    std::size_t i;
    for( i = 0; loader[ i ].detect; i++ )
        {
        hope( loader[ i ].load );

        // Detect format.

        if( loader[ i ].detect( data_object ) )
            {
            bool loaded_ok = false;
            try
                {
                loaded_ok = loader[ i ].load( *this, data_object );
                }
            catch( ... )
                {
                // Error loading the model.

                loaded_ok = false;
                }

            // Done?

            if( loaded_ok )
                {
                return true;
                }

            // No, try next loader.

            clear();
            }
        }

    // Can't load the model.

    return false;
    }


void Model::create_dummy()
    {
    const Float32 height = 1.8f;
    const Float32 half_width = 0.3f;
    std::size_t i;

    // Initialize simple box mesh.

    mesh.set_normal_stream_count( 1 );
    mesh.set_vertex_count( 6 * 4 );
    mesh.set_triangle_count( 6 * 2 );
    Renderer::Objects::initialize_box_24_vertices( mesh, 0,
                SVector3( -half_width, -half_width, 0.0f ),
                SVector3(  half_width,  half_width, height ) );
    Renderer::Objects::initialize_box_24_triangles( mesh, 0, 0 );
    mesh.update_normals( 0, 0 );

    // Two pipeline states:
    // - front side is white.
    // - remaining sides are bluish.

    render_template.pipeline_states.resize( 2 );
    PipelineState & state1 = render_template.pipeline_states[ 0 ];
    state1.depth_test.enable = true;
    state1.buffer_mask.depth = true;
    state1.lighting.color_material.enable = true;
    state1.lighting.color_material.mode = GL_AMBIENT_AND_DIFFUSE;
    state1.color = SColor( 1.0f, 1.0f, 1.0f );
    render_template.pipeline_states[ 1 ] = render_template.pipeline_states[ 0 ];
    PipelineState & state2 = render_template.pipeline_states[ 1 ];
    state2.color = SColor( 0.4f, 0.4f, 0.7f );

    // Inititalize two geometries:
    // - front quad.
    // - remaining quads.

    render_template.geometries.resize( 2 );
    RenderObject::Geometry & g1 = render_template.geometries[ 0 ];
    RenderObject::Geometry & g2 = render_template.geometries[ 1 ];

    g1.initialize_range( GL_QUADS, 2 * 4, 4 ); // third quad is north

    g2.initialize_index_array( GL_QUADS, 5 * 4 );
    g2.lock();
    for( i = 0; i < 2 * 4; i++ )
        {
        g2.indices[ i ] = i;
        }
    // skip north quad
    for( i = 0; i < 3 * 4; i++ )
        {
        g2.indices[ i + 2 * 4 ] = i + 3 * 4;
        }
    g2.unlock();

    // Initialize jobs.

    render_template.jobs.resize( 2 );
    render_template.jobs[ 0 ].pipeline_state = &state1;
    render_template.jobs[ 0 ].geometry = &g1;
    render_template.jobs[ 1 ].pipeline_state = &state2;
    render_template.jobs[ 1 ].geometry = &g2;
    }


void Model::initialize_render_object
    (
    RenderObject & render_object
    )
    {
    hope( mesh.get_vertex_stream_count() > 0 );
    const std::size_t normal_stream_count = ( mesh.get_normal_stream_count() ? 1 : 0 );
    const std::size_t texcoord_stream_count = mesh.get_texcoord_stream_count();
    const std::size_t color_stream_count = mesh.get_color_stream_count();
    const std::size_t vertex_count = mesh.get_vertex_count();
    std::size_t i;

    // Clear render object.

    render_object.clear();

    // Allocate vertex arrays.

    VertexArray & array = render_object.array;
    array.set_stream_count( VertexArray::TYPE_VECTOR2, texcoord_stream_count );
    array.set_stream_count( VertexArray::TYPE_VECTOR3, 1 + normal_stream_count );
    array.set_stream_count( VertexArray::TYPE_COLOR, color_stream_count );
    array.allocate( vertex_count );

    // Copy vertex streams.

    mesh.copy_vertex_coordinates( array, 0, 0 );
    if( normal_stream_count )
        {
        mesh.copy_vertex_normals( array, 1, 0 );
        }
    for( i = 0; i < texcoord_stream_count; i++ )
        {
        mesh.copy_texture_coordinates( array, i, i );
        }
    for( i = 0; i < color_stream_count; i++ )
        {
        mesh.copy_vertex_colors( array, i, i );
        }

    // Select array streams.

    array.select( VertexArray::VERTEX_ARRAY, VertexArray::TYPE_VECTOR3, 0 );
    if( normal_stream_count )
        {
        array.select( VertexArray::NORMAL_ARRAY, VertexArray::TYPE_VECTOR3, 1 );
        }
    if( color_stream_count )
        {
        array.select( VertexArray::COLOR_ARRAY, VertexArray::TYPE_COLOR, 0 );
        }
    for( i = 0; i < texcoord_stream_count && i < MAX_TEXTURE_UNITS; i++ )
        {
        array.select( VertexArray::texcoord_array_target( i ), VertexArray::TYPE_VECTOR2, i );
        }

    // Copy pipeline states.

    render_object.pipeline_states = render_template.pipeline_states;

    // Allocate and initialize jobs.

    const RenderObject::Jobs & src_jobs = render_template.jobs;
    RenderObject::Jobs & dst_jobs = render_object.jobs;
    dst_jobs.resize( src_jobs.size() );
    for( i = 0; i < src_jobs.size(); i++ )
        {
        dst_jobs[ i ].geometry = src_jobs[ i ].geometry;
        if( src_jobs[ i ].pipeline_state )
            {
            const std::size_t state_index = src_jobs[ i ].pipeline_state -
                        &render_template.pipeline_states.front();
            hope( state_index < render_template.pipeline_states.size() );
            dst_jobs[ i ].pipeline_state = &render_object.pipeline_states[ state_index ];
            }
        else
            {
            dst_jobs[ i ].pipeline_state = NULL;
            }
        }
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "model_inline.h"
#endif
