/** @file model.h
 * 3D model object - mesh + render jobs.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MODEL_H
#define CLIENT_MODEL_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_MESH_H
#include "renderer/mesh.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif

#include <iostream>


namespace Massiv {
namespace Core {

class DataObject;

}
}


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * 3D model object.
 *
 * Model represents 3D object. It consists of #Mesh and #RenderObject "template".
 * Information about model vertices and triangles is stored in the #Mesh. It is
 * used to create and initialize vertex arrays and to perform raycasting.
 * #RenderObject template is partially initialized render object - vertex arrays
 * are not initialized. Pipeline states and geometries contain information
 * about object "skin".
 *
 * Always use Model::Reference instead of native pointers to Model. Otherwise
 * ModelDataObject may destroy the Model. You should always use ModelDataObject
 * to obtain Model references. See #ModelDataObject for more information about
 * model versions and management.
 */
class Model : public ::Massiv::Core::SharedObject
    {
    NO_COPY_CONSTRUCTOR( Model );
    NO_ASSIGNMENT_OPERATOR( Model );

public:

    /**
     * Reference to Model.
     *
     * Reference counting is used to keep track of Model objects.
     */
    typedef ::Massiv::Core::Reference< Model > Reference;

public:

    /**
     * Constructor.
     */
    Model();

protected:

    /**
     * Destructor.
     */
    virtual ~Model();

public:

    /**
     * Clear model to initial state, but do not modify #version.
     */
    void clear();

    /**
     * Load model from @a data_object.
     *
     * @return @c true if @a data_object contains model data in supported format
     *         and model has been successfully loaded, @c false otherwise.
     */
    bool load
        (
        ::Massiv::Core::DataObject & data_object
        );

    /**
     * Create simple dummy model
     */
    void create_dummy();

public:

    /**
     * (Re)initialize given @a render_object from #mesh and #render_template.
     *
     * Mesh (first vertex stream) will be used to initialize vertex arrays and
     * #render_template will be used to initialize render jobs. Pipeline states
     * will be copied to @a render_object and can be safely modified, but
     * Geometry stored in #render_template will be referenced directly from
     * @a render_object render jobs. This means that the #Model must exist and
     * its geomtery can't be modified while render objects initialized by this
     * method are used.
     *
     * @note Only the first vertex coordinate and vertex normal stream will be
     *       copied from #mesh to @a render_object. All texture coordinate and
     *       vertex color streams will be copied.
     *
     * @pre #render_template jobs use only pipeline states owned by the template.
     */
    void initialize_render_object
        (
        RenderObject & render_object
        );

public:

    /**
     * Compute intersection of ray with the model.
     *
     * This is just a simple shorcut for <tt>mesh.cast_ray( ...);</tt>
     */
    INLINE bool cast_ray
        (
        Float32 &           distance,
        SVector3 &          normal,
        const SVector3 &    origin,
        const SVector3 &    direction,
        std::size_t         stream_index = 0
        ) const;

public:

    Mesh            mesh;
        /**< Model mesh. */

    RenderObject    render_template;
        /**< Render object template. */

    int             version;
        /**< Model version. See documentation of #ModelDataObject::get_model() for
             more information. */

    }; // class Model


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "model_inline.h"
#endif


#endif // CLIENT_MODEL_H
