/** @file model_data_object.cpp
 * Model data object (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/model_data_object.h"
#include "renderer/renderer.h"
#include "renderer/model_manager.h"
#include "data/data_manager.h"
#include "global/global.h"
#include <sstream>


namespace Demo {
namespace Client {
namespace Renderer {


ModelDataObject::ModelDataObject
    (
    const std::string &                         the_name,
    std::auto_ptr< Massiv::Core::DataObject > & the_data_object
    ) : reference_count( 0 ),
        name( the_name ),
        data_object( the_data_object )
    {
    hope( !the_data_object.get() );

    {//sob
    using namespace Massiv::Core;
    std::ostringstream oss;
    oss << "Model data object '" << name << "' created.";
    Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                Status::PRIORITY_LOW ), oss.str() );
    }//eob

    }


ModelDataObject::~ModelDataObject()
    {
    hope( reference_count == 0 );
    model = NULL;

    {//sob
    using namespace Massiv::Core;
    std::ostringstream oss;
    oss << "Model data object '" << name << "' destroyed.";
    Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                Status::PRIORITY_LOW ), oss.str() );
    }//eob

    }


Model::Reference ModelDataObject::get_model()
    {
    // Forget old model if data has been updated.

    int new_version = 1;
    if( data_object.get() && data_object->is_data_updated() )
        {
        data_object->refresh();
        if( !model.is_null() )
            {
            new_version = model->version + 1;
            model = NULL;
            }

        {//sob
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << "Model data object '" << name << "' updated.";
        Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                    Status::PRIORITY_HIGH ), oss.str() );
        }//eob
        
        }

    // Load model unless #model is valid.

    if( model.is_null() )
        {
        load_model( new_version );
        }
    hope( !model.is_null() );

    // Tell data manager to download new version if available.

    if( data_object.get() )
        {
        data_object->download_actual_version();
        }

    return model;
    }


void ModelDataObject::load_model
    (
    int version
    )
    {
    hope( model.is_null() );

    try
        {
        // Allocate the model.

        model = new Model;
        model->version = version;

        // Load the model.

        bool loaded = false;
        if( data_object.get() )
            {
            data_object->open();
            loaded = model->load( *data_object );
            data_object->close();
            }
        if( !loaded )
            {
            model->create_dummy();
            }
        }
    catch( ... )
        {
        if( data_object.get() )
            {
            data_object->close();
            }
        model = NULL;
        throw;
        }
    }


void ModelDataObject::zero_references()
    {
    ModelManager & model_manager = Global::get_model_manager();
    model_manager.unreferenced_data_object( this );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "model_data_object_inline.h"
#endif
