/** @file model_data_object.h
 * Model data object.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MODEL_DATA_OBJECT_H
#define CLIENT_MODEL_DATA_OBJECT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_MODELE_H
#include "renderer/model.h"
#endif

#include <memory>


namespace Massiv {
namespace Core {

class DataObject;

}
}


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Model data object.
 *
 * This object manages data object containing model stored in one of supported
 * formats. It should be used to create #Model objects from model data. Use
 * #get_model() method to obtain reference to the most recent model version.
 *
 * You must use ModelDataObject::Reference instead of plain pointers to
 * ModelDataObject. It ensures that the ModelDataObject will not be
 * destroyed by ModelManager. ModelManager is free to destroy unreferenced
 * data objects whenever it wants. However, unlike Model, ModelDataObject
 * can't outlive ModelManager. It's an error to destroy ModelManager while
 * references to ModelDataObjects still exist.
 *
 * @warning ModelDataObject tracks updates of single data object. It's expected
 *          that this data object contains all model data. Currently it's impossible
 *          to synchronize updates of multiple data objects. Model textures may
 *          be stored in different data objects, because PipelineState contains
 *          references to TextureDataObjects and will track texture updates
 *          automatically. Names of texture data object must be stored in the main
 *          model data object or it must be possible to deduce them from the name
 *          of the main data object.
 */
class ModelDataObject
    {
    NO_COPY_CONSTRUCTOR( ModelDataObject );
    NO_ASSIGNMENT_OPERATOR( ModelDataObject );

    /// Only ModelManager may create and destroy ModelDataObjects.
    friend class ModelManager;

    /// ModelManager uses autopointers to keep ModelDataObjects.
    FRIEND_AUTO_PTR( ModelDataObject );

public:

    /**
     * Reference counting traits.
     */
    class RefTraits
        {
    public:

        /** Adds reference to shared object. */
        static INLINE void add_reference
            (
            const ModelDataObject * ptr
            );

        /** Removes reference from shared object. */
        static INLINE void remove_reference
            (
            ModelDataObject *       ptr
            );

        /** Changes reference to shared object from @a ptr2 to @a ptr1. */
        static INLINE void change_reference
            (
            const ModelDataObject * ptr1,
            ModelDataObject *       ptr2
            );

        }; // class RefTraits

    /// RefTraits can hack into the data object.
    friend class RefTraits;

    /**
     * Reference to ModelDataObject.
     */
    typedef ::Massiv::Core::Reference< ModelDataObject, RefTraits > Reference;

protected:

    /**
     * Constructor.
     *
     * Ownership of @a the_data_object will be transferred to the ModelDataObject.
     * "Dummy" model data object will be created if @a the_data_object is @c NULL:
     * #get_model() will always return simple dummy model.
     *
     * If @a the_data_object is not @c NULL, @a the_name should be equal to
     * the name of @a the_data_object.
     */
    ModelDataObject
        (
        const std::string &                         the_name,
        std::auto_ptr< Massiv::Core::DataObject > & the_data_object
        );

    /**
     * Destructor.
     */
    ~ModelDataObject();

public:

    /**
     * Get model data object name.
     */
    INLINE const std::string & get_name() const;

    /**
     * Is this dummy model data object?
     */
    INLINE bool is_dummy() const;

    /**
     * Get #Model created from the most recent model data.
     *
     * Reference to the same object will be always returned unless the data object
     * is updated. #ModelDataObject will never modify existing #Model. When model
     * data is updated, new #Model object will be created. You can compare
     * #Model::version of "your" model and model returned by #get_model() to check
     * if model contents has been updated:
     * @code
     * Model::Reference current_model = ...;
     * ModelDataObject::Reference model_data = ...;
     * Model::Reference new_model = model_data->get_model();
     * if( current_model->version != new_model->version )
     *     {
     *     // Update all rendering structures.
     *     RenderObject & model_render_object = ...;
     *     current_model = new_model;
     *     current_model->initialize_render_object( model_render_object );
     *     }
     * @endcode
     *
     * To use #Model versions you must keep reference to #ModelDataObject instead
     * of calling #ModelManager::get_model_data_object() whenever you need the data,
     * because #ModelManager may destroy the #ModelDataObject if it's not used
     * (referenced) and create a new one when requested. In such case the model
     * versionwould drop back to 1.
     *
     * Models returned by ModelDataObject will have the following properties:
     * - Mesh will contain single stream of vertices. Other vertex streams will be
     *   present only if required by the skin (pipeline states).
     * - RenderObject will contain all information required to render the mesh as
     *   described in #Model documentation. Pipeline states will have depth testing
     *   and depth buffer writes enabled and lighting disabled.
     *
     * Currently transparent surfaces are not properly supported by the rendering
     * pipeline, blending should not be enabled in model pipeline states. Model
     * animation is not supported (neither morph-based nor skeleton-base), meshes
     * will always contain only single set of vertex coordinates.
     */
    Model::Reference get_model();

private:

    /**
     * Load #model with given @a version from #data_object.
     */
    void load_model
        (
        int version
        );

    /**
     * Inform the object that the reference count dropped to zero.
     */
    void zero_references();

private:

    mutable int                                 reference_count;
        /**< Reference count. */

    std::string                                 name;
        /**< ModelDataObject name. Always valid. */

    std::auto_ptr< Massiv::Core::DataObject >   data_object;
        /**< Model data. May be @c NULL. */

    Model::Reference                            model;
        /**< Reference to current model. May be @c NULL. */

    }; // class ModelDataObject


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "model_data_object_inline.h"
#endif


#endif // CLIENT_MODEL_DATA_OBJECT_H
