/** @file model_entity.h
 * Model-based render entity.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MODEL_ENTITY_H
#define CLIENT_MODEL_ENTITY_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDER_ENTITY_H
#include "renderer/render_entity.h"
#endif

#ifndef CLIENT_MODEL_DATA_OBJECT_H
#include "renderer/model_data_object.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Model-based render entity.
 *
 * This render entity uses #ModelDataObject to create #Model object and render it.
 */
class ModelEntity : public RenderEntity
    {
    NO_COPY_CONSTRUCTOR( ModelEntity );
    NO_ASSIGNMENT_OPERATOR( ModelEntity );

public:

    /**
     * Constructor.
     *
     * @param name Name of model data object to use.
     */
    ModelEntity
        (
        const std::string & name
        );

    /**
     * Destructor.
     */
    virtual ~ModelEntity();

public:

    /**
     * Render the entity.
     *
     * If model mesh contains vertex normals, it will be lit by #RenderManager lights.
     */
    virtual void render
        (
        const SystemTime & time_delta
        );

public:

    /**
     * Compute intersection of ray with the entity.
     *
     * @retvat distance Distance along the ray from the @a origin to the intersection.
     * @retvat normal Normal of the intersected surface (triangle).
     * @param origin Ray origin in entity-space (!).
     * @param direction Ray direction in entity-space (!).
     * @param stream_index Index of vertex coordinate and triangle normal streams to use.
     * @return @c true if ray intersects the entity, @c false otherwise.
     *
     * @note This method may have to reload the model if the data object has been updated.
     */
    INLINE bool cast_ray
        (
        Float32 &           distance,
        SVector3 &          normal,
        const SVector3 &    origin,
        const SVector3 &    direction,
        std::size_t         stream_index = 0
        );

private:

    /**
     * Refresh #model and #render_object from #model_data_object.
     */
    void refresh_model();

public:

    ModelDataObject::Reference  model_data_object;
        /**< The data object. */

    Model::Reference            model;
        /**< Model loaded from the data object. */
    
    RenderObject                render_object;
        /**< Render object build from #model. */

    }; // class ModelEntity


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "model_entity_inline.h"
#endif


#endif // CLIENT_MODEL_ENTITY_H
