/** @file model_manager.cpp
 * Model manager (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/model_manager.h"
#include "global/global.h"
#include "data/data_manager.h"
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


ModelManager::ModelManager()
    {
    }


ModelManager::~ModelManager()
    {
    // Check that list of reference models is empty.

    if( !referenced_models.empty() )
        {
        std::ostringstream oss;
        oss << "References to these ModelDataObjects still exist in ~ModelManager: ";
        ModelDataObjects::const_iterator it;
        bool first = true;
        for( it = referenced_models.begin(); it != referenced_models.end(); ++it )
            {
            if( !first )
                {
                oss << ", ";
                }
            else
                {
                first = false;
                }
            oss << it->first;
            }
        panic( oss.str().c_str() );
        }

    // Clear unreferenced models.

    clear_unreferenced_models();
    }


ModelDataObject::Reference ModelManager::get_model_data_object
    (
    const std::string & name
    )
    {
    // Try to find the model.

    ModelDataObjects::iterator it = referenced_models.find( name );
    if( it != referenced_models.end() )
        {
        // Active object found.

        ModelDataObject * const model = it->second;
        hope( model );
        hope( model->reference_count >= 1 );
        return model;
        }

    it = unreferenced_models.find( name );
    if( it != unreferenced_models.end() )
        {
        // Inactive object found. Remove it from "unreferenced" list.

        hope( it->second );
        std::auto_ptr< ModelDataObject > model( it->second );
        hope( model->reference_count == 0 );
        unreferenced_models.erase( it );

        // Move it back to "referenced" list.
        // If insert throws, tmodel will not be in neither list and it will be
        // destroyed by the auto_ptr.

        referenced_models.insert( ModelDataObjects::value_type( name, model.get() ) );
        return model.release();
        }

    // Create new model data object.

    Massiv::Core::DataManager & data_manager = Massiv::Core::Global::data_manager();
    std::auto_ptr< Massiv::Core::DataObject > data_object;
    try
        {
        data_object.reset( data_manager.get_data_object( name ) );
        }
    catch( ... )
        {
        // Ignore data manager exception, pass NULL data_object to ModelDataObject.
        }
    std::auto_ptr< ModelDataObject > model( new ModelDataObject( name, data_object ) );
    hope( !data_object.get() );

    // Insert it into list of referenced data objects.
    // See comment above about the insert.

    hope( model->reference_count == 0 );
    referenced_models.insert( ModelDataObjects::value_type( name, model.get() ) );
    return model.release();
    }


void ModelManager::collect_garbage()
    {
    clear_unreferenced_models();
    }


void ModelManager::unreferenced_data_object
    (
    ModelDataObject * model
    )
    {
    hope( model );
    hope( model->reference_count == 0 );
    const std::string name = model->get_name();

    // Remove model from "referenced" list.

    ModelDataObjects::iterator it = referenced_models.find( name );
    hope( it != referenced_models.end() );
    hope( it->second == model );
    std::auto_ptr< ModelDataObject > auto_model( it->second );
    referenced_models.erase( it );

    // Move it to "unreferenced" list.
    // If insert throws, model will be destroyed by the auto_ptr.

    unreferenced_models.insert( ModelDataObjects::value_type( name, model ) );
    auto_model.release();
    }


void ModelManager::clear_unreferenced_models()
    {
    ModelDataObjects::iterator it;
    for( it = unreferenced_models.begin(); it != unreferenced_models.end(); ++it )
        {
        hope( it->second );
        hope( it->second->reference_count == 0 );
        delete it->second;
        it->second = NULL;
        }

    unreferenced_models.clear();
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "model_manager_inline.h"
#endif
