/** @file model_manager.h
 * Model manager.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_MODEL_MANAGER_H
#define CLINET_MODEL_MANAGER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_MODEL_DATA_OBJECT_H
#include "renderer/model_data_object.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Manager of model data objects.
 *
 * Use ModelManager to get ModelDataObjects.
 *
 * @todo This class is nearly the same as TextureManager. Write template class
 *       for managers of data objects.
 */
class ModelManager
    {
    NO_COPY_CONSTRUCTOR( ModelManager );
    NO_ASSIGNMENT_OPERATOR( ModelManager );

public:

    /**
     * Constructor.
     */
    ModelManager();

    /**
     * Destructor.
     *
     * No model data objects may be referenced when destroying the manager.
     */
    ~ModelManager();

public:

    /**
     * Get reference to model data object with given @a name.
     *
     * NULL reference will never be returned. If no data object called @a name
     * exists, dummy data object will be returned.
     */
    ModelDataObject::Reference get_model_data_object
        (
        const std::string & name
        );

    /**
     * Get rid of some unused model data objects.
     *
     * Currently all unreferenced data objects will be destroyed. In the future
     * LRU-like algorithm should be used if too many data objects exists.
     * Sometimes it may be unwise to destroy model as soon as it becomes
     * unused.
     *
     * @note This is called automatically by Renderer.
     */
    void collect_garbage();

protected:

    /// ModelDataObject may inform ModelManager that its reference count is zero.
    friend class ModelDataObject;

    /**
     * Reference count of data object reached zero.
     */
    void unreferenced_data_object
        (
        ModelDataObject * model
        );

private:

    /**
     * Clear all unreferenced model data objects.
     */
    void clear_unreferenced_models();

private:

    /**
     * Hash of owning model data object pointers.
     *
     * @note Unfortunately std::auto_ptr has neither copy-constructor nor
     *       assignment operator with <tt>const &</tt> arguments and can't
     *       be used as a hash value type. So we have to destroy them manually.
     */
    typedef ext::hash_map< std::string, ModelDataObject * > ModelDataObjects;

    ModelDataObjects  referenced_models;
        /**< Models with non-zero reference count. */

    ModelDataObjects  unreferenced_models;
        /**< Models with zero reference count.
             @todo Implement wise algorithm to get rid of these. */

    }; // class ModelManager


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "model_manager_inline.h"
#endif


#endif // CLIENT_MODEL_MANAGER_H
