/** @file model_md2.cpp
 * Quake 2 model loading (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "client_exception.h"
#include "renderer/model_md2.h"
#include "renderer/model_md2_normals.h"
#include "renderer/texture_manager.h"
#include "renderer/renderer.h"
#include "data/data_manager.h"


namespace Demo {
namespace Client {
namespace Renderer {
namespace ModelLoad {


namespace {


/**
 * Read binary Int16 from @a stream.
 */
Int16 getw
    (
    std::istream & stream
    )
    {
    UInt16 x = static_cast< UInt16 >( stream.get() );
    x |= static_cast< UInt16 >( stream.get() << 8 );
    return static_cast< Int16 >( x );
    }


/**
 * Read binary Int32 from @a stream.
 */
Int32 getdw
    (
    std::istream & stream
    )
    {
    UInt32 x = static_cast< UInt32 >( stream.get() );
    x |= static_cast< UInt32 >( stream.get() << 8 );
    x |= static_cast< UInt32 >( stream.get() << 16 );
    x |= static_cast< UInt32 >( stream.get() << 24 );
    return static_cast< Int32 >( x );
    }


/**
 * Read binary Float32 from @a stream.
 */
Float32 getf
    (
    std::istream & stream
    )
    {
    union
        {
        UInt32  i;
        Float32 f;
        } value;
    value.i = static_cast< UInt32 >( getdw( stream ) );
    return value.f;
    }


/**
 * Read MD2 vertex from @a stream.
 */
SVector3 get_vertex
    (
    std::istream &      stream,
    const SVector3 &    scale,
    const SVector3 &    translate
    )
    {
    // MD2 models are looking in direction X axis, we must rotate them
    // 90 degrees counter-clockwise: y' = x, x' = -y.
    // The scale is just my guess.

    const Float32 SCALE = 1.0f / 15.0f;
    SVector3 result;
    result.set_y( ( stream.get() * scale.x() + translate.x() ) * SCALE );
    result.set_x( -( stream.get() * scale.y() + translate.y() ) * SCALE );
    result.set_z( ( stream.get() * scale.z() + translate.z() ) * SCALE );
    return result;
    }


/**
 * Read MD2 normal from @a stream.
 */
SVector3 get_normal
    (
    std::istream &      stream
    )
    {
    const UInt8 normal_index = static_cast< UInt8 >( stream.get() );
    const SVector3 normal = md2_normals[ normal_index ];
    return SVector3( -normal.y(), normal.x(), normal.z() );
    }


/**
 * MD2 header.
 */
class Header
    {
public:

    /**
     * Load MD2 header.
     *
     * @pre MD2 format has been detected.
     */
    void load
        (
        std::istream & stream
        )
        {
        stream.clear();
        stream.seekg( 8 );

        skin_width = getdw( stream );
        skin_height = getdw( stream );
        frame_size = getdw( stream );

        skin_count = getdw( stream );
        vertex_count = getdw( stream );
        texcoord_count = getdw( stream );
        triangle_count = getdw( stream );
        command_count = getdw( stream );
        frame_count = getdw( stream );

        skin_offset = getdw( stream );
        texcoord_offset = getdw( stream );
        triangle_offset = getdw( stream );
        frame_offset = getdw( stream );
        command_offset = getdw( stream );
        end_offset = getdw( stream );

        // Check header.

        if( skin_count > 256 || vertex_count > 65536 || texcoord_count > 65536 ||
            triangle_count > 65536 || command_count > 65536 || frame_count > 4096 )
            {
            throw ExRendererError( "Weird MD2 counts, probably damaged file." );
            }

        if( ( frame_size - 6 * 4 - 16 ) / 4 != vertex_count )
            {
            throw ExRendererError( "MD2 vertex count does not match frame size." );
            }
        }

public:

    Int32 skin_width;
    Int32 skin_height;
    Int32 frame_size;

    Int32 skin_count;
    Int32 vertex_count;
    Int32 texcoord_count;
    Int32 triangle_count;
    Int32 command_count;
    Int32 frame_count;

    Int32 skin_offset;
    Int32 texcoord_offset;
    Int32 triangle_offset;
    Int32 frame_offset;
    Int32 command_offset;
    Int32 end_offset;

    }; // class Header


/**
 * UV + index into vertex stream/array.
 */
class UVVertex
    {
public:

    /**
     * Default constructor.
     */
    UVVertex()
          : vertex_index( 0 )
        {
        }


    /**
     * Constructor.
     */
    UVVertex
        (
        Float32     the_u,
        Float32     the_v,
        std::size_t the_index
        ) : u( the_u ),
            v( the_v ),
            vertex_index( the_index )
        {
        }


    /**
     * Load U, V and vertcoord index from GL command element.
     *
     * @param stream Stream to load element from.
     * @return Vertcoord index (index info MD2 frame/vertex normals).
     *
     * #Vertex_index will not be modified.
     */
    std::size_t load_gl_command
        (
        std::istream &  stream
        )
        {
        u = getf( stream );
        v = getf( stream );
        return static_cast< std::size_t >( getdw( stream ) );
        }

public:

    Float32     u, v;
        /**< Texture coordinates. */

    std::size_t vertex_index;
        /**< Index into vertex stream/array. */

    }; // class UVVertex


/**
 * Vector of UV vertices.
 */
typedef std::vector< UVVertex > UVVertices;


/**
 * Vector of UVVertex vectors.
 */
typedef std::vector< UVVertices > VertexInfo;


/**
 * Load GL commands from @a stream.
 *
 * @param model Initialize geometries of this model.
 * @param vertex_info Store information about model vertices here.
 * @param stream Stream to read from. Current read position is at the
 *               first gl command.
 * @return Number of vertices.
 */
std::size_t load_gl_commands
    (
    Model &         model,
    VertexInfo &    vertex_info,
    std::istream &  stream
    )
    {
    std::size_t vertex_count = 0;
    while( true )
        {
        // Read command type and element count.

        const Int32 command_tag = getdw( stream );
        if( command_tag == 0 )
            {
            break;
            }
        const bool is_strip = ( command_tag > 0 );
        const std::size_t element_count = static_cast< std::size_t >
                    ( command_tag > 0 ? command_tag : -command_tag );
        if( element_count < 3 )
            {
            throw ExRendererError( "Less than 3 vertices in MD2 GL command." );
            }

        // Allocate new geometry.

        model.render_template.geometries.push_back( RenderObject::Geometry() );
        RenderObject::Geometry & geometry = model.render_template.geometries.back();
        geometry.initialize_index_array
                    (
                    is_strip ? GL_TRIANGLE_STRIP : GL_TRIANGLE_FAN,
                    element_count
                    );

        // Read command elements.

        geometry.lock();
        try 
            {
            std::size_t i;
            for( i = 0; i < element_count; i++ )
                {
                // Read element U, V, vertcoord index.

                UVVertex vertex;
                const std::size_t vertcoord = vertex.load_gl_command( stream );
                if( vertcoord >= vertex_info.size() )
                    {
                    // Vertcoord index out of range.

                    throw ExRendererError( "MD2 GL command vertex index out of range." );
                    }

                // Find/create new vertex.

                UVVertices & vertices = vertex_info[ vertcoord ];
                std::size_t vertex_index;
                std::size_t j;
                for( j = 0; j < vertices.size(); j++ )
                    {
                    const UVVertex & v = vertices[ j ];
                    if( Massiv::Core::FloatTraits< Float32 >::epsilon_equals( v.u, vertex.u ) &&
                        Massiv::Core::FloatTraits< Float32 >::epsilon_equals( v.v, vertex.v ) )
                        {
                        // We already know about this vertex.

                        vertex_index = v.vertex_index;
                        break;
                        }
                    }

                if( j == vertices.size() )
                    {
                    // New vertex.

                    vertex_index = vertex_count;
                    vertices.push_back( UVVertex( vertex.u, vertex.v, vertex_index ) );
                    vertex_count++;
                    }

                // Add geometry element.

                geometry.indices[ i ] = vertex_index;
                }
            }
        catch( ... )
            {
            geometry.unlock();
            throw;
            }
        geometry.unlock();
        }

    return vertex_count;
    }


/**
 * Initialize texture coordinate stream of @a mesh from @a vertex_info.
 */
void initialize_texcoord_stream
    (
    Mesh &              mesh,
    const VertexInfo &  vertex_info
    )
    {
    hope( mesh.get_texcoord_stream_count() >= 1 );
    hope( mesh.texture_coordinates[ 0 ] );
    Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ 0 ];

    std::size_t i;
    for( i = 0; i < vertex_info.size(); i++ )
        {
        const UVVertices & vertices = vertex_info[ i ];
        std::size_t j;
        for( j = 0; j < vertices.size(); j++ )
            {
            const UVVertex & vertex = vertices[ j ];
            hope( vertex.vertex_index < texcoords.size() );
            texcoords[ vertex.vertex_index ] = SVector2( vertex.u, vertex.v );
            }
        }
    }


/**
 * Load MD2 frame from @a stream and initialize @a stream_index-th vertex
 * coordanate and vertex normal streams of @a mesh.
 */
void load_frame
    (
    Mesh &              mesh,
    std::size_t         stream_index,
    const VertexInfo &  vertex_info,
    std::istream &      stream
    )
    {
    hope( mesh.get_vertex_stream_count() > stream_index );
    hope( mesh.get_normal_stream_count() > stream_index );
    hope( mesh.vertex_coordinates[ stream_index ] );
    hope( mesh.vertex_normals[ stream_index ] );
    Mesh::Vectors & coords = *mesh.vertex_coordinates[ stream_index ];
    Mesh::Vectors & normals = *mesh.vertex_normals[ stream_index ];
    hope( coords.size() == normals.size() );

    // Load frame header.

    SVector3 scale;
    scale.set_x( getf( stream ) );
    scale.set_y( getf( stream ) );
    scale.set_z( getf( stream ) );
    SVector3 translate;
    translate.set_x( getf( stream ) );
    translate.set_y( getf( stream ) );
    translate.set_z( getf( stream ) );

    // Skip frame name.

    char name[ 16 ];
    stream.read( reinterpret_cast< std::istream::char_type * >( name ), 16 );

    // Read frame vertices.

    std::size_t i;
    for( i = 0; i < vertex_info.size(); i++ )
        {
        const SVector3 position = get_vertex( stream, scale, translate );
        const SVector3 normal = get_normal( stream );
        const UVVertices & vertices = vertex_info[ i ];
        std::size_t j;
        for( j = 0; j < vertices.size(); j++ )
            {
            const std::size_t index = vertices[ j ].vertex_index;
            hope( index < coords.size() );
            coords[ index ] = position;
            normals[ index ] = normal;
            }
        }
    }


/**
 * Get height of the bottommost vertex.
 */
Float32 get_floor_level
    (
    const Mesh &    mesh,
    std::size_t     stream_index
    )
    {
    hope( mesh.get_vertex_stream_count() > stream_index );
    const Mesh::Vectors & coords = *mesh.vertex_coordinates[ stream_index ];

    if( coords.empty() )
        {
        return 0.0f;
        }
    Float32 floor_level = coords[ 0 ].z();
    std::size_t i;
    for( i = 1; i < coords.size(); i++ )
        {
        if( coords[ i ].z() < floor_level )
            {
            floor_level = coords[ i ].z();
            }
        }
    return floor_level;
    }


/**
 * Add given @a offset to vertex coordinates in stream @a stream_index.
 */
void move_vertices
    (
    Mesh &              mesh,
    std::size_t         stream_index,
    const SVector3 &    offset
    )
    {
    hope( mesh.get_vertex_stream_count() > stream_index );
    Mesh::Vectors & coords = *mesh.vertex_coordinates[ stream_index ];

    std::size_t i;
    for( i = 0; i < coords.size(); i++ )
        {
        coords[ i ] += offset;
        }
    }


/**
 * Interpret strip/fan geometries and initialize @a model triangles.
 */
void initialize_triangles_from_geometries
    (
    Model & model
    )
    {
    Mesh & mesh = model.mesh;
    RenderObject::Geometries & geometries = model.render_template.geometries;

    // Count triangles.

    std::size_t triangle_count = 0;
    std::size_t i;
    for( i = 0; i < geometries.size(); i++ )
        {
        triangle_count += geometries[ i ].get_size() - 2;
        }

    // Allocate triangles.

    mesh.set_triangle_count( triangle_count );

    // Initialize triangles.
    // Note: Reading index arrays is stupid. They might be in AGP memory in
    // the far far future.

    Mesh::Triangles & triangles = mesh.triangles;
    std::size_t gi;
    std::size_t ti = 0;
    for( gi = 0; gi < geometries.size(); gi++ )
        {
        const std::size_t vertex_count = geometries[ gi ].get_size();
        hope( vertex_count > 2 );
        hope( ti + vertex_count - 2 <= triangle_count );
        geometries[ gi ].lock();
        const RenderObject::Geometry::IndexArray & indices = geometries[ gi ].indices;

        if( geometries[ gi ].get_mode() == GL_TRIANGLE_STRIP )
            {
            // Interpret cw triangle strip as set of ccw triangles.

            std::size_t prev0 = indices[ 0 ];
            std::size_t prev1 = indices[ 1 ];
            for( i = 2; i < vertex_count; i++ )
                {
                if( ( i % 2 ) == 0 )
                    {
                    // Even triangle.

                    const std::size_t next = indices[ i ];
                    triangles[ ti ].v2 = prev0;
                    triangles[ ti ].v1 = prev1;
                    triangles[ ti ].v0 = next;
                    prev0 = next;
                    }
                else
                    {
                    // Odd triangle.

                    const std::size_t next = indices[ i ];
                    triangles[ ti ].v2 = prev0;
                    triangles[ ti ].v1 = prev1;
                    triangles[ ti ].v0 = next;
                    prev1 = next;
                    }
                ti++;
                }
            }
        else
            {
            // Interpret cw triangle fan as set of ccw triangles.

            hope( geometries[ gi ].get_mode() == GL_TRIANGLE_FAN );
            std::size_t center = indices[ 0 ];
            std::size_t prev = indices[ 1 ];
            for( i = 2; i < vertex_count; i++ )
                {
                const std::size_t next = indices[ i ];
                triangles[ ti ].v0 = center;
                triangles[ ti ].v1 = next;
                triangles[ ti ].v2 = prev;
                prev = next;
                ti++;
                }
            }

        geometries[ gi ].unlock();
        }

    hope( ti == triangles.size() );
    }


/**
 * Initialize model skin - pipeline states and jobs.
 *
 * @param render_object Render template to modify.
 * @param data_object_name Name of model data object. Used to guess texture
 *                         data object name.
 */
void initialize_skin
    (
    RenderObject &      render_object,
    const std::string & data_object_name
    )
    {
    // Simple pipeline state.

    render_object.pipeline_states.resize( 1 );
    PipelineState & state = render_object.pipeline_states.front();
    state.cull.front_face = GL_CW;
    state.depth_test.enable = true;
    state.buffer_mask.depth = true;
    state.lighting.color_material.enable = true;
    state.lighting.color_material.mode = GL_AMBIENT_AND_DIFFUSE;

    // Enable alpha testing.

    state.alpha_test.enable = true;
    state.alpha_test.function = GL_GREATER;
    state.alpha_test.value = 0.5f;

    // Try to load the texture.

    TextureDataObject::Reference texture;
    try
        {
        TextureManager & texture_manager = Global::get_texture_manager();
        const std::string texture_name = data_object_name + "_skin";
        texture = texture_manager.get_texture_data_object( texture_name );
        }
    catch( ... )
        {
        // No texturing if skin data object can't be loaded.

        texture = NULL;
        }
    if( !texture.is_null() )
        {
        // Use texturing.

        state.texture[ 0 ].data = texture;
        state.texture[ 0 ].flags = Texture::GENERATE_MIPMAPS;
        }

    // Render jobs.

    std::size_t job_count = render_object.geometries.size();
    render_object.jobs.resize( job_count );
    std::size_t i;
    for( i = 0; i < job_count; i++ )
        {
        render_object.jobs[ i ].pipeline_state = ( i == 0 ? &state : NULL );
        render_object.jobs[ i ].geometry = &render_object.geometries[ i ];
        }
    }


} // anonymous namespace


/**
 * Detect MD2 format.
 */
bool detect_md2
    (
    ::Massiv::Core::DataObject & data_object
    )
    {
    std::istream & stream = data_object.data;
    stream.clear();
    stream.seekg( 0 );
    if( stream.get() != 'I' )
        {
        return false;
        }
    if( stream.get() != 'D' )
        {
        return false;
        }
    if( stream.get() != 'P' )
        {
        return false;
        }
    if( stream.get() != '2' )
        {
        return false;
        }
    if( getdw( stream ) != 8 )
        {
        return false;
        }
    return true;
    }


/**
 * Load model in MD2 format.
 *
 * The name of data object with skin texture will be name of model data object
 * with "_skin" appended.
 *
 * @bug Separate triangles or triangle normals seem wrong - raycasting behaves
 *      really weird.
 */
bool load_md2
    (
    Model &                         model,
    ::Massiv::Core::DataObject &    data_object
    )
    {
    std::istream & stream = data_object.data;

    Header header;
    header.load( stream );
    if( !stream.good() )
        {
        return false;
        }

    // Information about model vertices.
    // Two-dimensional vector. Vector at [ vertcoord_index ] is set of all
    // vertices with different UVs sharing the same coordinate and normal.

    VertexInfo vertex_info;
    vertex_info.resize( header.vertex_count );

    // Load GL commands, build vertex_info and model geometries.

    stream.seekg( header.command_offset );
    const std::size_t vertex_count = load_gl_commands( model, vertex_info, stream );
    if( !stream.good() )
        {
        return false;
        }

    // Allocate vertex streams.

    Mesh & mesh = model.mesh;
    mesh.set_vertex_count( vertex_count );
    hope( mesh.get_vertex_stream_count() == 1 );
    mesh.set_normal_stream_count( 1 );
    mesh.set_texcoord_stream_count( 1 );

    // Initialize texture coordinate stream.

    initialize_texcoord_stream( mesh, vertex_info );

    // Load first frame into first vertex coordinate and vertex normal streams.

    stream.seekg( header.frame_offset );
    load_frame( mesh, 0, vertex_info, stream );
    if( !stream.good() )
        {
        return false;
        }

    // Move the model in Z-axis. Quake 2 model origin (0, 0, 0) is not at floor
    // level, I don't really know where exactly is it supposed to be. So guess
    // the floor level from the first frame (bottommost vertex) and use that
    // to move the model. Hack. Note that origin of egoboo models is on the floor.

    const Float32 floor_level = get_floor_level( mesh, 0 );
    move_vertices( mesh, 0, SVector3( 0.0f, 0.0f, -floor_level ) );

    // Initialize mesh triangles from geometries.

    initialize_triangles_from_geometries( model );

    // Compute triangle normals.

    mesh.update_triangle_normals( 0 );

    // Initialize pipeline states and jobs.

    initialize_skin( model.render_template, data_object.get_name() );

    return true;
    }


} // namespace Demo::Client::ModelLoad
} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo
