/** @file opengl.cpp
 * Helper OpenGL functions.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "renderer/opengl.h"
#include "sdl/client_sdl.h"
#include "client_exception.h"

// FIXME: current core.h does not expose logger, client facilities can't be added, buah
#include "global/global.h"

#include <cstring>


namespace Demo {
namespace Client {
namespace OpenGL {


namespace {


/**
 * Convert OpenGL error enum to string.
 */
const char * opengl_error_string
    (
    GLenum error
    )
    {
    switch( error )
        {
        case GL_NO_ERROR:
            return "no error";

        case GL_OUT_OF_MEMORY:
            return "out of memory";

        case GL_INVALID_ENUM:
            return "invalid enum";

        case GL_INVALID_VALUE:
            return "invalid value";

        case GL_INVALID_OPERATION:
            return "invalid operation";

// Not defined in OpenGL 1.1. FIXME.
//        case GL_TABLE_TOO_LARGE:
//            return "table too large";

        case GL_STACK_OVERFLOW:
            return "stack overflow";

        case GL_STACK_UNDERFLOW:
            return "stack underflow";

        default:
            return "unknown error";
        }
    }

    
/**
 * Is extension @a name present in EXTENSION gl string?
 */
bool extension_supported
    (
    const char *    name
    )
    {
    hope( name );
    const char * const extensions = reinterpret_cast< const char * >( glGetString( GL_EXTENSIONS ) );
    hope( extensions );

    // Find the extension name.
    // Don't bother creating std::strings, std::find can't be used, so use plain old strstr.
    // FIXME: using root namespace instead of std:: because of STLPort-MSVC problems.
    
    const char * const found = ::strstr( extensions, name );
    if( found )
        {
        const char c = found[ ::strlen( name ) ];
        if( ( c == ' ' ) || ( c == '\0' ) )
            {
            return true;
            }
        }

    return false;
    }

    
/**
 * Initialize extension pointers and *_supported variable.
 */
void initialize_extensions()
    {
    #define GLFN_BEGIN_EXTENSION( n ) \
        if( extension_supported( #n ) ) \
            { \
            GL::n ## _supported = true;

    #define GLFN( r, n, a) \
            { \
                void * const address = SDL_GL_GetProcAddress( "gl" #n ); \
                if( !address ) \
                    { \
                    throw ExOpenGLError( "Can't get address of gl" #n "." ); \
                    } \
                void * * const fn_ptr_ptr = reinterpret_cast< void * * >( &GL::n ); \
                *fn_ptr_ptr = address; \
            }

    #define GLFN_END_EXTENSION( n ) \
            } \
        else \
            { \
            GL::n ## _supported = false; \
            }

    #include "opengl_ext_functions.h"

    #undef GLFN_END_EXTENSION
    #undef GLFN
    #undef GLFN_BEGIN_EXTENSION
    }


/**
 * Deinitialize extension pointers.
 */
void deinitialize_extensions()
    {
    #define GLFN_BEGIN_EXTENSION( n ) GL::n ## _supported = false;
    #define GLFN( r, n, a) GL::n = NULL;
    #define GLFN_END_EXTENSION( n )

    #include "opengl_ext_functions.h"

    #undef GLFN_END_EXTENSION
    #undef GLFN
    #undef GLFN_BEGIN_EXTENSION
    }


} // anonymous namespace


void initialize()
    {
    initialize_extensions();
    check_errors( "Initialization: " );
    }


void shutdown()
    {
    check_errors( "Shutdown: " );
    deinitialize_extensions();
    }


void check_errors
    (
    const std::string & message
    )
    {
    int error_count = 0;
    while( error_count < 100 )
        {
        const GLenum error = glGetError();

        if( error == GL_NO_ERROR )
            {
            break;
            }

        // Log the error.
        
            {
            using namespace Massiv::Core;
            std::ostringstream oss;
            // FIXME: OpenGL facility, remove "OpenGL - " prefix
            oss << message << "OpenGL - " << opengl_error_string( error ) << ".";
            Global::log( Status::error( Status::FACILITY_GENERIC ), oss.str() );
            }

        error_count++;
        }

    // Throw exception?
    
    if( error_count )
        {
        if( error_count == 100 )
            {
            throw ExOpenGLError( "Infinite loop in OpenGL::check_errors(). "
                "Inside glBegin()-glEnd() block?" );
            }
        else
            {
            std::ostringstream oss;
            oss << error_count << " OpenGL error(s) collected in check_error(). "
                "See error log for more info.";
            throw ExOpenGLError( oss.str() );
            }
        }
    }


} // namespace Demo::Client::OpenGL
} // namespace Demo::Client
} // namespace Demo
