/** @file opengl.h
 * This header correctly includes system OpenGL headers and defines several
 * OpenGL extensions. Use it instead of plain GL/gl.h and GL/glu.h.
 *
 * Extension OpenGL functions are in @c GL namespace.
 * Global OpenGL-related functions (initialization, etc.) are in
 * @c Demo::Client::OpenGL namespace. The rest of the renderer is in
 * @c Demo::Client::Renderer namespace.
 *
 * @todo Currently standard OpenGL 1.1 functions should be called in the GL way
 *       (global namespace, @c gl prefix) but extensions are stored in @c GL
 *       namespace without @c gl prefix. Setup addresses of standard functions
 *       in the same way as extensions and put them in GL namespace too. Get
 *       rid of dependency on system gl.h, require OpenGL 1.3.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef CLIENT_OPENGL_H
#define CLIENT_OPENGL_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#include <string>


#if defined( GL_TRUE )
#error "GL/gl.h included before opengl.h"
#endif // GL_TRUE


// Include core opengl.

#ifdef SYSTEM_WIN32
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#endif // SYSTEM_WIN32

#include <GL/gl.h>

// OpenGL 1.3 functionality is required, but extensions are used instead of
// 1.3, because standard micro$oft headers are 1.1.

#if !defined( GL_VERSION_1_1 ) || !GL_VERSION_1_1
#error "OpenGL version >=1.1 required"
#endif // GL_VERSION_1_1

// OpenGL calling conventions.

#if !defined( GLAPIENTRY )
#  if defined( APIENTRY )
#    define GLAPIENTRY APIENTRY
#  else
#    error "Unknown OpenGL calling conventions."
#  endif
#endif


// Extensions

#include "opengl_extensions.h"


// GLU (only for gluScaleImage and gluBuild2DMipmaps)

#include <GL/glu.h>


// Our stuff


namespace Demo {
namespace Client {
namespace OpenGL {


/**
 * Initialize OpenGL.
 */
void initialize();


/**
 * Shutdown OpenGL.
 */
void shutdown();


/**
 * Check OpenGL errors.
 *
 * If there are OpenGL errors, log them and throw ExOpenGLError. It's not
 * recommended to catch ExOpenGLException from #check_errors() because it
 * does not contain enough information to recover from the error.
 * #check_errors() should never throw if you are using OpenGL correctly.
 *
 * @param message Message to prepend before each logged error message.
 */
void check_errors
    (
    const std::string & message = std::string()
    );


} // namespace Demo::Client::OpenGL
} // namespace Demo::Client
} // namespace Demo


#endif // CLIENT_OPENGL_H
