/** @file opengl_types.h
 * Structures representing standard OpenGL types. Their layout is compatible
 * with "vector" versions OpenGL commands.
 *
 * Note that all these types are defined in namespace Demo::Client::OpenGL, not
 * renderer.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_OPENGL_TYPES_H
#define CLIENT_OPENGL_TYPES_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_H
#include "renderer/opengl.h"
#endif

#ifndef MASSIV_MATH_H
#include "utility/math.h"
#endif


namespace Demo {
namespace Client {
namespace OpenGL {


/**
 * Vector - 2 floats (vertex, uv).
 */
class Vector2
    {
public:

    /** Default constructor. */
    INLINE Vector2();

    /** Copy constructor. */
    INLINE Vector2
        (
        const Vector2 & other
        );

    /** Construct Vector2 from SVector2. */
    explicit INLINE Vector2
        (
        const SVector2 & other
        );

    /** Construct Vector2 from coordinates. */
    INLINE Vector2
        (
        GLfloat x,
        GLfloat y
        );

    /** Assignment operator. */
    INLINE Vector2 & operator=
        (
        const Vector2 & other
        );

    /** Assign SVector2. */
    INLINE Vector2 & operator=
        (
        const SVector2 & other
        );

public:

    /** Compare vectors. */
    INLINE bool operator==
        (
        const Vector2 & other
        ) const;

    /** Compare vectors. */
    INLINE bool operator!=
        (
        const Vector2 & other
        ) const;

public:

    /** Get @a x coordinate, also known as @a u or @a s. */
    INLINE GLfloat x() const;

    /** Get @a y coordinate, also known as @a v or @a t */
    INLINE GLfloat y() const;

    /** Convert Vector2 to SVector2. */
    INLINE SVector2 svector() const;

public:

    /** Call @c glVertex2. */
    INLINE void gl_vertex() const;

    /** Call @c glTexCoord2. */
    INLINE void gl_tex_coord() const;

    /** Call @c glMultiTexCoord2. */
    void gl_tex_coord
        (
        GLenum texture
        ) const;

public:

    GLfloat v[ 2 ];
        /**< Vector data. */

    }; // class Vector2


/**
 * Vector - 3 floats (vertex, normal).
 */
class Vector3
    {
public:

    /** Default constructor. */
    INLINE Vector3();

    /** Copy constructor. */
    INLINE Vector3
        (
        const Vector3 & other
        );

    /** Construct Vector3 from SVector3. */
    explicit INLINE Vector3
        (
        const SVector3 & other
        );

    /** Construct Vector2 from coordinates. */
    INLINE Vector3
        (
        GLfloat x,
        GLfloat y,
        GLfloat z
        );

    /** Assignment operator. */
    INLINE Vector3 & operator=
        (
        const Vector3 & other
        );

    /** Assign SVector3. */
    INLINE Vector3 & operator=
        (
        const SVector3 & other
        );

public:

    /** Compare vectors. */
    INLINE bool operator==
        (
        const Vector3 & other
        ) const;

    /** Compare vectors. */
    INLINE bool operator!=
        (
        const Vector3 & other
        ) const;

public:

    /** Get @a x coordinate, also known as @a u, @a s or @a r. */
    INLINE GLfloat x() const;

    /** Get @a y coordinate, also known as @a v, @a t or @a g. */
    INLINE GLfloat y() const;

    /** Get @a z coordinate, also known as @a w, @a r or @a b. */
    INLINE GLfloat z() const;

    /** Convert Vector3 to SVector3. */
    INLINE SVector3 svector() const;

public:

    /** Call @c glVertex3. */
    INLINE void gl_vertex() const;

    /** Call @c glNormal3. */
    INLINE void gl_normal() const;

public:

    GLfloat v[ 3 ];
        /**< Vector data. */

    }; // class Vector3


/**
 * RGBA color - 4 unsigned bytes.
 *
 * Internal representation of the color components is unsigned byte, but
 * setters/getters expect floats in 0-1 range.
 */
class Color
    {
public:

    /** Default constructor - opaque black. */
    INLINE Color();

    /** Copy constructor. */
    INLINE Color
        (
        const Color & other
        );

    /** Construct from SColor. */
    explicit INLINE Color
        (
        const SColor & other
        );

    /** Construct from RGBA values. */
    INLINE Color
        (
        GLfloat red,
        GLfloat green,
        GLfloat blue,
        GLfloat alpha = 1.0f // opaque
        );
        
    /** Construct from RGBA values. */
    INLINE Color
        (
        GLubyte red,
        GLubyte green,
        GLubyte blue,
        GLubyte alpha = 255 // opaque
        );
        
    /** Construct from RGB vector and alpha value. */
    explicit INLINE Color
        (
        const Vector3 & rgb,
        GLfloat         alpha = 1.0f // opaque
        );

    /** Assignment operator. */
    INLINE Color & operator=
        (
        const Color & other
        );

    /** Assign SColor. */
    INLINE Color & operator=
        (
        const SColor & other
        );

public:

    /** Compare colors. */
    INLINE bool operator==
        (
        const Color & other
        ) const;

    /** Compare colors. */
    INLINE bool operator!=
        (
        const Color & other
        ) const;

    /** Lexicographically compare colors. */
    INLINE bool operator<
        (
        const Color & other
        ) const;

public:

    /** Get red component. */
    INLINE GLfloat red() const;

    /** Get green component. */
    INLINE GLfloat green() const;

    /** Get blue component. */
    INLINE GLfloat blue() const;

    /** Get alpha component. */
    INLINE GLfloat alpha() const;
    
    /** Get red component. */
    INLINE GLubyte red_byte() const;

    /** Get green component. */
    INLINE GLubyte green_byte() const;

    /** Get blue component. */
    INLINE GLubyte blue_byte() const;

    /** Get alpha component. */
    INLINE GLubyte alpha_byte() const;

    /** Get RGB vector. */
    INLINE Vector3 rgb() const;

    /** Convert to SColor. */
    INLINE SColor scolor() const;

public:

    /** Call @c glColor4. */
    INLINE void gl_color() const;

public:

    /** Set red component. */
    INLINE void set_red
        (
        GLfloat red
        );

    /** Set green component. */
    INLINE void set_green
        (
        GLfloat green
        );

    /** Set blue component. */
    INLINE void set_blue
        (
        GLfloat blue
        );

    /** Set alpha component. */
    INLINE void set_alpha
        (
        GLfloat alpha
        );
        
    /** Set red component. */
    INLINE void set_red
        (
        GLubyte red
        );

    /** Set green component. */
    INLINE void set_green
        (
        GLubyte green
        );

    /** Set blue component. */
    INLINE void set_blue
        (
        GLubyte blue
        );

    /** Set alpha component. */
    INLINE void set_alpha
        (
        GLubyte alpha
        );

    /** Set RGB vector. */
    INLINE void set_rgb
        (
        const Vector3 & rgb
        );

public:

    GLubyte rgba[ 4 ];
        /**< Color data. */

private:

    /** Convert GLfloat component value to GLubyte component value. */
    static INLINE GLubyte float2ubyte
        (
        GLfloat x
        );

    /** Convert GLubyte component value to GLfloat component value. */
    static INLINE GLfloat ubyte2float
        (
        GLubyte x
        );
    
    }; // class Color


/**
 * Floating-point RGBA color.
 *
 * Color represented by 4 floating-point values, no clamping is done.
 */
class FloatColor
    {
public:

    /** Default constructor - opaque black. */
    INLINE FloatColor();

    /** Copy constructor. */
    INLINE FloatColor
        (
        const FloatColor & other
        );

    /** Construct from SColor. */
    explicit INLINE FloatColor
        (
        const SColor & other
        );

    /** Construct from RGBA values. */
    INLINE FloatColor
        (
        GLfloat red,
        GLfloat green,
        GLfloat blue,
        GLfloat alpha = 1.0f // opaque
        );

    /** Construct from RGB vector and alpha value. */
    explicit INLINE FloatColor
        (
        const Vector3 & rgb,
        GLfloat         alpha = 1.0f // opaque
        );

    /** Assignment operator. */
    INLINE FloatColor & operator=
        (
        const FloatColor & other
        );

    /** Assign SColor. */
    INLINE FloatColor & operator=
        (
        const SColor & other
        );

public:

    /** Compare colors. */
    INLINE bool operator==
        (
        const FloatColor & other
        ) const;

    /** Compare colors. */
    INLINE bool operator!=
        (
        const FloatColor & other
        ) const;

public:

    /** Get red component. */
    INLINE GLfloat red() const;

    /** Get green component. */
    INLINE GLfloat green() const;

    /** Get blue component. */
    INLINE GLfloat blue() const;

    /** Get alpha component. */
    INLINE GLfloat alpha() const;

    /** Get RGB vector. */
    INLINE Vector3 rgb() const;

    /** Convert to SColor. */
    INLINE SColor scolor() const;

public:

    /** Call @c glColor4. */
    INLINE void gl_color() const;

public:

    /** Set red component. */
    INLINE void set_red
        (
        GLfloat red
        );

    /** Set green component. */
    INLINE void set_green
        (
        GLfloat green
        );

    /** Set blue component. */
    INLINE void set_blue
        (
        GLfloat blue
        );

    /** Set alpha component. */
    INLINE void set_alpha
        (
        GLfloat alpha
        );

    /** Set RGB vector. */
    INLINE void set_rgb
        (
        const Vector3 & rgb
        );

public:

    GLfloat rgba[ 4 ];
        /**< Color data. */

    }; // class FloatColor



} // namespace Demo::Client::OpenGL
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "opengl_types_inline.h"
#endif


#endif // CLIENT_OPENGL_TYPES_H
