/** @file overlay_animation.cpp
 * 2D object animation (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/overlay_animation.h"
#include "renderer/overlay_object.h"
#include "input/input_manager.h"
#include "client_global.h"
#include <iosfwd>


namespace Demo {
namespace Client {
namespace Renderer {


namespace {

/**
 * Animate single color/vector component.
 *
 * @return @c true if @a target_value has been reached, @c false otherwise.
 */
bool animate_component
    (
    Float32 &           result,
    Float32             value,
    Float32             target_value,
    Float32             speed,
    const SystemTime &  time_delta
    )
    {
    hope( speed > 0.0f );
    const Float32 delta = static_cast< Float32 >( time_delta.get_value() );
    result = value;
    if( delta <= 0.0f )
        {
        return false;
        }

    if( result < target_value )
        {
        result += delta * speed;
        if( result >= target_value )
            {
            result = target_value;
            return true;
            }
        }
    else if( result > target_value )
        {
        result -= delta * speed;
        if( result <= target_value )
            {
            result = target_value;
            return true;
            }
        }
    else
        {
        return true;
        }

    return false;
    }

} // anonymous namespace


// class OverlayAnimation

OverlayAnimation::~OverlayAnimation()
    {
    hope( !prev );
    hope( !next );
    }


// class OverlayOffsetAnimation

void OverlayOffsetAnimation::animate
    (
    OverlayObject *     object,
    const SystemTime &  time_delta
    )
    {
    hope( object );

    const SVector2 & offset = object->get_base_offset();
    if( offset != target_offset )
        {
        Float32 x, y;
        animate_component( x, offset.x(), target_offset.x(), speed.x(), time_delta );
        animate_component( y, offset.y(), target_offset.y(), speed.y(), time_delta );
        object->set_base_offset( SVector2( x, y ) );
        }
    }


// class OverlayOffsetDeltaAnimation

void OverlayOffsetDeltaAnimation::animate
    (
    OverlayObject *     object,
    const SystemTime &  time_delta
    )
    {
    hope( object );
    
    object->add_offset_delta( offset_delta );
    }


// class OverlayColorFactorAnimation

void OverlayColorFactorAnimation::animate
    (
    OverlayObject *     object,
    const SystemTime &  time_delta
    )
    {
    hope( object );

    // Animate.

    if( factor != target_factor )
        {
        Float32 r, g, b, a;
        animate_component( r, factor.red(), target_factor.red(), speed.red(), time_delta );
        animate_component( g, factor.green(), target_factor.green(), speed.green(), time_delta );
        animate_component( b, factor.blue(), target_factor.blue(), speed.blue(), time_delta );
        animate_component( a, factor.alpha(), target_factor.alpha(), speed.alpha(), time_delta );
        factor = SColor( r, g, b, a );
        }

    // Apply.

    object->multiply_color_factor( factor );
    }


// class OverlayMousePositionAnimation

void OverlayMousePositionAnimation::animate
    (
    OverlayObject *     object,
    const SystemTime &  time_delta
    )
    {
    hope( object );
    
    InputManager & input_manager = Demo::Client::Global::get_input_manager();
    
    object->set_base_offset( input_manager.get_mouse_position() );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "overlay_animation_inline.h"
#endif
