/** @file overlay_animation.h
 * 2D object animation.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_OVERLAY_ANIMATION_H
#define CLIENT_OVERLAY_ANIMATION_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Overlay object animation base class.
 */
class OverlayAnimation
    {
    /// OverlayObject can apply and link/unlink animations.
    friend class OverlayObject;

    NO_COPY_CONSTRUCTOR( OverlayAnimation );
    NO_ASSIGNMENT_OPERATOR( OverlayAnimation );

public:

    /**
     * Constructor.
     */
    INLINE OverlayAnimation();

    /**
     * Destructor.
     */
    virtual ~OverlayAnimation();

protected:

    /**
     * Apply animation to given overlay @a object.
     */
    virtual void animate
        (
        OverlayObject *     object,
        const SystemTime &  time_delta
        ) = 0;

protected:

    OverlayAnimation *  prev;
        /**< Previous animation in cyclic list.
             Directly modified by #OverlayObject. */

    OverlayAnimation *  next;
        /**< Previous animation in cyclic list.
             Directly modified by #OverlayObject. */
             
    }; // class OverlayAnimation


/**
 * Overlay object animation - linear movement of base offset.
 */
class OverlayOffsetAnimation : public OverlayAnimation
    {
    NO_COPY_CONSTRUCTOR( OverlayOffsetAnimation );
    NO_ASSIGNMENT_OPERATOR( OverlayOffsetAnimation );

public:

    /**
     * Constructor.
     *
     * @param the_speed Movement speed in pixels per second, positive vector.
     * @param the_offset Initial destination offset.
     */
    INLINE OverlayOffsetAnimation
        (
        const SVector2 & the_speed = SVector2( 100.0f, 100.0f ),
        const SVector2 & the_offset = SVector2()
        );

public:

    /**
     * Change movement speed.
     */
    INLINE void set_speed
        (
        const SVector2 & new_speed
        );

    /**
     * Change movement destination offset.
     */
    INLINE void set_target_offset
        (
        const SVector2 & new_offset
        );
        
private:

    /**
     */
    virtual void animate
        (
        OverlayObject *     object,
        const SystemTime &  time_delta
        );

private:

    SVector2        speed;
        /**< Movement speed in pixels per second. */

    SVector2        target_offset;
        /**< Offset towards which to move. */
    
    }; // class OverlayOffsetAnimation

    
/**
 * Overlay object animation - adds offset delta to
 * overlay object using its @c add_offset_delta() method.
 */
class OverlayOffsetDeltaAnimation : public OverlayAnimation
    {
    NO_COPY_CONSTRUCTOR( OverlayOffsetDeltaAnimation );
    NO_ASSIGNMENT_OPERATOR( OverlayOffsetDeltaAnimation );

public:

    /**
     * Constructor.
     *
     * @param the_speed Movement speed in pixels per second, positive vector.
     * @param the_offset Initial destination offset.
     */
    INLINE OverlayOffsetDeltaAnimation
        (
        const SVector2 & the_offset_delta
        );

public:

    /**
     * Change offset delta.
     */
    INLINE void set_offset_delta
        (
        const SVector2 & new_offset_delta
        );

private:

    /**
     */
    virtual void animate
        (
        OverlayObject *     object,
        const SystemTime &  time_delta
        );

private:

    SVector2        offset_delta;
        /**< Offset delta. */

    }; // class OverlayOffsetDeltaAnimation
    

/**
 * Overlay object animation - color factor multiplier with linear animation.
 */
class OverlayColorFactorAnimation : public OverlayAnimation
    {
    NO_COPY_CONSTRUCTOR( OverlayColorFactorAnimation );
    NO_ASSIGNMENT_OPERATOR( OverlayColorFactorAnimation );

public:

    /**
     * Constructor.
     *
     * @param the_speed Change speed in color component units per second, positive color.
     * @param the_factor Initial factor.
     */
    INLINE OverlayColorFactorAnimation
        (
        const SColor & the_speed = SColor( 1.0f, 1.0f, 1.0f, 1.0f ),
        const SColor & the_factor = SColor( 1.0f, 1.0f, 1.0f, 1.0f )
        );

public:

    /**
     * Change speed.
     */
    INLINE void set_speed
        (
        const SColor & new_speed
        );

    /**
     * Change color factor immediately.
     */
    INLINE void set_factor
        (
        const SColor & new_factor
        );

    /**
     * Change target color factor.
     */
    INLINE void set_target_factor
        (
        const SColor & new_target_factor
        );

private:

    /**
     */
    virtual void animate
        (
        OverlayObject *     object,
        const SystemTime &  time_delta
        );

private:

    SColor          factor;
        /**< Current factor. */

    SColor          speed;
        /**< Movement speed in pixels per second. */

    SColor          target_factor;
        /**< Target color factor. */

    }; // class OverlayColorFactorAnimation


/**
 * Overlay object animation.
 * Not a real animation, only set base offset to current mouse position
 * on screen.
 */
class OverlayMousePositionAnimation : public OverlayAnimation
    {
    NO_COPY_CONSTRUCTOR( OverlayMousePositionAnimation );
    NO_ASSIGNMENT_OPERATOR( OverlayMousePositionAnimation );

public:

    /**
     * Constructor.
     *
     */
    INLINE OverlayMousePositionAnimation();

private:

    /**
     */
    virtual void animate
        (
        OverlayObject *     object,
        const SystemTime &  time_delta
        );

    }; // class OverlayMousePositionAnimation

    

} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo



#ifdef ENABLE_INLINE
#include "overlay_animation_inline.h"
#endif


#endif // CLIENT_OVERLAY_ANIMATION_H
