/** @file overlay_manager.cpp
 * Manager of 2D overlay objects (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/overlay_manager.h"
#include "renderer/overlay_object.h"
#include "renderer/pipeline_state.h"
#include "sdl/client_sdl.h"
#include "global/global.h"
#include <sstream>


namespace Demo {
namespace Client {
namespace Renderer {


OverlayManager::OverlayManager()
    {
    }


OverlayManager::~OverlayManager()
    {
    // We never keep empty layers, so layers.empty() <=> no objects registered.
    
    if( !layers.empty() )
        {
        // Unregister objects.

        int object_count = 0;
        Layers::iterator layer_it;
        while( ( layer_it = layers.begin() ) != layers.end() )
            {
            Objects & objects = layer_it->second;
            const Objects::iterator object_it = objects.begin();
            hope( object_it != objects.end() );
            OverlayObject * const object = *object_it;
            hope( object );
            object->unregister_from_manager();
            object_count++;
            }

        // Log warning.

        {//sob
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << object_count << " objects still registered in ~OverlayManager.";
        ::Massiv::Core::Global::log( Status::warning( Status::FACILITY_GENERIC,
                    Status::PRIORITY_NORMAL), oss.str() );
        }//eob

        }
    }


void OverlayManager::render
    (
    const SystemTime & time_delta
    )
    {
    // Set screenspace projection.

        {
        const int screen_width = SDL::get_screen_width();
        const int screen_height = SDL::get_screen_height();
        glMatrixMode( GL_PROJECTION );
        glLoadIdentity();
        glOrtho( 0, screen_width, screen_height, 0, -1.0f, +1.0f );
        glMatrixMode( GL_MODELVIEW );
        }
    
    // Render registered objects.
    
    Layers::iterator layer_it;
    for( layer_it = layers.begin(); layer_it != layers.end(); ++layer_it )
        {
        Objects & objects = layer_it->second;
        Objects::iterator object_it;
        for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
            {
            OverlayObject * const object = *object_it;
            hope( object );
            
            render_overlay_object( object, false, time_delta );
            }
        }

    // Done.

    OpenGL::check_errors( "Overlay rendering: " );
    }


void OverlayManager::render_overlay_object
    (
    OverlayObject *     object,
    bool                animate_only,
    const SystemTime &  time_delta
    )
    {
    hope( object );
    
    // Animate the object.
    object->animate( time_delta );
    
    bool visible = object->is_visible();

    // Draw the object.
    if( visible && !animate_only )
        {
        object->render();
        }
    
    // Update animate_only for child overlay objects.
    if( !animate_only )
        {
        if( !visible )
            {
            // Do not draw child objects if parent is not visible.
            animate_only = true;
            }
        }
    
    // Render its child overlay objects.
        {
        OverlayObject::Children & children = object->children;
        OverlayObject::Children::iterator child_it;
        for( child_it = children.begin(); child_it != children.end(); ++child_it )
            {
            OverlayObject * child = *child_it;
            hope( child );
            render_overlay_object( child, animate_only, time_delta );
            }
        }
    }


void OverlayManager::initialize_pipeline_state
    (
    PipelineState & state
    )
    {
    state.alpha_test.enable = true;
    state.alpha_test.function = GL_GREATER;
    state.alpha_test.value = 0.05f;

    state.blend.enable = true;
    state.blend.source = GL_SRC_ALPHA;
    state.blend.destination = GL_ONE_MINUS_SRC_ALPHA;
    }


void OverlayManager::register_object
    (
    OverlayObject *     object,
    int                 layer
    )
    {
    hope( object );
    
    // Register the object.

    Objects & objects = layers[ layer ];
    objects.push_back( object );
    }


void OverlayManager::unregister_object
    (
    OverlayObject *     object,
    int                 layer
    )
    {
    hope( object );

    // Get the layer.

    Layers::iterator layer_it = layers.find( layer );
    hope( layer_it->first == layer );
    Objects & objects = layer_it->second;

    // Unregister the object.

    Objects::iterator object_it;
    for( object_it = objects.begin(); object_it != objects.end(); ++object_it )
        {
        OverlayObject * const registered_object = *object_it;
        hope( registered_object );
        if( registered_object == object )
            {
            objects.erase( object_it );

            // Remove layer if it's empty.

            if( objects.empty() )
                {
                layers.erase( layer_it );
                }

            // Done.
            
            return;
            }
        }

    panic( "Object not registered in unregister_object()." );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "overlay_manager_inline.h"
#endif
