/** @file overlay_manager.h
 * Manager of 2D overlay objects.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_OVERLAY_MANAGER_H
#define CLIENT_OVERLAY_MANAGER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#include <vector>
#include <map>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Manager of 2D overlay objects.
 *
 * Register #OverlayObject objects to manager's layers.
 * Manager's #render() method will then render registered objects in
 * bottom-to-front order.
 *
 * Parent overlay object is renderer before rendering of its child objects.
 * Child objects should not be registered in #OverlayManager.
 *
 * OverlayManager doesn't draw child overlay object if its parent
 * is not visible (but the OverlayManager animate the child object).
 *
 * Registered objects are not owned by the manager and should be unregistered
 * before the manager is destroyed.
 */
class OverlayManager
    {
    NO_COPY_CONSTRUCTOR( OverlayManager );
    NO_ASSIGNMENT_OPERATOR( OverlayManager );

public:

    /**
     * Constructor.
     */
    OverlayManager();

    /**
     * Destructor.
     *
     * All objects should be unregistered before destruction of the manager.
     * If some objects are still registered, they will be unregistered
     * automatically but warning will be logged.
     */
    ~OverlayManager();

public:

    /**
     * Render overlays.
     */
    void render
        (
        const SystemTime & time_delta
        );

private:

    /**
     * Render overlay object and its parent.
     * @param object Overlay object to draw.
     * @param animate_only Whether the object and its children should be
     *      animated only.
     */
    void render_overlay_object
        (
        OverlayObject *     object,
        bool                animate_only,
        const SystemTime &  time_delta
        );

public:

    /**
     * Change given standard pipeline state to overlay pipeline state.
     *
     * Several settings differ from standard state:
     * - Alpha testing is enabled, with function GL_GREATER and tiny
     *   reference value (ie. throw away invisible texels).
     * - Blending is enabled, with function set to
     *   <tt>src * SRC_ALPHA + dst * ONE_MINUS_SRC_ALPHA</tt>.
     *
     * @pre @a state is default-initialized.
     */
    static void initialize_pipeline_state
        (
        PipelineState & state
        );

protected:

    /// OverlayObject can register/unregister itself to/from the manager.
    friend class OverlayObject;

    /**
     * Register overlay object.
     */
    void register_object
        (
        OverlayObject *     object,
        int                 layer
        );

    /**
     * Unregister overlay object.
     */
    void unregister_object
        (
        OverlayObject *     object,
        int                 layer
        );
     
private:

    /**
     * Vector of aliasing OverlayObject pointers.
     */
    typedef std::vector< OverlayObject * > Objects;

    /**
     * Map from layer index to Objects.
     */
    typedef std::map< int, Objects > Layers;

private:

    Layers          layers;
        /**< Layers with registered objects. */
        
    }; // class OverlayManager


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "overlay_manager_inline.h"
#endif


#endif // CLIENT_OVERLAY_MANAGER_H
