/** @file overlay_object.cpp
 * Generic 2D object (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/overlay_object.h"
#include "renderer/overlay_manager.h"
#include "renderer/pass_manager.h"
#include "renderer/renderer.h"
#include "sdl/client_sdl.h"
#include "utility/debug.h"
#include <sstream>
#include <algorithm>


namespace Demo {
namespace Client {
namespace Renderer {


OverlayObject::OverlayObject()
      : origin( TOP_LEFT ),
        color( 1.0f, 1.0f, 1.0f, 1.0f ),
        color_factor( 1.0f, 1.0f, 1.0f, 1.0f ),
        visible( true ),
        position_dirty( true ),
        size_dirty( true ),
        color_dirty( true ),
        manager( NULL ),
        layer( 0 ),
        animation( NULL ),
        parent( NULL )
    {
    }


OverlayObject::~OverlayObject()
    {
    // Unlink self from parent objects.
    set_parent_overlay_object( NULL );

    // Unlink all child objects.
    while( !children.empty() )
        {
        OverlayObject * child = children[ 0 ];
        hope( child );
        child->set_parent_overlay_object( NULL );
        }
    
    destroy_animation_objects();
    unregister_from_manager();
    }


void OverlayObject::center_object()
    {
    const Float32 width = static_cast< Float32 >( SDL::get_screen_width() );
    const Float32 height = static_cast< Float32 >( SDL::get_screen_height() );
    set_origin( TOP_LEFT );
    set_base_offset( SVector2( 0.5f * ( width - size.x() ), 0.5f * ( height - size.y() ) ) );
    }


void OverlayObject::set_base_size_from_screen()
    {
    set_base_size( SVector2
        (
        static_cast< Float32 >( SDL::get_screen_width() ),
        static_cast< Float32 >( SDL::get_screen_height() )
        ) );
    }


void OverlayObject::register_to_manager
    (
    OverlayManager *    new_manager,
    int                 new_layer
    )
    {
    hope( new_layer >= 0 );
    if( manager )
        {
        manager->unregister_object( this, layer );
        manager = NULL;
        layer = 0;
        }
    if( new_manager )
        {
        manager = new_manager;
        layer = new_layer;
        manager->register_object( this, layer );
        }
    }


void OverlayObject::register_to_global_manager
    (
    int pass_index,
    int new_layer
    )
    {
    PassManager & pass_manager = Global::get_pass_manager();
    OverlayManager & overlay_manager = pass_manager.create_overlay_manager( pass_index );
    register_to_manager( &overlay_manager, new_layer );
    }


void OverlayObject::set_parent_overlay_object
    (
    OverlayObject * new_parent
    )
    {
    // Unlink from current parent.
    if( parent )
        {
        parent->unlink_child_overlay_object( this );
        parent = NULL;
        }

    // Link to new parent.
    if( new_parent )
        {
        new_parent->link_child_overlay_object( this );
        parent = new_parent;
        }
    }

    
void OverlayObject::add_animation_object
    (
    OverlayAnimation * animation_object
    )
    {
    hope( animation_object );
    hope( !animation_object->prev );
    hope( !animation_object->next );

    // Link it.

    if( animation )
        {
        OverlayAnimation * const last = animation->prev;
        hope( last );
        animation_object->prev = last;
        animation_object->next = animation;
        last->next = animation_object;
        animation->prev = animation_object;
        }
    else
        {
        animation = animation_object;
        animation_object->prev = animation_object->next = animation_object;
        }
    }


OverlayAnimation * OverlayObject::get_animation
    (
    const std::type_info &  type,
    int                     index
    )
    {
    hope( index >= 0 );

    OverlayAnimation * result = animation;
    while( result )
        {
        if( typeid( *result ) == type )
            {
            if( index == 0 )
                {
                return result;
                }
            index--;
            }
        result = result->next;
        }

    return NULL;
    }


OverlayAnimation * OverlayObject::require_animation
    (
    const std::type_info &  type,
    int                     index
    )
    {
    OverlayAnimation * const result = get_animation( type, index );
    if( !result )
        {
        std::ostringstream oss;
        oss << "There is no animation of type " << type.name() <<
            " with index " << index << ".";
        panic( oss.str().c_str() );
        }
    return result;
    }


void OverlayObject::remove_animation_object
    (
    OverlayAnimation * animation_object
    )
    {
    hope( animation_object );
    hope( animation_object->prev );
    hope( animation_object->next );

    // Unlink it.

    OverlayAnimation * const prev = animation_object->prev;
    hope( prev );
    OverlayAnimation * const next = animation_object->next;
    hope( next );

    if( prev == animation_object )
        {
        hope( next == animation_object );
        hope( animation == animation_object );
        animation = NULL;
        }
    else
        {
        hope( next != animation_object );
        prev->next = next;
        next->prev = prev;
        if( animation == animation_object )
            {
            animation = next;
            }
        }
    animation_object->prev = animation_object->next = NULL;
    }


void OverlayObject::destroy_animation_objects()
    {
    while( animation )
        {
        OverlayAnimation * const to_delete = animation;
        remove_animation_object( animation );
        delete to_delete;
        }
    }


void OverlayObject::animate
    (
    const SystemTime & time_delta
    )
    {
    if( animation == NULL )
        {
        return;
        }

    // Reset deltas and factors.

    if( offset_delta != SVector2() )
        {
        offset_delta = SVector2();
        position_dirty = true;
        }

    if( size_delta != SVector2() )
        {
        size_delta = SVector2();
        size_dirty = true;
        }

    if( ( color_delta != SColor() ) ||
        ( color_factor != SColor( 1.0f, 1.0f, 1.0f, 1.0f ) ) )
        {
        color_delta = SColor();
        color_factor = SColor( 1.0f, 1.0f, 1.0f, 1.0f );
        color_dirty = true;
        }

    // Animate.

    OverlayAnimation * a = animation;
    do  {
        a->animate( this, time_delta );
        a = a->next;
        } while( a != animation );
    }


OpenGL::Vector2 OverlayObject::screenspace_translation() const
    {
    GLfloat     parent_width;
    GLfloat     parent_height;
    
    if( parent )
        {
        const SVector2 parent_size = parent->get_current_size();
        parent_width = parent_size.x();
        parent_height = parent_size.y();
        }
    else
        {
        parent_width = static_cast< GLfloat >( SDL::get_screen_width() );
        parent_height = static_cast< GLfloat >( SDL::get_screen_height() );
        }
    
    // Compute object position relative to parent's top left corner.
    OpenGL::Vector2 object_position( get_current_offset() );
    const SVector2  object_size( get_current_size() );

    switch( origin )
        {
        case TOP_LEFT:
            break;

        case TOP_RIGHT:
            object_position = OpenGL::Vector2
                (
                parent_width - object_position.x() - object_size.x(),
                object_position.y()
                );
            break;

        case BOTTOM_LEFT:
            object_position = OpenGL::Vector2
                (
                object_position.x(),
                parent_height - object_position.y() - object_size.y()
                );
            break;

        case BOTTOM_RIGHT:
            object_position = OpenGL::Vector2
                (
                parent_width - object_position.x() - object_size.x(),
                parent_height - object_position.y() - object_size.y()
                );
            break;
        }
    
    if( parent )
        {
        const OpenGL::Vector2 parent_screen_position = parent->screenspace_translation();
        object_position = OpenGL::Vector2
            (
            object_position.x() + parent_screen_position.x(),
            object_position.y() + parent_screen_position.y()
            );
        }

    return object_position;
    }

    
SVector2 OverlayObject::inverse_screenspace_translation
    (
    const SVector2 screen_position
    ) const
    {
    const SVector2 parent_screen_position = parent->screenspace_translation().svector();
    const SVector2 screen_delta = screen_position - parent_screen_position;
    SVector2 result;
    
    GLfloat     parent_width;
    GLfloat     parent_height;
    
    if( parent )
        {
        const SVector2 parent_size = parent->get_current_size();
        parent_width = parent_size.x();
        parent_height = parent_size.y();
        }
    else
        {
        parent_width = static_cast< GLfloat >( SDL::get_screen_width() );
        parent_height = static_cast< GLfloat >( SDL::get_screen_height() );
        }
        
    const SVector2  object_size( get_current_size() );

    switch( origin )
        {
        case TOP_LEFT:
            result = screen_delta;
            break;

        case TOP_RIGHT:
            result = SVector2
                (
                parent_width - screen_delta.x() - object_size.x(),
                screen_delta.y()
                );
            break;

        case BOTTOM_LEFT:
            result = SVector2
                (
                screen_delta.x(),
                parent_height - screen_delta.y() - object_size.y()
                );
            break;

        case BOTTOM_RIGHT:
            result = SVector2
                (
                parent_width - screen_delta.x() - object_size.x(),
                parent_height - screen_delta.y() - object_size.y()
                );
            break;
        }
    
    return result;
    }
    

PipelineState & OverlayObject::get_pipeline_state()
    {
    return Global::get_overlay_pipeline_state();
    }


void OverlayObject::link_child_overlay_object
    (
    OverlayObject * child
    )
    {
    hope( child );
    hope( !child->get_parent_overlay_object() );
    hope( std::find( children.begin(), children.end(), child ) == children.end() );
    
    children.push_back( child );
    }

    
void OverlayObject::unlink_child_overlay_object
    (
    OverlayObject * child
    )
    {
    hope( child );
    hope( child->get_parent_overlay_object() == this );
    
    Children::iterator child_it = std::find( children.begin(), children.end(), child );
    hope( child_it != children.end() );
    
    children.erase( child_it );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "overlay_object_inline.h"
#endif
