/** @file overlay_object.h
 * Generic 2D object.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_OVERLAY_OBJECT_H
#define CLIENT_OVERLAY_OBJECT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_TYPES_H
#include "renderer/opengl_types.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_OVERLAY_ANIMATION_H
#include "renderer/overlay_animation.h"
#endif

#include <typeinfo>
#include <vector>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Generic 2D overlay object.
 *
 * Overlay objects have position (relative to one of the screen corners) and
 * color (including alpha) and are renderable.
 *
 * They can be used as standalone objects, but it's recommended to register
 * them to #OverlayManager and let the manager render them.
 *
 * Overlay object can have child overlay objects.
 * Position of child object is relative to its parent object.
 * Child overlay object can have only one parent. Parent object
 * can have unlimited number of child objects.
 *
 * Drawing of child overlay objects is managed by #OverlayManager.
 * Parent object is renderer before rendering of its child objects.
 * Child objects should not be registered in #OverlayManager.
 *
 * #OverlayManager doesn't draw child overlay object if its parent
 * is not visible (but the #OverlayManager animate the child object).
 */
class OverlayObject
    {
    NO_COPY_CONSTRUCTOR( OverlayObject );
    NO_ASSIGNMENT_OPERATOR( OverlayObject );
    
    friend class OverlayManager; // access to children

public:

    /**
     * One of the four corners of object/screen.
     */
    enum ScreenCorner
        {
        TOP_LEFT,       ///< Top-left corner.
        TOP_RIGHT,      ///< Top-right corner.
        BOTTOM_LEFT,    ///< Bottom-left corner.
        BOTTOM_RIGHT,   ///< You guess.
        };

public:

    /**
     * Constructor.
     *
     * Default OverlayObject is positioned at TOP_LEFT corner, it's opaque
     * white, and not registered to manager.
     */
    OverlayObject();

    /**
     * Destructor.
     * The objects unlinks self from its parent overlay object
     * and unlinks all child objects.
     *
     * Unregisters the object if it's registered. All registered
     * OverlayObjects must be destroyed before the manager.
     *
     * @warning Do not destroy overlay objects while rendering
     *      is in progress! See @c set_parent_overlay_object().
     */
    virtual ~OverlayObject();

public:

    /** Should OverlayManager draw the object? */
    INLINE bool is_visible() const;

    /** Get position origin. */
    INLINE ScreenCorner get_origin() const;

    /** Get base position offset (relative to parent). */
    INLINE const SVector2 & get_base_offset() const;

    /** Get base object width and height. */
    INLINE const SVector2 & get_base_size() const;

    /** Get base object color. */
    INLINE const SColor & get_base_color() const;

public:

    /** Get current position offset (#offset + #offset_delta) (relative to parent). */
    INLINE SVector2 get_current_offset() const;

    /** Get current object width and height (#size + #size_delta). */
    INLINE SVector2 get_current_size() const;

    /** Get current object color (#color * #color_factor + #color_delta). */
    INLINE SColor get_current_color() const;

public:

    /**
     * Enable or disable visibility of the object.
     * Note that this is ignored unless object is registered to OverlayManager.
     */
    INLINE void set_visibility
        (
        bool visible
        );

    /** Set position origin. */
    INLINE void set_origin
        (
        ScreenCorner new_origin
        );

    /** Set base position offset (relative to parent). */
    INLINE void set_base_offset
        (
        const SVector2 & new_offset
        );

    /**
     * Set object origin and offset to center it to screen.
     */
    void center_object();

    /**
     * Set object size.
     */
    INLINE void set_base_size
        (
        const SVector2 & new_size
        );

    /**
     * Set object size to size of screen.
     */
    void set_base_size_from_screen();

    /**
     * Set base object color.
     * This color will multiply any overlay object colors. Alpha 1.0 means
     * object will be opaque, objects with zero alpha will be invisible
     * (fully transparent).
     *
     * Not that if alpha is zero, #OverlayManager will not even call #render()
     * method of the object.
     */
    INLINE void set_base_color
        (
        const SColor & new_color
        );

public:

    /**
     * Add @a delta to position offset delta.
     *
     * This value should be modified by #OverlayAnimation objects.
     * It's will be set to zero vector by #animate() before
     * #OverlayAnimation::animate() is called for each registered animatior.
     */
    INLINE void add_offset_delta
        (
        const SVector2 & delta
        );

    /**
     * Add @a delta to size delta.
     *
     * This value should be modified by #OverlayAnimation objects.
     * It's will be set to zero vector by #animate() before
     * #OverlayAnimation::animate() is called for each registered animatior.
     *
     * Size delta is applied only to #size, not to #offset. This means that
     * corner specified by #origin will not move fixed, not object centerpoint.
     */
    INLINE void add_size_delta
        (
        const SVector2 & delta
        );

    /**
     * Add @a delta to color delta.
     *
     * This value should be modified by #OverlayAnimation objects.
     * It's will be set to transparent black (0,0,0,0) by #animate() before
     * #OverlayAnimation::animate() is called for each registered animatior.
     */
    INLINE void add_color_delta
        (
        const SColor & delta
        );

    /**
     * Multiply color factor by @a factor.
     *
     * This value should be modified by #OverlayAnimation objects.
     * It's will be set to opaque white (1,1,1,1) by #animate() before
     * #OverlayAnimation::animate() is called for each registered animatior.
     */
    INLINE void multiply_color_factor
        (
        const SColor & factor
        );

public:

    /**
     * Register overlay object to overlay manager.
     *
     * You can re-register the object. Use @c NULL as @a new_manager
     * to unregister the object.
     */
    void register_to_manager
        (
        OverlayManager *    new_manager,
        int                 new_layer
        );

    /**
     * Register overlay object to overlay manager owned by global pass @a pass_index.
     *
     * If there is not overlay manager allocated for given pass, it will be
     * automatically created.
     */
    void register_to_global_manager
        (
        int pass_index,
        int new_layer
        );

    /**
     * Unregister overlay object from the manager
     *
     * Shortcut for <tt>register_to_manager( NULL, 0 )</tt>.
     */
    INLINE void unregister_from_manager();
    
public:

    /**
     * Link the object to @a parent overlay objects as its child.
     * @param parent Parent overlay object. Pass @a NULL to unlink the object
     *      from its current parent.
     * @warning Do not link/unlink overlay objects while rendering
     *      is in progress!
     */
    void set_parent_overlay_object
        (
        OverlayObject * parent
        );
    
    /** 
     * Get parent overlay object.
     * @return Current parent overlay object or @c NULL.
     */
    INLINE OverlayObject * get_parent_overlay_object() const;

public:

    /**
     * Add new animation object to animate the overlay object.
     *
     * The animation object will be owned by the overlay object.
     */
    void add_animation_object
        (
        OverlayAnimation * animation_object
        );

    //@{
    /** Get pointer to @a index-th animation object of given type. */
    OverlayAnimation * get_animation
        (
        const std::type_info &  type,
        int                     index = 0
        );

#if 0
    // Nice, but can't be used because of MSVC6 stupidity.
    // So we implement all the methods manually.
    template< typename Type >
    Type * get_animation_type
        (
        int index = 0
        )
        {
        return static_cast< Type * >( get_animation( typeid( Type ), index ) );
        }
#endif

    INLINE OverlayOffsetAnimation * get_offset_animation
        (
        int index = 0
        );

    INLINE OverlayColorFactorAnimation * get_color_factor_animation
        (
        int index = 0
        );
    //@}

    //@{

    /**
     * Get pointer to @a index-th animation object of given type.
     * These methods panic if there is no such animation registered.
     */
    OverlayAnimation * require_animation
        (
        const std::type_info &  type,
        int                     index = 0
        );

    INLINE OverlayOffsetAnimation & require_offset_animation
        (
        int index = 0
        );

    INLINE OverlayColorFactorAnimation & require_color_factor_animation
        (
        int index = 0
        );
    //@}

    /**
     * Remove given @a animation_object from list of animations.
     *
     * Ownership of animation object will be returned to caller.
     */
    void remove_animation_object
        (
        OverlayAnimation * animation_object
        );

    /**
     * Destroy all registered animation objects.
     */
    void destroy_animation_objects();
    
public:

    /**
     * Render the object.
     *
     * When #OverlayManager calls #render(), it sets modelview matrix to
     * identity and projection matrix to screenspace ortho projection:
     * coordinates are in pixels, ( 0, 0 ) is top-left corner of the screen.
     */
    virtual void render() = 0;

    /**
     * Animate the object.
     *
     * This will be called by the #OverlayManager if the object is registered.
     * Deltas and factors will be reset to their default values and
     * #OverlayAnimation::animate() will be called for each registered
     * animation object.
     */
    virtual void animate
        (
        const SystemTime & time_delta
        );

protected:

    /**
     * Convert #origin and #offset to screenspace translation.
     *
     * Because #offset is relative to parent overlay object, this
     * function traverses hiearchy of parent overlay objects to properly
     * compute screnspace translation of child overlay objects.
     */
    OpenGL::Vector2 screenspace_translation() const;

    /**
     * Inverse function to screenspace_translation().
     * If you set offset of the overlay object to returned offset,
     * the screenspace_translation() will return @a screen_position.
     */
    SVector2 inverse_screenspace_translation
        (
        const SVector2 screen_position
        ) const;
    
    /**
     * Get global overlay pipeline state from #Renderer.
     */
    PipelineState & get_pipeline_state();

private:

    /**
     * Link @a child overlay object.
     */
    void link_child_overlay_object
        (
        OverlayObject * child
        );
    
    /**
     * Unlink @a child overlay object.
     */
    void unlink_child_overlay_object
        (
        OverlayObject * child
        );
    
private:

    ScreenCorner        origin;
        /**< Origin used to specify object position.
             Object position specifies position of @a origin corner of the
             object relative to @a origin corner of its parent overlay object
             or the screen (if the object doesn't have parent object). Positive
             coordinates move the object into the screen. For example if
             @a origin is @c BOTTOM_RIGHT and @a offset is ( 10, 20 ),
             object's bottom-right corner will be placed 10 pixels up from
             the bottom and 20 pixels left of the right edge of its parent
             or the screen. */

    SVector2            offset;
        /**< Current object position. */

    SVector2            size;
        /**< Object width and height. */

    SColor              color;
        /**< Current object color, including alpha.
             Alpha should be interpreted as transparency value by the object,
             1.0 means opaque, 0.0 fully transparent. */

private:

    SVector2            offset_delta;
        /**< Position offset delta.
             Set to zero vector by #animate(), changed by animation objects. */

    SVector2            size_delta;
        /**< Size delta.
             Set to zero vector by #animate(), changed by animation objects. */

    SColor              color_delta;
        /**< Color delta.
             Set to transparent black by #animate(), changed by animation objects. */

    SColor              color_factor;
        /**< Color multiplication factor.
             Set to opaque white by #animate(), changed by animation objects. */

protected:

    bool                visible : 1;
        /**< Draw this object?
             This is used only by #OverlayManager. */

    bool                position_dirty : 1;
        /**< #origin or #offset changed. */

    bool                size_dirty : 1;
        /**< #size changed. */

    bool                color_dirty : 1;
        /**< #color changed. */

private:

    OverlayManager *    manager;
        /**< Manager this object is registered to. It may be NULL. */

    int                 layer;
        /**< Overlay layer.
             0 is the bottommost layer, MAX_INT is the topmost layer.
             Ignored unless registered to manager. */

private:

    OverlayAnimation *  animation;
        /**< First element in cyclic list of owned animation objects. */
    
private:

    /**
     * Array of aliasing pointers to child overlay objects.
     */
    typedef std::vector< OverlayObject * > Children;

    OverlayObject *     parent;
        /**< Aliasing pointer to parent overlay object. */
    
    Children            children;
        /**< Aliasing pointers to child overlay objects. */

    }; // class OverlayObject


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "overlay_object_inline.h"
#endif


#endif // CLIENT_OVERLAY_OBJECT_H
