/** @file pass_manager.h
 * Manager of rendering passes.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_PASS_MANAGER_H
#define CLIENT_PASS_MANAGER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OVERLAY_MANAGER_H
#include "renderer/overlay_manager.h"
#endif

#ifndef CLIENT_RENDER_MANAGER_H
#include "renderer/render_manager.h"
#endif

#include <map>
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Manager of rendering passes.
 *
 * #PassManager manages rendering passes and renders them in specified order.
 * See #PassManager::Pass for info about pass contents.
 *
 * Use #create_pass() or shortcut #create_overlay_manager() and
 * #create_render_manager() to create new pass or pass with overlay manager
 * or pass with render manager. Reference to existing pass/manager will be
 * returned if it already exists. Use @c get_*() version of the methods to
 * get the reference if you are sure it already exists. Those methods
 * @c panic if you try to get object that does not exist.
 */
class PassManager
    {
    NO_COPY_CONSTRUCTOR( PassManager );
    NO_ASSIGNMENT_OPERATOR( PassManager );

public:

    /**
     * Rendering pass.
     *
     * Each pass may contain any of the following:
     * - 2D overlay managed by #OverlayManager.
     * - 3D entities managed by #RenderManager.
     * If both 2D and 3D objects exist in single pass, overlays will be
     * rendered first.
     */
    class Pass
        {
    public:

        /**
         * Map from pass index to Pass.
         */
        typedef std::map< int, Pass > Map;
    
    public:

        /**
         * Constructor.
         */
        INLINE Pass();

        /**
         * Destructor.
         */
        INLINE ~Pass();

        /**
         * Copy constructor.
         * Ownership of the managers will be transferred.
         */
        INLINE Pass
            (
            const Pass & other
            );

        /**
         * Assignment operator.
         * Ownership of the managers will be transferred.
         */
        INLINE Pass & operator=
            (
            const Pass & other
            );

    public:

        /**
         * Create overlay manager if it does not exist yet.
         */
        INLINE OverlayManager & create_overlay_manager();

        /**
         * Create render manager if it does not exist yet.
         */
        INLINE RenderManager & create_render_manager();

        /**
         * Does overlay manager exist in specified pass?
         */
        INLINE bool has_overlay_manager();

        /**
         * Does render manager exist in specified pass?
         */
        INLINE bool has_render_manager();

        /**
         * Get pass overlay manager.
         *
         * It's error to call this method if overlay manager does not exist
         * in this pass.
         */
        INLINE OverlayManager & get_overlay_manager();

        /**
         * Get pass render manager.
         *
         * It's error to call this method if render manager does not exist
         * in this pass.
         */
        INLINE RenderManager & get_render_manager();

    public:

        /**
         * Render the pass.
         */
        void render
            (
            const SystemTime & time_delta
            );

    private:
        // Hack: mutable because std::auto_ptr is not CopyConstructible and Assignable.

        mutable std::auto_ptr< OverlayManager > overlay_manager;
            /**< Owned overlay manager. */

        mutable std::auto_ptr< RenderManager >  render_manager;
            /**< Owned render manager. */
    
        }; // class Pass

public:

    /**
     * Construct empty pass manager.
     */
    PassManager();

    /**
     * Destroy all the passes.
     */
    ~PassManager();

public:

    /**
     * Create rendering pass.
     */
    INLINE Pass & create_pass
        (
        int pass_index
        );

    /**
     * Create overlay manager in given pass.
     */
    INLINE OverlayManager & create_overlay_manager
        (
        int pass_index
        );

    /**
     * Create render manager in given pass.
     */
    INLINE RenderManager & create_render_manager
        (
        int pass_index
        );

    /**
     * Does specified rendering pass exist?.
     */
    INLINE bool has_pass
        (
        int pass_index
        );

    /**
     * Does overlay manager exist in specified pass?
     */
    INLINE bool has_overlay_manager
        (
        int pass_index
        );

    /**
     * Does rendering manager exist in specified pass?
     */
    INLINE bool has_render_manager
        (
        int pass_index
        );

    /**
     * Get existing rendering pass or panic.
     */
    INLINE Pass & get_pass
        (
        int pass_index
        );

    /**
     * Get existing overlay manager or panic.
     */
    INLINE OverlayManager & get_overlay_manager
        (
        int pass_index
        );

    /**
     * Get existing rendering manager or panic.
     */
    INLINE RenderManager & get_render_manager
        (
        int pass_index
        );

public:

    /**
     * Render all the passes.
     */
    void render
        (
        const SystemTime & time_delta
        );

private:

    Pass::Map       passes;
        /**< Rendering passes. */

    }; // class PassManager


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "pass_manager_inline.h"
#endif


#endif // CLIENT_PASS_MANAGER_H
