/** @file pipeline_state.cpp
 * Rendering pipeline state (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/pipeline_state.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include "sdl/client_sdl.h"


namespace Demo {
namespace Client {
namespace Renderer {


// class PipelineState

PipelineState::PipelineState()
      : polygon_mode( GL_FILL ),
        shade_model( GL_SMOOTH ),
        color( 1.0f, 1.0f, 1.0f, 1.0f )
    {
    }


void PipelineState::apply_constant_defaults()
    {
    const int width = SDL::get_screen_width();
    const int height = SDL::get_screen_height();
    hope( width > 0 );
    hope( height > 0 );
    glViewport( 0, 0, width, height );

    glClearColor( 0, 0, 0, 0 );
    glEnable( GL_NORMALIZE ); // FIXME: Not always?

    OpenGL::check_errors( "Pipeline constant defaults: " );
    }


void PipelineState::apply_defaults()
    {
    Cull::apply_defaults();
    DepthOffset::apply_defaults();
    AlphaTest::apply_defaults();
    DepthTest::apply_defaults();
    BufferMask::apply_defaults();
    Blend::apply_defaults();

    TextureManager & texture_manager = Global::get_texture_manager();
    const int units = texture_manager.get_texture_units();
    int i;
    for( i = 0; i < units; i++ )
        {
        Texture::apply_defaults( i );
        }

    Lighting::apply_defaults();
    Fog::apply_defaults();

    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    glShadeModel( GL_SMOOTH );
    glColor4ub( 255, 255, 255, 255 );

    OpenGL::check_errors( "Pipeline defaults: " );
    }


void PipelineState::apply
    (
    const PipelineState & other
    )
    {
    cull.apply( other.cull );
    depth_offset.apply( other.depth_offset );
    alpha_test.apply( other.alpha_test );
    depth_test.apply( other.depth_test );
    buffer_mask.apply( other.buffer_mask );
    blend.apply( other.blend );

    TextureManager & texture_manager = Global::get_texture_manager();
    const int units = texture_manager.get_texture_units();
    int i;
    for( i = 0; i < units; i++ )
        {
        texture[ i ].apply( other.texture[ i ], i );
        }

    lighting.apply( other.lighting );
    fog.apply( other.fog );

    if( polygon_mode != other.polygon_mode )
        {
        polygon_mode = other.polygon_mode;
        glPolygonMode( GL_FRONT_AND_BACK, polygon_mode );
        }

    if( shade_model != other.shade_model )
        {
        shade_model = other.shade_model;
        glShadeModel( shade_model );
        }

    // Note: In current implementation we always set the color, because it
    //       could be changed by vertex arrays.

    color = other.color;
    color.gl_color();

    OpenGL::check_errors( "Apply pipeline state: " );
    }


// class PipelineState::Cull

PipelineState::Cull::Cull()
      : enable( true ),
        front_face( GL_CCW ),
        cull_face( GL_BACK )
    {
    }


void PipelineState::Cull::apply_defaults()
    {
    glFrontFace( GL_CCW );
    glCullFace( GL_BACK );
    glEnable( GL_CULL_FACE );
    }


void PipelineState::Cull::apply
    (
    const Cull & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_CULL_FACE );
            }
        else
            {
            glDisable( GL_CULL_FACE );
            }
        }
    if( enable )
        {
        if( front_face != other.front_face )
            {
            front_face = other.front_face;
            glFrontFace( front_face );
            }
        if( cull_face != other.cull_face )
            {
            cull_face = other.cull_face;
            glCullFace( cull_face );
            }
        }
    }


// class PipelineState::DepthOffset

PipelineState::DepthOffset::DepthOffset()
      : enable_point( false ),
        enable_line( false ),
        enable_polygon( false ),
        factor( 0.0f ),
        bias( 0.0f )
    {
    }


void PipelineState::DepthOffset::apply_defaults()
    {
    glPolygonOffset( 0.0f, 0.0f );
    glDisable( GL_POLYGON_OFFSET_POINT );
    glDisable( GL_POLYGON_OFFSET_LINE );
    glDisable( GL_POLYGON_OFFSET_FILL );
    }


void PipelineState::DepthOffset::apply
    (
    const DepthOffset & other
    )
    {
    if( enable_point != other.enable_point )
        {
        enable_point = other.enable_point;
        if( enable_point )
            {
            glEnable( GL_POLYGON_OFFSET_POINT );
            }
        else
            {
            glDisable( GL_POLYGON_OFFSET_POINT );
            }
        }

    if( enable_line != other.enable_line )
        {
        enable_line = other.enable_line;
        if( enable_line )
            {
            glEnable( GL_POLYGON_OFFSET_LINE );
            }
        else
            {
            glDisable( GL_POLYGON_OFFSET_LINE );
            }
        }

    if( enable_polygon != other.enable_polygon )
        {
        enable_polygon = other.enable_polygon;
        if( enable_polygon )
            {
            glEnable( GL_POLYGON_OFFSET_FILL );
            }
        else
            {
            glDisable( GL_POLYGON_OFFSET_FILL );
            }
        }

    if( enable_point || enable_line || enable_polygon )
        {
        if( ( factor != other.factor ) || ( bias != other.bias ) )
            {
            factor = other.factor;
            bias = other.bias;
            glPolygonOffset( factor, bias );
            }
        }
    }


// class PipelineState::AlphaTest

PipelineState::AlphaTest::AlphaTest()
      : enable( false ),
        function( GL_ALWAYS ),
        value( 0.0f )
    {
    }


void PipelineState::AlphaTest::apply_defaults()
    {
    glDisable( GL_ALPHA_TEST );
    glAlphaFunc( GL_ALWAYS, 0.0f );
    }


void PipelineState::AlphaTest::apply
    (
    const AlphaTest & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_ALPHA_TEST );
            }
        else
            {
            glDisable( GL_ALPHA_TEST );
            }
        }
    if( enable )
        {
        if( ( function != other.function ) || ( value != other.value ) )
            {
            function = other.function;
            value = other.value;
            glAlphaFunc( function, value );
            }
        }
    }


// class PipelineState::DepthTest

PipelineState::DepthTest::DepthTest()
      : enable( false ),
        function( GL_LESS )
    {
    }


void PipelineState::DepthTest::apply_defaults()
    {
    glDisable( GL_DEPTH_TEST );
    glDepthFunc( GL_LESS );
    }


void PipelineState::DepthTest::apply
    (
    const DepthTest & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_DEPTH_TEST );
            }
        else
            {
            glDisable( GL_DEPTH_TEST );
            }
        }
    if( enable )
        {
        if( function != other.function )
            {
            function = other.function;
            glDepthFunc( function );
            }
        }
    }


// class PipelineState::BufferMask

PipelineState::BufferMask::BufferMask()
      : color( true ),
        depth( false )
    {
    }


void PipelineState::BufferMask::apply_defaults()
    {
    glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
    glDepthMask( GL_FALSE );
    }


void PipelineState::BufferMask::apply
    (
    const BufferMask & other
    )
    {
    if( color != other.color )
        {
        color = other.color;
        GLboolean value = color ? GL_TRUE : GL_FALSE;
        glColorMask( value, value, value, value );
        }
    if( depth != other.depth )
        {
        depth = other.depth;
        GLboolean value = depth ? GL_TRUE : GL_FALSE;
        glDepthMask( value );
        }
    }


// class PipelineState::Blend

PipelineState::Blend::Blend()
      : enable( false ),
        source( GL_ONE ),
        destination( GL_ZERO )
    {
    }

    
void PipelineState::Blend::apply_defaults()
    {
    glDisable( GL_BLEND );
    glBlendFunc( GL_ONE, GL_ZERO );
    }


void PipelineState::Blend::apply
    (
    const Blend & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_BLEND );
            }
        else
            {
            glDisable( GL_BLEND );
            }
        }
    if( enable )
        {
        if( ( source != other.source ) || ( destination != other.destination ) )
            {
            source = other.source;
            destination = other.destination;
            glBlendFunc( source, destination );
            }
        }
    }


// class PipelineState::Texture

PipelineState::Texture::Texture()
      : flags( 0 ),
        mode( GL_MODULATE )
    {
    }

    
void PipelineState::Texture::apply_defaults
    (
    int unit
    )
    {
    TextureManager & texture_manager = Global::get_texture_manager();
    texture_manager.activate_texture_unit( unit );
    glDisable( GL_TEXTURE_2D );
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
    }

    
void PipelineState::Texture::apply
    (
    const Texture & other,
    int             unit
    )
    {
    TextureManager & texture_manager = Global::get_texture_manager();
    texture_manager.activate_texture_unit( unit );
    
    if( ( data != other.data ) || ( flags != other.flags ) )
        {
        data = other.data;
        flags = other.flags;
        if( data.is_null() )
            {
            texture_manager.unbind_texture();
            glDisable( GL_TEXTURE_2D );
            }
        else
            {
            texture_manager.bind_texture( data->get_texture( flags ) );
            glEnable( GL_TEXTURE_2D );
            }
        }

    if( mode != other.mode )
        {
        mode = other.mode;
        glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode );
        }
    }


// class PipelineState::Lighting::Model

PipelineState::Lighting::Model::Model()
      : ambient( 0.2f, 0.2f, 0.2f )
    {
    }


void PipelineState::Lighting::Model::apply_defaults()
    {
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, OpenGL::FloatColor( 0.2f, 0.2f, 0.2f ).rgba );
    // Leave usupported parameters at OpenGL defaults...
    }


void PipelineState::Lighting::Model::apply
    (
    const Model & other
    )
    {
    if( ambient != other.ambient )
        {
        ambient = other.ambient;
        glLightModelfv( GL_LIGHT_MODEL_AMBIENT, ambient.rgba );
        }
    }


// class PipelineState::Lighting::Material

PipelineState::Lighting::Material::Material()
      : ambient( 1.0f, 1.0f, 1.0f ),
        diffuse( 1.0f, 1.0f, 1.0f ),
        specular( 0.0f, 0.0f, 0.0f ),
        emission( 0.0f, 0.0f, 0.0f ),
        shininess( 0.0f )
    {
    }


void PipelineState::Lighting::Material::apply_defaults()
    {
    glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT,
        OpenGL::FloatColor( 1.0f, 1.0f, 1.0f ).rgba );    
    glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE,
        OpenGL::FloatColor( 1.0f, 1.0f, 1.0f ).rgba );
    glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR,
        OpenGL::FloatColor( 0.0f, 0.0f, 0.0f ).rgba );
    glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION,
        OpenGL::FloatColor( 0.0f, 0.0f, 0.0f ).rgba );
    glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 0.0f );
    }


void PipelineState::Lighting::Material::apply
    (
    const Material & other
    )
    {
    if( ambient != other.ambient )
        {
        ambient = other.ambient;
        glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, ambient.rgba );
        }
    if( diffuse != other.diffuse )
        {
        diffuse = other.diffuse;
        glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse.rgba );
        }
    if( specular != other.specular )
        {
        specular = other.specular;
        glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, specular.rgba );
        }
    if( emission != other.emission )
        {
        emission = other.emission;
        glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, emission.rgba );
        }
    if( shininess != other.shininess )
        {
        shininess = other.shininess;
        glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, shininess );
        }
    }


// class PipelineState::Lighting::ColorMaterial

PipelineState::Lighting::ColorMaterial::ColorMaterial()
      : enable( false ),
        mode( GL_AMBIENT_AND_DIFFUSE )
    {
    }


void PipelineState::Lighting::ColorMaterial::apply_defaults()
    {
    glDisable( GL_COLOR_MATERIAL );
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    }


void PipelineState::Lighting::ColorMaterial::apply
    (
    const ColorMaterial & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_COLOR_MATERIAL );
            }
        else
            {
            glDisable( GL_COLOR_MATERIAL );
            }
        }

    if( enable )
        {
        if( mode != other.mode )
            {
            mode = other.mode;
            glColorMaterial( GL_FRONT_AND_BACK, mode );
            }
        }
    }


// class PipelineState::Lighting::Light

PipelineState::Lighting::Light::Light()
      : enable( false ),
        ambient( 0.0f, 0.0f, 0.0f ),
        diffuse( 1.0f, 1.0f, 1.0f ),
        specular( 1.0f, 1.0f, 1.0f ),
        position( 0.0f, 0.0f, 1.0f ),
        infinite( true ),
        attenuation( 1.0f, 0.0f, 0.0f )
    {
    }


void PipelineState::Lighting::Light::apply_defaults
    (
    int light
    )
    {
    hope( light >= 0 );
    hope( light < MAX_LIGHT_SOURCES );
    const GLenum gl_light = GL_LIGHT0 + light;

    glDisable( gl_light );
    glLightfv( gl_light, GL_AMBIENT, OpenGL::FloatColor( 0.0f, 0.0f, 0.0f ).rgba );
    glLightfv( gl_light, GL_DIFFUSE, OpenGL::FloatColor( 1.0f, 1.0f, 1.0f ).rgba );
    glLightfv( gl_light, GL_SPECULAR, OpenGL::FloatColor( 1.0f, 1.0f, 1.0f ).rgba );
    const GLfloat gl_position[ 4 ] = { 0.0f, 0.0f, 1.0f, 0.0f };
    glLightfv( gl_light, GL_POSITION, gl_position );
    glLightf( gl_light, GL_CONSTANT_ATTENUATION, 1.0f );
    glLightf( gl_light, GL_LINEAR_ATTENUATION, 0.0f );
    glLightf( gl_light, GL_QUADRATIC_ATTENUATION, 0.0f );
    }


void PipelineState::Lighting::Light::apply
    (
    const Light &   other,
    int             light
    )
    {
    hope( light >= 0 );
    hope( light < MAX_LIGHT_SOURCES );
    const GLenum gl_light = GL_LIGHT0 + light;

    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( gl_light );
            }
        else
            {
            glDisable( gl_light );
            }
        }

    if( enable )
        {
        if( ambient != other.ambient )
            {
            ambient = other.ambient;
            glLightfv( gl_light, GL_AMBIENT, ambient.rgba );
            }
        if( diffuse != other.diffuse )
            {
            diffuse = other.diffuse;
            glLightfv( gl_light, GL_DIFFUSE, diffuse.rgba );
            }
        if( specular != other.specular )
            {
            specular = other.specular;
            glLightfv( gl_light, GL_SPECULAR, specular.rgba );
            }

        // Note: OpenGL transforms the light source to eye space when glLightfv is
        // called. Later changes to modelview matrix won't affect light position.
        // So we could:
        // - Always update light position regardless of this->position.
        // - Keep track of modification to modelview matrix and update position
        //   if it's dirty or modelview is dirty.
        // The second option seems better, but:
        // - We don't track modelview matrix changes.
        // - There would be problems with position which are not updated because
        //   light is disabled - we would have to change those to "unknown".

        //if( ( position != other.position ) || ( infinite != other.infinite ) )
            {
            position = other.position;
            infinite = other.infinite;
            const GLfloat gl_position[ 4 ] =
                {
                position.x(),
                position.y(),
                position.z(),
                infinite ? 0.0f : 1.0f
                };
            glLightfv( gl_light, GL_POSITION, gl_position );
            }

        if( attenuation != other.attenuation )
            {
            attenuation = other.attenuation;
            glLightf( gl_light, GL_CONSTANT_ATTENUATION, attenuation.x() );
            glLightf( gl_light, GL_LINEAR_ATTENUATION, attenuation.y() );
            glLightf( gl_light, GL_QUADRATIC_ATTENUATION, attenuation.z() );
            }
        }
    }


// class PipelineState::Lighting

PipelineState::Lighting::Lighting()
      : enable( false )
    {
    }


void PipelineState::Lighting::apply_defaults()
    {
    glDisable( GL_LIGHTING );
    Model::apply_defaults();
    Material::apply_defaults();
    ColorMaterial::apply_defaults();
    int i;
    for( i = 0; i < MAX_LIGHT_SOURCES; i++ )
        {
        Light::apply_defaults( i );
        }
    }


void PipelineState::Lighting::apply
    (
    const Lighting & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_LIGHTING );
            }
        else
            {
            glDisable( GL_LIGHTING );
            }
        }

    if( enable )
        {
        model.apply( other.model );
        material.apply( other.material );
        color_material.apply( other.color_material );
        int i;
        for( i = 0; i < MAX_LIGHT_SOURCES; i++ )
            {
            light[ i ].apply( other.light[ i ], i );
            }
        }
    }


// class PipelineState::Fog

PipelineState::Fog::Fog()
      : enable( false ),
        mode( GL_EXP ),
        density( 1.0f ),
        start( 0.0f ),
        end( 1.0f ),
        color( 0.0f, 0.0f, 0.0f, 0.0f )

    {
    }


void PipelineState::Fog::apply_defaults()
    {
    glDisable( GL_FOG );
    glFogi( GL_FOG_MODE, GL_EXP );
    glFogf( GL_FOG_DENSITY, 1.0f );
    glFogf( GL_FOG_START, 0.0f );
    glFogf( GL_FOG_END, 1.0f );
    const GLfloat color[ 4 ] = { 0.0f, 0.0f, 0.0f, 0.0f };
    glFogfv( GL_FOG_COLOR, color );
    }


void PipelineState::Fog::apply
    (
    const Fog & other
    )
    {
    if( enable != other.enable )
        {
        enable = other.enable;
        if( enable )
            {
            glEnable( GL_FOG );
            }
        else
            {
            glDisable( GL_FOG );
            }
        }
    if( enable )
        {
        if( mode != other.mode )
            {
            mode = other.mode;
            glFogi( GL_FOG_MODE, mode );
            }
        if( density != other.density )
            {
            density = other.density;
            glFogf( GL_FOG_DENSITY, density );
            }
        if( start != other.start )
            {
            start = other.start;
            glFogf( GL_FOG_START, start );
            }
        if( end != other.end )
            {
            end = other.end;
            glFogf( GL_FOG_END, end );
            }
        if( color != other.color )
            {
            color = other.color;
            glFogfv( GL_FOG_COLOR, color.rgba );
            }
        }
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "pipeline_state_inline.h"
#endif
