/** @file pipeline_state.h
 * Rendering pipeline state.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_PIPELINE_STATE_H
#define CLIENT_PIPELINE_STATE_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_TYPES_H
#include "renderer/opengl_types.h"
#endif

#ifndef CLIENT_TEXTURE_DATA_OBJECT_H
#include "renderer/texture_data_object.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Object representing state of OpenGL pipeline subset.
 *
 * This object is used to keep track of current state of rendering pipeline and
 * to change this state. There should be single PipelineState object per
 * rendering context representing current state (it's currently accesible via
 * class Renderer). To change the state, create new PipelineState object,
 * fill it to the state you need and call #apply() on the "current state"
 * PipelineState.
 *
 * PipelineState corresponds to subset of OpenGL pipeline state. Never change
 * parameters which are managed by PipelineState directly.
 *
 * Parameters not managed by PipelineState can be divided into three groups:
 *
 * - Parameters managed by other means. Currently these are:
 *   - Vertex arrays. Use VertexArray objects.
 *   - Modelview matrix stack. At rendering top-level it should be identity and
 *     push/pop should be used whenever it's changed. MatrixMode should be
 *     always modelview. See lighting notes below for more information about
 *     modelview stack handling.
 *   - Projection matrix stack. Set to well defined projection by each
 *     renderer subsystem. Push/pop should not be used. Remember to change
 *     MatrixMode back to modelview when changing projection.
 *
 * - "Constant" parameters. These are set during OpenGL initialization and
 *   should never be changed:
 *   - Viewport.
 *   - ClearColor is transparent black (probably should not be constant).
 *   - All normals will be normalized.
 *
 * - Other parameters. These are currently completely or mostly unused by the
 *   renderer. Nothing can be assumed about their state.
 *
 * Special care must be taken when OpenGL vertex lighting is used. Light
 * positions and orientations (not supported yet) are specified in object-space,
 * not world-space (there is no such thing in OpenGL) or camera-space (don't ask
 * me why). However, current modelview matrix is used to transform light to
 * camera-space when light properties are modified and further modelview changes
 * won't affect the position. This means you must ensure that:
 * - Modelview matrix is set to object-space to camera-space transformation
 *   before pipeline state with lighting is applied.
 * - Nothing else yet... - light positions are always updated regardless of
 *   "current" positions remembered by @c this in #apply().
 *
 * You could use world-space coordinates in pipeline state and apply the state
 * when modelview matrix contains world-space to camera-space transformation, but
 * don't do that.
 *
 * Also note that light source position in #Light objects should be stored in
 * world-space coordinates and transformed into object-space when pipeline state
 * is changed. #Light::set_pipeline_light() does this automatically when
 * @a entity argument is supplied. If you are using #RenderManager to manager
 * #RenderEntity and #Light objects, call #RenderManager::set_pipeline_lighting()
 * to update lighting parameters in pipeline state. All #RenderEntity classes
 * should do this automatically from their #RenderEntity::render() methods.
 */
class PipelineState
    {
public:

    /**
     * Default constructor.
     */
    PipelineState();

    /**
     * Reset "constant" OpenGL state (not managed by PipelineState) to defaults.
     */
    static void apply_constant_defaults();

    /**
     * Reset OpenGL pipeline state to defaults.
     */
    static void apply_defaults();

public:

    /**
     * Apply state from @a other.
     *
     * OpenGL pipeline will be set up as specified by @a other state and
     * contents of this PipelineState will be updated to match the new state.
     * Note that applied state may be slightly different from the requested one:
     * - Unnecessary state changes will not be performed. If feature should
     *   be disabled, its parameters specified in @a other will be ignored.
     *   The only exception is texturing - if requested texture binding is
     *   @c NULL texture, texturing will be disabled and current texture will
     *   be unbound.
     * - Unsupported parameters will be silently ignored. Currently this
     *   applies only to texturing parameters of units not supported by
     *   the hardware.
     */
    void apply
        (
        const PipelineState & other
        );

public:

    /**
     * Culling.
     */
    class Cull
        {
    public:

        /** Default constructor. */
        Cull();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const Cull & other
            );

    public:

        bool        enable;
            /**< @c GL_CULL_FACE. Default it @c true. */

        GLenum      front_face;
            /**< @c glFrontFace. Default is @c GL_CCW. */

        GLenum      cull_face;
            /**< @c glCullFace. Default is @c GL_BACK. */

        }; // class Cull

public:

    /**
     * Polygon depth offset.
     */
    class DepthOffset
        {
    public:

        /** Default constructor. */
        DepthOffset();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const DepthOffset & other
            );

    public:

        bool        enable_point;
            /**< @c GL_POLYGON_OFFSET_POINT. Default it @c flase. */

        bool        enable_line;
            /**< @c GL_POLYGON_OFFSET_LINE. Default it @c flase. */

        bool        enable_polygon;
            /**< @c GL_POLYGON_OFFSET_POLYGON. Default it @c flase. */

        GLfloat     factor;
            /**< First argument of @c glPolygonOffset. Default is 0.0. */

        GLfloat     bias;
            /**< Second argument of @c glPolygonOffset. Default is 0.0. */

        }; // class DepthOffset

public:

    /**
     * Alpha testing.
     */
    class AlphaTest
        {
    public:

        /** Default constructor. */
        AlphaTest();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const AlphaTest & other
            );

    public:

        bool        enable;
            /**< @c GL_ALPHA_TEST. Default is @c false. */

        GLenum      function;
            /**< @c glAlphaFunc::func. Default is @c GL_ALWAYS. */

        GLclampf    value;
            /**< @c glAlphaFunc::ref. Default is @c 0.0. */

        }; // class AlphaTest

public:

    /**
     * Depth testing.
     */
    class DepthTest
        {
    public:

        /** Default constructor. */
        DepthTest();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const DepthTest & other
            );

    public:

        bool        enable;
            /**< @c GL_DEPTH_TEST. Default is @c false. */

        GLenum      function;
            /**< @c glDepthFunc. Default is @c GL_LESS. */

        }; // class DepthTest

public:

    /**
     * Buffer update masking.
     */
    class BufferMask
        {
    public:

        /** Default constructor. */
        BufferMask();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const BufferMask & other
            );

    public:

        bool        color;
            /**< All parameters of @c glColorMask. Default is @c true. */

        bool        depth;
            /**< @c glDepthMask. Default is @c false. */

        }; // class BufferMask

public:

    /**
     * Blending.
     */
    class Blend
        {
    public:

        /** Default constructor. */
        Blend();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const Blend & other
            );

    public:

        bool        enable;
            /**< @c GL_BLEND. Default is @c false. */

        GLenum      source;
            /**< @c glBlendFunc::src. Default is @c GL_ONE. */

        GLenum      destination;
            /**< @c glBlendFunc::dst. Default is @c GL_ZERO. */

        }; // class Blend

public:

    /**
     * Texturing.
     */
    class Texture
        {
    public:

        /** Default constructor. */
        Texture();

        /** Apply defaults. */
        static void apply_defaults
            (
            int unit
            );

        /** Apply state from @a other. */
        void apply
            (
            const Texture & other,
            int             unit
            );

    public:

        TextureDataObject::Reference data;
            /**< Texture to bind. If NULL, no texture will be bound to this unit
                 and texturing will be disabled. Otherwise texture retreived
                 from the data object will be bound to this unit and 2D texturing
                 will be enabled. */

        int         flags;
            /**< See #TextureDataObject::get_texture(). Default is 0. */

        GLenum      mode;
            /**< <tt>glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, ... )</tt>.
                 Default is GL_MODULATE. */

        // GL_COMBINE mode not supported yet. It's part of 1.3 but we must
        // use extensions.

        //currently default texture parameters can't be changed here. FIXME?

        }; // class Texture

public:

    /**
     * Lighting.
     *
     * Using this is a bit tricky. Make sure you read and understand documentation
     * of #PipelineState class, especially the part about lighting.
     */
    class Lighting
        {
    public:

        /**
         * Lighting model parameters.
         *
         * These are all supported parameters set by <tt>glLightModel</tt>.
         * Hardcoded defaults (can't be changed by pipeline state) are:
         * - <tt>LIGHT_MODEL_LOCAL_VIEWER == FALSE </tt>.
         * - <tt>LIGHT_MODEL_TWO_SIDE == FALSE </tt>.
         * - <tt>LIGHT_MODEL_COLOR_CONTROL == SINGLE_COLOR </tt>.
         */
        class Model
            {
        public:

            /** Default constructor. */
            Model();

            /** Apply defaults. */
            static void apply_defaults();

            /** Apply state from @a other. */
            void apply
                (
                const Model & other
                );

        public:

            OpenGL::FloatColor  ambient;
                /**< Scene ambient (@c GL_LIGHT_MODEL_AMBIENT).
                     Default is (0.2, 0.2, 0.2, 1.0). */

            }; // class Model

    public:

        /**
         * Material parameters.
         *
         * All supported parameters set by <tt>glMaterial( GL_FRONT_AND_BACK, ...).
         * Two-sided lighting is not supported.
         */
        class Material
            {
        public:

            /** Default constructor. */
            Material();

            /** Apply defaults. */
            static void apply_defaults();

            /** Apply state from @a other. */
            void apply
                (
                const Material & other
                );

        public:

            OpenGL::FloatColor  ambient;
                /**< Ambient material color (@c GL_AMBIENT).
                     Default is (1.0, 1.0, 1.0, 1.0). */

            OpenGL::FloatColor  diffuse;
                /**< Diffuse material color (@c GL_DIFFUSE).
                     Default is (1.0, 1.0, 1.0, 1.0). */

            OpenGL::FloatColor  specular;
                /**< Specular material color (@c GL_SPECULAR).
                     Default is (0.0, 0.0, 0.0, 1.0). */

            OpenGL::FloatColor  emission;
                /**< Emission material color (@c GL_EMISSION).
                     Default is (0.0, 0.0, 0.0, 1.0). */

            GLfloat             shininess;
                /**< Material shininess (@c GL_SHININESS).
                     Default is 0.0. */

            }; // class Material

        /**
         * Current color -\> material color binding.
         */
        class ColorMaterial
            {
        public:

            /** Default constructor. */
            ColorMaterial();

            /** Apply defaults. */
            static void apply_defaults();

            /** Apply state from @a other. */
            void apply
                (
                const ColorMaterial & other
                );

        public:

            bool    enable;
                /**< Enable binding (@c GL_COLOR_MATERIAL).
                     Default is @c false. */

            GLenum  mode;
                /**< <tt>glColorMaterial( GL_FRONT_AND_BACK, ... ) parameter.
                     Default is @c GL_AMBIENT_AND_DIFFUSE. */

            }; // class ColorMaterial

    public:

        /**
         * Light source parameters.
         *
         * All supported parameters set by @c glLight.
         * Spotlights are not supported.
         *
         * Note that lights are specified in object-space.
         */
        class Light
            {
        public:

            /** Default constructor. */
            Light();

            /** Apply defaults. */
            static void apply_defaults
                (
                int light
                );

            /** Apply state from @a other. */
            void apply
                (
                const Light &   other,
                int             light
                );

        public:

            bool                enable;
                /**< Enable light source.
                     Default is @c false. */

            OpenGL::FloatColor  ambient;
                /**< Ambient intensity (@c GL_AMBIENT).
                     Default is (0.0, 0.0, 0.0, 1.0). */

            OpenGL::FloatColor  diffuse;
                /**< Diffuse intensity (@c GL_DIFFUSE).
                     Default is (1.0, 1.0, 1.0, 1.0). */

            OpenGL::FloatColor  specular;
                /**< Specular intensity (@c GL_SPECULAR).
                     Default is (1.0, 1.0, 1.0, 1.0). */

            OpenGL::Vector3     position;
                /**< Light position (@c GL_POSITION) in object-space(!).
                     Default is (0, 0, 1) */

            bool                infinite;
                /**< Infinite directional light source?
                     @a W coordinate of @c GL_POSITION will be 0.0 if #infinite
                     is @c true and 1.0 otherwise. Default is @c true. */

            OpenGL::Vector3     attenuation;
                /**< Constant (@a x), linear (@a y) and quadratic (@a z)
                     attenuation factors (@c GL_*_ATTENUATION).
                     Default is (1.0, 0.0, 0.0). */

            }; // class Light

    public:

        /** Default constructor. */
        Lighting();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const Lighting & other
            );

    public:

        bool                enable;
            /**< Enable lighting.
                 Default is @c false. */

        Model               model;
            /**< Lighting model parameters. */

        Material            material;
            /**< Material parameters. */

        ColorMaterial       color_material;
            /**< Current color -> material color binding. */

        Light               light[ MAX_LIGHT_SOURCES ];
            /**< Light sources parameters. */

        }; // class Lighting

public:

    /**
     * Fog.
     */
    class Fog
        {
    public:

        /** Default constructor. */
        Fog();

        /** Apply defaults. */
        static void apply_defaults();

        /** Apply state from @a other. */
        void apply
            (
            const Fog& other
            );

    public:

        bool                enable;
            /**< @c GL_FOG. Default is @c false. */

        GLenum              mode;
            /**< @C GL_FOG_MODE. GL_EXP (default), GL_EXP2 or GL_LINEAR. */

        GLfloat             density;
            /**< @c GL_FOG_DENSITY. Default is 1.0. */

        GLfloat             start;
            /**< @c GL_FOG_START. Default is 0.0. */

        GLfloat             end;
            /**< @c GL_FOG_END. Default is 1.0. */

        OpenGL::FloatColor  color;
            /**< @c GL_FOG_COLOR. Default is black. */

        }; // class Fog

public:

    Cull            cull;
        /**< Culling. */

    DepthOffset     depth_offset;
        /**< Point/line/polygon depth offset. */

    AlphaTest       alpha_test;
        /**< Alpha testing. */

    DepthTest       depth_test;
        /**< Depth testing. */

    BufferMask      buffer_mask;
        /**< Buffer update masking. */

    Blend           blend;
        /**< Blending. */

    Texture         texture[ MAX_TEXTURE_UNITS ];
        /**< Texturing parameters for each unit. */

    Lighting        lighting;
        /**< Lighting parameters. */

    Fog             fog;
        /**< Fog parameters. */

    GLenum          polygon_mode;
        /**< <tt>glPolygonMode( GL_FRONT_AND_BACK, ... )</tt>. Default is @c GL_FILL. */

    GLenum          shade_model;
        /**< @c glShadeModel. Default is @c GL_SMOOTH. */

    OpenGL::Color   color;
        /**< @c glColor. Default is opaque white.

             @bug This may be different from current OpenGL color if vertex
                  arrays are used. */

    }; // class PipelineState


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "pipeline_state_inline.h"
#endif


#endif // CLIENT_PIPELINE_STATE_H
