/** @file render_entity.h
 * Generic 3D renderable entity.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_RENDER_ENTITY_H
#define CLIENT_RENDER_ENTITY_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Generic 3D renderable entity.
 *
 * Render entities have position and orientation in our standard 3D
 * world-space and can be scaled.
 *
 * They can be used as standalone objects, but it's recommended to register
 * them to #RenderManager and let the manager render them.
 *
 * @todo Bounding box/sphere, visibility (frustum culling at least).
 */
class RenderEntity
    {
    NO_COPY_CONSTRUCTOR( RenderEntity );
    NO_ASSIGNMENT_OPERATOR( RenderEntity );

public:

    /**
     * Constructor.
     */
    RenderEntity();

    /**
     * Destructor.
     */
    virtual ~RenderEntity();

public:

    //@{
    /** Getter. */
    INLINE bool is_visible() const;

    INLINE const SVector3 & get_position() const;

    INLINE const SOrientation & get_orientation() const;

    INLINE const SVector3 & get_scale() const;
    //@}

    //@{
    /** Setter. */
    INLINE void set_visible
        (
        bool new_visible
        );

    INLINE void set_position
        (
        const SVector3 & new_position
        );

    INLINE void set_orientation
        (
        const SOrientation & new_orientation
        );

    INLINE void set_scale
        (
        const SVector3 & new_scale
        );
    //@}

public:

    /**
     * Register entity to render manager.
     *
     * It's allowed to register already registered entity. It will be
     * unregistered from current manager first.
     */
    void register_to_manager
        (
        RenderManager * new_manager
        );

    /**
     * Register entity to global render manager owned by pass manager.
     *
     * RenderManager must already exist for the pass. It will NOT be created
     * automatically. This is different from
     * #OverlayObject::register_to_global_manager(), because RenderManager
     * neeeds special setup (camera) before it can be used.
     */
    void register_to_global_manager
        (
        int pass_index
        );

    /**
     * Unregister render entity from the manager.
     *
     * Shortcut for <tt>register_to_manager( NULL )</tt>.
     */
    INLINE void unregister_from_manager();

public:

    /**
     * Render the entity.
     *
     * When called by #RenderManager, OpenGL matrices will be set to camera
     * transformation and projection. #RenderManager will not call this method
     * if #visible is @c false. The render entity will render itself regardless
     * of #visible when #render() is called directly.
     *
     * When render entity uses OpenGL lighting to apply light sources managed by
     * owning RenderManager, it must update its pipeline states before rendering
     * itself. Use #RenderManager::set_pipeline_state() to do this
     * (or other helper methods that update the state, such as
     * #RenderObject::set_pipeline_state()).
     */
    virtual void render
        (
        const SystemTime & time_delta
        ) = 0;

protected:

    /**
     * Apply entity transformation to OpenGL modelview matrix.
     */
    void apply_transformation() const;

    /**
     * Revert OpenGL modelview matrix to state before #apply_transformation().
     */
    void revert_transformation() const;

private:

    bool            visible;
        /**< Should entity be rendered? Initially @c true. */

    SVector3        position;
        /**< Entity position. */

    SOrientation    orientation;
        /**< Entity orientation. */

    SVector3        scale;
        /**< Entity scale. */

protected:

    RenderManager * manager;
        /**< Manager this object is registered to. It may be NULL. */

    }; // class RenderEntity


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "render_entity_inline.h"
#endif


#endif // CLIENT_RENDER_ENTITY_H
