/** @file render_manager.cpp
 * Manager and renderer of 3D render entities (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/render_manager.h"
#include "renderer/render_entity.h"
#include "renderer/light.h"
#include "renderer/camera.h"
#include "renderer/opengl.h"
#include "global/global.h"
#include <sstream>


namespace Demo {
namespace Client {
namespace Renderer {


RenderManager::RenderManager()
      : lighting( false ),
        camera( NULL )
    {
    }


RenderManager::~RenderManager()
    {
    if( !entities.empty() )
        {
        // Log warning.

        {//sob
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << entities.size() << " entities still registered in ~RenderManager.";
        Massiv::Core::Global::log( Status::warning( Status::FACILITY_GENERIC,
                    Status::PRIORITY_NORMAL), oss.str() );
        }//eob

        // Unregister entities.

        Entities::iterator it;
        while( ( it = entities.begin() ) != entities.end() )
            {
            RenderEntity * const entity = *it;
            hope( entity );
            entity->unregister_from_manager();
            }
        }
    }


void RenderManager::render
    (
    const SystemTime & time_delta
    )
    {
    // Set camera projection and transformation.
    // FIXME: allow passes with infinite objects - no camera translation.

    hope( camera );
    camera->apply_projection();
    camera->apply_transformation();

    // Render registered entities.

    Entities::iterator it;
    for( it = entities.begin(); it != entities.end(); ++it )
        {
        RenderEntity * const entity = *it;
        hope( entity );
        if( entity->is_visible() )
            {
            //FIXME: camera culling.
            entity->render( time_delta );
            }
        }

    // Done.

    camera->revert_transformation();
    OpenGL::check_errors( "Entity rendering: " );
    }


void RenderManager::set_pipeline_lighting
    (
    PipelineState &         state,
    const RenderEntity *    entity
    ) const
    {
    if( lighting )
        {
        state.lighting.enable = true;
        state.lighting.model.ambient = OpenGL::FloatColor( 0.0f, 0.0f, 0.0f );

        // FIXME: Use entity bounding box, select MAX_LIGHT_SOURCES most
        // intensive light sources.

        int i = 0;
        Lights::const_iterator it = lights.begin();
        while( ( i < MAX_LIGHT_SOURCES ) && ( it != lights.end() ) )
            {
            PipelineState::Lighting::Light & light_state = state.lighting.light[ i ];
            Light * const light_source = *it;
            hope( light_source );
            light_source->set_pipeline_light( light_state, entity );
            i++;
            ++it;
            }

        while( i < MAX_LIGHT_SOURCES )
            {
            state.lighting.light[ i ].enable = false;
            i++;
            }
        }
    else
        {
        state.lighting.enable = false;
        }
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "render_manager_inline.h"
#endif
