/** @file render_manager.h
 * Manager and renderer of 3D render entities.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_RENDER_MANAGER_H
#define CLIENT_RENDER_MANAGER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <list>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Manager of 3D render entities.
 *
 * Render manager keeps list of aliasing pointers to registered entities and
 * single camera object. Its #render() method sets transformation from camera
 * and renders all registered entities.
 */
class RenderManager
    {
    NO_COPY_CONSTRUCTOR( RenderManager );
    NO_ASSIGNMENT_OPERATOR( RenderManager );

public:

    /**
     * Constructor.
     */
    RenderManager();

    /**
     * Destructor.
     *
     * All entities should be unregistered before destruction of the manager.
     * If some entities are still registered, they will be unregistered
     * automatically but warning will be logged.
     */
    ~RenderManager();

public:

    /**
     * Is vertex lighting used?
     *
     * Initial default value is @c false.
     */
    INLINE bool is_lighting() const;

    /**
     * Should vertex lighting be used?
     */
    INLINE void set_lighting
        (
        bool new_lighting
        );

    /**
     * Change current camera.
     *
     * Use @c NULL to unset camera - identity transformation will be used.
     */
    INLINE void set_camera
        (
        Camera * new_camera
        );

public:

    /**
     * Render all registered entities.
     *
     * Note: Lighting will not be automatically set by the manager. Entities
     * should call #set_pipeline_lighting() from their #render() methods and
     * use the modified state to render themselves.
     *
     * @pre Camera is set.
     */
    void render
        (
        const SystemTime & time_delta
        );

    /**
     * Set pipeline state lighting.
     *
     * Only lighting @c enable flag, lighting model parameters and light sources
     * will be changed. Material parameters will not be touched by the manager.
     *
     * Scene ambient color will be set to "black" (0, 0, 0). Use ambient intensity
     * of "sun" light source (if there is such source) to control ambient lighting.
     *
     * See #Light::set_pipeline_light() for more info.
     */
    void set_pipeline_lighting
        (
        PipelineState &         state,
        const RenderEntity *    entity
        ) const;
    
protected:

    /// RenderEntity can register/unregister itself to/from the manager.
    friend class RenderEntity;

    /**
     * Register render entity.
     */
    INLINE void register_entity
        (
        RenderEntity * entity
        );

    /**
     * Unregister render entity.
     */
    INLINE void unregister_entity
        (
        RenderEntity * entity
        );

protected:

    /// Light can register/unregister itself to/from the manager.
    friend class Light;

    /**
     * Register light source.
     */
    INLINE void register_light
        (
        Light * light
        );

    /**
     * Unregister light source.
     */
    INLINE void unregister_light
        (
        Light * light
        );

private:

    /**
     * RenderEntity * hasher.
     */
    struct RenderEntityPtrHash
        {
        inline std::size_t operator()
            (
            RenderEntity * entity
            ) const
            {
            return reinterpret_cast< std::size_t >( entity );
            }
        };
        
    /**
     * Set of aliasing RenderEntity pointers.
     */
    typedef ext::hash_set< RenderEntity *, RenderEntityPtrHash > Entities;

    /**
     * Set of aliasing pointers to Light.
     */
    typedef std::list< Light * > Lights;

private:

    Entities    entities;
        /**< Set of aliasing RenderEntity pointers. */

    Lights      lights;
        /**< List of aliasing pointers to lights. */

    bool        lighting;
        /**< Use vertex lighting? */

    Camera *    camera;
        /**< Aliasing pointer to current camera. */

    }; // class RenderManager


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "render_manager_inline.h"
#endif


#endif // CLIENT_RENDER_MANAGER_H
