/** @file render_object.cpp
 * Vertex-array based object (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/render_object.h"
#include "renderer/renderer.h"
#include "renderer/render_manager.h"


namespace Demo {
namespace Client {
namespace Renderer {


// class RenderObject::Geometry

RenderObject::Geometry::~Geometry()
    {
    hope( !locked );
    delete [] memory;
    }


void RenderObject::Geometry::initialize_index_array
    (
    GLenum      the_mode,
    std::size_t array_size
    )
    {
    hope( !memory );
    hope( size == 0 );
    hope( array_size > 0 );

    // Currently system memory is always used.

    mode = the_mode;
    size = array_size;
    memory = new GLushort [ size ];
    }

    
void RenderObject::Geometry::lock()
    {
    hope( memory );
    hope( size > 0 );
    hope( !locked );

    indices.bind( memory, size );
    locked = true;
    }


void RenderObject::Geometry::unlock()
    {
    hope( memory );
    hope( size > 0 );
    hope( locked );

    indices.bind( NULL, 0 );
    locked = false;
    }


// class RenderObject

void RenderObject::clear()
    {
    jobs.clear();
    geometries.clear();
    pipeline_states.clear();
    array.free();
    array.reset_stream_counts();
    }


void RenderObject::initialize_simple_jobs
    (
    std::size_t job_count
    )
    {
    hope( job_count > 0 );
    hope( pipeline_states.empty() );
    hope( geometries.empty() );
    hope( jobs.empty() );

    pipeline_states.resize( job_count );
    geometries.resize( job_count );
    jobs.resize( job_count );

    std::size_t i;
    for( i = 0; i < job_count; i++ )
        {
        jobs[ i ].pipeline_state = &pipeline_states[ i ];
        jobs[ i ].geometry = &geometries[ i ];
        }
    }


void RenderObject::render()
    {
    if( jobs.empty() )
        {
        return;
        }

    // Bind vertex array.

    array.bind();

    // Render the jobs.

    Jobs::const_iterator it;
    for( it = jobs.begin(); it != jobs.end(); ++it )
        {
        const Job & job = *it;

        // Apply state.

        if( job.pipeline_state )
            {
            Global::apply_state( *job.pipeline_state );
            }

        // Render.

        if( job.geometry )
            {
            Geometry & g = *job.geometry;
            hope( g.get_size() > 0 );
            hope( !g.is_locked() );
            if( g.is_index_array() )
                {
                // Index array.

                array.render( g.get_mode(), g.get_memory(), g.get_size() );
                }
            else
                {
                // Range.

                array.render( g.get_mode(), g.get_first(), g.get_size() );
                }
            }
        }

    // Done.

    array.unbind();
    }


void RenderObject::set_pipeline_lighting
    (
    RenderManager &         manager,
    const RenderEntity *    entity
    )
    {
    PipelineStates::iterator it;
    for( it = pipeline_states.begin(); it != pipeline_states.end(); ++it )
        {
        manager.set_pipeline_lighting( *it, entity );
        }
    }


void RenderObject::disable_pipeline_lighting()
    {
    PipelineStates::iterator it;
    for( it = pipeline_states.begin(); it != pipeline_states.end(); ++it )
        {
        it->lighting.enable = false;
        }
    }

    
} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "render_object_inline.h"
#endif
