/** @file render_object.h
 * Vertex-array based object.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_RENDER_OBJECT_H
#define CLIENT_RENDER_OBJECT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_PIPELINE_STATE_H
#include "renderer/pipeline_state.h"
#endif

#ifndef CLIENT_VERTEX_ARRAY_H
#include "renderer/vertex_array.h"
#endif

#include <vector>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Vertex-array based renderable object.
 *
 * RenderObject encapsulates single vertex array and multiple render jobs
 * that use this array. Render job (see #Job) consists of pipeline state
 * (see #PipelineState) to use and geometry (see #Geometry) to draw.
 *
 * To initialize #RenderObject, access the #array, #pipeline_states, #geometries
 * and #jobs directly. You don't have to use #pipeline_states and #geometries,
 * pointers to #jobs might point anywhere, but they must be always valid or @c NULL.
 *
 * Call #render() to render all jobs.
 */
class RenderObject
    {
    NO_COPY_CONSTRUCTOR( RenderObject );
    NO_ASSIGNMENT_OPERATOR( RenderObject );

public:

    /**
     * Geometric element type plus index array or range which describe
     * geometry to render.
     *
     * Because index array could be allocated in AGP or video memory, Geometry
     * copying is forbidden. Pointer to this memory is owning. This means that
     * copy constructor and assignment operator have semantics similar to
     * auto_ptr - ownership of the memory is transferred, not contents.
     */
    class Geometry
        {
    public:

        /**
         * Array of 16bit indices.
         */
        typedef Massiv::Core::NormalizedArray
                    < Massiv::Core::UnmanagedStorage< GLushort > > IndexArray;
    
    public:

        /**
         * Construct empty geometry.
         */
        INLINE Geometry();

        /**
         * Destructor.
         */
        ~Geometry();

        /**
         * Copy constructor.
         *
         * Ownership of index array is transferred, not contents.
         */
        INLINE Geometry
            (
            const Geometry & other
            );

        /**
         * Assignment operator.
         *
         * Ownership of index array is transferred, not contents.
         */
        INLINE Geometry & operator=
            (
            const Geometry & other
            );

    public:

        /**
         * Initialize geometry with array range.
         *
         * @a mode should be OpenGL mode (see @c glBegin() documentation).
         *
         * @pre Geometry is not initialized.
         */
        INLINE void initialize_range
            (
            GLenum      mode,
            std::size_t first_element,
            std::size_t element_count
            );

        /**
         * Initialize geometry with index array.
         *
         * @pre Geometry is not initialized.
         */
        void initialize_index_array
            (
            GLenum      mode,
            std::size_t array_size
            );

    public:

        /** Is this index range geometry? */
        INLINE bool is_range() const;

        /** Is this index array geometry? */
        INLINE bool is_index_array() const;
        
        //@{
        /** Getter. */
        INLINE GLenum get_mode() const;
        INLINE const GLushort * get_memory() const;
        INLINE GLushort * get_memory();
        INLINE std::size_t get_size() const;
        INLINE std::size_t get_first() const;
        //@}
        
    public:

        /**
         * Lock index array.
         *
         * This binds the array to #indices.
         *
         * @pre This is index array geometry.
         */
        void lock();

        /**
         * Unlock index array.
         *
         * @pre This is index array job.
         *
         * @todo Update array range.
         */
        void unlock();

        /** Is index array locked? */
        INLINE bool is_locked() const;

    public:

        IndexArray          indices;
            /**< Index array.
                 Can be accessed only if locked. */

    private:

        GLenum              mode;
            /**< Mode used to render. */
    
        mutable GLushort *  memory;
            /**< Owning pointer to memory with #size indices.
                 If @c NULL this batch uses array range instead of index array. */

        std::size_t         size;
            /**< Number of indices in @a memory or length of the array range. */

        std::size_t         first;
            /**< Index of first element in the array range.
                 Used only if #memory is NULL. */

        bool                locked;
            /**< Is index array locked? */

        }; // class Geometry

public:

    /**
     * Render job.
     *
     * Job contains aliasing pointers to pipeline state to use and geometry
     * to draw. Any of these pointers may be @c NULL (but @c NULL geometry is
     * not quite useless). You don't have to use #RenderObject::pipeline_states
     * and RenderObject::geometries to store the data, but #geometry must index
     * vertex array owned by RenderObject owning this job.
     */
    class Job
        {
    public:

        /**
         * Construct empty job.
         */
        INLINE Job();

        /**
         * Destructor.
         */
        INLINE ~Job();

    public:

        PipelineState *     pipeline_state;
            /**< Aliasing pointer to pipeline state to apply before the geometry
                 is rendered.
                 If it's @c NULL, pipeline state won't be updated. First job
                 should always have non-NULL pipeline state. You can store
                 the state anywhere, but it's recommended to use
                 #RenderObject::pipeline_states vector. */

        Geometry *          geometry;
            /**< Aliasing pointer to geometry to draw.
                 If it's @c NULL, nothing will be drawn. You can store the geometry
                 anywhere, but it's recommended to use #RenderObject::geometries
                 vector. It must index #RenderObject::array. */

        }; // class Job

public:

    /**
     * Vector of pipeline states.
     */
    typedef std::vector< PipelineState > PipelineStates;

    /**
     * Vector of geometries.
     */
    typedef std::vector< Geometry > Geometries;

    /**
     * Vector of jobs.
     */
    typedef std::vector< Job > Jobs;

public:

    /**
     * Constructor.
     */
    INLINE RenderObject();

    /**
     * Destructor.
     */
    INLINE ~RenderObject();

public:

    /**
     * Reset render object to initial (clean) state.
     */
    void clear();

    /**
     * Initialize simple render object jobs.
     *
     * This method will allocate @a job_count jobs using pipeline states and
     * geometry objects stored in vectors owned by the render object.
     *
     * @pre No pipeline states, geometries or jobs are allocated.
     */
    void initialize_simple_jobs
        (
        std::size_t job_count = 1
        );

public:

    /**
     * Render all jobs.
     *
     * @pre Neither vertex array nor any index array is locked.
     */
    void render();

    /**
     * Set pipeline state lighting of all states in #pipeline_states.
     *
     * Only states stored in #pipeline_states vector will be updated.
     * States of jobs which use state not stored in the vector will not
     * be updated.
     *
     * See #RenderManager::set_pipeline_light() for more info.
     */
    void set_pipeline_lighting
        (
        RenderManager &         manager,
        const RenderEntity *    entity
        );

    /**
     * Disable lighting in all #pipeline_states.
     *
     * Use this method to disable lighting enabled by #set_pipeline_lighting().
     */
    void disable_pipeline_lighting();

    /**
     * Dump render object to given stream.
     *
     * @pre No jobs are locked.
     */
    void dump
        (
        std::ostream &      stream,
        const std::string & indent
        );

public:

    VertexArray     array;
        /**< Vertex array. */

    PipelineStates  pipeline_states;
        /**< Pipeline states. */

    Geometries      geometries;
        /**< Pipeline states. */
        
    Jobs            jobs;
        /**< Render jobs. */
    
    }; // class RenderObject


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "render_object_inline.h"
#endif


#endif // CLIENT_RENDER_OBJECT_H
