/** @file render_objects.cpp
 * Functions to create various simple #RenderObject objects (implementation).
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/render_objects.h"
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {
namespace Objects {


void initialize_vertex
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    point,
    const SVector2 &    texcoord
    )
    {
    // Initialize vertex.

    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + 1 );
    Mesh::Vectors & vertices = *mesh.vertex_coordinates[ 0 ];
    vertices[ vertex_index ] = point;

    // Initialize texcoord.

    std::size_t i;
    for( i = 0; i < mesh.get_texcoord_stream_count(); i++ )
        {
        Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ i ];
        texcoords[ vertex_index ] = texcoord;
        }
    }


void initialize_line_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    point_0,
    const SVector3 &    point_1
    )
    {
    // Initialize vertices.

    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + 2 );
    Mesh::Vectors & vertices = *mesh.vertex_coordinates[ 0 ];
    vertices[ vertex_index ] = point_0;
    vertices[ vertex_index + 1 ] = point_1;

    // Initialize texcoords.

    std::size_t i;
    for( i = 0; i < mesh.get_texcoord_stream_count(); i++ )
        {
        Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ i ];
        texcoords[ vertex_index ] = SVector2( 0.0f, 0.0f );
        texcoords[ vertex_index + 1 ] = SVector2( 1.0f, 0.0f );
        }
    }


void initialize_quad_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    corner,
    const SVector3 &    right,
    const SVector3 &    up
    )
    {
    // Initialize vertices.

    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + 4 );
    Mesh::Vectors & vertices = *mesh.vertex_coordinates[ 0 ];
    vertices[ vertex_index ] = corner;
    vertices[ vertex_index + 1 ] = corner + right;
    vertices[ vertex_index + 2 ] = corner + right + up;
    vertices[ vertex_index + 3 ] = corner + up;

    // Initialize texcoords.

    std::size_t i;
    for( i = 0; i < mesh.get_texcoord_stream_count(); i++ )
        {
        Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ i ];
        texcoords[ vertex_index ] = SVector2( 0.0f, 1.0f );
        texcoords[ vertex_index + 1 ] = SVector2( 1.0f, 1.0f );
        texcoords[ vertex_index + 2 ] = SVector2( 1.0f, 0.0f );
        texcoords[ vertex_index + 3 ] = SVector2( 0.0f, 0.0f );
        }
    }


void initialize_quad_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    )
    {
    hope( mesh.get_triangle_count() >= triangle_index + 2 );
    hope( mesh.get_vertex_count() >= vertex_index + 4 );
    mesh.triangles[ triangle_index ].v0 = vertex_index;
    mesh.triangles[ triangle_index ].v1 = vertex_index + 1;
    mesh.triangles[ triangle_index ].v2 = vertex_index + 3;
    mesh.triangles[ triangle_index + 1 ].v0 = vertex_index + 1;
    mesh.triangles[ triangle_index + 1 ].v1 = vertex_index + 2;
    mesh.triangles[ triangle_index + 1 ].v2 = vertex_index + 3;
    }


void initialize_quads_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         quad_count
    )
    {
    std::size_t i;
    for( i = 0; i < quad_count; i++ )
        {
        initialize_quad_triangles( mesh, triangle_index + 2 * i, vertex_index + 4 * i );
        }
    }


void initialize_box_8_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    min,
    const SVector3 &    max
    )
    {
    // Initialize vertices.

    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + 8 );
    Mesh::Vectors & vertices = *mesh.vertex_coordinates[ 0 ];
    vertices[ vertex_index + 0 ] = SVector3( min.x(), min.y(), min.z() );
    vertices[ vertex_index + 1 ] = SVector3( max.x(), min.y(), min.z() );
    vertices[ vertex_index + 2 ] = SVector3( max.x(), max.y(), min.z() );
    vertices[ vertex_index + 3 ] = SVector3( min.x(), max.y(), min.z() );
    vertices[ vertex_index + 4 ] = SVector3( min.x(), min.y(), max.z() );
    vertices[ vertex_index + 5 ] = SVector3( max.x(), min.y(), max.z() );
    vertices[ vertex_index + 6 ] = SVector3( max.x(), max.y(), max.z() );
    vertices[ vertex_index + 7 ] = SVector3( min.x(), max.y(), max.z() );

    // Initialize texcoords (pretty lame).

    std::size_t i;
    for( i = 0; i < mesh.get_texcoord_stream_count(); i++ )
        {
        Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ i ];
        texcoords[ vertex_index + 0 ] = SVector2( 0.0f, 1.0f );
        texcoords[ vertex_index + 1 ] = SVector2( 1.0f, 1.0f );
        texcoords[ vertex_index + 2 ] = SVector2( 0.0f, 1.0f );
        texcoords[ vertex_index + 3 ] = SVector2( 1.0f, 1.0f );
        texcoords[ vertex_index + 4 ] = SVector2( 0.0f, 0.0f );
        texcoords[ vertex_index + 5 ] = SVector2( 1.0f, 0.0f );
        texcoords[ vertex_index + 6 ] = SVector2( 0.0f, 0.0f );
        texcoords[ vertex_index + 7 ] = SVector2( 1.0f, 0.0f );
        }
    }


void initialize_box_24_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    min,
    const SVector3 &    max
    )
    {
    // Initialize vertices and texcoords.

    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + 6 * 4 );
    const Float32 width = max.x() - min.x();
    const Float32 depth = max.y() - min.y();
    const Float32 height = max.z() - min.z();
    initialize_quad_vertices( mesh, vertex_index + 0 * 4,
                SVector3( min.x(), min.y(), min.z() ),
                SVector3( width, 0.0f, 0.0f ),
                SVector3( 0.0f, 0.0f, height ) );
    initialize_quad_vertices( mesh, vertex_index + 1 * 4,
                SVector3( max.x(), min.y(), min.z() ),
                SVector3( 0.0f, depth, 0.0f ),
                SVector3( 0.0f, 0.0f, height ) );
    initialize_quad_vertices( mesh, vertex_index + 2 * 4,
                SVector3( max.x(), max.y(), min.z() ),
                SVector3( -width, 0.0f, 0.0f ),
                SVector3( 0.0f, 0.0f, height ) );
    initialize_quad_vertices( mesh, vertex_index + 3 * 4,
                SVector3( min.x(), max.y(), min.z() ),
                SVector3( 0.0f, -depth, 0.0f ),
                SVector3( 0.0f, 0.0f, height ) );
    initialize_quad_vertices( mesh, vertex_index + 4 * 4,
                SVector3( min.x(), max.y(), min.z() ),
                SVector3( width, 0.0f, 0.0f ),
                SVector3( 0.0f, -depth, 0.0f ) );
    initialize_quad_vertices( mesh, vertex_index + 5 * 4,
                SVector3( min.x(), min.y(), max.z() ),
                SVector3( width, 0.0f, 0.0f ),
                SVector3( 0.0f, depth, 0.0f ) );
    }

    
void initialize_box_8_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    )
    {
    hope( mesh.get_triangle_count() >= triangle_index + 24 );
    hope( mesh.get_vertex_count() >= vertex_index + 8 );

    #define Q( i, a, b, c, d ) \
        mesh.triangles[ triangle_index + 2 * i ].v0 = vertex_index + a; \
        mesh.triangles[ triangle_index + 2 * i ].v1 = vertex_index + b; \
        mesh.triangles[ triangle_index + 2 * i ].v2 = vertex_index + d; \
        mesh.triangles[ triangle_index + 2 * i + 1 ].v0 = vertex_index + b; \
        mesh.triangles[ triangle_index + 2 * i + 1 ].v1 = vertex_index + c; \
        mesh.triangles[ triangle_index + 2 * i + 1 ].v2 = vertex_index + d;
    Q( 0, 0, 1, 5, 4 );
    Q( 1, 1, 2, 6, 5 );
    Q( 2, 2, 3, 7, 6 );
    Q( 3, 3, 0, 4, 7 );
    Q( 4, 3, 2, 1, 0 );
    Q( 5, 4, 5, 6, 7 );
    #undef Q
    }


void initialize_box_24_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    )
    {
    initialize_quads_triangles( mesh, triangle_index, vertex_index, 6 );
    }


void initialize_circle_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    Float32             radius,
    Float32             z,
    std::size_t         vertex_count
    )
    {
    using namespace Massiv::Core;
    hope( mesh.get_vertex_stream_count() >= 1 );
    hope( mesh.get_vertex_count() >= vertex_index + vertex_count );

    Mesh::Vectors & vertices = *mesh.vertex_coordinates[ 0 ];
    const Float32 radians_per_vertex = 2.0f * MATHF_PI / static_cast< Float32 >( vertex_count );
    std::size_t i;
    for( i = 0; i < vertex_count; i++ )
        {
        // Vertex coordinates.

        const Float32 angle = static_cast< Float32 >( i ) * radians_per_vertex;
        const Float32 cos_a = FloatTraits< Float32 >::cos( angle );
        const Float32 sin_a = FloatTraits< Float32 >::sin( angle );
        vertices[ vertex_index + i ] = SVector3( radius * cos_a, radius * sin_a, z );

        // Texture coordinates.

        std::size_t j;
        for( j = 0; j < mesh.get_texcoord_stream_count(); j++ )
            {
            Mesh::TexCoords & texcoords = *mesh.texture_coordinates[ i ];
            texcoords[ vertex_index + i ] = SVector2
                        ( 0.5f * ( cos_a + 1.0f ), 0.5f * ( 1.0f - sin_a ) );
            }
        }
    }   


void initialize_polygon_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         vertex_count,
    bool                reverse
    )
    {
    hope( vertex_count >= 3 );
    hope( mesh.get_triangle_count() >= triangle_index + vertex_count - 2 );
    hope( mesh.get_vertex_count() >= vertex_index + vertex_count );

    std::size_t i;
    for( i = 0; i < vertex_count - 2; i++ )
        {
        mesh.triangles[ triangle_index + i ].v0 = vertex_index;
        mesh.triangles[ triangle_index + i ].v1 = vertex_index +
                    ( reverse ? vertex_count - i - 1 : i + 1 );
        mesh.triangles[ triangle_index + i ].v2 = vertex_index +
                    ( reverse ? vertex_count - i - 2 : i + 2 );
        }
    }


void initialize_pyramid_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         base_vertex_count
    )
    {
    hope( base_vertex_count >= 3 );
    hope( mesh.get_triangle_count() >= triangle_index + 2 * base_vertex_count - 2 );
    hope( mesh.get_vertex_count() >= vertex_index + base_vertex_count + 1 );

    // Pyramid base.

    initialize_polygon_triangles( mesh, triangle_index,
                vertex_index + 1, base_vertex_count, true );

    // Pyramid sides.

    const std::size_t t = triangle_index + base_vertex_count - 2;
    std::size_t i;
    for( i = 0; i < base_vertex_count; i++ )
        {
        mesh.triangles[ t + i ].v0 = vertex_index;
        mesh.triangles[ t + i ].v1 = vertex_index + i + 1;
        mesh.triangles[ t + i ].v2 = vertex_index +
                    ( i == base_vertex_count - 1 ? 1 : i + 2 );
        }
    }


} // namespace Demo::Client::Renderer::Objects
} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "render_objects_inline.h"
#endif
