/** @file render_objects.h
 * Functions to create various simple #RenderObject objects.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_RENDER_OBJECTS_H
#define CLIENT_RENDER_OBJECTS_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_MESH_H
#include "renderer/mesh.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {

/**
 * Namespace Objects contains functions which can be used to initialize
 * #Mesh objects to simple shapes.
 *
 * There are two groups of functions:
 * - Functions to initialize Mesh verices.
 * - Functions to initialize Mesh triangles.
 *
 * Use functions provided by Mesh to initialize vertex arrays and jobs
 * of RenderObject. Currently there is no way to create "optimized" jobs
 * (triangle strips or fans).
 *
 * FIXME: The following documentation is out of date...
 *
 * All functions have similar arguments:
 * - object - reference to object to modify.
 * - flags - combination of #CreateFlags and shape-specific flags.
 * - state - Pointer to #PipelineState. If it's @c NULL, the object will not contain
 *           any state and will be rendered using current state. Otherwise the state
 *           will be copied into the object and will be used by all render jobs.
 *
 * Functions initializing vertex arrays have the following arguments:
 * - object - reference to object to modify.
 * - flags - combination of #CreateFlags and shape-specific flags. Texture
 *      coordinates will be initialized if @c TEXCOORDS is set.
 * - element_index - index of first vertex element to modify.
 * - shape-specific arguments.
 *
 * Functions initializing render jobs have the following arguments:
 * - object - reference to object to modify.
 * - flags (optional) - combination of #CreateFlags and shape-specific
 *      flags. This argument will be expected only if it's required by the
 *      shape (for example box jobs are different for textured and untextured
 *      box).
 * - job_index - index of job to modify.
 * - element_index - index of first vertex element to use for the job.
 * - shape-specific arguments.
 *
 * Functions initializing complete object have the following arguments:
 * - object - reference to object to inititalize.
 * - flags - combination of #CreateFlags and shape-specific flags.
 * - state (optional) - pointer to pipeline state to use. If it's @c NULL,
 *      object will always use current pipeline state. Otherwise the state
 *      will be copied to #RenderObject::pipeline_states and will be used by
 *      the first job. Note that some shapes will create their own state
 *      and their @c create_*() functions don't have @a state argument.
 * - shape-specific arguments.
 *
 * When calling @c initialize_*_vertices() functions, the arrays must be already
 * allocated. When calling @c initialize_*_jobs() functions, the jobs must be
 * already allocated and vertex arrays must be initialized to expected values.
 * You can use @c initialize_object() to allocate arrays and jobs. @c create_*()
 * allocates everything automatically and expects unitialized object.
 */
namespace Objects {


/**
 * Initialize single vertex.
 *
 * Vertex count: 1.
 */
void initialize_vertex
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    point,
    const SVector2 &    texcoord = SVector2()
    );


/**
 * Initialize 2 vertices.
 *
 * Vertex count: 2.
 */
void initialize_line_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    point_0,
    const SVector3 &    point_1
    );


/**
 * Initialize CCW quad vertices.
 *
 * Vertex count: 4.
 */
void initialize_quad_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    corner,
    const SVector3 &    right,
    const SVector3 &    up
    );


/**
 * Initialize quad triangles.
 *
 * Triangle count: 2.
 */
void initialize_quad_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    );


/**
 * Initialize triangles of multiple quads.
 *
 * Triangle count: 2 * @a quad_count.
 */
void initialize_quads_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         quad_count
    );


/**
 * Initialize 8 vertices of axis-aligned box.
 *
 * Vertex count: 8 (2 quads - bottom and top.) Use box with 24 vertices
 * (6 separate quads) to get reasonable texture coordinates and vertex normals.
 */
void initialize_box_8_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    min_point,
    const SVector3 &    max_point
    );


/**
 * Initialize 8 vertices of axis-aligned box.
 *
 * Vertex count: 24 (6 quads - south, east, north, west, bottom and top.)
 */
void initialize_box_24_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    const SVector3 &    min_point,
    const SVector3 &    max_point
    );


/**
 * Initialize triangles of 8-vertex box.
 *
 * Triangle count: 12.
 */
void initialize_box_8_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    );


/**
 * Initialize triangles of 24-vertex box.
 *
 * Triangle count: 12.
 */
void initialize_box_24_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index
    );


/**
 * Initialize circle (regular polygon) vertices (in CCW order).
 *
 * The circle will lie on the X-Y plane.
 *
 * Vertex count: @a vertex_count.
 */
void initialize_circle_vertices
    (
    Mesh &              mesh,
    std::size_t         vertex_index,
    Float32             radius,
    Float32             z,
    std::size_t         vertex_count
    );


/**
 * Initialize triangles of planar convex polygon in fan layout.
 *
 * If @a reverse is @c true, vertices in vertex array will be used in reverse
 * order (clockwise instead of counter-clockwise if the vertices have been
 * initialized by #initialize_circle_vertices()).
 *
 * Triangle count: @a vertex_count - 2.
 */
void initialize_polygon_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         vertex_count,
    bool                reverse = false
    );


/**
 * Initialize pyramid triangles.
 *
 * It expects @c base_vertex_count + 1 vertices: pyramid tip vertex followed
 * by pyramid base in CCW order. You can use #initialize_vertex() and
 * #initialize_circle_vertices() to initialize the vertices.
 *
 * Triangle count: 2 * @a base_vertex_count - 2.
 */
void initialize_pyramid_triangles
    (
    Mesh &              mesh,
    std::size_t         triangle_index,
    std::size_t         vertex_index,
    std::size_t         base_vertex_count
    );


} // namespace Demo::Client::Renderer::Objects
} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "render_objects_inline.h"
#endif


#endif // CLIENT_RENDER_OBJECTS_H
