/** @file renderer.cpp
 * Renderer global objects.
 */

/*
 * begin        : Sun Mar 9 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "renderer/renderer.h"
#include "renderer/pipeline_state.h"
#include "renderer/texture_manager.h"
#include "renderer/model_manager.h"
#include "renderer/pass_manager.h"
#include "renderer/terrain_material.h"
#include "renderer/terrain_settings.h"
#include "sdl/client_sdl.h"
#include <sstream>


namespace Demo {
namespace Client {
namespace Renderer {


Global * Global::self = NULL;


Global::Global()
    {
    hope( self == NULL );
    self = this;
    
    // Initialize OpenGL.

    OpenGL::initialize();

    // Allocate global rendering objects.

    terrain_settings.reset( new TerrainSettings );
    texture_manager.reset( new TextureManager );
    pipeline_state.reset( new PipelineState );
    overlay_pipeline_state.reset( new PipelineState );
    model_manager.reset( new ModelManager );
    pass_manager.reset( new PassManager );
    terrain_material_database.reset( new TerrainMaterialDatabase( "material_database" ) );

    // Initialize global rendering objects.

    PipelineState::apply_constant_defaults();
    PipelineState::apply_defaults();
    OverlayManager::initialize_pipeline_state( get_overlay_pipeline_state() );

    // Default modelview and projection.

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    OpenGL::check_errors( "Renderer initialization: " );
    }
    
    
Global::~Global()
    {
    // Revert pipeline state to defaults.
    // This is needed to unbind all textures before the texture manager
    // is destroyed.

    if( pipeline_state.get() )
        {
        apply_state( PipelineState() );
        }

    // Destroy global rendering objects.

    terrain_material_database.reset();
    pass_manager.reset();
    model_manager.reset();
    overlay_pipeline_state.reset();
    pipeline_state.reset();
    texture_manager.reset();
    terrain_settings.reset();

    // Shutdown OpenGL.

    OpenGL::shutdown();

    hope( self == this );
    self = NULL;
    }

    
void Global::destroy()
    {
    delete self;
    self = NULL;
    }


void Global::apply_state
    (
    const PipelineState & new_state
    )
    {
    hope( self );
    hope( self->pipeline_state.get() );
    self->pipeline_state->apply( new_state );
    }


void Global::flip()
    {
    hope( self );
    
    // Finish per-frame work.

    get_texture_manager().collect_garbage();
    get_model_manager().collect_garbage();
    
    // Prevent random jerking on linux with ATI cards.
    // FIXME: use settings to turn this on/off.
    glFinish();
    SDL::flip();

    OpenGL::check_errors( "Swap buffers: " );
    }


void Global::render
    (
    const SystemTime & time_delta
    )
    {
    hope( self );

    // Render passes.

    hope( self->pass_manager.get() );
    self->pass_manager->render( time_delta );

    // Done.
    
    OpenGL::check_errors( "Render frame: " );
    }


} // namespace Renderer
} // namespace Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "renderer_inline.h"
#endif
