/** @file renderer.h
 * Renderer global objects.
 */

/*
 * begin        : Sun Mar 9 2003
 * copyright    : (C) 2003 by Petr Tovarys, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#ifndef CLIENT_RENDERER_H
#define CLIENT_RENDERER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_H
#include "renderer/opengl.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#include <string>
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Repository of global renderer objects.
 *
 * Only one instance of the renderer may exist. You can use
 * Global::get() to get reference to this instance.
 * All methods are @c static and can be called directly without
 * the reference.
 */
class Global
    {
public:

    /**
     * Constructor.
     *
     * Creating instance of Global allocates and initializes all global
     * rendering objects.
     *
     * Video mode must be alread opened. Video mode can't be changed while
     * Global exists. This means that if you want to change video mode,
     * all textures, overlays and other rendering objects must be destroyed
     * and then created again.
     */
    Global();
    
    /**
     * Destructor.
     *
     * Video mode must still be opened. All rendering objects (overlays,
     * models) should be already destroyed.
     */
    ~Global();

    /**
     * Destroy the renderer if it exists.
     */
    static void destroy();

public:

    /**
     * Get global renderer object.
     */
    static INLINE Global & get();

    /**
     * Get current rendering pipeline state.
     */
    static INLINE const PipelineState & get_pipeline_state();

    /**
     * Get reference to standard overlay pipeline state.
     *
     * See OverlayManager::initialize_pipeline_state() for more info.
     *
     * Feel free to modify this state, apply it and then revert all
     * changes you made.
     */
    static INLINE PipelineState & get_overlay_pipeline_state();

    /**
     * Get texture manager.
     */
    static INLINE TextureManager & get_texture_manager();

    /**
     * Get model manager.
     */
    static INLINE ModelManager & get_model_manager();

    /**
     * Get render pass manager.
     */
    static INLINE PassManager & get_pass_manager();

    /**
     * Get terrain material database.
     */
    static INLINE TerrainMaterialDatabase & get_terrain_material_database();

    /**
     * Get terrain settings.
     */
    static INLINE TerrainSettings & get_terrain_settings();

public:

    /**
     * Apply rendering pipeline state.
     */
    static void apply_state
        (
        const PipelineState & new_state
        );

public:

    /**
     * Render everything.
     */
    static void render
        (
        const SystemTime & time_delta
        );

    /**
     * Flip video buffers.
     */
    static void flip();

private:

    static Global *                             self;
        /**< The only instance Global. */

    std::auto_ptr< TextureManager >             texture_manager;
        /**< Texture manager. */

    std::auto_ptr< ModelManager >               model_manager;
        /**< Model manager. */

    std::auto_ptr< PipelineState >              pipeline_state;
        /**< Current rendering pipeline state. */

    std::auto_ptr< PipelineState >              overlay_pipeline_state;
        /**< Standard overlay pipeline state. */

    std::auto_ptr< PassManager >                pass_manager;
        /**< Render pass manager. */

    std::auto_ptr< TerrainMaterialDatabase >    terrain_material_database;
        /**< Terrain material database. */

    std::auto_ptr< TerrainSettings >            terrain_settings;
        /**< Terrain settings. */

    }; // class Global


} // namespace Renderer
} // namespace Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "renderer_inline.h"
#endif


#endif // CLIENT_RENDERER_H
