/** @file terrain.cpp
 * Heightmap-based 3D terrain (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/terrain.h"
#include "renderer/terrain_build.h"
#include "renderer/terrain_settings.h"

#if TERRAIN_DEBUG >= 1
#include <iostream>
#endif // TERRAIN_DEBUG


namespace Demo {
namespace Client {
namespace Renderer {


// class Heightmap

Heightmap::Heightmap
    (
    const TerrainSettings & the_settings
    ) : x_size( 0 ),
        y_size( 0 ),
        low_height( 0.0f ),
        high_height( 1.0f ),
        texture_scale( 1.0f ),
        settings( the_settings )
    {
    }


Heightmap::~Heightmap()
    {
    }


// class Terrain

Terrain::Terrain()
      : material_objects( NULL ),
        material_object_count( 0 ),
        patch_object( NULL ),
        lighting( true )
    {
    }


Terrain::~Terrain()
    {
    clear();
    }


void Terrain::build
    (
    const Heightmap & heightmap
    )
    {
    // Clear old terrain.

    clear();

    if( heightmap.x_size * heightmap.y_size > 65536 )
        {
        // FIXME: log error
        return;
        }

    // Build new terrain.

    TerrainBuild context( heightmap );
    material_objects = context.build_material( material_object_count );
    patch_object = context.build_patches();

    // Remember lighting on/off.

    lighting = heightmap.settings.enable_light;

#if TERRAIN_DEBUG >= 2
    std::cout << "Terrain dump: " << material_object_count << " shader passes.\n";
    int i;
    for( i = 0; i < material_object_count; i++ )
        {
        std::cout << "Shader pass " << i << ": " <<
            material_objects[ i ].jobs[ 0 ].pipeline_state->texture[ 0 ].data->get_name() << "\n";
        material_objects[ i ].dump( std::cout, "    " );
        std::cout << std::endl;
        }
#endif // TERRAIN_DEBUG
    }


void Terrain::clear()
    {
    delete [] material_objects;
    material_objects = NULL;
    material_object_count = 0;
    delete patch_object;
    patch_object = NULL;
    lighting = true;
    }


void Terrain::render
    (
    const SystemTime &
    )
    {
    hope( material_objects );
    apply_transformation();

    // Render material layer.

    std::size_t i;
    for( i = 0; i < material_object_count; i++ )
        {
        if( lighting && manager )
            {
            material_objects[ i ].set_pipeline_lighting( *manager, this );
            }
        else
            {
            material_objects[ i ].disable_pipeline_lighting();
            }
        material_objects[ i ].render();
        }

    // Render patch layer.

    if( patch_object )
        {
        // patch layer is currently _not_ lit.
        patch_object->render();
        }

    revert_transformation();
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "terrain_inline.h"
#endif
