/** @file terrain.h
 * Heightmap-based 3D terrain.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TERRAIN_H
#define CLIENT_TERRAIN_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_TYPES_H
#include "renderer/opengl_types.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_RENDER_ENTITY_H
#include "renderer/render_entity.h"
#endif

#ifndef CLIENT_TERRAIN_MATERIAL_H
#include "renderer/terrain_material.h"
#endif

#ifndef CLIENT_RENDER_OBJECT_H
#include "renderer/render_object.h"
#endif

#include <vector>


namespace Demo {
namespace Client {
namespace Renderer {


/** @def TERRAIN_DEBUG
 * Terrain debugging level:
 * - 0: No debugging.
 * - 1: Terrain RenderObject dumping.
 * - 2: Lots of messages.
 */
#define TERRAIN_DEBUG 0


/**
 * Heightmap.
 *
 * This structure is parameter of #Terrain::build().
 * To build the terrain, information about all neighboring tiles is needed.
 *
 * Heightmap can have no more than 16384 vertices (not counting neighbor
 * vertices; square heightmap with both #x_size and #y_size equal to 128 has
 * exactly 16384 vertices).
 */
class Heightmap
    {
public:

    /**
     * Heightmap vertex/tile element.
     */
    struct Element
        {
        Float32     height;
            /**< Vertex height.
                 @todo 8bit or 16bit integer. */

        Int32       material_id;
            /**< Vertex material id.
                 @todo 8bit integer.
                 @todo Two material layers? */

        Int32       tile_flags;
            /**< Tile flags.
                 See Demo::Lib::ClientSector. */
        };

    /**
     * Heightmap vertex/tile data.
     */
    typedef std::vector< Element > Elements;

public:

    /**
     * Constructor.
     */
    Heightmap
        (
        const TerrainSettings & the_settings
        );

    /**
     * Destructor.
     */
    ~Heightmap();

public:

    /**
     * Get vertex height.
     *
     * @param x Vertex @a x coordinate, -1 to #x_size.
     * @param y Vertex @a y coordinate, -1 to #y_size.
     */
    INLINE Float32 get_height
        (
        int x,
        int y
        ) const;

    /**
     * Get vertex material.
     *
     * @param x Vertex @a x coordinate, 0 to #x_size - 1.
     * @param y Vertex @a y coordinate, 0 to #y_size - 1.
     *
     * @note Materials are defined for neigbor sectors too, but they should
     *       not be needed by the build process.
     */
    INLINE Int32 get_material
        (
        int x,
        int y
        ) const;

    /**
     * Get tile flags.
     *
     * @param x Tile @a x coordinate, 0 to #x_size - 2.
     * @param y Tile @a y coordinate, 0 to #y_size - 2.
     */
    INLINE Int32 get_tile_flags
        (
        int x,
        int y
        ) const;

    /**
     * Set vertex height.
     *
     * @param x Vertex @a x coordinate, -1 to #x_size.
     * @param y Vertex @a y coordinate, -1 to #y_size.
     */
    INLINE void set_height
        (
        int     x,
        int     y,
        Float32 height
        );

    /**
     * Set vertex material.
     *
     * @param x Vertex @a x coordinate, 0 to #x_size - 1.
     * @param y Vertex @a y coordinate, 0 to #y_size - 1.
     */
    INLINE void set_material
        (
        int     x,
        int     y,
        Int32   material
        );

    /**
     * Set tile flags.
     *
     * @param x Tile @a x coordinate, 0 to #x_size - 2.
     * @param y Tile @a y coordinate, 0 to #y_size - 2.
     */
    INLINE void set_tile_flags
        (
        int     x,
        int     y,
        Int32   flags
        );

public:

    int         x_size;
        /**< Heightmap west-east size. Number of vertices. */

    int         y_size;
        /**< Heightmap south-north size. Number of vertices. */

    Elements    elements;
        /**< Vertex/tile data. ( x_size + 2 ) * ( y_size + 2 ) vector.
             The array starts with vertex/tile at ( -1, -1 ). Elements are stored
             "row" by "row", starting with southernmost row and westernmost
             vertex/tile. See get/set methods for valid index ranges for various
             components of Element. */

    Float32     low_height;
        /**< Lower bound of height range.
             Should be less than or equal to minimum vertex height. */

    Float32     high_height;
        /**< Upper bound of height range.
             Should be greater than or equal to maximum vertex height. */

    Float32     texture_scale;
        /**< Number of texture repeats per tile.
             @todo This (and more) should be specified in material/shared too, probably. */

public:

    const TerrainSettings & settings;
        /**< Terrain renderer settings. */

    }; // class Heightmap

 
/**
 * Heighmap-based terrain.
 *
 * Terrain is visual representation of a heightmap. It consists of two layers:
 * material and patches.
 *
 * Material is defined for each heighmap vertex (NOT tile). Set of textures
 * to mix at this vertex is determined from terrain height and slope.
 * See #TerrainMaterial info about the materials.
 *
 * Patch layer is drawn on top of material layer. Patches are not implemented
 * and not documented yet.
 *
 * @todo Check material updates in render(), report that terrain should be
 *       rebuilt if materials change.
 * @todo Patch layer.
 *
 * @todo Multiple (2) materials per vertex?
 *
 * @todo Support "none" material. Do not draw triangles with all "none"
 *       vertices at all.
 */
class Terrain : public RenderEntity
    {
    NO_COPY_CONSTRUCTOR( Terrain );
    NO_ASSIGNMENT_OPERATOR( Terrain );

public:

    /**
     * Constructor.
     */
    Terrain();

    /**
     * Destructor.
     */
    virtual ~Terrain();

public:

    /**
     * Rebuild terrain.
     */
    void build
        (
        const Heightmap & heightmap
        );

    /**
     * Destroy all terrain render objects.
     */
    void clear();

public:

    /**
     * Render the terrain.
     *
     * All terrains will be rendered using the following coordinates:
     * - @a x and @a y coordinate will be equal to @a x and - @a y index of
     *   the vertex. South-westernmost vertex will be rendered at ( 0, 0 ),
     *   north-easternmost at ( @a x_size - 1, @a y_size - 1 ).
     * - @a z coordinate will be equal to vertex height set by #build().
     *
     * Use RenderEntity transformation to move, scale and rotate the terrain.
     *
     * If texture data objects used by the terrain are updated, new texture
     * data will be used - this works automatically everywhere. However, if
     * terrain material database or material definition is updated, #Terrain
     * won't detect the change until #build() is called.
     *
     * @note It should be possible to detect material/database update. Just
     *       remember ids and versions of all used materials and check them
     *       (and add versioning similar to font or texture to materials).
     *       But we would have to remember the heightmap to use the new
     *       materials.
     */
    virtual void render
        (
        const SystemTime & time_delta
        );

private:

    RenderObject *  material_objects;
        /**< Array of material layer render objects. */

    std::size_t     material_object_count;
        /**< Number of material layer render objects. */

    RenderObject *  patch_object;
        /**< Patch render object. May be @c NULL. */

    bool            lighting;
        /**< Use OpenGL lighting. */

    }; // class Terrain


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "terrain_inline.h"
#endif


#endif // CLIENT_TERRAIN_H
