/** @file terrain_build.cpp
 * Terrain building (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "renderer/terrain_build.h"
#include "renderer/terrain_settings.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include "lib/shared/client_sector.h"

#if TERRAIN_DEBUG >= 1
#include <iostream>
#endif // TERRAIN_DEBUG


namespace Demo {
namespace Client {
namespace Renderer {


// class TerrainBuild::VertexInfo

void TerrainBuild::VertexInfo::add_texture
    (
    TextureDataObject::Reference    texture,
    Float32                         weight
    )
    {
    // Check if this vertex already has this texture.

    if( textures[ 0 ] == texture )
        {
        // Yep - add weights, done.

        weights[ 0 ] += weight;
        }
    else
        {
        // New texture.

        const int i = ( textures[ 0 ].is_null() ? 0 : 1 );
        hope( textures[ i ].is_null() );
        textures[ i ] = texture;
        weights[ i ] = weight;
        }
    }


// class TerrainBuild

TerrainBuild::TerrainBuild
    (
    const Heightmap &   the_heightmap
    ) : heightmap( the_heightmap ),
        triangle_strip_count( 0 ),
        material_objects( NULL ),
        patch_object( NULL )
    {
    hope( heightmap.x_size * heightmap.y_size <= 16384 );
    }


TerrainBuild::~TerrainBuild()
    {
    delete [] material_objects;
    delete patch_object;
    }


RenderObject * TerrainBuild::build_material
    (
    std::size_t & render_object_count
    )
    {
    gather_vertex_info();

    // Allocate single render object per texture.
    // Each unique terrain texture is drawn in separate pass. Multitexturing
    // is not used. (It could be, but we would have to use combine texenv, and
    // dermining which textures should be merged into the same pass is very
    // tricky. Multitexturing could be used in the future when simple textures
    // are replaced by shaders to render multi-layer shaders, but not to merge
    // multiple material layers into single pass.)

    render_object_count = used_textures.size();
    material_objects = new RenderObject [ render_object_count ];

    // Determine which texture covers the most vertices.

    TextureDataObject * first_texture = NULL;
    int first_texture_vertices = 0;
    TextureVertexCount::iterator tit;
    for( tit = used_textures.begin(); tit != used_textures.end(); ++tit )
        {
        hope( tit->second <= heightmap.x_size * heightmap.y_size );
        if( tit->second > first_texture_vertices )
            {
            first_texture = tit->first;
            first_texture_vertices = tit->second;
            }
        }
    hope( first_texture );

    // Draw all triangles using first_texture in the first pass.

    build_material_jobs( 0, first_texture );
    clear_material_jobs_temporaries();
    build_material_arrays( 0, first_texture );
    clear_material_arrays_temporaries();
    
    // For each used texture data object, walk through all triangles in triangle
    // strip order. Determine longest triangle strip run. Prepare triangle strip
    // and separate triangle arrays. Allocate arrays and initialize them. Repeat.
    // Note: We only try simple "west-east" strips to make build process simple.
    // Better arrays could be built using more complex algorithm.

    int layer_index = 1;
    for( tit = used_textures.begin(); tit != used_textures.end(); ++tit )
        {
        TextureDataObject * const texture = tit->first;
        hope( texture );
        if( texture != first_texture )
            {
            build_material_jobs( layer_index, texture );
            clear_material_jobs_temporaries();
            build_material_arrays( layer_index, texture );
            clear_material_arrays_temporaries();
            layer_index++;
            }
        }

    // Return generated render objects to caller.

    RenderObject * const result = material_objects;
    material_objects = NULL;
    return result;
    }


RenderObject * TerrainBuild::build_patches()
    {
    // Do nothing if marks are disabled - patches are the only marks supported.

    if( !heightmap.settings.draw_marks )
        {
        return NULL;
        }

    // Determine number of patch quads.

    std::size_t quad_count = 0;
    int y;
    for( y = 0; y < heightmap.y_size - 1; y++ )
        {
        int x;
        for( x = 0; x < heightmap.x_size - 1; x++ )
            {
            const Int32 flags = heightmap.get_tile_flags( x, y );
            if( flags & ::Demo::Lib::ClientSector::TILE_MARK )
                {
                quad_count++;
                }
            }
        }

    // Done if no quads.

    if( !quad_count )
        {
        return NULL;
        }

    // Allocate patch render object.

    patch_object = new RenderObject;

    // Allocate and initialize pipeline state (z tested, blended).
    // Note: used_textures.size() is equal to number of material layers.

    patch_object->pipeline_states.resize( 1 );
    PipelineState & state = patch_object->pipeline_states.front();
    state.depth_test.enable = true;
    state.depth_test.function = GL_LEQUAL;
    state.blend.enable = true;
    state.blend.source = GL_SRC_ALPHA;
    state.blend.destination = GL_ONE_MINUS_SRC_ALPHA;
    initialize_fog_pipeline_state( state, false );

    // Optional polygon offset.

    if( heightmap.settings.depth_offset_factor.read() || heightmap.settings.depth_offset_bias.read() )
        {
        state.depth_offset.enable_polygon = true;
        state.depth_offset.factor = used_textures.size() * heightmap.settings.depth_offset_factor;
        state.depth_offset.bias = used_textures.size() * heightmap.settings.depth_offset_bias;
        }

    // Allocate and initialize geometry.
    // Draw patches as lots of separate triangles. Quads can't be used, because
    // they are not planar and different cards draw non-planar quads differently.
    // Currently patches are used only to mark some tiles - polygon count should
    // be low and stripification should not be needed.

    patch_object->geometries.resize( 1 );
    RenderObject::Geometry & geometry = patch_object->geometries.front();
    
    #if 0
    
    geometry.initialize_range( GL_TRIANGLES, 0, 6 * quad_count );
    
    #else
    
    // Some lame drivers don't draw long array ranges correctly.
    
    int i;
    
    geometry.initialize_index_array( GL_TRIANGLES, 6 * quad_count );
    geometry.lock();
    for( i = 0; i < static_cast<int>( 6 * quad_count ); i++ )
        {
        geometry.indices[ i ] = i;
        }
    geometry.unlock();
    
    #endif

    // Allocate and initialize job.

    patch_object->jobs.resize( 1 );
    RenderObject::Job & job = patch_object->jobs.front();
    job.pipeline_state = &state;
    job.geometry = &geometry;

    // Allocate arrays.

    VertexArray & array = patch_object->array;
    array.set_stream_count( VertexArray::TYPE_VECTOR3, 1 );
    array.set_stream_count( VertexArray::TYPE_COLOR, 1 );
    array.allocate( 6 * quad_count );
    array.select( VertexArray::VERTEX_ARRAY, VertexArray::TYPE_VECTOR3, 0 );
    array.select( VertexArray::COLOR_ARRAY, VertexArray::TYPE_COLOR, 0 );

    // Initialize arrays.

    VertexArray::Vector3Stream positions;
    array.lock_vectors3( positions, 0 );
    int index = 0;
    for( y = 0; y < heightmap.y_size - 1; y++ )
        {
        int x;
        for( x = 0; x < heightmap.x_size - 1; x++ )
            {
            const Int32 flags = heightmap.get_tile_flags( x, y );
            if( flags & ::Demo::Lib::ClientSector::TILE_MARK )
                {
                #define SET_VERTEX( i, xo, yo ) \
                            positions[ index + i ] = get_heightmap_opengl_vertex \
                                ( x + xo, y + yo, heightmap.get_height( x + xo, y + yo ) );
                SET_VERTEX( 0, 0, 1 );
                SET_VERTEX( 1, 0, 0 );
                SET_VERTEX( 2, 1, 0 );
                
                SET_VERTEX( 3, 0, 1 );
                SET_VERTEX( 4, 1, 0 );
                SET_VERTEX( 5, 1, 1 );
                #undef SET_VERTEX
                index += 6;
                }
            }
        }
    array.unlock_vectors3( 0 );

    VertexArray::ColorStream colors;
    array.lock_colors( colors, 0 );
    index = 0;
    for( y = 0; y < heightmap.y_size - 1; y++ )
        {
        int x;
        for( x = 0; x < heightmap.x_size - 1; x++ )
            {
            const Int32 flags = heightmap.get_tile_flags( x, y );
            if( flags & ::Demo::Lib::ClientSector::TILE_MARK )
                {
                #define SET_VERTEX( i ) \
                            colors[ index + i ] = OpenGL::Color( 1.0f, 0.7f, 0.3f, 0.5f )
                SET_VERTEX( 0 );
                SET_VERTEX( 1 );
                SET_VERTEX( 2 );
                SET_VERTEX( 3 );
                SET_VERTEX( 4 );
                SET_VERTEX( 5 );
                #undef SET_VERTEX
                index += 6;
                }
            }
        }
    array.unlock_colors( 0 );

    // Return generated render object to caller.

    RenderObject * const result = patch_object;
    patch_object = NULL;
    return result;
    }


void TerrainBuild::gather_vertex_info()
    {
    // Array of material pointers. Used to cache results from get_material().
    const TerrainMaterialDatabase::MaterialInfo *
        used_materials[ TerrainMaterialDatabase::MAX_MATERIAL_ID + 1 ];

    {//sob
    int i;
    for( i = 0; i <= TerrainMaterialDatabase::MAX_MATERIAL_ID; i++ )
        {
        used_materials[ i ] = NULL;
        }
    }//eob

    TerrainMaterialDatabase & database = Global::get_terrain_material_database();

    const int x_size = heightmap.x_size;
    const int y_size = heightmap.y_size;
    vertex_infos.clear();
    vertex_infos.resize( x_size * y_size );
    int y;
    for( y = 0; y < y_size; y++ )
        {
        int x;
        for( x = 0; x < x_size; x++ )
            {
            #if TERRAIN_DEBUG >= 2
            std::cout << "Vertex ( " << x << ", " << y << " ):\n";
            #endif

            // Copy vertex height.

            VertexInfo & vertex = get_vertex_info( x, y );
            vertex.height = heightmap.get_height( x, y );

            #if TERRAIN_DEBUG >= 2
            std::cout << "    height " << vertex.height << "\n";
            #endif

            // Compute vertex normal.
            // If E, N, W and S are heights of the eastern, northern, western
            // and southern neighbor vertices, normals of the four neighbor
            // triangles are really simple:
            //   ( -E, -N, 1 ), ( W, -N, 1 ), ( W, S, 1 ), ( -E, S, 1 )
            // We want to compute vertex normal as combination of these normals.
            // Since their lenghts are approximately the same, we will just do:
            //   n = normalize( 0.25 * ( n1 + n2 + n3 + n4 ) )
            //     = normalize( 0.25 * ( 2W - 2E, 2S -2N, 4 ) )
            //     = normalize( W - E, S - N, 2 )

            vertex.normal = SVector3
                (
                heightmap.get_height( x - 1, y ) - heightmap.get_height( x + 1, y ),
                heightmap.get_height( x, y - 1 ) - heightmap.get_height( x, y + 1 ),
                2.0f
                ).normalize();


            #if TERRAIN_DEBUG >= 2
            std::cout << "    normal " << vertex.normal.to_string() << "\n";
            #endif

            // Compute terrain slope.

            const Float32 slope = Massiv::Core::clamp( 1.0f - vertex.normal.z(), 0.0f, 1.0f );

            #if TERRAIN_DEBUG >= 2
            std::cout << "    slope " << slope << "\n";
            #endif

            // Determine texturing.

            const Int32 material_id = heightmap.get_material( x, y );
            hope( material_id >= 0 );
            hope( material_id <= TerrainMaterialDatabase::MAX_MATERIAL_ID );
            if( used_materials[ material_id ] == NULL )
                {
                used_materials[ material_id ] = &database.get_material( material_id );
                }
            const TerrainMaterialDatabase::MaterialInfo * const info =
                used_materials[ material_id ];
            hope( info );

            // Get the two textures.

            TextureDataObject::Reference    texture1;
            TextureDataObject::Reference    texture2;
            Float32                         weight;
            const Float32                   normalized_height = Massiv::Core::clamp
                (
                ( vertex.height - heightmap.low_height ) /
                ( heightmap.high_height - heightmap.low_height ),
                0.0f,
                1.0f
                );
            info->material.get_textures
                ( texture1, texture2, weight, normalized_height, slope );

            // Add the textures to the vertex and update #used_textures.
            // Note: Updates of used_textures assume that ints are default-constructed to 0,
            //       which is behavior required by the standard.

            if( !texture1.is_null() )
                {
                #if TERRAIN_DEBUG >= 2
                std::cout << "    texture1 " << texture1->get_name() << " " << weight << "\n";
                #endif

                vertex.add_texture( texture1, weight );
                used_textures[ texture1.get() ]++;
                }
            if( !texture2.is_null() )
                {
                #if TERRAIN_DEBUG >= 2
                std::cout << "    texture2 " << texture2->get_name() << " " << 1.0f -  weight << "\n";
                #endif

                vertex.add_texture( texture2, 1.0f - weight );
                used_textures[ texture2.get() ]++;
                }
            } // for x
        } // for y
    }


void TerrainBuild::build_material_jobs
    (
    int                 layer_index,
    TextureDataObject * texture
    )
    {
    hope( layer_index >= 0 );
    hope( texture );
    const bool first_layer = ( layer_index == 0 ? true : false );

    // Check that temporary structures are clean.

    hope( array_vertices.empty() );
    hope( triangles.empty() );
    hope( triangle_strips.empty() );
    hope( triangle_strip_count ==  0 );

    // Clear vertex mapping.

    {//sob
    const int vertex_count = heightmap.x_size * heightmap.y_size;
    int vertex_index;
    for( vertex_index = 0; vertex_index < vertex_count; vertex_index++ )
        {
        VertexInfo & vertex_info = vertex_infos[ vertex_index ];
        vertex_info.array_index = -1;
        }
    }//eob

    // Build the jobs in separate structures first. Exact sizes are required
    // to allocate index arrays and we don't know them yet. And arrays might
    // be stored in uncached memory, so we just want to copy the results there.

    const int x_tris = 2 * ( heightmap.x_size - 1 );
    const int y_strips = heightmap.y_size - 1;
    int y;
    for( y = 0; y < y_strips; y++ )
        {
        if( first_layer )
            {
            // Draw all triangles in the first layer.

            prepare_triangle_strip( y, 0, x_tris - 1 );
            }
        else
            {
            // Determine which triangles use @a texture.

            int x = 0;
            while( x < x_tris )
                {
                // Find strip.

                x = strip_start( y, x, texture );
                if( x >= x_tris )
                    {
                    break;
                    }
                const int x_end = strip_end( y, x, texture );
                hope( x_end < x_tris );

                // Separate triangles or strip.
                // This updates vertex mapping.

                if( x_end - x < heightmap.settings.min_strip_length.read() )
                    {
                    prepare_separate_triangles( y, x, x_end );
                    }
                else
                    {
                    prepare_triangle_strip( y, x, x_end );
                    }

                // Try next strip.

                x = x_end + 1;
                }
            }
        }
    hope( static_cast< std::size_t >( triangle_strip_count ) == triangle_strips.size() );

    // Allocate and initialize pipeline states.
    // RenderObject for first pass has single state (all triangles are CW)
    // plus optional wireframe job pipeline state.
    // RenderObject for other passes has two states - CW (0) and CCW (1).

    PipelineState * cw_state = NULL;
    PipelineState * ccw_state = NULL;
    PipelineState * wireframe_state = NULL;
    const bool draw_wireframe = heightmap.settings.draw_wireframe;
    if( first_layer )
        {
        material_objects[ layer_index ].pipeline_states.resize( draw_wireframe ? 2 : 1 );
        cw_state = &material_objects[ layer_index ].pipeline_states[ 0 ];
        initialize_material_pipeline_state( *cw_state, layer_index, texture, GL_CW );
        if( draw_wireframe )
            {
            wireframe_state = &material_objects[ layer_index ].pipeline_states[ 1 ];
            initialize_wireframe_pipeline_state( *wireframe_state );
            }
        }
    else
        {
        material_objects[ layer_index ].pipeline_states.resize( 2 );
        cw_state = &material_objects[ layer_index ].pipeline_states[ 0 ];
        ccw_state = &material_objects[ layer_index ].pipeline_states[ 1 ];
        initialize_material_pipeline_state( *cw_state, layer_index, texture, GL_CW );
        initialize_material_pipeline_state( *ccw_state, layer_index, texture, GL_CCW );
        }

    // Allocate geometries and jobs.

    hope( material_objects );
    RenderObject::Geometries & geometries = material_objects[ layer_index ].geometries;
    RenderObject::Jobs & jobs = material_objects[ layer_index ].jobs;
    const std::size_t job_count = triangle_strip_count + ( triangles.empty() ? 0 : 1 );
    geometries.clear();
    jobs.clear();
    geometries.resize( job_count );
    jobs.resize( wireframe_state ? 2 * job_count : job_count );
    std::size_t job_index = 0;

    // Build separate triangle job.

    if( !triangles.empty() )
        {
        RenderObject::Geometry & geometry = geometries[ job_index ];
        RenderObject::Job & job = jobs[ job_index ];

        // Initialize geometry.

        geometry.initialize_index_array( GL_TRIANGLES, 3 * triangles.size() );
        geometry.lock();
        int vertex_i = 0;
        Triangles::const_iterator tri_it;
        for( tri_it = triangles.begin(); tri_it != triangles.end(); ++tri_it )
            {
            geometry.indices[ vertex_i++ ] = static_cast< GLushort >( tri_it->v1 );
            geometry.indices[ vertex_i++ ] = static_cast< GLushort >( tri_it->v2 );
            geometry.indices[ vertex_i++ ] = static_cast< GLushort >( tri_it->v3 );
            }
        geometry.unlock();

        // Initialize job.

        hope( cw_state );
        job.pipeline_state = cw_state;
        job.geometry = &geometry;

        // Next job.

        job_index++;
        }

    // Build tristrip jobs.

    std::size_t first_cw_job = job_index;
    initialize_tristrip_material_jobs( material_objects[ layer_index ], job_index, GL_CW );
    const std::size_t first_ccw_job = job_index;
    initialize_tristrip_material_jobs( material_objects[ layer_index ], job_index, GL_CCW );
    hope( job_index == job_count );

    // Build wireframe jobs.

    if( wireframe_state )
        {
        hope( first_ccw_job == job_count );

        if( !triangles.empty() )
            {
            jobs[ job_index ].pipeline_state = wireframe_state;
            jobs[ job_index ].geometry = &geometries[ 0 ];
            job_index++;
            }
        initialize_tristrip_wireframe_jobs( material_objects[ layer_index ], job_index, first_cw_job );
        hope( job_index == 2 * job_count );
        hope( first_cw_job == job_count );
        }
    }


void TerrainBuild::clear_material_jobs_temporaries()
    {
    triangles.clear();
    triangle_strips.clear();
    triangle_strip_count = 0;
    }


void TerrainBuild::build_material_arrays
    (
    int                 layer_index,
    TextureDataObject * texture
    )
    {
    hope( layer_index >= 0 );
    hope( texture );

    // Vertex mapping must be initialized.

    const std::size_t vertex_count = array_vertices.size();
    hope( vertex_count > 0 );

    // Allocate vertex array.

    hope( material_objects );
    VertexArray & array = material_objects[ layer_index ].array;
    if( array.get_size() != vertex_count )
        {
        array.free();
        array.reset_stream_counts();
        array.set_stream_count( VertexArray::TYPE_VECTOR3, 2 ); // position, normal
        array.set_stream_count( VertexArray::TYPE_VECTOR2, 1 ); // texcoord
        array.set_stream_count( VertexArray::TYPE_COLOR, 1 );   // weight * color
        array.allocate( vertex_count );
        }

    // Select streams.

    array.select( VertexArray::VERTEX_ARRAY, VertexArray::TYPE_VECTOR3, 0 );
    array.select( VertexArray::NORMAL_ARRAY, VertexArray::TYPE_VECTOR3, 1 );
    array.select( VertexArray::TEXCOORD_ARRAY_0, VertexArray::TYPE_VECTOR2, 0 );
    array.select( VertexArray::COLOR_ARRAY, VertexArray::TYPE_COLOR, 0 );

    // Initialize arrays. Because they are not interleaved and may be stored
    // in uncached memory, initialize each array individually (->sequentially).

    ArrayVertices::const_iterator it;
    int i;

    // Vertex positions.

    VertexArray::Vector3Stream positions;
    array.lock_vectors3( positions, 0 );
    for( it = array_vertices.begin(), i = 0; it != array_vertices.end(); ++it, i++ )
        {
        VertexInfo & vertex = get_vertex_info( it->x, it->y );
        positions[ i ] = get_heightmap_opengl_vertex( it->x, it->y, vertex.height );
        }
    array.unlock_vectors3( 0 );

    // Normals.

    VertexArray::Vector3Stream normals;
    array.lock_vectors3( normals, 1 );
    for( it = array_vertices.begin(), i = 0; it != array_vertices.end(); ++it, i++ )
        {
        VertexInfo & vertex = get_vertex_info( it->x, it->y );
        normals[ i ] = OpenGL::Vector3( vertex.normal );
        }
    array.unlock_vectors3( 1 );

    // Weights/colors.
    // Note: There is no vertex coloring yet.

    VertexArray::ColorStream colors;
    array.lock_colors( colors, 0 );
    for( it = array_vertices.begin(), i = 0; it != array_vertices.end(); ++it, i++ )
        {
        VertexInfo & vertex = get_vertex_info( it->x, it->y );
        const GLfloat weight = static_cast< GLfloat >( vertex.get_texture_weight( texture ) );
        colors[ i ] = OpenGL::Color( weight, weight, weight, 1.0f );
        }
    array.unlock_colors( 0 );

    // Texture coordinates.

    VertexArray::Vector2Stream tex_coords;
    array.lock_vectors2( tex_coords, 0 );
    for( it = array_vertices.begin(), i = 0; it != array_vertices.end(); ++it, i++ )
        {
        tex_coords[ i ] = OpenGL::Vector2
            (
            static_cast< GLfloat >( it->x * heightmap.texture_scale ),
            static_cast< GLfloat >( it->y * heightmap.texture_scale )
            );
        }
    array.unlock_vectors2( 0 );
    }


void TerrainBuild::clear_material_arrays_temporaries()
    {
    array_vertices.clear();
    }


void TerrainBuild::initialize_material_pipeline_state
    (
    PipelineState &     state,
    int                 layer_index,
    TextureDataObject * texture,
    GLenum              front_face
    )
    {
    hope( texture );

    // Front face.

    state.cull.front_face = front_face;

    // Optional polygon depth offset.

    if( ( layer_index > 0 ) && ( heightmap.settings.depth_offset_factor.read() ||
        heightmap.settings.depth_offset_bias.read() ) )
        {
        state.depth_offset.enable_polygon = true;
        state.depth_offset.factor = layer_index * heightmap.settings.depth_offset_factor;
        state.depth_offset.bias = layer_index * heightmap.settings.depth_offset_bias;
        }

    // Use z-buffering. Write z in first layer, test z in other layers.

    state.depth_test.enable = true;
    if( layer_index == 0 )
        {
        state.depth_test.function = GL_LEQUAL;
        state.buffer_mask.depth = true;
        }
    else
        {
        state.depth_test.function = GL_LEQUAL;
        state.buffer_mask.depth = false;
        }

    // Blending. Replace in first layer, add in other layers.

    if( layer_index == 0 )
        {
        state.blend.enable = false;
        }
    else
        {
        state.blend.enable = true;
        state.blend.source = GL_ONE;
        state.blend.destination = GL_ONE;
        }

    // Texture.

    if( heightmap.settings.enable_texture )
        {
        state.texture[ 0 ].data = texture;
        state.texture[ 0 ].flags = Texture::GENERATE_MIPMAPS;
        state.texture[ 0 ].mode = GL_MODULATE;
        }

    // Lighting material.

    if( heightmap.settings.enable_light )
        {
        state.lighting.color_material.enable = true;
        state.lighting.color_material.mode = GL_AMBIENT_AND_DIFFUSE;
        }

    // Fog.

    initialize_fog_pipeline_state( state, layer_index > 0 );
    }


void TerrainBuild::initialize_wireframe_pipeline_state
    (
    PipelineState &     state
    )
    {
    // Prepare pipeline state for the wireframe layer.
    // This is a nasty hack. Wireframe triangles are jobs of the first layer
    // render object. Its array includes vertex normals, vertex colors and
    // texturing coordinates, none of which is needed by wireframe layer.
    // But I'm too lazy to build new arrays just to draw wireframes.

    // Polygon mode.

    state.polygon_mode = GL_LINE;

    // Front face. Always CW.

    state.cull.front_face = GL_CW;

    // Optional line depth offset - between first and second layer.

    if( heightmap.settings.depth_offset_factor.read() || heightmap.settings.depth_offset_bias.read() )
        {
        state.depth_offset.enable_line = true;
        state.depth_offset.factor = 0.5f * heightmap.settings.depth_offset_factor;
        state.depth_offset.bias = 0.5f * heightmap.settings.depth_offset_bias;
        }

    // Depth testing.

    state.depth_test.enable = true;
    state.depth_test.function = GL_LEQUAL;
    state.buffer_mask.depth = false;

    // Additive blending.

    state.blend.enable = true;
    state.blend.source = GL_ONE;
    state.blend.destination = GL_ONE;

    // Texturing. Really nasty hack. The array we use has vertex colors. We don't
    // want them. So we use default texture in GL_REPLACE mode. Buah.

    state.texture[ 0 ].data = Global::get_texture_manager().
                get_texture_data_object( "image/default" );
    state.texture[ 0 ].flags = Texture::GENERATE_MIPMAPS;
    state.texture[ 0 ].mode = GL_REPLACE;

    // Fog.

    initialize_fog_pipeline_state( state, true );
    }


void TerrainBuild::initialize_fog_pipeline_state
    (
    PipelineState &     state,
    bool                additive_layer
    )
    {
    /*
    Note:
    Terrain is rendered in multiple additive passes. Without fogging, the
    equation look like this:
        Cframebuffer = sum( i = 0 .. N, Ctexture_i * Cweight_i * Clight ),
    where
        N is index of the last pass,
        Ctexture_i is color from texture used in the i-th pass,
        Cweight_i is interpolated weight of the fragment and
        Clight is color computed by OpenGL lighting.

    OpenGL fogging takes place after texturing and before blending:
        C = f * Cfragment + ( 1- f ) * Cfog,
    where
        f is fog depth of the fragment (1 = no fog, 0 = full fog)

    For terrain with colored fog we want:
        Cframebuffer = f * sum( i = 0 .. N, Ctexture_i * Cweight_i * Clight ) +
                       ( 1 - f ) * Cfog

    Because the sum represents multiple passes, we can use fogging only
    inside the sum. Fortunately, it's still possible to use it:
        Cframebuffer = sum( i = 0 .. N, f * Ctexture_i * Cweight_i * Clight ) +
                       ( 1 - f ) * Cfog
                     = f * Ctexture_0 * Cweight_0 * Clight +
                       sum( i = 1 .. N, f * Ctexture_i * Cweight_i * Clight ) +
                       ( 1 - f ) * Cfog
                     = f * Ctexture_0 * Cweight_0 * Clight + ( 1 - f ) * Cfog +
                       sum( i = 1 .. N, f * Ctexture_i * Cweight_i * Clight + ( 1 - f ) * 0 ),
    ie. we enable fogging in all passes, with real fog color in the first pass
    and black fog in the latter passes.

    Addition alpha-blended passes use real fog color too - fog works ok with standard
    alpha-blending.
    */
    heightmap.settings.apply_fog( state );
    if( state.fog.enable && additive_layer )
        {
        state.fog.color = OpenGL::FloatColor( 0.0f, 0.0f, 0.0f );
        }
    }


void TerrainBuild::initialize_tristrip_material_jobs
    (
    RenderObject &      render_object,
    std::size_t &       job_index,
    GLenum              front_face
    )
    {
    hope( front_face == GL_CW || front_face == GL_CCW );
    RenderObject::Geometries & geometries = render_object.geometries;
    RenderObject::Jobs & jobs = render_object.jobs;

    bool first_job = true;
    TriangleStrips::const_iterator strip_it;
    for( strip_it = triangle_strips.begin(); strip_it != triangle_strips.end(); ++strip_it )
        {
        const TriangleStrip & strip = *strip_it;
        if( strip.front_face == front_face )
            {
            hope( geometries.size() > job_index );
            hope( jobs.size() > job_index );

            // Initialize geometry.

            RenderObject::Geometry & geometry = geometries[ job_index ];
            const TriangleStrip::Indices & indices = strip.indices;
            geometry.initialize_index_array( GL_TRIANGLE_STRIP, indices.size() );
            geometry.lock();
            std::size_t i;
            for( i = 0; i < indices.size(); i++ )
                {
                geometry.indices[ i ] = static_cast< GLushort >( indices[ i ] );
                }
            geometry.unlock();

            // Initialize job.

            RenderObject::Job & job = jobs[ job_index ];
            if( first_job )
                {
                const std::size_t face_index = ( front_face == GL_CW ? 0 : 1 );
                hope( render_object.pipeline_states.size() >= face_index + 1 );
                job.pipeline_state = &render_object.pipeline_states[ face_index ];
                first_job = false;
                }
            job.geometry = &geometry;

            // Next job.

            job_index++;
            }
        }
    }


void TerrainBuild::initialize_tristrip_wireframe_jobs
    (
    RenderObject &      render_object,
    std::size_t &       job_index,
    std::size_t &       geometry_index
    )
    {
    RenderObject::Geometries & geometries = render_object.geometries;
    RenderObject::Jobs & jobs = render_object.jobs;

    bool first_job = true;
    TriangleStrips::const_iterator strip_it;
    for( strip_it = triangle_strips.begin(); strip_it != triangle_strips.end(); ++strip_it )
        {
        const TriangleStrip & strip = *strip_it;
        hope( strip.front_face == GL_CW );
        hope( geometries.size() > geometry_index );
        hope( jobs.size() > job_index );

        // Initialize job.

        RenderObject::Job & job = jobs[ job_index ];
        if( first_job )
            {
            hope( render_object.pipeline_states.size() >= 2 );
            job.pipeline_state = &render_object.pipeline_states[ 1 ];
            first_job = false;
            }
        job.geometry = &geometries[ geometry_index ];

        // Next job.

        job_index++;
        geometry_index++;
        }
    }


int TerrainBuild::strip_start
    (
    int                 strip_index,
    int                 triangle_index,
    TextureDataObject * texture
    ) const
    {
    // Note: Strips always start at even triangles, because mixing CW and CCW
    // strips causes Z-fighting on some hardware (ATI). Only strip_start() and
    // strip_end() methods have been modified, both CW and CCW strips are still
    // supported by the terrain builder. In the future cvar could control this
    // behavior.

    hope( strip_index >= 0 );
    hope( strip_index < heightmap.y_size - 1 );
    const int strip_vertices = 2 * heightmap.x_size;
    hope( triangle_index >= 0 );
    hope( triangle_index < strip_vertices - 2 );
    hope( texture );

    if( heightmap.settings.force_even_strips )
        {
        hope( !( triangle_index & 1 ) ); // start at even triangle
        }

    // Check the three triangle vertices.

    int x, y;
    int vertex_index;
    for( vertex_index = triangle_index; vertex_index < triangle_index + 3; vertex_index++ )
        {
        strip_vertex_coordinates( x, y, strip_index, vertex_index );

        if( get_vertex_info( x, y ).has_texture( texture ) )
            {
            return triangle_index;
            }
        }

    // Check the following triangles.

    triangle_index++;
    while( vertex_index < strip_vertices )
        {
        strip_vertex_coordinates( x, y, strip_index, vertex_index );

        if( get_vertex_info( x, y ).has_texture( texture ) )
            {
            if( heightmap.settings.force_even_strips )
                {
                return ( triangle_index & ~1 ); // start at even triangle
                }
            else
                {
                return triangle_index;
                }
            }

        vertex_index++;
        triangle_index++;
        }

    hope( vertex_index == strip_vertices );
    hope( triangle_index == strip_vertices - 2 );
    return triangle_index;
    }


int TerrainBuild::strip_end
    (
    int                 strip_index,
    int                 triangle_index,
    TextureDataObject * texture
    ) const
    {
    // All strips end at odd triangle. See strip_start() comment above...

    hope( strip_index >= 0 );
    hope( strip_index < heightmap.y_size - 1 );
    const int strip_triangles = 2 * ( heightmap.x_size - 1);
    hope( triangle_index >= 0 );
    hope( triangle_index < strip_triangles );

    if( heightmap.settings.force_even_strips )
        {
        hope( !( strip_triangles & 1 ) );
        hope( !( triangle_index & 1 ) );
        }

    triangle_index++;
    while( triangle_index < strip_triangles )
        {
        // Check all three vertices.

        int i;
        for( i = 0; i < 3; i++ )
            {
            int x, y;
            strip_vertex_coordinates( x, y, strip_index, triangle_index + i );

            if( get_vertex_info( x, y ).has_texture( texture ) )
                {
                break;
                }
            }
        if( i == 3 )
            {
            // Does not use texture, return previous triangle as the last one.

            if( heightmap.settings.force_even_strips )
                {
                return ( triangle_index - 1 ) | 1; // end at odd triangle
                }
            else
                {
                return triangle_index - 1;
                }
            }

        triangle_index++;
        }

    // Return last strip triangle.

    return triangle_index - 1;
    }


void TerrainBuild::prepare_separate_triangles
    (
    int strip_index,
    int triangle_start,
    int triangle_end
    )
    {
    hope( strip_index >= 0 );
    hope( strip_index < heightmap.y_size - 1 );
    DEBUG_ONLY( const int strip_triangles = 2 * ( heightmap.x_size - 1 ); )
    hope( triangle_start <= triangle_end );
    hope( triangle_start >= 0 );
    hope( triangle_start < strip_triangles );
    hope( triangle_end >= 0 );
    hope( triangle_end < strip_triangles );

    int ti;
    for( ti = triangle_start; ti <= triangle_end; ti++ )
        {
        int strip_vertices[ 3 ];
        strip_triangle_vertices
            (
            strip_vertices[ 0 ],
            strip_vertices[ 1 ],
            strip_vertices[ 2 ],
            ti
            );

        // Map triangle vertices into the array.

        int indices[ 3 ];
        int vi;
        for( vi = 0; vi < 3; vi++ )
            {
            int x, y;
            strip_vertex_coordinates( x, y, strip_index, strip_vertices[ vi ] );
            VertexInfo & vertex = get_vertex_info( x, y );
            if( vertex.array_index == -1 )
                {
                vertex.array_index = array_vertices.size();
                array_vertices.push_back( ArrayVertex( x, y ) );
                }
            hope( vertex.array_index >= 0 );
            hope( vertex.array_index < 65536 );
            indices[ vi ] = vertex.array_index;
            }

        // Remember the triangle.

        triangles.push_back( Triangle( indices[ 0 ], indices[ 1 ], indices[ 2 ] ) );
        }
    }


void TerrainBuild::prepare_triangle_strip
    (
    int strip_index,
    int triangle_start,
    int triangle_end
    )
    {
    hope( strip_index >= 0 );
    hope( strip_index < heightmap.y_size - 1 );
    DEBUG_ONLY( const int strip_triangles = 2 * ( heightmap.x_size - 1 ); )
    hope( triangle_start <= triangle_end );
    hope( triangle_start >= 0 );
    hope( triangle_start < strip_triangles );
    hope( triangle_end >= 0 );
    hope( triangle_end < strip_triangles );

    // First triangle is counter-clockwise if it's "odd", clockwise if it's "even".
    // See #strip_triangle_vertices() for picture.

    triangle_strips.push_back( TriangleStrip
        (
        triangle_end - triangle_start + 3,
        ( triangle_start & 1 ) ? GL_CCW : GL_CW
        ) );
    triangle_strip_count++;

    // Initialize vertex indices.

    TriangleStrip::Indices & indices = triangle_strips.back().indices;
    int vi;
    int i;
    for( vi = triangle_start, i = 0; vi <= triangle_end + 2; vi++, i++ )
        {
        // Map vertex.

        int x, y;
        strip_vertex_coordinates( x, y, strip_index, vi );
        VertexInfo & vertex = get_vertex_info( x, y );
        if( vertex.array_index == -1 )
            {
            vertex.array_index = array_vertices.size();
            array_vertices.push_back( ArrayVertex( x, y ) );
            }
        hope( vertex.array_index >= 0 );
        hope( vertex.array_index < 65536 );
        indices[ i ] = vertex.array_index;
        }
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "terrain_build_inline.h"
#endif
