/** @file terrain_build.h
 * Terrain building.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TERRAIN_BUILD_H
#define CLIENT_TERRAIN_BUILD_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_TERRAIN_H
#include "renderer/terrain.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <vector>
#include <list>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Structure containing all data using during #Terrain::build().
 *
 * This is private Terrain structure, users of Terrain can safely ignore it.
 *
 * @note TerrainBuild object can't be reused to build multiple terrains.
 *       Brand new TerrainBuild object must be used for each build.
 *
 * @bug Patches are just a quick hack to show "marked" triangles. They are
 *      drawn as separate quads which is not optimal and it could cause
 *      visual artifacts on some hw/drivers, OpenGL documentation does not
 *      say how exactly nonplanar quads should be drawn.
 */
class TerrainBuild
    {
    NO_COPY_CONSTRUCTOR( TerrainBuild );
    NO_ASSIGNMENT_OPERATOR( TerrainBuild );

public:

    /**
     * Constructor.
     */
    TerrainBuild
        (
        const Heightmap & the_heightmap
        );

    /**
     * Destructor.
     */
    ~TerrainBuild();

public:

    /**
     * Build material layer.
     *
     * This method will build and return array of render objects that can be used
     * to draw material layer of the terrain. @a Render_object_count will be set
     * to number of generated render objects. Each render object represents
     * material pass using single texture. The first pass will render the
     * whole terrain (some triangles textured, other black) and initialize the
     * depth buffer. The following passes will add more texture layers to the
     * terrain.
     */
    RenderObject * build_material
        (
        std::size_t & render_object_count
        );

    /**
     * Build patch layer.
     *
     * This method will build and return patch layer render object.
     * Currently no real tile patches are supported, some tiles may be "marked"
     * by this layer. Generated render object is not optimal (repeated vertices
     * in the array, etc.).
     *
     * Patches must be built immediately after material layers of the same terrain.
     *
     * @return Owning pointer to generated render object or @c NULL if no
     *         patches have been generated for the terrain.
     */
    RenderObject * build_patches();

private:

    /**
     * Pointer hasher.
     */
    struct PointerHash
        {
        INLINE std::size_t operator()
            (
            const TextureDataObject * ptr
            ) const;
        };

    /**
     * Map from texture data object pointer to number of vertices which
     * use the texture.
     */
    typedef ext::hash_map< TextureDataObject *, int, PointerHash > TextureVertexCount;

private:

    /**
     * Additional heightmap vertex information.
     */
    class VertexInfo
        {
    public:

        /**
         * Constructor.
         */
        INLINE VertexInfo();

        /**
         * Add texture to vertex.
         * There may be up to 2 textures per vertex.
         */
        void add_texture
            (
            TextureDataObject::Reference    texture,
            Float32                         weight
            );

        /**
         * Has @a texture been added to the vertex?
         */
        INLINE bool has_texture
            (
            TextureDataObject * texture
            ) const;

        /**
         * Return weight of given @a texture or 0 if vertex does not
         * contain the texture.
         */
        INLINE Float32 get_texture_weight
            (
            TextureDataObject * texture
            ) const;

    public:

        Float32                         height;
            /**< Copied from heightmap. */

        SVector3                        normal;
            /**< Vertex normal vector. */

        TextureDataObject::Reference    textures[ 2 ];
            /**< Textures. */

        Float32                         weights[ 2 ];
            /**< Texture weights. */

        int                             array_index;
            /**< Vertex array index, -1 if it's not mapped to current array. */

        }; // class VertexInfo


    /**
     * Infomation about each heightmap vertex.
     */
    typedef std::vector< VertexInfo > VertexInfos;

private:

    /**
     * X-Y vertex index.
     */
    class ArrayVertex
        {
    public:

        /**
         * Constructor.
         */
        INLINE ArrayVertex
            (
            int the_x,
            int the_y
            );

    public:

        int x;
            /**< Vertex x coordinate. */

        int y;
            /**< Vertex y coordinate. */

        }; // class ArrayVertex

    /**
     * Vector of x-y vertex indices.
     */
    typedef std::vector< ArrayVertex > ArrayVertices;

private:

    /**
     * Triangle.
     */
    class Triangle
        {
    public:

        /**
         * Constructor.
         */
        INLINE Triangle
            (
            int the_v1,
            int the_v2,
            int the_v3
            );

    public:

        //@{
        /// Vertex array index.
        int v1;
        int v2;
        int v3;
        //@}

        }; // class Triangle


    /**
     * Vector of triangles.
     */
    typedef std::vector< Triangle > Triangles;

private:
    
    /**
     * Triangle strip.
     */
    class TriangleStrip
        {
    public:

        /**
         * Constructor.
         */
        INLINE TriangleStrip
            (
            int     vertex_count,
            GLenum  the_front_face
            );

    public:

        /**
         * Vector of integers - indices into vertex array.
         */
        typedef std::vector< int > Indices;

        Indices     indices;
            /**< Indices into vertex array. */

        GLenum      front_face;
            /**< @c GL_CW if first triangle is odd, @c GL_CCW if it's even. */

        }; // class TriangleStrip


    /**
     * List of triangle strips.
     */
    typedef std::list< TriangleStrip > TriangleStrips;

private:

    /**
     * Gather vertex info.
     *
     * Initializes #used_textures and #vertex_infos.
     */
    void gather_vertex_info();

    /**
     * Get info about vertex at given coordinates.
     */
    INLINE VertexInfo & get_vertex_info
        (
        int x,
        int y
        );

    /**
     * Get info about vertex at given coordinates.
     */
    INLINE const VertexInfo & get_vertex_info
        (
        int x,
        int y
        ) const;

private:

    /**
     * Create jobs for <tt>material_objects[ @a layer_index ]</tt> using @a texture.
     *
     * As a side effect #array_vertices required by #build_material_arrays()
     * will be initialized.
     */
    void build_material_jobs
        (
        int                 layer_index,
        TextureDataObject * texture
        );

    /**
     * Clear temporary structures generated by #build_material_jobs() but
     * not required by #build_material_arrays().
     */
    void clear_material_jobs_temporaries();

    /**
     * Create vertex array from #array_vertices prepared by #build_material_jobs().
     */
    void build_material_arrays
        (
        int                 layer_index,
        TextureDataObject * texture
        );

    /**
     * Clear temporary structures generated by #build_material_jobs() and
     * #build_material_arrays().
     */
    void clear_material_arrays_temporaries();

    /**
     * Initialize material job pipeline state.
     *
     * @param job           Modify this state.
     * @param layer_index   Index of material layer. First layer is drawn using different settings
     *                      (depth of all terrain triangles must be written), consecutive layers
     *                      may optionally use polygon depth offset.
     * @param texture       Texture to use.
     * @param front_face    @c GL_CW or @c GL_CCW.
     */
    void initialize_material_pipeline_state
        (
        PipelineState &     state,
        int                 layer_index,
        TextureDataObject * texture,
        GLenum              front_face
        );

    /**
     * Initialize wireframe job pipeline state.
     * This is a really nasty hack. See sources for more info.
     */
    void initialize_wireframe_pipeline_state
        (
        PipelineState &     state
        );

    /**
     * Initialize fog pipeline state from heightmap settings.
     */
    void initialize_fog_pipeline_state
        (
        PipelineState &     state,
        bool                additive_layer
        );

    /**
     * Initialize material triangle strip geometries and jobs.
     *
     * @param render_object Render object to modify.
     * @param job_index     Index of first job to initialize. Will contain index
     *                      of job following the initialized jobs.
     * @param front_face    Initialize clockwise or counter-clockwise jobs?
     */
    void initialize_tristrip_material_jobs
        (
        RenderObject &      render_object,
        std::size_t &       job_index,
        GLenum              front_face
        );

    /**
     * Initialize wireframe triangle strip jobs.
     *
     * @param render_object     Render object to modify.
     * @param job_index         Index of first job to initialize. Will contain index
     *                          of job following the initialized jobs.
     * @param geometry_index    Index of first geometry to use by this job. Will increase
     *                          along with @a job_index.
     */
    void initialize_tristrip_wireframe_jobs
        (
        RenderObject &      render_object,
        std::size_t &       job_index,
        std::size_t &       geometry_index
        );

private:

    /**
     * Convert strip vertex index to heightmap integral x-y coordinates.
     *
     * @retval x @a x index into heightmap.
     * @retval y @a y index into heightmap.
     * @param strip_index Index of the strip. 0 for southernmost west-east strip.
     * @param vertex_index Index of strip vertex.
     */
    INLINE void strip_vertex_coordinates
        (
        int &   x,
        int &   y,
        int     strip_index,
        int     vertex_index
        ) const;

    /**
     * Get vertex indices of given strip triangle.
     *
     * This method sets @a i1, @a i2 and @a i3 to vertex indices of strip
     * triangle @a triangle_index, in clockwise order.
     *
     * Observe this beautiful image to understand what I mean by horizontal
     * strip vertex and triangle indices:
     * @verbatim

   1         3         5
   +---------+---------+--...
   |\        |\        |\
   |  \   1  |  \   3  |  \
   |    \    |    \    |
   |  0   \  |  2   \  |  4
   |        \|        \| 
   +---------+---------+--...
   0         2         4

@endverbatim
     *
     * As you can see, the set of vertex indices is really simple - @a ti,
     * @a ti + 1 and @a ti + 2 (where @a ti is @a triangle_index). Vertex
     * order is different for odd and event triangles. If order is important,
     * use this method.
     */
    INLINE void strip_triangle_vertices
        (
        int &   i1,
        int &   i2,
        int &   i3,
        int     triangle_index
        ) const;

    /**
     * Get triangle index of the first triangle using @a texture.
     *
     * Search strip @a strip_index, starting at triangle @a triangle_index.
     * Return number of strip triangles (2 * ( x_size - 1 ))  if no triangle
     * with @a texture has been found.
     */
    int strip_start
        (
        int                 strip_index,
        int                 triangle_index,
        TextureDataObject * texture
        ) const;

    /**
     * Find last triangle using @a texture in strip @a strip_index,
     * starting at triangle @a triangle_index.
     */
    int strip_end
        (
        int                 strip_index,
        int                 triangle_index,
        TextureDataObject * texture
        ) const;

    /**
     * Add triangles in given strip to #triangles.
     */
    void prepare_separate_triangles
        (
        int strip_index,
        int triangle_start,
        int triangle_end
        );

    /**
     * Add triangles in given strip to #triangle_strips.
     */
    void prepare_triangle_strip
        (
        int strip_index,
        int triangle_start,
        int triangle_end
        );

private:

    /**
     * Get OpenGL vertex for given heightmap vertex.
     */
    static INLINE OpenGL::Vector3 get_heightmap_opengl_vertex
        (
        int     x,
        int     y,
        Float32 height
        );

// source
private:

    const Heightmap &       heightmap;
        /**< Source heightmap. */

// vertex info set by gather_vertex_info
private:

    TextureVertexCount      used_textures;
        /**< Map of all textures used in the material layer.
             Value of each hash element is number of vertices used by the texture. */

    VertexInfos             vertex_infos;
        /**< Information about each vertex. */

// temp info for single material layer
private:

    ArrayVertices           array_vertices;
        /**< Vertices to put into vertex array. */

    Triangles               triangles;
        /**< Vector of separate triangles. */

    TriangleStrips          triangle_strips;
        /**< List of triangle strips. */

    int                     triangle_strip_count;
        /**< Number of triangle strips. */

// result
private:

    RenderObject *          material_objects;
        /**< Temporary array of material layer render objects. */

    RenderObject *          patch_object;
        /**< Temporary patch layer render object. */

    }; // class TerrainBuild


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "terrain_build_inline.h"
#endif


#endif // CLIENT_TERRAIN_BUILD_H
