/** @file terrain_material.cpp
 * Terrain material description and material database (implementation).
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/terrain_material.h"
#include "global/global.h"
#include "data/data_manager.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include <sstream>
#include <algorithm>


namespace Demo {
namespace Client {
namespace Renderer {


// class TerrainMaterial

void TerrainMaterial::text_read
    (
    Massiv::Core::TextReader & tr
    )
    {
    // Tag.

    tr.next_line_throw();
    tr.require_string( "material" );
    tr.require_string( "1" );

    // Equation.

    tr.next_line_throw();
    equation.height_factor = tr.read_float();
    equation.slope_factor = tr.read_float();
    equation.delta = tr.read_float();

    // Number of gradient elements.

    tr.next_line_throw();
    const std::size_t element_count = tr.read_int();
    texture_gradient.resize( element_count );

    // Gradient.

    TextureManager & texture_manager = Global::get_texture_manager();
    Float64 last_index;
    std::size_t i;
    for( i = 0; i < element_count; i++ )
        {
        tr.next_line_throw();
        const Float32 index = tr.read_float();
        const std::string texture_name = tr.read_quote_string();

        // Check index.

        if( ( i > 0 ) && ( index < last_index ) )
            {
            std::ostringstream ss;
            ss << "number greater than or equal to " << last_index;
            throw Massiv::Core::ExUnexpectedValue( tr.unexpected_value( ss.str(), index ) );
            }

        texture_gradient[ i ] = GradientElement
            (
            index,
            texture_manager.get_texture_data_object( texture_name )
            );

        last_index = index;
        }
    }


namespace {

/**
 * "Less" functor for GradientElement::index.
 */
struct GradientIndexLess
    {
    // Note: Standard does not state which objects will be compared
    // in std::lower_bound (usually it's only ( *it, value )) so you
    // should create fake GradientElement for the value to compare...
    inline bool operator()
        (
        const TerrainMaterial::GradientElement & first,
        const TerrainMaterial::GradientElement & second
        ) const
        {
        return first.index < second.index;
        }

    }; // struct GradientIndexLess

} // anonymous namespace


void TerrainMaterial::get_textures
    (
    TextureDataObject::Reference &  texture1,
    TextureDataObject::Reference &  texture2,
    Float32 &                       weight,
    Float32                         height,
    Float32                         slope
    ) const
    {
    if( texture_gradient.empty() )
        {
        texture1.set_null();
        texture2.set_null();
        weight = 1.0f;
        return;
        }

    // Compute index.

    const Float32 index = height * equation.height_factor +
        slope * equation.slope_factor + equation.delta;

    // Find first element with greater or equal index.

    TextureGradient::const_iterator it = std::lower_bound
        (
        texture_gradient.begin(),
        texture_gradient.end(),
        GradientElement( index, NULL ),
        GradientIndexLess()
        );

    // Last texture?

    if( it == texture_gradient.end() )
        {
        it--; // trick, it will satify ( it->index <= index ) below.
        }

    // First or last texture, or exact match.

    if( ( it == texture_gradient.begin() ) || ( it->index <= index ) )
        {
        texture1 = it->texture;
        texture2.set_null();
        weight = 1.0f;
        }
    else
        {
        // Two textures.
        
        hope( it != texture_gradient.begin() );
        const TextureGradient::const_iterator prev_it = it - 1;
        texture1 = prev_it->texture;
        texture2 = it->texture;
        
        const Float32 low_index = prev_it->index;
        const Float32 high_index = it->index;
        hope( high_index > index );
        hope( low_index < index );
        weight = Massiv::Core::clamp
            (
            ( high_index - index ) / ( high_index - low_index ),
            0.0f,
            1.0f
            );
        }
    }


// class TerrainMaterialDatabase::MaterialInfo

TerrainMaterialDatabase::MaterialInfo::MaterialInfo
    (
    const std::string & the_name,
    Int32               the_id,
    const std::string & data_name
    ) : name( the_name ),
        id( the_id )
    {
    // Get the data object.

    Massiv::Core::DataManager & data_manager = Massiv::Core::Global::data_manager();
    data_object.reset( data_manager.get_data_object( data_name ) );
    hope( data_object.get() );

    // Do not read the material yet, wait until first call to #get_material()...
    }


void TerrainMaterialDatabase::MaterialInfo::refresh
    (
    bool force
    )
    {
    hope( data_object.get() );
    
    if( force || data_object->is_data_updated() ||
        material.get_texture_gradient().empty() )
        {
        data_object->refresh();
        data_object->open();

        try
            {
            Massiv::Core::TextReader tr( data_object->data );
            material.text_read( tr );
            }
        catch( ... )
            {
            data_object->close();
            throw;
            }

        data_object->close();
        update_time = SystemTime::current_time();
        }

    // Start new material download if available.

    data_object->download_actual_version();
    }


// class TerrainMaterialDatabase

TerrainMaterialDatabase::TerrainMaterialDatabase
    (
    const std::string & data_name
    ) : id_to_material( MAX_MATERIAL_ID + 1 ),
        default_id( -1 )
    {
    // Clear id_to_material.

    int i;
    for( i = 0; i < MAX_MATERIAL_ID + 1; i++ )
        {
        id_to_material[ i ] = NULL;
        }
    
    // Get the data object.

    Massiv::Core::DataManager & data_manager = Massiv::Core::Global::data_manager();
    data_object.reset( data_manager.get_data_object( data_name ) );
    hope( data_object.get() );

    // Load the database.

    refresh( true );
    }


TerrainMaterialDatabase::~TerrainMaterialDatabase()
    {
    clear();
    }


const TerrainMaterialDatabase::MaterialInfo & TerrainMaterialDatabase::get_material
    (
    const std::string & name
    )
    {
    // Refresh the database, including all materials.
    
    refresh();

    // Find material.

    const NameToMaterial::iterator name_it = name_to_material.find( name );
    MaterialInfo * info = NULL;
    if( name_it == name_to_material.end() )
        {
        hope( default_id >= 0 );
        hope( default_id <= MAX_MATERIAL_ID );
        info = id_to_material[ default_id ];
        }
    else
        {
        info = name_it->second;
        }

    // Done.

    hope( info );
    return *info;
    }


const TerrainMaterialDatabase::MaterialInfo & TerrainMaterialDatabase::get_material
    (
    Int32 id
    )
    {
    hope( id >= 0 );
    hope( id <= MAX_MATERIAL_ID );

    // Refresh the database, including all materials

    refresh();

    // Find material.

    MaterialInfo * info = id_to_material[ id ];
    if( !info )
        {
        hope( default_id >= 0 );
        hope( default_id <= MAX_MATERIAL_ID );
        info = id_to_material[ default_id ];
        }

    // Done.

    hope( info );
    return *info;
    }


void TerrainMaterialDatabase::refresh
    (
    bool force
    )
    {
    // Check if the whole database should be reloaded.

    hope( data_object.get() );
    if( force || data_object->is_data_updated() )
        {
        // Reload the database.

        data_object->refresh();
        data_object->open();

        try
            {
            Massiv::Core::TextReader tr( data_object->data );
            text_read( tr );
            }
        catch( ... )
            {
            data_object->close();
            throw;
            }

        data_object->close();
        update_time = SystemTime::current_time();
        }
    else
        {
        // Refresh all materials.

        NameToMaterial::iterator it;
        for( it = name_to_material.begin(); it != name_to_material.end(); ++it )
            {
            MaterialInfo * const info = it->second;
            hope( info );
            info->refresh();
            if( info->get_update_time() > update_time )
                {
                update_time = info->get_update_time();
                }
            }
        }

    // Start new download if available.

    data_object->download_actual_version();
    }


void TerrainMaterialDatabase::text_read
    (
    Massiv::Core::TextReader & tr
    )
    {
    clear();

    // Tag.

    tr.next_line_throw();
    tr.require_string( "material_database" );
    tr.require_string( "1" );

    // Materials...

    while( tr.next_line() )
        {
        const std::string name = tr.read_quote_string();
        if( name_to_material.find( name ) != name_to_material.end() )
            {
            throw Massiv::Core::ExUnexpectedValue
                ( tr.unexpected_value( "unique material name", name ) );
            }

        const Int32 id = tr.read_int();
        if( ( id < 0 ) || ( id > MAX_MATERIAL_ID ) )
            {
            #define EXPECTED( max ) "integer between 0 and " #max
            throw Massiv::Core::ExUnexpectedValue
                ( tr.unexpected_value( EXPECTED( MAX_MATERIAL_ID ), id ) );
            #undef EXPECTED
            }
        if( id_to_material[ id ] )
            {
            throw Massiv::Core::ExUnexpectedValue
                ( tr.unexpected_value( "unique material id", id ) );
            }

        const std::string data_name = tr.read_quote_string();

        add_material( name, id, data_name );
        }

    // Check default material.

    if( default_id == -1 )
        {
        // Warning: This uses fileid in non-standard way.

        throw Massiv::Core::ExUnexpectedEof( ". Material \"default\" not defined" );
        }   
    }


void TerrainMaterialDatabase::clear()
    {
    while( !name_to_material.empty() )
        {
        const Int32 id = name_to_material.begin()->second->id;
        remove_material( id );
        }

    hope( name_to_material.empty() );
    #ifdef OPTION_DEBUG
    std::size_t i;
    for( i = 0; i < MAX_MATERIAL_ID; i++ )
        {
        hope( id_to_material[ i ] == NULL );
        }
    #endif
    hope( default_id == -1 );
    }


void TerrainMaterialDatabase::add_material
    (
    const std::string & name,
    Int32               id,
    const std::string & data_name
    )
    {
    hope( id >= 0 );
    hope( id <= MAX_MATERIAL_ID );
    hope( name_to_material.find( name ) == name_to_material.end() );
    hope( id_to_material[ id ] == NULL );

    // Create material info.

    std::auto_ptr< MaterialInfo > info( new MaterialInfo( name, id, data_name ) );

    // Add hash entry. Info will be deleted if it throws.

    name_to_material.insert( NameToMaterial::value_type( name, info.get() ) );

    // Add vector entry. Never throws.
    
    id_to_material[ id ] = info.get();

    // The info is now owned by the database...

    info.release();

    // Default id?

    if( name == "default" )
        {
        default_id = id;
        }
    }


void TerrainMaterialDatabase::remove_material
    (
    Int32               id
    )
    {
    hope( id >= 0 );
    hope( id <= MAX_MATERIAL_ID );
    const MaterialInfo * const info = id_to_material[ id ];
    hope( info );

    // Remove hash entry, reset vector entry.

    name_to_material.erase( info->name );
    id_to_material[ id ] = NULL;

    // Default id?

    if( id == default_id )
        {
        hope( info->name == "default" );
        default_id = -1;
        }

    // Destroy the info.

    delete info;
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "terrain_material_inline.h"
#endif
