/** @file terrain_material.h
 * Terrain material description and material database.
 */

/*
 * begin        : Jul 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TERRAIN_MATERIAL_H
#define CLIENT_TERRAIN_MATERIAL_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_TEXTURE_DATA_OBJECT_H
#include "renderer/texture_data_object.h"
#endif

#ifndef MASSIV_TEXT_READER_H
#include "utility/text_reader.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif

#include <vector>
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Heighmap terrain material.
 *
 * Terrain material is used to determine which textures should be used in the
 * material layer of #Terrain.
 *
 * Two structures describe the material. @a Vertex @a equation is used to
 * compute floating-point "index" from vertex height and terrain slope.
 * The index is used to determine which textures to use at the vertex.
 * @a Texture @a gradient is used to fetch this info.
 */
class TerrainMaterial
    {
public:

    /**
     * Equation coefficients.
     *
     * This structure contains coeffiecients of vertex equation. It's used to
     * compute value used to fetch texture from texture gradient:
     * @code
     * gradient_index = vertex_height * height_factor +
     *                  terrain_slope * slope_factor +
     *                  delta
     * @endcode
     * , where @a vertex_height is vertex @a z coordinate transformed to
     * \< 0.0, 1.0 \> range and @a terrain_slope is number between 0.0 (level
     * terrain, floor) and 1.0 (upright terrain, wall).
     *
     * @todo More powerful equation specification, any expression?
     */
    class VertexEquation
        {
    public:

        /** Set all coefficients to zero. */
        INLINE VertexEquation();

    public:

        //@{
        /// Equation coefficient.
        Float32     height_factor;
        Float32     slope_factor;
        Float32     delta;
        //@}

        }; // class VertexEquation.


    /**
     * Element of texture gradient.
     */
    class GradientElement
        {
    public:

        /** Default constructor. */
        INLINE GradientElement();

        /** Constructor. */
        INLINE GradientElement
            (
            Float32                         the_index,
            TextureDataObject::Reference    the_texture
            );

    public:
    
        Float32                         index;
            /**< "Index" or value. */

        TextureDataObject::Reference    texture;
            /**< Texture to return when fetching texture for @a index. */

        }; // struct GradientElement


    /**
     * Texture gradient.
     *
     * Vector of GradientElement, sorted by GradientElement::index.
     */
    typedef std::vector< GradientElement > TextureGradient;

public:

    /**
     * Construct empty material.
     */
    INLINE TerrainMaterial();

    /**
     * Copy constructor.
     */
    INLINE TerrainMaterial
        (
        const TerrainMaterial & other
        );

    /**
     * Assignment operator.
     */
    INLINE TerrainMaterial & operator=
        (
        const TerrainMaterial & other
        );

public:

    /**
     * Get equation coefficients.
     */
    INLINE const VertexEquation & get_vertex_equation() const;

    /**
     * Set equation coefficients.
     */
    INLINE void set_vertex_equation
        (
        const VertexEquation & new_equation
        );

public:

    /**
     * Get texture gradient.
     */
    INLINE const TextureGradient & get_texture_gradient() const;

    /**
     * Clear texture gradient.
     */
    INLINE void clear_texture_gradient();

    /**
     * Add @a element to texture gradient.
     *
     * Index of the @a element must be greater than or equal to current last
     * gradient element.
     */
    INLINE void add_texture_gradient_element
        (
        const GradientElement & element
        );

public:

    /**
     * Read material description from stream.
     *
     * See doc/formats/terrain_material.txt.
     */
    void text_read
        (
        Massiv::Core::TextReader & tr
        );

public:

    /**
     * Get textures from texture gradient using given vertex @a height
     * and terrain @a slope.
     *
     * The two textures should be combined using @a weight:
     * <tt>weight * texture1 + ( 1 - weight ) * texture2</tt>.
     * @a Weight will be always in range ( 0.0, 1.0 >. If @a weight is
     * 1.0, @a texture2 will be set to @c NULL. @a Texture1 will never be
     * NULL unless texture gradient is empty. @a Texture1 and @a texture2
     * will never point to the same texture data object.
     */
    void get_textures
        (
        TextureDataObject::Reference &  texture1,
        TextureDataObject::Reference &  texture2,
        Float32 &                       weight,
        Float32                         height,
        Float32                         slope
        ) const;

private:

    VertexEquation      equation;
        /**< Vertex equation. */

    TextureGradient     texture_gradient;
        /**< Texture gradient. Empty if material has not been read yet. */

    }; // class TerrainMaterial

    
/**
 * Material id hasher.
 */
struct TerrainMaterialIdHash
    {
    inline std::size_t operator()
        (
        Int32 id
        ) const
        {
        return ext::hash_int( id );
        }
    };


/**
 * Database of terrain materials.
 *
 * This database keeps list of all known materials, their ids and names.
 * See doc/formats/terrain_material_database.txt for description of
 * data object the database is read from.
 *
 * @todo Similar class should be implemented on the server too. It should read
 *       the same database, but it should only remember the ids and names,
 *       not the materials themselves. It could be implemented as base class of
 *       TerrainMaterialDatabase. Add directory src/demo/shared?
 */
class TerrainMaterialDatabase
    {
    NO_COPY_CONSTRUCTOR( TerrainMaterialDatabase );
    NO_ASSIGNMENT_OPERATOR( TerrainMaterialDatabase );

public:

    enum
        {
        MAX_MATERIAL_ID = 255
            /**< Maximal material id value. */
        };

public:

    /**
     * Information about material.
     */
    class MaterialInfo
        {
        NO_COPY_CONSTRUCTOR( MaterialInfo );
        NO_ASSIGNMENT_OPERATOR( MaterialInfo );

        // The database owns the info and can modify it.
        friend class TerrainMaterialDatabase;
        
    protected:

        /**
         * Constructor.
         *
         * #refresh() will not be called from constructor, #material of
         * constructed #MaterialInfo will be empty. Do not forget to call
         * #refresh() before accessing #material.
         */
        MaterialInfo
            (
            const std::string & the_name,
            Int32               the_id,
            const std::string & data_name
            );

        /**
         * Refresh #material from #data_object.
         *
         * Called automatically by #TerrainMaterialDatabase::get_material().
         */
        void refresh
            (
            bool force = false
            );

        /**
         * Get time of most recent data change.
         */
        INLINE const SystemTime & get_update_time() const;

    public:

        std::string         name;
            /**< Material name. */

        Int32               id;
            /**< Material id. */

        TerrainMaterial     material;
            /**< The material. */

    protected:
    
        std::auto_ptr< Massiv::Core::DataObject > data_object;
            /**< Material data object. */

        SystemTime          update_time;
            /**< Time of the most recent materialdata update. */

        }; // class MaterialInfo

public:

    /**
     * Constructor.
     *
     * @param name Database data object name.
     */
    TerrainMaterialDatabase
        (
        const std::string & data_name
        );

    /**
     * Destructor.
     */
    ~TerrainMaterialDatabase();

public:

    /**
     * Find material with given @a name.
     *
     * Returned reference will be valid until #refresh() is called.
     * If no material with @a name exists, reference to default material
     * will be returned.
     *
     * This method automatically checks if material database or material
     * definition has been updated.
     */
    const MaterialInfo & get_material
        (
        const std::string & name
        );

    /**
     * Find material with given @a id.
     *
     * Returned reference will be valid until #refresh() is called.
     * If no material with id @a id exists, reference to default material
     * will be returned.
     *
     * This method automatically checks if material database or material
     * definition has been updated.
     */
    const MaterialInfo & get_material
        (
        Int32 id
        );
    
    /**
     * Get time of the most recent update of database or material data.
     */
    INLINE const SystemTime & get_update_time() const;

    /**
     * Reload the database if #data_object has been updated.
     *
     * This will refresh all materials even if the database itself has
     * not been changed.
     */
    void refresh
        (
        bool force = false
        );

private:

    /**
     * Read material database from stream.
     */
    void text_read
        (
        Massiv::Core::TextReader & tr
        );

    /**
     * Clear material database.
     */
    void clear();

    /**
     * Add material to the database.
     *
     * Material must have unique id and name. If name is "default", #default_id
     * will be set to its id.
     */
    void add_material
        (
        const std::string & name,
        Int32               id,
        const std::string & data_name
        );

    /**
     * Remove material from the database.
     *
     * It its name is "default", #default_id will be set to -1.
     */
    void remove_material
        (
        Int32               id
        );

private:

    /**
     * Map string-\>MaterialInfo.
     */
    typedef ext::hash_map< std::string, MaterialInfo * > NameToMaterial;

    /**
     * Vector of pointers to MaterialInfo.
     */
    typedef std::vector< MaterialInfo * > MaterialPointers;

private:

    // Note: Materials are owned by the database. Use #add_material() and
    //       #remove_material() to modify the following hashes.

    NameToMaterial      name_to_material;
        /**< Map material name -> material info. */

    MaterialPointers    id_to_material;
        /**< Materials indexed by material id. */

    Int32               default_id;
        /**< Id of default material. */

private:

    std::auto_ptr< Massiv::Core::DataObject >   data_object;
        /**< Database data object. */
    
    SystemTime                                  update_time;
        /**< Time of the most recent database data or material data update. */

    }; // class TerrainMaterialDatabase


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "terrain_material_inline.h"
#endif


#endif // CLIENT_TERRAIN_MATERIAL_H
