/** @file text_rectangle.cpp
 * Overlay object - text rectangle (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/text_rectangle.h"
#include "renderer/renderer.h"
#include "renderer/pipeline_state.h"
#include "renderer/font.h"


namespace Demo {
namespace Client {
namespace Renderer {


TextRectangle::TextRectangle
    (
    Font *  the_font,
    int     the_columns,
    int     the_rows
    ) : font( the_font ),
        font_version( 0 ),
        rows( the_rows ),
        columns( the_columns ),
        font_color( 1.0f, 1.0f, 1.0f, 1.0f ),
        cleared( true ),
        dirty( true )
    {
    if( font )
        {
        set_size_from_font();
        }
    }


TextRectangle::~TextRectangle()
    {
    }


void TextRectangle::set_size_from_font()
    {
    hope( font );
    font->get_texture();
    set_base_size( SVector2
        (
        static_cast< Float32 >( font->get_character_width() * columns ),
        static_cast< Float32 >( font->get_character_height() * rows )
        ) );
    }


void TextRectangle::set_columns_rows_from_font()
    {
    hope( font );
    font->get_texture();
    hope( font->get_character_width() > 0 );
    hope( font->get_character_height() > 0 );

    // Use base size, not current size. We don't want to be affected by animations.

    const int columns = static_cast< int >( Massiv::Core::FloatTraits< Float32 >::floor
        ( get_base_size().x() / font->get_character_width() ) );
    const int rows = static_cast< int >( Massiv::Core::FloatTraits< Float32 >::floor
        ( get_base_size().y() / font->get_character_height() ) );
    set_columns_rows
        (
        columns > 0 ? columns : 1,
        rows > 0 ? rows : 1
        );
    }


void TextRectangle::print
    (
    int                 x,
    int                 y,
    const std::string & text
    )
    {
    hope( rows > 0 );
    hope( columns > 0 );

    if( cleared )
        {
        allocate_cells();
        }

    if( ( y >= 0 ) && ( y < rows ) )
        {
        std::string::const_iterator it;
        for( it = text.begin(); ( it != text.end() ) && ( x < columns ); ++it, x++ )
            {
            if( x < 0 )
                {
                continue;
                }

            // Update the cell.

            Cell & cell = cells[ y * columns + x ];
            cell.character = *it;
            cell.color = font_color;
            }
        }

    dirty = true;
    }


void TextRectangle::print_text
    (
    const std::string & text,
    Alignment           alignment
    )
    {
    // Count lines and find longest row.

    std::vector< std::string > lines;
    std::size_t max_line_length = 1;
    std::string::size_type old_pos = 0;
    while( old_pos < text.size() )
        {
        const std::string::size_type pos = text.find_first_of( "\n", old_pos );
        const std::string::size_type end = ( pos == std::string::npos ? text.size() : pos );
        const std::size_t length = end - old_pos;

        // Ignore line if it's the last one and it's empty.

        if( pos == std::string::npos && length == 0 )
            {
            break;
            }

        // Remember another line.

        lines.push_back( std::string( text, old_pos, length ) );
        if( length > max_line_length )
            {
            max_line_length = length;
            }

        // Next line.

        if( pos == std::string::npos )
            {
            break;
            }
        old_pos = pos + 1;
        }

    set_columns_rows( max_line_length, lines.size() ? lines.size() : 1 );
    set_size_from_font();

    // Print the text.

    std::size_t y;
    for( y = 0; y < lines.size(); y++ )
        {
        const std::string & line = lines[ y ];
        int x;
        if( alignment == ALIGN_LEFT )
            {
            x = 0;
            }
        else if( alignment == ALIGN_RIGHT )
            {
            x = ( max_line_length - line.size() );
            }
        else
            {
            // Unfortunately we can't print to half cells...
            hope( alignment == ALIGN_CENTER );
            x = ( max_line_length - line.size() ) / 2;
            }
        print( x, y, line );
        }    
    }


void TextRectangle::render()
    {
    hope( font );
    hope( rows > 0 );
    hope( columns > 0 );

    // Rebuild array if rectangle contents has changed.

    font->get_texture(); // update version
    if( font_version != font->get_version() )
        {
        font_version = font->get_version();
        dirty = true;
        }

    if( dirty )
        {
        rebuild_arrays();
        }

    // Update pipeline states if global color has changed.

    if( color_dirty )
        {
        hope( render_object.jobs.size() == job_colors.size() );
        
        std::size_t i;
        for( i = 0; i < job_colors.size(); i++ )
            {
            hope( render_object.jobs[ i ].pipeline_state );
            render_object.jobs[ i ].pipeline_state->color =
                job_colors[ i ].scolor() * get_current_color();
            }
        color_dirty = false;
        }

    // In current wacky hack implementation array include colors, so just draw them.

    if( !render_object.jobs.empty() )
        {
        // Set scale and tranlation.

        glPushMatrix();
        const OpenGL::Vector2 translation( screenspace_translation() );
        glTranslatef( translation.x(), translation.y(), 0.0f );
        glScalef( get_cell_width(), get_cell_height(), 1.0f );

        // Draw.

        render_object.render();

        // Revert tranformation.

        glPopMatrix();
        }
    }


void TextRectangle::allocate_cells()
    {
    hope( rows > 0 );
    hope( columns > 0 );

    cells.clear();
    cells.resize( rows * columns );
    cleared = false;
    }


void TextRectangle::rebuild_arrays()
    {
    hope( font );
    hope( rows > 0 );
    hope( columns > 0 );

    if( cleared )
        {
        allocate_cells();
        }

    // Clear render object.

    render_object.pipeline_states.clear();
    render_object.geometries.clear();
    render_object.jobs.clear();
    job_colors.clear();
    render_object.array.free();

    // Count elements.

    int element_count = 0;
    Cells::const_iterator it;
    for( it = cells.begin(); it != cells.end(); ++it )
        {
        const char c = it->character;
        if( font->is_valid( c ) )
            {
            element_count += 4;
            }
        }

    // Allocate arrays.

    if( element_count == 0 )
        {
        return;
        }

    render_object.array.set_stream_count( VertexArray::TYPE_VECTOR2, 2 );
    render_object.array.allocate( element_count );
    render_object.array.select( VertexArray::VERTEX_ARRAY, VertexArray::TYPE_VECTOR2, 0 );
    render_object.array.select( VertexArray::TEXCOORD_ARRAY_0, VertexArray::TYPE_VECTOR2, 1 );

    // Initialize arrays.

    VertexArray::Vector2Stream vertices;
    VertexArray::Vector2Stream tex_coords;
    render_object.array.lock_vectors2( vertices, 0 );
    render_object.array.lock_vectors2( tex_coords, 1 );
    int element = 0;
    OpenGL::Color last_color;
    bool last_color_valid = false;

    while( element < element_count )
        {
        int y;
        
        // Find next color to build job for.

        OpenGL::Color min_color;
        bool min_color_valid = false;
        it = cells.begin();
        for( y = 0; y < rows; y++ )
            {
            int x;
            for( x = 0; x < columns; x++, ++it )
                {
                if( font->is_valid( it->character ) )
                    {
                    if( ( !last_color_valid || ( last_color < it->color ) ) &&
                        ( !min_color_valid || ( it->color < min_color ) ) )
                        {
                        min_color = it->color;
                        min_color_valid = true;
                        }
                    }
                }
            }

        hope( min_color_valid );
        hope( it == cells.end() );

        // Build range for this color.

        int range_first_element = element;
        int range_element_count = 0;

        it = cells.begin();
        for( y = 0; y < rows; y++ )
            {
            int x;
            for( x = 0; x < columns; x++, ++it )
                {
                if( it->color != min_color )
                    {
                    continue;
                    }
                const char c = it->character;
                if( font->is_valid( c ) )
                    {
                    // Initialize character quad.

                    const GLfloat x0 = static_cast< GLfloat >( x );
                    const GLfloat y0 = static_cast< GLfloat >( y );
                    const GLfloat x1 = static_cast< GLfloat >( x + 1 );
                    const GLfloat y1 = static_cast< GLfloat >( y + 1 );
                    GLfloat u0, v0, u1, v1;
                    font->get_uv( c, u0, v0, u1, v1 );

                    vertices[ element ]     = OpenGL::Vector2( x0, y0 );
                    tex_coords[ element ]   = OpenGL::Vector2( u0, v0 );
                    element++;
                    vertices[ element ]     = OpenGL::Vector2( x0, y1 );
                    tex_coords[ element ]   = OpenGL::Vector2( u0, v1 );
                    element++;
                    vertices[ element ]     = OpenGL::Vector2( x1, y1 );
                    tex_coords[ element ]   = OpenGL::Vector2( u1, v1 );
                    element++;
                    vertices[ element ]     = OpenGL::Vector2( x1, y0 );
                    tex_coords[ element ]   = OpenGL::Vector2( u1, v0 );
                    element++;

                    range_element_count += 4;
                    }
                }
            }

        // Initialize geometry for this range.

        hope( range_element_count > 0 );
        hope( range_first_element + range_element_count == element );
        
        render_object.geometries.push_back( RenderObject::Geometry() );
        RenderObject::Geometry & geometry = render_object.geometries.back();
        geometry.initialize_range( GL_QUADS, range_first_element, range_element_count );

        job_colors.push_back( min_color );

        // Try next color.

        last_color = min_color;
        last_color_valid = true;
        }

    // Arrays done.

    hope( element == element_count );
    render_object.array.unlock_vectors2( 0 );
    render_object.array.unlock_vectors2( 1 );   

    // Initialize jobs.

    const std::size_t job_count = job_colors.size();
    hope( render_object.geometries.size() == job_count );
    render_object.pipeline_states.resize( job_count );
    render_object.jobs.resize( job_count );
    std::size_t i;
    for( i = 0; i < job_count; i++ )
        {
        // Initialize pipeline state.

        PipelineState & state = render_object.pipeline_states[ i ];
        state = get_pipeline_state();
        state.texture[ 0 ].data = font->get_texture();
        state.texture[ 0 ].flags = font->get_texture_flags();
        state.color = job_colors[ i ].scolor() * get_current_color();

        // Initialize job.

        render_object.jobs[ i ].pipeline_state = &state;
        render_object.jobs[ i ].geometry = &render_object.geometries[ i ];
        }

    // All done.

    dirty = false;
    color_dirty = false;
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "text_rectangle_inline.h"
#endif
