/** @file text_rectangle.h
 * Overlay object - text rectangle.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TEXT_RECTANGLE_H
#define CLIENT_TEXT_RECTANGLE_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OVERLAY_OBJECT_H
#include "renderer/overlay_object.h"
#endif

#ifndef CLIENT_RENDER_OBJECT_H
#include "renderer/render_object.h"
#endif

#include <vector>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Rectangle divided into evenly-sized cells containing specified text.
 *
 * Use #OverlayObject sizing functions to set size of the rectangle in
 * pixels and constructor/#set_rectangle_size() to set number of rows and
 * columns in the rectangle. Cell size will be determined from these sizes
 * automatically. Note that the text may look quite ugly if cell sizes are
 * not integral.
 *
 * @todo Document prerequisities for #print() and #render().
 *
 * @todo Allow zero number of columns / rows.
 */
class TextRectangle : public OverlayObject
    {
public:

    /**
     * Text alignment used by #print_text().
     */
    enum Alignment
        {
        ALIGN_LEFT      = 0,    /**< Align to the left. */
        ALIGN_CENTER    = 1,    /**< Center. */
        ALIGN_RIGHT     = 2,    /**< Align to the right. */
        };

public:

    /**
     * Constructor.
     *
     * If @a font is set, it will be used to determine cell sizes and overlay
     * object size will be set.
     */
    TextRectangle
        (
        Font *  the_font = NULL,
        int     the_columns = 1,
        int     the_rows = 1
        );

    /**
     * Destructor.
     */
    virtual ~TextRectangle();

public:

    //@{
    /** Getter. */
    INLINE Font * get_font() const;

    INLINE int get_columns() const;

    INLINE int get_rows() const;

    INLINE const OpenGL::Color & get_text_color() const;
    //@}

    //@{
    /**
     * Compute current cell size.
     *
     * This value is determined from current overlay object size (affected
     * by deltas set by animations) and number of columnts or rows.
     */
    INLINE Float32 get_cell_width() const;

    INLINE Float32 get_cell_height() const;
    //@}

public:

    /**
     * Set font to use.
     *
     * You must ensure the @a new_font will exist until another font is
     * set or the #TextRectangle is destroyed.
     */
    INLINE void set_font
        (
        Font * new_font
        );

    /**
     * Set base rectangle pixel size from font information.
     *
     * This method gets cell width and height from current font and sets
     * rectangle pixel width to <tt>cell width * columns</tt> and
     * pixel height to <tt>cell height * rows</tt>.
     *
     * @pre Valid font is set.
     */
    void set_size_from_font();

    /**
     * Set number of columns and rows.
     *
     * This method will clear text. Note that pixel size of the rectangle will
     * NOT be changed. Call #set_base_size() or #set_size_from_font().
     */
    INLINE void set_columns_rows
        (
        int new_columns,
        int new_rows
        );

    /**
     * Set number of columns and rows from font information.
     *
     * This method gets cell width and height from current font and sets
     * number of columns to <tt>pixel width / cell width</tt> and
     * number of rows to <tt>pixel height / cell height</tt>, where
     * <tt>pixel width</tt> and <tt>pixel height</tt> are base overlay
     * object sizes (not affected by deltas set by animation).
     *
     * @pre Valid font is set.
     */
    void set_columns_rows_from_font();

public:

    /**
     * Set text color.
     * Call this method to set color which will be used be following calls
     * to #print() and #print_line(). Call #OverlayObject::set_color() to
     * change global text rectangle color. These colors will be multiplied
     * when the text is rendered.
     *
     * Default color is opaque white.
     */
    INLINE void set_text_color
        (
        const SColor & new_color
        );

    /**
     * Print text using current color.
     *
     * Text will be clipped to the rectangle - it will never span
     * multiple lines.
     */
    void print
        (
        int                 x,
        int                 y,                  
        const std::string & text
        );

    /**
     * Create rectangle with single line and string length columns and print the text into it.
     *
     * #set_size_from_font() will be called automatically.
     */
    INLINE void print_line
        (
        const std::string & text
        );

    /**
     * Create text rectangle from @a text.
     *
     * Text may contain new-line characters. Rectangle will be resized to fit the text and
     * #set_size_from_font() will be called. Last new-line character at end of string will
     * be ignored (it will not begin new empty row).
     */
    void print_text
        (
        const std::string & text,
        Alignment           alignment = ALIGN_LEFT
        );

    /**
     * Clear all the text.
     */
    INLINE void clear();

public:

    /**
     * Render the text.
     */
    virtual void render();

private:

    /**
     * Reallocate cells.
     */
    void allocate_cells();

    /**
     * Rebuild vertex arrays.
     */
    void rebuild_arrays();

private:

    Font *          font;
        /**< Font to use when rendering the text. */

    int             font_version;
        /**< Implies #dirty if it does not match font->get_version(). */

    int             rows;
        /**< Number of rows in the rectangle. */

    int             columns;
        /**< Number of columns in the rectangle. */

    OpenGL::Color   font_color;
        /**< Color to use. */

    bool            cleared;
        /**< Reallocate #cells before accessing them? */

    bool            dirty;
        /**< Rebuild arrays before render?
             This must be set whenever text, font or number of rows/columns */

private:

    /**
     * Single cell.
     */
    struct Cell
        {
        /** Constructor. */
        INLINE Cell();

        char            character;
            /**< Character to print. */

        OpenGL::Color   color;
            /**< Color to use. */

        }; // struct Cell

    /**
     * Vector of Cell.
     */
    typedef std::vector< Cell > Cells;

    /**
     * Vector of OpenGL::Color.
     */
    typedef std::vector< OpenGL::Color > Colors;

private:

    Cells               cells;
        /**< #rows * #columns array of #Cell. */

    RenderObject        render_object;
        /**< Object encapsulating vertex array and render jobs. */

    Colors              job_colors;
        /**< Colors of the jobs, not multiplied by global color yet. */

    }; // class TextRectangle


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "text_rectangle_inline.h"
#endif


#endif // CLIENT_TEXT_RECTANGLE_H
