/** @file texture.cpp
 * Simple OpenGL texture (implementation).
 *
 * @todo Use Array< DynamicStorage > in scale() and extend()?
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/texture.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include "client_exception.h"
#include <sstream>
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


namespace {


/**
 * Compute nearest power of 2.
 */
int nearest_power_of_2
    (
    int value
    )
    {
    hope( value > 0 );
    hope( value <= static_cast< int >( ~0u >> 2 ) );

    unsigned n = value;
    int bit_count;
    for( bit_count = -1; n; bit_count++ )
        {
        n >>= 1;
        }

    const int s = 1 << bit_count;
    const int b = 1 << ( bit_count + 1 );
    if( value <= ( s + ( b - s ) / 2 ) )
        {
        return s;
        }
    else
        {
        return b;
        }
    }


/**
 * Round given sizes to powers of 2, lower or equal to maximum texture sizes.
 *
 * Rounds up if @a extend is @c true, rounds to nearest power of 2 otherwise.
 *
 * @return @c true if one of sizes has been clipped by maximum texture size,
 *         @c false otherwise.
 */
bool choose_texture_size
    (
    int &   width,
    int &   height,
    bool    extend
    )
    {
    bool force_scale = false;

    const TextureManager & texture_manager = Global::get_texture_manager();
    const int max_texture_size = texture_manager.get_max_texture_size();
    hope( max_texture_size > 0 );
    
    if( width <= 0 )
        {
        width = 1;
        }
    else
        {
        int new_width = nearest_power_of_2( width );
        if( extend && ( new_width < width ) )
            {
            new_width <<= 1;
            hope( new_width >= width );
            }
        if( new_width > max_texture_size )
            {
            new_width = max_texture_size;
            force_scale = true;
            }
        width = new_width;
        }

    if( height <= 0 )
        {
        height = 1;
        }
    else
        {
        int new_height = nearest_power_of_2( height );
        if( extend && ( new_height < height ) )
            {
            new_height <<= 1;
            hope( new_height >= height );
            }
        if( new_height > max_texture_size )
            {
            new_height = max_texture_size;
            force_scale = true;
            }
        height = new_height;
        }

    return force_scale;
    }

 
/**
 * Scale given image.
 *
 * Image must be 24bit or 32bit truecolor, with rows aligned to dword boundaries.
 */
UInt8 * scale_image
    (
    int             width,
    int             height,
    int             row_length,
    bool            has_alpha,
    const UInt8 *   data,
    int             new_width,
    int             new_height
    )
    {
    // Allocate new image.

    const int bytes_per_texel = ( has_alpha ? 4 : 3 );
    int new_pitch = new_width * bytes_per_texel;
    if( new_pitch & 3 )
        {
        new_pitch += 4 - ( new_pitch & 3 );
        }

    // Can't use std::auto_ptr, we must use delete [], not plain delete.
    UInt8 * const result = new UInt8 [ new_height * new_pitch ];

    // Scale the image.
    
    const GLint err = gluScaleImage
        (
        has_alpha ? GL_RGBA : GL_RGB,
        width,
        height,
        GL_UNSIGNED_BYTE,
        data,
        new_width,
        new_height,
        GL_UNSIGNED_BYTE,
        result
        );

    if( err != 0 )
        {
        delete [] result;

        const char * errstr = reinterpret_cast< const char * >( gluErrorString( err ) );
        errstr = errstr ? errstr : "no error info";
        std::ostringstream  ss;
        ss << "Unable to scale texture image (glu error: " << errstr << ").";
        throw ExOpenGLError( ss.str() );
        }

    return result;
    }


/**
 * Extend given image with "black" texels.
 */
UInt8 * extend_image
    (
    int             width,
    int             height,
    int             row_length,
    bool            has_alpha,
    const UInt8 *   data,
    int             new_width,
    int             new_height
    )
    {
    hope( new_width >= width );
    hope( new_height >= height );

    // Allocate new image.

    const int bytes_per_texel = ( has_alpha ? 4 : 3 );
    int old_pitch = row_length * bytes_per_texel;
    if( old_pitch & 3 )
        {
        // FIXME: assert instead?
        old_pitch += 4 - ( old_pitch & 3 );
        }
    int new_pitch = new_width * bytes_per_texel;
    if( new_pitch & 3 )
        {
        new_pitch += 4 - ( new_pitch & 3 );
        }

    // Can't use std::auto_ptr, we must use delete [], not plain delete.
    UInt8 * const result = new UInt8 [ new_height * new_pitch ];

    // Extend the image.

    int y;
    for( y = 0; y < new_height; y++ )
        {
        const UInt8 * source = data + y * old_pitch;
        UInt8 * destination = result + y * new_pitch;
        int x;
        for( x = 0; x < width; x++ )
            {
            int i;
            for( i = 0; i < bytes_per_texel; i++ )
                {
                *destination++ = *source++;
                }
            }
        for( ; x < new_width; x++ )
            {
            int i;
            for( i = 0; i < bytes_per_texel; i++ )
                {
                *destination++ = 0;
                }
            }
        }

    return result;
    }


} // anonymous namespace


Texture::Reference Texture::create
    (
    int             width,
    int             height,
    int             pitch,
    bool            has_alpha,
    const UInt8 *   data,
    int             flags
    )
    {
    hope( width > 0 );
    hope( height > 0 );
    hope( pitch > 0 );
    hope( !( pitch & 3 ) );
    hope( data );

    {//sob
        DEBUG_ONLY( GLint alignment; )
        DEBUG_ONLY( glGetIntegerv( GL_PACK_ALIGNMENT, &alignment ) );
        hope( alignment == 4 );
        DEBUG_ONLY( glGetIntegerv( GL_UNPACK_ALIGNMENT, &alignment ) );
        hope( alignment == 4 );
    }//eob

    const int bytes_per_texel = ( has_alpha ? 4 : 3 );
    int row_length = pitch / bytes_per_texel;
    hope( row_length >= width );
    const int format = ( has_alpha ? GL_RGBA : GL_RGB );
    const int internal_format = ( has_alpha ? GL_RGBA8 : GL_RGB5 );

    // Resize image?

    const UInt8 * source = data;
    UInt8 * new_data = NULL; // can't use std::auto_ptr, it's an array
    int new_width = width;
    int new_height = height;
    int valid_width = width;
    int valid_height = height;
    const bool force_scale = choose_texture_size
        ( new_width, new_height, ( flags & EXTEND_IMAGE ) ? true : false );
    if( ( new_width != width ) || ( new_height != height ) || force_scale )
        {
        if( ( flags & EXTEND_IMAGE ) && !force_scale )
            {
            new_data = extend_image
                (
                width,
                height,
                row_length,
                has_alpha,
                data,
                new_width,
                new_height
                );
            }
        else
            {
            new_data = scale_image
                (
                width,
                height,
                row_length,
                has_alpha,
                data,
                new_width,
                new_height
                );
            valid_width = new_width;
            valid_height = new_height;
            }
        row_length = new_width;
        source = new_data;
        hope( source );
        }

    // Remember current binding.

    TextureManager & texture_manager = Global::get_texture_manager();
    Texture::Reference old_bound_texture = texture_manager.get_bound_texture();

    // We must catch exceptions to destroy new_data.

    Texture::Reference texture;
    try
        {
        // Create texture.

        texture = new Texture
            (
            new_width,
            new_height,
            valid_width,
            valid_height,
            has_alpha,
            flags & GENERATE_MIPMAPS ? true : false
            );

        // Bind the texture and set default parameters.

        texture_manager.bind_texture( texture );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
        texture->set_filter( FILTER_DEFAULT );

        // Upload image.

        if( row_length != new_width )
            {
            glPixelStorei( GL_UNPACK_ROW_LENGTH, row_length );
            }

        if( flags & GENERATE_MIPMAPS )
            {
            const GLint err = gluBuild2DMipmaps
                (
                GL_TEXTURE_2D,
                internal_format,
                new_width,
                new_height,
                format,
                GL_UNSIGNED_BYTE,
                const_cast< UInt8 * >( source )
                );
            if( err != 0 )
                {
                const char * errstr = reinterpret_cast< const char * >( gluErrorString( err ) );
                errstr = errstr ? errstr : "no error info";
                std::ostringstream ss;
                ss << "Unable to build mipmaps (glu error: " << errstr << ").";
                throw ExOpenGLError( ss.str() );
                }
            }
        else
            {
            glTexImage2D
                (
                GL_TEXTURE_2D,
                0, // Mipmap level.
                internal_format,
                new_width,
                new_height,
                0, // border
                format,
                GL_UNSIGNED_BYTE,
                const_cast< UInt8 * >( source )
                );
            }

        if( row_length != new_width )
            {
            glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
            }
        }
    catch( ...  )
        {
        // Cleanup. The same stuff as below. Stupid C++.

        texture_manager.bind_texture( old_bound_texture );
        delete [] new_data;

        throw;
        }

    // Revert binding.

    texture_manager.bind_texture( old_bound_texture );

    // Destroy new_data.

    delete [] new_data;

    // Done.

    OpenGL::check_errors( "Create texture: " );

    return texture;
    }


Texture::Texture
    (
    int     the_width,
    int     the_height,
    int     the_valid_width,
    int     the_valid_height,
    bool    the_alpha,
    bool    the_mipmaps
    ) : width( the_width ),
        height( the_height ),
        valid_width( the_valid_width ),
        valid_height( the_valid_height ),
        alpha( the_alpha ),
        mipmaps( the_mipmaps ),
        filter( FILTER_INVALID )
    {
    glGenTextures( 1, &handle );
    }


Texture::~Texture()
    {
    // Note: TextureManager ensures that bound texture is never destroyed.
    // Delete the handle.

    glDeleteTextures( 1, &handle );
    }


void Texture::set_filter
    (
    Filter new_filter
    )
    {
    hope( new_filter != FILTER_INVALID );
    if( new_filter == filter )
        {
        return;
        }

    // Bind texture.
    
    TextureManager & texture_manager = Global::get_texture_manager();
    texture_manager.bind_texture( this );

    // Remember new filter.
    
    filter = new_filter;

    // Magnification filter.

    int mag_filter = GL_LINEAR;
    if( filter == FILTER_NEAREST )
        {
        mag_filter = GL_NEAREST;
        }

    // Minifiaction filter.

    GLint min_filter = mag_filter;
    if( mipmaps )
        {
        switch( filter )
            {
            case FILTER_NEAREST:
                min_filter = GL_NEAREST_MIPMAP_NEAREST;
                break;
            case FILTER_BILINEAR:
                min_filter = GL_LINEAR_MIPMAP_NEAREST;
                break;
            case FILTER_TRILINEAR:
                min_filter = GL_LINEAR_MIPMAP_LINEAR;
                break;
            default:
                panic( "Bad filter." );
            }
        }

    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "texture_inline.h"
#endif
