/** @file texture.h
 * Simple OpenGL texture.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TEXTURE_H
#define CLIENT_TEXTURE_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_OPENGL_H
#include "renderer/opengl.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef MASSIV_REFERENCE_H
#include "utility/reference.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Object representing OpenGL texture.
 *
 * Texture is quite limited and not efficient - it supports only basic RGB and
 * RGBA formats, 1D and cubemap textures are not supported, internal format is
 * chosen automatically, data may be converted/scaled several times, mipmaps
 * will generated on the fly and can't be supplied externally, texture contents
 * can't be updated, etc.
 *
 * Always use Texture::Reference instead of native pointers to Texture. However,
 * to ensure that the most recent version of the texture is used never keep
 * reference to Texture for too long. Use #TextureDataObject to acquire
 * the reference instead. Also note that #TextureDataObject::Reference is part
 * of #PipelineState, not #Texture::Reference.
 *
 * It's possible to keep #Texture::Reference when corresponding
 * #TextureDataObject no longer exists, but it's not recommended. All texture
 * references kept by user must be forgotten before #Renderer is destroyed.
 */
class Texture : public Massiv::Core::SharedObject
    {
    NO_COPY_CONSTRUCTOR( Texture );
    NO_ASSIGNMENT_OPERATOR( Texture );

public:

    /**
     * Reference to Texture.
     *
     * Reference counting is used to keep track of Texture objects.
     */
    typedef Massiv::Core::Reference< Texture > Reference;

public:

    /**
     * Flags controlling creation of a new texture from memory.
     */
    enum CreateFlags
        {
        GENERATE_MIPMAPS =  0x0001,
            /**< Generate mipmaps. Currently there is now way to specify own mipmaps. */
        EXTEND_IMAGE =      0x0002,
            /**< If image sizes are not power of 2, enlarge it and fill "empty"
                 space with black textels.  If this flag is not set, image will
                 be scaled to nearest power of 2 sizes. */

        CREATEFLAGS_MASK =  0x0003,
            /**< All the flags set. Used by TextureDataObject. */

        };

    /**
     * Filtering mode.
     */
    enum Filter
        {
        FILTER_INVALID,
        FILTER_NEAREST,
        FILTER_BILINEAR,
        FILTER_TRILINEAR,

        FILTER_DEFAULT = FILTER_BILINEAR
        };

public:

    /**
     * Create texture object from pixel data stored in memory.
     *
     * Current texture bindings and active texture unit will not be changed.
     *
     * @param width     Texture width in pixels.
     * @param height    Texture height in pixels.
     * @param pitch     Row length in bytes. It must be multiple of 4.
     * @param has_alpha Is alpha channel included?
     * @param data      Image data, either in RGB or RGBA format,
     *                  unsigned byte per component.
     * @param flags     Combination of #CreateFlags.
     *
     * @exception OpenGLException will be thrown on error.
     *
     * @note Image may be silently scaled down if its dimensions too large,
     *       even if ALLOW_SCALE is not set or ALLOW_EXTEND is set.
     *
     * @note Each row must be aligned on dword boundary. This might be a problem
     *       with 24bit images with dimensions which are not multiple of 4 texels,
     *       but no-one should create such weird things.
     *
     * @todo Allow 16bit packed pixel formats as input.
     */
    static Texture::Reference create
        (
        int             width,
        int             height,
        int             pitch,
        bool            has_alpha,
        const UInt8 *   data,
        int             flags = 0
        );

private:

    /**
     * Constructor.
     */
    Texture
        (
        int     the_width,
        int     the_height,
        int     the_valid_width,
        int     the_valid_height,
        bool    the_alpha,
        bool    the_mipmaps
        );

protected:

    /**
     * Destructor.
     */
    virtual ~Texture();

public:

    //@{
    /** Getter. */
    INLINE int get_width() const;

    INLINE int get_height() const;

    INLINE int get_valid_width() const;

    INLINE int get_valid_height() const;

    INLINE bool has_alpha() const;

    INLINE bool has_mipmaps() const;

    INLINE GLuint get_handle() const;

    INLINE Filter get_filter() const;
    //@}

public:

    /**
     * Set filtering mode.
     *
     * This method binds the texture to active texture unit. It should be
     * called only via PipelineState to ensure state consistency.
     */
    void set_filter
        (
        Filter new_filter
        );

private:

    /**
     * Get index of currently active texturing unit.
     */
    static int get_active_texture_unit();

private:

    int     width;
        /**< Texture width in texels. */

    int     height;
        /**< Texture height in texels. */

    int     valid_width;
        /**< Width of valid texture region in texels.
             May be < #width iff image has been extender. */

    int     valid_height;
        /**< Height of valid texture region in texels.
             May be < #height iff image has been extender. */

    bool    alpha : 1;
        /**< Has alpha channel? */

    bool    mipmaps : 1;
        /**< Have mipmaps been generated? */

    GLuint  handle;
        /**< OpenGL texture handle. */

    Filter  filter;
        /**< Current filtering mode. */

    }; // class Texture


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "texture_inline.h"
#endif


#endif // CLIENT_TEXTURE_H
