/** @file texture_data_object.cpp
 * Texture data object (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "client_exception.h"
#include "renderer/texture_data_object.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include "sdl/stream_rwops.h"
#include "data/data_manager.h"
#include "global/global.h"
#include <SDL_image.h>
#include <sstream>


namespace Demo {
namespace Client {
namespace Renderer {


TextureDataObject::TextureDataObject
    (
    std::auto_ptr< Massiv::Core::DataObject > & the_data_object
    ) : reference_count( 0 ),
        data_object( the_data_object ),
        version( 1 )
    {
    hope( data_object.get() );
    hope( !the_data_object.get() );

        {
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << "Texture data object '" << data_object->get_name() << "' created.";
        Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                    Status::PRIORITY_LOW ), oss.str() );
        }
    }


TextureDataObject::~TextureDataObject()
    {
    hope( reference_count == 0 );
    forget_textures();

        {
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << "Texture data object '" << data_object->get_name() << "' destroyed.";
        Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                    Status::PRIORITY_LOW ), oss.str() );
        }
    }


std::string TextureDataObject::get_name() const
    {
    hope( data_object.get() );
    return data_object->get_name();
    }


Texture::Reference TextureDataObject::get_texture
    (
    int flags
    )
    {
    // Invalidate all cached textures if object has changed.

    hope( data_object.get() );
    if( data_object->is_data_updated() )
        {
        data_object->refresh();
        forget_textures();
        version++;

            {
            using namespace Massiv::Core;
            std::ostringstream oss;
            oss << "Texture data object '" << data_object->get_name() << "' updated.";
            Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                        Status::PRIORITY_HIGH ), oss.str() );
            }
        }

    // Create new texture unless cached.

    int index = texture_index( flags );
    if( textures[ index ].is_null() )
        {
        textures[ index ] = create_texture( flags );
        }

    // Tell data manager to download new version if available.

    data_object->download_actual_version();

    return textures[ index ];
    }


Texture::Reference TextureDataObject::create_texture
    (
    int flags
    )
    {
    hope( data_object.get() );

    // Try to read the image.

    data_object->open();
    SDL_RWops * const rwops = SDL::create_stream_rwops( data_object->data );
    hope( rwops );
    SDL_Surface * const surface = IMG_Load_RW( rwops, 0 );
    rwops->close( rwops );
    data_object->close();

    if( !surface )
        {
        std::ostringstream oss;
        oss << "Unable to load texture '" << data_object->get_name() <<
            "'. " << SDL_GetError();
        throw ExSDLError( oss.str() );
        }

    // Convert the image to format compatible with Texture.

    SDL_PixelFormat format;
    format.palette = NULL;
    format.Rloss = 0;
    format.Gloss = 0;
    format.Bloss = 0;
    format.Aloss = 0;
    format.colorkey = 0;
    format.alpha = 0;

    hope( surface->format );
    const bool has_alpha = ( surface->format->Amask || ( flags & INTENSITY_ALPHA ) );
    if( has_alpha )
        {
        format.BitsPerPixel = 32;
        format.BytesPerPixel = 4;
        if( SDL_BYTEORDER == SDL_LIL_ENDIAN )
            {
            format.Rmask = 0x000000ff;
            format.Gmask = 0x0000ff00;
            format.Bmask = 0x00ff0000;
            format.Amask = 0xff000000;
            format.Rshift = 0;
            format.Gshift = 8;
            format.Bshift = 16;
            format.Ashift = 24;
            }
        else
            {
            format.Rmask = 0xff000000;
            format.Gmask = 0x00ff0000;
            format.Bmask = 0x0000ff00;
            format.Amask = 0x000000ff;
            format.Rshift = 24;
            format.Gshift = 16;
            format.Bshift = 8;
            format.Ashift = 0;
            }
        }
    else
        {
        format.BitsPerPixel = 24;
        format.BytesPerPixel = 3;
        if( SDL_BYTEORDER == SDL_LIL_ENDIAN )
            {
            format.Rmask = 0x000000ff;
            format.Gmask = 0x0000ff00;
            format.Bmask = 0x00ff0000;
            format.Rshift = 0;
            format.Gshift = 8;
            format.Bshift = 16;
            }
        else
            {
            format.Rmask = 0xff000000;
            format.Gmask = 0x00ff0000;
            format.Bmask = 0x0000ff00;
            format.Rshift = 16;
            format.Gshift = 8;
            format.Bshift = 0;
            }
        format.Amask = 0x00000000;
        format.Ashift = 0;
        }

    SDL_Surface * const new_surface = SDL_ConvertSurface( surface, &format, SDL_SWSURFACE );
    if( !new_surface )
        {
        SDL_FreeSurface( surface );
        std::ostringstream oss;
        oss << "Unable to convert pixel format of texture '" <<
            data_object->get_name() << "'. " << SDL_GetError();
        throw ExSDLError( oss.str() );
        }

    const int width = new_surface->w;
    const int height = new_surface->h;
    const int pitch = new_surface->pitch;

    if( pitch & 0x03 )
        {
        SDL_FreeSurface( new_surface );
        SDL_FreeSurface( surface );
        std::ostringstream oss;
        oss << "Unable to load texture '" << data_object->get_name() <<
            "', pitch is not multiple of 4. Try making it at least 4 pixels wide.";
        throw ExSDLError( oss.str() ); // I assume that SDL should create dword aligned image
        }

    // Lock surface.

    SDL_LockSurface( new_surface );
    UInt8 * const pixels = static_cast< UInt8 * >( new_surface->pixels );

    // Fill alpha channel?

    if( !surface->format->Amask && ( flags & INTENSITY_ALPHA ) )
        {
        hope( new_surface->format->BitsPerPixel == 32 );
        int y;
        for( y = 0; y < height; y++ )
            {
            UInt8 * row = pixels + y * pitch;
            int x;
            for( x = 0; x < width; x++ )
                {
                const int intensity = ( row[ 0 ] + row[ 1 ] + row[ 2 ] ) / 3; // or classical weights?
                hope( intensity < 256 );
                row[ 3 ] = static_cast< UInt8 >( intensity );
                row += 4;
                }
            }
        }

    Texture::Reference texture;
    try 
        {
        texture = Texture::create
            (
            width,
            height,
            pitch,
            has_alpha,
            pixels,
            flags & Texture::CREATEFLAGS_MASK
            );
        }
    catch( ... )
        {
        if( new_surface )
            {
            SDL_UnlockSurface( new_surface );
            SDL_FreeSurface( new_surface );
            }
        if( surface )
            {
            SDL_FreeSurface( surface );
            }
        throw;
        }

    SDL_UnlockSurface( new_surface );
    SDL_FreeSurface( new_surface );
    SDL_FreeSurface( surface );

        {
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << "Texture " << texture_index( flags ) <<
            " created from data object '" << data_object->get_name() << "'.";
        Massiv::Core::Global::log( Status::info( Status::FACILITY_GENERIC,
                    Status::PRIORITY_NORMAL ), oss.str() );
        }

    return texture;
    }


void TextureDataObject::forget_textures()
    {
    int i;
    for( i = 0; i < TEXTURE_COUNT; i++ )
        {
        textures[ i ].set_null();
        }
    }


void TextureDataObject::zero_references()
    {
    TextureManager & texture_manager = Global::get_texture_manager();
    texture_manager.unreferenced_data_object( this );
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "texture_data_object_inline.h"
#endif
