/** @file texture_data_object.h
 * Texture data object.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TEXTURE_DATA_OBJECT_H
#define CLIENT_TEXTURE_DATA_OBJECT_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_TEXTURE_H
#include "renderer/texture.h"
#endif

#include <memory>


namespace Massiv {
namespace Core {

class DataObject;

}
}


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Texture data object.
 *
 * Object that represents a texture data. You can request a Texture object
 * from it at any time. It will automatically return texture with the most
 * up-to-date data available. Because reference counting is used on Textures,
 * Texture objects may still be used even if new version of the texture is
 * available.
 *
 * You must use TextureDataObject::Reference instead of plain pointers to
 * TextureDataObject. It ensures that the TextureDataObject will not be
 * destroyed by TextureManager. TextureManager is free to destroy unreferenced
 * data objects whenever it wants. However unlike Texture, TextureDataObject
 * can't outlive TextureManager. It's an error to destroy TextureManager while
 * references to TextureDataObjects still exist.
 */
class TextureDataObject
    {
    NO_COPY_CONSTRUCTOR( TextureDataObject );
    NO_ASSIGNMENT_OPERATOR( TextureDataObject );
    
    /// Only TextureManager may create and destroy TextureDataObjects.
    friend class TextureManager;

    /// TextureManager uses autopointers to keep TextureDataObjects.
    FRIEND_AUTO_PTR( TextureDataObject );

public:

    /**
     * Flags additional to Texture::CreateFlags.
     */
    enum GetTextureFlags
        {
        INTENSITY_ALPHA = 0x1000,
            /**< If set and image has no alpha channel, texel intensity will be
              *  used as alpha. Hack useful only for font textures without alpha.
              */
        };
    
public:

    /**
     * Reference counting traits.
     */
    class RefTraits
        {
    public:
    
        /** Adds reference to shared object. */
        static INLINE void add_reference
            (
            const TextureDataObject *   ptr
            );

        /** Removes reference from shared object. */
        static INLINE void remove_reference
            (
            TextureDataObject *         ptr
            );

        /** Changes reference to shared object from @a ptr2 to @a ptr1. */
        static INLINE void change_reference
            (
            const TextureDataObject *   ptr1,
            TextureDataObject *         ptr2
            );

        }; // class RefTraits

    /// RefTraits can hack into the data object.
    friend class RefTraits;

    /**
     * Reference to TextureDataObject.
     */
    typedef Massiv::Core::Reference< TextureDataObject, RefTraits > Reference;

protected:

    /**
     * Constructor.
     *
     * @a Ownership of the_data_object will be transferred to the TextureDataObject.
     */
    TextureDataObject
        (
        std::auto_ptr< Massiv::Core::DataObject > & the_data_object
        );

    /**
     * Destructor.
     */
    ~TextureDataObject();

public:

    /**
     * Get texture data object name.
     */
    std::string get_name() const;
    
    /**
     * Get the texture.
     *
     * New texture will be created if texture object with given @a flags does
     * not exist yet. Otherwise new reference to already existing texture
     * will be returned. You should not keep the reference to the texture.
     * Keep TextureDataObject instead and call #get_texture() whenever you
     * need the texture object. You should never keep reference to Texture
     * without reference to corresponding TextureDataObject (it will work,
     * but Texture with the same contents could be created multiple times
     * if it's used this way from multiple places).
     *
     * Texture binding and active texture unit will not be changed.
     *
     * @param flags Combination of Texture::CreateFlags and GetTextureFlags.
     *
     * @exception ExSDLError Failed to read or convert the data.
     * @exception ExOpenGLError Failed to create OpenGL texture.
     *
     * @note ExSDLError muze byt hozeno kdyz je pitch spatne. Co s tim? Asi by
     *       to byla chyba v SDL, ale ja jsem nic o pitchi surface v dokumentaci
     *       SDL nenasel (boovie nekde psal ze musi bejt nasobek 4).
     */
    Texture::Reference get_texture
        (
        int flags
        );

    /**
     * Get version of the texture.
     *
     * Texture version is an integer that starts at 1 (not 0) and is increased
     * whenever #get_texture() detects that the data object has been updated and
     * discards all cached textures. Usage example:
     *
     * @code
     * static int old_version = 0;
     * Texture::Reference texture = texture_data.get_texture( flags );
     * int new_version = texture_data.get_version();
     * if( new_version != old_version )
     *     {
     *     // texture contents and attributes (such as width) may have changed...
     *     ...
     *     old_version = new_version;
     *     }
     * @endcode
     */
    INLINE int get_version() const;

private:

    /**
     * Number of textures kept by single TextureDataObject.
     *
     * Different Texture objects are kept for different texture flags.
     *
     * @note In same cases two texture objects with different flags will
     *       actually contain the same data. There is no easy way to guess
     *       this without actually creating the textures. And it's not a
     *       problem as we assume that different combinations of the flags
     *       won't be often used for the same texture.
     */
    enum
        {
        TEXTURE_COUNT = 8
        };

    /**
     * Convert texture flags (Texture::CreateFlags and GetTextureFlags)
     * to index into @a textures.
     */
    INLINE static int texture_index
        (
        int flags
        );

private:

    /**
     * Create new texture.
     *
     * @todo SDL_image does not support automatic Targa format recognition -\>
     *       Targa images can't be used as textures. Write our own recognition
     *       function? (even though it will be just a wild guess)
     */
    Texture::Reference create_texture
        (
        int flags
        );

    /**
     * Set all @a textures to NULL.
     */
    void forget_textures();

    /**
     * Inform the object that the reference count dropped to zero.
     */
    void zero_references();

private:

    mutable int                 reference_count;
        /**< Reference count. */

    std::auto_ptr< Massiv::Core::DataObject > data_object;
        /**< Texture data. */

    int                         version;
        /**< Texture version. */

    Texture::Reference          textures[ TEXTURE_COUNT ];
        /**< Textures. Some of them may be NULL. */

    }; // class TextureDataObject


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "texture_data_object_inline.h"
#endif


#endif // CLIENT_TEXTURE_DATA_OBJECT_H
