/** @file texture_manager.cpp
 * Texture manager (implementation).
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/texture_manager.h"
#include "global/global.h"
#include "data/data_manager.h"
#include <memory>


namespace Demo {
namespace Client {
namespace Renderer {


TextureManager::TextureManager()
      : max_texture_size( 0 ),
        texture_units( 0 ),
        active_texture_unit( 0 )
    {
    glGetIntegerv( GL_MAX_TEXTURE_SIZE, &max_texture_size );

    if( GL::ARB_multitexture_supported )
        {
        glGetIntegerv( GL_MAX_TEXTURE_UNITS_ARB, &texture_units );
        hope( texture_units >= 1 );
        if( texture_units > MAX_TEXTURE_UNITS )
            {
            texture_units = MAX_TEXTURE_UNITS;
            }
        GL::ActiveTextureARB( GL_TEXTURE0_ARB );
        GL::ClientActiveTextureARB( GL_TEXTURE0_ARB );
        }
    else
        {
        texture_units = 1;

        using namespace Massiv::Core;       
        std::ostringstream oss;
        Massiv::Core::Global::log( Status::warning( Status::FACILITY_GENERIC ),
            "Multitexturing is not supported, expect random garbage on screen." );
        }
    }


TextureManager::~TextureManager()
    {
    // Check that no textures are bound.

    int unit;
    for( unit = 0; unit < texture_units; unit++ )
        {
        if( !bound_textures[ unit ].is_null() )
            {
            panic( "Textures still bound in ~TextureManager." );
            }
        }
    
    // Check that list of referenced textures is empty.

    if( !referenced_textures.empty() )
        {
        std::ostringstream oss;
        oss << "References to these TextureDataObjects still exist in ~TextureManager: ";
        TextureDataObjects::const_iterator it;
        bool first = true;
        for( it = referenced_textures.begin(); it != referenced_textures.end(); ++it )
            {
            if( !first )
                {
                oss << ", ";
                }
            else
                {
                first = false;
                }
            oss << it->first;
            }
        panic( oss.str().c_str() );
        }

    // Clear unreferenced textures.

    clear_unreferenced_textures();
    }


void TextureManager::activate_texture_unit
    (
    int index
    )
    {
    hope( index < texture_units );
    if( index != active_texture_unit )
        {
        hope( GL::ARB_multitexture_supported );
        GL::ActiveTextureARB( GL_TEXTURE0_ARB + index );
        GL::ClientActiveTextureARB( GL_TEXTURE0_ARB + index );
        active_texture_unit = index;
        }
    }


void TextureManager::bind_texture
    (
    Texture::Reference texture
    )
    {
    if( bound_textures[ active_texture_unit ] != texture )
        {
        if( texture.is_null() )
            {
            glBindTexture( GL_TEXTURE_2D, 0 );
            }
        else
            {
            glBindTexture( GL_TEXTURE_2D, texture->get_handle() );
            }
        bound_textures[ active_texture_unit ] = texture;
        }
    }


TextureDataObject::Reference TextureManager::get_texture_data_object
    (
    const std::string & name
    )
    {
    // Try to find the texture.

    TextureDataObjects::iterator it = referenced_textures.find( name );
    if( it != referenced_textures.end() )
        {
        // Active object found.

        TextureDataObject * const texture = it->second;
        hope( texture );
        hope( texture->reference_count >= 1 );
        return texture;
        }

    it = unreferenced_textures.find( name );
    if( it != unreferenced_textures.end() )
        {
        // Inactive object found. Remove it from "unreferenced" list.

        hope( it->second );
        std::auto_ptr< TextureDataObject > texture( it->second );
        hope( texture->reference_count == 0 );
        unreferenced_textures.erase( it );

        // Move it back to "referenced" list.
        // If insert throws, texture will not be in neither list and it will be
        // destroyed by the auto_ptr.
        
        referenced_textures.insert( TextureDataObjects::value_type( name, texture.get() ) );
        return texture.release();
        }

    // Create new texture data object.

    Massiv::Core::DataManager & data_manager = Massiv::Core::Global::data_manager();
    std::auto_ptr< Massiv::Core::DataObject > data_object( data_manager.get_data_object( name ) );
    std::auto_ptr< TextureDataObject > texture( new TextureDataObject( data_object ) );
    hope( !data_object.get() );

    // Insert it into list of referenced data objects.
    // See comment above about the insert.

    hope( texture->reference_count == 0 );
    referenced_textures.insert( TextureDataObjects::value_type( name, texture.get() ) );
    return texture.release();
    }

    
void TextureManager::collect_garbage()
    {
    clear_unreferenced_textures();
    }

    
void TextureManager::unreferenced_data_object
    (
    TextureDataObject * texture
    )
    {
    hope( texture );
    hope( texture->reference_count == 0 );
    const std::string name = texture->get_name();

    // Remove texture from "referenced" list.

    TextureDataObjects::iterator it = referenced_textures.find( name );
    hope( it != referenced_textures.end() );
    hope( it->second == texture );
    std::auto_ptr< TextureDataObject > auto_texture( it->second );
    referenced_textures.erase( it );

    // Move it to "unreferenced" list.
    // If insert throws, texture will be destroyed by the auto_ptr.

    unreferenced_textures.insert( TextureDataObjects::value_type( name, texture ) );
    auto_texture.release();
    }


void TextureManager::clear_unreferenced_textures()
    {
    TextureDataObjects::iterator it;
    for( it = unreferenced_textures.begin(); it != unreferenced_textures.end(); ++it )
        {
        hope( it->second );
        hope( it->second->reference_count == 0 );
        delete it->second;
        it->second = NULL;
        }

    unreferenced_textures.clear();
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "texture_manager_inline.h"
#endif
