/** @file texture_manager.h
 * Texture manager.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TEXTURE_MANAGER_H
#define CLINET_TEXTURE_MANAGER_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_TEXTURE_DATA_OBJECT_H
#include "renderer/texture_data_object.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Manager of texture data objects.
 *
 * Use TextureManager to get TextureDataObjects.
 *
 * TextureManager also provides information about OpenGL texture units and
 * keeps track of currently active unit. This information is used by rendering
 * objects such as Texture or PipelineState, Renderer user can safely ignore
 * this stuff.
 */
class TextureManager
    {
    NO_COPY_CONSTRUCTOR( TextureManager );
    NO_ASSIGNMENT_OPERATOR( TextureManager );

public:

    /**
     * Constructor.
     *
     * @note Global PipelineState does not exists yet.
     */
    TextureManager();

    /**
     * Destructor.
     *
     * No texture data objects may be referenced when destroying the manager.
     * All textures should be unbound while global PipelineState still exists.
     *
     * @note Global PipelineState no longer exists.
     */
    ~TextureManager();

public:

    /**
     * Get maximal texture size in texels.
     */
    INLINE int get_max_texture_size() const;

    /**
     * Get number of texture units.
     */
    INLINE int get_texture_units() const;

    /**
     * Get index of currently active texture unit.
     */
    INLINE int get_active_texture_unit() const;

    /**
     * Activate given texture unit.
     */
    void activate_texture_unit
        (
        int index
        );

public:

    /**
     * Get texture bound to active unit.
     */
    INLINE Texture::Reference get_bound_texture() const;
     
    /**
     * Bind given texture to active unit.
     *
     * Use @c NULL @a texture to unbind current texture.
     */
    void bind_texture
        (
        Texture::Reference texture
        );

    /**
     * Unbind texture from active unit.
     *
     * Shortcut for <tt>bind_texture( NULL )</tt>.
     */
    INLINE void unbind_texture();

public:

    /**
     * Get reference to texture data object with given @a name.
     *
     * NULL reference will never be returned.
     *
     * @param name
     *      Name of image data object.
     *
     * @exception Anything data manager may throw. No one really knows what it may be.
     */
    TextureDataObject::Reference get_texture_data_object
        (
        const std::string & name
        );

    /**
     * Get rid of some unused texture data objects.
     *
     * Currently all unreferenced data objects will be destroyed. In the future
     * LRU-like algorithm should be used if too many data objects exists.
     * Sometimes it may be unwise to destroy texture as soon as it becomes
     * unused.
     *
     * @note This is called automatically by Renderer.
     */
    void collect_garbage();
     
protected:

    /// TextureDataObject may inform TextureManager that its reference count is zero.
    friend class TextureDataObject;

    /**
     * Reference count of data object reached zero.
     */
    void unreferenced_data_object
        (
        TextureDataObject * texture
        );

private:

    /**
     * Clear all unreferenced texture data objects.
     */
    void clear_unreferenced_textures();

private:

    int                 max_texture_size;
        /**< Maximal texture size in texels. */

    int                 texture_units;
        /**< Number of texture units. */

    int                 active_texture_unit;
        /** Index of currently active texture unit. */

    Texture::Reference  bound_textures[ MAX_TEXTURE_UNITS ];
        /**< Currently bound textures. */
        
private:

    /**
     * Hash of owning texture data object pointers.
     *
     * @note Unfortunately std::auto_ptr has neither copy-constructor nor
     *       assignment operator with <tt>const &</tt> arguments and can't
     *       be used as a hash value type. So we have to destroy them manually.
     */
    typedef ext::hash_map< std::string, TextureDataObject * > TextureDataObjects;

    TextureDataObjects  referenced_textures;
        /**< Textures with non-zero reference count. */

    TextureDataObjects  unreferenced_textures;
        /**< Textures with zero reference count.
             @todo Implement wise algorithm to get rid of these. */
    
    }; // class TextureManager


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "texture_manager_inline.h"
#endif


#endif // CLIENT_TEXTURE_MANAGER_H
