/** @file vertex_array.cpp
 * Vertex arrays (implementation).
 *
 * @todo Remove OpenGL::check_errors() and let array user to check the errors?
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "renderer/vertex_array.h"
#include "renderer/renderer.h"
#include "renderer/texture_manager.h"
#include "renderer/texture.h"


namespace Demo {
namespace Client {
namespace Renderer {


// Check if expected max. number of texture units matches number defined renderer_fwd.h.
namespace {
typedef int MaxTextureUnitsIsTwo[ ( MAX_TEXTURE_UNITS == 2 ) ? 1 : -1 ];
}


VertexArray * VertexArray::bound = NULL;


VertexArray::VertexArray()
      : memory( NULL ),
        size( 0 )
    {
    std::size_t i;
    for( i = 0; i < TYPE_COUNT; i++ )
        {
        allocated_streams[ i ] = 0;
        }
    for( i = 0; i < TARGET_COUNT; i++ )
        {
        selected_streams[ i ].index = -1;
        }
    }


VertexArray::~VertexArray()
    {
    free();
    }


void VertexArray::reset_stream_counts()
    {
    std::size_t i;
    for( i = 0; i < TYPE_COUNT; i++ )
        {
        set_stream_count( static_cast< StreamType >( i ), 0 );
        }
    }


void VertexArray::allocate
    (
    std::size_t element_count
    )
    {
    free();

    // Set size. Required by #stream_index and #stream_offset.

    size = element_count;

    // Allocate #locked_streams.

    locked_streams.resize( stream_index( TYPE_COUNT, 0 ) );

    // Allocate the array. Currently always using system memory.

    const std::size_t bytes = stream_offset( TYPE_COUNT, 0 );
    memory = new UInt8 [ bytes ];
    }


void VertexArray::free()
    {
    if( memory )
        {
        hope( !lock_count() );
        hope( bound != this );

        delete [] memory;
        memory = NULL;
        size = 0;

        std::size_t i;
        for( i = 0; i < TARGET_COUNT; i++ )
            {
            selected_streams[ i ].type = TYPE_COUNT; // invalid
            selected_streams[ i ].index = -1;        // nothing selected
            }

        // Remeber allocated_streams counts...
        }
    }


void VertexArray::lock_vectors2
    (
    Vector2Stream & stream,
    std::size_t    index
    )
    {
    hope( memory );
    hope( bound != this );
    hope( index < allocated_streams[ TYPE_VECTOR2 ] );

    // Lock it.
    
    const std::size_t bit_index = stream_index( TYPE_VECTOR2, index );
    hope( !locked_streams[ bit_index ] );
    locked_streams[ bit_index ] = true;

    // Bind the stream.

    stream.bind( reinterpret_cast< OpenGL::Vector2 * >
        ( memory + stream_offset( TYPE_VECTOR2, index ) ), size );
    }


void VertexArray::lock_vectors3
    (
    Vector3Stream & stream,
    std::size_t     index
    )
    {
    hope( memory );
    hope( bound != this );
    hope( index < allocated_streams[ TYPE_VECTOR3 ] );

    // Lock it.

    const std::size_t bit_index = stream_index( TYPE_VECTOR3, index );
    hope( !locked_streams[ bit_index ] );
    locked_streams[ bit_index ] = true;

    // Bind the stream.

    stream.bind( reinterpret_cast< OpenGL::Vector3 * >
        ( memory + stream_offset( TYPE_VECTOR3, index ) ), size );
    }


void VertexArray::lock_colors
    (
    ColorStream &   stream,
    std::size_t     index
    )
    {
    hope( memory );
    hope( bound != this );
    hope( index < allocated_streams[ TYPE_COLOR ] );

    // Lock it.

    const std::size_t bit_index = stream_index( TYPE_COLOR, index );
    hope( !locked_streams[ bit_index ] );
    locked_streams[ bit_index ] = true;

    // Bind the stream.

    stream.bind( reinterpret_cast< OpenGL::Color * >
        ( memory + stream_offset( TYPE_COLOR, index ) ), size );
    }


void VertexArray::unlock
    (
    StreamType  type,
    std::size_t index
    )
    {
    hope( memory );
    hope( bound != this );
    hope( type >= 0 );
    hope( type < TYPE_COUNT );
    hope( index < allocated_streams[ type ] );

    const std::size_t bit_index = stream_index( type, index );
    hope( locked_streams[ bit_index ] );
    locked_streams[ bit_index ] = false;
    }


void VertexArray::select
    (
    StreamTarget    target,
    StreamType      type,
    std::size_t     index
    )
    {
    hope( memory );
    hope( bound != this );
    hope( target >= 0 );
    hope( target < TARGET_COUNT );
    hope( type >= 0 );
    hope( type < TYPE_COUNT );
    hope( index < allocated_streams[ type ] );

    if( target == VERTEX_ARRAY )
        {
        hope( ( type == TYPE_VECTOR2 ) || ( type == TYPE_VECTOR3 ) );
        }
    else if( target == NORMAL_ARRAY )
        {
        hope( type == TYPE_VECTOR3 );
        }
    else if( target == COLOR_ARRAY )
        {
        hope( type == TYPE_COLOR );
        }
    else
        {
        //texcoord
        hope( type == TYPE_VECTOR2 );
        }

    selected_streams[ target ].type = type;
    selected_streams[ target ].index = static_cast< int >( index );
    }


void VertexArray::deselect
    (
    StreamTarget    target
    )
    {
    hope( memory );
    hope( bound != this );
    hope( target >= 0 );
    hope( target < TARGET_COUNT );

    selected_streams[ target ].type = TYPE_COUNT;
    selected_streams[ target ].index = -1;
    }

    
void VertexArray::bind()
    {
    hope( memory );
    hope( !lock_count() );
    hope( !bound );

    const StreamSelection & vertex = selected_streams[ VERTEX_ARRAY ];
    if( vertex.index >= 0 )
        {
        hope( ( vertex.type == TYPE_VECTOR2 ) || ( vertex.type == TYPE_VECTOR3 ) );
        glVertexPointer( ( vertex.type == TYPE_VECTOR2 ) ? 2 : 3, GL_FLOAT, 0,
            memory + stream_offset( vertex.type, vertex.index ) );
        glEnableClientState( GL_VERTEX_ARRAY );
        }

    const StreamSelection & normal = selected_streams[ NORMAL_ARRAY ];
    if( normal.index >= 0 )
        {
        hope( normal.type == TYPE_VECTOR3 );
        glNormalPointer( GL_FLOAT, 0,
            memory + stream_offset( TYPE_VECTOR3, normal.index ) );
        glEnableClientState( GL_NORMAL_ARRAY );
        }

    const StreamSelection & color = selected_streams[ COLOR_ARRAY ];
    if( color.index >= 0 )
        {
        hope( color.type == TYPE_COLOR );
        glColorPointer( 4, GL_UNSIGNED_BYTE, 0,
            memory + stream_offset( TYPE_COLOR, color.index ) );
        glEnableClientState( GL_COLOR_ARRAY );
        }

    TextureManager & texture_manager = Global::get_texture_manager();
    const std::size_t units = texture_manager.get_texture_units();
    std::size_t unit;
    for( unit = 0; unit < units; unit++ )
        {
        hope( unit < MAX_TEXTURE_UNITS );
        hope( TEXCOORD_ARRAY_0 + unit < TARGET_COUNT );
        const StreamSelection & texcoord = selected_streams[ TEXCOORD_ARRAY_0 + unit ];
        if( texcoord.index >= 0 )
            {
            hope( texcoord.type == TYPE_VECTOR2 );
            texture_manager.activate_texture_unit( unit );
            glTexCoordPointer( 2, GL_FLOAT, 0,
                memory + stream_offset( TYPE_VECTOR2, texcoord.index ) );
            glEnableClientState( GL_TEXTURE_COORD_ARRAY );
            }
        }

    OpenGL::check_errors( "Enable vertex array: " );
    bound = this;
    }


void VertexArray::unbind()
    {
    hope( bound == this );

    if( selected_streams[ VERTEX_ARRAY ].index >= 0 )
        {
        glDisableClientState( GL_VERTEX_ARRAY );
        }

    if( selected_streams[ NORMAL_ARRAY ].index >= 0 )
        {
        glDisableClientState( GL_NORMAL_ARRAY );
        }

    if( selected_streams[ COLOR_ARRAY ].index >= 0 )
        {
        glDisableClientState( GL_COLOR_ARRAY );
        }

    TextureManager & texture_manager = Global::get_texture_manager();
    const std::size_t units = texture_manager.get_texture_units();
    std::size_t unit;
    for( unit = 0; unit < units; unit++ )
        {
        hope( unit < MAX_TEXTURE_UNITS );
        hope( TEXCOORD_ARRAY_0 + unit < TARGET_COUNT );
        if( selected_streams[ TEXCOORD_ARRAY_0 + unit ].index >= 0 )
            {
            texture_manager.activate_texture_unit( unit );
            glDisableClientState( GL_TEXTURE_COORD_ARRAY );
            }
        }

    OpenGL::check_errors( "Disable vertex array: " );
    bound = NULL;
    }


void VertexArray::render
    (
    GLenum              mode,
    GLint               first,
    GLsizei             count
    )
    {
    hope( bound == this );

    if( count == 0 )
        {
        count = size;
        }
    hope( first >= 0 );
    hope( static_cast< std::size_t >( first ) + count <= size );
    glDrawArrays( mode, first, count );
        
    OpenGL::check_errors( "Render vertex array: " );
    }


void VertexArray::render
    (
    GLenum              mode,
    const GLushort *    indices,
    GLsizei             count,
    GLint               start,
    GLint               end
    )
    {
    hope( indices );
    hope( bound == this );

    void * const void_indices = const_cast< GLushort * >( indices );
    if( ( start >= 0 ) && ( end >= 0 ) && GL::EXT_draw_range_elements_supported )
        {
        hope( start < end );
        hope( static_cast< std::size_t >( start ) < size );
        hope( static_cast< std::size_t >( end ) < size );
        GL::DrawRangeElementsEXT( mode, start, end, count, GL_UNSIGNED_SHORT, void_indices );
        }
    else
        {
        glDrawElements( mode, count, GL_UNSIGNED_SHORT, void_indices );
        }

    OpenGL::check_errors( "Render vertex array: " );
    }


std::size_t VertexArray::element_size
    (
    StreamType type
    ) const
    {
    switch( type )
        {
        case TYPE_VECTOR2:
            return sizeof( OpenGL::Vector2 );

        case TYPE_VECTOR3:
            return sizeof( OpenGL::Vector3 );

        case TYPE_COLOR:
            return sizeof( OpenGL::Color );

        case TYPE_COUNT:
            break;
        }

    panic( "Bad stream type." );
    /*NOTREACHED*/
    }


std::size_t VertexArray::stream_offset
    (
    StreamType  type,
    std::size_t index
    ) const
    {
    hope( type >= 0 );
    hope( type <= TYPE_COUNT );

    // Compute how many bytes streams of type < @a type occupy.

    std::size_t offset = 0;
    std::size_t i;
    for( i = 0; i < static_cast< std::size_t >( type ); i++ )
        {
        offset += allocated_streams[ i ] * size *
            element_size( static_cast< StreamType >( i ) );
        }

    // Add size of streams of type @a stream before this stream.

    if( index > 0 )
        {
        hope( type < TYPE_COUNT );
        hope( index < allocated_streams[ type ] );
        offset += index * size * element_size( type );
        }

    return offset;
    }


std::size_t VertexArray::stream_index
    (
    StreamType  type,
    std::size_t index
    ) const
    {
    hope( type >= 0 );
    hope( ( ( type == TYPE_COUNT ) && ( index == 0 ) ) ||
            ( ( type < TYPE_COUNT ) && ( index < allocated_streams[ type ] ) ) );

    // Sum counts for streams of type < @a type.

    std::size_t result = 0;
    std::size_t i;
    for( i = 0; i < static_cast< std::size_t >( type ); i++ )
        {
        result += allocated_streams[ i ];
        }

    // Add index.

    return result + index;
    }


/**
 * Get number of locked streams.
 */
std::size_t VertexArray::lock_count() const
    {
    const std::size_t bits = stream_index( TYPE_COUNT, 0 );
    std::size_t count = 0;
    std::size_t i;
    for( i = 0; i < bits; i++ )
        {
        if( locked_streams[ i ] )
            {
            count++;
            }
        }

    return count;
    }


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "vertex_array_inline.h"
#endif
