/** @file vertex_array.h
 * Vertex arrays.
 */

/*
 * begin        : Jun 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_VERTEX_ARRAY_H
#define CLIENT_VERTEX_ARRAY_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef CLIENT_RENDERER_FWD_H
#include "renderer/renderer_fwd.h"
#endif

#ifndef CLIENT_OPENGL_TYPES_H
#include "renderer/opengl_types.h"
#endif

#ifndef MASSIV_ARRAY_H
#include "utility/array.h"
#endif

#include <vector>
#include <iosfwd>


namespace Demo {
namespace Client {
namespace Renderer {


/**
 * Object encapsulating single OpenGL vertex array.
 *
 * #VertexArray contains streams of several types (see #StreamType).
 * Any number (including zero) of streams of single type may be allocated.
 * Every stream contains the same number of elements (see #get_size()).
 *
 * Use #set_stream_count() or #choose_streams() and #allocate() to choose
 * streams and allocate memory for them. You can free the streams without
 * destroying the VertexArray by calling #free().
 *
 * For each stream type, single stream is "active". Use #select() to change
 * active stream.
 *
 * To access stream data, call #lock(). It will lock active streams of given
 * type. Then you can access stream data using #vertex3 or other public
 * array objects bound to the locked streams. Call #unlock() to return the
 * streams back to the #VertexArray object.
 *
 * To render the array, call #bind() to enable all active streams, #render()
 * to render them and #unbind() to disable them. Or use standard OpenGL
 * commands directly between #bind() and #unbind() to use enabled arrays.
 * Or use higher-level rendering objects which use these methods.
 *
 * @todo In current implementation arrays are always allocated in system
 *       memory. I think that it should be possible to write more optimal
 *       implementation (transparent array doublebuffering, etc.) with minimal
 *       changes to the API.
 */
class VertexArray
    {
    NO_COPY_CONSTRUCTOR( VertexArray );
    NO_ASSIGNMENT_OPERATOR( VertexArray );

public:

    #define ARRAY_TYPE Massiv::Core::NormalizedArray
    #define STORAGE_TYPE Massiv::Core::UnmanagedStorage

    /**
     * Stream of 2D vectors.
     */
    typedef ARRAY_TYPE< STORAGE_TYPE< OpenGL::Vector2 > > Vector2Stream;

    /**
     * Stream of 3D vectors.
     */
    typedef ARRAY_TYPE< STORAGE_TYPE< OpenGL::Vector3 > > Vector3Stream;

    /**
     * Stream of unsigend byte rgba colors.
     */
    typedef ARRAY_TYPE< STORAGE_TYPE< OpenGL::Color > > ColorStream;

    #undef STORAGE_TYPE
    #undef ARRAY_TYPE

public:

    /**
     * Stream types.
     */
    enum StreamType
        {
        TYPE_VECTOR2    = 0,
            /**< 2D vectors. */

        TYPE_VECTOR3    = 1,
            /**< 3D vectors. */

        TYPE_COLOR      = 2,
            /**< RGBA unsigned byte colors. */

        TYPE_COUNT
        };

    /**
     * Stream targets.
     */
    enum StreamTarget
        {
        VERTEX_ARRAY        = 0,
            /**< Vertex coordinates. Accepts VECTORS2 and VECTORS3 streams. */

        NORMAL_ARRAY        = 1,
            /**< Vertex normals. Accepts VECTORS3 streams. */

        COLOR_ARRAY         = 2,
            /**< Vertex colors. Accepts COLORS streams. */

        TEXCOORD_ARRAY_0    = 3,
            /**< Texture coordinates for unit 0. Accepts VECTORS2 streams. */

        TEXCOORD_ARRAY_1    = 4,
            /**< Texture coordinates for unit 1. Accepts VECTORS2 streams. */

        TARGET_COUNT
        };

public:

    /**
     * Convert texture unit index to StreamTarget.
     */
    static INLINE StreamTarget texcoord_array_target
        (
        std::size_t unit
        );

public:

    /**
     * Constructor.
     */
    VertexArray();

    /**
     * Destructor.
     */
    ~VertexArray();

public:

    /**
     * Choose how many of stream of given @a type will be allocated by #allocate().
     *
     * Initially this is 0 for each stream type.
     *
     * @pre Streams are not allocated.
     */
    INLINE void set_stream_count
        (
        StreamType  type,
        std::size_t count
        );

    /**
     * How many of stream of given @a type exist or will be allocated by #allocate()?
     */
    INLINE std::size_t get_stream_count
        (
        StreamType  type
        ) const;

    /**
     * Set all stream counts to zero.
     */
    void reset_stream_counts();

    /**
     * Allocate streams.
     *
     * Use #select() to bind allocated streams to vertex array targets.
     *
     * @param element_count
     *      Number of elements to allocate in each stream.
     *
     * @note It's possible to call #allocate() when streams are already
     *       allocated - #free() will be called automatically. This means
     *       that contents of the arrays will be lost.
     */
    void allocate
        (
        std::size_t element_count
        );

    /**
     * Free all the streams.
     *
     * Stream counts set by #set_stream_count() will not be changed.
     * Stream selections set by #select() will be forgotten.
     *
     * No streams may be locked while calling #free().
     * Array must not be bound.
     */
    void free();

public:

    //@{
    /**
     * Bind contents of stream of given type and @a index to array @a stream.
     *
     * @pre At least @a index + 1 streams of given type are allocated.
     *      Stream is not locked. Array is not bound.
     */
    void lock_vectors2
        (
        Vector2Stream & stream,
        std::size_t     index
        );

    void lock_vectors3
        (
        Vector3Stream & stream,
        std::size_t     index
        );

    void lock_colors
        (
        ColorStream &   stream,
        std::size_t     index
        );
    //@}

    //@{
    /**
     * Unlock stream of given @a type and @a index.
     *
     * @pre Stream is locked.
     * @post Stream binding set by @c lock_* method is no longer valid.
     */
    void unlock
        (
        StreamType  type,
        std::size_t index
        );

    INLINE void unlock_vectors2
        (
        std::size_t index
        );

    INLINE void unlock_vectors3
        (
        std::size_t index
        );

    INLINE void unlock_colors
        (
        std::size_t index
        );
    //@}

public:

    /**
     * Bind @a index-th stream of type @a type to vertex array @a target.
     *
     * Use this functions to select which streams will be bound by #bind().
     * @a Type must be valid type for @a target:
     * - VERTEX_ARRAY: TYPE_VECTOR2 or TYPE_VECTOR3.
     * - NORMAL_ARRAY: TYPE_VECTOR3.
     * - COLOR_ARRAY: TYPE_COLOR.
     * - @c TEXCOORD_ARRAY_*: TYPE_VECTOR2.
     *
     * @pre Array is not bound.
     */
    void select
        (
        StreamTarget    target,
        StreamType      type,
        std::size_t     index
        );

    /**
     * Unbind bound stream from vertex array @a target.
     */
    void deselect
        (
        StreamTarget    target
        );

public:

    /**
     * Enable bound streams of the array.
     *
     * @pre No streams are locked.
     */
    void bind();

    /**
     * Disable the array.
     */
    void unbind();

public:

    /**
     * Render the subset of the array.
     *
     * This method calls @c glDrawArrays(), see OpenGL documentation for more
     * info about the arguments.
     *
     * Default value of @a first is 0. If @a count is 0 (default), #get_size()
     * will be used instead.
     *
     * @pre Array is bound.
     */
    void render
        (
        GLenum              mode,
        GLint               first = 0,
        GLsizei             count = 0
        );

    /**
     * Render the array using given index array @a indices.
     *
     * This method calls @c glDrawElements() or @c glDrawRangeElements().
     * See OpenGL documentation for more info about the arguments.
     *
     * Indices must be of type GL_UNSIGNED_SHORT. If both @a start and @a end
     * are specified (they are not -1), @c glDrawRangeElements() will be
     * called if driver supports it, otherwise @c glDrawElements() will be
     * called.
     *
     * @pre Array is bound.
     */
    void render
        (
        GLenum              mode,
        const GLushort *    indices,
        GLsizei             count,
        GLint               start = -1,
        GLint               end = -1
        );

public:

    /**
     * Get number of elements in the array
     */
    INLINE std::size_t get_size() const;

public:

    /**
     * Dump currently selected streams of vertex array to given stream.
     *
     * @pre No streams are locked. Array is not bound.
     *
     * @bug Only vertex streams will be dumped.
     */
    void dump
        (
        std::ostream &      stream,
        const std::string & indent
        );

private:

    /**
     * Get element size in bytes of given stream type.
     */
    std::size_t element_size
        (
        StreamType  type
        ) const;

    /**
     * Get offset, in bytes, from the start of #memory to the begining of
     * @a index-th stream of type @a type.
     *
     * Use TYPE_COUNT and 0 to get size of vertex array memory.
     */
    std::size_t stream_offset
        (
        StreamType  type,
        std::size_t index
        ) const;

    /**
     * Get stream index into #locked_streams vector.
     *
     * Use TYPE_COUNT and 0 to get size of the vector.
     */
    std::size_t stream_index
        (
        StreamType  type,
        std::size_t index
        ) const;

    /**
     * Get number of locked streams.
     */
    std::size_t lock_count() const;

private:

    /**
     * Type-index pair.
     */
    struct StreamSelection
        {
        StreamType  type;
        int         index;
        };

    /**
     * Bitfield indicating which streams are locked.
     */
    typedef std::vector< bool > LockedStreams;

private:

    std::size_t         allocated_streams[ TYPE_COUNT ];
        /**< Number of allocated streams of each type.
             If @a memory is @c NULL, streams are not allocated yet. */

    LockedStreams       locked_streams;
        /**< Which streams are currently locked?
             Bitfield ( sum of allocated_streams counts ) bits long.
             First allocated_streams[ 0 ] bits correspond to streams of type 0. */

    StreamSelection     selected_streams[ TARGET_COUNT ];
        /**< Types and indices of selected (active) streams.
             -1 if no stream for given array is selected. */

    UInt8 *             memory;
        /**< Memory used to store the stream.
             All vertex array streams are stored sequentially in single
             memory chunk, in order given by StreamType enum.

             Currently this is always allocated in system memory.

             @c NULL if streams are not allocated. */

    std::size_t         size;
        /**< Number of elements in the array. */

private:

    static VertexArray * bound;
        /**< Currenly bound array. Used to detect error in array usage.
             (possible bug) We assume single rendering context is used. */
    
    }; // class VertexArray


} // namespace Demo::Client::Renderer
} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "vertex_array_inline.h"
#endif


#endif // CLIENT_VERTEX_ARRAY_H
