/** @file client_sdl.cpp
 * Helper SDL functions.
 */

/*
 * begin        : Sun Mar 9 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "client_sdl.h"
#include "client_exception.h"
#include "input/input_manager.h"
#include "client_global.h" // InputManager
#include <SDL_keyboard.h>
#include <sstream>
#include <cctype>


namespace Demo {
namespace Client {
namespace SDL {


// globals
namespace {

bool g_sdl_initialized              = false;
bool g_atexit_registered            = false;
bool g_video_subsystem_initialized  = false;
bool g_video_mode_opened            = false;
int g_screen_width                  = 0;
int g_screen_height                 = 0;

} // anonymous namespace


namespace {

void init_sdl_video_subsystem()
    {
    if( !g_video_subsystem_initialized )
        {
        if( SDL_InitSubSystem( SDL_INIT_VIDEO ) < 0 )
            {
            std::ostringstream ss;
            ss  << "Unable to init SDL video subsystem."
                << " " << SDL_GetError();
            throw ExInitSDLError( ss.str() );
            }

        g_video_subsystem_initialized = true;
        }
    }


void shutdown_sdl_video_subsystem()
    {
    if( g_video_subsystem_initialized )
        {
        SDL_QuitSubSystem( SDL_INIT_VIDEO );
        g_video_subsystem_initialized = false;
        }
    }

} // anonymous namespace


void initialize()
    {
    hope( !g_sdl_initialized );

    unsigned int sdl_flags = 0;

    // Prevent SDL from catching fatal signals,
    // so we will be able to debug the application
    // when an error occurs.
    sdl_flags |= SDL_INIT_NOPARACHUTE;

    if( SDL_Init( sdl_flags ) < 0 )
        {
        std::ostringstream ss;
        ss  << "Unable to initialize SDL library."
            << " " << SDL_GetError();
        throw ExInitSDLError( ss.str() );
        }

    init_sdl_video_subsystem();

    SDL_EnableKeyRepeat
        (
        SDL_DEFAULT_REPEAT_DELAY,
        SDL_DEFAULT_REPEAT_INTERVAL
        );

    if( !g_atexit_registered )
        {
        // FIXME: Using atexit is nasty. Moreover it's actually kinda weird to
        // quit SDL without destroying all the objects using it.
        atexit( &SDL_Quit );
        g_atexit_registered = true;
        }

    g_sdl_initialized = true;
    }


void shutdown()
    {
    if( g_sdl_initialized )
        {
        close_video_mode();
        SDL_Quit();
        g_sdl_initialized = false;
        }
    }


void open_video_mode
    (
    int width,
    int height,
    int bpp
    )
    {
    hope( width > 0 && height > 0 );
    hope( bpp == 0 || bpp == 16 || bpp == 32 );

    bool fullscreen = ( bpp != 0 );

    if( g_video_mode_opened )
        {
        std::ostringstream ss;
        ss  << "Video mode must be closed before opening new video mode.";
        throw ExOpenVideoModeError( ss.str() );
        }

    // Open videmode.

    unsigned int sdl_flags = 0;

    sdl_flags |= SDL_OPENGL;
    sdl_flags |= ( fullscreen ? SDL_FULLSCREEN : 0 );

    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    // Passing bpp != 0 to SDL_SetVideoMode() when trying to
    // open windowed mode leads to random crashes on windows 9x.
    hope( ( fullscreen && bpp != 0 ) || ( !fullscreen && bpp == 0 ) );

    if( SDL_SetVideoMode( width, height, bpp, sdl_flags ) == 0 )
        {
        std::ostringstream ss;
        ss  << "Unable to open video mode "
            << width << "x"<< height << "x" << bpp
            << " (" << ( fullscreen ? "fullscreen" : "windowed" ) << ")."
            << " " << SDL_GetError();
        throw ExOpenVideoModeError( ss.str() );
        }

    // Set window title bar text.

    SDL_WM_SetCaption( "Demo client.", NULL );

    // Hide/show cursor.

    SDL_ShowCursor( SDL_DISABLE );

    // FIXME: Temporary hack. What is the right time to call this
    // function to make key-repeat working? Also default delay seem
    // a bit too long to me and it seems to mess up global keyboard
    // settings under XFree86.
        SDL_EnableKeyRepeat
        (
        SDL_DEFAULT_REPEAT_DELAY,
        SDL_DEFAULT_REPEAT_INTERVAL
        );

    g_video_mode_opened = true;
    g_screen_width = width;
    g_screen_height = height;
    }


void close_video_mode()
    {
    if( g_video_mode_opened )
        {
        shutdown_sdl_video_subsystem();
        g_video_mode_opened = false;
        g_screen_width = 0;
        g_screen_height = 0;
        }
    }


int get_screen_width()
    {
    return g_screen_width;
    }


int get_screen_height()
    {
    return g_screen_height;
    }


void flip()
    {
    if( g_video_mode_opened )
        {
        SDL_GL_SwapBuffers();
        }
    }


bool process_events()
    {
    SDL_Event e;
    InputManager & input_manager = Global::get_input_manager();

    while( true )
        {
        if( !SDL_PollEvent( &e ) )
            break;

        switch( e.type )
            {
            case SDL_KEYDOWN:
                {
                SDL_KeyboardEvent & event = e.key;
                SDL_keysym          key = event.keysym;

                input_manager.key_down( key );
                break;
                }

            case SDL_KEYUP:
                {
                SDL_KeyboardEvent & event = e.key;
                SDL_keysym          key = event.keysym;

                input_manager.key_up( key );
                break;
                }

            case SDL_MOUSEBUTTONDOWN:
                {
                SDL_MouseButtonEvent & event = e.button;
                int                    button = event.button;

                input_manager.mouse_button_down( button );
                break;
                }

            case SDL_MOUSEBUTTONUP:
                {
                SDL_MouseButtonEvent & event = e.button;
                int                    button = event.button;

                input_manager.mouse_button_up( button );
                break;
                }

            case SDL_MOUSEMOTION:
                {
                SDL_MouseMotionEvent & event = e.motion;
                int                    dx = event.xrel;
                int                    dy = event.yrel;

                if( dx != 0 || dy != 0 )
                    {
                    input_manager.mouse_move( dx, dy );
                    }
                break;
                }

            case SDL_ACTIVEEVENT:
                {
                // Warning: e.active.state is a bit combination
                // of SDL_APPMOUSEFOCUS, SDL_APPINPUTFOCUS and SDL_APPACTIVE.
                break;
                }

            case SDL_QUIT:
                return true;

            default:
                // Skip this event.
                break;
            }
        }

    return false;
    }


int key_to_ascii_character
    (
    SDLKey  key,
    bool    caps,
    bool    shift
    )
    {
    // Check keypad.
    switch( key )
        {
        case SDLK_KP0:          return '0';
        case SDLK_KP1:          return '1';
        case SDLK_KP2:          return '2';
        case SDLK_KP3:          return '3';
        case SDLK_KP4:          return '4';
        case SDLK_KP5:          return '5';
        case SDLK_KP6:          return '6';
        case SDLK_KP7:          return '7';
        case SDLK_KP8:          return '8';
        case SDLK_KP9:          return '9';
        case SDLK_KP_PERIOD:    return '.';
        case SDLK_KP_DIVIDE:    return '/';
        case SDLK_KP_MULTIPLY:  return '*';
        case SDLK_KP_MINUS:     return '-';
        case SDLK_KP_PLUS:      return '+';
        case SDLK_KP_EQUALS:    return '=';
        default:
            break;
        }

    // SDL keys are mapped to lowercase ascii characters.
    if( key >= 32 && key < 127 )
        {
        int c = key;

        if( ( shift || caps ) && !( shift && caps ) )
            {
            c = std::toupper( c );
            }

        if( shift && c == key )
            {
            switch( c )
                {
                case '1':   c = '!'; break;
                case '2':   c = '@'; break;
                case '3':   c = '#'; break;
                case '4':   c = '$'; break;
                case '5':   c = '%'; break;
                case '6':   c = '^'; break;
                case '7':   c = '&'; break;
                case '8':   c = '*'; break;
                case '9':   c = '('; break;
                case '0':   c = ')'; break;
                case '-':   c = '_'; break;
                case '=':   c = '+'; break;

                case '[':   c = '{'; break;
                case ']':   c = '}'; break;
                case ';':   c = ':'; break;
                case '\'':  c = '\"'; break;
                case '\\':  c = '|'; break;
                case ',':   c = '<'; break;
                case '.':   c = '>'; break;
                case '/':   c = '?'; break;
                default:
                    break;
                }
            }

        return c;
        }
    else
        {
        return 0;
        }
    }


bool is_sdl_initialized()
    {
    return g_sdl_initialized;
    }


bool is_video_mode_opened()
    {
    return g_video_mode_opened;
    }


} // namespace SDL
} // namespace Client
} // namespace Demo
