/** @file tag_database.cpp
 * Database of string tags and tag matching (implementation).
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "tag_database.h"
#include "client_exception.h"
#include "global/global.h"
#include "data/data_manager.h"
#include "utility/quote_string.h"
#include "system/system.h"
#include <sstream>


namespace Demo {
namespace Client {


namespace {

const std::string COLON = ":";

}


// class TagSet

TagSet::TagSet
    (
    const std::string & tag_string
    )
    {
    std::istringstream tag_stream( tag_string );
    ::Massiv::Core::TextReader tr( tag_stream );

    if( !tr.next_line() )
        {
        // Emptry string -> no tags.

        return;
        }

    while( 1 )
        {
        // Copied from read_quote_string() - we don't want those eol exceptions
        // to pollute stdout and logs.

        std::string tag;
        tr.line() >> ::Massiv::Core::quote_string( tag );
        if( tr.line().fail() )
            {
            if( tr.line().eof() )
                {
                break;
                }
            else
                {
                throw ::Massiv::Core::ExUnexpectedToken( tr.unexpected_token( "quoted string" ) );
                }
            }

        insert( tag );
        }
    }


// class TagDatabase::Mapping

TagDatabase::Mapping::Mapping
    (
    const std::string & mapping,
    std::size_t         value_count
    ) : values( value_count )
    {
    std::istringstream mapping_stream( mapping );
    ::Massiv::Core::TextReader tr( mapping_stream );
    tr.next_line_throw();

    // Read key tag set.

    while( 1 )
        {
        const std::string tag = tr.read_quote_string();
        if( tag == COLON )
            {
            break;
            }
        key.insert( tag );
        }

    // Read value strings.

    std::size_t i;
    for( i = 0; i < value_count; i++ )
        {
        values[ i ] = tr.read_quote_string();
        }
    }


bool TagDatabase::Mapping::matches
    (
    const TagSet & tag_set
    ) const
    {
    TagSet::const_iterator it;
    for( it = key.begin(); it != key.end(); ++it )
        {
        if( tag_set.find( *it ) == tag_set.end() )
            {
            // #key contains tag not in @a tag_set -> not a subset.

            return false;
            }
        }

    return true;
    }


// class TagDatabase

TagDatabase::TagDatabase
    (
    std::size_t the_value_count
    ) : value_names( the_value_count ),
        value_count( the_value_count )
    {
    }


TagDatabase::TagDatabase
    (
    const std::string & name
    )
    {
    ::Massiv::Core::DataManager & data_manager = ::Massiv::Core::Global::data_manager();
    data_object.reset( data_manager.get_data_object( name ) );
    load();
    }


void TagDatabase::load
    (
    ::Massiv::Core::TextReader & tr
    )
    {
    mappings.clear();

    // Header.

    tr.next_line_throw();
    tr.require_string( "tag_database" );
    const int version = tr.read_int();
    if( version != 1 && version != 2 )
        {
        std::ostringstream oss;
        oss << "Unexpected tag database version, expected 1 or 2, found " << version;
        throw ExTagDatabaseError( oss.str() );
        }

    // Value names.

    if( version == 1 )
        {
        using namespace Massiv::Core;
        std::ostringstream oss;
        oss << "Loading old tag database version, update data files to prevent "
               "unexpected problems.";
        Massiv::Core::Global::log( Status::warning( Status::FACILITY_GENERIC,
                    Status::PRIORITY_HIGH ), oss.str() );

        value_count = 1;
        value_names.push_back( "unnamed" );
        }
    else
        {
        // Read count.

        tr.next_line_throw();
        const int count = tr.read_int();
        if( count < 1 )
            {
            std::ostringstream oss;
            oss << "Bad value count " << count;
            throw ExTagDatabaseError( oss.str() );
            }
        value_count = static_cast< std::size_t >( count );
        value_names.resize( value_count );

        // Read names.

        std::size_t i;
        for( i = 0; i < value_count; i++ )
            {
            value_names[ i ] = tr.read_quote_string();
            }
        }

    // Mappings.

    while( tr.next_line() )
        {
        const std::string line = tr.str();
        mappings.push_back( Mapping( line, value_count ) );
        }
    }


bool TagDatabase::refresh()
    {
    bool updated = false;

    // Reload database if we're using #data_object and it has been updated.

    if( data_object.get() )
        {
        if( data_object->is_data_updated() )
            {
            data_object->refresh();
            load();
            updated = true;
            }
        data_object->download_actual_version();
        }

    return updated;
    }


std::size_t TagDatabase::get_value_index
    (
    const std::string & value_name
    ) const
    {
    hope( value_count == value_names.size() );

    std::size_t i;
    for( i = 0; i < value_count; i++ )
        {
        if( value_names[ i ] == value_name )
            {
            return i;
            }
        }

    std::ostringstream oss;
    oss << "There is no value with name '" << value_name << "'.";
    throw ExTagDatabaseError( oss.str() );
    /*NOTREACHED*/
    }


TagDatabase::Match TagDatabase::match
    (
    const TagSet & tag_set
    ) const
    {
    std::size_t best_count = 0;
    Match best_match = mappings.end();

    Mappings::const_iterator it;
    for( it = mappings.begin(); it != mappings.end(); ++it )
        {
        if( it->matches( tag_set ) )
            {
            if( ( it->key.size() > best_count ) || ( best_match == mappings.end() ) )
                {
                best_count = it->key.size();
                best_match = it;
                }
            }
        }

    return best_match;
    }


std::string TagDatabase::get_value
    (
    Match               mapping_iterator,
    const std::string & value_name
    ) const
    {
    if( !match_found( mapping_iterator ) )
        {
        throw ExTagDatabaseError( "Tried to get_value() of invalid mapping." );
        }
    const std::size_t index = get_value_index( value_name );
    return mapping_iterator->values[ index ];
    }


void TagDatabase::load()
    {
    hope( data_object.get() );
    data_object->open();
    ::Massiv::Core::TextReader tr( data_object->data );
    load( tr );
    data_object->close();
    }


} // namespace Demo::Client
} // namespace Demo


#ifndef ENABLE_INLINE
#include "tag_database_inline.h"
#endif
