/** @file tag_database.h
 * Database of string tags and tag matching.
 */

/*
 * begin        : Oct 2003
 * copyright    : (C) 2003 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef CLIENT_TAG_DATABASE_H
#define CLIENT_TAG_DATABASE_H


#ifndef CLIENT_COMMON_H
#include "client_common.h"
#endif

#ifndef MASSIV_TEXT_READER_H
#include "utility/text_reader.h"
#endif

#ifndef MASSIV_HASH_H
#include "utility/hash.h"
#endif


#include <memory>
#include <string>
#include <vector>
#include <list>
#include <iostream>


namespace Massiv {
namespace Core {

class DataObject;

}
}


namespace Demo {
namespace Client {


/**
 * String tag.
 */
typedef std::string Tag;


/**
 * Set of tags.
 */
class TagSet : public ext::hash_set< std::string >
    {
public:

    /// Superclass.
    typedef ext::hash_set< std::string > Super;

public:

    /**
     * Default constructor.
     */
    INLINE TagSet();

    /**
     * Copy constructor.
     */
    INLINE TagSet
        (
        const TagSet & other
        );

    /**
     * Construct TagSet from string containing tags separated by whitespace.
     *
     * Each tag is either plain word (string with no whitespace) or quoted
     * string.
     *
     * @exception ExUnexpectedToken Thrown if @a tag_string contains non-string token.
     */
    TagSet
        (
        const std::string & tag_string
        );

    /**
     * Assignment operator.
     */
    INLINE TagSet & operator=
        (
        const TagSet & other
        );
    
    }; // class TagSet

    
/**
 * Database (list) of tag mappings.
 */
class TagDatabase
    {
    NO_COPY_CONSTRUCTOR( TagDatabase );
    NO_ASSIGNMENT_OPERATOR( TagDatabase );

public:

    /**
     * Vector of strings.
     */
    typedef std::vector< std::string > Strings;

public:

    /**
     * Tag mapping.
     *
     * This structure maps set of tags (key) to several strings (values).
     * The number of value strings is constant per TagDatabase.
     */
    class Mapping
        {
    public:

        /**
         * Default constructor.
         */
        INLINE Mapping();

        /**
         * Construct mapping from @a mapping string as described in #TagDatabase::load().
         *
         * @param mapping String containing the mapping.
         * @param value_count Expected number of mapping "values".
         *
         * @exception ExUnexpectedToken Thrown if @a tag_string contains non-string token.
         * @exception ExUnexpectedEol Thrown if end of @a mapping string is reached
         *            before colon or @a value_count values are parsed.
         */
        Mapping
            (
            const std::string & mapping,
            std::size_t         value_count
            );

        /**
         * Constructor.
         */
        INLINE Mapping
            (
            const TagSet &      the_key,
            const Strings &     the_values
            );

    public:

        /**
         * Does given @a tag_set match #key of this mapping?
         *
         * The @a tag_set matches the mapping if #Mapping::key tag set is subset
         * of @a tag_set.
         */
        bool matches
            (
            const TagSet & tag_set
            ) const;

    public:

        TagSet  key;
            /**< "Key" tag set. */

        Strings values;
            /**< Mapping values - vector of strings.
                 Each mapping should have the same number of value strings. */
            
        }; // class Mapping

public:

    /**
     * List of mappings.
     */
    typedef std::list< Mapping > Mappings;

    /**
     * Result of #match() - constant iterator to Mappings.
     */
    typedef Mappings::const_iterator Match;

public:

    /**
     * Construct empty database.
     *
     * The database won't be bound to data object. You must initialize
     * the list of mappings manually - access the #mappings and other variables
     * directly - or call #load() to load database from a stream. If you modify
     * the database manually, make sure that all mappings have @a the_value_count
     * value strings.
     *
     * @param the_value_count Number of values per mapping. #value_names will be
     *                        resized to @a the_value_count elements.
     */
    TagDatabase
        (
        std::size_t the_value_count
        );

    /**
     * Construct database from data object.
     *
     * Call #refresh() to ensure that the database is up-to-date. Do not modify
     * the database manually.
     */
    TagDatabase
        (
        const std::string & name
        );

public:

    /**
     * Load database using text reader @a tr.
     *
     * Tag database is standard line-oriented file. The first line should be
     * "tag_database 2" (where "2" is file version). The second line should
     * contain number of values of each mapping and value names. Single mapping
     * per line follows:
     *
     * @a key : @a value
     *
     * @a Key is tag set - string tags (quoted if they contain weird characters),
     * separated by whitespace. @a Values is list of string (may be quoted too),
     * their number must be the same for each mapping. The @a key can't contain
     * colon tag. The first separate color marks end of @a key. Anything
     * following expected number of @a value strings will be ignored.
     *
     * Example database:
     * @verbatim
tag_database 2
2 model name
player : model/player_default Player
goblin : model/goblin_default "Ordinary goblin"
goblin red : model/goblin_red "Big red goblin"
@endverbatim
     */
    void load
        (
        ::Massiv::Core::TextReader & tr
        );

    /**
     * Reload database from data object if it has been updated.
     *
     * Returns true if the database was updated.
     */
    bool refresh();

public:

    /**
     * Get number of values per mapping and size of #value_names.
     */
    INLINE std::size_t get_value_count() const;

    /**
     * Convert value name (type) to index.
     *
     * @return Index @a such that <tt>value_names[ i ] == value_name</tt>.
     *         Throws ExTagDatabaseError if there is no such index.
     */
    std::size_t get_value_index
        (
        const std::string & value_name
        ) const;

public:

    /**
     * Match given @a tag_set.
     *
     * Mapping @a mapping matches the @a tag_set iff @a mapping.tag_set is
     * subset of @a tag_set. This method return iterator to matching @a mapping
     * with the biggest @a mapping.tag_set.
     *
     * Example - the database contains the following mappings:
     * -# player : model/player_default
     * -# player red : model/player_red
     * -# goblin red : model/goblin_red
     *
     * The followin list shows which tag sets will match which mappings:
     * - player: 1.
     * - player red angry: 2.
     * - goblin red sleepy: 3.
     * - player blue: 1.
     * - goblin blue: none.
     * - dragon red: none.
     *
     * "goblin blue" does not match any mapping in the database, "goblin red" is
     * not subset of "goblin blue". The subset rule is important - it prevents
     * two possibly unrelated tag sets to match (if "goblin blue" matched
     * "goblin red", "dragon red" would match either "player red" or "goblin red").
     * The database should always contain generic mappings such as:
     * - goblin : model/goblin_red (if no generic goblin model is available)
     * - dragon : model/anything_resembling_dragon
     *
     * @return Iterator to best matching mapping or #mappings_end() if no mapping matches.
     *         "Best" mapping is mapping with the biggest "key" tag_set.
     */
    Match match
        (
        const TagSet & tag_set
        ) const;

    /**
     * Get #mappings_end().
     */
    INLINE Match mappings_end() const;

    /**
     * Check if iterator returned by #match() is invalid.
     */
    INLINE bool match_found
        (
        Match mapping_iterator
        ) const;

    /**
     * Get mapping value for given @a value_name.
     *
     * @exception ExTagDatabaseError Thrown if @a mapping_iterator is invalid or
     *                               if no value named @a value_name exists.
     */
    std::string get_value
        (
        Match               mapping_iterator,
        const std::string & value_name
        ) const;

private:

    /**
     * Load database from #data_object.
     */
    void load();

public:

    Strings                                     value_names;
        /**< Names (types) of values. */

    Mappings                                    mappings;
        /**< List of mappings. */

private:

    std::size_t                                 value_count;
        /**< Number of values per mapping and #value_names size. */

    std::auto_ptr< ::Massiv::Core::DataObject > data_object;
        /**< Database data object. May be @c NULL if database is not bound to
             a data object. */

    }; // class TagDatabase


} // namespace Demo::Client
} // namespace Demo


#ifdef ENABLE_INLINE
#include "tag_database_inline.h"
#endif


#endif // CLIENT_TAG_DATABASE_H
