/** @file client_node.cpp
 * Client node object implementation.
 */

/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/client/client_node.h"
#include "lib/client/client_node_rpc.h"
#include "lib/client/local_console.h"
#include "lib/shared/account_object_rpc.h"
#include "lib/shared/node_object_rpc.h"
#include "client/client_main.h"
#include "client/client_global.h"
#include "client/gui/client_full_console.h"
#include "client/gui/client_chat_console.h"
#include "client/gui/color_string.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();
    
namespace {

/** Tests if @c text is contained in the @c array. */
bool contains
    (
    const PValueArray<SString> &    array,
    const std::string &             text
    )
    {
    std::size_t i;

    for( i = 0; i < array.size(); i++ )
        {
        if( array[ i ] == text )
            {
            return true;
            }
        }

    return false;
    }


/** Merges arrays into one big array. Drops duplicate strings. */
PValueArray<SString> concatenate
    (
    const PValueArray<SString> & array1,
    const PValueArray<SString> & array2,
    const PValueArray<SString> & array3
    )
    {
    PValueArray<SString>    result;
    std::size_t             i;

    result = array1;

    for( i = 0; i < array2.size(); i++ )
        {
        if( !contains( result, array2[ i ] ) )
            {
            result.push_back( array2[ i ] );
            }
        }
    
    for( i = 0; i < array3.size(); i++ )
        {
        if( !contains( result, array3[ i ] ) )
            {
            result.push_back( array3[ i ] );
            }
        }

    return result;
    }

}
    

ClientNode::ClientNode() :
    local_console( NULL )
    {
    }


ClientNode::~ClientNode()
    {
    delete local_console;
    }


void ClientNode::initialize()
    {
    NodeObject::initialize();
    
    local_console = new LocalConsole();
    managed_node = get_object_id();
    connect_time = Massiv::System::system_time();
    
    update_console_prompt();
    check_if_linked_to_account();
    }


ClientNode::NodeType ClientNode::get_node_type() const
    {
    return NodeId::TYPE_CLIENT;
    }


void ClientNode::welcome_to_simulation
    (
    WeakPointer< AccountObject > the_account
    )
    {
    account = the_account;
    
    Demo::Client::Main::welcome_to_simulation( account );
    }


void ClientNode::player_selected
    (
    WeakPointer< ClientEntity > the_player
    )
    {
    Demo::Client::Main::player_selected( the_player );
    }


void ClientNode::handle_game_message
    (
    WeakPointer< ClientEntity > source_entity,
    const std::string &         message,
    const SColor &              message_color
    )
    {
    print_to_console( null, message, message_color );
    print_to_chat_window( message, message_color );
    }

    
void ClientNode::print_to_console
    (
    WeakPointer<NodeObject> sender,
    const std::string &     message,
    const SColor &          message_color
    )
    {
    Demo::Client::ColorString   cs;
    ObjectId                    sender_id = sender.get_object_id();
    
    // Write sender if it is not a local node.

    if( !sender_id.is_null() && sender_id != get_object_id() )
        {
        const std::size_t   line_size_limit = 60;
        std::ostringstream  ss;

        ss << sender_id << ":";

        if( message.size() > line_size_limit || message.find_first_of( '\n' ) != 
            std::string::npos )
            {
            ss << '\n';
            }
        else
            {
            ss << ' ';
            }

        cs.set_color( SColor( SColor::GREEN ) );
        cs += ss.str();
        }

    // Append the actual message.

    cs.set_color( message_color );
    cs += message;

    // Send the message to console.

    Demo::Client::Global::get_full_console().print( cs );
    }


void ClientNode::print_to_chat_window
    (
    const std::string &         message,
    const SColor &              message_color
    )
    {
    Demo::Client::ColorString   cs;

    cs.set_color( message_color );
    cs += message;

    Demo::Client::Global::get_chat_console().print( cs );
    }


void ClientNode::process_console_command
    (
    const std::string & command
    )
    {
    PValueArray<SString>    errors;
    PValueArray<SString>    local_errors;
    PValueArray<SString>    node_errors;
    PValueArray<SString>    account_errors;
    std::size_t             i;

    // Process command by client's local console first.

    local_console->process_command
        (
        account,
        command,
        false,
        local_errors,
        environment
        );

    // Process command by the node the console is connected to.

    if( managed_node == Pointer<NodeObject>( this ) )
        {
        // Call client's node object directly.

        node_errors = process_node_command( account, command, environment );
        }
    else
    if( account != null )    
        {
        // Forward the call to node object via client's account object.

        node_errors = account->sync_process_node_command
            ( managed_node, command, environment );
        }

    // Process command by @c account.

    if( account != null )
        {
        account_errors = account->sync_process_account_command( command, environment );
        }

    // If there is no console object that processed the command without
    // an error then print all errors that were logged by consoles that
    // recognized the command.

    if( local_errors.size() && node_errors.size() && account_errors.size() )
        {
        bool recognized_by_local = 
            !Console::is_unknown_command_error( local_errors );
        bool recognized_by_node = 
            !Console::is_unknown_command_error( node_errors );
        bool recognized_by_account = 
            !Console::is_unknown_command_error( account_errors );
            
        if( !recognized_by_local )
            {
            local_errors.clear();
            }
        
        if( !recognized_by_node )
            {
            node_errors.clear();
            }
        
        if( !recognized_by_account )
            {
            account_errors.clear();
            }

        errors = concatenate( local_errors, node_errors, account_errors );

        if( !errors.size() )
            {
            errors.push_back( "Unknown command." );
            }

        for( i = 0; i < errors.size(); i++ )
            {
            print_to_console( null, errors[ i ], SColor::RED );
            }
        }

    // Print returned value.
    Console::Environment::Iterator it = environment.find( std::string( "$" ) );
    if( it != environment.end() )
        {
        const std::string & temp_str = it->second;
        if( temp_str.size() )
            {
            print_to_console( null, std::string( "Returned value: " ) + temp_str, 
                SColor::GREEN );
            }
        }
    }


PValueArray<SString> ClientNode::auto_complete_console_command
    (
    const std::string & partial_command
    )
    {
    PValueArray<SString>    local_completions;
    PValueArray<SString>    node_completions;
    PValueArray<SString>    account_completions;

    // Complete command by client's local console first.

    local_console->process_command
        (
        account,
        partial_command,
        true,
        local_completions,
        environment
        );

    // Complete command by the node the console is connected to.

    if( managed_node == Pointer<NodeObject>( this ) )
        {
        // Call client's node object directly.

        node_completions = auto_complete_node_command
            ( account, partial_command );
        }
    else
    if( account != null )    
        {
        // Forward the call to node object via client's account object.

        node_completions = account->sync_auto_complete_node_command
            ( managed_node, partial_command );
        }

    // Complete command by @c account.

    if( account != null )
        {
        account_completions = account->sync_auto_complete_account_command
            ( partial_command );
        }

    // Return completions.

    return concatenate( local_completions, node_completions, account_completions );
    }
    

void ClientNode::connect_console_to
    (
    const ObjectId & object_id
    )
    {
    std::ostringstream ss;

    // Probe the requested node.

    bool can_connect = false;

    try
        {
        if( account != null )
            {
            can_connect = account->sync_can_connect_console_to( object_id );
            }
        }
    catch( Exception & )
        {
        }

    if( !can_connect )
        {
        ss << "Can not connect console to " << object_id << ".";
        print_to_console( this, ss.str(), SColor( SColor::RED ) );

        return;
        }

    // Change console connection.

    managed_node = object_id;
    update_console_prompt();

    // Done.

    ss << "You are connected to " << object_id << " now.";
    print_to_console( this, ss.str(), SColor( SColor::RED ) );
    }


void ClientNode::update_console_prompt()
    {
    std::ostringstream ss;

    ss << managed_node.get_object_id() << "> ";
    Demo::Client::Global::get_full_console().set_prompt( ss.str() );
    }


void ClientNode::check_if_linked_to_account()
    {
    Remote<ClientNode> self = this;

    if( account == null )
        {
        if( connect_time + Demo::Client::Global::account_timeout <
            Massiv::System::system_time() )
            {
            // Quit from the simulation (unmanaged exception must be thrown).

            throw ExLibError( "Unable to contact account object." );
            }
        else
            {
            // Schedule next recheck.
            // Can not be scheduled to account timeout once only since
            // the simulation time may not be synchronized with servers yet.

            self->param( RPC_DELAYED, 1.0f ).async_check_if_linked_to_account();
            }
        }
    }


} // namespace Demo::Lib
} // namespace Demo
