/** @file client_node.h
 * Client node object.
 */

/*
 * begin        : Thu Feb 27 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CLIENT_NODE_H
#define DEMO_CLIENT_NODE_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_CLIENT_FWD_H
#include "lib/client/client_fwd.h"
#endif

#ifndef DEMO_CLIENT_NODE_INTERFACE_H
#include "lib/shared/client_node_interface.h"
#endif

#ifndef DEMO_CONSOLE_H
#include "lib/shared/console.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * NodeObject for a client node.
 */
class ClientNode : public ClientNodeInterface
    {
public:

    //
    // NodeObjectInterface implementation.
    //

    /**
     * Constructor.
     */
    ClientNode();

    /**
     * Destructor.
     */
    ~ClientNode();

    /**
     * Initializes object.
     */
    void initialize();

    /**
     */
    virtual NodeType get_node_type() const;
    
public:

    //
    // ClientNodeInterface implementation.
    //
    
    // From ClientNodeInterface
    virtual void welcome_to_simulation
        (
        WeakPointer< AccountObject > account
        );
    
    // From ClientNodeInterface
    virtual void player_selected
        (
        WeakPointer< ClientEntity > player
        );

    // From ClientNodeInterface
    virtual void handle_game_message
        (
        WeakPointer< ClientEntity > source_entity,
        const std::string &         message,
        const SColor &              message_color
        );

    // From ClientNodeInterface
    virtual void print_to_console
        (
        WeakPointer<NodeObject>     sender,
        const std::string &         message,
        const SColor &              message_color = SColor::YELLOW
        );
    
    // From ClientNodeInterface
    virtual void print_to_chat_window
        (
        const std::string &         message,
        const SColor &              message_color = SColor::YELLOW
        );

public:

    //
    // Console.
    //

    /**
     * Processes @c command.
     * The command is first processed by @c client_console, then
     * by @c managed_node and finally by @c account object.
     */
    void process_console_command
        (
        const std::string & command
        );

    /**
     * Returns completions for @c partial_command.
     * The result is a concatenation of command completions returned
     * by @c client_console, @c managed_node and @c account object.
     */
    PValueArray<SString> auto_complete_console_command
        (
        const std::string & partial_command
        );
    
    /**
     * Connect client's console to given node, specified by its @c object_id.
     * By default console is connected to self.
     */
    void connect_console_to
        (
        const ObjectId & object_id
        );

private:

    /**
     * Updates console prompt.
     */
    void update_console_prompt();

public:

    /**
     * Checks if client node is linked to an account object and quits from
     * the simulation if it is yet unlinked and timeout expired.
     * Otherwise another check is scheduled to be performed soon.
     * Can not rely on simulation time being synchronized with servers.
     */
    void check_if_linked_to_account();

public:

    PRemote< AccountObject >    account;
        /**< Pointer to client's account object. */
    PRemote< NodeObject >       managed_node;
        /**< Node client's console is connected to. */
    LocalConsole *              local_console;
        /**< Console that implements client-specific (mostly console-related)
             commands. */
    Console::Environment        environment;
        /**< Console environment. */
    SystemTime                  connect_time;
        /**< When node object was created. */
    
    }; // ClientNode

} // namespace Demo::Lib
} // namespace Demo


#endif // DEMO_CLIENT_NODE_H
