/** @file local_console.cpp
 * Local console implementation.
 */

/*
 * begin        : Sat Jun 28 2003
 * copyright    : (C) 2003 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/client/local_console.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


LocalConsole::LocalConsole()
    {
    ConsoleCommand * cmd;

    // Command 'help'.

    cmd = command_list.new_command
        (
        "help",
        "HELP.",
        0,
        &execute_help
        );
    
    cmd->new_argument
        (
        "Command to show detail info about the console usage.",
        CommandArgument::FLAGS_STRING | CommandArgument::FLAGS_ARG_OPTIONAL
        );
    
    
    // Command 'quit'.

    cmd = command_list.new_command
        (
        "quit",
        "Quit client.",
        0,
        &execute_quit
        );
    
    
    // Command 'cls'.

    cmd = command_list.new_command
        (
        "cls",
        "Clear client's console display.",
        0,
        &execute_cls
        );


    // Command 'connect'.

    cmd = command_list.new_command
        (
        "connect",
        "Connect console to a given node.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_connect
        );

    cmd->new_argument
        (
        "Node to connect to.",
        CommandArgument::FLAGS_OBJECT_ID
        );
        
    // Command 'editor'.
    
    cmd = command_list.new_command
        (
        "editor",
        "Show/hide editor.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_editor
        );

    // Command 'exec'.

    cmd = command_list.new_command
        (
        "exec",
        "Exec local script.",
        0,
        &execute_exec
        );

    cmd->new_argument
        (
        "Command.",
        CommandArgument::FLAGS_STRING
        );

    // Command 'writefile'.

    cmd = command_list.new_command
        (
        "writefile",
        "Writes console output to a file.",
        0,
        &execute_writefile
        );
    
    cmd->new_argument
        (
        "Filename to write console output to.",
        CommandArgument::FLAGS_STRING
        );

    // Initialize script list from script data object.
    initialize_script_list();
    }


LocalConsole::~LocalConsole()
    {
    delete script_do;
    }


void LocalConsole::initialize_script_list()
    {
    Massiv::Core::DataManager & dm = Massiv::Core::Global::data_manager();
    script_do = dm.get_data_object( "script_list" );
    hope( script_do != NULL );
    script_do->download_actual_version();
    
    update_script_list();
    }


void LocalConsole::update_script_list()
    {
    hope( script_do != NULL );
    script_do->open();
    
    std::istream &              is = script_do->data;
    Massiv::Core::TextReader    tr( is );
    std::string                 token;
    
    script_list.clear();
    
    while( tr.next_line() )
        {
        tr.line() >> token;
        if( token != "script" )
            {
            std::ostringstream  ss;
            ss << "Error in data file \"script\", line " << tr.line_number();
            throw Massiv::Core::ExLibError( ss.str() );
            }
        // Load one script.        
        Script script;
        tr.line() >> script.name;
        while( tr.next_line() && tr.str() != "endscript" )
            {
            script.command_list.push_back( tr.str() );
            }
        script_list.push_back( script );
        }
        
    script_do->close();
    }


bool LocalConsole::find_script
    (
    const std::string &     script_name,
    LocalConsole::Script &  script
    )
    {
    // Try to update script data object if possible.
    if( script_do->is_data_updated() )
        {
        script_do->refresh();
        update_script_list();
        }
    std::vector< Script >::iterator it = script_list.begin();
    while( it != script_list.end() )
        {        
        if( it->name == script_name )
            {
            script = *it;
            return true;
            }
        ++it;
        }
    return false;
    }
    
// Definition of static attributes.

std::vector< LocalConsole::Script > LocalConsole::script_list;
Massiv::Core::DataObject *          LocalConsole::script_do;
    

} // namespace Demo::Lib
} // namespace Demo
