/** @file account_object_impl.cpp
 * Account object implementation.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/account_object_impl.h"
#include "lib/server/account_console.h"
#include "lib/shared/client_node_interface.h"
#include "lib/shared/client_node_interface_rpc.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/client_light.h"
#include "lib/server/sector_helper_objects.h"
#include "lib/server/account_object_impl_rpc.h"
#include "lib/server/sector.h"
#include "lib/server/sector_rpc.h"
#include "lib/server/remote_console_command.h"
#include "lib/server/entity_rpc.h"
#include "lib/server/generator.h"
#include "lib/server/generator_rpc.h"
#include "lib/server/entity_factory_rpc.h"

// Initial world creation.
#include "lib/server/world.h"
#include "lib/server/layer.h"
#include "lib/server/player.h"
#include "lib/server/player_rpc.h"
#include "lib/server/robot.h"
#include "lib/server/potion.h"
#include "lib/server/boots.h"
#include "lib/server/weapon.h"
#include "lib/server/naming_service.h"
#include "lib/server/naming_service_rpc.h"
#include "lib/server/entity_factory.h"
#include "lib/shared/random.h" // REMOVE

// Server global.
#include "../../server/server_global.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


namespace {

const Float32 nearby_sectors_replication_duration = 10.0f;
    /**< How long in seconds nearby sectors are replicated to a client. */
const Float32 replication_big_period = 2.0f;
    /**< How often replication of nearby sectors is requested. */
const Float32 replication_small_period = 1.0f;
    /**< If replication of nearby sectors can't be requested because the player
         is migrating to a remote sector, this value is used to schedule
         the next replication tick. */
const Int32   replication_neighborhood = 2;
    /**< Player's neighborhood size to be replicatated to a client. */

} // anonymous namespace


void AccountObjectImpl::initialize
    (
    bool the_root_account
    )
    {
    root_account = the_root_account;

        {
        std::cout
            << "**** Created "
            << ( root_account ? "ROOT " : "user " )
            << "account object. Object id: " << get_object_id() << "."
            << std::endl;
        }

    replication_tick_pending = false;
    }


void AccountObjectImpl::object_updated
    (
    Object::UpdateReason    reason
    )
    {
    if( reason == SIMULATION_STARTUP )
        {
        client_disconnected();
        }
    }


void AccountObjectImpl::client_connected
    (
    const WeakPointer< ::Massiv::Core::Lib::NodeObjectInterface > &   node_object
    )
    {
    client = node_object.convert();

    // Send welcome message to the client.
    client->async_welcome_to_simulation( this );

    // Create player object if the client connects for the first time.
    if( player == null )
        {
        create_player();
        }

    // Notify the client that its player was selected.
    // Because player_selected() must be delivered to client node object
    // _after_ welcome_to_simulation(), we must delay the player_selected()
    // a little, because system doesn't guarantee that two async calls
    // scheduled for same time will be delivered in order they are called!
    if( player != null )
        {
        client->param( RPC_DELAYED, STime( 0.1 ) ).async_player_selected
            ( player->get_client_entity() );
        }

    // Note for future:
    // Currently player and account are linked with strong pointer.
    // This ensures that the account migrates with the player.
    // It forces the client node to dynamically create connetion to the
    // server node which owns the player as the account migrates
    // with it.
    //
    // But if we want to support multiple players per one account,
    // we must make sure that the account is linked only to _one_
    // player at a time!
    // Implementation should be something like if new player is selected,
    // unlink from current player, migrate to new player and link
    // with it.
    // Note: It is not good idea to stay all players linked with the
    // account and manager linking/unlinking the players from world
    // depending on selected player, because that would put all players
    // into one migration group and force them to replicate to all clients.
    }


void AccountObjectImpl::client_disconnected()
    {
    client = null;
    }


bool AccountObjectImpl::check_client_rpc_request
    (
    const Pointer< ::Massiv::Core::RPCObject > & request
    ) const
    {
    hope( request->node_object != null );

    if( request->node_object != client )
        {
        // Request from different client.

        return false;
        }

    if( client == null )
        {
        // Client has disconnected meanwhile.

        return false;
        }

    return true;
    }


void AccountObjectImpl::log_system_messages
    (
    const WeakPointer< ::Massiv::Core::Lib::NodeObjectInterface > & sender,
    const PValueArray< SString > &                                  messages
    )
    {
    std::size_t i;

    for( i = 0; i < messages.size(); i++ )
        {
        print_to_console( sender.convert(), messages[ i ] );
        }
    }


void AccountObjectImpl::request_player_replication
    (
    WeakPointer< ClientEntity > the_player
    )
    {
    // Check if client is still connected.
    if( !is_client_connected() || !is_player_selected() )
        {
        return;
        }

    // Start replicating nearby sectors (including selected player).
    start_replication_of_nearby_sectors();
    }


void AccountObjectImpl::handle_move_action
    (
    const PlayerMoveAction & action
    )
    {
    // Still connected?
    if( !is_client_connected() )
        {
        return;
        }

    // Player selected?
    if( !is_player_selected() )
        {
        print_to_console( client, "No player entity selected." );
        return;
        }

    player->handle_move_action( action );
    }


void AccountObjectImpl::handle_item_action
    (
    const PlayerItemAction & action
    )
    {
    // Still connected?
    if( !is_client_connected() )
        {
        return;
        }

    // Player selected?
    if( !is_player_selected() )
        {
        print_to_console( client, "No player entity selected." );
        return;
        }

    player->handle_item_action( action );
    }


void AccountObjectImpl::set_name
    (
    const std::string & name
    )
    {
    if( player != null )
        {
        player->set_name( name );
        }
    }


void AccountObjectImpl::set_chat_color
    (
    const SColor & chat_color
    )
    {
    if( player != null )
        {
        player->set_chat_color( chat_color );
        }
    }


void AccountObjectImpl::say
    (
    const std::string & message
    )
    {
    if( player != null )
        {
        player->say( message, Player::MESSAGE_TALK );
        }
    }


bool AccountObjectImpl::is_root_account()
    {
    return root_account;
    }


PValueArray<SString> AccountObjectImpl::process_account_command
    (
    const std::string &             command,
    Console::Environment &          env
    )
    {
    PValueArray<SString> errors;

    Demo::Server::Global::get_account_console().process_command
        (
        this,
        command,
        false,
        errors,
        env
        );

    return errors;
    }


PValueArray<SString> AccountObjectImpl::auto_complete_account_command
    (
    const std::string & partial_command
    )
    {
    PValueArray<SString>    completions;
    Console::Environment    dummy_environment;

    Demo::Server::Global::get_account_console().process_command
        (
        this,
        partial_command,
        true,
        completions,
        dummy_environment
        );

    return completions;
    }


PValueArray<SString> AccountObjectImpl::process_node_command
    (
    Remote<NodeObject>          node,
    const std::string &         command,
    Console::Environment &      env
    )
    {
    if( node != null )
        {
        return node->sync_process_node_command( this, command, env );
        }
    else
        {
        throw Massiv::Core::Lib::RuntimeException( "Null pointer." );
        }
    }


PValueArray<SString> AccountObjectImpl::auto_complete_node_command
    (
    Remote<NodeObject>  node,
    const std::string & partial_command
    )
    {
    if( node != null )
        {
        return node->sync_auto_complete_node_command( this, partial_command );
        }
    else
        {
        throw Massiv::Core::Lib::RuntimeException( "Null pointer." );
        }
    }


void AccountObjectImpl::process_node_command_at_object
    (
    Remote<Object>          object,
    const std::string &     command
    )
    {
    // Create "command object" and let it migrate to @c object, then
    // process the command locally.

    CreateObject<RemoteConsoleCommand>( object, this, command );
    }


bool AccountObjectImpl::can_connect_console_to
    (
    Remote<NodeObject> node
    )
    {
    if( node != null )
        {
        if( node == client )
            {
            // Since SRPC to clients is not allowed @c node == client case
            // must be handled in this way.

            return true;
            }
        else
            {
            return node->sync_can_connect( this );
            }
        }
    else
        {
        return false;
        }
    }


void AccountObjectImpl::log_game_message
    (
    WeakPointer< ClientEntity > source_entity,
    const std::string &         message,
    const SColor &              message_color
    )
    {
    if( client != null )
        {
        client->async_handle_game_message( source_entity, message, message_color );
        }
    }


void AccountObjectImpl::print_to_console
    (
    WeakPointer< NodeObject >   sender,
    const std::string &         message,
    const SColor &              message_color
    )
    {
    if( client != null )
        {
        client->async_print_to_console( sender, message, message_color );
        }
    }


void AccountObjectImpl::modify_elevation_command
    (
    const SVector2 &                    center,
    ClientSector::ElevationMapKind      map_kind,
    ClientSector::ElevationMapFunction  function,
    ClientSector::ElevationMapOperation operation,
    Float32                             radius,
    Float32                             strength
    )
    {
    if( !is_root_account() )
        {
        return;
        }

    if( !is_player_selected() )
        {
        return;
        }

    Remote< Sector > sector = player->get_last_sector();
    Remote< Sector > originator = sector->sync_get_sector_from_same_layer( center );

    if( originator != null )
        {
        originator->async_modify_elevation
            (
            center,
            map_kind,
            function,
            operation,
            radius,
            strength
            );
        }
    }


void AccountObjectImpl::toggle_path_finding()
    {
    if( player != null )
        {
        player->log_game_message( std::string( "Path finding was " ) +
            ( player->toggle_path_finding() ? "enabled." : "disabled." ) );
        }
    }


void AccountObjectImpl::modify_material_command
    (
    const SVector2 &                    center,
    Float32                             radius,
    ClientSector::MaterialId            material_id
    )
    {
    if( !is_root_account() )
        {
        return;
        }

    if( !is_player_selected() )
        {
        return;
        }

    Remote< Sector > sector = player->get_last_sector();
    Remote< Sector > originator = sector->sync_get_sector_from_same_layer( center );

    if( originator != null )
        {
        originator->async_modify_material
            (
            center,
            radius,
            material_id
            );
        }
    }


void AccountObjectImpl::delete_entity_command
    (
    const SString &                 entity_object_id
    )
    {
    if( !is_root_account() )
        {
        return;
        }

    if( !is_player_selected() )
        {
        return;
        }

    Serializer::Description desc;
    std::istringstream ss( entity_object_id );
    TextReader  tr( ss );
    tr.next_line();

    Remote< Demo::Lib::Entity > entity;
    entity.text_read( tr, desc );

    if( entity != null )
        {
        entity->sync_remove_from_layer();
        }
    }


void AccountObjectImpl::edit_entity_command
    (
    const SString &                 entity_object_id,
    const SString &                 method_name,
    const SString &                 parameters
    )
    {
    if( !is_player_selected() )
        {
        return;
        }

    Serializer::Description desc;
    std::istringstream ss( entity_object_id );
    TextReader  tr( ss );
    tr.next_line();

    Remote< Demo::Lib::Entity > entity;
    entity.text_read( tr, desc );
    if( entity == null )
        {
        return;
        }

   if( entity != player && !is_root_account() )
        {
        return;
        }

    if( method_name == "translate_entity" )
        {
        SVector2 vector;
        std::istringstream ss( parameters );
        TextReader  tr( ss );
        tr.next_line();
        try
            {
            vector.text_read( tr, desc );
            }
        catch( ... )
            {
            return;
            }
        entity->sync_translate_entity( vector );
        return;
        }

    if( method_name == "rotate_entity" )
        {
        SOrientation orientation;
        std::istringstream ss( parameters );
        TextReader  tr( ss );
        tr.next_line();
        try
            {
            orientation.text_read( tr, desc );
            }
        catch( ... )
            {
            return;
            }
        entity->sync_rotate_entity( orientation );
        }

    if( method_name == "set_model_tags" )
        {
        entity->sync_set_model_tags( parameters );
        }
    }


void AccountObjectImpl::create_entity_command
    (
    const SString &                 parameters
    )
    {
    if( !is_root_account() )
        {
        return;
        }

    if( !is_player_selected() )
        {
        return;
        }

    const ObjectId & ns_id = Massiv::System::get_naming_service_object_id();
    Remote< NamingService > ns = ns_id;
    if( ns == null )
        {
        return;
        }
    Remote< EntityFactory > ef;
    ef = ns->sync_get_object( "EntityFactory" ).convert();
    if( ef == null )
        {
        return;
        }

    Serializer::Description desc;
    std::istringstream ss( parameters );
    TextReader  tr( ss );

    std::string entity_class;
    SVector2 origin;
    SOrientation orientation;
    Remote<Layer> layer;
    SBool solid;
    SVector2 bounding_box_size;
    SBool bounding_box_axis_aligned;
    std::string model_tags;
    try
        {    
        tr.next_line();

        tr.line() >> entity_class;

        origin.text_read( tr, desc );
        orientation.text_read( tr, desc );
        layer.text_read( tr, desc );
        solid.text_read( tr, desc );
        bounding_box_size.text_read( tr, desc );
        bounding_box_axis_aligned.text_read( tr, desc );

        tr.line() >> quote_string( model_tags );
        }
    catch( Exception & )
        {
        return;
        }
    
    Remote< Entity > entity = ef->sync_create_entity_in_layer
        (
        entity_class,
        origin,
        orientation,
        layer,
        solid,
        bounding_box_size,
        bounding_box_axis_aligned,
        model_tags
        );

    if( entity == null )
        {
        return;
        }

    if( entity_class == "Demo::Lib::ColorGenerator" )
        {
        // Setup color timer.
        STime tick_period;
        STime period;
        SBool enable_looping;
        WaveGenerator::WaveType wave;
        SColor color;        
        
        try
            {
            tick_period.text_read( tr, desc );
            period.text_read( tr, desc );
            enable_looping.text_read( tr, desc );

            // Read wave.
            std::string wave_str;
            tr.line() >> quote_string( wave_str );
            if( wave_str == "Sine Wave" )
                {
                wave = WaveGenerator::WAVE_SINE;
                }
            else
            if( wave_str == "Square Wave" )
                {
                wave = WaveGenerator::WAVE_SQUARE;
                }
            else
            if( wave_str == "Saw Wave" )
                {
                wave = WaveGenerator::WAVE_SAW;
                }
            else
            if( wave_str == "Random" )
                {
                wave = WaveGenerator::WAVE_RANDOM;
                }
            else
                {
                return;
                }

            color.text_read( tr, desc );
            } // try
        catch( Exception & )
            {
            return;
            }
        color_generator = entity.convert();
        color_generator->sync_setup_color_generator( tick_period, period, enable_looping, wave, color );
        color_generator->sync_start();
        }

    if( entity_class == "Demo::Lib::Lamp" )
        {
        // Plug to last color timer.
        if( color_generator != null )
            {
            Remote< Logic > logic = entity.convert();
            color_generator->sync_plug_output( "output", logic, "input" );
            }
        }
    }


bool AccountObjectImpl::is_client_connected() const
    {
    return client != null;
    }


bool AccountObjectImpl::is_player_selected() const
    {
    return player != null;
    }


void AccountObjectImpl::start_replication_of_nearby_sectors()
    {
    if( replication_tick_pending )
        {
        return;
        }

    Remote< AccountObjectImpl > self = this;
    self->async_replication_tick();
    replication_tick_pending = true;
    }


void AccountObjectImpl::replication_tick()
    {
    hope( replication_tick_pending.read() );

    if( !is_client_connected() || !is_player_selected() )
        {
        // Stop replication.
        replication_tick_pending = false;
        return;
        }

    hope( client != null );
    hope( player != null );

    STime   replication_period;
    const Float32 replication_duration = nearby_sectors_replication_duration;

    // Replicate player itself.
    player->get_client_entity()->replicate_to( client, STime::current_time() + 
        replication_duration );

    // Replicate nearby sectors.
    Pointer< Sector > sector = player->get_current_sector();

    if( sector != null )
        {
        // Replicate all nearby sectors to our client.
        CreateObject< ReplicateNearbySectorsTask >
            ( sector, client, replication_duration, replication_neighborhood );

        replication_period = STime( replication_big_period );
        }
    else
        {
        // The player is not linked to any sector. The player
        // is moving (migrating) to remote sector right now.
        // Delay the replication tick a little.
        replication_period = STime( replication_small_period );
        }

// #define FORCE_REPLICA_TIMEOUT

#ifndef FORCE_REPLICA_TIMEOUT

    // Schedule next tick.
    Remote< AccountObjectImpl > self = this;
    self->param( RPC_DELAYED, replication_period ).async_replication_tick();

#endif // FORCE_REPLICA_TIMEOUT
    }


void AccountObjectImpl::create_player()
    {
    hope( player == null );

    WeakPointer< Layer > default_layer;

    // Get pointer to default layer.
        {
        WeakPointer< Object > object = Massiv::System::get_naming_service_object_id();
        if( object != null )
            {
            Remote< NamingService > naming_service = object.convert();

            default_layer = naming_service->sync_get_object( "DefaultLayer" ).convert();
            }
        else
            {
            panic( "Unable to obtain naming service pointer." );
            }
        }

    if( default_layer == null )
        {
        panic( "Unable to get default layer." );
        }

   // Create player.
        {
        Pointer< Player >   p;
        SColor              chat_color( 1.0f, 0.4f, 0.4f, 1.0f );

        // Create the object.
        p = CreateObject< Player >();
        p->set_chat_color( chat_color );
        p->name_object( "player" );

        // Link the player with its account.
        p->account = this;
        player = p;

        // FIXME: Temp hack. Create player light source.

        p->create_light();
        Pointer< ClientLight > l = p->get_light();
        l->set_offset( SVector3( 0.0f, 0.0f, 1.0f ) );
        l->set_attenuation( SVector3( 0.5f, 0.1f, 0.0f ) );
        l->set_diffuse_intensity( SColor( 1.0f, 0.8f, 0.6f ) );
        }

    // Put player into the layer.
        {
        player->put_into_layer
            (
            EntityPosition
                (
                SVector2( 0.0f, 0.0f ),
                SOrientation(),
                default_layer
                )
            );
        }

    // Create inventory objects.
        {
        Pointer< Potion >   p;
        Pointer< Item >     item;
        PotionType          pt;
        BootsType           bt;
        WeaponType          wt;

        pt = POTION_HEALTH;
        p = CreateObject< Potion >( pt );
        player->put_item_to_inventory( p );

        pt = POTION_SPEED;
        p = CreateObject< Potion >( pt );
        player->put_item_to_inventory( p );

        bt = BOOTS_SPEED;
        item = CreateObject< Boots >( bt );
        player->put_item_to_inventory( item );

        wt = WEAPON_AXE;
        item = CreateObject< Weapon >( wt );
        player->put_item_to_inventory( item );
        }
    }


void AccountObjectImpl::create_initial_world()
    {
    if( !root_account )
        {
        return;
        }

    //
    // TODO: Move world creation outside account object.
    // To World class?
    //

    Remote< NamingService > naming_service;
    Pointer< World >        world;
    Pointer< Layer >        layer;

    try
        {
        WeakPointer< Object > object = Massiv::System::get_naming_service_object_id();
        naming_service = object.convert();
        }
    catch( Throwable & )
        {
        }

    if( naming_service == null )
        {
        panic( "Unable to obtain naming service pointer." );
        }

    // Create world.
        {
        int i, j;

        world = CreateObject<World>();
        layer = CreateObject<Layer>();

        world->name_object( "world" );
        layer->name_object( "layer" );

        world->register_layer( layer );

        for( i = -5; i <= 5; i++ )
            {
            for( j = -5; j <= 5; j++ )
                {
                Pointer< Sector >   sector = CreateObject<Sector>();
                std::ostringstream  ss;

                ss << "sector_" << i << "_" << j;
                sector->name_object( ss.str() );

                layer->register_sector( sector, SectorAddress( layer, i, j ) );
                }
            }
            
        naming_service->sync_register_object( "World", world, true );
        naming_service->sync_register_object( "DefaultLayer", layer, true );
        }

    // Create helper objects.
        {
        Pointer< EntityFactory > factory = CreateObject< EntityFactory >();

        factory->name_object( "entity_factory" );
        naming_service->sync_register_object( "EntityFactory", factory, true );
        }

    // Create robots. HACK. Should be done through the console.
        {
        #if 1

        // Boss.

        Pointer< Robot >    entity = CreateObject< Robot >();
        Pointer< Robot >    boss = entity;
        int                 i;

        entity->put_into_layer
            (
            EntityPosition
                (
                SVector2(),
                SOrientation(),
                layer
                )
            );

        for( i = 0; i < 10; i++ )
            {
            entity = CreateObject< Robot >();
            entity->put_into_layer
                (
                EntityPosition
                    (
                    Random::random_vector() * 15.0f,
                    SOrientation(),
                    layer
                    )
                );
            entity->link_to_boss( boss );
            }

        #else

        int i;

        for( i = 0; i < 20; i++ )
            {
            Pointer<Entity> entity = CreateObject< Player >();
            entity->put_into_layer
                (
                EntityPosition
                    (
                    Random::random_vector() * 15.0f,
                    SOrientation(),
                    layer
                    )
                );
            }

        #endif
        }
    }


} // namespace Demo::Lib
} // namespace Demo
