/** @file account_object_impl.h
 * Account object implementation.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_ACCOUNT_OBJECT_IMPL_H
#define DEMO_ACCOUNT_OBJECT_IMPL_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_ACCOUNT_OBJECT_H
#include "lib/shared/account_object.h"
#endif 

#ifndef DEMO_CLIENT_SECTOR_H
#include "lib/shared/client_sector.h"
#endif

#ifndef DEMO_CONSOLE_H
#include "lib/shared/console.h"
#endif

#ifndef DEMO_GENERATOR_H
#include "lib/server/generator.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Implementation of #AccountObject interface.
 */
class AccountObjectImpl : public AccountObject
    {
public:

    MASSIV_OBJECT( AccountObjectImpl );

    /** Initialize the object. */
    void initialize
        (
        bool root_account
        );
    
public:

    //
    // Object callbacks.
    //

    // From Object.
    virtual void object_updated
        (
        Object::UpdateReason    reason
        );

public:

    //
    // AccountObjectInterface implementation.
    //
    
    // From AccountObjectInterface.
    virtual void client_connected
        (
        const WeakPointer< ::Massiv::Core::Lib::NodeObjectInterface > & node_object
        );

    // From AccountObjectInterface.
    virtual void client_disconnected();

    // From AccountObjectInterface.
    virtual void log_system_messages
        (
        const WeakPointer< ::Massiv::Core::Lib::NodeObjectInterface > & sender,
        const PValueArray< SString > &                                  messages
        );

    // From AccountObjectInterface.
    virtual bool check_client_rpc_request
        (
        const Pointer< ::Massiv::Core::RPCObject > & request
        ) const;

public:

    //
    // AccountObject implementation.
    //

    // From AccountObject.
    virtual void request_player_replication
        (
        WeakPointer< ClientEntity > player
        );
    
    // From AccountObject.
    virtual void handle_move_action
        (
        const PlayerMoveAction & action
        );

    // From AccountObject.
    virtual void handle_item_action
        (
        const PlayerItemAction & action
        );

    // From AccountObject.
    virtual void say
        (
        const std::string & message
        );

    // From AccountObject.
    virtual bool is_root_account();
    
    // From AccountObject.
    virtual PValueArray<SString> process_account_command
        (
        const std::string &         command,
        Console::Environment &      env
        );
    
    // From AccountObject.
    virtual PValueArray<SString> auto_complete_account_command
        (
        const std::string & partial_command
        );
    
    // From AccountObject.
    virtual PValueArray<SString> process_node_command
        (
        Remote<NodeObject>          node,
        const std::string &         command,
        Console::Environment &      env
        );

    // From AccountObject.
    virtual PValueArray<SString> auto_complete_node_command
        (
        Remote<NodeObject>  node,
        const std::string & partial_command
        );
    
    // From AccountObject.
    virtual void process_node_command_at_object
        (
        Remote<Object>          object,
        const std::string &     command
        );
    
    // From AccountObject.
    virtual bool can_connect_console_to
        (
        Remote<NodeObject>  node
        );

    // From AccountObject.
    virtual void log_game_message
        (
        WeakPointer< ClientEntity > source_entity,
        const std::string &         message,
        const SColor &              message_color
        );
    
    // From AccountObject.
    virtual void print_to_console
        (
        WeakPointer< NodeObject >   sender,
        const std::string &         message,
        const SColor &              message_color = SColor::YELLOW
        );
    
    // From AccountObject.
    virtual void modify_material_command
        (
        const SVector2 &                    center,
        Float32                             radius,
        ClientSector::MaterialId            material_id
        );

    // From AccountObject.
    virtual void delete_entity_command
        (
        const SString &                     entity_object_id
        );

    // From AccountObject.
    virtual void edit_entity_command
        (
        const SString &                     entity_object_id,
        const SString &                     method_name,
        const SString &                     parameters
        );

    // From AccountObject.
    virtual void create_entity_command
        (
        const SString &                 parameters
        );

public:

    //
    // Commands (from AccountConsole).
    //
    
    /** Set player name. */
    void set_name
        (
        const std::string & name
        );
    
    /** Set player's chat color. */
    void set_chat_color
        (
        const SColor & chat_color
        );
    
    /** Modify elevation command. */
    void modify_elevation_command
        (
        const SVector2 &                    center,
        ClientSector::ElevationMapKind      map_kind,
        ClientSector::ElevationMapFunction  function,
        ClientSector::ElevationMapOperation operation,
        Float32                             radius,
        Float32                             strength
        );

    /** Enable or disable path finding for the selected player. */
    void toggle_path_finding();

public:

    /**
     * Is client connected to the account?
     * If true, @c client is not null pointer.
     */
    bool is_client_connected() const;
    
    /**
     * If any player selected?
     * If true, @c player is not null pointer.
     */
    bool is_player_selected() const;

protected:

    /** Start replicating nearby sectors. */
    void start_replication_of_nearby_sectors();

public: // FIXME: kvuli RPC

    /** Replicates all nearby sectors to client. */
    void replication_tick();
    
private:

    /**
     * Create player.
     */
    void create_player();

public:

    /**
     * Create initial world.
     * When simulation startup archive is beeing created,
     * this function is scheduled so when the simulation
     * is started from the archive, this function will be
     * invoked.
     * @note Does nothing if the object is not root account object.
     */
    void create_initial_world();

private:

    PBool                           root_account;
        /**< Is this account root account object? */

    PRemote< ClientNodeInterface >  client;
        /**< Pointer to client's node object that is associated with
             this account. @c null if client's not connected. */

    PPointer< Player >              player;
        /**< Currently selected player entity. This is the entity that is
             controled by the client.
             Player is linked to the account object by strong pointer.
             This ensures that the account migrates with the player.
             It forces the client node to dynamically create connection to the
             server node which owns the player as the account migrates
             with it. */

protected:

    PBool                           replication_tick_pending;
        /**< True if replication tick is scheduled. */

private:
    
    Remote< ColorGenerator >        color_generator;
        /** FIXME */
    
    }; // AccountObjectImpl


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_ACCOUNT_OBJECT_IMPL_H
