/** @file area.cpp
 * Area implementation.
 */

/*
 * begin        : Thu Aug 29 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/area.h"
#include "lib/server/area_rpc.h"
#include "lib/server/physical.h"
#include "lib/shared/client_entity.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of Area::Query
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Area::Query::initialize
    (
    Pointer<Area>       the_area_object,
    Pointer<Sector>     the_originator,
    const ConvexArea &  the_area,
    const SVector2 &    the_translation,
    Float32             the_rotation_angle,
    int                 the_flags
    )
    {
    area_object = the_area_object;
    SpatialQuery::initialize
        (
        the_originator,
        the_area,
        the_translation,
        the_translation,
        the_rotation_angle,
        the_flags,
        false
        );
    }


bool Area::Query::entity_reported
    (
    WeakPointer<Entity> entity
    )
    {
    // Since the spatial query operates in a sequential mode,
    // @c entity is automatically inserted to @c SpatialQuery::reported
    // list, which serves as a visitor list.

    return true;
    }


void Area::Query::enumeration_finished()
    {
    // Migrate the query object to @c area_object so that the query result
    // could be processed by it.

    deliver_asap_to( area_object );
    }


void Area::Query::delivered_to
    (
    WeakPointer<Object> object,
    const STime &       time
    )
    {
    if( object == area_object )
        {
        Pointer<Area> my_area = object.convert();
        my_area->scan_done( this );
        }
    else
        {
        Super::delivered_to( object, time );
        }
    }


// Implementation of Area
// ~~~~~~~~~~~~~~~~~~~~~~

void Area::initialize()
    {
    initialize( ConvexArea( SVector2(), SVector2( 16.0f, 16.0f ) ), 
        Entity::MONSTER );
    }


void Area::initialize
    (
    const ConvexArea &  the_area,
    int                 the_entity_flags
    )
    {
    SocketId id;
    
    Entity::initialize();

    id = create_input( "enable" );
    hope( id == get_enable_id() );

    id = create_output( "entered" );
    hope( id == get_entered_id() );
    
    id = create_output( "left" );
    hope( id == get_left_id() );
    
    id = create_output( "count" );
    hope( id == get_count_id() );

    monitored_area = the_area;
    monitored_entity_flags = the_entity_flags;

    enable();
    }


Area::SocketId Area::get_enable_id()
    {
    return 0;
    }


Area::SocketId Area::get_entered_id()
    {
    return 0;
    }


Area::SocketId Area::get_left_id()
    {
    return 1;
    }


Area::SocketId Area::get_count_id()
    {
    return 2;
    }


void Area::enable()
    {
    if( enabled )
        {
        request_disable = false;
        }
    else
        {
        schedule_next_tick();
        }
    }


void Area::disable()
    {
    if( enabled )
        {
        request_disable = true;
        }
    }


void Area::entered
    (
    WeakPointer<Physical> visitor
    )
    {
    }


void Area::left
    (
    WeakPointer<Physical> visitor
    )
    {
    }


Area::SocketId Area::translate_input
    (
    const std::string & input_name
    )
    {
    if( input_name == "enable" )
        {
        return get_enable_id();
        }
    else
        {
        return Logic::translate_input( input_name );
        }
    }


Area::SocketId Area::translate_output
    (
    const std::string & output_name
    )
    {
    if( output_name == "entered" )
        {
        return get_entered_id();
        }
    else
    if( output_name == "left" )
        {
        return get_left_id();
        }
    else
        {
        return Logic::translate_output( output_name );
        }
    }


void Area::operate_synced
    (
    const STime & delivery_time
    )
    {
    // Check "enable" input.

    Pointer<BooleanPacket> packet = get_input( get_enable_id() ).convert();
    
    if( packet != null )
        {
        bool value = packet->get_boolean();

        if( enabled.read() && !value )
            {
            disable();
            }
        else
        if( !enabled.read() && value )
            {
            enable();
            }
        }
    }


void Area::next_tick()
    {
    if( request_disable )
        {
        request_disable = false;
        enabled = false;
        }

    if( enabled )
        {
        start_scan();
        }
    }


void Area::start_scan()
    {
    if( current_sector != null )
        {
        Pointer<Query> query =
            CreateObject<Query>
                (
                this,
                current_sector,
                monitored_area,
                client_entity->current_position.origin.read(),
                client_entity->current_position.orientation.heading(),
                monitored_entity_flags
                );
        }
    else
        {
        scan_done( null );
        }
    }


void Area::scan_done
    (
    Pointer<Query> query
    )
    {
    visitor_count_difference = 0;

    // Report differences in visitor lists.

    if( query != null )
        {
        SpatialQuery::ReportedList::Iterator it;

        if( last_query != null )
            {
            // Pin the queries.

            LocalProxy<Query> q = query.dereference();
            LocalProxy<Query> prev_q = last_query.dereference();
            
            // Report differences.

            for( it = q->reported.begin(); it != q->reported.end(); ++it )
                {
                if( !prev_q->reported.contains( *it ) )
                    {
                    WeakPointer<Entity> visitor = *it;
                    
                    // Visitor was reported by the current query but
                    // not by the previous one.
                    
                    enter( visitor );
                    visitor_count_difference++; 
                    }
                }
            
            for( it = prev_q->reported.begin(); it != prev_q->reported.end(); ++it )
                {
                if( !q->reported.contains( *it ) )
                    {
                    WeakPointer<Entity> visitor = *it;
                    
                    // Visitor was reported by the previous query but
                    // not by the current one.
                    
                    leave( visitor );
                    }
                }
            }
        else
            {
            // Pin the query.

            LocalProxy<Query> q = query.dereference();

            // Report its contents.

            for( it = q->reported.begin(); it != q->reported.end(); ++it )
                {
                WeakPointer<Entity> visitor = *it;
                enter( visitor );
                }
            }
        }
    
    // Report how many entities are currently in the area.

    output
        (
        get_count_id(),
        CreateObject<FloatPacket>( static_cast<Float32>( num_visitors ) ),
        STime::current_time()
        );

    // Done.

    last_query = query;
    schedule_next_tick();
    }


namespace {

// Tick scheduler settings.

const Int32     lower_bound_to_visitor_count = 1;
const Int32     upper_bound_to_visitor_count = 8;
const Float32   lazy_factor = 1.2f;
const Float32   hurry_factor = 0.4f;
const Float32   min_tick_period = 0.5f;
const Float32   max_tick_period = 4.0f;

}


void Area::schedule_next_tick()
    {
    Remote<Area> self = this;
    
    // Adapt @c tick_period.

    // FIXME: abs is not in std namespace under STLPort even if
    // _STLP_DO_IMPORT_CSTD_FUNCTIONS is defined.

    if( ::abs( visitor_count_difference ) < lower_bound_to_visitor_count )
        {
        hope( lazy_factor > 1.0f );
        tick_period *= lazy_factor;
        }
    
    if( ::abs( visitor_count_difference ) > upper_bound_to_visitor_count )
        {
        hope( hurry_factor < 1.0f );
        tick_period *= hurry_factor;
        }

    if( tick_period < min_tick_period )
        {
        tick_period = min_tick_period;
        }
    
    if( tick_period < max_tick_period )
        {
        tick_period = max_tick_period;
        }

    // Schedule the call.

    hope( enabled.read() && !request_disable.read() );
    self->param( RPC_DELAYED, tick_period.read() ).async_next_tick();
    }


void Area::enter
    (
    WeakPointer<Entity> visitor
    )
    {
    if( !visitor.inherits( typeid( Physical ) ) )
        {
        return;
        }

    output
        (
        get_entered_id(),
        CreateObject<ObjectPacket>( visitor ),
        STime::current_time()
        );
    
    num_visitors++;
    visitor_count_difference++;

    entered( visitor.convert() );
    }


void Area::leave
    (
    WeakPointer<Entity> visitor
    )
    {
    if( !visitor.inherits( typeid( Physical ) ) )
        {
        return;
        }

    output
        (
        get_left_id(),
        CreateObject<ObjectPacket>( visitor ),
        STime::current_time()
        );
    
    num_visitors--;
    visitor_count_difference--;

    left( visitor.convert() );
    }


} // namespace Demo::Lib
} // namespace Demo
