/** @file area.h
 * Area.
 */

/*
 * begin        : Thu Aug 29 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_AREA_H
#define DEMO_AREA_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_ENTITY_H
#include "lib/server/entity.h"
#endif

#ifndef DEMO_SPATIAL_QUERY_H
#include "lib/server/spatial_query.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Area is an entity that monitors given convex area and maintains
 * a list of its visitors. The area is a logic object with a set of
 * input/output sockets.
 *
 * When a physical (of requested type) enters the monitored area, ObjectPacket
 * is sent to "entered" output. Similarly, ObjectPacket is sent to "left"
 * output upon physical exit.
 *
 * "count" output sends how many entities are currently inside the monitored
 * area. If routed to a TimerWithPeriod's "enable" input, animation can be
 * triggered if an entity enters the area.
 *
 * "enable" input enables or disables Area object functionality.
 * Only physicals are monitored.
 */
class Area : public Entity
    {
public:
    
    MASSIV_OBJECT( Area );

    /**
     * Query object that collects physicals reported by a sequential spatial
     * query.
     */
    class Query : public SpatialQuery
        {
    public:
            
        MASSIV_OBJECT( Query );

        typedef SpatialQuery Super;
    
        /** Initializes the query object. */
        void initialize
            (
            Pointer<Area>       the_area_object,
            Pointer<Sector>     the_originator,
            const ConvexArea &  the_area,
            const SVector2 &    the_translation,
            Float32             the_rotation_angle,
            int                 the_flags
            );

        /**
         * Reports the monitored entity.
         */
        virtual bool entity_reported
            (
            WeakPointer<Entity> entity
            );

        /**
         * Migrates the query to @c area_object so that it could be processed
         * by it.
         */
        virtual void enumeration_finished();

        /**
         * If query is delivered to an @c Area object it implies
         * that the the migration have been initiated by @c enumeration_finished().
         * In that case area object is notified that the area scan has
         * completed.
         */
        virtual void delivered_to
            (
            WeakPointer<Object> object,
            const STime &       time
            );

    public:

        PRemote<Area> area_object;
            /**< The area object. Weak/remote pointer so that the @c area
                 would not belong to the migration group of the @c Query
                 object. */

        }; // Query

    friend class Query;

protected:

    //
    // Area interface.
    //

    /** Default-initializes object. */
    virtual void initialize();

    /**
     * Initializes object.
     * [ @c the_area, @c the_entity_flags ] determine monitored area and
     * what physical types are reported.
     * See @c SpatialQuery for more information on the parameters.
     */
    void initialize
        (
        const ConvexArea &  the_area,
        int                 the_entity_flags
        );
    
    /** Returns "enable" input socket id. */
    static SocketId get_enable_id();
    
    /** Returns "entered" output socket id. */
    static SocketId get_entered_id();
    
    /** Returns "left" output socket id. */
    static SocketId get_left_id();
    
    /** Returns "count" output socket id. */
    static SocketId get_count_id();

    /** Enables object's logic. */
    void enable();

    /** Disables object's logic. */
    void disable();

    /** Called by @c Area to notify that @c visitor has entered the area. */
    virtual void entered
        (
        WeakPointer<Physical> visitor
        );
    
    /** Called by @c Area to notify that @c visitor has left the area. */
    virtual void left
        (
        WeakPointer<Physical> visitor
        );

protected:

    //
    // Logic interface implementation.
    //
    
    /**
     */
    virtual SocketId translate_input
        (
        const std::string & input_name
        );
    /**
     */
    virtual SocketId translate_output
        (
        const std::string & output_name
        );

    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

//private:
public: // FIXME: Kvuli RPC
    
    /** Called periodically if object's logic is enabled. */
    void next_tick();

private:
    
    /**
     * Starts a new scan of the monitored area to determine which entities
     * have entered or left it.
     */
    void start_scan();

    /**
     * Determines differences in visitor lists between the last query result
     * and the current query. Triggers "entered" / "left" outputs.
     * Called by the @c query object to notify @c Area that the area scan
     * has finished.
     */
    void scan_done
        (
        Pointer<Query> query
        );

    /** Schedules next call to @c next_tick(). */
    void schedule_next_tick();

    /**
     * Registers the visitor.
     * Sends @c visitor to "entered" output.
     * @note Non-physicals are ignored.
     */
    void enter
        (
        WeakPointer<Entity> visitor
        );
    
    /**
     * Unregisters the visitor.
     * Sends @c visitor to "left" output.
     * @note Non-physicals are ignored.
     */
    void leave
        (
        WeakPointer<Entity> visitor
        );

private:

    PBool           enabled;
        /**< Is object's logic enabled? */
    PBool           request_disable;
        /**< Was disable requested? */
    ConvexArea      monitored_area;
        /**< Monitored area. In layer coordinates relative to entity's
             current position and orientation. [ 0, 0 ] corresponds
             to entity center. */
    PVlInt32        monitored_entity_flags;
        /**< Only visitors whose flags match @c monitored_entity_flags are
             monitored. */
    PPointer<Query> last_query;
        /**< Result of the last query. Visitor list is stored there. */
    PVlInt32        num_visitors;
        /**< Current visitor count. */
    PVlInt32        visitor_count_difference;
        /**< Difference in the visitor count between the two last scans. */
    PFloat32        tick_period;
        /**< Current tick period. */

    }; // Area


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_AREA_H
