/** @fil brain.cpp
 * Brain implementation.
 */

/*
 * begin        : Sun Nov 16 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/server/brain.h"
#include "lib/server/collision_info.h"
#include "lib/server/monster.h"
#include "lib/server/brain_rpc.h"
#include "lib/server/lock.h"
#include "lib/shared/random.h"
#include <algorithm>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of Brain
// ~~~~~~~~~~~~~~~~~~~~~~~

void Brain::initialize
    (
    Pointer<Monster> the_monster
    )
    {
    monster = the_monster;
    }


void Brain::report_collision
    (
    const CollisionInfo & collision_info
    )
    {
    Brain::cancel_move();
    }


void Brain::report_visible
    (
    const VisibleList & visible
    )
    {
    }


bool Brain::move_to
    (
    const SVector2 &    position
    )
    {
    Brain::cancel_move();
    return monster->find_path( position );
    }


void Brain::move_to_along_line
    (
    const SVector2 &    position
    )
    {
    Brain::cancel_move();
    monster->add_way_point( position );
    }


void Brain::look_at
    (
    const SVector2 &    point
    )
    {
    Brain::cancel_move();
    monster->look_at( point );
    }


void Brain::cancel_move()
    {
    monster->clear_path();
    }


void Brain::think()
    {
    }


void Brain::think_task
    (
    int task_id
    )
    {
    if( think_task_counter != task_id )
        {
        // This task was obsoleted by @c rethink(). Cancel.

        return;
        }

    think_time = STime();
    think();
    }
    

void Brain::rethink
    (
    const STime &   delay,
    bool            allow_reschedule_to_future
    )
    {
    Remote<Brain>   self( this );
    STime           schedule_time = STime::current_time() + delay;
    
    if( !think_time.is_equal_to( STime() ) && schedule_time > think_time && 
        !allow_reschedule_to_future )
        {
        return;
        }

    // Schedule next call to @c think_task().
    // If another call to @c think_task() was also scheduled but the call
    // has not been dispatched yet, the task will be cancelled once
    // the call is dispatched.

    think_time = schedule_time;
    self->param( RPC_DELAYED, delay ).async_think_task( ++think_task_counter );
    }


void Brain::link_to_boss
    (
    Remote<Brain> new_boss
    )
    {
    if( linking_to_boss )
        {
        // Reject the request.

        return;
        }

    OPTIMISTIC_LOCK( linking_to_boss, NULL );

    if( boss != null )
        {
        boss->sync_unlink_slave( this );
        }

    if( new_boss != null )
        {
        new_boss->sync_link_slave( this );
        }
    
    boss = new_boss;
    }


void Brain::link_slave
    (
    Remote<Brain> slave
    )
    {
    slaves.push_back( slave );
    }


void Brain::unlink_slave
    (
    Remote<Brain> slave
    )
    {
    slaves.erase( std::find( slaves.begin(), slaves.end(), slave ) - 
        slaves.begin() );
    }


EntityPosition Brain::get_current_position() const
    {
    return monster->get_current_position();
    }


// Implementation of SimpleBrain
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

namespace {

const Float32   path_reschedule_delay = 30.0f;
    /**< Reschedule path every @c path_reschedule_delay seconds (mean value). */
const Float32   delay_after_collision = 0.25f;
    /**< Stand still for @c delay_after_collision seconds if entity collides
         with another entity (mean value). */
const Float32   move_target_epsilon = 1.0f / 8.0f;
    /**< Move target epsilon. */

} // anonymous namespace


void SimpleBrain::initialize
    (
    Pointer<Monster> the_monster
    )
    {
    Super::initialize( the_monster );
    }


void SimpleBrain::think()
    {
    EntityPosition position = get_current_position();

    Super::think();

    if( move_to_goal_set.read() && ( position.origin - move_to_position ).square_magnitude() < 
        move_target_epsilon * move_target_epsilon )
        {
        // We have reached the requested position.

        move_to_goal_set = false;
        }

    if( move_to_goal_set )
        {
        STime time = STime::current_time();

        if( time >= path_reschedule_time )
            {
            // Reschedule path to requested position.
            
            SimpleBrain::move_to( move_to_position.read() );
            }
        else
            {
            // Rethink.

            rethink( path_reschedule_time.read() - time );
            }
        }
    }


bool SimpleBrain::move_to
    (
    const SVector2 & position
    )
    {
    bool path_found;

    // Set move-to goal and initiate the move.
    
    move_to_position = position;
    path_found = Super::move_to( position );
    move_to_goal_set = true;

    // Schedule path reschedule.

    schedule_path_reschedule( path_reschedule_delay * Random::random() * 2.0f );

    return path_found;
    }


void SimpleBrain::look_at
    (
    const SVector2 & point
    )
    {
    // Clear move-to goal.

    move_to_goal_set = false;
    move_to_position.set_zero();
    
    // Look at @c point.

    Super::look_at( point );
    }


void SimpleBrain::cancel_move()
    {
    // Clear move-to goal.

    move_to_goal_set = false;
    Super::cancel_move();
    }


void SimpleBrain::report_collision
    (
    const CollisionInfo & collision_info
    )
    {
    if( collision_info.reason == CollisionInfo::REASON_COLLISION &&
        move_to_goal_set.read() )
        {
        // Wait for a while in hope that the entity that blocked us
        // will pass away meanwhile. Then reschedule the path.

        // Stop.

        Super::cancel_move();

        // Schedule path reschedule.
        
        STime delay = delay_after_collision * Random::random() * 2.0f;

        Global::log_info
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            monster.to_string() + " collided and will reschedule path to " +
            move_to_position.to_string() + " in " + delay.to_string() + " second(s)."
            );

        schedule_path_reschedule( delay );
        }
    }


void SimpleBrain::schedule_path_reschedule
    (
    const STime & delay
    )
    {
    path_reschedule_time = STime::current_time() + delay;
    rethink( delay );
    }


} // namespace Demo::Lib
} // namespace Demo
