/** @file brain.h
 * Brain.
 */

/*
 * begin        : Sun Nov 16 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_BRAIN_H
#define DEMO_BRAIN_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_ENTITY_POSITION_H
#include "lib/shared/entity_position.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Generic brain interface.
 */
class Brain : public Object
    {
public:

    MASSIV_OBJECT( Brain );

    typedef PSet< PWeakPointer<Entity> > VisibleList;

public:

    /** Initializes object. */
    void initialize
        (
        Pointer<Monster> the_monster
        );

public:
        
    //
    // Brain stimuli.
    //

    /**
     * Monster collided callback.
     * Cancels current move by default.
     */
    virtual void report_collision
        (
        const CollisionInfo & collision_info
        );

    /**
     * Monster sees @c physical callback.
     * It is up to monster logic to determine when to check for visible
     * entities and report them to Brain. Possibly Brain can trigger the
     * check by itself.
     * Does nothing by default.
     */
    virtual void report_visible
        (
        const VisibleList & visible
        );

public:

    //
    // Effector control.
    //

    /**
     * Makes monster walk to @c position (in current layer coordinates).
     * Return true if path was found.
     */
    virtual bool move_to
        (
        const SVector2 &    position
        );

    /**
     * Makes monster walk to @c position (in current layer coordinates)
     * along a direct line.
     */
    virtual void move_to_along_line
        (
        const SVector2 &    position
        );

    /** Makes monster look at @c point (current layer coordinates). */
    virtual void look_at
        (
        const SVector2 &    point
        );

    /** Cancels previous move. */
    virtual void cancel_move();
    
public:
    
    /** Main "think" method. */
    virtual void think();

protected:

    /**
     * Schedules next call to @c think().
     * If @c think() was already scheduled it is rescheduled.
     */
    void rethink
        (
        const STime &   delay,
        bool            allow_reschedule_to_future = false
        );

public:

    /** Makes this brain @c new_boss' slave. */
    void link_to_boss
        (
        Remote<Brain> new_boss
        );

//private: // FIXME: Kvuli RPC
public:

    /** Returns current position of the controlled monster. */
    EntityPosition get_current_position() const;

    /** Makes @c slave its slave brain. */
    void link_slave
        (
        Remote<Brain> slave
        );

    /** Unlinks @c slave from my slaves. */
    void unlink_slave
        (
        Remote<Brain> slave
        );

    /** Think task spawned by @c rethink(). */
    virtual void think_task
        (
        int task_id
        );

protected:

    PPointer<Monster>           monster;
        /**< Monster controlled by this Brain. We need to have corresponding
             Monster <-> Brain pair linked bidirectionally so that Monster
             can callback Brain (stimuli) and Brain call Monster (effector
             control). */
    PVlInt32                    think_task_counter;
        /**< Think-task counter incremented every time the new task is spawned.
             Id of the "most recent" think task. */
    PTime                       think_time;
        /**< Time when @c think() will be invoked. */
    PRemote<Brain>              boss;
        /**< My boss. Slave can stimule boss' brain. Can be @c null. */
    PArray< PRemote<Brain> >    slaves;
        /**< Slaves. Boss can control its slave brains. */
    PBool                       linking_to_boss;
        /**< Are we linking self to a boss. */

    }; // Brain


/**
 * Simple brain that handles "move-to" goal. It is able to reschedule path
 * if monster collides with another entity.
 */
class SimpleBrain : public Brain
    {
public:

    MASSIV_OBJECT( SimpleBrain );
    
    typedef Brain Super;

public:
    
    /** Initializes object. */
    void initialize
        (
        Pointer<Monster> the_monster
        );

protected:

    /**
     */
    virtual void think();

    /**
     */
    virtual bool move_to
        (
        const SVector2 & position
        );
    
    /**
     */
    virtual void look_at
        (
        const SVector2 & point
        );
    
    /**
     */
    virtual void cancel_move();

    /**
     */
    virtual void report_collision
        (
        const CollisionInfo & collision_info
        );

private:

    /**
     * Schedule path reschedule to take place after @c delay seconds have
     * elapsed.
     */
    void schedule_path_reschedule
        (
        const STime & delay
        );

protected:

    PBool       move_to_goal_set;
        /**< Should we move somewhere? */
    PVector2    move_to_position;
        /**< Actual @c move_to() goal. Ignored if @c move_to_goal_set is false. */
    PTime       path_reschedule_time;
        /**< When the path should be rescheduled next time. */

    }; // SimpleBrain


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_BRAIN_H
