/** @file convex_area.cpp
 * Convex area implementation.
 */

/*
 * begin        : Wed Aug 6 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/convex_area.h"
#include <limits>

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


namespace {

/**
 * Given an index @c i into a vector @c v compute an index of the successor
 * element, treat the vector as if it was wrapped around (ie. index
 * v.size() corresponds to index 0).
 */
template <typename IndexType, typename Vector>
IndexType next_index
    (
    IndexType       i,
    const Vector &  v
    )
    {
    hope( i >= 0 );
    return ( i + 1 ) % v.size();
    }

/**
 * Compute an index of the predeccessor element.
 */
template <typename IndexType, typename Vector>
IndexType prev_index
    (
    IndexType       i,
    const Vector &  v
    )
    {
    hope( i >= 0 );
    return ( i + v.size() - 1 ) % v.size();
    }

/**
 * Normalize the index so that it would belong to [ 0, v.size() - 1 ]
 * range.
 */
template <typename IndexType, typename Vector>
IndexType normalize_index
    (
    IndexType       i,
    const Vector &  v
    )
    {
    hope( i >= 0 );
    return i % v.size();
    }

} // anonymous namespace
    

// Implementation of ConvexArea::Edge
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ConvexArea::Edge::Edge()
    {
    initialize_object();
    default_constructed();
    }


ConvexArea::Edge::Edge
    (
    const SVector2 & first,
    const SVector2 & second
    )
    {
    initialize_object();
    init( first, second );
    }


ConvexArea::Edge::Edge
    (
    const Edge & edge
    ) : ValueType( edge ),
        normal( edge.normal ),
        displacement( edge.displacement )
    {
    initialize_object();
    }


void ConvexArea::Edge::init    
    (
    const SVector2 & first,
    const SVector2 & second
    )
    {
    // Compute edge direction and rotate it 90 degrees CCW
    // to get a normal.

    SVector2 direction = second - first;
    direction.normalize();
    
    normal.set_x( -direction.y() );
    normal.set_y( direction.x() );

    // Compute edge displacement so that "edge equation" would
    // hold for @c first.

    displacement = -first * normal;
    }


Float32 ConvexArea::Edge::distance
    (
    const SVector2 & point
    ) const
    {
    return normal * point + displacement;
    }


bool ConvexArea::Edge::distance_is_positive
    (
    const SVector2 & point
    ) const
    {
    return FloatTraits<Float32>::is_positive( distance( point ) );
    }


bool ConvexArea::Edge::distance_is_negative
    (
    const SVector2 & point
    ) const
    {
    return !distance_is_positive( point );
    }


// Implementation of ConvexArea::RasterizerCallback
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ConvexArea::RasterizerCallback::~RasterizerCallback()
    {
    }

    
// Implementation of ConvexArea
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ConvexArea::ConvexArea()
    {
    initialize_object();
    default_constructed();
    }


ConvexArea::ConvexArea
    (
    const SVector2 &    origin,
    const SVector2 &    bounding_box_size,
    bool                adjust_to_grid
    )
    {
    initialize_object();
    
    if( FloatTraits<Float32>::epsilon_zero( bounding_box_size.x() ) ||
        FloatTraits<Float32>::epsilon_zero( bounding_box_size.y() ) )
        {
        return;
        }

    Float32 west_x = origin.x() - bounding_box_size.x() / 2.0f;
    Float32 east_x = origin.x() + bounding_box_size.x() / 2.0f;
    Float32 south_y = origin.y() - bounding_box_size.y() / 2.0f;
    Float32 north_y = origin.y() + bounding_box_size.y() / 2.0f;

    if( adjust_to_grid )
        {
        west_x = FloatTraits<Float32>::round( west_x );
        east_x = FloatTraits<Float32>::round( east_x );
        south_y = FloatTraits<Float32>::round( south_y );
        north_y = FloatTraits<Float32>::round( north_y );
        }

    add_vertex( SVector2( east_x, north_y ) );
    add_vertex( SVector2( west_x, north_y ) );
    add_vertex( SVector2( west_x, south_y ) );
    add_vertex( SVector2( east_x, south_y ) );
    
    update_edges();
    }


ConvexArea::ConvexArea
    (
    const SVector2 &    origin,
    Float32             radius,
    Int32               num_vertices
    )
    {
    initialize_object();
    
    if( num_vertices < 3 )
        {
        num_vertices = 3;
        }
    
    Int32   i;
    Float32 angle_delta = 2.0f * MATHF_PI / static_cast< Float32 >( num_vertices );
    
    for( i = 0; i < num_vertices; ++i )
        {
        Float32 angle = angle_delta * static_cast< Float32 >( i );
        Float32 x = - FloatTraits< Float32 >::sin( angle );
        Float32 y = FloatTraits< Float32 >::cos( angle );
        add_vertex
            (
            SVector2
                (
                origin.x() + x * radius,
                origin.y() + y * radius
                )
            );
        }

    update_edges();
    }


ConvexArea::ConvexArea
    (
    const ConvexArea & area
    ) : ValueType( area ),
        vertices( area.vertices ),
        edges( area.edges )
    {
    initialize_object();
    }


void ConvexArea::add_vertex
    (
    const SVector2 & vertex
    )
    {
    vertices.push_back( vertex );
    }


void ConvexArea::rotate
    (
    const SVector2 &    center,
    Float32             angle
    )
    {
    SVector2 x_base
        ( 
        -FloatTraits<Float32>::sin( angle + MATHF_PI_2 ),
        FloatTraits<Float32>::cos( angle + MATHF_PI_2 )
        );
    
    SVector2 y_base
        ( 
        -x_base.y(),
        x_base.x()
        );

    std::size_t i;

    // Rotate vertices.

    for( i = 0; i < vertices.size(); i++ )
        {
        SVector2 v = vertices[ i ];

        v -= center;
        v = x_base * v.x() + y_base * v.y();
        v += center;

        vertices( i ) = v;
        }

    // Rotate edges.

    for( i = 0; i < edges.size(); i++ )
        {
        Edge & edge = edges[ i ];
        
        edge.normal = x_base * edge.normal.x() + y_base * edge.normal.y();
        edge.displacement = -vertices[ i ] * edge.normal;
        }
    }


void ConvexArea::translate
    (
    const SVector2 &    offset
    )
    {
    std::size_t i;
    
    // Translate vertices.

    for( i = 0; i < vertices.size(); i++ )
        {
        vertices( i ) = vertices[ i ] + offset;
        }

    // Update edges.

    for( i = 0; i < edges.size(); i++ )
        {
        Edge & edge = edges[ i ];
        edge.displacement = -vertices[ i ] * edge.normal;
        }
    }


void ConvexArea::update_edges()
    {
    std::size_t i;

    if( vertices.size() < 3 )
        {
        edges.clear();
        return;
        }

    edges.resize( vertices.size() );

    for( i = 0; i < vertices.size(); i++ )
        {
        const SVector2 & first = vertices[ i ];
        const SVector2 & second = vertices[ next_index( i, vertices ) ];

        edges[ i ].init( first, second );
        }
    }


bool ConvexArea::contains
    (
    const SVector2 & point
    ) const
    {
    std::size_t i;

    // @c point is inside iff it lays in all positive halfspaces
    // defined by the edges.

    for( i = 0; i < edges.size(); i++ )
        {
        if( edges[ i ].distance_is_negative( point ) )
            {
            // @c point lays "behind" this edge.

            return false;
            }
        }

    return true;
    }


bool ConvexArea::front_to
    (
    const Edge & edge
    ) const
    {
    std::size_t i;

    for( i = 0; i < vertices.size(); i++ )
        {
        if( edge.distance_is_negative( vertices[ i ] ) )
            {
            return false;
            }
        }
    
    return true;
    }


bool ConvexArea::back_to
    (
    const Edge & edge
    ) const
    {
    std::size_t i;

    for( i = 0; i < vertices.size(); i++ )
        {
        if( edge.distance_is_positive( vertices[ i ] ) )
            {
            return false;
            }
        }
    
    return true;
    }


bool ConvexArea::intersects
    (
    const ConvexArea & area
    ) const
    {
    std::size_t i;

    // The areas do not intersect if there is a plane (line)
    // that seperates the two areas.
    // If there is a such line then the two areas can be separated
    // by at least one of their edges.
    
    for( i = 0; i < edges.size(); i++ )
        {
        if( area.back_to( edges[ i ] ) )
            {
            return false;
            }
        }
    
    for( i = 0; i < area.edges.size(); i++ )
        {
        if( back_to( area.edges[ i ] ) )
            {
            return false;
            }
        }

    return true;
    }


bool ConvexArea::contains
    (
    const ConvexArea & area
    ) const
    {
    std::size_t count = 0;
    std::size_t i;

    for( i = 0; i < edges.size(); i++ )
        {
        if( area.front_to( edges[ i ] ) )
            {
            count++;
            }
        }

    return count == edges.size();
    }


ConvexArea ConvexArea::split
    (
    const Edge & edge
    ) const
    {
    ConvexArea  new_area;
    bool        previous_positive;
    std::size_t i;

    if( vertices.size() < 3 )
        {
        return new_area;
        }

    previous_positive = edge.distance_is_positive( vertices[ 0 ] );

    for( i = 1; i <= vertices.size(); i++ )
        {
        const SVector2 &    previous_vertex = vertices[ prev_index( i, vertices ) ];
        const SVector2 &    current_vertex = vertices[ normalize_index( i, vertices ) ];
        const bool          positive = edge.distance_is_positive( current_vertex );

        if( positive && previous_positive )
            {
            new_area.add_vertex( current_vertex );
            }
        else
        if( ( !positive && previous_positive ) || ( positive && !previous_positive ) )
            {
            // [ previous_vertex -> current_vertex ] edge is intersected by @c edge.
            // Compute the intersection point.

            SVector2    u = previous_vertex - current_vertex;
            Float32     t = -( edge.displacement + current_vertex * edge.normal ) /
                            ( u * edge.normal );
            SVector2    vertex = current_vertex + u * t;

            hope( FloatTraits<Float32>::epsilon_equals( edge.distance( vertex ), 0.0f ) );
            new_area.add_vertex( vertex );
            }

        previous_positive = positive;
        }

    // Done.

    if( new_area.vertices.size() < 3 )
        {
        new_area.vertices.clear();
        new_area.edges.clear();
        }

    new_area.update_edges();

    return new_area;
    }


ConvexArea ConvexArea::intersect_with
    (
    const ConvexArea & area
    ) const
    {
    ConvexArea  new_area( *this );
    std::size_t i;

    for( i = 0; i < area.edges.size(); i++ )
        {
        new_area = new_area.split( area.edges[ i ] );
        }

    return new_area;
    }


namespace {
    
const Float32   a_really_big_number = std::numeric_limits<Float32>::max();
const Int32     int_fixed_point_constant = 65536;
const Float32   fixed_point_constant = 65536.0f;

Int32 fixed_ceil
    (
    Int32 fixed_int
    )
    {
    return  ( fixed_int + ( fixed_int >= 0 ? int_fixed_point_constant - 1 : 0 ) ) / 
            int_fixed_point_constant * int_fixed_point_constant;
    }

}


void ConvexArea::rasterize
    (
    const RasterizerCallback &  callback,
    Int32                       clip_min_x,
    Int32                       clip_min_y,
    Int32                       clip_max_x,
    Int32                       clip_max_y
    ) const
    {
    Float32         min_y = a_really_big_number;
    Float32         max_y = -a_really_big_number;
    int             top_index = -1;
    int             bottom_index = -1;
    std::size_t     i;

    // Find "top" and "bottom" vertices.

    if( vertices.size() < 3 )
        {
        return;
        }

    for( i = 0; i < vertices.size(); i++ )
        {
        Float32 y = vertices[ i ].y();

        if( y < min_y )
            {
            min_y = y;
            bottom_index = i;
            }
        
        if( y > max_y )
            {
            max_y = y;
            top_index = i;
            }
        }

    // Initialize rasterizer.

    Int32   int_current_y = FloatTraits<Float32>::ceil_int( min_y );
    Int32   int_max_y = FloatTraits<Float32>::ceil_int( max_y );
    int     left_index = prev_index( bottom_index, vertices );
    int     right_index = next_index( bottom_index, vertices );
    Int32   int_left_stop_y = FloatTraits<Float32>::ceil_int( vertices[ left_index ].y() );
    Int32   int_right_stop_y = FloatTraits<Float32>::ceil_int( vertices[ right_index ].y() );
    Int32   fixed_left_x_step = FloatTraits<Float32>::to_integer
            ( ( vertices[ left_index ].x() - vertices[ bottom_index ].x() ) /
              ( vertices[ left_index ].y() - vertices[ bottom_index ].y() ) *
              fixed_point_constant );
    Int32   fixed_right_x_step = FloatTraits<Float32>::to_integer
            ( ( vertices[ right_index ].x() - vertices[ bottom_index ].x() ) /
              ( vertices[ right_index ].y() - vertices[ bottom_index ].y() ) *
              fixed_point_constant );
    Int32   fixed_left_x = FloatTraits<Float32>::to_integer
            ( vertices[ bottom_index ].x() * fixed_point_constant );
    Int32   fixed_right_x = FloatTraits<Float32>::to_integer
            ( vertices[ bottom_index ].x() * fixed_point_constant );
    Int32   int_left_x;
    Int32   int_right_x;
    
    // Clip.

    if( int_current_y < clip_min_y )
        {
        int_current_y = clip_min_y;
        }

    if( int_max_y > clip_max_y )
        {
        int_max_y = clip_max_y;
        }

    // Rasterize.
    
    fixed_left_x += FloatTraits<Float32>::to_integer
        ( fixed_left_x_step * ( int_current_y - vertices[ bottom_index ].y() ) );
    fixed_right_x += FloatTraits<Float32>::to_integer
        ( fixed_right_x_step * ( int_current_y - vertices[ bottom_index ].y() ) );

    while( int_current_y < int_max_y )
        {
        // Do we have reached an end of the current edge on the left or the right sides?
        // When such an event occurs we must already be "inside" the next edge.

        if( int_current_y >= int_left_stop_y )
            {
            // We have left the left edge.
            
            while( int_current_y >= int_left_stop_y && left_index != top_index )
                {
                int next_edge_start = left_index;
                
                hope( int_current_y - vertices[ next_edge_start ].y() >= 0 );

                left_index = prev_index( left_index, vertices );
                fixed_left_x = FloatTraits<Float32>::to_integer
                    ( vertices[ next_edge_start ].x() * fixed_point_constant );
                int_left_stop_y = FloatTraits<Float32>::ceil_int
                    ( vertices[ left_index ].y() );
                fixed_left_x_step = FloatTraits<Float32>::to_integer
                    ( ( vertices[ left_index ].x() - vertices[ next_edge_start ].x() ) /
                      ( vertices[ left_index ].y() - vertices[ next_edge_start ].y() ) *
                      fixed_point_constant );
                fixed_left_x += FloatTraits<Float32>::to_integer
                    ( fixed_left_x_step * ( int_current_y - vertices[ next_edge_start ].y() ) );
                }
            }
        
        if( int_current_y >= int_right_stop_y )
            {
            // We have left the right edge.
                
            while( int_current_y >= int_right_stop_y && right_index != top_index )
                {
                int next_edge_start = right_index;

                hope( int_current_y - vertices[ next_edge_start ].y() >= 0 );
            
                right_index = next_index( right_index, vertices );
                fixed_right_x = FloatTraits<Float32>::to_integer
                    ( vertices[ next_edge_start ].x() * fixed_point_constant );
                int_right_stop_y = FloatTraits<Float32>::ceil_int
                    ( vertices[ right_index ].y() );
                fixed_right_x_step = FloatTraits<Float32>::to_integer
                    ( ( vertices[ right_index ].x() - vertices[ next_edge_start ].x() ) /
                      ( vertices[ right_index ].y() - vertices[ next_edge_start ].y() ) *
                      fixed_point_constant );
                fixed_right_x += FloatTraits<Float32>::to_integer
                    ( fixed_right_x_step * ( int_current_y - vertices[ next_edge_start ].y() ) );
                }
            }

        // Compute current row bounds.
                
        int_left_x = fixed_ceil( fixed_left_x ) / int_fixed_point_constant;
        int_right_x = fixed_ceil( fixed_right_x ) / int_fixed_point_constant;

        if( int_left_x < clip_min_x )
            {
            int_left_x = clip_min_x;
            }

        if( int_right_x > clip_max_x )
            {
            int_right_x = clip_max_x;
            }

        // Fill current row.
        // It is ensured that SOUTH WEST corners of the reported cells
        // belong to the rasterized area.

        if( !callback( int_left_x, int_right_x, int_current_y ) )
            {
            break;
            }

        // Proceed to the next row.

        fixed_left_x += fixed_left_x_step;
        fixed_right_x += fixed_right_x_step;

        int_current_y++;
        }
    }


} // namespace Demo::Lib
} // namespace Demo
