/** @file convex_area.h
 * Convex area.
 */

/*
 * begin        : Wed Aug 6 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CONVEX_AREA_H
#define DEMO_CONVEX_AREA_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Convex area defined by its 2D vertices.
 * The area is used for bounding areas of sector tile maps.
 * The 2D coordinate system is oriented in the same way as the layer 2D
 * space (x coordinate EAST, y coordinate NORTH).
 */
class ConvexArea : public ValueType
    {
public:

    /**
     * Area edge.
     */
    class Edge : public ValueType
        {
    public:

        /** Default constructor. */
        Edge();

        /**
         * Construct an edge from @c first to @c second vertices.
         * See @c init().
         */
        Edge
            (
            const SVector2 & first,
            const SVector2 & second
            );

        /** Copy constructor. */
        Edge
            (
            const Edge & edge
            );

        /**
         * Constructs an edge from @c first to @c second vertices.
         * The edge orientation is chosen so that if a set of edges of an
         * polygon defined by vertices in CCW order is constructed, all
         * the edges are oriented inside the polygon.
         */
        void init    
            (
            const SVector2 & first,
            const SVector2 & second
            );
        
        /**
         * Computes distance from @c point to the edge.
         * Positive if @c point lays in front of the edge, negative otherwise.
         */
        Float32 distance
            (
            const SVector2 & point
            ) const;

        /** Tests if distance from @c point to the edge is positive. */
        bool distance_is_positive
            (
            const SVector2 & point
            ) const;
        
        /** Tests if distance from @c point to the edge is positive. */
        bool distance_is_negative
            (
            const SVector2 & point
            ) const;

    public:

        PVector2    normal;
            /**< Edge normal. Its direction determines edge orientation.
                 Always normalized. */
        PFloat32    displacement;
            /**< Edge displacement / distance from [ 0, 0 ]. */

        }; // Edge

    /**
     * Rasterizer call-back functor used by @c rasterize().
     * Not a managed object.
     */
    class RasterizerCallback
        {
    public:

        /**
         * Destructor.
         */
        virtual ~RasterizerCallback();
            
        /**
         * Called by the rasterizer to notify that SOUTH WEST corners of
         * reported cells [ x, y ] (where begin_x <= x < end_x) were classified
         * "inside" the area.
         * The callback returns true if the rasterization should continue.
         */
        virtual bool operator()
            (
            Int32   begin_x,
            Int32   end_x,
            Int32   y
            ) const = 0;

        }; // RasterizerCallback

public:
        
    /**
     * Default constructor.
     * Constructs an empty area.
     */
    ConvexArea();

    /**
     * Constructs area that is actually an axis-aligned bounding box
     * with @c origin and @c bounding_box_size sizes.
     * If @c ajdust_to_grid is true, bounding box corners will be rounded
     * so that their coordinates would fit into the grid.
     */
    ConvexArea
        (
        const SVector2 &    origin,
        const SVector2 &    bounding_box_size,
        bool                adjust_to_grid = false
        );
    
    /**
     * Construct area that aproximates a circle.
     * @param origin Origin of the circle.
     * @param radius Radius of the circle.
     * @param num_vertices Number of vertices to generate.
     *      If lesser than three, three will be used instead.
     */
    ConvexArea
        (
        const SVector2 &    origin,
        Float32             radius,
        Int32               num_vertices
        );

    /** Copy constructor. */
    ConvexArea
        (
        const ConvexArea & area
        );

    /**
     * Adds vertex to the area.
     * The resulting area must still be convex, the vertices must
     * be inserted in CCW order.
     * Once all vertices have been added, call @c update_edges()
     * to cache edge info.
     */
    void add_vertex
        (
        const SVector2 & vertex
        );
    
    /**
     * Rotates the area around @c center by given @c angle.
     * The edges are updated automatically.
     */
    void rotate
        (
        const SVector2 &    center,
        Float32             angle
        );

    /**
     * Translates the area by @c offset.
     */
    void translate
        (
        const SVector2 &    offset
        );

    /**
     * Generates edges from @c vertices list.
     */
    void update_edges();

    /** Tests if @c point lays inside the area. */
    bool contains
        (
        const SVector2 & point
        ) const;
    
    /** Tests if area lays in @c edge's positive halfspace. */
    bool front_to
        (
        const Edge & edge
        ) const;

    /** Tests if area lays in @c edge's negative halfspace. */
    bool back_to
        (
        const Edge & edge
        ) const;

    /** Tests if the area intersects other @c area. */
    bool intersects
        (
        const ConvexArea & area
        ) const;

    /** Tests if @c area is a subset of this area. */
    bool contains
        (
        const ConvexArea & area
        ) const;
    
    /** Returns intersection of the area and @c edge's positive halfspace. */
    ConvexArea split
        (
        const Edge & edge
        ) const;
    
    /** Returns intersection of the two areas. */
    ConvexArea intersect_with
        (
        const ConvexArea & area
        ) const;

    /**
     * Enumerates cells of a discrete raster with unit (1.0) scales that are
     * contained in the area, a cell is assumed to be contained in the area
     * if its SOUTH WEST corner lies inside the area.
     *
     * A generic convex polygon filler.
     *
     * [ @c clip_min_y, @c clip_max_y ) determines which "y-values" the
     * caller is interested in. Similary for @c clip_min_x, @c clip_max_x.
     */
    void rasterize
        (
        const RasterizerCallback &  callback,
        Int32                       clip_min_x = -10000,
        Int32                       clip_min_y = -10000,
        Int32                       clip_max_x = 10000,
        Int32                       clip_max_y = 10000
        ) const;

public:

    PValueArray<SVector2>   vertices;
        /**< List of area vertices in layer 2D coordinates. */
    PArray<Edge>            edges;
        /**< List of area edges. Always computed from @c vertices
             by @c update_edges(). */
    
    }; // ConvexArea


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_CONVEX_AREA_H
