/** @file entity.cpp
 * Entity implementation.
 */

/*
 * begin        : Mon May 19 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/server/entity.h"
#include "lib/server/entity_rpc.h"
#include "lib/server/sector_rpc.h"
#include "lib/server/layer_rpc.h"
#include "lib/server/ability_modifier.h"
#include "lib/server/player.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/client_light.h"
#include <sstream>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of Entity
// ~~~~~~~~~~~~~~~~~~~~~~~~

void Entity::initialize()
    {
    client_entity = CreateObject< ClientEntity >( this );
    entity_flags = CLASS_UNKNOWN;
    }


bool Entity::is_player() const
    {
    return match_entity_flags( CLASS_PLAYER );
    }


bool Entity::match_entity_flags
    (
    int flags
    ) const
    {
    return flags_on( entity_flags, flags );
    }


void Entity::put_into_layer
    (
    const EntityPosition & position
    )
    {
    move_entity( position );
    }


void Entity::remove_from_layer()
    {
    move_entity( EntityPosition() );
    }


EntityPosition Entity::get_current_position() const
    {
    return client_entity->current_position;
    }


Pointer<Sector> Entity::get_current_sector() const
    {
    return current_sector;
    }


Remote<Sector> Entity::get_last_sector() const
    {
    if( current_sector != null )
        {
        return current_sector;
        }
    else
    if( last_sector != null )    
        {
        return last_sector;
        }
    else
        {
        throw EntityErrorException( "Entity was not put into map yet." );
        }
    }


Pointer<ClientEntity> Entity::get_client_entity() const
    {
    return client_entity;
    }


std::string Entity::get_name() const
    {
    return client_entity->name.read();
    }
    

void Entity::set_name
    (
    const std::string & new_name
    )
    {
    client_entity->name = new_name;
    }


bool Entity::is_light() const
    {
    return !client_entity->light.is_null();
    }


Pointer< ClientLight > Entity::get_light() const
    {
    return client_entity->light;
    }


void Entity::create_light()
    {
    client_entity->light = CreateObject< ClientLight >();
    }


void Entity::apply_powerup
    (
    AbilityId       ability,
    Int32           ability_value_delta,
    Int32           ability_upper_limit_delta,
    PowerUpType     powerup_type,
    const STime &   duration
    )
    {
    if( !client_entity->supports_ability( ability ) )
        {
        return;
        }
    
    Int32 old_value = client_entity->get_ability_value( ability );
    Int32 old_upper_limit = client_entity->get_ability_upper_limit( ability );

    switch( powerup_type )
        {
        case POWERUP_PERMANENT:
            client_entity->add_ability_upper_limit( ability, ability_upper_limit_delta );
            client_entity->add_ability_value( ability, ability_value_delta );
            break;

        case POWERUP_TEMPORARY:
            CreateObject< AbilityModifier >
                (
                this,
                ability,
                ability_value_delta,
                ability_upper_limit_delta,
                duration
                );
            break;
        }
    
    if( is_player() )
        {
        Int32 new_value = client_entity->get_ability_value( ability );
        Int32 new_upper_limit = client_entity->get_ability_upper_limit( ability );
        Pointer< Player > player = Pointer< Entity >( this ).convert();
        std::ostringstream ss;
        ss  << "Your " << ClientEntity::ability_to_string( ability )
            << " changed from " << old_value << "/" << old_upper_limit
            << " to " << new_value << "/" << new_upper_limit << ".";
        player->log_game_message( ss.str() );
        }
    }


void Entity::move_entity
    (
    const EntityPosition & target
    )
    {
    if( moving )
        {
        // Previous move is still in progress.
        // Remember the new target.

        destination_position = target;
        destination_dirty = true;
        }
    else
        {
        move_position = target;
        do_move();
        }
    }


void Entity::translate_entity
    (
    const SVector2 & origin
    )
    {
    EntityPosition position = client_entity->current_position;

    position.origin = origin;
    move_entity( position );
    }


void Entity::rotate_entity
    (
    const SOrientation & orientation
    )
    {
    EntityPosition position = client_entity->current_position;
    
    position.orientation = orientation;
    move_entity( position );
    }


void Entity::set_model_tags
    (
    const SString & model_tags
    )
    {
    client_entity->model_tags = model_tags;
    }
    

void Entity::entity_moved()
    {
    }


void Entity::move_failed
    (
    const CollisionInfo & collision_info
    )
    {
    }


void Entity::operate_synced
    (
    const STime & delivery_time
    )
    {
    }


Remote<Sector> Entity::get_sector
    (
    const SectorAddress & address
    )
    {
    if( address.layer != null )
        {
        if( current_sector != null )
            {
            return current_sector->get_sector( address );
            }
        else
            {
            Remote< Layer > layer = address.layer.convert();
            return layer->sync_get_sector( address );
            }
        }
    else
        {
        return null;
        }
    }


void Entity::do_move()
    {
    // Initiate move.

    moving = true;
    
    // Initialize.
    
    bool            unlink_from_layer = move_position.layer == null;
    SectorAddress   new_address( move_position.layer, move_position.origin.read() );
    Remote<Sector>  new_sector( get_sector( new_address ) );

    // Reserve new tiles.
    // If unlinking from the layer then the reservation will automatically
    // succeed but no reservation will actually be done.

    bool            reservation_succeeded = unlink_from_layer;
    CollisionInfo   collision_info;
    
    if( new_sector != null )
        {
        hope( !unlink_from_layer );

        reservation_succeeded = new_sector->sync_reserve_tiles
            (
            move_position,
            physical_info,
            entity_flags,
            this,
            collision_info
            );
        }

    if( reservation_succeeded )
        {
        // The reservation succeeded so the old reservation for current
        // position can be released.
        // If @c current_sector is null then there are no tiles to release.

        if( current_sector != null )
            {
            current_sector->release_tiles
                (
                client_entity->current_position,
                physical_info,
                entity_flags,
                this
                );
            }

        if( !client_entity->current_position.is_same_sector( new_address ) &&
            !unlink_from_layer )
            {
            // Moving to a different sector.
            // Unlink entity from the current sector.

            if( current_sector != null )
                {
                current_sector->unlink_entity( this );
                }
            
            // Migrate the entity to the new sector.
            // When the migration finishes @c delivered_to() will be called
            // and entity linked to the new sector.

            hope( new_sector != null );
            deliver_asap_to( new_sector );
                
            return;
            }
        else
        if( !unlink_from_layer )    
            {
            // Moving to the same sector.

            client_entity->current_position = move_position;
            }
        else
            {
            // Unlinking from layer.
            // Unlink entity from the current sector and reset its
            // position.

            if( current_sector != null )
                {
                current_sector->unlink_entity( this );
                }

            client_entity->current_position = EntityPosition();
            last_sector = null;
            }
        }

    // Move finished.

    moving = false;

    // If the entity wants to move elsewhere initiate next move.
    // This happens if @c move_entity() was called while previous
    // move was still in progress.

    if( destination_dirty )
        {
        move_position = destination_position;
        destination_dirty = false;

        do_move();
        }

    // Notify client.

    if( reservation_succeeded )
        {
        entity_moved();
        }
    else
        {
        move_failed( collision_info );
        }
    }


void Entity::delivered_to
    (
    WeakPointer<Object> object,
    const STime &       time
    )
    {
    // The entity migration finished.
    // Link the entity to the new sector.

    Pointer<Sector> sector = object.convert();
    sector->link_entity( this );

    // Finish the move.

    client_entity->current_position = move_position;
    moving = false;

    if( destination_dirty )
        {
        move_position = destination_position;
        destination_dirty = false;
        
        do_move();
        }

    // Notify client.

    entity_moved();
    }
   

} // namespace Demo::Lib
} // namespace Demo
