/** @file entity.h
 * Entity.
 */

/*
 * begin        : Mon May 19 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_ENTITY_H
#define DEMO_ENTITY_H

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_LOGIC_H
#include "lib/server/logic.h"
#endif

#ifndef DEMO_PHYSICAL_INFO_H
#include "lib/server/physical_info.h"
#endif

#ifndef DEMO_ENTITY_POSIION_H
#include "lib/shared/entity_position.h"
#endif

#ifndef DEMO_GAME_DEFS_H
#include "lib/shared/game_defs.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();

BASIC_LIB_EXCEPTION( EntityErrorException, ::Massiv::Core::Lib::RuntimeException );


/**
 * Object within a virtual world.
 */
class Entity : public Logic
    {
public:

    MASSIV_OBJECT( Entity );

    friend class Sector;
    friend class AbilityModifier; // Access to client_entity.
    friend class EntityFactory; // Access to client_entity, physical_info.
    friend class EntityTest;

    /**
     * Entity classes flags.
     */
    enum EntityClass
        {
        // Class flags.

        CLASS_UNKNOWN   = 1,
            /**< Unknown entity class. */
        CLASS_MONSTER   = 2,
            /**< The entity is ancestor of #Monster class. */
        CLASS_PLAYER    = 4,
            /**< The entity is ancestor of #Player class. */

        // Class flags combinations.

        MONSTER         = CLASS_MONSTER,
            /**< Monster. */
        PLAYER          = CLASS_MONSTER | CLASS_PLAYER
            /**< Player. */
        }; // EntityClass

public:

    /** Ability power up type. */
    enum PowerUpType
        {
        POWERUP_PERMANENT,
            /**< Pemanent powerup duration. */
        POWERUP_TEMPORARY
            /**< Time limited powerup duration. */
        }; // PowerUpType

public:

    /**
     * Initializes entity.
     * The initialization leaves the object default-constructed.
     * Especially @c physical_info is [ CLASS_PENETRABLE, no bounding box ].
     *
     * Any subclass must properly override this if needed, which allows
     * to automated entity creation and initialization.
     */
    virtual void initialize();

public:

    /**
     * Is this entity a player?
     * @return True if and only if the entity is convertible
     *      to #Player. False otherwise.
     *
     * Example:
     * @code
     * Pointer< Entity > entity;
     * if( entity->is_player() )
     *     {
     *     Pointer< Player > player = entity.conver();
     *     player->log_game_message( "Hello! The entity is player." );
     *     }
     * @endcode
     */
    bool is_player() const;

    /**
     * Tests if entity's flags match @c flags.
     */
    bool match_entity_flags
        (
        int flags
        ) const;

public:

    /**
     * Register the entity into layer.
     * This function tries to link the entity to some sector based on
     * @a position. If target position can not be reserved, @c move_failed()
     * will be called and entity will not be linked to any sector.
     *
     * When an entity is created it is not placed into map (linked
     * to a sector) by default. This function forces the entity
     * to do first step in the map.
     *
     * @warning Calls SRPC.
     */
    virtual void put_into_layer
        (
        const EntityPosition & position
        );

    /**
     * Remove entity from layer.
     * Unlinks the entity from the current sector.
     */
    virtual void remove_from_layer();

public:

    /**
     * Get current position of the entity.
     */
    EntityPosition get_current_position() const;

    /**
     * Get sector the entity is currently linked to.
     * @warning May return null if entity is moving from sector to a sector.
     */
    Pointer<Sector> get_current_sector() const;

    /**
     * Get sector the entity was linked to for the last time.
     * Always returns a non-null pointer.
     * Throws EntityErrorException if entity was not put into the map yet.
     */
    Remote<Sector> get_last_sector() const;

    /**
     * Get client entity object associated with this entity.
     */
    Pointer<ClientEntity> get_client_entity() const;

    /**
     * Get name of the entity.
     */
    std::string get_name() const;

    /**
     * Set name of the entity.
     */
    void set_name
        (
        const std::string & new_name
        );

public:

    /**
     * Is this entity light source?
     */
    bool is_light() const;

    /**
     * Get pointer to light source.
     *
     * Returns @c NULL pointer if <tt>!is_light()</tt>.
     */
    Pointer< ClientLight > get_light() const;

    /**
     * Create light source for this entity.
     *
     * Replaces current light source by new default-constructed one if entity
     * already has light source.
     */
    void create_light();

public:

    /**
     * Apply powerup to the monster.
     * @param ability Ability to modify.
     * @param ability_value_delta This value is added to current value
     *      of the ability.
     * @param ability_upper_limit_delta This value is added to current
     *      upper limit of the ability.
     * @param powerup_type Powerup duration type.
     * @param duration Duration of the powerup. If @a powerup_type
     *      is @a POWERUP_PERMANENT, the @a duration is ignored.
     *      If @a powerup_type is @a POWERUP_TEMPORARY, the ability
     *      is changed only for @a duration seconds, then the effect
     *      of the powerup is cancelled.
     * @note Upper limit of the ability is changes before uptaing
     *      current value of the ability.
     */
    void apply_powerup
        (
        AbilityId       ability,
        Int32           ability_value_delta,
        Int32           ability_upper_limit_delta,
        PowerUpType     powerup_type,
        const STime &   duration
        );

public:

    //
    // Entity interface.
    //

    /**
     * Schedules the entity to be moved to @c target.
     * When done @c entity_moved() call-back will be invoked.
     * If entity can not move to @c target @c move_failed() will be called.
     * It is safe to call this when a previous move is in progress.
     * In that case move target is changed.
     */
    void move_entity
        (
        const EntityPosition & target
        );

    /**
     * Restricted version of @c move_entity().
     * Allows to translate the entity within the same layer.
     */
    void translate_entity
        (
        const SVector2 & origin
        );

    /**
     * Restricted version of @c move_entity().
     * Allows to rotate the entity.
     */
    void rotate_entity
        (
        const SOrientation & orientation
        );

    /**
     * Set new model tags for client entity.
     */
    void set_model_tags
        (
        const SString & model_tags
        );

protected:

    /**
     * Call-backed when move request (submitted by @c move_entity())
     * completed.
     */
    virtual void entity_moved();

    /**
     * Call-backed when move request (submitted by @c move_entity())
     * can not complete. @c collision_info describes why move failed.
     */
    virtual void move_failed
        (
        const CollisionInfo & collision_info
        );

protected:

    //
    // Logic interface implementation.
    //

    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

private:

    /**
     * Returns pointer to sector identified by @c address.
     */
    Remote<Sector> get_sector
        (
        const SectorAddress & address
        );

    /**
     * Moves the entity to @c move_position.
     * Used by @c move_entity() as a helper method.
     */
    void do_move();

    /**
     * Called by system when entity was "delivered" to @c object.
     * @c object is assumed to be a Sector.
     */
    virtual void delivered_to
        (
        WeakPointer<Object> object,
        const STime &       time
        );

protected:

    PPointer<ClientEntity>  client_entity;
        /**< Client visible part of the entity. */
    PPointer<Sector>        current_sector;
        /**< Current sector the entity is linked to. The entity may
             be unlinked (@c current_sector == null) if it is moving from
             sector to a sector. */
    PRemote<Sector>         last_sector;
        /**< Last non-null value of @c current_sector. Sector the entity
             was linked to for the last time. */
    PPointer<Enumerator>    entity_list_enumerator;
        /**< Enumerator to @c current_sector->entity_list. Managed by Sector. */
    EntityPosition          move_position;
        /**< Where the entity is just moving.
             Valid if @c moving. */
    EntityPosition          destination_position;
        /**< Where the entity wants to move.
             Valid if @c destination_dirty. */
    PBool                   moving;
        /**< True if the entity moves to @c move_position. */
    PBool                   destination_dirty;
        /**< True if the entity wants to move to @c destination_position. */
    PhysicalInfo            physical_info;
        /**< Entity physical info. Static properties only. */
    PVlInt32                entity_flags;
        /**< Entity class flags. An ORed combination of @c EntityClass flags. */

    }; // Entity


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_ENTITY_H
