/** @file entity_test.cpp
 * Entity test implementation.
 */

/*
 * begin        : Mon June 13 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/server/device.h"
#include "lib/server/generator.h"
#include "lib/server/pipe.h"
#include "lib/server/convex_area.h"
#include "lib/server/spatial_query.h"
#include "lib/server/sector.h"
#include "lib/server/world.h"
#include "lib/server/layer.h"
#include "lib/shared/random.h"

#include "lib/server/entity_test.h"
#include "lib/server/entity_test_rpc.h"

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Vertex test.
    
void Vertex::create()
    {
    VertexList  list;
    std::size_t i;

    for( i = 0; i < 8; i++ )
        {
        list.push_back( CreateObject<Vertex>() );
        }

    for( i = 0; i < list.size(); i++ )
        {
        Pointer<Vertex> vertex = list[ i ];
        Remote<Vertex> v = vertex.convert();

        vertex->all_vertices = list;
        v->param( RPC_DELAYED, 3.0 ).async_change_edges();
        }
    }


void Vertex::initialize()
    {
    }


void Vertex::change_edges()
    {
    int num_edges_to_delete = Random::random_integer( 4 );
    int num_edges_to_insert = Random::random_integer( 4 );
    int i;

    for( i = 0; i < num_edges_to_delete; i++ )
        {
        const std::size_t index = Random::random_integer( neighbors.size() - 1 );
        hope( index < neighbors.size() );
        neighbors.erase( index );
        }
    
    if( neighbors.size() < all_vertices.size() / 2 )
        {
        for( i = 0; i < num_edges_to_insert; i++ )
            {
            const std::size_t index = Random::random_integer( all_vertices.size() - 1 );
            hope( index < all_vertices.size() );
            neighbors.push_back( all_vertices[ index ].convert() );
            }
        }

    Remote<Vertex> self = this;
    self->param( RPC_DELAYED, Random::random() * 5.0f ).async_change_edges();
    }


void EntityTest::generator_test()
    {
    STime   tick_period = 0.25f;
    STime   timer_period = 5.0f;
    bool    loop = true;
    bool    auto_start = false;

    // Create logic units.

    Pointer<TimerWithPeriod> timer =
        CreateObject<TimerWithPeriod>
            ( tick_period, timer_period, loop, auto_start );
    Pointer<FloatInterpolator> interpolator =
        CreateObject<FloatInterpolator>();
    Pointer<PrintMessageDevice> device =
        CreateObject<PrintMessageDevice>();

    // Plug units.

    timer->plug_output
        (
        TimerWithPeriod::get_output_name(),
        interpolator,
        FloatInterpolator::get_input_name()
        );

    interpolator->plug_output
        (
        FloatInterpolator::get_output_name(),
        device,
        PrintMessageDevice::get_input_name()
        );

    // Set interpolator function (square wave).

    interpolator->add_key_value( 0.0f, 0.0f );
    interpolator->add_key_value( 0.5f, 0.0f );
    interpolator->add_key_value( 0.5f, 1.0f );
    interpolator->add_key_value( 1.0f, 1.0f );

    // Start timer.

    timer->start();
    }


void EntityTest::sector_test()
    {
    // Build world.

    Pointer<World>  world = CreateObject<World>();
    Pointer<Layer>  layer = CreateObject<Layer>();
    int             i, j;

    for( i = -5; i <= 5; i++ )
        {
        for( j = -5; j <= 5; j++ )
            {
            Pointer<Sector> sector = CreateObject<Sector>();
            layer->register_sector( sector, SectorAddress( layer, i, j ) );
            }
        }

    // Test sector connectivity.

    Pointer<Sector> master_sector = layer->get_sector( SectorAddress( layer, 0, 0 ) );
    Pointer<Sector> north_west_corner = layer->get_sector( SectorAddress( layer, -5, 5 ) );
    Pointer<Sector> south_east_corner = layer->get_sector( SectorAddress( layer, 5, -5 ) );
    Pointer<Sector> sector = north_west_corner;

    for( i = 0; i < 10; i++ )
        {
        sector = sector->get_neighbor( SOUTH_EAST );
        }

    hope( sector == south_east_corner );

    // Test entity linkage.

    Pointer<Entity> entity1 = CreateObject<Entity>();
    Pointer<Entity> entity2 = CreateObject<Entity>();

    entity1->physical_info.bounding_box_size = SVector2( 1.0f, 1.0f );
    entity2->physical_info.bounding_box_size = SVector2( 1.0f, 1.0f );

    entity1->move_entity( EntityPosition( SVector2( 0, 0 ), SOrientation(), layer ) );
    entity1->move_entity( EntityPosition( SVector2( 3, 0 ), SOrientation(), layer ) );

    entity2->move_entity( EntityPosition( SVector2( 3, 0 ), SOrientation(), layer ) );

    // Test sector enumerators.

    master_sector->modify_elevation
        (
        SVector2(),
        ClientSector::KIND_DELTA_ELEVATION_MAP,
        ClientSector::FUNCTION_GAUSSIAN,
        ClientSector::OPERATION_ADD,
        2.0f,
        1.0f
        );
    }


// Convex area test.

void EntityTest::area_test()
    {
    ConvexArea area1( SVector2( 50.0f, 50.0f ), SVector2( 10.0f, 10.0f ) );
    ConvexArea area2( SVector2( 40.0f, 50.0f ), SVector2( 9.0f, 9.0f ) );

    hope( !area1.intersects( area2 ) );
    area1.rotate( SVector2( 50.0f, 50.0f ), MATHF_PI_4 );
    hope( area1.intersects( area2 ) );

    struct Callback : public ConvexArea::RasterizerCallback
        {
        virtual bool operator()
            (
            Int32   x1,
            Int32   x2,
            Int32   y
            ) const
            {
            return true;
            }
        } cb;

    area1.rasterize( cb );
    
    ConvexArea area3( SVector2( 50.0f, 50.0f ), SVector2( 20.0f, 20.0f ) );
    area3.split( ConvexArea::Edge( SVector2( 55.0f, 0.f ), SVector2( 60.0f, 100.0f ) ) );
    }


void EntityTest::test()
    {
    static bool initialized = false;

    if( Massiv::System::get_phase() == Massiv::System::PHASE_RUNNING &&
        !initialized )
        {
        initialized = true;

        //Vertex::create();
        //generator_test();
        //area_test();
        sector_test();
        }
    }

} // namespace Demo::Lib
} // namespace Demo
