/** @file generator.cpp
 * Generator implementation.
 */

/*
 * begin        : Mon June 13 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/server/generator.h"
#include "lib/server/generator_rpc.h"
#include "lib/shared/random.h"
#include "lib/shared/client_entity.h"
#include <algorithm>


namespace Demo {
namespace Lib {

using namespace Massiv::Core;


// Implementation of Generator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Generator::initialize()
    {
    Entity::initialize();
    create_output( get_output_name() );
    }


Logic::SocketId Generator::get_output_id()
    {
    // Generator has hardwired default output called "output"
    // whose socket id is always 0.

    return 0;
    }


std::string Generator::get_output_name()
    {
    return "output";
    }


void Generator::generate
    (
    Pointer<Packet> packet,
    const STime &   delivery_time
    )
    {
    output( get_output_id(), packet, delivery_time );
    }


Logic::SocketId Generator::translate_output
    (
    const std::string & output_name
    )
    {
    if( output_name == get_output_name() )
        {
        return get_output_id();
        }

    return Logic::translate_output( output_name );
    }


Logic::SocketId Generator::create_output
    (
    const std::string & dynamic_name
    )
    {
    if( dynamic_name != get_output_name() || outputs.size() )
        {
        throw GeneratorHasDefaultOutputOnlyException();
        }

    SocketId id = Logic::create_output();
    hope( id == get_output_id() );

    return id;
    }


void Generator::operate_synced
    (
    const STime & delivery_time
    )
    {
    }


// Implementation of Timer
// ~~~~~~~~~~~~~~~~~~~~~~~

void Timer::initialize()
    {
    initialize( STime(), false );
    }


void Timer::initialize
    (
    const STime &   period,
    bool            auto_start
    )
    {
    Generator::initialize();

    tick_period = period;
    running = false;

    enable_id = create_input();

    if( auto_start )
        {
        start();
        }
    }


void Timer::start()
    {
    start_time = STime::current_time();

    if( !running )
        {
        schedule_next_tick();
        }
    else
    if( request_stop )
        {
        request_stop = false;
        }
    }


void Timer::stop()
    {
    if( running )
        {
        // The timer will be stopped by @c timer_ticked() which is already
        // scheduled.

        request_stop = true;
        }
    }


void Timer::setup_timer
    (
    const STime & new_period
    )
    {
    tick_period = new_period;
    }


void Timer::timer_ticked()
    {
    if( request_stop )
        {
        request_stop = false;
        running = false;
        }

    if( running )
        {
        next_packet( STime::current_time() - start_time );
        schedule_next_tick();
        }
    }


Logic::SocketId Timer::translate_input
    (
    const std::string & input_name
    )
    {
    if( input_name == "enable" )
        {
        return enable_id;
        }

    return Generator::translate_input( input_name );
    }


void Timer::operate_synced
    (
    const STime & delivery_time
    )
    {
    // Check "enable" input.

    Pointer<BooleanPacket> packet = get_input( enable_id ).convert();

    if( packet != null )
        {
        bool value = packet->get_boolean();

        if( running.read() && !value )
            {
            stop();
            }
        else
        if( !running && value )
            {
            start();
            }
        }
    }


void Timer::schedule_next_tick()
    {
    Remote<Timer> self = this;

    running = true;
    self->param( RPC_DELAYED, tick_period ).async_timer_ticked();
    }


// Implementation of TimerWithPeriod
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void TimerWithPeriod::initialize()
    {
    initialize( STime(), STime(), true, false );
    }


void TimerWithPeriod::initialize
    (
    const STime &   tick_period,
    const STime &   period,
    bool            enable_looping,
    bool            auto_start
    )
    {
    timer_period = period;
    loop = enable_looping;
    Timer::initialize( tick_period, auto_start );
    }


void TimerWithPeriod::setup_timer_with_period
    (
    const STime &   new_tick_period,
    const STime &   new_period,
    bool            enable_looping
    )
    {
    timer_period = new_period;
    loop = enable_looping;
    setup_timer( new_tick_period );
    }


Float32 TimerWithPeriod::get_fraction
    (
    const STime & delta
    )
    {
    Float64 fraction = delta.get_value() / timer_period.read().get_value();

    if( fraction >= 1.0f )
        {
        if( !loop )
            {
            stop();
            return 0.0f;
            }

        fraction -= FloatTraits<Float64>::floor( fraction );
        }

    return static_cast<Float32>( fraction );
    }


void TimerWithPeriod::next_packet
    (
    const STime & delta
    )
    {
    generate
        (
        CreateObject<FloatPacket>( get_fraction( delta ) ),
        STime::current_time()
        );
    }


void TimerWithPeriod::operate_synced
    (
    const STime & delivery_time
    )
    {
    // Check "enable" input.

    Pointer<BooleanPacket> packet = get_input( enable_id ).convert();

    if( packet != null )
        {
        bool value = packet->get_boolean();

        if( running.read() && !value )
            {
            // Disable looping which makes the timer to stop at the end
            // of its period.

            loop = false;
            return;
            }
        }

    Timer::operate_synced( delivery_time );
    }


// Implementation of WaveGenerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void WaveGenerator::initialize()
    {
    TimerWithPeriod::initialize();
    }


void WaveGenerator::setup_wave_generator
    (
    const STime &   new_tick_period,
    const STime &   new_period,
    bool            enable_looping,
    WaveType        wave
    )
    {
    wave_type = wave;
    setup_timer_with_period( new_tick_period, new_period, enable_looping );
    }


Float32 WaveGenerator::get_function_value
    (
    const STime & delta
    )
    {
    Float32 fraction = get_fraction( delta );
    Float32 value;

    switch( wave_type )
        {
        case WAVE_SINE:
            value = FloatTraits<Float32>::sin( fraction * 2.0f * MATHF_PI );
            break;

        case WAVE_SQUARE:
            value = fraction >= 0.5f ? 1.0f : -1.0f;
            break;

        case WAVE_SAW:
            value = fraction >= 0.5f ? 1.0f - 4.0f * ( fraction - 0.5f ) :
                -1.0f + fraction * 4.0f;
            break;

        case WAVE_RANDOM:
            value = Random::random() * 2.0f - 1.0f;
            break;

        default:
            value = 0.0f;
        }

    return value;
    }


void WaveGenerator::next_packet
    (
    const STime & delta
    )
    {
    generate
        (
        CreateObject<FloatPacket>( get_function_value( delta ) ),
        STime::current_time()
        );
    }


// Implementation of ColorGenerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ColorGenerator::initialize()
    {
    WaveGenerator::initialize();
    client_entity->type = ClientEntity::TYPE_COLOR_GENERATOR; // To be visible in map.
    }


void ColorGenerator::setup_color_generator
    (
    const STime &   new_tick_period,
    const STime &   new_period,
    bool            enable_looping,
    WaveType        wave,
    const SColor &  color
    )
    {
    base_color = color;
    setup_wave_generator( new_tick_period, new_period, enable_looping, wave );
    }


void ColorGenerator::next_packet
    (
    const STime & delta
    )
    {
    SColor color = base_color.read();
    color.multiply_rgb( ( 1.0f + get_function_value( delta ) ) * 0.5f );
    generate
        (
        CreateObject<ColorPacket>( color.read() ),
        STime::current_time()
        );
    }


// Implementation of ColorMultiplexer
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ColorMultiplexer::initialize()
    {
    Generator::initialize();

    red_id = create_input();
    green_id = create_input();
    blue_id = create_input();
    alpha_id = create_input();
    }


Logic::SocketId ColorMultiplexer::translate_input
    (
    const std::string & input_name
    )
    {
    if( input_name == "red" )
        {
        return red_id;
        }

    if( input_name == "green" )
        {
        return green_id;
        }

    if( input_name == "blue" )
        {
        return blue_id;
        }

    if( input_name == "alpha" )
        {
        return alpha_id;
        }

    return Generator::translate_input( input_name );
    }


void ColorMultiplexer::operate
    (
    SocketId        input_id,
    const STime &   delivery_time
    )
    {
    Pointer<FloatPacket> packet = get_input( input_id ).convert();

    if( input_id == red_id )
        {
        last_color.set_red( packet->get_float() );
        }
    else
    if( input_id == green_id )
        {
        last_color.set_green( packet->get_float() );
        }
    else
    if( input_id == blue_id )
        {
        last_color.set_blue( packet->get_float() );
        }
    else
    if( input_id == alpha_id )
        {
        last_color.set_alpha( packet->get_float() );
        }

    generate( CreateObject<ColorPacket>( last_color.read() ), delivery_time );
    }


void ColorMultiplexer::operate_synced
    (
    const STime & delivery_time
    )
    {
    Pointer<FloatPacket> red = get_input( red_id ).convert();
    Pointer<FloatPacket> green = get_input( red_id ).convert();
    Pointer<FloatPacket> blue = get_input( red_id ).convert();
    Pointer<FloatPacket> alpha = get_input( red_id ).convert();

    last_color = SColor
        (
        red->get_float(),
        green->get_float(),
        blue->get_float(),
        alpha->get_float()
        );

    generate( CreateObject<ColorPacket>( last_color.read() ), delivery_time );
    }


} // namespace Demo::Lib
} // namespace Demo
