/** @file generator.h
 * Generator.
 */

/*
 * begin        : Mon June 13 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_GENERATOR_H
#define DEMO_GENERATOR_H

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_ENTITY_H
#include "lib/server/entity.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();

BASIC_LIB_EXCEPTION( GeneratorErrorException, ::Massiv::Core::Lib::RuntimeException );
BASIC_LIB_EXCEPTION( GeneratorHasDefaultOutputOnlyException, GeneratorErrorException );


/**
 * Generator object.
 * An entity whose logic has one default output ("output").
 * Generator can have (multiple) inputs that can be used to control
 * the generator.
 */
class Generator : public Entity
    {
protected:
    
    MASSIV_OBJECT( Generator );

    //
    // Generator interface.
    //

    /** Initializes object. */
    virtual void initialize();

public:    

    /**
     * Returns default output socket id.
     */
    static SocketId get_output_id();

    /**
     * Returns default output name.
     */
    static std::string get_output_name();

    /**
     * Distribute @c packet to output listeners so that it would be
     * delivered at @c delivery_time.
     */
    void generate
        (
        Pointer<Packet> packet,
        const STime &   delivery_time
        );
    
protected:

    //
    // Logic interface implementation.
    //
    
    /**
     */
    virtual Logic::SocketId translate_output
        (
        const std::string & output_name
        );
    
    /**
     */
    virtual SocketId create_output
        (
        const std::string & dynamic_name = ""
        );

    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

    }; // Generator


/**
 * Timer object that pulses in discrete ticks.
 * Each timer tick triggers the generator to generate its output.
 * Timer has a special boolean input "enable" that causes the timer to start
 * or stop according to the input change.
 */
class Timer : public Generator
    {
protected:

    MASSIV_OBJECT( Timer );

    //
    // Timer interface.
    //

    /**
     * Initializes stopped timer with zero period.
     */
    virtual void initialize();

    /**
     * Initializes the timer.
     * The delay between two timer ticks is determined by @c period.
     * @c auto_start defines whether the timer can be started
     * immediatelly. If false, the timer must be started manually
     * by @c start().
     */
    void initialize
        (
        const STime &   period,
        bool            auto_start
        );

public:

    /**
     * Starts the timer.
     * The timer can be started/stopped multiple times.
     */
    virtual void start();

    /**
     * Stops the timer.
     * @note If called in the middle of a tick, the timer will be stopped
     * at the end of the tick.
     */
    virtual void stop();

    /**
     * Changes timer parameters.
     * It is not neccessary to stop the generator before calling this.
     */
    virtual void setup_timer
        (
        const STime & new_period
        );

public: // FIXME: Kvuli RPC

    /**
     * Called by the system at the beginning of each timer tick.
     * The implementation calls @c next_packet() and schedules next
     * timer tick.
     */
    virtual void timer_ticked();

protected:
        
    /**
     * Called by @c timer_ticked() so that subclass may generate
     * new output. This should be done by calling @c Generator::generate().
     * @c delta defines time delay since the timer was started.
     */
    virtual void next_packet
        (
        const STime & delta
        ) = 0;

protected:

    //
    // Logic interface implementation.
    // Handles Timer inputs ("enable").
    //
    
    /**
     */
    virtual Logic::SocketId translate_input
        (
        const std::string & input_name
        );
    
    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

private:

    /**
     * Schedules next timer tick.
     * When the timer ticks @c timer_ticked() will be called by the system.
     */
    void schedule_next_tick();

protected:

    PTime       tick_period;
        /**< Time between two timer ticks. */
    PTime       start_time;
        /**< When the timer started. */
    PBool       running;
        /**< Is the timer running? */
    PBool       request_stop;
        /**< Was timer requested to stop? */
    PSocketId   enable_id;
        /**< "Enable" input socket id. */
    }; // Timer


/**
 * Timer object that measures how much time has elapsed since the start
 * of the timer period relatively to the length of the period (time fraction).
 * Resulting time fraction (a float value from [ 0, 1 ] range) is sent to
 * generator default output.
 *
 * TimerWithPeriod changes interpretation of the "enable" input.
 * If set to true, timer will be started (unless already runs).
 * If set to false, looping will be disabled which makes the timer to stop
 * at the end of the period.
 */
class TimerWithPeriod : public Timer
    {
public:
        
    MASSIV_OBJECT( TimerWithPeriod );

    //
    // TimerWithPeriod interface.
    //
    
    /**
     * Initializes stopped timer object.
     */
    virtual void initialize();

    /**
     * Initializes timer object.
     * @c tick_period, @c period define timer tick period (delay between
     * two timer ticks) and timer cycle period.
     * At the end of timer cycle the timer either stops (@c enable_looping
     * was false) or starts over.
     */
    void initialize
        (
        const STime &   tick_period,
        const STime &   period,
        bool            enable_looping,
        bool            auto_start
        );
    
    /**
     * Changes timer parameters.
     * It is not neccessary to stop the generator before calling this.
     */
    virtual void setup_timer_with_period
        (
        const STime &   new_tick_period,
        const STime &   new_period,
        bool            enable_looping
        );

protected:

    //
    // Timer interface implementation.
    //
    
    /**
     * Returns current time fraction.
     * @c delta measures time since the start of the timer.
     */
    Float32 get_fraction
        (
        const STime & delta
        );

    /**
     * Outputs time fraction.
     * @c delta measures time since the start of the timer.
     */
    virtual void next_packet
        (
        const STime & delta
        );

protected:

    //
    // Logic interface implementation.
    //

    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

    
protected:

    PTime       timer_period;
        /**< Timer period. */
    PBool       loop;
        /**< Enable loops? */

    }; // TimerWithPeriod


/**
 * Wave generator.
 */
class WaveGenerator : public TimerWithPeriod
    {
public:
        
    MASSIV_OBJECT( WaveGenerator );

    /**
     * Wave type.
     * All waves are defined on [ 0, 1 ] range with function values ranging
     * from -1.0f to 1.0f (single period).
     */
    enum WaveType
        {
        WAVE_SINE =     0,
            /**< Sine wave. */
        WAVE_SQUARE =   1,
            /**< Square wave. */
        WAVE_SAW =      2,
            /**< Saw wave. */
        WAVE_RANDOM =   3
            /**< Random function. */
        };

public:

    //
    // WaveGenerator interface.
    //
    
    /**
     * Initializes a stopped generator object.
     */
    virtual void initialize();
        
    /**
     * Changes wave generator parameters.
     */
    virtual void setup_wave_generator
        (
        const STime &   new_tick_period,
        const STime &   new_period,
        bool            enable_looping,
        WaveType        wave
        );

protected:

    //
    // Timer interface implementation.
    //
    
    /**
     * Returns function value according to the current time fraction.
     * @c delta measures time since the start of the timer.
     */
    Float32 get_function_value
        (
        const STime & delta
        );

    /**
     * Outputs value according to a predefined wave function and current
     * time fraction.
     * @c delta measures time since the start of the timer.
     */
    virtual void next_packet
        (
        const STime & delta
        );

private:

    PEnum<WaveType> wave_type;
        /**< Type of the wave. */

    }; // WaveGenerator


/**
 * Simple monochrome color generator.
 * Outputs the base color scaled by a function value generated by the super
 * class.
 */
class ColorGenerator : public WaveGenerator
    {
public:

    MASSIV_OBJECT( ColorGenerator );

public:

    //
    // ColorGenerator interface.
    //
    
    /**
     * Initializes a stopped generator object.
     */
    virtual void initialize();
        
    /**
     * Changes color generator parameters.
     */
    virtual void setup_color_generator
        (
        const STime &   new_tick_period,
        const STime &   new_period,
        bool            enable_looping,
        WaveType        wave,
        const SColor &  color
        );

protected:

    //
    // Timer interface implementation.
    //
    
    /**
     */
    virtual void next_packet
        (
        const STime & delta
        );

private:

    PColor  base_color;
        /**< Base color. */

    }; // ColorGenerator


/**
 * A generator with float inputs "red", "green", "blue", "alpha".
 * If at least one of the input values change, color composed of the four
 * components is sent to default output.
 */
class ColorMultiplexer : public Generator
    {
public:

    MASSIV_OBJECT( ColorMultiplexer );
    
    /** Initializes object. */
    virtual void initialize();

protected:

    //
    // Logic interface implementation.
    //

    /**
     */
    virtual Logic::SocketId translate_input
        (
        const std::string & input_name
        );
    
    /**
     */
    virtual void operate
        (
        SocketId        input_id,
        const STime &   delivery_time
        );
    
    /**
     */
    virtual void operate_synced
        (
        const STime & delivery_time
        );

private:

    PSocketId   red_id;
        /**< Input socket for RED component. */
    PSocketId   green_id;
        /**< Input socket for GREEN component. */
    PSocketId   blue_id;
        /**< Input socket for BLUE component. */
    PSocketId   alpha_id;
        /**< Input socket for ALPHA component. */
    PColor      last_color;
        /**< Last output. */

    }; // ColorMultiplexer


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_GENERATOR_H
