/** @file item.h
 * Item.
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_ITEM_H
#define DEMO_ITEM_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_GAME_DEFS_H
#include "lib/shared/game_defs.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Item.
 * Item that can be put into containers (like inventory
 * or chest).
 *
 * Items can be used on monsters (to modify its attributes etc.)
 * or equipped.
 * Equipable items should set @a equip_type to value other than
 * @a EQUIP_NONE and redefine @a equip() and @a unequip() method.
 * Items that are not equipable should leave @a equip_type to
 * @a EQUIP_NONE and redefine @a use_on_monster() method.
 * If item is equipable, monster will call @a equip() and @a unequip().
 * If item is not equipable, monster will call @a use_on_monster().
 *
 * TODO: Ma byt potomek entity? Nebo itemy, ktere se budou
 *  povalovat ve svete budou jineho typu, a pri sebrani se promeni
 *  na Item?
 *
 * Moznost:
 * class ItemEntity
 *     {
 *     strong_ptr< Item > item;
 *
 *     Pointer< Item > pick_up()
 *         {
 *         unlink_from_world();
 *         Pointer< Item > i = item;
 *         item = null;
 *         return i;
 *         }
 *     }
 */
class Item : public Object
    {
public:

    MASSIV_OBJECT( Item );

    /** Initialize the object. */
    void initialize();

public:

    /**
     * Create item of given type.
     */
    static Pointer< Item > create
        (
        ItemType   item_type,
        int        item_variant
        );

public:

    /**
     * Use the item on @a monster.
     * @param player Player who used the item. Never null.
     * @param monster Monster to use the item on.
     * @return Item that replaces "this" item. Caller
     *      must forget the pointer that was used
     *      to call the function and use returned item
     *      instead. The returned item replaces the original one.
     *      Warning - returned item may be different type
     *      than the original one.
     *
     * Default action does nothing and returns self.
     *
     * @warning Item may <b>not</b> modify existing items in player's
     *      inventory from this function. It can only put new items in
     *      empty slots of the inventory.
     *      The reason is that @c use_on_monster() is often called when
     *      manipulating the inventory.
     *      May return null.
     */
    virtual Pointer< Item > use_on_monster
        (
        Pointer< Player >       player,
        WeakPointer< Monster >  monster
        );

    /**
     * Use the item on sector.
     * @param player Player who used the item. Never null.
     * @return Item that replaces "this" item. Caller
     *      must forget the pointer that was used
     *      to call the function and use returned item
     *      instead. The returned item replaces the original one.
     *      Warning - returned item may be different type
     *      than the original one.
     *
     * Default action does nothing and returns self.
     *
     * @warning Item may <b>not</b> modify existing items in player's
     *      inventory from this function. It can only put new items in
     *      empty slots of the inventory.
     *      The reason is that @c use_on_monster() is often called when
     *      manipulating the inventory.
     *      May return null.
     */
    virtual Pointer< Item > use_on_sector
        (
        Pointer< Player > player
        );

public:

    /**
     * Equip the item.
     * This is callback from monster's item management and can be
     * redefined by ancestors.
     * Default action returns self.
     * @param monster Monster to equip the item on.
     * @return Item that replaces "this" item. Caller
     *      must forget the pointer that was used
     *      to call the function and use returned item
     *      instead. The returned item replaces the original one.
     *      Returned item may be different type than the original one.
     *      May return null.
     * @pre The item is not equipped.
     * @warning Equipped flag is not modified! Caller must use
     *      @c set_equipped() method to modify the flag.
     */
    virtual Pointer< Item > equip
        (
        Pointer< Monster > monster
        );

    /**
     * Unequip the item from the @a monster.
     * This is callback from monster's item management and can be
     * redefined by ancestors.
     * Default action returns self.
     * @param monster Monster to unequip the item from.
     * @return Item that replaces "this" item. Caller
     *      must forget the pointer that was used
     *      to call the function and use returned item
     *      instead. The returned item replaces the original one.
     *      Returned item may be different type than the original one.
     *      May return null.
     * @pre The item is equipped.
     * @warning Equipped flag is not modified! Caller must use
     *      @c set_equipped() method to modify the flag.
     */
    virtual Pointer< Item > unequip
        (
        Pointer< Monster > monster
        );

    /**
     * Mark the item equipped or unequipped.
     */
    void set_equipped
        (
        bool equipped
        );

    /**
     * Is the item equipped?
     */
    bool is_equipped() const;

    /**
     * Is the item equipable?
     * @return True if @a equip_type is not @c EQUIP_NONE.
     */
    bool is_equipable() const;

public:

    PPointer< ClientItem >  client_item;
        /**< Client visible part of the item. */

    }; // Item


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_ITEM_H
