/** @file layer.h
 * Layer.
 */

/*
 * begin        : Mon Jul 14 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LAYER_H
#define DEMO_LAYER_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_SECTOR_ADDRESS_H
#include "lib/shared/sector_address.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();

BASIC_LIB_EXCEPTION( LayerErrorException, ::Massiv::Core::Lib::RuntimeException );
BASIC_LIB_EXCEPTION( SectorAlreadyDefinedException, LayerErrorException );


/**
 * Layer object.
 * Represents a layer of a virtual world.
 * States basic layer properties: metrics, physical attributes, layer
 * organization, used data structures, etc.
 * The layer is subdivided into a number of box-sized sectors, areas
 * of constant volume. 
 */
class Layer : public Object
    {
protected:

    MASSIV_OBJECT( Layer );

    friend class Sector;
    friend class World;

    typedef PDictionary< SectorAddress, PWeakPointer<Sector> > SectorList;

public:

    //
    // Layer interface.
    //

    /** Creates new Layer. */
    void initialize();
    
    /**
     * Lookups global sector map and returns sector that is identified
     * by @c address. Returns @c null if there is no such sector.
     */
    Remote<Sector> get_sector
        (
        const SectorAddress & address
        );

    /**
     * Registers @c sector to the layer under given @c address.
     * Once the method returns the sector is properly registered and linked
     * with its neighbors.
     * The operation is not atomic. Sector neighbors may see temporal
     * inconsistent state (improper sector linkage)!
     * @note Must not run in parallel with another @c register_sector().
     */
    void register_sector
        (
        Pointer<Sector>         sector,
        const SectorAddress &   address
        );

    /**
     * Creates and registers a new sector to a given [ column, row ]
     * position in the layer.
     */
    void register_new_sector
        (
        int column,
        int row
        );

protected:

    PPointer<World>         world;
        /**< World that owns this layer. */
    PPointer<Enumerator>    layer_list_enumerator;
        /**< Enumerator to @c world->layer_list that references this layer. */
    SectorList              sector_list;
        /**< List of all registered sectors that belong to this layer. */
    PBool                   sector_list_lock;
        /**< True if @c sector_list is locked. */
    }; // Layer


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_LAYER_H
