/** @file lock.h
 * Lock.
 */

/*
 * begin        : Mon Jul 14 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LOCK_H
#define DEMO_LOCK_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();

BASIC_LIB_EXCEPTION( LockErrorException, ::Massiv::Core::Lib::RuntimeException );


/**
 * Simple optimistic lock with standard constructor/destructor
 * (lock/unlock) semantics. If lock can not be granted @c LockErrorException
 * is thrown instead of waiting for the lock.
 */
class OptimisticLock
    {
public:

    /**
     * Locks the lock.
     * @c locked should point to an external lock state variable
     * which is expected to a be property of an object to be locked.
     * If @c message is not @c NULL LockErrorException with @c message
     * description will be thrown if lock can not be locked.
     * If @c message is @c NULL and lock can not be thrown, the
     * lock will assert.
     */
    INLINE OptimisticLock
        (
        PBool &         locked,
        const char *    message
        );

    /**
     * Releases the lock.
     */
    INLINE ~OptimisticLock();

private:

    PBool & locked;
        /**< External lock state variable. */

    }; // OptimisticLock


/**
 * Use this macro to lock an object that owns @c state property.
 * When macro "goes out of its scope", the lock is automatically unlocked.
 */
#define OPTIMISTIC_LOCK( state, message ) \
    OptimisticLock __lock( state, message )
        

} // namespace Demo::Lib
} // namespace Demo

#ifdef ENABLE_INLINE
#include "lib/server/lock_inline.h"
#endif

#endif // DEMO_LOCK_H
