/** @file logic.cpp
 * Logic implementation.
 */

/*
 * begin        : Mon May 19 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/server/logic.h"
#include "lib/server/logic_rpc.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of Logic::SocketListener
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::SocketListener::initialize
    (
    const Remote< Logic >   logic,
    const SocketId          id
    )
    {
    input_logic = logic;
    input_id = id;
    }


// Implementation of Logic::InputSocket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::InputSocket::initialize()
    {
    }


// Implementation of Logic::OutputSocket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::OutputSocket::initialize()
    {
    }


void Logic::OutputSocket::add_listener
    (
    const Remote<Logic> &   input_logic,
    const SocketId &        input_id
    )
    {
    listeners.push_back
        ( ObjectOnStack<SocketListener>( input_logic, input_id ) );
    }


void Logic::OutputSocket::remove_listener
    (
    const Remote<Logic> &   input_logic,
    const SocketId &        input_id
    )
    {
    std::size_t i;

    for( i = 0; i < listeners.size(); i++ )
        {
        SocketListener & listener = listeners[ i ];
        
        if( listener.input_logic == input_logic && listener.input_id == input_id )
            {
            listeners.erase( i );
            }
        }
    }


void Logic::OutputSocket::remove_listeners
    (
    const Remote<Logic> &   input_logic
    )
    {
    std::size_t i = 0;

    while( i < listeners.size() )
        {
        if( listeners[ i ].input_logic == input_logic )
            {
            listeners.erase( i );
            }
        else
            {
            i++;
            }
        }
    }


// Implementation of Logic::Packet
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::Packet::initialize()
    {
    }


void Logic::Packet::delivered_to
    (
    WeakPointer<Object> object,
    const STime &       time
    )
    {
    Pointer<Logic> logic = object.convert();
    logic->input( input_id, this, time );
    }


void Logic::Packet::delivery_failed
    (
    DeliveryFailure failure,
    const STime &   deliver_time
    )
    {
    if( failure == Object::DESTINATION_NOT_ACCESSIBLE )
        {
        output_logic->async_unplug_outputs( input_logic );
        }
    }


// Implementation of Logic::FloatPacket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::FloatPacket::initialize
    (
    Float32 the_value
    )
    {
    value = the_value;
    }


void Logic::FloatPacket::initialize
    (
    bool the_value
    )
    {
    if( the_value == true )
        {
        value = 1.0f;
        }
    else
        {
        value = 0.0f;
        }
    }


Float32 Logic::FloatPacket::get_float()
    {
    return value;
    }


bool Logic::FloatPacket::get_boolean()
    {
    return value >= 0.5f;
    }


// Implementation of Logic::ColorPacket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::ColorPacket::initialize
    (
    const SColor & the_value
    )
    {
    value = the_value;
    }


SColor Logic::ColorPacket::get_color()
    {
    return value.read();
    }


// Implementation of Logic::ObjectPacket
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Logic::ObjectPacket::initialize
    (
    const WeakPointer<Object> & the_value
    )
    {
    value = the_value;
    }


WeakPointer<Object> Logic::ObjectPacket::get_object()
    {
    return value;
    }


// Implementation of Logic
// ~~~~~~~~~~~~~~~~~~~~~~~

void Logic::initialize()
    {
    const int num_buckets = 4;

    input_names.clear( num_buckets );
    output_names.clear( num_buckets );
    }


Logic::SocketId Logic::translate_input
    (
    const std::string & input_name
    )
    {
    const PSocketId temp = input_names[ input_name ];
    return temp;
    }


Logic::SocketId Logic::translate_output
    (
    const std::string & output_name
    )
    {
    const PSocketId temp = output_names[ output_name ];
    return temp;
    }


Logic::SocketId Logic::create_input
    (
    const std::string & dynamic_name
    )
    {
    SocketId id = inputs.size();
    
    if( dynamic_name.size() )
        {
        input_names[ dynamic_name ] = id;
        }
        
    inputs.push_back( ObjectOnStack<InputSocket>() );
    
    return id;
    }


Logic::SocketId Logic::create_output
    (
    const std::string & dynamic_name
    )
    {
    SocketId id = outputs.size();

    if( dynamic_name.size() )
        {
        output_names[ dynamic_name ] = id;
        }
    
    outputs.push_back( ObjectOnStack<OutputSocket>() );
    
    return id;
    }


Pointer<Logic::Packet> Logic::get_input
    (
    const SocketId & input_id
    )
    {
    return inputs[ input_id ].received;
    }


Pointer<Logic::Packet> Logic::get_output
    (
    const SocketId & output_id
    )
    {
    return outputs[ output_id ].sent;
    }


void Logic::plug_output
    (
    const std::string &     output,
    const Remote<Logic> &   input_logic,
    const std::string &     input
    )
    {
    outputs[ translate_output( output ) ].add_listener
        ( input_logic, input_logic->sync_translate_input( input ) );
    }


void Logic::plug_input
    (
    const std::string &     input,
    const Remote<Logic> &   output_logic,
    const std::string &     output
    )
    {
    output_logic->sync_plug_output( output, this, input );
    }


void Logic::unplug_output
    (
    const std::string &     output,
    const Remote<Logic> &   input_logic,
    const std::string &     input
    )
    {
    outputs[ translate_output( output ) ].remove_listener
        ( input_logic, input_logic->sync_translate_input( input ) );
    }


void Logic::unplug_input
    (
    const std::string &     input,
    const Remote<Logic> &   output_logic,
    const std::string &     output
    )
    {
    output_logic->sync_unplug_output( output, this, input );
    }


void Logic::unplug_outputs
    (
    const Remote<Logic> &   input_logic
    )
    {
    std::size_t i;

    for( i = 0; i < outputs.size(); i++ )
        {
        OutputSocket & output = outputs[ i ];
        output.remove_listeners( input_logic );
        }
    }


void Logic::output
    (
    const SocketId &        output_id,
    const Pointer<Packet> & packet,
    const STime &           delivery_time
    )
    {
    OutputSocket &  socket = outputs[ output_id ];
    std::size_t     i;
    
    // Initialize the packet.

    packet->output_logic = this;
    packet->output_id = output_id;

    socket.sent = packet;
    socket.send_time = STime::current_time();

    // Distribute the packet.

    for( i = 0; i < socket.listeners.size(); i++ )
        {
        SocketListener &    listener = socket.listeners[ i ];
        Pointer<Packet>     send_this;

        send_this = ( packet->clone() ).convert();
        send_this->input_logic = listener.input_logic;
        send_this->input_id = listener.input_id;
        
        send_this->migrate_to( listener.input_logic, delivery_time );
        }
    }
    

void Logic::input
    (
    const SocketId &        input_id,
    const Pointer<Packet> & packet,
    const STime &           delivery_time
    )
    {
    std::size_t i;

    // Update input socket.

    inputs[ input_id ].received = packet;
    inputs[ input_id ].receive_time = delivery_time;
    
    // Test synchronousness.

    for( i = 0; i < inputs.size(); i++ )
        {
        InputSocket &   input = inputs[ i ];
        const Float64   threshold = 0.005f;
        STime           delay = input.receive_time.read() - delivery_time;
        
        if( FloatTraits<Float64>::abs( delay.get_value() ) > threshold ||
            input.receive_time.is_equal_to( STime() ) )
            {
            operate( input_id, delivery_time );
            return;
            }

        hope( input.received != null );
        }

    operate_synced( delivery_time );
    }


void Logic::operate
    (
    SocketId        input_id,
    const STime &   delivery_time
    )
    {
    operate_synced( delivery_time );
    }


} // namespace Demo::Lib
} // namespace Demo
