/** @file logic.h
 * Logic.
 */

/*
 * begin        : Mon May 19 2003
 * copyright    : (C) 2003 by Martin Havlista, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LOGIC_H
#define DEMO_LOGIC_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Logic is an object that defines set of "inputs" and "outputs".
 * Outputs can be linked with inputs of another @c Logic objects.
 * When an output changes all @c Logic objects whose inputs "listen" to this
 * output are notified and updated (the @c Logic output is propagated to
 * the inputs).
 * The inputs are implemented by InputSockets, outputs by OutputSockets.
 */
class Logic : public Object
    {
public:
        
    MASSIV_OBJECT( Logic );

    class Socket;
    class SocketListener;
    class InputSocket;
    class OutputSocket;
    class Packet;

    friend class Packet;
                
    typedef PVlInt32                PSocketId;
    typedef Int32                   SocketId;
    typedef PArray<InputSocket>     InputSockets;
    typedef PArray<OutputSocket>    OutputSockets;
    typedef PArray<SocketListener>  SocketListeners;
    
    typedef PDictionary
                <
                PString,
                PSocketId,
                ext::hash<std::string>,
                std::equal_to<std::string>
                >  NameToPSocketId;

    /**
     * Representation of Logic input or output.
     */
    class Socket : public Object
        {
        MASSIV_OBJECT( Socket );
        }; // Socket

    /**
     * Each Logic output can be plugged into multiple inputs.
     * SocketListener represents an established link between
     * an OutputSocket (that owns SocketListener) and an InputSocket.
     */
    class SocketListener : public Object
        {
    public:
        
        MASSIV_OBJECT( SocketListener );
        friend class Logic;
        friend class OutputSocket;

        /** Initializes object. */
        void initialize
            (
            const Remote< Logic >   input,
            const SocketId          id
            );
        
    protected:

        PRemote<Logic>  input_logic;
            /**< Logic that listens to the output. */
        PSocketId       input_id;
            /**< Input socket the output is routed to. */
        }; // SocketListener

    /**
     * Logic input.
     */
    class InputSocket : public Socket
        {
    public:
        
        MASSIV_OBJECT( InputSocket );
        friend class Logic;

        /** Initializes object. */
        void initialize();

    protected:

        PPointer<Packet>    received;
            /**< Last received packet. */
        PTime               receive_time;
            /**< Time to what @c received was scheduled to be delivered. */
        }; // InputSocket

    /**
     * Logic output.
     */
    class OutputSocket : public Socket
        {
    public:

        MASSIV_OBJECT( OutputSocket );
        friend class Logic;

        /** Initializes object. */
        void initialize();

        /** Adds socket listener. */
        void add_listener
            (
            const Remote<Logic> &   input_logic,
            const SocketId &        input_id
            );
        
        /** Remove socket listener. */
        void remove_listener
            (
            const Remote<Logic> &   input_logic,
            const SocketId &        input_id
            );
        
        /** Remove all @c input_logic listeners. */
        void remove_listeners
            (
            const Remote<Logic> &   input_logic
            );

    protected:

        SocketListeners     listeners;
            /**< Who listens to this output. */
        PPointer<Packet>    sent;
            /**< Last sent packet. */
        PTime               send_time;
            /**< When @c sent was sent. */
        }; // Output socket

    /**
     * Packet is a base class for client-defined packets. Each packet
     * represents a "value" that is sent from an OutputSocket
     * to an InputSocket. The packets are created by Logic subclasses
     * which are responsible for packet distribution whenever their
     * outputs change.
     */
    class Packet : public Object
        {
    public:

        MASSIV_OBJECT( Packet );
        friend class Logic;

        /** Initializes object.*/
        void initialize();
    
    protected:

        /**
         */
        virtual void delivered_to
            (
            WeakPointer<Object> object,
            const STime &       time
            );

        /**
         */
        virtual void delivery_failed
            (
            DeliveryFailure failure,
            const STime &   deliver_time
            );

    protected:
            
        // Filled by sender (@c Logic::output()).

        PRemote<Logic>  output_logic;
            /**< Packet sender. */
        PSocketId       output_id;
            /**< Packet sender. */
        PRemote<Logic>  input_logic;
            /**< Packet consumer. */
        PSocketId       input_id;
            /**< Packet consumer. */
        
        }; // Packet

    /**
     * Packet that holds a floating point value.
     * The packet is also used to transmit boolean values according to
     * this mapping:
     * - @c value >= 0.5f -\> true
     * - @c value < 0.5f -\> false
     * .
     * This is to allow automatic conversions float \<-\> boolean where needed.
     * @see BooleanPacket which is an alias to @c FloatPacket.
     */
    class FloatPacket : public Packet
        {
    public:
            
        MASSIV_OBJECT( FloatPacket );

        /** Initializes packet. */
        void initialize
            (
            Float32 the_value
            );
        
        /** Initializes packet. */
        void initialize
            (
            bool the_value
            );

        /** Returns stored float. */
        Float32 get_float();

        /** Converts stored float to boolean. */
        bool get_boolean();

    protected:

        PFloat32 value;
            /**< Stored value. */

        }; // FloatPacket
    
    typedef FloatPacket BooleanPacket;

    /**
     * Packet that holds a color value.
     */
    class ColorPacket : public Packet
        {
    public:
            
        MASSIV_OBJECT( ColorPacket );

        /** Initializes packet. */
        void initialize
            (
            const SColor & the_value
            );
        
        /** Returns stored color. */
        SColor get_color();

    protected:

        PColor value;
            /**< Stored color. */

        }; // ColorPacket
    
    /**
     * Packet that holds a pointer to an object.
     */
    class ObjectPacket : public Packet
        {
    public:
            
        MASSIV_OBJECT( ObjectPacket );

        /** Initializes packet. */
        void initialize
            (
            const WeakPointer<Object> & the_value
            );
        
        /** Returns stored pointer. */
        WeakPointer<Object> get_object();

    protected:

        PWeakPointer<Object> value;
            /**< Stored pointer. */

        }; // ObjectPacket

protected:

    /**
     * Initializes object.
     */
    void initialize();

//protected:
public: // FIXME: Kvuli RPC
    
    //
    // Socket management.
    //

    /**
     * Translates input socket name to input socket id.
     * The general implementation looks up the optional @c input_names
     * map. It can be overriden so that the resolution is static.
     */
    virtual Logic::SocketId translate_input
        (
        const std::string & input_name
        );
    
    /**
     * Translates output socket name to output socket id.
     * The general implementation looks up the optional @c output_names
     * map. It can be overriden so that the resolution is static.
     */
    virtual Logic::SocketId translate_output
        (
        const std::string & output_name
        );

    /**
     * Creates an input socket.
     * Returned SocketId is simply an index to @c inputs, the new
     * socket is created at the end of the array.
     * If @c dynamic_name is non-empty the socket is registered
     * under that name into @c input_names.
     */
    virtual SocketId create_input
        (
        const std::string & dynamic_name = ""
        );
    
    /**
     * Creates an output socket.
     * Returned SocketId is simply an index to @c outputs, the new
     * socket is created at the end of the array.
     * If @c dynamic_name is non-empty the socket is registered
     * under that name into @c output_names.
     */
    virtual SocketId create_output
        (
        const std::string & dynamic_name = ""
        );

    /**
     * Returns packet that was received by input @c input_id for the last
     * time.
     */
    virtual Pointer<Packet> get_input
        (
        const SocketId &    input_id
        );
    
    /**
     * Returns packet that was sent by output @c output_id for the last time.
     */
    virtual Pointer<Packet> get_output
        (
        const SocketId &    output_id
        );

public:

    //
    // Plug protocol.
    // Uses SRPC to initiate the connection.
    //

    /**
     * Plugs @c output to [ @c input_logic, @c input ].
     * The connection is fully estabilished when method returns.
     */
    virtual void plug_output
        (
        const std::string &     output,
        const Remote<Logic> &   input_logic,
        const std::string &     input
        );

    /**
     * Plugs @c input to [ @c output_logic, @c output ].
     * The connection is fully estabilished when method returns.
     */
    virtual void plug_input
        (
        const std::string &     input,
        const Remote<Logic> &   output_logic,
        const std::string &     output
        );

    /**
     * Unplugs @c output from [ @c input_logic, @c input ].
     * The connection is fully dropped when method returns.
     */
    virtual void unplug_output
        (
        const std::string &     output,
        const Remote<Logic> &   input_logic,
        const std::string &     input
        );

    /**
     * Unplugs @c input from [ @c output_logic, @c output ].
     * The connection is fully dropped when method returns.
     */
    virtual void unplug_input
        (
        const std::string &     input,
        const Remote<Logic> &   output_logic,
        const std::string &     output
        );
    
    /**
     * Unplugs all @c outputs routed to @c input_logic.
     */
    virtual void unplug_outputs
        (
        const Remote<Logic> &   input_logic
        );

//protected:
public: // FIMXE: Kvuli RPC

    //
    // Packet dispatching.
    //

    /** Sends @c packet from output @c output_id to all listeners. */
    virtual void output
        (
        const SocketId &        output_id,
        const Pointer<Packet> & packet,
        const STime &           delivery_time
        );

    /**
     * Called by @c Packet when @c packet is delivered to @c Logic.
     * Depending on update time of another inputs either @c operate()
     * or @c operate_synced() is called.
     */
    virtual void input
        (
        const SocketId &        input_id,
        const Pointer<Packet> & packet,
        const STime &           delivery_time
        );

    /**
     * Called by @c input to trigger Logic operation.
     * @c input_id and @c time determine what input changed and when.
     * If there are more inputs each input triggers the operation
     * independently on each other.
     * This is known as "asynchronous" update.
     */
    virtual void operate
        (
        SocketId        input_id,
        const STime &   delivery_time
        );

    /**
     * Called by @c input to trigger Logic operation.
     * The operation is triggered only if all the inputs have changed in
     * approximatelly the same time slot.
     * This is known as "synchronous" update.
     * If there is only one input there is no difference from @c operate().
     */
    virtual void operate_synced
        (
        const STime &   delivery_time
        ) = 0;

protected:

    NameToPSocketId     input_names;
        /**< Optional map that translates input socket names
             to input socket ids. */
    NameToPSocketId     output_names;
        /**< Optional map that translates output socket names
             to output socket ids. */
    InputSockets        inputs;
        /**< Input sockets. */
    OutputSockets       outputs;
        /**< Output sockets. */
    }; // Logic


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_LOGIC_H
