/** @file monster.cpp
 * Monster (implementation).
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/monster.h"
#include "lib/server/monster_rpc.h"
#include "lib/server/sector_helper_objects.h"
#include "lib/server/brain.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/random.h"


/**
 * Enable this to log collisions.
 */
//#define LOG_COLLISIONS


namespace Demo {
namespace Lib {


void Monster::initialize()
    {
    Super::initialize();
    entity_flags = MONSTER;
    client_entity->type = ClientEntity::TYPE_MONSTER;
    set_chat_color( SColor( 0.8f, 0.8f, 0.8f ) );
    set_name( "monster" );
    }


void Monster::set_chat_color
    (
    const SColor & color
    )
    {
    chat_color.set_red( color.red() );
    chat_color.set_green( color.green() );
    chat_color.set_blue( color.blue() );
    chat_color.set_alpha( color.alpha() );
    }


void Monster::say
    (
    const std::string &     text,
    MessageType             type
    )
    {
    Pointer< Sector > sector = current_sector;

    // FIXME: Use get_last_sector()? Or is sector == null delay
    // say() a bit?
    if( sector != null )
        {
        Float32 radius = 16.0f;
        SColor  color( chat_color.red(), chat_color.green(), chat_color.blue(), chat_color.alpha() );

        std::string message;
        if( type == MESSAGE_ACTION )
            {
            message += "* ";
            }
        message += get_name();
        if( type == MESSAGE_TALK )
            {
            message += ":";
            }
        message += " ";
        message += text;

        // Send the text to all nearby clients/players.
        Pointer< MessageEventTask > task =
            CreateObject< MessageEventTask >
                (
                sector,
                client_entity,
                get_current_position().origin.read(),
                radius,
                message,
                color
                );
        }
    }


void Monster::check_visible()
    {
    }


Remote<Brain> Monster::get_brain() const
    {
    return brain;
    }


Monster::AttackResult Monster::attack_by
    (
    PPair< PEnum< ItemType >, PVlInt32 > & acquired_item,
    const EntityPosition &                 attacker_position,
    WeaponType                             weapon_type
    )
    {
    if( client_entity->get_ability_value( ABILITY_HEALTH ) <= 0 )
        {
        return ATTACK_ALREADY_DEAD;
        }

    float weapon_range = 0.0f;
    int weapon_damage = 0;
    if( weapon_type == WEAPON_AXE )
        {
        weapon_range = 4.0f;
        weapon_damage = 10;
        }

    const SVector2 my_position = client_entity->current_position.origin.read();
    const SVector2 attacker_to_self = my_position - attacker_position.origin.read();
    if( attacker_to_self.magnitude() < weapon_range )
        {
        Remote< Monster > self( this );
        apply_powerup( ABILITY_HEALTH, -weapon_damage, 0, POWERUP_PERMANENT, STime() );
        if( client_entity->get_ability_value( ABILITY_HEALTH ) <= 0 )
            {
            self->param( RPC_DELAYED, 0.2f ).async_say( "aaargh!", MESSAGE_TALK );
            ItemType it = ITEM_UNKNOWN;
            Int32 iv = 0;
            monster_died( it, iv );
            acquired_item.first = it;
            acquired_item.second = iv;
            return ATTACK_KILLED;
            }
        else
            {
            self->param( RPC_DELAYED, 0.2f ).async_say( "ouch!", MESSAGE_TALK );
            return ATTACK_HIT;
            }
        }

    if( attacker_to_self.magnitude() < 2.0f * weapon_range )
        {
        return ATTACK_TOO_FAR;
        }
    return ATTACK_FAILED;
    }


void Monster::monster_died
    (
    ItemType & item_type,
    Int32 &    item_variant
    )
    {
    item_type = ITEM_UNKNOWN;
    item_variant = 0;
    }


void Monster::put_into_layer
    (
    const EntityPosition & position
    )
    {
    deploying = true;
    deploy_position = position;

    Super::put_into_layer( position );
    }


void Monster::entity_moved()
    {
    if( deploying )
        {
        std::ostringstream ss;

        // Monster was successfully deployed into the layer.

        ss  << get_object_id() << " was successfully deployed: "
            << get_current_position().to_string();

        Global::log_debug
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_LOW,
            ss.str()
            );

        deploying = false;
        }

    Super::entity_moved();
    }


void Monster::move_failed
    (
    const CollisionInfo & collision_info
    )
    {
    if( deploying )
        {
        std::ostringstream ss;

        // Retry to deploy the monster to a random nearby position.

        EntityPosition position = deploy_position;
        position.origin += Random::random_vector() * 30.0f;

        ss  << get_object_id() << " could not be deployed. "
            << "Retrying to deploy to a random nearby position: "
            << position.to_string();

        Global::log_warning
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_HIGHEST,
            ss.str()
            );

        Super::put_into_layer( position );
        }
    else
        {
        #ifdef LOG_COLLISIONS

        std::ostringstream ss;

        ss  << get_object_id() << " collided. Details: "
            << collision_info.to_string();

        Global::log_info
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_LOW,
            ss.str()
            );

        #endif

        Super::move_failed( collision_info );

        if( brain != null )
            {
            brain->report_collision( collision_info );
            }
        }
    }


} // namespace Demo::Lib
} // namespace Demo
