/** @file monster.h
 * Monster.
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys, Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_MONSTER_H
#define DEMO_MONSTER_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_PHYSICAL_H
#include "lib/server/physical.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Monster, a Physical with AI controlled via a Brain object.
 * Monster routes external stimuli to associated Brain object which in turn
 * calls Monster to control itself.
 */
class Monster : public Physical
    {
    typedef Physical Super; // Leave it private.

public:

    /**
     * Parameter of say().
     */
    enum MessageType
        {
        MESSAGE_TALK,           ///< Entity said something.
        MESSAGE_ACTION          ///< Entity performed an action.
        }; // enum TalkType

    /**
     * Result of attack_by().
     */
    enum AttackResult
        {
        ATTACK_HIT,             ///< Successful attack.
        ATTACK_KILLED,          ///< Successful attack, target killed.
        ATTACK_TOO_FAR,         ///< Missed because the target is too far away.
        ATTACK_ALREADY_DEAD,    ///< Missed because the target is already dead.
        ATTACK_FAILED,          ///< Generic attack failure.
        }; // enum AttackResult

public:

    MASSIV_OBJECT( Monster );

    /** Initialize the object. */
    virtual void initialize();

public:

    /** Set @a chat_color of the monster. */
    void set_chat_color
        (
        const SColor & color
        );

    /**
     * Say @a text.
     * Monster uses this function to talk.
     * The @a text is sent to all nearby players.
     */
    void say
        (
        const std::string & text,
        MessageType         type
        );

    /**
     * Check for entities visible by the monster and report
     * them to Brain.
     * May operate asynchronously.
     * The default implementation does nothing.
     */
    virtual void check_visible();

    /**
     * Return monster's brain.
     */
    Remote<Brain> get_brain() const;

    /**
     * Someone attacked the monster from @a attacker_position using weapon of type @a weapon_type.
     *
     * Default implementation does nothing if monster is dead, hurts monster if it's reachable,
     * and calls monster_dies() if the attack killed the monster.
     *
     * @return Result of the attack. Ff the attack succeeded (@c ATTACK_HIT), @a acquired_item
     *         may contain a new item (no item if ItemType is ITEM_UNKNOWN).
     *
     * @bug Should be handled by the brain object.
     */
    virtual AttackResult attack_by
        (
        PPair< PEnum< ItemType >, PVlInt32 > & acquired_item,
        const EntityPosition &                 attacker_position,
        WeaponType                             weapon_type
        );

    /**
     * A monster died. Return earned item.
     *
     * Default implementation sets type to @c ITEM_UNKNOWN.
     */
    virtual void monster_died
        (
        ItemType & item_type,
        Int32 &    item_variant
        );

public:

    /** Deploys player to the layer. */
    virtual void put_into_layer
        (
        const EntityPosition & position
        );

protected:

    /** Entity moved call-back. */
    virtual void entity_moved();

    /** Entity collided call-back. */
    virtual void move_failed
        (
        const CollisionInfo & collision_info
        );

public:

    PPointer<Brain> brain;
        /**< Brain that is responsible for monster's AI. */
    PColor          chat_color;
        /**< Color used by @c say() method. */
    PBool           deploying;
        /**< Has monster not been deployed to the layer yet? */
    EntityPosition  deploy_position;
        /**< Position the monster should be deployed to. If monster can not
             be deployed to that position, a random neighbor position will
             be chosen. */

    }; // Monster


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_MONSTER_H
