/** @file naming_service.h
 * Naming service.
 */

/*
 * begin        : Sun Oct 12 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_NAMING_SERVICE_H
#define DEMO_NAMING_SERVICE_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Naming service.
 *
 * These objects are registered when initial world is created:
 * <ul>
 * <li> "World" - World object. Object of type #World.
 * <li> "DefaultLayer" - Default world layer. Object of type #Layer.
 * </ul>
 */
class NamingService : public Object
    {
public:

    MASSIV_OBJECT( NamingService );

    typedef PDictionary
        <
        PString,
        PWeakPointer< Object >,
        ext::hash<std::string>,
        std::equal_to<std::string>
        >  NameToObject;

public:

    /** Creates new NamingService. */
    void initialize();

    /**
     * Get pointer to object by name the object is registered in the naming service.
     * Returns null pointer if there is no object registered under specified name.
     *
     * @note Not const because PDictionary has no constant iterators.
     */
    WeakPointer< Object > get_object
        (
        const std::string & name
        );

    /**
     * Register object to the naming service.
     *
     * @param name    Name under which the object should be registered.
     * @param object  Object to register.
     * @param replace If another object of given @a name is already registered, this
     *                argument determines behavior of the method. If @a replace is
     *                @c true, @a object will be registered under the @a name and
     *                old registration will be forgotten (unregistered object won't
     *                be notified of the change). If @a replace is @c false, the
     *                call will fail.
     * @return @c true if the object has successfully been registered, @c false otherwise.
     *
     * @note If the @a object is already registered under the @a name, @c true will
     *       be returned.
     */
    bool register_object
        (
        const std::string &     name,
        WeakPointer< Object >   object,
        bool                    replace
        );

    /**
     * Unregister object from the naming service.
     *
     * If object registered under the @a name is @a object, it will be unregistered
     * and the method will return @c true. Otherwise no change to registration
     * will be made and @c false will be returned.
     */
    bool unregister_object
        (
        const std::string &     name,
        WeakPointer< Object >   object
        );

private:

    NameToObject    name_to_object;
        /**< Map that translates object name to object pointer. */

    }; // class NamingService


} // namespace Demo::Lib
} // namespace Demo


#endif // DEMO_NAMING_SERVICE_H
