/** @file physical.cpp
 * Physical object implementation.
 */

/*
 * begin        : Mon Aug 18 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/physical.h"
#include "lib/server/physical_rpc.h"
#include "lib/shared/client_sector.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/random.h"
#include "lib/server/sector.h"
#include "lib/server/sector_rpc.h"
#include "lib/server/lock.h"
#include "lib/server/collision_info.h"

#include "utility/heap.h"
#include "utility/heap_index.h"
#include "utility/heap_impl.h"

#include <limits>

/** @def USE_HEURISTICS
 * Define this to enable A-star heuristics.
 */
#define USE_HEURISTICS

/** @def LOG_RESERVATION_MAP
 * Define this to enable reservation map logging.
 */
//#define LOG_RESERVATION_MAP


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


namespace {

//
// Physics tick scheduler settings.
//

const Float32 min_tick_period = 0.1f;
const Float32 max_tick_period = 1.0f;

const Float32 upper_distance_limit = 1.0f / 2.0f; // m/physics tick
const Float32 lower_distance_limit = 1.0f / 4.0f; // m/physics tick

const Float32 lazy_tick_factor = 1.5f;
const Float32 speedy_tick_factor = 0.4f;

//
// Physical's logic settings.
//

const Float32 logic_period = 0.25f;
const Float32 last_way_point_accuracy = 1.0f / 3.0f;
const Float32 way_point_accuracy = 1.0f / 2.0f;
const Float32 acceleration_smoothness = 1.0f / 8.0f;
const Float32 braking_smoothness = 2.0f;
const Float32 steering_smoothness = 0.3f;
const Float32 distance_smoothness = 3.0f;
const Float32 steering_distance = 0.25f;
const Float32 inner_way_point_speed_coef = 0.3f;
const Float32 min_steering_angle = MATHF_PI / 4.0f;
const Float32 max_steering_angle = MATHF_PI / 2.0f;
const Float32 bad_heading_coef = 8.0f;
const Float32 slope_speed_coef = 1.0f;

//
// Delta map deformation settings.
//

const Float32 mass_radius_coef = 2.0f / 100.0f;
const Float32 mass_height_coef = 0.2f / 100.0f;
const Float32 modify_terrain_probability = 0.02f;

//
// Helper methods.
//

/** Returns heading of @c vector. */
Float32 get_heading
    (
    const SVector2 & vector
    )
    {
    return FloatTraits<Float32>::atan2( vector.y(), vector.x() ) - MATHF_PI_2;
    }

/** Returns unit vector with the given orientation. */
SVector2 get_direction
    (
    Float32 heading
    )
    {
    return SVector2
        (
        -FloatTraits<Float32>::sin( heading ),
        FloatTraits<Float32>::cos( heading )
        );
    }

/** Returns angle distance between @c angle1 and @c angle2. */
Float32 angle_distance
    (
    Float32 angle1,
    Float32 angle2
    )
    {
    Float32 difference = FloatTraits< Float32 >::clamp_periodic
        (
        angle1 - angle2,
        0.0f,
        2.0f * MATHF_PI
        );
    if( difference < MATHF_PI )
        {
        return difference;
        }
    return 2.0f * MATHF_PI - difference;
    }

//
// Helper data structures for @c Physical::find_path().
//

const Int32     max_distance = std::numeric_limits<Int32>::max();
const Int32     max_search_distance = ClientSector::NUM_X_TILES * 4;
const Float32   bounding_box_epsilon = 0.5f;
const Float32   slope_epsilon_coef = 0.9f;
const Float32   min_way_point_distance = 0.5f;

/**
 * Helper class that computes shortest path to a given tile from
 * the originator tile (while accounting entity's physical properties).
 * The implementation uses simple Dijkstra algorithm.
 */
class FindPath
    {
public:

    friend class Demo::Lib::Physical;

    enum Parameters
        {
        FIXED_POINT_CONSTANT = 256
            /**< Path distances are represented by fixed-point integers with
                 log2( FIXED_POINT_CONSTANT ) bits used for the fraction part. */
        };

    /**
     * Fixed-point distance. The actual floating-point distance can be
     * computed by dividing @c Distance by @C FIXED_POINT_CONSTANT.
     */
    typedef Int32 Distance;

    /**
     * Tile coordinates.
     *
     * x-y tile coordinate maps directly to x-y sector-relative coordinates, ie.
     * tile 0,0 is actually south-western corner of the sector, not center
     * of south-western tile.
     */
    struct Tile
        {
    public:

        /** HashKey functor for @c Tile. */
        struct HashKey
            {
            std::size_t operator()
                (
                const Tile & tile
                ) const
                {
                return tile.x * 1235678 + tile.y * 87654321;
                }
            };

        /** Constructor. */
        Tile
            (
            int x = 0,
            int y = 0
            ) : x( x ), y( y )
            {
            }

        /** Returns a neighboring tile. */
        Tile get_neighbor
            (
            Direction direction
            ) const
            {
            switch( direction )
                {
                case EAST:
                    return Tile( x + 1, y );
                case NORTH_EAST:
                    return Tile( x + 1, y + 1 );
                case NORTH:
                    return Tile( x, y + 1 );
                case NORTH_WEST:
                    return Tile( x - 1, y + 1 );
                case WEST:
                    return Tile( x - 1, y );
                case SOUTH_WEST:
                    return Tile( x - 1, y - 1 );
                case SOUTH:
                    return Tile( x, y - 1 );
                case SOUTH_EAST:
                    return Tile( x + 1, y - 1 );
                }

            panic( "Invalid direction." );
            }

        /** Tests if the two tiles are the same. */
        bool operator==
            (
            const Tile & other_tile
            ) const
            {
            return x == other_tile.x && y == other_tile.y;
            }

        /** Tests if the two tiles are not the same. */
        bool operator!=
            (
            const Tile & other_tile
            ) const
            {
            return !operator==( other_tile );
            }

        /**
         * Returns tile "difference", a lower-bound estimation of the shortest path
         * from this tile to @c tile. Tile difference to target tile can be used as
         * a heuristics that penalizes not promising paths.
         */
        Distance estimate_shortest_path_to
            (
            const Tile & tile
            ) const
            {
            int dx = tile.x - x;
            int dy = tile.y - y;

            // In order to make the heuristics work it must be a lower bound
            // estimation of the shortest path from this tile to @c tile.
            // Suppose dy >= dx >= 0. Then if there were no obstacles
            // the shortest path would consist of @c dx diagonal steps and @c dy - dx
            // orthogonal steps, which yields @c dx * sqrt( 2.0f ) + ( dy - dx ) length.
            // Note that we can use simple Euclidean distance as a heuristics as it is
            // a lower bound estimation of the shortest path to @c tile (it can be easily
            // show that @c dx * sqrt( 2.0f ) + ( dy - dx ) >= sqrt( dx * dx + dy * dy )).

            return static_cast<Int32>( FloatTraits<Float32>::sqrt( static_cast<Float32>(
                dx * dx + dy * dy ) ) * FIXED_POINT_CONSTANT );
            }

    public:

        int x;
            /**< X tile coordinate. In @c sector's sector tile coordinates. */
        int y;
            /**< Y tile coordinate. In @c sector's sector tile coordinates. */
        }; // Tile

    /** Information about a tile. */
    struct TileInfo
        {
    public:

        /** Constructor. */
        TileInfo() :
            distance( max_distance ),
            from_direction( NORTH ),
            unreachable( false )
            {
            }

    public:

        Distance    distance;
            /**< Tile distance. Length of the shortest path to this tile
                 found so far. */
        Direction   from_direction;
            /**< Direction from neighbor on the shortest path. */
        HeapIndex   index;
            /**< Index into @c ActiveTiles heap. */
        bool        unreachable;
            /**< Is the tile unreachable (ie. can not be reserved)? */

        }; // TileInfo

    /** Tile map. */
    typedef ext::hash_map
        <
        Tile,
        TileInfo,
        Tile::HashKey
        > TileMap;

    /**
     * Heap index changed functor.
     * Used to keep proper correspondence between @c active_tile heap
     * and @c tiles items.
     */
    struct HeapIndexChanged
        {
        void operator()
            (
            const TileMap::iterator &   it,
            HeapIndex                   new_index
            ) const
            {
            TileInfo & info = it->second;

            info.index = new_index;
            }
        }; // HeapIndexChanged

    /** Heap of active tiles, sorted by tile distances. */
    typedef Heap
        <
        Distance,
        TileMap::iterator,
        std::less<Distance>,
        HeapIndexChanged
        > ActiveTiles;

    typedef Tile LightWeightAddress;

    /** Map that translates (light weight) sector addresses to sectors. */
    typedef ext::hash_map
        <
        LightWeightAddress,
        WeakPointer<Sector>,
        Tile::HashKey
        > SectorMap;

    typedef Physical::WayPoint      WayPoint;
    typedef std::vector<WayPoint>   WayPointPath;

public:

    /**
     * Constructor.
     * [ @c entity, @c physical_info ] describe physical properties of
     * the entity that wants to move from @c from to @c to position
     * (in layer coordinates).
     * @c sector should point to the sector @c from belongs to.
     */
    FindPath
        (
        Pointer<Sector>         sector,
        Pointer<Physical>       entity,
        const PhysicalInfo &    info,
        const WayPoint &        from,
        const WayPoint &        to
        ) : from( from ),
            to( to ),
            sector( sector ),
            entity( entity ),
            physical_info( info )
        {
        sector_address = sector->get_address();
        origin = way_point_to_tile( from );
        target = way_point_to_tile( to );

        // Tweak entity's physical properties:
        // enlarge entity's bounding box and apply "stricter" physical
        // restrictions.
        // This is to avoid problems when entity does not follow the
        // path precisely. Tested box will be larger than the box tested
        // by sector physics (tile reservation) and the entity will not
        // be able to climb all slopes.

        physical_info.bounding_box_size +=
            SVector2( bounding_box_epsilon, bounding_box_epsilon );

        physical_info.max_slope *= slope_epsilon_coef;

        // Draw tiles that can and can not be reserved.

        #ifdef LOG_RESERVATION_MAP

        std::ostringstream  ss;
        int                 y, x;

        ss << "origin = [ " << origin.x << ", " << origin.y << "]\n";
        ss << "target = [ " << target.x << ", " << target.y << "]\n";

        for( y = max_search_distance + origin.y; y >= -max_search_distance + origin.y; y-- )
            {
            for( x = -max_search_distance + origin.x; x <= max_search_distance + origin.x; x++ )
                {
                if( x == origin.x && y == origin.y )
                    {
                    ss << '*';
                    }
                else
                if( can_reserve( x, y ) )
                    {
                    ss << ' ';
                    }
                else
                    {
                    ss << 'X';
                    }
                }
            ss << '\n';
            }

        Global::log_debug
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            ss.str()
            );

        #endif
        }

    /** Runs the algorithm. */
    void go()
        {
        bool path_found = false;

        // Initialize.

        #ifndef USE_HEURISTICS

        change_distance( origin, 0, NORTH );

        #else

        change_distance( origin, origin.estimate_shortest_path_to( target ), NORTH );

        #endif

        // Compute distances to reachable "nearby" tiles.
        // Assume that entity can move to @c origin tile.

        while( !active_tiles.is_empty() )
            {
            int i;

            // Extract the closest tile.

            TileMap::iterator   it = extract_closest_tile();
            const Tile &        closest_tile = it->first;
            const TileInfo &    tile_info = it->second;

            if( tile_info.distance > max_search_distance * FIXED_POINT_CONSTANT )
                {
                // Cancel this path because we are too far from the
                // origin tile.

                continue;
                }

            if( closest_tile == target )
                {
                // We have found some path to the @c target.
                // Stop the algorithm.

                path_found = true;
                break;
                }

            // Update neighboring tiles.

            for( i = 0; i < 8; i++ )
                {
                Direction   direction = static_cast<Direction>( i );
                Tile        neighbor = closest_tile.get_neighbor( direction );
                Distance    neighbor_distance = tile_info.distance +
                    get_direction_distance( direction );

                if( !can_move_to( closest_tile, direction ) )
                    {
                    continue;
                    }

                #ifdef USE_HEURISTICS

                // We can safely penalize current length of the shortest path to
                // @c neighbor by tile difference from @c neighbor to @c target,
                // because even if the shortest path from @c origin to @c target
                // went through @c neighbor its total distance would be
                // at least as big as the penalized distance to @c neighbor.
                // Compute the penalty incrementally (penalty for @c closest_tile
                // has already been accounted to @c neighbor_distance), we want
                // @c neighbor_distance to be equal to the "actual neighbor distance
                // plus penalty to target".

                neighbor_distance += neighbor.estimate_shortest_path_to( target ) -
                    closest_tile.estimate_shortest_path_to( target );

                #endif

                if( neighbor_distance < get_distance( neighbor ) )
                    {
                    // We have found a closer path to @c neighbor.

                    change_distance( neighbor, neighbor_distance, direction );
                    }
                }
            }

        if( !path_found )
            {
            return;
            }

        // Construct (backtrack) shortest path from tile distances.
        // The path will lead from @c from to @c to along the discrete
        // tile grid.

            {
            Tile            tile = target;
            Direction       previous_direction = NORTH;
            Direction       direction = SOUTH;
            WayPointPath    backward_path;
            int             i;

            // Backtrack path.

            backward_path.push_back( to );

            while( tile != origin )
                {
                // Find tile we came from.

                TileMap::const_iterator it = tiles.find( tile );
                hope( it != tiles.end() );

                previous_direction = direction;
                direction = Sector::opposite_direction( it->second.from_direction );

                // Add tile to path if it is a "corner tile".

                if( previous_direction != direction )
                    {
                    backward_path.push_back( tile_to_way_point( tile ) );
                    }

                // Go to the that tile.

                tile = tile.get_neighbor( direction );
                }

            backward_path.push_back( from );

            // Reverse the path and drop those way points that are too close
            // to each other.

            for( i = backward_path.size() - 1; i >= 0; i-- )
                {
                const SVector2 & way_point = backward_path[ i ];

                if( path.size() && ( path.back() - way_point ).square_magnitude() <
                    min_way_point_distance * min_way_point_distance )
                    {
                    continue;
                    }

                path.push_back( way_point );
                }
            }
        }

    /** Returns distance to a neighboring tile in the particular direction. */
    Distance get_direction_distance
        (
        Direction direction
        )
        {
        // Return a fixed-point distance.

        switch( direction )
            {
            case EAST:
            case NORTH:
            case WEST:
            case SOUTH:
                return static_cast<Distance>( 1.0f * FIXED_POINT_CONSTANT );

            case NORTH_EAST:
            case NORTH_WEST:
            case SOUTH_WEST:
            case SOUTH_EAST:
                return static_cast<Distance>( MATHF_SQRT2 * FIXED_POINT_CONSTANT );
            }

        panic( "Invalid direction." );
        }

    /** Extracts the closest tile from the @c active_tiles. */
    TileMap::iterator extract_closest_tile()
        {
        TileMap::iterator it = active_tiles[ 0 ];

        hope( !it->second.unreachable );
        active_tiles.pop();

        return it;
        }

    /**
     * Changes distance of @a tile to @a distance, which is distance to
     * origin to @a tile when coming via tile in direction opposite to
     * @a from_direction,
     * The @a tile must already be reachable or if it has not been processed
     * yet it is made reachable.
     */
    void change_distance
        (
        const Tile &    tile,
        Distance        distance,
        Direction       from_direction
        )
        {
        TileMap::iterator it = tiles.find( tile );

        if( it == tiles.end() )
            {
            TileInfo info;

            // Register tile into @c tiles.

            info.distance = distance;
            info.from_direction = from_direction;
            info.unreachable = false;

            it = tiles.insert( TileMap::value_type( tile, info ) ).first;

            // Register tile into @c active_tiles heap.

            active_tiles.push( info.distance, it );
            hope( it->second.index.is_valid() );
            }
        else
            {
            TileInfo & info = it->second;

            // Tile should still be present in the @c active_tiles heap.

            if( !info.index.is_valid() )
                {
                std::ostringstream ss;

                ss  << "Tile [ " << tile.x << ", " << tile.y << " ] is closed, "
                    << "but we have found a shorter path to it. Tile distance = "
                    << info.distance << ". New distance = " << distance
                    << ". Hrmmh.";

                Global::log_warning
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_HIGHEST,
                    ss.str()
                    );

                return;
                }

            hope( info.index.is_valid() );
            hope( !info.unreachable );
            hope( info.distance >= distance );

            // Update.

            active_tiles.change_key( info.index, distance );
            info.distance = distance;
            info.from_direction = from_direction;
            }
        }

    /** Returns distance of @c tile. */
    Distance get_distance
        (
        const Tile & tile
        )
        {
        TileMap::iterator it = tiles.find( tile );

        if( it == tiles.end() )
            {
            return max_distance;
            }
        else
            {
            const TileInfo & info = it->second;

            hope( !info.unreachable || info.distance == max_distance );

            return info.distance;
            }
        }

    /**
     * Tests if we can move (teleport) the entity from @a tile in
     * @a direction. It is assumed that entity can be teleported to
     * @a tile. Entity bounding box is not rotated.
     */
    bool can_move_to
        (
        const Tile &    tile,
        Direction       direction
        )
        {
        Int32   x = tile.x;
        Int32   y = tile.y;
        Int32   i;

        // Get entity's enlarged bounding box.

        Int32 x_half_size = FloatTraits<Float32>::ceil_int
            ( 0.5f * physical_info.bounding_box_size.x() );
        Int32 y_half_size = FloatTraits<Float32>::ceil_int
            ( 0.5f * physical_info.bounding_box_size.y() );

        // Bounding box centered to the @ tile is assumed to be checked already.
        // Now, bounding box centered to @a tile + @a direction should be checked.
        // It is sufficient to check the difference of the two boxes only
        // (which reduces to a column and/or a row check).

        // Note: x and y are vertex coordinates, can_reserve() takes tile coordinates.

        switch( direction )
            {
            case EAST:
                {
                // Check if [ x + x_half_size ] column can be reserved.

                for( i = -y_half_size; i < y_half_size; i++ )
                    {
                    if( !can_reserve( x + x_half_size, y + i ) )
                        {
                        return false;
                        }
                    }

                return true;
                }

            case NORTH_EAST:
                {
                return can_move_to( tile, NORTH ) &&
                    can_move_to( tile, EAST ) &&
                    can_reserve( x + x_half_size, y + y_half_size );
                }

            case NORTH:
                {
                // Check if [ y + y_half_size ] row can be reserved.

                for( i = -x_half_size; i < x_half_size; i++ )
                    {
                    if( !can_reserve( x + i, y + y_half_size ) )
                        {
                        return false;
                        }
                    }

                return true;
                }

            case NORTH_WEST:
                {
                return can_move_to( tile, NORTH ) &&
                    can_move_to( tile, WEST ) &&
                    can_reserve( x - x_half_size - 1, y + y_half_size );
                }

            case WEST:
                {
                // Check if [ x - x_half_size - 1 ] column can be reserved.

                for( i = -y_half_size; i < y_half_size; i++ )
                    {
                    if( !can_reserve( x - x_half_size - 1, y + i ) )
                        {
                        return false;
                        }
                    }

                return true;
                }

            case SOUTH_WEST:
                {
                return can_move_to( tile, SOUTH ) &&
                    can_move_to( tile, WEST ) &&
                    can_reserve( x - x_half_size - 1, y - y_half_size - 1 );
                }

            case SOUTH:
                {
                // Check if [ y - y_half_size - 1 ] row can be reserved.

                for( i = -x_half_size; i < x_half_size; i++ )
                    {
                    if( !can_reserve( x + i, y - y_half_size - 1 ) )
                        {
                        return false;
                        }
                    }

                return true;
                }

            case SOUTH_EAST:
                {
                return can_move_to( tile, SOUTH ) &&
                    can_move_to( tile, EAST ) &&
                    can_reserve( x + x_half_size, y - y_half_size - 1 );
                }
            }

        return true;
        }

    /**
     * Tests if [ x_tile, y_tile ] (in @c sector's tile coordinates) can
     * be reserved.
     */
    bool can_reserve
        (
        int x_tile,
        int y_tile
        )
        {
        Tile                    tile( x_tile, y_tile );
        TileMap::const_iterator it = tiles.find( tile );

        if( it != tiles.end() )
            {
            // The tile was present in @c tiles.
            // Use the cached result.

            const TileInfo & info = it->second;

            return !info.unreachable;
            }
        else
            {
            // The tile was not present in @c tiles.

            int sector_x_tile = x_tile;
            int sector_y_tile = y_tile;
            WeakPointer<Sector> the_sector = get_sector( sector_x_tile, sector_y_tile );

            if( the_sector.is_local() && the_sector->is_consistent() )
                {
                CollisionInfo collision_info;

                bool result = the_sector->can_reserve_tile
                    (
                    sector_x_tile,
                    sector_y_tile,
                    physical_info,
                    entity,
                    collision_info
                    );

                if( result )
                    {
                    // The tile is reachable.
                    // Update both the @c tiles and @c active_tiles.
                    // This also automatically caches the result.

                    change_distance( tile, max_distance, NORTH );

                    return true;
                    }
                }

            // The tile is unreachable.
            // Cache the result.

            TileInfo info;

            info.distance = max_distance;
            info.unreachable = true;

            tiles.insert( TileMap::value_type( tile, info ) );

            return false;
            }
        }

    /**
     * Returns sector [ x_tile, y_tile ] belongs to.
     * At the time of call [ x_tile, y_tile ] are expressed in @c sector's
     * tile coordinates, the method converts them to the returned sector's
     * tile coordinates.
     * Initiates sector replication to the current node.
     *
     * @warning May return a pointer to an inconsistent sector replica.
     */
    WeakPointer<Sector> get_sector
        (
        int & x_tile,
        int & y_tile
        )
        {
        // Compute address of a sector [ x_tile, y_tile ] belongs to.
        // Translate [ x_tile, y_tile ] to that sector's tile coordinates.

        LightWeightAddress address
            (
            FloatTraits<Float32>::floor_int
                ( x_tile / static_cast<Float32>( ClientSector::NUM_X_TILES ) ),
            FloatTraits<Float32>::floor_int
                ( y_tile / static_cast<Float32>( ClientSector::NUM_Y_TILES ) )
            );

        x_tile -= address.x * ClientSector::NUM_X_TILES;
        y_tile -= address.y * ClientSector::NUM_Y_TILES;

        address.x += sector_address.column;
        address.y += sector_address.row;

        // Return pointer to that sector.

        SectorMap::iterator it = sectors.find( address );

        if( it != sectors.end() )
            {
            // Return cached value.

            return it->second;
            }
        else
            {
            // Resolve sector address.

            Remote<Sector> the_sector = sector->get_sector
                (
                SectorAddress
                    (
                    sector_address.layer,
                    address.x,
                    address.y
                    )
                );
            WeakPointer<Sector> weak_sector = the_sector;

            // Update cache.

            sectors.insert( SectorMap::value_type( address, weak_sector ) );

            // Request sector replication.

            if( the_sector != null )
                {
                the_sector->async_replicate_to
                    (
                    Massiv::System::get_local_node_id(),
                    STime::current_time() + 20.0
                    );
                }

            return weak_sector;
            }
        }

    /** Converts @c tile, in @c sector tile coordinates, to a way-point. */
    WayPoint tile_to_way_point
        (
        const Tile & tile
        )
        {
        WayPoint    way_point;
        SVector2    tile_xy
            (
            static_cast<Float32>( tile.x ),
            static_cast<Float32>( tile.y )
            );

        sector->tile_xy_to_layer_xy( tile_xy, way_point );

        return way_point;
        }

    /** Converts way-point to tile, in @c sector tile coordinates. */
    Tile way_point_to_tile
        (
        const WayPoint & way_point
        )
        {
        SVector2 tile_xy;

        sector->layer_xy_to_tile_xy( way_point, tile_xy );

        return Tile
            (
            FloatTraits<Float32>::to_integer( tile_xy.x() ),
            FloatTraits<Float32>::to_integer( tile_xy.y() )
            );
        }

private:

    Tile                origin;
        /**< The origin tile. The tile we are searching the path from.
             In @c sector tile coordinates. */
    Tile                target;
        /**< The target tile. In @c sector tile coordinates. */
    WayPoint            from;
        /**< The original position. In layer coordinates. */
    WayPoint            to;
        /**< The target position. In layer coordinates. */
    Pointer<Sector>     sector;
        /**< Sector @c origin belongs to. */
    SectorAddress       sector_address;
        /**< Address of @c sector. */
    Pointer<Physical>   entity;
        /**< Entity. */
    PhysicalInfo        physical_info;
        /**< Physical info of @c entity. */
    TileMap             tiles;
        /** Information about processed tiles. */
    ActiveTiles         active_tiles;
        /**< Active tiles. Reachable tiles only. Always subset of @c tiles. */
    SectorMap           sectors;
        /**< Sector address to pointer to sector map cache. */
    WayPointPath        path;
        /**< The constructed path. */
    }; // FindPath

}


void Physical::initialize()
    {
    Remote<Physical> self = this;

    // Initialize.

    Super::initialize();
    client_entity->type = ClientEntity::TYPE_PHYSICAL;
    physics_period = max_tick_period;
    path_finding_disabled = false;

    // Spawn physics and logic loops.

    self->async_physics_tick();
    self->async_logic_tick();
    }


void Physical::add_way_point
    (
    const WayPoint & way_point
    )
    {
    std::ostringstream ss;

    // Add way-point.

    do_not_move = false;
    path.push_back( way_point );

    // Log.

    #ifdef OPTION_DEBUG

    ss  << get_object_id() << " scheduled " << way_point.to_string()
        << " as its way-point number " << path.size() << ".";

    Global::log_debug
        (
        Status::FACILITY_LIB,
        Status::PRIORITY_NORMAL,
        ss.str()
        );

    #endif
    }


void Physical::clear_path()
    {
    path.clear();
    }


bool Physical::find_path
    (
    const WayPoint & way_point
    )
    {
    if( path_finding_disabled )
        {
        add_way_point( way_point );
        return true;
        }

    if( current_sector == null )
        {
        #ifdef OPTION_DEBUG

        std::ostringstream ss;

        ss  << get_object_id() << " found no path to " << way_point.to_string()
            << " because entity is currently unlinked from the map.";

        Global::log_debug
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            ss.str()
            );

        #endif

        return false;
        }

    FindPath request
        (
        current_sector,
        this,
        physical_info,
        get_current_position().origin.read(),
        way_point
        );

    request.go();

    if( !request.path.size() )
        {
        #ifdef OPTION_DEBUG

        std::ostringstream ss;

        ss << get_object_id() << " found no path to " << way_point.to_string();

        Global::log_debug
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            ss.str()
            );

        #endif

        return false;
        }
    else
        {
        std::size_t i;

        for( i = 0; i < request.path.size(); i++ )
            {
            add_way_point( request.path[ i ] );
            }

        return true;
        }
    }


void Physical::look_at
    (
    const WayPoint & way_point
    )
    {
    clear_path();
    add_way_point( way_point );
    do_not_move = true;
    }


bool Physical::is_path_empty() const
    {
    return path.empty();
    }


void Physical::teleport_to
    (
    const EntityPosition & position
    )
    {
    if( teleporting )
        {
        return;
        }

    do_physics();

    // Turn on "teleport mode" so that physics would not interfere with
    // the move.

    teleporting = true;
    move_entity( position );
    }


bool Physical::toggle_path_finding()
    {
    if( path_finding_disabled )
        {
        path_finding_disabled = false;
        return true;
        }
    else
        {
        path_finding_disabled = true;
        return false;
        }
    }


void Physical::do_physics()
    {
    // Lock method.

    if( inside_do_physics )
        {
        // Avoid recursion.

        return;
        }

    OPTIMISTIC_LOCK( inside_do_physics, NULL );

    // Initialize.

    STime   current_time = STime::current_time();
    Float32 delta = static_cast<Float32>( ( current_time - old_time ).get_value() );

    if( old_time.is_equal_to( STime() ) )
        {
        // @c do_physics() was called for the first time.

        old_time = current_time;
        return;
        }

    if( delta == 0.0f )
        {
        return;
        }

    if( current_sector == null )
        {
        return;
        }

    if( teleporting )
        {
        // Do not touch entity position if it is teleporting on behalf
        // of @c teleport_to() request.

        return;
        }

    // Gather information on entity's current position.

    EntityPosition      position = get_current_position();
    SVector2            current_tile;
    Float32             current_slope;

    current_sector->layer_xy_to_tile_xy
        ( position.origin.read(), current_tile );

    current_slope = current_sector->get_tile_slope
        (
        FloatTraits<Float32>::to_integer( current_tile.x() ),
        FloatTraits<Float32>::to_integer( current_tile.y() )
        );

    // Account for external powers.

    Float32 friction_coef = 0.0f;
    Float32 slope_coef = current_slope * slope_speed_coef;
    Float32 speed_factor = FloatTraits<Float32>::exp( -friction_coef * delta ) *
        FloatTraits<Float32>::exp( -slope_coef * delta );

    // Modify terrain.

    if( !FloatTraits<Float32>::epsilon_zero( physical_info.mass *
        mass_height_coef ) && Random::random() < modify_terrain_probability )
        {
        current_sector->modify_elevation
            (
            position.origin.read(),
            ClientSector::KIND_DELTA_ELEVATION_MAP,
            ClientSector::FUNCTION_FUZZY_CIRCLE,
            ClientSector::OPERATION_ADD,
            physical_info.mass * mass_radius_coef,
            -physical_info.mass * mass_height_coef * delta
            );
        }

    // Update entity's position and its physical properties.
    // We are assuming that physical properties have been constant
    // for [ @c old_time, @c current_time ] period.

    position.origin += current_speed * delta * speed_factor;
    position.orientation.set_heading( current_heading );

    // Adapt physics tick period so that slowly moving entities would not
    // be processed too often and fast entities would not pass through
    // obstacles.

    Float32 distance_squared = current_speed.square_magnitude() * delta * delta;

    if( distance_squared > upper_distance_limit * upper_distance_limit )
        {
        physics_period *= speedy_tick_factor;

        if( physics_period < min_tick_period )
            {
            physics_period = min_tick_period;
            }
        }
    else
    if( distance_squared < lower_distance_limit * lower_distance_limit )
        {
        physics_period *= lazy_tick_factor;

        if( physics_period > max_tick_period )
            {
            physics_period = max_tick_period;
            }
        }

    if( FloatTraits<Float32>::epsilon_zero( distance_squared ) )
        {
        current_speed.set_zero();
        }

    // "Teleport" the entity to the new position.
    // There is no check whether the entity can really move that way
    // (if there is an obstacle).
    // The move fails if @c position is already occupied by another
    // colliding entity.
    // Warning! From now on the entity can be unlinked from the sector.

    if( position.epsilon_equals( get_current_position() ) )
        {
        if( !client_entity->stopped )
            {
            client_entity->stopped = true;
            }
        }
    else
        {
        move_entity( position );
        if( client_entity->stopped )
            {
            client_entity->stopped = false;
            }
        }

    // Done.

    old_time = current_time;
    }


void Physical::physics_tick()
    {
    Remote<Physical> self = this;

    // Do physics.

    do_physics();

    // Schedule next tick.

    self->param( RPC_DELAYED, physics_period.read() ).async_physics_tick();
    }


void Physical::logic_tick()
    {
    EntityPosition  my_position = get_current_position();
    SVector2        origin = my_position.origin.read();

    // Set requested properties.

    Float32 requested_speed = 0.0f;
    Float32 requested_heading = current_heading;

    // Control speed with the aim to reach the first way-point.

    while( path.size() && !do_not_move.read() )
        {
        // Check the first way-point.
        // Move to that way-point if we are not there already.

        WayPoint        point = path[ 0 ];
        SVector2        direction = point - origin;
        Float32         distance_squared = direction.square_magnitude();

        if( ( path.size() == 1 && distance_squared <
              last_way_point_accuracy * last_way_point_accuracy ) ||
            ( path.size() > 1 && distance_squared <
              way_point_accuracy * way_point_accuracy ) )
            {
            // We are already there.

            path.erase( 0 );

            // Log.

            #ifdef OPTION_DEBUG

            std::ostringstream ss;

            ss  << get_object_id() << " reached way-point at " << point.to_string()
                << ". " << path.size() << " way-point(s) left.";

            Global::log_debug
                (
                Status::FACILITY_LIB,
                Status::PRIORITY_NORMAL,
                ss.str()
                );

            #endif

            continue;
            }

        // Set requested speed according to Physical's abilities.

        Float32 max_speed = 1.0f + 0.1f * client_entity->get_ability_value( ABILITY_SPEED );
        requested_speed = max_speed;

        // Braking.

        Float32 angle_penalty = angle_distance( current_heading,
            get_heading( direction ) ) / MATHF_PI;

        if( path.size() > 1 )
            {
            requested_speed *= 1.0f - FloatTraits<Float32>::exp
                ( -distance_squared / distance_smoothness ) *
                inner_way_point_speed_coef;
            }
        else
            {
            requested_speed *= 1.0f - FloatTraits<Float32>::exp
                ( -distance_squared / distance_smoothness );
            }

        requested_speed *=
            FloatTraits<Float32>::exp( -bad_heading_coef * angle_penalty );

        // Smooth speed changes.

        if( current_speed.magnitude() < requested_speed )
            {
            Float32 speed = current_speed.magnitude() + acceleration_smoothness;

            if( speed < requested_speed )
                {
                requested_speed = speed;
                }
            }
        else
        if( current_speed.magnitude() > requested_speed )
            {
            Float32 speed = current_speed.magnitude() - braking_smoothness;

            if( speed > requested_speed )
                {
                requested_speed = speed;
                }
            }

        break;
        }

    // Control heading with the aim to look at the first way-point.

    if( path.size() )
        {
        Float32             steering_speed =
            1.0f + 0.1f * client_entity->get_ability_value( ABILITY_SPEED );

        const SVector2 &    look_at = path[ 0 ];
        SVector2            direction = look_at - origin;
        ConvexArea::Edge    axis( origin, origin + get_direction( current_heading ) );
        Float32             distance = axis.distance( look_at );
        Float32             absolute_distance = FloatTraits<Float32>::abs( distance );
        Float32             angle = steering_smoothness * absolute_distance * steering_speed;

        // Clamp @c angle by limits.

        if( angle < min_steering_angle )
            {
            angle = min_steering_angle;
            }

        if( angle > max_steering_angle )
            {
            angle = max_steering_angle;
            }

        // Change heading by +- @c angle.

        if( distance > steering_distance )
            {
            // @c look_at is on our left side.
            // Turn left or set exact direction if we are too "close".

            requested_heading = current_heading + angle;
            axis.init( origin, origin + get_direction( requested_heading ) );

            if( axis.distance_is_negative( look_at ) )
                {
                requested_heading = get_heading( direction );
                }
            }
        else
        if( distance < -steering_distance )
            {
            // @c look_at is on our right side.
            // Turn right or set exact direction if we are too "close".

            requested_heading = current_heading - angle;
            axis.init( origin, origin + get_direction( requested_heading ) );

            if( axis.distance_is_positive( look_at ) )
                {
                requested_heading = get_heading( direction );
                }
            }
        else
            {
            // Our axis leads to @c look_at.
            // Turn left as much as possible if we are not heading
            // towards @c look_at (need to do 180 degree "U turn").

            axis.init
                (
                origin,
                origin + get_direction( current_heading + MATHF_PI_2 )
                );

            if( axis.distance_is_positive( look_at ) )
                {
                requested_heading += max_steering_angle;
                }
            }
        }

    // Change physical properties according to what is "requested".

    current_heading = requested_heading;
    current_speed = get_direction( current_heading ) * requested_speed;

    do_physics();

    // Schedule next tick.
    // Unlike physics tick, logic tick must be scheduled regularly
    // (each @c logic_period seconds). That is because of Physical control,
    // most of the parameters have been tuned to the particular tick period.

    Remote<Physical> self = this;
    self->param( RPC_DELAYED, logic_period ).async_logic_tick();
    }


void Physical::entity_moved()
    {
    if( teleporting )
        {
        // Entity has teleported to the requested position successfully.
        // Cleanup the state.

        teleporting = false;

        current_speed.set_zero();
        clear_path();
        }
    }


void Physical::move_failed
    (
    const CollisionInfo & collision_info
    )
    {
    collision_time = STime::current_time();

    if( teleporting )
        {
        // Entity could not be teleported.

        teleporting = false;
        }
    }


} // namespace Demo::Lib
} // namespace Demo
