/** @file physical.h
 * Physical.
 */

/*
 * begin        : Mon Aug 18 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_PHYSICAL_H
#define DEMO_PHYSICAL_H


#include "lib/shared/lib.h"
#include "lib/server/entity.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Physical object.
 * An entity that is able to move and is influenced by the physics.
 * There is no general physics engine, the physics is implemented in
 * Physical object's logic instead.
 */
class Physical : public Entity
    {
    typedef Entity  Super; // Leave it private.

public:

    MASSIV_OBJECT( Physical );

    typedef SVector2                WayPoint;
    typedef PValueArray<WayPoint>   Path;

protected:

    /**
     * Initializes object.
     */
    virtual void initialize();

public:

    /**
     * Adds way-point to the current path.
     */
    void add_way_point
        (
        const WayPoint & way_point
        );

    /**
     * Clears current path.
     * The entity will cancel the move.
     */
    void clear_path();

    /**
     * Finds shortest path to @c way_point, decomposes it to straight
     * components and adds appropriate way-points to the current path.
     * Returns true if path to @c way_point was found.
     *
     * If path finding is disabled @c direct path will be scheduled
     * and possible obstacles not taken into account.
     */
    bool find_path
        (
        const WayPoint & way_point
        );

    /**
     * Makes the entity look at given way-point.
     * The entity will not move until next call to @c add_way_point()
     * or @c find_path().
     * Clears current path.
     */
    void look_at
        (
        const WayPoint & way_point
        );

    /**
     * Tests if current path is empty.
     */
    bool is_path_empty() const;

    /**
     * Teleports entity to @c position.
     * If entity can not be moved to that position, the entity state
     * is left unchanged.
     * If previous teleport has not finished yet, the new @c position
     * is ignored.
     */
    void teleport_to
        (
        const EntityPosition & position
        );
    
    /**
     * Enables/disables path finding for this physical.
     * Returns true if path finding was enabled.
     */
    bool toggle_path_finding();

//protected:
public: // FIXME: Kvuli RPC

    /**
     * Handles the physics.
     * To get absolutely correct results @c do_physics() must be called
     * before any change of the dynamic physical properties.
     * The method must be called quite often so that the entity would
     * not pass through obstacles.
     */
    void do_physics();

    /**
     * Handles physics tick, invokes @c do_physics().
     * Once done, next call to @c physics_tick() is scheduled automatically.
     */
    virtual void physics_tick();

    /**
     * Changes requested properties according to physical entity's logic.
     * Must be called periodically, the tick period must be fixed.
     * Schedules next @c do_logic().
     */
    virtual void logic_tick();

    /**
     * Invoked by Entity when the last move has finished successfully.
     */
    virtual void entity_moved();

    /**
     * Called by Entity when entity can not be teleported to the target
     * position (ie. because of collision).
     */
    virtual void move_failed
        (
        const CollisionInfo & collision_info
        );

protected:
    
    // State.

    PBool       inside_do_physics;
        /**< Are we inside @c do_physics()? */
    PBool       teleporting;
        /**< Was @c teleport_to() requested? */
        
    // Current dynamic properties.

    PVector2    current_speed;
        /**< Current speed vector. In meters per second. */
    PFloat32    current_heading;
        /**< Current heading. */
    PTime       old_time;
        /**< When the physics was computed for the last time. */
    PFloat32    physics_period;
        /**< Current period of physics tick. */
    PTime       collision_time;
        /**< Time of the last collision. */
    
    // Path.

    PBool       path_finding_disabled;
        /**< Is path finding disabled for this physical? */
             
    Path        path;
        /**< List of scheduled way-points the entity should go through.
             The entity moves from way-point to way-point along a straight
             line. */
    PBool       do_not_move;
        /**< If set to true the entity will not move and @c path[ 0 ] will
             be interpreted as a way-point the entity should look at. */

    }; // Physical


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_PHYSICAL_H
