/** @file physical_info.h
 * Physical info.
 */

/*
 * begin        : Mon Jul 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_PHYSICAL_INFO_H
#define DEMO_PHYSICAL_INFO_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Static physical properties of an entity.
 */
class PhysicalInfo : public ValueType
    {
public:

    /**
     * PhysicalClass.
     *
     * All entities are penetrable by default.
     */
    enum PhysicalClass
        {
        CLASS_PENETRABLE = 0,
            /**< Penetrable entity. */
        CLASS_SOLID = 1,
            /**< Solid entity. */
        CLASS_FLYING = 2,
            /**< Flying entity. */
        }; // PhysicalClass

    /** Default constructor. */
    PhysicalInfo();

    /** Copy constructor. */
    PhysicalInfo
        (
        const PhysicalInfo & info
        );

    /**
     * Tests if entity with this PhysicalInfo properties can collide
     * with another entity with @c info properties.
     */
    bool collides_with
        (
        const PhysicalInfo & info
        ) const;

    /**
     * Test if entity with this PhysicalInfo properties can collide
     * with another entity of @c other_class physical class.
     */
    bool collides_with
        (
        PhysicalClass other_class
        ) const;

    /**
     * Tests if entity with this PhysicalInfo properties can handle
     * specified tile @c slope.
     */
    bool can_handle_slope
        (
        Float32 slope
        ) const;

public:

    PEnum<PhysicalClass>    physical_class;
        /**< Entity physical class. */
    PVector2                bounding_box_size;
        /**< Entity bounding box size in sector tile xy coordinates.
             @warning Must be at least 1 x 1 tiles. Otherwise entities
             whose position is not aligned with the tile grid will not
             reserve any tile. See ConvexArea for information on what
             tiles are reported during area rasterization. */
    PBool                   bounding_box_axis_aligned;
        /**< True if bounding box of this entity should not be rotated
             according to entity's heading. */
    PFloat32                max_slope;
        /**< Maximum slope the entity is able to climb. Any slope is
             expressed as a floating point value from [ 0, 1 ] range. */
    PFloat32                mass;
        /**< Physical's mass in kg. Used for deformation of the delta
             elevation map when physical moves. */

    }; // PhysicalInfo


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_PHYSICAL_INFO_H
