/** @file pipe.cpp
 * Pipe implementation.
 */

/*
 * begin        : Sat June 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/pipe.h"
#include "lib/shared/client_entity.h"
#include <algorithm>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


namespace {
    
/**
 * Less predicate for XXXInterpolator::KeyValue.
 */
template <typename KeyValue>
struct KeyValueLess
    {
    bool operator()
        (
        const KeyValue &    value1,
        const KeyValue &    value2
        ) const
        {
        return value1.input < value2.input;
        }
    
    bool operator()
        (
        const KeyValue &                value,
        typename KeyValue::InputType    input
        ) const
        {
        return value.input < input;
        }
    
    bool operator()
        (
        typename KeyValue::InputType    input,
        const KeyValue &                value
        ) const
        {
        return input < value.input;
        }
    };


/**
 * Returns function value for @c input of a picewise linear function
 * defined by @c key_values.
 */
template <typename KeyValue>
typename KeyValue::OutputType interpolate
    (
    typename KeyValue::InputType    input,
    const PArray<KeyValue> &        key_values
    )
    {
    typedef typename KeyValue::OutputType       OutputType;
    OutputType                                  output;
    typename PArray<KeyValue>::ConstIterator    it;
    KeyValueLess<KeyValue>                      less;

    it = std::lower_bound( key_values.begin(), key_values.end(), input, less );

    if( it == key_values.end() )
        {
        if( key_values.size() )
            {
            output = key_values[ key_values.size() - 1 ].requested_output;
            }
        else
            {
            output = OutputType();
            }
        }
    else
    if( it == key_values.begin() )
        {
        if( key_values.size() )
            {
            output = key_values[ 0 ].requested_output;
            }
        else
            {
            output = OutputType();
            }
        }
    else
        {
        const KeyValue &                first = *( it - 1 );
        const KeyValue &                second = *it;
        typename KeyValue::InputType    weight;

        hope( input >= first.input && input <= second.input );

        weight = ( input - first.input ) / ( second.input - first.input );
        output = ( 1.0f - weight ) * first.requested_output +
                 weight * second.requested_output;
        }

    return output;
    }

/**
 * Adds key value to the @c key_values.
 */
template <typename KeyValue>
void add_value
    (
    PArray<KeyValue> &              key_values,
    typename KeyValue::InputType    input,
    typename KeyValue::OutputType   requested_output
    )
    {
    if( key_values.size() )
        {
        const KeyValue & back = key_values[ key_values.size() - 1 ];

        if( back.input > input )
            {
            throw InvalidKeyValueException();
            }
        }

    key_values.push_back( ObjectOnStack<KeyValue>( input, requested_output ) );
    }

}


// Implementation of Pipe
// ~~~~~~~~~~~~~~~~~~~~~~

void Pipe::initialize()
    {
    Entity::initialize();
    client_entity->type = ClientEntity::TYPE_LOGIC;
    create_input( get_input_name() );
    create_output( get_output_name() );
    }


Logic::SocketId Pipe::get_input_id()
    {
    // Pipe has hardwired default input called "input"
    // whose input socket id is always 0.
    
    return 0;
    }


Logic::SocketId Pipe::get_output_id()
    {
    // Pipe has hardwired default output called "output"
    // whose output socket id is always 0.
    
    return 0;
    }


std::string Pipe::get_input_name()
    {
    return "input";
    }


std::string Pipe::get_output_name()
    {
    return "output";
    }


Pointer<Logic::Packet> Pipe::get_input()
    {
    return Logic::get_input( get_input_id() );
    }


Logic::SocketId Pipe::translate_input
    (
    const std::string & input_name
    )
    {
    if( input_name == get_input_name() )
        {
        return get_input_id();
        }

    return Logic::translate_input( input_name );
    }


Logic::SocketId Pipe::translate_output
    (
    const std::string & output_name
    )
    {
    if( output_name == get_output_name() )
        {
        return get_output_id();
        }

    return Logic::translate_output( output_name );
    }


Logic::SocketId Pipe::create_input
    (
    const std::string & dynamic_name
    )
    {
    SocketId id;

    if( dynamic_name == get_input_name() )
        {
        id = Logic::create_input();
        hope( id == get_input_id() );
        }
    else
        {
        id = Logic::create_input( dynamic_name );
        }

    return id;
    }


Logic::SocketId Pipe::create_output
    (
    const std::string & dynamic_name
    )
    {
    SocketId id;

    if( dynamic_name == get_output_name() )
        {
        id = Logic::create_output();
        hope( id == get_output_id() );
        }
    else
        {
        id = Logic::create_output( dynamic_name );
        }

    return id;
    }


// Implementation of FloatInterpolator::KeyValue
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void FloatInterpolator::KeyValue::initialize
    (
    Float32 input_value,
    Float32 output_value
    )
    {
    input = input_value;
    requested_output = output_value;
    }


// Implementation of FloatInterpolator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void FloatInterpolator::initialize()
    {
    Pipe::initialize();
    }


void FloatInterpolator::add_key_value
    (
    Float32 input,
    Float32 requested_output
    )
    {
    add_value( key_values, input, requested_output );
    }


void FloatInterpolator::operate_synced
    (
    const STime &   delivery_time
    )
    {
    Pointer<FloatPacket> packet = get_input().convert();

    Logic::output
        (
        get_output_id(),
        CreateObject<FloatPacket>
            ( interpolate( packet->get_float(), key_values ) ),
        delivery_time
        );
    }


// Implementation of ColorInterpolator::KeyValue
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ColorInterpolator::KeyValue::initialize
    (
    Float32         input_value,
    const SColor &  output_value
    )
    {
    input = input_value;
    requested_output = output_value;
    }


// Implementation of ColorInterpolator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ColorInterpolator::initialize()
    {
    Pipe::initialize();
    }


void ColorInterpolator::add_key_value
    (
    Float32         input,
    const SColor &  requested_output
    )
    {
    add_value( key_values, input, requested_output );
    }


void ColorInterpolator::operate_synced
    (
    const STime &   delivery_time
    )
    {
    Pointer<FloatPacket> packet = get_input().convert();

    Logic::output
        (
        get_output_id(),
        CreateObject<ColorPacket>
            ( interpolate( packet->get_float(), key_values ) ),
        delivery_time
        );
    }


} // namespace Demo::Lib
} // namespace Demo
