/** @file pipe.h
 * Pipe.
 */

/*
 * begin        : Sat June 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_PIPE_H
#define DEMO_PIPE_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_ENTITY_H
#include "lib/server/entity.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Pipe object.
 * An entity whose logic has one default input ("input") and a default
 * output ("output").
 * There can be additional inputs and outputs that can be used to
 * control the pipe or other entities.
 */
class Pipe : public Entity
    {
protected:

    MASSIV_OBJECT( Pipe );

    //
    // Pipe interface.
    //

    /** Initializes object.*/
    virtual void initialize();

public:

    /**
     * Returns default input socket id.
     */
    static SocketId get_input_id();
    
    /**
     * Returns default output socket id.
     */
    static SocketId get_output_id();
    
    /**
     * Returns default input name.
     */
    static std::string get_input_name();

    /**
     * Returns default output name.
     */
    static std::string get_output_name();
    
    /**
     * Returns pipe default input.
     */
    Pointer<Packet> get_input();
    
protected:

    //
    // Logic interface implementation.
    //
    
    /**
     */
    virtual Logic::SocketId translate_input
        (
        const std::string & input_name
        );
    
    /**
     */
    virtual Logic::SocketId translate_output
        (
        const std::string & output_name
        );
    
    /**
     */
    virtual SocketId create_input
        (
        const std::string & dynamic_name = ""
        );
    
    /**
     */
    virtual SocketId create_output
        (
        const std::string & dynamic_name = ""
        );
    
    }; // Pipe


BASIC_LIB_EXCEPTION( InterpolatorErrorException, ::Massiv::Core::Lib::RuntimeException );
BASIC_LIB_EXCEPTION( InvalidKeyValueException, InterpolatorErrorException );


/**
 * Float interpolator.
 * The interpolator transforms a float input to a float output.
 * The output is computed using linear interpolation from a set of given
 * key values.
 */
class FloatInterpolator : public Pipe
    {
public:
    
    MASSIV_OBJECT( FloatInterpolator );

    /**
     * Key value.
     * For given @c input the interpolator is expected to produce
     * @c requested_output.
     * Other values are computed from key values using interpolation.
     */
    class KeyValue : public Object
        {
    public:

        typedef SFloat32    InputType;
        typedef SFloat32    OutputType;

        /** Initializes object. */
        void initialize
            (
            Float32 input_value,
            Float32 output_value
            );

    public:
            
        PFloat32    input;
            /**< Input value. */
        PFloat32    requested_output;
            /**< Requested output value. */
        }; // KeyValue

    typedef PArray<KeyValue>    KeyValues;

public:

    //
    // FloatInterpolator interface.
    //

    /**
     * Initializes the interpolator.
     * Initially there are no key values and the interpolator outputs
     * zero for any input.
     */
    virtual void initialize();

    /**
     * Adds a key value.
     * @c input is expected to be greater than any other @c KeyValue::input
     * inserted so far.
     */
    void add_key_value
        (
        Float32 input,
        Float32 requested_output
        );
    
protected:

    //
    // Logic interface implementation.
    //

    /**
     */
    virtual void operate_synced
        (
        const STime &   delivery_time
        );

protected:

    KeyValues   key_values;
        /**< Key values. */

    }; // FloatInterpolator


/**
 * Color interpolator.
 * The interpolator transforms a float input to a color output.
 * The output is computed using linear interpolation from a set of given
 * key values.
 */
class ColorInterpolator : public Pipe
    {
public:
    
    MASSIV_OBJECT( ColorInterpolator );

    /**
     * Key value.
     * For given @c input the interpolator is expected to produce
     * @c requested_output.
     * Other values are computed from key values using interpolation.
     */
    class KeyValue : public Object
        {
    public:

        typedef SFloat32    InputType;
        typedef SColor      OutputType;

        /** Initializes object. */
        void initialize
            (
            Float32         input_value,
            const SColor &  output_value
            );

    public:
            
        PFloat32    input;
            /**< Input value. */
        PColor      requested_output;
            /**< Requested output value. */
        }; // KeyValue

    typedef PArray<KeyValue>    KeyValues;

public:

    //
    // ColorInterpolator interface.
    //

    /**
     * Initializes the interpolator.
     * Initially there are no key values and the interpolator outputs
     * SColor() for any input.
     */
    virtual void initialize();

    /**
     * Adds a key value.
     * @c input is expected to be greater than any other @c KeyValue::input
     * inserted so far.
     */
    void add_key_value
        (
        Float32         input,
        const SColor &  requested_output
        );
    
protected:

    //
    // Logic interface implementation.
    //

    /**
     */
    virtual void operate_synced
        (
        const STime &   delivery_time
        );

protected:

    KeyValues   key_values;
        /**< Key values. */

    }; // ColorInterpolator


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_PIPE_H
