/** @file player.cpp
 * Player (implementation).
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/player.h"
#include "lib/server/item_slot_container.h"
#include "lib/server/item.h"
#include "lib/server/brain.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/client_item.h"
#include "lib/shared/account_object_rpc.h"


namespace Demo {
namespace Lib {


void Player::initialize()
    {
    // Initialize monster.
    Super::initialize();
    entity_flags = PLAYER;
    client_entity->type = ClientEntity::TYPE_PLAYER;
    brain = CreateObject< SimpleBrain >( this );
    physical_info.physical_class = PhysicalInfo::CLASS_PENETRABLE;
    physical_info.bounding_box_size = SVector2( 1.0f, 1.0f );
    physical_info.bounding_box_axis_aligned = true;
    physical_info.max_slope = 0.2f;

    // Name player.
    set_name( "unnamed player" );

    // Initialize abilities.
    client_entity->abilities.resize( ABILITY_COUNT );
    client_entity->set_ability_upper_limit( ABILITY_HEALTH, 100 );
    client_entity->set_ability_value( ABILITY_HEALTH, 60 );

    // Create inventory.
    inventory = CreateObject< ItemSlotContainer >( 5, 3 );
    client_entity->inventory = inventory->client_item_slot_container;

    // Start thinking.
    brain->think();
    }


void Player::handle_move_action
    (
    const PlayerMoveAction & action
    )
    {
    if( !brain->move_to( action.target_position.read() ) )
        {
        return;
        }

#ifdef OPTION_DEBUG
        {
        std::ostringstream oss;
        oss << get_object_id() << " moving to ("
            << action.target_position.x() << ", "
            << action.target_position.y() << ") "
            << "as requested by client.";

        Global::log_debug
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            oss.str()
            );
        }
#endif
    }


void Player::handle_item_action
    (
    const PlayerItemAction & action
    )
    {
    if( !verify_item_action( action ) )
        {
        return;
        }

    if( action.source == PlayerItemAction::SOURCE_INVENTORY )
        {
        // Source item slot must always contains an item.
        int source_slot_index = action.source_slot_index;
        hope( inventory->is_valid_slot_index( source_slot_index ) );
        hope( inventory->get_item( source_slot_index ) != null );
        hope( inventory->get_item( source_slot_index )->client_item == action.source_object );

        if( action.target == PlayerItemAction::TARGET_INVENTORY )
            {
            int target_slot_index = action.target_slot_index;
            hope( inventory->is_valid_slot_index( target_slot_index ) );

            if( inventory->get_item( target_slot_index ) == null )
                {
                // Use source item to empty slot -> move source item from one slot to another.
                inventory->put_item( target_slot_index, inventory->get_item( source_slot_index ) );
                inventory->put_item( source_slot_index, null );
                }
            }
        else if( action.target == PlayerItemAction::TARGET_ENTITY )
            {
            // Use source item on entity.

            hope( inventory->get_item( source_slot_index ) != null );

            // Get the item.
            Pointer< Item > item = inventory->get_item( source_slot_index );
            if( item == null )
                {
                return;
                }

            // Use the item.

            if( item->is_equipable() )
                {
                if( action.target_object == WeakPointer< Object >( this ) )
                    {
                    // Equip/unequip the item.
                    if( item->is_equipped()  )
                        {
                        say( std::string( "removes " ) + item->client_item->get_long_name(), MESSAGE_ACTION );
                        Pointer< Item > new_item = item->unequip( this );
                        item->set_equipped( false );
                        inventory->put_item( source_slot_index, new_item );
                        }
                    else
                        {
                        say( std::string( "equips " ) + item->client_item->get_long_name(), MESSAGE_ACTION );
                        Pointer< Item > new_item = item->equip( this );
                        if( new_item != null )
                            {
                            new_item->set_equipped( true );
                            }
                        inventory->put_item( source_slot_index, new_item );
                        }
                    }
                else
                    {
                    // Can't use equippable items on other entities

                    std::string target_name = "an entity";
                    try
                        {
                        WeakPointer< Monster > monster = action.target_object.convert();
                        if( monster.is_local() )
                            {
                            target_name = monster->get_name();
                            }
                        }
                    catch( ... )
                        {
                        }

                    say( std::string( "refuses to equip " ) + item->client_item->get_long_name() +
                         " on " + target_name, MESSAGE_ACTION );
                    }
                }
            else
                {
                // Use the item.
                Pointer< Item > new_item = item->use_on_monster( this, action.target_object.convert() );
                inventory->put_item( source_slot_index, new_item );
                }
            }
        else if( action.target == PlayerItemAction::TARGET_WORLD )
            {
            Pointer< Item > item = inventory->get_item( source_slot_index );
            if( item == null )
                {
                return;
                }
            Pointer< Item > new_item = item->use_on_sector( this );
            inventory->put_item( source_slot_index, new_item );
            }
        }
    }


void Player::log_game_message
    (
    const std::string & message,
    const SColor &      message_color
    ) const
    {
    log_message_event( null, message, message_color );
    }


void Player::log_message_event
    (
    WeakPointer< ClientEntity > entity,
    const std::string &         message,
    const SColor &              message_color
    ) const
    {
    if( account != null )
        {
        account->async_log_game_message( entity, message, message_color );
        }
    }


bool Player::put_item_to_inventory
    (
    const Pointer< Item > & item
    )
    {
    hope( item != null );

    unsigned int i;

    for( i = 0; i < inventory->slots.size(); ++i )
        {
        if( inventory->get_item( i ) == null )
            {
            inventory->put_item( i, item );
            return true;
            }
        }

    return false;
    }


Pointer< Item > Player::remove_item_from_inventory
    (
    int slot_index
    )
    {
    hope( inventory->is_valid_slot_index( slot_index ) );

    // Get the item.
    Pointer< Item > item = inventory->get_item( slot_index );

    // Clear the slot.
    inventory->put_item( slot_index, null );

    // If the item is equipped, unequip it.
    if( item != null && item->is_equipable() && item->is_equipped() )
        {
        item = item->unequip( this );
        if( item != null )
            {
            item->set_equipped( false );
            }
        }

    return item;
    }


bool Player::verify_item_action
    (
    const PlayerItemAction & action
    )
    {
    bool source_ok = false;
    bool target_ok = false;
    PlayerItemAction::Source source = action.source;
    PlayerItemAction::Target target = action.target;

    switch( source )
        {
        case PlayerItemAction::SOURCE_INVENTORY:
            // Verify source slot index.
            if( !inventory->is_valid_slot_index( action.source_slot_index ) )
                {
                log_game_message( "Invalid source inventory slot index." );
                return false;
                }

            // Verify source slot object.
                {
                // Source item slot must always contains an item.
                Pointer< Item > item = inventory->get_item( action.source_slot_index );
                if( item == null || item->client_item != action.source_object )
                    {
                    log_game_message( "Requested action with inventory item but the source item "
                        "is not at specified inventory position" );
                    return false;
                    }
                }

            source_ok = true;
            break;

        case PlayerItemAction::SOURCE_ABILITIES:
            break;

        default:
            break;
        }

    if( !source_ok )
        {
        log_game_message( "Unsupported action (unsupported source)." );
        return false;
        }

    switch( target )
        {
        case PlayerItemAction::TARGET_INVENTORY:
            // Verify target slot index.
            if( !inventory->is_valid_slot_index( action.target_slot_index ) )
                {
                log_game_message( "Invalid target inventory slot index." );
                return false;
                }

            // Verify target slot object.
                {
                Pointer< Item > item = inventory->get_item( action.target_slot_index );
                bool invalid_target;

                if( item == null )
                    {
                    if( action.target_object == null )
                        {
                        invalid_target = false;
                        }
                    else
                        {
                        invalid_target = true;
                        }
                    }
                else
                    {
                    if( item->client_item == action.target_object )
                        {
                        invalid_target = false;
                        }
                    else
                        {
                        invalid_target = true;
                        }
                    }

                if( invalid_target )
                    {
                    log_game_message( "Requested action with inventory item but the target item "
                        "is not at specified inventory position" );
                    return false;
                    }
                }

            target_ok = true;
            break;

        case PlayerItemAction::TARGET_ENTITY:
            if( action.target_object != null )
                {
                target_ok = true;
                }
            break;

        case PlayerItemAction::TARGET_WORLD:
            target_ok = true;
            break;
        }

    if( !target_ok )
        {
        log_game_message( "Unsupported action (unsupported target)." );
        return false;
        }

    return true;
    }


} // namespace Demo::Lib
} // namespace Demo
