/** @file player.h
 * Player.
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_PLAYER_H
#define DEMO_PLAYER_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_PLAYER_ACTION_H
#include "lib/shared/player_action.h"
#endif

#ifndef DEMO_MONSTER_H
#include "lib/server/monster.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Player.
 */
class Player : public Monster
    {
    typedef Monster Super; // Leave it private.

public:

    MASSIV_OBJECT( Player );

    /** Initialize player object. */
    virtual void initialize();

public:

    /**
     * Handle move action requested by client.
     */
    void handle_move_action
        (
        const PlayerMoveAction & action
        );

    /**
     * Handle item action requested by client.
     */
    void handle_item_action
        (
        const PlayerItemAction & action
        );

public:

    /**
     * Log message to client that owns this player object.
     * This method should be used to log messages notifying
     * the client of some ingame actions of the player,
     * for example "health potion picked/used" etc.
     * The message is sent through client's account object.
     */
    void log_game_message
        (
        const std::string & message,
        const SColor &      message_color = SColor( 1.0f, 1.0f, 1.0f )
        ) const;

    /**
     * Log message event to client that owns this player object.
     * Message event is text event when somebody nearby the player
     * say something.
     * If @a entity is null, the function is same as @c log_game_message().
     * @param entity Entity that generated the message. May be null.
     *      Client can use this pointer to place the message above
     *      the entity on screen.
     * @param message Text message,
     * @param message_color Color of the message.
     */
    void log_message_event
        (
        WeakPointer< ClientEntity > entity,
        const std::string &         message,
        const SColor &              message_color = SColor( 1.0f, 1.0f, 1.0f )
        ) const;

public:

    /**
     * Put item into player's inventory.
     * @return @c true on success, @c false if the inventory is full.
     */
    bool put_item_to_inventory
        (
        const Pointer< Item > & item
        );

private:

    /**
     * Remove item from specified inventory slot.
     * The item is unequipped if equipped.
     * May return null.
     */
    Pointer< Item > remove_item_from_inventory
        (
        int slot_index
        );

    /**
     * Verify player item action from client.
     */
    bool verify_item_action
        (
        const PlayerItemAction & action
        );

public:

    PRemote< AccountObject >        account;
        /**< Player's account. */
    PPointer< ItemSlotContainer >   inventory;
        /**< Player's inventory. */

    }; // Player


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_PLAYER_H
