/** @file robot.cpp
 * Robot implementation.
 */

/*
 * begin        : Mon Nov 17 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/robot.h"
#include "lib/server/robot_rpc.h"
#include "lib/server/brain.h"
#include "lib/server/lock.h"
#include "lib/server/entity_rpc.h"
#include "lib/server/sector.h"
#include "lib/server/naming_service.h"
#include "lib/server/naming_service_rpc.h"
#include "lib/server/entity_factory_rpc.h"
#include "lib/shared/client_entity.h"
#include "lib/shared/random.h"
#include <limits>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of Robot::AreaScan
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Robot::AreaScan::initialize
    (
    Pointer<Robot>      the_robot,
    Pointer<Sector>     the_current_sector,
    const ConvexArea &  the_area,
    const SVector2 &    the_position,
    Float32             the_heading,
    int                 the_entity_flags
    )
    {
    robot = the_robot;

    SpatialQuery::initialize
        (
        the_current_sector,
        the_area,
        the_position,
        the_position,
        the_heading,
        the_entity_flags,
        false
        );
    }


bool Robot::AreaScan::entity_reported
    (
    WeakPointer<Entity> entity
    )
    {
    // @c entity is automatically put into @c reported lists
    // (SpatialQuery operates "synchronously").

    return true;
    }


void Robot::AreaScan::enumeration_finished()
    {
    deliver_asap_to( robot );
    }


void Robot::AreaScan::delivered_to
    (
    WeakPointer<Object> object,
    const STime &       time
    )
    {
    if( object == robot )
        {
        Pointer<Robot> my_robot = object.convert();
        my_robot->scan_done( this );
        }
    else
        {
        Super::delivered_to( object, time );
        }
    }


// Implementation of Robot
// ~~~~~~~~~~~~~~~~~~~~~~~

void Robot::initialize()
    {
    const Float32   fov = 3.0f * MATHF_PI_4;
    const Float32   distance = 40.0f;
    const Float32   parallel_vision_size = distance * 0.3f;
    const Float32   back_vision_size = distance * 0.3f;

    // Initialize monster.

    Super::initialize();
    brain = CreateObject< RobotBrain >( this );
    physical_info.physical_class = PhysicalInfo::CLASS_PENETRABLE;
    physical_info.bounding_box_size = SVector2( 1.0f, 1.0f );
    physical_info.bounding_box_axis_aligned = true;
    physical_info.max_slope = 0.2f;
    physical_info.mass = 100.0f;

    // Initialize abilities.

    client_entity->abilities.resize( ABILITY_COUNT );
    client_entity->set_ability_upper_limit( ABILITY_HEALTH, 50 );
    client_entity->set_ability_value( ABILITY_HEALTH, 50 );

    // Initialize robot.

    set_name( "sheep" );
    area.add_vertex( SVector2( -parallel_vision_size, -back_vision_size ) );
    area.add_vertex( SVector2( parallel_vision_size, -back_vision_size ) );
    area.add_vertex( SVector2(
        FloatTraits<Float32>::cos( ( MATHF_PI - fov ) ) * distance,
        FloatTraits<Float32>::sin( ( MATHF_PI - fov ) ) * distance
        ) );
    area.add_vertex( SVector2(
        -FloatTraits<Float32>::cos( ( MATHF_PI - fov ) ) * distance,
        FloatTraits<Float32>::sin( ( MATHF_PI - fov ) ) * distance
        ) );
    area.update_edges();

    area_entity_flags = PLAYER;

    // Start thinking.

    brain->think();
    }


void Robot::link_to_boss
    (
    Remote<Robot> boss
    )
    {
    brain->link_to_boss( boss->sync_get_brain() );
    }


void Robot::check_visible()
    {
    if( current_sector == null )
        {
        return;
        }

    if( scan_in_progress )
        {
        return;
        }

    scan_in_progress = true;

    CreateObject<AreaScan>
        (
        this,
        current_sector,
        area,
        client_entity->current_position.origin.read(),
        client_entity->current_position.orientation.heading(),
        area_entity_flags
        );
    }


void Robot::monster_died
    (
    ItemType & item_type,
    Int32 &    item_variant
    )
    {
    // Return meat.

    item_type = ITEM_POTION;
    item_variant = POTION_MEAT;

    // We want to remove the sheep from the map and put it back in 30 seconds.
    // Unfortunately that's nearly impossible to do - if an entity is removed
    // from the layer, GC will collect it in the next run. To prevent it, we
    // would have to migrate to a special object ("limbo") and wait there until
    // respawn time. The brain would have to understand that an entity is
    // "in limbo", delivered_to() would have to be rewritten, etc.
    // So just commit suicide and schedule task that will spawn another entity.

    Remote< NamingService > naming( Massiv::System::get_naming_service_object_id() );
    Remote< EntityFactory > factory( naming->sync_get_object( "EntityFactory" ).convert() );
    if( factory == null )
        {
        // No factory is registered - can't spawn new entity.
        }
    else
        {
        const SVector2 position_delta
            (
            32.0f * Random::random() - 16.0f,
            32.0f * Random::random() - 16.0f
            );
        factory->param( RPC_DELAYED, 30.0f ).async_create_entity_in_layer
            (
            "Demo::Lib::Robot",
            ( client_entity->current_position.origin + position_delta ).read(),
            client_entity->current_position.orientation.read(),
            client_entity->current_position.layer.convert(),
            physical_info.physical_class == PhysicalInfo::CLASS_SOLID,
            physical_info.bounding_box_size.read(),
            physical_info.bounding_box_axis_aligned,
            client_entity->model_tags
            );
        }

    // Stop and schedule selfdestruction.
    // Delay must be larger than "say" delay in Monster::attack_by()!

    brain->cancel_move();
    Remote< Robot > self( this );
    self->param( RPC_DELAYED, 2.0f ).async_remove_from_layer();
    }


void Robot::scan_done
    (
    Pointer<AreaScan> scan
    )
    {
    Pointer<Brain> my_brain = brain.convert();

    scan_in_progress = false;
    my_brain->report_visible( scan->reported );
    }


// Implementation of RobotBrain
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

namespace {

const Float32   sniff_around_to_look_around_probability = 0.2f;
    /**< Sniff-around to look-around goal change probability. */
const Float32   look_around_to_sniff_around_probability = 0.2f;
    /**< Look-around to sniff-around goal change probability. */
const Float32   follow_to_look_around_probability = 0.1f;
    /**< Follow to look-around goal change probability. */
const Float32   guard_boss_to_look_around_probability = 0.1f;
    /**< Guard-boss to look-around goal change probability. */
const Float32   sniff_around_distance = 10.0f;
    /**< Sniff-around distance (mean value). */
const Float32   sniff_around_think_delay = 60.0f;
    /**< Think delay for @c GOAL_SNIFF_AROUND (mean value). */
const Float32   look_around_think_delay = 8.0f;
    /**< Think delay for @c GOAL_LOOK_AROUND (mean value). */
const Float32   check_visible_delay = 10.0f;
    /**< Check visible delay. */
const Float32   retreat_think_delay = 10.0f;
    /**< Think delay for @c GOAL_RETREAT (mean value). */
const Float32   follow_think_delay = 40.0f;
    /**< Think delay for @c GOAL_FOLLOW_ENTITY (mean value). */
const Float32   simple_follow_think_delay = 20.0f;
    /**< Think delay for @c GOAL_FOLLOW_ENTITY in "simple" mode (mean value). */
const Float32   guard_boss_think_delay = 10.0f;
    /**< Think delay for @c GOAL_GUARD_BOSS (mean value). */
const Float32   idle_delay = 10.0f;
    /**< Idle delay. */
const Float32   random_target_retry_delay = 10.0f;
    /**< If path to a random target could not be found retry in
         given number of seconds (mean value). */
const Float32   replication_timeout = 30.0f;
    /**< Replication timeout. */
const Float32   retreat_if_distance_less_than = 2.5f;
    /**< Retreat if distance to a closest visible entity is less than the
         given constant. */
const Float32   summon_slaves_if_distance_less_than = 2.0f;
    /**< Summon slaves if distance to a closest visible entity is less than the
         given constant. */
const Float32   follow_if_distance_greater_than = 6.0f;
    /**< Follow entity if distance to it is greater than the given constant. */
const Float32   stop_following_if_distance_greater_than = 20.0f;
    /**< Stop following entity if distance to it is greated than the constant. */
const Float32   follow_distance = 4.0f;
    /**< Follow distance. */
const Float32   simple_follow_probability = 0.5f;
    /**< Probability of "simple" following over to "advanced" following. */
const Float32   retreat_probability = 0.8f;
    /**< Retreat probability. */
const Float32   retreat_run_period = 5.0f;
    /**< How long monster is able to run if retreating (mean value). */
const Float32   follow_probability = 0.25f;
    /**< Follow probability. */
const Int32     run_speed_increment = 30;
    /**< @c ABILITY_SPEED increment if monster should run. */
const Int32     num_material_tests = 4;
    /**< How many tests to perform when searching for @c prefer_this_material_id. */

/**
 * Adjusts probability as a function of time.
 * The function should be increasing.
 */
Float32 adjust_probability
    (
    Float32         probability,
    const STime &   time
    )
    {
    probability *= FloatTraits<Float32>::exp( static_cast<Float32>(
        time.get_value() / 10.0f - 1.0f ) );

    if( probability > 1.0f )
        {
        probability = 1.0f;
        }

    return probability;
    }


} // anonymous namespace


void RobotBrain::initialize
    (
    Pointer<Monster> the_monster
    )
    {
    Super::initialize( the_monster );
    prefer_this_material_id = 1; // FIXME: This is hardcoded!
    }


void RobotBrain::think()
    {
    // Check for recursion.

    if( thinking )
        {
        return;
        }

    OPTIMISTIC_LOCK( thinking, NULL );

    // Initialize.

    EntityPosition  current_position = monster->get_current_position();
    STime           current_time = STime::current_time();
    STime           goal_time = current_time - goal_change_time;

    Super::think();

    // Check for visible entities.

    if( current_time < check_visible_time )
        {
        rethink( check_visible_time.read() - current_time );
        }
    else
        {
        monster->check_visible();
        schedule_check_visible( check_visible_delay );
        }

    Remote<Entity>  closest_entity;
    Float32         closest_distance = get_distance_to_closest_visible_entity
        ( current_position.origin.read(), closest_entity );

    // Think.

    if( current_time < robot_think_time )
        {
        rethink( robot_think_time.read() - current_time );
        return;
        }

    if( robot_goal == GOAL_SNIFF_AROUND || robot_goal == GOAL_LOOK_AROUND )
        {
        if( closest_distance < retreat_if_distance_less_than &&
            Random::random() < retreat_probability )
            {
            // Retreat.

            monster->apply_powerup
                (
                ABILITY_SPEED,
                run_speed_increment,
                0,
                Entity::POWERUP_TEMPORARY,
                retreat_run_period * Random::random() * 2.0f
                );

            #ifdef OPTION_DEBUG

            Global::log_debug
                (
                Status::FACILITY_LIB,
                Status::PRIORITY_NORMAL,
                monster.to_string() + " changes its goal to GOAL_RETREAT."
                );

            #endif

            robot_goal = GOAL_RETREAT;
            goal_change_time = current_time;
            }
        else
        if( closest_distance > follow_if_distance_greater_than &&
            closest_distance < stop_following_if_distance_greater_than &&
            closest_entity != null && Random::random() < follow_probability )
            {
            // Follow.

            #ifdef OPTION_DEBUG

            Global::log_debug
                (
                Status::FACILITY_LIB,
                Status::PRIORITY_NORMAL,
                monster.to_string() + " changes its goal to GOAL_FOLLOW_ENTITY. " +
                closest_entity.to_string() + " will be followed."
                );

            #endif

            followed_entity = closest_entity;

            robot_goal = GOAL_FOLLOW_ENTITY;
            goal_change_time = current_time;
            }
        }

    switch( robot_goal )
        {
        case GOAL_SNIFF_AROUND:
            {
            // Change state to @c GOAL_LOOK_AROUND with a given probability.

            if( adjust_probability( sniff_around_to_look_around_probability, goal_time ) >
                Random::random() )
                {
                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " changes its goal to GOAL_LOOK_AROUND."
                    );

                #endif

                robot_goal = GOAL_LOOK_AROUND;
                goal_change_time = current_time;

                schedule_robot_think( STime() );
                break;
                }

            // Choose move target.

            Pointer<Sector> current_sector = monster->get_current_sector();
            SVector2        my_position;
            SVector2        new_target;
            SVector2        tile_xy;
            int             material;
            int             i;

            if( boss == null )
                {
                my_position = current_position.origin.read();
                }
            else
                {
                my_position = boss->optimize_replica( replication_timeout ).
                    sync_get_current_position().origin.read();
                }

            for( i = 0; i < num_material_tests; i++ )
                {
                new_target = my_position + Random::random_vector() *
                    sniff_around_distance * 2.0f;

                if( current_sector != null )
                    {
                    current_sector->layer_xy_to_tile_xy( new_target, tile_xy );
                    material = current_sector->get_tile_material
                        (
                        FloatTraits<Float32>::floor_int( tile_xy.x() ),
                        FloatTraits<Float32>::floor_int( tile_xy.y() )
                        );

                    if( material != prefer_this_material_id )
                        {
                        continue;
                        }
                    }

                break;
                }

            Brain::move_to_along_line( new_target );

            // Done.

            schedule_robot_think( Random::random() * sniff_around_think_delay * 2.0f );
            break;
            }

        case GOAL_LOOK_AROUND:
            {
            // Change state to @c GOAL_SNIFF_AROUND with a given probability.

            if( adjust_probability( look_around_to_sniff_around_probability, goal_time ) >
                Random::random() )
                {
                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " changes its goal to GOAL_SNIFF_AROUND."
                    );

                #endif

                robot_goal = GOAL_SNIFF_AROUND;
                goal_change_time = current_time;

                schedule_robot_think( STime() );
                break;
                }

            // Select a random point to look at.

            SVector2 new_target = current_position.origin.read();
            new_target += Random::random_vector();

            // Look at that position and schedule next @c think().

            look_at( new_target );

            schedule_robot_think( Random::random() * look_around_think_delay * 2.0f );
            break;
            }

        case GOAL_RETREAT:
            {
            // Change state to @c GOAL_LOOK_AROUND if there is no need to retreat
            // any more.

            if( closest_distance > retreat_if_distance_less_than )
                {
                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " changes its goal to GOAL_LOOK_AROUND."
                    );

                #endif

                robot_goal = GOAL_LOOK_AROUND;
                goal_change_time = current_time;

                schedule_robot_think( STime() );
                break;
                }

            // Summon slaves to guard self.

            if( closest_distance < summon_slaves_if_distance_less_than )
                {
                std::size_t i;

                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " summons its slaves to guard self."
                    );

                #endif

                for( i = 0; i < slaves.size(); i++ )
                    {
                    try
                        {
                        Remote<RobotBrain> slave_brain = slaves[ i ].convert();
                        slave_brain->async_report_guard_boss();
                        }
                    catch( Throwable & )
                        {
                        }
                    }
                }

            // Select a nearby "safe" position.

            SVector2 new_target = current_position.origin.read();
            new_target += Random::random_vector() * retreat_if_distance_less_than * 2.0f;

            // Move to the new position if it is safe, the path itself
            // need not be safe.

            Remote<Entity>  dummy;
            STime           delay;

            if( !Brain::move_to( new_target ) )
                {
                delay = STime( Random::random() * random_target_retry_delay * 2.0f );

                if( get_distance_to_closest_visible_entity( new_target, dummy ) <
                    retreat_if_distance_less_than )
                    {
                    delay = STime();
                    }
                }
            else
                {
                delay = STime( Random::random() * retreat_think_delay * 2.0f );
                }

            // Done.

            schedule_robot_think( delay );
            break;
            }

        case GOAL_FOLLOW_ENTITY:
            {
            EntityPosition position = followed_entity->optimize_replica
                ( replication_timeout ).sync_get_current_position();

            // Change state to @c GOAL_LOOK_AROUND with a given probability.

            if( adjust_probability( follow_to_look_around_probability, goal_time ) >
                Random::random() || ( position.origin - current_position.origin ).magnitude() >
                stop_following_if_distance_greater_than )
                {
                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " changes its goal to GOAL_LOOK_AROUND."
                    );

                #endif

                robot_goal = GOAL_LOOK_AROUND;
                goal_change_time = current_time;

                schedule_robot_think( STime() );
                break;
                }

            // Move to a safe random position close to the followed entity.

            SVector2 new_target = position.origin.read();
            new_target += Random::random_vector() * follow_distance;

            // Move to the new position if it is safe, the path itself
            // need not be safe.

            Remote<Entity>  dummy;
            STime           delay;

            if( get_distance_to_closest_visible_entity( new_target, dummy ) <
                retreat_if_distance_less_than )
                {
                delay = STime();
                }
            else
                {
                if( Random::random() < simple_follow_probability )
                    {
                    // "Simple" following.

                    Brain::move_to_along_line( new_target );
                    delay = STime( Random::random() * simple_follow_think_delay * 2.0f );
                    }
                else
                    {
                    // "Advanced" following.

                    if( !Brain::move_to( new_target ) )
                        {
                        delay = STime( Random::random() * random_target_retry_delay * 2.0f );
                        }
                    else
                        {
                        delay = STime( Random::random() * follow_think_delay * 2.0f );
                        }
                    }
                }

            // Done.

            schedule_robot_think( delay );
            break;
            }

        case GOAL_GUARD_BOSS:
            {
            EntityPosition position = boss->optimize_replica
                ( replication_timeout ).sync_get_current_position();

            // Change state to @c GOAL_LOOK_AROUND with a given probability.

            if( adjust_probability( guard_boss_to_look_around_probability, goal_time ) >
                Random::random() )
                {
                #ifdef OPTION_DEBUG

                Global::log_debug
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    monster.to_string() + " changes its goal to GOAL_LOOK_AROUND."
                    );

                #endif

                robot_goal = GOAL_LOOK_AROUND;
                goal_change_time = current_time;

                schedule_robot_think( STime() );
                break;
                }

            // Run to boss.

            // Move to the new position if it is safe, the path itself
            // need not be safe.

            SVector2 new_target = position.origin.read();
            new_target += Random::random_vector() * follow_distance;

            Remote<Entity>  dummy;
            STime           delay;

            if( get_distance_to_closest_visible_entity( new_target, dummy ) <
                retreat_if_distance_less_than )
                {
                delay = STime();
                }
            else
                {
                if( !Brain::move_to( new_target ) )
                    {
                    delay = STime( Random::random() * random_target_retry_delay * 2.0f );
                    }
                else
                    {
                    delay = STime( Random::random() * guard_boss_think_delay * 2.0f );
                    }
                }

            monster->apply_powerup
                ( ABILITY_SPEED, run_speed_increment, 0, Entity::POWERUP_TEMPORARY, delay );

            // Done.

            schedule_robot_think( delay );
            break;
            }
        }

    // Delay thinking if there is no important entity nearby.

    // TODO: There should be two different tests, one that checks for
    //       visible entities, the other that checks for all nearby
    //       entities (connected players).

    if( !visible_entities.size() )
        {
        rethink( idle_delay * Random::random() * 2.0f, true );
        }
    }


void RobotBrain::report_visible
    (
    const VisibleList & visible
    )
    {
    #ifdef OPTION_DEBUG

    std::ostringstream ss;

    ss << monster << " sees " << visible << ".";

    Global::log_debug
        (
        Status::FACILITY_LIB,
        Status::PRIORITY_NORMAL,
        ss.str()
        );

    #endif

    visible_entities = visible;
    schedule_robot_think( STime() );
    }


void RobotBrain::report_guard_boss()
    {
    if( thinking )
        {
        return;
        }

    #ifdef OPTION_DEBUG

    Global::log_debug
        (
        Status::FACILITY_LIB,
        Status::PRIORITY_NORMAL,
        monster.to_string() + " changes its goal to GOAL_GUARD_BOSS."
        );

    #endif

    robot_goal = GOAL_GUARD_BOSS;
    goal_change_time = STime::current_time();

    schedule_robot_think( STime() );
    }


Float32 RobotBrain::get_distance_to_closest_visible_entity
    (
    const SVector2 &    position,
    Remote<Entity> &    closest_entity
    )
    {
    VisibleList::Iterator   it = visible_entities.begin();
    Float32                 closest_distance = std::numeric_limits<Float32>::max();

    closest_entity = null;

    while( it != visible_entities.end() )
        {
        Remote<Entity>  entity = *it;
        EntityPosition  entity_position = entity->optimize_replica
            ( replication_timeout ).sync_get_current_position();
        SVector2        delta = ( entity_position.origin - position ).read();
        Float32         distance = delta.magnitude();

        if( distance < closest_distance )
            {
            closest_distance = distance;
            closest_entity = entity;
            }

        ++it;
        }

    return closest_distance;
    }


void RobotBrain::schedule_robot_think
    (
    const STime & delay
    )
    {
    robot_think_time = STime::current_time() + delay;
    rethink( delay );
    }


void RobotBrain::schedule_check_visible
    (
    const STime & delay
    )
    {
    check_visible_time = STime::current_time() + delay;
    rethink( delay );
    }


} // namespace Demo::Lib
} // namespace Demo
