/** @file robot.h
 * Robot.
 */

/*
 * begin        : Mon Nov 17 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_ROBOT_H
#define DEMO_ROBOT_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_MONSTER_H
#include "lib/server/monster.h"
#endif

#ifndef DEMO_SPATIAL_QUERY_H
#include "lib/server/spatial_query.h"
#endif

#ifndef DEMO_BRAIN_H
#include "lib/server/brain.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Robot, a monster controlled by a simple AI.
 */
class Robot : public Monster
    {
public:

    typedef Monster Super;

    MASSIV_OBJECT( Robot );

    /**
     * Object that collects physicals belonging to a particular area.
     * The objects are then reported to robot's brain.
     */
    class AreaScan : public SpatialQuery
        {
    public:

        MASSIV_OBJECT( AreaScan );

        typedef SpatialQuery Super;

        /** Initializes object. */
        void initialize
            (
            Pointer<Robot>      the_robot,
            Pointer<Sector>     the_current_sector,
            const ConvexArea &  the_area,
            const SVector2 &    the_position,
            Float32             the_heading,
            int                 the_entity_flags
            );

        /** Stores reported entity to a cache. */
        virtual bool entity_reported
            (
            WeakPointer<Entity> entity
            );

        /** Migrates scan object to @c robot. */
        virtual void enumeration_finished();

        /**
         * Notifies Robot that @c enumeration_finished() if object was
         * delivered to a @c Robot object (in that case object migration
         * must have been initiated by @c enumeration_finished).
         */
        virtual void delivered_to
            (
            WeakPointer<Object> object,
            const STime &       time
            );


    public:

        PRemote<Robot>  robot;
            /**< Robot that initiated the scan. Should be in a different
                 migration group than the scan object. */

        }; // AreaScan

    friend class AreaScan;

public:

    /**
     * Initializes robot object.
     */
    virtual void initialize();

    /**
     * Links the robot to its boss.
     */
    void link_to_boss
        (
        Remote<Robot> boss
        );

    /**
     * Triggers new area scan to check what entities are visible.
     * Once done, @c scan_done() will be called which in turn
     * notifies robot's brain via @c Brain::report_visible().
     */
    virtual void check_visible();

    /**
     * Returns a meat.
     */
    virtual void monster_died
        (
        ItemType & item_type,
        Int32 &    item_variant
        );

private:

    /**
     * Called by @c AreaScan when the scan has finished.
     * The implementation calls @c Brain::report_visible().
     */
    void scan_done
        (
        Pointer<AreaScan> scan
        );

protected:

    ConvexArea  area;
        /**< Area that bounds Robot's field of visibility. Defined relative
             to robot's position and "neutral" orientation. */
    PVlInt32    area_entity_flags;
        /**< Entity flags of physicals that are "monitored" by Robot. */
    PBool       scan_in_progress;
        /**< Is area scan in progess? */

    }; // Robot


/**
 * More sophisticated brain that can be used to control "robots".
 * Besides simple low-level "move-to" goal it handles advanced "robot goals",
 * such as "follow entity", etc.
 */
class RobotBrain : public SimpleBrain
    {
public:

    MASSIV_OBJECT( RobotBrain );

    typedef SimpleBrain Super;

    enum RobotGoal
        {
        GOAL_SNIFF_AROUND   =   0,
            /**< Sniff-around. Move to random nearby targets. */
        GOAL_LOOK_AROUND    =   1,
            /**< Stand-still and look around. */
        GOAL_FOLLOW_ENTITY  =   2,
            /**< Follow selected entity. */
        GOAL_GUARD_BOSS     =   3,
            /**< Guard my boss. */
        GOAL_RETREAT        =   4
            /**< Move further from the visible entities. */
        };

public:

    /** Initializes object. */
    void initialize
        (
        Pointer<Monster> the_monster
        );

protected:

    /**
     */
    virtual void think();

    /**
     */
    virtual void report_visible
        (
        const VisibleList & visible
        );

public:

    /**
     * Forces monster controlled by this brain to guard its boss.
     */
    void report_guard_boss();

private:

    /**
     * Returns closest visible entity from @c position and distance from
     * @c position to that entity.
     */
    Float32 get_distance_to_closest_visible_entity
        (
        const SVector2 & position,
        Remote<Entity> & closest_entity
        );

    /** Schedule robot's think in @c delay seconds. */
    void schedule_robot_think
        (
        const STime & delay
        );

    /** Schedule check for visible entities in @c delay seconds. */
    void schedule_check_visible
        (
        const STime & delay
        );

public:

    PEnum<RobotGoal>    robot_goal;
        /**< Current goal. */
    PTime               robot_think_time;
        /**< When to do next decision. */
    PTime               goal_change_time;
        /**< When @c robot_goal was changed for the last time. */
    PTime               check_visible_time;
        /**< When next check for visible entities should be performed. */
    VisibleList         visible_entities;
        /**< List of entities visible when the visibility check was done
             for the last time. */
    PRemote<Entity>     followed_entity;
        /**< Entity being followed. */
    PBool               thinking;
        /**< Is @c think() active now? */
    PVlInt32            prefer_this_material_id;
        /**< Prefer this material id when thinking where to go. */

    }; // RobotBrain


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_ROBOT_H
