/** @file sector.cpp
 * Sector implementation.
 */

/*
 * begin        : Mon Jun 31 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/sector.h"
#include "lib/server/sector_rpc.h"
#include "lib/server/layer_rpc.h"
#include "lib/server/sector_helper_objects.h"
#include "lib/server/collision_info.h"
#include "lib/shared/client_entity.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/** @def MARK_RESERVED_TILES
 * Define to set TILE_MARK bit of reserved tiles.
 */
#define MARK_RESERVED_TILES

/** @def LOG_FAILED_RESERVATIONS
 * Define to enable logging of failed reservation attempts.
 */
//#define LOG_FAILED_RESERVATIONS


// Implementation of Sector::TileReservation
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Sector::TileReservation::TileReservation()
    {
    initialize_object();
    default_constructed();
    }


Sector::TileReservation::TileReservation
    (
    Remote<Entity>              entity,
    PhysicalInfo::PhysicalClass physical_class,
    int                         entity_flags
    ) : entity( entity ),
        physical_class( physical_class ),
        entity_flags( entity_flags )
    {
    initialize_object();
    }


Sector::TileReservation::TileReservation
    (
    const TileReservation & reservation
    ) : ValueType( reservation ),
        entity( reservation.entity ),
        physical_class( reservation.physical_class ),
        entity_flags( reservation.entity_flags )
    {
    initialize_object();
    }


// Implementation of Sector
// ~~~~~~~~~~~~~~~~~~~~~~~~

void Sector::initialize()
    {
    client_sector = CreateObject<ClientSector>();
    
    neighbors.resize( 8 );
    entity_list = CreateObject<List>();
    
    client_sector->elevation_map.resize( ClientSector::NUM_X_TILES * ClientSector::NUM_Y_TILES );
    client_sector->delta_elevation_map.resize( ClientSector::NUM_X_TILES * ClientSector::NUM_Y_TILES );
    client_sector->material_map.resize( ClientSector::NUM_X_TILES * ClientSector::NUM_Y_TILES );
    client_sector->tile_flag_map.resize( ClientSector::NUM_X_TILES * ClientSector::NUM_Y_TILES );

    delta_map_decay_factor = 0.9f;
    delta_map_decay_period = PTime( 10.0 );
    max_delta_map_magnitude = 1.0f; // FIXME?

    Remote<Sector> self = this;
    self->async_update_delta_elevation_map();
    }


SectorAddress Sector::get_address() const
    {
    return client_sector->sector_address;
    }


SVector2 Sector::get_origin() const
    {
    return sector_origin.read();
    }


Pointer<ClientSector> Sector::get_client_sector() const
    {
    return client_sector;
    }


void Sector::set_address
    (
    const SectorAddress & the_sector_address
    )
    {
    client_sector->sector_address = the_sector_address;
    }


Direction Sector::opposite_direction
    (
    Direction direction
    )
    {
    #ifndef OPTION_OPTIMIZE

    switch( direction )
        {
        case EAST:
            return WEST;
        case NORTH_EAST:
            return SOUTH_WEST;
        case NORTH:
            return SOUTH;
        case NORTH_WEST:
            return SOUTH_EAST;
        case WEST:
            return EAST;
        case SOUTH_WEST:
            return NORTH_EAST;
        case SOUTH:
            return NORTH;
        case SOUTH_EAST:
            return NORTH_WEST;
        }
    
    panic( "Invalid direction." );
    
    #else
    
    return static_cast<Direction>( ( direction + 4 ) & 7 );

    #endif
    }


Remote<Sector> Sector::get_neighbor
    (
    Direction neighbor
    ) const
    {
    return neighbors[ neighbor ].convert();
    }
        
    
Remote<Sector> Sector::get_sector
    (
    const SectorAddress & address
    ) const
    {
    Int32 x_difference = address.column - client_sector->sector_address.column;
    Int32 y_difference = address.row - client_sector->sector_address.row;

    // Test neighboring sectors.

    if( address.layer == client_sector->sector_address.layer )
        {
        if( x_difference == -1 )
            {
            if( y_difference == -1 )
                {
                return get_neighbor( SOUTH_WEST );
                }
            else
            if( y_difference == 0 )
                {
                return get_neighbor( WEST );
                }
            else
            if( y_difference == 1 )
                {
                return get_neighbor( NORTH_WEST );
                }
            }
        else
        if( x_difference == 0 )
            {
            if( y_difference == -1 )
                {
                return get_neighbor( SOUTH );
                }
            else
            if( y_difference == 0 )
                {
                return this;
                }
            else
            if( y_difference == 1 )
                {
                return get_neighbor( NORTH );
                }            
            }
        else
        if( x_difference == 1 )
            {
            if( y_difference == -1 )
                {
                return get_neighbor( SOUTH_EAST );
                }
            else
            if( y_difference == 0 )
                {
                return get_neighbor( EAST );
                }
            else
            if( y_difference == 1 )
                {
                return get_neighbor( NORTH_EAST );
                }
            }
        }

    // Lookup global sector map.       

    Remote<Layer> layer = address.layer.convert();
    return layer->optimize_replica( 20.0 ).sync_get_sector( address );
    }


Remote<Sector> Sector::get_sector_from_same_layer
    (
    const SVector2 & layer_position
    ) const
    {
    return get_sector
        ( SectorAddress( client_sector->sector_address.layer, layer_position ) );
    }
    

bool Sector::is_in_this_sector
    (
    const SVector2 & layer_position
    ) const
    {
    if( layer_position.x() < sector_origin.x() ||
        layer_position.y() < sector_origin.y() )
        {
        return false;
        }

    if( layer_position.x() >= sector_origin.x() + ClientSector::SECTOR_X_SIZE ||
        layer_position.y() >= sector_origin.y() + ClientSector::SECTOR_Y_SIZE )
        {
        return false;
        }

    return true;
    }


void Sector::link_entity
    (
    Pointer<Entity> entity
    )
    {
    hope( entity->current_sector == null );
    
    entity_list->push_front( entity.convert() );
    
    entity->last_sector = this;
    entity->current_sector = this;
    entity->client_entity->current_sector = this->client_sector;
    entity->entity_list_enumerator = entity_list->begin();
    }


void Sector::unlink_entity
    (
    Pointer<Entity> entity
    )
    {
    hope( entity->current_sector == this );
    hope( entity->last_sector == this );
    
    entity->current_sector = null;
    entity->client_entity->current_sector = null;
    entity->entity_list_enumerator->erase();
    entity->entity_list_enumerator = null;
    }


Pointer<Enumerator> Sector::list_entities()
    {
    return entity_list->begin();
    }


void Sector::request_replication_to_client
    (
    WeakPointer< Object > client,
    const STime &         timeout
    )
    {
    // Replicate self.
    get_client_sector()->replicate_to( client, timeout );
    
    // Replicate entities.
    ObjectOnStack< List::ListEnumerator > it( entity_list );
    for( ; !it->is_end(); it->next() )
        {
        Pointer< Entity > entity = it->current().convert();
        entity->get_client_entity()->replicate_to( client, timeout );
        }
    }


Sector::TileIndex Sector::get_tile_index
    (
    int x_tile,
    int y_tile
    )
    {
    hope( x_tile >= 0 && x_tile < ClientSector::NUM_X_TILES );
    hope( y_tile >= 0 && y_tile < ClientSector::NUM_Y_TILES );

    return y_tile * ClientSector::NUM_X_TILES + x_tile;
    }


Float32 Sector::get_tile_elevation
    (
    int x_tile,
    int y_tile
    ) const
    {
    WeakPointer<Sector> sector = access_tile( x_tile, y_tile );
    TileIndex index = get_tile_index( x_tile, y_tile );

    return sector->client_sector->elevation_map[ index ] + 
        sector->client_sector->delta_elevation_map[ index ];
    }


Float32 Sector::get_elevation
    (
    const SVector2 & sector_xy
    ) const
    {
    SVector2 tile_xy;
    
    sector_xy_to_tile_xy( sector_xy, tile_xy );
    
    int     int_x_tile = static_cast<int>( FloatTraits< Float32 >::floor( tile_xy.x() ) );
    int     int_y_tile = static_cast<int>( FloatTraits< Float32 >::floor( tile_xy.y() ) );
    Float32 x_tile_fraction = tile_xy.x() - int_x_tile;
    Float32 y_tile_fraction = tile_xy.y() - int_y_tile;
    Float32 a = get_tile_elevation( int_x_tile, int_y_tile );
    Float32 b = get_tile_elevation( int_x_tile + 1, int_y_tile );
    Float32 c = get_tile_elevation( int_x_tile, int_y_tile + 1 );
    Float32 d = get_tile_elevation( int_x_tile + 1, int_y_tile + 1 );

    a = a * ( 1 - x_tile_fraction ) + b * x_tile_fraction;
    c = c * ( 1 - x_tile_fraction ) + d * x_tile_fraction;
    a = a * ( 1 - y_tile_fraction ) + c * y_tile_fraction;

    return a;
    }


namespace {

/**
 * Applies @c operation on @c elevation and @c new_elevation.
 * Writes result to @c elevation.
 */
void apply_operation
    (
    ClientSector::Height &          elevation,
    Sector::ElevationMapOperation   operation,
    Float32                         new_elevation,
    Float32                         min_elevation,
    Float32                         max_elevation
    )
    {
    switch( operation )
        {
        case ClientSector::OPERATION_SET:
            elevation = new_elevation;
            break;
        case ClientSector::OPERATION_ADD:
            elevation += new_elevation;
            break;
        case ClientSector::OPERATION_MULTIPLY:
            elevation *= new_elevation;
            break;
        }
    
    elevation = clamp< Float32 >( elevation, min_elevation, max_elevation );
    }

}


bool Sector::modify_elevation
    (
    const SVector2 &        center,
    ElevationMapKind        map_kind,
    ElevationMapFunction    function,
    ElevationMapOperation   operation,
    Float32                 radius,
    Float32                 strength,
    bool                    current_sector_only
    )
    {
    // Initiate distributed operation if requested.

    if( !current_sector_only )
        {
        Pointer<ModifyElevationTask> task =
            CreateObject<ModifyElevationTask>
                (
                this,
                center,
                map_kind,
                function,
                operation,
                radius,
                strength
                );

        return false;
        }

    // Pin @c client_sector.
    
    LocalProxy<ClientSector> sector = client_sector.dereference();

    // Operate on current sector.
    
    bool    modified = false;
    int     y_tile;
    for( y_tile = 0; y_tile < ClientSector::NUM_Y_TILES; y_tile++ )
        {
        int x_tile;
        for( x_tile = 0; x_tile < ClientSector::NUM_X_TILES; x_tile++ )
            {
            SVector2    layer_xy;
            Float32     distance;
            Float32     elevation;
            TileIndex   index;

            // Compute tile distance from @c center.

            tile_xy_to_layer_xy
                (
                SVector2
                    (
                    static_cast<Float32>( x_tile ),
                    static_cast<Float32>( y_tile )
                    ),
                layer_xy
                );

            distance = ( layer_xy - center ).magnitude();

            // Evaluate @c function.

            bool apply = false;
            switch( function )
                {
                case ClientSector::FUNCTION_CIRCLE:
                    {
                    // Modify vertices inside circle of given radius. The +0.01 hack
                    // is there to make integral radius values behave nicely if snap
                    // to grid is on. Radius 0 = single vertex, radius 1 = central
                    // vertex + its neighbors, etc.

                    if( distance <= radius + 0.01f )
                        {
                        elevation = strength;
                        apply = true;
                        }
                    break;
                    }

                case ClientSector::FUNCTION_FUZZY_CIRCLE:
                    {
                    if( radius > 0 && distance <= radius + 0.01f ) // See above.
                        {
                        const Float32 dist_factor = 1 - ( distance * distance ) / ( radius * radius );
                        if( dist_factor > 0.0f )
                            {
                            elevation = strength * FloatTraits< Float32 >::sqrt( dist_factor );
                            apply = true;
                            }
                        }
                    break;
                    }

                case ClientSector::FUNCTION_GAUSSIAN:
                    {
                    // Gaussian is applied in circle of radius = 3 * deviation.

                    if( radius >= FloatTraits< Float32 >::epsilon() && distance <= 3.0f * radius )
                        {
                        elevation = FloatTraits<Float32>::exp( -distance / ( 2 * radius * radius ) );
                        elevation *= strength;
                        apply = true;
                        }
                    break;
                    }
                }

            if( apply )
                {
                // Modify tile elevation.

                index = get_tile_index( x_tile, y_tile );

                switch( map_kind )
                    {
                    case ClientSector::KIND_ELEVATION_MAP:
                        {
                        apply_operation
                            (
                            sector->elevation_map( index ),
                            operation,
                            elevation,
                            ClientSector::min_height,
                            ClientSector::max_height
                            );

                        break;
                        }

                    case ClientSector::KIND_DELTA_ELEVATION_MAP:
                        {
                        apply_operation
                            (
                            sector->delta_elevation_map( index ),
                            operation,
                            elevation,
                            -max_delta_map_magnitude,
                            max_delta_map_magnitude
                            );

                        break;
                        }
                    }

                modified = true;
                }
            }
        }

    // Done.

    if( map_kind == ClientSector::KIND_DELTA_ELEVATION_MAP && modified )
        {
        delta_map_dirty = true;
        }

    return modified;
    }


bool Sector::modify_material
    (
    const SVector2 &            center,
    Float32                     radius,
    ClientSector::MaterialId    material_id,
    bool                        current_sector_only
    )
    {
    // Initiate distributed operation if requested.

    if( !current_sector_only )
        {
        Pointer< ModifyMaterialTask > task =
            CreateObject< ModifyMaterialTask >
                (
                this,
                center,
                radius,
                material_id
                );

        return false;
        }

    // Pin @c client_sector.
    
    LocalProxy< ClientSector > sector = client_sector.dereference();

    // Operate on current sector.
    
    int     y_tile;
    bool    modified = false;
    for( y_tile = 0; y_tile < ClientSector::NUM_Y_TILES; y_tile++ )
        {
        int x_tile;
        for( x_tile = 0; x_tile < ClientSector::NUM_X_TILES; x_tile++ )
            {
            SVector2    layer_xy;
            Float32     distance;
            TileIndex   index;

            // Compute tile distance from @c center.

            tile_xy_to_layer_xy
                (
                SVector2
                    (
                    static_cast<Float32>( x_tile ),
                    static_cast<Float32>( y_tile )
                    ),
                layer_xy
                );

            distance = ( layer_xy - center ).magnitude();
            
            if( distance <= radius + 0.01f ) // see comment in modify_elevation().
                {
                // Modify material.
                index = get_tile_index( x_tile, y_tile );
                sector->material_map( index ) = material_id;
                modified = true;
                }
            }
        }

    return modified;
    }


Float32 Sector::get_tile_slope
    (
    int x_tile,
    int y_tile
    ) const
    {
    // Compute normal of [ x_tile, y_tile ] vertex.
    // Warning! The normal is computed in the SOUTH_WEST corner of the tile.

    SVector3 normal = SVector3
        (
        get_tile_elevation( x_tile - 1, y_tile ) - 
        get_tile_elevation( x_tile + 1, y_tile ),
        get_tile_elevation( x_tile, y_tile - 1 ) - 
        get_tile_elevation( x_tile, y_tile + 1 ),
        2.0f
        ).normalize();

    // Compute slope.

    return clamp( 1.0f - normal.z(), 0.0f, 1.0f );
    }
    

ClientSector::MaterialId Sector::get_tile_material
    (
    int x_tile,
    int y_tile
    ) const
    {
    WeakPointer<Sector> sector = access_tile( x_tile, y_tile );
    TileIndex index = get_tile_index( x_tile, y_tile );
    
    return sector->client_sector->material_map[ index ];
    }


bool Sector::reserve_tiles
    (
    const EntityPosition &  position,
    const PhysicalInfo &    physical_info,
    int                     entity_flags,
    Remote<Entity>          entity,
    CollisionInfo &         collision_info
    )
    {
    return reserve_or_release_tiles
        ( position, physical_info, entity_flags, entity, true, collision_info );
    }


void Sector::release_tiles
    (
    const EntityPosition &  position,
    const PhysicalInfo &    physical_info,
    int                     entity_flags,
    Remote<Entity>          entity
    )
    {
    CollisionInfo collision_info;

    reserve_or_release_tiles
        ( position, physical_info, entity_flags, entity, false, collision_info );
    }


namespace {    

/**
 * Rasterization call-back.
 * Either reserves or releases the tiles.
 */
class ReserveOrReleaseTiles : public ConvexArea::RasterizerCallback
    {
public:
        
    /** Constructor. */
    ReserveOrReleaseTiles
        (
        Sector &                sector,
        const PhysicalInfo &    physical_info,
        int                     entity_flags,
        Remote<Entity>          entity,
        bool                    reserve,
        CollisionInfo &         collision_info
        ) : sector( sector ),
            physical_info( physical_info ),
            entity_flags( entity_flags ),
            entity( entity ),
            reserve( reserve ),
            collision_info( collision_info )
        {
        reservation_status.failed = false;
        }

    /** Rasterizer callback. */
    virtual bool operator()
        (
        Int32   begin_x,
        Int32   end_x,
        Int32   y
        ) const
        {
        if( reserve )
            {
            // Reserve tiles.

            if( !reservation_status.failed )
                {
                // Normal mode (first rasterization pass):
                // Try to reserve all tiles. All tiles have to be reserved
                // or none of them.
                // If a tile can not be reserved all the previous reservations
                // have to be cancelled. This is done by the second
                // rasterization pass.

                while( begin_x < end_x )
                    {
                    // Try to reserve [ begin_x, y ] tile.

                    if( !sector.reserve_tile( begin_x, y, physical_info, 
                        entity_flags, entity, collision_info ) )
                        {
                        // The reservation failed.
                        
                        reservation_status.failed = true;

                        // Remember the tile so that the previous reservations
                        // could be undoed by the second rasterization pass.
                        
                        reservation_status.failing_x_tile = begin_x;
                        reservation_status.failing_y_tile = y;

                        // Finish the rasterization pass.

                        return false;
                        }

                    begin_x++;
                    }
                }
            else
                {
                // Undo mode (second rasterization pass):
                // Release all tiles that were reserved in the first
                // rasterization pass.

                while( begin_x < end_x )
                    {
                    if( begin_x == reservation_status.failing_x_tile && 
                        y == reservation_status.failing_y_tile )
                        {
                        // This is the tile that could not be (and was not)
                        // reserved in the first rasterization pass.
                        // All the work from the first pass has been
                        // undoed.
                        // Finish the rasterization pass.

                        return false;
                        }

                    // Release the [ begin_x, y ] tile that was reserved in
                    // the first rasterization pass.

                    sector.release_tile( begin_x, y, entity );

                    begin_x++;
                    }
                }
            }
        else
            {
            // Releases tiles.

            while( begin_x < end_x )
                {
                // Release the [ begin_x, y ] tile.

                sector.release_tile( begin_x, y, entity );

                begin_x++;
                }
            }

        return true;
        }

public:

    Sector &                sector;
        /**< Sector being processed. */
    const PhysicalInfo &    physical_info;
        /**< Entity physical info. */
    int                     entity_flags;
        /**< Entity class flags. */
    Remote<Entity>          entity;
        /**< Entity under discussion. */
    bool                    reserve;
        /**< Reserve or release? */
    CollisionInfo &         collision_info;
        /**< Collision info. */
    mutable struct
        {
        bool    failed;
            /**< Set to true if the reservation failed. */
        Int32   failing_x_tile;
            /**< X coordinate of the tile that could not be reserved. */
        Int32   failing_y_tile;
            /**< Y coordinate of the tile that could not be reserved. */
        } reservation_status;
            /**< Reservation status. */

    }; // ReserveOrReleaseTiles

}    


bool Sector::reserve_or_release_tiles
    (
    const EntityPosition &  position,
    const PhysicalInfo &    physical_info,
    int                     entity_flags,
    Remote<Entity>          entity,
    bool                    reserve,
    CollisionInfo &         collision_info
    )
    {
    // Get entity's bounding box.

    ConvexArea area = get_bounding_box( position, physical_info );

    // Process tiles in the current sector.

    ReserveOrReleaseTiles callback
        (
        *this,
        physical_info,
        entity_flags,
        entity,
        reserve,
        collision_info
        );
    
    area.rasterize( callback, 0, 0, ClientSector::NUM_X_TILES, ClientSector::NUM_Y_TILES );

    if( callback.reservation_status.failed )
        {
        // The reservation failed.
        // Undo the previous work. This is simply done by starting a new
        // rasterization pass that automatically undoes effect of the first
        // (incomplete) pass.

        area.rasterize( callback, 0, 0, ClientSector::NUM_X_TILES, ClientSector::NUM_Y_TILES );
        
        // Done.

        return false;
        }

    // Process tiles in the neighboring sectors.
    // Process the sectors if this is the "originator" sector only.
    // Since this processes sectors neighboring to the sector containing
    // @c origin only, the rasterized bounding box must not be larger than a
    // single sector.

    if( !get_sector_area().contains( area ) &&
        is_in_this_sector( position.origin.read() ) )
        {
        bool        sector_reserved[ 8 ];
        std::size_t i;

        // Process neighbors.

        for( i = 0; i < 8; i++ )
            {
            Direction   direction = static_cast<Direction>( i );
            ConvexArea  sector_area = get_neighbor_area( direction );

            sector_reserved[ i ] = false;

            if( area.intersects( sector_area ) )
                {
                Remote<Sector> sector = get_neighbor( direction );

                // Process tiles in @c sector.

                if( reserve )
                    {
                    // Reserve tiles in @c sector.

                    if( sector == null || !sector->sync_reserve_tiles( position, 
                        physical_info, entity_flags, entity, collision_info ) )
                        {
                        // Failed to reserve tiles in that sector.
                        // Undo previous work.

                        // Undo work in the current sector.
                        
                        ReserveOrReleaseTiles release
                            (
                            *this,
                            physical_info,
                            entity_flags,
                            entity,
                            false,
                            collision_info
                            );

                        area.rasterize
                            (
                            release,
                            0,
                            0,
                            ClientSector::NUM_X_TILES,
                            ClientSector::NUM_Y_TILES
                            );

                        // Undo work in the neighboring sectors processed so far.

                        std::size_t j;

                        for( j = 0; j < i; j++ )
                            {
                            if( sector_reserved[ j ] )
                                {
                                sector = get_neighbor( static_cast<Direction>( j ) );

                                if( sector == null )
                                    {
                                    // This may happen only if @c sector was detached
                                    // from the current sector meanwhile.

                                    continue;
                                    }

                                sector->sync_release_tiles
                                    ( position, physical_info, entity_flags, entity );
                                }
                            }

                        // Done.

                        return false;
                        }

                    sector_reserved[ i ] = true;
                    }
                else
                    {
                    // Release tiles in @c sector.

                    if( sector != null )
                        {
                        sector->sync_release_tiles
                            ( position, physical_info, entity_flags, entity );
                        }
                    }
                }
            }
        }

    // Done.

    return true;
    }


bool Sector::can_reserve_tile
    (
    int                     x_tile,
    int                     y_tile,
    const PhysicalInfo &    physical_info,
    Remote<Entity>          entity,
    CollisionInfo &         collision_info
    )
    {
    TileIndex                   index = get_tile_index( x_tile, y_tile );
    ReservationMap::KeyRange    range = reservation_map.equal_range( index );
    ReservationMap::Iterator    it = range.first;

    // Check tile material.

    switch( get_tile_material( x_tile, y_tile ) )
        {
        case ClientSector::MATERIAL_OBSTACLE:
            {
            // Tile can not be reserved by any entity.
            
            collision_info.reason = CollisionInfo::REASON_OBSTACLE;

            // Log.

            #ifdef LOG_FAILED_RESERVATIONS

            std::ostringstream  ss;
            
            ss  << entity << " can not reserve [ " << x_tile << ", " << y_tile
                << " @ " << get_object_id() << " ] because the tile is MATERIAL_OBSTACLE.";

            Global::log_info
                (
                Status::FACILITY_LIB,
                Status::PRIORITY_NORMAL,
                ss.str()
                );

            #endif

            return false;
            }
        }

    // Check tile slope.

    if( !physical_info.can_handle_slope( get_tile_slope( x_tile, y_tile ) ) )
        {
        // The tile is too steep.

        collision_info.reason = CollisionInfo::REASON_SLOPE_TOO_HIGH;
        
        // Log.

        #ifdef LOG_FAILED_RESERVATIONS

        std::ostringstream ss;
        
        ss  << entity << " can not reserve [ " << x_tile << ", " << y_tile
            << " @ " << get_object_id() << " ] because the tile is too steep (slope = " 
            << get_tile_slope( x_tile, y_tile ) << ", max_slope = " 
            << physical_info.max_slope << ").";

        Global::log_info
            (
            Status::FACILITY_LIB,
            Status::PRIORITY_NORMAL,
            ss.str()
            );

        #endif

        return false;
        }

    // Check current reservations for tile.

    while( it != range.second )
        {
        const TileReservation & reservation = it->second;

        if( reservation.entity != entity )
            {
            if( physical_info.collides_with( reservation.physical_class ) )
                {
                // Tile can not be reserved by @c entity because
                // it collides with an existing reservation.

                collision_info.reason = CollisionInfo::REASON_COLLISION;
                collision_info.entity = reservation.entity;
                collision_info.entity_flags = reservation.entity_flags;

                // Log.
                
                #ifdef LOG_FAILED_RESERVATIONS

                std::ostringstream ss;
                
                ss  << entity << " can not reserve [ " << x_tile << ", " << y_tile
                    << " @ " << get_object_id()
                    << " ] because the tile is already reserved by a colliding "
                    << collision_info.entity << ".";

                Global::log_info
                    (
                    Status::FACILITY_LIB,
                    Status::PRIORITY_NORMAL,
                    ss.str()
                    );

                #endif

                return false;
                }
            }
        else
            {
            // If the tile has already been reserved by @c entity, the
            // tile can be reserved for the second time too.
            
            break;
            }

        ++it;
        }
    
    // Tile can be reserved.

    return true;
    }


bool Sector::reserve_tile
    (
    int                     x_tile,
    int                     y_tile,
    const PhysicalInfo &    physical_info,
    int                     entity_flags,
    Remote<Entity>          entity,
    CollisionInfo &         collision_info
    )
    {
    if( !can_reserve_tile( x_tile, y_tile, physical_info, entity, collision_info ) )
        {
        return false;
        }
    else
        {
        TileIndex index = get_tile_index( x_tile, y_tile );

        reservation_map.insert
            (
            index,
            TileReservation
                (
                entity,
                physical_info.physical_class,
                entity_flags
                )
            );

#ifdef MARK_RESERVED_TILES
        client_sector->tile_flag_map( index ) |= ClientSector::TILE_MARK;
#endif
        
        return true;
        }
    }


void Sector::release_tile
    (
    int             x_tile,
    int             y_tile,
    Remote<Entity>  entity
    )
    {
    TileIndex                   index = get_tile_index( x_tile, y_tile );
    ReservationMap::KeyRange    range = reservation_map.equal_range( index );
    ReservationMap::Iterator    it = range.first;
    
    while( it != range.second )
        {
        const TileReservation & reservation = it->second;

        if( reservation.entity == entity )
            {
            reservation_map.erase( it );
            
            // If the tile was reserved by @c entity multiple times,
            // the first reservation is released only.

#ifdef MARK_RESERVED_TILES
            if( reservation_map.find( index ) == reservation_map.end() )
                {
                client_sector->tile_flag_map( index ) &= ~ClientSector::TILE_MARK;
                }
#endif

            return;
            }
        
        ++it;
        }

    Global::log_warning
        (
        Status::FACILITY_LIB,
        Status::PRIORITY_HIGHEST,
        "Trying to release a tile that was not reserved. Either rasterization "
        "bug, tile reservation bug or floating point serialization inaccuracy. "
        "Should not happen."
        );
    }


void Sector::enumerate_entities
    (
    int                 x_tile,
    int                 y_tile,
    int                 entity_flags,
    EntityList &        list
    )
    {
    TileIndex                   index = get_tile_index( x_tile, y_tile );
    ReservationMap::KeyRange    range = reservation_map.equal_range( index );
    ReservationMap::Iterator    it = range.first;
    
    list.clear();

    while( it != range.second )
        {
        const TileReservation & reservation = it->second;
        
        if( flags_on( reservation.entity_flags, entity_flags ) )
            {
            list.push_back( reservation.entity );
            }
        
        ++it;
        }
    }


ConvexArea Sector::get_bounding_box
    (
    const EntityPosition &  position,
    const PhysicalInfo &    physical_info
    ) const
    {
    SVector2 origin_tile_xy;
    
    layer_xy_to_tile_xy( position.origin.read(), origin_tile_xy );
    
    // Entity's bounding box must not be larger than a single sector for
    // various reasons. That is mostly because of tile reservation or
    // entity enumeration (spatial queries).

    hope( physical_info.bounding_box_size.x() < ClientSector::NUM_X_TILES );
    hope( physical_info.bounding_box_size.y() < ClientSector::NUM_Y_TILES );
    
    ConvexArea area
        (
        origin_tile_xy,
        physical_info.bounding_box_size.read(),
        physical_info.bounding_box_axis_aligned
        );
    
    if( physical_info.bounding_box_axis_aligned == false )
        {
        area.rotate( origin_tile_xy, position.orientation.heading() );
        }

    return area;
    }


ConvexArea Sector::get_bounding_box
    (
    WeakPointer<Entity> entity
    ) const
    {
    return get_bounding_box( entity->get_current_position(), entity->physical_info );
    }


ConvexArea Sector::get_sector_area() const
    {
    return get_sector_area( client_sector->sector_address );
    }


ConvexArea Sector::get_sector_area
    (
    const SectorAddress & address
    ) const
    {
    ConvexArea  area;
    SVector2    origin
        (
        static_cast<Float32>( ClientSector::NUM_X_TILES * ( address.column - client_sector->sector_address.column ) ),
        static_cast<Float32>( ClientSector::NUM_Y_TILES * ( address.row - client_sector->sector_address.row ) )
        );
    
    area.add_vertex( origin );
    area.add_vertex( origin + SVector2( ClientSector::NUM_X_TILES, 0 ) );
    area.add_vertex( origin + SVector2( ClientSector::NUM_X_TILES, ClientSector::NUM_Y_TILES ) );
    area.add_vertex( origin + SVector2( 0, ClientSector::NUM_Y_TILES ) );
    area.update_edges();

    return area;
    }


ConvexArea Sector::get_neighbor_area
    (
    Direction neighbor
    ) const
    {
    return get_sector_area( client_sector->sector_address.get_neighbor( neighbor ) );
    }


void Sector::layer_xy_to_sector_xy
    (
    const SVector2 &    layer_xy,
    SVector2 &          sector_xy
    ) const
    {
    sector_xy.set_x( layer_xy.x() - sector_origin.x() );
    sector_xy.set_y( layer_xy.y() - sector_origin.y() );
    }


void Sector::sector_xy_to_layer_xy
    (
    const SVector2 &    sector_xy,
    SVector2 &          layer_xy
    ) const
    {
    layer_xy.set_x( sector_xy.x() + sector_origin.x() );
    layer_xy.set_y( sector_xy.y() + sector_origin.y() );
    }


void Sector::sector_xy_to_tile_xy
    (
    const SVector2 &    sector_xy,
    SVector2 &          tile_xy
    )
    {
    tile_xy.set_x( sector_xy.x() * ClientSector::NUM_X_TILES / ClientSector::SECTOR_X_SIZE );
    tile_xy.set_y( sector_xy.y() * ClientSector::NUM_Y_TILES / ClientSector::SECTOR_Y_SIZE );
    }


void Sector::tile_xy_to_sector_xy
    (
    const SVector2 &    tile_xy,
    SVector2 &          sector_xy
    )
    {
    sector_xy.set_x( tile_xy.x() * ClientSector::SECTOR_X_SIZE / ClientSector::NUM_X_TILES );
    sector_xy.set_y( tile_xy.y() * ClientSector::SECTOR_Y_SIZE / ClientSector::NUM_Y_TILES );
    }


void Sector::tile_xy_to_layer_xy
    (
    const SVector2 &    tile_xy,
    SVector2 &          layer_xy
    ) const
    {
    layer_xy.set_x
        ( tile_xy.x() * ClientSector::SECTOR_X_SIZE / ClientSector::NUM_X_TILES + sector_origin.x() );
    layer_xy.set_y
        ( tile_xy.y() * ClientSector::SECTOR_Y_SIZE / ClientSector::NUM_Y_TILES + sector_origin.y() );
    }


void Sector::layer_xy_to_tile_xy
    (
    const SVector2 &    layer_xy,
    SVector2 &          tile_xy
    ) const
    {
    tile_xy.set_x
        ( ( layer_xy.x() - sector_origin.x() ) * ClientSector::NUM_X_TILES / ClientSector::SECTOR_X_SIZE );
    tile_xy.set_y
        ( ( layer_xy.y() - sector_origin.y() ) * ClientSector::NUM_Y_TILES / ClientSector::SECTOR_Y_SIZE );
    }


void Sector::single_link_neighbor
    (
    Direction           neighbor,
    WeakPointer<Sector> sector
    )
    {
    neighbors[ neighbor ] = sector;
    }


void Sector::double_link_neighbor
    (
    Direction           neighbor,
    WeakPointer<Sector> sector
    )
    {
    Remote<Sector> s = sector.convert();

    single_link_neighbor( neighbor, sector );
    s->sync_single_link_neighbor( opposite_direction( neighbor ), this );
    }


void Sector::update_delta_elevation_map()
    {
    std::size_t index;
    bool        is_zeroed = true;

    // Flatten @c delta_elevation_map.

    if( delta_map_dirty )
        {
        hope( delta_map_decay_factor < 1.0f );

        // Optimization:
        // Use @c sector proxy so that @c client_sector could be dereferenced
        // once only.

        LocalProxy<ClientSector> sector = client_sector.dereference();

        for( index = 0; index < sector->delta_elevation_map.size(); index++ )
            {
            Float32 elevation = sector->delta_elevation_map[ index ];
            
            elevation *= delta_map_decay_factor;

            if( !FloatTraits<Float32>::epsilon_equals( elevation, 0.0f ) )
                {
                is_zeroed = false;
                }
            else
                {
                elevation = 0.0f;
                }

            sector->delta_elevation_map( index ) = elevation;
            }

        delta_map_dirty = !is_zeroed;
        }

    // Schedule next update.

    Remote<Sector>  self = this;
    STime           delay;

    if( delta_map_dirty )
        {
        // Stoupik: gcc3.2 says it's ambiguous without the read()
        delay = delta_map_decay_period.read();
        }
    else
        {
        delay = STime( 60.0f );
        }
    
    self->param( RPC_DELAYED, delay ).async_update_delta_elevation_map();
    }


WeakPointer<Sector> Sector::access_tile
    (
    int & x_tile,
    int & y_tile
    ) const
    {
    WeakPointer<Sector> current_sector = this;
    hope( is_consistent() );

    do
        {
        Direction neighbor;

        // Find out if [ x_tile, y_tile ] points to a neighboring sector
        // and recompute the coordinates relative to that sector.

        if( x_tile < 0 && y_tile < 0 )
            {
            neighbor = SOUTH_WEST;
            x_tile += ClientSector::NUM_X_TILES;
            y_tile += ClientSector::NUM_Y_TILES;
            }
        else
        if( x_tile < 0 && y_tile < ClientSector::NUM_Y_TILES )
            {
            neighbor = WEST;
            x_tile += ClientSector::NUM_X_TILES;
            }
        else
        if( x_tile < 0 && y_tile >= ClientSector::NUM_Y_TILES )
            {
            neighbor = NORTH_WEST;
            x_tile += ClientSector::NUM_X_TILES;
            y_tile -= ClientSector::NUM_Y_TILES;
            }
        else
        if( x_tile < ClientSector::NUM_X_TILES && y_tile < 0 )
            {
            neighbor = SOUTH;
            y_tile += ClientSector::NUM_Y_TILES;
            }
        else
        if( x_tile < ClientSector::NUM_X_TILES && y_tile >= ClientSector::NUM_Y_TILES )
            {
            neighbor = NORTH;
            y_tile -= ClientSector::NUM_Y_TILES;
            }
        else
        if( x_tile >= ClientSector::NUM_X_TILES && y_tile < 0 )
            {
            neighbor = SOUTH_EAST;
            x_tile -= ClientSector::NUM_X_TILES;
            y_tile += ClientSector::NUM_Y_TILES;
            }
        else
        if( x_tile >= ClientSector::NUM_X_TILES && y_tile < ClientSector::NUM_Y_TILES )
            {
            neighbor = EAST;
            x_tile -= ClientSector::NUM_X_TILES;
            }
        else
        if( x_tile >= ClientSector::NUM_X_TILES && y_tile >= ClientSector::NUM_Y_TILES )
            {
            neighbor = NORTH_EAST;
            x_tile -= ClientSector::NUM_X_TILES;
            y_tile -= ClientSector::NUM_Y_TILES;
            }
        else
            {
            // [ x_tile, y_tile ] points to @c current_sector.

            break;
            }

        WeakPointer<Sector> sector = current_sector->neighbors[ neighbor ];

        if( !sector.is_local() || !sector->is_consistent() )
            {
            // [ x_tile, y_tile ] points to a neighboring sector which is not
            // local or is not consistent. Request its replication and return
            // closest tile from the current sector.
            
            switch( neighbor )
                {
                case EAST:
                    x_tile = ClientSector::NUM_X_TILES - 1;
                    break;
                case NORTH_EAST:
                    x_tile = ClientSector::NUM_X_TILES - 1;
                    y_tile = ClientSector::NUM_Y_TILES - 1;
                    break;
                case NORTH:
                    y_tile = ClientSector::NUM_Y_TILES - 1;
                    break;
                case NORTH_WEST:
                    x_tile = 0;
                    y_tile = ClientSector::NUM_Y_TILES - 1;
                    break;
                case WEST:
                    x_tile = 0;
                    break;
                case SOUTH_WEST:
                    x_tile = 0;
                    y_tile = 0;
                    break;
                case SOUTH:
                    y_tile = 0;
                    break;
                case SOUTH_EAST:
                    x_tile = ClientSector::NUM_X_TILES - 1;
                    y_tile =  0;
                    break;
                }
            
            if( sector != null )
                {
                Remote<Sector> ptr = sector;
                ptr->async_replicate_to
                    (
                    Massiv::System::get_local_node_id(), 
                    STime::current_time() + 20.0
                    );
                }

            break;
            }

        // [ x_tile, y_tile ] was expressed relative to the local neighboring
        // sector. Move to that sector.

        current_sector = sector;
        }
    while( true );

    // Done.

    hope( current_sector.is_local() && current_sector->is_consistent() );
    return current_sector;
    }


} // namespace Demo::Lib
} // namespace Demo
