/** @file sector.h
 * Sector.
 */

/*
 * begin        : Mon Jun 31 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_SECTOR_H
#define DEMO_SECTOR_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_CONVEX_AREA_H
#include "lib/server/convex_area.h"
#endif

#ifndef DEMO_LIST_CONTAINER_H
#include "lib/shared/list_container.h"
#endif

#ifndef DEMO_DIRECTION_H
#include "lib/shared/direction.h"
#endif

#ifndef DEMO_SECTOR_ADDRESS_H
#include "lib/shared/sector_address.h"
#endif

#ifndef DEMO_PHYSICAL_INFO_H
#include "lib/server/physical_info.h"
#endif

#ifndef DEMO_COLLISION_INFO_H
#include "lib/server/collision_info.h"
#endif

#ifndef DEMO_ENTITY_H
#include "lib/server/entity.h"
#endif

#ifndef DEMO_CLIENT_SECTOR_H
#include "lib/shared/client_sector.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();

BASIC_LIB_EXCEPTION( SectorErrorException, ::Massiv::Core::Lib::RuntimeException );


/**
 * World area of constant volume, single migration group of objects.
 * Manages list of entities that geographically belong (their origins)
 * to that sector, characterizes the area.
 * The sector area is split into small rectangular tiles that form ground
 * level of that area. Each tile holds an entry for sector elevation
 * map (sampled in the south-west corner of the tile) and material map.
 */
class Sector : public Object
    {
public:

    MASSIV_OBJECT( Sector );

    friend class Layer;
    friend class Entity;
    friend class SpatialQuery;
            
    typedef ClientSector::ElevationMapKind      ElevationMapKind;
    typedef ClientSector::ElevationMapOperation ElevationMapOperation;
    typedef ClientSector::ElevationMapFunction  ElevationMapFunction;

    typedef PArray< PWeakPointer<Sector> >  SectorList;
    typedef PArray< PWeakPointer<Entity> >  EntityList;
    
    typedef int     TileIndex;
    typedef PInt32  PTileIndex;

    /**
     * Our entity "physics" is tile-based like in Dune 2.
     * Entities reserve tiles they want to move to. If a tile, that need
     * be reserved, is already reserved by another entity that collides
     * with the current entity, the reservation fails and the entity
     * will not move.
     * TileReservation holds information about what entity reserved
     * the tile and its physical info.
     * @note Tile can be reserved by multiple entities if they do not
     *       collide. An entity never collides with self, which means
     *       that tile can be reserved by the same entity multiple times.
     */
    class TileReservation : public ValueType
        {
    public:

        MASSIV_OBJECT( TileReservation );

        /** Default constructor. */
        TileReservation();

        /** Constructor. */
        TileReservation
            (
            Remote<Entity>              entity,
            PhysicalInfo::PhysicalClass physical_class,
            int                         entity_flags
            );

        /** Copy constructor. */
        TileReservation
            (
            const TileReservation & reservation
            );

    public:

        PRemote<Entity>                     entity;
            /**< The entity that reserved the tile. */
        PEnum<PhysicalInfo::PhysicalClass>  physical_class;
            /**< Physical class of @c entity. */
        PVlInt32                            entity_flags;
            /**< Entity class flags. */
        }; // TileReservation

    typedef PMultiDictionary<PTileIndex, TileReservation> ReservationMap;

public:

    /** Initializes default sector. */
    void initialize();
    
public:

    //
    // Getters.
    //

    /**
     * Get sector address.
     */
    SectorAddress get_address() const;
    
    /**
     * Get sector origin.
     */
    SVector2 get_origin() const;

    /**
     * Get client sector object associated with this sector.
     */
    Pointer<ClientSector> get_client_sector() const;

protected:

    /**
     * Set address of the sector.
     * Used by #Layer when registering new sector to it.
     */
    void set_address
        (
        const SectorAddress & sector_address
        );
    
public:

    //
    // Neighboring sectors management.
    //
    
    /** Returns opposite direction. */
    static Direction opposite_direction
        (
        Direction direction
        );

    /** Returns pointer to the neighboring sector. */
    Remote<Sector> get_neighbor
        (
        Direction neighbor
        ) const;
    
    /**
     * Returns sector identified by @c address.
     * If @c address does not refer to a neighboring sector global
     * sector map is searched.
     */
    Remote<Sector> get_sector
        (
        const SectorAddress & address
        ) const;
    
    /**
     * Returns sector from this sector's layer the @c layer_position belongs to.
     * If @c layer_position does not refer to a neighboring sector global
     * sector map is searched.
     */
    Remote<Sector> get_sector_from_same_layer
        (
        const SVector2 & layer_position
        ) const;
    
    /** Tests if @c position geographically belongs to this sector. */
    bool is_in_this_sector
        (
        const SVector2 & layer_position
        ) const;

protected:

    //
    // Linked entities management.
    //

    /**
     * Links the entity to this sector (regardless of entity's
     * real position).
     * The entity will be owned by this sector and will belong to its
     * migration group.
     * Assumes that entity is currently unlinked from all sectors.
     */
    void link_entity
        (
        Pointer<Entity> entity
        );

    /**
     * Unlinks the entity from this sector.
     * Assumes that the entity is linked to this sector.
     */
    void unlink_entity
        (
        Pointer<Entity> entity
        );

public:

    /**
     * Returns enumerator pointing to the list of owned entities.
     */
    Pointer<Enumerator> list_entities();

public:

    //
    // Replication.
    //
    
    /**
     * Create replication request for the sector and all
     * of its entities to @a client node.
     * Replication is requested for all client visible
     * parts of the objects (#ClientSector for the sector
     * and #ClientEntity for the entities).
     */
    void request_replication_to_client
        (
        WeakPointer< Object > client,
        const STime &         timeout
        );

public:

    //
    // Tile maps.
    //

    /** Returns index into tile maps (elevation map, ...). */
    static TileIndex get_tile_index
        (
        int x_tile,
        int y_tile
        );

    /**
     * Returns elevation for [ @c x_tile, @c y_tile ] (south-west corner).
     * Tile coordinates may span sector boundaries.
     * In that case sector replica is accessed to obtain the result.
     * If replica is not available an estimated value is returned.
     */
    Float32 get_tile_elevation
        (
        int x_tile,
        int y_tile
        ) const;

    /**
     * Returns elevation for @c sector_xy coordinates.
     * Uses bilinear interpolation to retrieve result from the discrete
     * elevation map.
     */
    Float32 get_elevation
        (
        const SVector2 & sector_xy
        ) const;

    /**
     * Modifies elevation map specified by @c map_kind parameter
     * in @c center's neighborhood. @c function is applied to tiles
     * in the processed neighborhood.
     * @c radius and @c strength parametrize @c function properties.
     * @c radius affects size of the modified neighborhood, @c strength
     * determines how "much" the map is modified.
     * The actual neighborhood size (correspondence with the actual
     * neighborhood size and @c radius parameter) and how the map is modified
     * is @c function-specific.
     * @c operation defines how @c function result is combined with
     * the current elevation.
     * @c center is defined in layer coordinates.
     * If @c current_sector_only is true, then current sector's elevation
     * map is modified only.
     * Returns true if @c function was evaluated to non-zero in at least
     * one tile and @c current_sector_only was true.
     * @note If @c current_sector_only is @c false then the operation
     * is executed asynchronously with no synchronization with another
     * elevation operations.
     */
    bool modify_elevation
        (
        const SVector2 &        center,
        ElevationMapKind        map_kind,
        ElevationMapFunction    function,
        ElevationMapOperation   operation,
        Float32                 radius,
        Float32                 strength,
        bool                    current_sector_only = false
        );
        
    /**
     * Modifies material map in @c center's neighborhood.
     * Material of all sector verticel in specified @a radius
     * are set to @a material_id.
     * @c center is defined in layer coordinates.
     *
     * If @c current_sector_only is true, then current sector's elevation
     * map is modified only.
     * @note If @c current_sector_only is @c false then the operation
     * is executed asynchronously with no synchronization with another
     * material operations.
     *
     * @return @c true if neighbor sectors should be visited, @c false if
     *         enumeration should be stopped at this sector.
     */
    bool modify_material
        (
        const SVector2 &            center,
        Float32                     radius,
        ClientSector::MaterialId    material_id,
        bool                        current_sector_only = false
        );
    
    /**
     * Compute tile slope.
     * Slope is expressed as a floating point value from [ 0, 1 ] range,
     * where 0.0f corresponds to a flat tile (parallel to ground level).
     * Tile coordinates may span sector boundaries.
     */
    Float32 get_tile_slope
        (
        int x_tile,
        int y_tile
        ) const;

    /**
     * Returns tile material for south-western vertex of [ x_tile, y_tile ]
     * tile. Tile coordinates must point into this sector.
     * Tile coordinates may span sector boundaries.
     * In that case sector replica is accessed to obtain the result.
     * If replica is not available an estimated value is returned.
     */
    ClientSector::MaterialId get_tile_material
        (
        int x_tile,
        int y_tile
        ) const;

    /**
     * Reserves tiles for @c entity at @c position according to entity's
     * @c physical_info. Once the tiles are reserved they can not be
     * reserved by another entity that would collide with the entity
     * that did the reservation.
     * Returns true if tiles were reserved. If tiles were not reserved
     * @c collision_info describes the reason why the reservation failed.
     */
    bool reserve_tiles
        (
        const EntityPosition &  position,
        const PhysicalInfo &    physical_info,
        int                     entity_flags,
        Remote<Entity>          entity,
        CollisionInfo &         collision_info
        );

    /**
     * Releases tiles that were previously reserved by @c reserve_tiles().
     */
    void release_tiles
        (
        const EntityPosition &  position,
        const PhysicalInfo &    physical_info,
        int                     entity_flags,
        Remote<Entity>          entity
        );
    
    /**
     * Reserves or releases tiles according to @c reserve parameter.
     * Returns true on success.
     * Used by @c reserve_tiles() and @c release_tiles().
     */
    bool reserve_or_release_tiles
        (
        const EntityPosition &  position,
        const PhysicalInfo &    physical_info,
        int                     entity_flags,
        Remote<Entity>          entity,
        bool                    reserve,
        CollisionInfo &         collision_info
        );

    /**
     * Tests if [ x_tile, y_tile ] could be reserved by @c entity according
     * to its @c physical_info. Fills @c collision_info if tile can not be
     * reserved.
     * Used by @c reserve_tile().
     */
    bool can_reserve_tile
        (
        int                     x_tile,
        int                     y_tile,
        const PhysicalInfo &    physical_info,
        Remote<Entity>          entity,
        CollisionInfo &         collision_info
        );

    /**
     * Reserves tile [ x_tile, y_tile ] to @c entity according to its
     * @c physical_info. Fills @c collision_info if tile can not be reserved.
     * Used by @c reserve_tiles().
     */
    bool reserve_tile
        (
        int                     x_tile,
        int                     y_tile,
        const PhysicalInfo &    physical_info,
        int                     entity_flags,
        Remote<Entity>          entity,
        CollisionInfo &         collision_info
        );

    /**
     * Releases tile [ x_tile, y_tile ] that was previously reserved
     * by @c reserve_tile(). Used by @c release_tiles().
     */
    void release_tile
        (
        int             x_tile,
        int             y_tile,
        Remote<Entity>  entity
        );

    /**
     * Enumerates entities that have reserved [ x_tile, y_tile ] tile
     * and their entity flags apply for the following formula
     * @c flags_on( entity-\>entity_flags, entity_flags ).
     * Set @c entity_flags to 0 to enumerate all entities.
     * Fills @c list (which is cleared first).
     */
    void enumerate_entities
        (
        int                 x_tile,
        int                 y_tile,
        int                 entity_flags,
        EntityList &        list
        );

public:
    
    //
    // Areas.
    //
    
    /**
     * Returns bounding box for an entity with given position and physical
     * info (bounding box sizes).
     * The area is constructed in this sector's tile coordinates.
     */
    ConvexArea get_bounding_box
        (
        const EntityPosition &  position,
        const PhysicalInfo &    physical_info
        ) const;
    
    /**
     * Returns bounding box of @c entity. The entity must be local however
     * it can be an object replica thus a weak pointer must be used.
     * The area is constructed in this sector's tile coordinates.
     */
    ConvexArea get_bounding_box
        (
        WeakPointer<Entity> entity
        ) const;
    
    /**
     * Returns this sector's area.
     * The area is constructed in this sector's tile coordinates.
     */
    ConvexArea get_sector_area() const;
    
    /**
     * Returns area of a sector with given @c address.
     * The area is constructed in this sector's tile coordinates.
     */
    ConvexArea get_sector_area
        (
        const SectorAddress & address
        ) const;
    
    /**
     * Returns a neighboring sector area.
     * The area is constructed in this sector's tile coordinates.
     */
    ConvexArea get_neighbor_area
        (
        Direction neighbor
        ) const;

public:

    //
    // Coordinate systems.
    //

    /**
     * Translates @c layer_xy coordinates (layer 2D space) to
     * @c sector_xy (sector 2D space).
     */
    void layer_xy_to_sector_xy
        (
        const SVector2 &    layer_xy,
        SVector2 &          sector_xy
        ) const;
    
    /**
     * Translates @c sector_xy (sector 2D space) to @c layer_xy
     * (layer 2D space).
     */
    void sector_xy_to_layer_xy
        (
        const SVector2 &    sector_xy,
        SVector2 &          layer_xy
        ) const;
    
    /**
     * Translates @c sector_xy coordinates into sector tile coordinates.
     */
    static void sector_xy_to_tile_xy
        (
        const SVector2 &    sector_xy,
        SVector2 &          tile_xy
        );
    
    /**
     * Translates tile sector coordinates into @c sector_xy coordinates.
     */
    static void tile_xy_to_sector_xy
        (
        const SVector2 &    tile_xy,
        SVector2 &          sector_xy
        );
    
    /**
     * Translates @c tile_xy coordinates (sector tile 2D space) to
     * @c layer_xy coordinates (layer 2D space).
     */
    void tile_xy_to_layer_xy
        (
        const SVector2 &    tile_xy,
        SVector2 &          layer_xy
        ) const;
    
    /**
     * Translates @c layer_xy coordinates (layer 2D space) to
     * sector tile coordinates.
     */
    void layer_xy_to_tile_xy
        (
        const SVector2 &    layer_xy,
        SVector2 &          tile_xy
        ) const;

//protected:
public: // FIXME: kvuli RPC

    //
    // Internal stuff.
    //
    
    /**
     * Single-links @c sector to the neighbor list.
     * Does not update other sector's lists.
     * Used by Layer when it registers/unregisters sector from the
     * world.
     */
    void single_link_neighbor
        (
        Direction           neighbor,
        WeakPointer<Sector> sector
        );
    
    /**
     * Double-links @c sector to the neighbor list.
     * Does not update other sector's lists.
     * Used by Layer when it registers/unregisters sector from the
     * world.
     * Operates synchronously.
     */
    void double_link_neighbor
        (
        Direction           neighbor,
        WeakPointer<Sector> sector
        );

    /**
     * Flattens @c delta_elevation_map.
     * Called periodically with @c delta_map_decay_period, should not be
     * invoked manually.
     */
    void update_delta_elevation_map();

private:

    /**
     * Accesses tile [ @c x_tile, @c y_tile ] that was specified in this
     * sector's tile coordinates. The tile need not point to this sector.
     * If it points to a different sector, @c x_tile and @c y_tile are
     * recomputed so that they would be expressed in that sector's tile
     * coordinates.
     * Returns pointer to the sector in which out [ @c x_tile, @c y_tile ]
     * coordinates were expressed.
     * If the sector where [ @c x_tile, @c y_tile ] points to can not be
     * accessed locally, its replication will be requested and the "closest"
     * local tile will be returned.
     */
    WeakPointer<Sector> access_tile
        (
        int & x_tile,
        int & y_tile
        ) const;
    
protected:

    PPointer<ClientSector>  client_sector;
        /**< Client-visible part of the sector. */
    SectorList              neighbors;
        /**< Neighboring sectors. Exactly one-neighborhood in
             eight-conectivity metrics in discrete 2D. */
    PVector2                sector_origin;
        /**< Position of the sector in the Layer (using layer 2D coordinates).
             Centered to [ 0, 0 ] tile. */
    PPointer<List>          entity_list;
        /**< List of linked entities that belong to this sector.
             The entities will belong to the sector migration group. */
    PFloat32                delta_map_decay_factor;
        /**< A float factor less than 1.0f. Determines how much
             @c delta_elevation_map is flattened after each update iteration. */
    PTime                   delta_map_decay_period;
        /**< Time difference between two @c delta_elevation_map updates
             if @c delta_map_dirty. */
    PFloat32                max_delta_map_magnitude;
        /**< Maximum absolute value of a value stored in @c delta_elevation_map.
             Determines how much @c delta_elevation_map may deform the
             terrain. */
    PBool                   delta_map_dirty;
        /**< True if @c delta_elevation_map may not be filled by zeroes. */
    ReservationMap          reservation_map;
        /**< Tile reservation map.
             The structure is used both for physics and spatial queries. */
    }; // Sector


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_SECTOR_H
