/** @file sector_enumerator.cpp
 * Sector enumerator implementation.
 */

/*
 * begin        : Mon Jul 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/sector_enumerator.h"
#include "lib/server/sector.h"
#include "lib/server/layer_rpc.h"

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of SectorEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void SectorEnumerator::initialize
    (
    Pointer<Sector> originator,
    Int32           max_neighborhood
    )
    {
    originator_sector = originator;
    current_sector = null;
    current_address.layer = originator->get_address().layer;
    flags = FLAGS_INITIAL_STATE;
    neighborhood_size = 0;
    max_neighborhood_size = max_neighborhood;
    }


void SectorEnumerator::finish()
    {
    check_busy_flag();
    
    flags |= FLAGS_FINISHED;
    move_to_sector( originator_sector );
    }


void SectorEnumerator::rewind()
    {
    check_busy_flag();
    flags |= FLAGS_INITIAL_STATE;
    flags &= ~FLAGS_FINISHED;
    }


void SectorEnumerator::sector_visited()
    {
    if( neighborhood_size >= max_neighborhood_size )
        {
        finish();
        }
    else
        {
        next();
        }
    }


void SectorEnumerator::sector_unavailable()
    {
    if( neighborhood_size >= max_neighborhood_size )
        {
        finish();
        }
    else
        {
        next();
        }
    }


void SectorEnumerator::enumeration_finished()
    {
    }


void SectorEnumerator::move_to_sector
    (
    Remote<Sector> sector
    )
    {
    // Test parameters.

    if( sector == null )
        {
        finish();
        }

    // Lock enumerator.

    check_busy_flag();
    
    flags &= ~FLAGS_INITIAL_STATE;
    flags |= FLAGS_BUSY;

    // Unlink itself from the current sector.

    current_sector = null;
    current_address = SectorAddress();
        
    // Migrate to the new sector.

    if( flags_on( flags, FLAGS_ALLOW_LOCAL_CALLS ) )
        {
        deliver_asap_to( sector, flags_on( flags, FLAGS_ALLOW_CALLS_ON_REPLICAS ) );
        }
    else
        {
        migrate_to( sector, STime::current_time() );
        }
    }


void SectorEnumerator::move_to_sector
    (
    const SectorAddress & address
    )
    {
    Remote<Sector> sector;

    // Resolve sector address.

    if( address.layer != current_address.layer )
        {
        throw SectorErrorException
            ( "move_to_sector() to a different layer is not allowed." );
        }

    if( current_sector != null )
        {
        // If @c current_sector is available let the sector do the
        // resolution (@c address might be a neighbor of @c current_sector).

        sector = current_sector->get_sector( address );
        }
    else
        {
        // @c current_sector is unavailable. The resolution must be
        // performed by inspecting global sector map.
        
        Remote< Layer > layer = current_address.layer.convert();        
        sector = layer->optimize_replica( 20.0 ).sync_get_sector( address );
        }

    if( sector != null )
        {
        // Move to the sector.

        move_to_sector( sector );
        }
    else
        {
        // No such sector exists.
        // Finish the move.

        hope( !flags_on( flags, FLAGS_FINISHED ) );
        check_busy_flag();

        current_sector = null;
        current_address = address;

        // Notify about move to an unavailable sector.

        sector_unavailable();
        }
    }


void SectorEnumerator::move_to_sector
    (
    Direction direction
    )
    {
    move_to_sector( current_address.get_neighbor( direction ) );
    }


void SectorEnumerator::delivered_to
    (
    WeakPointer<Object> object,
    const STime &       time
    )
    {
    // Link itself to the new sector.

    current_sector = object.convert();
    current_address = current_sector->get_address();
    
    // Unlock the enumerator.

    flags &= ~FLAGS_BUSY;
    
    // Notify about sector change.

    if( flags_on( flags, FLAGS_FINISHED ) )
        {
        hope( current_sector == originator_sector );
        enumeration_finished();
        }
    else
        {
        if( current_sector->is_replica() )
            {
            hope( current_sector->is_consistent() );
            sector_visited();
            }
        else
            {
            // Create a strong pointer to the sector. This ensures
            // that the sector will not migrate to a remote node while
            // in @c sector_visited().

            Pointer<Object> ptr = object;
            sector_visited();
            }
        }
    }


void SectorEnumerator::check_busy_flag()
    {
    hope( !flags_on( flags, FLAGS_BUSY ) );
    }


void SectorEnumerator::allow_local_calls()
    {
    flags |= FLAGS_ALLOW_LOCAL_CALLS;
    }


void SectorEnumerator::allow_calls_on_replicas()
    {
    flags |= FLAGS_ALLOW_CALLS_ON_REPLICAS;
    }


// Implementation of FloodingEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void FloodingEnumerator::next()
    {
    check_busy_flag();

    if( flags_on( flags, FLAGS_INITIAL_STATE ) )
        {
        // Initial sector.

        neighborhood_size = 0;
        direction = NORTH;
        
        move_to_sector( originator_sector );
        }
    else
    if( steps_left > 0 )
        {
        // Move to neighboring sector in the current enumeration
        // direction.

        steps_left--;
        move_to_sector( direction.read() );
        }
    else
        {
        // We are at a vertex of the neighborhood.
        // Turn left (relatively to current direction).

        switch( direction )
            {
            case WEST:
                {
                direction = SOUTH;
                steps_left = neighborhood_size - 1;

                move_to_sector( direction.read() );
                break;
                }
            
            case SOUTH:
                {
                direction = EAST;
                steps_left = neighborhood_size - 1;
                
                move_to_sector( direction.read() );
                break;
                }

            case EAST:
                {
                direction = NORTH;
                steps_left = neighborhood_size - 1;
                
                move_to_sector( direction.read() );
                break;
                }

            case NORTH:
                {
                // We have enumerated the whole neighborhood.
                // Increase neighborhood size.

                direction = WEST;
                neighborhood_size += 2;
                steps_left = neighborhood_size;

                move_to_sector( NORTH_EAST );
                break;
                }

            default:
                {
                panic( "Invalid direction." );
                }
            }
        }
    }


// Implementation of ParallelFloodingEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ParallelFloodingEnumerator::next()
    {
    check_busy_flag();

    if( flags_on( flags, FLAGS_INITIAL_STATE ) )
        {
        // Initial sector.

        neighborhood_size = 0;
        move_to_sector( originator_sector );
        }
    else
    if( neighborhood_size == 0 )
        {
        int i;

        // Replace current enumerator by 8 enumerators.
        // 7 new enumerators are created.
        // Setup the new enumerators and spread them in the 8 directions.

        for( i = 0; i < 8; i++ )
            {
            Pointer<ParallelFloodingEnumerator> enumerator;
            Direction direction = static_cast<Direction>( i );

            if( i == 7 )
                {
                enumerator = this;
                }
            else
                {
                enumerator = clone().convert();
                }

            enumerator->direction = direction;
            enumerator->neighborhood_size = 1;
            enumerator->move_to_sector( direction );
            }
        }
    else
        {
        Direction   directions[ 3 ];
        int         i;

        switch( direction )
            {
            case EAST:
            case NORTH:
            case WEST:
            case SOUTH:
                {
                // Enumerator that is spread in one of the 4 major directions
                // is simply moved in that direction.

                neighborhood_size++;
                move_to_sector( direction.read() );

                return;
                }

            case NORTH_EAST:
                {
                // NORTH_EAST enumerator is spread into EAST, NORTH_EAST, NORTH.
                
                directions[ 0 ] = EAST;
                directions[ 1 ] = NORTH_EAST;
                directions[ 2 ] = NORTH;

                break;
                }

            case NORTH_WEST:
                {
                // NORTH_WEST enumerator is spread into NORTH, NORTH_WEST, WEST.
                
                directions[ 0 ] = NORTH;
                directions[ 1 ] = NORTH_WEST;
                directions[ 2 ] = WEST;

                break;
                }
            
            case SOUTH_WEST:
                {
                // SOUTH_WEST enumerator is spread into WEST, SOUTH_WEST, SOUTH.
                
                directions[ 0 ] = WEST;
                directions[ 1 ] = SOUTH_WEST;
                directions[ 2 ] = SOUTH;

                break;
                }
            
            case SOUTH_EAST:
                {
                // SOUTH_EAST enumerator is spread into SOUTH, SOUTH_EAST, EAST.
                
                directions[ 0 ] = SOUTH;
                directions[ 1 ] = SOUTH_EAST;
                directions[ 2 ] = EAST;

                break;
                }
            
            default:
                {
                panic( "Invalid direction." );
                }
            }

        // "Corner" enumerators (enumerators spread in the 4 minor directions)
        // are replaced by 3 enumerators spread in the 3 closest
        // directions (one minor and two majors).

        for( i = 0; i < 3; i++ )
            {
            Pointer<ParallelFloodingEnumerator> enumerator;
            
            if( i == 2 )
                {
                enumerator = this;
                }
            else
                {
                enumerator = clone().convert();
                }

            enumerator->direction = directions[ i ];
            enumerator->neighborhood_size++;
            enumerator->move_to_sector( directions[ i ] );
            }
        }
    }


// Implementation of BimodalFloodingEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void BimodalFloodingEnumerator::initialize
    (
    Pointer<Sector> originator,
    bool            parallel,
    Int32           max_neighborhood
    )
    {
    parallel_mode = parallel;
    SectorEnumerator::initialize( originator, max_neighborhood );
    }


void BimodalFloodingEnumerator::next()
    {
    if( parallel_mode )
        {
        ParallelFloodingEnumerator::next();
        }
    else
        {
        FloodingEnumerator::next();
        }
    }


} // namespace Demo::Lib
} // namespace Demo
