/** @file sector_enumerator.h
 * Sector enumerator.
 */

/*
 * begin        : Mon Jul 28 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_SECTOR_ENUMERATOR_H
#define DEMO_SECTOR_ENUMERATOR_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SERVER_FWD_H
#include "lib/server/server_fwd.h"
#endif

#ifndef DEMO_SECTOR_ADDRESS_H
#include "lib/shared/sector_address.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Class that holds a context of an "asynchronous query".
 * Note that Sectors are different migration groups and thus may be
 * located on different nodes. To simplify query implementation
 * Enumerator interface is provided.
 *
 * Enumerator visits layer sectors in a well-defined (enumerator-specific)
 * order starting at @c originator_sector. On each visit @c sector_visited()
 * is called. When the enumeration finishes, the enumerator is returned
 * back to the originator sector and @c enumeration_finished() is called.
 *
 * Client manually calls @c next() to transfer the enumerator to
 * the next sector. If the sector does not exist the enumerator
 * call-backs @c sector_unavailable() and finishes the move.
 * It is then okay to submit another @c next() to skip the missing sector.
 *
 * Current enumerator position can be restored back to initial position
 * by @c rewind(). This allows to implement two-phase queries according
 * to this pattern:
 *  - gather information from Sectors (@c next(), first phase)
 *  - send the information to originator for central processing (@c rewind())
 *  - distribute the results (@c next(), second phase)
 *
 * Since the enumeration works asynchronously, some synchronization may
 * be needed. The first phase can be used to "reserve" (lock, etc)
 * the sectors.
 *
 * SectorEnumerator is not a GC root object (by-default).
 */
class SectorEnumerator : public Object
    {
public:
    
    MASSIV_OBJECT( SectorEnumerator );

    //
    // SectorEnumerator interface.
    //

    enum Flags
        {
        FLAGS_INITIAL_STATE =           1,
            /**< The enumerator is in initial state.
                 Rewind turns this flag on. */
        FLAGS_BUSY =                    2,
            /**< The enumerator is busy. */
        FLAGS_FINISHED =                4,
            /**< The enumeration has finished. */
        FLAGS_ALLOW_LOCAL_CALLS =       8,
            /**< Are direct calls to local sectors allowed?
                 If yes, local sectors are visited immediatelly instead
                 of scheduling migrations and visits.
                 Note that this changes enumerator semantics because
                 remote sectors are handled in a different way than
                 local ones.
                 The feature is turned off by default.
                 To enable this call @c allow_local_calls(). */
        FLAGS_ALLOW_CALLS_ON_REPLICAS = FLAGS_ALLOW_LOCAL_CALLS | 16
            /**< Are direct calls to sector replicas allowed?
                 If enabled client code must be aware that it may work
                 on a sector replica, not an original data.
                 To enable call @c allow_calls_on_replicas(). */
        };
    
    /**
     * Parameters.
     */
    enum Parameters
        {
        DEFAULT_MAX_NEIGHBORHOOD_SIZE = 4
            /**< Max neighborhood size used by default implementation
                 of @c sector_visited() and @c sector_unavailable(). */
        }; // Parameters

    /**
     * Initializes object.
     */
    void initialize
        (
        Pointer<Sector> originator,
        Int32           max_neighborhood = DEFAULT_MAX_NEIGHBORHOOD_SIZE
        );

    /**
     * Requests the enumerator to move to the next sector.
     * When done @c sector_visited(), @c sector_unavailable() or
     * enumeration_finished() is called.
     * Until done, the enumerator is busy and rejects any operation.
     */
    virtual void next() = 0;
    
    /**
     * Requests the enumeration to return back to the originator.
     * When done @c enumeration_finished() is called.
     * Until done, the enumerator is busy and rejects any operation.
     */
    virtual void finish();

    /**
     * Restores enumerator initial state.
     * Next call to @c next() will transfer the iterator to the
     * @c originator_sector.
     */
    virtual void rewind();
    
    /**
     * Called by the enumerator when a new sector is visited.
     * Subclass can use @c current_sector to access the visited sector.
     * It is guaranteed that @c current_sector points to a local object.
     * The default implementation stops the enumeration if
     * current neighborhood size >= @c max_neighborhood_size.
     */
    virtual void sector_visited();

    /**
     * Called by the enumerator when an unavailable sector is visited.
     * This happens if enumerator reaches edge of the layer or there
     * is a "hole" in the layer (missing neighboring sector).
     * Subclass can use @c current_address to examine address of
     * the unavailable sector. @c current_sector is null.
     * The default implementation stops the enumeration if
     * current neighborhood size >= @c max_neighborhood_size.
     */
    virtual void sector_unavailable();

    /**
     * Called by the enumerator when the enumeration has finished.
     * Subclass can use @c originator_sector to access the originator
     * sector (which is guaranteed to be a local object).
     */
    virtual void enumeration_finished();

protected:

    /**
     * Moves enumerator to @c sector.
     * This is done be migrating the enumerator to that sector.
     * When the enumerator is "delivered" (the move finishes),
     * @c current_sector and @c current_address is updated.
     * If @c sector is @c null the enumeration finishes.
     * @note Until previous move is finished, another call to
     * to @c move_to_sector() will fail.
     */
    virtual void move_to_sector
        (
        Remote<Sector> sector
        );
    
    /**
     * Moves enumerator to a sector identified by @c address.
     * If no such sector exists the enumerator will point to
     * an undefined sector.
     */
    virtual void move_to_sector
        (
        const SectorAddress & address
        );
    
    /**
     * Moves enumerator to a neighboring sector relative to
     * @c current_sector (or @c current_address).
     * If no such sector exists the enumerator will point to
     * an undefined sector.
     */
    virtual void move_to_sector
        (
        Direction direction
        );

    /**
     * Called by the system when the enumerator is transfered to @c object.
     * @c object is assumed to be pointing to a Sector.
     */
    virtual void delivered_to
        (
        WeakPointer<Object> object,
        const STime &       time
        );

    /**
     * Checks if the enumerator is not busy.
     * The enumerator is busy if a move has been scheduled but the move
     * is still in progress.
     */
    void check_busy_flag();

    /**
     * Enables local call optimization.
     * If sector to be visited is local, @c move_to_sector() will immediatelly
     * call @c sector_visited() instead of scheduling the move.
     */
    void allow_local_calls();

    /**
     * Enables local call optimization on sector replicas.
     * If sector to be visited is replicated to current node, @c move_to_sector()
     * will immediatelly call @c sector_visited() on the replica instead of
     * scheduling the move to the original sector.
     */
    void allow_calls_on_replicas();

protected:
    
    PWeakPointer<Sector>        originator_sector;
        /**< The sector that originated the enumeration. */
    PWeakPointer<Sector>        current_sector;
        /**< Sector the enumerator has just visited.
             The pointer is weak so that sector replicas could be visited
             too. */
    SectorAddress               current_address;
        /**< Current sector address.
             Note that the address is always valid even if
             @c current_sector is @c null. */
    PVlInt32                    flags;
        /**< Enumerator flags. */
    PVlInt32                    neighborhood_size;
        /**< Current neighborhood size minus one.
             How far is @c current_sector from the @c originator_sector. */
    PVlInt32                    max_neighborhood_size;
        /**< Maximum neighborhood size. The enumeration automatically
             stops when a sector of that neighborhood is visited
             (under a condition that @c sector_visited() and
             @c sector_unavailable() call the original SectorEnumeration
             versions of the callbacks). */
    }; // SectorEnumerator


/**
 * The enumerator systematically visits sectors in originator's
 * 0-neighborhood, 1-neighborhood, ...
 * The neighborhoods correspond to eight-conectivity neighborhoods
 * with discrete minimum metrics.
 */
class FloodingEnumerator : public virtual SectorEnumerator
    {
public:

    MASSIV_OBJECT( FloodingEnumerator );

    /**
     * Assignment operator.
     *
     * @bug Quickhack fix for gcc 3.3.2pre3. Should be removed when gcc is fixed.
     */
    inline FloodingEnumerator & operator=
        (
        const FloodingEnumerator & other
        )
        {
        steps_left = other.steps_left;
        direction = other.direction;
        SectorEnumerator::operator=( other );
        return *this;
        }

protected:

    //
    // SectorEnumerator interface implementation.
    //

    /**
     */
    virtual void next();

protected:

    PVlInt32            steps_left;
        /**< How many steps are there left until enumeration direction
             is changed. */
    PEnum<Direction>    direction;
        /**< What direction is the enumerator just moving. */
    }; // FloodingEnumerator


/**
 * The enumerator visits sectors in originator's neighborhood.
 * The difference from the FloodingEnumerator is that the enumeration
 * is performed in parallel. That means that @c next() may clone
 * existing enumerator to create new instances of the enumerator that
 * will run in parallel with the parent enumerator.
 * The only property that is guaranteed is that each sector will be
 * visited at most once. The order in which the sectors are visited
 * is random.
 * The implementation uses @c clone() to clone the enumerator context.
 * @code
 * // The following code visits all sectors that lie in 0, 1, ..,
 * // n-neighborhood. Each sector is visisted exactly once.
 * class MyEnumrator : ParallelFloodingEnumerator
 *  {
 *  virtual void sector_visited()
 *      {
 *      // Add your code here that does some job on @c current_sector.
 *      if( neighborhood_size < n )
 *          {
 *          // Continue.
 *          next();
 *          }
 *      }
 *  };
 * Pointer<MyEnumerator> e = CreateObject::enumerator( ... );
 * e->next(); // Start flooding.
 * @endcode
 */
class ParallelFloodingEnumerator : public virtual SectorEnumerator
    {
public:

    MASSIV_OBJECT( ParallelFloodingEnumerator );

    /**
     * Assignment operator.
     *
     * @bug Quickhack fix for gcc 3.3.2pre3. Should be removed when gcc is fixed.
     */
    inline ParallelFloodingEnumerator & operator=
        (
        const ParallelFloodingEnumerator & other
        )
        {
        direction = other.direction;
        SectorEnumerator::operator=( other );
        return *this;
        }

protected:

    //
    // SectorEnumerator interface implementation.
    //

    /**
     */
    virtual void next();

protected:

    PEnum<Direction>    direction;
        /**< Classifies enumerator type and direction the flood is
             spread into from the current sector. */
    }; // ParallelFloodingEnumerator


/**
 * Bimodal enumerator that operates either like FloodingEnumerator or
 * ParallelFloodingEnumerator.
 * @note In the first version it was implemented without the virtual
 * inheritace. In fact the inheritance was just "emulated" by hand and the
 * code was horrible. Later it was rewritten using clean C++ approach.
 */
class BimodalFloodingEnumerator : public FloodingEnumerator, public ParallelFloodingEnumerator
    {
public:

    MASSIV_OBJECT( BimodalFloodingEnumerator );

    //
    // BimodalFloodingEnumerator interface.
    //
    
    /**
     * Initializes object.
     */
    void initialize
        (
        Pointer<Sector> originator,
        bool            parallel,
        Int32           max_neighborhood = DEFAULT_MAX_NEIGHBORHOOD_SIZE
        );

protected:

    //
    // SectorEnumerator interface implementation.
    //

    /**
     * According to @c parallel_mode the call is redirected into one of
     * the base classes.
     */
    virtual void next();

protected:
        
    PBool parallel_mode;
        /**< Is the enumerator switched into parallel mode? */

    }; // BimodalFloodingEnumerator


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_SECTOR_ENUMERATOR_H
