/** @file sector_helper_objects.cpp
 * Sector helper objects implementation.
 */

/*
 * begin        : Mon Aug 4 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/sector_helper_objects.h"
#include "lib/server/player.h"
#include "lib/server/player_rpc.h"
#include "lib/shared/client_sector.h"


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


// Implementation of ModifyElevationTask
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ModifyElevationTask::initialize
    (
    Pointer<Sector>                 the_originator,
    const SVector2 &                the_center,
    Sector::ElevationMapKind        the_map_kind,
    Sector::ElevationMapFunction    the_function,
    Sector::ElevationMapOperation   the_operation,
    Float32                         the_radius,
    Float32                         the_strength,
    Int32                           the_max_neighborhood_size
    )
    {
    // Initialize context.

    center = the_center;
    map_kind = the_map_kind;
    function = the_function;
    operation = the_operation;
    radius = the_radius;
    strength = the_strength;
    
    ParallelFloodingEnumerator::initialize( the_originator, the_max_neighborhood_size );
    allow_local_calls();

    // Start enumeration.

    next();
    }


void ModifyElevationTask::sector_visited()
    {
    // If current sector has not been modified by the function, do not
    // visit next sectors as they are further away.

    const bool modified = current_sector->modify_elevation
        (
        center.read(),
        map_kind,
        function,
        operation,
        radius,
        strength,
        true
        );

    if( modified )
        {
        ParallelFloodingEnumerator::sector_visited();
        }
    else
        {
        finish();
        }
    }


// Implementation of ModifyMaterialTask
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ModifyMaterialTask::initialize
    (
    Pointer<Sector>                 the_originator,
    const SVector2 &                the_center,
    Float32                         the_radius,
    ClientSector::MaterialId        the_material_id,
    Int32                           the_max_neighborhood_size
    )
    {
    // Initialize context.

    center = the_center;
    radius = the_radius;
    material_id = the_material_id;
    
    ParallelFloodingEnumerator::initialize( the_originator, the_max_neighborhood_size );
    allow_local_calls();

    // Start enumeration.

    next();
    }

    
void ModifyMaterialTask::sector_visited()
    {
    // See ModifyElevationTask::sector_visited().

    const bool modified = current_sector->modify_material
        (
        center.read(),
        radius,
        material_id.read(),
        true
        );

    if( modified )
        {
        ParallelFloodingEnumerator::sector_visited();
        }
    else
        {
        finish();
        }
    }


// Implementation of MessageEventTask
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


void MessageEventTask::initialize
    (
    Pointer< Sector >           the_originator,
    WeakPointer< ClientEntity > the_client_entity,
    const SVector2 &            the_center,
    Float32                     the_radius,
    const std::string &         the_message,
    const SColor &              the_message_color
    )
    {
    int num_vertices = 8;
    
    // Create circle area.
    ConvexArea  area( the_center, the_radius, num_vertices );
    
    // Initialize context.
    client_entity = the_client_entity;
    message = the_message;
    // fixme
    message_color.set_red( the_message_color.red() );
    message_color.set_green( the_message_color.green() );
    message_color.set_blue( the_message_color.blue() );
    message_color.set_alpha( the_message_color.alpha() );
    
    // Initialize spatial query.
    // Spatial query object must be initialized last, because
    // it automatically starts enumeration.
    SpatialQuery::initialize
        (
        the_originator,
        area,
        Entity::CLASS_PLAYER, // Enumerate only players.
        false // Sequential mode - the event must be deliver to one player mostly once.
        );
    }

    
bool MessageEventTask::entity_reported
    (
    WeakPointer< Entity > entity
    )
    {
    hope( entity != null );
    
    // todo: Sometime in the future player object may have flag whether its client
    // is connected or not - in that case send the message only to connected players.
    Remote< Player >    player_remote = entity.convert();
    SColor              color = message_color.read();
    player_remote->async_log_message_event( client_entity, message, color );
    
    return true;
    }


// Implementation of ReplicateNearbySectorsTask
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


void ReplicateNearbySectorsTask::initialize
    (
    Pointer< Sector >       the_originator,
    WeakPointer< Object >   the_destination_node,
    const STime &           the_duration,
    Int32                   the_max_neighborhood
    )
    {
    destination_node = the_destination_node;
    duration = the_duration;
    
    Super::initialize( the_originator, the_max_neighborhood );
    allow_local_calls();
    
    // Start enumeration.
    next();
    }
    

void ReplicateNearbySectorsTask::sector_visited()
    {
    hope( current_sector.is_local() );
    
    // While an entity is moving between two sectors, its unlinked
    // from its sector for a little time. If replication update
    // is sent to the client while the entity is not linked, and only
    // the sector is requested to replicate, the entity is not added to
    // the replication update because it is not in the replication group
    // of the sector any more so the replication update contains
    // command that the client should destroy the replica of the entity.
    // 
    // To solve this problem, we must create replication requests not
    // only for the sector itself, but for all entities of the sector.
    
    current_sector->request_replication_to_client
        ( destination_node, STime::current_time() + duration );
    
    Super::sector_visited();
    }


} // namespace Demo::Lib
} // namespace Demo
