/** @file sector_helper_objects.h
 * Sector helper objects that need not be visible to user.
 */

/*
 * begin        : Mon Aug 4 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_SECTOR_HELPER_OBJECTS_H
#define DEMO_SECTOR_HELPER_OBJECTS_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SECTOR_ENUMERATOR_H
#include "lib/server/sector_enumerator.h"
#endif

#ifndef DEMO_SPATIAL_QUERY_H
#include "lib/server/spatial_query.h"
#endif

#ifndef DEMO_SECTOR_H
#include "lib/server/sector.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Helper enumerator that modifies elevation map of each visited
 * sector by calling @c Sector::modify_elevation().
 * The enumerator holds context of the "distributed query"
 * (see Sector for argument description).
 * The current enumerator instance is stopped if elevation function
 * is constant-zero on the current sector or the sector is too
 * far from the originator sector.
 */
class ModifyElevationTask : public ParallelFloodingEnumerator
    {
public:

    MASSIV_OBJECT( ModifyElevationTask );

    /**
     * Initializes task context.
     * Starts the task.
     */
    void initialize
        (
        Pointer<Sector>                 the_originator,
        const SVector2 &                the_center,
        Sector::ElevationMapKind        the_map_kind,
        Sector::ElevationMapFunction    the_function,
        Sector::ElevationMapOperation   the_operation,
        Float32                         the_radius,
        Float32                         the_strength,
        Int32                           the_max_neighborhood_size = DEFAULT_MAX_NEIGHBORHOOD_SIZE
        );

protected:
    
    /**
     * Modifies elevation map of the current sector.
     * Proceeds to the next sectors.
     */
    virtual void sector_visited();

protected:
        
    PVector2                                center;
        /**< Center-point. */
    PEnum<Sector::ElevationMapKind>         map_kind;
        /**< Elevation map being modified. */
    PEnum<Sector::ElevationMapFunction>     function;
        /**< Elevation function. */
    PEnum<Sector::ElevationMapOperation>    operation;
        /**< Elevation operation applied on the current elevation. */
    PFloat32                                radius;
        /**< Elevation function parameter. */
    PFloat32                                strength;
        /**< Elevation function parameter. */

    }; // ModifyElevationTask


/**
 * Helper enumerator that modifies material map of each visited
 * sector by calling @c Sector::modify_material().
 * The enumerator holds context of the "distributed query"
 * (see Sector for argument description).
 * The current enumerator instance is the sector is too
 * far from the originator sector.
 */
class ModifyMaterialTask : public ParallelFloodingEnumerator
    {
public:

    MASSIV_OBJECT( ModifyMaterialTask );

    /**
     * Initializes task context.
     * Starts the task.
     */
    void initialize
        (
        Pointer<Sector>                 the_originator,
        const SVector2 &                the_center,
        Float32                         the_radius,
        ClientSector::MaterialId        the_material_id,
        Int32                           the_max_neighborhood_size = DEFAULT_MAX_NEIGHBORHOOD_SIZE
        );

protected:
    
    /**
     * Modifies material map of the current sector.
     * Proceeds to the next sectors.
     */
    virtual void sector_visited();

protected:
        
    PVector2                                center;
        /**< Center-point. */
    PFloat32                                radius;
        /**< Elevation function parameter. */
    ClientSector::PMaterialId               material_id;
        /**< Material to set. */

    }; // ModifyMaterialTask
    
    
/**
 * Message event task - enumerates all clients (through their players)
 * in specified circle area (specified by center and radius)
 * and sends a message to them.
 */
class MessageEventTask : public SpatialQuery
    {
public:
    
    MASSIV_OBJECT( MessageEventTask );

    /**
     * Initialize task context and starts the task.
     * The enumeration will start at @c the_originator.
     * @param client_entity Specifies entity that generated the event.
     *      May be null.
     */
    void initialize
        (
        Pointer< Sector >           the_originator,
        WeakPointer< ClientEntity > client_entity,
        const SVector2 &            center,
        Float32                     radius,
        const std::string &         message,
        const SColor &              message_color
        );
    
protected:

    // From SpatialQuery.
    bool entity_reported
        (
        WeakPointer< Entity > entity
        );

protected:

    PWeakPointer< ClientEntity >    client_entity;
        /**< Pointer to entity that generated the event. May be null.
             May be used by client to place the event message above
             the entity on screen. */
    PString                         message;
        /**< Text of the message. */
    PColor                          message_color;
        /**< Color of the message. */

    }; // MessageEventTask


/**
 * Helper enumerator that requests replication of nearby sectors to a node
 * for specified amount of time.
 */
class ReplicateNearbySectorsTask : public ParallelFloodingEnumerator
    {
    typedef ParallelFloodingEnumerator  Super;

public:

    MASSIV_OBJECT( ReplicateNearbySectorsTask );
    
    /**
     * Initialize the object.
     * Requests replication of sectors inside originator's neighborhood
     * to node specified by @a destination_node proxy node object
     * for @a duration seconds from current time.
     */
    void initialize
        (
        Pointer< Sector >       originator,
        WeakPointer< Object >   destination_node,
        const STime &           duration,
        Int32                   max_neighborhood
        );

protected:

    /**
     * Replicates current sector to destination node.
     */
    virtual void sector_visited();

protected:

    PWeakPointer< Object >      destination_node;
        /**< Target node for replication. */
    
    PTime                       duration;
        /**< Replication duration. */

    }; // ReplicateNearbySectorsTask


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_SECTOR_HELPER_OBJECTS_H
