/** @file spatial_query.cpp
 * Spatial query implementation.
 */

/*
 * begin        : Fri Aug 15 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/server/spatial_query.h"
#include "lib/server/sector.h"
#include "lib/shared/client_sector.h"

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/** @def RESERVATION_MAP_BASED_ENUMERATION
 * Define this to enable reservation map based entity enumeration.
 */
//#define RESERVATION_MAP_BASED_ENUMERATION


namespace {

/**
 * Computes the minimum size of an @c origin-centered sector neighborhood
 * that will cover @c area.
 * The @c area must be defined in layer coordinates.
 */
Int32 compute_neighborhood_size
    (
    const SectorAddress &   origin,
    const ConvexArea &      area
    )
    {
    std::size_t i;
    Int32       radius;

    radius = 0;

    for( i = 0; i < area.vertices.size(); i++ )
        {
        #ifndef OPTION_OPTIMIZE

        SectorAddress   address( null, area.vertices[ i ].read() );
        Int32           column_difference = address.column - origin.column;
        Int32           row_difference = address.row - origin.row;

        #else

        Int32 column_difference = FloatTraits<Float32>::floor_int
            ( area.vertices[ i ].x() / ClientSector::SECTOR_X_SIZE ) -
            origin.column;
        Int32 row_difference = FloatTraits<Float32>::floor_int
            ( area.vertices[ i ].y() / ClientSector::SECTOR_Y_SIZE ) -
            origin.column;
        
        // Note: The code above has to be kept up in sync with the
        //       implementation of SectorAddress.

        #endif

        if( column_difference > radius )
            {
            radius = column_difference;
            }
        
        if( row_difference > radius )
            {
            radius = row_difference;
            }
        }

    return radius + 1;
    }
    

/**
 * Translates convex area defined in layer coordinates to another area
 * defined in @c sector's tile coordinates.
 */
void layer_area_to_tile_area
    (
    const ConvexArea &  layer_area,
    ConvexArea &        tile_area,
    WeakPointer<Sector> sector
    )
    {
    std::size_t i;

    for( i = 0; i < layer_area.vertices.size(); i++ )
        {
        SVector2 vertex;

        sector->layer_xy_to_tile_xy( layer_area.vertices[ i ], vertex );
        tile_area.add_vertex( vertex );
        }

    tile_area.update_edges();
    }


#ifdef RESERVATION_MAP_BASED_ENUMERATION

/**
 * Query enumerator base class.
 * Enumerates entities that pass @c query filter and belong to @c sector.
 */
class QueryEnumerator : public ConvexArea::RasterizerCallback
    {
protected:
    
    friend class SpatialQuery;

    /** Constructor. */
    QueryEnumerator
        (
        SpatialQuery &      query,
        const ConvexArea &  tile_area,
        WeakPointer<Sector> sector
        ) : query( query ),
            sector( sector ),
            tile_area( tile_area ),
            finished( false )
        {
        }

    /** Enumerates entities from the tile range. */
    virtual bool operator()
        (
        int begin_x_tile,
        int end_x_tile,
        int y_tile
        ) const
        {
        while( begin_x_tile < end_x_tile )
            {
            // Enumerate entities from [ begin_x_tile, y_tile ].

            std::size_t i;

            sector->enumerate_entities
                ( begin_x_tile, y_tile, query.entity_flags, list );

            for( i = 0; i < list.size(); i++ )
                {
                if( !report( list[ i ] ) )
                    {
                    finished = true;
                    return false;
                    }
                }

            // Next tile.
            
            if( !sector.is_local() || !sector->is_consistent() )
                {
                // Sector replica was deleted or became inconsistent meanwhile.
                // Can not continue.
                
                finished = true;
                return false;
                }

            ++begin_x_tile;
            }

        return true;
        }
    
    /**
     * Reports @c entity.
     * Returns true if the enumeration should continue.
     */
    virtual bool report
        (
        WeakPointer<Entity> entity
        ) const = 0;

protected:
    
    SpatialQuery &              query;
        /**< Query context. */
    WeakPointer<Sector>         sector;
        /**< Current sector. */
    const ConvexArea &          tile_area;
        /**< The area. In @c sector's tile coordinates. */
    mutable Sector::EntityList  list;
        /**< Helper entity list. */
    mutable bool                finished;
        /**< Set to true if client requested to stop the enumeration. */
        
    }; // QueryEnumerator


/**
 * Enumerator implementation for the sequential query.
 */
class EnumeratorForSequentialQuery : public QueryEnumerator
    {
public:
        
    /** Constructor. */
    EnumeratorForSequentialQuery
        (
        SpatialQuery &      query,
        const ConvexArea &  tile_area,
        Pointer<Sector>     sector
        ) : QueryEnumerator( query, tile_area, sector )
        {
        }
    
    /**
     */
    virtual bool report
        (
        WeakPointer<Entity> entity
        ) const
        {
        // Already reported?

        if( query.reported.find( entity ) != query.reported.end() )
            {
            return true;
            }

        // Should the entity be reported now?
        
        if( entity.is_local() )
            {
            SVector2 origin = entity->get_current_position().origin.read();

            if( !sector->is_in_this_sector( origin ) )
                {
                if( query.area.contains( origin ) )
                    {
                    // @c entity will be enumerated by the sector its
                    // origin belongs to.

                    return true;
                    }
                }
            }

        // Report.

        query.reported.insert( entity );
        
        return query.entity_reported( entity );
        }

    }; // EnumeratorForSequentialQuery


/**
 * Enumerator implementation for the parallel query.
 * Entities that belong to the enumerated area partially only need
 * be handled with extra care so that they would be enumerated only once
 * (if they do not move, of course).
 */
class EnumeratorForParallelQuery : public QueryEnumerator
    {
public:
        
    typedef ext::hash_set<ObjectId, ObjectId::Hash, ObjectId::EqualKey> ReportedList;

    /** Constructor. */
    EnumeratorForParallelQuery
        (
        SpatialQuery &      query,
        const ConvexArea &  tile_area,
        Pointer<Sector>     sector
        ) : QueryEnumerator( query, tile_area, sector )
        {
        }

    /**
     */
    virtual bool report
        (
        WeakPointer<Entity> entity
        ) const
        {
        // Already reported?

        if( reported.find( entity.get_object_id() ) != reported.end() )
            {
            return true;
            }

        // Check if the entity is local.
        // We hope that we have at least entity replica on this node.
        // Otherwise we can not do anything (except costy SRPC).

        if( entity.is_local() )
            {
            // Does @c entity's origin belong to this sector?

            SVector2 origin = entity->get_current_position().origin.read();

            if( !sector->is_in_this_sector( origin ) )
                {
                // No. But @c entity's bounding box does and intersects
                // @c tile_area.
                // This implies that the entity should be enumerated.
                // Note that we want to report the entity once only
                // (in exactly one sector).
                // Choose which sector will report the entity.
                // All the sectors involved must do the same decision.
                
                SectorAddress   address( sector->get_address().layer, origin );
                ConvexArea      bounding_box = sector->get_bounding_box( entity );
                ConvexArea      owner_sector_area = sector->get_sector_area( address );
                ConvexArea      intersection_of_bounding_box_and_tile_area = 
                                bounding_box.intersect_with( tile_area );

                if( intersection_of_bounding_box_and_tile_area.intersects( owner_sector_area ) )
                    {
                    // Owner sector contains entity's origin and
                    // will enumerate that entity, because @c tile_area intersects
                    // both the owner sector and the bounding box.

                    return true;
                    }
                else
                    {
                    // Owner sector can not enumerate @c entity, because
                    // its area is outside the processed @c tile_area.
                    // Find a neighboring sector with the "best" SectorAddress
                    // that intersects entity's bounding box clipped by @c tile_area.
                    // This is the sector that will enumerate that entity.
                    
                    SectorAddress   elected_address = address;
                    bool            elected = false;
                    int             i;

                    for( i = 0; i < 8; i++ )
                        {
                        Direction       direction = static_cast<Direction>( i );
                        SectorAddress   neighbor_address = address.get_neighbor( direction );
                        ConvexArea      neighbor_area = sector->get_sector_area( neighbor_address );
                        
                        if( intersection_of_bounding_box_and_tile_area.intersects( neighbor_area ) )
                            {
                            // @c neighbor_address sector will also try to
                            // report @c entity. This is one of the sector
                            // "candidates".

                            if( neighbor_address.row <= elected_address.row ||
                                neighbor_address.column <= elected_address.column ||
                                !elected )
                                {
                                elected_address = neighbor_address;
                                elected = true;
                                }
                            }
                        }
                    
                    // Test if current sector is the elected one.

                    if( sector->get_address() != elected_address )
                        {
                        return true;
                        }
                    }
                }
            }

        // Report.

        reported.insert( entity.get_object_id() );
        
        return query.entity_reported( entity );
        }

public:

    mutable ReportedList reported;
        /**< List of entities reported in this @c sector.
             Native data structure is used instead of @c query.reported
             for better performance. */

    }; // EnumeratorForParallelQuery

#endif


}    


void SpatialQuery::initialize
    (
    Pointer<Sector>     the_originator,
    const ConvexArea &  the_area,
    int                 the_entity_flags,
    bool                the_parallel_mode,
    bool                the_use_replicas
    )
    {
    initialize
        (
        the_originator,
        the_area,
        SVector2(),
        SVector2(),
        0.0f,
        the_entity_flags,
        the_parallel_mode,
        the_use_replicas
        );
    }


void SpatialQuery::initialize
    (
    Pointer<Sector>     the_originator,
    const ConvexArea &  the_area,
    const SVector2 &    the_offset,
    const SVector2 &    the_rotation_center,
    const Float32       the_rotation_angle,
    int                 the_entity_flags,
    bool                the_parallel_mode,
    bool                the_use_replicas
    )
    {
    // Initialize query.

    area = the_area;
    
    if( !the_offset.epsilon_zero() )
        {
        area.translate( the_offset );
        }

    if( !FloatTraits<Float32>::epsilon_zero( the_rotation_angle ) )
        {
        area.rotate( the_rotation_center, the_rotation_angle );
        }

    entity_flags = the_entity_flags;

    BimodalFloodingEnumerator::initialize
        (
        the_originator,
        the_parallel_mode,
        compute_neighborhood_size( the_originator->get_address(), area )
        );
    
    allow_local_calls();
    
    if( the_use_replicas )
        {
        allow_calls_on_replicas();
        }

    // Start enumeration.

    next();
    }


void SpatialQuery::sector_visited()
    {
    bool finished;

    // Convert @c area to @c current_sector's tile coordinates.

    ConvexArea tile_area;
    
    layer_area_to_tile_area( area, tile_area, current_sector );

    // Enumerate entities that belong to @c tile_area in the current sector.

    #ifdef RESERVATION_MAP_BASED_ENUMERATION

    // Reservation map based enumeration.
    // Uses rasterizer to determine tiles belonging to @c tile_area and
    // lookups reservations for those tiles.

    if( parallel_mode )
        {
        EnumeratorForParallelQuery callback( *this, tile_area, current_sector );
        
        tile_area.rasterize
            (
            callback,
            0,
            0,
            ClientSector::NUM_X_TILES,
            ClientSector::NUM_Y_TILES
            );

        finished = callback.finished;
        }
    else
        {
        EnumeratorForSequentialQuery callback( *this, tile_area, current_sector );
        
        tile_area.rasterize
            (
            callback,
            0,
            0,
            ClientSector::NUM_X_TILES,
            ClientSector::NUM_Y_TILES
            );
        
        finished = callback.finished;
        }

    #else

    // Entity list based enumeration.
    // Iterates through all entities registered to @c current_sector.
    // Since any entity is registered to at most one sector, the enumeration
    // ensures the "at most once" semantics (unless entities move meanwhile).
    
    finished = false;
    hope( current_sector->is_consistent() );

    // Since @c entity_reported() can potentially invoke SRPC, @c current_sector
    // could be changed during the enumeration which might invalidate our
    // enumerator. To simplify the enumeration process, we will enumerate to
    // a private list first and than report the entities from the list.
    // Note that this approach is essential if enumeration should work on
    // @c current_sector replicas. Because of this ListEnumerator can not be
    // used.

    WeakPointer<List>                   entity_list = current_sector->entity_list;
    WeakPointer<List::ListEntry>        enumerator = entity_list->get_first();
    std::vector< WeakPointer<Entity> >  to_report;
    std::size_t                         i;

    // Enumerate entities.    

    while( enumerator != entity_list->get_last() )
        {
        WeakPointer<Entity> entity = enumerator->current().convert();
        hope( entity.is_local() );

        if( entity->match_entity_flags( entity_flags ) && tile_area.intersects
            ( current_sector->get_bounding_box( entity ) ) )
            {
            to_report.push_back( entity );
            }

        enumerator = enumerator->get_next();
        }

    // Report.

    for( i = 0; i < to_report.size(); i++ )
        {
        WeakPointer<Entity> entity = to_report[ i ];

        if( !parallel_mode )
            {
            if( reported.find( entity ) == reported.end() )
                {
                reported.insert( entity );
                
                if( !entity_reported( entity ) )
                    {
                    finished = true;
                    break;
                    }
                }
            }
        else
            {
            if( !entity_reported( entity ) )
                {
                finished = true;
                break;
                }
            }
        }

    #endif

    // Visit next sector.

    if( !finished )
        {
        SectorEnumerator::sector_visited();
        }
    else
        {
        finish();
        }
    }


} // namespace Demo::Lib
} // namespace Demo
