/** @file spatial_query.h
 * Spatial query.
 */

/*
 * begin        : Fri Aug 15 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_SPATIAL_QUERY_H
#define DEMO_SPATIAL_QUERY_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SECTOR_ENUMERATOR_H
#include "lib/server/sector_enumerator.h"
#endif

#ifndef DEMO_CONVEX_AREA_H
#include "lib/server/convex_area.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Spatial query.
 * Enumerates entities that lay (at least partially) inside a given area.
 * Once the object is initialized the enumeration starts automatically.
 * The enumerated entities are reported by @c entity_reported().
 * While handling the query all the sectors, that participate in the query,
 * can either be processed sequentially or in parallel.
 * In sequential mode each entity is reported at most once, one
 * @c SpatialQuery instance is used to visit all the involved sectors.
 * In parallel mode entities can be reported more than once (if they have
 * moved meanwhile), @c SpatialQuery instance is cloned and each instance
 * operates on a different sector (in parallel).
 * Note that if entities move while the enumeration is in progress,
 * the query result may be inconsistent.
 */
class SpatialQuery : public BimodalFloodingEnumerator
    {
public:
        
    MASSIV_OBJECT( SpatialQuery );
   
    typedef PSet< PWeakPointer< Entity > > ReportedList;

public:

    /**
     * Initializes the object and initiates the query.
     * The enumeration flood will start at @c the_originator.
     * [ @c the_area, @c the_entity_flags ] define query criteria.
     * Only entites that belong to the area and are of the requested class
     * are reported. See @c Sector::enumerate_entities() for
     * @c the_entity_flags description.
     * The @c area is specified in layer 2D coordinates.
     * @c the_use_replicas specifies whether it is okay to access sector
     * replicas. Note that even if @c the_use_replicas is false the query
     * result still may be inconsistent (the neighboring sectors are not
     * locked during the query processing, the entity can move meanwhile).
     * @c the_parallel_mode determines whether the query object can be
     * cloned and its copies dispatched to multiple sectors in a time.
     */
    void initialize
        (
        Pointer<Sector>     the_originator,
        const ConvexArea &  the_area,
        int                 the_entity_flags,
        bool                the_parallel_mode = true,
        bool                the_use_replicas = true
        );
    
    /**
     * @c initialize() variant that allows to transform @c the_area
     * before it is used by the enumerator.
     * The transformation consists of simple translation by @c offest
     * and rotation by @c the_rotation_angle around @c the_rotation_center
     * (in this order).
     */
    void initialize
        (
        Pointer<Sector>     the_originator,
        const ConvexArea &  the_area,
        const SVector2 &    the_offset,
        const SVector2 &    the_rotation_center,
        const Float32       the_rotation_angle,
        int                 the_entity_flags,
        bool                the_parallel_mode = true,
        bool                the_use_replicas = true
        );

public:
    
    //
    // Spatial query interface.
    //

    /**
     * Called by the enumerator to report an entity that passed the
     * query criteria.
     * The call-back returns true if the enumeration should continue.
     * @note Reported entities still may not be local objects. This is in
     * those cases when entities belong to the area partially only.
     */
    virtual bool entity_reported
        (
        WeakPointer<Entity> entity
        ) = 0;

protected:

    //
    // Sector enumerator interface implementation.
    //

    /**
     * Enumerates entities from the current sector.
     */
    virtual void sector_visited();

public:

    ConvexArea                  area;
        /**< The area. In layer 2D coordinates. */
    PVlInt32                    entity_flags;
        /**< Report only those entities that match these entity flags.
             See @c Sector::enumerate_entities() to get information on
             how the flags are interpreted. */
    ReportedList                reported;
        /**< List of reported entities. Used by the sequential query
             to ensure the "at most once" semantics. */

    }; // SpatialQuery


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_SPATIAL_QUERY_H
