/** @file account_object.h
 * Account object.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_ACCOUNT_OBJECT_H
#define DEMO_ACCOUNT_OBJECT_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#ifndef DEMO_PLAYER_ACTION_H
#include "lib/shared/player_action.h"
#endif

#ifndef DEMO_CLIENT_SECTOR_H
#include "lib/shared/client_sector.h"
#endif

#ifndef DEMO_CONSOLE_H
#include "lib/shared/console.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Account object.
 * AccountObject is an interface visible by a client to access the account 
 * object's implementation in server lib.
 */
class AccountObject : public Massiv::Core::Lib::AccountObjectInterface
    {
public:

    //--------------------------------------------------------------------------
    // Interface for CLIENT -> SERVER communication.
    //--------------------------------------------------------------------------

public:

    /**
     * Request replication of @a player to client of this account.
     */
    virtual void request_player_replication
        (
        WeakPointer< ClientEntity > player
        ) = 0;
    
    /**
     * Handle move action requested by client.
     */
    virtual void handle_move_action
        (
        const PlayerMoveAction & action
        ) = 0;

    /**
     * Handle item action requested by client.
     */
    virtual void handle_item_action
        (
        const PlayerItemAction & action
        ) = 0;

    /**
     * Say message.
     */
    virtual void say
        (
        const std::string & message
        ) = 0;

    /**
     * Is this root account?
     */
    virtual bool is_root_account() = 0;

    /**
     * Process account object specific @c command.
     * Returns error descriptions that occured during command parsing.
     */
    virtual PValueArray<SString> process_account_command
        (
        const std::string &         command,
        Console::Environment &      env        
        ) = 0;
    
    /**
     * Complete account object specific @c partial_command.
     */
    virtual PValueArray<SString> auto_complete_account_command
        (
        const std::string & partial_command
        ) = 0;
    
    /**
     * Process node object specific @c command.
     * Because of security reasons client can not connect @c node directly.
     * Returns error descriptions that occured during command parsing.
     * @note Fails if @c node refers to a client node. Direct calls
     * from client to its client node are okay.
     */
    virtual PValueArray<SString> process_node_command
        (
        Remote<NodeObject>          node,
        const std::string &         command,
        Console::Environment &      env        
        ) = 0;

    /**
     * Complete node object specific @c partial_command.
     * Because of security reasons client can not connect @c node directly.
     * @note Fails if @c node refers to a client node. Direct calls
     * from client to its client node are okay.
     */
    virtual PValueArray<SString> auto_complete_node_command
        (
        Remote<NodeObject>  node,
        const std::string & partial_command
        ) = 0;

    /**
     * Process node object specific @c command on the node where @c object
     * currently lives.
     * Because of implementation, the method is not able to return results
     * nor environment.
     * Operates asynchronously.
     */
    virtual void process_node_command_at_object
        (
        Remote<Object>          object,
        const std::string &     command
        ) = 0;

    /**
     * Tests if client's console can connect to @c node.
     */
    virtual bool can_connect_console_to
        (
        Remote<NodeObject> node
        ) = 0;

    // FIXME?
    // boovie: Maybe modify_material_command should be removed
    // from AccountObject as modify_elevation_command was
    // (and sending commands should be done through console only).
    // If there is anybody who has much time and nothing to do
    // and who likes parsing args from a string, he can do it.

    /** Modify material command. */
    virtual void modify_material_command
        (
        const SVector2 &            center,
        Float32                     radius,
        ClientSector::MaterialId    material_id
        ) = 0;

    /** Delete entity command. */
    virtual void delete_entity_command
        (
        const SString & entity_object_id
        ) = 0;

    /** Edit entity command. */
    virtual void edit_entity_command
        (
        const SString & entity_object_id,
        const SString & method_name,
        const SString & parameters
        ) = 0;

    /** Create entity command. */
    virtual void create_entity_command
        (
        const SString & parameters
        ) = 0;

public:
    
    //--------------------------------------------------------------------------
    // Interface for SERVER -> CLIENT communication.
    //--------------------------------------------------------------------------
    
public:
    
    /**
     * Log message to client that owns this account object.
     * This method should be used to log messages notifying
     * the client of some ingame events.
     * @param source_entity Entity that generated the message.
     *      Null if the message is generic ingame message.
     * @param message Message itself.
     * @param message_color Color that should be used to draw the message.
     */
    virtual void log_game_message
        (
        WeakPointer< ClientEntity >     source_entity,
        const std::string &             message,
        const SColor &                  message_color
        ) = 0;

    /**
     * Print message to client's console.
     */
    virtual void print_to_console
        (
        WeakPointer< NodeObject >   sender,
        const std::string &         message,
        const SColor &              message_color = SColor::YELLOW
        ) = 0;

    }; // AccountObject

    
} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_ACCOUNT_OBJECT_H
