/** @file client_entity.h
 * ClientEntity.
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CLIENT_ENTITY_H
#define DEMO_CLIENT_ENTITY_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#ifndef DEMO_GAME_DEFS_H
#include "lib/shared/game_defs.h"
#endif

#ifndef DEMO_ABILITY_H
#include "lib/shared/ability.h"
#endif

#ifndef DEMO_ENTITY_POSITION_H
#include "lib/shared/entity_position.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Client visible part of entity.
 */
class ClientEntity : public Object
    {
public:
    
    MASSIV_OBJECT( ClientEntity );

public:

    /**
     * Entity type.
     *
     * This affects client-side presentation of the entity.
     *
     * @todo Merge with Entity::EntityClass? Currently it contains nearly
     *       the same info but it's used for different purposes.
     */
    enum Type
        {
        TYPE_DEFAULT,
            /**< Generic entity. Default entity type. Invisible. */

        TYPE_LOGIC,
            /**< Logic-only. No visualisation on client. */

        TYPE_PHYSICAL,
            /**< Generic physical entity. */

        TYPE_MONSTER,
            /**< Generic monster entity. */

        TYPE_PLAYER,
            /**< Player entity. */

        TYPE_LAMP,
            /**< Lamp entity. */

        TYPE_COLOR_GENERATOR,
            /**< Visible logic entity. */

        }; // enum Type
    
public:

    /**
     * Initialize the object.
     */
    void initialize
        (
        Pointer<Object> the_entity
        );

public:

    /**
     * Check whether the entity support @a ability.
     */
    bool supports_ability
        (
        AbilityId   ability
        ) const;
    
    /**
     * Convert #AbilityId to string.
     */
    static std::string ability_to_string
        (
        AbilityId   ability
        );

    /**
     * Get current value of @a ability.
     * If the ability is not supported by the entity, zero is returned.
     */
    Int32 get_ability_value
        (
        AbilityId   ability
        ) const;
    
    /**
     * Get current upper limit of @a ability.
     * If the ability is not supported by the entity, zero is returned.
     */
    Int32 get_ability_upper_limit
        (
        AbilityId   ability
        ) const;

public:

    /**
     * Set value of @a ability.
     * The value is clampled against upper limit of the ability.
     * If the ability is not supported by the entity, nothing is done.
     */
    void set_ability_value
        (
        AbilityId   ability,
        Int32       new_value
        );
    
    /**
     * Add @a delta to current value of the ability.
     * The @a delta may be negative.
     * The value is clampled against upper limit of the ability.
     * If the ability is not supported by the entity, nothing is done.
     */
    void add_ability_value
        (
        AbilityId   ability,
        Int32       delta
        );
    
    /**
     * Set upper limit of @a ability.
     * Current value of the ability is clampled against
     * the new upper limit.
     * If the ability is not supported by the entity, nothing is done.
     */
    void set_ability_upper_limit
        (
        AbilityId   ability,
        Int32       new_upper_limit
        );

    /**
     * Add @a delta to current upper limit of @a ability.
     * The @a delta may be negative.
     * Current value of the ability is clampled against
     * the new upper limit.
     * If the ability is not supported by the entity, nothing is done.
     */
    void add_ability_upper_limit
        (
        AbilityId   ability,
        Int32       delta
        );

public:

    /**
     * Should this entity have a visual representation?
     *
     * If @c admin_mode is set to true some of the invisible entities will
     * be drawn too.
     */
    bool is_visual
        (
        bool admin_mode
        ) const;

    /**
     * Construct model tag string for this entity.
     *
     * This combines #model_tags string with tags generated from entity
     * #type and current state.
     *
     * @pre Entity #is_visual().
     */
    std::string get_model_tags() const;
    
public:

    typedef PArray< Ability >           Abilities;

    PPointer< Object >                  entity;
        /**< Reference to the corresponding Entity object.
             Since Demo::Lib::Entity is from a different class kind, generic
             pointer to object must be used. */

    Abilities                           abilities;
        /**< Array of abilities of the entity. If the entity doesn't
             support any ability, this array is empty. If the entity support
             only a subset of abilities, this array is smaller than
             @c ABILITY_COUNT. */

    EntityPosition                      current_position;
        /**< Current entity position. */

    PPointer< ClientSector >            current_sector;
        /**< Current sector the entity is linked to. */

    PPointer< ClientItemSlotContainer > inventory;
        /**< Inventory, if the entity has one. */

    PEnum< Type >                       type;
        /**< Entity type. */

    PString                             name;
        /**< Name of the entity. Empty if the entity has no name. */

    PString                             model_tags;
        /**< Optional whitespace-separated tags used to create entity model. */

    PPointer< ClientLight >             light;
        /**< Optional omnidirectional light source parameters.
             If #light is @c NULL, entity will not emit any light. */

    PColor                              emissive_color;
        /**< Entity self-lighting.
             It will be added to color received from light sources. */

    PBool                               stopped;
        /**< Set to true if entity is not moving. Utilized by prediction
             code on clients. */

    }; // ClientEntity


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_CLIENT_ENTITY_H
