/** @file client_node_interface.h
 * Client node object interface.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CLIENT_NODE_INTERFACE_H
#define DEMO_CLIENT_NODE_INTERFACE_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#ifndef DEMO_NODE_OBJECT_H
#include "lib/shared/node_object.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * This is the only public interface for accessing client node object.
 * Defines console "push" interface (ie. call-backs used for printing text
 * to console, notifying about completed commands, etc.).
 * Console "pull" interface (ie. submitting new commands, etc.) is
 * implemented by AccountObject.
 */
class ClientNodeInterface : public NodeObject
    {
public:

    /**
     * Welcome message from account object.
     * This function is called by #AccountObject object notifying
     * the client that it was successfully connected to simulation.
     * The main purpose of this method is to link client node
     * with its account object, because account object is the only
     * way how the client can communicate with the simulation.
     * @param account Pointer to #AccountObject of the client.
     */
    virtual void welcome_to_simulation
        (
        WeakPointer< AccountObject > account
        ) = 0;
    
    /**
     * This function is called when account object of the client
     * have selected a player object to control.
     *
     * Before client can control its player, the player object must
     * be selected by client's account object. Then the client manipulates
     * the player object using its account object.
     * This function is used to notify the client that a player object
     * was successfully selected.
     *
     * @param player Pointer to #ClientEntity object. Null if currently selected
     *      player was deselected by the account object.
     */
    virtual void player_selected
        (
        WeakPointer< ClientEntity > player
        ) = 0;

public:

    /**
     * Callback used by account object to notify the client
     * of some ingame message.
     * @param source_entity Entity that generated the message.
     *      Null if the message is generic ingame message.
     * @param message Message itself.
     * @param message_color Color that should be used to draw the message.
     */
    virtual void handle_game_message
        (
        WeakPointer< ClientEntity > source_entity,
        const std::string &         message,
        const SColor &              message_color
        ) = 0;

public:

    /**
     * Prints @c text to client's console.
     */
    virtual void print_to_console
        (
        WeakPointer<NodeObject> sender,
        const std::string &     message,
        const SColor &          message_color = SColor::YELLOW
        ) = 0;
    
    /**
     * Prints @c text to client's chat window.
     */
    virtual void print_to_chat_window
        (
        const std::string &     message,
        const SColor &          message_color = SColor::YELLOW
        ) = 0;
        
    }; // ClientNodeInterface

} // namespace Demo::Lib
} // namespace Demo


#endif // DEMO_CLIENT_NODE_INTERFACE_H      
