/** @file client_sector.h
 * ClientSector.
 */

/*
 * begin        : Aug 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CLIENT_SECTOR_H
#define DEMO_CLIENT_SECTOR_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SECTOR_ADDRESS_H
#include "lib/shared/sector_address.h"
#endif



namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Client visible part of sector.
 */
class ClientSector : public Object
    {
public:
    
    MASSIV_OBJECT( ClientSector );

    /**
     * Static parameters.
     */
    enum Parameters
        {
        NUM_X_TILES =   8,
            /**< Number of tiles in a "row". */
        NUM_Y_TILES =   8,
            /**< Nummber of tiles in a "column". */
        SECTOR_X_SIZE = 8,
            /**< How many units Sector occupies in layer coordinates in 
                 east direction. */
        SECTOR_Y_SIZE = 8,
            /**< How many units Sector occupies in layer coordinates in
                 north direction. */
        }; // Parameters

    static const Float32 min_height;
        /**< Minimum terrain height. */

    static const Float32 max_height;
        /**< Maximum terrain height. */

    /**
     * Height quantization traits.
     */
    struct HeightTraits
        {
        static const Float32 min_value;
        static const Float32 max_value;
        enum { bits = 8 };
        };
        
    /**
     * Identifies what kind of elevation map should be accessed.
     * Used as a parameter to elevation map-related methods.
     */
    enum ElevationMapKind
        {
        KIND_ELEVATION_MAP =        0,
            /**< Access @c elevation_map. */
        KIND_DELTA_ELEVATION_MAP =  1,
            /**< Access @c delta_elevation_map. */
        }; // ElevationMapKind

    /**
     * Identifies type of the operation to be applied on specified
     * elevation map.
     * Used as a parameter to elevation map-related methods.
     */
    enum ElevationMapOperation
        {
        OPERATION_SET =         0,
            /**< The new elevation overwrites current value. */
        OPERATION_ADD =         1,
            /**< The new elevation is added to the current value. */
        OPERATION_MULTIPLY =    2
            /**< The current value is multiplied by the new elevation. */
        }; // ElevationMapOperation

    /**
     * Function that is applied to tiles in a center-point's neighborhood
     * to modify an elevation map.
     * Each function is parametrized by @c radius (horizontal influence)
     * and @c strength (vertical influence).
     */
    enum ElevationMapFunction
        {
        FUNCTION_CIRCLE =           0,
            /**< Apply @a strength in circular region with given @a radius. */
        FUNCTION_FUZZY_CIRCLE =     1,
            /**< Apply @a strength at @a origin and zero strength at edge of
                 circle with given @a radius. */
        FUNCTION_GAUSSIAN =         2,
            /**< Symetric gaussian with @c radius standard deviation. */
        }; // ElevationMapFunction
    
    /**
     * Terrain height type.
     */
    typedef SFloatQ< Float32, HeightTraits > Height;

    /**
     * Elevation map (heightmap).
     */
    typedef PValueArray< Height > ElevationMap;

public:
    
    enum
        {
        MATERIAL_ID_NUM_BITS = 8,
            /**< Number of bits used to serialize #MaterialId. */
        };
    
    /**
     * Material id.
     * Identifies a material in material database.
     */
    typedef SInt< UInt32, MATERIAL_ID_NUM_BITS > MaterialId;
    
    /**
     * Property material id.
     */
    typedef PInt< UInt32, MATERIAL_ID_NUM_BITS > PMaterialId;
    
    /**
     * Material map.
     */
    typedef PValueArray< MaterialId > MaterialMap;
    
    /**
     * Material ids of special materials interpreted by the lib.
     */
    enum SpecialMaterialId
        {
        MATERIAL_OBSTACLE = 1 << MATERIAL_ID_NUM_BITS,
            /**< Obstacle. */
        MATERIAL_DEFAULT = 0
            /**< Default material. */
        }; // SpecialMaterialId

public:

    enum TileFlagsEnum
        {
        TILE_MARK               = 0x01,
            /**< "Mark" tile. */

        TILE_FLAGS_BIT_COUNT    = 4,
            /**< Number of serialized per-tile flag bits. */
        };

    /**
     * Tile flags.
     * Combination of TileFlagsEnum bits.
     */
    typedef SInt< UInt32, TILE_FLAGS_BIT_COUNT > TileFlags;

    /**
     * Tile flag map.
     */
    typedef PValueArray< TileFlags > TileFlagMap;

public:

    /**
     * Initialize the object.
     *
     * Elevation and other maps will not be resized by initialize().
     * This is done by #ClientSector owner - #Sector.
     */
    void initialize();
    
public:

    /**
     * Convert #ElevationMapKind to string.
     */
    static std::string map_kind_to_string
        (
        ElevationMapKind map_kind
        );

    /**
     * Convert @a str to #ElevationMapKind.
     * @return True if the conversion was successfull, false otherwise.
     *     If successfull, @a map_kind holds extracted #ElevationMapKind.
     */
    static bool string_to_map_kind
        (
        const std::string & str,
        ElevationMapKind &  map_kind
        );
    
    /**
     * Convert #ElevationMapFunction to string.
     */
    static std::string map_function_to_string
        (
        ElevationMapFunction map_function
        );

    /**
     * Convert @a str to #ElevationMapFunction.
     * @return True if the conversion was successfull, false otherwise.
     *     If successfull, @a map_function holds extracted #ElevationMapFunction.
     */
    static bool string_to_map_function
        (
        const std::string &     str,
        ElevationMapFunction &  map_function
        );
    
    /**
     * Convert #ElevationMapOperation to string.
     */
    static std::string map_operation_to_string
        (
        ElevationMapOperation map_operation
        );

    /**
     * Convert @a str to #ElevationMapOperation.
     * @return True if the conversion was successfull, false otherwise.
     *     If successfull, @a map_operation holds extracted #ElevationMapOperation.
     */
    static bool string_to_map_operation
        (
        const std::string &     str,
        ElevationMapOperation & map_operation
        );

public:

    SectorAddress   sector_address;
        /**< This sector's address. */

    ElevationMap    elevation_map;
        /**< Elevation map.
             NUM_X_TILES * NUM_Y_TILES entries - heights of sector vertices,
             excluding eastern and northern "borders". Those heights must be
             read from sector neighbors.
             This is the "original" terrain. */

    ElevationMap    delta_elevation_map;
        /**< Map that represents temporary differences to @c elevation_map,
             such as local terrain deformations caused by explosions, etc.
             @c delta_elevation_map is added to @c elevation_map to get
             actual terrain elevation.
             The values are periodically multiplied by @c Sector::delta_map_decay_factor
             so that their influence to actual terrain elevation decreases
             after each update iteration. */

    MaterialMap     material_map;
        /**< Material map.
             Materials are defined at sector vertices, not tiles. The array
             contains NUM_X_TILES * NUM_Y_TILES entries and does not include
             materials at eastern and northern "borders". */

    TileFlagMap     tile_flag_map;
        /**< Per-tile flags.
             NUM_X_TILES * NUM_Y_TILES per-tile entries. */

    }; // ClientSector


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_CLIENT_SECTOR_H
