/** @file console.cpp
 * Console implementation.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "lib/shared/console.h"
#include "lib/shared/node_object.h"
#include "lib/shared/account_object_rpc.h"
#include <map>
#include <sstream>
#include <iomanip>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();



// Implementation of Console
// ~~~~~~~~~~~~~~~~~~~~~~~~~

Console::Console()
    {
    }


Console::~Console()
    {
    }


namespace {

/**
 * FIXME - okomentovat
 */
static void get_variable_name
    (
    const std::string &             temp_command,
    const std::string::size_type    start,
    const char                      end_char,
    std::string &                   var_name,
    std::string::size_type &        end
    )
    {
    std::string::size_type pos = temp_command.find_first_of( end_char, start );
    if( pos == std::string::npos )
        {
        // End of string.
        var_name = temp_command.substr( start );
        end = temp_command.size() - 1;
        return;
        }
    var_name = temp_command.substr( start, pos - start );
    end = pos - 1;    
    }

} // anonymous namespace


void Console::process_command
    (
    const Remote<AccountObject> &   account,
    const std::string &             command,
    bool                            auto_complete,
    PValueArray<SString> &          result,
    Environment &                   env
    )
    {
    // Initialize.
    
    PValueArray<SString>    dummy_array;
    PValueArray<SString> &  completions = auto_complete ? result : dummy_array;
    PValueArray<SString> &  errors = auto_complete ? dummy_array : result;

    Serializer::Description desc;

    std::string temp_command = command;
    const std::string result_substring = " > ";
    std::string::size_type result_pos = temp_command.find( result_substring );
    if( result_pos != std::string::npos )
        {
        temp_command = temp_command.substr( 0, result_pos );
        }
    
    // Replace occurence of "$var_name or ${var_name}".
    
    typedef std::string::size_type size_type; 
    size_type pos = size_type( -1 );
    while( ( pos = temp_command.find( "$", pos + 1 ) ) != std::string::npos )
        {
        // Replace next occurence.
        char end_char = ' ';
        size_type start = pos + 1;
        if( temp_command[ pos + 1 ] == '{' )
            {
            start = pos + 2;
            end_char = '}';
            }
        size_type end;
        std::string var_name;
        get_variable_name( temp_command, start, end_char, var_name, end );

        // Get replacement string.
        Console::Environment::Iterator it = env.find( var_name );
        std::string value_string = "UNKNOWN_VALUE";
        if( it != env.end() )
            {
            value_string = it->second;
            }
        // Replace.
        size_type length;
        length = end - pos + 1;
        if( start - pos == 2 )
            {
            // '{' case.
            length++;
            }
        temp_command.replace( pos, length, value_string );
        }
        
    std::istringstream      command_stream( temp_command );
    TextReader              tr( command_stream );

    // Parse command.

    SString cmd_string;
    
    try
        {
        tr.next_line();
        cmd_string.text_read( tr, desc );
        }
    catch( Exception & e )
        {
        if( !auto_complete )
            {
            std::ostringstream os;
            
            os << "Error parsing command (" << e.what() << ")";
              
            errors.push_back( os.str() );
            return;
            }
        else
            {
            complete_command
                (
                account->sync_is_root_account(),
                cmd_string,
                completions
                );

            return;
            }
        }
    
    if( cmd_string[ 0 ] != '/' && !auto_complete )
        {
        errors.push_back( SString( "Expected \"/<command>\"." ) );
        return;
        }

    // Lookup command list.

    CommandList::Iterator c_it = command_list.find( cmd_string.read().substr( 1 ) );
    
    if( c_it == command_list.end() )
        {
        if( !auto_complete )
            {
            // Log "Unknown command." error.
             
            errors.push_back( get_unknown_command_tag() );
            return;
            }
        else
            {
            std::string dummy_string;

            // Test eof.
            
            tr.line() >> dummy_string;

            if( !tr.line().fail() )
                {
                // @c cmd_string seems to have some arguments, but such command
                // is not recognized by this console so do not offer any completions.

                return;
                }
            
            // Complete command.

            complete_command
                (
                account->sync_is_root_account(),
                cmd_string,
                completions
                );

            return;
            }
        }

    // Copy command.
    // We need to operate on our copy of the command so that
    // other method instances may run in parallel (we will be writing to
    // command arguments once they are parsed).

    ConsoleCommand cmd = c_it->second;

    // Test account privileges.

    if( cmd.flags & ConsoleCommand::FLAGS_ROOT_ACCOUNT )
        {
        if( !account->sync_is_root_account() )
            {
            errors.push_back( SString( "Access denied." ) );
            return;
            }
        }

    // Parse arguments and write their values to @c cmd.

    ConsoleCommand::ArgumentList::iterator a_it = cmd.args.begin();
    std::size_t arg_no = 0;

    while( a_it != cmd.args.end() )
        {
        CommandArgument &   arg_description = *a_it;
        std::size_t         arg_position = tr.current_position();
        
        ++arg_no;

        try
            {
            if( arg_description.flags & CommandArgument::FLAGS_ARG_OPTIONAL )
                {
                try
                    {
                    fill_default_value( arg_description );
                    }
                catch( Exception & )
                    {
                    }
                }
            
            parse_argument( arg_description, tr, desc );
            
            if( arg_no == cmd.args.size() && auto_complete )
                {
                // This is the last argument.
                // Trigger its auto-completion even if it was parsed
                // successfully.

                // FIXME: korektne, ne pres vyjimku

                throw Exception( "<trigger-auto-completion>" );
                }
            }
        catch( Exception & e )
            {
            if( !( arg_description.flags & CommandArgument::FLAGS_ARG_OPTIONAL && !auto_complete ) )
                {
                if( !auto_complete )
                    {
                    std::ostringstream os;

                    os << "Error parsing argument #" << arg_no << " (" << e.what() << ")";
                    
                    errors.push_back( os.str() );
                    return;
                    }
                else
                    {
                    std::size_t i;

                    // Get possible completions for argument that failed
                    // to parse successfully.

                    std::istringstream      command_tail( temp_command.substr( arg_position ) );
                    std::string             prefix( temp_command.substr( 0, arg_position ) + " " );

                    try
                        {
                        hope( cmd.complete );
                        cmd.complete
                            (
                            CommandContext( this, account, cmd.args, arg_no ),
                            command_tail,
                            completions
                            );
                        
                        // Extend argument completions to the whole command completions.

                        for( i = 0; i < completions.size(); i++ )
                            {
                            completions( i ) = prefix + completions[ i ].read();
                            }
                        }
                    catch( ... )
                        {
                        completions.clear();
                        }

                    // Finished.

                    return;
                    }
                }
            }
            
        ++a_it;
        }

    // All arguments have been parsed successfully.

    if( auto_complete )    
        {
        // There is nothing to complete.
        return;
        }
    
    // Test eof.

    std::string dummy_string;

    tr.line() >> dummy_string;

    if( !tr.line().fail() )
        {
        errors.push_back( SString( "Too many arguments." ) );
        return;
        }

    // Execute command.
    // All objects involved in the call must be pinned.
    // Note that another instances of command may run simultaneously.

    env[ std::string( "$" ) ] = std::string( "" );
    std::string result_string = "";
    try
        {
        hope( cmd.execute );
        CommandContext cc( this, account, cmd.args, arg_no );
        
        std::string print_command = std::string( "Console execution: " ) + 
            temp_command + ( result_pos != std::string::npos ? command.substr
            ( result_pos ) : "" );
            
        cc.print( print_command, SColor::YELLOW );
            
        cmd.execute( cc );
        result_string = cc.result;
        }
    catch( Exception & e )
        {
        std::ostringstream os;

        os << "Exception during command processing:\n" << e.what();

        errors.push_back( os.str() );
        return;
        }
    
    // Set result to environment.
    env[ std::string( "$" ) ] = result_string;
    if( result_pos != std::string::npos )
        {
        // Set result to other environment variable.
        std::string variable = command.substr( result_pos + result_substring.size() );
        if( variable[ 0 ] == '$' )
            {
            // Environment variable must begin with '$'
            variable.erase( 0, 1 ); // Delete char '$'.
            if( variable[ 0 ] == '{' )
                {
                variable.erase( 0, 1 ); // Delete char '{'.
                }
            // Delete spaces from the end.
            while( variable[ variable.size() - 1 ] == ' ' )
                {
                variable.erase( variable.size() - 1, 1 ); // Delete char ' '.
                }
            // Delete char '}'.
            if( variable[ variable.size() - 1 ] == '}' )
                {
                variable.erase( variable.size() - 1, 1 ); // Delete char '}'.
                }                            
            env[ variable ] = result_string;            
            }
        }
    }


std::string Console::get_unknown_command_tag()
    {
    return "$@UNKNOWN_COMMAND";
    }


bool Console::is_unknown_command_error
    (
    const PValueArray<SString> &    errors
    )
    {
    return errors.size() == 1 && errors[ 0 ] == Console::get_unknown_command_tag();
    }


std::string Console::print_command_help()
    {
    std::ostringstream  ss;
    
    ss.setf( std::ios_base::left );

    CommandList::Iterator c_it = command_list.begin();

    while( c_it != command_list.end() )
        {
        const std::string &     spelling = c_it->first;
        const ConsoleCommand &  cmd = c_it->second;

        ss << std::setw( 1 ) << "/" << std::setw( 13 ) << spelling;
        ss << " ... " << cmd.description << "\n";
        
        ++c_it;
        }
    
    return ss.str();
    }


std::string Console::print_command_help
    (
    const std::string & spelling
    )
    {
    std::ostringstream      ss;
    CommandList::Iterator   c_it = command_list.find( spelling );

    if( c_it == command_list.end() )
        {
        return "Unknown command.";
        }
    
    const ConsoleCommand &                          cmd = c_it->second;
    ConsoleCommand::ArgumentList::const_iterator    a_it = cmd.args.begin();

    ss << cmd.description << "\n";
        
    while( a_it != cmd.args.end() )
        {
        const CommandArgument & arg = *a_it;
                    
        ss << std::setw( 2 ) << a_it - cmd.args.begin() << " ";
        ss << arg.description << " [ " << arg.flags_to_string() << " ]";
        
        ss << "\n";

        ++a_it;
        }

    return ss.str();
    }


void Console::complete_command
    (
    bool                    is_root_account,
    const std::string &     command,
    PValueArray<SString> &  completions
    )
    {
    CommandList::Iterator   c_it;
    std::string             mask;
    
    if( command.size() )
        {
        if( command[ 0 ] == '/' )
            {
            mask = command.substr( 1 ) + "*";    
            }
        else
            {
            mask = "*";
            }
        }
    else
        {
        mask = "*";
        }

    for( c_it = command_list.begin(); c_it != command_list.end(); ++c_it )
        {
        ConsoleCommand & cmd = c_it->second;
    
        if( cmd.flags & ConsoleCommand::FLAGS_ROOT_ACCOUNT )
            {
            if( !is_root_account )
                {
                continue;
                }
            }

        if( Wildcards::match( mask.c_str(), cmd.spelling.c_str() ) )
            {
            completions.push_back( "/" + cmd.spelling );
            }
        }
    }


void Console::complete_args
    (
    const CommandContext &  context,
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    Console &               console = *context.console;
    const CommandArgument & arg = context.get_last_arg();

    if( arg.flags & CommandArgument::FLAGS_BOOLEAN )
        {
        console.complete_boolean( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_STRING )
        {
        console.complete_string( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_OBJECT_ID )
        {
        console.complete_object_id( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_STATISTICS )
        {
        console.complete_registry_node( stream, result, std::string( "Statistics")  );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_CONFIG_VARIABLE )
        {
        console.complete_config_variable( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_CONFIG_VARIABLE_SET )
        {
        console.complete_config_variable( stream, result, std::string( "Settings" ), true );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_CLASS_NAME )
        {
        console.complete_class_name( stream, result, true );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_WILDCARD )
        {
        console.complete_wildcard( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_PROPERTY )
        {
        console.complete_property( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_METHOD )
        {
        console.complete_method( stream, result );
        }
    else
    if( arg.flags & CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT )
        {
        console.complete_class_name( stream, result, false );
        }
    }


void Console::parse_argument
    (
    CommandArgument &           arg_description,
    TextReader &                tr,
    Serializer::Description &   desc
    )
    {
    PValueArray<SString> dummy_array;
    
    if( arg_description.flags & CommandArgument::FLAGS_INTEGER )
        {
        arg_description.bound_integer.text_read( tr, desc );
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_BOOLEAN )
        {
        arg_description.bound_boolean.text_read( tr, desc );
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_FLOAT )
        {
        arg_description.bound_float.text_read( tr, desc );
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_STRING ||
        arg_description.flags & CommandArgument::FLAGS_WILDCARD ) 
        {
        arg_description.bound_string.text_read( tr, desc );
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_OBJECT_ID )
        {
        arg_description.bound_object_id.text_read( tr, desc );
        arg_description.bound_object = 
            Global::garbage_collector().find_local_object
                ( arg_description.bound_object_id );
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_STATISTICS )
        {
        ParseRegnodeResult      result;
        SString                 param;

        param.text_read( tr, desc );

        parse_or_complete_regnode
            (
            param,
            false,
            VariantPointer( &result ),
            dummy_array,
            "Statistics"
            );

        arg_description.bound_regnode = result.regnode;
        arg_description.bound_string = param;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_CONFIG_VARIABLE )
        {
        ParseCvarResult         result;
        SString                 param;

        param.text_read( tr, desc );

        parse_or_complete_cvar
            (
            param,
            false,
            VariantPointer( &result ),
            dummy_array
            );

        arg_description.bound_cvar = result.cvar;
        arg_description.bound_string = param;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_CONFIG_VARIABLE_SET )
        {
        ParseCvarResult         result;
        SString                 param;

        param.text_read( tr, desc );

        parse_or_complete_cvar
            (
            param,
            false,
            VariantPointer( &result ),
            dummy_array,
            "Settings",
            true
            );

        arg_description.bound_cvar = result.cvar;
        arg_description.bound_string = param;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_CLASS_NAME )
        {
        ParseMetaobjectResult   result;
        SString                 class_name;

        class_name.text_read( tr, desc );

        parse_or_complete_metaobject
            (
            class_name,
            false,
            VariantPointer( &result ),
            dummy_array,
            true
            );

        arg_description.bound_metaobject = result.metaobject;
        arg_description.bound_string = class_name;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT )
        {
        ParseMetaobjectResult   result;
        SString                 class_name;

        class_name.text_read( tr, desc );

        parse_or_complete_metaobject
            (
            class_name,
            false,
            VariantPointer( &result ),
            dummy_array,
            false
            );

        arg_description.bound_metaobject = result.metaobject;
        arg_description.bound_string = class_name;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_PROPERTY )
        {
        ParsePropertyResult     result;
        SString                 scoped_member;

        scoped_member.text_read( tr, desc );

        parse_or_complete_object_member
            (
            scoped_member,
            PARSE_PROPERTY,
            false,
            VariantPointer( &result ),
            dummy_array
            );

        // Note: By filling @c bound_object we ensure that the object that
        //       owns the property will be pinned until command completes.

        arg_description.bound_property = result.property;
        arg_description.bound_object = result.object;
        }
    else
    if( arg_description.flags & CommandArgument::FLAGS_METHOD )
        {
        ParseMethodResult       result;
        SString                 scoped_member;

        scoped_member.text_read( tr, desc );

        parse_or_complete_object_member
            (
            scoped_member,
            PARSE_METHOD,
            false,
            VariantPointer( &result ),
            dummy_array
            );
        
        // Note: By filling @c bound_object we ensure that the object that
        //       the method will be called on will be pinned until command
        //       completes.
        
        arg_description.bound_component = result.component;
        arg_description.bound_metaobject = result.metaobject;
        arg_description.bound_method = result.method;
        arg_description.bound_object = result.object;
        }
    else
        {
        panic( "Unknown argument type." );
        }
    }


void Console::fill_default_value
    (
    CommandArgument & cmdarg
    )
    {
    hope( cmdarg.flags & CommandArgument::FLAGS_ARG_OPTIONAL );
    
    Serializer::Description desc;
    std::istringstream      command_stream( cmdarg.default_value );
    TextReader              tr( command_stream );
    
    if( !cmdarg.default_value.size() )
        {
        return;
        }
    
    tr.next_line();
    parse_argument( cmdarg, tr, desc );
    }


void Console::parse_or_complete_cvar
    (
    const std::string &     param,
    bool                    auto_complete,
    const VariantPointer &  parse_result,
    PValueArray<SString> &  completions,
    const std::string &     subsystem,
    bool                    show_value
    )
    {
    Massiv::Core::Registry & registry = Massiv::Core::Global::registry();
    Registry::RegistryNode * node = NULL;
    ConfigVariable *         cvar;

    std::string              variable( param );
    unsigned int             pos;

    Registry::SearchContext  sc( &registry );

    try
        {
        if( auto_complete )
            {
            throw ExCVarNotFound( param.c_str() );
            }

        pos = variable.find_last_of( "/" );
        if( pos != variable.npos )
            {
            node = registry.find_node_for_path( variable.substr( 0, pos ) );
            variable = variable.substr( pos + 1, variable.size() - pos - 1 );
            }
        else
            {
            node = registry.find_node_for_path( "" );
            }
        if( variable.size() == 0 )
            {
            throw ExCVarNotFound( param.c_str() );
            }
        registry.initialize_search_context( sc, *node );

        if( !registry.find_variable( variable.c_str(), sc, *node ) )
            {
            throw ExCVarNotFound( param.c_str() );
            }
        else
            {
            cvar = &sc.get_variable();
            ParseCvarResult * r = variant_cast<ParseCvarResult *>( parse_result );
            r->cvar = cvar;
            }
        }
    catch( ... )
        {
        if( !auto_complete )
            {
            throw ExCVarNotFound( param );
            }
        else
            {
            std::string     look_ahead = param;
            std::string     node_name;

            std::string             base_registry_node_name = subsystem;
            ConfigVariable *        cvar;

            Registry::SearchContext sc( &registry );
         
            look_ahead += "*";

            node = registry.find_node_for_path( base_registry_node_name );
            registry.initialize_search_context( sc, *node );
              
            while( registry.find_variable( 0, sc, *node ) )
                {
                cvar = &sc.get_variable();
                std::string match = "";
            
                match += sc.get_path() + "/" + sc.get_name();

                if( Wildcards::match( look_ahead.c_str(), match.c_str() ) )
                    {
                    std::ostringstream ss;
            
                    ss << sc.get_path() << "/" << sc.get_name();

                    if( show_value )
                        {
                        ss << " " << quote_const_string( cvar->read_string() );
                        }
                    completions.push_back( ss.str() );
                    }
                }
            }
        }
    }


void Console::parse_or_complete_regnode
    (
    const std::string &     param,
    bool                    auto_complete,
    const VariantPointer &  parse_result,
    PValueArray<SString> &  completions,
    const std::string &     subsystem
    )
    {
    Massiv::Core::Registry & registry = Massiv::Core::Global::registry();
    Registry::RegistryNode * regnode = NULL;
    std::string              base_registry_node_name;

    try
        {
        if( auto_complete )
            throw ExIllegalRegistryNode( param );
        if( param == "all" )
            {
            base_registry_node_name = subsystem;
            }
        else
            {
            if( param.size() > 0 )
                base_registry_node_name = param;
            else
                throw ExIllegalRegistryNode( param );
            }
        regnode = registry.find_node_for_path( base_registry_node_name );
        ParseRegnodeResult * r = variant_cast<ParseRegnodeResult *>( parse_result );
        r->regnode = regnode;
        }
    catch( ... )
        {
        if( !auto_complete )
            {
            throw ExIllegalRegistryNode( param );
            }
        else
            {
            std::string     look_ahead;
            std::string     node_name;

            base_registry_node_name = subsystem;
            Registry::RegistryNode *        node;
            Registry::RegistryNode *        current_node = NULL;

            Registry::SearchContext         sc( &registry );
         
            look_ahead = param + "*";

            if( Wildcards::match( look_ahead.c_str(), "all" ) && look_ahead.size() > 1 )
                {
                completions.push_back( SString( "all") );
                return;
                }
            
            node = registry.find_node_for_path( subsystem );
            registry.initialize_search_context( sc, *node );
            current_node = node;
            if( Wildcards::match( look_ahead.c_str(), "all" ) )
                {
                completions.push_back( SString( "all") );
                }
            while( registry.find_variable( 0, sc, *node ) )
                {
                if( sc.get_node() != current_node )
                    {
                    if( Wildcards::match( look_ahead.c_str(), sc.get_path() ) )
                        {
                        completions.push_back( sc.get_path() );
                        }
          
                    current_node = sc.get_node();
                    }
                }

            }
        }
    }


void Console::parse_or_complete_metaobject
    (
    const std::string &     param,
    bool                    auto_complete,
    const VariantPointer &  parse_result,
    PValueArray<SString> &  completions,
    bool                    abstract_included
    )
    {
    ClassManager &      class_manager = Global::class_manager();
    const MetaObject *  metaobject = NULL;

    metaobject = class_manager.get_metaobject_no_throw( param );
    
    if( !metaobject || auto_complete )
        {
        if( !auto_complete )
            {
            throw ::Massiv::Core::Lib::UndefinedClassTypeException( param );
            }
        else
            {
            std::vector<std::string>    classes;
            std::size_t                 index;
            std::string                 class_name = param;
            class_name += "*";
            class_manager.find_classes( classes, class_name );
            
            for( index = 0; index < classes.size(); index++ )
                {
                if( !abstract_included )
                    {
                    const MetaObject * mo = class_manager.get_metaobject( classes[ index ] );
                    if( !mo->get_attribute_abstract() )
                        {
                        completions.push_back( classes[ index ] );
                        }
                    }
                else
                    {
                    completions.push_back( classes[ index ] );
                    }
                }
            return;
            }
        }
    else
        {
        ParseMetaobjectResult * r = variant_cast<ParseMetaobjectResult *>( parse_result );
            
        r->metaobject = metaobject;
        }

    }


void Console::parse_or_complete_object_member
    (
    const std::string &     scoped_member,
    ParseMemberType         member_type,
    bool                    auto_complete,
    const VariantPointer &  parse_result,
    PValueArray<SString> &  completions
    )
    {
    // Member lookup context.

    PtrToObject                     object;
    VariantPointer                  component;
    const MetaObject *              metaobject;
    Property *                      property;
    MetaObject::MethodInfoIterator  method;
    bool                            method_found;
    
    // Parser.
    
    // Note:    Due to simple parsing method based on iostreams extractors
    //          we have to seperate "tokens" by spaces in the input stream.

    std::string                     tokenized_member
        ( replace_substrings( scoped_member, ".", " . " ) );
    std::istringstream              stream( tokenized_member );
    Serializer::Description         desc;
    TextReader                      tr( stream );
    std::string                     look_ahead;
    std::string                     matched_prefix;

    tr.next_line();

    // Parse object id.

    ObjectId    object_id;

    try
        {
        object_id.text_read( tr, desc );
        }
    catch( ... )
        {
        if( !auto_complete )
            {
            throw;
            }
        else
            {
            // Rewind stream and complete object id.

            tr.line().clear();
            tr.line().seekg( 0 );

            complete_object_id( tr.line(), completions );
            return;
            }
        }
    
    matched_prefix = object_id_to_string( object_id );

    // Initialize lookup context.
    
    object = Global::garbage_collector().find_local_object( object_id );
    
    if( !object )
        {
        if( !auto_complete )
            {
            throw ExCommonError( "NULL_ID or remote object." );
            }
        else
            {
            return;
            }
        }

    component = object->get_complete_object();
    metaobject = object->get_metaobject();
    property = object;
    method_found = false;
    
    // Resolve scoped member.

    while( true )
        {        
        std::istream &  stream = tr.line();

        // Look-ahead.

        stream >> look_ahead;
        
        if( stream.fail() )
            {
            if( !auto_complete )
                {
                // Quit parser loop.

                break;
                }
            
            look_ahead.clear();
            }

        // Update context.
        
        if( property->get_type() == Property::TYPE_OBJECT_PROPERTY )
            {
            // This is an object.
            // Update lookup context and descent into the object.
            
            object = property->cast_to_object();
            component = object->get_complete_object();
            metaobject = object->get_metaobject();
            }
        else
        if( property->get_type() == Property::TYPE_POINTER )
            {
            // This is a pointer.
            // Update lookup context and dereference the pointer.

            PointerInfo * info = property->get_pointer_info();

            object = info->ptr_get_object();
            
            if( !object )
                {
                if( !auto_complete )
                    {    
                    throw ExCommonError( "NULL_ID or remote object." );
                    }
                else
                    {
                    return;
                    }
                }

            property = object;
            component = object->get_complete_object();
            metaobject = object->get_metaobject();
            }
        
        // Parse object component.

        if( property->get_type() == Property::TYPE_OBJECT_PROPERTY )
            {
            while( look_ahead == "as" )
                {
                std::string component_name;

                // Read component name.
                
                stream >> component_name;

                if( stream.fail() && !auto_complete )
                    {
                    throw ExCommonError( "Expected class component." );
                    }
                
                // Upcast and update lookup context.

                MetaObject::UpcastResult upcast_result = metaobject->upcast( component, component_name );
                if( !component_name.size() || upcast_result.second < 0 )
                    {
                    if( !auto_complete )
                        {
                        throw ExCommonError( "Invalid class component." );
                        }
                    else
                        {
                        // Complete component name.

                        complete_class_component
                            (
                            component_name + "*",
                            metaobject,
                            matched_prefix + " as ",
                            completions
                            );

                        return;
                        }
                    }

                // Update component and metaobject.

                component = upcast_result.first;
                metaobject = metaobject->get_component_infos()[ upcast_result.second ].metaobject;
                
                matched_prefix += " as " + component_name;
                
                // Look ahead.

                stream >> look_ahead;

                if( stream.fail() )
                    {
                    look_ahead.clear();
                    break;
                    }
                }
            }
        
        // Parse object member.
        
        if( property->get_type() == Property::TYPE_OBJECT_PROPERTY )
            {
            // Note:    Property::TYPE_POINTER context is automatically
            //          dereferenced to Property::TYPE_OBJECT_PROPERTY
            //          at the beginning of the loop.

            if( look_ahead == "." )
                {
                MetaObject::MemberContexts  member_contexts;
                std::string                 member_name;

                // Read object member name.

                stream >> member_name;
                
                if( stream.fail() && !auto_complete )
                    {
                    throw ExCommonError( "Expected class member." );
                    }

                // Try accessing object method.

                if( member_type == PARSE_METHOD )
                    {
                    metaobject->find_methods
                        (
                        member_contexts,
                        component,
                        member_name
                        );

                    if( member_contexts.size() > 1 )
                        {
                        if( !auto_complete )
                            {
                            throw ExCommonError
                                ( "Ambiguous method (object component need be specified)." );
                            }
                        else
                            {
                            return;
                            }
                        }

                    if( member_contexts.size() == 1 )
                        {
                        const MetaObject::MemberContext & context = member_contexts.front();
                        
                        // Match method.

                        matched_prefix += "." + member_name;

                        // Update context.

                        component = context.object;
                        metaobject = context.metaobject;
                        method = context.method_it;
                        method_found = true;

                        break;
                        }
                    }

                // Try accessing property.

                MetaObject::PropertyPointers properties;
                
                metaobject->find_properties
                    (
                    properties,
                    component,
                    member_name
                    );

                if( properties.size() > 1 )
                    {
                    if( !auto_complete )
                        {
                        throw ExCommonError
                            ( "Ambiguous property (object component need be specified)." );
                        }
                    else
                        {
                        return;
                        }
                    }

                if( properties.size() == 1 )
                    {
                    // Match property.

                    matched_prefix += "." + member_name;

                    // Update context.

                    property = properties.front();

                    // Component and metaobject become obsolete in property context.
                    // If property is object or pointer, these will be set again in
                    // next while cycle. Containers don't need them of course.

                    component = VariantPointer();
                    metaobject = NULL;

                    // Object properties, pointer properties and containers can
                    // further be scoped.

                    if( property->get_type() == Property::TYPE_OBJECT_PROPERTY ||
                        property->get_type() == Property::TYPE_POINTER ||
                        property->get_type() == Property::TYPE_CONTAINER )
                        {
                        continue;
                        }

                    break;
                    }

                // Failed to parse object member.
                
                if( !auto_complete )
                    {
                    if( member_type == PARSE_METHOD )
                        {
                        throw ExCommonError( "Not an object member." );
                        }
                    else
                        {
                        throw ExCommonError( "Not a property." );
                        }
                    }
                else
                    {
                    // Complete object member.

                    complete_object_class_member
                        (
                        member_name + "*",
                        metaobject,
                        member_type,
                        matched_prefix + ".",
                        completions
                        );

                    return;
                    }
                }
            else
            if( look_ahead.size() || auto_complete )
                {
                // Invalid token.
                // Determine current context.
                
                const MetaObject * context = NULL;
                
                if( property->get_type() == Property::TYPE_OBJECT_PROPERTY )
                    {
                    // Use metaobject of the current object property.

                    context = metaobject;
                    }
                else
                if( property->get_type() == Property::TYPE_POINTER )
                    {
                    // Use metaobject of the referenced object.

                    PointerInfo *   info = property->get_pointer_info();
                    PtrToObject     dereferenced = info->ptr_get_object();
                    
                    if( dereferenced != NULL )
                        {
                        context = dereferenced->get_metaobject();
                        }
                    else
                        {
                        context = NULL;
                        }
                    }    
                
                // Find out what tokens are valid in this context.
                
                bool can_use_as = false;
                bool can_use_dot = false;

                if( context != NULL )    
                    {
                    can_use_as = context->get_component_infos().size() > 1;
                        // First component is self, parents follow.
                    
                    if( member_type == PARSE_PROPERTY )
                        {
                        can_use_dot = context->get_property_infos().size() > 0;
                        }
                    else
                        {
                        // FIXME: can_use_dot should be true only iff there is
                        //        a method defined in this component or its parents.
                        // TODO: MetaObject::get_method_infos()

                        can_use_dot = true;
                        }
                    }

                if( !auto_complete )
                    {
                    std::string msg;

                    msg = "Expected ";
                    
                    if( can_use_as )
                        {
                        msg += "'as', ";
                        }

                    if( can_use_dot )
                        {
                        msg += "'.', ";
                        }

                    msg += "EOF";

                    throw ExCommonError( msg );
                    }
                else
                    {
                    // Complete/suggest next token.

                    completions.push_back( matched_prefix );

                    if( can_use_as )    
                        {
                        completions.push_back( matched_prefix + " as" );
                        }

                    if( can_use_dot )
                        {
                        completions.push_back( matched_prefix + "." );
                        }
                    }
                }
            }
        else
        if( property->get_type() == Property::TYPE_CONTAINER )
            {
            ArrayInfo *         array = property->get_array_info();
            DictionaryInfo *    dictionary = property->get_dictionary_info();
            SetInfo *           set = property->get_set_info();

            // Process container.
            // The parser is expected to access container element
            // and appropriatelly update context.

            if( array )
                {
                // Process array.
                
                if( look_ahead == "index" )
                    {
                    UInt32 index;

                    // Parse array index.

                    stream >> index;
                
                    if( stream.fail() )
                        {
                        if( !auto_complete )
                            {
                            throw ExCommonError( "Expected array index." );
                            }
                        else
                            {
                            return;
                            }
                        }

                    // Access array element.
                    // Update context.
                    
                    property = array->array_get_element( index );
                    
                    if( !property )
                        {
                        throw ExCommonError( "Can not access such element." );
                        }

                    matched_prefix += " index ";
                    matched_prefix += integer_to_string( index );
                    }
                else
                if( auto_complete )
                    {
                    completions.push_back( matched_prefix );
                    completions.push_back( matched_prefix + " index" );

                    return;
                    }
                else
                    {
                    throw ExCommonError( "Expected EOF or 'index'." );
                    }
                }
            else
            if( dictionary )
                {
                // Process dictionary.

                if( look_ahead == "key" )
                    {
                    std::string key;

                    // Parse dictionary key.

                    stream >> quote_string( key );

                    if( stream.fail() )
                        {
                        if( !auto_complete )
                            {
                            throw ExCommonError( "Expected dictionary key." );
                            }
                        else
                            {
                            return;
                            }
                        }

                    // Access associated data.
                    // Update context.

                    std::istringstream  key_stream( key );
                    TextReader          key_reader( key_stream );

                    property = dictionary->dictionary_look_up( key_reader );

                    if( !property )
                        {
                        throw ExCommonError( "Can not look-up such key." );
                        }

                    matched_prefix += " key ";
                    matched_prefix += string_to_quoted_string( key );
                    }
                else
                if( auto_complete )
                    {
                    completions.push_back( matched_prefix );
                    completions.push_back( matched_prefix + " key" );

                    return;
                    }
                else
                    {
                    throw ExCommonError( "Expected EOF or 'key'." );
                    }
                }
            else
            if( set )
                {
                // Process set.

                if( auto_complete )
                    {
                    completions.push_back( matched_prefix );
                    return;
                    }
                else
                    {
                    throw ExCommonError( "EOF Expected." );
                    }
                }
            else
                {
                throw ExCommonError( "Unsupported container." );
                }

            // Object properties, pointer properties and containers can
            // further be scoped.

            if( property->get_type() == Property::TYPE_OBJECT_PROPERTY ||
                property->get_type() == Property::TYPE_POINTER ||
                property->get_type() == Property::TYPE_CONTAINER )
                {
                continue;
                }
            }

        break;
        }
    
    // The object member is fully qualified and matched now.
    
    if( auto_complete )
        {
        return;
        }

    // EOF expected.

    tr.line() >> look_ahead;

    if( !tr.line().fail() )
        {
        throw ExCommonError( "EOF expected." );
        }
        
    // Test if valid member was found.
    
    if( member_type == PARSE_METHOD )
        {
        if( !method_found )
            {
            throw ExCommonError( "This is not a method." );
            }
        }

    // Fill result.

    switch( member_type )
        {
        case PARSE_PROPERTY:
            {
            ParsePropertyResult * r = variant_cast<ParsePropertyResult *>( parse_result );
            
            r->property = property;
            r->object = object;
            break;
            }
        
        case PARSE_METHOD:
            {
            ParseMethodResult * r = variant_cast<ParseMethodResult *>( parse_result );

            r->object = object;
            r->component = component;
            r->metaobject = metaobject;
            r->method = method;
            break;
            }
        }
    }


void Console::complete_class_component
    (
    const std::string &     component_pattern,
    const MetaObject *      metaobject,
    const std::string &     completion_prefix,
    PValueArray<SString> &  completions
    )
    {
    hope( metaobject );
    const MetaObject::ComponentInfos &  components = metaobject->get_component_infos();
    MetaObject::ComponentInfoIterator   it = components.begin();

    while( it != components.end() )
        {
        const MetaObject::ComponentInfo & component = *it;
        if( !component.is_complete_component() )
            {
            if( Wildcards::match( component_pattern, component.name ) )
                {
                completions.push_back( completion_prefix + component.name );
                }
            }

        ++it;
        }
    }


void Console::complete_object_class_member
    (
    const std::string &     member_pattern,
    const MetaObject *      metaobject,
    ParseMemberType         member_type,
    const std::string &     completion_prefix,
    PValueArray<SString> &  completions
    )
    {
    MetaObject::PropertyInfoIterators   property_infos;
    MetaObject::MethodInfoIterators     method_infos;
    std::size_t                         i;

    // Complete property members.

    metaobject->find_property_infos
        (
        property_infos,
        member_pattern
        );

    for( i = 0; i < property_infos.size(); i++ )
        {
        completions.push_back
            ( completion_prefix + property_infos[ i ]->name );
        }

    // Complete method members.

    if( member_type == PARSE_METHOD )
        {
        metaobject->find_method_infos
            (
            method_infos,
            member_pattern
            );
        
        for( i = 0; i < method_infos.size(); i++ )
            {
            completions.push_back
                ( completion_prefix + method_infos[ i ]->name );
            }
        }
    }


bool Console::complete_object_id
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    std::ostringstream  ss;
    std::string         look_ahead;
    std::string         class_name;
    UInt32              provider_id;
    UInt32              relative_id;
    bool                tracked_flag;

    // Expect leading '['.

    const GarbageCollector::NameToObjectIdMap & object_names =
        Global::garbage_collector().get_names_to_objects();

    stream >> look_ahead;

    if( look_ahead != "[" )
        {
        // Try interpreting @c look_ahead as an object name.

        if( object_names.find( look_ahead ) != object_names.end() )
            {
            return true;
            }
        
        // Complete leading '['.

        if( look_ahead.empty() )
            {
            result.push_back( SString( "[" ) );
            }

        // Complete object names.
        
        GarbageCollector::NameToObjectIdMap::const_iterator it;
        std::string pattern = look_ahead + "*";

        for( it = object_names.begin(); it != object_names.end(); ++it )
            {
            const std::string & name = it->first;

            if( Wildcards::match( pattern, name ) )
                {
                result.push_back( name );
                }
            }
        
        return false;
        }
    
    // Expect provider_id.

    stream >> provider_id;

    if( stream.fail() )
        {
        // Complete provider_id.

        ss << "[ ";
        result.push_back( ss.str() + std::string( "0" ) );
        result.push_back( ss.str() + std::string( "1" ) );

        return false;
        }

    // Expect relative_id.

    stream >> relative_id;

    if( stream.fail() )
        {
        // Complete relative_id.

        ss << "[ " << provider_id << " ";
        
        result.push_back( ss.str() + std::string( "0" ) );
        result.push_back( ss.str() + std::string( "1" ) );

        return false;
        }

    // Expect tracked_flag.

    stream >> tracked_flag;

    if( stream.fail() )
        {
        // Complete tracked_flag.

        ss  << "[ " << provider_id << " "
            << relative_id << " ";

        result.push_back( ss.str() + std::string( "0" ) );
        result.push_back( ss.str() + std::string( "1" ) );

        return false;
        }
    
    // Complete trailing ']'.

    stream >> look_ahead;

    ss  << "[ " << provider_id << " "
        << relative_id << " " << tracked_flag << " ]";

    result.push_back( ss.str() );

    return !stream.fail() && look_ahead == "]";
    }


std::string Console::object_id_to_string
    (
    const ObjectId & object_id
    )
    {
    std::ostringstream ss;
    ss << object_id;
    return ss.str();
    }


std::string Console::integer_to_string
    (
    const Int32 &   integer
    )
    {
    std::ostringstream ss;
    ss << integer;
    return ss.str();
    }


std::string Console::string_to_quoted_string
    (
    const std::string & str
    )
    {
    std::ostringstream ss;
    ss << quote_const_string( str );
    return ss.str();
    }


std::string Console::replace_substrings
    (
    const std::string & str,
    const std::string & search,
    const std::string & replace
    )
    {
    std::string s = str;
    std::size_t min_pos = 0;
        
    do
        {
        std::size_t pos = s.find_first_of( search, min_pos );
        
        if( pos == std::string::npos )
            {
            break;
            }

        s.replace( pos, search.size(), replace );
        min_pos = pos + replace.size();
        }
    while( true );

    return s;
    }


void Console::complete_boolean
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    std::string look_ahead;

    stream >> look_ahead;
    look_ahead += "*";

    if( Wildcards::match( look_ahead.c_str(), "false" ) )
        {
        result.push_back( SString( "false" ) );
        }

    if( Wildcards::match( look_ahead.c_str(), "true" ) )
        {
        result.push_back( SString( "true" ) );
        }
    }


void Console::complete_string
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    std::string         s;
    std::ostringstream  ss;

    stream >> quote_string( s );
    ss << quote_const_string( s );

    result.push_back( ss.str() );
    }


void Console::complete_registry_node
    (
    std::istream &          stream,
    PValueArray<SString> &  result,
    const std::string &     subsystem
    )
    {
    std::string param;

    stream >> param;
    parse_or_complete_regnode
        (
        param,
        true,
        VariantPointer(),
        result,
        subsystem
        );
    
    }


void Console::complete_config_variable
    (
    std::istream &          stream,
    PValueArray<SString> &  result,
    const std::string &     subsystem,
    bool                    editable
    )
    {
    std::string param;

    stream >> quote_string( param );
    parse_or_complete_cvar
        (
        param,
        true,
        VariantPointer(),
        result,
        subsystem,
        editable
        );

    }


void Console::complete_class_name
    (
    std::istream &          stream,
    PValueArray<SString> &  result,
    bool                    abstract_included
    )
    {
    std::string class_name;

    stream >> class_name;
    parse_or_complete_metaobject
        (
        class_name,
        true,
        VariantPointer(),
        result,
        abstract_included
        );
    }


void Console::complete_wildcard
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    std::string look_ahead;

    stream >> look_ahead;

    result.push_back( look_ahead + "*" );
    }


void Console::complete_object_member
    (
    std::istream &          stream,
    ParseMemberType         member_type,
    PValueArray<SString> &  result
    )
    {
    std::string member;
    std::size_t i;

    stream >> quote_string( member );

    // Complete object member.

    parse_or_complete_object_member
        (
        member,
        member_type,
        true,
        VariantPointer(),
        result
        );
    
    // Quote completions.

    for( i = 0; i < result.size(); i++ )
        {
        std::ostringstream ss;
        
        ss << quote_const_string( result[ i ] );
        result( i ) = ss.str();
        }
    }


void Console::complete_property
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    complete_object_member( stream, PARSE_PROPERTY, result );
    }


void Console::complete_method
    (
    std::istream &          stream,
    PValueArray<SString> &  result
    )
    {
    complete_object_member( stream, PARSE_METHOD, result );
    }



// Implemetation of CommandArgument
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Console::CommandArgument::CommandArgument()
    {
    unbind();
    }

    
Console::CommandArgument::CommandArgument
    (
    const CommandArgument & arg
    ) : description( arg.description ),
        flags( arg.flags ),
        default_value( arg.default_value )
    {
    unbind();
    }


void Console::CommandArgument::unbind()
    {
    bound_integer = 0;
    bound_float = 0.0;
    bound_boolean = false;
    bound_string = std::string();
    bound_object_id = ObjectId();
    bound_property = NULL;
    bound_component.set_null();
    bound_object = NULL;
    bound_metaobject = NULL;
    bound_regnode = NULL;
    bound_cvar = NULL;
    }
    

std::string Console::CommandArgument::flags_to_string() const
    {
    std::string type;
    
    if( flags & FLAGS_ARG_OPTIONAL )
        {
        type = "optional ";
        }
    
    if( flags & FLAGS_INTEGER )
        {
        type += "integer";
        }
    else if( flags & FLAGS_FLOAT )
        {
        type += "float";
        }
    else if( flags & FLAGS_BOOLEAN )
        {
        type += "boolean";
        }
    else if( flags & FLAGS_STRING )
        {
        type += "string";
        }
    else if( flags & FLAGS_OBJECT_ID )
        {
        type += "object id";
        }
    else if( flags & FLAGS_STATISTICS )
        {
        type += "statistics registry node ";
        }
    else if( flags & FLAGS_CONFIG_VARIABLE )
        {
        type += "config variable (read-only)";
        }
    else if( flags & FLAGS_CONFIG_VARIABLE_SET )
        {
        type += "config variable";
        }
    else if( flags & FLAGS_CLASS_NAME )
        {
        type += "class name";
        } 
    else if( flags & FLAGS_CLASS_NAME_NOT_ABSTRACT )
        {
        type += "name of a not abstract class";
        } 
    else if( flags & FLAGS_WILDCARD )
        {
        type += "wildcard";
        } 
    else if( flags & FLAGS_PROPERTY )
        {
        type += "class property";
        }
    else if( flags & FLAGS_METHOD )
        {
        type += "class method";
        }       
    else 
        {
        type = "<unknown>";
        }

    return type;
    }


// Implemetation of Console::ConsoleCommand
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Console::ConsoleCommand::ConsoleCommand()
      : flags( 0 ),
        execute( NULL ),
        complete( NULL )
    {
    }


void Console::ConsoleCommand::new_argument
    (
    const std::string & description,
    int                 flags,
    const std::string & default_value
    )
    {
    if( args.size() && !( flags & CommandArgument::FLAGS_ARG_OPTIONAL ) )
        {
        if( args.back().flags & CommandArgument::FLAGS_ARG_OPTIONAL )
            {
            panic( "Non optional command parameter cannot follow optional." );
            }
        }
    args.push_back( CommandArgument() );
    args.back().description = description;
    args.back().flags = flags;
    args.back().default_value = default_value;
    }


// Implementation of Console::CommandList
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        
void Console::CommandList::clear()
    {
    Map::clear();
    }
    

Console::ConsoleCommand * Console::CommandList::new_command
    (
    const std::string &         spelling,
    const std::string &         description,
    int                         flags,
    ConsoleCommand::Execute *   execute,
    ConsoleCommand::Complete *  complete
    )
    {
    hope( execute );
    hope( complete );

    ConsoleCommand cmd;

    cmd.spelling = spelling;
    cmd.description = description;
    cmd.flags = flags;
    cmd.execute = execute;
    cmd.complete = complete;

    Iterator it = insert( Map::value_type( spelling, cmd ) ).first;
    return &( it->second );
    }


Console::CommandList::Iterator Console::CommandList::begin()
    {
    return Map::begin();
    }


Console::CommandList::Iterator Console::CommandList::end()
    {
    return Map::end();
    }


Console::CommandList::Iterator Console::CommandList::find
    (
    const std::string & spelling
    )
    {
    return Map::find( spelling );
    }


// Implemetation of Console::CommandContext
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Console::CommandContext::CommandContext
    (
    Console *                                       console,
    const Remote<AccountObject> &                   account,
    const Console::ConsoleCommand::ArgumentList &   args,
    int                                             num_args
    ) : console( console ),
        account( account ),
        args( args ),
        num_args( num_args )
    {
    hope( console );
    }


const SInt64 & Console::CommandContext::get_integer
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_INTEGER );
    return args[ index ].bound_integer;
    }


const SFloat32 & Console::CommandContext::get_float
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_FLOAT );
    return args[ index ].bound_float;
    }


const SBool & Console::CommandContext::get_boolean
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_BOOLEAN );
    return args[ index ].bound_boolean;
    }


const SString & Console::CommandContext::get_string
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_STRING ||
            args[ index ].flags & CommandArgument::FLAGS_STATISTICS ||
            args[ index ].flags & CommandArgument::FLAGS_CONFIG_VARIABLE ||
            args[ index ].flags & CommandArgument::FLAGS_CONFIG_VARIABLE_SET ||
            args[ index ].flags & CommandArgument::FLAGS_CLASS_NAME ||
            args[ index ].flags & CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT ||
            args[ index ].flags & CommandArgument::FLAGS_WILDCARD );

    return args[ index ].bound_string;
    }


const ObjectId & Console::CommandContext::get_object_id
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_OBJECT_ID );
    return args[ index ].bound_object_id;
    }


Property * Console::CommandContext::get_property
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_PROPERTY );
    hope( args[ index ].bound_object != NULL );
    return args[ index ].bound_property;
    }


const VariantPointer & Console::CommandContext::get_component
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_METHOD );
    hope( args[ index ].bound_object != NULL );
    return args[ index ].bound_component;
    }


const PtrToObject & Console::CommandContext::get_object
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_METHOD || 
            args[ index ].flags & CommandArgument::FLAGS_PROPERTY ||
            args[ index ].flags & CommandArgument::FLAGS_OBJECT_ID );
    return args[ index ].bound_object;
    }


const MetaObject * Console::CommandContext::get_metaobject
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( ( args[ index ].flags & CommandArgument::FLAGS_METHOD ) ||
            ( args[ index ].flags & CommandArgument::FLAGS_CLASS_NAME ) ||
            ( args[ index ].flags & CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT ) );
    return args[ index ].bound_metaobject;
    }


const MetaObject::MethodInfoIterator & Console::CommandContext::get_method
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( args[ index ].flags & CommandArgument::FLAGS_METHOD );
    hope( args[ index ].bound_object != NULL );
    return args[ index ].bound_method;
    }


Registry::RegistryNode * Console::CommandContext::get_regnode
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( ( args[ index ].flags & CommandArgument::FLAGS_STATISTICS ) );
    return args[ index ].bound_regnode;
    }


ConfigVariable * Console::CommandContext::get_cvar
    (
    int index
    ) const
    {
    hope( index < num_args );
    hope( ( args[ index ].flags & CommandArgument::FLAGS_CONFIG_VARIABLE ) ||
            ( args[ index ].flags & CommandArgument::FLAGS_CONFIG_VARIABLE_SET ) );
    return args[ index ].bound_cvar;
    }


const Console::CommandArgument & Console::CommandContext::get_last_arg() const
    {
    return args[ num_args - 1 ];
    }


void Console::CommandContext::print
    (
    const std::string & s,
    const SColor      & color
    ) const
    {
    account->async_print_to_console( ::Massiv::System::get_local_node_id(), s, color );
    }


} // namespace Demo::Lib
} // namespace Demo
