/** @file console.h
 * Console.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_CONSOLE_H
#define DEMO_CONSOLE_H

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_SHARED_FWD_H
#include "lib/shared/shared_fwd.h"
#endif

#ifndef MASSIV_CORE_MANAGEMENT_H
#include "core_management.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Console.
 */
class Console
    {
public:

    typedef Massiv::Core::ConfigVariable    ConfigVariable;
    typedef Massiv::Core::Registry          Registry;
    typedef PDictionary
                <
                PString,
                PString
                > Environment;

    
    /**
     * Default constructor.
     */
    Console();
    
    /**
     * Default destructor.
     */
    virtual ~Console();

public:

    // Forward declarations.
    class CommandArgument;
    class ConsoleCommand;
    class CommandList;
    class CommandContext;

public:

    enum ParseMemberType
        {
        PARSE_PROPERTY,
            /**< Scoped member that is to be parsed can be any property. */
        PARSE_METHOD
            /**< Scoped member can be a RPC method. */
        };

    struct ParseCvarResult
        {
        ConfigVariable *    cvar;
            /**< Parsed config variable. */
        };

    struct ParseMethodResult
        {
        PtrToObject                     object;
            /**< Parsed object. */
        VariantPointer                  component;
            /**< Parsed object component. The component is pinned thanks to 
                 @c object. */
        const MetaObject *              metaobject;
            /**< @c component's metaobject. */
        MetaObject::MethodInfoIterator  method;
            /**< Parsed method. */
        };

    struct ParseMetaobjectResult
        {
        const MetaObject *  metaobject;
            /**< Parsed metaobject. */
        };

    struct ParseRegnodeResult
        {
        Registry::RegistryNode *    regnode;
            /**< Parsed registry node. */
        };

    struct ParsePropertyResult
        {
        Property *  property;
            /**< Parsed property. Always pinned thanks to @c object. */
        PtrToObject object;
            /**< Object that owns @c property. */
        };
    
    /**
     * Console command argument.
     */
    class CommandArgument
        {
    public:

        enum Flags
            {
            FLAGS_INTEGER                 = 0x0001,
                /** The argument is integer. */
            FLAGS_FLOAT                   = 0x0002,
                /**< The argument is float. */
            FLAGS_BOOLEAN                 = 0x0004,
                /**< The argument is boolean. */
            FLAGS_STRING                  = 0x0008,
                /**< The argument is string. */
            FLAGS_OBJECT_ID               = 0x0010,
                /**< The argument is object id. */
            FLAGS_STATISTICS              = 0x0020,
                /**< The argument is statisics registry node. */
            FLAGS_CONFIG_VARIABLE         = 0x0040,
                /**< The argument is a config variable to show. */
            FLAGS_CONFIG_VARIABLE_SET     = 0x0080,
                /**< The argument is a config variable.
                     We must be able to distinguish between writable cvars
                     and read-only ones. */
            FLAGS_CLASS_NAME              = 0x0100,
                /**< The argument is a class name. */
            FLAGS_CLASS_NAME_NOT_ABSTRACT = 0x0200,
                /**< The argument is a name of a not-abstract class. */
            FLAGS_WILDCARD                = 0x0400,
                /**< The argument is a wildcard pattern. */
            FLAGS_PROPERTY                = 0x0800,
                /**< The argument is a property. */
            FLAGS_METHOD                  = 0x1000,
                /**< The argument is a method. */
            FLAGS_ARG_OPTIONAL            = 0x2000
                /**< This argument is optinal. */
            };

        /** Default constructor. */
        CommandArgument();
            
        /** Copy constructor. */
        CommandArgument
            (
            const CommandArgument & arg
            );

    public:

        /** Releases bound data. */
        void unbind();

        /** Converts @c flags to a printable string. */
        std::string flags_to_string() const;

    public:
            
        //
        // Argument description.
        //

        std::string     description;
            /**< Argument description. */
        int             flags;
            /**< Argument flags. */
        std::string     default_value;
            /**< Default argument value. */
        
        //
        // Bound value.
        //

        SInt64                          bound_integer;
            /**< Bound integer. */
        SFloat32                        bound_float;
            /**< Bound float. */
        SBool                           bound_boolean;
            /**< Bound bool. */
        SString                         bound_string;
            /**< Bound string. */
        ObjectId                        bound_object_id;
            /**< Bound object id. */
        Property *                      bound_property;
            /**< Bound property. Filled together with @c bound_object which
                 ensures proper pinning. */
        VariantPointer                  bound_component;
            /**< Bound object component. Filled together with @c bound_object. */
        PtrToObject                     bound_object;
            /**< Bound object. */
        const MetaObject *              bound_metaobject;
            /**< Bound metaobject. */
        MetaObject::MethodInfoIterator  bound_method;
            /**< Bound method. Filled together with @c bound_object. */
        Registry::RegistryNode *        bound_regnode;
            /**< Bound registry node. */
        ConfigVariable *                bound_cvar;

        }; // CommandArgument

    /**
     * Console command.
     * Maintains command description and list of its arguments holding
     * currently parsed values.
     * To be able to process commands in parallel, @c ConsoleCommand
     * must be used "by value" (each instance has its own copy).
     */
    class ConsoleCommand
        {
    public:

        enum Flags
            {
            FLAGS_ROOT_ACCOUNT = 1
                /**< The command must be processed on behalf of root account. */    
            };

        typedef std::vector< CommandArgument >  ArgumentList;
        
        /**
         * Executes command according to @c context.
         */
        typedef void Execute
            (
            const CommandContext &  context
            );

        /**
         * Returns possible completions for the last "available" argument.
         * The result is to be written to @c completions.
         * @c stream is initialized to point to the input stream, where
         * the last argument is written. This can be used to implement
         * context-sensitive argument completions.
         */
        typedef void Complete
            (
            const CommandContext &  context,
            std::istream &          stream,
            PValueArray<SString> &  completions
            );

    public:

        /** Constructor. */
        ConsoleCommand();
        
        /** Registers new command argument. */
        void new_argument
            (
            const std::string & description,
            int                 flags,
            const std::string & default_value = ""
            );

    public:
            
        std::string     spelling;
            /**< Command spelling. */
        std::string     description;
            /**< Command description. */
        int             flags;
            /**< Command flags. */
        ArgumentList    args;
            /**< Command arguments. */
        Execute *       execute;
            /**< Handler that executes command. */
        Complete *      complete;
            /**< Handler that auto completes command. */
        }; // ConsoleCommand

    /**
     * Command list.
     */
    class CommandList :
        private std::map
            <
            std::string,
            ConsoleCommand,
            std::less<std::string>
            >
        {
    public:
            
        typedef std::map
            <
            std::string,
            ConsoleCommand, 
            std::less<std::string>
            > Map;
        
        typedef Map::iterator   Iterator;

        /** Erases all registered commands. */
        void clear();
            
        /** Registers new console command that is applicable for the local node. */
        ConsoleCommand * new_command
            (
            const std::string &         spelling,
            const std::string &         description,
            int                         flags,
            ConsoleCommand::Execute *   execute,
            ConsoleCommand::Complete *  complete = &complete_args
            );

        /** Returns iterator pointing to the beginning of command list. */
        Iterator begin();

        /** Returns iterator pointing to the end of command list. */
        Iterator end();

        /** Returns iterator pointing to the command that match @c spelling. */
        Iterator find
            (
            const std::string & spelling
            );

        }; // CommandList

    /**
     * Command execution context.
     */
    class CommandContext
        {
    public:

        /** Constructor. */
        CommandContext
            (
            Console *                               console,
            const Remote<AccountObject> &           account,
            const ConsoleCommand::ArgumentList &    args,
            int                                     num_args
            );

    public:

        /** Returns bound value to argument @c index. */
        const SInt64 & get_integer
            (
            int index
            ) const;
            
        /** Returns bound value to argument @c index. */
        const SFloat32 & get_float
            (
            int index
            ) const;
        
        /** Returns bound value to argument @c index. */
        const SBool & get_boolean
            (
            int index
            ) const;
        
        /** Returns bound value to argument @c index. */
        const SString & get_string
            (
            int index
            ) const;
            
        /** Returns bound value to argument @c index. */
        const ObjectId & get_object_id
            (
            int index
            ) const;
            
        /** Returns bound value to argument @c index. */
        Property * get_property
            (
            int index
            ) const;
            
        /** Returns bound value to argument @c index. */
        const VariantPointer & get_component
            (
            int index
            ) const;

        /** Returns bound value to argument @c index. */
        const PtrToObject & get_object
            (
            int index
            ) const;
            
        /** Returns bound value to argument @c index. */
        const MetaObject * get_metaobject
            (
            int index
            ) const;
        
        /** Returns bound value to argument @c index. */
        const MetaObject::MethodInfoIterator & get_method
            (
            int index
            ) const;

        /** Returns bound value to argument @c index. */
        Registry::RegistryNode * get_regnode
            (
            int index
            ) const;
            

        /** Returns bound value to argument @c index. */
        ConfigVariable * get_cvar
            (
            int index
            ) const;

        /** Returns last available argument. */
        const CommandArgument & get_last_arg() const;

    public:

        /** Prints string to command requestor's console. */
        void print
            (
            const std::string & s,
            const SColor      & color  = SColor( 1.0f, 1.0f, 0.0f )
            ) const;

    public:

        Console *                               console;
            /**< Console object executing this command. */

        Remote<AccountObject>                   account;
            /**< Under what account the command is executed. */

        const ConsoleCommand::ArgumentList &    args;
            /**< Argument list with bound values. */

        int                                     num_args;
            /**< Number of arguments available. */

        mutable std::string                     result; // FIXME - odstranit mutable
            /**< Result value. */

        }; // CommandContext

public:

    /**
     * Analyzes and processes or autocompletes @c command.
     * Depending on @c auto_complete either command completions or
     * parser errors are written to @c result.
     * Parser errors should not be written to client console directly,
     * they are logged to @c result instead. Use special tags to
     * log "standardized" error messages so that consumer can filter them
     * as needed.
     * If command was parsed properly, command handler prints results
     * to client's console through @c account object and @c result
     * is empty.
     */
    void process_command
        (
        const Remote<AccountObject> &   account,
        const std::string &             command,
        bool                            auto_complete,
        PValueArray<SString> &          result,
        Environment &                   env
        );

    /**
     * Returns a tag that represents a "Unknown command." message.
     * If Console does not recognize the command @c process_command,
     * @c result contains a single error message consisting
     * of this special tag (if not auto-completing).
     */
    static std::string get_unknown_command_tag();

    /**
     * Tests if @c result returned by @c process_command (if not auto-completing)
     * represents the "Unknown command" error.
     */
    static bool is_unknown_command_error
        (
        const PValueArray<SString> &    errors
        );

public:

    /**
     * Formats message that describes commands accepted by console.
     */
    std::string print_command_help();

    /**
     * Formats message that describes command arguments for a command
     * with given @c spelling.
     */
    std::string print_command_help
        (
        const std::string & spelling
        );
    
private:

    /** Parses and bounds value to argument. */
    void parse_argument
        (
        CommandArgument &           arg_description,
        TextReader &                tr,
        Serializer::Description &   desc
        );

    /** Fills default value to bound type. */
    void fill_default_value
        (
        CommandArgument & cmdarg
        );

    /** Replaces each occurence of @c search in @c str by @c replace. */
    std::string replace_substrings
        (
        const std::string & str,
        const std::string & search,
        const std::string & replace
        );

    /** Converts an integer to a string. */
    std::string string_to_quoted_string
        (
        const std::string & str
        );

    /** Converts an integer to a string. */
    std::string integer_to_string
        (
        const Int32 &   integer
        );

    /** Prints object id into the string. */
    std::string object_id_to_string
        (
        const ObjectId & object_id
        );
    
    /** Completes command @c command. */
    void complete_command
        (
        bool                    is_root_account,
        const std::string &     command,
        PValueArray<SString> &  completions
        );
    
    /** Default argument completion handler. */
    static void complete_args
        (
        const CommandContext &  context,
        std::istream &          stream,
        PValueArray<SString> &  result
        );
    
    /**
     * Parses or auto-completes config variable.
     * Depending on @c auto_complete either @c parse_result or @c completions
     * (if @c auto_complete is true) is filled.
     * Specific subsystem can be filled when auto_complete should be more
     * accurate ( eg. "Settings" ). When @c show_value is true, a value
     * contained in the config variable is also displayed.
     */
    void parse_or_complete_cvar
        (
        const std::string &     param,
        bool                    auto_complete,
        const VariantPointer &  parse_result,
        PValueArray<SString> &  completions,
        const std::string &     subsystem = "",
        bool                    show_value = false
        );

    /**
     * Parses or auto-completes registry node.
     * Depending on @c auto_complete either @c parse_result or @c completions
     * (if @c auto_complete is true) is filled.
     * Specific subsystem can be filled when auto_complete should be more
     * accurate.
     */
    void parse_or_complete_regnode
        (
        const std::string &     param,
        bool                    auto_complete,
        const VariantPointer &  parse_result,
        PValueArray<SString> &  completions,
        const std::string &     subsystem = ""
        );

    /**
     * Parses or auto-completes metaobject ( class name ).
     * Depending on @c auto_complete either @c parse_result or @c completions
     * (if @c auto_complete is true) is filled.
     */
    void parse_or_complete_metaobject
        (
        const std::string &     param,
        bool                    auto_complete,
        const VariantPointer &  parse_result,
        PValueArray<SString> &  completions,
        bool                    abstract_included
        );

    /**
     * Parses or auto-completes object member.
     * Depending on @c auto_complete either @c parse_result or @c completions
     * (if @c auto_complete is true) is filled.
     */
    void parse_or_complete_object_member
        (
        const std::string &     scoped_member,
        ParseMemberType         member_type,
        bool                    auto_complete,
        const VariantPointer &  parse_result,
        PValueArray<SString> &  completions
        );

    /**
     * Auto-completes class member of a class described by @c metaobject that
     * matches @c member_pattern.
     * @c member_type determines what types of class members should be considered.
     * @c completion_prefix is a prefix that is prepended to any completion
     * which is put to @c completions.
     */
    void complete_object_class_member
        (
        const std::string &     member_pattern,
        const MetaObject *      metaobject,
        ParseMemberType         member_type,
        const std::string &     completion_prefix,
        PValueArray<SString> &  completions
        );

    /**
     * Auto-completes class component.
     * @c component_pattern specifies pattern for components of @c metaobject
     * that should be matched. @c completion_prefix is prepended to any
     * positive match that is put into @c completions.
     */
    static void complete_class_component
        (
        const std::string &     component_pattern,
        const MetaObject *      metaobject,
        const std::string &     completion_prefix,
        PValueArray<SString> &  completions
        );

    /** Completes object id. Returns true if object_id is fully completed. */
    bool complete_object_id
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );
        
    /** Completes object method. */
    void complete_method
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );

    /** Completes object property. */    
    void complete_property
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );

    /** Completes object member. */
    void complete_object_member
        (
        std::istream &          stream,
        ParseMemberType         member_type,
        PValueArray<SString> &  result
        );
        
    /** Completes wildcard pattern. */
    void complete_wildcard
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );

    /** Completes class name. */
    void complete_class_name
        (
        std::istream &          stream,
        PValueArray<SString> &  result,
        bool                    abstract_included
        );

    /** Completes registry node. */
    void complete_config_variable
        (
        std::istream &          stream,
        PValueArray<SString> &  result,
        const std::string &     subsystem = "",
        bool                    editable = false
        );
        
    /** Completes registry node. */
    void complete_registry_node
        (
        std::istream &          stream,
        PValueArray<SString> &  result,
        const std::string &     subsystem
        );
        
    /** Completes string. */
    void complete_string
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );

    /** Completes boolean. */
    void complete_boolean
        (
        std::istream &          stream,
        PValueArray<SString> &  result
        );

protected:

    CommandList command_list;
        /**< Command list. */

    }; // Console

} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_CONSOLE_H
