/** @file container.h
 * Container interfaces.
 */

/*
 * begin        : Wed Feb 19 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LIB_CONTAINER_H
#define DEMO_LIB_CONTAINER_H

#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Interface that must implement every enumerator.
 * Enumerator is a special object that allows client to enumerate
 * objects stored in a container.
 */
class BasicEnumerator : public Object
    {
public:
    
    MASSIV_OBJECT( BasicEnumerator );

    /** Tests if enumerator is at the end of the sequence. */
    virtual bool is_end() = 0;

    /** Tests if enumerator is at the beginning of the sequence. */
    virtual bool is_begin() = 0;

    /** Moves enumerator forward in the sequence. */
    virtual void next() = 0;

    /** Moves enumerator backward in the sequence. */
    virtual void prev() = 0;

    /** Erases current element from the sequence. Invalidates enumerator. */
    virtual void erase() = 0;

    /** Erases current element and moves enumerator forward. */
    virtual void erase_and_next();

    }; // BasicEnumerator

/**
 * Enumerator for simple containers.
 */
class Enumerator : public BasicEnumerator
    {
public:

    MASSIV_OBJECT( Enumerator );

    /** Returns current element from the enumerated sequence. */
    virtual WeakPointer<Object> current() = 0;

    }; // Enumerator

/**
 * Enumerator for associative containers.
 */
class AssociativeEnumerator : public BasicEnumerator
    {
public:

    MASSIV_OBJECT( AssociativeEnumerator );

    /** Returns key of the current element from the enumerated sequence. */
    virtual WeakPointer<Object> current_key() = 0;

    /** Returns data associated with the current element. */
    virtual WeakPointer<Object> current_data() = 0;

    }; // AssociativeEnumerator

/**
 * Interface that must implement every container.
 * Container holds pointers to objects that are stored inside it.
 * There are two types of containers. Strong containers
 * own objects that are stored inside them, strong pointers to stored
 * items are used. Weak containers do not own those objects,
 * weak pointers are used.
 */
class BasicContainer : public Object
    {
public:

    MASSIV_OBJECT( BasicContainer );

    /** Returns number of items stored in the container. */
    virtual int size() = 0;

    /** Returns current capacity of the container. */
    virtual int capacity();

    /** Tests if container is empty. */
    virtual bool empty();

    /** Clears the container. */
    virtual void clear() = 0;

    }; // BasicContainer

/**
 * Container that holds objects.
 * Container may allow to store duplicate entries in it. To determine
 * whether two objects are equal (and thus the two entries duplicate)
 * @c equal() method is used.
 */
class Container : public BasicContainer
    {
public:

    MASSIV_OBJECT( Container );

    /**
     * Returns enumerator that enumerates all objects stored in
     * the container. Enumerator points to the beginning of the sequence.
     * The enumerator is usually invalidated by an erase operation that
     * erased the referenced item.
     */
    virtual Pointer<Enumerator> begin() = 0;

    /**
     * Returns enumerator that enumerates all objects stored in
     * the container. Enumerator points to the end of the sequence.
     * The enumerator is usually invalidated by an erase operation that
     * erased the referenced item.
     */
    virtual Pointer<Enumerator> end() = 0;

    /**
     * Tests if @c object is stored in the container.
     */
    virtual bool contains
        (
        const WeakPointer<Object> & object
        );

    /**
     * Removes first object, that equals to @c object, from the container.
     * Returns true if object was removed from the container.
     */
    virtual bool erase
        (
        const WeakPointer<Object> & object
        );

    /**
     * Removes all objects, that equal to @c object, from the container.
     * Returns number of objects removed.
     */
    virtual int erase_all
        (
        const WeakPointer<Object> & object
        );

    /**
     * Returns "range" enumerator that enumerates sequence of elements
     * that are equal to @c object.
     * Range enumerators are usually invalidated by both insert and erase
     * operations.
     */
    virtual Pointer<Enumerator> equal_range
        (
        const WeakPointer<Object> & object
        ) = 0;

    /**
     * Inserts @c object into the container.
     * Returns true if object was inserted to the container.
     */
    virtual bool insert
        (
        const WeakPointer<Object> & object
        );

    /**
     * Inserts @c object into the container and returns @c enumerator
     * that points to the stored object. Enumerator can be used later
     * to remove the object from the container.
     */
    virtual void insert
        (
        const WeakPointer<Object> & object,
        Pointer<Enumerator> &       enumerator
        ) = 0;

    /**
     * Tests whether the two objects are equal.
     */
    virtual bool equal
        (
        const WeakPointer<Object> & object1,
        const WeakPointer<Object> & object2
        ) const = 0;

    }; // Container

/**
 * Sequence interface.
 */
class SequenceContainer : public Container
    {
public:

    MASSIV_OBJECT( SequenceContainer );

    /** Puts @c object to the end of the sequence. */
    virtual void push_back
        (
        const WeakPointer<Object> & object
        ) = 0;

    /** Puts @c object to the front of the sequence. */
    virtual void push_front
        (
        const WeakPointer<Object> & object
        ) = 0;

    /** Inserts @c object before @c before_this element. */
    virtual void insert_before
        (
        const Pointer<Enumerator> & before_this,
        const WeakPointer<Object> & object
        ) = 0;

    /** Inserts @c object after @c after_this element. */
    virtual void insert_after
        (
        const Pointer<Enumerator> & after_this,
        const WeakPointer<Object> & object
        ) = 0;

    }; // SequenceContainer

/**
 * Associative container.
 * The container holds [ key, data ] pairs, where each entry is identified
 * by its @c key. @c data is any object associated with the @c key.
 * If the container allows to store multiple pairs with the same
 * @c key, then the key does not identify the entry uniquely.
 * In that case if client requests the container to find an entry,
 * both the key and data must be supplied. The container then uses
 * @c equal_key and @c equal_data to compare the entries.
 */
class AssociativeContainer : public BasicContainer
    {
public:

    MASSIV_OBJECT( AssociativeContainer );

    /**
     * Returns enumerator that enumerates all [ key, data ] pairs
     * stored in the container. Enumerator points to the beginning
     * of the sequence.
     * The enumerator is usually invalidated by erase operations
     * that erased the pair referenced by the enumerator.
     */
    virtual Pointer<AssociativeEnumerator> begin() = 0;

    /**
     * Returns enumerator that enumerates all [ key, data ] pairs
     * stored in the container. Enumerator points to the end
     * of the sequence.
     * The enumerator is usually invalidated by erase operations
     * that erased the pair referenced by the enumerator.
     */
    virtual Pointer<AssociativeEnumerator> end() = 0;

    /**
     * Tests if there is a pair with key equal to @c key stored
     * in the container.
     */
    virtual bool contains
        (
        const WeakPointer<Object> & key
        );

    /**
     * Tests if there is a (@c key, @c data) pair stored in
     * the container.
     */
    virtual bool contains
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     * Looks up the container and returns data associated with
     * the first pair that matches @c key.
     * If no such pair exists @c null is returned.
     * To get "consistent" results the method should only be called on
     * containers that do not support insertion of multiple pairs with
     * the same key and that refuse to associate @c null with stored keys.
     */
    virtual WeakPointer<Object> look_up
        (
        const WeakPointer<Object> & key
        );

    /**
     * Removes first pair whose key equals to @c key from the container.
     */
    virtual bool erase
        (
        const WeakPointer<Object> & key
        );

    /**
     * Removes [ @c key, @c data ] pair from the container.
     */
    virtual bool erase
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     * Removes all pairs whose keys equal to @c key from the container.
     * If container does not support insertion of multiple pairs with
     * the same key, then the method is equal to @c erase().
     */
    virtual int erase_all
        (
        const WeakPointer<Object> & key
        );

    /**
     * Returns "range" enumerator that enumerates sequence of pairs
     * whose keys match @c key.
     * Range enumerators are usually invalidated by both insert and erase
     * operations.
     */
    virtual Pointer<AssociativeEnumerator> equal_range
        (
        const WeakPointer<Object> & key
        ) = 0;

    /**
     * Inserts [ @c key, @c data ] pair into the container.
     * Returns true if pair was inserted.
     */
    virtual bool insert
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     * Inserts [ @c key, @c data ] pair into the container and returns
     * @c enumerator that points to the stored pair. Enumerator can
     * be used later to remove the object from the container.
     * The enumerator is usually invalidated by only the erase operation
     * that erased the referenced pair.
     * The enumerator can enumerate all the items stored in the container
     * (using @c next(), @c prev()).
     */
    virtual void insert
        (
        const WeakPointer<Object> &         key,
        const WeakPointer<Object> &         data,
        Pointer<AssociativeEnumerator> &    enumerator
        ) = 0;

    /**
     * Tests if the two keys are equal.
     */
    virtual bool equal_key
        (
        const WeakPointer<Object> & key1,
        const WeakPointer<Object> & key2
        ) const = 0;

    /**
     * Tests if associated data are the same.
     */
    virtual bool equal_data
        (
        const WeakPointer<Object> & data1,
        const WeakPointer<Object> & data2
        ) const = 0;

    }; // AssociativeContainer


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_LIB_CONTAINER_H
