/** @file hash_table_container.h
 * Hash table container.
 */

/*
 * begin        : Wed Feb 19 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LIB_HASH_TABLE_CONTAINER_H
#define DEMO_LIB_HASH_TABLE_CONTAINER_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_CONTAINER_H
#include "lib/shared/container.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


//
// HashTable.
//

/**
 * Basic hashed associative container.
 * The container stores [ @c key, @c data ] pairs. The pairs are organized
 * into buckets. Each bucket has assigned an @c Address. A bucket holds
 * those pairs whose addresses correspond to the address of the bucket.
 * The bucket is represented by double-linked list of the pairs
 * (these pairs form a collision chain).
 * Erase operations invalidate enumerators pointing to the removed
 * items. Range enumerators (returned by @c equal_range()) are invalidated
 * by both erase and insert operations.
 * Can not be instantiated as a member object or on the stack.
 */
class HashTable : public AssociativeContainer
    {
public:

    MASSIV_OBJECT( HashTable );

    /**
     * Container flags.
     */
    enum Flags
        {
        FLAGS_ALLOW_DUPLICATE_KEYS =    1,
            /**< Duplicate keys are allowed. */
        FLAGS_ALLOW_NULL_DATA =         2,
            /**< @c null data is allowed to associate with stored keys. */
        FLAGS_STRONG_REFERENCED_KEYS =  4,
            /**< Keys should be strong-referenced. */
        FLAGS_STRONG_REFERENCED_DATA =  8
            /**< Associated data should be strong-referenced. */
        }; // Flags

    /**
     * Predefined container flags.
     */
    enum PredefinedFlags
        {
        PURE_WEAK_HASH_MAP =            0,
            /**< PureWeakHashMap. */
        WEAK_HASH_MAP =                 FLAGS_STRONG_REFERENCED_KEYS,
            /**< WeakHashMap. */
        HASH_MAP =                      FLAGS_STRONG_REFERENCED_KEYS | FLAGS_STRONG_REFERENCED_DATA,
            /**< HashMap. */
        PURE_WEAK_HASH_MULTI_MAP =      PURE_WEAK_HASH_MAP | FLAGS_ALLOW_DUPLICATE_KEYS,
            /**< PureWeakHashMultiMap. */
        WEAK_HASH_MULTI_MAP =           WEAK_HASH_MAP | FLAGS_ALLOW_DUPLICATE_KEYS,
            /**< WeakHashMultiMap. */
        HASH_MULTI_MAP =                HASH_MAP | FLAGS_ALLOW_DUPLICATE_KEYS
            /**< HashMultiMap. */
        }; // PredefinedFlags
    
    /**
     * Parameters.
     */
    enum Parameters
        {
        DEFAULT_TABLE_SIZE =            101
            /**< Default hash table size. */
        }; // Parameters
    
private:

    /**
     * Constants.
     */
    enum Constants
        {
        INVALID_FIRST_CHAIN =   0x7fffffff,
            /**< Special address that tags @c first_chain invalid.
                 It is chosen so that it is always greater than any @c Address. */
        INVALID_LAST_CHAIN =    -1,
            /**< Special address that tags @c last_chain invalid.
                 It is chosen so that it is always less than any @c Address. */
        INVALID_CHAIN =         -1
            /**< Invalid chain. */
        }; // Constants

    typedef int     Address;    /// Address of a bucket.

public:

    struct BeginTag
        {
        };

    struct EndTag
        {
        };

    class ChainEntryEnumerator;
    class HashTableEnumerator;

    /**
     * Collision chain entry.
     * Maintains links to other entries in the same chain,
     * entry key and associated data.
     * Each collision chain is represented as a double-linked
     * list with tail.
     */
    class ChainEntry : public Object
        {
    public:
        
        MASSIV_OBJECT( ChainEntry )
        friend class HashTable;
        friend class ChainEntryEnumerator;
        friend class HashTableEnumerator;
    
    public:

        /** Returns key of this entry. */
        virtual WeakPointer<Object> get_key() = 0;

        /** Returns data associated with this entry. */
        virtual WeakPointer<Object> get_data() = 0;

    protected:

        PPointer<ChainEntry>    next;
            /**< Next entry in the collision chain. */
        PPointer<ChainEntry>    prev;
            /**< Previous entry in the collision chain. */
        }; // ChainEntry

    /**
     * Collision chain tail entry.
     * If collision chain consists of TailChainEntry only
     * (ie. the double-linked list is empty), the chain must be deleted.
     */
    class TailChainEntry : public ChainEntry
        {
    public:

        /**
         * Initializes object.
         */
        void initialize();

        /**
         */
        virtual WeakPointer<Object> get_key();

        /**
         */
        virtual WeakPointer<Object> get_data();

        }; // TailChainEntry

    /**
     * Collision chain entry where both entry key and associated data
     * are weak-referenced.
     */
    class PureWeakChainEntry : public ChainEntry
        {
    public:

        MASSIV_OBJECT( PureWeakChainEntry );
        friend class HashTable;

    public:

        /**
         * Initializes object.
         */
        void initialize
            (
            const WeakPointer<Object> & the_key,
            const WeakPointer<Object> & the_data
            );

        /**
         */
        virtual WeakPointer<Object> get_key();

        /**
         */
        virtual WeakPointer<Object> get_data();

    protected:

        PWeakPointer<Object>    key;
            /**< Entry key. */
        PWeakPointer<Object>    data;
            /**< Associated data. */
        }; // PureWeakChainEntry

    /**
     * Collision chain entry where key is strong-referenced
     * and associated data weak-referenced.
     */
    class WeakChainEntry : public ChainEntry
        {
    public:

        MASSIV_OBJECT( WeakChainEntry );
        friend class HashTable;

    public:

        /**
         * Initializes object.
         */
        void initialize
            (
            const WeakPointer<Object> & the_key,
            const WeakPointer<Object> & the_data
            );

        /**
         */
        virtual WeakPointer<Object> get_key();

        /**
         */
        virtual WeakPointer<Object> get_data();

    protected:

        PPointer<Object>        key;
            /**< Entry key. */
        PWeakPointer<Object>    data;
            /**< Associated data. */
        }; // WeakChainEntry

    /**
     * Collision chain entry where both entry key and associated data
     * are strong-referenced.
     */
    class StrongChainEntry : public ChainEntry
        {
    public:
        
        MASSIV_OBJECT( StrongChainEntry );
        friend class HashTable;

    public:
        
        /**
         * Initializes object.
         */
        void initialize
            (
            const WeakPointer<Object> & the_key,
            const WeakPointer<Object> & the_data
            );

        /**
         */
        virtual WeakPointer<Object> get_key();

        /**
         */
        virtual WeakPointer<Object> get_data();

    protected:

        PPointer<Object>    key;
            /**< Entry key. */
        PPointer<Object>    data;
            /**< Associated data. */
        }; // StrongChainEntry

    /**
     * Simple list-like enumerator that enumerates entries linked
     * in a collision chain for given address.
     */
    class ChainEntryEnumerator : public AssociativeEnumerator
        {
    public:

        MASSIV_OBJECT( ChainEntryEnumerator );
        friend class HashTable;

    public:

        /** Initializes enumerator to point to an empty sequence. */
        void initialize();

        /** Initializes enumerator to point to the beginning of @c chain. */
        void initialize
            (
            const Pointer<HashTable> &  table,
            Address                     chain,
            BeginTag
            );

        /** Initializes enumerator to point to the end of @c chain. */
        void initialize
            (
            const Pointer<HashTable> &  table,
            Address                     chain,
            EndTag
            );

        /**
         * Initializes enumerator to point to @c this_entry, that
         * is expected to be stored in the @c chain.
         */
        void initialize
            (
            const Pointer<HashTable> &  table,
            Address                     chain,
            const Pointer<ChainEntry> & this_entry
            );

    public:

        //
        // BasicEnumerator implementation.
        //

        /**
         */
        virtual bool is_end();

        /**
         */
        virtual bool is_begin();

        /**
         */
        virtual void next();

        /**
         */
        virtual void prev();

        /**
         */
        virtual void erase();

        /**
         */
        virtual void erase_and_next();

    public:

        //
        // AssociativeEnumerator implementation.
        //

        /**
         */
        virtual WeakPointer<Object> current_key();

        /**
         */
        virtual WeakPointer<Object> current_data();

    protected:

        //
        // Protected.
        //

        /** Updates hash table after erase operation. */
        void update_table();

    protected:

        PPointer<HashTable>     hash_table;
            /**< Pointer to hash table. */
        PInt32                  collision_chain;
            /**< Index of collision chain being enumerated. */
        PPointer<ChainEntry>    entry;
            /**< Entry that the enumerator points at. */
        PPointer<ChainEntry>    chain_tail;
            /**< Collision chain tail. */

        }; // ChainEntryEnumerator

    /**
     * RangeEnumerator enumerates items with given @c key.
     * The enumerated sequence is always a subset of one collision chain
     * because all the items have to produce the same hash code.
     * The enumerator is invalidated by both insert and erase operations.
     */
    class RangeEnumerator : public ChainEntryEnumerator
        {
    public:
        
        MASSIV_OBJECT( RangeEnumerator );
        friend class HashTable;

    public:
        
        /**
         * Initializes enumerator to point to the first element
         * that matches @c search_key.
         * @c search_key is expected to be found in @c chain.
         */
        void initialize
            (
            const Pointer<HashTable> &  table,
            const Pointer<Object> &     search_key,
            Address                     chain
            );

    public:

        /**
         */
        virtual bool is_begin();

        /**
         */
        virtual void next();

        /**
         */
        virtual void prev();

    private:

        PWeakPointer<Object>    key;
            /** Searched key. */
        PPointer<ChainEntry>    begin;
            /**< First entry that matches @c key. */

        }; // RangeEnumerator

    /**
     * HashTableEnumerator enumerates all elements stored in the
     * hash table.
     * The enumerator is invalidated by erase operation that
     * erases the element referenced by the enumerator.
     */
    class HashTableEnumerator : public ChainEntryEnumerator
        {
    public:
        
        /**
         * Initializes enumerator to point to the beginning of the
         * sequence.
         */
        void initialize
            (
            const Pointer<HashTable> &  table,
            BeginTag
            );

        /** Initializes enumerator to point to the end of the sequence. */
        void initialize
            (
            const Pointer<HashTable> &  table,
            EndTag
            );

    public:

        /**
         */
        virtual bool is_end();

        /**
         */
        virtual bool is_begin();

        /**
         */
        virtual void next();

        /**
         */
        virtual void prev();

        /**
         */
        virtual void erase_and_next();

    private:

        /**
         * Moves enumerator to the beginning of the next non-empty
         * chain if the enumerator is at the end of current chain.
         */
        void next_chain();

        /**
         * Moves enumerator to the end of the previous non-empty
         * chain if the enumerator is at the beginning of current chain.
         */
        void prev_chain();

        }; // HashTableEnumerator

    friend class ChainEntryEnumerator;
    friend class RangeEnumerator;
    friend class HashTableEnumerator;

public:

    //
    // HashTable interface.
    //

    /**
     * Initializes table.
     */
    void initialize
        (
        int container_flags,
        int table_size = DEFAULT_TABLE_SIZE
        );

public:

    //
    // BasicContainer implementation.
    //

    /**
     */
    virtual int size();

    /**
     */
    virtual void clear();

public:

    //
    // AssociativeContainer implementation.
    //

    /**
     */
    virtual Pointer<AssociativeEnumerator> begin();

    /**
     */
    virtual Pointer<AssociativeEnumerator> end();

    /**
     */
    virtual bool contains
        (
        const WeakPointer<Object> & key
        );

    /**
     */
    virtual bool contains
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     */
    virtual WeakPointer<Object> look_up
        (
        const WeakPointer<Object> & key
        );

    /**
     */
    virtual bool erase
        (
        const WeakPointer<Object> & key
        );

    /**
     */
    virtual bool erase
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     */
    virtual int erase_all
        (
        const WeakPointer<Object> & key
        );

    /**
     */
    virtual Pointer<AssociativeEnumerator> equal_range
        (
        const WeakPointer<Object> & key
        );

    /**
     */
    virtual bool insert
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /**
     */
    virtual void insert
        (
        const WeakPointer<Object> &         key,
        const WeakPointer<Object> &         data,
        Pointer<AssociativeEnumerator> &    enumerator
        );

    /**
     */
    virtual bool equal_key
        (
        const WeakPointer<Object> & key1,
        const WeakPointer<Object> & key2
        ) const;
    
    /**
     */
    virtual bool equal_data
        (
        const WeakPointer<Object> & data1,
        const WeakPointer<Object> & data2
        ) const;

public:

    //
    // HashTable extensions.
    //

    /** Returns hash code for given @c key. */
    virtual int hash_key
        (
        const WeakPointer<Object> & key
        ) const;

private:

    /** Creates chain entry that holds [ @c key, @c data ] pair. */
    virtual Pointer<ChainEntry> create_entry
        (
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /** Determines which collision chain should hold @c key. */
    Address get_address
        (
        const WeakPointer<Object> & key
        ) const;

    /** Creates collision @c chain that will hold [ @c key, @c data ] pair. */
    void create_chain
        (
        Address                     chain,
        const WeakPointer<Object> & key,
        const WeakPointer<Object> & data
        );

    /** Deletes collision @c chain. */
    void delete_chain
        (
        Address chain
        );

private:

    PArray< PPointer<ChainEntry> >  buckets;
        /**< The hash table is organized into buckets each being linked
             to those entries whose addresses belong to the same bucket.
             The bucket is represented as a double-linked list with tail.
             If the bucket is empty then the corresponding pointer is @c null.
             If the bucket is non-empty then the pointer points to the tail
             of the double-linked list that represents the bucket.
             The bucket is deleted if the list contains tail only. */
    PInt32                          num_items;
        /**< Number of items stored in the hash table. */
    PInt32                          first_chain;
        /**< First non-empty collision chain (bucket). */
    PInt32                          last_chain;
        /**< Last non-empty collision chain (bucket). */
    PInt32                          flags;
        /**< Flags. */

    }; // HashTable

} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_LIB_HASH_TABLE_CONTAINER_H
