/** @file list_container.cpp
 * List container implementation.
 */

/*
 * begin        : Wed Feb 19 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "core.h"
#include "lib/shared/list_container.h"

namespace Demo {
namespace Lib {


// Implementation of ListBase::ListEntry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::ListEntry::initialize()
    {
    }


Pointer<ListBase::ListEntry> ListBase::ListEntry::erase()
    {
    Pointer<ListEntry> old_next = next;

    prev->next = next;
    next->prev = prev;

    next = prev = this;

    return old_next;
    }


WeakPointer<ListBase::ListEntry> ListBase::ListEntry::get_next()
    {
    return next;
    }


// Implementation of ListBase::WeakListEntry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::WeakListEntry::initialize
    (
    const WeakPointer<Object> & stored_object
    )
    {
    ListEntry::initialize();
    object = stored_object;
    }


WeakPointer<Object> ListBase::WeakListEntry::current()
    {
    return object;
    }


// Implementation of ListBase::StrongListEntry
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::StrongListEntry::initialize
    (
    const WeakPointer<Object> & stored_object
    )
    {
    ListEntry::initialize();
    object = stored_object;
    }


WeakPointer<Object> ListBase::StrongListEntry::current()
    {
    return object;
    }


// Implementation of ListBase::ListEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::ListEnumerator::initialize
    (
    const Pointer<ListBase> &   the_list,
    BeginTag
    )
    {
    list = the_list;
    tail = list->tail;
    entry = tail->next;
    }


void ListBase::ListEnumerator::initialize
    (
    const Pointer<ListBase> &   the_list,
    EndTag
    )
    {
    list = the_list;
    tail = list->tail;
    entry = tail;
    }


bool ListBase::ListEnumerator::is_end()
    {
    return entry == tail;
    }


bool ListBase::ListEnumerator::is_begin()
    {
    return entry == tail->next;
    }


void ListBase::ListEnumerator::next()
    {
    hope( !is_end() );
    entry = entry->next;
    }


void ListBase::ListEnumerator::prev()
    {
    hope( !is_begin() );
    entry = entry->prev;
    }


void ListBase::ListEnumerator::erase()
    {
    hope( !is_end() );

    // Check that we are not trying to remove an entry that was already removed
    // from the list.
    hope( entry->next != entry );
    hope( entry->prev != entry );
    
    entry->erase();
    entry = tail;
    list->num_items--;
    }


void ListBase::ListEnumerator::erase_and_next()
    {
    hope( !is_end() );

    // Check that we are not trying to remove an entry that was already removed
    // from the list.
    hope( entry->next != entry );
    hope( entry->prev != entry );
    
    entry = entry->erase();
    list->num_items--;
    }


WeakPointer<Object> ListBase::ListEnumerator::current()
    {
    return entry->current();
    }


// Implementation of ListBase::ListRangeEnumerator
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::ListRangeEnumerator::initialize
    (
    const Pointer<ListBase> &   list,
    const Pointer<Object> &     search_key
    )
    {
    ListEnumerator::initialize( list, BeginTag() );
    key = search_key;

    while( !is_end() && !list->equal( current(), key ) )
        {
        ListEnumerator::next();
        }

    begin = entry;
    }


bool ListBase::ListRangeEnumerator::is_begin()
    {
    return entry == begin;
    }


void ListBase::ListRangeEnumerator::next()
    {
    ListEnumerator::next();

    while( !is_end() && !list->equal( current(), key ) )
        {
        ListEnumerator::next();
        }
    }


void ListBase::ListRangeEnumerator::prev()
    {
    ListEnumerator::prev();

    while( !is_begin() && !list->equal( current(), key ) )
        {
        ListEnumerator::prev();
        }
    }


// Implementation of ListBase
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

void ListBase::initialize()
    {
    WeakPointer<Object> null;

    tail = create_entry( null );
    tail->next = tail->prev = tail;

    num_items = 0;
    }


int ListBase::size()
    {
    return num_items;
    }


void ListBase::clear()
    {
    initialize();
    }


Pointer<Enumerator> ListBase::begin()
    {
    hope( !is_replica() );
    return CreateObject<ListEnumerator>( this, BeginTag() );
    }


Pointer<Enumerator> ListBase::end()
    {
    hope( !is_replica() );
    return CreateObject<ListEnumerator>( this, EndTag() );
    }


bool ListBase::contains
    (
    const WeakPointer<Object> & object
    )
    {
    hope( object != null ); // Trying to erase tail.
    Pointer<ListEntry> entry = tail->next;

    while( entry != tail )
        {
        if( equal( entry->current(), object ) )
            {
            return true;
            }

        entry = entry->next;
        }

    return false;
    }


bool ListBase::erase
    (
    const WeakPointer<Object> & object
    )
    {
    hope( object != null ); // Trying to erase tail.
    Pointer<ListEntry> entry = tail->next;

    while( entry != tail )
        {
        if( equal( entry->current(), object ) )
            {
            entry->erase();
            hope( num_items > 0 );
            num_items--;
            return true;
            }

        entry = entry->next;
        }

    return false;
    }


int ListBase::erase_all
    (
    const WeakPointer<Object> & object
    )
    {
    Pointer<ListEntry> entry = tail->next;
    int count = 0;

    while( entry != tail )
        {
        if( equal( entry->current(), object ) )
            {
            entry = entry->erase();
            count++;
            }
        else
            {
            entry = entry->next;
            }
        }

    hope( num_items >= count );
    num_items -= count;
#ifdef OPTION_DEBUG
    if( num_items == 0 )
        {
        hope( tail->next == tail );
        hope( tail->prev == tail );
        }
#endif
    return true;
    }


Pointer<Enumerator> ListBase::equal_range
    (
    const WeakPointer<Object> & object
    )
    {
    return CreateObject<ListRangeEnumerator>( this, object );
    }


bool ListBase::insert
    (
    const WeakPointer<Object> & object
    )
    {
    push_back( object );
    return true;
    }


void ListBase::insert
    (
    const WeakPointer<Object> & object,
    Pointer<Enumerator> &       enumerator
    )
    {
    push_back( object );

    enumerator = end();
    enumerator->prev();
    }


void ListBase::push_back
    (
    const WeakPointer<Object> & object
    )
    {
    internal_insert_before( tail, object );
    }


void ListBase::push_front
    (
    const WeakPointer<Object> & object
    )
    {
    internal_insert_after( tail, object );
    }


void ListBase::insert_before
    (
    const Pointer<Enumerator> & before_this,
    const WeakPointer<Object> & object
    )
    {
    Pointer<ListEnumerator> enumerator = before_this.convert();
    internal_insert_before( enumerator->entry, object );
    }


void ListBase::insert_after
    (
    const Pointer<Enumerator> & after_this,
    const WeakPointer<Object> & object
    )
    {
    Pointer<ListEnumerator> enumerator = after_this.convert();
    internal_insert_after( enumerator->entry, object );
    }


WeakPointer<ListBase::ListEntry> ListBase::get_first()
    {
    return tail->next;
    }


WeakPointer<ListBase::ListEntry> ListBase::get_last()
    {
    return tail;
    }


void ListBase::internal_insert_before
    (
    const Pointer<ListEntry> &  before_this,
    const WeakPointer<Object> & object
    )
    {
    Pointer<ListEntry> entry = create_entry( object );

    entry->next = before_this;
    entry->prev = before_this->prev;
    entry->prev->next = entry;
    entry->next->prev = entry;

    num_items++;
    }


void ListBase::internal_insert_after
    (
    const Pointer<ListEntry> &  after_this,
    const WeakPointer<Object> & object
    )
    {
    Pointer<ListEntry> entry = create_entry( object );

    entry->next = after_this->next;
    entry->prev = after_this;
    entry->next->prev = entry;
    entry->prev->next = entry;

    num_items++;
    }


// Implementation of List
// ~~~~~~~~~~~~~~~~~~~~~~

Pointer<List::ListEntry> List::create_entry
    (
    const WeakPointer<Object> & object
    )
    {
    return CreateObject<StrongListEntry>( object );
    }


bool List::equal
    (
    const WeakPointer<Object> & object1,
    const WeakPointer<Object> & object2
    ) const
    {
    return object1->equals( object2 );
    }


// Implementation of WeakList
// ~~~~~~~~~~~~~~~~~~~~~~~~~~

Pointer<List::ListEntry> WeakList::create_entry
    (
    const WeakPointer<Object> & object
    )
    {
    return CreateObject<WeakListEntry>( object );
    }


bool WeakList::equal
    (
    const WeakPointer<Object> & object1,
    const WeakPointer<Object> & object2
    ) const
    {
    return object1 == object2;
    }


} // namespace Demo::Lib
} // namespace Demo
