/** @file list_container.h
 * List container.
 */

/*
 * begin        : Wed Feb 19 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_LIST_CONTAINER_H
#define DEMO_LIST_CONTAINER_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif

#ifndef DEMO_CONSTAINER_H
#include "lib/shared/container.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Base class for a "list" container.
 * The list is implemented as a double-linked list with tail.
 * Erase operations invalidate enumerators pointing to the removed
 * items. Range enumerators (returned by @c equal_range()) are invalidated
 * by both erase and insert operations.
 * Can not be instantiated as a member object or on the stack.
 */
class ListBase : public SequenceContainer
    {
public:

    MASSIV_OBJECT( ListBase );

    struct BeginTag
        {
        };

    struct EndTag
        {
        };

    class ListEnumerator;

    /**
     * List entry represents an entry in double-linked
     * cyclic list with @c tail.
     */
    class ListEntry : public Object
        {
    public:
            
        MASSIV_OBJECT( ListEntry );
        friend class ListBase;
        friend class ListEnumerator;
        
    public:

        /** Initializes object. */
        void initialize();

    protected:

        /** Removes entry from the list. */
        Pointer<ListEntry> erase();

    public:

        /** Returns pointer to the stored object. */
        virtual WeakPointer<Object> current() = 0;

        /** Returns pointer to the next entry. */
        WeakPointer<ListEntry> get_next();

    protected:

        PPointer<ListEntry> next;
            /**< Pointer to the next entry. */
        PPointer<ListEntry> prev;
            /**< Pointer to the previous entry. */
        }; // ListEntry

    /**
     * ListEntry used by WeakList.
     */
    class WeakListEntry : public ListEntry
        {
    public:
        
        MASSIV_OBJECT( WeakListEntry );

    public:

        /** Initializes object. */
        void initialize
            (
            const WeakPointer<Object> & stored_object
            );

        /**
         */
        virtual WeakPointer<Object> current();

    protected:

        PWeakPointer<Object>    object;
            /**< Weak pointer to stored object. */
        }; // WeakListEntry

    /**
     * ListEntry used by List.
     */
    class StrongListEntry : public ListEntry
        {
    public:
        
        MASSIV_OBJECT( StrongListEntry );
    
    public:

        /** Initializes object. */
        void initialize
            (
            const WeakPointer<Object> & stored_object
            );

        /**
         */
        virtual WeakPointer<Object> current();

    protected:

        PPointer<Object>    object;
            /**< Strong pointer to stored object. */
        }; // StrongListEntry

    /**
     * ListEnumerator enumerates elements stored in the list.
     */
    class ListEnumerator : public Enumerator
        {
    public:

        MASSIV_OBJECT( ListEnumerator );
        friend class ListBase;

        //
        // Initialization.
        //

        /** Initializes enumerator to point to the beginning of @c the_list. */
        void initialize
            (
            const Pointer<ListBase> &   the_list,
            BeginTag = BeginTag()
            );

        /** Initializes enumerator to point to the end of @c the_list. */
        void initialize
            (
            const Pointer<ListBase> &   the_list,
            EndTag
            );

    public:

        //
        // BasicEnumerator implementation.
        //

        /**
         */
        virtual bool is_end();

        /**
         */
        virtual bool is_begin();

        /**
         */
        virtual void next();

        /**
         */
        virtual void prev();

        /**
         */
        virtual void erase();

        /**
         */
        virtual void erase_and_next();

    public:

        //
        // Enumerator implementation.
        //

        /**
         */
        virtual WeakPointer<Object> current();

    protected:

        PPointer<ListEntry> entry;
            /**< Current list entry. */
        PPointer<ListEntry> tail;
            /**< List tail. */
        PPointer<ListBase>  list;
            /**< Pointer to list. */
        }; // ListEnumerator

    /**
     * ListRangeEnumerator enumerates list items that are equal
     * to @c key.
     */
    class ListRangeEnumerator : public ListEnumerator
        {
    public:

        MASSIV_OBJECT( ListRangeEnumerator );

        //
        // Initialization.
        //

        /**
         * Creates enumerator that points the beginning of sequence
         * of items that are equal to @c search_key.
         */
        void initialize
            (
            const Pointer<ListBase> &   list,
            const Pointer<Object> &     search_key
            );

    public:

        //
        // ListEnumerator extensions.
        //

        /**
         */
        virtual bool is_begin();

        /**
         */
        virtual void next();

        /**
         */
        virtual void prev();

    protected:

        PWeakPointer<Object>    key;
            /**< Key that is searched. */
        PPointer<ListEntry>     begin;
            /**< First entry that match @c key. */
        }; // ListRangeEnumerator

    friend class ListEnumerator;

public:

    //
    // Initialization.
    //

    /**
     * Initializes object.
     */
    void initialize();

public:

    //
    // BasicContainer implementation.
    //

    /**
     */
    virtual int size();

    /**
     */
    virtual void clear();

public:

    //
    // SimpleContainer implementation.
    //

    /**
     */
    virtual Pointer<Enumerator> begin();

    /**
     */
    virtual Pointer<Enumerator> end();

    /**
     */
    virtual bool contains
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual bool erase
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual int erase_all
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual Pointer<Enumerator> equal_range
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual bool insert
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual void insert
        (
        const WeakPointer<Object> & object,
        Pointer<Enumerator> &       enumerator
        );

public:

    //
    // SequenceContainer implementation.
    //

    /**
     */
    virtual void push_back
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual void push_front
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual void insert_before
        (
        const Pointer<Enumerator> & before_this,
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual void insert_after
        (
        const Pointer<Enumerator> & after_this,
        const WeakPointer<Object> & object
        );

public:

    //
    // Extensions.
    //

    /**
     * Returns a pointer to the first entry stored in the list.
     *
     * Together with @c end() and @c ListEntry::get_next() can be used
     * to enumerate entries stored in the list and provides an alternative
     * enumeration method to more powerful ListEnumerator.
     *
     * Since generic ListEnumerator can not be used to enumerate objects
     * stored in list replicas, this is the only way how such objects can
     * be retrieved. However one has to be careful since the replica entries
     * are volatile and strong pointers can not be used to reference replica
     * entries.
     *
     * @code
     *  // The following code works as long as list->is_consistent() and
     *  // the list replica is not updated during the enumeration
     *  // (report must not invoke a SRPC).
     *
     *  WeakPointer<ListEntry> e = list->get_first();
     *  while( e != list->get_last()() )
     *      {
     *      report( e->current() );
     *      e = e->get_next();
     *      }
     * @endcode
     */
    WeakPointer<ListEntry> get_first();

    /**
     * Returns pointer to the past of the end of the sequence entry.
     */
    WeakPointer<ListEntry> get_last();

private:

    //
    // Internal stuff.
    //

    /** Creates list entry linked to @c object. */
    virtual Pointer<ListEntry> create_entry
        (
        const WeakPointer<Object> & object
        ) = 0;

    /** Inserts @c object before @c before_this entry in the list. */
    void internal_insert_before
        (
        const Pointer<ListEntry> &  before_this,
        const WeakPointer<Object> & object
        );

    /** Inserts @c object after @c after_this entry in the list. */
    void internal_insert_after
        (
        const Pointer<ListEntry> &  after_this,
        const WeakPointer<Object> & object
        );

protected:

    PPointer<ListEntry> tail;
        /**< List tail. Unused entry. */
    PInt32              num_items;
        /**< Number of items stored in the list. */

    }; // ListBase

/**
 * List container.
 */
class List : public ListBase
    {
private:

    /**
     */
    virtual Pointer<ListEntry> create_entry
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual bool equal
        (
        const WeakPointer<Object> & object1,
        const WeakPointer<Object> & object2
        ) const;

    }; // List

/**
 * Weak list container.
 */
class WeakList : public ListBase
    {
private:

    /**
     */
    virtual Pointer<ListEntry> create_entry
        (
        const WeakPointer<Object> & object
        );

    /**
     */
    virtual bool equal
        (
        const WeakPointer<Object> & object1,
        const WeakPointer<Object> & object2
        ) const;

    }; // WeakList


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_LIST_CONTAINER_H
