/** @file node_console.cpp
 * Node console implementation.
 */

/*
 * begin        : Sat Jun 28 2003
 * copyright    : (C) 2003 by Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "core_management.h"
#include "lib/shared/node_console.h"

namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


NodeConsole::NodeConsole()    
    {
    ConsoleCommand * cmd;

    // Command 'help'.

    cmd = command_list.new_command
        (
        "help",
        "HELP.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_help
        );

    cmd->new_argument
        (
        "Command to show detail info about.",
        CommandArgument::FLAGS_STRING | CommandArgument::FLAGS_ARG_OPTIONAL
        );

    // Command 'stat'.

    cmd = command_list.new_command
        (
        "stat",
        "Print statistics of a specified subsystem.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_stat
        );

    cmd->new_argument
        (
        "Subsystem to show statistics.",
        CommandArgument::FLAGS_STATISTICS | CommandArgument::FLAGS_ARG_OPTIONAL,
        "all"
        );


    // Command 'getvar'.

    cmd = command_list.new_command
        (
        "getvar",
        "Returns config variable.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_getvar
        );

    cmd->new_argument
        (
        "Config variable to return.",
        CommandArgument::FLAGS_CONFIG_VARIABLE
        );


    // Command 'setvar'.

    cmd = command_list.new_command
        (
        "setvar",
        "Sets config variable.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_setvar
        );

    cmd->new_argument
        (
        "Config variable to set.",
        CommandArgument::FLAGS_CONFIG_VARIABLE_SET
        );

    cmd->new_argument
        (
        "New value.",
        CommandArgument::FLAGS_STRING
        );


    // Command 'createobject'.

    cmd = command_list.new_command
        (
        "createobject",
        "Creates instance of given class and returns its object id.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_createobject
        );

    cmd->new_argument
        (
        "Name of an instantiable class.",
        CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT
        );


    // Command 'eraseobject'.

    cmd = command_list.new_command
        (
        "eraseobject",
        "Erases object.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_eraseobject
        );

    cmd->new_argument
        (
        "Object to erase.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'gc'.

    cmd = command_list.new_command
        (
        "gc",
        "Initiates next GC run.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_gc
        );


    // Command 'objects'.

    cmd = command_list.new_command
        (
        "objects",
        "Dumps registered objects that match given class name.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_objects
        );

    cmd->new_argument
        (
        "Class name pattern.",
        CommandArgument::FLAGS_WILDCARD | CommandArgument::FLAGS_ARG_OPTIONAL,
        "*"
        );

    cmd->new_argument
        (
        "Include subclasses?",
        CommandArgument::FLAGS_BOOLEAN | CommandArgument::FLAGS_ARG_OPTIONAL,
        "true"
        );


    // Command 'rgroup'.

    cmd = command_list.new_command
        (
        "rgroup",
        "Dumps replication group of a given object.\n",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_rgroup
        );

    cmd->new_argument
        (
        "Object whose group should be dumped.",
        CommandArgument::FLAGS_OBJECT_ID
        );

    cmd->new_argument
        (
        "Replication flags: repflags_yes.",
        CommandArgument::FLAGS_INTEGER
        );

    cmd->new_argument
        (
        "Replication flags: repflags_no.",
        CommandArgument::FLAGS_INTEGER
        );


    // Command 'mgroup'.

    cmd = command_list.new_command
        (
        "mgroup",
        "Dumps migration group of given object.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_mgroup
        );

    cmd->new_argument
        (
        "Object whose group should be dumped.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'snapmemory'.

    cmd = command_list.new_command
        (
        "snapmemory",
        "Saves current memory \"snapshot\".",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_snapmemory
        );

    cmd->new_argument
        (
        "Snapshot file name.",
        CommandArgument::FLAGS_STRING
        );


    // Command 'dump'.

    cmd = command_list.new_command
        (
        "dump",
        "Dumps property.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_dump
        );

    cmd->new_argument
        (
        "Property that should be dumped.",
        CommandArgument::FLAGS_PROPERTY
        );


    // Command 'get'.

    cmd = command_list.new_command
        (
        "get",
        "Returns property in a textual serialization format.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_get
        );

    cmd->new_argument
        (
        "Property that should be printed.",
        CommandArgument::FLAGS_PROPERTY
        );


    // Command 'set'.

    cmd = command_list.new_command
        (
        "set",
        "Initializes property from a textual serialization.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_set
        );

    cmd->new_argument
        (
        "Property that should be set.",
        CommandArgument::FLAGS_PROPERTY
        );

    cmd->new_argument
        (
        "Textual serialization.",
        CommandArgument::FLAGS_STRING
        );


    // Command 'copy'.

    cmd = command_list.new_command
        (
        "copy",
        "Copies property.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_copy
        );

    cmd->new_argument
        (
        "Property that should be read.",
        CommandArgument::FLAGS_PROPERTY
        );

    cmd->new_argument
        (
        "Property that should be set.",
        CommandArgument::FLAGS_PROPERTY
        );


    // Command 'call'.

    cmd = command_list.new_command
        (
        "call",
        "Calls method on an object.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_call
        );

    cmd->new_argument
        (
        "Method that should be called. An expression that evaluates to a "
        "component of a local object.",
        CommandArgument::FLAGS_METHOD
        );

    cmd->new_argument
        (
        "Method arguments (textual serialization).",
        CommandArgument::FLAGS_STRING
        );


    // Command 'rcall'.

    cmd = command_list.new_command
        (
        "rcall",
        "Calls rpc method on an object synchronously and returns result.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_rcall
        );

    cmd->new_argument
        (
        "Object id of an object.",
        CommandArgument::FLAGS_OBJECT_ID
        );

    cmd->new_argument
        (
        "Class interface.",
        CommandArgument::FLAGS_CLASS_NAME
        );

    cmd->new_argument
        (
        "Method that should be called.",
        CommandArgument::FLAGS_STRING
        );

    cmd->new_argument
        (
        "Method arguments (textual serialization).",
        CommandArgument::FLAGS_STRING
        );

    // Command 'get_naming_service'.

    cmd = command_list.new_command
        (
        "get_naming_service",
        "Return naming service object id.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_get_naming_service
        );


    // Command 'echo'.

    cmd = command_list.new_command
        (
        "echo",
        "Print and return parameter %1.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_echo
        );

    cmd->new_argument
        (
        "Arbitrary string.",
        CommandArgument::FLAGS_STRING
        );

    
    // Command 'classinfo'.

    cmd = command_list.new_command
        (
        "classinfo",
        "Shows info about a class.",
        0,
        &execute_classinfo
        );

    cmd->new_argument
        (
        "Class name.",
        CommandArgument::FLAGS_CLASS_NAME
        );


    // Command 'name'.

    cmd = command_list.new_command
        (
        "name",
        "Names object.",
        0,
        &execute_name
        );

    cmd->new_argument
        (
        "Object to name.",
        CommandArgument::FLAGS_OBJECT_ID
        );

    cmd->new_argument
        (
        "Name to be assigned.",
        CommandArgument::FLAGS_STRING
        );


    // Command 'savesettings'.

    cmd = command_list.new_command
        (
        "savesettings",
        "Saves settings to file massiv.settings.conf.",
        0,
        &execute_savesettings
        );


    // Command 'archive'.

    cmd = command_list.new_command
        (
        "archive",
        "Creates new archive.",
        0,
        &execute_archive
        );


    // Command 'swapout'.

    cmd = command_list.new_command
        (
        "swapout",
        "Swaps object out.",
        0,
        &execute_swapout
        );

    cmd->new_argument
        (
        "Object to swap out.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'swapin'.

    cmd = command_list.new_command
        (
        "swapin",
        "Swaps objectin.",
        0,
        &execute_swapin
        );

    cmd->new_argument
        (
        "Object to swap in.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'swapinall'.

    cmd = command_list.new_command
        (
        "swapinall",
        "Swaps all objects in.",
        0,
        &execute_swapinall
        );


    // Command 'pin'.

    cmd = command_list.new_command
        (
        "pin",
        "Pins local object.",
        0,
        &execute_pin
        );

    cmd->new_argument
        (
        "Local object to pin.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'unpin'.

    cmd = command_list.new_command
        (
        "unpin",
        "Unpins local object.",
        0,
        &execute_unpin
        );

    cmd->new_argument
        (
        "Local object to unpin.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'migrate'.

    cmd = command_list.new_command
        (
        "migrate",
        "Migrates local object without notifying it.",
        0,
        &execute_migrate
        );

    cmd->new_argument
        (
        "Local object to migrate.",
        CommandArgument::FLAGS_OBJECT_ID
        );
    
    cmd->new_argument
        (
        "Object to migrate to.",
        CommandArgument::FLAGS_OBJECT_ID
        );


    // Command 'createpinned'.

    cmd = command_list.new_command
        (
        "createpinned",
        "Creates pinned object with a given local name and returns its object id.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_createpinned
        );

    cmd->new_argument
        (
        "Class name.",
        CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT
        );

    cmd->new_argument
        (
        "Local object name.",
        CommandArgument::FLAGS_STRING | CommandArgument::FLAGS_ARG_OPTIONAL
        );

    
    // Command 'createbound'.

    cmd = command_list.new_command
        (
        "createbound",
        "Creates instance of a class and writes pointer to it to a property. "
        "Returns its object id.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_createbound
        );
    
    cmd->new_argument
        (
        "Class name.",
        CommandArgument::FLAGS_CLASS_NAME_NOT_ABSTRACT
        );

    cmd->new_argument
        (
        "Pointer property the pointer to created object will be written to.",
        CommandArgument::FLAGS_PROPERTY
        );
    
    
    // Command 'data_update'.

    cmd = command_list.new_command
        (
        "data_update",
        "Update data on master data service.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_data_update
        );

    
    // Command 'log'.

    cmd = command_list.new_command
        (
        "log",
        "Log system messages from current node to my console.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_log
        );
    
    cmd->new_argument
        (
        "Enable/disable logging.",
        CommandArgument::FLAGS_BOOLEAN | CommandArgument::FLAGS_ARG_OPTIONAL,
        "true"
        );
    
    
    // Command 'server_quit'.
    
    cmd = command_list.new_command
        (
        "server_quit",
        "Archivate simulation and then shutdown all server nodes.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_server_quit
        );
        
        
    // Command 'rcommand'.
    
    cmd = command_list.new_command
        (
        "rcommand",
        "Send command to the node where given object currently lives and then "
        "process the command locally. "
        "Operates asynchronously, does not update environment and does not "
        "return method result.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_rcommand
        );

    cmd->new_argument
        (
        "Migrate the command to object.",
        CommandArgument::FLAGS_OBJECT_ID
        );

    cmd->new_argument
        (
        "Command to be processed.",
        CommandArgument::FLAGS_STRING
        );

    
    // Command 'gather_replicated_objects'

    cmd = command_list.new_command
        (
        "gather_replicated",
        "Prints information about replicated objects owned by this node. "
        "Only those objects whose class name matches a selected pattern are "
        "dumped.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_gather_replicated
        );
    
    cmd->new_argument
        (
        "Class name pattern.",
        CommandArgument::FLAGS_WILDCARD | CommandArgument::FLAGS_ARG_OPTIONAL,
        "*"
        );

    cmd->new_argument
        (
        "Include subclasses?",
        CommandArgument::FLAGS_BOOLEAN | CommandArgument::FLAGS_ARG_OPTIONAL,
        "true"
        );

    
    // Command 'replication_info'
    
    cmd = command_list.new_command
        (
        "replication_info",
        "Print replication related information about a specified object.",
        ConsoleCommand::FLAGS_ROOT_ACCOUNT,
        &execute_replication_info
        );
        
    cmd->new_argument
        (
        "The object to dump the information about.",
        CommandArgument::FLAGS_OBJECT_ID
        );
    }


NodeConsole::~NodeConsole()
    {
    }


} // namespace Demo::Lib
} // namespace Demo
