/** @file node_console_commands.cpp
 * Node console commands implementation.
 */

/*
 * begin        : Mon Feb 17 2003
 * copyright    : (C) 2003 by Marek Vondrak, Ondrej Pecta
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "core_management.h"
#include "lib/shared/node_console.h"
#include "lib/shared/account_object.h"
#include "lib/shared/account_object_rpc.h"
#include "lib/shared/node_object.h"
#include "node/node_action_manager.h" // Should be included in core_management.h?
#include <sstream>
#include <iomanip>


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


void NodeConsole::execute_help
    (
    const CommandContext & context
    )
    {
    const std::string & cmd = context.get_string( 0 );
    std::string         help;

    if( cmd.size() )
        {
        help = "Detail info about command " + cmd + "\n"
            + context.console->print_command_help( cmd );
        }
    else
        {
        help = "Commands supported by node console:\n"
            + context.console->print_command_help();
        }

    context.print( help );
    }


void NodeConsole::execute_stat
    (
    const CommandContext &  context
    )
    {
    std::ostringstream          ss;

    Registry &                  registry = Global::registry();
    Registry::RegistryNode *    node = context.get_regnode( 0 );
    const std::string &         subsystem = context.get_string( 0 );
    Registry::RegistryNode *    current_node = NULL;
    ConfigVariable *            cvar;

    Registry::SearchContext     sc( &registry );

    registry.initialize_search_context( sc, *node );
    while( registry.find_variable( 0, sc, *node ) )
        {
        if( current_node != sc.get_node() )
            {
            ss << "\n[ ";
            if( sc.get_path() == "" )
                ss << subsystem;
            else
                ss << sc.get_path();

            ss << " statistics: ]\n";
            }
        cvar = &sc.get_variable();

        ss << sc.get_name() << " : " << cvar->read_string() << "\n";
        current_node = sc.get_node();
        }
    context.print( ss.str() );
    }


void NodeConsole::execute_getvar
    (
    const CommandContext &  context
    )
    {
    std::ostringstream      ss;
    const std::string &     variable = context.get_string( 0 );
    ConfigVariable *        cvar = context.get_cvar( 0 );
    unsigned int            pos;
    std::string             var = variable;

    try
        {
        pos = variable.find_last_of( '/' );
        if( pos != variable.npos )
            {
            var = variable.substr( pos + 1, variable.size() - pos - 1 );
            }

        ss << quote_const_string( cvar->read_string() );

        context.print( var + " : " + ss.str() );
        context.result = ss.str();
        }
    catch( Exception & )
        {
        ss << "Error printing variable " << variable;
        context.print( ss.str() );
        }
    }


void NodeConsole::execute_setvar
    (
    const CommandContext &  context
    )
    {
    std::ostringstream          ss;
    const std::string &         variable = context.get_string( 0 );
    const std::string &         new_value = context.get_string( 1 );
    Registry &                  registry = Global::registry();
    std::string                 base_registry_node_name = variable;
    Registry::RegistryNode *    node;
    ConfigVariable *            cvar;
    unsigned int                pos;
    std::string                 var = variable;

    Registry::SearchContext     sc( &registry );

    try
        {
        pos = variable.find_last_of( "/" );

        node = registry.find_node_for_path( variable.substr( 0, pos ) );
        var = variable.substr( pos + 1, variable.size() - pos - 1 );
        ss << "[ " << variable.substr( 0, pos ) << " ]\n";

        registry.initialize_search_context( sc, *node );
        registry.find_variable( var.c_str(), sc, *node );

        cvar = &sc.get_variable();

        ss << "Old value: " << sc.get_name() << " : " << cvar->read_string() << "\n";
        cvar->write_string( new_value );
        ss << "New value: " << sc.get_name() << " : " << cvar->read_string() << "\n";
        context.print( ss.str() );
        }
    catch( ... )
        {
        ss << "Error finding variable " << variable << ".";
        context.print( ss.str() );
        }
    }


void NodeConsole::execute_createobject
    (
    const CommandContext &  context
    )
    {
    const std::string &     class_name = context.get_string( 0 );
    ClassManager &          class_manager = Global::class_manager();
    const ObjectFactory *   factory = class_manager.get_factory( class_name );
    PtrToObject             object = factory->create_object();
    std::ostringstream      ss;

    ss << "Object " << class_name << "created\n:" << *object;
    context.print( ss.str() );
    context.result = object->get_object_id().to_string();
    }


void NodeConsole::execute_createpinned
    (
    const CommandContext &     context
    )
    {
    const std::string &     class_name = context.get_string( 0 );
    const std::string &     name = context.get_string( 1 );
    ClassManager &          class_manager = Global::class_manager();
    const ObjectFactory *   factory = class_manager.get_factory( class_name );
    PtrToObject             object = factory->create_object();
    std::ostringstream      ss;

    object->pin();

    ss << "Object " << class_name << " created and pinned:\n" << *object;

    // Name object.

    if( name.size() )
        {
        Global::garbage_collector().name_object
            (
            object->get_object_id(),
            name
            );
        ss << "\nObject was named \"" << name << "\".\n";
        }
    else
        {
        ss << "\nObject was not named.\n";
        }

    context.print( ss.str() );
    context.result = object->get_object_id().to_string();
    }


void NodeConsole::execute_createbound
    (
    const CommandContext &     context
    )
    {
    const std::string &     class_name = context.get_string( 0 );
    Property *              property = context.get_property( 1 );
    ClassManager &          class_manager = Global::class_manager();
    const ObjectFactory *   factory = class_manager.get_factory( class_name );
    std::ostringstream      ss;

    if( property->get_type() == Property::TYPE_POINTER )
        {
        PtrToObject         object = factory->create_object();
        PPointer<Object>    pointer = object->get_object_id();

        property->assign( pointer );
        context.result = object->get_object_id().to_string();

        ss << "Object " << class_name << " created and assigned:\n" << *object;
        }
    else
        {
        ss << "Property " << property << " is not a pointer.";
        }

    context.print( ss.str() );
    }


void NodeConsole::execute_eraseobject
    (
    const CommandContext &  context
    )
    {
    PtrToObject object = context.get_object( 0 );
    Global::garbage_collector().collect_object( object );
    }


void NodeConsole::execute_gc
    (
    const CommandContext &  context
    )
    {
    Global::garbage_collector().force_run();
    }


void NodeConsole::execute_objects
    (
    const CommandContext &  context
    )
    {
    const std::string & pattern = context.get_string( 0 );
    const bool &        include_subclasses = context.get_boolean( 1 );
    std::ostringstream  ss;

    Global::garbage_collector().dump_registered_objects
        (
        ss,
        pattern,
        include_subclasses
        );

    context.print( ss.str() );
    }


void NodeConsole::execute_rgroup
    (
    const CommandContext &  context
    )
    {
    const ObjectId &    object_id = context.get_object_id( 0 );
    const Int64 &       repflags_yes = context.get_integer( 1 );
    const Int64 &       repflags_no = context.get_integer( 2 );
    std::ostringstream  ss;

    Global::garbage_collector().dump_replication_group
        (
        ss,
        object_id,
        static_cast<int>( repflags_yes ),
        static_cast<int>( repflags_no )
        );

    context.print( ss.str() );
    }


void NodeConsole::execute_mgroup
    (
    const CommandContext &  context
    )
    {
    const ObjectId &    object_id = context.get_object_id( 0 );
    std::ostringstream  ss;

    Global::garbage_collector().dump_migration_group
        (
        ss,
        object_id
        );

    context.print( ss.str() );
    }


void NodeConsole::execute_snapmemory
    (
    const CommandContext &  context
    )
    {
    const std::string & file_name = context.get_string( 0 );

    Global::garbage_collector().save_memory_snapshot( file_name );
    }


void NodeConsole::execute_dump
    (
    const CommandContext &  context
    )
    {
    Property *          property = context.get_property( 0 );
    std::ostringstream  ss;

    if( property->get_type() == Property::TYPE_OBJECT_PROPERTY )
        {
        // Ensure proper pinning.

        PtrToObject object = property->cast_to_object();
        ss << *object;
        }
    else
        {
        // Property can not (and need not) be pinned.

        ss << *property;
        }

    context.print( ss.str() );
    }


void NodeConsole::execute_get
    (
    const CommandContext &  context
    )
    {
    Property *      property = context.get_property( 0 );
    std::string     content = property->get();
    context.print( content );
    context.result = content;
    }


void NodeConsole::execute_set
    (
    const CommandContext &  context
    )
    {
    Property *  property = context.get_property( 0 );
    property->set( context.get_string( 1 ) );
    }


void NodeConsole::execute_copy
    (
    const CommandContext &  context
    )
    {
    Property *  source_property = context.get_property( 0 );
    Property *  target_property = context.get_property( 1 );

    target_property->assign( *source_property );
    }


void NodeConsole::execute_call
    (
    const CommandContext &  context
    )
    {
    const VariantPointer &                  component = context.get_component( 0 );
    const MetaObject *                      metaobject = context.get_metaobject( 0 );
    const MetaObject::MethodInfoIterator &  method = context.get_method( 0 );
    const std::string &                     method_args = context.get_string( 1 );
    std::istringstream                      stream( method_args );
    TextReader                              tr( stream );

    tr.next_line();

    metaobject->call_method
        (
        component,
        method,
        tr
        );
    }


void NodeConsole::execute_rcall
    (
    const CommandContext &  context
    )
    {
    const ObjectId &    object = context.get_object_id( 0 );
    const MetaObject *  metaobject =
        Global::class_manager().get_metaobject( context.get_string( 1 ) );
    const std::string & method = context.get_string( 2 );
    const std::string & method_args = context.get_string( 3 );
    std::istringstream  stream( method_args );
    TextReader          tr( stream );

    tr.next_line();

    std::auto_ptr<MethodPacket> mp( metaobject->remote_call_method
        (
        object,
        method,
        tr
        ) );

    try
        {
        context.result = mp->get_argument_value( -1 );
        }
    catch( Exception & )
        {
        }
    }


void NodeConsole::execute_get_naming_service
    (
    const CommandContext & context
    )
    {
    const ObjectId & ns = Massiv::System::get_naming_service_object_id();
    std::ostringstream  ss;
    ss << ns;
    context.result = ss.str();
    }


void NodeConsole::execute_echo
    (
    const CommandContext &  context
    )
    {
    std::string echo = context.get_string( 0 );
    context.print( echo );
    context.result = echo;
    }


void NodeConsole::execute_classinfo
    (
    const CommandContext &  context
    )
    {
    const MetaObject *  metaobject = context.get_metaobject( 0 );
    std::ostringstream  ss;

    ss << *metaobject;
    context.print( ss.str() );
    }


void NodeConsole::execute_name
    (
    const CommandContext &  context
    )
    {
    const ObjectId &    object_id = context.get_object_id( 0 );
    const SString &     name = context.get_string( 1 );

    Global::garbage_collector().name_object
        (
        object_id,
        name
        );
    }


void NodeConsole::execute_savesettings
    (
    const CommandContext &  context
    )
    {
    const std::string           file =
        Global::path_manager().get_work_directory() + "massiv.settings.conf";
    Registry &                  registry = Global::registry();
    Registry::RegistryNode *    node = registry.find_node_for_path( "Settings" );

    try
        {
        registry.save_config_file( file, *node );
        context.print( "Settings succefully saved to " + file + "." );
        }
    catch( ... )
        {
        context.print( "Error saving settings to " + file + "!" );
        }
    }


void NodeConsole::execute_archive
    (
    const CommandContext &  context
    )
    {
    ArchiveManager & archive_manager = Global::archive_manager();

    if( archive_manager.archivation_in_progress() )
        {
        context.print( "Some archivation is already running." );
        return;
        }

    try
        {
        switch( archive_manager.request_start_archivation() )
            {
            case ArchiveManager::RESULT_NODE_IS_NOT_SERVER:
                context.print( "Archivation can be started only on server nodes." );
                break;

            case ArchiveManager::RESULT_ARCHIVATION_IN_PROGRESS:
                context.print( "Archivation started successfully." );
                break;

            case ArchiveManager::RESULT_REQUEST_DELAYED_BECAUSE_OF_SRPC:
                context.print( "Archivation start delayed because of pending sync rpc. "
                    "Archivation will automatically start as soon as possible." );
                break;
            }
        }
    catch( ... )
        {
        context.print( "There was a problem starting archivation!" );
        }
    }


void NodeConsole::execute_swapout
    (
    const CommandContext &  context
    )
    {
    ObjectManager &                 manager = Global::object_manager();
    const ObjectId &                object_id = context.get_object_id( 0 );
    ObjectManager::SwapOutResult    result;

    result = manager.try_to_swap_out( object_id );

    switch( result )
        {
        case ObjectManager::SWAP_OUT_RESULT_SUCCESS:
            context.print( "Swap-out successful." );
            break;

        case ObjectManager::SWAP_OUT_RESULT_OBJECTS_PINNED_OR_STACK_REFERENCED:
            context.print( "Error - object pinned or stack strong referenced!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_ARCHIVATION_IN_PROGRESS:
            context.print( "Error - archivation is in progress!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_NO_SWAP_FILE:
            context.print( "Error - no swap file is opened!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_IS_REPLICA:
            context.print( "Error - replicas cannot be swapped!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_HAS_REPLICA_REQUESTS:
            context.print( "Error - objects with replication requests cannot be swapped!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_UNREGISTERED_OBJECT:
            context.print( "Error - unregistered object!" );
            break;

        case ObjectManager::SWAP_OUT_RESULT_OBJECT_NOT_ARCHIVABLE:
            context.print( "Error - object is not archivable!" );
            break;

        // Do not add default case so we will get compiler warning
        // if new swap out result is added in the future.
        }
    }


void NodeConsole::execute_swapin
    (
    const CommandContext &  context
    )
    {
    ObjectManager &     manager = Global::object_manager();
    const ObjectId &    object_id = context.get_object_id( 0 );
    PtrToObject         object = NULL;

    object = manager.try_to_swap_in( object_id );

    if( object != NULL )
        {
        std::ostringstream ss;
        ss << "Object successfully swapped in:\n";
        ss << *object;
        context.print( ss.str() );
        }
    else
        {
        context.print( "Error - object cannot be swapped in!" );
        }
    }


void NodeConsole::execute_swapinall
    (
    const CommandContext &  context
    )
    {
    ObjectManager & manager = Global::object_manager();

    manager.swap_in_all_objects();
    context.print( "Swapping in all objects..." );
    }


void NodeConsole::execute_pin
    (
    const CommandContext &  context
    )
    {
    PtrToObject object = context.get_object( 0 );

    if( object != NULL )
        {
        object->pin();
        }
    else
        {
        context.print( "Object not local." );
        }

    }


void NodeConsole::execute_unpin
    (
    const CommandContext &  context
    )
    {
    PtrToObject object = context.get_object( 0 );

    if( object != NULL )
        {
        object->unpin();
        }
    else
        {
        context.print( "Object not local." );
        }
    }


void NodeConsole::execute_migrate
    (
    const CommandContext &  context
    )
    {
    PtrToObject         object = context.get_object( 0 );
    WeakPointer<Object> addressee = context.get_object_id( 1 );

    if( object != NULL )
        {
        object->migrate_to( addressee, STime(), false );
        }
    else
        {
        context.print( "Object not local." );
        }
    }


void NodeConsole::execute_data_update
    (
    const CommandContext &  context
    )
    {
    DataManager &       data_manager = Global::data_manager();
    std::ostringstream  ss;

    if( data_manager.request_master_data_update() )
        {
        // Server node.
        ss << "Data update request was sent to master data service.";
        }
    else
        {
        // Client node.
        ss << "Client node cannot send data update request to master data service.";
        }

    context.print( ss.str() );
    }


void NodeConsole::execute_log
    (
    const CommandContext &  context
    )
    {
    bool                    enable = context.get_boolean( 0 );
    Logger &                logger = Global::logger();
    LogDestination *        dest = logger.find_destination( "clients" );
    LogDestinationClients * dest_clients = NULL;

    if( !dest )
        {
        context.print( "Current node does not define any logger rules for clients." );
        return;
        }

    dest_clients = checked_cast<LogDestinationClients *>( dest );

    if( enable )
        {
        dest_clients->subscribe( context.account );
        context.print( "Subscribed." );
        }
    else
        {
        dest_clients->unsubscribe( context.account );
        context.print( "Unsubscribed." );
        }
    }


void NodeConsole::execute_server_quit
    (
    const Console::CommandContext & context
    )
    {
    using Massiv::Core::NodeActionManager;
    using Massiv::Core::NodeAction;

    NodeActionManager & node_action_manager = Global::node_action_manager();

    if( !Massiv::System::is_server() )
        {
        context.print( "Server quit action can be initiated only on server nodes." );
        return;
        }

    if( !node_action_manager.is_ready() )
        {
        context.print( "Can't initiate server quit. Node action manager is not yet ready "
            "to initiate the action." );
        return;
        }

    node_action_manager.initiate_action( NodeAction::SERVER_QUIT );
    }


void NodeConsole::execute_rcommand
    (
    const Console::CommandContext & context
    )
    {
    const ObjectId &        object_id = context.get_object_id( 0 );
    const std::string &     command = context.get_string( 1 );
    Remote<AccountObject>   account = context.account;

    account->async_process_node_command_at_object( object_id, command );
    }


void NodeConsole::execute_gather_replicated
    (
    const Console::CommandContext & context
    )
    {
    const std::string &         pattern = context.get_string( 0 );
    const bool &                include_subclasses = context.get_boolean( 1 );
    ReplicaServer &             replica_server = Global::replica_server();
    ObjectShared::ObjectIdSet   objects;
    std::ostringstream          ss;

    replica_server.gather_replicated_objects( objects, pattern,
        include_subclasses );

    ObjectShared::ObjectIdSet::iterator it_object( objects.begin() );

    if( it_object == objects.end() )
        {
        ss << "There are no replicated objects matching the criteria.";
        }
    else
        {
        ss << "Replicated objects: " << std::endl;

        while( it_object != objects.end() )
            {
            ss << *it_object << '\n';
            ++it_object;
            }
        }

    context.print( ss.str() );
    }


void NodeConsole::execute_replication_info
    (
    const Console::CommandContext & context
    )
    {
    const ReplicaServer &   replica_server = Global::replica_server();
    const ObjectId &        object_id = context.get_object_id( 0 );
    std::ostringstream      result_string;

    replica_server.dump_object_info( result_string, object_id );
    context.print( result_string.str() );
    }


} // namespace Demo::Lib
} // namespace Demo
