/** @file player_action.h
 * Player actions.
 */

/*
 * begin        : Tue Aug 5 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef DEMO_PLAYER_ACTION_H
#define DEMO_PLAYER_ACTION_H


#ifndef DEMO_LIB_H
#include "lib/shared/lib.h"
#endif


namespace Demo {
namespace Lib {

INJECT_CORE_SYMBOLS();


/**
 * Action used to move player entity.
 * @todo Add pointer to layer to verify that the player is at
 *      same layer as it was when the action was generated on client?
 */
class PlayerMoveAction : public ValueType
    {
public:

    /** Default constructor. */
    PlayerMoveAction();

    /** Copy constructor. */
    PlayerMoveAction
        (
        const PlayerMoveAction & other
        );

public:

    PVector2    target_position;
        /**< Position the player wants to move to. */

    }; // PlayerMoveAction


/**
 * Represents an action that manipulates some item
 * from inventory or some ability.
 */
class PlayerItemAction : public ValueType
    {
public:

    /**
     * Specifies source of the action.
     */
    enum Source
        {
        SOURCE_NONE         = 0,
            /**< No source. */

        SOURCE_INVENTORY    = 1,
            /**< Source of the action is some item from inventory. */

        SOURCE_ABILITIES    = 2
            /**< Source of the action is some ability. */
        }; // Source

    /**
     * Specifies target of the action.
     */
    enum Target
        {
        TARGET_INVENTORY    = 0,
            /**< Target of the action is some item or empty slot from inventory. */

        TARGET_ENTITY       = 1,
            /**< Target of the action is an entity.
                 If the player icon is clicked, this target is used too, with player
                 entity object as the target. */

        TARGET_WORLD        = 2
            /**< Target of the action is some position in the world. */

        }; // Target

public:

    /** Default constructor. */
    PlayerItemAction();

    /** Copy constructor. */
    PlayerItemAction
        (
        const PlayerItemAction & other
        );

public:

    PEnum< Source >             source;
        /**< Identification of container holding source item. */

    PInt32                      source_slot_index;
        /**< Slot index where @a source_object is stored.
         * Identifies source item.
         */

    PWeakPointer< Object >      source_object;
        /**< Pointer to source of the action.
         * If @a source is @c SOURCE_INVENTORY, @a source_object points to ClientItem.
         * Otherwise it's undefined.
         *
         * @note This property is used only for verification that the source slot
         *       still contains the same object as the one when the action
         *       was generated.
         */

    PEnum< Target >             target;
        /**< Identification of target of the action. */

    PInt32                      target_slot_index;
        /**< Slot index where @a target_object is stored.
         * Identifies target item if @a target is @c TARGET_INVENTORY
         * or @c TARGET_ABILITIES.
         */

    PWeakPointer< Object >      target_object;
        /**< Pointer to target of the action.
         * If @a target is @c TARGET_INVENTORY, @a target_object points to ClientItem
         * or it's @c NULL if target slot is empty. If @a target is @c TARGET_ENTITY,
         * @a target_object is ClientEntity::entity of the target entity.
         */

    PVector2                    target_position;
        /**< Target position if @a target is @c TARGET_WORLD. */

    }; // PlayerItemAction


} // namespace Demo::Lib
} // namespace Demo

#endif // DEMO_PLAYER_ACTION_H
