/** @file server_main.cpp
 * Demo server.
 */

/*
 * begin        : Wed Feb 19 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

// Core

#include "core.h"
#include "config.h"
#include "cmdline/cmdline_parser.h"
#include "data/path_manager.h"

// Global

#include "server_global.h"

// Server

#include "config.h"
#include "server_main.h"
#include "server_global.h"

// Lib

#include "lib/server/class_list.h"
#include "lib/server/server_node.h"
#include "lib/server/account_object_impl.h"
#include "lib/server/account_object_impl_rpc.h"
#include "lib/server/naming_service.h"

// STD C++

#include <iosfwd>
#include <fstream>


namespace Demo {
namespace Server {

INJECT_CORE_SYMBOLS();

using namespace Massiv;
using namespace Demo::Lib;


// Anonymous namespace.
namespace {


/**
 * Server options.
 */
struct ServerOptions
    {
    ServerOptions() :
        create_startup_archive( false ),
        startup_archive_is_initial_archive( false ),
        startup_archive_provider_id( 1 ),
        pack_archives( false ),
        unpack_archives( false ),
        config_file_name( CONFIG_MASSIV_CONF_PATH )
        {
        }

    bool                        create_startup_archive;
        /**< Whether to create startup archive. */

    bool                        startup_archive_is_initial_archive;
        /**< Whether the startup archive is also initial archive. */

    UInt32                      startup_archive_provider_id;
        /**< Id of object provider which will be created in startup archive. */
    
    bool                        pack_archives;
        /**< Whether pack unpacked archives. */
    
    bool                        unpack_archives;
        /**< Whether unpack a packed archive. */
    
    std::vector< std::string >  directories;
        /**< List of directories. */
    
    std::string                 archive_id;
        /**< Stores ArchivationId in textual format. Specifies what archive
             should be packed / unpacked. */

    std::string                 config_file_name;
        /**< Config file name variable. Can be changed from command line. */

    }; // ServerOptions


std::string read_local_node_config()
    {
    std::string         config_file_name = Massiv::Core::Global::path_manager().
        get_login_directory() + "/default";
    std::ifstream       file( config_file_name.c_str() );
    std::ostringstream  oss;
    oss << file.rdbuf();
    if( !file.good() )
        {
        std::ostringstream ss;
        ss  << "Error reading '" << config_file_name << "' file.";
        throw Massiv::Core::ExIOError( ss.str() );
        }

    return oss.str();
    }


void print_usage()
    {
    std::cout <<
"Massiv demo server.\n"
"Usage: demo_server [options]\n"
"Options:\n"
"-h (-?)            This help.\n"
"-i                 Archive to be created should be 'initial'. That means that\n"
"                   archive will not be empty and global objects will be created\n"
"                   and archived.\n"
"                   To be specified together with -s.\n"
"-s <providerid>    Create startup archive. Object provider with given\n"
"                   <providerid> will be created and written to the archive.\n"
"                   Provider ids must be unique, ie. there can not be two\n"
"                   archives on different servers, both with the same provider\n"
"                   id.\n"
"                   Note that provider id need not be equal to the number of\n"
"                   server, one archive can contain more object providers.\n"
"-p <archivationid> Create new archive with given <archivationid> from one or\n"
"                   more unpacked archives. Use -d to specify directories where\n"
"                   unpacked archives are located.\n"
"-u <archivationid> Unpack archive with the given <archivationid>.\n"
"                   Use -d to specify the directory where to unpack to.\n"
"-d                 Directory containing unpacked archive or where the archive\n"
"                   should be unpacked to.\n"
"                   Can be specified multiple times if more directories should\n"
"                   be packed.\n"
"-c <file name>     Override default config file name.\n"
"\n"
"Examples:\n"
"-s 1 -i            Create initial startup archive. Global objects will be\n"
"                   created by object provider 1 and then archived here.\n"
"-s n               Create empty startup archive. Object provider n will be\n"
"                   stored into the archive.\n"
"-p massiv_0000000001.archive -d c:\\dir\n"
"                   Create archive from unpacked archive stored in c:\\dir.\n"
"-u massiv_0000000001.archive -d c:\\dir\n"
"                   Unpack the specified archive to c:\\dir.\n"
"\n"
"Basic tasks:\n"
"1) Create initial archives for new simulation running on k servers:\n"
"On one of the servers, [ Server n ]: -s n -i\n"
"On the other servers, [ Server m ]: -s m\n"
"2) Create initial archive for a new server:\n"
"On the new server: -s <providerid> (<providerid> must be unique)\n"
"3) Merge archives from [ Server k ], [ Server l ]:\n"
"On [ Server k ]: -u <id> -d c:\\unpacked_k\n"
"On [ Server l ]: -u <id> -d c:\\unpacked_l\n"
"On [ Server k ]: -p <id> -d c:\\unpacked_k -d c:\\unpacked_l\n"
    << "\n";
    }

    
void parse_arguments
    (
    int             argc,
    const char **   argv,
    ServerOptions & options
    )
    {    
    try 
        {
        Massiv::CmdLineParser   cmd_parser( argc, argv, "?hi", "cdpsu" );
        char            cswitch;
        std::string     parameter;    
        while( ( cswitch = cmd_parser.get_next( parameter ) ) )
            {
            switch( cswitch )
                {
                case 'c':
                    {
                    options.config_file_name = parameter;
                    break;
                    }
                case 'd':
                    {
                    options.directories.push_back( parameter );
                    break;
                    }
                case '?':
                case 'h':
                    {
                    print_usage();
                    exit( 1 );
                    break;
                    }
                case 'i':
                    {
                    options.startup_archive_is_initial_archive = true;
                    break;
                    }
                case 'p':
                    {
                    hope( !options.unpack_archives );

                    options.archive_id = parameter;
                    options.pack_archives = true;
                    break;
                    }
                case 's':
                    {
                    UInt32  startup_provider_id;
                    std::istringstream ss( parameter );
                    ss >> startup_provider_id;
                    hope( !ss.fail() );
                    
                    options.create_startup_archive = true;
                    options.startup_archive_provider_id = startup_provider_id;
                    break;
                    }
                case 'u':
                    {
                    hope( !options.pack_archives );

                    options.archive_id = parameter;
                    options.unpack_archives = true;
                    break;
                    }
                default:
                    {
                    std::ostringstream  ss;
                    ss << "Error: Invalid command line options.";
                    throw ExCommandLineError( ss.str() );
                    }
                } // switch
            } // while
        } // try
    catch( ExCommandLineError & e )
        {
        print_usage();
        throw e;
        }
    
    if( options.pack_archives && options.directories.empty() )
        {
        panic( "You must specify at least one directory holding unpacked archive(s)." );
        }
    
    if( options.unpack_archives && options.directories.empty() )
        {
        panic( "You must specify a directory where to unpack to." );
        }

    if( !options.directories.empty() && !options.pack_archives && 
        !options.unpack_archives )
        {
        panic( "What to do with the directories? Either pack or unpack must be specified." );
        }
    }


void create_initial_objects()
    {
    Pointer< NamingService >        naming_service;
    Pointer< AccountObjectImpl >    root_account;
    Remote< AccountObjectImpl >     root_account_remote;
    
    // Create naming service.
    std::cout << "Creating naming service object..." << std::endl;
    naming_service = CreateObject< NamingService >();
    
    // Report naming service id to admin.
        {
        std::string file_name = "naming_service.txt";
        std::ofstream f( file_name.c_str() );
        f << "naming_service_object : string = \"" << naming_service.get_object_id() << "\"" << std::endl;
        std::cout << "**** Naming service id: " << naming_service.get_object_id() << std::endl
            << "Naming service id was written to '" << file_name << "' file." << std::endl;
        }

    // Create root account object for administrator.
    std::cout << "Creating root account object..." << std::endl;
    bool yes = true;
    root_account = CreateObject< AccountObjectImpl >( yes );
    
    // Schedule initial world creation. The world will be created
    // once the simulation is started from the archive.
    std::cout << "Scheduling creation of initial world." << std::endl;
    root_account_remote = root_account.remote();
    root_account_remote->async_create_initial_world();
    }


void create_startup_archive
    (
    bool    startup_archive_is_initial_archive,
    UInt32  startup_archive_provider_id
    )
    {
    std::cout << "Creating startup archive..." << std::endl;

    // Create startup object provider.
    // Must be created before any Object is created.
        {
        UInt32  provider_id = startup_archive_provider_id;

        std::cout << "Creating startup archive provider (id "
            << provider_id << ")..." << std::endl;

        System::create_initial_object_provider( provider_id );
        }

    // If we are creating initial archive, create objects which are
    // necessary for the simulation to run.
    if( startup_archive_is_initial_archive )
        {
        create_initial_objects();
        }

    // Create startup archive.
    System::create_initial_archive();

    // Everything done.
    std::cout << "Startup archive successfully created." << std::endl;
    }

    
void create_node_object_callback()
    {
    Pointer<ServerNode>     server_node;
    
    server_node = CreateObject< ServerNode >();
    
    std::cout << "**** ObjectId of my node object: " << server_node.get_object_id() << std::endl;
    }


void next_tick_callback()
    {
    }
    

void shutdown()
    {
    System::shutdown();
    Demo::Server::Global::shutdown();
    }


void initialize
    (
    bool                creating_startup_archive,
    ServerOptions &     options
    )
    {
    std::cout << "Initializing Massiv core..." << std::endl;

    System::initialize
        (
        System::StartUpInfo
            (
            Massiv::Core::NodeId::TYPE_SERVER,
            options.config_file_name,
            &Massiv::Generated::register_classes,
            &create_node_object_callback,
            &next_tick_callback
            )
        );
    
    Demo::Server::Global::initialize();
    }


} // Anonymous namespace.


void server_main
    (
    int             argc,
    const char **   argv
    )
    {
    ServerOptions   options;

    parse_arguments( argc, argv, options );

    if( options.create_startup_archive )
        {
        initialize( true, options );
        // FIXME: Should we download minimal required data?
        create_startup_archive
            (
            options.startup_archive_is_initial_archive,
            options.startup_archive_provider_id
            );
        System::shutdown();
        }
    else
    if( options.pack_archives )
        {
        hope( !options.directories.empty() );
        
        initialize( true, options );
        System::pack_archives( options.directories, options.archive_id );
        System::shutdown();
        }
    else
    if( options.unpack_archives )
        {
        hope( !options.directories.empty() );

        initialize( true, options );
        System::unpack_archive( options.archive_id, options.directories[ 0 ] );
        System::shutdown();
        }
    else
        {
        initialize( false, options );
        
        try
            {                        
            std::string         config = read_local_node_config();
            std::istringstream  config_stream( config );

            System::download_data();
            System::connect( config_stream );            
            std::cout << "------------------------ RUNNING ------------------------" << std::endl;
            System::run();
            System::disconnect();
            }
        catch( Exception & )
            {
            }
        
        System::shutdown();
        }
    }


} // namespace Server
} // namespace Demo
