#! /usr/bin/perl -w
#
# Modify a file using snippets of source code converted to docbook format.

=head1 NAME

docbookize.pl - insert source snippets to docbook document

=head1 SYNOPSIS

B<docbookize.pl> S< I<file> >

=head1 DESCRIPTION

Modify I<file> using snippets of source code converted to docbook format.
Each set of lines in the I<file> between the lines containing special comments
looking like this

    <!--REPLACE type source tag-->
    ...
    <!--ENDREPLACE-->

will be modified in the following way:

=over 3

=item *

All text between the special comments will be removed. Text before the REPLACE
comment and everything on the line containing the ENDREPLACE comment will be
preserved.

=item *

I<source> will be opened. Its name should be relative to current directory.

=item *

The I<source> will be searched for block beggining with the specified
tag. The way the tags are written in the source depends on I<type>
of the source file:

=over 3

=item shell

Comments begin with a '#' character. Blocks beginnings are marked with
C<##BEGIN tag>, block ends with C<##END tag> and line numbering changes
with C<##LINE number>.

=item c

Comments begin with C</*> and end with C<*/>. Start of a block is marked
with C</*BEGIN tag*/>, end of a block with with C</*END tag*/>, line
number change with C</*LINE tag*/>.

=back

Because the block ending must contain the tag too, you can easily nest
and even interleave code blocks. If you define multiple blocks with the
same tag, they will be merged together.

=item *

All lines inside the block will be converted to docbook format: special
characters will be properly replaced, comments enclosed in
E<lt>lineannotationE<gt> and line numbers prepended.
Special comments will be skipped and not be included in the line count.
Line numbering can be changed using special comment.

=item *

The coverted lines will be inserted between the special docbook comments
in I<file>. The first line will be written directly after the REPLACE
special comment.

=back

=head1 NOTE

Run this script using C<B<perl> -i I<file>> to perform the replacements
inplace. Make sure the special comments in the docbook source are well-balanced
and do not enclose any important text, otherwise the file might get seriously
damaged if inplace replacement is used.

=head1 EXAMPLE

If the docbook XML to modify looks like this:

    <para>
    This is an example:
      <informalexample>
    <programlisting><!--REPLACE c foo.c function-->
    </programlisting><!--ENDREPLACE-->
      </informalexample>
    </para>

and F<foo.c> like this:

    ...
    /*BEGIN function*/
    void fn( void )
    {
        /* Print it. */
        printf( "Hi\n" ) ;
    }
    /*END function*/
    ...

the modified XML will look like this

    <para>
    This is an example:
      <informalexample>
    <programlisting><!--REPLACE c foo.c function--><lineannotation>    (1)</lineannotation>  void fn( void )
    <lineannotation>     (2)</lineannotation>  {
    <lineannotation>     (3)</lineannotation>      <lineannotation>/* Print it. */</lineannotation>
    <lineannotation>     (4)</lineannotation>      printf( "Hi\n" ) ;
    <lineannotation>     (5)</lineannotation>  }
    </programlisting><!--ENDREPLACE-->
      </informalexample>
    </para>

=head1 KNOWN BUGS

The source should not contain tabs as F<docbookize.pl> does not replace tabs
by spaces. Resulting layout of such sources would probably look really weird.

=cut

use warnings ;
use strict ;

use FileHandle ;


###
# Docbookize given line.
#
# type - 'shell' or 'c'.
# line - contents of current line (with trailing new-line)
# index - index of current line
# in_comment - non-zero iff the line begins in a multi-line comment
#
# Returns non-zero iff the line ends in a multi-line comment.
sub docbookize_line( $$$$ ) {
    my $type = shift ;
    my $line = shift ;
    my $index = shift ;
    my $in_comment = shift ;

    chomp( $line ) ;

    # Print line number.

    my $index_len = length( $index ) ;
    print(
        "<lineannotation>" .
        ( ' ' x ( 5 - $index_len ) ) .
        "(${index})</lineannotation>"
    ) ;

    # Strip initial whitespace and remember it.

    my $starting_space ;
    ( $starting_space, $line ) = $line =~ m/^(\s*)(.*)$/ ;

    # Replace special symbols on the line.

    $line =~ s/&/&amp;/g ;
    $line =~ s/</&lt;/g ;
    $line =~ s/>/&gt;/g ;

    # <lineannotation> around comments.

    if ( $type eq 'shell' ) {
        # Shell/perl/whatever # comments.

        $line =~ s{#(.*)$}{<lineannotation>#$1</lineannotation>} ;
        $in_comment = 0 ;
    }
    elsif ( $type eq 'c' ) {
        # C/C++ // and /* */ comments.
        # FIXME: this will not be valid if comments are nested.

        if ( $in_comment ) {
            $line = "<lineannotation>" . $line ;
        }
        $line =~ s{/\*}{<lineannotation>/*}g ;
        $line =~ s{\*/}{*/</lineannotation>}g ;
        $line =~ s{//(.*)$}{<lineannotation>//$1</lineannotation>} ;

        # Determine if we're in comment at EOL.

        if ( $in_comment ) {
            # Left a comment if there is */ with no /* afterwards.

            if ( $line =~ m{\*/(?!.*/\*)} ) {
                $in_comment = 0 ;
            }
        }
        else {
            # Entered a comment if there is /* with no */ afterwards.

            if ( $line =~ m{/\*(?!.*\*/)} ) {
                $in_comment = 1 ;
            }
        }

        if ( $in_comment ) {
            $line = $line . "</lineannotation>" ;
        }
    }

    # Put initial whitespace back + 2 extra spaces.

    $line = '  ' . $starting_space . $line ;

    # Strip trailing whitespace.

    $line =~ s/\s*$// ;

    # Done.

    print $line . "\n" ;

    return $in_comment ;
}


###
# Print docbookized lines from $source of type $type, block $tag.
#
# type - 'shell' or 'c'.
# source - source file name.
# tag - block tag.
#
# Return 0 if $source can't be opened, 1 otherwise.
# Always print at least a new-line if returning 1, never print anything otherwise.
sub docbookize_block( $$$ ) {
    my $type = shift ;
    my $source = shift ;
    my $tag = shift ;

    # Known $type?

    if ( ( $type ne 'shell' ) and ( $type ne 'c' ) ) {
        warn( "Unknown type '${type}' of file '${source}'.\n" ) ;
        return 0 ;
    }

    # Try to open the source.

    my $file = FileHandle->new( $source, "r" ) ;
    unless ( defined( $file ) ) {
        # Failed to open.

        warn( "Can't open '${source}'.\n" ) ;
        return 0 ;
    }

    # Try to find the block.

    my $real_line_index = 1;     # Counting all lines.
    my $virtual_line_index = 1 ; # Counting lines without special comments.
    my $lines_converted = 0 ;
    my $in_block = 0 ;
    my $in_comment = 0 ;
    LINE : while ( my $line = $file->getline() ) {
        # Check for special comments.

        my( $blockpos, $blocktag, $line_change ) ;
        if ( $type eq 'shell' ) {
            ( $blockpos, $blocktag ) = $line =~ m{^\s*##(BEGIN|END)\s+(\S+)\s*$} ;
            ( $line_change ) = $line =~ m{^\s*##LINE\s+(\d+)\s*$} ;
        }
        elsif ( $type eq 'c' ) {
            ( $blockpos, $blocktag ) = $line =~ m{^\s*/\*(BEGIN|END)\s+(\S+)\*/} ;
            ( $line_change ) = $line =~ m{^\s*/\*LINE\s+(\d+)\*/} ;
        }

        # Process the line.

        if ( defined( $blockpos ) ) {
            # Block beginning or ending. Never count it as a line and
            # never print it.

            if ( $blocktag eq $tag ) {
                # Block with our tag. Switch $in_block.

                if ( $blockpos eq 'BEGIN' ) {
                    warn( "${source} (${real_line_index}): Unbalanced /*BEGIN ${tag}*/.\n" )
                        if $in_block ;
                    $in_block = 1 ;
                    $in_comment = 0 ;
                }
                else {
                    warn( "${source} (${real_line_index}): Unbalanced /*END ${tag}*/.\n" )
                        unless $in_block ;
                    $in_block = 0 ;
                }
            }
        }
        elsif ( defined( $line_change ) ) {
            # Change line number special comment.

            $virtual_line_index = $line_change ;
        }
        else {
            # Not a special comment.
            # Print the line if $in_block.

            if ( $in_block ) {
                $in_comment = docbookize_line( $type, $line, $virtual_line_index, $in_comment ) ;
                $lines_converted++ ;
            }

            # Next line without special comment.

            $virtual_line_index++ ;
        }

        # Next line.

        $real_line_index++ ;
    }

    # Warn?

    if ( $in_block ) {
        warn( "In block '${tag}' at '${source}' EOF.\n" ) ;
    }

    unless ( $lines_converted ) {
        # Always print at least a single new-line.

        warn( "Block '${tag}' not found in '${source}'.\n" ) ;
        print "\n" ;
    }

    $file->close() ;
    return 1 ;
}


###
# Main.
while ( <> ) {
    # Is this begginning of REPLACE block?

    my( $type, $source, $tag ) = m/<!--REPLACE\s+(\S+)\s+(\S+)\s+(\S+)-->/ ;
    if ( defined( $type ) and defined( $source ) and defined( $tag ) ) {
        # Ok, strip everything after the comment, inluding the new-line
        # and print it.

        my( $stripped ) = m/(^.*<!--REPLACE\s+\S+\s+\S+\s+\S+-->)/ ;
        print( $stripped ) ;

        # Try to perform the replacement.

        if ( ! docbookize_block( $type, $source, $tag ) ) {
            # Failed, so put back the new-line.

            print( "\n" ) ;
        }

        # Always remove old contents of the block.

        BLOCKLINE : while ( <> ) {
            if ( m/<!--ENDREPLACE-->/) {
                # Done. Print the line and break the loop.

                print ;
                last BLOCKLINE ;
            }
        }
    }
    else {
        # Normal line, print it.

        print ;
    }
}
