/** @file objects.cpp
 * Testing objects.
 *
 * @todo Automatic casts to ObjectPointers?
 * @todo IOStream operators?
 */

/*
 * begin        : Mon Oct 22 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "objects.h"
#include <iostream>
#include <sstream>


using namespace Massiv::Core;


namespace Massiv {
namespace ServerTest {
namespace LocalMigration {


namespace { // anonymous namespace

    void print
        (
        const std::string & object_name,
        const std::string & method_name
        )
        {
        std::cout << System::time().get_value() << ": " << object_name << "::" <<
            method_name << "()" << std::endl;
        }

    template< typename Type1 >
    void print
        (
        const std::string & object_name,
        const std::string & method_name,
        const Type1 &       param1
        )
        {
        std::cout << System::time().get_value() << ": " << object_name << "::" <<
            method_name << "( " << param1 << " )" << std::endl;
        }

    template< typename Type1, typename Type2 >
    void print
        (
        const std::string & object_name,
        const std::string & method_name,
        const Type1 &       param1,
        const Type2 &       param2
        )
        {
        std::cout << System::time().get_value() << ": " << object_name << "::" <<
            method_name << "( " << param1 << ", " << param2 << " )" << std::endl;
        }

} // anonymous namespace


// class Null

Null::Null()
    : migrate_period( 1.0 )
    {
    print( name_string(), "Null" );
    }


Null::~Null()
    {
    print( name_string(), "~Null" );
    }


void Null::finalize()
    {
    print( name_string(), "finalize" );
    }


void Null::initialize()
    {
    }


void Null::delivered_to
    (
    Pointer<Object> destination,
    const STime &   delivery_time
    )
    {
    WeakPointer< Root > root = destination.convert();
    hope( ! root.is_null() );

    print( name_string(), "delivered_to", root->name_string(), delivery_time.get_value() );

    // "Register" to owner.

    hope( ! owner.is_null() );
    hope( owner.is_local() );
    hope( owner->null.is_null() );
    if ( owner->set_null( this ) )
        {
        // Ok, Null should not be destroyed yet.

        hope( owner->null == this );

        // Inform Root.

        Pointer< Null > self( this );
        root->null_delivered( self, delivery_time );
        }
    else
        {
        // Destroy Null...

        hope( owner->null.is_null() );
        }
    }


void Null::delivery_failed
    (
    DeliveryFailure,
    const STime &
    )
    {
    print( name_string(), "delivery_failed" );
    }


std::string Null::name_string() const
    {
    std::stringstream stream;
    stream << "Null";
    if( name.size() )
        {
        stream << "( " << name.read() << " )";
        }
    return stream.str();
    }


void Null::send_to
    (
    Root *          root,
    const STime &   time
    )
    {
    hope( root );

    print( name_string(), "send_to", root->name_string(), time.get_value() );

    // Keep strong pointer to self.

    Pointer< Null > self = this;

    // Inform owner to forget this Null.

    hope( !owner.is_null() );
    owner->set_null( NULL );

    // "self" StackStrong keeps this Null alive...

    // Migrate. NullOwner no longer points to this Null so only this Null will migrate.

    migrate_to( root, time );

    // Forget strong pointer to self. There will be no strong pointer to this Null
    // left but that's ok since migrate_to() has been already called...
    }


// class NullOwner

NullOwner::NullOwner()
      : destroy( false )
    {
    print( name_string(), "NullOwner" );
    }


NullOwner::~NullOwner()
    {
    print( name_string(), "~NullOwner" );
    }


void NullOwner::finalize()
    {
    print( name_string(), "finalize" );
    }


void NullOwner::initialize()
    {
    }


std::string NullOwner::name_string() const
    {
    std::stringstream stream;
    stream << "NullOwner";
    if( name.size() )
        {
        stream << "( " << name.read() << " )";
        }
    return stream.str();
    }


void NullOwner::create_null()
    {
    hope( null.is_null() );

    print( name_string(), "create_null" );

    //  Create new Null.

    null = CreateObject< Null >::create();
    null->owner = this;
    }


void NullOwner::destroy_null()
    {
    if ( null.is_null() )
        {
        // Null is not here. Set #destroy flag to true.

        destroy = true;
        }
    else
        {
        // Null is here. Kill it.

        null.set_null();
        }
    }


bool NullOwner::set_null
    (
    Null * const new_null
    )
    {
    print( name_string(), "set_null", ( new_null ? new_null->name_string() : "NULL" ) );

    if ( new_null )
        {
        hope( null.is_null() );

        if ( destroy )
            {
            // Null should be destroyed, so do not set it.

            destroy = false;
            return false;
            }
        else
            {
            // Set null.

            null = new_null;
            return true;
            }
        }
    else
        {
        hope( ! null.is_null() );

        // Reset null.

        null.set_null();
        return true;
        }
    }


// class Root

Root::Root()
    {
    print( name_string(), "Root" );
    }


Root::~Root()
    {
    print( name_string(), "~Root" );
    }


void Root::finalize()
    {
    print( name_string(), "finalize" );
    }


void Root::initialize()
    {
    }


void Root::null_delivered
    (
    Pointer< Null > &   delivered_null,
    const STime &       delivery_time
    )
    {
    delivered_null->send_to( this, STime( delivery_time + delivered_null->migrate_period.read() ) );
    }


std::string Root::name_string() const
    {
    std::stringstream stream;
    stream << "Root";
    if( name.size() )
        {
        stream << "( " << name.read() << " )";
        }
    return stream.str();
    }


} // namespace Massiv::ServerTest::LocalMigration
} // namespace Massiv::ServerTest
} // namespace Massiv
