/** @file objects.h
 * Testing objects.
 */

/*
 * begin        : Mon Oct 22 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef _OBJECTS_H_
#define _OBJECTS_H_


#include "core.h"
#include <string>


namespace Massiv {
namespace ServerTest {
namespace LocalMigration {


using namespace Massiv::Core;


// Forward declarations.

class Null;
class NullOwner;
class Root;


/**
 * Plain object.
 */
class Null : public Object
    {
    /*
    IDL_BEGIN
    class Massiv::ServerTest::LocalMigration::Null
        parent Massiv::Core::Object
        property name
        property migrate_period
        property owner
    endclass
    IDL_END
    */

    NO_COPY_CONSTRUCTOR( Null );
    NO_ASSIGNMENT_OPERATOR( Null );

public:

    /**
     * Default constructor.
     */
    Null();

    /**
     * Destructor.
     */
    virtual ~Null();

    /**
     * Finalize Null.
     */
    virtual void finalize();

    /**
     * Initializes object.
     */
    void initialize();

protected:

    /**
     * Null delivered to specified object.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   delivery_time
        );

    /**
     * Null delivery failed.
     */
    virtual void delivery_failed
        (
        DeliveryFailure failure,
        const STime &   delivery_time
        );

public:

    /**
     * Print Null's name.
     */
    std::string name_string() const;

    /**
     * Send Null to given Root.
     * Schedule Null's delivery to @a time.
     */
    void send_to
        (
        Root * const    root,
        const STime &   time
        );

public:

    PString                     name;
        /**< Null name. */

    PTime                       migrate_period;
        /**< Delay between two deliveries #deliver_to(). */

    PWeakPointer< NullOwner >   owner;
        /**< Owning NullOwner. Always local. */

    }; // class Null


/**
 * Owner of single Null.
 */
class NullOwner : public Object
    {
    NO_COPY_CONSTRUCTOR( NullOwner );
    NO_ASSIGNMENT_OPERATOR( NullOwner );

public:

    /**
     * Default constructor.
     */
    NullOwner();

    /**
     * Destructor.
     */
    virtual ~NullOwner();

    /**
     * Finalize NullOwner.
     */
    virtual void finalize();

    /**
     * Initializes object.
     */
    void initialize();

public:

    /**
     * Print NullOwner's name.
     */
    std::string name_string() const;

    /**
     * Create owned Null.
     */
    void create_null();

    /**
     * Destoroy Null as soon as it's delivered.
     */
    void destroy_null();

    /**
     * Set Null to own.
     * Called by Null when it migrates (with @a new_null equal to NULL)
     * and when it's delivered to unregister/register Null to NullOwner.
     *
     * @return true on success, false if Null should be destroyed.
     */
    bool set_null
        (
        Null * const new_null
        );

public:

    PString                 name;
        /**< NullOwner name. */

    PPointer< Null >        null;
        /**< Owned Null. */

    PBool                   destroy;
        /**< Destroy Null as soon as it's delivered. */

    }; // class NullOwner


/**
 * GC root object.
 *
 * Sends Nulls to itself.
 */
class Root : public Object
    {
    NO_COPY_CONSTRUCTOR( Root );
    NO_ASSIGNMENT_OPERATOR( Root );

public:

    /**
     * Default constructor.
     */
    Root();

    /**
     * Destructor.
     */
    virtual ~Root();

    /**
     * Finalize Root.
     */
    virtual void finalize();

    /**
     * Initializes object.
     */
    void initialize();

public:

    /**
     * Null object delivered to Root.
     */
    void null_delivered
        (
        Pointer< Null > &   delivered_null,
        const STime &       delivery_time
        );

public:

    /**
     * Print Root's name.
     */
    std::string name_string() const;

public:

    PString                 name;
        /**< Root name. */

    }; // class Root


} // namespace Massiv::ServerTest::LocalMigration
} // namespace Massiv::ServerTest
} // namespace Massiv


#endif // _OBJECTS_H_
