/** @file server_main.cpp
 * Testing server.
 */

/*
 * begin        : Tue Nov 19 2002
 * copyright    : (C) 2002 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "server_common.h"
#include "server_main.h"
#include "class_list.h"
#include "objects.h"
#include "registry/statistics.h"
#include "object/class_manager.h"
#include "object/archive_manager.h"
#include "object/object_manager.h"
#include <iostream>


#if 0
/// Print regular info about current time?
#define PRINT_TIME_INFO
#endif // 0


using namespace Massiv::Core;
using namespace Massiv::Generated;


namespace Massiv {
namespace ServerTest {
namespace LocalMigration {


namespace { // anonymous namespace


    void start_archivation()
        {
        if( ServerCommon::options().enable_archivation )
            {
            hope( !Global::archive_manager().archivation_in_progress() );
            Global::archive_manager().start_archivation();
            }
        }


    void finish_archivation()
        {
        while( Global::archive_manager().archivation_in_progress() )
            {
            System::next_tick();
            }
        }

    /**
     * Run simulation for given number of seconds.
     */
    void run
        (
        const SystemTime period
        )
        {
        static SystemTime zero_time = System::system_time();
        const SystemTime old_time = System::system_time() - zero_time;
#ifdef PRINT_TIME_INFO
        SystemTime time = old_time;
#endif // PRINT_TIME_INFO

        start_archivation();

        for( ; ; )
            {
            // Done?

            const SystemTime current_time = System::system_time() - zero_time;
            if( current_time >= ( old_time + period ) )
                {
                break;
                }

            // Nope. Next tick.

            System::next_tick();

            // Print time info.

#ifdef PRINT_TIME_INFO
            if( current_time >= ( time + 0.2 ) )
                {
                std::cout << System::time().get_value() << ": system time is " <<
                    current_time.get_value() << "." << std::endl;
                time = current_time;
                }
#endif // PRINT_TIME_INFO
            }

        finish_archivation();
        }

} // anonymous namespace


/**
 * Server main.
 * Runs various local migration tests.
 */
void server_main()
    {
    // Initialize core.

    std::cout << "Initializing core..." << std::endl;

    System::register_classes( &register_classes );

    if( ServerCommon::options().create_initial_archive )
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            | System::INIT_STANDALONE_NODE
            );

        UInt32 provider_id = ServerCommon::options().initial_provider_id;

        std::cout << "Creating initial provider (" << provider_id << ")..." << std::endl;
        System::create_initial_object_provider( provider_id );

        std::cout << "Creating initial archive..." << std::endl;
        System::create_initial_archive();

        std::cout << "Initial archive successfully created. Exiting..." << std::endl;
        return;
        }
    else
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            );
        }

    std::cout << "Connecting..." << std::endl;

    System::connect();

    while( System::get_phase() != System::PHASE_CONNECTED )
        {
        System::next_tick();
        }

    // Instantiate factories.

    std::cout << "Instantiating factories..." << std::endl;

    register_classes();

    // Create objects.

    std::cout << "Creating objects..." << std::endl;

    WeakPointer< Root >         root1;
    WeakPointer< Root  >        root2;
    WeakPointer< NullOwner >    null1_owner;
    WeakPointer< NullOwner >    null2_owner;

    root1 = CreateObject< Root >::create();
    root1->name = "root1";

    null1_owner = CreateObject< NullOwner >::create();
    null1_owner->name = "null1_owner";

    null1_owner->create_null();
    null1_owner->null->name = "null1";
    null1_owner->null->migrate_period = 1.0;

    // Run for a while.

    std::cout << "Running for a while..." << std::endl;
    run( 1.0 );

    // Migrate null1.

    std::cout << "Starting migration..." << std::endl;
    null1_owner->null->send_to( root1.get_native_pointer(), System::time() );
    run( 5.0 );

    // Two nulls migrating to root1.

    std::cout << "Migrating two nulls to root1..." << std::endl;

    null2_owner = CreateObject< NullOwner >::create();
    null2_owner->name = "null2_owner";

    null2_owner->create_null();
    null2_owner->null->name = "null2";
    null2_owner->null->migrate_period = 0.4;

    null2_owner->null->send_to( root1.get_native_pointer(), System::time() );
    run( 5.0 );

    // Destroy null1.

    { // sob
    std::cout << "Destroying null1..." << std::endl;
    null1_owner->destroy_null();
    run( 1.5 );
    } // eob

// This test can't. null2_owner->null.is_null() is true, because null2 is migrating.
// Actually it's more or less impossible to migrate object to multiple destinations
// unless it's GCRootObject.

#if 0
    // Migrate null2 to two roots.

    std::cout << "Migrating null2 to root1 and root2..." << std::endl;

    root2 = CreateObject< Root >::create();
    root2->name = "root2";

    null2_owner->null->send_to( root2.get_native_pointer(), System::time() );
    run( 5.0 );
#endif

    // Destroy null2.

    std::cout << "Destroying null2..." << std::endl;
    null2_owner->destroy_null();

    // Give GC enough time to collect unreferenced objects.

    Global::force_gc();
    run( 2.0 );

    // Print statistics.

    std::cout << Global::object_manager().get_statistics() << std::endl;

    // Done.
    System::shutdown();
    }


} // namespace Massiv::ServerTest::LocalMigration
} // namespace Massiv::ServerTest
} // namespace Massiv
