/** @file objects.cpp
 * Testing objects.
 */

/*
 * begin        : Sat Dec 8 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "objects.h"
#include <iostream>


namespace Massiv {
namespace ServerTest {
namespace Philosophers {

#if 1
#define PRINT( text ) std::cout << text
#endif

// Implementation of Philosopher
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void Philosopher::initialize
    (
    Float64 _think_period,
    Float64 _eat_period,
    Float64 _take_fork_period
    )
    {
    think_period = _think_period;
    eat_period = _eat_period;
    take_fork_period = _take_fork_period;
    }


void Philosopher::try_to_sit
    (
    WeakPointer<Table>  _table,
    const STime &       _time
    )
    {
    hope( chair.is_null() );
    hope( table.is_null() );

    // Send sit request to the table.

    PRINT( get_object_id() << " is trying to sit." << std::endl );

    Pointer<WantToSit> r = CreateObject<WantToSit>::create( this );
    r->migrate_to( _table, _time );
    }


void Philosopher::chair_allocated
    (
    WeakPointer<Chair>  _chair,
    WeakPointer<Table>  _table
    )
    {
    PRINT( get_object_id() << " was notified about allocated chair." << std::endl );

    chair = _chair.convert();
    table = _table.convert();

    think();
    }


void Philosopher::think()
    {
    PRINT( get_object_id() << " starts thinking." << std::endl );

    // Schedule stop thinking request.

    Pointer<StopThinking> r = CreateObject<StopThinking>::create();
    r->migrate_to( this, STime( Massiv::System::time() + think_period.read() ) );
    }


void Philosopher::eat()
    {
    PRINT( get_object_id() << " is hungry and waiting for forks." << std::endl );

    // Take forks.

    Pointer<TakeForks> r = CreateObject<TakeForks>::create( this, chair );
    r->migrate_to( table, STime( Massiv::System::time() + take_fork_period.read() ) );
    }


void Philosopher::forks_taken()
    {
    PRINT( get_object_id() << " acquired forks and is eating now." << std::endl );

    // Schedule stop eating request.

    Pointer<StopEating> r = CreateObject<StopEating>::create();
    r->migrate_to( this, STime( Massiv::System::time() + eat_period.read() ) );
    }


void Philosopher::can_not_take_forks()
    {
    PRINT( get_object_id() << " failed to acquire forks and is starving." << std::endl );

    eat();
    }


void Philosopher::stop_eating()
    {
    PRINT( get_object_id() << " is full and feels dizzy." << std::endl );

    // Release forks.

    Pointer<ReleaseForks> r = CreateObject<ReleaseForks>::create( chair );
    r->migrate_to( table, Massiv::System::time() );

    think();
    }


// Implementation of Chair
// ~~~~~~~~~~~~~~~~~~~~~~~

void Chair::initialize()
    {
    }


// Implementation of Table
// ~~~~~~~~~~~~~~~~~~~~~~~

void Table::initialize()
    {
    dummy_chair = CreateObject<Chair>::create();
    dummy_chair->right_chair = dummy_chair;
    dummy_chair->left_chair = dummy_chair;
    dummy_chair->table = this;

    num_chairs = 0;
    }


Pointer<Chair> Table::create_chair()
    {
    Pointer<Chair> chair;

    // Create chair.

    PRINT( "Creating a new chair." << std::endl );

    chair = CreateObject<Chair>::create();

    // Insert chair to double-linked cyclic list.

    chair->right_chair = dummy_chair->right_chair;
    chair->left_chair = dummy_chair;
    dummy_chair->right_chair->left_chair = chair;
    dummy_chair->right_chair = chair;
    chair->table = this;

    num_chairs++;

    // Done.

    return chair;
    }


void Table::want_to_sit
    (
    WeakPointer<Philosopher>    philosopher
    )
    {
    Pointer<Chair> chair;

    // Sit philosopher.

    PRINT( philosopher.get_object_id() << " wants to sit down." << std::endl );

    chair = create_chair();
    chair->philosopher = philosopher.convert();

    // Notify philosopher that he is sitting at the table.

    Pointer<ChairAllocated> notify = CreateObject<ChairAllocated>::create( chair, this );
    notify->migrate_to( philosopher, Massiv::System::time() );
    }


void Table::take_forks
    (
    WeakPointer<Chair>          chair,
    WeakPointer<Philosopher>    philosopher
    )
    {
    PRINT( philosopher.get_object_id() << " wants to take forks." << std::endl );
    PRINT( "left fork at: " << chair.get_object_id() << std::endl );
    PRINT( "right fork at: " << chair->right_chair.get_object_id() << std::endl );

    if( chair->fork_acquired == true || chair->right_chair->fork_acquired == true )
        {
        // Forks are in use.

        PRINT( "Forks are in use." << std::endl );

        Pointer<CanNotTakeForks> r = CreateObject<CanNotTakeForks>::create();
        r->migrate_to( philosopher, Massiv::System::time() );
        }
    else
        {
        // Take forks.

        PRINT( "Acquiring forks." << std::endl );

        chair->fork_acquired = true;
        chair->right_chair->fork_acquired = true;

        Pointer<ForksTaken> r = CreateObject<ForksTaken>::create();
        r->migrate_to( philosopher, Massiv::System::time() );
        }
    }


void Table::release_forks
    (
    WeakPointer<Chair> chair
    )
    {
    PRINT( "Releasing forks:" << std::endl );
    PRINT( "left fork at: " << chair.get_object_id() << std::endl );
    PRINT( "right fork at: " << chair->right_chair.get_object_id() << std::endl );

    chair->fork_acquired = false;
    chair->right_chair->fork_acquired = false;
    }


// Implementation of WantToSit
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

void WantToSit::initialize
    (
    const WeakPointer<Philosopher> &    _philosopher
    )
    {
    philosopher = _philosopher;
    }


void WantToSit::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Table> table = destination.convert();
    table->want_to_sit( philosopher );
    }


// Implementation of ChairAllocated
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ChairAllocated::initialize
    (
    const WeakPointer<Chair> &  _chair,
    const WeakPointer<Table> &  _table
    )
    {
    chair = _chair;
    table = _table;
    }


void ChairAllocated::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Philosopher> philosopher = destination.convert();
    philosopher->chair_allocated( chair, table );
    }


// Implementation of StopThinking
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void StopThinking::initialize()
    {
    }


void StopThinking::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Philosopher> philosopher = destination.convert();
    philosopher->eat();
    }


// Implementation of StopEating
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void StopEating::initialize()
    {
    }


void StopEating::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Philosopher> philosopher = destination.convert();
    philosopher->stop_eating();
    }


// Implementation of TakeForks
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

void TakeForks::initialize
    (
    const WeakPointer<Philosopher> &    _philosopher,
    const WeakPointer<Chair> &          _chair
    )
    {
    philosopher = _philosopher;
    chair = _chair;
    }


void TakeForks::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Table> table = destination.convert();
    table->take_forks( chair, philosopher );
    }


// Implementation of ForksTaken
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ForksTaken::initialize()
    {
    }


void ForksTaken::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Philosopher> philosopher = destination.convert();
    philosopher->forks_taken();
    }


// Implementation of CanNotTakeForks
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void CanNotTakeForks::initialize()
    {
    }


void CanNotTakeForks::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Philosopher> philosopher = destination.convert();
    philosopher->can_not_take_forks();
    }


// Implementation of TakeForks
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~

void ReleaseForks::initialize
    (
    const WeakPointer<Chair> & _chair
    )
    {
    chair = _chair;
    }


void ReleaseForks::delivered_to
    (
    Pointer<Object> destination,
    const STime &   time
    )
    {
    Pointer<Table> table = destination.convert();
    table->release_forks( chair );
    }

} // namespace Massiv::ServerTest::Philosophers
} // namespace Massiv::ServerTest
} // namespace Massiv
