/** @file objects.h
 * Testing objects.
 * Eating philosophers classic.
 */

/*
 * begin        : Sat Dec 8 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef _OBJECTS_H_
#define _OBJECTS_H_


#include "core.h"


using namespace Massiv::Core;


namespace Massiv {
namespace ServerTest {
namespace Philosophers {

//
// Philosopher classes.
//

class Philosopher;
class Table;
class Chair;

/**
 * Philosopher.
 */
class Philosopher : public Object
    {
    friend class ChairAllocated;

public:

    /**
     * Sets up philosopher.
     */
    void initialize
        (
        Float64 _think_period,
        Float64 _eat_period,
        Float64 _take_fork_period
        );

    /**
     * Tries philosopher to sit at @c table.
     */
    void try_to_sit
        (
        WeakPointer<Table>  _table,
        const STime &       _time
        );

    /**
     * Notifies philosopher that he is sitting on chair.
     */
    void chair_allocated
        (
        WeakPointer<Chair>  _chair,
        WeakPointer<Table>  _table
        );

    /**
     * Enters think phase.
     */
    void think();

    /**
     * Enters eat phase.
     */
    void eat();

    /**
     * Forks were taken.
     */
    void forks_taken();

    /**
     * Forks could not be taken.
     */
    void can_not_take_forks();

    /**
     * Philosopher stops eating.
     */
    void stop_eating();

public:

    PWeakPointer<Chair>     chair;
        /**< Chair the philosopher is sitting on. */
    PWeakPointer<Table>     table;
        /**< Table the philosopher is sitting at. */
    PTime                   think_period;
        /**< How long philosopher thinks. */
    PTime                   eat_period;
        /**< How long philosopher eats. */
    PTime                   take_fork_period;
        /**< How long philosopher waits if he did not succeed in
             acquiring forks. */

    }; // Philosopher

/**
 * Chair.
 */
class Chair : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize();

public:

    PWeakPointer<Philosopher>   philosopher;
        /**< Who is sitting on the chair. */
    PPointer<Chair>             right_chair;
        /**< Chair on the right. */
    PPointer<Chair>             left_chair;
        /**< Chair on the left. */
    PPointer<Table>             table;
        /**< Table the chair stands at. */
    PBool                       fork_acquired;
        /**< Just one fork is assigned to each chair.
             If a philosopher wants to eat he must take both
             the forks (this fork and fork on the right).
             Is this fork acquired by a philosopher? */

    }; // Chair

/**
 * Round table where Philosophers think and eat.
 */
class Table : public Object
    {
    friend class WantToSit;
    friend class TakeForks;
    friend class ReleaseForks;

public:

    /**
     * Initializes table.
     */
    void initialize();

private:

    /**
     * Creates empty chair.
     */
    Pointer<Chair> create_chair();

    /**
     * Sits philosopher at the table.
     */
    void want_to_sit
        (
        WeakPointer<Philosopher>    philosopher
        );

    /**
     * Philosopher asks for forks.
     */
    void take_forks
        (
        WeakPointer<Chair>          chair,
        WeakPointer<Philosopher>    philosopher
        );

    /**
     * Philosopher releases for forks.
     */
    void release_forks
        (
        WeakPointer<Chair>  chair
        );

public:

    PPointer<Chair> dummy_chair;
        /**< Dummy chair. */
    PInt<Int32, 32> num_chairs;
        /**< Number of chairs at the table. */

    }; // Table

//
// Messages.
//

/**
 * Philosopher wants to sit at the table.
 * Sent to table by philosopher.
 */
class WantToSit : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize
        (
        const WeakPointer<Philosopher> &    _philosopher
        );

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   deliver_time
        );

public:

    PWeakPointer<Philosopher>   philosopher;
        /**< Philosopher. */
    }; // WantToSit

/**
 * A chair was allocated to the philosopher.
 * Sent to philosopher by table.
 */
class ChairAllocated : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize
        (
        const WeakPointer<Chair> &  _chair,
        const WeakPointer<Table> &  _table
        );

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

public:

    PWeakPointer<Chair> chair;
        /**< Chair. */
    PWeakPointer<Table> table;
        /**< Table. */

    }; // ChairAllocated

/**
 * Sent to philosopher to stop thinking.
 */
class StopThinking : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize();

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

    }; // StopThinking

/**
 * Sent to philosopher to stop eaiting.
 */
class StopEating : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize();

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

    }; // StopEating

/**
 * Philosopher is trying to take forks.
 * Sent to table by philosopher.
 */
class TakeForks : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize
        (
        const WeakPointer<Philosopher> &    _philosopher,
        const WeakPointer<Chair> &          _chair
        );

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

public:

    PWeakPointer<Philosopher>   philosopher;
        /**< Philosopher. */
    PWeakPointer<Chair>         chair;
        /**< Chair. */

    }; // TakeForks

/**
 * Philosopher acquires forks.
 * Sent by table to philosopher.
 */
class ForksTaken : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize();

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

    }; // ForksTaken

/**
 * Philosopher failed to acquire forks.
 * Sent by table to philosopher.
 */
class CanNotTakeForks : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize();

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

    }; // CanNotTakeForks

/**
 * Philosopher is relasing forks.
 * Sent to table by philosopher.
 */
class ReleaseForks : public Object
    {
public:

    /**
     * Initializes object.
     */
    void initialize
        (
        const WeakPointer<Chair> & _chair
        );

    /**
     * Dispatch.
     */
    virtual void delivered_to
        (
        Pointer<Object> destination,
        const STime &   time
        );

public:

    PWeakPointer<Chair>         chair;
        /**< Chair. */

    }; // ReleaseForks


} // namespace Massiv::ServerTest::Philosophers
} // namespace Massiv::ServerTest
} // namespace Massiv


#endif // _OBJECTS_H_
