/** @file server_main.cpp
 * Testing server.
 */

/*
 * begin        : Sat Dec 8 2002
 * copyright    : (C) 2002 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "server_common.h"
#include "server_main.h"
#include "class_list.h"
#include "objects.h"
#include "object/class_manager.h"
#include "object/archive_manager.h"
#include "object/object_manager.h"
#include <iostream>


using namespace Massiv::Core;
using namespace Massiv::Generated;


namespace Massiv {
namespace ServerTest {
namespace Philosophers {

namespace { // anonymous namespace


void start_archivation()
    {
    if( ServerCommon::options().enable_archivation )
        {
        hope( !Global::archive_manager().archivation_in_progress() );
        Global::archive_manager().start_archivation();
        }
    }


void finish_archivation()
    {
    while( Global::archive_manager().archivation_in_progress() )
        {
        System::next_tick();
        }
    }


void run
    (
    float period
    )
    {
    const SystemTime old_time = System::system_time();

    while( System::system_time() < ( old_time + period ) )
        {
        SystemTime          t = System::system_time();
        std::ostringstream  ss;
        static int          counter;
        static bool         archivation_started = false;

        if( !archivation_started && System::system_time() >= ( old_time + period / 2 ) )
            {
            start_archivation();
            archivation_started = true;
            }

        while( System::system_time() < ( old_time + period ) &&
                System::system_time() < ( t + 5.0 ) )
            {
            System::next_tick();
            }

        ss << "memory_" << counter++ << ".dot";
        Global::garbage_collector().save_memory_snapshot( ss.str() );
        }

    finish_archivation();
    }

} // anonymous namespace

enum Parameters
    {
    NUM_PHILOSOPHERS    =   16
    };


static void load_simulation()
    {
    std::string archive_file_name = ServerCommon::options().archive_file_name;

    std::cout << "Loading simulation from archive \"" << archive_file_name << "\"." << std::endl;

    System::start_simulation_from_archive( archive_file_name );

    // Run for a while.

    run( 60.0 );
    }


static void create_simulation()
    {
    // Create GC root objects.

    { // sob
    Pointer<Table>          table;
    int                     i;

    std::cout << "Creating objects..." << std::endl;

    table = CreateObject<Table>::create();

    for( i = 0; i < NUM_PHILOSOPHERS; i++ )
        {
        Pointer<Philosopher> philosopher;

        philosopher = CreateObject<Philosopher>::create
            ( 1.0 + i * 3.0 / NUM_PHILOSOPHERS, 5.0, 1.0 );
        philosopher->try_to_sit( table, STime( System::time() + i * 0.003 ) );
        }
    } // eob

    // Run for a while.

    run( 60.0 );
    }


void server_main()
    {
    // Initialize core.

    std::cout << "Initializing core..." << std::endl;

    System::register_classes( &register_classes );

    if( ServerCommon::options().create_initial_archive )
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            | System::INIT_STANDALONE_NODE
            );

        UInt32 provider_id = ServerCommon::options().initial_provider_id;

        std::cout << "Creating initial provider (" << provider_id << ")..." << std::endl;
        System::create_initial_object_provider( provider_id );

        std::cout << "Creating initial archive..." << std::endl;
        System::create_initial_archive();

        std::cout << "Initial archive successfully created. Exiting..." << std::endl;
        return;
        }
    else
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            );
        }

    std::cout << "Connecting..." << std::endl;

    System::connect();

    while( System::get_phase() != System::PHASE_CONNECTED )
        {
        System::next_tick();
        }

    if( ServerCommon::options().load_simulation_from_archive )
        {
        load_simulation();
        }
    else
        {
        create_simulation();
        }

    // Done.
    System::shutdown();
    }


} // namespace Massiv::ServerTest::Philosophers
} // namespace Massiv::ServerTest
} // namespace Massiv
