/** @file objects.cpp
 * Testing objects.
 * Eating philosophers classic, using async RPC.
 */

/*
 * begin        : Sat Dec 12 2002
 * copyright    : (C) 2002 by Marek Vondrak, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "objects.h"
#include "objects_rpc.h"
#include <iostream>


namespace Massiv {
namespace ServerTest {
namespace PhilosophersRPC {


// class Philosopher

void Philosopher::initialize
    (
    Float64 the_think_period,
    Float64 the_eat_period,
    Float64 the_take_fork_period
    )
    {
    think_period = the_think_period;
    eat_period = the_eat_period;
    take_fork_period = the_take_fork_period;
    }


void Philosopher::try_to_sit
    (
    WeakPointer<Table>  the_table,
    const STime &       the_time
    )
    {
    hope( chair.is_null() );
    hope( table.is_null() );

    // Send sit request to the table.

    std::cout << get_object_id() << " is trying to sit." << std::endl;

    Remote< Table > remote_table( the_table );
    remote_table->want_to_sit( this );
    }


void Philosopher::chair_allocated
    (
    WeakPointer< Chair >    the_chair,
    WeakPointer< Table >    the_table
    )
    {
    std::cout << get_object_id() << " was notified about allocated chair." << std::endl;

    chair = the_chair.convert();
    table = the_table.convert();

    think();
    }


void Philosopher::think()
    {
    std::cout << get_object_id() << " starts thinking." << std::endl;

    // Schedule stop thinking request.

    Remote< Philosopher > remote_self( this );
    remote_self->delayed_eat( think_period );
    }


void Philosopher::eat()
    {
    std::cout << get_object_id() << " is hungry and waiting for forks." << std::endl;

    // Take forks.

    Remote< Table > remote_table( table );
    remote_table->delayed_take_forks( take_fork_period, chair, this );
    }


void Philosopher::forks_taken()
    {
    std::cout << get_object_id() << " acquired forks and is eating now." << std::endl;

    // Schedule stop eating request.

    Remote< Philosopher > remote_self( this );
    remote_self->delayed_stop_eating( eat_period );
    }


void Philosopher::can_not_take_forks()
    {
    std::cout << get_object_id() << " failed to acquire forks and is starving." << std::endl;

    eat();
    }


void Philosopher::stop_eating()
    {
    std::cout << get_object_id() << " is full and feels dizzy." << std::endl;

    // Release forks.

    Remote< Table > remote_table( table );
    remote_table->release_forks( chair );

    think();
    }


// class Chair

void Chair::initialize()
    {
    }


// class Table

void Table::initialize()
    {
    dummy_chair = CreateObject< Chair >::create();
    dummy_chair->right_chair = dummy_chair;
    dummy_chair->left_chair = dummy_chair;
    dummy_chair->table = this;

    num_chairs = 0;
    }


Pointer<Chair> Table::create_chair()
    {
    Pointer< Chair > chair;

    // Create chair.

    std::cout << "Creating a new chair." << std::endl;

    chair = CreateObject< Chair >::create();

    // Insert chair to double-linked cyclic list.

    chair->right_chair = dummy_chair->right_chair;
    chair->left_chair = dummy_chair;
    dummy_chair->right_chair->left_chair = chair;
    dummy_chair->right_chair = chair;
    chair->table = this;

    num_chairs++;

    // Done.

    return chair;
    }


void Table::want_to_sit
    (
    WeakPointer<Philosopher>    philosopher
    )
    {
    Pointer< Chair > chair;

    // Sit philosopher.

    std::cout << philosopher.get_object_id() << " wants to sit down." << std::endl;

    chair = create_chair();
    chair->philosopher = philosopher.convert();

    // Notify philosopher that he is sitting at the table.

    Remote< Philosopher > remote_philosopher( chair->philosopher );
    remote_philosopher->chair_allocated( chair, this );
    }


void Table::take_forks
    (
    WeakPointer< Chair >        chair,
    WeakPointer< Philosopher >  philosopher
    )
    {
    std::cout << philosopher.get_object_id() << " wants to take forks." << std::endl;
    std::cout << "left fork at: " << chair.get_object_id() << std::endl;
    std::cout << "right fork at: " << chair->right_chair.get_object_id() << std::endl;

    Remote< Philosopher > remote_philosopher = philosopher.convert();
    if( chair->fork_acquired == true || chair->right_chair->fork_acquired == true )
        {
        // Forks are in use.

        std::cout << "Forks are in use." << std::endl;
        remote_philosopher->can_not_take_forks();
        }
    else
        {
        // Take forks.

        std::cout << "Acquiring forks." << std::endl;

        chair->fork_acquired = true;
        chair->right_chair->fork_acquired = true;

        remote_philosopher->forks_taken();
        }
    }


void Table::release_forks
    (
    WeakPointer< Chair > chair
    )
    {
    std::cout << "Releasing forks:" << std::endl;
    std::cout << "left fork at: " << chair.get_object_id() << std::endl;
    std::cout << "right fork at: " << chair->right_chair.get_object_id() << std::endl;

    chair->fork_acquired = false;
    chair->right_chair->fork_acquired = false;
    }


} // namespace Massiv::ServerTest::PhilosophersRPC
} // namespace Massiv::ServerTest
} // namespace Massiv
