/** @file server_main.cpp
 * Testing pointer functionality & performance.
 */

/*
 * begin        : Fri Jan 17 2003
 * copyright    : (C) 2003 by Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/


#include "core.h"
#include "server_main.h"
#include "server_common.h"
#include "objects.h"
#include "class_list.h"
#include <iostream>

using namespace Massiv::Core;
using namespace Massiv::Generated;

namespace Massiv {
namespace ServerTest {
namespace PointerTest {


enum Parameters
    {
    LIST_SIZE = 10000,
    ITERATION_COUNT = 20,
    CONVERSION_COUNT = 300000,
    WRITE_COUNT = 300000
    };


void test_list()
    {
    Pointer<List>   list = CreateObject<List>::create();
    int             i;

    // Allocation.

    const SystemTime allocation_start = System::system_time();

    for( i = 0; i < LIST_SIZE; i++ )
        {
        Pointer<ObjectA> object = CreateObject<ObjectA>::create();

        list->push_back( object );
        }

    // Iteration.

    const SystemTime iteration_start = System::system_time();

    for( i = 0; i < ITERATION_COUNT; i++ )
        {
        Pointer<Enumerator> e = list->elements();

        while( e->has_next() )
            {
            e->next();
            }
        }

    // Evaluation.

    const SystemTime test_end = System::system_time();

    const SystemTime::TimeType allocation_duration =
        ( iteration_start - allocation_start ).get_value();
    const SystemTime::TimeType iteration_duration =
        ( test_end - iteration_start ).get_value();
    const SystemTime::TimeType total_duration =
        ( test_end - allocation_start ).get_value();

    std::cout << "Allocation: " << allocation_duration << "s, " <<
        ( 1000.0 * allocation_duration / LIST_SIZE ) <<
        "ms per element." << std::endl;
    std::cout << "Iteration: " << iteration_duration << "s, " <<
        ( 1000.0 * iteration_duration / ( ITERATION_COUNT * LIST_SIZE ) ) <<
        "ms per element." << std::endl;
    std::cout << "Total: " << total_duration << "s." << std::endl;
    }


void test_ppointer_to_pointer()
    {
    Pointer<List>           list = CreateObject<List>::create();
    PPointer<List::Entry> & property_pointer = list->first;
    Pointer<List::Entry>    stack_pointer;
    int                     i;

    // PPointer to Pointer.

    const SystemTime test_start = System::system_time();

    for( i = 0; i < CONVERSION_COUNT; i++ )
        {
        stack_pointer = property_pointer;
        }

    const SystemTime test_end = System::system_time();

    // Evaluation.

    const SystemTime::TimeType test_duration =
        ( test_end - test_start ).get_value();

    std::cout << "Property-pointer to stack-pointer: " << test_duration << "s, " <<
        ( 1000.0 * test_duration / CONVERSION_COUNT ) <<
        "ms per element." << std::endl;
    }

void test_pointer_cast()
    {
    Pointer<ObjectC>    ptr_c = CreateObject<ObjectC>::create();
    Pointer<ObjectB>    ptr_b = ptr_c;
    Pointer<ObjectB>    ptr_a = ptr_b;
    Pointer<ObjectA>    upcasted;
    Pointer<ObjectB>    downcasted;
    Pointer<ObjectC>    most_derived;
    int                 i;

    // Upcast (ObjectB -> ObjectA).

    const SystemTime upcast_start = System::system_time();

    for( i = 0; i < CONVERSION_COUNT; i++ )
        {
        upcasted = ptr_b;
        }

    // Downcast (ObjectA -> ObjectB).

    const SystemTime upcast_end = System::system_time();

    for( i = 0; i < CONVERSION_COUNT; i++ )
        {
        downcasted = ptr_a.convert();
        }

    // Evaluation.

    const SystemTime downcast_end = System::system_time();

    const SystemTime::TimeType upcast_duration =
        ( upcast_end - upcast_start ).get_value();
    const SystemTime::TimeType downcast_duration =
        ( downcast_end - upcast_end ).get_value();

    std::cout << "Pointer upcast: " << upcast_duration << "s, " <<
        ( 1000.0 * upcast_duration / CONVERSION_COUNT ) <<
        "ms per element." << std::endl;
    std::cout << "Pointer downcast: " << downcast_duration << "s, " <<
        ( 1000.0 * downcast_duration / CONVERSION_COUNT ) <<
        "ms per element." << std::endl;
    }


void test_ppointer_write()
    {
    Pointer<List>           list = CreateObject<List>::create();
    PPointer<List::Entry> & first = list->first;
    PInt32 &                num_items = list->num_items;
    Pointer<List::Entry>    old_first = first;
    int                     old_num_items = num_items;
    Pointer<List::Entry>    null;
    int                     i;

    // PPointer write.

    const SystemTime ptr_write_start = System::system_time();

    for( i = 0; i < WRITE_COUNT; i++ )
        {
        first = null;
        first = old_first;
        }

    // PInteger write.

    const SystemTime ptr_write_end = System::system_time();

    for( i = 0; i < WRITE_COUNT; i++ )
        {
        num_items = 0;
        num_items = old_num_items;
        }

    // Evaluation.

    const SystemTime int_write_end = System::system_time();

    const SystemTime::TimeType ptr_write_duration =
        ( ptr_write_end - ptr_write_start ).get_value();
    const SystemTime::TimeType int_write_duration =
        ( int_write_end - ptr_write_end ).get_value();

    std::cout << "PPointer write: " << ptr_write_duration << "s, " <<
        ( 1000.0 * ptr_write_duration / WRITE_COUNT ) <<
        "ms per element." << std::endl;
    std::cout << "PInteger write: " << int_write_duration << "s, " <<
        ( 1000.0 * int_write_duration / WRITE_COUNT ) <<
        "ms per element." << std::endl;
    }


void server_main()
    {
    // Init.

    std::cout << "Global initialization..." << std::endl;

    System::register_classes( &register_classes );

    if( ServerCommon::options().create_initial_archive )
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            | System::INIT_STANDALONE_NODE
            );

        UInt32 provider_id = ServerCommon::options().initial_provider_id;

        std::cout << "Creating initial provider (" << provider_id << ")..." << std::endl;
        System::create_initial_object_provider( provider_id );

        std::cout << "Creating initial archive..." << std::endl;
        System::create_initial_archive();

        std::cout << "Initial archive successfully created. Exiting..." << std::endl;
        return;
        }
    else
        {
        System::initialize
            (
            0
            | System::INIT_COUT_LOGGER_DESTINATION
            );
        }

    std::cout << "Connecting..." << std::endl;

    System::connect();

    while( System::get_phase() != System::PHASE_CONNECTED )
        {
        System::next_tick();
        }

    // Tests.

    test_list();
    test_ppointer_to_pointer();
    test_pointer_cast();
    test_ppointer_write();
    }


} // namespace Massiv::ServerTest::PointerTest
} // namespace Massiv::ServerTest
} // namespace Massiv
