/** @file gencfg.cpp
 * Command line utility for generationg configs for massiv nodes.
 */

/*
 * begin        : Sun Sep 21 2003
 * copyright    : (C) 2003 by Petr Tovarys
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "cmdline/cmdline_parser.h"
#include "registry/registry.h"
#include "registry/conversion.h"
#include "filesys/volume_manager.h"
#include "utility/quote_string.h"
#include "net/massiv_rsa.h"
#include "node/nodeid.h"
#include "utility/debug.h"
#include "system/system.h"
#include "data/account_manager.h"
#include <memory>
#include <vector>
#include <iostream>
#include <ctime>
#include <cctype>


using Massiv::Core::NodeId;
using Massiv::Core::RSACipher;
using Massiv::Core::Registry;
using Massiv::Core::ConfigVariable;
using Massiv::Core::VolumeManager;
using Massiv::Core::Int32;
using Massiv::Core::UInt32;
using Massiv::CmdLineParser;
using Massiv::Core::flags_on;


//------------------------------------------------------------------------------


enum WriteFlags
    {
    WRITE_NODE_ID =         0x01,
    WRITE_PUBLIC_KEY =      0x02,
    WRITE_PRIVATE_KEY =     0x04,
    WRITE_HOST_ADDRESS =    0x08,
    WRITE_ACCOUNT_OBJECT =  0x10,
    WRITE_NAME =            0x20,
    
    WRITE_ALL_EXCEPT_RSA_PRIVATE_KEY =
        WRITE_NODE_ID |
        WRITE_PUBLIC_KEY |
        WRITE_HOST_ADDRESS |
        WRITE_ACCOUNT_OBJECT |
        WRITE_NAME,
        
    WRITE_ALL =
        WRITE_NODE_ID |
        WRITE_PUBLIC_KEY |
        WRITE_PRIVATE_KEY |
        WRITE_HOST_ADDRESS |
        WRITE_ACCOUNT_OBJECT |
        WRITE_NAME
        
    }; // WriteFlags


/**
 * Misc generated data.
 */
class Node
    {
public:

    /**
     * Get directory of the node. Returned string is terminated by '/'
     * character so when appending file name do not add useless '/'.
     */
    std::string get_directory() const
        {
        std::ostringstream oss;
        oss << NodeId::type_to_string( node_id.get_type() )
            << node_id.get_relative_id()
            << '/';
        std::string temp = oss.str();
        temp[ 0 ] = tolower( temp[ 0 ] );
        
        // FIXME. Little hack.
        if( temp == "service1/" )
            {
            temp = "master_data_service/";
            }
        return temp;
        }
        
public:
    NodeId          node_id;
    std::string     rsa_public_key;
    std::string     rsa_private_key;
    std::string     address;
    int             port;
    std::string     account_object;
    std::string     name;
    }; // Node


//------------------------------------------------------------------------------


std::vector< Node >     servers;
std::vector< Node >     clients;
Node                    master_data_service;
Node                    anonymous;
int                     num_servers = 0;
int                     num_clients = 1;

bool                    nsis = false; // Create config files for nsis?
std::string             ip_address_port; // IP addresses and ports string.
std::string             admin_password; // Administrator password.
//------------------------------------------------------------------------------
    

void error_write_open
    (
    const std::string & file_name
    )
    {
    std::ostringstream oss;
    oss << "Failed to open '" << file_name << "' file for writing.";
    panic( oss.str().c_str() );
    }

    
void generate_rsa_pair( std::string & rsa_public_key, std::string & rsa_private_key )
    {
    unsigned int bitlen = RSACipher::DEFAULT_PRIVATE_KEY_LENGTH;
    unsigned int seedlen = RSACipher::DEFAULT_PRIVATE_KEY_SEED_LENGTH;
    RSACipher rsa( bitlen, RSACipher::generate_seed( seedlen ) );
    
    rsa_public_key = rsa.get_public_key();
    rsa_private_key = rsa.get_private_key();
    }


void create_int
    (
    Registry &                  registry,
    Registry::RegistryNode *    reg_node,
    const std::string &         var_name,
    int                         value
    )
    {
    registry.create_variable
            (
            var_name,
            reg_node,
            ConfigVariable::CVAR_TYPE_INTEGER,
            Massiv::Core::ManagedConversion< std::string, Int32 >( value )
            );
    }


void create_string
    (
    Registry &                  registry,
    Registry::RegistryNode *    reg_node,
    const std::string &         var_name,
    const std::string &         value
    )
    {
    registry.create_variable
            (
            var_name,
            reg_node,
            ConfigVariable::CVAR_TYPE_STRING,
            value
            );
    }


void generate_server
    (
    int i
    )
    {
    hope( i >= 0 && i < static_cast< int >( servers.size() ) );
    UInt32 server_index = static_cast< UInt32 >( i + 1 );
    Node server;
    
    server.node_id = NodeId::create_from_type_and_relative_id( NodeId::TYPE_SERVER, server_index );
    generate_rsa_pair( server.rsa_public_key, server.rsa_private_key );
    server.address = "127.0.0.1";
    server.port = 10000 + server_index;
    server.account_object = "[ null ]";
    
        {
        std::ostringstream oss;
        oss << "Server" << server.node_id.get_relative_id();
        server.name = oss.str();
        }
    
    servers[ i ] = server;
    }


void generate_client
    (
    int i
    )
    {
    hope( i >= 0 && i < static_cast< int >( clients.size() ) );
    UInt32 client_index = static_cast< UInt32 >( i + 1 );
    Node client;
    
    client.node_id = NodeId::create_from_type_and_relative_id( NodeId::TYPE_CLIENT, client_index );
    generate_rsa_pair( client.rsa_public_key, client.rsa_private_key );
    client.address = "IGNORED BY CLIENT NODES";
    client.port = 0;

    if( !i )
        {
        client.account_object = "[ 1 2 0 ]";
        }
    else
        {
        client.account_object = "TO BE EDITED BY CLIENT";
        }
    
        {
        std::ostringstream oss;
        oss << "Client" << client.node_id.get_relative_id();
        client.name = oss.str();
        }
    
    clients[ i ] = client;
    }


void generate_master_data_service()
    {
    UInt32 service_index = 1;
    Node service;
    
    service.node_id = NodeId::create_from_type_and_relative_id( NodeId::TYPE_SERVICE, service_index );
    generate_rsa_pair( service.rsa_public_key, service.rsa_private_key );
    service.address = "127.0.0.1";
    service.port = 20000 + service_index;
    service.account_object = "[ null ]";
    service.name = "MasterDataService";
    
    master_data_service = service;
    }

    
void generate_anonymous()
    {
    anonymous.node_id = NodeId::create_from_type_and_relative_id( NodeId::TYPE_ANONYMOUS, 1 );
    generate_rsa_pair( anonymous.rsa_public_key, anonymous.rsa_private_key );
    anonymous.address = "IGNORED";
    anonymous.port = 0;
    anonymous.account_object = "[ null ]";
    anonymous.name = "anonymous";
    }


void write_node
    (
    const Node &                node,
    Registry &                  registry,
    Registry::RegistryNode *    reg_node,
    UInt32                      flags
    )
    {
    if( flags_on( flags, WRITE_NODE_ID ) )
        {
        create_string( registry, reg_node, "node_id", node.node_id.to_string() );
        }
        
    if( flags_on( flags, WRITE_PUBLIC_KEY ) )
        {
        create_string( registry, reg_node, "rsa_public_key", node.rsa_public_key );
        }
        
    if( flags_on( flags, WRITE_PRIVATE_KEY ) )
        {
        create_string( registry, reg_node, "rsa_private_key", node.rsa_private_key );
        }
        
    if( flags_on( flags, WRITE_HOST_ADDRESS ) )
        {
        create_string( registry, reg_node, "address", node.address );
        create_int( registry, reg_node, "port", node.port );
        }
    if( flags_on( flags, WRITE_ACCOUNT_OBJECT ) )
        {
        create_string( registry, reg_node, "account_object", node.account_object );
        }
    if( flags_on( flags, WRITE_NAME ) )
        {
        create_string( registry, reg_node, "name", node.name );
        }
    }


void write_configs
    (
    const Node &    node
    )
    {
    if( nsis )
        {
        return;
        }
    std::string             file_name;
    VolumeManager::OFStream file;
    
    file_name = node.get_directory() + "massiv.conf";
    file.open( file_name, std::ios_base::out | std::ios_base::trunc );
    if( !file )
        {
        error_write_open( file_name );
        }
    
    file
        << "# Paths settings. \n"
        << "!include <massiv.paths.conf>\n\n"
        << "# Various settings. \n"
        << "!include <massiv.settings.conf> \n\n"
        << "# Server nodes info. \n"
        << "!include <../massiv.server_nodes.conf> \n\n"
        << "# Logger settings. \n"
        << "!include <../massiv.logger.conf> \n\n"
        << "# Replication settings. \n"
        << "!include <../massiv.replication.conf> \n";
    
    file.close();
    
    file_name = node.get_directory() + "massiv.settings.conf";
    file.open( file_name, std::ios_base::out | std::ios_base::trunc );
    if( !file )
        {
        error_write_open( file_name );
        }
    
    switch( node.node_id.get_type() )
        {
        case NodeId::TYPE_CLIENT:
            file << "# Settings skeleton...\n";
            file <<
                "[ Settings/DataManager ]\n" <<
                "mount_data_volume : boolean = true\n";
            break;
        
        case NodeId::TYPE_SERVICE:
            file << "# Settings skeleton...\n";
            file <<
                "[ Settings/DataManager ]\n" <<
                "download_type : integer = 0\n";
            break;

        case NodeId::TYPE_SERVER:
            file << "# Settings skeleton... TO BE EDITED BY USER!\n\n";
            file <<
"[ Settings/WellKnownObjectIdDatabase ]\n"
"naming_service_object : string = \"[ 1 1 0 Demo::Lib::NamingService true ]\"\n"
"    # ObjectId of naming service object.\n"
"    # Warning: This object id is valid as long as the object is created\n"
"    # during initial archive creation on the server which owns object\n"
"    # provider with provider id \"1\". Must be edited otherwise.\n";
            break;

        default:
            panic( "Invalid node type." );
        }
    
    file.close();

    file_name = node.get_directory() + "massiv.paths.conf";
    file.open( file_name, std::ios_base::out | std::ios_base::trunc );
    if( !file )
        {
        error_write_open( file_name );
        }
    
    file << "# Settings skeleton...\n\n";
    file <<
"[ Settings/Paths ]\n"
"#login_directory : string = \"$HOME/massiv/" << node.get_directory() << "login\"\n" <<
"#source_data_directory : string = \"$HOME/massiv/" << node.get_directory() << "source_data\"\n" <<
"#data_directory : string = \"$HOME/massiv/" << node.get_directory() << "data\"\n" <<
"#archive_directory : string = \"$HOME/massiv/" << node.get_directory() << "archives\"\n" <<
"#log_directory : string = \"$HOME/massiv/" << node.get_directory() << "log\"\n" <<
"#work_directory : string = \"$HOME/massiv/" << node.get_directory() << "work\"\n" <<
"login_directory : string = \"./login\"\n" <<
"source_data_directory : string = \"./source_data\"\n" <<
"data_directory : string = \"./data\"\n" <<
"archive_directory : string = \"./archives\"\n" <<
"log_directory : string = \"./log\"\n" <<
"work_directory : string = \"./\"\n";

    file.close();
    }


void write_login
    (
    const Node & node
    )
    {
    std::string                 file_name = node.get_directory() + "login/default";
    VolumeManager::OFStream     file;
    Registry                    registry;
    Registry::RegistryNode *    reg_node = registry.find_node_for_path( "" );
    hope( reg_node );
    UInt32                      flags = WRITE_NODE_ID | WRITE_PUBLIC_KEY | 
        WRITE_PRIVATE_KEY;
    
    // Write content of node to registry.
    write_node( node, registry, reg_node, flags );
        
    switch( node.node_id.get_type() )
        {
        case NodeId::TYPE_SERVER:
        case NodeId::TYPE_SERVICE:
            // Open output file.
            file.open( file_name, std::ios_base::out | std::ios_base::trunc );
            if( !file )
                {
                error_write_open( file_name );
                }
            // Write registry to the file.
            registry.save_config_file( file );
            break;
        case NodeId::TYPE_CLIENT:
            {
            // Encrypt login file.
            std::stringstream   ss;
            registry.save_config_file( ss );
            Massiv::Core::AccountManager::encrypt_file( file_name, admin_password, ss.str() );
            }
            break;
        case NodeId::TYPE_ANONYMOUS:
            panic( "Can't create login for anonymous nods." );
            break;
        }       
    }


void write_source_data_description
    (
    const Node & node
    )
    {
    if( nsis )
        {
        return;
        }

    std::string                 file_name;
    VolumeManager::OFStream     file;
    
    file_name = node.get_directory() + "source_data/.update_description";
    file.open( file_name, std::ios_base::out | std::ios_base::trunc );
    if( !file )
        {
        error_write_open( file_name );
        }
    
    file
        << "###\n"
        << "# Configs\n\n"
        << "node config_nodes_private\n"
        << "    type necessary_client\n"
        << "    type necessary_server\n"
        << "    file \"config_nodes_private\"\n"
        << "endnode\n\n"
        << "node config_nodes_public\n"
        << "    type necessary_client\n"
        << "    type necessary_server\n"
        << "    file \"config_nodes_public\"\n"
        << "endnode\n\n"
        << "dataobject config/nodes/private = /config_nodes_private\n"
        << "dataobject config/nodes/public = /config_nodes_public\n";
    }


void write_source_data
    (
    const Node & node
    )
    {
    write_source_data_description( node );
    
    // Public nodes - server and service nodes.
        {
        Registry    registry;
        
        int i;
        for( i = 0; i < num_servers; ++i )
            {
            const Node & node = servers[ i ];
            UInt32 flags = WRITE_ALL_EXCEPT_RSA_PRIVATE_KEY;
            std::string reg_node_name = node.name;
            Registry::RegistryNode * reg_node = registry.create_node_path( reg_node_name );
            hope( reg_node );
            
            write_node( node, registry, reg_node, flags );
            }
        
            {
            const Node & node = master_data_service;
            UInt32 flags = WRITE_ALL_EXCEPT_RSA_PRIVATE_KEY;
            std::string reg_node_name = node.name;
            Registry::RegistryNode * reg_node = registry.create_node_path( reg_node_name );
            hope( reg_node );
            
            write_node( node, registry, reg_node, flags );
            }
        
        std::string                 file_name;
        VolumeManager::OFStream     file;
        
        file_name = node.get_directory() + "source_data/config_nodes_public";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }
        
        file <<
"    # Address and port fields must be edited by client to respect the\n"
"    # the actual node configuration.\n\n"
"    # Account object ids should be set to [ null ] as they are ignored by\n"
"    # non-client nodes.\n\n";
        
        registry.save_config_file( file );
        }

    // Private nodes - client nodes.
        {
        Registry    registry;
        
        int i;
        for( i = 0; i < num_clients; ++i )
            {
            const Node & node = clients[ i ];
            UInt32 flags = WRITE_ALL_EXCEPT_RSA_PRIVATE_KEY;
            std::string reg_node_name = node.name;
            Registry::RegistryNode * reg_node = registry.create_node_path( reg_node_name );
            hope( reg_node );
            
            write_node( node, registry, reg_node, flags );
            }

        std::string                 file_name;
        VolumeManager::OFStream     file;
        
        file_name = node.get_directory() + "source_data/config_nodes_private";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }
        
        file <<
"    # Account object ids must be edited by hand.\n"
"    # However for Client1 it is preset and points to a valid root\n"
"    # account as long as the initial archive was created on a server\n"
"    # which owns the object provider with id \"1\".\n\n"
"    # Address and port fields are ignored for client nodes.\n\n";

        registry.save_config_file( file );
        }
    }

    
void generate_clients()
    {
    clients.resize( num_clients );
    
    int i;
    for( i = 0; i < num_clients; ++i )
        {
        generate_client( i );
        }
    }

    
void generate_servers()
    {
    servers.resize( num_servers );
    
    int i;
    for( i = 0; i < num_servers; ++i )
        {
        generate_server( i );
        }
    }


void write_servers()
    {
    int i;
    for( i = 0; i < num_servers; ++i )
        {
        const Node & node = servers[ i ];
        write_login( node );
        write_configs( node );
        }
    }


void write_clients()
    {
    int i;
    hope( num_clients == 1 );
    for( i = 0; i < num_clients; ++i )
        {
        const Node & node = clients[ i ];
        write_login( node );
        write_configs( node );
        }
    }


void write_master_data_service()
    {
    write_login( master_data_service );
    write_configs( master_data_service );
    write_source_data( master_data_service );
    }


void fill_ip_addresses_and_ports()
    {
    if( ip_address_port.empty() )
        {
        return;
        }
        
    std::istringstream ss( ip_address_port );
    std::string ip;
    std::string port;
    
    char panic_string[] = "Error in ip addresses and ports parameter.";
    
    // First one is service1.
    ss >> master_data_service.address;
    if( ss.fail() )
        {
        panic( panic_string );
        }
    ss >> master_data_service.port;
    if( ss.fail() )
        {
        panic( panic_string );
        }

    for( int i = 0; i < num_servers; i++ )
        {
        ss >> servers[ i ].address;
        if( ss.fail() )
            {
            panic( panic_string );
            }
        ss >> servers[ i ].port;        
        if( ss.fail() )
            {
            panic( panic_string );
            }
        }
    if( !ss.eof() )
        {
        panic( panic_string );
        }    
    }
    
        
void generate()
    {
    // Generate server nodes.
    generate_servers();
    
    // Generate configs for master data service.
    generate_master_data_service();
    
    // Generate client nodes.
    generate_clients();

    // Generate anonymous node.
    generate_anonymous();
    
    // Fill ip addresses and ports.
    fill_ip_addresses_and_ports();
    }


void write()
    {
    // Write server nodes.
    write_servers();
    
    // Write master data service.
    write_master_data_service();
    
    // Write client nodes.
    write_clients();
    
    if( nsis )
        {
        // Write number of servers to file - used by the NSIS installer.
        std::string                 file_name;
        VolumeManager::OFStream     file;
        
        file_name = "nsis";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }
        
        file << num_servers;
        file.close();
        }        

    // Write additional configs.
        {
        std::string                 file_name;
        VolumeManager::OFStream     file;
        
        file_name = "massiv.server_nodes.conf";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }        
        std::ostringstream master_data_service_node_id;
        master_data_service_node_id << master_data_service.node_id;
        std::ostringstream anonymous_node_id;
        anonymous_node_id << anonymous.node_id;
        
        file
            << "# Statically distributed node credentials...\n\n"
            << "[ Settings/NodeDatabase ]\n"
            << "\n"
            << "[ Settings/NodeDatabase/master_data_node ]\n"
            << "node_id : string = " << Massiv::Core::quote_const_string( master_data_service_node_id.str(), true ) << "\n"
            << "rsa_public_key : string = " << Massiv::Core::quote_const_string( master_data_service.rsa_public_key, true ) << "\n"
            << "address : string = " << Massiv::Core::quote_const_string( master_data_service.address, true ) << "\n"
            << "port : integer = " << master_data_service.port << "\n"                
            << "name : string = " << Massiv::Core::quote_const_string( master_data_service.name, true ) << "\n"
            << "account_object : string = " << Massiv::Core::quote_const_string( master_data_service.account_object, true ) << "\n"
            << "\n"
            << "[ Settings/NodeDatabase/anonymous_node ]\n"
            << "node_id : string = " << Massiv::Core::quote_const_string( anonymous_node_id.str(), true ) << "\n"
            << "rsa_public_key : string = " << Massiv::Core::quote_const_string( anonymous.rsa_public_key, true ) << "\n"
            << "rsa_private_key : string = " << Massiv::Core::quote_const_string( anonymous.rsa_private_key, true ) << "\n"
            << "address : string = " << Massiv::Core::quote_const_string( anonymous.address, true ) << "\n"
            << "port : integer = " << anonymous.port << "\n"                
            << "name : string = " << Massiv::Core::quote_const_string( anonymous.name, true ) << "\n"
            << "account_object : string = " << Massiv::Core::quote_const_string( anonymous.account_object, true ) << "\n"
            << "";
            
        file.close();
        
        file_name = "massiv.logger.conf";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }

        file
            << "# Logger settings skeleton...\n\n"
            << "[ Settings/Logger/Destinations ]\n"
            << "dest_cout : string = \"cout\"\n"
            << "[ Settings/Logger/SelectorElements/Cout ]\n"
            << "priority : string = \">= normal\"\n"
            << "type : string = \">= info\"\n"
            << "facility : string = \"any\"\n"
            << "[ Settings/Logger/Selectors/Rule1/Elements ]\n"
            << "element1 : string = \"Cout\"\n"
            << "[ Settings/Logger/Selectors/Rule1/Destinations ]\n"
            << "destination1 : symlink = Settings/Logger/Destinations/dest_cout\n";

        file.close();

        file_name = "massiv.replication.conf";
        file.open( file_name, std::ios_base::out | std::ios_base::trunc );
        if( !file )
            {
            error_write_open( file_name );
            }

        file
            << "# Replication settings skeleton...\n\n"
            << "[ Settings/ObjectManagement/Replication ]\n"
            << "client_repflags_no : integer = 0\n"
            << "client_repflags_yes : integer = 5\n";

        file.close();
        }
    }


void print_usage
    (
    const std::string & argv0
    )
    {
    std::cout
        << "\nDescription: Create massiv simulation config files for one service,\nN servers and one administrator client.\n"
        << "\nUsage: " << argv0 << " [switches]\n"
        << "  -s <N> ... Create configs for N server nodes.\n"
        << "  -i \"<service ip> <service port> <server1 ip> <server1 port>... \n"
        << "       <serverN ip> <serverN port>\"\n"
        << "         ... Set ip addresses and names for one service and N servers.\n"
        << "  -n     ... Generate simulation config files for NSIS\n"
        << "  -p     ... Set administrator password.\n"
        << "\n";
    }


void print_done()
    {
    std::cout
        << "Done...\n"
        << "Now you must:\n"
        << " - For server and service nodes update their addresses and ports in "
        << " source_data/config_nodes_public.\n"
        << " - For each node set its settings in massiv.settings.conf file, edit\n"
        << "massiv.server_nodes.conf.\n"
        << " - For master data service fill its source_data with minimal data.\n"
        << " - Create startup archives.\n"
        << " - Create account objects and put their object ids to\n"
        << " source_data/config_nodes_private to corresponding clients.\n"
        << " - Add object id naming service to settings files.\n"
        ;
        
    }


bool parse_args
    (
    int     argc,
    char ** argv
    )
    {
    std::string non_parameter_switches = "n";
    std::string parameter_switches = "sip";
    CmdLineParser   cmd_line( argc, argv, non_parameter_switches, parameter_switches );
    bool parse = true;
    while( parse )
        {
        std::string parameter;
        char option;
        try
            {
            option = cmd_line.get_next( parameter );
            }
        catch( Massiv::Core::Exception )
            {        
            std::cerr << "Unknown option.\n";
            print_usage( argv[ 0 ] );
            return false;
            }
        switch( option )
            {
            case '-':
            case '\0':
                parse = false;
                break;
            
            case 's':
                num_servers = Massiv::Core::ManagedConversion< int, std::string >( parameter );
                break;
                        
            case 'n':
                nsis = true;
                break;

            case 'i':
                // IP adresses and ports for service and servers.
                // Format is "<service ip> <service port> <server1 ip> <server1 port> ...".
                ip_address_port = parameter;
                break;

            case 'p':
                // Administrator password.
                admin_password = parameter;
                break;

            default:
                std::cerr << "Unknown option " << option << "." << std::endl;
                print_usage( argv[ 0 ] );
                return false;
            }
        }
    
    return true;
    }


int main
    (
    int        argc,
    char **    argv
    )
    {
    try
        {
        ::srand( static_cast< unsigned int >( std::time( NULL ) ) );
        
        // Read arguments.
        if( !parse_args( argc, argv ) )
            {
            return 1;
            }
        
        if( argc == 1 )
            {
            // No argument.
            print_usage( argv[ 0 ] );
            return 2;
            }
            
        if( num_servers <= 0 )
            {
            // Number of server was not specified.
            std::cerr << "Invalid number of server nodes. Must be at least one." << std::endl;
            print_usage( argv[ 0 ] );
            return 3;
            }
        // Initialize massiv core.
        Massiv::System::StartUpInfo info;
        info.base_services = true;
        info.file_systems = true;
        Massiv::System::initialize( Massiv::System::StartUpInfo( info ) );

        // Generate configs.
        generate();

        // Write configs.
        write();
    
        // Shutdown massiv core.
        Massiv::System::shutdown();
        }
    catch( Massiv::Core::Exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        return 3;
        }
    catch( std::exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        return 4;
        }
    catch( ... )
        {
        std::cout << "Unknown exception. " << std::endl;
        throw;
        }
    
    print_done();
    return 0;
    }
