/** @file data_update.cpp
 * Command line utility for master data server data updating.
 */

/*
 * begin        : Sun Mar 30 2003
 * copyright    : (C) 2003 by Martin Havlista
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "utility/debug.h"
#include "global/global.h"
#include "utility/text_reader.h"
#include "utility/text_writer.h"
#include "status/exception.h"
#include "cmdline/cmdline_parser.h"
#include "data/data_manager.h"
#include "data/copy_file.h"
#include "serializer/serializer_int.h"
#include "thread/thread.h"
#include "thread/synch.h"
#include "thread/shmem.h"
#include <stdio.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>

using namespace Massiv::Core;
using namespace Massiv;
using namespace std;

#define VERSION                         "1.1"
    /**< Version of data update utility. */
#define MASSIV_SETTINGS_CONF            "massiv.settings.conf"
    /**< Massiv settings configuration file name. */
    
// Global variables.
string              settings_directory;
    /**< Path to directory with massiv settings file MASSIV_SETTINGS_CONF. */
string              description_file_name;
    /**< Full file name to data description file. */
string              data_directory;
    /**< Data directory for data files update. */
DataManager *       dm = NULL;
    /**< Instance of DataManager. */
string              source_directory;
    /**< Relative path to a directory with new versions of data files. */
CriticalSection *   cs = NULL;
    /**< Critical section used to detect if any massiv server is running
         and using data directory for update. */
SharedMemory *      shmem = NULL;
    /**< Shared memory used to communicate with running massiv process. */
CriticalSection *   update_lock = NULL;
    /**< Critical section used for exlusive access of data update
         utility to shared memory. */
bool                check;
    /**< True when data update utility just check if it's possible to update data. */
DataManager::DataVersion         data_version;
    /**< Version of data to update. Value is read from configuration file. */
        
/**
 * Assert within realase mode.
 */
#define assert_release( condition, text ) \
    if( condition == 0 ) \
        { \
        std::ostringstream  ss; \
        ss << text << std::endl << "Source file: " << __FILE__ << " line: " << __LINE__ << std::endl; \
        throw ExDataManagementError( ss.str() ); \
        }

/**
 * Print usage of data update utility.
 */
void print_usage
    (
    const std::string & argv0
    )
    {
    cout << "Usage: data_update " << argv0 << " [ -u ] [ -c ] [ -n ]" << endl << endl;
    cout << "u      process update" << endl;
    cout << "c      check if it's possible to update data" << endl;
    cout << "n      create empty update config file \"" << DM_UPDATE_FILE_NAME 
         << "\" with an example." << endl;
    }

/**
 * Throws an exception while parsing massiv configuration file with custom 
 * @param message.
 */
void settings_parse_error
    (
    const string &  message = ""
    )
    {
    string  msg = message;
    if( !message.empty() )
         msg = ": " + message;
    throw ExCommonError( "Error occured while parsing configuration file" + msg );
    }    

/**
 * Parse @param line where the first part of string is @param field_name,
 * and the value is stored to @param value.
 */
int parse_settings_file_string
    (
    const string &  line,
    const string &  field_name,
    string &        value
    )
    {
    if( line.find( field_name.c_str(), 0, field_name.size() ) != std::string::npos )
        {
        string::size_type i1 = line.find( "\"" );
        if( i1 == std::string::npos )
            {                
            settings_parse_error( "quotation marks are missing" );
            }
        i1++;
        string::size_type i2 = line.find( "\"", i1 );
        if( i2 == std::string::npos )
            {                
            settings_parse_error( "quotation marks are missing" );
            }
        i2--;
        value.assign( line, i1, i2 - i1 + 1 );            
        return true;
        }
    return false;
    }

/**
 * Parse @param line where the first part of string is @param field_name,
 * and the value is stored to @param value.
 */
int parse_settings_file_int
    (
    const string &  line,
    const string &  field_name,
    int &           value
    )
    {
    if( line.find( field_name.c_str(), 0, field_name.size() ) != std::string::npos )
        {
        string::size_type i1 = line.rfind( "=" );
        if( i1 == std::string::npos )
            {                
            settings_parse_error();
            }
        i1++;
        string  temp( line, i1, string::npos );
        istringstream ss( temp );
        ss >> value;
        return true;
        }
    return false;
    }

/**
 * Check if massiv settings configuration file has valid data manager/settings 
 * information to pass data update.
 * This function also sets some global variables.
 */
void check_settings
    (
    const std::string & argv0
    )
    {
    if( settings_directory[ settings_directory.size() - 1 ] != '/' )
        {
        settings_directory += "/";
        }
    ifstream ifs( ( settings_directory + MASSIV_SETTINGS_CONF ).c_str() );
    if( !ifs.is_open() )
        {
        print_usage( argv0 );
        ostringstream   ss;
        ss << endl << "Error: cannot open configuration file " << settings_directory 
           << MASSIV_SETTINGS_CONF << endl;
        throw ExCommonError( ss.str() );
        }
    
    // Parse configuration file and find data manager settings values.
    string  line;
    
    data_directory =  ".";
    int     data_server_type = -1;
    
    bool    search = false;
    while( getline( ifs, line ) )
        {
        if( !search && line.compare( "[ Settings/DataManager ]" ) != 0 )
            {
            continue;
            }
        else
            {
            search = true;
            }
        parse_settings_file_string( line, DM_SETTINGS_DATA_DIRECTORY, data_directory );
        parse_settings_file_int( line, DM_SETTINGS_TYPE, data_server_type );
        }
    if( data_server_type == -1 )
        {
        settings_parse_error( "DataManager type is missing" );
        }
    if( data_server_type != 0 ) // 0 == MASTER_DATA_SERVER
        {
        cout << "Error: DataManager is not MASTER_DATA_SERVER.";
        }
    // Let's find file DESCRIPTION
    ifs.close();
    string  desc_dir;
    get_absolute_path( data_directory, desc_dir, settings_directory );
    data_directory = desc_dir;
    description_file_name = data_directory + "/" + DM_DD_FILE_NAME;
    ifs.open( description_file_name.c_str() );
    if( !ifs.is_open() )
        {
        print_usage( argv0 );
        ostringstream   ss;
        ss << endl << "Error: cannot open data description file " << description_file_name << endl; 
        throw ExCommonError( ss.str() );
        }
    ifs.close();
    }

/**
 * Initialize data manager of data update utility.
 * @param exclusive_access - false to check validity od data update information.
 *                           true to process real data update.
 * Massiv process is not running. Data update will be processed
 * by data manager of data update utility.
 */
void initialize_data_manager
    (
    bool            exclusive_access
    )
    {
    // FIXME - initialize global VolumeManager
    static VolumeManager & vm = Global::volume_manager();
    if( !dm )
        {
        dm = new DataManager();
        if( !dm )
            throw ExCommonError( "Memory allocation error." );
        }
    dm->_initialize
        (
        data_directory,
        DataManager::MASTER_DATA_SERVICE,
        DataManager::NO_DOWNLOAD,
        true,
        exclusive_access
        );
    while( !dm->is_ready_for_use() )
        {
        dm->next_tick();
        }
    // Check if data version is valid.
    if( data_version != dm->get_data_version() )
        {
        throw ExCommonError( "Wrong data version for update. Check update configuration file." );
        }
    }

/**
 * Set upload provider to test mode and check if data update information
 * from data update config file are valid.
 * Throws an exception when an error occurs.
 */
void check_data_update_config()
    {
    dm->updatep_set_test();
    dm->switch_to_update_phase( DataManager::AUTOMATIC, DM_UPDATE_FILE_NAME );
    while( !dm->update_finished() )
        {
        dm->next_tick();
        }
    }


/**
 * Process data update.
 * This function is called when data update information are checked for validity
 * and it's possible to update data.
 */
void process_data_update()
    {
    dm->switch_to_update_phase( DataManager::AUTOMATIC, DM_UPDATE_FILE_NAME );
    while( !dm->update_finished() )
        {
        dm->next_tick();
        }
    }

    
/**
 * Tests if any massiv process is running and using data directory.
 * Returns true when no such process is running - so data update will
 * be processed by data manager of data update utility.
 */
bool provide_secure_access_to_data_directory()
    {
    std::string     data_dir( data_directory );
    to_unix_slash( data_dir );
    
    data_dir = DM_PROCESS_COMMUNICATION_ACCESS_LOCK + data_dir;
    static CriticalSectionManager & csm = Global::critical_section_manager();
    cs = csm.create( true, data_dir.c_str() );
    if( !cs->try_enter() )
        {
        return false;
        }
    return true;
    }    

/**
 * Create data update config file with update information in data
 * directory. Data manager of running massiv process will use this
 * information for data update.
 */
void copy_update_cfg()
    {
    string  source_dir;
    get_absolute_path( source_directory, source_dir );
    string  name = data_directory + "/" + DM_UPDATE_FILE_NAME;
    ofstream    ofs( name.c_str() );
    if( !ofs.is_open() )
        {
        ostringstream   ss;
        ss << "Cannot create file \"" << name << "\".";
        throw ExCommonError( ss.str() );
        }
    ifstream    ifs( DM_UPDATE_FILE_NAME );
    if( !ifs.is_open() )
        {
        ostringstream   ss;
        ss << "Cannot open file \"" << DM_UPDATE_FILE_NAME << "\".";
        throw ExCommonError( ss.str() );
        }
    TextWriter  tw( ofs );
    TextReader  tr( ifs );
    
    // Copy every line to new file except field source directory - source directory
    // is relative path and must be changed to full path.
    while( tr.next_line() )
        {
        if( tr.str() == DM_UPDATE_CFG_SETUP )
            {
            // Change source directory to full directory.
            tw.line() << tr.str();
            tw.end_line();
            tw.line() << DM_UPDATE_CFG_SOURCE_DATA_DIRECTORY << "=" << source_dir; 
            tw.end_line();
            while( tr.next_line() && tr.str()[0] != '[' );
            }
        tw.line() << tr.str();
        tw.end_line();
        }
    
    ofs.close();
    ifs.close();
    }

/**
 * Load some information from data update config file and set
 * global variables with this values.
 */
void load_config()
    {
    ifstream    ifs( DM_UPDATE_FILE_NAME );
    if( !ifs.is_open() )
        {
        ostringstream   ss;
        ss << "Cannot config file \"" << DM_UPDATE_FILE_NAME << "\".";
        throw ExCommonError( ss.str() );
        }
    TextReader  tr( ifs );
    while( tr.next_line() )
        {
        if( tr.str() == DM_UPDATE_CFG_SETUP )
            {
            while( tr.next_line() && tr.str()[0] != '[' )
                {
                std::string     temp1, temp2;
                parse_equality_line( tr.str(), temp1, temp2 );
                if( temp1 == DM_UPDATE_CFG_SOURCE_DATA_DIRECTORY )
                    {
                    source_directory = temp2;
                    }
                if( temp1 == DM_UPDATE_CFG_SETTINGS_DIRECTORY )
                    {
                    settings_directory = temp2;
                    }
                if( temp1 == DM_UPDATE_CFG_DATA_VERSION_UPDATE )
                    {
                    std::stringstream ss( temp2 );
                    TextReader  tr2( ss );
                    tr2.next_line();
                    data_version.text_read( tr2 );
                    }                    
                }
            }
        }
    if( settings_directory == "" )
        {
        throw ExCommonError( "Update config error - cannot find field for settings directory." );
        }
    if( source_directory == "" )
        {
        throw ExCommonError( "Update config error - cannot find field for source directory." );
        }
    ifs.close();
    }

/**
 * Notify data manager of running massiv process about data update.
 * Data update file must be copied to data directory before calling this function.
 */
void notify_about_update()
    {
    try 
        {
        std::string     data_dir( data_directory );
        to_unix_slash( data_dir );
            
        data_dir = DM_PROCESS_COMMUNICATION_UPDATE_MEMORY + data_dir;
        shmem = new SharedMemory( sizeof( DataManager::UpdateNotify ), data_dir );
        assert_release( shmem, "Memory allocation error." );
        volatile DataManager::UpdateNotify & un = *( static_cast< DataManager::UpdateNotify * >( shmem->get_memory() ) );
        
        // Create update lock.
        std::string     lock_name( data_directory );
        to_unix_slash( lock_name );
        lock_name = DM_PROCESS_COMMUNICATION_UPDATE_LOCK + lock_name;
        static CriticalSectionManager & csm = Global::critical_section_manager();
        update_lock = csm.create( true, lock_name.c_str() );

        if( !update_lock->try_enter() )
            {
            throw ExCommonError( "Data update is probably in progress." );
            }
        if( un != DataManager::UPDATE_NOTIFY_IDLE )
            {
            throw ExCommonError( "Data update fatal error." );
            }
        un = DataManager::UPDATE_NOTIFY_START_UPDATE;
        
        int timeout = 10;
        while( un != DataManager::UPDATE_NOTIFY_UPDATE_IN_PROGRESS && 
            un != DataManager::UPDATE_NOTIFY_UPDATE_FINISHED && 
            un != DataManager::UPDATE_NOTIFY_UPDATE_ERROR )
            {
            Thread::sleep( 1000 );
            timeout--;
            if( timeout == 0 )
                {
                throw ExCommonError( "Timeout: running massiv process is not able to update data." );
                }
            }            
        cout << "Update is processed by running massiv process..." << endl;
        while( un != DataManager::UPDATE_NOTIFY_UPDATE_FINISHED && 
            un != DataManager::UPDATE_NOTIFY_UPDATE_ERROR )
            {
            Thread::sleep( 1000 );
            }            
        if( un == DataManager::UPDATE_NOTIFY_UPDATE_ERROR )
            {
            ostringstream ss;
            ss << "Update fatal error: data update failed. See log file of massiv process "
               << "for more information.";
            throw ExCommonError( ss.str() );
            }
        } // try
    catch( Exception & e )
        {
        update_lock->leave();
        delete update_lock;
        delete shmem;
        throw e;
        }
    }

/**
 * Create empty data update config file with en example how to use it.
 * If will be created in actual directory.
 */ 
void create_empty_update_config()
    {
    ifstream    ifs( DM_UPDATE_FILE_NAME );
    if( ifs.is_open() )
        {
        ifs.close();
        ostringstream   ss;
        ss << "Config file \"" << DM_UPDATE_FILE_NAME << "\" already exist. Delete it first.";
        throw ExCommonError( ss.str() );
        }
    ofstream    ofs( DM_UPDATE_FILE_NAME );
    if( !ofs.is_open() )
        {
        ostringstream   ss;
        ss << "Cannot create config file \"" << DM_UPDATE_FILE_NAME << "\".";
        throw ExCommonError( ss.str() );
        }
    ofs << 
    "# Configuration file for data_update utility, version " << VERSION << endl <<
    "#" << endl <<
    "# Example of configuration file:" << endl <<
    "#" << DM_UPDATE_CFG_SETUP << endl <<
    "#" << DM_UPDATE_CFG_SOURCE_DATA_DIRECTORY << "=./source_data" << endl <<
    "#" << DM_UPDATE_CFG_SETTINGS_DIRECTORY << "=./server" << endl <<
    "#" << DM_UPDATE_CFG_DATA_VERSION_UPDATE << "=1.1" << endl <<
    "#" << endl <<
    "#" << DM_UPDATE_CFG_DATA_FILE_UPDATE << endl <<
    "#figure.do" << endl <<
    "#man.do" << endl <<
    "#" << endl <<
    "#" << DM_UPDATE_CFG_ADD_DATA_NODES << endl <<
    "#{" << endl <<
    "#/graphic/graphic_update" << endl <<
    "#0" << endl <<
    "#graphic_update.do" << endl <<
    "#}" << endl <<
    "#" << endl <<
    "#" << DM_UPDATE_CFG_ADD_DATA_OBJECTS << endl <<
    "#graphic_update=/graphic/graphic_update" << endl <<
    "#" << endl;

    ofs.close();
    cout << "Update config file \"" << DM_UPDATE_FILE_NAME << "\" successfuly created." << endl;
    }

int main
    (
    int        argc,
    char **    argv
    )
    {        
    cout << "Massiv data update utility, version " << VERSION << endl << endl;
    
    if( argc == 1 )
        {
        print_usage( argv[ 0 ] );
        return 1;
        }
    
    CmdLineParser   cmd_parser( argc, argv, "ucn", "" );
    char            cswitch;
    string          parameter;
    
    try
        {
        while( ( cswitch = cmd_parser.get_next( parameter ) ) )
            {
            switch( cswitch )
                {
                case 'u':
                    {
                    check = false;
                    break;
                    }
                case 'c':
                    {
                    check = true;
                    break;
                    }
                case 'n':
                    {
                    create_empty_update_config();
                    return 0;
                    }
                default:
                    {
                    cout << endl << "Error: Invalid switch." << endl;
                    print_usage( argv[ 0 ] );
                    return 1;
                    }
                }
            }

        load_config();
        
        // Check if it is possible to update data files.
        check_settings( argv[ 0 ] );
        initialize_data_manager( false );

        check_data_update_config();
        
        if( check )
            {
            cout << "Everything looks good. Run data update utility with parameter ' -u ' to process data update." << endl;
            return 0;
            }
        cout << endl << "Everything looks good. Do you really want to process data update? <Y/N>";
        char    c;
        cin >> c;
        switch( c )
            {
            case 'y':
            case 'Y':
                {
                break;
                }
            case 'n':
            case 'N':
                {
                cout << "Key 'N' pressed. Data update WILL NOT be processed." << endl;
                return 0;
                }
            default:
                cout << "Error: press key 'Y' or 'N'..." << endl;
                return 1;
            }
                
        delete dm;
        // Now everything is ready to update master data server.
        if( !provide_secure_access_to_data_directory() )
            {
            // Massiv process is running - update will be processed by data manager
            // of running master data server.
            
            copy_update_cfg();
            
            // Notify running process about update.
            
            notify_about_update();
            
            // We must wait some time till data manager of running massiv process
            // updates data. 
            }
        else
            {
            // Massiv process is not running - data update utility will process data update.
            // Access to data directory is now granted.
            
            // Initialize data manager again. Last initialization was to check
            // if it is possible to process data update.
            initialize_data_manager( true );
            process_data_update();
            delete dm;
            cs->leave();
            delete cs;
            }
            
        cout << endl << "Data succesfully updated." << endl;

        }
    catch( ExCommandLineError & e )
        {
        cout << "Commandline syntax error: " << e.what() << endl << endl;
        return 1;
        }
    catch( Exception & e )
        {
        cout << "Unexpected error: " << e.what() << endl;
        return 1;
        }
    return( 0 );
    }
