/** @file list_volume.cpp
 * Command line utility for listing contents of a volume (BasicVolume or its
 * subclass).
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "global/global.h"
#include "logger/logger.h"
#include "utility/debug.h"
#include "system/system.h"
#include "status/exception.h"
#include "cmdline/cmdline_parser.h"
#include "filesys/volume.h"
#include "filesys/compact_volume.h"
#include <iomanip>
#include <iostream>
#include <sstream>
#include <memory>

using namespace Massiv;
using namespace Massiv::Core;


static  std::string     volume_image;
    /**< Volume image to be listed. */
static  std::string     path_in_volume;
    /**< Path of the volume to list. */
static  std::string     mask( "*" );
    /**< Mask for filtering files. */
static  bool            recursive = false;
    /**< true = list recursively all subdirectories; false = list only one. */
static  bool            interpret_compact_volume = false;
    /**< Interpret volume image as CompactVolume image. */
static  bool            print_sizes = false;
    /**< Print stream sizes. */
static  bool            print_directories = false;
    /**< Print directories even in non-recursive mode. */
static  bool            force_recursive = false;
    /**< Recurse into subdirectories without checking if they match the mask.*/
static  int             size_width = 10;
    /**< Minimum width of the size information column. */
static  int             omit_header = false;
    /**< Don't print header with general volume information. */


/**
 * Prints the utility usage information.
 */
void print_info
    (
    const std::string & argv0
    );

/**
 * Processes the command line.
 * @return true if list_volume() should be processed.
 */
bool process_cmdline
    (
    int             argc,
    char **         argv
    );

/**
 * The actual volume listing utility.
 */
void list_volume();

/**
 * Lists one directory or whole directory tree (according to @c recursive 
 * global variable).
 * @param it Volume iterator bound with directory that has to be listed.
 * @param relative_path Relative path to the directory being listed - relative
 *      to directory given using the -p switch or to the volume root if the
 *      switch has been omitted.
 */
void list_directory
    (
    FileIterator    it,
    std::string     relative_path
    );


int main
    (
    int             argc,
    char **         argv
    )
    {
    try
        {
        // Initialize massiv core.
        System::StartUpInfo startup_info( NodeId::TYPE_ANONYMOUS, "" );
        startup_info.load_config = false;
        startup_info.quiet_start = true;
        System::initialize( startup_info );
        Global::logger().enable( true );

        if( process_cmdline( argc, argv ) )
            {
            list_volume();
            }

        // Shutdown massiv core.
        System::shutdown();
        }
    catch( std::exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        }
    catch( ... )
        {
        std::cout << "Unknown exception leaked from list_volume " << std::endl;
        }

    return 0;
    }  // main()


void print_info
    (
    const std::string & argv0
    )
    {
    std::cout << "Volume listing utility." << std::endl
              << "Usage: " << argv0 << " [switches] <volume_image>"
              << std::endl << "Switches:"
              << std::endl << "  -m <mask>  ... specifies a mask for filtering files"
              << std::endl << "  -p <path>  ... specifies a directory in the volume to be listed"
              << std::endl << "  -w <width> ... width of the size information column (used with -s)"
              << std::endl << "  -c ... interpet volume image as CompactVolume"
              << std::endl << "  -d ... print directory names even in non-recursive mode"
              << std::endl << "  -h ... show this message"
              << std::endl << "  -o ... omit header with general volume image information"
              << std::endl << "  -r ... list directories recursively"
              << std::endl << "  -R ... list all directories recursively"
              << std::endl << "         (don't try to match subdirectory names with mask)"
              << std::endl << "  -s ... print stream sizes"
              << std::endl << "  -? ... show this message" 
              << std::endl << std::endl
              << "Example: " << argv0 << " -m t* -rs -p /img/textures ../../volume.image"
              << std::endl << std::endl;
    }  // print_info()


bool process_cmdline
    (
    int             argc,
    char **         argv
    )
    {
    CmdLineParser   cmd_parser( argc, argv, "cdhorRs?", "mpw" );
    char            cswitch;
    std::string     parameter;

    try
        {
        while( ( cswitch = cmd_parser.get_next( parameter ) ) )
            {
            switch( cswitch )
                {
            case 'c':
                interpret_compact_volume = true;
                break;

            case 'd':
                print_directories = true;
                break;

            case 'm':
                mask = parameter;
                break;

            case 'p':
                path_in_volume = parameter;
                break;

            case 'r':
                recursive = true;
                break;

            case 'R':
                recursive = true;
                force_recursive = true;
                break;
        
            case 's':
                print_sizes = true;
                break;

            case 'h':
            case '?':
                print_info( argv[ 0 ] );
                return false;

            case 'o':
                omit_header = true;
                break;

            case 'w':
                {
                std::istringstream ss( parameter );
                ss >> size_width;
                break;
                }
        
            case '-':
                if( volume_image.size() )
                    {
                    std::cout << "Error: Duplicate volume_image argument." 
                              << std::endl << std::endl;

                    print_info( argv[ 0 ] );
                    return false;
                    }

                volume_image = parameter;
                break;

            default:
                panic( "Invalid switch." );
                break;
                }
            }
        }
    catch( ExCommandLineError & e )
        {
        std::cout << "Commandline syntax error: " << e.what() 
                  << std::endl << std::endl;

        print_info( argv[ 0 ] );
        return false;
        }
    catch( Exception & e )
        {
        std::cout << "Unexpected error: " << e.what() << std::endl;
        return false;
        }

    if( !volume_image.size() )
        {
        // Volume image was not given on the command line.
        print_info( argv[ 0 ] );
        return false;
        }

    return true;
    }  // process_cmdline()


void list_volume()
    {
    std::auto_ptr<BasicVolume> volume;
        
    if( interpret_compact_volume )
        {
        volume = std::auto_ptr<BasicVolume>
            ( new CompactVolume( volume_image, std::ios_base::in ) );
        }
    else
        {
        volume = std::auto_ptr<BasicVolume>
            ( new Volume( volume_image, std::ios_base::in ) );
        }

    FileIterator it = volume->open_directory
        ( 
        path_in_volume, 
        mask,
        force_recursive ? 
            FileIterator::ENUMERATE_ALL_DIRS :
            FileIterator::MATCH_DIRS_NAMES
        );

    if( !omit_header )
        {
        std::string volume_type;

        if( interpret_compact_volume )
            {
            volume_type = "compact";
            }
        else
            {
            volume_type = "volume";
            }

        std::cout << "Volume name: " << volume->get_volume_name() 
                  << "; Type: " << volume_type << std::endl;
        std::cout << "---------------------------------------------------"
                  << "----------------------------" << std::endl;
        }

    list_directory( it, "" );
    
    it.unbind();
    }  // list_volume()


void list_directory
    (
    FileIterator    it,
    std::string     relative_path
    )
    {
    while( it.is_valid() )
        {
        if( it.is_file() )
            {
            if( print_sizes )
                {
                std::cout << std::setw( size_width ) << it.size() << " ";
                }
            else
            if( print_directories )
                {
                std::cout << "      ";
                }

            std::cout << relative_path + it.name() << std::endl;
            }
        else
        if( it.is_directory() )
            {
            if( print_directories )
                {
                if( print_sizes )
                    {
                    std::cout << std::setw( size_width ) << "<DIR>" << " ";
                    }
                else
                    {
                    std::cout << "<DIR> ";
                    }

                std::cout << relative_path + it.name() << std::endl;
                }
            
            if( recursive )
                {
                list_directory( it.enter(), relative_path + it.name() + '/' );
                }
            }

        it++;
        }
    }  // list_directory()
