/** @file pack_volume.cpp
 * Command line utility for packing files from OS filesystem into a volume
 * (managed by BasicVolume or its subclass).
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "global/global.h"
#include "logger/logger.h"
#include "utility/debug.h"
#include "system/system.h"
#include "status/exception.h"
#include "cmdline/cmdline_parser.h"
#include "filesys/volume.h"
#include "filesys/compact_volume.h"
#include <iostream>
#include <sstream>
#include <memory>

using namespace Massiv;
using namespace Massiv::Core;


static  std::string     volume_image;
    /**< Path to the volume to pack files into. */
static  std::string     volume_name( "noname" );
    /**< Name of the volume to pack files into. Used only if new volume is
         being created. */
static  std::string     path_in_volume;
    /**< Specifies where the files should be packed into in the volume. */
static  std::string     mask( "*" );
    /**< Mask for filtering files. */
static  bool            recursive = false;
    /**< true = pack recursively all subdirectories; false = pack only one. */
static  bool            force_recursive = false;
    /**< Recurse into subdirectories without checking if they match the mask.*/
static  bool            compact = false;
    /**< Pack files into the CompactVolume. */
static  bool            erase = false;
    /**< true - erase files after being packed into the volume. */
static  bool            format = false;
    /**< Format the volume first. */
static  std::list< std::string >    sources;
    /**< List of names of all directories and files to pack. */


/**
 * Prints the utility usage information.
 */
void print_info
    (
    const std::string & argv0
    );

/**
 * Processes the command line.
 * @return true if pack_volume() should be processed.
 */
bool process_cmdline
    (
    int             argc,
    char **         argv
    );

/**
 * The actual volume packing utility.
 */
void pack_volume();



int main
    (
    int        argc,
    char **    argv
    )
    {
    try
        {
        // Initialize massiv core.
        System::StartUpInfo startup_info( NodeId::TYPE_ANONYMOUS, "" );
        startup_info.load_config = false;
        startup_info.quiet_start = true;
        System::initialize( startup_info );
        Global::logger().enable( true );

        if( process_cmdline( argc, argv ) )
            {
            pack_volume();
            }
        
        // Shutdown massiv core.
        System::shutdown();
        }
    catch( Exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        }
    catch( std::exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        }
    catch( ... )
        {
        std::cout << "Unknown exception leaked list_volume " << std::endl;
        }

    return 0;
    } // main()


void print_info
    (
    const std::string & argv0
    )
    {
    std::cout << "Utility for packing files into volume." << std::endl
              << "Usage: " << argv0 << " [switches] <volume_image> <list_of_sources>"
              << std::endl
              << std::endl << "<volume_image>    ... specifies a image file to pack to"
              << std::endl << "<list_of_sources> ... list of directories/files to pack into the volume"
              << std::endl << std::endl << "Switches:"
              << std::endl << "  -m <mask>  ... specifies a mask for filtering files"
              << std::endl << "  -n <name>  ... name of the volume (used only when new volume is created)"
              << std::endl << "  -p <path>  ... specifies a destination path in the volume"
              << std::endl << "  -c ... pack files into compact volume"
              << std::endl << "  -f ... format the volume first (you must use this when creating new volume)"
              << std::endl << "  -e ... erase source files after being packed"
              << std::endl << "  -h ... show this message"
              << std::endl << "  -r ... pack directories recursively"
              << std::endl << "  -R ... pack all directories recursively"
              << std::endl << "         (don't try to match subdirectory names with mask)"
              << std::endl << "  -? ... show this message" 
              << std::endl << std::endl
              << "Example: " << argv0 << " -m t* -r ../../volume.image data/source"
              << std::endl << std::endl;
    }  // print_info()


bool process_cmdline
    (
    int             argc,
    char **         argv
    )
    {
    CmdLineParser   cmd_parser( argc, argv, "cefhrR?", "mnp" );
    char            cswitch;
    std::string     parameter;

    try
        {
        while( ( cswitch = cmd_parser.get_next( parameter ) ) )
            {
            switch( cswitch )
                {
            case 'm':
                mask = parameter;
                break;

            case 'n':
                volume_name = parameter;
                break;

            case 'p':
                path_in_volume = parameter;
                break;

            case 'c':
                compact = true;
                break;

            case 'e':
                erase = true;
                break;
            
            case 'f':
                format = true;
                break;

            case 'r':
                recursive = true;
                break;

            case 'R':
                recursive = true;
                force_recursive = true;
                break;
        
            case 'h':
            case '?':
                print_info( argv[ 0 ] );
                return false;

            case '-':
                if( !volume_image.size() )
                    {
                    volume_image = parameter;
                    }
                else
                    {
                    sources.push_back( parameter );
                    }

                break;

            default:
                panic( "Invalid switch." );
                break;
                }
            }
        }
    catch( ExCommandLineError & e )
        {
        std::cout << "Commandline syntax error: " << e.what() 
                  << std::endl << std::endl;

        print_info( argv[ 0 ] );
        return false;
        }
    catch( Exception & e )
        {
        std::cout << "Unexpected error: " << e.what() << std::endl;
        return false;
        }

    if( !volume_image.size() || ( !sources.size() && !format ) )
        {
        // Some of mandatory parameters were not given
        print_info( argv[ 0 ] );
        return false;
        }

    return true;
    }  // process_cmdline()


void pack_volume()
    {
    std::auto_ptr<BasicVolume> volume;

    if( compact )
        {
        if( format )
            {
            volume = std::auto_ptr<BasicVolume>
                ( new CompactVolume( volume_image, volume_name ) );
            }
        else
            {
            volume = std::auto_ptr<BasicVolume>
                ( new CompactVolume( volume_image ) );
            }
        }
    else
        {
        if( format )
            {
            volume = std::auto_ptr<BasicVolume>
                ( new Volume( volume_image, volume_name ) );
            }
        else
            {
            volume = std::auto_ptr<BasicVolume>
                ( new Volume( volume_image ) );
            }
        }

    while( !sources.empty() )
        {
        volume->pack
            (
            path_in_volume, 
            sources.front(), 
            mask, 
            BasicVolume::DEFAULT_STREAM_FLAGS, 
            erase, 
            recursive, 
            force_recursive ? 
                FileIterator::ENUMERATE_ALL_DIRS :
                FileIterator::MATCH_DIRS_NAMES
            );

        sources.pop_front();
        }
    }  // pack_volume()
