/** @file pack_volume.cpp
 * Command line utility for unpacking files from a volume (managed by 
 * BasicVolumePacking or some of its descendants) to an OS filesystem.
 */

/*
 * begin        : Tue Mar 11 2003
 * copyright    : (C) 2003 by Marek Svantner
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "global/common.h"
#include "global/global.h"
#include "logger/logger.h"
#include "utility/debug.h"
#include "system/system.h"
#include "status/exception.h"
#include "cmdline/cmdline_parser.h"
#include "filesys/volume.h"
#include "filesys/compact_volume.h"
#include <iostream>
#include <sstream>
#include <memory>

using namespace Massiv;
using namespace Massiv::Core;


static  std::string     volume_image;
    /**< Path to the volume to unpack files from. */
static  std::string     path_in_volume;
    /**< Specifies where the files should be unpacked from. */
static  std::string     mask( "*" );
    /**< Mask for filtering files. */
static  bool            interpret_compact_volume = false;
    /**< Interpret volume image as compact volume. */
static  bool            recursive = false;
    /**< true = pack recursively all subdirectories; false = pack only one. */
static  bool            force_recursive = false;
    /**< Recurse into subdirectories without checking if they match the mask.*/
static  bool            erase = false;
    /**< true - erase files after being packed into the volume. */
static  std::list< std::string >    sources;
    /**< List of names of all directories and files to pack. */
static  std::string     destination_path;
    /**< Path where to unpack the files to. */


/**
 * Prints the utility usage information.
 */
void print_info
    (
    const std::string & argv0
    );

/**
 * Processes the command line.
 * @return true if unpack_volume() should be processed.
 */
bool process_cmdline
    (
    int             argc,
    char **         argv
    );

/**
 * The actual volume unpacking utility.
 */
void unpack_volume();



int main
    (
    int        argc,
    char **    argv
    )
    {
    try
        {
        // Initialize massiv core.
        System::StartUpInfo startup_info( NodeId::TYPE_ANONYMOUS, "" );
        startup_info.load_config = false;
        startup_info.quiet_start = true;
        System::initialize( startup_info );
        Global::logger().enable( true );

        if( process_cmdline( argc, argv ) )
            {
            unpack_volume();
            }
        
        // Shutdown massiv core.
        System::shutdown();
        }
    catch( Exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        }
    catch( std::exception & e )
        {
        std::cout << "Error: " << e.what() << std::endl;
        }
    catch( ... )
        {
        std::cout << "Unknown exception leaked list_volume " << std::endl;
        }

    return 0;
    }


void print_info
    (
    const std::string & argv0
    )
    {
    std::cout << "Utility for unpacking files from volume." << std::endl
              << "Usage: " << argv0 << " [switches] <volume_image> <list_of_src_paths>"
              << std::endl
              << std::endl << "<volume_path>       ... specifies a image file to unpack from"
              << std::endl << "<list_of_src_paths> ... list of paths to unpack from the volume"
              << std::endl << std::endl << "Switches:"
              << std::endl << "  -d <path> ... specifies a directory where to unpack the files into"
              << std::endl << "  -m <mask> ... specifies a mask for filtering files"
              << std::endl << "  -c ... interpret volume image as CompactVolume image"
              << std::endl << "  -e ... erase source files after being packed"
              << std::endl << "  -h ... show this message"
              << std::endl << "  -r ... unpack directories recursively"
              << std::endl << "  -R ... unpack all directories recursively"
              << std::endl << "         (don't try to match subdirectory names with mask)"
              << std::endl << "  -? ... show this message" 
              << std::endl << std::endl
              << "Example: " << argv0 << " -m t* -r ../../volume.image data/source"
              << std::endl << std::endl;
    }  // print_info()


bool process_cmdline
    (
    int             argc,
    char **         argv
    )
    {
    CmdLineParser   cmd_parser( argc, argv, "cehrR?", "dm" );
    char            cswitch;
    std::string     parameter;

    try
        {
        while( ( cswitch = cmd_parser.get_next( parameter ) ) )
            {
            switch( cswitch )
                {
            case 'c':
                interpret_compact_volume = true;
                break;

            case 'd':
                destination_path = parameter;
                break;

            case 'm':
                mask = parameter;
                break;

            case 'e':
                erase = true;
                break;

            case 'r':
                recursive = true;
                break;

            case 'R':
                recursive = true;
                force_recursive = true;
                break;
        
            case 'h':
            case '?':
                print_info( argv[ 0 ] );
                return false;

            case '-':
                if( !volume_image.size() )
                    {
                    volume_image = parameter;
                    }
                else
                    {
                    sources.push_back( parameter );
                    }

                break;

            default:
                panic( "Invalid switch." );
                break;
                }
            }
        }
    catch( ExCommandLineError & e )
        {
        std::cout << "Commandline syntax error: " << e.what() 
                  << std::endl << std::endl;

        print_info( argv[ 0 ] );
        return false;
        }
    catch( Exception & e )
        {
        std::cout << "Unexpected error: " << e.what() << std::endl;
        return false;
        }

    if( !volume_image.size() )
        {
        // Some of mandatory parameters were not given
        print_info( argv[ 0 ] );
        return false;
        }

    if( !sources.size() )
        {
        sources.push_back( std::string( "" ) );
        }

    return true;
    }  // process_cmdline()


void unpack_volume()
    {
    std::auto_ptr<BasicVolume>  volume;
    std::ios_base::openmode     mode;
    
    mode = std::ios_base::in;

    if( erase )
        {
        mode |= std::ios_base::out;
        }

    if( interpret_compact_volume )
        {
        volume = std::auto_ptr<BasicVolume>
            ( new CompactVolume( volume_image, mode ) );
        }
    else
        {
        volume = std::auto_ptr<BasicVolume>
            ( new Volume( volume_image, mode ) );
        }
    
    while( !sources.empty() )
        {
        volume->unpack
            (
            destination_path, 
            sources.front(), 
            mask, 
            erase, 
            recursive, 
            force_recursive ? 
                FileIterator::ENUMERATE_ALL_DIRS :
                FileIterator::MATCH_DIRS_NAMES
            );

        sources.pop_front();
        }
    }  // unpack_volume()
